﻿using Photon.Pun;
using System.Collections;
using System.Collections.Generic;
using TMPro;
using UnityEngine;
using CymaticLabs.Unity3D.Amqp;

public class PlayerController : MonoBehaviour
{
    static public Camera playerCamera;

    [SerializeField] GameObject cameraHolder;

    [SerializeField] GameObject chatUI;

    [SerializeField] float mouseSensitivity, sprintSpeed, walkSpeed, jumpForce, smoothTime;
    
    float verticalLookRotation;
    bool grounded;
    Vector3 smoothMoveVelocity;
    Vector3 moveAmount;

    Rigidbody rb;

    PhotonView PV;

    bool saltar;

    string actualRoom;

    private void Awake()
    {
        rb = GetComponent<Rigidbody>();
        PV = GetComponent<PhotonView>();
        //var subscription = new UnityAmqpExchangeSubscription("topic_logs", exchangeType, routingKey, null, AmqpClient.Instance.UnityEventDebugExhangeMessageHandler);

    }

    // Start is called before the first frame update
    void Start()
    {
        if (!PV.IsMine)
        {
            Destroy(GetComponentInChildren<Camera>().gameObject);
            Destroy(chatUI);
            Destroy(rb);
        }
        else 
        {
            // gameObject.tag = "Player";
            GetComponent<Rigidbody>().tag = "Player";
            PlayerController.playerCamera = GetComponentInChildren<Camera>();
            TMP_InputField aux = chatUI.GetComponentInChildren<TMP_InputField>();
            aux.Select();
            aux.ActivateInputField();
        }

        var subscription = new UnityAmqpExchangeSubscription("topic_logs", AmqpExchangeTypes.Topic, "Player.Movimientos",HandleExchangeMessageReceived, null);
        //var subscription = new UnityAmqpExchangeSubscription("topic_logs", AmqpExchangeTypes.Topic, "Player.Movimientos",null, HandleExchangeMessageReceived);
        /*
            * Add the subscription to the client. If you are using multiple AmqpClient instances then
            * using the static methods won't work. In that case add a inspector property of type 'AmqpClient'
            * and assigned a reference to the connection you want to work with and call the 'SubscribeToExchange()'
            * non-static method instead.
        */
        AmqpClient.Subscribe(subscription);
        // Create a new exchange subscription using the inspector values
 
        //MessageHandler.Subscribe(ExchangeName, ExchangeType, RoutingKey, HandleMessageRecivedFinDesarrolloTarea)
    }

    // Update is called once per frame
    void Update()
    {
        if (!PV.IsMine)
            return;
        Look();
        Move();
        Jump();
    }

    void Look()
    {
        transform.Rotate(Vector3.up * Input.GetAxisRaw("Mouse X") * mouseSensitivity);

        verticalLookRotation += Input.GetAxisRaw("Mouse Y") * mouseSensitivity;
        verticalLookRotation = Mathf.Clamp(verticalLookRotation, -90f, 90f);

        cameraHolder.transform.localEulerAngles = Vector3.left * verticalLookRotation;
    }

    void Move()
    {
        Vector3 moveDir = new Vector3(Input.GetAxisRaw("Horizontal"), 0, Input.GetAxisRaw("Vertical")).normalized;
        moveAmount = Vector3.SmoothDamp(moveAmount, moveDir * (Input.GetKey(KeyCode.LeftShift) ? sprintSpeed : walkSpeed), ref smoothMoveVelocity, smoothTime);
    }

    void Jump()
    {
        if (Input.GetKeyDown(KeyCode.Space) && grounded || saltar)
        {
            rb.AddForce(transform.up * jumpForce);
            saltar=false;
        }

    }

    public void SetGroundedState(bool _grounded)
    {
        grounded = _grounded;
    }

    void FixedUpdate()
    {
        if (!PV.IsMine)
            return;
        rb.MovePosition(rb.position + transform.TransformDirection(moveAmount) * Time.fixedDeltaTime);
    }

    public void setRoom(string room)
    {
        var exchangeName = "topic_logs";
        var routingKey = "Player.Sala.Cambio";
        var message = "Cambio de la habitacion \"" + this.actualRoom + "\" a \"" + room + "\"";
        AmqpClient.Publish(exchangeName, routingKey, message);
        //Debug.Log(message);
        this.actualRoom = room;
    }

    //
    /**
    * Handles messages receieved from this object's subscription based on the exchange name,
    * exchange type, and routing key used. You could also write an anonymous delegate in line
    * when creating the subscription like: (received) => { Debug.Log(received.Message.Body.Length); }
    */
    void HandleExchangeMessageReceived(AmqpExchangeReceivedMessage received)
    {
        var receivedJson = System.Text.Encoding.UTF8.GetString(received.Message.Body);
        if(receivedJson.Equals("Saltar")){
            saltar=true;
        }
        var msg = CymaticLabs.Unity3D.Amqp.SimpleJSON.JSON.Parse(receivedJson);
        //Process the msg and make an action
        Debug.LogFormat("From AMQP : {0}",msg);
    }
}
