// The Module object: Our interface to the outside world. We import
// and export values on it, and do the work to get that through
// closure compiler if necessary. There are various ways Module can be used:
// 1. Not defined. We create it here
// 2. A function parameter, function(Module) { ..generated code.. }
// 3. pre-run appended it, var Module = {}; ..generated code..
// 4. External script tag defines var Module.
// We need to do an eval in order to handle the closure compiler
// case, where this code here is minified but Module was defined
// elsewhere (e.g. case 4 above). We also need to check if Module
// already exists (e.g. case 3 above).
// Note that if you want to run closure, and also to use Module
// after the generated code, you will need to define   var Module = {};
// before the code. Then that object will be used in the code, and you
// can continue to use Module afterwards as well.
var Module;
if (!Module) Module = (typeof Module !== 'undefined' ? Module : null) || {};

// Sometimes an existing Module object exists with properties
// meant to overwrite the default module functionality. Here
// we collect those properties and reapply _after_ we configure
// the current environment's defaults to avoid having to be so
// defensive during initialization.
var moduleOverrides = {};
for (var key in Module) {
  if (Module.hasOwnProperty(key)) {
    moduleOverrides[key] = Module[key];
  }
}

// The environment setup code below is customized to use Module.
// *** Environment setup code ***
var ENVIRONMENT_IS_WEB = false;
var ENVIRONMENT_IS_WORKER = false;
var ENVIRONMENT_IS_NODE = false;
var ENVIRONMENT_IS_SHELL = false;

// Three configurations we can be running in:
// 1) We could be the application main() thread running in the main JS UI thread. (ENVIRONMENT_IS_WORKER == false and ENVIRONMENT_IS_PTHREAD == false)
// 2) We could be the application main() thread proxied to worker. (with Emscripten -s PROXY_TO_WORKER=1) (ENVIRONMENT_IS_WORKER == true, ENVIRONMENT_IS_PTHREAD == false)
// 3) We could be an application pthread running in a worker. (ENVIRONMENT_IS_WORKER == true and ENVIRONMENT_IS_PTHREAD == true)

if (Module['ENVIRONMENT']) {
  if (Module['ENVIRONMENT'] === 'WEB') {
    ENVIRONMENT_IS_WEB = true;
  } else if (Module['ENVIRONMENT'] === 'WORKER') {
    ENVIRONMENT_IS_WORKER = true;
  } else if (Module['ENVIRONMENT'] === 'NODE') {
    ENVIRONMENT_IS_NODE = true;
  } else if (Module['ENVIRONMENT'] === 'SHELL') {
    ENVIRONMENT_IS_SHELL = true;
  } else {
    throw new Error('The provided Module[\'ENVIRONMENT\'] value is not valid. It must be one of: WEB|WORKER|NODE|SHELL.');
  }
} else {
  ENVIRONMENT_IS_WEB = typeof window === 'object';
  ENVIRONMENT_IS_WORKER = typeof importScripts === 'function';
  ENVIRONMENT_IS_NODE = typeof process === 'object' && typeof require === 'function' && !ENVIRONMENT_IS_WEB && !ENVIRONMENT_IS_WORKER;
  ENVIRONMENT_IS_SHELL = !ENVIRONMENT_IS_WEB && !ENVIRONMENT_IS_NODE && !ENVIRONMENT_IS_WORKER;
}


if (ENVIRONMENT_IS_NODE) {
  // Expose functionality in the same simple way that the shells work
  // Note that we pollute the global namespace here, otherwise we break in node
  if (!Module['print']) Module['print'] = console.log;
  if (!Module['printErr']) Module['printErr'] = console.warn;

  var nodeFS;
  var nodePath;

  Module['read'] = function shell_read(filename, binary) {
    if (!nodeFS) nodeFS = require('fs');
    if (!nodePath) nodePath = require('path');
    filename = nodePath['normalize'](filename);
    var ret = nodeFS['readFileSync'](filename);
    return binary ? ret : ret.toString();
  };

  Module['readBinary'] = function readBinary(filename) {
    var ret = Module['read'](filename, true);
    if (!ret.buffer) {
      ret = new Uint8Array(ret);
    }
    assert(ret.buffer);
    return ret;
  };

  Module['load'] = function load(f) {
    globalEval(read(f));
  };

  if (!Module['thisProgram']) {
    if (process['argv'].length > 1) {
      Module['thisProgram'] = process['argv'][1].replace(/\\/g, '/');
    } else {
      Module['thisProgram'] = 'unknown-program';
    }
  }

  Module['arguments'] = process['argv'].slice(2);

  if (typeof module !== 'undefined') {
    module['exports'] = Module;
  }

  process['on']('uncaughtException', function(ex) {
    // suppress ExitStatus exceptions from showing an error
    if (!(ex instanceof ExitStatus)) {
      throw ex;
    }
  });

  Module['inspect'] = function () { return '[Emscripten Module object]'; };
}
else if (ENVIRONMENT_IS_SHELL) {
  if (!Module['print']) Module['print'] = print;
  if (typeof printErr != 'undefined') Module['printErr'] = printErr; // not present in v8 or older sm

  if (typeof read != 'undefined') {
    Module['read'] = read;
  } else {
    Module['read'] = function shell_read() { throw 'no read() available' };
  }

  Module['readBinary'] = function readBinary(f) {
    if (typeof readbuffer === 'function') {
      return new Uint8Array(readbuffer(f));
    }
    var data = read(f, 'binary');
    assert(typeof data === 'object');
    return data;
  };

  if (typeof scriptArgs != 'undefined') {
    Module['arguments'] = scriptArgs;
  } else if (typeof arguments != 'undefined') {
    Module['arguments'] = arguments;
  }

  if (typeof quit === 'function') {
    Module['quit'] = function(status, toThrow) {
      quit(status);
    }
  }

}
else if (ENVIRONMENT_IS_WEB || ENVIRONMENT_IS_WORKER) {
  Module['read'] = function shell_read(url) {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', url, false);
    xhr.send(null);
    return xhr.responseText;
  };

  if (ENVIRONMENT_IS_WORKER) {
    Module['readBinary'] = function readBinary(url) {
      var xhr = new XMLHttpRequest();
      xhr.open('GET', url, false);
      xhr.responseType = 'arraybuffer';
      xhr.send(null);
      return new Uint8Array(xhr.response);
    };
  }

  Module['readAsync'] = function readAsync(url, onload, onerror) {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', url, true);
    xhr.responseType = 'arraybuffer';
    xhr.onload = function xhr_onload() {
      if (xhr.status == 200 || (xhr.status == 0 && xhr.response)) { // file URLs can return 0
        onload(xhr.response);
      } else {
        onerror();
      }
    };
    xhr.onerror = onerror;
    xhr.send(null);
  };

  if (typeof arguments != 'undefined') {
    Module['arguments'] = arguments;
  }

  if (typeof console !== 'undefined') {
    if (!Module['print']) Module['print'] = function shell_print(x) {
      console.log(x);
    };
    if (!Module['printErr']) Module['printErr'] = function shell_printErr(x) {
      console.warn(x);
    };
  } else {
    // Probably a worker, and without console.log. We can do very little here...
    var TRY_USE_DUMP = false;
    if (!Module['print']) Module['print'] = (TRY_USE_DUMP && (typeof(dump) !== "undefined") ? (function(x) {
      dump(x);
    }) : (function(x) {
      // self.postMessage(x); // enable this if you want stdout to be sent as messages
    }));
  }

  if (ENVIRONMENT_IS_WORKER) {
    Module['load'] = importScripts;
  }

  if (typeof Module['setWindowTitle'] === 'undefined') {
    Module['setWindowTitle'] = function(title) { document.title = title };
  }
}
else {
  // Unreachable because SHELL is dependant on the others
  throw 'Unknown runtime environment. Where are we?';
}

function globalEval(x) {
  eval.call(null, x);
}
if (!Module['load'] && Module['read']) {
  Module['load'] = function load(f) {
    globalEval(Module['read'](f));
  };
}
if (!Module['print']) {
  Module['print'] = function(){};
}
if (!Module['printErr']) {
  Module['printErr'] = Module['print'];
}
if (!Module['arguments']) {
  Module['arguments'] = [];
}
if (!Module['thisProgram']) {
  Module['thisProgram'] = './this.program';
}
if (!Module['quit']) {
  Module['quit'] = function(status, toThrow) {
    throw toThrow;
  }
}

// *** Environment setup code ***

// Closure helpers
Module.print = Module['print'];
Module.printErr = Module['printErr'];

// Callbacks
Module['preRun'] = [];
Module['postRun'] = [];

// Merge back in the overrides
for (var key in moduleOverrides) {
  if (moduleOverrides.hasOwnProperty(key)) {
    Module[key] = moduleOverrides[key];
  }
}
// Free the object hierarchy contained in the overrides, this lets the GC
// reclaim data used e.g. in memoryInitializerRequest, which is a large typed array.
moduleOverrides = undefined;



// {{PREAMBLE_ADDITIONS}}

// === Preamble library stuff ===

// Documentation for the public APIs defined in this file must be updated in:
//    site/source/docs/api_reference/preamble.js.rst
// A prebuilt local version of the documentation is available at:
//    site/build/text/docs/api_reference/preamble.js.txt
// You can also build docs locally as HTML or other formats in site/
// An online HTML version (which may be of a different version of Emscripten)
//    is up at http://kripken.github.io/emscripten-site/docs/api_reference/preamble.js.html

//========================================
// Runtime code shared with compiler
//========================================

var Runtime = {
  setTempRet0: function (value) {
    tempRet0 = value;
    return value;
  },
  getTempRet0: function () {
    return tempRet0;
  },
  stackSave: function () {
    return STACKTOP;
  },
  stackRestore: function (stackTop) {
    STACKTOP = stackTop;
  },
  getNativeTypeSize: function (type) {
    switch (type) {
      case 'i1': case 'i8': return 1;
      case 'i16': return 2;
      case 'i32': return 4;
      case 'i64': return 8;
      case 'float': return 4;
      case 'double': return 8;
      default: {
        if (type[type.length-1] === '*') {
          return Runtime.QUANTUM_SIZE; // A pointer
        } else if (type[0] === 'i') {
          var bits = parseInt(type.substr(1));
          assert(bits % 8 === 0);
          return bits/8;
        } else {
          return 0;
        }
      }
    }
  },
  getNativeFieldSize: function (type) {
    return Math.max(Runtime.getNativeTypeSize(type), Runtime.QUANTUM_SIZE);
  },
  STACK_ALIGN: 16,
  prepVararg: function (ptr, type) {
    if (type === 'double' || type === 'i64') {
      // move so the load is aligned
      if (ptr & 7) {
        assert((ptr & 7) === 4);
        ptr += 4;
      }
    } else {
      assert((ptr & 3) === 0);
    }
    return ptr;
  },
  getAlignSize: function (type, size, vararg) {
    // we align i64s and doubles on 64-bit boundaries, unlike x86
    if (!vararg && (type == 'i64' || type == 'double')) return 8;
    if (!type) return Math.min(size, 8); // align structures internally to 64 bits
    return Math.min(size || (type ? Runtime.getNativeFieldSize(type) : 0), Runtime.QUANTUM_SIZE);
  },
  dynCall: function (sig, ptr, args) {
    if (args && args.length) {
      assert(args.length == sig.length-1);
      assert(('dynCall_' + sig) in Module, 'bad function pointer type - no table for sig \'' + sig + '\'');
      return Module['dynCall_' + sig].apply(null, [ptr].concat(args));
    } else {
      assert(sig.length == 1);
      assert(('dynCall_' + sig) in Module, 'bad function pointer type - no table for sig \'' + sig + '\'');
      return Module['dynCall_' + sig].call(null, ptr);
    }
  },
  functionPointers: [null,null,null,null,null,null,null,null,null,null,null,null,null,null,null,null,null,null,null,null],
  addFunction: function (func) {
    for (var i = 0; i < Runtime.functionPointers.length; i++) {
      if (!Runtime.functionPointers[i]) {
        Runtime.functionPointers[i] = func;
        return 2*(1 + i);
      }
    }
    throw 'Finished up all reserved function pointers. Use a higher value for RESERVED_FUNCTION_POINTERS.';
  },
  removeFunction: function (index) {
    Runtime.functionPointers[(index-2)/2] = null;
  },
  warnOnce: function (text) {
    if (!Runtime.warnOnce.shown) Runtime.warnOnce.shown = {};
    if (!Runtime.warnOnce.shown[text]) {
      Runtime.warnOnce.shown[text] = 1;
      Module.printErr(text);
    }
  },
  funcWrappers: {},
  getFuncWrapper: function (func, sig) {
    assert(sig);
    if (!Runtime.funcWrappers[sig]) {
      Runtime.funcWrappers[sig] = {};
    }
    var sigCache = Runtime.funcWrappers[sig];
    if (!sigCache[func]) {
      // optimize away arguments usage in common cases
      if (sig.length === 1) {
        sigCache[func] = function dynCall_wrapper() {
          return Runtime.dynCall(sig, func);
        };
      } else if (sig.length === 2) {
        sigCache[func] = function dynCall_wrapper(arg) {
          return Runtime.dynCall(sig, func, [arg]);
        };
      } else {
        // general case
        sigCache[func] = function dynCall_wrapper() {
          return Runtime.dynCall(sig, func, Array.prototype.slice.call(arguments));
        };
      }
    }
    return sigCache[func];
  },
  getCompilerSetting: function (name) {
    throw 'You must build with -s RETAIN_COMPILER_SETTINGS=1 for Runtime.getCompilerSetting or emscripten_get_compiler_setting to work';
  },
  stackAlloc: function (size) { var ret = STACKTOP;STACKTOP = (STACKTOP + size)|0;STACKTOP = (((STACKTOP)+15)&-16);(assert((((STACKTOP|0) < (STACK_MAX|0))|0))|0); return ret; },
  staticAlloc: function (size) { var ret = STATICTOP;STATICTOP = (STATICTOP + (assert(!staticSealed),size))|0;STATICTOP = (((STATICTOP)+15)&-16); return ret; },
  dynamicAlloc: function (size) { assert(DYNAMICTOP_PTR);var ret = HEAP32[DYNAMICTOP_PTR>>2];var end = (((ret + size + 15)|0) & -16);HEAP32[DYNAMICTOP_PTR>>2] = end;if (end >= TOTAL_MEMORY) {var success = enlargeMemory();if (!success) {HEAP32[DYNAMICTOP_PTR>>2] = ret;return 0;}}return ret;},
  alignMemory: function (size,quantum) { var ret = size = Math.ceil((size)/(quantum ? quantum : 16))*(quantum ? quantum : 16); return ret; },
  makeBigInt: function (low,high,unsigned) { var ret = (unsigned ? ((+((low>>>0)))+((+((high>>>0)))*4294967296.0)) : ((+((low>>>0)))+((+((high|0)))*4294967296.0))); return ret; },
  GLOBAL_BASE: 8,
  QUANTUM_SIZE: 4,
  __dummy__: 0
}



Module["Runtime"] = Runtime;



//========================================
// Runtime essentials
//========================================

var ABORT = 0; // whether we are quitting the application. no code should run after this. set in exit() and abort()
var EXITSTATUS = 0;

/** @type {function(*, string=)} */
function assert(condition, text) {
  if (!condition) {
    abort('Assertion failed: ' + text);
  }
}

var globalScope = this;

// Returns the C function with a specified identifier (for C++, you need to do manual name mangling)
function getCFunc(ident) {
  var func = Module['_' + ident]; // closure exported function
  if (!func) {
    try { func = eval('_' + ident); } catch(e) {}
  }
  assert(func, 'Cannot call unknown function ' + ident + ' (perhaps LLVM optimizations or closure removed it?)');
  return func;
}

var cwrap, ccall;
(function(){
  var JSfuncs = {
    // Helpers for cwrap -- it can't refer to Runtime directly because it might
    // be renamed by closure, instead it calls JSfuncs['stackSave'].body to find
    // out what the minified function name is.
    'stackSave': function() {
      Runtime.stackSave()
    },
    'stackRestore': function() {
      Runtime.stackRestore()
    },
    // type conversion from js to c
    'arrayToC' : function(arr) {
      var ret = Runtime.stackAlloc(arr.length);
      writeArrayToMemory(arr, ret);
      return ret;
    },
    'stringToC' : function(str) {
      var ret = 0;
      if (str !== null && str !== undefined && str !== 0) { // null string
        // at most 4 bytes per UTF-8 code point, +1 for the trailing '\0'
        var len = (str.length << 2) + 1;
        ret = Runtime.stackAlloc(len);
        stringToUTF8(str, ret, len);
      }
      return ret;
    }
  };
  // For fast lookup of conversion functions
  var toC = {'string' : JSfuncs['stringToC'], 'array' : JSfuncs['arrayToC']};

  // C calling interface.
  ccall = function ccallFunc(ident, returnType, argTypes, args, opts) {
    var func = getCFunc(ident);
    var cArgs = [];
    var stack = 0;
    assert(returnType !== 'array', 'Return type should not be "array".');
    if (args) {
      for (var i = 0; i < args.length; i++) {
        var converter = toC[argTypes[i]];
        if (converter) {
          if (stack === 0) stack = Runtime.stackSave();
          cArgs[i] = converter(args[i]);
        } else {
          cArgs[i] = args[i];
        }
      }
    }
    var ret = func.apply(null, cArgs);
    if ((!opts || !opts.async) && typeof EmterpreterAsync === 'object') {
      assert(!EmterpreterAsync.state, 'cannot start async op with normal JS calling ccall');
    }
    if (opts && opts.async) assert(!returnType, 'async ccalls cannot return values');
    if (returnType === 'string') ret = Pointer_stringify(ret);
    if (stack !== 0) {
      if (opts && opts.async) {
        EmterpreterAsync.asyncFinalizers.push(function() {
          Runtime.stackRestore(stack);
        });
        return;
      }
      Runtime.stackRestore(stack);
    }
    return ret;
  }

  var sourceRegex = /^function\s*[a-zA-Z$_0-9]*\s*\(([^)]*)\)\s*{\s*([^*]*?)[\s;]*(?:return\s*(.*?)[;\s]*)?}$/;
  function parseJSFunc(jsfunc) {
    // Match the body and the return value of a javascript function source
    var parsed = jsfunc.toString().match(sourceRegex).slice(1);
    return {arguments : parsed[0], body : parsed[1], returnValue: parsed[2]}
  }

  // sources of useful functions. we create this lazily as it can trigger a source decompression on this entire file
  var JSsource = null;
  function ensureJSsource() {
    if (!JSsource) {
      JSsource = {};
      for (var fun in JSfuncs) {
        if (JSfuncs.hasOwnProperty(fun)) {
          // Elements of toCsource are arrays of three items:
          // the code, and the return value
          JSsource[fun] = parseJSFunc(JSfuncs[fun]);
        }
      }
    }
  }

  cwrap = function cwrap(ident, returnType, argTypes) {
    argTypes = argTypes || [];
    var cfunc = getCFunc(ident);
    // When the function takes numbers and returns a number, we can just return
    // the original function
    var numericArgs = argTypes.every(function(type){ return type === 'number'});
    var numericRet = (returnType !== 'string');
    if ( numericRet && numericArgs) {
      return cfunc;
    }
    // Creation of the arguments list (["$1","$2",...,"$nargs"])
    var argNames = argTypes.map(function(x,i){return '$'+i});
    var funcstr = "(function(" + argNames.join(',') + ") {";
    var nargs = argTypes.length;
    if (!numericArgs) {
      // Generate the code needed to convert the arguments from javascript
      // values to pointers
      ensureJSsource();
      funcstr += 'var stack = ' + JSsource['stackSave'].body + ';';
      for (var i = 0; i < nargs; i++) {
        var arg = argNames[i], type = argTypes[i];
        if (type === 'number') continue;
        var convertCode = JSsource[type + 'ToC']; // [code, return]
        funcstr += 'var ' + convertCode.arguments + ' = ' + arg + ';';
        funcstr += convertCode.body + ';';
        funcstr += arg + '=(' + convertCode.returnValue + ');';
      }
    }

    // When the code is compressed, the name of cfunc is not literally 'cfunc' anymore
    var cfuncname = parseJSFunc(function(){return cfunc}).returnValue;
    // Call the function
    funcstr += 'var ret = ' + cfuncname + '(' + argNames.join(',') + ');';
    if (!numericRet) { // Return type can only by 'string' or 'number'
      // Convert the result to a string
      var strgfy = parseJSFunc(function(){return Pointer_stringify}).returnValue;
      funcstr += 'ret = ' + strgfy + '(ret);';
    }
    funcstr += "if (typeof EmterpreterAsync === 'object') { assert(!EmterpreterAsync.state, 'cannot start async op with normal JS calling cwrap') }";
    if (!numericArgs) {
      // If we had a stack, restore it
      ensureJSsource();
      funcstr += JSsource['stackRestore'].body.replace('()', '(stack)') + ';';
    }
    funcstr += 'return ret})';
    return eval(funcstr);
  };
})();
Module["ccall"] = ccall;
Module["cwrap"] = cwrap;

/** @type {function(number, number, string, boolean=)} */
function setValue(ptr, value, type, noSafe) {
  type = type || 'i8';
  if (type.charAt(type.length-1) === '*') type = 'i32'; // pointers are 32-bit
    switch(type) {
      case 'i1': HEAP8[((ptr)>>0)]=value; break;
      case 'i8': HEAP8[((ptr)>>0)]=value; break;
      case 'i16': HEAP16[((ptr)>>1)]=value; break;
      case 'i32': HEAP32[((ptr)>>2)]=value; break;
      case 'i64': (tempI64 = [value>>>0,(tempDouble=value,(+(Math_abs(tempDouble))) >= 1.0 ? (tempDouble > 0.0 ? ((Math_min((+(Math_floor((tempDouble)/4294967296.0))), 4294967295.0))|0)>>>0 : (~~((+(Math_ceil((tempDouble - +(((~~(tempDouble)))>>>0))/4294967296.0)))))>>>0) : 0)],HEAP32[((ptr)>>2)]=tempI64[0],HEAP32[(((ptr)+(4))>>2)]=tempI64[1]); break;
      case 'float': HEAPF32[((ptr)>>2)]=value; break;
      case 'double': HEAPF64[((ptr)>>3)]=value; break;
      default: abort('invalid type for setValue: ' + type);
    }
}
Module["setValue"] = setValue;

/** @type {function(number, string, boolean=)} */
function getValue(ptr, type, noSafe) {
  type = type || 'i8';
  if (type.charAt(type.length-1) === '*') type = 'i32'; // pointers are 32-bit
    switch(type) {
      case 'i1': return HEAP8[((ptr)>>0)];
      case 'i8': return HEAP8[((ptr)>>0)];
      case 'i16': return HEAP16[((ptr)>>1)];
      case 'i32': return HEAP32[((ptr)>>2)];
      case 'i64': return HEAP32[((ptr)>>2)];
      case 'float': return HEAPF32[((ptr)>>2)];
      case 'double': return HEAPF64[((ptr)>>3)];
      default: abort('invalid type for setValue: ' + type);
    }
  return null;
}
Module["getValue"] = getValue;

var ALLOC_NORMAL = 0; // Tries to use _malloc()
var ALLOC_STACK = 1; // Lives for the duration of the current function call
var ALLOC_STATIC = 2; // Cannot be freed
var ALLOC_DYNAMIC = 3; // Cannot be freed except through sbrk
var ALLOC_NONE = 4; // Do not allocate
Module["ALLOC_NORMAL"] = ALLOC_NORMAL;
Module["ALLOC_STACK"] = ALLOC_STACK;
Module["ALLOC_STATIC"] = ALLOC_STATIC;
Module["ALLOC_DYNAMIC"] = ALLOC_DYNAMIC;
Module["ALLOC_NONE"] = ALLOC_NONE;

// allocate(): This is for internal use. You can use it yourself as well, but the interface
//             is a little tricky (see docs right below). The reason is that it is optimized
//             for multiple syntaxes to save space in generated code. So you should
//             normally not use allocate(), and instead allocate memory using _malloc(),
//             initialize it with setValue(), and so forth.
// @slab: An array of data, or a number. If a number, then the size of the block to allocate,
//        in *bytes* (note that this is sometimes confusing: the next parameter does not
//        affect this!)
// @types: Either an array of types, one for each byte (or 0 if no type at that position),
//         or a single type which is used for the entire block. This only matters if there
//         is initial data - if @slab is a number, then this does not matter at all and is
//         ignored.
// @allocator: How to allocate memory, see ALLOC_*
/** @type {function((TypedArray|Array<number>|number), string, number, number=)} */
function allocate(slab, types, allocator, ptr) {
  var zeroinit, size;
  if (typeof slab === 'number') {
    zeroinit = true;
    size = slab;
  } else {
    zeroinit = false;
    size = slab.length;
  }

  var singleType = typeof types === 'string' ? types : null;

  var ret;
  if (allocator == ALLOC_NONE) {
    ret = ptr;
  } else {
    ret = [typeof _malloc === 'function' ? _malloc : Runtime.staticAlloc, Runtime.stackAlloc, Runtime.staticAlloc, Runtime.dynamicAlloc][allocator === undefined ? ALLOC_STATIC : allocator](Math.max(size, singleType ? 1 : types.length));
  }

  if (zeroinit) {
    var ptr = ret, stop;
    assert((ret & 3) == 0);
    stop = ret + (size & ~3);
    for (; ptr < stop; ptr += 4) {
      HEAP32[((ptr)>>2)]=0;
    }
    stop = ret + size;
    while (ptr < stop) {
      HEAP8[((ptr++)>>0)]=0;
    }
    return ret;
  }

  if (singleType === 'i8') {
    if (slab.subarray || slab.slice) {
      HEAPU8.set(/** @type {!Uint8Array} */ (slab), ret);
    } else {
      HEAPU8.set(new Uint8Array(slab), ret);
    }
    return ret;
  }

  var i = 0, type, typeSize, previousType;
  while (i < size) {
    var curr = slab[i];

    if (typeof curr === 'function') {
      curr = Runtime.getFunctionIndex(curr);
    }

    type = singleType || types[i];
    if (type === 0) {
      i++;
      continue;
    }
    assert(type, 'Must know what type to store in allocate!');

    if (type == 'i64') type = 'i32'; // special case: we have one i32 here, and one i32 later

    setValue(ret+i, curr, type);

    // no need to look up size unless type changes, so cache it
    if (previousType !== type) {
      typeSize = Runtime.getNativeTypeSize(type);
      previousType = type;
    }
    i += typeSize;
  }

  return ret;
}
Module["allocate"] = allocate;

// Allocate memory during any stage of startup - static memory early on, dynamic memory later, malloc when ready
function getMemory(size) {
  if (!staticSealed) return Runtime.staticAlloc(size);
  if (!runtimeInitialized) return Runtime.dynamicAlloc(size);
  return _malloc(size);
}
Module["getMemory"] = getMemory;

/** @type {function(number, number=)} */
function Pointer_stringify(ptr, length) {
  if (length === 0 || !ptr) return '';
  // TODO: use TextDecoder
  // Find the length, and check for UTF while doing so
  var hasUtf = 0;
  var t;
  var i = 0;
  while (1) {
    assert(ptr + i < TOTAL_MEMORY);
    t = HEAPU8[(((ptr)+(i))>>0)];
    hasUtf |= t;
    if (t == 0 && !length) break;
    i++;
    if (length && i == length) break;
  }
  if (!length) length = i;

  var ret = '';

  if (hasUtf < 128) {
    var MAX_CHUNK = 1024; // split up into chunks, because .apply on a huge string can overflow the stack
    var curr;
    while (length > 0) {
      curr = String.fromCharCode.apply(String, HEAPU8.subarray(ptr, ptr + Math.min(length, MAX_CHUNK)));
      ret = ret ? ret + curr : curr;
      ptr += MAX_CHUNK;
      length -= MAX_CHUNK;
    }
    return ret;
  }
  return Module['UTF8ToString'](ptr);
}
Module["Pointer_stringify"] = Pointer_stringify;

// Given a pointer 'ptr' to a null-terminated ASCII-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

function AsciiToString(ptr) {
  var str = '';
  while (1) {
    var ch = HEAP8[((ptr++)>>0)];
    if (!ch) return str;
    str += String.fromCharCode(ch);
  }
}
Module["AsciiToString"] = AsciiToString;

// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in ASCII form. The copy will require at most str.length+1 bytes of space in the HEAP.

function stringToAscii(str, outPtr) {
  return writeAsciiToMemory(str, outPtr, false);
}
Module["stringToAscii"] = stringToAscii;

// Given a pointer 'ptr' to a null-terminated UTF8-encoded string in the given array that contains uint8 values, returns
// a copy of that string as a Javascript String object.

var UTF8Decoder = typeof TextDecoder !== 'undefined' ? new TextDecoder('utf8') : undefined;
function UTF8ArrayToString(u8Array, idx) {
  var endPtr = idx;
  // TextDecoder needs to know the byte length in advance, it doesn't stop on null terminator by itself.
  // Also, use the length info to avoid running tiny strings through TextDecoder, since .subarray() allocates garbage.
  while (u8Array[endPtr]) ++endPtr;

  if (endPtr - idx > 16 && u8Array.subarray && UTF8Decoder) {
    return UTF8Decoder.decode(u8Array.subarray(idx, endPtr));
  } else {
    var u0, u1, u2, u3, u4, u5;

    var str = '';
    while (1) {
      // For UTF8 byte structure, see http://en.wikipedia.org/wiki/UTF-8#Description and https://www.ietf.org/rfc/rfc2279.txt and https://tools.ietf.org/html/rfc3629
      u0 = u8Array[idx++];
      if (!u0) return str;
      if (!(u0 & 0x80)) { str += String.fromCharCode(u0); continue; }
      u1 = u8Array[idx++] & 63;
      if ((u0 & 0xE0) == 0xC0) { str += String.fromCharCode(((u0 & 31) << 6) | u1); continue; }
      u2 = u8Array[idx++] & 63;
      if ((u0 & 0xF0) == 0xE0) {
        u0 = ((u0 & 15) << 12) | (u1 << 6) | u2;
      } else {
        u3 = u8Array[idx++] & 63;
        if ((u0 & 0xF8) == 0xF0) {
          u0 = ((u0 & 7) << 18) | (u1 << 12) | (u2 << 6) | u3;
        } else {
          u4 = u8Array[idx++] & 63;
          if ((u0 & 0xFC) == 0xF8) {
            u0 = ((u0 & 3) << 24) | (u1 << 18) | (u2 << 12) | (u3 << 6) | u4;
          } else {
            u5 = u8Array[idx++] & 63;
            u0 = ((u0 & 1) << 30) | (u1 << 24) | (u2 << 18) | (u3 << 12) | (u4 << 6) | u5;
          }
        }
      }
      if (u0 < 0x10000) {
        str += String.fromCharCode(u0);
      } else {
        var ch = u0 - 0x10000;
        str += String.fromCharCode(0xD800 | (ch >> 10), 0xDC00 | (ch & 0x3FF));
      }
    }
  }
}
Module["UTF8ArrayToString"] = UTF8ArrayToString;

// Given a pointer 'ptr' to a null-terminated UTF8-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

function UTF8ToString(ptr) {
  return UTF8ArrayToString(HEAPU8,ptr);
}
Module["UTF8ToString"] = UTF8ToString;

// Copies the given Javascript String object 'str' to the given byte array at address 'outIdx',
// encoded in UTF8 form and null-terminated. The copy will require at most str.length*4+1 bytes of space in the HEAP.
// Use the function lengthBytesUTF8 to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outU8Array: the array to copy to. Each index in this array is assumed to be one 8-byte element.
//   outIdx: The starting offset in the array to begin the copying.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null
//                    terminator, i.e. if maxBytesToWrite=1, only the null terminator will be written and nothing else.
//                    maxBytesToWrite=0 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF8Array(str, outU8Array, outIdx, maxBytesToWrite) {
  if (!(maxBytesToWrite > 0)) // Parameter maxBytesToWrite is not optional. Negative values, 0, null, undefined and false each don't write out any bytes.
    return 0;

  var startIdx = outIdx;
  var endIdx = outIdx + maxBytesToWrite - 1; // -1 for string null terminator.
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! So decode UTF16->UTF32->UTF8.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    // For UTF8 byte structure, see http://en.wikipedia.org/wiki/UTF-8#Description and https://www.ietf.org/rfc/rfc2279.txt and https://tools.ietf.org/html/rfc3629
    var u = str.charCodeAt(i); // possibly a lead surrogate
    if (u >= 0xD800 && u <= 0xDFFF) u = 0x10000 + ((u & 0x3FF) << 10) | (str.charCodeAt(++i) & 0x3FF);
    if (u <= 0x7F) {
      if (outIdx >= endIdx) break;
      outU8Array[outIdx++] = u;
    } else if (u <= 0x7FF) {
      if (outIdx + 1 >= endIdx) break;
      outU8Array[outIdx++] = 0xC0 | (u >> 6);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0xFFFF) {
      if (outIdx + 2 >= endIdx) break;
      outU8Array[outIdx++] = 0xE0 | (u >> 12);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0x1FFFFF) {
      if (outIdx + 3 >= endIdx) break;
      outU8Array[outIdx++] = 0xF0 | (u >> 18);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else if (u <= 0x3FFFFFF) {
      if (outIdx + 4 >= endIdx) break;
      outU8Array[outIdx++] = 0xF8 | (u >> 24);
      outU8Array[outIdx++] = 0x80 | ((u >> 18) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    } else {
      if (outIdx + 5 >= endIdx) break;
      outU8Array[outIdx++] = 0xFC | (u >> 30);
      outU8Array[outIdx++] = 0x80 | ((u >> 24) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 18) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 12) & 63);
      outU8Array[outIdx++] = 0x80 | ((u >> 6) & 63);
      outU8Array[outIdx++] = 0x80 | (u & 63);
    }
  }
  // Null-terminate the pointer to the buffer.
  outU8Array[outIdx] = 0;
  return outIdx - startIdx;
}
Module["stringToUTF8Array"] = stringToUTF8Array;

// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF8 form. The copy will require at most str.length*4+1 bytes of space in the HEAP.
// Use the function lengthBytesUTF8 to compute the exact number of bytes (excluding null terminator) that this function will write.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF8(str, outPtr, maxBytesToWrite) {
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF8(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  return stringToUTF8Array(str, HEAPU8,outPtr, maxBytesToWrite);
}
Module["stringToUTF8"] = stringToUTF8;

// Returns the number of bytes the given Javascript string takes if encoded as a UTF8 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF8(str) {
  var len = 0;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! So decode UTF16->UTF32->UTF8.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var u = str.charCodeAt(i); // possibly a lead surrogate
    if (u >= 0xD800 && u <= 0xDFFF) u = 0x10000 + ((u & 0x3FF) << 10) | (str.charCodeAt(++i) & 0x3FF);
    if (u <= 0x7F) {
      ++len;
    } else if (u <= 0x7FF) {
      len += 2;
    } else if (u <= 0xFFFF) {
      len += 3;
    } else if (u <= 0x1FFFFF) {
      len += 4;
    } else if (u <= 0x3FFFFFF) {
      len += 5;
    } else {
      len += 6;
    }
  }
  return len;
}
Module["lengthBytesUTF8"] = lengthBytesUTF8;

// Given a pointer 'ptr' to a null-terminated UTF16LE-encoded string in the emscripten HEAP, returns
// a copy of that string as a Javascript String object.

var UTF16Decoder = typeof TextDecoder !== 'undefined' ? new TextDecoder('utf-16le') : undefined;
function UTF16ToString(ptr) {
  assert(ptr % 2 == 0, 'Pointer passed to UTF16ToString must be aligned to two bytes!');
  var endPtr = ptr;
  // TextDecoder needs to know the byte length in advance, it doesn't stop on null terminator by itself.
  // Also, use the length info to avoid running tiny strings through TextDecoder, since .subarray() allocates garbage.
  var idx = endPtr >> 1;
  while (HEAP16[idx]) ++idx;
  endPtr = idx << 1;

  if (endPtr - ptr > 32 && UTF16Decoder) {
    return UTF16Decoder.decode(HEAPU8.subarray(ptr, endPtr));
  } else {
    var i = 0;

    var str = '';
    while (1) {
      var codeUnit = HEAP16[(((ptr)+(i*2))>>1)];
      if (codeUnit == 0) return str;
      ++i;
      // fromCharCode constructs a character from a UTF-16 code unit, so we can pass the UTF16 string right through.
      str += String.fromCharCode(codeUnit);
    }
  }
}


// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF16 form. The copy will require at most str.length*4+2 bytes of space in the HEAP.
// Use the function lengthBytesUTF16() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outPtr: Byte address in Emscripten HEAP where to write the string to.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null
//                    terminator, i.e. if maxBytesToWrite=2, only the null terminator will be written and nothing else.
//                    maxBytesToWrite<2 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF16(str, outPtr, maxBytesToWrite) {
  assert(outPtr % 2 == 0, 'Pointer passed to stringToUTF16 must be aligned to two bytes!');
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF16(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  // Backwards compatibility: if max bytes is not specified, assume unsafe unbounded write is allowed.
  if (maxBytesToWrite === undefined) {
    maxBytesToWrite = 0x7FFFFFFF;
  }
  if (maxBytesToWrite < 2) return 0;
  maxBytesToWrite -= 2; // Null terminator.
  var startPtr = outPtr;
  var numCharsToWrite = (maxBytesToWrite < str.length*2) ? (maxBytesToWrite / 2) : str.length;
  for (var i = 0; i < numCharsToWrite; ++i) {
    // charCodeAt returns a UTF-16 encoded code unit, so it can be directly written to the HEAP.
    var codeUnit = str.charCodeAt(i); // possibly a lead surrogate
    HEAP16[((outPtr)>>1)]=codeUnit;
    outPtr += 2;
  }
  // Null-terminate the pointer to the HEAP.
  HEAP16[((outPtr)>>1)]=0;
  return outPtr - startPtr;
}


// Returns the number of bytes the given Javascript string takes if encoded as a UTF16 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF16(str) {
  return str.length*2;
}


function UTF32ToString(ptr) {
  assert(ptr % 4 == 0, 'Pointer passed to UTF32ToString must be aligned to four bytes!');
  var i = 0;

  var str = '';
  while (1) {
    var utf32 = HEAP32[(((ptr)+(i*4))>>2)];
    if (utf32 == 0)
      return str;
    ++i;
    // Gotcha: fromCharCode constructs a character from a UTF-16 encoded code (pair), not from a Unicode code point! So encode the code point to UTF-16 for constructing.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    if (utf32 >= 0x10000) {
      var ch = utf32 - 0x10000;
      str += String.fromCharCode(0xD800 | (ch >> 10), 0xDC00 | (ch & 0x3FF));
    } else {
      str += String.fromCharCode(utf32);
    }
  }
}


// Copies the given Javascript String object 'str' to the emscripten HEAP at address 'outPtr',
// null-terminated and encoded in UTF32 form. The copy will require at most str.length*4+4 bytes of space in the HEAP.
// Use the function lengthBytesUTF32() to compute the exact number of bytes (excluding null terminator) that this function will write.
// Parameters:
//   str: the Javascript string to copy.
//   outPtr: Byte address in Emscripten HEAP where to write the string to.
//   maxBytesToWrite: The maximum number of bytes this function can write to the array. This count should include the null
//                    terminator, i.e. if maxBytesToWrite=4, only the null terminator will be written and nothing else.
//                    maxBytesToWrite<4 does not write any bytes to the output, not even the null terminator.
// Returns the number of bytes written, EXCLUDING the null terminator.

function stringToUTF32(str, outPtr, maxBytesToWrite) {
  assert(outPtr % 4 == 0, 'Pointer passed to stringToUTF32 must be aligned to four bytes!');
  assert(typeof maxBytesToWrite == 'number', 'stringToUTF32(str, outPtr, maxBytesToWrite) is missing the third parameter that specifies the length of the output buffer!');
  // Backwards compatibility: if max bytes is not specified, assume unsafe unbounded write is allowed.
  if (maxBytesToWrite === undefined) {
    maxBytesToWrite = 0x7FFFFFFF;
  }
  if (maxBytesToWrite < 4) return 0;
  var startPtr = outPtr;
  var endPtr = startPtr + maxBytesToWrite - 4;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! We must decode the string to UTF-32 to the heap.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var codeUnit = str.charCodeAt(i); // possibly a lead surrogate
    if (codeUnit >= 0xD800 && codeUnit <= 0xDFFF) {
      var trailSurrogate = str.charCodeAt(++i);
      codeUnit = 0x10000 + ((codeUnit & 0x3FF) << 10) | (trailSurrogate & 0x3FF);
    }
    HEAP32[((outPtr)>>2)]=codeUnit;
    outPtr += 4;
    if (outPtr + 4 > endPtr) break;
  }
  // Null-terminate the pointer to the HEAP.
  HEAP32[((outPtr)>>2)]=0;
  return outPtr - startPtr;
}


// Returns the number of bytes the given Javascript string takes if encoded as a UTF16 byte array, EXCLUDING the null terminator byte.

function lengthBytesUTF32(str) {
  var len = 0;
  for (var i = 0; i < str.length; ++i) {
    // Gotcha: charCodeAt returns a 16-bit word that is a UTF-16 encoded code unit, not a Unicode code point of the character! We must decode the string to UTF-32 to the heap.
    // See http://unicode.org/faq/utf_bom.html#utf16-3
    var codeUnit = str.charCodeAt(i);
    if (codeUnit >= 0xD800 && codeUnit <= 0xDFFF) ++i; // possibly a lead surrogate, so skip over the tail surrogate.
    len += 4;
  }

  return len;
}


function demangle(func) {
  var __cxa_demangle_func = Module['___cxa_demangle'] || Module['__cxa_demangle'];
  if (__cxa_demangle_func) {
    try {
      var s =
        func.substr(1);
      var len = lengthBytesUTF8(s)+1;
      var buf = _malloc(len);
      stringToUTF8(s, buf, len);
      var status = _malloc(4);
      var ret = __cxa_demangle_func(buf, 0, 0, status);
      if (getValue(status, 'i32') === 0 && ret) {
        return Pointer_stringify(ret);
      }
      // otherwise, libcxxabi failed
    } catch(e) {
      // ignore problems here
    } finally {
      if (buf) _free(buf);
      if (status) _free(status);
      if (ret) _free(ret);
    }
    // failure when using libcxxabi, don't demangle
    return func;
  }
  Runtime.warnOnce('warning: build with  -s DEMANGLE_SUPPORT=1  to link in libcxxabi demangling');
  return func;
}

function demangleAll(text) {
  var regex =
    /__Z[\w\d_]+/g;
  return text.replace(regex,
    function(x) {
      var y = demangle(x);
      return x === y ? x : (x + ' [' + y + ']');
    });
}

function jsStackTrace() {
  var err = new Error();
  if (!err.stack) {
    // IE10+ special cases: It does have callstack info, but it is only populated if an Error object is thrown,
    // so try that as a special-case.
    try {
      throw new Error(0);
    } catch(e) {
      err = e;
    }
    if (!err.stack) {
      return '(no stack trace available)';
    }
  }
  return err.stack.toString();
}

function stackTrace() {
  var js = jsStackTrace();
  if (Module['extraStackTrace']) js += '\n' + Module['extraStackTrace']();
  return demangleAll(js);
}
Module["stackTrace"] = stackTrace;

// Memory management

var PAGE_SIZE = 16384;
var WASM_PAGE_SIZE = 65536;
var ASMJS_PAGE_SIZE = 16777216;
var MIN_TOTAL_MEMORY = 16777216;

function alignUp(x, multiple) {
  if (x % multiple > 0) {
    x += multiple - (x % multiple);
  }
  return x;
}

var HEAP,
/** @type {ArrayBuffer} */
  buffer,
/** @type {Int8Array} */
  HEAP8,
/** @type {Uint8Array} */
  HEAPU8,
/** @type {Int16Array} */
  HEAP16,
/** @type {Uint16Array} */
  HEAPU16,
/** @type {Int32Array} */
  HEAP32,
/** @type {Uint32Array} */
  HEAPU32,
/** @type {Float32Array} */
  HEAPF32,
/** @type {Float64Array} */
  HEAPF64;

function updateGlobalBuffer(buf) {
  Module['buffer'] = buffer = buf;
}

function updateGlobalBufferViews() {
  Module['HEAP8'] = HEAP8 = new Int8Array(buffer);
  Module['HEAP16'] = HEAP16 = new Int16Array(buffer);
  Module['HEAP32'] = HEAP32 = new Int32Array(buffer);
  Module['HEAPU8'] = HEAPU8 = new Uint8Array(buffer);
  Module['HEAPU16'] = HEAPU16 = new Uint16Array(buffer);
  Module['HEAPU32'] = HEAPU32 = new Uint32Array(buffer);
  Module['HEAPF32'] = HEAPF32 = new Float32Array(buffer);
  Module['HEAPF64'] = HEAPF64 = new Float64Array(buffer);
}

var STATIC_BASE, STATICTOP, staticSealed; // static area
var STACK_BASE, STACKTOP, STACK_MAX; // stack area
var DYNAMIC_BASE, DYNAMICTOP_PTR; // dynamic area handled by sbrk

  STATIC_BASE = STATICTOP = STACK_BASE = STACKTOP = STACK_MAX = DYNAMIC_BASE = DYNAMICTOP_PTR = 0;
  staticSealed = false;


// Initializes the stack cookie. Called at the startup of main and at the startup of each thread in pthreads mode.
function writeStackCookie() {
  assert((STACK_MAX & 3) == 0);
  HEAPU32[(STACK_MAX >> 2)-1] = 0x02135467;
  HEAPU32[(STACK_MAX >> 2)-2] = 0x89BACDFE;
}

function checkStackCookie() {
  if (HEAPU32[(STACK_MAX >> 2)-1] != 0x02135467 || HEAPU32[(STACK_MAX >> 2)-2] != 0x89BACDFE) {
    abort('Stack overflow! Stack cookie has been overwritten, expected hex dwords 0x89BACDFE and 0x02135467, but received 0x' + HEAPU32[(STACK_MAX >> 2)-2].toString(16) + ' ' + HEAPU32[(STACK_MAX >> 2)-1].toString(16));
  }
  // Also test the global address 0 for integrity. This check is not compatible with SAFE_SPLIT_MEMORY though, since that mode already tests all address 0 accesses on its own.
  if (HEAP32[0] !== 0x63736d65 /* 'emsc' */) throw 'Runtime error: The application has corrupted its heap memory area (address zero)!';
}

function abortStackOverflow(allocSize) {
  abort('Stack overflow! Attempted to allocate ' + allocSize + ' bytes on the stack, but stack has only ' + (STACK_MAX - Module['asm'].stackSave() + allocSize) + ' bytes available!');
}

function abortOnCannotGrowMemory() {
  abort('Cannot enlarge memory arrays. Either (1) compile with  -s TOTAL_MEMORY=X  with X higher than the current value ' + TOTAL_MEMORY + ', (2) compile with  -s ALLOW_MEMORY_GROWTH=1  which allows increasing the size at runtime but prevents some optimizations, (3) set Module.TOTAL_MEMORY to a higher value before the program runs, or (4) if you want malloc to return NULL (0) instead of this abort, compile with  -s ABORTING_MALLOC=0 ');
}

if (!Module['reallocBuffer']) Module['reallocBuffer'] = function(size) {
  var ret;
  try {
    if (ArrayBuffer.transfer) {
      ret = ArrayBuffer.transfer(buffer, size);
    } else {
      var oldHEAP8 = HEAP8;
      ret = new ArrayBuffer(size);
      var temp = new Int8Array(ret);
      temp.set(oldHEAP8);
    }
  } catch(e) {
    return false;
  }
  var success = _emscripten_replace_memory(ret);
  if (!success) return false;
  return ret;
};

function enlargeMemory() {
  // TOTAL_MEMORY is the current size of the actual array, and DYNAMICTOP is the new top.
  assert(HEAP32[DYNAMICTOP_PTR>>2] > TOTAL_MEMORY); // This function should only ever be called after the ceiling of the dynamic heap has already been bumped to exceed the current total size of the asm.js heap.


  var PAGE_MULTIPLE = Module["usingWasm"] ? WASM_PAGE_SIZE : ASMJS_PAGE_SIZE; // In wasm, heap size must be a multiple of 64KB. In asm.js, they need to be multiples of 16MB.
  var LIMIT = 2147483648 - PAGE_MULTIPLE; // We can do one page short of 2GB as theoretical maximum.

  if (HEAP32[DYNAMICTOP_PTR>>2] > LIMIT) {
    Module.printErr('Cannot enlarge memory, asked to go up to ' + HEAP32[DYNAMICTOP_PTR>>2] + ' bytes, but the limit is ' + LIMIT + ' bytes!');
    return false;
  }

  var OLD_TOTAL_MEMORY = TOTAL_MEMORY;
  TOTAL_MEMORY = Math.max(TOTAL_MEMORY, MIN_TOTAL_MEMORY); // So the loop below will not be infinite, and minimum asm.js memory size is 16MB.

  while (TOTAL_MEMORY < HEAP32[DYNAMICTOP_PTR>>2]) { // Keep incrementing the heap size as long as it's less than what is requested.
    if (TOTAL_MEMORY <= 536870912) {
      TOTAL_MEMORY = alignUp(2 * TOTAL_MEMORY, PAGE_MULTIPLE); // Simple heuristic: double until 1GB...
    } else {
      TOTAL_MEMORY = Math.min(alignUp((3 * TOTAL_MEMORY + 2147483648) / 4, PAGE_MULTIPLE), LIMIT); // ..., but after that, add smaller increments towards 2GB, which we cannot reach
    }
  }

  var start = Date.now();

  var replacement = Module['reallocBuffer'](TOTAL_MEMORY);
  if (!replacement || replacement.byteLength != TOTAL_MEMORY) {
    Module.printErr('Failed to grow the heap from ' + OLD_TOTAL_MEMORY + ' bytes to ' + TOTAL_MEMORY + ' bytes, not enough memory!');
    if (replacement) {
      Module.printErr('Expected to get back a buffer of size ' + TOTAL_MEMORY + ' bytes, but instead got back a buffer of size ' + replacement.byteLength);
    }
    // restore the state to before this call, we failed
    TOTAL_MEMORY = OLD_TOTAL_MEMORY;
    return false;
  }

  // everything worked

  updateGlobalBuffer(replacement);
  updateGlobalBufferViews();

  Module.printErr('enlarged memory arrays from ' + OLD_TOTAL_MEMORY + ' to ' + TOTAL_MEMORY + ', took ' + (Date.now() - start) + ' ms (has ArrayBuffer.transfer? ' + (!!ArrayBuffer.transfer) + ')');

  if (!Module["usingWasm"]) {
    Module.printErr('Warning: Enlarging memory arrays, this is not fast! ' + [OLD_TOTAL_MEMORY, TOTAL_MEMORY]);
  }


  return true;
}

var byteLength;
try {
  byteLength = Function.prototype.call.bind(Object.getOwnPropertyDescriptor(ArrayBuffer.prototype, 'byteLength').get);
  byteLength(new ArrayBuffer(4)); // can fail on older ie
} catch(e) { // can fail on older node/v8
  byteLength = function(buffer) { return buffer.byteLength; };
}

var TOTAL_STACK = Module['TOTAL_STACK'] || 5242880;
var TOTAL_MEMORY = Module['TOTAL_MEMORY'] || 16777216;
if (TOTAL_MEMORY < TOTAL_STACK) Module.printErr('TOTAL_MEMORY should be larger than TOTAL_STACK, was ' + TOTAL_MEMORY + '! (TOTAL_STACK=' + TOTAL_STACK + ')');

// Initialize the runtime's memory
// check for full engine support (use string 'subarray' to avoid closure compiler confusion)
assert(typeof Int32Array !== 'undefined' && typeof Float64Array !== 'undefined' && Int32Array.prototype.subarray !== undefined && Int32Array.prototype.set !== undefined,
       'JS engine does not provide full typed array support');



// Use a provided buffer, if there is one, or else allocate a new one
if (Module['buffer']) {
  buffer = Module['buffer'];
  assert(buffer.byteLength === TOTAL_MEMORY, 'provided buffer should be ' + TOTAL_MEMORY + ' bytes, but it is ' + buffer.byteLength);
} else {
  // Use a WebAssembly memory where available
  {
    buffer = new ArrayBuffer(TOTAL_MEMORY);
  }
  assert(buffer.byteLength === TOTAL_MEMORY);
}
updateGlobalBufferViews();


function getTotalMemory() {
  return TOTAL_MEMORY;
}

// Endianness check (note: assumes compiler arch was little-endian)
  HEAP32[0] = 0x63736d65; /* 'emsc' */
HEAP16[1] = 0x6373;
if (HEAPU8[2] !== 0x73 || HEAPU8[3] !== 0x63) throw 'Runtime error: expected the system to be little-endian!';

Module['HEAP'] = HEAP;
Module['buffer'] = buffer;
Module['HEAP8'] = HEAP8;
Module['HEAP16'] = HEAP16;
Module['HEAP32'] = HEAP32;
Module['HEAPU8'] = HEAPU8;
Module['HEAPU16'] = HEAPU16;
Module['HEAPU32'] = HEAPU32;
Module['HEAPF32'] = HEAPF32;
Module['HEAPF64'] = HEAPF64;

function callRuntimeCallbacks(callbacks) {
  while(callbacks.length > 0) {
    var callback = callbacks.shift();
    if (typeof callback == 'function') {
      callback();
      continue;
    }
    var func = callback.func;
    if (typeof func === 'number') {
      if (callback.arg === undefined) {
        Module['dynCall_v'](func);
      } else {
        Module['dynCall_vi'](func, callback.arg);
      }
    } else {
      func(callback.arg === undefined ? null : callback.arg);
    }
  }
}

var __ATPRERUN__  = []; // functions called before the runtime is initialized
var __ATINIT__    = []; // functions called during startup
var __ATMAIN__    = []; // functions called when main() is to be run
var __ATEXIT__    = []; // functions called during shutdown
var __ATPOSTRUN__ = []; // functions called after the runtime has exited

var runtimeInitialized = false;
var runtimeExited = false;


function preRun() {
  // compatibility - merge in anything from Module['preRun'] at this time
  if (Module['preRun']) {
    if (typeof Module['preRun'] == 'function') Module['preRun'] = [Module['preRun']];
    while (Module['preRun'].length) {
      addOnPreRun(Module['preRun'].shift());
    }
  }
  callRuntimeCallbacks(__ATPRERUN__);
}

function ensureInitRuntime() {
  checkStackCookie();
  if (runtimeInitialized) return;
  runtimeInitialized = true;
  callRuntimeCallbacks(__ATINIT__);
}

function preMain() {
  checkStackCookie();
  callRuntimeCallbacks(__ATMAIN__);
}

function exitRuntime() {
  checkStackCookie();
  callRuntimeCallbacks(__ATEXIT__);
  runtimeExited = true;
}

function postRun() {
  checkStackCookie();
  // compatibility - merge in anything from Module['postRun'] at this time
  if (Module['postRun']) {
    if (typeof Module['postRun'] == 'function') Module['postRun'] = [Module['postRun']];
    while (Module['postRun'].length) {
      addOnPostRun(Module['postRun'].shift());
    }
  }
  callRuntimeCallbacks(__ATPOSTRUN__);
}

function addOnPreRun(cb) {
  __ATPRERUN__.unshift(cb);
}
Module["addOnPreRun"] = addOnPreRun;

function addOnInit(cb) {
  __ATINIT__.unshift(cb);
}
Module["addOnInit"] = addOnInit;

function addOnPreMain(cb) {
  __ATMAIN__.unshift(cb);
}
Module["addOnPreMain"] = addOnPreMain;

function addOnExit(cb) {
  __ATEXIT__.unshift(cb);
}
Module["addOnExit"] = addOnExit;

function addOnPostRun(cb) {
  __ATPOSTRUN__.unshift(cb);
}
Module["addOnPostRun"] = addOnPostRun;

// Tools

/** @type {function(string, boolean=, number=)} */
function intArrayFromString(stringy, dontAddNull, length) {
  var len = length > 0 ? length : lengthBytesUTF8(stringy)+1;
  var u8array = new Array(len);
  var numBytesWritten = stringToUTF8Array(stringy, u8array, 0, u8array.length);
  if (dontAddNull) u8array.length = numBytesWritten;
  return u8array;
}
Module["intArrayFromString"] = intArrayFromString;

function intArrayToString(array) {
  var ret = [];
  for (var i = 0; i < array.length; i++) {
    var chr = array[i];
    if (chr > 0xFF) {
      assert(false, 'Character code ' + chr + ' (' + String.fromCharCode(chr) + ')  at offset ' + i + ' not in 0x00-0xFF.');
      chr &= 0xFF;
    }
    ret.push(String.fromCharCode(chr));
  }
  return ret.join('');
}
Module["intArrayToString"] = intArrayToString;

// Deprecated: This function should not be called because it is unsafe and does not provide
// a maximum length limit of how many bytes it is allowed to write. Prefer calling the
// function stringToUTF8Array() instead, which takes in a maximum length that can be used
// to be secure from out of bounds writes.
/** @deprecated */
function writeStringToMemory(string, buffer, dontAddNull) {
  Runtime.warnOnce('writeStringToMemory is deprecated and should not be called! Use stringToUTF8() instead!');

  var /** @type {number} */ lastChar, /** @type {number} */ end;
  if (dontAddNull) {
    // stringToUTF8Array always appends null. If we don't want to do that, remember the
    // character that existed at the location where the null will be placed, and restore
    // that after the write (below).
    end = buffer + lengthBytesUTF8(string);
    lastChar = HEAP8[end];
  }
  stringToUTF8(string, buffer, Infinity);
  if (dontAddNull) HEAP8[end] = lastChar; // Restore the value under the null character.
}
Module["writeStringToMemory"] = writeStringToMemory;

function writeArrayToMemory(array, buffer) {
  assert(array.length >= 0, 'writeArrayToMemory array must have a length (should be an array or typed array)')
  HEAP8.set(array, buffer);
}
Module["writeArrayToMemory"] = writeArrayToMemory;

function writeAsciiToMemory(str, buffer, dontAddNull) {
  for (var i = 0; i < str.length; ++i) {
    assert(str.charCodeAt(i) === str.charCodeAt(i)&0xff);
    HEAP8[((buffer++)>>0)]=str.charCodeAt(i);
  }
  // Null-terminate the pointer to the HEAP.
  if (!dontAddNull) HEAP8[((buffer)>>0)]=0;
}
Module["writeAsciiToMemory"] = writeAsciiToMemory;

function unSign(value, bits, ignore) {
  if (value >= 0) {
    return value;
  }
  return bits <= 32 ? 2*Math.abs(1 << (bits-1)) + value // Need some trickery, since if bits == 32, we are right at the limit of the bits JS uses in bitshifts
                    : Math.pow(2, bits)         + value;
}
function reSign(value, bits, ignore) {
  if (value <= 0) {
    return value;
  }
  var half = bits <= 32 ? Math.abs(1 << (bits-1)) // abs is needed if bits == 32
                        : Math.pow(2, bits-1);
  if (value >= half && (bits <= 32 || value > half)) { // for huge values, we can hit the precision limit and always get true here. so don't do that
                                                       // but, in general there is no perfect solution here. With 64-bit ints, we get rounding and errors
                                                       // TODO: In i64 mode 1, resign the two parts separately and safely
    value = -2*half + value; // Cannot bitshift half, as it may be at the limit of the bits JS uses in bitshifts
  }
  return value;
}

// check for imul support, and also for correctness ( https://bugs.webkit.org/show_bug.cgi?id=126345 )
if (!Math['imul'] || Math['imul'](0xffffffff, 5) !== -5) Math['imul'] = function imul(a, b) {
  var ah  = a >>> 16;
  var al = a & 0xffff;
  var bh  = b >>> 16;
  var bl = b & 0xffff;
  return (al*bl + ((ah*bl + al*bh) << 16))|0;
};
Math.imul = Math['imul'];


if (!Math['clz32']) Math['clz32'] = function(x) {
  x = x >>> 0;
  for (var i = 0; i < 32; i++) {
    if (x & (1 << (31 - i))) return i;
  }
  return 32;
};
Math.clz32 = Math['clz32']

if (!Math['trunc']) Math['trunc'] = function(x) {
  return x < 0 ? Math.ceil(x) : Math.floor(x);
};
Math.trunc = Math['trunc'];

var Math_abs = Math.abs;
var Math_cos = Math.cos;
var Math_sin = Math.sin;
var Math_tan = Math.tan;
var Math_acos = Math.acos;
var Math_asin = Math.asin;
var Math_atan = Math.atan;
var Math_atan2 = Math.atan2;
var Math_exp = Math.exp;
var Math_log = Math.log;
var Math_sqrt = Math.sqrt;
var Math_ceil = Math.ceil;
var Math_floor = Math.floor;
var Math_pow = Math.pow;
var Math_imul = Math.imul;
var Math_fround = Math.fround;
var Math_round = Math.round;
var Math_min = Math.min;
var Math_clz32 = Math.clz32;
var Math_trunc = Math.trunc;

// A counter of dependencies for calling run(). If we need to
// do asynchronous work before running, increment this and
// decrement it. Incrementing must happen in a place like
// PRE_RUN_ADDITIONS (used by emcc to add file preloading).
// Note that you can add dependencies in preRun, even though
// it happens right before run - run will be postponed until
// the dependencies are met.
var runDependencies = 0;
var runDependencyWatcher = null;
var dependenciesFulfilled = null; // overridden to take different actions when all run dependencies are fulfilled
var runDependencyTracking = {};

function getUniqueRunDependency(id) {
  var orig = id;
  while (1) {
    if (!runDependencyTracking[id]) return id;
    id = orig + Math.random();
  }
  return id;
}

function addRunDependency(id) {
  runDependencies++;
  if (Module['monitorRunDependencies']) {
    Module['monitorRunDependencies'](runDependencies);
  }
  if (id) {
    assert(!runDependencyTracking[id]);
    runDependencyTracking[id] = 1;
    if (runDependencyWatcher === null && typeof setInterval !== 'undefined') {
      // Check for missing dependencies every few seconds
      runDependencyWatcher = setInterval(function() {
        if (ABORT) {
          clearInterval(runDependencyWatcher);
          runDependencyWatcher = null;
          return;
        }
        var shown = false;
        for (var dep in runDependencyTracking) {
          if (!shown) {
            shown = true;
            Module.printErr('still waiting on run dependencies:');
          }
          Module.printErr('dependency: ' + dep);
        }
        if (shown) {
          Module.printErr('(end of list)');
        }
      }, 10000);
    }
  } else {
    Module.printErr('warning: run dependency added without ID');
  }
}
Module["addRunDependency"] = addRunDependency;

function removeRunDependency(id) {
  runDependencies--;
  if (Module['monitorRunDependencies']) {
    Module['monitorRunDependencies'](runDependencies);
  }
  if (id) {
    assert(runDependencyTracking[id]);
    delete runDependencyTracking[id];
  } else {
    Module.printErr('warning: run dependency removed without ID');
  }
  if (runDependencies == 0) {
    if (runDependencyWatcher !== null) {
      clearInterval(runDependencyWatcher);
      runDependencyWatcher = null;
    }
    if (dependenciesFulfilled) {
      var callback = dependenciesFulfilled;
      dependenciesFulfilled = null;
      callback(); // can add another dependenciesFulfilled
    }
  }
}
Module["removeRunDependency"] = removeRunDependency;

Module["preloadedImages"] = {}; // maps url to image data
Module["preloadedAudios"] = {}; // maps url to audio data



var memoryInitializer = null;






// === Body ===

var ASM_CONSTS = [function($0, $1) { Module['websocket'].subprotocol = Pointer_stringify($0); Module['websocket'].url = Pointer_stringify($1); }];

function _emscripten_asm_const_iii(code, a0, a1) {
  return ASM_CONSTS[code](a0, a1);
}



STATIC_BASE = Runtime.GLOBAL_BASE;

STATICTOP = STATIC_BASE + 169472;
/* global initializers */  __ATINIT__.push({ func: function() { __GLOBAL__I_000101() } }, { func: function() { __GLOBAL__sub_I_PhotonPeer_cpp() } }, { func: function() { __GLOBAL__sub_I_Base_cpp() } }, { func: function() { __GLOBAL__sub_I_ToString_cpp() } }, { func: function() { __GLOBAL__sub_I_PeerData_cpp() } }, { func: function() { __GLOBAL__sub_I_iostream_cpp() } });


/* memory initializer */ allocate([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,72,0,0,0,51,0,0,0,52,0,0,0,50,0,0,0,50,0,0,0,50,0,0,0,50,0,0,0,0,0,0,0,16,86,0,0,59,50,2,0,216,86,0,0,23,50,2,0,0,0,0,0,1,0,0,0,64,0,0,0,3,236,255,255,56,86,0,0,93,50,2,0,72,0,0,0,0,0,0,0,56,86,0,0,178,50,2,0,224,2,0,0,0,0,0,0,56,86,0,0,225,50,2,0,176,2,0,0,0,0,0,0,56,86,0,0,2,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,38,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,74,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,110,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,146,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,182,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,218,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,254,51,2,0,176,2,0,0,0,0,0,0,56,86,0,0,46,52,2,0,176,2,0,0,0,0,0,0,56,86,0,0,96,52,2,0,224,2,0,0,0,0,0,0,56,86,0,0,143,52,2,0,224,2,0,0,0,0,0,0,56,86,0,0,192,52,2,0,176,2,0,0,0,0,0,0,56,86,0,0,3,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,70,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,139,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,177,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,214,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,251,53,2,0,176,2,0,0,0,0,0,0,56,86,0,0,32,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,69,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,106,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,155,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,192,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,240,54,2,0,176,2,0,0,0,0,0,0,56,86,0,0,17,55,2,0,160,2,0,0,0,0,0,0,56,86,0,0,48,55,2,0,160,2,0,0,0,0,0,0,56,86,0,0,196,57,2,0,144,2,0,0,0,0,0,0,56,86,0,0,90,55,2,0,80,2,0,0,0,0,0,0,16,86,0,0,118,55,2,0,56,86,0,0,165,55,2,0,104,2,0,0,0,0,0,0,16,86,0,0,229,55,2,0,56,86,0,0,31,57,2,0,104,2,0,0,0,0,0,0,56,86,0,0,103,57,2,0,144,2,0,0,0,0,0,0,56,86,0,0,135,57,2,0,160,2,0,0,0,0,0,0,56,86,0,0,171,57,2,0,80,2,0,0,0,0,0,0,56,86,0,0,97,58,2,0,160,2,0,0,0,0,0,0,56,86,0,0,32,58,2,0,160,2,0,0,0,0,0,0,56,86,0,0,171,58,2,0,160,2,0,0,0,0,0,0,56,86,0,0,230,58,2,0,160,2,0,0,0,0,0,0,56,86,0,0,34,59,2,0,160,2,0,0,0,0,0,0,56,86,0,0,90,59,2,0,160,2,0,0,0,0,0,0,56,86,0,0,144,59,2,0,160,2,0,0,0,0,0,0,56,86,0,0,176,59,2,0,176,2,0,0,0,0,0,0,56,86,0,0,213,59,2,0,64,3,0,0,0,0,0,0,16,86,0,0,1,60,2,0,56,86,0,0,95,63,2,0,88,3,0,0,0,0,0,0,16,86,0,0,133,63,2,0,56,86,0,0,185,63,2,0,160,2,0,0,0,0,0,0,56,86,0,0,228,63,2,0,48,3,0,0,0,0,0,0,56,86,0,0,194,64,2,0,72,3,0,0,0,0,0,0,56,86,0,0,134,64,2,0,160,2,0,0,0,0,0,0,56,86,0,0,229,64,2,0,176,3,0,0,0,0,0,0,56,86,0,0,23,65,2,0,160,2,0,0,0,0,0,0,56,86,0,0,57,65,2,0,208,3,0,0,0,0,0,0,56,86,0,0,116,65,2,0,160,2,0,0,0,0,0,0,16,86,0,0,174,65,2,0,56,86,0,0,3,94,2,0,248,3,0,0,0,0,0,0,16,86,0,0,241,93,2,0,56,86,0,0,45,94,2,0,248,3,0,0,0,0,0,0,16,86,0,0,87,94,2,0,16,86,0,0,136,94,2,0,216,86,0,0,185,94,2,0,0,0,0,0,1,0,0,0,232,3,0,0,3,244,255,255,216,86,0,0,232,94,2,0,0,0,0,0,1,0,0,0,0,4,0,0,3,244,255,255,216,86,0,0,23,95,2,0,0,0,0,0,1,0,0,0,232,3,0,0,3,244,255,255,216,86,0,0,70,95,2,0,0,0,0,0,1,0,0,0,0,4,0,0,3,244,255,255,56,86,0,0,162,95,2,0,72,9,0,0,0,0,0,0,56,86,0,0,192,95,2,0,96,9,0,0,0,0,0,0,56,86,0,0,234,95,2,0,24,4,0,0,0,0,0,0,56,86,0,0,3,96,2,0,16,4,0,0,0,0,0,0,56,86,0,0,66,96,2,0,24,4,0,0,0,0,0,0,56,86,0,0,90,96,2,0,16,4,0,0,0,0,0,0,56,86,0,0,114,96,2,0,240,4,0,0,0,0,0,0,56,86,0,0,134,96,2,0,64,9,0,0,0,0,0,0,56,86,0,0,156,96,2,0,240,4,0,0,0,0,0,0,216,86,0,0,214,96,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,48,5,0,0,0,0,0,0,216,86,0,0,26,97,2,0,0,0,0,0,1,0,0,0,72,5,0,0,0,0,0,0,16,86,0,0,48,97,2,0,216,86,0,0,73,97,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,112,5,0,0,0,0,0,0,216,86,0,0,141,97,2,0,0,0,0,0,1,0,0,0,72,5,0,0,0,0,0,0,216,86,0,0,177,97,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,168,5,0,0,0,0,0,0,216,86,0,0,245,97,2,0,0,0,0,0,1,0,0,0,192,5,0,0,0,0,0,0,16,86,0,0,11,98,2,0,216,86,0,0,36,98,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,232,5,0,0,0,0,0,0,216,86,0,0,104,98,2,0,0,0,0,0,1,0,0,0,192,5,0,0,0,0,0,0,216,86,0,0,190,99,2,0,0,0,0,0,3,0,0,0,240,4,0,0,2,0,0,0,40,6,0,0,2,0,0,0,48,6,0,0,0,8,0,0,16,86,0,0,37,100,2,0,16,86,0,0,3,100,2,0,216,86,0,0,56,100,2,0,0,0,0,0,3,0,0,0,240,4,0,0,2,0,0,0,40,6,0,0,2,0,0,0,96,6,0,0,0,8,0,0,16,86,0,0,125,100,2,0,216,86,0,0,159,100,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,136,6,0,0,0,8,0,0,16,86,0,0,228,100,2,0,216,86,0,0,14,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,136,6,0,0,0,8,0,0,216,86,0,0,83,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,208,6,0,0,2,0,0,0,16,86,0,0,111,101,2,0,216,86,0,0,132,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,208,6,0,0,2,0,0,0,216,86,0,0,160,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,208,6,0,0,2,0,0,0,216,86,0,0,188,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,208,6,0,0,2,0,0,0,216,86,0,0,247,101,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,88,7,0,0,0,0,0,0,16,86,0,0,61,102,2,0,216,86,0,0,97,102,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,128,7,0,0,0,0,0,0,16,86,0,0,167,102,2,0,216,86,0,0,198,102,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,168,7,0,0,0,0,0,0,16,86,0,0,12,103,2,0,216,86,0,0,37,103,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,208,7,0,0,0,0,0,0,16,86,0,0,107,103,2,0,216,86,0,0,132,103,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,248,7,0,0,2,0,0,0,16,86,0,0,153,103,2,0,216,86,0,0,48,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,248,7,0,0,2,0,0,0,56,86,0,0,177,103,2,0,48,8,0,0,0,0,0,0,216,86,0,0,212,103,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,80,8,0,0,2,0,0,0,16,86,0,0,247,103,2,0,56,86,0,0,14,104,2,0,48,8,0,0,0,0,0,0,216,86,0,0,69,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,80,8,0,0,2,0,0,0,216,86,0,0,103,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,80,8,0,0,2,0,0,0,216,86,0,0,137,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,80,8,0,0,2,0,0,0,56,86,0,0,172,104,2,0,240,4,0,0,0,0,0,0,216,86,0,0,194,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,248,8,0,0,2,0,0,0,16,86,0,0,212,104,2,0,216,86,0,0,233,104,2,0,0,0,0,0,2,0,0,0,240,4,0,0,2,0,0,0,248,8,0,0,2,0,0,0,56,86,0,0,251,104,2,0,240,4,0,0,0,0,0,0,56,86,0,0,16,105,2,0,240,4,0,0,0,0,0,0,16,86,0,0,44,105,2,0,56,86,0,0,255,105,2,0,88,9,0,0,0,0,0,0,16,86,0,0,22,106,2,0,56,86,0,0,47,106,2,0,208,9,0,0,0,0,0,0,16,86,0,0,218,106,2,0,56,86,0,0,58,107,2,0,136,9,0,0,0,0,0,0,56,86,0,0,231,106,2,0,152,9,0,0,0,0,0,0,16,86,0,0,8,107,2,0,56,86,0,0,21,107,2,0,120,9,0,0,0,0,0,0,56,86,0,0,93,108,2,0,112,9,0,0,0,0,0,0,56,86,0,0,106,108,2,0,112,9,0,0,0,0,0,0,56,86,0,0,122,108,2,0,112,9,0,0,0,0,0,0,56,86,0,0,140,108,2,0,192,9,0,0,0,0,0,0,56,86,0,0,157,108,2,0,192,9,0,0,0,0,0,0,56,86,0,0,188,108,2,0,112,9,0,0,0,0,0,0,56,86,0,0,236,108,2,0,136,9,0,0,0,0,0,0,56,86,0,0,200,108,2,0,16,10,0,0,0,0,0,0,56,86,0,0,14,109,2,0,120,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,96,0,0,0,42,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,47,0,0,0,84,10,0,0,36,0,0,0,36,0,0,0,84,10,0,0,0,0,0,0,64,0,0,0,48,0,0,0,49,0,0,0,50,0,0,0,36,0,0,0,36,0,0,0,0,0,0,0,112,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,0,0,0,0,128,0,0,0,63,0,0,0,64,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,0,0,0,0,144,0,0,0,70,0,0,0,71,0,0,0,65,0,0,0,66,0,0,0,72,0,0,0,73,0,0,0,74,0,0,0,0,0,0,0,160,0,0,0,75,0,0,0,76,0,0,0,65,0,0,0,66,0,0,0,77,0,0,0,78,0,0,0,79,0,0,0,0,0,0,0,176,0,0,0,80,0,0,0,81,0,0,0,65,0,0,0,66,0,0,0,82,0,0,0,83,0,0,0,84,0,0,0,0,0,0,0,192,0,0,0,85,0,0,0,86,0,0,0,65,0,0,0,66,0,0,0,87,0,0,0,88,0,0,0,89,0,0,0,0,0,0,0,208,0,0,0,90,0,0,0,91,0,0,0,65,0,0,0,66,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,0,0,0,0,224,0,0,0,95,0,0,0,96,0,0,0,65,0,0,0,66,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,0,0,0,0,240,0,0,0,100,0,0,0,101,0,0,0,65,0,0,0,66,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,0,0,0,0,0,1,0,0,105,0,0,0,106,0,0,0,65,0,0,0,66,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,0,0,0,0,16,1,0,0,110,0,0,0,111,0,0,0,65,0,0,0,66,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,0,0,0,0,32,1,0,0,115,0,0,0,116,0,0,0,55,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,123,0,0,0,0,0,0,0,48,1,0,0,124,0,0,0,125,0,0,0,55,0,0,0,126,0,0,0,127,0,0,0,128,0,0,0,129,0,0,0,130,0,0,0,131,0,0,0,132,0,0,0,0,0,0,0,64,1,0,0,133,0,0,0,134,0,0,0,65,0,0,0,66,0,0,0,135,0,0,0,136,0,0,0,137,0,0,0,0,0,0,0,80,1,0,0,138,0,0,0,139,0,0,0,65,0,0,0,66,0,0,0,140,0,0,0,141,0,0,0,142,0,0,0,0,0,0,0,96,1,0,0,143,0,0,0,144,0,0,0,65,0,0,0,66,0,0,0,145,0,0,0,146,0,0,0,147,0,0,0,0,0,0,0,112,1,0,0,148,0,0,0,149,0,0,0,65,0,0,0,66,0,0,0,150,0,0,0,151,0,0,0,152,0,0,0,0,0,0,0,128,1,0,0,153,0,0,0,154,0,0,0,65,0,0,0,66,0,0,0,155,0,0,0,156,0,0,0,157,0,0,0,0,0,0,0,144,1,0,0,158,0,0,0,159,0,0,0,65,0,0,0,66,0,0,0,160,0,0,0,161,0,0,0,162,0,0,0,0,0,0,0,160,1,0,0,163,0,0,0,164,0,0,0,65,0,0,0,66,0,0,0,165,0,0,0,166,0,0,0,167,0,0,0,0,0,0,0,176,1,0,0,168,0,0,0,169,0,0,0,65,0,0,0,66,0,0,0,170,0,0,0,171,0,0,0,172,0,0,0,0,0,0,0,192,1,0,0,173,0,0,0,174,0,0,0,65,0,0,0,66,0,0,0,175,0,0,0,176,0,0,0,177,0,0,0,0,0,0,0,208,1,0,0,178,0,0,0,179,0,0,0,65,0,0,0,66,0,0,0,180,0,0,0,181,0,0,0,182,0,0,0,0,0,0,0,224,1,0,0,183,0,0,0,184,0,0,0,65,0,0,0,66,0,0,0,185,0,0,0,186,0,0,0,187,0,0,0,0,0,0,0,240,1,0,0,188,0,0,0,189,0,0,0,65,0,0,0,66,0,0,0,190,0,0,0,191,0,0,0,192,0,0,0,101,0,0,0,109,0,0,0,45,0,0,0,115,0,0,0,100,0,0,0,107,0,0,0,92,0,0,0,67,0,0,0,111,0,0,0,109,0,0,0,109,0,0,0,111,0,0,0,110,0,0,0,45,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,47,0,0,0,105,0,0,0,110,0,0,0,99,0,0,0,47,0,0,0,74,0,0,0,86,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,111,0,0,0,114,0,0,0,46,0,0,0,104,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,193,0,0,0,194,0,0,0,65,0,0,0,66,0,0,0,195,0,0,0,196,0,0,0,197,0,0,0,0,0,0,0,16,2,0,0,198,0,0,0,199,0,0,0,65,0,0,0,200,0,0,0,0,0,0,0,32,2,0,0,201,0,0,0,202,0,0,0,65,0,0,0,203,0,0,0,0,0,0,0,48,2,0,0,204,0,0,0,205,0,0,0,65,0,0,0,206,0,0,0,207,0,0,0,208,0,0,0,209,0,0,0,210,0,0,0,0,0,0,0,64,2,0,0,211,0,0,0,212,0,0,0,65,0,0,0,213,0,0,0,34,0,0,0,0,0,0,0,156,55,2,0,147,55,2,0,0,0,0,0,88,2,0,0,214,0,0,0,215,0,0,0,216,0,0,0,217,0,0,0,218,0,0,0,219,0,0,0,0,128,0,0,0,0,0,0,112,2,0,0,220,0,0,0,221,0,0,0,222,0,0,0,223,0,0,0,224,0,0,0,225,0,0,0,99,0,0,0,108,0,0,0,97,0,0,0,115,0,0,0,115,0,0,0,32,0,0,0,0,0,0,0,115,0,0,0,116,0,0,0,114,0,0,0,117,0,0,0,99,0,0,0,116,0,0,0,32,0,0,0,0,0,0,0,117,0,0,0,110,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,32,0,0,0,0,0,0,0,101,0,0,0,110,0,0,0,117,0,0,0,109,0,0,0,32,0,0,0,0,0,0,0,91,0,0,0,93,0,0,0,0,0,0,0,0,0,0,0,128,2,0,0,226,0,0,0,227,0,0,0,65,0,0,0,206,0,0,0,228,0,0,0,229,0,0,0,230,0,0,0,231,0,0,0,97,57,2,0,0,0,0,0,144,2,0,0,232,0,0,0,233,0,0,0,65,0,0,0,206,0,0,0,50,0,0,0,50,0,0,0,50,0,0,0,50,0,0,0,0,0,0,0,160,2,0,0,234,0,0,0,235,0,0,0,65,0,0,0,50,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,114,0,0,0,0,0,0,0,115,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,114,0,0,0,0,0,0,0,117,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,114,0,0,0,0,0,0,0,115,0,0,0,104,0,0,0,111,0,0,0,114,0,0,0,116,0,0,0,0,0,0,0,117,0,0,0,115,0,0,0,104,0,0,0,111,0,0,0,114,0,0,0,116,0,0,0,0,0,0,0,105,0,0,0,110,0,0,0,116,0,0,0,0,0,0,0,117,0,0,0,105,0,0,0,110,0,0,0,116,0,0,0,0,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,0,0,0,0,117,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,0,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,0,0,0,0,117,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,0,0,0,0,102,0,0,0,108,0,0,0,111,0,0,0,97,0,0,0,116,0,0,0,0,0,0,0,100,0,0,0,111,0,0,0,117,0,0,0,98,0,0,0,108,0,0,0,101,0,0,0,0,0,0,0,108,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,100,0,0,0,111,0,0,0,117,0,0,0,98,0,0,0,108,0,0,0,101,0,0,0,0,0,0,0,98,0,0,0,111,0,0,0,111,0,0,0,108,0,0,0,0,0,0,0,0,0,0,0,176,2,0,0,236,0,0,0,237,0,0,0,65,0,0,0,66,0,0,0,238,0,0,0,239,0,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,67,0,0,0,117,0,0,0,115,0,0,0,116,0,0,0,111,0,0,0,109,0,0,0,84,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,66,0,0,0,97,0,0,0,115,0,0,0,101,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,67,0,0,0,117,0,0,0,115,0,0,0,116,0,0,0,111,0,0,0,109,0,0,0,32,0,0,0,116,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,110,0,0,0,111,0,0,0,116,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,103,0,0,0,105,0,0,0,115,0,0,0,116,0,0,0,101,0,0,0,114,0,0,0,101,0,0,0,100,0,0,0,33,0,0,0,0,0,0,0,40,0,0,0,41,0,0,0,0,0,0,0,85,0,0,0,78,0,0,0,75,0,0,0,78,0,0,0,79,0,0,0,87,0,0,0,78,0,0,0,0,0,0,0,68,0,0,0,69,0,0,0,66,0,0,0,85,0,0,0,71,0,0,0,0,0,0,0,73,0,0,0,78,0,0,0,70,0,0,0,79,0,0,0,0,0,0,0,87,0,0,0,65,0,0,0,82,0,0,0,78,0,0,0,73,0,0,0,78,0,0,0,71,0,0,0,0,0,0,0,69,0,0,0,82,0,0,0,82,0,0,0,79,0,0,0,82,0,0,0,0,0,0,0,70,0,0,0,65,0,0,0,84,0,0,0,65,0,0,0,76,0,0,0,0,0,0,0,37,0,0,0,104,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,45,0,0,0,55,0,0,0,108,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,45,0,0,0,51,0,0,0,48,0,0,0,108,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,45,0,0,0,54,0,0,0,48,0,0,0,108,0,0,0,115,0,0,0,32,0,0,0,108,0,0,0,105,0,0,0,110,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,53,0,0,0,117,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,37,0,0,0,108,0,0,0,115,0,0,0,0,0,0,0,0,0,0,0,192,2,0,0,240,0,0,0,241,0,0,0,65,0,0,0,242,0,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,79,0,0,0,98,0,0,0,106,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,69,0,0,0,114,0,0,0,114,0,0,0,111,0,0,0,114,0,0,0,58,0,0,0,32,0,0,0,117,0,0,0,110,0,0,0,115,0,0,0,117,0,0,0,112,0,0,0,112,0,0,0,111,0,0,0,114,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,116,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,0,0,0,0,37,0,0,0,102,0,0,0,0,0,0,0,37,0,0,0,108,0,0,0,108,0,0,0,100,0,0,0,0,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,37,0,0,0,104,0,0,0,100,0,0,0,0,0,0,0,37,0,0,0,104,0,0,0,104,0,0,0,117,0,0,0,0,0,0,0,40,0,0,0,78,0,0,0,85,0,0,0,76,0,0,0,76,0,0,0,41,0,0,0,0,0,0,0,0,0,0,0,224,2,0,0,243,0,0,0,244,0,0,0,55,0,0,0,245,0,0,0,246,0,0,0,247,0,0,0,248,0,0,0,249,0,0,0,250,0,0,0,251,0,0,0,0,0,0,0,208,2,0,0,252,0,0,0,253,0,0,0,65,0,0,0,254,0,0,0,42,0,0,0,0,0,0,0,60,0,0,0,0,0,0,0,62,0,0,0,0,0,0,0,78,0,0,0,85,0,0,0,76,0,0,0,76,0,0,0,0,0,0,0,117,0,0,0,110,0,0,0,107,0,0,0,110,0,0,0,111,0,0,0,119,0,0,0,110,0,0,0,32,0,0,0,116,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,0,0,0,0,123,0,0,0,0,0,0,0,62,0,0,0,41,0,0,0,0,0,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,72,0,0,0,101,0,0,0,108,0,0,0,112,0,0,0,101,0,0,0,114,0,0,0,115,0,0,0,47,0,0,0,68,0,0,0,101,0,0,0,83,0,0,0,101,0,0,0,114,0,0,0,105,0,0,0,97,0,0,0,108,0,0,0,105,0,0,0,122,0,0,0,101,0,0,0,114,0,0,0,73,0,0,0,109,0,0,0,112,0,0,0,108,0,0,0,101,0,0,0,109,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,73,0,0,0,110,0,0,0,116,0,0,0,101,0,0,0,114,0,0,0,110,0,0,0,97,0,0,0,108,0,0,0,32,0,0,0,100,0,0,0,101,0,0,0,115,0,0,0,101,0,0,0,114,0,0,0,105,0,0,0,97,0,0,0,108,0,0,0,105,0,0,0,122,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,32,0,0,0,101,0,0,0,114,0,0,0,114,0,0,0,111,0,0,0,114,0,0,0,33,0,0,0,32,0,0,0,119,0,0,0,114,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,32,0,0,0,116,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,32,0,0,0,104,0,0,0,97,0,0,0,110,0,0,0,100,0,0,0,108,0,0,0,101,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,0,0,0,0,240,2,0,0,234,0,0,0,255,0,0,0,65,0,0,0,0,1,0,0,0,0,0,0,0,3,0,0,1,1,0,0,2,1,0,0,65,0,0,0,3,1,0,0,0,0,0,0,16,3,0,0,4,1,0,0,5,1,0,0,65,0,0,0,6,1,0,0,0,0,0,0,80,2,0,0,7,1,0,0,8,1,0,0,65,0,0,0,50,0,0,0,0,0,0,0,32,3,0,0,236,0,0,0,9,1,0,0,65,0,0,0,66,0,0,0,10,1,0,0,11,1,0,0,12,1,0,0,37,0,0,0,104,0,0,0,117,0,0,0,0,0,0,0,0,0,0,0,48,3,0,0,13,1,0,0,14,1,0,0,15,1,0,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,1,0,0,23,1,0,0,50,0,0,0,24,1,0,0,50,0,0,0,50,0,0,0,47,0,0,0,0,0,0,0,63,0,0,0,112,0,0,0,114,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,99,0,0,0,111,0,0,0,108,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,108,0,0,0,105,0,0,0,98,0,0,0,105,0,0,0,100,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,108,0,0,0,105,0,0,0,98,0,0,0,118,0,0,0,101,0,0,0,114,0,0,0,115,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,61,0,0,0,0,0,0,0,108,0,0,0,111,0,0,0,99,0,0,0,97,0,0,0,108,0,0,0,104,0,0,0,111,0,0,0,115,0,0,0,116,0,0,0,0,0,0,0,58,0,0,0,0,0,0,0,119,0,0,0,115,0,0,0,115,0,0,0,0,0,0,0,93,0,0,0,58,0,0,0,0,0,0,0,0,0,0,0,64,3,0,0,50,0,0,0,50,0,0,0,50,0,0,0,50,0,0,0,17,1,0,0,18,1,0,0,50,0,0,0,50,0,0,0,50,0,0,0,25,1,0,0,26,1,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,73,0,0,0,110,0,0,0,116,0,0,0,101,0,0,0,114,0,0,0,110,0,0,0,97,0,0,0,108,0,0,0,47,0,0,0,80,0,0,0,101,0,0,0,101,0,0,0,114,0,0,0,66,0,0,0,97,0,0,0,115,0,0,0,101,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,60,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,97,0,0,0,100,0,0,0,32,0,0,0,105,0,0,0,110,0,0,0,105,0,0,0,116,0,0,0,0,0,0,0,101,0,0,0,115,0,0,0,116,0,0,0,97,0,0,0,98,0,0,0,108,0,0,0,105,0,0,0,115,0,0,0,104,0,0,0,105,0,0,0,110,0,0,0,103,0,0,0,32,0,0,0,101,0,0,0,110,0,0,0,99,0,0,0,114,0,0,0,121,0,0,0,112,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,32,0,0,0,107,0,0,0,101,0,0,0,121,0,0,0,115,0,0,0,32,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,46,0,0,0,32,0,0,0,83,0,0,0,101,0,0,0,114,0,0,0,118,0,0,0,101,0,0,0,114,0,0,0,39,0,0,0,115,0,0,0,32,0,0,0,112,0,0,0,117,0,0,0,98,0,0,0,108,0,0,0,105,0,0,0,99,0,0,0,32,0,0,0,107,0,0,0,101,0,0,0,121,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,78,0,0,0,85,0,0,0,76,0,0,0,76,0,0,0,32,0,0,0,111,0,0,0,114,0,0,0,32,0,0,0,104,0,0,0,97,0,0,0,115,0,0,0,32,0,0,0,97,0,0,0,110,0,0,0,32,0,0,0,117,0,0,0,110,0,0,0,101,0,0,0,120,0,0,0,112,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,115,0,0,0,105,0,0,0,122,0,0,0,101,0,0,0,46,0,0,0,32,0,0,0,75,0,0,0,101,0,0,0,121,0,0,0,115,0,0,0,105,0,0,0,122,0,0,0,101,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,152,47,138,66,145,68,55,113,207,251,192,181,165,219,181,233,91,194,86,57,241,17,241,89,164,130,63,146,213,94,28,171,152,170,7,216,1,91,131,18,190,133,49,36,195,125,12,85,116,93,190,114,254,177,222,128,167,6,220,155,116,241,155,193,193,105,155,228,134,71,190,239,198,157,193,15,204,161,12,36,111,44,233,45,170,132,116,74,220,169,176,92,218,136,249,118,82,81,62,152,109,198,49,168,200,39,3,176,199,127,89,191,243,11,224,198,71,145,167,213,81,99,202,6,103,41,41,20,133,10,183,39,56,33,27,46,252,109,44,77,19,13,56,83,84,115,10,101,187,10,106,118,46,201,194,129,133,44,114,146,161,232,191,162,75,102,26,168,112,139,75,194,163,81,108,199,25,232,146,209,36,6,153,214,133,53,14,244,112,160,106,16,22,193,164,25,8,108,55,30,76,119,72,39,181,188,176,52,179,12,28,57,74,170,216,78,79,202,156,91,243,111,46,104,238,130,143,116,111,99,165,120,20,120,200,132,8,2,199,140,250,255,190,144,235,108,80,164,247,163,249,190,242,120,113,198,103,230,9,106,133,174,103,187,114,243,110,60,58,245,79,165,127,82,14,81,140,104,5,155,171,217,131,31,25,205,224,91,20,29,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,109,0,0,0,115,0,0,0,103,0,0,0,84,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,32,0,0,0,110,0,0,0,111,0,0,0,116,0,0,0,32,0,0,0,101,0,0,0,120,0,0,0,112,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,104,0,0,0,101,0,0,0,114,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,85,0,0,0,68,0,0,0,80,0,0,0,47,0,0,0,84,0,0,0,67,0,0,0,80,0,0,0,32,0,0,0,100,0,0,0,97,0,0,0,116,0,0,0,97,0,0,0,32,0,0,0,116,0,0,0,111,0,0,0,111,0,0,0,32,0,0,0,115,0,0,0,104,0,0,0,111,0,0,0,114,0,0,0,116,0,0,0,33,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,77,0,0,0,97,0,0,0,103,0,0,0,105,0,0,0,99,0,0,0,78,0,0,0,117,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,32,0,0,0,115,0,0,0,104,0,0,0,111,0,0,0,117,0,0,0,108,0,0,0,100,0,0,0,32,0,0,0,98,0,0,0,101,0,0,0,32,0,0,0,48,0,0,0,120,0,0,0,70,0,0,0,51,0,0,0,44,0,0,0,32,0,0,0,98,0,0,0,117,0,0,0,116,0,0,0,32,0,0,0,105,0,0,0,116,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,88,0,0,0,32,0,0,0,0,0,0,0,98,0,0,0,111,0,0,0,100,0,0,0,121,0,0,0,66,0,0,0,117,0,0,0,102,0,0,0,102,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,44,0,0,0,32,0,0,0,109,0,0,0,115,0,0,0,103,0,0,0,84,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,40,0,0,0,101,0,0,0,118,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,32,0,0,0,61,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,41,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,100,0,0,0,117,0,0,0,101,0,0,0,32,0,0,0,116,0,0,0,111,0,0,0,32,0,0,0,97,0,0,0,32,0,0,0,100,0,0,0,97,0,0,0,116,0,0,0,97,0,0,0,32,0,0,0,116,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,44,0,0,0,32,0,0,0,110,0,0,0,111,0,0,0,116,0,0,0,32,0,0,0,115,0,0,0,117,0,0,0,112,0,0,0,112,0,0,0,111,0,0,0,114,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,98,0,0,0,121,0,0,0,32,0,0,0,112,0,0,0,114,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,99,0,0,0,111,0,0,0,108,0,0,0,0,0,0,0,84,0,0,0,104,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,102,0,0,0,101,0,0,0,97,0,0,0,116,0,0,0,117,0,0,0,114,0,0,0,101,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,110,0,0,0,111,0,0,0,116,0,0,0,32,0,0,0,99,0,0,0,117,0,0,0,114,0,0,0,114,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,108,0,0,0,121,0,0,0,32,0,0,0,105,0,0,0,109,0,0,0,112,0,0,0,108,0,0,0,101,0,0,0,109,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,111,0,0,0,110,0,0,0,32,0,0,0,116,0,0,0,104,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,112,0,0,0,108,0,0,0,97,0,0,0,116,0,0,0,102,0,0,0,111,0,0,0,114,0,0,0,109,0,0,0,33,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,101,0,0,0,120,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,110,0,0,0,103,0,0,0,101,0,0,0,32,0,0,0,107,0,0,0,101,0,0,0,121,0,0,0,115,0,0,0,32,0,0,0,102,0,0,0,105,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,32,0,0,0,116,0,0,0,111,0,0,0,32,0,0,0,101,0,0,0,110,0,0,0,97,0,0,0,98,0,0,0,108,0,0,0,101,0,0,0,32,0,0,0,101,0,0,0,110,0,0,0,99,0,0,0,114,0,0,0,121,0,0,0,112,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,33,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,99,0,0,0,97,0,0,0,108,0,0,0,108,0,0,0,32,0,0,0,67,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,40,0,0,0,41,0,0,0,32,0,0,0,102,0,0,0,105,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,33,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,108,0,0,0,73,0,0,0,100,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,111,0,0,0,117,0,0,0,116,0,0,0,32,0,0,0,111,0,0,0,102,0,0,0,32,0,0,0,99,0,0,0,104,0,0,0,97,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,108,0,0,0,45,0,0,0,99,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,116,0,0,0,32,0,0,0,98,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,100,0,0,0,115,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,45,0,0,0,37,0,0,0,100,0,0,0,41,0,0,0,0,0,0,0,100,0,0,0,105,0,0,0,115,0,0,0,112,0,0,0,97,0,0,0,116,0,0,0,99,0,0,0,104,0,0,0,32,0,0,0,61,0,0,0,61,0,0,0,32,0,0,0,116,0,0,0,114,0,0,0,117,0,0,0,101,0,0,0,0,0,0,0,100,0,0,0,105,0,0,0,115,0,0,0,112,0,0,0,97,0,0,0,116,0,0,0,99,0,0,0,104,0,0,0,32,0,0,0,61,0,0,0,61,0,0,0,32,0,0,0,102,0,0,0,97,0,0,0,108,0,0,0,115,0,0,0,101,0,0,0,0,0,0,0,97,0,0,0,100,0,0,0,100,0,0,0,114,0,0,0,101,0,0,0,115,0,0,0,115,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,108,0,0,0,115,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,105,0,0,0,110,0,0,0,118,0,0,0,97,0,0,0,108,0,0,0,105,0,0,0,100,0,0,0,32,0,0,0,112,0,0,0,97,0,0,0,114,0,0,0,97,0,0,0,109,0,0,0,101,0,0,0,116,0,0,0,101,0,0,0,114,0,0,0,115,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,80,0,0,0,101,0,0,0,101,0,0,0,114,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,97,0,0,0,108,0,0,0,114,0,0,0,101,0,0,0,97,0,0,0,100,0,0,0,121,0,0,0,32,0,0,0,99,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,33,0,0,0,0,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,80,0,0,0,104,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,110,0,0,0,67,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,58,0,0,0,58,0,0,0,115,0,0,0,116,0,0,0,97,0,0,0,114,0,0,0,116,0,0,0,67,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,40,0,0,0,41,0,0,0,32,0,0,0,102,0,0,0,97,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,46,0,0,0,0,0,0,0,37,0,0,0,117,0,0,0,0,0,0,0,105,0,0,0,110,0,0,0,105,0,0,0,116,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,97,0,0,0,112,0,0,0,112,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,99,0,0,0,108,0,0,0,105,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,118,0,0,0,101,0,0,0,114,0,0,0,115,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,112,0,0,0,114,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,99,0,0,0,111,0,0,0,108,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,115,0,0,0,105,0,0,0,100,0,0,0,61,0,0,0,0,0,0,0,38,0,0,0,73,0,0,0,80,0,0,0,118,0,0,0,54,0,0,0,0,0,0,0,80,0,0,0,79,0,0,0,83,0,0,0,84,0,0,0,32,0,0,0,47,0,0,0,63,0,0,0,0,0,0,0,32,0,0,0,72,0,0,0,84,0,0,0,84,0,0,0,80,0,0,0,47,0,0,0,49,0,0,0,46,0,0,0,49,0,0,0,13,0,0,0,10,0,0,0,0,0,0,0,72,0,0,0,111,0,0,0,115,0,0,0,116,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,13,0,0,0,10,0,0,0,0,0,0,0,67,0,0,0,111,0,0,0,110,0,0,0,116,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,45,0,0,0,76,0,0,0,101,0,0,0,110,0,0,0,103,0,0,0,116,0,0,0,104,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,13,0,0,0,10,0,0,0,13,0,0,0,10,0,0,0,0,0,0,0,0,0,0,0,72,3,0,0,27,1,0,0,28,1,0,0,29,1,0,0,50,0,0,0,30,1,0,0,50,0,0,0,50,0,0,0,31,1,0,0,32,1,0,0,50,0,0,0,33,1,0,0,50,0,0,0,34,1,0,0,50,0,0,0,35,1,0,0,36,1,0,0,50,0,0,0,50,0,0,0,50,0,0,0,37,1,0,0,38,1,0,0,39,1,0,0,40,1,0,0,50], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE);
/* memory initializer */ allocate([41,1,0,0,42,1,0,0,43,1,0,0,44,1,0,0,45,1,0,0,50,0,0,0,46,1,0,0,47,1,0,0,48,1,0,0,49,1,0,0,50,0,0,0,50,0,0,0,0,0,0,0,96,3,0,0,50,1,0,0,51,1,0,0,65,0,0,0,52,1,0,0,123,0,0,0,79,0,0,0,112,0,0,0,101,0,0,0,114,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,46,0,0,0,32,0,0,0,83,0,0,0,105,0,0,0,122,0,0,0,101,0,0,0,61,0,0,0,0,0,0,0,125,0,0,0,0,0,0,0,71,0,0,0,112,0,0,0,66,0,0,0,105,0,0,0,110,0,0,0,97,0,0,0,114,0,0,0,121,0,0,0,86,0,0,0,0,0,0,0,0,0,0,0,112,3,0,0,13,1,0,0,14,1,0,0,15,1,0,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,53,1,0,0,54,1,0,0,55,1,0,0,24,1,0,0,56,1,0,0,57,1,0,0,0,0,0,0,128,3,0,0,58,1,0,0,59,1,0,0,60,1,0,0,61,1,0,0,30,1,0,0,62,1,0,0,63,1,0,0,31,1,0,0,64,1,0,0,65,1,0,0,66,1,0,0,67,1,0,0,34,1,0,0,68,1,0,0,35,1,0,0,36,1,0,0,69,1,0,0,70,1,0,0,71,1,0,0,37,1,0,0,72,1,0,0,73,1,0,0,40,1,0,0,74,1,0,0,75,1,0,0,42,1,0,0,43,1,0,0,44,1,0,0,45,1,0,0,76,1,0,0,46,1,0,0,47,1,0,0,77,1,0,0,78,1,0,0,79,1,0,0,80,1,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,73,0,0,0,110,0,0,0,116,0,0,0,101,0,0,0,114,0,0,0,110,0,0,0,97,0,0,0,108,0,0,0,47,0,0,0,84,0,0,0,80,0,0,0,101,0,0,0,101,0,0,0,114,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,99,0,0,0,84,0,0,0,121,0,0,0,112,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,112,0,0,0,97,0,0,0,121,0,0,0,108,0,0,0,111,0,0,0,97,0,0,0,100,0,0,0,83,0,0,0,105,0,0,0,122,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,46,0,0,0,46,0,0,0,92,0,0,0,67,0,0,0,111,0,0,0,109,0,0,0,109,0,0,0,111,0,0,0,110,0,0,0,45,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,47,0,0,0,105,0,0,0,110,0,0,0,99,0,0,0,47,0,0,0,74,0,0,0,86,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,111,0,0,0,114,0,0,0,46,0,0,0,104,0,0,0,0,0,0,0,74,0,0,0,86,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,111,0,0,0,114,0,0,0,58,0,0,0,32,0,0,0,73,0,0,0,110,0,0,0,100,0,0,0,101,0,0,0,120,0,0,0,32,0,0,0,79,0,0,0,117,0,0,0,116,0,0,0,32,0,0,0,79,0,0,0,102,0,0,0,32,0,0,0,66,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,100,0,0,0,115,0,0,0,0,0,0,0,99,0,0,0,97,0,0,0,108,0,0,0,108,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,105,0,0,0,110,0,0,0,32,0,0,0,97,0,0,0,32,0,0,0,100,0,0,0,105,0,0,0,115,0,0,0,99,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,115,0,0,0,116,0,0,0,97,0,0,0,116,0,0,0,101,0,0,0,46,0,0,0,32,0,0,0,80,0,0,0,104,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,110,0,0,0,32,0,0,0,99,0,0,0,97,0,0,0,110,0,0,0,39,0,0,0,116,0,0,0,32,0,0,0,102,0,0,0,101,0,0,0,116,0,0,0,99,0,0,0,104,0,0,0,32,0,0,0,116,0,0,0,104,0,0,0,101,0,0,0,32,0,0,0,115,0,0,0,101,0,0,0,114,0,0,0,118,0,0,0,101,0,0,0,114,0,0,0,116,0,0,0,105,0,0,0,109,0,0,0,101,0,0,0,32,0,0,0,119,0,0,0,104,0,0,0,105,0,0,0,108,0,0,0,101,0,0,0,32,0,0,0,100,0,0,0,105,0,0,0,115,0,0,0,99,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,32,0,0,0,97,0,0,0,110,0,0,0,100,0,0,0,32,0,0,0,119,0,0,0,105,0,0,0,108,0,0,0,108,0,0,0,32,0,0,0,105,0,0,0,103,0,0,0,110,0,0,0,111,0,0,0,114,0,0,0,101,0,0,0,32,0,0,0,116,0,0,0,104,0,0,0,105,0,0,0,115,0,0,0,32,0,0,0,99,0,0,0,97,0,0,0,108,0,0,0,108,0,0,0,46,0,0,0,32,0,0,0,80,0,0,0,108,0,0,0,101,0,0,0,97,0,0,0,115,0,0,0,101,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,112,0,0,0,101,0,0,0,97,0,0,0,116,0,0,0,32,0,0,0,105,0,0,0,116,0,0,0,44,0,0,0,32,0,0,0,119,0,0,0,104,0,0,0,101,0,0,0,110,0,0,0,32,0,0,0,121,0,0,0,111,0,0,0,117,0,0,0,32,0,0,0,97,0,0,0,114,0,0,0,101,0,0,0,32,0,0,0,99,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,101,0,0,0,100,0,0,0,46,0,0,0,0,0,0,0,97,0,0,0,100,0,0,0,100,0,0,0,105,0,0,0,110,0,0,0,103,0,0,0,32,0,0,0,80,0,0,0,73,0,0,0,78,0,0,0,71,0,0,0,0,0,0,0,100,0,0,0,105,0,0,0,115,0,0,0,99,0,0,0,111,0,0,0,110,0,0,0,110,0,0,0,101,0,0,0,99,0,0,0,116,0,0,0,32,0,0,0,100,0,0,0,117,0,0,0,101,0,0,0,32,0,0,0,116,0,0,0,111,0,0,0,32,0,0,0,97,0,0,0,32,0,0,0,116,0,0,0,105,0,0,0,109,0,0,0,101,0,0,0,111,0,0,0,117,0,0,0,116,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,116,0,0,0,105,0,0,0,109,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,32,0,0,0,32,0,0,0,116,0,0,0,105,0,0,0,109,0,0,0,101,0,0,0,32,0,0,0,111,0,0,0,102,0,0,0,32,0,0,0,108,0,0,0,97,0,0,0,115,0,0,0,116,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,99,0,0,0,101,0,0,0,105,0,0,0,118,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,108,0,0,0,101,0,0,0,110,0,0,0,103,0,0,0,116,0,0,0,104,0,0,0,32,0,0,0,61,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,44,0,0,0,32,0,0,0,101,0,0,0,114,0,0,0,114,0,0,0,111,0,0,0,114,0,0,0,32,0,0,0,61,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,103,0,0,0,105,0,0,0,99,0,0,0,32,0,0,0,110,0,0,0,117,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,32,0,0,0,115,0,0,0,104,0,0,0,111,0,0,0,117,0,0,0,108,0,0,0,100,0,0,0,32,0,0,0,98,0,0,0,101,0,0,0,32,0,0,0,48,0,0,0,120,0,0,0,70,0,0,0,48,0,0,0,32,0,0,0,111,0,0,0,114,0,0,0,32,0,0,0,48,0,0,0,120,0,0,0,70,0,0,0,51,0,0,0,44,0,0,0,32,0,0,0,98,0,0,0,117,0,0,0,116,0,0,0,32,0,0,0,105,0,0,0,116,0,0,0,32,0,0,0,105,0,0,0,115,0,0,0,58,0,0,0,32,0,0,0,48,0,0,0,120,0,0,0,37,0,0,0,88,0,0,0,0,0,0,0,0,0,0,0,144,3,0,0,81,1,0,0,82,1,0,0,65,0,0,0,83,1,0,0,91,0,0,0,0,0,0,0,44,0,0,0,32,0,0,0,0,0,0,0,93,0,0,0,0,0,0,0,0,0,0,0,160,3,0,0,84,1,0,0,85,1,0,0,65,0,0,0,86,1,0,0,84,0,0,0,111,0,0,0,116,0,0,0,97,0,0,0,108,0,0,0,80,0,0,0,97,0,0,0,99,0,0,0,107,0,0,0,101,0,0,0,116,0,0,0,66,0,0,0,121,0,0,0,116,0,0,0,101,0,0,0,115,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,84,0,0,0,111,0,0,0,116,0,0,0,97,0,0,0,108,0,0,0,67,0,0,0,111,0,0,0,109,0,0,0,109,0,0,0,97,0,0,0,110,0,0,0,100,0,0,0,66,0,0,0,121,0,0,0,116,0,0,0,101,0,0,0,115,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,84,0,0,0,111,0,0,0,116,0,0,0,97,0,0,0,108,0,0,0,80,0,0,0,97,0,0,0,99,0,0,0,107,0,0,0,101,0,0,0,116,0,0,0,67,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,116,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,84,0,0,0,111,0,0,0,116,0,0,0,97,0,0,0,108,0,0,0,67,0,0,0,111,0,0,0,109,0,0,0,109,0,0,0,97,0,0,0,110,0,0,0,100,0,0,0,115,0,0,0,73,0,0,0,110,0,0,0,80,0,0,0,97,0,0,0,99,0,0,0,107,0,0,0,101,0,0,0,116,0,0,0,115,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,0,0,0,0,192,3,0,0,87,1,0,0,88,1,0,0,65,0,0,0,89,1,0,0,90,1,0,0,76,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,101,0,0,0,115,0,0,0,116,0,0,0,32,0,0,0,100,0,0,0,101,0,0,0,108,0,0,0,116,0,0,0,97,0,0,0,32,0,0,0,98,0,0,0,101,0,0,0,116,0,0,0,119,0,0,0,101,0,0,0,101,0,0,0,110,0,0,0,32,0,0,0,83,0,0,0,101,0,0,0,110,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,109,0,0,0,115,0,0,0,32,0,0,0,68,0,0,0,105,0,0,0,115,0,0,0,112,0,0,0,97,0,0,0,116,0,0,0,99,0,0,0,104,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,109,0,0,0,115,0,0,0,46,0,0,0,32,0,0,0,76,0,0,0,111,0,0,0,110,0,0,0,103,0,0,0,101,0,0,0,115,0,0,0,116,0,0,0,32,0,0,0,99,0,0,0,97,0,0,0,108,0,0,0,108,0,0,0,98,0,0,0,97,0,0,0,99,0,0,0,107,0,0,0,32,0,0,0,79,0,0,0,110,0,0,0,69,0,0,0,118,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,61,0,0,0,0,0,0,0,109,0,0,0,115,0,0,0,32,0,0,0,79,0,0,0,110,0,0,0,82,0,0,0,101,0,0,0,115,0,0,0,112,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,109,0,0,0,115,0,0,0,46,0,0,0,32,0,0,0,67,0,0,0,97,0,0,0,108,0,0,0,108,0,0,0,115,0,0,0,32,0,0,0,111,0,0,0,102,0,0,0,32,0,0,0,83,0,0,0,101,0,0,0,110,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,68,0,0,0,105,0,0,0,115,0,0,0,112,0,0,0,97,0,0,0,116,0,0,0,99,0,0,0,104,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,46,0,0,0,0,0,0,0,79,0,0,0,112,0,0,0,101,0,0,0,114,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,66,0,0,0,121,0,0,0,116,0,0,0,101,0,0,0,67,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,116,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,82,0,0,0,101,0,0,0,115,0,0,0,117,0,0,0,108,0,0,0,116,0,0,0,66,0,0,0,121,0,0,0,116,0,0,0,101,0,0,0,67,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,116,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,69,0,0,0,118,0,0,0,101,0,0,0,110,0,0,0,116,0,0,0,66,0,0,0,121,0,0,0,116,0,0,0,101,0,0,0,67,0,0,0,111,0,0,0,117,0,0,0,110,0,0,0,116,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,58,0,0,0,47,0,0,0,47,0,0,0,0,0,0,0,119,0,0,0,115,0,0,0,115,0,0,0,58,0,0,0,47,0,0,0,47,0,0,0,0,0,0,0,119,0,0,0,115,0,0,0,58,0,0,0,47,0,0,0,47,0,0,0,0,0,0,0,0,0,0,0,208,3,0,0,91,1,0,0,92,1,0,0,65,0,0,0,89,1,0,0,90,1,0,0,0,0,0,0,176,3,0,0,93,1,0,0,94,1,0,0,65,0,0,0,86,1,0,0,82,0,0,0,116,0,0,0,116,0,0,0,40,0,0,0,118,0,0,0,97,0,0,0,114,0,0,0,105,0,0,0,97,0,0,0,110,0,0,0,99,0,0,0,101,0,0,0,41,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,40,0,0,0,0,0,0,0,41,0,0,0,46,0,0,0,32,0,0,0,77,0,0,0,115,0,0,0,32,0,0,0,115,0,0,0,105,0,0,0,110,0,0,0,99,0,0,0,101,0,0,0,32,0,0,0,108,0,0,0,97,0,0,0,115,0,0,0,116,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,99,0,0,0,101,0,0,0,105,0,0,0,118,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,46,0,0,0,32,0,0,0,83,0,0,0,116,0,0,0,97,0,0,0,116,0,0,0,115,0,0,0,32,0,0,0,101,0,0,0,108,0,0,0,97,0,0,0,112,0,0,0,115,0,0,0,101,0,0,0,100,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,115,0,0,0,101,0,0,0,99,0,0,0,46,0,0,0,32,0,0,0,0,0,0,0,73,0,0,0,110,0,0,0,99,0,0,0,111,0,0,0,109,0,0,0,105,0,0,0,110,0,0,0,103,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,46,0,0,0,32,0,0,0,79,0,0,0,117,0,0,0,116,0,0,0,103,0,0,0,111,0,0,0,105,0,0,0,110,0,0,0,103,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,115,0,0,0,114,0,0,0,99,0,0,0,47,0,0,0,80,0,0,0,104,0,0,0,111,0,0,0,116,0,0,0,111,0,0,0,110,0,0,0,80,0,0,0,101,0,0,0,101,0,0,0,114,0,0,0,46,0,0,0,99,0,0,0,112,0,0,0,112,0,0,0,0,0,0,0,0,0,0,0,224,3,0,0,95,1,0,0,96,1,0,0,97,1,0,0,98,1,0,0,99,1,0,0,100,1,0,0,101,1,0,0,102,1,0,0,103,1,0,0,104,1,0,0,105,1,0,0,106,1,0,0,107,1,0,0,108,1,0,0,109,1,0,0,110,1,0,0,111,1,0,0,79,0,0,0,112,0,0,0,101,0,0,0,114,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,82,0,0,0,101,0,0,0,115,0,0,0,112,0,0,0,111,0,0,0,110,0,0,0,115,0,0,0,101,0,0,0,32,0,0,0,45,0,0,0,32,0,0,0,111,0,0,0,112,0,0,0,101,0,0,0,114,0,0,0,97,0,0,0,116,0,0,0,105,0,0,0,111,0,0,0,110,0,0,0,67,0,0,0,111,0,0,0,100,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,44,0,0,0,32,0,0,0,114,0,0,0,101,0,0,0,116,0,0,0,117,0,0,0,114,0,0,0,110,0,0,0,67,0,0,0,111,0,0,0,100,0,0,0,101,0,0,0,58,0,0,0,32,0,0,0,0,0,0,0,32,0,0,0,40,0,0,0,0,0,0,0,41,0,0,0,0,0,0,0,32,0,0,0,0,0,0,0,44,56,0,0,20,0,0,0,67,46,85,84,70,45,56,0,0,0,0,0,0,0,0,0,0,0,0,0,222,18,4,149,0,0,0,0,255,255,255,255,255,255,255,255,255,255,255,255,16,56,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,192,3,0,0,192,4,0,0,192,5,0,0,192,6,0,0,192,7,0,0,192,8,0,0,192,9,0,0,192,10,0,0,192,11,0,0,192,12,0,0,192,13,0,0,192,14,0,0,192,15,0,0,192,16,0,0,192,17,0,0,192,18,0,0,192,19,0,0,192,20,0,0,192,21,0,0,192,22,0,0,192,23,0,0,192,24,0,0,192,25,0,0,192,26,0,0,192,27,0,0,192,28,0,0,192,29,0,0,192,30,0,0,192,31,0,0,192,0,0,0,179,1,0,0,195,2,0,0,195,3,0,0,195,4,0,0,195,5,0,0,195,6,0,0,195,7,0,0,195,8,0,0,195,9,0,0,195,10,0,0,195,11,0,0,195,12,0,0,195,13,0,0,211,14,0,0,195,15,0,0,195,0,0,12,187,1,0,12,195,2,0,12,195,3,0,12,195,4,0,12,211,40,57,0,0,5,0,0,0,0,0,0,0,0,0,0,0,112,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,113,1,0,0,114,1,0,0,224,141,2,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,216,129,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,156,58,0,0,9,0,0,0,0,0,0,0,0,0,0,0,112,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,115,1,0,0,0,0,0,0,114,1,0,0,232,141,2,0,0,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,28,59,0,0,5,0,0,0,0,0,0,0,0,0,0,0,112,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,116,1,0,0,114,1,0,0,240,145,2,0,0,4,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,28,59,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,117,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,255,255,255,255,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,252,87,0,0,32,62,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,30,0,0,0,31,0,0,0,32,0,0,0,33,0,0,0,34,0,0,0,35,0,0,0,36,0,0,0,37,0,0,0,38,0,0,0,39,0,0,0,40,0,0,0,41,0,0,0,42,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,47,0,0,0,48,0,0,0,49,0,0,0,50,0,0,0,51,0,0,0,52,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,63,0,0,0,64,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,100,0,0,0,101,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,91,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,95,0,0,0,96,0,0,0,97,0,0,0,98,0,0,0,99,0,0,0,100,0,0,0,101,0,0,0,102,0,0,0,103,0,0,0,104,0,0,0,105,0,0,0,106,0,0,0,107,0,0,0,108,0,0,0,109,0,0,0,110,0,0,0,111,0,0,0,112,0,0,0,113,0,0,0,114,0,0,0,115,0,0,0,116,0,0,0,117,0,0,0,118,0,0,0,119,0,0,0,120,0,0,0,121,0,0,0,122,0,0,0,123,0,0,0,124,0,0,0,125,0,0,0,126,0,0,0,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,36,68,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,2,0,0,0,3,0,0,0,4,0,0,0,5,0,0,0,6,0,0,0,7,0,0,0,8,0,0,0,9,0,0,0,10,0,0,0,11,0,0,0,12,0,0,0,13,0,0,0,14,0,0,0,15,0,0,0,16,0,0,0,17,0,0,0,18,0,0,0,19,0,0,0,20,0,0,0,21,0,0,0,22,0,0,0,23,0,0,0,24,0,0,0,25,0,0,0,26,0,0,0,27,0,0,0,28,0,0,0,29,0,0,0,30,0,0,0,31,0,0,0,32,0,0,0,33,0,0,0,34,0,0,0,35,0,0,0,36,0,0,0,37,0,0,0,38,0,0,0,39,0,0,0,40,0,0,0,41,0,0,0,42,0,0,0,43,0,0,0,44,0,0,0,45,0,0,0,46,0,0,0,47,0,0,0,48,0,0,0,49,0,0,0,50,0,0,0,51,0,0,0,52,0,0,0,53,0,0,0,54,0,0,0,55,0,0,0,56,0,0,0,57,0,0,0,58,0,0,0,59,0,0,0,60,0,0,0,61,0,0,0,62,0,0,0,63,0,0,0,64,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,70,0,0,0,71,0,0,0,72,0,0,0,73,0,0,0,74,0,0,0,75,0,0,0,76,0,0,0,77,0,0,0,78,0,0,0,79,0,0,0,80,0,0,0,81,0,0,0,82,0,0,0,83,0,0,0,84,0,0,0,85,0,0,0,86,0,0,0,87,0,0,0,88,0,0,0,89,0,0,0,90,0,0,0,91,0,0,0,92,0,0,0,93,0,0,0,94,0,0,0,95,0,0,0,96,0,0,0,65,0,0,0,66,0,0,0,67,0,0,0,68,0,0,0,69,0,0,0,70,0,0,0,71,0,0,0,72,0,0,0,73,0,0,0,74,0,0,0,75,0,0,0,76,0,0,0,77,0,0,0,78,0,0,0,79,0,0,0,80,0,0,0,81,0,0,0,82,0,0,0,83,0,0,0,84,0,0,0,85,0,0,0,86,0,0,0,87,0,0,0,88,0,0,0,89,0,0,0,90,0,0,0,123,0,0,0,124,0,0,0,125,0,0,0,126,0,0,0,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,100,0,0,0,232,3,0,0,16,39,0,0,160,134,1,0,64,66,15,0,128,150,152,0,0,225,245,5,95,112,137,0,255,9,47,15,0,0,0,0,248,3,0,0,118,1,0,0,119,1,0,0,0,0,0,0,16,4,0,0,120,1,0,0,121,1,0,0,122,1,0,0,123,1,0,0,124,1,0,0,125,1,0,0,126,1,0,0,127,1,0,0,128,1,0,0,129,1,0,0,130,1,0,0,131,1,0,0,132,1,0,0,133,1,0,0,0,0,0,0,24,4,0,0,134,1,0,0,135,1,0,0,136,1,0,0,137,1,0,0,138,1,0,0,139,1,0,0,140,1,0,0,141,1,0,0,142,1,0,0,143,1,0,0,144,1,0,0,145,1,0,0,146,1,0,0,147,1,0,0,8,0,0,0,0,0,0,0,32,4,0,0,148,1,0,0,149,1,0,0,248,255,255,255,248,255,255,255,32,4,0,0,150,1,0,0,151,1,0,0,232,72,0,0,252,72,0,0,8,0,0,0,0,0,0,0,56,4,0,0,152,1,0,0,153,1,0,0,248,255,255,255,248,255,255,255,56,4,0,0,154,1,0,0,155,1,0,0,24,73,0,0,44,73,0,0,4,0,0,0,0,0,0,0,80,4,0,0,156,1,0,0,157,1,0,0,252,255,255,255,252,255,255,255,80,4,0,0,158,1,0,0,159,1,0,0,72,73,0,0,92,73,0,0,4,0,0,0,0,0,0,0,104,4,0,0,160,1,0,0,161,1,0,0,252,255,255,255,252,255,255,255,104,4,0,0,162,1,0,0,163,1,0,0,120,73,0,0,140,73,0,0,0,0,0,0,128,4,0,0,164,1,0,0,165,1,0,0,166,1,0,0,167,1,0,0,168,1,0,0,169,1,0,0,170,1,0,0,0,0,0,0,144,4,0,0,171,1,0,0,172,1,0,0,173,1,0,0,164,73,0,0,0,0,0,0,160,4,0,0,174,1,0,0,175,1,0,0,176,1,0,0,137,1,0,0,138,1,0,0,139,1,0,0,177,1,0,0,141,1,0,0,142,1,0,0,143,1,0,0,144,1,0,0,145,1,0,0,178,1,0,0,179,1,0,0,0,0,0,0,176,4,0,0,180,1,0,0,181,1,0,0,182,1,0,0,123,1,0,0,124,1,0,0,125,1,0,0,183,1,0,0,127,1,0,0,128,1,0,0,129,1,0,0,130,1,0,0,131,1,0,0,184,1,0,0,185,1,0,0,0,0,0,0,192,4,0,0,186,1,0,0,187,1,0,0,188,1,0,0,137,1,0,0,138,1,0,0,139,1,0,0,140,1,0,0,141,1,0,0,142,1,0,0,189,1,0,0,190,1,0,0,191,1,0,0,146,1,0,0,147,1,0,0,0,0,0,0,208,4,0,0,192,1,0,0,193,1,0,0,194,1,0,0,123,1,0,0,124,1,0,0,125,1,0,0,126,1,0,0,127,1,0,0,128,1,0,0,195,1,0,0,196,1,0,0,197,1,0,0,132,1,0,0,133,1,0,0,0,0,0,0,224,4,0,0,198,1,0,0,199,1,0,0,200,1,0,0,201,1,0,0,202,1,0,0,203,1,0,0,0,0,0,0,0,5,0,0,204,1,0,0,205,1,0,0,200,1,0,0,206,1,0,0,207,1,0,0,208,1,0,0,0,0,0,0,16,5,0,0,209,1,0,0,210,1,0,0,200,1,0,0,211,1,0,0,212,1,0,0,213,1,0,0,214,1,0,0,215,1,0,0,216,1,0,0,217,1,0,0,218,1,0,0,219,1,0,0,220,1,0,0,221,1,0,0,0,0,0,0,80,5,0,0,222,1,0,0,223,1,0,0,200,1,0,0,224,1,0,0,225,1,0,0,226,1,0,0,227,1,0,0,228,1,0,0,229,1,0,0,230,1,0,0,231,1,0,0,232,1,0,0,233,1,0,0,234,1,0,0,0,0,0,0,136,5,0,0,235,1,0,0,236,1,0,0,200,1,0,0,237,1,0,0,238,1,0,0,239,1,0,0,240,1,0,0,241,1,0,0,242,1,0,0,243,1,0,0,244,1,0,0,0,0,0,0,200,5,0,0,245,1,0,0,246,1,0,0,200,1,0,0,247,1,0,0,248,1,0,0,249,1,0,0,250,1,0,0,251,1,0,0,252,1,0,0,253,1,0,0,254,1,0,0,0,0,0,0,0,6,0,0,255,1,0,0,0,2,0,0,200,1,0,0,1,2,0,0,2,2,0,0,3,2,0,0,4,2,0,0,5,2,0,0,6,2,0,0,7,2,0,0,248,255,255,255,0,6,0,0,8,2,0,0,9,2,0,0,10,2,0,0,11,2,0,0,12,2,0,0,13,2,0,0,14,2,0,0,0,0,0,0,56,6,0,0,15,2,0,0,16,2,0,0,200,1,0,0,17,2,0,0,18,2,0,0,19,2,0,0,20,2,0,0,21,2,0,0,22,2,0,0,23,2,0,0,248,255,255,255,56,6,0,0,24,2,0,0,25,2,0,0,26,2,0,0,27,2,0,0,28,2,0,0,29,2,0,0,30,2,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,0,0,0,0,37,0,0,0,109,0,0,0,47,0,0,0,37,0,0,0,100,0,0,0,47,0,0,0,37,0,0,0,121,0,0,0,0,0,0,0,37,0,0,0,73,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,112,0,0,0,0,0,0,0,37,0,0,0,97,0,0,0,32,0,0,0,37,0,0,0,98,0,0,0,32,0,0,0,37,0,0,0,100,0,0,0,32,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,89,0,0,0,0,0,0,0,65,0,0,0,77,0,0,0,0,0,0,0,80,0,0,0,77,0,0,0,0,0,0,0,74,0,0,0,97,0,0,0,110,0,0,0,117,0,0,0,97,0,0,0,114,0,0,0,121,0,0,0,0,0,0,0,70,0,0,0,101,0,0,0,98,0,0,0,114,0,0,0,117,0,0,0,97,0,0,0,114,0,0,0,121,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,114,0,0,0,99,0,0,0,104,0,0,0,0,0,0,0,65,0,0,0,112,0,0,0,114,0,0,0,105,0,0,0,108,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,110,0,0,0,101,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,108,0,0,0,121,0,0,0,0,0,0,0,65,0,0,0,117,0,0,0,103,0,0,0,117,0,0,0,115,0,0,0,116,0,0,0,0,0,0,0,83,0,0,0,101,0,0,0,112,0,0,0,116,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,79,0,0,0,99,0,0,0,116,0,0,0,111,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,78,0,0,0,111,0,0,0,118,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,68,0,0,0,101,0,0,0,99,0,0,0,101,0,0,0,109,0,0,0,98,0,0,0,101,0,0,0,114,0,0,0,0,0,0,0,74,0,0,0,97,0,0,0,110,0,0,0,0,0,0,0,70,0,0,0,101,0,0,0,98,0,0,0,0,0,0,0,77,0,0,0,97,0,0,0,114,0,0,0,0,0,0,0,65,0,0,0,112,0,0,0,114,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,110,0,0,0,0,0,0,0,74,0,0,0,117,0,0,0,108,0,0,0,0,0,0,0,65,0,0,0,117,0,0,0,103,0,0,0,0,0,0,0,83,0,0,0,101,0,0,0,112,0,0,0,0,0,0,0,79,0,0,0,99,0,0,0,116,0,0,0,0,0,0,0,78,0,0,0,111,0,0,0,118,0,0,0,0,0,0,0,68,0,0,0,101,0,0,0,99,0,0,0,0,0,0,0,83,0,0,0,117,0,0,0,110,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,77,0,0,0,111,0,0,0,110,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,84,0,0,0,117,0,0,0,101,0,0,0,115,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,87,0,0,0,101,0,0,0,100,0,0,0,110,0,0,0,101,0,0,0,115,0,0,0,100], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+10240);
/* memory initializer */ allocate([97,0,0,0,121,0,0,0,0,0,0,0,84,0,0,0,104,0,0,0,117,0,0,0,114,0,0,0,115,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,70,0,0,0,114,0,0,0,105,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,83,0,0,0,97,0,0,0,116,0,0,0,117,0,0,0,114,0,0,0,100,0,0,0,97,0,0,0,121,0,0,0,0,0,0,0,83,0,0,0,117,0,0,0,110,0,0,0,0,0,0,0,77,0,0,0,111,0,0,0,110,0,0,0,0,0,0,0,84,0,0,0,117,0,0,0,101,0,0,0,0,0,0,0,87,0,0,0,101,0,0,0,100,0,0,0,0,0,0,0,84,0,0,0,104,0,0,0,117,0,0,0,0,0,0,0,70,0,0,0,114,0,0,0,105,0,0,0,0,0,0,0,83,0,0,0,97,0,0,0,116,0,0,0,0,0,0,0,37,0,0,0,109,0,0,0,47,0,0,0,37,0,0,0,100,0,0,0,47,0,0,0,37,0,0,0,121,0,0,0,37,0,0,0,89,0,0,0,45,0,0,0,37,0,0,0,109,0,0,0,45,0,0,0,37,0,0,0,100,0,0,0,37,0,0,0,73,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,32,0,0,0,37,0,0,0,112,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,37,0,0,0,72,0,0,0,58,0,0,0,37,0,0,0,77,0,0,0,58,0,0,0,37,0,0,0,83,0,0,0,0,0,0,0,104,6,0,0,31,2,0,0,32,2,0,0,200,1,0,0,33,2,0,0,0,0,0,0,144,6,0,0,34,2,0,0,35,2,0,0,200,1,0,0,36,2,0,0,0,0,0,0,176,6,0,0,37,2,0,0,38,2,0,0,200,1,0,0,39,2,0,0,40,2,0,0,41,2,0,0,42,2,0,0,43,2,0,0,44,2,0,0,45,2,0,0,46,2,0,0,47,2,0,0,0,0,0,0,216,6,0,0,48,2,0,0,49,2,0,0,200,1,0,0,50,2,0,0,51,2,0,0,52,2,0,0,53,2,0,0,54,2,0,0,55,2,0,0,56,2,0,0,57,2,0,0,58,2,0,0,0,0,0,0,248,6,0,0,59,2,0,0,60,2,0,0,200,1,0,0,61,2,0,0,62,2,0,0,63,2,0,0,64,2,0,0,65,2,0,0,66,2,0,0,67,2,0,0,68,2,0,0,69,2,0,0,0,0,0,0,24,7,0,0,70,2,0,0,71,2,0,0,200,1,0,0,72,2,0,0,73,2,0,0,74,2,0,0,75,2,0,0,76,2,0,0,77,2,0,0,78,2,0,0,79,2,0,0,80,2,0,0,0,0,0,0,56,7,0,0,81,2,0,0,82,2,0,0,200,1,0,0,83,2,0,0,84,2,0,0,0,0,0,0,96,7,0,0,85,2,0,0,86,2,0,0,200,1,0,0,87,2,0,0,88,2,0,0,0,0,0,0,136,7,0,0,89,2,0,0,90,2,0,0,200,1,0,0,91,2,0,0,92,2,0,0,0,0,0,0,176,7,0,0,93,2,0,0,94,2,0,0,200,1,0,0,95,2,0,0,96,2,0,0,0,0,0,0,216,7,0,0,97,2,0,0,98,2,0,0,200,1,0,0,99,2,0,0,100,2,0,0,101,2,0,0,0,0,0,0,0,8,0,0,102,2,0,0,103,2,0,0,200,1,0,0,104,2,0,0,105,2,0,0,106,2,0,0,0,0,0,0,88,8,0,0,107,2,0,0,108,2,0,0,200,1,0,0,109,2,0,0,110,2,0,0,111,2,0,0,112,2,0,0,113,2,0,0,114,2,0,0,115,2,0,0,0,0,0,0,32,8,0,0,107,2,0,0,116,2,0,0,200,1,0,0,109,2,0,0,110,2,0,0,111,2,0,0,112,2,0,0,113,2,0,0,114,2,0,0,115,2,0,0,0,0,0,0,136,8,0,0,117,2,0,0,118,2,0,0,200,1,0,0,119,2,0,0,120,2,0,0,121,2,0,0,122,2,0,0,123,2,0,0,124,2,0,0,125,2,0,0,0,0,0,0,200,8,0,0,126,2,0,0,127,2,0,0,200,1,0,0,0,0,0,0,216,8,0,0,128,2,0,0,129,2,0,0,200,1,0,0,130,2,0,0,131,2,0,0,132,2,0,0,133,2,0,0,134,2,0,0,135,2,0,0,136,2,0,0,137,2,0,0,0,0,0,0,32,9,0,0,138,2,0,0,139,2,0,0,200,1,0,0,140,2,0,0,141,2,0,0,142,2,0,0,143,2,0,0,144,2,0,0,0,0,0,0,48,9,0,0,145,2,0,0,146,2,0,0,200,1,0,0,147,2,0,0,148,2,0,0,149,2,0,0,150,2,0,0,151,2,0,0,102,0,0,0,97,0,0,0,108,0,0,0,115,0,0,0,101,0,0,0,0,0,0,0,116,0,0,0,114,0,0,0,117,0,0,0,101,0,0,0,0,0,0,0,0,0,0,0,240,4,0,0,107,2,0,0,152,2,0,0,200,1,0,0,0,0,0,0,0,9,0,0,107,2,0,0,153,2,0,0,200,1,0,0,154,2,0,0,155,2,0,0,156,2,0,0,157,2,0,0,158,2,0,0,159,2,0,0,160,2,0,0,161,2,0,0,162,2,0,0,163,2,0,0,164,2,0,0,165,2,0,0,0,0,0,0,104,8,0,0,107,2,0,0,166,2,0,0,200,1,0,0,167,2,0,0,168,2,0,0,169,2,0,0,170,2,0,0,171,2,0,0,172,2,0,0,173,2,0,0,0,0,0,0,168,8,0,0,107,2,0,0,174,2,0,0,200,1,0,0,175,2,0,0,176,2,0,0,177,2,0,0,178,2,0,0,179,2,0,0,180,2,0,0,181,2,0,0,0,0,0,0,48,8,0,0,107,2,0,0,182,2,0,0,200,1,0,0,109,2,0,0,110,2,0,0,111,2,0,0,112,2,0,0,113,2,0,0,114,2,0,0,115,2,0,0,0,0,0,0,48,6,0,0,8,2,0,0,9,2,0,0,10,2,0,0,11,2,0,0,12,2,0,0,13,2,0,0,14,2,0,0,0,0,0,0,96,6,0,0,24,2,0,0,25,2,0,0,26,2,0,0,27,2,0,0,28,2,0,0,29,2,0,0,30,2,0,0,0,0,0,0,96,9,0,0,183,2,0,0,184,2,0,0,173,1,0,0,185,2,0,0,209,106,2,0,0,0,0,0,120,9,0,0,186,2,0,0,187,2,0,0,188,2,0,0,189,2,0,0,190,2,0,0,191,2,0,0,192,2,0,0,193,2,0,0,0,0,0,0,160,9,0,0,186,2,0,0,194,2,0,0,188,2,0,0,189,2,0,0,190,2,0,0,195,2,0,0,196,2,0,0,197,2,0,0,0,0,0,0,176,9,0,0,198,2,0,0,199,2,0,0,200,2,0,0,0,0,0,0,192,9,0,0,201,2,0,0,202,2,0,0,203,2,0,0,0,0,0,0,208,9,0,0,204,2,0,0,205,2,0,0,173,1,0,0,0,0,0,0,224,9,0,0,201,2,0,0,206,2,0,0,203,2,0,0,0,0,0,0,240,9,0,0,201,2,0,0,207,2,0,0,203,2,0,0,0,0,0,0,0,10,0,0,208,2,0,0,209,2,0,0,210,2,0,0,0,0,0,0,48,10,0,0,186,2,0,0,211,2,0,0,188,2,0,0,189,2,0,0,190,2,0,0,212,2,0,0,213,2,0,0,214,2,0,0,146,74,130,35,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,3,32,2,32,2,32,2,32,2,32,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,1,96,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,4,192,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,8,216,4,192,4,192,4,192,4,192,4,192,4,192,4,192,8,213,8,213,8,213,8,213,8,213,8,213,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,8,197,4,192,4,192,4,192,4,192,4,192,4,192,8,214,8,214,8,214,8,214,8,214,8,214,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,8,198,4,192,4,192,4,192,4,192,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,48,1,48,2,48,12,255,14,255,251,48,26,255,27,255,31,255,1,255,155,48,156,48,180,0,64,255,168,0,62,255,227,255,63,255,253,48,254,48,157,48,158,48,3,48,221,78,5,48,6,48,7,48,252,48,21,32,16,32,15,255,92,0,28,48,22,32,92,255,38,32,37,32,24,32,25,32,28,32,29,32,8,255,9,255,20,48,21,48,59,255,61,255,91,255,93,255,8,48,9,48,10,48,11,48,12,48,13,48,14,48,15,48,16,48,17,48,11,255,18,34,177,0,215,0,247,0,29,255,96,34,28,255,30,255,102,34,103,34,30,34,52,34,66,38,64,38,176,0,50,32,51,32,3,33,229,255,4,255,162,0,163,0,5,255,3,255,6,255,10,255,32,255,167,0,6,38,5,38,203,37,207,37,206,37,199,37,198,37,161,37,160,37,179,37,178,37,189,37,188,37,59,32,18,48,146,33,144,33,145,33,147,33,19,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8,34,11,34,134,34,135,34,130,34,131,34,42,34,41,34,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,39,34,40,34,172,0,210,33,212,33,0,34,3,34,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,34,165,34,18,35,2,34,7,34,97,34,82,34,106,34,107,34,26,34,61,34,29,34,53,34,43,34,44,34,0,0,0,0,0,0,0,0,0,0,0,0,0,0,43,33,48,32,111,38,109,38,106,38,32,32,33,32,182,0,0,0,0,0,0,0,0,0,239,37,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,255,17,255,18,255,19,255,20,255,21,255,22,255,23,255,24,255,25,255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33,255,34,255,35,255,36,255,37,255,38,255,39,255,40,255,41,255,42,255,43,255,44,255,45,255,46,255,47,255,48,255,49,255,50,255,51,255,52,255,53,255,54,255,55,255,56,255,57,255,58,255,0,0,0,0,0,0,0,0,0,0,0,0,65,255,66,255,67,255,68,255,69,255,70,255,71,255,72,255,73,255,74,255,75,255,76,255,77,255,78,255,79,255,80,255,81,255,82,255,83,255,84,255,85,255,86,255,87,255,88,255,89,255,90,255,0,0,0,0,0,0,0,0,65,48,66,48,67,48,68,48,69,48,70,48,71,48,72,48,73,48,74,48,75,48,76,48,77,48,78,48,79,48,80,48,81,48,82,48,83,48,84,48,85,48,86,48,87,48,88,48,89,48,90,48,91,48,92,48,93,48,94,48,95,48,96,48,97,48,98,48,99,48,100,48,101,48,102,48,103,48,104,48,105,48,106,48,107,48,108,48,109,48,110,48,111,48,112,48,113,48,114,48,115,48,116,48,117,48,118,48,119,48,120,48,121,48,122,48,123,48,124,48,125,48,126,48,127,48,128,48,129,48,130,48,131,48,132,48,133,48,134,48,135,48,136,48,137,48,138,48,139,48,140,48,141,48,142,48,143,48,144,48,145,48,146,48,147,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161,48,162,48,163,48,164,48,165,48,166,48,167,48,168,48,169,48,170,48,171,48,172,48,173,48,174,48,175,48,176,48,177,48,178,48,179,48,180,48,181,48,182,48,183,48,184,48,185,48,186,48,187,48,188,48,189,48,190,48,191,48,192,48,193,48,194,48,195,48,196,48,197,48,198,48,199,48,200,48,201,48,202,48,203,48,204,48,205,48,206,48,207,48,208,48,209,48,210,48,211,48,212,48,213,48,214,48,215,48,216,48,217,48,218,48,219,48,220,48,221,48,222,48,223,48,224,48,225,48,226,48,227,48,228,48,229,48,230,48,231,48,232,48,233,48,234,48,235,48,236,48,237,48,238,48,239,48,240,48,241,48,242,48,243,48,244,48,245,48,246,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,145,3,146,3,147,3,148,3,149,3,150,3,151,3,152,3,153,3,154,3,155,3,156,3,157,3,158,3,159,3,160,3,161,3,163,3,164,3,165,3,166,3,167,3,168,3,169,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,177,3,178,3,179,3,180,3,181,3,182,3,183,3,184,3,185,3,186,3,187,3,188,3,189,3,190,3,191,3,192,3,193,3,195,3,196,3,197,3,198,3,199,3,200,3,201,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,4,17,4,18,4,19,4,20,4,21,4,1,4,22,4,23,4,24,4,25,4,26,4,27,4,28,4,29,4,30,4,31,4,32,4,33,4,34,4,35,4,36,4,37,4,38,4,39,4,40,4,41,4,42,4,43,4,44,4,45,4,46,4,47,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,48,4,49,4,50,4,51,4,52,4,53,4,81,4,54,4,55,4,56,4,57,4,58,4,59,4,60,4,61,4,62,4,63,4,64,4,65,4,66,4,67,4,68,4,69,4,70,4,71,4,72,4,73,4,74,4,75,4,76,4,77,4,78,4,79,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,37,2,37,12,37,16,37,24,37,20,37,28,37,44,37,36,37,52,37,60,37,1,37,3,37,15,37,19,37,27,37,23,37,35,37,51,37,43,37,59,37,75,37,32,37,47,37,40,37,55,37,63,37,29,37,48,37,37,37,56,37,66,37], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+20480);
/* memory initializer */ allocate([156,78,22,85,3,90,63,150,192,84,27,97,40,99,246,89,34,144,117,132,28,131,80,122,170,96,225,99,37,110,237,101,102,132,166,130,245,155,147,104,39,87,161,101,113,98,155,91,208,89,123,134,244,152,98,125,190,125,142,155,22,98,159,124,183,136,137,91,181,94,9,99,151,102,72,104,199,149,141,151,79,103,229,78,10,79,77,79,157,79,73,80,242,86,55,89,212,89,1,90,9,92,223,96,15,97,112,97,19,102,5,105,186,112,79,117,112,117,251,121,173,125,239,125,195,128,14,132,99,136,2,139,85,144,122,144,59,83,149,78,165,78,223,87,178,128,193,144,239,120,0,78,241,88,162,110,56,144,50,122,40,131,139,130,47,156,65,81,112,83,189,84,225,84,224,86,251,89,21,95,242,152,235,109,228,128,45,133,98,150,112,150,160,150,251,151,11,84,243,83,135,91,207,112,189,127,194,143,232,150,111,83,92,157,186,122,17,78,147,120,252,129,38,110,24,86,4,85,29,107,26,133,59,156,229,89,169,83,102,109,220,116,143,149,66,86,145,78,75,144,242,150,79,131,12,153,225,83,182,85,48,91,113,95,32,102,243,102,4,104,56,108,243,108,41,109,91,116,200,118,78,122,52,152,241,130,91,136,96,138,237,146,178,109,171,117,202,118,197,153,166,96,1,139,138,141,178,149,142,105,173,83,134,81,18,87,48,88,68,89,180,91,246,94,40,96,169,99,244,99,191,108,20,111,142,112,20,113,89,113,213,113,63,115,1,126,118,130,209,130,151,133,96,144,91,146,27,157,105,88,188,101,90,108,37,117,249,81,46,89,101,89,128,95,220,95,188,98,250,101,42,106,39,107,180,107,139,115,193,127,86,137,44,157,14,157,196,158,161,92,150,108,123,131,4,81,75,92,182,97,198,129,118,104,97,114,89,78,250,79,120,83,105,96,41,110,79,122,243,151,11,78,22,83,238,78,85,79,61,79,161,79,115,79,160,82,239,83,9,86,15,89,193,90,182,91,225,91,209,121,135,102,156,103,182,103,76,107,179,108,107,112,194,115,141,121,190,121,60,122,135,123,177,130,219,130,4,131,119,131,239,131,211,131,102,135,178,138,41,86,168,140,230,143,78,144,30,151,138,134,196,79,232,92,17,98,89,114,59,117,229,129,189,130,254,134,192,140,197,150,19,153,213,153,203,78,26,79,227,137,222,86,74,88,202,88,251,94,235,95,42,96,148,96,98,96,208,97,18,98,208,98,57,101,65,155,102,102,176,104,119,109,112,112,76,117,134,118,117,125,165,130,249,135,139,149,142,150,157,140,241,81,190,82,22,89,179,84,179,91,22,93,104,97,130,105,175,109,141,120,203,132,87,136,114,138,167,147,184,154,108,109,168,153,217,134,163,87,255,103,206,134,14,146,131,82,135,86,4,84,211,94,225,98,185,100,60,104,56,104,187,107,114,115,186,120,107,122,154,137,210,137,107,141,3,143,237,144,163,149,148,150,105,151,102,91,179,92,125,105,77,152,78,152,155,99,32,123,43,106,127,106,182,104,13,156,95,111,114,82,157,85,112,96,236,98,59,109,7,110,209,110,91,132,16,137,68,143,20,78,57,156,246,83,27,105,58,106,132,151,42,104,92,81,195,122,178,132,220,145,140,147,91,86,40,157,34,104,5,131,49,132,165,124,8,82,197,130,230,116,126,78,131,79,160,81,210,91,10,82,216,82,231,82,251,93,154,85,42,88,230,89,140,91,152,91,219,91,114,94,121,94,163,96,31,97,99,97,190,97,219,99,98,101,209,103,83,104,250,104,62,107,83,107,87,108,34,111,151,111,69,111,176,116,24,117,227,118,11,119,255,122,161,123,33,124,233,125,54,127,240,127,157,128,102,130,158,131,179,137,204,138,171,140,132,144,81,148,147,149,145,149,162,149,101,150,211,151,40,153,24,130,56,78,43,84,184,92,204,93,169,115,76,118,60,119,169,92,235,127,11,141,193,150,17,152,84,152,88,152,1,79,14,79,113,83,156,85,104,86,250,87,71,89,9,91,196,91,144,92,12,94,126,94,204,95,238,99,58,103,215,101,226,101,31,103,203,104,196,104,95,106,48,94,197,107,23,108,125,108,127,117,72,121,99,91,0,122,0,125,189,95,143,137,24,138,180,140,119,141,204,142,29,143,226,152,14,154,60,155,128,78,125,80,0,81,147,89,156,91,47,98,128,98,236,100,58,107,160,114,145,117,71,121,169,127,251,135,188,138,112,139,172,99,202,131,160,151,9,84,3,84,171,85,84,104,88,106,112,138,39,120,117,103,205,158,116,83,162,91,26,129,80,134,6,144,24,78,69,78,199,78,17,79,202,83,56,84,174,91,19,95,37,96,81,101,61,103,66,108,114,108,227,108,120,112,3,116,118,122,174,122,8,123,26,125,254,124,102,125,231,101,91,114,187,83,69,92,232,93,210,98,224,98,25,99,32,110,90,134,49,138,221,141,248,146,1,111,166,121,90,155,168,78,171,78,172,78,155,79,160,79,209,80,71,81,246,122,113,81,246,81,84,83,33,83,127,83,235,83,172,85,131,88,225,92,55,95,74,95,47,96,80,96,109,96,31,99,89,101,75,106,193,108,194,114,237,114,239,119,248,128,5,129,8,130,78,133,247,144,225,147,255,151,87,153,90,154,240,78,221,81,45,92,129,102,109,105,64,92,242,102,117,105,137,115,80,104,129,124,197,80,228,82,71,87,254,93,38,147,164,101,35,107,61,107,52,116,129,121,189,121,75,123,202,125,185,130,204,131,127,136,95,137,57,139,209,143,209,145,31,84,128,146,93,78,54,80,229,83,58,83,215,114,150,115,233,119,230,130,175,142,198,153,200,153,210,153,119,81,26,97,94,134,176,85,122,122,118,80,211,91,71,144,133,150,50,78,219,106,231,145,81,92,72,92,152,99,159,122,147,108,116,151,97,143,170,122,138,113,136,150,130,124,23,104,112,126,81,104,108,147,242,82,27,84,171,133,19,138,164,127,205,142,225,144,102,83,136,136,65,121,194,79,190,80,17,82,68,81,83,85,45,87,234,115,139,87,81,89,98,95,132,95,117,96,118,97,103,97,169,97,178,99,58,100,108,101,111,102,66,104,19,110,102,117,61,122,251,124,76,125,153,125,75,126,107,127,14,131,74,131,205,134,8,138,99,138,102,139,253,142,26,152,143,157,184,130,206,143,232,155,135,82,31,98,131,100,192,111,153,150,65,104,145,80,32,107,122,108,84,111,116,122,80,125,64,136,35,138,8,103,246,78,57,80,38,80,101,80,124,81,56,82,99,82,167,85,15,87,5,88,204,90,250,94,178,97,248,97,243,98,114,99,28,105,41,106,125,114,172,114,46,115,20,120,111,120,121,125,12,119,169,128,139,137,25,139,226,140,210,142,99,144,117,147,122,150,85,152,19,154,120,158,67,81,159,83,179,83,123,94,38,95,27,110,144,110,132,115,254,115,67,125,55,130,0,138,250,138,80,150,78,78,11,80,228,83,124,84,250,86,209,89,100,91,241,93,171,94,39,95,56,98,69,101,175,103,86,110,208,114,202,124,180,136,161,128,225,128,240,131,78,134,135,138,232,141,55,146,199,150,103,152,19,159,148,78,146,78,13,79,72,83,73,84,62,84,47,90,140,95,161,95,159,96,167,104,142,106,90,116,129,120,158,138,164,138,119,139,144,145,94,78,201,155,164,78,124,79,175,79,25,80,22,80,73,81,108,81,159,82,185,82,254,82,154,83,227,83,17,84,14,84,137,85,81,87,162,87,125,89,84,91,93,91,143,91,229,93,231,93,247,93,120,94,131,94,154,94,183,94,24,95,82,96,76,97,151,98,216,98,167,99,59,101,2,102,67,102,244,102,109,103,33,104,151,104,203,105,95,108,42,109,105,109,47,110,157,110,50,117,135,118,108,120,63,122,224,124,5,125,24,125,94,125,177,125,21,128,3,128,175,128,177,128,84,129,143,129,42,130,82,131,76,136,97,136,27,139,162,140,252,140,202,144,117,145,113,146,63,120,252,146,164,149,77,150,5,152,153,153,216,154,59,157,91,82,171,82,247,83,8,84,213,88,247,98,224,111,106,140,95,143,185,158,75,81,59,82,74,84,253,86,64,122,119,145,96,157,210,158,68,115,9,111,112,129,17,117,253,95,218,96,168,154,219,114,188,143,100,107,3,152,202,78,240,86,100,87,190,88,90,90,104,96,199,97,15,102,6,102,57,104,177,104,247,109,213,117,58,125,110,130,66,155,155,78,80,79,201,83,6,85,111,93,230,93,238,93,251,103,153,108,115,116,2,120,80,138,150,147,223,136,80,87,167,94,43,99,181,80,172,80,141,81,0,103,201,84,94,88,187,89,176,91,105,95,77,98,161,99,61,104,115,107,8,110,125,112,199,145,128,114,21,120,38,120,109,121,142,101,48,125,220,131,193,136,9,143,155,150,100,82,40,87,80,103,106,127,161,140,180,81,66,87,42,150,58,88,138,105,180,128,178,84,14,93,252,87,149,120,250,157,92,79,74,82,139,84,62,100,40,102,20,103,245,103,132,122,86,123,34,125,47,147,92,104,173,155,57,123,25,83,138,81,55,82,223,91,246,98,174,100,230,100,45,103,186,107,169,133,209,150,144,118,214,155,76,99,6,147,171,155,191,118,82,102,9,78,152,80,194,83,113,92,232,96,146,100,99,101,95,104,230,113,202,115,35,117,151,123,130,126,149,134,131,139,219,140,120,145,16,153,172,101,171,102,139,107,213,78,212,78,58,79,127,79,58,82,248,83,242,83,227,85,219,86,235,88,203,89,201,89,255,89,80,91,77,92,2,94,43,94,215,95,29,96,7,99,47,101,92,91,175,101,189,101,232,101,157,103,98,107,123,107,15,108,69,115,73,121,193,121,248,124,25,125,43,125,162,128,2,129,243,129,150,137,94,138,105,138,102,138,140,138,238,138,199,140,220,140,204,150,252,152,111,107,139,78,60,79,141,79,80,81,87,91,250,91,72,97,1,99,66,102,33,107,203,110,187,108,62,114,189,116,212,117,193,120,58,121,12,128,51,128,234,129,148,132,158,143,80,108,127,158,15,95,88,139,43,157,250,122,248,142,141,91,235,150,3,78,241,83,247,87,49,89,201,90,164,91,137,96,127,110,6,111,190,117,234,140,159,91,0,133,224,123,114,80,244,103,157,130,97,92,74,133,30,126,14,130,153,81,4,92,104,99,102,141,156,101,110,113,62,121,23,125,5,128,29,139,202,142,110,144,199,134,170,144,31,80,250,82,58,92,83,103,124,112,53,114,76,145,200,145,43,147,229,130,194,91,49,95,249,96,59,78,214,83,136,91,75,98,49,103,138,107,233,114,224,115,46,122,107,129,163,141,82,145,150,153,18,81,215,83,106,84,255,91,136,99,57,106,172,125,0,151,218,86,206,83,104,84,151,91,49,92,222,93,238,79,1,97,254,98,50,109,192,121,203,121,66,125,77,126,210,127,237,129,31,130,144,132,70,136,114,137,144,139,116,142,47,143,49,144,75,145,108,145,198,150,156,145,192,78,79,79,69,81,65,83,147,95,14,98,212,103,65,108,11,110,99,115,38,126,205,145,131,146,212,83,25,89,191,91,209,109,93,121,46,126,155,124,126,88,159,113,250,81,83,136,240,143,202,79,251,92,37,102,172,119,227,122,28,130,255,153,198,81,170,95,236,101,111,105,137,107,243,109,150,110,100,111,254,118,20,125,225,93,117,144,135,145,6,152,230,81,29,82,64,98,145,102,217,102,26,110,182,94,210,125,114,127,248,102,175,133,247,133,248,138,169,82,217,83,115,89,143,94,144,95,85,96,228,146,100,150,183,80,31,81,221,82,32,83,71,83,236,83,232,84,70,85,49,85,23,86,104,89,190,89,60,90,181,91,6,92,15,92,17,92,26,92,132,94,138,94,224,94,112,95,127,98,132,98,219,98,140,99,119,99,7,102,12,102,45,102,118,102,126,103,162,104,31,106,53,106,188,108,136,109,9,110,88,110,60,113,38,113,103,113,199,117,1,119,93,120,1,121,101,121,240,121,224,122,17,123,167,124,57,125,150,128,214,131,139,132,73,133,93,136,243,136,31,138,60,138,84,138,115,138,97,140,222,140,164,145,102,146,126,147,24,148,156,150,152,151,10,78,8,78,30,78,87,78,151,81,112,82,206,87,52,88,204,88,34,91,56,94,197,96,254,100,97,103,86,103,68,109,182,114,115,117,99,122,184,132,114,139,184,145,32,147,49,86,244,87,254,152,237,98,13,105,150,107,237,113,84,126,119,128,114,130,230,137,223,152,85,135,177,143,59,92,56,79,225,79,181,79,7,85,32,90,221,91,233,91,195,95,78,97,47,99,176,101,75,102,238,104,155,105,120,109,241,109,51,117,185,117,31,119,94,121,230,121,51,125,227,129,175,130,170,133,170,137,58,138,171,142,155,143,50,144,221,145,7,151,186,78,193,78,3,82,117,88,236,88,11,92,26,117,61,92,78,129,10,138,197,143,99,150,109,151,37,123,207,138,8,152,98,145,243,86,168,83,23,144,57,84,130,87,37,94,168,99,52,108,138,112,97,119,139,124,224,127,112,136,66,144,84,145,16,147,24,147,143,150,94,116,196,154,7,93,105,93,112,101,162,103,168,141,219,150,110,99,73,103,25,105,197,131,23,152,192,150,254,136,132,111,122,100,248,91,22,78,44,112,93,117,47,102,196,81,54,82,226,82,211,89,129,95,39,96,16,98,63,101,116,101,31,102,116,102,242,104,22,104,99,107,5,110,114,114,31,117,219,118,190,124,86,128,240,88,253,136,127,137,160,138,147,138,203,138,29,144,146,145,82,151,89,151,137,101,14,122,6,129,187,150,45,94,220,96,26,98,165,101,20,102,144,103,243,119,77,122,77,124,62,126,10,129,172,140,100,141,225,141,95,142,169,120,7,82,217,98,165,99,66,100,152,98,45,138,131,122,192,123,172,138,234,150,118,125,12,130,73,135,217,78,72,81,67,83,96,83,163,91,2,92,22,92,221,93,38,98,71,98,176,100,19,104,52,104,201,108,69,109,23,109,211,103,92,111,78,113,125,113,203,101,127,122,173,123,218,125,74,126,168,127,122,129,27,130,57,130,166,133,110,138,206,140,245,141,120,144,119,144,173,146,145,146,131,149,174,155,77,82,132,85,56,111,54,113,104,81,133,121,85,126,179,129,206,124,76,86,81,88,168,92,170,99,254,102,253,102,90,105,217,114,143,117,142,117,14,121,86,121,223,121,151,124,32,125,68,125,7,134,52,138,59,150,97,144,32,159,231,80,117,82,204,83,226,83,9,80,170,85,238,88,79,89,61,114,139,91,100,92,29,83,227,96,243,96,92,99,131,99,63,99,187,99,205,100,233,101,249,102,227,93,205,105,253,105,21,111,229,113,137,78,233,117,248,118,147,122,223,124,207,125,156,125,97,128,73,131,88,131,108,132,188,132,251,133,197,136,112,141,1,144,109,144,151,147,28,151,18,154,207,80,151,88,142,97,211,129,53,133,8,141,32,144,195,79,116,80,71,82,115,83,111,96,73,99,95,103,44,110,179,141,31,144,215,79,94,92,202,140,207,101,154,125,82,83,150,136,118,81,195,99,88,91,107,91,10,92,13,100,81,103,92,144,214,78,26,89,42,89,112,108,81,138,62,85,21,88,165,89,240,96,83,98,193,103,53,130,85,105,64,150,196,153,40,154,83,79,6,88,254,91,16,128,177,92,47,94,133,95,32,96,75,97,52,98,255,102,240,108,222,110,206,128,127,129,212,130,139,136,184,140,0,144,46,144,138,150,219,158,219,155,227,78,240,83,39,89,44,123,141,145,76,152,249,157,221,110,39,112,83,83,68,85,133,91,88,98,158,98,211,98,162,108,239,111,34,116,23,138,56,148,193,111,254,138,56,131,231,81,248,134,234,83,233,83,70,79,84,144,176,143,106,89,49,129,253,93,234,122,191,143,218,104,55,140,248,114,72,156,61,106,176,138,57,78,88,83,6,86,102,87,197,98,162,99,230,101,78,107,225,109,91,110,173,112,237,119,239,122,170,123,187,125,61,128,198,128,203,134,149,138,91,147,227,86,199,88,62,95,173,101,150,102,128,106,181,107,55,117,199,138,36,80,229,119,48,87,27,95,101,96,122,102,96,108,244,117,26,122,110,127,244,129,24,135,69,144,179,153,201,123,92,117,249,122,81,123,196,132,16,144,233,121,146,122,54,131,225,90,64,119,45,78,242,78,153,91,224,95,189,98,60,102,241,103,232,108,107,134,119,136,59,138,78,145,243,146,208,153,23,106,38,112,42,115,231,130,87,132,175,140,1,78,70,81,203,81,139,85,245,91,22,94,51,94,129,94,20,95,53,95,107,95,180,95,242,97,17,99,162,102,29,103,110,111,82,114,58,117,58,119,116,128,57,129,120,129,118,135,191,138,220,138,133,141,243,141,154,146,119,149,2,152,229,156,197,82,87,99,244,118,21,103,136,108,205,115,195,140,174,147,115,150,37,109,156,88,14,105,204,105,253,143,154,147,219,117,26,144,90,88,2,104,180,99,251,105,67,79,44,111,216,103,187,143,38,133,180,125,84,147,63,105,112,111,106,87,247,88,44,91,44,125,42,114,10,84,227,145,180,157,173,78,78,79,92,80,117,80,67,82,158,140,72,84,36,88,154,91,29,94,149,94,173,94,247,94,31,95,140,96,181,98,58,99,208,99,175,104,64,108,135,120,142,121,11,122,224,125,71,130,2,138,230,138,68,142,19,144,184,144,45,145,216,145,14,159,229,108,88,100,226,100,117,101,244,110,132,118,27,123,105,144,209,147,186,110,242,84,185,95,164,100,77,143,237,143,68,146,120,81,107,88,41,89,85,92,151,94,251,109,143,126,28,117,188,140,226,142,91,152,185,112,29,79,191,107,177,111,48,117,251,150,78,81,16,84,53,88,87,88,172,89,96,92,146,95,151,101,92,103,33,110,123,118,223,131,237,140,20,144,253,144,77,147,37,120,58,120,170,82,166,94,31,87,116,89,18,96,18,80,90,81,172,81,205,81,0,82,16,85,84,88,88,88,87,89,149,91,246,92,139,93,188,96,149,98,45,100,113,103,67,104,188,104,223,104,215,118,216,109,111,110,155,109,111,112,200,113,83,95,216,117,119,121,73,123,84,123,82,123,214,124,113,125,48,82,99,132,105,133,228,133,14,138,4,139,70,140,15,142,3,144,15,144,25,148,118,150,45,152,48,154,216,149,205,80,213,82,12,84,2,88,14,92,167,97,158,100,30,109,179,119,229,122,244,128,4,132,83,144,133,146,224,92,7,157,63,83,151,95,179,95,156,109,121,114,99,119,191,121,228,123,210,107,236,114,173,138,3,104,97,106,248,81,129,122,52,105,74,92,246,156,235,130,197,91,73,145,30,112,120,86,111,92,199,96,102,101,140,108,90,140,65,144,19,152,81,84,199,102,13,146,72,89,163,144,133,81,77,78,234,81,153,133,14,139,88,112,122,99,75,147,98,105,180,153,4,126,119,117,87,83,96,105,223,142,227,150,93,108,140,78,60,92,16,95,233,143,2,83,209,140,137,128,121,134,255,94,229,101,115,78,101,81,130,89,63,92,238,151,251,78,138,89,205,95,141,138,225,111,176,121,98,121,231,91,113,132,43,115,177,113,116,94,245,95,123,99,154,100,195,113,152,124,67,78,252,94,75,78,220,87,162,86,169,96,195,111,13,125,253,128,51,129,191,129,178,143,151,137,164,134,244,93,138,98,173,100,135,137,119,103,226,108,62,109,54,116,52,120,70,90,117,127,173,130,172,153,243,79,195,94,221,98,146,99,87,101,111,103,195,118,76,114,204,128,186,128,41,143,77,145,13,80,249,87,146,90,133,104,115,105,100,113,253,114,183,140,242,88,224,140,106,150,25,144,127,135,228,121,231,119,41,132,47,79,101,82,90,83,205,98,207,103,202,108,125,118,148,123,149,124,54,130,132,133,235,143,221,102,32,111,6,114,27,126,171,131,193,153,166,158,253,81,177,123,114,120,184,123,135,128,72,123,232,106,97,94,140,128,81,117,96,117,107,81,98,146,140,110,122,118,151,145,234,154,16,79,112,127,156,98,79,123,165,149,233,156,122,86,89,88,228,134,188,150,52,79,36,82,74,83,205,83,219,83,6,94,44,100,145,101,127,103,62,108,78,108,72,114,175,114,237,115,84,117,65,126,44,130,233,133,169,140,196,123,198,145,105,113,18,152,239,152,61,99,105,102,106,117,228,118,208,120,67,133,238,134,42,83,81,83,38,84,131,89,135,94,124,95,178,96,73,98,121,98,171,98,144,101,212,107,204,108,178,117,174,118,145,120,216,121,203,125,119,127,165,128,171,136,185,138,187,140,127,144,94,151,219,152,11,106,56,124,153,80,62,92,174,95,135,103,216,107,53,116,9,119,142,127,59,159,202,103,23,122,57,83,139,117,237,154,102,95,157,129,241,131,152,128,60,95,197,95,98,117,70,123,60,144,103,104,235,89,155,90,16,125,126,118,44,139,245,79,106,95,25,106,55,108,2,111,226,116,104,121,104,136,85,138,121,140,223,94,207,99,197,117,210,121,215,130,40,147,242,146,156,132,237,134,45,156,193,84,108,95,140,101,92,109,21,112,167,140,211,140,59,152,79,101,246,116,13,78,216,78,224,87,43,89,102,90,204,91,168,81,3,94,156,94,22,96,118,98,119,101,167,101,110,102,110,109,54,114,38,123,80,129,154,129,153,130,92,139,160,140,230,140,116,141,28,150,68,150,174,79,171,100,102,107,30,130,97,132,106,133,232,144,1,92,83,105,168,152,122,132,87,133,15,79,111,82,169,95,69,94,13,103,143,121,121,129,7,137,134,137,245,109,23,95,85,98,184,108,207,78,105,114,146,155,6,82,59,84,116,86,179,88,164,97,110,98,26,113,110,89,137,124,222,124,27,125,240,150,135,101,94,128,25,78,117,79,117,81,64,88,99,94,115,94,10,95,196,103,38,78,61,133,137,149,91,150,115,124,1,152,251,80,193,88,86,118,167,120,37,82,165,119,17,133,134,123,79,80,9,89,71,114,199,123,232,125,186,143,212,143,77,144,191,79,201,82,41,90,1,95,173,151,221,79,23,130,234,146,3,87,85,99,105,107,43,117,220,136,20,143,66,122,223,82,147,88,85,97,10,98,174,102,205,107,63,124,233,131,35,80,248,79,5,83,70,84,49,88,73,89,157,91,240,92,239,92,41,93,150,94,177,98,103,99,62,101,185,101,11,103,213,108,225,108,249,112,50,120,43,126,222,128,179,130,12,132,236,132,2,135,18,137,42,138,74,140,166,144,210,146,253,152,243,156,108,157,79,78,161,78,141,80,86,82,74,87,168,89,61,94,216,95,217,95,63,98,180,102,27,103,208,103,210,104,146,81,33,125,170,128,168,129,0,139,140,140,191,140,126,146,50,150,32,84,44,152,23,83,213,80,92,83,168,88,178,100,52,103,103,114,102,119,70,122,230,145,195,82,161,108,134,107,0,88,76,94,84,89,44,103,251,127,225,81,198,118,105,100,232,120,84,155,187,158,203,87,185,89,39,102,154,103,206,107,233,84,217,105,85,94,156,129,149,103,170,155,254,103,82,156,93,104,166,78,227,79,200,83,185,98,43,103,171,108,196,143,173,79,109,126,191,158,7,78,98,97,128,110,43,111,19,133,115,84,42,103,69,155,243,93,149,123,172,92,198,91,28,135,74,110,209,132,20,122,8,129,153,89,141,124,17,108,32,119,217,82,34,89,33,113,95,114,219,119,39,151,97,157,11,105,127,90,24,90,165,81,13,84,125,84,14,102,223,118,247,143,152,146,244,156,234,89,93,114,197,110,77,81,201,104,191,125,236,125,98,151,186,158,120,100,33,106,2,131,132,89,95,91,219,107,27,115,242,118,178,125,23,128,153,132,50,81,40,103,217,158,238,118,98,103,255,82,5,153,36,92,59,98,126,124,176,140,79,85,182,96,11,125,128,149,1,83,95,78,182,81,28,89,58,114,54,128,206,145,37,95,226,119,132,83,121,95,4,125,172,133,51,138,141,142,86,151,243,103,174,133,83,148,9,97,8,97,185,108,82,118,237,138,56,143,47,85,81,79,42,81,199,82,203,83,165,91,125,94,160,96,130,97,214,99,9,103,218,103,103,110,140,109,54,115,55,115,49,117,80,121,213,136,152,138,74,144,145,144,245,144,196,150,141,135,21,89,136,78,89,79,14,78,137,138,63,143,16,152,173,80,124,94,150,89,185,91,184,94,218,99,250,99,193,100,220,102,74,105,216,105,11,109,182,110,148,113,40,117,175,122,138,127,0,128,73,132,201,132,129,137,33,139,10,142,101,144,125,150,10,153,126,97,145,98,50,107,131,108,116,109,204,127,252,127,192,109,133,127,186,135,248,136,101,103,177,131,60,152,247,150,27,109,97,125,61,132,106,145,113,78,117,83,80,93,4,107,235,111,205,133,45,134,167,137,41,82,15,84,101,92,78,103,168,104,6,116,131,116,226,117,207,136,225,136,204,145,226,150,120,150,139,95,135,115,203,122,78,132,160,99,101,117,137,82,65,109,156,110,9,116,89,117,107,120,146,124,134,150,220,122,141,159,182,79,110,97,197,101,92,134,134,78,174,78,218,80,33,78,204,81,238,91,153,101,129,104,188,109,31,115,66,118,173,119,28,122,231,124,111,130,210,138,124,144,207,145,117,150,24,152,155,82,209,125,43,80,152,83,151,103,203,109,208,113,51,116,232,129,42,143,163,150,87,156,159,158,96,116,65,88,153,109,47,125,94,152,228,78,54,79,139,79,183,81,177,82,186,93,28,96,178,115,60,121,211,130,52,146,183,150,246,150,10,151,151,158,98,159,166,102,116,107,23,82,163,82,200,112,194,136,201,94,75,96,144,97,35,111,73,113,62,124,244,125,111,128,238,132,35,144,44,147,66,84,111,155,211,106,137,112,194,140,239,141,50,151,180,82,65,90,202,94,4,95,23,103,124,105,148,105,106,109,15,111,98,114,252,114,237,123,1,128,126,128,75,135,206,144,109,81,147,158,132,121,139,128,50,147,214,138,45,80,140,84,113,138,106,107,196,140,7,129,209,96,160,103,242,157,153,78,152,78,16,156,107,138,193,133,104,133,0,105,126,110,151,120,85,129,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,95,16,78,21,78,42,78,49,78,54,78,60,78,63,78,66,78,86,78,88,78,130,78,133,78,107,140,138,78,18,130,13,95,142,78,158,78,159,78,160,78,162,78,176,78,179,78,182,78,206,78,205,78,196,78,198,78,194,78,215,78,222,78,237,78,223,78,247,78,9,79,90,79,48,79,91,79,93,79,87,79,71,79,118,79,136,79,143,79,152,79,123,79,105,79,112,79,145,79,111,79,134,79,150,79,24,81,212,79,223,79,206,79,216,79,219,79,209,79,218,79,208,79,228,79,229,79,26,80,40,80,20,80,42,80,37,80,5,80,28,79,246,79,33,80,41,80,44,80,254,79,239,79,17,80,6,80,67,80,71,80,3,103,85,80,80,80,72,80,90,80,86,80,108,80,120,80,128,80,154,80,133,80,180,80,178,80,201,80,202,80,179,80,194,80,214,80,222,80,229,80,237,80,227,80,238,80,249,80,245,80,9,81,1,81,2,81,22,81,21,81,20,81,26,81,33,81,58,81,55,81,60,81,59,81,63,81,64,81,82,81,76,81,84,81,98,81,248,122,105,81,106,81,110,81,128,81,130,81,216,86,140,81,137,81,143,81,145,81,147,81,149,81,150,81,164,81,166,81,162,81,169,81,170,81,171,81,179,81,177,81,178,81,176,81,181,81,189,81,197,81,201,81,219,81,224,81,85,134,233,81,237,81,240,81,245,81,254,81,4,82,11,82,20,82,14,82,39,82,42,82,46,82,51,82,57,82,79,82,68,82,75,82,76,82,94,82,84,82,106,82,116,82,105,82,115,82,127,82,125,82,141,82,148,82,146,82,113,82,136,82,145,82,168,143,167,143,172,82,173,82,188,82,181,82,193,82,205,82,215,82,222,82,227,82,230,82,237,152,224,82,243,82,245,82,248,82,249,82,6,83,8,83,56,117,13,83,16,83,15,83,21,83,26,83,35,83,47,83,49,83,51,83,56,83,64,83,70,83,69,83,23,78,73,83,77,83,214,81,94,83,105,83,110,83,24,89,123,83,119,83,130,83,150,83,160,83,166,83,165,83,174,83,176,83,182,83,195,83,18,124,217,150,223,83,252,102,238,113,238,83,232,83,237,83,250,83,1,84,61,84,64,84,44,84,45,84,60,84,46,84,54,84,41,84,29,84,78,84,143,84,117,84,142,84,95,84,113,84,119,84,112,84,146,84,123,84,128,84,118,84,132,84,144,84,134,84,199,84,162,84,184,84,165,84,172,84,196,84,200,84,168,84,171,84,194,84,164,84,190,84,188,84,216,84,229,84,230,84,15,85,20,85,253,84,238,84,237,84,250,84,226,84,57,85,64,85,99,85,76,85,46,85,92,85,69,85,86,85,87,85,56,85,51,85,93,85,153,85,128,85,175,84,138,85,159,85,123,85,126,85,152,85,158,85,174,85,124,85,131,85,169,85,135,85,168,85,218,85,197,85,223,85,196,85,220,85,228,85,212,85,20,86,247,85,22,86,254,85,253,85,27,86,249,85,78,86,80,86,223,113,52,86,54,86,50,86,56,86,107,86,100,86,47,86,108,86,106,86,134,86,128,86,138,86,160,86,148,86,143,86,165,86,174,86,182,86,180,86,194,86,188,86,193,86,195,86,192,86,200,86,206,86,209,86,211,86,215,86,238,86,249,86,0,87,255,86,4,87,9,87,8,87,11,87,13,87,19,87,24,87,22,87,199,85,28,87,38,87,55,87,56,87,78,87,59,87,64,87,79,87,105,87,192,87,136,87,97,87,127,87,137,87,147,87,160,87,179,87,164,87,170,87,176,87,195,87,198,87,212,87,210,87,211,87,10,88,214,87,227,87,11,88,25,88,29,88,114,88,33,88,98,88,75,88,112,88,192,107,82,88,61,88,121,88,133,88,185,88,159,88,171,88,186,88,222,88,187,88,184,88,174,88,197,88,211,88,209,88,215,88,217,88,216,88,229,88,220,88,228,88,223,88,239,88,250,88,249,88,251,88,252,88,253,88,2,89,10,89,16,89,27,89,166,104,37,89,44,89,45,89,50,89,56,89,62,89,210,122,85,89,80,89,78,89,90,89,88,89,98,89,96,89,103,89,108,89,105,89,120,89,129,89,157,89,94,79,171,79,163,89,178,89,198,89,232,89,220,89,141,89,217,89,218,89,37,90,31,90,17,90,28,90,9,90,26,90,64,90,108,90,73,90,53,90,54,90,98,90,106,90,154,90,188,90,190,90,203,90,194,90,189,90,227,90,215,90,230,90,233,90,214,90,250,90,251,90,12,91,11,91,22,91,50,91,208,90,42,91,54,91,62,91,67,91,69,91,64,91,81,91,85,91,90,91,91,91,101,91,105,91,112,91,115,91,117,91,120,91,136,101,122,91,128,91,131,91,166,91,184,91,195,91,199,91,201,91,212,91,208,91,228,91,230,91,226,91,222,91,229,91,235,91,240,91,246,91,243,91,5,92,7,92,8,92,13,92,19,92,32,92,34,92,40,92,56,92,57,92,65,92,70,92,78,92,83,92,80,92,79,92,113,91,108,92,110,92,98,78,118,92,121,92,140,92,145,92,148,92,155,89,171,92,187,92,182,92,188,92,183,92,197,92,190,92,199,92,217,92,233,92,253,92,250,92,237,92,140,93,234,92,11,93,21,93,23,93,92,93,31,93,27,93,17,93,20,93,34,93,26,93,25,93,24,93,76,93,82,93,78,93,75,93,108,93,115,93,118,93,135,93,132,93,130,93,162,93,157,93,172,93,174,93,189,93,144,93,183,93,188,93,201,93,205,93,211,93,210,93,214,93,219,93,235,93,242,93,245,93,11,94,26,94,25,94,17,94,27,94,54,94,55,94,68,94,67,94,64,94,78,94,87,94,84,94,95,94,98,94,100,94,71,94,117,94,118,94,122,94,188,158,127,94,160,94,193,94,194,94,200,94,208,94,207,94,214,94,227,94,221,94,218,94,219,94,226,94,225,94,232,94,233,94,236,94,241,94,243,94,240,94,244,94,248,94,254,94,3,95,9,95,93,95,92,95,11,95,17,95,22,95,41,95,45,95,56,95,65,95,72,95,76,95,78,95,47,95,81,95,86,95,87,95,89,95,97,95,109,95,115,95,119,95,131,95,130,95,127,95,138,95,136,95,145,95,135,95,158,95,153,95,152,95,160,95,168,95,173,95,188,95,214,95,251,95,228,95,248,95,241,95,221,95,179,96,255,95,33,96,96,96,25,96,16,96,41,96,14,96,49,96,27,96,21,96,43,96,38,96,15,96,58,96,90,96,65,96,106,96,119,96,95,96,74,96,70,96,77,96,99,96,67,96,100,96,66,96,108,96,107,96,89,96,129,96,141,96,231,96,131,96,154,96,132,96,155,96,150,96,151,96,146,96,167,96,139,96,225,96,184,96,224,96,211,96,180,96,240,95,189,96,198,96,181,96,216,96,77,97,21,97,6,97,246,96,247,96,0,97,244,96,250,96,3,97,33,97,251,96,241,96,13,97,14,97,71,97,62,97,40,97,39,97,74,97,63,97,60,97,44,97,52,97,61,97,66,97,68,97,115,97,119,97,88,97,89,97,90,97,107,97,116,97,111,97,101,97,113,97,95,97,93,97,83,97,117,97,153,97,150,97,135,97,172,97,148,97,154,97,138,97,145,97,171,97,174,97,204,97,202,97,201,97,247,97,200,97,195,97,198,97,186,97,203,97,121,127,205,97,230,97,227,97,246,97,250,97,244,97,255,97,253,97,252,97,254,97,0,98,8,98,9,98,13,98,12,98,20,98,27,98,30,98,33,98,42,98,46,98,48,98,50,98,51,98,65,98,78,98,94,98,99,98,91,98,96,98,104,98,124,98,130,98,137,98,126,98,146,98,147,98,150,98,212,98,131,98,148,98,215,98,209,98,187,98,207,98,255,98,198,98,212,100,200,98,220,98,204,98,202,98,194,98,199,98,155,98,201,98,12,99,238,98,241,98,39,99,2,99,8,99,239,98,245,98,80,99,62,99,77,99,28,100,79,99,150,99,142,99,128,99,171,99,118,99,163,99,143,99,137,99,159,99,181,99,107,99,105,99,190,99,233,99,192,99,198,99,227,99,201,99,210,99,246,99,196,99,22,100,52,100,6,100,19,100,38,100,54,100,29,101,23,100,40,100,15,100,103,100,111,100,118,100,78,100,42,101,149,100,147,100,165,100,169,100,136,100,188,100,218,100,210,100,197,100,199,100,187,100,216,100,194,100,241,100,231,100,9,130,224,100,225,100,172,98,227,100,239,100,44,101,246,100,244,100,242,100,250,100,0,101,253,100,24,101,28,101,5,101,36,101,35,101,43,101,52,101,53,101,55,101,54,101,56,101,75,117,72,101,86,101,85,101,77,101,88,101,94,101,93,101,114,101,120,101,130,101,131,101,138,139,155,101,159,101,171,101,183,101,195,101,198,101,193,101,196,101,204,101,210,101,219,101,217,101,224,101,225,101,241,101,114,103,10,102,3,102,251,101,115,103,53,102,54,102,52,102,28,102,79,102,68,102,73,102,65,102,94,102,93,102,100,102,103,102,104,102,95,102,98,102,112,102,131,102,136,102,142,102,137,102,132,102,152,102,157,102,193,102,185,102,201,102,190,102,188,102,196,102,184,102,214,102,218,102,224,102,63,102,230,102,233,102,240,102,245,102,247,102,15,103,22,103,30,103,38,103,39,103,56,151,46,103,63,103,54,103,65,103,56,103,55,103,70,103,94,103,96,103,89,103,99,103,100,103,137,103,112,103,169,103,124,103,106,103,140,103,139,103,166,103,161,103,133,103,183,103,239,103,180,103,236,103,179,103,233,103,184,103,228,103,222,103,221,103,226,103,238,103,185,103,206,103,198,103,231,103,156,106,30,104,70,104,41,104,64,104,77,104,50,104,78,104,179,104,43,104,89,104,99,104,119,104,127,104,159,104,143,104,173,104,148,104,157,104,155,104,131,104,174,106,185,104,116,104,181,104,160,104,186,104,15,105,141,104,126,104,1,105,202,104,8,105,216,104,34,105,38,105,225,104,12,105,205,104,212,104,231,104,213,104,54,105,18,105,4,105,215,104,227,104,37,105,249,104,224,104,239,104,40,105,42,105,26,105,35,105,33,105,198,104,121,105,119,105,92,105,120,105,107,105,84,105,126,105,110,105,57,105,116,105,61,105,89,105,48,105,97,105,94,105,93,105,129,105,106,105,178,105,174,105,208,105,191,105,193,105,211,105,190,105,206,105,232,91,202,105,221,105,187,105,195,105,167,105,46,106,145,105,160,105,156,105,149,105,180,105,222,105,232,105,2,106,27,106,255,105,10,107,249,105,242,105,231,105,5,106,177,105,30,106,237,105,20,106,235,105,10,106,18,106,193,106,35,106,19,106,68,106,12,106,114,106,54,106,120,106,71,106,98,106,89,106,102,106,72,106,56,106,34,106,144,106,141,106,160,106,132,106,162,106,163,106,151,106,23,134,187,106,195,106,194,106,184,106,179,106,172,106,222,106,209,106,223,106,170,106,218,106,234,106,251,106,5,107,22,134,250,106,18,107,22,107,49,155,31,107,56,107,55,107,220,118,57,107,238,152,71,107,67,107,73,107,80,107,89,107,84,107,91,107,95,107,97,107,120,107,121,107,127,107,128,107,132,107,131,107,141,107,152,107,149,107,158,107,164,107,170,107,171,107,175,107,178,107,177,107,179,107,183,107,188,107,198,107,203,107,211,107,223,107,236,107,235,107,243,107,239,107,190,158,8,108,19,108,20,108,27,108,36,108,35,108,94,108,85,108,98,108,106,108,130,108,141,108,154,108,129,108,155,108,126,108,104,108,115,108,146,108,144,108,196,108,241,108,211,108,189,108,215,108,197,108,221,108,174,108,177,108,190,108,186,108,219,108,239,108,217,108,234,108,31,109,77,136,54,109,43,109,61,109,56,109,25,109,53,109,51,109,18,109,12,109,99,109,147,109,100,109,90,109,121,109,89,109,142,109,149,109,228,111,133,109,249,109,21,110,10,110,181,109,199,109,230,109,184,109,198,109,236,109,222,109,204,109,232,109,210,109,197,109,250,109,217,109,228,109,213,109,234,109,238,109,45,110,110,110,46,110,25,110,114,110,95,110,62,110,35,110,107,110,43,110,118,110,77,110,31,110,67,110,58,110,78,110,36,110,255,110,29,110,56,110,130,110,170,110,152,110,201,110,183,110,211,110,189,110,175,110,196,110,178,110,212,110,213,110,143,110,165,110,194,110,159,110,65,111,17,111,76,112,236,110,248,110,254,110,63,111,242,110,49,111,239,110,50,111,204,110,62,111,19,111,247,110,134,111,122,111,120,111,129,111,128,111,111,111,91,111,243,111,109,111,130,111,124,111,88,111,142,111,145,111,194,111,102,111,179,111,163,111,161,111,164,111,185,111,198,111,170,111,223,111,213,111,236,111,212,111,216,111,241,111,238,111,219,111,9,112,11,112,250,111,17,112,1,112,15,112,254,111,27,112,26,112,116,111,29,112,24,112,31,112,48,112,62,112,50,112,81,112,99,112,153,112,146,112,175,112,241,112,172,112,184,112,179,112,174,112,223,112,203,112,221,112,217,112,9,113,253,112,28,113,25,113,101,113,85,113,136,113,102,113,98,113,76,113,86,113,108,113,143,113,251,113,132,113,149,113,168,113,172,113,215,113,185,113,190,113,210,113,201,113,212,113,206,113,224,113,236,113,231,113,245,113,252,113,249,113,255,113,13,114,16,114,27,114,40,114,45,114,44,114,48,114,50,114,59,114,60,114,63,114,64,114,70,114,75,114,88,114,116,114,126,114,130,114,129,114,135,114,146,114,150,114,162,114,167,114,185,114,178,114,195,114,198,114,196,114,206,114,210,114,226,114,224,114,225,114,249,114,247,114,15,80,23,115,10,115,28,115,22,115,29,115,52,115,47,115,41,115,37,115,62,115,78,115,79,115,216,158,87,115,106,115,104,115,112,115,120,115,117,115,123,115,122,115,200,115,179,115,206,115,187,115,192,115,229,115,238,115,222,115,162,116,5,116,111,116,37,116,248,115,50,116,58,116,85,116,63,116,95,116,89,116,65,116,92,116,105,116,112,116,99,116,106,116,118,116,126,116,139,116,158,116,167,116,202,116,207,116,212,116,241,115,224,116,227,116,231,116,233,116,238,116,242,116,240,116,241,116,248,116,247,116,4,117,3,117,5,117,12,117,14,117,13,117,21,117,19,117,30,117,38,117,44,117,60,117,68,117,77,117,74,117,73,117,91,117,70,117,90,117,105,117,100,117,103,117,107,117,109,117,120,117,118,117,134,117,135,117,116,117,138,117,137,117,130,117,148,117,154,117,157,117,165,117,163,117,194,117,179,117,195,117,181,117,189,117,184,117,188,117,177,117,205,117,202,117,210,117,217,117,227,117,222,117,254,117,255,117,252,117,1,118,240,117,250,117,242,117,243,117,11,118,13,118,9,118,31,118,39,118,32,118,33,118,34,118,36,118,52,118,48,118,59,118,71,118,72,118,70,118,92,118,88,118,97,118,98,118,104,118,105,118,106,118,103,118,108,118,112,118,114,118,118,118,120,118,124,118,128,118,131,118,136,118,139,118,142,118,150,118,147,118,153,118,154,118,176,118,180,118,184,118,185,118,186,118,194,118,205,118,214,118,210,118,222,118,225,118,229,118,231,118,234,118,47,134,251,118,8,119,7,119,4,119,41,119,36,119,30,119,37,119,38,119,27,119,55,119,56,119,71,119,90,119,104,119,107,119,91,119,101,119,127,119,126,119,121,119,142,119,139,119,145,119,160,119,158,119,176,119,182,119,185,119,191,119,188,119,189,119,187,119,199,119,205,119,215,119,218,119,220,119,227,119,238,119,252,119,12,120,18,120,38,121,32,120,42,121,69,120,142,120,116,120,134,120,124,120,154,120,140,120,163,120,181,120,170,120,175,120,209,120,198,120,203,120,212,120,190,120,188,120,197,120,202,120,236,120,231,120,218,120,253,120,244,120,7,121,18,121,17,121,25,121,44,121,43,121,64,121,96,121,87,121,95,121,90,121,85,121,83,121,122,121,127,121,138,121,157,121,167,121,75,159,170,121,174,121,179,121,185,121,186,121,201,121,213,121,231,121,236,121,225,121,227,121,8,122,13,122,24,122,25,122,32,122,31,122,128,121,49,122,59,122,62,122,55,122,67,122,87,122,73,122,97,122,98,122,105,122,157,159,112,122,121,122,125,122,136,122,151,122,149,122,152,122,150,122,169,122,200,122,176,122,182,122,197,122,196,122,191,122,131,144,199,122,202,122,205,122,207,122,213,122,211,122,217,122,218,122,221,122,225,122,226,122,230,122,237,122,240,122,2,123,15,123,10,123,6,123,51,123,24,123,25,123,30,123,53,123,40,123,54,123,80,123,122,123,4,123,77,123,11,123,76,123,69,123,117,123,101,123,116,123,103,123,112,123,113,123,108,123,110,123,157,123,152,123,159,123,141,123,156,123,154,123,139,123,146,123,143,123,93,123,153,123,203,123,193,123,204,123,207,123,180,123,198,123,221,123,233,123,17,124,20,124,230,123,229,123,96,124,0,124,7,124,19,124,243,123,247,123,23,124,13,124,246,123,35,124,39,124,42,124,31,124,55,124,43,124,61,124,76,124,67,124,84,124,79,124,64,124,80,124,88,124,95,124,100,124,86,124,101,124,108,124,117,124,131,124,144,124,164,124,173,124,162,124,171,124,161,124,168,124,179,124,178,124,177,124,174,124,185,124,189,124,192,124,197,124,194,124,216,124,210,124,220,124,226,124,59,155,239,124,242,124,244,124,246,124,250,124,6,125,2,125,28,125,21,125,10,125,69,125,75,125,46,125,50,125,63,125,53,125,70,125,115,125,86,125,78,125,114,125,104,125,110,125,79,125,99,125,147,125,137,125,91,125,143,125,125,125,155,125,186,125,174,125,163,125,181,125,199,125,189,125,171,125,61,126,162,125,175,125,220,125,184,125,159,125,176,125,216,125,221,125,228,125,222,125,251,125,242,125,225,125,5,126,10,126,35,126,33,126,18,126,49,126,31,126,9,126,11,126,34,126,70,126,102,126,59,126,53,126,57,126,67,126,55,126,50,126,58,126,103,126,93,126,86,126,94,126,89,126,90,126,121,126,106,126,105,126,124,126,123,126,131,126,213,125,125,126,174,143,127,126,136,126,137,126,140,126,146,126,144,126,147,126,148,126,150,126,142,126,155,126,156,126,56,127,58,127,69,127,76,127,77,127,78,127,80,127,81,127,85,127,84,127,88,127,95,127,96,127,104,127,105,127,103,127,120,127,130,127,134,127,131,127,136,127,135,127,140,127,148,127,158,127,157,127,154,127,163,127,175,127,178,127,185,127,174,127,182,127,184,127,113,139,197,127,198,127,202,127,213,127,212,127,225,127,230,127,233,127,243,127,249,127,220,152], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+25848);
/* memory initializer */ allocate([6,128,4,128,11,128,18,128,24,128,25,128,28,128,33,128,40,128,63,128,59,128,74,128,70,128,82,128,88,128,90,128,95,128,98,128,104,128,115,128,114,128,112,128,118,128,121,128,125,128,127,128,132,128,134,128,133,128,155,128,147,128,154,128,173,128,144,81,172,128,219,128,229,128,217,128,221,128,196,128,218,128,214,128,9,129,239,128,241,128,27,129,41,129,35,129,47,129,75,129,139,150,70,129,62,129,83,129,81,129,252,128,113,129,110,129,101,129,102,129,116,129,131,129,136,129,138,129,128,129,130,129,160,129,149,129,164,129,163,129,95,129,147,129,169,129,176,129,181,129,190,129,184,129,189,129,192,129,194,129,186,129,201,129,205,129,209,129,217,129,216,129,200,129,218,129,223,129,224,129,231,129,250,129,251,129,254,129,1,130,2,130,5,130,7,130,10,130,13,130,16,130,22,130,41,130,43,130,56,130,51,130,64,130,89,130,88,130,93,130,90,130,95,130,100,130,98,130,104,130,106,130,107,130,46,130,113,130,119,130,120,130,126,130,141,130,146,130,171,130,159,130,187,130,172,130,225,130,227,130,223,130,210,130,244,130,243,130,250,130,147,131,3,131,251,130,249,130,222,130,6,131,220,130,9,131,217,130,53,131,52,131,22,131,50,131,49,131,64,131,57,131,80,131,69,131,47,131,43,131,23,131,24,131,133,131,154,131,170,131,159,131,162,131,150,131,35,131,142,131,135,131,138,131,124,131,181,131,115,131,117,131,160,131,137,131,168,131,244,131,19,132,235,131,206,131,253,131,3,132,216,131,11,132,193,131,247,131,7,132,224,131,242,131,13,132,34,132,32,132,189,131,56,132,6,133,251,131,109,132,42,132,60,132,90,133,132,132,119,132,107,132,173,132,110,132,130,132,105,132,70,132,44,132,111,132,121,132,53,132,202,132,98,132,185,132,191,132,159,132,217,132,205,132,187,132,218,132,208,132,193,132,198,132,214,132,161,132,33,133,255,132,244,132,23,133,24,133,44,133,31,133,21,133,20,133,252,132,64,133,99,133,88,133,72,133,65,133,2,134,75,133,85,133,128,133,164,133,136,133,145,133,138,133,168,133,109,133,148,133,155,133,234,133,135,133,156,133,119,133,126,133,144,133,201,133,186,133,207,133,185,133,208,133,213,133,221,133,229,133,220,133,249,133,10,134,19,134,11,134,254,133,250,133,6,134,34,134,26,134,48,134,63,134,77,134,85,78,84,134,95,134,103,134,113,134,147,134,163,134,169,134,170,134,139,134,140,134,182,134,175,134,196,134,198,134,176,134,201,134,35,136,171,134,212,134,222,134,233,134,236,134,223,134,219,134,239,134,18,135,6,135,8,135,0,135,3,135,251,134,17,135,9,135,13,135,249,134,10,135,52,135,63,135,55,135,59,135,37,135,41,135,26,135,96,135,95,135,120,135,76,135,78,135,116,135,87,135,104,135,110,135,89,135,83,135,99,135,106,135,5,136,162,135,159,135,130,135,175,135,203,135,189,135,192,135,208,135,214,150,171,135,196,135,179,135,199,135,198,135,187,135,239,135,242,135,224,135,15,136,13,136,254,135,246,135,247,135,14,136,210,135,17,136,22,136,21,136,34,136,33,136,49,136,54,136,57,136,39,136,59,136,68,136,66,136,82,136,89,136,94,136,98,136,107,136,129,136,126,136,158,136,117,136,125,136,181,136,114,136,130,136,151,136,146,136,174,136,153,136,162,136,141,136,164,136,176,136,191,136,177,136,195,136,196,136,212,136,216,136,217,136,221,136,249,136,2,137,252,136,244,136,232,136,242,136,4,137,12,137,10,137,19,137,67,137,30,137,37,137,42,137,43,137,65,137,68,137,59,137,54,137,56,137,76,137,29,137,96,137,94,137,102,137,100,137,109,137,106,137,111,137,116,137,119,137,126,137,131,137,136,137,138,137,147,137,152,137,161,137,169,137,166,137,172,137,175,137,178,137,186,137,189,137,191,137,192,137,218,137,220,137,221,137,231,137,244,137,248,137,3,138,22,138,16,138,12,138,27,138,29,138,37,138,54,138,65,138,91,138,82,138,70,138,72,138,124,138,109,138,108,138,98,138,133,138,130,138,132,138,168,138,161,138,145,138,165,138,166,138,154,138,163,138,196,138,205,138,194,138,218,138,235,138,243,138,231,138,228,138,241,138,20,139,224,138,226,138,247,138,222,138,219,138,12,139,7,139,26,139,225,138,22,139,16,139,23,139,32,139,51,139,171,151,38,139,43,139,62,139,40,139,65,139,76,139,79,139,78,139,73,139,86,139,91,139,90,139,107,139,95,139,108,139,111,139,116,139,125,139,128,139,140,139,142,139,146,139,147,139,150,139,153,139,154,139,58,140,65,140,63,140,72,140,76,140,78,140,80,140,85,140,98,140,108,140,120,140,122,140,130,140,137,140,133,140,138,140,141,140,142,140,148,140,124,140,152,140,29,98,173,140,170,140,189,140,178,140,179,140,174,140,182,140,200,140,193,140,228,140,227,140,218,140,253,140,250,140,251,140,4,141,5,141,10,141,7,141,15,141,13,141,16,141,78,159,19,141,205,140,20,141,22,141,103,141,109,141,113,141,115,141,129,141,153,141,194,141,190,141,186,141,207,141,218,141,214,141,204,141,219,141,203,141,234,141,235,141,223,141,227,141,252,141,8,142,9,142,255,141,29,142,30,142,16,142,31,142,66,142,53,142,48,142,52,142,74,142,71,142,73,142,76,142,80,142,72,142,89,142,100,142,96,142,42,142,99,142,85,142,118,142,114,142,124,142,129,142,135,142,133,142,132,142,139,142,138,142,147,142,145,142,148,142,153,142,170,142,161,142,172,142,176,142,198,142,177,142,190,142,197,142,200,142,203,142,219,142,227,142,252,142,251,142,235,142,254,142,10,143,5,143,21,143,18,143,25,143,19,143,28,143,31,143,27,143,12,143,38,143,51,143,59,143,57,143,69,143,66,143,62,143,76,143,73,143,70,143,78,143,87,143,92,143,98,143,99,143,100,143,156,143,159,143,163,143,173,143,175,143,183,143,218,143,229,143,226,143,234,143,239,143,135,144,244,143,5,144,249,143,250,143,17,144,21,144,33,144,13,144,30,144,22,144,11,144,39,144,54,144,53,144,57,144,248,143,79,144,80,144,81,144,82,144,14,144,73,144,62,144,86,144,88,144,94,144,104,144,111,144,118,144,168,150,114,144,130,144,125,144,129,144,128,144,138,144,137,144,143,144,168,144,175,144,177,144,181,144,226,144,228,144,72,98,219,144,2,145,18,145,25,145,50,145,48,145,74,145,86,145,88,145,99,145,101,145,105,145,115,145,114,145,139,145,137,145,130,145,162,145,171,145,175,145,170,145,181,145,180,145,186,145,192,145,193,145,201,145,203,145,208,145,214,145,223,145,225,145,219,145,252,145,245,145,246,145,30,146,255,145,20,146,44,146,21,146,17,146,94,146,87,146,69,146,73,146,100,146,72,146,149,146,63,146,75,146,80,146,156,146,150,146,147,146,155,146,90,146,207,146,185,146,183,146,233,146,15,147,250,146,68,147,46,147,25,147,34,147,26,147,35,147,58,147,53,147,59,147,92,147,96,147,124,147,110,147,86,147,176,147,172,147,173,147,148,147,185,147,214,147,215,147,232,147,229,147,216,147,195,147,221,147,208,147,200,147,228,147,26,148,20,148,19,148,3,148,7,148,16,148,54,148,43,148,53,148,33,148,58,148,65,148,82,148,68,148,91,148,96,148,98,148,94,148,106,148,41,146,112,148,117,148,119,148,125,148,90,148,124,148,126,148,129,148,127,148,130,149,135,149,138,149,148,149,150,149,152,149,153,149,160,149,168,149,167,149,173,149,188,149,187,149,185,149,190,149,202,149,246,111,195,149,205,149,204,149,213,149,212,149,214,149,220,149,225,149,229,149,226,149,33,150,40,150,46,150,47,150,66,150,76,150,79,150,75,150,119,150,92,150,94,150,93,150,95,150,102,150,114,150,108,150,141,150,152,150,149,150,151,150,170,150,167,150,177,150,178,150,176,150,180,150,182,150,184,150,185,150,206,150,203,150,201,150,205,150,77,137,220,150,13,151,213,150,249,150,4,151,6,151,8,151,19,151,14,151,17,151,15,151,22,151,25,151,36,151,42,151,48,151,57,151,61,151,62,151,68,151,70,151,72,151,66,151,73,151,92,151,96,151,100,151,102,151,104,151,210,82,107,151,113,151,121,151,133,151,124,151,129,151,122,151,134,151,139,151,143,151,144,151,156,151,168,151,166,151,163,151,179,151,180,151,195,151,198,151,200,151,203,151,220,151,237,151,79,159,242,151,223,122,246,151,245,151,15,152,12,152,56,152,36,152,33,152,55,152,61,152,70,152,79,152,75,152,107,152,111,152,112,152,113,152,116,152,115,152,170,152,175,152,177,152,182,152,196,152,195,152,198,152,233,152,235,152,3,153,9,153,18,153,20,153,24,153,33,153,29,153,30,153,36,153,32,153,44,153,46,153,61,153,62,153,66,153,73,153,69,153,80,153,75,153,81,153,82,153,76,153,85,153,151,153,152,153,165,153,173,153,174,153,188,153,223,153,219,153,221,153,216,153,209,153,237,153,238,153,241,153,242,153,251,153,248,153,1,154,15,154,5,154,226,153,25,154,43,154,55,154,69,154,66,154,64,154,67,154,62,154,85,154,77,154,91,154,87,154,95,154,98,154,101,154,100,154,105,154,107,154,106,154,173,154,176,154,188,154,192,154,207,154,209,154,211,154,212,154,222,154,223,154,226,154,227,154,230,154,239,154,235,154,238,154,244,154,241,154,247,154,251,154,6,155,24,155,26,155,31,155,34,155,35,155,37,155,39,155,40,155,41,155,42,155,46,155,47,155,50,155,68,155,67,155,79,155,77,155,78,155,81,155,88,155,116,155,147,155,131,155,145,155,150,155,151,155,159,155,160,155,168,155,180,155,192,155,202,155,185,155,198,155,207,155,209,155,210,155,227,155,226,155,228,155,212,155,225,155,58,156,242,155,241,155,240,155,21,156,20,156,9,156,19,156,12,156,6,156,8,156,18,156,10,156,4,156,46,156,27,156,37,156,36,156,33,156,48,156,71,156,50,156,70,156,62,156,90,156,96,156,103,156,118,156,120,156,231,156,236,156,240,156,9,157,8,157,235,156,3,157,6,157,42,157,38,157,175,157,35,157,31,157,68,157,21,157,18,157,65,157,63,157,62,157,70,157,72,157,93,157,94,157,100,157,81,157,80,157,89,157,114,157,137,157,135,157,171,157,111,157,122,157,154,157,164,157,169,157,178,157,196,157,193,157,187,157,184,157,186,157,198,157,207,157,194,157,217,157,211,157,248,157,230,157,237,157,239,157,253,157,26,158,27,158,30,158,117,158,121,158,125,158,129,158,136,158,139,158,140,158,146,158,149,158,145,158,157,158,165,158,169,158,184,158,170,158,173,158,97,151,204,158,206,158,207,158,208,158,212,158,220,158,222,158,221,158,224,158,229,158,232,158,239,158,244,158,246,158,247,158,249,158,251,158,252,158,253,158,7,159,8,159,183,118,21,159,33,159,44,159,62,159,74,159,82,159,84,159,99,159,95,159,96,159,97,159,102,159,103,159,108,159,106,159,119,159,114,159,118,159,149,159,156,159,160,159,47,88,199,105,89,144,100,116,220,81,153,113,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,78,4,78,5,78,6,78,15,78,18,78,23,78,31,78,32,78,33,78,35,78,38,78,41,78,46,78,47,78,49,78,51,78,53,78,55,78,60,78,64,78,65,78,66,78,68,78,70,78,74,78,81,78,85,78,87,78,90,78,91,78,98,78,99,78,100,78,101,78,103,78,104,78,106,78,107,78,108,78,109,78,110,78,111,78,114,78,116,78,117,78,118,78,119,78,120,78,121,78,122,78,123,78,124,78,125,78,127,78,128,78,129,78,130,78,131,78,132,78,133,78,135,78,138,78,144,78,150,78,151,78,153,78,156,78,157,78,158,78,163,78,170,78,175,78,176,78,177,78,180,78,182,78,183,78,184,78,185,78,188,78,189,78,190,78,200,78,204,78,207,78,208,78,210,78,218,78,219,78,220,78,224,78,226,78,230,78,231,78,233,78,237,78,238,78,239,78,241,78,244,78,248,78,249,78,250,78,252,78,254,78,0,79,2,79,3,79,4,79,5,79,6,79,7,79,8,79,11,79,12,79,18,79,19,79,20,79,21,79,22,79,28,79,29,79,33,79,35,79,40,79,41,79,44,79,45,79,46,79,49,79,51,79,53,79,55,79,57,79,59,79,62,79,63,79,64,79,65,79,66,79,68,79,69,79,71,79,72,79,73,79,74,79,75,79,76,79,82,79,84,79,86,79,97,79,98,79,102,79,104,79,106,79,107,79,109,79,110,79,113,79,114,79,117,79,119,79,120,79,121,79,122,79,125,79,128,79,129,79,130,79,133,79,134,79,135,79,138,79,140,79,142,79,144,79,146,79,147,79,149,79,150,79,152,79,153,79,154,79,156,79,158,79,159,79,161,79,162,79,164,79,171,79,173,79,176,79,177,79,178,79,179,79,180,79,182,79,183,79,184,79,185,79,186,79,187,79,188,79,189,79,190,79,192,79,193,79,194,79,198,79,199,79,200,79,201,79,203,79,204,79,205,79,210,79,211,79,212,79,213,79,214,79,217,79,219,79,224,79,226,79,228,79,229,79,231,79,235,79,236,79,240,79,242,79,244,79,245,79,246,79,247,79,249,79,251,79,252,79,253,79,255,79,0,80,1,80,2,80,3,80,4,80,5,80,6,80,7,80,8,80,9,80,10,80,11,80,14,80,16,80,17,80,19,80,21,80,22,80,23,80,27,80,29,80,30,80,32,80,34,80,35,80,36,80,39,80,43,80,47,80,48,80,49,80,50,80,51,80,52,80,53,80,54,80,55,80,56,80,57,80,59,80,61,80,63,80,64,80,65,80,66,80,68,80,69,80,70,80,73,80,74,80,75,80,77,80,80,80,81,80,82,80,83,80,84,80,86,80,87,80,88,80,89,80,91,80,93,80,94,80,95,80,96,80,97,80,98,80,99,80,100,80,102,80,103,80,104,80,105,80,106,80,107,80,109,80,110,80,111,80,112,80,113,80,114,80,115,80,116,80,117,80,120,80,121,80,122,80,124,80,125,80,129,80,130,80,131,80,132,80,134,80,135,80,137,80,138,80,139,80,140,80,142,80,143,80,144,80,145,80,146,80,147,80,148,80,149,80,150,80,151,80,152,80,153,80,154,80,155,80,156,80,157,80,158,80,159,80,160,80,161,80,162,80,164,80,166,80,170,80,171,80,173,80,174,80,175,80,176,80,177,80,179,80,180,80,181,80,182,80,183,80,184,80,185,80,188,80,189,80,190,80,191,80,192,80,193,80,194,80,195,80,196,80,197,80,198,80,199,80,200,80,201,80,202,80,203,80,204,80,205,80,206,80,208,80,209,80,210,80,211,80,212,80,213,80,215,80,216,80,217,80,219,80,220,80,221,80,222,80,223,80,224,80,225,80,226,80,227,80,228,80,229,80,232,80,233,80,234,80,235,80,239,80,240,80,241,80,242,80,244,80,246,80,247,80,248,80,249,80,250,80,252,80,253,80,254,80,255,80,0,81,1,81,2,81,3,81,4,81,5,81,8,81,9,81,10,81,12,81,13,81,14,81,15,81,16,81,17,81,19,81,20,81,21,81,22,81,23,81,24,81,25,81,26,81,27,81,28,81,29,81,30,81,31,81,32,81,34,81,35,81,36,81,37,81,38,81,39,81,40,81,41,81,42,81,43,81,44,81,45,81,46,81,47,81,48,81,49,81,50,81,51,81,52,81,53,81,54,81,55,81,56,81,57,81,58,81,59,81,60,81,61,81,62,81,66,81,71,81,74,81,76,81,78,81,79,81,80,81,82,81,83,81,87,81,88,81,89,81,91,81,93,81,94,81,95,81,96,81,97,81,99,81,100,81,102,81,103,81,105,81,106,81,111,81,114,81,122,81,126,81,127,81,131,81,132,81,134,81,135,81,138,81,139,81,142,81,143,81,144,81,145,81,147,81,148,81,152,81,154,81,157,81,158,81,159,81,161,81,163,81,166,81,167,81,168,81,169,81,170,81,173,81,174,81,180,81,184,81,185,81,186,81,190,81,191,81,193,81,194,81,195,81,197,81,200,81,202,81,205,81,206,81,208,81,210,81,211,81,212,81,213,81,214,81,215,81,216,81,217,81,218,81,220,81,222,81,223,81,226,81,227,81,229,81,230,81,231,81,232,81,233,81,234,81,236,81,238,81,241,81,242,81,244,81,247,81,254,81,4,82,5,82,9,82,11,82,12,82,15,82,16,82,19,82,20,82,21,82,28,82,30,82,31,82,33,82,34,82,35,82,37,82,38,82,39,82,42,82,44,82,47,82,49,82,50,82,52,82,53,82,60,82,62,82,68,82,69,82,70,82,71,82,72,82,73,82,75,82,78,82,79,82,82,82,83,82,85,82,87,82,88,82,89,82,90,82,91,82,93,82,95,82,96,82,98,82,99,82,100,82,102,82,104,82,107,82,108,82,109,82,110,82,112,82,113,82,115,82,116,82,117,82,118,82,119,82,120,82,121,82,122,82,123,82,124,82,126,82,128,82,131,82,132,82,133,82,134,82,135,82,137,82,138,82,139,82,140,82,141,82,142,82,143,82,145,82,146,82,148,82,149,82,150,82,151,82,152,82,153,82,154,82,156,82,164,82,165,82,166,82,167,82,174,82,175,82,176,82,180,82,181,82,182,82,183,82,184,82,185,82,186,82,187,82,188,82,189,82,192,82,193,82,194,82,196,82,197,82,198,82,200,82,202,82,204,82,205,82,206,82,207,82,209,82,211,82,212,82,213,82,215,82,217,82,218,82,219,82,220,82,221,82,222,82,224,82,225,82,226,82,227,82,229,82,230,82,231,82,232,82,233,82,234,82,235,82,236,82,237,82,238,82,239,82,241,82,242,82,243,82,244,82,245,82,246,82,247,82,248,82,251,82,252,82,253,82,1,83,2,83,3,83,4,83,7,83,9,83,10,83,11,83,12,83,14,83,17,83,18,83,19,83,20,83,24,83,27,83,28,83,30,83,31,83,34,83,36,83,37,83,39,83,40,83,41,83,43,83,44,83,45,83,47,83,48,83,49,83,50,83,51,83,52,83,53,83,54,83,55,83,56,83,60,83,61,83,64,83,66,83,68,83,70,83,75,83,76,83,77,83,80,83,84,83,88,83,89,83,91,83,93,83,101,83,104,83,106,83,108,83,109,83,114,83,118,83,121,83,123,83,124,83,125,83,126,83,128,83,129,83,131,83,135,83,136,83,138,83,142,83,143,83,144,83,145,83,146,83,147,83,148,83,150,83,151,83,153,83,155,83,156,83,158,83,160,83,161,83,164,83,167,83,170,83,171,83,172,83,173,83,175,83,176,83,177,83,178,83,179,83,180,83,181,83,183,83,184,83,185,83,186,83,188,83,189,83,190,83,192,83,195,83,196,83,197,83,198,83,199,83,206,83,207,83,208,83,210,83,211,83,213,83,218,83,220,83,221,83,222,83,225,83,226,83,231,83,244,83,250,83,254,83,255,83,0,84,2,84,5,84,7,84,11,84,20,84,24,84,25,84,26,84,28,84,34,84,36,84,37,84,42,84,48,84,51,84,54,84,55,84,58,84,61,84,63,84,65,84,66,84,68,84,69,84,71,84,73,84,76,84,77,84,78,84,79,84,81,84,90,84,93,84,94,84,95,84,96,84,97,84,99,84,101,84,103,84,105,84,106,84,107,84,108,84,109,84,110,84,111,84,112,84,116,84,121,84,122,84,126,84,127,84,129,84,131,84,133,84,135,84,136,84,137,84,138,84,141,84,145,84,147,84,151,84,152,84,156,84,158,84,159,84,160,84,161,84,162,84,165,84,174,84,176,84,178,84,181,84,182,84,183,84,185,84,186,84,188,84,190,84,195,84,197,84,202,84,203,84,214,84,216,84,219,84,224,84,225,84,226,84,227,84,228,84,235,84,236,84,239,84,240,84,241,84,244,84,245,84,246,84,247,84,248,84,249,84,251,84,254,84,0,85,2,85,3,85,4,85,5,85,8,85,10,85,11,85,12,85,13,85,14,85,18,85,19,85,21,85,22,85,23,85,24,85,25,85,26,85,28,85,29,85,30,85,31,85,33,85,37,85,38,85,40,85,41,85,43,85,45,85,50,85,52,85,53,85,54,85,56,85,57,85,58,85,59,85,61,85,64,85,66,85,69,85,71,85,72,85,75,85,76,85,77,85,78,85,79,85,81,85,82,85,83,85,84,85,87,85,88,85,89,85,90,85,91,85,93,85,94,85,95,85,96,85,98,85,99,85,104,85,105,85,107,85,111,85,112,85,113,85,114,85,115,85,116,85,121,85,122,85,125,85,127,85,133,85,134,85,140,85,141,85,142,85,144,85,146,85,147,85,149,85,150,85,151,85,154,85,155,85,158,85,160,85,161,85,162,85,163,85,164,85,165,85,166,85,168,85,169,85,170,85,171,85,172,85,173,85,174,85,175,85,176,85,178,85,180,85,182,85,184,85,186,85,188,85,191,85,192,85,193,85,194,85,195,85,198,85,199,85,200,85,202,85,203,85,206,85,207,85,208,85,213,85,215,85,216,85,217,85,218,85,219,85,222,85,224,85,226,85,231,85,233,85,237,85,238,85,240,85,241,85,244,85,246,85,248,85,249,85,250,85,251,85,252,85,255,85,2,86,3,86,4,86,5,86,6,86,7,86,10,86,11,86,13,86,16,86,17,86,18,86,19,86,20,86,21,86,22,86,23,86,25,86,26,86,28,86,29,86,32,86,33,86,34,86,37,86,38,86,40,86,41,86,42,86,43,86,46,86,47,86,48,86,51,86,53,86,55,86,56,86,58,86,60,86,61,86,62,86,64,86,65,86,66,86,67,86,68,86,69,86,70,86,71,86,72,86,73,86,74,86,75,86,79,86,80,86,81,86,82,86,83,86,85,86,86,86,90,86,91,86,93,86,94,86,95,86,96,86,97,86,99,86,101,86,102,86,103,86,109,86,110,86,111,86,112,86,114,86,115,86,116,86,117,86,119,86,120,86,121,86,122,86,125,86,126,86,127,86,128,86,129,86,130,86,131,86,132,86,135,86,136,86,137,86,138,86,139,86,140,86,141,86,144,86,145,86,146,86,148,86,149,86,150,86,151,86,152,86,153,86,154,86,155,86,156,86,157,86,158,86,159,86,160,86,161,86,162,86,164,86,165,86,166,86,167,86,168,86,169,86,170,86,171,86,172,86,173,86,174,86,176,86,177,86,178,86,179,86,180,86,181,86,182,86,184,86,185,86,186,86,187,86,189,86,190,86,191,86,192,86,193,86,194,86,195,86,196,86,197,86,198,86,199,86,200,86,201,86,203,86,204,86,205,86,206,86,207,86,208,86,209,86,210,86,211,86,213,86,214,86,216,86,217,86,220,86,227,86,229,86,230,86,231,86,232,86,233,86,234,86,236,86,238,86,239,86,242,86,243,86,246,86,247,86,248,86,251,86,252,86,0,87,1,87,2,87,5,87,7,87,11,87,12,87,13,87,14,87,15,87,16,87,17,87,18,87,19,87,20,87,21,87,22,87,23,87,24,87,25,87,26,87,27,87,29,87,30,87,32,87,33,87,34,87,36,87,37,87,38,87,39,87,43,87,49,87,50,87,52,87,53,87,54,87,55,87,56,87,60,87,61,87,63,87,65,87,67,87,68,87,69,87,70,87,72,87,73,87,75,87,82,87,83,87,84,87,85,87,86,87,88,87,89,87,98,87,99,87,101,87,103,87,108,87,110,87,112,87,113,87,114,87,116,87,117,87,120,87,121,87,122,87,125,87,126,87,127,87,128,87,129,87,135,87,136,87,137,87,138,87,141,87,142,87,143,87,144,87,145,87,148,87,149,87,150,87,151,87,152,87,153,87,154,87,156,87,157,87,158,87,159,87,165,87,168,87,170,87,172,87,175,87,176,87,177,87,179,87,181,87,182,87,183,87,185,87,186,87,187,87,188,87,189,87,190,87,191,87,192,87,193,87,196,87,197,87,198,87,199,87,200,87,201,87,202,87,204,87,205,87,208,87,209,87,211,87,214,87,215,87,219,87,220,87,222,87,225,87,226,87,227,87,229,87,230,87,231,87,232,87,233,87,234,87,235,87,236,87,238,87,240,87,241,87,242,87,243,87,245,87,246,87,247,87,251,87,252,87,254,87,255,87,1,88,3,88,4,88,5,88,8,88,9,88,10,88,12,88,14,88,15,88,16,88,18,88,19,88,20,88,22,88,23,88,24,88,26,88,27,88,28,88,29,88,31,88,34,88,35,88,37,88,38,88,39,88,40,88,41,88,43,88,44,88,45,88,46,88,47,88,49,88,50,88,51,88,52,88,54,88,55,88,56,88,57,88,58,88,59,88,60,88,61,88,62,88,63,88,64,88,65,88,66,88,67,88,69,88,70,88,71,88,72,88,73,88,74,88,75,88,78,88,79,88,80,88,82,88,83,88,85,88,86,88,87,88,89,88,90,88,91,88,92,88,93,88,95,88,96,88,97,88,98,88,99,88,100,88,102,88,103,88,104,88,105,88,106,88,109,88,110,88,111,88,112,88,113,88,114,88,115,88,116,88,117,88,118,88,119,88,120,88,121,88,122,88,123,88,124,88,125,88,127,88,130,88,132,88,134,88,135,88,136,88,138,88,139,88,140,88,141,88,142,88,143,88,144,88,145,88,148,88,149,88,150,88,151,88,152,88,155,88,156,88,157,88,160,88,161,88,162,88,163,88,164,88,165,88,166,88,167,88,170,88,171,88,172,88,173,88,174,88,175,88,176,88,177,88,178,88,179,88,180,88,181,88,182,88,183,88,184,88,185,88,186,88,187,88,189,88,190,88,191,88,192,88,194,88,195,88,196,88,198,88,199,88,200,88,201,88,202,88,203,88,204,88,205,88,206,88,207,88,208,88,210,88,211,88,212,88,214,88,215,88,216,88,217,88,218,88,219,88,220,88,221,88,222,88,223,88,224,88,225,88,226,88,227,88,229,88,230,88,231,88,232,88,233,88,234,88,237,88,239,88,241,88,242,88,244,88,245,88,247,88,248,88,250,88,251,88,252,88,253,88,254,88,255,88,0,89,1,89,3,89,5,89,6,89,8,89,9,89,10,89,11,89,12,89,14,89,16,89,17,89,18,89,19,89,23,89,24,89,27,89,29,89,30,89,32,89,33,89,34,89,35,89,38,89,40,89,44,89,48,89,50,89,51,89,53,89,54,89,59,89,61,89,62,89,63,89,64,89,67,89,69,89,70,89,74,89,76,89,77,89,80,89,82,89,83,89,89,89,91,89,92,89,93,89,94,89,95,89,97,89,99,89,100,89,102,89,103,89,104,89,105,89,106,89,107,89,108,89,109,89,110,89,111,89,112,89,113,89,114,89,117,89,119,89,122,89,123,89,124,89,126,89,127,89,128,89,133,89,137,89,139,89,140,89,142,89,143,89,144,89,145,89,148,89,149,89,152,89,154,89,155,89,156,89,157,89,159,89,160,89,161,89,162,89,166,89,167,89,172,89,173,89,176,89,177,89,179,89,180,89,181,89,182,89,183,89,184,89,186,89,188,89,189,89,191,89,192,89,193,89,194,89,195,89,196,89,197,89,199,89,200,89,201,89,204,89,205,89,206,89,207,89,213,89,214,89,217,89,219,89,222,89,223,89,224,89,225,89,226,89,228,89,230,89,231,89,233,89,234,89,235,89,237,89,238,89,239,89,240,89,241,89,242,89,243,89,244,89,245,89,246,89,247,89,248,89,250,89,252,89,253,89,254,89,0,90,2,90,10,90,11,90,13,90,14,90,15,90,16,90,18,90,20,90,21,90,22,90,23,90,25,90,26,90,27,90,29,90,30,90,33,90,34,90,36,90,38,90,39,90,40,90,42,90,43,90,44,90,45,90,46,90,47,90,48,90,51,90,53,90,55,90,56,90,57,90,58,90,59,90,61,90,62,90,63,90,65,90,66,90,67,90,68,90,69,90,71,90,72,90,75,90,76,90,77,90,78,90,79,90,80,90,81,90,82,90,83,90,84,90,86,90,87,90,88,90,89,90,91,90,92,90,93,90,94,90,95,90,96,90,97,90,99,90,100,90,101,90,102,90,104,90,105,90,107,90,108,90,109,90,110,90,111,90,112,90,113,90,114,90,115,90,120,90,121,90,123,90,124,90,125,90,126,90,128,90,129,90,130,90,131,90,132,90,133,90,134,90,135,90,136,90,137,90,138,90,139,90,140,90,141,90,142,90,143,90,144,90,145,90,147,90,148,90,149,90,150,90,151,90,152,90,153,90,156,90,157,90,158,90,159,90,160,90,161,90,162,90,163,90,164,90,165,90,166,90,167,90,168,90,169,90,171,90,172,90,173,90,174,90,175,90,176,90,177,90,180,90,182,90,183,90,185,90,186,90,187,90,188,90,189,90,191,90,192,90,195,90,196,90,197,90,198,90,199,90,200,90,202,90,203,90,205,90,206,90,207,90,208,90,209,90,211,90,213,90,215,90,217,90,218,90,219,90,221,90,222,90,223,90,226,90,228,90,229,90,231,90,232,90,234,90,236,90,237,90,238,90,239,90,240,90,242,90,243,90,244,90,245,90,246,90,247,90,248,90,249,90,250,90,251,90,252,90,253,90,254,90,255,90,0,91,1,91,2,91,3,91,4,91,5,91,6,91,7,91,8,91,10,91,11,91,12,91,13,91,14,91,15,91,16,91,17,91,18,91,19,91,20,91,21,91,24,91,25,91,26,91,27,91,28,91,29,91,30,91,31,91,32,91,33,91,34,91,35,91,36,91,37,91,38,91,39,91,40,91,41,91,42,91,43,91,44,91,45,91,46,91,47,91,48,91,49,91,51,91,53,91,54,91,56,91,57,91,58,91,59,91,60,91,61,91,62,91,63,91,65,91,66,91,67,91,68,91,69,91,70,91,71,91,72,91,73,91,74,91,75,91,76,91,77,91,78,91,79,91,82,91,86,91,94,91,96,91,97,91,103,91,104,91,107,91,109,91,110,91,111,91,114,91,116,91,118,91,119,91,120,91,121,91,123,91,124,91,126,91,127,91,130,91,134,91,138,91,141,91,142,91,144,91,145,91,146,91,148,91,150,91,159,91,167,91,168,91,169,91,172,91,173,91,174,91,175,91,177,91,178,91,183,91,186,91,187,91,188,91,192,91,193,91,195,91,200,91,201,91,202,91,203,91,205,91,206,91,207,91,209,91,212,91,213,91,214,91,215,91,216,91,217,91,218,91,219,91,220,91,224,91,226,91,227,91,230,91,231,91,233,91,234,91,235,91,236,91,237,91,239,91,241,91,242,91,243,91,244,91,245,91,246,91,247,91,253,91,254,91,0,92,2,92,3,92,5,92,7,92,8,92,11,92,12,92,13,92,14,92,16,92,18,92,19,92,23,92,25,92,27,92,30,92,31,92,32,92,33,92,35,92,38,92,40,92,41,92,42,92,43,92,45,92,46,92,47,92,48,92,50,92,51,92,53,92,54,92,55,92,67,92,68,92,70,92,71,92,76,92,77,92,82,92,83,92,84,92,86,92,87,92,88,92,90,92,91,92,92,92,93,92,95,92,98,92,100,92,103,92,104,92,105,92,106,92,107,92,108,92,109,92,112,92,114,92,115,92,116,92,117,92,118,92,119,92,120,92,123,92,124,92,125,92,126,92,128,92,131,92,132,92,133,92,134,92,135,92,137,92,138,92,139,92,142,92,143,92,146,92,147,92,149,92,157,92,158,92,159,92,160,92,161,92,164,92,165,92,166,92,167,92,168,92,170,92,174,92,175,92,176,92,178,92,180,92,182,92,185,92,186,92,187,92,188,92,190,92,192,92,194,92,195,92,197,92,198,92,199,92,200,92,201,92,202,92,204,92,205,92,206,92,207,92,208,92,209,92,211,92,212,92,213,92,214,92,215,92,216,92,218,92,219,92,220,92,221,92,222,92,223,92,224,92,226,92,227,92,231,92,233,92,235,92,236,92,238,92,239,92,241,92,242,92,243,92,244,92,245,92,246,92,247,92,248,92,249,92,250,92,252,92,253,92,254,92,255,92,0,93,1,93,4,93,5,93,8,93,9,93,10,93,11,93,12,93,13,93,15,93,16,93,17,93,18,93,19,93,21,93,23,93,24,93,25,93,26,93,28,93,29,93,31,93,32,93,33,93,34,93,35,93,37,93,40,93,42,93,43,93,44,93,47,93,48,93,49,93,50,93,51,93,53,93,54,93,55,93,56,93,57,93,58,93,59,93,60,93,63,93,64,93,65,93,66,93,67,93,68,93,69,93,70,93,72,93,73,93,77,93,78,93,79,93,80,93,81,93,82,93,83,93,84,93,85,93,86,93,87,93,89,93,90,93,92,93,94,93,95,93,96,93,97,93,98,93,99,93,100,93,101,93,102,93,103,93,104,93,106,93,109,93,110,93,112,93,113,93,114,93,115,93,117,93,118,93,119,93,120,93,121,93,122,93,123,93,124,93,125,93,126,93,127,93,128,93,129,93,131,93,132,93,133,93,134,93,135,93,136,93,137,93,138,93,139,93,140,93,141,93,142,93,143,93,144,93,145,93,146,93,147,93,148,93,149,93,150,93,151,93,152,93,154,93,155,93,156,93,158,93,159,93,160,93,161,93,162,93,163,93,164,93,165,93,166,93,167,93,168,93,169,93,170,93,171,93,172,93,173,93,174,93,175,93,176,93,177,93,178,93,179,93,180,93,181,93,182,93,184,93,185,93,186,93,187,93,188,93,189,93,190,93,191,93,192,93,193,93,194,93,195,93,196,93,198,93,199,93,200,93,201,93,202,93,203,93,204,93,206,93,207,93,208,93,209,93,210,93,211,93,212,93,213,93,214,93,215,93,216,93,217,93,218,93,220,93,223,93,224,93,227,93,228,93,234,93,236,93,237,93,240,93,245,93,246,93,248,93,249,93,250,93,251,93,252,93,255,93,0,94,4,94,7,94,9,94,10,94,11,94,13,94,14,94,18,94,19,94,23,94,30,94,31,94,32,94,33,94,34,94,35,94,36,94,37,94,40,94,41,94,42,94,43,94,44,94,47,94,48,94,50,94,51,94,52,94,53,94,54,94,57,94,58,94,62,94,63,94,64,94,65,94,67,94,70,94,71,94,72,94,73,94,74,94,75,94,77,94,78,94,79,94,80,94,81,94,82,94,83,94,86,94,87,94,88,94,89,94,90,94,92,94,93,94,95,94,96,94,99,94,100,94,101,94,102,94,103,94,104,94,105,94,106,94,107,94,108,94,109,94,110,94,111,94,112,94,113,94,117,94,119,94,121,94,126,94,129,94,130,94,131,94,133,94,136,94,137,94,140,94,141,94,142,94,146,94,152,94,155,94,157,94,161,94,162,94,163,94,164,94,168,94,169,94,170,94,171,94,172,94,174,94,175,94,176,94,177,94,178,94,180,94,186,94,187,94,188,94,189,94,191,94,192,94,193,94,194,94,195,94,196,94,197,94,198,94,199,94,200,94,203,94,204,94,205,94,206,94,207,94,208,94,212,94,213,94,215,94,216,94,217,94,218,94,220,94,221,94,222,94,223,94,224,94,225,94,226,94,227,94,228,94,229,94,230,94,231,94,233,94,235,94,236,94,237,94,238,94,239,94,240,94,241,94,242,94,243,94,245,94,248,94,249,94,251,94,252,94,253,94,5,95,6,95,7,95,9,95,12,95,13,95,14,95,16,95,18,95,20,95,22,95,25,95,26,95,28,95,29,95,30,95,33,95,34,95,35,95,36,95,40,95,43,95,44,95,46,95,48,95,50,95,51,95,52,95,53,95,54,95,55,95,56,95,59,95,61,95,62,95,63,95,65,95,66,95,67,95,68,95,69,95,70,95,71,95,72,95,73,95,74,95,75,95,76,95,77,95,78,95,79,95,81,95,84,95,89,95,90,95,91,95,92,95,94,95,95,95,96,95,99,95,101,95,103,95,104,95,107,95,110,95,111,95,114,95,116,95,117,95,118,95,120,95,122,95,125,95,126,95,127,95,131,95,134,95,141,95,142,95,143,95,145,95,147,95,148,95,150,95,154,95,155,95,157,95,158,95,159,95,160,95,162,95,163,95,164,95,165,95,166,95,167,95,169,95,171,95,172,95,175,95,176,95,177,95,178,95,179,95,180,95,182,95,184,95,185,95,186,95,187,95,190,95,191,95,192,95,193,95,194,95,199,95,200,95,202,95,203,95,206,95,211,95,212,95,213,95,218,95,219,95,220,95,222,95,223,95,226,95,227,95,229,95,230,95,232,95,233,95,236,95,239,95,240,95,242,95,243,95,244,95,246,95,247,95,249,95,250,95,252,95,7,96,8,96,9,96,11,96,12,96,16,96,17,96,19,96,23,96,24,96,26,96,30,96,31,96,34,96,35,96,36,96,44,96,45,96,46,96,48,96,49,96,50,96,51,96,52,96,54,96,55,96,56,96,57,96,58,96,61,96,62,96,64,96,68,96,69,96,70,96,71,96,72,96,73,96,74,96,76,96,78,96,79,96,81,96,83,96,84,96,86,96,87,96,88,96,91,96,92,96,94,96,95,96,96,96,97,96,101,96,102,96,110,96,113,96,114,96,116,96,117,96,119,96,126,96,128,96,129,96,130,96,133,96,134,96,135,96,136,96,138,96,139,96,142,96,143,96,144,96,145,96,147,96,149,96,151,96,152,96,153,96,156,96,158,96,161,96,162,96,164,96,165,96,167,96,169,96,170,96,174,96,176,96,179,96,181,96,182,96,183,96,185,96,186,96,189,96,190,96,191,96,192,96,193,96,194,96,195,96,196,96,199,96,200,96,201,96,204,96,205,96,206,96,207,96,208,96,210,96,211,96,212,96,214,96,215,96,217,96,219,96,222,96,225,96,226,96,227,96,228,96,229,96,234,96,241,96,242,96,245,96,247,96,248,96,251,96,252,96,253,96,254,96,255,96,2,97,3,97,4,97,5,97,7,97,10,97,11,97,12,97,16,97,17,97,18,97,19,97,20,97,22,97,23,97,24,97,25,97,27,97,28,97,29,97,30,97,33,97,34,97,37,97,40,97,41,97,42,97,44,97,45,97,46,97,47,97,48,97,49,97,50,97,51,97,52,97,53,97,54,97,55,97,56,97,57,97,58,97,59,97,60,97,61,97,62,97,64,97,65,97,66,97,67,97,68,97,69,97,70,97,71,97,73,97,75,97,77,97,79,97,80,97,82,97,83,97,84,97,86,97,87,97,88,97,89,97,90,97,91,97,92,97,94,97,95,97,96,97,97,97,99,97,100,97,101,97,102,97,105,97,106,97,107,97,108,97,109,97,110,97,111,97,113,97,114,97,115,97,116,97,118,97,120,97,121,97,122,97,123,97,124,97,125,97,126,97,127,97,128,97,129,97,130,97,131,97,132,97,133,97,134,97,135,97,136,97,137,97,138,97,140,97,141,97,143,97,144,97,145,97,146,97,147,97,149,97,150,97,151,97,152,97,153,97,154,97,155,97,156,97,158,97,159,97,160,97,161,97,162,97,163,97,164,97,165,97,166,97,170,97,171,97,173,97,174,97,175,97,176,97,177,97,178,97,179,97,180,97,181,97,182,97,184,97,185,97,186,97,187,97,188,97,189,97,191,97,192,97,193,97,195,97,196,97,197,97,198,97,199,97,201,97,204,97,205,97,206,97,207,97,208,97,211,97,213,97,214,97,215,97,216,97,217,97,218,97,219,97,220,97,221,97,222,97,223,97,224,97,225,97,226,97,227,97,228,97,229,97,231,97,232,97,233,97,234,97,235,97,236,97,237,97,238,97,239,97,240,97,241,97,242,97,243,97,244,97,246,97,247,97,248,97,249,97,250,97,251,97,252,97,253,97,254,97,0,98,1,98,2,98,3,98,4,98,5,98,7,98,9,98,19,98,20,98,25,98,28,98,29,98,30,98,32,98,35,98,38,98,39,98,40,98,41,98,43,98,45,98,47,98,48,98,49,98,50,98,53,98,54,98,56,98,57,98,58,98,59,98,60,98,66,98,68,98,69,98,70,98,74,98,79,98,80,98,85,98,86,98,87,98,89,98,90,98,92,98,93,98,94,98,95,98,96,98,97,98,98,98,100,98,101,98,104,98,113,98,114,98,116,98,117,98,119,98,120,98,122,98,123,98,125,98,129,98,130,98,131,98,133,98,134,98,135,98,136,98,139,98,140,98,141,98,142,98,143,98,144,98,148,98,153,98,156,98,157,98,158,98,163,98,166,98,167,98,169,98,170,98,173,98,174,98,175,98,176,98,178,98,179,98,180,98,182,98,183,98,184,98,186,98,190,98,192,98,193,98,195,98,203,98,207,98,209,98,213,98,221,98,222,98,224,98,225,98,228,98,234,98,235,98,240,98,242,98,245,98,248,98,249,98,250,98,251,98,0,99,3,99,4,99,5,99,6,99,10,99,11,99,12,99,13,99,15,99,16,99,18,99,19,99,20,99,21,99,23,99,24,99,25,99,28,99,38,99,39,99,41,99,44,99,45,99,46,99,48,99,49,99,51,99,52,99,53,99,54,99,55,99,56,99,59,99,60,99,62,99,63,99,64,99,65,99,68,99,71,99,72,99,74,99,81,99,82,99,83,99,84,99,86,99,87,99,88,99,89,99,90,99,91,99,92,99,93,99,96,99,100,99,101,99,102,99,104,99,106,99,107,99,108,99,111,99,112,99,114,99,115,99,116,99,117,99,120,99,121,99,124,99,125,99,126,99,127,99,129,99,131,99,132,99,133,99,134,99,139,99,141,99,145,99,147,99,148,99,149,99,151,99,153,99,154,99,155,99,156,99,157,99,158,99,159,99,161,99,164,99,166,99,171,99,175,99,177,99,178,99,181,99,182,99,185,99,187,99,189,99,191,99,192,99,193,99,194,99,195,99,197,99,199,99,200,99,202,99,203,99,204,99,209,99,211,99,212,99,213,99,215,99,216,99,217,99,218,99,219,99,220,99,221,99,223,99,226,99,228,99,229,99,230,99,231,99,232,99,235,99,236,99,238,99,239,99,240,99,241,99,243,99,245,99,247,99,249,99,250,99,251,99,252,99,254,99,3,100,4,100,6,100,7,100,8,100,9,100,10,100,13,100,14,100,17,100,18,100,21,100,22,100,23,100,24,100,25,100,26,100,29,100,31,100,34,100,35,100,36,100,37,100,39,100,40,100,41,100,43,100,46,100,47,100,48,100,49,100,50,100,51,100,53,100,54,100,55,100,56,100,57,100,59,100,60,100,62,100,64,100,66,100,67,100,73,100,75,100,76,100,77,100,78,100,79,100,80,100,81,100,83,100,85,100,86,100,87,100,89,100,90,100,91,100,92,100,93,100,95,100,96,100,97,100,98,100,99,100,100,100,101,100,102,100,104,100,106,100,107,100,108,100,110,100,111,100,112,100,113,100,114,100,115,100,116,100,117,100,118,100,119,100,123,100,124,100,125,100,126,100,127,100,128,100,129,100,131,100,134,100,136,100,137,100,138,100,139,100,140,100,141,100,142,100,143,100,144,100,147,100,148,100,151,100,152,100,154,100,155,100,156,100,157,100,159,100,160,100,161,100,162,100,163,100,165,100,166,100,167,100,168,100,170,100,171,100,175,100,177,100,178,100,179,100,180,100,182,100,185,100,187,100,189,100,190,100,191,100,193,100,195,100,196,100,198,100,199,100,200,100,201,100,202,100,203,100,204,100,207,100,209,100,211,100,212,100,213,100,214,100,217,100,218,100,219,100,220,100,221,100,223,100,224,100,225,100,227,100,229,100,231,100,232,100,233,100,234,100,235,100,236,100,237,100,238,100,239,100,240,100,241,100,242,100,243,100,244,100,245,100,246,100,247,100,248,100,249,100,250,100,251,100,252,100,253,100,254,100,255,100,1,101,2,101,3,101,4,101,5,101,6,101,7,101,8,101,10,101,11,101,12,101,13,101,14,101,15,101,16,101,17,101,19,101,20,101,21,101,22,101,23,101,25,101,26,101,27,101,28,101,29,101,30,101,31,101,32,101,33,101,34,101,35,101,36,101,38,101,39,101,40,101,41,101,42,101,44,101,45,101,48,101,49,101,50,101,51,101,55,101,58,101,60,101,61,101,64,101,65,101,66,101,67,101,68,101,70,101,71,101,74,101,75,101,77,101,78,101,80,101,82,101,83,101,84,101,87,101,88,101,90,101,92,101,95,101,96,101,97,101,100,101,101,101,103,101,104,101,105,101,106,101,109,101,110,101,111,101,113,101,115,101,117,101,118,101,120,101,121,101,122,101,123,101,124,101,125,101,126,101,127,101,128,101,129,101,130,101,131,101,132,101,133,101,134,101,136,101,137,101,138,101,141,101,142,101,143,101,146,101,148,101,149,101,150,101,152,101,154,101,157,101], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+36088);
/* memory initializer */ allocate([158,101,160,101,162,101,163,101,166,101,168,101,170,101,172,101,174,101,177,101,178,101,179,101,180,101,181,101,182,101,183,101,184,101,186,101,187,101,190,101,191,101,192,101,194,101,199,101,200,101,201,101,202,101,205,101,208,101,209,101,211,101,212,101,213,101,216,101,217,101,218,101,219,101,220,101,221,101,222,101,223,101,225,101,227,101,228,101,234,101,235,101,242,101,243,101,244,101,245,101,248,101,249,101,251,101,252,101,253,101,254,101,255,101,1,102,4,102,5,102,7,102,8,102,9,102,11,102,13,102,16,102,17,102,18,102,22,102,23,102,24,102,26,102,27,102,28,102,30,102,33,102,34,102,35,102,36,102,38,102,41,102,42,102,43,102,44,102,46,102,48,102,50,102,51,102,55,102,56,102,57,102,58,102,59,102,61,102,63,102,64,102,66,102,68,102,69,102,70,102,71,102,72,102,73,102,74,102,77,102,78,102,80,102,81,102,88,102,89,102,91,102,92,102,93,102,94,102,96,102,98,102,99,102,101,102,103,102,105,102,106,102,107,102,108,102,109,102,113,102,114,102,115,102,117,102,120,102,121,102,123,102,124,102,125,102,127,102,128,102,129,102,131,102,133,102,134,102,136,102,137,102,138,102,139,102,141,102,142,102,143,102,144,102,146,102,147,102,148,102,149,102,152,102,153,102,154,102,155,102,156,102,158,102,159,102,160,102,161,102,162,102,163,102,164,102,165,102,166,102,169,102,170,102,171,102,172,102,173,102,175,102,176,102,177,102,178,102,179,102,181,102,182,102,183,102,184,102,186,102,187,102,188,102,189,102,191,102,192,102,193,102,194,102,195,102,196,102,197,102,198,102,199,102,200,102,201,102,202,102,203,102,204,102,205,102,206,102,207,102,208,102,209,102,210,102,211,102,212,102,213,102,214,102,215,102,216,102,218,102,222,102,223,102,224,102,225,102,226,102,227,102,228,102,229,102,231,102,232,102,234,102,235,102,236,102,237,102,238,102,239,102,241,102,245,102,246,102,248,102,250,102,251,102,253,102,1,103,2,103,3,103,4,103,5,103,6,103,7,103,12,103,14,103,15,103,17,103,18,103,19,103,22,103,24,103,25,103,26,103,28,103,30,103,32,103,33,103,34,103,35,103,36,103,37,103,39,103,41,103,46,103,48,103,50,103,51,103,54,103,55,103,56,103,57,103,59,103,60,103,62,103,63,103,65,103,68,103,69,103,71,103,74,103,75,103,77,103,82,103,84,103,85,103,87,103,88,103,89,103,90,103,91,103,93,103,98,103,99,103,100,103,102,103,103,103,107,103,108,103,110,103,113,103,116,103,118,103,120,103,121,103,122,103,123,103,125,103,128,103,130,103,131,103,133,103,134,103,136,103,138,103,140,103,141,103,142,103,143,103,145,103,146,103,147,103,148,103,150,103,153,103,155,103,159,103,160,103,161,103,164,103,166,103,169,103,172,103,174,103,177,103,178,103,180,103,185,103,186,103,187,103,188,103,189,103,190,103,191,103,192,103,194,103,197,103,198,103,199,103,200,103,201,103,202,103,203,103,204,103,205,103,206,103,213,103,214,103,215,103,219,103,223,103,225,103,227,103,228,103,230,103,231,103,232,103,234,103,235,103,237,103,238,103,242,103,245,103,246,103,247,103,248,103,249,103,250,103,251,103,252,103,254,103,1,104,2,104,3,104,4,104,6,104,13,104,16,104,18,104,20,104,21,104,24,104,25,104,26,104,27,104,28,104,30,104,31,104,32,104,34,104,35,104,36,104,37,104,38,104,39,104,40,104,43,104,44,104,45,104,46,104,47,104,48,104,49,104,52,104,53,104,54,104,58,104,59,104,63,104,71,104,75,104,77,104,79,104,82,104,86,104,87,104,88,104,89,104,90,104,91,104,92,104,93,104,94,104,95,104,106,104,108,104,109,104,110,104,111,104,112,104,113,104,114,104,115,104,117,104,120,104,121,104,122,104,123,104,124,104,125,104,126,104,127,104,128,104,130,104,132,104,135,104,136,104,137,104,138,104,139,104,140,104,141,104,142,104,144,104,145,104,146,104,148,104,149,104,150,104,152,104,153,104,154,104,155,104,156,104,157,104,158,104,159,104,160,104,161,104,163,104,164,104,165,104,169,104,170,104,171,104,172,104,174,104,177,104,178,104,180,104,182,104,183,104,184,104,185,104,186,104,187,104,188,104,189,104,190,104,191,104,193,104,195,104,196,104,197,104,198,104,199,104,200,104,202,104,204,104,206,104,207,104,208,104,209,104,211,104,212,104,214,104,215,104,217,104,219,104,220,104,221,104,222,104,223,104,225,104,226,104,228,104,229,104,230,104,231,104,232,104,233,104,234,104,235,104,236,104,237,104,239,104,242,104,243,104,244,104,246,104,247,104,248,104,251,104,253,104,254,104,255,104,0,105,2,105,3,105,4,105,6,105,7,105,8,105,9,105,10,105,12,105,15,105,17,105,19,105,20,105,21,105,22,105,23,105,24,105,25,105,26,105,27,105,28,105,29,105,30,105,33,105,34,105,35,105,37,105,38,105,39,105,40,105,41,105,42,105,43,105,44,105,46,105,47,105,49,105,50,105,51,105,53,105,54,105,55,105,56,105,58,105,59,105,60,105,62,105,64,105,65,105,67,105,68,105,69,105,70,105,71,105,72,105,73,105,74,105,75,105,76,105,77,105,78,105,79,105,80,105,81,105,82,105,83,105,85,105,86,105,88,105,89,105,91,105,92,105,95,105,97,105,98,105,100,105,101,105,103,105,104,105,105,105,106,105,108,105,109,105,111,105,112,105,114,105,115,105,116,105,117,105,118,105,122,105,123,105,125,105,126,105,127,105,129,105,131,105,133,105,138,105,139,105,140,105,142,105,143,105,144,105,145,105,146,105,147,105,150,105,151,105,153,105,154,105,157,105,158,105,159,105,160,105,161,105,162,105,163,105,164,105,165,105,166,105,169,105,170,105,172,105,174,105,175,105,176,105,178,105,179,105,181,105,182,105,184,105,185,105,186,105,188,105,189,105,190,105,191,105,192,105,194,105,195,105,196,105,197,105,198,105,199,105,200,105,201,105,203,105,205,105,207,105,209,105,210,105,211,105,213,105,214,105,215,105,216,105,217,105,218,105,220,105,221,105,222,105,225,105,226,105,227,105,228,105,229,105,230,105,231,105,232,105,233,105,234,105,235,105,236,105,238,105,239,105,240,105,241,105,243,105,244,105,245,105,246,105,247,105,248,105,249,105,250,105,251,105,252,105,254,105,0,106,1,106,2,106,3,106,4,106,5,106,6,106,7,106,8,106,9,106,11,106,12,106,13,106,14,106,15,106,16,106,17,106,18,106,19,106,20,106,21,106,22,106,25,106,26,106,27,106,28,106,29,106,30,106,32,106,34,106,35,106,36,106,37,106,38,106,39,106,41,106,43,106,44,106,45,106,46,106,48,106,50,106,51,106,52,106,54,106,55,106,56,106,57,106,58,106,59,106,60,106,63,106,64,106,65,106,66,106,67,106,69,106,70,106,72,106,73,106,74,106,75,106,76,106,77,106,78,106,79,106,81,106,82,106,83,106,84,106,85,106,86,106,87,106,90,106,92,106,93,106,94,106,95,106,96,106,98,106,99,106,100,106,102,106,103,106,104,106,105,106,106,106,107,106,108,106,109,106,110,106,111,106,112,106,114,106,115,106,116,106,117,106,118,106,119,106,120,106,122,106,123,106,125,106,126,106,127,106,129,106,130,106,131,106,133,106,134,106,135,106,136,106,137,106,138,106,139,106,140,106,141,106,143,106,146,106,147,106,148,106,149,106,150,106,152,106,153,106,154,106,155,106,156,106,157,106,158,106,159,106,161,106,162,106,163,106,164,106,165,106,166,106,167,106,168,106,170,106,173,106,174,106,175,106,176,106,177,106,178,106,179,106,180,106,181,106,182,106,183,106,184,106,185,106,186,106,187,106,188,106,189,106,190,106,191,106,192,106,193,106,194,106,195,106,196,106,197,106,198,106,199,106,200,106,201,106,202,106,203,106,204,106,205,106,206,106,207,106,208,106,209,106,210,106,211,106,212,106,213,106,214,106,215,106,216,106,217,106,218,106,219,106,220,106,221,106,222,106,223,106,224,106,225,106,226,106,227,106,228,106,229,106,230,106,231,106,232,106,233,106,234,106,235,106,236,106,237,106,238,106,239,106,240,106,241,106,242,106,243,106,244,106,245,106,246,106,247,106,248,106,249,106,250,106,251,106,252,106,253,106,254,106,255,106,0,107,1,107,2,107,3,107,4,107,5,107,6,107,7,107,8,107,9,107,10,107,11,107,12,107,13,107,14,107,15,107,16,107,17,107,18,107,19,107,20,107,21,107,22,107,23,107,24,107,25,107,26,107,27,107,28,107,29,107,30,107,31,107,37,107,38,107,40,107,41,107,42,107,43,107,44,107,45,107,46,107,47,107,48,107,49,107,51,107,52,107,53,107,54,107,56,107,59,107,60,107,61,107,63,107,64,107,65,107,66,107,68,107,69,107,72,107,74,107,75,107,77,107,78,107,79,107,80,107,81,107,82,107,83,107,84,107,85,107,86,107,87,107,88,107,90,107,91,107,92,107,93,107,94,107,95,107,96,107,97,107,104,107,105,107,107,107,108,107,109,107,110,107,111,107,112,107,113,107,114,107,115,107,116,107,117,107,118,107,119,107,120,107,122,107,125,107,126,107,127,107,128,107,133,107,136,107,140,107,142,107,143,107,144,107,145,107,148,107,149,107,151,107,152,107,153,107,156,107,157,107,158,107,159,107,160,107,162,107,163,107,164,107,165,107,166,107,167,107,168,107,169,107,171,107,172,107,173,107,174,107,175,107,176,107,177,107,178,107,182,107,184,107,185,107,186,107,187,107,188,107,189,107,190,107,192,107,195,107,196,107,198,107,199,107,200,107,201,107,202,107,204,107,206,107,208,107,209,107,216,107,218,107,220,107,221,107,222,107,223,107,224,107,226,107,227,107,228,107,229,107,230,107,231,107,232,107,233,107,236,107,237,107,238,107,240,107,241,107,242,107,244,107,246,107,247,107,248,107,250,107,251,107,252,107,254,107,255,107,0,108,1,108,2,108,3,108,4,108,8,108,9,108,10,108,11,108,12,108,14,108,18,108,23,108,28,108,29,108,30,108,32,108,35,108,37,108,43,108,44,108,45,108,49,108,51,108,54,108,55,108,57,108,58,108,59,108,60,108,62,108,63,108,67,108,68,108,69,108,72,108,75,108,76,108,77,108,78,108,79,108,81,108,82,108,83,108,86,108,88,108,89,108,90,108,98,108,99,108,101,108,102,108,103,108,107,108,108,108,109,108,110,108,111,108,113,108,115,108,117,108,119,108,120,108,122,108,123,108,124,108,127,108,128,108,132,108,135,108,138,108,139,108,141,108,142,108,145,108,146,108,149,108,150,108,151,108,152,108,154,108,156,108,157,108,158,108,160,108,162,108,168,108,172,108,175,108,176,108,180,108,181,108,182,108,183,108,186,108,192,108,193,108,194,108,195,108,198,108,199,108,200,108,203,108,205,108,206,108,207,108,209,108,210,108,216,108,217,108,218,108,220,108,221,108,223,108,228,108,230,108,231,108,233,108,236,108,237,108,242,108,244,108,249,108,255,108,0,109,2,109,3,109,5,109,6,109,8,109,9,109,10,109,13,109,15,109,16,109,17,109,19,109,20,109,21,109,22,109,24,109,28,109,29,109,31,109,32,109,33,109,34,109,35,109,36,109,38,109,40,109,41,109,44,109,45,109,47,109,48,109,52,109,54,109,55,109,56,109,58,109,63,109,64,109,66,109,68,109,73,109,76,109,80,109,85,109,86,109,87,109,88,109,91,109,93,109,95,109,97,109,98,109,100,109,101,109,103,109,104,109,107,109,108,109,109,109,112,109,113,109,114,109,115,109,117,109,118,109,121,109,122,109,123,109,125,109,126,109,127,109,128,109,129,109,131,109,132,109,134,109,135,109,138,109,139,109,141,109,143,109,144,109,146,109,150,109,151,109,152,109,153,109,154,109,156,109,162,109,165,109,172,109,173,109,176,109,177,109,179,109,180,109,182,109,183,109,185,109,186,109,187,109,188,109,189,109,190,109,193,109,194,109,195,109,200,109,201,109,202,109,205,109,206,109,207,109,208,109,210,109,211,109,212,109,213,109,215,109,218,109,219,109,220,109,223,109,226,109,227,109,229,109,231,109,232,109,233,109,234,109,237,109,239,109,240,109,242,109,244,109,245,109,246,109,248,109,250,109,253,109,254,109,255,109,0,110,1,110,2,110,3,110,4,110,6,110,7,110,8,110,9,110,11,110,15,110,18,110,19,110,21,110,24,110,25,110,27,110,28,110,30,110,31,110,34,110,38,110,39,110,40,110,42,110,44,110,46,110,48,110,49,110,51,110,53,110,54,110,55,110,57,110,59,110,60,110,61,110,62,110,63,110,64,110,65,110,66,110,69,110,70,110,71,110,72,110,73,110,74,110,75,110,76,110,79,110,80,110,81,110,82,110,85,110,87,110,89,110,90,110,92,110,93,110,94,110,96,110,97,110,98,110,99,110,100,110,101,110,102,110,103,110,104,110,105,110,106,110,108,110,109,110,111,110,112,110,113,110,114,110,115,110,116,110,117,110,118,110,119,110,120,110,121,110,122,110,123,110,124,110,125,110,128,110,129,110,130,110,132,110,135,110,136,110,138,110,139,110,140,110,141,110,142,110,145,110,146,110,147,110,148,110,149,110,150,110,151,110,153,110,154,110,155,110,157,110,158,110,160,110,161,110,163,110,164,110,166,110,168,110,169,110,171,110,172,110,173,110,174,110,176,110,179,110,181,110,184,110,185,110,188,110,190,110,191,110,192,110,195,110,196,110,197,110,198,110,200,110,201,110,202,110,204,110,205,110,206,110,208,110,210,110,214,110,216,110,217,110,219,110,220,110,221,110,227,110,231,110,234,110,235,110,236,110,237,110,238,110,239,110,240,110,241,110,242,110,243,110,245,110,246,110,247,110,248,110,250,110,251,110,252,110,253,110,254,110,255,110,0,111,1,111,3,111,4,111,5,111,7,111,8,111,10,111,11,111,12,111,13,111,14,111,16,111,17,111,18,111,22,111,23,111,24,111,25,111,26,111,27,111,28,111,29,111,30,111,31,111,33,111,34,111,35,111,37,111,38,111,39,111,40,111,44,111,46,111,48,111,50,111,52,111,53,111,55,111,56,111,57,111,58,111,59,111,60,111,61,111,63,111,64,111,65,111,66,111,67,111,68,111,69,111,72,111,73,111,74,111,76,111,78,111,79,111,80,111,81,111,82,111,83,111,84,111,85,111,86,111,87,111,89,111,90,111,91,111,93,111,95,111,96,111,97,111,99,111,100,111,101,111,103,111,104,111,105,111,106,111,107,111,108,111,111,111,112,111,113,111,115,111,117,111,118,111,119,111,121,111,123,111,125,111,126,111,127,111,128,111,129,111,130,111,131,111,133,111,134,111,135,111,138,111,139,111,143,111,144,111,145,111,146,111,147,111,148,111,149,111,150,111,151,111,152,111,153,111,154,111,155,111,157,111,158,111,159,111,160,111,162,111,163,111,164,111,165,111,166,111,168,111,169,111,170,111,171,111,172,111,173,111,174,111,175,111,176,111,177,111,178,111,180,111,181,111,183,111,184,111,186,111,187,111,188,111,189,111,190,111,191,111,193,111,195,111,196,111,197,111,198,111,199,111,200,111,202,111,203,111,204,111,205,111,206,111,207,111,208,111,211,111,212,111,213,111,214,111,215,111,216,111,217,111,218,111,219,111,220,111,221,111,223,111,226,111,227,111,228,111,229,111,230,111,231,111,232,111,233,111,234,111,235,111,236,111,237,111,240,111,241,111,242,111,243,111,244,111,245,111,246,111,247,111,248,111,249,111,250,111,251,111,252,111,253,111,254,111,255,111,0,112,1,112,2,112,3,112,4,112,5,112,6,112,7,112,8,112,9,112,10,112,11,112,12,112,13,112,14,112,15,112,16,112,18,112,19,112,20,112,21,112,22,112,23,112,24,112,25,112,28,112,29,112,30,112,31,112,32,112,33,112,34,112,36,112,37,112,38,112,39,112,40,112,41,112,42,112,43,112,44,112,45,112,46,112,47,112,48,112,49,112,50,112,51,112,52,112,54,112,55,112,56,112,58,112,59,112,60,112,61,112,62,112,63,112,64,112,65,112,66,112,67,112,68,112,69,112,70,112,71,112,72,112,73,112,74,112,75,112,77,112,78,112,80,112,81,112,82,112,83,112,84,112,85,112,86,112,87,112,88,112,89,112,90,112,91,112,92,112,93,112,95,112,96,112,97,112,98,112,99,112,100,112,101,112,102,112,103,112,104,112,105,112,106,112,110,112,113,112,114,112,115,112,116,112,119,112,121,112,122,112,123,112,125,112,129,112,130,112,131,112,132,112,134,112,135,112,136,112,139,112,140,112,141,112,143,112,144,112,145,112,147,112,151,112,152,112,154,112,155,112,158,112,159,112,160,112,161,112,162,112,163,112,164,112,165,112,166,112,167,112,168,112,169,112,170,112,176,112,178,112,180,112,181,112,182,112,186,112,190,112,191,112,196,112,197,112,198,112,199,112,201,112,203,112,204,112,205,112,206,112,207,112,208,112,209,112,210,112,211,112,212,112,213,112,214,112,215,112,218,112,220,112,221,112,222,112,224,112,225,112,226,112,227,112,229,112,234,112,238,112,240,112,241,112,242,112,243,112,244,112,245,112,246,112,248,112,250,112,251,112,252,112,254,112,255,112,0,113,1,113,2,113,3,113,4,113,5,113,6,113,7,113,8,113,11,113,12,113,13,113,14,113,15,113,17,113,18,113,20,113,23,113,27,113,28,113,29,113,30,113,31,113,32,113,33,113,34,113,35,113,36,113,37,113,39,113,40,113,41,113,42,113,43,113,44,113,45,113,46,113,50,113,51,113,52,113,53,113,55,113,56,113,57,113,58,113,59,113,60,113,61,113,62,113,63,113,64,113,65,113,66,113,67,113,68,113,70,113,71,113,72,113,73,113,75,113,77,113,79,113,80,113,81,113,82,113,83,113,84,113,85,113,86,113,87,113,88,113,89,113,90,113,91,113,93,113,95,113,96,113,97,113,98,113,99,113,101,113,105,113,106,113,107,113,108,113,109,113,111,113,112,113,113,113,116,113,117,113,118,113,119,113,121,113,123,113,124,113,126,113,127,113,128,113,129,113,130,113,131,113,133,113,134,113,135,113,136,113,137,113,139,113,140,113,141,113,142,113,144,113,145,113,146,113,147,113,149,113,150,113,151,113,154,113,155,113,156,113,157,113,158,113,161,113,162,113,163,113,164,113,165,113,166,113,167,113,169,113,170,113,171,113,173,113,174,113,175,113,176,113,177,113,178,113,180,113,182,113,183,113,184,113,186,113,187,113,188,113,189,113,190,113,191,113,192,113,193,113,194,113,196,113,197,113,198,113,199,113,200,113,201,113,202,113,203,113,204,113,205,113,207,113,208,113,209,113,210,113,211,113,214,113,215,113,216,113,217,113,218,113,219,113,220,113,221,113,222,113,223,113,225,113,226,113,227,113,228,113,230,113,232,113,233,113,234,113,235,113,236,113,237,113,239,113,240,113,241,113,242,113,243,113,244,113,245,113,246,113,247,113,248,113,250,113,251,113,252,113,253,113,254,113,255,113,0,114,1,114,2,114,3,114,4,114,5,114,7,114,8,114,9,114,10,114,11,114,12,114,13,114,14,114,15,114,16,114,17,114,18,114,19,114,20,114,21,114,22,114,23,114,24,114,25,114,26,114,27,114,28,114,30,114,31,114,32,114,33,114,34,114,35,114,36,114,37,114,38,114,39,114,41,114,43,114,45,114,46,114,47,114,50,114,51,114,52,114,58,114,60,114,62,114,64,114,65,114,66,114,67,114,68,114,69,114,70,114,73,114,74,114,75,114,78,114,79,114,80,114,81,114,83,114,84,114,85,114,87,114,88,114,90,114,92,114,94,114,96,114,99,114,100,114,101,114,104,114,106,114,107,114,108,114,109,114,112,114,113,114,115,114,116,114,118,114,119,114,120,114,123,114,124,114,125,114,130,114,131,114,133,114,134,114,135,114,136,114,137,114,140,114,142,114,144,114,145,114,147,114,148,114,149,114,150,114,151,114,152,114,153,114,154,114,155,114,156,114,157,114,158,114,160,114,161,114,162,114,163,114,164,114,165,114,166,114,167,114,168,114,169,114,170,114,171,114,174,114,177,114,178,114,179,114,181,114,186,114,187,114,188,114,189,114,190,114,191,114,192,114,197,114,198,114,199,114,201,114,202,114,203,114,204,114,207,114,209,114,211,114,212,114,213,114,214,114,216,114,218,114,219,114,198,228,199,228,200,228,201,228,202,228,203,228,204,228,205,228,206,228,207,228,208,228,209,228,210,228,211,228,212,228,213,228,214,228,215,228,216,228,217,228,218,228,219,228,220,228,221,228,222,228,223,228,224,228,225,228,226,228,227,228,228,228,229,228,230,228,231,228,232,228,233,228,234,228,235,228,236,228,237,228,238,228,239,228,240,228,241,228,242,228,243,228,244,228,245,228,246,228,247,228,248,228,249,228,250,228,251,228,252,228,253,228,254,228,255,228,0,229,1,229,2,229,3,229,4,229,5,229,6,229,7,229,8,229,9,229,10,229,11,229,12,229,13,229,14,229,15,229,16,229,17,229,18,229,19,229,20,229,21,229,22,229,23,229,24,229,25,229,26,229,27,229,28,229,29,229,30,229,31,229,32,229,33,229,34,229,35,229,36,229,37,229,0,48,1,48,2,48,183,0,201,2,199,2,168,0,3,48,5,48,20,32,94,255,22,32,38,32,24,32,25,32,28,32,29,32,20,48,21,48,8,48,9,48,10,48,11,48,12,48,13,48,14,48,15,48,22,48,23,48,16,48,17,48,177,0,215,0,247,0,54,34,39,34,40,34,17,34,15,34,42,34,41,34,8,34,55,34,26,34,165,34,37,34,32,34,18,35,153,34,43,34,46,34,97,34,76,34,72,34,61,34,29,34,96,34,110,34,111,34,100,34,101,34,30,34,53,34,52,34,66,38,64,38,176,0,50,32,51,32,3,33,4,255,164,0,224,255,225,255,48,32,167,0,22,33,6,38,5,38,203,37,207,37,206,37,199,37,198,37,161,37,160,37,179,37,178,37,59,32,146,33,144,33,145,33,147,33,19,48,38,229,39,229,40,229,41,229,42,229,43,229,44,229,45,229,46,229,47,229,48,229,49,229,50,229,51,229,52,229,53,229,54,229,55,229,56,229,57,229,58,229,59,229,60,229,61,229,62,229,63,229,64,229,65,229,66,229,67,229,68,229,69,229,70,229,71,229,72,229,73,229,74,229,75,229,76,229,77,229,78,229,79,229,80,229,81,229,82,229,83,229,84,229,85,229,86,229,87,229,88,229,89,229,90,229,91,229,92,229,93,229,94,229,95,229,96,229,97,229,98,229,99,229,100,229,101,229,102,229,103,229,104,229,105,229,106,229,107,229,108,229,109,229,110,229,111,229,112,229,113,229,114,229,115,229,116,229,117,229,118,229,119,229,120,229,121,229,122,229,123,229,124,229,125,229,126,229,127,229,128,229,129,229,130,229,131,229,132,229,133,229,112,33,113,33,114,33,115,33,116,33,117,33,118,33,119,33,120,33,121,33,102,231,103,231,104,231,105,231,106,231,107,231,136,36,137,36,138,36,139,36,140,36,141,36,142,36,143,36,144,36,145,36,146,36,147,36,148,36,149,36,150,36,151,36,152,36,153,36,154,36,155,36,116,36,117,36,118,36,119,36,120,36,121,36,122,36,123,36,124,36,125,36,126,36,127,36,128,36,129,36,130,36,131,36,132,36,133,36,134,36,135,36,96,36,97,36,98,36,99,36,100,36,101,36,102,36,103,36,104,36,105,36,172,32,109,231,32,50,33,50,34,50,35,50,36,50,37,50,38,50,39,50,40,50,41,50,110,231,111,231,96,33,97,33,98,33,99,33,100,33,101,33,102,33,103,33,104,33,105,33,106,33,107,33,112,231,113,231,134,229,135,229,136,229,137,229,138,229,139,229,140,229,141,229,142,229,143,229,144,229,145,229,146,229,147,229,148,229,149,229,150,229,151,229,152,229,153,229,154,229,155,229,156,229,157,229,158,229,159,229,160,229,161,229,162,229,163,229,164,229,165,229,166,229,167,229,168,229,169,229,170,229,171,229,172,229,173,229,174,229,175,229,176,229,177,229,178,229,179,229,180,229,181,229,182,229,183,229,184,229,185,229,186,229,187,229,188,229,189,229,190,229,191,229,192,229,193,229,194,229,195,229,196,229,197,229,198,229,199,229,200,229,201,229,202,229,203,229,204,229,205,229,206,229,207,229,208,229,209,229,210,229,211,229,212,229,213,229,214,229,215,229,216,229,217,229,218,229,219,229,220,229,221,229,222,229,223,229,224,229,225,229,226,229,227,229,228,229,229,229,1,255,2,255,3,255,229,255,5,255,6,255,7,255,8,255,9,255,10,255,11,255,12,255,13,255,14,255,15,255,16,255,17,255,18,255,19,255,20,255,21,255,22,255,23,255,24,255,25,255,26,255,27,255,28,255,29,255,30,255,31,255,32,255,33,255,34,255,35,255,36,255,37,255,38,255,39,255,40,255,41,255,42,255,43,255,44,255,45,255,46,255,47,255,48,255,49,255,50,255,51,255,52,255,53,255,54,255,55,255,56,255,57,255,58,255,59,255,60,255,61,255,62,255,63,255,64,255,65,255,66,255,67,255,68,255,69,255,70,255,71,255,72,255,73,255,74,255,75,255,76,255,77,255,78,255,79,255,80,255,81,255,82,255,83,255,84,255,85,255,86,255,87,255,88,255,89,255,90,255,91,255,92,255,93,255,227,255,230,229,231,229,232,229,233,229,234,229,235,229,236,229,237,229,238,229,239,229,240,229,241,229,242,229,243,229,244,229,245,229,246,229,247,229,248,229,249,229,250,229,251,229,252,229,253,229,254,229,255,229,0,230,1,230,2,230,3,230,4,230,5,230,6,230,7,230,8,230,9,230,10,230,11,230,12,230,13,230,14,230,15,230,16,230,17,230,18,230,19,230,20,230,21,230,22,230,23,230,24,230,25,230,26,230,27,230,28,230,29,230,30,230,31,230,32,230,33,230,34,230,35,230,36,230,37,230,38,230,39,230,40,230,41,230,42,230,43,230,44,230,45,230,46,230,47,230,48,230,49,230,50,230,51,230,52,230,53,230,54,230,55,230,56,230,57,230,58,230,59,230,60,230,61,230,62,230,63,230,64,230,65,230,66,230,67,230,68,230,69,230,65,48,66,48,67,48,68,48,69,48,70,48,71,48,72,48,73,48,74,48,75,48,76,48,77,48,78,48,79,48,80,48,81,48,82,48,83,48,84,48,85,48,86,48,87,48,88,48,89,48,90,48,91,48,92,48,93,48,94,48,95,48,96,48,97,48,98,48,99,48,100,48,101,48,102,48,103,48,104,48,105,48,106,48,107,48,108,48,109,48,110,48,111,48,112,48,113,48,114,48,115,48,116,48,117,48,118,48,119,48,120,48,121,48,122,48,123,48,124,48,125,48,126,48,127,48,128,48,129,48,130,48,131,48,132,48,133,48,134,48,135,48,136,48,137,48,138,48,139,48,140,48,141,48,142,48,143,48,144,48,145,48,146,48,147,48,114,231,115,231,116,231,117,231,118,231,119,231,120,231,121,231,122,231,123,231,124,231,70,230,71,230,72,230,73,230,74,230,75,230,76,230,77,230,78,230,79,230,80,230,81,230,82,230,83,230,84,230,85,230,86,230,87,230,88,230,89,230,90,230,91,230,92,230,93,230,94,230,95,230,96,230,97,230,98,230,99,230,100,230,101,230,102,230,103,230,104,230,105,230,106,230,107,230,108,230,109,230,110,230,111,230,112,230,113,230,114,230,115,230,116,230,117,230,118,230,119,230,120,230,121,230,122,230,123,230,124,230,125,230,126,230,127,230,128,230,129,230,130,230,131,230,132,230,133,230,134,230,135,230,136,230,137,230,138,230,139,230,140,230,141,230,142,230,143,230,144,230,145,230,146,230,147,230,148,230,149,230,150,230,151,230,152,230,153,230,154,230,155,230,156,230,157,230,158,230,159,230,160,230,161,230,162,230,163,230,164,230,165,230,161,48,162,48,163,48,164,48,165,48,166,48,167,48,168,48,169,48,170,48,171,48,172,48,173,48,174,48,175,48,176,48,177,48,178,48,179,48,180,48,181,48,182,48,183,48,184,48,185,48,186,48,187,48,188,48,189,48,190,48,191,48,192,48,193,48,194,48,195,48,196,48,197,48,198,48,199,48,200,48,201,48,202,48,203,48,204,48,205,48,206,48,207,48,208,48,209,48,210,48,211,48,212,48,213,48,214,48,215,48,216,48,217,48,218,48,219,48,220,48,221,48,222,48,223,48,224,48,225,48,226,48,227,48,228,48,229,48,230,48,231,48,232,48,233,48,234,48,235,48,236,48,237,48,238,48,239,48,240,48,241,48,242,48,243,48,244,48,245,48,246,48,125,231,126,231,127,231,128,231,129,231,130,231,131,231,132,231,166,230,167,230,168,230,169,230,170,230,171,230,172,230,173,230,174,230,175,230,176,230,177,230,178,230,179,230,180,230,181,230,182,230,183,230,184,230,185,230,186,230,187,230,188,230,189,230,190,230,191,230,192,230,193,230,194,230,195,230,196,230,197,230,198,230,199,230,200,230,201,230,202,230,203,230,204,230,205,230,206,230,207,230,208,230,209,230,210,230,211,230,212,230,213,230,214,230,215,230,216,230,217,230,218,230,219,230,220,230,221,230,222,230,223,230,224,230,225,230,226,230,227,230,228,230,229,230,230,230,231,230,232,230,233,230,234,230,235,230,236,230,237,230,238,230,239,230,240,230,241,230,242,230,243,230,244,230,245,230,246,230,247,230,248,230,249,230,250,230,251,230,252,230,253,230,254,230,255,230,0,231,1,231,2,231,3,231,4,231,5,231,145,3,146,3,147,3,148,3,149,3,150,3,151,3,152,3,153,3,154,3,155,3,156,3,157,3,158,3,159,3,160,3,161,3,163,3,164,3,165,3,166,3,167,3,168,3,169,3,133,231,134,231,135,231,136,231,137,231,138,231,139,231,140,231,177,3,178,3,179,3,180,3,181,3,182,3,183,3,184,3,185,3,186,3,187,3,188,3,189,3,190,3,191,3,192,3,193,3,195,3,196,3,197,3,198,3,199,3,200,3,201,3,141,231,142,231,143,231,144,231,145,231,146,231,147,231,53,254,54,254,57,254,58,254,63,254,64,254,61,254,62,254,65,254,66,254,67,254,68,254,148,231,149,231,59,254,60,254,55,254,56,254,49,254,150,231,51,254,52,254,151,231,152,231,153,231,154,231,155,231,156,231,157,231,158,231,159,231,6,231,7,231,8,231,9,231,10,231,11,231,12,231,13,231,14,231,15,231,16,231,17,231,18,231,19,231,20,231,21,231,22,231,23,231,24,231,25,231,26,231,27,231,28,231,29,231,30,231,31,231,32,231,33,231,34,231,35,231,36,231,37,231,38,231,39,231,40,231,41,231,42,231,43,231,44,231,45,231,46,231,47,231,48,231,49,231,50,231,51,231,52,231,53,231,54,231,55,231,56,231,57,231,58,231,59,231,60,231,61,231,62,231,63,231,64,231,65,231,66,231,67,231,68,231,69,231,70,231,71,231,72,231,73,231,74,231,75,231,76,231,77,231,78,231,79,231,80,231,81,231,82,231,83,231,84,231,85,231,86,231,87,231,88,231,89,231,90,231,91,231,92,231,93,231,94,231,95,231,96,231,97,231,98,231,99,231,100,231,101,231,16,4,17,4,18,4,19,4,20,4,21,4,1,4,22,4,23,4,24,4,25,4,26,4,27,4,28,4,29,4,30,4,31,4,32,4,33,4,34,4,35,4,36,4,37,4,38,4,39,4,40,4,41,4,42,4,43,4,44,4,45,4,46,4,47,4,160,231,161,231,162,231,163,231,164,231,165,231,166,231,167,231,168,231,169,231,170,231,171,231,172,231,173,231,174,231,48,4,49,4,50,4,51,4,52,4,53,4,81,4,54,4,55,4,56,4,57,4,58,4,59,4,60,4,61,4,62,4,63,4,64,4,65,4,66,4,67,4,68,4,69,4,70,4,71,4,72,4,73,4,74,4,75,4,76,4,77,4,78,4,79,4,175,231,176,231,177,231,178,231,179,231,180,231,181,231,182,231,183,231,184,231,185,231,186,231,187,231,202,2,203,2,217,2,19,32,21,32,37,32,53,32,5,33,9,33,150,33,151,33,152,33,153,33,21,34,31,34,35,34,82,34,102,34,103,34,191,34,80,37,81,37,82,37,83,37,84,37,85,37,86,37,87,37,88,37,89,37,90,37,91,37,92,37,93,37,94,37,95,37,96,37,97,37,98,37,99,37,100,37,101,37,102,37,103,37,104,37,105,37,106,37,107,37,108,37,109,37,110,37,111,37,112,37,113,37,114,37,115,37,129,37,130,37,131,37,132,37,133,37,134,37,135,37,136,37,137,37,138,37,139,37,140,37,141,37,142,37,143,37,147,37,148,37,149,37,188,37,189,37,226,37,227,37,228,37,229,37,9,38,149,34,18,48,29,48,30,48,188,231,189,231,190,231,191,231,192,231,193,231,194,231,195,231,196,231,197,231,198,231,1,1,225,0,206,1,224,0,19,1,233,0,27,1,232,0,43,1,237,0,208,1,236,0,77,1,243,0,210,1,242,0,107,1,250,0,212,1,249,0,214,1,216,1,218,1,220,1,252,0,234,0,81,2,199,231,68,1,72,1,249,1,97,2,201,231,202,231,203,231,204,231,5,49,6,49,7,49,8,49,9,49,10,49,11,49,12,49,13,49,14,49,15,49,16,49,17,49,18,49,19,49,20,49,21,49,22,49,23,49,24,49,25,49,26,49,27,49,28,49,29,49,30,49,31,49,32,49,33,49,34,49,35,49,36,49,37,49,38,49,39,49,40,49,41,49,205,231,206,231,207,231,208,231,209,231,210,231,211,231,212,231,213,231,214,231,215,231,216,231,217,231,218,231,219,231,220,231,221,231,222,231,223,231,224,231,225,231,33,48,34,48,35,48,36,48,37,48,38,48,39,48,40,48,41,48,163,50,142,51,143,51,156,51,157,51,158,51,161,51,196,51,206,51,209,51,210,51,213,51,48,254,226,255,228,255,226,231,33,33,49,50,227,231,16,32,228,231,229,231,230,231,252,48,155,48,156,48,253,48,254,48,6,48,157,48,158,48,73,254,74,254,75,254,76,254,77,254,78,254,79,254,80,254,81,254,82,254,84,254,85,254,86,254,87,254,89,254,90,254,91,254,92,254,93,254,94,254,95,254,96,254,97,254,98,254,99,254,100,254,101,254,102,254,104,254,105,254,106,254,107,254,62,48,240,47,241,47,242,47,243,47,244,47,245,47,246,47,247,47,248,47,249,47,250,47,251,47,7,48,244,231,245,231,246,231,247,231,248,231,249,231,250,231,251,231,252,231,253,231,254,231,255,231,0,232,0,37,1,37,2,37,3,37,4,37,5,37,6,37,7,37,8,37,9,37,10,37,11,37,12,37,13,37,14,37,15,37,16,37,17,37,18,37,19,37,20,37,21,37,22,37,23,37,24,37,25,37,26,37,27,37,28,37,29,37,30,37,31,37,32,37,33,37,34,37,35,37,36,37,37,37,38,37,39,37,40,37,41,37,42,37,43,37,44,37,45,37,46,37,47,37,48,37,49,37,50,37,51,37,52,37,53,37,54,37,55,37,56,37,57,37,58,37,59,37,60,37,61,37,62,37,63,37,64,37,65,37,66,37,67,37,68,37,69,37,70,37,71,37,72,37,73,37,74,37,75,37,1,232,2,232,3,232,4,232,5,232,6,232,7,232,8,232,9,232,10,232,11,232,12,232,13,232,14,232,15,232,220,114,221,114,223,114,226,114,227,114,228,114,229,114,230,114,231,114,234,114,235,114,245,114,246,114,249,114,253,114,254,114,255,114,0,115,2,115,4,115,5,115,6,115,7,115,8,115,9,115,11,115,12,115,13,115,15,115,16,115,17,115,18,115,20,115,24,115,25,115,26,115,31,115,32,115,35,115,36,115,38,115,39,115,40,115,45,115,47,115,48,115,50,115,51,115,53,115,54,115,58,115,59,115,60,115,61,115,64,115,65,115,66,115,67,115,68,115,69,115,70,115,71,115,72,115,73,115,74,115,75,115,76,115,78,115,79,115,81,115,83,115,84,115,85,115,86,115,88,115,89,115,90,115,91,115,92,115,93,115,94,115,95,115,97,115,98,115,99,115,100,115,101,115,102,115,103,115,104,115,105,115,106,115,107,115,110,115,112,115,113,115,0,224,1,224,2,224,3,224,4,224,5,224,6,224,7,224,8,224,9,224,10,224,11,224,12,224,13,224,14,224,15,224,16,224,17,224,18,224,19,224,20,224,21,224,22,224,23,224,24,224,25,224,26,224,27,224,28,224,29,224,30,224,31,224,32,224,33,224,34,224,35,224,36,224,37,224,38,224,39,224,40,224,41,224,42,224,43,224,44,224,45,224,46,224,47,224,48,224,49,224,50,224,51,224,52,224,53,224,54,224,55,224,56,224,57,224,58,224,59,224,60,224,61,224,62,224,63,224,64,224,65,224,66,224,67,224,68,224,69,224,70,224,71,224,72,224,73,224,74,224,75,224,76,224,77,224,78,224,79,224,80,224,81,224,82,224,83,224,84,224,85,224,86,224,87,224,88,224,89,224,90,224,91,224,92,224,93,224,114,115,115,115,116,115,117,115,118,115,119,115,120,115,121,115,122,115,123,115,124,115,125,115,127,115,128,115,129,115,130,115,131,115,133,115,134,115,136,115,138,115,140,115,141,115,143,115,144,115,146,115,147,115,148,115,149,115,151,115,152,115,153,115,154,115,156,115,157,115,158,115,160,115,161,115,163,115,164,115,165,115,166,115,167,115,168,115,170,115,172,115,173,115,177,115,180,115,181,115,182,115,184,115,185,115,188,115,189,115,190,115,191,115,193,115,195,115,196,115,197,115,198,115,199,115,203,115,204,115,206,115,210,115,211,115,212,115,213,115,214,115,215,115,216,115,218,115,219,115,220,115,221,115,223,115,225,115,226,115,227,115,228,115,230,115,232,115,234,115,235,115,236,115,238,115,239,115,240,115,241,115,243,115,244,115,245,115,246,115,247,115,94,224,95,224,96,224,97,224,98,224,99,224,100,224,101,224,102,224,103,224,104,224,105,224,106,224,107,224,108,224,109,224,110,224,111,224,112,224,113,224,114,224,115,224,116,224,117,224,118,224,119,224,120,224,121,224,122,224,123,224,124,224,125,224,126,224,127,224,128,224,129,224,130,224,131,224,132,224,133,224,134,224,135,224,136,224,137,224,138,224,139,224,140,224,141,224,142,224,143,224,144,224,145,224,146,224,147,224,148,224,149,224,150,224,151,224,152,224,153,224,154,224,155,224,156,224,157,224,158,224,159,224,160,224,161,224,162,224,163,224,164,224,165,224,166,224,167,224,168,224,169,224,170,224,171,224,172,224,173,224,174,224,175,224,176,224,177,224,178,224,179,224,180,224,181,224,182,224,183,224,184,224,185,224,186,224,187,224,248,115,249,115,250,115,251,115,252,115,253,115,254,115,255,115,0,116,1,116,2,116,4,116,7,116,8,116,11,116,12,116,13,116,14,116,17,116,18,116,19,116,20,116,21,116,22,116,23,116,24,116,25,116,28,116,29,116,30,116,31,116,32,116,33,116,35,116,36,116,39,116,41,116,43,116,45,116,47,116,49,116,50,116,55,116,56,116,57,116,58,116,59,116,61,116,62,116,63,116,64,116,66,116,67,116,68,116,69,116,70,116,71,116,72,116,73,116,74,116,75,116,76,116,77,116,78,116,79,116,80,116,81,116,82,116,83,116,84,116,86,116,88,116,93,116,96,116,97,116,98,116,99,116,100,116,101,116,102,116,103,116,104,116,105,116,106,116,107,116,108,116,110,116,111,116,113,116,114,116,115,116,116,116,117,116,120,116,121,116,122,116,188,224,189,224,190,224,191,224,192,224,193,224,194,224,195,224,196,224,197,224,198,224,199,224,200,224,201,224,202,224,203,224,204,224,205,224,206,224,207,224,208,224,209,224,210,224,211,224,212,224,213,224,214,224,215,224,216,224,217,224,218,224,219,224,220,224,221,224,222,224,223,224,224,224,225,224,226,224,227,224,228,224,229,224,230,224,231,224,232,224,233,224,234,224,235,224,236,224,237,224,238,224,239,224,240,224,241,224,242,224,243,224,244,224,245,224,246,224,247,224,248,224,249,224,250,224,251,224,252,224,253,224,254,224,255,224,0,225,1,225,2,225,3,225,4,225,5,225,6,225,7,225,8,225,9,225,10,225,11,225,12,225,13,225,14,225,15,225,16,225,17,225,18,225,19,225,20,225,21,225,22,225,23,225,24,225,25,225,123,116,124,116,125,116,127,116,130,116,132,116,133,116,134,116,136,116,137,116,138,116,140,116,141,116,143,116,145,116,146,116,147,116,148,116,149,116,150,116,151,116,152,116,153,116,154,116,155,116,157,116,159,116,160,116,161,116,162,116,163,116,164,116,165,116,166,116,170,116,171,116,172,116,173,116,174,116,175,116,176,116,177,116,178,116,179,116,180,116,181,116,182,116,183,116,184,116,185,116,187,116,188,116,189,116,190,116,191,116,192,116,193,116,194,116,195,116,196,116,197,116,198,116,199,116,200,116,201,116,202,116,203,116,204,116,205,116,206,116,207,116,208,116,209,116,211,116,212,116,213,116,214,116,215,116,216,116,217,116,218,116,219,116,221,116,223,116,225,116,229,116,231,116,232,116,233,116,234,116,235,116,236,116,237,116,240,116,241,116,242,116,26,225,27,225,28,225,29,225,30,225,31,225,32,225,33,225,34,225,35,225,36,225,37,225,38,225,39,225,40,225,41,225,42,225,43,225,44,225,45,225,46,225,47,225,48,225,49,225,50,225,51,225,52,225,53,225,54,225,55,225,56,225,57,225,58,225,59,225,60,225,61,225,62,225,63,225,64,225,65,225,66,225,67,225,68,225,69,225,70,225,71,225,72,225,73,225,74,225,75,225,76,225,77,225,78,225,79,225,80,225,81,225,82,225,83,225,84,225,85,225,86,225,87,225,88,225,89,225,90,225,91,225,92,225,93,225,94,225,95,225,96,225,97,225,98,225,99,225,100,225,101,225,102,225,103,225,104,225,105,225,106,225,107,225,108,225,109,225,110,225,111,225,112,225,113,225,114,225,115,225,116,225,117,225,118,225,119,225,243,116,245,116,248,116,249,116,250,116,251,116,252,116,253,116,254,116,0,117,1,117,2,117,3,117,5,117,6,117,7,117,8,117,9,117,10,117,11,117,12,117,14,117,16,117,18,117,20,117,21,117,22,117,23,117,27,117,29,117,30,117,32,117,33,117,34,117,35,117,36,117,38,117,39,117,42,117,46,117,52,117,54,117,57,117,60,117,61,117,63,117,65,117,66,117,67,117,68,117,70,117,71,117,73,117,74,117,77,117,80,117,81,117,82,117,83,117,85,117,86,117,87,117,88,117,93,117,94,117,95,117,96,117,97,117,98,117,99,117,100,117,103,117,104,117,105,117,107,117,108,117,109,117,110,117,111,117,112,117,113,117,115,117,117,117,118,117,119,117,122,117,123,117,124,117,125,117,126,117,128,117,129,117,130,117,132,117,133,117,135,117,120,225,121,225,122,225,123,225,124,225,125,225,126,225,127,225,128,225,129,225,130,225,131,225,132,225,133,225,134,225,135,225,136,225,137,225,138,225,139,225,140,225,141,225,142,225,143,225,144,225,145,225,146,225,147,225,148,225,149,225,150,225,151,225,152,225,153,225,154,225,155,225,156,225,157,225,158,225,159,225,160,225,161,225,162,225,163,225,164,225,165,225,166,225,167,225,168,225,169,225,170,225,171,225,172,225,173,225,174,225,175,225,176,225,177,225,178,225,179,225,180,225,181,225,182,225,183,225,184,225,185,225,186,225,187,225,188,225,189,225,190,225,191,225,192,225,193,225,194,225,195,225,196,225,197,225,198,225,199,225,200,225,201,225,202,225,203,225,204,225,205,225,206,225,207,225,208,225,209,225,210,225,211,225,212,225,213,225,136,117,137,117,138,117,140,117,141,117,142,117,144,117,147,117,149,117,152,117,155,117,156,117,158,117,162,117,166,117,167,117,168,117,169,117,170,117,173,117,182,117,183,117,186,117,187,117,191,117,192,117,193,117,198,117,203,117,204,117,206,117,207,117,208,117,209,117,211,117,215,117,217,117,218,117,220,117,221,117,223,117,224,117,225,117,229,117,233,117,236,117,237,117,238,117,239,117,242,117,243,117,245,117,246,117,247,117,248,117,250,117,251,117,253,117,254,117,2,118,4,118,6,118,7,118,8,118,9,118,11,118,13,118,14,118,15,118,17,118,18,118,19,118,20,118,22,118,26,118,28,118,29,118,30,118,33,118,35,118,39,118,40,118,44,118,46,118,47,118,49,118,50,118,54,118,55,118,57,118,58,118,59,118,61,118,65,118,66,118,68,118,214,225,215,225,216,225,217,225,218,225,219,225,220,225,221,225,222,225,223,225,224,225,225,225,226,225,227,225,228,225,229,225,230,225,231,225,232,225,233,225,234,225,235,225,236,225,237,225,238,225,239,225,240,225,241,225,242,225,243,225,244,225,245,225,246,225,247,225,248,225,249,225,250,225,251,225], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+46328);
/* memory initializer */ allocate([252,225,253,225,254,225,255,225,0,226,1,226,2,226,3,226,4,226,5,226,6,226,7,226,8,226,9,226,10,226,11,226,12,226,13,226,14,226,15,226,16,226,17,226,18,226,19,226,20,226,21,226,22,226,23,226,24,226,25,226,26,226,27,226,28,226,29,226,30,226,31,226,32,226,33,226,34,226,35,226,36,226,37,226,38,226,39,226,40,226,41,226,42,226,43,226,44,226,45,226,46,226,47,226,48,226,49,226,50,226,51,226,69,118,70,118,71,118,72,118,73,118,74,118,75,118,78,118,79,118,80,118,81,118,82,118,83,118,85,118,87,118,88,118,89,118,90,118,91,118,93,118,95,118,96,118,97,118,98,118,100,118,101,118,102,118,103,118,104,118,105,118,106,118,108,118,109,118,110,118,112,118,113,118,114,118,115,118,116,118,117,118,118,118,119,118,121,118,122,118,124,118,127,118,128,118,129,118,131,118,133,118,137,118,138,118,140,118,141,118,143,118,144,118,146,118,148,118,149,118,151,118,152,118,154,118,155,118,156,118,157,118,158,118,159,118,160,118,161,118,162,118,163,118,165,118,166,118,167,118,168,118,169,118,170,118,171,118,172,118,173,118,175,118,176,118,179,118,181,118,182,118,183,118,184,118,185,118,186,118,187,118,188,118,189,118,190,118,192,118,193,118,195,118,74,85,63,150,195,87,40,99,206,84,9,85,192,84,145,118,76,118,60,133,238,119,126,130,141,120,49,114,152,150,141,151,40,108,137,91,250,79,9,99,151,102,184,92,250,128,72,104,174,128,2,102,206,118,249,81,86,101,172,113,241,127,132,136,178,80,101,89,202,97,179,111,173,130,76,99,82,98,237,83,39,84,6,123,107,81,164,117,244,93,212,98,203,141,118,151,138,98,25,128,93,87,56,151,98,127,56,114,125,118,207,103,126,118,70,100,112,79,37,141,220,98,23,122,145,101,237,115,44,100,115,98,44,130,129,152,127,103,72,114,110,98,204,98,52,79,227,116,74,83,158,82,202,126,166,144,46,94,134,104,156,105,128,129,209,126,210,104,197,120,140,134,81,149,141,80,36,140,222,130,222,128,5,83,18,137,101,82,196,118,199,118,201,118,203,118,204,118,211,118,213,118,217,118,218,118,220,118,221,118,222,118,224,118,225,118,226,118,227,118,228,118,230,118,231,118,232,118,233,118,234,118,235,118,236,118,237,118,240,118,243,118,245,118,246,118,247,118,250,118,251,118,253,118,255,118,0,119,2,119,3,119,5,119,6,119,10,119,12,119,14,119,15,119,16,119,17,119,18,119,19,119,20,119,21,119,22,119,23,119,24,119,27,119,28,119,29,119,30,119,33,119,35,119,36,119,37,119,39,119,42,119,43,119,44,119,46,119,48,119,49,119,50,119,51,119,52,119,57,119,59,119,61,119,62,119,63,119,66,119,68,119,69,119,70,119,72,119,73,119,74,119,75,119,76,119,77,119,78,119,79,119,82,119,83,119,84,119,85,119,86,119,87,119,88,119,89,119,92,119,132,133,249,150,221,79,33,88,113,153,157,91,177,98,165,98,180,102,121,140,141,156,6,114,111,103,145,120,178,96,81,83,23,83,136,143,204,128,29,141,161,148,13,80,200,114,7,89,235,96,25,113,171,136,84,89,239,130,44,103,40,123,41,93,247,126,45,117,245,108,102,142,248,143,60,144,59,159,212,107,25,145,20,123,124,95,167,120,214,132,61,133,213,107,217,107,214,107,1,94,135,94,249,117,237,149,93,101,10,95,197,95,159,143,193,88,194,129,127,144,91,150,173,151,185,143,22,127,44,141,65,98,191,79,216,83,94,83,168,143,169,143,171,143,77,144,7,104,106,95,152,129,104,136,214,156,139,97,43,82,42,118,108,95,140,101,210,111,232,110,190,91,72,100,117,81,176,81,196,103,25,78,201,121,124,153,179,112,93,119,94,119,95,119,96,119,100,119,103,119,105,119,106,119,109,119,110,119,111,119,112,119,113,119,114,119,115,119,116,119,117,119,118,119,119,119,120,119,122,119,123,119,124,119,129,119,130,119,131,119,134,119,135,119,136,119,137,119,138,119,139,119,143,119,144,119,147,119,148,119,149,119,150,119,151,119,152,119,153,119,154,119,155,119,156,119,157,119,158,119,161,119,163,119,164,119,166,119,168,119,171,119,173,119,174,119,175,119,177,119,178,119,180,119,182,119,183,119,184,119,185,119,186,119,188,119,190,119,192,119,193,119,194,119,195,119,196,119,197,119,198,119,199,119,200,119,201,119,202,119,203,119,204,119,206,119,207,119,208,119,209,119,210,119,211,119,212,119,213,119,214,119,216,119,217,119,218,119,221,119,222,119,223,119,224,119,225,119,228,119,197,117,118,94,187,115,224,131,173,100,232,98,181,148,226,108,90,83,195,82,15,100,194,148,148,123,47,79,27,94,54,130,22,129,138,129,36,110,202,108,115,154,85,99,92,83,250,84,101,136,224,87,13,78,3,94,101,107,63,124,232,144,22,96,230,100,28,115,193,136,80,103,77,98,34,141,108,119,41,142,199,145,105,95,220,131,33,133,16,153,194,83,149,134,139,107,237,96,232,96,127,112,205,130,49,130,211,78,167,108,207,133,205,100,217,124,253,105,249,102,73,131,149,83,86,123,167,79,140,81,75,109,66,92,109,142,210,99,201,83,44,131,54,131,229,103,180,120,61,100,223,91,148,92,238,93,231,139,198,98,244,103,122,140,0,100,186,99,73,135,139,153,23,140,32,127,242,148,167,78,16,150,164,152,12,102,22,115,230,119,232,119,234,119,239,119,240,119,241,119,242,119,244,119,245,119,247,119,249,119,250,119,251,119,252,119,3,120,4,120,5,120,6,120,7,120,8,120,10,120,11,120,14,120,15,120,16,120,19,120,21,120,25,120,27,120,30,120,32,120,33,120,34,120,36,120,40,120,42,120,43,120,46,120,47,120,49,120,50,120,51,120,53,120,54,120,61,120,63,120,65,120,66,120,67,120,68,120,70,120,72,120,73,120,74,120,75,120,77,120,79,120,81,120,83,120,84,120,88,120,89,120,90,120,91,120,92,120,94,120,95,120,96,120,97,120,98,120,99,120,100,120,101,120,102,120,103,120,104,120,105,120,111,120,112,120,113,120,114,120,115,120,116,120,117,120,118,120,120,120,121,120,122,120,123,120,125,120,126,120,127,120,128,120,129,120,130,120,131,120,58,87,29,92,56,94,127,149,127,80,160,128,130,83,94,101,69,117,49,85,33,80,133,141,132,98,158,148,29,103,50,86,110,111,226,93,53,84,146,112,102,143,111,98,164,100,163,99,123,95,136,111,244,144,227,129,176,143,24,92,104,102,241,95,137,108,72,150,129,141,108,136,145,100,240,121,206,87,89,106,16,98,72,84,88,78,11,122,233,96,132,111,218,139,127,98,30,144,139,154,228,121,3,84,244,117,1,99,25,83,96,108,223,143,27,95,112,154,59,128,127,159,136,79,58,92,100,141,197,127,165,101,189,112,69,81,178,81,107,134,7,93,160,91,189,98,108,145,116,117,12,142,32,122,1,97,121,123,199,78,248,126,133,119,17,78,237,129,29,82,250,81,113,106,168,83,135,142,4,149,207,150,193,110,100,150,90,105,132,120,133,120,134,120,136,120,138,120,139,120,143,120,144,120,146,120,148,120,149,120,150,120,153,120,157,120,158,120,160,120,162,120,164,120,166,120,168,120,169,120,170,120,171,120,172,120,173,120,174,120,175,120,181,120,182,120,183,120,184,120,186,120,187,120,188,120,189,120,191,120,192,120,194,120,195,120,196,120,198,120,199,120,200,120,204,120,205,120,206,120,207,120,209,120,210,120,211,120,214,120,215,120,216,120,218,120,219,120,220,120,221,120,222,120,223,120,224,120,225,120,226,120,227,120,228,120,229,120,230,120,231,120,233,120,234,120,235,120,237,120,238,120,239,120,240,120,241,120,243,120,245,120,246,120,248,120,249,120,251,120,252,120,253,120,254,120,255,120,0,121,2,121,3,121,4,121,6,121,7,121,8,121,9,121,10,121,11,121,12,121,64,120,168,80,215,119,16,100,230,137,4,89,227,99,221,93,127,122,61,105,32,79,57,130,152,85,50,78,174,117,151,122,98,94,138,94,239,149,27,82,57,84,138,112,118,99,36,149,130,87,37,102,63,105,135,145,7,85,243,109,175,126,34,136,51,98,240,126,181,117,40,131,193,120,204,150,158,143,72,97,247,116,205,139,100,107,58,82,80,141,33,107,106,128,113,132,241,86,6,83,206,78,27,78,209,81,151,124,139,145,7,124,195,79,127,142,225,123,156,122,103,100,20,93,172,80,6,129,1,118,185,124,236,109,224,127,81,103,88,91,248,91,203,120,174,100,19,100,170,99,43,99,25,149,45,100,190,143,84,123,41,118,83,98,39,89,70,84,121,107,163,80,52,98,38,94,134,107,227,78,55,141,139,136,133,95,46,144,13,121,14,121,15,121,16,121,17,121,18,121,20,121,21,121,22,121,23,121,24,121,25,121,26,121,27,121,28,121,29,121,31,121,32,121,33,121,34,121,35,121,37,121,38,121,39,121,40,121,41,121,42,121,43,121,44,121,45,121,46,121,47,121,48,121,49,121,50,121,51,121,53,121,54,121,55,121,56,121,57,121,61,121,63,121,66,121,67,121,68,121,69,121,71,121,74,121,75,121,76,121,77,121,78,121,79,121,80,121,81,121,82,121,84,121,85,121,88,121,89,121,97,121,99,121,100,121,102,121,105,121,106,121,107,121,108,121,110,121,112,121,113,121,114,121,115,121,116,121,117,121,118,121,121,121,123,121,124,121,125,121,126,121,127,121,130,121,131,121,134,121,135,121,136,121,137,121,139,121,140,121,141,121,142,121,144,121,145,121,146,121,32,96,61,128,197,98,57,78,85,83,248,144,184,99,198,128,230,101,46,108,70,79,238,96,225,109,222,139,57,95,203,134,83,95,33,99,90,81,97,131,99,104,0,82,99,99,72,142,18,80,155,92,119,121,252,91,48,82,59,122,188,96,83,144,215,118,183,95,151,95,132,118,108,142,111,112,123,118,73,123,170,119,243,81,147,144,36,88,78,79,244,110,234,143,76,101,27,123,196,114,164,109,223,127,225,90,181,98,149,94,48,87,130,132,44,123,29,94,31,95,18,144,20,127,160,152,130,99,199,110,152,120,185,112,120,81,91,151,171,87,53,117,67,79,56,117,151,94,230,96,96,89,192,109,191,107,137,120,252,83,213,150,203,81,1,82,137,99,10,84,147,148,3,140,204,141,57,114,159,120,118,135,237,143,13,140,224,83,147,121,148,121,149,121,150,121,151,121,152,121,153,121,155,121,156,121,157,121,158,121,159,121,160,121,161,121,162,121,163,121,164,121,165,121,166,121,168,121,169,121,170,121,171,121,172,121,173,121,174,121,175,121,176,121,177,121,178,121,180,121,181,121,182,121,183,121,184,121,188,121,191,121,194,121,196,121,197,121,199,121,200,121,202,121,204,121,206,121,207,121,208,121,211,121,212,121,214,121,215,121,217,121,218,121,219,121,220,121,221,121,222,121,224,121,225,121,226,121,229,121,232,121,234,121,236,121,238,121,241,121,242,121,243,121,244,121,245,121,246,121,247,121,249,121,250,121,252,121,254,121,255,121,1,122,4,122,5,122,7,122,8,122,9,122,10,122,12,122,15,122,16,122,17,122,18,122,19,122,21,122,22,122,24,122,25,122,27,122,28,122,1,78,239,118,238,83,137,148,118,152,14,159,45,149,154,91,162,139,34,78,28,78,172,81,99,132,194,97,168,82,11,104,151,79,107,96,187,81,30,109,92,81,150,98,151,101,97,150,70,140,23,144,216,117,253,144,99,119,210,107,138,114,236,114,251,139,53,88,121,119,76,141,92,103,64,149,154,128,166,94,33,110,146,89,239,122,237,119,59,149,181,107,173,101,14,127,6,88,81,81,31,150,249,91,169,88,40,84,114,142,102,101,127,152,228,86,157,148,254,118,65,144,135,99,198,84,26,89,58,89,155,87,178,142,53,103,250,141,53,130,65,82,240,96,21,88,254,134,232,92,69,158,196,79,157,152,185,139,37,90,118,96,132,83,124,98,79,144,2,145,127,153,105,96,12,128,63,81,51,128,20,92,117,153,49,109,140,78,29,122,31,122,33,122,34,122,36,122,37,122,38,122,39,122,40,122,41,122,42,122,43,122,44,122,45,122,46,122,47,122,48,122,49,122,50,122,52,122,53,122,54,122,56,122,58,122,62,122,64,122,65,122,66,122,67,122,68,122,69,122,71,122,72,122,73,122,74,122,75,122,76,122,77,122,78,122,79,122,80,122,82,122,83,122,84,122,85,122,86,122,88,122,89,122,90,122,91,122,92,122,93,122,94,122,95,122,96,122,97,122,98,122,99,122,100,122,101,122,102,122,103,122,104,122,105,122,106,122,107,122,108,122,109,122,110,122,111,122,113,122,114,122,115,122,117,122,123,122,124,122,125,122,126,122,130,122,133,122,135,122,137,122,138,122,139,122,140,122,142,122,143,122,144,122,147,122,148,122,153,122,154,122,155,122,158,122,161,122,162,122,48,141,209,83,90,127,79,123,16,79,79,78,0,150,213,108,208,115,233,133,6,94,106,117,251,127,10,106,254,119,146,148,65,126,225,81,230,112,205,83,212,143,3,131,41,141,175,114,109,153,219,108,74,87,179,130,185,101,170,128,63,98,50,150,168,89,255,78,191,139,186,126,62,101,242,131,94,151,97,85,222,152,165,128,42,83,253,139,32,84,186,128,159,94,184,108,57,141,172,130,90,145,41,84,27,108,6,82,183,126,95,87,26,113,126,108,137,124,75,89,253,78,255,95,36,97,170,124,48,78,1,92,171,103,2,135,240,92,11,149,206,152,175,117,253,112,34,144,175,81,29,127,189,139,73,89,228,81,91,79,38,84,43,89,119,101,164,128,117,91,118,98,194,98,144,143,69,94,31,108,38,123,15,79,216,79,13,103,163,122,164,122,167,122,169,122,170,122,171,122,174,122,175,122,176,122,177,122,178,122,180,122,181,122,182,122,183,122,184,122,185,122,186,122,187,122,188,122,189,122,190,122,192,122,193,122,194,122,195,122,196,122,197,122,198,122,199,122,200,122,201,122,202,122,204,122,205,122,206,122,207,122,208,122,209,122,210,122,211,122,212,122,213,122,215,122,216,122,218,122,219,122,220,122,221,122,225,122,226,122,228,122,231,122,232,122,233,122,234,122,235,122,236,122,238,122,240,122,241,122,242,122,243,122,244,122,245,122,246,122,247,122,248,122,251,122,252,122,254,122,0,123,1,123,2,123,5,123,7,123,9,123,12,123,13,123,14,123,16,123,18,123,19,123,22,123,23,123,24,123,26,123,28,123,29,123,31,123,33,123,34,123,35,123,39,123,41,123,45,123,110,109,170,109,143,121,177,136,23,95,43,117,154,98,133,143,239,79,220,145,167,101,47,129,81,129,156,94,80,129,116,141,111,82,134,137,75,141,13,89,133,80,216,78,28,150,54,114,121,129,31,141,204,91,163,139,68,150,135,89,26,127,144,84,118,86,14,86,229,139,57,101,130,105,153,148,214,118,137,110,114,94,24,117,70,103,209,103,255,122,157,128,118,141,31,97,198,121,98,101,99,141,136,81,26,82,162,148,56,127,155,128,178,126,151,92,47,110,96,103,217,123,139,118,216,154,143,129,148,127,213,124,30,100,80,149,63,122,74,84,229,84,76,107,1,100,8,98,61,158,243,128,153,117,114,82,105,151,91,132,60,104,228,134,1,150,148,150,236,148,42,78,4,84,217,126,57,104,223,141,21,128,244,102,154,94,185,127,47,123,48,123,50,123,52,123,53,123,54,123,55,123,57,123,59,123,61,123,63,123,64,123,65,123,66,123,67,123,68,123,70,123,72,123,74,123,77,123,78,123,83,123,85,123,87,123,89,123,92,123,94,123,95,123,97,123,99,123,100,123,101,123,102,123,103,123,104,123,105,123,106,123,107,123,108,123,109,123,111,123,112,123,115,123,116,123,118,123,120,123,122,123,124,123,125,123,127,123,129,123,130,123,131,123,132,123,134,123,135,123,136,123,137,123,138,123,139,123,140,123,142,123,143,123,145,123,146,123,147,123,150,123,152,123,153,123,154,123,155,123,158,123,159,123,160,123,163,123,164,123,165,123,174,123,175,123,176,123,178,123,179,123,181,123,182,123,183,123,185,123,186,123,187,123,188,123,189,123,190,123,191,123,192,123,194,123,195,123,196,123,194,87,63,128,151,104,229,93,59,101,159,82,109,96,154,159,155,79,172,142,108,81,171,91,19,95,233,93,94,108,241,98,33,141,113,81,169,148,254,82,159,108,223,130,215,114,162,87,132,103,45,141,31,89,156,143,199,131,149,84,141,123,48,79,189,108,100,91,209,89,19,159,228,83,202,134,168,154,55,140,161,128,69,101,126,152,250,86,199,150,46,82,220,116,80,82,225,91,2,99,2,137,86,78,208,98,42,96,250,104,115,81,152,91,160,81,194,137,161,123,134,153,80,127,239,96,76,112,47,141,73,81,127,94,27,144,112,116,196,137,45,87,69,120,82,95,159,159,250,149,104,143,60,155,225,139,120,118,66,104,220,103,234,141,53,141,61,82,138,143,218,110,205,104,5,149,237,144,253,86,156,103,249,136,199,143,200,84,197,123,200,123,201,123,202,123,203,123,205,123,206,123,207,123,208,123,210,123,212,123,213,123,214,123,215,123,216,123,219,123,220,123,222,123,223,123,224,123,226,123,227,123,228,123,231,123,232,123,233,123,235,123,236,123,237,123,239,123,240,123,242,123,243,123,244,123,245,123,246,123,248,123,249,123,250,123,251,123,253,123,255,123,0,124,1,124,2,124,3,124,4,124,5,124,6,124,8,124,9,124,10,124,13,124,14,124,16,124,17,124,18,124,19,124,20,124,21,124,23,124,24,124,25,124,26,124,27,124,28,124,29,124,30,124,32,124,33,124,34,124,35,124,36,124,37,124,40,124,41,124,43,124,44,124,45,124,46,124,47,124,48,124,49,124,50,124,51,124,52,124,53,124,54,124,55,124,57,124,58,124,59,124,60,124,61,124,62,124,66,124,184,154,105,91,119,109,38,108,165,78,179,91,135,154,99,145,168,97,175,144,233,151,43,84,181,109,210,91,253,81,138,85,85,127,240,127,188,100,77,99,241,101,190,97,141,96,10,113,87,108,73,108,47,89,109,103,42,130,213,88,142,86,106,140,235,107,221,144,125,89,23,128,247,83,105,109,117,84,157,85,119,131,207,131,56,104,190,121,140,84,85,79,8,84,210,118,137,140,2,150,179,108,184,109,107,141,16,137,100,158,58,141,63,86,209,158,213,117,136,95,224,114,104,96,252,84,168,78,42,106,97,136,82,96,112,143,196,84,216,112,121,134,63,158,42,109,143,91,24,95,162,126,137,85,175,79,52,115,60,84,154,83,25,80,14,84,124,84,78,78,253,95,90,116,246,88,107,132,225,128,116,135,208,114,202,124,86,110,67,124,68,124,69,124,70,124,71,124,72,124,73,124,74,124,75,124,76,124,78,124,79,124,80,124,81,124,82,124,83,124,84,124,85,124,86,124,87,124,88,124,89,124,90,124,91,124,92,124,93,124,94,124,95,124,96,124,97,124,98,124,99,124,100,124,101,124,102,124,103,124,104,124,105,124,106,124,107,124,108,124,109,124,110,124,111,124,112,124,113,124,114,124,117,124,118,124,119,124,120,124,121,124,122,124,126,124,127,124,128,124,129,124,130,124,131,124,132,124,133,124,134,124,135,124,136,124,138,124,139,124,140,124,141,124,142,124,143,124,144,124,147,124,148,124,150,124,153,124,154,124,155,124,160,124,161,124,163,124,166,124,167,124,168,124,169,124,171,124,172,124,173,124,175,124,176,124,180,124,181,124,182,124,183,124,184,124,186,124,187,124,39,95,78,134,44,85,164,98,146,78,170,108,55,98,177,130,215,84,78,83,62,115,209,110,59,117,18,82,22,83,221,139,208,105,138,95,0,96,238,109,79,87,34,107,175,115,83,104,216,143,19,127,98,99,163,96,36,85,234,117,98,140,21,113,163,109,166,91,123,94,82,131,76,97,196,158,250,120,87,135,39,124,135,118,240,81,246,96,76,113,67,102,76,94,77,96,14,140,112,112,37,99,137,143,189,95,98,96,212,134,222,86,193,107,148,96,103,97,73,83,224,96,102,102,63,141,253,121,26,79,233,112,71,108,179,139,242,139,216,126,100,131,15,102,90,90,66,155,81,109,247,109,65,140,59,109,25,79,107,112,183,131,22,98,209,96,13,151,39,141,120,121,251,81,62,87,250,87,58,103,120,117,61,122,239,121,149,123,191,124,192,124,194,124,195,124,196,124,198,124,201,124,203,124,206,124,207,124,208,124,209,124,210,124,211,124,212,124,216,124,218,124,219,124,221,124,222,124,225,124,226,124,227,124,228,124,229,124,230,124,231,124,233,124,234,124,235,124,236,124,237,124,238,124,240,124,241,124,242,124,243,124,244,124,245,124,246,124,247,124,249,124,250,124,252,124,253,124,254,124,255,124,0,125,1,125,2,125,3,125,4,125,5,125,6,125,7,125,8,125,9,125,11,125,12,125,13,125,14,125,15,125,16,125,17,125,18,125,19,125,20,125,21,125,22,125,23,125,24,125,25,125,26,125,27,125,28,125,29,125,30,125,31,125,33,125,35,125,36,125,37,125,38,125,40,125,41,125,42,125,44,125,45,125,46,125,48,125,49,125,50,125,51,125,52,125,53,125,54,125,140,128,101,153,249,143,192,111,165,139,33,158,236,89,233,126,9,127,9,84,129,103,216,104,145,143,77,124,198,150,202,83,37,96,190,117,114,108,115,83,201,90,167,126,36,99,224,81,10,129,241,93,223,132,128,98,128,81,99,91,14,79,109,121,66,82,184,96,78,109,196,91,194,91,161,139,176,139,226,101,204,95,69,150,147,89,231,126,170,126,9,86,183,103,57,89,115,79,182,91,160,82,90,131,138,152,62,141,50,117,190,148,71,80,60,122,247,78,182,103,126,154,193,90,124,107,209,118,90,87,22,92,58,123,244,149,78,113,124,81,169,128,112,130,120,89,4,127,39,131,192,104,236,103,177,120,119,120,227,98,97,99,128,123,237,79,106,82,207,81,80,131,219,105,116,146,245,141,49,141,193,137,46,149,173,123,246,78,55,125,56,125,57,125,58,125,59,125,60,125,61,125,62,125,63,125,64,125,65,125,66,125,67,125,68,125,69,125,70,125,71,125,72,125,73,125,74,125,75,125,76,125,77,125,78,125,79,125,80,125,81,125,82,125,83,125,84,125,85,125,86,125,87,125,88,125,89,125,90,125,91,125,92,125,93,125,94,125,95,125,96,125,97,125,98,125,99,125,100,125,101,125,102,125,103,125,104,125,105,125,106,125,107,125,108,125,109,125,111,125,112,125,113,125,114,125,115,125,116,125,117,125,118,125,120,125,121,125,122,125,123,125,124,125,125,125,126,125,127,125,128,125,129,125,130,125,131,125,132,125,133,125,134,125,135,125,136,125,137,125,138,125,139,125,140,125,141,125,142,125,143,125,144,125,145,125,146,125,147,125,148,125,149,125,150,125,151,125,152,125,101,80,48,130,81,82,111,153,16,110,133,110,167,109,250,94,245,80,220,89,6,92,70,109,95,108,134,117,139,132,104,104,86,89,178,139,32,83,113,145,77,150,73,133,18,105,1,121,38,113,246,128,164,78,202,144,71,109,132,154,7,90,188,86,5,100,240,148,235,119,165,79,26,129,225,114,210,137,122,153,52,127,222,126,127,82,89,101,117,145,127,143,131,143,235,83,150,122,237,99,165,99,134,118,248,121,87,136,54,150,42,98,171,82,130,130,84,104,112,103,119,99,107,119,237,122,1,109,211,126,227,137,208,89,18,98,201,133,165,130,76,117,31,80,203,78,165,117,235,139,74,92,254,93,75,123,164,101,209,145,202,78,37,109,95,137,39,125,38,149,197,78,40,140,219,143,115,151,75,102,129,121,209,143,236,112,120,109,153,125,154,125,155,125,156,125,157,125,158,125,159,125,160,125,161,125,162,125,163,125,164,125,165,125,167,125,168,125,169,125,170,125,171,125,172,125,173,125,175,125,176,125,177,125,178,125,179,125,180,125,181,125,182,125,183,125,184,125,185,125,186,125,187,125,188,125,189,125,190,125,191,125,192,125,193,125,194,125,195,125,196,125,197,125,198,125,199,125,200,125,201,125,202,125,203,125,204,125,205,125,206,125,207,125,208,125,209,125,210,125,211,125,212,125,213,125,214,125,215,125,216,125,217,125,218,125,219,125,220,125,221,125,222,125,223,125,224,125,225,125,226,125,227,125,228,125,229,125,230,125,231,125,232,125,233,125,234,125,235,125,236,125,237,125,238,125,239,125,240,125,241,125,242,125,243,125,244,125,245,125,246,125,247,125,248,125,249,125,250,125,61,92,178,82,70,131,98,81,14,131,91,119,118,102,184,156,172,78,202,96,190,124,179,124,207,126,149,78,102,139,111,102,136,152,89,151,131,88,108,101,92,149,132,95,201,117,86,151,223,122,222,122,192,81,175,112,152,122,234,99,118,122,160,126,150,115,237,151,69,78,120,112,93,78,82,145,169,83,81,101,231,101,252,129,5,130,142,84,49,92,154,117,160,151,216,98,217,114,189,117,69,92,121,154,202,131,64,92,128,84,233,119,62,78,174,108,90,128,210,98,110,99,232,93,119,81,221,141,30,142,47,149,241,79,229,83,231,96,172,112,103,82,80,99,67,158,31,90,38,80,55,119,119,83,226,126,133,100,43,101,137,98,152,99,20,80,53,114,201,137,179,81,192,139,221,126,71,87,204,131,167,148,155,81,27,84,251,92,251,125,252,125,253,125,254,125,255,125,0,126,1,126,2,126,3,126,4,126,5,126,6,126,7,126,8,126,9,126,10,126,11,126,12,126,13,126,14,126,15,126,16,126,17,126,18,126,19,126,20,126,21,126,22,126,23,126,24,126,25,126,26,126,27,126,28,126,29,126,30,126,31,126,32,126,33,126,34,126,35,126,36,126,37,126,38,126,39,126,40,126,41,126,42,126,43,126,44,126,45,126,46,126,47,126,48,126,49,126,50,126,51,126,52,126,53,126,54,126,55,126,56,126,57,126,58,126,60,126,61,126,62,126,63,126,64,126,66,126,67,126,68,126,69,126,70,126,72,126,73,126,74,126,75,126,76,126,77,126,78,126,79,126,80,126,81,126,82,126,83,126,84,126,85,126,86,126,87,126,88,126,89,126,90,126,91,126,92,126,93,126,202,79,227,122,90,109,225,144,143,154,128,85,150,84,97,83,175,84,0,95,233,99,119,105,239,81,104,97,10,82,42,88,216,82,78,87,13,120,11,119,183,94,119,97,224,124,91,98,151,98,162,78,149,112,3,128,247,98,228,112,96,151,119,87,219,130,239,103,245,104,213,120,151,152,209,121,243,88,179,84,239,83,52,110,75,81,59,82,162,91,254,139,175,128,67,85,166,87,115,96,81,87,45,84,122,122,80,96,84,91,167,99,160,98,227,83,99,98,199,91,175,103,237,84,159,122,230,130,119,145,147,94,228,136,56,89,174,87,14,99,232,141,239,128,87,87,119,123,169,79,235,95,189,91,62,107,33,83,80,123,194,114,70,104,255,119,54,119,247,101,181,81,143,78,212,118,191,92,165,122,117,132,78,89,65,155,128,80,94,126,95,126,96,126,97,126,98,126,99,126,100,126,101,126,102,126,103,126,104,126,105,126,106,126,107,126,108,126,109,126,110,126,111,126,112,126,113,126,114,126,115,126,116,126,117,126,118,126,119,126,120,126,121,126,122,126,123,126,124,126,125,126,126,126,127,126,128,126,129,126,131,126,132,126,133,126,134,126,135,126,136,126,137,126,138,126,139,126,140,126,141,126,142,126,143,126,144,126,145,126,146,126,147,126,148,126,149,126,150,126,151,126,152,126,153,126,154,126,156,126,157,126,158,126,174,126,180,126,187,126,188,126,214,126,228,126,236,126,249,126,10,127,16,127,30,127,55,127,57,127,59,127,60,127,61,127,62,127,63,127,64,127,65,127,67,127,70,127,71,127,72,127,73,127,74,127,75,127,76,127,77,127,78,127,79,127,82,127,83,127,136,153,39,97,131,110,100,87,6,102,70,99,240,86,236,98,105,98,211,94,20,150,131,87,201,98,135,85,33,135,74,129,163,143,102,85,177,131,101,103,86,141,221,132,106,90,15,104,230,98,238,123,17,150,112,81,156,111,48,140,253,99,200,137,210,97,6,127,194,112,229,110,5,116,148,105,252,114,202,94,206,144,23,103,106,109,94,99,179,82,98,114,1,128,108,79,229,89,106,145,217,112,157,109,210,82,80,78,247,150,109,149,126,133,202,120,47,125,33,81,146,87,194,100,139,128,123,124,234,108,241,104,94,105,183,81,152,83,168,104,129,114,206,158,241,123,248,114,187,121,19,111,6,116,78,103,204,145,164,156,60,121,137,131,84,131,15,84,23,104,61,78,137,83,177,82,62,120,134,83,41,82,136,80,139,79,208,79,86,127,89,127,91,127,92,127,93,127,94,127,96,127,99,127,100,127,101,127,102,127,103,127,107,127,108,127,109,127,111,127,112,127,115,127,117,127,118,127,119,127,120,127,122,127,123,127,124,127,125,127,127,127,128,127,130,127,131,127,132,127,133,127,134,127,135,127,136,127,137,127,139,127,141,127,143,127,144,127,145,127,146,127,147,127,149,127,150,127,151,127,152,127,153,127,155,127,156,127,160,127,162,127,163,127,165,127,166,127,168,127,169,127,170,127,171,127,172,127,173,127,174,127,177,127,179,127,180,127,181,127,182,127,183,127,186,127,187,127,190,127,192,127,194,127,195,127,196,127,198,127,199,127,200,127,201,127,203,127,205,127,207,127,208,127,209,127,210,127,211,127,214,127,215,127,217,127,218,127,219,127,220,127,221,127,222,127,226,127,227,127,226,117,203,122,146,124,165,108,182,150,155,82,131,116,233,84,233,79,84,128,178,131,222,143,112,149,201,94,28,96,159,109,24,94,91,101,56,129,254,148,75,96,188,112,195,126,174,124,201,81,129,104,177,124,111,130,36,78,134,143,207,145,126,102,174,78,5,140,169,100,74,128,218,80,151,117,206,113,229,91,189,143,102,111,134,78,130,100,99,149,214,94,153,101,23,82,194,136,200,112,163,82,14,115,51,116,151,103,247,120,22,151,52,78,187,144,222,156,203,109,219,81,65,141,29,84,206,98,178,115,241,131,246,150,132,159,195,148,54,79,154,127,204,81,117,112,117,150,173,92,134,152,230,83,228,78,156,110,9,116,180,105,107,120,143,153,89,117,24,82,36,118,65,109,243,103,109,81,153,159,75,128,153,84,60,123,191,122,228,127,231,127,232,127,234,127,235,127,236,127,237,127,239,127,242,127,244,127,245,127,246,127,247,127,248,127,249,127,250,127,253,127,254,127,255,127,2,128,7,128,8,128,9,128,10,128,14,128,15,128,17,128,19,128,26,128,27,128,29,128,30,128,31,128,33,128,35,128,36,128,43,128,44,128,45,128,46,128,47,128,48,128,50,128,52,128,57,128,58,128,60,128,62,128,64,128,65,128,68,128,69,128,71,128,72,128,73,128,78,128,79,128,80,128,81,128,83,128,85,128,86,128,87,128,89,128,91,128,92,128,93,128,94,128,95,128,96,128,97,128,98,128,99,128,100,128,101,128,102,128,103,128,104,128,107,128,108,128,109,128,110,128,111,128,112,128,114,128,115,128,116,128,117,128,118,128,119,128,120,128,121,128,122,128,123,128,124,128,125,128,134,150,132,87,226,98,71,150,124,105,4,90,2,100,211,123,15,111,75,150,166,130,98,83,133,152,144,94,137,112,179,99,100,83,79,134,129,156,147,158,140,120,50,151,239,141,66,141,127,158,94,111,132,121,85,95,70,150,46,98,116,154,21,84,221,148,163,79,197,101,101,92,97,92,21,127,81,134,47,108,139,95,135,115,228,110,255,126,230,92,27,99,106,91,230,110,117,83,113,78,160,99,101,117,161,98,110,143,38,79,209,78,166,108,182,126,186,139,29,132,186,135,87,127,59,144,35,149,169,123,161,154,248,136,61,132,27,109,134,154,220,126,136,89,187,158,155,115,1,120,130,134,108,154,130,154,27,86,23,84,203,87,112,78,166,158,86,83,200,143,9,129,146,119,146,153,238,134,225,110,19,133,252,102,98,97,43,111,126,128,129,128,130,128,133,128,136,128,138,128,141,128,142,128,143,128,144,128,145,128,146,128,148,128,149,128,151,128,153,128,158,128,163,128,166,128,167,128,168,128,172,128,176,128,179,128,181,128,182,128,184,128,185,128,187,128,197,128,199,128,200,128,201,128,202,128,203,128,207,128,208,128,209,128,210,128,211,128,212,128,213,128,216,128,223,128,224,128,226,128,227,128,230,128,238,128,245,128,247,128,249,128,251,128,254,128,255,128,0,129,1,129,3,129,4,129,5,129,7,129,8,129,11,129,12,129,21,129,23,129,25,129,27,129,28,129,29,129,31,129,32,129,33,129,34,129,35,129,36,129,37,129,38,129,39,129,40,129,41,129,42,129,43,129,45,129,46,129,48,129,51,129,52,129,53,129,55,129,57,129,58,129,59,129,60,129,61,129,63,129,41,140,146,130,43,131,242,118,19,108,217,95,189,131,43,115,5,131,26,149,219,107,219,119,198,148,111,83,2,131,146,81,61,94,140,140,56,141,72,78,171,115,154,103,133,104,118,145,9,151,100,113,161,108,9,119,146,90,65,149,207,107,142,127,39,102,208,91,185,89,154,90,232,149,247,149,236,78,12,132,153,132,172,106,223,118,48,149,27,115,166,104,95,91,47,119,154,145,97,151,220,124,247,143,28,140,37,95,115,124,216,121,197,137,204,108,28,135,198,91,66,94,201,104,32,119,245,126,149,81,77,81,201,82,41,90,5,127,98,151,215,130,207,99,132,119,208,133,210,121,58,110,153,94,153,89,17,133,109,112,17,108,191,98,191,118,79,101,175,96,253,149,14,102,159,135,35,158,237,148,13,84,125,84,44,140,120,100,64,129,65,129,66,129,67,129,68,129,69,129,71,129,73,129,77,129,78,129,79,129,82,129,86,129,87,129,88,129,91,129,92,129,93,129,94,129,95,129,97,129,98,129,99,129,100,129,102,129,104,129,106,129,107,129,108,129,111,129,114,129,115,129,117,129,118,129,119,129,120,129,129,129,131,129,132,129,133,129,134,129,135,129,137,129,139,129,140,129,141,129,142,129,144,129,146,129,147,129,148,129,149,129,150,129,151,129,153,129,154,129,158,129,159,129,160,129,161,129,162,129,164,129,165,129,167,129,169,129,171,129,172,129,173,129,174,129,175,129,176,129,177,129,178,129,180,129,181,129,182,129,183,129,184,129,185,129,188,129,189,129,190,129,191,129,196,129,197,129,199,129,200,129,201,129,203,129,205,129,206,129,207,129,208,129,209,129,210,129,211,129,121,100,17,134,33,106,156,129,232,120,105,100,84,155,185,98,43,103,171,131,168,88,216,158,171,108,32,111,222,91,76,150,11,140,95,114,208,103,199,98,97,114,169,78,198,89,205,107,147,88,174,102,85,94,223,82,85,97,40,103,238,118,102,119,103,114,70,122,255,98,234,84,80,84,160,148,163,144,28,90,179,126,22,108,67,78,118,89,16,128,72,89,87,83,55,117,190,150,202,86,32,99,17,129,124,96,249,149,214,109,98,84,129,153,133,81,233,90,253,128,174,89,19,151,42,80,229,108,60,92,223,98,96,79,63,83,123,129,6,144,186,110,43,133,200,98,116,94,190,120,181,100,123,99,245,95,24,90,127,145,31,158,63,92,79,99,66,128,125,91,110,85,74,149,77,149,133,109,168,96,224,103,222,114,221,81,129,91,212,129,213,129,214,129,215,129,216,129,217,129,218,129,219,129,220,129,221,129,222,129,223,129,224,129,225,129,226,129,228,129,229,129,230,129,232,129,233,129,235,129,238,129,239,129,240,129,241,129,242,129,245,129,246,129,247,129,248,129,249,129,250,129,253,129,255,129,3,130,7,130,8,130,9,130,10,130,11,130,14,130,15,130,17,130,19,130,21,130,22,130,23,130,24,130,25,130,26,130,29,130,32,130,36,130,37,130,38,130,39,130,41,130,46,130,50,130,58,130,60,130,61,130,63,130,64,130,65,130,66,130,67,130,69,130,70,130,72,130,74,130,76,130,77,130,78,130,80,130,81,130,82,130,83,130,84,130,85,130,86,130,87,130,89,130,91,130,92,130,93,130,94,130,96,130,97,130,98,130,99,130,100,130,101,130,102,130,103,130,105,130,231,98,222,108,91,114,109,98,174,148,189,126,19,129,83,109,156,81,4,95,116,89,170,82,18,96,115,89,150,102,80,134,159,117,42,99,230,97,239,124,250,139,230,84,39,107,37,158,180,107,213,133,85,84,118,80,164,108,106,85,180,141,44,114,21,94,21,96,54,116,205,98,146,99,76,114,152,95,67,110,62,109,0,101,88,111,216,118,208,120,252,118,84,117,36,82,219,83,83,78,158,94,193,101,42,128,214,128,155,98,134,84,40,82,174,112,141,136,209,141,225,108,120,84,218,128,249,87,244,136,84,141,106,150,77,145,105,79,155,108,183,85,198,118,48,120,168,98,249,112,142,111,109,95,236,132,218,104,124,120,247,123,168,129,11,103,79,158,103,99,176,120,111,87,18,120,57,151,121,98,171,98,136,82,53,116,215,107,106,130,107,130,108,130,109,130,113,130,117,130,118,130,119,130,120,130,123,130,124,130,128,130,129,130,131,130,133,130,134,130,135,130,137,130,140,130,144,130,147,130,148,130,149,130,150,130,154,130,155,130,158,130,160,130,162,130,163,130,167,130,178,130,181,130,182,130,186,130,187,130,188,130,191,130,192,130,194,130,195,130,197,130,198,130,201,130,208,130,214,130,217,130,218,130,221,130,226,130,231,130,232,130,233,130,234,130,236,130,237,130,238,130,240,130,242,130,243,130,245,130,246,130,248,130,250,130,252,130,253,130,254,130,255,130,0,131,10,131,11,131,13,131,16,131,18,131,19,131,22,131,24,131,25,131,29,131,30,131,31,131,32,131,33,131,34,131,35,131,36,131,37,131,38,131,41,131,42,131,46,131,48,131,50,131,55,131,59,131,61,131,100,85,62,129,178,117,174,118,57,83,222,117,251,80,65,92,108,139,199,123,79,80,71,114,151,154,216,152,2,111,226,116,104,121,135,100,165,119,252,98,145,152,43,141,193,84,88,128,82,78,106,87,249,130,13,132,115,94,237,81,246,116,196,139,79,92,97,87,252,108,135,152,70,90,52,120,68,155,235,143,149,124,86,82,81,98,250,148,198,78,134,131,97,132,233,131,178,132,212,87,52,103,3,87,110,102,102,109,49,140,221,102,17,112,31,103,58,107,22,104,26,98,187,89,3,78,196,81,6,111,210,103,143,108,118,81,203,104,71,89,103,107,102,117,14,93,16,129,80,159,215,101,72,121,65,121,145,154,119,141,130,92,94,78,1,79,47,84,81,89,12,120,104,86,20,108,196,143,3,95,125,108,227,108,171,139,144,99,62,131,63,131,65,131,66,131,68,131,69,131,72,131,74,131,75,131,76,131,77,131,78,131,83,131,85,131,86,131,87,131,88,131,89,131,93,131,98,131,112,131,113,131,114,131,115,131,116,131,117,131,118,131,121,131,122,131,126,131,127,131,128,131,129,131,130,131,131,131,132,131,135,131,136,131,138,131,139,131,140,131,141,131,143,131,144,131,145,131,148,131,149,131,150,131,151,131,153,131,154,131,157,131,159,131,161,131,162,131,163,131,164,131,165,131,166,131,167,131,172,131,173,131,174,131,175,131,181,131,187,131,190,131,191,131,194,131,195,131,196,131,198,131,200,131,201,131,203,131,205,131,206,131,208,131,209,131,210,131,211,131,213,131,215,131,217,131,218,131,219,131,222,131,226,131,227,131,228,131,230,131,231,131,232,131,235,131,236,131,237,131,112,96,61,109,117,114,102,98,142,148,197,148,67,83,193,143,126,123,223,78,38,140,126,78,212,158,177,148,179,148,77,82,92,111,99,144,69,109,52,140,17,88,76,93,32,107,73,107,170,103,91,84,84,129,140,127,153,88,55,133,58,95,162,98,71,106,57,149,114,101,132,96,101,104,167,119,84,78,168,79,231,93,152,151,172,100,216,127,237,92,207,79,141,122,7,82,4,131,20,78,47,96,131,122,166,148,181,79,178,78,230,121,52,116,228,82,185,130,210,100,189,121,221,91,129,108,82,151,123,143,34,108,62,80,127,83,5,110,206,100,116,102,48,108,197,96,119,152,247,139,134,94,60,116,119,122,203,121,24,78,177,144,3,116,66,108,218,86,75,145,197,108,139,141,58,83,198,134,242,102,175,142,72,92,113,154,32,110,238,131,239,131,243,131,244,131,245,131,246,131,247,131,250,131,251,131,252,131,254,131,255,131,0,132,2,132,5,132,7,132,8,132,9,132,10,132,16,132,18,132,19,132,20,132,21,132,22,132,23,132,25,132,26,132,27,132,30,132,31,132,32,132,33,132,34,132,35,132,41,132,42,132,43,132,44,132,45,132,46,132,47,132,48,132,50,132,51,132,52,132,53,132,54,132,55,132,57,132,58,132,59,132,62,132,63,132,64,132,65,132,66,132,67,132,68,132,69,132,71,132,72,132,73,132,74,132,75,132,76,132,77,132,78,132,79,132,80,132,82,132,83,132,84,132,85,132,86,132,88,132,93,132,94,132,95,132,96,132,98,132,100,132,101,132,102,132,103,132,104,132,106,132,110,132,111,132,112,132,114,132,116,132,119,132,121,132,123,132,124,132,214,83,54,90,139,159,163,141,187,83,8,87,167,152,67,103,155,145,201,108,104,81,202,117,243,98,172,114,56,82,157,82,58,127,148,112,56,118,116,83,74,158,183,105,110,120,192,150,217,136,164,127,54,113,195,113,137,81,211,103,228,116,228,88,24,101,183,86,169,139,118,153,112,98,213,126,249,96,237,112,236,88,193,78,186,78,205,95,231,151,251,78,164,139,3,82,138,89,171,126,84,98,205,78,229,101,14,98,56,131,201,132,99,131,141,135,148,113,182,110,185,91,210,126,151,81,201,99,212,103,137,128,57,131,21,136,18,81,122,91,130,89,177,143,115,78,93,108,101,81,37,137,111,143,46,150,74,133,94,116,16,149,240,149,166,109,229,130,49,95,146,100,18,109,40,132,110,129,195,156,94,88,91,141,9,78,193,83,125,132,126,132,127,132,128,132,129,132,131,132,132,132,133,132,134,132,138,132,141,132,143,132,144,132,145,132,146,132,147,132,148,132,149,132,150,132,152,132,154,132,155,132,157,132,158,132,159,132,160,132,162,132,163,132,164,132,165,132,166,132,167,132,168,132,169,132,170,132,171,132,172,132,173,132,174,132,176,132,177,132,179,132,181,132,182,132,183,132,187,132,188,132,190,132,192,132,194,132,195,132,197,132,198,132,199,132,200,132,203,132,204,132,206,132,207,132,210,132,212,132,213,132,215,132,216,132,217,132,218,132,219,132,220,132,222,132,225,132,226,132,228,132,231,132,232,132,233,132,234,132,235,132,237,132,238,132,239,132,241,132,242,132,243,132,244,132,245,132,246,132,247,132,248,132,249,132,250,132,251,132,253,132,254,132,0,133,1,133,2,133,30,79,99,101,81,104,211,85,39,78,20,100,154,154,107,98,194,90,95,116,114,130,169,109,238,104,231,80,142,131,2,120,64,103,57,82,153,108,177,126,187,80,101,85,94,113,91,123,82,102,202,115,235,130,73,103,113,92,32,82,125,113,107,136,234,149,85,150,197,100,97,141,179,129,132,85,85,108,71,98,46,127,146,88,36,79,70,85,79,141,76,102,10,78,26,92,243,136,162,104,78,99,13,122,231,112,141,130,250,82,246,151,17,92,232,84,181,144,205,126,98,89,74,141,199,134,12,130,13,130,102,141,68,100,4,92,81,97,137,109,62,121,190,139,55,120,51,117,123,84,56,79,171,142,241,109,32,90,197,126,94,121,136,108,161,91,118,90,26,117,190,128,78,97,23,110,240,88,31,117,37,117,114,114,71,83,243,126,3,133,4,133,5,133,6,133,7,133,8,133,9,133,10,133,11,133,13,133,14,133,15,133,16,133,18,133,20,133,21,133,22,133,24,133,25,133,27,133,28,133,29,133,30,133,32,133,34,133,35,133,36,133,37,133,38,133,39,133,40,133,41,133,42,133,45,133,46,133,47,133,48,133,49,133,50,133,51,133,52,133,53,133,54,133,62,133,63,133,64,133,65,133,66,133,68,133,69,133,70,133,71,133,75,133,76,133,77,133,78,133,79,133,80,133,81,133,82,133,83,133,84,133,85,133,87,133,88,133,90,133,91,133,92,133,93,133,95,133,96,133,97,133,98,133,99,133,101,133,102,133,103,133,105,133,106,133,107,133,108,133,109,133,110,133,111,133,112,133,113,133,115,133,117,133,118,133,119,133,120,133,124,133,125,133,127,133,128,133,129,133,1,119,219,118,105,82,220,128,35,87,8,94,49,89,238,114,189,101,127,110,215,139,56,92,113,134,65,83,243,119,254,98,246,101,192,78,223,152,128,134,158,91,198,139,242,83,226,119,127,79,78,92,118,154,203,89], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+56568);
/* memory initializer */ allocate([15,95,58,121,235,88,22,78,255,103,139,78,237,98,147,138,29,144,191,82,47,102,220,85,108,86,2,144,213,78,141,79,202,145,112,153,15,108,2,94,67,96,164,91,198,137,213,139,54,101,75,98,150,153,136,91,255,91,136,99,46,85,215,83,38,118,125,81,44,133,162,103,179,104,138,107,146,98,147,143,212,83,18,130,209,109,143,117,102,78,78,141,112,91,159,113,175,133,145,102,217,102,114,127,0,135,205,158,32,159,94,92,47,103,240,143,17,104,95,103,13,98,214,122,133,88,182,94,112,101,49,111,130,133,131,133,134,133,136,133,137,133,138,133,139,133,140,133,141,133,142,133,144,133,145,133,146,133,147,133,148,133,149,133,150,133,151,133,152,133,153,133,154,133,157,133,158,133,159,133,160,133,161,133,162,133,163,133,165,133,166,133,167,133,169,133,171,133,172,133,173,133,177,133,178,133,179,133,180,133,181,133,182,133,184,133,186,133,187,133,188,133,189,133,190,133,191,133,192,133,194,133,195,133,196,133,197,133,198,133,199,133,200,133,202,133,203,133,204,133,205,133,206,133,209,133,210,133,212,133,214,133,215,133,216,133,217,133,218,133,219,133,221,133,222,133,223,133,224,133,225,133,226,133,227,133,229,133,230,133,231,133,232,133,234,133,235,133,236,133,237,133,238,133,239,133,240,133,241,133,242,133,243,133,244,133,245,133,246,133,247,133,248,133,85,96,55,82,13,128,84,100,112,136,41,117,5,94,19,104,244,98,28,151,204,83,61,114,1,140,52,108,97,119,14,122,46,84,172,119,122,152,28,130,244,139,85,120,20,103,193,112,175,101,149,100,54,86,29,96,193,121,248,83,29,78,123,107,134,128,250,91,227,85,219,86,58,79,60,79,114,153,243,93,126,103,56,128,2,96,130,152,1,144,139,91,188,139,245,139,28,100,88,130,222,100,253,85,207,130,101,145,215,79,32,125,31,144,159,124,243,80,81,88,175,110,191,91,201,139,131,128,120,145,156,132,151,123,125,134,139,150,143,150,229,126,211,154,142,120,129,92,87,122,66,144,167,150,95,121,89,91,95,99,11,123,209,132,173,104,6,85,41,127,16,116,34,125,1,149,64,98,76,88,214,78,131,91,121,89,84,88,249,133,250,133,252,133,253,133,254,133,0,134,1,134,2,134,3,134,4,134,6,134,7,134,8,134,9,134,10,134,11,134,12,134,13,134,14,134,15,134,16,134,18,134,19,134,20,134,21,134,23,134,24,134,25,134,26,134,27,134,28,134,29,134,30,134,31,134,32,134,33,134,34,134,35,134,36,134,37,134,38,134,40,134,42,134,43,134,44,134,45,134,46,134,47,134,48,134,49,134,50,134,51,134,52,134,53,134,54,134,55,134,57,134,58,134,59,134,61,134,62,134,63,134,64,134,65,134,66,134,67,134,68,134,69,134,70,134,71,134,72,134,73,134,74,134,75,134,76,134,82,134,83,134,85,134,86,134,87,134,88,134,89,134,91,134,92,134,93,134,95,134,96,134,97,134,99,134,100,134,101,134,102,134,103,134,104,134,105,134,106,134,109,115,30,99,75,142,15,142,206,128,212,130,172,98,240,83,240,108,94,145,42,89,1,96,112,108,77,87,74,100,42,141,43,118,233,110,91,87,128,106,240,117,109,111,45,140,8,140,102,87,239,107,146,136,179,120,162,99,249,83,173,112,100,108,88,88,42,100,2,88,224,104,155,129,16,85,214,124,24,80,186,142,204,109,159,141,235,112,143,99,155,109,212,110,230,126,4,132,67,104,3,144,216,109,118,150,168,139,87,89,121,114,228,133,126,129,188,117,138,138,175,104,84,82,34,142,17,149,208,99,152,152,68,142,124,85,83,79,255,102,143,86,213,96,149,109,67,82,73,92,41,89,251,109,107,88,48,117,28,117,108,96,20,130,70,129,17,99,97,103,226,143,58,119,243,141,52,141,193,148,22,94,133,83,44,84,195,112,109,134,111,134,112,134,114,134,115,134,116,134,117,134,118,134,119,134,120,134,131,134,132,134,133,134,134,134,135,134,136,134,137,134,142,134,143,134,144,134,145,134,146,134,148,134,150,134,151,134,152,134,153,134,154,134,155,134,158,134,159,134,160,134,161,134,162,134,165,134,166,134,171,134,173,134,174,134,178,134,179,134,183,134,184,134,185,134,187,134,188,134,189,134,190,134,191,134,193,134,194,134,195,134,197,134,200,134,204,134,205,134,210,134,211,134,213,134,214,134,215,134,218,134,220,134,221,134,224,134,225,134,226,134,227,134,229,134,230,134,231,134,232,134,234,134,235,134,236,134,239,134,245,134,246,134,247,134,250,134,251,134,252,134,253,134,255,134,1,135,4,135,5,135,6,135,11,135,12,135,14,135,15,135,16,135,17,135,20,135,22,135,64,108,247,94,92,80,173,78,173,94,58,99,71,130,26,144,80,104,110,145,179,119,12,84,220,148,100,95,229,122,118,104,69,99,82,123,223,126,219,117,119,80,149,98,52,89,15,144,248,81,195,121,129,122,254,86,146,95,20,144,130,109,96,92,31,87,16,84,84,81,77,110,226,86,168,99,147,152,127,129,21,135,42,137,0,144,30,84,111,92,192,129,214,98,88,98,49,129,53,158,64,150,110,154,124,154,45,105,165,89,211,98,62,85,22,99,199,84,217,134,60,109,3,90,230,116,156,136,106,107,22,89,76,140,47,95,126,110,169,115,125,152,56,78,247,112,140,91,151,120,61,99,90,102,150,118,203,96,155,91,73,90,7,78,85,129,106,108,139,115,161,78,137,103,81,127,128,95,250,101,27,103,216,95,132,89,1,90,25,135,27,135,29,135,31,135,32,135,36,135,38,135,39,135,40,135,42,135,43,135,44,135,45,135,47,135,48,135,50,135,51,135,53,135,54,135,56,135,57,135,58,135,60,135,61,135,64,135,65,135,66,135,67,135,68,135,69,135,70,135,74,135,75,135,77,135,79,135,80,135,81,135,82,135,84,135,85,135,86,135,88,135,90,135,91,135,92,135,93,135,94,135,95,135,97,135,98,135,102,135,103,135,104,135,105,135,106,135,107,135,108,135,109,135,111,135,113,135,114,135,115,135,117,135,119,135,120,135,121,135,122,135,127,135,128,135,129,135,132,135,134,135,135,135,137,135,138,135,140,135,142,135,143,135,144,135,145,135,146,135,148,135,149,135,150,135,152,135,153,135,154,135,155,135,156,135,157,135,158,135,160,135,161,135,162,135,163,135,164,135,205,93,174,95,113,83,230,151,221,143,69,104,244,86,47,85,223,96,58,78,77,111,244,126,199,130,14,132,212,89,31,79,42,79,62,92,172,126,42,103,26,133,115,84,79,117,195,128,130,85,79,155,77,79,45,110,19,140,9,92,112,97,107,83,31,118,41,110,138,134,135,101,251,149,185,126,59,84,51,122,10,125,238,149,225,85,193,127,238,116,29,99,23,135,161,109,157,122,17,98,161,101,103,83,225,99,131,108,235,93,92,84,168,148,76,78,97,108,236,139,75,92,224,101,156,130,167,104,62,84,52,84,203,107,102,107,148,78,66,99,72,83,30,130,13,79,174,79,94,87,10,98,254,150,100,102,105,114,255,82,161,82,159,96,239,139,20,102,153,113,144,103,127,137,82,120,253,119,112,102,59,86,56,84,33,149,122,114,165,135,166,135,167,135,169,135,170,135,174,135,176,135,177,135,178,135,180,135,182,135,183,135,184,135,185,135,187,135,188,135,190,135,191,135,193,135,194,135,195,135,196,135,197,135,199,135,200,135,201,135,204,135,205,135,206,135,207,135,208,135,212,135,213,135,214,135,215,135,216,135,217,135,218,135,220,135,221,135,222,135,223,135,225,135,226,135,227,135,228,135,230,135,231,135,232,135,233,135,235,135,236,135,237,135,239,135,240,135,241,135,242,135,243,135,244,135,245,135,246,135,247,135,248,135,250,135,251,135,252,135,253,135,255,135,0,136,1,136,2,136,4,136,5,136,6,136,7,136,8,136,9,136,11,136,12,136,13,136,14,136,15,136,16,136,17,136,18,136,20,136,23,136,24,136,25,136,26,136,28,136,29,136,30,136,31,136,32,136,35,136,0,122,111,96,12,94,137,96,157,129,21,89,220,96,132,113,239,112,170,110,80,108,128,114,132,106,173,136,45,94,96,78,179,90,156,85,227,148,23,109,251,124,153,150,15,98,198,126,142,119,126,134,35,83,30,151,150,143,135,102,225,92,160,79,237,114,11,78,166,83,15,89,19,84,128,99,40,149,72,81,217,78,156,156,164,126,184,84,36,141,84,136,55,130,242,149,142,109,38,95,204,90,62,102,105,150,176,115,46,115,191,83,122,129,133,153,161,127,170,91,119,150,80,150,191,126,248,118,162,83,118,149,153,153,177,123,68,137,88,110,97,78,212,127,101,121,230,139,243,96,205,84,171,78,121,152,247,93,97,106,207,80,17,84,97,140,39,132,93,120,4,151,74,82,238,84,163,86,0,149,136,109,181,91,198,109,83,102,36,136,37,136,38,136,39,136,40,136,41,136,42,136,43,136,44,136,45,136,46,136,47,136,48,136,49,136,51,136,52,136,53,136,54,136,55,136,56,136,58,136,59,136,61,136,62,136,63,136,65,136,66,136,67,136,70,136,71,136,72,136,73,136,74,136,75,136,78,136,79,136,80,136,81,136,82,136,83,136,85,136,86,136,88,136,90,136,91,136,92,136,93,136,94,136,95,136,96,136,102,136,103,136,106,136,109,136,111,136,113,136,115,136,116,136,117,136,118,136,120,136,121,136,122,136,123,136,124,136,128,136,131,136,134,136,135,136,137,136,138,136,140,136,142,136,143,136,144,136,145,136,147,136,148,136,149,136,151,136,152,136,153,136,154,136,155,136,157,136,158,136,159,136,160,136,161,136,163,136,165,136,166,136,167,136,168,136,169,136,170,136,15,92,93,91,33,104,150,128,120,85,17,123,72,101,84,105,155,78,71,107,78,135,139,151,79,83,31,99,58,100,170,144,156,101,193,128,16,140,153,81,176,104,120,83,249,135,200,97,196,108,251,108,34,140,81,92,170,133,175,130,12,149,35,107,155,143,176,101,251,95,195,95,225,79,69,136,31,102,101,129,41,115,250,96,116,81,17,82,139,87,98,95,162,144,76,136,146,145,120,94,79,103,39,96,211,89,68,81,246,81,248,128,8,83,121,108,196,150,138,113,17,79,238,79,158,127,61,103,197,85,8,149,192,121,150,136,227,126,159,88,12,98,0,151,90,134,24,86,123,152,144,95,184,139,196,132,87,145,217,83,237,101,143,94,92,117,100,96,110,125,127,90,234,126,237,126,105,143,167,85,163,91,172,96,203,101,132,115,172,136,174,136,175,136,176,136,178,136,179,136,180,136,181,136,182,136,184,136,185,136,186,136,187,136,189,136,190,136,191,136,192,136,195,136,196,136,199,136,200,136,202,136,203,136,204,136,205,136,207,136,208,136,209,136,211,136,214,136,215,136,218,136,219,136,220,136,221,136,222,136,224,136,225,136,230,136,231,136,233,136,234,136,235,136,236,136,237,136,238,136,239,136,242,136,245,136,246,136,247,136,250,136,251,136,253,136,255,136,0,137,1,137,3,137,4,137,5,137,6,137,7,137,8,137,9,137,11,137,12,137,13,137,14,137,15,137,17,137,20,137,21,137,22,137,23,137,24,137,28,137,29,137,30,137,31,137,32,137,34,137,35,137,36,137,38,137,39,137,40,137,41,137,44,137,45,137,46,137,47,137,49,137,50,137,51,137,53,137,55,137,9,144,99,118,41,119,218,126,116,151,155,133,102,91,116,122,234,150,64,136,203,82,143,113,170,95,236,101,226,139,251,91,111,154,225,93,137,107,91,108,173,139,175,139,10,144,197,143,139,83,188,98,38,158,45,158,64,84,43,78,189,130,89,114,156,134,22,93,89,136,175,109,197,150,209,84,154,78,182,139,9,113,189,84,9,150,223,112,249,109,208,118,37,78,20,120,18,135,169,92,246,94,0,138,156,152,14,150,142,112,191,108,68,89,169,99,60,119,77,136,20,111,115,130,48,88,213,113,140,83,26,120,193,150,1,85,102,95,48,113,180,91,26,140,140,154,131,107,46,89,47,158,231,121,104,103,108,98,111,79,161,117,138,127,11,109,51,150,39,108,240,78,210,117,123,81,55,104,62,111,128,144,112,129,150,89,118,116,56,137,57,137,58,137,59,137,60,137,61,137,62,137,63,137,64,137,66,137,67,137,69,137,70,137,71,137,72,137,73,137,74,137,75,137,76,137,77,137,78,137,79,137,80,137,81,137,82,137,83,137,84,137,85,137,86,137,87,137,88,137,89,137,90,137,91,137,92,137,93,137,96,137,97,137,98,137,99,137,100,137,101,137,103,137,104,137,105,137,106,137,107,137,108,137,109,137,110,137,111,137,112,137,113,137,114,137,115,137,116,137,117,137,118,137,119,137,120,137,121,137,122,137,124,137,125,137,126,137,128,137,130,137,132,137,133,137,135,137,136,137,137,137,138,137,139,137,140,137,141,137,142,137,143,137,144,137,145,137,146,137,147,137,148,137,149,137,150,137,151,137,152,137,153,137,154,137,155,137,156,137,157,137,158,137,159,137,160,137,161,137,71,100,39,92,101,144,145,122,35,140,218,89,172,84,0,130,111,131,129,137,0,128,48,105,78,86,54,128,55,114,206,145,182,81,95,78,117,152,150,99,26,78,246,83,243,102,75,129,28,89,178,109,0,78,249,88,59,83,214,99,241,148,157,79,10,79,99,136,144,152,55,89,87,144,251,121,234,78,240,128,145,117,130,108,156,91,232,89,93,95,5,105,129,134,26,80,242,93,89,78,227,119,229,78,122,130,145,98,19,102,145,144,121,92,191,78,121,95,198,129,56,144,132,128,171,117,166,78,212,136,15,97,197,107,198,95,73,78,202,118,162,110,227,139,174,139,10,140,209,139,2,95,252,127,204,127,206,126,53,131,107,131,224,86,183,107,243,151,52,150,251,89,31,84,246,148,235,109,197,91,110,153,57,92,21,95,144,150,162,137,163,137,164,137,165,137,166,137,167,137,168,137,169,137,170,137,171,137,172,137,173,137,174,137,175,137,176,137,177,137,178,137,179,137,180,137,181,137,182,137,183,137,184,137,185,137,186,137,187,137,188,137,189,137,190,137,191,137,192,137,195,137,205,137,211,137,212,137,213,137,215,137,216,137,217,137,219,137,221,137,223,137,224,137,225,137,226,137,228,137,231,137,232,137,233,137,234,137,236,137,237,137,238,137,240,137,241,137,242,137,244,137,245,137,246,137,247,137,248,137,249,137,250,137,251,137,252,137,253,137,254,137,255,137,1,138,2,138,3,138,4,138,5,138,6,138,8,138,9,138,10,138,11,138,12,138,13,138,14,138,15,138,16,138,17,138,18,138,19,138,20,138,21,138,22,138,23,138,24,138,25,138,26,138,27,138,28,138,29,138,112,83,241,130,49,106,116,90,112,158,148,94,40,127,185,131,36,132,37,132,103,131,71,135,206,143,98,141,200,118,113,95,150,152,108,120,32,102,223,84,229,98,99,79,195,129,200,117,184,94,205,150,10,142,249,134,143,84,243,108,140,109,56,108,127,96,199,82,40,117,125,94,24,79,160,96,231,95,36,92,49,117,174,144,192,148,185,114,185,108,56,110,73,145,9,103,203,83,243,83,81,79,201,145,241,139,200,83,124,94,194,143,228,109,142,78,194,118,134,105,94,134,26,97,6,130,89,79,222,79,62,144,124,156,9,97,29,110,20,110,133,150,136,78,49,90,232,150,14,78,127,92,185,121,135,91,237,139,189,127,137,115,223,87,139,130,193,144,1,84,71,144,187,85,234,92,161,95,8,97,50,107,241,114,178,128,137,138,30,138,31,138,32,138,33,138,34,138,35,138,36,138,37,138,38,138,39,138,40,138,41,138,42,138,43,138,44,138,45,138,46,138,47,138,48,138,49,138,50,138,51,138,52,138,53,138,54,138,55,138,56,138,57,138,58,138,59,138,60,138,61,138,63,138,64,138,65,138,66,138,67,138,68,138,69,138,70,138,71,138,73,138,74,138,75,138,76,138,77,138,78,138,79,138,80,138,81,138,82,138,83,138,84,138,85,138,86,138,87,138,88,138,89,138,90,138,91,138,92,138,93,138,94,138,95,138,96,138,97,138,98,138,99,138,100,138,101,138,102,138,103,138,104,138,105,138,106,138,107,138,108,138,109,138,110,138,111,138,112,138,113,138,114,138,115,138,116,138,117,138,118,138,119,138,120,138,122,138,123,138,124,138,125,138,126,138,127,138,128,138,116,109,211,91,213,136,132,152,107,140,109,154,51,158,10,110,164,81,67,81,163,87,129,136,159,83,244,99,149,143,237,86,88,84,6,87,63,115,144,110,24,127,220,143,209,130,63,97,40,96,98,150,240,102,166,126,138,141,195,141,165,148,179,92,164,124,8,103,166,96,5,150,24,128,145,78,231,144,0,83,104,150,65,81,208,143,116,133,93,145,85,102,245,151,85,91,29,83,56,120,66,103,61,104,201,84,126,112,176,91,125,143,141,81,40,87,177,84,18,101,130,102,94,141,67,141,15,129,108,132,109,144,223,124,255,81,251,133,163,103,233,101,161,111,164,134,129,142,106,86,32,144,130,118,118,112,229,113,35,141,233,98,25,82,253,108,60,141,14,96,158,88,142,97,254,102,96,141,78,98,179,85,35,110,45,103,103,143,129,138,130,138,131,138,132,138,133,138,134,138,135,138,136,138,139,138,140,138,141,138,142,138,143,138,144,138,145,138,146,138,148,138,149,138,150,138,151,138,152,138,153,138,154,138,155,138,156,138,157,138,158,138,159,138,160,138,161,138,162,138,163,138,164,138,165,138,166,138,167,138,168,138,169,138,170,138,171,138,172,138,173,138,174,138,175,138,176,138,177,138,178,138,179,138,180,138,181,138,182,138,183,138,184,138,185,138,186,138,187,138,188,138,189,138,190,138,191,138,192,138,193,138,194,138,195,138,196,138,197,138,198,138,199,138,200,138,201,138,202,138,203,138,204,138,205,138,206,138,207,138,208,138,209,138,210,138,211,138,212,138,213,138,214,138,215,138,216,138,217,138,218,138,219,138,220,138,221,138,222,138,223,138,224,138,225,138,226,138,227,138,225,148,248,149,40,119,5,104,168,105,139,84,77,78,184,112,200,139,88,100,139,101,133,91,132,122,58,80,232,91,187,119,225,107,121,138,152,124,190,108,207,118,169,101,151,143,45,93,85,92,56,134,8,104,96,83,24,98,217,122,91,110,253,126,31,106,224,122,112,95,51,111,32,95,140,99,168,109,86,103,8,78,16,94,38,141,215,78,192,128,52,118,156,150,219,98,45,102,126,98,188,108,117,141,103,113,105,127,70,81,135,128,236,83,110,144,152,98,242,84,240,134,153,143,5,128,23,149,23,133,217,143,89,109,205,115,159,101,31,119,4,117,39,120,251,129,30,141,136,148,166,79,149,103,185,117,202,139,7,151,47,99,71,149,53,150,184,132,35,99,65,119,129,95,240,114,137,78,20,96,116,101,239,98,99,107,63,101,228,138,229,138,230,138,231,138,232,138,233,138,234,138,235,138,236,138,237,138,238,138,239,138,240,138,241,138,242,138,243,138,244,138,245,138,246,138,247,138,248,138,249,138,250,138,251,138,252,138,253,138,254,138,255,138,0,139,1,139,2,139,3,139,4,139,5,139,6,139,8,139,9,139,10,139,11,139,12,139,13,139,14,139,15,139,16,139,17,139,18,139,19,139,20,139,21,139,22,139,23,139,24,139,25,139,26,139,27,139,28,139,29,139,30,139,31,139,32,139,33,139,34,139,35,139,36,139,37,139,39,139,40,139,41,139,42,139,43,139,44,139,45,139,46,139,47,139,48,139,49,139,50,139,51,139,52,139,53,139,54,139,55,139,56,139,57,139,58,139,59,139,60,139,61,139,62,139,63,139,64,139,65,139,66,139,67,139,68,139,69,139,39,94,199,117,209,144,193,139,157,130,157,103,47,101,49,84,24,135,229,119,162,128,2,129,65,108,75,78,199,126,76,128,244,118,13,105,150,107,103,98,60,80,132,79,64,87,7,99,98,107,190,141,234,83,232,101,184,126,215,95,26,99,183,99,243,129,244,129,110,127,28,94,217,92,54,82,122,102,233,121,26,122,40,141,153,112,212,117,222,110,187,108,146,122,45,78,197,118,224,95,159,148,119,136,200,126,205,121,191,128,205,145,242,78,23,79,31,130,104,84,222,93,50,109,204,139,165,124,116,143,152,128,26,94,146,84,177,118,153,91,60,102,164,154,224,115,42,104,219,134,49,103,42,115,248,139,219,139,16,144,249,122,219,112,110,113,196,98,169,119,49,86,59,78,87,132,241,103,169,82,192,134,46,141,248,148,81,123,70,139,71,139,72,139,73,139,74,139,75,139,76,139,77,139,78,139,79,139,80,139,81,139,82,139,83,139,84,139,85,139,86,139,87,139,88,139,89,139,90,139,91,139,92,139,93,139,94,139,95,139,96,139,97,139,98,139,99,139,100,139,101,139,103,139,104,139,105,139,106,139,107,139,109,139,110,139,111,139,112,139,113,139,114,139,115,139,116,139,117,139,118,139,119,139,120,139,121,139,122,139,123,139,124,139,125,139,126,139,127,139,128,139,129,139,130,139,131,139,132,139,133,139,134,139,135,139,136,139,137,139,138,139,139,139,140,139,141,139,142,139,143,139,144,139,145,139,146,139,147,139,148,139,149,139,150,139,151,139,152,139,153,139,154,139,155,139,156,139,157,139,158,139,159,139,172,139,177,139,187,139,199,139,208,139,234,139,9,140,30,140,79,79,232,108,93,121,123,154,147,98,42,114,253,98,19,78,22,120,108,143,176,100,90,141,198,123,105,104,132,94,197,136,134,89,158,100,238,88,182,114,14,105,37,149,253,143,88,141,96,87,0,127,6,140,198,81,73,99,217,98,83,83,76,104,34,116,1,131,76,145,68,85,64,119,124,112,74,109,121,81,168,84,68,141,255,89,203,110,196,109,92,91,43,125,212,78,125,124,211,110,80,91,234,129,13,110,87,91,3,155,213,104,42,142,151,91,252,126,59,96,181,126,185,144,112,141,79,89,205,99,223,121,179,141,82,83,207,101,86,121,197,139,59,150,196,126,187,148,130,126,52,86,137,145,0,103,106,127,10,92,117,144,40,102,230,93,80,79,222,103,90,80,92,79,80,87,167,94,16,232,17,232,18,232,19,232,20,232,56,140,57,140,58,140,59,140,60,140,61,140,62,140,63,140,64,140,66,140,67,140,68,140,69,140,72,140,74,140,75,140,77,140,78,140,79,140,80,140,81,140,82,140,83,140,84,140,86,140,87,140,88,140,89,140,91,140,92,140,93,140,94,140,95,140,96,140,99,140,100,140,101,140,102,140,103,140,104,140,105,140,108,140,109,140,110,140,111,140,112,140,113,140,114,140,116,140,117,140,118,140,119,140,123,140,124,140,125,140,126,140,127,140,128,140,129,140,131,140,132,140,134,140,135,140,136,140,139,140,141,140,142,140,143,140,144,140,145,140,146,140,147,140,149,140,150,140,151,140,153,140,154,140,155,140,156,140,157,140,158,140,159,140,160,140,161,140,162,140,163,140,164,140,165,140,166,140,167,140,168,140,169,140,170,140,171,140,172,140,173,140,141,78,12,78,64,81,16,78,255,94,69,83,21,78,152,78,30,78,50,155,108,91,105,86,40,78,186,121,63,78,21,83,71,78,45,89,59,114,110,83,16,108,223,86,228,128,151,153,211,107,126,119,23,159,54,78,159,78,16,159,92,78,105,78,147,78,136,130,91,91,108,85,15,86,196,78,141,83,157,83,163,83,165,83,174,83,101,151,93,141,26,83,245,83,38,83,46,83,62,83,92,141,102,83,99,83,2,82,8,82,14,82,45,82,51,82,63,82,64,82,76,82,94,82,97,82,92,82,175,132,125,82,130,82,129,82,144,82,147,82,130,81,84,127,187,78,195,78,201,78,194,78,232,78,225,78,235,78,222,78,27,79,243,78,34,79,100,79,245,78,37,79,39,79,9,79,43,79,94,79,103,79,56,101,90,79,93,79,174,140,175,140,176,140,177,140,178,140,179,140,180,140,181,140,182,140,183,140,184,140,185,140,186,140,187,140,188,140,189,140,190,140,191,140,192,140,193,140,194,140,195,140,196,140,197,140,198,140,199,140,200,140,201,140,202,140,203,140,204,140,205,140,206,140,207,140,208,140,209,140,210,140,211,140,212,140,213,140,214,140,215,140,216,140,217,140,218,140,219,140,220,140,221,140,222,140,223,140,224,140,225,140,226,140,227,140,228,140,229,140,230,140,231,140,232,140,233,140,234,140,235,140,236,140,237,140,238,140,239,140,240,140,241,140,242,140,243,140,244,140,245,140,246,140,247,140,248,140,249,140,250,140,251,140,252,140,253,140,254,140,255,140,0,141,1,141,2,141,3,141,4,141,5,141,6,141,7,141,8,141,9,141,10,141,11,141,12,141,13,141,95,79,87,79,50,79,61,79,118,79,116,79,145,79,137,79,131,79,143,79,126,79,123,79,170,79,124,79,172,79,148,79,230,79,232,79,234,79,197,79,218,79,227,79,220,79,209,79,223,79,248,79,41,80,76,80,243,79,44,80,15,80,46,80,45,80,254,79,28,80,12,80,37,80,40,80,126,80,67,80,85,80,72,80,78,80,108,80,123,80,165,80,167,80,169,80,186,80,214,80,6,81,237,80,236,80,230,80,238,80,7,81,11,81,221,78,61,108,88,79,101,79,206,79,160,159,70,108,116,124,110,81,253,93,201,158,152,153,129,81,20,89,249,82,13,83,7,138,16,83,235,81,25,89,85,81,160,78,86,81,179,78,110,136,164,136,181,78,20,129,210,136,128,121,52,91,3,136,184,127,171,81,177,81,189,81,188,81,14,141,15,141,16,141,17,141,18,141,19,141,20,141,21,141,22,141,23,141,24,141,25,141,26,141,27,141,28,141,32,141,81,141,82,141,87,141,95,141,101,141,104,141,105,141,106,141,108,141,110,141,111,141,113,141,114,141,120,141,121,141,122,141,123,141,124,141,125,141,126,141,127,141,128,141,130,141,131,141,134,141,135,141,136,141,137,141,140,141,141,141,142,141,143,141,144,141,146,141,147,141,149,141,150,141,151,141,152,141,153,141,154,141,155,141,156,141,157,141,158,141,160,141,161,141,162,141,164,141,165,141,166,141,167,141,168,141,169,141,170,141,171,141,172,141,173,141,174,141,175,141,176,141,178,141,182,141,183,141,185,141,187,141,189,141,192,141,193,141,194,141,197,141,199,141,200,141,201,141,202,141,205,141,208,141,210,141,211,141,212,141,199,81,150,81,162,81,165,81,160,139,166,139,167,139,170,139,180,139,181,139,183,139,194,139,195,139,203,139,207,139,206,139,210,139,211,139,212,139,214,139,216,139,217,139,220,139,223,139,224,139,228,139,232,139,233,139,238,139,240,139,243,139,246,139,249,139,252,139,255,139,0,140,2,140,4,140,7,140,12,140,15,140,17,140,18,140,20,140,21,140,22,140,25,140,27,140,24,140,29,140,31,140,32,140,33,140,37,140,39,140,42,140,43,140,46,140,47,140,50,140,51,140,53,140,54,140,105,83,122,83,29,150,34,150,33,150,49,150,42,150,61,150,60,150,66,150,73,150,84,150,95,150,103,150,108,150,114,150,116,150,136,150,141,150,151,150,176,150,151,144,155,144,157,144,153,144,172,144,161,144,180,144,179,144,182,144,186,144,213,141,216,141,217,141,220,141,224,141,225,141,226,141,229,141,230,141,231,141,233,141,237,141,238,141,240,141,241,141,242,141,244,141,246,141,252,141,254,141,255,141,0,142,1,142,2,142,3,142,4,142,6,142,7,142,8,142,11,142,13,142,14,142,16,142,17,142,18,142,19,142,21,142,22,142,23,142,24,142,25,142,26,142,27,142,28,142,32,142,33,142,36,142,37,142,38,142,39,142,40,142,43,142,45,142,48,142,50,142,51,142,52,142,54,142,55,142,56,142,59,142,60,142,62,142,63,142,67,142,69,142,70,142,76,142,77,142,78,142,79,142,80,142,83,142,84,142,85,142,86,142,87,142,88,142,90,142,91,142,92,142,93,142,94,142,95,142,96,142,97,142,98,142,99,142,100,142,101,142,103,142,104,142,106,142,107,142,110,142,113,142,184,144,176,144,207,144,197,144,190,144,208,144,196,144,199,144,211,144,230,144,226,144,220,144,215,144,219,144,235,144,239,144,254,144,4,145,34,145,30,145,35,145,49,145,47,145,57,145,67,145,70,145,13,82,66,89,162,82,172,82,173,82,190,82,255,84,208,82,214,82,240,82,223,83,238,113,205,119,244,94,245,81,252,81,47,155,182,83,1,95,90,117,239,93,76,87,169,87,161,87,126,88,188,88,197,88,209,88,41,87,44,87,42,87,51,87,57,87,46,87,47,87,92,87,59,87,66,87,105,87,133,87,107,87,134,87,124,87,123,87,104,87,109,87,118,87,115,87,173,87,164,87,140,87,178,87,207,87,167,87,180,87,147,87,160,87,213,87,216,87,218,87,217,87,210,87,184,87,244,87,239,87,248,87,228,87,221,87,115,142,117,142,119,142,120,142,121,142,122,142,123,142,125,142,126,142,128,142,130,142,131,142,132,142,134,142,136,142,137,142,138,142,139,142,140,142,141,142,142,142,145,142,146,142,147,142,149,142,150,142,151,142,152,142,153,142,154,142,155,142,157,142,159,142,160,142,161,142,162,142,163,142,164,142,165,142,166,142,167,142,168,142,169,142,170,142,173,142,174,142,176,142,177,142,179,142,180,142,181,142,182,142,183,142,184,142,185,142,187,142,188,142,189,142,190,142,191,142,192,142,193,142,194,142,195,142,196,142,197,142,198,142,199,142,200,142,201,142,202,142,203,142,204,142,205,142,207,142,208,142,209,142,210,142,211,142,212,142,213,142,214,142,215,142,216,142,217,142,218,142,219,142,220,142,221,142,222,142,223,142,224,142,225,142,226,142,227,142,228,142,11,88,13,88,253,87,237,87,0,88,30,88,25,88,68,88,32,88,101,88,108,88,129,88,137,88,154,88,128,88,168,153,25,159,255,97,121,130,125,130,127,130,143,130,138,130,168,130,132,130,142,130,145,130,151,130,153,130,171,130,184,130,190,130,176,130,200,130,202,130,227,130,152,130,183,130,174,130,203,130,204,130,193,130,169,130,180,130,161,130,170,130,159,130,196,130,206,130,164,130,225,130,9,131,247,130,228,130,15,131,7,131,220,130,244,130,210,130,216,130,12,131,251,130,211,130,17,131,26,131,6,131,20,131,21,131,224,130,213,130,28,131,81,131,91,131,92,131,8,131,146,131,60,131,52,131,49,131,155,131,94,131,47,131,79,131,71,131,67,131,95,131,64,131,23,131,96,131,45,131,58,131,51,131,102,131,101,131,229,142,230,142,231,142,232,142,233,142,234,142,235,142,236,142,237,142,238,142,239,142,240,142,241,142,242,142,243,142,244,142,245,142,246,142,247,142,248,142,249,142,250,142,251,142,252,142,253,142,254,142,255,142,0,143,1,143,2,143,3,143,4,143,5,143,6,143,7,143,8,143,9,143,10,143,11,143,12,143,13,143,14,143,15,143,16,143,17,143,18,143,19,143,20,143,21,143,22,143,23,143,24,143,25,143,26,143,27,143,28,143,29,143,30,143,31,143,32,143,33,143,34,143,35,143,36,143,37,143,38,143,39,143,40,143,41,143,42,143,43,143,44,143,45,143,46,143,47,143,48,143,49,143,50,143,51,143,52,143,53,143,54,143,55,143,56,143,57,143,58,143,59,143,60,143,61,143,62,143,63,143,64,143,65,143,66,143,67,143,68,143,104,131,27,131,105,131,108,131,106,131,109,131,110,131,176,131,120,131,179,131,180,131,160,131,170,131,147,131,156,131,133,131,124,131,182,131,169,131,125,131,184,131,123,131,152,131,158,131,168,131,186,131,188,131,193,131,1,132,229,131,216,131,7,88,24,132,11,132,221,131,253,131,214,131,28,132,56,132,17,132,6,132,212,131,223,131,15,132,3,132,248,131,249,131,234,131,197,131,192,131,38,132,240,131,225,131,92,132,81,132,90,132,89,132,115,132,135,132,136,132,122,132,137,132,120,132,60,132,70,132,105,132,118,132,140,132,142,132,49,132,109,132,193,132,205,132,208,132,230,132,189,132,211,132,202,132,191,132,186,132,224,132,161,132,185,132,180,132,151,132,229,132,227,132,12,133,13,117,56,133,240,132,57,133,31,133,58,133,69,143,70,143,71,143,72,143,73,143,74,143,75,143,76,143,77,143,78,143,79,143,80,143,81,143,82,143,83,143,84,143,85,143,86,143,87,143,88,143,89,143,90,143,91,143,92,143,93,143,94,143,95,143,96,143,97,143,98,143,99,143,100,143,101,143,106,143,128,143,140,143,146,143,157,143,160,143,161,143,162,143,164,143,165,143,166,143,167,143,170,143,172,143,173,143,174,143,175,143,178,143,179,143,180,143,181,143,183,143,184,143,186,143,187,143,188,143,191,143,192,143,195,143,198,143,201,143,202,143,203,143,204,143,205,143,207,143,210,143,214,143,215,143,218,143,224,143,225,143,227,143,231,143,236,143,239,143,241,143,242,143,244,143,245,143,246,143,250,143,251,143,252,143,254,143,255,143,7,144,8,144,12,144,14,144,19,144,21,144,24,144,86,133,59,133,255,132,252,132,89,133,72,133,104,133,100,133,94,133,122,133,162,119,67,133,114,133,123,133,164,133,168,133,135,133,143,133,121,133,174,133,156,133,133,133,185,133,183,133,176,133,211,133,193,133,220,133,255,133,39,134,5,134,41,134,22,134,60,134,254,94,8,95,60,89,65,89,55,128,85,89,90,89,88,89,15,83,34,92,37,92,44,92,52,92,76,98,106,98,159,98,187,98,202,98,218,98,215,98,238,98,34,99,246,98,57,99,75,99,67,99,173,99,246,99,113,99,122,99,142,99,180,99,109,99,172,99,138,99,105,99,174,99,188,99,242,99,248,99,224,99,255,99,196,99,222,99,206,99,82,100,198,99,190,99,69,100,65,100,11,100,27,100,32,100,12,100,38,100,33,100,94,100,132,100,109,100,150,100,25,144,28,144,35,144,36,144,37,144,39,144,40,144,41,144,42,144,43,144,44,144,48,144,49,144,50,144,51,144,52,144,55,144,57,144,58,144,61,144,63,144,64,144,67,144,69,144,70,144,72,144,73,144,74,144,75,144,76,144,78,144,84,144,85,144,86,144,89,144,90,144,92,144,93,144,94,144,95,144,96,144,97,144,100,144,102,144,103,144,105,144,106,144,107,144,108,144,111,144,112,144,113,144,114,144,115,144,118,144,119,144,120,144,121,144,122,144,123,144,124,144,126,144,129,144,132,144,133,144,134,144,135,144,137,144,138,144,140,144,141,144,142,144,143,144,144,144,146,144,148,144,150,144,152,144,154,144,156,144,158,144,159,144,160,144,164,144,165,144,167,144,168,144,169,144,171,144,173,144,178,144,183,144,188,144,189,144,191,144,192,144,122,100,183,100,184,100,153,100,186,100,192,100,208,100,215,100,228,100,226,100,9,101,37,101,46,101,11,95,210,95,25,117,17,95,95,83,241,83,253,83,233,83,232,83,251,83,18,84,22,84,6,84,75,84,82,84,83,84,84,84,86,84,67,84,33,84,87,84,89,84,35,84,50,84,130,84,148,84,119,84,113,84,100,84,154,84,155,84,132,84,118,84,102,84,157,84,208,84,173,84,194,84,180,84,210,84,167,84,166,84,211,84,212,84,114,84,163,84,213,84,187,84,191,84,204,84,217,84,218,84,220,84,169,84,170,84,164,84,221,84,207,84,222,84,27,85,231,84,32,85,253,84,20,85,243,84,34,85,35,85,15,85,17,85,39,85,42,85,103,85,143,85,181,85,73,85,109,85,65,85,85,85,63,85,80,85,60,85,194,144,195,144,198,144,200,144,201,144,203,144,204,144,205,144,210,144,212,144,213,144,214,144,216,144,217,144,218,144,222,144,223,144,224,144,227,144,228,144,229,144,233,144,234,144,236,144,238,144,240,144,241,144,242,144,243,144,245,144,246,144,247,144,249,144,250,144,251,144,252,144,255,144,0,145,1,145,3,145,5,145,6,145,7,145,8,145,9,145,10,145,11,145,12,145,13,145,14,145,15,145,16,145,17,145,18,145,19,145,20,145,21,145,22,145,23,145,24,145,26,145,27,145,28,145,29,145,31,145,32,145,33,145,36,145,37,145,38,145,39,145,40,145,41,145,42,145,43,145,44,145,45,145,46,145,48,145,50,145,51,145,52,145,53,145,54,145,55,145,56,145,58,145,59,145,60,145,61,145,62,145,63,145,64,145,65,145,66,145,68,145,55,85,86,85,117,85,118,85,119,85,51,85,48,85,92,85,139,85,210,85,131,85,177,85,185,85,136,85,129,85,159,85,126,85,214,85,145,85,123,85,223,85,189,85,190,85,148,85,153,85,234,85,247,85,201,85,31,86,209,85,235,85,236,85,212,85,230,85,221,85,196,85,239,85,229,85,242,85,243,85,204,85,205,85,232,85,245,85,228,85,148,143,30,86,8,86,12,86,1,86,36,86,35,86,254,85,0,86,39,86,45,86,88,86,57,86,87,86,44,86,77,86,98,86,89,86,92,86,76,86,84,86,134,86,100,86,113,86,107,86,123,86,124,86,133,86,147,86,175,86,212,86,215,86,221,86,225,86,245,86,235,86,249,86,255,86,4,87,10,87,9,87,28,87,15,94,25,94,20,94,17,94,49,94,59,94,60,94,69,145,71,145,72,145,81,145,83,145,84,145,85,145,86,145,88,145,89,145,91,145,92,145,95,145,96,145,102,145,103,145,104,145,107,145,109,145,115,145,122,145,123,145,124,145,128,145,129,145,130,145,131,145,132,145,134,145,136,145,138,145,142,145,143,145,147,145,148,145,149,145,150,145,151,145,152,145,153,145,156,145,157,145,158,145,159,145,160,145,161,145,164,145,165,145,166,145,167,145,168,145,169,145,171,145,172,145,176,145,177,145,178,145,179,145,182,145,183,145,184,145,185,145,187,145,188,145,189,145,190,145,191,145,192,145,193,145,194,145,195,145,196,145,197,145,198,145,200,145,203,145,208,145,210,145,211,145,212,145,213,145,214,145,215,145,216,145,217,145,218,145,219,145,221,145,222,145,223,145,224,145,225,145,226,145,227,145,228,145,229,145,55,94,68,94,84,94,91,94,94,94,97,94,140,92,122,92,141,92,144,92,150,92,136,92,152,92,153,92,145,92,154,92,156,92,181,92,162,92,189,92,172,92,171,92,177,92,163,92,193,92,183,92,196,92,210,92,228,92,203,92,229,92,2,93,3,93,39,93,38,93,46,93,36,93,30,93,6,93,27,93,88,93,62,93,52,93,61,93,108,93,91,93,111,93,93,93,107,93,75,93,74,93,105,93,116,93,130,93,153,93,157,93,115,140,183,93,197,93,115,95,119,95,130,95,135,95,137,95,140,95,149,95,153,95,156,95,168,95,173,95,181,95,188,95,98,136,97,95,173,114,176,114,180,114,183,114,184,114,195,114,193,114,206,114,205,114,210,114,232,114,239,114,233,114,242,114,244,114,247,114,1,115,243,114,3,115,250,114,230,145,231,145,232,145,233,145,234,145,235,145,236,145,237,145,238,145,239,145,240,145,241,145,242,145,243,145,244,145,245,145,246,145,247,145,248,145,249,145,250,145,251,145,252,145,253,145,254,145,255,145,0,146,1,146,2,146,3,146,4,146,5,146,6,146,7,146,8,146,9,146,10,146,11,146,12,146,13,146,14,146,15,146,16,146,17,146,18,146,19,146,20,146,21,146,22,146,23,146,24,146,25,146,26,146,27,146,28,146,29,146,30,146,31,146,32,146,33,146,34,146,35,146,36,146,37,146,38,146,39,146,40,146,41,146,42,146,43,146,44,146,45,146,46,146,47,146,48,146,49,146,50,146,51,146,52,146,53,146,54,146,55,146,56,146,57,146,58,146,59,146,60,146,61,146,62,146,63,146,64,146,65,146,66,146,67,146,68,146,69,146,251,114,23,115,19,115,33,115,10,115,30,115,29,115,21,115,34,115,57,115,37,115,44,115,56,115,49,115,80,115,77,115,87,115,96,115,108,115,111,115,126,115,27,130,37,89,231,152,36,89,2,89,99,153,103,153,104,153,105,153,106,153,107,153,108,153,116,153,119,153,125,153,128,153,132,153,135,153,138,153,141,153,144,153,145,153,147,153,148,153,149,153,128,94,145,94,139,94,150,94,165,94,160,94,185,94,181,94,190,94,179,94,83,141,210,94,209,94,219,94,232,94,234,94,186,129,196,95,201,95,214,95,207,95,3,96,238,95,4,96,225,95,228,95,254,95,5,96,6,96,234,95,237,95,248,95,25,96,53,96,38,96,27,96,15,96,13,96,41,96,43,96,10,96,63,96,33,96,120,96,121,96,123,96,122,96,66,96,70,146,71,146,72,146,73,146,74,146,75,146,76,146,77,146,78,146,79,146,80,146,81,146,82,146,83,146,84,146,85,146,86,146,87,146,88,146,89,146,90,146,91,146,92,146,93,146,94,146,95,146,96,146,97,146,98,146,99,146,100,146,101,146,102,146,103,146,104,146,105,146,106,146,107,146,108,146,109,146,110,146,111,146,112,146,113,146,114,146,115,146,117,146,118,146,119,146,120,146,121,146,122,146,123,146,124,146,125,146,126,146,127,146,128,146,129,146,130,146,131,146,132,146,133,146,134,146,135,146,136,146,137,146,138,146,139,146,140,146,141,146,143,146,144,146,145,146,146,146,147,146,148,146,149,146,150,146,151,146,152,146,153,146,154,146,155,146,156,146,157,146,158,146,159,146,160,146,161,146,162,146,163,146,164,146,165,146,166,146,167,146,106,96,125,96,150,96,154,96,173,96,157,96,131,96,146,96,140,96,155,96,236,96,187,96,177,96,221,96,216,96,198,96,218,96,180,96,32,97,38,97,21,97,35,97,244,96,0,97,14,97,43,97,74,97,117,97,172,97,148,97,167,97,183,97,212,97,245,97,221,95,179,150,233,149,235,149,241,149,243,149,245,149,246,149,252,149,254,149,3,150,4,150,6,150,8,150,10,150,11,150,12,150,13,150,15,150,18,150,21,150,22,150,23,150,25,150,26,150,44,78,63,114,21,98,53,108,84,108,92,108,74,108,163,108,133,108,144,108,148,108,140,108,104,108,105,108,116,108,118,108,134,108,169,108,208,108,212,108,173,108,247,108,248,108,241,108,215,108,178,108,224,108,214,108,250,108,235,108,238,108,177,108,211,108,239,108,254,108,168,146,169,146,170,146,171,146,172,146,173,146,175,146,176,146,177,146,178,146,179,146,180,146,181,146,182,146,183,146,184,146,185,146,186,146,187,146,188,146,189,146,190,146,191,146,192,146,193,146,194,146,195,146,196,146,197,146,198,146,199,146,201,146,202,146,203,146,204,146,205,146,206,146,207,146,208,146,209,146,210,146,211,146,212,146,213,146,214,146,215,146,216,146,217,146,218,146,219,146,220,146,221,146,222,146,223,146,224,146,225,146,226,146,227,146,228,146,229,146,230,146,231,146,232,146,233,146,234,146,235,146,236,146,237,146,238,146,239,146,240,146,241,146,242,146,243,146,244,146,245,146,246,146,247,146,248,146,249,146,250,146,251,146,252,146,253,146,254,146,255,146,0,147,1,147,2,147,3,147,4,147,5,147,6,147,7,147,8,147,9,147,57,109,39,109,12,109,67,109,72,109,7,109,4,109,25,109,14,109,43,109,77,109,46,109,53,109,26,109,79,109,82,109,84,109,51,109,145,109,111,109,158,109,160,109,94,109,147,109,148,109,92,109,96,109,124,109,99,109,26,110,199,109,197,109,222,109,14,110,191,109,224,109,17,110,230,109,221,109,217,109,22,110,171,109,12,110,174,109,43,110,110,110,78,110,107,110,178,110,95,110,134,110,83,110,84,110,50,110,37,110,68,110,223,110,177,110,152,110,224,110,45,111,226,110,165,110,167,110,189,110,187,110,183,110,215,110,180,110,207,110,143,110,194,110,159,110,98,111,70,111,71,111,36,111,21,111,249,110,47,111,54,111,75,111,116,111,42,111,9,111,41,111,137,111,141,111,140,111,120,111,114,111,124,111,122,111,209,111,10,147,11,147,12,147,13,147,14,147,15,147,16,147,17,147,18,147,19,147,20,147,21,147,22,147,23,147,24,147,25,147,26,147,27,147,28,147,29,147,30,147,31,147,32,147,33,147,34,147,35,147,36,147,37,147,38,147,39,147,40,147,41,147,42,147,43,147,44,147,45,147,46,147,47,147,48,147,49,147,50,147,51,147,52,147,53,147,54,147,55,147,56,147,57,147,58,147,59,147,60,147,61,147,63,147,64,147,65,147,66,147,67,147,68,147,69,147,70,147,71,147,72,147,73,147,74,147,75,147,76,147,77,147,78,147,79,147,80,147,81,147,82,147,83,147,84,147,85,147,86,147,87,147,88,147,89,147,90,147,91,147,92,147,93,147,94,147,95,147,96,147,97,147,98,147,99,147,100,147,101,147,102,147,103,147,104,147,105,147,107,147,201,111,167,111,185,111,182,111,194,111,225,111,238,111,222,111,224,111,239,111,26,112,35,112,27,112,57,112,53,112,79,112,94,112,128,91], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+66808);
/* memory initializer */ allocate([132,91,149,91,147,91,165,91,184,91,47,117,158,154,52,100,228,91,238,91,48,137,240,91,71,142,7,139,182,143,211,143,213,143,229,143,238,143,228,143,233,143,230,143,243,143,232,143,5,144,4,144,11,144,38,144,17,144,13,144,22,144,33,144,53,144,54,144,45,144,47,144,68,144,81,144,82,144,80,144,104,144,88,144,98,144,91,144,185,102,116,144,125,144,130,144,136,144,131,144,139,144,80,95,87,95,86,95,88,95,59,92,171,84,80,92,89,92,113,91,99,92,102,92,188,127,42,95,41,95,45,95,116,130,60,95,59,155,110,92,129,89,131,89,141,89,169,89,170,89,163,89,108,147,109,147,110,147,111,147,112,147,113,147,114,147,115,147,116,147,117,147,118,147,119,147,120,147,121,147,122,147,123,147,124,147,125,147,126,147,127,147,128,147,129,147,130,147,131,147,132,147,133,147,134,147,135,147,136,147,137,147,138,147,139,147,140,147,141,147,142,147,144,147,145,147,146,147,147,147,148,147,149,147,150,147,151,147,152,147,153,147,154,147,155,147,156,147,157,147,158,147,159,147,160,147,161,147,162,147,163,147,164,147,165,147,166,147,167,147,168,147,169,147,170,147,171,147,172,147,173,147,174,147,175,147,176,147,177,147,178,147,179,147,180,147,181,147,182,147,183,147,184,147,185,147,186,147,187,147,188,147,189,147,190,147,191,147,192,147,193,147,194,147,195,147,196,147,197,147,198,147,199,147,200,147,201,147,203,147,204,147,205,147,151,89,202,89,171,89,158,89,164,89,210,89,178,89,175,89,215,89,190,89,5,90,6,90,221,89,8,90,227,89,216,89,249,89,12,90,9,90,50,90,52,90,17,90,35,90,19,90,64,90,103,90,74,90,85,90,60,90,98,90,117,90,236,128,170,90,155,90,119,90,122,90,190,90,235,90,178,90,210,90,212,90,184,90,224,90,227,90,241,90,214,90,230,90,216,90,220,90,9,91,23,91,22,91,50,91,55,91,64,91,21,92,28,92,90,91,101,91,115,91,81,91,83,91,98,91,117,154,119,154,120,154,122,154,127,154,125,154,128,154,129,154,133,154,136,154,138,154,144,154,146,154,147,154,150,154,152,154,155,154,156,154,157,154,159,154,160,154,162,154,163,154,165,154,167,154,159,126,161,126,163,126,165,126,168,126,169,126,206,147,207,147,208,147,209,147,210,147,211,147,212,147,213,147,215,147,216,147,217,147,218,147,219,147,220,147,221,147,222,147,223,147,224,147,225,147,226,147,227,147,228,147,229,147,230,147,231,147,232,147,233,147,234,147,235,147,236,147,237,147,238,147,239,147,240,147,241,147,242,147,243,147,244,147,245,147,246,147,247,147,248,147,249,147,250,147,251,147,252,147,253,147,254,147,255,147,0,148,1,148,2,148,3,148,4,148,5,148,6,148,7,148,8,148,9,148,10,148,11,148,12,148,13,148,14,148,15,148,16,148,17,148,18,148,19,148,20,148,21,148,22,148,23,148,24,148,25,148,26,148,27,148,28,148,29,148,30,148,31,148,32,148,33,148,34,148,35,148,36,148,37,148,38,148,39,148,40,148,41,148,42,148,43,148,44,148,45,148,46,148,173,126,176,126,190,126,192,126,193,126,194,126,201,126,203,126,204,126,208,126,212,126,215,126,219,126,224,126,225,126,232,126,235,126,238,126,239,126,241,126,242,126,13,127,246,126,250,126,251,126,254,126,1,127,2,127,3,127,7,127,8,127,11,127,12,127,15,127,17,127,18,127,23,127,25,127,28,127,27,127,31,127,33,127,34,127,35,127,36,127,37,127,38,127,39,127,42,127,43,127,44,127,45,127,47,127,48,127,49,127,50,127,51,127,53,127,122,94,127,117,219,93,62,117,149,144,142,115,145,115,174,115,162,115,159,115,207,115,194,115,209,115,183,115,179,115,192,115,201,115,200,115,229,115,217,115,124,152,10,116,233,115,231,115,222,115,186,115,242,115,15,116,42,116,91,116,38,116,37,116,40,116,48,116,46,116,44,116,47,148,48,148,49,148,50,148,51,148,52,148,53,148,54,148,55,148,56,148,57,148,58,148,59,148,60,148,61,148,63,148,64,148,65,148,66,148,67,148,68,148,69,148,70,148,71,148,72,148,73,148,74,148,75,148,76,148,77,148,78,148,79,148,80,148,81,148,82,148,83,148,84,148,85,148,86,148,87,148,88,148,89,148,90,148,91,148,92,148,93,148,94,148,95,148,96,148,97,148,98,148,99,148,100,148,101,148,102,148,103,148,104,148,105,148,106,148,108,148,109,148,110,148,111,148,112,148,113,148,114,148,115,148,116,148,117,148,118,148,119,148,120,148,121,148,122,148,123,148,124,148,125,148,126,148,127,148,128,148,129,148,130,148,131,148,132,148,145,148,150,148,152,148,199,148,207,148,211,148,212,148,218,148,230,148,251,148,28,149,32,149,27,116,26,116,65,116,92,116,87,116,85,116,89,116,119,116,109,116,126,116,156,116,142,116,128,116,129,116,135,116,139,116,158,116,168,116,169,116,144,116,167,116,210,116,186,116,234,151,235,151,236,151,76,103,83,103,94,103,72,103,105,103,165,103,135,103,106,103,115,103,152,103,167,103,117,103,168,103,158,103,173,103,139,103,119,103,124,103,240,103,9,104,216,103,10,104,233,103,176,103,12,104,217,103,181,103,218,103,179,103,221,103,0,104,195,103,184,103,226,103,14,104,193,103,253,103,50,104,51,104,96,104,97,104,78,104,98,104,68,104,100,104,131,104,29,104,85,104,102,104,65,104,103,104,64,104,62,104,74,104,73,104,41,104,181,104,143,104,116,104,119,104,147,104,107,104,194,104,110,105,252,104,31,105,32,105,249,104,39,149,51,149,61,149,67,149,72,149,75,149,85,149,90,149,96,149,110,149,116,149,117,149,119,149,120,149,121,149,122,149,123,149,124,149,125,149,126,149,128,149,129,149,130,149,131,149,132,149,133,149,134,149,135,149,136,149,137,149,138,149,139,149,140,149,141,149,142,149,143,149,144,149,145,149,146,149,147,149,148,149,149,149,150,149,151,149,152,149,153,149,154,149,155,149,156,149,157,149,158,149,159,149,160,149,161,149,162,149,163,149,164,149,165,149,166,149,167,149,168,149,169,149,170,149,171,149,172,149,173,149,174,149,175,149,176,149,177,149,178,149,179,149,180,149,181,149,182,149,183,149,184,149,185,149,186,149,187,149,188,149,189,149,190,149,191,149,192,149,193,149,194,149,195,149,196,149,197,149,198,149,199,149,200,149,201,149,202,149,203,149,36,105,240,104,11,105,1,105,87,105,227,104,16,105,113,105,57,105,96,105,66,105,93,105,132,105,107,105,128,105,152,105,120,105,52,105,204,105,135,105,136,105,206,105,137,105,102,105,99,105,121,105,155,105,167,105,187,105,171,105,173,105,212,105,177,105,193,105,202,105,223,105,149,105,224,105,141,105,255,105,47,106,237,105,23,106,24,106,101,106,242,105,68,106,62,106,160,106,80,106,91,106,53,106,142,106,121,106,61,106,40,106,88,106,124,106,145,106,144,106,169,106,151,106,171,106,55,115,82,115,129,107,130,107,135,107,132,107,146,107,147,107,141,107,154,107,155,107,161,107,170,107,107,143,109,143,113,143,114,143,115,143,117,143,118,143,120,143,119,143,121,143,122,143,124,143,126,143,129,143,130,143,132,143,135,143,139,143,204,149,205,149,206,149,207,149,208,149,209,149,210,149,211,149,212,149,213,149,214,149,215,149,216,149,217,149,218,149,219,149,220,149,221,149,222,149,223,149,224,149,225,149,226,149,227,149,228,149,229,149,230,149,231,149,236,149,255,149,7,150,19,150,24,150,27,150,30,150,32,150,35,150,36,150,37,150,38,150,39,150,40,150,41,150,43,150,44,150,45,150,47,150,48,150,55,150,56,150,57,150,58,150,62,150,65,150,67,150,74,150,78,150,79,150,81,150,82,150,83,150,86,150,87,150,88,150,89,150,90,150,92,150,93,150,94,150,96,150,99,150,101,150,102,150,107,150,109,150,110,150,111,150,112,150,113,150,115,150,120,150,121,150,122,150,123,150,124,150,125,150,126,150,127,150,128,150,129,150,130,150,131,150,132,150,135,150,137,150,138,150,141,143,142,143,143,143,152,143,154,143,206,142,11,98,23,98,27,98,31,98,34,98,33,98,37,98,36,98,44,98,231,129,239,116,244,116,255,116,15,117,17,117,19,117,52,101,238,101,239,101,240,101,10,102,25,102,114,103,3,102,21,102,0,102,133,112,247,102,29,102,52,102,49,102,54,102,53,102,6,128,95,102,84,102,65,102,79,102,86,102,97,102,87,102,119,102,132,102,140,102,167,102,157,102,190,102,219,102,220,102,230,102,233,102,50,141,51,141,54,141,59,141,61,141,64,141,69,141,70,141,72,141,73,141,71,141,77,141,85,141,89,141,199,137,202,137,203,137,204,137,206,137,207,137,208,137,209,137,110,114,159,114,93,114,102,114,111,114,126,114,127,114,132,114,139,114,141,114,143,114,146,114,8,99,50,99,176,99,140,150,142,150,145,150,146,150,147,150,149,150,150,150,154,150,155,150,157,150,158,150,159,150,160,150,161,150,162,150,163,150,164,150,165,150,166,150,168,150,169,150,170,150,171,150,172,150,173,150,174,150,175,150,177,150,178,150,180,150,181,150,183,150,184,150,186,150,187,150,191,150,194,150,195,150,200,150,202,150,203,150,208,150,209,150,211,150,212,150,214,150,215,150,216,150,217,150,218,150,219,150,220,150,221,150,222,150,223,150,225,150,226,150,227,150,228,150,229,150,230,150,231,150,235,150,236,150,237,150,238,150,240,150,241,150,242,150,244,150,245,150,248,150,250,150,251,150,252,150,253,150,255,150,2,151,3,151,5,151,10,151,11,151,12,151,16,151,17,151,18,151,20,151,21,151,23,151,24,151,25,151,26,151,27,151,29,151,31,151,32,151,63,100,216,100,4,128,234,107,243,107,253,107,245,107,249,107,5,108,7,108,6,108,13,108,21,108,24,108,25,108,26,108,33,108,41,108,36,108,42,108,50,108,53,101,85,101,107,101,77,114,82,114,86,114,48,114,98,134,22,82,159,128,156,128,147,128,188,128,10,103,189,128,177,128,171,128,173,128,180,128,183,128,231,128,232,128,233,128,234,128,219,128,194,128,196,128,217,128,205,128,215,128,16,103,221,128,235,128,241,128,244,128,237,128,13,129,14,129,242,128,252,128,21,103,18,129,90,140,54,129,30,129,44,129,24,129,50,129,72,129,76,129,83,129,116,129,89,129,90,129,113,129,96,129,105,129,124,129,125,129,109,129,103,129,77,88,181,90,136,129,130,129,145,129,213,110,163,129,170,129,204,129,38,103,202,129,187,129,33,151,34,151,35,151,36,151,37,151,38,151,39,151,40,151,41,151,43,151,44,151,46,151,47,151,49,151,51,151,52,151,53,151,54,151,55,151,58,151,59,151,60,151,61,151,63,151,64,151,65,151,66,151,67,151,68,151,69,151,70,151,71,151,72,151,73,151,74,151,75,151,76,151,77,151,78,151,79,151,80,151,81,151,84,151,85,151,87,151,88,151,90,151,92,151,93,151,95,151,99,151,100,151,102,151,103,151,104,151,106,151,107,151,108,151,109,151,110,151,111,151,112,151,113,151,114,151,117,151,119,151,120,151,121,151,122,151,123,151,125,151,126,151,127,151,128,151,129,151,130,151,131,151,132,151,134,151,135,151,136,151,137,151,138,151,140,151,142,151,143,151,144,151,147,151,149,151,150,151,151,151,153,151,154,151,155,151,156,151,157,151,193,129,166,129,36,107,55,107,57,107,67,107,70,107,89,107,209,152,210,152,211,152,213,152,217,152,218,152,179,107,64,95,194,107,243,137,144,101,81,159,147,101,188,101,198,101,196,101,195,101,204,101,206,101,210,101,214,101,128,112,156,112,150,112,157,112,187,112,192,112,183,112,171,112,177,112,232,112,202,112,16,113,19,113,22,113,47,113,49,113,115,113,92,113,104,113,69,113,114,113,74,113,120,113,122,113,152,113,179,113,181,113,168,113,160,113,224,113,212,113,231,113,249,113,29,114,40,114,108,112,24,113,102,113,185,113,62,98,61,98,67,98,72,98,73,98,59,121,64,121,70,121,73,121,91,121,92,121,83,121,90,121,98,121,87,121,96,121,111,121,103,121,122,121,133,121,138,121,154,121,167,121,179,121,209,95,208,95,158,151,159,151,161,151,162,151,164,151,165,151,166,151,167,151,168,151,169,151,170,151,172,151,174,151,176,151,177,151,179,151,181,151,182,151,183,151,184,151,185,151,186,151,187,151,188,151,189,151,190,151,191,151,192,151,193,151,194,151,195,151,196,151,197,151,198,151,199,151,200,151,201,151,202,151,203,151,204,151,205,151,206,151,207,151,208,151,209,151,210,151,211,151,212,151,213,151,214,151,215,151,216,151,217,151,218,151,219,151,220,151,221,151,222,151,223,151,224,151,225,151,226,151,227,151,228,151,229,151,232,151,238,151,239,151,240,151,241,151,242,151,244,151,247,151,248,151,249,151,250,151,251,151,252,151,253,151,254,151,255,151,0,152,1,152,2,152,3,152,4,152,5,152,6,152,7,152,8,152,9,152,10,152,11,152,12,152,13,152,14,152,60,96,93,96,90,96,103,96,65,96,89,96,99,96,171,96,6,97,13,97,93,97,169,97,157,97,203,97,209,97,6,98,128,128,127,128,147,108,246,108,252,109,246,119,248,119,0,120,9,120,23,120,24,120,17,120,171,101,45,120,28,120,29,120,57,120,58,120,59,120,31,120,60,120,37,120,44,120,35,120,41,120,78,120,109,120,86,120,87,120,38,120,80,120,71,120,76,120,106,120,155,120,147,120,154,120,135,120,156,120,161,120,163,120,178,120,185,120,165,120,212,120,217,120,201,120,236,120,242,120,5,121,244,120,19,121,36,121,30,121,52,121,155,159,249,158,251,158,252,158,241,118,4,119,13,119,249,118,7,119,8,119,26,119,34,119,25,119,45,119,38,119,53,119,56,119,80,119,81,119,71,119,67,119,90,119,104,119,15,152,16,152,17,152,18,152,19,152,20,152,21,152,22,152,23,152,24,152,25,152,26,152,27,152,28,152,29,152,30,152,31,152,32,152,33,152,34,152,35,152,36,152,37,152,38,152,39,152,40,152,41,152,42,152,43,152,44,152,45,152,46,152,47,152,48,152,49,152,50,152,51,152,52,152,53,152,54,152,55,152,56,152,57,152,58,152,59,152,60,152,61,152,62,152,63,152,64,152,65,152,66,152,67,152,68,152,69,152,70,152,71,152,72,152,73,152,74,152,75,152,76,152,77,152,78,152,79,152,80,152,81,152,82,152,83,152,84,152,85,152,86,152,87,152,88,152,89,152,90,152,91,152,92,152,93,152,94,152,95,152,96,152,97,152,98,152,99,152,100,152,101,152,102,152,103,152,104,152,105,152,106,152,107,152,108,152,109,152,110,152,98,119,101,119,127,119,141,119,125,119,128,119,140,119,145,119,159,119,160,119,176,119,181,119,189,119,58,117,64,117,78,117,75,117,72,117,91,117,114,117,121,117,131,117,88,127,97,127,95,127,72,138,104,127,116,127,113,127,121,127,129,127,126,127,205,118,229,118,50,136,133,148,134,148,135,148,139,148,138,148,140,148,141,148,143,148,144,148,148,148,151,148,149,148,154,148,155,148,156,148,163,148,164,148,171,148,170,148,173,148,172,148,175,148,176,148,178,148,180,148,182,148,183,148,184,148,185,148,186,148,188,148,189,148,191,148,196,148,200,148,201,148,202,148,203,148,204,148,205,148,206,148,208,148,209,148,210,148,213,148,214,148,215,148,217,148,216,148,219,148,222,148,223,148,224,148,226,148,228,148,229,148,231,148,232,148,234,148,111,152,112,152,113,152,114,152,115,152,116,152,139,152,142,152,146,152,149,152,153,152,163,152,168,152,169,152,170,152,171,152,172,152,173,152,174,152,175,152,176,152,177,152,178,152,179,152,180,152,181,152,182,152,183,152,184,152,185,152,186,152,187,152,188,152,189,152,190,152,191,152,192,152,193,152,194,152,195,152,196,152,197,152,198,152,199,152,200,152,201,152,202,152,203,152,204,152,205,152,207,152,208,152,212,152,214,152,215,152,219,152,220,152,221,152,224,152,225,152,226,152,227,152,228,152,229,152,230,152,233,152,234,152,235,152,236,152,237,152,238,152,239,152,240,152,241,152,242,152,243,152,244,152,245,152,246,152,247,152,248,152,249,152,250,152,251,152,252,152,253,152,254,152,255,152,0,153,1,153,2,153,3,153,4,153,5,153,6,153,7,153,233,148,235,148,238,148,239,148,243,148,244,148,245,148,247,148,249,148,252,148,253,148,255,148,3,149,2,149,6,149,7,149,9,149,10,149,13,149,14,149,15,149,18,149,19,149,20,149,21,149,22,149,24,149,27,149,29,149,30,149,31,149,34,149,42,149,43,149,41,149,44,149,49,149,50,149,52,149,54,149,55,149,56,149,60,149,62,149,63,149,66,149,53,149,68,149,69,149,70,149,73,149,76,149,78,149,79,149,82,149,83,149,84,149,86,149,87,149,88,149,89,149,91,149,94,149,95,149,93,149,97,149,98,149,100,149,101,149,102,149,103,149,104,149,105,149,106,149,107,149,108,149,111,149,113,149,114,149,115,149,58,149,231,119,236,119,201,150,213,121,237,121,227,121,235,121,6,122,71,93,3,122,2,122,30,122,20,122,8,153,9,153,10,153,11,153,12,153,14,153,15,153,17,153,18,153,19,153,20,153,21,153,22,153,23,153,24,153,25,153,26,153,27,153,28,153,29,153,30,153,31,153,32,153,33,153,34,153,35,153,36,153,37,153,38,153,39,153,40,153,41,153,42,153,43,153,44,153,45,153,47,153,48,153,49,153,50,153,51,153,52,153,53,153,54,153,55,153,56,153,57,153,58,153,59,153,60,153,61,153,62,153,63,153,64,153,65,153,66,153,67,153,68,153,69,153,70,153,71,153,72,153,73,153,74,153,75,153,76,153,77,153,78,153,79,153,80,153,81,153,82,153,83,153,86,153,87,153,88,153,89,153,90,153,91,153,92,153,93,153,94,153,95,153,96,153,97,153,98,153,100,153,102,153,115,153,120,153,121,153,123,153,126,153,130,153,131,153,137,153,57,122,55,122,81,122,207,158,165,153,112,122,136,118,142,118,147,118,153,118,164,118,222,116,224,116,44,117,32,158,34,158,40,158,41,158,42,158,43,158,44,158,50,158,49,158,54,158,56,158,55,158,57,158,58,158,62,158,65,158,66,158,68,158,70,158,71,158,72,158,73,158,75,158,76,158,78,158,81,158,85,158,87,158,90,158,91,158,92,158,94,158,99,158,102,158,103,158,104,158,105,158,106,158,107,158,108,158,113,158,109,158,115,158,146,117,148,117,150,117,160,117,157,117,172,117,163,117,179,117,180,117,184,117,196,117,177,117,176,117,195,117,194,117,214,117,205,117,227,117,232,117,230,117,228,117,235,117,231,117,3,118,241,117,252,117,255,117,16,118,0,118,5,118,12,118,23,118,10,118,37,118,24,118,21,118,25,118,140,153,142,153,154,153,155,153,156,153,157,153,158,153,159,153,160,153,161,153,162,153,163,153,164,153,166,153,167,153,169,153,170,153,171,153,172,153,173,153,174,153,175,153,176,153,177,153,178,153,179,153,180,153,181,153,182,153,183,153,184,153,185,153,186,153,187,153,188,153,189,153,190,153,191,153,192,153,193,153,194,153,195,153,196,153,197,153,198,153,199,153,200,153,201,153,202,153,203,153,204,153,205,153,206,153,207,153,208,153,209,153,210,153,211,153,212,153,213,153,214,153,215,153,216,153,217,153,218,153,219,153,220,153,221,153,222,153,223,153,224,153,225,153,226,153,227,153,228,153,229,153,230,153,231,153,232,153,233,153,234,153,235,153,236,153,237,153,238,153,239,153,240,153,241,153,242,153,243,153,244,153,245,153,246,153,247,153,248,153,249,153,27,118,60,118,34,118,32,118,64,118,45,118,48,118,63,118,53,118,67,118,62,118,51,118,77,118,94,118,84,118,92,118,86,118,107,118,111,118,202,127,230,122,120,122,121,122,128,122,134,122,136,122,149,122,166,122,160,122,172,122,168,122,173,122,179,122,100,136,105,136,114,136,125,136,127,136,130,136,162,136,198,136,183,136,188,136,201,136,226,136,206,136,227,136,229,136,241,136,26,137,252,136,232,136,254,136,240,136,33,137,25,137,19,137,27,137,10,137,52,137,43,137,54,137,65,137,102,137,123,137,139,117,229,128,178,118,180,118,220,119,18,128,20,128,22,128,28,128,32,128,34,128,37,128,38,128,39,128,41,128,40,128,49,128,11,128,53,128,67,128,70,128,77,128,82,128,105,128,113,128,131,137,120,152,128,152,131,152,250,153,251,153,252,153,253,153,254,153,255,153,0,154,1,154,2,154,3,154,4,154,5,154,6,154,7,154,8,154,9,154,10,154,11,154,12,154,13,154,14,154,15,154,16,154,17,154,18,154,19,154,20,154,21,154,22,154,23,154,24,154,25,154,26,154,27,154,28,154,29,154,30,154,31,154,32,154,33,154,34,154,35,154,36,154,37,154,38,154,39,154,40,154,41,154,42,154,43,154,44,154,45,154,46,154,47,154,48,154,49,154,50,154,51,154,52,154,53,154,54,154,55,154,56,154,57,154,58,154,59,154,60,154,61,154,62,154,63,154,64,154,65,154,66,154,67,154,68,154,69,154,70,154,71,154,72,154,73,154,74,154,75,154,76,154,77,154,78,154,79,154,80,154,81,154,82,154,83,154,84,154,85,154,86,154,87,154,88,154,89,154,137,152,140,152,141,152,143,152,148,152,154,152,155,152,158,152,159,152,161,152,162,152,165,152,166,152,77,134,84,134,108,134,110,134,127,134,122,134,124,134,123,134,168,134,141,134,139,134,172,134,157,134,167,134,163,134,170,134,147,134,169,134,182,134,196,134,181,134,206,134,176,134,186,134,177,134,175,134,201,134,207,134,180,134,233,134,241,134,242,134,237,134,243,134,208,134,19,135,222,134,244,134,223,134,216,134,209,134,3,135,7,135,248,134,8,135,10,135,13,135,9,135,35,135,59,135,30,135,37,135,46,135,26,135,62,135,72,135,52,135,49,135,41,135,55,135,63,135,130,135,34,135,125,135,126,135,123,135,96,135,112,135,76,135,110,135,139,135,83,135,99,135,124,135,100,135,89,135,101,135,147,135,175,135,168,135,210,135,90,154,91,154,92,154,93,154,94,154,95,154,96,154,97,154,98,154,99,154,100,154,101,154,102,154,103,154,104,154,105,154,106,154,107,154,114,154,131,154,137,154,141,154,142,154,148,154,149,154,153,154,166,154,169,154,170,154,171,154,172,154,173,154,174,154,175,154,178,154,179,154,180,154,181,154,185,154,187,154,189,154,190,154,191,154,195,154,196,154,198,154,199,154,200,154,201,154,202,154,205,154,206,154,207,154,208,154,210,154,212,154,213,154,214,154,215,154,217,154,218,154,219,154,220,154,221,154,222,154,224,154,226,154,227,154,228,154,229,154,231,154,232,154,233,154,234,154,236,154,238,154,240,154,241,154,242,154,243,154,244,154,245,154,246,154,247,154,248,154,250,154,252,154,253,154,254,154,255,154,0,155,1,155,2,155,4,155,5,155,6,155,198,135,136,135,133,135,173,135,151,135,131,135,171,135,229,135,172,135,181,135,179,135,203,135,211,135,189,135,209,135,192,135,202,135,219,135,234,135,224,135,238,135,22,136,19,136,254,135,10,136,27,136,33,136,57,136,60,136,54,127,66,127,68,127,69,127,16,130,250,122,253,122,8,123,3,123,4,123,21,123,10,123,43,123,15,123,71,123,56,123,42,123,25,123,46,123,49,123,32,123,37,123,36,123,51,123,62,123,30,123,88,123,90,123,69,123,117,123,76,123,93,123,96,123,110,123,123,123,98,123,114,123,113,123,144,123,166,123,167,123,184,123,172,123,157,123,168,123,133,123,170,123,156,123,162,123,171,123,180,123,209,123,193,123,204,123,221,123,218,123,229,123,230,123,234,123,12,124,254,123,252,123,15,124,22,124,11,124,7,155,9,155,10,155,11,155,12,155,13,155,14,155,16,155,17,155,18,155,20,155,21,155,22,155,23,155,24,155,25,155,26,155,27,155,28,155,29,155,30,155,32,155,33,155,34,155,36,155,37,155,38,155,39,155,40,155,41,155,42,155,43,155,44,155,45,155,46,155,48,155,49,155,51,155,52,155,53,155,54,155,55,155,56,155,57,155,58,155,61,155,62,155,63,155,64,155,70,155,74,155,75,155,76,155,78,155,80,155,82,155,83,155,85,155,86,155,87,155,88,155,89,155,90,155,91,155,92,155,93,155,94,155,95,155,96,155,97,155,98,155,99,155,100,155,101,155,102,155,103,155,104,155,105,155,106,155,107,155,108,155,109,155,110,155,111,155,112,155,113,155,114,155,115,155,116,155,117,155,118,155,119,155,120,155,121,155,122,155,123,155,31,124,42,124,38,124,56,124,65,124,64,124,254,129,1,130,2,130,4,130,236,129,68,136,33,130,34,130,35,130,45,130,47,130,40,130,43,130,56,130,59,130,51,130,52,130,62,130,68,130,73,130,75,130,79,130,90,130,95,130,104,130,126,136,133,136,136,136,216,136,223,136,94,137,157,127,159,127,167,127,175,127,176,127,178,127,124,124,73,101,145,124,157,124,156,124,158,124,162,124,178,124,188,124,189,124,193,124,199,124,204,124,205,124,200,124,197,124,215,124,232,124,110,130,168,102,191,127,206,127,213,127,229,127,225,127,230,127,233,127,238,127,243,127,248,124,119,125,166,125,174,125,71,126,155,126,184,158,180,158,115,141,132,141,148,141,145,141,177,141,103,141,109,141,71,140,73,140,74,145,80,145,78,145,79,145,100,145,124,155,125,155,126,155,127,155,128,155,129,155,130,155,131,155,132,155,133,155,134,155,135,155,136,155,137,155,138,155,139,155,140,155,141,155,142,155,143,155,144,155,145,155,146,155,147,155,148,155,149,155,150,155,151,155,152,155,153,155,154,155,155,155,156,155,157,155,158,155,159,155,160,155,161,155,162,155,163,155,164,155,165,155,166,155,167,155,168,155,169,155,170,155,171,155,172,155,173,155,174,155,175,155,176,155,177,155,178,155,179,155,180,155,181,155,182,155,183,155,184,155,185,155,186,155,187,155,188,155,189,155,190,155,191,155,192,155,193,155,194,155,195,155,196,155,197,155,198,155,199,155,200,155,201,155,202,155,203,155,204,155,205,155,206,155,207,155,208,155,209,155,210,155,211,155,212,155,213,155,214,155,215,155,216,155,217,155,218,155,219,155,98,145,97,145,112,145,105,145,111,145,125,145,126,145,114,145,116,145,121,145,140,145,133,145,144,145,141,145,145,145,162,145,163,145,170,145,173,145,174,145,175,145,181,145,180,145,186,145,85,140,126,158,184,141,235,141,5,142,89,142,105,142,181,141,191,141,188,141,186,141,196,141,214,141,215,141,218,141,222,141,206,141,207,141,219,141,198,141,236,141,247,141,248,141,227,141,249,141,251,141,228,141,9,142,253,141,20,142,29,142,31,142,44,142,46,142,35,142,47,142,58,142,64,142,57,142,53,142,61,142,49,142,73,142,65,142,66,142,81,142,82,142,74,142,112,142,118,142,124,142,111,142,116,142,133,142,143,142,148,142,144,142,156,142,158,142,120,140,130,140,138,140,133,140,152,140,148,140,155,101,214,137,222,137,218,137,220,137,220,155,221,155,222,155,223,155,224,155,225,155,226,155,227,155,228,155,229,155,230,155,231,155,232,155,233,155,234,155,235,155,236,155,237,155,238,155,239,155,240,155,241,155,242,155,243,155,244,155,245,155,246,155,247,155,248,155,249,155,250,155,251,155,252,155,253,155,254,155,255,155,0,156,1,156,2,156,3,156,4,156,5,156,6,156,7,156,8,156,9,156,10,156,11,156,12,156,13,156,14,156,15,156,16,156,17,156,18,156,19,156,20,156,21,156,22,156,23,156,24,156,25,156,26,156,27,156,28,156,29,156,30,156,31,156,32,156,33,156,34,156,35,156,36,156,37,156,38,156,39,156,40,156,41,156,42,156,43,156,44,156,45,156,46,156,47,156,48,156,49,156,50,156,51,156,52,156,53,156,54,156,55,156,56,156,57,156,58,156,59,156,229,137,235,137,239,137,62,138,38,139,83,151,233,150,243,150,239,150,6,151,1,151,8,151,15,151,14,151,42,151,45,151,48,151,62,151,128,159,131,159,133,159,134,159,135,159,136,159,137,159,138,159,140,159,254,158,11,159,13,159,185,150,188,150,189,150,206,150,210,150,191,119,224,150,142,146,174,146,200,146,62,147,106,147,202,147,143,147,62,148,107,148,127,156,130,156,133,156,134,156,135,156,136,156,35,122,139,156,142,156,144,156,145,156,146,156,148,156,149,156,154,156,155,156,158,156,159,156,160,156,161,156,162,156,163,156,165,156,166,156,167,156,168,156,169,156,171,156,173,156,174,156,176,156,177,156,178,156,179,156,180,156,181,156,182,156,183,156,186,156,187,156,188,156,189,156,196,156,197,156,198,156,199,156,202,156,203,156,60,156,61,156,62,156,63,156,64,156,65,156,66,156,67,156,68,156,69,156,70,156,71,156,72,156,73,156,74,156,75,156,76,156,77,156,78,156,79,156,80,156,81,156,82,156,83,156,84,156,85,156,86,156,87,156,88,156,89,156,90,156,91,156,92,156,93,156,94,156,95,156,96,156,97,156,98,156,99,156,100,156,101,156,102,156,103,156,104,156,105,156,106,156,107,156,108,156,109,156,110,156,111,156,112,156,113,156,114,156,115,156,116,156,117,156,118,156,119,156,120,156,121,156,122,156,123,156,125,156,126,156,128,156,131,156,132,156,137,156,138,156,140,156,143,156,147,156,150,156,151,156,152,156,153,156,157,156,170,156,172,156,175,156,185,156,190,156,191,156,192,156,193,156,194,156,200,156,201,156,209,156,210,156,218,156,219,156,224,156,225,156,204,156,205,156,206,156,207,156,208,156,211,156,212,156,213,156,215,156,216,156,217,156,220,156,221,156,223,156,226,156,124,151,133,151,145,151,146,151,148,151,175,151,171,151,163,151,178,151,180,151,177,154,176,154,183,154,88,158,182,154,186,154,188,154,193,154,192,154,197,154,194,154,203,154,204,154,209,154,69,155,67,155,71,155,73,155,72,155,77,155,81,155,232,152,13,153,46,153,85,153,84,153,223,154,225,154,230,154,239,154,235,154,251,154,237,154,249,154,8,155,15,155,19,155,31,155,35,155,189,158,190,158,59,126,130,158,135,158,136,158,139,158,146,158,214,147,157,158,159,158,219,158,220,158,221,158,224,158,223,158,226,158,233,158,231,158,229,158,234,158,239,158,34,159,44,159,47,159,57,159,55,159,61,159,62,159,68,159,227,156,228,156,229,156,230,156,231,156,232,156,233,156,234,156,235,156,236,156,237,156,238,156,239,156,240,156,241,156,242,156,243,156,244,156,245,156,246,156,247,156,248,156,249,156,250,156,251,156,252,156,253,156,254,156,255,156,0,157,1,157,2,157,3,157,4,157,5,157,6,157,7,157,8,157,9,157,10,157,11,157,12,157,13,157,14,157,15,157,16,157,17,157,18,157,19,157,20,157,21,157,22,157,23,157,24,157,25,157,26,157,27,157,28,157,29,157,30,157,31,157,32,157,33,157,34,157,35,157,36,157,37,157,38,157,39,157,40,157,41,157,42,157,43,157,44,157,45,157,46,157,47,157,48,157,49,157,50,157,51,157,52,157,53,157,54,157,55,157,56,157,57,157,58,157,59,157,60,157,61,157,62,157,63,157,64,157,65,157,66,157,52,226,53,226,54,226,55,226,56,226,57,226,58,226,59,226,60,226,61,226,62,226,63,226,64,226,65,226,66,226,67,226,68,226,69,226,70,226,71,226,72,226,73,226,74,226,75,226,76,226,77,226,78,226,79,226,80,226,81,226,82,226,83,226,84,226,85,226,86,226,87,226,88,226,89,226,90,226,91,226,92,226,93,226,94,226,95,226,96,226,97,226,98,226,99,226,100,226,101,226,102,226,103,226,104,226,105,226,106,226,107,226,108,226,109,226,110,226,111,226,112,226,113,226,114,226,115,226,116,226,117,226,118,226,119,226,120,226,121,226,122,226,123,226,124,226,125,226,126,226,127,226,128,226,129,226,130,226,131,226,132,226,133,226,134,226,135,226,136,226,137,226,138,226,139,226,140,226,141,226,142,226,143,226,144,226,145,226,67,157,68,157,69,157,70,157,71,157,72,157,73,157,74,157,75,157,76,157,77,157,78,157,79,157,80,157,81,157,82,157,83,157,84,157,85,157,86,157,87,157,88,157,89,157,90,157,91,157,92,157,93,157,94,157,95,157,96,157,97,157,98,157,99,157,100,157,101,157,102,157,103,157,104,157,105,157,106,157,107,157,108,157,109,157,110,157,111,157,112,157,113,157,114,157,115,157,116,157,117,157,118,157,119,157,120,157,121,157,122,157,123,157,124,157,125,157,126,157,127,157,128,157,129,157,130,157,131,157,132,157,133,157,134,157,135,157,136,157,137,157,138,157,139,157,140,157,141,157,142,157,143,157,144,157,145,157,146,157,147,157,148,157,149,157,150,157,151,157,152,157,153,157,154,157,155,157,156,157,157,157,158,157,159,157,160,157,161,157,162,157,146,226,147,226,148,226,149,226,150,226,151,226,152,226,153,226,154,226,155,226,156,226,157,226,158,226,159,226,160,226,161,226,162,226,163,226,164,226,165,226,166,226,167,226,168,226,169,226,170,226,171,226,172,226,173,226,174,226,175,226,176,226,177,226,178,226,179,226,180,226,181,226,182,226,183,226,184,226,185,226,186,226,187,226,188,226,189,226,190,226,191,226,192,226,193,226,194,226,195,226,196,226,197,226,198,226,199,226,200,226,201,226,202,226,203,226,204,226,205,226,206,226,207,226,208,226,209,226,210,226,211,226,212,226,213,226,214,226,215,226,216,226,217,226,218,226,219,226,220,226,221,226,222,226,223,226,224,226,225,226,226,226,227,226,228,226,229,226,230,226,231,226,232,226,233,226,234,226,235,226,236,226,237,226,238,226,239,226,163,157,164,157,165,157,166,157,167,157,168,157,169,157,170,157,171,157,172,157,173,157,174,157,175,157,176,157,177,157,178,157,179,157,180,157,181,157,182,157,183,157,184,157,185,157,186,157,187,157,188,157,189,157,190,157,191,157,192,157,193,157,194,157,195,157,196,157,197,157,198,157,199,157,200,157,201,157,202,157,203,157,204,157,205,157,206,157,207,157,208,157,209,157,210,157,211,157,212,157,213,157,214,157,215,157,216,157,217,157,218,157,219,157,220,157,221,157,222,157,223,157,224,157,225,157,226,157,227,157,228,157,229,157,230,157,231,157,232,157,233,157,234,157,235,157,236,157,237,157,238,157,239,157,240,157,241,157,242,157,243,157,244,157,245,157,246,157,247,157,248,157,249,157,250,157,251,157,252,157,253,157,254,157,255,157,0,158,1,158,2,158,240,226,241,226,242,226,243,226,244,226,245,226,246,226,247,226,248,226,249,226,250,226,251,226,252,226,253,226,254,226,255,226,0,227,1,227,2,227,3,227,4,227,5,227,6,227,7,227,8,227,9,227,10,227,11,227,12,227,13,227,14,227,15,227,16,227,17,227,18,227,19,227,20,227,21,227,22,227,23,227,24,227,25,227,26,227,27,227,28,227,29,227,30,227,31,227,32,227,33,227,34,227,35,227,36,227,37,227,38,227,39,227,40,227,41,227,42,227,43,227,44,227,45,227,46,227,47,227,48,227,49,227,50,227,51,227,52,227,53,227,54,227,55,227,56,227,57,227,58,227,59,227,60,227,61,227,62,227,63,227,64,227,65,227,66,227,67,227,68,227,69,227,70,227,71,227,72,227,73,227,74,227,75,227,76,227,77,227,3,158,4,158,5,158,6,158,7,158,8,158,9,158,10,158,11,158,12,158,13,158,14,158,15,158,16,158,17,158,18,158,19,158,20,158,21,158,22,158,23,158,24,158,25,158,26,158,27,158,28,158,29,158,30,158,36,158,39,158,46,158,48,158,52,158,59,158,60,158,64,158,77,158,80,158,82,158,83,158,84,158,86,158,89,158,93,158,95,158,96,158,97,158,98,158,101,158,110,158,111,158,114,158,116,158,117,158,118,158,119,158,120,158,121,158,122,158,123,158,124,158,125,158,128,158,129,158,131,158,132,158,133,158,134,158,137,158,138,158,140,158,141,158,142,158,143,158,144,158,145,158,148,158,149,158,150,158,151,158,152,158,153,158,154,158,155,158,156,158,158,158,160,158,161,158,162,158,163,158,164,158,165,158,167,158,168,158,169,158,170,158,78,227,79,227,80,227,81,227,82,227,83,227,84,227,85,227,86,227,87,227,88,227,89,227,90,227,91,227,92,227,93,227,94,227,95,227,96,227,97,227,98,227,99,227,100,227,101,227,102,227,103,227,104,227,105,227,106,227,107,227,108,227,109,227,110,227,111,227,112,227,113,227,114,227,115,227,116,227,117,227,118,227,119,227,120,227,121,227,122,227,123,227,124,227,125,227,126,227,127,227,128,227,129,227,130,227,131,227,132,227,133,227,134,227,135,227,136,227,137,227,138,227,139,227,140,227,141,227,142,227,143,227,144,227,145,227,146,227,147,227,148,227,149,227,150,227,151,227,152,227,153,227,154,227,155,227,156,227,157,227,158,227,159,227,160,227,161,227,162,227,163,227,164,227,165,227,166,227,167,227,168,227,169,227,170,227,171,227,171,158,172,158,173,158,174,158,175,158,176,158,177,158,178,158,179,158,181,158,182,158,183,158,185,158,186,158,188,158,191,158,192,158,193,158,194,158,195,158,197,158,198,158,199,158,200,158,202,158,203,158,204,158,208,158,210,158,211,158,213,158,214,158,215,158,217,158,218,158,222,158,225,158,227,158,228,158,230,158,232,158,235,158,236,158,237,158,238,158,240,158,241,158,242,158,243,158,244,158,245,158,246,158,247,158,248,158,250,158,253,158,255,158,0,159,1,159,2,159,3,159,4,159,5,159,6,159,7,159,8,159,9,159,10,159,12,159,15,159,17,159,18,159,20,159,21,159,22,159,24,159,26,159,27,159,28,159,29,159,30,159,31,159,33,159,35,159,36,159,37,159,38,159,39,159,40,159,41,159,42,159,43,159,45,159,46,159,48,159,49,159,172,227,173,227,174,227,175,227,176,227,177,227,178,227,179,227,180,227,181,227,182,227,183,227,184,227,185,227,186,227,187,227,188,227,189,227,190,227,191,227,192,227,193,227,194,227,195,227,196,227,197,227,198,227,199,227,200,227,201,227,202,227,203,227,204,227,205,227,206,227,207,227,208,227,209,227,210,227,211,227,212,227,213,227,214,227,215,227,216,227,217,227,218,227,219,227,220,227,221,227,222,227,223,227,224,227,225,227,226,227,227,227,228,227,229,227,230,227,231,227,232,227,233,227,234,227,235,227,236,227,237,227,238,227,239,227,240,227,241,227,242,227,243,227,244,227,245,227,246,227,247,227,248,227,249,227,250,227,251,227,252,227,253,227,254,227,255,227,0,228,1,228,2,228,3,228,4,228,5,228,6,228,7,228,8,228,9,228,50,159,51,159,52,159,53,159,54,159,56,159,58,159,60,159,63,159,64,159,65,159,66,159,67,159,69,159,70,159,71,159,72,159,73,159,74,159,75,159,76,159,77,159,78,159,79,159,82,159,83,159,84,159,85,159,86,159,87,159,88,159,89,159,90,159,91,159,92,159,93,159,94,159,95,159,96,159,97,159,98,159,99,159,100,159,101,159,102,159,103,159,104,159,105,159,106,159,107,159,108,159,109,159,110,159,111,159,112,159,113,159,114,159,115,159,116,159,117,159,118,159,119,159,120,159,121,159,122,159,123,159,124,159,125,159,126,159,129,159,130,159,141,159,142,159,143,159,144,159,145,159,146,159,147,159,148,159,149,159,150,159,151,159,152,159,156,159,157,159,158,159,161,159,162,159,163,159,164,159,165,159,44,249,121,249,149,249,231,249,241,249,10,228,11,228,12,228,13,228,14,228,15,228,16,228,17,228,18,228,19,228,20,228,21,228,22,228,23,228,24,228,25,228,26,228,27,228,28,228,29,228,30,228,31,228,32,228,33,228,34,228,35,228,36,228,37,228,38,228,39,228,40,228,41,228,42,228,43,228,44,228,45,228,46,228,47,228,48,228,49,228,50,228,51,228,52,228,53,228,54,228,55,228,56,228,57,228,58,228,59,228,60,228,61,228,62,228,63,228,64,228,65,228,66,228,67,228,68,228,69,228,70,228,71,228,72,228,73,228,74,228,75,228,76,228,77,228,78,228,79,228,80,228,81,228,82,228,83,228,84,228,85,228,86,228,87,228,88,228,89,228,90,228,91,228,92,228,93,228,94,228,95,228,96,228,97,228,98,228,99,228,100,228,101,228,102,228,103,228,12,250,13,250,14,250,15,250,17,250,19,250,20,250,24,250,31,250,32,250,33,250,35,250,36,250,39,250,40,250,41,250,129,46,22,232,23,232,24,232,132,46,115,52,71,52,136,46,139,46,30,232,158,53,26,54,14,54,140,46,151,46,110,57,24,57,38,232,207,57,223,57,115,58,208,57,43,232,44,232,78,59,110,60,224,60,167,46,49,232,50,232,170,46,86,64,95,65,174,46,55,67,179,46,182,46,183,46,59,232,177,67,172,67,187,46,221,67,214,68,97,70,76,70,67,232,35,71,41,71,124,71,141,71,202,46,71,73,122,73,125,73,130,73,131,73,133,73,134,73,159,73,155,73,183,73,182,73,84,232,85,232,163,76,159,76,160,76,161,76,119,76,162,76,19,77,20,77,21,77,22,77,23,77,24,77,25,77,174,77,100,232,104,228,105,228,106,228,107,228,108,228,109,228,110,228,111,228,112,228,113,228,114,228,115,228,116,228,117,228,118,228,119,228,120,228,121,228,122,228,123,228,124,228,125,228,126,228,127,228,128,228,129,228,130,228,131,228,132,228,133,228,134,228,135,228,136,228,137,228,138,228,139,228,140,228,141,228,142,228,143,228,144,228,145,228,146,228,147,228,148,228,149,228,150,228,151,228,152,228,153,228,154,228,155,228,156,228,157,228,158,228,159,228,160,228,161,228,162,228,163,228,164,228,165,228,166,228,167,228,168,228,169,228,170,228,171,228,172,228,173,228,174,228,175,228,176,228,177,228,178,228,179,228,180,228,181,228,182,228,183,228,184,228,185,228,186,228,187,228,188,228,189,228,190,228,191,228,192,228,193,228,194,228,195,228,196,228,197,228,240,67,50,76,3,70,166,69,120,69,103,114,119,77,179,69,177,124,226,76,197,124,149,59,54,71,68,71,71,76,64,76,191,66,23,54,82,115,139,110,210,112,87,76,81,163,79,71,218,69,133,76,108,124,7,77,164,74,161,70,35,107,37,114,84,90,99,26,6,62,97,63,77,102,251,86,0,0,149,125,29,89,185,139,244,61,52,151,239,123,219,91,94,29,164,90,37,54,176,158,209,90,183,91,252,92,110,103,147,133,69,153,97,116,157,116,117,56,83,29,158,54,33,96,236,62,222,88,245,58,252,122,151,159,97,65,13,137,234,49,138,10,94,50,10,67,132,132,150,159,47,148,48,73,19,134,150,88,74,151,24,146,208,121,50,122,96,102,41,106,157,136,76,116,197,123,130,103,44,122,79,82,70,144,230,52,196,115,185,93,198,116,199,159,179,87,47,73,76,84,49,65,142,54,24,88,114,122,101,123,143,139,174,70,136,110,129,65,153,93,174,123,188,36,200,159,193,36,201,36,204,36,201,159,4,133,187,53,180,64,202,159,225,68,255,173,193,98,110,112,203,159,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,192,49,193,49,194,49,195,49,196,49,12,1,197,49,209,0,205,0,198,49,199,49,203,0,232,31,200,49,202,0,201,49,202,49,203,49,204,49,14,1,205,49,206,49,0,1,193,0,205,1,192,0,18,1,201,0,26,1,200,0,76,1,211,0,209,1,210,0,0,220,190,30,4,220,192,30,202,0,1,1,225,0,206,1,224,0,81,2,19,1,233,0,27,1,232,0,43,1,237,0,208,1,236,0,77,1,243,0,210,1,242,0,107,1,250,0,212,1,249,0,214,1,216,1,218,1,220,1,252,0,8,220,191,30,12,220,193,30,234,0,97,2,218,35,219,35], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+77048);
/* memory initializer */ allocate([169,163,69,17,0,0,10,101,0,0,0,0,61,78,221,110,78,157,223,145,0,0,0,0,53,119,145,100,26,79,40,79,168,79,86,81,116,81,156,81,228,81,161,82,168,82,59,83,78,83,209,83,216,83,226,86,240,88,4,89,7,89,50,89,52,89,102,91,158,91,159,91,154,92,134,94,59,96,137,101,254,103,4,104,101,104,78,109,188,112,53,117,164,126,172,126,186,126,199,126,207,126,223,126,6,127,55,127,122,130,207,130,111,131,198,137,190,139,226,139,102,143,103,143,110,143,17,116,252,124,205,125,70,105,201,122,39,82,0,0,0,0,0,0,0,0,140,145,184,120,94,145,188,128,0,0,11,141,246,128,231,9,0,0,0,0,159,128,199,158,205,76,201,157,12,158,62,76,246,157,14,112,10,158,51,161,193,53,0,0,154,110,62,130,25,117,0,0,17,73,108,154,143,154,153,159,135,121,108,132,202,29,208,5,230,42,36,78,129,78,128,78,135,78,191,78,235,78,55,79,76,52,189,79,72,62,3,80,136,80,125,52,147,52,165,52,134,81,5,89,219,81,252,81,5,82,137,78,121,82,144,82,39,83,199,53,169,83,81,53,176,83,83,53,194,83,35,84,109,53,114,53,129,54,147,84,163,84,180,84,185,84,208,84,239,84,24,85,35,85,40,85,152,53,63,85,165,53,191,53,215,85,197,53,132,125,37,85,0,0,66,12,21,13,43,81,144,85,198,44,236,57,65,3,70,142,184,77,229,148,83,64,190,128,122,119,56,44,52,58,213,71,93,129,242,105,234,77,221,100,124,13,180,15,213,12,244,16,141,100,126,142,150,14,11,12,100,15,169,44,86,130,211,68,0,0,70,13,77,154,233,128,244,71,167,78,194,44,178,154,103,58,244,149,237,63,6,53,199,82,212,151,200,120,68,45,110,157,21,152,0,0,217,67,165,96,180,100,227,84,76,45,202,43,119,16,251,57,111,16,218,102,22,103,160,121,234,100,82,80,67,12,104,142,161,33,76,139,49,7,0,0,11,72,169,1,250,63,115,88,141,45,0,0,200,69,252,4,151,96,76,15,150,13,121,85,187,64,186,67,0,0,180,74,102,42,157,16,170,129,245,152,156,13,121,99,254,57,117,39,192,141,161,86,124,100,67,62,0,0,1,166,9,14,207,42,201,44,0,0,200,16,194,57,146,57,6,58,155,130,120,53,73,94,199,32,82,86,49,15,178,44,32,151,188,52,61,108,59,78,0,0,0,0,116,117,139,46,8,34,91,166,205,140,122,14,52,12,28,104,147,127,207,16,3,40,57,41,251,53,227,81,140,14,141,15,170,14,147,63,48,15,71,13,79,17,76,14,0,0,171,14,169,11,72,13,192,16,61,17,249,63,150,38,50,100,173,15,244,51,57,118,206,43,126,13,127,13,81,44,85,44,24,58,152,14,199,16,46,15,50,166,80,107,210,140,153,141,202,140,170,149,204,84,196,130,185,85,0,0,195,158,38,156,182,154,94,119,238,45,64,113,109,129,236,128,28,92,114,101,52,129,151,55,95,83,189,128,182,145,250,14,15,14,119,14,251,14,221,53,235,77,9,54,214,12,175,86,181,39,201,16,16,14,120,14,120,16,72,17,7,130,85,20,121,14,80,78,164,45,84,90,29,16,30,16,245,16,246,16,156,87,17,14,148,118,205,130,181,15,123,14,126,81,3,55,182,15,128,17,216,82,189,162,218,73,58,24,119,65,124,130,153,88,104,82,26,54,61,87,178,123,104,91,0,72,44,75,39,159,231,73,31,156,141,155,116,91,61,49,251,85,242,53,137,86,40,78,2,89,193,27,120,248,81,151,134,0,91,78,187,78,62,53,35,92,81,95,196,95,250,56,76,98,53,101,122,107,53,108,58,108,108,112,43,114,44,78,173,114,233,72,82,127,59,121,249,124,83,127,106,98,193,52,0,0,75,99,2,128,128,128,18,102,81,105,93,83,100,136,193,137,178,120,160,139,29,141,133,148,120,149,127,149,232,149,15,142,230,151,117,152,206,152,222,152,99,153,16,152,124,156,31,158,196,158,111,107,7,249,55,78,135,0,29,150,55,98,162,148,0,0,59,80,254,109,115,156,166,159,201,61,143,136,78,65,119,112,245,92,32,75,205,81,89,53,48,93,34,97,50,138,167,143,246,145,145,113,25,103,186,115,129,50,7,161,139,60,128,25,16,75,228,120,2,116,174,81,15,135,9,64,99,106,186,162,35,66,15,134,111,10,42,122,71,153,234,138,85,151,77,112,36,83,126,32,244,147,217,118,227,137,167,159,221,119,163,78,240,79,188,80,47,78,23,79,168,159,52,84,139,125,146,88,208,88,182,29,146,94,153,94,194,95,18,39,139,101,249,51,25,105,67,106,99,60,255,108,0,0,0,114,5,69,140,115,219,62,19,74,21,91,185,116,131,139,164,92,149,86,147,122,236,123,195,124,108,126,248,130,151,133,169,159,144,136,170,159,185,142,171,159,207,143,95,133,224,153,33,146,172,159,185,141,63,20,113,64,162,66,26,90,0,0,0,0,0,0,104,152,107,103,118,66,61,87,0,0,214,133,123,73,191,130,13,113,129,76,116,109,123,93,21,107,190,111,173,159,174,159,150,91,175,159,231,102,91,126,87,110,202,121,136,61,195,68,86,50,150,39,154,67,54,69,0,0,213,92,26,59,249,138,120,92,18,61,81,53,120,93,178,159,87,113,88,69,236,64,35,30,119,76,120,57,74,52,164,1,65,108,204,138,180,79,57,2,191,89,108,129,86,152,250,152,59,95,159,11,0,0,193,33,109,137,2,65,187,70,121,144,7,63,179,159,181,161,248,64,214,55,247,70,70,108,124,65,178,134,255,115,109,69,212,56,154,84,97,69,27,69,137,77,123,76,118,77,234,69,200,63,15,75,97,54,222,68,189,68,237,65,62,93,72,93,86,93,252,61,15,56,164,93,185,93,32,56,56,56,66,94,189,94,37,95,131,95,8,57,20,57,63,57,77,57,215,96,61,97,229,92,137,57,183,97,185,97,207,97,184,57,44,98,144,98,229,98,24,99,248,57,177,86,3,58,226,99,251,99,7,100,90,100,75,58,192,100,21,93,33,86,159,159,151,58,134,101,189,58,255,101,83,102,242,58,146,102,34,59,22,103,66,59,164,103,0,104,88,59,74,104,132,104,114,59,113,59,123,59,9,105,67,105,92,114,100,105,159,105,133,105,188,59,214,105,221,59,101,106,116,106,113,106,130,106,236,59,153,106,242,59,171,106,181,106,212,106,246,106,129,107,193,107,234,107,117,108,170,108,203,60,2,109,6,109,38,109,129,109,239,60,164,109,177,109,21,110,24,110,41,110,134,110,192,137,187,110,226,110,218,110,127,159,232,110,233,110,36,111,52,111,70,61,65,63,129,111,190,111,106,61,117,61,183,113,153,92,138,61,44,112,145,61,80,112,84,112,111,112,127,112,137,112,37,3,193,67,241,53,216,14,215,62,190,87,211,110,62,113,224,87,78,54,162,105,233,139,116,91,73,122,225,88,217,148,101,122,125,122,172,89,187,122,176,122,194,122,195,122,209,113,141,100,202,65,218,122,221,122,234,122,239,65,178,84,1,92,11,123,85,123,41,123,14,83,254,92,162,123,111,123,156,131,180,91,127,108,208,123,33,132,146,123,0,0,32,93,173,61,101,92,146,132,250,123,0,0,53,124,193,92,68,124,131,124,130,72,166,124,125,102,120,69,201,124,199,124,230,124,116,124,243,124,245,124,0,0,103,126,29,69,68,110,93,125,214,110,141,116,137,125,171,125,53,113,179,125,0,0,87,64,41,96,228,125,19,61,245,125,249,23,229,125,109,131,0,0,33,97,90,97,110,126,146,126,43,67,108,148,39,126,64,127,65,127,71,127,54,121,208,98,225,153,151,127,81,99,163,127,97,22,104,0,92,69,102,55,3,69,58,131,250,127,137,100,0,0,8,128,29,128,0,0,47,128,135,160,195,108,59,128,60,128,97,128,20,39,137,73,38,102,227,61,232,102,37,103,167,128,72,138,7,129,26,129,176,88,246,38,127,108,152,100,184,79,231,100,138,20,24,130,94,24,83,106,101,74,149,74,122,68,41,130,13,11,82,106,126,61,249,79,253,20,226,132,98,131,10,107,167,73,48,53,115,23,248,61,170,130,27,105,148,249,219,65,75,133,208,130,26,131,22,14,180,23,193,54,125,49,90,53,123,130,226,130,24,131,139,62,163,109,5,107,151,107,206,53,191,61,29,131,236,85,133,131,11,69,165,109,172,131,0,0,211,131,126,52,212,110,87,106,90,133,150,52,66,110,239,46,88,132,228,91,113,132,211,61,228,68,167,106,74,132,181,60,88,121,0,0,150,107,119,110,67,110,222,132,0,0,145,131,160,68,147,132,228,132,145,92,64,66,192,92,67,69,52,133,242,90,153,110,39,69,115,133,22,69,191,103,22,134,37,134,59,134,193,133,136,112,2,134,130,21,205,112,178,249,106,69,40,134,72,54,162,24,247,83,154,115,126,134,113,135,248,160,238,135,39,44,177,135,218,135,15,136,97,86,108,134,86,104,15,70,69,136,70,136,224,117,185,61,228,117,94,136,156,136,91,70,180,136,181,136,193,99,197,136,119,119,15,119,135,137,138,137,0,0,0,0,167,137,188,137,37,138,231,137,36,121,189,122,156,138,147,119,254,145,144,138,89,122,233,122,58,123,143,63,19,71,56,123,124,113,12,139,31,139,48,84,101,85,63,139,76,139,77,139,169,138,122,74,144,139,155,139,175,138,223,22,21,70,79,136,155,140,84,125,143,125,212,249,37,55,83,125,214,140,152,125,189,125,18,141,3,141,16,25,219,140,92,112,17,141,201,76,208,62,0,0,169,141,2,128,20,16,138,73,124,59,188,129,12,113,231,122,173,142,182,142,195,142,212,146,25,143,45,143,101,131,18,132,165,143,3,147,159,162,80,10,179,143,42,73,222,137,61,133,187,61,248,94,98,50,249,143,20,160,188,134,1,133,37,35,128,57,215,110,55,144,60,133,190,122,97,144,108,133,11,134,168,144,19,135,196,144,230,134,174,144,0,0,103,145,240,58,169,145,196,145,172,124,51,137,137,30,14,146,159,108,65,146,98,146,185,85,0,0,198,138,155,60,12,139,219,85,49,13,44,147,107,147,225,138,235,139,143,112,195,90,226,138,229,138,101,73,68,146,236,139,57,140,255,139,115,147,91,148,188,142,133,149,166,149,38,148,160,149,246,111,185,66,122,38,216,134,124,18,46,62,223,73,28,108,123,150,150,150,108,65,163,150,213,110,218,97,182,150,245,120,224,138,189,150,204,83,161,73,184,108,116,2,16,100,175,144,229,144,209,74,21,25,10,51,49,151,66,134,54,151,15,74,61,69,133,69,233,74,117,112,65,91,27,151,0,0,213,145,87,151,74,91,235,145,95,151,37,148,208,80,183,48,188,48,137,151,159,151,177,151,190,151,192,151,210,151,224,151,108,84,238,151,28,116,51,148,0,0,245,151,29,148,122,121,209,74,52,152,51,152,75,152,102,152,14,59,117,113,81,61,48,6,92,65,6,87,202,152,183,152,200,152,199,152,255,74,39,109,211,22,176,85,225,152,230,152,236,152,120,147,57,153,41,74,114,75,87,152,5,153,245,153,12,154,59,154,16,154,88,154,37,87,196,54,177,144,213,155,224,154,226,154,5,155,244,154,14,76,20,155,45,155,0,134,52,80,52,155,168,105,195,56,125,48,80,155,64,155,62,157,69,90,99,24,142,155,75,66,2,156,255,155,12,156,104,158,212,157,183,159,146,161,171,161,225,160,35,161,223,161,126,157,131,157,52,161,14,158,136,104,196,157,91,33,147,161,32,162,59,25,51,162,57,157,185,160,180,162,144,158,149,158,158,158,162,158,52,77,170,158,175,158,100,67,193,158,96,59,229,57,29,61,50,79,190,55,43,140,2,159,8,159,150,75,36,148,162,109,23,159,0,0,57,159,159,86,138,86,69,159,184,153,139,144,242,151,127,132,98,159,105,159,220,122,142,159,22,114,190,75,117,73,187,73,119,113,248,73,72,67,81,74,158,115,218,139,250,24,159,121,126,137,54,142,105,147,243,147,68,138,236,146,129,147,203,147,108,137,185,68,23,114,235,62,114,119,67,122,208,112,115,68,248,67,126,113,239,23,163,112,190,24,153,53,199,62,133,24,47,84,248,23,34,55,251,22,57,24,225,54,116,23,209,24,75,95,35,55,192,22,91,87,37,74,254,19,168,18,198,19,182,20,3,133,166,54,0,0,85,132,148,73,101,113,49,62,92,85,251,62,82,112,244,68,238,54,157,153,38,111,249,103,51,55,21,60,231,61,108,88,34,25,16,104,87,64,63,55,225,64,139,64,15,65,33,108,203,84,158,86,177,102,146,86,223,15,168,11,13,14,198,147,19,139,156,147,248,78,43,81,25,56,54,68,188,78,101,4,127,3,75,79,138,79,81,86,104,90,171,1,203,3,153,57,10,3,20,4,53,52,41,79,192,2,179,142,117,2,218,138,12,2,152,78,205,80,13,81,162,79,3,79,14,74,138,62,66,79,46,80,108,80,129,80,204,79,229,79,88,80,252,80,0,0,0,0,0,0,0,0,118,110,149,53,57,62,191,62,114,109,132,24,137,62,168,81,195,81,224,5,221,68,163,4,146,4,145,4,122,141,156,138,14,7,89,82,164,82,115,8,225,82,0,0,122,70,140,113,140,67,32,12,172,73,228,16,209,105,29,14,0,0,222,62,153,116,20,116,86,116,152,115,142,75,188,74,141,64,208,83,132,53,15,114,201,64,180,85,69,3,205,84,198,11,29,87,93,146,244,150,102,147,221,87,141,87,127,87,62,54,203,88,153,90,70,138,250,22,111,23,16,23,44,90,184,89,143,146,126,90,207,90,18,90,70,89,243,25,97,24,149,66,245,54,5,109,67,116,33,90,131,94,129,90,215,139,19,4,224,147,140,116,3,19,5,113,114,73,8,148,251,137,189,147,160,55,30,92,158,92,94,94,72,94,150,25,124,25,238,58,205,94,79,91,3,25,4,25,1,55,160,24,221,54,254,22,211,54,42,129,71,138,186,29,114,52,168,137,12,95,14,95,39,25,171,23,107,90,59,23,68,91,20,134,253,117,96,136,126,96,96,40,43,38,219,95,184,62,175,37,190,37,136,144,115,111,192,97,62,0,70,0,27,38,153,97,152,97,117,96,155,44,7,45,212,70,77,145,113,100,101,70,106,43,41,58,34,43,80,52,234,152,120,46,55,99,91,164,182,100,49,99,209,99,227,73,103,45,164,98,161,44,59,100,107,101,114,105,244,59,142,48,173,50,137,73,171,50,13,85,224,50,217,24,63,148,206,102,137,50,179,49,224,58,144,65,132,85,34,139,143,85,252,22,91,85,37,84,238,120,3,49,42,24,52,50,100,52,15,50,130,49,201,66,142,102,36,109,107,102,147,75,48,102,112,120,235,29,99,102,210,50,225,50,30,102,114,88,209,56,58,56,188,55,153,59,162,55,254,51,208,116,150,59,143,103,42,70,182,104,30,104,196,59,190,106,99,56,213,55,135,68,51,106,82,106,201,106,5,107,18,25,17,101,152,104,76,106,215,59,122,106,87,107,192,63,154,60,160,147,242,146,234,139,203,138,137,146,30,128,220,137,103,148,165,109,11,111,236,73,0,0,127,63,143,61,4,110,60,64,61,90,10,110,71,88,36,109,66,120,59,113,26,67,118,66,241,112,80,114,135,114,148,114,143,71,37,71,121,81,164,74,235,5,122,116,248,62,95,54,74,74,23,73,225,95,6,63,177,62,223,74,35,140,53,63,167,96,243,62,204,116,60,116,135,147,55,116,159,68,234,109,81,69,131,117,99,63,217,76,6,77,88,63,85,117,115,118,198,165,25,59,104,116,204,138,171,73,142,73,251,58,205,61,78,74,255,62,197,73,243,72,250,145,50,87,66,147,227,138,100,24,223,80,33,82,231,81,120,119,50,50,14,119,15,119,123,119,151,70,129,55,94,58,240,72,56,116,155,116,191,62,186,74,199,74,200,64,150,74,174,97,7,147,129,85,30,120,141,120,136,120,210,120,208,115,89,121,65,119,227,86,14,65,0,0,150,132,165,121,45,106,250,62,58,122,244,121,110,65,230,22,50,65,53,146,241,121,76,13,140,73,153,2,186,61,110,23,151,53,107,85,112,53,170,54,212,1,13,12,226,122,89,90,245,38,175,90,156,90,13,90,91,2,240,120,42,90,198,91,254,122,249,65,93,124,109,124,17,66,179,91,188,94,166,94,205,124,249,73,176,23,142,124,124,124,174,124,178,106,220,125,7,126,211,125,78,127,97,98,92,97,72,123,151,125,130,94,106,66,117,107,22,9,214,103,78,0,207,53,196,87,18,100,248,99,98,73,221,127,39,123,44,8,233,90,67,93,12,123,14,94,230,153,69,134,99,154,28,106,63,52,226,57,247,73,173,101,31,154,160,101,128,132,39,113,209,108,234,68,55,129,2,68,198,128,9,129,66,129,180,103,195,152,66,106,98,130,101,130,81,106,83,132,167,109,16,134,27,114,134,90,127,65,64,24,43,91,161,24,228,90,216,24,160,134,188,249,143,61,45,136,34,116,2,90,110,136,69,79,135,136,191,136,230,136,101,137,77,137,131,86,84,137,133,119,132,119,245,139,217,139,156,139,249,137,173,62,163,132,245,70,207,70,242,55,61,138,28,138,72,148,77,95,43,146,132,66,212,101,41,113,196,112,69,24,109,157,159,140,233,140,220,125,154,89,195,119,240,89,110,67,212,54,42,142,167,142,9,76,48,143,74,143,244,66,88,108,187,111,33,35,155,72,121,111,139,110,218,23,233,155,181,54,47,73,187,144,0,0,113,85,6,73,187,145,4,148,75,138,98,64,252,138,39,148,29,140,59,140,229,132,43,138,153,149,167,149,151,149,150,149,52,141,69,116,194,62,255,72,66,74,234,67,231,62,37,50,143,150,231,142,102,142,101,142,204,62,237,73,120,74,238,63,18,116,107,116,252,62,65,151,176,144,71,104,29,74,147,144,223,87,0,0,104,147,137,137,38,140,47,139,190,99,186,146,17,91,105,139,60,73,249,115,27,66,155,151,113,151,56,153,38,15,193,93,197,139,178,74,31,152,218,148,246,146,215,149,229,145,192,68,80,139,103,74,100,139,220,152,69,138,0,63,42,146,37,73,20,132,59,153,77,153,6,123,253,61,155,153,111,75,170,153,92,154,101,139,200,88,143,106,33,154,254,90,47,154,241,152,144,75,72,153,188,153,189,75,151,75,125,147,114,88,2,19,34,88,184,73,232,20,68,120,31,39,184,61,197,104,125,61,88,148,39,57,80,97,129,39,107,41,7,97,79,156,83,156,123,156,53,156,16,156,127,155,207,155,45,158,159,155,245,161,254,160,33,157,174,76,4,65,24,158,176,76,12,157,180,161,237,160,243,160,47,153,165,157,189,132,18,110,223,111,130,107,252,133,51,69,164,109,132,110,240,109,32,132,238,133,0,110,215,55,100,96,226,121,156,53,64,54,45,73,222,73,98,61,219,147,190,146,72,147,191,2,185,120,119,146,77,148,228,79,64,52,100,144,93,85,61,120,84,120,182,120,75,120,87,23,201,49,65,73,154,54,114,79,218,111,217,111,0,0,30,112,20,84,181,65,187,87,243,88,138,87,22,157,215,87,52,113,175,52,172,65,235,113,64,108,151,79,0,0,181,23,73,138,12,97,206,90,11,90,188,66,136,68,44,55,123,75,252,137,187,147,184,147,214,24,29,15,114,132,192,108,19,20,250,66,38,44,193,67,148,89,183,61,65,103,168,125,91,97,164,96,185,73,139,73,250,137,229,146,226,115,233,62,180,116,99,139,159,24,225,62,179,74,216,106,243,115,251,115,214,62,62,74,148,74,217,23,102,74,167,3,36,20,229,73,72,116,22,73,165,112,118,73,132,146,230,115,95,147,254,4,49,147,206,138,22,138,134,147,231,139,213,85,53,73,130,138,107,113,67,73,255,12,164,86,26,6,235,11,184,12,2,85,196,121,250,23,254,125,194,22,80,74,82,24,46,69,1,148,10,55,192,138,173,73,176,89,191,24,131,24,132,116,161,90,226,54,91,61,176,54,95,146,121,90,129,138,98,24,116,147,205,60,180,10,150,74,138,57,244,80,105,61,76,61,156,19,117,113,251,66,24,130,15,110,228,144,235,68,87,109,79,126,103,112,175,108,214,60,237,63,45,62,2,110,12,111,111,61,245,3,81,117,188,54,200,52,128,70,218,62,113,72,196,89,110,146,62,73,65,143,28,140,192,107,18,88,200,87,214,54,82,20,254,112,98,67,113,74,227,47,176,18,189,35,185,104,103,105,152,19,229,52,244,123,223,54,131,138,214,55,250,51,159,76,26,106,173,54,183,108,62,132,223,68,206,68,38,109,81,109,130,108,222,111,23,111,9,113,61,131,58,23,237,131,128,108,83,112,219,23,137,89,130,90,179,23,97,90,113,90,5,25,252,65,45,55,239,89,60,23,199,54,142,113,144,147,154,102,165,66,110,90,43,90,147,66,43,106,249,62,54,119,91,68,202,66,29,113,89,66,225,137,176,79,40,109,194,92,206,68,77,126,189,67,12,106,86,66,4,19,166,112,51,113,233,67,165,61,223,108,37,248,79,74,101,126,235,89,47,93,243,61,92,95,93,74,223,23,164,125,38,132,133,84,250,58,0,51,20,2,126,87,213,8,25,6,229,63,158,31,182,162,3,112,91,145,112,93,143,115,211,124,89,138,32,148,200,79,231,127,205,114,16,115,244,122,56,115,57,115,246,86,65,115,72,115,169,62,24,123,108,144,245,113,242,72,225,115,246,129,202,62,12,119,209,62,162,108,253,86,25,116,30,116,31,116,226,62,240,62,244,62,250,62,211,116,14,63,83,63,66,117,109,117,114,117,141,117,124,63,200,117,220,117,192,63,77,118,215,63,116,118,220,63,122,118,92,79,136,113,35,86,128,137,105,88,29,64,67,119,57,64,97,103,69,64,219,53,152,119,106,64,111,64,94,92,190,119,203,119,242,88,24,120,185,112,28,120,168,64,57,120,71,120,81,120,102,120,72,132,53,85,51,121,3,104,50,121,3,65,9,65,145,121,153,121,187,143,6,122,188,143,103,65,145,122,178,65,188,122,121,130,196,65,207,122,219,122,207,65,33,78,98,123,108,123,123,123,18,124,27,124,96,66,122,66,123,124,156,124,140,66,184,124,148,66,237,124,147,143,192,112,207,12,207,125,212,125,208,125,253,125,174,127,180,127,159,114,151,67,32,128,37,128,57,123,46,128,49,128,84,128,204,61,180,87,160,112,183,128,233,128,237,67,12,129,42,115,14,129,18,129,96,117,20,129,1,68,57,59,86,129,89,129,90,129,19,68,58,88,124,129,132,129,37,68,147,129,45,68,165,129,239,87,193,129,228,129,84,130,143,68,166,130,118,130,202,130,216,130,255,130,176,68,87,131,105,150,138,105,5,132,245,112,100,132,227,96,136,132,4,69,190,132,225,132,248,132,16,133,56,133,82,133,59,69,111,133,112,133,224,133,119,69,114,134,146,134,178,134,239,134,69,150,139,135,6,70,23,70,174,136,255,136,36,137,71,137,145,137,103,121,41,138,56,138,148,138,180,138,81,140,212,140,242,140,28,141,152,71,95,88,195,141,237,71,238,78,58,142,216,85,84,87,113,142,245,85,176,142,55,72,206,142,226,142,228,142,237,142,242,142,183,143,193,143,202,143,204,143,51,144,196,153,173,72,224,152,19,146,30,73,40,146,88,146,107,146,177,146,174,146,191,146,227,146,235,146,243,146,244,146,253,146,67,147,132,147,173,147,69,73,81,73,191,158,23,148,1,83,29,148,45,148,62,148,106,73,84,148,121,148,45,149,162,149,167,73,244,149,51,150,229,73,160,103,36,74,64,151,53,74,178,151,194,151,84,86,228,74,232,96,185,152,25,75,241,152,68,88,14,153,25,153,180,81,28,153,55,153,66,153,93,153,98,153,112,75,197,153,157,75,60,154,15,155,131,122,105,155,129,155,221,155,241,155,244,155,109,76,32,156,111,55,194,27,73,157,58,156,254,158,80,86,147,157,189,157,192,157,252,157,246,148,182,143,123,158,172,158,177,158,189,158,198,158,220,148,226,158,241,158,248,158,200,122,68,159,148,0,183,2,160,3,26,105,195,148,172,89,215,4,64,88,193,148,185,55,213,5,21,6,118,6,186,22,87,87,115,113,194,10,205,10,191,11,106,84,59,248,203,11,158,84,251,11,59,12,83,12,101,12,124,12,231,96,141,12,122,86,181,12,221,12,237,12,111,13,178,13,200,13,85,105,47,156,165,135,4,14,14,14,215,14,144,15,45,15,115,14,32,92,188,15,11,94,92,16,79,16,118,16,30,103,123,16,136,16,150,16,71,54,191,16,211,16,47,17,59,17,100,83,173,132,227,18,117,19,54,19,129,139,119,21,25,22,195,23,199,23,120,78,187,112,45,24,106,25,45,26,69,26,42,28,112,28,172,28,200,30,195,98,213,30,21,31,152,113,85,104,69,32,233,105,200,54,124,34,215,35,250,35,42,39,113,40,79,41,253,130,103,41,147,41,213,42,165,137,232,42,160,143,14,43,184,151,63,43,71,152,189,154,76,44,0,0,136,44,183,44,232,91,8,45,18,45,183,45,149,45,66,46,116,47,204,47,51,48,102,48,31,51,222,51,177,95,72,102,191,102,121,122,103,53,243,53,0,0,186,73,0,0,26,54,22,55,0,0,70,3,181,88,14,103,24,105,167,58,87,118,226,95,17,62,185,62,254,117,154,32,208,72,184,74,25,65,154,138,238,66,13,67,59,64,52,67,150,67,69,74,202,5,210,81,17,6,159,89,168,30,190,59,255,60,4,68,214,68,136,87,116,70,155,57,47,71,232,133,201,153,98,55,195,33,94,139,78,139,0,0,18,72,251,72,21,74,9,114,192,74,120,12,101,89,165,78,134,79,121,7,218,142,44,80,143,82,63,87,113,113,153,82,25,84,74,63,167,74,188,85,70,84,110,84,82,107,0,0,115,52,63,85,50,118,94,85,24,71,98,85,102,85,199,87,63,73,93,88,102,80,251,52,204,51,0,0,3,89,124,71,72,137,174,90,137,91,6,92,144,29,161,87,81,113,0,0,2,97,18,124,86,144,178,97,154,79,98,139,2,100,74,100,91,93,247,107,0,0,132,100,28,25,234,138,246,73,136,100,239,63,18,101,192,75,191,101,181,102,27,39,101,148,225,87,149,97,39,90,205,248,0,0,185,86,33,69,252,102,106,78,52,73,86,150,143,109,189,108,24,54,119,137,153,103,110,104,17,100,94,104,0,0,199,104,66,123,192,144,17,10,38,105,0,0,57,105,69,122,0,0,250,105,38,154,45,106,95,54,105,100,33,0,131,121,52,106,91,107,44,93,25,53,0,0,157,107,208,70,164,108,59,117,101,136,174,109,182,88,28,55,141,37,75,112,205,113,84,60,128,114,133,114,129,146,122,33,139,114,48,147,230,114,208,73,57,108,159,148,80,116,248,14,39,136,245,136,38,41,115,132,177,23,184,110,42,74,32,24,164,57,185,54,0,0,0,0,63,69,182,102,173,156,164,152,67,137,204,119,88,120,214,86,223,64,10,22,161,57,47,55,232,128,197,19,173,113,102,131,221,121,168,145,0,0,183,76,175,112,171,137,253,121,10,122,11,123,102,125,122,65,67,123,126,121,9,128,181,111,223,162,3,106,24,131,162,83,7,110,191,147,54,104,93,151,111,129,35,128,181,105,237,19,47,50,72,128,133,93,48,140,131,128,21,87,35,152,73,137,171,93,136,73,190,101,213,105,210,83,165,74,129,63,17,60,54,103,144,128,244,128,46,129,161,31,79,129,137,129,175,129,26,130,6,131,47,131,138,131,202,53,104,132,170,134,250,72,230,99,86,137,8,120,85,146,184,137,242,67,231,137,223,67,232,137,70,139,212,139,248,89,9,140,0,0,197,143,236,144,0,0,16,145,60,145,247,61,94,145,202,74,208,143,143,114,139,86,231,148,233,149,176,149,184,149,50,151,209,152,73,153,106,153,195,153,40,154,14,155,90,157,155,157,159,126,248,158,35,159,164,76,71,149,147,162,162,113,255,162,145,77,18,144,203,165,156,77,156,12,190,143,193,85,186,143,176,36,185,143,147,74,9,69,127,126,86,111,177,106,234,78,228,52,44,139,157,120,58,55,128,142,245,23,36,128,108,139,153,139,62,122,175,102,235,61,85,118,183,60,53,86,86,89,154,78,129,94,88,98,191,86,109,14,14,142,109,91,136,62,158,76,222,99,0,0,246,23,123,24,48,101,45,86,74,92,26,84,17,83,198,61,152,157,125,76,34,86,30,86,73,127,216,94,117,89,64,61,112,135,28,78,234,15,73,13,186,54,23,129,94,157,24,141,59,118,69,156,78,118,185,119,69,147,50,84,72,129,247,130,37,86,50,129,24,132,189,128,234,85,98,121,67,86,22,84,157,14,206,53,5,86,241,85,241,102,226,130,45,54,52,117,240,85,186,85,151,84,114,85,65,12,150,12,208,94,72,81,118,14,98,44,162,14,171,158,90,125,222,85,117,16,157,98,109,151,148,84,205,140,246,113,118,145,252,99,185,99,254,99,105,85,67,43,114,156,179,46,154,81,223,52,167,13,167,81,77,84,30,85,19,85,102,118,45,142,138,104,177,117,182,128,4,136,134,135,199,136,182,129,28,132,193,16,236,68,4,115,6,71,144,91,11,131,147,104,123,86,244,38,47,125,163,65,115,125,208,110,182,114,112,145,217,17,8,146,252,60,169,166,172,14,249,14,102,114,162,28,78,71,194,79,249,127,235,15,250,64,93,156,31,101,160,45,243,72,224,71,124,157,236,15,10,14,0,0,163,117,237,15,0,0,72,96,135,17,163,113,142,126,80,157,26,78,4,78,119,53,13,91,178,108,103,83,172,54,220,57,125,83,165,54,24,70,154,88,110,75,45,130,75,84,170,87,149,90,121,9,0,0,82,58,101,36,116,115,172,158,9,77,237,155,254,60,48,159,91,76,169,79,158,149,222,159,92,132,182,61,178,114,179,103,32,55,46,99,37,125,247,62,44,62,42,58,8,144,204,82,116,62,122,54,233,69,142,4,64,118,240,90,182,14,122,120,46,127,167,88,191,64,124,86,139,155,116,93,84,118,52,164,133,158,225,76,0,0,251,55,25,97,218,48,242,67,0,0,93,86,169,18,167,87,99,73,6,158,52,82,174,112,173,53,0,0,124,157,86,124,57,155,222,87,108,23,83,92,211,100,208,148,53,99,100,113,173,134,40,13,34,109,226,74,113,13,0,0,254,81,15,31,142,93,3,151,209,29,129,158,76,144,31,123,2,155,209,92,163,123,104,98,53,99,255,154,207,123,42,155,126,124,0,0,66,124,134,124,21,156,252,123,9,155,0,0,27,156,62,73,90,159,115,85,195,91,253,79,152,158,242,79,96,82,6,62,209,82,103,87,86,80,183,89,18,94,200,151,171,157,92,143,105,84,180,151,64,153,186,151,44,83,48,97,44,105,218,83,10,156,2,157,59,76,65,150,128,105,166,80,70,117,109,23,218,153,115,82,0,0,89,145,129,150,92,145,0,0,81,145,151,142,127,99,35,109,202,106,17,86,142,145,122,117,133,98,252,3,79,115,112,124,33,92,253,60,0,0,25,73,214,118,157,155,42,78,212,12,190,131,66,136,0,0,74,92,192,105,0,0,122,87,31,82,245,93,206,78,49,108,242,1,57,79,156,84,218,84,154,82,130,141,254,53,0,0,243,53,0,0,82,107,124,145,165,159,151,155,46,152,180,152,186,154,168,158,132,158,122,113,20,123,0,0,250,107,24,136,120,127,0,0,32,86,74,166,119,142,83,159,0,0,212,141,79,142,28,158,1,142,130,98,125,131,40,142,117,142,211,122,119,74,62,122,216,120,234,108,103,138,7,118,90,138,38,159,206,108,214,135,195,117,178,162,83,120,64,248,12,141,226,114,113,115,45,139,2,115,241,116,235,140,187,74,47,134,186,95,160,136,183,68,0,0,59,24,5,110,0,0,126,138,27,37,0,0,253,96,103,118,215,154,68,157,110,147,143,155,245,135,0,0,0,0,247,140,44,115,33,151,176,155,214,53,178,114,7,76,81,124,74,153,89,97,89,97,4,76,150,158,125,97,0,0,95,87,111,97,166,98,57,98,0,0,92,58,226,97,170,83,245,51,100,99,2,104,210,53,87,93,194,139,218,143,57,142,0,0,217,80,70,29,6,121,50,83,56,150,59,15,101,64,0,0,254,119,0,0,194,124,26,95,218,124,45,122,102,128,99,128,77,125,5,117,242,116,148,137,26,130,12,103,98,128,134,116,91,128,240,116,3,129,36,119,137,137,204,103,83,117,209,110,169,135,206,135,200,129,140,135,73,138,173,140,67,139,43,119,248,116,218,132,53,54,178,105,166,141,0,0,169,137,0,0,185,109,193,135,17,64,231,116,219,61,118,113,164,96,156,97,209,60,0,0,119,96,0,0,113,127,45,139,0,0,233,96,126,75,32,82,24,60,199,60,215,94,86,118,49,85,68,25,254,18,3,153,220,109,173,112,193,92,173,97,15,138,119,54,238,0,70,104,14,79,98,69,31,91,76,99,80,159,166,158,107,98,71,5,219,146,223,5,197,63,76,133,181,66,239,115,181,81,73,54,66,73,228,137,68,147,219,25,238,130,200,60,60,120,68,103,223,98,51,73,170,137,160,2,179,107,5,19,171,79,237,36,8,80,41,109,132,122,0,54,177,74,19,37,0,0,126,3,164,95,128,3,71,3,219,110,31,4,0,0,1,81,122,52,14,81,108,152,67,55,22,132,164,73,135,4,96,81,180,51,106,81,255,11,252,32,229,2,48,37,142,5,51,50,131,25,130,91,125,135,179,5,153,60,178,81,184,81,52,157,201,81,207,81,209,81,220,60,211,81,166,74,179,81,226,81,66,83,237,81,205,131,62,105,45,55,123,95,11,82,38,82,60,82,181,82,87,82,148,82,185,82,197,82,21,124,66,133,224,82,13,134,19,107,0,0,222,138,73,85,217,110,128,63,84,9,236,63,51,83,0,0,226,11,203,108,38,23,27,104,213,115,74,96,170,62,204,56,232,22,221,113,162,68,109,83,116,83,171,134,126,83,0,0,150,21,19,22,230,119,147,83,155,138,160,83,171,83,174,83,167,115,114,87,89,63,156,115,193,83,197,83,73,108,73,78,254,87,217,83,171,58,143,11,224,83,235,63,163,45,246,83,119,12,19,84,121,112,43,85,87,102,91,109,109,84,83,107,116,13,93,85,143,84,164,84,166,71,13,23,221,14,180,61,77,13,188,137,152,38,71,85,237,76,47,84,23,116,134,85,169,85,0,0,215,24,58,64,82,69,53,68,179,102,180,16,55,86,205,102,138,50,164,102,173,102,77,86,79,86,241,120,241,86,135,151,254,83,0,87,239,86,237,86,102,139,35,54,79,18,70,87,165,65,110,108,139,112,66,87,177,54,126,108,230,87,22,20,3,88,84,20,99,67,38,88,245,75,92,88,170,88,97,53,224,88,220,88,60,18,251,88,255,91,67,87,80,161,120,66,211,147,161,53,31,89,166,104,195,54,89,110,62,22,36,90,83,85,146,22,5,133,201,89,78,13,129,108,42,109,220,23,217,89,251,23,178,23,166,109,113,109,40,24,213,22,249,89,69,110,171,90,99,90,230,54,169,73,0,0,8,55,150,90,101,116,211,90,161,111,84,37,133,61,17,25,50,55,184,22,131,94,208,82,118,91,136,101,124,91,14,122,4,64,93,72,4,2,213,91,96,97,52,26,204,89,165,5,243,91,157,91,16,77,5,92,68,27,19,92,206,115,20,92,165,28,40,107,73,92,221,72,133,92,233,92,239,92,139,93,249,29,55,30,16,93,24,93,70,93,164,30,186,92,215,93,252,130,45,56,1,73,73,32,115,33,135,130,54,56,194,59,46,94,138,106,0,0,122,94,188,68,211,12,166,83,183,78,0,0,168,83,113,23,9,94,244,94,130,132,249,94,251,94,160,56,252,94,62,104,27,148,13,95,193,1,148,248,222,58,174,72,58,19,58,95,136,104,208,35,0,0,113,36,99,95,189,151,110,110,114,95,64,147,54,138,167,95,182,93,95,61,80,82,106,31,248,112,104,38,214,145,158,2,41,138,49,96,133,102,119,24,99,57,199,61,57,54,144,87,180,39,113,121,64,62,158,96,0,0,179,96,130,73,143,73,83,122,164,116,225,80,160,90,100,97,36,132,66,97,166,248,210,110,129,97,244,81,86,6,135,97,170,91,183,63,95,40,211,97,157,139,93,153,208,97,50,57,128,41,193,40,35,96,92,97,30,101,139,99,24,1,197,98,112,23,213,98,13,46,108,99,223,73,23,58,56,100,248,99,142,19,252,23,0,0,138,111,54,46,20,152,140,64,29,87,225,100,229,100,123,148,102,58,58,100,87,58,77,101,22,111,40,74,35,74,133,101,109,101,95,101,126,48,181,101,64,73,55,75,209,101,216,64,41,24,224,101,227,101,223,95,0,52,24,102,247,49,248,49,68,102,164,49,165,49,75,102,117,14,103,102,230,81,115,102,0,0,61,30,49,50,244,133,200,49,19,83,197,119,247,40,164,153,2,103,156,67,33,74,43,59,250,105,194,55,0,0,103,103,98,103,205,65,237,144,215,103,233,68,34,104,80,110,60,146,1,104,230,51,160,109,93,104,111,52,225,105,11,106,223,138,115,105,195,104,205,53,1,105,0,105,50,61,1,58,60,54,128,59,172,103,97,105,74,138,252,66,54,105,152,105,161,59,201,3,99,131,144,80,249,105,89,54,42,33,69,106,3,55,157,106,243,59,177,103,200,106,156,145,13,60,29,107,35,9,222,96,53,107,116,107,205,39,181,110,219,58,181,3,88,25,64,55,33,84,90,59,225,107,252,62,220,107,55,108,139,36,241,72,81,107,90,108,38,130,121,108,188,61,197,68,189,61,164,65,12,73,0,73,201,60,229,54,235,60,50,13,131,155,249,49,145,36,143,127,55,104,37,109,161,109,235,109,150,109,92,109,124,110,4,111,127,73,133,64,114,110,51,133,116,111,199,81,0,0,0,0,46,132,33,139,0,0,47,62,83,116,130,63,204,121,79,110,145,90,75,48,248,111,13,55,157,111,48,62,250,110,151,20,61,64,85,69,240,147,68,111,92,111,78,61,116,111,112,145,59,61,159,111,68,65,211,111,145,64,85,65,57,64,240,63,180,63,63,65,223,81,86,65,87,65,64,65,221,97,75,112,126,112,167,112,129,112,204,112,213,112,214,112,223,112,4,65,232,61,180,113,150,113,119,66,43,113,69,113,136,90,74,113,0,0,156,92,101,67,79,113,98,147,193,66,44,113,90,68,39,74,34,74,186,113,232,139,189,112,14,114,66,148,21,114,17,89,67,148,36,114,65,147,5,86,46,114,64,114,116,73,189,104,85,114,87,114,85,62,68,48,13,104,61,111,130,114,0,0,43,115,35,72,43,136,237,72,4,136,40,115,46,115,207,115,170,115,58,12,46,106,201,115,73,116,226,65,231,22,36,74,35,102,197,54,183,73,141,73,251,73,247,115,21,116,3,105,38,74,57,116,195,5,215,62,0,0,173,40,96,116,178,142,71,116,228,115,118,116,185,131,108,116,48,55,116,116,241,147,44,106,130,116,83,73,140,74,95,65,121,74,143,139,70,91,3,140,158,24,200,116,136,25,14,117,0,0,30,117,217,142,75,26,215,91,172,142,133,147,77,117,74,117,103,117,110,117,130,79,4,63,19,77,142,117,93,116,158,117,180,117,2,118,44,118,81,118,79,118,111,118,118,118,245,99,144,118,239,129,248,55,17,105,14,105,161,118,165,118,183,118,204,118,159,111,98,132,157,80,125,81,28,30,30,119,38,119,64,119,175,100,32,82,88,119,172,50,175,119,100,137,104,137,193,22,244,119,0,0,118,19,18,74,202,104,175,120,199,120,211,120,165,150,46,121,224,85,215,120,52,121,177,120,12,118,184,143,132,136,43,139,131,96,28,38,134,121,0,137,2,105,128,121,87,88,157,121,57,123,60,121,169,121,42,110,38,113,168,62,198,121,13,145,212,121,32,5,95,68,15,82,130,184,248,0,0,64,32,169,78,4,0,0,0,0,52,11,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,64,0,0,0,0,0,0,0,0,4,88,0,60,0,0,0,0,0,128,92,173,220,243,187,201,67,238,237,164,15,38,193,155,118,242,239,222,251,250,247,15,50,68,175,254,239,253,254,17,96,176,0,0,33,25,129,32,16,136,168,96,33,105,36,0,68,137,196,0,0,3,64,53,0,67,132,49,81,147,104,2,2,0,0,0,0,0,0,0,0,0,0,0,64,0,1,128,172,100,130,64,140,161,144,4,80,96,212,0,210,42,24,157,104,53,151,79,216,248,32,196,130,28,40,130,117,148,2,56,64,160,161,197,16,113,77,196,13,152,185,94,91,1,67,159,153,3,120,32,226,31,8,137,1,201,64,7,3,13,154,144,82,209,79,124,3,2,33,32,192,110,19,131,181,219,245,229,119,144,79,23,154,54,23,12,0,110,186,9,47,96,84,0,30,120,104,34,112,86,158,37,243,142,183,23,11,239,230,119,23,91,27,193,16,152,141,18,24,182,207,1,49,114,26,91,44,182,192,32,226,195,92,130,27,224,218,206,133,214,64,169,84,11,208,15,72,68,64,16,9,212,64,174,59,100,30,68,43,23,8,134,130,54,24,160,137,231,156,91,8,132,3,2,218,144,236,219,135,133,191,162,173,157,142,157,152,32,37,70,0,70,159,127,227,213,201,132,244,214,77,6,227,182,24,18,9,0,0,0,4,0,0,32,0,0,0,0,2,0,0,0,0,0,0,0,0,0,32,0,0,0,0,0,0,0,0,0,0,0,0,8,0,0,142,120,110,95,63,126,221,61,247,252,155,252,174,210,255,143,107,241,231,191,186,174,110,23,239,220,167,62,182,181,119,201,146,215,165,182,74,110,155,185,129,77,103,254,84,241,155,98,254,111,82,215,54,254,191,41,69,129,249,189,101,166,66,57,0,0,66,96,143,0,148,64,64,210,175,215,209,219,0,160,152,178,123,6,82,32,152,22,200,61,9,0,0,1,0,0,0,1,10,178,8,128,0,0,0,4,136,32,20,16,76,0,0,4,0,161,68,16,0,65,1,0,8,64,240,223,79,54,88,241,245,181,128,245,103,128,64,0,80,78,65,200,132,0,90,96,216,1,43,1,80,161,22,196,64,242,110,17,172,32,57,196,192,8,7,12,9,152,150,196,155,16,28,152,108,134,42,166,1,70,17,109,229,203,196,192,146,16,33,22,18,157,114,244,105,14,23,42,162,129,244,123,0,8,164,11,144,16,44,204,57,74,1,240,22,11,10,64,12,29,234,12,68,14,69,2,0,0,48,12,255,1,48,2,48,14,255,39,32,27,255,26,255,31,255,1,255,48,254,38,32,37,32,80,254,81,254,82,254,183,0,84,254,85,254,86,254,87,254,92,255,19,32,49,254,20,32,51,254,116,37,52,254,79,254,8,255,9,255,53,254,54,254,91,255,93,255,55,254,56,254,20,48,21,48,57,254,58,254,16,48,17,48,59,254,60,254,10,48,11,48,61,254,62,254,8,48,9,48,63,254,64,254,12,48,13,48,65,254,66,254,14,48,15,48,67,254,68,254,89,254,90,254,91,254,92,254,93,254,94,254,24,32,25,32,28,32,29,32,29,48,30,48,53,32,50,32,3,255,6,255,10,255,59,32,167,0,3,48,203,37,207,37,179,37,178,37,206,37,6,38,5,38,199,37,198,37,161,37,160,37,189,37,188,37,163,50,5,33,175,0,227,255,63,255,205,2,73,254,74,254,77,254,78,254,75,254,76,254,95,254,96,254,97,254,11,255,13,255,215,0,247,0,177,0,26,34,28,255,30,255,29,255,102,34,103,34,96,34,30,34,82,34,97,34,98,254,99,254,100,254,101,254,102,254,94,255,41,34,42,34,165,34,32,34,31,34,191,34,210,51,209,51,43,34,46,34,53,34,52,34,64,38,66,38,149,34,153,34,145,33,147,33,144,33,146,33,150,33,151,33,153,33,152,33,37,34,35,34,15,255,60,255,21,34,104,254,4,255,229,255,18,48,224,255,225,255,5,255,32,255,3,33,9,33,105,254,106,254,107,254,213,51,156,51,157,51,158,51,206,51,161,51,142,51,143,51,196,51,176,0,89,81,91,81,94,81,93,81,97,81,99,81,231,85,233,116,206,124,129,37,130,37,131,37,132,37,133,37,134,37,135,37,136,37,143,37,142,37,141,37,140,37,139,37,138,37,137,37,60,37,52,37,44,37,36,37,28,37,148,37,0,37,2,37,149,37,12,37,16,37,20,37,24,37,109,37,110,37,112,37,111,37,80,37,94,37,106,37,97,37,226,37,227,37,229,37,228,37,113,37,114,37,115,37,16,255,17,255,18,255,19,255,20,255,21,255,22,255,23,255,24,255,25,255,96,33,97,33,98,33,99,33,100,33,101,33,102,33,103,33,104,33,105,33,33,48,34,48,35,48,36,48,37,48,38,48,39,48,40,48], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+87328);
/* memory initializer */ allocate([41,48,65,83,68,83,69,83,33,255,34,255,35,255,36,255,37,255,38,255,39,255,40,255,41,255,42,255,43,255,44,255,45,255,46,255,47,255,48,255,49,255,50,255,51,255,52,255,53,255,54,255,55,255,56,255,57,255,58,255,65,255,66,255,67,255,68,255,69,255,70,255,71,255,72,255,73,255,74,255,75,255,76,255,77,255,78,255,79,255,80,255,81,255,82,255,83,255,84,255,85,255,86,255,87,255,88,255,89,255,90,255,145,3,146,3,147,3,148,3,149,3,150,3,151,3,152,3,153,3,154,3,155,3,156,3,157,3,158,3,159,3,160,3,161,3,163,3,164,3,165,3,166,3,167,3,168,3,169,3,177,3,178,3,179,3,180,3,181,3,182,3,183,3,184,3,185,3,186,3,187,3,188,3,189,3,190,3,191,3,192,3,193,3,195,3,196,3,197,3,198,3,199,3,200,3,201,3,5,49,6,49,7,49,8,49,9,49,10,49,11,49,12,49,13,49,14,49,15,49,16,49,17,49,18,49,19,49,20,49,21,49,22,49,23,49,24,49,25,49,26,49,27,49,28,49,29,49,30,49,31,49,32,49,33,49,34,49,35,49,36,49,37,49,38,49,39,49,40,49,41,49,217,2,201,2,202,2,199,2,203,2,0,36,1,36,2,36,3,36,4,36,5,36,6,36,7,36,8,36,9,36,10,36,11,36,12,36,13,36,14,36,15,36,16,36,17,36,18,36,19,36,20,36,21,36,22,36,23,36,24,36,25,36,26,36,27,36,28,36,29,36,30,36,31,36,33,36,172,32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,78,89,78,1,78,3,78,67,78,93,78,134,78,140,78,186,78,63,81,101,81,107,81,224,81,0,82,1,82,155,82,21,83,65,83,92,83,200,83,9,78,11,78,8,78,10,78,43,78,56,78,225,81,69,78,72,78,95,78,94,78,142,78,161,78,64,81,3,82,250,82,67,83,201,83,227,83,31,87,235,88,21,89,39,89,115,89,80,91,81,91,83,91,248,91,15,92,34,92,56,92,113,92,221,93,229,93,241,93,242,93,243,93,254,93,114,94,254,94,11,95,19,95,77,98,17,78,16,78,13,78,45,78,48,78,57,78,75,78,57,92,136,78,145,78,149,78,146,78,148,78,162,78,193,78,192,78,195,78,198,78,199,78,205,78,202,78,203,78,196,78,67,81,65,81,103,81,109,81,110,81,108,81,151,81,246,81,6,82,7,82,8,82,251,82,254,82,255,82,22,83,57,83,72,83,71,83,69,83,94,83,132,83,203,83,202,83,205,83,236,88,41,89,43,89,42,89,45,89,84,91,17,92,36,92,58,92,111,92,244,93,123,94,255,94,20,95,21,95,195,95,8,98,54,98,75,98,78,98,47,101,135,101,151,101,164,101,185,101,229,101,240,102,8,103,40,103,32,107,98,107,121,107,203,107,212,107,219,107,15,108,52,108,107,112,42,114,54,114,59,114,71,114,89,114,91,114,172,114,139,115,25,78,22,78,21,78,20,78,24,78,59,78,77,78,79,78,78,78,229,78,216,78,212,78,213,78,214,78,215,78,227,78,228,78,217,78,222,78,69,81,68,81,137,81,138,81,172,81,249,81,250,81,248,81,10,82,160,82,159,82,5,83,6,83,23,83,29,83,223,78,74,83,73,83,97,83,96,83,111,83,110,83,187,83,239,83,228,83,243,83,236,83,238,83,233,83,232,83,252,83,248,83,245,83,235,83,230,83,234,83,242,83,241,83,240,83,229,83,237,83,251,83,219,86,218,86,22,89,46,89,49,89,116,89,118,89,85,91,131,91,60,92,232,93,231,93,230,93,2,94,3,94,115,94,124,94,1,95,24,95,23,95,197,95,10,98,83,98,84,98,82,98,81,98,165,101,230,101,46,103,44,103,42,103,43,103,45,103,99,107,205,107,17,108,16,108,56,108,65,108,64,108,62,108,175,114,132,115,137,115,220,116,230,116,24,117,31,117,40,117,41,117,48,117,49,117,50,117,51,117,139,117,125,118,174,118,191,118,238,118,219,119,226,119,243,119,58,121,190,121,116,122,203,122,30,78,31,78,82,78,83,78,105,78,153,78,164,78,166,78,165,78,255,78,9,79,25,79,10,79,21,79,13,79,16,79,17,79,15,79,242,78,246,78,251,78,240,78,243,78,253,78,1,79,11,79,73,81,71,81,70,81,72,81,104,81,113,81,141,81,176,81,23,82,17,82,18,82,14,82,22,82,163,82,8,83,33,83,32,83,112,83,113,83,9,84,15,84,12,84,10,84,16,84,1,84,11,84,4,84,17,84,13,84,8,84,3,84,14,84,6,84,18,84,224,86,222,86,221,86,51,87,48,87,40,87,45,87,44,87,47,87,41,87,25,89,26,89,55,89,56,89,132,89,120,89,131,89,125,89,121,89,130,89,129,89,87,91,88,91,135,91,136,91,133,91,137,91,250,91,22,92,121,92,222,93,6,94,118,94,116,94,15,95,27,95,217,95,214,95,14,98,12,98,13,98,16,98,99,98,91,98,88,98,54,101,233,101,232,101,236,101,237,101,242,102,243,102,9,103,61,103,52,103,49,103,53,103,33,107,100,107,123,107,22,108,93,108,87,108,89,108,95,108,96,108,80,108,85,108,97,108,91,108,77,108,78,108,112,112,95,114,93,114,126,118,249,122,115,124,248,124,54,127,138,127,189,127,1,128,3,128,12,128,18,128,51,128,127,128,137,128,139,128,140,128,227,129,234,129,243,129,252,129,12,130,27,130,31,130,110,130,114,130,126,130,107,134,64,136,76,136,99,136,127,137,33,150,50,78,168,78,77,79,79,79,71,79,87,79,94,79,52,79,91,79,85,79,48,79,80,79,81,79,61,79,58,79,56,79,67,79,84,79,60,79,70,79,99,79,92,79,96,79,47,79,78,79,54,79,89,79,93,79,72,79,90,79,76,81,75,81,77,81,117,81,182,81,183,81,37,82,36,82,41,82,42,82,40,82,171,82,169,82,170,82,172,82,35,83,115,83,117,83,29,84,45,84,30,84,62,84,38,84,78,84,39,84,70,84,67,84,51,84,72,84,66,84,27,84,41,84,74,84,57,84,59,84,56,84,46,84,53,84,54,84,32,84,60,84,64,84,49,84,43,84,31,84,44,84,234,86,240,86,228,86,235,86,74,87,81,87,64,87,77,87,71,87,78,87,62,87,80,87,79,87,59,87,239,88,62,89,157,89,146,89,168,89,158,89,163,89,153,89,150,89,141,89,164,89,147,89,138,89,165,89,93,91,92,91,90,91,91,91,140,91,139,91,143,91,44,92,64,92,65,92,63,92,62,92,144,92,145,92,148,92,140,92,235,93,12,94,143,94,135,94,138,94,247,94,4,95,31,95,100,95,98,95,119,95,121,95,216,95,204,95,215,95,205,95,241,95,235,95,248,95,234,95,18,98,17,98,132,98,151,98,150,98,128,98,118,98,137,98,109,98,138,98,124,98,126,98,121,98,115,98,146,98,111,98,152,98,110,98,149,98,147,98,145,98,134,98,57,101,59,101,56,101,241,101,244,102,95,103,78,103,79,103,80,103,81,103,92,103,86,103,94,103,73,103,70,103,96,103,83,103,87,103,101,107,207,107,66,108,94,108,153,108,129,108,136,108,137,108,133,108,155,108,106,108,122,108,144,108,112,108,140,108,104,108,150,108,146,108,125,108,131,108,114,108,126,108,116,108,134,108,118,108,141,108,148,108,152,108,130,108,118,112,124,112,125,112,120,112,98,114,97,114,96,114,196,114,194,114,150,115,44,117,43,117,55,117,56,117,130,118,239,118,227,119,193,121,192,121,191,121,118,122,251,124,85,127,150,128,147,128,157,128,152,128,155,128,154,128,178,128,111,130,146,130,139,130,141,130,139,137,210,137,0,138,55,140,70,140,85,140,157,140,100,141,112,141,179,141,171,142,202,142,155,143,176,143,194,143,198,143,197,143,196,143,225,93,145,144,162,144,170,144,166,144,163,144,73,145,198,145,204,145,50,150,46,150,49,150,42,150,44,150,38,78,86,78,115,78,139,78,155,78,158,78,171,78,172,78,111,79,157,79,141,79,115,79,127,79,108,79,155,79,139,79,134,79,131,79,112,79,117,79,136,79,105,79,123,79,150,79,126,79,143,79,145,79,122,79,84,81,82,81,85,81,105,81,119,81,118,81,120,81,189,81,253,81,59,82,56,82,55,82,58,82,48,82,46,82,54,82,65,82,190,82,187,82,82,83,84,83,83,83,81,83,102,83,119,83,120,83,121,83,214,83,212,83,215,83,115,84,117,84,150,84,120,84,149,84,128,84,123,84,119,84,132,84,146,84,134,84,124,84,144,84,113,84,118,84,140,84,154,84,98,84,104,84,139,84,125,84,142,84,250,86,131,87,119,87,106,87,105,87,97,87,102,87,100,87,124,87,28,89,73,89,71,89,72,89,68,89,84,89,190,89,187,89,212,89,185,89,174,89,209,89,198,89,208,89,205,89,203,89,211,89,202,89,175,89,179,89,210,89,197,89,95,91,100,91,99,91,151,91,154,91,152,91,156,91,153,91,155,91,26,92,72,92,69,92,70,92,183,92,161,92,184,92,169,92,171,92,177,92,179,92,24,94,26,94,22,94,21,94,27,94,17,94,120,94,154,94,151,94,156,94,149,94,150,94,246,94,38,95,39,95,41,95,128,95,129,95,127,95,124,95,221,95,224,95,253,95,245,95,255,95,15,96,20,96,47,96,53,96,22,96,42,96,21,96,33,96,39,96,41,96,43,96,27,96,22,98,21,98,63,98,62,98,64,98,127,98,201,98,204,98,196,98,191,98,194,98,185,98,210,98,219,98,171,98,211,98,212,98,203,98,200,98,168,98,189,98,188,98,208,98,217,98,199,98,205,98,181,98,218,98,177,98,216,98,214,98,215,98,198,98,172,98,206,98,62,101,167,101,188,101,250,101,20,102,19,102,12,102,6,102,2,102,14,102,0,102,15,102,21,102,10,102,7,102,13,103,11,103,109,103,139,103,149,103,113,103,156,103,115,103,119,103,135,103,157,103,151,103,111,103,112,103,127,103,137,103,126,103,144,103,117,103,154,103,147,103,124,103,106,103,114,103,35,107,102,107,103,107,127,107,19,108,27,108,227,108,232,108,243,108,177,108,204,108,229,108,179,108,189,108,190,108,188,108,226,108,171,108,213,108,211,108,184,108,196,108,185,108,193,108,174,108,215,108,197,108,241,108,191,108,187,108,225,108,219,108,202,108,172,108,239,108,220,108,214,108,224,108,149,112,142,112,146,112,138,112,153,112,44,114,45,114,56,114,72,114,103,114,105,114,192,114,206,114,217,114,215,114,208,114,169,115,168,115,159,115,171,115,165,115,61,117,157,117,153,117,154,117,132,118,194,118,242,118,244,118,229,119,253,119,62,121,64,121,65,121,201,121,200,121,122,122,121,122,250,122,254,124,84,127,140,127,139,127,5,128,186,128,165,128,162,128,177,128,161,128,171,128,169,128,180,128,170,128,175,128,229,129,254,129,13,130,179,130,157,130,153,130,173,130,189,130,159,130,185,130,177,130,172,130,165,130,175,130,184,130,163,130,176,130,190,130,183,130,78,134,113,134,29,82,104,136,203,142,206,143,212,143,209,143,181,144,184,144,177,144,182,144,199,145,209,145,119,149,128,149,28,150,64,150,63,150,59,150,68,150,66,150,185,150,232,150,82,151,94,151,159,78,173,78,174,78,225,79,181,79,175,79,191,79,224,79,209,79,207,79,221,79,195,79,182,79,216,79,223,79,202,79,215,79,174,79,208,79,196,79,194,79,218,79,206,79,222,79,183,79,87,81,146,81,145,81,160,81,78,82,67,82,74,82,77,82,76,82,75,82,71,82,199,82,201,82,195,82,193,82,13,83,87,83,123,83,154,83,219,83,172,84,192,84,168,84,206,84,201,84,184,84,166,84,179,84,199,84,194,84,189,84,170,84,193,84,196,84,200,84,175,84,171,84,177,84,187,84,169,84,167,84,191,84,255,86,130,87,139,87,160,87,163,87,162,87,206,87,174,87,147,87,85,89,81,89,79,89,78,89,80,89,220,89,216,89,255,89,227,89,232,89,3,90,229,89,234,89,218,89,230,89,1,90,251,89,105,91,163,91,166,91,164,91,162,91,165,91,1,92,78,92,79,92,77,92,75,92,217,92,210,92,247,93,29,94,37,94,31,94,125,94,160,94,166,94,250,94,8,95,45,95,101,95,136,95,133,95,138,95,139,95,135,95,140,95,137,95,18,96,29,96,32,96,37,96,14,96,40,96,77,96,112,96,104,96,98,96,70,96,67,96,108,96,107,96,106,96,100,96,65,98,220,98,22,99,9,99,252,98,237,98,1,99,238,98,253,98,7,99,241,98,247,98,239,98,236,98,254,98,244,98,17,99,2,99,63,101,69,101,171,101,189,101,226,101,37,102,45,102,32,102,39,102,47,102,31,102,40,102,49,102,36,102,247,102,255,103,211,103,241,103,212,103,208,103,236,103,182,103,175,103,245,103,233,103,239,103,196,103,209,103,180,103,218,103,229,103,184,103,207,103,222,103,243,103,176,103,217,103,226,103,221,103,210,103,106,107,131,107,134,107,181,107,210,107,215,107,31,108,201,108,11,109,50,109,42,109,65,109,37,109,12,109,49,109,30,109,23,109,59,109,61,109,62,109,54,109,27,109,245,108,57,109,39,109,56,109,41,109,46,109,53,109,14,109,43,109,171,112,186,112,179,112,172,112,175,112,173,112,184,112,174,112,164,112,48,114,114,114,111,114,116,114,233,114,224,114,225,114,183,115,202,115,187,115,178,115,205,115,192,115,179,115,26,117,45,117,79,117,76,117,78,117,75,117,171,117,164,117,165,117,162,117,163,117,120,118,134,118,135,118,136,118,200,118,198,118,195,118,197,118,1,119,249,118,248,118,9,119,11,119,254,118,252,118,7,119,220,119,2,120,20,120,12,120,13,120,70,121,73,121,72,121,71,121,185,121,186,121,209,121,210,121,203,121,127,122,129,122,255,122,253,122,125,124,2,125,5,125,0,125,9,125,7,125,4,125,6,125,56,127,142,127,191,127,4,128,16,128,13,128,17,128,54,128,214,128,229,128,218,128,195,128,196,128,204,128,225,128,219,128,206,128,222,128,228,128,221,128,244,129,34,130,231,130,3,131,5,131,227,130,219,130,230,130,4,131,229,130,2,131,9,131,210,130,215,130,241,130,1,131,220,130,212,130,209,130,222,130,211,130,223,130,239,130,6,131,80,134,121,134,123,134,122,134,77,136,107,136,129,137,212,137,8,138,2,138,3,138,158,140,160,140,116,141,115,141,180,141,205,142,204,142,240,143,230,143,226,143,234,143,229,143,237,143,235,143,228,143,232,143,202,144,206,144,193,144,195,144,75,145,74,145,205,145,130,149,80,150,75,150,76,150,77,150,98,151,105,151,203,151,237,151,243,151,1,152,168,152,219,152,223,152,150,153,153,153,88,78,179,78,12,80,13,80,35,80,239,79,38,80,37,80,248,79,41,80,22,80,6,80,60,80,31,80,26,80,18,80,17,80,250,79,0,80,20,80,40,80,241,79,33,80,11,80,25,80,24,80,243,79,238,79,45,80,42,80,254,79,43,80,9,80,124,81,164,81,165,81,162,81,205,81,204,81,198,81,203,81,86,82,92,82,84,82,91,82,93,82,42,83,127,83,159,83,157,83,223,83,232,84,16,85,1,85,55,85,252,84,229,84,242,84,6,85,250,84,20,85,233,84,237,84,225,84,9,85,238,84,234,84,230,84,39,85,7,85,253,84,15,85,3,87,4,87,194,87,212,87,203,87,195,87,9,88,15,89,87,89,88,89,90,89,17,90,24,90,28,90,31,90,27,90,19,90,236,89,32,90,35,90,41,90,37,90,12,90,9,90,107,91,88,92,176,91,179,91,182,91,180,91,174,91,181,91,185,91,184,91,4,92,81,92,85,92,80,92,237,92,253,92,251,92,234,92,232,92,240,92,246,92,1,93,244,92,238,93,45,94,43,94,171,94,173,94,167,94,49,95,146,95,145,95,144,95,89,96,99,96,101,96,80,96,85,96,109,96,105,96,111,96,132,96,159,96,154,96,141,96,148,96,140,96,133,96,150,96,71,98,243,98,8,99,255,98,78,99,62,99,47,99,85,99,66,99,70,99,79,99,73,99,58,99,80,99,61,99,42,99,43,99,40,99,77,99,76,99,72,101,73,101,153,101,193,101,197,101,66,102,73,102,79,102,67,102,82,102,76,102,69,102,65,102,248,102,20,103,21,103,23,103,33,104,56,104,72,104,70,104,83,104,57,104,66,104,84,104,41,104,179,104,23,104,76,104,81,104,61,104,244,103,80,104,64,104,60,104,67,104,42,104,69,104,19,104,24,104,65,104,138,107,137,107,183,107,35,108,39,108,40,108,38,108,36,108,240,108,106,109,149,109,136,109,135,109,102,109,120,109,119,109,89,109,147,109,108,109,137,109,110,109,90,109,116,109,105,109,140,109,138,109,121,109,133,109,101,109,148,109,202,112,216,112,228,112,217,112,200,112,207,112,57,114,121,114,252,114,249,114,253,114,248,114,247,114,134,115,237,115,9,116,238,115,224,115,234,115,222,115,84,117,93,117,92,117,90,117,89,117,190,117,197,117,199,117,178,117,179,117,189,117,188,117,185,117,194,117,184,117,139,118,176,118,202,118,205,118,206,118,41,119,31,119,32,119,40,119,233,119,48,120,39,120,56,120,29,120,52,120,55,120,37,120,45,120,32,120,31,120,50,120,85,121,80,121,96,121,95,121,86,121,94,121,93,121,87,121,90,121,228,121,227,121,231,121,223,121,230,121,233,121,216,121,132,122,136,122,217,122,6,123,17,123,137,124,33,125,23,125,11,125,10,125,32,125,34,125,20,125,16,125,21,125,26,125,28,125,13,125,25,125,27,125,58,127,95,127,148,127,197,127,193,127,6,128,24,128,21,128,25,128,23,128,61,128,63,128,241,128,2,129,240,128,5,129,237,128,244,128,6,129,248,128,243,128,8,129,253,128,10,129,252,128,239,128,237,129,236,129,0,130,16,130,42,130,43,130,40,130,44,130,187,130,43,131,82,131,84,131,74,131,56,131,80,131,73,131,53,131,52,131,79,131,50,131,57,131,54,131,23,131,64,131,49,131,40,131,67,131,84,134,138,134,170,134,147,134,164,134,169,134,140,134,163,134,156,134,112,136,119,136,129,136,130,136,125,136,121,136,24,138,16,138,14,138,12,138,21,138,10,138,23,138,19,138,22,138,15,138,17,138,72,140,122,140,121,140,161,140,162,140,119,141,172,142,210,142,212,142,207,142,177,143,1,144,6,144,247,143,0,144,250,143,244,143,3,144,253,143,5,144,248,143,149,144,225,144,221,144,226,144,82,145,77,145,76,145,216,145,221,145,215,145,220,145,217,145,131,149,98,150,99,150,97,150,91,150,93,150,100,150,88,150,94,150,187,150,226,152,172,153,168,154,216,154,37,155,50,155,60,155,126,78,122,80,125,80,92,80,71,80,67,80,76,80,90,80,73,80,101,80,118,80,78,80,85,80,117,80,116,80,119,80,79,80,15,80,111,80,109,80,92,81,149,81,240,81,106,82,111,82,210,82,217,82,216,82,213,82,16,83,15,83,25,83,63,83,64,83,62,83,195,83,252,102,70,85,106,85,102,85,68,85,94,85,97,85,67,85,74,85,49,85,86,85,79,85,85,85,47,85,100,85,56,85,46,85,92,85,44,85,99,85,51,85,65,85,87,85,8,87,11,87,9,87,223,87,5,88,10,88,6,88,224,87,228,87,250,87,2,88,53,88,247,87,249,87,32,89,98,89,54,90,65,90,73,90,102,90,106,90,64,90,60,90,98,90,90,90,70,90,74,90,112,91,199,91,197,91,196,91,194,91,191,91,198,91,9,92,8,92,7,92,96,92,92,92,93,92,7,93,6,93,14,93,27,93,22,93,34,93,17,93,41,93,20,93,25,93,36,93,39,93,23,93,226,93,56,94,54,94,51,94,55,94,183,94,184,94,182,94,181,94,190,94,53,95,55,95,87,95,108,95,105,95,107,95,151,95,153,95,158,95,152,95,161,95,160,95,156,95,127,96,163,96,137,96,160,96,168,96,203,96,180,96,230,96,189,96,197,96,187,96,181,96,220,96,188,96,216,96,213,96,198,96,223,96,184,96,218,96,199,96,26,98,27,98,72,98,160,99,167,99,114,99,150,99,162,99,165,99,119,99,103,99,152,99,170,99,113,99,169,99,137,99,131,99,155,99,107,99,168,99,132,99,136,99,153,99,161,99,172,99,146,99,143,99,128,99,123,99,105,99,104,99,122,99,93,101,86,101,81,101,89,101,87,101,95,85,79,101,88,101,85,101,84,101,156,101,155,101,172,101,207,101,203,101,204,101,206,101,93,102,90,102,100,102,104,102,102,102,94,102,249,102,215,82,27,103,129,104,175,104,162,104,147,104,181,104,127,104,118,104,177,104,167,104,151,104,176,104,131,104,196,104,173,104,134,104,133,104,148,104,157,104,168,104,159,104,161,104,130,104,50,107,186,107,235,107,236,107,43,108,142,109,188,109,243,109,217,109,178,109,225,109,204,109,228,109,251,109,250,109,5,110,199,109,203,109,175,109,209,109,174,109,222,109,249,109,184,109,247,109,245,109,197,109,210,109,26,110,181,109,218,109,235,109,216,109,234,109,241,109,238,109,232,109,198,109,196,109,170,109,236,109,191,109,230,109,249,112,9,113,10,113,253,112,239,112,61,114,125,114,129,114,28,115,27,115,22,115,19,115,25,115,135,115,5,116,10,116,3,116,6,116,254,115,13,116,224,116,246,116,247,116,28,117,34,117,101,117,102,117,98,117,112,117,143,117,212,117,213,117,181,117,202,117,205,117,142,118,212,118,210,118,219,118,55,119,62,119,60,119,54,119,56,119,58,119,107,120,67,120,78,120,101,121,104,121,109,121,251,121,146,122,149,122,32,123,40,123,27,123,44,123,38,123,25,123,30,123,46,123,146,124,151,124,149,124,70,125,67,125,113,125,46,125,57,125,60,125,64,125,48,125,51,125,68,125,47,125,66,125,50,125,49,125,61,127,158,127,154,127,204,127,206,127,210,127,28,128,74,128,70,128,47,129,22,129,35,129,43,129,41,129,48,129,36,129,2,130,53,130,55,130,54,130,57,130,142,131,158,131,152,131,120,131,162,131,150,131,189,131,171,131,146,131,138,131,147,131,137,131,160,131,119,131,123,131,124,131,134,131,167,131,85,134,106,95,199,134,192,134,182,134,196,134,181,134,198,134,203,134,177,134,175,134,201,134,83,136,158,136,136,136,171,136,146,136,150,136,141,136,139,136,147,137,143,137,42,138,29,138,35,138,37,138,49,138,45,138,31,138,27,138,34,138,73,140,90,140,169,140,172,140,171,140,168,140,170,140,167,140,103,141,102,141,190,141,186,141,219,142,223,142,25,144,13,144,26,144,23,144,35,144,31,144,29,144,16,144,21,144,30,144,32,144,15,144,34,144,22,144,27,144,20,144,232,144,237,144,253,144,87,145,206,145,245,145,230,145,227,145,231,145,237,145,233,145,137,149,106,150,117,150,115,150,120,150,112,150,116,150,118,150,119,150,108,150,192,150,234,150,233,150,224,122,223,122,2,152,3,152,90,155,229,156,117,158,127,158,165,158,187,158,162,80,141,80,133,80,153,80,145,80,128,80,150,80,152,80,154,80,0,103,241,81,114,82,116,82,117,82,105,82,222,82,221,82,219,82,90,83,165,83,123,85,128,85,167,85,124,85,138,85,157,85,152,85,130,85,156,85,170,85,148,85,135,85,139,85,131,85,179,85,174,85,159,85,62,85,178,85,154,85,187,85,172,85,177,85,126,85,137,85,171,85,153,85,13,87,47,88,42,88,52,88,36,88,48,88,49,88,33,88,29,88,32,88,249,88,250,88,96,89,119,90,154,90,127,90,146,90,155,90,167,90,115,91,113,91,210,91,204,91,211,91,208,91,10,92,11,92,49,92,76,93,80,93,52,93,71,93,253,93,69,94,61,94,64,94,67,94,126,94,202,94,193,94,194,94,196,94,60,95,109,95,169,95,170,95,168,95,209,96,225,96,178,96,182,96,224,96,28,97,35,97,250,96,21,97,240,96,251,96,244,96,104,97,241,96,14,97,246,96,9,97,0,97,18,97,31,98,73,98,163,99,140,99,207,99,192,99,233,99,201,99,198,99,205,99,210,99,227,99,208,99,225,99,214,99,237,99,238,99,118,99,244,99,234,99,219,99,82,100,218,99,249,99,94,101,102,101,98,101,99,101,145,101,144,101,175,101,110,102,112,102,116,102,118,102,111,102,145,102,122,102,126,102,119,102,254,102,255,102,31,103,29,103,250,104,213,104,224,104,216,104,215,104,5,105,223,104,245,104,238,104,231,104,249,104,210,104,242,104,227,104,203,104,205,104,13,105,18,105,14,105,201,104,218,104,110,105,251,104,62,107,58,107,61,107,152,107,150,107,188,107,239,107,46,108,47,108,44,108,47,110,56,110,84,110,33,110,50,110,103,110,74,110,32,110,37,110,35,110,27,110,91,110,88,110,36,110,86,110,110,110,45,110,38,110,111,110,52,110,77,110,58,110,44,110,67,110,29,110,62,110,203,110,137,110,25,110,78,110,99,110,68,110,114,110,105,110,95,110,25,113,26,113,38,113,48,113,33,113,54,113,110,113,28,113,76,114,132,114,128,114,54,115,37,115,52,115,41,115,58,116,42,116,51,116,34,116,37,116,53,116,54,116,52,116,47,116,27,116,38,116,40,116,37,117,38,117,107,117,106,117,226,117,219,117,227,117,217,117,216,117,222,117,224,117,123,118,124,118,150,118,147,118,180,118,220,118,79,119,237,119,93,120,108,120,111,120,13,122,8,122,11,122,5,122,0,122,152,122,151,122,150,122,229,122,227,122,73,123,86,123,70,123,80,123,82,123,84,123,77,123,75,123,79,123,81,123,159,124,165,124,94,125,80,125,104,125,85,125,43,125,110,125,114,125,97,125,102,125,98,125,112,125,115,125,132,85,212,127,213,127,11,128,82,128,133,128,85,129,84,129,75,129,81,129,78,129,57,129,70,129,62,129,76,129,83,129,116,129,18,130,28,130,233,131,3,132,248,131,13,132,224,131,197,131,11,132,193,131,239,131,241,131,244,131,87,132,10,132,240,131,12,132,204,131,253,131,242,131,202,131,56,132,14,132,4,132,220,131,7,132,212,131,223,131,91,134,223,134,217,134,237,134,212,134,219,134,228,134,208,134,222,134,87,136,193,136,194,136,177,136,131,137,150,137,59,138,96,138,85,138,94,138,60,138,65,138,84,138,91,138,80,138,70,138,52,138,58,138,54,138,86,138,97,140,130,140,175,140,188,140,179,140,189,140,193,140,187,140,192,140,180,140,183,140,182,140,191,140,184,140,138,141,133,141,129,141,206,141,221,141,203,141,218,141,209,141,204,141,219,141,198,141,251,142,248,142,252,142,156,143,46,144,53,144,49,144,56,144,50,144,54,144,2,145,245,144,9,145,254,144,99,145,101,145,207,145,20,146,21,146,35,146,9,146,30,146,13,146,16,146,7,146,17,146,148,149,143,149,139,149,145,149,147,149,146,149,142,149,138,150,142,150,139,150,125,150,133,150,134,150,141,150,114,150,132,150,193,150,197,150,196,150,198,150,199,150,239,150,242,150,204,151,5,152,6,152,8,152,231,152,234,152,239,152,233,152,242,152,237,152,174,153,173,153,195,158,205,158,209,158,130,78,173,80,181,80,178,80,179,80,197,80,190,80,172,80,183,80,187,80,175,80,199,80,127,82,119,82,125,82,223,82,230,82,228,82,226,82,227,82,47,83,223,85,232,85,211,85,230,85,206,85,220,85,199,85,209,85,227,85,228,85,239,85,218,85,225,85,197,85,198,85,229,85,201,85,18,87,19,87,94,88,81,88,88,88,87,88,90,88,84,88,107,88,76,88,109,88,74,88,98,88,82,88,75,88,103,89,193,90,201,90,204,90,190,90,189,90,188,90,179,90,194,90,178,90,105,93,111,93,76,94,121,94,201,94,200,94,18,95,89,95,172,95,174,95,26,97,15,97,72,97,31,97,243,96,27,97,249,96,1,97,8,97,78,97,76,97,68,97,77,97,62,97,52,97,39,97,13,97,6,97,55,97,33,98,34,98,19,100,62,100,30,100,42,100,45,100,61,100,44,100,15,100,28,100,20,100,13,100,54,100,22,100,23,100,6,100,108,101,159,101,176,101,151,102,137,102,135,102,136,102,150,102,132,102,152,102,141,102,3,103,148,105,109,105,90,105,119,105,96,105,84,105,117,105,48,105,130,105,74,105,104,105,107,105,94,105,83,105,121,105,134,105,93,105,99,105,91,105,71,107,114,107,192,107,191,107,211,107,253,107,162,110,175,110,211,110,182,110,194,110,144,110,157,110,199,110,197,110,165,110,152,110,188,110,186,110,171,110,209,110,150,110,156,110,196,110,212,110,170,110,167,110,180,110,78,113,89,113,105,113,100,113,73,113,103,113,92,113,108,113,102,113,76,113,101,113,94,113,70,113,104,113,86,113,58,114,82,114,55,115,69,115,63,115,62,115,111,116,90,116,85,116,95,116,94,116,65,116,63,116,89,116,91,116,92,116,118,117,120,117,0,118,240,117,1,118,242,117,241,117,250,117,255,117,244,117,243,117,222,118,223,118,91,119,107,119,102,119,94,119,99,119,121,119,106,119,108,119,92,119,101,119,104,119,98,119,238,119,142,120,176,120,151,120,152,120,140,120,137,120,124,120,145,120,147,120,127,120,122,121,127,121,129,121,44,132,189,121,28,122,26,122,32,122,20,122,31,122,30,122,159,122,160,122,119,123,192,123,96,123,110,123,103,123,177,124,179,124,181,124,147,125,121,125,145,125,129,125,143,125,91,125,110,127,105,127,106,127,114,127,169,127,168,127,164,127,86,128,88,128,134,128,132,128,113,129,112,129,120,129,101,129,110,129,115,129,107,129,121,129,122,129,102,129,5,130,71,130,130,132,119,132,61,132,49,132,117,132,102,132,107,132,73,132,108,132,91,132,60,132,53,132,97,132,99,132,105,132,109,132,70,132,94,134,92,134,95,134,249,134,19,135,8,135,7,135,0,135,254,134,251,134,2,135,3,135,6,135,10,135,89,136,223,136,212,136,217,136,220,136,216,136,221,136,225,136,202,136,213,136,210,136,156,137,227,137,107,138,114,138,115,138,102,138,105,138,112,138,135,138,124,138,99,138,160,138,113,138,133,138,109,138,98,138,110,138,108,138,121,138,123,138,62,138,104,138,98,140,138,140,137,140,202,140,199,140,200,140,196,140,178,140,195,140,194,140,197,140,225,141,223,141,232,141,239,141,243,141,250,141,234,141,228,141,230,141,178,142,3,143,9,143,254,142,10,143,159,143,178,143,75,144,74,144,83,144,66,144,84,144,60,144,85,144,80,144,71,144,79,144,78,144,77,144,81,144,62,144,65,144,18,145,23,145,108,145,106,145,105,145,201,145,55,146,87,146,56,146,61,146,64,146,62,146,91,146,75,146,100,146,81,146,52,146,73,146,77,146,69,146,57,146,63,146,90,146,152,149,152,150,148,150,149,150,205,150,203,150,201,150,202,150,247,150,251,150,249,150,246,150,86,151,116,151,118,151,16,152,17,152,19,152,10,152,18,152,12,152,252,152,244,152,253,152,254,152,179,153,177,153,180,153,225,154,233,156,130,158,14,159,19,159,32,159,231,80,238,80,229,80,214,80,237,80,218,80,213,80,207,80,209,80,241,80,206,80,233,80,98,81,243,81,131,82,130,82,49,83,173,83,254,85,0,86,27,86,23,86,253,85,20,86,6,86,9,86,13,86,14,86,247,85,22,86,31,86,8,86,16,86,246,85,24,87,22,87,117,88,126,88,131,88,147,88,138,88,121,88,133,88,125,88,253,88,37,89,34,89,36,89,106,89,105,89,225,90,230,90,233,90,215,90,214,90,216,90,227,90,117,91,222,91,231,91,225,91,229,91,230,91,232,91,226,91,228,91,223,91,13,92,98,92,132,93,135,93,91,94,99,94,85,94,87,94,84,94,211,94,214,94,10,95,70,95,112,95,185,95,71,97,63,97,75,97,119,97,98,97,99,97,95,97,90,97,88,97,117,97,42,98,135,100,88,100,84,100,164,100,120,100,95,100,122,100,81,100,103,100,52,100,109,100,123,100,114,101,161,101,215,101,214,101,162,102,168,102,157,102,156,105,168,105,149,105,193,105,174,105,211,105,203,105,155,105,183,105,187,105,171,105,180,105,208,105,205,105,173,105,204,105,166,105,195,105,163,105,73,107,76,107,51,108,51,111,20,111,254,110,19,111,244,110,41,111,62,111,32,111,44,111,15,111,2,111,34,111,255,110,239,110,6,111,49,111,56,111,50,111,35,111,21,111,43,111,47,111,136,111,42,111,236,110,1,111,242,110,204,110,247,110,148,113,153,113,125,113,138,113,132,113,146,113,62,114,146,114,150,114,68,115,80,115,100,116,99,116,106,116,112,116,109,116,4,117,145,117,39,118,13,118,11,118,9,118,19,118,225,118,227,118,132,119,125,119,127,119,97,119,193,120,159,120,167,120,179,120,169,120,163,120,142,121,143,121,141,121,46,122,49,122,170,122,169,122,237,122,239,122,161,123,149,123,139,123,117,123,151,123,157,123,148,123,143,123,184,123,135,123,132,123,185,124,189,124,190,124,187,125,176,125,156,125,189,125,190,125,160,125,202,125,180,125,178,125,177,125,186,125,162,125,191,125,181,125,184,125,173,125,210,125,199,125,172,125,112,127,224,127,225,127,223,127,94,128,90,128,135,128,80,129,128,129,143,129,136,129,138,129,127,129,130,129,231,129,250,129,7,130,20,130,30,130,75,130,201,132,191,132,198,132,196,132,153,132,158,132,178,132,156,132,203,132,184,132,192,132,211,132,144,132,188,132,209,132,202,132,63,135,28,135,59,135,34,135,37,135,52,135,24,135,85,135,55,135,41,135,243,136,2,137,244,136,249,136,248,136,253,136,232,136,26,137,239,136,166,138,140,138,158,138,163,138,141,138,161,138,147,138,164,138,170,138,165,138,168,138,152,138,145,138,154,138,167,138,106,140,141,140,140,140,211,140,209,140,210,140,107,141,153,141,149,141,252,141,20,143,18,143,21,143,19,143,163,143,96,144,88,144,92,144,99,144,89,144,94,144,98,144,93,144,91,144,25,145,24,145,30,145,117,145,120,145,119,145,116,145,120,146,128,146,133,146,152,146,150,146,123,146,147,146,156,146,168,146,124,146,145,146,161,149,168,149,169,149,163,149,165,149,164,149,153,150,156,150,155,150,204,150,210,150,0,151,124,151,133,151,246,151,23,152,24,152,175,152,177,152,3,153,5,153,12,153,9,153,193,153,175,154,176,154,230,154,65,155,66,155,244,156,246,156,243,156,188,158,59,159,74,159,4,81,0,81,251,80,245,80,249,80,2,81,8,81,9,81,5,81,220,81,135,82,136,82,137,82,141,82,138,82,240,82,178,83,46,86,59,86,57,86,50,86,63,86,52,86,41,86,83,86,78,86,87,86,116,86,54,86,47,86,48,86,128,88,159,88,158,88,179,88,156,88,174,88,169,88,166,88,109,89,9,91,251,90,11,91,245,90,12,91,8,91,238,91,236,91,233,91,235,91,100,92,101,92,157,93,148,93,98,94,95,94,97,94,226,94,218,94,223,94,221,94,227,94,224,94,72,95,113,95,183,95,181,95,118,97,103,97,110,97,93,97,85,97,130,97,124,97,112,97,107,97,126,97,167,97,144,97,171,97,142,97,172,97,154,97,164,97,148,97,174,97,46,98,105,100,111,100,121,100,158,100,178,100,136,100,144,100,176,100,165,100,147,100,149,100,169,100,146,100,174,100,173,100,171,100,154,100,172,100,153,100,162,100,179,100,117,101,119,101,120,101,174,102,171,102,180,102,177,102,35,106,31,106,232,105,1,106,30,106,25,106,253,105,33,106,19,106,10,106,243,105,2,106,5,106,237,105,17,106,80,107,78,107,164,107,197,107,198,107,63,111,124,111,132,111,81,111,102,111,84,111,134,111,109,111,91,111,120,111,110,111,142,111,122,111,112,111,100,111,151,111,88,111,213,110,111,111,96,111,95,111,159,113,172,113,177,113,168,113,86,114,155,114,78,115,87,115,105,116,139,116,131,116,126,116,128,116,127,117,32,118,41,118,31,118,36,118,38,118,33,118,34,118,154,118,186,118,228,118,142,119,135,119,140,119,145,119,139,119,203,120,197,120,186,120,202,120,190,120,213,120,188,120,208,120,63,122,60,122,64,122,61,122,55,122,59,122,175,122,174,122,173,123,177,123,196,123,180,123,198,123,199,123,193,123,160,123,204,123,202,124,224,125,244,125,239,125,251,125,216,125,236,125,221,125,232,125,227,125,218,125,222,125,233,125,158,125,217,125,242,125,249,125,117,127,119,127,175,127,233,127,38,128,155,129,156,129,157,129,160,129,154,129,152,129,23,133,61,133,26,133,238,132,44,133,45,133,19,133,17,133,35,133,33,133,20,133,236,132,37,133,255,132,6,133,130,135,116,135,118,135,96,135,102,135,120,135,104,135,89,135,87,135,76,135,83,135,91,136,93,136,16,137,7,137,18,137,19,137,21,137,10,137,188,138,210,138,199,138,196,138,149,138,203,138,248,138,178,138,201,138,194,138,191,138,176,138,214,138,205,138,182,138,185,138,219,138,76,140,78,140,108,140,224,140,222,140,230,140,228,140,236,140,237,140,226,140,227,140,220,140,234,140,225,140,109,141,159,141,163,141,43,142,16,142,29,142,34,142,15,142,41,142,31,142,33,142,30,142,186,142,29,143,27,143,31,143,41,143,38,143,42,143,28,143,30,143,37,143,105,144,110,144,104,144,109,144,119,144,48,145,45,145,39,145,49,145,135,145,137,145,139,145,131,145,197,146,187,146,183,146,234,146,172,146,228,146,193,146,179,146,188,146,210,146,199,146,240,146,178,146,173,149,177,149,4,151,6,151,7,151,9,151,96,151,141,151,139,151,143,151,33,152,43,152,28,152,179,152,10,153,19,153,18,153,24,153,221,153,208,153,223,153,219,153,209,153,213,153,210,153,217,153,183,154,238,154,239,154,39,155,69,155,68,155,119,155,111,155,6,157,9,157,3,157,169,158,190,158,206,158,168,88,82,159,18,81,24,81,20,81,16,81,21,81,128,81,170,81,221,81,145,82,147,82,243,82,89,86,107,86,121,86,105,86,100,86,120,86,106,86,104,86,101,86,113,86,111,86,108,86,98,86,118,86,193,88,190,88,199,88,197,88,110,89,29,91,52,91,120,91,240,91,14,92,74,95,178,97,145,97,169,97,138,97,205,97,182,97,190,97,202,97,200,97,48,98,197,100,193,100,203,100,187,100,188,100,218,100,196,100,199,100,194,100,205,100,191,100,210,100,212,100,190,100,116,101,198,102,201,102,185,102,196,102,199,102,184,102,61,106,56,106,58,106,89,106,107,106,88,106,57,106,68,106,98,106,97,106,75,106,71,106,53,106,95,106,72,106,89,107,119,107,5,108,194,111,177,111,161,111,195,111,164,111,193,111,167,111,179,111,192,111,185,111,182,111,166,111,160,111,180,111,190,113,201,113,208,113,210,113,200,113,213,113,185,113,206,113,217,113,220,113,195,113,196,113,104,115,156,116,163,116,152,116,159,116,158,116,226,116,12,117,13,117,52,118,56,118,58,118,231,118,229,118,160,119,158,119,159,119,165,119,232,120,218,120,236,120,231,120,166,121,77,122,78,122,70,122,76,122,75,122,186,122,217,123,17,124,201,123,228,123,219,123,225,123,233,123,230,123,213,124,214,124,10,126,17,126,8,126,27,126,35,126,30,126,29,126,9,126,16,126,121,127,178,127,240,127,241,127,238,127,40,128,179,129,169,129,168,129,251,129,8,130,88,130,89,130,74,133,89,133,72,133,104,133,105,133,67,133,73,133,109,133,106,133,94,133,131,135,159,135,158,135,162,135,141,135,97,136,42,137,50,137,37,137,43,137,33,137,170,137,166,137,230,138,250,138,235,138,241,138,0,139,220,138,231,138,238,138,254,138,1,139,2,139,247,138,237,138,243,138,246,138,252,138,107,140,109,140,147,140,244,140,68,142,49,142,52,142,66,142,57,142,53,142,59,143,47,143,56,143,51,143,168,143,166,143,117,144,116,144,120,144,114,144,124,144,122,144,52,145,146,145,32,147,54,147,248,146,51,147,47,147,34,147,252,146,43,147,4,147,26,147,16,147,38,147,33,147,21,147,46,147,25,147,187,149,167,150,168,150,170,150,213,150,14,151,17,151,22,151,13,151,19,151,15,151,91,151,92,151,102,151,152,151,48,152,56,152,59,152,55,152,45,152,57,152,36,152,16,153,40,153,30,153,27,153,33,153,26,153,237,153,226,153,241,153,184,154,188,154,251,154,237,154,40,155,145,155,21,157,35,157,38,157,40,157,18,157,27,157,216,158,212,158,141,159,156,159,42,81,31,81,33,81,50,81,245,82,142,86,128,86,144,86,133,86,135,86,143,86,213,88,211,88,209,88,206,88,48,91,42,91,36,91,122,91,55,92,104,92,188,93,186,93,189,93,184,93,107,94,76,95,189,95,201,97,194,97,199,97,230,97,203,97,50,98,52,98,206,100,202,100,216,100,224,100,240,100,230,100,236,100,241,100,226,100,237,100,130,101,131,101,217,102,214,102,128,106,148,106,132,106,162,106,156,106,219,106,163,106,126,106,151,106,144,106,160,106,92,107,174,107,218,107,8,108,216,111,241,111,223,111,224,111,219,111,228,111,235,111,239,111,128,111,236,111,225,111,233,111,213,111,238,111,240,111,231,113,223,113,238,113,230,113,229,113,237,113,236,113,244,113,224,113,53,114,70,114,112,115,114,115,169,116,176,116,166,116,168,116,70,118,66,118,76,118,234,118,179,119,170,119,176,119,172,119,167,119,173,119,239,119,247,120,250,120,244,120,239,120,1,121,167,121,170,121,87,122,191,122,7,124,13,124,254,123,247,123,12,124,224,123,224,124,220,124,222,124,226,124,223,124,217,124,221,124,46,126,62,126,70,126,55,126,50,126,67,126,43,126,61,126,49,126,69,126,65,126,52,126,57,126,72,126,53,126,63,126,47,126,68,127,243,127,252,127,113,128,114,128,112,128,111,128,115,128,198,129,195,129,186,129,194,129,192,129,191,129,189,129,201,129,190,129,232,129,9,130,113,130,170,133,132,133,126,133,156,133,145,133,148,133,175,133,155,133,135,133,168,133,138,133,103,134,192,135,209,135,179,135,210,135,198,135,171,135,187,135,186,135,200,135,203,135,59,137,54,137,68,137,56,137,61,137,172,137,14,139,23,139,25,139,27,139,10,139,32,139,29,139,4,139,16,139,65,140,63,140,115,140,250,140,253,140,252,140,248,140,251,140,168,141,73,142,75,142,72,142,74,142,68,143,62,143,66,143,69,143,63,143,127,144,125,144,132,144,129,144,130,144,128,144,57,145,163,145,158,145,156,145,77,147,130,147,40,147,117,147,74,147,101,147,75,147,24,147,126,147,108,147,91,147,112,147,90,147,84,147,202,149,203,149,204,149,200,149,198,149,177,150,184,150,214,150,28,151,30,151,160,151,211,151,70,152,182,152,53,153,1,154,255,153,174,155,171,155,170,155,173,155,59,157,63,157,139,158,207,158,222,158,220,158,221,158,219,158,62,159,75,159,226,83,149,86,174,86,217,88,216,88,56,91,93,95,227,97,51,98,244,100,242,100,254,100,6,101,250,100,251,100,247,100,183,101,220,102,38,103,179,106,172,106,195,106,187,106,184,106,194,106,174,106,175,106,95,107,120,107,175,107,9,112,11,112,254,111,6,112,250,111,17,112,15,112,251,113,252,113,254,113,248,113,119,115,117,115,167,116,191,116,21,117,86,118,88,118,82,118,189,119,191,119,187,119,188,119,14,121,174,121,97,122,98,122,96,122,196,122,197,122,43,124,39,124,42,124,30,124,35,124,33,124,231,124,84,126,85,126,94,126,90,126,97,126,82,126,89,126,72,127,249,127,251,127,119,128,118,128,205,129,207,129,10,130,207,133,169,133,205,133,208,133,201,133,176,133,186,133,185,133,166,133,239,135,236,135,242,135,224,135,134,137,178,137,244,137,40,139,57,139,44,139,43,139,80,140,5,141,89,142,99,142,102,142,100,142,95,142,85,142,192,142,73,143,77,143,135,144,131,144,136,144,171,145,172,145,208,145,148,147,138,147,150,147,162,147,179,147,174,147,172,147,176,147,152,147,154,147,151,147,212,149,214,149,208,149,213,149,226,150,220,150,217,150,219,150,222,150,36,151,163,151,166,151,173,151,249,151,77,152,79,152,76,152,78,152,83,152,186,152,62,153,63,153,61,153,46,153,165,153,14,154,193,154,3,155,6,155,79,155,78,155,77,155,202,155,201,155,253,155,200,155,192,155,81,157,93,157,96,157,224,158,21,159,44,159,51,81,165,86,222,88,223,88,226,88,245,91,144,159,236,94,242,97,247,97,246,97,245,97,0,101], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+97568);
/* memory initializer */ allocate([15,101,224,102,221,102,229,106,221,106,218,106,211,106,27,112,31,112,40,112,26,112,29,112,21,112,24,112,6,114,13,114,88,114,162,114,120,115,122,115,189,116,202,116,227,116,135,117,134,117,95,118,97,118,199,119,25,121,177,121,107,122,105,122,62,124,63,124,56,124,61,124,55,124,64,124,107,126,109,126,121,126,105,126,106,126,133,127,115,126,182,127,185,127,184,127,216,129,233,133,221,133,234,133,213,133,228,133,229,133,247,133,251,135,5,136,13,136,249,135,254,135,96,137,95,137,86,137,94,137,65,139,92,139,88,139,73,139,90,139,78,139,79,139,70,139,89,139,8,141,10,141,124,142,114,142,135,142,118,142,108,142,122,142,116,142,84,143,78,143,173,143,138,144,139,144,177,145,174,145,225,147,209,147,223,147,195,147,200,147,220,147,221,147,214,147,226,147,205,147,216,147,228,147,215,147,232,147,220,149,180,150,227,150,42,151,39,151,97,151,220,151,251,151,94,152,88,152,91,152,188,152,69,153,73,153,22,154,25,154,13,155,232,155,231,155,214,155,219,155,137,157,97,157,114,157,106,157,108,157,146,158,151,158,147,158,180,158,248,82,168,86,183,86,182,86,180,86,188,86,228,88,64,91,67,91,125,91,246,91,201,93,248,97,250,97,24,101,20,101,25,101,230,102,39,103,236,106,62,112,48,112,50,112,16,114,123,115,207,116,98,118,101,118,38,121,42,121,44,121,43,121,199,122,246,122,76,124,67,124,77,124,239,124,240,124,174,143,125,126,124,126,130,126,76,127,0,128,218,129,102,130,251,133,249,133,17,134,250,133,6,134,11,134,7,134,10,134,20,136,21,136,100,137,186,137,248,137,112,139,108,139,102,139,111,139,95,139,107,139,15,141,13,141,137,142,129,142,133,142,130,142,180,145,203,145,24,148,3,148,253,147,225,149,48,151,196,152,82,153,81,153,168,153,43,154,48,154,55,154,53,154,19,156,13,156,121,158,181,158,232,158,47,159,95,159,99,159,97,159,55,81,56,81,193,86,192,86,194,86,20,89,108,92,205,93,252,97,254,97,29,101,28,101,149,101,233,102,251,106,4,107,250,106,178,107,76,112,27,114,167,114,214,116,212,116,105,118,211,119,80,124,143,126,140,126,188,127,23,134,45,134,26,134,35,136,34,136,33,136,31,136,106,137,108,137,189,137,116,139,119,139,125,139,19,141,138,142,141,142,139,142,95,143,175,143,186,145,46,148,51,148,53,148,58,148,56,148,50,148,43,148,226,149,56,151,57,151,50,151,255,151,103,152,101,152,87,153,69,154,67,154,64,154,62,154,207,154,84,155,81,155,45,156,37,156,175,157,180,157,194,157,184,157,157,158,239,158,25,159,92,159,102,159,103,159,60,81,59,81,200,86,202,86,201,86,127,91,212,93,210,93,78,95,255,97,36,101,10,107,97,107,81,112,88,112,128,115,228,116,138,117,110,118,108,118,179,121,96,124,95,124,126,128,125,128,223,129,114,137,111,137,252,137,128,139,22,141,23,141,145,142,147,142,97,143,72,145,68,148,81,148,82,148,61,151,62,151,195,151,193,151,107,152,85,153,85,154,77,154,210,154,26,155,73,156,49,156,62,156,59,156,211,157,215,157,52,159,108,159,106,159,148,159,204,86,214,93,0,98,35,101,43,101,42,101,236,102,16,107,218,116,202,122,100,124,99,124,101,124,147,126,150,126,148,126,226,129,56,134,63,134,49,136,138,139,144,144,143,144,99,148,96,148,100,148,104,151,111,152,92,153,90,154,91,154,87,154,211,154,212,154,209,154,84,156,87,156,86,156,229,157,159,158,244,158,209,86,233,88,44,101,94,112,113,118,114,118,215,119,80,127,136,127,54,136,57,136,98,136,147,139,146,139,150,139,119,130,27,141,192,145,106,148,66,151,72,151,68,151,198,151,112,152,95,154,34,155,88,155,95,156,249,157,250,157,124,158,125,158,7,159,119,159,114,159,243,94,22,107,99,112,108,124,110,124,59,136,192,137,161,142,193,145,114,148,112,148,113,152,94,153,214,154,35,155,204,158,100,112,218,119,154,139,119,148,201,151,98,154,101,154,156,126,156,139,170,142,197,145,125,148,126,148,124,148,119,156,120,156,247,158,84,140,127,148,26,158,40,114,106,154,49,155,27,158,30,158,114,124,96,36,97,36,98,36,99,36,100,36,101,36,102,36,103,36,104,36,105,36,116,36,117,36,118,36,119,36,120,36,121,36,122,36,123,36,124,36,125,36,112,33,113,33,114,33,115,33,116,33,117,33,118,33,119,33,120,33,121,33,54,78,63,78,133,78,160,78,130,81,150,81,171,81,249,82,56,83,105,83,182,83,10,89,128,91,219,93,122,94,127,94,244,94,80,95,97,95,52,101,224,101,146,117,118,118,181,143,182,150,168,0,198,2,253,48,254,48,157,48,158,48,0,0,0,0,5,48,6,48,7,48,252,48,59,255,61,255,61,39,65,48,66,48,67,48,68,48,69,48,70,48,71,48,72,48,73,48,74,48,75,48,76,48,77,48,78,48,79,48,80,48,81,48,82,48,83,48,84,48,85,48,86,48,87,48,88,48,89,48,90,48,91,48,92,48,93,48,94,48,95,48,96,48,97,48,98,48,99,48,100,48,101,48,102,48,103,48,104,48,105,48,106,48,107,48,108,48,109,48,110,48,111,48,112,48,113,48,114,48,115,48,116,48,117,48,118,48,119,48,120,48,121,48,122,48,123,48,124,48,125,48,126,48,127,48,128,48,129,48,130,48,131,48,132,48,133,48,134,48,135,48,136,48,137,48,138,48,139,48,140,48,141,48,142,48,143,48,144,48,145,48,146,48,147,48,161,48,162,48,163,48,164,48,165,48,166,48,167,48,168,48,169,48,170,48,171,48,172,48,173,48,174,48,175,48,176,48,177,48,178,48,179,48,180,48,181,48,182,48,183,48,184,48,185,48,186,48,187,48,188,48,189,48,190,48,191,48,192,48,193,48,194,48,195,48,196,48,197,48,198,48,199,48,200,48,201,48,202,48,203,48,204,48,205,48,206,48,207,48,208,48,209,48,210,48,211,48,212,48,213,48,214,48,215,48,216,48,217,48,218,48,219,48,220,48,221,48,222,48,223,48,224,48,225,48,226,48,227,48,228,48,229,48,230,48,231,48,232,48,233,48,234,48,235,48,236,48,237,48,238,48,239,48,240,48,241,48,242,48,243,48,244,48,245,48,246,48,16,4,17,4,18,4,19,4,20,4,21,4,1,4,22,4,23,4,24,4,25,4,26,4,27,4,28,4,29,4,30,4,31,4,32,4,33,4,34,4,35,4,36,4,37,4,38,4,39,4,40,4,41,4,42,4,43,4,44,4,45,4,46,4,47,4,48,4,49,4,50,4,51,4,52,4,53,4,81,4,54,4,55,4,56,4,57,4,58,4,59,4,60,4,61,4,62,4,63,4,64,4,65,4,66,4,67,4,68,4,69,4,70,4,71,4,72,4,73,4,74,4,75,4,76,4,77,4,78,4,79,4,231,33,184,33,185,33,207,49,204,0,90,78,138,0,2,82,145,68,176,159,136,81,177,159,7,118,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,226,255,228,255,7,255,2,255,49,50,22,33,33,33,155,48,156,48,128,46,132,46,134,46,135,46,136,46,138,46,140,46,141,46,149,46,156,46,157,46,165,46,167,46,170,46,172,46,174,46,182,46,188,46,190,46,198,46,202,46,204,46,205,46,207,46,214,46,215,46,222,46,227,46,0,0,0,0,0,0,131,2,80,2,91,2,84,2,117,2,83,1,248,0,75,1,138,2,106,2,66,78,92,78,245,81,26,83,130,83,7,78,12,78,71,78,141,78,215,86,12,250,110,92,115,95,15,78,135,81,14,78,46,78,147,78,194,78,201,78,200,78,152,81,252,82,108,83,185,83,32,87,3,89,44,89,16,92,255,93,225,101,179,107,204,107,20,108,63,114,49,78,60,78,232,78,220,78,233,78,225,78,221,78,218,78,12,82,28,83,76,83,34,87,35,87,23,89,47,89,129,91,132,91,18,92,59,92,116,92,115,92,4,94,128,94,130,94,201,95,9,98,80,98,21,108,54,108,67,108,63,108,59,108,174,114,176,114,138,115,184,121,138,128,30,150,14,79,24,79,44,79,245,78,20,79,241,78,0,79,247,78,8,79,29,79,2,79,5,79,34,79,19,79,4,79,244,78,18,79,177,81,19,82,9,82,16,82,166,82,34,83,31,83,77,83,138,83,7,84,225,86,223,86,46,87,42,87,52,87,60,89,128,89,124,89,133,89,123,89,126,89,119,89,127,89,86,91,21,92,37,92,124,92,122,92,123,92,126,92,223,93,117,94,132,94,2,95,26,95,116,95,213,95,212,95,207,95,92,98,94,98,100,98,97,98,102,98,98,98,89,98,96,98,90,98,101,98,239,101,238,101,62,103,57,103,56,103,59,103,58,103,63,103,60,103,51,103,24,108,70,108,82,108,92,108,79,108,74,108,84,108,75,108,76,108,113,112,94,114,180,114,181,114,142,115,42,117,127,118,117,122,81,127,120,130,124,130,128,130,125,130,127,130,77,134,126,137,153,144,151,144,152,144,155,144,148,144,34,150,36,150,32,150,35,150,86,79,59,79,98,79,73,79,83,79,100,79,62,79,103,79,82,79,95,79,65,79,88,79,45,79,51,79,63,79,97,79,143,81,185,81,28,82,30,82,33,82,173,82,174,82,9,83,99,83,114,83,142,83,143,83,48,84,55,84,42,84,84,84,69,84,25,84,28,84,37,84,24,84,61,84,79,84,65,84,40,84,36,84,71,84,238,86,231,86,229,86,65,87,69,87,76,87,73,87,75,87,82,87,6,89,64,89,166,89,152,89,160,89,151,89,142,89,162,89,144,89,143,89,167,89,161,89,142,91,146,91,40,92,42,92,141,92,143,92,136,92,139,92,137,92,146,92,138,92,134,92,147,92,149,92,224,93,10,94,14,94,139,94,137,94,140,94,136,94,141,94,5,95,29,95,120,95,118,95,210,95,209,95,208,95,237,95,232,95,238,95,243,95,225,95,228,95,227,95,250,95,239,95,247,95,251,95,0,96,244,95,58,98,131,98,140,98,142,98,143,98,148,98,135,98,113,98,123,98,122,98,112,98,129,98,136,98,119,98,125,98,114,98,116,98,55,101,240,101,244,101,243,101,242,101,245,101,69,103,71,103,89,103,85,103,76,103,72,103,93,103,77,103,90,103,75,103,208,107,25,108,26,108,120,108,103,108,107,108,132,108,139,108,143,108,113,108,111,108,105,108,154,108,109,108,135,108,149,108,156,108,102,108,115,108,101,108,123,108,142,108,116,112,122,112,99,114,191,114,189,114,195,114,198,114,193,114,186,114,197,114,149,115,151,115,147,115,148,115,146,115,58,117,57,117,148,117,149,117,129,118,61,121,52,128,149,128,153,128,144,128,146,128,156,128,144,130,143,130,133,130,142,130,145,130,147,130,138,130,131,130,132,130,120,140,201,143,191,143,159,144,161,144,165,144,158,144,167,144,160,144,48,150,40,150,47,150,45,150,51,78,152,79,124,79,133,79,125,79,128,79,135,79,118,79,116,79,137,79,132,79,119,79,76,79,151,79,106,79,154,79,121,79,129,79,120,79,144,79,156,79,148,79,158,79,146,79,130,79,149,79,107,79,110,79,158,81,188,81,190,81,53,82,50,82,51,82,70,82,49,82,188,82,10,83,11,83,60,83,146,83,148,83,135,84,127,84,129,84,145,84,130,84,136,84,107,84,122,84,126,84,101,84,108,84,116,84,102,84,141,84,111,84,97,84,96,84,152,84,99,84,103,84,100,84,247,86,249,86,111,87,114,87,109,87,107,87,113,87,112,87,118,87,128,87,117,87,123,87,115,87,116,87,98,87,104,87,125,87,12,89,69,89,181,89,186,89,207,89,206,89,178,89,204,89,193,89,182,89,188,89,195,89,214,89,177,89,189,89,192,89,200,89,180,89,199,89,98,91,101,91,147,91,149,91,68,92,71,92,174,92,164,92,160,92,181,92,175,92,168,92,172,92,159,92,163,92,173,92,162,92,170,92,167,92,157,92,165,92,182,92,176,92,166,92,23,94,20,94,25,94,40,95,34,95,35,95,36,95,84,95,130,95,126,95,125,95,222,95,229,95,45,96,38,96,25,96,50,96,11,96,52,96,10,96,23,96,51,96,26,96,30,96,44,96,34,96,13,96,16,96,46,96,19,96,17,96,12,96,9,96,28,96,20,98,61,98,173,98,180,98,209,98,190,98,170,98,182,98,202,98,174,98,179,98,175,98,187,98,169,98,176,98,184,98,61,101,168,101,187,101,9,102,252,101,4,102,18,102,8,102,251,101,3,102,11,102,13,102,5,102,253,101,17,102,16,102,246,102,10,103,133,103,108,103,142,103,146,103,118,103,123,103,152,103,134,103,132,103,116,103,141,103,140,103,122,103,159,103,145,103,153,103,131,103,125,103,129,103,120,103,121,103,148,103,37,107,128,107,126,107,222,107,29,108,147,108,236,108,235,108,238,108,217,108,182,108,212,108,173,108,231,108,183,108,208,108,194,108,186,108,195,108,198,108,237,108,242,108,210,108,221,108,180,108,138,108,157,108,128,108,222,108,192,108,48,109,205,108,199,108,176,108,249,108,207,108,233,108,209,108,148,112,152,112,133,112,147,112,134,112,132,112,145,112,150,112,130,112,154,112,131,112,106,114,214,114,203,114,216,114,201,114,220,114,210,114,212,114,218,114,204,114,209,114,164,115,161,115,173,115,166,115,162,115,160,115,172,115,157,115,221,116,232,116,63,117,64,117,62,117,140,117,152,117,175,118,243,118,241,118,240,118,245,118,248,119,252,119,249,119,251,119,250,119,247,119,66,121,63,121,197,121,120,122,123,122,251,122,117,124,253,124,53,128,143,128,174,128,163,128,184,128,181,128,173,128,32,130,160,130,192,130,171,130,154,130,152,130,155,130,181,130,167,130,174,130,188,130,158,130,186,130,180,130,168,130,161,130,169,130,194,130,164,130,195,130,182,130,162,130,112,134,111,134,109,134,110,134,86,140,210,143,203,143,211,143,205,143,214,143,213,143,215,143,178,144,180,144,175,144,179,144,176,144,57,150,61,150,60,150,58,150,67,150,205,79,197,79,211,79,178,79,201,79,203,79,193,79,212,79,220,79,217,79,187,79,179,79,219,79,199,79,214,79,186,79,192,79,185,79,236,79,68,82,73,82,192,82,194,82,61,83,124,83,151,83,150,83,153,83,152,83,186,84,161,84,173,84,165,84,207,84,195,84,13,131,183,84,174,84,214,84,182,84,197,84,198,84,160,84,112,84,188,84,162,84,190,84,114,84,222,84,176,84,181,87,158,87,159,87,164,87,140,87,151,87,157,87,155,87,148,87,152,87,143,87,153,87,165,87,154,87,149,87,244,88,13,89,83,89,225,89,222,89,238,89,0,90,241,89,221,89,250,89,253,89,252,89,246,89,228,89,242,89,247,89,219,89,233,89,243,89,245,89,224,89,254,89,244,89,237,89,168,91,76,92,208,92,216,92,204,92,215,92,203,92,219,92,222,92,218,92,201,92,199,92,202,92,214,92,211,92,212,92,207,92,200,92,198,92,206,92,223,92,248,92,249,93,33,94,34,94,35,94,32,94,36,94,176,94,164,94,162,94,155,94,163,94,165,94,7,95,46,95,86,95,134,95,55,96,57,96,84,96,114,96,94,96,69,96,83,96,71,96,73,96,91,96,76,96,64,96,66,96,95,96,36,96,68,96,88,96,102,96,110,96,66,98,67,98,207,98,13,99,11,99,245,98,14,99,3,99,235,98,249,98,15,99,12,99,248,98,246,98,0,99,19,99,20,99,250,98,21,99,251,98,240,98,65,101,67,101,170,101,191,101,54,102,33,102,50,102,53,102,28,102,38,102,34,102,51,102,43,102,58,102,29,102,52,102,57,102,46,102,15,103,16,103,193,103,242,103,200,103,186,103,220,103,187,103,248,103,216,103,192,103,183,103,197,103,235,103,228,103,223,103,181,103,205,103,179,103,247,103,246,103,238,103,227,103,194,103,185,103,206,103,231,103,240,103,178,103,252,103,198,103,237,103,204,103,174,103,230,103,219,103,250,103,201,103,202,103,195,103,234,103,203,103,40,107,130,107,132,107,182,107,214,107,216,107,224,107,32,108,33,108,40,109,52,109,45,109,31,109,60,109,63,109,18,109,10,109,218,108,51,109,4,109,25,109,58,109,26,109,17,109,0,109,29,109,66,109,1,109,24,109,55,109,3,109,15,109,64,109,7,109,32,109,44,109,8,109,34,109,9,109,16,109,183,112,159,112,190,112,177,112,176,112,161,112,180,112,181,112,169,112,65,114,73,114,74,114,108,114,112,114,115,114,110,114,202,114,228,114,232,114,235,114,223,114,234,114,230,114,227,114,133,115,204,115,194,115,200,115,197,115,185,115,182,115,181,115,180,115,235,115,191,115,199,115,190,115,195,115,198,115,184,115,203,115,236,116,238,116,46,117,71,117,72,117,167,117,170,117,121,118,196,118,8,119,3,119,4,119,5,119,10,119,247,118,251,118,250,118,231,119,232,119,6,120,17,120,18,120,5,120,16,120,15,120,14,120,9,120,3,120,19,120,74,121,76,121,75,121,69,121,68,121,213,121,205,121,207,121,214,121,206,121,128,122,126,122,209,122,0,123,1,123,122,124,120,124,121,124,127,124,128,124,129,124,3,125,8,125,1,125,88,127,145,127,141,127,190,127,7,128,14,128,15,128,20,128,55,128,216,128,199,128,224,128,209,128,200,128,194,128,208,128,197,128,227,128,217,128,220,128,202,128,213,128,201,128,207,128,215,128,230,128,205,128,255,129,33,130,148,130,217,130,254,130,249,130,7,131,232,130,0,131,213,130,58,131,235,130,214,130,244,130,236,130,225,130,242,130,245,130,12,131,251,130,246,130,240,130,234,130,228,130,224,130,250,130,243,130,237,130,119,134,116,134,124,134,115,134,65,136,78,136,103,136,106,136,105,136,211,137,4,138,7,138,114,141,227,143,225,143,238,143,224,143,241,144,189,144,191,144,213,144,197,144,190,144,199,144,203,144,200,144,212,145,211,145,84,150,79,150,81,150,83,150,74,150,78,150,30,80,5,80,7,80,19,80,34,80,48,80,27,80,245,79,244,79,51,80,55,80,44,80,246,79,247,79,23,80,28,80,32,80,39,80,53,80,47,80,49,80,14,80,90,81,148,81,147,81,202,81,196,81,197,81,200,81,206,81,97,82,90,82,82,82,94,82,95,82,85,82,98,82,205,82,14,83,158,83,38,85,226,84,23,85,18,85,231,84,243,84,228,84,26,85,255,84,4,85,8,85,235,84,17,85,5,85,241,84,10,85,251,84,247,84,248,84,224,84,14,85,3,85,11,85,1,87,2,87,204,87,50,88,213,87,210,87,186,87,198,87,189,87,188,87,184,87,182,87,191,87,199,87,208,87,185,87,193,87,14,89,74,89,25,90,22,90,45,90,46,90,21,90,15,90,23,90,10,90,30,90,51,90,108,91,167,91,173,91,172,91,3,92,86,92,84,92,236,92,255,92,238,92,241,92,247,92,0,93,249,92,41,94,40,94,168,94,174,94,170,94,172,94,51,95,48,95,103,95,93,96,90,96,103,96,65,96,162,96,136,96,128,96,146,96,129,96,157,96,131,96,149,96,155,96,151,96,135,96,156,96,142,96,25,98,70,98,242,98,16,99,86,99,44,99,68,99,69,99,54,99,67,99,228,99,57,99,75,99,74,99,60,99,41,99,65,99,52,99,88,99,84,99,89,99,45,99,71,99,51,99,90,99,81,99,56,99,87,99,64,99,72,99,74,101,70,101,198,101,195,101,196,101,194,101,74,102,95,102,71,102,81,102,18,103,19,103,31,104,26,104,73,104,50,104,51,104,59,104,75,104,79,104,22,104,49,104,28,104,53,104,43,104,45,104,47,104,78,104,68,104,52,104,29,104,18,104,20,104,38,104,40,104,46,104,77,104,58,104,37,104,32,104,44,107,47,107,45,107,49,107,52,107,109,107,130,128,136,107,230,107,228,107,232,107,227,107,226,107,231,107,37,108,122,109,99,109,100,109,118,109,13,109,97,109,146,109,88,109,98,109,109,109,111,109,145,109,141,109,239,109,127,109,134,109,94,109,103,109,96,109,151,109,112,109,124,109,95,109,130,109,152,109,47,109,104,109,139,109,126,109,128,109,132,109,22,109,131,109,123,109,125,109,117,109,144,109,220,112,211,112,209,112,221,112,203,112,57,127,226,112,215,112,210,112,222,112,224,112,212,112,205,112,197,112,198,112,199,112,218,112,206,112,225,112,66,114,120,114,119,114,118,114,0,115,250,114,244,114,254,114,246,114,243,114,251,114,1,115,211,115,217,115,229,115,214,115,188,115,231,115,227,115,233,115,220,115,210,115,219,115,212,115,221,115,218,115,215,115,216,115,232,115,222,116,223,116,244,116,245,116,33,117,91,117,95,117,176,117,193,117,187,117,196,117,192,117,191,117,182,117,186,117,138,118,201,118,29,119,27,119,16,119,19,119,18,119,35,119,17,119,21,119,25,119,26,119,34,119,39,119,35,120,44,120,34,120,53,120,47,120,40,120,46,120,43,120,33,120,41,120,51,120,42,120,49,120,84,121,91,121,79,121,92,121,83,121,82,121,81,121,235,121,236,121,224,121,238,121,237,121,234,121,220,121,222,121,221,121,134,122,137,122,133,122,139,122,140,122,138,122,135,122,216,122,16,123,4,123,19,123,5,123,15,123,8,123,10,123,14,123,9,123,18,123,132,124,145,124,138,124,140,124,136,124,141,124,133,124,30,125,29,125,17,125,14,125,24,125,22,125,19,125,31,125,18,125,15,125,12,125,92,127,97,127,94,127,96,127,93,127,91,127,150,127,146,127,195,127,194,127,192,127,22,128,62,128,57,128,250,128,242,128,249,128,245,128,1,129,251,128,0,129,1,130,47,130,37,130,51,131,45,131,68,131,25,131,81,131,37,131,86,131,63,131,65,131,38,131,28,131,34,131,66,131,78,131,27,131,42,131,8,131,60,131,77,131,22,131,36,131,32,131,55,131,47,131,41,131,71,131,69,131,76,131,83,131,30,131,44,131,75,131,39,131,72,131,83,134,82,134,162,134,168,134,150,134,141,134,145,134,158,134,135,134,151,134,134,134,139,134,154,134,133,134,165,134,153,134,161,134,167,134,149,134,152,134,142,134,157,134,144,134,148,134,67,136,68,136,109,136,117,136,118,136,114,136,128,136,113,136,127,136,111,136,131,136,126,136,116,136,124,136,18,138,71,140,87,140,123,140,164,140,163,140,118,141,120,141,181,141,183,141,182,141,209,142,211,142,254,143,245,143,2,144,255,143,251,143,4,144,252,143,246,143,214,144,224,144,217,144,218,144,227,144,223,144,229,144,216,144,219,144,215,144,220,144,228,144,80,145,78,145,79,145,213,145,226,145,218,145,92,150,95,150,188,150,227,152,223,154,47,155,127,78,112,80,106,80,97,80,94,80,96,80,83,80,75,80,93,80,114,80,72,80,77,80,65,80,91,80,74,80,98,80,21,80,69,80,95,80,105,80,107,80,99,80,100,80,70,80,64,80,110,80,115,80,87,80,81,80,208,81,107,82,109,82,108,82,110,82,214,82,211,82,45,83,156,83,117,85,118,85,60,85,77,85,80,85,52,85,42,85,81,85,98,85,54,85,53,85,48,85,82,85,69,85,12,85,50,85,101,85,78,85,57,85,72,85,45,85,59,85,64,85,75,85,10,87,7,87,251,87,20,88,226,87,246,87,220,87,244,87,0,88,237,87,253,87,8,88,248,87,11,88,243,87,207,87,7,88,238,87,227,87,242,87,229,87,236,87,225,87,14,88,252,87,16,88,231,87,1,88,12,88,241,87,233,87,240,87,13,88,4,88,92,89,96,90,88,90,85,90,103,90,94,90,56,90,53,90,109,90,80,90,95,90,101,90,108,90,83,90,100,90,87,90,67,90,93,90,82,90,68,90,91,90,72,90,142,90,62,90,77,90,57,90,76,90,112,90,105,90,71,90,81,90,86,90,66,90,92,90,114,91,110,91,193,91,192,91,89,92,30,93,11,93,29,93,26,93,32,93,12,93,40,93,13,93,38,93,37,93,15,93,48,93,18,93,35,93,31,93,46,93,62,94,52,94,177,94,180,94,185,94,178,94,179,94,54,95,56,95,155,95,150,95,159,95,138,96,144,96,134,96,190,96,176,96,186,96,211,96,212,96,207,96,228,96,217,96,221,96,200,96,177,96,219,96,183,96,202,96,191,96,195,96,205,96,192,96,50,99,101,99,138,99,130,99,125,99,189,99,158,99,173,99,157,99,151,99,171,99,142,99,111,99,135,99,144,99,110,99,175,99,117,99,156,99,109,99,174,99,124,99,164,99,59,99,159,99,120,99,133,99,129,99,145,99,141,99,112,99,83,101,205,101,101,102,97,102,91,102,89,102,92,102,98,102,24,103,121,104,135,104,144,104,156,104,109,104,110,104,174,104,171,104,86,105,111,104,163,104,172,104,169,104,117,104,116,104,178,104,143,104,119,104,146,104,124,104,107,104,114,104,170,104,128,104,113,104,126,104,155,104,150,104,139,104,160,104,137,104,164,104,120,104,123,104,145,104,140,104,138,104,125,104,54,107,51,107,55,107,56,107,145,107,143,107,141,107,142,107,140,107,42,108,192,109,171,109,180,109,179,109,116,110,172,109,233,109,226,109,183,109,246,109,212,109,0,110,200,109,224,109,223,109,214,109,190,109,229,109,220,109,221,109,219,109,244,109,202,109,189,109,237,109,240,109,186,109,213,109,194,109,207,109,201,109,208,109,242,109,211,109,253,109,215,109,205,109,227,109,187,109,250,112,13,113,247,112,23,113,244,112,12,113,240,112,4,113,243,112,16,113,252,112,255,112,6,113,19,113,0,113,248,112,246,112,11,113,2,113,14,113,126,114,123,114,124,114,127,114,29,115,23,115,7,115,17,115,24,115,10,115,8,115,255,114,15,115,30,115,136,115,246,115,248,115,245,115,4,116,1,116,253,115,7,116,0,116,250,115,252,115,255,115,12,116,11,116,244,115,8,116,100,117,99,117,206,117,210,117,207,117,203,117,204,117,209,117,208,117,143,118,137,118,211,118,57,119,47,119,45,119,49,119,50,119,52,119,51,119,61,119,37,119,59,119,53,119,72,120,82,120,73,120,77,120,74,120,76,120,38,120,69,120,80,120,100,121,103,121,105,121,106,121,99,121,107,121,97,121,187,121,250,121,248,121,246,121,247,121,143,122,148,122,144,122,53,123,71,123,52,123,37,123,48,123,34,123,36,123,51,123,24,123,42,123,29,123,49,123,43,123,45,123,47,123,50,123,56,123,26,123,35,123,148,124,152,124,150,124,163,124,53,125,61,125,56,125,54,125,58,125,69,125,44,125,41,125,65,125,71,125,62,125,63,125,74,125,59,125,40,125,99,127,149,127,156,127,157,127,155,127,202,127,203,127,205,127,208,127,209,127,199,127,207,127,201,127,31,128,30,128,27,128,71,128,67,128,72,128,24,129,37,129,25,129,27,129,45,129,31,129,44,129,30,129,33,129,21,129,39,129,29,129,34,129,17,130,56,130,51,130,58,130,52,130,50,130,116,130,144,131,163,131,168,131,141,131,122,131,115,131,164,131,116,131,143,131,129,131,149,131,153,131,117,131,148,131,169,131,125,131,131,131,140,131,157,131,155,131,170,131,139,131,126,131,165,131,175,131,136,131,151,131,176,131,127,131,166,131,135,131,174,131,118,131,154,131,89,134,86,134,191,134,183,134,194,134,193,134,197,134,186,134,176,134,200,134,185,134,179,134,184,134,204,134,180,134,187,134,188,134,195,134,189,134,190,134,82,136,137,136,149,136,168,136,162,136,170,136,154,136,145,136,161,136,159,136,152,136,167,136,153,136,155,136,151,136,164,136,172,136,140,136,147,136,142,136,130,137,214,137,217,137,213,137,48,138,39,138,44,138,30,138,57,140,59,140,92,140,93,140,125,140,165,140,125,141,123,141,121,141,188,141,194,141,185,141,191,141,193,141,216,142,222,142,221,142,220,142,215,142,224,142,225,142,36,144,11,144,17,144,28,144,12,144,33,144,239,144,234,144,240,144,244,144,242,144,243,144,212,144,235,144,236,144,233,144,86,145,88,145,90,145,83,145,85,145,236,145,244,145,241,145,243,145,248,145,228,145,249,145,234,145,235,145,247,145,232,145,238,145,122,149,134,149,136,149,124,150,109,150,107,150,113,150,111,150,191,150,106,151,4,152,229,152,151,153,155,80,149,80,148,80,158,80,139,80,163,80,131,80,140,80,142,80,157,80,104,80,156,80,146,80,130,80,135,80,95,81,212,81,18,83,17,83,164,83,167,83,145,85,168,85,165,85,173,85,119,85,69,86,162,85,147,85,136,85,143,85,181,85,129,85,163,85,146,85,164,85,125,85,140,85,166,85,127,85,149,85,161,85,142,85,12,87,41,88,55,88,25,88,30,88,39,88,35,88,40,88,245,87,72,88,37,88,28,88,27,88,51,88,63,88,54,88,46,88,57,88,56,88,45,88,44,88,59,88,97,89,175,90,148,90,159,90,122,90,162,90,158,90,120,90,166,90,124,90,165,90,172,90,149,90,174,90,55,90,132,90,138,90,151,90,131,90,139,90,169,90,123,90,125,90,140,90,156,90,143,90,147,90,157,90,234,91,205,91,203,91,212,91,209,91,202,91,206,91,12,92,48,92,55,93,67,93,107,93,65,93,75,93,63,93,53,93,81,93,78,93,85,93,51,93,58,93,82,93,61,93,49,93,89,93,66,93,57,93,73,93,56,93,60,93,50,93,54,93,64,93,69,93,68,94,65,94,88,95,166,95,165,95,171,95,201,96,185,96,204,96,226,96,206,96,196,96,20,97,242,96,10,97,22,97,5,97,245,96,19,97,248,96,252,96,254,96,193,96,3,97,24,97,29,97,16,97,255,96,4,97,11,97,74,98,148,99,177,99,176,99,206,99,229,99,232,99,239,99,195,99,157,100,243,99,202,99,224,99,246,99,213,99,242,99,245,99,97,100,223,99,190,99,221,99,220,99,196,99,216,99,211,99,194,99,199,99,204,99,203,99,200,99,240,99,215,99,217,99,50,101,103,101,106,101,100,101,92,101,104,101,101,101,140,101,157,101,158,101,174,101,208,101,210,101,124,102,108,102,123,102,128,102,113,102,121,102,106,102,114,102,1,103,12,105,211,104,4,105,220,104,42,105,236,104,234,104,241,104,15,105,214,104,247,104,235,104,228,104,246,104,19,105,16,105,243,104,225,104,7,105,204,104,8,105,112,105,180,104,17,105,239,104,198,104,20,105,248,104,208,104,253,104,252,104,232,104,11,105,10,105,23,105,206,104,200,104,221,104,222,104,230,104,244,104,209,104,6,105,212,104,233,104,21,105,37,105,199,104,57,107,59,107,63,107,60,107,148,107,151,107,153,107,149,107,189,107,240,107,242,107,243,107,48,108,252,109,70,110,71,110,31,110,73,110,136,110,60,110,61,110,69,110,98,110,43,110,63,110,65,110,93,110,115,110,28,110,51,110,75,110,64,110,81,110,59,110,3,110,46,110,94,110,104,110,92,110,97,110,49,110,40,110,96,110,113,110,107,110,57,110,34,110,48,110,83,110,101,110,39,110,120,110,100,110,119,110,85,110,121,110,82,110,102,110,53,110,54,110,90,110,32,113,30,113,47,113,251,112,46,113,49,113,35,113,37,113,34,113,50,113,31,113,40,113,58,113,27,113,75,114,90,114,136,114,137,114,134,114,133,114,139,114,18,115,11,115,48,115,34,115,49,115,51,115,39,115,50,115,45,115,38,115,35,115,53,115,12,115,46,116,44,116,48,116,43,116,22,116,26,116,33,116,45,116,49,116,36,116,35,116,29,116,41,116,32,116,50,116,251,116,47,117,111,117,108,117,231,117,218,117,225,117,230,117,221,117,223,117,228,117,215,117,149,118,146,118,218,118,70,119,71,119,68,119,77,119,69,119,74,119,78,119,75,119,76,119,222,119,236,119,96,120,100,120,101,120,92,120,109,120,113,120,106,120,110,120,112,120,105,120,104,120,94,120,98,120,116,121,115,121,114,121,112,121,2,122,10,122,3,122,12,122,4,122,153,122,230,122,228,122,74,123,59,123,68,123,72,123,76,123,78,123,64,123,88,123,69,123,162,124,158,124,168,124,161,124,88,125,111,125,99,125,83,125,86,125,103,125,106,125,79,125,109,125,92,125,107,125,82,125,84,125,105,125,81,125,95,125,78,125,62,127,63,127,101,127,102,127,162,127,160,127,161,127,215,127,81,128,79,128,80,128,254,128,212,128,67,129,74,129,82,129,79,129,71,129,61,129,77,129,58,129,230,129,238,129,247,129,248,129,249,129,4,130,60,130,61,130,63,130,117,130,59,131,207,131,249,131,35,132,192,131,232,131,18,132,231,131,228,131,252,131,246,131,16,132,198,131,200,131,235,131,227,131,191,131,1,132,221,131,229,131,216,131,255,131,225,131,203,131,206,131,214,131,245,131,201,131,9,132,15,132,222,131,17,132,6,132,194,131,243,131,213,131,250,131,199,131,209,131,234,131,19,132,195,131,236,131,238,131,196,131,251,131,215,131,226,131,27,132,219,131,254,131,216,134,226,134,230,134,211,134,227,134,218,134,234,134,221,134,235,134,220,134,236,134,233,134,215,134,232,134,209,134,72,136,86,136,85,136,186,136,215,136,185,136,184,136,192,136,190,136,182,136,188,136,183,136,189,136,178,136,1,137,201,136,149,137,152,137,151,137,221,137,218,137,219,137,78,138,77,138,57,138,89,138,64,138,87,138,88,138,68,138,69,138,82,138,72,138,81,138,74,138,76,138,79,138,95,140,129,140,128,140,186,140,190,140,176,140,185,140,181,140,132,141,128,141,137,141,216,141,211,141,205,141,199,141,214,141,220,141,207,141,213,141,217,141,200,141,215,141,197,141,239,142,247,142,250,142,249,142,230,142,238,142,229,142,245,142,231,142,232,142,246,142,235,142,241,142,236,142,244,142,233,142,45,144,52,144,47,144,6,145,44,145,4,145,255,144,252,144,8,145,249,144,251,144,1,145,0,145,7,145,5,145,3,145,97,145,100,145,95,145,98,145,96,145,1,146,10,146,37,146,3,146,26,146,38,146,15,146,12,146,0,146,18,146,255,145,253,145,6,146,4,146,39,146,2,146,28,146,36,146,25,146,23,146,5,146,22,146,123,149,141,149,140,149,144,149,135,150,126,150,136,150,137,150,131,150,128,150,194,150,200,150,195,150,241,150,240,150,108,151,112,151,110,151,7,152,169,152,235,152,230,156,249,158,131,78,132,78,182,78,189,80,191,80,198,80,174,80,196,80,202,80,180,80,200,80,194,80,176,80,193,80,186,80,177,80,203,80,201,80,182,80,184,80,215,81,122,82,120,82,123,82,124,82,195,85,219,85,204,85,208,85,203,85,202,85,221,85,192,85,212,85,196,85,233,85,191,85,210,85,141,85,207,85,213,85,226,85,214,85,200,85,242,85,205,85,217,85,194,85,20,87,83,88,104,88,100,88,79,88,77,88,73,88,111,88,85,88,78,88,93,88,89,88,101,88,91,88,61,88,99,88,113,88,252,88,199,90,196,90,203,90,186,90,184,90,177,90,181,90,176,90,191,90,200,90,187,90,198,90,183,90,192,90,202,90,180,90,182,90,205,90,185,90,144,90,214,91,216,91,217,91,31,92,51,92,113,93,99,93,74,93,101,93,114,93,108,93,94,93,104,93,103,93,98,93,240,93,79,94,78,94,74,94,77,94,75,94,197,94,204,94,198,94,203,94,199,94,64,95,175,95,173,95,247,96,73,97,74,97,43,97,69,97,54,97,50,97,46,97,70,97,47,97,79,97,41,97,64,97,32,98,104,145,35,98,37,98,36,98,197,99,241,99,235,99,16,100,18,100,9,100,32,100,36,100,51,100,67,100,31,100,21,100,24,100,57,100,55,100,34,100,35,100,12,100,38,100,48,100,40,100,65,100,53,100,47,100,10,100,26,100,64,100,37,100,39,100,11,100,231,99,27,100,46,100,33,100,14,100,111,101,146,101,211,101,134,102,140,102,149,102,144,102,139,102,138,102,153,102,148,102,120,102,32,103,102,105,95,105,56,105,78,105,98,105,113,105,63,105,69,105,106,105,57,105,66,105,87,105,89,105,122,105,72,105,73,105,53,105,108,105,51,105,61,105,101,105,240,104,120,105,52,105,105,105,64,105,111,105,68,105,118,105,88,105,65,105,116,105,76,105,59,105,75,105,55,105,92,105,79,105,81,105,50,105,82,105,47,105,123,105,60,105,70,107,69,107,67,107,66,107,72,107,65,107,155,107,13,250,251,107,252,107,249,107,247,107,248,107,155,110,214,110,200,110,143,110,192,110,159,110,147,110,148,110,160,110,177,110,185,110,198,110,210,110,189,110,193,110,158,110,201,110,183,110,176,110,205,110,166,110,207,110,178,110,190,110,195,110,220,110,216,110,153,110,146,110,142,110,141,110,164,110,161,110,191,110,179,110,208,110,202,110,151,110,174,110,163,110,71,113,84,113,82,113,99,113,96,113,65,113,93,113,98,113,114,113,120,113,106,113,97,113,66,113,88,113,67,113,75,113,112,113,95,113,80,113,83,113,68,113,77,113,90,113,79,114,141,114,140,114,145,114,144,114,142,114,60,115,66,115,59,115,58,115,64,115,74,115,73,115,68,116,74,116,75,116,82,116,81,116,87,116,64,116,79,116,80,116,78,116,66,116,70,116,77,116,84,116,225,116,255,116,254,116,253,116,29,117,121,117,119,117,131,105,239,117,15,118,3,118,247,117,254,117,252,117,249,117,248,117,16,118,251,117,246,117,237,117,245,117,253,117,153,118,181,118,221,118,85,119,95,119,96,119,82,119,86,119,90,119,105,119,103,119,84,119,89,119,109,119,224,119,135,120,154,120,148,120,143,120,132,120,149,120,133,120,134,120,161,120,131,120,121,120,153,120,128,120,150,120,123,120,124,121,130,121,125,121,121,121,17,122,24,122,25,122,18,122,23,122,21,122,34,122,19,122,27,122,16,122,163,122,162,122,158,122,235,122,102,123,100,123,109,123,116,123,105,123,114,123,101,123,115,123,113,123,112,123,97,123,120,123,118,123,99,123,178,124,180,124,175,124,136,125,134,125,128,125,141,125,127,125,133,125,122,125,142,125,123,125,131,125,124,125,140,125,148,125,132,125,125,125,146,125,109,127,107,127,103,127,104,127,108,127,166,127,165,127,167,127,219,127,220,127,33,128,100,129,96,129,119,129,92,129,105,129,91,129,98,129,114,129,33,103,94,129,118,129,103,129,111,129,68,129,97,129,29,130,73,130,68,130,64,130,66,130,69,130,241,132,63,132,86,132,118,132,121,132,143,132,141,132,101,132,81,132,64,132,134,132,103,132,48,132,77,132,125,132,90,132,89,132,116,132,115,132,93,132,7,133,94,132,55,132,58,132,52,132,122,132,67,132,120,132,50,132,69,132,41,132,217,131,75,132,47,132,66,132,45,132,95,132,112,132,57,132,78,132,76,132,82,132,111,132,197,132,142,132,59,132,71,132,54,132,51,132,104,132,126,132,68,132,43,132,96,132,84,132,110,132,80,132,11,135,4,135,247,134,12,135,250,134,214,134,245,134,77,135,248,134,14,135,9,135,1,135,246,134,13,135,5,135,214,136,203,136,205,136,206,136,222,136,219,136,218,136,204,136,208,136,133,137,155,137,223,137,229,137,228,137,225,137,224,137,226,137,220,137,230,137,118,138,134,138,127,138,97,138,63,138,119,138,130,138,132,138,117,138,131,138,129,138,116,138,122,138,60,140,75,140,74,140,101,140,100,140,102,140,134,140,132,140,133,140,204,140,104,141,105,141,145,141,140,141,142,141,143,141,141,141,147,141,148,141,144,141,146,141,240,141,224,141,236,141,241,141,238,141,208,141,233,141,227,141,226,141,231,141,242,141,235,141,244,141,6,143,255,142,1,143,0,143,5,143,7,143,8,143,2,143,11,143,82,144,63,144,68,144,73,144,61,144,16,145,13,145,15,145,17,145,22,145,20,145,11,145,14,145,110,145,111,145,72,146,82,146,48,146,58,146,102,146,51,146,101,146,94,146,131,146,46,146,74,146,70,146,109,146,108,146,79,146,96,146,103,146,111,146,54,146,97,146,112,146,49,146,84,146,99,146,80,146,114,146,78,146,83,146,76,146,86,146,50,146,159,149,156,149,158,149,155,149,146,150,147,150,145,150,151,150,206,150,250,150,253,150,248,150,245,150,115,151,119,151,120,151,114,151,15,152,13,152,14,152,172,152,246,152,249,152,175,153,178,153,176,153,181,153,173,154,171,154,91,155,234,156,237,156,231,156,128,158,253,158,230,80,212,80,215,80,232,80,243,80,219,80,234,80,221,80,228,80,211,80,236,80,240,80,239,80,227,80,224,80,216,81,128,82,129,82,233,82,235,82,48,83,172,83,39,86,21,86,12,86,18,86,252,85,15,86,28,86,1,86,19,86,2,86,250,85,29,86,4,86,255,85,249,85,137,88,124,88,144,88,152,88,134,88,129,88,127,88,116,88,139,88,122,88,135,88,145,88,142,88,118,88,130,88,136,88,123,88,148,88,143,88,254,88,107,89,220,90,238,90,229,90,213,90,234,90,218,90,237,90,235,90,243,90,226,90,224,90,219,90,236,90,222,90,221,90,217,90,232,90,223,90,119,91,224,91,227,91,99,92,130,93,128,93,125,93,134,93,122,93,129,93,119,93,138,93,137,93,136,93,126,93,124,93,141,93,121,93,127,93,88,94,89,94,83,94,216,94,209,94,215,94,206,94,220,94,213,94,217,94,210,94,212,94,68,95,67,95,111,95,182,95,44,97,40,97,65,97,94,97,113,97,115,97,82,97,83,97,114,97,108,97,128,97,116,97,84,97,122,97,91,97,101,97,59,97,106,97,97,97,86,97,41,98,39,98,43,98,43,100,77,100,91,100,93,100,116,100,118,100,114,100,115,100,125,100,117,100,102,100,166,100,78,100,130,100,94,100,92,100,75,100,83,100,96,100,80,100,127,100,63,100,108,100,107,100,89,100,101,100,119,100,115,101,160,101,161,102,160,102,159,102,5,103,4,103,34,103,177,105,182,105,201,105,160,105,206,105,150,105,176,105,172,105,188,105,145,105,153,105,142,105,167,105,141,105,169,105,190,105,175,105,191,105,196,105,189,105,164,105,212,105,185,105,202,105,154,105,207,105,179,105,147,105,170,105,161,105,158,105,217,105,151,105,144,105,194,105,181,105,165,105,198,105,74,107,77,107,75,107,158,107,159,107,160,107,195,107,196,107,254,107,206,110,245,110,241,110,3,111,37,111,248,110,55,111,251,110,46,111,9,111,78,111,25,111,26,111,39,111,24,111,59,111,18,111,237,110,10,111,54,111,115,111,249,110,238,110,45,111,64,111,48,111,60,111,53,111,235,110,7,111,14,111,67,111,5,111,253,110,246,110,57,111,28,111,252,110,58,111,31,111,13,111,30,111,8,111,33,111,135,113,144,113,137,113,128,113,133,113,130,113,143,113,123,113,134,113,129,113,151,113,68,114,83,114,151,114,149,114,147,114,67,115,77,115,81,115,76,115,98,116,115,116,113,116,117,116,114,116,103,116,110,116,0,117,2,117,3,117,125,117,144,117,22,118,8,118,12,118,21,118,17,118,10,118,20,118,184,118,129,119,124,119,133,119,130,119,110,119,128,119,111,119,126,119,131,119,178,120,170,120,180,120,173,120,168,120,126,120,171,120,158,120,165,120,160,120,172,120,162,120,164,120,152,121,138,121,139,121,150,121,149,121,148,121,147,121,151,121,136,121,146,121,144,121,43,122,74,122,48,122,47,122,40,122,38,122,168,122,171,122,172,122,238,122,136,123,156,123,138,123,145,123,144,123,150,123,141,123,140,123,155,123,142,123,133,123,152,123,132,82,153,123,164,123,130,123,187,124,191,124,188,124,186,124,167,125,183,125,194,125,163,125,170,125,193,125,192,125,197,125,157,125,206,125,196,125,198,125,203,125,204,125,175,125,185,125,150,125,188,125,159,125,166,125,174,125,169,125,161,125,201,125,115,127,226,127,227,127,229,127,222,127,36,128,93,128,92,128,137,129,134,129,131,129,135,129,141,129,140,129,139,129,21,130,151,132,164,132,161,132,159,132,186,132,206,132,194,132,172,132,174,132,171,132,185,132,180,132,193,132,205,132,170,132,154,132,177,132,208,132,157,132,167,132,187,132,162,132,148,132,199,132,204,132,155,132,169,132,175,132,168,132,214,132,152,132,182,132,207,132,160,132,215,132,212,132,210,132,219,132,176,132,145,132,97,134,51,135,35,135,40,135,107,135,64,135,46,135,30,135,33,135,25,135,27,135,67,135,44,135,65,135,62,135,70,135,32,135,50,135,42,135,45,135,60,135,18,135,58,135,49,135,53,135,66,135], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+107808);
/* memory initializer */ allocate([38,135,39,135,56,135,36,135,26,135,48,135,17,135,247,136,231,136,241,136,242,136,250,136,254,136,238,136,252,136,246,136,251,136,240,136,236,136,235,136,157,137,161,137,159,137,158,137,233,137,235,137,232,137,171,138,153,138,139,138,146,138,143,138,150,138,61,140,104,140,105,140,213,140,207,140,215,140,150,141,9,142,2,142,255,141,13,142,253,141,10,142,3,142,7,142,6,142,5,142,254,141,0,142,4,142,16,143,17,143,14,143,13,143,35,145,28,145,32,145,34,145,31,145,29,145,26,145,36,145,33,145,27,145,122,145,114,145,121,145,115,145,165,146,164,146,118,146,155,146,122,146,160,146,148,146,170,146,141,146,166,146,154,146,171,146,121,146,151,146,127,146,163,146,238,146,142,146,130,146,149,146,162,146,125,146,136,146,161,146,138,146,134,146,140,146,153,146,167,146,126,146,135,146,169,146,157,146,139,146,45,146,158,150,161,150,255,150,88,151,125,151,122,151,126,151,131,151,128,151,130,151,123,151,132,151,129,151,127,151,206,151,205,151,22,152,173,152,174,152,2,153,0,153,7,153,157,153,156,153,195,153,185,153,187,153,186,153,194,153,189,153,199,153,177,154,227,154,231,154,62,155,63,155,96,155,97,155,95,155,241,156,242,156,245,156,167,158,255,80,3,81,48,81,248,80,6,81,7,81,246,80,254,80,11,81,12,81,253,80,10,81,139,82,140,82,241,82,239,82,72,86,66,86,76,86,53,86,65,86,74,86,73,86,70,86,88,86,90,86,64,86,51,86,61,86,44,86,62,86,56,86,42,86,58,86,26,87,171,88,157,88,177,88,160,88,163,88,175,88,172,88,165,88,161,88,255,88,255,90,244,90,253,90,247,90,246,90,3,91,248,90,2,91,249,90,1,91,7,91,5,91,15,91,103,92,153,93,151,93,159,93,146,93,162,93,147,93,149,93,160,93,156,93,161,93,154,93,158,93,105,94,93,94,96,94,92,94,243,125,219,94,222,94,225,94,73,95,178,95,139,97,131,97,121,97,177,97,176,97,162,97,137,97,155,97,147,97,175,97,173,97,159,97,146,97,170,97,161,97,141,97,102,97,179,97,45,98,110,100,112,100,150,100,160,100,133,100,151,100,156,100,143,100,139,100,138,100,140,100,163,100,159,100,104,100,177,100,152,100,118,101,122,101,121,101,123,101,178,101,179,101,181,102,176,102,169,102,178,102,183,102,170,102,175,102,0,106,6,106,23,106,229,105,248,105,21,106,241,105,228,105,32,106,255,105,236,105,226,105,27,106,29,106,254,105,39,106,242,105,238,105,20,106,247,105,231,105,64,106,8,106,230,105,251,105,13,106,252,105,235,105,9,106,4,106,24,106,37,106,15,106,246,105,38,106,7,106,244,105,22,106,81,107,165,107,163,107,162,107,166,107,1,108,0,108,255,107,2,108,65,111,38,111,126,111,135,111,198,111,146,111,141,111,137,111,140,111,98,111,79,111,133,111,90,111,150,111,118,111,108,111,130,111,85,111,114,111,82,111,80,111,87,111,148,111,147,111,93,111,0,111,97,111,107,111,125,111,103,111,144,111,83,111,139,111,105,111,127,111,149,111,99,111,119,111,106,111,123,111,178,113,175,113,155,113,176,113,160,113,154,113,169,113,181,113,157,113,165,113,158,113,164,113,161,113,170,113,156,113,167,113,179,113,152,114,154,114,88,115,82,115,94,115,95,115,96,115,93,115,91,115,97,115,90,115,89,115,98,115,135,116,137,116,138,116,134,116,129,116,125,116,133,116,136,116,124,116,121,116,8,117,7,117,126,117,37,118,30,118,25,118,29,118,28,118,35,118,26,118,40,118,27,118,156,118,157,118,158,118,155,118,141,119,143,119,137,119,136,119,205,120,187,120,207,120,204,120,209,120,206,120,212,120,200,120,195,120,196,120,201,120,154,121,161,121,160,121,156,121,162,121,155,121,118,107,57,122,178,122,180,122,179,122,183,123,203,123,190,123,172,123,206,123,175,123,185,123,202,123,181,123,197,124,200,124,204,124,203,124,247,125,219,125,234,125,231,125,215,125,225,125,3,126,250,125,230,125,246,125,241,125,240,125,238,125,223,125,118,127,172,127,176,127,173,127,237,127,235,127,234,127,236,127,230,127,232,127,100,128,103,128,163,129,159,129,158,129,149,129,162,129,153,129,151,129,22,130,79,130,83,130,82,130,80,130,78,130,81,130,36,133,59,133,15,133,0,133,41,133,14,133,9,133,13,133,31,133,10,133,39,133,28,133,251,132,43,133,250,132,8,133,12,133,244,132,42,133,242,132,21,133,247,132,235,132,243,132,252,132,18,133,234,132,233,132,22,133,254,132,40,133,29,133,46,133,2,133,253,132,30,133,246,132,49,133,38,133,231,132,232,132,240,132,239,132,249,132,24,133,32,133,48,133,11,133,25,133,47,133,98,134,86,135,99,135,100,135,119,135,225,135,115,135,88,135,84,135,91,135,82,135,97,135,90,135,81,135,94,135,109,135,106,135,80,135,78,135,95,135,93,135,111,135,108,135,122,135,110,135,92,135,101,135,79,135,123,135,117,135,98,135,103,135,105,135,90,136,5,137,12,137,20,137,11,137,23,137,24,137,25,137,6,137,22,137,17,137,14,137,9,137,162,137,164,137,163,137,237,137,240,137,236,137,207,138,198,138,184,138,211,138,209,138,212,138,213,138,187,138,215,138,190,138,192,138,197,138,216,138,195,138,186,138,189,138,217,138,62,140,77,140,143,140,229,140,223,140,217,140,232,140,218,140,221,140,231,140,160,141,156,141,161,141,155,141,32,142,35,142,37,142,36,142,46,142,21,142,27,142,22,142,17,142,25,142,38,142,39,142,20,142,18,142,24,142,19,142,28,142,23,142,26,142,44,143,36,143,24,143,26,143,32,143,35,143,22,143,23,143,115,144,112,144,111,144,103,144,107,144,47,145,43,145,41,145,42,145,50,145,38,145,46,145,133,145,134,145,138,145,129,145,130,145,132,145,128,145,208,146,195,146,196,146,192,146,217,146,182,146,207,146,241,146,223,146,216,146,233,146,215,146,221,146,204,146,239,146,194,146,232,146,202,146,200,146,206,146,230,146,205,146,213,146,201,146,224,146,222,146,231,146,209,146,211,146,181,146,225,146,198,146,180,146,124,149,172,149,171,149,174,149,176,149,164,150,162,150,211,150,5,151,8,151,2,151,90,151,138,151,142,151,136,151,208,151,207,151,30,152,29,152,38,152,41,152,40,152,32,152,27,152,39,152,178,152,8,153,250,152,17,153,20,153,22,153,23,153,21,153,220,153,205,153,207,153,211,153,212,153,206,153,201,153,214,153,216,153,203,153,215,153,204,153,179,154,236,154,235,154,243,154,242,154,241,154,70,155,67,155,103,155,116,155,113,155,102,155,118,155,117,155,112,155,104,155,100,155,108,155,252,156,250,156,253,156,255,156,247,156,7,157,0,157,249,156,251,156,8,157,5,157,4,157,131,158,211,158,15,159,16,159,28,81,19,81,23,81,26,81,17,81,222,81,52,83,225,83,112,86,96,86,110,86,115,86,102,86,99,86,109,86,114,86,94,86,119,86,28,87,27,87,200,88,189,88,201,88,191,88,186,88,194,88,188,88,198,88,23,91,25,91,27,91,33,91,20,91,19,91,16,91,22,91,40,91,26,91,32,91,30,91,239,91,172,93,177,93,169,93,167,93,181,93,176,93,174,93,170,93,168,93,178,93,173,93,175,93,180,93,103,94,104,94,102,94,111,94,233,94,231,94,230,94,232,94,229,94,75,95,188,95,157,97,168,97,150,97,197,97,180,97,198,97,193,97,204,97,186,97,191,97,184,97,140,97,215,100,214,100,208,100,207,100,201,100,189,100,137,100,195,100,219,100,243,100,217,100,51,101,127,101,124,101,162,101,200,102,190,102,192,102,202,102,203,102,207,102,189,102,187,102,186,102,204,102,35,103,52,106,102,106,73,106,103,106,50,106,104,106,62,106,93,106,109,106,118,106,91,106,81,106,40,106,90,106,59,106,63,106,65,106,106,106,100,106,80,106,79,106,84,106,111,106,105,106,96,106,60,106,94,106,86,106,85,106,77,106,78,106,70,106,85,107,84,107,86,107,167,107,170,107,171,107,200,107,199,107,4,108,3,108,6,108,173,111,203,111,163,111,199,111,188,111,206,111,200,111,94,111,196,111,189,111,158,111,202,111,168,111,4,112,165,111,174,111,186,111,172,111,170,111,207,111,191,111,184,111,162,111,201,111,171,111,205,111,175,111,178,111,176,111,197,113,194,113,191,113,184,113,214,113,192,113,193,113,203,113,212,113,202,113,199,113,207,113,189,113,216,113,188,113,198,113,218,113,219,113,157,114,158,114,105,115,102,115,103,115,108,115,101,115,107,115,106,115,127,116,154,116,160,116,148,116,146,116,149,116,161,116,11,117,128,117,47,118,45,118,49,118,61,118,51,118,60,118,53,118,50,118,48,118,187,118,230,118,154,119,157,119,161,119,156,119,155,119,162,119,163,119,149,119,153,119,151,119,221,120,233,120,229,120,234,120,222,120,227,120,219,120,225,120,226,120,237,120,223,120,224,120,164,121,68,122,72,122,71,122,182,122,184,122,181,122,177,122,183,122,222,123,227,123,231,123,221,123,213,123,229,123,218,123,232,123,249,123,212,123,234,123,226,123,220,123,235,123,216,123,223,123,210,124,212,124,215,124,208,124,209,124,18,126,33,126,23,126,12,126,31,126,32,126,19,126,14,126,28,126,21,126,26,126,34,126,11,126,15,126,22,126,13,126,20,126,37,126,36,126,67,127,123,127,124,127,122,127,177,127,239,127,42,128,41,128,108,128,177,129,166,129,174,129,185,129,181,129,171,129,176,129,172,129,180,129,178,129,183,129,167,129,242,129,85,130,86,130,87,130,86,133,69,133,107,133,77,133,83,133,97,133,88,133,64,133,70,133,100,133,65,133,98,133,68,133,81,133,71,133,99,133,62,133,91,133,113,133,78,133,110,133,117,133,85,133,103,133,96,133,140,133,102,133,93,133,84,133,101,133,108,133,99,134,101,134,100,134,155,135,143,135,151,135,147,135,146,135,136,135,129,135,150,135,152,135,121,135,135,135,163,135,133,135,144,135,145,135,157,135,132,135,148,135,156,135,154,135,137,135,30,137,38,137,48,137,45,137,46,137,39,137,49,137,34,137,41,137,35,137,47,137,44,137,31,137,241,137,224,138,226,138,242,138,244,138,245,138,221,138,20,139,228,138,223,138,240,138,200,138,222,138,225,138,232,138,255,138,239,138,251,138,145,140,146,140,144,140,245,140,238,140,241,140,240,140,243,140,108,141,110,141,165,141,167,141,51,142,62,142,56,142,64,142,69,142,54,142,60,142,61,142,65,142,48,142,63,142,189,142,54,143,46,143,53,143,50,143,57,143,55,143,52,143,118,144,121,144,123,144,134,144,250,144,51,145,53,145,54,145,147,145,144,145,145,145,141,145,143,145,39,147,30,147,8,147,31,147,6,147,15,147,122,147,56,147,60,147,27,147,35,147,18,147,1,147,70,147,45,147,14,147,13,147,203,146,29,147,250,146,37,147,19,147,249,146,247,146,52,147,2,147,36,147,255,146,41,147,57,147,53,147,42,147,20,147,12,147,11,147,254,146,9,147,0,147,251,146,22,147,188,149,205,149,190,149,185,149,186,149,182,149,191,149,181,149,189,149,169,150,212,150,11,151,18,151,16,151,153,151,151,151,148,151,240,151,248,151,53,152,47,152,50,152,36,153,31,153,39,153,41,153,158,153,238,153,236,153,229,153,228,153,240,153,227,153,234,153,233,153,231,153,185,154,191,154,180,154,187,154,246,154,250,154,249,154,247,154,51,155,128,155,133,155,135,155,124,155,126,155,123,155,130,155,147,155,146,155,144,155,122,155,149,155,125,155,136,155,37,157,23,157,32,157,30,157,20,157,41,157,29,157,24,157,34,157,16,157,25,157,31,157,136,158,134,158,135,158,174,158,173,158,213,158,214,158,250,158,18,159,61,159,38,81,37,81,34,81,36,81,32,81,41,81,244,82,147,86,140,86,141,86,134,86,132,86,131,86,126,86,130,86,127,86,129,86,214,88,212,88,207,88,210,88,45,91,37,91,50,91,35,91,44,91,39,91,38,91,47,91,46,91,123,91,241,91,242,91,183,93,108,94,106,94,190,95,187,95,195,97,181,97,188,97,231,97,224,97,229,97,228,97,232,97,222,97,239,100,233,100,227,100,235,100,228,100,232,100,129,101,128,101,182,101,218,101,210,102,141,106,150,106,129,106,165,106,137,106,159,106,155,106,161,106,158,106,135,106,147,106,142,106,149,106,131,106,168,106,164,106,145,106,127,106,166,106,154,106,133,106,140,106,146,106,91,107,173,107,9,108,204,111,169,111,244,111,212,111,227,111,220,111,237,111,231,111,230,111,222,111,242,111,221,111,226,111,232,111,225,113,241,113,232,113,242,113,228,113,240,113,226,113,115,115,110,115,111,115,151,116,178,116,171,116,144,116,170,116,173,116,177,116,165,116,175,116,16,117,17,117,18,117,15,117,132,117,67,118,72,118,73,118,71,118,164,118,233,118,181,119,171,119,178,119,183,119,182,119,180,119,177,119,168,119,240,119,243,120,253,120,2,121,251,120,252,120,242,120,5,121,249,120,254,120,4,121,171,121,168,121,92,122,91,122,86,122,88,122,84,122,90,122,190,122,192,122,193,122,5,124,15,124,242,123,0,124,255,123,251,123,14,124,244,123,11,124,243,123,2,124,9,124,3,124,1,124,248,123,253,123,6,124,240,123,241,123,16,124,10,124,232,124,45,126,60,126,66,126,51,126,72,152,56,126,42,126,73,126,64,126,71,126,41,126,76,126,48,126,59,126,54,126,68,126,58,126,69,127,127,127,126,127,125,127,244,127,242,127,44,128,187,129,196,129,204,129,202,129,197,129,199,129,188,129,233,129,91,130,90,130,92,130,131,133,128,133,143,133,167,133,149,133,160,133,139,133,163,133,123,133,164,133,154,133,158,133,119,133,124,133,137,133,161,133,122,133,120,133,87,133,142,133,150,133,134,133,141,133,153,133,157,133,129,133,162,133,130,133,136,133,133,133,121,133,118,133,152,133,144,133,159,133,104,134,190,135,170,135,173,135,197,135,176,135,172,135,185,135,181,135,188,135,174,135,201,135,195,135,194,135,204,135,183,135,175,135,196,135,202,135,180,135,182,135,191,135,184,135,189,135,222,135,178,135,53,137,51,137,60,137,62,137,65,137,82,137,55,137,66,137,173,137,175,137,174,137,242,137,243,137,30,139,24,139,22,139,17,139,5,139,11,139,34,139,15,139,18,139,21,139,7,139,13,139,8,139,6,139,28,139,19,139,26,139,79,140,112,140,114,140,113,140,111,140,149,140,148,140,249,140,111,141,78,142,77,142,83,142,80,142,76,142,71,142,67,143,64,143,133,144,126,144,56,145,154,145,162,145,155,145,153,145,159,145,161,145,157,145,160,145,161,147,131,147,175,147,100,147,86,147,71,147,124,147,88,147,92,147,118,147,73,147,80,147,81,147,96,147,109,147,143,147,76,147,106,147,121,147,87,147,85,147,82,147,79,147,113,147,119,147,123,147,97,147,94,147,99,147,103,147,128,147,78,147,89,147,199,149,192,149,201,149,195,149,197,149,183,149,174,150,176,150,172,150,32,151,31,151,24,151,29,151,25,151,154,151,161,151,156,151,158,151,157,151,213,151,212,151,241,151,65,152,68,152,74,152,73,152,69,152,67,152,37,153,43,153,44,153,42,153,51,153,50,153,47,153,45,153,49,153,48,153,152,153,163,153,161,153,2,154,250,153,244,153,247,153,249,153,248,153,246,153,251,153,253,153,254,153,252,153,3,154,190,154,254,154,253,154,1,155,252,154,72,155,154,155,168,155,158,155,155,155,166,155,161,155,165,155,164,155,134,155,162,155,160,155,175,155,51,157,65,157,103,157,54,157,46,157,47,157,49,157,56,157,48,157,69,157,66,157,67,157,62,157,55,157,64,157,61,157,245,127,45,157,138,158,137,158,141,158,176,158,200,158,218,158,251,158,255,158,36,159,35,159,34,159,84,159,160,159,49,81,45,81,46,81,152,86,156,86,151,86,154,86,157,86,153,86,112,89,60,91,105,92,106,92,192,93,109,94,110,94,216,97,223,97,237,97,238,97,241,97,234,97,240,97,235,97,214,97,233,97,255,100,4,101,253,100,248,100,1,101,3,101,252,100,148,101,219,101,218,102,219,102,216,102,197,106,185,106,189,106,225,106,198,106,186,106,182,106,183,106,199,106,180,106,173,106,94,107,201,107,11,108,7,112,12,112,13,112,1,112,5,112,20,112,14,112,255,111,0,112,251,111,38,112,252,111,247,111,10,112,1,114,255,113,249,113,3,114,253,113,118,115,184,116,192,116,181,116,193,116,190,116,182,116,187,116,194,116,20,117,19,117,92,118,100,118,89,118,80,118,83,118,87,118,90,118,166,118,189,118,236,118,194,119,186,119,255,120,12,121,19,121,20,121,9,121,16,121,18,121,17,121,173,121,172,121,95,122,28,124,41,124,25,124,32,124,31,124,45,124,29,124,38,124,40,124,34,124,37,124,48,124,92,126,80,126,86,126,99,126,88,126,98,126,95,126,81,126,96,126,87,126,83,126,181,127,179,127,247,127,248,127,117,128,209,129,210,129,208,129,95,130,94,130,180,133,198,133,192,133,195,133,194,133,179,133,181,133,189,133,199,133,196,133,191,133,203,133,206,133,200,133,197,133,177,133,182,133,210,133,36,134,184,133,183,133,190,133,105,134,231,135,230,135,226,135,219,135,235,135,234,135,229,135,223,135,243,135,228,135,212,135,220,135,211,135,237,135,216,135,227,135,164,135,215,135,217,135,1,136,244,135,232,135,221,135,83,137,75,137,79,137,76,137,70,137,80,137,81,137,73,137,42,139,39,139,35,139,51,139,48,139,53,139,71,139,47,139,60,139,62,139,49,139,37,139,55,139,38,139,54,139,46,139,36,139,59,139,61,139,58,139,66,140,117,140,153,140,152,140,151,140,254,140,4,141,2,141,0,141,92,142,98,142,96,142,87,142,86,142,94,142,101,142,103,142,91,142,90,142,97,142,93,142,105,142,84,142,70,143,71,143,72,143,75,143,40,145,58,145,59,145,62,145,168,145,165,145,167,145,175,145,170,145,181,147,140,147,146,147,183,147,155,147,157,147,137,147,167,147,142,147,170,147,158,147,166,147,149,147,136,147,153,147,159,147,141,147,177,147,145,147,178,147,164,147,168,147,180,147,163,147,165,147,210,149,211,149,209,149,179,150,215,150,218,150,194,93,223,150,216,150,221,150,35,151,34,151,37,151,172,151,174,151,168,151,171,151,164,151,170,151,162,151,165,151,215,151,217,151,214,151,216,151,250,151,80,152,81,152,82,152,184,152,65,153,60,153,58,153,15,154,11,154,9,154,13,154,4,154,17,154,10,154,5,154,7,154,6,154,192,154,220,154,8,155,4,155,5,155,41,155,53,155,74,155,76,155,75,155,199,155,198,155,195,155,191,155,193,155,181,155,184,155,211,155,182,155,196,155,185,155,189,155,92,157,83,157,79,157,74,157,91,157,75,157,89,157,86,157,76,157,87,157,82,157,84,157,95,157,88,157,90,157,142,158,140,158,223,158,1,159,0,159,22,159,37,159,43,159,42,159,41,159,40,159,76,159,85,159,52,81,53,81,150,82,247,82,180,83,171,86,173,86,166,86,167,86,170,86,172,86,218,88,221,88,219,88,18,89,61,91,62,91,63,91,195,93,112,94,191,95,251,97,7,101,16,101,13,101,9,101,12,101,14,101,132,101,222,101,221,101,222,102,231,106,224,106,204,106,209,106,217,106,203,106,223,106,220,106,208,106,235,106,207,106,205,106,222,106,96,107,176,107,12,108,25,112,39,112,32,112,22,112,43,112,33,112,34,112,35,112,41,112,23,112,36,112,28,112,42,112,12,114,10,114,7,114,2,114,5,114,165,114,166,114,164,114,163,114,161,114,203,116,197,116,183,116,195,116,22,117,96,118,201,119,202,119,196,119,241,119,29,121,27,121,33,121,28,121,23,121,30,121,176,121,103,122,104,122,51,124,60,124,57,124,44,124,59,124,236,124,234,124,118,126,117,126,120,126,112,126,119,126,111,126,122,126,114,126,116,126,104,126,75,127,74,127,131,127,134,127,183,127,253,127,254,127,120,128,215,129,213,129,100,130,97,130,99,130,235,133,241,133,237,133,217,133,225,133,232,133,218,133,215,133,236,133,242,133,248,133,216,133,223,133,227,133,220,133,209,133,240,133,230,133,239,133,222,133,226,133,0,136,250,135,3,136,246,135,247,135,9,136,12,136,11,136,6,136,252,135,8,136,255,135,10,136,2,136,98,137,90,137,91,137,87,137,97,137,92,137,88,137,93,137,89,137,136,137,183,137,182,137,246,137,80,139,72,139,74,139,64,139,83,139,86,139,84,139,75,139,85,139,81,139,66,139,82,139,87,139,67,140,119,140,118,140,154,140,6,141,7,141,9,141,172,141,170,141,173,141,171,141,109,142,120,142,115,142,106,142,111,142,123,142,194,142,82,143,81,143,79,143,80,143,83,143,180,143,64,145,63,145,176,145,173,145,222,147,199,147,207,147,194,147,218,147,208,147,249,147,236,147,204,147,217,147,169,147,230,147,202,147,212,147,238,147,227,147,213,147,196,147,206,147,192,147,210,147,231,147,125,149,218,149,219,149,225,150,41,151,43,151,44,151,40,151,38,151,179,151,183,151,182,151,221,151,222,151,223,151,92,152,89,152,93,152,87,152,191,152,189,152,187,152,190,152,72,153,71,153,67,153,166,153,167,153,26,154,21,154,37,154,29,154,36,154,27,154,34,154,32,154,39,154,35,154,30,154,28,154,20,154,194,154,11,155,10,155,14,155,12,155,55,155,234,155,235,155,224,155,222,155,228,155,230,155,226,155,240,155,212,155,215,155,236,155,220,155,217,155,229,155,213,155,225,155,218,155,119,157,129,157,138,157,132,157,136,157,113,157,128,157,120,157,134,157,139,157,140,157,125,157,107,157,116,157,117,157,112,157,105,157,133,157,115,157,123,157,130,157,111,157,121,157,127,157,135,157,104,157,148,158,145,158,192,158,252,158,45,159,64,159,65,159,77,159,86,159,87,159,88,159,55,83,178,86,181,86,179,86,227,88,69,91,198,93,199,93,238,94,239,94,192,95,193,95,249,97,23,101,22,101,21,101,19,101,223,101,232,102,227,102,228,102,243,106,240,106,234,106,232,106,249,106,241,106,238,106,239,106,60,112,53,112,47,112,55,112,52,112,49,112,66,112,56,112,63,112,58,112,57,112,64,112,59,112,51,112,65,112,19,114,20,114,168,114,125,115,124,115,186,116,171,118,170,118,190,118,237,118,204,119,206,119,207,119,205,119,242,119,37,121,35,121,39,121,40,121,36,121,41,121,178,121,110,122,108,122,109,122,247,122,73,124,72,124,74,124,71,124,69,124,238,124,123,126,126,126,129,126,128,126,186,127,255,127,121,128,219,129,217,129,11,130,104,130,105,130,34,134,255,133,1,134,254,133,27,134,0,134,246,133,4,134,9,134,5,134,12,134,253,133,25,136,16,136,17,136,23,136,19,136,22,136,99,137,102,137,185,137,247,137,96,139,106,139,93,139,104,139,99,139,101,139,103,139,109,139,174,141,134,142,136,142,132,142,89,143,86,143,87,143,85,143,88,143,90,143,141,144,67,145,65,145,183,145,181,145,178,145,179,145,11,148,19,148,251,147,32,148,15,148,20,148,254,147,21,148,16,148,40,148,25,148,13,148,245,147,0,148,247,147,7,148,14,148,22,148,18,148,250,147,9,148,248,147,10,148,255,147,252,147,12,148,246,147,17,148,6,148,222,149,224,149,223,149,46,151,47,151,185,151,187,151,253,151,254,151,96,152,98,152,99,152,95,152,193,152,194,152,80,153,78,153,89,153,76,153,75,153,83,153,50,154,52,154,49,154,44,154,42,154,54,154,41,154,46,154,56,154,45,154,199,154,202,154,198,154,16,155,18,155,17,155,11,156,8,156,247,155,5,156,18,156,248,155,64,156,7,156,14,156,6,156,23,156,20,156,9,156,159,157,153,157,164,157,157,157,146,157,152,157,144,157,155,157,160,157,148,157,156,157,170,157,151,157,161,157,154,157,162,157,168,157,158,157,163,157,191,157,169,157,150,157,166,157,167,157,153,158,155,158,154,158,229,158,228,158,231,158,230,158,48,159,46,159,91,159,96,159,94,159,93,159,89,159,145,159,58,81,57,81,152,82,151,82,195,86,189,86,190,86,72,91,71,91,203,93,207,93,241,94,253,97,27,101,2,107,252,106,3,107,248,106,0,107,67,112,68,112,74,112,72,112,73,112,69,112,70,112,29,114,26,114,25,114,126,115,23,117,106,118,208,119,45,121,49,121,47,121,84,124,83,124,242,124,138,126,135,126,136,126,139,126,134,126,141,126,77,127,187,127,48,128,221,129,24,134,42,134,38,134,31,134,35,134,28,134,25,134,39,134,46,134,33,134,32,134,41,134,30,134,37,134,41,136,29,136,27,136,32,136,36,136,28,136,43,136,74,136,109,137,105,137,110,137,107,137,250,137,121,139,120,139,69,139,122,139,123,139,16,141,20,141,175,141,142,142,140,142,94,143,91,143,93,143,70,145,68,145,69,145,185,145,63,148,59,148,54,148,41,148,61,148,60,148,48,148,57,148,42,148,55,148,44,148,64,148,49,148,229,149,228,149,227,149,53,151,58,151,191,151,225,151,100,152,201,152,198,152,192,152,88,153,86,153,57,154,61,154,70,154,68,154,66,154,65,154,58,154,63,154,205,154,21,155,23,155,24,155,22,155,58,155,82,155,43,156,29,156,28,156,44,156,35,156,40,156,41,156,36,156,33,156,183,157,182,157,188,157,193,157,199,157,202,157,207,157,190,157,197,157,195,157,187,157,181,157,206,157,185,157,186,157,172,157,200,157,177,157,173,157,204,157,179,157,205,157,178,157,122,158,156,158,235,158,238,158,237,158,27,159,24,159,26,159,49,159,78,159,101,159,100,159,146,159,185,78,198,86,197,86,203,86,113,89,75,91,76,91,213,93,209,93,242,94,33,101,32,101,38,101,34,101,11,107,8,107,9,107,13,108,85,112,86,112,87,112,82,112,30,114,31,114,169,114,127,115,216,116,213,116,217,116,215,116,109,118,173,118,53,121,180,121,112,122,113,122,87,124,92,124,89,124,91,124,90,124,244,124,241,124,145,126,79,127,135,127,222,129,107,130,52,134,53,134,51,134,44,134,50,134,54,134,44,136,40,136,38,136,42,136,37,136,113,137,191,137,190,137,251,137,126,139,132,139,130,139,134,139,133,139,127,139,21,141,149,142,148,142,154,142,146,142,144,142,150,142,151,142,96,143,98,143,71,145,76,148,80,148,74,148,75,148,79,148,71,148,69,148,72,148,73,148,70,148,63,151,227,151,106,152,105,152,203,152,84,153,91,153,78,154,83,154,84,154,76,154,79,154,72,154,74,154,73,154,82,154,80,154,208,154,25,155,43,155,59,155,86,155,85,155,70,156,72,156,63,156,68,156,57,156,51,156,65,156,60,156,55,156,52,156,50,156,61,156,54,156,219,157,210,157,222,157,218,157,203,157,208,157,220,157,209,157,223,157,233,157,217,157,216,157,214,157,245,157,213,157,221,157,182,158,240,158,53,159,51,159,50,159,66,159,107,159,149,159,162,159,61,81,153,82,232,88,231,88,114,89,77,91,216,93,47,136,79,95,1,98,3,98,4,98,41,101,37,101,150,101,235,102,17,107,18,107,15,107,202,107,91,112,90,112,34,114,130,115,129,115,131,115,112,118,212,119,103,124,102,124,149,126,108,130,58,134,64,134,57,134,60,134,49,134,59,134,62,134,48,136,50,136,46,136,51,136,118,137,116,137,115,137,254,137,140,139,142,139,139,139,136,139,69,140,25,141,152,142,100,143,99,143,188,145,98,148,85,148,93,148,87,148,94,148,196,151,197,151,0,152,86,154,89,154,30,155,31,155,32,155,82,156,88,156,80,156,74,156,77,156,75,156,85,156,89,156,76,156,78,156,251,157,247,157,239,157,227,157,235,157,248,157,228,157,246,157,225,157,238,157,230,157,242,157,240,157,226,157,236,157,244,157,243,157,232,157,237,157,194,158,208,158,242,158,243,158,6,159,28,159,56,159,55,159,54,159,67,159,79,159,113,159,112,159,110,159,111,159,211,86,205,86,78,91,109,92,45,101,237,102,238,102,19,107,95,112,97,112,93,112,96,112,35,114,219,116,229,116,213,119,56,121,183,121,182,121,106,124,151,126,137,127,109,130,67,134,56,136,55,136,53,136,75,136,148,139,149,139,158,142,159,142,160,142,157,142,190,145,189,145,194,145,107,148,104,148,105,148,229,150,70,151,67,151,71,151,199,151,229,151,94,154,213,154,89,155,99,156,103,156,102,156,98,156,94,156,96,156,2,158,254,157,7,158,3,158,6,158,5,158,0,158,1,158,9,158,255,157,253,157,4,158,160,158,30,159,70,159,116,159,117,159,118,159,212,86,46,101,184,101,24,107,25,107,23,107,26,107,98,112,38,114,170,114,216,119,217,119,57,121,105,124,107,124,246,124,154,126,152,126,155,126,153,126,224,129,225,129,70,134,71,134,72,134,121,137,122,137,124,137,123,137,255,137,152,139,153,139,165,142,164,142,163,142,110,148,109,148,111,148,113,148,115,148,73,151,114,152,95,153,104,156,110,156,109,156,11,158,13,158,16,158,15,158,18,158,17,158,161,158,245,158,9,159,71,159,120,159,123,159,122,159,121,159,30,87,102,112,111,124,60,136,178,141,166,142,195,145,116,148,120,148,118,148,117,148,96,154,116,156,115,156,113,156,117,156,20,158,19,158,246,158,10,159,164,159,104,112,101,112,247,124,106,134,62,136,61,136,63,136,158,139,156,140,169,142,201,142,75,151,115,152,116,152,204,152,97,153,171,153,100,154,102,154,103,154,36,155,21,158,23,158,72,159,7,98,30,107,39,114,76,134,168,142,130,148,128,148,129,148,105,154,104,154,46,155,25,158,41,114,75,134,159,139,131,148,121,156,183,158,117,118,107,154,122,156,29,158,105,112,106,112,164,158,126,159,73,159,152,159,129,120,185,146,207,136,187,88,82,96,167,124,250,90,84,37,102,37,87,37,96,37,108,37,99,37,90,37,105,37,93,37,82,37,100,37,85,37,94,37,106,37,97,37,88,37,103,37,91,37,83,37,101,37,86,37,95,37,107,37,98,37,89,37,104,37,92,37,81,37,80,37,109,37,110,37,112,37,111,37,147,37,0,48,1,48,2,48,183,0,37,32,38,32,168,0,3,48,173,0,21,32,37,34,60,255,60,34,24,32,25,32,28,32,29,32,20,48,21,48,8,48,9,48,10,48,11,48,12,48,13,48,14,48,15,48,16,48,17,48,177,0,215,0,247,0,96,34,100,34,101,34,30,34,52,34,176,0,50,32,51,32,3,33,43,33,224,255,225,255,229,255,66,38,64,38,32,34,165,34,18,35,2,34,7,34,97,34,82,34,167,0,59,32,6,38,5,38,203,37,207,37,206,37,199,37,198,37,161,37,160,37,179,37,178,37,189,37,188,37,146,33,144,33,145,33,147,33,148,33,19,48,106,34,107,34,26,34,61,34,29,34,53,34,43,34,44,34,8,34,11,34,134,34,135,34,130,34,131,34,42,34,41,34,39,34,40,34,226,255,210,33,212,33,0,34,3,34,180,0,94,255,199,2,216,2,221,2,218,2,217,2,184,0,219,2,161,0,191,0,208,2,46,34,17,34,15,34,164,0,9,33,48,32,193,37,192,37,183,37,182,37,100,38,96,38,97,38,101,38,103,38,99,38,153,34,200,37,163,37,208,37,209,37,146,37,164,37,165,37,168,37,167,37,166,37,169,37,104,38,15,38,14,38,28,38,30,38,182,0,32,32,33,32,149,33,151,33,153,33,150,33,152,33,109,38,105,38,106,38,108,38,127,50,28,50,22,33,199,51,34,33,194,51,216,51,33,33,172,32,174,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,255,2,255,3,255,4,255,5,255,6,255,7,255,8,255,9,255,10,255,11,255,12,255,13,255,14,255,15,255,16,255,17,255,18,255,19,255,20,255,21,255,22,255,23,255,24,255,25,255,26,255,27,255,28,255,29,255,30,255,31,255,32,255,33,255,34,255,35,255,36,255,37,255,38,255,39,255,40,255,41,255,42,255,43,255,44,255,45,255,46,255,47,255,48,255,49,255,50,255,51,255,52,255,53,255,54,255,55,255,56,255,57,255,58,255,59,255,230,255,61,255,62,255,63,255,64,255,65,255,66,255,67,255,68,255,69,255,70,255,71,255,72,255,73,255,74,255,75,255,76,255,77,255,78,255,79,255,80,255,81,255,82,255,83,255,84,255,85,255,86,255,87,255,88,255,89,255,90,255,91,255,92,255,93,255,227,255,49,49,50,49,51,49,52,49,53,49,54,49,55,49,56,49,57,49,58,49,59,49,60,49,61,49,62,49,63,49,64,49,65,49,66,49,67,49,68,49,69,49,70,49,71,49,72,49,73,49,74,49,75,49,76,49,77,49,78,49,79,49,80,49,81,49,82,49,83,49,84,49,85,49,86,49,87,49,88,49,89,49,90,49,91,49,92,49,93,49,94,49,95,49,96,49,97,49,98,49,99,49,100,49,101,49,102,49,103,49,104,49,105,49,106,49,107,49,108,49,109,49,110,49,111,49,112,49,113,49,114,49,115,49,116,49,117,49,118,49,119,49,120,49,121,49,122,49,123,49,124,49,125,49,126,49,127,49,128,49,129,49,130,49,131,49,132,49,133,49,134,49,135,49,136,49,137,49,138,49,139,49,140,49,141,49,142,49,112,33,113,33,114,33,115,33,116,33,117,33,118,33,119,33,120,33,121,33,0,0,0,0,0,0,0,0,0,0,96,33,97,33,98,33,99,33,100,33,101,33,102,33,103,33,104,33,105,33,0,0,0,0,0,0,0,0,0,0,0,0,0,0,145,3,146,3,147,3,148,3,149,3,150,3,151,3,152,3,153,3,154,3,155,3,156,3,157,3,158,3,159,3,160,3,161,3,163,3,164,3,165,3,166,3,167,3,168,3,169,3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,177,3,178,3,179,3,180,3,181,3,182,3,183,3,184,3,185,3,186,3,187,3,188,3,189,3,190,3,191,3,192,3,193,3,195,3,196,3,197,3,198,3,199,3,200,3,201,3,0,0,0,0,0,0,0,0,0,0,0,0,0,37,2,37,12,37,16,37,24,37,20,37,28,37,44,37,36,37,52,37,60,37,1,37,3,37,15,37,19,37,27,37,23,37,35,37,51,37,43,37,59,37,75,37,32,37,47,37,40,37,55,37,63,37,29,37,48,37,37,37,56,37,66,37,18,37,17,37,26,37,25,37,22,37,21,37,14,37,13,37,30,37,31,37,33,37,34,37,38,37,39,37,41,37,42,37,45,37,46,37,49,37,50,37,53,37,54,37,57,37,58,37,61,37,62,37,64,37,65,37,67,37,68,37,69,37,70,37,71,37,72,37,73,37,74,37,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,149,51,150,51,151,51,19,33,152,51,196,51,163,51,164,51,165,51,166,51,153,51,154,51,155,51,156,51,157,51,158,51,159,51,160,51,161,51,162,51,202,51,141,51,142,51,143,51,207,51,136,51,137,51,200,51,167,51,168,51,176,51,177,51,178,51,179,51,180,51,181,51,182,51,183,51,184,51,185,51,128,51,129,51,130,51,131,51,132,51,186,51,187,51,188,51,189,51,190,51,191,51,144,51,145,51,146,51,147,51,148,51,38,33,192,51,193,51,138,51,139,51,140,51,214,51,197,51,173,51,174,51,175,51,219,51,169,51,170,51,171,51,172,51,221,51,208,51,211,51,195,51,201,51,220,51,198,51,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,198,0,208,0,170,0,38,1,0,0,50,1,0,0,63,1,65,1,216,0,82,1,186,0,222,0,102,1,74,1,0,0,96,50,97,50,98,50,99,50,100,50,101,50,102,50,103,50,104,50,105,50,106,50,107,50,108,50,109,50,110,50,111,50,112,50,113,50,114,50,115,50,116,50,117,50,118,50,119,50,120,50,121,50,122,50,123,50,208,36,209,36,210,36,211,36,212,36,213,36,214,36,215,36,216,36,217,36,218,36,219,36,220,36,221,36,222,36,223,36,224,36,225,36,226,36,227,36,228,36,229,36,230,36,231,36,232,36,233,36,96,36,97,36,98,36,99,36,100,36,101,36,102,36,103,36,104,36,105,36,106,36,107,36,108,36,109,36,110,36,189,0,83,33,84,33,188,0,190,0,91,33,92,33,93,33,94,33,230,0,17,1,240,0,39,1,49,1,51,1,56,1,64,1,66,1,248,0,83,1,223,0,254,0,103,1,75,1,73,1,0,50,1,50,2,50,3,50,4,50,5,50,6,50,7,50,8,50,9,50,10,50,11,50,12,50,13,50,14,50,15,50,16,50,17,50,18,50,19,50,20,50,21,50,22,50,23,50,24,50,25,50,26,50,27,50,156,36,157,36,158,36,159,36,160,36,161,36,162,36,163,36,164,36,165,36,166,36,167,36,168,36,169,36,170,36,171,36,172,36,173,36,174,36,175,36,176,36,177,36,178,36,179,36,180,36,181,36,116,36,117,36,118,36,119,36,120,36,121,36,122,36,123,36,124,36,125,36,126,36,127,36,128,36,129,36,130,36,185,0,178,0,179,0,116,32,127,32,129,32,130,32,131,32,132,32,65,48,66,48,67,48,68,48,69,48,70,48,71,48,72,48,73,48,74,48,75,48,76,48,77,48,78,48,79,48,80,48,81,48,82,48,83,48,84,48,85,48,86,48,87,48,88,48,89,48,90,48,91,48,92,48,93,48,94,48,95,48,96,48,97,48,98,48,99,48,100,48,101,48,102,48,103,48,104,48,105,48,106,48,107,48,108,48,109,48,110,48,111,48,112,48,113,48,114,48,115,48,116,48,117,48,118,48,119,48,120,48,121,48,122,48,123,48,124,48,125,48,126,48,127,48,128,48,129,48,130,48,131,48,132,48,133,48,134,48,135,48,136,48,137,48,138,48,139,48,140,48,141,48,142,48,143,48,144,48,145,48,146,48,147,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161,48,162,48,163,48,164,48,165,48,166,48,167,48,168,48,169,48,170,48,171,48,172,48,173,48,174,48,175,48,176,48,177,48,178,48,179,48,180,48,181,48,182,48,183,48,184,48,185,48,186,48,187,48,188,48,189,48,190,48,191,48,192,48,193,48,194,48,195,48,196,48,197,48,198,48,199,48,200,48,201,48,202,48,203,48,204,48,205,48,206,48,207,48,208,48,209,48,210,48,211,48,212,48,213,48,214,48,215,48,216,48,217,48,218,48,219,48,220,48,221,48,222,48,223,48,224,48,225,48,226,48,227,48,228,48,229,48,230,48,231,48,232,48,233,48,234,48,235,48,236,48,237,48,238,48,239,48,240,48,241,48,242,48,243,48,244,48,245,48,246,48,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,4,17,4,18,4,19,4,20,4,21,4,1,4,22,4,23,4,24,4,25,4,26,4,27,4,28,4,29,4,30,4,31,4,32,4,33,4,34,4,35,4,36,4,37,4,38,4,39,4,40,4,41,4,42,4,43,4,44,4,45,4,46,4,47,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,48,4,49,4,50,4,51,4,52,4,53,4,81,4,54,4,55,4,56,4,57,4,58,4,59,4,60,4,61,4,62,4,63,4,64,4,65,4,66,4,67,4,68,4,69,4,70,4,71,4,72,4,73,4,74,4,75,4,76,4,77,4,78,4,79,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,172,1,172,4,172,7,172,8,172,9,172,10,172,16,172,17,172,18,172,19,172,20,172,21,172,22,172,23,172,25,172,26,172,27,172,28,172,29,172,32,172,36,172,44,172,45,172,47,172,48,172,49,172,56,172,57,172,60,172,64,172,75,172,77,172,84,172,88,172,92,172,112,172,113,172,116,172,119,172,120,172,122,172,128,172,129,172,131,172,132,172,133,172,134,172,137,172,138,172,139,172,140,172,144,172,148,172,156,172,157,172,159,172,160,172,161,172,168,172,169,172,170,172,172,172,175,172,176,172,184,172,185,172,187,172,188,172,189,172,193,172,196,172,200,172,204,172,213,172,215,172,224,172,225,172,228,172,231,172,232,172,234,172,236,172,239,172,240,172,241,172,243,172,245,172,246,172,252,172,253,172,0,173,4,173,6,173,12,173,13,173,15,173,17,173,24,173,28,173,32,173,41,173,44,173,45,173,52,173,53,173,56,173,60,173,68,173,69,173,71,173,73,173,80,173,84,173,88,173,97,173,99,173,108,173,109,173,112,173,115,173,116,173,117,173,118,173,123,173,124,173,125,173,127,173,129,173,130,173,136,173,137,173,140,173,144,173,156,173,157,173,164,173,183,173,192,173,193,173,196,173,200,173,208,173,209,173,211,173,220,173,224,173,228,173,248,173,249,173,252,173,255,173,0,174,1,174,8,174,9,174,11,174,13,174,20,174,48,174,49,174,52,174,55,174,56,174,58,174,64,174,65,174,67,174,69,174,70,174,74,174,76,174,77,174,78,174,80,174,84,174,86,174,92,174,93,174,95,174,96,174,97,174,101,174,104,174,105,174,108,174,112,174,120,174,121,174,123,174,124,174,125,174,132,174,133,174,140,174,188,174,189,174,190,174,192,174,196,174,204,174,205,174,207,174,208,174,209,174,216,174,217,174,220,174,232,174,235,174,237,174,244,174,248,174,252,174,7,175,8,175,13,175,16,175,44,175,45,175,48,175,50,175,52,175,60,175,61,175,63,175,65,175,66,175,67,175,72,175,73,175,80,175,92,175,93,175,100,175,101,175,121,175,128,175,132,175], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+118048);
/* memory initializer */ allocate([136,175,144,175,145,175,149,175,156,175,184,175,185,175,188,175,192,175,199,175,200,175,201,175,203,175,205,175,206,175,212,175,220,175,232,175,233,175,240,175,241,175,244,175,248,175,0,176,1,176,4,176,12,176,16,176,20,176,28,176,29,176,40,176,68,176,69,176,72,176,74,176,76,176,78,176,83,176,84,176,85,176,87,176,89,176,93,176,124,176,125,176,128,176,132,176,140,176,141,176,143,176,145,176,152,176,153,176,154,176,156,176,159,176,160,176,161,176,162,176,168,176,169,176,171,176,172,176,173,176,174,176,175,176,177,176,179,176,180,176,181,176,184,176,188,176,196,176,197,176,199,176,200,176,201,176,208,176,209,176,212,176,216,176,224,176,229,176,8,177,9,177,11,177,12,177,16,177,18,177,19,177,24,177,25,177,27,177,28,177,29,177,35,177,36,177,37,177,40,177,44,177,52,177,53,177,55,177,56,177,57,177,64,177,65,177,68,177,72,177,80,177,81,177,84,177,85,177,88,177,92,177,96,177,120,177,121,177,124,177,128,177,130,177,136,177,137,177,139,177,141,177,146,177,147,177,148,177,152,177,156,177,168,177,204,177,208,177,212,177,220,177,221,177,223,177,232,177,233,177,236,177,240,177,249,177,251,177,253,177,4,178,5,178,8,178,11,178,12,178,20,178,21,178,23,178,25,178,32,178,52,178,60,178,88,178,92,178,96,178,104,178,105,178,116,178,117,178,124,178,132,178,133,178,137,178,144,178,145,178,148,178,152,178,153,178,154,178,160,178,161,178,163,178,165,178,166,178,170,178,172,178,176,178,180,178,200,178,201,178,204,178,208,178,210,178,216,178,217,178,219,178,221,178,226,178,228,178,229,178,230,178,232,178,235,178,236,178,237,178,238,178,239,178,243,178,244,178,245,178,247,178,248,178,249,178,250,178,251,178,255,178,0,179,1,179,4,179,8,179,16,179,17,179,19,179,20,179,21,179,28,179,84,179,85,179,86,179,88,179,91,179,92,179,94,179,95,179,100,179,101,179,103,179,105,179,107,179,110,179,112,179,113,179,116,179,120,179,128,179,129,179,131,179,132,179,133,179,140,179,144,179,148,179,160,179,161,179,168,179,172,179,196,179,197,179,200,179,203,179,204,179,206,179,208,179,212,179,213,179,215,179,217,179,219,179,221,179,224,179,228,179,232,179,252,179,16,180,24,180,28,180,32,180,40,180,41,180,43,180,52,180,80,180,81,180,84,180,88,180,96,180,97,180,99,180,101,180,108,180,128,180,136,180,157,180,164,180,168,180,172,180,181,180,183,180,185,180,192,180,196,180,200,180,208,180,213,180,220,180,221,180,224,180,227,180,228,180,230,180,236,180,237,180,239,180,241,180,248,180,20,181,21,181,24,181,27,181,28,181,36,181,37,181,39,181,40,181,41,181,42,181,48,181,49,181,52,181,56,181,64,181,65,181,67,181,68,181,69,181,75,181,76,181,77,181,80,181,84,181,92,181,93,181,95,181,96,181,97,181,160,181,161,181,164,181,168,181,170,181,171,181,176,181,177,181,179,181,180,181,181,181,187,181,188,181,189,181,192,181,196,181,204,181,205,181,207,181,208,181,209,181,216,181,236,181,16,182,17,182,20,182,24,182,37,182,44,182,52,182,72,182,100,182,104,182,156,182,157,182,160,182,164,182,171,182,172,182,177,182,212,182,240,182,244,182,248,182,0,183,1,183,5,183,40,183,41,183,44,183,47,183,48,183,56,183,57,183,59,183,68,183,72,183,76,183,84,183,85,183,96,183,100,183,104,183,112,183,113,183,115,183,117,183,124,183,125,183,128,183,132,183,140,183,141,183,143,183,144,183,145,183,146,183,150,183,151,183,152,183,153,183,156,183,160,183,168,183,169,183,171,183,172,183,173,183,180,183,181,183,184,183,199,183,201,183,236,183,237,183,240,183,244,183,252,183,253,183,255,183,0,184,1,184,7,184,8,184,9,184,12,184,16,184,24,184,25,184,27,184,29,184,36,184,37,184,40,184,44,184,52,184,53,184,55,184,56,184,57,184,64,184,68,184,81,184,83,184,92,184,93,184,96,184,100,184,108,184,109,184,111,184,113,184,120,184,124,184,141,184,168,184,176,184,180,184,184,184,192,184,193,184,195,184,197,184,204,184,208,184,212,184,221,184,223,184,225,184,232,184,233,184,236,184,240,184,248,184,249,184,251,184,253,184,4,185,24,185,32,185,60,185,61,185,64,185,68,185,76,185,79,185,81,185,88,185,89,185,92,185,96,185,104,185,105,185,107,185,109,185,116,185,117,185,120,185,124,185,132,185,133,185,135,185,137,185,138,185,141,185,142,185,172,185,173,185,176,185,180,185,188,185,189,185,191,185,193,185,200,185,201,185,204,185,206,185,207,185,208,185,209,185,210,185,216,185,217,185,219,185,221,185,222,185,225,185,227,185,228,185,229,185,232,185,236,185,244,185,245,185,247,185,248,185,249,185,250,185,0,186,1,186,8,186,21,186,56,186,57,186,60,186,64,186,66,186,72,186,73,186,75,186,77,186,78,186,83,186,84,186,85,186,88,186,92,186,100,186,101,186,103,186,104,186,105,186,112,186,113,186,116,186,120,186,131,186,132,186,133,186,135,186,140,186,168,186,169,186,171,186,172,186,176,186,178,186,184,186,185,186,187,186,189,186,196,186,200,186,216,186,217,186,252,186,0,187,4,187,13,187,15,187,17,187,24,187,28,187,32,187,41,187,43,187,52,187,53,187,54,187,56,187,59,187,60,187,61,187,62,187,68,187,69,187,71,187,73,187,77,187,79,187,80,187,84,187,88,187,97,187,99,187,108,187,136,187,140,187,144,187,164,187,168,187,172,187,180,187,183,187,192,187,196,187,200,187,208,187,211,187,248,187,249,187,252,187,255,187,0,188,2,188,8,188,9,188,11,188,12,188,13,188,15,188,17,188,20,188,21,188,22,188,23,188,24,188,27,188,28,188,29,188,30,188,31,188,36,188,37,188,39,188,41,188,45,188,48,188,49,188,52,188,56,188,64,188,65,188,67,188,68,188,69,188,73,188,76,188,77,188,80,188,93,188,132,188,133,188,136,188,139,188,140,188,142,188,148,188,149,188,151,188,153,188,154,188,160,188,161,188,164,188,167,188,168,188,176,188,177,188,179,188,180,188,181,188,188,188,189,188,192,188,196,188,205,188,207,188,208,188,209,188,213,188,216,188,220,188,244,188,245,188,246,188,248,188,252,188,4,189,5,189,7,189,9,189,16,189,20,189,36,189,44,189,64,189,72,189,73,189,76,189,80,189,88,189,89,189,100,189,104,189,128,189,129,189,132,189,135,189,136,189,137,189,138,189,144,189,145,189,147,189,149,189,153,189,154,189,156,189,164,189,176,189,184,189,212,189,213,189,216,189,220,189,233,189,240,189,244,189,248,189,0,190,3,190,5,190,12,190,13,190,16,190,20,190,28,190,29,190,31,190,68,190,69,190,72,190,76,190,78,190,84,190,85,190,87,190,89,190,90,190,91,190,96,190,97,190,100,190,104,190,106,190,112,190,113,190,115,190,116,190,117,190,123,190,124,190,125,190,128,190,132,190,140,190,141,190,143,190,144,190,145,190,152,190,153,190,168,190,208,190,209,190,212,190,215,190,216,190,224,190,227,190,228,190,229,190,236,190,1,191,8,191,9,191,24,191,25,191,27,191,28,191,29,191,64,191,65,191,68,191,72,191,80,191,81,191,85,191,148,191,176,191,197,191,204,191,205,191,208,191,212,191,220,191,223,191,225,191,60,192,81,192,88,192,92,192,96,192,104,192,105,192,144,192,145,192,148,192,152,192,160,192,161,192,163,192,165,192,172,192,173,192,175,192,176,192,179,192,180,192,181,192,182,192,188,192,189,192,191,192,192,192,193,192,197,192,200,192,201,192,204,192,208,192,216,192,217,192,219,192,220,192,221,192,228,192,229,192,232,192,236,192,244,192,245,192,247,192,249,192,0,193,4,193,8,193,16,193,21,193,28,193,29,193,30,193,31,193,32,193,35,193,36,193,38,193,39,193,44,193,45,193,47,193,48,193,49,193,54,193,56,193,57,193,60,193,64,193,72,193,73,193,75,193,76,193,77,193,84,193,85,193,88,193,92,193,100,193,101,193,103,193,104,193,105,193,112,193,116,193,120,193,133,193,140,193,141,193,142,193,144,193,148,193,150,193,156,193,157,193,159,193,161,193,165,193,168,193,169,193,172,193,176,193,189,193,196,193,200,193,204,193,212,193,215,193,216,193,224,193,228,193,232,193,240,193,241,193,243,193,252,193,253,193,0,194,4,194,12,194,13,194,15,194,17,194,24,194,25,194,28,194,31,194,32,194,40,194,41,194,43,194,45,194,47,194,49,194,50,194,52,194,72,194,80,194,81,194,84,194,88,194,96,194,101,194,108,194,109,194,112,194,116,194,124,194,125,194,127,194,129,194,136,194,137,194,144,194,152,194,155,194,157,194,164,194,165,194,168,194,172,194,173,194,180,194,181,194,183,194,185,194,220,194,221,194,224,194,227,194,228,194,235,194,236,194,237,194,239,194,241,194,246,194,248,194,249,194,251,194,252,194,0,195,8,195,9,195,12,195,13,195,19,195,20,195,21,195,24,195,28,195,36,195,37,195,40,195,41,195,69,195,104,195,105,195,108,195,112,195,114,195,120,195,121,195,124,195,125,195,132,195,136,195,140,195,192,195,216,195,217,195,220,195,223,195,224,195,226,195,232,195,233,195,237,195,244,195,245,195,248,195,8,196,16,196,36,196,44,196,48,196,52,196,60,196,61,196,72,196,100,196,101,196,104,196,108,196,116,196,117,196,121,196,128,196,148,196,156,196,184,196,188,196,233,196,240,196,241,196,244,196,248,196,250,196,255,196,0,197,1,197,12,197,16,197,20,197,28,197,40,197,41,197,44,197,48,197,56,197,57,197,59,197,61,197,68,197,69,197,72,197,73,197,74,197,76,197,77,197,78,197,83,197,84,197,85,197,87,197,88,197,89,197,93,197,94,197,96,197,97,197,100,197,104,197,112,197,113,197,115,197,116,197,117,197,124,197,125,197,128,197,132,197,135,197,140,197,141,197,143,197,145,197,149,197,151,197,152,197,156,197,160,197,169,197,180,197,181,197,184,197,185,197,187,197,188,197,189,197,190,197,196,197,197,197,198,197,199,197,200,197,201,197,202,197,204,197,206,197,208,197,209,197,212,197,216,197,224,197,225,197,227,197,229,197,236,197,237,197,238,197,240,197,244,197,246,197,247,197,252,197,253,197,254,197,255,197,0,198,1,198,5,198,6,198,7,198,8,198,12,198,16,198,24,198,25,198,27,198,28,198,36,198,37,198,40,198,44,198,45,198,46,198,48,198,51,198,52,198,53,198,55,198,57,198,59,198,64,198,65,198,68,198,72,198,80,198,81,198,83,198,84,198,85,198,92,198,93,198,96,198,108,198,111,198,113,198,120,198,121,198,124,198,128,198,136,198,137,198,139,198,141,198,148,198,149,198,152,198,156,198,164,198,165,198,167,198,169,198,176,198,177,198,180,198,184,198,185,198,186,198,192,198,193,198,195,198,197,198,204,198,205,198,208,198,212,198,220,198,221,198,224,198,225,198,232,198,233,198,236,198,240,198,248,198,249,198,253,198,4,199,5,199,8,199,12,199,20,199,21,199,23,199,25,199,32,199,33,199,36,199,40,199,48,199,49,199,51,199,53,199,55,199,60,199,61,199,64,199,68,199,74,199,76,199,77,199,79,199,81,199,82,199,83,199,84,199,85,199,86,199,87,199,88,199,92,199,96,199,104,199,107,199,116,199,117,199,120,199,124,199,125,199,126,199,131,199,132,199,133,199,135,199,136,199,137,199,138,199,142,199,144,199,145,199,148,199,150,199,151,199,152,199,154,199,160,199,161,199,163,199,164,199,165,199,166,199,172,199,173,199,176,199,180,199,188,199,189,199,191,199,192,199,193,199,200,199,201,199,204,199,206,199,208,199,216,199,221,199,228,199,232,199,236,199,0,200,1,200,4,200,8,200,10,200,16,200,17,200,19,200,21,200,22,200,28,200,29,200,32,200,36,200,44,200,45,200,47,200,49,200,56,200,60,200,64,200,72,200,73,200,76,200,77,200,84,200,112,200,113,200,116,200,120,200,122,200,128,200,129,200,131,200,133,200,134,200,135,200,139,200,140,200,141,200,148,200,157,200,159,200,161,200,168,200,188,200,189,200,196,200,200,200,204,200,212,200,213,200,215,200,217,200,224,200,225,200,228,200,245,200,252,200,253,200,0,201,4,201,5,201,6,201,12,201,13,201,15,201,17,201,24,201,44,201,52,201,80,201,81,201,84,201,88,201,96,201,97,201,99,201,108,201,112,201,116,201,124,201,136,201,137,201,140,201,144,201,152,201,153,201,155,201,157,201,192,201,193,201,196,201,199,201,200,201,202,201,208,201,209,201,211,201,213,201,214,201,217,201,218,201,220,201,221,201,224,201,226,201,228,201,231,201,236,201,237,201,239,201,240,201,241,201,248,201,249,201,252,201,0,202,8,202,9,202,11,202,12,202,13,202,20,202,24,202,41,202,76,202,77,202,80,202,84,202,92,202,93,202,95,202,96,202,97,202,104,202,125,202,132,202,152,202,188,202,189,202,192,202,196,202,204,202,205,202,207,202,209,202,211,202,216,202,217,202,224,202,236,202,244,202,8,203,16,203,20,203,24,203,32,203,33,203,65,203,72,203,73,203,76,203,80,203,88,203,89,203,93,203,100,203,120,203,121,203,156,203,184,203,212,203,228,203,231,203,233,203,12,204,13,204,16,204,20,204,28,204,29,204,33,204,34,204,39,204,40,204,41,204,44,204,46,204,48,204,56,204,57,204,59,204,60,204,61,204,62,204,68,204,69,204,72,204,76,204,84,204,85,204,87,204,88,204,89,204,96,204,100,204,102,204,104,204,112,204,117,204,152,204,153,204,156,204,160,204,168,204,169,204,171,204,172,204,173,204,180,204,181,204,184,204,188,204,196,204,197,204,199,204,201,204,208,204,212,204,228,204,236,204,240,204,1,205,8,205,9,205,12,205,16,205,24,205,25,205,27,205,29,205,36,205,40,205,44,205,57,205,92,205,96,205,100,205,108,205,109,205,111,205,113,205,120,205,136,205,148,205,149,205,152,205,156,205,164,205,165,205,167,205,169,205,176,205,196,205,204,205,208,205,232,205,236,205,240,205,248,205,249,205,251,205,253,205,4,206,8,206,12,206,20,206,25,206,32,206,33,206,36,206,40,206,48,206,49,206,51,206,53,206,88,206,89,206,92,206,95,206,96,206,97,206,104,206,105,206,107,206,109,206,116,206,117,206,120,206,124,206,132,206,133,206,135,206,137,206,144,206,145,206,148,206,152,206,160,206,161,206,163,206,164,206,165,206,172,206,173,206,193,206,228,206,229,206,232,206,235,206,236,206,244,206,245,206,247,206,248,206,249,206,0,207,1,207,4,207,8,207,16,207,17,207,19,207,21,207,28,207,32,207,36,207,44,207,45,207,47,207,48,207,49,207,56,207,84,207,85,207,88,207,92,207,100,207,101,207,103,207,105,207,112,207,113,207,116,207,120,207,128,207,133,207,140,207,161,207,168,207,176,207,196,207,224,207,225,207,228,207,232,207,240,207,241,207,243,207,245,207,252,207,0,208,4,208,17,208,24,208,45,208,52,208,53,208,56,208,60,208,68,208,69,208,71,208,73,208,80,208,84,208,88,208,96,208,108,208,109,208,112,208,116,208,124,208,125,208,129,208,164,208,165,208,168,208,172,208,180,208,181,208,183,208,185,208,192,208,193,208,196,208,200,208,201,208,208,208,209,208,211,208,212,208,213,208,220,208,221,208,224,208,228,208,236,208,237,208,239,208,240,208,241,208,248,208,13,209,48,209,49,209,52,209,56,209,58,209,64,209,65,209,67,209,68,209,69,209,76,209,77,209,80,209,84,209,92,209,93,209,95,209,97,209,104,209,108,209,124,209,132,209,136,209,160,209,161,209,164,209,168,209,176,209,177,209,179,209,181,209,186,209,188,209,192,209,216,209,244,209,248,209,7,210,9,210,16,210,44,210,45,210,48,210,52,210,60,210,61,210,63,210,65,210,72,210,92,210,100,210,128,210,129,210,132,210,136,210,144,210,145,210,149,210,156,210,160,210,164,210,172,210,177,210,184,210,185,210,188,210,191,210,192,210,194,210,200,210,201,210,203,210,212,210,216,210,220,210,228,210,229,210,240,210,241,210,244,210,248,210,0,211,1,211,3,211,5,211,12,211,13,211,14,211,16,211,20,211,22,211,28,211,29,211,31,211,32,211,33,211,37,211,40,211,41,211,44,211,48,211,56,211,57,211,59,211,60,211,61,211,68,211,69,211,124,211,125,211,128,211,132,211,140,211,141,211,143,211,144,211,145,211,152,211,153,211,156,211,160,211,168,211,169,211,171,211,173,211,180,211,184,211,188,211,196,211,197,211,200,211,201,211,208,211,216,211,225,211,227,211,236,211,237,211,240,211,244,211,252,211,253,211,255,211,1,212,8,212,29,212,64,212,68,212,92,212,96,212,100,212,109,212,111,212,120,212,121,212,124,212,127,212,128,212,130,212,136,212,137,212,139,212,141,212,148,212,169,212,204,212,208,212,212,212,220,212,223,212,232,212,236,212,240,212,248,212,251,212,253,212,4,213,8,213,12,213,20,213,21,213,23,213,60,213,61,213,64,213,68,213,76,213,77,213,79,213,81,213,88,213,89,213,92,213,96,213,101,213,104,213,105,213,107,213,109,213,116,213,117,213,120,213,124,213,132,213,133,213,135,213,136,213,137,213,144,213,165,213,200,213,201,213,204,213,208,213,210,213,216,213,217,213,219,213,221,213,228,213,229,213,232,213,236,213,244,213,245,213,247,213,249,213,0,214,1,214,4,214,8,214,16,214,17,214,19,214,20,214,21,214,28,214,32,214,36,214,45,214,56,214,57,214,60,214,64,214,69,214,72,214,73,214,75,214,77,214,81,214,84,214,85,214,88,214,92,214,103,214,105,214,112,214,113,214,116,214,131,214,133,214,140,214,141,214,144,214,148,214,157,214,159,214,161,214,168,214,172,214,176,214,185,214,187,214,196,214,197,214,200,214,204,214,209,214,212,214,215,214,217,214,224,214,228,214,232,214,240,214,245,214,252,214,253,214,0,215,4,215,17,215,24,215,25,215,28,215,32,215,40,215,41,215,43,215,45,215,52,215,53,215,56,215,60,215,68,215,71,215,73,215,80,215,81,215,84,215,86,215,87,215,88,215,89,215,96,215,97,215,99,215,101,215,105,215,108,215,112,215,116,215,124,215,125,215,129,215,136,215,137,215,140,215,144,215,152,215,153,215,155,215,157,215,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,61,79,115,79,71,80,249,80,160,82,239,83,117,84,229,84,9,86,193,90,182,91,135,102,182,103,183,103,239,103,76,107,194,115,194,117,60,122,219,130,4,131,87,136,136,136,54,138,200,140,207,141,251,142,230,143,213,153,59,82,116,83,4,84,106,96,100,97,188,107,207,115,26,129,186,137,210,137,163,149,131,79,10,82,190,88,120,89,230,89,114,94,121,94,199,97,192,99,70,103,236,103,127,104,151,111,78,118,11,119,245,120,8,122,255,122,33,124,157,128,110,130,113,130,235,138,147,149,107,78,157,85,247,102,52,110,163,120,237,122,91,132,16,137,78,135,168,151,216,82,78,87,42,88,76,93,31,97,190,97,33,98,98,101,209,103,68,106,27,110,24,117,179,117,227,118,176,119,58,125,175,144,81,148,82,148,149,159,35,83,172,92,50,117,219,128,64,146,152,149,91,82,8,88,220,89,161,92,23,93,183,94,58,95,74,95,119,97,95,108,122,117,134,117,224,124,115,125,177,125,140,127,84,129,33,130,145,133,65,137,27,139,252,146,77,150,71,156,203,78,247,78,11,80,241,81,79,88,55,97,62,97,104,97,57,101,234,105,17,111,165,117,134,118,214,118,135,123,165,130,203,132,0,249,167,147,139,149,128,85,162,91,81,87,1,249,179,124,185,127,181,145,40,80,187,83,69,92,232,93,210,98,110,99,218,100,231,100,32,110,172,112,91,121,221,141,30,142,2,249,125,144,69,146,248,146,126,78,246,78,101,80,254,93,250,94,6,97,87,105,113,129,84,134,71,142,117,147,43,154,94,78,145,80,112,103,64,104,9,81,141,82,146,82,162,106,188,119,16,146,212,158,171,82,47,96,242,143,72,80,169,97,237,99,202,100,60,104,132,106,192,111,136,129,161,137,148,150,5,88,125,114,172,114,4,117,121,125,109,126,169,128,139,137,116,139,99,144,81,157,137,98,122,108,84,111,80,125,58,127,35,138,124,81,74,97,157,123,25,139,87,146,140,147,172,78,211,79,30,80,190,80,6,81,193,82,205,82,127,83,112,87,131,88,154,94,145,95,118,97,172,97,206,100,108,101,111,102,187,102,244,102,151,104,135,109,133,112,241,112,159,116,165,116,202,116,217,117,108,120,236,120,223,122,246,122,69,125,147,125,21,128,63,128,27,129,150,131,102,139,21,143,21,144,225,147,3,152,56,152,90,154,232,155,194,79,83,85,58,88,81,89,99,91,70,92,184,96,18,98,66,104,176,104,232,104,170,110,76,117,120,118,206,120,61,122,251,124,107,126,124,126,8,138,161,138,63,140,142,150,196,157,228,83,233,83,74,84,113,84,250,86,209,89,100,91,59,92,171,94,247,98,55,101,69,101,114,101,160,102,175,103,193,105,189,108,252,117,144,118,126,119,63,122,148,127,3,128,161,128,143,129,230,130,253,130,240,131,193,133,49,136,180,136,165,138,3,249,156,143,46,147,199,150,103,152,216,154,19,159,237,84,155,101,242,102,143,104,64,122,55,140,96,157,240,86,100,87,17,93,6,102,177,104,205,104,254,110,40,116,158,136,228,155,104,108,4,249,168,154,155,79,108,81,113,81,159,82,84,91,229,93,80,96,109,96,241,98,167,99,59,101,217,115,122,122,163,134,162,140,143,151,50,78,225,91,8,98,156,103,220,116,209,121,211,131,135,138,178,138,232,141,78,144,75,147,70,152,211,94,232,105,255,133,237,144,5,249,160,81,152,91,236,91,99,97,250,104,62,107,76,112,47,116,216,116,161,123,80,127,197,131,192,137,171,140,220,149,40,153,46,82,93,96,236,98,2,144,138,79,73,81,33,83,217,88,227,94,224,102,56,109,154,112,194,114,214,115,80,123,241,128,91,148,102,83,155,99,107,127,86,78,128,80,74,88,222,88,42,96,39,97,208,98,208,105,65,155,143,91,24,125,177,128,95,143,164,78,209,80,172,84,172,85,12,91,160,93,231,93,42,101,78,101,33,104,75,106,225,114,142,118,239,119,94,125,249,127,160,129,78,133,223,134,3,143,78,143,202,144,3,153,85,154,171,155,24,78,69,78,93,78,199,78,241,79,119,81,254,82,64,83,227,83,229,83,142,84,20,86,117,87,162,87,199,91,135,93,208,94,252,97,216,98,81,101,184,103,233,103,203,105,80,107,198,107,236,107,66,108,157,110,120,112,215,114,150,115,3,116,191,119,233,119,118,122,127,125,9,128,252,129,5,130,10,130,223,130,98,136,51,139,252,140,192,142,17,144,177,144,100,146,182,146,210,153,69,154,233,156,215,157,156,159,11,87,64,92,202,131,160,151,171,151,180,158,27,84,152,122,164,127,217,136,205,142,225,144,0,88,72,92,152,99,159,122,174,91,19,95,121,122,174,122,142,130,172,142,38,80,56,82,248,82,119,83,8,87,243,98,114,99,10,107,195,109,55,119,165,83,87,115,104,133,118,142,213,149,58,103,195,106,112,111,109,138,204,142,75,153,6,249,119,102,120,107,180,140,60,155,7,249,235,83,45,87,78,89,198,99,251,105,234,115,69,120,186,122,197,122,254,124,117,132,143,137,115,141,53,144,168,149,251,82,71,87,71,117,96,123,204,131,30,146,8,249,88,106,75,81,75,82,135,82,31,98,216,104,117,105,153,150,197,80,164,82,228,82,195,97,164,101,57,104,255,105,126,116,75,123,185,130,235,131,178,137,57,139,209,143,73,153,9,249,202,78,151,89,210,100,17,102,142,106,52,116,129,121,189,121,169,130,126,136,127,136,95,137,10,249,38,147,11,79,202,83,37,96,113,98,114,108,26,125,102,125,152,78,98,81,220,119,175,128,1,79,14,79,118,81,128,81,220,85,104,86,59,87,250,87,252,87,20,89,71,89,147,89,196,91,144,92,14,93,241,93,126,94,204,95,128,98,215,101,227,101,30,103,31,103,94,103,203,104,196,104,95,106,58,107,35,108,125,108,130,108,199,109,152,115,38,116,42,116,130,116,163,116,120,117,127,117,129,120,239,120,65,121,71,121,72,121,122,121,149,123,0,125,186,125,136,127,6,128,45,128,140,128,24,138,79,139,72,140,119,141,33,147,36,147,226,152,81,153,14,154,15,154,101,154,146,158,202,125,118,79,9,84,238,98,84,104,209,145,171,85,58,81,11,249,12,249,28,90,230,97,13,249,207,98,255,98,14,249,15,249,16,249,17,249,18,249,19,249,163,144,20,249,21,249,22,249,23,249,24,249,254,138,25,249,26,249,27,249,28,249,150,102,29,249,86,113,30,249,31,249,227,150,32,249,79,99,122,99,87,83,33,249,143,103,96,105,115,110,34,249,55,117,35,249,36,249,37,249,13,125,38,249,39,249,114,136,202,86,24,90,40,249,41,249,42,249,43,249,44,249,67,78,45,249,103,81,72,89,240,103,16,128,46,249,115,89,116,94,154,100,202,121,245,95,108,96,200,98,123,99,231,91,215,91,170,82,47,249,116,89,41,95,18,96,48,249,49,249,50,249,89,116,51,249,52,249,53,249,54,249,55,249,56,249,209,153,57,249,58,249,59,249,60,249,61,249,62,249,63,249,64,249,65,249,66,249,67,249,195,111,68,249,69,249,191,129,178,143,241,96,70,249,71,249,102,129,72,249,73,249,63,92,74,249,75,249,76,249,77,249,78,249,79,249,80,249,81,249,233,90,37,138,123,103,16,125,82,249,83,249,84,249,85,249,86,249,87,249,253,128,88,249,89,249,60,92,229,108,63,83,186,110,26,89,54,131,57,78,182,78,70,79,174,85,24,87,199,88,86,95,183,101,230,101,128,106,181,107,77,110,237,119,239,122,30,124,222,125,203,134,146,136,50,145,91,147,187,100,190,111,122,115,184,117,84,144,86,85,77,87,186,97,212,100,199,102,225,109,91,110,109,111,185,111,240,117,67,128,189,129,65,133,131,137,199,138,90,139,31,147,147,108,83,117,84,123,15,142,93,144,16,85,2,88,88,88,98,94,7,98,158,100,224,104,118,117,214,124,179,135,232,158,227,78,136,87,110,87,39,89,13,92,177,92,54,94,133,95,52,98,225,100,179,115,250,129,139,136,184,140,138,150,219,158,133,91,183,95,179,96,18,80,0,82,48,82,22,87,53,88,87,88,14,92,96,92,246,92,139,93,166,94,146,95,188,96,17,99,137,99,23,100,67,104,249,104,194,106,216,109,33,110,212,110,228,111,254,113,220,118,121,119,177,121,59,122,4,132,169,137,237,140,243,141,72,142,3,144,20,144,83,144,253,144,77,147,118,150,220,151,210,107,6,112,88,114,162,114,104,115,99,119,191,121,228,123,155,126,128,139,169,88,199,96,102,101,253,101,190,102,140,108,30,113,201,113,90,140,19,152,109,78,129,122,221,78,172,81,205,81,213,82,12,84,167,97,113,103,80,104,223,104,30,109,124,111,188,117,179,119,229,122,244,128,99,132,133,146,92,81,151,101,92,103,147,103,216,117,199,122,115,131,90,249,70,140,23,144,45,152,111,92,192,129,154,130,65,144,111,144,13,146,151,95,157,93,89,106,200,113,123,118,73,123,228,133,4,139,39,145,48,154,135,85,246,97,91,249,105,118,133,127,63,134,186,135,248,136,143,144,92,249,27,109,217,112,222,115,97,125,61,132,93,249,106,145,241,153,94,249,130,78,117,83,4,107,18,107,62,112,27,114,45,134,30,158,76,82,163,143,80,93,229,100,44,101,22,107,235,111,67,124,156,126,205,133,100,137,189,137,201,98,216,129,31,136,202,94,23,103,106,109,252,114,5,116,111,116,130,135,222,144,134,79,13,93,160,95,10,132,183,81,160,99,101,117,174,78,6,80,105,81,201,81,129,104,17,106,174,124,177,124,231,124,111,130,210,138,27,143,207,145,182,79,55,81,245,82,66,84,236,94,110,97,62,98,197,101,218,106,254,111,42,121,220,133,35,136,173,149,98,154,106,154,151,158,206,158,155,82,198,102,119,107,29,112,43,121,98,143,66,151,144,97,0,98,35,101,35,111,73,113,137,116,244,125,111,128,238,132,38,143,35,144,74,147,189,81,23,82,163,82,12,109,200,112,194,136,201,94,130,101,174,107,194,111,62,124,117,115,228,78,54,79,249,86,95,249,186,92,186,93,28,96,178,115,45,123,154,127,206,127,70,128,30,144,52,146,246,150,72,151,24,152,97,159,139,79,167,111,174,121,180,145,183,150,222,82,96,249,136,100,196,100,211,106,94,111,24,112,16,114,231,118,1,128,6,134,92,134,239,141,5,143,50,151,111,155,250,157,117,158,140,120,127,121,160,125,201,131,4,147,127,158,147,158,214,138,223,88,4,95,39,103,39,112,207,116,96,124,126,128,33,81,40,112,98,114,202,120,194,140,218,140,244,140,247,150,134,78,218,80,238,91,214,94,153,101,206,113,66,118,173,119,74,128,252,132,124,144,39,155,141,159,216,88,65,90,98,92,19,106,218,109,15,111,59,118,47,125,55,126,30,133,56,137,228,147,75,150,137,82,210,101,243,103,180,105,65,109,156,110,15,112,9,116,96,116,89,117,36,118,107,120,44,139,94,152,109,81,46,98,120,150,150,79,43,80,25,93,234,109,184,125,42,143,139,95,68,97,23,104,97,249,134,150,210,82,139,128,220,81,204,81,94,105,28,122,190,125,241,131,117,150,218,79,41,82,152,83,15,84,14,85,101,92,167,96,78,103,168,104,108,109,129,114,248,114,6,116,131,116,98,249,226,117,108,124,121,127,184,127,137,131,207,136,225,136,204,145,208,145,226,150,201,155,29,84,126,111,208,113,152,116,250,133,170,142,163,150,87,156,159,158,151,103,203,109,51,116,232,129,22,151,44,120,203,122,32,123,146,124,105,100,106,116,242,117,188,120,232,120,172,153,84,155,187,158,222,91,85,94,32,111,156,129,171,131,136,144,7,78,77,83,41,90,210,93,78,95,98,97,61,99,105,102,252,102,255,110,43,111,99,112,158,119,44,132,19,133,59,136,19,143,69,153,59,156,28,85,185,98,43,103,171,108,9,131,106,137,122,151,161,78,132,89,216,95,217,95,27,103,178,125,84,127,146,130,43,131,189,131,30,143,153,144,203,87,185,89,146,90,208,91,39,102,154,103,133,104,207,107,100,113,117,127,183,140,227,140,129,144,69,155,8,129,138,140,76,150,64,154,165,158,95,91,19,108,27,115,242,118,223,118,12,132,170,81,147,137,77,81,149,81,201,82,201,104,148,108,4,119,32,119,191,125,236,125,98,151,181,158,197,110,17,133,165,81,13,84,125,84,14,102,157,102,39,105,159,110,191,118,145,119,23,131,194,132,159,135,105,145,152,146,244,156,130,136,174,79,146,81,223,82,198,89,61,94,85,97,120,100,121,100,174,102,208,103,33,106,205,107,219,107,95,114,97,114,65,116,56,119,219,119,23,128,188,130,5,131,0,139,40,139,140,140,40,103,144,108,103,114,238,118,102,119,70,122,169,157,127,107,146,108,34,89,38,103,153,132,111,83,147,88,153,89,223,94,207,99,52,102,115,103,58,110,43,115,215,122,215,130,40,147,217,82,235,93,174,97,203,97,10,98,199,98,171,100,224,101,89,105,102,107,203,107,33,113,247,115,93,117,70,126,30,130,2,131,106,133,163,138,191,140,39,151,97,157,168,88,216,158,17,80,14,82,59,84,79,85,135,101,118,108,10,125,11,125,94,128,138,134,128,149,239,150,255,82,149,108,105,114,115,84,154,90,62,92,75,93,76,95,174,95,42,103,182,104,99,105,60,110,68,110,9,119,115,124,142,127,135,133,14,139,247,143,97,151,244,158,183,92,182,96,13,97,171,97,79,101,251,101,252,101,17,108,239,108,159,115,201,115,225,125,148,149,198,91,28,135,16,139,93,82,90,83,205,98,15,100,178,100,52,103,56,106,202,108,192,115,158,116,148,123,149,124,27,126,138,129,54,130,132,133,235,143,249,150,193,153,52,79,74,83,205,83,219,83,204,98,44,100,0,101,145,101,195,105,238,108,88,111,237,115,84,117,34,118,228,118,252,118,208,120,251,120,44,121,70,125,44,130,224,135,212,143,18,152,239,152,195,82,212,98,165,100,36,110,81,111,124,118,203,141,177,145,98,146,238,154,67,155,35,80,141,80,74,87,168,89,40,92,71,94,119,95,63,98,62,101,185,101,193,101,9,102,139,103,156,105,194,110,197,120,33,125,170,128,128,129,43,130,179,130,161,132,140,134,42,138,23,139,166,144,50,150,144,159,13,80,243,79,99,249,249,87,152,95,220,98,146,99,111,103,67,110,25,113,195,118,204,128,218,128,244,136,245,136,25,137,224,140,41,143,77,145,106,150,47,79,112,79,27,94,207,103,34,104,125,118,126,118,68,155,97,94,10,106,105,113,212,113,106,117,100,249,65,126,67,133,233,133,220,152,16,79,79,123,112,127,165,149,225,81,6,94,181,104,62,108,78,108,219,108,175,114,196,123,3,131,213,108,58,116,251,80,136,82,193,88,216,100,151,106,167,116,86,118,167,120,23,134,226,149,57,151,101,249,94,83,1,95,138,139,168,143,175,143,138,144,37,82,165,119,73,156,8,159,25,78,2,80,117,81,91,92,119,94,30,102,58,102,196,103,197,104,179,112,1,117,197,117,201,121,221,122,39,143,32,153,8,154,221,79,33,88,49,88,246,91,110,102,101,107,17,109,122,110,125,111,228,115,43,117,233,131,220,136,19,137,92,139,20,143,15,79,213,80,16,83,92,83,147,91,169,95,13,103,143,121,121,129,47,131,20,133,7,137,134,137,57,143,59,143,165,153,18,156,44,103,118,78,248,79,73,89,1,92,239,92,240,92,103,99,210,104,253,112,162,113,43,116,43,126,236,132,2,135,34,144,210,146,243,156,13,78,216,78,239,79,133,80,86,82,111,82,38,84,144,84,224,87,43,89,102,90,90,91,117,91,204,91,156,94,102,249,118,98,119,101,167,101,110,109,165,110,54,114,38,123,63,124,54,127,80,129,81,129,154,129,64,130,153,130,169,131,3,138,160,140,230,140,251,140,116,141,186,141,232,144,220,145,28,150,68,150,217,153,231,156,23,83,6,82,41,84,116,86,179,88,84,89,110,89,255,95,164,97,110,98,16,102,126,108,26,113,198,118,137,124,222,124,27,125,172,130,193,140,240,150,103,249,91,79,23,95,127,95,194,98,41,93,11,103,218,104,124,120,67,126,108,157,21,78,153,80,21,83,42,83,81,83,131,89,98,90,135,94,178,96,138,97,73,98,121,98,144,101,135,103,167,105,212,107,214,107,215,107,216,107,184,108,104,249,53,116,250,117,18,120,145,120,213,121,216,121,131,124,203,125,225,127,165,128,62,129,194,129,242,131,26,135,232,136,185,138,108,139,187,140,25,145,94,151,219,152,59,159,172,86,42,91,108,95,140,101,179,106,175,107,92,109,241,111,21,112,93,114,173,115,167,140,211,140,59,152,145,97,55,108,88,128,1,154,77,78,139,78,155,78,213,78,58,79,60,79,127,79,223,79,255,80,242,83,248,83,6,85,227,85,219,86,235,88,98,89,17,90,235,91,250,91,4,92,243,93,43,94,153,95,29,96,104,99,156,101,175,101,246,103,251,103,173,104,123,107,153,108,215,108,35,110,9,112,69,115,2,120,62,121,64,121,96,121,193,121,233,123,23,125,114,125,134,128,13,130,142,131,209,132,199,134,223,136,80,138,94,138,29,139,220,140,102,141,173,143,170,144,252,152,223,153,157,158,74,82,105,249,20,103,106,249,152,80,42,82,113,92,99,101,85,108,202,115,35,117,157,117,151,123,156,132,120,145,48,151,119,78,146,100,186,107,94,113,169,133,9,78,107,249,73,103,238,104,23,110,159,130,24,133,107,136,247,99,129,111,18,146,175,152,10,78,183,80,207,80,31,81,70,85,170,85,23,86,64,91,25,92,224,92,56,94,138,94,160,94,194,94,243,96,81,104,97,106,88,110,61,114,64,114,192,114,248,118,101,121,177,123,212,127,243,136,244,137,115,138,97,140,222,140,28,151,94,88,189,116,253,140,199,85,108,249,97,122,34,125,114,130,114,114,31,117,37,117,109,249,25,123,133,88,251,88,188,93,143,94,182,94,144,95,85,96,146,98,127,99,77,101,145,102,217,102,248,102,22,104,242,104,128,114,94,116,110,123,110,125,214,125,114,127,229,128,18,130,175,133,127,137,147,138,29,144,228,146,205,158,32,159,21,89,109,89,45,94,220,96,20,102,115,102,144,103,80,108,197,109,95,111,243,119,169,120,198,132,203,145,43,147,217,78,202,80,72,81,132,85,11,91,163,91,71,98,126,101,203,101,50,110,125,113,1,116,68,116,135,116,191,116,108,118,170,121,218,125,85,126,168,127,122,129,179,129,57,130,26,134,236,135,117,138,227,141,120,144,145,146,37,148,77,153,174,155,104,83,81,92,84,105,196,108,41,109,43,110,12,130,155,133,59,137,45,138,170,138,234,150,103,159,97,82,185,102,178,107,150,126,254,135,13,141,131,149,93,150,29,101,137,109,238,113,110,249,206,87,211,89,172,91,39,96,250,96,16,98,31,102,95,102,41,115,249,115,219,118,1,119,108,123,86,128,114,128,101,129,160,138,146,145,22,78,226,82,114,107,23,109,5,122,57,123,48,125,111,249,176,140,236,83,47,86,81,88,181,91,15,92,17,92,226,93,64,98,131,99,20,100,45,102,179,104,188,108,136,109,175,110,31,112,164,112,210,113,38,117,143,117,142,117,25,118,17,123,224,123,43,124,32,125,57,125,44,133,109,133,7,134,52,138,13,144,97,144,181,144,183,146,246,151,55,154,215,79,108,92,95,103,145,109,159,124,140,126,22,139,22,141,31,144,107,91,253,93,13,100,192,132,92,144,225,152,135,115,139,91,154,96,126,103,222,109,31,138,166,138,1,144,12,152,55,82,112,249,81,112,142,120,150,147,112,136,215,145,238,79,215,83,253,85,218,86,130,87,253,88,194,90,136,91,171,92,192,92,37,94,1,97,13,98,75,98,136,99,28,100,54,101,120,101,57,106,138,107,52,108,25,109,49,111,231,113,233,114,120,115,7,116,178,116,38,118,97,119,192,121,87,122,234,122,185,124,143,125,172,125,97,126,158,127,41,129,49,131,144,132,218,132,234,133,150,136,176,138,144,139,56,143,66,144,131,144,108,145,150,146,185,146,139,150,167,150,168,150,214,150,0,151,8,152,150,153,211,154,26,155,212,83,126,88,25,89,112,91,191,91,209,109,90,111,159,113,33,116,185,116,133,128,253,131,225,93,135,95,170,95,66,96,236,101,18,104,111,105,83,106,137,107,53,109,243,109,227,115,254,118,172,119,77,123,20,125,35,129,28,130,64,131,244,132,99,133,98,138,196,138,135,145,30,147,6,152,180,153,12,98,83,136,240,143,101,146,7,93,39,93,105,93,95,116,157,129,104,135,213,111,254,98,210,127,54,137,114,137,30,78,88,78,231,80,221,82,71,83,127,98,7,102,105,126,5,136,94,150,141,79,25,83,54,86,203,89,164,90,56,92,78,92,77,92,2,94,17,95,67,96,189,101,47,102,66,102,190,103,244,103,28,115,226,119,58,121,197,127,148,132,205,132,150,137,102,138,105,138,225,138,85,140,122,140,244,87,212,91,15,95,111,96,237,98,13,105,150,107,92,110,132,113,210,123,85,135,88,139,254,142,223,152,254,152,56,79,129,79,225,79,123,84,32,90,184,91,60,97,176,101,104,102,252,113,51,117,94,121,51,125,78,129,227,129,152,131,170,133,206,133,3,135,10,138,171,142,155,143,113,249,197,143,49,89,164,91,230,91,137,96,233,91,11,92,195,95,129,108,114,249,241,109,11,112,26,117,175,130,246,138,192,78,65,83,115,249,217,150,15,108,158,78,196,79,82,81,94,85,37,90,232,92,17,98,89,114,189,130,170,131,254,134,89,136,29,138,63,150,197,150,19,153,9,157,93,157,10,88,179,92,189,93,68,94,225,96,21,97,225,99,2,106,37,110,2,145,84,147,78,152,16,156,119,159,137,91,184,92,9,99,79,102,72,104,60,119,193,150,141,151,84,152,159,155,161,101,1,139,203,142,188,149,53,85,169,92,214,93,181,94,151,102,76,118,244,131,199,149,211,88,188,98,206,114,40,157,240,78,46,89,15,96,59,102,131,107,231,121,38,157,147,83,192,84,195,87,22,93,27,97,214,102,175,109,141,120,126,130,152,150,68,151,132,83,124,98,150,99,178,109,10,126,75,129,77,152,251,106,76,127,175,157,26,158,95,78,59,80,182,81,28,89,249,96,246,99,48,105,58,114,54,128,116,249,206,145,49,95,117,249,118,249,4,125,229,130,111,132,187,132,229,133,141,142,119,249,111,79,120,249,121,249,228,88,67,91,89,96,218,99,24,101,109,101,152,102,122,249,74,105,35,106,11,109,1,112,108,113,210,117,13,118,179,121,112,122,123,249,138,127,124,249,68,137,125,249,147,139,192,145,125,150,126,249,10,153,4,87,161,95,188,101,1,111,0,118,166,121,158,138,173,153,90,155,108,159,4,81,182,97,145,98,141,106,198,129,67,80,48,88,102,95,9,113,0,138,250,138,124,91,22,134,250,79,60,81,180,86,68,89,169,99,249,109,170,93,109,105,134,81,136,78,89,79,127,249,128,249,129,249,130,89,130,249,131,249,95,107,93,108,132,249,181,116,22,121,133,249,7,130,69,130,57,131,63,143,93,143,134,249,24,153,135,249,136,249,137,249,166,78,138,249,223,87,121,95,19,102,139,249,140,249,171,117,121,126,111,139,141,249,6,144,91,154,165,86,39,88,248,89,31,90,180,91,142,249,246,94,143,249,144,249,80,99,59,99,145,249,61,105,135,108,191,108,142,109,147,109,245,109,20,111,146,249,223,112,54,113,89,113,147,249,195,113,213,113,148,249,79,120,111,120,149,249,117,123,227,125,150,249,47,126,151,249,77,136,223,142,152,249,153,249,154,249,91,146,155,249,246,156,156,249,157,249,158,249,133,96,133,109,159,249,177,113,160,249,161,249,177,149,173,83,162,249,163,249,164,249,211,103,165,249,142,112,48,113,48,116,118,130,210,130,166,249,187,149,229,154,125,158,196,102,167,249,193,113,73,132,168,249,169,249,75,88,170,249,171,249,184,93,113,95,172,249,32,102,142,102,121,105,174,105,56,108,243,108,54,110,65,111,218,111,27,112,47,112,80,113,223,113,112,115,173,249,91,116,174,249,212,116,200,118,78,122,147,126,175,249,176,249,241,130,96,138,206,143,177,249,72,147,178,249,25,151,179,249,180,249,66,78,42,80,181,249,8,82,225,83,243,102,109,108,202,111,10,115,127,119,98,122,174,130,221,133,2,134,182,249,212,136,99,138,125,139,107,140,183,249,179,146,184,249,19,151,16,152,148,78,13,79,201,79,178,80,72,83,62,84,51,84,218,85,98,88,186,88,103,89,27,90,228,91,159,96,185,249,202,97,86,101,255,101,100,102,167,104,90,108,179,111,207,112,172,113,82,115,125,123,8,135,164,138,50,156,7,159,75,92,131,108,68,115,137,115,58,146,171,110,101,116,31,118,105,122,21,126,10,134,64,81,197,88,193,100,238,116,21,117,112,118,193,127,149,144,205,150,84,153,38,110,230,116,169,122,170,122,229,129,217,134,120,135,27,138,73,90,140,91,155,91,161,104,0,105,99,109,169,115,19,116,44,116,151,120,233,125,235,127,24,129,85,129,158,131,76,140,46,150,17,152,240,102,128,95,250,101,137,103,106,108,139,115,45,80,3,90,106,107,238,119,22,89,108,93,205,93,37,115,79,117,186,249,187,249,229,80,249,81,47,88,45,89,150,89,218,89,229,91,188,249,189,249,162,93,215,98,22,100,147,100,254,100,190,249,220,102,191,249,72,106,192,249,255,113,100,116,193,249,136,122], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+128288);
/* memory initializer */ allocate([175,122,71,126,94,126,0,128,112,129,194,249,239,135,129,137,32,139,89,144,195,249,128,144,82,153,126,97,50,107,116,109,31,126,37,137,177,143,209,79,173,80,151,81,199,82,199,87,137,88,185,91,184,94,66,97,149,105,140,109,103,110,182,110,148,113,98,116,40,117,44,117,115,128,56,131,201,132,10,142,148,147,222,147,196,249,142,78,81,79,118,80,42,81,200,83,203,83,243,83,135,91,211,91,36,92,26,97,130,97,244,101,91,114,151,115,64,116,194,118,80,121,145,121,185,121,6,125,189,127,139,130,213,133,94,134,194,143,71,144,245,144,234,145,133,150,232,150,233,150,214,82,103,95,237,101,49,102,47,104,92,113,54,122,193,144,10,152,145,78,197,249,82,106,158,107,144,111,137,113,24,128,184,130,83,133,75,144,149,150,242,150,251,151,26,133,49,155,144,78,138,113,196,150,67,81,159,83,225,84,19,87,18,87,163,87,155,90,196,90,195,91,40,96,63,97,244,99,133,108,57,109,114,110,144,110,48,114,63,115,87,116,209,130,129,136,69,143,96,144,198,249,98,150,88,152,27,157,8,103,138,141,94,146,77,79,73,80,222,80,113,83,13,87,212,89,1,90,9,92,112,97,144,102,45,110,50,114,75,116,239,125,195,128,14,132,102,132,63,133,95,135,91,136,24,137,2,139,85,144,203,151,79,155,115,78,145,79,18,81,106,81,199,249,47,85,169,85,122,91,165,91,124,94,125,94,190,94,160,96,223,96,8,97,9,97,196,99,56,101,9,103,200,249,212,103,218,103,201,249,97,105,98,105,185,108,39,109,202,249,56,110,203,249,225,111,54,115,55,115,204,249,92,116,49,117,205,249,82,118,206,249,207,249,173,125,254,129,56,132,213,136,152,138,219,138,237,138,48,142,66,142,74,144,62,144,122,144,73,145,201,145,110,147,208,249,209,249,9,88,210,249,211,107,137,128,178,128,211,249,212,249,65,81,107,89,57,92,213,249,214,249,100,111,167,115,228,128,7,141,215,249,23,146,143,149,216,249,217,249,218,249,219,249,127,128,14,98,28,112,104,125,141,135,220,249,160,87,105,96,71,97,183,107,190,138,128,146,177,150,89,78,31,84,235,109,45,133,112,150,243,151,238,152,214,99,227,108,145,144,221,81,201,97,186,129,249,157,157,79,26,80,0,81,156,91,15,97,255,97,236,100,5,105,197,107,145,117,227,119,169,127,100,130,143,133,251,135,99,136,188,138,112,139,171,145,140,78,229,78,10,79,221,249,222,249,55,89,232,89,223,249,242,93,27,95,91,95,33,96,224,249,225,249,226,249,227,249,62,114,229,115,228,249,112,117,205,117,229,249,251,121,230,249,12,128,51,128,132,128,225,130,81,131,231,249,232,249,189,140,179,140,135,144,233,249,234,249,244,152,12,153,235,249,236,249,55,112,202,118,202,127,204,127,252,127,26,139,186,78,193,78,3,82,112,83,237,249,189,84,224,86,251,89,197,91,21,95,205,95,110,110,238,249,239,249,106,125,53,131,240,249,147,134,141,138,241,249,109,151,119,151,242,249,243,249,0,78,90,79,126,79,249,88,229,101,162,110,56,144,176,147,185,153,251,78,236,88,138,89,217,89,65,96,244,249,245,249,20,122,246,249,79,131,195,140,101,81,68,83,247,249,248,249,249,249,205,78,105,82,85,91,191,130,212,78,58,82,168,84,201,89,255,89,80,91,87,91,92,91,99,96,72,97,203,110,153,112,110,113,134,115,247,116,181,117,193,120,43,125,5,128,234,129,40,131,23,133,201,133,238,138,199,140,204,150,92,79,250,82,188,86,171,101,40,102,124,112,184,112,53,114,189,125,141,130,76,145,192,150,114,157,113,91,231,104,152,107,122,111,222,118,145,92,171,102,91,111,180,123,42,124,54,136,220,150,8,78,215,78,32,83,52,88,187,88,239,88,108,89,7,92,51,94,132,94,53,95,140,99,178,102,86,103,31,106,163,106,12,107,63,111,70,114,250,249,80,115,139,116,224,122,167,124,120,129,223,129,231,129,138,131,108,132,35,133,148,133,207,133,221,136,19,141,172,145,119,149,156,150,141,81,201,84,40,87,176,91,77,98,80,103,61,104,147,104,61,110,211,110,125,112,33,126,193,136,161,140,9,143,75,159,78,159,45,114,143,123,205,138,26,147,71,79,78,79,50,81,128,84,208,89,149,94,181,98,117,103,110,105,23,106,174,108,26,110,217,114,42,115,189,117,184,123,53,125,231,130,249,131,87,132,247,133,91,138,175,140,135,142,25,144,184,144,206,150,95,159,227,82,10,84,225,90,194,91,88,100,117,101,244,110,196,114,251,249,132,118,77,122,27,123,77,124,62,126,223,127,123,131,43,139,202,140,100,141,225,141,95,142,234,143,249,143,105,144,209,147,67,79,122,79,179,80,104,81,120,81,77,82,106,82,97,88,124,88,96,89,8,92,85,92,219,94,155,96,48,98,19,104,191,107,8,108,177,111,78,113,32,116,48,117,56,117,81,117,114,118,76,123,139,123,173,123,198,123,143,126,110,138,62,143,73,143,63,146,147,146,34,147,43,148,251,150,90,152,107,152,30,153,7,82,42,98,152,98,89,109,100,118,202,122,192,123,118,125,96,83,190,92,151,94,56,111,185,112,152,124,17,151,142,155,222,158,165,99,122,100,118,135,1,78,149,78,173,78,92,80,117,80,72,84,195,89,154,91,64,94,173,94,247,94,129,95,197,96,58,99,63,101,116,101,204,101,118,102,120,102,254,103,104,105,137,106,99,107,64,108,192,109,232,109,31,110,94,110,30,112,161,112,142,115,253,115,58,117,91,119,135,120,142,121,11,122,125,122,190,124,142,125,71,130,2,138,234,138,158,140,45,145,74,145,216,145,102,146,204,146,32,147,6,151,86,151,92,151,2,152,14,159,54,82,145,82,124,85,36,88,29,94,31,95,140,96,208,99,175,104,223,111,109,121,44,123,205,129,186,133,253,136,248,138,68,142,141,145,100,150,155,150,61,151,76,152,74,159,206,79,70,81,203,81,169,82,50,86,20,95,107,95,170,99,205,100,233,101,65,102,250,102,249,102,29,103,157,104,215,104,253,105,21,111,110,111,103,113,229,113,42,114,170,116,58,119,86,121,90,121,223,121,32,122,149,122,151,124,223,124,68,125,112,126,135,128,251,133,164,134,84,138,191,138,153,141,129,142,32,144,109,144,227,145,59,150,213,150,229,156,207,101,7,124,179,141,195,147,88,91,10,92,82,83,217,98,29,115,39,80,151,91,158,95,176,96,107,97,213,104,217,109,46,116,46,122,66,125,156,125,49,126,107,129,42,142,53,142,126,147,24,148,80,79,80,87,230,93,167,94,43,99,106,127,59,78,79,79,143,79,90,80,221,89,196,128,106,84,104,84,254,85,79,89,153,91,222,93,218,94,93,102,49,103,241,103,42,104,232,108,50,109,74,110,141,111,183,112,224,115,135,117,76,124,2,125,44,125,162,125,31,130,219,134,59,138,133,138,112,141,138,142,51,143,49,144,78,145,82,145,68,148,208,153,249,122,165,124,202,79,1,81,198,81,200,87,239,91,251,92,89,102,61,106,90,109,150,110,236,111,12,113,111,117,227,122,34,136,33,144,117,144,203,150,255,153,1,131,45,78,242,78,70,136,205,145,125,83,219,106,107,105,65,108,122,132,158,88,142,97,254,102,239,98,221,112,17,117,199,117,82,126,184,132,73,139,8,141,75,78,234,83,171,84,48,87,64,87,215,95,1,99,7,99,111,100,47,101,232,101,122,102,157,103,179,103,98,107,96,108,154,108,44,111,229,119,37,120,73,121,87,121,25,125,162,128,2,129,243,129,157,130,183,130,24,135,140,138,252,249,4,141,190,141,114,144,244,118,25,122,55,122,84,126,119,128,7,85,212,85,117,88,47,99,34,100,73,102,75,102,109,104,155,105,132,107,37,109,177,110,205,115,104,116,161,116,91,117,185,117,225,118,30,119,139,119,230,121,9,126,29,126,251,129,47,133,151,136,58,138,209,140,235,142,176,143,50,144,173,147,99,150,115,150,7,151,132,79,241,83,234,89,201,90,25,94,78,104,198,116,190,117,233,121,146,122,163,129,237,134,234,140,204,141,237,143,159,101,21,103,253,249,247,87,87,111,221,125,47,143,246,147,198,150,181,95,242,97,132,111,20,78,152,79,31,80,201,83,223,85,111,93,238,93,33,107,100,107,203,120,154,123,254,249,73,142,202,142,110,144,73,99,62,100,64,119,132,122,47,147,127,148,106,159,176,100,175,111,230,113,168,116,218,116,196,122,18,124,130,126,178,124,152,126,154,139,10,141,125,148,16,153,76,153,57,82,223,91,230,100,45,103,46,125,237,80,195,83,121,88,88,97,89,97,250,97,172,101,217,122,146,139,150,139,9,80,33,80,117,82,49,85,60,90,224,94,112,95,52,97,94,101,12,102,54,102,162,102,205,105,196,110,50,111,22,115,33,118,147,122,57,129,89,130,214,131,188,132,181,80,240,87,192,91,232,91,105,95,161,99,38,120,181,125,220,131,33,133,199,145,245,145,138,81,245,103,86,123,172,140,196,81,187,89,189,96,85,134,28,80,255,249,84,82,58,92,125,97,26,98,211,98,242,100,165,101,204,110,32,118,10,129,96,142,95,150,187,150,223,78,67,83,152,85,41,89,221,93,197,100,201,108,250,109,148,115,127,122,27,130,166,133,228,140,16,142,119,144,231,145,225,149,33,150,198,151,248,81,242,84,134,85,185,95,164,100,136,111,180,125,31,143,77,143,53,148,201,80,22,92,190,108,251,109,27,117,187,119,61,124,100,124,121,138,194,138,30,88,190,89,22,94,119,99,82,114,138,117,107,119,220,138,188,140,18,143,243,94,116,102,248,109,125,128,193,131,203,138,81,151,214,155,0,250,67,82,255,102,149,109,239,110,224,125,230,138,46,144,94,144,212,154,29,82,127,82,232,84,148,97,132,98,219,98,162,104,18,105,90,105,53,106,146,112,38,113,93,120,1,121,14,121,210,121,13,122,150,128,120,130,213,130,73,131,73,133,130,140,133,141,98,145,139,145,174,145,195,79,209,86,237,113,215,119,0,135,248,137,248,91,214,95,81,103,168,144,226,83,90,88,245,91,164,96,129,97,96,100,61,126,112,128,37,133,131,146,174,100,172,80,20,93,0,103,156,88,189,98,168,99,14,105,120,105,30,106,107,110,186,118,203,121,187,130,41,132,207,138,168,141,253,143,18,145,75,145,156,145,16,147,24,147,154,147,219,150,54,154,13,156,17,78,92,117,93,121,250,122,81,123,201,123,46,126,196,132,89,142,116,142,248,142,16,144,37,102,63,105,67,116,250,81,46,103,220,158,69,81,224,95,150,108,242,135,93,136,119,136,180,96,181,129,3,132,5,141,214,83,57,84,52,86,54,90,49,92,138,112,224,127,90,128,6,129,237,129,163,141,137,145,95,154,242,157,116,80,196,78,160,83,251,96,44,110,100,92,136,79,36,80,228,85,217,92,95,94,101,96,148,104,187,108,196,109,190,113,212,117,244,117,97,118,26,122,73,122,199,125,251,125,110,127,244,129,169,134,28,143,201,150,179,153,82,159,71,82,197,82,237,152,170,137,3,78,210,103,6,111,181,79,226,91,149,103,136,108,120,109,27,116,39,120,221,145,124,147,196,135,228,121,49,122,235,95,214,78,164,84,62,85,174,88,165,89,240,96,83,98,214,98,54,103,85,105,53,130,64,150,177,153,221,153,44,80,83,83,68,85,124,87,1,250,88,98,2,250,226,100,107,102,221,103,193,111,239,111,34,116,56,116,23,138,56,148,81,84,6,86,102,87,72,95,154,97,78,107,88,112,173,112,187,125,149,138,106,89,43,129,162,99,8,119,61,128,170,140,84,88,45,100,187,105,149,91,17,94,111,110,3,250,105,133,76,81,240,83,42,89,32,96,75,97,134,107,112,108,240,108,30,123,206,128,212,130,198,141,176,144,177,152,4,250,199,100,164,111,145,100,4,101,78,81,16,84,31,87,14,138,95,97,118,104,5,250,219,117,82,123,113,125,26,144,6,88,204,105,127,129,42,137,0,144,57,152,120,80,87,89,172,89,149,98,15,144,42,155,93,97,121,114,214,149,97,87,70,90,244,93,138,98,173,100,250,100,119,103,226,108,62,109,44,114,54,116,52,120,119,127,173,130,219,141,23,152,36,82,66,87,127,103,72,114,227,116,169,140,166,143,17,146,42,150,107,81,237,83,76,99,105,79,4,85,150,96,87,101,155,108,127,109,76,114,253,114,23,122,135,137,157,140,109,95,142,111,249,112,168,129,14,97,191,79,79,80,65,98,71,114,199,123,232,125,233,127,77,144,173,151,25,154,182,140,106,87,115,94,176,103,13,132,85,138,32,84,22,91,99,94,226,94,10,95,131,101,186,128,61,133,137,149,91,150,72,79,5,83,13,83,15,83,134,84,250,84,3,87,3,94,22,96,155,98,177,98,85,99,6,250,225,108,102,109,177,117,50,120,222,128,47,129,222,130,97,132,178,132,141,136,18,137,11,144,234,146,253,152,145,155,69,94,180,102,221,102,17,112,6,114,7,250,245,79,125,82,106,95,83,97,83,103,25,106,2,111,226,116,104,121,104,136,121,140,199,152,196,152,67,154,193,84,31,122,83,105,247,138,74,140,168,152,174,153,124,95,171,98,178,117,174,118,171,136,127,144,66,150,57,83,60,95,197,95,204,108,204,115,98,117,139,117,70,123,254,130,157,153,79,78,60,144,11,78,85,79,166,83,15,89,200,94,48,102,179,108,85,116,119,131,102,135,192,140,80,144,30,151,21,156,209,88,120,91,80,134,20,139,180,157,210,91,104,96,141,96,241,101,87,108,34,111,163,111,26,112,85,127,240,127,145,149,146,149,80,150,211,151,114,82,68,143,253,81,43,84,184,84,99,85,138,85,187,106,181,109,216,125,102,130,156,146,119,150,121,158,8,84,200,84,210,118,228,134,164,149,212,149,92,150,162,78,9,79,238,89,230,90,247,93,82,96,151,98,109,103,65,104,134,108,47,110,56,127,155,128,42,130,8,250,9,250,5,152,165,78,85,80,179,84,147,87,90,89,105,91,179,91,200,97,119,105,119,109,35,112,249,135,227,137,114,138,231,138,130,144,237,153,184,154,190,82,56,104,22,80,120,94,79,103,71,131,76,136,171,78,17,84,174,86,230,115,21,145,255,151,9,153,87,153,153,153,83,86,159,88,91,134,49,138,178,97,246,106,123,115,210,142,71,107,170,150,87,154,85,89,0,114,107,141,105,151,212,79,244,92,38,95,248,97,91,102,235,108,171,112,132,115,185,115,254,115,41,119,77,119,67,125,98,125,35,126,55,130,82,136,10,250,226,140,73,146,111,152,81,91,116,122,64,136,1,152,204,90,224,79,84,83,62,89,253,92,62,99,121,109,249,114,5,129,7,129,162,131,207,146,48,152,168,78,68,81,17,82,139,87,98,95,194,108,206,110,5,112,80,112,175,112,146,113,233,115,105,116,74,131,162,135,97,136,8,144,162,144,163,147,168,153,110,81,87,95,224,96,103,97,179,102,89,133,74,142,175,145,139,151,78,78,146,78,124,84,213,88,250,88,125,89,181,92,39,95,54,98,72,98,10,102,103,102,235,107,105,109,207,109,86,110,248,110,148,111,224,111,233,111,93,112,208,114,37,116,90,116,224,116,147,118,92,121,202,124,30,126,225,128,166,130,107,132,191,132,78,134,95,134,116,135,119,139,106,140,172,147,0,152,101,152,209,96,22,98,119,145,90,90,15,102,247,109,62,110,63,116,66,155,253,95,218,96,15,123,196,84,24,95,94,108,211,108,42,109,216,112,5,125,121,134,12,138,59,157,22,83,140,84,5,91,58,106,107,112,117,117,141,121,190,121,177,130,239,131,113,138,65,139,168,140,116,151,11,250,244,100,43,101,186,120,187,120,107,122,56,78,154,85,80,89,166,91,123,94,163,96,219,99,97,107,101,102,83,104,25,110,101,113,176,116,8,125,132,144,105,154,37,156,59,109,209,110,62,115,65,140,202,149,240,81,76,94,168,95,77,96,246,96,48,97,76,97,67,102,68,102,165,105,193,108,95,110,201,110,98,111,76,113,156,116,135,118,193,123,39,124,82,131,87,135,81,144,141,150,195,158,47,83,222,86,251,94,138,95,98,96,148,96,247,97,102,102,3,103,156,106,238,109,174,111,112,112,106,115,106,126,190,129,52,131,212,134,168,138,196,140,131,82,114,115,150,91,107,106,4,148,238,84,134,86,93,91,72,101,133,101,201,102,159,104,141,109,198,109,59,114,180,128,117,145,77,154,175,79,25,80,154,83,14,84,60,84,137,85,197,85,63,94,140,95,61,103,102,113,221,115,5,144,219,82,243,82,100,88,206,88,4,113,143,113,251,113,176,133,19,138,136,102,168,133,167,85,132,102,74,113,49,132,73,83,153,85,193,107,89,95,189,95,238,99,137,102,71,113,241,138,29,143,190,158,17,79,58,100,203,112,102,117,103,134,100,96,78,139,248,157,71,81,246,81,8,83,54,109,248,128,209,158,21,102,35,107,152,112,213,117,3,84,121,92,7,125,22,138,32,107,61,107,70,107,56,84,112,96,61,109,213,127,8,130,214,80,222,81,156,85,107,86,205,86,236,89,9,91,12,94,153,97,152,97,49,98,94,102,230,102,153,113,185,113,186,113,167,114,167,121,0,122,178,127,112,138,0,0,1,0,160,0,167,0,168,0,169,0,175,0,176,0,178,0,183,0,184,0,198,0,215,0,216,0,230,0,247,0,248,0,162,0,163,0,165,0,196,0,197,0,198,0,199,0,201,0,214,0,215,0,216,0,220,0,224,0,226,0,228,0,229,0,230,0,231,0,232,0,233,0,234,0,235,0,236,0,238,0,239,0,242,0,244,0,246,0,248,0,249,0,251,0,252,0,255,0,0,1,1,1,2,1,3,1,4,1,5,1,6,1,7,1,8,1,9,1,10,1,11,1,12,1,13,1,14,1,15,1,16,1,17,1,18,1,19,1,22,1,23,1,24,1,25,1,26,1,27,1,28,1,29,1,30,1,31,1,32,1,33,1,34,1,35,1,36,1,37,1,38,1,39,1,40,1,41,1,42,1,43,1,46,1,47,1,48,1,49,1,52,1,53,1,54,1,55,1,56,1,57,1,58,1,59,1,60,1,61,1,62,1,65,1,66,1,67,1,68,1,69,1,70,1,71,1,72,1,74,1,75,1,76,1,77,1,80,1,81,1,82,1,83,1,84,1,85,1,86,1,87,1,88,1,89,1,90,1,91,1,92,1,93,1,94,1,95,1,96,1,97,1,98,1,99,1,100,1,101,1,102,1,103,1,104,1,105,1,106,1,107,1,108,1,109,1,110,1,111,1,112,1,113,1,114,1,115,1,116,1,117,1,118,1,119,1,120,1,121,1,122,1,123,1,124,1,125,1,126,1,146,1,160,1,161,1,175,1,176,1,24,2,25,2,26,2,27,2,198,2,199,2,216,2,217,2,219,2,220,2,221,2,0,3,1,3,3,3,9,3,35,3,122,3,132,3,133,3,134,3,136,3,137,3,138,3,140,3,142,3,143,3,144,3,145,3,146,3,147,3,148,3,149,3,150,3,151,3,152,3,153,3,154,3,155,3,156,3,157,3,158,3,159,3,160,3,161,3,163,3,164,3,165,3,166,3,167,3,168,3,169,3,170,3,171,3,172,3,173,3,174,3,175,3,176,3,177,3,178,3,179,3,180,3,181,3,182,3,183,3,184,3,185,3,186,3,187,3,188,3,189,3,190,3,191,3,192,3,193,3,194,3,195,3,196,3,197,3,198,3,199,3,200,3,201,3,202,3,203,3,204,3,205,3,206,3,1,4,2,4,3,4,4,4,5,4,6,4,7,4,8,4,9,4,10,4,11,4,12,4,14,4,15,4,16,4,17,4,18,4,19,4,20,4,21,4,22,4,23,4,24,4,25,4,26,4,27,4,28,4,29,4,30,4,31,4,32,4,33,4,34,4,35,4,36,4,37,4,38,4,39,4,40,4,41,4,42,4,43,4,44,4,45,4,46,4,47,4,48,4,49,4,50,4,51,4,52,4,53,4,54,4,55,4,56,4,57,4,58,4,59,4,60,4,61,4,62,4,63,4,64,4,65,4,66,4,67,4,68,4,69,4,70,4,71,4,72,4,73,4,74,4,75,4,76,4,77,4,78,4,79,4,81,4,82,4,83,4,84,4,85,4,86,4,87,4,88,4,89,4,90,4,91,4,92,4,94,4,95,4,144,4,145,4,176,5,177,5,178,5,179,5,180,5,181,5,182,5,183,5,184,5,185,5,187,5,188,5,189,5,190,5,191,5,192,5,193,5,194,5,195,5,208,5,209,5,210,5,211,5,212,5,213,5,214,5,215,5,216,5,217,5,218,5,219,5,220,5,221,5,222,5,223,5,224,5,225,5,226,5,227,5,228,5,229,5,230,5,231,5,232,5,233,5,234,5,240,5,241,5,242,5,243,5,244,5,12,6,27,6,31,6,33,6,34,6,35,6,36,6,37,6,38,6,39,6,40,6,41,6,42,6,43,6,44,6,45,6,46,6,47,6,48,6,49,6,50,6,51,6,52,6,53,6,54,6,55,6,56,6,57,6,58,6,64,6,65,6,66,6,67,6,68,6,69,6,70,6,71,6,72,6,73,6,74,6,75,6,76,6,77,6,78,6,79,6,80,6,81,6,82,6,121,6,126,6,134,6,136,6,145,6,152,6,169,6,175,6,186,6,190,6,193,6,210,6,1,14,2,14,3,14,4,14,5,14,6,14,7,14,8,14,9,14,10,14,11,14,12,14,13,14,14,14,15,14,16,14,17,14,18,14,19,14,20,14,21,14,22,14,23,14,24,14,25,14,26,14,27,14,28,14,29,14,30,14,31,14,32,14,33,14,34,14,35,14,36,14,37,14,38,14,39,14,40,14,41,14,42,14,43,14,44,14,45,14,46,14,47,14,48,14,49,14,50,14,51,14,52,14,53,14,54,14,55,14,56,14,57,14,58,14,63,14,64,14,65,14,66,14,67,14,68,14,69,14,70,14,71,14,72,14,73,14,74,14,75,14,76,14,77,14,78,14,79,14,80,14,81,14,82,14,83,14,84,14,85,14,86,14,87,14,88,14,89,14,90,14,91,14,2,30,3,30,10,30,11,30,30,30,31,30,64,30,65,30,86,30,87,30,96,30,97,30,106,30,107,30,128,30,129,30,130,30,131,30,132,30,133,30,242,30,243,30,12,32,13,32,14,32,15,32,19,32,20,32,21,32,23,32,24,32,25,32,26,32,28,32,29,32,30,32,32,32,33,32,34,32,38,32,48,32,57,32,58,32,170,32,171,32,172,32,175,32,167,32,22,33,34,33,25,34,26,34,72,34,100,34,101,34,32,35,33,35,0,37,2,37,12,37,16,37,20,37,24,37,28,37,36,37,44,37,52,37,60,37,80,37,81,37,82,37,83,37,84,37,85,37,86,37,87,37,88,37,89,37,90,37,91,37,92,37,93,37,94,37,95,37,96,37,97,37,98,37,99,37,100,37,101,37,102,37,103,37,104,37,105,37,106,37,107,37,108,37,128,37,132,37,136,37,140,37,144,37,145,37,146,37,147,37,160,37,91,69,109,93,32,80,104,111,116,111,110,80,101,101,114,67,114,101,97,116,101,32,0,5,4,91,69,109,93,32,80,104,111,116,111,110,80,101,101,114,68,101,115,116,114,111,121,0,91,69,109,93,32,80,104,111,116,111,110,80,101,101,114,67,111,110,110,101,99,116,32,0,91,69,109,93,32,80,104,111,116,111,110,80,101,101,114,68,105,115,99,111,110,110,101,99,116,32,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,49,52,80,104,111,116,111,110,76,105,115,116,101,110,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,50,66,97,115,101,76,105,115,116,101,110,101,114,69,0,49,54,80,104,111,116,111,110,76,105,115,116,101,110,101,114,69,77,0,91,69,109,93,32,100,101,98,117,103,82,101,116,117,114,110,32,0,91,69,109,93,32,111,110,79,112,101,114,97,116,105,111,110,82,101,115,112,111,110,115,101,32,0,91,69,109,93,32,111,110,83,116,97,116,117,115,67,104,97,110,103,101,100,32,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,68,105,99,116,105,111,110,97,114,121,73,104,78,83,48,95,54,79,98,106,101,99,116,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,57,75,101,121,79,98,106,101,99,116,73,104,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,104,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,115,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,105,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,120,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,102,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,100,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,98,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,78,83,48,95,55,74,83,116,114,105,110,103,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,78,83,48,95,57,72,97,115,104,116,97,98,108,101,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,68,105,99,116,105,111,110,97,114,121,73,105,78,83,48,95,54,79,98,106,101,99,116,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,68,105,99,116,105,111,110,97,114,121,73,78,83,48,95,54,79,98,106,101,99,116,69,83,50,95,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,78,83,48,95,49,48,68,105,99,116,105,111,110,97,114,121,73,104,78,83,48,95,54,79,98,106,101,99,116,69,69,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,78,83,48,95,49,48,68,105,99,116,105,111,110,97,114,121,73,105,78,83,48,95,54,79,98,106,101,99,116,69,69,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,78,83,48,95,49,48,68,105,99,116,105,111,110,97,114,121,73,78,83,48,95,54,79,98,106,101,99,116,69,83,51,95,69,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,75,104,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,115,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,105,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,120,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,102,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,100,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,78,83,48,95,55,74,83,116,114,105,110,103,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,98,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,78,83,48,95,54,79,98,106,101,99,116,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,57,75,101,121,79,98,106,101,99,116,73,105,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,74,86,101,99,116,111,114,73,104,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,74,86,101,99,116,111,114,73,78,83,48,95,54,79,98,106,101,99,116,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,74,83,116,114,105,110,103,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,56,84,111,83,116,114,105,110,103,69,0,85,83,45,65,83,67,73,73,0,85,84,70,45,51,50,76,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,54,77,101,109,111,114,121,77,97,110,97,103,101,109,101,110,116,56,73,110,116,101,114,110,97,108,49,53,68,105,114,101,99,116,65,108,108,111,99,97,116,111,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,54,77,101,109,111,114,121,77,97,110,97,103,101,109,101,110,116,49,56,65,108,108,111,99,97,116,111,114,73,110,116,101,114,102,97,99,101,69,0,255,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,54,77,101,109,111,114,121,77,97,110,97,103,101,109,101,110,116,56,73,110,116,101,114,110,97,108,49,55,77,101,109,111,114,121,80,111,111,108,77,97,110,97,103,101,114,69,0,85,84,70,45,56,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,85,84,70,56,83,116,114,105,110,103,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,52,66,97,115,101,67,104,97,114,83,116,114,105,110,103,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,52,66,97,115,101,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,65,78,83,73,83,116,114,105,110,103,69,0,99,104,101,99,107,84,121,112,101,0,37,89,45,37,109,45,37,100,32,37,72,58,37,77,58,37,83,0,37,115,44,37,48,51,100,0,37,115,0,99,100,105,111,117,120,88,101,69,102,103,71,112,110,37,115,83,104,108,76,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,57,72,97,115,104,116,97,98,108,101,69,0,116,111,83,116,114,105,110,103,72,101,108,112,101,114,0,112,97,121,108,111,97,100,84,121,112,101,84,111,83,116,114,105,110,103,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,54,79,98,106,101,99,116,69,0,74,83,116,114,105,110,103,0,72,97,115,104,116,97,98,108,101,0,68,105,99,116,105,111,110,97,114,121,0,79,98,106,101,99,116,0,67,117,115,116,111,109,84,121,112,101,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,52,68,105,99,116,105,111,110,97,114,121,66,97,115,101,56,84,121,112,101,73,110,102,111,69,0,99,111,109,112,97,114,101,72,101,108,112,101,114,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,52,68,105,99,116,105,111,110,97,114,121,66,97,115,101,69,0,112,111,112,79,98,106,101,99,116,65,114,114,97,121,0,112,111,112,65,114,114,97,121,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,72,101,108,112,101,114,115,50,54,68,101,83,101,114,105,97,108,105,122,101,114,73,109,112,108,101,109,101,110,116,97,116,105,111,110,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,72,101,108,112,101,114,115,50,52,83,101,114,105,97,108,105,122,101,114,73,109,112,108,101,109,101,110,116,97,116,105,111,110,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,48,83,101,114,105,97,108,105,122,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,49,49,86,97,108,117,101,79,98,106,101,99,116,73,80,104,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,49,51,83,111,99,107,101,116,67,111,110,110,101,99,116,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,49,51,80,104,111,116,111,110,67,111,110,110,101,99,116,69,0,1,6,123,32,77,111,100,117,108,101,91,39,119,101,98,115,111,99,107,101,116,39,93,46,115,117,98,112,114,111,116,111,99,111,108,32,61,32,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,40,36,48,41,59,32,77,111,100,117,108,101,91,39,119,101,98,115,111,99,107,101,116,39,93,46,117,114,108,32,61,32,80,111,105,110,116,101,114,95,115,116,114,105,110,103,105,102,121,40,36,49,41,59,32,125,0,105,110,105,116,67,97,108,108,98,97,99,107,0,100,101,114,105,118,101,83,104,97,114,101,100,75,101,121,0,255,255,255,255,255,255,255,255,201,15,218,162,33,104,194,52,196,198,98,139,128,220,28,209,41,2,78,8,138,103,204,116,2,11,190,166,59,19,155,34,81,74,8,121,142,52,4,221,239,149,25,179,205,58,67,27,48,43,10,109,242,95,20,55,79,225,53,109,109,81,194,69,228,133,181,118,98,94,126,198,244,76,66,233,166,58,54,32,255,255,255,255,255,255,255,255,99,111,110,116,101,120,116,32,33,61,32,40,83,72,65,50,53,54,95,67,84,88,42,41,48,0,115,114,99,47,73,110,116,101,114,110,97,108,47,69,110,99,114,121,112,116,105,111,110,47,115,104,97,50,46,99,0,69,71,83,72,65,50,53,54,95,69,110,100,0,69,71,83,72,65,50,53,54,95,70,105,110,97,108,0,99,111,110,116,101,120,116,32,33,61,32,40,83,72,65,50,53,54,95,67,84,88,42,41,48,32,38,38,32,100,97,116,97,32,33,61,32,40,115,104,97,50,95,98,121,116,101,42,41,48,0,69,71,83,72,65,50,53,54,95,85,112,100,97,116,101,0,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,1,117,112,100,97,116,101,82,111,117,110,100,84,114,105,112,84,105,109,101,65,110,100,86,97,114,105,97,110,99,101,0,100,101,115,101,114,105,97,108,105,122,101,79,112,101,114,97,116,105,111,110,82,101,115,112,111,110,115,101,0,100,101,115,101,114,105,97,108,105,122,101,79,112,101,114,97,116,105,111,110,0,115,101,114,105,97,108,105,122,101,79,112,101,114,97,116,105,111,110,0,111,110,68,105,115,99,111,110,110,101,99,116,101,100,0,114,101,115,101,116,0,112,105,110,103,83,101,114,118,101,114,0,111,112,67,117,115,116,111,109,0,115,101,114,118,105,99,101,0,99,111,110,110,101,99,116,0,76,105,116,101,0,111,110,67,111,110,110,101,99,116,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,56,80,101,101,114,66,97,115,101,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,50,49,80,104,111,116,111,110,67,111,110,110,101,99,116,76,105,115,116,101,110,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,57,79,112,101,114,97,116,105,111,110,69,0,4,1,9,2,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,56,84,67,111,110,110,101,99,116,69,0,115,101,110,100,0,118,101,114,105,102,121,73,110,100,101,120,0,102,101,116,99,104,83,101,114,118,101,114,84,105,109,101,115,116,97,109,112,0,100,105,115,112,97,116,99,104,73,110,99,111,109,105,110,103,67,111,109,109,97,110,100,115,0,115,101,110,100,65,99,107,115,79,110,108,121,0,115,101,110,100,79,117,116,103,111,105,110,103,67,111,109,109,97,110,100,115,0,115,101,114,118,105,99,101,66,97,115,105,99,0,111,110,82,101,99,101,105,118,101,68,97,116,97,0,78,57,69,120,105,116,71,97,109,101,115,54,67,111,109,109,111,110,55,74,86,101,99,116,111,114,73,78,83,95,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,57,79,112,101,114,97,116,105,111,110,69,69,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,53,84,80,101,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,49,57,84,114,97,102,102,105,99,83,116,97,116,115,67,111,117,110,116,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,49,50,84,114,97,102,102,105,99,83,116,97,116,115,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,56,73,110,116,101,114,110,97,108,50,56,84,114,97,102,102,105,99,83,116,97,116,115,71,97,109,101,76,101,118,101,108,67,111,117,110,116,101,114,69,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,50,49,84,114,97,102,102,105,99,83,116,97,116,115,71,97,109,101,76,101,118,101,108,69,0,58,47,47,0,100,105,115,99,111,110,110,101,99,116,0,78,57,69,120,105,116,71,97,109,101,115,54,80,104,111,116,111,110,49,48,80,104,111,116,111,110,80,101,101,114,69,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,3,4,5,6,7,8,9,255,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,0,1,2,4,7,3,6,5,0,17,0,10,0,17,17,17,0,0,0,0,5,0,0,0,0,0,0,9,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,15,10,17,17,17,3,10,7,0,1,19,9,11,11,0,0,9,6,11,0,0,11,0,6,17,0,0,0,17,17,17,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,17,0,10,10,17,17,17,0,10,0,0,2,0,9,11,0,0,0,9,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,14,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,4,13,0,0,0,0,9,14,0,0,0,0,0,14,0,0,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,0,0,0,0,0,0,0,0,0,0,0,15,0,0,0,0,15,0,0,0,0,9,16,0,0,0,0,0,16,0,0,16,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18,0,0,0,18,18,18,0,0,0,0,0,0,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,11,0,0,0,0,0,0,0,0,0,0,0,10,0,0,0,0,10,0,0,0,0,9,11,0,0,0,0,0,11,0,0,11,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,0,0,0,0,0,0,0,12,0,0,0,0,12,0,0,0,0,9,12,0,0,0,0,0,12,0,0,12,0,0,45,43,32,32,32,48,88,48,120,0,40,110,117,108,108,41,0,45,48,88,43,48,88,32,48,88,45,48,120,43,48,120,32,48,120,0,105,110,102,0,73,78,70,0,78,65,78,0,48,49,50,51,52,53,54,55,56,57,65,66,67,68,69,70,84,33,34,25,13,1,2,3,17,75,28,12,16,4,11,29,18,30,39,104,110,111,112,113,98,32,5,6,15,19,20,21,26,8,22,7,40,36,23,24,9,10,14,27,31,37,35,131,130,125,38,42,43,60,61,62,63,67,71,74,77,88,89,90], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+138528);
/* memory initializer */ allocate([91,92,93,94,95,96,97,99,100,101,102,103,105,106,107,108,114,115,116,121,122,123,124,0,73,108,108,101,103,97,108,32,98,121,116,101,32,115,101,113,117,101,110,99,101,0,68,111,109,97,105,110,32,101,114,114,111,114,0,82,101,115,117,108,116,32,110,111,116,32,114,101,112,114,101,115,101,110,116,97,98,108,101,0,78,111,116,32,97,32,116,116,121,0,80,101,114,109,105,115,115,105,111,110,32,100,101,110,105,101,100,0,79,112,101,114,97,116,105,111,110,32,110,111,116,32,112,101,114,109,105,116,116,101,100,0,78,111,32,115,117,99,104,32,102,105,108,101,32,111,114,32,100,105,114,101,99,116,111,114,121,0,78,111,32,115,117,99,104,32,112,114,111,99,101,115,115,0,70,105,108,101,32,101,120,105,115,116,115,0,86,97,108,117,101,32,116,111,111,32,108,97,114,103,101,32,102,111,114,32,100,97,116,97,32,116,121,112,101,0,78,111,32,115,112,97,99,101,32,108,101,102,116,32,111,110,32,100,101,118,105,99,101,0,79,117,116,32,111,102,32,109,101,109,111,114,121,0,82,101,115,111,117,114,99,101,32,98,117,115,121,0,73,110,116,101,114,114,117,112,116,101,100,32,115,121,115,116,101,109,32,99,97,108,108,0,82,101,115,111,117,114,99,101,32,116,101,109,112,111,114,97,114,105,108,121,32,117,110,97,118,97,105,108,97,98,108,101,0,73,110,118,97,108,105,100,32,115,101,101,107,0,67,114,111,115,115,45,100,101,118,105,99,101,32,108,105,110,107,0,82,101,97,100,45,111,110,108,121,32,102,105,108,101,32,115,121,115,116,101,109,0,68,105,114,101,99,116,111,114,121,32,110,111,116,32,101,109,112,116,121,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,112,101,101,114,0,79,112,101,114,97,116,105,111,110,32,116,105,109,101,100,32,111,117,116,0,67,111,110,110,101,99,116,105,111,110,32,114,101,102,117,115,101,100,0,72,111,115,116,32,105,115,32,100,111,119,110,0,72,111,115,116,32,105,115,32,117,110,114,101,97,99,104,97,98,108,101,0,65,100,100,114,101,115,115,32,105,110,32,117,115,101,0,66,114,111,107,101,110,32,112,105,112,101,0,73,47,79,32,101,114,114,111,114,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,32,111,114,32,97,100,100,114,101,115,115,0,66,108,111,99,107,32,100,101,118,105,99,101,32,114,101,113,117,105,114,101,100,0,78,111,32,115,117,99,104,32,100,101,118,105,99,101,0,78,111,116,32,97,32,100,105,114,101,99,116,111,114,121,0,73,115,32,97,32,100,105,114,101,99,116,111,114,121,0,84,101,120,116,32,102,105,108,101,32,98,117,115,121,0,69,120,101,99,32,102,111,114,109,97,116,32,101,114,114,111,114,0,73,110,118,97,108,105,100,32,97,114,103,117,109,101,110,116,0,65,114,103,117,109,101,110,116,32,108,105,115,116,32,116,111,111,32,108,111,110,103,0,83,121,109,98,111,108,105,99,32,108,105,110,107,32,108,111,111,112,0,70,105,108,101,110,97,109,101,32,116,111,111,32,108,111,110,103,0,84,111,111,32,109,97,110,121,32,111,112,101,110,32,102,105,108,101,115,32,105,110,32,115,121,115,116,101,109,0,78,111,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,115,32,97,118,97,105,108,97,98,108,101,0,66,97,100,32,102,105,108,101,32,100,101,115,99,114,105,112,116,111,114,0,78,111,32,99,104,105,108,100,32,112,114,111,99,101,115,115,0,66,97,100,32,97,100,100,114,101,115,115,0,70,105,108,101,32,116,111,111,32,108,97,114,103,101,0,84,111,111,32,109,97,110,121,32,108,105,110,107,115,0,78,111,32,108,111,99,107,115,32,97,118,97,105,108,97,98,108,101,0,82,101,115,111,117,114,99,101,32,100,101,97,100,108,111,99,107,32,119,111,117,108,100,32,111,99,99,117,114,0,83,116,97,116,101,32,110,111,116,32,114,101,99,111,118,101,114,97,98,108,101,0,80,114,101,118,105,111,117,115,32,111,119,110,101,114,32,100,105,101,100,0,79,112,101,114,97,116,105,111,110,32,99,97,110,99,101,108,101,100,0,70,117,110,99,116,105,111,110,32,110,111,116,32,105,109,112,108,101,109,101,110,116,101,100,0,78,111,32,109,101,115,115,97,103,101,32,111,102,32,100,101,115,105,114,101,100,32,116,121,112,101,0,73,100,101,110,116,105,102,105,101,114,32,114,101,109,111,118,101,100,0,68,101,118,105,99,101,32,110,111,116,32,97,32,115,116,114,101,97,109,0,78,111,32,100,97,116,97,32,97,118,97,105,108,97,98,108,101,0,68,101,118,105,99,101,32,116,105,109,101,111,117,116,0,79,117,116,32,111,102,32,115,116,114,101,97,109,115,32,114,101,115,111,117,114,99,101,115,0,76,105,110,107,32,104,97,115,32,98,101,101,110,32,115,101,118,101,114,101,100,0,80,114,111,116,111,99,111,108,32,101,114,114,111,114,0,66,97,100,32,109,101,115,115,97,103,101,0,70,105,108,101,32,100,101,115,99,114,105,112,116,111,114,32,105,110,32,98,97,100,32,115,116,97,116,101,0,78,111,116,32,97,32,115,111,99,107,101,116,0,68,101,115,116,105,110,97,116,105,111,110,32,97,100,100,114,101,115,115,32,114,101,113,117,105,114,101,100,0,77,101,115,115,97,103,101,32,116,111,111,32,108,97,114,103,101,0,80,114,111,116,111,99,111,108,32,119,114,111,110,103,32,116,121,112,101,32,102,111,114,32,115,111,99,107,101,116,0,80,114,111,116,111,99,111,108,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,80,114,111,116,111,99,111,108,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,83,111,99,107,101,116,32,116,121,112,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,111,116,32,115,117,112,112,111,114,116,101,100,0,80,114,111,116,111,99,111,108,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,65,100,100,114,101,115,115,32,102,97,109,105,108,121,32,110,111,116,32,115,117,112,112,111,114,116,101,100,32,98,121,32,112,114,111,116,111,99,111,108,0,65,100,100,114,101,115,115,32,110,111,116,32,97,118,97,105,108,97,98,108,101,0,78,101,116,119,111,114,107,32,105,115,32,100,111,119,110,0,78,101,116,119,111,114,107,32,117,110,114,101,97,99,104,97,98,108,101,0,67,111,110,110,101,99,116,105,111,110,32,114,101,115,101,116,32,98,121,32,110,101,116,119,111,114,107,0,67,111,110,110,101,99,116,105,111,110,32,97,98,111,114,116,101,100,0,78,111,32,98,117,102,102,101,114,32,115,112,97,99,101,32,97,118,97,105,108,97,98,108,101,0,83,111,99,107,101,116,32,105,115,32,99,111,110,110,101,99,116,101,100,0,83,111,99,107,101,116,32,110,111,116,32,99,111,110,110,101,99,116,101,100,0,67,97,110,110,111,116,32,115,101,110,100,32,97,102,116,101,114,32,115,111,99,107,101,116,32,115,104,117,116,100,111,119,110,0,79,112,101,114,97,116,105,111,110,32,97,108,114,101,97,100,121,32,105,110,32,112,114,111,103,114,101,115,115,0,79,112,101,114,97,116,105,111,110,32,105,110,32,112,114,111,103,114,101,115,115,0,83,116,97,108,101,32,102,105,108,101,32,104,97,110,100,108,101,0,82,101,109,111,116,101,32,73,47,79,32,101,114,114,111,114,0,81,117,111,116,97,32,101,120,99,101,101,100,101,100,0,78,111,32,109,101,100,105,117,109,32,102,111,117,110,100,0,87,114,111,110,103,32,109,101,100,105,117,109,32,116,121,112,101,0,78,111,32,101,114,114,111,114,32,105,110,102,111,114,109,97,116,105,111,110,0,0,105,110,102,105,110,105,116,121,0,110,97,110,0,117,116,102,56,0,99,104,97,114,0,0,200,119,99,104,97,114,116,0,0,198,117,99,115,50,0,117,99,115,50,98,101,0,0,196,117,99,115,50,108,101,0,0,197,117,116,102,49,54,0,117,116,102,49,54,98,101,0,0,194,117,116,102,49,54,108,101,0,0,193,117,99,115,52,0,117,99,115,52,98,101,0,117,116,102,51,50,0,117,116,102,51,50,98,101,0,0,192,117,99,115,52,108,101,0,117,116,102,51,50,108,101,0,0,195,97,115,99,105,105,0,117,115,97,115,99,105,105,0,105,115,111,54,52,54,0,105,115,111,54,52,54,117,115,0,0,199,101,117,99,106,112,0,0,208,115,104,105,102,116,106,105,115,0,115,106,105,115,0,0,209,103,98,49,56,48,51,48,0,0,216,103,98,107,0,0,217,103,98,50,51,49,50,0,0,218,98,105,103,53,0,98,105,103,102,105,118,101,0,99,112,57,53,48,0,98,105,103,53,104,107,115,99,115,0,0,224,101,117,99,107,114,0,107,115,99,53,54,48,49,0,107,115,120,49,48,48,49,0,99,112,57,52,57,0,0,232,105,115,111,56,56,53,57,49,0,108,97,116,105,110,49,0,0,128,105,115,111,56,56,53,57,50,0,0,32,0,216,16,203,26,0,164,17,8,0,0,28,82,200,34,160,0,64,138,40,0,220,48,11,27,0,168,33,8,44,0,32,98,8,35,161,212,82,202,40,123,0,0,0,13,0,148,129,3,0,62,0,128,4,0,74,0,0,0,16,66,180,17,7,0,0,220,1,0,0,127,84,2,192,37,0,0,144,8,0,124,0,0,64,13,0,152,145,3,0,63,0,144,4,0,75,0,0,64,16,67,184,33,7,0,0,224,1,0,0,128,88,2,0,38,0,0,160,136,44,105,115,111,56,56,53,57,51,0,0,32,0,88,17,11,0,0,4,64,5,0,0,120,81,136,19,96,0,16,128,40,0,92,1,0,0,0,0,80,5,0,0,124,97,200,19,97,0,16,192,40,0,0,0,64,0,0,240,160,3,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,64,1,0,0,76,0,0,0,0,0,76,50,8,0,0,0,0,64,0,0,244,176,3,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,68,1,0,0,77,0,0,0,0,0,80,66,136,44,105,115,111,56,56,53,57,52,0,0,32,0,216,64,70,31,0,96,113,6,0,0,28,66,132,20,141,0,64,10,0,0,220,48,139,31,0,100,129,6,44,0,32,82,196,20,142,204,81,10,29,50,0,0,0,0,0,0,0,0,23,62,0,128,4,0,70,0,0,128,22,66,188,81,135,24,0,0,0,0,0,0,100,2,0,0,0,60,18,9,0,51,0,0,0,0,0,0,0,64,23,63,0,144,4,0,71,0,0,192,22,67,192,97,199,24,0,0,0,0,0,0,104,2,0,0,0,64,34,137,44,105,115,111,56,56,53,57,53,0,0,32,0,12,68,80,65,6,29,132,80,66,10,45,196,80,67,14,1,240,16,68,17,73,52,17,69,21,89,116,17,70,25,105,180,17,71,29,121,244,17,72,33,137,52,18,73,37,153,116,18,74,41,169,180,18,75,45,185,244,18,76,49,201,52,19,77,53,217,116,19,78,57,233,180,19,79,61,249,244,19,80,65,9,53,20,81,69,25,117,20,82,73,41,181,20,83,77,57,245,20,84,87,70,37,213,84,84,85,101,213,85,88,101,165,213,86,92,13,208,149,87,105,115,111,56,56,53,57,54,0,0,32,0,4,16,64,0,0,4,16,64,0,1,4,16,64,0,148,1,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,101,1,4,16,128,101,1,92,134,89,102,154,109,198,89,103,158,125,6,90,104,162,141,70,90,105,166,157,134,90,106,170,173,198,90,107,174,189,6,91,0,1,4,16,64,0,177,201,54,27,109,181,217,118,27,110,185,233,182,27,111,189,249,246,27,112,193,9,55,92,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,105,115,111,56,56,53,57,55,0,0,32,0,20,105,36,0,84,86,9,0,0,0,0,176,11,0,0,0,16,192,144,0,0,0,0,0,188,244,226,11,0,191,0,19,12,0,194,0,48,12,49,197,24,115,12,50,201,40,179,12,51,205,56,243,12,52,209,72,51,13,53,213,88,19,192,53,216,100,163,205,54,220,116,227,205,55,224,132,35,206,56,228,148,99,206,57,232,164,163,206,58,236,180,227,206,59,240,196,35,207,60,244,212,99,207,61,248,228,163,207,62,252,244,227,207,63,0,5,36,80,0,105,115,111,56,56,53,57,56,0,0,32,0,4,0,0,0,0,0,0,0,0,0,0,192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,0,0,0,0,0,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,64,0,1,4,16,0,145,116,213,101,215,93,120,229,165,215,94,124,245,229,215,95,128,5,38,216,96,132,21,102,216,97,136,37,166,216,98,140,53,230,88,0,1,252,8,100,0,105,115,111,56,56,53,57,57,0,0,80,78,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,120,81,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,79,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,124,97,8,0,105,115,111,56,56,53,57,49,48,0,0,32,0,216,64,132,20,90,96,33,6,0,103,8,113,72,35,164,0,16,201,28,0,220,80,196,20,91,100,49,6,0,104,12,129,136,35,165,12,41,9,29,50,0,0,0,0,0,0,0,0,23,62,0,128,4,0,70,0,0,0,0,0,188,81,7,0,0,0,0,192,35,0,100,2,0,0,0,0,0,0,0,51,0,0,0,0,0,0,0,64,23,63,0,144,4,0,71,0,0,0,0,0,192,97,7,0,0,0,0,0,36,0,104,2,0,0,0,0,0,0,25,105,115,111,56,56,53,57,49,49,0,116,105,115,54,50,48,0,0,32,0,64,23,157,116,211,81,87,157,117,215,97,151,157,118,219,113,215,157,119,223,129,23,158,120,227,145,87,158,121,231,161,151,158,122,235,177,215,158,123,239,193,23,159,124,243,209,87,159,125,247,225,151,159,126,251,241,215,159,127,255,1,24,160,128,3,18,88,160,129,7,34,152,96,0,1,4,16,128,130,11,50,216,160,131,15,66,24,161,132,19,82,88,161,133,23,98,152,161,134,27,114,216,161,135,31,130,24,162,136,35,146,88,162,137,1,4,16,64,0,105,115,111,56,56,53,57,49,51,0,0,32,0,36,9,0,0,0,40,9,0,0,13,0,208,7,0,0,0,0,192,2,0,0,0,0,0,72,2,0,0,0,16,0,224,7,0,0,0,0,128,3,54,112,33,3,14,0,0,128,4,17,62,0,0,138,17,82,136,161,197,25,135,180,241,6,0,117,0,0,0,0,153,172,17,72,36,0,136,66,10,0,55,116,49,67,14,0,0,144,68,17,63,0,16,202,17,83,140,177,5,26,136,184,1,7,0,118,0,0,0,0,154,176,33,136,36,0,140,82,138,145,105,115,111,56,56,53,57,49,52,0,0,32,0,156,136,34,0,60,244,144,34,0,53,2,112,163,138,59,2,0,192,39,43,178,8,69,20,45,186,8,192,139,54,194,136,99,140,60,230,168,163,140,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,155,0,0,0,0,0,0,0,192,140,0,0,0,0,0,0,0,208,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,156,0,0,0,0,0,0,0,0,141,0,0,0,0,0,0,0,224,9,0,105,115,111,56,56,53,57,49,53,0,108,97,116,105,110,57,0,0,36,84,2,112,8,0,136,0,0,0,0,0,0,0,0,0,0,0,0,0,0,164,0,0,0,0,165,0,0,0,0,121,232,241,9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,105,115,111,56,56,53,57,49,54,0,0,32,0,216,112,195,26,84,42,121,8,0,136,0,176,10,0,160,0,16,138,40,0,0,224,3,27,164,36,9,0,0,165,252,192,10,0,121,232,241,201,40,0,0,0,0,13,0,224,0,0,0,0,0,0,0,0,0,0,0,0,0,66,180,1,0,0,0,220,1,64,32,151,0,0,0,0,0,32,209,10,0,0,0,0,64,13,0,228,0,0,0,0,0,0,0,0,0,0,0,0,0,67,184,1,0,0,0,224,1,128,32,152,0,0,0,0,0,36,225,10,0,99,112,49,50,53,48,0,119,105,110,100,111,119,115,49,50,53,48,0,0,0,84,6,112,100,0,74,58,185,36,147,1,60,121,8,148,129,44,66,10,40,1,20,105,36,146,73,54,25,164,144,1,96,137,72,148,130,48,82,74,40,0,192,18,203,26,0,216,0,0,0,0,0,80,8,0,0,0,0,128,40,0,0,48,11,27,0,0,0,0,0,0,220,96,8,0,105,212,162,198,40,123,0,0,0,13,0,148,129,3,0,62,0,128,4,0,74,0,0,0,16,66,180,17,7,0,0,220,1,0,0,127,84,2,192,37,0,0,144,8,0,124,0,0,64,13,0,152,145,3,0,63,0,144,4,0,75,0,0,64,16,67,184,33,7,0,0,224,1,0,0,128,88,2,0,38,0,0,160,136,44,99,112,49,50,53,49,0,119,105,110,100,111,119,115,49,50,53,49,0,0,0,4,21,116,228,84,74,58,185,36,147,84,62,185,16,148,12,57,212,16,68,82,21,105,36,146,73,54,25,164,144,1,96,153,85,148,90,113,181,149,87,0,60,212,149,66,0,124,5,0,0,3,1,96,16,0,0,0,0,64,66,0,0,128,144,85,96,1,0,0,0,81,93,73,21,0,88,29,84,213,85,17,73,52,17,69,21,89,116,17,70,25,105,180,17,71,29,121,244,17,72,33,137,52,18,73,37,153,116,18,74,41,169,180,18,75,45,185,244,18,76,49,201,52,19,77,53,217,116,19,78,57,233,180,19,79,61,249,244,19,80,65,9,53,20,81,69,25,117,20,82,73,41,181,20,83,77,57,245,20,84,99,112,49,50,53,50,0,119,105,110,100,111,119,115,49,50,53,50,0,0,0,84,6,112,164,41,74,58,185,36,147,175,60,121,8,148,121,4,64,74,0,1,20,105,36,146,73,54,25,164,144,180,96,137,72,148,122,4,80,202,39,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,99,112,49,50,53,51,0,119,105,110,100,111,119,115,49,50,53,51,0,0,0,84,6,112,164,41,74,58,185,36,147,1,60,25,0,148,1,4,16,64,0,1,20,105,36,146,73,54,25,164,144,1,96,25,64,148,1,4,16,64,0,0,244,226,11,0,0,0,0,0,0,0,0,16,0,0,0,0,0,192,144,0,0,0,0,0,188,0,0,0,0,191,0,19,12,0,194,0,48,12,49,197,24,115,12,50,201,40,179,12,51,205,56,243,12,52,209,72,51,13,53,213,88,19,192,53,216,100,163,205,54,220,116,227,205,55,224,132,35,206,56,228,148,99,206,57,232,164,163,206,58,236,180,227,206,59,240,196,35,207,60,244,212,99,207,61,248,228,163,207,62,252,244,227,207,63,0,5,36,80,0,99,112,49,50,53,52,0,119,105,110,100,111,119,115,49,50,53,52,0,0,0,84,6,112,164,41,74,58,185,36,147,175,60,121,8,148,121,4,16,64,0,1,20,105,36,146,73,54,25,164,144,180,96,137,72,148,122,4,16,192,39,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,78,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,120,81,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,79,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,124,97,8,0,99,112,49,50,53,53,0,119,105,110,100,111,119,115,49,50,53,53,0,0,0,84,6,112,164,41,74,58,185,36,147,175,60,25,0,148,1,4,16,64,0,1,20,105,36,146,73,54,25,164,144,180,96,25,64,148,1,4,16,64,0,0,0,0,0,0,82,2,0,0,0,0,0,192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,240,0,0,0,0,0,0,0,97,137,53,22,89,101,153,117,22,90,105,169,21,192,90,108,181,229,214,91,112,197,37,215,92,143,65,22,153,100,147,5,16,64,0,1,4,16,64,0,116,213,101,215,93,120,229,165,215,94,124,245,229,215,95,128,5,38,216,96,132,21,102,216,97,136,37,166,216,98,140,53,230,88,0,1,252,8,100,0,99,112,49,50,53,54,0,119,105,110,100,111,119,115,49,50,53,54,0,0,0,84,22,119,164,41,74,58,185,36,147,175,60,73,28,148,121,24,151,220,113,203,21,105,36,146,73,54,25,164,144,202,97,137,92,148,122,244,232,35,115,0,80,6,0,0,0,0,0,0,0,0,0,208,28,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,80,25,0,0,0,0,128,101,206,93,134,89,102,154,109,198,89,103,158,125,6,90,104,162,141,70,90,105,166,157,134,90,106,170,173,198,26,0,173,185,246,26,108,177,201,54,27,109,0,212,6,128,109,183,225,150,27,0,0,0,0,0,0,186,237,6,0,0,188,245,230,219,111,0,0,23,28,0,194,1,48,28,0,0,252,8,228,115,99,112,49,50,53,55,0,119,105,110,100,111,119,115,49,50,53,55,0,0,0,84,6,112,100,0,74,58,185,36,147,1,60,25,0,148,1,16,0,139,2,1,20,105,36,146,73,54,25,164,144,1,96,25,64,148,1,24,48,75,0,0,4,0,0,0,0,4,0,0,0,13,0,208,7,0,0,0,0,192,2,0,0,0,0,0,0,0,0,0,0,16,0,224,7,0,0,0,0,128,3,54,112,33,3,14,0,0,128,4,17,62,0,0,138,17,82,136,161,197,25,135,180,241,6,0,117,0,0,0,0,153,172,17,72,36,0,136,66,10,0,55,116,49,67,14,0,0,144,68,17,63,0,16,202,17,83,140,177,5,26,136,184,1,7,0,118,0,0,0,0,154,176,33,136,36,0,140,82,138,44,99,112,49,50,53,56,0,119,105,110,100,111,119,115,49,50,53,56,0,0,0,84,6,112,164,41,74,58,185,36,147,175,60,25,0,148,121,4,16,64,0,1,20,105,36,146,73,54,25,164,144,180,96,25,64,148,122,4,16,192,39,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,13,0,0,0,0,0,0,0,0,0,0,182,0,0,0,0,66,0,144,11,0,0,156,2,0,0,0,0,0,0,0,0,164,130,11,0,0,0,0,64,13,0,0,0,0,0,0,0,0,0,0,183,0,0,0,0,67,0,160,11,0,0,160,2,0,0,0,0,0,0,0,0,168,50,37,0,107,111,105,56,114,0,0,0,96,134,41,230,152,100,150,105,230,153,104,166,169,38,162,137,42,186,40,163,141,58,250,168,151,144,102,169,229,150,92,118,41,192,151,7,32,144,192,3,107,178,217,102,84,110,190,9,103,156,114,206,73,103,157,118,222,137,103,158,122,238,201,231,64,125,250,249,39,160,129,10,58,40,161,133,26,122,104,1,79,197,36,211,81,53,217,84,20,77,70,229,164,211,78,60,245,228,211,79,64,65,21,148,80,67,17,117,211,76,77,49,133,83,82,78,41,133,212,82,47,69,36,209,73,21,89,84,18,69,38,101,164,209,70,28,117,228,209,71,32,193,20,146,72,35,145,116,209,68,45,177,132,81,74,46,169,132,210,74,107,111,105,56,117,0,0,0,96,134,41,230,152,100,150,105,230,153,104,166,169,38,162,137,42,186,40,163,141,58,250,168,151,144,102,169,229,150,92,118,41,192,151,7,32,144,192,3,107,178,217,102,84,84,189,105,213,85,114,206,73,103,157,118,130,133,103,158,122,238,201,231,64,6,249,137,80,66,129,10,58,40,161,133,126,117,104,1,79,197,36,211,81,53,217,84,20,77,70,229,164,211,78,60,245,228,211,79,64,65,21,148,80,67,17,117,211,76,77,49,133,83,82,78,41,133,212,82,47,69,36,209,73,21,89,84,18,69,38,101,164,209,70,28,117,228,209,71,32,193,20,146,72,35,145,116,209,68,45,177,132,81,74,46,169,132,210,74,99,112,52,51,55,0,0,0,23,192,64,130,7,31,116,0,130,8,37,152,48,66,10,40,156,64,65,5,24,56,176,192,10,44,168,240,130,11,49,100,192,65,4,18,76,96,165,41,145,70,26,105,164,145,70,26,105,164,145,70,26,105,164,145,70,26,105,164,141,58,250,104,152,103,242,217,103,156,112,250,201,166,156,120,222,105,231,152,100,166,137,166,153,96,170,153,167,158,117,190,73,104,160,123,174,121,168,160,131,254,9,40,157,115,182,233,166,161,133,150,41,166,162,137,46,202,40,162,229,68,138,12,61,215,220,19,41,62,218,52,211,13,58,145,234,147,78,164,145,70,218,37,151,94,126,249,192,150,7,100,153,128,150,145,34,0,169,0,99,112,56,53,48,0,0,0,23,192,64,130,7,31,116,0,130,8,37,152,48,66,10,40,156,64,65,5,24,56,176,192,10,44,168,240,130,11,49,100,192,1,4,18,52,192,128,41,145,70,26,105,164,145,70,26,105,164,145,70,26,105,164,145,70,26,105,164,141,58,250,104,152,103,70,26,105,164,5,248,201,166,156,120,70,48,193,152,100,166,137,166,153,96,170,25,105,164,117,190,73,104,160,123,174,121,104,164,145,70,26,105,164,145,126,17,105,164,145,150,41,166,162,137,70,26,41,162,145,70,26,105,164,145,70,26,105,164,145,70,26,105,164,145,70,106,64,164,145,70,74,100,164,145,14,240,128,2,7,16,144,64,164,145,34,0,169,0,0,8,4,4,8,4,4,8,2,4,0,195,138,204,132,195,138,204,140,195,170,204,132,195,170,204,140,0,76,67,95,65,76,76,0,76,67,95,67,84,89,80,69,0,0,0,0,76,67,95,78,85,77,69,82,73,67,0,0,76,67,95,84,73,77,69,0,0,0,0,0,76,67,95,67,79,76,76,65,84,69,0,0,76,67,95,77,79,78,69,84,65,82,89,0,76,67,95,77,69,83,83,65,71,69,83,0,76,65,78,71,0,67,46,85,84,70,45,56,0,80,79,83,73,88,0,77,85,83,76,95,76,79,67,80,65,84,72,0,78,83,116,51,95,95,50,56,105,111,115,95,98,97,115,101,69,0,78,83,116,51,95,95,50,57,98,97,115,105,99,95,105,111,115,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,50,57,98,97,115,105,99,95,105,111,115,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,50,49,53,98,97,115,105,99,95,115,116,114,101,97,109,98,117,102,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,50,49,53,98,97,115,105,99,95,115,116,114,101,97,109,98,117,102,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,50,49,51,98,97,115,105,99,95,105,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,50,49,51,98,97,115,105,99,95,105,115,116,114,101,97,109,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,78,83,116,51,95,95,50,49,51,98,97,115,105,99,95,111,115,116,114,101,97,109,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,0,78,83,116,51,95,95,50,49,51,98,97,115,105,99,95,111,115,116,114,101,97,109,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,0,117,110,115,112,101,99,105,102,105,101,100,32,105,111,115,116,114,101,97,109,95,99,97,116,101,103,111,114,121,32,101,114,114,111,114,0,105,111,115,116,114,101,97,109,0,78,83,116,51,95,95,50,49,57,95,95,105,111,115,116,114,101,97,109,95,99,97,116,101,103,111,114,121,69,0,78,83,116,51,95,95,50,56,105,111,115,95,98,97,115,101,55,102,97,105,108,117,114,101,69,0,105,111,115,95,98,97,115,101,58,58,99,108,101,97,114,0,78,83,116,51,95,95,50,49,49,95,95,115,116,100,111,117,116,98,117,102,73,119,69,69,0,78,83,116,51,95,95,50,49,49,95,95,115,116,100,111,117,116,98,117,102,73,99,69,69,0,117,110,115,117,112,112,111,114,116,101,100,32,108,111,99,97,108,101,32,102,111,114,32,115,116,97,110,100,97,114,100,32,105,110,112,117,116,0,78,83,116,51,95,95,50,49,48,95,95,115,116,100,105,110,98,117,102,73,119,69,69,0,78,83,116,51,95,95,50,49,48,95,95,115,116,100,105,110,98,117,102,73,99,69,69,0,78,83,116,51,95,95,50,55,99,111,108,108,97,116,101,73,99,69,69,0,78,83,116,51,95,95,50,54,108,111,99,97,108,101,53,102,97,99,101,116,69,0,78,83,116,51,95,95,50,55,99,111,108,108,97,116,101,73,119,69,69,0,48,49,50,51,52,53,54,55,56,57,97,98,99,100,101,102,65,66,67,68,69,70,120,88,43,45,112,80,105,73,110,78,0,37,112,0,67,0,78,83,116,51,95,95,50,55,110,117,109,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,57,95,95,110,117,109,95,103,101,116,73,99,69,69,0,78,83,116,51,95,95,50,49,52,95,95,110,117,109,95,103,101,116,95,98,97,115,101,69,0,78,83,116,51,95,95,50,55,110,117,109,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,57,95,95,110,117,109,95,103,101,116,73,119,69,69,0,37,112,0,0,0,0,76,0,37,0,0,0,0,0,78,83,116,51,95,95,50,55,110,117,109,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,57,95,95,110,117,109,95,112,117,116,73,99,69,69,0,78,83,116,51,95,95,50,49,52,95,95,110,117,109,95,112,117,116,95,98,97,115,101,69,0,78,83,116,51,95,95,50,55,110,117,109,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,57,95,95,110,117,109,95,112,117,116,73,119,69,69,0,37,72,58,37,77,58,37,83,0,37,109,47,37,100,47,37,121,0,37,73,58,37,77,58,37,83,32,37,112,0,37,97,32,37,98,32,37,100,32,37,72,58,37,77,58,37,83,32,37,89,0,65,77,0,80,77,0,74,97,110,117,97,114,121,0,70,101,98,114,117,97,114,121,0,77,97,114,99,104,0,65,112,114,105,108,0,77,97,121,0,74,117,110,101,0,74,117,108,121,0,65,117,103,117,115,116,0,83,101,112,116,101,109,98,101,114,0,79,99,116,111,98,101,114,0,78,111,118,101,109,98,101,114,0,68,101,99,101,109,98,101,114,0,74,97,110,0,70,101,98,0,77,97,114,0,65,112,114,0,74,117,110,0,74,117,108,0,65,117,103,0,83,101,112,0,79,99,116,0,78,111,118,0,68,101,99,0,83,117,110,100,97,121,0,77,111,110,100,97,121,0,84,117,101,115,100,97,121,0,87,101,100,110,101,115,100,97,121,0,84,104,117,114,115,100,97,121,0,70,114,105,100,97,121,0,83,97,116,117,114,100,97,121,0,83,117,110,0,77,111,110,0,84,117,101,0,87,101,100,0,84,104,117,0,70,114,105,0,83,97,116,0,37,109,47,37,100,47,37,121,37,89,45,37,109,45,37,100,37,73,58,37,77,58,37,83,32,37,112,37,72,58,37,77,37,72,58,37,77,58,37,83,37,72,58,37,77,58,37,83,78,83,116,51,95,95,50,56,116,105,109,101,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,50,48,95,95,116,105,109,101,95,103,101,116,95,99,95,115,116,111,114,97,103,101,73,99,69,69,0,78,83,116,51,95,95,50,57,116,105,109,101,95,98,97,115,101,69,0,78,83,116,51,95,95,50,56,116,105,109,101,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,50,48,95,95,116,105,109,101,95,103,101,116,95,99,95,115,116,111,114,97,103,101,73,119,69,69,0,78,83,116,51,95,95,50,56,116,105,109,101,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,48,95,95,116,105,109,101,95,112,117,116,69,0,108,111,99,97,108,101,32,110,111,116,32,115,117,112,112,111,114,116,101,100,0,78,83,116,51,95,95,50,56,116,105,109,101,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,48,109,111,110,101,121,112,117,110,99,116,73,99,76,98,48,69,69,69,0,78,83,116,51,95,95,50,49,48,109,111,110,101,121,95,98,97,115,101,69,0,78,83,116,51,95,95,50,49,48,109,111,110,101,121,112,117,110,99,116,73,99,76,98,49,69,69,69,0,78,83,116,51,95,95,50,49,48,109,111,110,101,121,112,117,110,99,116,73,119,76,98,48,69,69,69,0,78,83,116,51,95,95,50,49,48,109,111,110,101,121,112,117,110,99,116,73,119,76,98,49,69,69,69,0,48,49,50,51,52,53,54,55,56,57,0,37,76,102,0,109,111,110,101,121,95,103,101,116,32,101,114,114,111,114,0,78,83,116,51,95,95,50,57,109,111,110,101,121,95,103,101,116,73,99,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,49,95,95,109,111,110,101,121,95,103,101,116,73,99,69,69,0,48,49,50,51,52,53,54,55,56,57,0,78,83,116,51,95,95,50,57,109,111,110,101,121,95,103,101,116,73,119,78,83,95,49,57,105,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,49,95,95,109,111,110,101,121,95,103,101,116,73,119,69,69,0,37,46,48,76,102,0,78,83,116,51,95,95,50,57,109,111,110,101,121,95,112,117,116,73,99,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,99,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,99,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,49,95,95,109,111,110,101,121,95,112,117,116,73,99,69,69,0,78,83,116,51,95,95,50,57,109,111,110,101,121,95,112,117,116,73,119,78,83,95,49,57,111,115,116,114,101,97,109,98,117,102,95,105,116,101,114,97,116,111,114,73,119,78,83,95,49,49,99,104,97,114,95,116,114,97,105,116,115,73,119,69,69,69,69,69,69,0,78,83,116,51,95,95,50,49,49,95,95,109,111,110,101,121,95,112,117,116,73,119,69,69,0,78,83,116,51,95,95,50,56,109,101,115,115,97,103,101,115,73,99,69,69,0,78,83,116,51,95,95,50,49,51,109,101,115,115,97,103,101,115,95,98,97,115,101,69,0,78,83,116,51,95,95,50,49,55,95,95,119,105,100,101,110,95,102,114,111,109,95,117,116,102,56,73,76,106,51,50,69,69,69,0,78,83,116,51,95,95,50,55,99,111,100,101,99,118,116,73,68,105,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,50,49,50,99,111,100,101,99,118,116,95,98,97,115,101,69,0,78,83,116,51,95,95,50,49,54,95,95,110,97,114,114,111,119,95,116,111,95,117,116,102,56,73,76,106,51,50,69,69,69,0,78,83,116,51,95,95,50,56,109,101,115,115,97,103,101,115,73,119,69,69,0,78,83,116,51,95,95,50,55,99,111,100,101,99,118,116,73,99,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,50,55,99,111,100,101,99,118,116,73,119,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,50,55,99,111,100,101,99,118,116,73,68,115,99,49,49,95,95,109,98,115,116,97,116,101,95,116,69,69,0,78,83,116,51,95,95,50,54,108,111,99,97,108,101,53,95,95,105,109,112,69,0,78,83,116,51,95,95,50,53,99,116,121,112,101,73,99,69,69,0,78,83,116,51,95,95,50,49,48,99,116,121,112,101,95,98,97,115,101,69,0,78,83,116,51,95,95,50,53,99,116,121,112,101,73,119,69,69,0,78,83,116,51,95,95,50,56,110,117,109,112,117,110,99,116,73,99,69,69,0,78,83,116,51,95,95,50,56,110,117,109,112,117,110,99,116,73,119,69,69,0,118,101,99,116,111,114,0,78,83,116,51,95,95,50,49,52,95,95,115,104,97,114,101,100,95,99,111,117,110,116,69,0,97,108,108,111,99,97,116,111,114,60,84,62,58,58,97,108,108,111,99,97,116,101,40,115,105,122,101,95,116,32,110,41,32,39,110,39,32,101,120,99,101,101,100,115,32,109,97,120,105,109,117,109,32,115,117,112,112,111,114,116,101,100,32,115,105,122,101,0,85,110,107,110,111,119,110,32,101,114,114,111,114,32,37,100,0,110,101,119,95,101,114,114,110,111,32,61,61,32,69,82,65,78,71,69,0,68,58,92,68,101,118,92,101,109,115,100,107,92,101,109,115,99,114,105,112,116,101,110,92,49,46,51,55,46,50,49,92,115,121,115,116,101,109,92,108,105,98,92,108,105,98,99,120,120,92,115,121,115,116,101,109,95,101,114,114,111,114,46,99,112,112,0,100,111,95,115,116,114,101,114,114,111,114,95,114,0,78,83,116,51,95,95,50,49,50,95,95,100,111,95,109,101,115,115,97,103,101,69,0,78,83,116,51,95,95,50,49,52,101,114,114,111,114,95,99,97,116,101,103,111,114,121,69,0,78,83,116,51,95,95,50,49,50,115,121,115,116,101,109,95,101,114,114,111,114,69,0,58,32,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,101,120,99,101,112,116,105,111,110,32,111,102,32,116,121,112,101,32,37,115,58,32,37,115,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,101,120,99,101,112,116,105,111,110,32,111,102,32,116,121,112,101,32,37,115,0,116,101,114,109,105,110,97,116,105,110,103,32,119,105,116,104,32,37,115,32,102,111,114,101,105,103,110,32,101,120,99,101,112,116,105,111,110,0,116,101,114,109,105,110,97,116,105,110,103,0,117,110,99,97,117,103,104,116,0,83,116,57,101,120,99,101,112,116,105,111,110,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,54,95,95,115,104,105,109,95,116,121,112,101,95,105,110,102,111,69,0,83,116,57,116,121,112,101,95,105,110,102,111,0,78,49,48,95,95,99,120,120,97,98,105,118,49,50,48,95,95,115,105,95,99,108,97,115,115,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,55,95,95,99,108,97,115,115,95,116,121,112,101,95,105,110,102,111,69,0,112,116,104,114,101,97,100,95,111,110,99,101,32,102,97,105,108,117,114,101,32,105,110,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,95,102,97,115,116,40,41,0,99,97,110,110,111,116,32,99,114,101,97,116,101,32,112,116,104,114,101,97,100,32,107,101,121,32,102,111,114,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,40,41,0,99,97,110,110,111,116,32,122,101,114,111,32,111,117,116,32,116,104,114,101,97,100,32,118,97,108,117,101,32,102,111,114,32,95,95,99,120,97,95,103,101,116,95,103,108,111,98,97,108,115,40,41,0,116,101,114,109,105,110,97,116,101,95,104,97,110,100,108,101,114,32,117,110,101,120,112,101,99,116,101,100,108,121,32,114,101,116,117,114,110,101,100,0,116,101,114,109,105,110,97,116,101,95,104,97,110,100,108,101,114,32,117,110,101,120,112,101,99,116,101,100,108,121,32,116,104,114,101,119,32,97,110,32,101,120,99,101,112,116,105,111,110,0,115,116,100,58,58,98,97,100,95,97,108,108,111,99,0,83,116,57,98,97,100,95,97,108,108,111,99,0,83,116,49,49,108,111,103,105,99,95,101,114,114,111,114,0,83,116,49,51,114,117,110,116,105,109,101,95,101,114,114,111,114,0,83,116,49,50,108,101,110,103,116,104,95,101,114,114,111,114,0,83,116,49,50,111,117,116,95,111,102,95,114,97,110,103,101,0,115,116,100,58,58,98,97,100,95,99,97,115,116,0,83,116,56,98,97,100,95,99,97,115,116,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,57,95,95,112,111,105,110,116,101,114,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,49,55,95,95,112,98,97,115,101,95,116,121,112,101,95,105,110,102,111,69,0,78,49,48,95,95,99,120,120,97,98,105,118,49,50,49,95,95,118,109,105,95,99,108,97,115,115], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+148768);
/* memory initializer */ allocate([95,116,121,112,101,95,105,110,102,111,69,0,32,99,111,110,115,116,0,32,118,111,108,97,116,105,108,101,0,32,114,101,115,116,114,105,99,116,0,32,99,111,109,112,108,101,120,0,32,105,109,97,103,105,110,97,114,121,0,32,91,0,32,40,0,41,0,40,0,38,38,0,111,98,106,99,95,111,98,106,101,99,116,60,0,42,0,38,0,111,98,106,99,112,114,111,116,111,0,32,0,60,0,62,0,32,118,101,99,116,111,114,91,0,93,0,112,105,120,101,108,32,118,101,99,116,111,114,91,0,38,61,0,61,0,44,0,126,0,58,58,0,100,101,108,101,116,101,91,93,32,0,100,101,108,101,116,101,32,0,47,0,47,61,0,94,0,94,61,0,61,61,0,62,61,0,41,91,0,60,61,0,60,60,0,60,60,61,0,45,0,45,61,0,42,61,0,45,45,0,41,45,45,0,33,61,0,33,0,124,124,0,124,0,124,61,0,45,62,42,0,43,0,43,61,0,43,43,0,41,43,43,0,41,32,63,32,40,0,41,32,58,32,40,0,37,0,37,61,0,62,62,0,62,62,61,0,116,104,114,111,119,0,116,104,114,111,119,32,0,116,121,112,101,105,100,40,0,115,105,122,101,111,102,46,46,46,40,0,44,32,0,115,105,122,101,111,102,32,40,0,115,116,97,116,105,99,95,99,97,115,116,60,0,62,40,0,114,101,105,110,116,101,114,112,114,101,116,95,99,97,115,116,60,0,45,62,0,110,111,101,120,99,101,112,116,32,40,0,91,93,32,0,41,32,0,46,0,46,42,0,115,116,100,58,58,0,111,112,101,114,97,116,111,114,38,38,0,111,112,101,114,97,116,111,114,38,0,111,112,101,114,97,116,111,114,38,61,0,111,112,101,114,97,116,111,114,61,0,111,112,101,114,97,116,111,114,40,41,0,111,112,101,114,97,116,111,114,44,0,111,112,101,114,97,116,111,114,126,0,111,112,101,114,97,116,111,114,32,0,111,112,101,114,97,116,111,114,32,100,101,108,101,116,101,91,93,0,111,112,101,114,97,116,111,114,42,0,111,112,101,114,97,116,111,114,47,0,111,112,101,114,97,116,111,114,47,61,0,111,112,101,114,97,116,111,114,94,0,111,112,101,114,97,116,111,114,94,61,0,111,112,101,114,97,116,111,114,61,61,0,111,112,101,114,97,116,111,114,62,61,0,111,112,101,114,97,116,111,114,62,0,111,112,101,114,97,116,111,114,91,93,0,111,112,101,114,97,116,111,114,60,61,0,111,112,101,114,97,116,111,114,34,34,32,0,111,112,101,114,97,116,111,114,60,60,0,111,112,101,114,97,116,111,114,60,60,61,0,111,112,101,114,97,116,111,114,60,0,111,112,101,114,97,116,111,114,45,0,111,112,101,114,97,116,111,114,45,61,0,111,112,101,114,97,116,111,114,42,61,0,111,112,101,114,97,116,111,114,45,45,0,111,112,101,114,97,116,111,114,32,110,101,119,91,93,0,111,112,101,114,97,116,111,114,33,61,0,111,112,101,114,97,116,111,114,33,0,111,112,101,114,97,116,111,114,32,110,101,119,0,111,112,101,114,97,116,111,114,124,124,0,111,112,101,114,97,116,111,114,124,0,111,112,101,114,97,116,111,114,124,61,0,111,112,101,114,97,116,111,114,45,62,42,0,111,112,101,114,97,116,111,114,43,0,111,112,101,114,97,116,111,114,43,61,0,111,112,101,114,97,116,111,114,43,43,0,111,112,101,114,97,116,111,114,45,62,0,111,112,101,114,97,116,111,114,63,0,111,112,101,114,97,116,111,114,37,0,111,112,101,114,97,116,111,114,37,61,0,111,112,101,114,97,116,111,114,62,62,0,111,112,101,114,97,116,111,114,62,62,61,0,111,112,101,114,97,116,111,114,32,100,101,108,101,116,101,0,39,117,110,110,97,109,101,100,0,39,108,97,109,98,100,97,39,40,0,115,116,100,58,58,115,116,114,105,110,103,0,115,116,100,58,58,98,97,115,105,99,95,115,116,114,105,110,103,60,99,104,97,114,44,32,115,116,100,58,58,99,104,97,114,95,116,114,97,105,116,115,60,99,104,97,114,62,44,32,115,116,100,58,58,97,108,108,111,99,97,116,111,114,60,99,104,97,114,62,32,62,0,98,97,115,105,99,95,115,116,114,105,110,103,0,115,116,100,58,58,105,115,116,114,101,97,109,0,115,116,100,58,58,98,97,115,105,99,95,105,115,116,114,101,97,109,60,99,104,97,114,44,32,115,116,100,58,58,99,104,97,114,95,116,114,97,105,116,115,60,99,104,97,114,62,32,62,0,98,97,115,105,99,95,105,115,116,114,101,97,109,0,115,116,100,58,58,111,115,116,114,101,97,109,0,115,116,100,58,58,98,97,115,105,99,95,111,115,116,114,101,97,109,60,99,104,97,114,44,32,115,116,100,58,58,99,104,97,114,95,116,114,97,105,116,115,60,99,104,97,114,62,32,62,0,98,97,115,105,99,95,111,115,116,114,101,97,109,0,115,116,100,58,58,105,111,115,116,114,101,97,109,0,115,116,100,58,58,98,97,115,105,99,95,105,111,115,116,114,101,97,109,60,99,104,97,114,44,32,115,116,100,58,58,99,104,97,114,95,116,114,97,105,116,115,60,99,104,97,114,62,32,62,0,98,97,115,105,99,95,105,111,115,116,114,101,97,109,0,100,121,110,97,109,105,99,95,99,97,115,116,60,0,41,40,0,99,111,110,115,116,95,99,97,115,116,60,0,97,108,105,103,110,111,102,32,40,0,102,112,0,119,99,104,97,114,95,116,0,102,97,108,115,101,0,116,114,117,101,0,99,104,97,114,0,115,105,103,110,101,100,32,99,104,97,114,0,117,110,115,105,103,110,101,100,32,99,104,97,114,0,115,104,111,114,116,0,117,110,115,105,103,110,101,100,32,115,104,111,114,116,0,117,0,108,0,117,108,0,108,108,0,117,108,108,0,95,95,105,110,116,49,50,56,0,117,110,115,105,103,110,101,100,32,95,95,105,110,116,49,50,56,0,37,76,97,76,0,37,97,0,37,97,102,0,100,101,99,108,116,121,112,101,40,0,115,116,100,58,58,97,108,108,111,99,97,116,111,114,0,115,116,100,58,58,98,97,115,105,99,95,115,116,114,105,110,103,0,58,58,115,116,114,105,110,103,32,108,105,116,101,114,97,108,0,115,116,100,0,95,71,76,79,66,65,76,95,95,78,0,40,97,110,111,110,121,109,111,117,115,32,110,97,109,101,115,112,97,99,101,41,0,32,62,0,84,95,0,105,100,0,58,58,42,0,32,38,0,32,38,38,0,32,91,93,0,118,111,105,100,0,98,111,111,108,0,105,110,116,0,117,110,115,105,103,110,101,100,32,105,110,116,0,108,111,110,103,0,117,110,115,105,103,110,101,100,32,108,111,110,103,0,108,111,110,103,32,108,111,110,103,0,102,108,111,97,116,0,108,111,110,103,32,100,111,117,98,108,101,0,95,95,102,108,111,97,116,49,50,56,0,46,46,46,0,100,101,99,105,109,97,108,54,52,0,100,101,99,105,109,97,108,49,50,56,0,100,101,99,105,109,97,108,51,50,0,100,101,99,105,109,97,108,49,54,0,99,104,97,114,51,50,95,116,0,99,104,97,114,49,54,95,116,0,97,117,116,111,0,115,116,100,58,58,110,117,108,108,112,116,114,95,116,0,100,111,117,98,108,101,0,117,110,115,105,103,110,101,100,32,108,111,110,103,32,108,111,110,103,0,95,98,108,111,99,107,95,105,110,118,111,107,101,0,105,110,118,111,99,97,116,105,111,110,32,102,117,110,99,116,105,111,110,32,102,111,114,32,98,108,111,99,107,32,105,110,32,0,118,116,97,98,108,101,32,102,111,114,32,0,86,84,84,32,102,111,114,32,0,116,121,112,101,105,110,102,111,32,102,111,114,32,0,116,121,112,101,105,110,102,111,32,110,97,109,101,32,102,111,114,32,0,99,111,118,97,114,105,97,110,116,32,114,101,116,117,114,110,32,116,104,117,110,107,32,116,111,32,0,99,111,110,115,116,114,117,99,116,105,111,110,32,118,116,97,98,108,101,32,102,111,114,32,0,45,105,110,45,0,118,105,114,116,117,97,108,32,116,104,117,110,107,32,116,111,32,0,110,111,110,45,118,105,114,116,117,97,108,32,116,104,117,110,107,32,116,111,32,0,103,117,97,114,100,32,118,97,114,105,97,98,108,101,32,102,111,114,32,0,114,101,102,101,114,101,110,99,101,32,116,101,109,112,111,114,97,114,121,32,102,111,114,32,0], "i8", ALLOC_NONE, Runtime.GLOBAL_BASE+159008);





/* no memory initializer */
var tempDoublePtr = STATICTOP; STATICTOP += 16;

assert(tempDoublePtr % 8 == 0);

function copyTempFloat(ptr) { // functions, because inlining this code increases code size too much

  HEAP8[tempDoublePtr] = HEAP8[ptr];

  HEAP8[tempDoublePtr+1] = HEAP8[ptr+1];

  HEAP8[tempDoublePtr+2] = HEAP8[ptr+2];

  HEAP8[tempDoublePtr+3] = HEAP8[ptr+3];

}

function copyTempDouble(ptr) {

  HEAP8[tempDoublePtr] = HEAP8[ptr];

  HEAP8[tempDoublePtr+1] = HEAP8[ptr+1];

  HEAP8[tempDoublePtr+2] = HEAP8[ptr+2];

  HEAP8[tempDoublePtr+3] = HEAP8[ptr+3];

  HEAP8[tempDoublePtr+4] = HEAP8[ptr+4];

  HEAP8[tempDoublePtr+5] = HEAP8[ptr+5];

  HEAP8[tempDoublePtr+6] = HEAP8[ptr+6];

  HEAP8[tempDoublePtr+7] = HEAP8[ptr+7];

}

// {{PRE_LIBRARY}}


  
  function _atexit(func, arg) {
      __ATEXIT__.unshift({ func: func, arg: arg });
    }function ___cxa_atexit() {
  return _atexit.apply(null, arguments)
  }

   

  function ___assert_fail(condition, filename, line, func) {
      ABORT = true;
      throw 'Assertion failed: ' + Pointer_stringify(condition) + ', at: ' + [filename ? Pointer_stringify(filename) : 'unknown filename', line, func ? Pointer_stringify(func) : 'unknown function'] + ' at ' + stackTrace();
    }

  
  function __ZSt18uncaught_exceptionv() { // std::uncaught_exception()
      return !!__ZSt18uncaught_exceptionv.uncaught_exception;
    }
  
  
  
  var EXCEPTIONS={last:0,caught:[],infos:{},deAdjust:function (adjusted) {
        if (!adjusted || EXCEPTIONS.infos[adjusted]) return adjusted;
        for (var ptr in EXCEPTIONS.infos) {
          var info = EXCEPTIONS.infos[ptr];
          if (info.adjusted === adjusted) {
            return ptr;
          }
        }
        return adjusted;
      },addRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        info.refcount++;
      },decRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        assert(info.refcount > 0);
        info.refcount--;
        // A rethrown exception can reach refcount 0; it must not be discarded
        // Its next handler will clear the rethrown flag and addRef it, prior to
        // final decRef and destruction here
        if (info.refcount === 0 && !info.rethrown) {
          if (info.destructor) {
            Module['dynCall_vi'](info.destructor, ptr);
          }
          delete EXCEPTIONS.infos[ptr];
          ___cxa_free_exception(ptr);
        }
      },clearRef:function (ptr) {
        if (!ptr) return;
        var info = EXCEPTIONS.infos[ptr];
        info.refcount = 0;
      }};
  function ___resumeException(ptr) {
      if (!EXCEPTIONS.last) { EXCEPTIONS.last = ptr; }
      throw ptr;
    }function ___cxa_find_matching_catch() {
      var thrown = EXCEPTIONS.last;
      if (!thrown) {
        // just pass through the null ptr
        return ((Runtime.setTempRet0(0),0)|0);
      }
      var info = EXCEPTIONS.infos[thrown];
      var throwntype = info.type;
      if (!throwntype) {
        // just pass through the thrown ptr
        return ((Runtime.setTempRet0(0),thrown)|0);
      }
      var typeArray = Array.prototype.slice.call(arguments);
  
      var pointer = Module['___cxa_is_pointer_type'](throwntype);
      // can_catch receives a **, add indirection
      if (!___cxa_find_matching_catch.buffer) ___cxa_find_matching_catch.buffer = _malloc(4);
      HEAP32[((___cxa_find_matching_catch.buffer)>>2)]=thrown;
      thrown = ___cxa_find_matching_catch.buffer;
      // The different catch blocks are denoted by different types.
      // Due to inheritance, those types may not precisely match the
      // type of the thrown object. Find one which matches, and
      // return the type of the catch block which should be called.
      for (var i = 0; i < typeArray.length; i++) {
        if (typeArray[i] && Module['___cxa_can_catch'](typeArray[i], throwntype, thrown)) {
          thrown = HEAP32[((thrown)>>2)]; // undo indirection
          info.adjusted = thrown;
          return ((Runtime.setTempRet0(typeArray[i]),thrown)|0);
        }
      }
      // Shouldn't happen unless we have bogus data in typeArray
      // or encounter a type for which emscripten doesn't have suitable
      // typeinfo defined. Best-efforts match just in case.
      thrown = HEAP32[((thrown)>>2)]; // undo indirection
      return ((Runtime.setTempRet0(throwntype),thrown)|0);
    }function ___cxa_throw(ptr, type, destructor) {
      EXCEPTIONS.infos[ptr] = {
        ptr: ptr,
        adjusted: ptr,
        type: type,
        destructor: destructor,
        refcount: 0,
        caught: false,
        rethrown: false
      };
      EXCEPTIONS.last = ptr;
      if (!("uncaught_exception" in __ZSt18uncaught_exceptionv)) {
        __ZSt18uncaught_exceptionv.uncaught_exception = 1;
      } else {
        __ZSt18uncaught_exceptionv.uncaught_exception++;
      }
      throw ptr;
    }

   

   

  
  
  function __isLeapYear(year) {
        return year%4 === 0 && (year%100 !== 0 || year%400 === 0);
    }
  
  function __arraySum(array, index) {
      var sum = 0;
      for (var i = 0; i <= index; sum += array[i++]);
      return sum;
    }
  
  
  var __MONTH_DAYS_LEAP=[31,29,31,30,31,30,31,31,30,31,30,31];
  
  var __MONTH_DAYS_REGULAR=[31,28,31,30,31,30,31,31,30,31,30,31];function __addDays(date, days) {
      var newDate = new Date(date.getTime());
      while(days > 0) {
        var leap = __isLeapYear(newDate.getFullYear());
        var currentMonth = newDate.getMonth();
        var daysInCurrentMonth = (leap ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR)[currentMonth];
  
        if (days > daysInCurrentMonth-newDate.getDate()) {
          // we spill over to next month
          days -= (daysInCurrentMonth-newDate.getDate()+1);
          newDate.setDate(1);
          if (currentMonth < 11) {
            newDate.setMonth(currentMonth+1)
          } else {
            newDate.setMonth(0);
            newDate.setFullYear(newDate.getFullYear()+1);
          }
        } else {
          // we stay in current month 
          newDate.setDate(newDate.getDate()+days);
          return newDate;
        }
      }
  
      return newDate;
    }function _strftime(s, maxsize, format, tm) {
      // size_t strftime(char *restrict s, size_t maxsize, const char *restrict format, const struct tm *restrict timeptr);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/strftime.html
  
      var tm_zone = HEAP32[(((tm)+(40))>>2)];
  
      var date = {
        tm_sec: HEAP32[((tm)>>2)],
        tm_min: HEAP32[(((tm)+(4))>>2)],
        tm_hour: HEAP32[(((tm)+(8))>>2)],
        tm_mday: HEAP32[(((tm)+(12))>>2)],
        tm_mon: HEAP32[(((tm)+(16))>>2)],
        tm_year: HEAP32[(((tm)+(20))>>2)],
        tm_wday: HEAP32[(((tm)+(24))>>2)],
        tm_yday: HEAP32[(((tm)+(28))>>2)],
        tm_isdst: HEAP32[(((tm)+(32))>>2)],
        tm_gmtoff: HEAP32[(((tm)+(36))>>2)],
        tm_zone: tm_zone ? Pointer_stringify(tm_zone) : ''
      };
  
      var pattern = Pointer_stringify(format);
  
      // expand format
      var EXPANSION_RULES_1 = {
        '%c': '%a %b %d %H:%M:%S %Y',     // Replaced by the locale's appropriate date and time representation - e.g., Mon Aug  3 14:02:01 2013
        '%D': '%m/%d/%y',                 // Equivalent to %m / %d / %y
        '%F': '%Y-%m-%d',                 // Equivalent to %Y - %m - %d
        '%h': '%b',                       // Equivalent to %b
        '%r': '%I:%M:%S %p',              // Replaced by the time in a.m. and p.m. notation
        '%R': '%H:%M',                    // Replaced by the time in 24-hour notation
        '%T': '%H:%M:%S',                 // Replaced by the time
        '%x': '%m/%d/%y',                 // Replaced by the locale's appropriate date representation
        '%X': '%H:%M:%S'                  // Replaced by the locale's appropriate date representation
      };
      for (var rule in EXPANSION_RULES_1) {
        pattern = pattern.replace(new RegExp(rule, 'g'), EXPANSION_RULES_1[rule]);
      }
  
      var WEEKDAYS = ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'];
      var MONTHS = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
  
      function leadingSomething(value, digits, character) {
        var str = typeof value === 'number' ? value.toString() : (value || '');
        while (str.length < digits) {
          str = character[0]+str;
        }
        return str;
      };
  
      function leadingNulls(value, digits) {
        return leadingSomething(value, digits, '0');
      };
  
      function compareByDay(date1, date2) {
        function sgn(value) {
          return value < 0 ? -1 : (value > 0 ? 1 : 0);
        };
  
        var compare;
        if ((compare = sgn(date1.getFullYear()-date2.getFullYear())) === 0) {
          if ((compare = sgn(date1.getMonth()-date2.getMonth())) === 0) {
            compare = sgn(date1.getDate()-date2.getDate());
          }
        }
        return compare;
      };
  
      function getFirstWeekStartDate(janFourth) {
          switch (janFourth.getDay()) {
            case 0: // Sunday
              return new Date(janFourth.getFullYear()-1, 11, 29);
            case 1: // Monday
              return janFourth;
            case 2: // Tuesday
              return new Date(janFourth.getFullYear(), 0, 3);
            case 3: // Wednesday
              return new Date(janFourth.getFullYear(), 0, 2);
            case 4: // Thursday
              return new Date(janFourth.getFullYear(), 0, 1);
            case 5: // Friday
              return new Date(janFourth.getFullYear()-1, 11, 31);
            case 6: // Saturday
              return new Date(janFourth.getFullYear()-1, 11, 30);
          }
      };
  
      function getWeekBasedYear(date) {
          var thisDate = __addDays(new Date(date.tm_year+1900, 0, 1), date.tm_yday);
  
          var janFourthThisYear = new Date(thisDate.getFullYear(), 0, 4);
          var janFourthNextYear = new Date(thisDate.getFullYear()+1, 0, 4);
  
          var firstWeekStartThisYear = getFirstWeekStartDate(janFourthThisYear);
          var firstWeekStartNextYear = getFirstWeekStartDate(janFourthNextYear);
  
          if (compareByDay(firstWeekStartThisYear, thisDate) <= 0) {
            // this date is after the start of the first week of this year
            if (compareByDay(firstWeekStartNextYear, thisDate) <= 0) {
              return thisDate.getFullYear()+1;
            } else {
              return thisDate.getFullYear();
            }
          } else { 
            return thisDate.getFullYear()-1;
          }
      };
  
      var EXPANSION_RULES_2 = {
        '%a': function(date) {
          return WEEKDAYS[date.tm_wday].substring(0,3);
        },
        '%A': function(date) {
          return WEEKDAYS[date.tm_wday];
        },
        '%b': function(date) {
          return MONTHS[date.tm_mon].substring(0,3);
        },
        '%B': function(date) {
          return MONTHS[date.tm_mon];
        },
        '%C': function(date) {
          var year = date.tm_year+1900;
          return leadingNulls((year/100)|0,2);
        },
        '%d': function(date) {
          return leadingNulls(date.tm_mday, 2);
        },
        '%e': function(date) {
          return leadingSomething(date.tm_mday, 2, ' ');
        },
        '%g': function(date) {
          // %g, %G, and %V give values according to the ISO 8601:2000 standard week-based year. 
          // In this system, weeks begin on a Monday and week 1 of the year is the week that includes 
          // January 4th, which is also the week that includes the first Thursday of the year, and 
          // is also the first week that contains at least four days in the year. 
          // If the first Monday of January is the 2nd, 3rd, or 4th, the preceding days are part of 
          // the last week of the preceding year; thus, for Saturday 2nd January 1999, 
          // %G is replaced by 1998 and %V is replaced by 53. If December 29th, 30th, 
          // or 31st is a Monday, it and any following days are part of week 1 of the following year. 
          // Thus, for Tuesday 30th December 1997, %G is replaced by 1998 and %V is replaced by 01.
          
          return getWeekBasedYear(date).toString().substring(2);
        },
        '%G': function(date) {
          return getWeekBasedYear(date);
        },
        '%H': function(date) {
          return leadingNulls(date.tm_hour, 2);
        },
        '%I': function(date) {
          var twelveHour = date.tm_hour;
          if (twelveHour == 0) twelveHour = 12;
          else if (twelveHour > 12) twelveHour -= 12;
          return leadingNulls(twelveHour, 2);
        },
        '%j': function(date) {
          // Day of the year (001-366)
          return leadingNulls(date.tm_mday+__arraySum(__isLeapYear(date.tm_year+1900) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, date.tm_mon-1), 3);
        },
        '%m': function(date) {
          return leadingNulls(date.tm_mon+1, 2);
        },
        '%M': function(date) {
          return leadingNulls(date.tm_min, 2);
        },
        '%n': function() {
          return '\n';
        },
        '%p': function(date) {
          if (date.tm_hour >= 0 && date.tm_hour < 12) {
            return 'AM';
          } else {
            return 'PM';
          }
        },
        '%S': function(date) {
          return leadingNulls(date.tm_sec, 2);
        },
        '%t': function() {
          return '\t';
        },
        '%u': function(date) {
          var day = new Date(date.tm_year+1900, date.tm_mon+1, date.tm_mday, 0, 0, 0, 0);
          return day.getDay() || 7;
        },
        '%U': function(date) {
          // Replaced by the week number of the year as a decimal number [00,53]. 
          // The first Sunday of January is the first day of week 1; 
          // days in the new year before this are in week 0. [ tm_year, tm_wday, tm_yday]
          var janFirst = new Date(date.tm_year+1900, 0, 1);
          var firstSunday = janFirst.getDay() === 0 ? janFirst : __addDays(janFirst, 7-janFirst.getDay());
          var endDate = new Date(date.tm_year+1900, date.tm_mon, date.tm_mday);
          
          // is target date after the first Sunday?
          if (compareByDay(firstSunday, endDate) < 0) {
            // calculate difference in days between first Sunday and endDate
            var februaryFirstUntilEndMonth = __arraySum(__isLeapYear(endDate.getFullYear()) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, endDate.getMonth()-1)-31;
            var firstSundayUntilEndJanuary = 31-firstSunday.getDate();
            var days = firstSundayUntilEndJanuary+februaryFirstUntilEndMonth+endDate.getDate();
            return leadingNulls(Math.ceil(days/7), 2);
          }
  
          return compareByDay(firstSunday, janFirst) === 0 ? '01': '00';
        },
        '%V': function(date) {
          // Replaced by the week number of the year (Monday as the first day of the week) 
          // as a decimal number [01,53]. If the week containing 1 January has four 
          // or more days in the new year, then it is considered week 1. 
          // Otherwise, it is the last week of the previous year, and the next week is week 1. 
          // Both January 4th and the first Thursday of January are always in week 1. [ tm_year, tm_wday, tm_yday]
          var janFourthThisYear = new Date(date.tm_year+1900, 0, 4);
          var janFourthNextYear = new Date(date.tm_year+1901, 0, 4);
  
          var firstWeekStartThisYear = getFirstWeekStartDate(janFourthThisYear);
          var firstWeekStartNextYear = getFirstWeekStartDate(janFourthNextYear);
  
          var endDate = __addDays(new Date(date.tm_year+1900, 0, 1), date.tm_yday);
  
          if (compareByDay(endDate, firstWeekStartThisYear) < 0) {
            // if given date is before this years first week, then it belongs to the 53rd week of last year
            return '53';
          } 
  
          if (compareByDay(firstWeekStartNextYear, endDate) <= 0) {
            // if given date is after next years first week, then it belongs to the 01th week of next year
            return '01';
          }
  
          // given date is in between CW 01..53 of this calendar year
          var daysDifference;
          if (firstWeekStartThisYear.getFullYear() < date.tm_year+1900) {
            // first CW of this year starts last year
            daysDifference = date.tm_yday+32-firstWeekStartThisYear.getDate()
          } else {
            // first CW of this year starts this year
            daysDifference = date.tm_yday+1-firstWeekStartThisYear.getDate();
          }
          return leadingNulls(Math.ceil(daysDifference/7), 2);
        },
        '%w': function(date) {
          var day = new Date(date.tm_year+1900, date.tm_mon+1, date.tm_mday, 0, 0, 0, 0);
          return day.getDay();
        },
        '%W': function(date) {
          // Replaced by the week number of the year as a decimal number [00,53]. 
          // The first Monday of January is the first day of week 1; 
          // days in the new year before this are in week 0. [ tm_year, tm_wday, tm_yday]
          var janFirst = new Date(date.tm_year, 0, 1);
          var firstMonday = janFirst.getDay() === 1 ? janFirst : __addDays(janFirst, janFirst.getDay() === 0 ? 1 : 7-janFirst.getDay()+1);
          var endDate = new Date(date.tm_year+1900, date.tm_mon, date.tm_mday);
  
          // is target date after the first Monday?
          if (compareByDay(firstMonday, endDate) < 0) {
            var februaryFirstUntilEndMonth = __arraySum(__isLeapYear(endDate.getFullYear()) ? __MONTH_DAYS_LEAP : __MONTH_DAYS_REGULAR, endDate.getMonth()-1)-31;
            var firstMondayUntilEndJanuary = 31-firstMonday.getDate();
            var days = firstMondayUntilEndJanuary+februaryFirstUntilEndMonth+endDate.getDate();
            return leadingNulls(Math.ceil(days/7), 2);
          }
          return compareByDay(firstMonday, janFirst) === 0 ? '01': '00';
        },
        '%y': function(date) {
          // Replaced by the last two digits of the year as a decimal number [00,99]. [ tm_year]
          return (date.tm_year+1900).toString().substring(2);
        },
        '%Y': function(date) {
          // Replaced by the year as a decimal number (for example, 1997). [ tm_year]
          return date.tm_year+1900;
        },
        '%z': function(date) {
          // Replaced by the offset from UTC in the ISO 8601:2000 standard format ( +hhmm or -hhmm ).
          // For example, "-0430" means 4 hours 30 minutes behind UTC (west of Greenwich).
          var off = date.tm_gmtoff;
          var ahead = off >= 0;
          off = Math.abs(off) / 60;
          // convert from minutes into hhmm format (which means 60 minutes = 100 units)
          off = (off / 60)*100 + (off % 60);
          return (ahead ? '+' : '-') + String("0000" + off).slice(-4);
        },
        '%Z': function(date) {
          return date.tm_zone;
        },
        '%%': function() {
          return '%';
        }
      };
      for (var rule in EXPANSION_RULES_2) {
        if (pattern.indexOf(rule) >= 0) {
          pattern = pattern.replace(new RegExp(rule, 'g'), EXPANSION_RULES_2[rule](date));
        }
      }
  
      var bytes = intArrayFromString(pattern, false);
      if (bytes.length > maxsize) {
        return 0;
      } 
  
      writeArrayToMemory(bytes, s);
      return bytes.length-1;
    }function _strftime_l(s, maxsize, format, tm) {
      return _strftime(s, maxsize, format, tm); // no locale support yet
    }

  function _abort() {
      Module['abort']();
    }

  
  function ___cxa_free_exception(ptr) {
      try {
        return _free(ptr);
      } catch(e) { // XXX FIXME
        Module.printErr('exception during cxa_free_exception: ' + e);
      }
    }function ___cxa_end_catch() {
      // Clear state flag.
      Module['setThrew'](0);
      // Call destructor if one is registered then clear it.
      var ptr = EXCEPTIONS.caught.pop();
      if (ptr) {
        EXCEPTIONS.decRef(EXCEPTIONS.deAdjust(ptr));
        EXCEPTIONS.last = 0; // XXX in decRef?
      }
    }


  function _pthread_once(ptr, func) {
      if (!_pthread_once.seen) _pthread_once.seen = {};
      if (ptr in _pthread_once.seen) return;
      Module['dynCall_v'](func);
      _pthread_once.seen[ptr] = 1;
    }

  function ___lock() {}

  function ___unlock() {}

  
  var PTHREAD_SPECIFIC={};function _pthread_getspecific(key) {
      return PTHREAD_SPECIFIC[key] || 0;
    }

   

  function _clock() {
      if (_clock.start === undefined) _clock.start = Date.now();
      return ((Date.now() - _clock.start) * (1000000 / 1000))|0;
    }

  
  var PTHREAD_SPECIFIC_NEXT_KEY=1;
  
  var ERRNO_CODES={EPERM:1,ENOENT:2,ESRCH:3,EINTR:4,EIO:5,ENXIO:6,E2BIG:7,ENOEXEC:8,EBADF:9,ECHILD:10,EAGAIN:11,EWOULDBLOCK:11,ENOMEM:12,EACCES:13,EFAULT:14,ENOTBLK:15,EBUSY:16,EEXIST:17,EXDEV:18,ENODEV:19,ENOTDIR:20,EISDIR:21,EINVAL:22,ENFILE:23,EMFILE:24,ENOTTY:25,ETXTBSY:26,EFBIG:27,ENOSPC:28,ESPIPE:29,EROFS:30,EMLINK:31,EPIPE:32,EDOM:33,ERANGE:34,ENOMSG:42,EIDRM:43,ECHRNG:44,EL2NSYNC:45,EL3HLT:46,EL3RST:47,ELNRNG:48,EUNATCH:49,ENOCSI:50,EL2HLT:51,EDEADLK:35,ENOLCK:37,EBADE:52,EBADR:53,EXFULL:54,ENOANO:55,EBADRQC:56,EBADSLT:57,EDEADLOCK:35,EBFONT:59,ENOSTR:60,ENODATA:61,ETIME:62,ENOSR:63,ENONET:64,ENOPKG:65,EREMOTE:66,ENOLINK:67,EADV:68,ESRMNT:69,ECOMM:70,EPROTO:71,EMULTIHOP:72,EDOTDOT:73,EBADMSG:74,ENOTUNIQ:76,EBADFD:77,EREMCHG:78,ELIBACC:79,ELIBBAD:80,ELIBSCN:81,ELIBMAX:82,ELIBEXEC:83,ENOSYS:38,ENOTEMPTY:39,ENAMETOOLONG:36,ELOOP:40,EOPNOTSUPP:95,EPFNOSUPPORT:96,ECONNRESET:104,ENOBUFS:105,EAFNOSUPPORT:97,EPROTOTYPE:91,ENOTSOCK:88,ENOPROTOOPT:92,ESHUTDOWN:108,ECONNREFUSED:111,EADDRINUSE:98,ECONNABORTED:103,ENETUNREACH:101,ENETDOWN:100,ETIMEDOUT:110,EHOSTDOWN:112,EHOSTUNREACH:113,EINPROGRESS:115,EALREADY:114,EDESTADDRREQ:89,EMSGSIZE:90,EPROTONOSUPPORT:93,ESOCKTNOSUPPORT:94,EADDRNOTAVAIL:99,ENETRESET:102,EISCONN:106,ENOTCONN:107,ETOOMANYREFS:109,EUSERS:87,EDQUOT:122,ESTALE:116,ENOTSUP:95,ENOMEDIUM:123,EILSEQ:84,EOVERFLOW:75,ECANCELED:125,ENOTRECOVERABLE:131,EOWNERDEAD:130,ESTRPIPE:86};function _pthread_key_create(key, destructor) {
      if (key == 0) {
        return ERRNO_CODES.EINVAL;
      }
      HEAP32[((key)>>2)]=PTHREAD_SPECIFIC_NEXT_KEY;
      // values start at 0
      PTHREAD_SPECIFIC[PTHREAD_SPECIFIC_NEXT_KEY] = 0;
      PTHREAD_SPECIFIC_NEXT_KEY++;
      return 0;
    }

  var _emscripten_asm_const_int=true;


  
  
  
  var ERRNO_MESSAGES={0:"Success",1:"Not super-user",2:"No such file or directory",3:"No such process",4:"Interrupted system call",5:"I/O error",6:"No such device or address",7:"Arg list too long",8:"Exec format error",9:"Bad file number",10:"No children",11:"No more processes",12:"Not enough core",13:"Permission denied",14:"Bad address",15:"Block device required",16:"Mount device busy",17:"File exists",18:"Cross-device link",19:"No such device",20:"Not a directory",21:"Is a directory",22:"Invalid argument",23:"Too many open files in system",24:"Too many open files",25:"Not a typewriter",26:"Text file busy",27:"File too large",28:"No space left on device",29:"Illegal seek",30:"Read only file system",31:"Too many links",32:"Broken pipe",33:"Math arg out of domain of func",34:"Math result not representable",35:"File locking deadlock error",36:"File or path name too long",37:"No record locks available",38:"Function not implemented",39:"Directory not empty",40:"Too many symbolic links",42:"No message of desired type",43:"Identifier removed",44:"Channel number out of range",45:"Level 2 not synchronized",46:"Level 3 halted",47:"Level 3 reset",48:"Link number out of range",49:"Protocol driver not attached",50:"No CSI structure available",51:"Level 2 halted",52:"Invalid exchange",53:"Invalid request descriptor",54:"Exchange full",55:"No anode",56:"Invalid request code",57:"Invalid slot",59:"Bad font file fmt",60:"Device not a stream",61:"No data (for no delay io)",62:"Timer expired",63:"Out of streams resources",64:"Machine is not on the network",65:"Package not installed",66:"The object is remote",67:"The link has been severed",68:"Advertise error",69:"Srmount error",70:"Communication error on send",71:"Protocol error",72:"Multihop attempted",73:"Cross mount point (not really error)",74:"Trying to read unreadable message",75:"Value too large for defined data type",76:"Given log. name not unique",77:"f.d. invalid for this operation",78:"Remote address changed",79:"Can   access a needed shared lib",80:"Accessing a corrupted shared lib",81:".lib section in a.out corrupted",82:"Attempting to link in too many libs",83:"Attempting to exec a shared library",84:"Illegal byte sequence",86:"Streams pipe error",87:"Too many users",88:"Socket operation on non-socket",89:"Destination address required",90:"Message too long",91:"Protocol wrong type for socket",92:"Protocol not available",93:"Unknown protocol",94:"Socket type not supported",95:"Not supported",96:"Protocol family not supported",97:"Address family not supported by protocol family",98:"Address already in use",99:"Address not available",100:"Network interface is not configured",101:"Network is unreachable",102:"Connection reset by network",103:"Connection aborted",104:"Connection reset by peer",105:"No buffer space available",106:"Socket is already connected",107:"Socket is not connected",108:"Can't send after socket shutdown",109:"Too many references",110:"Connection timed out",111:"Connection refused",112:"Host is down",113:"Host is unreachable",114:"Socket already connected",115:"Connection already in progress",116:"Stale file handle",122:"Quota exceeded",123:"No medium (in tape drive)",125:"Operation canceled",130:"Previous owner died",131:"State not recoverable"};
  
  function ___setErrNo(value) {
      if (Module['___errno_location']) HEAP32[((Module['___errno_location']())>>2)]=value;
      else Module.printErr('failed to set errno from JS');
      return value;
    }
  
  var PATH={splitPath:function (filename) {
        var splitPathRe = /^(\/?|)([\s\S]*?)((?:\.{1,2}|[^\/]+?|)(\.[^.\/]*|))(?:[\/]*)$/;
        return splitPathRe.exec(filename).slice(1);
      },normalizeArray:function (parts, allowAboveRoot) {
        // if the path tries to go above the root, `up` ends up > 0
        var up = 0;
        for (var i = parts.length - 1; i >= 0; i--) {
          var last = parts[i];
          if (last === '.') {
            parts.splice(i, 1);
          } else if (last === '..') {
            parts.splice(i, 1);
            up++;
          } else if (up) {
            parts.splice(i, 1);
            up--;
          }
        }
        // if the path is allowed to go above the root, restore leading ..s
        if (allowAboveRoot) {
          for (; up; up--) {
            parts.unshift('..');
          }
        }
        return parts;
      },normalize:function (path) {
        var isAbsolute = path.charAt(0) === '/',
            trailingSlash = path.substr(-1) === '/';
        // Normalize the path
        path = PATH.normalizeArray(path.split('/').filter(function(p) {
          return !!p;
        }), !isAbsolute).join('/');
        if (!path && !isAbsolute) {
          path = '.';
        }
        if (path && trailingSlash) {
          path += '/';
        }
        return (isAbsolute ? '/' : '') + path;
      },dirname:function (path) {
        var result = PATH.splitPath(path),
            root = result[0],
            dir = result[1];
        if (!root && !dir) {
          // No dirname whatsoever
          return '.';
        }
        if (dir) {
          // It has a dirname, strip trailing slash
          dir = dir.substr(0, dir.length - 1);
        }
        return root + dir;
      },basename:function (path) {
        // EMSCRIPTEN return '/'' for '/', not an empty string
        if (path === '/') return '/';
        var lastSlash = path.lastIndexOf('/');
        if (lastSlash === -1) return path;
        return path.substr(lastSlash+1);
      },extname:function (path) {
        return PATH.splitPath(path)[3];
      },join:function () {
        var paths = Array.prototype.slice.call(arguments, 0);
        return PATH.normalize(paths.join('/'));
      },join2:function (l, r) {
        return PATH.normalize(l + '/' + r);
      },resolve:function () {
        var resolvedPath = '',
          resolvedAbsolute = false;
        for (var i = arguments.length - 1; i >= -1 && !resolvedAbsolute; i--) {
          var path = (i >= 0) ? arguments[i] : FS.cwd();
          // Skip empty and invalid entries
          if (typeof path !== 'string') {
            throw new TypeError('Arguments to path.resolve must be strings');
          } else if (!path) {
            return ''; // an invalid portion invalidates the whole thing
          }
          resolvedPath = path + '/' + resolvedPath;
          resolvedAbsolute = path.charAt(0) === '/';
        }
        // At this point the path should be resolved to a full absolute path, but
        // handle relative paths to be safe (might happen when process.cwd() fails)
        resolvedPath = PATH.normalizeArray(resolvedPath.split('/').filter(function(p) {
          return !!p;
        }), !resolvedAbsolute).join('/');
        return ((resolvedAbsolute ? '/' : '') + resolvedPath) || '.';
      },relative:function (from, to) {
        from = PATH.resolve(from).substr(1);
        to = PATH.resolve(to).substr(1);
        function trim(arr) {
          var start = 0;
          for (; start < arr.length; start++) {
            if (arr[start] !== '') break;
          }
          var end = arr.length - 1;
          for (; end >= 0; end--) {
            if (arr[end] !== '') break;
          }
          if (start > end) return [];
          return arr.slice(start, end - start + 1);
        }
        var fromParts = trim(from.split('/'));
        var toParts = trim(to.split('/'));
        var length = Math.min(fromParts.length, toParts.length);
        var samePartsLength = length;
        for (var i = 0; i < length; i++) {
          if (fromParts[i] !== toParts[i]) {
            samePartsLength = i;
            break;
          }
        }
        var outputParts = [];
        for (var i = samePartsLength; i < fromParts.length; i++) {
          outputParts.push('..');
        }
        outputParts = outputParts.concat(toParts.slice(samePartsLength));
        return outputParts.join('/');
      }};
  
  var TTY={ttys:[],init:function () {
        // https://github.com/kripken/emscripten/pull/1555
        // if (ENVIRONMENT_IS_NODE) {
        //   // currently, FS.init does not distinguish if process.stdin is a file or TTY
        //   // device, it always assumes it's a TTY device. because of this, we're forcing
        //   // process.stdin to UTF8 encoding to at least make stdin reading compatible
        //   // with text files until FS.init can be refactored.
        //   process['stdin']['setEncoding']('utf8');
        // }
      },shutdown:function () {
        // https://github.com/kripken/emscripten/pull/1555
        // if (ENVIRONMENT_IS_NODE) {
        //   // inolen: any idea as to why node -e 'process.stdin.read()' wouldn't exit immediately (with process.stdin being a tty)?
        //   // isaacs: because now it's reading from the stream, you've expressed interest in it, so that read() kicks off a _read() which creates a ReadReq operation
        //   // inolen: I thought read() in that case was a synchronous operation that just grabbed some amount of buffered data if it exists?
        //   // isaacs: it is. but it also triggers a _read() call, which calls readStart() on the handle
        //   // isaacs: do process.stdin.pause() and i'd think it'd probably close the pending call
        //   process['stdin']['pause']();
        // }
      },register:function (dev, ops) {
        TTY.ttys[dev] = { input: [], output: [], ops: ops };
        FS.registerDevice(dev, TTY.stream_ops);
      },stream_ops:{open:function (stream) {
          var tty = TTY.ttys[stream.node.rdev];
          if (!tty) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          stream.tty = tty;
          stream.seekable = false;
        },close:function (stream) {
          // flush any pending line data
          stream.tty.ops.flush(stream.tty);
        },flush:function (stream) {
          stream.tty.ops.flush(stream.tty);
        },read:function (stream, buffer, offset, length, pos /* ignored */) {
          if (!stream.tty || !stream.tty.ops.get_char) {
            throw new FS.ErrnoError(ERRNO_CODES.ENXIO);
          }
          var bytesRead = 0;
          for (var i = 0; i < length; i++) {
            var result;
            try {
              result = stream.tty.ops.get_char(stream.tty);
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
            if (result === undefined && bytesRead === 0) {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
            if (result === null || result === undefined) break;
            bytesRead++;
            buffer[offset+i] = result;
          }
          if (bytesRead) {
            stream.node.timestamp = Date.now();
          }
          return bytesRead;
        },write:function (stream, buffer, offset, length, pos) {
          if (!stream.tty || !stream.tty.ops.put_char) {
            throw new FS.ErrnoError(ERRNO_CODES.ENXIO);
          }
          for (var i = 0; i < length; i++) {
            try {
              stream.tty.ops.put_char(stream.tty, buffer[offset+i]);
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
          }
          if (length) {
            stream.node.timestamp = Date.now();
          }
          return i;
        }},default_tty_ops:{get_char:function (tty) {
          if (!tty.input.length) {
            var result = null;
            if (ENVIRONMENT_IS_NODE) {
              // we will read data by chunks of BUFSIZE
              var BUFSIZE = 256;
              var buf = new Buffer(BUFSIZE);
              var bytesRead = 0;
  
              var isPosixPlatform = (process.platform != 'win32'); // Node doesn't offer a direct check, so test by exclusion
  
              var fd = process.stdin.fd;
              if (isPosixPlatform) {
                // Linux and Mac cannot use process.stdin.fd (which isn't set up as sync)
                var usingDevice = false;
                try {
                  fd = fs.openSync('/dev/stdin', 'r');
                  usingDevice = true;
                } catch (e) {}
              }
  
              try {
                bytesRead = fs.readSync(fd, buf, 0, BUFSIZE, null);
              } catch(e) {
                // Cross-platform differences: on Windows, reading EOF throws an exception, but on other OSes,
                // reading EOF returns 0. Uniformize behavior by treating the EOF exception to return 0.
                if (e.toString().indexOf('EOF') != -1) bytesRead = 0;
                else throw e;
              }
  
              if (usingDevice) { fs.closeSync(fd); }
              if (bytesRead > 0) {
                result = buf.slice(0, bytesRead).toString('utf-8');
              } else {
                result = null;
              }
  
            } else if (typeof window != 'undefined' &&
              typeof window.prompt == 'function') {
              // Browser.
              result = window.prompt('Input: ');  // returns null on cancel
              if (result !== null) {
                result += '\n';
              }
            } else if (typeof readline == 'function') {
              // Command line.
              result = readline();
              if (result !== null) {
                result += '\n';
              }
            }
            if (!result) {
              return null;
            }
            tty.input = intArrayFromString(result, true);
          }
          return tty.input.shift();
        },put_char:function (tty, val) {
          if (val === null || val === 10) {
            Module['print'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          } else {
            if (val != 0) tty.output.push(val); // val == 0 would cut text output off in the middle.
          }
        },flush:function (tty) {
          if (tty.output && tty.output.length > 0) {
            Module['print'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          }
        }},default_tty1_ops:{put_char:function (tty, val) {
          if (val === null || val === 10) {
            Module['printErr'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          } else {
            if (val != 0) tty.output.push(val);
          }
        },flush:function (tty) {
          if (tty.output && tty.output.length > 0) {
            Module['printErr'](UTF8ArrayToString(tty.output, 0));
            tty.output = [];
          }
        }}};
  
  var MEMFS={ops_table:null,mount:function (mount) {
        return MEMFS.createNode(null, '/', 16384 | 511 /* 0777 */, 0);
      },createNode:function (parent, name, mode, dev) {
        if (FS.isBlkdev(mode) || FS.isFIFO(mode)) {
          // no supported
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (!MEMFS.ops_table) {
          MEMFS.ops_table = {
            dir: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr,
                lookup: MEMFS.node_ops.lookup,
                mknod: MEMFS.node_ops.mknod,
                rename: MEMFS.node_ops.rename,
                unlink: MEMFS.node_ops.unlink,
                rmdir: MEMFS.node_ops.rmdir,
                readdir: MEMFS.node_ops.readdir,
                symlink: MEMFS.node_ops.symlink
              },
              stream: {
                llseek: MEMFS.stream_ops.llseek
              }
            },
            file: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr
              },
              stream: {
                llseek: MEMFS.stream_ops.llseek,
                read: MEMFS.stream_ops.read,
                write: MEMFS.stream_ops.write,
                allocate: MEMFS.stream_ops.allocate,
                mmap: MEMFS.stream_ops.mmap,
                msync: MEMFS.stream_ops.msync
              }
            },
            link: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr,
                readlink: MEMFS.node_ops.readlink
              },
              stream: {}
            },
            chrdev: {
              node: {
                getattr: MEMFS.node_ops.getattr,
                setattr: MEMFS.node_ops.setattr
              },
              stream: FS.chrdev_stream_ops
            }
          };
        }
        var node = FS.createNode(parent, name, mode, dev);
        if (FS.isDir(node.mode)) {
          node.node_ops = MEMFS.ops_table.dir.node;
          node.stream_ops = MEMFS.ops_table.dir.stream;
          node.contents = {};
        } else if (FS.isFile(node.mode)) {
          node.node_ops = MEMFS.ops_table.file.node;
          node.stream_ops = MEMFS.ops_table.file.stream;
          node.usedBytes = 0; // The actual number of bytes used in the typed array, as opposed to contents.length which gives the whole capacity.
          // When the byte data of the file is populated, this will point to either a typed array, or a normal JS array. Typed arrays are preferred
          // for performance, and used by default. However, typed arrays are not resizable like normal JS arrays are, so there is a small disk size
          // penalty involved for appending file writes that continuously grow a file similar to std::vector capacity vs used -scheme.
          node.contents = null; 
        } else if (FS.isLink(node.mode)) {
          node.node_ops = MEMFS.ops_table.link.node;
          node.stream_ops = MEMFS.ops_table.link.stream;
        } else if (FS.isChrdev(node.mode)) {
          node.node_ops = MEMFS.ops_table.chrdev.node;
          node.stream_ops = MEMFS.ops_table.chrdev.stream;
        }
        node.timestamp = Date.now();
        // add the new node to the parent
        if (parent) {
          parent.contents[name] = node;
        }
        return node;
      },getFileDataAsRegularArray:function (node) {
        if (node.contents && node.contents.subarray) {
          var arr = [];
          for (var i = 0; i < node.usedBytes; ++i) arr.push(node.contents[i]);
          return arr; // Returns a copy of the original data.
        }
        return node.contents; // No-op, the file contents are already in a JS array. Return as-is.
      },getFileDataAsTypedArray:function (node) {
        if (!node.contents) return new Uint8Array;
        if (node.contents.subarray) return node.contents.subarray(0, node.usedBytes); // Make sure to not return excess unused bytes.
        return new Uint8Array(node.contents);
      },expandFileStorage:function (node, newCapacity) {
        // If we are asked to expand the size of a file that already exists, revert to using a standard JS array to store the file
        // instead of a typed array. This makes resizing the array more flexible because we can just .push() elements at the back to
        // increase the size.
        if (node.contents && node.contents.subarray && newCapacity > node.contents.length) {
          node.contents = MEMFS.getFileDataAsRegularArray(node);
          node.usedBytes = node.contents.length; // We might be writing to a lazy-loaded file which had overridden this property, so force-reset it.
        }
  
        if (!node.contents || node.contents.subarray) { // Keep using a typed array if creating a new storage, or if old one was a typed array as well.
          var prevCapacity = node.contents ? node.contents.length : 0;
          if (prevCapacity >= newCapacity) return; // No need to expand, the storage was already large enough.
          // Don't expand strictly to the given requested limit if it's only a very small increase, but instead geometrically grow capacity.
          // For small filesizes (<1MB), perform size*2 geometric increase, but for large sizes, do a much more conservative size*1.125 increase to
          // avoid overshooting the allocation cap by a very large margin.
          var CAPACITY_DOUBLING_MAX = 1024 * 1024;
          newCapacity = Math.max(newCapacity, (prevCapacity * (prevCapacity < CAPACITY_DOUBLING_MAX ? 2.0 : 1.125)) | 0);
          if (prevCapacity != 0) newCapacity = Math.max(newCapacity, 256); // At minimum allocate 256b for each file when expanding.
          var oldContents = node.contents;
          node.contents = new Uint8Array(newCapacity); // Allocate new storage.
          if (node.usedBytes > 0) node.contents.set(oldContents.subarray(0, node.usedBytes), 0); // Copy old data over to the new storage.
          return;
        }
        // Not using a typed array to back the file storage. Use a standard JS array instead.
        if (!node.contents && newCapacity > 0) node.contents = [];
        while (node.contents.length < newCapacity) node.contents.push(0);
      },resizeFileStorage:function (node, newSize) {
        if (node.usedBytes == newSize) return;
        if (newSize == 0) {
          node.contents = null; // Fully decommit when requesting a resize to zero.
          node.usedBytes = 0;
          return;
        }
        if (!node.contents || node.contents.subarray) { // Resize a typed array if that is being used as the backing store.
          var oldContents = node.contents;
          node.contents = new Uint8Array(new ArrayBuffer(newSize)); // Allocate new storage.
          if (oldContents) {
            node.contents.set(oldContents.subarray(0, Math.min(newSize, node.usedBytes))); // Copy old data over to the new storage.
          }
          node.usedBytes = newSize;
          return;
        }
        // Backing with a JS array.
        if (!node.contents) node.contents = [];
        if (node.contents.length > newSize) node.contents.length = newSize;
        else while (node.contents.length < newSize) node.contents.push(0);
        node.usedBytes = newSize;
      },node_ops:{getattr:function (node) {
          var attr = {};
          // device numbers reuse inode numbers.
          attr.dev = FS.isChrdev(node.mode) ? node.id : 1;
          attr.ino = node.id;
          attr.mode = node.mode;
          attr.nlink = 1;
          attr.uid = 0;
          attr.gid = 0;
          attr.rdev = node.rdev;
          if (FS.isDir(node.mode)) {
            attr.size = 4096;
          } else if (FS.isFile(node.mode)) {
            attr.size = node.usedBytes;
          } else if (FS.isLink(node.mode)) {
            attr.size = node.link.length;
          } else {
            attr.size = 0;
          }
          attr.atime = new Date(node.timestamp);
          attr.mtime = new Date(node.timestamp);
          attr.ctime = new Date(node.timestamp);
          // NOTE: In our implementation, st_blocks = Math.ceil(st_size/st_blksize),
          //       but this is not required by the standard.
          attr.blksize = 4096;
          attr.blocks = Math.ceil(attr.size / attr.blksize);
          return attr;
        },setattr:function (node, attr) {
          if (attr.mode !== undefined) {
            node.mode = attr.mode;
          }
          if (attr.timestamp !== undefined) {
            node.timestamp = attr.timestamp;
          }
          if (attr.size !== undefined) {
            MEMFS.resizeFileStorage(node, attr.size);
          }
        },lookup:function (parent, name) {
          throw FS.genericErrors[ERRNO_CODES.ENOENT];
        },mknod:function (parent, name, mode, dev) {
          return MEMFS.createNode(parent, name, mode, dev);
        },rename:function (old_node, new_dir, new_name) {
          // if we're overwriting a directory at new_name, make sure it's empty.
          if (FS.isDir(old_node.mode)) {
            var new_node;
            try {
              new_node = FS.lookupNode(new_dir, new_name);
            } catch (e) {
            }
            if (new_node) {
              for (var i in new_node.contents) {
                throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
              }
            }
          }
          // do the internal rewiring
          delete old_node.parent.contents[old_node.name];
          old_node.name = new_name;
          new_dir.contents[new_name] = old_node;
          old_node.parent = new_dir;
        },unlink:function (parent, name) {
          delete parent.contents[name];
        },rmdir:function (parent, name) {
          var node = FS.lookupNode(parent, name);
          for (var i in node.contents) {
            throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
          }
          delete parent.contents[name];
        },readdir:function (node) {
          var entries = ['.', '..']
          for (var key in node.contents) {
            if (!node.contents.hasOwnProperty(key)) {
              continue;
            }
            entries.push(key);
          }
          return entries;
        },symlink:function (parent, newname, oldpath) {
          var node = MEMFS.createNode(parent, newname, 511 /* 0777 */ | 40960, 0);
          node.link = oldpath;
          return node;
        },readlink:function (node) {
          if (!FS.isLink(node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return node.link;
        }},stream_ops:{read:function (stream, buffer, offset, length, position) {
          var contents = stream.node.contents;
          if (position >= stream.node.usedBytes) return 0;
          var size = Math.min(stream.node.usedBytes - position, length);
          assert(size >= 0);
          if (size > 8 && contents.subarray) { // non-trivial, and typed array
            buffer.set(contents.subarray(position, position + size), offset);
          } else {
            for (var i = 0; i < size; i++) buffer[offset + i] = contents[position + i];
          }
          return size;
        },write:function (stream, buffer, offset, length, position, canOwn) {
          if (!length) return 0;
          var node = stream.node;
          node.timestamp = Date.now();
  
          if (buffer.subarray && (!node.contents || node.contents.subarray)) { // This write is from a typed array to a typed array?
            if (canOwn) {
              assert(position === 0, 'canOwn must imply no weird position inside the file');
              node.contents = buffer.subarray(offset, offset + length);
              node.usedBytes = length;
              return length;
            } else if (node.usedBytes === 0 && position === 0) { // If this is a simple first write to an empty file, do a fast set since we don't need to care about old data.
              node.contents = new Uint8Array(buffer.subarray(offset, offset + length));
              node.usedBytes = length;
              return length;
            } else if (position + length <= node.usedBytes) { // Writing to an already allocated and used subrange of the file?
              node.contents.set(buffer.subarray(offset, offset + length), position);
              return length;
            }
          }
  
          // Appending to an existing file and we need to reallocate, or source data did not come as a typed array.
          MEMFS.expandFileStorage(node, position+length);
          if (node.contents.subarray && buffer.subarray) node.contents.set(buffer.subarray(offset, offset + length), position); // Use typed array write if available.
          else {
            for (var i = 0; i < length; i++) {
             node.contents[position + i] = buffer[offset + i]; // Or fall back to manual write if not.
            }
          }
          node.usedBytes = Math.max(node.usedBytes, position+length);
          return length;
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              position += stream.node.usedBytes;
            }
          }
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return position;
        },allocate:function (stream, offset, length) {
          MEMFS.expandFileStorage(stream.node, offset + length);
          stream.node.usedBytes = Math.max(stream.node.usedBytes, offset + length);
        },mmap:function (stream, buffer, offset, length, position, prot, flags) {
          if (!FS.isFile(stream.node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          var ptr;
          var allocated;
          var contents = stream.node.contents;
          // Only make a new copy when MAP_PRIVATE is specified.
          if ( !(flags & 2) &&
                (contents.buffer === buffer || contents.buffer === buffer.buffer) ) {
            // We can't emulate MAP_SHARED when the file is not backed by the buffer
            // we're mapping to (e.g. the HEAP buffer).
            allocated = false;
            ptr = contents.byteOffset;
          } else {
            // Try to avoid unnecessary slices.
            if (position > 0 || position + length < stream.node.usedBytes) {
              if (contents.subarray) {
                contents = contents.subarray(position, position + length);
              } else {
                contents = Array.prototype.slice.call(contents, position, position + length);
              }
            }
            allocated = true;
            ptr = _malloc(length);
            if (!ptr) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOMEM);
            }
            buffer.set(contents, ptr);
          }
          return { ptr: ptr, allocated: allocated };
        },msync:function (stream, buffer, offset, length, mmapFlags) {
          if (!FS.isFile(stream.node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
          }
          if (mmapFlags & 2) {
            // MAP_PRIVATE calls need not to be synced back to underlying fs
            return 0;
          }
  
          var bytesWritten = MEMFS.stream_ops.write(stream, buffer, 0, length, offset, false);
          // should we check if bytesWritten and length are the same?
          return 0;
        }}};
  
  var IDBFS={dbs:{},indexedDB:function () {
        if (typeof indexedDB !== 'undefined') return indexedDB;
        var ret = null;
        if (typeof window === 'object') ret = window.indexedDB || window.mozIndexedDB || window.webkitIndexedDB || window.msIndexedDB;
        assert(ret, 'IDBFS used, but indexedDB not supported');
        return ret;
      },DB_VERSION:21,DB_STORE_NAME:"FILE_DATA",mount:function (mount) {
        // reuse all of the core MEMFS functionality
        return MEMFS.mount.apply(null, arguments);
      },syncfs:function (mount, populate, callback) {
        IDBFS.getLocalSet(mount, function(err, local) {
          if (err) return callback(err);
  
          IDBFS.getRemoteSet(mount, function(err, remote) {
            if (err) return callback(err);
  
            var src = populate ? remote : local;
            var dst = populate ? local : remote;
  
            IDBFS.reconcile(src, dst, callback);
          });
        });
      },getDB:function (name, callback) {
        // check the cache first
        var db = IDBFS.dbs[name];
        if (db) {
          return callback(null, db);
        }
  
        var req;
        try {
          req = IDBFS.indexedDB().open(name, IDBFS.DB_VERSION);
        } catch (e) {
          return callback(e);
        }
        if (!req) {
          return callback("Unable to connect to IndexedDB");
        }
        req.onupgradeneeded = function(e) {
          var db = e.target.result;
          var transaction = e.target.transaction;
  
          var fileStore;
  
          if (db.objectStoreNames.contains(IDBFS.DB_STORE_NAME)) {
            fileStore = transaction.objectStore(IDBFS.DB_STORE_NAME);
          } else {
            fileStore = db.createObjectStore(IDBFS.DB_STORE_NAME);
          }
  
          if (!fileStore.indexNames.contains('timestamp')) {
            fileStore.createIndex('timestamp', 'timestamp', { unique: false });
          }
        };
        req.onsuccess = function() {
          db = req.result;
  
          // add to the cache
          IDBFS.dbs[name] = db;
          callback(null, db);
        };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },getLocalSet:function (mount, callback) {
        var entries = {};
  
        function isRealDir(p) {
          return p !== '.' && p !== '..';
        };
        function toAbsolute(root) {
          return function(p) {
            return PATH.join2(root, p);
          }
        };
  
        var check = FS.readdir(mount.mountpoint).filter(isRealDir).map(toAbsolute(mount.mountpoint));
  
        while (check.length) {
          var path = check.pop();
          var stat;
  
          try {
            stat = FS.stat(path);
          } catch (e) {
            return callback(e);
          }
  
          if (FS.isDir(stat.mode)) {
            check.push.apply(check, FS.readdir(path).filter(isRealDir).map(toAbsolute(path)));
          }
  
          entries[path] = { timestamp: stat.mtime };
        }
  
        return callback(null, { type: 'local', entries: entries });
      },getRemoteSet:function (mount, callback) {
        var entries = {};
  
        IDBFS.getDB(mount.mountpoint, function(err, db) {
          if (err) return callback(err);
  
          var transaction = db.transaction([IDBFS.DB_STORE_NAME], 'readonly');
          transaction.onerror = function(e) {
            callback(this.error);
            e.preventDefault();
          };
  
          var store = transaction.objectStore(IDBFS.DB_STORE_NAME);
          var index = store.index('timestamp');
  
          index.openKeyCursor().onsuccess = function(event) {
            var cursor = event.target.result;
  
            if (!cursor) {
              return callback(null, { type: 'remote', db: db, entries: entries });
            }
  
            entries[cursor.primaryKey] = { timestamp: cursor.key };
  
            cursor.continue();
          };
        });
      },loadLocalEntry:function (path, callback) {
        var stat, node;
  
        try {
          var lookup = FS.lookupPath(path);
          node = lookup.node;
          stat = FS.stat(path);
        } catch (e) {
          return callback(e);
        }
  
        if (FS.isDir(stat.mode)) {
          return callback(null, { timestamp: stat.mtime, mode: stat.mode });
        } else if (FS.isFile(stat.mode)) {
          // Performance consideration: storing a normal JavaScript array to a IndexedDB is much slower than storing a typed array.
          // Therefore always convert the file contents to a typed array first before writing the data to IndexedDB.
          node.contents = MEMFS.getFileDataAsTypedArray(node);
          return callback(null, { timestamp: stat.mtime, mode: stat.mode, contents: node.contents });
        } else {
          return callback(new Error('node type not supported'));
        }
      },storeLocalEntry:function (path, entry, callback) {
        try {
          if (FS.isDir(entry.mode)) {
            FS.mkdir(path, entry.mode);
          } else if (FS.isFile(entry.mode)) {
            FS.writeFile(path, entry.contents, { encoding: 'binary', canOwn: true });
          } else {
            return callback(new Error('node type not supported'));
          }
  
          FS.chmod(path, entry.mode);
          FS.utime(path, entry.timestamp, entry.timestamp);
        } catch (e) {
          return callback(e);
        }
  
        callback(null);
      },removeLocalEntry:function (path, callback) {
        try {
          var lookup = FS.lookupPath(path);
          var stat = FS.stat(path);
  
          if (FS.isDir(stat.mode)) {
            FS.rmdir(path);
          } else if (FS.isFile(stat.mode)) {
            FS.unlink(path);
          }
        } catch (e) {
          return callback(e);
        }
  
        callback(null);
      },loadRemoteEntry:function (store, path, callback) {
        var req = store.get(path);
        req.onsuccess = function(event) { callback(null, event.target.result); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },storeRemoteEntry:function (store, path, entry, callback) {
        var req = store.put(entry, path);
        req.onsuccess = function() { callback(null); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },removeRemoteEntry:function (store, path, callback) {
        var req = store.delete(path);
        req.onsuccess = function() { callback(null); };
        req.onerror = function(e) {
          callback(this.error);
          e.preventDefault();
        };
      },reconcile:function (src, dst, callback) {
        var total = 0;
  
        var create = [];
        Object.keys(src.entries).forEach(function (key) {
          var e = src.entries[key];
          var e2 = dst.entries[key];
          if (!e2 || e.timestamp > e2.timestamp) {
            create.push(key);
            total++;
          }
        });
  
        var remove = [];
        Object.keys(dst.entries).forEach(function (key) {
          var e = dst.entries[key];
          var e2 = src.entries[key];
          if (!e2) {
            remove.push(key);
            total++;
          }
        });
  
        if (!total) {
          return callback(null);
        }
  
        var errored = false;
        var completed = 0;
        var db = src.type === 'remote' ? src.db : dst.db;
        var transaction = db.transaction([IDBFS.DB_STORE_NAME], 'readwrite');
        var store = transaction.objectStore(IDBFS.DB_STORE_NAME);
  
        function done(err) {
          if (err) {
            if (!done.errored) {
              done.errored = true;
              return callback(err);
            }
            return;
          }
          if (++completed >= total) {
            return callback(null);
          }
        };
  
        transaction.onerror = function(e) {
          done(this.error);
          e.preventDefault();
        };
  
        // sort paths in ascending order so directory entries are created
        // before the files inside them
        create.sort().forEach(function (path) {
          if (dst.type === 'local') {
            IDBFS.loadRemoteEntry(store, path, function (err, entry) {
              if (err) return done(err);
              IDBFS.storeLocalEntry(path, entry, done);
            });
          } else {
            IDBFS.loadLocalEntry(path, function (err, entry) {
              if (err) return done(err);
              IDBFS.storeRemoteEntry(store, path, entry, done);
            });
          }
        });
  
        // sort paths in descending order so files are deleted before their
        // parent directories
        remove.sort().reverse().forEach(function(path) {
          if (dst.type === 'local') {
            IDBFS.removeLocalEntry(path, done);
          } else {
            IDBFS.removeRemoteEntry(store, path, done);
          }
        });
      }};
  
  var NODEFS={isWindows:false,staticInit:function () {
        NODEFS.isWindows = !!process.platform.match(/^win/);
      },mount:function (mount) {
        assert(ENVIRONMENT_IS_NODE);
        return NODEFS.createNode(null, '/', NODEFS.getMode(mount.opts.root), 0);
      },createNode:function (parent, name, mode, dev) {
        if (!FS.isDir(mode) && !FS.isFile(mode) && !FS.isLink(mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var node = FS.createNode(parent, name, mode);
        node.node_ops = NODEFS.node_ops;
        node.stream_ops = NODEFS.stream_ops;
        return node;
      },getMode:function (path) {
        var stat;
        try {
          stat = fs.lstatSync(path);
          if (NODEFS.isWindows) {
            // On Windows, directories return permission bits 'rw-rw-rw-', even though they have 'rwxrwxrwx', so
            // propagate write bits to execute bits.
            stat.mode = stat.mode | ((stat.mode & 146) >> 1);
          }
        } catch (e) {
          if (!e.code) throw e;
          throw new FS.ErrnoError(ERRNO_CODES[e.code]);
        }
        return stat.mode;
      },realPath:function (node) {
        var parts = [];
        while (node.parent !== node) {
          parts.push(node.name);
          node = node.parent;
        }
        parts.push(node.mount.opts.root);
        parts.reverse();
        return PATH.join.apply(null, parts);
      },flagsToPermissionStringMap:{0:"r",1:"r+",2:"r+",64:"r",65:"r+",66:"r+",129:"rx+",193:"rx+",514:"w+",577:"w",578:"w+",705:"wx",706:"wx+",1024:"a",1025:"a",1026:"a+",1089:"a",1090:"a+",1153:"ax",1154:"ax+",1217:"ax",1218:"ax+",4096:"rs",4098:"rs+"},flagsToPermissionString:function (flags) {
        flags &= ~0x200000 /*O_PATH*/; // Ignore this flag from musl, otherwise node.js fails to open the file.
        flags &= ~0x800 /*O_NONBLOCK*/; // Ignore this flag from musl, otherwise node.js fails to open the file.
        flags &= ~0x8000 /*O_LARGEFILE*/; // Ignore this flag from musl, otherwise node.js fails to open the file.
        flags &= ~0x80000 /*O_CLOEXEC*/; // Some applications may pass it; it makes no sense for a single process.
        if (flags in NODEFS.flagsToPermissionStringMap) {
          return NODEFS.flagsToPermissionStringMap[flags];
        } else {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
      },node_ops:{getattr:function (node) {
          var path = NODEFS.realPath(node);
          var stat;
          try {
            stat = fs.lstatSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          // node.js v0.10.20 doesn't report blksize and blocks on Windows. Fake them with default blksize of 4096.
          // See http://support.microsoft.com/kb/140365
          if (NODEFS.isWindows && !stat.blksize) {
            stat.blksize = 4096;
          }
          if (NODEFS.isWindows && !stat.blocks) {
            stat.blocks = (stat.size+stat.blksize-1)/stat.blksize|0;
          }
          return {
            dev: stat.dev,
            ino: stat.ino,
            mode: stat.mode,
            nlink: stat.nlink,
            uid: stat.uid,
            gid: stat.gid,
            rdev: stat.rdev,
            size: stat.size,
            atime: stat.atime,
            mtime: stat.mtime,
            ctime: stat.ctime,
            blksize: stat.blksize,
            blocks: stat.blocks
          };
        },setattr:function (node, attr) {
          var path = NODEFS.realPath(node);
          try {
            if (attr.mode !== undefined) {
              fs.chmodSync(path, attr.mode);
              // update the common node structure mode as well
              node.mode = attr.mode;
            }
            if (attr.timestamp !== undefined) {
              var date = new Date(attr.timestamp);
              fs.utimesSync(path, date, date);
            }
            if (attr.size !== undefined) {
              fs.truncateSync(path, attr.size);
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },lookup:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          var mode = NODEFS.getMode(path);
          return NODEFS.createNode(parent, name, mode);
        },mknod:function (parent, name, mode, dev) {
          var node = NODEFS.createNode(parent, name, mode, dev);
          // create the backing node for this in the fs root as well
          var path = NODEFS.realPath(node);
          try {
            if (FS.isDir(node.mode)) {
              fs.mkdirSync(path, node.mode);
            } else {
              fs.writeFileSync(path, '', { mode: node.mode });
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          return node;
        },rename:function (oldNode, newDir, newName) {
          var oldPath = NODEFS.realPath(oldNode);
          var newPath = PATH.join2(NODEFS.realPath(newDir), newName);
          try {
            fs.renameSync(oldPath, newPath);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },unlink:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          try {
            fs.unlinkSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },rmdir:function (parent, name) {
          var path = PATH.join2(NODEFS.realPath(parent), name);
          try {
            fs.rmdirSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },readdir:function (node) {
          var path = NODEFS.realPath(node);
          try {
            return fs.readdirSync(path);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },symlink:function (parent, newName, oldPath) {
          var newPath = PATH.join2(NODEFS.realPath(parent), newName);
          try {
            fs.symlinkSync(oldPath, newPath);
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },readlink:function (node) {
          var path = NODEFS.realPath(node);
          try {
            path = fs.readlinkSync(path);
            path = NODEJS_PATH.relative(NODEJS_PATH.resolve(node.mount.opts.root), path);
            return path;
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        }},stream_ops:{open:function (stream) {
          var path = NODEFS.realPath(stream.node);
          try {
            if (FS.isFile(stream.node.mode)) {
              stream.nfd = fs.openSync(path, NODEFS.flagsToPermissionString(stream.flags));
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },close:function (stream) {
          try {
            if (FS.isFile(stream.node.mode) && stream.nfd) {
              fs.closeSync(stream.nfd);
            }
          } catch (e) {
            if (!e.code) throw e;
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
        },read:function (stream, buffer, offset, length, position) {
          if (length === 0) return 0; // node errors on 0 length reads
          // FIXME this is terrible.
          var nbuffer = new Buffer(length);
          var res;
          try {
            res = fs.readSync(stream.nfd, nbuffer, 0, length, position);
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          if (res > 0) {
            for (var i = 0; i < res; i++) {
              buffer[offset + i] = nbuffer[i];
            }
          }
          return res;
        },write:function (stream, buffer, offset, length, position) {
          // FIXME this is terrible.
          var nbuffer = new Buffer(buffer.subarray(offset, offset + length));
          var res;
          try {
            res = fs.writeSync(stream.nfd, nbuffer, 0, length, position);
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES[e.code]);
          }
          return res;
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              try {
                var stat = fs.fstatSync(stream.nfd);
                position += stat.size;
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES[e.code]);
              }
            }
          }
  
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
  
          return position;
        }}};
  
  var WORKERFS={DIR_MODE:16895,FILE_MODE:33279,reader:null,mount:function (mount) {
        assert(ENVIRONMENT_IS_WORKER);
        if (!WORKERFS.reader) WORKERFS.reader = new FileReaderSync();
        var root = WORKERFS.createNode(null, '/', WORKERFS.DIR_MODE, 0);
        var createdParents = {};
        function ensureParent(path) {
          // return the parent node, creating subdirs as necessary
          var parts = path.split('/');
          var parent = root;
          for (var i = 0; i < parts.length-1; i++) {
            var curr = parts.slice(0, i+1).join('/');
            // Issue 4254: Using curr as a node name will prevent the node
            // from being found in FS.nameTable when FS.open is called on
            // a path which holds a child of this node,
            // given that all FS functions assume node names
            // are just their corresponding parts within their given path,
            // rather than incremental aggregates which include their parent's
            // directories.
            if (!createdParents[curr]) {
              createdParents[curr] = WORKERFS.createNode(parent, parts[i], WORKERFS.DIR_MODE, 0);
            }
            parent = createdParents[curr];
          }
          return parent;
        }
        function base(path) {
          var parts = path.split('/');
          return parts[parts.length-1];
        }
        // We also accept FileList here, by using Array.prototype
        Array.prototype.forEach.call(mount.opts["files"] || [], function(file) {
          WORKERFS.createNode(ensureParent(file.name), base(file.name), WORKERFS.FILE_MODE, 0, file, file.lastModifiedDate);
        });
        (mount.opts["blobs"] || []).forEach(function(obj) {
          WORKERFS.createNode(ensureParent(obj["name"]), base(obj["name"]), WORKERFS.FILE_MODE, 0, obj["data"]);
        });
        (mount.opts["packages"] || []).forEach(function(pack) {
          pack['metadata'].files.forEach(function(file) {
            var name = file.filename.substr(1); // remove initial slash
            WORKERFS.createNode(ensureParent(name), base(name), WORKERFS.FILE_MODE, 0, pack['blob'].slice(file.start, file.end));
          });
        });
        return root;
      },createNode:function (parent, name, mode, dev, contents, mtime) {
        var node = FS.createNode(parent, name, mode);
        node.mode = mode;
        node.node_ops = WORKERFS.node_ops;
        node.stream_ops = WORKERFS.stream_ops;
        node.timestamp = (mtime || new Date).getTime();
        assert(WORKERFS.FILE_MODE !== WORKERFS.DIR_MODE);
        if (mode === WORKERFS.FILE_MODE) {
          node.size = contents.size;
          node.contents = contents;
        } else {
          node.size = 4096;
          node.contents = {};
        }
        if (parent) {
          parent.contents[name] = node;
        }
        return node;
      },node_ops:{getattr:function (node) {
          return {
            dev: 1,
            ino: undefined,
            mode: node.mode,
            nlink: 1,
            uid: 0,
            gid: 0,
            rdev: undefined,
            size: node.size,
            atime: new Date(node.timestamp),
            mtime: new Date(node.timestamp),
            ctime: new Date(node.timestamp),
            blksize: 4096,
            blocks: Math.ceil(node.size / 4096),
          };
        },setattr:function (node, attr) {
          if (attr.mode !== undefined) {
            node.mode = attr.mode;
          }
          if (attr.timestamp !== undefined) {
            node.timestamp = attr.timestamp;
          }
        },lookup:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        },mknod:function (parent, name, mode, dev) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },rename:function (oldNode, newDir, newName) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },unlink:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },rmdir:function (parent, name) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },readdir:function (node) {
          var entries = ['.', '..'];
          for (var key in node.contents) {
            if (!node.contents.hasOwnProperty(key)) {
              continue;
            }
            entries.push(key);
          }
          return entries;
        },symlink:function (parent, newName, oldPath) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        },readlink:function (node) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }},stream_ops:{read:function (stream, buffer, offset, length, position) {
          if (position >= stream.node.size) return 0;
          var chunk = stream.node.contents.slice(position, position + length);
          var ab = WORKERFS.reader.readAsArrayBuffer(chunk);
          buffer.set(new Uint8Array(ab), offset);
          return chunk.size;
        },write:function (stream, buffer, offset, length, position) {
          throw new FS.ErrnoError(ERRNO_CODES.EIO);
        },llseek:function (stream, offset, whence) {
          var position = offset;
          if (whence === 1) {  // SEEK_CUR.
            position += stream.position;
          } else if (whence === 2) {  // SEEK_END.
            if (FS.isFile(stream.node.mode)) {
              position += stream.node.size;
            }
          }
          if (position < 0) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          return position;
        }}};
  
  var _stdin=STATICTOP; STATICTOP += 16;;
  
  var _stdout=STATICTOP; STATICTOP += 16;;
  
  var _stderr=STATICTOP; STATICTOP += 16;;var FS={root:null,mounts:[],devices:[null],streams:[],nextInode:1,nameTable:null,currentPath:"/",initialized:false,ignorePermissions:true,trackingDelegate:{},tracking:{openFlags:{READ:1,WRITE:2}},ErrnoError:null,genericErrors:{},filesystems:null,syncFSRequests:0,handleFSError:function (e) {
        if (!(e instanceof FS.ErrnoError)) throw e + ' : ' + stackTrace();
        return ___setErrNo(e.errno);
      },lookupPath:function (path, opts) {
        path = PATH.resolve(FS.cwd(), path);
        opts = opts || {};
  
        if (!path) return { path: '', node: null };
  
        var defaults = {
          follow_mount: true,
          recurse_count: 0
        };
        for (var key in defaults) {
          if (opts[key] === undefined) {
            opts[key] = defaults[key];
          }
        }
  
        if (opts.recurse_count > 8) {  // max recursive lookup of 8
          throw new FS.ErrnoError(ERRNO_CODES.ELOOP);
        }
  
        // split the path
        var parts = PATH.normalizeArray(path.split('/').filter(function(p) {
          return !!p;
        }), false);
  
        // start at the root
        var current = FS.root;
        var current_path = '/';
  
        for (var i = 0; i < parts.length; i++) {
          var islast = (i === parts.length-1);
          if (islast && opts.parent) {
            // stop resolving
            break;
          }
  
          current = FS.lookupNode(current, parts[i]);
          current_path = PATH.join2(current_path, parts[i]);
  
          // jump to the mount's root node if this is a mountpoint
          if (FS.isMountpoint(current)) {
            if (!islast || (islast && opts.follow_mount)) {
              current = current.mounted.root;
            }
          }
  
          // by default, lookupPath will not follow a symlink if it is the final path component.
          // setting opts.follow = true will override this behavior.
          if (!islast || opts.follow) {
            var count = 0;
            while (FS.isLink(current.mode)) {
              var link = FS.readlink(current_path);
              current_path = PATH.resolve(PATH.dirname(current_path), link);
  
              var lookup = FS.lookupPath(current_path, { recurse_count: opts.recurse_count });
              current = lookup.node;
  
              if (count++ > 40) {  // limit max consecutive symlinks to 40 (SYMLOOP_MAX).
                throw new FS.ErrnoError(ERRNO_CODES.ELOOP);
              }
            }
          }
        }
  
        return { path: current_path, node: current };
      },getPath:function (node) {
        var path;
        while (true) {
          if (FS.isRoot(node)) {
            var mount = node.mount.mountpoint;
            if (!path) return mount;
            return mount[mount.length-1] !== '/' ? mount + '/' + path : mount + path;
          }
          path = path ? node.name + '/' + path : node.name;
          node = node.parent;
        }
      },hashName:function (parentid, name) {
        var hash = 0;
  
  
        for (var i = 0; i < name.length; i++) {
          hash = ((hash << 5) - hash + name.charCodeAt(i)) | 0;
        }
        return ((parentid + hash) >>> 0) % FS.nameTable.length;
      },hashAddNode:function (node) {
        var hash = FS.hashName(node.parent.id, node.name);
        node.name_next = FS.nameTable[hash];
        FS.nameTable[hash] = node;
      },hashRemoveNode:function (node) {
        var hash = FS.hashName(node.parent.id, node.name);
        if (FS.nameTable[hash] === node) {
          FS.nameTable[hash] = node.name_next;
        } else {
          var current = FS.nameTable[hash];
          while (current) {
            if (current.name_next === node) {
              current.name_next = node.name_next;
              break;
            }
            current = current.name_next;
          }
        }
      },lookupNode:function (parent, name) {
        var err = FS.mayLookup(parent);
        if (err) {
          throw new FS.ErrnoError(err, parent);
        }
        var hash = FS.hashName(parent.id, name);
        for (var node = FS.nameTable[hash]; node; node = node.name_next) {
          var nodeName = node.name;
          if (node.parent.id === parent.id && nodeName === name) {
            return node;
          }
        }
        // if we failed to find it in the cache, call into the VFS
        return FS.lookup(parent, name);
      },createNode:function (parent, name, mode, rdev) {
        if (!FS.FSNode) {
          FS.FSNode = function(parent, name, mode, rdev) {
            if (!parent) {
              parent = this;  // root node sets parent to itself
            }
            this.parent = parent;
            this.mount = parent.mount;
            this.mounted = null;
            this.id = FS.nextInode++;
            this.name = name;
            this.mode = mode;
            this.node_ops = {};
            this.stream_ops = {};
            this.rdev = rdev;
          };
  
          FS.FSNode.prototype = {};
  
          // compatibility
          var readMode = 292 | 73;
          var writeMode = 146;
  
          // NOTE we must use Object.defineProperties instead of individual calls to
          // Object.defineProperty in order to make closure compiler happy
          Object.defineProperties(FS.FSNode.prototype, {
            read: {
              get: function() { return (this.mode & readMode) === readMode; },
              set: function(val) { val ? this.mode |= readMode : this.mode &= ~readMode; }
            },
            write: {
              get: function() { return (this.mode & writeMode) === writeMode; },
              set: function(val) { val ? this.mode |= writeMode : this.mode &= ~writeMode; }
            },
            isFolder: {
              get: function() { return FS.isDir(this.mode); }
            },
            isDevice: {
              get: function() { return FS.isChrdev(this.mode); }
            }
          });
        }
  
        var node = new FS.FSNode(parent, name, mode, rdev);
  
        FS.hashAddNode(node);
  
        return node;
      },destroyNode:function (node) {
        FS.hashRemoveNode(node);
      },isRoot:function (node) {
        return node === node.parent;
      },isMountpoint:function (node) {
        return !!node.mounted;
      },isFile:function (mode) {
        return (mode & 61440) === 32768;
      },isDir:function (mode) {
        return (mode & 61440) === 16384;
      },isLink:function (mode) {
        return (mode & 61440) === 40960;
      },isChrdev:function (mode) {
        return (mode & 61440) === 8192;
      },isBlkdev:function (mode) {
        return (mode & 61440) === 24576;
      },isFIFO:function (mode) {
        return (mode & 61440) === 4096;
      },isSocket:function (mode) {
        return (mode & 49152) === 49152;
      },flagModes:{"r":0,"rs":1052672,"r+":2,"w":577,"wx":705,"xw":705,"w+":578,"wx+":706,"xw+":706,"a":1089,"ax":1217,"xa":1217,"a+":1090,"ax+":1218,"xa+":1218},modeStringToFlags:function (str) {
        var flags = FS.flagModes[str];
        if (typeof flags === 'undefined') {
          throw new Error('Unknown file open mode: ' + str);
        }
        return flags;
      },flagsToPermissionString:function (flag) {
        var perms = ['r', 'w', 'rw'][flag & 3];
        if ((flag & 512)) {
          perms += 'w';
        }
        return perms;
      },nodePermissions:function (node, perms) {
        if (FS.ignorePermissions) {
          return 0;
        }
        // return 0 if any user, group or owner bits are set.
        if (perms.indexOf('r') !== -1 && !(node.mode & 292)) {
          return ERRNO_CODES.EACCES;
        } else if (perms.indexOf('w') !== -1 && !(node.mode & 146)) {
          return ERRNO_CODES.EACCES;
        } else if (perms.indexOf('x') !== -1 && !(node.mode & 73)) {
          return ERRNO_CODES.EACCES;
        }
        return 0;
      },mayLookup:function (dir) {
        var err = FS.nodePermissions(dir, 'x');
        if (err) return err;
        if (!dir.node_ops.lookup) return ERRNO_CODES.EACCES;
        return 0;
      },mayCreate:function (dir, name) {
        try {
          var node = FS.lookupNode(dir, name);
          return ERRNO_CODES.EEXIST;
        } catch (e) {
        }
        return FS.nodePermissions(dir, 'wx');
      },mayDelete:function (dir, name, isdir) {
        var node;
        try {
          node = FS.lookupNode(dir, name);
        } catch (e) {
          return e.errno;
        }
        var err = FS.nodePermissions(dir, 'wx');
        if (err) {
          return err;
        }
        if (isdir) {
          if (!FS.isDir(node.mode)) {
            return ERRNO_CODES.ENOTDIR;
          }
          if (FS.isRoot(node) || FS.getPath(node) === FS.cwd()) {
            return ERRNO_CODES.EBUSY;
          }
        } else {
          if (FS.isDir(node.mode)) {
            return ERRNO_CODES.EISDIR;
          }
        }
        return 0;
      },mayOpen:function (node, flags) {
        if (!node) {
          return ERRNO_CODES.ENOENT;
        }
        if (FS.isLink(node.mode)) {
          return ERRNO_CODES.ELOOP;
        } else if (FS.isDir(node.mode)) {
          if (FS.flagsToPermissionString(flags) !== 'r' || // opening for write
              (flags & 512)) { // TODO: check for O_SEARCH? (== search for dir only)
            return ERRNO_CODES.EISDIR;
          }
        }
        return FS.nodePermissions(node, FS.flagsToPermissionString(flags));
      },MAX_OPEN_FDS:4096,nextfd:function (fd_start, fd_end) {
        fd_start = fd_start || 0;
        fd_end = fd_end || FS.MAX_OPEN_FDS;
        for (var fd = fd_start; fd <= fd_end; fd++) {
          if (!FS.streams[fd]) {
            return fd;
          }
        }
        throw new FS.ErrnoError(ERRNO_CODES.EMFILE);
      },getStream:function (fd) {
        return FS.streams[fd];
      },createStream:function (stream, fd_start, fd_end) {
        if (!FS.FSStream) {
          FS.FSStream = function(){};
          FS.FSStream.prototype = {};
          // compatibility
          Object.defineProperties(FS.FSStream.prototype, {
            object: {
              get: function() { return this.node; },
              set: function(val) { this.node = val; }
            },
            isRead: {
              get: function() { return (this.flags & 2097155) !== 1; }
            },
            isWrite: {
              get: function() { return (this.flags & 2097155) !== 0; }
            },
            isAppend: {
              get: function() { return (this.flags & 1024); }
            }
          });
        }
        // clone it, so we can return an instance of FSStream
        var newStream = new FS.FSStream();
        for (var p in stream) {
          newStream[p] = stream[p];
        }
        stream = newStream;
        var fd = FS.nextfd(fd_start, fd_end);
        stream.fd = fd;
        FS.streams[fd] = stream;
        return stream;
      },closeStream:function (fd) {
        FS.streams[fd] = null;
      },chrdev_stream_ops:{open:function (stream) {
          var device = FS.getDevice(stream.node.rdev);
          // override node's stream ops with the device's
          stream.stream_ops = device.stream_ops;
          // forward the open call
          if (stream.stream_ops.open) {
            stream.stream_ops.open(stream);
          }
        },llseek:function () {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }},major:function (dev) {
        return ((dev) >> 8);
      },minor:function (dev) {
        return ((dev) & 0xff);
      },makedev:function (ma, mi) {
        return ((ma) << 8 | (mi));
      },registerDevice:function (dev, ops) {
        FS.devices[dev] = { stream_ops: ops };
      },getDevice:function (dev) {
        return FS.devices[dev];
      },getMounts:function (mount) {
        var mounts = [];
        var check = [mount];
  
        while (check.length) {
          var m = check.pop();
  
          mounts.push(m);
  
          check.push.apply(check, m.mounts);
        }
  
        return mounts;
      },syncfs:function (populate, callback) {
        if (typeof(populate) === 'function') {
          callback = populate;
          populate = false;
        }
  
        FS.syncFSRequests++;
  
        if (FS.syncFSRequests > 1) {
          console.log('warning: ' + FS.syncFSRequests + ' FS.syncfs operations in flight at once, probably just doing extra work');
        }
  
        var mounts = FS.getMounts(FS.root.mount);
        var completed = 0;
  
        function doCallback(err) {
          assert(FS.syncFSRequests > 0);
          FS.syncFSRequests--;
          return callback(err);
        }
  
        function done(err) {
          if (err) {
            if (!done.errored) {
              done.errored = true;
              return doCallback(err);
            }
            return;
          }
          if (++completed >= mounts.length) {
            doCallback(null);
          }
        };
  
        // sync all mounts
        mounts.forEach(function (mount) {
          if (!mount.type.syncfs) {
            return done(null);
          }
          mount.type.syncfs(mount, populate, done);
        });
      },mount:function (type, opts, mountpoint) {
        var root = mountpoint === '/';
        var pseudo = !mountpoint;
        var node;
  
        if (root && FS.root) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        } else if (!root && !pseudo) {
          var lookup = FS.lookupPath(mountpoint, { follow_mount: false });
  
          mountpoint = lookup.path;  // use the absolute path
          node = lookup.node;
  
          if (FS.isMountpoint(node)) {
            throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
          }
  
          if (!FS.isDir(node.mode)) {
            throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
          }
        }
  
        var mount = {
          type: type,
          opts: opts,
          mountpoint: mountpoint,
          mounts: []
        };
  
        // create a root node for the fs
        var mountRoot = type.mount(mount);
        mountRoot.mount = mount;
        mount.root = mountRoot;
  
        if (root) {
          FS.root = mountRoot;
        } else if (node) {
          // set as a mountpoint
          node.mounted = mount;
  
          // add the new mount to the current mount's children
          if (node.mount) {
            node.mount.mounts.push(mount);
          }
        }
  
        return mountRoot;
      },unmount:function (mountpoint) {
        var lookup = FS.lookupPath(mountpoint, { follow_mount: false });
  
        if (!FS.isMountpoint(lookup.node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
  
        // destroy the nodes for this mount, and all its child mounts
        var node = lookup.node;
        var mount = node.mounted;
        var mounts = FS.getMounts(mount);
  
        Object.keys(FS.nameTable).forEach(function (hash) {
          var current = FS.nameTable[hash];
  
          while (current) {
            var next = current.name_next;
  
            if (mounts.indexOf(current.mount) !== -1) {
              FS.destroyNode(current);
            }
  
            current = next;
          }
        });
  
        // no longer a mountpoint
        node.mounted = null;
  
        // remove this mount from the child mounts
        var idx = node.mount.mounts.indexOf(mount);
        assert(idx !== -1);
        node.mount.mounts.splice(idx, 1);
      },lookup:function (parent, name) {
        return parent.node_ops.lookup(parent, name);
      },mknod:function (path, mode, dev) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        if (!name || name === '.' || name === '..') {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var err = FS.mayCreate(parent, name);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.mknod) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return parent.node_ops.mknod(parent, name, mode, dev);
      },create:function (path, mode) {
        mode = mode !== undefined ? mode : 438 /* 0666 */;
        mode &= 4095;
        mode |= 32768;
        return FS.mknod(path, mode, 0);
      },mkdir:function (path, mode) {
        mode = mode !== undefined ? mode : 511 /* 0777 */;
        mode &= 511 | 512;
        mode |= 16384;
        return FS.mknod(path, mode, 0);
      },mkdirTree:function (path, mode) {
        var dirs = path.split('/');
        var d = '';
        for (var i = 0; i < dirs.length; ++i) {
          if (!dirs[i]) continue;
          d += '/' + dirs[i];
          try {
            FS.mkdir(d, mode);
          } catch(e) {
            if (e.errno != ERRNO_CODES.EEXIST) throw e;
          }
        }
      },mkdev:function (path, mode, dev) {
        if (typeof(dev) === 'undefined') {
          dev = mode;
          mode = 438 /* 0666 */;
        }
        mode |= 8192;
        return FS.mknod(path, mode, dev);
      },symlink:function (oldpath, newpath) {
        if (!PATH.resolve(oldpath)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        var lookup = FS.lookupPath(newpath, { parent: true });
        var parent = lookup.node;
        if (!parent) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        var newname = PATH.basename(newpath);
        var err = FS.mayCreate(parent, newname);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.symlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return parent.node_ops.symlink(parent, newname, oldpath);
      },rename:function (old_path, new_path) {
        var old_dirname = PATH.dirname(old_path);
        var new_dirname = PATH.dirname(new_path);
        var old_name = PATH.basename(old_path);
        var new_name = PATH.basename(new_path);
        // parents must exist
        var lookup, old_dir, new_dir;
        try {
          lookup = FS.lookupPath(old_path, { parent: true });
          old_dir = lookup.node;
          lookup = FS.lookupPath(new_path, { parent: true });
          new_dir = lookup.node;
        } catch (e) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        if (!old_dir || !new_dir) throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        // need to be part of the same mount
        if (old_dir.mount !== new_dir.mount) {
          throw new FS.ErrnoError(ERRNO_CODES.EXDEV);
        }
        // source must exist
        var old_node = FS.lookupNode(old_dir, old_name);
        // old path should not be an ancestor of the new path
        var relative = PATH.relative(old_path, new_dirname);
        if (relative.charAt(0) !== '.') {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        // new path should not be an ancestor of the old path
        relative = PATH.relative(new_path, old_dirname);
        if (relative.charAt(0) !== '.') {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTEMPTY);
        }
        // see if the new path already exists
        var new_node;
        try {
          new_node = FS.lookupNode(new_dir, new_name);
        } catch (e) {
          // not fatal
        }
        // early out if nothing needs to change
        if (old_node === new_node) {
          return;
        }
        // we'll need to delete the old entry
        var isdir = FS.isDir(old_node.mode);
        var err = FS.mayDelete(old_dir, old_name, isdir);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        // need delete permissions if we'll be overwriting.
        // need create permissions if new doesn't already exist.
        err = new_node ?
          FS.mayDelete(new_dir, new_name, isdir) :
          FS.mayCreate(new_dir, new_name);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!old_dir.node_ops.rename) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(old_node) || (new_node && FS.isMountpoint(new_node))) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        // if we are going to change the parent, check write permissions
        if (new_dir !== old_dir) {
          err = FS.nodePermissions(old_dir, 'w');
          if (err) {
            throw new FS.ErrnoError(err);
          }
        }
        try {
          if (FS.trackingDelegate['willMovePath']) {
            FS.trackingDelegate['willMovePath'](old_path, new_path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
        }
        // remove the node from the lookup hash
        FS.hashRemoveNode(old_node);
        // do the underlying fs rename
        try {
          old_dir.node_ops.rename(old_node, new_dir, new_name);
        } catch (e) {
          throw e;
        } finally {
          // add the node back to the hash (in case node_ops.rename
          // changed its name)
          FS.hashAddNode(old_node);
        }
        try {
          if (FS.trackingDelegate['onMovePath']) FS.trackingDelegate['onMovePath'](old_path, new_path);
        } catch(e) {
          console.log("FS.trackingDelegate['onMovePath']('"+old_path+"', '"+new_path+"') threw an exception: " + e.message);
        }
      },rmdir:function (path) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        var node = FS.lookupNode(parent, name);
        var err = FS.mayDelete(parent, name, true);
        if (err) {
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.rmdir) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        try {
          if (FS.trackingDelegate['willDeletePath']) {
            FS.trackingDelegate['willDeletePath'](path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
        }
        parent.node_ops.rmdir(parent, name);
        FS.destroyNode(node);
        try {
          if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
        } catch(e) {
          console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
        }
      },readdir:function (path) {
        var lookup = FS.lookupPath(path, { follow: true });
        var node = lookup.node;
        if (!node.node_ops.readdir) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        return node.node_ops.readdir(node);
      },unlink:function (path) {
        var lookup = FS.lookupPath(path, { parent: true });
        var parent = lookup.node;
        var name = PATH.basename(path);
        var node = FS.lookupNode(parent, name);
        var err = FS.mayDelete(parent, name, false);
        if (err) {
          // According to POSIX, we should map EISDIR to EPERM, but
          // we instead do what Linux does (and we must, as we use
          // the musl linux libc).
          throw new FS.ErrnoError(err);
        }
        if (!parent.node_ops.unlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isMountpoint(node)) {
          throw new FS.ErrnoError(ERRNO_CODES.EBUSY);
        }
        try {
          if (FS.trackingDelegate['willDeletePath']) {
            FS.trackingDelegate['willDeletePath'](path);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['willDeletePath']('"+path+"') threw an exception: " + e.message);
        }
        parent.node_ops.unlink(parent, name);
        FS.destroyNode(node);
        try {
          if (FS.trackingDelegate['onDeletePath']) FS.trackingDelegate['onDeletePath'](path);
        } catch(e) {
          console.log("FS.trackingDelegate['onDeletePath']('"+path+"') threw an exception: " + e.message);
        }
      },readlink:function (path) {
        var lookup = FS.lookupPath(path);
        var link = lookup.node;
        if (!link) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        if (!link.node_ops.readlink) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        return PATH.resolve(FS.getPath(link.parent), link.node_ops.readlink(link));
      },stat:function (path, dontFollow) {
        var lookup = FS.lookupPath(path, { follow: !dontFollow });
        var node = lookup.node;
        if (!node) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        if (!node.node_ops.getattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        return node.node_ops.getattr(node);
      },lstat:function (path) {
        return FS.stat(path, true);
      },chmod:function (path, mode, dontFollow) {
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: !dontFollow });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        node.node_ops.setattr(node, {
          mode: (mode & 4095) | (node.mode & ~4095),
          timestamp: Date.now()
        });
      },lchmod:function (path, mode) {
        FS.chmod(path, mode, true);
      },fchmod:function (fd, mode) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        FS.chmod(stream.node, mode);
      },chown:function (path, uid, gid, dontFollow) {
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: !dontFollow });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        node.node_ops.setattr(node, {
          timestamp: Date.now()
          // we ignore the uid / gid for now
        });
      },lchown:function (path, uid, gid) {
        FS.chown(path, uid, gid, true);
      },fchown:function (fd, uid, gid) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        FS.chown(stream.node, uid, gid);
      },truncate:function (path, len) {
        if (len < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var node;
        if (typeof path === 'string') {
          var lookup = FS.lookupPath(path, { follow: true });
          node = lookup.node;
        } else {
          node = path;
        }
        if (!node.node_ops.setattr) {
          throw new FS.ErrnoError(ERRNO_CODES.EPERM);
        }
        if (FS.isDir(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!FS.isFile(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var err = FS.nodePermissions(node, 'w');
        if (err) {
          throw new FS.ErrnoError(err);
        }
        node.node_ops.setattr(node, {
          size: len,
          timestamp: Date.now()
        });
      },ftruncate:function (fd, len) {
        var stream = FS.getStream(fd);
        if (!stream) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        FS.truncate(stream.node, len);
      },utime:function (path, atime, mtime) {
        var lookup = FS.lookupPath(path, { follow: true });
        var node = lookup.node;
        node.node_ops.setattr(node, {
          timestamp: Math.max(atime, mtime)
        });
      },open:function (path, flags, mode, fd_start, fd_end) {
        if (path === "") {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        flags = typeof flags === 'string' ? FS.modeStringToFlags(flags) : flags;
        mode = typeof mode === 'undefined' ? 438 /* 0666 */ : mode;
        if ((flags & 64)) {
          mode = (mode & 4095) | 32768;
        } else {
          mode = 0;
        }
        var node;
        if (typeof path === 'object') {
          node = path;
        } else {
          path = PATH.normalize(path);
          try {
            var lookup = FS.lookupPath(path, {
              follow: !(flags & 131072)
            });
            node = lookup.node;
          } catch (e) {
            // ignore
          }
        }
        // perhaps we need to create the node
        var created = false;
        if ((flags & 64)) {
          if (node) {
            // if O_CREAT and O_EXCL are set, error out if the node already exists
            if ((flags & 128)) {
              throw new FS.ErrnoError(ERRNO_CODES.EEXIST);
            }
          } else {
            // node doesn't exist, try to create it
            node = FS.mknod(path, mode, 0);
            created = true;
          }
        }
        if (!node) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        // can't truncate a device
        if (FS.isChrdev(node.mode)) {
          flags &= ~512;
        }
        // if asked only for a directory, then this must be one
        if ((flags & 65536) && !FS.isDir(node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        // check permissions, if this is not a file we just created now (it is ok to
        // create and write to a file with read-only permissions; it is read-only
        // for later use)
        if (!created) {
          var err = FS.mayOpen(node, flags);
          if (err) {
            throw new FS.ErrnoError(err);
          }
        }
        // do truncation if necessary
        if ((flags & 512)) {
          FS.truncate(node, 0);
        }
        // we've already handled these, don't pass down to the underlying vfs
        flags &= ~(128 | 512);
  
        // register the stream with the filesystem
        var stream = FS.createStream({
          node: node,
          path: FS.getPath(node),  // we want the absolute path to the node
          flags: flags,
          seekable: true,
          position: 0,
          stream_ops: node.stream_ops,
          // used by the file family libc calls (fopen, fwrite, ferror, etc.)
          ungotten: [],
          error: false
        }, fd_start, fd_end);
        // call the new stream's open function
        if (stream.stream_ops.open) {
          stream.stream_ops.open(stream);
        }
        if (Module['logReadFiles'] && !(flags & 1)) {
          if (!FS.readFiles) FS.readFiles = {};
          if (!(path in FS.readFiles)) {
            FS.readFiles[path] = 1;
            Module['printErr']('read file: ' + path);
          }
        }
        try {
          if (FS.trackingDelegate['onOpenFile']) {
            var trackingFlags = 0;
            if ((flags & 2097155) !== 1) {
              trackingFlags |= FS.tracking.openFlags.READ;
            }
            if ((flags & 2097155) !== 0) {
              trackingFlags |= FS.tracking.openFlags.WRITE;
            }
            FS.trackingDelegate['onOpenFile'](path, trackingFlags);
          }
        } catch(e) {
          console.log("FS.trackingDelegate['onOpenFile']('"+path+"', flags) threw an exception: " + e.message);
        }
        return stream;
      },close:function (stream) {
        if (stream.getdents) stream.getdents = null; // free readdir state
        try {
          if (stream.stream_ops.close) {
            stream.stream_ops.close(stream);
          }
        } catch (e) {
          throw e;
        } finally {
          FS.closeStream(stream.fd);
        }
      },llseek:function (stream, offset, whence) {
        if (!stream.seekable || !stream.stream_ops.llseek) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        stream.position = stream.stream_ops.llseek(stream, offset, whence);
        stream.ungotten = [];
        return stream.position;
      },read:function (stream, buffer, offset, length, position) {
        if (length < 0 || position < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 1) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (FS.isDir(stream.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!stream.stream_ops.read) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        var seeking = true;
        if (typeof position === 'undefined') {
          position = stream.position;
          seeking = false;
        } else if (!stream.seekable) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        var bytesRead = stream.stream_ops.read(stream, buffer, offset, length, position);
        if (!seeking) stream.position += bytesRead;
        return bytesRead;
      },write:function (stream, buffer, offset, length, position, canOwn) {
        if (length < 0 || position < 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (FS.isDir(stream.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.EISDIR);
        }
        if (!stream.stream_ops.write) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if (stream.flags & 1024) {
          // seek to the end before writing in append mode
          FS.llseek(stream, 0, 2);
        }
        var seeking = true;
        if (typeof position === 'undefined') {
          position = stream.position;
          seeking = false;
        } else if (!stream.seekable) {
          throw new FS.ErrnoError(ERRNO_CODES.ESPIPE);
        }
        var bytesWritten = stream.stream_ops.write(stream, buffer, offset, length, position, canOwn);
        if (!seeking) stream.position += bytesWritten;
        try {
          if (stream.path && FS.trackingDelegate['onWriteToFile']) FS.trackingDelegate['onWriteToFile'](stream.path);
        } catch(e) {
          console.log("FS.trackingDelegate['onWriteToFile']('"+path+"') threw an exception: " + e.message);
        }
        return bytesWritten;
      },allocate:function (stream, offset, length) {
        if (offset < 0 || length <= 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
        }
        if ((stream.flags & 2097155) === 0) {
          throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        }
        if (!FS.isFile(stream.node.mode) && !FS.isDir(stream.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
        }
        if (!stream.stream_ops.allocate) {
          throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
        }
        stream.stream_ops.allocate(stream, offset, length);
      },mmap:function (stream, buffer, offset, length, position, prot, flags) {
        // TODO if PROT is PROT_WRITE, make sure we have write access
        if ((stream.flags & 2097155) === 1) {
          throw new FS.ErrnoError(ERRNO_CODES.EACCES);
        }
        if (!stream.stream_ops.mmap) {
          throw new FS.ErrnoError(ERRNO_CODES.ENODEV);
        }
        return stream.stream_ops.mmap(stream, buffer, offset, length, position, prot, flags);
      },msync:function (stream, buffer, offset, length, mmapFlags) {
        if (!stream || !stream.stream_ops.msync) {
          return 0;
        }
        return stream.stream_ops.msync(stream, buffer, offset, length, mmapFlags);
      },munmap:function (stream) {
        return 0;
      },ioctl:function (stream, cmd, arg) {
        if (!stream.stream_ops.ioctl) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTTY);
        }
        return stream.stream_ops.ioctl(stream, cmd, arg);
      },readFile:function (path, opts) {
        opts = opts || {};
        opts.flags = opts.flags || 'r';
        opts.encoding = opts.encoding || 'binary';
        if (opts.encoding !== 'utf8' && opts.encoding !== 'binary') {
          throw new Error('Invalid encoding type "' + opts.encoding + '"');
        }
        var ret;
        var stream = FS.open(path, opts.flags);
        var stat = FS.stat(path);
        var length = stat.size;
        var buf = new Uint8Array(length);
        FS.read(stream, buf, 0, length, 0);
        if (opts.encoding === 'utf8') {
          ret = UTF8ArrayToString(buf, 0);
        } else if (opts.encoding === 'binary') {
          ret = buf;
        }
        FS.close(stream);
        return ret;
      },writeFile:function (path, data, opts) {
        opts = opts || {};
        opts.flags = opts.flags || 'w';
        opts.encoding = opts.encoding || 'utf8';
        if (opts.encoding !== 'utf8' && opts.encoding !== 'binary') {
          throw new Error('Invalid encoding type "' + opts.encoding + '"');
        }
        var stream = FS.open(path, opts.flags, opts.mode);
        if (opts.encoding === 'utf8') {
          var buf = new Uint8Array(lengthBytesUTF8(data)+1);
          var actualNumBytes = stringToUTF8Array(data, buf, 0, buf.length);
          FS.write(stream, buf, 0, actualNumBytes, 0, opts.canOwn);
        } else if (opts.encoding === 'binary') {
          FS.write(stream, data, 0, data.length, 0, opts.canOwn);
        }
        FS.close(stream);
      },cwd:function () {
        return FS.currentPath;
      },chdir:function (path) {
        var lookup = FS.lookupPath(path, { follow: true });
        if (lookup.node === null) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOENT);
        }
        if (!FS.isDir(lookup.node.mode)) {
          throw new FS.ErrnoError(ERRNO_CODES.ENOTDIR);
        }
        var err = FS.nodePermissions(lookup.node, 'x');
        if (err) {
          throw new FS.ErrnoError(err);
        }
        FS.currentPath = lookup.path;
      },createDefaultDirectories:function () {
        FS.mkdir('/tmp');
        FS.mkdir('/home');
        FS.mkdir('/home/web_user');
      },createDefaultDevices:function () {
        // create /dev
        FS.mkdir('/dev');
        // setup /dev/null
        FS.registerDevice(FS.makedev(1, 3), {
          read: function() { return 0; },
          write: function(stream, buffer, offset, length, pos) { return length; }
        });
        FS.mkdev('/dev/null', FS.makedev(1, 3));
        // setup /dev/tty and /dev/tty1
        // stderr needs to print output using Module['printErr']
        // so we register a second tty just for it.
        TTY.register(FS.makedev(5, 0), TTY.default_tty_ops);
        TTY.register(FS.makedev(6, 0), TTY.default_tty1_ops);
        FS.mkdev('/dev/tty', FS.makedev(5, 0));
        FS.mkdev('/dev/tty1', FS.makedev(6, 0));
        // setup /dev/[u]random
        var random_device;
        if (typeof crypto !== 'undefined') {
          // for modern web browsers
          var randomBuffer = new Uint8Array(1);
          random_device = function() { crypto.getRandomValues(randomBuffer); return randomBuffer[0]; };
        } else if (ENVIRONMENT_IS_NODE) {
          // for nodejs
          random_device = function() { return require('crypto').randomBytes(1)[0]; };
        } else {
          // default for ES5 platforms
          random_device = function() { return (Math.random()*256)|0; };
        }
        FS.createDevice('/dev', 'random', random_device);
        FS.createDevice('/dev', 'urandom', random_device);
        // we're not going to emulate the actual shm device,
        // just create the tmp dirs that reside in it commonly
        FS.mkdir('/dev/shm');
        FS.mkdir('/dev/shm/tmp');
      },createSpecialDirectories:function () {
        // create /proc/self/fd which allows /proc/self/fd/6 => readlink gives the name of the stream for fd 6 (see test_unistd_ttyname)
        FS.mkdir('/proc');
        FS.mkdir('/proc/self');
        FS.mkdir('/proc/self/fd');
        FS.mount({
          mount: function() {
            var node = FS.createNode('/proc/self', 'fd', 16384 | 511 /* 0777 */, 73);
            node.node_ops = {
              lookup: function(parent, name) {
                var fd = +name;
                var stream = FS.getStream(fd);
                if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
                var ret = {
                  parent: null,
                  mount: { mountpoint: 'fake' },
                  node_ops: { readlink: function() { return stream.path } }
                };
                ret.parent = ret; // make it look like a simple root node
                return ret;
              }
            };
            return node;
          }
        }, {}, '/proc/self/fd');
      },createStandardStreams:function () {
        // TODO deprecate the old functionality of a single
        // input / output callback and that utilizes FS.createDevice
        // and instead require a unique set of stream ops
  
        // by default, we symlink the standard streams to the
        // default tty devices. however, if the standard streams
        // have been overwritten we create a unique device for
        // them instead.
        if (Module['stdin']) {
          FS.createDevice('/dev', 'stdin', Module['stdin']);
        } else {
          FS.symlink('/dev/tty', '/dev/stdin');
        }
        if (Module['stdout']) {
          FS.createDevice('/dev', 'stdout', null, Module['stdout']);
        } else {
          FS.symlink('/dev/tty', '/dev/stdout');
        }
        if (Module['stderr']) {
          FS.createDevice('/dev', 'stderr', null, Module['stderr']);
        } else {
          FS.symlink('/dev/tty1', '/dev/stderr');
        }
  
        // open default streams for the stdin, stdout and stderr devices
        var stdin = FS.open('/dev/stdin', 'r');
        assert(stdin.fd === 0, 'invalid handle for stdin (' + stdin.fd + ')');
  
        var stdout = FS.open('/dev/stdout', 'w');
        assert(stdout.fd === 1, 'invalid handle for stdout (' + stdout.fd + ')');
  
        var stderr = FS.open('/dev/stderr', 'w');
        assert(stderr.fd === 2, 'invalid handle for stderr (' + stderr.fd + ')');
      },ensureErrnoError:function () {
        if (FS.ErrnoError) return;
        FS.ErrnoError = function ErrnoError(errno, node) {
          //Module.printErr(stackTrace()); // useful for debugging
          this.node = node;
          this.setErrno = function(errno) {
            this.errno = errno;
            for (var key in ERRNO_CODES) {
              if (ERRNO_CODES[key] === errno) {
                this.code = key;
                break;
              }
            }
          };
          this.setErrno(errno);
          this.message = ERRNO_MESSAGES[errno];
          if (this.stack) this.stack = demangleAll(this.stack);
        };
        FS.ErrnoError.prototype = new Error();
        FS.ErrnoError.prototype.constructor = FS.ErrnoError;
        // Some errors may happen quite a bit, to avoid overhead we reuse them (and suffer a lack of stack info)
        [ERRNO_CODES.ENOENT].forEach(function(code) {
          FS.genericErrors[code] = new FS.ErrnoError(code);
          FS.genericErrors[code].stack = '<generic error, no stack>';
        });
      },staticInit:function () {
        FS.ensureErrnoError();
  
        FS.nameTable = new Array(4096);
  
        FS.mount(MEMFS, {}, '/');
  
        FS.createDefaultDirectories();
        FS.createDefaultDevices();
        FS.createSpecialDirectories();
  
        FS.filesystems = {
          'MEMFS': MEMFS,
          'IDBFS': IDBFS,
          'NODEFS': NODEFS,
          'WORKERFS': WORKERFS,
        };
      },init:function (input, output, error) {
        assert(!FS.init.initialized, 'FS.init was previously called. If you want to initialize later with custom parameters, remove any earlier calls (note that one is automatically added to the generated code)');
        FS.init.initialized = true;
  
        FS.ensureErrnoError();
  
        // Allow Module.stdin etc. to provide defaults, if none explicitly passed to us here
        Module['stdin'] = input || Module['stdin'];
        Module['stdout'] = output || Module['stdout'];
        Module['stderr'] = error || Module['stderr'];
  
        FS.createStandardStreams();
      },quit:function () {
        FS.init.initialized = false;
        // force-flush all streams, so we get musl std streams printed out
        var fflush = Module['_fflush'];
        if (fflush) fflush(0);
        // close all of our streams
        for (var i = 0; i < FS.streams.length; i++) {
          var stream = FS.streams[i];
          if (!stream) {
            continue;
          }
          FS.close(stream);
        }
      },getMode:function (canRead, canWrite) {
        var mode = 0;
        if (canRead) mode |= 292 | 73;
        if (canWrite) mode |= 146;
        return mode;
      },joinPath:function (parts, forceRelative) {
        var path = PATH.join.apply(null, parts);
        if (forceRelative && path[0] == '/') path = path.substr(1);
        return path;
      },absolutePath:function (relative, base) {
        return PATH.resolve(base, relative);
      },standardizePath:function (path) {
        return PATH.normalize(path);
      },findObject:function (path, dontResolveLastLink) {
        var ret = FS.analyzePath(path, dontResolveLastLink);
        if (ret.exists) {
          return ret.object;
        } else {
          ___setErrNo(ret.error);
          return null;
        }
      },analyzePath:function (path, dontResolveLastLink) {
        // operate from within the context of the symlink's target
        try {
          var lookup = FS.lookupPath(path, { follow: !dontResolveLastLink });
          path = lookup.path;
        } catch (e) {
        }
        var ret = {
          isRoot: false, exists: false, error: 0, name: null, path: null, object: null,
          parentExists: false, parentPath: null, parentObject: null
        };
        try {
          var lookup = FS.lookupPath(path, { parent: true });
          ret.parentExists = true;
          ret.parentPath = lookup.path;
          ret.parentObject = lookup.node;
          ret.name = PATH.basename(path);
          lookup = FS.lookupPath(path, { follow: !dontResolveLastLink });
          ret.exists = true;
          ret.path = lookup.path;
          ret.object = lookup.node;
          ret.name = lookup.node.name;
          ret.isRoot = lookup.path === '/';
        } catch (e) {
          ret.error = e.errno;
        };
        return ret;
      },createFolder:function (parent, name, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(canRead, canWrite);
        return FS.mkdir(path, mode);
      },createPath:function (parent, path, canRead, canWrite) {
        parent = typeof parent === 'string' ? parent : FS.getPath(parent);
        var parts = path.split('/').reverse();
        while (parts.length) {
          var part = parts.pop();
          if (!part) continue;
          var current = PATH.join2(parent, part);
          try {
            FS.mkdir(current);
          } catch (e) {
            // ignore EEXIST
          }
          parent = current;
        }
        return current;
      },createFile:function (parent, name, properties, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(canRead, canWrite);
        return FS.create(path, mode);
      },createDataFile:function (parent, name, data, canRead, canWrite, canOwn) {
        var path = name ? PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name) : parent;
        var mode = FS.getMode(canRead, canWrite);
        var node = FS.create(path, mode);
        if (data) {
          if (typeof data === 'string') {
            var arr = new Array(data.length);
            for (var i = 0, len = data.length; i < len; ++i) arr[i] = data.charCodeAt(i);
            data = arr;
          }
          // make sure we can write to the file
          FS.chmod(node, mode | 146);
          var stream = FS.open(node, 'w');
          FS.write(stream, data, 0, data.length, 0, canOwn);
          FS.close(stream);
          FS.chmod(node, mode);
        }
        return node;
      },createDevice:function (parent, name, input, output) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        var mode = FS.getMode(!!input, !!output);
        if (!FS.createDevice.major) FS.createDevice.major = 64;
        var dev = FS.makedev(FS.createDevice.major++, 0);
        // Create a fake device that a set of stream ops to emulate
        // the old behavior.
        FS.registerDevice(dev, {
          open: function(stream) {
            stream.seekable = false;
          },
          close: function(stream) {
            // flush any pending line data
            if (output && output.buffer && output.buffer.length) {
              output(10);
            }
          },
          read: function(stream, buffer, offset, length, pos /* ignored */) {
            var bytesRead = 0;
            for (var i = 0; i < length; i++) {
              var result;
              try {
                result = input();
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES.EIO);
              }
              if (result === undefined && bytesRead === 0) {
                throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
              }
              if (result === null || result === undefined) break;
              bytesRead++;
              buffer[offset+i] = result;
            }
            if (bytesRead) {
              stream.node.timestamp = Date.now();
            }
            return bytesRead;
          },
          write: function(stream, buffer, offset, length, pos) {
            for (var i = 0; i < length; i++) {
              try {
                output(buffer[offset+i]);
              } catch (e) {
                throw new FS.ErrnoError(ERRNO_CODES.EIO);
              }
            }
            if (length) {
              stream.node.timestamp = Date.now();
            }
            return i;
          }
        });
        return FS.mkdev(path, mode, dev);
      },createLink:function (parent, name, target, canRead, canWrite) {
        var path = PATH.join2(typeof parent === 'string' ? parent : FS.getPath(parent), name);
        return FS.symlink(target, path);
      },forceLoadFile:function (obj) {
        if (obj.isDevice || obj.isFolder || obj.link || obj.contents) return true;
        var success = true;
        if (typeof XMLHttpRequest !== 'undefined') {
          throw new Error("Lazy loading should have been performed (contents set) in createLazyFile, but it was not. Lazy loading only works in web workers. Use --embed-file or --preload-file in emcc on the main thread.");
        } else if (Module['read']) {
          // Command-line.
          try {
            // WARNING: Can't read binary files in V8's d8 or tracemonkey's js, as
            //          read() will try to parse UTF8.
            obj.contents = intArrayFromString(Module['read'](obj.url), true);
            obj.usedBytes = obj.contents.length;
          } catch (e) {
            success = false;
          }
        } else {
          throw new Error('Cannot load without read() or XMLHttpRequest.');
        }
        if (!success) ___setErrNo(ERRNO_CODES.EIO);
        return success;
      },createLazyFile:function (parent, name, url, canRead, canWrite) {
        // Lazy chunked Uint8Array (implements get and length from Uint8Array). Actual getting is abstracted away for eventual reuse.
        function LazyUint8Array() {
          this.lengthKnown = false;
          this.chunks = []; // Loaded chunks. Index is the chunk number
        }
        LazyUint8Array.prototype.get = function LazyUint8Array_get(idx) {
          if (idx > this.length-1 || idx < 0) {
            return undefined;
          }
          var chunkOffset = idx % this.chunkSize;
          var chunkNum = (idx / this.chunkSize)|0;
          return this.getter(chunkNum)[chunkOffset];
        }
        LazyUint8Array.prototype.setDataGetter = function LazyUint8Array_setDataGetter(getter) {
          this.getter = getter;
        }
        LazyUint8Array.prototype.cacheLength = function LazyUint8Array_cacheLength() {
          // Find length
          var xhr = new XMLHttpRequest();
          xhr.open('HEAD', url, false);
          xhr.send(null);
          if (!(xhr.status >= 200 && xhr.status < 300 || xhr.status === 304)) throw new Error("Couldn't load " + url + ". Status: " + xhr.status);
          var datalength = Number(xhr.getResponseHeader("Content-length"));
          var header;
          var hasByteServing = (header = xhr.getResponseHeader("Accept-Ranges")) && header === "bytes";
          var usesGzip = (header = xhr.getResponseHeader("Content-Encoding")) && header === "gzip";
  
          var chunkSize = 1024*1024; // Chunk size in bytes
  
          if (!hasByteServing) chunkSize = datalength;
  
          // Function to get a range from the remote URL.
          var doXHR = (function(from, to) {
            if (from > to) throw new Error("invalid range (" + from + ", " + to + ") or no bytes requested!");
            if (to > datalength-1) throw new Error("only " + datalength + " bytes available! programmer error!");
  
            // TODO: Use mozResponseArrayBuffer, responseStream, etc. if available.
            var xhr = new XMLHttpRequest();
            xhr.open('GET', url, false);
            if (datalength !== chunkSize) xhr.setRequestHeader("Range", "bytes=" + from + "-" + to);
  
            // Some hints to the browser that we want binary data.
            if (typeof Uint8Array != 'undefined') xhr.responseType = 'arraybuffer';
            if (xhr.overrideMimeType) {
              xhr.overrideMimeType('text/plain; charset=x-user-defined');
            }
  
            xhr.send(null);
            if (!(xhr.status >= 200 && xhr.status < 300 || xhr.status === 304)) throw new Error("Couldn't load " + url + ". Status: " + xhr.status);
            if (xhr.response !== undefined) {
              return new Uint8Array(xhr.response || []);
            } else {
              return intArrayFromString(xhr.responseText || '', true);
            }
          });
          var lazyArray = this;
          lazyArray.setDataGetter(function(chunkNum) {
            var start = chunkNum * chunkSize;
            var end = (chunkNum+1) * chunkSize - 1; // including this byte
            end = Math.min(end, datalength-1); // if datalength-1 is selected, this is the last block
            if (typeof(lazyArray.chunks[chunkNum]) === "undefined") {
              lazyArray.chunks[chunkNum] = doXHR(start, end);
            }
            if (typeof(lazyArray.chunks[chunkNum]) === "undefined") throw new Error("doXHR failed!");
            return lazyArray.chunks[chunkNum];
          });
  
          if (usesGzip || !datalength) {
            // if the server uses gzip or doesn't supply the length, we have to download the whole file to get the (uncompressed) length
            chunkSize = datalength = 1; // this will force getter(0)/doXHR do download the whole file
            datalength = this.getter(0).length;
            chunkSize = datalength;
            console.log("LazyFiles on gzip forces download of the whole file when length is accessed");
          }
  
          this._length = datalength;
          this._chunkSize = chunkSize;
          this.lengthKnown = true;
        }
        if (typeof XMLHttpRequest !== 'undefined') {
          if (!ENVIRONMENT_IS_WORKER) throw 'Cannot do synchronous binary XHRs outside webworkers in modern browsers. Use --embed-file or --preload-file in emcc';
          var lazyArray = new LazyUint8Array();
          Object.defineProperties(lazyArray, {
            length: {
              get: function() {
                if(!this.lengthKnown) {
                  this.cacheLength();
                }
                return this._length;
              }
            },
            chunkSize: {
              get: function() {
                if(!this.lengthKnown) {
                  this.cacheLength();
                }
                return this._chunkSize;
              }
            }
          });
  
          var properties = { isDevice: false, contents: lazyArray };
        } else {
          var properties = { isDevice: false, url: url };
        }
  
        var node = FS.createFile(parent, name, properties, canRead, canWrite);
        // This is a total hack, but I want to get this lazy file code out of the
        // core of MEMFS. If we want to keep this lazy file concept I feel it should
        // be its own thin LAZYFS proxying calls to MEMFS.
        if (properties.contents) {
          node.contents = properties.contents;
        } else if (properties.url) {
          node.contents = null;
          node.url = properties.url;
        }
        // Add a function that defers querying the file size until it is asked the first time.
        Object.defineProperties(node, {
          usedBytes: {
            get: function() { return this.contents.length; }
          }
        });
        // override each stream op with one that tries to force load the lazy file first
        var stream_ops = {};
        var keys = Object.keys(node.stream_ops);
        keys.forEach(function(key) {
          var fn = node.stream_ops[key];
          stream_ops[key] = function forceLoadLazyFile() {
            if (!FS.forceLoadFile(node)) {
              throw new FS.ErrnoError(ERRNO_CODES.EIO);
            }
            return fn.apply(null, arguments);
          };
        });
        // use a custom read function
        stream_ops.read = function stream_ops_read(stream, buffer, offset, length, position) {
          if (!FS.forceLoadFile(node)) {
            throw new FS.ErrnoError(ERRNO_CODES.EIO);
          }
          var contents = stream.node.contents;
          if (position >= contents.length)
            return 0;
          var size = Math.min(contents.length - position, length);
          assert(size >= 0);
          if (contents.slice) { // normal array
            for (var i = 0; i < size; i++) {
              buffer[offset + i] = contents[position + i];
            }
          } else {
            for (var i = 0; i < size; i++) { // LazyUint8Array from sync binary XHR
              buffer[offset + i] = contents.get(position + i);
            }
          }
          return size;
        };
        node.stream_ops = stream_ops;
        return node;
      },createPreloadedFile:function (parent, name, url, canRead, canWrite, onload, onerror, dontCreateFile, canOwn, preFinish) {
        Browser.init(); // XXX perhaps this method should move onto Browser?
        // TODO we should allow people to just pass in a complete filename instead
        // of parent and name being that we just join them anyways
        var fullname = name ? PATH.resolve(PATH.join2(parent, name)) : parent;
        var dep = getUniqueRunDependency('cp ' + fullname); // might have several active requests for the same fullname
        function processData(byteArray) {
          function finish(byteArray) {
            if (preFinish) preFinish();
            if (!dontCreateFile) {
              FS.createDataFile(parent, name, byteArray, canRead, canWrite, canOwn);
            }
            if (onload) onload();
            removeRunDependency(dep);
          }
          var handled = false;
          Module['preloadPlugins'].forEach(function(plugin) {
            if (handled) return;
            if (plugin['canHandle'](fullname)) {
              plugin['handle'](byteArray, fullname, finish, function() {
                if (onerror) onerror();
                removeRunDependency(dep);
              });
              handled = true;
            }
          });
          if (!handled) finish(byteArray);
        }
        addRunDependency(dep);
        if (typeof url == 'string') {
          Browser.asyncLoad(url, function(byteArray) {
            processData(byteArray);
          }, onerror);
        } else {
          processData(url);
        }
      },indexedDB:function () {
        return window.indexedDB || window.mozIndexedDB || window.webkitIndexedDB || window.msIndexedDB;
      },DB_NAME:function () {
        return 'EM_FS_' + window.location.pathname;
      },DB_VERSION:20,DB_STORE_NAME:"FILE_DATA",saveFilesToDB:function (paths, onload, onerror) {
        onload = onload || function(){};
        onerror = onerror || function(){};
        var indexedDB = FS.indexedDB();
        try {
          var openRequest = indexedDB.open(FS.DB_NAME(), FS.DB_VERSION);
        } catch (e) {
          return onerror(e);
        }
        openRequest.onupgradeneeded = function openRequest_onupgradeneeded() {
          console.log('creating db');
          var db = openRequest.result;
          db.createObjectStore(FS.DB_STORE_NAME);
        };
        openRequest.onsuccess = function openRequest_onsuccess() {
          var db = openRequest.result;
          var transaction = db.transaction([FS.DB_STORE_NAME], 'readwrite');
          var files = transaction.objectStore(FS.DB_STORE_NAME);
          var ok = 0, fail = 0, total = paths.length;
          function finish() {
            if (fail == 0) onload(); else onerror();
          }
          paths.forEach(function(path) {
            var putRequest = files.put(FS.analyzePath(path).object.contents, path);
            putRequest.onsuccess = function putRequest_onsuccess() { ok++; if (ok + fail == total) finish() };
            putRequest.onerror = function putRequest_onerror() { fail++; if (ok + fail == total) finish() };
          });
          transaction.onerror = onerror;
        };
        openRequest.onerror = onerror;
      },loadFilesFromDB:function (paths, onload, onerror) {
        onload = onload || function(){};
        onerror = onerror || function(){};
        var indexedDB = FS.indexedDB();
        try {
          var openRequest = indexedDB.open(FS.DB_NAME(), FS.DB_VERSION);
        } catch (e) {
          return onerror(e);
        }
        openRequest.onupgradeneeded = onerror; // no database to load from
        openRequest.onsuccess = function openRequest_onsuccess() {
          var db = openRequest.result;
          try {
            var transaction = db.transaction([FS.DB_STORE_NAME], 'readonly');
          } catch(e) {
            onerror(e);
            return;
          }
          var files = transaction.objectStore(FS.DB_STORE_NAME);
          var ok = 0, fail = 0, total = paths.length;
          function finish() {
            if (fail == 0) onload(); else onerror();
          }
          paths.forEach(function(path) {
            var getRequest = files.get(path);
            getRequest.onsuccess = function getRequest_onsuccess() {
              if (FS.analyzePath(path).exists) {
                FS.unlink(path);
              }
              FS.createDataFile(PATH.dirname(path), PATH.basename(path), getRequest.result, true, true, true);
              ok++;
              if (ok + fail == total) finish();
            };
            getRequest.onerror = function getRequest_onerror() { fail++; if (ok + fail == total) finish() };
          });
          transaction.onerror = onerror;
        };
        openRequest.onerror = onerror;
      }};var SYSCALLS={DEFAULT_POLLMASK:5,mappings:{},umask:511,calculateAt:function (dirfd, path) {
        if (path[0] !== '/') {
          // relative path
          var dir;
          if (dirfd === -100) {
            dir = FS.cwd();
          } else {
            var dirstream = FS.getStream(dirfd);
            if (!dirstream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
            dir = dirstream.path;
          }
          path = PATH.join2(dir, path);
        }
        return path;
      },doStat:function (func, path, buf) {
        try {
          var stat = func(path);
        } catch (e) {
          if (e && e.node && PATH.normalize(path) !== PATH.normalize(FS.getPath(e.node))) {
            // an error occurred while trying to look up the path; we should just report ENOTDIR
            return -ERRNO_CODES.ENOTDIR;
          }
          throw e;
        }
        HEAP32[((buf)>>2)]=stat.dev;
        HEAP32[(((buf)+(4))>>2)]=0;
        HEAP32[(((buf)+(8))>>2)]=stat.ino;
        HEAP32[(((buf)+(12))>>2)]=stat.mode;
        HEAP32[(((buf)+(16))>>2)]=stat.nlink;
        HEAP32[(((buf)+(20))>>2)]=stat.uid;
        HEAP32[(((buf)+(24))>>2)]=stat.gid;
        HEAP32[(((buf)+(28))>>2)]=stat.rdev;
        HEAP32[(((buf)+(32))>>2)]=0;
        HEAP32[(((buf)+(36))>>2)]=stat.size;
        HEAP32[(((buf)+(40))>>2)]=4096;
        HEAP32[(((buf)+(44))>>2)]=stat.blocks;
        HEAP32[(((buf)+(48))>>2)]=(stat.atime.getTime() / 1000)|0;
        HEAP32[(((buf)+(52))>>2)]=0;
        HEAP32[(((buf)+(56))>>2)]=(stat.mtime.getTime() / 1000)|0;
        HEAP32[(((buf)+(60))>>2)]=0;
        HEAP32[(((buf)+(64))>>2)]=(stat.ctime.getTime() / 1000)|0;
        HEAP32[(((buf)+(68))>>2)]=0;
        HEAP32[(((buf)+(72))>>2)]=stat.ino;
        return 0;
      },doMsync:function (addr, stream, len, flags) {
        var buffer = new Uint8Array(HEAPU8.subarray(addr, addr + len));
        FS.msync(stream, buffer, 0, len, flags);
      },doMkdir:function (path, mode) {
        // remove a trailing slash, if one - /a/b/ has basename of '', but
        // we want to create b in the context of this function
        path = PATH.normalize(path);
        if (path[path.length-1] === '/') path = path.substr(0, path.length-1);
        FS.mkdir(path, mode, 0);
        return 0;
      },doMknod:function (path, mode, dev) {
        // we don't want this in the JS API as it uses mknod to create all nodes.
        switch (mode & 61440) {
          case 32768:
          case 8192:
          case 24576:
          case 4096:
          case 49152:
            break;
          default: return -ERRNO_CODES.EINVAL;
        }
        FS.mknod(path, mode, dev);
        return 0;
      },doReadlink:function (path, buf, bufsize) {
        if (bufsize <= 0) return -ERRNO_CODES.EINVAL;
        var ret = FS.readlink(path);
  
        var len = Math.min(bufsize, lengthBytesUTF8(ret));
        var endChar = HEAP8[buf+len];
        stringToUTF8(ret, buf, bufsize+1);
        // readlink is one of the rare functions that write out a C string, but does never append a null to the output buffer(!)
        // stringToUTF8() always appends a null byte, so restore the character under the null byte after the write.
        HEAP8[buf+len] = endChar;
  
        return len;
      },doAccess:function (path, amode) {
        if (amode & ~7) {
          // need a valid mode
          return -ERRNO_CODES.EINVAL;
        }
        var node;
        var lookup = FS.lookupPath(path, { follow: true });
        node = lookup.node;
        var perms = '';
        if (amode & 4) perms += 'r';
        if (amode & 2) perms += 'w';
        if (amode & 1) perms += 'x';
        if (perms /* otherwise, they've just passed F_OK */ && FS.nodePermissions(node, perms)) {
          return -ERRNO_CODES.EACCES;
        }
        return 0;
      },doDup:function (path, flags, suggestFD) {
        var suggest = FS.getStream(suggestFD);
        if (suggest) FS.close(suggest);
        return FS.open(path, flags, 0, suggestFD, suggestFD).fd;
      },doReadv:function (stream, iov, iovcnt, offset) {
        var ret = 0;
        for (var i = 0; i < iovcnt; i++) {
          var ptr = HEAP32[(((iov)+(i*8))>>2)];
          var len = HEAP32[(((iov)+(i*8 + 4))>>2)];
          var curr = FS.read(stream, HEAP8,ptr, len, offset);
          if (curr < 0) return -1;
          ret += curr;
          if (curr < len) break; // nothing more to read
        }
        return ret;
      },doWritev:function (stream, iov, iovcnt, offset) {
        var ret = 0;
        for (var i = 0; i < iovcnt; i++) {
          var ptr = HEAP32[(((iov)+(i*8))>>2)];
          var len = HEAP32[(((iov)+(i*8 + 4))>>2)];
          var curr = FS.write(stream, HEAP8,ptr, len, offset);
          if (curr < 0) return -1;
          ret += curr;
        }
        return ret;
      },varargs:0,get:function (varargs) {
        SYSCALLS.varargs += 4;
        var ret = HEAP32[(((SYSCALLS.varargs)-(4))>>2)];
        return ret;
      },getStr:function () {
        var ret = Pointer_stringify(SYSCALLS.get());
        return ret;
      },getStreamFromFD:function () {
        var stream = FS.getStream(SYSCALLS.get());
        if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        return stream;
      },getSocketFromFD:function () {
        var socket = SOCKFS.getSocket(SYSCALLS.get());
        if (!socket) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
        return socket;
      },getSocketAddress:function (allowNull) {
        var addrp = SYSCALLS.get(), addrlen = SYSCALLS.get();
        if (allowNull && addrp === 0) return null;
        var info = __read_sockaddr(addrp, addrlen);
        if (info.errno) throw new FS.ErrnoError(info.errno);
        info.addr = DNS.lookup_addr(info.addr) || info.addr;
        return info;
      },get64:function () {
        var low = SYSCALLS.get(), high = SYSCALLS.get();
        if (low >= 0) assert(high === 0);
        else assert(high === -1);
        return low;
      },getZero:function () {
        assert(SYSCALLS.get() === 0);
      }};function ___syscall91(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // munmap
      var addr = SYSCALLS.get(), len = SYSCALLS.get();
      // TODO: support unmmap'ing parts of allocations
      var info = SYSCALLS.mappings[addr];
      if (!info) return 0;
      if (len === info.len) {
        var stream = FS.getStream(info.fd);
        SYSCALLS.doMsync(addr, stream, len, info.flags)
        FS.munmap(stream);
        SYSCALLS.mappings[addr] = null;
        if (info.allocated) {
          _free(info.malloc);
        }
      }
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function _pthread_setspecific(key, value) {
      if (!(key in PTHREAD_SPECIFIC)) {
        return ERRNO_CODES.EINVAL;
      }
      PTHREAD_SPECIFIC[key] = value;
      return 0;
    }

  function ___cxa_allocate_exception(size) {
      return _malloc(size);
    }

  function ___syscall54(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // ioctl
      var stream = SYSCALLS.getStreamFromFD(), op = SYSCALLS.get();
      switch (op) {
        case 21505: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return 0;
        }
        case 21506: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return 0; // no-op, not actually adjusting terminal settings
        }
        case 21519: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          var argp = SYSCALLS.get();
          HEAP32[((argp)>>2)]=0;
          return 0;
        }
        case 21520: {
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return -ERRNO_CODES.EINVAL; // not supported
        }
        case 21531: {
          var argp = SYSCALLS.get();
          return FS.ioctl(stream, op, argp);
        }
        case 21523: {
          // TODO: in theory we should write to the winsize struct that gets
          // passed in, but for now musl doesn't read anything on it
          if (!stream.tty) return -ERRNO_CODES.ENOTTY;
          return 0;
        }
        default: abort('bad ioctl syscall ' + op);
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  var ___tm_current=STATICTOP; STATICTOP += 48;;
  
  
  var ___tm_timezone=allocate(intArrayFromString("GMT"), "i8", ALLOC_STATIC);
  
  
  var _tzname=STATICTOP; STATICTOP += 16;;
  
  var _daylight=STATICTOP; STATICTOP += 16;;
  
  var _timezone=STATICTOP; STATICTOP += 16;;function _tzset() {
      // TODO: Use (malleable) environment variables instead of system settings.
      if (_tzset.called) return;
      _tzset.called = true;
  
      HEAP32[((_timezone)>>2)]=-(new Date()).getTimezoneOffset() * 60;
  
      var winter = new Date(2000, 0, 1);
      var summer = new Date(2000, 6, 1);
      HEAP32[((_daylight)>>2)]=Number(winter.getTimezoneOffset() != summer.getTimezoneOffset());
  
      function extractZone(date) {
        var match = date.toTimeString().match(/\(([A-Za-z ]+)\)$/);
        return match ? match[1] : "GMT";
      };
      var winterName = extractZone(winter);
      var summerName = extractZone(summer);
      var winterNamePtr = allocate(intArrayFromString(winterName), 'i8', ALLOC_NORMAL);
      var summerNamePtr = allocate(intArrayFromString(summerName), 'i8', ALLOC_NORMAL);
      if (summer.getTimezoneOffset() < winter.getTimezoneOffset()) {
        // Northern hemisphere
        HEAP32[((_tzname)>>2)]=winterNamePtr;
        HEAP32[(((_tzname)+(4))>>2)]=summerNamePtr;
      } else {
        HEAP32[((_tzname)>>2)]=summerNamePtr;
        HEAP32[(((_tzname)+(4))>>2)]=winterNamePtr;
      }
    }function _localtime_r(time, tmPtr) {
      _tzset();
      var date = new Date(HEAP32[((time)>>2)]*1000);
      HEAP32[((tmPtr)>>2)]=date.getSeconds();
      HEAP32[(((tmPtr)+(4))>>2)]=date.getMinutes();
      HEAP32[(((tmPtr)+(8))>>2)]=date.getHours();
      HEAP32[(((tmPtr)+(12))>>2)]=date.getDate();
      HEAP32[(((tmPtr)+(16))>>2)]=date.getMonth();
      HEAP32[(((tmPtr)+(20))>>2)]=date.getFullYear()-1900;
      HEAP32[(((tmPtr)+(24))>>2)]=date.getDay();
  
      var start = new Date(date.getFullYear(), 0, 1);
      var yday = ((date.getTime() - start.getTime()) / (1000 * 60 * 60 * 24))|0;
      HEAP32[(((tmPtr)+(28))>>2)]=yday;
      HEAP32[(((tmPtr)+(36))>>2)]=-(date.getTimezoneOffset() * 60);
  
      // DST is in December in South
      var summerOffset = new Date(2000, 6, 1).getTimezoneOffset();
      var winterOffset = start.getTimezoneOffset();
      var dst = (date.getTimezoneOffset() == Math.min(winterOffset, summerOffset))|0;
      HEAP32[(((tmPtr)+(32))>>2)]=dst;
  
      var zonePtr = HEAP32[(((_tzname)+(dst ? Runtime.QUANTUM_SIZE : 0))>>2)];
      HEAP32[(((tmPtr)+(40))>>2)]=zonePtr;
  
      return tmPtr;
    }function _localtime(time) {
      return _localtime_r(time, ___tm_current);
    }

   

   

  function ___cxa_pure_virtual() {
      ABORT = true;
      throw 'Pure virtual function called!';
    }


  function _time(ptr) {
      var ret = (Date.now()/1000)|0;
      if (ptr) {
        HEAP32[((ptr)>>2)]=ret;
      }
      return ret;
    }

   

  
  
  
  
  var _environ=STATICTOP; STATICTOP += 16;;var ___environ=_environ;function ___buildEnvironment(env) {
      // WARNING: Arbitrary limit!
      var MAX_ENV_VALUES = 64;
      var TOTAL_ENV_SIZE = 1024;
  
      // Statically allocate memory for the environment.
      var poolPtr;
      var envPtr;
      if (!___buildEnvironment.called) {
        ___buildEnvironment.called = true;
        // Set default values. Use string keys for Closure Compiler compatibility.
        ENV['USER'] = ENV['LOGNAME'] = 'web_user';
        ENV['PATH'] = '/';
        ENV['PWD'] = '/';
        ENV['HOME'] = '/home/web_user';
        ENV['LANG'] = 'C';
        ENV['_'] = Module['thisProgram'];
        // Allocate memory.
        poolPtr = allocate(TOTAL_ENV_SIZE, 'i8', ALLOC_STATIC);
        envPtr = allocate(MAX_ENV_VALUES * 4,
                          'i8*', ALLOC_STATIC);
        HEAP32[((envPtr)>>2)]=poolPtr;
        HEAP32[((_environ)>>2)]=envPtr;
      } else {
        envPtr = HEAP32[((_environ)>>2)];
        poolPtr = HEAP32[((envPtr)>>2)];
      }
  
      // Collect key=value lines.
      var strings = [];
      var totalSize = 0;
      for (var key in env) {
        if (typeof env[key] === 'string') {
          var line = key + '=' + env[key];
          strings.push(line);
          totalSize += line.length;
        }
      }
      if (totalSize > TOTAL_ENV_SIZE) {
        throw new Error('Environment size exceeded TOTAL_ENV_SIZE!');
      }
  
      // Make new.
      var ptrSize = 4;
      for (var i = 0; i < strings.length; i++) {
        var line = strings[i];
        writeAsciiToMemory(line, poolPtr);
        HEAP32[(((envPtr)+(i * ptrSize))>>2)]=poolPtr;
        poolPtr += line.length + 1;
      }
      HEAP32[(((envPtr)+(strings.length * ptrSize))>>2)]=0;
    }var ENV={};function _getenv(name) {
      // char *getenv(const char *name);
      // http://pubs.opengroup.org/onlinepubs/009695399/functions/getenv.html
      if (name === 0) return 0;
      name = Pointer_stringify(name);
      if (!ENV.hasOwnProperty(name)) return 0;
  
      if (_getenv.ret) _free(_getenv.ret);
      _getenv.ret = allocate(intArrayFromString(ENV[name]), 'i8', ALLOC_NORMAL);
      return _getenv.ret;
    }

  function ___map_file(pathname, size) {
      ___setErrNo(ERRNO_CODES.EPERM);
      return -1;
    }

  function ___cxa_find_matching_catch_2() {
          return ___cxa_find_matching_catch.apply(null, arguments);
        }

  function ___cxa_find_matching_catch_3() {
          return ___cxa_find_matching_catch.apply(null, arguments);
        }

  function ___cxa_begin_catch(ptr) {
      var info = EXCEPTIONS.infos[ptr];
      if (info && !info.caught) {
        info.caught = true;
        __ZSt18uncaught_exceptionv.uncaught_exception--;
      }
      if (info) info.rethrown = false;
      EXCEPTIONS.caught.push(ptr);
      EXCEPTIONS.addRef(EXCEPTIONS.deAdjust(ptr));
      return ptr;
    }

  
  function _emscripten_memcpy_big(dest, src, num) {
      HEAPU8.set(HEAPU8.subarray(src, src+num), dest);
      return dest;
    } 

  function ___syscall6(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // close
      var stream = SYSCALLS.getStreamFromFD();
      FS.close(stream);
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  
  
  var cttz_i8 = allocate([8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0], "i8", ALLOC_STATIC);   

  
    

   

   

   

  
  var SOCKFS={mount:function (mount) {
        // If Module['websocket'] has already been defined (e.g. for configuring
        // the subprotocol/url) use that, if not initialise it to a new object.
        Module['websocket'] = (Module['websocket'] && 
                               ('object' === typeof Module['websocket'])) ? Module['websocket'] : {};
  
        // Add the Event registration mechanism to the exported websocket configuration
        // object so we can register network callbacks from native JavaScript too.
        // For more documentation see system/include/emscripten/emscripten.h
        Module['websocket']._callbacks = {};
        Module['websocket']['on'] = function(event, callback) {
  	    if ('function' === typeof callback) {
  		  this._callbacks[event] = callback;
          }
  	    return this;
        };
  
        Module['websocket'].emit = function(event, param) {
  	    if ('function' === typeof this._callbacks[event]) {
  		  this._callbacks[event].call(this, param);
          }
        };
  
        // If debug is enabled register simple default logging callbacks for each Event.
  
        return FS.createNode(null, '/', 16384 | 511 /* 0777 */, 0);
      },createSocket:function (family, type, protocol) {
        var streaming = type == 1;
        if (protocol) {
          assert(streaming == (protocol == 6)); // if SOCK_STREAM, must be tcp
        }
  
        // create our internal socket structure
        var sock = {
          family: family,
          type: type,
          protocol: protocol,
          server: null,
          error: null, // Used in getsockopt for SOL_SOCKET/SO_ERROR test
          peers: {},
          pending: [],
          recv_queue: [],
          sock_ops: SOCKFS.websocket_sock_ops
        };
  
        // create the filesystem node to store the socket structure
        var name = SOCKFS.nextname();
        var node = FS.createNode(SOCKFS.root, name, 49152, 0);
        node.sock = sock;
  
        // and the wrapping stream that enables library functions such
        // as read and write to indirectly interact with the socket
        var stream = FS.createStream({
          path: name,
          node: node,
          flags: FS.modeStringToFlags('r+'),
          seekable: false,
          stream_ops: SOCKFS.stream_ops
        });
  
        // map the new stream to the socket structure (sockets have a 1:1
        // relationship with a stream)
        sock.stream = stream;
  
        return sock;
      },getSocket:function (fd) {
        var stream = FS.getStream(fd);
        if (!stream || !FS.isSocket(stream.node.mode)) {
          return null;
        }
        return stream.node.sock;
      },stream_ops:{poll:function (stream) {
          var sock = stream.node.sock;
          return sock.sock_ops.poll(sock);
        },ioctl:function (stream, request, varargs) {
          var sock = stream.node.sock;
          return sock.sock_ops.ioctl(sock, request, varargs);
        },read:function (stream, buffer, offset, length, position /* ignored */) {
          var sock = stream.node.sock;
          var msg = sock.sock_ops.recvmsg(sock, length);
          if (!msg) {
            // socket is closed
            return 0;
          }
          buffer.set(msg.buffer, offset);
          return msg.buffer.length;
        },write:function (stream, buffer, offset, length, position /* ignored */) {
          var sock = stream.node.sock;
          return sock.sock_ops.sendmsg(sock, buffer, offset, length);
        },close:function (stream) {
          var sock = stream.node.sock;
          sock.sock_ops.close(sock);
        }},nextname:function () {
        if (!SOCKFS.nextname.current) {
          SOCKFS.nextname.current = 0;
        }
        return 'socket[' + (SOCKFS.nextname.current++) + ']';
      },websocket_sock_ops:{createPeer:function (sock, addr, port) {
          var ws;
  
          if (typeof addr === 'object') {
            ws = addr;
            addr = null;
            port = null;
          }
  
          if (ws) {
            // for sockets that've already connected (e.g. we're the server)
            // we can inspect the _socket property for the address
            if (ws._socket) {
              addr = ws._socket.remoteAddress;
              port = ws._socket.remotePort;
            }
            // if we're just now initializing a connection to the remote,
            // inspect the url property
            else {
              var result = /ws[s]?:\/\/([^:]+):(\d+)/.exec(ws.url);
              if (!result) {
                throw new Error('WebSocket URL must be in the format ws(s)://address:port');
              }
              addr = result[1];
              port = parseInt(result[2], 10);
            }
          } else {
            // create the actual websocket object and connect
            try {
              // runtimeConfig gets set to true if WebSocket runtime configuration is available.
              var runtimeConfig = (Module['websocket'] && ('object' === typeof Module['websocket']));
  
              // The default value is 'ws://' the replace is needed because the compiler replaces '//' comments with '#'
              // comments without checking context, so we'd end up with ws:#, the replace swaps the '#' for '//' again.
              var url = 'ws:#'.replace('#', '//');
  
              if (runtimeConfig) {
                if ('string' === typeof Module['websocket']['url']) {
                  url = Module['websocket']['url']; // Fetch runtime WebSocket URL config.
                }
              }
  
              if (url === 'ws://' || url === 'wss://') { // Is the supplied URL config just a prefix, if so complete it.
                var parts = addr.split('/');
                url = url + parts[0] + ":" + port + "/" + parts.slice(1).join('/');
              }
  
              // Make the WebSocket subprotocol (Sec-WebSocket-Protocol) default to binary if no configuration is set.
              var subProtocols = 'binary'; // The default value is 'binary'
  
              if (runtimeConfig) {
                if ('string' === typeof Module['websocket']['subprotocol']) {
                  subProtocols = Module['websocket']['subprotocol']; // Fetch runtime WebSocket subprotocol config.
                }
              }
  
              // The regex trims the string (removes spaces at the beginning and end, then splits the string by
              // <any space>,<any space> into an Array. Whitespace removal is important for Websockify and ws.
              subProtocols = subProtocols.replace(/^ +| +$/g,"").split(/ *, */);
  
              // The node ws library API for specifying optional subprotocol is slightly different than the browser's.
              var opts = ENVIRONMENT_IS_NODE ? {'protocol': subProtocols.toString()} : subProtocols;
  
              // If node we use the ws library.
              var WebSocketConstructor;
              if (ENVIRONMENT_IS_NODE) {
                WebSocketConstructor = require('ws');
              } else if (ENVIRONMENT_IS_WEB) {
                WebSocketConstructor = window['WebSocket'];
              } else {
                WebSocketConstructor = WebSocket;
              }
              ws = new WebSocketConstructor(url, opts);
              ws.binaryType = 'arraybuffer';
            } catch (e) {
              throw new FS.ErrnoError(ERRNO_CODES.EHOSTUNREACH);
            }
          }
  
  
          var peer = {
            addr: addr,
            port: port,
            socket: ws,
            dgram_send_queue: []
          };
  
          SOCKFS.websocket_sock_ops.addPeer(sock, peer);
          SOCKFS.websocket_sock_ops.handlePeerEvents(sock, peer);
  
          // if this is a bound dgram socket, send the port number first to allow
          // us to override the ephemeral port reported to us by remotePort on the
          // remote end.
          if (sock.type === 2 && typeof sock.sport !== 'undefined') {
            peer.dgram_send_queue.push(new Uint8Array([
                255, 255, 255, 255,
                'p'.charCodeAt(0), 'o'.charCodeAt(0), 'r'.charCodeAt(0), 't'.charCodeAt(0),
                ((sock.sport & 0xff00) >> 8) , (sock.sport & 0xff)
            ]));
          }
  
          return peer;
        },getPeer:function (sock, addr, port) {
          return sock.peers[addr + ':' + port];
        },addPeer:function (sock, peer) {
          sock.peers[peer.addr + ':' + peer.port] = peer;
        },removePeer:function (sock, peer) {
          delete sock.peers[peer.addr + ':' + peer.port];
        },handlePeerEvents:function (sock, peer) {
          var first = true;
  
          var handleOpen = function () {
  
            Module['websocket'].emit('open', sock.stream.fd);
  
            try {
              var queued = peer.dgram_send_queue.shift();
              while (queued) {
                peer.socket.send(queued);
                queued = peer.dgram_send_queue.shift();
              }
            } catch (e) {
              // not much we can do here in the way of proper error handling as we've already
              // lied and said this data was sent. shut it down.
              peer.socket.close();
            }
          };
  
          function handleMessage(data) {
            assert(typeof data !== 'string' && data.byteLength !== undefined);  // must receive an ArrayBuffer
  
            // An empty ArrayBuffer will emit a pseudo disconnect event
            // as recv/recvmsg will return zero which indicates that a socket
            // has performed a shutdown although the connection has not been disconnected yet.
            if (data.byteLength == 0) {
              return;
            }
            data = new Uint8Array(data);  // make a typed array view on the array buffer
  
  
            // if this is the port message, override the peer's port with it
            var wasfirst = first;
            first = false;
            if (wasfirst &&
                data.length === 10 &&
                data[0] === 255 && data[1] === 255 && data[2] === 255 && data[3] === 255 &&
                data[4] === 'p'.charCodeAt(0) && data[5] === 'o'.charCodeAt(0) && data[6] === 'r'.charCodeAt(0) && data[7] === 't'.charCodeAt(0)) {
              // update the peer's port and it's key in the peer map
              var newport = ((data[8] << 8) | data[9]);
              SOCKFS.websocket_sock_ops.removePeer(sock, peer);
              peer.port = newport;
              SOCKFS.websocket_sock_ops.addPeer(sock, peer);
              return;
            }
  
            sock.recv_queue.push({ addr: peer.addr, port: peer.port, data: data });
            Module['websocket'].emit('message', sock.stream.fd);
          };
  
          if (ENVIRONMENT_IS_NODE) {
            peer.socket.on('open', handleOpen);
            peer.socket.on('message', function(data, flags) {
              if (!flags.binary) {
                return;
              }
              handleMessage((new Uint8Array(data)).buffer);  // copy from node Buffer -> ArrayBuffer
            });
            peer.socket.on('close', function() {
              Module['websocket'].emit('close', sock.stream.fd);
            });
            peer.socket.on('error', function(error) {
              // Although the ws library may pass errors that may be more descriptive than
              // ECONNREFUSED they are not necessarily the expected error code e.g. 
              // ENOTFOUND on getaddrinfo seems to be node.js specific, so using ECONNREFUSED
              // is still probably the most useful thing to do.
              sock.error = ERRNO_CODES.ECONNREFUSED; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
              Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'ECONNREFUSED: Connection refused']);
              // don't throw
            });
          } else {
            peer.socket.onopen = handleOpen;
            peer.socket.onclose = function() {
              Module['websocket'].emit('close', sock.stream.fd);
            };
            peer.socket.onmessage = function peer_socket_onmessage(event) {
              handleMessage(event.data);
            };
            peer.socket.onerror = function(error) {
              // The WebSocket spec only allows a 'simple event' to be thrown on error,
              // so we only really know as much as ECONNREFUSED.
              sock.error = ERRNO_CODES.ECONNREFUSED; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
              Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'ECONNREFUSED: Connection refused']);
            };
          }
        },poll:function (sock) {
          if (sock.type === 1 && sock.server) {
            // listen sockets should only say they're available for reading
            // if there are pending clients.
            return sock.pending.length ? (64 | 1) : 0;
          }
  
          var mask = 0;
          var dest = sock.type === 1 ?  // we only care about the socket state for connection-based sockets
            SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport) :
            null;
  
          if (sock.recv_queue.length ||
              !dest ||  // connection-less sockets are always ready to read
              (dest && dest.socket.readyState === dest.socket.CLOSING) ||
              (dest && dest.socket.readyState === dest.socket.CLOSED)) {  // let recv return 0 once closed
            mask |= (64 | 1);
          }
  
          if (!dest ||  // connection-less sockets are always ready to write
              (dest && dest.socket.readyState === dest.socket.OPEN)) {
            mask |= 4;
          }
  
          if ((dest && dest.socket.readyState === dest.socket.CLOSING) ||
              (dest && dest.socket.readyState === dest.socket.CLOSED)) {
            mask |= 16;
          }
  
          return mask;
        },ioctl:function (sock, request, arg) {
          switch (request) {
            case 21531:
              var bytes = 0;
              if (sock.recv_queue.length) {
                bytes = sock.recv_queue[0].data.length;
              }
              HEAP32[((arg)>>2)]=bytes;
              return 0;
            default:
              return ERRNO_CODES.EINVAL;
          }
        },close:function (sock) {
          // if we've spawned a listen server, close it
          if (sock.server) {
            try {
              sock.server.close();
            } catch (e) {
            }
            sock.server = null;
          }
          // close any peer connections
          var peers = Object.keys(sock.peers);
          for (var i = 0; i < peers.length; i++) {
            var peer = sock.peers[peers[i]];
            try {
              peer.socket.close();
            } catch (e) {
            }
            SOCKFS.websocket_sock_ops.removePeer(sock, peer);
          }
          return 0;
        },bind:function (sock, addr, port) {
          if (typeof sock.saddr !== 'undefined' || typeof sock.sport !== 'undefined') {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);  // already bound
          }
          sock.saddr = addr;
          sock.sport = port;
          // in order to emulate dgram sockets, we need to launch a listen server when
          // binding on a connection-less socket
          // note: this is only required on the server side
          if (sock.type === 2) {
            // close the existing server if it exists
            if (sock.server) {
              sock.server.close();
              sock.server = null;
            }
            // swallow error operation not supported error that occurs when binding in the
            // browser where this isn't supported
            try {
              sock.sock_ops.listen(sock, 0);
            } catch (e) {
              if (!(e instanceof FS.ErrnoError)) throw e;
              if (e.errno !== ERRNO_CODES.EOPNOTSUPP) throw e;
            }
          }
        },connect:function (sock, addr, port) {
          if (sock.server) {
            throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
          }
  
          // TODO autobind
          // if (!sock.addr && sock.type == 2) {
          // }
  
          // early out if we're already connected / in the middle of connecting
          if (typeof sock.daddr !== 'undefined' && typeof sock.dport !== 'undefined') {
            var dest = SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport);
            if (dest) {
              if (dest.socket.readyState === dest.socket.CONNECTING) {
                throw new FS.ErrnoError(ERRNO_CODES.EALREADY);
              } else {
                throw new FS.ErrnoError(ERRNO_CODES.EISCONN);
              }
            }
          }
  
          // add the socket to our peer list and set our
          // destination address / port to match
          var peer = SOCKFS.websocket_sock_ops.createPeer(sock, addr, port);
          sock.daddr = peer.addr;
          sock.dport = peer.port;
  
          // always "fail" in non-blocking mode
          throw new FS.ErrnoError(ERRNO_CODES.EINPROGRESS);
        },listen:function (sock, backlog) {
          if (!ENVIRONMENT_IS_NODE) {
            throw new FS.ErrnoError(ERRNO_CODES.EOPNOTSUPP);
          }
          if (sock.server) {
             throw new FS.ErrnoError(ERRNO_CODES.EINVAL);  // already listening
          }
          var WebSocketServer = require('ws').Server;
          var host = sock.saddr;
          sock.server = new WebSocketServer({
            host: host,
            port: sock.sport
            // TODO support backlog
          });
          Module['websocket'].emit('listen', sock.stream.fd); // Send Event with listen fd.
  
          sock.server.on('connection', function(ws) {
            if (sock.type === 1) {
              var newsock = SOCKFS.createSocket(sock.family, sock.type, sock.protocol);
  
              // create a peer on the new socket
              var peer = SOCKFS.websocket_sock_ops.createPeer(newsock, ws);
              newsock.daddr = peer.addr;
              newsock.dport = peer.port;
  
              // push to queue for accept to pick up
              sock.pending.push(newsock);
              Module['websocket'].emit('connection', newsock.stream.fd);
            } else {
              // create a peer on the listen socket so calling sendto
              // with the listen socket and an address will resolve
              // to the correct client
              SOCKFS.websocket_sock_ops.createPeer(sock, ws);
              Module['websocket'].emit('connection', sock.stream.fd);
            }
          });
          sock.server.on('closed', function() {
            Module['websocket'].emit('close', sock.stream.fd);
            sock.server = null;
          });
          sock.server.on('error', function(error) {
            // Although the ws library may pass errors that may be more descriptive than
            // ECONNREFUSED they are not necessarily the expected error code e.g. 
            // ENOTFOUND on getaddrinfo seems to be node.js specific, so using EHOSTUNREACH
            // is still probably the most useful thing to do. This error shouldn't
            // occur in a well written app as errors should get trapped in the compiled
            // app's own getaddrinfo call.
            sock.error = ERRNO_CODES.EHOSTUNREACH; // Used in getsockopt for SOL_SOCKET/SO_ERROR test.
            Module['websocket'].emit('error', [sock.stream.fd, sock.error, 'EHOSTUNREACH: Host is unreachable']);
            // don't throw
          });
        },accept:function (listensock) {
          if (!listensock.server) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
          var newsock = listensock.pending.shift();
          newsock.stream.flags = listensock.stream.flags;
          return newsock;
        },getname:function (sock, peer) {
          var addr, port;
          if (peer) {
            if (sock.daddr === undefined || sock.dport === undefined) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
            }
            addr = sock.daddr;
            port = sock.dport;
          } else {
            // TODO saddr and sport will be set for bind()'d UDP sockets, but what
            // should we be returning for TCP sockets that've been connect()'d?
            addr = sock.saddr || 0;
            port = sock.sport || 0;
          }
          return { addr: addr, port: port };
        },sendmsg:function (sock, buffer, offset, length, addr, port) {
          if (sock.type === 2) {
            // connection-less sockets will honor the message address,
            // and otherwise fall back to the bound destination address
            if (addr === undefined || port === undefined) {
              addr = sock.daddr;
              port = sock.dport;
            }
            // if there was no address to fall back to, error out
            if (addr === undefined || port === undefined) {
              throw new FS.ErrnoError(ERRNO_CODES.EDESTADDRREQ);
            }
          } else {
            // connection-based sockets will only use the bound
            addr = sock.daddr;
            port = sock.dport;
          }
  
          // find the peer for the destination address
          var dest = SOCKFS.websocket_sock_ops.getPeer(sock, addr, port);
  
          // early out if not connected with a connection-based socket
          if (sock.type === 1) {
            if (!dest || dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
              throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
            } else if (dest.socket.readyState === dest.socket.CONNECTING) {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
          }
  
          // create a copy of the incoming data to send, as the WebSocket API
          // doesn't work entirely with an ArrayBufferView, it'll just send
          // the entire underlying buffer
          var data;
          if (buffer instanceof Array || buffer instanceof ArrayBuffer) {
            data = buffer.slice(offset, offset + length);
          } else {  // ArrayBufferView
            data = buffer.buffer.slice(buffer.byteOffset + offset, buffer.byteOffset + offset + length);
          }
  
          // if we're emulating a connection-less dgram socket and don't have
          // a cached connection, queue the buffer to send upon connect and
          // lie, saying the data was sent now.
          if (sock.type === 2) {
            if (!dest || dest.socket.readyState !== dest.socket.OPEN) {
              // if we're not connected, open a new connection
              if (!dest || dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
                dest = SOCKFS.websocket_sock_ops.createPeer(sock, addr, port);
              }
              dest.dgram_send_queue.push(data);
              return length;
            }
          }
  
          try {
            // send the actual data
            dest.socket.send(data);
            return length;
          } catch (e) {
            throw new FS.ErrnoError(ERRNO_CODES.EINVAL);
          }
        },recvmsg:function (sock, length) {
          // http://pubs.opengroup.org/onlinepubs/7908799/xns/recvmsg.html
          if (sock.type === 1 && sock.server) {
            // tcp servers should not be recv()'ing on the listen socket
            throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
          }
  
          var queued = sock.recv_queue.shift();
          if (!queued) {
            if (sock.type === 1) {
              var dest = SOCKFS.websocket_sock_ops.getPeer(sock, sock.daddr, sock.dport);
  
              if (!dest) {
                // if we have a destination address but are not connected, error out
                throw new FS.ErrnoError(ERRNO_CODES.ENOTCONN);
              }
              else if (dest.socket.readyState === dest.socket.CLOSING || dest.socket.readyState === dest.socket.CLOSED) {
                // return null if the socket has closed
                return null;
              }
              else {
                // else, our socket is in a valid state but truly has nothing available
                throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
              }
            } else {
              throw new FS.ErrnoError(ERRNO_CODES.EAGAIN);
            }
          }
  
          // queued.data will be an ArrayBuffer if it's unadulterated, but if it's
          // requeued TCP data it'll be an ArrayBufferView
          var queuedLength = queued.data.byteLength || queued.data.length;
          var queuedOffset = queued.data.byteOffset || 0;
          var queuedBuffer = queued.data.buffer || queued.data;
          var bytesRead = Math.min(length, queuedLength);
          var res = {
            buffer: new Uint8Array(queuedBuffer, queuedOffset, bytesRead),
            addr: queued.addr,
            port: queued.port
          };
  
  
          // push back any unread data for TCP connections
          if (sock.type === 1 && bytesRead < queuedLength) {
            var bytesRemaining = queuedLength - bytesRead;
            queued.data = new Uint8Array(queuedBuffer, queuedOffset + bytesRead, bytesRemaining);
            sock.recv_queue.unshift(queued);
          }
  
          return res;
        }}};
  
  
  function __inet_pton4_raw(str) {
      var b = str.split('.');
      for (var i = 0; i < 4; i++) {
        var tmp = Number(b[i]);
        if (isNaN(tmp)) return null;
        b[i] = tmp;
      }
      return (b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24)) >>> 0;
    }
  
  function __inet_pton6_raw(str) {
      var words;
      var w, offset, z, i;
      /* http://home.deds.nl/~aeron/regex/ */
      var valid6regx = /^((?=.*::)(?!.*::.+::)(::)?([\dA-F]{1,4}:(:|\b)|){5}|([\dA-F]{1,4}:){6})((([\dA-F]{1,4}((?!\3)::|:\b|$))|(?!\2\3)){2}|(((2[0-4]|1\d|[1-9])?\d|25[0-5])\.?\b){4})$/i
      var parts = [];
      if (!valid6regx.test(str)) {
        return null;
      }
      if (str === "::") {
        return [0, 0, 0, 0, 0, 0, 0, 0];
      }
      // Z placeholder to keep track of zeros when splitting the string on ":"
      if (str.indexOf("::") === 0) {
        str = str.replace("::", "Z:"); // leading zeros case
      } else {
        str = str.replace("::", ":Z:");
      }
  
      if (str.indexOf(".") > 0) {
        // parse IPv4 embedded stress
        str = str.replace(new RegExp('[.]', 'g'), ":");
        words = str.split(":");
        words[words.length-4] = parseInt(words[words.length-4]) + parseInt(words[words.length-3])*256;
        words[words.length-3] = parseInt(words[words.length-2]) + parseInt(words[words.length-1])*256;
        words = words.slice(0, words.length-2);
      } else {
        words = str.split(":");
      }
  
      offset = 0; z = 0;
      for (w=0; w < words.length; w++) {
        if (typeof words[w] === 'string') {
          if (words[w] === 'Z') {
            // compressed zeros - write appropriate number of zero words
            for (z = 0; z < (8 - words.length+1); z++) {
              parts[w+z] = 0;
            }
            offset = z-1;
          } else {
            // parse hex to field to 16-bit value and write it in network byte-order
            parts[w+offset] = _htons(parseInt(words[w],16));
          }
        } else {
          // parsed IPv4 words
          parts[w+offset] = words[w];
        }
      }
      return [
        (parts[1] << 16) | parts[0],
        (parts[3] << 16) | parts[2],
        (parts[5] << 16) | parts[4],
        (parts[7] << 16) | parts[6]
      ];
    }var DNS={address_map:{id:1,addrs:{},names:{}},lookup_name:function (name) {
        // If the name is already a valid ipv4 / ipv6 address, don't generate a fake one.
        var res = __inet_pton4_raw(name);
        if (res !== null) {
          return name;
        }
        res = __inet_pton6_raw(name);
        if (res !== null) {
          return name;
        }
  
        // See if this name is already mapped.
        var addr;
  
        if (DNS.address_map.addrs[name]) {
          addr = DNS.address_map.addrs[name];
        } else {
          var id = DNS.address_map.id++;
          assert(id < 65535, 'exceeded max address mappings of 65535');
  
          addr = '172.29.' + (id & 0xff) + '.' + (id & 0xff00);
  
          DNS.address_map.names[addr] = name;
          DNS.address_map.addrs[name] = addr;
        }
  
        return addr;
      },lookup_addr:function (addr) {
        if (DNS.address_map.names[addr]) {
          return DNS.address_map.names[addr];
        }
  
        return null;
      }};
  
  
  var Sockets={BUFFER_SIZE:10240,MAX_BUFFER_SIZE:10485760,nextFd:1,fds:{},nextport:1,maxport:65535,peer:null,connections:{},portmap:{},localAddr:4261412874,addrPool:[33554442,50331658,67108874,83886090,100663306,117440522,134217738,150994954,167772170,184549386,201326602,218103818,234881034]};
  
  function __inet_ntop4_raw(addr) {
      return (addr & 0xff) + '.' + ((addr >> 8) & 0xff) + '.' + ((addr >> 16) & 0xff) + '.' + ((addr >> 24) & 0xff)
    }
  
  function __inet_ntop6_raw(ints) {
      //  ref:  http://www.ietf.org/rfc/rfc2373.txt - section 2.5.4
      //  Format for IPv4 compatible and mapped  128-bit IPv6 Addresses
      //  128-bits are split into eight 16-bit words
      //  stored in network byte order (big-endian)
      //  |                80 bits               | 16 |      32 bits        |
      //  +-----------------------------------------------------------------+
      //  |               10 bytes               |  2 |      4 bytes        |
      //  +--------------------------------------+--------------------------+
      //  +               5 words                |  1 |      2 words        |
      //  +--------------------------------------+--------------------------+
      //  |0000..............................0000|0000|    IPv4 ADDRESS     | (compatible)
      //  +--------------------------------------+----+---------------------+
      //  |0000..............................0000|FFFF|    IPv4 ADDRESS     | (mapped)
      //  +--------------------------------------+----+---------------------+
      var str = "";
      var word = 0;
      var longest = 0;
      var lastzero = 0;
      var zstart = 0;
      var len = 0;
      var i = 0;
      var parts = [
        ints[0] & 0xffff,
        (ints[0] >> 16),
        ints[1] & 0xffff,
        (ints[1] >> 16),
        ints[2] & 0xffff,
        (ints[2] >> 16),
        ints[3] & 0xffff,
        (ints[3] >> 16)
      ];
  
      // Handle IPv4-compatible, IPv4-mapped, loopback and any/unspecified addresses
  
      var hasipv4 = true;
      var v4part = "";
      // check if the 10 high-order bytes are all zeros (first 5 words)
      for (i = 0; i < 5; i++) {
        if (parts[i] !== 0) { hasipv4 = false; break; }
      }
  
      if (hasipv4) {
        // low-order 32-bits store an IPv4 address (bytes 13 to 16) (last 2 words)
        v4part = __inet_ntop4_raw(parts[6] | (parts[7] << 16));
        // IPv4-mapped IPv6 address if 16-bit value (bytes 11 and 12) == 0xFFFF (6th word)
        if (parts[5] === -1) {
          str = "::ffff:";
          str += v4part;
          return str;
        }
        // IPv4-compatible IPv6 address if 16-bit value (bytes 11 and 12) == 0x0000 (6th word)
        if (parts[5] === 0) {
          str = "::";
          //special case IPv6 addresses
          if(v4part === "0.0.0.0") v4part = ""; // any/unspecified address
          if(v4part === "0.0.0.1") v4part = "1";// loopback address
          str += v4part;
          return str;
        }
      }
  
      // Handle all other IPv6 addresses
  
      // first run to find the longest contiguous zero words
      for (word = 0; word < 8; word++) {
        if (parts[word] === 0) {
          if (word - lastzero > 1) {
            len = 0;
          }
          lastzero = word;
          len++;
        }
        if (len > longest) {
          longest = len;
          zstart = word - longest + 1;
        }
      }
  
      for (word = 0; word < 8; word++) {
        if (longest > 1) {
          // compress contiguous zeros - to produce "::"
          if (parts[word] === 0 && word >= zstart && word < (zstart + longest) ) {
            if (word === zstart) {
              str += ":";
              if (zstart === 0) str += ":"; //leading zeros case
            }
            continue;
          }
        }
        // converts 16-bit words from big-endian to little-endian before converting to hex string
        str += Number(_ntohs(parts[word] & 0xffff)).toString(16);
        str += word < 7 ? ":" : "";
      }
      return str;
    }function __read_sockaddr(sa, salen) {
      // family / port offsets are common to both sockaddr_in and sockaddr_in6
      var family = HEAP16[((sa)>>1)];
      var port = _ntohs(HEAP16[(((sa)+(2))>>1)]);
      var addr;
  
      switch (family) {
        case 2:
          if (salen !== 16) {
            return { errno: ERRNO_CODES.EINVAL };
          }
          addr = HEAP32[(((sa)+(4))>>2)];
          addr = __inet_ntop4_raw(addr);
          break;
        case 10:
          if (salen !== 28) {
            return { errno: ERRNO_CODES.EINVAL };
          }
          addr = [
            HEAP32[(((sa)+(8))>>2)],
            HEAP32[(((sa)+(12))>>2)],
            HEAP32[(((sa)+(16))>>2)],
            HEAP32[(((sa)+(20))>>2)]
          ];
          addr = __inet_ntop6_raw(addr);
          break;
        default:
          return { errno: ERRNO_CODES.EAFNOSUPPORT };
      }
  
      return { family: family, addr: addr, port: port };
    }
  
  function __write_sockaddr(sa, family, addr, port) {
      switch (family) {
        case 2:
          addr = __inet_pton4_raw(addr);
          HEAP16[((sa)>>1)]=family;
          HEAP32[(((sa)+(4))>>2)]=addr;
          HEAP16[(((sa)+(2))>>1)]=_htons(port);
          break;
        case 10:
          addr = __inet_pton6_raw(addr);
          HEAP32[((sa)>>2)]=family;
          HEAP32[(((sa)+(8))>>2)]=addr[0];
          HEAP32[(((sa)+(12))>>2)]=addr[1];
          HEAP32[(((sa)+(16))>>2)]=addr[2];
          HEAP32[(((sa)+(20))>>2)]=addr[3];
          HEAP16[(((sa)+(2))>>1)]=_htons(port);
          HEAP32[(((sa)+(4))>>2)]=0;
          HEAP32[(((sa)+(24))>>2)]=0;
          break;
        default:
          return { errno: ERRNO_CODES.EAFNOSUPPORT };
      }
      // kind of lame, but let's match _read_sockaddr's interface
      return {};
    }function ___syscall102(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // socketcall
      var call = SYSCALLS.get(), socketvararg = SYSCALLS.get();
      // socketcalls pass the rest of the arguments in a struct
      SYSCALLS.varargs = socketvararg;
      switch (call) {
        case 1: { // socket
          var domain = SYSCALLS.get(), type = SYSCALLS.get(), protocol = SYSCALLS.get();
          var sock = SOCKFS.createSocket(domain, type, protocol);
          assert(sock.stream.fd < 64); // XXX ? select() assumes socket fd values are in 0..63
          return sock.stream.fd;
        }
        case 2: { // bind
          var sock = SYSCALLS.getSocketFromFD(), info = SYSCALLS.getSocketAddress();
          sock.sock_ops.bind(sock, info.addr, info.port);
          return 0;
        }
        case 3: { // connect
          var sock = SYSCALLS.getSocketFromFD(), info = SYSCALLS.getSocketAddress();
          sock.sock_ops.connect(sock, info.addr, info.port);
          return 0;
        }
        case 4: { // listen
          var sock = SYSCALLS.getSocketFromFD(), backlog = SYSCALLS.get();
          sock.sock_ops.listen(sock, backlog);
          return 0;
        }
        case 5: { // accept
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          var newsock = sock.sock_ops.accept(sock);
          if (addr) {
            var res = __write_sockaddr(addr, newsock.family, DNS.lookup_name(newsock.daddr), newsock.dport);
            assert(!res.errno);
          }
          return newsock.stream.fd;
        }
        case 6: { // getsockname
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          // TODO: sock.saddr should never be undefined, see TODO in websocket_sock_ops.getname
          var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(sock.saddr || '0.0.0.0'), sock.sport);
          assert(!res.errno);
          return 0;
        }
        case 7: { // getpeername
          var sock = SYSCALLS.getSocketFromFD(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          if (!sock.daddr) {
            return -ERRNO_CODES.ENOTCONN; // The socket is not connected.
          }
          var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(sock.daddr), sock.dport);
          assert(!res.errno);
          return 0;
        }
        case 11: { // sendto
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), length = SYSCALLS.get(), flags = SYSCALLS.get(), dest = SYSCALLS.getSocketAddress(true);
          if (!dest) {
            // send, no address provided
            return FS.write(sock.stream, HEAP8,message, length);
          } else {
            // sendto an address
            return sock.sock_ops.sendmsg(sock, HEAP8,message, length, dest.addr, dest.port);
          }
        }
        case 12: { // recvfrom
          var sock = SYSCALLS.getSocketFromFD(), buf = SYSCALLS.get(), len = SYSCALLS.get(), flags = SYSCALLS.get(), addr = SYSCALLS.get(), addrlen = SYSCALLS.get();
          var msg = sock.sock_ops.recvmsg(sock, len);
          if (!msg) return 0; // socket is closed
          if (addr) {
            var res = __write_sockaddr(addr, sock.family, DNS.lookup_name(msg.addr), msg.port);
            assert(!res.errno);
          }
          HEAPU8.set(msg.buffer, buf);
          return msg.buffer.byteLength;
        }
        case 14: { // setsockopt
          return -ERRNO_CODES.ENOPROTOOPT; // The option is unknown at the level indicated.
        }
        case 15: { // getsockopt
          var sock = SYSCALLS.getSocketFromFD(), level = SYSCALLS.get(), optname = SYSCALLS.get(), optval = SYSCALLS.get(), optlen = SYSCALLS.get();
          // Minimal getsockopt aimed at resolving https://github.com/kripken/emscripten/issues/2211
          // so only supports SOL_SOCKET with SO_ERROR.
          if (level === 1) {
            if (optname === 4) {
              HEAP32[((optval)>>2)]=sock.error;
              HEAP32[((optlen)>>2)]=4;
              sock.error = null; // Clear the error (The SO_ERROR option obtains and then clears this field).
              return 0;
            }
          }
          return -ERRNO_CODES.ENOPROTOOPT; // The option is unknown at the level indicated.
        }
        case 16: { // sendmsg
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), flags = SYSCALLS.get();
          var iov = HEAP32[(((message)+(8))>>2)];
          var num = HEAP32[(((message)+(12))>>2)];
          // read the address and port to send to
          var addr, port;
          var name = HEAP32[((message)>>2)];
          var namelen = HEAP32[(((message)+(4))>>2)];
          if (name) {
            var info = __read_sockaddr(name, namelen);
            if (info.errno) return -info.errno;
            port = info.port;
            addr = DNS.lookup_addr(info.addr) || info.addr;
          }
          // concatenate scatter-gather arrays into one message buffer
          var total = 0;
          for (var i = 0; i < num; i++) {
            total += HEAP32[(((iov)+((8 * i) + 4))>>2)];
          }
          var view = new Uint8Array(total);
          var offset = 0;
          for (var i = 0; i < num; i++) {
            var iovbase = HEAP32[(((iov)+((8 * i) + 0))>>2)];
            var iovlen = HEAP32[(((iov)+((8 * i) + 4))>>2)];
            for (var j = 0; j < iovlen; j++) {  
              view[offset++] = HEAP8[(((iovbase)+(j))>>0)];
            }
          }
          // write the buffer
          return sock.sock_ops.sendmsg(sock, view, 0, total, addr, port);
        }
        case 17: { // recvmsg
          var sock = SYSCALLS.getSocketFromFD(), message = SYSCALLS.get(), flags = SYSCALLS.get();
          var iov = HEAP32[(((message)+(8))>>2)];
          var num = HEAP32[(((message)+(12))>>2)];
          // get the total amount of data we can read across all arrays
          var total = 0;
          for (var i = 0; i < num; i++) {
            total += HEAP32[(((iov)+((8 * i) + 4))>>2)];
          }
          // try to read total data
          var msg = sock.sock_ops.recvmsg(sock, total);
          if (!msg) return 0; // socket is closed
  
          // TODO honor flags:
          // MSG_OOB
          // Requests out-of-band data. The significance and semantics of out-of-band data are protocol-specific.
          // MSG_PEEK
          // Peeks at the incoming message.
          // MSG_WAITALL
          // Requests that the function block until the full amount of data requested can be returned. The function may return a smaller amount of data if a signal is caught, if the connection is terminated, if MSG_PEEK was specified, or if an error is pending for the socket.
  
          // write the source address out
          var name = HEAP32[((message)>>2)];
          if (name) {
            var res = __write_sockaddr(name, sock.family, DNS.lookup_name(msg.addr), msg.port);
            assert(!res.errno);
          }
          // write the buffer out to the scatter-gather arrays
          var bytesRead = 0;
          var bytesRemaining = msg.buffer.byteLength;
          for (var i = 0; bytesRemaining > 0 && i < num; i++) {
            var iovbase = HEAP32[(((iov)+((8 * i) + 0))>>2)];
            var iovlen = HEAP32[(((iov)+((8 * i) + 4))>>2)];
            if (!iovlen) {
              continue;
            }
            var length = Math.min(iovlen, bytesRemaining);
            var buf = msg.buffer.subarray(bytesRead, bytesRead + length);
            HEAPU8.set(buf, iovbase + bytesRead);
            bytesRead += length;
            bytesRemaining -= length;
          }
  
          // TODO set msghdr.msg_flags
          // MSG_EOR
          // End of record was received (if supported by the protocol).
          // MSG_OOB
          // Out-of-band data was received.
          // MSG_TRUNC
          // Normal data was truncated.
          // MSG_CTRUNC
  
          return bytesRead;
        }
        default: abort('unsupported socketcall syscall ' + call);
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___gxx_personality_v0() {
    }

   

  function _pthread_cond_wait() { return 0; }

  function ___cxa_rethrow() {
      var ptr = EXCEPTIONS.caught.pop();
      if (!EXCEPTIONS.infos[ptr].rethrown) {
        // Only pop if the corresponding push was through rethrow_primary_exception
        EXCEPTIONS.caught.push(ptr)
        EXCEPTIONS.infos[ptr].rethrown = true;
      }
      EXCEPTIONS.last = ptr;
      throw ptr;
    }

   

   

  function ___syscall221(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // fcntl64
      var stream = SYSCALLS.getStreamFromFD(), cmd = SYSCALLS.get();
      switch (cmd) {
        case 0: {
          var arg = SYSCALLS.get();
          if (arg < 0) {
            return -ERRNO_CODES.EINVAL;
          }
          var newStream;
          newStream = FS.open(stream.path, stream.flags, 0, arg);
          return newStream.fd;
        }
        case 1:
        case 2:
          return 0;  // FD_CLOEXEC makes no sense for a single process.
        case 3:
          return stream.flags;
        case 4: {
          var arg = SYSCALLS.get();
          stream.flags |= arg;
          return 0;
        }
        case 12:
        case 12: {
          var arg = SYSCALLS.get();
          var offset = 0;
          // We're always unlocked.
          HEAP16[(((arg)+(offset))>>1)]=2;
          return 0;
        }
        case 13:
        case 14:
        case 13:
        case 14:
          return 0; // Pretend that the locking is successful.
        case 16:
        case 8:
          return -ERRNO_CODES.EINVAL; // These are for sockets. We don't have them fully implemented yet.
        case 9:
          // musl trusts getown return values, due to a bug where they must be, as they overlap with errors. just return -1 here, so fnctl() returns that, and we set errno ourselves.
          ___setErrNo(ERRNO_CODES.EINVAL);
          return -1;
        default: {
          return -ERRNO_CODES.EINVAL;
        }
      }
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

   


  function ___syscall142(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // newselect
      // readfds are supported,
      // writefds checks socket open status
      // exceptfds not supported
      // timeout is always 0 - fully async
      var nfds = SYSCALLS.get(), readfds = SYSCALLS.get(), writefds = SYSCALLS.get(), exceptfds = SYSCALLS.get(), timeout = SYSCALLS.get();
  
      assert(nfds <= 64, 'nfds must be less than or equal to 64');  // fd sets have 64 bits // TODO: this could be 1024 based on current musl headers
      assert(!exceptfds, 'exceptfds not supported');
  
      var total = 0;
      
      var srcReadLow = (readfds ? HEAP32[((readfds)>>2)] : 0),
          srcReadHigh = (readfds ? HEAP32[(((readfds)+(4))>>2)] : 0);
      var srcWriteLow = (writefds ? HEAP32[((writefds)>>2)] : 0),
          srcWriteHigh = (writefds ? HEAP32[(((writefds)+(4))>>2)] : 0);
      var srcExceptLow = (exceptfds ? HEAP32[((exceptfds)>>2)] : 0),
          srcExceptHigh = (exceptfds ? HEAP32[(((exceptfds)+(4))>>2)] : 0);
  
      var dstReadLow = 0,
          dstReadHigh = 0;
      var dstWriteLow = 0,
          dstWriteHigh = 0;
      var dstExceptLow = 0,
          dstExceptHigh = 0;
  
      var allLow = (readfds ? HEAP32[((readfds)>>2)] : 0) |
                   (writefds ? HEAP32[((writefds)>>2)] : 0) |
                   (exceptfds ? HEAP32[((exceptfds)>>2)] : 0);
      var allHigh = (readfds ? HEAP32[(((readfds)+(4))>>2)] : 0) |
                    (writefds ? HEAP32[(((writefds)+(4))>>2)] : 0) |
                    (exceptfds ? HEAP32[(((exceptfds)+(4))>>2)] : 0);
  
      function check(fd, low, high, val) {
        return (fd < 32 ? (low & val) : (high & val));
      }
  
      for (var fd = 0; fd < nfds; fd++) {
        var mask = 1 << (fd % 32);
        if (!(check(fd, allLow, allHigh, mask))) {
          continue;  // index isn't in the set
        }
  
        var stream = FS.getStream(fd);
        if (!stream) throw new FS.ErrnoError(ERRNO_CODES.EBADF);
  
        var flags = SYSCALLS.DEFAULT_POLLMASK;
  
        if (stream.stream_ops.poll) {
          flags = stream.stream_ops.poll(stream);
        }
  
        if ((flags & 1) && check(fd, srcReadLow, srcReadHigh, mask)) {
          fd < 32 ? (dstReadLow = dstReadLow | mask) : (dstReadHigh = dstReadHigh | mask);
          total++;
        }
        if ((flags & 4) && check(fd, srcWriteLow, srcWriteHigh, mask)) {
          fd < 32 ? (dstWriteLow = dstWriteLow | mask) : (dstWriteHigh = dstWriteHigh | mask);
          total++;
        }
        if ((flags & 2) && check(fd, srcExceptLow, srcExceptHigh, mask)) {
          fd < 32 ? (dstExceptLow = dstExceptLow | mask) : (dstExceptHigh = dstExceptHigh | mask);
          total++;
        }
      }
  
      if (readfds) {
        HEAP32[((readfds)>>2)]=dstReadLow;
        HEAP32[(((readfds)+(4))>>2)]=dstReadHigh;
      }
      if (writefds) {
        HEAP32[((writefds)>>2)]=dstWriteLow;
        HEAP32[(((writefds)+(4))>>2)]=dstWriteHigh;
      }
      if (exceptfds) {
        HEAP32[((exceptfds)>>2)]=dstExceptLow;
        HEAP32[(((exceptfds)+(4))>>2)]=dstExceptHigh;
      }
      
      return total;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall140(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // llseek
      var stream = SYSCALLS.getStreamFromFD(), offset_high = SYSCALLS.get(), offset_low = SYSCALLS.get(), result = SYSCALLS.get(), whence = SYSCALLS.get();
      // NOTE: offset_high is unused - Emscripten's off_t is 32-bit
      var offset = offset_low;
      FS.llseek(stream, offset, whence);
      HEAP32[((result)>>2)]=stream.position;
      if (stream.getdents && offset === 0 && whence === 0) stream.getdents = null; // reset readdir state
      return 0;
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  function ___syscall146(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // writev
      var stream = SYSCALLS.getStreamFromFD(), iov = SYSCALLS.get(), iovcnt = SYSCALLS.get();
      return SYSCALLS.doWritev(stream, iov, iovcnt);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  var _llvm_nacl_atomic_cmpxchg_i32=undefined;

  function ___syscall145(which, varargs) {SYSCALLS.varargs = varargs;
  try {
   // readv
      var stream = SYSCALLS.getStreamFromFD(), iov = SYSCALLS.get(), iovcnt = SYSCALLS.get();
      return SYSCALLS.doReadv(stream, iov, iovcnt);
    } catch (e) {
    if (typeof FS === 'undefined' || !(e instanceof FS.ErrnoError)) abort(e);
    return -e.errno;
  }
  }

  var ___dso_handle=STATICTOP; STATICTOP += 16;;
FS.staticInit();__ATINIT__.unshift(function() { if (!Module["noFSInit"] && !FS.init.initialized) FS.init() });__ATMAIN__.push(function() { FS.ignorePermissions = false });__ATEXIT__.push(function() { FS.quit() });Module["FS_createFolder"] = FS.createFolder;Module["FS_createPath"] = FS.createPath;Module["FS_createDataFile"] = FS.createDataFile;Module["FS_createPreloadedFile"] = FS.createPreloadedFile;Module["FS_createLazyFile"] = FS.createLazyFile;Module["FS_createLink"] = FS.createLink;Module["FS_createDevice"] = FS.createDevice;Module["FS_unlink"] = FS.unlink;;
__ATINIT__.unshift(function() { TTY.init() });__ATEXIT__.push(function() { TTY.shutdown() });;
if (ENVIRONMENT_IS_NODE) { var fs = require("fs"); var NODEJS_PATH = require("path"); NODEFS.staticInit(); };
___buildEnvironment(ENV);;
__ATINIT__.push(function() { SOCKFS.root = FS.mount(SOCKFS, {}, null); });;
DYNAMICTOP_PTR = allocate(1, "i32", ALLOC_STATIC);

STACK_BASE = STACKTOP = Runtime.alignMemory(STATICTOP);

STACK_MAX = STACK_BASE + TOTAL_STACK;

DYNAMIC_BASE = Runtime.alignMemory(STACK_MAX);

HEAP32[DYNAMICTOP_PTR>>2] = DYNAMIC_BASE;

staticSealed = true; // seal the static portion of memory

assert(DYNAMIC_BASE < TOTAL_MEMORY, "TOTAL_MEMORY not big enough for stack");


function nullFunc_iiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vi(x) { Module["printErr"]("Invalid function pointer called with signature 'vi'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_vii(x) { Module["printErr"]("Invalid function pointer called with signature 'vii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_ii(x) { Module["printErr"]("Invalid function pointer called with signature 'ii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iid(x) { Module["printErr"]("Invalid function pointer called with signature 'iid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iii(x) { Module["printErr"]("Invalid function pointer called with signature 'iii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_diii(x) { Module["printErr"]("Invalid function pointer called with signature 'diii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_i(x) { Module["printErr"]("Invalid function pointer called with signature 'i'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viii(x) { Module["printErr"]("Invalid function pointer called with signature 'viii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_v(x) { Module["printErr"]("Invalid function pointer called with signature 'v'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiiiiii(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiiiiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_iiiiid(x) { Module["printErr"]("Invalid function pointer called with signature 'iiiiid'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function nullFunc_viiii(x) { Module["printErr"]("Invalid function pointer called with signature 'viiii'. Perhaps this is an invalid value (e.g. caused by calling a virtual method on a NULL pointer)? Or calling a function with an incorrect type, which will fail? (it is worth building your source files with -Werror (warnings are errors), as warnings can indicate undefined behavior which can cause this)");  Module["printErr"]("Build with ASSERTIONS=2 for more info.");abort(x) }

function invoke_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    return Module["dynCall_iiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7);
}

function invoke_viiiii(index,a1,a2,a3,a4,a5) {
  try {
    Module["dynCall_viiiii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiii(index,a1,a2,a3,a4,a5) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5);
}

function invoke_iiiiiid(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiiiiid"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiid(index,a1,a2,a3,a4,a5,a6) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6);
}

function invoke_vi(index,a1) {
  try {
    Module["dynCall_vi"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_vi(index,a1) {
    Runtime.functionPointers[index](a1);
}

function invoke_vii(index,a1,a2) {
  try {
    Module["dynCall_vii"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_vii(index,a1,a2) {
    Runtime.functionPointers[index](a1,a2);
}

function invoke_iiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    return Module["dynCall_iiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
}

function invoke_ii(index,a1) {
  try {
    return Module["dynCall_ii"](index,a1);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_ii(index,a1) {
    return Runtime.functionPointers[index](a1);
}

function invoke_iiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  try {
    return Module["dynCall_iiiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
}

function invoke_iiii(index,a1,a2,a3) {
  try {
    return Module["dynCall_iiii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiii(index,a1,a2,a3) {
    return Runtime.functionPointers[index](a1,a2,a3);
}

function invoke_viiiiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  try {
    Module["dynCall_viiiiiiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15);
}

function invoke_viiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    Module["dynCall_viiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8);
}

function invoke_viiiiii(index,a1,a2,a3,a4,a5,a6) {
  try {
    Module["dynCall_viiiiii"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiiii(index,a1,a2,a3,a4,a5,a6) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6);
}

function invoke_iiiiiii(index,a1,a2,a3,a4,a5,a6) {
  try {
    return Module["dynCall_iiiiiii"](index,a1,a2,a3,a4,a5,a6);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiii(index,a1,a2,a3,a4,a5,a6) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6);
}

function invoke_iiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  try {
    return Module["dynCall_iiiiiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12);
}

function invoke_iid(index,a1,a2) {
  try {
    return Module["dynCall_iid"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iid(index,a1,a2) {
    return Runtime.functionPointers[index](a1,a2);
}

function invoke_viiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  try {
    Module["dynCall_viiiiiii"](index,a1,a2,a3,a4,a5,a6,a7);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7);
}

function invoke_viiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  try {
    Module["dynCall_viiiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
    Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8,a9,a10);
}

function invoke_iii(index,a1,a2) {
  try {
    return Module["dynCall_iii"](index,a1,a2);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iii(index,a1,a2) {
    return Runtime.functionPointers[index](a1,a2);
}

function invoke_iiiiii(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiiii"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiii(index,a1,a2,a3,a4,a5) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5);
}

function invoke_diii(index,a1,a2,a3) {
  try {
    return Module["dynCall_diii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_diii(index,a1,a2,a3) {
    return Runtime.functionPointers[index](a1,a2,a3);
}

function invoke_i(index) {
  try {
    return Module["dynCall_i"](index);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_i(index) {
    return Runtime.functionPointers[index]();
}

function invoke_iiiii(index,a1,a2,a3,a4) {
  try {
    return Module["dynCall_iiiii"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiii(index,a1,a2,a3,a4) {
    return Runtime.functionPointers[index](a1,a2,a3,a4);
}

function invoke_viii(index,a1,a2,a3) {
  try {
    Module["dynCall_viii"](index,a1,a2,a3);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viii(index,a1,a2,a3) {
    Runtime.functionPointers[index](a1,a2,a3);
}

function invoke_v(index) {
  try {
    Module["dynCall_v"](index);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_v(index) {
    Runtime.functionPointers[index]();
}

function invoke_iiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  try {
    return Module["dynCall_iiiiiiiii"](index,a1,a2,a3,a4,a5,a6,a7,a8);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5,a6,a7,a8);
}

function invoke_iiiiid(index,a1,a2,a3,a4,a5) {
  try {
    return Module["dynCall_iiiiid"](index,a1,a2,a3,a4,a5);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_iiiiid(index,a1,a2,a3,a4,a5) {
    return Runtime.functionPointers[index](a1,a2,a3,a4,a5);
}

function invoke_viiii(index,a1,a2,a3,a4) {
  try {
    Module["dynCall_viiii"](index,a1,a2,a3,a4);
  } catch(e) {
    if (typeof e !== 'number' && e !== 'longjmp') throw e;
    Module["setThrew"](1, 0);
  }
}

function jsCall_viiii(index,a1,a2,a3,a4) {
    Runtime.functionPointers[index](a1,a2,a3,a4);
}

Module.asmGlobalArg = { "Math": Math, "Int8Array": Int8Array, "Int16Array": Int16Array, "Int32Array": Int32Array, "Uint8Array": Uint8Array, "Uint16Array": Uint16Array, "Uint32Array": Uint32Array, "Float32Array": Float32Array, "Float64Array": Float64Array, "NaN": NaN, "Infinity": Infinity, "byteLength": byteLength };

Module.asmLibraryArg = { "abort": abort, "assert": assert, "enlargeMemory": enlargeMemory, "getTotalMemory": getTotalMemory, "abortOnCannotGrowMemory": abortOnCannotGrowMemory, "abortStackOverflow": abortStackOverflow, "nullFunc_iiiiiiii": nullFunc_iiiiiiii, "nullFunc_viiiii": nullFunc_viiiii, "nullFunc_iiiiiid": nullFunc_iiiiiid, "nullFunc_vi": nullFunc_vi, "nullFunc_vii": nullFunc_vii, "nullFunc_iiiiiiiiiii": nullFunc_iiiiiiiiiii, "nullFunc_ii": nullFunc_ii, "nullFunc_iiiiiiiiiiii": nullFunc_iiiiiiiiiiii, "nullFunc_iiii": nullFunc_iiii, "nullFunc_viiiiiiiiiiiiiii": nullFunc_viiiiiiiiiiiiiii, "nullFunc_viiiiiiii": nullFunc_viiiiiiii, "nullFunc_viiiiii": nullFunc_viiiiii, "nullFunc_iiiiiii": nullFunc_iiiiiii, "nullFunc_iiiiiiiiiiiii": nullFunc_iiiiiiiiiiiii, "nullFunc_iid": nullFunc_iid, "nullFunc_viiiiiii": nullFunc_viiiiiii, "nullFunc_viiiiiiiiii": nullFunc_viiiiiiiiii, "nullFunc_iii": nullFunc_iii, "nullFunc_iiiiii": nullFunc_iiiiii, "nullFunc_diii": nullFunc_diii, "nullFunc_i": nullFunc_i, "nullFunc_iiiii": nullFunc_iiiii, "nullFunc_viii": nullFunc_viii, "nullFunc_v": nullFunc_v, "nullFunc_iiiiiiiii": nullFunc_iiiiiiiii, "nullFunc_iiiiid": nullFunc_iiiiid, "nullFunc_viiii": nullFunc_viiii, "invoke_iiiiiiii": invoke_iiiiiiii, "jsCall_iiiiiiii": jsCall_iiiiiiii, "invoke_viiiii": invoke_viiiii, "jsCall_viiiii": jsCall_viiiii, "invoke_iiiiiid": invoke_iiiiiid, "jsCall_iiiiiid": jsCall_iiiiiid, "invoke_vi": invoke_vi, "jsCall_vi": jsCall_vi, "invoke_vii": invoke_vii, "jsCall_vii": jsCall_vii, "invoke_iiiiiiiiiii": invoke_iiiiiiiiiii, "jsCall_iiiiiiiiiii": jsCall_iiiiiiiiiii, "invoke_ii": invoke_ii, "jsCall_ii": jsCall_ii, "invoke_iiiiiiiiiiii": invoke_iiiiiiiiiiii, "jsCall_iiiiiiiiiiii": jsCall_iiiiiiiiiiii, "invoke_iiii": invoke_iiii, "jsCall_iiii": jsCall_iiii, "invoke_viiiiiiiiiiiiiii": invoke_viiiiiiiiiiiiiii, "jsCall_viiiiiiiiiiiiiii": jsCall_viiiiiiiiiiiiiii, "invoke_viiiiiiii": invoke_viiiiiiii, "jsCall_viiiiiiii": jsCall_viiiiiiii, "invoke_viiiiii": invoke_viiiiii, "jsCall_viiiiii": jsCall_viiiiii, "invoke_iiiiiii": invoke_iiiiiii, "jsCall_iiiiiii": jsCall_iiiiiii, "invoke_iiiiiiiiiiiii": invoke_iiiiiiiiiiiii, "jsCall_iiiiiiiiiiiii": jsCall_iiiiiiiiiiiii, "invoke_iid": invoke_iid, "jsCall_iid": jsCall_iid, "invoke_viiiiiii": invoke_viiiiiii, "jsCall_viiiiiii": jsCall_viiiiiii, "invoke_viiiiiiiiii": invoke_viiiiiiiiii, "jsCall_viiiiiiiiii": jsCall_viiiiiiiiii, "invoke_iii": invoke_iii, "jsCall_iii": jsCall_iii, "invoke_iiiiii": invoke_iiiiii, "jsCall_iiiiii": jsCall_iiiiii, "invoke_diii": invoke_diii, "jsCall_diii": jsCall_diii, "invoke_i": invoke_i, "jsCall_i": jsCall_i, "invoke_iiiii": invoke_iiiii, "jsCall_iiiii": jsCall_iiiii, "invoke_viii": invoke_viii, "jsCall_viii": jsCall_viii, "invoke_v": invoke_v, "jsCall_v": jsCall_v, "invoke_iiiiiiiii": invoke_iiiiiiiii, "jsCall_iiiiiiiii": jsCall_iiiiiiiii, "invoke_iiiiid": invoke_iiiiid, "jsCall_iiiiid": jsCall_iiiiid, "invoke_viiii": invoke_viiii, "jsCall_viiii": jsCall_viiii, "___syscall221": ___syscall221, "_strftime": _strftime, "_pthread_cond_wait": _pthread_cond_wait, "__inet_pton4_raw": __inet_pton4_raw, "__inet_ntop6_raw": __inet_ntop6_raw, "___syscall102": ___syscall102, "_pthread_key_create": _pthread_key_create, "__inet_ntop4_raw": __inet_ntop4_raw, "___gxx_personality_v0": ___gxx_personality_v0, "__inet_pton6_raw": __inet_pton6_raw, "__read_sockaddr": __read_sockaddr, "___assert_fail": ___assert_fail, "___cxa_free_exception": ___cxa_free_exception, "___cxa_allocate_exception": ___cxa_allocate_exception, "___cxa_find_matching_catch": ___cxa_find_matching_catch, "___buildEnvironment": ___buildEnvironment, "__isLeapYear": __isLeapYear, "__addDays": __addDays, "_strftime_l": _strftime_l, "_localtime_r": _localtime_r, "_tzset": _tzset, "___setErrNo": ___setErrNo, "___cxa_find_matching_catch_2": ___cxa_find_matching_catch_2, "_emscripten_memcpy_big": _emscripten_memcpy_big, "___cxa_end_catch": ___cxa_end_catch, "___resumeException": ___resumeException, "__ZSt18uncaught_exceptionv": __ZSt18uncaught_exceptionv, "___cxa_begin_catch": ___cxa_begin_catch, "_pthread_getspecific": _pthread_getspecific, "_clock": _clock, "__arraySum": __arraySum, "___syscall91": ___syscall91, "___cxa_find_matching_catch_3": ___cxa_find_matching_catch_3, "_pthread_once": _pthread_once, "_abort": _abort, "_getenv": _getenv, "___map_file": ___map_file, "___syscall54": ___syscall54, "___unlock": ___unlock, "___lock": ___lock, "__write_sockaddr": __write_sockaddr, "_emscripten_asm_const_iii": _emscripten_asm_const_iii, "_pthread_setspecific": _pthread_setspecific, "___cxa_atexit": ___cxa_atexit, "___cxa_throw": ___cxa_throw, "___cxa_rethrow": ___cxa_rethrow, "___syscall6": ___syscall6, "___cxa_pure_virtual": ___cxa_pure_virtual, "_time": _time, "_atexit": _atexit, "___syscall140": ___syscall140, "_localtime": _localtime, "___syscall142": ___syscall142, "___syscall145": ___syscall145, "___syscall146": ___syscall146, "DYNAMICTOP_PTR": DYNAMICTOP_PTR, "tempDoublePtr": tempDoublePtr, "ABORT": ABORT, "STACKTOP": STACKTOP, "STACK_MAX": STACK_MAX, "cttz_i8": cttz_i8, "___dso_handle": ___dso_handle };
// EMSCRIPTEN_START_ASM
var asm = (function(global, env, buffer) {
'almost asm';


  var Int8View = global.Int8Array;
  var HEAP8 = new Int8View(buffer);
  var Int16View = global.Int16Array;
  var HEAP16 = new Int16View(buffer);
  var Int32View = global.Int32Array;
  var HEAP32 = new Int32View(buffer);
  var Uint8View = global.Uint8Array;
  var HEAPU8 = new Uint8View(buffer);
  var Uint16View = global.Uint16Array;
  var HEAPU16 = new Uint16View(buffer);
  var Uint32View = global.Uint32Array;
  var HEAPU32 = new Uint32View(buffer);
  var Float32View = global.Float32Array;
  var HEAPF32 = new Float32View(buffer);
  var Float64View = global.Float64Array;
  var HEAPF64 = new Float64View(buffer);
  var byteLength = global.byteLength;

  var DYNAMICTOP_PTR=env.DYNAMICTOP_PTR|0;
  var tempDoublePtr=env.tempDoublePtr|0;
  var ABORT=env.ABORT|0;
  var STACKTOP=env.STACKTOP|0;
  var STACK_MAX=env.STACK_MAX|0;
  var cttz_i8=env.cttz_i8|0;
  var ___dso_handle=env.___dso_handle|0;

  var __THREW__ = 0;
  var threwValue = 0;
  var setjmpId = 0;
  var undef = 0;
  var nan = global.NaN, inf = global.Infinity;
  var tempInt = 0, tempBigInt = 0, tempBigIntS = 0, tempValue = 0, tempDouble = 0.0;
  var tempRet0 = 0;

  var Math_floor=global.Math.floor;
  var Math_abs=global.Math.abs;
  var Math_sqrt=global.Math.sqrt;
  var Math_pow=global.Math.pow;
  var Math_cos=global.Math.cos;
  var Math_sin=global.Math.sin;
  var Math_tan=global.Math.tan;
  var Math_acos=global.Math.acos;
  var Math_asin=global.Math.asin;
  var Math_atan=global.Math.atan;
  var Math_atan2=global.Math.atan2;
  var Math_exp=global.Math.exp;
  var Math_log=global.Math.log;
  var Math_ceil=global.Math.ceil;
  var Math_imul=global.Math.imul;
  var Math_min=global.Math.min;
  var Math_max=global.Math.max;
  var Math_clz32=global.Math.clz32;
  var abort=env.abort;
  var assert=env.assert;
  var enlargeMemory=env.enlargeMemory;
  var getTotalMemory=env.getTotalMemory;
  var abortOnCannotGrowMemory=env.abortOnCannotGrowMemory;
  var abortStackOverflow=env.abortStackOverflow;
  var nullFunc_iiiiiiii=env.nullFunc_iiiiiiii;
  var nullFunc_viiiii=env.nullFunc_viiiii;
  var nullFunc_iiiiiid=env.nullFunc_iiiiiid;
  var nullFunc_vi=env.nullFunc_vi;
  var nullFunc_vii=env.nullFunc_vii;
  var nullFunc_iiiiiiiiiii=env.nullFunc_iiiiiiiiiii;
  var nullFunc_ii=env.nullFunc_ii;
  var nullFunc_iiiiiiiiiiii=env.nullFunc_iiiiiiiiiiii;
  var nullFunc_iiii=env.nullFunc_iiii;
  var nullFunc_viiiiiiiiiiiiiii=env.nullFunc_viiiiiiiiiiiiiii;
  var nullFunc_viiiiiiii=env.nullFunc_viiiiiiii;
  var nullFunc_viiiiii=env.nullFunc_viiiiii;
  var nullFunc_iiiiiii=env.nullFunc_iiiiiii;
  var nullFunc_iiiiiiiiiiiii=env.nullFunc_iiiiiiiiiiiii;
  var nullFunc_iid=env.nullFunc_iid;
  var nullFunc_viiiiiii=env.nullFunc_viiiiiii;
  var nullFunc_viiiiiiiiii=env.nullFunc_viiiiiiiiii;
  var nullFunc_iii=env.nullFunc_iii;
  var nullFunc_iiiiii=env.nullFunc_iiiiii;
  var nullFunc_diii=env.nullFunc_diii;
  var nullFunc_i=env.nullFunc_i;
  var nullFunc_iiiii=env.nullFunc_iiiii;
  var nullFunc_viii=env.nullFunc_viii;
  var nullFunc_v=env.nullFunc_v;
  var nullFunc_iiiiiiiii=env.nullFunc_iiiiiiiii;
  var nullFunc_iiiiid=env.nullFunc_iiiiid;
  var nullFunc_viiii=env.nullFunc_viiii;
  var invoke_iiiiiiii=env.invoke_iiiiiiii;
  var jsCall_iiiiiiii=env.jsCall_iiiiiiii;
  var invoke_viiiii=env.invoke_viiiii;
  var jsCall_viiiii=env.jsCall_viiiii;
  var invoke_iiiiiid=env.invoke_iiiiiid;
  var jsCall_iiiiiid=env.jsCall_iiiiiid;
  var invoke_vi=env.invoke_vi;
  var jsCall_vi=env.jsCall_vi;
  var invoke_vii=env.invoke_vii;
  var jsCall_vii=env.jsCall_vii;
  var invoke_iiiiiiiiiii=env.invoke_iiiiiiiiiii;
  var jsCall_iiiiiiiiiii=env.jsCall_iiiiiiiiiii;
  var invoke_ii=env.invoke_ii;
  var jsCall_ii=env.jsCall_ii;
  var invoke_iiiiiiiiiiii=env.invoke_iiiiiiiiiiii;
  var jsCall_iiiiiiiiiiii=env.jsCall_iiiiiiiiiiii;
  var invoke_iiii=env.invoke_iiii;
  var jsCall_iiii=env.jsCall_iiii;
  var invoke_viiiiiiiiiiiiiii=env.invoke_viiiiiiiiiiiiiii;
  var jsCall_viiiiiiiiiiiiiii=env.jsCall_viiiiiiiiiiiiiii;
  var invoke_viiiiiiii=env.invoke_viiiiiiii;
  var jsCall_viiiiiiii=env.jsCall_viiiiiiii;
  var invoke_viiiiii=env.invoke_viiiiii;
  var jsCall_viiiiii=env.jsCall_viiiiii;
  var invoke_iiiiiii=env.invoke_iiiiiii;
  var jsCall_iiiiiii=env.jsCall_iiiiiii;
  var invoke_iiiiiiiiiiiii=env.invoke_iiiiiiiiiiiii;
  var jsCall_iiiiiiiiiiiii=env.jsCall_iiiiiiiiiiiii;
  var invoke_iid=env.invoke_iid;
  var jsCall_iid=env.jsCall_iid;
  var invoke_viiiiiii=env.invoke_viiiiiii;
  var jsCall_viiiiiii=env.jsCall_viiiiiii;
  var invoke_viiiiiiiiii=env.invoke_viiiiiiiiii;
  var jsCall_viiiiiiiiii=env.jsCall_viiiiiiiiii;
  var invoke_iii=env.invoke_iii;
  var jsCall_iii=env.jsCall_iii;
  var invoke_iiiiii=env.invoke_iiiiii;
  var jsCall_iiiiii=env.jsCall_iiiiii;
  var invoke_diii=env.invoke_diii;
  var jsCall_diii=env.jsCall_diii;
  var invoke_i=env.invoke_i;
  var jsCall_i=env.jsCall_i;
  var invoke_iiiii=env.invoke_iiiii;
  var jsCall_iiiii=env.jsCall_iiiii;
  var invoke_viii=env.invoke_viii;
  var jsCall_viii=env.jsCall_viii;
  var invoke_v=env.invoke_v;
  var jsCall_v=env.jsCall_v;
  var invoke_iiiiiiiii=env.invoke_iiiiiiiii;
  var jsCall_iiiiiiiii=env.jsCall_iiiiiiiii;
  var invoke_iiiiid=env.invoke_iiiiid;
  var jsCall_iiiiid=env.jsCall_iiiiid;
  var invoke_viiii=env.invoke_viiii;
  var jsCall_viiii=env.jsCall_viiii;
  var ___syscall221=env.___syscall221;
  var _strftime=env._strftime;
  var _pthread_cond_wait=env._pthread_cond_wait;
  var __inet_pton4_raw=env.__inet_pton4_raw;
  var __inet_ntop6_raw=env.__inet_ntop6_raw;
  var ___syscall102=env.___syscall102;
  var _pthread_key_create=env._pthread_key_create;
  var __inet_ntop4_raw=env.__inet_ntop4_raw;
  var ___gxx_personality_v0=env.___gxx_personality_v0;
  var __inet_pton6_raw=env.__inet_pton6_raw;
  var __read_sockaddr=env.__read_sockaddr;
  var ___assert_fail=env.___assert_fail;
  var ___cxa_free_exception=env.___cxa_free_exception;
  var ___cxa_allocate_exception=env.___cxa_allocate_exception;
  var ___cxa_find_matching_catch=env.___cxa_find_matching_catch;
  var ___buildEnvironment=env.___buildEnvironment;
  var __isLeapYear=env.__isLeapYear;
  var __addDays=env.__addDays;
  var _strftime_l=env._strftime_l;
  var _localtime_r=env._localtime_r;
  var _tzset=env._tzset;
  var ___setErrNo=env.___setErrNo;
  var ___cxa_find_matching_catch_2=env.___cxa_find_matching_catch_2;
  var _emscripten_memcpy_big=env._emscripten_memcpy_big;
  var ___cxa_end_catch=env.___cxa_end_catch;
  var ___resumeException=env.___resumeException;
  var __ZSt18uncaught_exceptionv=env.__ZSt18uncaught_exceptionv;
  var ___cxa_begin_catch=env.___cxa_begin_catch;
  var _pthread_getspecific=env._pthread_getspecific;
  var _clock=env._clock;
  var __arraySum=env.__arraySum;
  var ___syscall91=env.___syscall91;
  var ___cxa_find_matching_catch_3=env.___cxa_find_matching_catch_3;
  var _pthread_once=env._pthread_once;
  var _abort=env._abort;
  var _getenv=env._getenv;
  var ___map_file=env.___map_file;
  var ___syscall54=env.___syscall54;
  var ___unlock=env.___unlock;
  var ___lock=env.___lock;
  var __write_sockaddr=env.__write_sockaddr;
  var _emscripten_asm_const_iii=env._emscripten_asm_const_iii;
  var _pthread_setspecific=env._pthread_setspecific;
  var ___cxa_atexit=env.___cxa_atexit;
  var ___cxa_throw=env.___cxa_throw;
  var ___cxa_rethrow=env.___cxa_rethrow;
  var ___syscall6=env.___syscall6;
  var ___cxa_pure_virtual=env.___cxa_pure_virtual;
  var _time=env._time;
  var _atexit=env._atexit;
  var ___syscall140=env.___syscall140;
  var _localtime=env._localtime;
  var ___syscall142=env.___syscall142;
  var ___syscall145=env.___syscall145;
  var ___syscall146=env.___syscall146;
  var tempFloat = 0.0;

function _emscripten_replace_memory(newBuffer) {
  if ((byteLength(newBuffer) & 0xffffff || byteLength(newBuffer) <= 0xffffff) || byteLength(newBuffer) > 0x80000000) return false;
  HEAP8 = new Int8View(newBuffer);
  HEAP16 = new Int16View(newBuffer);
  HEAP32 = new Int32View(newBuffer);
  HEAPU8 = new Uint8View(newBuffer);
  HEAPU16 = new Uint16View(newBuffer);
  HEAPU32 = new Uint32View(newBuffer);
  HEAPF32 = new Float32View(newBuffer);
  HEAPF64 = new Float64View(newBuffer);
  buffer = newBuffer;
  return true;
}

// EMSCRIPTEN_START_FUNCS

function stackAlloc(size) {
  size = size|0;
  var ret = 0;
  ret = STACKTOP;
  STACKTOP = (STACKTOP + size)|0;
  STACKTOP = (STACKTOP + 15)&-16;
  if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(size|0);

  return ret|0;
}
function stackSave() {
  return STACKTOP|0;
}
function stackRestore(top) {
  top = top|0;
  STACKTOP = top;
}
function establishStackSpace(stackBase, stackMax) {
  stackBase = stackBase|0;
  stackMax = stackMax|0;
  STACKTOP = stackBase;
  STACK_MAX = stackMax;
}

function setThrew(threw, value) {
  threw = threw|0;
  value = value|0;
  if ((__THREW__|0) == 0) {
    __THREW__ = threw;
    threwValue = value;
  }
}

function setTempRet0(value) {
  value = value|0;
  tempRet0 = value;
}
function getTempRet0() {
  return tempRet0|0;
}

function ___cxx_global_var_init() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $3 = sp + 13|0;
 $4 = sp + 12|0;
 $2 = 161656;
 $5 = $2;
 ;HEAP8[$1>>0]=HEAP8[$4>>0]|0;
 $0 = $3;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEEC2ERKSC_($5,$3);
 (___cxa_atexit((727|0),(161656|0),(___dso_handle|0))|0);
 STACKTOP = sp;return;
}
function __ZNSt3__23mapIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMNS_4lessIS4_EENS_9allocatorINS_4pairIKS4_S6_EEEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEED2Ev($2);
 STACKTOP = sp;return;
}
function _PhotonPeerCreate($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $_ZN9ExitGames6Photon18ConnectionProtocolL3WSSE$_ZN9ExitGames6Photon18ConnectionProtocolL2WSE = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $8 = sp + 20|0;
 $9 = sp + 16|0;
 $10 = sp + 12|0;
 $11 = sp + 8|0;
 $13 = sp;
 $14 = $0&1;
 $7 = $14;
 HEAP32[$8>>2] = $1;
 HEAP32[$9>>2] = $2;
 HEAP32[$10>>2] = $3;
 HEAP32[$11>>2] = $4;
 $15 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,143796)|0);
 $16 = HEAP32[$8>>2]|0;
 $17 = ($16|0)!=(0|0);
 $18 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEb($15,$17)|0);
 $19 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($18,159115)|0);
 $20 = HEAP32[$9>>2]|0;
 $21 = ($20|0)!=(0|0);
 $22 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEb($19,$21)|0);
 $23 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($22,159115)|0);
 $24 = HEAP32[$10>>2]|0;
 $25 = ($24|0)!=(0|0);
 $26 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEb($23,$25)|0);
 $27 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($26,159115)|0);
 $28 = HEAP32[$11>>2]|0;
 $29 = ($28|0)!=(0|0);
 $30 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEb($27,$29)|0);
 $5 = $30;
 $6 = 728;
 $31 = $5;
 $32 = $6;
 (FUNCTION_TABLE_ii[$32 & 2047]($31)|0);
 $33 = (__ZN9ExitGames6Common16MemoryManagement8allocateI16PhotonListenerEMPFviEPFviPKciPKNS0_10DictionaryIhNS0_6ObjectEEEEPFviSC_EPFviS7_EEEPT_RT0_RT1_RT2_RT3_($8,$9,$10,$11)|0);
 $12 = $33;
 $34 = $12;
 $35 = $7;
 $36 = $35&1;
 $_ZN9ExitGames6Photon18ConnectionProtocolL3WSSE$_ZN9ExitGames6Photon18ConnectionProtocolL2WSE = $36 ? 143819 : 143820;
 $37 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS_6Photon10PhotonPeerE16PhotonListenerEMKhEEPT_RT0_RT1_($34,$_ZN9ExitGames6Photon18ConnectionProtocolL3WSSE$_ZN9ExitGames6Photon18ConnectionProtocolL2WSE)|0);
 HEAP32[$13>>2] = $37;
 $38 = $12;
 $39 = (__ZNSt3__23mapIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMNS_4lessIS4_EENS_9allocatorINS_4pairIKS4_S6_EEEEEixERSB_(161656,$13)|0);
 HEAP32[$39>>2] = $38;
 $40 = HEAP32[$13>>2]|0;
 STACKTOP = sp;return ($40|0);
}
function __ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$byval_copy = 0, $$sink3 = 0, $$sink4 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 752|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(752|0);
 $$byval_copy = sp + 732|0;
 $12 = sp + 696|0;
 $26 = sp + 648|0;
 $62 = sp + 504|0;
 $63 = sp + 500|0;
 $66 = sp + 488|0;
 $67 = sp + 484|0;
 $75 = sp + 448|0;
 $80 = sp + 32|0;
 $82 = sp + 16|0;
 $83 = sp + 12|0;
 $85 = sp + 4|0;
 $86 = sp;
 $73 = $0;
 $74 = $1;
 $87 = $73;
 __THREW__ = 0;
 invoke_vii(729,($75|0),($87|0));
 $88 = __THREW__; __THREW__ = 0;
 $89 = $88&1;
 L1: do {
  if ($89) {
   $121 = ___cxa_find_matching_catch_3(0|0)|0;
   $122 = tempRet0;
   $76 = $121;
   $77 = $122;
  } else {
   $72 = $75;
   $90 = $72;
   $91 = HEAP8[$90>>0]|0;
   $92 = $91&1;
   L4: do {
    if ($92) {
     $93 = $74;
     $94 = (__ZNSt3__211char_traitsIcE6lengthEPKc($93)|0);
     $78 = $94;
     $79 = 100;
     $81 = $80;
     HEAP32[$83>>2] = 730;
     $69 = $82;
     $70 = 0;
     $71 = $83;
     $95 = $69;
     $96 = $70;
     $97 = $71;
     $68 = $97;
     $98 = $68;
     $99 = HEAP32[$98>>2]|0;
     $65 = $95;
     HEAP32[$66>>2] = $96;
     HEAP32[$67>>2] = $99;
     $100 = $65;
     $64 = $66;
     $101 = $64;
     $102 = HEAP32[$101>>2]|0;
     $58 = $67;
     $103 = $58;
     $104 = HEAP32[$103>>2]|0;
     $61 = $100;
     HEAP32[$62>>2] = $102;
     HEAP32[$63>>2] = $104;
     $105 = $61;
     $60 = $62;
     $106 = $60;
     $107 = HEAP32[$106>>2]|0;
     HEAP32[$105>>2] = $107;
     $108 = ((($105)) + 4|0);
     $59 = $63;
     $109 = $59;
     $110 = HEAP32[$109>>2]|0;
     HEAP32[$108>>2] = $110;
     $111 = $78;
     $112 = ($111>>>0)>(100);
     do {
      if ($112) {
       $113 = $78;
       $114 = $113<<2;
       __THREW__ = 0;
       $115 = (invoke_ii(731,($114|0))|0);
       $116 = __THREW__; __THREW__ = 0;
       $117 = $116&1;
       if ($117) {
        label = 8;
       } else {
        $81 = $115;
        $118 = $81;
        $119 = ($118|0)==(0|0);
        if ($119) {
         __THREW__ = 0;
         invoke_v(732);
         $120 = __THREW__; __THREW__ = 0;
         label = 8;
         break;
        }
        $144 = $81;
        $35 = $82;
        $36 = $144;
        $145 = $35;
        $34 = $145;
        $146 = $34;
        $33 = $146;
        $147 = $33;
        $148 = HEAP32[$147>>2]|0;
        $37 = $148;
        $149 = $36;
        $32 = $145;
        $150 = $32;
        $31 = $150;
        $151 = $31;
        HEAP32[$151>>2] = $149;
        $152 = $37;
        $153 = ($152|0)!=(0|0);
        if ($153) {
         $30 = $145;
         $154 = $30;
         $29 = $154;
         $155 = $29;
         $156 = ((($155)) + 4|0);
         $157 = HEAP32[$156>>2]|0;
         $158 = $37;
         __THREW__ = 0;
         invoke_vi($157|0,($158|0));
         $159 = __THREW__; __THREW__ = 0;
         $160 = $159&1;
         if ($160) {
          $161 = ___cxa_find_matching_catch_3(0|0)|0;
          $162 = tempRet0;
          ___clang_call_terminate($161);
          // unreachable;
         } else {
          label = 15;
         }
        } else {
         label = 15;
        }
       }
      } else {
       label = 15;
      }
     } while(0);
     L14: do {
      if ((label|0) == 15) {
       $163 = $81;
       $84 = $163;
       while(1) {
        $164 = $74;
        $165 = HEAP8[$164>>0]|0;
        $166 = $165 << 24 >> 24;
        $167 = ($166|0)!=(0);
        $168 = $73;
        if (!($167)) {
         break;
        }
        $169 = HEAP32[$168>>2]|0;
        $170 = ((($169)) + -12|0);
        $171 = HEAP32[$170>>2]|0;
        $172 = (($168) + ($171)|0);
        $173 = $74;
        $174 = HEAP8[$173>>0]|0;
        $24 = $172;
        $25 = $174;
        $175 = $24;
        __THREW__ = 0;
        invoke_vii(733,($26|0),($175|0));
        $176 = __THREW__; __THREW__ = 0;
        $177 = $176&1;
        if ($177) {
         label = 8;
         break L14;
        }
        $23 = $26;
        $178 = $23;
        __THREW__ = 0;
        $179 = (invoke_iii(734,($178|0),(165472|0))|0);
        $180 = __THREW__; __THREW__ = 0;
        $181 = $180&1;
        if ($181) {
         label = 20;
         break;
        }
        $182 = $25;
        $21 = $179;
        $22 = $182;
        $183 = $21;
        $184 = HEAP32[$183>>2]|0;
        $185 = ((($184)) + 44|0);
        $186 = HEAP32[$185>>2]|0;
        $187 = $22;
        __THREW__ = 0;
        $188 = (invoke_iii($186|0,($183|0),($187|0))|0);
        $189 = __THREW__; __THREW__ = 0;
        $190 = $189&1;
        if ($190) {
         label = 20;
         break;
        }
        __ZNSt3__26localeD2Ev($26);
        $195 = $84;
        HEAP32[$195>>2] = $188;
        $196 = $74;
        $197 = ((($196)) + 1|0);
        $74 = $197;
        $198 = $84;
        $199 = ((($198)) + 4|0);
        $84 = $199;
       }
       if ((label|0) == 20) {
        $191 = ___cxa_find_matching_catch_3(0|0)|0;
        $192 = tempRet0;
        $27 = $191;
        $28 = $192;
        __ZNSt3__26localeD2Ev($26);
        $193 = $27;
        $194 = $28;
        $$sink3 = $194;$$sink4 = $193;
        break;
       }
       $19 = $86;
       $20 = $168;
       $200 = $19;
       $201 = $20;
       $202 = HEAP32[$201>>2]|0;
       $203 = ((($202)) + -12|0);
       $204 = HEAP32[$203>>2]|0;
       $205 = (($201) + ($204)|0);
       $18 = $205;
       $206 = $18;
       $17 = $206;
       $207 = $17;
       $208 = ((($207)) + 24|0);
       $209 = HEAP32[$208>>2]|0;
       HEAP32[$200>>2] = $209;
       $210 = $81;
       $211 = $73;
       $212 = HEAP32[$211>>2]|0;
       $213 = ((($212)) + -12|0);
       $214 = HEAP32[$213>>2]|0;
       $215 = (($211) + ($214)|0);
       $16 = $215;
       $216 = $16;
       $217 = ((($216)) + 4|0);
       $218 = HEAP32[$217>>2]|0;
       $219 = $218 & 176;
       $220 = ($219|0)==(32);
       $221 = $81;
       $222 = $78;
       $223 = (($221) + ($222<<2)|0);
       $224 = $220 ? $223 : $221;
       $225 = $81;
       $226 = $78;
       $227 = (($225) + ($226<<2)|0);
       $228 = $73;
       $229 = HEAP32[$228>>2]|0;
       $230 = ((($229)) + -12|0);
       $231 = HEAP32[$230>>2]|0;
       $232 = (($228) + ($231)|0);
       $233 = $73;
       $234 = HEAP32[$233>>2]|0;
       $235 = ((($234)) + -12|0);
       $236 = HEAP32[$235>>2]|0;
       $237 = (($233) + ($236)|0);
       $15 = $237;
       $238 = $15;
       $239 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $240 = ((($238)) + 76|0);
       $241 = HEAP32[$240>>2]|0;
       $242 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($239,$241)|0);
       do {
        if ($242) {
         $10 = $238;
         $11 = 32;
         $243 = $10;
         __THREW__ = 0;
         invoke_vii(733,($12|0),($243|0));
         $244 = __THREW__; __THREW__ = 0;
         $245 = $244&1;
         if ($245) {
          label = 8;
          break L14;
         }
         $9 = $12;
         $246 = $9;
         __THREW__ = 0;
         $247 = (invoke_iii(734,($246|0),(165472|0))|0);
         $248 = __THREW__; __THREW__ = 0;
         $249 = $248&1;
         if (!($249)) {
          $250 = $11;
          $7 = $247;
          $8 = $250;
          $251 = $7;
          $252 = HEAP32[$251>>2]|0;
          $253 = ((($252)) + 44|0);
          $254 = HEAP32[$253>>2]|0;
          $255 = $8;
          __THREW__ = 0;
          $256 = (invoke_iii($254|0,($251|0),($255|0))|0);
          $257 = __THREW__; __THREW__ = 0;
          $258 = $257&1;
          if (!($258)) {
           __ZNSt3__26localeD2Ev($12);
           $263 = ((($238)) + 76|0);
           HEAP32[$263>>2] = $256;
           break;
          }
         }
         $259 = ___cxa_find_matching_catch_3(0|0)|0;
         $260 = tempRet0;
         $13 = $259;
         $14 = $260;
         __ZNSt3__26localeD2Ev($12);
         $261 = $13;
         $262 = $14;
         $$sink3 = $262;$$sink4 = $261;
         break L14;
        }
       } while(0);
       $264 = ((($238)) + 76|0);
       $265 = HEAP32[$264>>2]|0;
       __THREW__ = 0;
       ;HEAP32[$$byval_copy>>2]=HEAP32[$86>>2]|0;
       $266 = (invoke_iiiiiii(735,($$byval_copy|0),($210|0),($224|0),($227|0),($232|0),($265|0))|0);
       $267 = __THREW__; __THREW__ = 0;
       $268 = $267&1;
       if ($268) {
        label = 8;
       } else {
        HEAP32[$85>>2] = $266;
        $6 = $85;
        $269 = $6;
        $270 = HEAP32[$269>>2]|0;
        $271 = ($270|0)==(0|0);
        if ($271) {
         $272 = $73;
         $273 = HEAP32[$272>>2]|0;
         $274 = ((($273)) + -12|0);
         $275 = HEAP32[$274>>2]|0;
         $276 = (($272) + ($275)|0);
         $4 = $276;
         $5 = 5;
         $277 = $4;
         $278 = $5;
         $2 = $277;
         $3 = $278;
         $279 = $2;
         $280 = ((($279)) + 16|0);
         $281 = HEAP32[$280>>2]|0;
         $282 = $3;
         $283 = $281 | $282;
         __THREW__ = 0;
         invoke_vii(736,($279|0),($283|0));
         $284 = __THREW__; __THREW__ = 0;
         $285 = $284&1;
         if ($285) {
          label = 8;
          break;
         }
        }
        $57 = $82;
        $286 = $57;
        $54 = $286;
        $55 = 0;
        $287 = $54;
        $53 = $287;
        $288 = $53;
        $52 = $288;
        $289 = $52;
        $290 = HEAP32[$289>>2]|0;
        $56 = $290;
        $291 = $55;
        $51 = $287;
        $292 = $51;
        $50 = $292;
        $293 = $50;
        HEAP32[$293>>2] = $291;
        $294 = $56;
        $295 = ($294|0)!=(0|0);
        if (!($295)) {
         break L4;
        }
        $49 = $287;
        $296 = $49;
        $48 = $296;
        $297 = $48;
        $298 = ((($297)) + 4|0);
        $299 = HEAP32[$298>>2]|0;
        $300 = $56;
        __THREW__ = 0;
        invoke_vi($299|0,($300|0));
        $301 = __THREW__; __THREW__ = 0;
        $302 = $301&1;
        if (!($302)) {
         break L4;
        }
        $303 = ___cxa_find_matching_catch_3(0|0)|0;
        $304 = tempRet0;
        ___clang_call_terminate($303);
        // unreachable;
       }
      }
     } while(0);
     if ((label|0) == 8) {
      $123 = ___cxa_find_matching_catch_3(0|0)|0;
      $124 = tempRet0;
      $$sink3 = $124;$$sink4 = $123;
     }
     $76 = $$sink4;
     $77 = $$sink3;
     $47 = $82;
     $125 = $47;
     $44 = $125;
     $45 = 0;
     $126 = $44;
     $43 = $126;
     $127 = $43;
     $42 = $127;
     $128 = $42;
     $129 = HEAP32[$128>>2]|0;
     $46 = $129;
     $130 = $45;
     $41 = $126;
     $131 = $41;
     $40 = $131;
     $132 = $40;
     HEAP32[$132>>2] = $130;
     $133 = $46;
     $134 = ($133|0)!=(0|0);
     if ($134) {
      $39 = $126;
      $135 = $39;
      $38 = $135;
      $136 = $38;
      $137 = ((($136)) + 4|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = $46;
      __THREW__ = 0;
      invoke_vi($138|0,($139|0));
      $140 = __THREW__; __THREW__ = 0;
      $141 = $140&1;
      if ($141) {
       $142 = ___cxa_find_matching_catch_3(0|0)|0;
       $143 = tempRet0;
       ___clang_call_terminate($142);
       // unreachable;
      }
     }
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($75);
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($75);
   $313 = $73;
   STACKTOP = sp;return ($313|0);
  }
 } while(0);
 $305 = $76;
 (___cxa_begin_catch(($305|0))|0);
 $306 = $73;
 $307 = HEAP32[$306>>2]|0;
 $308 = ((($307)) + -12|0);
 $309 = HEAP32[$308>>2]|0;
 $310 = (($306) + ($309)|0);
 __THREW__ = 0;
 invoke_vi(737,($310|0));
 $311 = __THREW__; __THREW__ = 0;
 $312 = $311&1;
 if (!($312)) {
  ___cxa_end_catch();
  $313 = $73;
  STACKTOP = sp;return ($313|0);
 }
 $314 = ___cxa_find_matching_catch_2()|0;
 $315 = tempRet0;
 $76 = $314;
 $77 = $315;
 __THREW__ = 0;
 invoke_v(738);
 $316 = __THREW__; __THREW__ = 0;
 $317 = $316&1;
 if ($317) {
  $320 = ___cxa_find_matching_catch_3(0|0)|0;
  $321 = tempRet0;
  ___clang_call_terminate($320);
  // unreachable;
 } else {
  $318 = $76;
  $319 = $77;
  ___resumeException($318|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__24endlIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $6 = sp + 12|0;
 $9 = $0;
 $10 = $9;
 $11 = $9;
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + -12|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (($11) + ($14)|0);
 $4 = $15;
 $5 = 10;
 $16 = $4;
 __ZNKSt3__28ios_base6getlocEv($6,$16);
 $3 = $6;
 $17 = $3;
 __THREW__ = 0;
 $18 = (invoke_iii(734,($17|0),(165472|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  $7 = $30;
  $8 = $31;
  __ZNSt3__26localeD2Ev($6);
  $32 = $7;
  $33 = $8;
  ___resumeException($32|0);
  // unreachable;
 }
 $21 = $5;
 $1 = $18;
 $2 = $21;
 $22 = $1;
 $23 = HEAP32[$22>>2]|0;
 $24 = ((($23)) + 44|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = $2;
 __THREW__ = 0;
 $27 = (invoke_iii($25|0,($22|0),($26|0))|0);
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  $7 = $30;
  $8 = $31;
  __ZNSt3__26localeD2Ev($6);
  $32 = $7;
  $33 = $8;
  ___resumeException($32|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE3putEw($10,$27)|0);
  $34 = $9;
  (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE5flushEv($34)|0);
  $35 = $9;
  STACKTOP = sp;return ($35|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8allocateI16PhotonListenerEMPFviEPFviPKciPKNS0_10DictionaryIhNS0_6ObjectEEEEPFviSC_EPFviS7_EEEPT_RT0_RT1_RT2_RT3_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $9 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(20)|0);
 $8 = $9;
 $10 = $8;
 $11 = $4;
 $12 = HEAP32[$11>>2]|0;
 $13 = $5;
 $14 = HEAP32[$13>>2]|0;
 $15 = $6;
 $16 = HEAP32[$15>>2]|0;
 $17 = $7;
 $18 = HEAP32[$17>>2]|0;
 __ZN16PhotonListenerEMC1EPFviEPFviPKciPKN9ExitGames6Common10DictionaryIhNS5_6ObjectEEEEPFviSA_EPFviS3_E($10,$12,$14,$16,$18);
 $19 = $8;
 STACKTOP = sp;return ($19|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS_6Photon10PhotonPeerE16PhotonListenerEMKhEEPT_RT0_RT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(32)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = $3;
 $9 = HEAP8[$8>>0]|0;
 __ZN9ExitGames6Photon10PhotonPeerC2ERNS0_14PhotonListenerEh($6,$7,$9);
 $10 = $4;
 STACKTOP = sp;return ($10|0);
}
function __ZNSt3__23mapIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMNS_4lessIS4_EENS_9allocatorINS_4pairIKS4_S6_EEEEEixERSB_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $3 = sp + 117|0;
 $14 = sp + 24|0;
 $15 = sp + 16|0;
 $16 = sp + 8|0;
 $17 = sp;
 $20 = sp + 116|0;
 $21 = sp + 115|0;
 $22 = sp + 114|0;
 $23 = sp + 113|0;
 $25 = sp + 52|0;
 $29 = sp + 32|0;
 $30 = sp + 28|0;
 $31 = sp + 112|0;
 $27 = $0;
 $28 = $1;
 $32 = $27;
 $33 = $28;
 $34 = $28;
 $26 = $34;
 $35 = $26;
 $24 = $35;
 $36 = $24;
 $18 = $25;
 $19 = $36;
 $37 = $18;
 $38 = $19;
 ;HEAP8[$14>>0]=HEAP8[$23>>0]|0;
 ;HEAP8[$15>>0]=HEAP8[$22>>0]|0;
 ;HEAP8[$16>>0]=HEAP8[$21>>0]|0;
 ;HEAP8[$17>>0]=HEAP8[$20>>0]|0;
 $12 = $37;
 $13 = $38;
 $39 = $12;
 $40 = $13;
 $11 = $40;
 $41 = $11;
 $9 = $39;
 $10 = $41;
 $42 = $9;
 $43 = $10;
 $8 = $43;
 $44 = $8;
 HEAP32[$42>>2] = $44;
 $45 = HEAP32[$25>>2]|0;
 HEAP32[$30>>2] = $45;
 $2 = $3;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE25__emplace_unique_key_argsIS5_JRKNS_21piecewise_construct_tENS_5tupleIJRKS5_EEENSK_IJEEEEEENS_4pairINS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEEbEERKT_DpOT0_($29,$32,$33,167372,$30,$31);
 $7 = $29;
 $46 = $7;
 $6 = $46;
 $47 = $6;
 $48 = HEAP32[$47>>2]|0;
 $49 = ((($48)) + 16|0);
 $5 = $49;
 $50 = $5;
 $4 = $50;
 $51 = $4;
 $52 = ((($51)) + 4|0);
 STACKTOP = sp;return ($52|0);
}
function _PhotonPeerDestroy($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(160|0);
 $15 = sp + 8|0;
 $23 = sp + 60|0;
 $25 = sp + 52|0;
 $27 = sp + 44|0;
 $29 = sp;
 $30 = sp + 36|0;
 $33 = sp + 24|0;
 $34 = sp + 20|0;
 $35 = sp + 16|0;
 $36 = sp + 12|0;
 HEAP32[$34>>2] = $0;
 (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,143821)|0);
 $31 = 161656;
 $32 = $34;
 $37 = $31;
 $38 = $32;
 $39 = (__ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE4findIS5_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_($37,$38)|0);
 HEAP32[$33>>2] = $39;
 ;HEAP8[$29>>0]=HEAP8[$33>>0]|0;HEAP8[$29+1>>0]=HEAP8[$33+1>>0]|0;HEAP8[$29+2>>0]=HEAP8[$33+2>>0]|0;HEAP8[$29+3>>0]=HEAP8[$33+3>>0]|0;
 $28 = $30;
 $40 = $28;
 ;HEAP32[$40>>2]=HEAP32[$29>>2]|0;
 $41 = HEAP32[$30>>2]|0;
 HEAP32[$35>>2] = $41;
 $26 = 161656;
 $42 = $26;
 $24 = $42;
 $43 = $24;
 $22 = $43;
 $44 = $22;
 $45 = ((($44)) + 4|0);
 $21 = $45;
 $46 = $21;
 $20 = $46;
 $47 = $20;
 $19 = $47;
 $48 = $19;
 $18 = $48;
 $49 = $18;
 $16 = $23;
 $17 = $49;
 $50 = $16;
 $51 = $17;
 HEAP32[$50>>2] = $51;
 $52 = HEAP32[$23>>2]|0;
 HEAP32[$27>>2] = $52;
 ;HEAP8[$15>>0]=HEAP8[$27>>0]|0;HEAP8[$15+1>>0]=HEAP8[$27+1>>0]|0;HEAP8[$15+2>>0]=HEAP8[$27+2>>0]|0;HEAP8[$15+3>>0]=HEAP8[$27+3>>0]|0;
 $14 = $25;
 $53 = $14;
 ;HEAP32[$53>>2]=HEAP32[$15>>2]|0;
 $54 = HEAP32[$25>>2]|0;
 HEAP32[$36>>2] = $54;
 $12 = $35;
 $13 = $36;
 $55 = $12;
 $56 = $13;
 $10 = $55;
 $11 = $56;
 $57 = $10;
 $58 = $11;
 $8 = $57;
 $9 = $58;
 $59 = $8;
 $60 = HEAP32[$59>>2]|0;
 $61 = $9;
 $62 = HEAP32[$61>>2]|0;
 $63 = ($60|0)==($62|0);
 $64 = $63 ^ 1;
 if (!($64)) {
  $76 = HEAP32[$34>>2]|0;
  __ZN9ExitGames6Common16MemoryManagement10deallocateINS_6Photon10PhotonPeerEEEvPKT_($76);
  STACKTOP = sp;return;
 }
 $7 = $35;
 $65 = $7;
 $6 = $65;
 $66 = $6;
 $5 = $66;
 $67 = $5;
 $68 = HEAP32[$67>>2]|0;
 $69 = ((($68)) + 16|0);
 $4 = $69;
 $70 = $4;
 $3 = $70;
 $71 = $3;
 $2 = $71;
 $72 = $2;
 $1 = $72;
 $73 = $1;
 $74 = ((($73)) + 4|0);
 $75 = HEAP32[$74>>2]|0;
 __ZN9ExitGames6Common16MemoryManagement10deallocateI16PhotonListenerEMEEvPKT_($75);
 $76 = HEAP32[$34>>2]|0;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS_6Photon10PhotonPeerEEEvPKT_($76);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement10deallocateI16PhotonListenerEMEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $1;
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$6 & 2047]($4);
 $7 = $1;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($7);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement10deallocateINS_6Photon10PhotonPeerEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $1;
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$6 & 2047]($4);
 $7 = $1;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($7);
 STACKTOP = sp;return;
}
function _PhotonPeerConnect($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $8 = sp + 24|0;
 $9 = sp + 8|0;
 $5 = $0;
 $6 = $1;
 $7 = $2;
 $12 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,143844)|0);
 $13 = $5;
 $14 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEPKv($12,$13)|0);
 $15 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($14,159115)|0);
 $16 = $6;
 $17 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($15,$16)|0);
 $18 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($17,159115)|0);
 $19 = $7;
 $20 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc($18,$19)|0);
 $3 = $20;
 $4 = 728;
 $21 = $3;
 $22 = $4;
 (FUNCTION_TABLE_ii[$22 & 2047]($21)|0);
 $23 = $5;
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($24)) + 8|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = $6;
 __ZN9ExitGames6Common7JStringC2EPKc($8,$27);
 $28 = $7;
 __THREW__ = 0;
 invoke_vii(739,($9|0),($28|0));
 $29 = __THREW__; __THREW__ = 0;
 $30 = $29&1;
 if ($30) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  $10 = $34;
  $11 = $35;
  __ZN9ExitGames6Common7JStringD2Ev($8);
  $38 = $10;
  $39 = $11;
  ___resumeException($38|0);
  // unreachable;
 }
 __THREW__ = 0;
 $31 = (invoke_iiii($26|0,($23|0),($8|0),($9|0))|0);
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if (!($33)) {
  __ZN9ExitGames6Common7JStringD2Ev($9);
  __ZN9ExitGames6Common7JStringD2Ev($8);
  STACKTOP = sp;return ($31|0);
 }
 $36 = ___cxa_find_matching_catch_2()|0;
 $37 = tempRet0;
 $10 = $36;
 $11 = $37;
 __ZN9ExitGames6Common7JStringD2Ev($9);
 __ZN9ExitGames6Common7JStringD2Ev($8);
 $38 = $10;
 $39 = $11;
 ___resumeException($38|0);
 // unreachable;
 return (0)|0;
}
function _PhotonPeerDisconnect($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,143868)|0);
 $5 = $3;
 $6 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEPKv($4,$5)|0);
 $1 = $6;
 $2 = 728;
 $7 = $1;
 $8 = $2;
 (FUNCTION_TABLE_ii[$8 & 2047]($7)|0);
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 12|0);
 $12 = HEAP32[$11>>2]|0;
 FUNCTION_TABLE_vi[$12 & 2047]($9);
 STACKTOP = sp;return;
}
function _PhotonPeerState($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Photon10PhotonPeer12getPeerStateEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _PhotonPeerService($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 FUNCTION_TABLE_vii[$5 & 2047]($2,1);
 STACKTOP = sp;return;
}
function _PhotonPeerSetDebugOutputLevel($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZN9ExitGames6Photon10PhotonPeer19setDebugOutputLevelEi($4,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _PhotonPeerOpCustom($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $5 = $0;
 $6 = $1;
 $10 = $2&1;
 $7 = $10;
 $8 = $3;
 $11 = $4&1;
 $9 = $11;
 $12 = $5;
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 24|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $6;
 $17 = $7;
 $18 = $17&1;
 $19 = $8;
 $20 = $9;
 $21 = $20&1;
 $22 = (FUNCTION_TABLE_iiiiii[$15 & 2047]($12,$16,$18,$19,$21)|0);
 STACKTOP = sp;return ($22|0);
}
function _PhotonOperationRequestCreate($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAP8[$1>>0] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS_6Photon16OperationRequestEhEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS_6Photon16OperationRequestEhEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $3 = sp + 8|0;
 $1 = $0;
 $6 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(56)|0);
 $2 = $6;
 $7 = $2;
 $8 = $1;
 $9 = HEAP8[$8>>0]|0;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEC2Ev($3);
 __THREW__ = 0;
 invoke_viii(740,($7|0),($9|0),($3|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $4 = $13;
  $5 = $14;
  __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED2Ev($3);
  $15 = $4;
  $16 = $5;
  ___resumeException($15|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED2Ev($3);
  $12 = $2;
  STACKTOP = sp;return ($12|0);
 }
 return (0)|0;
}
function _PhotonOperationRequestDestroy($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS_6Photon16OperationRequestEEEvPKT_($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement10deallocateINS_6Photon16OperationRequestEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $1;
 __ZN9ExitGames6Photon16OperationRequestD2Ev($4);
 $5 = $1;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($5);
 STACKTOP = sp;return;
}
function _PhotonOperationRequestSetParam($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp + 8|0;
 $3 = $0;
 HEAP8[$4>>0] = $1;
 $5 = $2;
 $6 = $3;
 $7 = (__ZN9ExitGames6Photon16OperationRequest13getParametersEv($6)|0);
 $8 = $5;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE3putERKhRKS2_($7,$4,$8);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE3putERKhRKS2_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 __ZN9ExitGames6Common14DictionaryBase3putIhNS0_6ObjectEEEvRKT_RKT0_($6,$7,$8);
 STACKTOP = sp;return;
}
function _PhotonTypesByte($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAP8[$1>>0] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIhEEhEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIhEEhEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIhEC2ERKh($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesShort($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAP32[$1>>2] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIsEEiEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIsEEiEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp + 8|0;
 $1 = $0;
 $4 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $4;
 $5 = $2;
 $6 = $1;
 $7 = HEAP32[$6>>2]|0;
 $8 = $7&65535;
 HEAP16[$3>>1] = $8;
 __ZN9ExitGames6Common11ValueObjectIsEC2ERKs($5,$3);
 $9 = $2;
 STACKTOP = sp;return ($9|0);
}
function _PhotonTypesInt($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAP32[$1>>2] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIiEEiEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIiEEiEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIiEC2ERKi($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesLong($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $4 = sp + 16|0;
 $5 = sp + 8|0;
 $6 = sp;
 $7 = $4;
 $8 = $7;
 HEAP32[$8>>2] = $0;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = $1;
 $11 = $5;
 $12 = $11;
 HEAP32[$12>>2] = $2;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = $3;
 $15 = $4;
 $16 = $15;
 $17 = HEAP32[$16>>2]|0;
 $18 = (($15) + 4)|0;
 $19 = $18;
 $20 = HEAP32[$19>>2]|0;
 $21 = $5;
 $22 = $21;
 $23 = HEAP32[$22>>2]|0;
 $24 = (($21) + 4)|0;
 $25 = $24;
 $26 = HEAP32[$25>>2]|0;
 $27 = (_i64Add(($17|0),($20|0),0,($23|0))|0);
 $28 = tempRet0;
 $29 = $6;
 $30 = $29;
 HEAP32[$30>>2] = $27;
 $31 = (($29) + 4)|0;
 $32 = $31;
 HEAP32[$32>>2] = $28;
 $33 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIxEExEEPT_RKT0_($6)|0);
 STACKTOP = sp;return ($33|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIxEExEEPT_RKT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIxEC2ERKx($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesFloat($0) {
 $0 = +$0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAPF32[$1>>2] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIfEEfEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIfEEfEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIfEC2ERKf($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesDouble($0) {
 $0 = +$0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAPF64[$1>>3] = $0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIdEEdEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIdEEdEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIdEC2ERKd($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesBool($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = $0&1;
 HEAP8[$1>>0] = $2;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIbEEbEEPT_RT0_($1)|0);
 STACKTOP = sp;return ($3|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIbEEbEEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectIbEC2ERKb($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesString($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 8|0;
 $1 = $0;
 $5 = $1;
 __ZN9ExitGames6Common7JStringC2EPKc($2,$5);
 __THREW__ = 0;
 $6 = (invoke_ii(741,($2|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  $3 = $9;
  $4 = $10;
  __ZN9ExitGames6Common7JStringD2Ev($2);
  $11 = $3;
  $12 = $4;
  ___resumeException($11|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($2);
  STACKTOP = sp;return ($6|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_7JStringEEES4_EEPT_RKT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_7JStringEEC2ERKS2_($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesTable() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $0 = sp + 8|0;
 __ZN9ExitGames6Common9HashtableC2Ev($0);
 __THREW__ = 0;
 $3 = (invoke_ii(742,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  $1 = $6;
  $2 = $7;
  __ZN9ExitGames6Common9HashtableD2Ev($0);
  $8 = $1;
  $9 = $2;
  ___resumeException($8|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common9HashtableD2Ev($0);
  STACKTOP = sp;return ($3|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_9HashtableEEES4_EEPT_RKT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEC2ERKS2_($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesDictBase($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $4;
 $6 = $5&255;
 $7 = ($6|0)==(122);
 if ($7) {
  $8 = $3;
  $9 = $8&255;
  $10 = ($9|0)==(98);
  if ($10) {
   $11 = (__ZL13dictBaseAllocIhN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEv()|0);
   $2 = $11;
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   $12 = $3;
   $13 = $12&255;
   $14 = ($13|0)==(105);
   if ($14) {
    $15 = (__ZL13dictBaseAllocIiN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEv()|0);
    $2 = $15;
    break;
   }
   $16 = $3;
   $17 = $16&255;
   $18 = ($17|0)==(122);
   if ($18) {
    $19 = (__ZL13dictBaseAllocIN9ExitGames6Common6ObjectES2_EPNS1_14DictionaryBaseEv()|0);
    $2 = $19;
    break;
   } else {
    $2 = 0;
    break;
   }
  }
 } while(0);
 $20 = $2;
 STACKTOP = sp;return ($20|0);
}
function __ZL13dictBaseAllocIhN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryIhNS0_6ObjectEEEEEPT_v()|0);
 return ($0|0);
}
function __ZL13dictBaseAllocIiN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryIiNS0_6ObjectEEEEEPT_v()|0);
 return ($0|0);
}
function __ZL13dictBaseAllocIN9ExitGames6Common6ObjectES2_EPNS1_14DictionaryBaseEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryINS0_6ObjectES4_EEEEPT_v()|0);
 return ($0|0);
}
function _PhotonTypesDictBaseFromDict($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $5;
 $8 = $7&255;
 $9 = ($8|0)==(122);
 if ($9) {
  $10 = $4;
  $11 = $10&255;
  $12 = ($11|0)==(98);
  if ($12) {
   $13 = $6;
   $14 = (__ZL16dictBaseFromDictIhN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($13)|0);
   $3 = $14;
   $25 = $3;
   STACKTOP = sp;return ($25|0);
  }
 }
 $15 = $4;
 $16 = $15&255;
 $17 = ($16|0)==(105);
 if ($17) {
  $18 = $6;
  $19 = (__ZL16dictBaseFromDictIiN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($18)|0);
  $3 = $19;
  $25 = $3;
  STACKTOP = sp;return ($25|0);
 }
 $20 = $4;
 $21 = $20&255;
 $22 = ($21|0)==(122);
 if ($22) {
  $23 = $6;
  $24 = (__ZL16dictBaseFromDictIN9ExitGames6Common6ObjectES2_EPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($23)|0);
  $3 = $24;
  $25 = $3;
  STACKTOP = sp;return ($25|0);
 } else {
  $3 = 0;
  $25 = $3;
  STACKTOP = sp;return ($25|0);
 }
 return (0)|0;
}
function __ZL16dictBaseFromDictIhN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE14getDataAddressEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function __ZL16dictBaseFromDictIiN9ExitGames6Common6ObjectEEPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE14getDataAddressEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function __ZL16dictBaseFromDictIN9ExitGames6Common6ObjectES2_EPNS1_14DictionaryBaseEPNS1_11ValueObjectINS1_10DictionaryIT_T0_EEEE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE14getDataAddressEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _PhotonTypesDict($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $5 = $2;
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (FUNCTION_TABLE_ii[$8 & 2047]($5)|0);
 $10 = HEAP8[$9>>0]|0;
 $3 = $10;
 $11 = $2;
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($12)) + 20|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = (FUNCTION_TABLE_ii[$14 & 2047]($11)|0);
 $16 = HEAP8[$15>>0]|0;
 $4 = $16;
 $17 = $4;
 $18 = $17&255;
 $19 = ($18|0)==(122);
 if ($19) {
  $20 = $3;
  $21 = $20&255;
  $22 = ($21|0)==(98);
  if ($22) {
   $23 = $2;
   $24 = (__ZL4dictIhN9ExitGames6Common6ObjectEEPS2_PNS1_14DictionaryBaseE($23)|0);
   $1 = $24;
   $35 = $1;
   STACKTOP = sp;return ($35|0);
  }
 }
 $25 = $3;
 $26 = $25&255;
 $27 = ($26|0)==(105);
 if ($27) {
  $28 = $2;
  $29 = (__ZL4dictIiN9ExitGames6Common6ObjectEEPS2_PNS1_14DictionaryBaseE($28)|0);
  $1 = $29;
  $35 = $1;
  STACKTOP = sp;return ($35|0);
 }
 $30 = $3;
 $31 = $30&255;
 $32 = ($31|0)==(122);
 if ($32) {
  $33 = $2;
  $34 = (__ZL4dictIN9ExitGames6Common6ObjectES2_EPS2_PNS1_14DictionaryBaseE($33)|0);
  $1 = $34;
  $35 = $1;
  STACKTOP = sp;return ($35|0);
 } else {
  $1 = 0;
  $35 = $1;
  STACKTOP = sp;return ($35|0);
 }
 return (0)|0;
}
function __ZL4dictIhN9ExitGames6Common6ObjectEEPS2_PNS1_14DictionaryBaseE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = $1;
 $4 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEES6_EEPT_RT0_($3)|0);
 $2 = $4;
 $5 = $1;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_14DictionaryBaseEEEvPKT_($5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function __ZL4dictIiN9ExitGames6Common6ObjectEEPS2_PNS1_14DictionaryBaseE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = $1;
 $4 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEES6_EEPT_RT0_($3)|0);
 $2 = $4;
 $5 = $1;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_14DictionaryBaseEEEvPKT_($5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function __ZL4dictIN9ExitGames6Common6ObjectES2_EPS2_PNS1_14DictionaryBaseE($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = $1;
 $4 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryINS0_6ObjectES5_EEEES6_EEPT_RT0_($3)|0);
 $2 = $4;
 $5 = $1;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_14DictionaryBaseEEEvPKT_($5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesArray($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $5 = sp + 96|0;
 $7 = sp + 88|0;
 $8 = sp + 84|0;
 $9 = sp + 80|0;
 $10 = sp + 76|0;
 $11 = sp + 72|0;
 $12 = sp + 68|0;
 $13 = sp + 64|0;
 $15 = sp + 60|0;
 $17 = sp + 52|0;
 $19 = sp;
 $3 = $0;
 $4 = $1;
 HEAP32[$5>>2] = $2;
 $6 = 0;
 $22 = $3;
 $23 = $22&255;
 do {
  switch ($23|0) {
  case 98:  {
   $24 = $4;
   HEAP32[$7>>2] = $24;
   $25 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPKhEEPhiEEPT_RKT0_RKT1_($7,$5)|0);
   $6 = $25;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 107:  {
   $26 = $4;
   HEAP32[$8>>2] = $26;
   $27 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPsEES4_iEEPT_RKT0_RKT1_($8,$5)|0);
   $6 = $27;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 105:  {
   $28 = $4;
   HEAP32[$9>>2] = $28;
   $29 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPiEES4_iEEPT_RKT0_RKT1_($9,$5)|0);
   $6 = $29;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 108:  {
   $30 = $4;
   HEAP32[$10>>2] = $30;
   $31 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPxEES4_iEEPT_RKT0_RKT1_($10,$5)|0);
   $6 = $31;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 102:  {
   $32 = $4;
   HEAP32[$11>>2] = $32;
   $33 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPfEES4_iEEPT_RKT0_RKT1_($11,$5)|0);
   $6 = $33;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 100:  {
   $34 = $4;
   HEAP32[$12>>2] = $34;
   $35 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPdEES4_iEEPT_RKT0_RKT1_($12,$5)|0);
   $6 = $35;
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 115:  {
   $36 = HEAP32[$5>>2]|0;
   $37 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_7JStringEEEPT_j($36)|0);
   HEAP32[$13>>2] = $37;
   $14 = 0;
   while(1) {
    $38 = $14;
    $39 = $38 << 16 >> 16;
    $40 = HEAP32[$5>>2]|0;
    $41 = ($39|0)<($40|0);
    if (!($41)) {
     break;
    }
    $42 = $4;
    $43 = $14;
    $44 = $43 << 16 >> 16;
    $45 = (($42) + ($44<<2)|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = HEAP32[$13>>2]|0;
    $48 = $14;
    $49 = $48 << 16 >> 16;
    $50 = (($47) + ($49<<4)|0);
    (__ZN9ExitGames6Common7JStringaSEPKc($50,$46)|0);
    $51 = $14;
    $52 = (($51) + 1)<<16>>16;
    $14 = $52;
   }
   $53 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPNS0_7JStringEEES5_iEEPT_RT0_RT1_($13,$5)|0);
   $6 = $53;
   $54 = HEAP32[$13>>2]|0;
   __ZN9ExitGames6Common16MemoryManagement15deallocateArrayINS0_7JStringEEEvPKT_($54);
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 111:  {
   $55 = HEAP32[$5>>2]|0;
   $56 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayIbEEPT_j($55)|0);
   HEAP32[$15>>2] = $56;
   $16 = 0;
   while(1) {
    $57 = $16;
    $58 = HEAP32[$5>>2]|0;
    $59 = ($57|0)<($58|0);
    if (!($59)) {
     break;
    }
    $60 = $4;
    $61 = $16;
    $62 = (($60) + ($61)|0);
    $63 = HEAP8[$62>>0]|0;
    $64 = $63&1;
    $65 = HEAP32[$15>>2]|0;
    $66 = $16;
    $67 = (($65) + ($66)|0);
    $68 = $64&1;
    HEAP8[$67>>0] = $68;
    $69 = $16;
    $70 = (($69) + 1)|0;
    $16 = $70;
   }
   $71 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPbEES4_iEEPT_RT0_RT1_($15,$5)|0);
   $6 = $71;
   $72 = HEAP32[$15>>2]|0;
   __ZN9ExitGames6Common16MemoryManagement15deallocateArrayIbEEvPKT_($72);
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  case 122:  {
   $73 = HEAP32[$5>>2]|0;
   $74 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_6ObjectEEEPT_j($73)|0);
   HEAP32[$17>>2] = $74;
   $18 = 0;
   while(1) {
    $75 = $18;
    $76 = HEAP32[$5>>2]|0;
    $77 = ($75|0)<($76|0);
    if (!($77)) {
     break;
    }
    $78 = $4;
    $79 = $18;
    $80 = (($78) + ($79<<2)|0);
    $81 = HEAP32[$80>>2]|0;
    __ZN9ExitGames6Common6ObjectC2ERKS1_($19,$81);
    $82 = HEAP32[$17>>2]|0;
    $83 = $18;
    $84 = (($82) + (($83*40)|0)|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = ((($85)) + 16|0);
    $87 = HEAP32[$86>>2]|0;
    __THREW__ = 0;
    (invoke_iii($87|0,($84|0),($19|0))|0);
    $88 = __THREW__; __THREW__ = 0;
    $89 = $88&1;
    if ($89) {
     label = 20;
     break;
    }
    __ZN9ExitGames6Common6ObjectD2Ev($19);
    $90 = $18;
    $91 = (($90) + 1)|0;
    $18 = $91;
   }
   if ((label|0) == 20) {
    $92 = ___cxa_find_matching_catch_2()|0;
    $93 = tempRet0;
    $20 = $92;
    $21 = $93;
    __ZN9ExitGames6Common6ObjectD2Ev($19);
    $94 = $20;
    $95 = $21;
    ___resumeException($94|0);
    // unreachable;
   }
   $96 = (__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPNS0_6ObjectEEES5_iEEPT_RT0_RT1_($17,$5)|0);
   $6 = $96;
   $97 = HEAP32[$17>>2]|0;
   __ZN9ExitGames6Common16MemoryManagement15deallocateArrayINS0_6ObjectEEEvPKT_($97);
   $98 = $6;
   STACKTOP = sp;return ($98|0);
   break;
  }
  default: {
   $98 = $6;
   STACKTOP = sp;return ($98|0);
  }
  }
 } while(0);
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPKhEEPhiEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 __ZN9ExitGames6Common11ValueObjectIPKhEC2ES3_i($6,$8,$10);
 $11 = $4;
 STACKTOP = sp;return ($11|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPsEES4_iEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPsEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPiEES4_iEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPiEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPxEES4_iEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPxEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPfEES4_iEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPfEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPdEES4_iEEPT_RKT0_RKT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPdEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_7JStringEEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = $4<<4;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + ($35<<4)|0);
  __ZN9ExitGames6Common7JStringC2Ej($39,0);
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPNS0_7JStringEEES5_iEEPT_RT0_RT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEC2ES3_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement15deallocateArrayINS0_7JStringEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4|0)!=(0|0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = $1;
 $7 = ((($6)) + -8|0);
 $2 = $7;
 $8 = $2;
 $9 = $8;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = $3;
 $16 = $15;
 HEAP32[$16>>2] = $11;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = $14;
 while(1) {
  $19 = $3;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (($19) + 4)|0;
  $23 = $22;
  $24 = HEAP32[$23>>2]|0;
  $25 = (_i64Add(($21|0),($24|0),-1,-1)|0);
  $26 = tempRet0;
  $27 = $3;
  $28 = $27;
  HEAP32[$28>>2] = $25;
  $29 = (($27) + 4)|0;
  $30 = $29;
  HEAP32[$30>>2] = $26;
  $31 = ($24>>>0)>(0);
  $32 = ($21>>>0)>(0);
  $33 = ($24|0)==(0);
  $34 = $33 & $32;
  $35 = $31 | $34;
  if (!($35)) {
   break;
  }
  $36 = $1;
  $37 = $3;
  $38 = $37;
  $39 = HEAP32[$38>>2]|0;
  $40 = (($37) + 4)|0;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($36) + ($39<<4)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = HEAP32[$44>>2]|0;
  FUNCTION_TABLE_vi[$45 & 2047]($43);
 }
 $46 = $2;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($46);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayIbEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = $4;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + ($35)|0);
  HEAP8[$39>>0] = 0;
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPbEES4_iEEPT_RT0_RT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPbEC2ES2_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement15deallocateArrayIbEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4|0)!=(0|0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = $1;
 $7 = ((($6)) + -8|0);
 $2 = $7;
 $8 = $2;
 $9 = $8;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = $3;
 $16 = $15;
 HEAP32[$16>>2] = $11;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = $14;
 while(1) {
  $19 = $3;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (($19) + 4)|0;
  $23 = $22;
  $24 = HEAP32[$23>>2]|0;
  $25 = (_i64Add(($21|0),($24|0),-1,-1)|0);
  $26 = tempRet0;
  $27 = $3;
  $28 = $27;
  HEAP32[$28>>2] = $25;
  $29 = (($27) + 4)|0;
  $30 = $29;
  HEAP32[$30>>2] = $26;
  $31 = ($24>>>0)>(0);
  $32 = ($21>>>0)>(0);
  $33 = ($24|0)==(0);
  $34 = $33 & $32;
  $35 = $31 | $34;
  if (!($35)) {
   break;
  }
 }
 $36 = $2;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($36);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_6ObjectEEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4*40)|0;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + (($35*40)|0)|0);
  __ZN9ExitGames6Common6ObjectC2Ev($39);
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectIPNS0_6ObjectEEES5_iEEPT_RT0_RT1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $4 = $5;
 $6 = $4;
 $7 = $2;
 $8 = HEAP32[$7>>2]|0;
 $9 = $3;
 $10 = HEAP32[$9>>2]|0;
 $11 = $10&65535;
 __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEC2ES3_s($6,$8,$11);
 $12 = $4;
 STACKTOP = sp;return ($12|0);
}
function __ZN9ExitGames6Common16MemoryManagement15deallocateArrayINS0_6ObjectEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4|0)!=(0|0);
 if (!($5)) {
  STACKTOP = sp;return;
 }
 $6 = $1;
 $7 = ((($6)) + -8|0);
 $2 = $7;
 $8 = $2;
 $9 = $8;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = $3;
 $16 = $15;
 HEAP32[$16>>2] = $11;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = $14;
 while(1) {
  $19 = $3;
  $20 = $19;
  $21 = HEAP32[$20>>2]|0;
  $22 = (($19) + 4)|0;
  $23 = $22;
  $24 = HEAP32[$23>>2]|0;
  $25 = (_i64Add(($21|0),($24|0),-1,-1)|0);
  $26 = tempRet0;
  $27 = $3;
  $28 = $27;
  HEAP32[$28>>2] = $25;
  $29 = (($27) + 4)|0;
  $30 = $29;
  HEAP32[$30>>2] = $26;
  $31 = ($24>>>0)>(0);
  $32 = ($21>>>0)>(0);
  $33 = ($24|0)==(0);
  $34 = $33 & $32;
  $35 = $31 | $34;
  if (!($35)) {
   break;
  }
  $36 = $1;
  $37 = $3;
  $38 = $37;
  $39 = HEAP32[$38>>2]|0;
  $40 = (($37) + 4)|0;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($36) + (($39*40)|0)|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = HEAP32[$44>>2]|0;
  FUNCTION_TABLE_vi[$45 & 2047]($43);
 }
 $46 = $2;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($46);
 STACKTOP = sp;return;
}
function _PhotonTypesDestroy($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_6ObjectEEEvPKT_($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_6ObjectEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $1;
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$6 & 2047]($4);
 $7 = $1;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($7);
 STACKTOP = sp;return;
}
function _PhotonTypesGetType($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common6Object7getTypeEv($2)|0);
 $4 = $3&255;
 STACKTOP = sp;return ($4|0);
}
function _PhotonTypesGetByte($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIhE14getDataAddressEv($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&255;
 STACKTOP = sp;return ($5|0);
}
function __ZNK9ExitGames6Common11ValueObjectIhE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(98);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIhE28getDataAddressImplementationIhEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetShort($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIsE14getDataAddressEv($2)|0);
 $4 = HEAP16[$3>>1]|0;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIsE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(107);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIsE28getDataAddressImplementationIsEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetInt($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIiE14getDataAddressEv($2)|0);
 $4 = HEAP32[$3>>2]|0;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIiE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(105);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIiE28getDataAddressImplementationIiEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetLongLow($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIxE14getDataAddressEv($2)|0);
 $4 = $3;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) + 4)|0;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 tempRet0 = (0);
 STACKTOP = sp;return ($6|0);
}
function __ZNK9ExitGames6Common11ValueObjectIxE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(108);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIxE28getDataAddressImplementationIxEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetLongHigh($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIxE14getDataAddressEv($2)|0);
 $4 = $3;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) + 4)|0;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 $10 = (_bitshift64Ashr(($6|0),($9|0),32)|0);
 $11 = tempRet0;
 tempRet0 = ($11);
 STACKTOP = sp;return ($10|0);
}
function _PhotonTypesGetFloat($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIfE14getDataAddressEv($2)|0);
 $4 = +HEAPF32[$3>>2];
 STACKTOP = sp;return (+$4);
}
function __ZNK9ExitGames6Common11ValueObjectIfE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(102);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIfE28getDataAddressImplementationIfEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetDouble($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIdE14getDataAddressEv($2)|0);
 $4 = +HEAPF64[$3>>3];
 STACKTOP = sp;return (+$4);
}
function __ZNK9ExitGames6Common11ValueObjectIdE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(100);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIdE28getDataAddressImplementationIdEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetString($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 8|0;
 $1 = $0;
 $5 = $1;
 $6 = (__ZNK9ExitGames6Common11ValueObjectINS0_7JStringEE14getDataAddressEv($5)|0);
 __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($2,$6);
 __THREW__ = 0;
 $7 = (invoke_ii(228,($2|0))|0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  $3 = $10;
  $4 = $11;
  __ZN9ExitGames6Common10UTF8StringD2Ev($2);
  $12 = $3;
  $13 = $4;
  ___resumeException($12|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10UTF8StringD2Ev($2);
  STACKTOP = sp;return ($7|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common11ValueObjectINS0_7JStringEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(115);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectINS0_7JStringEE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetBool($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectIbE14getDataAddressEv($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&1;
 STACKTOP = sp;return ($5|0);
}
function __ZNK9ExitGames6Common11ValueObjectIbE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(111);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIbE28getDataAddressImplementationIbEclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesGetTable($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common11ValueObjectINS0_9HashtableEE14getDataAddressEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_9HashtableEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(104);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectINS0_9HashtableEE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesByteAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPhE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 STACKTOP = sp;return ($10|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPhE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(98);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPhE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesShortAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPsE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<1)|0);
 $9 = HEAP16[$8>>1]|0;
 STACKTOP = sp;return ($9|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPsE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(107);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPsE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesIntAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPiE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 STACKTOP = sp;return ($9|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPiE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(105);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPiE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesLongLowAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPxE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<3)|0);
 $9 = $8;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 STACKTOP = sp;return ($11|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPxE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(108);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPxE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesLongHighAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPxE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<3)|0);
 $9 = $8;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = (_bitshift64Ashr(($11|0),($14|0),32)|0);
 $16 = tempRet0;
 STACKTOP = sp;return ($15|0);
}
function _PhotonTypesFloatAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPfE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<2)|0);
 $9 = +HEAPF32[$8>>2];
 STACKTOP = sp;return (+$9);
}
function __ZNK9ExitGames6Common11ValueObjectIPfE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(102);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPfE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesDoubleAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPdE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7<<3)|0);
 $9 = +HEAPF64[$8>>3];
 STACKTOP = sp;return (+$9);
}
function __ZNK9ExitGames6Common11ValueObjectIPdE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(100);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPdE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesStringAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $5 = sp + 8|0;
 $2 = $0;
 $3 = $1;
 $8 = $2;
 $9 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_7JStringEE14getDataAddressEv($8)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $3;
 $12 = (($10) + ($11<<4)|0);
 $4 = $12;
 $13 = $4;
 __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($5,$13);
 __THREW__ = 0;
 $14 = (invoke_ii(228,($5|0))|0);
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $6 = $17;
  $7 = $18;
  __ZN9ExitGames6Common10UTF8StringD2Ev($5);
  $19 = $6;
  $20 = $7;
  ___resumeException($19|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10UTF8StringD2Ev($5);
  STACKTOP = sp;return ($14|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_7JStringEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(115);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_7JStringEE28getDataAddressImplementationIS3_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesBoolAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPbE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&1;
 STACKTOP = sp;return ($10|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPbE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(111);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPbE28getDataAddressImplementationIS2_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesTableAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_9HashtableEE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + (($7*44)|0)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_9HashtableEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(104);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_9HashtableEE28getDataAddressImplementationIS3_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesObjectAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_6ObjectEE14getDataAddressEv($4)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (($6) + (($7*40)|0)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_6ObjectEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(122);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(1);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectIPNS0_6ObjectEE28getDataAddressImplementationIS3_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function _PhotonTypesTableSize($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _PhotonTypesTablePut($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $7 = $3;
 $8 = (__ZNK9ExitGames6Common11ValueObjectINS0_9HashtableEE14getDataAddressEv($7)|0);
 $6 = $8;
 $9 = $6;
 $10 = $4;
 $11 = $5;
 __ZN9ExitGames6Common9Hashtable3putINS0_6ObjectES3_EEvRKT_RKT0_($9,$10,$11);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9Hashtable3putINS0_6ObjectES3_EEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(272|0);
 $6 = sp + 200|0;
 $7 = sp + 160|0;
 $10 = sp + 120|0;
 $11 = sp + 80|0;
 $12 = sp + 40|0;
 $13 = sp;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $14 = $3;
 __ZN9ExitGames6Common6ObjectC2Ev($6);
 __THREW__ = 0;
 invoke_vi(743,($7|0));
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  $8 = $29;
  $9 = $30;
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  $41 = $8;
  $42 = $9;
  ___resumeException($41|0);
  // unreachable;
 }
 $17 = $4;
 __THREW__ = 0;
 invoke_vii(744,($11|0),($17|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $8 = $31;
  $9 = $32;
 } else {
  __THREW__ = 0;
  invoke_vii(745,($10|0),($11|0));
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  if ($21) {
   $33 = ___cxa_find_matching_catch_2()|0;
   $34 = tempRet0;
   $8 = $33;
   $9 = $34;
  } else {
   $22 = $5;
   __THREW__ = 0;
   invoke_vii(744,($13|0),($22|0));
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    $8 = $35;
    $9 = $36;
   } else {
    __THREW__ = 0;
    invoke_vii(746,($12|0),($13|0));
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    do {
     if ($26) {
      $37 = ___cxa_find_matching_catch_2()|0;
      $38 = tempRet0;
      $8 = $37;
      $9 = $38;
     } else {
      __THREW__ = 0;
      invoke_viii(747,($14|0),($10|0),($12|0));
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       $39 = ___cxa_find_matching_catch_2()|0;
       $40 = tempRet0;
       $8 = $39;
       $9 = $40;
       __ZN9ExitGames6Common6ObjectD2Ev($12);
       break;
      } else {
       __ZN9ExitGames6Common6ObjectD2Ev($12);
       __ZN9ExitGames6Common6ObjectD2Ev($13);
       __ZN9ExitGames6Common6ObjectD2Ev($10);
       __ZN9ExitGames6Common6ObjectD2Ev($11);
       __ZN9ExitGames6Common6ObjectD2Ev($7);
       __ZN9ExitGames6Common6ObjectD2Ev($6);
       STACKTOP = sp;return;
      }
     }
    } while(0);
    __ZN9ExitGames6Common6ObjectD2Ev($13);
   }
   __ZN9ExitGames6Common6ObjectD2Ev($10);
  }
  __ZN9ExitGames6Common6ObjectD2Ev($11);
 }
 __ZN9ExitGames6Common6ObjectD2Ev($7);
 __ZN9ExitGames6Common6ObjectD2Ev($6);
 $41 = $8;
 $42 = $9;
 ___resumeException($41|0);
 // unreachable;
}
function _PhotonTypesTableKeyAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($4)|0);
 $6 = $3;
 $7 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($5,$6)|0);
 STACKTOP = sp;return ($7|0);
}
function __ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEE12getElementAtEj($4,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesTableValueAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZN9ExitGames6Common9HashtableixEj($4,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesDictSize($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _PhotonTypesDictBasePut($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $8 = $3;
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 16|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (FUNCTION_TABLE_ii[$11 & 2047]($8)|0);
 $13 = HEAP8[$12>>0]|0;
 $6 = $13;
 $14 = $3;
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (FUNCTION_TABLE_ii[$17 & 2047]($14)|0);
 $19 = HEAP8[$18>>0]|0;
 $7 = $19;
 $20 = $7;
 $21 = $20&255;
 $22 = ($21|0)==(122);
 if ($22) {
  $23 = $6;
  $24 = $23&255;
  $25 = ($24|0)==(98);
  if ($25) {
   $26 = $3;
   $27 = $4;
   $28 = (__ZNK9ExitGames6Common11ValueObjectIhE14getDataAddressEv($27)|0);
   $29 = $5;
   __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE3putERKhRKS2_($26,$28,$29);
  }
 }
 $30 = $6;
 $31 = $30&255;
 $32 = ($31|0)==(105);
 if ($32) {
  $33 = $3;
  $34 = $4;
  $35 = (__ZNK9ExitGames6Common11ValueObjectIiE14getDataAddressEv($34)|0);
  $36 = $5;
  __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE3putERKiRKS2_($33,$35,$36);
 }
 $37 = $6;
 $38 = $37&255;
 $39 = ($38|0)==(122);
 if (!($39)) {
  STACKTOP = sp;return;
 }
 $40 = $3;
 $41 = $4;
 $42 = $5;
 __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E3putERKS2_S5_($40,$41,$42);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE3putERKiRKS2_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 __ZN9ExitGames6Common14DictionaryBase3putIiNS0_6ObjectEEEvRKT_RKT0_($6,$7,$8);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E3putERKS2_S5_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 __ZN9ExitGames6Common14DictionaryBase3putINS0_6ObjectES3_EEvRKT_RKT0_($6,$7,$8);
 STACKTOP = sp;return;
}
function _PhotonTypesDictKeyAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = $2;
 $6 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($5)|0);
 $4 = $6;
 $7 = $4;
 $8 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($7)|0);
 $9 = $3;
 $10 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($8,$9)|0);
 STACKTOP = sp;return ($10|0);
}
function _PhotonTypesDictValueAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = $2;
 $6 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($5)|0);
 $4 = $6;
 $7 = $4;
 $8 = $3;
 $9 = (__ZNK9ExitGames6Common9HashtableixEj($7,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function _PhotonTypesByteDictSize($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($2)|0);
 STACKTOP = sp;return ($3|0);
}
function _PhotonTypesByteDictPut($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp + 8|0;
 $3 = $0;
 HEAP8[$4>>0] = $1;
 $5 = $2;
 $6 = $3;
 $7 = $5;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE3putERKhRKS2_($6,$4,$7);
 STACKTOP = sp;return;
}
function _PhotonTypesByteDictKeyAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $4 = sp + 8|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE7getKeysEv($4,$7);
 $8 = $3;
 __THREW__ = 0;
 $9 = (invoke_iii(748,($4|0),($8|0))|0);
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  $5 = $14;
  $6 = $15;
  __ZN9ExitGames6Common7JVectorIhED2Ev($4);
  $16 = $5;
  $17 = $6;
  ___resumeException($16|0);
  // unreachable;
 } else {
  $12 = HEAP8[$9>>0]|0;
  $13 = $12&255;
  __ZN9ExitGames6Common7JVectorIhED2Ev($4);
  STACKTOP = sp;return ($13|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE7getKeysEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $1;
 $3 = $2;
 __ZNK9ExitGames6Common14DictionaryBase7getKeysIhEENS0_7JVectorIT_EEPKS4_($0,$3,0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7JVectorIhEixEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZNK9ExitGames6Common7JVectorIhE11verifyIndexEj($4,$5);
 $6 = ((($4)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (($7) + ($8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common7JVectorIhED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $4 = $1;
 HEAP32[$4>>2] = (3820);
 __THREW__ = 0;
 invoke_vi(749,($4|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = ((($4)) + 16|0);
  $8 = HEAP32[$7>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($8|0));
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   __ZN9ExitGames6Common4BaseD2Ev($4);
   STACKTOP = sp;return;
  }
 }
 $11 = ___cxa_find_matching_catch_3(0|0)|0;
 $12 = tempRet0;
 $2 = $11;
 $3 = $12;
 __ZN9ExitGames6Common4BaseD2Ev($4);
 $13 = $2;
 ___clang_call_terminate($13);
 // unreachable;
}
function _PhotonTypesByteDictValueAt($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEixEj($4,$5)|0);
 STACKTOP = sp;return ($6|0);
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEixEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZN9ExitGames6Common14DictionaryBase12getElementAtEjPKNS0_6ObjectE($4,$5,0)|0);
 STACKTOP = sp;return ($6|0);
}
function _PhotonTypesIsArray($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($2)|0);
 $4 = ($3|0)!=(0);
 STACKTOP = sp;return ($4|0);
}
function _PhotonTypesArraySize($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (__ZNK9ExitGames6Common6Object8getSizesEv($2)|0);
 $4 = HEAP16[$3>>1]|0;
 $5 = $4 << 16 >> 16;
 STACKTOP = sp;return ($5|0);
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $7 = $0;
 $8 = $7;
 $6 = $8;
 $9 = $6;
 $5 = $9;
 $10 = $5;
 $11 = ((($10)) + 4|0);
 $4 = $11;
 $12 = $4;
 $3 = $12;
 $13 = $3;
 $2 = $13;
 $14 = $2;
 $1 = $14;
 $15 = $1;
 $16 = HEAP32[$15>>2]|0;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($8,$16);
 STACKTOP = sp;return;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $4 = sp;
 $7 = sp + 80|0;
 $20 = $0;
 $21 = $1;
 $23 = $20;
 $24 = $21;
 $25 = ($24|0)!=(0|0);
 if (!($25)) {
  STACKTOP = sp;return;
 }
 $26 = $21;
 $27 = HEAP32[$26>>2]|0;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($23,$27);
 $28 = $21;
 $29 = ((($28)) + 4|0);
 $30 = HEAP32[$29>>2]|0;
 __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE7destroyEPNS_11__tree_nodeIS8_PvEE($23,$30);
 $12 = $23;
 $31 = $12;
 $32 = ((($31)) + 4|0);
 $11 = $32;
 $33 = $11;
 $10 = $33;
 $34 = $10;
 $22 = $34;
 $35 = $22;
 $36 = $21;
 $37 = ((($36)) + 16|0);
 $9 = $37;
 $38 = $9;
 $8 = $38;
 $39 = $8;
 $5 = $35;
 $6 = $39;
 $40 = $5;
 $41 = $6;
 ;HEAP8[$4>>0]=HEAP8[$7>>0]|0;
 $2 = $40;
 $3 = $41;
 $42 = $22;
 $43 = $21;
 $17 = $42;
 $18 = $43;
 $19 = 1;
 $44 = $17;
 $45 = $18;
 $46 = $19;
 $14 = $44;
 $15 = $45;
 $16 = $46;
 $47 = $15;
 $13 = $47;
 $48 = $13;
 __ZdlPv($48);
 STACKTOP = sp;return;
}
function ___clang_call_terminate($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (___cxa_begin_catch(($0|0))|0);
 __ZSt9terminatev();
 // unreachable;
}
function __ZNSt3__211char_traitsIcE6lengthEPKc($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = (_strlen($2)|0);
 STACKTOP = sp;return ($3|0);
}
function __ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $switch = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(416|0);
 $64 = sp;
 $66 = sp + 408|0;
 $77 = sp + 128|0;
 $94 = sp + 60|0;
 $103 = sp + 16|0;
 $95 = $1;
 $96 = $2;
 $97 = $3;
 $98 = $4;
 $99 = $5;
 $107 = HEAP32[$0>>2]|0;
 $108 = ($107|0)==(0|0);
 if ($108) {
  ;HEAP32[$94>>2]=HEAP32[$0>>2]|0;
  $305 = HEAP32[$94>>2]|0;
  STACKTOP = sp;return ($305|0);
 }
 $109 = $97;
 $110 = $95;
 $111 = $109;
 $112 = $110;
 $113 = (($111) - ($112))|0;
 $114 = (($113|0) / 4)&-1;
 $100 = $114;
 $115 = $98;
 $93 = $115;
 $116 = $93;
 $117 = ((($116)) + 12|0);
 $118 = HEAP32[$117>>2]|0;
 $101 = $118;
 $119 = $101;
 $120 = $100;
 $121 = ($119|0)>($120|0);
 if ($121) {
  $122 = $100;
  $123 = $101;
  $124 = (($123) - ($122))|0;
  $101 = $124;
 } else {
  $101 = 0;
 }
 $125 = $96;
 $126 = $95;
 $127 = $125;
 $128 = $126;
 $129 = (($127) - ($128))|0;
 $130 = (($129|0) / 4)&-1;
 $102 = $130;
 $131 = $102;
 $132 = ($131|0)>(0);
 if ($132) {
  $133 = HEAP32[$0>>2]|0;
  $134 = $95;
  $135 = $102;
  $84 = $133;
  $85 = $134;
  $86 = $135;
  $136 = $84;
  $137 = HEAP32[$136>>2]|0;
  $138 = ((($137)) + 48|0);
  $139 = HEAP32[$138>>2]|0;
  $140 = $85;
  $141 = $86;
  $142 = (FUNCTION_TABLE_iiii[$139 & 2047]($136,$140,$141)|0);
  $143 = $102;
  $144 = ($142|0)!=($143|0);
  if ($144) {
   HEAP32[$0>>2] = 0;
   ;HEAP32[$94>>2]=HEAP32[$0>>2]|0;
   $305 = HEAP32[$94>>2]|0;
   STACKTOP = sp;return ($305|0);
  }
 }
 $145 = $101;
 $146 = ($145|0)>(0);
 if ($146) {
  $147 = $101;
  $148 = $99;
  $81 = $103;
  $82 = $147;
  $83 = $148;
  $149 = $81;
  $80 = $149;
  $150 = $80;
  $79 = $150;
  $151 = $79;
  $78 = $151;
  ;HEAP32[$151>>2]=0|0;HEAP32[$151+4>>2]=0|0;HEAP32[$151+8>>2]=0|0;
  $152 = $82;
  $153 = $83;
  $72 = $149;
  $73 = $152;
  $74 = $153;
  $154 = $72;
  $155 = $73;
  $70 = $154;
  $156 = $70;
  $69 = $156;
  $157 = $69;
  $68 = $157;
  $158 = $68;
  $67 = $158;
  $159 = $67;
  $65 = $159;
  $160 = $65;
  ;HEAP8[$64>>0]=HEAP8[$66>>0]|0;
  $63 = $160;
  $161 = $63;
  $62 = $161;
  $71 = 1073741823;
  $162 = $71;
  $163 = (($162) - 16)|0;
  $164 = ($155>>>0)>($163>>>0);
  if ($164) {
   __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($154);
   // unreachable;
  }
  $165 = $73;
  $166 = ($165>>>0)<(2);
  $167 = $73;
  do {
   if ($166) {
    $60 = $154;
    $61 = $167;
    $168 = $60;
    $169 = $61;
    $170 = $169&255;
    $59 = $168;
    $171 = $59;
    $58 = $171;
    $172 = $58;
    $173 = ((($172)) + 8|0);
    $174 = ((($173)) + 3|0);
    HEAP8[$174>>0] = $170;
    $57 = $154;
    $175 = $57;
    $56 = $175;
    $176 = $56;
    $55 = $176;
    $177 = $55;
    $54 = $177;
    $178 = $54;
    $53 = $178;
    $179 = $53;
    $75 = $179;
   } else {
    $27 = $167;
    $180 = $27;
    $181 = ($180>>>0)<(2);
    if ($181) {
     $188 = 2;
    } else {
     $182 = $27;
     $183 = (($182) + 1)|0;
     $26 = $183;
     $184 = $26;
     $185 = (($184) + 3)|0;
     $186 = $185 & -4;
     $188 = $186;
    }
    $187 = (($188) - 1)|0;
    $76 = $187;
    $25 = $154;
    $189 = $25;
    $24 = $189;
    $190 = $24;
    $23 = $190;
    $191 = $23;
    $192 = $76;
    $193 = (($192) + 1)|0;
    $38 = $191;
    $39 = $193;
    $194 = $38;
    $195 = $39;
    $35 = $194;
    $36 = $195;
    $37 = 0;
    $196 = $35;
    $197 = $36;
    $34 = $196;
    $198 = ($197>>>0)>(1073741823);
    if (!($198)) {
     $209 = $36;
     $210 = $209<<2;
     $33 = $210;
     $211 = $33;
     $212 = (__Znwj($211)|0);
     $75 = $212;
     $213 = $75;
     $42 = $154;
     $43 = $213;
     $214 = $42;
     $215 = $43;
     $41 = $214;
     $216 = $41;
     $40 = $216;
     $217 = $40;
     HEAP32[$217>>2] = $215;
     $218 = $76;
     $219 = (($218) + 1)|0;
     $46 = $154;
     $47 = $219;
     $220 = $46;
     $221 = $47;
     $222 = -2147483648 | $221;
     $45 = $220;
     $223 = $45;
     $44 = $223;
     $224 = $44;
     $225 = ((($224)) + 8|0);
     HEAP32[$225>>2] = $222;
     $226 = $73;
     $50 = $154;
     $51 = $226;
     $227 = $50;
     $228 = $51;
     $49 = $227;
     $229 = $49;
     $48 = $229;
     $230 = $48;
     $231 = ((($230)) + 4|0);
     HEAP32[$231>>2] = $228;
     break;
    }
    $30 = 158021;
    $199 = (___cxa_allocate_exception(8)|0);
    $200 = $30;
    $28 = $199;
    $29 = $200;
    $201 = $28;
    $202 = $29;
    __THREW__ = 0;
    invoke_vii(751,($201|0),($202|0));
    $203 = __THREW__; __THREW__ = 0;
    $204 = $203&1;
    if ($204) {
     $205 = ___cxa_find_matching_catch_2()|0;
     $206 = tempRet0;
     $31 = $205;
     $32 = $206;
     ___cxa_free_exception(($199|0));
     $207 = $31;
     $208 = $32;
     ___resumeException($207|0);
     // unreachable;
    } else {
     HEAP32[$201>>2] = (22172);
     ___cxa_throw(($199|0),(2528|0),(713|0));
     // unreachable;
    }
   }
  } while(0);
  $232 = $75;
  $52 = $232;
  $233 = $52;
  $234 = $73;
  $235 = $74;
  (__ZNSt3__211char_traitsIwE6assignEPwjw($233,$234,$235)|0);
  $236 = $75;
  $237 = $73;
  $238 = (($236) + ($237<<2)|0);
  HEAP32[$77>>2] = 0;
  __ZNSt3__211char_traitsIwE6assignERwRKw($238,$77);
  $239 = HEAP32[$0>>2]|0;
  $22 = $103;
  $240 = $22;
  $21 = $240;
  $241 = $21;
  $20 = $241;
  $242 = $20;
  $19 = $242;
  $243 = $19;
  $18 = $243;
  $244 = $18;
  $245 = ((($244)) + 8|0);
  $246 = ((($245)) + 3|0);
  $247 = HEAP8[$246>>0]|0;
  $248 = $247&255;
  $249 = $248 & 128;
  $250 = ($249|0)!=(0);
  if ($250) {
   $12 = $241;
   $251 = $12;
   $11 = $251;
   $252 = $11;
   $10 = $252;
   $253 = $10;
   $254 = HEAP32[$253>>2]|0;
   $260 = $254;
  } else {
   $17 = $241;
   $255 = $17;
   $16 = $255;
   $256 = $16;
   $15 = $256;
   $257 = $15;
   $14 = $257;
   $258 = $14;
   $13 = $258;
   $259 = $13;
   $260 = $259;
  }
  $9 = $260;
  $261 = $9;
  $262 = $101;
  $6 = $239;
  $7 = $261;
  $8 = $262;
  $263 = $6;
  $264 = HEAP32[$263>>2]|0;
  $265 = ((($264)) + 48|0);
  $266 = HEAP32[$265>>2]|0;
  $267 = $7;
  $268 = $8;
  __THREW__ = 0;
  $269 = (invoke_iiii($266|0,($263|0),($267|0),($268|0))|0);
  $270 = __THREW__; __THREW__ = 0;
  $271 = $270&1;
  if ($271) {
   $274 = ___cxa_find_matching_catch_2()|0;
   $275 = tempRet0;
   $104 = $274;
   $105 = $275;
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($103);
   $276 = $104;
   $277 = $105;
   ___resumeException($276|0);
   // unreachable;
  }
  $272 = $101;
  $273 = ($269|0)!=($272|0);
  if ($273) {
   HEAP32[$0>>2] = 0;
   ;HEAP32[$94>>2]=HEAP32[$0>>2]|0;
   $106 = 1;
  } else {
   $106 = 0;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($103);
  $278 = $106;
  $switch = ($278>>>0)<(1);
  if (!($switch)) {
   $305 = HEAP32[$94>>2]|0;
   STACKTOP = sp;return ($305|0);
  }
 }
 $279 = $97;
 $280 = $96;
 $281 = $279;
 $282 = $280;
 $283 = (($281) - ($282))|0;
 $284 = (($283|0) / 4)&-1;
 $102 = $284;
 $285 = $102;
 $286 = ($285|0)>(0);
 if ($286) {
  $287 = HEAP32[$0>>2]|0;
  $288 = $96;
  $289 = $102;
  $87 = $287;
  $88 = $288;
  $89 = $289;
  $290 = $87;
  $291 = HEAP32[$290>>2]|0;
  $292 = ((($291)) + 48|0);
  $293 = HEAP32[$292>>2]|0;
  $294 = $88;
  $295 = $89;
  $296 = (FUNCTION_TABLE_iiii[$293 & 2047]($290,$294,$295)|0);
  $297 = $102;
  $298 = ($296|0)!=($297|0);
  if ($298) {
   HEAP32[$0>>2] = 0;
   ;HEAP32[$94>>2]=HEAP32[$0>>2]|0;
   $305 = HEAP32[$94>>2]|0;
   STACKTOP = sp;return ($305|0);
  }
 }
 $299 = $98;
 $90 = $299;
 $91 = 0;
 $300 = $90;
 $301 = ((($300)) + 12|0);
 $302 = HEAP32[$301>>2]|0;
 $92 = $302;
 $303 = $91;
 $304 = ((($300)) + 12|0);
 HEAP32[$304>>2] = $303;
 ;HEAP32[$94>>2]=HEAP32[$0>>2]|0;
 $305 = HEAP32[$94>>2]|0;
 STACKTOP = sp;return ($305|0);
}
function __ZNSt3__211char_traitsIwE6assignEPwjw($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $4;
 $7 = ($6|0)==(0);
 $8 = $3;
 if ($7) {
  $14 = $8;
  STACKTOP = sp;return ($14|0);
 }
 $9 = $5;
 $10 = $4;
 __THREW__ = 0;
 $11 = (invoke_iiii(752,($8|0),($9|0),($10|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $15 = ___cxa_find_matching_catch_3(0|0)|0;
  $16 = tempRet0;
  ___clang_call_terminate($15);
  // unreachable;
 } else {
  $14 = $8;
  STACKTOP = sp;return ($14|0);
 }
 return (0)|0;
}
function __ZNSt3__211char_traitsIwE6assignERwRKw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $6 = $2;
 HEAP32[$6>>2] = $5;
 STACKTOP = sp;return;
}
function __ZNSt3__211char_traitsIwE11eq_int_typeEjj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = ($4|0)==($5|0);
 STACKTOP = sp;return ($6|0);
}
function __ZNSt3__211char_traitsIwE3eofEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEEC2ERKSC_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $12 = sp + 48|0;
 $13 = sp + 8|0;
 $16 = sp + 36|0;
 $17 = sp + 93|0;
 $18 = sp;
 $25 = sp + 92|0;
 $23 = $0;
 $24 = $1;
 $26 = $23;
 $27 = ((($26)) + 4|0);
 $22 = $27;
 $28 = $22;
 $21 = $28;
 $29 = $21;
 $20 = $29;
 $19 = $29;
 $30 = $19;
 HEAP32[$30>>2] = 0;
 $31 = ((($26)) + 8|0);
 ;HEAP8[$18>>0]=HEAP8[$25>>0]|0;
 $15 = $31;
 HEAP32[$16>>2] = 0;
 $32 = $15;
 $14 = $16;
 $33 = $14;
 $34 = HEAP32[$33>>2]|0;
 $8 = $18;
 ;HEAP8[$13>>0]=HEAP8[$17>>0]|0;
 $11 = $32;
 HEAP32[$12>>2] = $34;
 $35 = $11;
 $10 = $13;
 $9 = $12;
 $36 = $9;
 $37 = HEAP32[$36>>2]|0;
 HEAP32[$35>>2] = $37;
 $6 = $26;
 $38 = $6;
 $39 = ((($38)) + 4|0);
 $5 = $39;
 $40 = $5;
 $4 = $40;
 $41 = $4;
 $3 = $41;
 $42 = $3;
 $2 = $42;
 $43 = $2;
 $7 = $26;
 $44 = $7;
 HEAP32[$44>>2] = $43;
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEMC1EPFviEPFviPKciPKN9ExitGames6Common10DictionaryIhNS5_6ObjectEEEEPFviSA_EPFviS3_E($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $5 = $0;
 $6 = $1;
 $7 = $2;
 $8 = $3;
 $9 = $4;
 $10 = $5;
 __ZN9ExitGames6Common12BaseListenerC2Ev($10);
 $11 = 2672;
 __ZN9ExitGames6Photon14PhotonListenerC2Ev($10,$11);
 HEAP32[$10>>2] = (2644);
 HEAP32[$10>>2] = (2644);
 $12 = ((($10)) + 4|0);
 $13 = $6;
 HEAP32[$12>>2] = $13;
 $14 = ((($10)) + 8|0);
 $15 = $7;
 HEAP32[$14>>2] = $15;
 $16 = ((($10)) + 12|0);
 $17 = $8;
 HEAP32[$16>>2] = $17;
 $18 = ((($10)) + 16|0);
 $19 = $9;
 HEAP32[$18>>2] = $19;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common12BaseListenerC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 HEAP32[$2>>2] = (2692);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon14PhotonListenerC2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$4>>2] = $6;
 $7 = ((($5)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = HEAP32[$4>>2]|0;
 $10 = ((($9)) + -20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (($4) + ($11)|0);
 HEAP32[$12>>2] = $8;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon14PhotonListenerD1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Photon14PhotonListenerD2Ev($2,2704);
 __ZN9ExitGames6Common12BaseListenerD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon14PhotonListenerD0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Photon14PhotonListenerD1Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEMD1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN16PhotonListenerEMD2Ev($2,2668);
 __ZN9ExitGames6Common12BaseListenerD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEMD0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN16PhotonListenerEMD1Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEM11debugReturnEiRKN9ExitGames6Common7JStringE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $8 = sp + 8|0;
 $5 = $0;
 $6 = $1;
 $7 = $2;
 $11 = $5;
 $12 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,143984)|0);
 $13 = $6;
 $14 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEi($12,$13)|0);
 $15 = $7;
 $16 = (__ZN9ExitGames6CommonlsIwNSt3__211char_traitsIwEEEERNS2_13basic_ostreamIT_T0_EES9_RKNS0_7JStringE($14,$15)|0);
 $3 = $16;
 $4 = 728;
 $17 = $3;
 $18 = $4;
 (FUNCTION_TABLE_ii[$18 & 2047]($17)|0);
 $19 = ((($11)) + 16|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = $6;
 $22 = $7;
 __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($8,$22);
 __THREW__ = 0;
 $23 = (invoke_ii(228,($8|0))|0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $9 = $28;
  $10 = $29;
  __ZN9ExitGames6Common10UTF8StringD2Ev($8);
  $30 = $9;
  $31 = $10;
  ___resumeException($30|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_vii($20|0,($21|0),($23|0));
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $9 = $28;
  $10 = $29;
  __ZN9ExitGames6Common10UTF8StringD2Ev($8);
  $30 = $9;
  $31 = $10;
  ___resumeException($30|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10UTF8StringD2Ev($8);
  STACKTOP = sp;return;
 }
}
function __ZN16PhotonListenerEM19onOperationResponseERKN9ExitGames6Photon17OperationResponseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $6 = sp + 24|0;
 $9 = sp;
 $4 = $0;
 $5 = $1;
 $10 = $4;
 $11 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,144002)|0);
 $12 = $5;
 __ZNK9ExitGames6Photon17OperationResponse8toStringEbbb($6,$12,1,1,0);
 __THREW__ = 0;
 $13 = (invoke_iii(753,($11|0),($6|0))|0);
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  $2 = $13;
  $3 = 728;
  $16 = $2;
  $17 = $3;
  __THREW__ = 0;
  (invoke_ii($17|0,($16|0))|0);
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   __ZN9ExitGames6Common7JStringD2Ev($6);
   $20 = ((($10)) + 8|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = $5;
   $23 = (__ZNK9ExitGames6Photon17OperationResponse13getReturnCodeEv($22)|0);
   $24 = $23 << 16 >> 16;
   $25 = $5;
   $26 = (__ZNK9ExitGames6Photon17OperationResponse15getDebugMessageEv($25)|0);
   __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($9,$26);
   __THREW__ = 0;
   $27 = (invoke_ii(228,($9|0))|0);
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if (!($29)) {
    $30 = $5;
    __THREW__ = 0;
    $31 = (invoke_ii(754,($30|0))|0);
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if (!($33)) {
     $34 = $31&255;
     $35 = $5;
     __THREW__ = 0;
     $36 = (invoke_ii(755,($35|0))|0);
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if (!($38)) {
      __THREW__ = 0;
      invoke_viiii($21|0,($24|0),($27|0),($34|0),($36|0));
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if (!($40)) {
       __ZN9ExitGames6Common10UTF8StringD2Ev($9);
       STACKTOP = sp;return;
      }
     }
    }
   }
   $43 = ___cxa_find_matching_catch_2()|0;
   $44 = tempRet0;
   $7 = $43;
   $8 = $44;
   __ZN9ExitGames6Common10UTF8StringD2Ev($9);
   $45 = $7;
   $46 = $8;
   ___resumeException($45|0);
   // unreachable;
  }
 }
 $41 = ___cxa_find_matching_catch_2()|0;
 $42 = tempRet0;
 $7 = $41;
 $8 = $42;
 __ZN9ExitGames6Common7JStringD2Ev($6);
 $45 = $7;
 $46 = $8;
 ___resumeException($45|0);
 // unreachable;
}
function __ZN16PhotonListenerEM15onStatusChangedEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $4;
 $7 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKc(164652,144028)|0);
 $8 = $5;
 $9 = (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEi($7,$8)|0);
 $2 = $9;
 $3 = 728;
 $10 = $2;
 $11 = $3;
 (FUNCTION_TABLE_ii[$11 & 2047]($10)|0);
 $12 = ((($6)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $5;
 FUNCTION_TABLE_vi[$13 & 2047]($14);
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEM7onEventERKN9ExitGames6Photon9EventDataE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = ((($4)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3;
 $8 = (__ZNK9ExitGames6Photon9EventData7getCodeEv($7)|0);
 $9 = $8&255;
 $10 = $3;
 $11 = (__ZNK9ExitGames6Photon9EventData13getParametersEv($10)|0);
 FUNCTION_TABLE_vii[$6 & 2047]($9,$11);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common12BaseListenerD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common12BaseListenerD0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common12BaseListenerD2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon14PhotonListenerD2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 STACKTOP = sp;return;
}
function __ZN16PhotonListenerEMD2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = ((($5)) + 4|0);
 __ZN9ExitGames6Photon14PhotonListenerD2Ev($4,$6);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6CommonlsIwNSt3__211char_traitsIwEEEERNS2_13basic_ostreamIT_T0_EES9_RKNS0_7JStringE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common7JString4cstrEv($5)|0);
 $7 = (__ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_($4,$6)|0);
 STACKTOP = sp;return ($7|0);
}
function __ZNSt3__2lsIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = $3;
 $7 = (__ZNSt3__211char_traitsIwE6lengthEPKw($6)|0);
 $8 = (__ZNSt3__224__put_character_sequenceIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_j($4,$5,$7)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNSt3__224__put_character_sequenceIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_PKS4_j($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$byval_copy = 0, $$sink1 = 0, $$sink2 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $$byval_copy = sp + 108|0;
 $8 = sp + 92|0;
 $26 = sp + 16|0;
 $29 = sp + 4|0;
 $30 = sp;
 $23 = $0;
 $24 = $1;
 $25 = $2;
 $31 = $23;
 __THREW__ = 0;
 invoke_vii(729,($26|0),($31|0));
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 L1: do {
  if ($33) {
   $124 = ___cxa_find_matching_catch_3(0|0)|0;
   $125 = tempRet0;
   $27 = $124;
   $28 = $125;
  } else {
   $22 = $26;
   $34 = $22;
   $35 = HEAP8[$34>>0]|0;
   $36 = $35&1;
   do {
    if ($36) {
     $37 = $23;
     $20 = $30;
     $21 = $37;
     $38 = $20;
     $39 = $21;
     $40 = HEAP32[$39>>2]|0;
     $41 = ((($40)) + -12|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = (($39) + ($42)|0);
     $19 = $43;
     $44 = $19;
     $18 = $44;
     $45 = $18;
     $46 = ((($45)) + 24|0);
     $47 = HEAP32[$46>>2]|0;
     HEAP32[$38>>2] = $47;
     $48 = $24;
     $49 = $23;
     $50 = HEAP32[$49>>2]|0;
     $51 = ((($50)) + -12|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = (($49) + ($52)|0);
     $13 = $53;
     $54 = $13;
     $55 = ((($54)) + 4|0);
     $56 = HEAP32[$55>>2]|0;
     $57 = $56 & 176;
     $58 = ($57|0)==(32);
     $59 = $24;
     $60 = $25;
     $61 = (($59) + ($60<<2)|0);
     $62 = $58 ? $61 : $59;
     $63 = $24;
     $64 = $25;
     $65 = (($63) + ($64<<2)|0);
     $66 = $23;
     $67 = HEAP32[$66>>2]|0;
     $68 = ((($67)) + -12|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = (($66) + ($69)|0);
     $71 = $23;
     $72 = HEAP32[$71>>2]|0;
     $73 = ((($72)) + -12|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = (($71) + ($74)|0);
     $11 = $75;
     $76 = $11;
     $77 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $78 = ((($76)) + 76|0);
     $79 = HEAP32[$78>>2]|0;
     $80 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($77,$79)|0);
     do {
      if ($80) {
       $6 = $76;
       $7 = 32;
       $81 = $6;
       __THREW__ = 0;
       invoke_vii(733,($8|0),($81|0));
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        label = 13;
       } else {
        $5 = $8;
        $84 = $5;
        __THREW__ = 0;
        $85 = (invoke_iii(734,($84|0),(165472|0))|0);
        $86 = __THREW__; __THREW__ = 0;
        $87 = $86&1;
        if (!($87)) {
         $88 = $7;
         $3 = $85;
         $4 = $88;
         $89 = $3;
         $90 = HEAP32[$89>>2]|0;
         $91 = ((($90)) + 44|0);
         $92 = HEAP32[$91>>2]|0;
         $93 = $4;
         __THREW__ = 0;
         $94 = (invoke_iii($92|0,($89|0),($93|0))|0);
         $95 = __THREW__; __THREW__ = 0;
         $96 = $95&1;
         if (!($96)) {
          __ZNSt3__26localeD2Ev($8);
          $101 = ((($76)) + 76|0);
          HEAP32[$101>>2] = $94;
          label = 9;
          break;
         }
        }
        $97 = ___cxa_find_matching_catch_3(0|0)|0;
        $98 = tempRet0;
        $9 = $97;
        $10 = $98;
        __ZNSt3__26localeD2Ev($8);
        $99 = $9;
        $100 = $10;
        $$sink1 = $100;$$sink2 = $99;
       }
      } else {
       label = 9;
      }
     } while(0);
     if ((label|0) == 9) {
      $102 = ((($76)) + 76|0);
      $103 = HEAP32[$102>>2]|0;
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$30>>2]|0;
      $104 = (invoke_iiiiiii(735,($$byval_copy|0),($48|0),($62|0),($65|0),($70|0),($103|0))|0);
      $105 = __THREW__; __THREW__ = 0;
      $106 = $105&1;
      if ($106) {
       label = 13;
      } else {
       HEAP32[$29>>2] = $104;
       $12 = $29;
       $107 = $12;
       $108 = HEAP32[$107>>2]|0;
       $109 = ($108|0)==(0|0);
       if (!($109)) {
        break;
       }
       $110 = $23;
       $111 = HEAP32[$110>>2]|0;
       $112 = ((($111)) + -12|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = (($110) + ($113)|0);
       $16 = $114;
       $17 = 5;
       $115 = $16;
       $116 = $17;
       $14 = $115;
       $15 = $116;
       $117 = $14;
       $118 = ((($117)) + 16|0);
       $119 = HEAP32[$118>>2]|0;
       $120 = $15;
       $121 = $119 | $120;
       __THREW__ = 0;
       invoke_vii(736,($117|0),($121|0));
       $122 = __THREW__; __THREW__ = 0;
       $123 = $122&1;
       if ($123) {
        label = 13;
       } else {
        break;
       }
      }
     }
     if ((label|0) == 13) {
      $126 = ___cxa_find_matching_catch_3(0|0)|0;
      $127 = tempRet0;
      $$sink1 = $127;$$sink2 = $126;
     }
     $27 = $$sink2;
     $28 = $$sink1;
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($26);
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($26);
   $136 = $23;
   STACKTOP = sp;return ($136|0);
  }
 } while(0);
 $128 = $27;
 (___cxa_begin_catch(($128|0))|0);
 $129 = $23;
 $130 = HEAP32[$129>>2]|0;
 $131 = ((($130)) + -12|0);
 $132 = HEAP32[$131>>2]|0;
 $133 = (($129) + ($132)|0);
 __THREW__ = 0;
 invoke_vi(737,($133|0));
 $134 = __THREW__; __THREW__ = 0;
 $135 = $134&1;
 if (!($135)) {
  ___cxa_end_catch();
  $136 = $23;
  STACKTOP = sp;return ($136|0);
 }
 $137 = ___cxa_find_matching_catch_2()|0;
 $138 = tempRet0;
 $27 = $137;
 $28 = $138;
 __THREW__ = 0;
 invoke_v(738);
 $139 = __THREW__; __THREW__ = 0;
 $140 = $139&1;
 if ($140) {
  $143 = ___cxa_find_matching_catch_3(0|0)|0;
  $144 = tempRet0;
  ___clang_call_terminate($143);
  // unreachable;
 } else {
  $141 = $27;
  $142 = $28;
  ___resumeException($141|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZNSt3__211char_traitsIwE6lengthEPKw($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __THREW__ = 0;
 $3 = (invoke_ii(756,($2|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  STACKTOP = sp;return ($3|0);
 }
 return (0)|0;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE25__emplace_unique_key_argsIS5_JRKNS_21piecewise_construct_tENS_5tupleIJRKS5_EEENSK_IJEEEEEENS_4pairINS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEEbEERKT_DpOT0_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $29 = sp + 8|0;
 $32 = sp + 346|0;
 $57 = sp;
 $60 = sp + 345|0;
 $84 = sp + 44|0;
 $87 = sp + 344|0;
 $88 = sp + 24|0;
 $91 = sp + 12|0;
 $79 = $1;
 $80 = $2;
 $81 = $3;
 $82 = $4;
 $83 = $5;
 $92 = $79;
 $93 = $80;
 $94 = (__ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE12__find_equalIS5_EERPNS_16__tree_node_baseIPvEERPNS_15__tree_end_nodeISK_EERKT_($92,$84,$93)|0);
 $85 = $94;
 $95 = $85;
 $96 = HEAP32[$95>>2]|0;
 $86 = $96;
 HEAP8[$87>>0] = 0;
 $97 = $85;
 $98 = HEAP32[$97>>2]|0;
 $99 = ($98|0)==(0|0);
 do {
  if ($99) {
   $100 = $81;
   $78 = $100;
   $101 = $78;
   $102 = $82;
   $43 = $102;
   $103 = $43;
   $104 = $83;
   $15 = $104;
   $105 = $15;
   __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE16__construct_nodeIJRKNS_21piecewise_construct_tENS_5tupleIJRKS5_EEENSK_IJEEEEEENS_10unique_ptrINS_11__tree_nodeIS8_PvEENS_22__tree_node_destructorINSD_ISS_EEEEEEDpOT_($88,$92,$101,$103,$105);
   $106 = HEAP32[$84>>2]|0;
   $107 = $85;
   $8 = $88;
   $108 = $8;
   $7 = $108;
   $109 = $7;
   $6 = $109;
   $110 = $6;
   $111 = HEAP32[$110>>2]|0;
   __THREW__ = 0;
   invoke_viiii(757,($92|0),($106|0),($107|0),($111|0));
   $112 = __THREW__; __THREW__ = 0;
   $113 = $112&1;
   if (!($113)) {
    $13 = $88;
    $114 = $13;
    $12 = $114;
    $115 = $12;
    $11 = $115;
    $116 = $11;
    $117 = HEAP32[$116>>2]|0;
    $14 = $117;
    $10 = $114;
    $118 = $10;
    $9 = $118;
    $119 = $9;
    HEAP32[$119>>2] = 0;
    $120 = $14;
    $86 = $120;
    HEAP8[$87>>0] = 1;
    $42 = $88;
    $121 = $42;
    $39 = $121;
    $40 = 0;
    $122 = $39;
    $38 = $122;
    $123 = $38;
    $37 = $123;
    $124 = $37;
    $125 = HEAP32[$124>>2]|0;
    $41 = $125;
    $126 = $40;
    $19 = $122;
    $127 = $19;
    $18 = $127;
    $128 = $18;
    HEAP32[$128>>2] = $126;
    $129 = $41;
    $130 = ($129|0)!=(0|0);
    if (!($130)) {
     break;
    }
    $17 = $122;
    $131 = $17;
    $16 = $131;
    $132 = $16;
    $133 = ((($132)) + 4|0);
    $134 = $41;
    $35 = $133;
    $36 = $134;
    $135 = $35;
    $136 = ((($135)) + 4|0);
    $137 = HEAP8[$136>>0]|0;
    $138 = $137&1;
    if ($138) {
     $139 = HEAP32[$135>>2]|0;
     $140 = $36;
     $141 = ((($140)) + 16|0);
     $34 = $141;
     $142 = $34;
     $33 = $142;
     $143 = $33;
     $30 = $139;
     $31 = $143;
     $144 = $30;
     $145 = $31;
     ;HEAP8[$29>>0]=HEAP8[$32>>0]|0;
     $27 = $144;
     $28 = $145;
    }
    $146 = $36;
    $147 = ($146|0)!=(0|0);
    if (!($147)) {
     break;
    }
    $148 = HEAP32[$135>>2]|0;
    $149 = $36;
    $24 = $148;
    $25 = $149;
    $26 = 1;
    $150 = $24;
    $151 = $25;
    $152 = $26;
    $21 = $150;
    $22 = $151;
    $23 = $152;
    $153 = $22;
    $20 = $153;
    $154 = $20;
    __ZdlPv($154);
    break;
   }
   $155 = ___cxa_find_matching_catch_2()|0;
   $156 = tempRet0;
   $89 = $155;
   $90 = $156;
   $70 = $88;
   $157 = $70;
   $67 = $157;
   $68 = 0;
   $158 = $67;
   $66 = $158;
   $159 = $66;
   $65 = $159;
   $160 = $65;
   $161 = HEAP32[$160>>2]|0;
   $69 = $161;
   $162 = $68;
   $47 = $158;
   $163 = $47;
   $46 = $163;
   $164 = $46;
   HEAP32[$164>>2] = $162;
   $165 = $69;
   $166 = ($165|0)!=(0|0);
   if (!($166)) {
    $203 = $89;
    $204 = $90;
    ___resumeException($203|0);
    // unreachable;
   }
   $45 = $158;
   $167 = $45;
   $44 = $167;
   $168 = $44;
   $169 = ((($168)) + 4|0);
   $170 = $69;
   $63 = $169;
   $64 = $170;
   $171 = $63;
   $172 = ((($171)) + 4|0);
   $173 = HEAP8[$172>>0]|0;
   $174 = $173&1;
   if ($174) {
    $175 = HEAP32[$171>>2]|0;
    $176 = $64;
    $177 = ((($176)) + 16|0);
    $62 = $177;
    $178 = $62;
    $61 = $178;
    $179 = $61;
    $58 = $175;
    $59 = $179;
    $180 = $58;
    $181 = $59;
    ;HEAP8[$57>>0]=HEAP8[$60>>0]|0;
    $55 = $180;
    $56 = $181;
   }
   $182 = $64;
   $183 = ($182|0)!=(0|0);
   if (!($183)) {
    $203 = $89;
    $204 = $90;
    ___resumeException($203|0);
    // unreachable;
   }
   $184 = HEAP32[$171>>2]|0;
   $185 = $64;
   $52 = $184;
   $53 = $185;
   $54 = 1;
   $186 = $52;
   $187 = $53;
   $188 = $54;
   $49 = $186;
   $50 = $187;
   $51 = $188;
   $189 = $50;
   $48 = $189;
   $190 = $48;
   __ZdlPv($190);
   $203 = $89;
   $204 = $90;
   ___resumeException($203|0);
   // unreachable;
  }
 } while(0);
 $191 = $86;
 $71 = $91;
 $72 = $191;
 $192 = $71;
 $193 = $72;
 HEAP32[$192>>2] = $193;
 $75 = $0;
 $76 = $91;
 $77 = $87;
 $194 = $75;
 $195 = $76;
 $74 = $195;
 $196 = $74;
 ;HEAP32[$194>>2]=HEAP32[$196>>2]|0;
 $197 = ((($194)) + 4|0);
 $198 = $77;
 $73 = $198;
 $199 = $73;
 $200 = HEAP8[$199>>0]|0;
 $201 = $200&1;
 $202 = $201&1;
 HEAP8[$197>>0] = $202;
 STACKTOP = sp;return;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE12__find_equalIS5_EERPNS_16__tree_node_baseIPvEERPNS_15__tree_end_nodeISK_EERKT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(160|0);
 $35 = $0;
 $36 = $1;
 $37 = $2;
 $40 = $35;
 $33 = $40;
 $41 = $33;
 $32 = $41;
 $42 = $32;
 $43 = ((($42)) + 4|0);
 $31 = $43;
 $44 = $31;
 $30 = $44;
 $45 = $30;
 $29 = $45;
 $46 = $29;
 $28 = $46;
 $47 = $28;
 $48 = HEAP32[$47>>2]|0;
 $38 = $48;
 $49 = (__ZNKSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE10__root_ptrEv($40)|0);
 $39 = $49;
 $50 = $38;
 $51 = ($50|0)!=(0|0);
 if (!($51)) {
  $27 = $40;
  $107 = $27;
  $108 = ((($107)) + 4|0);
  $26 = $108;
  $109 = $26;
  $25 = $109;
  $110 = $25;
  $24 = $110;
  $111 = $24;
  $23 = $111;
  $112 = $23;
  $113 = $36;
  HEAP32[$113>>2] = $112;
  $114 = $36;
  $115 = HEAP32[$114>>2]|0;
  $34 = $115;
  $116 = $34;
  STACKTOP = sp;return ($116|0);
 }
 while(1) {
  $5 = $40;
  $52 = $5;
  $53 = ((($52)) + 8|0);
  $4 = $53;
  $54 = $4;
  $3 = $54;
  $55 = $3;
  $56 = $37;
  $57 = $38;
  $58 = ((($57)) + 16|0);
  $9 = $55;
  $10 = $56;
  $11 = $58;
  $59 = $9;
  $60 = $10;
  $61 = $11;
  $6 = $59;
  $7 = $60;
  $8 = $61;
  $62 = $7;
  $63 = HEAP32[$62>>2]|0;
  $64 = $8;
  $65 = HEAP32[$64>>2]|0;
  $66 = ($63>>>0)<($65>>>0);
  if ($66) {
   $67 = $38;
   $68 = HEAP32[$67>>2]|0;
   $69 = ($68|0)!=(0|0);
   $70 = $38;
   if (!($69)) {
    label = 5;
    break;
   }
   $12 = $70;
   $71 = $12;
   $39 = $71;
   $72 = $38;
   $73 = HEAP32[$72>>2]|0;
   $38 = $73;
   continue;
  }
  $15 = $40;
  $77 = $15;
  $78 = ((($77)) + 8|0);
  $14 = $78;
  $79 = $14;
  $13 = $79;
  $80 = $13;
  $81 = $38;
  $82 = ((($81)) + 16|0);
  $83 = $37;
  $19 = $80;
  $20 = $82;
  $21 = $83;
  $84 = $19;
  $85 = $20;
  $86 = $21;
  $16 = $84;
  $17 = $85;
  $18 = $86;
  $87 = $17;
  $88 = HEAP32[$87>>2]|0;
  $89 = $18;
  $90 = HEAP32[$89>>2]|0;
  $91 = ($88>>>0)<($90>>>0);
  $92 = $38;
  if (!($91)) {
   label = 10;
   break;
  }
  $93 = ((($92)) + 4|0);
  $94 = HEAP32[$93>>2]|0;
  $95 = ($94|0)!=(0|0);
  $96 = $38;
  if (!($95)) {
   label = 9;
   break;
  }
  $97 = ((($96)) + 4|0);
  $22 = $97;
  $98 = $22;
  $39 = $98;
  $99 = $38;
  $100 = ((($99)) + 4|0);
  $101 = HEAP32[$100>>2]|0;
  $38 = $101;
 }
 if ((label|0) == 5) {
  $74 = $36;
  HEAP32[$74>>2] = $70;
  $75 = $36;
  $76 = HEAP32[$75>>2]|0;
  $34 = $76;
  $116 = $34;
  STACKTOP = sp;return ($116|0);
 }
 else if ((label|0) == 9) {
  $102 = $36;
  HEAP32[$102>>2] = $96;
  $103 = $38;
  $104 = ((($103)) + 4|0);
  $34 = $104;
  $116 = $34;
  STACKTOP = sp;return ($116|0);
 }
 else if ((label|0) == 10) {
  $105 = $36;
  HEAP32[$105>>2] = $92;
  $106 = $39;
  $34 = $106;
  $116 = $34;
  STACKTOP = sp;return ($116|0);
 }
 return (0)|0;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE16__construct_nodeIJRKNS_21piecewise_construct_tENS_5tupleIJRKS5_EEENSK_IJEEEEEENS_10unique_ptrINS_11__tree_nodeIS8_PvEENS_22__tree_node_destructorINSD_ISS_EEEEEEDpOT_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$expand_i1_val = 0, $$expand_i1_val2 = 0, $$pre_trunc = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 528|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(528|0);
 $18 = sp + 72|0;
 $19 = sp + 64|0;
 $20 = sp + 56|0;
 $22 = sp + 524|0;
 $23 = sp + 523|0;
 $24 = sp + 522|0;
 $25 = sp + 48|0;
 $26 = sp + 40|0;
 $27 = sp + 32|0;
 $34 = sp + 521|0;
 $35 = sp + 432|0;
 $36 = sp + 520|0;
 $43 = sp + 24|0;
 $50 = sp + 519|0;
 $65 = sp + 324|0;
 $66 = sp + 16|0;
 $69 = sp + 312|0;
 $70 = sp + 304|0;
 $71 = sp + 8|0;
 $76 = sp + 280|0;
 $105 = sp;
 $108 = sp + 517|0;
 $127 = sp + 516|0;
 $128 = sp + 80|0;
 $122 = $1;
 $123 = $2;
 $124 = $3;
 $125 = $4;
 $129 = $122;
 $121 = $129;
 $130 = $121;
 $131 = ((($130)) + 4|0);
 $120 = $131;
 $132 = $120;
 $119 = $132;
 $133 = $119;
 $126 = $133;
 $$expand_i1_val = 0;
 HEAP8[$127>>0] = $$expand_i1_val;
 $134 = $126;
 $90 = $134;
 $91 = 1;
 $135 = $90;
 $136 = $91;
 $87 = $135;
 $88 = $136;
 $89 = 0;
 $137 = $87;
 $138 = $88;
 $86 = $137;
 $139 = ($138>>>0)>(178956970);
 if ($139) {
  $82 = 158021;
  $140 = (___cxa_allocate_exception(8)|0);
  $141 = $82;
  $80 = $140;
  $81 = $141;
  $142 = $80;
  $143 = $81;
  __THREW__ = 0;
  invoke_vii(751,($142|0),($143|0));
  $144 = __THREW__; __THREW__ = 0;
  $145 = $144&1;
  if ($145) {
   $146 = ___cxa_find_matching_catch_2()|0;
   $147 = tempRet0;
   $83 = $146;
   $84 = $147;
   ___cxa_free_exception(($140|0));
   $148 = $83;
   $149 = $84;
   ___resumeException($148|0);
   // unreachable;
  } else {
   HEAP32[$142>>2] = (22172);
   ___cxa_throw(($140|0),(2528|0),(713|0));
   // unreachable;
  }
 }
 $150 = $88;
 $151 = ($150*24)|0;
 $85 = $151;
 $152 = $85;
 $153 = (__Znwj($152)|0);
 $154 = $126;
 $77 = $128;
 $78 = $154;
 $79 = 0;
 $155 = $77;
 $156 = $78;
 HEAP32[$155>>2] = $156;
 $157 = ((($155)) + 4|0);
 $158 = $79;
 $159 = $158&1;
 $160 = $159&1;
 HEAP8[$157>>0] = $160;
 $73 = $0;
 $74 = $153;
 $75 = $128;
 $161 = $73;
 $162 = $74;
 $163 = $75;
 $72 = $163;
 $164 = $72;
 ;HEAP32[$76>>2]=HEAP32[$164>>2]|0;HEAP32[$76+4>>2]=HEAP32[$164+4>>2]|0;
 ;HEAP8[$71>>0]=HEAP8[$76>>0]|0;HEAP8[$71+1>>0]=HEAP8[$76+1>>0]|0;HEAP8[$71+2>>0]=HEAP8[$76+2>>0]|0;HEAP8[$71+3>>0]=HEAP8[$76+3>>0]|0;HEAP8[$71+4>>0]=HEAP8[$76+4>>0]|0;HEAP8[$71+5>>0]=HEAP8[$76+5>>0]|0;HEAP8[$71+6>>0]=HEAP8[$76+6>>0]|0;HEAP8[$71+7>>0]=HEAP8[$76+7>>0]|0;
 $68 = $161;
 HEAP32[$69>>2] = $162;
 $165 = $68;
 $67 = $69;
 $166 = $67;
 $167 = HEAP32[$166>>2]|0;
 $61 = $71;
 $168 = $61;
 ;HEAP32[$70>>2]=HEAP32[$168>>2]|0;HEAP32[$70+4>>2]=HEAP32[$168+4>>2]|0;
 ;HEAP8[$66>>0]=HEAP8[$70>>0]|0;HEAP8[$66+1>>0]=HEAP8[$70+1>>0]|0;HEAP8[$66+2>>0]=HEAP8[$70+2>>0]|0;HEAP8[$66+3>>0]=HEAP8[$70+3>>0]|0;HEAP8[$66+4>>0]=HEAP8[$70+4>>0]|0;HEAP8[$66+5>>0]=HEAP8[$70+5>>0]|0;HEAP8[$66+6>>0]=HEAP8[$70+6>>0]|0;HEAP8[$66+7>>0]=HEAP8[$70+7>>0]|0;
 $64 = $165;
 HEAP32[$65>>2] = $167;
 $169 = $64;
 $63 = $65;
 $170 = $63;
 $171 = HEAP32[$170>>2]|0;
 HEAP32[$169>>2] = $171;
 $172 = ((($169)) + 4|0);
 $62 = $66;
 $173 = $62;
 ;HEAP32[$172>>2]=HEAP32[$173>>2]|0;HEAP32[$172+4>>2]=HEAP32[$173+4>>2]|0;
 $174 = $126;
 $60 = $0;
 $175 = $60;
 $59 = $175;
 $176 = $59;
 $58 = $176;
 $177 = $58;
 $178 = HEAP32[$177>>2]|0;
 $179 = ((($178)) + 16|0);
 $57 = $179;
 $180 = $57;
 $56 = $180;
 $181 = $56;
 $182 = $123;
 $55 = $182;
 $183 = $55;
 $184 = $124;
 $54 = $184;
 $185 = $54;
 $186 = $125;
 $5 = $186;
 $187 = $5;
 $45 = $174;
 $46 = $181;
 $47 = $183;
 $48 = $185;
 $49 = $187;
 $188 = $45;
 $189 = $46;
 $190 = $47;
 $44 = $190;
 $191 = $44;
 $192 = $48;
 $6 = $192;
 $193 = $6;
 $194 = $49;
 $7 = $194;
 $195 = $7;
 ;HEAP8[$43>>0]=HEAP8[$50>>0]|0;
 $38 = $188;
 $39 = $189;
 $40 = $191;
 $41 = $193;
 $42 = $195;
 $196 = $38;
 $197 = $39;
 $198 = $40;
 $37 = $198;
 $199 = $37;
 $200 = $41;
 $8 = $200;
 $201 = $8;
 $202 = $42;
 $9 = $202;
 $203 = $9;
 $29 = $196;
 $30 = $197;
 $31 = $199;
 $32 = $201;
 $33 = $203;
 $204 = $30;
 $205 = $31;
 $28 = $205;
 $206 = $32;
 $10 = $206;
 $207 = $10;
 ;HEAP32[$35>>2]=HEAP32[$207>>2]|0;
 $208 = $33;
 $11 = $208;
 ;HEAP8[$25>>0]=HEAP8[$36>>0]|0;
 ;HEAP8[$26>>0]=HEAP8[$35>>0]|0;HEAP8[$26+1>>0]=HEAP8[$35+1>>0]|0;HEAP8[$26+2>>0]=HEAP8[$35+2>>0]|0;HEAP8[$26+3>>0]=HEAP8[$35+3>>0]|0;
 ;HEAP8[$27>>0]=HEAP8[$34>>0]|0;
 $21 = $204;
 $209 = $21;
 ;HEAP8[$18>>0]=HEAP8[$24>>0]|0;
 ;HEAP8[$19>>0]=HEAP8[$23>>0]|0;
 ;HEAP8[$20>>0]=HEAP8[$22>>0]|0;
 $15 = $209;
 $16 = $26;
 $17 = $25;
 $210 = $15;
 $211 = $16;
 $14 = $211;
 $212 = $14;
 $13 = $212;
 $213 = $13;
 $214 = HEAP32[$213>>2]|0;
 $12 = $214;
 $215 = $12;
 $216 = HEAP32[$215>>2]|0;
 HEAP32[$210>>2] = $216;
 $217 = ((($210)) + 4|0);
 HEAP32[$217>>2] = 0;
 $53 = $0;
 $218 = $53;
 $52 = $218;
 $219 = $52;
 $51 = $219;
 $220 = $51;
 $221 = ((($220)) + 4|0);
 $222 = ((($221)) + 4|0);
 HEAP8[$222>>0] = 1;
 $$expand_i1_val2 = 1;
 HEAP8[$127>>0] = $$expand_i1_val2;
 $$pre_trunc = HEAP8[$127>>0]|0;
 $223 = $$pre_trunc&1;
 if ($223) {
  STACKTOP = sp;return;
 }
 $118 = $0;
 $224 = $118;
 $115 = $224;
 $116 = 0;
 $225 = $115;
 $114 = $225;
 $226 = $114;
 $113 = $226;
 $227 = $113;
 $228 = HEAP32[$227>>2]|0;
 $117 = $228;
 $229 = $116;
 $95 = $225;
 $230 = $95;
 $94 = $230;
 $231 = $94;
 HEAP32[$231>>2] = $229;
 $232 = $117;
 $233 = ($232|0)!=(0|0);
 if (!($233)) {
  STACKTOP = sp;return;
 }
 $93 = $225;
 $234 = $93;
 $92 = $234;
 $235 = $92;
 $236 = ((($235)) + 4|0);
 $237 = $117;
 $111 = $236;
 $112 = $237;
 $238 = $111;
 $239 = ((($238)) + 4|0);
 $240 = HEAP8[$239>>0]|0;
 $241 = $240&1;
 if ($241) {
  $242 = HEAP32[$238>>2]|0;
  $243 = $112;
  $244 = ((($243)) + 16|0);
  $110 = $244;
  $245 = $110;
  $109 = $245;
  $246 = $109;
  $106 = $242;
  $107 = $246;
  $247 = $106;
  $248 = $107;
  ;HEAP8[$105>>0]=HEAP8[$108>>0]|0;
  $103 = $247;
  $104 = $248;
 }
 $249 = $112;
 $250 = ($249|0)!=(0|0);
 if (!($250)) {
  STACKTOP = sp;return;
 }
 $251 = HEAP32[$238>>2]|0;
 $252 = $112;
 $100 = $251;
 $101 = $252;
 $102 = 1;
 $253 = $100;
 $254 = $101;
 $255 = $102;
 $97 = $253;
 $98 = $254;
 $99 = $255;
 $256 = $98;
 $96 = $256;
 $257 = $96;
 __ZdlPv($257);
 STACKTOP = sp;return;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE16__insert_node_atEPNS_15__tree_end_nodeIPNS_16__tree_node_baseIPvEEEERSK_SK_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $15 = $0;
 $16 = $1;
 $17 = $2;
 $18 = $3;
 $19 = $15;
 $20 = $18;
 HEAP32[$20>>2] = 0;
 $21 = $18;
 $22 = ((($21)) + 4|0);
 HEAP32[$22>>2] = 0;
 $23 = $16;
 $24 = $18;
 $25 = ((($24)) + 8|0);
 HEAP32[$25>>2] = $23;
 $26 = $18;
 $27 = $17;
 HEAP32[$27>>2] = $26;
 $14 = $19;
 $28 = $14;
 $29 = HEAP32[$28>>2]|0;
 $30 = HEAP32[$29>>2]|0;
 $31 = ($30|0)!=(0|0);
 if ($31) {
  $10 = $19;
  $32 = $10;
  $33 = HEAP32[$32>>2]|0;
  $34 = HEAP32[$33>>2]|0;
  $4 = $19;
  $35 = $4;
  HEAP32[$35>>2] = $34;
 }
 $9 = $19;
 $36 = $9;
 $37 = ((($36)) + 4|0);
 $8 = $37;
 $38 = $8;
 $7 = $38;
 $39 = $7;
 $6 = $39;
 $40 = $6;
 $5 = $40;
 $41 = $5;
 $42 = HEAP32[$41>>2]|0;
 $43 = $17;
 $44 = HEAP32[$43>>2]|0;
 __ZNSt3__227__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($42,$44);
 $13 = $19;
 $45 = $13;
 $46 = ((($45)) + 8|0);
 $12 = $46;
 $47 = $12;
 $11 = $47;
 $48 = $11;
 $49 = HEAP32[$48>>2]|0;
 $50 = (($49) + 1)|0;
 HEAP32[$48>>2] = $50;
 STACKTOP = sp;return;
}
function __ZNKSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE10__root_ptrEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $7 = $0;
 $8 = $7;
 $6 = $8;
 $9 = $6;
 $10 = ((($9)) + 4|0);
 $5 = $10;
 $11 = $5;
 $4 = $11;
 $12 = $4;
 $3 = $12;
 $13 = $3;
 $2 = $13;
 $14 = $2;
 $1 = $14;
 $15 = $1;
 STACKTOP = sp;return ($15|0);
}
function __ZNSt3__227__tree_balance_after_insertIPNS_16__tree_node_baseIPvEEEEvT_S5_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $20 = $0;
 $21 = $1;
 $24 = $21;
 $25 = $20;
 $26 = ($24|0)==($25|0);
 $27 = $21;
 $28 = ((($27)) + 12|0);
 $29 = $26&1;
 HEAP8[$28>>0] = $29;
 while(1) {
  $30 = $21;
  $31 = $20;
  $32 = ($30|0)!=($31|0);
  if (!($32)) {
   label = 17;
   break;
  }
  $33 = $21;
  $19 = $33;
  $34 = $19;
  $35 = ((($34)) + 8|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ((($36)) + 12|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = $38&1;
  $40 = $39 ^ 1;
  if (!($40)) {
   label = 17;
   break;
  }
  $41 = $21;
  $18 = $41;
  $42 = $18;
  $43 = ((($42)) + 8|0);
  $44 = HEAP32[$43>>2]|0;
  $17 = $44;
  $45 = $17;
  $46 = $17;
  $47 = ((($46)) + 8|0);
  $48 = HEAP32[$47>>2]|0;
  $49 = HEAP32[$48>>2]|0;
  $50 = ($45|0)==($49|0);
  $51 = $21;
  if ($50) {
   $14 = $51;
   $52 = $14;
   $53 = ((($52)) + 8|0);
   $54 = HEAP32[$53>>2]|0;
   $8 = $54;
   $55 = $8;
   $56 = ((($55)) + 8|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = ((($57)) + 4|0);
   $59 = HEAP32[$58>>2]|0;
   $22 = $59;
   $60 = $22;
   $61 = ($60|0)!=(0|0);
   if (!($61)) {
    label = 8;
    break;
   }
   $62 = $22;
   $63 = ((($62)) + 12|0);
   $64 = HEAP8[$63>>0]|0;
   $65 = $64&1;
   if ($65) {
    label = 8;
    break;
   }
   $66 = $21;
   $5 = $66;
   $67 = $5;
   $68 = ((($67)) + 8|0);
   $69 = HEAP32[$68>>2]|0;
   $21 = $69;
   $70 = $21;
   $71 = ((($70)) + 12|0);
   HEAP8[$71>>0] = 1;
   $72 = $21;
   $2 = $72;
   $73 = $2;
   $74 = ((($73)) + 8|0);
   $75 = HEAP32[$74>>2]|0;
   $21 = $75;
   $76 = $21;
   $77 = $20;
   $78 = ($76|0)==($77|0);
   $79 = $21;
   $80 = ((($79)) + 12|0);
   $81 = $78&1;
   HEAP8[$80>>0] = $81;
   $82 = $22;
   $83 = ((($82)) + 12|0);
   HEAP8[$83>>0] = 1;
   continue;
  } else {
   $9 = $51;
   $109 = $9;
   $110 = ((($109)) + 8|0);
   $111 = HEAP32[$110>>2]|0;
   $112 = ((($111)) + 8|0);
   $113 = HEAP32[$112>>2]|0;
   $114 = HEAP32[$113>>2]|0;
   $23 = $114;
   $115 = $23;
   $116 = ($115|0)!=(0|0);
   if (!($116)) {
    label = 14;
    break;
   }
   $117 = $23;
   $118 = ((($117)) + 12|0);
   $119 = HEAP8[$118>>0]|0;
   $120 = $119&1;
   if ($120) {
    label = 14;
    break;
   }
   $121 = $21;
   $10 = $121;
   $122 = $10;
   $123 = ((($122)) + 8|0);
   $124 = HEAP32[$123>>2]|0;
   $21 = $124;
   $125 = $21;
   $126 = ((($125)) + 12|0);
   HEAP8[$126>>0] = 1;
   $127 = $21;
   $11 = $127;
   $128 = $11;
   $129 = ((($128)) + 8|0);
   $130 = HEAP32[$129>>2]|0;
   $21 = $130;
   $131 = $21;
   $132 = $20;
   $133 = ($131|0)==($132|0);
   $134 = $21;
   $135 = ((($134)) + 12|0);
   $136 = $133&1;
   HEAP8[$135>>0] = $136;
   $137 = $23;
   $138 = ((($137)) + 12|0);
   HEAP8[$138>>0] = 1;
   continue;
  }
 }
 if ((label|0) == 8) {
  $84 = $21;
  $3 = $84;
  $85 = $3;
  $86 = $3;
  $87 = ((($86)) + 8|0);
  $88 = HEAP32[$87>>2]|0;
  $89 = HEAP32[$88>>2]|0;
  $90 = ($85|0)==($89|0);
  if (!($90)) {
   $91 = $21;
   $4 = $91;
   $92 = $4;
   $93 = ((($92)) + 8|0);
   $94 = HEAP32[$93>>2]|0;
   $21 = $94;
   $95 = $21;
   __ZNSt3__218__tree_left_rotateIPNS_16__tree_node_baseIPvEEEEvT_($95);
  }
  $96 = $21;
  $6 = $96;
  $97 = $6;
  $98 = ((($97)) + 8|0);
  $99 = HEAP32[$98>>2]|0;
  $21 = $99;
  $100 = $21;
  $101 = ((($100)) + 12|0);
  HEAP8[$101>>0] = 1;
  $102 = $21;
  $7 = $102;
  $103 = $7;
  $104 = ((($103)) + 8|0);
  $105 = HEAP32[$104>>2]|0;
  $21 = $105;
  $106 = $21;
  $107 = ((($106)) + 12|0);
  HEAP8[$107>>0] = 0;
  $108 = $21;
  __ZNSt3__219__tree_right_rotateIPNS_16__tree_node_baseIPvEEEEvT_($108);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 14) {
  $139 = $21;
  $12 = $139;
  $140 = $12;
  $141 = $12;
  $142 = ((($141)) + 8|0);
  $143 = HEAP32[$142>>2]|0;
  $144 = HEAP32[$143>>2]|0;
  $145 = ($140|0)==($144|0);
  if ($145) {
   $146 = $21;
   $13 = $146;
   $147 = $13;
   $148 = ((($147)) + 8|0);
   $149 = HEAP32[$148>>2]|0;
   $21 = $149;
   $150 = $21;
   __ZNSt3__219__tree_right_rotateIPNS_16__tree_node_baseIPvEEEEvT_($150);
  }
  $151 = $21;
  $15 = $151;
  $152 = $15;
  $153 = ((($152)) + 8|0);
  $154 = HEAP32[$153>>2]|0;
  $21 = $154;
  $155 = $21;
  $156 = ((($155)) + 12|0);
  HEAP8[$156>>0] = 1;
  $157 = $21;
  $16 = $157;
  $158 = $16;
  $159 = ((($158)) + 8|0);
  $160 = HEAP32[$159>>2]|0;
  $21 = $160;
  $161 = $21;
  $162 = ((($161)) + 12|0);
  HEAP8[$162>>0] = 0;
  $163 = $21;
  __ZNSt3__218__tree_left_rotateIPNS_16__tree_node_baseIPvEEEEvT_($163);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 17) {
  STACKTOP = sp;return;
 }
}
function __ZNSt3__218__tree_left_rotateIPNS_16__tree_node_baseIPvEEEEvT_($0) {
 $0 = $0|0;
 var $$sink = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $7 = $0;
 $9 = $7;
 $10 = ((($9)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $8 = $11;
 $12 = $8;
 $13 = HEAP32[$12>>2]|0;
 $14 = $7;
 $15 = ((($14)) + 4|0);
 HEAP32[$15>>2] = $13;
 $16 = $7;
 $17 = ((($16)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)!=(0|0);
 if ($19) {
  $20 = $7;
  $21 = ((($20)) + 4|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = $7;
  $5 = $22;
  $6 = $23;
  $24 = $5;
  $25 = $6;
  $26 = ((($24)) + 8|0);
  HEAP32[$26>>2] = $25;
 }
 $27 = $7;
 $28 = ((($27)) + 8|0);
 $29 = HEAP32[$28>>2]|0;
 $30 = $8;
 $31 = ((($30)) + 8|0);
 HEAP32[$31>>2] = $29;
 $32 = $7;
 $4 = $32;
 $33 = $4;
 $34 = $4;
 $35 = ((($34)) + 8|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = HEAP32[$36>>2]|0;
 $38 = ($33|0)==($37|0);
 $39 = $8;
 $40 = $7;
 if ($38) {
  $41 = ((($40)) + 8|0);
  $42 = HEAP32[$41>>2]|0;
  $$sink = $42;
  HEAP32[$$sink>>2] = $39;
  $47 = $7;
  $48 = $8;
  HEAP32[$48>>2] = $47;
  $49 = $7;
  $50 = $8;
  $2 = $49;
  $3 = $50;
  $51 = $2;
  $52 = $3;
  $53 = ((($51)) + 8|0);
  HEAP32[$53>>2] = $52;
  STACKTOP = sp;return;
 } else {
  $1 = $40;
  $43 = $1;
  $44 = ((($43)) + 8|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ((($45)) + 4|0);
  $$sink = $46;
  HEAP32[$$sink>>2] = $39;
  $47 = $7;
  $48 = $8;
  HEAP32[$48>>2] = $47;
  $49 = $7;
  $50 = $8;
  $2 = $49;
  $3 = $50;
  $51 = $2;
  $52 = $3;
  $53 = ((($51)) + 8|0);
  HEAP32[$53>>2] = $52;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__219__tree_right_rotateIPNS_16__tree_node_baseIPvEEEEvT_($0) {
 $0 = $0|0;
 var $$sink = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $7 = $0;
 $9 = $7;
 $10 = HEAP32[$9>>2]|0;
 $8 = $10;
 $11 = $8;
 $12 = ((($11)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $7;
 HEAP32[$14>>2] = $13;
 $15 = $7;
 $16 = HEAP32[$15>>2]|0;
 $17 = ($16|0)!=(0|0);
 if ($17) {
  $18 = $7;
  $19 = HEAP32[$18>>2]|0;
  $20 = $7;
  $5 = $19;
  $6 = $20;
  $21 = $5;
  $22 = $6;
  $23 = ((($21)) + 8|0);
  HEAP32[$23>>2] = $22;
 }
 $24 = $7;
 $25 = ((($24)) + 8|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = $8;
 $28 = ((($27)) + 8|0);
 HEAP32[$28>>2] = $26;
 $29 = $7;
 $4 = $29;
 $30 = $4;
 $31 = $4;
 $32 = ((($31)) + 8|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = HEAP32[$33>>2]|0;
 $35 = ($30|0)==($34|0);
 $36 = $8;
 $37 = $7;
 if ($35) {
  $38 = ((($37)) + 8|0);
  $39 = HEAP32[$38>>2]|0;
  $$sink = $39;
 } else {
  $1 = $37;
  $40 = $1;
  $41 = ((($40)) + 8|0);
  $42 = HEAP32[$41>>2]|0;
  $43 = ((($42)) + 4|0);
  $$sink = $43;
 }
 HEAP32[$$sink>>2] = $36;
 $44 = $7;
 $45 = $8;
 $46 = ((($45)) + 4|0);
 HEAP32[$46>>2] = $44;
 $47 = $7;
 $48 = $8;
 $2 = $47;
 $3 = $48;
 $49 = $2;
 $50 = $3;
 $51 = ((($49)) + 8|0);
 HEAP32[$51>>2] = $50;
 STACKTOP = sp;return;
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE4findIS5_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(208|0);
 $14 = sp + 144|0;
 $38 = sp + 48|0;
 $46 = sp + 16|0;
 $49 = sp + 4|0;
 $50 = sp;
 $47 = $0;
 $48 = $1;
 $51 = $47;
 $52 = $48;
 $45 = $51;
 $53 = $45;
 $44 = $53;
 $54 = $44;
 $55 = ((($54)) + 4|0);
 $43 = $55;
 $56 = $43;
 $42 = $56;
 $57 = $42;
 $41 = $57;
 $58 = $41;
 $40 = $58;
 $59 = $40;
 $60 = HEAP32[$59>>2]|0;
 $6 = $51;
 $61 = $6;
 $62 = ((($61)) + 4|0);
 $5 = $62;
 $63 = $5;
 $4 = $63;
 $64 = $4;
 $3 = $64;
 $65 = $3;
 $2 = $65;
 $66 = $2;
 $67 = (__ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE13__lower_boundIS5_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_SL_PNS_15__tree_end_nodeIPNS_16__tree_node_baseISJ_EEEE($51,$52,$60,$66)|0);
 HEAP32[$49>>2] = $67;
 $15 = $51;
 $68 = $15;
 $13 = $68;
 $69 = $13;
 $70 = ((($69)) + 4|0);
 $12 = $70;
 $71 = $12;
 $11 = $71;
 $72 = $11;
 $10 = $72;
 $73 = $10;
 $9 = $73;
 $74 = $9;
 $7 = $14;
 $8 = $74;
 $75 = $7;
 $76 = $8;
 HEAP32[$75>>2] = $76;
 $77 = HEAP32[$14>>2]|0;
 HEAP32[$50>>2] = $77;
 $18 = $49;
 $19 = $50;
 $78 = $18;
 $79 = $19;
 $16 = $78;
 $17 = $79;
 $80 = $16;
 $81 = HEAP32[$80>>2]|0;
 $82 = $17;
 $83 = HEAP32[$82>>2]|0;
 $84 = ($81|0)==($83|0);
 $85 = $84 ^ 1;
 if ($85) {
  $22 = $51;
  $86 = $22;
  $87 = ((($86)) + 8|0);
  $21 = $87;
  $88 = $21;
  $20 = $88;
  $89 = $20;
  $90 = $48;
  $24 = $49;
  $91 = $24;
  $23 = $91;
  $92 = $23;
  $93 = HEAP32[$92>>2]|0;
  $94 = ((($93)) + 16|0);
  $28 = $89;
  $29 = $90;
  $30 = $94;
  $95 = $28;
  $96 = $29;
  $97 = $30;
  $25 = $95;
  $26 = $96;
  $27 = $97;
  $98 = $26;
  $99 = HEAP32[$98>>2]|0;
  $100 = $27;
  $101 = HEAP32[$100>>2]|0;
  $102 = ($99>>>0)<($101>>>0);
  $103 = $102 ^ 1;
  if ($103) {
   ;HEAP32[$46>>2]=HEAP32[$49>>2]|0;
   $114 = HEAP32[$46>>2]|0;
   STACKTOP = sp;return ($114|0);
  }
 }
 $39 = $51;
 $104 = $39;
 $37 = $104;
 $105 = $37;
 $106 = ((($105)) + 4|0);
 $36 = $106;
 $107 = $36;
 $35 = $107;
 $108 = $35;
 $34 = $108;
 $109 = $34;
 $33 = $109;
 $110 = $33;
 $31 = $38;
 $32 = $110;
 $111 = $31;
 $112 = $32;
 HEAP32[$111>>2] = $112;
 $113 = HEAP32[$38>>2]|0;
 HEAP32[$46>>2] = $113;
 $114 = HEAP32[$46>>2]|0;
 STACKTOP = sp;return ($114|0);
}
function __ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE13__lower_boundIS5_EENS_15__tree_iteratorIS8_PNS_11__tree_nodeIS8_PvEEiEERKT_SL_PNS_15__tree_end_nodeIPNS_16__tree_node_baseISJ_EEEE($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $15 = sp + 16|0;
 $16 = $0;
 $17 = $1;
 $18 = $2;
 $19 = $3;
 $20 = $16;
 while(1) {
  $21 = $18;
  $22 = ($21|0)!=(0|0);
  if (!($22)) {
   break;
  }
  $14 = $20;
  $23 = $14;
  $24 = ((($23)) + 8|0);
  $13 = $24;
  $25 = $13;
  $12 = $25;
  $26 = $12;
  $27 = $18;
  $28 = ((($27)) + 16|0);
  $29 = $17;
  $7 = $26;
  $8 = $28;
  $9 = $29;
  $30 = $7;
  $31 = $8;
  $32 = $9;
  $4 = $30;
  $5 = $31;
  $6 = $32;
  $33 = $5;
  $34 = HEAP32[$33>>2]|0;
  $35 = $6;
  $36 = HEAP32[$35>>2]|0;
  $37 = ($34>>>0)<($36>>>0);
  $38 = $18;
  if ($37) {
   $41 = ((($38)) + 4|0);
   $42 = HEAP32[$41>>2]|0;
   $18 = $42;
   continue;
  } else {
   $19 = $38;
   $39 = $18;
   $40 = HEAP32[$39>>2]|0;
   $18 = $40;
   continue;
  }
 }
 $43 = $19;
 $10 = $15;
 $11 = $43;
 $44 = $10;
 $45 = $11;
 HEAP32[$44>>2] = $45;
 $46 = HEAP32[$15>>2]|0;
 STACKTOP = sp;return ($46|0);
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $4 = $1;
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($4);
 HEAP32[$4>>2] = (2720);
 $5 = HEAP8[167373]|0;
 $6 = $5&1;
 if ($6) {
  STACKTOP = sp;return;
 }
 __THREW__ = 0;
 invoke_viiii(758,(167374|0),(167375|0),(161668|0),0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  $2 = $9;
  $3 = $10;
  __ZN9ExitGames6Common14DictionaryBaseD2Ev($4);
  $11 = $2;
  $12 = $3;
  ___resumeException($11|0);
  // unreachable;
 }
 HEAP8[167373] = 1;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $4;
 $9 = $7;
 $10 = (($8) + ($9)|0);
 HEAP8[$10>>0] = 98;
 $11 = $5;
 $12 = $7;
 $13 = (($11) + ($12)|0);
 HEAP8[$13>>0] = 122;
 $14 = $6;
 $15 = $7;
 $16 = (($14) + ($15<<2)|0);
 HEAP32[$16>>2] = 0;
 $17 = $7;
 $18 = ($17>>>0)<(1);
 if (!($18)) {
  STACKTOP = sp;return;
 }
 $19 = $4;
 $20 = $5;
 $21 = $6;
 $22 = $7;
 $23 = (($22) + 1)|0;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($19,$20,$21,$23);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE8toStringERNS0_7JStringEb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $7 = $2&1;
 $5 = $7;
 $8 = $3;
 $9 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($9,5676)|0);
 $6 = 0;
 while(1) {
  $10 = $6;
  $11 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $12 = ($10>>>0)<($11>>>0);
  if (!($12)) {
   break;
  }
  $13 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $14 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($13)|0);
  $15 = $6;
  $16 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($14,$15)|0);
  $17 = $4;
  $18 = $5;
  $19 = $18&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($16,$17,$19,1)|0);
  $20 = $4;
  (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($20,13052)|0);
  $21 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $22 = $6;
  $23 = (__ZNK9ExitGames6Common9HashtableixEj($21,$22)|0);
  $24 = $4;
  $25 = $5;
  $26 = $25&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($23,$24,$26,0)|0);
  $27 = $6;
  $28 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $29 = (($28) - 1)|0;
  $30 = ($27>>>0)<($29>>>0);
  if ($30) {
   $31 = $4;
   (__ZN9ExitGames6Common7JStringpLIA3_wEERS1_RKT_($31,12352)|0);
  }
  $32 = $6;
  $33 = (($32) + 1)|0;
  $6 = $33;
 }
 $34 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($34,10392)|0);
 $35 = $4;
 STACKTOP = sp;return ($35|0);
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE11getKeyTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167374|0);
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE13getValueTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167375|0);
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE18getValueDimensionsEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (161668|0);
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE6assignERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 (__ZN9ExitGames6Common14DictionaryBase6assignERKS1_($4,$5)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE7compareERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($4)|0);
 $6 = $3;
 $7 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($6)|0);
 $8 = (__ZNK9ExitGames6Common9HashtableeqERKS1_($5,$7)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE4copyEs($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $7 << 16 >> 16;
 $9 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryIhNS0_6ObjectEEEEEPT_j($8)|0);
 $4 = $9;
 $5 = 0;
 while(1) {
  $10 = $5;
  $11 = $10 << 16 >> 16;
  $12 = $3;
  $13 = $12 << 16 >> 16;
  $14 = ($11|0)<($13|0);
  if (!($14)) {
   break;
  }
  $15 = $5;
  $16 = $15 << 16 >> 16;
  $17 = (($6) + (($16*52)|0)|0);
  $18 = $4;
  $19 = $5;
  $20 = $19 << 16 >> 16;
  $21 = (($18) + (($20*52)|0)|0);
  (__ZN9ExitGames6Common14DictionaryBaseaSERKS1_($21,$17)|0);
  $22 = $5;
  $23 = (($22) + 1)<<16>>16;
  $5 = $23;
 }
 $24 = $4;
 STACKTOP = sp;return ($24|0);
}
function __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $4 = sp + 8|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 __ZN9ExitGames6Common7JStringC2Ej($4,0);
 $8 = $3;
 __THREW__ = 0;
 (invoke_iii(759,($4|0),($8|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  __THREW__ = 0;
  $11 = (invoke_iii(760,($7|0),($4|0))|0);
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   __ZN9ExitGames6Common7JStringD2Ev($4);
   STACKTOP = sp;return ($11|0);
  }
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 $5 = $14;
 $6 = $15;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 $16 = $5;
 $17 = $6;
 ___resumeException($16|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common7JStringpLIA3_wEERS1_RKT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $4 = sp + 8|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 __ZN9ExitGames6Common7JStringC2Ej($4,0);
 $8 = $3;
 __THREW__ = 0;
 (invoke_iii(759,($4|0),($8|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  __THREW__ = 0;
  $11 = (invoke_iii(760,($7|0),($4|0))|0);
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   __ZN9ExitGames6Common7JStringD2Ev($4);
   STACKTOP = sp;return ($11|0);
  }
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 $5 = $14;
 $6 = $15;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 $16 = $5;
 $17 = $6;
 ___resumeException($16|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryIhNS0_6ObjectEEEEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4*52)|0;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + (($35*52)|0)|0);
  __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEC2Ev($39);
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZN9ExitGames6Common14DictionaryBase3putIhNS0_6ObjectEEEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = ((($6)) + 4|0);
 $8 = $4;
 $9 = $5;
 __ZN9ExitGames6Common9Hashtable3putIhNS0_6ObjectEEEvRKT_RKT0_($7,$8,$9);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9Hashtable3putIhNS0_6ObjectEEEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(192|0);
 $6 = sp + 120|0;
 $7 = sp + 80|0;
 $10 = sp + 40|0;
 $11 = sp;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $12 = $3;
 __ZN9ExitGames6Common6ObjectC2Ev($6);
 $13 = $4;
 $14 = HEAP8[$13>>0]|0;
 __THREW__ = 0;
 invoke_vii(761,($7|0),($14|0));
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $8 = $24;
  $9 = $25;
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  $32 = $8;
  $33 = $9;
  ___resumeException($32|0);
  // unreachable;
 }
 $17 = $5;
 __THREW__ = 0;
 invoke_vii(744,($11|0),($17|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  $8 = $26;
  $9 = $27;
 } else {
  __THREW__ = 0;
  invoke_vii(746,($10|0),($11|0));
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  do {
   if ($21) {
    $28 = ___cxa_find_matching_catch_2()|0;
    $29 = tempRet0;
    $8 = $28;
    $9 = $29;
   } else {
    __THREW__ = 0;
    invoke_viii(747,($12|0),($7|0),($10|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $30 = ___cxa_find_matching_catch_2()|0;
     $31 = tempRet0;
     $8 = $30;
     $9 = $31;
     __ZN9ExitGames6Common6ObjectD2Ev($10);
     break;
    } else {
     __ZN9ExitGames6Common6ObjectD2Ev($10);
     __ZN9ExitGames6Common6ObjectD2Ev($11);
     __ZN9ExitGames6Common6ObjectD2Ev($7);
     __ZN9ExitGames6Common6ObjectD2Ev($6);
     STACKTOP = sp;return;
    }
   }
  } while(0);
  __ZN9ExitGames6Common6ObjectD2Ev($11);
 }
 __ZN9ExitGames6Common6ObjectD2Ev($7);
 __ZN9ExitGames6Common6ObjectD2Ev($6);
 $32 = $8;
 $33 = $9;
 ___resumeException($32|0);
 // unreachable;
}
function __ZN9ExitGames6Common7Helpers11KeyToObject3getIhEENS0_6ObjectET_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $2 = sp + 48|0;
 $3 = sp;
 HEAP8[$2>>0] = $1;
 __ZN9ExitGames6Common9KeyObjectIhEC2ERKh($3,$2);
 __THREW__ = 0;
 invoke_vii(744,($0|0),($3|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  $4 = $8;
  $5 = $9;
  __ZN9ExitGames6Common9KeyObjectIhED2Ev($3);
  $10 = $4;
  $11 = $5;
  ___resumeException($10|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common9KeyObjectIhED2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common9KeyObjectIhEC2ERKh($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,98,0,1);
 HEAP32[$4>>2] = (2768);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIhED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIhED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common9KeyObjectIhED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIhEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common9KeyObjectIhE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (98)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common9KeyObjectIhEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIhEC2ERKh($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,98,0,1);
 HEAP32[$4>>2] = (2804);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIhED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIhED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIhED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIhEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIhE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (98)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIhEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIsEC2ERKs($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,107,0,1);
 HEAP32[$4>>2] = (2840);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIsED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIsED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIsED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIsEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIsE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (107)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIsEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIiEC2ERKi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,105,0,1);
 HEAP32[$4>>2] = (2876);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIiED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIiED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIiED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIiEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIiE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (105)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIiEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIxEC2ERKx($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,108,0,1);
 HEAP32[$4>>2] = (2912);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIxED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIxED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIxED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIxEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIxE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (108)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIxEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIfEC2ERKf($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,102,0,1);
 HEAP32[$4>>2] = (2948);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIfED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIfED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIfED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIfEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIfE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (102)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIfEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIdEC2ERKd($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,100,0,1);
 HEAP32[$4>>2] = (2984);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIdED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIdED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIdED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIdEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIdE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (100)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIdEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIbEC2ERKb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,111,0,1);
 HEAP32[$4>>2] = (3020);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIbED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIbED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIbED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIbEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIbE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (111)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIbEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEEC2ERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,115,0,1);
 HEAP32[$4>>2] = (3056);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_7JStringEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (115)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_7JStringEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEC2ERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,104,0,1);
 HEAP32[$4>>2] = (3092);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (104)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryIhNS0_6ObjectEEEEEPT_v() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(52)|0);
 $0 = $1;
 $2 = $0;
 __ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEEC2Ev($2);
 $3 = $0;
 STACKTOP = sp;return ($3|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryIiNS0_6ObjectEEEEEPT_v() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(52)|0);
 $0 = $1;
 $2 = $0;
 __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEEC2Ev($2);
 $3 = $0;
 STACKTOP = sp;return ($3|0);
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEEC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $4 = $1;
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($4);
 HEAP32[$4>>2] = (3128);
 $5 = HEAP8[167376]|0;
 $6 = $5&1;
 if ($6) {
  STACKTOP = sp;return;
 }
 __THREW__ = 0;
 invoke_viiii(762,(167377|0),(167378|0),(161672|0),0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  $2 = $9;
  $3 = $10;
  __ZN9ExitGames6Common14DictionaryBaseD2Ev($4);
  $11 = $2;
  $12 = $3;
  ___resumeException($11|0);
  // unreachable;
 }
 HEAP8[167376] = 1;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $4;
 $9 = $7;
 $10 = (($8) + ($9)|0);
 HEAP8[$10>>0] = 105;
 $11 = $5;
 $12 = $7;
 $13 = (($11) + ($12)|0);
 HEAP8[$13>>0] = 122;
 $14 = $6;
 $15 = $7;
 $16 = (($14) + ($15<<2)|0);
 HEAP32[$16>>2] = 0;
 $17 = $7;
 $18 = ($17>>>0)<(1);
 if (!($18)) {
  STACKTOP = sp;return;
 }
 $19 = $4;
 $20 = $5;
 $21 = $6;
 $22 = $7;
 $23 = (($22) + 1)|0;
 __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($19,$20,$21,$23);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE8toStringERNS0_7JStringEb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $7 = $2&1;
 $5 = $7;
 $8 = $3;
 $9 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($9,5676)|0);
 $6 = 0;
 while(1) {
  $10 = $6;
  $11 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $12 = ($10>>>0)<($11>>>0);
  if (!($12)) {
   break;
  }
  $13 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $14 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($13)|0);
  $15 = $6;
  $16 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($14,$15)|0);
  $17 = $4;
  $18 = $5;
  $19 = $18&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($16,$17,$19,1)|0);
  $20 = $4;
  (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($20,13052)|0);
  $21 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $22 = $6;
  $23 = (__ZNK9ExitGames6Common9HashtableixEj($21,$22)|0);
  $24 = $4;
  $25 = $5;
  $26 = $25&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($23,$24,$26,0)|0);
  $27 = $6;
  $28 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $29 = (($28) - 1)|0;
  $30 = ($27>>>0)<($29>>>0);
  if ($30) {
   $31 = $4;
   (__ZN9ExitGames6Common7JStringpLIA3_wEERS1_RKT_($31,12352)|0);
  }
  $32 = $6;
  $33 = (($32) + 1)|0;
  $6 = $33;
 }
 $34 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($34,10392)|0);
 $35 = $4;
 STACKTOP = sp;return ($35|0);
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE11getKeyTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167377|0);
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE13getValueTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167378|0);
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE18getValueDimensionsEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (161672|0);
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE6assignERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 (__ZN9ExitGames6Common14DictionaryBase6assignERKS1_($4,$5)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE7compareERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($4)|0);
 $6 = $3;
 $7 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($6)|0);
 $8 = (__ZNK9ExitGames6Common9HashtableeqERKS1_($5,$7)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE4copyEs($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $7 << 16 >> 16;
 $9 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryIiNS0_6ObjectEEEEEPT_j($8)|0);
 $4 = $9;
 $5 = 0;
 while(1) {
  $10 = $5;
  $11 = $10 << 16 >> 16;
  $12 = $3;
  $13 = $12 << 16 >> 16;
  $14 = ($11|0)<($13|0);
  if (!($14)) {
   break;
  }
  $15 = $5;
  $16 = $15 << 16 >> 16;
  $17 = (($6) + (($16*52)|0)|0);
  $18 = $4;
  $19 = $5;
  $20 = $19 << 16 >> 16;
  $21 = (($18) + (($20*52)|0)|0);
  (__ZN9ExitGames6Common14DictionaryBaseaSERKS1_($21,$17)|0);
  $22 = $5;
  $23 = (($22) + 1)<<16>>16;
  $5 = $23;
 }
 $24 = $4;
 STACKTOP = sp;return ($24|0);
}
function __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryIiNS0_6ObjectEEEEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4*52)|0;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + (($35*52)|0)|0);
  __ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEEC2Ev($39);
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_10DictionaryINS0_6ObjectES4_EEEEPT_v() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(52)|0);
 $0 = $1;
 $2 = $0;
 __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_EC2Ev($2);
 $3 = $0;
 STACKTOP = sp;return ($3|0);
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_EC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $4 = $1;
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($4);
 HEAP32[$4>>2] = (3176);
 $5 = HEAP8[167379]|0;
 $6 = $5&1;
 if ($6) {
  STACKTOP = sp;return;
 }
 __THREW__ = 0;
 invoke_viiii(763,(167380|0),(167381|0),(161676|0),0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  $2 = $9;
  $3 = $10;
  __ZN9ExitGames6Common14DictionaryBaseD2Ev($4);
  $11 = $2;
  $12 = $3;
  ___resumeException($11|0);
  // unreachable;
 }
 HEAP8[167379] = 1;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $4;
 $9 = $7;
 $10 = (($8) + ($9)|0);
 HEAP8[$10>>0] = 122;
 $11 = $5;
 $12 = $7;
 $13 = (($11) + ($12)|0);
 HEAP8[$13>>0] = 122;
 $14 = $6;
 $15 = $7;
 $16 = (($14) + ($15<<2)|0);
 HEAP32[$16>>2] = 0;
 $17 = $7;
 $18 = ($17>>>0)<(1);
 if (!($18)) {
  STACKTOP = sp;return;
 }
 $19 = $4;
 $20 = $5;
 $21 = $6;
 $22 = $7;
 $23 = (($22) + 1)|0;
 __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($19,$20,$21,$23);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_ED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_ED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_ED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E8toStringERNS0_7JStringEb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $7 = $2&1;
 $5 = $7;
 $8 = $3;
 $9 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($9,5676)|0);
 $6 = 0;
 while(1) {
  $10 = $6;
  $11 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $12 = ($10>>>0)<($11>>>0);
  if (!($12)) {
   break;
  }
  $13 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $14 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($13)|0);
  $15 = $6;
  $16 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEEixEj($14,$15)|0);
  $17 = $4;
  $18 = $5;
  $19 = $18&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($16,$17,$19,0)|0);
  $20 = $4;
  (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($20,13052)|0);
  $21 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($8)|0);
  $22 = $6;
  $23 = (__ZNK9ExitGames6Common9HashtableixEj($21,$22)|0);
  $24 = $4;
  $25 = $5;
  $26 = $25&1;
  (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($23,$24,$26,0)|0);
  $27 = $6;
  $28 = (__ZNK9ExitGames6Common14DictionaryBase7getSizeEv($8)|0);
  $29 = (($28) - 1)|0;
  $30 = ($27>>>0)<($29>>>0);
  if ($30) {
   $31 = $4;
   (__ZN9ExitGames6Common7JStringpLIA3_wEERS1_RKT_($31,12352)|0);
  }
  $32 = $6;
  $33 = (($32) + 1)|0;
  $6 = $33;
 }
 $34 = $4;
 (__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_($34,10392)|0);
 $35 = $4;
 STACKTOP = sp;return ($35|0);
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E11getKeyTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167380|0);
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E13getValueTypesEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (167381|0);
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E18getValueDimensionsEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 STACKTOP = sp;return (161676|0);
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E6assignERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 (__ZN9ExitGames6Common14DictionaryBase6assignERKS1_($4,$5)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E7compareERKNS0_14DictionaryBaseE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($4)|0);
 $6 = $3;
 $7 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($6)|0);
 $8 = (__ZNK9ExitGames6Common9HashtableeqERKS1_($5,$7)|0);
 STACKTOP = sp;return ($8|0);
}
function __ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E4copyEs($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $7 << 16 >> 16;
 $9 = (__ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryINS0_6ObjectES4_EEEEPT_j($8)|0);
 $4 = $9;
 $5 = 0;
 while(1) {
  $10 = $5;
  $11 = $10 << 16 >> 16;
  $12 = $3;
  $13 = $12 << 16 >> 16;
  $14 = ($11|0)<($13|0);
  if (!($14)) {
   break;
  }
  $15 = $5;
  $16 = $15 << 16 >> 16;
  $17 = (($6) + (($16*52)|0)|0);
  $18 = $4;
  $19 = $5;
  $20 = $19 << 16 >> 16;
  $21 = (($18) + (($20*52)|0)|0);
  (__ZN9ExitGames6Common14DictionaryBaseaSERKS1_($21,$17)|0);
  $22 = $5;
  $23 = (($22) + 1)<<16>>16;
  $5 = $23;
 }
 $24 = $4;
 STACKTOP = sp;return ($24|0);
}
function __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E11NestedInfosIS2_S2_E10fillArraysEPhS6_Pjj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = $0;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement13allocateArrayINS0_10DictionaryINS0_6ObjectES4_EEEEPT_j($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $1 = $0;
 $4 = $1;
 $5 = ($4*52)|0;
 $6 = (($5) + 8)|0;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $2 = $8;
 $9 = $1;
 $10 = $2;
 $11 = ((($10)) + -8|0);
 $12 = $11;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = $3;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = (($16) + 4)|0;
 $19 = $18;
 HEAP32[$19>>2] = 0;
 while(1) {
  $20 = $3;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + 4)|0;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = $1;
  $27 = ($25>>>0)<(0);
  $28 = ($22>>>0)<($26>>>0);
  $29 = ($25|0)==(0);
  $30 = $29 & $28;
  $31 = $27 | $30;
  $32 = $2;
  if (!($31)) {
   break;
  }
  $33 = $3;
  $34 = $33;
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) + 4)|0;
  $37 = $36;
  $38 = HEAP32[$37>>2]|0;
  $39 = (($32) + (($35*52)|0)|0);
  __ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_EC2Ev($39);
  $40 = $3;
  $41 = $40;
  $42 = HEAP32[$41>>2]|0;
  $43 = (($40) + 4)|0;
  $44 = $43;
  $45 = HEAP32[$44>>2]|0;
  $46 = (_i64Add(($42|0),($45|0),1,0)|0);
  $47 = tempRet0;
  $48 = $3;
  $49 = $48;
  HEAP32[$49>>2] = $46;
  $50 = (($48) + 4)|0;
  $51 = $50;
  HEAP32[$51>>2] = $47;
 }
 STACKTOP = sp;return ($32|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(68);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE28getDataAddressImplementationIS4_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE28getDataAddressImplementationIS4_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(68);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE28getDataAddressImplementationIS4_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE28getDataAddressImplementationIS4_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE14getDataAddressEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $1 = $0;
 $3 = $1;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(68);
 if (!($6)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $7 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($3)|0);
 $8 = ($7|0)==(0);
 if (!($8)) {
  $11 = 0;
  STACKTOP = sp;return ($11|0);
 }
 $9 = (__ZNK9ExitGames6Common6Object14getDataPointerEv($3)|0);
 $10 = (__ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE28getDataAddressImplementationIS4_EclEPKPKv($2,$9)|0);
 $11 = $10;
 STACKTOP = sp;return ($11|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE28getDataAddressImplementationIS4_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEES6_EEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEC2ERKS4_($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function __ZN9ExitGames6Common16MemoryManagement10deallocateINS0_14DictionaryBaseEEEvPKT_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ($2|0)!=(0|0);
 if (!($3)) {
  STACKTOP = sp;return;
 }
 $4 = $1;
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_vi[$6 & 2047]($4);
 $7 = $1;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($7);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEC2ERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,68,0,1);
 HEAP32[$4>>2] = (3224);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE6assignERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (68)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEaSERKS5_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEES6_EEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEC2ERKS4_($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEC2ERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,68,0,1);
 HEAP32[$4>>2] = (3260);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE6assignERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (68)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEaSERKS5_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_10DictionaryINS0_6ObjectES5_EEEES6_EEPT_RT0_($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $2 = $3;
 $4 = $2;
 $5 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEC2ERKS4_($4,$5);
 $6 = $2;
 STACKTOP = sp;return ($6|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEC2ERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,68,0,1);
 HEAP32[$4>>2] = (3296);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE6assignERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (68)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEaSERKS5_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPKhEC2ES3_i($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,98,0,$8,1);
 HEAP32[$6>>2] = (3332);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPKhED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPKhED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPKhED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPKhEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPKhE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (98)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPKhEaSERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPsEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,107,0,$9,1);
 HEAP32[$6>>2] = (3368);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPsED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPsED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPsED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPsEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPsE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (107)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPsEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPiEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,105,0,$9,1);
 HEAP32[$6>>2] = (3404);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPiED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPiED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPiED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPiEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPiE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (105)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPiEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPxEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,108,0,$9,1);
 HEAP32[$6>>2] = (3440);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPxED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPxED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPxED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPxEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPxE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (108)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPxEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPfEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,102,0,$9,1);
 HEAP32[$6>>2] = (3476);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPfED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPfED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPfED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPfEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPfE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (102)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPfEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPdEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,100,0,$9,1);
 HEAP32[$6>>2] = (3512);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPdED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPdED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPdED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPdEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPdE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (100)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPdEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEC2ES3_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,115,0,$9,1);
 HEAP32[$6>>2] = (3548);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (115)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEaSERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPbEC2ES2_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,111,0,$9,1);
 HEAP32[$6>>2] = (3584);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPbED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPbED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPbED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPbEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPbE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (111)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPbEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEC2ES3_s($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = $4;
 $8 = $5;
 $9 = $8 << 16 >> 16;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($6,$7,122,0,$9,1);
 HEAP32[$6>>2] = (3620);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEE6assignERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (122)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEaSERKS4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZNK9ExitGames6Common11ValueObjectIhE28getDataAddressImplementationIhEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIsE28getDataAddressImplementationIsEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIiE28getDataAddressImplementationIiEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIxE28getDataAddressImplementationIxEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIfE28getDataAddressImplementationIfEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIdE28getDataAddressImplementationIdEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_7JStringEE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIbE28getDataAddressImplementationIbEclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectINS0_9HashtableEE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPhE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPsE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPiE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPxE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPfE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPdE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_7JStringEE28getDataAddressImplementationIS3_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPbE28getDataAddressImplementationIS2_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_9HashtableEE28getDataAddressImplementationIS3_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common11ValueObjectIPNS0_6ObjectEE28getDataAddressImplementationIS3_EclEPKPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $3;
 STACKTOP = sp;return ($4|0);
}
function __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE12getElementAtEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE11verifyIndexEj($4,$5);
 $6 = ((($4)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (($7) + (($8*40)|0)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE11verifyIndexEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer = sp;
 $4 = sp + 16|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 $8 = $3;
 $9 = ((($7)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>=($10>>>0);
 if (!($11)) {
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKc($4,147471);
 __THREW__ = 0;
 $12 = (invoke_ii(764,($4|0))|0);
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,(162472|0),1,(3648|0),($12|0),1,566,(10968|0),($vararg_buffer|0));
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if (!($16)) {
   __ZN9ExitGames6Common7JStringD2Ev($4);
   STACKTOP = sp;return;
  }
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 $5 = $17;
 $6 = $18;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 $19 = $5;
 $20 = $6;
 ___resumeException($19|0);
 // unreachable;
}
function __ZN9ExitGames6Common14DictionaryBase3putIiNS0_6ObjectEEEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = ((($6)) + 4|0);
 $8 = $4;
 $9 = $5;
 __ZN9ExitGames6Common9Hashtable3putIiNS0_6ObjectEEEvRKT_RKT0_($7,$8,$9);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9Hashtable3putIiNS0_6ObjectEEEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(192|0);
 $6 = sp + 120|0;
 $7 = sp + 80|0;
 $10 = sp + 40|0;
 $11 = sp;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $12 = $3;
 __ZN9ExitGames6Common6ObjectC2Ev($6);
 $13 = $4;
 $14 = HEAP32[$13>>2]|0;
 __THREW__ = 0;
 invoke_vii(766,($7|0),($14|0));
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $8 = $24;
  $9 = $25;
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  $32 = $8;
  $33 = $9;
  ___resumeException($32|0);
  // unreachable;
 }
 $17 = $5;
 __THREW__ = 0;
 invoke_vii(744,($11|0),($17|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  $8 = $26;
  $9 = $27;
 } else {
  __THREW__ = 0;
  invoke_vii(746,($10|0),($11|0));
  $20 = __THREW__; __THREW__ = 0;
  $21 = $20&1;
  do {
   if ($21) {
    $28 = ___cxa_find_matching_catch_2()|0;
    $29 = tempRet0;
    $8 = $28;
    $9 = $29;
   } else {
    __THREW__ = 0;
    invoke_viii(747,($12|0),($7|0),($10|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $30 = ___cxa_find_matching_catch_2()|0;
     $31 = tempRet0;
     $8 = $30;
     $9 = $31;
     __ZN9ExitGames6Common6ObjectD2Ev($10);
     break;
    } else {
     __ZN9ExitGames6Common6ObjectD2Ev($10);
     __ZN9ExitGames6Common6ObjectD2Ev($11);
     __ZN9ExitGames6Common6ObjectD2Ev($7);
     __ZN9ExitGames6Common6ObjectD2Ev($6);
     STACKTOP = sp;return;
    }
   }
  } while(0);
  __ZN9ExitGames6Common6ObjectD2Ev($11);
 }
 __ZN9ExitGames6Common6ObjectD2Ev($7);
 __ZN9ExitGames6Common6ObjectD2Ev($6);
 $32 = $8;
 $33 = $9;
 ___resumeException($32|0);
 // unreachable;
}
function __ZN9ExitGames6Common7Helpers11KeyToObject3getIiEENS0_6ObjectET_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $2 = sp + 48|0;
 $3 = sp;
 HEAP32[$2>>2] = $1;
 __ZN9ExitGames6Common9KeyObjectIiEC2ERKi($3,$2);
 __THREW__ = 0;
 invoke_vii(744,($0|0),($3|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  $4 = $8;
  $5 = $9;
  __ZN9ExitGames6Common9KeyObjectIiED2Ev($3);
  $10 = $4;
  $11 = $5;
  ___resumeException($10|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common9KeyObjectIiED2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common9KeyObjectIiEC2ERKi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($4,$5,105,0,1);
 HEAP32[$4>>2] = (3784);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIiED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIiED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common9KeyObjectIiED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIiEaSERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common9KeyObjectIiE6assignERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 $6 = (__ZNK9ExitGames6Common6Object7getTypeEv($5)|0);
 $7 = $6&255;
 $8 = (105)==($7|0);
 if (!($8)) {
  STACKTOP = sp;return ($4|0);
 }
 $9 = $3;
 (__ZN9ExitGames6Common6Object6assignERKS1_($4,$9)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common9KeyObjectIiEaSERKS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (FUNCTION_TABLE_iii[$7 & 2047]($4,$8)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common14DictionaryBase3putINS0_6ObjectES3_EEvRKT_RKT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $6 = $3;
 $7 = ((($6)) + 4|0);
 $8 = $4;
 $9 = $5;
 __ZN9ExitGames6Common9Hashtable3putINS0_6ObjectES3_EEvRKT_RKT0_($7,$8,$9);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common14DictionaryBase7getKeysIhEENS0_7JVectorIT_EEPKS4_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$expand_i1_val = 0, $$expand_i1_val2 = 0, $$pre_trunc = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $5 = sp + 81|0;
 $6 = sp + 52|0;
 $10 = sp + 80|0;
 $11 = sp;
 $3 = $1;
 $4 = $2;
 $12 = $3;
 $$expand_i1_val = 0;
 HEAP8[$5>>0] = $$expand_i1_val;
 __ZN9ExitGames6Common7JVectorIhEC2Ejj($0,40,10);
 __THREW__ = 0;
 $13 = (invoke_ii(767,($12|0))|0);
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if (!($15)) {
  __THREW__ = 0;
  $16 = (invoke_ii(768,($13|0))|0);
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if (!($18)) {
   __THREW__ = 0;
   invoke_vii(769,($6|0),($16|0));
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    $9 = 0;
    while(1) {
     $21 = $9;
     __THREW__ = 0;
     $22 = (invoke_ii(770,($6|0))|0);
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      label = 13;
      break;
     }
     $25 = ($21>>>0)<($22>>>0);
     if (!($25)) {
      label = 15;
      break;
     }
     $26 = $9;
     __THREW__ = 0;
     $27 = (invoke_iii(771,($6|0),($26|0))|0);
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      label = 13;
      break;
     }
     __THREW__ = 0;
     invoke_vii(772,($11|0),($27|0));
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      label = 13;
      break;
     }
     __THREW__ = 0;
     $32 = (invoke_ii(773,($11|0))|0);
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      label = 14;
      break;
     }
     HEAP8[$10>>0] = $32;
     __THREW__ = 0;
     invoke_vii(774,($0|0),($10|0));
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      label = 14;
      break;
     }
     __ZN9ExitGames6Common9KeyObjectIhED2Ev($11);
     $37 = $9;
     $38 = (($37) + 1)|0;
     $9 = $38;
    }
    if ((label|0) == 13) {
     $41 = ___cxa_find_matching_catch_2()|0;
     $42 = tempRet0;
     $7 = $41;
     $8 = $42;
    }
    else if ((label|0) == 14) {
     $43 = ___cxa_find_matching_catch_2()|0;
     $44 = tempRet0;
     $7 = $43;
     $8 = $44;
     __ZN9ExitGames6Common9KeyObjectIhED2Ev($11);
    }
    else if ((label|0) == 15) {
     $$expand_i1_val2 = 1;
     HEAP8[$5>>0] = $$expand_i1_val2;
     __ZN9ExitGames6Common7JVectorINS0_6ObjectEED2Ev($6);
     $$pre_trunc = HEAP8[$5>>0]|0;
     $45 = $$pre_trunc&1;
     if ($45) {
      STACKTOP = sp;return;
     }
     __ZN9ExitGames6Common7JVectorIhED2Ev($0);
     STACKTOP = sp;return;
    }
    __ZN9ExitGames6Common7JVectorINS0_6ObjectEED2Ev($6);
    __ZN9ExitGames6Common7JVectorIhED2Ev($0);
    $46 = $7;
    $47 = $8;
    ___resumeException($46|0);
    // unreachable;
   }
  }
 }
 $39 = ___cxa_find_matching_catch_2()|0;
 $40 = tempRet0;
 $7 = $39;
 $8 = $40;
 __ZN9ExitGames6Common7JVectorIhED2Ev($0);
 $46 = $7;
 $47 = $8;
 ___resumeException($46|0);
 // unreachable;
}
function __ZN9ExitGames6Common7JVectorIhEC2Ejj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $8 = $3;
 __ZN9ExitGames6Common4BaseC2Ev($8);
 HEAP32[$8>>2] = (3820);
 $9 = ((($8)) + 4|0);
 HEAP32[$9>>2] = 0;
 $10 = $4;
 $11 = ((($8)) + 8|0);
 HEAP32[$11>>2] = $10;
 $12 = ((($8)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = $13;
 __THREW__ = 0;
 $15 = (invoke_ii(775,($14|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $6 = $21;
  $7 = $22;
  __ZN9ExitGames6Common4BaseD2Ev($8);
  $23 = $6;
  $24 = $7;
  ___resumeException($23|0);
  // unreachable;
 } else {
  $18 = ((($8)) + 16|0);
  HEAP32[$18>>2] = $15;
  $19 = $5;
  $20 = ((($8)) + 12|0);
  HEAP32[$20>>2] = $19;
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEEC2ERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 __ZN9ExitGames6Common4BaseC2Ev($6);
 HEAP32[$6>>2] = (3844);
 $7 = ((($6)) + 4|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($6)) + 8|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($6)) + 12|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($6)) + 16|0);
 HEAP32[$10>>2] = 0;
 $11 = $3;
 __THREW__ = 0;
 (invoke_iii(776,($6|0),($11|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $14 = ___cxa_find_matching_catch_2()|0;
  $15 = tempRet0;
  $4 = $14;
  $5 = $15;
  __ZN9ExitGames6Common4BaseD2Ev($6);
  $16 = $4;
  $17 = $5;
  ___resumeException($16|0);
  // unreachable;
 } else {
  STACKTOP = sp;return;
 }
}
function __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE7getSizeEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common7JVectorIhE10addElementERKh($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = ((($4)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($4)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6|0)==($8|0);
 if ($9) {
  $10 = ((($4)) + 8|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($4)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($11) + ($13))|0;
  __ZN9ExitGames6Common7JVectorIhE14ensureCapacityEj($4,$14);
 }
 $15 = ((($4)) + 16|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($4)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (($16) + ($18)|0);
 $20 = $3;
 $21 = HEAP8[$20>>0]|0;
 HEAP8[$19>>0] = $21;
 $22 = ((($4)) + 4|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + 1)|0;
 HEAP32[$22>>2] = $24;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEEixEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $4 = $2;
 $5 = $3;
 __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE11verifyIndexEj($4,$5);
 $6 = ((($4)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $3;
 $9 = (($7) + (($8*40)|0)|0);
 STACKTOP = sp;return ($9|0);
}
function __ZN9ExitGames6Common9KeyObjectIhEC2ERKNS0_6ObjectE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 __ZN9ExitGames6Common6ObjectC2Ev($6);
 HEAP32[$6>>2] = (2768);
 $7 = $3;
 __THREW__ = 0;
 invoke_viii(777,($6|0),($7|0),98);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  $4 = $10;
  $5 = $11;
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  $12 = $4;
  $13 = $5;
  ___resumeException($12|0);
  // unreachable;
 } else {
  STACKTOP = sp;return;
 }
}
function __ZNK9ExitGames6Common9KeyObjectIhE11getDataCopyEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $2;
 $4 = (__ZNK9ExitGames6Common6Object7getTypeEv($3)|0);
 $5 = $4&255;
 $6 = ($5|0)==(98);
 if ($6) {
  $7 = (__ZNK9ExitGames6Common6Object7getDataEv($3)|0);
  $8 = HEAP8[$7>>0]|0;
  $1 = $8;
  $9 = $1;
  STACKTOP = sp;return ($9|0);
 } else {
  $1 = 0;
  $9 = $1;
  STACKTOP = sp;return ($9|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $4 = $1;
 HEAP32[$4>>2] = (3844);
 __THREW__ = 0;
 invoke_vi(778,($4|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  $7 = ((($4)) + 16|0);
  $8 = HEAP32[$7>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($8|0));
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   __ZN9ExitGames6Common4BaseD2Ev($4);
   STACKTOP = sp;return;
  }
 }
 $11 = ___cxa_find_matching_catch_3(0|0)|0;
 $12 = tempRet0;
 $2 = $11;
 $3 = $12;
 __ZN9ExitGames6Common4BaseD2Ev($4);
 $13 = $2;
 ___clang_call_terminate($13);
 // unreachable;
}
function __ZN9ExitGames6Common4BaseC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common8ToStringC2Ev($2);
 HEAP32[$2>>2] = (4224);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7JVectorIhED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common7JVectorIhED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common7JVectorIhE8toStringERNS0_7JStringEb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $6 = sp + 8|0;
 $3 = $0;
 $4 = $1;
 $9 = $2&1;
 $5 = $9;
 $10 = $3;
 $11 = ((($10)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (__ZNK9ExitGames6Common7JVectorIhE7getSizeEv($10)|0);
 __ZN9ExitGames6Common7Helpers26PrimitiveToStringConverterIhE8toStringEPhjb($6,$12,$13,0);
 $14 = $4;
 __THREW__ = 0;
 $15 = (invoke_iii(760,($14|0),($6|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $7 = $18;
  $8 = $19;
  __ZN9ExitGames6Common7JStringD2Ev($6);
  $20 = $7;
  $21 = $8;
  ___resumeException($20|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($6);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common8ToStringC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 HEAP32[$2>>2] = (6160);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26PrimitiveToStringConverterIhE8toStringEPhjb($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$expand_i1_val = 0, $$expand_i1_val2 = 0, $$pre_trunc = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $7 = sp + 20|0;
 $4 = $1;
 $5 = $2;
 $11 = $3&1;
 $6 = $11;
 $$expand_i1_val = 0;
 HEAP8[$7>>0] = $$expand_i1_val;
 __ZN9ExitGames6Common7JStringC2EPKw($0,12344);
 $8 = 0;
 while(1) {
  $12 = $8;
  $13 = $5;
  $14 = ($12>>>0)<($13>>>0);
  if (!($14)) {
   label = 8;
   break;
  }
  $15 = $4;
  $16 = $8;
  $17 = (($15) + ($16)|0);
  __THREW__ = 0;
  (invoke_iii(779,($0|0),($17|0))|0);
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if ($19) {
   break;
  }
  $20 = $8;
  $21 = $5;
  $22 = (($21) - 1)|0;
  $23 = ($20>>>0)<($22>>>0);
  if ($23) {
   __THREW__ = 0;
   (invoke_iii(780,($0|0),(12352|0))|0);
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    break;
   }
  }
  $30 = $8;
  $31 = (($30) + 1)|0;
  $8 = $31;
 }
 if ((label|0) == 8) {
  __THREW__ = 0;
  (invoke_iii(781,($0|0),(12364|0))|0);
  $32 = __THREW__; __THREW__ = 0;
  $33 = $32&1;
  if (!($33)) {
   $$expand_i1_val2 = 1;
   HEAP8[$7>>0] = $$expand_i1_val2;
   $$pre_trunc = HEAP8[$7>>0]|0;
   $34 = $$pre_trunc&1;
   if ($34) {
    STACKTOP = sp;return;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
   STACKTOP = sp;return;
  }
 }
 $26 = ___cxa_find_matching_catch_2()|0;
 $27 = tempRet0;
 $9 = $26;
 $10 = $27;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $28 = $9;
 $29 = $10;
 ___resumeException($28|0);
 // unreachable;
}
function __ZNK9ExitGames6Common7JVectorIhE7getSizeEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common7JStringpLIhEERS1_RKT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $4 = sp + 8|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 __ZN9ExitGames6Common7JStringC2Ej($4,0);
 $8 = $3;
 $9 = HEAP8[$8>>0]|0;
 __THREW__ = 0;
 (invoke_iii(782,($4|0),($9|0))|0);
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if (!($11)) {
  __THREW__ = 0;
  $12 = (invoke_iii(760,($7|0),($4|0))|0);
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if (!($14)) {
   __ZN9ExitGames6Common7JStringD2Ev($4);
   STACKTOP = sp;return ($12|0);
  }
 }
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 $5 = $15;
 $6 = $16;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 $17 = $5;
 $18 = $6;
 ___resumeException($17|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEEaSERKS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $5 = $2;
 $6 = ((($5)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7>>>0)>(0);
 if ($8) {
  label = 3;
 } else {
  $9 = ((($5)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = $3;
  $12 = ((($11)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($10>>>0)<($13>>>0);
  if ($14) {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  __ZN9ExitGames6Common7JVectorINS0_6ObjectEE17removeAllElementsEv($5);
  $15 = ((($5)) + 16|0);
  $16 = HEAP32[$15>>2]|0;
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($16);
  $17 = $3;
  $18 = ((($17)) + 8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($5)) + 8|0);
  HEAP32[$20>>2] = $19;
  $21 = ((($5)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22*40)|0;
  $24 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($23)|0);
  $25 = ((($5)) + 16|0);
  HEAP32[$25>>2] = $24;
 }
 $26 = $3;
 $27 = ((($26)) + 4|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ((($5)) + 4|0);
 HEAP32[$29>>2] = $28;
 $30 = $3;
 $31 = ((($30)) + 12|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ((($5)) + 12|0);
 HEAP32[$33>>2] = $32;
 $4 = 0;
 while(1) {
  $34 = $4;
  $35 = ((($5)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($34>>>0)<($36>>>0);
  if (!($37)) {
   break;
  }
  $38 = ((($5)) + 16|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = $4;
  $41 = (($39) + (($40*40)|0)|0);
  $42 = $3;
  $43 = ((($42)) + 16|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = $4;
  $46 = (($44) + (($45*40)|0)|0);
  __ZN9ExitGames6Common6ObjectC2ERKS1_($41,$46);
  $47 = $4;
  $48 = (($47) + 1)|0;
  $4 = $48;
 }
 STACKTOP = sp;return ($5|0);
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $2 = $1;
 __ZN9ExitGames6Common7JVectorINS0_6ObjectEED2Ev($2);
 __ZdlPv($2);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common7JVectorINS0_6ObjectEE8toStringERNS0_7JStringEb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $6 = sp + 8|0;
 $3 = $0;
 $4 = $1;
 $9 = $2&1;
 $5 = $9;
 $10 = $3;
 $11 = ((($10)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (__ZNK9ExitGames6Common7JVectorINS0_6ObjectEE7getSizeEv($10)|0);
 __ZN9ExitGames6Common7Helpers23ObjectToStringConverterINS0_6ObjectEE8toStringEPS3_jb($6,$12,$13,0);
 $14 = $4;
 __THREW__ = 0;
 $15 = (invoke_iii(760,($14|0),($6|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  $7 = $18;
  $8 = $19;
  __ZN9ExitGames6Common7JStringD2Ev($6);
  $20 = $7;
  $21 = $8;
  ___resumeException($20|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($6);
  STACKTOP = sp;return ($15|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEE17removeAllElementsEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = $1;
 $2 = 0;
 while(1) {
  $4 = $2;
  $5 = ((($3)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($4>>>0)<($6>>>0);
  if (!($7)) {
   break;
  }
  $8 = ((($3)) + 16|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $2;
  $11 = (($9) + (($10*40)|0)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$12>>2]|0;
  FUNCTION_TABLE_vi[$13 & 2047]($11);
  $14 = $2;
  $15 = (($14) + 1)|0;
  $2 = $15;
 }
 $16 = ((($3)) + 4|0);
 HEAP32[$16>>2] = 0;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers23ObjectToStringConverterINS0_6ObjectEE8toStringEPS3_jb($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$expand_i1_val = 0, $$expand_i1_val2 = 0, $$pre_trunc = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $7 = sp + 36|0;
 $9 = sp + 8|0;
 $4 = $1;
 $5 = $2;
 $12 = $3&1;
 $6 = $12;
 $$expand_i1_val = 0;
 HEAP8[$7>>0] = $$expand_i1_val;
 __ZN9ExitGames6Common7JStringC2EPKw($0,12344);
 $8 = 0;
 while(1) {
  $13 = $8;
  $14 = $5;
  $15 = ($13>>>0)<($14>>>0);
  if (!($15)) {
   label = 10;
   break;
  }
  $16 = $4;
  $17 = $8;
  $18 = (($16) + (($17*40)|0)|0);
  $19 = $6;
  $20 = $19&1;
  __THREW__ = 0;
  invoke_viii(783,($9|0),($18|0),($20|0));
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   break;
  }
  __THREW__ = 0;
  (invoke_iii(760,($0|0),($9|0))|0);
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if ($24) {
   label = 8;
   break;
  }
  __ZN9ExitGames6Common7JStringD2Ev($9);
  $25 = $8;
  $26 = $5;
  $27 = (($26) - 1)|0;
  $28 = ($25>>>0)<($27>>>0);
  if ($28) {
   __THREW__ = 0;
   (invoke_iii(780,($0|0),(12352|0))|0);
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    break;
   }
  }
  $35 = $8;
  $36 = (($35) + 1)|0;
  $8 = $36;
 }
 if ((label|0) == 8) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $10 = $33;
  $11 = $34;
  __ZN9ExitGames6Common7JStringD2Ev($9);
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $40 = $10;
  $41 = $11;
  ___resumeException($40|0);
  // unreachable;
 }
 if ((label|0) == 10) {
  __THREW__ = 0;
  (invoke_iii(781,($0|0),(12364|0))|0);
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if (!($38)) {
   $$expand_i1_val2 = 1;
   HEAP8[$7>>0] = $$expand_i1_val2;
   $$pre_trunc = HEAP8[$7>>0]|0;
   $39 = $$pre_trunc&1;
   if ($39) {
    STACKTOP = sp;return;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
   STACKTOP = sp;return;
  }
 }
 $31 = ___cxa_find_matching_catch_2()|0;
 $32 = tempRet0;
 $10 = $31;
 $11 = $32;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $40 = $10;
 $41 = $11;
 ___resumeException($40|0);
 // unreachable;
}
function __ZN9ExitGames6Common7JVectorIhE14ensureCapacityEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = $0;
 $3 = $1;
 $6 = $2;
 $7 = $3;
 $8 = ((($6)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($7>>>0)>($9>>>0);
 if (!($10)) {
  STACKTOP = sp;return;
 }
 $11 = $3;
 $12 = ((($6)) + 8|0);
 HEAP32[$12>>2] = $11;
 $13 = $11;
 $14 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($13)|0);
 $4 = $14;
 $5 = 0;
 while(1) {
  $15 = $5;
  $16 = ((($6)) + 4|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($15>>>0)<($17>>>0);
  if (!($18)) {
   break;
  }
  $19 = $4;
  $20 = $5;
  $21 = (($19) + ($20)|0);
  $22 = ((($6)) + 16|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $5;
  $25 = (($23) + ($24)|0);
  $26 = HEAP8[$25>>0]|0;
  HEAP8[$21>>0] = $26;
  $27 = $5;
  $28 = (($27) + 1)|0;
  $5 = $28;
 }
 $29 = ((($6)) + 16|0);
 $30 = HEAP32[$29>>2]|0;
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($30);
 $31 = $4;
 $32 = ((($6)) + 16|0);
 HEAP32[$32>>2] = $31;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common9KeyObjectIhE7convertEPKNS0_6ObjectEh($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $6 = sp;
 $3 = $0;
 $4 = $1;
 $5 = $2;
 $9 = $3;
 $10 = $4;
 $11 = ($10|0)!=(0|0);
 if ($11) {
  $12 = $5;
  $13 = $12&255;
  $14 = $4;
  $15 = (__ZNK9ExitGames6Common6Object7getTypeEv($14)|0);
  $16 = $15&255;
  $17 = ($13|0)==($16|0);
  if ($17) {
   $18 = $4;
   __ZN9ExitGames6Common6ObjectC2ERKS1_($6,$18);
  } else {
   label = 4;
  }
 } else {
  label = 4;
 }
 if ((label|0) == 4) {
  __ZN9ExitGames6Common6ObjectC2Ev($6);
 }
 __THREW__ = 0;
 (invoke_iii(239,($9|0),($6|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $7 = $21;
  $8 = $22;
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  $23 = $7;
  $24 = $8;
  ___resumeException($23|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common6ObjectD2Ev($6);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common7JVectorIhE17removeAllElementsEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = $0;
 $3 = $1;
 $2 = 0;
 while(1) {
  $4 = $2;
  $5 = ((($3)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($4>>>0)<($6>>>0);
  if (!($7)) {
   break;
  }
  $8 = $2;
  $9 = (($8) + 1)|0;
  $2 = $9;
 }
 $10 = ((($3)) + 4|0);
 HEAP32[$10>>2] = 0;
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common7JVectorIhE11verifyIndexEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer = sp;
 $4 = sp + 16|0;
 $2 = $0;
 $3 = $1;
 $7 = $2;
 $8 = $3;
 $9 = ((($7)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>=($10>>>0);
 if (!($11)) {
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKc($4,147471);
 __THREW__ = 0;
 $12 = (invoke_ii(764,($4|0))|0);
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,(162472|0),1,(3648|0),($12|0),1,566,(10968|0),($vararg_buffer|0));
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if (!($16)) {
   __ZN9ExitGames6Common7JStringD2Ev($4);
   STACKTOP = sp;return;
  }
 }
 $17 = ___cxa_find_matching_catch_2()|0;
 $18 = tempRet0;
 $5 = $17;
 $6 = $18;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 $19 = $5;
 $20 = $6;
 ___resumeException($19|0);
 // unreachable;
}
function __GLOBAL__sub_I_PhotonPeer_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 ___cxx_global_var_init();
 return;
}
function __ZN9ExitGames6Common10ANSIStringD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3868);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common10ANSIStringD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3868);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  __ZdlPv($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common8ToString12typeToStringEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $eh$lpad$body5$index2Z2D = 0, $eh$lpad$body5$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $str$i = sp + 64|0;
 $0 = sp + 48|0;
 $1 = sp + 32|0;
 $2 = sp + 16|0;
 $3 = sp;
 __ZN9ExitGames6Common7JStringC2EPKw($2,13660);
 $4 = HEAP32[$this>>2]|0;
 $5 = ((($4)) + -4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($6)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 __THREW__ = 0;
 invoke_vii(784,($3|0),($8|0));
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  $$1 = $29;$$12 = $28;
  __ZN9ExitGames6Common7JStringD2Ev($2);
  ___resumeException($$12|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viii(785,($1|0),($2|0),($3|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  $$0 = $31;$$01 = $30;
 } else {
  __THREW__ = 0;
  invoke_vii(786,($0|0),($1|0));
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $32 = ___cxa_find_matching_catch_2()|0;
   $33 = tempRet0;
   $eh$lpad$body$index7Z2D = $33;$eh$lpad$body$indexZ2D = $32;
  } else {
   __THREW__ = 0;
   invoke_vii(787,($str$i|0),0);
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   do {
    if ($16) {
     label = 9;
    } else {
     __THREW__ = 0;
     (invoke_iii(759,($str$i|0),(14336|0))|0);
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if (!($18)) {
      __THREW__ = 0;
      $19 = (invoke_iii(760,($0|0),($str$i|0))|0);
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if (!($21)) {
       __ZN9ExitGames6Common7JStringD2Ev($str$i);
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($19|0));
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        label = 9;
        break;
       }
       __ZN9ExitGames6Common7JStringD2Ev($0);
       __ZN9ExitGames6Common7JStringD2Ev($1);
       __ZN9ExitGames6Common7JStringD2Ev($3);
       __ZN9ExitGames6Common7JStringD2Ev($2);
       STACKTOP = sp;return;
      }
     }
     $22 = ___cxa_find_matching_catch_2()|0;
     $23 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     $eh$lpad$body5$index2Z2D = $23;$eh$lpad$body5$indexZ2D = $22;
    }
   } while(0);
   if ((label|0) == 9) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    $eh$lpad$body5$index2Z2D = $27;$eh$lpad$body5$indexZ2D = $26;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $eh$lpad$body$index7Z2D = $eh$lpad$body5$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body5$indexZ2D;
  }
  __ZN9ExitGames6Common7JStringD2Ev($1);
  $$0 = $eh$lpad$body$index7Z2D;$$01 = $eh$lpad$body$indexZ2D;
 }
 __ZN9ExitGames6Common7JStringD2Ev($3);
 $$1 = $$0;$$12 = $$01;
 __ZN9ExitGames6Common7JStringD2Ev($2);
 ___resumeException($$12|0);
 // unreachable;
}
function __ZNK9ExitGames6Common14BaseCharString8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = HEAP32[$this>>2]|0;
 $2 = ((($1)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_vii[$3 & 2047]($0,$this);
 __THREW__ = 0;
 $4 = (invoke_iii(760,($retStr|0),($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  STACKTOP = sp;return ($4|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common10ANSIStringcvPKcEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common10ANSIStringcvNS0_7JStringEEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 FUNCTION_TABLE_vii[$2 & 2047]($agg$result,$this);
 return;
}
function __ZNK9ExitGames6Common10ANSIString21JStringRepresentationEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common14BaseCharString6lengthEv($this)|0);
 $1 = (($0) + 1)|0;
 $2 = $1 << 2;
 $3 = (($2) + 8)|0;
 $4 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($3)|0);
 $5 = ((($4)) + 8|0);
 $6 = $4;
 $7 = $6;
 HEAP32[$7>>2] = $1;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP32[$9>>2] = 0;
 $10 = ($1|0)==(0);
 if (!($10)) {
  _memset(($5|0),0,($2|0))|0;
 }
 $11 = (__ZNK9ExitGames6Common14BaseCharString4cstrEv($this)|0);
 $12 = (__ZNK9ExitGames6Common14BaseCharString6lengthEv($this)|0);
 $13 = (($12) + 1)|0;
 (__ZN9ExitGames6Common10ANSIString13ANSIConverter12ANSI2UnicodeEPKcPwj($11,$5,$13)|0);
 __ZN9ExitGames6Common7JStringC2EPKw($agg$result,$5);
 __THREW__ = 0;
 invoke_vi(750,($4|0));
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($agg$result);
  ___resumeException($16|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNK9ExitGames6Common10ANSIString4sizeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common14BaseCharString6lengthEv($this)|0);
 return ($0|0);
}
function __ZNK9ExitGames6Common14BaseCharString6lengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($size) {
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($size) + 8)|0;
 $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface12getAllocatorEj($0)|0);
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface12getAllocatorEj($0)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (FUNCTION_TABLE_iii[$5 & 2047]($2,$0)|0);
 HEAP32[$6>>2] = $size;
 $7 = ((($6)) + 4|0);
 HEAP32[$7>>2] = $1;
 $8 = ((($6)) + 8|0);
 return ($8|0);
}
function __ZNK9ExitGames6Common14BaseCharString4cstrEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Common10ANSIString13ANSIConverter12ANSI2UnicodeEPKcPwj($str,$wstr,$wstrLen) {
 $str = $str|0;
 $wstr = $wstr|0;
 $wstrLen = $wstrLen|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $len = 0, $size = 0, $wrptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp + 12|0;
 $wrptr = sp + 8|0;
 $size = sp + 4|0;
 $len = sp;
 HEAP32[$0>>2] = $str;
 $1 = HEAP32[983]|0;
 $2 = HEAP32[984]|0;
 $3 = (_iconv_open($1,$2)|0);
 HEAP32[$wrptr>>2] = $wstr;
 $4 = $wstrLen << 2;
 HEAP32[$size>>2] = $4;
 $5 = (_strlen($str)|0);
 HEAP32[$len>>2] = $5;
 (_iconv($3,$0,$len,$wrptr,$size)|0);
 $6 = HEAP32[$wrptr>>2]|0;
 HEAP32[$6>>2] = 0;
 STACKTOP = sp;return ($wstr|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($p) {
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($p|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($p)) + -4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($p)) + -8|0);
 FUNCTION_TABLE_vii[$5 & 2047]($2,$6);
 return;
}
function __ZN9ExitGames6Common7JStringD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3908);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common8ToStringD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common7JStringD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3908);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZdlPv($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common7JString8toStringERS1_b($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $0 = sp + 16|0;
 $1 = sp;
 __ZN9ExitGames6CommonplIA2_wEENS0_7JStringERKT_RKS3_($1,3924,$this);
 __THREW__ = 0;
 invoke_viii(788,($0|0),($1|0),(3924|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $52 = ___cxa_find_matching_catch_2()|0;
  $53 = tempRet0;
  $$0 = $53;$$01 = $52;
 } else {
  $4 = ((($0)) + 12|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($retStr)) + 12|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (($7) + ($5))|0;
  HEAP32[$6>>2] = $8;
  $9 = ((($retStr)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($8>>>0)>($10>>>0);
  $12 = ((($retStr)) + 4|0);
  if ($11) {
   $13 = HEAP32[$12>>2]|0;
   HEAP32[$9>>2] = $8;
   $14 = (($8) + 1)|0;
   $15 = $14 << 2;
   $16 = (($15) + 8)|0;
   __THREW__ = 0;
   $17 = (invoke_ii(775,($16|0))|0);
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    $20 = ((($17)) + 8|0);
    $21 = $17;
    $22 = $21;
    HEAP32[$22>>2] = $14;
    $23 = (($21) + 4)|0;
    $24 = $23;
    HEAP32[$24>>2] = 0;
    $25 = ($14|0)==(0);
    if (!($25)) {
     _memset(($20|0),0,($15|0))|0;
    }
    HEAP32[$12>>2] = $20;
    __THREW__ = 0;
    (invoke_iii(789,($20|0),($13|0))|0);
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if (!($27)) {
     $28 = ($13|0)==(0|0);
     if ($28) {
      label = 9;
     } else {
      $29 = ((($13)) + -8|0);
      __THREW__ = 0;
      invoke_vi(750,($29|0));
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if (!($31)) {
       label = 9;
      }
     }
    }
   }
  } else {
   label = 9;
  }
  if ((label|0) == 9) {
   $32 = HEAP32[$12>>2]|0;
   $33 = ((($0)) + 4|0);
   $34 = HEAP32[$33>>2]|0;
   __THREW__ = 0;
   (invoke_iii(790,($32|0),($34|0))|0);
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if (!($36)) {
    HEAP32[$0>>2] = (3908);
    $37 = HEAP32[$33>>2]|0;
    $38 = ($37|0)==(0|0);
    if (!($38)) {
     $39 = ((($37)) + -8|0);
     __THREW__ = 0;
     invoke_vi(750,($39|0));
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if ($41) {
      $42 = ___cxa_find_matching_catch_3(0|0)|0;
      $43 = tempRet0;
      ___clang_call_terminate($42);
      // unreachable;
     }
    }
    HEAP32[$1>>2] = (3908);
    $44 = ((($1)) + 4|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = ($45|0)==(0|0);
    if ($46) {
     STACKTOP = sp;return ($retStr|0);
    }
    $47 = ((($45)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($47|0));
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $50 = ___cxa_find_matching_catch_3(0|0)|0;
     $51 = tempRet0;
     ___clang_call_terminate($50);
     // unreachable;
    } else {
     STACKTOP = sp;return ($retStr|0);
    }
   }
  }
  $54 = ___cxa_find_matching_catch_2()|0;
  $55 = tempRet0;
  HEAP32[$0>>2] = (3908);
  $56 = ((($0)) + 4|0);
  $57 = HEAP32[$56>>2]|0;
  $58 = ($57|0)==(0|0);
  if (!($58)) {
   $59 = ((($57)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($59|0));
   $60 = __THREW__; __THREW__ = 0;
   $61 = $60&1;
   if ($61) {
    $62 = ___cxa_find_matching_catch_3(0|0)|0;
    $63 = tempRet0;
    ___clang_call_terminate($62);
    // unreachable;
   }
  }
  $$0 = $55;$$01 = $54;
 }
 HEAP32[$1>>2] = (3908);
 $64 = ((($1)) + 4|0);
 $65 = HEAP32[$64>>2]|0;
 $66 = ($65|0)==(0|0);
 if ($66) {
  ___resumeException($$01|0);
  // unreachable;
 }
 $67 = ((($65)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($67|0));
 $68 = __THREW__; __THREW__ = 0;
 $69 = $68&1;
 if ($69) {
  $70 = ___cxa_find_matching_catch_3(0|0)|0;
  $71 = tempRet0;
  ___clang_call_terminate($70);
  // unreachable;
 } else {
  ___resumeException($$01|0);
  // unreachable;
 }
 return (0)|0;
}
function __ZN9ExitGames6CommonplIA2_wEENS0_7JStringERKT_RKS3_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($0,0);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($Lsh|0))|0);
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     $10 = (invoke_iii(760,($5|0),($Rsh|0))|0);
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      break;
     }
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($10|0));
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $15 = ___cxa_find_matching_catch_2()|0;
 $16 = tempRet0;
 $eh$lpad$body$index2Z2D = $16;$eh$lpad$body$indexZ2D = $15;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplIA2_wEENS0_7JStringERKS3_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($Rsh|0))|0);
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($5|0));
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 $eh$lpad$body$index2Z2D = $13;$eh$lpad$body$indexZ2D = $12;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Common7JStringaSEPKw($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = ($Rhs|0)==(0|0);
 if (!($1)) {
  $65 = ((($this)) + 4|0);
  $66 = HEAP32[$65>>2]|0;
  $67 = ($66|0)==(0|0);
  if (!($67)) {
   $68 = ((($66)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($68);
  }
  $69 = (_wcslen($Rhs)|0);
  $70 = ((($this)) + 12|0);
  HEAP32[$70>>2] = $69;
  $71 = ((($this)) + 8|0);
  HEAP32[$71>>2] = $69;
  $72 = (($69) + 1)|0;
  $73 = $72 << 2;
  $74 = (($73) + 8)|0;
  $75 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($74)|0);
  $76 = ((($75)) + 8|0);
  $77 = $75;
  $78 = $77;
  HEAP32[$78>>2] = $72;
  $79 = (($77) + 4)|0;
  $80 = $79;
  HEAP32[$80>>2] = 0;
  $81 = ($72|0)==(0);
  if (!($81)) {
   _memset(($76|0),0,($73|0))|0;
  }
  HEAP32[$65>>2] = $76;
  (_wcscpy($76,$Rhs)|0);
  STACKTOP = sp;return ($this|0);
 }
 HEAP32[$0>>2] = (3908);
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 __THREW__ = 0;
 $5 = (invoke_ii(756,(163772|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  ___resumeException($22|0);
  // unreachable;
 }
 HEAP32[$4>>2] = $5;
 HEAP32[$3>>2] = $5;
 $8 = (($5) + 1)|0;
 $9 = $8 << 2;
 $10 = (($9) + 8)|0;
 __THREW__ = 0;
 $11 = (invoke_ii(775,($10|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  ___resumeException($22|0);
  // unreachable;
 }
 $14 = ((($11)) + 8|0);
 $15 = $11;
 $16 = $15;
 HEAP32[$16>>2] = $8;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = 0;
 $19 = ($8|0)==(0);
 if (!($19)) {
  _memset(($14|0),0,($9|0))|0;
 }
 HEAP32[$2>>2] = $14;
 __THREW__ = 0;
 (invoke_iii(789,($14|0),(163772|0))|0);
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  ___resumeException($22|0);
  // unreachable;
 }
 $24 = ($0|0)==($this|0);
 do {
  if (!($24)) {
   $25 = ((($this)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = ($26|0)==(0|0);
   if ($27) {
    label = 11;
   } else {
    $28 = ((($26)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($28|0));
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if (!($30)) {
     label = 11;
    }
   }
   if ((label|0) == 11) {
    $31 = HEAP32[$4>>2]|0;
    $32 = ((($this)) + 12|0);
    HEAP32[$32>>2] = $31;
    $33 = ((($this)) + 8|0);
    HEAP32[$33>>2] = $31;
    $34 = (($31) + 1)|0;
    $35 = $34 << 2;
    $36 = (($35) + 8)|0;
    __THREW__ = 0;
    $37 = (invoke_ii(775,($36|0))|0);
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     $40 = ((($37)) + 8|0);
     $41 = $37;
     $42 = $41;
     HEAP32[$42>>2] = $34;
     $43 = (($41) + 4)|0;
     $44 = $43;
     HEAP32[$44>>2] = 0;
     $45 = ($34|0)==(0);
     if (!($45)) {
      _memset(($40|0),0,($35|0))|0;
     }
     HEAP32[$25>>2] = $40;
     $46 = HEAP32[$2>>2]|0;
     __THREW__ = 0;
     (invoke_iii(789,($40|0),($46|0))|0);
     $47 = __THREW__; __THREW__ = 0;
     $48 = $47&1;
     if (!($48)) {
      break;
     }
    }
   }
   $56 = ___cxa_find_matching_catch_2()|0;
   $57 = tempRet0;
   HEAP32[$0>>2] = (3908);
   $58 = HEAP32[$2>>2]|0;
   $59 = ($58|0)==(0|0);
   if ($59) {
    ___resumeException($56|0);
    // unreachable;
   }
   $60 = ((($58)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($60|0));
   $61 = __THREW__; __THREW__ = 0;
   $62 = $61&1;
   if ($62) {
    $63 = ___cxa_find_matching_catch_3(0|0)|0;
    $64 = tempRet0;
    ___clang_call_terminate($63);
    // unreachable;
   } else {
    ___resumeException($56|0);
    // unreachable;
   }
  }
 } while(0);
 HEAP32[$0>>2] = (3908);
 $49 = HEAP32[$2>>2]|0;
 $50 = ($49|0)==(0|0);
 if (!($50)) {
  $51 = ((($49)) + -8|0);
  __THREW__ = 0;
  invoke_vi(750,($51|0));
  $52 = __THREW__; __THREW__ = 0;
  $53 = $52&1;
  if ($53) {
   $54 = ___cxa_find_matching_catch_3(0|0)|0;
   $55 = tempRet0;
   ___clang_call_terminate($54);
   // unreachable;
  }
 }
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JStringpLERKS1_($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($Rhs)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($3) + ($1))|0;
 HEAP32[$2>>2] = $4;
 $5 = ((($this)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($4>>>0)>($6>>>0);
 $8 = ((($this)) + 4|0);
 if ($7) {
  $9 = HEAP32[$8>>2]|0;
  HEAP32[$5>>2] = $4;
  $10 = (($4) + 1)|0;
  $11 = $10 << 2;
  $12 = (($11) + 8)|0;
  $13 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($12)|0);
  $14 = ((($13)) + 8|0);
  $15 = $13;
  $16 = $15;
  HEAP32[$16>>2] = $10;
  $17 = (($15) + 4)|0;
  $18 = $17;
  HEAP32[$18>>2] = 0;
  $19 = ($10|0)==(0);
  if (!($19)) {
   _memset(($14|0),0,($11|0))|0;
  }
  HEAP32[$8>>2] = $14;
  (_wcscpy($14,$9)|0);
  $20 = ($9|0)==(0|0);
  if (!($20)) {
   $21 = ((($9)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($21);
  }
 }
 $22 = HEAP32[$8>>2]|0;
 $23 = ((($Rhs)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 (_wcscat($22,$24)|0);
 return ($this|0);
}
function __ZN9ExitGames6Common7JStringC2Ej($this,$bufferlen) {
 $this = $this|0;
 $bufferlen = $bufferlen|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $bufferlen$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3908);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = 0;
 __THREW__ = 0;
 $3 = (invoke_ii(756,(163772|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  HEAP32[$2>>2] = $3;
  $6 = ($3>>>0)<($bufferlen>>>0);
  $bufferlen$ = $6 ? $bufferlen : $3;
  HEAP32[$1>>2] = $bufferlen$;
  $7 = (($bufferlen$) + 1)|0;
  $8 = $7 << 2;
  $9 = (($8) + 8)|0;
  __THREW__ = 0;
  $10 = (invoke_ii(775,($9|0))|0);
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if (!($12)) {
   $13 = ((($10)) + 8|0);
   $14 = $10;
   $15 = $14;
   HEAP32[$15>>2] = $7;
   $16 = (($14) + 4)|0;
   $17 = $16;
   HEAP32[$17>>2] = 0;
   $18 = ($7|0)==(0);
   if (!($18)) {
    _memset(($13|0),0,($8|0))|0;
   }
   HEAP32[$0>>2] = $13;
   __THREW__ = 0;
   (invoke_iii(789,($13|0),(163772|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    return;
   }
  }
 }
 $21 = ___cxa_find_matching_catch_2()|0;
 $22 = tempRet0;
 ___resumeException($21|0);
 // unreachable;
}
function __ZN9ExitGames6Common7JStringC2ERKS1_($this,$Value) {
 $this = $this|0;
 $Value = $Value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3908);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($Value)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 HEAP32[$2>>2] = $4;
 HEAP32[$1>>2] = $4;
 $5 = (($4) + 1)|0;
 $6 = $5 << 2;
 $7 = (($6) + 8)|0;
 __THREW__ = 0;
 $8 = (invoke_ii(775,($7|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___resumeException($21|0);
  // unreachable;
 }
 $11 = ((($8)) + 8|0);
 $12 = $8;
 $13 = $12;
 HEAP32[$13>>2] = $5;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 $16 = ($5|0)==(0);
 if (!($16)) {
  _memset(($11|0),0,($6|0))|0;
 }
 HEAP32[$0>>2] = $11;
 $17 = ((($Value)) + 4|0);
 $18 = HEAP32[$17>>2]|0;
 __THREW__ = 0;
 (invoke_iii(789,($11|0),($18|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___resumeException($21|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common7JStringC2EPKw($this,$Value) {
 $this = $this|0;
 $Value = $Value|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3908);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = 0;
 $3 = ($Value|0)!=(0|0);
 $4 = $3 ? $Value : 163772;
 __THREW__ = 0;
 $5 = (invoke_ii(756,($4|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  HEAP32[$2>>2] = $5;
  HEAP32[$1>>2] = $5;
  $8 = (($5) + 1)|0;
  $9 = $8 << 2;
  $10 = (($9) + 8)|0;
  __THREW__ = 0;
  $11 = (invoke_ii(775,($10|0))|0);
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   $14 = ((($11)) + 8|0);
   $15 = $11;
   $16 = $15;
   HEAP32[$16>>2] = $8;
   $17 = (($15) + 4)|0;
   $18 = $17;
   HEAP32[$18>>2] = 0;
   $19 = ($8|0)==(0);
   if (!($19)) {
    _memset(($14|0),0,($9|0))|0;
   }
   HEAP32[$0>>2] = $14;
   __THREW__ = 0;
   (invoke_iii(789,($14|0),($4|0))|0);
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if (!($21)) {
    return;
   }
  }
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 ___resumeException($22|0);
 // unreachable;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal9Interface12getAllocatorEj($size) {
 $size = $size|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[160960]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire(160960)|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_i(791)|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $15 = ___cxa_find_matching_catch_2()|0;
     $16 = tempRet0;
     ___resumeException($15|0);
     // unreachable;
    } else {
     HEAP32[40420] = $4;
     HEAP32[40421] = $4;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP8[160968]|0;
 $8 = ($7<<24>>24)==(0);
 if ($8) {
  $9 = (___cxa_guard_acquire(160968)|0);
  $10 = ($9|0)==(0);
  if (!($10)) {
   HEAP32[40422] = (3948);
   (___cxa_atexit((214|0),(161688|0),(___dso_handle|0))|0);
  }
 }
 $11 = HEAP32[993]|0;
 $12 = ($11>>>0)<($size>>>0);
 $13 = HEAP32[40420]|0;
 $14 = $12 ? 161688 : $13;
 return ($14|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager3getEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[160976]|0;
 $1 = ($0<<24>>24)==(0);
 if (!($1)) {
  return (161692|0);
 }
 $2 = (___cxa_guard_acquire(160976)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  return (161692|0);
 }
 __THREW__ = 0;
 invoke_vi(792,(161692|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  ___resumeException($6|0);
  // unreachable;
 }
 (___cxa_atexit((220|0),(161692|0),(___dso_handle|0))|0);
 return (161692|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocatorD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocatorD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator15setMaxAllocSizeEj($this,$maxAllocSize) {
 $this = $this|0;
 $maxAllocSize = $maxAllocSize|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator5allocEj($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znaj($size)|0);
 return ($0|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator7deallocEPv($this,$p) {
 $this = $this|0;
 $p = $p|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($p|0)==(0|0);
 if ($0) {
  return;
 }
 __ZdaPv($p);
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator6resizeEPvj($this,$p,$size) {
 $this = $this|0;
 $p = $p|0;
 $size = $size|0;
 var $$size = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Znaj($size)|0);
 $1 = HEAP32[$p>>2]|0;
 $2 = ($1>>>0)<($size>>>0);
 $$size = $2 ? $1 : $size;
 _memcpy(($0|0),($p|0),($$size|0))|0;
 $3 = ($p|0)==(0|0);
 if ($3) {
  return ($0|0);
 }
 __ZdaPv($p);
 return ($0|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $indvars$iv = 0, $indvars$iv$next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3984);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vi(793,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if ($2) {
   $$lcssa = $0;
  } else {
   $3 = ((($this)) + 28|0);
   __THREW__ = 0;
   invoke_vi(793,($3|0));
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if ($5) {
    $$lcssa = $3;
   } else {
    $80 = ((($this)) + 52|0);
    __THREW__ = 0;
    invoke_vi(793,($80|0));
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     $$lcssa = $80;
    } else {
     $83 = ((($this)) + 76|0);
     __THREW__ = 0;
     invoke_vi(793,($83|0));
     $84 = __THREW__; __THREW__ = 0;
     $85 = $84&1;
     if ($85) {
      $$lcssa = $83;
     } else {
      $86 = ((($this)) + 100|0);
      __THREW__ = 0;
      invoke_vi(793,($86|0));
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if ($88) {
       $$lcssa = $86;
      } else {
       $89 = ((($this)) + 124|0);
       __THREW__ = 0;
       invoke_vi(793,($89|0));
       $90 = __THREW__; __THREW__ = 0;
       $91 = $90&1;
       if ($91) {
        $$lcssa = $89;
       } else {
        $92 = ((($this)) + 148|0);
        __THREW__ = 0;
        invoke_vi(793,($92|0));
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         $$lcssa = $92;
        } else {
         $95 = ((($this)) + 172|0);
         __THREW__ = 0;
         invoke_vi(793,($95|0));
         $96 = __THREW__; __THREW__ = 0;
         $97 = $96&1;
         if ($97) {
          $$lcssa = $95;
         } else {
          $98 = ((($this)) + 196|0);
          __THREW__ = 0;
          invoke_vi(793,($98|0));
          $99 = __THREW__; __THREW__ = 0;
          $100 = $99&1;
          if ($100) {
           $$lcssa = $98;
          } else {
           $101 = ((($this)) + 220|0);
           __THREW__ = 0;
           invoke_vi(793,($101|0));
           $102 = __THREW__; __THREW__ = 0;
           $103 = $102&1;
           if ($103) {
            $$lcssa = $101;
           } else {
            $104 = ((($this)) + 244|0);
            __THREW__ = 0;
            invoke_vi(793,($104|0));
            $105 = __THREW__; __THREW__ = 0;
            $106 = $105&1;
            if ($106) {
             $$lcssa = $104;
            } else {
             $107 = ((($this)) + 268|0);
             __THREW__ = 0;
             invoke_vi(793,($107|0));
             $108 = __THREW__; __THREW__ = 0;
             $109 = $108&1;
             if ($109) {
              $$lcssa = $107;
             } else {
              $110 = ((($this)) + 292|0);
              __THREW__ = 0;
              invoke_vi(793,($110|0));
              $111 = __THREW__; __THREW__ = 0;
              $112 = $111&1;
              if ($112) {
               $$lcssa = $110;
              } else {
               $113 = ((($this)) + 316|0);
               __THREW__ = 0;
               invoke_vi(793,($113|0));
               $114 = __THREW__; __THREW__ = 0;
               $115 = $114&1;
               if ($115) {
                $$lcssa = $113;
               } else {
                $116 = ((($this)) + 340|0);
                __THREW__ = 0;
                invoke_vi(793,($116|0));
                $117 = __THREW__; __THREW__ = 0;
                $118 = $117&1;
                if ($118) {
                 $$lcssa = $116;
                } else {
                 $119 = ((($this)) + 364|0);
                 __THREW__ = 0;
                 invoke_vi(793,($119|0));
                 $120 = __THREW__; __THREW__ = 0;
                 $121 = $120&1;
                 if ($121) {
                  $$lcssa = $119;
                 } else {
                  $122 = ((($this)) + 388|0);
                  __THREW__ = 0;
                  invoke_vi(793,($122|0));
                  $123 = __THREW__; __THREW__ = 0;
                  $124 = $123&1;
                  if ($124) {
                   $$lcssa = $122;
                  } else {
                   $125 = ((($this)) + 412|0);
                   __THREW__ = 0;
                   invoke_vi(793,($125|0));
                   $126 = __THREW__; __THREW__ = 0;
                   $127 = $126&1;
                   if ($127) {
                    $$lcssa = $125;
                   } else {
                    $128 = ((($this)) + 436|0);
                    __THREW__ = 0;
                    invoke_vi(793,($128|0));
                    $129 = __THREW__; __THREW__ = 0;
                    $130 = $129&1;
                    if ($130) {
                     $$lcssa = $128;
                    } else {
                     $131 = ((($this)) + 460|0);
                     __THREW__ = 0;
                     invoke_vi(793,($131|0));
                     $132 = __THREW__; __THREW__ = 0;
                     $133 = $132&1;
                     if ($133) {
                      $$lcssa = $131;
                      break;
                     }
                     $134 = ((($this)) + 484|0);
                     __THREW__ = 0;
                     invoke_vi(793,($134|0));
                     $135 = __THREW__; __THREW__ = 0;
                     $136 = $135&1;
                     if ($136) {
                      $$lcssa = $134;
                      break;
                     }
                     $137 = ((($this)) + 508|0);
                     __THREW__ = 0;
                     invoke_vi(793,($137|0));
                     $138 = __THREW__; __THREW__ = 0;
                     $139 = $138&1;
                     if ($139) {
                      $$lcssa = $137;
                      break;
                     }
                     $140 = ((($this)) + 532|0);
                     __THREW__ = 0;
                     invoke_vi(793,($140|0));
                     $141 = __THREW__; __THREW__ = 0;
                     $142 = $141&1;
                     if ($142) {
                      $$lcssa = $140;
                      break;
                     }
                     $143 = ((($this)) + 556|0);
                     __THREW__ = 0;
                     invoke_vi(793,($143|0));
                     $144 = __THREW__; __THREW__ = 0;
                     $145 = $144&1;
                     if ($145) {
                      $$lcssa = $143;
                      break;
                     }
                     $146 = ((($this)) + 580|0);
                     __THREW__ = 0;
                     invoke_vi(793,($146|0));
                     $147 = __THREW__; __THREW__ = 0;
                     $148 = $147&1;
                     if ($148) {
                      $$lcssa = $146;
                      break;
                     }
                     $149 = ((($this)) + 604|0);
                     __THREW__ = 0;
                     invoke_vi(793,($149|0));
                     $150 = __THREW__; __THREW__ = 0;
                     $151 = $150&1;
                     if ($151) {
                      $$lcssa = $149;
                      break;
                     }
                     $152 = ((($this)) + 628|0);
                     __THREW__ = 0;
                     invoke_vi(793,($152|0));
                     $153 = __THREW__; __THREW__ = 0;
                     $154 = $153&1;
                     if ($154) {
                      $$lcssa = $152;
                      break;
                     }
                     $155 = ((($this)) + 652|0);
                     __THREW__ = 0;
                     invoke_vi(793,($155|0));
                     $156 = __THREW__; __THREW__ = 0;
                     $157 = $156&1;
                     if ($157) {
                      $$lcssa = $155;
                      break;
                     }
                     $158 = ((($this)) + 676|0);
                     __THREW__ = 0;
                     invoke_vi(793,($158|0));
                     $159 = __THREW__; __THREW__ = 0;
                     $160 = $159&1;
                     if ($160) {
                      $$lcssa = $158;
                      break;
                     }
                     $161 = ((($this)) + 700|0);
                     __THREW__ = 0;
                     invoke_vi(793,($161|0));
                     $162 = __THREW__; __THREW__ = 0;
                     $163 = $162&1;
                     if ($163) {
                      $$lcssa = $161;
                      break;
                     }
                     $164 = ((($this)) + 724|0);
                     __THREW__ = 0;
                     invoke_vi(793,($164|0));
                     $165 = __THREW__; __THREW__ = 0;
                     $166 = $165&1;
                     if ($166) {
                      $$lcssa = $164;
                      break;
                     }
                     $167 = ((($this)) + 748|0);
                     __THREW__ = 0;
                     invoke_vi(793,($167|0));
                     $168 = __THREW__; __THREW__ = 0;
                     $169 = $168&1;
                     if ($169) {
                      $$lcssa = $167;
                      break;
                     }
                     $170 = ((($this)) + 772|0);
                     __THREW__ = 0;
                     invoke_vi(795,($170|0));
                     $171 = __THREW__; __THREW__ = 0;
                     $172 = $171&1;
                     L34: do {
                      if (!($172)) {
                       __THREW__ = 0;
                       invoke_viiii(794,($0|0),1,1024,0);
                       $6 = __THREW__; __THREW__ = 0;
                       $7 = $6&1;
                       if ($7) {
                        break;
                       }
                       $8 = ((($this)) + 28|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($8|0),2,512,0);
                       $9 = __THREW__; __THREW__ = 0;
                       $10 = $9&1;
                       if ($10) {
                        break;
                       }
                       $11 = ((($this)) + 52|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($11|0),4,256,0);
                       $12 = __THREW__; __THREW__ = 0;
                       $13 = $12&1;
                       if ($13) {
                        break;
                       }
                       $14 = ((($this)) + 76|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($14|0),8,128,0);
                       $15 = __THREW__; __THREW__ = 0;
                       $16 = $15&1;
                       if ($16) {
                        break;
                       }
                       $17 = ((($this)) + 100|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($17|0),16,64,0);
                       $18 = __THREW__; __THREW__ = 0;
                       $19 = $18&1;
                       if ($19) {
                        break;
                       }
                       $20 = ((($this)) + 124|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($20|0),32,32,0);
                       $21 = __THREW__; __THREW__ = 0;
                       $22 = $21&1;
                       if ($22) {
                        break;
                       }
                       $23 = ((($this)) + 148|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($23|0),64,16,0);
                       $24 = __THREW__; __THREW__ = 0;
                       $25 = $24&1;
                       if ($25) {
                        break;
                       }
                       $26 = ((($this)) + 172|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($26|0),128,8,0);
                       $27 = __THREW__; __THREW__ = 0;
                       $28 = $27&1;
                       if ($28) {
                        break;
                       }
                       $29 = ((($this)) + 196|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($29|0),256,4,0);
                       $30 = __THREW__; __THREW__ = 0;
                       $31 = $30&1;
                       if ($31) {
                        break;
                       }
                       $32 = ((($this)) + 220|0);
                       __THREW__ = 0;
                       invoke_viiii(794,($32|0),512,2,0);
                       $33 = __THREW__; __THREW__ = 0;
                       $34 = $33&1;
                       if ($34) {
                        break;
                       } else {
                        $indvars$iv = 10;
                       }
                       while(1) {
                        $35 = (((($this)) + 4|0) + (($indvars$iv*24)|0)|0);
                        $36 = 1 << $indvars$iv;
                        __THREW__ = 0;
                        invoke_viiii(794,($35|0),($36|0),1,0);
                        $37 = __THREW__; __THREW__ = 0;
                        $38 = $37&1;
                        if ($38) {
                         break L34;
                        }
                        $indvars$iv$next = (($indvars$iv) + 1)|0;
                        $39 = ($indvars$iv$next>>>0)<(32);
                        if ($39) {
                         $indvars$iv = $indvars$iv$next;
                        } else {
                         break;
                        }
                       }
                       return;
                      }
                     } while(0);
                     $46 = ___cxa_find_matching_catch_2()|0;
                     $47 = tempRet0;
                     $48 = ((($this)) + 748|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($48);
                     $49 = ((($this)) + 724|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($49);
                     $50 = ((($this)) + 700|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($50);
                     $51 = ((($this)) + 676|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($51);
                     $52 = ((($this)) + 652|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($52);
                     $53 = ((($this)) + 628|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($53);
                     $54 = ((($this)) + 604|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($54);
                     $55 = ((($this)) + 580|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($55);
                     $56 = ((($this)) + 556|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($56);
                     $57 = ((($this)) + 532|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($57);
                     $58 = ((($this)) + 508|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($58);
                     $59 = ((($this)) + 484|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($59);
                     $60 = ((($this)) + 460|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($60);
                     $61 = ((($this)) + 436|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($61);
                     $62 = ((($this)) + 412|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($62);
                     $63 = ((($this)) + 388|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($63);
                     $64 = ((($this)) + 364|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($64);
                     $65 = ((($this)) + 340|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($65);
                     $66 = ((($this)) + 316|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($66);
                     $67 = ((($this)) + 292|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($67);
                     $68 = ((($this)) + 268|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($68);
                     $69 = ((($this)) + 244|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($69);
                     $70 = ((($this)) + 220|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($70);
                     $71 = ((($this)) + 196|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($71);
                     $72 = ((($this)) + 172|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($72);
                     $73 = ((($this)) + 148|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($73);
                     $74 = ((($this)) + 124|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($74);
                     $75 = ((($this)) + 100|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($75);
                     $76 = ((($this)) + 76|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($76);
                     $77 = ((($this)) + 52|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($77);
                     $78 = ((($this)) + 28|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($78);
                     $79 = ((($this)) + 4|0);
                     __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($79);
                     $$0 = $47;$$01 = $46;
                     ___resumeException($$01|0);
                     // unreachable;
                    }
                   }
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 $40 = ___cxa_find_matching_catch_2()|0;
 $41 = tempRet0;
 $42 = ($0|0)==($$lcssa|0);
 if ($42) {
  $$0 = $41;$$01 = $40;
  ___resumeException($$01|0);
  // unreachable;
 } else {
  $44 = $$lcssa;
 }
 while(1) {
  $43 = ((($44)) + -24|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($43);
  $45 = ($43|0)==($0|0);
  if ($45) {
   $$0 = $41;$$01 = $40;
   break;
  } else {
   $44 = $43;
  }
 }
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (3984);
 $0 = ((($this)) + 748|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($0);
 $1 = ((($this)) + 724|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($1);
 $2 = ((($this)) + 700|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($2);
 $3 = ((($this)) + 676|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($3);
 $4 = ((($this)) + 652|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($4);
 $5 = ((($this)) + 628|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($5);
 $6 = ((($this)) + 604|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($6);
 $7 = ((($this)) + 580|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($7);
 $8 = ((($this)) + 556|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($8);
 $9 = ((($this)) + 532|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($9);
 $10 = ((($this)) + 508|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($10);
 $11 = ((($this)) + 484|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($11);
 $12 = ((($this)) + 460|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($12);
 $13 = ((($this)) + 436|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($13);
 $14 = ((($this)) + 412|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($14);
 $15 = ((($this)) + 388|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($15);
 $16 = ((($this)) + 364|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($16);
 $17 = ((($this)) + 340|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($17);
 $18 = ((($this)) + 316|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($18);
 $19 = ((($this)) + 292|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($19);
 $20 = ((($this)) + 268|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($20);
 $21 = ((($this)) + 244|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($21);
 $22 = ((($this)) + 220|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($22);
 $23 = ((($this)) + 196|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($23);
 $24 = ((($this)) + 172|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($24);
 $25 = ((($this)) + 148|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($25);
 $26 = ((($this)) + 124|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($26);
 $27 = ((($this)) + 100|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($27);
 $28 = ((($this)) + 76|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($28);
 $29 = ((($this)) + 52|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($29);
 $30 = ((($this)) + 28|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($30);
 $31 = ((($this)) + 4|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($31);
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolD2Ev($this) {
 $this = $this|0;
 var $$cast = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $4 = $1;
 while(1) {
  $3 = HEAP32[$4>>2]|0;
  __ZdaPv($4);
  HEAP32[$0>>2] = $3;
  $$cast = $3;
  $5 = ($3|0)==(0);
  if ($5) {
   break;
  } else {
   $4 = $$cast;
  }
 }
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager15setMaxAllocSizeEj($this,$maxAllocSize) {
 $this = $this|0;
 $maxAllocSize = $maxAllocSize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $indvars$iv = 0;
 while(1) {
  $0 = (((($this)) + 4|0) + (($indvars$iv*24)|0)|0);
  $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool12getBlockSizeEv($0)|0);
  $2 = (($maxAllocSize>>>0) / ($1>>>0))&-1;
  $3 = ($2|0)==(0);
  if ($3) {
   $6 = 1;
  } else {
   $4 = (__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool12getBlockSizeEv($0)|0);
   $5 = (($maxAllocSize>>>0) / ($4>>>0))&-1;
   $6 = $5;
  }
  __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool23setMaxCapacityIncrementEj($0,$6);
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $exitcond = ($indvars$iv$next|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $indvars$iv = $indvars$iv$next;
  }
 }
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager5allocEj($this,$s) {
 $this = $this|0;
 $s = $s|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $lock = sp;
 $0 = ($s|0)==(0);
 if ($0) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $1 = ((($this)) + 772|0);
 __ZN9ExitGames6Common7Helpers9LockguardC2ERNS1_8SpinlockE($lock,$1);
 $2 = (($s) + -1)|0;
 __THREW__ = 0;
 $3 = (invoke_ii(796,($2|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  $6 = (($3) + 1)|0;
  $7 = $6 & 255;
  $8 = (((($this)) + 4|0) + (($7*24)|0)|0);
  __THREW__ = 0;
  $9 = (invoke_ii(797,($8|0))|0);
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
   $$0 = $9;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
 ___resumeException($12|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager7deallocEPv($this,$p) {
 $this = $this|0;
 $p = $p|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $lock = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $lock = sp;
 $0 = ($p|0)==(0|0);
 if ($0) {
  STACKTOP = sp;return;
 }
 $1 = ((($this)) + 772|0);
 __ZN9ExitGames6Common7Helpers9LockguardC2ERNS1_8SpinlockE($lock,$1);
 $2 = ((($p)) + -4|0);
 $3 = HEAP32[$2>>2]|0;
 __THREW__ = 0;
 invoke_vii(798,($3|0),($p|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
  ___resumeException($6|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager6resizeEPvj($this,$p,$size) {
 $this = $this|0;
 $p = $p|0;
 $size = $size|0;
 var $$0 = 0, $$not3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $lock = 0, $or$cond$not = 0, $or$cond2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $lock = sp;
 $0 = ((($this)) + 772|0);
 __ZN9ExitGames6Common7Helpers9LockguardC2ERNS1_8SpinlockE($lock,$0);
 $1 = ($p|0)!=(0|0);
 if ($1) {
  $2 = ((($p)) + -4|0);
  $3 = HEAP32[$2>>2]|0;
  __THREW__ = 0;
  $4 = (invoke_ii(799,($3|0))|0);
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   $$not3 = ($4>>>0)>=($size>>>0);
   $or$cond$not = $1 & $$not3;
   $7 = $4 >>> 1;
   $8 = ($7>>>0)<($size>>>0);
   $or$cond2 = $or$cond$not & $8;
   if ($or$cond2) {
    $$0 = $p;
    __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
    STACKTOP = sp;return ($$0|0);
   } else {
    $16 = $4;
    label = 4;
   }
  }
 } else {
  $16 = 0;
  label = 4;
 }
 if ((label|0) == 4) {
  $9 = HEAP32[$this>>2]|0;
  $10 = ((($9)) + 12|0);
  $11 = HEAP32[$10>>2]|0;
  __THREW__ = 0;
  $12 = (invoke_iii($11|0,($this|0),($size|0))|0);
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if (!($14)) {
   if (!($1)) {
    $$0 = $12;
    __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
    STACKTOP = sp;return ($$0|0);
   }
   $15 = ($16>>>0)>($size>>>0);
   $17 = $15 ? $size : $16;
   _memcpy(($12|0),($p|0),($17|0))|0;
   $18 = HEAP32[$this>>2]|0;
   $19 = ((($18)) + 16|0);
   $20 = HEAP32[$19>>2]|0;
   __THREW__ = 0;
   invoke_vii($20|0,($this|0),($p|0));
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if (!($22)) {
    $$0 = $12;
    __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
    STACKTOP = sp;return ($$0|0);
   }
  }
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __ZN9ExitGames6Common7Helpers9LockguardD2Ev($lock);
 ___resumeException($23|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool12getBlockSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Common7Helpers9LockguardD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 __THREW__ = 0;
 invoke_vi(800,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_3(0|0)|0;
  $4 = tempRet0;
  ___clang_call_terminate($3);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common7Helpers8Spinlock6unlockEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = ((($this)) + 4|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = (($2) + -1)<<24>>24;
 HEAP8[$1>>0] = $3;
 $4 = ($3<<24>>24)==(0);
 if (!($4)) {
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($0);
 $5 = HEAP32[$0>>2]|0;
 HEAP32[$this>>2] = $5;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = 0;
 return;
}
function __ZN9ExitGames6Common7Helpers9LockguardC2ERNS1_8SpinlockE($this,$lock) {
 $this = $this|0;
 $lock = $lock|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $lock;
 __ZN9ExitGames6Common7Helpers8Spinlock4lockEv($lock);
 return;
}
function __ZN9ExitGames6Common7Helpers8Spinlock4lockEv($this) {
 $this = $this|0;
 var $$index = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $success = 0, $success6 = 0;
 var $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp + 12|0;
 $1 = sp + 8|0;
 $temp = sp + 4|0;
 $2 = sp;
 $3 = HEAP32[$this>>2]|0;
 HEAP32[$0>>2] = $3;
 $4 = (__ZN9ExitGames6Common7Helpers6Thread16getLocalThreadIDEv()|0);
 HEAP32[$1>>2] = $4;
 $5 = (__ZNK9ExitGames6Common7Helpers6Thread2IDneERKS3_($0,$1)|0);
 if ($5) {
  __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($temp);
  __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($2);
  $6 = HEAP32[$2>>2]|0;
  HEAP32[$temp>>2] = $6;
  $7 = (__ZN9ExitGames6Common7Helpers6Thread16getLocalThreadIDEv()|0);
  $8 = HEAP32[$temp>>2]|0;
  $9 = HEAP32[$this>>2]|0;if (($9|0) == ($8|0)) HEAP32[$this>>2] = $7;
  $success = ($9|0)==($8|0);
  if (!($success)) {
   $$index = $9;
   while(1) {
    HEAP32[$temp>>2] = $$index;
    __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($2);
    $10 = HEAP32[$2>>2]|0;
    HEAP32[$temp>>2] = $10;
    $11 = (__ZN9ExitGames6Common7Helpers6Thread16getLocalThreadIDEv()|0);
    $12 = HEAP32[$temp>>2]|0;
    $13 = HEAP32[$this>>2]|0;if (($13|0) == ($12|0)) HEAP32[$this>>2] = $11;
    $success6 = ($13|0)==($12|0);
    if ($success6) {
     break;
    } else {
     $$index = $13;
    }
   }
  }
 }
 $14 = ((($this)) + 4|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = (($15) + 1)<<24>>24;
 HEAP8[$14>>0] = $16;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers6Thread16getLocalThreadIDEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function __ZNK9ExitGames6Common7Helpers6Thread2IDneERKS3_($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = HEAP32[$Rhs>>2]|0;
 $2 = ($0|0)!=($1|0);
 return ($2|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool7deallocEPv($this,$p) {
 $this = $this|0;
 $p = $p|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($p)) + -8|0);
 $1 = HEAP32[$this>>2]|0;
 HEAP32[$0>>2] = $1;
 HEAP32[$this>>2] = $0;
 return;
}
function __ZN9ExitGames6Common7Helpers10MathHelper7getLog2Ej($v) {
 $v = $v|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $v >>> 16;
 $1 = ($0|0)==(0);
 if ($1) {
  $12 = $v >>> 8;
  $13 = ($12|0)==(0);
  if ($13) {
   $18 = (145439 + ($v)|0);
   $19 = HEAP8[$18>>0]|0;
   $20 = $19 << 24 >> 24;
   $21 = $20;
   return ($21|0);
  } else {
   $14 = (145439 + ($12)|0);
   $15 = HEAP8[$14>>0]|0;
   $16 = $15 << 24 >> 24;
   $17 = (($16) + 8)|0;
   $21 = $17;
   return ($21|0);
  }
 } else {
  $2 = $v >>> 24;
  $3 = ($2|0)==(0);
  if ($3) {
   $8 = (145439 + ($0)|0);
   $9 = HEAP8[$8>>0]|0;
   $10 = $9 << 24 >> 24;
   $11 = (($10) + 16)|0;
   $21 = $11;
   return ($21|0);
  } else {
   $4 = (145439 + ($2)|0);
   $5 = HEAP8[$4>>0]|0;
   $6 = $5 << 24 >> 24;
   $7 = (($6) + 24)|0;
   $21 = $7;
   return ($21|0);
  }
 }
 return (0)|0;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool5allocEv($this) {
 $this = $this|0;
 var $$lcssa$i = 0, $$lcssa$i$i = 0, $$phi$trans$insert$i = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $8 = 0, $9 = 0, $or$cond$i = 0, $p$0$lcssa$i = 0, $p$0$lcssa$i$i = 0, $p$01$i = 0, $p$01$i$i = 0, $pI$02$i = 0, $pI$02$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $2 = ((($this)) + 8|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = ((($this)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + 8)|0;
  $7 = Math_imul($6, $3)|0;
  $8 = (($7) + 8)|0;
  $9 = (__Znaj($8)|0);
  $10 = ((($this)) + 16|0);
  $11 = $10;
  $12 = $11;
  $13 = HEAP32[$12>>2]|0;
  $14 = (($11) + 4)|0;
  $15 = $14;
  $16 = HEAP32[$15>>2]|0;
  $17 = $9;
  $18 = $17;
  HEAP32[$18>>2] = $13;
  $19 = (($17) + 4)|0;
  $20 = $19;
  HEAP32[$20>>2] = $16;
  HEAP32[$10>>2] = $9;
  $21 = ((($9)) + 8|0);
  HEAP32[$this>>2] = $21;
  $22 = (($21) + ($6<<3)|0);
  $23 = (($21) + ($5<<3)|0);
  HEAP32[$21>>2] = $23;
  $24 = $22;
  $25 = $21;
  $26 = (($24) - ($25))|0;
  $27 = ($26>>>0)<($7>>>0);
  if ($27) {
   $30 = $21;$p$01$i = $21;$pI$02$i = $22;
   while(1) {
    HEAP32[$p$01$i>>2] = $pI$02$i;
    $28 = ((($p$01$i)) + 4|0);
    HEAP32[$28>>2] = $this;
    $29 = HEAP32[$30>>2]|0;
    $31 = (($pI$02$i) + ($6)|0);
    $32 = $31;
    $33 = (($32) - ($25))|0;
    $34 = ($33>>>0)<($7>>>0);
    if ($34) {
     $30 = $29;$p$01$i = $29;$pI$02$i = $31;
    } else {
     $$lcssa$i = $29;$p$0$lcssa$i = $29;
     break;
    }
   }
  } else {
   $$lcssa$i = $21;$p$0$lcssa$i = $21;
  }
  HEAP32[$$lcssa$i>>2] = 0;
  $35 = ((($p$0$lcssa$i)) + 4|0);
  HEAP32[$35>>2] = $this;
  $37 = $21;
 } else {
  $37 = $0;
 }
 $36 = HEAP32[$37>>2]|0;
 HEAP32[$this>>2] = $36;
 $38 = ($36|0)==(0|0);
 if (!($38)) {
  $77 = ((($37)) + 8|0);
  return ($77|0);
 }
 $39 = ((($this)) + 12|0);
 $40 = HEAP32[$39>>2]|0;
 $41 = ($40|0)!=(0);
 $$phi$trans$insert$i = ((($this)) + 8|0);
 $$pre$i = HEAP32[$$phi$trans$insert$i>>2]|0;
 $42 = ($40>>>0)<($$pre$i>>>0);
 $or$cond$i = $41 & $42;
 $43 = $or$cond$i ? $40 : $$pre$i;
 $44 = ((($this)) + 4|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (($45) + 8)|0;
 $47 = Math_imul($43, $46)|0;
 $48 = (($47) + 8)|0;
 $49 = (__Znaj($48)|0);
 $50 = ((($this)) + 16|0);
 $51 = $50;
 $52 = $51;
 $53 = HEAP32[$52>>2]|0;
 $54 = (($51) + 4)|0;
 $55 = $54;
 $56 = HEAP32[$55>>2]|0;
 $57 = $49;
 $58 = $57;
 HEAP32[$58>>2] = $53;
 $59 = (($57) + 4)|0;
 $60 = $59;
 HEAP32[$60>>2] = $56;
 HEAP32[$50>>2] = $49;
 $61 = ((($49)) + 8|0);
 HEAP32[$this>>2] = $61;
 $62 = (($61) + ($46<<3)|0);
 $63 = (($61) + ($45<<3)|0);
 HEAP32[$61>>2] = $63;
 $64 = $62;
 $65 = $61;
 $66 = (($64) - ($65))|0;
 $67 = ($66>>>0)<($47>>>0);
 if ($67) {
  $70 = $61;$p$01$i$i = $61;$pI$02$i$i = $62;
  while(1) {
   HEAP32[$p$01$i$i>>2] = $pI$02$i$i;
   $68 = ((($p$01$i$i)) + 4|0);
   HEAP32[$68>>2] = $this;
   $69 = HEAP32[$70>>2]|0;
   $71 = (($pI$02$i$i) + ($46)|0);
   $72 = $71;
   $73 = (($72) - ($65))|0;
   $74 = ($73>>>0)<($47>>>0);
   if ($74) {
    $70 = $69;$p$01$i$i = $69;$pI$02$i$i = $71;
   } else {
    $$lcssa$i$i = $69;$p$0$lcssa$i$i = $69;
    break;
   }
  }
 } else {
  $$lcssa$i$i = $61;$p$0$lcssa$i$i = $61;
 }
 HEAP32[$$lcssa$i$i>>2] = 0;
 $75 = ((($p$0$lcssa$i$i)) + 4|0);
 HEAP32[$75>>2] = $this;
 $76 = (($43) + ($$pre$i))|0;
 HEAP32[$$phi$trans$insert$i>>2] = $76;
 HEAP32[$37>>2] = $65;
 $77 = ((($37)) + 8|0);
 return ($77|0);
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool23setMaxCapacityIncrementEj($this,$maxCapacityIncrement) {
 $this = $this|0;
 $maxCapacityIncrement = $maxCapacityIncrement|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 HEAP32[$0>>2] = $maxCapacityIncrement;
 return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool4initEjjj($this,$blockSize,$initialCapacity,$maxCapacityIncrement) {
 $this = $this|0;
 $blockSize = $blockSize|0;
 $initialCapacity = $initialCapacity|0;
 $maxCapacityIncrement = $maxCapacityIncrement|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $blockSize;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = $initialCapacity;
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = $maxCapacityIncrement;
 return;
}
function __ZN9ExitGames6Common7Helpers8SpinlockC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 __ZN9ExitGames6Common7Helpers6Thread2IDC2Ev($0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$this>>2] = $1;
 $2 = ((($this)) + 4|0);
 HEAP8[$2>>0] = 0;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolC2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common7Helpers8TypeName3cutEPKc($agg$result,$rawTypeName) {
 $agg$result = $agg$result|0;
 $rawTypeName = $rawTypeName|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$10 = 0, $$1019 = 0, $$11 = 0, $$110 = 0, $$1120 = 0, $$12 = 0, $$1221 = 0, $$13 = 0, $$1322 = 0, $$14 = 0, $$1423 = 0, $$15 = 0, $$1524 = 0, $$16 = 0, $$1625 = 0, $$17 = 0, $$1726 = 0;
 var $$18 = 0, $$1827 = 0, $$19 = 0, $$1928 = 0, $$2 = 0, $$20 = 0, $$2029 = 0, $$211 = 0, $$3 = 0, $$312 = 0, $$4 = 0, $$413 = 0, $$5 = 0, $$514 = 0, $$6 = 0, $$615 = 0, $$7 = 0, $$716 = 0, $$8 = 0, $$817 = 0;
 var $$9 = 0, $$918 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $0 = sp + 336|0;
 $1 = sp + 320|0;
 $2 = sp + 304|0;
 $3 = sp + 288|0;
 $4 = sp + 272|0;
 $5 = sp + 256|0;
 $6 = sp + 240|0;
 $7 = sp + 224|0;
 $8 = sp + 208|0;
 $9 = sp + 192|0;
 $10 = sp + 176|0;
 $11 = sp + 160|0;
 $12 = sp + 144|0;
 $13 = sp + 128|0;
 $14 = sp + 112|0;
 $15 = sp + 96|0;
 $16 = sp + 80|0;
 $17 = sp + 64|0;
 $18 = sp + 48|0;
 $19 = sp + 32|0;
 $20 = sp + 16|0;
 $21 = sp;
 __ZN9ExitGames6Common7JStringC2EPKc($agg$result,$rawTypeName);
 __THREW__ = 0;
 invoke_vii(801,($6|0),(4008|0));
 $22 = __THREW__; __THREW__ = 0;
 $23 = $22&1;
 L1: do {
  if (!($23)) {
   __THREW__ = 0;
   invoke_vii(801,($7|0),(163772|0));
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $98 = ___cxa_find_matching_catch_2()|0;
    $99 = tempRet0;
    $$16 = $99;$$1625 = $98;
   } else {
    __THREW__ = 0;
    invoke_viiii(802,($5|0),($agg$result|0),($6|0),($7|0));
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if ($27) {
     $100 = ___cxa_find_matching_catch_2()|0;
     $101 = tempRet0;
     $$15 = $101;$$1524 = $100;
    } else {
     __THREW__ = 0;
     invoke_vii(801,($8|0),(4036|0));
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $102 = ___cxa_find_matching_catch_2()|0;
      $103 = tempRet0;
      $$14 = $103;$$1423 = $102;
     } else {
      __THREW__ = 0;
      invoke_vii(801,($9|0),(163772|0));
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      if ($31) {
       $104 = ___cxa_find_matching_catch_2()|0;
       $105 = tempRet0;
       $$13 = $105;$$1322 = $104;
      } else {
       __THREW__ = 0;
       invoke_viiii(802,($4|0),($5|0),($8|0),($9|0));
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        $106 = ___cxa_find_matching_catch_2()|0;
        $107 = tempRet0;
        $$12 = $107;$$1221 = $106;
       } else {
        __THREW__ = 0;
        invoke_vii(801,($10|0),(4068|0));
        $34 = __THREW__; __THREW__ = 0;
        $35 = $34&1;
        if ($35) {
         $108 = ___cxa_find_matching_catch_2()|0;
         $109 = tempRet0;
         $$11 = $109;$$1120 = $108;
        } else {
         __THREW__ = 0;
         invoke_vii(801,($11|0),(163772|0));
         $36 = __THREW__; __THREW__ = 0;
         $37 = $36&1;
         if ($37) {
          $110 = ___cxa_find_matching_catch_2()|0;
          $111 = tempRet0;
          $$10 = $111;$$1019 = $110;
         } else {
          __THREW__ = 0;
          invoke_viiii(802,($3|0),($4|0),($10|0),($11|0));
          $38 = __THREW__; __THREW__ = 0;
          $39 = $38&1;
          if ($39) {
           $112 = ___cxa_find_matching_catch_2()|0;
           $113 = tempRet0;
           $$9 = $113;$$918 = $112;
          } else {
           __THREW__ = 0;
           invoke_vii(801,($12|0),(4096|0));
           $40 = __THREW__; __THREW__ = 0;
           $41 = $40&1;
           if ($41) {
            $114 = ___cxa_find_matching_catch_2()|0;
            $115 = tempRet0;
            $$8 = $115;$$817 = $114;
           } else {
            __THREW__ = 0;
            invoke_vii(801,($13|0),(163772|0));
            $42 = __THREW__; __THREW__ = 0;
            $43 = $42&1;
            if ($43) {
             $116 = ___cxa_find_matching_catch_2()|0;
             $117 = tempRet0;
             $$7 = $117;$$716 = $116;
            } else {
             __THREW__ = 0;
             invoke_viiii(802,($2|0),($3|0),($12|0),($13|0));
             $44 = __THREW__; __THREW__ = 0;
             $45 = $44&1;
             if ($45) {
              $118 = ___cxa_find_matching_catch_2()|0;
              $119 = tempRet0;
              $$6 = $119;$$615 = $118;
             } else {
              __THREW__ = 0;
              invoke_vii(801,($14|0),(14344|0));
              $46 = __THREW__; __THREW__ = 0;
              $47 = $46&1;
              if ($47) {
               $120 = ___cxa_find_matching_catch_2()|0;
               $121 = tempRet0;
               $$5 = $121;$$514 = $120;
              } else {
               __THREW__ = 0;
               invoke_vii(801,($15|0),(163772|0));
               $48 = __THREW__; __THREW__ = 0;
               $49 = $48&1;
               if ($49) {
                $122 = ___cxa_find_matching_catch_2()|0;
                $123 = tempRet0;
                $$4 = $123;$$413 = $122;
               } else {
                __THREW__ = 0;
                invoke_viiii(802,($1|0),($2|0),($14|0),($15|0));
                $50 = __THREW__; __THREW__ = 0;
                $51 = $50&1;
                if ($51) {
                 $124 = ___cxa_find_matching_catch_2()|0;
                 $125 = tempRet0;
                 $$3 = $125;$$312 = $124;
                } else {
                 __THREW__ = 0;
                 invoke_vii(801,($16|0),(5580|0));
                 $52 = __THREW__; __THREW__ = 0;
                 $53 = $52&1;
                 if ($53) {
                  $126 = ___cxa_find_matching_catch_2()|0;
                  $127 = tempRet0;
                  $$2 = $127;$$211 = $126;
                 } else {
                  __THREW__ = 0;
                  invoke_vii(801,($17|0),(4120|0));
                  $54 = __THREW__; __THREW__ = 0;
                  $55 = $54&1;
                  if ($55) {
                   $128 = ___cxa_find_matching_catch_2()|0;
                   $129 = tempRet0;
                   $$1 = $129;$$110 = $128;
                  } else {
                   __THREW__ = 0;
                   invoke_viiii(802,($0|0),($1|0),($16|0),($17|0));
                   $56 = __THREW__; __THREW__ = 0;
                   $57 = $56&1;
                   do {
                    if ($57) {
                     $130 = ___cxa_find_matching_catch_2()|0;
                     $131 = tempRet0;
                     $$0 = $131;$$01 = $130;
                    } else {
                     __THREW__ = 0;
                     (invoke_iii(803,($agg$result|0),($0|0))|0);
                     $58 = __THREW__; __THREW__ = 0;
                     $59 = $58&1;
                     if ($59) {
                      $132 = ___cxa_find_matching_catch_2()|0;
                      $133 = tempRet0;
                      __ZN9ExitGames6Common7JStringD2Ev($0);
                      $$0 = $133;$$01 = $132;
                      break;
                     }
                     __ZN9ExitGames6Common7JStringD2Ev($0);
                     __ZN9ExitGames6Common7JStringD2Ev($17);
                     __ZN9ExitGames6Common7JStringD2Ev($16);
                     __ZN9ExitGames6Common7JStringD2Ev($1);
                     __ZN9ExitGames6Common7JStringD2Ev($15);
                     __ZN9ExitGames6Common7JStringD2Ev($14);
                     __ZN9ExitGames6Common7JStringD2Ev($2);
                     __ZN9ExitGames6Common7JStringD2Ev($13);
                     __ZN9ExitGames6Common7JStringD2Ev($12);
                     __ZN9ExitGames6Common7JStringD2Ev($3);
                     __ZN9ExitGames6Common7JStringD2Ev($11);
                     __ZN9ExitGames6Common7JStringD2Ev($10);
                     __ZN9ExitGames6Common7JStringD2Ev($4);
                     __ZN9ExitGames6Common7JStringD2Ev($9);
                     __ZN9ExitGames6Common7JStringD2Ev($8);
                     __ZN9ExitGames6Common7JStringD2Ev($5);
                     __ZN9ExitGames6Common7JStringD2Ev($7);
                     __ZN9ExitGames6Common7JStringD2Ev($6);
                     while(1) {
                      __THREW__ = 0;
                      $60 = (invoke_iii(804,($agg$result|0),58)|0);
                      $61 = __THREW__; __THREW__ = 0;
                      $62 = $61&1;
                      if ($62) {
                       break L1;
                      }
                      $63 = ($60|0)==(-1);
                      if ($63) {
                       label = 79;
                       break;
                      }
                      __THREW__ = 0;
                      $64 = (invoke_iii(804,($agg$result|0),58)|0);
                      $65 = __THREW__; __THREW__ = 0;
                      $66 = $65&1;
                      if ($66) {
                       break L1;
                      }
                      __THREW__ = 0;
                      invoke_viiii(805,($20|0),($agg$result|0),0,($64|0));
                      $67 = __THREW__; __THREW__ = 0;
                      $68 = $67&1;
                      if ($68) {
                       break L1;
                      }
                      __THREW__ = 0;
                      $69 = (invoke_iii(804,($agg$result|0),60)|0);
                      $70 = __THREW__; __THREW__ = 0;
                      $71 = $70&1;
                      if ($71) {
                       label = 72;
                       break;
                      }
                      __THREW__ = 0;
                      $72 = (invoke_iii(804,($agg$result|0),44)|0);
                      $73 = __THREW__; __THREW__ = 0;
                      $74 = $73&1;
                      if ($74) {
                       label = 72;
                       break;
                      }
                      $75 = ($69|0)>($72|0);
                      if ($75) {
                       __THREW__ = 0;
                       $76 = (invoke_iii(804,($agg$result|0),60)|0);
                       $77 = __THREW__; __THREW__ = 0;
                       $78 = $77&1;
                       if ($78) {
                        label = 72;
                        break;
                       } else {
                        $83 = $76;
                       }
                      } else {
                       __THREW__ = 0;
                       $79 = (invoke_iii(804,($agg$result|0),44)|0);
                       $80 = __THREW__; __THREW__ = 0;
                       $81 = $80&1;
                       if ($81) {
                        label = 72;
                        break;
                       } else {
                        $83 = $79;
                       }
                      }
                      $82 = (($83) + 1)|0;
                      __THREW__ = 0;
                      invoke_viiii(805,($19|0),($20|0),0,($82|0));
                      $84 = __THREW__; __THREW__ = 0;
                      $85 = $84&1;
                      if ($85) {
                       label = 72;
                       break;
                      }
                      __THREW__ = 0;
                      $86 = (invoke_iii(804,($agg$result|0),58)|0);
                      $87 = __THREW__; __THREW__ = 0;
                      $88 = $87&1;
                      if ($88) {
                       label = 73;
                       break;
                      }
                      $89 = (($86) + 1)|0;
                      __THREW__ = 0;
                      invoke_viii(806,($21|0),($agg$result|0),($89|0));
                      $90 = __THREW__; __THREW__ = 0;
                      $91 = $90&1;
                      if ($91) {
                       label = 73;
                       break;
                      }
                      __THREW__ = 0;
                      invoke_viii(785,($18|0),($19|0),($21|0));
                      $92 = __THREW__; __THREW__ = 0;
                      $93 = $92&1;
                      if ($93) {
                       label = 74;
                       break;
                      }
                      __THREW__ = 0;
                      (invoke_iii(803,($agg$result|0),($18|0))|0);
                      $94 = __THREW__; __THREW__ = 0;
                      $95 = $94&1;
                      if ($95) {
                       label = 75;
                       break;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($18);
                      __ZN9ExitGames6Common7JStringD2Ev($21);
                      __ZN9ExitGames6Common7JStringD2Ev($19);
                      __ZN9ExitGames6Common7JStringD2Ev($20);
                     }
                     if ((label|0) == 72) {
                      $134 = ___cxa_find_matching_catch_2()|0;
                      $135 = tempRet0;
                      $$19 = $135;$$1928 = $134;
                     }
                     else if ((label|0) == 73) {
                      $136 = ___cxa_find_matching_catch_2()|0;
                      $137 = tempRet0;
                      $$18 = $137;$$1827 = $136;
                      label = 77;
                     }
                     else if ((label|0) == 74) {
                      $138 = ___cxa_find_matching_catch_2()|0;
                      $139 = tempRet0;
                      $$17 = $139;$$1726 = $138;
                      label = 76;
                     }
                     else if ((label|0) == 75) {
                      $140 = ___cxa_find_matching_catch_2()|0;
                      $141 = tempRet0;
                      __ZN9ExitGames6Common7JStringD2Ev($18);
                      $$17 = $141;$$1726 = $140;
                      label = 76;
                     }
                     else if ((label|0) == 79) {
                      STACKTOP = sp;return;
                     }
                     if ((label|0) == 76) {
                      __ZN9ExitGames6Common7JStringD2Ev($21);
                      $$18 = $$17;$$1827 = $$1726;
                      label = 77;
                     }
                     if ((label|0) == 77) {
                      __ZN9ExitGames6Common7JStringD2Ev($19);
                      $$19 = $$18;$$1928 = $$1827;
                     }
                     __ZN9ExitGames6Common7JStringD2Ev($20);
                     $$20 = $$19;$$2029 = $$1928;
                     __ZN9ExitGames6Common7JStringD2Ev($agg$result);
                     ___resumeException($$2029|0);
                     // unreachable;
                    }
                   } while(0);
                   __ZN9ExitGames6Common7JStringD2Ev($17);
                   $$1 = $$0;$$110 = $$01;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($16);
                  $$2 = $$1;$$211 = $$110;
                 }
                 __ZN9ExitGames6Common7JStringD2Ev($1);
                 $$3 = $$2;$$312 = $$211;
                }
                __ZN9ExitGames6Common7JStringD2Ev($15);
                $$4 = $$3;$$413 = $$312;
               }
               __ZN9ExitGames6Common7JStringD2Ev($14);
               $$5 = $$4;$$514 = $$413;
              }
              __ZN9ExitGames6Common7JStringD2Ev($2);
              $$6 = $$5;$$615 = $$514;
             }
             __ZN9ExitGames6Common7JStringD2Ev($13);
             $$7 = $$6;$$716 = $$615;
            }
            __ZN9ExitGames6Common7JStringD2Ev($12);
            $$8 = $$7;$$817 = $$716;
           }
           __ZN9ExitGames6Common7JStringD2Ev($3);
           $$9 = $$8;$$918 = $$817;
          }
          __ZN9ExitGames6Common7JStringD2Ev($11);
          $$10 = $$9;$$1019 = $$918;
         }
         __ZN9ExitGames6Common7JStringD2Ev($10);
         $$11 = $$10;$$1120 = $$1019;
        }
        __ZN9ExitGames6Common7JStringD2Ev($4);
        $$12 = $$11;$$1221 = $$1120;
       }
       __ZN9ExitGames6Common7JStringD2Ev($9);
       $$13 = $$12;$$1322 = $$1221;
      }
      __ZN9ExitGames6Common7JStringD2Ev($8);
      $$14 = $$13;$$1423 = $$1322;
     }
     __ZN9ExitGames6Common7JStringD2Ev($5);
     $$15 = $$14;$$1524 = $$1423;
    }
    __ZN9ExitGames6Common7JStringD2Ev($7);
    $$16 = $$15;$$1625 = $$1524;
   }
   __ZN9ExitGames6Common7JStringD2Ev($6);
   $$20 = $$16;$$2029 = $$1625;
   __ZN9ExitGames6Common7JStringD2Ev($agg$result);
   ___resumeException($$2029|0);
   // unreachable;
  }
 } while(0);
 $96 = ___cxa_find_matching_catch_2()|0;
 $97 = tempRet0;
 $$20 = $97;$$2029 = $96;
 __ZN9ExitGames6Common7JStringD2Ev($agg$result);
 ___resumeException($$2029|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplERKNS0_7JStringES3_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$0>>2] = (3908);
 $1 = ((($0)) + 4|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($0)) + 8|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 12|0);
 HEAP32[$3>>2] = 0;
 __THREW__ = 0;
 $4 = (invoke_ii(756,(163772|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___resumeException($21|0);
  // unreachable;
 }
 HEAP32[$3>>2] = $4;
 HEAP32[$2>>2] = $4;
 $7 = (($4) + 1)|0;
 $8 = $7 << 2;
 $9 = (($8) + 8)|0;
 __THREW__ = 0;
 $10 = (invoke_ii(775,($9|0))|0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___resumeException($21|0);
  // unreachable;
 }
 $13 = ((($10)) + 8|0);
 $14 = $10;
 $15 = $14;
 HEAP32[$15>>2] = $7;
 $16 = (($14) + 4)|0;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = ($7|0)==(0);
 if (!($18)) {
  _memset(($13|0),0,($8|0))|0;
 }
 HEAP32[$1>>2] = $13;
 __THREW__ = 0;
 (invoke_iii(789,($13|0),(163772|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  ___resumeException($21|0);
  // unreachable;
 }
 $23 = ((($Lsh)) + 12|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$3>>2]|0;
 $26 = (($25) + ($24))|0;
 HEAP32[$3>>2] = $26;
 $27 = HEAP32[$2>>2]|0;
 $28 = ($26>>>0)>($27>>>0);
 if ($28) {
  $29 = HEAP32[$1>>2]|0;
  HEAP32[$2>>2] = $26;
  $30 = (($26) + 1)|0;
  $31 = $30 << 2;
  $32 = (($31) + 8)|0;
  __THREW__ = 0;
  $33 = (invoke_ii(775,($32|0))|0);
  $34 = __THREW__; __THREW__ = 0;
  $35 = $34&1;
  if ($35) {
   label = 32;
  } else {
   $36 = ((($33)) + 8|0);
   $37 = $33;
   $38 = $37;
   HEAP32[$38>>2] = $30;
   $39 = (($37) + 4)|0;
   $40 = $39;
   HEAP32[$40>>2] = 0;
   $41 = ($30|0)==(0);
   if (!($41)) {
    _memset(($36|0),0,($31|0))|0;
   }
   HEAP32[$1>>2] = $36;
   __THREW__ = 0;
   (invoke_iii(789,($36|0),($29|0))|0);
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   if ($43) {
    label = 32;
   } else {
    $44 = ($29|0)==(0|0);
    if ($44) {
     label = 14;
    } else {
     $45 = ((($29)) + -8|0);
     __THREW__ = 0;
     invoke_vi(750,($45|0));
     $46 = __THREW__; __THREW__ = 0;
     $47 = $46&1;
     if ($47) {
      label = 32;
     } else {
      label = 14;
     }
    }
   }
  }
 } else {
  label = 14;
 }
 do {
  if ((label|0) == 14) {
   $48 = HEAP32[$1>>2]|0;
   $49 = ((($Lsh)) + 4|0);
   $50 = HEAP32[$49>>2]|0;
   __THREW__ = 0;
   (invoke_iii(790,($48|0),($50|0))|0);
   $51 = __THREW__; __THREW__ = 0;
   $52 = $51&1;
   if ($52) {
    label = 32;
   } else {
    $53 = ((($Rsh)) + 12|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = HEAP32[$3>>2]|0;
    $56 = (($55) + ($54))|0;
    HEAP32[$3>>2] = $56;
    $57 = HEAP32[$2>>2]|0;
    $58 = ($56>>>0)>($57>>>0);
    if ($58) {
     $59 = HEAP32[$1>>2]|0;
     HEAP32[$2>>2] = $56;
     $60 = (($56) + 1)|0;
     $61 = $60 << 2;
     $62 = (($61) + 8)|0;
     __THREW__ = 0;
     $63 = (invoke_ii(775,($62|0))|0);
     $64 = __THREW__; __THREW__ = 0;
     $65 = $64&1;
     if ($65) {
      label = 32;
      break;
     }
     $66 = ((($63)) + 8|0);
     $67 = $63;
     $68 = $67;
     HEAP32[$68>>2] = $60;
     $69 = (($67) + 4)|0;
     $70 = $69;
     HEAP32[$70>>2] = 0;
     $71 = ($60|0)==(0);
     if (!($71)) {
      _memset(($66|0),0,($61|0))|0;
     }
     HEAP32[$1>>2] = $66;
     __THREW__ = 0;
     (invoke_iii(789,($66|0),($59|0))|0);
     $72 = __THREW__; __THREW__ = 0;
     $73 = $72&1;
     if ($73) {
      label = 32;
      break;
     }
     $74 = ($59|0)==(0|0);
     if (!($74)) {
      $75 = ((($59)) + -8|0);
      __THREW__ = 0;
      invoke_vi(750,($75|0));
      $76 = __THREW__; __THREW__ = 0;
      $77 = $76&1;
      if ($77) {
       label = 32;
       break;
      }
     }
    }
    $78 = HEAP32[$1>>2]|0;
    $79 = ((($Rsh)) + 4|0);
    $80 = HEAP32[$79>>2]|0;
    __THREW__ = 0;
    (invoke_iii(790,($78|0),($80|0))|0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     label = 32;
    } else {
     HEAP32[$agg$result>>2] = (3908);
     $83 = ((($agg$result)) + 4|0);
     HEAP32[$83>>2] = 0;
     $84 = ((($agg$result)) + 8|0);
     $85 = ((($agg$result)) + 12|0);
     $86 = HEAP32[$3>>2]|0;
     HEAP32[$85>>2] = $86;
     HEAP32[$84>>2] = $86;
     $87 = (($86) + 1)|0;
     $88 = $87 << 2;
     $89 = (($88) + 8)|0;
     __THREW__ = 0;
     $90 = (invoke_ii(775,($89|0))|0);
     $91 = __THREW__; __THREW__ = 0;
     $92 = $91&1;
     if (!($92)) {
      $93 = ((($90)) + 8|0);
      $94 = $90;
      $95 = $94;
      HEAP32[$95>>2] = $87;
      $96 = (($94) + 4)|0;
      $97 = $96;
      HEAP32[$97>>2] = 0;
      $98 = ($87|0)==(0);
      if (!($98)) {
       _memset(($93|0),0,($88|0))|0;
      }
      HEAP32[$83>>2] = $93;
      $99 = HEAP32[$1>>2]|0;
      __THREW__ = 0;
      (invoke_iii(789,($93|0),($99|0))|0);
      $100 = __THREW__; __THREW__ = 0;
      $101 = $100&1;
      if (!($101)) {
       HEAP32[$0>>2] = (3908);
       $104 = HEAP32[$1>>2]|0;
       $105 = ($104|0)==(0|0);
       if ($105) {
        STACKTOP = sp;return;
       }
       $106 = ((($104)) + -8|0);
       __THREW__ = 0;
       invoke_vi(750,($106|0));
       $107 = __THREW__; __THREW__ = 0;
       $108 = $107&1;
       if ($108) {
        $109 = ___cxa_find_matching_catch_3(0|0)|0;
        $110 = tempRet0;
        ___clang_call_terminate($109);
        // unreachable;
       } else {
        STACKTOP = sp;return;
       }
      }
     }
     $102 = ___cxa_find_matching_catch_2()|0;
     $103 = tempRet0;
     $eh$lpad$body$index2Z2D = $103;$eh$lpad$body$indexZ2D = $102;
    }
   }
  }
 } while(0);
 if ((label|0) == 32) {
  $111 = ___cxa_find_matching_catch_2()|0;
  $112 = tempRet0;
  $eh$lpad$body$index2Z2D = $112;$eh$lpad$body$indexZ2D = $111;
 }
 HEAP32[$0>>2] = (3908);
 $113 = HEAP32[$1>>2]|0;
 $114 = ($113|0)==(0|0);
 if ($114) {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 $115 = ((($113)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($115|0));
 $116 = __THREW__; __THREW__ = 0;
 $117 = $116&1;
 if ($117) {
  $118 = ___cxa_find_matching_catch_3(0|0)|0;
  $119 = tempRet0;
  ___clang_call_terminate($118);
  // unreachable;
 } else {
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
}
function __ZNK9ExitGames6Common7JString7replaceERKS1_S3_($agg$result,$this,$match,$replacement) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $match = $match|0;
 $replacement = $replacement|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $temp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $temp = sp + 32|0;
 $0 = sp + 16|0;
 $1 = sp;
 $2 = ((($this)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 HEAP32[$temp>>2] = (3908);
 $4 = ((($temp)) + 4|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($temp)) + 8|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($temp)) + 12|0);
 HEAP32[$6>>2] = 0;
 $7 = ($3|0)!=(0|0);
 $8 = $7 ? $3 : 163772;
 __THREW__ = 0;
 $9 = (invoke_ii(756,($8|0))|0);
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  ___resumeException($26|0);
  // unreachable;
 }
 HEAP32[$6>>2] = $9;
 HEAP32[$5>>2] = $9;
 $12 = (($9) + 1)|0;
 $13 = $12 << 2;
 $14 = (($13) + 8)|0;
 __THREW__ = 0;
 $15 = (invoke_ii(775,($14|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  ___resumeException($26|0);
  // unreachable;
 }
 $18 = ((($15)) + 8|0);
 $19 = $15;
 $20 = $19;
 HEAP32[$20>>2] = $12;
 $21 = (($19) + 4)|0;
 $22 = $21;
 HEAP32[$22>>2] = 0;
 $23 = ($12|0)==(0);
 if (!($23)) {
  _memset(($18|0),0,($13|0))|0;
 }
 HEAP32[$4>>2] = $18;
 __THREW__ = 0;
 (invoke_iii(789,($18|0),($8|0))|0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  ___resumeException($26|0);
  // unreachable;
 }
 HEAP32[$agg$result>>2] = (3908);
 $28 = ((($agg$result)) + 4|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($agg$result)) + 8|0);
 HEAP32[$29>>2] = 0;
 $30 = ((($agg$result)) + 12|0);
 HEAP32[$30>>2] = 0;
 __THREW__ = 0;
 $31 = (invoke_ii(756,(163772|0))|0);
 $32 = __THREW__; __THREW__ = 0;
 $33 = $32&1;
 if ($33) {
  label = 48;
 } else {
  HEAP32[$30>>2] = $31;
  HEAP32[$29>>2] = $31;
  $34 = (($31) + 1)|0;
  $35 = $34 << 2;
  $36 = (($35) + 8)|0;
  __THREW__ = 0;
  $37 = (invoke_ii(775,($36|0))|0);
  $38 = __THREW__; __THREW__ = 0;
  $39 = $38&1;
  if ($39) {
   label = 48;
  } else {
   $40 = ((($37)) + 8|0);
   $41 = $37;
   $42 = $41;
   HEAP32[$42>>2] = $34;
   $43 = (($41) + 4)|0;
   $44 = $43;
   HEAP32[$44>>2] = 0;
   $45 = ($34|0)==(0);
   if (!($45)) {
    _memset(($40|0),0,($35|0))|0;
   }
   HEAP32[$28>>2] = $40;
   __THREW__ = 0;
   (invoke_iii(789,($40|0),(163772|0))|0);
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    label = 48;
   } else {
    $48 = HEAP32[$6>>2]|0;
    $49 = ($48|0)==(0);
    L20: do {
     if ($49) {
      label = 58;
     } else {
      $50 = ((($match)) + 4|0);
      $51 = ((($0)) + 12|0);
      $52 = ((($0)) + 4|0);
      $53 = ((($replacement)) + 12|0);
      $54 = ((($replacement)) + 4|0);
      $55 = ((($match)) + 12|0);
      $56 = ((($1)) + 12|0);
      $57 = ((($1)) + 4|0);
      L22: while(1) {
       $58 = HEAP32[$4>>2]|0;
       $59 = HEAP32[$50>>2]|0;
       __THREW__ = 0;
       $60 = (invoke_iii(807,($58|0),($59|0))|0);
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        label = 49;
        break L20;
       }
       $63 = ($60|0)==(0|0);
       if ($63) {
        label = 58;
        break L20;
       }
       $64 = HEAP32[$4>>2]|0;
       $65 = $60;
       $66 = $64;
       $67 = (($65) - ($66))|0;
       $68 = $67 >> 2;
       $69 = ($68|0)==(-1);
       if ($69) {
        label = 58;
        break L20;
       }
       __THREW__ = 0;
       invoke_viiii(805,($0|0),($temp|0),0,($68|0));
       $70 = __THREW__; __THREW__ = 0;
       $71 = $70&1;
       if ($71) {
        label = 49;
        break L20;
       }
       $72 = HEAP32[$51>>2]|0;
       $73 = HEAP32[$30>>2]|0;
       $74 = (($73) + ($72))|0;
       HEAP32[$30>>2] = $74;
       $75 = HEAP32[$29>>2]|0;
       $76 = ($74>>>0)>($75>>>0);
       do {
        if ($76) {
         $77 = HEAP32[$28>>2]|0;
         HEAP32[$29>>2] = $74;
         $78 = (($74) + 1)|0;
         $79 = $78 << 2;
         $80 = (($79) + 8)|0;
         __THREW__ = 0;
         $81 = (invoke_ii(775,($80|0))|0);
         $82 = __THREW__; __THREW__ = 0;
         $83 = $82&1;
         if ($83) {
          label = 50;
          break L22;
         }
         $84 = ((($81)) + 8|0);
         $85 = $81;
         $86 = $85;
         HEAP32[$86>>2] = $78;
         $87 = (($85) + 4)|0;
         $88 = $87;
         HEAP32[$88>>2] = 0;
         $89 = ($78|0)==(0);
         if (!($89)) {
          _memset(($84|0),0,($79|0))|0;
         }
         HEAP32[$28>>2] = $84;
         __THREW__ = 0;
         (invoke_iii(789,($84|0),($77|0))|0);
         $90 = __THREW__; __THREW__ = 0;
         $91 = $90&1;
         if ($91) {
          label = 50;
          break L22;
         }
         $92 = ($77|0)==(0|0);
         if ($92) {
          break;
         }
         $93 = ((($77)) + -8|0);
         __THREW__ = 0;
         invoke_vi(750,($93|0));
         $94 = __THREW__; __THREW__ = 0;
         $95 = $94&1;
         if ($95) {
          label = 50;
          break L22;
         }
        }
       } while(0);
       $96 = HEAP32[$28>>2]|0;
       $97 = HEAP32[$52>>2]|0;
       __THREW__ = 0;
       (invoke_iii(790,($96|0),($97|0))|0);
       $98 = __THREW__; __THREW__ = 0;
       $99 = $98&1;
       if ($99) {
        label = 50;
        break;
       }
       HEAP32[$0>>2] = (3908);
       $100 = HEAP32[$52>>2]|0;
       $101 = ($100|0)==(0|0);
       if (!($101)) {
        $102 = ((($100)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($102|0));
        $103 = __THREW__; __THREW__ = 0;
        $104 = $103&1;
        if ($104) {
         label = 28;
         break;
        }
       }
       $107 = HEAP32[$53>>2]|0;
       $108 = HEAP32[$30>>2]|0;
       $109 = (($108) + ($107))|0;
       HEAP32[$30>>2] = $109;
       $110 = HEAP32[$29>>2]|0;
       $111 = ($109>>>0)>($110>>>0);
       do {
        if ($111) {
         $112 = HEAP32[$28>>2]|0;
         HEAP32[$29>>2] = $109;
         $113 = (($109) + 1)|0;
         $114 = $113 << 2;
         $115 = (($114) + 8)|0;
         __THREW__ = 0;
         $116 = (invoke_ii(775,($115|0))|0);
         $117 = __THREW__; __THREW__ = 0;
         $118 = $117&1;
         if ($118) {
          label = 49;
          break L20;
         }
         $119 = ((($116)) + 8|0);
         $120 = $116;
         $121 = $120;
         HEAP32[$121>>2] = $113;
         $122 = (($120) + 4)|0;
         $123 = $122;
         HEAP32[$123>>2] = 0;
         $124 = ($113|0)==(0);
         if (!($124)) {
          _memset(($119|0),0,($114|0))|0;
         }
         HEAP32[$28>>2] = $119;
         __THREW__ = 0;
         (invoke_iii(789,($119|0),($112|0))|0);
         $125 = __THREW__; __THREW__ = 0;
         $126 = $125&1;
         if ($126) {
          label = 49;
          break L20;
         }
         $127 = ($112|0)==(0|0);
         if ($127) {
          break;
         }
         $128 = ((($112)) + -8|0);
         __THREW__ = 0;
         invoke_vi(750,($128|0));
         $129 = __THREW__; __THREW__ = 0;
         $130 = $129&1;
         if ($130) {
          label = 49;
          break L20;
         }
        }
       } while(0);
       $131 = HEAP32[$28>>2]|0;
       $132 = HEAP32[$54>>2]|0;
       __THREW__ = 0;
       (invoke_iii(790,($131|0),($132|0))|0);
       $133 = __THREW__; __THREW__ = 0;
       $134 = $133&1;
       if ($134) {
        label = 49;
        break L20;
       }
       $135 = HEAP32[$55>>2]|0;
       $136 = (($135) + ($68))|0;
       $137 = HEAP32[$6>>2]|0;
       __THREW__ = 0;
       invoke_viiii(805,($1|0),($temp|0),($136|0),($137|0));
       $138 = __THREW__; __THREW__ = 0;
       $139 = $138&1;
       if ($139) {
        label = 49;
        break L20;
       }
       $140 = HEAP32[$4>>2]|0;
       $141 = ($140|0)==(0|0);
       if (!($141)) {
        $142 = ((($140)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($142|0));
        $143 = __THREW__; __THREW__ = 0;
        $144 = $143&1;
        if ($144) {
         label = 54;
         break;
        }
       }
       $145 = HEAP32[$56>>2]|0;
       HEAP32[$6>>2] = $145;
       HEAP32[$5>>2] = $145;
       $146 = (($145) + 1)|0;
       $147 = $146 << 2;
       $148 = (($147) + 8)|0;
       __THREW__ = 0;
       $149 = (invoke_ii(775,($148|0))|0);
       $150 = __THREW__; __THREW__ = 0;
       $151 = $150&1;
       if ($151) {
        label = 54;
        break;
       }
       $152 = ((($149)) + 8|0);
       $153 = $149;
       $154 = $153;
       HEAP32[$154>>2] = $146;
       $155 = (($153) + 4)|0;
       $156 = $155;
       HEAP32[$156>>2] = 0;
       $157 = ($146|0)==(0);
       if (!($157)) {
        _memset(($152|0),0,($147|0))|0;
       }
       HEAP32[$4>>2] = $152;
       $158 = HEAP32[$57>>2]|0;
       __THREW__ = 0;
       (invoke_iii(789,($152|0),($158|0))|0);
       $159 = __THREW__; __THREW__ = 0;
       $160 = $159&1;
       if ($160) {
        label = 54;
        break;
       }
       HEAP32[$1>>2] = (3908);
       $161 = HEAP32[$57>>2]|0;
       $162 = ($161|0)==(0|0);
       if (!($162)) {
        $163 = ((($161)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($163|0));
        $164 = __THREW__; __THREW__ = 0;
        $165 = $164&1;
        if ($165) {
         label = 46;
         break;
        }
       }
       $168 = HEAP32[$6>>2]|0;
       $169 = ($168|0)==(0);
       if ($169) {
        label = 58;
        break L20;
       }
      }
      if ((label|0) == 28) {
       $105 = ___cxa_find_matching_catch_3(0|0)|0;
       $106 = tempRet0;
       ___clang_call_terminate($105);
       // unreachable;
      }
      else if ((label|0) == 46) {
       $166 = ___cxa_find_matching_catch_3(0|0)|0;
       $167 = tempRet0;
       ___clang_call_terminate($166);
       // unreachable;
      }
      else if ((label|0) == 50) {
       $174 = ___cxa_find_matching_catch_2()|0;
       $175 = tempRet0;
       HEAP32[$0>>2] = (3908);
       $176 = HEAP32[$52>>2]|0;
       $177 = ($176|0)==(0|0);
       if (!($177)) {
        $178 = ((($176)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($178|0));
        $179 = __THREW__; __THREW__ = 0;
        $180 = $179&1;
        if ($180) {
         $181 = ___cxa_find_matching_catch_3(0|0)|0;
         $182 = tempRet0;
         ___clang_call_terminate($181);
         // unreachable;
        }
       }
       $$0 = $174;$$01 = $175;
       break;
      }
      else if ((label|0) == 54) {
       $183 = ___cxa_find_matching_catch_2()|0;
       $184 = tempRet0;
       HEAP32[$1>>2] = (3908);
       $185 = HEAP32[$57>>2]|0;
       $186 = ($185|0)==(0|0);
       if (!($186)) {
        $187 = ((($185)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($187|0));
        $188 = __THREW__; __THREW__ = 0;
        $189 = $188&1;
        if ($189) {
         $190 = ___cxa_find_matching_catch_3(0|0)|0;
         $191 = tempRet0;
         ___clang_call_terminate($190);
         // unreachable;
        }
       }
       $$0 = $183;$$01 = $184;
       break;
      }
     }
    } while(0);
    do {
     if ((label|0) == 58) {
      $192 = HEAP32[$6>>2]|0;
      $193 = HEAP32[$30>>2]|0;
      $194 = (($193) + ($192))|0;
      HEAP32[$30>>2] = $194;
      $195 = HEAP32[$29>>2]|0;
      $196 = ($194>>>0)>($195>>>0);
      if ($196) {
       $197 = HEAP32[$28>>2]|0;
       HEAP32[$29>>2] = $194;
       $198 = (($194) + 1)|0;
       $199 = $198 << 2;
       $200 = (($199) + 8)|0;
       __THREW__ = 0;
       $201 = (invoke_ii(775,($200|0))|0);
       $202 = __THREW__; __THREW__ = 0;
       $203 = $202&1;
       if ($203) {
        label = 49;
        break;
       }
       $204 = ((($201)) + 8|0);
       $205 = $201;
       $206 = $205;
       HEAP32[$206>>2] = $198;
       $207 = (($205) + 4)|0;
       $208 = $207;
       HEAP32[$208>>2] = 0;
       $209 = ($198|0)==(0);
       if (!($209)) {
        _memset(($204|0),0,($199|0))|0;
       }
       HEAP32[$28>>2] = $204;
       __THREW__ = 0;
       (invoke_iii(789,($204|0),($197|0))|0);
       $210 = __THREW__; __THREW__ = 0;
       $211 = $210&1;
       if ($211) {
        label = 49;
        break;
       }
       $212 = ($197|0)==(0|0);
       if (!($212)) {
        $213 = ((($197)) + -8|0);
        __THREW__ = 0;
        invoke_vi(750,($213|0));
        $214 = __THREW__; __THREW__ = 0;
        $215 = $214&1;
        if ($215) {
         label = 49;
         break;
        }
       }
      }
      $216 = HEAP32[$28>>2]|0;
      $217 = HEAP32[$4>>2]|0;
      __THREW__ = 0;
      (invoke_iii(790,($216|0),($217|0))|0);
      $218 = __THREW__; __THREW__ = 0;
      $219 = $218&1;
      if ($219) {
       label = 49;
      } else {
       HEAP32[$temp>>2] = (3908);
       $220 = HEAP32[$4>>2]|0;
       $221 = ($220|0)==(0|0);
       if ($221) {
        STACKTOP = sp;return;
       }
       $222 = ((($220)) + -8|0);
       __THREW__ = 0;
       invoke_vi(750,($222|0));
       $223 = __THREW__; __THREW__ = 0;
       $224 = $223&1;
       if ($224) {
        $225 = ___cxa_find_matching_catch_3(0|0)|0;
        $226 = tempRet0;
        ___clang_call_terminate($225);
        // unreachable;
       } else {
        STACKTOP = sp;return;
       }
      }
     }
    } while(0);
    if ((label|0) == 49) {
     $172 = ___cxa_find_matching_catch_2()|0;
     $173 = tempRet0;
     $$0 = $172;$$01 = $173;
    }
    HEAP32[$agg$result>>2] = (3908);
    $227 = HEAP32[$28>>2]|0;
    $228 = ($227|0)==(0|0);
    if (!($228)) {
     $229 = ((($227)) + -8|0);
     __THREW__ = 0;
     invoke_vi(750,($229|0));
     $230 = __THREW__; __THREW__ = 0;
     $231 = $230&1;
     if ($231) {
      $232 = ___cxa_find_matching_catch_3(0|0)|0;
      $233 = tempRet0;
      ___clang_call_terminate($232);
      // unreachable;
     }
    }
    $$1 = $$0;$$12 = $$01;
   }
  }
 }
 if ((label|0) == 48) {
  $170 = ___cxa_find_matching_catch_2()|0;
  $171 = tempRet0;
  $$1 = $170;$$12 = $171;
 }
 HEAP32[$temp>>2] = (3908);
 $234 = HEAP32[$4>>2]|0;
 $235 = ($234|0)==(0|0);
 if ($235) {
  ___resumeException($$1|0);
  // unreachable;
 }
 $236 = ((($234)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($236|0));
 $237 = __THREW__; __THREW__ = 0;
 $238 = $237&1;
 if ($238) {
  $239 = ___cxa_find_matching_catch_3(0|0)|0;
  $240 = tempRet0;
  ___clang_call_terminate($239);
  // unreachable;
 } else {
  ___resumeException($$1|0);
  // unreachable;
 }
}
function __ZN9ExitGames6Common7JStringaSERKS1_($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($this|0)==($Rhs|0);
 if ($0) {
  return ($this|0);
 }
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = ((($2)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($4);
 }
 $5 = ((($Rhs)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($this)) + 12|0);
 HEAP32[$7>>2] = $6;
 $8 = ((($this)) + 8|0);
 HEAP32[$8>>2] = $6;
 $9 = (($6) + 1)|0;
 $10 = $9 << 2;
 $11 = (($10) + 8)|0;
 $12 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($11)|0);
 $13 = ((($12)) + 8|0);
 $14 = $12;
 $15 = $14;
 HEAP32[$15>>2] = $9;
 $16 = (($14) + 4)|0;
 $17 = $16;
 HEAP32[$17>>2] = 0;
 $18 = ($9|0)==(0);
 if (!($18)) {
  _memset(($13|0),0,($10|0))|0;
 }
 HEAP32[$1>>2] = $13;
 $19 = ((($Rhs)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 (_wcscpy($13,$20)|0);
 return ($this|0);
}
function __ZNK9ExitGames6Common7JString11lastIndexOfEw($this,$ch) {
 $this = $this|0;
 $ch = $ch|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0$i = -1;
  return ($$0$i|0);
 }
 $3 = ((($this)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($1<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 HEAP32[$5>>2] = 0;
 $7 = (_wcsrchr($4,$ch)|0);
 $8 = HEAP32[$3>>2]|0;
 $9 = (($8) + ($1<<2)|0);
 HEAP32[$9>>2] = $6;
 $10 = ($7|0)==(0|0);
 if ($10) {
  $$0$i = -1;
  return ($$0$i|0);
 }
 $11 = $7;
 $12 = $8;
 $13 = (($11) - ($12))|0;
 $14 = $13 >> 2;
 $$0$i = $14;
 return ($$0$i|0);
}
function __ZNK9ExitGames6Common7JString9substringEjj($agg$result,$this,$beginIndex,$endIndex) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $beginIndex = $beginIndex|0;
 $endIndex = $endIndex|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $beginIndex$endIndex = 0, $endIndex$beginIndex = 0, $outPut = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $outPut = sp;
 $0 = ($beginIndex>>>0)>($endIndex>>>0);
 $beginIndex$endIndex = $0 ? $beginIndex : $endIndex;
 $1 = ((($this)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($beginIndex$endIndex>>>0)>($2>>>0);
 if ($3) {
  HEAP32[$agg$result>>2] = (3908);
  $4 = ((($agg$result)) + 4|0);
  HEAP32[$4>>2] = 0;
  $5 = ((($agg$result)) + 8|0);
  HEAP32[$5>>2] = 0;
  $6 = ((($agg$result)) + 12|0);
  HEAP32[$6>>2] = 0;
  __THREW__ = 0;
  $7 = (invoke_ii(756,(163772|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   ___resumeException($24|0);
   // unreachable;
  }
  HEAP32[$6>>2] = $7;
  HEAP32[$5>>2] = $7;
  $10 = (($7) + 1)|0;
  $11 = $10 << 2;
  $12 = (($11) + 8)|0;
  __THREW__ = 0;
  $13 = (invoke_ii(775,($12|0))|0);
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   ___resumeException($24|0);
   // unreachable;
  }
  $16 = ((($13)) + 8|0);
  $17 = $13;
  $18 = $17;
  HEAP32[$18>>2] = $10;
  $19 = (($17) + 4)|0;
  $20 = $19;
  HEAP32[$20>>2] = 0;
  $21 = ($10|0)==(0);
  if (!($21)) {
   _memset(($16|0),0,($11|0))|0;
  }
  HEAP32[$4>>2] = $16;
  __THREW__ = 0;
  (invoke_iii(789,($16|0),(163772|0))|0);
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   ___resumeException($24|0);
   // unreachable;
  } else {
   STACKTOP = sp;return;
  }
 }
 $endIndex$beginIndex = $0 ? $endIndex : $beginIndex;
 $26 = ((($this)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (($27) + ($beginIndex$endIndex<<2)|0);
 $29 = HEAP32[$28>>2]|0;
 HEAP32[$28>>2] = 0;
 $30 = (($27) + ($endIndex$beginIndex<<2)|0);
 HEAP32[$outPut>>2] = (3908);
 $31 = ((($outPut)) + 4|0);
 HEAP32[$31>>2] = 0;
 $32 = ((($outPut)) + 8|0);
 HEAP32[$32>>2] = 0;
 $33 = ((($outPut)) + 12|0);
 HEAP32[$33>>2] = 0;
 $34 = ($30|0)!=(0|0);
 $35 = $34 ? $30 : 163772;
 __THREW__ = 0;
 $36 = (invoke_ii(756,($35|0))|0);
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  ___resumeException($53|0);
  // unreachable;
 }
 HEAP32[$33>>2] = $36;
 HEAP32[$32>>2] = $36;
 $39 = (($36) + 1)|0;
 $40 = $39 << 2;
 $41 = (($40) + 8)|0;
 __THREW__ = 0;
 $42 = (invoke_ii(775,($41|0))|0);
 $43 = __THREW__; __THREW__ = 0;
 $44 = $43&1;
 if ($44) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  ___resumeException($53|0);
  // unreachable;
 }
 $45 = ((($42)) + 8|0);
 $46 = $42;
 $47 = $46;
 HEAP32[$47>>2] = $39;
 $48 = (($46) + 4)|0;
 $49 = $48;
 HEAP32[$49>>2] = 0;
 $50 = ($39|0)==(0);
 if (!($50)) {
  _memset(($45|0),0,($40|0))|0;
 }
 HEAP32[$31>>2] = $45;
 __THREW__ = 0;
 (invoke_iii(789,($45|0),($35|0))|0);
 $51 = __THREW__; __THREW__ = 0;
 $52 = $51&1;
 if ($52) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  ___resumeException($53|0);
  // unreachable;
 }
 $55 = HEAP32[$26>>2]|0;
 $56 = (($55) + ($beginIndex$endIndex<<2)|0);
 HEAP32[$56>>2] = $29;
 HEAP32[$agg$result>>2] = (3908);
 $57 = ((($agg$result)) + 4|0);
 HEAP32[$57>>2] = 0;
 $58 = ((($agg$result)) + 8|0);
 $59 = ((($agg$result)) + 12|0);
 $60 = HEAP32[$33>>2]|0;
 HEAP32[$59>>2] = $60;
 HEAP32[$58>>2] = $60;
 $61 = (($60) + 1)|0;
 $62 = $61 << 2;
 $63 = (($62) + 8)|0;
 __THREW__ = 0;
 $64 = (invoke_ii(775,($63|0))|0);
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if (!($66)) {
  $67 = ((($64)) + 8|0);
  $68 = $64;
  $69 = $68;
  HEAP32[$69>>2] = $61;
  $70 = (($68) + 4)|0;
  $71 = $70;
  HEAP32[$71>>2] = 0;
  $72 = ($61|0)==(0);
  if (!($72)) {
   _memset(($67|0),0,($62|0))|0;
  }
  HEAP32[$57>>2] = $67;
  $73 = HEAP32[$31>>2]|0;
  __THREW__ = 0;
  (invoke_iii(789,($67|0),($73|0))|0);
  $74 = __THREW__; __THREW__ = 0;
  $75 = $74&1;
  if (!($75)) {
   HEAP32[$outPut>>2] = (3908);
   $76 = HEAP32[$31>>2]|0;
   $77 = ($76|0)==(0|0);
   if (!($77)) {
    $78 = ((($76)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($78|0));
    $79 = __THREW__; __THREW__ = 0;
    $80 = $79&1;
    if ($80) {
     $81 = ___cxa_find_matching_catch_3(0|0)|0;
     $82 = tempRet0;
     ___clang_call_terminate($81);
     // unreachable;
    }
   }
   STACKTOP = sp;return;
  }
 }
 $83 = ___cxa_find_matching_catch_2()|0;
 $84 = tempRet0;
 HEAP32[$outPut>>2] = (3908);
 $85 = HEAP32[$31>>2]|0;
 $86 = ($85|0)==(0|0);
 if ($86) {
  ___resumeException($83|0);
  // unreachable;
 }
 $87 = ((($85)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($87|0));
 $88 = __THREW__; __THREW__ = 0;
 $89 = $88&1;
 if ($89) {
  $90 = ___cxa_find_matching_catch_3(0|0)|0;
  $91 = tempRet0;
  ___clang_call_terminate($90);
  // unreachable;
 } else {
  ___resumeException($83|0);
  // unreachable;
 }
}
function __ZNK9ExitGames6Common7JString9substringEj($agg$result,$this,$beginIndex) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $beginIndex = $beginIndex|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 __ZNK9ExitGames6Common7JString9substringEjj($agg$result,$this,$beginIndex,$1);
 return;
}
function __ZN9ExitGames6Common7JStringC2EPKc($this,$Value) {
 $this = $this|0;
 $Value = $Value|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $0 = sp + 16|0;
 $1 = sp;
 HEAP32[$this>>2] = (3908);
 $2 = ((($this)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($this)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($this)) + 12|0);
 HEAP32[$4>>2] = 0;
 __THREW__ = 0;
 invoke_vii(808,($1|0),($Value|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $43 = ___cxa_find_matching_catch_2()|0;
  $44 = tempRet0;
  $$1 = $44;$$12 = $43;
  ___resumeException($$12|0);
  // unreachable;
 }
 $7 = HEAP32[$1>>2]|0;
 $8 = ((($7)) + 24|0);
 $9 = HEAP32[$8>>2]|0;
 __THREW__ = 0;
 invoke_vii($9|0,($0|0),($1|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 L5: do {
  if ($11) {
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   $$0 = $46;$$01 = $45;
  } else {
   $12 = ($0|0)==($this|0);
   do {
    if ($12) {
     $$pre = ((($0)) + 4|0);
     $$pre$phiZ2D = $$pre;
    } else {
     $13 = HEAP32[$2>>2]|0;
     $14 = ($13|0)==(0|0);
     if ($14) {
      label = 7;
     } else {
      $15 = ((($13)) + -8|0);
      __THREW__ = 0;
      invoke_vi(750,($15|0));
      $16 = __THREW__; __THREW__ = 0;
      $17 = $16&1;
      if (!($17)) {
       label = 7;
      }
     }
     if ((label|0) == 7) {
      $18 = ((($0)) + 12|0);
      $19 = HEAP32[$18>>2]|0;
      HEAP32[$4>>2] = $19;
      HEAP32[$3>>2] = $19;
      $20 = (($19) + 1)|0;
      $21 = $20 << 2;
      $22 = (($21) + 8)|0;
      __THREW__ = 0;
      $23 = (invoke_ii(775,($22|0))|0);
      $24 = __THREW__; __THREW__ = 0;
      $25 = $24&1;
      if (!($25)) {
       $26 = ((($23)) + 8|0);
       $27 = $23;
       $28 = $27;
       HEAP32[$28>>2] = $20;
       $29 = (($27) + 4)|0;
       $30 = $29;
       HEAP32[$30>>2] = 0;
       $31 = ($20|0)==(0);
       if (!($31)) {
        _memset(($26|0),0,($21|0))|0;
       }
       HEAP32[$2>>2] = $26;
       $32 = ((($0)) + 4|0);
       $33 = HEAP32[$32>>2]|0;
       __THREW__ = 0;
       (invoke_iii(789,($26|0),($33|0))|0);
       $34 = __THREW__; __THREW__ = 0;
       $35 = $34&1;
       if (!($35)) {
        $$pre$phiZ2D = $32;
        break;
       }
      }
     }
     $47 = ___cxa_find_matching_catch_2()|0;
     $48 = tempRet0;
     HEAP32[$0>>2] = (3908);
     $49 = ((($0)) + 4|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = ($50|0)==(0|0);
     if (!($51)) {
      $52 = ((($50)) + -8|0);
      __THREW__ = 0;
      invoke_vi(750,($52|0));
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       $55 = ___cxa_find_matching_catch_3(0|0)|0;
       $56 = tempRet0;
       ___clang_call_terminate($55);
       // unreachable;
      }
     }
     $$0 = $48;$$01 = $47;
     break L5;
    }
   } while(0);
   HEAP32[$0>>2] = (3908);
   $36 = HEAP32[$$pre$phiZ2D>>2]|0;
   $37 = ($36|0)==(0|0);
   if ($37) {
    __ZN9ExitGames6Common10UTF8StringD2Ev($1);
    STACKTOP = sp;return;
   }
   $38 = ((($36)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($38|0));
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    $41 = ___cxa_find_matching_catch_3(0|0)|0;
    $42 = tempRet0;
    ___clang_call_terminate($41);
    // unreachable;
   } else {
    __ZN9ExitGames6Common10UTF8StringD2Ev($1);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 __ZN9ExitGames6Common10UTF8StringD2Ev($1);
 $$1 = $$0;$$12 = $$01;
 ___resumeException($$12|0);
 // unreachable;
}
function __ZN9ExitGames6Common10UTF8StringD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4140);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common14BaseCharStringD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 return;
}
function __ZN9ExitGames6Common4BaseD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common10UTF8StringD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4140);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  __ZdlPv($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common10UTF8StringcvPKcEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common10UTF8StringcvNS0_7JStringEEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 24|0);
 $2 = HEAP32[$1>>2]|0;
 FUNCTION_TABLE_vii[$2 & 2047]($agg$result,$this);
 return;
}
function __ZNK9ExitGames6Common10UTF8String21JStringRepresentationEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$lcssa = 0, $$lcssa17 = 0, $$mask$i$i = 0, $$mask$i$i5 = 0, $$mask1$i$i = 0, $$mask1$i$i6 = 0, $$mask2$i$i = 0, $$mask2$i$i7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $count$0$lcssa$i = 0, $count$0$lcssa$i9 = 0, $count$01$i = 0, $count$01$i3 = 0, $i$02$i = 0, $i$02$i2 = 0;
 var $phitmp = 0, $phitmp12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$this>>2]|0;
 $3 = ((($2)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 2047]($this)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $count$0$lcssa$i = 1;
 } else {
  $count$01$i = 0;$i$02$i = 0;
  while(1) {
   $7 = (($1) + ($i$02$i)|0);
   $8 = HEAP8[$7>>0]|0;
   $9 = $8 << 24 >> 24;
   $10 = ($8<<24>>24)<(0);
   if ($10) {
    $$mask$i$i = $9 & 64;
    $11 = ($$mask$i$i|0)==(0);
    if ($11) {
     $16 = 0;
    } else {
     $$mask1$i$i = $9 & 32;
     $12 = ($$mask1$i$i|0)==(0);
     if ($12) {
      $16 = 2;
     } else {
      $$mask2$i$i = $9 >>> 4;
      $13 = $$mask2$i$i & 1;
      $14 = (($13) + 3)|0;
      $16 = $14;
     }
    }
   } else {
    $16 = 1;
   }
   $15 = $16 >>> 2;
   $17 = $16 & 3;
   $18 = ($17|0)!=(0);
   $19 = $18&1;
   $20 = (($15) + ($count$01$i))|0;
   $21 = (($20) + ($19))|0;
   $22 = (($16) + ($i$02$i))|0;
   $23 = ($22>>>0)<($5>>>0);
   if ($23) {
    $count$01$i = $21;$i$02$i = $22;
   } else {
    $$lcssa17 = $21;
    break;
   }
  }
  $phitmp = (($$lcssa17) + 1)|0;
  $count$0$lcssa$i = $phitmp;
 }
 $24 = $count$0$lcssa$i << 2;
 $25 = (($24) + 8)|0;
 $26 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($25)|0);
 $27 = ((($26)) + 8|0);
 $28 = $26;
 $29 = $28;
 HEAP32[$29>>2] = $count$0$lcssa$i;
 $30 = (($28) + 4)|0;
 $31 = $30;
 HEAP32[$31>>2] = 0;
 $32 = ($count$0$lcssa$i|0)==(0);
 if (!($32)) {
  _memset(($27|0),0,($24|0))|0;
 }
 $33 = HEAP32[$0>>2]|0;
 $34 = HEAP32[$this>>2]|0;
 $35 = ((($34)) + 28|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (FUNCTION_TABLE_ii[$36 & 2047]($this)|0);
 $38 = (($37) + 1)|0;
 $39 = HEAP32[$0>>2]|0;
 $40 = HEAP32[$this>>2]|0;
 $41 = ((($40)) + 28|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = (FUNCTION_TABLE_ii[$42 & 2047]($this)|0);
 $44 = ($43|0)==(0);
 if ($44) {
  $count$0$lcssa$i9 = 1;
 } else {
  $count$01$i3 = 0;$i$02$i2 = 0;
  while(1) {
   $45 = (($39) + ($i$02$i2)|0);
   $46 = HEAP8[$45>>0]|0;
   $47 = $46 << 24 >> 24;
   $48 = ($46<<24>>24)<(0);
   if ($48) {
    $$mask$i$i5 = $47 & 64;
    $49 = ($$mask$i$i5|0)==(0);
    if ($49) {
     $54 = 0;
    } else {
     $$mask1$i$i6 = $47 & 32;
     $50 = ($$mask1$i$i6|0)==(0);
     if ($50) {
      $54 = 2;
     } else {
      $$mask2$i$i7 = $47 >>> 4;
      $51 = $$mask2$i$i7 & 1;
      $52 = (($51) + 3)|0;
      $54 = $52;
     }
    }
   } else {
    $54 = 1;
   }
   $53 = $54 >>> 2;
   $55 = $54 & 3;
   $56 = ($55|0)!=(0);
   $57 = $56&1;
   $58 = (($53) + ($count$01$i3))|0;
   $59 = (($58) + ($57))|0;
   $60 = (($54) + ($i$02$i2))|0;
   $61 = ($60>>>0)<($43>>>0);
   if ($61) {
    $count$01$i3 = $59;$i$02$i2 = $60;
   } else {
    $$lcssa = $59;
    break;
   }
  }
  $phitmp12 = (($$lcssa) + 1)|0;
  $count$0$lcssa$i9 = $phitmp12;
 }
 (__ZN9ExitGames6Common10UTF8String13UTF8Converter12UTF82UnicodeEPKhjPwj($33,$38,$27,$count$0$lcssa$i9)|0);
 __ZN9ExitGames6Common7JStringC2EPKw($agg$result,$27);
 __THREW__ = 0;
 invoke_vi(750,($26|0));
 $62 = __THREW__; __THREW__ = 0;
 $63 = $62&1;
 if ($63) {
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($agg$result);
  ___resumeException($64|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNK9ExitGames6Common10UTF8String4sizeEv($this) {
 $this = $this|0;
 var $$mask$i$i = 0, $$mask1$i$i = 0, $$mask2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$02$i = 0, $size$0$lcssa$i = 0;
 var $size$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common14BaseCharString4cstrEv($this)|0);
 $1 = (__ZNK9ExitGames6Common14BaseCharString6lengthEv($this)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  $size$0$lcssa$i = 0;
  return ($size$0$lcssa$i|0);
 } else {
  $i$02$i = 0;$size$01$i = 0;
 }
 while(1) {
  $3 = (($0) + ($size$01$i)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4 << 24 >> 24;
  $6 = ($4<<24>>24)<(0);
  if ($6) {
   $$mask$i$i = $5 & 64;
   $7 = ($$mask$i$i|0)==(0);
   if ($7) {
    $12 = 0;
   } else {
    $$mask1$i$i = $5 & 32;
    $8 = ($$mask1$i$i|0)==(0);
    if ($8) {
     $12 = 2;
    } else {
     $$mask2$i$i = $5 >>> 4;
     $9 = $$mask2$i$i & 1;
     $10 = (($9) + 3)|0;
     $12 = $10;
    }
   }
  } else {
   $12 = 1;
  }
  $11 = (($12) + ($size$01$i))|0;
  $13 = (($i$02$i) + 1)|0;
  $exitcond$i = ($13|0)==($1|0);
  if ($exitcond$i) {
   $size$0$lcssa$i = $11;
   break;
  } else {
   $i$02$i = $13;$size$01$i = $11;
  }
 }
 return ($size$0$lcssa$i|0);
}
function __ZN9ExitGames6Common10UTF8String13UTF8Converter12UTF82UnicodeEPKhjPwj($str,$src_size,$dst,$dst_len) {
 $str = $str|0;
 $src_size = $src_size|0;
 $dst = $dst|0;
 $dst_len = $dst_len|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $size = 0, $tmpSrcSize = 0, $wrptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp + 12|0;
 $wrptr = sp + 8|0;
 $size = sp + 4|0;
 $tmpSrcSize = sp;
 HEAP32[$0>>2] = $str;
 $1 = HEAP32[983]|0;
 $2 = HEAP32[1043]|0;
 $3 = (_iconv_open($1,$2)|0);
 HEAP32[$wrptr>>2] = $dst;
 $4 = $dst_len << 2;
 HEAP32[$size>>2] = $4;
 HEAP32[$tmpSrcSize>>2] = $src_size;
 (_iconv($3,$0,$tmpSrcSize,$wrptr,$size)|0);
 STACKTOP = sp;return ($dst|0);
}
function __ZN9ExitGames6Common10UTF8StringC2EPKc($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (4140);
 __THREW__ = 0;
 (invoke_iii(809,($this|0),($str|0))|0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_2()|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___resumeException($2|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common14BaseCharStringC2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4184);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 return;
}
function __ZN9ExitGames6Common10UTF8StringaSEPKc($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $$mask$i$i = 0, $$mask1$i$i = 0, $$mask2$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$02$i = 0, $len$0$lcssa$i = 0, $len$01$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = ((($2)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($4);
 }
 __ZN9ExitGames6Common10ANSIStringC2EPKc($0,$Rhs);
 __THREW__ = 0;
 $5 = (invoke_ii(810,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  __ZN9ExitGames6Common10ANSIStringD2Ev($0);
  ___resumeException($31|0);
  // unreachable;
 }
 __ZN9ExitGames6Common10ANSIStringD2Ev($0);
 $8 = ($5|0)==(0);
 if ($8) {
  $len$0$lcssa$i = 0;
 } else {
  $i$02$i = 0;$len$01$i = 0;
  while(1) {
   $9 = (($Rhs) + ($i$02$i)|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = $10 << 24 >> 24;
   $12 = ($10<<24>>24)<(0);
   if ($12) {
    $$mask$i$i = $11 & 64;
    $13 = ($$mask$i$i|0)==(0);
    if ($13) {
     $18 = 0;
    } else {
     $$mask1$i$i = $11 & 32;
     $14 = ($$mask1$i$i|0)==(0);
     if ($14) {
      $18 = 2;
     } else {
      $$mask2$i$i = $11 >>> 4;
      $15 = $$mask2$i$i & 1;
      $16 = (($15) + 3)|0;
      $18 = $16;
     }
    }
   } else {
    $18 = 1;
   }
   $17 = (($18) + ($i$02$i))|0;
   $19 = (($len$01$i) + 1)|0;
   $20 = ($17>>>0)<($5>>>0);
   if ($20) {
    $i$02$i = $17;$len$01$i = $19;
   } else {
    $len$0$lcssa$i = $19;
    break;
   }
  }
 }
 $21 = ((($this)) + 8|0);
 HEAP32[$21>>2] = $len$0$lcssa$i;
 $22 = (($5) + 1)|0;
 $23 = (($5) + 9)|0;
 $24 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($23)|0);
 $25 = ((($24)) + 8|0);
 $26 = $24;
 $27 = $26;
 HEAP32[$27>>2] = $22;
 $28 = (($26) + 4)|0;
 $29 = $28;
 HEAP32[$29>>2] = 0;
 $30 = ($22|0)==(0);
 if ($30) {
  HEAP32[$1>>2] = $25;
  _memcpy(($25|0),($Rhs|0),($5|0))|0;
  STACKTOP = sp;return ($this|0);
 }
 _memset(($25|0),0,($22|0))|0;
 HEAP32[$1>>2] = $25;
 _memcpy(($25|0),($Rhs|0),($5|0))|0;
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common10ANSIStringC2EPKc($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (3868);
 __THREW__ = 0;
 (invoke_iii(811,($this|0),($str|0))|0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_2()|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___resumeException($2|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common10ANSIStringaSEPKc($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = ($Rhs|0)==(0|0);
 if (!($1)) {
  $39 = ((($this)) + 4|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ($40|0)==(0|0);
  if (!($41)) {
   $42 = ((($40)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($42);
  }
  $43 = (_strlen($Rhs)|0);
  $44 = ((($this)) + 8|0);
  HEAP32[$44>>2] = $43;
  $45 = (($43) + 1)|0;
  $46 = (($43) + 9)|0;
  $47 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($46)|0);
  $48 = ((($47)) + 8|0);
  $49 = $47;
  $50 = $49;
  HEAP32[$50>>2] = $45;
  $51 = (($49) + 4)|0;
  $52 = $51;
  HEAP32[$52>>2] = 0;
  $53 = ($45|0)==(0);
  if (!($53)) {
   _memset(($48|0),0,($45|0))|0;
  }
  HEAP32[$39>>2] = $48;
  (_strcpy($48,$Rhs)|0);
  $$0 = $this;
  STACKTOP = sp;return ($$0|0);
 }
 __ZN9ExitGames6Common14BaseCharStringC2Ev($0);
 HEAP32[$0>>2] = (3868);
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = ((($3)) + -8|0);
  __THREW__ = 0;
  invoke_vi(750,($5|0));
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if ($7) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
   ___resumeException($21|0);
   // unreachable;
  }
 }
 $8 = ((($0)) + 8|0);
 HEAP32[$8>>2] = 0;
 __THREW__ = 0;
 $9 = (invoke_ii(775,9)|0);
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
  ___resumeException($21|0);
  // unreachable;
 }
 $12 = ((($9)) + 8|0);
 $13 = $9;
 $14 = $13;
 HEAP32[$14>>2] = 1;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 HEAP32[$2>>2] = $12;
 HEAP8[$12>>0] = 0;
 $17 = HEAP32[$2>>2]|0;
 __THREW__ = 0;
 $18 = (invoke_iii(811,($this|0),($17|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  HEAP32[$0>>2] = (3868);
  $32 = HEAP32[$2>>2]|0;
  $33 = ($32|0)==(0|0);
  if ($33) {
   __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
   ___resumeException($30|0);
   // unreachable;
  }
  $34 = ((($32)) + -8|0);
  __THREW__ = 0;
  invoke_vi(750,($34|0));
  $35 = __THREW__; __THREW__ = 0;
  $36 = $35&1;
  if ($36) {
   $37 = ___cxa_find_matching_catch_3(0|0)|0;
   $38 = tempRet0;
   __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
   ___clang_call_terminate($37);
   // unreachable;
  } else {
   __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
   ___resumeException($30|0);
   // unreachable;
  }
 } else {
  HEAP32[$0>>2] = (3868);
  $23 = HEAP32[$2>>2]|0;
  $24 = ($23|0)==(0|0);
  if (!($24)) {
   $25 = ((($23)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($25|0));
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_3(0|0)|0;
    $29 = tempRet0;
    __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
    ___clang_call_terminate($28);
    // unreachable;
   }
  }
  __ZN9ExitGames6Common14BaseCharStringD2Ev($0);
  $$0 = $18;
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common14BaseCharStringD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common4BaseD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __GLOBAL__sub_I_Base_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common6LoggerC2Ei(162472,2);
 (___cxa_atexit((812|0),(162472|0),(___dso_handle|0))|0);
 return;
}
function __GLOBAL__sub_I_ToString_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common7JStringC2EPKw(162480,4240);
 (___cxa_atexit((211|0),(162480|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162496,4260);
 (___cxa_atexit((211|0),(162496|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162512,4284);
 (___cxa_atexit((211|0),(162512|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162528,4308);
 (___cxa_atexit((211|0),(162528|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162544,4332);
 (___cxa_atexit((211|0),(162544|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162560,4360);
 (___cxa_atexit((211|0),(162560|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162576,4376);
 (___cxa_atexit((211|0),(162576|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162592,4396);
 (___cxa_atexit((211|0),(162592|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162608,4416);
 (___cxa_atexit((211|0),(162608|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162624,4440);
 (___cxa_atexit((211|0),(162624|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162640,4476);
 (___cxa_atexit((211|0),(162640|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162656,4516);
 (___cxa_atexit((211|0),(162656|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162672,4540);
 (___cxa_atexit((211|0),(162672|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162688,4568);
 (___cxa_atexit((211|0),(162688|0),(___dso_handle|0))|0);
 __ZN9ExitGames6Common7JStringC2EPKw(162704,4612);
 (___cxa_atexit((211|0),(162704|0),(___dso_handle|0))|0);
 return;
}
function __ZN9ExitGames6Common6LoggerD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Common6LoggerC2Ei($this,$debugLevel) {
 $this = $this|0;
 $debugLevel = $debugLevel|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $debugLevel;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 return;
}
function __ZNK9ExitGames6Common7Helpers26DeSerializerImplementation8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $obj = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $obj = sp;
 $0 = sp + 64|0;
 $1 = sp + 56|0;
 $2 = sp + 40|0;
 __ZN9ExitGames6Common6ObjectC2Ev($obj);
 $3 = ((($this)) + 4|0);
 $4 = ((($this)) + 8|0);
 $i$0 = 0;
 while(1) {
  $5 = HEAP32[$3>>2]|0;
  $6 = (($5) + 1)|0;
  HEAP32[$3>>2] = $6;
  $7 = HEAP32[$4>>2]|0;
  $8 = (($7) + ($6)|0);
  $9 = HEAP8[$8>>0]|0;
  __THREW__ = 0;
  $10 = (invoke_iiii(813,($this|0),($9|0),($obj|0))|0);
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if ($12) {
   label = 8;
   break;
  }
  if (!($10)) {
   label = 12;
   break;
  }
  $13 = ($i$0|0)!=(0);
  $14 = $13 ? 12352 : 163772;
  HEAP32[$1>>2] = $14;
  __THREW__ = 0;
  invoke_viii(783,($2|0),($obj|0),($withTypes|0));
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   label = 8;
   break;
  }
  __THREW__ = 0;
  invoke_viii(814,($0|0),($1|0),($2|0));
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   label = 9;
   break;
  }
  __THREW__ = 0;
  (invoke_iii(760,($retStr|0),($0|0))|0);
  $19 = __THREW__; __THREW__ = 0;
  $20 = $19&1;
  if ($20) {
   label = 10;
   break;
  }
  __ZN9ExitGames6Common7JStringD2Ev($0);
  __ZN9ExitGames6Common7JStringD2Ev($2);
  $21 = (($i$0) + 1)|0;
  $i$0 = $21;
 }
 if ((label|0) == 8) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  $$1 = $23;$$12 = $22;
  __ZN9ExitGames6Common6ObjectD2Ev($obj);
  ___resumeException($$12|0);
  // unreachable;
 }
 else if ((label|0) == 9) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $$0 = $25;$$01 = $24;
 }
 else if ((label|0) == 10) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $$0 = $27;$$01 = $26;
 }
 else if ((label|0) == 12) {
  HEAP32[$3>>2] = -1;
  __ZN9ExitGames6Common6ObjectD2Ev($obj);
  STACKTOP = sp;return ($retStr|0);
 }
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$1 = $$0;$$12 = $$01;
 __ZN9ExitGames6Common6ObjectD2Ev($obj);
 ___resumeException($$12|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation7popTypeEhRNS0_6ObjectE($this,$param,$object) {
 $this = $this|0;
 $param = $param|0;
 $object = $object|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $hash$i = 0, $res$i = 0, $res$i1 = 0, $res$i2 = 0, $res$i3 = 0, $res$i4 = 0, $res$i5 = 0, $res$i6 = 0, $str$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $hash$i = sp + 44|0;
 $res$i6 = sp + 40|0;
 $str$i = sp + 24|0;
 $res$i5 = sp + 91|0;
 $res$i4 = sp + 88|0;
 $res$i3 = sp + 8|0;
 $res$i2 = sp + 90|0;
 $res$i1 = sp + 16|0;
 $res$i = sp;
 $0 = $param&255;
 do {
  switch ($0|0) {
  case 105:  {
   $1 = ((($this)) + 4|0);
   $2 = ((($this)) + 8|0);
   $3 = HEAP32[$1>>2]|0;
   $4 = (($3) + 1)|0;
   HEAP32[$1>>2] = $4;
   $5 = HEAP32[$2>>2]|0;
   $6 = (($5) + ($4)|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = (($3) + 2)|0;
   HEAP32[$1>>2] = $8;
   $9 = (($5) + ($8)|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = (($3) + 3)|0;
   HEAP32[$1>>2] = $11;
   $12 = (($5) + ($11)|0);
   $13 = HEAP8[$12>>0]|0;
   $14 = (($3) + 4)|0;
   HEAP32[$1>>2] = $14;
   $15 = (($5) + ($14)|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = $7&255;
   $18 = $17 << 24;
   $19 = $10&255;
   $20 = $19 << 16;
   $21 = $20 | $18;
   $22 = $13&255;
   $23 = $22 << 8;
   $24 = $21 | $23;
   $25 = $16&255;
   $26 = $24 | $25;
   HEAP32[$res$i6>>2] = $26;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i6,105,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 115:  {
   __ZN9ExitGames6Common7JStringC2Ej($str$i,0);
   __THREW__ = 0;
   invoke_vii(815,($this|0),($str$i|0));
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    ___resumeException($31|0);
    // unreachable;
   }
   __THREW__ = 0;
   invoke_viiiii(816,($object|0),($str$i|0),115,0,1);
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    ___resumeException($31|0);
    // unreachable;
   }
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 104:  {
   __ZN9ExitGames6Common9HashtableC2Ev($hash$i);
   __THREW__ = 0;
   invoke_vii(817,($this|0),($hash$i|0));
   $33 = __THREW__; __THREW__ = 0;
   $34 = $33&1;
   if ($34) {
    $37 = ___cxa_find_matching_catch_2()|0;
    $38 = tempRet0;
    __ZN9ExitGames6Common9HashtableD2Ev($hash$i);
    ___resumeException($37|0);
    // unreachable;
   }
   __THREW__ = 0;
   invoke_viiiii(816,($object|0),($hash$i|0),104,0,1);
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    $37 = ___cxa_find_matching_catch_2()|0;
    $38 = tempRet0;
    __ZN9ExitGames6Common9HashtableD2Ev($hash$i);
    ___resumeException($37|0);
    // unreachable;
   }
   __ZN9ExitGames6Common9HashtableD2Ev($hash$i);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 111:  {
   $39 = ((($this)) + 4|0);
   $40 = HEAP32[$39>>2]|0;
   $41 = (($40) + 1)|0;
   HEAP32[$39>>2] = $41;
   $42 = ((($this)) + 8|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (($43) + ($41)|0);
   $45 = HEAP8[$44>>0]|0;
   $46 = ($45<<24>>24)!=(0);
   $47 = $46&1;
   HEAP8[$res$i5>>0] = $47;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i5,111,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 107:  {
   $48 = ((($this)) + 4|0);
   $49 = ((($this)) + 8|0);
   $50 = HEAP32[$48>>2]|0;
   $51 = (($50) + 1)|0;
   HEAP32[$48>>2] = $51;
   $52 = HEAP32[$49>>2]|0;
   $53 = (($52) + ($51)|0);
   $54 = HEAP8[$53>>0]|0;
   $55 = (($50) + 2)|0;
   HEAP32[$48>>2] = $55;
   $56 = (($52) + ($55)|0);
   $57 = HEAP8[$56>>0]|0;
   $58 = $54&255;
   $59 = ($58 << 8)&65535;
   $60 = $57&255;
   $61 = $59 | $60;
   HEAP16[$res$i4>>1] = $61;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i4,107,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 108:  {
   $62 = ((($this)) + 4|0);
   $63 = ((($this)) + 8|0);
   $64 = HEAP32[$62>>2]|0;
   $65 = (($64) + 1)|0;
   HEAP32[$62>>2] = $65;
   $66 = HEAP32[$63>>2]|0;
   $67 = (($66) + ($65)|0);
   $68 = HEAP8[$67>>0]|0;
   $69 = (($64) + 2)|0;
   HEAP32[$62>>2] = $69;
   $70 = (($66) + ($69)|0);
   $71 = HEAP8[$70>>0]|0;
   $72 = (($64) + 3)|0;
   HEAP32[$62>>2] = $72;
   $73 = (($66) + ($72)|0);
   $74 = HEAP8[$73>>0]|0;
   $75 = (($64) + 4)|0;
   HEAP32[$62>>2] = $75;
   $76 = (($66) + ($75)|0);
   $77 = HEAP8[$76>>0]|0;
   $78 = (($64) + 5)|0;
   HEAP32[$62>>2] = $78;
   $79 = (($66) + ($78)|0);
   $80 = HEAP8[$79>>0]|0;
   $81 = (($64) + 6)|0;
   HEAP32[$62>>2] = $81;
   $82 = (($66) + ($81)|0);
   $83 = HEAP8[$82>>0]|0;
   $84 = (($64) + 7)|0;
   HEAP32[$62>>2] = $84;
   $85 = (($66) + ($84)|0);
   $86 = HEAP8[$85>>0]|0;
   $87 = (($64) + 8)|0;
   HEAP32[$62>>2] = $87;
   $88 = (($66) + ($87)|0);
   $89 = HEAP8[$88>>0]|0;
   $90 = $68&255;
   $91 = (_bitshift64Shl(($90|0),0,56)|0);
   $92 = tempRet0;
   $93 = $71&255;
   $94 = (_bitshift64Shl(($93|0),0,48)|0);
   $95 = tempRet0;
   $96 = $94 | $91;
   $97 = $95 | $92;
   $98 = $74&255;
   $99 = (_bitshift64Shl(($98|0),0,40)|0);
   $100 = tempRet0;
   $101 = $96 | $99;
   $102 = $97 | $100;
   $103 = $77&255;
   $104 = $102 | $103;
   $105 = $80&255;
   $106 = (_bitshift64Shl(($105|0),0,24)|0);
   $107 = tempRet0;
   $108 = $101 | $106;
   $109 = $104 | $107;
   $110 = $83&255;
   $111 = (_bitshift64Shl(($110|0),0,16)|0);
   $112 = tempRet0;
   $113 = $86&255;
   $114 = (_bitshift64Shl(($113|0),0,8)|0);
   $115 = tempRet0;
   $116 = $108 | $111;
   $117 = $109 | $112;
   $118 = $89&255;
   $119 = $116 | $114;
   $120 = $117 | $115;
   $121 = $119 | $118;
   $122 = $res$i3;
   $123 = $122;
   HEAP32[$123>>2] = $121;
   $124 = (($122) + 4)|0;
   $125 = $124;
   HEAP32[$125>>2] = $120;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i3,108,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 98:  {
   $126 = ((($this)) + 4|0);
   $127 = HEAP32[$126>>2]|0;
   $128 = (($127) + 1)|0;
   HEAP32[$126>>2] = $128;
   $129 = ((($this)) + 8|0);
   $130 = HEAP32[$129>>2]|0;
   $131 = (($130) + ($128)|0);
   $132 = HEAP8[$131>>0]|0;
   HEAP8[$res$i2>>0] = $132;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i2,98,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 120:  {
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation12popByteArrayERNS0_6ObjectE($this,$object);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 102:  {
   $133 = ((($this)) + 4|0);
   $134 = ((($this)) + 8|0);
   $135 = HEAP32[$133>>2]|0;
   $136 = (($135) + 1)|0;
   HEAP32[$133>>2] = $136;
   $137 = HEAP32[$134>>2]|0;
   $138 = (($137) + ($136)|0);
   $139 = HEAP8[$138>>0]|0;
   $140 = $139&255;
   $141 = $140 << 24;
   $142 = (($135) + 2)|0;
   HEAP32[$133>>2] = $142;
   $143 = (($137) + ($142)|0);
   $144 = HEAP8[$143>>0]|0;
   $145 = $144&255;
   $146 = $145 << 16;
   $147 = $146 | $141;
   $148 = (($135) + 3)|0;
   HEAP32[$133>>2] = $148;
   $149 = (($137) + ($148)|0);
   $150 = HEAP8[$149>>0]|0;
   $151 = $150&255;
   $152 = $151 << 8;
   $153 = $147 | $152;
   $154 = (($135) + 4)|0;
   HEAP32[$133>>2] = $154;
   $155 = (($137) + ($154)|0);
   $156 = HEAP8[$155>>0]|0;
   $157 = $156&255;
   $158 = $153 | $157;
   HEAP32[$res$i1>>2] = $158;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i1,102,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 100:  {
   $159 = ((($this)) + 4|0);
   $160 = ((($this)) + 8|0);
   $161 = HEAP32[$159>>2]|0;
   $162 = (($161) + 1)|0;
   HEAP32[$159>>2] = $162;
   $163 = HEAP32[$160>>2]|0;
   $164 = (($163) + ($162)|0);
   $165 = HEAP8[$164>>0]|0;
   $166 = $165&255;
   $167 = (_bitshift64Shl(($166|0),0,56)|0);
   $168 = tempRet0;
   $169 = (($161) + 2)|0;
   HEAP32[$159>>2] = $169;
   $170 = (($163) + ($169)|0);
   $171 = HEAP8[$170>>0]|0;
   $172 = $171&255;
   $173 = (_bitshift64Shl(($172|0),0,48)|0);
   $174 = tempRet0;
   $175 = $173 | $167;
   $176 = $174 | $168;
   $177 = (($161) + 3)|0;
   HEAP32[$159>>2] = $177;
   $178 = (($163) + ($177)|0);
   $179 = HEAP8[$178>>0]|0;
   $180 = $179&255;
   $181 = (_bitshift64Shl(($180|0),0,40)|0);
   $182 = tempRet0;
   $183 = $175 | $181;
   $184 = $176 | $182;
   $185 = (($161) + 4)|0;
   HEAP32[$159>>2] = $185;
   $186 = (($163) + ($185)|0);
   $187 = HEAP8[$186>>0]|0;
   $188 = $187&255;
   $189 = $184 | $188;
   $190 = (($161) + 5)|0;
   HEAP32[$159>>2] = $190;
   $191 = (($163) + ($190)|0);
   $192 = HEAP8[$191>>0]|0;
   $193 = $192&255;
   $194 = (_bitshift64Shl(($193|0),0,24)|0);
   $195 = tempRet0;
   $196 = $183 | $194;
   $197 = $189 | $195;
   $198 = (($161) + 6)|0;
   HEAP32[$159>>2] = $198;
   $199 = (($163) + ($198)|0);
   $200 = HEAP8[$199>>0]|0;
   $201 = $200&255;
   $202 = (_bitshift64Shl(($201|0),0,16)|0);
   $203 = tempRet0;
   $204 = (($161) + 7)|0;
   HEAP32[$159>>2] = $204;
   $205 = (($163) + ($204)|0);
   $206 = HEAP8[$205>>0]|0;
   $207 = $206&255;
   $208 = (_bitshift64Shl(($207|0),0,8)|0);
   $209 = tempRet0;
   $210 = $196 | $202;
   $211 = $197 | $203;
   $212 = (($161) + 8)|0;
   HEAP32[$159>>2] = $212;
   $213 = (($163) + ($212)|0);
   $214 = HEAP8[$213>>0]|0;
   $215 = $214&255;
   $216 = $210 | $208;
   $217 = $211 | $209;
   $218 = $216 | $215;
   $219 = $res$i;
   $220 = $219;
   HEAP32[$220>>2] = $218;
   $221 = (($219) + 4)|0;
   $222 = $221;
   HEAP32[$222>>2] = $217;
   __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res$i,100,0,1);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 121:  {
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation8popArrayERNS0_6ObjectE($this,$object);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 122:  {
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation14popObjectArrayERNS0_6ObjectE($this,$object);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 42:  {
   __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,0,42,0,0,0,0);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 99:  {
   $223 = ((($this)) + 4|0);
   $224 = HEAP32[$223>>2]|0;
   $225 = (($224) + 1)|0;
   HEAP32[$223>>2] = $225;
   $226 = ((($this)) + 8|0);
   $227 = HEAP32[$226>>2]|0;
   $228 = (($227) + ($225)|0);
   $229 = HEAP8[$228>>0]|0;
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popCustomHelperERNS0_6ObjectEhsi($this,$object,$229,1,0);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  case 68:  {
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation22popArrayOfDictionariesERNS0_6ObjectEsi($this,$object,1,0);
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
   break;
  }
  default: {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
  }
 } while(0);
 return (0)|0;
}
function __ZNK9ExitGames6Common8ToString8toStringEb($agg$result,$this,$withTypes) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $withTypes = $withTypes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $retStr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $retStr = sp;
 __ZN9ExitGames6Common7JStringC2Ej($retStr,0);
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 __THREW__ = 0;
 $3 = (invoke_iiii($2|0,($this|0),($retStr|0),($withTypes|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  __THREW__ = 0;
  invoke_vii(786,($agg$result|0),($3|0));
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   __ZN9ExitGames6Common7JStringD2Ev($retStr);
   STACKTOP = sp;return;
  }
 }
 $8 = ___cxa_find_matching_catch_2()|0;
 $9 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($retStr);
 ___resumeException($8|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplIPKwEENS0_7JStringERKT_RKS4_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0;
 var $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($0,0);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   $3 = HEAP32[$Lsh>>2]|0;
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($3|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    $6 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if (!($8)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     $11 = (invoke_iii(760,($6|0),($Rsh|0))|0);
     $12 = __THREW__; __THREW__ = 0;
     $13 = $12&1;
     if ($13) {
      break;
     }
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($11|0));
     $14 = __THREW__; __THREW__ = 0;
     $15 = $14&1;
     if ($15) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $10;$eh$lpad$body$indexZ2D = $9;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 $eh$lpad$body$index2Z2D = $17;$eh$lpad$body$indexZ2D = $16;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Common6ObjectD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4640);
 __THREW__ = 0;
 invoke_viii(818,($this|0),(0|0),0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_3(0|0)|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($2);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common6Object7cleanupEPKvj($this,$pData,$recursionDepth) {
 $this = $this|0;
 $pData = $pData|0;
 $recursionDepth = $recursionDepth|0;
 var $$0 = 0, $$not = 0, $$not22 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $i$021 = 0, $or$cond = 0, $or$cond14 = 0, $or$cond15 = 0, $or$cond16 = 0, $or$cond17 = 0, $or$cond18 = 0, $or$cond19 = 0, $or$cond20 = 0, $pData$mux = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = ((($this)) + 28|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(42);
  if (!($5)) {
   $6 = ((($this)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if ($8) {
    label = 4;
   }
  }
 } else {
  label = 4;
 }
 L4: do {
  if ((label|0) == 4) {
   $$not = ($pData|0)!=(0|0);
   $$not22 = $2 ^ 1;
   $brmerge = $$not | $$not22;
   $pData$mux = $$not ? $pData : $1;
   if ($brmerge) {
    $$0 = $pData$mux;
   } else {
    $9 = ((($this)) + 28|0);
    $10 = HEAP8[$9>>0]|0;
    $11 = ($10<<24>>24)==(42);
    if ($11) {
     $$0 = 0;
    } else {
     $12 = ((($this)) + 32|0);
     $13 = HEAP32[$12>>2]|0;
     $14 = ($13|0)==(0);
     if ($14) {
      $15 = ((($this)) + 8|0);
      $$0 = $15;
     } else {
      $$0 = 0;
     }
    }
   }
   $16 = ((($this)) + 32|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)!=(0);
   $19 = (($17) + -1)|0;
   $20 = ($19>>>0)>($recursionDepth>>>0);
   $or$cond = $18 & $20;
   if ($or$cond) {
    $21 = ((($this)) + 24|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ($22|0)==(0|0);
    $24 = ((($this)) + 20|0);
    $25 = $23 ? $24 : $22;
    $26 = (($25) + ($recursionDepth<<1)|0);
    $27 = HEAP16[$26>>1]|0;
    $28 = ($27<<16>>16)>(0);
    if ($28) {
     $29 = (($recursionDepth) + 1)|0;
     $i$021 = 0;
     while(1) {
      $30 = (($$0) + ($i$021<<2)|0);
      $31 = HEAP32[$30>>2]|0;
      __ZN9ExitGames6Common6Object7cleanupEPKvj($this,$31,$29);
      $32 = (($i$021) + 1)|0;
      $33 = HEAP32[$21>>2]|0;
      $34 = ($33|0)==(0|0);
      $35 = $34 ? $24 : $33;
      $36 = (($35) + ($recursionDepth<<1)|0);
      $37 = HEAP16[$36>>1]|0;
      $38 = $37 << 16 >> 16;
      $39 = ($32|0)<($38|0);
      if ($39) {
       $i$021 = $32;
      } else {
       break;
      }
     }
    } else {
     $40 = ($$0|0)==(0|0);
     if ($40) {
      break;
     }
    }
    $41 = ((($$0)) + -8|0);
    __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($41);
    break;
   }
   $42 = ((($this)) + 28|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = $43&255;
   do {
    switch ($44|0) {
    case 98:  {
     $45 = ($17|0)==(0);
     $46 = ($$0|0)==(0|0);
     $or$cond14 = $46 | $45;
     if ($or$cond14) {
      break L4;
     }
     $47 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($47);
     break L4;
     break;
    }
    case 107:  {
     $48 = ($17|0)==(0);
     $49 = ($$0|0)==(0|0);
     $or$cond16 = $49 | $48;
     if ($or$cond16) {
      break L4;
     }
     $50 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($50);
     break L4;
     break;
    }
    case 105:  {
     $51 = ($17|0)==(0);
     $52 = ($$0|0)==(0|0);
     $or$cond17 = $52 | $51;
     if ($or$cond17) {
      break L4;
     }
     $53 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($53);
     break L4;
     break;
    }
    case 108:  {
     $54 = ($17|0)==(0);
     $55 = ($$0|0)==(0|0);
     $or$cond18 = $55 | $54;
     if ($or$cond18) {
      break L4;
     }
     $56 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($56);
     break L4;
     break;
    }
    case 102:  {
     $57 = ($17|0)==(0);
     $58 = ($$0|0)==(0|0);
     $or$cond19 = $58 | $57;
     if ($or$cond19) {
      break L4;
     }
     $59 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($59);
     break L4;
     break;
    }
    case 100:  {
     $60 = ($17|0)==(0);
     $61 = ($$0|0)==(0|0);
     $or$cond20 = $61 | $60;
     if ($or$cond20) {
      break L4;
     }
     $62 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($62);
     break L4;
     break;
    }
    case 111:  {
     $63 = ($17|0)==(0);
     $64 = ($$0|0)==(0|0);
     $or$cond15 = $64 | $63;
     if ($or$cond15) {
      break L4;
     }
     $65 = ((($$0)) + -8|0);
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($65);
     break L4;
     break;
    }
    case 99:  {
     $66 = ($43<<24>>24)==(99);
     if ($66) {
      $67 = ((($this)) + 29|0);
      $68 = HEAP8[$67>>0]|0;
      $69 = $68;
     } else {
      $69 = 0;
     }
     __ZN9ExitGames6Common14CustomTypeBase10freeObjectEPKvh($$0,$69);
     break L4;
     break;
    }
    case 115:  {
     $70 = ($$0|0)==(0|0);
     if ($70) {
      break L4;
     }
     $71 = ((($$0)) + -8|0);
     $72 = $71;
     $73 = $72;
     $74 = HEAP32[$73>>2]|0;
     $75 = (($72) + 4)|0;
     $76 = $75;
     $77 = HEAP32[$76>>2]|0;
     $78 = ($74|0)==(0);
     $79 = ($77|0)==(0);
     $80 = $78 & $79;
     if (!($80)) {
      $81 = $74;$82 = $77;
      while(1) {
       $83 = (_i64Add(($81|0),($82|0),-1,-1)|0);
       $84 = tempRet0;
       $85 = (($$0) + ($83<<4)|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = HEAP32[$86>>2]|0;
       FUNCTION_TABLE_vi[$87 & 2047]($85);
       $88 = ($83|0)==(0);
       $89 = ($84|0)==(0);
       $90 = $88 & $89;
       if ($90) {
        break;
       } else {
        $81 = $83;$82 = $84;
       }
      }
     }
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($71);
     break L4;
     break;
    }
    case 104:  {
     $91 = ($$0|0)==(0|0);
     if ($91) {
      break L4;
     }
     $92 = ((($$0)) + -8|0);
     $93 = $92;
     $94 = $93;
     $95 = HEAP32[$94>>2]|0;
     $96 = (($93) + 4)|0;
     $97 = $96;
     $98 = HEAP32[$97>>2]|0;
     $99 = ($95|0)==(0);
     $100 = ($98|0)==(0);
     $101 = $99 & $100;
     if (!($101)) {
      $102 = $95;$103 = $98;
      while(1) {
       $104 = (_i64Add(($102|0),($103|0),-1,-1)|0);
       $105 = tempRet0;
       $106 = (($$0) + (($104*44)|0)|0);
       $107 = HEAP32[$106>>2]|0;
       $108 = HEAP32[$107>>2]|0;
       FUNCTION_TABLE_vi[$108 & 2047]($106);
       $109 = ($104|0)==(0);
       $110 = ($105|0)==(0);
       $111 = $109 & $110;
       if ($111) {
        break;
       } else {
        $102 = $104;$103 = $105;
       }
      }
     }
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($92);
     break L4;
     break;
    }
    case 122:  {
     $112 = ($$0|0)==(0|0);
     if ($112) {
      break L4;
     }
     $113 = ((($$0)) + -8|0);
     $114 = $113;
     $115 = $114;
     $116 = HEAP32[$115>>2]|0;
     $117 = (($114) + 4)|0;
     $118 = $117;
     $119 = HEAP32[$118>>2]|0;
     $120 = ($116|0)==(0);
     $121 = ($119|0)==(0);
     $122 = $120 & $121;
     if (!($122)) {
      $123 = $116;$124 = $119;
      while(1) {
       $125 = (_i64Add(($123|0),($124|0),-1,-1)|0);
       $126 = tempRet0;
       $127 = (($$0) + (($125*40)|0)|0);
       $128 = HEAP32[$127>>2]|0;
       $129 = HEAP32[$128>>2]|0;
       FUNCTION_TABLE_vi[$129 & 2047]($127);
       $130 = ($125|0)==(0);
       $131 = ($126|0)==(0);
       $132 = $130 & $131;
       if ($132) {
        break;
       } else {
        $123 = $125;$124 = $126;
       }
      }
     }
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($113);
     break L4;
     break;
    }
    case 68:  {
     $133 = ($$0|0)==(0|0);
     if ($133) {
      break L4;
     }
     $134 = ((($$0)) + -8|0);
     $135 = $134;
     $136 = $135;
     $137 = HEAP32[$136>>2]|0;
     $138 = (($135) + 4)|0;
     $139 = $138;
     $140 = HEAP32[$139>>2]|0;
     $141 = ($137|0)==(0);
     $142 = ($140|0)==(0);
     $143 = $141 & $142;
     if (!($143)) {
      $144 = $137;$145 = $140;
      while(1) {
       $146 = (_i64Add(($144|0),($145|0),-1,-1)|0);
       $147 = tempRet0;
       $148 = (($$0) + (($146*52)|0)|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = HEAP32[$149>>2]|0;
       FUNCTION_TABLE_vi[$150 & 2047]($148);
       $151 = ($146|0)==(0);
       $152 = ($147|0)==(0);
       $153 = $151 & $152;
       if ($153) {
        break;
       } else {
        $144 = $146;$145 = $147;
       }
      }
     }
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($134);
     break L4;
     break;
    }
    default: {
     break L4;
    }
    }
   } while(0);
  }
 } while(0);
 $154 = ($recursionDepth|0)==(0);
 if (!($154)) {
  return;
 }
 $155 = ((($this)) + 32|0);
 $156 = HEAP32[$155>>2]|0;
 $157 = ($156>>>0)>(1);
 if ($157) {
  $158 = ((($this)) + 24|0);
  $159 = HEAP32[$158>>2]|0;
  $160 = ($159|0)==(0|0);
  $161 = ((($this)) + 20|0);
  $162 = $160 ? $161 : $159;
  $163 = ($162|0)==(0|0);
  if (!($163)) {
   $164 = ((($162)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($164);
  }
  HEAP32[$155>>2] = 0;
  HEAP32[$161>>2] = 0;
  HEAP32[$158>>2] = 0;
 }
 HEAP32[$0>>2] = 0;
 return;
}
function __ZN9ExitGames6Common14CustomTypeBase10freeObjectEPKvh($pData,$typeCode) {
 $pData = $pData|0;
 $typeCode = $typeCode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  __ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper4freeEPKvh($3,$pData,$typeCode);
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Common7JString4cstrEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common6Logger3logEiPKwS3_bjS3_z($this,$debugLevel,$file,$function,$printBrackets,$line,$dbgMsg,$varargs) {
 $this = $this|0;
 $debugLevel = $debugLevel|0;
 $file = $file|0;
 $function = $function|0;
 $printBrackets = $printBrackets|0;
 $line = $line|0;
 $dbgMsg = $dbgMsg|0;
 $varargs = $varargs|0;
 var $args = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $args = sp;
 HEAP32[$args>>2] = $varargs;
 __ZNK9ExitGames6Common6Logger4vlogEiPKwS3_bjS3_Pi($this,$debugLevel,$file,$function,$printBrackets,$line,$dbgMsg,$args);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper4freeEPKvh($this,$pData,$customTypeCode) {
 $this = $this|0;
 $pData = $pData|0;
 $customTypeCode = $customTypeCode|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 FUNCTION_TABLE_vii[$1 & 2047]($pData,$customTypeCode);
 return;
}
function __ZNK9ExitGames6Common6Logger4vlogEiPKwS3_bjS3_Pi($this,$debugLevel,$file,$function,$printBrackets,$line,$dbgMsg,$args) {
 $this = $this|0;
 $debugLevel = $debugLevel|0;
 $file = $file|0;
 $function = $function|0;
 $printBrackets = $printBrackets|0;
 $line = $line|0;
 $dbgMsg = $dbgMsg|0;
 $args = $args|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$op$op = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0;
 var $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0;
 var $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buffer1 = 0, $currentTime = 0, $dateTime = 0, $dbgLvlStr$0 = 0, $eh$lpad$body$index11Z2D = 0, $eh$lpad$body$indexZ2D = 0, $fileWithPath = 0, $str$i = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, $vararg_ptr7 = 0;
 var $vararg_ptr8 = 0, $vararg_ptr9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(208|0);
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $str$i = sp + 192|0;
 $fileWithPath = sp + 176|0;
 $0 = sp + 160|0;
 $1 = sp + 144|0;
 $2 = sp + 128|0;
 $buffer1 = sp + 112|0;
 $currentTime = sp + 108|0;
 $dateTime = sp + 64|0;
 $3 = sp + 48|0;
 $4 = sp + 32|0;
 $5 = HEAP32[$this>>2]|0;
 $6 = ($5|0)<($debugLevel|0);
 if ($6) {
  STACKTOP = sp;return;
 }
 $7 = ((($this)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  STACKTOP = sp;return;
 }
 $10 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(4104)|0);
 $11 = ((($10)) + 8|0);
 $12 = $10;
 $13 = $12;
 HEAP32[$13>>2] = 1024;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = 0;
 _memset(($11|0),0,4096)|0;
 (__Z12EG_vswprintfPwjPKwPi($11,1024,$dbgMsg,$args)|0);
 $16 = $printBrackets ? 4892 : 163772;
 __ZN9ExitGames6Common7JStringC2EPKw($fileWithPath,$file);
 __THREW__ = 0;
 $17 = (invoke_iii(819,($fileWithPath|0),47)|0);
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 L7: do {
  if (!($19)) {
   $20 = ($17|0)>(0);
   if ($20) {
    __THREW__ = 0;
    $21 = (invoke_iii(819,($fileWithPath|0),47)|0);
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     break;
    }
    $24 = (($21) + 1)|0;
    $37 = $24;
   } else {
    __THREW__ = 0;
    $25 = (invoke_iii(819,($fileWithPath|0),92)|0);
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if ($27) {
     break;
    }
    $28 = ($25|0)>(0);
    if ($28) {
     __THREW__ = 0;
     $29 = (invoke_iii(819,($fileWithPath|0),92)|0);
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if ($31) {
      break;
     }
     $32 = (($29) + 1)|0;
     $37 = $32;
    } else {
     $37 = 0;
    }
   }
   __THREW__ = 0;
   $33 = (invoke_ii(764,($fileWithPath|0))|0);
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if (!($35)) {
    $36 = (($33) + ($37<<2)|0);
    __THREW__ = 0;
    invoke_vii(801,($0|0),($36|0));
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     __THREW__ = 0;
     $40 = (invoke_ii(820,($0|0))|0);
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if ($42) {
      $108 = ___cxa_find_matching_catch_2()|0;
      $109 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($0);
      $$2 = $108;$$23 = $109;
      __ZN9ExitGames6Common7JStringD2Ev($fileWithPath);
      ___resumeException($$2|0);
      // unreachable;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     __THREW__ = 0;
     invoke_vii(801,($1|0),($function|0));
     $43 = __THREW__; __THREW__ = 0;
     $44 = $43&1;
     if (!($44)) {
      __THREW__ = 0;
      $45 = (invoke_ii(820,($1|0))|0);
      $46 = __THREW__; __THREW__ = 0;
      $47 = $46&1;
      do {
       if ($47) {
        label = 36;
       } else {
        __THREW__ = 0;
        invoke_vii(801,($2|0),($16|0));
        $48 = __THREW__; __THREW__ = 0;
        $49 = $48&1;
        if ($49) {
         label = 36;
        } else {
         __THREW__ = 0;
         $50 = (invoke_ii(820,($2|0))|0);
         $51 = __THREW__; __THREW__ = 0;
         $52 = $51&1;
         if ($52) {
          $112 = ___cxa_find_matching_catch_2()|0;
          $113 = tempRet0;
          __ZN9ExitGames6Common7JStringD2Ev($2);
          $$0 = $112;$$01 = $113;
          break;
         }
         $53 = (($50) + ($45))|0;
         __ZN9ExitGames6Common7JStringD2Ev($2);
         __ZN9ExitGames6Common7JStringD2Ev($1);
         __THREW__ = 0;
         invoke_vii(801,($buffer1|0),($function|0));
         $54 = __THREW__; __THREW__ = 0;
         $55 = $54&1;
         if ($55) {
          break L7;
         }
         __THREW__ = 0;
         invoke_vii(787,($str$i|0),0);
         $56 = __THREW__; __THREW__ = 0;
         $57 = $56&1;
         do {
          if ($57) {
           label = 39;
          } else {
           __THREW__ = 0;
           (invoke_iii(759,($str$i|0),($16|0))|0);
           $58 = __THREW__; __THREW__ = 0;
           $59 = $58&1;
           if (!($59)) {
            __THREW__ = 0;
            (invoke_iii(760,($buffer1|0),($str$i|0))|0);
            $60 = __THREW__; __THREW__ = 0;
            $61 = $60&1;
            if (!($61)) {
             __ZN9ExitGames6Common7JStringD2Ev($str$i);
             __THREW__ = 0;
             $64 = (invoke_ii(821,(0|0))|0);
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              label = 39;
              break;
             }
             HEAP32[$currentTime>>2] = $64;
             __THREW__ = 0;
             $67 = (invoke_ii(822,($currentTime|0))|0);
             $68 = __THREW__; __THREW__ = 0;
             $69 = $68&1;
             if ($69) {
              label = 39;
              break;
             }
             dest=$dateTime; src=$67; stop=dest+44|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
             __THREW__ = 0;
             $70 = (invoke_i(823)|0);
             $71 = __THREW__; __THREW__ = 0;
             $72 = $71&1;
             if ($72) {
              label = 39;
              break;
             }
             $73 = (($70|0) % 1000)&-1;
             __THREW__ = 0;
             $74 = (invoke_ii(775,32)|0);
             $75 = __THREW__; __THREW__ = 0;
             $76 = $75&1;
             if ($76) {
              label = 39;
              break;
             }
             $77 = ((($74)) + 8|0);
             $78 = $74;
             $79 = $78;
             HEAP32[$79>>2] = 24;
             $80 = (($78) + 4)|0;
             $81 = $80;
             HEAP32[$81>>2] = 0;
             dest=$77; stop=dest+24|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
             __THREW__ = 0;
             (invoke_iiiii(824,($77|0),24,(145902|0),($dateTime|0))|0);
             $82 = __THREW__; __THREW__ = 0;
             $83 = $82&1;
             if ($83) {
              label = 39;
              break;
             }
             HEAP32[$vararg_buffer>>2] = $77;
             $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
             HEAP32[$vararg_ptr1>>2] = $73;
             (_snprintf($77,24,145920,$vararg_buffer)|0);
             $84 = ($40>>>0)>(30);
             $85 = ($53>>>0)>(60);
             $86 = $85 ? $53 : 60;
             __THREW__ = 0;
             invoke_vii(801,($3|0),($11|0));
             $87 = __THREW__; __THREW__ = 0;
             $88 = $87&1;
             if ($88) {
              label = 39;
              break;
             }
             __THREW__ = 0;
             $89 = (invoke_ii(820,($3|0))|0);
             $90 = __THREW__; __THREW__ = 0;
             $91 = $90&1;
             if ($91) {
              $116 = ___cxa_find_matching_catch_2()|0;
              $117 = tempRet0;
              __ZN9ExitGames6Common7JStringD2Ev($3);
              $$1 = $116;$$12 = $117;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($3);
             $$op$op = (($40) + 56)|0;
             $92 = $84 ? $$op$op : 86;
             $93 = (($92) + ($86))|0;
             $94 = (($93) + ($89))|0;
             $95 = $94 << 2;
             $96 = (($95) + 8)|0;
             __THREW__ = 0;
             $97 = (invoke_ii(775,($96|0))|0);
             $98 = __THREW__; __THREW__ = 0;
             $99 = $98&1;
             if ($99) {
              label = 39;
              break;
             }
             $100 = ((($97)) + 8|0);
             $101 = $97;
             $102 = $101;
             HEAP32[$102>>2] = $94;
             $103 = (($101) + 4)|0;
             $104 = $103;
             HEAP32[$104>>2] = 0;
             $105 = ($94|0)==(0);
             if (!($105)) {
              _memset(($100|0),0,($95|0))|0;
             }
             switch ($debugLevel|0) {
             case 0:  {
              $dbgLvlStr$0 = 5036;
              break;
             }
             case 1:  {
              $dbgLvlStr$0 = 5012;
              break;
             }
             case 2:  {
              $dbgLvlStr$0 = 4980;
              break;
             }
             case 3:  {
              $dbgLvlStr$0 = 4960;
              break;
             }
             case 4:  {
              $dbgLvlStr$0 = 4936;
              break;
             }
             default: {
              $dbgLvlStr$0 = 4904;
             }
             }
             __THREW__ = 0;
             $118 = (invoke_ii(764,($buffer1|0))|0);
             $119 = __THREW__; __THREW__ = 0;
             $120 = $119&1;
             if ($120) {
              label = 39;
              break;
             }
             __THREW__ = 0;
             HEAP32[$vararg_buffer2>>2] = $77;
             $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
             HEAP32[$vararg_ptr5>>2] = $dbgLvlStr$0;
             $vararg_ptr6 = ((($vararg_buffer2)) + 8|0);
             HEAP32[$vararg_ptr6>>2] = $36;
             $vararg_ptr7 = ((($vararg_buffer2)) + 12|0);
             HEAP32[$vararg_ptr7>>2] = $118;
             $vararg_ptr8 = ((($vararg_buffer2)) + 16|0);
             HEAP32[$vararg_ptr8>>2] = $line;
             $vararg_ptr9 = ((($vararg_buffer2)) + 20|0);
             HEAP32[$vararg_ptr9>>2] = $11;
             (invoke_iiiii(825,($100|0),($94|0),(5060|0),($vararg_buffer2|0))|0);
             $121 = __THREW__; __THREW__ = 0;
             $122 = $121&1;
             if ($122) {
              label = 39;
              break;
             }
             __THREW__ = 0;
             invoke_vi(750,($74|0));
             $123 = __THREW__; __THREW__ = 0;
             $124 = $123&1;
             if ($124) {
              label = 39;
              break;
             }
             __THREW__ = 0;
             invoke_vi(750,($10|0));
             $125 = __THREW__; __THREW__ = 0;
             $126 = $125&1;
             if ($126) {
              label = 39;
              break;
             }
             $127 = HEAP32[$7>>2]|0;
             $128 = HEAP32[$127>>2]|0;
             $129 = ((($128)) + 8|0);
             $130 = HEAP32[$129>>2]|0;
             __THREW__ = 0;
             invoke_vii(801,($4|0),($100|0));
             $131 = __THREW__; __THREW__ = 0;
             $132 = $131&1;
             if ($132) {
              label = 39;
              break;
             }
             __THREW__ = 0;
             invoke_viii($130|0,($127|0),($debugLevel|0),($4|0));
             $133 = __THREW__; __THREW__ = 0;
             $134 = $133&1;
             if ($134) {
              $137 = ___cxa_find_matching_catch_2()|0;
              $138 = tempRet0;
              __ZN9ExitGames6Common7JStringD2Ev($4);
              $$1 = $137;$$12 = $138;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($4);
             __THREW__ = 0;
             invoke_vi(750,($97|0));
             $135 = __THREW__; __THREW__ = 0;
             $136 = $135&1;
             if ($136) {
              label = 39;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($buffer1);
             __ZN9ExitGames6Common7JStringD2Ev($fileWithPath);
             STACKTOP = sp;return;
            }
           }
           $62 = ___cxa_find_matching_catch_2()|0;
           $63 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($str$i);
           $eh$lpad$body$index11Z2D = $63;$eh$lpad$body$indexZ2D = $62;
           label = 40;
          }
         } while(0);
         if ((label|0) == 39) {
          $114 = ___cxa_find_matching_catch_2()|0;
          $115 = tempRet0;
          $eh$lpad$body$index11Z2D = $115;$eh$lpad$body$indexZ2D = $114;
          label = 40;
         }
         if ((label|0) == 40) {
          $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index11Z2D;
         }
         __ZN9ExitGames6Common7JStringD2Ev($buffer1);
         $$2 = $$1;$$23 = $$12;
         __ZN9ExitGames6Common7JStringD2Ev($fileWithPath);
         ___resumeException($$2|0);
         // unreachable;
        }
       }
      } while(0);
      if ((label|0) == 36) {
       $110 = ___cxa_find_matching_catch_2()|0;
       $111 = tempRet0;
       $$0 = $110;$$01 = $111;
      }
      __ZN9ExitGames6Common7JStringD2Ev($1);
      $$2 = $$0;$$23 = $$01;
      __ZN9ExitGames6Common7JStringD2Ev($fileWithPath);
      ___resumeException($$2|0);
      // unreachable;
     }
    }
   }
  }
 } while(0);
 $106 = ___cxa_find_matching_catch_2()|0;
 $107 = tempRet0;
 $$2 = $106;$$23 = $107;
 __ZN9ExitGames6Common7JStringD2Ev($fileWithPath);
 ___resumeException($$2|0);
 // unreachable;
}
function __Z12EG_vswprintfPwjPKwPi($wcs,$maxlen,$format,$args) {
 $wcs = $wcs|0;
 $maxlen = $maxlen|0;
 $format = $format|0;
 $args = $args|0;
 var $$0 = 0, $$01 = 0, $$phi$trans$insert = 0, $$phi$trans$insert38 = 0, $$pre = 0, $$pre36 = 0, $$pre39 = 0, $$sum = 0, $$sum10 = 0, $$sum11 = 0, $$sum12 = 0, $$sum13 = 0, $$sum14 = 0, $$sum15 = 0, $$sum16 = 0, $$sum17 = 0, $$sum18 = 0, $$sum19 = 0, $$sum20 = 0, $$sum21 = 0;
 var $$sum22 = 0, $$sum23 = 0, $$sum9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0.0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arglist_current = 0, $arglist_current11 = 0, $arglist_current17 = 0, $arglist_current23 = 0, $arglist_current29 = 0, $arglist_current35 = 0, $arglist_current41 = 0, $arglist_current47 = 0, $arglist_current5 = 0, $arglist_current53 = 0, $arglist_current59 = 0;
 var $arglist_current65 = 0, $arglist_current71 = 0, $arglist_current77 = 0, $arglist_next = 0, $arglist_next12 = 0, $arglist_next18 = 0, $arglist_next24 = 0, $arglist_next30 = 0, $arglist_next36 = 0, $arglist_next42 = 0, $arglist_next48 = 0, $arglist_next54 = 0, $arglist_next6 = 0, $arglist_next60 = 0, $arglist_next66 = 0, $arglist_next72 = 0, $arglist_next78 = 0, $dst$i = 0, $dst$i3 = 0, $eh$lpad$body$index88Z2D = 0;
 var $eh$lpad$body$indexZ2D = 0, $expanded = 0, $expanded100 = 0, $expanded101 = 0, $expanded103 = 0, $expanded104 = 0, $expanded105 = 0, $expanded107 = 0, $expanded108 = 0, $expanded110 = 0, $expanded111 = 0, $expanded112 = 0, $expanded114 = 0, $expanded115 = 0, $expanded117 = 0, $expanded118 = 0, $expanded119 = 0, $expanded121 = 0, $expanded122 = 0, $expanded124 = 0;
 var $expanded125 = 0, $expanded126 = 0, $expanded128 = 0, $expanded129 = 0, $expanded131 = 0, $expanded132 = 0, $expanded133 = 0, $expanded135 = 0, $expanded136 = 0, $expanded138 = 0, $expanded139 = 0, $expanded140 = 0, $expanded142 = 0, $expanded143 = 0, $expanded145 = 0, $expanded146 = 0, $expanded147 = 0, $expanded149 = 0, $expanded150 = 0, $expanded152 = 0;
 var $expanded153 = 0, $expanded154 = 0, $expanded156 = 0, $expanded157 = 0, $expanded159 = 0, $expanded160 = 0, $expanded161 = 0, $expanded163 = 0, $expanded164 = 0, $expanded166 = 0, $expanded167 = 0, $expanded168 = 0, $expanded170 = 0, $expanded171 = 0, $expanded173 = 0, $expanded174 = 0, $expanded175 = 0, $expanded177 = 0, $expanded178 = 0, $expanded180 = 0;
 var $expanded181 = 0, $expanded182 = 0, $expanded184 = 0, $expanded185 = 0, $expanded187 = 0, $expanded188 = 0, $expanded189 = 0, $expanded94 = 0, $expanded96 = 0, $expanded97 = 0, $expanded98 = 0, $formatPart$0$lcssa = 0, $formatPart$032 = 0, $offset$0 = 0, $offset$1 = 0, $offset$2 = 0, $ptr$1 = 0, $ptr2$0 = 0, $ptr2$1 = 0, $ptr2$2 = 0;
 var $ptr3$042 = 0, $ptr3$140 = 0, $resStr = 0, $retVal$0$lcssa = 0, $retVal$031 = 0, $retVal$1 = 0, $retVal$2 = 0, $retVal$3 = 0, $retVal$4 = 0, $strchr = 0, $strchr30 = 0, $strchr33 = 0, $tmpChar$0 = 0, $tmpChar$1 = 0, $tmpChar$2 = 0, $utf8Str = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer13 = 0, $vararg_buffer19 = 0;
 var $vararg_buffer25 = 0, $vararg_buffer31 = 0, $vararg_buffer37 = 0, $vararg_buffer43 = 0, $vararg_buffer49 = 0, $vararg_buffer55 = 0, $vararg_buffer61 = 0, $vararg_buffer67 = 0, $vararg_buffer7 = 0, $vararg_buffer73 = 0, $vararg_buffer79 = 0, $vararg_buffer82 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(192|0);
 $vararg_buffer82 = sp + 120|0;
 $vararg_buffer79 = sp + 112|0;
 $vararg_buffer73 = sp + 104|0;
 $vararg_buffer67 = sp + 96|0;
 $vararg_buffer61 = sp + 88|0;
 $vararg_buffer55 = sp + 80|0;
 $vararg_buffer49 = sp + 72|0;
 $vararg_buffer43 = sp + 64|0;
 $vararg_buffer37 = sp + 56|0;
 $vararg_buffer31 = sp + 48|0;
 $vararg_buffer25 = sp + 40|0;
 $vararg_buffer19 = sp + 32|0;
 $vararg_buffer13 = sp + 24|0;
 $vararg_buffer7 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $dst$i3 = sp + 168|0;
 $dst$i = sp + 156|0;
 $utf8Str = sp + 144|0;
 $resStr = sp + 128|0;
 __ZN9ExitGames6Common10UTF8StringC2EPKw($utf8Str,$format);
 __THREW__ = 0;
 $0 = (invoke_ii(826,($utf8Str|0))|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 L1: do {
  if ($2) {
   label = 9;
  } else {
   $3 = $maxlen << 2;
   $4 = (($3) + 8)|0;
   __THREW__ = 0;
   $5 = (invoke_ii(775,($4|0))|0);
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    label = 9;
   } else {
    $8 = ((($5)) + 8|0);
    $9 = $5;
    $10 = $9;
    HEAP32[$10>>2] = $3;
    $11 = (($9) + 4)|0;
    $12 = $11;
    HEAP32[$12>>2] = 0;
    $13 = ($3|0)==(0);
    if (!($13)) {
     _memset(($8|0),0,($3|0))|0;
    }
    $strchr30 = (_strchr($0,37)|0);
    $14 = ($strchr30|0)==(0|0);
    L7: do {
     if ($14) {
      $formatPart$0$lcssa = $0;$retVal$0$lcssa = 0;
     } else {
      $formatPart$032 = $0;$retVal$031 = 0;$strchr33 = $strchr30;
      L9: while(1) {
       HEAP8[$strchr33>>0] = 0;
       $$sum = (($retVal$031) + 8)|0;
       $15 = (($5) + ($$sum)|0);
       $16 = (($3) - ($retVal$031))|0;
       HEAP32[$vararg_buffer>>2] = $formatPart$032;
       $17 = (_snprintf($15,$16,145928,$vararg_buffer)|0);
       $18 = (($17) + ($retVal$031))|0;
       HEAP8[$strchr33>>0] = 37;
       $19 = ((($strchr33)) + 1|0);
       $20 = (_strpbrk($19,145931)|0);
       $21 = ((($20)) + 1|0);
       $22 = HEAP8[$21>>0]|0;
       HEAP8[$21>>0] = 0;
       $23 = HEAP8[$20>>0]|0;
       $24 = $23 << 24 >> 24;
       do {
        switch ($24|0) {
        case 99:  {
         $$sum22 = (($18) + 8)|0;
         $29 = (($5) + ($$sum22)|0);
         $30 = (($3) - ($18))|0;
         $arglist_current = HEAP32[$args>>2]|0;
         $31 = $arglist_current;
         $32 = ((0) + 4|0);
         $expanded94 = $32;
         $expanded = (($expanded94) - 1)|0;
         $33 = (($31) + ($expanded))|0;
         $34 = ((0) + 4|0);
         $expanded98 = $34;
         $expanded97 = (($expanded98) - 1)|0;
         $expanded96 = $expanded97 ^ -1;
         $35 = $33 & $expanded96;
         $36 = $35;
         $37 = HEAP32[$36>>2]|0;
         $arglist_next = ((($36)) + 4|0);
         HEAP32[$args>>2] = $arglist_next;
         HEAP32[$vararg_buffer1>>2] = $37;
         $38 = (_snprintf($29,$30,$strchr33,$vararg_buffer1)|0);
         $39 = (($38) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $39;$tmpChar$2 = $22;
         break;
        }
        case 88: case 120: case 117: case 111: case 105: case 100:  {
         $$sum21 = (($18) + 8)|0;
         $40 = (($5) + ($$sum21)|0);
         $41 = (($3) - ($18))|0;
         $arglist_current5 = HEAP32[$args>>2]|0;
         $42 = $arglist_current5;
         $43 = ((0) + 4|0);
         $expanded101 = $43;
         $expanded100 = (($expanded101) - 1)|0;
         $44 = (($42) + ($expanded100))|0;
         $45 = ((0) + 4|0);
         $expanded105 = $45;
         $expanded104 = (($expanded105) - 1)|0;
         $expanded103 = $expanded104 ^ -1;
         $46 = $44 & $expanded103;
         $47 = $46;
         $48 = HEAP32[$47>>2]|0;
         $arglist_next6 = ((($47)) + 4|0);
         HEAP32[$args>>2] = $arglist_next6;
         HEAP32[$vararg_buffer7>>2] = $48;
         $49 = (_snprintf($40,$41,$strchr33,$vararg_buffer7)|0);
         $50 = (($49) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $50;$tmpChar$2 = $22;
         break;
        }
        case 71: case 103: case 102: case 69: case 101:  {
         $$sum20 = (($18) + 8)|0;
         $51 = (($5) + ($$sum20)|0);
         $52 = (($3) - ($18))|0;
         $arglist_current11 = HEAP32[$args>>2]|0;
         $53 = $arglist_current11;
         $54 = ((0) + 8|0);
         $expanded108 = $54;
         $expanded107 = (($expanded108) - 1)|0;
         $55 = (($53) + ($expanded107))|0;
         $56 = ((0) + 8|0);
         $expanded112 = $56;
         $expanded111 = (($expanded112) - 1)|0;
         $expanded110 = $expanded111 ^ -1;
         $57 = $55 & $expanded110;
         $58 = $57;
         $59 = +HEAPF64[$58>>3];
         $arglist_next12 = ((($58)) + 8|0);
         HEAP32[$args>>2] = $arglist_next12;
         HEAPF64[$vararg_buffer13>>3] = $59;
         $60 = (_snprintf($51,$52,$strchr33,$vararg_buffer13)|0);
         $61 = (($60) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $61;$tmpChar$2 = $22;
         break;
        }
        case 112:  {
         $$sum19 = (($18) + 8)|0;
         $62 = (($5) + ($$sum19)|0);
         $63 = (($3) - ($18))|0;
         $arglist_current17 = HEAP32[$args>>2]|0;
         $64 = $arglist_current17;
         $65 = ((0) + 4|0);
         $expanded115 = $65;
         $expanded114 = (($expanded115) - 1)|0;
         $66 = (($64) + ($expanded114))|0;
         $67 = ((0) + 4|0);
         $expanded119 = $67;
         $expanded118 = (($expanded119) - 1)|0;
         $expanded117 = $expanded118 ^ -1;
         $68 = $66 & $expanded117;
         $69 = $68;
         $70 = HEAP32[$69>>2]|0;
         $arglist_next18 = ((($69)) + 4|0);
         HEAP32[$args>>2] = $arglist_next18;
         HEAP32[$vararg_buffer19>>2] = $70;
         $71 = (_snprintf($62,$63,$strchr33,$vararg_buffer19)|0);
         $72 = (($71) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $72;$tmpChar$2 = $22;
         break;
        }
        case 110:  {
         $$sum18 = (($18) + 8)|0;
         $73 = (($5) + ($$sum18)|0);
         $74 = (($3) - ($18))|0;
         $arglist_current23 = HEAP32[$args>>2]|0;
         $75 = $arglist_current23;
         $76 = ((0) + 4|0);
         $expanded122 = $76;
         $expanded121 = (($expanded122) - 1)|0;
         $77 = (($75) + ($expanded121))|0;
         $78 = ((0) + 4|0);
         $expanded126 = $78;
         $expanded125 = (($expanded126) - 1)|0;
         $expanded124 = $expanded125 ^ -1;
         $79 = $77 & $expanded124;
         $80 = $79;
         $81 = HEAP32[$80>>2]|0;
         $arglist_next24 = ((($80)) + 4|0);
         HEAP32[$args>>2] = $arglist_next24;
         HEAP32[$vararg_buffer25>>2] = $81;
         $82 = (_snprintf($73,$74,$strchr33,$vararg_buffer25)|0);
         $83 = (($82) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $83;$tmpChar$2 = $22;
         break;
        }
        case 37:  {
         $offset$2 = 0;$ptr$1 = $20;$ptr2$2 = $20;$retVal$4 = $18;$tmpChar$2 = $22;
         break;
        }
        case 115:  {
         $$sum17 = (($18) + 8)|0;
         $84 = (($5) + ($$sum17)|0);
         $85 = (($3) - ($18))|0;
         $arglist_current29 = HEAP32[$args>>2]|0;
         $86 = $arglist_current29;
         $87 = ((0) + 4|0);
         $expanded129 = $87;
         $expanded128 = (($expanded129) - 1)|0;
         $88 = (($86) + ($expanded128))|0;
         $89 = ((0) + 4|0);
         $expanded133 = $89;
         $expanded132 = (($expanded133) - 1)|0;
         $expanded131 = $expanded132 ^ -1;
         $90 = $88 & $expanded131;
         $91 = $90;
         $92 = HEAP32[$91>>2]|0;
         $arglist_next30 = ((($91)) + 4|0);
         HEAP32[$args>>2] = $arglist_next30;
         HEAP32[$vararg_buffer31>>2] = $92;
         $93 = (_snprintf($84,$85,$strchr33,$vararg_buffer31)|0);
         $94 = (($93) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $94;$tmpChar$2 = $22;
         break;
        }
        case 83:  {
         $$sum16 = (($18) + 8)|0;
         $95 = (($5) + ($$sum16)|0);
         $96 = (($3) - ($18))|0;
         $arglist_current35 = HEAP32[$args>>2]|0;
         $97 = $arglist_current35;
         $98 = ((0) + 4|0);
         $expanded136 = $98;
         $expanded135 = (($expanded136) - 1)|0;
         $99 = (($97) + ($expanded135))|0;
         $100 = ((0) + 4|0);
         $expanded140 = $100;
         $expanded139 = (($expanded140) - 1)|0;
         $expanded138 = $expanded139 ^ -1;
         $101 = $99 & $expanded138;
         $102 = $101;
         $103 = HEAP32[$102>>2]|0;
         $arglist_next36 = ((($102)) + 4|0);
         HEAP32[$args>>2] = $arglist_next36;
         __THREW__ = 0;
         invoke_vii(827,($dst$i|0),($103|0));
         $104 = __THREW__; __THREW__ = 0;
         $105 = $104&1;
         if ($105) {
          label = 8;
          break L9;
         }
         __THREW__ = 0;
         $106 = (invoke_ii(826,($dst$i|0))|0);
         $107 = __THREW__; __THREW__ = 0;
         $108 = $107&1;
         if ($108) {
          label = 20;
          break L9;
         }
         HEAP32[$vararg_buffer37>>2] = $106;
         $111 = (_snprintf($95,$96,$strchr33,$vararg_buffer37)|0);
         __ZN9ExitGames6Common10UTF8StringD2Ev($dst$i);
         $112 = (($111) + ($18))|0;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $112;$tmpChar$2 = $22;
         break;
        }
        case 104:  {
         HEAP8[$21>>0] = $22;
         $113 = ((($20)) + 2|0);
         $114 = HEAP8[$113>>0]|0;
         HEAP8[$113>>0] = 0;
         $115 = $22 << 24 >> 24;
         switch ($115|0) {
         case 88: case 120: case 117: case 111: case 105: case 100:  {
          $$sum15 = (($18) + 8)|0;
          $116 = (($5) + ($$sum15)|0);
          $117 = (($3) - ($18))|0;
          $arglist_current41 = HEAP32[$args>>2]|0;
          $118 = $arglist_current41;
          $119 = ((0) + 4|0);
          $expanded143 = $119;
          $expanded142 = (($expanded143) - 1)|0;
          $120 = (($118) + ($expanded142))|0;
          $121 = ((0) + 4|0);
          $expanded147 = $121;
          $expanded146 = (($expanded147) - 1)|0;
          $expanded145 = $expanded146 ^ -1;
          $122 = $120 & $expanded145;
          $123 = $122;
          $124 = HEAP32[$123>>2]|0;
          $arglist_next42 = ((($123)) + 4|0);
          HEAP32[$args>>2] = $arglist_next42;
          HEAP32[$vararg_buffer43>>2] = $124;
          $125 = (_snprintf($116,$117,$strchr33,$vararg_buffer43)|0);
          $126 = (($125) + ($18))|0;
          $offset$0 = 1;$ptr2$0 = $21;$retVal$1 = $126;$tmpChar$0 = $114;
          break;
         }
         case 104:  {
          HEAP8[$113>>0] = $114;
          $127 = ((($20)) + 3|0);
          $128 = HEAP8[$127>>0]|0;
          HEAP8[$127>>0] = 0;
          $$sum14 = (($18) + 8)|0;
          $129 = (($5) + ($$sum14)|0);
          $130 = (($3) - ($18))|0;
          $arglist_current47 = HEAP32[$args>>2]|0;
          $131 = $arglist_current47;
          $132 = ((0) + 4|0);
          $expanded150 = $132;
          $expanded149 = (($expanded150) - 1)|0;
          $133 = (($131) + ($expanded149))|0;
          $134 = ((0) + 4|0);
          $expanded154 = $134;
          $expanded153 = (($expanded154) - 1)|0;
          $expanded152 = $expanded153 ^ -1;
          $135 = $133 & $expanded152;
          $136 = $135;
          $137 = HEAP32[$136>>2]|0;
          $arglist_next48 = ((($136)) + 4|0);
          HEAP32[$args>>2] = $arglist_next48;
          HEAP32[$vararg_buffer49>>2] = $137;
          $138 = (_snprintf($129,$130,$strchr33,$vararg_buffer49)|0);
          $139 = (($138) + ($18))|0;
          HEAP8[$127>>0] = $128;
          $140 = HEAP8[$113>>0]|0;
          $offset$0 = 2;$ptr2$0 = $113;$retVal$1 = $139;$tmpChar$0 = $140;
          break;
         }
         case 83:  {
          HEAP8[$21>>0] = 115;
          $141 = 115;
          label = 26;
          break;
         }
         case 115:  {
          $141 = $22;
          label = 26;
          break;
         }
         default: {
          $offset$0 = 0;$ptr2$0 = $20;$retVal$1 = $18;$tmpChar$0 = $114;
         }
         }
         if ((label|0) == 26) {
          label = 0;
          HEAP8[$113>>0] = $114;
          HEAP8[$20>>0] = $141;
          $142 = ($141<<24>>24)==(0);
          if (!($142)) {
           $ptr3$042 = $20;
           while(1) {
            $143 = ((($ptr3$042)) + 1|0);
            $$phi$trans$insert38 = ((($ptr3$042)) + 2|0);
            $$pre39 = HEAP8[$$phi$trans$insert38>>0]|0;
            HEAP8[$143>>0] = $$pre39;
            $144 = ($$pre39<<24>>24)==(0);
            if ($144) {
             break;
            } else {
             $ptr3$042 = $143;
            }
           }
          }
          HEAP8[$21>>0] = 0;
          $$sum13 = (($18) + 8)|0;
          $145 = (($5) + ($$sum13)|0);
          $146 = (($3) - ($18))|0;
          $arglist_current53 = HEAP32[$args>>2]|0;
          $147 = $arglist_current53;
          $148 = ((0) + 4|0);
          $expanded157 = $148;
          $expanded156 = (($expanded157) - 1)|0;
          $149 = (($147) + ($expanded156))|0;
          $150 = ((0) + 4|0);
          $expanded161 = $150;
          $expanded160 = (($expanded161) - 1)|0;
          $expanded159 = $expanded160 ^ -1;
          $151 = $149 & $expanded159;
          $152 = $151;
          $153 = HEAP32[$152>>2]|0;
          $arglist_next54 = ((($152)) + 4|0);
          HEAP32[$args>>2] = $arglist_next54;
          HEAP32[$vararg_buffer55>>2] = $153;
          $154 = (_snprintf($145,$146,$strchr33,$vararg_buffer55)|0);
          $155 = (($154) + ($18))|0;
          $offset$0 = 1;$ptr2$0 = $20;$retVal$1 = $155;$tmpChar$0 = $114;
         }
         $156 = (2 - ($offset$0))|0;
         $157 = (($ptr2$0) + ($156)|0);
         HEAP8[$157>>0] = $tmpChar$0;
         $158 = (1 - ($offset$0))|0;
         $159 = (($ptr2$0) + ($158)|0);
         $160 = HEAP8[$159>>0]|0;
         $offset$2 = $offset$0;$ptr$1 = $strchr33;$ptr2$2 = $ptr2$0;$retVal$4 = $retVal$1;$tmpChar$2 = $160;
         break;
        }
        case 108:  {
         HEAP8[$21>>0] = $22;
         $161 = ((($20)) + 2|0);
         $162 = HEAP8[$161>>0]|0;
         HEAP8[$161>>0] = 0;
         $163 = $22 << 24 >> 24;
         switch ($163|0) {
         case 88: case 120: case 117: case 111: case 105: case 100:  {
          $$sum12 = (($18) + 8)|0;
          $164 = (($5) + ($$sum12)|0);
          $165 = (($3) - ($18))|0;
          $arglist_current59 = HEAP32[$args>>2]|0;
          $166 = $arglist_current59;
          $167 = ((0) + 4|0);
          $expanded164 = $167;
          $expanded163 = (($expanded164) - 1)|0;
          $168 = (($166) + ($expanded163))|0;
          $169 = ((0) + 4|0);
          $expanded168 = $169;
          $expanded167 = (($expanded168) - 1)|0;
          $expanded166 = $expanded167 ^ -1;
          $170 = $168 & $expanded166;
          $171 = $170;
          $172 = HEAP32[$171>>2]|0;
          $arglist_next60 = ((($171)) + 4|0);
          HEAP32[$args>>2] = $arglist_next60;
          HEAP32[$vararg_buffer61>>2] = $172;
          $173 = (_snprintf($164,$165,$strchr33,$vararg_buffer61)|0);
          $174 = (($173) + ($18))|0;
          $offset$1 = 1;$ptr2$1 = $21;$retVal$2 = $174;$tmpChar$1 = $162;
          break;
         }
         case 108:  {
          HEAP8[$161>>0] = $162;
          $175 = ((($20)) + 3|0);
          $176 = HEAP8[$175>>0]|0;
          HEAP8[$175>>0] = 0;
          $$sum11 = (($18) + 8)|0;
          $177 = (($5) + ($$sum11)|0);
          $178 = (($3) - ($18))|0;
          $arglist_current65 = HEAP32[$args>>2]|0;
          $179 = $arglist_current65;
          $180 = ((0) + 8|0);
          $expanded171 = $180;
          $expanded170 = (($expanded171) - 1)|0;
          $181 = (($179) + ($expanded170))|0;
          $182 = ((0) + 8|0);
          $expanded175 = $182;
          $expanded174 = (($expanded175) - 1)|0;
          $expanded173 = $expanded174 ^ -1;
          $183 = $181 & $expanded173;
          $184 = $183;
          $185 = $184;
          $186 = $185;
          $187 = HEAP32[$186>>2]|0;
          $188 = (($185) + 4)|0;
          $189 = $188;
          $190 = HEAP32[$189>>2]|0;
          $arglist_next66 = ((($184)) + 8|0);
          HEAP32[$args>>2] = $arglist_next66;
          $191 = $vararg_buffer67;
          $192 = $191;
          HEAP32[$192>>2] = $187;
          $193 = (($191) + 4)|0;
          $194 = $193;
          HEAP32[$194>>2] = $190;
          $195 = (_snprintf($177,$178,$strchr33,$vararg_buffer67)|0);
          $196 = (($195) + ($18))|0;
          HEAP8[$175>>0] = $176;
          $197 = HEAP8[$161>>0]|0;
          $offset$1 = 2;$ptr2$1 = $161;$retVal$2 = $196;$tmpChar$1 = $197;
          break;
         }
         case 83:  {
          HEAP8[$21>>0] = 115;
          $198 = 115;
          label = 34;
          break;
         }
         case 115:  {
          $198 = $22;
          label = 34;
          break;
         }
         default: {
          $offset$1 = 0;$ptr2$1 = $20;$retVal$2 = $18;$tmpChar$1 = $162;
         }
         }
         if ((label|0) == 34) {
          label = 0;
          HEAP8[$161>>0] = $162;
          HEAP8[$20>>0] = $198;
          $199 = ($198<<24>>24)==(0);
          if (!($199)) {
           $ptr3$140 = $20;
           while(1) {
            $200 = ((($ptr3$140)) + 1|0);
            $$phi$trans$insert = ((($ptr3$140)) + 2|0);
            $$pre36 = HEAP8[$$phi$trans$insert>>0]|0;
            HEAP8[$200>>0] = $$pre36;
            $201 = ($$pre36<<24>>24)==(0);
            if ($201) {
             break;
            } else {
             $ptr3$140 = $200;
            }
           }
          }
          HEAP8[$21>>0] = 0;
          $$sum10 = (($18) + 8)|0;
          $202 = (($5) + ($$sum10)|0);
          $203 = (($3) - ($18))|0;
          $arglist_current71 = HEAP32[$args>>2]|0;
          $204 = $arglist_current71;
          $205 = ((0) + 4|0);
          $expanded178 = $205;
          $expanded177 = (($expanded178) - 1)|0;
          $206 = (($204) + ($expanded177))|0;
          $207 = ((0) + 4|0);
          $expanded182 = $207;
          $expanded181 = (($expanded182) - 1)|0;
          $expanded180 = $expanded181 ^ -1;
          $208 = $206 & $expanded180;
          $209 = $208;
          $210 = HEAP32[$209>>2]|0;
          $arglist_next72 = ((($209)) + 4|0);
          HEAP32[$args>>2] = $arglist_next72;
          __THREW__ = 0;
          invoke_vii(827,($dst$i3|0),($210|0));
          $211 = __THREW__; __THREW__ = 0;
          $212 = $211&1;
          if ($212) {
           label = 8;
           break L9;
          }
          __THREW__ = 0;
          $213 = (invoke_ii(826,($dst$i3|0))|0);
          $214 = __THREW__; __THREW__ = 0;
          $215 = $214&1;
          if ($215) {
           label = 38;
           break L9;
          }
          HEAP32[$vararg_buffer73>>2] = $213;
          $218 = (_snprintf($202,$203,$strchr33,$vararg_buffer73)|0);
          __ZN9ExitGames6Common10UTF8StringD2Ev($dst$i3);
          $219 = (($218) + ($18))|0;
          $offset$1 = 1;$ptr2$1 = $20;$retVal$2 = $219;$tmpChar$1 = $162;
         }
         $220 = (2 - ($offset$1))|0;
         $221 = (($ptr2$1) + ($220)|0);
         HEAP8[$221>>0] = $tmpChar$1;
         $222 = (1 - ($offset$1))|0;
         $223 = (($ptr2$1) + ($222)|0);
         $224 = HEAP8[$223>>0]|0;
         $offset$2 = $offset$1;$ptr$1 = $strchr33;$ptr2$2 = $ptr2$1;$retVal$4 = $retVal$2;$tmpChar$2 = $224;
         break;
        }
        case 76:  {
         HEAP8[$21>>0] = $22;
         $225 = ((($20)) + 2|0);
         $226 = HEAP8[$225>>0]|0;
         HEAP8[$225>>0] = 0;
         $227 = $22 << 24 >> 24;
         switch ($227|0) {
         case 71: case 103: case 102: case 69: case 101:  {
          $$sum9 = (($18) + 8)|0;
          $228 = (($5) + ($$sum9)|0);
          $229 = (($3) - ($18))|0;
          $arglist_current77 = HEAP32[$args>>2]|0;
          $230 = $arglist_current77;
          $231 = ((0) + 8|0);
          $expanded185 = $231;
          $expanded184 = (($expanded185) - 1)|0;
          $232 = (($230) + ($expanded184))|0;
          $233 = ((0) + 8|0);
          $expanded189 = $233;
          $expanded188 = (($expanded189) - 1)|0;
          $expanded187 = $expanded188 ^ -1;
          $234 = $232 & $expanded187;
          $235 = $234;
          $236 = +HEAPF64[$235>>3];
          $arglist_next78 = ((($235)) + 8|0);
          HEAP32[$args>>2] = $arglist_next78;
          HEAPF64[$vararg_buffer79>>3] = $236;
          $237 = (_snprintf($228,$229,$strchr33,$vararg_buffer79)|0);
          $238 = (($237) + ($18))|0;
          $$pre = HEAP8[$21>>0]|0;
          $265 = $$pre;$retVal$3 = $238;
          break;
         }
         default: {
          $265 = $22;$retVal$3 = $18;
         }
         }
         HEAP8[$225>>0] = $226;
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $retVal$3;$tmpChar$2 = $265;
         break;
        }
        default: {
         $offset$2 = 0;$ptr$1 = $strchr33;$ptr2$2 = $20;$retVal$4 = $18;$tmpChar$2 = $22;
        }
        }
       } while(0);
       $239 = (1 - ($offset$2))|0;
       $240 = (($ptr2$2) + ($239)|0);
       HEAP8[$240>>0] = $tmpChar$2;
       $241 = ((($ptr2$2)) + 1|0);
       $242 = ((($ptr$1)) + 1|0);
       $strchr = (_strchr($242,37)|0);
       $243 = ($strchr|0)==(0|0);
       if ($243) {
        $formatPart$0$lcssa = $241;$retVal$0$lcssa = $retVal$4;
        break L7;
       } else {
        $formatPart$032 = $241;$retVal$031 = $retVal$4;$strchr33 = $strchr;
       }
      }
      if ((label|0) == 8) {
       $25 = ___cxa_find_matching_catch_2()|0;
       $26 = tempRet0;
       $eh$lpad$body$index88Z2D = $26;$eh$lpad$body$indexZ2D = $25;
       break L1;
      }
      else if ((label|0) == 20) {
       $109 = ___cxa_find_matching_catch_2()|0;
       $110 = tempRet0;
       __ZN9ExitGames6Common10UTF8StringD2Ev($dst$i);
       $eh$lpad$body$index88Z2D = $110;$eh$lpad$body$indexZ2D = $109;
       break L1;
      }
      else if ((label|0) == 38) {
       $216 = ___cxa_find_matching_catch_2()|0;
       $217 = tempRet0;
       __ZN9ExitGames6Common10UTF8StringD2Ev($dst$i3);
       $eh$lpad$body$index88Z2D = $217;$eh$lpad$body$indexZ2D = $216;
       break L1;
      }
     }
    } while(0);
    $$sum23 = (($retVal$0$lcssa) + 8)|0;
    $244 = (($5) + ($$sum23)|0);
    $245 = (($3) - ($retVal$0$lcssa))|0;
    HEAP32[$vararg_buffer82>>2] = $formatPart$0$lcssa;
    $246 = (_snprintf($244,$245,145928,$vararg_buffer82)|0);
    $247 = (($246) + ($retVal$0$lcssa))|0;
    __THREW__ = 0;
    invoke_vii(739,($resStr|0),($8|0));
    $248 = __THREW__; __THREW__ = 0;
    $249 = $248&1;
    if ($249) {
     label = 9;
    } else {
     __THREW__ = 0;
     $250 = (invoke_ii(764,($resStr|0))|0);
     $251 = __THREW__; __THREW__ = 0;
     $252 = $251&1;
     if (!($252)) {
      __THREW__ = 0;
      $253 = (invoke_ii(820,($resStr|0))|0);
      $254 = __THREW__; __THREW__ = 0;
      $255 = $254&1;
      if (!($255)) {
       $256 = $253 << 2;
       _memcpy(($wcs|0),($250|0),($256|0))|0;
       __THREW__ = 0;
       $257 = (invoke_ii(820,($resStr|0))|0);
       $258 = __THREW__; __THREW__ = 0;
       $259 = $258&1;
       if (!($259)) {
        $260 = (($wcs) + ($257<<2)|0);
        HEAP32[$260>>2] = 0;
        __THREW__ = 0;
        invoke_vi(750,($5|0));
        $261 = __THREW__; __THREW__ = 0;
        $262 = $261&1;
        if (!($262)) {
         __ZN9ExitGames6Common7JStringD2Ev($resStr);
         __ZN9ExitGames6Common10UTF8StringD2Ev($utf8Str);
         STACKTOP = sp;return ($247|0);
        }
       }
      }
     }
     $263 = ___cxa_find_matching_catch_2()|0;
     $264 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($resStr);
     $$0 = $263;$$01 = $264;
     __ZN9ExitGames6Common10UTF8StringD2Ev($utf8Str);
     ___resumeException($$0|0);
     // unreachable;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  $eh$lpad$body$index88Z2D = $28;$eh$lpad$body$indexZ2D = $27;
 }
 $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index88Z2D;
 __ZN9ExitGames6Common10UTF8StringD2Ev($utf8Str);
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZNK9ExitGames6Common7JString11lastIndexOfEc($this,$ch) {
 $this = $this|0;
 $ch = $ch|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + -1)|0;
 $3 = (__ZNK9ExitGames6Common7JString11lastIndexOfEcj($this,$ch,$2)|0);
 return ($3|0);
}
function __ZNK9ExitGames6Common7JString6lengthEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __Z11getTimeUnixv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_clock()|0);
 $1 = (($0|0) / 1000)&-1;
 return ($1|0);
}
function __Z11EG_swprintfPwjPKwz($wcs,$maxlen,$format,$varargs) {
 $wcs = $wcs|0;
 $maxlen = $maxlen|0;
 $format = $format|0;
 $varargs = $varargs|0;
 var $0 = 0, $argptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $argptr = sp;
 HEAP32[$argptr>>2] = $varargs;
 $0 = (__Z12EG_vswprintfPwjPKwPi($wcs,$maxlen,$format,$argptr)|0);
 STACKTOP = sp;return ($0|0);
}
function __ZNK9ExitGames6Common7JString11lastIndexOfEcj($this,$ch,$fromIndex) {
 $this = $this|0;
 $ch = $ch|0;
 $fromIndex = $fromIndex|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0;
 var $s = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $buf = sp + 16|0;
 $s = sp;
 HEAP8[$buf>>0] = $ch;
 $0 = ((($buf)) + 1|0);
 HEAP8[$0>>0] = 0;
 __ZN9ExitGames6Common7JStringC2EPKc($s,$buf);
 $1 = ((($s)) + 4|0);
 $2 = ((($this)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3>>>0)>($fromIndex>>>0);
 do {
  if ($4) {
   $5 = HEAP32[$1>>2]|0;
   $6 = HEAP32[$5>>2]|0;
   $7 = (($fromIndex) + 1)|0;
   $8 = ((($this)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (($9) + ($7<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   HEAP32[$10>>2] = 0;
   __THREW__ = 0;
   $12 = (invoke_iii(828,($9|0),($6|0))|0);
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if (!($14)) {
    $15 = HEAP32[$8>>2]|0;
    $16 = (($15) + ($7<<2)|0);
    HEAP32[$16>>2] = $11;
    $17 = ($12|0)==(0|0);
    if ($17) {
     $$0$i = -1;
     break;
    }
    $18 = $12;
    $19 = $15;
    $20 = (($18) - ($19))|0;
    $21 = $20 >> 2;
    $$0$i = $21;
    break;
   }
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   HEAP32[$s>>2] = (3908);
   $31 = HEAP32[$1>>2]|0;
   $32 = ($31|0)==(0|0);
   if ($32) {
    ___resumeException($29|0);
    // unreachable;
   }
   $33 = ((($31)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($33|0));
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if ($35) {
    $36 = ___cxa_find_matching_catch_3(0|0)|0;
    $37 = tempRet0;
    ___clang_call_terminate($36);
    // unreachable;
   } else {
    ___resumeException($29|0);
    // unreachable;
   }
  } else {
   $$0$i = -1;
  }
 } while(0);
 HEAP32[$s>>2] = (3908);
 $22 = HEAP32[$1>>2]|0;
 $23 = ($22|0)==(0|0);
 if ($23) {
  STACKTOP = sp;return ($$0$i|0);
 }
 $24 = ((($22)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($24|0));
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27);
  // unreachable;
 } else {
  STACKTOP = sp;return ($$0$i|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common10UTF8StringC2EPKw($this,$wstr) {
 $this = $this|0;
 $wstr = $wstr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (4140);
 __THREW__ = 0;
 invoke_vii(801,($0|0),($wstr|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  $eh$lpad$body$index2Z2D = $8;$eh$lpad$body$indexZ2D = $7;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___resumeException($eh$lpad$body$indexZ2D|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(829,($this|0),($0|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  STACKTOP = sp;return;
 }
 $5 = ___cxa_find_matching_catch_2()|0;
 $6 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $eh$lpad$body$index2Z2D = $6;$eh$lpad$body$indexZ2D = $5;
 __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Common10UTF8StringaSERKNS0_7JStringE($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $$lcssa13 = 0, $$mask$i$i = 0, $$mask$i$i4 = 0, $$mask1$i$i = 0, $$mask1$i$i5 = 0, $$mask2$i$i = 0, $$mask2$i$i6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i = 0, $i$02$i = 0, $i$02$i2 = 0, $len$0$lcssa$i = 0, $len$01$i = 0;
 var $size$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($1)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($3);
 }
 $4 = (__ZNK9ExitGames6Common7JString4cstrEv($Rhs)|0);
 $5 = (__ZNK9ExitGames6Common7JString6lengthEv($Rhs)|0);
 $6 = (__ZN9ExitGames6Common10UTF8String13UTF8Converter16SizeOfWStrAsUTF8EPKwj($4,$5)|0);
 $7 = (($6) + 1)|0;
 $8 = (($6) + 9)|0;
 $9 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($8)|0);
 $10 = ((($9)) + 8|0);
 $11 = $9;
 $12 = $11;
 HEAP32[$12>>2] = $7;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = 0;
 $15 = ($7|0)==(0);
 if (!($15)) {
  _memset(($10|0),0,($7|0))|0;
 }
 HEAP32[$0>>2] = $10;
 $16 = (__ZNK9ExitGames6Common7JString4cstrEv($Rhs)|0);
 $17 = (__ZNK9ExitGames6Common7JString6lengthEv($Rhs)|0);
 $18 = (($17) + 1)|0;
 $19 = HEAP32[$0>>2]|0;
 (__ZN9ExitGames6Common10UTF8String13UTF8Converter12Unicode2UTF8EPKwjPcj($16,$18,$19,$7)|0);
 $20 = HEAP32[$0>>2]|0;
 $21 = (__ZNK9ExitGames6Common7JString6lengthEv($Rhs)|0);
 $22 = ($21|0)==(0);
 if ($22) {
  $len$0$lcssa$i = 0;
  $47 = ((($this)) + 8|0);
  HEAP32[$47>>2] = $len$0$lcssa$i;
  return ($this|0);
 } else {
  $i$02$i = 0;$size$01$i = 0;
 }
 while(1) {
  $23 = (($20) + ($size$01$i)|0);
  $24 = HEAP8[$23>>0]|0;
  $25 = $24 << 24 >> 24;
  $26 = ($24<<24>>24)<(0);
  if ($26) {
   $$mask$i$i = $25 & 64;
   $27 = ($$mask$i$i|0)==(0);
   if ($27) {
    $32 = 0;
   } else {
    $$mask1$i$i = $25 & 32;
    $28 = ($$mask1$i$i|0)==(0);
    if ($28) {
     $32 = 2;
    } else {
     $$mask2$i$i = $25 >>> 4;
     $29 = $$mask2$i$i & 1;
     $30 = (($29) + 3)|0;
     $32 = $30;
    }
   }
  } else {
   $32 = 1;
  }
  $31 = (($32) + ($size$01$i))|0;
  $33 = (($i$02$i) + 1)|0;
  $exitcond$i = ($33|0)==($21|0);
  if ($exitcond$i) {
   $$lcssa13 = $31;
   break;
  } else {
   $i$02$i = $33;$size$01$i = $31;
  }
 }
 $34 = ($$lcssa13|0)==(0);
 if ($34) {
  $len$0$lcssa$i = 0;
  $47 = ((($this)) + 8|0);
  HEAP32[$47>>2] = $len$0$lcssa$i;
  return ($this|0);
 } else {
  $i$02$i2 = 0;$len$01$i = 0;
 }
 while(1) {
  $35 = (($20) + ($i$02$i2)|0);
  $36 = HEAP8[$35>>0]|0;
  $37 = $36 << 24 >> 24;
  $38 = ($36<<24>>24)<(0);
  if ($38) {
   $$mask$i$i4 = $37 & 64;
   $39 = ($$mask$i$i4|0)==(0);
   if ($39) {
    $44 = 0;
   } else {
    $$mask1$i$i5 = $37 & 32;
    $40 = ($$mask1$i$i5|0)==(0);
    if ($40) {
     $44 = 2;
    } else {
     $$mask2$i$i6 = $37 >>> 4;
     $41 = $$mask2$i$i6 & 1;
     $42 = (($41) + 3)|0;
     $44 = $42;
    }
   }
  } else {
   $44 = 1;
  }
  $43 = (($44) + ($i$02$i2))|0;
  $45 = (($len$01$i) + 1)|0;
  $46 = ($43>>>0)<($$lcssa13>>>0);
  if ($46) {
   $i$02$i2 = $43;$len$01$i = $45;
  } else {
   $len$0$lcssa$i = $45;
   break;
  }
 }
 $47 = ((($this)) + 8|0);
 HEAP32[$47>>2] = $len$0$lcssa$i;
 return ($this|0);
}
function __ZN9ExitGames6Common10UTF8String13UTF8Converter16SizeOfWStrAsUTF8EPKwj($wstr,$len) {
 $wstr = $wstr|0;
 $len = $len|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $dst_size = 0, $size = 0, $strptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $0 = sp + 24|0;
 $dst_size = sp + 20|0;
 $strptr = sp + 16|0;
 $size = sp + 12|0;
 $1 = sp;
 HEAP32[$0>>2] = $wstr;
 $2 = HEAP32[1043]|0;
 $3 = HEAP32[983]|0;
 $4 = (_iconv_open($2,$3)|0);
 $5 = $len << 2;
 HEAP32[$dst_size>>2] = $5;
 $6 = (($5) + 4)|0;
 $7 = (($5) + 12)|0;
 $8 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($7)|0);
 $9 = ((($8)) + 8|0);
 $10 = $8;
 $11 = $10;
 HEAP32[$11>>2] = $6;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = 0;
 $14 = ($6|0)==(0);
 if (!($14)) {
  _memset(($9|0),0,($6|0))|0;
 }
 HEAP32[$strptr>>2] = $9;
 HEAP32[$size>>2] = $5;
 (_iconv($4,$0,$size,$strptr,$dst_size)|0);
 $15 = HEAP32[$strptr>>2]|0;
 HEAP8[$15>>0] = 0;
 __ZN9ExitGames6Common10ANSIStringC2EPKc($1,$9);
 __THREW__ = 0;
 $16 = (invoke_ii(810,($1|0))|0);
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN9ExitGames6Common10ANSIStringD2Ev($1);
  ___resumeException($19|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10ANSIStringD2Ev($1);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($8);
  STACKTOP = sp;return ($16|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common10UTF8String13UTF8Converter12Unicode2UTF8EPKwjPcj($wstr,$src_len,$dst,$dst_size) {
 $wstr = $wstr|0;
 $src_len = $src_len|0;
 $dst = $dst|0;
 $dst_size = $dst_size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $size = 0, $strptr = 0, $tmpDstSize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp + 12|0;
 $strptr = sp + 8|0;
 $size = sp + 4|0;
 $tmpDstSize = sp;
 HEAP32[$0>>2] = $wstr;
 $1 = HEAP32[1043]|0;
 $2 = HEAP32[983]|0;
 $3 = (_iconv_open($1,$2)|0);
 HEAP32[$strptr>>2] = $dst;
 $4 = $src_len << 2;
 HEAP32[$size>>2] = $4;
 HEAP32[$tmpDstSize>>2] = $dst_size;
 (_iconv($3,$0,$size,$strptr,$tmpDstSize)|0);
 STACKTOP = sp;return ($dst|0);
}
function __ZN9ExitGames6Common6ObjectD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4640);
 __THREW__ = 0;
 invoke_viii(818,($this|0),(0|0),0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_3(0|0)|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($2);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common6Object8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($this,$retStr,$withTypes,0)|0);
 return ($retStr|0);
}
function __ZN9ExitGames6Common6ObjectaSERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 2047]($this,$toCopy)|0);
 return ($3|0);
}
function __ZN9ExitGames6Common6Object6assignERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($toCopy)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ((($toCopy)) + 28|0);
 $4 = HEAP8[$3>>0]|0;
 if ($2) {
  $5 = ($4<<24>>24)==(42);
  if ($5) {
   $11 = 42;$25 = 0;
  } else {
   $6 = ((($toCopy)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if ($8) {
    $9 = ((($toCopy)) + 8|0);
    $11 = $4;$25 = $9;
   } else {
    $11 = $4;$25 = 0;
   }
  }
 } else {
  $11 = $4;$25 = $1;
 }
 $10 = ($11<<24>>24)==(99);
 $12 = ((($toCopy)) + 29|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $10 ? $13 : 0;
 $15 = ((($toCopy)) + 32|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($toCopy)) + 24|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0|0);
 $20 = ((($toCopy)) + 20|0);
 $21 = $19 ? $20 : $18;
 __ZN9ExitGames6Common6Object7cleanupEPKvj($this,0,0);
 $22 = ($11<<24>>24)==(42);
 do {
  if (!($22)) {
   $23 = ($21|0)==(0|0);
   if (!($23)) {
    $24 = ($25|0)==(0|0);
    if (!($24)) {
     break;
    }
    $26 = HEAP16[$21>>1]|0;
    $27 = ($26<<16>>16)==(0);
    if ($27) {
     break;
    }
   }
   $28 = ((($this)) + 28|0);
   HEAP8[$28>>0] = 42;
   $29 = ((($this)) + 29|0);
   HEAP8[$29>>0] = 0;
   $30 = ((($this)) + 32|0);
   HEAP32[$30>>2] = 0;
   $31 = ((($this)) + 16|0);
   ;HEAP32[$31>>2]=0|0;HEAP32[$31+4>>2]=0|0;HEAP32[$31+8>>2]=0|0;
   return ($this|0);
  }
 } while(0);
 $32 = ((($this)) + 28|0);
 HEAP8[$32>>0] = $11;
 $33 = ((($this)) + 29|0);
 HEAP8[$33>>0] = $14;
 $34 = ((($this)) + 24|0);
 HEAP32[$34>>2] = 0;
 $35 = ($21|0)==(0|0);
 $36 = ((($this)) + 32|0);
 do {
  if ($35) {
   HEAP32[$36>>2] = 0;
   $52 = ((($this)) + 20|0);
   HEAP32[$52>>2] = 0;
   HEAP32[$34>>2] = 0;
  } else {
   HEAP32[$36>>2] = $16;
   $37 = ($16>>>0)<(2);
   if ($37) {
    $38 = HEAP32[$21>>2]|0;
    $39 = ((($this)) + 20|0);
    HEAP32[$39>>2] = $38;
    break;
   }
   $40 = $16 << 1;
   $41 = (($40) + 8)|0;
   $42 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($41)|0);
   $43 = ((($42)) + 8|0);
   $44 = $42;
   $45 = $44;
   HEAP32[$45>>2] = $16;
   $46 = (($44) + 4)|0;
   $47 = $46;
   HEAP32[$47>>2] = 0;
   _memset(($43|0),0,($40|0))|0;
   HEAP32[$34>>2] = $43;
   $i$01$i$i = 0;
   while(1) {
    $48 = (($21) + ($i$01$i$i<<1)|0);
    $49 = HEAP16[$48>>1]|0;
    $50 = (($43) + ($i$01$i$i<<1)|0);
    HEAP16[$50>>1] = $49;
    $51 = (($i$01$i$i) + 1)|0;
    $exitcond$i$i = ($51|0)==($16|0);
    if ($exitcond$i$i) {
     break;
    } else {
     $i$01$i$i = $51;
    }
   }
  }
 } while(0);
 $53 = ((($this)) + 16|0);
 HEAP32[$53>>2] = 0;
 __ZN9ExitGames6Common6Object7setDataEPKv($this,$25);
 return ($this|0);
}
function __ZN9ExitGames6Common6Object7setDataEPKv($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($data|0)==(0|0);
 if ($0) {
  $1 = ((($this)) + 16|0);
  HEAP32[$1>>2] = 0;
  return;
 }
 $2 = ((($this)) + 28|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $3&255;
 switch ($4|0) {
 case 100: case 102: case 111: case 108: case 105: case 107: case 98:  {
  $5 = ((($this)) + 32|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0);
  if (!($7)) {
   $8 = ((($this)) + 16|0);
   __ZNK9ExitGames6Common6Object9copyArrayEPKvPPvj($this,$data,$8,0);
   return;
  }
  switch ($4|0) {
  case 98:  {
   $9 = HEAP8[$data>>0]|0;
   $10 = ((($this)) + 8|0);
   $11 = $9&255;
   $12 = $10;
   $13 = $12;
   HEAP32[$13>>2] = $11;
   $14 = (($12) + 4)|0;
   $15 = $14;
   HEAP32[$15>>2] = 0;
   return;
   break;
  }
  case 107:  {
   $16 = HEAP16[$data>>1]|0;
   $17 = ((($this)) + 8|0);
   $18 = $16&65535;
   $19 = $17;
   $20 = $19;
   HEAP32[$20>>2] = $18;
   $21 = (($19) + 4)|0;
   $22 = $21;
   HEAP32[$22>>2] = 0;
   return;
   break;
  }
  case 105:  {
   $23 = HEAP32[$data>>2]|0;
   $24 = ((($this)) + 8|0);
   $25 = $24;
   $26 = $25;
   HEAP32[$26>>2] = $23;
   $27 = (($25) + 4)|0;
   $28 = $27;
   HEAP32[$28>>2] = 0;
   return;
   break;
  }
  case 108:  {
   $29 = $data;
   $30 = $29;
   $31 = HEAP32[$30>>2]|0;
   $32 = (($29) + 4)|0;
   $33 = $32;
   $34 = HEAP32[$33>>2]|0;
   $35 = ((($this)) + 8|0);
   $36 = $35;
   $37 = $36;
   HEAP32[$37>>2] = $31;
   $38 = (($36) + 4)|0;
   $39 = $38;
   HEAP32[$39>>2] = $34;
   return;
   break;
  }
  case 102:  {
   $40 = HEAP32[$data>>2]|0;
   $41 = ((($this)) + 8|0);
   HEAP32[$41>>2] = $40;
   return;
   break;
  }
  case 100:  {
   $42 = $data;
   $43 = $42;
   $44 = HEAP32[$43>>2]|0;
   $45 = (($42) + 4)|0;
   $46 = $45;
   $47 = HEAP32[$46>>2]|0;
   $48 = ((($this)) + 8|0);
   $49 = $48;
   $50 = $49;
   HEAP32[$50>>2] = $44;
   $51 = (($49) + 4)|0;
   $52 = $51;
   HEAP32[$52>>2] = $47;
   return;
   break;
  }
  case 111:  {
   $53 = HEAP8[$data>>0]|0;
   $54 = ((($this)) + 8|0);
   $55 = $53&255;
   $56 = $54;
   $57 = $56;
   HEAP32[$57>>2] = $55;
   $58 = (($56) + 4)|0;
   $59 = $58;
   HEAP32[$59>>2] = 0;
   return;
   break;
  }
  case 104: case 99:  {
   $60 = ((($this)) + 16|0);
   HEAP32[$60>>2] = $data;
   return;
   break;
  }
  default: {
   return;
  }
  }
  break;
 }
 case 99: case 68: case 122: case 104: case 115:  {
  $61 = ((($this)) + 16|0);
  __ZNK9ExitGames6Common6Object9copyArrayEPKvPPvj($this,$data,$61,0);
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function __ZNK9ExitGames6Common6Object9copyArrayEPKvPPvj($this,$pDataIn,$ppDataOut,$recursionDepth) {
 $this = $this|0;
 $pDataIn = $pDataIn|0;
 $ppDataOut = $ppDataOut|0;
 $recursionDepth = $recursionDepth|0;
 var $$ = 0, $$129 = 0, $$18 = 0, $$1957 = 0, $$20 = 0, $$21 = 0, $$2160 = 0, $$22 = 0, $$2363 = 0, $$24 = 0, $$2566 = 0, $$26 = 0, $$2769 = 0, $$28 = 0, $$2972 = 0, $$30 = 0, $$31 = 0, $$3175 = 0, $$32 = 0, $$33 = 0;
 var $$3378 = 0, $$34 = 0, $$35 = 0, $$3581 = 0, $$36 = 0, $$37 = 0, $$38 = 0, $$3886 = 0, $$52 = 0, $$pre = 0, $$pre100 = 0, $$pre102 = 0, $$pre104 = 0, $$pre96 = 0, $$pre98 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0.0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0;
 var $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0;
 var $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0;
 var $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0;
 var $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0;
 var $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0;
 var $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0;
 var $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0;
 var $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0;
 var $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0;
 var $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0;
 var $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$051 = 0, $i10$079 = 0, $i11$082 = 0, $i12$087 = 0, $i3$058 = 0, $i4$061 = 0, $i9$076 = 0, $indvars$iv = 0, $indvars$iv$next = 0, $indvars$iv$next90 = 0, $indvars$iv$next92 = 0, $indvars$iv$next94 = 0;
 var $indvars$iv89 = 0, $indvars$iv91 = 0, $indvars$iv93 = 0, $or$cond = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pDataIn|0)!=(0|0);
 $1 = ($ppDataOut|0)!=(0|0);
 $or$cond = $0 & $1;
 if (!($or$cond)) {
  return;
 }
 $2 = ((($this)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)!=(0);
 $5 = (($3) + -1)|0;
 $6 = ($5>>>0)>($recursionDepth>>>0);
 $or$cond3 = $4 & $6;
 if ($or$cond3) {
  $7 = ((($this)) + 24|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(0|0);
  $10 = ((($this)) + 20|0);
  $11 = $9 ? $10 : $8;
  $12 = (($11) + ($recursionDepth<<1)|0);
  $13 = HEAP16[$12>>1]|0;
  $14 = $13 << 16 >> 16;
  $15 = $14 << 2;
  $16 = (($15) + 8)|0;
  $17 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($16)|0);
  $18 = ((($17)) + 8|0);
  $19 = $17;
  $20 = $19;
  HEAP32[$20>>2] = $14;
  $21 = (($19) + 4)|0;
  $22 = $21;
  HEAP32[$22>>2] = 0;
  $23 = ($13<<16>>16)==(0);
  if (!($23)) {
   _memset(($18|0),0,($15|0))|0;
  }
  HEAP32[$ppDataOut>>2] = $18;
  $24 = HEAP32[$7>>2]|0;
  $25 = ($24|0)==(0|0);
  $26 = $25 ? $10 : $24;
  $27 = (($26) + ($recursionDepth<<1)|0);
  $28 = HEAP16[$27>>1]|0;
  $29 = ($28<<16>>16)>(0);
  if (!($29)) {
   return;
  }
  $30 = (($recursionDepth) + 1)|0;
  $34 = $18;$i$051 = 0;
  while(1) {
   $31 = (($pDataIn) + ($i$051<<2)|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($34) + ($i$051<<2)|0);
   __ZNK9ExitGames6Common6Object9copyArrayEPKvPPvj($this,$32,$33,$30);
   $35 = (($i$051) + 1)|0;
   $36 = HEAP32[$7>>2]|0;
   $37 = ($36|0)==(0|0);
   $38 = $37 ? $10 : $36;
   $39 = (($38) + ($recursionDepth<<1)|0);
   $40 = HEAP16[$39>>1]|0;
   $41 = $40 << 16 >> 16;
   $42 = ($35|0)<($41|0);
   if (!($42)) {
    break;
   }
   $$pre104 = HEAP32[$ppDataOut>>2]|0;
   $34 = $$pre104;$i$051 = $35;
  }
  return;
 }
 $43 = ((($this)) + 28|0);
 $44 = HEAP8[$43>>0]|0;
 $45 = $44&255;
 do {
  switch ($45|0) {
  case 98:  {
   switch ($3|0) {
   case 1:  {
    $46 = ((($this)) + 20|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = (($47) + 8)|0;
    $49 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($48)|0);
    $50 = ((($49)) + 8|0);
    $51 = $49;
    $52 = $51;
    HEAP32[$52>>2] = $47;
    $53 = (($51) + 4)|0;
    $54 = $53;
    HEAP32[$54>>2] = 0;
    $55 = ($47|0)==(0);
    if (!($55)) {
     _memset(($50|0),0,($47|0))|0;
    }
    HEAP32[$ppDataOut>>2] = $50;
    $56 = HEAP32[$46>>2]|0;
    $57 = ($56|0)>(0);
    if (!($57)) {
     return;
    }
    $58 = HEAP8[$pDataIn>>0]|0;
    HEAP8[$50>>0] = $58;
    $59 = HEAP32[$46>>2]|0;
    $60 = ($59|0)>(1);
    if ($60) {
     $62 = 1;
    } else {
     return;
    }
    while(1) {
     $$pre100 = HEAP32[$ppDataOut>>2]|0;
     $61 = (($pDataIn) + ($62)|0);
     $63 = HEAP8[$61>>0]|0;
     $64 = (($$pre100) + ($62)|0);
     HEAP8[$64>>0] = $63;
     $65 = (($62) + 1)|0;
     $66 = HEAP32[$46>>2]|0;
     $67 = ($65|0)<($66|0);
     if ($67) {
      $62 = $65;
     } else {
      break;
     }
    }
    return;
    break;
   }
   case 0:  {
    $74 = 0;
    break;
   }
   default: {
    $74 = $5;
   }
   }
   $68 = ((($this)) + 24|0);
   $69 = HEAP32[$68>>2]|0;
   $70 = ($69|0)==(0|0);
   $71 = ((($this)) + 20|0);
   $72 = $70 ? $71 : $69;
   $73 = (($72) + ($74<<1)|0);
   $75 = HEAP16[$73>>1]|0;
   $76 = $75 << 16 >> 16;
   $77 = (($76) + 8)|0;
   $78 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($77)|0);
   $79 = ((($78)) + 8|0);
   $80 = $78;
   $81 = $80;
   HEAP32[$81>>2] = $76;
   $82 = (($80) + 4)|0;
   $83 = $82;
   HEAP32[$83>>2] = 0;
   $84 = ($75<<16>>16)==(0);
   if (!($84)) {
    _memset(($79|0),0,($76|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $79;
   $85 = HEAP32[$2>>2]|0;
   $86 = ($85|0)==(0);
   $87 = (($85) + -1)|0;
   $$52 = $86 ? 0 : $87;
   $88 = HEAP32[$68>>2]|0;
   $89 = ($88|0)==(0|0);
   $90 = $89 ? $71 : $88;
   $91 = (($90) + ($$52<<1)|0);
   $92 = HEAP16[$91>>1]|0;
   $93 = ($92<<16>>16)>(0);
   if (!($93)) {
    return;
   }
   $94 = HEAP8[$pDataIn>>0]|0;
   HEAP8[$79>>0] = $94;
   $95 = HEAP32[$2>>2]|0;
   $96 = ($95|0)==(0);
   $97 = (($95) + -1)|0;
   $$129 = $96 ? 0 : $97;
   $98 = HEAP32[$68>>2]|0;
   $99 = ($98|0)==(0|0);
   $100 = $99 ? $71 : $98;
   $101 = (($100) + ($$129<<1)|0);
   $102 = HEAP16[$101>>1]|0;
   $103 = ($102<<16>>16)>(1);
   if ($103) {
    $105 = 1;
   } else {
    return;
   }
   while(1) {
    $104 = $105 << 16 >> 16;
    $$pre102 = HEAP32[$ppDataOut>>2]|0;
    $106 = (($pDataIn) + ($104)|0);
    $107 = HEAP8[$106>>0]|0;
    $108 = (($$pre102) + ($104)|0);
    HEAP8[$108>>0] = $107;
    $109 = (($105) + 1)<<16>>16;
    $110 = HEAP32[$2>>2]|0;
    $111 = ($110|0)==(0);
    $112 = (($110) + -1)|0;
    $$ = $111 ? 0 : $112;
    $113 = HEAP32[$68>>2]|0;
    $114 = ($113|0)==(0|0);
    $115 = $114 ? $71 : $113;
    $116 = (($115) + ($$<<1)|0);
    $117 = HEAP16[$116>>1]|0;
    $118 = ($109<<16>>16)<($117<<16>>16);
    if ($118) {
     $105 = $109;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 107:  {
   $119 = ($3|0)==(0);
   $$18 = $119 ? 0 : $5;
   $120 = ((($this)) + 24|0);
   $121 = HEAP32[$120>>2]|0;
   $122 = ($121|0)==(0|0);
   $123 = ((($this)) + 20|0);
   $124 = $122 ? $123 : $121;
   $125 = (($124) + ($$18<<1)|0);
   $126 = HEAP16[$125>>1]|0;
   $127 = $126 << 16 >> 16;
   $128 = $127 << 1;
   $129 = (($128) + 8)|0;
   $130 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($129)|0);
   $131 = ((($130)) + 8|0);
   $132 = $130;
   $133 = $132;
   HEAP32[$133>>2] = $127;
   $134 = (($132) + 4)|0;
   $135 = $134;
   HEAP32[$135>>2] = 0;
   $136 = ($126<<16>>16)==(0);
   if (!($136)) {
    _memset(($131|0),0,($128|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $131;
   $137 = HEAP32[$2>>2]|0;
   $138 = ($137|0)==(0);
   $139 = (($137) + -1)|0;
   $$1957 = $138 ? 0 : $139;
   $140 = HEAP32[$120>>2]|0;
   $141 = ($140|0)==(0|0);
   $142 = $141 ? $123 : $140;
   $143 = (($142) + ($$1957<<1)|0);
   $144 = HEAP16[$143>>1]|0;
   $145 = ($144<<16>>16)>(0);
   if (!($145)) {
    return;
   }
   $147 = 0;$i3$058 = 0;
   while(1) {
    $146 = (($pDataIn) + ($147<<1)|0);
    $148 = HEAP16[$146>>1]|0;
    $149 = (($131) + ($147<<1)|0);
    HEAP16[$149>>1] = $148;
    $150 = (($i3$058) + 1)<<16>>16;
    $151 = $150 << 16 >> 16;
    $152 = HEAP16[$143>>1]|0;
    $153 = ($150<<16>>16)<($152<<16>>16);
    if ($153) {
     $147 = $151;$i3$058 = $150;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 105:  {
   $154 = ($3|0)==(0);
   $$20 = $154 ? 0 : $5;
   $155 = ((($this)) + 24|0);
   $156 = HEAP32[$155>>2]|0;
   $157 = ($156|0)==(0|0);
   $158 = ((($this)) + 20|0);
   $159 = $157 ? $158 : $156;
   $160 = (($159) + ($$20<<1)|0);
   $161 = HEAP16[$160>>1]|0;
   $162 = $161 << 16 >> 16;
   $163 = $162 << 2;
   $164 = (($163) + 8)|0;
   $165 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($164)|0);
   $166 = ((($165)) + 8|0);
   $167 = $165;
   $168 = $167;
   HEAP32[$168>>2] = $162;
   $169 = (($167) + 4)|0;
   $170 = $169;
   HEAP32[$170>>2] = 0;
   $171 = ($161<<16>>16)==(0);
   if (!($171)) {
    _memset(($166|0),0,($163|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $166;
   $172 = HEAP32[$2>>2]|0;
   $173 = ($172|0)==(0);
   $174 = (($172) + -1)|0;
   $$2160 = $173 ? 0 : $174;
   $175 = HEAP32[$155>>2]|0;
   $176 = ($175|0)==(0|0);
   $177 = $176 ? $158 : $175;
   $178 = (($177) + ($$2160<<1)|0);
   $179 = HEAP16[$178>>1]|0;
   $180 = ($179<<16>>16)>(0);
   if (!($180)) {
    return;
   }
   $182 = 0;$i4$061 = 0;
   while(1) {
    $181 = (($pDataIn) + ($182<<2)|0);
    $183 = HEAP32[$181>>2]|0;
    $184 = (($166) + ($182<<2)|0);
    HEAP32[$184>>2] = $183;
    $185 = (($i4$061) + 1)<<16>>16;
    $186 = $185 << 16 >> 16;
    $187 = HEAP32[$2>>2]|0;
    $188 = ($187|0)==(0);
    $189 = (($187) + -1)|0;
    $$21 = $188 ? 0 : $189;
    $190 = (($177) + ($$21<<1)|0);
    $191 = HEAP16[$190>>1]|0;
    $192 = ($185<<16>>16)<($191<<16>>16);
    if ($192) {
     $182 = $186;$i4$061 = $185;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 108:  {
   $193 = ($3|0)==(0);
   $$22 = $193 ? 0 : $5;
   $194 = ((($this)) + 24|0);
   $195 = HEAP32[$194>>2]|0;
   $196 = ($195|0)==(0|0);
   $197 = ((($this)) + 20|0);
   $198 = $196 ? $197 : $195;
   $199 = (($198) + ($$22<<1)|0);
   $200 = HEAP16[$199>>1]|0;
   $201 = $200 << 16 >> 16;
   $202 = $201 << 3;
   $203 = (($202) + 8)|0;
   $204 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($203)|0);
   $205 = ((($204)) + 8|0);
   $206 = $204;
   $207 = $206;
   HEAP32[$207>>2] = $201;
   $208 = (($206) + 4)|0;
   $209 = $208;
   HEAP32[$209>>2] = 0;
   $210 = ($200<<16>>16)==(0);
   if (!($210)) {
    _memset(($205|0),0,($202|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $205;
   $211 = HEAP32[$2>>2]|0;
   $212 = ($211|0)==(0);
   $213 = (($211) + -1)|0;
   $$2363 = $212 ? 0 : $213;
   $214 = HEAP32[$194>>2]|0;
   $215 = ($214|0)==(0|0);
   $216 = $215 ? $197 : $214;
   $217 = (($216) + ($$2363<<1)|0);
   $218 = HEAP16[$217>>1]|0;
   $219 = ($218<<16>>16)>(0);
   if (!($219)) {
    return;
   }
   $220 = $218 << 16 >> 16;
   $indvars$iv = 0;
   while(1) {
    $221 = (($pDataIn) + ($indvars$iv<<3)|0);
    $222 = $221;
    $223 = $222;
    $224 = HEAP32[$223>>2]|0;
    $225 = (($222) + 4)|0;
    $226 = $225;
    $227 = HEAP32[$226>>2]|0;
    $228 = (($205) + ($indvars$iv<<3)|0);
    $229 = $228;
    $230 = $229;
    HEAP32[$230>>2] = $224;
    $231 = (($229) + 4)|0;
    $232 = $231;
    HEAP32[$232>>2] = $227;
    $indvars$iv$next = (($indvars$iv) + 1)|0;
    $233 = ($indvars$iv$next|0)<($220|0);
    if ($233) {
     $indvars$iv = $indvars$iv$next;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 102:  {
   $234 = ($3|0)==(0);
   $$24 = $234 ? 0 : $5;
   $235 = ((($this)) + 24|0);
   $236 = HEAP32[$235>>2]|0;
   $237 = ($236|0)==(0|0);
   $238 = ((($this)) + 20|0);
   $239 = $237 ? $238 : $236;
   $240 = (($239) + ($$24<<1)|0);
   $241 = HEAP16[$240>>1]|0;
   $242 = $241 << 16 >> 16;
   $243 = $242 << 2;
   $244 = (($243) + 8)|0;
   $245 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($244)|0);
   $246 = ((($245)) + 8|0);
   $247 = $245;
   $248 = $247;
   HEAP32[$248>>2] = $242;
   $249 = (($247) + 4)|0;
   $250 = $249;
   HEAP32[$250>>2] = 0;
   $251 = ($241<<16>>16)==(0);
   if (!($251)) {
    _memset(($246|0),0,($243|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $246;
   $252 = HEAP32[$2>>2]|0;
   $253 = ($252|0)==(0);
   $254 = (($252) + -1)|0;
   $$2566 = $253 ? 0 : $254;
   $255 = HEAP32[$235>>2]|0;
   $256 = ($255|0)==(0|0);
   $257 = $256 ? $238 : $255;
   $258 = (($257) + ($$2566<<1)|0);
   $259 = HEAP16[$258>>1]|0;
   $260 = ($259<<16>>16)>(0);
   if (!($260)) {
    return;
   }
   $261 = $259 << 16 >> 16;
   $indvars$iv89 = 0;
   while(1) {
    $262 = (($pDataIn) + ($indvars$iv89<<2)|0);
    $263 = HEAP32[$262>>2]|0;
    $264 = (($246) + ($indvars$iv89<<2)|0);
    HEAP32[$264>>2] = $263;
    $indvars$iv$next90 = (($indvars$iv89) + 1)|0;
    $265 = ($indvars$iv$next90|0)<($261|0);
    if ($265) {
     $indvars$iv89 = $indvars$iv$next90;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 100:  {
   $266 = ($3|0)==(0);
   $$26 = $266 ? 0 : $5;
   $267 = ((($this)) + 24|0);
   $268 = HEAP32[$267>>2]|0;
   $269 = ($268|0)==(0|0);
   $270 = ((($this)) + 20|0);
   $271 = $269 ? $270 : $268;
   $272 = (($271) + ($$26<<1)|0);
   $273 = HEAP16[$272>>1]|0;
   $274 = $273 << 16 >> 16;
   $275 = $274 << 3;
   $276 = (($275) + 8)|0;
   $277 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($276)|0);
   $278 = ((($277)) + 8|0);
   $279 = $277;
   $280 = $279;
   HEAP32[$280>>2] = $274;
   $281 = (($279) + 4)|0;
   $282 = $281;
   HEAP32[$282>>2] = 0;
   $283 = ($273<<16>>16)==(0);
   if (!($283)) {
    _memset(($278|0),0,($275|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $278;
   $284 = HEAP32[$2>>2]|0;
   $285 = ($284|0)==(0);
   $286 = (($284) + -1)|0;
   $$2769 = $285 ? 0 : $286;
   $287 = HEAP32[$267>>2]|0;
   $288 = ($287|0)==(0|0);
   $289 = $288 ? $270 : $287;
   $290 = (($289) + ($$2769<<1)|0);
   $291 = HEAP16[$290>>1]|0;
   $292 = ($291<<16>>16)>(0);
   if (!($292)) {
    return;
   }
   $293 = $291 << 16 >> 16;
   $indvars$iv91 = 0;
   while(1) {
    $294 = (($pDataIn) + ($indvars$iv91<<3)|0);
    $295 = +HEAPF64[$294>>3];
    $296 = (($278) + ($indvars$iv91<<3)|0);
    HEAPF64[$296>>3] = $295;
    $indvars$iv$next92 = (($indvars$iv91) + 1)|0;
    $297 = ($indvars$iv$next92|0)<($293|0);
    if ($297) {
     $indvars$iv91 = $indvars$iv$next92;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 111:  {
   $298 = ($3|0)==(0);
   $$28 = $298 ? 0 : $5;
   $299 = ((($this)) + 24|0);
   $300 = HEAP32[$299>>2]|0;
   $301 = ($300|0)==(0|0);
   $302 = ((($this)) + 20|0);
   $303 = $301 ? $302 : $300;
   $304 = (($303) + ($$28<<1)|0);
   $305 = HEAP16[$304>>1]|0;
   $306 = $305 << 16 >> 16;
   $307 = (($306) + 8)|0;
   $308 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($307)|0);
   $309 = ((($308)) + 8|0);
   $310 = $308;
   $311 = $310;
   HEAP32[$311>>2] = $306;
   $312 = (($310) + 4)|0;
   $313 = $312;
   HEAP32[$313>>2] = 0;
   $314 = ($305<<16>>16)==(0);
   if (!($314)) {
    _memset(($309|0),0,($306|0))|0;
   }
   HEAP32[$ppDataOut>>2] = $309;
   $315 = HEAP32[$2>>2]|0;
   $316 = ($315|0)==(0);
   $317 = (($315) + -1)|0;
   $$2972 = $316 ? 0 : $317;
   $318 = HEAP32[$299>>2]|0;
   $319 = ($318|0)==(0|0);
   $320 = $319 ? $302 : $318;
   $321 = (($320) + ($$2972<<1)|0);
   $322 = HEAP16[$321>>1]|0;
   $323 = ($322<<16>>16)>(0);
   if (!($323)) {
    return;
   }
   $324 = $322 << 16 >> 16;
   $indvars$iv93 = 0;
   while(1) {
    $325 = (($pDataIn) + ($indvars$iv93)|0);
    $326 = HEAP8[$325>>0]|0;
    $$sum = (($indvars$iv93) + 8)|0;
    $327 = (($308) + ($$sum)|0);
    HEAP8[$327>>0] = $326;
    $indvars$iv$next94 = (($indvars$iv93) + 1)|0;
    $328 = ($indvars$iv$next94|0)<($324|0);
    if ($328) {
     $indvars$iv93 = $indvars$iv$next94;
    } else {
     break;
    }
   }
   return;
   break;
  }
  case 115:  {
   $329 = ($3|0)==(0);
   $$30 = $329 ? 0 : $5;
   $330 = ((($this)) + 24|0);
   $331 = HEAP32[$330>>2]|0;
   $332 = ($331|0)==(0|0);
   $333 = ((($this)) + 20|0);
   $334 = $332 ? $333 : $331;
   $335 = (($334) + ($$30<<1)|0);
   $336 = HEAP16[$335>>1]|0;
   $337 = $336 << 16 >> 16;
   $338 = $337 << 4;
   $339 = $338 | 8;
   $340 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($339)|0);
   $341 = ((($340)) + 8|0);
   $342 = $340;
   $343 = $342;
   HEAP32[$343>>2] = $337;
   $344 = (($342) + 4)|0;
   $345 = $344;
   HEAP32[$345>>2] = 0;
   $346 = ($336<<16>>16)==(0);
   if (!($346)) {
    $348 = 0;$349 = 0;
    while(1) {
     $347 = (($341) + ($348<<4)|0);
     __ZN9ExitGames6Common7JStringC2Ej($347,0);
     $350 = (_i64Add(($348|0),($349|0),1,0)|0);
     $351 = tempRet0;
     $352 = ($351>>>0)<(0);
     $353 = ($350>>>0)<($337>>>0);
     $354 = ($351|0)==(0);
     $355 = $354 & $353;
     $356 = $352 | $355;
     if ($356) {
      $348 = $350;$349 = $351;
     } else {
      break;
     }
    }
   }
   HEAP32[$ppDataOut>>2] = $341;
   $357 = HEAP32[$2>>2]|0;
   $358 = ($357|0)==(0);
   $359 = (($357) + -1)|0;
   $$3175 = $358 ? 0 : $359;
   $360 = HEAP32[$330>>2]|0;
   $361 = ($360|0)==(0|0);
   $362 = $361 ? $333 : $360;
   $363 = (($362) + ($$3175<<1)|0);
   $364 = HEAP16[$363>>1]|0;
   $365 = ($364<<16>>16)>(0);
   if (!($365)) {
    return;
   }
   $367 = $341;$368 = 0;$i9$076 = 0;
   while(1) {
    $366 = (($367) + ($368<<4)|0);
    $369 = (($pDataIn) + ($368<<4)|0);
    (__ZN9ExitGames6Common7JStringaSERKS1_($366,$369)|0);
    $370 = (($i9$076) + 1)<<16>>16;
    $371 = HEAP32[$2>>2]|0;
    $372 = ($371|0)==(0);
    $373 = (($371) + -1)|0;
    $$31 = $372 ? 0 : $373;
    $374 = HEAP32[$330>>2]|0;
    $375 = ($374|0)==(0|0);
    $376 = $375 ? $333 : $374;
    $377 = (($376) + ($$31<<1)|0);
    $378 = HEAP16[$377>>1]|0;
    $379 = ($370<<16>>16)<($378<<16>>16);
    if (!($379)) {
     break;
    }
    $380 = $370 << 16 >> 16;
    $$pre98 = HEAP32[$ppDataOut>>2]|0;
    $367 = $$pre98;$368 = $380;$i9$076 = $370;
   }
   return;
   break;
  }
  case 104:  {
   $381 = ($3|0)==(0);
   $$32 = $381 ? 0 : $5;
   $382 = ((($this)) + 24|0);
   $383 = HEAP32[$382>>2]|0;
   $384 = ($383|0)==(0|0);
   $385 = ((($this)) + 20|0);
   $386 = $384 ? $385 : $383;
   $387 = (($386) + ($$32<<1)|0);
   $388 = HEAP16[$387>>1]|0;
   $389 = $388 << 16 >> 16;
   $390 = ($389*44)|0;
   $391 = (($390) + 8)|0;
   $392 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($391)|0);
   $393 = ((($392)) + 8|0);
   $394 = $392;
   $395 = $394;
   HEAP32[$395>>2] = $389;
   $396 = (($394) + 4)|0;
   $397 = $396;
   HEAP32[$397>>2] = 0;
   $398 = ($388<<16>>16)==(0);
   if (!($398)) {
    $400 = 0;$401 = 0;
    while(1) {
     $399 = (($393) + (($400*44)|0)|0);
     __ZN9ExitGames6Common9HashtableC2Ev($399);
     $402 = (_i64Add(($400|0),($401|0),1,0)|0);
     $403 = tempRet0;
     $404 = ($403>>>0)<(0);
     $405 = ($402>>>0)<($389>>>0);
     $406 = ($403|0)==(0);
     $407 = $406 & $405;
     $408 = $404 | $407;
     if ($408) {
      $400 = $402;$401 = $403;
     } else {
      break;
     }
    }
   }
   HEAP32[$ppDataOut>>2] = $393;
   $409 = HEAP32[$2>>2]|0;
   $410 = ($409|0)==(0);
   $411 = (($409) + -1)|0;
   $$3378 = $410 ? 0 : $411;
   $412 = HEAP32[$382>>2]|0;
   $413 = ($412|0)==(0|0);
   $414 = $413 ? $385 : $412;
   $415 = (($414) + ($$3378<<1)|0);
   $416 = HEAP16[$415>>1]|0;
   $417 = ($416<<16>>16)>(0);
   if (!($417)) {
    return;
   }
   $419 = $393;$420 = 0;$i10$079 = 0;
   while(1) {
    $418 = (($419) + (($420*44)|0)|0);
    $421 = (($pDataIn) + (($420*44)|0)|0);
    (__ZN9ExitGames6Common9HashtableaSERKS1_($418,$421)|0);
    $422 = (($i10$079) + 1)<<16>>16;
    $423 = HEAP32[$2>>2]|0;
    $424 = ($423|0)==(0);
    $425 = (($423) + -1)|0;
    $$33 = $424 ? 0 : $425;
    $426 = HEAP32[$382>>2]|0;
    $427 = ($426|0)==(0|0);
    $428 = $427 ? $385 : $426;
    $429 = (($428) + ($$33<<1)|0);
    $430 = HEAP16[$429>>1]|0;
    $431 = ($422<<16>>16)<($430<<16>>16);
    if (!($431)) {
     break;
    }
    $432 = $422 << 16 >> 16;
    $$pre96 = HEAP32[$ppDataOut>>2]|0;
    $419 = $$pre96;$420 = $432;$i10$079 = $422;
   }
   return;
   break;
  }
  case 122:  {
   $433 = ($3|0)==(0);
   $$34 = $433 ? 0 : $5;
   $434 = ((($this)) + 24|0);
   $435 = HEAP32[$434>>2]|0;
   $436 = ($435|0)==(0|0);
   $437 = ((($this)) + 20|0);
   $438 = $436 ? $437 : $435;
   $439 = (($438) + ($$34<<1)|0);
   $440 = HEAP16[$439>>1]|0;
   $441 = $440 << 16 >> 16;
   $442 = ($441*40)|0;
   $443 = (($442) + 8)|0;
   $444 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($443)|0);
   $445 = ((($444)) + 8|0);
   $446 = $444;
   $447 = $446;
   HEAP32[$447>>2] = $441;
   $448 = (($446) + 4)|0;
   $449 = $448;
   HEAP32[$449>>2] = 0;
   $450 = ($440<<16>>16)==(0);
   if (!($450)) {
    $452 = 0;$457 = 0;
    while(1) {
     $451 = (($445) + (($452*40)|0)|0);
     HEAP32[$451>>2] = (4640);
     $453 = (((($445) + (($452*40)|0)|0)) + 28|0);
     HEAP8[$453>>0] = 42;
     $454 = (((($445) + (($452*40)|0)|0)) + 29|0);
     HEAP8[$454>>0] = 0;
     $455 = (((($445) + (($452*40)|0)|0)) + 32|0);
     HEAP32[$455>>2] = 0;
     $456 = (((($445) + (($452*40)|0)|0)) + 16|0);
     $458 = (_i64Add(($452|0),($457|0),1,0)|0);
     $459 = tempRet0;
     $460 = ($459>>>0)<(0);
     $461 = ($458>>>0)<($441>>>0);
     $462 = ($459|0)==(0);
     $463 = $462 & $461;
     $464 = $460 | $463;
     ;HEAP32[$456>>2]=0|0;HEAP32[$456+4>>2]=0|0;HEAP32[$456+8>>2]=0|0;
     if ($464) {
      $452 = $458;$457 = $459;
     } else {
      break;
     }
    }
   }
   HEAP32[$ppDataOut>>2] = $445;
   $465 = HEAP32[$2>>2]|0;
   $466 = ($465|0)==(0);
   $467 = (($465) + -1)|0;
   $$3581 = $466 ? 0 : $467;
   $468 = HEAP32[$434>>2]|0;
   $469 = ($468|0)==(0|0);
   $470 = $469 ? $437 : $468;
   $471 = (($470) + ($$3581<<1)|0);
   $472 = HEAP16[$471>>1]|0;
   $473 = ($472<<16>>16)>(0);
   if (!($473)) {
    return;
   }
   $475 = $445;$476 = 0;$i11$082 = 0;
   while(1) {
    $474 = (($475) + (($476*40)|0)|0);
    $477 = HEAP32[$474>>2]|0;
    $478 = ((($477)) + 16|0);
    $479 = HEAP32[$478>>2]|0;
    $480 = (($pDataIn) + (($476*40)|0)|0);
    (FUNCTION_TABLE_iii[$479 & 2047]($474,$480)|0);
    $481 = (($i11$082) + 1)<<16>>16;
    $482 = HEAP32[$2>>2]|0;
    $483 = ($482|0)==(0);
    $484 = (($482) + -1)|0;
    $$35 = $483 ? 0 : $484;
    $485 = HEAP32[$434>>2]|0;
    $486 = ($485|0)==(0|0);
    $487 = $486 ? $437 : $485;
    $488 = (($487) + ($$35<<1)|0);
    $489 = HEAP16[$488>>1]|0;
    $490 = ($481<<16>>16)<($489<<16>>16);
    if (!($490)) {
     break;
    }
    $491 = $481 << 16 >> 16;
    $$pre = HEAP32[$ppDataOut>>2]|0;
    $475 = $$pre;$476 = $491;$i11$082 = $481;
   }
   return;
   break;
  }
  case 68:  {
   $492 = HEAP32[$pDataIn>>2]|0;
   $493 = ((($492)) + 36|0);
   $494 = HEAP32[$493>>2]|0;
   $495 = ($3|0)==(0);
   $$36 = $495 ? 0 : $5;
   $496 = ((($this)) + 24|0);
   $497 = HEAP32[$496>>2]|0;
   $498 = ($497|0)==(0|0);
   $499 = ((($this)) + 20|0);
   $500 = $498 ? $499 : $497;
   $501 = (($500) + ($$36<<1)|0);
   $502 = HEAP16[$501>>1]|0;
   $503 = (FUNCTION_TABLE_iii[$494 & 2047]($pDataIn,$502)|0);
   HEAP32[$ppDataOut>>2] = $503;
   return;
   break;
  }
  case 99:  {
   $504 = ($3|0)==(0);
   $$37 = $504 ? 0 : $5;
   $505 = ((($this)) + 24|0);
   $506 = HEAP32[$505>>2]|0;
   $507 = ($506|0)==(0|0);
   $508 = ((($this)) + 20|0);
   $509 = $507 ? $508 : $506;
   $510 = (($509) + ($$37<<1)|0);
   $511 = HEAP16[$510>>1]|0;
   $512 = ($44<<24>>24)==(99);
   if ($512) {
    $513 = ((($this)) + 29|0);
    $514 = HEAP8[$513>>0]|0;
    $515 = $514;
   } else {
    $515 = 0;
   }
   $516 = (__ZN9ExitGames6Common14CustomTypeBase11allocObjectEsh($511,$515)|0);
   HEAP32[$ppDataOut>>2] = $516;
   $517 = HEAP8[$43>>0]|0;
   $518 = ($517<<24>>24)==(99);
   if ($518) {
    $519 = ((($this)) + 29|0);
    $520 = HEAP8[$519>>0]|0;
    $521 = $520;
   } else {
    $521 = 0;
   }
   $522 = (__ZN9ExitGames6Common14CustomTypeBase9getSizeofEh($521)|0);
   $523 = HEAP32[$2>>2]|0;
   $524 = ($523|0)==(0);
   $525 = (($523) + -1)|0;
   $$3886 = $524 ? 0 : $525;
   $526 = HEAP32[$505>>2]|0;
   $527 = ($526|0)==(0|0);
   $528 = $527 ? $508 : $526;
   $529 = (($528) + ($$3886<<1)|0);
   $530 = HEAP16[$529>>1]|0;
   $531 = ($530<<16>>16)>(0);
   if (!($531)) {
    return;
   }
   $532 = ((($this)) + 29|0);
   $i12$087 = 0;
   while(1) {
    $533 = Math_imul($i12$087, $522)|0;
    $534 = (($pDataIn) + ($533)|0);
    $535 = HEAP32[$ppDataOut>>2]|0;
    $536 = (($535) + ($533)|0);
    $537 = HEAP8[$43>>0]|0;
    $538 = ($537<<24>>24)==(99);
    if ($538) {
     $539 = HEAP8[$532>>0]|0;
     $540 = $539;
    } else {
     $540 = 0;
    }
    __ZN9ExitGames6Common14CustomTypeBase9duplicateEPvS2_h($534,$536,$540);
    $541 = (($i12$087) + 1)|0;
    $542 = HEAP32[$2>>2]|0;
    $543 = ($542|0)==(0);
    $544 = (($542) + -1)|0;
    $$38 = $543 ? 0 : $544;
    $545 = HEAP32[$505>>2]|0;
    $546 = ($545|0)==(0|0);
    $547 = $546 ? $508 : $545;
    $548 = (($547) + ($$38<<1)|0);
    $549 = HEAP16[$548>>1]|0;
    $550 = $549 << 16 >> 16;
    $551 = ($541|0)<($550|0);
    if ($551) {
     $i12$087 = $541;
    } else {
     break;
    }
   }
   return;
   break;
  }
  default: {
   return;
  }
  }
 } while(0);
}
function __ZN9ExitGames6Common9HashtableaSERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = ((($toCopy)) + 4|0);
 (__ZN9ExitGames6Common7JVectorINS0_6ObjectEEaSERKS3_($0,$1)|0);
 $2 = ((($this)) + 24|0);
 $3 = ((($toCopy)) + 24|0);
 (__ZN9ExitGames6Common7JVectorINS0_6ObjectEEaSERKS3_($2,$3)|0);
 return ($this|0);
}
function __ZN9ExitGames6Common14CustomTypeBase11allocObjectEsh($count,$typeCode) {
 $count = $count|0;
 $typeCode = $typeCode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $12 = (__ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper5allocEsh($3,$count,$typeCode)|0);
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Common14CustomTypeBase9getSizeofEh($typeCode) {
 $typeCode = $typeCode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $12 = (__ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper9getSizeofEh($3,$typeCode)|0);
  $$0 = $12;
  STACKTOP = sp;return ($$0|0);
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Common14CustomTypeBase9duplicateEPvS2_h($pData,$retVal,$typeCode) {
 $pData = $pData|0;
 $retVal = $retVal|0;
 $typeCode = $typeCode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $12 = HEAP32[$3>>2]|0;
  $13 = ((($12)) + 16|0);
  $14 = HEAP32[$13>>2]|0;
  FUNCTION_TABLE_viii[$14 & 2047]($3,$pData,$retVal);
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper9getSizeofEh($this,$customTypeCode) {
 $this = $this|0;
 $customTypeCode = $customTypeCode|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_ii[$1 & 2047]($customTypeCode)|0);
 return ($2|0);
}
function __ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper5allocEsh($this,$count,$customTypeCode) {
 $this = $this|0;
 $count = $count|0;
 $customTypeCode = $customTypeCode|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (FUNCTION_TABLE_iii[$1 & 2047]($count,$customTypeCode)|0);
 return ($2|0);
}
function __ZN9ExitGames6Common9HashtableC2Ev($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5228);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = (3844);
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = 40;
 __THREW__ = 0;
 $3 = (invoke_ii(775,1600)|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($0);
  $$0 = $17;$$01 = $16;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($$01|0);
  // unreachable;
 }
 $6 = ((($this)) + 20|0);
 HEAP32[$6>>2] = $3;
 $7 = ((($this)) + 16|0);
 HEAP32[$7>>2] = 10;
 $8 = ((($this)) + 24|0);
 HEAP32[$8>>2] = (3844);
 $9 = ((($this)) + 28|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($this)) + 32|0);
 HEAP32[$10>>2] = 40;
 __THREW__ = 0;
 $11 = (invoke_ii(775,1600)|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  $14 = ((($this)) + 40|0);
  HEAP32[$14>>2] = $11;
  $15 = ((($this)) + 36|0);
  HEAP32[$15>>2] = 10;
  return;
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __ZN9ExitGames6Common4BaseD2Ev($8);
 HEAP32[$0>>2] = (3844);
 $20 = HEAP32[$1>>2]|0;
 $21 = ($20|0)==(0);
 if (!($21)) {
  $i$01$i$i = 0;
  while(1) {
   $22 = HEAP32[$6>>2]|0;
   $23 = (($22) + (($i$01$i$i*40)|0)|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = HEAP32[$24>>2]|0;
   FUNCTION_TABLE_vi[$25 & 2047]($23);
   $26 = (($i$01$i$i) + 1)|0;
   $27 = HEAP32[$1>>2]|0;
   $28 = ($26>>>0)<($27>>>0);
   if ($28) {
    $i$01$i$i = $26;
   } else {
    break;
   }
  }
 }
 HEAP32[$1>>2] = 0;
 $29 = HEAP32[$6>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($29|0));
 $30 = __THREW__; __THREW__ = 0;
 $31 = $30&1;
 if ($31) {
  $32 = ___cxa_find_matching_catch_3(0|0)|0;
  $33 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($0);
  ___clang_call_terminate($32);
  // unreachable;
 }
 __ZN9ExitGames6Common4BaseD2Ev($0);
 $$0 = $19;$$01 = $18;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Common9HashtableD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i$i = 0, $i$01$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5228);
 $0 = ((($this)) + 24|0);
 HEAP32[$0>>2] = (3844);
 $1 = ((($this)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 $$pre$i = ((($this)) + 40|0);
 if (!($3)) {
  $i$01$i$i = 0;
  while(1) {
   $4 = HEAP32[$$pre$i>>2]|0;
   $5 = (($4) + (($i$01$i$i*40)|0)|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = HEAP32[$6>>2]|0;
   FUNCTION_TABLE_vi[$7 & 2047]($5);
   $8 = (($i$01$i$i) + 1)|0;
   $9 = HEAP32[$1>>2]|0;
   $10 = ($8>>>0)<($9>>>0);
   if ($10) {
    $i$01$i$i = $8;
   } else {
    break;
   }
  }
 }
 HEAP32[$1>>2] = 0;
 $11 = HEAP32[$$pre$i>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($11|0));
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $14 = ___cxa_find_matching_catch_3(0|0)|0;
  $15 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($0);
  ___clang_call_terminate($14);
  // unreachable;
 }
 __ZN9ExitGames6Common4BaseD2Ev($0);
 $16 = ((($this)) + 4|0);
 HEAP32[$16>>2] = (3844);
 $17 = ((($this)) + 8|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0);
 $$pre$i1 = ((($this)) + 20|0);
 if (!($19)) {
  $i$01$i$i2 = 0;
  while(1) {
   $20 = HEAP32[$$pre$i1>>2]|0;
   $21 = (($20) + (($i$01$i$i2*40)|0)|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = HEAP32[$22>>2]|0;
   FUNCTION_TABLE_vi[$23 & 2047]($21);
   $24 = (($i$01$i$i2) + 1)|0;
   $25 = HEAP32[$17>>2]|0;
   $26 = ($24>>>0)<($25>>>0);
   if ($26) {
    $i$01$i$i2 = $24;
   } else {
    break;
   }
  }
 }
 HEAP32[$17>>2] = 0;
 $27 = HEAP32[$$pre$i1>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($27|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $30 = ___cxa_find_matching_catch_3(0|0)|0;
  $31 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($16);
  ___clang_call_terminate($30);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($16);
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common9HashtableD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common9HashtableD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Common9Hashtable8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $i$07 = 0, $str$i = 0, $str$i1 = 0, $str$i4 = 0, $str$i5 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $str$i5 = sp + 96|0;
 $str$i4 = sp + 80|0;
 $0 = sp + 64|0;
 $str$i1 = sp + 48|0;
 $1 = sp + 32|0;
 $str$i = sp + 16|0;
 __ZN9ExitGames6Common7JStringC2Ej($str$i,0);
 __THREW__ = 0;
 (invoke_iii(759,($str$i|0),(5676|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i);
  ___resumeException($6|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(760,($retStr|0),($str$i|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i);
  ___resumeException($6|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($str$i);
 $8 = ((($this)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(0);
 L7: do {
  if (!($10)) {
   $11 = ((($this)) + 20|0);
   $12 = ((($this)) + 28|0);
   $13 = ((($this)) + 40|0);
   $15 = $9;$i$07 = 0;
   while(1) {
    $14 = ($15>>>0)>($i$07>>>0);
    if (!($14)) {
     __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
     __THREW__ = 0;
     $16 = (invoke_ii(764,($1|0))|0);
     $17 = __THREW__; __THREW__ = 0;
     $18 = $17&1;
     if ($18) {
      label = 10;
      break;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($16|0),1,566,(10968|0),($vararg_buffer|0));
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     if ($20) {
      label = 10;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($1);
    }
    $23 = HEAP32[$11>>2]|0;
    $24 = (($23) + (($i$07*40)|0)|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ((($25)) + 12|0);
    $27 = HEAP32[$26>>2]|0;
    (FUNCTION_TABLE_iiii[$27 & 2047]($24,$retStr,$withTypes)|0);
    __ZN9ExitGames6Common7JStringC2Ej($str$i1,0);
    __THREW__ = 0;
    (invoke_iii(759,($str$i1|0),(13052|0))|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if ($29) {
     label = 13;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(760,($retStr|0),($str$i1|0))|0);
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if ($31) {
     label = 13;
     break;
    }
    __ZN9ExitGames6Common7JStringD2Ev($str$i1);
    $34 = HEAP32[$12>>2]|0;
    $35 = ($34>>>0)>($i$07>>>0);
    if (!($35)) {
     __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
     __THREW__ = 0;
     $36 = (invoke_ii(764,($0|0))|0);
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if ($38) {
      label = 18;
      break;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($36|0),1,566,(10968|0),($vararg_buffer1|0));
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      label = 18;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
    }
    $43 = HEAP32[$13>>2]|0;
    $44 = (($43) + (($i$07*40)|0)|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = ((($45)) + 12|0);
    $47 = HEAP32[$46>>2]|0;
    (FUNCTION_TABLE_iiii[$47 & 2047]($44,$retStr,$withTypes)|0);
    $48 = HEAP32[$8>>2]|0;
    $49 = (($48) + -1)|0;
    $50 = ($i$07>>>0)<($49>>>0);
    if ($50) {
     __ZN9ExitGames6Common7JStringC2Ej($str$i4,0);
     __THREW__ = 0;
     (invoke_iii(759,($str$i4|0),(12352|0))|0);
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 22;
      break;
     }
     __THREW__ = 0;
     (invoke_iii(760,($retStr|0),($str$i4|0))|0);
     $53 = __THREW__; __THREW__ = 0;
     $54 = $53&1;
     if ($54) {
      label = 22;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($str$i4);
     $$pre = HEAP32[$8>>2]|0;
     $59 = $$pre;
    } else {
     $59 = $48;
    }
    $57 = (($i$07) + 1)|0;
    $58 = ($57>>>0)<($59>>>0);
    if ($58) {
     $15 = $59;$i$07 = $57;
    } else {
     break L7;
    }
   }
   if ((label|0) == 10) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($1);
    ___resumeException($21|0);
    // unreachable;
   }
   else if ((label|0) == 13) {
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i1);
    ___resumeException($32|0);
    // unreachable;
   }
   else if ((label|0) == 18) {
    $41 = ___cxa_find_matching_catch_2()|0;
    $42 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($41|0);
    // unreachable;
   }
   else if ((label|0) == 22) {
    $55 = ___cxa_find_matching_catch_2()|0;
    $56 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i4);
    ___resumeException($55|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN9ExitGames6Common7JStringC2Ej($str$i5,0);
 __THREW__ = 0;
 (invoke_iii(759,($str$i5|0),(10392|0))|0);
 $60 = __THREW__; __THREW__ = 0;
 $61 = $60&1;
 if ($61) {
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i5);
  ___resumeException($64|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(760,($retStr|0),($str$i5|0))|0);
 $62 = __THREW__; __THREW__ = 0;
 $63 = $62&1;
 if ($63) {
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i5);
  ___resumeException($64|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($str$i5);
  STACKTOP = sp;return ($retStr|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($object,$retStr,$withTypes,$butWithOutTopLevelTypes) {
 $object = $object|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 $butWithOutTopLevelTypes = $butWithOutTopLevelTypes|0;
 var $$ = 0, $$0 = 0, $$01 = 0, $$88 = 0, $$89 = 0, $$90 = 0, $$91 = 0, $$92 = 0, $$93 = 0, $$94 = 0, $$95 = 0, $$96 = 0, $$97 = 0, $$98 = 0, $$pre = 0, $$pre71 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0.0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0.0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0;
 var $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0;
 var $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0;
 var $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $brmerge = 0, $exitcond$i$i = 0, $i$01$i$i = 0, $i$070 = 0, $i1$0 = 0, $or$cond38 = 0, $or$cond40 = 0, $or$cond42 = 0, $or$cond44 = 0, $or$cond46 = 0, $or$cond48 = 0, $or$cond50 = 0, $or$cond52 = 0, $or$cond54 = 0, $or$cond56 = 0, $or$cond58 = 0, $or$cond60 = 0, $or$cond62 = 0, $or$cond64 = 0;
 var $or$cond66 = 0, $size = 0, $str$i = 0, $str$i10 = 0, $str$i14 = 0, $str$i16 = 0, $str$i18 = 0, $str$i2 = 0, $str$i20 = 0, $str$i22 = 0, $str$i24 = 0, $str$i26 = 0, $str$i33 = 0, $str$i34 = 0, $str$i4 = 0, $str$i6 = 0, $str$i8 = 0, $vararg_buffer = 0, $withTypes$not = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(416|0);
 $vararg_buffer = sp + 40|0;
 $str$i34 = sp + 392|0;
 $str$i33 = sp + 376|0;
 $str$i26 = sp + 360|0;
 $str$i24 = sp + 344|0;
 $str$i22 = sp + 328|0;
 $str$i20 = sp + 312|0;
 $str$i18 = sp + 296|0;
 $str$i16 = sp + 280|0;
 $str$i14 = sp + 264|0;
 $str$i10 = sp + 248|0;
 $str$i8 = sp + 232|0;
 $str$i6 = sp + 216|0;
 $str$i4 = sp + 200|0;
 $str$i2 = sp + 184|0;
 $str$i = sp + 168|0;
 $0 = sp + 152|0;
 $size = sp + 144|0;
 $1 = sp;
 $2 = sp + 128|0;
 $3 = sp + 112|0;
 $4 = sp + 96|0;
 $5 = sp + 80|0;
 $6 = sp + 64|0;
 $7 = sp + 48|0;
 $withTypes$not = $withTypes ^ 1;
 $brmerge = $withTypes$not | $butWithOutTopLevelTypes;
 do {
  if (!($brmerge)) {
   __ZNK9ExitGames6Common6Object19payloadTypeToStringEv($0,$object);
   __THREW__ = 0;
   (invoke_iii(760,($retStr|0),($0|0))|0);
   $8 = __THREW__; __THREW__ = 0;
   $9 = $8&1;
   if (!($9)) {
    __ZN9ExitGames6Common7JStringD2Ev($0);
    break;
   }
   $10 = ___cxa_find_matching_catch_2()|0;
   $11 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $$0 = $11;$$01 = $10;
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 $12 = ((($object)) + 32|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13>>>0)>(1);
 if ($14) {
  __ZN9ExitGames6Common7JStringC2Ej($str$i,0);
  __THREW__ = 0;
  (invoke_iii(759,($str$i|0),(12344|0))|0);
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   ___resumeException($19|0);
   // unreachable;
  }
  __THREW__ = 0;
  (invoke_iii(760,($retStr|0),($str$i|0))|0);
  $17 = __THREW__; __THREW__ = 0;
  $18 = $17&1;
  if ($18) {
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   ___resumeException($19|0);
   // unreachable;
  }
  __ZN9ExitGames6Common7JStringD2Ev($str$i);
  $21 = ((($object)) + 24|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22|0)==(0|0);
  $24 = ((($object)) + 20|0);
  $25 = $23 ? $24 : $22;
  $26 = HEAP16[$25>>1]|0;
  $27 = ($26<<16>>16)>(0);
  L16: do {
   if ($27) {
    $28 = ((($object)) + 16|0);
    $29 = ((($object)) + 8|0);
    $30 = ((($object)) + 28|0);
    $31 = ((($object)) + 29|0);
    $32 = ((($1)) + 28|0);
    $33 = ((($1)) + 29|0);
    $34 = ((($1)) + 32|0);
    $35 = ((($1)) + 16|0);
    $36 = ((($1)) + 24|0);
    $37 = ((($1)) + 20|0);
    $51 = $22;$i$070 = 0;
    while(1) {
     $38 = HEAP32[$28>>2]|0;
     $39 = ($38|0)==(0|0);
     $40 = $39 ? $29 : $38;
     $41 = (($40) + ($i$070<<2)|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = HEAP8[$30>>0]|0;
     $44 = ($43<<24>>24)==(99);
     $45 = HEAP8[$31>>0]|0;
     $46 = $44 ? $45 : 0;
     $47 = HEAP32[$12>>2]|0;
     $48 = (($47) + -1)|0;
     $49 = ($47|0)==(2);
     $50 = ($51|0)==(0|0);
     $52 = $50 ? $24 : $51;
     $53 = ((($52)) + 2|0);
     if ($49) {
      $54 = HEAP16[$53>>1]|0;
      $55 = $54 << 16 >> 16;
      HEAP32[$size>>2] = $55;
      $59 = $size;
     } else {
      $59 = $53;
     }
     HEAP32[$1>>2] = (4640);
     $56 = ($43<<24>>24)!=(42);
     $57 = ($42|0)==(0|0);
     $or$cond38 = $57 & $56;
     if ($or$cond38) {
      $58 = HEAP16[$59>>1]|0;
      $60 = ($58<<16>>16)==(0);
      if ($60) {
       label = 16;
      } else {
       HEAP8[$32>>0] = 42;
       HEAP8[$33>>0] = 0;
       HEAP32[$34>>2] = 0;
       ;HEAP32[$35>>2]=0|0;HEAP32[$35+4>>2]=0|0;HEAP32[$35+8>>2]=0|0;
      }
     } else {
      label = 16;
     }
     if ((label|0) == 16) {
      label = 0;
      HEAP8[$32>>0] = $43;
      HEAP8[$33>>0] = $46;
      HEAP32[$36>>2] = 0;
      HEAP32[$34>>2] = $48;
      $61 = ($48>>>0)<(2);
      if ($61) {
       $62 = HEAP32[$59>>2]|0;
       HEAP32[$37>>2] = $62;
      } else {
       $63 = $48 << 1;
       $64 = (($63) + 8)|0;
       __THREW__ = 0;
       $65 = (invoke_ii(775,($64|0))|0);
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        label = 22;
        break;
       }
       $68 = ((($65)) + 8|0);
       $69 = $65;
       $70 = $69;
       HEAP32[$70>>2] = $48;
       $71 = (($69) + 4)|0;
       $72 = $71;
       HEAP32[$72>>2] = 0;
       _memset(($68|0),0,($63|0))|0;
       HEAP32[$36>>2] = $68;
       $i$01$i$i = 0;
       while(1) {
        $73 = (($59) + ($i$01$i$i<<1)|0);
        $74 = HEAP16[$73>>1]|0;
        $75 = (($68) + ($i$01$i$i<<1)|0);
        HEAP16[$75>>1] = $74;
        $76 = (($i$01$i$i) + 1)|0;
        $exitcond$i$i = ($76|0)==($48|0);
        if ($exitcond$i$i) {
         break;
        } else {
         $i$01$i$i = $76;
        }
       }
      }
      HEAP32[$35>>2] = 0;
      __THREW__ = 0;
      invoke_vii(830,($1|0),($42|0));
      $77 = __THREW__; __THREW__ = 0;
      $78 = $77&1;
      if ($78) {
       label = 22;
       break;
      }
     }
     __THREW__ = 0;
     (invoke_iiiii(831,($1|0),($retStr|0),($withTypes|0),1)|0);
     $81 = __THREW__; __THREW__ = 0;
     $82 = $81&1;
     if ($82) {
      label = 31;
      break;
     }
     HEAP32[$1>>2] = (4640);
     __THREW__ = 0;
     invoke_viii(818,($1|0),(0|0),0);
     $83 = __THREW__; __THREW__ = 0;
     $84 = $83&1;
     if ($84) {
      label = 25;
      break;
     }
     __ZN9ExitGames6Common4BaseD2Ev($1);
     $87 = HEAP32[$21>>2]|0;
     $88 = ($87|0)==(0|0);
     $89 = $88 ? $24 : $87;
     $90 = HEAP16[$89>>1]|0;
     $91 = $90 << 16 >> 16;
     $92 = (($91) + -1)|0;
     $93 = ($i$070|0)<($92|0);
     if ($93) {
      __ZN9ExitGames6Common7JStringC2Ej($str$i6,0);
      __THREW__ = 0;
      (invoke_iii(759,($str$i6|0),(12352|0))|0);
      $94 = __THREW__; __THREW__ = 0;
      $95 = $94&1;
      if ($95) {
       label = 29;
       break;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i6|0))|0);
      $96 = __THREW__; __THREW__ = 0;
      $97 = $96&1;
      if ($97) {
       label = 29;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i6);
      $$pre71 = HEAP32[$21>>2]|0;
      $108 = $$pre71;
     } else {
      $108 = $87;
     }
     $106 = (($i$070) + 1)|0;
     $107 = ($108|0)==(0|0);
     $109 = $107 ? $24 : $108;
     $110 = HEAP16[$109>>1]|0;
     $111 = $110 << 16 >> 16;
     $112 = ($106|0)<($111|0);
     if ($112) {
      $51 = $108;$i$070 = $106;
     } else {
      break L16;
     }
    }
    if ((label|0) == 22) {
     $79 = ___cxa_find_matching_catch_2()|0;
     $80 = tempRet0;
     __ZN9ExitGames6Common4BaseD2Ev($1);
     ___resumeException($79|0);
     // unreachable;
    }
    else if ((label|0) == 25) {
     $85 = ___cxa_find_matching_catch_3(0|0)|0;
     $86 = tempRet0;
     __ZN9ExitGames6Common4BaseD2Ev($1);
     ___clang_call_terminate($85);
     // unreachable;
    }
    else if ((label|0) == 29) {
     $98 = ___cxa_find_matching_catch_2()|0;
     $99 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i6);
     ___resumeException($98|0);
     // unreachable;
    }
    else if ((label|0) == 31) {
     $100 = ___cxa_find_matching_catch_2()|0;
     $101 = tempRet0;
     HEAP32[$1>>2] = (4640);
     __THREW__ = 0;
     invoke_viii(818,($1|0),(0|0),0);
     $102 = __THREW__; __THREW__ = 0;
     $103 = $102&1;
     if ($103) {
      $104 = ___cxa_find_matching_catch_3(0|0)|0;
      $105 = tempRet0;
      __ZN9ExitGames6Common4BaseD2Ev($1);
      ___clang_call_terminate($104);
      // unreachable;
     }
     __ZN9ExitGames6Common4BaseD2Ev($1);
     $$0 = $101;$$01 = $100;
     ___resumeException($$01|0);
     // unreachable;
    }
   }
  } while(0);
  __ZN9ExitGames6Common7JStringC2Ej($str$i8,0);
  __THREW__ = 0;
  (invoke_iii(759,($str$i8|0),(12364|0))|0);
  $113 = __THREW__; __THREW__ = 0;
  $114 = $113&1;
  if ($114) {
   $117 = ___cxa_find_matching_catch_2()|0;
   $118 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i8);
   ___resumeException($117|0);
   // unreachable;
  }
  __THREW__ = 0;
  (invoke_iii(760,($retStr|0),($str$i8|0))|0);
  $115 = __THREW__; __THREW__ = 0;
  $116 = $115&1;
  if ($116) {
   $117 = ___cxa_find_matching_catch_2()|0;
   $118 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i8);
   ___resumeException($117|0);
   // unreachable;
  }
  __ZN9ExitGames6Common7JStringD2Ev($str$i8);
  STACKTOP = sp;return ($retStr|0);
 }
 $119 = ($13|0)==(0);
 do {
  if (!($119)) {
   __ZN9ExitGames6Common7JStringC2Ej($str$i10,0);
   __THREW__ = 0;
   (invoke_iii(759,($str$i10|0),(12344|0))|0);
   $120 = __THREW__; __THREW__ = 0;
   $121 = $120&1;
   if ($121) {
    $124 = ___cxa_find_matching_catch_2()|0;
    $125 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i10);
    ___resumeException($124|0);
    // unreachable;
   }
   __THREW__ = 0;
   (invoke_iii(760,($retStr|0),($str$i10|0))|0);
   $122 = __THREW__; __THREW__ = 0;
   $123 = $122&1;
   if ($123) {
    $124 = ___cxa_find_matching_catch_2()|0;
    $125 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i10);
    ___resumeException($124|0);
    // unreachable;
   } else {
    __ZN9ExitGames6Common7JStringD2Ev($str$i10);
    break;
   }
  }
 } while(0);
 $126 = ((($object)) + 16|0);
 $127 = HEAP32[$126>>2]|0;
 $128 = ($127|0)==(0|0);
 $129 = ((($object)) + 28|0);
 do {
  if ($128) {
   $130 = HEAP8[$129>>0]|0;
   $131 = ($130<<24>>24)!=(42);
   $132 = HEAP32[$12>>2]|0;
   $133 = ($132|0)==(0);
   $or$cond40 = $131 & $133;
   if ($or$cond40) {
    label = 46;
   } else {
    $330 = ($130<<24>>24)==(42);
    if ($330) {
     __ZN9ExitGames6Common7JStringC2Ej($str$i2,0);
     __THREW__ = 0;
     (invoke_iii(759,($str$i2|0),(5604|0))|0);
     $331 = __THREW__; __THREW__ = 0;
     $332 = $331&1;
     if ($332) {
      $335 = ___cxa_find_matching_catch_2()|0;
      $336 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($str$i2);
      ___resumeException($335|0);
      // unreachable;
     }
     __THREW__ = 0;
     (invoke_iii(760,($retStr|0),($str$i2|0))|0);
     $333 = __THREW__; __THREW__ = 0;
     $334 = $333&1;
     if ($334) {
      $335 = ___cxa_find_matching_catch_2()|0;
      $336 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($str$i2);
      ___resumeException($335|0);
      // unreachable;
     } else {
      __ZN9ExitGames6Common7JStringD2Ev($str$i2);
      $$pre = HEAP32[$12>>2]|0;
      $338 = $$pre;
      break;
     }
    } else {
     $338 = $132;
    }
   }
  } else {
   label = 46;
  }
 } while(0);
 L82: do {
  if ((label|0) == 46) {
   $134 = ((($object)) + 24|0);
   $135 = ((($object)) + 20|0);
   $136 = ((($object)) + 8|0);
   $137 = ((($object)) + 29|0);
   $i1$0 = 0;
   L84: while(1) {
    $138 = HEAP8[$129>>0]|0;
    $139 = ($138<<24>>24)==(98);
    $140 = HEAP32[$12>>2]|0;
    $141 = ($140|0)==(1);
    $or$cond42 = $139 & $141;
    $142 = HEAP32[$134>>2]|0;
    $143 = ($142|0)==(0|0);
    $144 = $143 ? $135 : $142;
    if ($or$cond42) {
     $145 = HEAP32[$144>>2]|0;
     $149 = $145;
    } else {
     $146 = HEAP16[$144>>1]|0;
     $147 = $146 << 16 >> 16;
     $149 = $147;
    }
    $148 = ($i1$0|0)<($149|0);
    if (!($148)) {
     $338 = $140;
     break L82;
    }
    $150 = $138&255;
    do {
     switch ($150|0) {
     case 98:  {
      $151 = HEAP32[$126>>2]|0;
      $152 = ($151|0)==(0|0);
      if ($152) {
       $153 = ($138<<24>>24)!=(42);
       $154 = ($140|0)==(0);
       $or$cond44 = $153 & $154;
       $$ = $or$cond44 ? $136 : 0;
       $156 = $$;
      } else {
       $156 = $151;
      }
      $155 = (($156) + ($i1$0)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i14,0);
      $157 = HEAP8[$155>>0]|0;
      __THREW__ = 0;
      (invoke_iii(782,($str$i14|0),($157|0))|0);
      $158 = __THREW__; __THREW__ = 0;
      $159 = $158&1;
      if ($159) {
       label = 56;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i14|0))|0);
      $160 = __THREW__; __THREW__ = 0;
      $161 = $160&1;
      if ($161) {
       label = 56;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i14);
      break;
     }
     case 107:  {
      $164 = HEAP32[$126>>2]|0;
      $165 = ($164|0)==(0|0);
      if ($165) {
       $166 = ($138<<24>>24)!=(42);
       $167 = ($140|0)==(0);
       $or$cond46 = $166 & $167;
       $$88 = $or$cond46 ? $136 : 0;
       $169 = $$88;
      } else {
       $169 = $164;
      }
      $168 = (($169) + ($i1$0<<1)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i16,0);
      $170 = HEAP16[$168>>1]|0;
      __THREW__ = 0;
      (invoke_iii(832,($str$i16|0),($170|0))|0);
      $171 = __THREW__; __THREW__ = 0;
      $172 = $171&1;
      if ($172) {
       label = 62;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i16|0))|0);
      $173 = __THREW__; __THREW__ = 0;
      $174 = $173&1;
      if ($174) {
       label = 62;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i16);
      break;
     }
     case 105:  {
      $177 = HEAP32[$126>>2]|0;
      $178 = ($177|0)==(0|0);
      if ($178) {
       $179 = ($138<<24>>24)!=(42);
       $180 = ($140|0)==(0);
       $or$cond48 = $179 & $180;
       $$89 = $or$cond48 ? $136 : 0;
       $182 = $$89;
      } else {
       $182 = $177;
      }
      $181 = (($182) + ($i1$0<<2)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i18,0);
      $183 = HEAP32[$181>>2]|0;
      __THREW__ = 0;
      (invoke_iii(833,($str$i18|0),($183|0))|0);
      $184 = __THREW__; __THREW__ = 0;
      $185 = $184&1;
      if ($185) {
       label = 68;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i18|0))|0);
      $186 = __THREW__; __THREW__ = 0;
      $187 = $186&1;
      if ($187) {
       label = 68;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i18);
      break;
     }
     case 108:  {
      $190 = HEAP32[$126>>2]|0;
      $191 = ($190|0)==(0|0);
      if ($191) {
       $192 = ($138<<24>>24)!=(42);
       $193 = ($140|0)==(0);
       $or$cond50 = $192 & $193;
       $$90 = $or$cond50 ? $136 : 0;
       $195 = $$90;
      } else {
       $195 = $190;
      }
      $194 = (($195) + ($i1$0<<3)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i20,0);
      $196 = $194;
      $197 = $196;
      $198 = HEAP32[$197>>2]|0;
      $199 = (($196) + 4)|0;
      $200 = $199;
      $201 = HEAP32[$200>>2]|0;
      __THREW__ = 0;
      (invoke_iiii(834,($str$i20|0),($198|0),($201|0))|0);
      $202 = __THREW__; __THREW__ = 0;
      $203 = $202&1;
      if ($203) {
       label = 74;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i20|0))|0);
      $204 = __THREW__; __THREW__ = 0;
      $205 = $204&1;
      if ($205) {
       label = 74;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i20);
      break;
     }
     case 102:  {
      $208 = HEAP32[$126>>2]|0;
      $209 = ($208|0)==(0|0);
      if ($209) {
       $210 = ($138<<24>>24)!=(42);
       $211 = ($140|0)==(0);
       $or$cond52 = $210 & $211;
       $$91 = $or$cond52 ? $136 : 0;
       $213 = $$91;
      } else {
       $213 = $208;
      }
      $212 = (($213) + ($i1$0<<2)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i22,0);
      $214 = +HEAPF32[$212>>2];
      __THREW__ = 0;
      (invoke_iid(835,($str$i22|0),(+$214))|0);
      $215 = __THREW__; __THREW__ = 0;
      $216 = $215&1;
      if ($216) {
       label = 80;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i22|0))|0);
      $217 = __THREW__; __THREW__ = 0;
      $218 = $217&1;
      if ($218) {
       label = 80;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i22);
      break;
     }
     case 100:  {
      $221 = HEAP32[$126>>2]|0;
      $222 = ($221|0)==(0|0);
      if ($222) {
       $223 = ($138<<24>>24)!=(42);
       $224 = ($140|0)==(0);
       $or$cond54 = $223 & $224;
       $$92 = $or$cond54 ? $136 : 0;
       $226 = $$92;
      } else {
       $226 = $221;
      }
      $225 = (($226) + ($i1$0<<3)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i24,0);
      $227 = +HEAPF64[$225>>3];
      __THREW__ = 0;
      (invoke_iid(836,($str$i24|0),(+$227))|0);
      $228 = __THREW__; __THREW__ = 0;
      $229 = $228&1;
      if ($229) {
       label = 86;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i24|0))|0);
      $230 = __THREW__; __THREW__ = 0;
      $231 = $230&1;
      if ($231) {
       label = 86;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i24);
      break;
     }
     case 111:  {
      $234 = HEAP32[$126>>2]|0;
      $235 = ($234|0)==(0|0);
      if ($235) {
       $236 = ($138<<24>>24)!=(42);
       $237 = ($140|0)==(0);
       $or$cond56 = $236 & $237;
       $$93 = $or$cond56 ? $136 : 0;
       $239 = $$93;
      } else {
       $239 = $234;
      }
      $238 = (($239) + ($i1$0)|0);
      __ZN9ExitGames6Common7JStringC2Ej($str$i26,0);
      $240 = HEAP8[$238>>0]|0;
      $241 = ($240<<24>>24)!=(0);
      __THREW__ = 0;
      (invoke_iii(837,($str$i26|0),($241|0))|0);
      $242 = __THREW__; __THREW__ = 0;
      $243 = $242&1;
      if ($243) {
       label = 92;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i26|0))|0);
      $244 = __THREW__; __THREW__ = 0;
      $245 = $244&1;
      if ($245) {
       label = 92;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i26);
      break;
     }
     case 115:  {
      $248 = HEAP32[$126>>2]|0;
      $249 = ($248|0)==(0|0);
      if ($249) {
       $250 = ($138<<24>>24)!=(42);
       $251 = ($140|0)==(0);
       $or$cond58 = $250 & $251;
       $$94 = $or$cond58 ? $136 : 0;
       $253 = $$94;
      } else {
       $253 = $248;
      }
      $252 = (($253) + ($i1$0<<4)|0);
      __ZNK9ExitGames6Common8ToString8toStringEb($2,$252,$withTypes);
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($2|0))|0);
      $254 = __THREW__; __THREW__ = 0;
      $255 = $254&1;
      if ($255) {
       label = 98;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($2);
      break;
     }
     case 104:  {
      $258 = HEAP32[$126>>2]|0;
      $259 = ($258|0)==(0|0);
      if ($259) {
       $260 = ($138<<24>>24)!=(42);
       $261 = ($140|0)==(0);
       $or$cond60 = $260 & $261;
       $$95 = $or$cond60 ? $136 : 0;
       $263 = $$95;
      } else {
       $263 = $258;
      }
      $262 = (($263) + (($i1$0*44)|0)|0);
      __ZNK9ExitGames6Common8ToString8toStringEb($3,$262,$withTypes);
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($3|0))|0);
      $264 = __THREW__; __THREW__ = 0;
      $265 = $264&1;
      if ($265) {
       label = 103;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($3);
      break;
     }
     case 68:  {
      $268 = HEAP32[$126>>2]|0;
      $269 = ($268|0)==(0|0);
      if ($269) {
       $270 = ($138<<24>>24)!=(42);
       $271 = ($140|0)==(0);
       $or$cond62 = $270 & $271;
       $$96 = $or$cond62 ? $136 : 0;
       $273 = $$96;
      } else {
       $273 = $268;
      }
      $272 = (($273) + (($i1$0*52)|0)|0);
      __ZNK9ExitGames6Common8ToString8toStringEb($4,$272,$withTypes);
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($4|0))|0);
      $274 = __THREW__; __THREW__ = 0;
      $275 = $274&1;
      if ($275) {
       label = 108;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($4);
      break;
     }
     case 122:  {
      $278 = HEAP32[$126>>2]|0;
      $279 = ($278|0)==(0|0);
      if ($279) {
       $280 = ($138<<24>>24)!=(42);
       $281 = ($140|0)==(0);
       $or$cond64 = $280 & $281;
       $$97 = $or$cond64 ? $136 : 0;
       $283 = $$97;
      } else {
       $283 = $278;
      }
      $282 = (($283) + (($i1$0*40)|0)|0);
      __ZNK9ExitGames6Common8ToString8toStringEb($5,$282,$withTypes);
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($5|0))|0);
      $284 = __THREW__; __THREW__ = 0;
      $285 = $284&1;
      if ($285) {
       label = 113;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($5);
      break;
     }
     case 99:  {
      $288 = HEAP32[$126>>2]|0;
      $289 = ($288|0)==(0|0);
      if ($289) {
       $290 = ($138<<24>>24)!=(42);
       $291 = ($140|0)==(0);
       $or$cond66 = $290 & $291;
       $$98 = $or$cond66 ? $136 : 0;
       $298 = $$98;
      } else {
       $298 = $288;
      }
      $292 = ($138<<24>>24)==(99);
      $293 = HEAP8[$137>>0]|0;
      $294 = $292 ? $293 : 0;
      $295 = (__ZN9ExitGames6Common14CustomTypeBase9getSizeofEh($294)|0);
      $296 = Math_imul($295, $i1$0)|0;
      $297 = (($298) + ($296)|0);
      __ZNK9ExitGames6Common8ToString8toStringEb($6,$297,$withTypes);
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($6|0))|0);
      $299 = __THREW__; __THREW__ = 0;
      $300 = $299&1;
      if ($300) {
       label = 118;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($6);
      break;
     }
     default: {
      __ZN9ExitGames6Common7JStringC2EPKc($7,145982);
      __THREW__ = 0;
      $303 = (invoke_ii(764,($7|0))|0);
      $304 = __THREW__; __THREW__ = 0;
      $305 = $304&1;
      if ($305) {
       label = 125;
       break L84;
      }
      __THREW__ = 0;
      invoke_viiiiiiii(765,(162472|0),1,(5244|0),($303|0),1,597,(5304|0),($vararg_buffer|0));
      $306 = __THREW__; __THREW__ = 0;
      $307 = $306&1;
      if ($307) {
       label = 125;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($7);
      __ZN9ExitGames6Common7JStringC2Ej($str$i33,0);
      __THREW__ = 0;
      (invoke_iii(759,($str$i33|0),(5304|0))|0);
      $308 = __THREW__; __THREW__ = 0;
      $309 = $308&1;
      if ($309) {
       label = 123;
       break L84;
      }
      __THREW__ = 0;
      (invoke_iii(760,($retStr|0),($str$i33|0))|0);
      $310 = __THREW__; __THREW__ = 0;
      $311 = $310&1;
      if ($311) {
       label = 123;
       break L84;
      }
      __ZN9ExitGames6Common7JStringD2Ev($str$i33);
     }
     }
    } while(0);
    $316 = HEAP32[$134>>2]|0;
    $317 = ($316|0)==(0|0);
    $318 = $317 ? $135 : $316;
    $319 = HEAP16[$318>>1]|0;
    $320 = $319 << 16 >> 16;
    $321 = (($320) + -1)|0;
    $322 = ($i1$0|0)<($321|0);
    if ($322) {
     __ZN9ExitGames6Common7JStringC2Ej($str$i34,0);
     __THREW__ = 0;
     (invoke_iii(759,($str$i34|0),(12352|0))|0);
     $323 = __THREW__; __THREW__ = 0;
     $324 = $323&1;
     if ($324) {
      label = 129;
      break;
     }
     __THREW__ = 0;
     (invoke_iii(760,($retStr|0),($str$i34|0))|0);
     $325 = __THREW__; __THREW__ = 0;
     $326 = $325&1;
     if ($326) {
      label = 129;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($str$i34);
    }
    $329 = (($i1$0) + 1)|0;
    $i1$0 = $329;
   }
   switch (label|0) {
    case 56: {
     $162 = ___cxa_find_matching_catch_2()|0;
     $163 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i14);
     ___resumeException($162|0);
     // unreachable;
     break;
    }
    case 62: {
     $175 = ___cxa_find_matching_catch_2()|0;
     $176 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i16);
     ___resumeException($175|0);
     // unreachable;
     break;
    }
    case 68: {
     $188 = ___cxa_find_matching_catch_2()|0;
     $189 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i18);
     ___resumeException($188|0);
     // unreachable;
     break;
    }
    case 74: {
     $206 = ___cxa_find_matching_catch_2()|0;
     $207 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i20);
     ___resumeException($206|0);
     // unreachable;
     break;
    }
    case 80: {
     $219 = ___cxa_find_matching_catch_2()|0;
     $220 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i22);
     ___resumeException($219|0);
     // unreachable;
     break;
    }
    case 86: {
     $232 = ___cxa_find_matching_catch_2()|0;
     $233 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i24);
     ___resumeException($232|0);
     // unreachable;
     break;
    }
    case 92: {
     $246 = ___cxa_find_matching_catch_2()|0;
     $247 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i26);
     ___resumeException($246|0);
     // unreachable;
     break;
    }
    case 98: {
     $256 = ___cxa_find_matching_catch_2()|0;
     $257 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($2);
     $$0 = $257;$$01 = $256;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 103: {
     $266 = ___cxa_find_matching_catch_2()|0;
     $267 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($3);
     $$0 = $267;$$01 = $266;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 108: {
     $276 = ___cxa_find_matching_catch_2()|0;
     $277 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($4);
     $$0 = $277;$$01 = $276;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 113: {
     $286 = ___cxa_find_matching_catch_2()|0;
     $287 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($5);
     $$0 = $287;$$01 = $286;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 118: {
     $301 = ___cxa_find_matching_catch_2()|0;
     $302 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($6);
     $$0 = $302;$$01 = $301;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 123: {
     $312 = ___cxa_find_matching_catch_2()|0;
     $313 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i33);
     ___resumeException($312|0);
     // unreachable;
     break;
    }
    case 125: {
     $314 = ___cxa_find_matching_catch_2()|0;
     $315 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($7);
     $$0 = $315;$$01 = $314;
     ___resumeException($$01|0);
     // unreachable;
     break;
    }
    case 129: {
     $327 = ___cxa_find_matching_catch_2()|0;
     $328 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i34);
     ___resumeException($327|0);
     // unreachable;
     break;
    }
   }
  }
 } while(0);
 $337 = ($338|0)==(0);
 if ($337) {
  STACKTOP = sp;return ($retStr|0);
 }
 __ZN9ExitGames6Common7JStringC2Ej($str$i4,0);
 __THREW__ = 0;
 (invoke_iii(759,($str$i4|0),(12364|0))|0);
 $339 = __THREW__; __THREW__ = 0;
 $340 = $339&1;
 if ($340) {
  $343 = ___cxa_find_matching_catch_2()|0;
  $344 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i4);
  ___resumeException($343|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(760,($retStr|0),($str$i4|0))|0);
 $341 = __THREW__; __THREW__ = 0;
 $342 = $341&1;
 if ($342) {
  $343 = ___cxa_find_matching_catch_2()|0;
  $344 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i4);
  ___resumeException($343|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($str$i4);
 STACKTOP = sp;return ($retStr|0);
}
function __ZNK9ExitGames6Common6Object19payloadTypeToStringEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$10 = 0, $$1017 = 0, $$11 = 0, $$1118 = 0, $$12 = 0, $$1219 = 0, $$13 = 0, $$1320 = 0, $$14 = 0, $$1421 = 0, $$15 = 0, $$1522 = 0, $$16 = 0, $$1623 = 0, $$17 = 0, $$1724 = 0, $$18 = 0;
 var $$2 = 0, $$29 = 0, $$3 = 0, $$310 = 0, $$4 = 0, $$411 = 0, $$5 = 0, $$512 = 0, $$6 = 0, $$613 = 0, $$7 = 0, $$714 = 0, $$8 = 0, $$815 = 0, $$9 = 0, $$916 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0;
 var $282 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index2Z2D = 0;
 var $eh$lpad$body$indexZ2D = 0, $pData$026 = 0, $pData$1 = 0, $retStr = 0, $str$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 576|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(576|0);
 $vararg_buffer = sp + 40|0;
 $str$i = sp + 560|0;
 $retStr = sp + 544|0;
 $0 = sp + 528|0;
 $1 = sp + 512|0;
 $2 = sp + 496|0;
 $3 = sp + 480|0;
 $4 = sp + 464|0;
 $5 = sp + 448|0;
 $6 = sp + 432|0;
 $7 = sp + 416|0;
 $8 = sp + 400|0;
 $9 = sp + 384|0;
 $10 = sp + 368|0;
 $11 = sp + 352|0;
 $12 = sp + 336|0;
 $13 = sp + 320|0;
 $14 = sp + 304|0;
 $15 = sp + 288|0;
 $16 = sp + 272|0;
 $17 = sp + 256|0;
 $18 = sp + 240|0;
 $19 = sp + 224|0;
 $20 = sp + 208|0;
 $21 = sp + 192|0;
 $22 = sp + 176|0;
 $23 = sp + 160|0;
 $24 = sp + 112|0;
 $25 = sp + 96|0;
 $26 = sp + 80|0;
 $27 = sp;
 $28 = sp + 64|0;
 $29 = sp + 48|0;
 __ZN9ExitGames6Common7JStringC2Ej($retStr,0);
 $30 = ((($this)) + 28|0);
 $31 = HEAP8[$30>>0]|0;
 $32 = $31&255;
 L1: do {
  switch ($32|0) {
  case 98:  {
   __THREW__ = 0;
   invoke_vii(801,($2|0),(13660|0));
   $35 = __THREW__; __THREW__ = 0;
   $36 = $35&1;
   if ($36) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($1|0),($2|0),(162512|0));
    $37 = __THREW__; __THREW__ = 0;
    $38 = $37&1;
    if ($38) {
     $46 = ___cxa_find_matching_catch_2()|0;
     $47 = tempRet0;
     $$1 = $46;$$18 = $47;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($0|0),($1|0),(14336|0));
     $39 = __THREW__; __THREW__ = 0;
     $40 = $39&1;
     if ($40) {
      $48 = ___cxa_find_matching_catch_2()|0;
      $49 = tempRet0;
      $$0 = $48;$$01 = $49;
     } else {
      __THREW__ = 0;
      $41 = (invoke_iii(803,($retStr|0),($0|0))|0);
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if (!($43)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($41|0));
       $44 = __THREW__; __THREW__ = 0;
       $45 = $44&1;
       if (!($45)) {
        __ZN9ExitGames6Common7JStringD2Ev($0);
        __ZN9ExitGames6Common7JStringD2Ev($1);
        __ZN9ExitGames6Common7JStringD2Ev($2);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $50 = ___cxa_find_matching_catch_2()|0;
      $51 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($0);
      $$0 = $50;$$01 = $51;
     }
     __ZN9ExitGames6Common7JStringD2Ev($1);
     $$1 = $$0;$$18 = $$01;
    }
    __ZN9ExitGames6Common7JStringD2Ev($2);
    $$17 = $$1;$$1724 = $$18;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 107:  {
   __THREW__ = 0;
   invoke_vii(801,($5|0),(13660|0));
   $52 = __THREW__; __THREW__ = 0;
   $53 = $52&1;
   if ($53) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($4|0),($5|0),(162528|0));
    $54 = __THREW__; __THREW__ = 0;
    $55 = $54&1;
    if ($55) {
     $63 = ___cxa_find_matching_catch_2()|0;
     $64 = tempRet0;
     $$3 = $63;$$310 = $64;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($3|0),($4|0),(14336|0));
     $56 = __THREW__; __THREW__ = 0;
     $57 = $56&1;
     if ($57) {
      $65 = ___cxa_find_matching_catch_2()|0;
      $66 = tempRet0;
      $$2 = $65;$$29 = $66;
     } else {
      __THREW__ = 0;
      $58 = (invoke_iii(803,($retStr|0),($3|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if (!($60)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($58|0));
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if (!($62)) {
        __ZN9ExitGames6Common7JStringD2Ev($3);
        __ZN9ExitGames6Common7JStringD2Ev($4);
        __ZN9ExitGames6Common7JStringD2Ev($5);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $67 = ___cxa_find_matching_catch_2()|0;
      $68 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($3);
      $$2 = $67;$$29 = $68;
     }
     __ZN9ExitGames6Common7JStringD2Ev($4);
     $$3 = $$2;$$310 = $$29;
    }
    __ZN9ExitGames6Common7JStringD2Ev($5);
    $$17 = $$3;$$1724 = $$310;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 105:  {
   __THREW__ = 0;
   invoke_vii(801,($8|0),(13660|0));
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($7|0),($8|0),(162560|0));
    $71 = __THREW__; __THREW__ = 0;
    $72 = $71&1;
    if ($72) {
     $80 = ___cxa_find_matching_catch_2()|0;
     $81 = tempRet0;
     $$5 = $80;$$512 = $81;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($6|0),($7|0),(14336|0));
     $73 = __THREW__; __THREW__ = 0;
     $74 = $73&1;
     if ($74) {
      $82 = ___cxa_find_matching_catch_2()|0;
      $83 = tempRet0;
      $$4 = $82;$$411 = $83;
     } else {
      __THREW__ = 0;
      $75 = (invoke_iii(803,($retStr|0),($6|0))|0);
      $76 = __THREW__; __THREW__ = 0;
      $77 = $76&1;
      if (!($77)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($75|0));
       $78 = __THREW__; __THREW__ = 0;
       $79 = $78&1;
       if (!($79)) {
        __ZN9ExitGames6Common7JStringD2Ev($6);
        __ZN9ExitGames6Common7JStringD2Ev($7);
        __ZN9ExitGames6Common7JStringD2Ev($8);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $84 = ___cxa_find_matching_catch_2()|0;
      $85 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($6);
      $$4 = $84;$$411 = $85;
     }
     __ZN9ExitGames6Common7JStringD2Ev($7);
     $$5 = $$4;$$512 = $$411;
    }
    __ZN9ExitGames6Common7JStringD2Ev($8);
    $$17 = $$5;$$1724 = $$512;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 108:  {
   __THREW__ = 0;
   invoke_vii(801,($11|0),(13660|0));
   $86 = __THREW__; __THREW__ = 0;
   $87 = $86&1;
   if ($87) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($10|0),($11|0),(162624|0));
    $88 = __THREW__; __THREW__ = 0;
    $89 = $88&1;
    if ($89) {
     $97 = ___cxa_find_matching_catch_2()|0;
     $98 = tempRet0;
     $$7 = $97;$$714 = $98;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($9|0),($10|0),(14336|0));
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      $99 = ___cxa_find_matching_catch_2()|0;
      $100 = tempRet0;
      $$6 = $99;$$613 = $100;
     } else {
      __THREW__ = 0;
      $92 = (invoke_iii(803,($retStr|0),($9|0))|0);
      $93 = __THREW__; __THREW__ = 0;
      $94 = $93&1;
      if (!($94)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($92|0));
       $95 = __THREW__; __THREW__ = 0;
       $96 = $95&1;
       if (!($96)) {
        __ZN9ExitGames6Common7JStringD2Ev($9);
        __ZN9ExitGames6Common7JStringD2Ev($10);
        __ZN9ExitGames6Common7JStringD2Ev($11);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $101 = ___cxa_find_matching_catch_2()|0;
      $102 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($9);
      $$6 = $101;$$613 = $102;
     }
     __ZN9ExitGames6Common7JStringD2Ev($10);
     $$7 = $$6;$$714 = $$613;
    }
    __ZN9ExitGames6Common7JStringD2Ev($11);
    $$17 = $$7;$$1724 = $$714;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 102:  {
   __THREW__ = 0;
   invoke_vii(801,($14|0),(13660|0));
   $103 = __THREW__; __THREW__ = 0;
   $104 = $103&1;
   if ($104) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($13|0),($14|0),(162656|0));
    $105 = __THREW__; __THREW__ = 0;
    $106 = $105&1;
    if ($106) {
     $114 = ___cxa_find_matching_catch_2()|0;
     $115 = tempRet0;
     $$9 = $114;$$916 = $115;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($12|0),($13|0),(14336|0));
     $107 = __THREW__; __THREW__ = 0;
     $108 = $107&1;
     if ($108) {
      $116 = ___cxa_find_matching_catch_2()|0;
      $117 = tempRet0;
      $$8 = $116;$$815 = $117;
     } else {
      __THREW__ = 0;
      $109 = (invoke_iii(803,($retStr|0),($12|0))|0);
      $110 = __THREW__; __THREW__ = 0;
      $111 = $110&1;
      if (!($111)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($109|0));
       $112 = __THREW__; __THREW__ = 0;
       $113 = $112&1;
       if (!($113)) {
        __ZN9ExitGames6Common7JStringD2Ev($12);
        __ZN9ExitGames6Common7JStringD2Ev($13);
        __ZN9ExitGames6Common7JStringD2Ev($14);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $118 = ___cxa_find_matching_catch_2()|0;
      $119 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($12);
      $$8 = $118;$$815 = $119;
     }
     __ZN9ExitGames6Common7JStringD2Ev($13);
     $$9 = $$8;$$916 = $$815;
    }
    __ZN9ExitGames6Common7JStringD2Ev($14);
    $$17 = $$9;$$1724 = $$916;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 100:  {
   __THREW__ = 0;
   invoke_vii(801,($17|0),(13660|0));
   $120 = __THREW__; __THREW__ = 0;
   $121 = $120&1;
   if ($121) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($16|0),($17|0),(162672|0));
    $122 = __THREW__; __THREW__ = 0;
    $123 = $122&1;
    if ($123) {
     $131 = ___cxa_find_matching_catch_2()|0;
     $132 = tempRet0;
     $$11 = $131;$$1118 = $132;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($15|0),($16|0),(14336|0));
     $124 = __THREW__; __THREW__ = 0;
     $125 = $124&1;
     if ($125) {
      $133 = ___cxa_find_matching_catch_2()|0;
      $134 = tempRet0;
      $$10 = $133;$$1017 = $134;
     } else {
      __THREW__ = 0;
      $126 = (invoke_iii(803,($retStr|0),($15|0))|0);
      $127 = __THREW__; __THREW__ = 0;
      $128 = $127&1;
      if (!($128)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($126|0));
       $129 = __THREW__; __THREW__ = 0;
       $130 = $129&1;
       if (!($130)) {
        __ZN9ExitGames6Common7JStringD2Ev($15);
        __ZN9ExitGames6Common7JStringD2Ev($16);
        __ZN9ExitGames6Common7JStringD2Ev($17);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $135 = ___cxa_find_matching_catch_2()|0;
      $136 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($15);
      $$10 = $135;$$1017 = $136;
     }
     __ZN9ExitGames6Common7JStringD2Ev($16);
     $$11 = $$10;$$1118 = $$1017;
    }
    __ZN9ExitGames6Common7JStringD2Ev($17);
    $$17 = $$11;$$1724 = $$1118;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 111:  {
   __THREW__ = 0;
   invoke_vii(801,($20|0),(13660|0));
   $137 = __THREW__; __THREW__ = 0;
   $138 = $137&1;
   if ($138) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($19|0),($20|0),(162704|0));
    $139 = __THREW__; __THREW__ = 0;
    $140 = $139&1;
    if ($140) {
     $148 = ___cxa_find_matching_catch_2()|0;
     $149 = tempRet0;
     $$13 = $148;$$1320 = $149;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($18|0),($19|0),(14336|0));
     $141 = __THREW__; __THREW__ = 0;
     $142 = $141&1;
     if ($142) {
      $150 = ___cxa_find_matching_catch_2()|0;
      $151 = tempRet0;
      $$12 = $150;$$1219 = $151;
     } else {
      __THREW__ = 0;
      $143 = (invoke_iii(803,($retStr|0),($18|0))|0);
      $144 = __THREW__; __THREW__ = 0;
      $145 = $144&1;
      if (!($145)) {
       __THREW__ = 0;
       invoke_vii(786,($agg$result|0),($143|0));
       $146 = __THREW__; __THREW__ = 0;
       $147 = $146&1;
       if (!($147)) {
        __ZN9ExitGames6Common7JStringD2Ev($18);
        __ZN9ExitGames6Common7JStringD2Ev($19);
        __ZN9ExitGames6Common7JStringD2Ev($20);
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $152 = ___cxa_find_matching_catch_2()|0;
      $153 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($18);
      $$12 = $152;$$1219 = $153;
     }
     __ZN9ExitGames6Common7JStringD2Ev($19);
     $$13 = $$12;$$1320 = $$1219;
    }
    __ZN9ExitGames6Common7JStringD2Ev($20);
    $$17 = $$13;$$1724 = $$1320;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 42:  {
   __THREW__ = 0;
   $154 = (invoke_iii(759,($retStr|0),(5480|0))|0);
   $155 = __THREW__; __THREW__ = 0;
   $156 = $155&1;
   if ($156) {
    label = 2;
   } else {
    __THREW__ = 0;
    invoke_vii(786,($agg$result|0),($154|0));
    $157 = __THREW__; __THREW__ = 0;
    $158 = $157&1;
    if ($158) {
     label = 2;
    } else {
     __ZN9ExitGames6Common7JStringD2Ev($retStr);
     STACKTOP = sp;return;
    }
   }
   break;
  }
  case 115:  {
   __THREW__ = 0;
   invoke_vii(787,($22|0),0);
   $159 = __THREW__; __THREW__ = 0;
   $160 = $159&1;
   if ($160) {
    label = 2;
   } else {
    $161 = HEAP32[$22>>2]|0;
    $162 = ((($161)) + 8|0);
    $163 = HEAP32[$162>>2]|0;
    __THREW__ = 0;
    invoke_vii($163|0,($21|0),($22|0));
    $164 = __THREW__; __THREW__ = 0;
    $165 = $164&1;
    if ($165) {
     $171 = ___cxa_find_matching_catch_2()|0;
     $172 = tempRet0;
     $$14 = $171;$$1421 = $172;
    } else {
     __THREW__ = 0;
     $166 = (invoke_iii(803,($retStr|0),($21|0))|0);
     $167 = __THREW__; __THREW__ = 0;
     $168 = $167&1;
     if (!($168)) {
      __THREW__ = 0;
      invoke_vii(786,($agg$result|0),($166|0));
      $169 = __THREW__; __THREW__ = 0;
      $170 = $169&1;
      if (!($170)) {
       __ZN9ExitGames6Common7JStringD2Ev($21);
       __ZN9ExitGames6Common7JStringD2Ev($22);
       __ZN9ExitGames6Common7JStringD2Ev($retStr);
       STACKTOP = sp;return;
      }
     }
     $173 = ___cxa_find_matching_catch_2()|0;
     $174 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($21);
     $$14 = $173;$$1421 = $174;
    }
    __ZN9ExitGames6Common7JStringD2Ev($22);
    $$17 = $$14;$$1724 = $$1421;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 104:  {
   __THREW__ = 0;
   invoke_vi(838,($24|0));
   $175 = __THREW__; __THREW__ = 0;
   $176 = $175&1;
   if ($176) {
    label = 2;
   } else {
    $177 = HEAP32[$24>>2]|0;
    $178 = ((($177)) + 8|0);
    $179 = HEAP32[$178>>2]|0;
    __THREW__ = 0;
    invoke_vii($179|0,($23|0),($24|0));
    $180 = __THREW__; __THREW__ = 0;
    $181 = $180&1;
    if ($181) {
     $187 = ___cxa_find_matching_catch_2()|0;
     $188 = tempRet0;
     $$15 = $187;$$1522 = $188;
    } else {
     __THREW__ = 0;
     $182 = (invoke_iii(803,($retStr|0),($23|0))|0);
     $183 = __THREW__; __THREW__ = 0;
     $184 = $183&1;
     if (!($184)) {
      __THREW__ = 0;
      invoke_vii(786,($agg$result|0),($182|0));
      $185 = __THREW__; __THREW__ = 0;
      $186 = $185&1;
      if (!($186)) {
       __ZN9ExitGames6Common7JStringD2Ev($23);
       __ZN9ExitGames6Common9HashtableD2Ev($24);
       __ZN9ExitGames6Common7JStringD2Ev($retStr);
       STACKTOP = sp;return;
      }
     }
     $189 = ___cxa_find_matching_catch_2()|0;
     $190 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($23);
     $$15 = $189;$$1522 = $190;
    }
    __ZN9ExitGames6Common9HashtableD2Ev($24);
    $$17 = $$15;$$1724 = $$1522;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 68:  {
   $191 = ((($this)) + 16|0);
   $192 = HEAP32[$191>>2]|0;
   $193 = ($192|0)==(0|0);
   if ($193) {
    $194 = ($31<<24>>24)==(42);
    if ($194) {
     $282 = 0;
    } else {
     $195 = ((($this)) + 32|0);
     $196 = HEAP32[$195>>2]|0;
     $197 = ($196|0)==(0);
     if ($197) {
      $198 = ((($this)) + 8|0);
      $282 = $198;
     } else {
      $282 = 0;
     }
    }
   } else {
    $282 = $192;
   }
   $199 = ((($this)) + 32|0);
   $200 = HEAP32[$199>>2]|0;
   $201 = ($200|0)==(0);
   if ($201) {
    $pData$1 = $282;
   } else {
    $202 = (($200) + -1)|0;
    $203 = ($202|0)==(0);
    if ($203) {
     $pData$1 = $282;
    } else {
     $206 = $202;$pData$026 = $282;
     while(1) {
      $204 = HEAP32[$pData$026>>2]|0;
      $205 = (($206) + -1)|0;
      $207 = ($205|0)==(0);
      if ($207) {
       $pData$1 = $204;
       break;
      } else {
       $206 = $205;$pData$026 = $204;
      }
     }
    }
   }
   $208 = HEAP32[$pData$1>>2]|0;
   $209 = ((($208)) + 8|0);
   $210 = HEAP32[$209>>2]|0;
   __THREW__ = 0;
   invoke_vii($210|0,($25|0),($pData$1|0));
   $211 = __THREW__; __THREW__ = 0;
   $212 = $211&1;
   if ($212) {
    label = 2;
   } else {
    __THREW__ = 0;
    $213 = (invoke_iii(803,($retStr|0),($25|0))|0);
    $214 = __THREW__; __THREW__ = 0;
    $215 = $214&1;
    if (!($215)) {
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($213|0));
     $216 = __THREW__; __THREW__ = 0;
     $217 = $216&1;
     if (!($217)) {
      __ZN9ExitGames6Common7JStringD2Ev($25);
      __ZN9ExitGames6Common7JStringD2Ev($retStr);
      STACKTOP = sp;return;
     }
    }
    $218 = ___cxa_find_matching_catch_2()|0;
    $219 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($25);
    $$17 = $218;$$1724 = $219;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
   break;
  }
  case 122:  {
   HEAP32[$27>>2] = (4640);
   $220 = ((($27)) + 28|0);
   HEAP8[$220>>0] = 42;
   $221 = ((($27)) + 29|0);
   HEAP8[$221>>0] = 0;
   $222 = ((($27)) + 32|0);
   HEAP32[$222>>2] = 0;
   $223 = ((($27)) + 16|0);
   ;HEAP32[$223>>2]=0|0;HEAP32[$223+4>>2]=0|0;HEAP32[$223+8>>2]=0|0;
   __THREW__ = 0;
   invoke_vii(65,($26|0),($27|0));
   $224 = __THREW__; __THREW__ = 0;
   $225 = $224&1;
   if ($225) {
    $235 = ___cxa_find_matching_catch_2()|0;
    $236 = tempRet0;
    $$16 = $235;$$1623 = $236;
   } else {
    __THREW__ = 0;
    $226 = (invoke_iii(803,($retStr|0),($26|0))|0);
    $227 = __THREW__; __THREW__ = 0;
    $228 = $227&1;
    if (!($228)) {
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($226|0));
     $229 = __THREW__; __THREW__ = 0;
     $230 = $229&1;
     if (!($230)) {
      __ZN9ExitGames6Common7JStringD2Ev($26);
      HEAP32[$27>>2] = (4640);
      __THREW__ = 0;
      invoke_viii(818,($27|0),(0|0),0);
      $231 = __THREW__; __THREW__ = 0;
      $232 = $231&1;
      if ($232) {
       $233 = ___cxa_find_matching_catch_3(0|0)|0;
       $234 = tempRet0;
       __ZN9ExitGames6Common4BaseD2Ev($27);
       ___clang_call_terminate($233);
       // unreachable;
      }
      __ZN9ExitGames6Common4BaseD2Ev($27);
      __ZN9ExitGames6Common7JStringD2Ev($retStr);
      STACKTOP = sp;return;
     }
    }
    $237 = ___cxa_find_matching_catch_2()|0;
    $238 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($26);
    $$16 = $237;$$1623 = $238;
   }
   HEAP32[$27>>2] = (4640);
   __THREW__ = 0;
   invoke_viii(818,($27|0),(0|0),0);
   $239 = __THREW__; __THREW__ = 0;
   $240 = $239&1;
   if ($240) {
    $241 = ___cxa_find_matching_catch_3(0|0)|0;
    $242 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($27);
    ___clang_call_terminate($241);
    // unreachable;
   }
   __ZN9ExitGames6Common4BaseD2Ev($27);
   $$17 = $$16;$$1724 = $$1623;
   __ZN9ExitGames6Common7JStringD2Ev($retStr);
   ___resumeException($$17|0);
   // unreachable;
   break;
  }
  case 99:  {
   $243 = ((($this)) + 29|0);
   $244 = HEAP8[$243>>0]|0;
   __THREW__ = 0;
   $245 = (invoke_iii(839,1,($244|0))|0);
   $246 = __THREW__; __THREW__ = 0;
   $247 = $246&1;
   if ($247) {
    label = 2;
   } else {
    $248 = HEAP32[$245>>2]|0;
    $249 = ((($248)) + 8|0);
    $250 = HEAP32[$249>>2]|0;
    __THREW__ = 0;
    invoke_vii($250|0,($28|0),($245|0));
    $251 = __THREW__; __THREW__ = 0;
    $252 = $251&1;
    if ($252) {
     label = 2;
    } else {
     __THREW__ = 0;
     (invoke_iii(760,($retStr|0),($28|0))|0);
     $253 = __THREW__; __THREW__ = 0;
     $254 = $253&1;
     if ($254) {
      $260 = ___cxa_find_matching_catch_2()|0;
      $261 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($28);
      $$17 = $260;$$1724 = $261;
      __ZN9ExitGames6Common7JStringD2Ev($retStr);
      ___resumeException($$17|0);
      // unreachable;
     }
     __ZN9ExitGames6Common7JStringD2Ev($28);
     $255 = HEAP8[$243>>0]|0;
     __THREW__ = 0;
     invoke_vii(840,($245|0),($255|0));
     $256 = __THREW__; __THREW__ = 0;
     $257 = $256&1;
     if ($257) {
      label = 2;
     } else {
      __THREW__ = 0;
      invoke_vii(786,($agg$result|0),($retStr|0));
      $258 = __THREW__; __THREW__ = 0;
      $259 = $258&1;
      if ($259) {
       label = 2;
      } else {
       __ZN9ExitGames6Common7JStringD2Ev($retStr);
       STACKTOP = sp;return;
      }
     }
    }
   }
   break;
  }
  default: {
   __THREW__ = 0;
   invoke_vii(739,($29|0),(145997|0));
   $262 = __THREW__; __THREW__ = 0;
   $263 = $262&1;
   if ($263) {
    label = 2;
   } else {
    __THREW__ = 0;
    $264 = (invoke_ii(764,($29|0))|0);
    $265 = __THREW__; __THREW__ = 0;
    $266 = $265&1;
    if (!($266)) {
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(5244|0),($264|0),1,523,(5304|0),($vararg_buffer|0));
     $267 = __THREW__; __THREW__ = 0;
     $268 = $267&1;
     if (!($268)) {
      __ZN9ExitGames6Common7JStringD2Ev($29);
      __THREW__ = 0;
      invoke_vii(787,($str$i|0),0);
      $269 = __THREW__; __THREW__ = 0;
      $270 = $269&1;
      if ($270) {
       label = 2;
       break L1;
      }
      __THREW__ = 0;
      (invoke_iii(759,($str$i|0),(5304|0))|0);
      $271 = __THREW__; __THREW__ = 0;
      $272 = $271&1;
      if (!($272)) {
       __THREW__ = 0;
       $273 = (invoke_iii(760,($retStr|0),($str$i|0))|0);
       $274 = __THREW__; __THREW__ = 0;
       $275 = $274&1;
       if (!($275)) {
        __ZN9ExitGames6Common7JStringD2Ev($str$i);
        __THREW__ = 0;
        invoke_vii(786,($agg$result|0),($273|0));
        $278 = __THREW__; __THREW__ = 0;
        $279 = $278&1;
        if ($279) {
         label = 2;
         break L1;
        }
        __ZN9ExitGames6Common7JStringD2Ev($retStr);
        STACKTOP = sp;return;
       }
      }
      $276 = ___cxa_find_matching_catch_2()|0;
      $277 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      $eh$lpad$body$index2Z2D = $277;$eh$lpad$body$indexZ2D = $276;
      break L1;
     }
    }
    $280 = ___cxa_find_matching_catch_2()|0;
    $281 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($29);
    $$17 = $280;$$1724 = $281;
    __ZN9ExitGames6Common7JStringD2Ev($retStr);
    ___resumeException($$17|0);
    // unreachable;
   }
  }
  }
 } while(0);
 if ((label|0) == 2) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $eh$lpad$body$index2Z2D = $34;$eh$lpad$body$indexZ2D = $33;
 }
 $$17 = $eh$lpad$body$indexZ2D;$$1724 = $eh$lpad$body$index2Z2D;
 __ZN9ExitGames6Common7JStringD2Ev($retStr);
 ___resumeException($$17|0);
 // unreachable;
}
function __ZN9ExitGames6Common7JStringaSEh($this,$aChar) {
 $this = $this|0;
 $aChar = $aChar|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP8[$0>>0] = $aChar;
 $1 = (__ZN9ExitGames6Common7JString6assignIhEERS1_RKT_jPKw($this,$0,4,5460)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JStringaSEs($this,$aNum) {
 $this = $this|0;
 $aNum = $aNum|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP16[$0>>1] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignIsEERS1_RKT_jPKw($this,$0,7,5444)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JStringaSEi($this,$aNum) {
 $this = $this|0;
 $aNum = $aNum|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$0>>2] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignIiEERS1_RKT_jPKw($this,$0,12,5432)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JStringaSEx($this,$0,$1) {
 $this = $this|0;
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $3 = $2;
 $4 = $3;
 HEAP32[$4>>2] = $0;
 $5 = (($3) + 4)|0;
 $6 = $5;
 HEAP32[$6>>2] = $1;
 $7 = (__ZN9ExitGames6Common7JString6assignIxEERS1_RKT_jPKw($this,$2,21,5412)|0);
 STACKTOP = sp;return ($7|0);
}
function __ZN9ExitGames6Common7JStringaSEf($this,$aNum) {
 $this = $this|0;
 $aNum = +$aNum;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAPF32[$0>>2] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignIfEERS1_RKT_jPKw($this,$0,30,5400)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JStringaSEd($this,$aNum) {
 $this = $this|0;
 $aNum = +$aNum;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAPF64[$0>>3] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignIdEERS1_RKT_jPKw($this,$0,60,5400)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JStringaSEb($this,$aBool) {
 $this = $this|0;
 $aBool = $aBool|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 $1 = $aBool&1;
 HEAP8[$0>>0] = $1;
 $2 = $aBool ? 5 : 6;
 $3 = $aBool ? 21672 : 21648;
 $4 = (__ZN9ExitGames6Common7JString6assignIbEERS1_RKT_jPKw($this,$0,$2,$3)|0);
 STACKTOP = sp;return ($4|0);
}
function __ZN9ExitGames6Common7JString6assignIbEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP8[$Rhs>>0]|0;
 $10 = $9&255;
 HEAP32[$vararg_buffer>>2] = $10;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $11 = ((($this)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $14 = ((($12)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($14);
 }
 $15 = (_wcslen($3)|0);
 $16 = ((($this)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($this)) + 8|0);
 HEAP32[$17>>2] = $15;
 $18 = (($15) + 1)|0;
 $19 = $18 << 2;
 $20 = (($19) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $18;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($18|0)==(0);
 if ($27) {
  HEAP32[$11>>2] = $22;
  (_wcscpy($22,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($22|0),0,($19|0))|0;
 HEAP32[$11>>2] = $22;
 (_wcscpy($22,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIdEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = +HEAPF64[$Rhs>>3];
 HEAPF64[$vararg_buffer>>3] = $9;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $10 = ((($this)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if (!($12)) {
  $13 = ((($11)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($13);
 }
 $14 = (_wcslen($3)|0);
 $15 = ((($this)) + 12|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($this)) + 8|0);
 HEAP32[$16>>2] = $14;
 $17 = (($14) + 1)|0;
 $18 = $17 << 2;
 $19 = (($18) + 8)|0;
 $20 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($19)|0);
 $21 = ((($20)) + 8|0);
 $22 = $20;
 $23 = $22;
 HEAP32[$23>>2] = $17;
 $24 = (($22) + 4)|0;
 $25 = $24;
 HEAP32[$25>>2] = 0;
 $26 = ($17|0)==(0);
 if ($26) {
  HEAP32[$10>>2] = $21;
  (_wcscpy($21,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($21|0),0,($18|0))|0;
 HEAP32[$10>>2] = $21;
 (_wcscpy($21,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIfEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = +HEAPF32[$Rhs>>2];
 $10 = $9;
 HEAPF64[$vararg_buffer>>3] = $10;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $11 = ((($this)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $14 = ((($12)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($14);
 }
 $15 = (_wcslen($3)|0);
 $16 = ((($this)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($this)) + 8|0);
 HEAP32[$17>>2] = $15;
 $18 = (($15) + 1)|0;
 $19 = $18 << 2;
 $20 = (($19) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $18;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($18|0)==(0);
 if ($27) {
  HEAP32[$11>>2] = $22;
  (_wcscpy($22,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($22|0),0,($19|0))|0;
 HEAP32[$11>>2] = $22;
 (_wcscpy($22,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIxEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = $Rhs;
 $10 = $9;
 $11 = HEAP32[$10>>2]|0;
 $12 = (($9) + 4)|0;
 $13 = $12;
 $14 = HEAP32[$13>>2]|0;
 $15 = $vararg_buffer;
 $16 = $15;
 HEAP32[$16>>2] = $11;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = $14;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $19 = ((($this)) + 4|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(0|0);
 if (!($21)) {
  $22 = ((($20)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($22);
 }
 $23 = (_wcslen($3)|0);
 $24 = ((($this)) + 12|0);
 HEAP32[$24>>2] = $23;
 $25 = ((($this)) + 8|0);
 HEAP32[$25>>2] = $23;
 $26 = (($23) + 1)|0;
 $27 = $26 << 2;
 $28 = (($27) + 8)|0;
 $29 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($28)|0);
 $30 = ((($29)) + 8|0);
 $31 = $29;
 $32 = $31;
 HEAP32[$32>>2] = $26;
 $33 = (($31) + 4)|0;
 $34 = $33;
 HEAP32[$34>>2] = 0;
 $35 = ($26|0)==(0);
 if ($35) {
  HEAP32[$19>>2] = $30;
  (_wcscpy($30,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($30|0),0,($27|0))|0;
 HEAP32[$19>>2] = $30;
 (_wcscpy($30,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIiEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP32[$Rhs>>2]|0;
 HEAP32[$vararg_buffer>>2] = $9;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $10 = ((($this)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if (!($12)) {
  $13 = ((($11)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($13);
 }
 $14 = (_wcslen($3)|0);
 $15 = ((($this)) + 12|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($this)) + 8|0);
 HEAP32[$16>>2] = $14;
 $17 = (($14) + 1)|0;
 $18 = $17 << 2;
 $19 = (($18) + 8)|0;
 $20 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($19)|0);
 $21 = ((($20)) + 8|0);
 $22 = $20;
 $23 = $22;
 HEAP32[$23>>2] = $17;
 $24 = (($22) + 4)|0;
 $25 = $24;
 HEAP32[$25>>2] = 0;
 $26 = ($17|0)==(0);
 if ($26) {
  HEAP32[$10>>2] = $21;
  (_wcscpy($21,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($21|0),0,($18|0))|0;
 HEAP32[$10>>2] = $21;
 (_wcscpy($21,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIsEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP16[$Rhs>>1]|0;
 $10 = $9 << 16 >> 16;
 HEAP32[$vararg_buffer>>2] = $10;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $11 = ((($this)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $14 = ((($12)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($14);
 }
 $15 = (_wcslen($3)|0);
 $16 = ((($this)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($this)) + 8|0);
 HEAP32[$17>>2] = $15;
 $18 = (($15) + 1)|0;
 $19 = $18 << 2;
 $20 = (($19) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $18;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($18|0)==(0);
 if ($27) {
  HEAP32[$11>>2] = $22;
  (_wcscpy($22,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($22|0),0,($19|0))|0;
 HEAP32[$11>>2] = $22;
 (_wcscpy($22,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JString6assignIhEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP8[$Rhs>>0]|0;
 $10 = $9&255;
 HEAP32[$vararg_buffer>>2] = $10;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $11 = ((($this)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $14 = ((($12)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($14);
 }
 $15 = (_wcslen($3)|0);
 $16 = ((($this)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($this)) + 8|0);
 HEAP32[$17>>2] = $15;
 $18 = (($15) + 1)|0;
 $19 = $18 << 2;
 $20 = (($19) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $18;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($18|0)==(0);
 if ($27) {
  HEAP32[$11>>2] = $22;
  (_wcscpy($22,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($22|0),0,($19|0))|0;
 HEAP32[$11>>2] = $22;
 (_wcscpy($22,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common6Object3setEPKvhhb($this,$data,$type,$customType,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $makeCopy = $makeCopy|0;
 var $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $size = sp;
 HEAP32[$size>>2] = 1;
 __ZN9ExitGames6Common6Object7cleanupEPKvj($this,0,0);
 __ZN9ExitGames6Common6Object17setWithoutCleanupEPKvhhjPKsb($this,$data,$type,$customType,0,$size,$makeCopy);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popStringHelperERNS0_7JStringE($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $$sum = 0, $$sum1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $lftr$wideiv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$0>>2]|0;
 $3 = (($2) + 1)|0;
 HEAP32[$0>>2] = $3;
 $4 = HEAP32[$1>>2]|0;
 $5 = (($4) + ($3)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = (($2) + 2)|0;
 HEAP32[$0>>2] = $7;
 $8 = (($4) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $6&255;
 $11 = ($10 << 8)&65535;
 $12 = $9&255;
 $13 = $11 | $12;
 $14 = $13 << 16 >> 16;
 $15 = (($14) + 1)|0;
 $16 = (($14) + 9)|0;
 $17 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($16)|0);
 $18 = ((($17)) + 8|0);
 $19 = $17;
 $20 = $19;
 HEAP32[$20>>2] = $15;
 $21 = (($19) + 4)|0;
 $22 = $21;
 HEAP32[$22>>2] = 0;
 $23 = ($15|0)==(0);
 if (!($23)) {
  _memset(($18|0),0,($15|0))|0;
 }
 $24 = ($13<<16>>16)>(0);
 if ($24) {
  $indvars$iv = 0;
 } else {
  $$sum = (($14) + 8)|0;
  $31 = (($17) + ($$sum)|0);
  HEAP8[$31>>0] = 0;
  (__ZN9ExitGames6Common7JStringaSEPKc($str,$18)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($17);
  return;
 }
 while(1) {
  $25 = HEAP32[$0>>2]|0;
  $26 = (($25) + 1)|0;
  HEAP32[$0>>2] = $26;
  $27 = HEAP32[$1>>2]|0;
  $28 = (($27) + ($26)|0);
  $29 = HEAP8[$28>>0]|0;
  $$sum1 = (($indvars$iv) + 8)|0;
  $30 = (($17) + ($$sum1)|0);
  HEAP8[$30>>0] = $29;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $lftr$wideiv = $indvars$iv$next&65535;
  $exitcond = ($lftr$wideiv<<16>>16)==($13<<16>>16);
  if ($exitcond) {
   break;
  } else {
   $indvars$iv = $indvars$iv$next;
  }
 }
 $$sum = (($14) + 8)|0;
 $31 = (($17) + ($$sum)|0);
 HEAP8[$31>>0] = 0;
 (__ZN9ExitGames6Common7JStringaSEPKc($str,$18)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($17);
 return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation18popHashTableHelperERNS0_9HashtableE($this,$hashtable) {
 $this = $this|0;
 $hashtable = $hashtable|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $i$03 = 0, $k = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $k = sp + 40|0;
 $v = sp;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$0>>2]|0;
 $3 = (($2) + 1)|0;
 HEAP32[$0>>2] = $3;
 $4 = HEAP32[$1>>2]|0;
 $5 = (($4) + ($3)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = (($2) + 2)|0;
 HEAP32[$0>>2] = $7;
 $8 = (($4) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $6&255;
 $11 = ($10 << 8)&65535;
 $12 = $9&255;
 $13 = $11 | $12;
 $14 = ($13<<16>>16)>(0);
 if (!($14)) {
  STACKTOP = sp;return;
 }
 $i$03 = 0;
 while(1) {
  __ZN9ExitGames6Common6ObjectC2Ev($k);
  __THREW__ = 0;
  invoke_vi(743,($v|0));
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   label = 8;
   break;
  }
  $17 = HEAP32[$0>>2]|0;
  $18 = (($17) + 1)|0;
  HEAP32[$0>>2] = $18;
  $19 = HEAP32[$1>>2]|0;
  $20 = (($19) + ($18)|0);
  $21 = HEAP8[$20>>0]|0;
  __THREW__ = 0;
  (invoke_iiii(813,($this|0),($21|0),($k|0))|0);
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   label = 9;
   break;
  }
  $24 = HEAP32[$0>>2]|0;
  $25 = (($24) + 1)|0;
  HEAP32[$0>>2] = $25;
  $26 = HEAP32[$1>>2]|0;
  $27 = (($26) + ($25)|0);
  $28 = HEAP8[$27>>0]|0;
  __THREW__ = 0;
  (invoke_iiii(813,($this|0),($28|0),($v|0))|0);
  $29 = __THREW__; __THREW__ = 0;
  $30 = $29&1;
  if ($30) {
   label = 9;
   break;
  }
  __THREW__ = 0;
  invoke_viii(841,($hashtable|0),($k|0),($v|0));
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if ($32) {
   label = 9;
   break;
  }
  __ZN9ExitGames6Common6ObjectD2Ev($v);
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  $33 = (($i$03) + 1)<<16>>16;
  $34 = ($33<<16>>16)<($13<<16>>16);
  if ($34) {
   $i$03 = $33;
  } else {
   label = 11;
   break;
  }
 }
 if ((label|0) == 8) {
  $35 = ___cxa_find_matching_catch_2()|0;
  $36 = tempRet0;
  $$0 = $36;$$01 = $35;
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  ___resumeException($$01|0);
  // unreachable;
 }
 else if ((label|0) == 9) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZN9ExitGames6Common6ObjectD2Ev($v);
  $$0 = $38;$$01 = $37;
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  ___resumeException($$01|0);
  // unreachable;
 }
 else if ((label|0) == 11) {
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation12popByteArrayERNS0_6ObjectE($this,$object) {
 $this = $this|0;
 $object = $object|0;
 var $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01 = 0, $size = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $size = sp;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$0>>2]|0;
 $3 = (($2) + 1)|0;
 HEAP32[$0>>2] = $3;
 $4 = HEAP32[$1>>2]|0;
 $5 = (($4) + ($3)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = (($2) + 2)|0;
 HEAP32[$0>>2] = $7;
 $8 = (($4) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = (($2) + 3)|0;
 HEAP32[$0>>2] = $10;
 $11 = (($4) + ($10)|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = (($2) + 4)|0;
 HEAP32[$0>>2] = $13;
 $14 = (($4) + ($13)|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $6&255;
 $17 = $16 << 24;
 $18 = $9&255;
 $19 = $18 << 16;
 $20 = $19 | $17;
 $21 = $12&255;
 $22 = $21 << 8;
 $23 = $20 | $22;
 $24 = $15&255;
 $25 = $23 | $24;
 HEAP32[$size>>2] = $25;
 $26 = (($25) + 8)|0;
 $27 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($26)|0);
 $28 = ((($27)) + 8|0);
 $29 = $27;
 $30 = $29;
 HEAP32[$30>>2] = $25;
 $31 = (($29) + 4)|0;
 $32 = $31;
 HEAP32[$32>>2] = 0;
 $33 = ($25|0)==(0);
 if ($33) {
  __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,$28,98,0,1,$size,0);
  STACKTOP = sp;return;
 }
 _memset(($28|0),0,($25|0))|0;
 $34 = ($25|0)>(0);
 if ($34) {
  $i$01 = 0;
 } else {
  __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,$28,98,0,1,$size,0);
  STACKTOP = sp;return;
 }
 while(1) {
  $35 = HEAP32[$0>>2]|0;
  $36 = (($35) + 1)|0;
  HEAP32[$0>>2] = $36;
  $37 = HEAP32[$1>>2]|0;
  $38 = (($37) + ($36)|0);
  $39 = HEAP8[$38>>0]|0;
  $$sum = (($i$01) + 8)|0;
  $40 = (($27) + ($$sum)|0);
  HEAP8[$40>>0] = $39;
  $41 = (($i$01) + 1)|0;
  $42 = HEAP32[$size>>2]|0;
  $43 = ($41|0)<($42|0);
  if ($43) {
   $i$01 = $41;
  } else {
   break;
  }
 }
 __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,$28,98,0,1,$size,0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation8popArrayERNS0_6ObjectE($this,$array) {
 $this = $this|0;
 $array = $array|0;
 var $$0 = 0, $$01 = 0, $$pre = 0, $$pre68 = 0, $$pre69 = 0, $$pre70 = 0, $$pre71 = 0, $$pre72 = 0, $$pre73 = 0, $$pre74 = 0, $$pre75 = 0, $$pre76 = 0, $$pre77 = 0, $$sum = 0, $$sum34 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0;
 var $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0;
 var $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0;
 var $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0;
 var $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0;
 var $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0;
 var $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0;
 var $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0;
 var $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $exitcond59 = 0, $exitcond60 = 0, $exitcond61 = 0, $exitcond62 = 0, $exitcond63 = 0, $exitcond64 = 0, $exitcond65 = 0, $exitcond66 = 0, $exitcond67 = 0, $i$038 = 0, $i$1057 = 0;
 var $i$139 = 0, $i$241 = 0, $i$343 = 0, $i$445 = 0, $i$547 = 0, $i$649 = 0, $i$751 = 0, $i$853 = 0, $i$955 = 0, $innerArray = 0, $j$0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $vararg_buffer = sp + 40|0;
 $innerArray = sp;
 $0 = sp + 48|0;
 $1 = ((($this)) + 4|0);
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$1>>2]|0;
 $4 = (($3) + 1)|0;
 HEAP32[$1>>2] = $4;
 $5 = HEAP32[$2>>2]|0;
 $6 = (($5) + ($4)|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = (($3) + 2)|0;
 HEAP32[$1>>2] = $8;
 $9 = (($5) + ($8)|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $7&255;
 $12 = ($11 << 8)&65535;
 $13 = $10&255;
 $14 = $12 | $13;
 $15 = (($3) + 3)|0;
 HEAP32[$1>>2] = $15;
 $16 = (($5) + ($15)|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = ($14<<16>>16)==(0);
 if ($18) {
  __ZN9ExitGames6Common6Object3setEPKvhhib($array,0,$17,0,0,0);
  STACKTOP = sp;return;
 }
 $19 = $17&255;
 do {
  switch ($19|0) {
  case 98:  {
   $20 = $14 << 16 >> 16;
   $21 = (($20) + 8)|0;
   $22 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($21)|0);
   $23 = ((($22)) + 8|0);
   $24 = $22;
   $25 = $24;
   HEAP32[$25>>2] = $20;
   $26 = (($24) + 4)|0;
   $27 = $26;
   HEAP32[$27>>2] = 0;
   _memset(($23|0),0,($20|0))|0;
   $28 = ($14<<16>>16)>(0);
   if ($28) {
    $i$038 = 0;
    while(1) {
     $29 = HEAP32[$1>>2]|0;
     $30 = (($29) + 1)|0;
     HEAP32[$1>>2] = $30;
     $31 = HEAP32[$2>>2]|0;
     $32 = (($31) + ($30)|0);
     $33 = HEAP8[$32>>0]|0;
     $$sum34 = (($i$038) + 8)|0;
     $34 = (($22) + ($$sum34)|0);
     HEAP8[$34>>0] = $33;
     $35 = (($i$038) + 1)|0;
     $exitcond = ($35|0)==($20|0);
     if ($exitcond) {
      break;
     } else {
      $i$038 = $35;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$23,$17,0,$20,0);
   STACKTOP = sp;return;
   break;
  }
  case 107:  {
   $36 = $14 << 16 >> 16;
   $37 = $36 << 1;
   $38 = (($37) + 8)|0;
   $39 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($38)|0);
   $40 = ((($39)) + 8|0);
   $41 = $39;
   $42 = $41;
   HEAP32[$42>>2] = $36;
   $43 = (($41) + 4)|0;
   $44 = $43;
   HEAP32[$44>>2] = 0;
   _memset(($40|0),0,($37|0))|0;
   $45 = ($14<<16>>16)>(0);
   if ($45) {
    $$pre76 = HEAP32[$1>>2]|0;
    $$pre77 = HEAP32[$2>>2]|0;
    $47 = $$pre76;$i$139 = 0;
    while(1) {
     $46 = (($47) + 1)|0;
     HEAP32[$1>>2] = $46;
     $48 = (($$pre77) + ($46)|0);
     $49 = HEAP8[$48>>0]|0;
     $50 = (($47) + 2)|0;
     HEAP32[$1>>2] = $50;
     $51 = (($$pre77) + ($50)|0);
     $52 = HEAP8[$51>>0]|0;
     $53 = $49&255;
     $54 = ($53 << 8)&65535;
     $55 = $52&255;
     $56 = $54 | $55;
     $57 = (($40) + ($i$139<<1)|0);
     HEAP16[$57>>1] = $56;
     $58 = (($i$139) + 1)|0;
     $exitcond59 = ($58|0)==($36|0);
     if ($exitcond59) {
      break;
     } else {
      $47 = $50;$i$139 = $58;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$40,$17,0,$36,0);
   STACKTOP = sp;return;
   break;
  }
  case 105:  {
   $59 = $14 << 16 >> 16;
   $60 = $59 << 2;
   $61 = (($60) + 8)|0;
   $62 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($61)|0);
   $63 = ((($62)) + 8|0);
   $64 = $62;
   $65 = $64;
   HEAP32[$65>>2] = $59;
   $66 = (($64) + 4)|0;
   $67 = $66;
   HEAP32[$67>>2] = 0;
   _memset(($63|0),0,($60|0))|0;
   $68 = ($14<<16>>16)>(0);
   if ($68) {
    $$pre75 = HEAP32[$2>>2]|0;
    $i$241 = 0;
    while(1) {
     $69 = HEAP32[$1>>2]|0;
     $70 = (($69) + 1)|0;
     HEAP32[$1>>2] = $70;
     $71 = (($$pre75) + ($70)|0);
     $72 = HEAP8[$71>>0]|0;
     $73 = (($69) + 2)|0;
     HEAP32[$1>>2] = $73;
     $74 = (($$pre75) + ($73)|0);
     $75 = HEAP8[$74>>0]|0;
     $76 = (($69) + 3)|0;
     HEAP32[$1>>2] = $76;
     $77 = (($$pre75) + ($76)|0);
     $78 = HEAP8[$77>>0]|0;
     $79 = (($69) + 4)|0;
     HEAP32[$1>>2] = $79;
     $80 = (($$pre75) + ($79)|0);
     $81 = HEAP8[$80>>0]|0;
     $82 = $72&255;
     $83 = $82 << 24;
     $84 = $75&255;
     $85 = $84 << 16;
     $86 = $85 | $83;
     $87 = $78&255;
     $88 = $87 << 8;
     $89 = $86 | $88;
     $90 = $81&255;
     $91 = $89 | $90;
     $92 = (($63) + ($i$241<<2)|0);
     HEAP32[$92>>2] = $91;
     $93 = (($i$241) + 1)|0;
     $exitcond60 = ($93|0)==($59|0);
     if ($exitcond60) {
      break;
     } else {
      $i$241 = $93;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$63,$17,0,$59,0);
   STACKTOP = sp;return;
   break;
  }
  case 108:  {
   $94 = $14 << 16 >> 16;
   $95 = $94 << 3;
   $96 = (($95) + 8)|0;
   $97 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($96)|0);
   $98 = ((($97)) + 8|0);
   $99 = $97;
   $100 = $99;
   HEAP32[$100>>2] = $94;
   $101 = (($99) + 4)|0;
   $102 = $101;
   HEAP32[$102>>2] = 0;
   _memset(($98|0),0,($95|0))|0;
   $103 = ($14<<16>>16)>(0);
   if ($103) {
    $$pre73 = HEAP32[$1>>2]|0;
    $$pre74 = HEAP32[$2>>2]|0;
    $105 = $$pre73;$i$343 = 0;
    while(1) {
     $104 = (($105) + 1)|0;
     HEAP32[$1>>2] = $104;
     $106 = (($$pre74) + ($104)|0);
     $107 = HEAP8[$106>>0]|0;
     $108 = (($105) + 2)|0;
     HEAP32[$1>>2] = $108;
     $109 = (($$pre74) + ($108)|0);
     $110 = HEAP8[$109>>0]|0;
     $111 = (($105) + 3)|0;
     HEAP32[$1>>2] = $111;
     $112 = (($$pre74) + ($111)|0);
     $113 = HEAP8[$112>>0]|0;
     $114 = (($105) + 4)|0;
     HEAP32[$1>>2] = $114;
     $115 = (($$pre74) + ($114)|0);
     $116 = HEAP8[$115>>0]|0;
     $117 = (($105) + 5)|0;
     HEAP32[$1>>2] = $117;
     $118 = (($$pre74) + ($117)|0);
     $119 = HEAP8[$118>>0]|0;
     $120 = (($105) + 6)|0;
     HEAP32[$1>>2] = $120;
     $121 = (($$pre74) + ($120)|0);
     $122 = HEAP8[$121>>0]|0;
     $123 = (($105) + 7)|0;
     HEAP32[$1>>2] = $123;
     $124 = (($$pre74) + ($123)|0);
     $125 = HEAP8[$124>>0]|0;
     $126 = (($105) + 8)|0;
     HEAP32[$1>>2] = $126;
     $127 = (($$pre74) + ($126)|0);
     $128 = HEAP8[$127>>0]|0;
     $129 = $107&255;
     $130 = (_bitshift64Shl(($129|0),0,56)|0);
     $131 = tempRet0;
     $132 = $110&255;
     $133 = (_bitshift64Shl(($132|0),0,48)|0);
     $134 = tempRet0;
     $135 = $133 | $130;
     $136 = $134 | $131;
     $137 = $113&255;
     $138 = (_bitshift64Shl(($137|0),0,40)|0);
     $139 = tempRet0;
     $140 = $135 | $138;
     $141 = $136 | $139;
     $142 = $116&255;
     $143 = $141 | $142;
     $144 = $119&255;
     $145 = (_bitshift64Shl(($144|0),0,24)|0);
     $146 = tempRet0;
     $147 = $140 | $145;
     $148 = $143 | $146;
     $149 = $122&255;
     $150 = (_bitshift64Shl(($149|0),0,16)|0);
     $151 = tempRet0;
     $152 = $125&255;
     $153 = (_bitshift64Shl(($152|0),0,8)|0);
     $154 = tempRet0;
     $155 = $147 | $150;
     $156 = $148 | $151;
     $157 = $128&255;
     $158 = $155 | $153;
     $159 = $156 | $154;
     $160 = $158 | $157;
     $161 = (($98) + ($i$343<<3)|0);
     $162 = $161;
     $163 = $162;
     HEAP32[$163>>2] = $160;
     $164 = (($162) + 4)|0;
     $165 = $164;
     HEAP32[$165>>2] = $159;
     $166 = (($i$343) + 1)|0;
     $exitcond61 = ($166|0)==($94|0);
     if ($exitcond61) {
      break;
     } else {
      $105 = $126;$i$343 = $166;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$98,$17,0,$94,0);
   STACKTOP = sp;return;
   break;
  }
  case 102:  {
   $167 = $14 << 16 >> 16;
   $168 = $167 << 2;
   $169 = (($168) + 8)|0;
   $170 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($169)|0);
   $171 = ((($170)) + 8|0);
   $172 = $170;
   $173 = $172;
   HEAP32[$173>>2] = $167;
   $174 = (($172) + 4)|0;
   $175 = $174;
   HEAP32[$175>>2] = 0;
   _memset(($171|0),0,($168|0))|0;
   $176 = ($14<<16>>16)>(0);
   if ($176) {
    $$pre71 = HEAP32[$1>>2]|0;
    $$pre72 = HEAP32[$2>>2]|0;
    $178 = $$pre71;$i$445 = 0;
    while(1) {
     $177 = (($178) + 1)|0;
     HEAP32[$1>>2] = $177;
     $179 = (($$pre72) + ($177)|0);
     $180 = HEAP8[$179>>0]|0;
     $181 = $180&255;
     $182 = $181 << 24;
     $183 = (($178) + 2)|0;
     HEAP32[$1>>2] = $183;
     $184 = (($$pre72) + ($183)|0);
     $185 = HEAP8[$184>>0]|0;
     $186 = $185&255;
     $187 = $186 << 16;
     $188 = (($178) + 3)|0;
     HEAP32[$1>>2] = $188;
     $189 = (($$pre72) + ($188)|0);
     $190 = HEAP8[$189>>0]|0;
     $191 = $190&255;
     $192 = $191 << 8;
     $193 = $187 | $182;
     $194 = $193 | $192;
     $195 = (($178) + 4)|0;
     HEAP32[$1>>2] = $195;
     $196 = (($$pre72) + ($195)|0);
     $197 = HEAP8[$196>>0]|0;
     $198 = $197&255;
     $199 = $194 | $198;
     $200 = (($171) + ($i$445<<2)|0);
     HEAP32[$200>>2] = $199;
     $201 = (($i$445) + 1)|0;
     $exitcond62 = ($201|0)==($167|0);
     if ($exitcond62) {
      break;
     } else {
      $178 = $195;$i$445 = $201;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$171,$17,0,$167,0);
   STACKTOP = sp;return;
   break;
  }
  case 100:  {
   $202 = $14 << 16 >> 16;
   $203 = $202 << 3;
   $204 = (($203) + 8)|0;
   $205 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($204)|0);
   $206 = ((($205)) + 8|0);
   $207 = $205;
   $208 = $207;
   HEAP32[$208>>2] = $202;
   $209 = (($207) + 4)|0;
   $210 = $209;
   HEAP32[$210>>2] = 0;
   _memset(($206|0),0,($203|0))|0;
   $211 = ($14<<16>>16)>(0);
   if ($211) {
    $$pre69 = HEAP32[$1>>2]|0;
    $$pre70 = HEAP32[$2>>2]|0;
    $213 = $$pre69;$i$547 = 0;
    while(1) {
     $212 = (($213) + 1)|0;
     HEAP32[$1>>2] = $212;
     $214 = (($$pre70) + ($212)|0);
     $215 = HEAP8[$214>>0]|0;
     $216 = $215&255;
     $217 = (_bitshift64Shl(($216|0),0,56)|0);
     $218 = tempRet0;
     $219 = (($213) + 2)|0;
     HEAP32[$1>>2] = $219;
     $220 = (($$pre70) + ($219)|0);
     $221 = HEAP8[$220>>0]|0;
     $222 = $221&255;
     $223 = (_bitshift64Shl(($222|0),0,48)|0);
     $224 = tempRet0;
     $225 = (($213) + 3)|0;
     HEAP32[$1>>2] = $225;
     $226 = (($$pre70) + ($225)|0);
     $227 = HEAP8[$226>>0]|0;
     $228 = $227&255;
     $229 = (_bitshift64Shl(($228|0),0,40)|0);
     $230 = tempRet0;
     $231 = $223 | $217;
     $232 = $224 | $218;
     $233 = (($213) + 4)|0;
     HEAP32[$1>>2] = $233;
     $234 = (($$pre70) + ($233)|0);
     $235 = HEAP8[$234>>0]|0;
     $236 = $235&255;
     $237 = $231 | $229;
     $238 = $232 | $230;
     $239 = (($213) + 5)|0;
     HEAP32[$1>>2] = $239;
     $240 = (($$pre70) + ($239)|0);
     $241 = HEAP8[$240>>0]|0;
     $242 = $241&255;
     $243 = (_bitshift64Shl(($242|0),0,24)|0);
     $244 = tempRet0;
     $245 = (($213) + 6)|0;
     HEAP32[$1>>2] = $245;
     $246 = (($$pre70) + ($245)|0);
     $247 = HEAP8[$246>>0]|0;
     $248 = $247&255;
     $249 = (_bitshift64Shl(($248|0),0,16)|0);
     $250 = tempRet0;
     $251 = (($213) + 7)|0;
     HEAP32[$1>>2] = $251;
     $252 = (($$pre70) + ($251)|0);
     $253 = HEAP8[$252>>0]|0;
     $254 = $253&255;
     $255 = (_bitshift64Shl(($254|0),0,8)|0);
     $256 = tempRet0;
     $257 = $238 | $236;
     $258 = (($213) + 8)|0;
     HEAP32[$1>>2] = $258;
     $259 = (($$pre70) + ($258)|0);
     $260 = HEAP8[$259>>0]|0;
     $261 = $260&255;
     $262 = $237 | $243;
     $263 = $257 | $244;
     $264 = $262 | $249;
     $265 = $263 | $250;
     $266 = $264 | $255;
     $267 = $265 | $256;
     $268 = $266 | $261;
     $269 = (($206) + ($i$547<<3)|0);
     $270 = $269;
     $271 = $270;
     HEAP32[$271>>2] = $268;
     $272 = (($270) + 4)|0;
     $273 = $272;
     HEAP32[$273>>2] = $267;
     $274 = (($i$547) + 1)|0;
     $exitcond63 = ($274|0)==($202|0);
     if ($exitcond63) {
      break;
     } else {
      $213 = $258;$i$547 = $274;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$206,$17,0,$202,0);
   STACKTOP = sp;return;
   break;
  }
  case 111:  {
   $275 = $14 << 16 >> 16;
   $276 = (($275) + 8)|0;
   $277 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($276)|0);
   $278 = ((($277)) + 8|0);
   $279 = $277;
   $280 = $279;
   HEAP32[$280>>2] = $275;
   $281 = (($279) + 4)|0;
   $282 = $281;
   HEAP32[$282>>2] = 0;
   _memset(($278|0),0,($275|0))|0;
   $283 = ($14<<16>>16)>(0);
   if ($283) {
    $$pre = HEAP32[$1>>2]|0;
    $$pre68 = HEAP32[$2>>2]|0;
    $285 = $$pre;$i$649 = 0;
    while(1) {
     $284 = (($285) + 1)|0;
     HEAP32[$1>>2] = $284;
     $286 = (($$pre68) + ($284)|0);
     $287 = HEAP8[$286>>0]|0;
     $288 = ($287<<24>>24)!=(0);
     $$sum = (($i$649) + 8)|0;
     $289 = (($277) + ($$sum)|0);
     $290 = $288&1;
     HEAP8[$289>>0] = $290;
     $291 = (($i$649) + 1)|0;
     $exitcond64 = ($291|0)==($275|0);
     if ($exitcond64) {
      break;
     } else {
      $285 = $284;$i$649 = $291;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$278,$17,0,$275,0);
   STACKTOP = sp;return;
   break;
  }
  case 115:  {
   $292 = $14 << 16 >> 16;
   $293 = $292 << 4;
   $294 = $293 | 8;
   $295 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($294)|0);
   $296 = ((($295)) + 8|0);
   $297 = $295;
   $298 = $297;
   HEAP32[$298>>2] = $292;
   $299 = (($297) + 4)|0;
   $300 = $299;
   HEAP32[$300>>2] = 0;
   $302 = 0;$303 = 0;
   while(1) {
    $301 = (($296) + ($302<<4)|0);
    __ZN9ExitGames6Common7JStringC2Ej($301,0);
    $304 = (_i64Add(($302|0),($303|0),1,0)|0);
    $305 = tempRet0;
    $306 = ($305>>>0)<(0);
    $307 = ($304>>>0)<($292>>>0);
    $308 = ($305|0)==(0);
    $309 = $308 & $307;
    $310 = $306 | $309;
    if ($310) {
     $302 = $304;$303 = $305;
    } else {
     break;
    }
   }
   $311 = ($14<<16>>16)>(0);
   if ($311) {
    $i$751 = 0;
    while(1) {
     $312 = (($296) + ($i$751<<4)|0);
     __ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popStringHelperERNS0_7JStringE($this,$312);
     $313 = (($i$751) + 1)|0;
     $exitcond65 = ($313|0)==($292|0);
     if ($exitcond65) {
      break;
     } else {
      $i$751 = $313;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$296,$17,0,$292,0);
   STACKTOP = sp;return;
   break;
  }
  case 104:  {
   $314 = $14 << 16 >> 16;
   $315 = ($314*44)|0;
   $316 = (($315) + 8)|0;
   $317 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($316)|0);
   $318 = ((($317)) + 8|0);
   $319 = $317;
   $320 = $319;
   HEAP32[$320>>2] = $314;
   $321 = (($319) + 4)|0;
   $322 = $321;
   HEAP32[$322>>2] = 0;
   $324 = 0;$325 = 0;
   while(1) {
    $323 = (($318) + (($324*44)|0)|0);
    __ZN9ExitGames6Common9HashtableC2Ev($323);
    $326 = (_i64Add(($324|0),($325|0),1,0)|0);
    $327 = tempRet0;
    $328 = ($327>>>0)<(0);
    $329 = ($326>>>0)<($314>>>0);
    $330 = ($327|0)==(0);
    $331 = $330 & $329;
    $332 = $328 | $331;
    if ($332) {
     $324 = $326;$325 = $327;
    } else {
     break;
    }
   }
   $333 = ($14<<16>>16)>(0);
   if ($333) {
    $i$853 = 0;
    while(1) {
     $334 = (($318) + (($i$853*44)|0)|0);
     __ZN9ExitGames6Common7Helpers26DeSerializerImplementation18popHashTableHelperERNS0_9HashtableE($this,$334);
     $335 = (($i$853) + 1)|0;
     $exitcond66 = ($335|0)==($314|0);
     if ($exitcond66) {
      break;
     } else {
      $i$853 = $335;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$318,$17,0,$314,0);
   STACKTOP = sp;return;
   break;
  }
  case 121:  {
   __ZN9ExitGames6Common6ObjectC2Ev($innerArray);
   $336 = $14 << 16 >> 16;
   $337 = $336 << 2;
   $338 = (($337) + 8)|0;
   __THREW__ = 0;
   $339 = (invoke_ii(775,($338|0))|0);
   $340 = __THREW__; __THREW__ = 0;
   $341 = $340&1;
   L80: do {
    if (!($341)) {
     $342 = ((($339)) + 8|0);
     $343 = $339;
     $344 = $343;
     HEAP32[$344>>2] = $336;
     $345 = (($343) + 4)|0;
     $346 = $345;
     HEAP32[$346>>2] = 0;
     _memset(($342|0),0,($337|0))|0;
     $347 = ($14<<16>>16)>(0);
     L82: do {
      if ($347) {
       $348 = (($336) + -1)|0;
       $i$955 = 0;
       while(1) {
        __THREW__ = 0;
        invoke_vii(842,($this|0),($innerArray|0));
        $349 = __THREW__; __THREW__ = 0;
        $350 = $349&1;
        if ($350) {
         break L80;
        }
        __THREW__ = 0;
        $351 = (invoke_ii(843,($innerArray|0))|0);
        $352 = __THREW__; __THREW__ = 0;
        $353 = $352&1;
        if ($353) {
         break L80;
        }
        $354 = (($342) + ($i$955<<2)|0);
        HEAP32[$354>>2] = $351;
        $355 = ($i$955|0)==($348|0);
        if ($355) {
         break L82;
        }
        __THREW__ = 0;
        invoke_vii(844,($innerArray|0),(0|0));
        $358 = __THREW__; __THREW__ = 0;
        $359 = $358&1;
        if ($359) {
         break L80;
        }
        __THREW__ = 0;
        invoke_vii(845,($innerArray|0),42);
        $360 = __THREW__; __THREW__ = 0;
        $361 = $360&1;
        if ($361) {
         break L80;
        }
        $362 = (($i$955) + 1)|0;
        $363 = ($362|0)<($336|0);
        if ($363) {
         $i$955 = $362;
        } else {
         break;
        }
       }
      }
     } while(0);
     __THREW__ = 0;
     $364 = (invoke_ii(846,($innerArray|0))|0);
     $365 = __THREW__; __THREW__ = 0;
     $366 = $365&1;
     if (!($366)) {
      $367 = (($364) + 1)|0;
      $368 = $367 << 1;
      $369 = (($368) + 8)|0;
      __THREW__ = 0;
      $370 = (invoke_ii(775,($369|0))|0);
      $371 = __THREW__; __THREW__ = 0;
      $372 = $371&1;
      if (!($372)) {
       $373 = ((($370)) + 8|0);
       $374 = $370;
       $375 = $374;
       HEAP32[$375>>2] = $367;
       $376 = (($374) + 4)|0;
       $377 = $376;
       HEAP32[$377>>2] = 0;
       $378 = ($367|0)==(0);
       if (!($378)) {
        _memset(($373|0),0,($368|0))|0;
       }
       HEAP16[$373>>1] = $14;
       $j$0 = 1;
       while(1) {
        __THREW__ = 0;
        $379 = (invoke_ii(846,($innerArray|0))|0);
        $380 = __THREW__; __THREW__ = 0;
        $381 = $380&1;
        if ($381) {
         break L80;
        }
        $382 = (($379) + 1)|0;
        $383 = ($j$0>>>0)<($382>>>0);
        if (!($383)) {
         break;
        }
        __THREW__ = 0;
        $384 = (invoke_ii(847,($innerArray|0))|0);
        $385 = __THREW__; __THREW__ = 0;
        $386 = $385&1;
        if ($386) {
         break L80;
        }
        $387 = (($j$0) + -1)|0;
        $388 = (($384) + ($387<<1)|0);
        $389 = HEAP16[$388>>1]|0;
        $390 = (($373) + ($j$0<<1)|0);
        HEAP16[$390>>1] = $389;
        $391 = (($j$0) + 1)|0;
        $j$0 = $391;
       }
       __THREW__ = 0;
       $392 = (invoke_ii(848,($innerArray|0))|0);
       $393 = __THREW__; __THREW__ = 0;
       $394 = $393&1;
       if (!($394)) {
        __THREW__ = 0;
        $395 = (invoke_ii(849,($innerArray|0))|0);
        $396 = __THREW__; __THREW__ = 0;
        $397 = $396&1;
        if (!($397)) {
         __THREW__ = 0;
         $398 = (invoke_ii(846,($innerArray|0))|0);
         $399 = __THREW__; __THREW__ = 0;
         $400 = $399&1;
         if (!($400)) {
          $401 = (($398) + 1)|0;
          __THREW__ = 0;
          invoke_viiiiiii(850,($array|0),($342|0),($392|0),($395|0),($401|0),($373|0),0);
          $402 = __THREW__; __THREW__ = 0;
          $403 = $402&1;
          if (!($403)) {
           __THREW__ = 0;
           invoke_vii(844,($innerArray|0),(0|0));
           $404 = __THREW__; __THREW__ = 0;
           $405 = $404&1;
           if (!($405)) {
            __THREW__ = 0;
            invoke_vii(845,($innerArray|0),42);
            $406 = __THREW__; __THREW__ = 0;
            $407 = $406&1;
            if (!($407)) {
             __THREW__ = 0;
             invoke_vi(750,($370|0));
             $408 = __THREW__; __THREW__ = 0;
             $409 = $408&1;
             if (!($409)) {
              __ZN9ExitGames6Common6ObjectD2Ev($innerArray);
              STACKTOP = sp;return;
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   } while(0);
   $356 = ___cxa_find_matching_catch_2()|0;
   $357 = tempRet0;
   __ZN9ExitGames6Common6ObjectD2Ev($innerArray);
   $$0 = $357;$$01 = $356;
   ___resumeException($$01|0);
   // unreachable;
   break;
  }
  case 122:  {
   $410 = $14 << 16 >> 16;
   $411 = $410 << 2;
   $412 = (($411) + 8)|0;
   $413 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($412)|0);
   $414 = ((($413)) + 8|0);
   $415 = $413;
   $416 = $415;
   HEAP32[$416>>2] = $410;
   $417 = (($415) + 4)|0;
   $418 = $417;
   HEAP32[$418>>2] = 0;
   _memset(($414|0),0,($411|0))|0;
   $419 = ($14<<16>>16)>(0);
   if ($419) {
    $i$1057 = 0;
    while(1) {
     $420 = (($414) + ($i$1057<<2)|0);
     $421 = HEAP32[$420>>2]|0;
     __ZN9ExitGames6Common7Helpers26DeSerializerImplementation14popObjectArrayERNS0_6ObjectE($this,$421);
     $422 = (($i$1057) + 1)|0;
     $exitcond67 = ($422|0)==($410|0);
     if ($exitcond67) {
      break;
     } else {
      $i$1057 = $422;
     }
    }
   }
   __ZN9ExitGames6Common6Object3setEPKvhhib($array,$414,$17,0,$410,0);
   STACKTOP = sp;return;
   break;
  }
  case 99:  {
   $423 = (($3) + 4)|0;
   HEAP32[$1>>2] = $423;
   $424 = (($5) + ($423)|0);
   $425 = HEAP8[$424>>0]|0;
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popCustomHelperERNS0_6ObjectEhsi($this,$array,$425,$14,1);
   STACKTOP = sp;return;
   break;
  }
  case 68:  {
   __ZN9ExitGames6Common7Helpers26DeSerializerImplementation22popArrayOfDictionariesERNS0_6ObjectEsi($this,$array,$14,1);
   STACKTOP = sp;return;
   break;
  }
  default: {
   __ZN9ExitGames6Common7JStringC2EPKc($0,146201);
   __THREW__ = 0;
   $426 = (invoke_ii(764,($0|0))|0);
   $427 = __THREW__; __THREW__ = 0;
   $428 = $427&1;
   if (!($428)) {
    __THREW__ = 0;
    HEAP32[$vararg_buffer>>2] = $19;
    invoke_viiiiiiii(765,(162472|0),1,(5696|0),($426|0),1,356,(5868|0),($vararg_buffer|0));
    $429 = __THREW__; __THREW__ = 0;
    $430 = $429&1;
    if (!($430)) {
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $431 = ___cxa_find_matching_catch_2()|0;
   $432 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $$0 = $432;$$01 = $431;
   ___resumeException($$01|0);
   // unreachable;
  }
  }
 } while(0);
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation14popObjectArrayERNS0_6ObjectE($this,$array) {
 $this = $this|0;
 $array = $array|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $boolValue = 0, $byteValue = 0, $data$0 = 0, $doubleValue = 0, $floatValue = 0, $i1$023 = 0, $innerDimensions$0 = 0, $innerSize$0 = 0, $innerType$0 = 0, $intValue = 0, $longValue = 0, $shortValue = 0, $size1 = 0, $temp = 0, $useTemp$0$off0 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer = sp + 56|0;
 $size1 = sp + 88|0;
 $temp = sp + 16|0;
 $byteValue = sp + 95|0;
 $shortValue = sp + 92|0;
 $intValue = sp + 84|0;
 $longValue = sp + 8|0;
 $floatValue = sp + 80|0;
 $doubleValue = sp;
 $boolValue = sp + 94|0;
 $0 = sp + 64|0;
 $1 = ((($this)) + 4|0);
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$1>>2]|0;
 $4 = (($3) + 1)|0;
 HEAP32[$1>>2] = $4;
 $5 = HEAP32[$2>>2]|0;
 $6 = (($5) + ($4)|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = (($3) + 2)|0;
 HEAP32[$1>>2] = $8;
 $9 = (($5) + ($8)|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = $7&255;
 $12 = ($11 << 8)&65535;
 $13 = $10&255;
 $14 = $12 | $13;
 $15 = ($14<<16>>16)==(0);
 if ($15) {
  __ZN9ExitGames6Common6Object3setEPKvhhb($array,0,122,0,0);
  STACKTOP = sp;return;
 }
 $16 = $14 << 16 >> 16;
 $17 = ($16*40)|0;
 $18 = (($17) + 8)|0;
 $19 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($18)|0);
 $20 = ((($19)) + 8|0);
 $21 = $19;
 $22 = $21;
 HEAP32[$22>>2] = $16;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = 0;
 $27 = 0;$28 = 0;
 while(1) {
  $26 = (($20) + (($27*40)|0)|0);
  __ZN9ExitGames6Common6ObjectC2Ev($26);
  $29 = (_i64Add(($27|0),($28|0),1,0)|0);
  $30 = tempRet0;
  $31 = ($30>>>0)<(0);
  $32 = ($29>>>0)<($16>>>0);
  $33 = ($30|0)==(0);
  $34 = $33 & $32;
  $35 = $31 | $34;
  if ($35) {
   $27 = $29;$28 = $30;
  } else {
   break;
  }
 }
 $25 = ($14<<16>>16)>(0);
 L8: do {
  if ($25) {
   $i1$023 = 0;
   L10: while(1) {
    $36 = HEAP32[$1>>2]|0;
    $37 = (($36) + 1)|0;
    HEAP32[$1>>2] = $37;
    $38 = HEAP32[$2>>2]|0;
    $39 = (($38) + ($37)|0);
    $40 = HEAP8[$39>>0]|0;
    HEAP32[$size1>>2] = 1;
    __ZN9ExitGames6Common6ObjectC2Ev($temp);
    $41 = $40&255;
    do {
     switch ($41|0) {
     case 98:  {
      $42 = HEAP32[$1>>2]|0;
      $43 = (($42) + 1)|0;
      HEAP32[$1>>2] = $43;
      $44 = HEAP32[$2>>2]|0;
      $45 = (($44) + ($43)|0);
      $46 = HEAP8[$45>>0]|0;
      HEAP8[$byteValue>>0] = $46;
      $data$0 = $byteValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 107:  {
      $49 = HEAP32[$1>>2]|0;
      $50 = (($49) + 1)|0;
      HEAP32[$1>>2] = $50;
      $51 = HEAP32[$2>>2]|0;
      $52 = (($51) + ($50)|0);
      $53 = HEAP8[$52>>0]|0;
      $54 = (($49) + 2)|0;
      HEAP32[$1>>2] = $54;
      $55 = (($51) + ($54)|0);
      $56 = HEAP8[$55>>0]|0;
      $57 = $53&255;
      $58 = ($57 << 8)&65535;
      $59 = $56&255;
      $60 = $58 | $59;
      HEAP16[$shortValue>>1] = $60;
      $data$0 = $shortValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 105:  {
      $61 = HEAP32[$1>>2]|0;
      $62 = (($61) + 1)|0;
      HEAP32[$1>>2] = $62;
      $63 = HEAP32[$2>>2]|0;
      $64 = (($63) + ($62)|0);
      $65 = HEAP8[$64>>0]|0;
      $66 = (($61) + 2)|0;
      HEAP32[$1>>2] = $66;
      $67 = (($63) + ($66)|0);
      $68 = HEAP8[$67>>0]|0;
      $69 = (($61) + 3)|0;
      HEAP32[$1>>2] = $69;
      $70 = (($63) + ($69)|0);
      $71 = HEAP8[$70>>0]|0;
      $72 = (($61) + 4)|0;
      HEAP32[$1>>2] = $72;
      $73 = (($63) + ($72)|0);
      $74 = HEAP8[$73>>0]|0;
      $75 = $65&255;
      $76 = $75 << 24;
      $77 = $68&255;
      $78 = $77 << 16;
      $79 = $78 | $76;
      $80 = $71&255;
      $81 = $80 << 8;
      $82 = $79 | $81;
      $83 = $74&255;
      $84 = $82 | $83;
      HEAP32[$intValue>>2] = $84;
      $data$0 = $intValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 108:  {
      $85 = HEAP32[$1>>2]|0;
      $86 = (($85) + 1)|0;
      HEAP32[$1>>2] = $86;
      $87 = HEAP32[$2>>2]|0;
      $88 = (($87) + ($86)|0);
      $89 = HEAP8[$88>>0]|0;
      $90 = (($85) + 2)|0;
      HEAP32[$1>>2] = $90;
      $91 = (($87) + ($90)|0);
      $92 = HEAP8[$91>>0]|0;
      $93 = (($85) + 3)|0;
      HEAP32[$1>>2] = $93;
      $94 = (($87) + ($93)|0);
      $95 = HEAP8[$94>>0]|0;
      $96 = (($85) + 4)|0;
      HEAP32[$1>>2] = $96;
      $97 = (($87) + ($96)|0);
      $98 = HEAP8[$97>>0]|0;
      $99 = (($85) + 5)|0;
      HEAP32[$1>>2] = $99;
      $100 = (($87) + ($99)|0);
      $101 = HEAP8[$100>>0]|0;
      $102 = (($85) + 6)|0;
      HEAP32[$1>>2] = $102;
      $103 = (($87) + ($102)|0);
      $104 = HEAP8[$103>>0]|0;
      $105 = (($85) + 7)|0;
      HEAP32[$1>>2] = $105;
      $106 = (($87) + ($105)|0);
      $107 = HEAP8[$106>>0]|0;
      $108 = (($85) + 8)|0;
      HEAP32[$1>>2] = $108;
      $109 = (($87) + ($108)|0);
      $110 = HEAP8[$109>>0]|0;
      $111 = $89&255;
      $112 = (_bitshift64Shl(($111|0),0,56)|0);
      $113 = tempRet0;
      $114 = $92&255;
      $115 = (_bitshift64Shl(($114|0),0,48)|0);
      $116 = tempRet0;
      $117 = $115 | $112;
      $118 = $116 | $113;
      $119 = $95&255;
      $120 = (_bitshift64Shl(($119|0),0,40)|0);
      $121 = tempRet0;
      $122 = $117 | $120;
      $123 = $118 | $121;
      $124 = $98&255;
      $125 = $123 | $124;
      $126 = $101&255;
      $127 = (_bitshift64Shl(($126|0),0,24)|0);
      $128 = tempRet0;
      $129 = $122 | $127;
      $130 = $125 | $128;
      $131 = $104&255;
      $132 = (_bitshift64Shl(($131|0),0,16)|0);
      $133 = tempRet0;
      $134 = $107&255;
      $135 = (_bitshift64Shl(($134|0),0,8)|0);
      $136 = tempRet0;
      $137 = $129 | $132;
      $138 = $130 | $133;
      $139 = $110&255;
      $140 = $137 | $135;
      $141 = $138 | $136;
      $142 = $140 | $139;
      $143 = $longValue;
      $144 = $143;
      HEAP32[$144>>2] = $142;
      $145 = (($143) + 4)|0;
      $146 = $145;
      HEAP32[$146>>2] = $141;
      $data$0 = $longValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 102:  {
      $147 = HEAP32[$1>>2]|0;
      $148 = (($147) + 1)|0;
      HEAP32[$1>>2] = $148;
      $149 = HEAP32[$2>>2]|0;
      $150 = (($149) + ($148)|0);
      $151 = HEAP8[$150>>0]|0;
      $152 = $151&255;
      $153 = $152 << 24;
      $154 = (($147) + 2)|0;
      HEAP32[$1>>2] = $154;
      $155 = (($149) + ($154)|0);
      $156 = HEAP8[$155>>0]|0;
      $157 = $156&255;
      $158 = $157 << 16;
      $159 = (($147) + 3)|0;
      HEAP32[$1>>2] = $159;
      $160 = (($149) + ($159)|0);
      $161 = HEAP8[$160>>0]|0;
      $162 = $161&255;
      $163 = $162 << 8;
      $164 = $158 | $153;
      $165 = $164 | $163;
      $166 = (($147) + 4)|0;
      HEAP32[$1>>2] = $166;
      $167 = (($149) + ($166)|0);
      $168 = HEAP8[$167>>0]|0;
      $169 = $168&255;
      $170 = $165 | $169;
      HEAP32[$floatValue>>2] = $170;
      $data$0 = $floatValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 100:  {
      $171 = HEAP32[$1>>2]|0;
      $172 = (($171) + 1)|0;
      HEAP32[$1>>2] = $172;
      $173 = HEAP32[$2>>2]|0;
      $174 = (($173) + ($172)|0);
      $175 = HEAP8[$174>>0]|0;
      $176 = $175&255;
      $177 = (_bitshift64Shl(($176|0),0,56)|0);
      $178 = tempRet0;
      $179 = (($171) + 2)|0;
      HEAP32[$1>>2] = $179;
      $180 = (($173) + ($179)|0);
      $181 = HEAP8[$180>>0]|0;
      $182 = $181&255;
      $183 = (_bitshift64Shl(($182|0),0,48)|0);
      $184 = tempRet0;
      $185 = (($171) + 3)|0;
      HEAP32[$1>>2] = $185;
      $186 = (($173) + ($185)|0);
      $187 = HEAP8[$186>>0]|0;
      $188 = $187&255;
      $189 = (_bitshift64Shl(($188|0),0,40)|0);
      $190 = tempRet0;
      $191 = $183 | $177;
      $192 = $184 | $178;
      $193 = (($171) + 4)|0;
      HEAP32[$1>>2] = $193;
      $194 = (($173) + ($193)|0);
      $195 = HEAP8[$194>>0]|0;
      $196 = $195&255;
      $197 = $191 | $189;
      $198 = $192 | $190;
      $199 = (($171) + 5)|0;
      HEAP32[$1>>2] = $199;
      $200 = (($173) + ($199)|0);
      $201 = HEAP8[$200>>0]|0;
      $202 = $201&255;
      $203 = (_bitshift64Shl(($202|0),0,24)|0);
      $204 = tempRet0;
      $205 = (($171) + 6)|0;
      HEAP32[$1>>2] = $205;
      $206 = (($173) + ($205)|0);
      $207 = HEAP8[$206>>0]|0;
      $208 = $207&255;
      $209 = (_bitshift64Shl(($208|0),0,16)|0);
      $210 = tempRet0;
      $211 = (($171) + 7)|0;
      HEAP32[$1>>2] = $211;
      $212 = (($173) + ($211)|0);
      $213 = HEAP8[$212>>0]|0;
      $214 = $213&255;
      $215 = (_bitshift64Shl(($214|0),0,8)|0);
      $216 = tempRet0;
      $217 = $198 | $196;
      $218 = (($171) + 8)|0;
      HEAP32[$1>>2] = $218;
      $219 = (($173) + ($218)|0);
      $220 = HEAP8[$219>>0]|0;
      $221 = $220&255;
      $222 = $197 | $203;
      $223 = $217 | $204;
      $224 = $222 | $209;
      $225 = $223 | $210;
      $226 = $224 | $215;
      $227 = $225 | $216;
      $228 = $226 | $221;
      $229 = $doubleValue;
      $230 = $229;
      HEAP32[$230>>2] = $228;
      $231 = (($229) + 4)|0;
      $232 = $231;
      HEAP32[$232>>2] = $227;
      $data$0 = $doubleValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 111:  {
      $233 = HEAP32[$1>>2]|0;
      $234 = (($233) + 1)|0;
      HEAP32[$1>>2] = $234;
      $235 = HEAP32[$2>>2]|0;
      $236 = (($235) + ($234)|0);
      $237 = HEAP8[$236>>0]|0;
      $238 = ($237<<24>>24)!=(0);
      $239 = $238&1;
      HEAP8[$boolValue>>0] = $239;
      $data$0 = $boolValue;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      break;
     }
     case 115:  {
      __THREW__ = 0;
      $240 = (invoke_ii(775,24)|0);
      $241 = __THREW__; __THREW__ = 0;
      $242 = $241&1;
      if ($242) {
       label = 9;
       break L10;
      }
      $243 = ((($240)) + 8|0);
      $244 = $240;
      $245 = $244;
      HEAP32[$245>>2] = 1;
      $246 = (($244) + 4)|0;
      $247 = $246;
      HEAP32[$247>>2] = 0;
      __THREW__ = 0;
      invoke_vii(787,($243|0),0);
      $248 = __THREW__; __THREW__ = 0;
      $249 = $248&1;
      if ($249) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      invoke_vii(815,($this|0),($243|0));
      $250 = __THREW__; __THREW__ = 0;
      $251 = $250&1;
      if ($251) {
       label = 9;
       break L10;
      } else {
       $data$0 = $243;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      }
      break;
     }
     case 104:  {
      __THREW__ = 0;
      $252 = (invoke_ii(775,52)|0);
      $253 = __THREW__; __THREW__ = 0;
      $254 = $253&1;
      if ($254) {
       label = 9;
       break L10;
      }
      $255 = ((($252)) + 8|0);
      $256 = $252;
      $257 = $256;
      HEAP32[$257>>2] = 1;
      $258 = (($256) + 4)|0;
      $259 = $258;
      HEAP32[$259>>2] = 0;
      __THREW__ = 0;
      invoke_vi(838,($255|0));
      $260 = __THREW__; __THREW__ = 0;
      $261 = $260&1;
      if ($261) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      invoke_vii(817,($this|0),($255|0));
      $262 = __THREW__; __THREW__ = 0;
      $263 = $262&1;
      if ($263) {
       label = 9;
       break L10;
      } else {
       $data$0 = $255;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
      }
      break;
     }
     case 121:  {
      __THREW__ = 0;
      invoke_vii(842,($this|0),($temp|0));
      $264 = __THREW__; __THREW__ = 0;
      $265 = $264&1;
      if ($265) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $266 = (invoke_ii(843,($temp|0))|0);
      $267 = __THREW__; __THREW__ = 0;
      $268 = $267&1;
      if ($268) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $269 = (invoke_ii(848,($temp|0))|0);
      $270 = __THREW__; __THREW__ = 0;
      $271 = $270&1;
      if ($271) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $272 = (invoke_ii(846,($temp|0))|0);
      $273 = __THREW__; __THREW__ = 0;
      $274 = $273&1;
      if ($274) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $275 = (invoke_ii(847,($temp|0))|0);
      $276 = __THREW__; __THREW__ = 0;
      $277 = $276&1;
      if ($277) {
       label = 9;
       break L10;
      }
      $data$0 = $266;$innerDimensions$0 = $272;$innerSize$0 = $275;$innerType$0 = $269;$useTemp$0$off0 = 1;
      break;
     }
     case 120:  {
      __THREW__ = 0;
      invoke_vii(851,($this|0),($temp|0));
      $278 = __THREW__; __THREW__ = 0;
      $279 = $278&1;
      if ($279) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $280 = (invoke_ii(843,($temp|0))|0);
      $281 = __THREW__; __THREW__ = 0;
      $282 = $281&1;
      if ($282) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $283 = (invoke_ii(848,($temp|0))|0);
      $284 = __THREW__; __THREW__ = 0;
      $285 = $284&1;
      if ($285) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $286 = (invoke_ii(846,($temp|0))|0);
      $287 = __THREW__; __THREW__ = 0;
      $288 = $287&1;
      if ($288) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $289 = (invoke_ii(847,($temp|0))|0);
      $290 = __THREW__; __THREW__ = 0;
      $291 = $290&1;
      if ($291) {
       label = 9;
       break L10;
      }
      $data$0 = $280;$innerDimensions$0 = $286;$innerSize$0 = $289;$innerType$0 = $283;$useTemp$0$off0 = 1;
      break;
     }
     case 122:  {
      __THREW__ = 0;
      invoke_vii(852,($this|0),($temp|0));
      $292 = __THREW__; __THREW__ = 0;
      $293 = $292&1;
      if ($293) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $294 = (invoke_ii(843,($temp|0))|0);
      $295 = __THREW__; __THREW__ = 0;
      $296 = $295&1;
      if ($296) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $297 = (invoke_ii(848,($temp|0))|0);
      $298 = __THREW__; __THREW__ = 0;
      $299 = $298&1;
      if ($299) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $300 = (invoke_ii(846,($temp|0))|0);
      $301 = __THREW__; __THREW__ = 0;
      $302 = $301&1;
      if ($302) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $303 = (invoke_ii(847,($temp|0))|0);
      $304 = __THREW__; __THREW__ = 0;
      $305 = $304&1;
      if ($305) {
       label = 9;
       break L10;
      }
      $data$0 = $294;$innerDimensions$0 = $300;$innerSize$0 = $303;$innerType$0 = $297;$useTemp$0$off0 = 1;
      break;
     }
     case 99:  {
      $306 = HEAP32[$1>>2]|0;
      $307 = (($306) + 1)|0;
      HEAP32[$1>>2] = $307;
      $308 = HEAP32[$2>>2]|0;
      $309 = (($308) + ($307)|0);
      $310 = HEAP8[$309>>0]|0;
      __THREW__ = 0;
      invoke_viiiii(853,($this|0),($temp|0),($310|0),1,0);
      $311 = __THREW__; __THREW__ = 0;
      $312 = $311&1;
      if ($312) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $313 = (invoke_ii(843,($temp|0))|0);
      $314 = __THREW__; __THREW__ = 0;
      $315 = $314&1;
      if ($315) {
       label = 9;
       break L10;
      } else {
       $data$0 = $313;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $310;$useTemp$0$off0 = 1;
      }
      break;
     }
     case 68:  {
      __THREW__ = 0;
      invoke_viiii(854,($this|0),($temp|0),1,0);
      $316 = __THREW__; __THREW__ = 0;
      $317 = $316&1;
      if ($317) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $318 = (invoke_ii(843,($temp|0))|0);
      $319 = __THREW__; __THREW__ = 0;
      $320 = $319&1;
      if ($320) {
       label = 9;
       break L10;
      } else {
       $data$0 = $318;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 1;
      }
      break;
     }
     default: {
      __THREW__ = 0;
      invoke_vii(739,($0|0),(146186|0));
      $321 = __THREW__; __THREW__ = 0;
      $322 = $321&1;
      if ($322) {
       label = 9;
       break L10;
      }
      __THREW__ = 0;
      $323 = (invoke_ii(764,($0|0))|0);
      $324 = __THREW__; __THREW__ = 0;
      $325 = $324&1;
      if ($325) {
       label = 48;
       break L10;
      }
      __THREW__ = 0;
      HEAP32[$vararg_buffer>>2] = $41;
      invoke_viiiiiiii(765,(162472|0),1,(5696|0),($323|0),1,456,(5868|0),($vararg_buffer|0));
      $326 = __THREW__; __THREW__ = 0;
      $327 = $326&1;
      if ($327) {
       label = 48;
       break L10;
      }
      __ZN9ExitGames6Common7JStringD2Ev($0);
      $data$0 = 0;$innerDimensions$0 = 0;$innerSize$0 = $size1;$innerType$0 = $40;$useTemp$0$off0 = 0;
     }
     }
    } while(0);
    $330 = ($40<<24>>24)==(99);
    $331 = $330 ? $innerType$0 : 0;
    $332 = $330 ? 99 : $innerType$0;
    $333 = (($20) + (($i1$023*40)|0)|0);
    __THREW__ = 0;
    invoke_viiiiiii(855,($333|0),($data$0|0),($332|0),($331|0),($innerDimensions$0|0),($innerSize$0|0),($useTemp$0$off0|0));
    $334 = __THREW__; __THREW__ = 0;
    $335 = $334&1;
    if ($335) {
     label = 9;
     break;
    }
    __ZN9ExitGames6Common6ObjectD2Ev($temp);
    $336 = (($i1$023) + 1)|0;
    $337 = ($336|0)<($16|0);
    if ($337) {
     $i1$023 = $336;
    } else {
     break L8;
    }
   }
   if ((label|0) == 9) {
    $47 = ___cxa_find_matching_catch_2()|0;
    $48 = tempRet0;
    $$0 = $48;$$01 = $47;
    __ZN9ExitGames6Common6ObjectD2Ev($temp);
    ___resumeException($$01|0);
    // unreachable;
   }
   else if ((label|0) == 48) {
    $328 = ___cxa_find_matching_catch_2()|0;
    $329 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    $$0 = $329;$$01 = $328;
    __ZN9ExitGames6Common6ObjectD2Ev($temp);
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN9ExitGames6Common6Object3setEPKvhhib($array,$20,122,0,$16,0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($this,$data,$type,$customType,$dimensions,$sizes,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $dimensions = $dimensions|0;
 $sizes = $sizes|0;
 $makeCopy = $makeCopy|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common6Object7cleanupEPKvj($this,0,0);
 __ZN9ExitGames6Common6Object17setWithoutCleanupEPKvhhjPKsb($this,$data,$type,$customType,$dimensions,$sizes,$makeCopy);
 return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popCustomHelperERNS0_6ObjectEhsi($this,$object,$customType,$arraySize,$dimensions) {
 $this = $this|0;
 $object = $object|0;
 $customType = $customType|0;
 $arraySize = $arraySize|0;
 $dimensions = $dimensions|0;
 var $$pre = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $j$02 = 0, $lftr$wideiv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP16[$0>>1] = $arraySize;
 $1 = (__ZN9ExitGames6Common14CustomTypeBase11allocObjectEsh($arraySize,$customType)|0);
 $2 = ($arraySize<<16>>16)>(0);
 if (!($2)) {
  __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,$1,99,$customType,$dimensions,$0,0);
  STACKTOP = sp;return;
 }
 $3 = ((($this)) + 4|0);
 $4 = ((($this)) + 8|0);
 $35 = 0;$41 = $arraySize;$j$02 = 0;
 while(1) {
  $5 = HEAP32[$3>>2]|0;
  $6 = (($5) + 1)|0;
  HEAP32[$3>>2] = $6;
  $7 = HEAP32[$4>>2]|0;
  $8 = (($7) + ($6)|0);
  $9 = HEAP8[$8>>0]|0;
  $10 = (($5) + 2)|0;
  HEAP32[$3>>2] = $10;
  $11 = (($7) + ($10)|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = $9&255;
  $14 = ($13 << 8)&65535;
  $15 = $12&255;
  $16 = $14 | $15;
  $17 = $16 << 16 >> 16;
  $18 = (($17) + 8)|0;
  $19 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($18)|0);
  $20 = ((($19)) + 8|0);
  $21 = $19;
  $22 = $21;
  HEAP32[$22>>2] = $17;
  $23 = (($21) + 4)|0;
  $24 = $23;
  HEAP32[$24>>2] = 0;
  $25 = ($16<<16>>16)==(0);
  if ($25) {
   $40 = $41;
  } else {
   _memset(($20|0),0,($17|0))|0;
   $26 = ($16<<16>>16)>(0);
   if ($26) {
    $indvars$iv = 0;
    while(1) {
     $27 = HEAP32[$3>>2]|0;
     $28 = (($27) + 1)|0;
     HEAP32[$3>>2] = $28;
     $29 = HEAP32[$4>>2]|0;
     $30 = (($29) + ($28)|0);
     $31 = HEAP8[$30>>0]|0;
     $$sum = (($indvars$iv) + 8)|0;
     $32 = (($19) + ($$sum)|0);
     HEAP8[$32>>0] = $31;
     $indvars$iv$next = (($indvars$iv) + 1)|0;
     $lftr$wideiv = $indvars$iv$next&65535;
     $exitcond = ($lftr$wideiv<<16>>16)==($16<<16>>16);
     if ($exitcond) {
      break;
     } else {
      $indvars$iv = $indvars$iv$next;
     }
    }
    $$pre = HEAP16[$0>>1]|0;
    $40 = $$pre;
   } else {
    $40 = $41;
   }
  }
  $33 = (__ZN9ExitGames6Common14CustomTypeBase9getSizeofEh($customType)|0);
  $34 = Math_imul($33, $35)|0;
  $36 = (($1) + ($34)|0);
  __ZN9ExitGames6Common14CustomTypeBase11deserializeEPhsPvh($20,$16,$36,$customType);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($19);
  $37 = (($j$02) + 1)<<16>>16;
  $38 = $37 << 16 >> 16;
  $39 = ($37<<16>>16)<($40<<16>>16);
  if ($39) {
   $35 = $38;$41 = $40;$j$02 = $37;
  } else {
   break;
  }
 }
 __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($object,$1,99,$customType,$dimensions,$0,0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation22popArrayOfDictionariesERNS0_6ObjectEsi($this,$dict,$arraySize,$dimensions) {
 $this = $this|0;
 $dict = $dict|0;
 $arraySize = $arraySize|0;
 $dimensions = $dimensions|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arrSize = 0, $exitcond = 0, $indvars$iv = 0;
 var $indvars$iv$next = 0, $lftr$wideiv = 0, $pDimensionsOfNestedValues = 0, $pKeyTypes = 0, $pValTypes = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $pKeyTypes = sp + 12|0;
 $pValTypes = sp + 8|0;
 $pDimensionsOfNestedValues = sp + 4|0;
 $arrSize = sp;
 HEAP32[$pKeyTypes>>2] = 0;
 HEAP32[$pValTypes>>2] = 0;
 HEAP32[$pDimensionsOfNestedValues>>2] = 0;
 __ZN9ExitGames6Common7Helpers26DeSerializerImplementation19readDictionaryTypesEPPhS4_PPj($this,$pKeyTypes,$pValTypes,$pDimensionsOfNestedValues);
 $0 = $arraySize << 16 >> 16;
 $1 = ($0*52)|0;
 $2 = (($1) + 8)|0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($2)|0);
 $4 = ((($3)) + 8|0);
 $5 = $3;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = 0;
 $9 = ($arraySize<<16>>16)==(0);
 if ($9) {
  label = 3;
 } else {
  $15 = 0;$16 = 0;
  while(1) {
   $14 = (($4) + (($15*52)|0)|0);
   __ZN9ExitGames6Common14DictionaryBaseC2Ev($14);
   $17 = (_i64Add(($15|0),($16|0),1,0)|0);
   $18 = tempRet0;
   $19 = ($18>>>0)<(0);
   $20 = ($17>>>0)<($0>>>0);
   $21 = ($18|0)==(0);
   $22 = $21 & $20;
   $23 = $19 | $22;
   if ($23) {
    $15 = $17;$16 = $18;
   } else {
    break;
   }
  }
  $10 = ($arraySize<<16>>16)>(0);
  if ($10) {
   $11 = HEAP32[$pKeyTypes>>2]|0;
   $12 = HEAP32[$pValTypes>>2]|0;
   $13 = HEAP32[$pDimensionsOfNestedValues>>2]|0;
   $indvars$iv = 0;
   while(1) {
    $24 = (__ZN9ExitGames6Common7Helpers26DeSerializerImplementation19popDictionaryHelperEPhS3_Pj($this,$11,$12,$13)|0);
    $25 = (($4) + (($indvars$iv*52)|0)|0);
    (__ZN9ExitGames6Common14DictionaryBaseaSERKS1_($25,$24)|0);
    $26 = ($24|0)==(0|0);
    if (!($26)) {
     $27 = ((($24)) + -8|0);
     $28 = $27;
     $29 = $28;
     $30 = HEAP32[$29>>2]|0;
     $31 = (($28) + 4)|0;
     $32 = $31;
     $33 = HEAP32[$32>>2]|0;
     $34 = ($30|0)==(0);
     $35 = ($33|0)==(0);
     $36 = $34 & $35;
     if (!($36)) {
      $37 = $30;$38 = $33;
      while(1) {
       $39 = (_i64Add(($37|0),($38|0),-1,-1)|0);
       $40 = tempRet0;
       $41 = (($24) + (($39*52)|0)|0);
       $42 = HEAP32[$41>>2]|0;
       $43 = HEAP32[$42>>2]|0;
       FUNCTION_TABLE_vi[$43 & 2047]($41);
       $44 = ($39|0)==(0);
       $45 = ($40|0)==(0);
       $46 = $44 & $45;
       if ($46) {
        break;
       } else {
        $37 = $39;$38 = $40;
       }
      }
     }
     __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($27);
    }
    $indvars$iv$next = (($indvars$iv) + 1)|0;
    $lftr$wideiv = $indvars$iv$next&65535;
    $exitcond = ($lftr$wideiv<<16>>16)==($arraySize<<16>>16);
    if ($exitcond) {
     $48 = $11;
     break;
    } else {
     $indvars$iv = $indvars$iv$next;
    }
   }
  } else {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  $$pre = HEAP32[$pKeyTypes>>2]|0;
  $48 = $$pre;
 }
 HEAP32[$arrSize>>2] = $0;
 __ZN9ExitGames6Common6Object3setEPKvhhjPKsb($dict,$4,68,0,$dimensions,$arrSize,0);
 $47 = ($48|0)==(0|0);
 if (!($47)) {
  $49 = ((($48)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($49);
 }
 $50 = HEAP32[$pValTypes>>2]|0;
 $51 = ($50|0)==(0|0);
 if (!($51)) {
  $52 = ((($50)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($52);
 }
 $53 = HEAP32[$pDimensionsOfNestedValues>>2]|0;
 $54 = ($53|0)==(0|0);
 if ($54) {
  STACKTOP = sp;return;
 }
 $55 = ((($53)) + -8|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($55);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation19readDictionaryTypesEPPhS4_PPj($this,$ppKeyTypes,$ppValTypes,$ppDimensionsOfNestedValues) {
 $this = $this|0;
 $ppKeyTypes = $ppKeyTypes|0;
 $ppValTypes = $ppValTypes|0;
 $ppDimensionsOfNestedValues = $ppDimensionsOfNestedValues|0;
 var $$sum = 0, $$sum1$i = 0, $$sum1$i10 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $dimensions$0$lcssa = 0, $dimensions$018 = 0, $nestingDepthIncludingTopLevel$0 = 0, $or$cond$i = 0, $or$cond$i11 = 0, $or$cond$i4 = 0, $or$cond2$i = 0, $or$cond3$i = 0, $or$cond3$i5 = 0, $scevgep$i = 0, $scevgep$i2 = 0, $scevgep$i7 = 0, $valueType$0 = 0, $valueType$0$lcssa = 0;
 var $valueType$017 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $nestingDepthIncludingTopLevel$0 = 1;
 while(1) {
  $2 = ($nestingDepthIncludingTopLevel$0>>>0)<(2);
  if (!($2)) {
   $3 = (($nestingDepthIncludingTopLevel$0) + -2)|0;
   $4 = HEAP32[$ppValTypes>>2]|0;
   $5 = (($4) + ($3)|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = ($6<<24>>24)==(68);
   if (!($7)) {
    break;
   }
  }
  $8 = HEAP32[$0>>2]|0;
  $9 = (($8) + 1)|0;
  HEAP32[$0>>2] = $9;
  $10 = HEAP32[$1>>2]|0;
  $11 = (($10) + ($9)|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = (($8) + 2)|0;
  HEAP32[$0>>2] = $13;
  $14 = (($10) + ($13)|0);
  $valueType$017 = HEAP8[$14>>0]|0;
  $15 = ($valueType$017<<24>>24)==(121);
  if ($15) {
   $18 = $13;$dimensions$018 = 0;
   while(1) {
    $16 = (($dimensions$018) + 1)|0;
    $17 = (($18) + 1)|0;
    HEAP32[$0>>2] = $17;
    $19 = (($10) + ($17)|0);
    $valueType$0 = HEAP8[$19>>0]|0;
    $20 = ($valueType$0<<24>>24)==(121);
    if ($20) {
     $18 = $17;$dimensions$018 = $16;
    } else {
     $dimensions$0$lcssa = $16;$valueType$0$lcssa = $valueType$0;
     break;
    }
   }
  } else {
   $dimensions$0$lcssa = 0;$valueType$0$lcssa = $valueType$017;
  }
  $21 = ($12<<24>>24)==(0);
  $22 = $21 ? 122 : $12;
  $23 = (($nestingDepthIncludingTopLevel$0) + -1)|0;
  $24 = HEAP32[$ppKeyTypes>>2]|0;
  $25 = ($24|0)!=(0|0);
  if ($25) {
   $26 = ((($24)) + -8|0);
   $27 = $26;
   $28 = $27;
   $29 = HEAP32[$28>>2]|0;
   $30 = (($27) + 4)|0;
   $31 = $30;
   $32 = HEAP32[$31>>2]|0;
   $41 = $29;$43 = $32;
  } else {
   $41 = 0;$43 = 0;
  }
  $33 = (($nestingDepthIncludingTopLevel$0) + 8)|0;
  $34 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($33)|0);
  $35 = ((($34)) + 8|0);
  $36 = $34;
  $37 = $36;
  HEAP32[$37>>2] = $nestingDepthIncludingTopLevel$0;
  $38 = (($36) + 4)|0;
  $39 = $38;
  HEAP32[$39>>2] = 0;
  $40 = ($41|0)!=(0);
  $42 = ($43|0)!=(0);
  $44 = $40 | $42;
  $45 = ($nestingDepthIncludingTopLevel$0|0)!=(0);
  $or$cond3$i5 = $45 & $44;
  if ($or$cond3$i5) {
   $57 = 0;$61 = 0;
   while(1) {
    $$sum1$i10 = (($57) + 8)|0;
    $58 = (($34) + ($$sum1$i10)|0);
    $59 = (($24) + ($57)|0);
    $60 = HEAP8[$59>>0]|0;
    HEAP8[$58>>0] = $60;
    $62 = (_i64Add(($57|0),($61|0),1,0)|0);
    $63 = tempRet0;
    $64 = ($63>>>0)<($43>>>0);
    $65 = ($62>>>0)<($41>>>0);
    $66 = ($63|0)==($43|0);
    $67 = $66 & $65;
    $68 = $64 | $67;
    $69 = ($63>>>0)<(0);
    $70 = ($62>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
    $71 = ($63|0)==(0);
    $72 = $71 & $70;
    $73 = $69 | $72;
    $or$cond$i11 = $68 & $73;
    if ($or$cond$i11) {
     $57 = $62;$61 = $63;
    } else {
     break;
    }
   }
  }
  $46 = ($43>>>0)<(0);
  $47 = ($41>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
  $48 = ($43|0)==(0);
  $49 = $48 & $47;
  $50 = $46 | $49;
  if ($50) {
   $51 = (($41) + 8)|0;
   $scevgep$i7 = (($34) + ($51)|0);
   $52 = (_i64Add(($nestingDepthIncludingTopLevel$0|0),0,-1,0)|0);
   $53 = tempRet0;
   $54 = (_i64Subtract(($52|0),($53|0),($41|0),($43|0))|0);
   $55 = tempRet0;
   $56 = (($54) + 1)|0;
   _memset(($scevgep$i7|0),0,($56|0))|0;
  }
  if ($25) {
   $74 = ((($24)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($74);
  }
  HEAP32[$ppKeyTypes>>2] = $35;
  $$sum = (($nestingDepthIncludingTopLevel$0) + 7)|0;
  $75 = (($34) + ($$sum)|0);
  HEAP8[$75>>0] = $22;
  $76 = ($valueType$0$lcssa<<24>>24)==(0);
  $77 = $76 ? 122 : $valueType$0$lcssa;
  $78 = HEAP32[$ppValTypes>>2]|0;
  $79 = ($78|0)!=(0|0);
  if ($79) {
   $80 = ((($78)) + -8|0);
   $81 = $80;
   $82 = $81;
   $83 = HEAP32[$82>>2]|0;
   $84 = (($81) + 4)|0;
   $85 = $84;
   $86 = HEAP32[$85>>2]|0;
   $94 = $83;$96 = $86;
  } else {
   $94 = 0;$96 = 0;
  }
  $87 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($33)|0);
  $88 = ((($87)) + 8|0);
  $89 = $87;
  $90 = $89;
  HEAP32[$90>>2] = $nestingDepthIncludingTopLevel$0;
  $91 = (($89) + 4)|0;
  $92 = $91;
  HEAP32[$92>>2] = 0;
  $93 = ($94|0)!=(0);
  $95 = ($96|0)!=(0);
  $97 = $93 | $95;
  $or$cond3$i = $45 & $97;
  if ($or$cond3$i) {
   $109 = 0;$113 = 0;
   while(1) {
    $$sum1$i = (($109) + 8)|0;
    $110 = (($87) + ($$sum1$i)|0);
    $111 = (($78) + ($109)|0);
    $112 = HEAP8[$111>>0]|0;
    HEAP8[$110>>0] = $112;
    $114 = (_i64Add(($109|0),($113|0),1,0)|0);
    $115 = tempRet0;
    $116 = ($115>>>0)<($96>>>0);
    $117 = ($114>>>0)<($94>>>0);
    $118 = ($115|0)==($96|0);
    $119 = $118 & $117;
    $120 = $116 | $119;
    $121 = ($115>>>0)<(0);
    $122 = ($114>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
    $123 = ($115|0)==(0);
    $124 = $123 & $122;
    $125 = $121 | $124;
    $or$cond$i4 = $120 & $125;
    if ($or$cond$i4) {
     $109 = $114;$113 = $115;
    } else {
     break;
    }
   }
  }
  $98 = ($96>>>0)<(0);
  $99 = ($94>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
  $100 = ($96|0)==(0);
  $101 = $100 & $99;
  $102 = $98 | $101;
  if ($102) {
   $103 = (($94) + 8)|0;
   $scevgep$i2 = (($87) + ($103)|0);
   $104 = (_i64Add(($nestingDepthIncludingTopLevel$0|0),0,-1,0)|0);
   $105 = tempRet0;
   $106 = (_i64Subtract(($104|0),($105|0),($94|0),($96|0))|0);
   $107 = tempRet0;
   $108 = (($106) + 1)|0;
   _memset(($scevgep$i2|0),0,($108|0))|0;
  }
  if ($79) {
   $126 = ((($78)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($126);
  }
  HEAP32[$ppValTypes>>2] = $88;
  $127 = (($87) + ($$sum)|0);
  HEAP8[$127>>0] = $77;
  $128 = HEAP32[$ppDimensionsOfNestedValues>>2]|0;
  $129 = ($128|0)!=(0|0);
  if ($129) {
   $130 = ((($128)) + -8|0);
   $131 = $130;
   $132 = $131;
   $133 = HEAP32[$132>>2]|0;
   $134 = (($131) + 4)|0;
   $135 = $134;
   $136 = HEAP32[$135>>2]|0;
   $146 = $133;$148 = $136;
  } else {
   $146 = 0;$148 = 0;
  }
  $137 = $nestingDepthIncludingTopLevel$0 << 2;
  $138 = (($137) + 8)|0;
  $139 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($138)|0);
  $140 = ((($139)) + 8|0);
  $141 = $139;
  $142 = $141;
  HEAP32[$142>>2] = $nestingDepthIncludingTopLevel$0;
  $143 = (($141) + 4)|0;
  $144 = $143;
  HEAP32[$144>>2] = 0;
  $145 = ($146|0)!=(0);
  $147 = ($148|0)!=(0);
  $149 = $145 | $147;
  $or$cond2$i = $45 & $149;
  if ($or$cond2$i) {
   $164 = 0;$167 = 0;
   while(1) {
    $163 = (($140) + ($164<<2)|0);
    $165 = (($128) + ($164<<2)|0);
    $166 = HEAP32[$165>>2]|0;
    HEAP32[$163>>2] = $166;
    $168 = (_i64Add(($164|0),($167|0),1,0)|0);
    $169 = tempRet0;
    $170 = ($169>>>0)<($148>>>0);
    $171 = ($168>>>0)<($146>>>0);
    $172 = ($169|0)==($148|0);
    $173 = $172 & $171;
    $174 = $170 | $173;
    $175 = ($169>>>0)<(0);
    $176 = ($168>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
    $177 = ($169|0)==(0);
    $178 = $177 & $176;
    $179 = $175 | $178;
    $or$cond$i = $174 & $179;
    if ($or$cond$i) {
     $164 = $168;$167 = $169;
    } else {
     break;
    }
   }
  }
  $150 = ($148>>>0)<(0);
  $151 = ($146>>>0)<($nestingDepthIncludingTopLevel$0>>>0);
  $152 = ($148|0)==(0);
  $153 = $152 & $151;
  $154 = $150 | $153;
  if ($154) {
   $155 = $146 << 2;
   $156 = (($155) + 8)|0;
   $scevgep$i = (($139) + ($156)|0);
   $157 = (_i64Add(($nestingDepthIncludingTopLevel$0|0),0,1073741823,0)|0);
   $158 = tempRet0;
   $159 = (_i64Subtract(($157|0),($158|0),($146|0),($148|0))|0);
   $160 = tempRet0;
   $161 = $159 << 2;
   $162 = (($161) + 4)|0;
   _memset(($scevgep$i|0),0,($162|0))|0;
  }
  if ($129) {
   $180 = ((($128)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($180);
  }
  HEAP32[$ppDimensionsOfNestedValues>>2] = $140;
  $181 = (($140) + ($23<<2)|0);
  HEAP32[$181>>2] = $dimensions$0$lcssa;
  $182 = (($nestingDepthIncludingTopLevel$0) + 1)|0;
  $nestingDepthIncludingTopLevel$0 = $182;
 }
 return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation19popDictionaryHelperEPhS3_Pj($this,$pKeyTypes,$pValueTypes,$pDimensionsOfNestedValues) {
 $this = $this|0;
 $pKeyTypes = $pKeyTypes|0;
 $pValueTypes = $pValueTypes|0;
 $pDimensionsOfNestedValues = $pDimensionsOfNestedValues|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i1$03 = 0, $k = 0;
 var $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $k = sp + 40|0;
 $v = sp;
 $0 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(60)|0);
 $1 = ((($0)) + 8|0);
 $2 = $0;
 $3 = $2;
 HEAP32[$3>>2] = 1;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 __ZN9ExitGames6Common14DictionaryBaseC2EPKhS3_PKj($1,$pKeyTypes,$pValueTypes,$pDimensionsOfNestedValues);
 $6 = ((($this)) + 4|0);
 $7 = ((($this)) + 8|0);
 $8 = HEAP32[$6>>2]|0;
 $9 = (($8) + 1)|0;
 HEAP32[$6>>2] = $9;
 $10 = HEAP32[$7>>2]|0;
 $11 = (($10) + ($9)|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = (($8) + 2)|0;
 HEAP32[$6>>2] = $13;
 $14 = (($10) + ($13)|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $12&255;
 $17 = ($16 << 8)&65535;
 $18 = $15&255;
 $19 = $17 | $18;
 $20 = ($19<<16>>16)>(0);
 if (!($20)) {
  STACKTOP = sp;return ($1|0);
 }
 $21 = ((($0)) + 12|0);
 $i1$03 = 0;
 while(1) {
  __ZN9ExitGames6Common6ObjectC2Ev($k);
  __THREW__ = 0;
  invoke_vi(743,($v|0));
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   label = 6;
   break;
  }
  $24 = HEAP8[$pKeyTypes>>0]|0;
  $25 = ($24<<24>>24)==(122);
  if ($25) {
   $26 = HEAP32[$6>>2]|0;
   $27 = (($26) + 1)|0;
   HEAP32[$6>>2] = $27;
   $28 = HEAP32[$7>>2]|0;
   $29 = (($28) + ($27)|0);
   $30 = HEAP8[$29>>0]|0;
   __THREW__ = 0;
   (invoke_iiii(813,($this|0),($30|0),($k|0))|0);
   $31 = __THREW__; __THREW__ = 0;
   $32 = $31&1;
   if ($32) {
    label = 7;
    break;
   }
  } else {
   __THREW__ = 0;
   (invoke_iiii(813,($this|0),($24|0),($k|0))|0);
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if ($38) {
    label = 7;
    break;
   }
  }
  $39 = HEAP8[$pValueTypes>>0]|0;
  $40 = ($39<<24>>24)==(122);
  if ($40) {
   label = 11;
  } else {
   $41 = HEAP32[$pDimensionsOfNestedValues>>2]|0;
   $42 = ($41|0)==(0);
   if ($42) {
    __THREW__ = 0;
    (invoke_iiii(813,($this|0),($39|0),($v|0))|0);
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     label = 7;
     break;
    }
   } else {
    label = 11;
   }
  }
  if ((label|0) == 11) {
   label = 0;
   $43 = HEAP32[$6>>2]|0;
   $44 = (($43) + 1)|0;
   HEAP32[$6>>2] = $44;
   $45 = HEAP32[$7>>2]|0;
   $46 = (($45) + ($44)|0);
   $47 = HEAP8[$46>>0]|0;
   __THREW__ = 0;
   (invoke_iiii(813,($this|0),($47|0),($v|0))|0);
   $48 = __THREW__; __THREW__ = 0;
   $49 = $48&1;
   if ($49) {
    label = 7;
    break;
   }
  }
  __THREW__ = 0;
  invoke_viii(841,($21|0),($k|0),($v|0));
  $52 = __THREW__; __THREW__ = 0;
  $53 = $52&1;
  if ($53) {
   label = 7;
   break;
  }
  __ZN9ExitGames6Common6ObjectD2Ev($v);
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  $54 = (($i1$03) + 1)<<16>>16;
  $55 = ($54<<16>>16)<($19<<16>>16);
  if ($55) {
   $i1$03 = $54;
  } else {
   label = 16;
   break;
  }
 }
 if ((label|0) == 6) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  $$0 = $34;$$01 = $33;
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  ___resumeException($$01|0);
  // unreachable;
 }
 else if ((label|0) == 7) {
  $35 = ___cxa_find_matching_catch_2()|0;
  $36 = tempRet0;
  __ZN9ExitGames6Common6ObjectD2Ev($v);
  $$0 = $36;$$01 = $35;
  __ZN9ExitGames6Common6ObjectD2Ev($k);
  ___resumeException($$01|0);
  // unreachable;
 }
 else if ((label|0) == 16) {
  STACKTOP = sp;return ($1|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common14DictionaryBaseaSERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 2047]($this,$toCopy)|0);
 return ($3|0);
}
function __ZN9ExitGames6Common14DictionaryBaseC2EPKhS3_PKj($this,$pKeyTypes,$pValueTypes,$pValueDimensions) {
 $this = $this|0;
 $pKeyTypes = $pKeyTypes|0;
 $pValueTypes = $pValueTypes|0;
 $pValueDimensions = $pValueDimensions|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5516);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vi(838,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  $$0 = $12;$$01 = $11;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 $3 = (invoke_ii(775,16)|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   $13 = ___cxa_find_matching_catch_2()|0;
   $14 = tempRet0;
   $eh$lpad$body$index2Z2D = $14;$eh$lpad$body$indexZ2D = $13;
  } else {
   HEAP32[$3>>2] = (5564);
   __THREW__ = 0;
   invoke_viiii(856,($3|0),($pKeyTypes|0),($pValueTypes|0),($pValueDimensions|0));
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if ($7) {
    $8 = ___cxa_find_matching_catch_2()|0;
    $9 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($3);
    $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
    break;
   } else {
    $10 = ((($this)) + 48|0);
    HEAP32[$10>>2] = $3;
    return;
   }
  }
 } while(0);
 __ZN9ExitGames6Common9HashtableD2Ev($0);
 $$0 = $eh$lpad$body$index2Z2D;$$01 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Common14DictionaryBase8TypeInfo4initEPKhS4_PKj($this,$pKeyTypes,$pValueTypes,$pValueDimensions) {
 $this = $this|0;
 $pKeyTypes = $pKeyTypes|0;
 $pValueTypes = $pValueTypes|0;
 $pValueDimensions = $pValueDimensions|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $i$04 = 0, $nestingDepth$0 = 0, $nestingDepth$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $nestingDepth$0 = 0;
 while(1) {
  $0 = (($pValueTypes) + ($nestingDepth$0)|0);
  $1 = HEAP8[$0>>0]|0;
  $2 = ($1<<24>>24)==(68);
  $3 = (($nestingDepth$0) + 1)|0;
  if ($2) {
   $nestingDepth$0 = $3;
  } else {
   $$lcssa = $3;$nestingDepth$0$lcssa = $nestingDepth$0;
   break;
  }
 }
 $4 = (($nestingDepth$0$lcssa) + 9)|0;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($4)|0);
 $6 = ((($5)) + 8|0);
 $7 = $5;
 $8 = $7;
 HEAP32[$8>>2] = $$lcssa;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = 0;
 $11 = ($$lcssa|0)==(0);
 if (!($11)) {
  _memset(($6|0),0,($$lcssa|0))|0;
 }
 $12 = ((($this)) + 4|0);
 HEAP32[$12>>2] = $6;
 $13 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($4)|0);
 $14 = ((($13)) + 8|0);
 $15 = $13;
 $16 = $15;
 HEAP32[$16>>2] = $$lcssa;
 $17 = (($15) + 4)|0;
 $18 = $17;
 HEAP32[$18>>2] = 0;
 if (!($11)) {
  _memset(($14|0),0,($$lcssa|0))|0;
 }
 $19 = ((($this)) + 8|0);
 HEAP32[$19>>2] = $14;
 $20 = $$lcssa << 2;
 $21 = (($20) + 8)|0;
 $22 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($21)|0);
 $23 = ((($22)) + 8|0);
 $24 = $22;
 $25 = $24;
 HEAP32[$25>>2] = $$lcssa;
 $26 = (($24) + 4)|0;
 $27 = $26;
 HEAP32[$27>>2] = 0;
 if ($11) {
  $28 = ((($this)) + 12|0);
  HEAP32[$28>>2] = $23;
  return;
 }
 _memset(($23|0),0,($20|0))|0;
 $29 = ((($this)) + 12|0);
 HEAP32[$29>>2] = $23;
 $i$04 = 0;
 while(1) {
  $30 = (($pKeyTypes) + ($i$04)|0);
  $31 = HEAP8[$30>>0]|0;
  $32 = HEAP32[$12>>2]|0;
  $33 = (($32) + ($i$04)|0);
  HEAP8[$33>>0] = $31;
  $34 = (($pValueTypes) + ($i$04)|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = HEAP32[$19>>2]|0;
  $37 = (($36) + ($i$04)|0);
  HEAP8[$37>>0] = $35;
  $38 = (($pValueDimensions) + ($i$04<<2)|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = HEAP32[$29>>2]|0;
  $41 = (($40) + ($i$04<<2)|0);
  HEAP32[$41>>2] = $39;
  $42 = (($i$04) + 1)|0;
  $exitcond = ($42|0)==($$lcssa|0);
  if ($exitcond) {
   break;
  } else {
   $i$04 = $42;
  }
 }
 return;
}
function __ZN9ExitGames6Common14DictionaryBase8TypeInfoD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5564);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  label = 3;
 } else {
  $3 = ((($1)) + -8|0);
  __THREW__ = 0;
  invoke_vi(750,($3|0));
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   label = 3;
  }
 }
 do {
  if ((label|0) == 3) {
   HEAP32[$0>>2] = 0;
   $6 = ((($this)) + 8|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if (!($8)) {
    $9 = ((($7)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($9|0));
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     break;
    }
   }
   HEAP32[$6>>2] = 0;
   $12 = ((($this)) + 12|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==(0|0);
   if ($14) {
    HEAP32[$12>>2] = 0;
    __ZN9ExitGames6Common4BaseD2Ev($this);
    return;
   }
   $15 = ((($13)) + -8|0);
   __THREW__ = 0;
   invoke_vi(750,($15|0));
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if (!($17)) {
    HEAP32[$12>>2] = 0;
    __ZN9ExitGames6Common4BaseD2Ev($this);
    return;
   }
  }
 } while(0);
 $18 = ___cxa_find_matching_catch_3(0|0)|0;
 $19 = tempRet0;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 ___clang_call_terminate($18);
 // unreachable;
}
function __ZN9ExitGames6Common14DictionaryBase8TypeInfoD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14DictionaryBase8TypeInfoD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Common14DictionaryBase8TypeInfo8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 __ZNK9ExitGames6Common14DictionaryBase8TypeInfo16toStringRecursorEbj($0,$this,$withTypes,0);
 __THREW__ = 0;
 $1 = (invoke_iii(760,($retStr|0),($0|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($4|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  STACKTOP = sp;return ($1|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common14DictionaryBase8TypeInfo16toStringRecursorEbj($agg$result,$this,$withTypes,$recursionDepth) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $withTypes = $withTypes|0;
 $recursionDepth = $recursionDepth|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $$5 = 0, $$56 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index3Z2D = 0;
 var $eh$lpad$body$indexZ2D = 0, $i$07 = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $str$i = sp + 96|0;
 $0 = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp + 48|0;
 $3 = sp + 32|0;
 $4 = sp + 16|0;
 $5 = sp;
 $6 = ((($this)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($7) + ($recursionDepth)|0);
 $9 = HEAP8[$8>>0]|0;
 __ZN9ExitGames6Common14DictionaryBase8TypeInfo12typeToStringEh($1,$9);
 __THREW__ = 0;
 invoke_viii(857,($0|0),($1|0),(12352|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $38 = ___cxa_find_matching_catch_2()|0;
  $39 = tempRet0;
  $$1 = $39;$$12 = $38;
 } else {
  $12 = ((($this)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) + ($recursionDepth)|0);
  $15 = HEAP8[$14>>0]|0;
  __THREW__ = 0;
  invoke_vii(858,($2|0),($15|0));
  $16 = __THREW__; __THREW__ = 0;
  $17 = $16&1;
  do {
   if ($17) {
    $40 = ___cxa_find_matching_catch_2()|0;
    $41 = tempRet0;
    $$0 = $41;$$01 = $40;
   } else {
    __THREW__ = 0;
    invoke_viii(785,($agg$result|0),($0|0),($2|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $42 = ___cxa_find_matching_catch_2()|0;
     $43 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($2);
     $$0 = $43;$$01 = $42;
     break;
    }
    __ZN9ExitGames6Common7JStringD2Ev($2);
    __ZN9ExitGames6Common7JStringD2Ev($0);
    __ZN9ExitGames6Common7JStringD2Ev($1);
    $20 = ((($this)) + 12|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = (($21) + ($recursionDepth<<2)|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($23|0)==(0);
    L10: do {
     if ($24) {
      label = 19;
     } else {
      $i$07 = 0;
      while(1) {
       __THREW__ = 0;
       invoke_vii(787,($str$i|0),0);
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       if ($26) {
        label = 16;
        break;
       }
       __THREW__ = 0;
       (invoke_iii(759,($str$i|0),(5580|0))|0);
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        label = 9;
        break;
       }
       __THREW__ = 0;
       (invoke_iii(760,($agg$result|0),($str$i|0))|0);
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       if ($30) {
        label = 9;
        break;
       }
       __ZN9ExitGames6Common7JStringD2Ev($str$i);
       $33 = (($i$07) + 1)|0;
       $34 = HEAP32[$20>>2]|0;
       $35 = (($34) + ($recursionDepth<<2)|0);
       $36 = HEAP32[$35>>2]|0;
       $37 = ($33>>>0)<($36>>>0);
       if ($37) {
        $i$07 = $33;
       } else {
        label = 19;
        break L10;
       }
      }
      if ((label|0) == 9) {
       $31 = ___cxa_find_matching_catch_2()|0;
       $32 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($str$i);
       $eh$lpad$body$index3Z2D = $32;$eh$lpad$body$indexZ2D = $31;
       label = 18;
       break;
      }
      else if ((label|0) == 16) {
       $44 = ___cxa_find_matching_catch_2()|0;
       $45 = tempRet0;
       $eh$lpad$body$index3Z2D = $45;$eh$lpad$body$indexZ2D = $44;
       label = 18;
       break;
      }
     }
    } while(0);
    do {
     if ((label|0) == 19) {
      $48 = HEAP32[$12>>2]|0;
      $49 = (($48) + ($recursionDepth)|0);
      $50 = HEAP8[$49>>0]|0;
      $51 = ($50<<24>>24)==(68);
      if (!($51)) {
       STACKTOP = sp;return;
      }
      $52 = (($recursionDepth) + 1)|0;
      __THREW__ = 0;
      invoke_viiii(859,($5|0),($this|0),($withTypes|0),($52|0));
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       $46 = ___cxa_find_matching_catch_2()|0;
       $47 = tempRet0;
       $eh$lpad$body$index3Z2D = $47;$eh$lpad$body$indexZ2D = $46;
       label = 18;
       break;
      }
      __THREW__ = 0;
      invoke_viii(860,($4|0),(5588|0),($5|0));
      $55 = __THREW__; __THREW__ = 0;
      $56 = $55&1;
      if ($56) {
       $61 = ___cxa_find_matching_catch_2()|0;
       $62 = tempRet0;
       $$3 = $62;$$34 = $61;
      } else {
       __THREW__ = 0;
       invoke_viii(788,($3|0),($4|0),(5596|0));
       $57 = __THREW__; __THREW__ = 0;
       $58 = $57&1;
       do {
        if ($58) {
         $63 = ___cxa_find_matching_catch_2()|0;
         $64 = tempRet0;
         $$2 = $64;$$23 = $63;
        } else {
         __THREW__ = 0;
         (invoke_iii(760,($agg$result|0),($3|0))|0);
         $59 = __THREW__; __THREW__ = 0;
         $60 = $59&1;
         if ($60) {
          $65 = ___cxa_find_matching_catch_2()|0;
          $66 = tempRet0;
          __ZN9ExitGames6Common7JStringD2Ev($3);
          $$2 = $66;$$23 = $65;
          break;
         }
         __ZN9ExitGames6Common7JStringD2Ev($3);
         __ZN9ExitGames6Common7JStringD2Ev($4);
         __ZN9ExitGames6Common7JStringD2Ev($5);
         STACKTOP = sp;return;
        }
       } while(0);
       __ZN9ExitGames6Common7JStringD2Ev($4);
       $$3 = $$2;$$34 = $$23;
      }
      __ZN9ExitGames6Common7JStringD2Ev($5);
      $$4 = $$3;$$45 = $$34;
     }
    } while(0);
    if ((label|0) == 18) {
     $$4 = $eh$lpad$body$index3Z2D;$$45 = $eh$lpad$body$indexZ2D;
    }
    __ZN9ExitGames6Common7JStringD2Ev($agg$result);
    $$5 = $$4;$$56 = $$45;
    ___resumeException($$56|0);
    // unreachable;
   }
  } while(0);
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $$1 = $$0;$$12 = $$01;
 }
 __ZN9ExitGames6Common7JStringD2Ev($1);
 $$5 = $$1;$$56 = $$12;
 ___resumeException($$56|0);
 // unreachable;
}
function __ZN9ExitGames6Common14DictionaryBase8TypeInfo12typeToStringEh($agg$result,$typeCode) {
 $agg$result = $agg$result|0;
 $typeCode = $typeCode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $retStr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $retStr = sp;
 __ZN9ExitGames6Common7JStringC2Ej($retStr,0);
 $0 = $typeCode&255;
 do {
  switch ($0|0) {
  case 98:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162512|0));
   $1 = __THREW__; __THREW__ = 0;
   $2 = $1&1;
   if ($2) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 107:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162528|0));
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 105:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162560|0));
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 108:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162624|0));
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 102:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162656|0));
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if ($12) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 100:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162672|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 111:  {
   __THREW__ = 0;
   invoke_vii(786,($agg$result|0),(162704|0));
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 42:  {
   __THREW__ = 0;
   invoke_vii(801,($agg$result|0),(5604|0));
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 115:  {
   __THREW__ = 0;
   invoke_vii(739,($agg$result|0),(146044|0));
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if ($20) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 104:  {
   __THREW__ = 0;
   invoke_vii(739,($agg$result|0),(146052|0));
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 68:  {
   __THREW__ = 0;
   invoke_vii(739,($agg$result|0),(146062|0));
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 122:  {
   __THREW__ = 0;
   invoke_vii(739,($agg$result|0),(146073|0));
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  case 99:  {
   __THREW__ = 0;
   invoke_vii(739,($agg$result|0),(146080|0));
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    label = 3;
   } else {
    label = 17;
   }
   break;
  }
  default: {
   __THREW__ = 0;
   invoke_vii(801,($agg$result|0),(5624|0));
   $29 = __THREW__; __THREW__ = 0;
   $30 = $29&1;
   if ($30) {
    label = 3;
   } else {
    label = 17;
   }
  }
  }
 } while(0);
 if ((label|0) == 3) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($retStr);
  ___resumeException($3|0);
  // unreachable;
 }
 else if ((label|0) == 17) {
  __ZN9ExitGames6Common7JStringD2Ev($retStr);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6CommonplIA3_wEENS0_7JStringERKS3_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($Rsh|0))|0);
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($5|0));
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 $eh$lpad$body$index2Z2D = $13;$eh$lpad$body$indexZ2D = $12;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Common14DictionaryBaseD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5516);
 $0 = ((($this)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $7 = ((($this)) + 4|0);
  __ZN9ExitGames6Common9HashtableD2Ev($7);
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
 $3 = HEAP32[$1>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vi[$4 & 2047]($1);
 __THREW__ = 0;
 invoke_vi(750,($1|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $8 = ___cxa_find_matching_catch_3(0|0)|0;
  $9 = tempRet0;
  $10 = ((($this)) + 4|0);
  __ZN9ExitGames6Common9HashtableD2Ev($10);
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($8);
  // unreachable;
 } else {
  $7 = ((($this)) + 4|0);
  __ZN9ExitGames6Common9HashtableD2Ev($7);
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common14DictionaryBaseD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5516);
 $0 = ((($this)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0;
  $4 = HEAP32[$3>>2]|0;
  FUNCTION_TABLE_vi[$4 & 2047]($1);
  __THREW__ = 0;
  invoke_vi(750,($1|0));
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $7 = ___cxa_find_matching_catch_3(0|0)|0;
   $8 = tempRet0;
   $9 = ((($this)) + 4|0);
   __ZN9ExitGames6Common9HashtableD2Ev($9);
   __ZN9ExitGames6Common4BaseD2Ev($this);
   ___clang_call_terminate($7);
   // unreachable;
  }
 }
 $10 = ((($this)) + 4|0);
 __ZN9ExitGames6Common9HashtableD2Ev($10);
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Common14DictionaryBase12typeToStringEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body7$index2Z2D = 0, $eh$lpad$body7$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $str$i = sp + 96|0;
 $0 = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp + 48|0;
 $3 = sp + 32|0;
 $4 = sp + 16|0;
 $5 = sp;
 __ZN9ExitGames6Common7JStringC2EPKw($4,13660);
 __THREW__ = 0;
 invoke_vii(786,($0|0),($4|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $31 = ___cxa_find_matching_catch_2()|0;
  $32 = tempRet0;
  $eh$lpad$body$index7Z2D = $32;$eh$lpad$body$indexZ2D = $31;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  do {
   if ($9) {
    label = 7;
   } else {
    __THREW__ = 0;
    (invoke_iii(861,($str$i|0),(146062|0))|0);
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if (!($11)) {
     __THREW__ = 0;
     $12 = (invoke_iii(760,($0|0),($str$i|0))|0);
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if (!($14)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($3|0),($12|0));
      $17 = __THREW__; __THREW__ = 0;
      $18 = $17&1;
      if ($18) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($0);
      __THREW__ = 0;
      invoke_viii(788,($2|0),($3|0),(5588|0));
      $21 = __THREW__; __THREW__ = 0;
      $22 = $21&1;
      if ($22) {
       $33 = ___cxa_find_matching_catch_2()|0;
       $34 = tempRet0;
       $$2 = $34;$$23 = $33;
      } else {
       $23 = ((($this)) + 48|0);
       $24 = HEAP32[$23>>2]|0;
       __THREW__ = 0;
       invoke_viii(783,($5|0),($24|0),0);
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       if ($26) {
        $35 = ___cxa_find_matching_catch_2()|0;
        $36 = tempRet0;
        $$1 = $36;$$12 = $35;
       } else {
        __THREW__ = 0;
        invoke_viii(785,($1|0),($2|0),($5|0));
        $27 = __THREW__; __THREW__ = 0;
        $28 = $27&1;
        do {
         if ($28) {
          $37 = ___cxa_find_matching_catch_2()|0;
          $38 = tempRet0;
          $$0 = $38;$$01 = $37;
         } else {
          __THREW__ = 0;
          invoke_viii(857,($agg$result|0),($1|0),(5684|0));
          $29 = __THREW__; __THREW__ = 0;
          $30 = $29&1;
          if ($30) {
           $39 = ___cxa_find_matching_catch_2()|0;
           $40 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($1);
           $$0 = $40;$$01 = $39;
           break;
          } else {
           __ZN9ExitGames6Common7JStringD2Ev($1);
           __ZN9ExitGames6Common7JStringD2Ev($5);
           __ZN9ExitGames6Common7JStringD2Ev($2);
           __ZN9ExitGames6Common7JStringD2Ev($3);
           __ZN9ExitGames6Common7JStringD2Ev($4);
           STACKTOP = sp;return;
          }
         }
        } while(0);
        __ZN9ExitGames6Common7JStringD2Ev($5);
        $$1 = $$0;$$12 = $$01;
       }
       __ZN9ExitGames6Common7JStringD2Ev($2);
       $$2 = $$1;$$23 = $$12;
      }
      __ZN9ExitGames6Common7JStringD2Ev($3);
      $$3 = $$2;$$34 = $$23;
      __ZN9ExitGames6Common7JStringD2Ev($4);
      ___resumeException($$34|0);
      // unreachable;
     }
    }
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body7$index2Z2D = $16;$eh$lpad$body7$indexZ2D = $15;
   }
  } while(0);
  if ((label|0) == 7) {
   $19 = ___cxa_find_matching_catch_2()|0;
   $20 = tempRet0;
   $eh$lpad$body7$index2Z2D = $20;$eh$lpad$body7$indexZ2D = $19;
  }
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $eh$lpad$body$index7Z2D = $eh$lpad$body7$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body7$indexZ2D;
 }
 $$3 = $eh$lpad$body$index7Z2D;$$34 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($4);
 ___resumeException($$34|0);
 // unreachable;
}
function __ZNK9ExitGames6Common14DictionaryBase8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $7 = 0, $8 = 0, $9 = 0, $i$06 = 0, $str$i = 0, $str$i1 = 0, $str$i3 = 0, $str$i4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer = sp;
 $str$i4 = sp + 72|0;
 $str$i3 = sp + 56|0;
 $str$i1 = sp + 40|0;
 $0 = sp + 24|0;
 $str$i = sp + 8|0;
 __ZN9ExitGames6Common7JStringC2Ej($str$i,0);
 __THREW__ = 0;
 (invoke_iii(759,($str$i|0),(5676|0))|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i);
  ___resumeException($5|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(760,($retStr|0),($str$i|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i);
  ___resumeException($5|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($str$i);
 $7 = ((($this)) + 4|0);
 $8 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($7)|0);
 $9 = ($8|0)==(0);
 L7: do {
  if (!($9)) {
   $10 = ((($this)) + 48|0);
   $i$06 = 0;
   while(1) {
    $11 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($7)|0);
    $12 = ((($11)) + 4|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ($13>>>0)>($i$06>>>0);
    if (!($14)) {
     __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
     __THREW__ = 0;
     $15 = (invoke_ii(764,($0|0))|0);
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if ($17) {
      label = 10;
      break;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($15|0),1,566,(10968|0),($vararg_buffer|0));
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      label = 10;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
    }
    $22 = ((($11)) + 16|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (($23) + (($i$06*40)|0)|0);
    $25 = HEAP32[$10>>2]|0;
    $26 = ($25|0)==(0|0);
    if ($26) {
     $31 = 0;
    } else {
     $27 = ((($25)) + 4|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = HEAP8[$28>>0]|0;
     $30 = ($29<<24>>24)!=(0);
     $31 = $30;
    }
    (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($24,$retStr,$withTypes,$31)|0);
    __ZN9ExitGames6Common7JStringC2Ej($str$i1,0);
    __THREW__ = 0;
    (invoke_iii(759,($str$i1|0),(13052|0))|0);
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     label = 15;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(760,($retStr|0),($str$i1|0))|0);
    $34 = __THREW__; __THREW__ = 0;
    $35 = $34&1;
    if ($35) {
     label = 15;
     break;
    }
    __ZN9ExitGames6Common7JStringD2Ev($str$i1);
    $38 = (__ZNK9ExitGames6Common9HashtableixEj($7,$i$06)|0);
    $39 = HEAP32[$10>>2]|0;
    $40 = ($39|0)==(0|0);
    if ($40) {
     $45 = 0;
    } else {
     $41 = ((($39)) + 8|0);
     $42 = HEAP32[$41>>2]|0;
     $43 = HEAP8[$42>>0]|0;
     $44 = ($43<<24>>24)!=(0);
     $45 = $44;
    }
    (__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb($38,$retStr,$withTypes,$45)|0);
    $46 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($7)|0);
    $47 = (($46) + -1)|0;
    $48 = ($i$06>>>0)<($47>>>0);
    if ($48) {
     __ZN9ExitGames6Common7JStringC2Ej($str$i3,0);
     __THREW__ = 0;
     (invoke_iii(759,($str$i3|0),(12352|0))|0);
     $49 = __THREW__; __THREW__ = 0;
     $50 = $49&1;
     if ($50) {
      label = 21;
      break;
     }
     __THREW__ = 0;
     (invoke_iii(760,($retStr|0),($str$i3|0))|0);
     $51 = __THREW__; __THREW__ = 0;
     $52 = $51&1;
     if ($52) {
      label = 21;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($str$i3);
    }
    $55 = (($i$06) + 1)|0;
    $56 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($7)|0);
    $57 = ($55>>>0)<($56>>>0);
    if ($57) {
     $i$06 = $55;
    } else {
     break L7;
    }
   }
   if ((label|0) == 10) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($20|0);
    // unreachable;
   }
   else if ((label|0) == 15) {
    $36 = ___cxa_find_matching_catch_2()|0;
    $37 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i1);
    ___resumeException($36|0);
    // unreachable;
   }
   else if ((label|0) == 21) {
    $53 = ___cxa_find_matching_catch_2()|0;
    $54 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i3);
    ___resumeException($53|0);
    // unreachable;
   }
  }
 } while(0);
 __ZN9ExitGames6Common7JStringC2Ej($str$i4,0);
 __THREW__ = 0;
 (invoke_iii(759,($str$i4|0),(10392|0))|0);
 $58 = __THREW__; __THREW__ = 0;
 $59 = $58&1;
 if ($59) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i4);
  ___resumeException($62|0);
  // unreachable;
 }
 __THREW__ = 0;
 (invoke_iii(760,($retStr|0),($str$i4|0))|0);
 $60 = __THREW__; __THREW__ = 0;
 $61 = $60&1;
 if ($61) {
  $62 = ___cxa_find_matching_catch_2()|0;
  $63 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($str$i4);
  ___resumeException($62|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($str$i4);
  STACKTOP = sp;return ($retStr|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common14DictionaryBase11getKeyTypesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function __ZNK9ExitGames6Common14DictionaryBase13getValueTypesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function __ZNK9ExitGames6Common14DictionaryBase18getValueDimensionsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 48|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 return ($3|0);
}
function __ZN9ExitGames6Common14DictionaryBase6assignERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = ((($toCopy)) + 4|0);
 (__ZN9ExitGames6Common9HashtableaSERKS1_($0,$1)|0);
 $2 = ((($this)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = HEAP32[$3>>2]|0;
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vi[$6 & 2047]($3);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($3);
 }
 $7 = ((($toCopy)) + 48|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if ($9) {
  $21 = 0;
  HEAP32[$2>>2] = $21;
  return ($this|0);
 }
 $10 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(16)|0);
 HEAP32[$10>>2] = (5564);
 $11 = ((($8)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ((($8)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($8)) + 12|0);
 $16 = HEAP32[$15>>2]|0;
 __THREW__ = 0;
 invoke_viiii(856,($10|0),($12|0),($14|0),($16|0));
 $17 = __THREW__; __THREW__ = 0;
 $18 = $17&1;
 if ($18) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($10);
  ___resumeException($19|0);
  // unreachable;
 } else {
  $21 = $10;
  HEAP32[$2>>2] = $21;
  return ($this|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common14DictionaryBase7compareERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = ((($toCompare)) + 4|0);
 $2 = (__ZNK9ExitGames6Common9HashtableeqERKS1_($0,$1)|0);
 return ($2|0);
}
function __ZNK9ExitGames6Common14DictionaryBase4copyEs($this,$amount) {
 $this = $this|0;
 $amount = $amount|0;
 var $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $lftr$wideiv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $amount << 16 >> 16;
 $1 = ($0*52)|0;
 $2 = (($1) + 8)|0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($2)|0);
 $4 = ((($3)) + 8|0);
 $5 = $3;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = 0;
 $9 = ($amount<<16>>16)==(0);
 if ($9) {
  return ($4|0);
 } else {
  $12 = 0;$19 = 0;
 }
 while(1) {
  $11 = (($4) + (($12*52)|0)|0);
  HEAP32[$11>>2] = (5516);
  $13 = (((($4) + (($12*52)|0)|0)) + 4|0);
  __THREW__ = 0;
  invoke_vi(838,($13|0));
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   $$lcssa = $11;
   label = 4;
   break;
  }
  $18 = (((($4) + (($12*52)|0)|0)) + 48|0);
  HEAP32[$18>>2] = 0;
  $20 = (_i64Add(($12|0),($19|0),1,0)|0);
  $21 = tempRet0;
  $22 = ($21>>>0)<(0);
  $23 = ($20>>>0)<($0>>>0);
  $24 = ($21|0)==(0);
  $25 = $24 & $23;
  $26 = $22 | $25;
  if ($26) {
   $12 = $20;$19 = $21;
  } else {
   break;
  }
 }
 if ((label|0) == 4) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($$lcssa);
  ___resumeException($16|0);
  // unreachable;
 }
 $10 = ($amount<<16>>16)>(0);
 if ($10) {
  $indvars$iv = 0;
 } else {
  return ($4|0);
 }
 while(1) {
  $27 = (($4) + (($indvars$iv*52)|0)|0);
  $28 = (($this) + (($indvars$iv*52)|0)|0);
  $29 = HEAP32[$27>>2]|0;
  $30 = ((($29)) + 28|0);
  $31 = HEAP32[$30>>2]|0;
  (FUNCTION_TABLE_iii[$31 & 2047]($27,$28)|0);
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $lftr$wideiv = $indvars$iv$next&65535;
  $exitcond = ($lftr$wideiv<<16>>16)==($amount<<16>>16);
  if ($exitcond) {
   break;
  } else {
   $indvars$iv = $indvars$iv$next;
  }
 }
 return ($4|0);
}
function __ZNK9ExitGames6Common9HashtableeqERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i$02 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($toCompare)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3|0)==($5|0);
 if (!($6)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = ($3|0)==(0);
 if ($7) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($this)) + 20|0);
 $13 = $3;$i$02 = 0;
 while(1) {
  $12 = ($13>>>0)>($i$02>>>0);
  if (!($12)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $14 = (invoke_ii(764,($0|0))|0);
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    label = 9;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($14|0),1,566,(10968|0),($vararg_buffer|0));
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 9;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  $21 = HEAP32[$8>>2]|0;
  $22 = (($21) + (($i$02*40)|0)|0);
  $23 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($this,$22)|0);
  $24 = ($23|0)==(0|0);
  if ($24) {
   $$0 = 0;
   label = 18;
   break;
  }
  $25 = HEAP32[$2>>2]|0;
  $26 = ($25>>>0)>($i$02>>>0);
  if (!($26)) {
   __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
   __THREW__ = 0;
   $27 = (invoke_ii(764,($1|0))|0);
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    label = 15;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($27|0),1,566,(10968|0),($vararg_buffer1|0));
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    label = 15;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($1);
  }
  $34 = HEAP32[$8>>2]|0;
  $35 = (($34) + (($i$02*40)|0)|0);
  $36 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($toCompare,$35)|0);
  $37 = ($36|0)==(0|0);
  if ($37) {
   $$0 = 0;
   label = 18;
   break;
  }
  $38 = (__ZNK9ExitGames6Common6ObjectneERKS1_($23,$36)|0);
  $11 = (($i$02) + 1)|0;
  if ($38) {
   $$0 = 0;
   label = 18;
   break;
  }
  $9 = HEAP32[$2>>2]|0;
  $10 = ($11>>>0)<($9>>>0);
  if ($10) {
   $13 = $9;$i$02 = $11;
  } else {
   $$0 = 1;
   label = 18;
   break;
  }
 }
 if ((label|0) == 9) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($19|0);
  // unreachable;
 }
 else if ((label|0) == 15) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($32|0);
  // unreachable;
 }
 else if ((label|0) == 18) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($this,$key) {
 $this = $this|0;
 $key = $key|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $0 = sp + 40|0;
 $1 = sp;
 __ZN9ExitGames6Common6ObjectC2ERKS1_($1,$key);
 __THREW__ = 0;
 invoke_vii(745,($0|0),($1|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  $$0 = $8;$$01 = $7;
  __ZN9ExitGames6Common6ObjectD2Ev($1);
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 $4 = (invoke_iii(862,($this|0),($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __ZN9ExitGames6Common6ObjectD2Ev($0);
  __ZN9ExitGames6Common6ObjectD2Ev($1);
  STACKTOP = sp;return ($4|0);
 }
 $9 = ___cxa_find_matching_catch_2()|0;
 $10 = tempRet0;
 __ZN9ExitGames6Common6ObjectD2Ev($0);
 $$0 = $10;$$01 = $9;
 __ZN9ExitGames6Common6ObjectD2Ev($1);
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZNK9ExitGames6Common6ObjectneERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common6ObjecteqERKS1_($this,$toCompare)|0);
 $1 = $0 ^ 1;
 return ($1|0);
}
function __ZNK9ExitGames6Common6ObjecteqERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $i$018 = 0, $or$cond = 0, $or$cond16 = 0, $or$cond19 = 0, $or$cond20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $3 = ((($this)) + 28|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(42);
  if ($5) {
   label = 4;
  } else {
   $6 = ((($this)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if (!($8)) {
    label = 4;
   }
  }
  L5: do {
   if ((label|0) == 4) {
    $9 = ((($toCompare)) + 16|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)==(0|0);
    do {
     if ($11) {
      $12 = ((($toCompare)) + 28|0);
      $13 = HEAP8[$12>>0]|0;
      $14 = ($13<<24>>24)==(42);
      if (!($14)) {
       $15 = ((($toCompare)) + 32|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = ($16|0)==(0);
       if ($17) {
        break;
       }
      }
      $18 = ((($this)) + 24|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = ($19|0)==(0|0);
      $21 = ((($this)) + 20|0);
      $22 = $20 ? $21 : $19;
      $23 = ($22|0)==(0|0);
      if ($23) {
       $24 = ((($toCompare)) + 24|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = ($25|0)==(0|0);
       $27 = ((($toCompare)) + 20|0);
       $28 = $26 ? $27 : $25;
       $29 = ($28|0)==(0|0);
       $brmerge = $29 | $5;
       if ($brmerge) {
        $$0 = $29;
       } else {
        break L5;
       }
       return ($$0|0);
      }
     }
    } while(0);
    if ($5) {
     $$0 = 0;
     return ($$0|0);
    }
   }
  } while(0);
  $30 = ((($this)) + 32|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ($31|0)==(0);
  if (!($32)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $33 = ((($toCompare)) + 16|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = ($34|0)==(0|0);
 if ($35) {
  $36 = ((($toCompare)) + 28|0);
  $37 = HEAP8[$36>>0]|0;
  $38 = ($37<<24>>24)==(42);
  if ($38) {
   $$0 = 0;
   return ($$0|0);
  }
  $39 = ((($toCompare)) + 32|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ($40|0)==(0);
  if (!($41)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $42 = ((($this)) + 24|0);
 $43 = HEAP32[$42>>2]|0;
 $44 = ($43|0)==(0|0);
 $45 = ((($this)) + 20|0);
 $46 = $44 ? $45 : $43;
 $47 = ($46|0)==(0|0);
 if ($47) {
  $$0 = 0;
  return ($$0|0);
 }
 $48 = ((($toCompare)) + 24|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = ($49|0)==(0|0);
 $51 = ((($toCompare)) + 20|0);
 $52 = $50 ? $51 : $49;
 $53 = ($52|0)==(0|0);
 if ($53) {
  $$0 = 0;
  return ($$0|0);
 }
 $54 = ((($this)) + 28|0);
 $55 = HEAP8[$54>>0]|0;
 $56 = ((($toCompare)) + 28|0);
 $57 = HEAP8[$56>>0]|0;
 $58 = ($55<<24>>24)==($57<<24>>24);
 if (!($58)) {
  $$0 = 0;
  return ($$0|0);
 }
 $59 = ((($this)) + 32|0);
 $60 = HEAP32[$59>>2]|0;
 $61 = ((($toCompare)) + 32|0);
 $62 = HEAP32[$61>>2]|0;
 $63 = ($60|0)==($62|0);
 if (!($63)) {
  $$0 = 0;
  return ($$0|0);
 }
 $64 = ($55<<24>>24)==(99);
 if ($64) {
  $65 = ((($this)) + 29|0);
  $66 = HEAP8[$65>>0]|0;
  $71 = $66;
 } else {
  $71 = 0;
 }
 $67 = ((($toCompare)) + 29|0);
 $68 = HEAP8[$67>>0]|0;
 $69 = $64 ? $68 : 0;
 $70 = ($71<<24>>24)==($69<<24>>24);
 if (!($70)) {
  $$0 = 0;
  return ($$0|0);
 }
 $72 = ($60|0)==(0);
 $$ = $72 ? 1 : $60;
 $i$018 = 0;
 while(1) {
  $75 = (($46) + ($i$018<<1)|0);
  $76 = HEAP16[$75>>1]|0;
  $77 = (($52) + ($i$018<<1)|0);
  $78 = HEAP16[$77>>1]|0;
  $79 = ($76<<16>>16)==($78<<16>>16);
  $74 = (($i$018) + 1)|0;
  if (!($79)) {
   $$0 = 0;
   label = 41;
   break;
  }
  $73 = ($74>>>0)<($$>>>0);
  if ($73) {
   $i$018 = $74;
  } else {
   break;
  }
 }
 if ((label|0) == 41) {
  return ($$0|0);
 }
 $80 = $55&255;
 switch ($80|0) {
 case 99: case 100: case 102: case 111: case 108: case 105: case 107: case 98:  {
  if ($2) {
   $81 = ($55<<24>>24)!=(42);
   $82 = ($60|0)==(0);
   $or$cond19 = $81 & $82;
   if ($or$cond19) {
    $83 = ((($this)) + 8|0);
    $89 = $83;
   } else {
    $89 = 0;
   }
  } else {
   $89 = $1;
  }
  if ($35) {
   $84 = ($55<<24>>24)!=(42);
   $85 = ($60|0)==(0);
   $or$cond = $84 & $85;
   if ($or$cond) {
    $86 = ((($toCompare)) + 8|0);
    $90 = $86;
   } else {
    $90 = 0;
   }
  } else {
   $90 = $34;
  }
  if ($64) {
   $87 = ((($this)) + 29|0);
   $88 = HEAP8[$87>>0]|0;
   $91 = $88;
  } else {
   $91 = 0;
  }
  $92 = (__ZNK9ExitGames6Common6Object13compareHelperEPKvS3_hhjPKsj($this,$89,$90,$55,$91,$60,$46,0)|0);
  $$0 = $92;
  return ($$0|0);
  break;
 }
 case 122: case 68: case 104: case 115:  {
  if ($2) {
   $93 = ($55<<24>>24)!=(42);
   $94 = ($60|0)==(0);
   $or$cond20 = $93 & $94;
   if ($or$cond20) {
    $95 = ((($this)) + 8|0);
    $99 = $95;
   } else {
    $99 = 0;
   }
  } else {
   $99 = $1;
  }
  if ($35) {
   $96 = ($55<<24>>24)!=(42);
   $97 = ($60|0)==(0);
   $or$cond16 = $96 & $97;
   if ($or$cond16) {
    $98 = ((($toCompare)) + 8|0);
    $100 = $98;
   } else {
    $100 = 0;
   }
  } else {
   $100 = $34;
  }
  $101 = (__ZNK9ExitGames6Common6Object11equalsArrayEPKvS3_j($this,$99,$100,0)|0);
  $$0 = $101;
  return ($$0|0);
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common6Object13compareHelperEPKvS3_hhjPKsj($this,$pData1,$pData2,$type,$customType,$dimensions,$arraySizes,$recursionDepth) {
 $this = $this|0;
 $pData1 = $pData1|0;
 $pData2 = $pData2|0;
 $type = $type|0;
 $customType = $customType|0;
 $dimensions = $dimensions|0;
 $arraySizes = $arraySizes|0;
 $recursionDepth = $recursionDepth|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0.0, $6 = 0, $60 = 0, $61 = 0.0;
 var $62 = 0, $63 = 0, $64 = 0.0, $65 = 0, $66 = 0.0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $i$05 = 0, $i1$08 = 0, $i1$08$us = 0, $or$cond = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ($dimensions|0)!=(0);
 $2 = (($dimensions) + -1)|0;
 $3 = ($2>>>0)>($recursionDepth>>>0);
 $or$cond = $1 & $3;
 if ($or$cond) {
  $18 = (($arraySizes) + ($recursionDepth<<1)|0);
  $19 = HEAP16[$18>>1]|0;
  $20 = ($19<<16>>16)>(0);
  if (!($20)) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  $21 = (($recursionDepth) + 1)|0;
  $i$05 = 0;
  while(1) {
   $26 = (($pData1) + ($i$05<<2)|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = (($pData2) + ($i$05<<2)|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (__ZNK9ExitGames6Common6Object13compareHelperEPKvS3_hhjPKsj($this,$27,$29,$type,$customType,$dimensions,$arraySizes,$21)|0);
   $25 = (($i$05) + 1)|0;
   if (!($30)) {
    $$0 = 0;
    label = 24;
    break;
   }
   $22 = HEAP16[$18>>1]|0;
   $23 = $22 << 16 >> 16;
   $24 = ($25|0)<($23|0);
   if ($24) {
    $i$05 = $25;
   } else {
    $$0 = 1;
    label = 24;
    break;
   }
  }
  if ((label|0) == 24) {
   STACKTOP = sp;return ($$0|0);
  }
 }
 $4 = $1 ? $2 : 0;
 $5 = (($arraySizes) + ($4<<1)|0);
 $6 = HEAP16[$5>>1]|0;
 $7 = ($6<<16>>16)>(0);
 if (!($7)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = $type&255;
 $9 = ($type<<24>>24)==(98);
 if ($9) {
  $10 = $6 << 16 >> 16;
  $i1$08$us = 0;
  while(1) {
   $11 = (($pData1) + ($i1$08$us)|0);
   $12 = HEAP8[$11>>0]|0;
   $13 = (($pData2) + ($i1$08$us)|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = ($12<<24>>24)==($14<<24>>24);
   if (!($15)) {
    $$0 = 0;
    label = 24;
    break;
   }
   $16 = (($i1$08$us) + 1)|0;
   $17 = ($16|0)<($10|0);
   if ($17) {
    $i1$08$us = $16;
   } else {
    $$0 = 1;
    label = 24;
    break;
   }
  }
  if ((label|0) == 24) {
   STACKTOP = sp;return ($$0|0);
  }
 } else {
  $i1$08 = 0;
 }
 L24: while(1) {
  switch ($8|0) {
  case 99:  {
   $73 = (($pData1) + ($i1$08<<2)|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ((($74)) + 20|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = (($pData2) + ($i1$08<<2)|0);
   $78 = (FUNCTION_TABLE_iii[$76 & 2047]($73,$77)|0);
   if (!($78)) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 107:  {
   $31 = (($pData1) + ($i1$08<<1)|0);
   $32 = HEAP16[$31>>1]|0;
   $33 = (($pData2) + ($i1$08<<1)|0);
   $34 = HEAP16[$33>>1]|0;
   $35 = ($32<<16>>16)==($34<<16>>16);
   if (!($35)) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 105:  {
   $36 = (($pData1) + ($i1$08<<2)|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = (($pData2) + ($i1$08<<2)|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($37|0)==($39|0);
   if (!($40)) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 108:  {
   $41 = (($pData1) + ($i1$08<<3)|0);
   $42 = $41;
   $43 = $42;
   $44 = HEAP32[$43>>2]|0;
   $45 = (($42) + 4)|0;
   $46 = $45;
   $47 = HEAP32[$46>>2]|0;
   $48 = (($pData2) + ($i1$08<<3)|0);
   $49 = $48;
   $50 = $49;
   $51 = HEAP32[$50>>2]|0;
   $52 = (($49) + 4)|0;
   $53 = $52;
   $54 = HEAP32[$53>>2]|0;
   $55 = ($44|0)==($51|0);
   $56 = ($47|0)==($54|0);
   $57 = $55 & $56;
   if (!($57)) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 102:  {
   $58 = (($pData1) + ($i1$08<<2)|0);
   $59 = +HEAPF32[$58>>2];
   $60 = (($pData2) + ($i1$08<<2)|0);
   $61 = +HEAPF32[$60>>2];
   $62 = $59 != $61;
   if ($62) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 100:  {
   $63 = (($pData1) + ($i1$08<<3)|0);
   $64 = +HEAPF64[$63>>3];
   $65 = (($pData2) + ($i1$08<<3)|0);
   $66 = +HEAPF64[$65>>3];
   $67 = $64 != $66;
   if ($67) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  case 111:  {
   $68 = (($pData1) + ($i1$08)|0);
   $69 = HEAP8[$68>>0]|0;
   $70 = (($pData2) + ($i1$08)|0);
   $71 = HEAP8[$70>>0]|0;
   $72 = ($69<<24>>24)==($71<<24>>24);
   if (!($72)) {
    $$0 = 0;
    label = 24;
    break L24;
   }
   break;
  }
  default: {
   __ZN9ExitGames6Common7JStringC2EPKc($0,146136);
   __THREW__ = 0;
   $79 = (invoke_ii(764,($0|0))|0);
   $80 = __THREW__; __THREW__ = 0;
   $81 = $80&1;
   if ($81) {
    label = 22;
    break L24;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(5244|0),($79|0),1,786,(5304|0),($vararg_buffer|0));
   $82 = __THREW__; __THREW__ = 0;
   $83 = $82&1;
   if ($83) {
    label = 22;
    break L24;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  }
  $86 = (($i1$08) + 1)|0;
  $87 = HEAP16[$5>>1]|0;
  $88 = $87 << 16 >> 16;
  $89 = ($86|0)<($88|0);
  if ($89) {
   $i1$08 = $86;
  } else {
   $$0 = 1;
   label = 24;
   break;
  }
 }
 if ((label|0) == 22) {
  $84 = ___cxa_find_matching_catch_2()|0;
  $85 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($84|0);
  // unreachable;
 }
 else if ((label|0) == 24) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common6Object11equalsArrayEPKvS3_j($this,$pData1,$pData2,$recursionDepth) {
 $this = $this|0;
 $pData1 = $pData1|0;
 $pData2 = $pData2|0;
 $recursionDepth = $recursionDepth|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $i$014 = 0;
 var $i1$017 = 0, $i2$022 = 0, $i3$027 = 0, $i4$032 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)!=(0);
 $3 = (($1) + -1)|0;
 $4 = ($3>>>0)>($recursionDepth>>>0);
 $or$cond = $2 & $4;
 if ($or$cond) {
  $5 = ((($this)) + 24|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0|0);
  $8 = ((($this)) + 20|0);
  $9 = $7 ? $8 : $6;
  $10 = (($9) + ($recursionDepth<<1)|0);
  $11 = HEAP16[$10>>1]|0;
  $12 = ($11<<16>>16)>(0);
  if (!($12)) {
   $$0 = 1;
   return ($$0|0);
  }
  $13 = (($recursionDepth) + 1)|0;
  $i$014 = 0;
  while(1) {
   $22 = (($pData1) + ($i$014<<2)|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = (($pData2) + ($i$014<<2)|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (__ZNK9ExitGames6Common6Object11equalsArrayEPKvS3_j($this,$23,$25,$13)|0);
   $21 = (($i$014) + 1)|0;
   if (!($26)) {
    $$0 = 0;
    label = 23;
    break;
   }
   $14 = HEAP32[$5>>2]|0;
   $15 = ($14|0)==(0|0);
   $16 = $15 ? $8 : $14;
   $17 = (($16) + ($recursionDepth<<1)|0);
   $18 = HEAP16[$17>>1]|0;
   $19 = $18 << 16 >> 16;
   $20 = ($21|0)<($19|0);
   if ($20) {
    $i$014 = $21;
   } else {
    $$0 = 1;
    label = 23;
    break;
   }
  }
  if ((label|0) == 23) {
   return ($$0|0);
  }
 }
 $27 = ((($this)) + 28|0);
 $28 = HEAP8[$27>>0]|0;
 $29 = $28&255;
 switch ($29|0) {
 case 115:  {
  $51 = ((($this)) + 24|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = ($52|0)==(0|0);
  $54 = ((($this)) + 20|0);
  $55 = $53 ? $54 : $52;
  $56 = HEAP16[$55>>1]|0;
  $57 = ($56<<16>>16)>(0);
  if (!($57)) {
   $$0 = 1;
   return ($$0|0);
  }
  $i1$017 = 0;
  while(1) {
   $65 = (($pData1) + ($i1$017<<4)|0);
   $66 = (($pData2) + ($i1$017<<4)|0);
   $67 = (__ZNK9ExitGames6Common7JStringneERKS1_($65,$66)|0);
   $64 = (($i1$017) + 1)|0;
   if ($67) {
    $$0 = 0;
    label = 23;
    break;
   }
   $58 = HEAP32[$51>>2]|0;
   $59 = ($58|0)==(0|0);
   $60 = $59 ? $54 : $58;
   $61 = HEAP16[$60>>1]|0;
   $62 = $61 << 16 >> 16;
   $63 = ($64|0)<($62|0);
   if ($63) {
    $i1$017 = $64;
   } else {
    $$0 = 1;
    label = 23;
    break;
   }
  }
  if ((label|0) == 23) {
   return ($$0|0);
  }
  break;
 }
 case 104:  {
  $44 = ((($this)) + 24|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = ($45|0)==(0|0);
  $47 = ((($this)) + 20|0);
  $48 = $46 ? $47 : $45;
  $49 = HEAP16[$48>>1]|0;
  $50 = ($49<<16>>16)>(0);
  if (!($50)) {
   $$0 = 1;
   return ($$0|0);
  }
  $i2$022 = 0;
  while(1) {
   $75 = (($pData1) + (($i2$022*44)|0)|0);
   $76 = (($pData2) + (($i2$022*44)|0)|0);
   $77 = (__ZNK9ExitGames6Common9HashtableneERKS1_($75,$76)|0);
   $74 = (($i2$022) + 1)|0;
   if ($77) {
    $$0 = 0;
    label = 23;
    break;
   }
   $68 = HEAP32[$44>>2]|0;
   $69 = ($68|0)==(0|0);
   $70 = $69 ? $47 : $68;
   $71 = HEAP16[$70>>1]|0;
   $72 = $71 << 16 >> 16;
   $73 = ($74|0)<($72|0);
   if ($73) {
    $i2$022 = $74;
   } else {
    $$0 = 1;
    label = 23;
    break;
   }
  }
  if ((label|0) == 23) {
   return ($$0|0);
  }
  break;
 }
 case 68:  {
  $37 = ((($this)) + 24|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ($38|0)==(0|0);
  $40 = ((($this)) + 20|0);
  $41 = $39 ? $40 : $38;
  $42 = HEAP16[$41>>1]|0;
  $43 = ($42<<16>>16)>(0);
  if (!($43)) {
   $$0 = 1;
   return ($$0|0);
  }
  $i3$027 = 0;
  while(1) {
   $85 = (($pData1) + (($i3$027*52)|0)|0);
   $86 = (($pData2) + (($i3$027*52)|0)|0);
   $87 = (__ZNK9ExitGames6Common14DictionaryBaseneERKS1_($85,$86)|0);
   $84 = (($i3$027) + 1)|0;
   if ($87) {
    $$0 = 0;
    label = 23;
    break;
   }
   $78 = HEAP32[$37>>2]|0;
   $79 = ($78|0)==(0|0);
   $80 = $79 ? $40 : $78;
   $81 = HEAP16[$80>>1]|0;
   $82 = $81 << 16 >> 16;
   $83 = ($84|0)<($82|0);
   if ($83) {
    $i3$027 = $84;
   } else {
    $$0 = 1;
    label = 23;
    break;
   }
  }
  if ((label|0) == 23) {
   return ($$0|0);
  }
  break;
 }
 case 122:  {
  $30 = ((($this)) + 24|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ($31|0)==(0|0);
  $33 = ((($this)) + 20|0);
  $34 = $32 ? $33 : $31;
  $35 = HEAP16[$34>>1]|0;
  $36 = ($35<<16>>16)>(0);
  if (!($36)) {
   $$0 = 1;
   return ($$0|0);
  }
  $i4$032 = 0;
  while(1) {
   $95 = (($pData1) + (($i4$032*40)|0)|0);
   $96 = (($pData2) + (($i4$032*40)|0)|0);
   $97 = (__ZNK9ExitGames6Common6ObjecteqERKS1_($95,$96)|0);
   $94 = (($i4$032) + 1)|0;
   if (!($97)) {
    $$0 = 0;
    label = 23;
    break;
   }
   $88 = HEAP32[$30>>2]|0;
   $89 = ($88|0)==(0|0);
   $90 = $89 ? $33 : $88;
   $91 = HEAP16[$90>>1]|0;
   $92 = $91 << 16 >> 16;
   $93 = ($94|0)<($92|0);
   if ($93) {
    $i4$032 = $94;
   } else {
    $$0 = 1;
    label = 23;
    break;
   }
  }
  if ((label|0) == 23) {
   return ($$0|0);
  }
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common7JStringneERKS1_($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($Rhs)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  $11 = 1;
  return ($11|0);
 }
 $5 = ((($this)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($Rhs)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_wcscmp($6,$8)|0);
 $10 = ($9|0)!=(0);
 $11 = $10;
 return ($11|0);
}
function __ZNK9ExitGames6Common9HashtableneERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common9HashtableeqERKS1_($this,$toCompare)|0);
 $1 = $0 ^ 1;
 return ($1|0);
}
function __ZNK9ExitGames6Common14DictionaryBaseneERKS1_($this,$toCompare) {
 $this = $this|0;
 $toCompare = $toCompare|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 32|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 2047]($this,$toCompare)|0);
 $4 = $3 ^ 1;
 return ($4|0);
}
function __ZN9ExitGames6Common7Helpers11KeyToObject3getENS0_6ObjectE($agg$result,$key) {
 $agg$result = $agg$result|0;
 $key = $key|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common6ObjectC2ERKS1_($agg$result,$key);
 return;
}
function __ZNK9ExitGames6Common9Hashtable22getValueImplementationERKNS0_6ObjectE($this,$key) {
 $this = $this|0;
 $key = $key|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, $i$01$i$lcssa = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $25 = 0;
  STACKTOP = sp;return ($25|0);
 }
 $4 = ((($this)) + 20|0);
 $i$01$i = 0;
 while(1) {
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + (($i$01$i*40)|0)|0);
  $7 = (__ZNK9ExitGames6Common6ObjecteqERKS1_($6,$key)|0);
  if ($7) {
   $i$01$i$lcssa = $i$01$i;
   break;
  }
  $8 = (($i$01$i) + 1)|0;
  $9 = HEAP32[$1>>2]|0;
  $10 = ($8>>>0)<($9>>>0);
  if ($10) {
   $i$01$i = $8;
  } else {
   $25 = 0;
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  STACKTOP = sp;return ($25|0);
 }
 $11 = ($i$01$i$lcssa|0)==(-1);
 if ($11) {
  $25 = 0;
  STACKTOP = sp;return ($25|0);
 }
 $12 = ((($this)) + 28|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13>>>0)>($i$01$i$lcssa>>>0);
 do {
  if (!($14)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $15 = (invoke_ii(764,($0|0))|0);
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if ($17) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($20|0);
    // unreachable;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($15|0),1,566,(10968|0),($vararg_buffer|0));
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($20|0);
    // unreachable;
   } else {
    __ZN9ExitGames6Common7JStringD2Ev($0);
    break;
   }
  }
 } while(0);
 $22 = ((($this)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = (($23) + (($i$01$i$lcssa*40)|0)|0);
 $25 = $24;
 STACKTOP = sp;return ($25|0);
}
function __ZN9ExitGames6Common6ObjectC2ERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $exitcond$i$i = 0, $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4640);
 $0 = ((($toCopy)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ((($toCopy)) + 28|0);
 $4 = HEAP8[$3>>0]|0;
 if ($2) {
  $5 = ($4<<24>>24)==(42);
  if ($5) {
   $11 = 42;$25 = 0;
  } else {
   $6 = ((($toCopy)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0);
   if ($8) {
    $9 = ((($toCopy)) + 8|0);
    $11 = $4;$25 = $9;
   } else {
    $11 = $4;$25 = 0;
   }
  }
 } else {
  $11 = $4;$25 = $1;
 }
 $10 = ($11<<24>>24)==(99);
 $12 = ((($toCopy)) + 29|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = $10 ? $13 : 0;
 $15 = ((($toCopy)) + 32|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($toCopy)) + 24|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ($18|0)==(0|0);
 $20 = ((($toCopy)) + 20|0);
 $21 = $19 ? $20 : $18;
 $22 = ($11<<24>>24)==(42);
 do {
  if (!($22)) {
   $23 = ($21|0)==(0|0);
   if (!($23)) {
    $24 = ($25|0)==(0|0);
    if (!($24)) {
     break;
    }
    $26 = HEAP16[$21>>1]|0;
    $27 = ($26<<16>>16)==(0);
    if ($27) {
     break;
    }
   }
   $28 = ((($this)) + 28|0);
   HEAP8[$28>>0] = 42;
   $29 = ((($this)) + 29|0);
   HEAP8[$29>>0] = 0;
   $30 = ((($this)) + 32|0);
   HEAP32[$30>>2] = 0;
   $31 = ((($this)) + 16|0);
   ;HEAP32[$31>>2]=0|0;HEAP32[$31+4>>2]=0|0;HEAP32[$31+8>>2]=0|0;
   return;
  }
 } while(0);
 $32 = ((($this)) + 28|0);
 HEAP8[$32>>0] = $11;
 $33 = ((($this)) + 29|0);
 HEAP8[$33>>0] = $14;
 $34 = ((($this)) + 24|0);
 HEAP32[$34>>2] = 0;
 $35 = ($21|0)==(0|0);
 $36 = ((($this)) + 32|0);
 do {
  if ($35) {
   HEAP32[$36>>2] = 0;
   $54 = ((($this)) + 20|0);
   HEAP32[$54>>2] = 0;
   HEAP32[$34>>2] = 0;
  } else {
   HEAP32[$36>>2] = $16;
   $37 = ($16>>>0)<(2);
   if ($37) {
    $38 = HEAP32[$21>>2]|0;
    $39 = ((($this)) + 20|0);
    HEAP32[$39>>2] = $38;
    break;
   }
   $40 = $16 << 1;
   $41 = (($40) + 8)|0;
   __THREW__ = 0;
   $42 = (invoke_ii(775,($41|0))|0);
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if ($44) {
    $58 = ___cxa_find_matching_catch_2()|0;
    $59 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($this);
    ___resumeException($58|0);
    // unreachable;
   }
   $45 = ((($42)) + 8|0);
   $46 = $42;
   $47 = $46;
   HEAP32[$47>>2] = $16;
   $48 = (($46) + 4)|0;
   $49 = $48;
   HEAP32[$49>>2] = 0;
   _memset(($45|0),0,($40|0))|0;
   HEAP32[$34>>2] = $45;
   $i$01$i$i = 0;
   while(1) {
    $50 = (($21) + ($i$01$i$i<<1)|0);
    $51 = HEAP16[$50>>1]|0;
    $52 = (($45) + ($i$01$i$i<<1)|0);
    HEAP16[$52>>1] = $51;
    $53 = (($i$01$i$i) + 1)|0;
    $exitcond$i$i = ($53|0)==($16|0);
    if ($exitcond$i$i) {
     break;
    } else {
     $i$01$i$i = $53;
    }
   }
  }
 } while(0);
 $55 = ((($this)) + 16|0);
 HEAP32[$55>>2] = 0;
 __THREW__ = 0;
 invoke_vii(830,($this|0),($25|0));
 $56 = __THREW__; __THREW__ = 0;
 $57 = $56&1;
 if ($57) {
  $58 = ___cxa_find_matching_catch_2()|0;
  $59 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($58|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNK9ExitGames6Common9Hashtable7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common9Hashtable7getKeysEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZNK9ExitGames6Common9HashtableixEj($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2>>>0)>($index>>>0);
 if ($3) {
  $11 = ((($this)) + 40|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12) + (($index*40)|0)|0);
  STACKTOP = sp;return ($13|0);
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
 __THREW__ = 0;
 $4 = (invoke_ii(764,($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,(162472|0),1,(10856|0),($4|0),1,566,(10968|0),($vararg_buffer|0));
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $11 = ((($this)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) + (($index*40)|0)|0);
 STACKTOP = sp;return ($13|0);
}
function __ZN9ExitGames6Common7JStringaSEPKc($this,$Rhs) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 var $$0 = 0, $$01 = 0, $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $0 = sp + 16|0;
 $1 = sp;
 __ZN9ExitGames6Common10UTF8StringC2EPKc($1,$Rhs);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 __THREW__ = 0;
 invoke_vii($4|0,($0|0),($1|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $$0 = $42;$$01 = $41;
  __ZN9ExitGames6Common10UTF8StringD2Ev($1);
  ___resumeException($$01|0);
  // unreachable;
 }
 $7 = ($0|0)==($this|0);
 do {
  if ($7) {
   $$pre = ((($0)) + 4|0);
   $$pre$phiZ2D = $$pre;
  } else {
   $8 = ((($this)) + 4|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($9|0)==(0|0);
   if ($10) {
    label = 6;
   } else {
    $11 = ((($9)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($11|0));
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     label = 6;
    }
   }
   if ((label|0) == 6) {
    $14 = ((($0)) + 12|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($this)) + 12|0);
    HEAP32[$16>>2] = $15;
    $17 = ((($this)) + 8|0);
    HEAP32[$17>>2] = $15;
    $18 = (($15) + 1)|0;
    $19 = $18 << 2;
    $20 = (($19) + 8)|0;
    __THREW__ = 0;
    $21 = (invoke_ii(775,($20|0))|0);
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if (!($23)) {
     $24 = ((($21)) + 8|0);
     $25 = $21;
     $26 = $25;
     HEAP32[$26>>2] = $18;
     $27 = (($25) + 4)|0;
     $28 = $27;
     HEAP32[$28>>2] = 0;
     $29 = ($18|0)==(0);
     if (!($29)) {
      _memset(($24|0),0,($19|0))|0;
     }
     HEAP32[$8>>2] = $24;
     $30 = ((($0)) + 4|0);
     $31 = HEAP32[$30>>2]|0;
     __THREW__ = 0;
     (invoke_iii(789,($24|0),($31|0))|0);
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if (!($33)) {
      $$pre$phiZ2D = $30;
      break;
     }
    }
   }
   $43 = ___cxa_find_matching_catch_2()|0;
   $44 = tempRet0;
   HEAP32[$0>>2] = (3908);
   $45 = ((($0)) + 4|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($46|0)==(0|0);
   if (!($47)) {
    $48 = ((($46)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($48|0));
    $49 = __THREW__; __THREW__ = 0;
    $50 = $49&1;
    if ($50) {
     $51 = ___cxa_find_matching_catch_3(0|0)|0;
     $52 = tempRet0;
     ___clang_call_terminate($51);
     // unreachable;
    }
   }
   $$0 = $44;$$01 = $43;
   __ZN9ExitGames6Common10UTF8StringD2Ev($1);
   ___resumeException($$01|0);
   // unreachable;
  }
 } while(0);
 HEAP32[$0>>2] = (3908);
 $34 = HEAP32[$$pre$phiZ2D>>2]|0;
 $35 = ($34|0)==(0|0);
 if ($35) {
  __ZN9ExitGames6Common10UTF8StringD2Ev($1);
  STACKTOP = sp;return ($this|0);
 }
 $36 = ((($34)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($36|0));
 $37 = __THREW__; __THREW__ = 0;
 $38 = $37&1;
 if ($38) {
  $39 = ___cxa_find_matching_catch_3(0|0)|0;
  $40 = tempRet0;
  ___clang_call_terminate($39);
  // unreachable;
 } else {
  __ZN9ExitGames6Common10UTF8StringD2Ev($1);
  STACKTOP = sp;return ($this|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common14DictionaryBaseC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5516);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vi(838,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($4|0);
  // unreachable;
 } else {
  $3 = ((($this)) + 48|0);
  HEAP32[$3>>2] = 0;
  return;
 }
}
function __ZN9ExitGames6Common14CustomTypeBase11deserializeEPhsPvh($pData,$length,$retVal,$typeCode) {
 $pData = $pData|0;
 $length = $length|0;
 $retVal = $retVal|0;
 $typeCode = $typeCode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $12 = HEAP32[$3>>2]|0;
  $13 = ((($12)) + 20|0);
  $14 = HEAP32[$13>>2]|0;
  FUNCTION_TABLE_viiii[$14 & 2047]($3,$pData,$length,$retVal);
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common6Object17setWithoutCleanupEPKvhhjPKsb($this,$data,$type,$customType,$dimensions,$sizes,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $dimensions = $dimensions|0;
 $sizes = $sizes|0;
 $makeCopy = $makeCopy|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $9 = 0, $exitcond$i = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($type<<24>>24)==(42);
 do {
  if (!($0)) {
   $1 = ($sizes|0)==(0|0);
   if (!($1)) {
    $2 = ($data|0)==(0|0);
    if (!($2)) {
     break;
    }
    $3 = HEAP16[$sizes>>1]|0;
    $4 = ($3<<16>>16)==(0);
    if ($4) {
     break;
    }
   }
   $5 = ((($this)) + 28|0);
   HEAP8[$5>>0] = 42;
   $6 = ((($this)) + 29|0);
   HEAP8[$6>>0] = 0;
   $7 = ((($this)) + 32|0);
   HEAP32[$7>>2] = 0;
   $8 = ((($this)) + 16|0);
   ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;
   return;
  }
 } while(0);
 $9 = ((($this)) + 28|0);
 HEAP8[$9>>0] = $type;
 $10 = ((($this)) + 29|0);
 HEAP8[$10>>0] = $customType;
 $11 = ((($this)) + 24|0);
 HEAP32[$11>>2] = 0;
 $12 = ($sizes|0)==(0|0);
 $13 = ((($this)) + 32|0);
 do {
  if ($12) {
   HEAP32[$13>>2] = 0;
   $29 = ((($this)) + 20|0);
   HEAP32[$29>>2] = 0;
   HEAP32[$11>>2] = 0;
  } else {
   HEAP32[$13>>2] = $dimensions;
   $14 = ($dimensions>>>0)<(2);
   if ($14) {
    $15 = HEAP32[$sizes>>2]|0;
    $16 = ((($this)) + 20|0);
    HEAP32[$16>>2] = $15;
    break;
   }
   $17 = $dimensions << 1;
   $18 = (($17) + 8)|0;
   $19 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($18)|0);
   $20 = ((($19)) + 8|0);
   $21 = $19;
   $22 = $21;
   HEAP32[$22>>2] = $dimensions;
   $23 = (($21) + 4)|0;
   $24 = $23;
   HEAP32[$24>>2] = 0;
   _memset(($20|0),0,($17|0))|0;
   HEAP32[$11>>2] = $20;
   $i$01$i = 0;
   while(1) {
    $25 = (($sizes) + ($i$01$i<<1)|0);
    $26 = HEAP16[$25>>1]|0;
    $27 = (($20) + ($i$01$i<<1)|0);
    HEAP16[$27>>1] = $26;
    $28 = (($i$01$i) + 1)|0;
    $exitcond$i = ($28|0)==($dimensions|0);
    if ($exitcond$i) {
     break;
    } else {
     $i$01$i = $28;
    }
   }
  }
 } while(0);
 $30 = ((($this)) + 16|0);
 HEAP32[$30>>2] = 0;
 if ($makeCopy) {
  __ZN9ExitGames6Common6Object7setDataEPKv($this,$data);
  return;
 }
 $31 = ($data|0)==(0|0);
 if ($31) {
  HEAP32[$30>>2] = 0;
  return;
 }
 $32 = HEAP32[$13>>2]|0;
 $33 = ($32|0)==(0);
 if (!($33)) {
  HEAP32[$30>>2] = $data;
  return;
 }
 $34 = HEAP8[$9>>0]|0;
 $35 = $34&255;
 switch ($35|0) {
 case 98:  {
  $36 = HEAP8[$data>>0]|0;
  $37 = ((($this)) + 8|0);
  $38 = $36&255;
  $39 = $37;
  $40 = $39;
  HEAP32[$40>>2] = $38;
  $41 = (($39) + 4)|0;
  $42 = $41;
  HEAP32[$42>>2] = 0;
  return;
  break;
 }
 case 107:  {
  $43 = HEAP16[$data>>1]|0;
  $44 = ((($this)) + 8|0);
  $45 = $43&65535;
  $46 = $44;
  $47 = $46;
  HEAP32[$47>>2] = $45;
  $48 = (($46) + 4)|0;
  $49 = $48;
  HEAP32[$49>>2] = 0;
  return;
  break;
 }
 case 105:  {
  $50 = HEAP32[$data>>2]|0;
  $51 = ((($this)) + 8|0);
  $52 = $51;
  $53 = $52;
  HEAP32[$53>>2] = $50;
  $54 = (($52) + 4)|0;
  $55 = $54;
  HEAP32[$55>>2] = 0;
  return;
  break;
 }
 case 108:  {
  $56 = $data;
  $57 = $56;
  $58 = HEAP32[$57>>2]|0;
  $59 = (($56) + 4)|0;
  $60 = $59;
  $61 = HEAP32[$60>>2]|0;
  $62 = ((($this)) + 8|0);
  $63 = $62;
  $64 = $63;
  HEAP32[$64>>2] = $58;
  $65 = (($63) + 4)|0;
  $66 = $65;
  HEAP32[$66>>2] = $61;
  return;
  break;
 }
 case 102:  {
  $67 = HEAP32[$data>>2]|0;
  $68 = ((($this)) + 8|0);
  HEAP32[$68>>2] = $67;
  return;
  break;
 }
 case 100:  {
  $69 = $data;
  $70 = $69;
  $71 = HEAP32[$70>>2]|0;
  $72 = (($69) + 4)|0;
  $73 = $72;
  $74 = HEAP32[$73>>2]|0;
  $75 = ((($this)) + 8|0);
  $76 = $75;
  $77 = $76;
  HEAP32[$77>>2] = $71;
  $78 = (($76) + 4)|0;
  $79 = $78;
  HEAP32[$79>>2] = $74;
  return;
  break;
 }
 case 111:  {
  $80 = HEAP8[$data>>0]|0;
  $81 = ((($this)) + 8|0);
  $82 = $80&255;
  $83 = $81;
  $84 = $83;
  HEAP32[$84>>2] = $82;
  $85 = (($83) + 4)|0;
  $86 = $85;
  HEAP32[$86>>2] = 0;
  return;
  break;
 }
 case 99: case 68: case 122: case 104: case 115:  {
  HEAP32[$30>>2] = $data;
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function __ZNK9ExitGames6Common6Object7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 16|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $10 = $1;
  return ($10|0);
 }
 $3 = ((($this)) + 28|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(42);
 if ($5) {
  $10 = 0;
  return ($10|0);
 }
 $6 = ((($this)) + 32|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0);
 if (!($8)) {
  $10 = 0;
  return ($10|0);
 }
 $9 = ((($this)) + 8|0);
 $10 = $9;
 return ($10|0);
}
function __ZNK9ExitGames6Common6Object7getTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0);
 $1 = HEAP8[$0>>0]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common6Object13getDimensionsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common6Object8getSizesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 $3 = ((($this)) + 20|0);
 $4 = $2 ? $3 : $1;
 return ($4|0);
}
function __ZN9ExitGames6Common6Object3setEPKvhhib($this,$data,$type,$customType,$size,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $size = $size|0;
 $makeCopy = $makeCopy|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$0>>2] = $size;
 __ZN9ExitGames6Common6Object7cleanupEPKvj($this,0,0);
 __ZN9ExitGames6Common6Object17setWithoutCleanupEPKvhhjPKsb($this,$data,$type,$customType,1,$0,$makeCopy);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common6Object13setDataNoCopyEPv($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($data|0)==(0|0);
 if ($0) {
  $1 = ((($this)) + 16|0);
  HEAP32[$1>>2] = 0;
  return;
 }
 $2 = ((($this)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if (!($4)) {
  $5 = ((($this)) + 16|0);
  HEAP32[$5>>2] = $data;
  return;
 }
 $6 = ((($this)) + 28|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = $7&255;
 switch ($8|0) {
 case 98:  {
  $9 = HEAP8[$data>>0]|0;
  $10 = ((($this)) + 8|0);
  $11 = $9&255;
  $12 = $10;
  $13 = $12;
  HEAP32[$13>>2] = $11;
  $14 = (($12) + 4)|0;
  $15 = $14;
  HEAP32[$15>>2] = 0;
  return;
  break;
 }
 case 107:  {
  $16 = HEAP16[$data>>1]|0;
  $17 = ((($this)) + 8|0);
  $18 = $16&65535;
  $19 = $17;
  $20 = $19;
  HEAP32[$20>>2] = $18;
  $21 = (($19) + 4)|0;
  $22 = $21;
  HEAP32[$22>>2] = 0;
  return;
  break;
 }
 case 105:  {
  $23 = HEAP32[$data>>2]|0;
  $24 = ((($this)) + 8|0);
  $25 = $24;
  $26 = $25;
  HEAP32[$26>>2] = $23;
  $27 = (($25) + 4)|0;
  $28 = $27;
  HEAP32[$28>>2] = 0;
  return;
  break;
 }
 case 108:  {
  $29 = $data;
  $30 = $29;
  $31 = HEAP32[$30>>2]|0;
  $32 = (($29) + 4)|0;
  $33 = $32;
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($this)) + 8|0);
  $36 = $35;
  $37 = $36;
  HEAP32[$37>>2] = $31;
  $38 = (($36) + 4)|0;
  $39 = $38;
  HEAP32[$39>>2] = $34;
  return;
  break;
 }
 case 102:  {
  $40 = HEAP32[$data>>2]|0;
  $41 = ((($this)) + 8|0);
  HEAP32[$41>>2] = $40;
  return;
  break;
 }
 case 100:  {
  $42 = $data;
  $43 = $42;
  $44 = HEAP32[$43>>2]|0;
  $45 = (($42) + 4)|0;
  $46 = $45;
  $47 = HEAP32[$46>>2]|0;
  $48 = ((($this)) + 8|0);
  $49 = $48;
  $50 = $49;
  HEAP32[$50>>2] = $44;
  $51 = (($49) + 4)|0;
  $52 = $51;
  HEAP32[$52>>2] = $47;
  return;
  break;
 }
 case 111:  {
  $53 = HEAP8[$data>>0]|0;
  $54 = ((($this)) + 8|0);
  $55 = $53&255;
  $56 = $54;
  $57 = $56;
  HEAP32[$57>>2] = $55;
  $58 = (($56) + 4)|0;
  $59 = $58;
  HEAP32[$59>>2] = 0;
  return;
  break;
 }
 case 99: case 68: case 122: case 104: case 115:  {
  $60 = ((($this)) + 16|0);
  HEAP32[$60>>2] = $data;
  return;
  break;
 }
 default: {
  return;
 }
 }
}
function __ZN9ExitGames6Common6Object7setTypeEh($this,$type) {
 $this = $this|0;
 $type = $type|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0);
 HEAP8[$0>>0] = $type;
 return;
}
function __ZNK9ExitGames6Common6Object13getCustomTypeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(99);
 if (!($2)) {
  $5 = 0;
  return ($5|0);
 }
 $3 = ((($this)) + 29|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4;
 return ($5|0);
}
function __ZN9ExitGames6Common6ObjectC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (4640);
 $0 = ((($this)) + 28|0);
 HEAP8[$0>>0] = 42;
 $1 = ((($this)) + 29|0);
 HEAP8[$1>>0] = 0;
 $2 = ((($this)) + 32|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($this)) + 16|0);
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;
 return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementationD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6112);
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementationD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6112);
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $0 = sp + 40|0;
 $1 = sp;
 $2 = ((($this)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common6ObjectC2EPKvhhib($1,$3,98,0,$5,1);
 HEAP32[$1>>2] = (3332);
 __THREW__ = 0;
 invoke_viii(783,($0|0),($1|0),($withTypes|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $11 = ___cxa_find_matching_catch_2()|0;
  $12 = tempRet0;
  $$0 = $12;$$01 = $11;
  __ZN9ExitGames6Common6ObjectD2Ev($1);
  ___resumeException($$01|0);
  // unreachable;
 }
 __THREW__ = 0;
 $8 = (invoke_iii(760,($retStr|0),($0|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if (!($10)) {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  __ZN9ExitGames6Common6ObjectD2Ev($1);
  STACKTOP = sp;return ($8|0);
 }
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = $14;$$01 = $13;
 __ZN9ExitGames6Common6ObjectD2Ev($1);
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Common6ObjectC2EPKvhhib($this,$data,$type,$customType,$size,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $size = $size|0;
 $makeCopy = $makeCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$this>>2] = (4640);
 HEAP32[$0>>2] = $size;
 __THREW__ = 0;
 invoke_viiiiiii(855,($this|0),($data|0),($type|0),($customType|0),1,($0|0),($makeCopy|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($3|0);
  // unreachable;
 } else {
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common10SerializerD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6136);
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($0);
 __ZN9ExitGames6Common4BaseD2Ev($this);
 return;
}
function __ZN9ExitGames6Common10SerializerD0Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6136);
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($0);
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Common10Serializer8toStringERNS0_7JStringEb($this,$retStr,$withTypes) {
 $this = $this|0;
 $retStr = $retStr|0;
 $withTypes = $withTypes|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation8toStringERNS0_7JStringEb($0,$retStr,$withTypes)|0);
 return ($1|0);
}
function __ZN9ExitGames6Common8ToStringD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common10ANSIStringC2ERKNS0_7JStringE($this,$wstr) {
 $this = $this|0;
 $wstr = $wstr|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (3868);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  label = 3;
 } else {
  $3 = ((($1)) + -8|0);
  __THREW__ = 0;
  invoke_vi(750,($3|0));
  $4 = __THREW__; __THREW__ = 0;
  $5 = $4&1;
  if (!($5)) {
   label = 3;
  }
 }
 if ((label|0) == 3) {
  __THREW__ = 0;
  $6 = (invoke_ii(820,($wstr|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   $9 = ((($this)) + 8|0);
   HEAP32[$9>>2] = $6;
   $10 = (($6) + 1)|0;
   $11 = (($6) + 9)|0;
   __THREW__ = 0;
   $12 = (invoke_ii(775,($11|0))|0);
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if (!($14)) {
    $15 = ((($12)) + 8|0);
    $16 = $12;
    $17 = $16;
    HEAP32[$17>>2] = $10;
    $18 = (($16) + 4)|0;
    $19 = $18;
    HEAP32[$19>>2] = 0;
    $20 = ($10|0)==(0);
    if (!($20)) {
     _memset(($15|0),0,($10|0))|0;
    }
    HEAP32[$0>>2] = $15;
    __THREW__ = 0;
    $21 = (invoke_ii(764,($wstr|0))|0);
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if (!($23)) {
     $24 = HEAP32[$0>>2]|0;
     $25 = HEAP32[$9>>2]|0;
     $26 = (($25) + 1)|0;
     __THREW__ = 0;
     (invoke_iiii(863,($21|0),($24|0),($26|0))|0);
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if (!($28)) {
      return;
     }
    }
   }
  }
 }
 $29 = ___cxa_find_matching_catch_2()|0;
 $30 = tempRet0;
 __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
 ___resumeException($29|0);
 // unreachable;
}
function __ZN9ExitGames6Common10ANSIString13ANSIConverter12Unicode2ANSIEPKwPcj($wstr,$str,$strSize) {
 $wstr = $wstr|0;
 $str = $str|0;
 $strSize = $strSize|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $size = 0, $strptr = 0, $tmpStrSize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp + 12|0;
 $strptr = sp + 8|0;
 $size = sp + 4|0;
 $tmpStrSize = sp;
 HEAP32[$0>>2] = $wstr;
 $1 = HEAP32[984]|0;
 $2 = HEAP32[983]|0;
 $3 = (_iconv_open($1,$2)|0);
 HEAP32[$strptr>>2] = $str;
 $4 = $strSize << 2;
 HEAP32[$size>>2] = $4;
 HEAP32[$tmpStrSize>>2] = $strSize;
 (_iconv($3,$0,$size,$strptr,$tmpStrSize)|0);
 STACKTOP = sp;return ($str|0);
}
function __ZN9ExitGames6Common6Logger11setListenerERKNS0_12BaseListenerE($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $listener;
 return;
}
function __ZN9ExitGames6Common6Logger19setDebugOutputLevelEi($this,$debugLevel) {
 $this = $this|0;
 $debugLevel = $debugLevel|0;
 var $$0 = 0, $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($debugLevel|0)>(4);
 if ($0) {
  $$0 = 0;
 } else {
  HEAP32[$this>>2] = $debugLevel;
  $$0 = 1;
 }
 return ($$0|0);
}
function __ZN9ExitGames6Common14CustomTypeBase9serializeEPKvPhh($pData,$retVal,$typeCode) {
 $pData = $pData|0;
 $retVal = $retVal|0;
 $typeCode = $typeCode|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = $typeCode&255;
 $2 = (162720 + ($1<<2)|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $12 = HEAP32[$3>>2]|0;
  $13 = ((($12)) + 24|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (FUNCTION_TABLE_iiii[$14 & 2047]($3,$pData,$retVal)|0);
  $$0 = $15;
  STACKTOP = sp;return ($$0|0);
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,145892);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $1;
 invoke_viiiiiiii(765,(162472|0),1,(4664|0),($5|0),1,172,(4756|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($10|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementationC2EPKh($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6088);
 $0 = ((($this)) + 8|0);
 HEAP32[$0>>2] = $data;
 $1 = ((($this)) + 4|0);
 HEAP32[$1>>2] = -1;
 return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation3popERNS0_6ObjectE($this,$object) {
 $this = $this|0;
 $object = $object|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($2)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = (__ZN9ExitGames6Common7Helpers26DeSerializerImplementation7popTypeEhRNS0_6ObjectE($this,$6,$object)|0);
 return ($7|0);
}
function __ZN9ExitGames6Common14DictionaryBaseC2ERKS1_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (5516);
 $0 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_vi(838,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $24 = ___cxa_find_matching_catch_2()|0;
  $25 = tempRet0;
  $$0 = $25;$$01 = $24;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($$01|0);
  // unreachable;
 }
 $3 = ((($toCopy)) + 4|0);
 __THREW__ = 0;
 (invoke_iii(864,($0|0),($3|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   label = 9;
  } else {
   $6 = ((($toCopy)) + 48|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $23 = 0;
    $22 = ((($this)) + 48|0);
    HEAP32[$22>>2] = $23;
    return;
   }
   __THREW__ = 0;
   $9 = (invoke_ii(775,16)|0);
   $10 = __THREW__; __THREW__ = 0;
   $11 = $10&1;
   if ($11) {
    label = 9;
   } else {
    HEAP32[$9>>2] = (5564);
    $12 = ((($7)) + 4|0);
    $13 = HEAP32[$12>>2]|0;
    $14 = ((($7)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($7)) + 12|0);
    $17 = HEAP32[$16>>2]|0;
    __THREW__ = 0;
    invoke_viiii(856,($9|0),($13|0),($15|0),($17|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $20 = ___cxa_find_matching_catch_2()|0;
     $21 = tempRet0;
     __ZN9ExitGames6Common4BaseD2Ev($9);
     $eh$lpad$body$index2Z2D = $21;$eh$lpad$body$indexZ2D = $20;
     break;
    } else {
     $23 = $9;
     $22 = ((($this)) + 48|0);
     HEAP32[$22>>2] = $23;
     return;
    }
   }
  }
 } while(0);
 if ((label|0) == 9) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  $eh$lpad$body$index2Z2D = $27;$eh$lpad$body$indexZ2D = $26;
 }
 __ZN9ExitGames6Common9HashtableD2Ev($0);
 $$0 = $eh$lpad$body$index2Z2D;$$01 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Common9Hashtable17putImplementationERKNS0_6ObjectES4_($this,$key,$val) {
 $this = $this|0;
 $key = $key|0;
 $val = $val|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, $i$01$i$lcssa = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = (__ZNK9ExitGames6Common6Object7getTypeEv($key)|0);
 $3 = ($2<<24>>24)==(42);
 if ($3) {
  STACKTOP = sp;return;
 }
 $4 = ((($this)) + 4|0);
 $5 = ((($this)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 L4: do {
  if (!($7)) {
   $8 = ((($this)) + 20|0);
   $i$01$i = 0;
   while(1) {
    $9 = HEAP32[$8>>2]|0;
    $10 = (($9) + (($i$01$i*40)|0)|0);
    $11 = (__ZNK9ExitGames6Common6ObjecteqERKS1_($10,$key)|0);
    if ($11) {
     $i$01$i$lcssa = $i$01$i;
     break;
    }
    $12 = (($i$01$i) + 1)|0;
    $13 = HEAP32[$5>>2]|0;
    $14 = ($12>>>0)<($13>>>0);
    if ($14) {
     $i$01$i = $12;
    } else {
     break L4;
    }
   }
   $15 = ($i$01$i$lcssa|0)==(-1);
   if (!($15)) {
    $17 = HEAP32[$5>>2]|0;
    $18 = ($17>>>0)>($i$01$i$lcssa>>>0);
    do {
     if (!($18)) {
      __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
      __THREW__ = 0;
      $19 = (invoke_ii(764,($1|0))|0);
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if ($21) {
       $24 = ___cxa_find_matching_catch_2()|0;
       $25 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($1);
       ___resumeException($24|0);
       // unreachable;
      }
      __THREW__ = 0;
      invoke_viiiiiiii(765,(162472|0),1,(10856|0),($19|0),1,566,(10968|0),($vararg_buffer|0));
      $22 = __THREW__; __THREW__ = 0;
      $23 = $22&1;
      if ($23) {
       $24 = ___cxa_find_matching_catch_2()|0;
       $25 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($1);
       ___resumeException($24|0);
       // unreachable;
      } else {
       __ZN9ExitGames6Common7JStringD2Ev($1);
       break;
      }
     }
    } while(0);
    $26 = HEAP32[$8>>2]|0;
    $27 = (($26) + (($i$01$i$lcssa*40)|0)|0);
    $28 = HEAP32[$27>>2]|0;
    $29 = ((($28)) + 16|0);
    $30 = HEAP32[$29>>2]|0;
    (FUNCTION_TABLE_iii[$30 & 2047]($27,$key)|0);
    $31 = ((($this)) + 28|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = ($32>>>0)>($i$01$i$lcssa>>>0);
    do {
     if (!($33)) {
      __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
      __THREW__ = 0;
      $34 = (invoke_ii(764,($0|0))|0);
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       $39 = ___cxa_find_matching_catch_2()|0;
       $40 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($0);
       ___resumeException($39|0);
       // unreachable;
      }
      __THREW__ = 0;
      invoke_viiiiiiii(765,(162472|0),1,(10856|0),($34|0),1,566,(10968|0),($vararg_buffer1|0));
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       $39 = ___cxa_find_matching_catch_2()|0;
       $40 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($0);
       ___resumeException($39|0);
       // unreachable;
      } else {
       __ZN9ExitGames6Common7JStringD2Ev($0);
       break;
      }
     }
    } while(0);
    $41 = ((($this)) + 40|0);
    $42 = HEAP32[$41>>2]|0;
    $43 = (($42) + (($i$01$i$lcssa*40)|0)|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ((($44)) + 16|0);
    $46 = HEAP32[$45>>2]|0;
    (FUNCTION_TABLE_iii[$46 & 2047]($43,$val)|0);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 __ZN9ExitGames6Common7JVectorINS0_6ObjectEE10addElementERKS2_($4,$key);
 $16 = ((($this)) + 24|0);
 __ZN9ExitGames6Common7JVectorINS0_6ObjectEE10addElementERKS2_($16,$val);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7JVectorINS0_6ObjectEE10addElementERKS2_($this,$elem) {
 $this = $this|0;
 $elem = $elem|0;
 var $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if ($4) {
  $5 = ((($this)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($6) + ($1))|0;
  $8 = ($1>>>0)<($7>>>0);
  if ($8) {
   HEAP32[$2>>2] = $7;
   $9 = ($7*40)|0;
   $10 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($9)|0);
   $11 = HEAP32[$0>>2]|0;
   $12 = ($11|0)==(0);
   $$pre$i = ((($this)) + 16|0);
   if (!($12)) {
    $i$01$i = 0;
    while(1) {
     $13 = (($10) + (($i$01$i*40)|0)|0);
     $14 = HEAP32[$$pre$i>>2]|0;
     $15 = (($14) + (($i$01$i*40)|0)|0);
     __ZN9ExitGames6Common6ObjectC2ERKS1_($13,$15);
     $16 = HEAP32[$$pre$i>>2]|0;
     $17 = (($16) + (($i$01$i*40)|0)|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = HEAP32[$18>>2]|0;
     FUNCTION_TABLE_vi[$19 & 2047]($17);
     $20 = (($i$01$i) + 1)|0;
     $21 = HEAP32[$0>>2]|0;
     $22 = ($20>>>0)<($21>>>0);
     if ($22) {
      $i$01$i = $20;
     } else {
      break;
     }
    }
   }
   $23 = HEAP32[$$pre$i>>2]|0;
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($23);
   HEAP32[$$pre$i>>2] = $10;
   $$pre = HEAP32[$0>>2]|0;
   $27 = $$pre;
  } else {
   $27 = $1;
  }
 } else {
  $27 = $1;
 }
 $24 = ((($this)) + 16|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (($25) + (($27*40)|0)|0);
 __ZN9ExitGames6Common6ObjectC2ERKS1_($26,$elem);
 $28 = HEAP32[$0>>2]|0;
 $29 = (($28) + 1)|0;
 HEAP32[$0>>2] = $29;
 return;
}
function __ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZNK9ExitGames6Common14DictionaryBase7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($0)|0);
 return ($1|0);
}
function __ZN9ExitGames6Common14DictionaryBase12getElementAtEjPKNS0_6ObjectE($this,$index,$0) {
 $this = $this|0;
 $index = $index|0;
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($this)) + 4|0);
 $2 = (__ZN9ExitGames6Common9HashtableixEj($1,$index)|0);
 return ($2|0);
}
function __ZN9ExitGames6Common9HashtableixEj($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 28|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2>>>0)>($index>>>0);
 if ($3) {
  $11 = ((($this)) + 40|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (($12) + (($index*40)|0)|0);
  STACKTOP = sp;return ($13|0);
 }
 __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
 __THREW__ = 0;
 $4 = (invoke_ii(764,($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,(162472|0),1,(10856|0),($4|0),1,566,(10968|0),($vararg_buffer|0));
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $11 = ((($this)) + 40|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($12) + (($index*40)|0)|0);
 STACKTOP = sp;return ($13|0);
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation8readByteEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($2)|0);
 $6 = HEAP8[$5>>0]|0;
 return ($6|0);
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation7popByteERNS0_6ObjectE($this,$object) {
 $this = $this|0;
 $object = $object|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $res = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $res = sp;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + ($2)|0);
 $6 = HEAP8[$5>>0]|0;
 HEAP8[$res>>0] = $6;
 __ZN9ExitGames6Common6Object3setEPKvhhb($object,$res,98,0,1);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Common7Helpers26DeSerializerImplementation9readShortEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $by = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $by = sp;
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$0>>2]|0;
 $3 = (($2) + 1)|0;
 HEAP32[$0>>2] = $3;
 $4 = HEAP32[$1>>2]|0;
 $5 = (($4) + ($3)|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ((($by)) + 1|0);
 HEAP8[$7>>0] = $6;
 $8 = (($2) + 2)|0;
 HEAP32[$0>>2] = $8;
 $9 = (($4) + ($8)|0);
 $10 = HEAP8[$9>>0]|0;
 HEAP8[$by>>0] = $10;
 $11 = HEAP16[$by>>1]|0;
 STACKTOP = sp;return ($11|0);
}
function __ZN9ExitGames6Common6ObjectC2EPKvhhb($this,$data,$type,$customType,$makeCopy) {
 $this = $this|0;
 $data = $data|0;
 $type = $type|0;
 $customType = $customType|0;
 $makeCopy = $makeCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $size$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $size$i = sp;
 HEAP32[$this>>2] = (4640);
 HEAP32[$size$i>>2] = 1;
 __THREW__ = 0;
 invoke_viiiiiii(855,($this|0),($data|0),($type|0),($customType|0),0,($size$i|0),($makeCopy|0));
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_2()|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($2|0);
  // unreachable;
 } else {
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementationC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6112);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 12|0);
 HEAP32[$2>>2] = 0;
 return;
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation13getDataOffsetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation10pushObjectEPKNS0_6ObjectEb($this,$serObject,$setType) {
 $this = $this|0;
 $serObject = $serObject|0;
 $setType = $setType|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$serObject,$setType)|0);
 $1 = $setType&1;
 $2 = (($0) + ($1))|0;
 __ZN9ExitGames6Common7Helpers24SerializerImplementation20extendInternalBufferEi($this,$2);
 $3 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$serObject,$setType)|0);
 return ($3|0);
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$serObject,$setType) {
 $this = $this|0;
 $serObject = $serObject|0;
 $setType = $setType|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $factor = 0, $sz = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $sz = sp;
 $0 = ($serObject|0)==(0|0);
 if ($0) {
  label = 3;
 } else {
  $1 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $2 = ($1<<24>>24)==(42);
  if ($2) {
   $$0 = 1;
  } else {
   label = 3;
  }
 }
 L3: do {
  if ((label|0) == 3) {
   $3 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
   $4 = ($3|0)==(0|0);
   if ($4) {
    $5 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
    $6 = HEAP16[$5>>1]|0;
    $7 = ($6<<16>>16)==(0);
    if (!($7)) {
     $$0 = 0;
     break;
    }
   }
   $8 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($serObject)|0);
   switch ($8|0) {
   case 0:  {
    $9 = $setType&1;
    $10 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
    $11 = $10&255;
    do {
     switch ($11|0) {
     case 98:  {
      $12 = $setType ? 2 : 1;
      $$0 = $12;
      break L3;
      break;
     }
     case 107:  {
      $13 = $9 | 2;
      $$0 = $13;
      break L3;
      break;
     }
     case 105:  {
      $14 = $9 | 4;
      $$0 = $14;
      break L3;
      break;
     }
     case 108:  {
      $15 = $9 | 8;
      $$0 = $15;
      break L3;
      break;
     }
     case 102:  {
      $16 = $9 | 4;
      $$0 = $16;
      break L3;
      break;
     }
     case 100:  {
      $17 = $9 | 8;
      $$0 = $17;
      break L3;
      break;
     }
     case 111:  {
      $18 = $setType ? 2 : 1;
      $$0 = $18;
      break L3;
      break;
     }
     case 115:  {
      $19 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
      $20 = (__ZN9ExitGames6Common10UTF8String4sizeERKNS0_7JStringE($19)|0);
      $21 = $9 | 2;
      $22 = (($21) + ($20))|0;
      $$0 = $22;
      break L3;
      break;
     }
     case 104:  {
      $23 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
      $24 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_9HashtableE($this,$23)|0);
      $25 = (($24) + ($9))|0;
      $$0 = $25;
      break L3;
      break;
     }
     case 68:  {
      $26 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
      $27 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_14DictionaryBaseE($this,$26)|0);
      $28 = (($27) + ($9))|0;
      $$0 = $28;
      break L3;
      break;
     }
     case 99:  {
      $29 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
      $30 = (__ZNK9ExitGames6Common6Object13getCustomTypeEv($serObject)|0);
      $31 = (__ZN9ExitGames6Common14CustomTypeBase9serializeEPKvPhh($29,0,$30)|0);
      $32 = $31 << 16 >> 16;
      $factor = $9 << 1;
      $33 = (($32) + 2)|0;
      $34 = (($33) + ($factor))|0;
      $$0 = $34;
      break L3;
      break;
     }
     default: {
      $$0 = 0;
      break L3;
     }
     }
    } while(0);
    break;
   }
   case 1:  {
    $35 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
    $36 = ($35<<24>>24)==(98);
    if ($36) {
     $37 = $setType&1;
     $38 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = $37 | 4;
     $41 = (($40) + ($39))|0;
     $$0 = $41;
     break L3;
    }
    break;
   }
   default: {
   }
   }
   HEAP32[$sz>>2] = 1;
   $42 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
   $43 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
   $44 = (__ZNK9ExitGames6Common6Object13getCustomTypeEv($serObject)|0);
   $45 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($serObject)|0);
   $46 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
   (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13calcArraySizeEPKvhhjPKsjRj($this,$42,$43,$44,$45,$46,0,$sz)|0);
   $47 = HEAP32[$sz>>2]|0;
   $$0 = $47;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation20extendInternalBufferEi($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $$sum1$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond$i = 0, $or$cond3$i = 0, $scevgep$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($1) - ($3))|0;
 $5 = ($4|0)<($size|0);
 if (!($5)) {
  return;
 }
 $6 = (__ZN9ExitGames6Common7Helpers10MathHelper7getLog2Ej($size)|0);
 $7 = 2 << $6;
 $8 = HEAP32[$0>>2]|0;
 $9 = (($8) + ($7))|0;
 HEAP32[$0>>2] = $9;
 $10 = ((($this)) + 12|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)!=(0|0);
 if ($12) {
  $13 = ((($11)) + -8|0);
  $14 = $13;
  $15 = $14;
  $16 = HEAP32[$15>>2]|0;
  $17 = (($14) + 4)|0;
  $18 = $17;
  $19 = HEAP32[$18>>2]|0;
  $28 = $16;$30 = $19;
 } else {
  $28 = 0;$30 = 0;
 }
 $20 = (($9) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $9;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($28|0)!=(0);
 $29 = ($30|0)!=(0);
 $31 = $27 | $29;
 $32 = ($9|0)!=(0);
 $or$cond3$i = $32 & $31;
 if ($or$cond3$i) {
  $44 = 0;$48 = 0;
  while(1) {
   $$sum1$i = (($44) + 8)|0;
   $45 = (($21) + ($$sum1$i)|0);
   $46 = (($11) + ($44)|0);
   $47 = HEAP8[$46>>0]|0;
   HEAP8[$45>>0] = $47;
   $49 = (_i64Add(($44|0),($48|0),1,0)|0);
   $50 = tempRet0;
   $51 = ($50>>>0)<($30>>>0);
   $52 = ($49>>>0)<($28>>>0);
   $53 = ($50|0)==($30|0);
   $54 = $53 & $52;
   $55 = $51 | $54;
   $56 = ($50>>>0)<(0);
   $57 = ($49>>>0)<($9>>>0);
   $58 = ($50|0)==(0);
   $59 = $58 & $57;
   $60 = $56 | $59;
   $or$cond$i = $55 & $60;
   if ($or$cond$i) {
    $44 = $49;$48 = $50;
   } else {
    break;
   }
  }
 }
 $33 = ($30>>>0)<(0);
 $34 = ($28>>>0)<($9>>>0);
 $35 = ($30|0)==(0);
 $36 = $35 & $34;
 $37 = $33 | $36;
 if ($37) {
  $38 = (($28) + 8)|0;
  $scevgep$i = (($21) + ($38)|0);
  $39 = (_i64Add(($9|0),0,-1,0)|0);
  $40 = tempRet0;
  $41 = (_i64Subtract(($39|0),($40|0),($28|0),($30|0))|0);
  $42 = tempRet0;
  $43 = (($41) + 1)|0;
  _memset(($scevgep$i|0),0,($43|0))|0;
 }
 if ($12) {
  $61 = ((($11)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($61);
 }
 HEAP32[$10>>2] = $22;
 return;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$serObject,$setType) {
 $this = $this|0;
 $serObject = $serObject|0;
 $setType = $setType|0;
 var $$0 = 0, $$sum$i$i = 0, $$sum2$i$i = 0, $$sum3$i$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $exitcond$i = 0, $i$02$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($serObject|0)==(0|0);
 if (!($0)) {
  $1 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $2 = ($1<<24>>24)==(42);
  if ($2) {
   $3 = ((($this)) + 8|0);
   $4 = ((($this)) + 12|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = HEAP32[$3>>2]|0;
   $7 = (($5) + ($6)|0);
   HEAP8[$7>>0] = 42;
   $8 = HEAP32[$3>>2]|0;
   $9 = (($8) + 1)|0;
   HEAP32[$3>>2] = $9;
   $$0 = 1;
   return ($$0|0);
  }
 }
 $10 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
 $11 = ($10|0)==(0|0);
 if ($11) {
  $12 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
  $13 = HEAP16[$12>>1]|0;
  $14 = ($13<<16>>16)==(0);
  if (!($14)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $15 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($serObject)|0);
 switch ($15|0) {
 case 0:  {
  if ($setType) {
   $16 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
   $17 = ((($this)) + 8|0);
   $18 = ((($this)) + 12|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = HEAP32[$17>>2]|0;
   $21 = (($19) + ($20)|0);
   HEAP8[$21>>0] = $16;
   $22 = HEAP32[$17>>2]|0;
   $23 = (($22) + 1)|0;
   HEAP32[$17>>2] = $23;
   $24 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
   $25 = ($24<<24>>24)==(99);
   if ($25) {
    $26 = (__ZNK9ExitGames6Common6Object13getCustomTypeEv($serObject)|0);
    $27 = HEAP32[$18>>2]|0;
    $28 = HEAP32[$17>>2]|0;
    $29 = (($27) + ($28)|0);
    HEAP8[$29>>0] = $26;
    $30 = HEAP32[$17>>2]|0;
    $31 = (($30) + 1)|0;
    HEAP32[$17>>2] = $31;
   }
  }
  $32 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
  $33 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $34 = (__ZNK9ExitGames6Common6Object13getCustomTypeEv($serObject)|0);
  $35 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation14writeArrayTypeEPKvhhs($this,$32,$33,$34,0)|0);
  if ($35) {
   label = 17;
  } else {
   $$0 = 0;
   return ($$0|0);
  }
  break;
 }
 case 1:  {
  $36 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $37 = ($36<<24>>24)==(98);
  if ($37) {
   $38 = ((($this)) + 8|0);
   $39 = ((($this)) + 12|0);
   if ($setType) {
    $40 = HEAP32[$39>>2]|0;
    $41 = HEAP32[$38>>2]|0;
    $42 = (($40) + ($41)|0);
    HEAP8[$42>>0] = 120;
    $43 = HEAP32[$38>>2]|0;
    $44 = (($43) + 1)|0;
    HEAP32[$38>>2] = $44;
   }
   $45 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
   $46 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = $47&255;
   $49 = $47 >>> 8;
   $50 = $49&255;
   $51 = $47 >>> 16;
   $52 = $51&255;
   $53 = $47 >>> 24;
   $54 = $53&255;
   $55 = HEAP32[$39>>2]|0;
   $56 = HEAP32[$38>>2]|0;
   $57 = (($55) + ($56)|0);
   HEAP8[$57>>0] = $54;
   $$sum$i$i = (($56) + 1)|0;
   $58 = (($55) + ($$sum$i$i)|0);
   HEAP8[$58>>0] = $52;
   $$sum2$i$i = (($56) + 2)|0;
   $59 = (($55) + ($$sum2$i$i)|0);
   HEAP8[$59>>0] = $50;
   $$sum3$i$i = (($56) + 3)|0;
   $60 = (($55) + ($$sum3$i$i)|0);
   HEAP8[$60>>0] = $48;
   $61 = HEAP32[$38>>2]|0;
   $62 = (($61) + 4)|0;
   HEAP32[$38>>2] = $62;
   $63 = $53 << 24;
   $64 = $47 & 16777215;
   $65 = $63 | $64;
   $66 = ($65|0)>(0);
   if ($66) {
    $70 = $62;$i$02$i = 0;
    while(1) {
     $67 = (($45) + ($i$02$i)|0);
     $68 = HEAP32[$39>>2]|0;
     $69 = (($68) + ($70)|0);
     $71 = HEAP8[$67>>0]|0;
     HEAP8[$69>>0] = $71;
     $72 = HEAP32[$38>>2]|0;
     $73 = (($72) + 1)|0;
     HEAP32[$38>>2] = $73;
     $74 = (($i$02$i) + 1)|0;
     $exitcond$i = ($74|0)==($65|0);
     if ($exitcond$i) {
      label = 17;
      break;
     } else {
      $70 = $73;$i$02$i = $74;
     }
    }
   } else {
    label = 17;
   }
  } else {
   label = 16;
  }
  break;
 }
 default: {
  label = 16;
 }
 }
 if ((label|0) == 16) {
  $75 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $76 = ($75<<24>>24)==(122);
  $77 = ((($this)) + 8|0);
  $78 = ((($this)) + 12|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = HEAP32[$77>>2]|0;
  $81 = (($79) + ($80)|0);
  $82 = $76 ? 122 : 121;
  HEAP8[$81>>0] = $82;
  $83 = HEAP32[$77>>2]|0;
  $84 = (($83) + 1)|0;
  HEAP32[$77>>2] = $84;
  $85 = (__ZNK9ExitGames6Common6Object7getDataEv($serObject)|0);
  $86 = (__ZNK9ExitGames6Common6Object7getTypeEv($serObject)|0);
  $87 = (__ZNK9ExitGames6Common6Object13getCustomTypeEv($serObject)|0);
  $88 = (__ZNK9ExitGames6Common6Object13getDimensionsEv($serObject)|0);
  $89 = (__ZNK9ExitGames6Common6Object8getSizesEv($serObject)|0);
  $90 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation16writeArrayHelperEPKvhhjPsj($this,$85,$86,$87,$88,$89,0)|0);
  $$0 = $90;
  return ($$0|0);
 }
 else if ((label|0) == 17) {
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation14writeArrayTypeEPKvhhs($this,$pData,$type,$customType,$i) {
 $this = $this|0;
 $pData = $pData|0;
 $type = $type|0;
 $customType = $customType|0;
 $i = $i|0;
 var $$0 = 0, $$sum$i = 0, $$sum$i1 = 0, $$sum$i2 = 0, $$sum$i5 = 0, $$sum$i8 = 0, $$sum2$i = 0, $$sum2$i3 = 0, $$sum2$i6 = 0, $$sum2$i9 = 0, $$sum3$i = 0, $$sum3$i10 = 0, $$sum3$i4 = 0, $$sum3$i7 = 0, $$sum4$i = 0, $$sum4$i11 = 0, $$sum5$i = 0, $$sum5$i12 = 0, $$sum6$i = 0, $$sum6$i13 = 0;
 var $$sum7$i = 0, $$sum7$i14 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $type&255;
 do {
  switch ($0|0) {
  case 98:  {
   $1 = $i << 16 >> 16;
   $2 = (($pData) + ($1)|0);
   $3 = ((($this)) + 8|0);
   $4 = ((($this)) + 12|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = HEAP32[$3>>2]|0;
   $7 = (($5) + ($6)|0);
   $8 = HEAP8[$2>>0]|0;
   HEAP8[$7>>0] = $8;
   $9 = HEAP32[$3>>2]|0;
   $10 = (($9) + 1)|0;
   HEAP32[$3>>2] = $10;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 107:  {
   $11 = $i << 16 >> 16;
   $12 = (($pData) + ($11<<1)|0);
   $13 = ((($this)) + 8|0);
   $14 = ((($this)) + 12|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = HEAP32[$13>>2]|0;
   $17 = (($15) + ($16)|0);
   $18 = ((($12)) + 1|0);
   $19 = HEAP8[$18>>0]|0;
   HEAP8[$17>>0] = $19;
   $$sum$i = (($16) + 1)|0;
   $20 = (($15) + ($$sum$i)|0);
   $21 = HEAP8[$12>>0]|0;
   HEAP8[$20>>0] = $21;
   $22 = HEAP32[$13>>2]|0;
   $23 = (($22) + 2)|0;
   HEAP32[$13>>2] = $23;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 105:  {
   $24 = $i << 16 >> 16;
   $25 = (($pData) + ($24<<2)|0);
   $26 = ((($this)) + 8|0);
   $27 = ((($this)) + 12|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = HEAP32[$26>>2]|0;
   $30 = (($28) + ($29)|0);
   $31 = ((($25)) + 3|0);
   $32 = HEAP8[$31>>0]|0;
   HEAP8[$30>>0] = $32;
   $$sum$i1 = (($29) + 1)|0;
   $33 = (($28) + ($$sum$i1)|0);
   $34 = ((($25)) + 2|0);
   $35 = HEAP8[$34>>0]|0;
   HEAP8[$33>>0] = $35;
   $$sum2$i = (($29) + 2)|0;
   $36 = (($28) + ($$sum2$i)|0);
   $37 = ((($25)) + 1|0);
   $38 = HEAP8[$37>>0]|0;
   HEAP8[$36>>0] = $38;
   $$sum3$i = (($29) + 3)|0;
   $39 = (($28) + ($$sum3$i)|0);
   $40 = HEAP8[$25>>0]|0;
   HEAP8[$39>>0] = $40;
   $41 = HEAP32[$26>>2]|0;
   $42 = (($41) + 4)|0;
   HEAP32[$26>>2] = $42;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 108:  {
   $43 = $i << 16 >> 16;
   $44 = (($pData) + ($43<<3)|0);
   $45 = ((($this)) + 8|0);
   $46 = ((($this)) + 12|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = HEAP32[$45>>2]|0;
   $49 = (($47) + ($48)|0);
   $50 = ((($44)) + 7|0);
   $51 = HEAP8[$50>>0]|0;
   HEAP8[$49>>0] = $51;
   $$sum$i2 = (($48) + 1)|0;
   $52 = (($47) + ($$sum$i2)|0);
   $53 = ((($44)) + 6|0);
   $54 = HEAP8[$53>>0]|0;
   HEAP8[$52>>0] = $54;
   $$sum2$i3 = (($48) + 2)|0;
   $55 = (($47) + ($$sum2$i3)|0);
   $56 = ((($44)) + 5|0);
   $57 = HEAP8[$56>>0]|0;
   HEAP8[$55>>0] = $57;
   $$sum3$i4 = (($48) + 3)|0;
   $58 = (($47) + ($$sum3$i4)|0);
   $59 = ((($44)) + 4|0);
   $60 = HEAP8[$59>>0]|0;
   HEAP8[$58>>0] = $60;
   $$sum4$i = (($48) + 4)|0;
   $61 = (($47) + ($$sum4$i)|0);
   $62 = ((($44)) + 3|0);
   $63 = HEAP8[$62>>0]|0;
   HEAP8[$61>>0] = $63;
   $$sum5$i = (($48) + 5)|0;
   $64 = (($47) + ($$sum5$i)|0);
   $65 = ((($44)) + 2|0);
   $66 = HEAP8[$65>>0]|0;
   HEAP8[$64>>0] = $66;
   $$sum6$i = (($48) + 6)|0;
   $67 = (($47) + ($$sum6$i)|0);
   $68 = ((($44)) + 1|0);
   $69 = HEAP8[$68>>0]|0;
   HEAP8[$67>>0] = $69;
   $$sum7$i = (($48) + 7)|0;
   $70 = (($47) + ($$sum7$i)|0);
   $71 = HEAP8[$44>>0]|0;
   HEAP8[$70>>0] = $71;
   $72 = HEAP32[$45>>2]|0;
   $73 = (($72) + 8)|0;
   HEAP32[$45>>2] = $73;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 102:  {
   $74 = $i << 16 >> 16;
   $75 = (($pData) + ($74<<2)|0);
   $76 = ((($this)) + 8|0);
   $77 = ((($this)) + 12|0);
   $78 = HEAP32[$77>>2]|0;
   $79 = HEAP32[$76>>2]|0;
   $80 = (($78) + ($79)|0);
   $81 = ((($75)) + 3|0);
   $82 = HEAP8[$81>>0]|0;
   HEAP8[$80>>0] = $82;
   $$sum$i5 = (($79) + 1)|0;
   $83 = (($78) + ($$sum$i5)|0);
   $84 = ((($75)) + 2|0);
   $85 = HEAP8[$84>>0]|0;
   HEAP8[$83>>0] = $85;
   $$sum2$i6 = (($79) + 2)|0;
   $86 = (($78) + ($$sum2$i6)|0);
   $87 = ((($75)) + 1|0);
   $88 = HEAP8[$87>>0]|0;
   HEAP8[$86>>0] = $88;
   $$sum3$i7 = (($79) + 3)|0;
   $89 = (($78) + ($$sum3$i7)|0);
   $90 = HEAP8[$75>>0]|0;
   HEAP8[$89>>0] = $90;
   $91 = HEAP32[$76>>2]|0;
   $92 = (($91) + 4)|0;
   HEAP32[$76>>2] = $92;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 100:  {
   $93 = $i << 16 >> 16;
   $94 = (($pData) + ($93<<3)|0);
   $95 = ((($this)) + 8|0);
   $96 = ((($this)) + 12|0);
   $97 = HEAP32[$96>>2]|0;
   $98 = HEAP32[$95>>2]|0;
   $99 = (($97) + ($98)|0);
   $100 = ((($94)) + 7|0);
   $101 = HEAP8[$100>>0]|0;
   HEAP8[$99>>0] = $101;
   $$sum$i8 = (($98) + 1)|0;
   $102 = (($97) + ($$sum$i8)|0);
   $103 = ((($94)) + 6|0);
   $104 = HEAP8[$103>>0]|0;
   HEAP8[$102>>0] = $104;
   $$sum2$i9 = (($98) + 2)|0;
   $105 = (($97) + ($$sum2$i9)|0);
   $106 = ((($94)) + 5|0);
   $107 = HEAP8[$106>>0]|0;
   HEAP8[$105>>0] = $107;
   $$sum3$i10 = (($98) + 3)|0;
   $108 = (($97) + ($$sum3$i10)|0);
   $109 = ((($94)) + 4|0);
   $110 = HEAP8[$109>>0]|0;
   HEAP8[$108>>0] = $110;
   $$sum4$i11 = (($98) + 4)|0;
   $111 = (($97) + ($$sum4$i11)|0);
   $112 = ((($94)) + 3|0);
   $113 = HEAP8[$112>>0]|0;
   HEAP8[$111>>0] = $113;
   $$sum5$i12 = (($98) + 5)|0;
   $114 = (($97) + ($$sum5$i12)|0);
   $115 = ((($94)) + 2|0);
   $116 = HEAP8[$115>>0]|0;
   HEAP8[$114>>0] = $116;
   $$sum6$i13 = (($98) + 6)|0;
   $117 = (($97) + ($$sum6$i13)|0);
   $118 = ((($94)) + 1|0);
   $119 = HEAP8[$118>>0]|0;
   HEAP8[$117>>0] = $119;
   $$sum7$i14 = (($98) + 7)|0;
   $120 = (($97) + ($$sum7$i14)|0);
   $121 = HEAP8[$94>>0]|0;
   HEAP8[$120>>0] = $121;
   $122 = HEAP32[$95>>2]|0;
   $123 = (($122) + 8)|0;
   HEAP32[$95>>2] = $123;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 111:  {
   $124 = $i << 16 >> 16;
   $125 = (($pData) + ($124)|0);
   $126 = ((($this)) + 8|0);
   $127 = ((($this)) + 12|0);
   $128 = HEAP32[$127>>2]|0;
   $129 = HEAP32[$126>>2]|0;
   $130 = (($128) + ($129)|0);
   $131 = HEAP8[$125>>0]|0;
   HEAP8[$130>>0] = $131;
   $132 = HEAP32[$126>>2]|0;
   $133 = (($132) + 1)|0;
   HEAP32[$126>>2] = $133;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 115:  {
   $134 = $i << 16 >> 16;
   $135 = (($pData) + ($134<<4)|0);
   (__ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_7JStringE($this,$135)|0);
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 104:  {
   $136 = $i << 16 >> 16;
   $137 = (($pData) + (($136*44)|0)|0);
   $138 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_9HashtableE($this,$137)|0);
   $$0 = $138;
   return ($$0|0);
   break;
  }
  case 68:  {
   $139 = $i << 16 >> 16;
   $140 = (($pData) + (($139*52)|0)|0);
   $141 = ($i<<16>>16)==(0);
   $142 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_14DictionaryBaseEb($this,$140,$141)|0);
   $$0 = $142;
   return ($$0|0);
   break;
  }
  case 99:  {
   (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeCustomEPKvh($this,$pData,$customType)|0);
   $$0 = 1;
   return ($$0|0);
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0);
  }
  }
 } while(0);
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation16writeArrayHelperEPKvhhjPsj($this,$pData,$type,$customType,$dimensions,$arraySizes,$recursionDepth) {
 $this = $this|0;
 $pData = $pData|0;
 $type = $type|0;
 $customType = $customType|0;
 $dimensions = $dimensions|0;
 $arraySizes = $arraySizes|0;
 $recursionDepth = $recursionDepth|0;
 var $$0 = 0, $$pre = 0, $$pre8 = 0, $$sink3 = 0, $$sum$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $customType$sink = 0, $i$03 = 0;
 var $i$16 = 0, $i$16$us = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pData|0)==(0|0);
 $1 = (($arraySizes) + ($recursionDepth<<1)|0);
 if ($0) {
  $2 = HEAP16[$1>>1]|0;
  $3 = ($2<<16>>16)==(0);
  if (!($3)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $4 = ((($this)) + 8|0);
 $5 = ((($this)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$4>>2]|0;
 $8 = (($6) + ($7)|0);
 $9 = ((($1)) + 1|0);
 $10 = HEAP8[$9>>0]|0;
 HEAP8[$8>>0] = $10;
 $$sum$i = (($7) + 1)|0;
 $11 = (($6) + ($$sum$i)|0);
 $12 = HEAP8[$1>>0]|0;
 HEAP8[$11>>0] = $12;
 $13 = HEAP32[$4>>2]|0;
 $14 = (($13) + 2)|0;
 HEAP32[$4>>2] = $14;
 $15 = ($dimensions|0)!=(0);
 $16 = (($dimensions) + -1)|0;
 $17 = ($16>>>0)>($recursionDepth>>>0);
 $or$cond = $15 & $17;
 if ($or$cond) {
  $18 = ($type<<24>>24)==(122);
  $19 = HEAP32[$5>>2]|0;
  $20 = (($19) + ($14)|0);
  $21 = $18 ? 122 : 121;
  HEAP8[$20>>0] = $21;
  $22 = HEAP32[$4>>2]|0;
  $23 = (($22) + 1)|0;
  HEAP32[$4>>2] = $23;
  $24 = HEAP16[$1>>1]|0;
  $25 = ($24<<16>>16)>(0);
  if (!($25)) {
   $$0 = 1;
   return ($$0|0);
  }
  $26 = (($recursionDepth) + 1)|0;
  $i$03 = 0;
  while(1) {
   $30 = $i$03 << 16 >> 16;
   $31 = (($pData) + ($30<<2)|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation16writeArrayHelperEPKvhhjPsj($this,$32,$type,$customType,$dimensions,$arraySizes,$26)|0);
   $29 = (($i$03) + 1)<<16>>16;
   if (!($33)) {
    $$0 = 0;
    label = 15;
    break;
   }
   $27 = HEAP16[$1>>1]|0;
   $28 = ($29<<16>>16)<($27<<16>>16);
   if ($28) {
    $i$03 = $29;
   } else {
    $$0 = 1;
    label = 15;
    break;
   }
  }
  if ((label|0) == 15) {
   return ($$0|0);
  }
 }
 $34 = $type&255;
 switch ($34|0) {
 case 68: case 104: case 115: case 111: case 100: case 102: case 108: case 105: case 107: case 98:  {
  $$sink3 = $14;$customType$sink = $type;
  label = 10;
  break;
 }
 case 99:  {
  $35 = HEAP32[$5>>2]|0;
  $36 = (($35) + ($14)|0);
  HEAP8[$36>>0] = $type;
  $37 = HEAP32[$4>>2]|0;
  $38 = (($37) + 1)|0;
  HEAP32[$4>>2] = $38;
  $$sink3 = $38;$customType$sink = $customType;
  label = 10;
  break;
 }
 case 122:  {
  break;
 }
 default: {
  $$0 = 0;
  return ($$0|0);
 }
 }
 if ((label|0) == 10) {
  $39 = HEAP32[$5>>2]|0;
  $40 = (($39) + ($$sink3)|0);
  HEAP8[$40>>0] = $customType$sink;
  $41 = HEAP32[$4>>2]|0;
  $42 = (($41) + 1)|0;
  HEAP32[$4>>2] = $42;
 }
 $43 = HEAP16[$1>>1]|0;
 $44 = ($43<<16>>16)>(0);
 if (!($44)) {
  $$0 = 1;
  return ($$0|0);
 }
 $45 = ($type<<24>>24)==(122);
 if ($45) {
  $47 = 0;$i$16$us = 0;
  while(1) {
   $46 = (($pData) + (($47*40)|0)|0);
   (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$46,1)|0);
   $$pre8 = HEAP16[$1>>1]|0;
   $48 = (($i$16$us) + 1)<<16>>16;
   $49 = $48 << 16 >> 16;
   $50 = ($48<<16>>16)<($$pre8<<16>>16);
   if ($50) {
    $47 = $49;$i$16$us = $48;
   } else {
    $$0 = 1;
    break;
   }
  }
  return ($$0|0);
 } else {
  $i$16 = 0;
  while(1) {
   (__ZN9ExitGames6Common7Helpers24SerializerImplementation14writeArrayTypeEPKvhhs($this,$pData,$type,$customType,$i$16)|0);
   $$pre = HEAP16[$1>>1]|0;
   $51 = (($i$16) + 1)<<16>>16;
   $52 = ($51<<16>>16)<($$pre<<16>>16);
   if ($52) {
    $i$16 = $51;
   } else {
    $$0 = 1;
    break;
   }
  }
  return ($$0|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_7JStringE($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $$pre = 0, $$sum$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $lftr$wideiv = 0, $utfString = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $utfString = sp;
 __ZN9ExitGames6Common10UTF8StringC2ERKNS0_7JStringE($utfString,$str);
 __THREW__ = 0;
 $0 = (invoke_ii(231,($utfString|0))|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN9ExitGames6Common10UTF8StringD2Ev($utfString);
  ___resumeException($26|0);
  // unreachable;
 }
 $3 = $0&65535;
 $4 = $0&255;
 $5 = ($3&65535) >>> 8;
 $6 = $5&255;
 $7 = ((($this)) + 8|0);
 $8 = ((($this)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$7>>2]|0;
 $11 = (($9) + ($10)|0);
 HEAP8[$11>>0] = $6;
 $$sum$i = (($10) + 1)|0;
 $12 = (($9) + ($$sum$i)|0);
 HEAP8[$12>>0] = $4;
 $13 = HEAP32[$7>>2]|0;
 $14 = (($13) + 2)|0;
 HEAP32[$7>>2] = $14;
 __THREW__ = 0;
 $15 = (invoke_ii(826,($utfString|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZN9ExitGames6Common10UTF8StringD2Ev($utfString);
  ___resumeException($26|0);
  // unreachable;
 }
 $18 = ($3<<16>>16)>(0);
 if (!($18)) {
  __ZN9ExitGames6Common10UTF8StringD2Ev($utfString);
  STACKTOP = sp;return 1;
 }
 $$pre = HEAP32[$7>>2]|0;
 $22 = $$pre;$indvars$iv = 0;
 while(1) {
  $19 = (($15) + ($indvars$iv)|0);
  $20 = HEAP32[$8>>2]|0;
  $21 = (($20) + ($22)|0);
  $23 = HEAP8[$19>>0]|0;
  HEAP8[$21>>0] = $23;
  $24 = HEAP32[$7>>2]|0;
  $25 = (($24) + 1)|0;
  HEAP32[$7>>2] = $25;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $lftr$wideiv = $indvars$iv$next&65535;
  $exitcond = ($lftr$wideiv<<16>>16)==($3<<16>>16);
  if ($exitcond) {
   break;
  } else {
   $22 = $25;$indvars$iv = $indvars$iv$next;
  }
 }
 __ZN9ExitGames6Common10UTF8StringD2Ev($utfString);
 STACKTOP = sp;return 1;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_9HashtableE($this,$hash) {
 $this = $this|0;
 $hash = $hash|0;
 var $$0 = 0, $$sum$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $indvars$iv = 0, $indvars$iv$next = 0, $sext = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($hash)|0);
 $3 = $2&65535;
 $4 = $2&255;
 $5 = ($3&65535) >>> 8;
 $6 = $5&255;
 $7 = ((($this)) + 8|0);
 $8 = ((($this)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = HEAP32[$7>>2]|0;
 $11 = (($9) + ($10)|0);
 HEAP8[$11>>0] = $6;
 $$sum$i = (($10) + 1)|0;
 $12 = (($9) + ($$sum$i)|0);
 HEAP8[$12>>0] = $4;
 $13 = HEAP32[$7>>2]|0;
 $14 = (($13) + 2)|0;
 HEAP32[$7>>2] = $14;
 $15 = ($3<<16>>16)>(0);
 if (!($15)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $sext = $2 << 16;
 $16 = $sext >> 16;
 $indvars$iv = 0;
 while(1) {
  $18 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($hash)|0);
  $19 = ((($18)) + 4|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ($20>>>0)>($indvars$iv>>>0);
  if (!($21)) {
   __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
   __THREW__ = 0;
   $22 = (invoke_ii(764,($1|0))|0);
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    label = 8;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($22|0),1,566,(10968|0),($vararg_buffer|0));
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    label = 8;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($1);
  }
  $29 = ((($18)) + 16|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = (($30) + (($indvars$iv*40)|0)|0);
  $32 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$31,1)|0);
  if (!($32)) {
   $$0 = 0;
   label = 16;
   break;
  }
  $33 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($hash)|0);
  $34 = ((($33)) + 4|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ($35>>>0)>($indvars$iv>>>0);
  if (!($36)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $37 = (invoke_ii(764,($0|0))|0);
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    label = 14;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($37|0),1,566,(10968|0),($vararg_buffer1|0));
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    label = 14;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  $44 = ((($33)) + 16|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($45) + (($indvars$iv*40)|0)|0);
  $47 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($hash,$46)|0);
  $48 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$47,1)|0);
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  if (!($48)) {
   $$0 = 0;
   label = 16;
   break;
  }
  $17 = ($indvars$iv$next|0)<($16|0);
  if ($17) {
   $indvars$iv = $indvars$iv$next;
  } else {
   $$0 = 1;
   label = 16;
   break;
  }
 }
 if ((label|0) == 8) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($27|0);
  // unreachable;
 }
 else if ((label|0) == 14) {
  $42 = ___cxa_find_matching_catch_2()|0;
  $43 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($42|0);
  // unreachable;
 }
 else if ((label|0) == 16) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation5writeERKNS0_14DictionaryBaseEb($this,$dict,$setKeyValueTypes) {
 $this = $this|0;
 $dict = $dict|0;
 $setKeyValueTypes = $setKeyValueTypes|0;
 var $$0 = 0, $$pre = 0, $$pre$phi11Z2D = 0, $$pre$phiZ2D = 0, $$pre10 = 0, $$sum$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$07 = 0, $indvars$iv = 0, $indvars$iv$next = 0, $j$0 = 0, $sext = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 L1: do {
  if ($setKeyValueTypes) {
   $2 = ((($this)) + 8|0);
   $3 = ((($this)) + 12|0);
   $j$0 = 0;
   while(1) {
    $4 = ($j$0|0)==(0);
    if (!($4)) {
     $5 = (($j$0) + -1)|0;
     $6 = HEAP32[$dict>>2]|0;
     $7 = ((($6)) + 20|0);
     $8 = HEAP32[$7>>2]|0;
     $9 = (FUNCTION_TABLE_ii[$8 & 2047]($dict)|0);
     $10 = (($9) + ($5)|0);
     $11 = HEAP8[$10>>0]|0;
     $12 = ($11<<24>>24)==(68);
     if (!($12)) {
      $$pre$phi11Z2D = $3;$$pre$phiZ2D = $2;
      break L1;
     }
    }
    $13 = HEAP32[$dict>>2]|0;
    $14 = ((($13)) + 16|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = (FUNCTION_TABLE_ii[$15 & 2047]($dict)|0);
    $17 = (($16) + ($j$0)|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = ($18<<24>>24)==(122);
    if ($19) {
     $29 = 167382;
    } else {
     $20 = HEAP32[$dict>>2]|0;
     $21 = ((($20)) + 16|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = (FUNCTION_TABLE_ii[$22 & 2047]($dict)|0);
     $24 = (($23) + ($j$0)|0);
     $29 = $24;
    }
    $25 = HEAP32[$3>>2]|0;
    $26 = HEAP32[$2>>2]|0;
    $27 = (($25) + ($26)|0);
    $28 = HEAP8[$29>>0]|0;
    HEAP8[$27>>0] = $28;
    $30 = HEAP32[$2>>2]|0;
    $31 = (($30) + 1)|0;
    HEAP32[$2>>2] = $31;
    $32 = HEAP32[$dict>>2]|0;
    $33 = ((($32)) + 20|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (FUNCTION_TABLE_ii[$34 & 2047]($dict)|0);
    $36 = (($35) + ($j$0)|0);
    $37 = HEAP8[$36>>0]|0;
    $38 = ($37<<24>>24)==(122);
    if (!($38)) {
     $39 = HEAP32[$dict>>2]|0;
     $40 = ((($39)) + 24|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = (FUNCTION_TABLE_ii[$41 & 2047]($dict)|0);
     $43 = (($42) + ($j$0<<2)|0);
     $44 = HEAP32[$43>>2]|0;
     $45 = ($44|0)==(0);
     if (!($45)) {
      $i$07 = 0;
      while(1) {
       $46 = HEAP32[$3>>2]|0;
       $47 = HEAP32[$2>>2]|0;
       $48 = (($46) + ($47)|0);
       HEAP8[$48>>0] = 121;
       $49 = HEAP32[$2>>2]|0;
       $50 = (($49) + 1)|0;
       HEAP32[$2>>2] = $50;
       $51 = (($i$07) + 1)|0;
       $52 = HEAP32[$dict>>2]|0;
       $53 = ((($52)) + 24|0);
       $54 = HEAP32[$53>>2]|0;
       $55 = (FUNCTION_TABLE_ii[$54 & 2047]($dict)|0);
       $56 = (($55) + ($j$0<<2)|0);
       $57 = HEAP32[$56>>2]|0;
       $58 = ($51>>>0)<($57>>>0);
       if ($58) {
        $i$07 = $51;
       } else {
        break;
       }
      }
     }
    }
    $59 = HEAP32[$dict>>2]|0;
    $60 = ((($59)) + 20|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = (FUNCTION_TABLE_ii[$61 & 2047]($dict)|0);
    $63 = (($62) + ($j$0)|0);
    $64 = HEAP8[$63>>0]|0;
    $65 = ($64<<24>>24)==(122);
    if ($65) {
     $66 = HEAP32[$dict>>2]|0;
     $67 = ((($66)) + 24|0);
     $68 = HEAP32[$67>>2]|0;
     $69 = (FUNCTION_TABLE_ii[$68 & 2047]($dict)|0);
     $70 = (($69) + ($j$0<<2)|0);
     $71 = HEAP32[$70>>2]|0;
     $72 = ($71|0)==(0);
     if ($72) {
      $82 = 167382;
     } else {
      label = 13;
     }
    } else {
     label = 13;
    }
    if ((label|0) == 13) {
     label = 0;
     $73 = HEAP32[$dict>>2]|0;
     $74 = ((($73)) + 20|0);
     $75 = HEAP32[$74>>2]|0;
     $76 = (FUNCTION_TABLE_ii[$75 & 2047]($dict)|0);
     $77 = (($76) + ($j$0)|0);
     $82 = $77;
    }
    $78 = HEAP32[$3>>2]|0;
    $79 = HEAP32[$2>>2]|0;
    $80 = (($78) + ($79)|0);
    $81 = HEAP8[$82>>0]|0;
    HEAP8[$80>>0] = $81;
    $83 = HEAP32[$2>>2]|0;
    $84 = (($83) + 1)|0;
    HEAP32[$2>>2] = $84;
    $85 = (($j$0) + 1)|0;
    $j$0 = $85;
   }
  } else {
   $$pre = ((($this)) + 8|0);
   $$pre10 = ((($this)) + 12|0);
   $$pre$phi11Z2D = $$pre10;$$pre$phiZ2D = $$pre;
  }
 } while(0);
 $86 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($dict)|0);
 $87 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($86)|0);
 $88 = $87&65535;
 $89 = $87&255;
 $90 = ($88&65535) >>> 8;
 $91 = $90&255;
 $92 = HEAP32[$$pre$phi11Z2D>>2]|0;
 $93 = HEAP32[$$pre$phiZ2D>>2]|0;
 $94 = (($92) + ($93)|0);
 HEAP8[$94>>0] = $91;
 $$sum$i = (($93) + 1)|0;
 $95 = (($92) + ($$sum$i)|0);
 HEAP8[$95>>0] = $89;
 $96 = HEAP32[$$pre$phiZ2D>>2]|0;
 $97 = (($96) + 2)|0;
 HEAP32[$$pre$phiZ2D>>2] = $97;
 $98 = ($88<<16>>16)>(0);
 if (!($98)) {
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 $sext = $87 << 16;
 $99 = $sext >> 16;
 $indvars$iv = 0;
 while(1) {
  $101 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($86)|0);
  $102 = ((($101)) + 4|0);
  $103 = HEAP32[$102>>2]|0;
  $104 = ($103>>>0)>($indvars$iv>>>0);
  if (!($104)) {
   __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
   __THREW__ = 0;
   $105 = (invoke_ii(764,($1|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    label = 22;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($105|0),1,566,(10968|0),($vararg_buffer|0));
   $108 = __THREW__; __THREW__ = 0;
   $109 = $108&1;
   if ($109) {
    label = 22;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($1);
  }
  $112 = ((($101)) + 16|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = (($113) + (($indvars$iv*40)|0)|0);
  $115 = HEAP32[$dict>>2]|0;
  $116 = ((($115)) + 16|0);
  $117 = HEAP32[$116>>2]|0;
  $118 = (FUNCTION_TABLE_ii[$117 & 2047]($dict)|0);
  $119 = HEAP8[$118>>0]|0;
  $120 = ($119<<24>>24)==(122);
  $121 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$114,$120)|0);
  if (!($121)) {
   $$0 = 0;
   label = 30;
   break;
  }
  $122 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($86)|0);
  $123 = ((($122)) + 4|0);
  $124 = HEAP32[$123>>2]|0;
  $125 = ($124>>>0)>($indvars$iv>>>0);
  if (!($125)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $126 = (invoke_ii(764,($0|0))|0);
   $127 = __THREW__; __THREW__ = 0;
   $128 = $127&1;
   if ($128) {
    label = 28;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($126|0),1,566,(10968|0),($vararg_buffer1|0));
   $129 = __THREW__; __THREW__ = 0;
   $130 = $129&1;
   if ($130) {
    label = 28;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  $133 = ((($122)) + 16|0);
  $134 = HEAP32[$133>>2]|0;
  $135 = (($134) + (($indvars$iv*40)|0)|0);
  $136 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($86,$135)|0);
  $137 = HEAP32[$dict>>2]|0;
  $138 = ((($137)) + 20|0);
  $139 = HEAP32[$138>>2]|0;
  $140 = (FUNCTION_TABLE_ii[$139 & 2047]($dict)|0);
  $141 = HEAP8[$140>>0]|0;
  $142 = ($141<<24>>24)==(122);
  $143 = (__ZN9ExitGames6Common7Helpers24SerializerImplementation11writeObjectEPKNS0_6ObjectEb($this,$136,$142)|0);
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  if (!($143)) {
   $$0 = 0;
   label = 30;
   break;
  }
  $100 = ($indvars$iv$next|0)<($99|0);
  if ($100) {
   $indvars$iv = $indvars$iv$next;
  } else {
   $$0 = 1;
   label = 30;
   break;
  }
 }
 if ((label|0) == 22) {
  $110 = ___cxa_find_matching_catch_2()|0;
  $111 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($110|0);
  // unreachable;
 }
 else if ((label|0) == 28) {
  $131 = ___cxa_find_matching_catch_2()|0;
  $132 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($131|0);
  // unreachable;
 }
 else if ((label|0) == 30) {
  STACKTOP = sp;return ($$0|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation11writeCustomEPKvh($this,$data,$customType) {
 $this = $this|0;
 $data = $data|0;
 $customType = $customType|0;
 var $$pre = 0, $$sum = 0, $$sum$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $lftr$wideiv = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Common14CustomTypeBase9serializeEPKvPhh($data,0,$customType)|0);
 $1 = $0&255;
 $2 = ($0&65535) >>> 8;
 $3 = $2&255;
 $4 = $0 << 16 >> 16;
 $5 = (($4) + 8)|0;
 $6 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($5)|0);
 $7 = ((($6)) + 8|0);
 $8 = $6;
 $9 = $8;
 HEAP32[$9>>2] = $4;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = 0;
 $12 = ($0<<16>>16)==(0);
 if (!($12)) {
  _memset(($7|0),0,($4|0))|0;
 }
 $13 = ((($this)) + 8|0);
 $14 = ((($this)) + 12|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = HEAP32[$13>>2]|0;
 $17 = (($15) + ($16)|0);
 HEAP8[$17>>0] = $3;
 $$sum$i = (($16) + 1)|0;
 $18 = (($15) + ($$sum$i)|0);
 HEAP8[$18>>0] = $1;
 $19 = HEAP32[$13>>2]|0;
 $20 = (($19) + 2)|0;
 HEAP32[$13>>2] = $20;
 (__ZN9ExitGames6Common14CustomTypeBase9serializeEPKvPhh($data,$7,$customType)|0);
 $21 = ($0<<16>>16)>(0);
 if (!($21)) {
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($6);
  return 1;
 }
 $$pre = HEAP32[$13>>2]|0;
 $25 = $$pre;$indvars$iv = 0;
 while(1) {
  $$sum = (($indvars$iv) + 8)|0;
  $22 = (($6) + ($$sum)|0);
  $23 = HEAP32[$14>>2]|0;
  $24 = (($23) + ($25)|0);
  $26 = HEAP8[$22>>0]|0;
  HEAP8[$24>>0] = $26;
  $27 = HEAP32[$13>>2]|0;
  $28 = (($27) + 1)|0;
  HEAP32[$13>>2] = $28;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $lftr$wideiv = $indvars$iv$next&65535;
  $exitcond = ($lftr$wideiv<<16>>16)==($0<<16>>16);
  if ($exitcond) {
   break;
  } else {
   $25 = $28;$indvars$iv = $indvars$iv$next;
  }
 }
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($6);
 return 1;
}
function __ZN9ExitGames6Common10UTF8StringC2ERKNS0_7JStringE($this,$wstr) {
 $this = $this|0;
 $wstr = $wstr|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (4140);
 __THREW__ = 0;
 (invoke_iii(829,($this|0),($wstr|0))|0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_2()|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___resumeException($2|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Common10UTF8String4sizeERKNS0_7JStringE($str) {
 $str = $str|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common7JString4cstrEv($str)|0);
 $1 = (__ZNK9ExitGames6Common7JString6lengthEv($str)|0);
 $2 = (__ZN9ExitGames6Common10UTF8String13UTF8Converter16SizeOfWStrAsUTF8EPKwj($0,$1)|0);
 return ($2|0);
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_9HashtableE($this,$hash) {
 $this = $this|0;
 $hash = $hash|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$05 = 0, $result$0$lcssa = 0, $result$04 = 0, $sext = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($hash)|0);
 $sext = $2 << 16;
 $3 = $sext >> 16;
 $4 = ($sext|0)>(0);
 if (!($4)) {
  $result$0$lcssa = 2;
  STACKTOP = sp;return ($result$0$lcssa|0);
 }
 $9 = 0;$i$05 = 0;$result$04 = 2;
 while(1) {
  $5 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($hash)|0);
  $6 = ((($5)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($7>>>0)>($9>>>0);
  if (!($8)) {
   __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
   __THREW__ = 0;
   $10 = (invoke_ii(764,($1|0))|0);
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if ($12) {
    label = 7;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($10|0),1,566,(10968|0),($vararg_buffer|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    label = 7;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($1);
  }
  $17 = ((($5)) + 16|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($18) + (($9*40)|0)|0);
  $20 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$19,1)|0);
  $21 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($hash)|0);
  $22 = ((($21)) + 4|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = ($23>>>0)>($9>>>0);
  if (!($24)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $25 = (invoke_ii(764,($0|0))|0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    label = 12;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($25|0),1,566,(10968|0),($vararg_buffer1|0));
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    label = 12;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  $32 = ((($21)) + 16|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = (($33) + (($9*40)|0)|0);
  $35 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($hash,$34)|0);
  $36 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$35,1)|0);
  $37 = (($20) + ($result$04))|0;
  $38 = (($37) + ($36))|0;
  $39 = (($i$05) + 1)<<16>>16;
  $40 = $39 << 16 >> 16;
  $41 = ($40|0)<($3|0);
  if ($41) {
   $9 = $40;$i$05 = $39;$result$04 = $38;
  } else {
   $result$0$lcssa = $38;
   label = 14;
   break;
  }
 }
 if ((label|0) == 7) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($15|0);
  // unreachable;
 }
 else if ((label|0) == 12) {
  $30 = ___cxa_find_matching_catch_2()|0;
  $31 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($30|0);
  // unreachable;
 }
 else if ((label|0) == 14) {
  STACKTOP = sp;return ($result$0$lcssa|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_14DictionaryBaseE($this,$dict) {
 $this = $this|0;
 $dict = $dict|0;
 var $$lcssa9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$0 = 0, $i1$03 = 0, $j$0 = 0, $result$0 = 0, $result$0$lcssa = 0, $result$1 = 0, $result$2$lcssa = 0, $result$24 = 0, $sext = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $j$0 = 0;$result$0 = 0;
 while(1) {
  $1 = ($j$0|0)==(0);
  if (!($1)) {
   $2 = (($j$0) + -1)|0;
   $3 = HEAP32[$dict>>2]|0;
   $4 = ((($3)) + 20|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = (FUNCTION_TABLE_ii[$5 & 2047]($dict)|0);
   $7 = (($6) + ($2)|0);
   $8 = HEAP8[$7>>0]|0;
   $9 = ($8<<24>>24)==(68);
   if (!($9)) {
    $result$0$lcssa = $result$0;
    break;
   }
  }
  $10 = (($result$0) + 1)|0;
  $i$0 = 0;$result$1 = $10;
  while(1) {
   $11 = HEAP32[$dict>>2]|0;
   $12 = ((($11)) + 24|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (FUNCTION_TABLE_ii[$13 & 2047]($dict)|0);
   $15 = (($14) + ($j$0<<2)|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ($i$0>>>0)<($16>>>0);
   $18 = (($result$1) + 1)|0;
   $19 = (($i$0) + 1)|0;
   if ($17) {
    $i$0 = $19;$result$1 = $18;
   } else {
    $$lcssa9 = $18;
    break;
   }
  }
  $20 = (($j$0) + 1)|0;
  $j$0 = $20;$result$0 = $$lcssa9;
 }
 $21 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($dict)|0);
 $22 = (($result$0$lcssa) + 2)|0;
 $23 = (__ZNK9ExitGames6Common9Hashtable7getSizeEv($21)|0);
 $sext = $23 << 16;
 $24 = $sext >> 16;
 $25 = ($sext|0)>(0);
 if (!($25)) {
  $result$2$lcssa = $22;
  STACKTOP = sp;return ($result$2$lcssa|0);
 }
 $30 = 0;$i1$03 = 0;$result$24 = $22;
 while(1) {
  $26 = (__ZNK9ExitGames6Common9Hashtable7getKeysEv($21)|0);
  $27 = ((($26)) + 4|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ($28>>>0)>($30>>>0);
  if (!($29)) {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $31 = (invoke_ii(764,($0|0))|0);
   $32 = __THREW__; __THREW__ = 0;
   $33 = $32&1;
   if ($33) {
    label = 13;
    break;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($31|0),1,566,(10968|0),($vararg_buffer|0));
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if ($35) {
    label = 13;
    break;
   }
   __ZN9ExitGames6Common7JStringD2Ev($0);
  }
  $38 = ((($26)) + 16|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = (($39) + (($30*40)|0)|0);
  $41 = (__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_($21,$40)|0);
  $42 = HEAP32[$dict>>2]|0;
  $43 = ((($42)) + 16|0);
  $44 = HEAP32[$43>>2]|0;
  $45 = (FUNCTION_TABLE_ii[$44 & 2047]($dict)|0);
  $46 = HEAP8[$45>>0]|0;
  $47 = ($46<<24>>24)==(122);
  $48 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$40,$47)|0);
  $49 = HEAP32[$dict>>2]|0;
  $50 = ((($49)) + 20|0);
  $51 = HEAP32[$50>>2]|0;
  $52 = (FUNCTION_TABLE_ii[$51 & 2047]($dict)|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = ($53<<24>>24)==(122);
  $55 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$41,$54)|0);
  $56 = (($48) + ($result$24))|0;
  $57 = (($56) + ($55))|0;
  $58 = (($i1$03) + 1)<<16>>16;
  $59 = $58 << 16 >> 16;
  $60 = ($59|0)<($24|0);
  if ($60) {
   $30 = $59;$i1$03 = $58;$result$24 = $57;
  } else {
   $result$2$lcssa = $57;
   label = 15;
   break;
  }
 }
 if ((label|0) == 13) {
  $36 = ___cxa_find_matching_catch_2()|0;
  $37 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($36|0);
  // unreachable;
 }
 else if ((label|0) == 15) {
  STACKTOP = sp;return ($result$2$lcssa|0);
 }
 return (0)|0;
}
function __ZNK9ExitGames6Common7Helpers24SerializerImplementation13calcArraySizeEPKvhhjPKsjRj($this,$pData,$type,$customType,$dimensions,$arraySizes,$recursionDepth,$result) {
 $this = $this|0;
 $pData = $pData|0;
 $type = $type|0;
 $customType = $customType|0;
 $dimensions = $dimensions|0;
 $arraySizes = $arraySizes|0;
 $recursionDepth = $recursionDepth|0;
 $result = $result|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, $i$011 = 0, $i1$014 = 0, $i2$016 = 0, $i3$018 = 0, $i4$020 = 0, $i5$022 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($pData|0)==(0|0);
 if ($0) {
  $1 = (($arraySizes) + ($recursionDepth<<1)|0);
  $2 = HEAP16[$1>>1]|0;
  $3 = ($2<<16>>16)==(0);
  if (!($3)) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $4 = HEAP32[$result>>2]|0;
 $5 = (($4) + 2)|0;
 $6 = ($dimensions|0)!=(0);
 $7 = (($dimensions) + -1)|0;
 $8 = ($7>>>0)>($recursionDepth>>>0);
 $or$cond = $6 & $8;
 $9 = (($4) + 3)|0;
 HEAP32[$result>>2] = $9;
 $10 = (($arraySizes) + ($recursionDepth<<1)|0);
 $11 = HEAP16[$10>>1]|0;
 if ($or$cond) {
  $12 = ($11<<16>>16)>(0);
  if (!($12)) {
   $$0 = 1;
   return ($$0|0);
  }
  $13 = (($recursionDepth) + 1)|0;
  $i$011 = 0;
  while(1) {
   $17 = $i$011 << 16 >> 16;
   $18 = (($pData) + ($17<<2)|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13calcArraySizeEPKvhhjPKsjRj($this,$19,$type,$customType,$dimensions,$arraySizes,$13,$result)|0);
   $16 = (($i$011) + 1)<<16>>16;
   if (!($20)) {
    $$0 = 0;
    label = 27;
    break;
   }
   $14 = HEAP16[$10>>1]|0;
   $15 = ($16<<16>>16)<($14<<16>>16);
   if ($15) {
    $i$011 = $16;
   } else {
    $$0 = 1;
    label = 27;
    break;
   }
  }
  if ((label|0) == 27) {
   return ($$0|0);
  }
 }
 $21 = $11 << 16 >> 16;
 $22 = ($11<<16>>16)>(0);
 if (!($22)) {
  $$0 = 1;
  return ($$0|0);
 }
 $23 = $type&255;
 do {
  switch ($23|0) {
  case 98:  {
   $26 = (($21) + ($9))|0;
   HEAP32[$result>>2] = $26;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 107:  {
   $27 = $21 << 1;
   $28 = (($27) + ($9))|0;
   HEAP32[$result>>2] = $28;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 105:  {
   $29 = $21 << 2;
   $30 = (($29) + ($9))|0;
   HEAP32[$result>>2] = $30;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 108:  {
   $31 = $21 << 3;
   $32 = (($31) + ($9))|0;
   HEAP32[$result>>2] = $32;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 102:  {
   $33 = $21 << 2;
   $34 = (($33) + ($9))|0;
   HEAP32[$result>>2] = $34;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 100:  {
   $35 = $21 << 3;
   $36 = (($35) + ($9))|0;
   HEAP32[$result>>2] = $36;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 111:  {
   $37 = (($21) + ($9))|0;
   HEAP32[$result>>2] = $37;
   $$0 = 1;
   return ($$0|0);
   break;
  }
  case 115:  {
   $i1$014 = 0;
   while(1) {
    $38 = $i1$014 << 16 >> 16;
    $39 = (($pData) + ($38<<4)|0);
    $40 = (__ZN9ExitGames6Common10UTF8String4sizeERKNS0_7JStringE($39)|0);
    $41 = (($40) + 2)|0;
    $42 = HEAP32[$result>>2]|0;
    $43 = (($41) + ($42))|0;
    HEAP32[$result>>2] = $43;
    $44 = (($i1$014) + 1)<<16>>16;
    $45 = HEAP16[$10>>1]|0;
    $46 = ($44<<16>>16)<($45<<16>>16);
    if ($46) {
     $i1$014 = $44;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
   break;
  }
  case 104:  {
   $i2$016 = 0;
   while(1) {
    $47 = $i2$016 << 16 >> 16;
    $48 = (($pData) + (($47*44)|0)|0);
    $49 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_9HashtableE($this,$48)|0);
    $50 = HEAP32[$result>>2]|0;
    $51 = (($50) + ($49))|0;
    HEAP32[$result>>2] = $51;
    $52 = (($i2$016) + 1)<<16>>16;
    $53 = HEAP16[$10>>1]|0;
    $54 = ($52<<16>>16)<($53<<16>>16);
    if ($54) {
     $i2$016 = $52;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
   break;
  }
  case 68:  {
   $i3$018 = 0;
   while(1) {
    $55 = $i3$018 << 16 >> 16;
    $56 = (($pData) + (($55*52)|0)|0);
    $57 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation11getTypeSizeEPKNS0_14DictionaryBaseE($this,$56)|0);
    $58 = HEAP32[$result>>2]|0;
    $59 = (($58) + ($57))|0;
    HEAP32[$result>>2] = $59;
    $60 = (($i3$018) + 1)<<16>>16;
    $61 = HEAP16[$10>>1]|0;
    $62 = ($60<<16>>16)<($61<<16>>16);
    if ($62) {
     $i3$018 = $60;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
   break;
  }
  case 99:  {
   $24 = $customType&255;
   $25 = (162720 + ($24<<2)|0);
   $i4$020 = 0;
   while(1) {
    $63 = $i4$020 << 16 >> 16;
    $64 = HEAP32[$25>>2]|0;
    $65 = (__ZN9ExitGames6Common7Helpers25CustomTypeCallbackWrapper9getSizeofEh($64,$customType)|0);
    $66 = Math_imul($65, $63)|0;
    $67 = (($pData) + ($66)|0);
    $68 = (__ZN9ExitGames6Common14CustomTypeBase9serializeEPKvPhh($67,0,$customType)|0);
    $69 = $68 << 16 >> 16;
    $70 = (($69) + 2)|0;
    $71 = HEAP32[$result>>2]|0;
    $72 = (($70) + ($71))|0;
    HEAP32[$result>>2] = $72;
    $73 = (($i4$020) + 1)<<16>>16;
    $74 = HEAP16[$10>>1]|0;
    $75 = ($73<<16>>16)<($74<<16>>16);
    if ($75) {
     $i4$020 = $73;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
   break;
  }
  case 122:  {
   HEAP32[$result>>2] = $5;
   $i5$022 = 0;
   while(1) {
    $76 = $i5$022 << 16 >> 16;
    $77 = (($pData) + (($76*40)|0)|0);
    $78 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getObjectSizeEPKNS0_6ObjectEb($this,$77,1)|0);
    $79 = HEAP32[$result>>2]|0;
    $80 = (($79) + ($78))|0;
    HEAP32[$result>>2] = $80;
    $81 = (($i5$022) + 1)<<16>>16;
    $82 = HEAP16[$10>>1]|0;
    $83 = ($81<<16>>16)<($82<<16>>16);
    if ($83) {
     $i5$022 = $81;
    } else {
     $$0 = 1;
     break;
    }
   }
   return ($$0|0);
   break;
  }
  default: {
   $$0 = 0;
   return ($$0|0);
  }
  }
 } while(0);
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers24SerializerImplementation17writeInvertedDataEPKvi($this,$data,$size) {
 $this = $this|0;
 $data = $data|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$03 = 0, $i$03$in = 0, $ptrForWriting$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = ($size|0)>(0);
 if ($1) {
  $2 = ((($this)) + 12|0);
  $3 = HEAP32[$2>>2]|0;
  $4 = HEAP32[$0>>2]|0;
  $5 = (($3) + ($4)|0);
  $i$03$in = $size;$ptrForWriting$02 = $5;
  while(1) {
   $i$03 = (($i$03$in) + -1)|0;
   $6 = (($data) + ($i$03)|0);
   $7 = HEAP8[$6>>0]|0;
   HEAP8[$ptrForWriting$02>>0] = $7;
   $8 = ((($ptrForWriting$02)) + 1|0);
   $9 = ($i$03$in|0)>(1);
   if ($9) {
    $i$03$in = $i$03;$ptrForWriting$02 = $8;
   } else {
    break;
   }
  }
 }
 $10 = HEAP32[$0>>2]|0;
 $11 = (($10) + ($size))|0;
 HEAP32[$0>>2] = $11;
 return;
}
function __ZN9ExitGames6Common7Helpers13ValueToObject3getENS0_6ObjectE($agg$result,$value) {
 $agg$result = $agg$result|0;
 $value = $value|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common6ObjectC2ERKS1_($agg$result,$value);
 return;
}
function __ZN9ExitGames6Common11ValueObjectIPhED0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common6ObjectD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Common11ValueObjectIPhEaSERKNS0_6ObjectE($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 2047]($this,$toCopy)|0);
 return ($3|0);
}
function __ZN9ExitGames6Common11ValueObjectIPhE6assignERKNS0_6ObjectE($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNK9ExitGames6Common6Object7getTypeEv($toCopy)|0);
 $1 = ($0<<24>>24)==(98);
 if (!($1)) {
  return ($this|0);
 }
 (__ZN9ExitGames6Common6Object6assignERKS1_($this,$toCopy)|0);
 return ($this|0);
}
function __ZN9ExitGames6Common11ValueObjectIPhEaSERKS3_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_iii[$2 & 2047]($this,$toCopy)|0);
 return ($3|0);
}
function __ZN9ExitGames6Common7JStringaSEt($this,$aNum) {
 $this = $this|0;
 $aNum = $aNum|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP16[$0>>1] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignItEERS1_RKT_jPKw($this,$0,6,6212)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JString6assignItEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP16[$Rhs>>1]|0;
 $10 = $9&65535;
 HEAP32[$vararg_buffer>>2] = $10;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $11 = ((($this)) + 4|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0|0);
 if (!($13)) {
  $14 = ((($12)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($14);
 }
 $15 = (_wcslen($3)|0);
 $16 = ((($this)) + 12|0);
 HEAP32[$16>>2] = $15;
 $17 = ((($this)) + 8|0);
 HEAP32[$17>>2] = $15;
 $18 = (($15) + 1)|0;
 $19 = $18 << 2;
 $20 = (($19) + 8)|0;
 $21 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($20)|0);
 $22 = ((($21)) + 8|0);
 $23 = $21;
 $24 = $23;
 HEAP32[$24>>2] = $18;
 $25 = (($23) + 4)|0;
 $26 = $25;
 HEAP32[$26>>2] = 0;
 $27 = ($18|0)==(0);
 if ($27) {
  HEAP32[$11>>2] = $22;
  (_wcscpy($22,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($22|0),0,($19|0))|0;
 HEAP32[$11>>2] = $22;
 (_wcscpy($22,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZN9ExitGames6Common7JStringaSEj($this,$aNum) {
 $this = $this|0;
 $aNum = $aNum|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$0>>2] = $aNum;
 $1 = (__ZN9ExitGames6Common7JString6assignIjEERS1_RKT_jPKw($this,$0,11,9720)|0);
 STACKTOP = sp;return ($1|0);
}
function __ZN9ExitGames6Common7JString6assignIjEERS1_RKT_jPKw($this,$Rhs,$bufferLen,$formatSpecifier) {
 $this = $this|0;
 $Rhs = $Rhs|0;
 $bufferLen = $bufferLen|0;
 $formatSpecifier = $formatSpecifier|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = $bufferLen << 2;
 $1 = (($0) + 8)|0;
 $2 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($1)|0);
 $3 = ((($2)) + 8|0);
 $4 = $2;
 $5 = $4;
 HEAP32[$5>>2] = $bufferLen;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ($bufferLen|0)==(0);
 if (!($8)) {
  _memset(($3|0),0,($0|0))|0;
 }
 $9 = HEAP32[$Rhs>>2]|0;
 HEAP32[$vararg_buffer>>2] = $9;
 (__Z11EG_swprintfPwjPKwz($3,$bufferLen,$formatSpecifier,$vararg_buffer)|0);
 $10 = ((($this)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0|0);
 if (!($12)) {
  $13 = ((($11)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($13);
 }
 $14 = (_wcslen($3)|0);
 $15 = ((($this)) + 12|0);
 HEAP32[$15>>2] = $14;
 $16 = ((($this)) + 8|0);
 HEAP32[$16>>2] = $14;
 $17 = (($14) + 1)|0;
 $18 = $17 << 2;
 $19 = (($18) + 8)|0;
 $20 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($19)|0);
 $21 = ((($20)) + 8|0);
 $22 = $20;
 $23 = $22;
 HEAP32[$23>>2] = $17;
 $24 = (($22) + 4)|0;
 $25 = $24;
 HEAP32[$25>>2] = 0;
 $26 = ($17|0)==(0);
 if ($26) {
  HEAP32[$10>>2] = $21;
  (_wcscpy($21,$3)|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
  STACKTOP = sp;return ($this|0);
 }
 _memset(($21|0),0,($18|0))|0;
 HEAP32[$10>>2] = $21;
 (_wcscpy($21,$3)|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($2);
 STACKTOP = sp;return ($this|0);
}
function __ZNK9ExitGames6Common7JStringcvPKwEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Common7JString7indexOfEc($this,$ch) {
 $this = $this|0;
 $ch = $ch|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $buf$i = sp + 12|0;
 $0 = sp;
 $1 = ((($this)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $$0$i = -1;
  STACKTOP = sp;return ($$0$i|0);
 }
 HEAP8[$buf$i>>0] = $ch;
 $4 = ((($buf$i)) + 1|0);
 HEAP8[$4>>0] = 0;
 $5 = ((($this)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 __ZN9ExitGames6Common10UTF8StringC2EPKc($0,$buf$i);
 __THREW__ = 0;
 $7 = (invoke_ii(826,($0|0))|0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common10UTF8StringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 $10 = HEAP8[$7>>0]|0;
 $11 = $10 << 24 >> 24;
 __THREW__ = 0;
 $12 = (invoke_iii(865,($6|0),($11|0))|0);
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common10UTF8StringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 __ZN9ExitGames6Common10UTF8StringD2Ev($0);
 $15 = ($12|0)==(0|0);
 if ($15) {
  $$0$i = -1;
  STACKTOP = sp;return ($$0$i|0);
 }
 $18 = HEAP32[$5>>2]|0;
 $19 = $12;
 $20 = $18;
 $21 = (($19) - ($20))|0;
 $22 = $21 >> 2;
 $$0$i = $22;
 STACKTOP = sp;return ($$0$i|0);
}
function __ZNK9ExitGames6Common7JString7indexOfERKS1_($this,$str) {
 $this = $this|0;
 $str = $str|0;
 var $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0$i = -1;
  return ($$0$i|0);
 }
 $3 = ((($this)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($str)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_wcsstr($4,$6)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0$i = -1;
  return ($$0$i|0);
 }
 $9 = HEAP32[$3>>2]|0;
 $10 = $7;
 $11 = $9;
 $12 = (($10) - ($11))|0;
 $13 = $12 >> 2;
 $$0$i = $13;
 return ($$0$i|0);
}
function __ZNK9ExitGames6Common7JString10startsWithERKS1_($this,$prefix) {
 $this = $this|0;
 $prefix = $prefix|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($prefix)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1>>>0)<($3>>>0);
 if ($4) {
  $11 = 0;
  return ($11|0);
 }
 $5 = ((($this)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($prefix)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (_wcsncmp($6,$8,$3)|0);
 $10 = ($9|0)==(0);
 $11 = $10;
 return ($11|0);
}
function __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common10UTF8StringC2ERKNS0_7JStringE($agg$result,$this);
 return;
}
function __ZNK9ExitGames6Common7JString18ANSIRepresentationEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common10ANSIStringC2ERKNS0_7JStringE($agg$result,$this);
 return;
}
function __ZNK9ExitGames6Common6Object14getDataPointerEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $3 = ((($this)) + 16|0);
 if (!($2)) {
  $10 = $3;
  return ($10|0);
 }
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if (!($5)) {
  $10 = $4;
  return ($10|0);
 }
 $6 = ((($this)) + 28|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)==(42);
 if ($8) {
  $10 = 0;
  return ($10|0);
 }
 $9 = ((($this)) + 8|0);
 $10 = $9;
 return ($10|0);
}
function __ZNK9ExitGames6Common10Serializer7getDataEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation7getDataEv($0)|0);
 return ($1|0);
}
function __ZNK9ExitGames6Common10Serializer7getSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = (__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getDataOffsetEv($0)|0);
 return ($1|0);
}
function __ZN9ExitGames6Common10UTF8StringC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common14BaseCharStringC2Ev($this);
 HEAP32[$this>>2] = (4140);
 __THREW__ = 0;
 (invoke_iii(809,($this|0),(169457|0))|0);
 $0 = __THREW__; __THREW__ = 0;
 $1 = $0&1;
 if ($1) {
  $2 = ___cxa_find_matching_catch_2()|0;
  $3 = tempRet0;
  __ZN9ExitGames6Common14BaseCharStringD2Ev($this);
  ___resumeException($2|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Photon8Internal13SocketConnect7serviceEv($this) {
 $this = $this|0;
 var $$$i = 0, $$phi$trans$insert = 0, $$pre = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $bytesReceived = 0, $or$cond = 0, $pRecvBuffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $bytesReceived = sp + 4|0;
 $pRecvBuffer = sp;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 L1: do {
  switch ($1|0) {
  case 5:  {
   break;
  }
  case 3:  {
   $34 = HEAP32[$this>>2]|0;
   $35 = ((($34)) + 56|0);
   $36 = HEAP32[$35>>2]|0;
   $37 = (FUNCTION_TABLE_ii[$36 & 2047]($this)|0);
   if ($37) {
    HEAP32[$0>>2] = 5;
    $38 = ((($this)) + 4|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = HEAP32[$39>>2]|0;
    $41 = ((($40)) + 8|0);
    $42 = HEAP32[$41>>2]|0;
    FUNCTION_TABLE_vii[$42 & 2047]($39,0);
    break L1;
   } else {
    STACKTOP = sp;return;
   }
   break;
  }
  case 2:  {
   $2 = ((($this)) + 20|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)<(0);
   if ($4) {
    $5 = HEAP32[$this>>2]|0;
    $6 = ((($5)) + 52|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = (FUNCTION_TABLE_ii[$7 & 2047]($this)|0);
    HEAP32[$2>>2] = $8;
    $9 = ($8|0)<(0);
    if ($9) {
     $10 = ((($this)) + 16|0);
     HEAP32[$10>>2] = 1;
    } else {
     $16 = $8;
     label = 5;
    }
   } else {
    $16 = $3;
    label = 5;
   }
   L10: do {
    if ((label|0) == 5) {
     $11 = ((($this)) + 48|0);
     $12 = ((($this)) + 76|0);
     $13 = HEAP8[$12>>0]|0;
     $14 = ($13<<24>>24)!=(0);
     $15 = $14 ? 28 : 16;
     $17 = (_connect($16,$11,$15)|0);
     $18 = ($17|0)<(0);
     L12: do {
      if ($18) {
       $19 = (___errno_location()|0);
       $20 = HEAP32[$19>>2]|0;
       switch ($20|0) {
       case 11: case 115:  {
        break L12;
        break;
       }
       default: {
       }
       }
       $21 = HEAP32[$this>>2]|0;
       $22 = ((($21)) + 8|0);
       $23 = HEAP32[$22>>2]|0;
       (FUNCTION_TABLE_ii[$23 & 2047]($this)|0);
       $24 = $20 & -2;
       $25 = ($24|0)==(100);
       $26 = ((($this)) + 16|0);
       $$$i = $25 ? 534 : -1;
       HEAP32[$26>>2] = $$$i;
       break L10;
      }
     } while(0);
     HEAP32[$0>>2] = 3;
     $27 = ((($this)) + 16|0);
     HEAP32[$27>>2] = 0;
     STACKTOP = sp;return;
    }
   } while(0);
   HEAP32[$0>>2] = 0;
   label = 12;
   break;
  }
  case 4:  {
   $$phi$trans$insert = ((($this)) + 16|0);
   $$pre = HEAP32[$$phi$trans$insert>>2]|0;
   HEAP32[$0>>2] = 0;
   $28 = ($$pre|0)==(0);
   if ($28) {
    HEAP32[$$phi$trans$insert>>2] = -1;
    label = 12;
   } else {
    label = 12;
   }
   break;
  }
  default: {
   STACKTOP = sp;return;
  }
  }
 } while(0);
 if ((label|0) == 12) {
  $29 = ((($this)) + 4|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($31)) + 8|0);
  $33 = HEAP32[$32>>2]|0;
  FUNCTION_TABLE_vii[$33 & 2047]($30,-1);
  STACKTOP = sp;return;
 }
 $43 = ((($this)) + 24|0);
 $44 = HEAP32[$43>>2]|0;
 L25: do {
  switch ($44|0) {
  case 3:  {
   $45 = ((($this)) + 16|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = ($46|0)==(0);
   if ($47) {
    HEAP32[$45>>2] = -1;
    $53 = -1;
   } else {
    $53 = $46;
   }
   $48 = ((($this)) + 4|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = HEAP32[$49>>2]|0;
   $51 = ((($50)) + 16|0);
   $52 = HEAP32[$51>>2]|0;
   FUNCTION_TABLE_vii[$52 & 2047]($49,$53);
   HEAP32[$43>>2] = 0;
   break;
  }
  case 2:  {
   $54 = ((($this)) + 16|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = ($55|0)==(0);
   if (!($56)) {
    HEAP32[$54>>2] = 0;
   }
   $57 = ((($this)) + 4|0);
   $58 = HEAP32[$57>>2]|0;
   $59 = HEAP32[$58>>2]|0;
   $60 = ((($59)) + 16|0);
   $61 = HEAP32[$60>>2]|0;
   FUNCTION_TABLE_vii[$61 & 2047]($58,0);
   HEAP32[$43>>2] = 0;
   break;
  }
  case 1:  {
   $62 = ((($this)) + 36|0);
   $63 = HEAP32[$62>>2]|0;
   $64 = HEAP32[$this>>2]|0;
   $65 = ((($64)) + 48|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = ((($this)) + 28|0);
   $68 = HEAP32[$67>>2]|0;
   $69 = (FUNCTION_TABLE_iiii[$66 & 2047]($this,$68,$62)|0);
   $70 = HEAP32[$62>>2]|0;
   $71 = ($70|0)!=(0);
   $72 = ($69|0)==(0);
   $or$cond = $72 & $71;
   if ($or$cond) {
    $73 = HEAP32[$67>>2]|0;
    $$sum = (($63) - ($70))|0;
    $74 = (($73) + ($$sum)|0);
    _memcpy(($73|0),($74|0),($70|0))|0;
    break L25;
   }
   switch ($69|0) {
   case 0: case -1:  {
    break;
   }
   default: {
    break L25;
   }
   }
   $75 = ((($this)) + 4|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = HEAP32[$76>>2]|0;
   $78 = ((($77)) + 16|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = ((($this)) + 16|0);
   $81 = HEAP32[$80>>2]|0;
   FUNCTION_TABLE_vii[$79 & 2047]($76,$81);
   HEAP32[$43>>2] = 0;
   break;
  }
  default: {
  }
  }
 } while(0);
 $82 = ((($this)) + 4|0);
 while(1) {
  $83 = HEAP32[$0>>2]|0;
  $84 = ($83|0)==(5);
  if (!($84)) {
   label = 33;
   break;
  }
  HEAP32[$bytesReceived>>2] = 0;
  HEAP32[$pRecvBuffer>>2] = 0;
  $85 = HEAP32[$this>>2]|0;
  $86 = ((($85)) + 44|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = (FUNCTION_TABLE_iiii[$87 & 2047]($this,$pRecvBuffer,$bytesReceived)|0);
  switch ($88|0) {
  case -1: case 0: case 540:  {
   $89 = HEAP32[$82>>2]|0;
   $90 = HEAP32[$89>>2]|0;
   $91 = ((($90)) + 12|0);
   $92 = HEAP32[$91>>2]|0;
   $93 = HEAP32[$pRecvBuffer>>2]|0;
   $94 = HEAP32[$bytesReceived>>2]|0;
   FUNCTION_TABLE_viiii[$92 & 2047]($89,$93,$94,$88);
   break;
  }
  default: {
  }
  }
  $95 = HEAP32[$pRecvBuffer>>2]|0;
  $96 = ($95|0)==(0|0);
  if (!($96)) {
   $98 = ((($95)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($98);
  }
  $97 = ($88|0)==(0);
  if (!($97)) {
   label = 33;
   break;
  }
 }
 if ((label|0) == 33) {
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Photon8Internal13SocketConnect15startConnectionERKNS_6Common7JStringES6_($this,$ipAddr,$appID) {
 $this = $this|0;
 $ipAddr = $ipAddr|0;
 $appID = $appID|0;
 var $$0 = 0, $$044 = 0, $$054 = 0, $$065 = 0, $$1064$ph = 0, $$1075$ph = 0, $$11 = 0, $$1176 = 0, $$12 = 0, $$1277 = 0, $$155 = 0, $$166 = 0, $$256131 = 0, $$267130 = 0, $$357$ph = 0, $$368$ph = 0, $$458$ph = 0, $$469$ph = 0, $$559$ph = 0, $$570$ph = 0;
 var $$660$ph = 0, $$671$ph = 0, $$761$ph = 0, $$772$ph = 0, $$862$ph = 0, $$873$ph = 0, $$963$ph = 0, $$974$ph = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $eh$lpad$body$index17Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body115$index22Z2D = 0, $eh$lpad$body115$indexZ2D = 0, $eh$lpad$body118$index27Z2D = 0, $eh$lpad$body118$indexZ2D = 0, $eh$lpad$body121$index2Z2D = 0, $eh$lpad$body121$indexZ2D = 0, $eh$lpad$body125$index7Z2D = 0, $eh$lpad$body125$indexZ2D = 0, $eh$lpad$body129$index12Z2D = 0, $eh$lpad$body129$indexZ2D = 0, $str$i = 0, $str$i122 = 0, $str$i126 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(336|0);
 $str$i126 = sp + 304|0;
 $str$i122 = sp + 288|0;
 $str$i = sp + 272|0;
 $0 = sp + 256|0;
 $1 = sp + 240|0;
 $2 = sp + 224|0;
 $3 = sp + 208|0;
 $4 = sp + 192|0;
 $5 = sp + 176|0;
 $6 = sp + 160|0;
 $7 = sp + 144|0;
 $8 = sp + 128|0;
 $9 = sp + 112|0;
 $10 = sp + 96|0;
 $11 = sp + 80|0;
 $12 = sp + 64|0;
 $13 = sp + 48|0;
 $14 = sp + 32|0;
 $15 = sp + 16|0;
 $16 = sp + 320|0;
 $17 = sp;
 $18 = ((($this)) + 8|0);
 HEAP32[$18>>2] = 4;
 $19 = (__ZNK9ExitGames6Common7JString6lengthEv($ipAddr)|0);
 $20 = ($19|0)==(0);
 if ($20) {
  $21 = ((($this)) + 16|0);
  HEAP32[$21>>2] = 14;
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 __ZN9ExitGames6Photon8Internal13SocketConnect16addPortToAddressERKNS_6Common7JStringE($5,$ipAddr);
 __THREW__ = 0;
 $22 = (invoke_ii(820,($appID|0))|0);
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 L5: do {
  if ($24) {
   label = 46;
  } else {
   $25 = ($22|0)==(0);
   L7: do {
    if ($25) {
     __THREW__ = 0;
     invoke_vii(801,($6|0),(163772|0));
     $90 = __THREW__; __THREW__ = 0;
     $91 = $90&1;
     if ($91) {
      $128 = ___cxa_find_matching_catch_2()|0;
      $129 = tempRet0;
      $$11 = $129;$$1176 = $128;
      break L5;
     } else {
      $$044 = 0;
      label = 39;
     }
    } else {
     __THREW__ = 0;
     invoke_viii(860,($14|0),(6296|0),($appID|0));
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      label = 46;
      break L5;
     }
     __THREW__ = 0;
     invoke_vii(786,($2|0),($14|0));
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $106 = ___cxa_find_matching_catch_2()|0;
      $107 = tempRet0;
      $eh$lpad$body$index17Z2D = $107;$eh$lpad$body$indexZ2D = $106;
     } else {
      __THREW__ = 0;
      invoke_vii(787,($str$i|0),0);
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      do {
       if ($31) {
        label = 12;
       } else {
        __THREW__ = 0;
        (invoke_iii(759,($str$i|0),(6304|0))|0);
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        if (!($33)) {
         __THREW__ = 0;
         $34 = (invoke_iii(760,($2|0),($str$i|0))|0);
         $35 = __THREW__; __THREW__ = 0;
         $36 = $35&1;
         if (!($36)) {
          __ZN9ExitGames6Common7JStringD2Ev($str$i);
          __THREW__ = 0;
          invoke_vii(786,($13|0),($34|0));
          $39 = __THREW__; __THREW__ = 0;
          $40 = $39&1;
          if ($40) {
           label = 12;
           break;
          }
          __ZN9ExitGames6Common7JStringD2Ev($2);
          __THREW__ = 0;
          invoke_viii(866,($12|0),($13|0),(146477|0));
          $43 = __THREW__; __THREW__ = 0;
          $44 = $43&1;
          if ($44) {
           $108 = ___cxa_find_matching_catch_2()|0;
           $109 = tempRet0;
           $$963$ph = $109;$$974$ph = $108;
           label = 71;
           break L7;
          }
          __THREW__ = 0;
          invoke_vii(786,($1|0),($12|0));
          $45 = __THREW__; __THREW__ = 0;
          $46 = $45&1;
          if ($46) {
           $110 = ___cxa_find_matching_catch_2()|0;
           $111 = tempRet0;
           $eh$lpad$body115$index22Z2D = $111;$eh$lpad$body115$indexZ2D = $110;
          } else {
           __THREW__ = 0;
           invoke_vii(787,($str$i122|0),0);
           $47 = __THREW__; __THREW__ = 0;
           $48 = $47&1;
           L27: do {
            if ($48) {
             label = 21;
            } else {
             __THREW__ = 0;
             (invoke_iii(759,($str$i122|0),(13236|0))|0);
             $49 = __THREW__; __THREW__ = 0;
             $50 = $49&1;
             do {
              if (!($50)) {
               __THREW__ = 0;
               $51 = (invoke_iii(760,($1|0),($str$i122|0))|0);
               $52 = __THREW__; __THREW__ = 0;
               $53 = $52&1;
               if ($53) {
                break;
               }
               __ZN9ExitGames6Common7JStringD2Ev($str$i122);
               __THREW__ = 0;
               invoke_vii(786,($11|0),($51|0));
               $56 = __THREW__; __THREW__ = 0;
               $57 = $56&1;
               if ($57) {
                label = 21;
                break L27;
               }
               __ZN9ExitGames6Common7JStringD2Ev($1);
               __THREW__ = 0;
               invoke_viii(866,($10|0),($11|0),(146478|0));
               $60 = __THREW__; __THREW__ = 0;
               $61 = $60&1;
               if ($61) {
                $112 = ___cxa_find_matching_catch_2()|0;
                $113 = tempRet0;
                $$761$ph = $113;$$772$ph = $112;
                label = 69;
                break L7;
               }
               __THREW__ = 0;
               invoke_vii(801,($15|0),(6348|0));
               $62 = __THREW__; __THREW__ = 0;
               $63 = $62&1;
               if ($63) {
                $114 = ___cxa_find_matching_catch_2()|0;
                $115 = tempRet0;
                $$660$ph = $115;$$671$ph = $114;
                label = 68;
                break L7;
               }
               __THREW__ = 0;
               invoke_viii(785,($9|0),($10|0),($15|0));
               $64 = __THREW__; __THREW__ = 0;
               $65 = $64&1;
               if ($65) {
                $116 = ___cxa_find_matching_catch_2()|0;
                $117 = tempRet0;
                $$559$ph = $117;$$570$ph = $116;
                label = 67;
                break L7;
               }
               __THREW__ = 0;
               $66 = (invoke_i(867)|0);
               $67 = __THREW__; __THREW__ = 0;
               $68 = $67&1;
               do {
                if (!($68)) {
                 HEAP8[$16>>0] = $66;
                 __THREW__ = 0;
                 invoke_viii(866,($8|0),($9|0),($16|0));
                 $69 = __THREW__; __THREW__ = 0;
                 $70 = $69&1;
                 if ($70) {
                  break;
                 }
                 __THREW__ = 0;
                 invoke_vii(786,($0|0),($8|0));
                 $71 = __THREW__; __THREW__ = 0;
                 $72 = $71&1;
                 if ($72) {
                  $120 = ___cxa_find_matching_catch_2()|0;
                  $121 = tempRet0;
                  $eh$lpad$body118$index27Z2D = $121;$eh$lpad$body118$indexZ2D = $120;
                 } else {
                  __THREW__ = 0;
                  invoke_vii(787,($str$i126|0),0);
                  $73 = __THREW__; __THREW__ = 0;
                  $74 = $73&1;
                  L48: do {
                   if ($74) {
                    label = 34;
                   } else {
                    __THREW__ = 0;
                    (invoke_iii(759,($str$i126|0),(6380|0))|0);
                    $75 = __THREW__; __THREW__ = 0;
                    $76 = $75&1;
                    do {
                     if (!($76)) {
                      __THREW__ = 0;
                      $77 = (invoke_iii(760,($0|0),($str$i126|0))|0);
                      $78 = __THREW__; __THREW__ = 0;
                      $79 = $78&1;
                      if ($79) {
                       break;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($str$i126);
                      __THREW__ = 0;
                      invoke_vii(786,($7|0),($77|0));
                      $82 = __THREW__; __THREW__ = 0;
                      $83 = $82&1;
                      if ($83) {
                       label = 34;
                       break L48;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($0);
                      __THREW__ = 0;
                      invoke_viii(785,($6|0),($7|0),(163740|0));
                      $86 = __THREW__; __THREW__ = 0;
                      $87 = $86&1;
                      if (!($87)) {
                       $$044 = 1;
                       label = 39;
                       break L7;
                      }
                      $88 = ___cxa_find_matching_catch_2()|0;
                      $89 = tempRet0;
                      $$256131 = $89;$$267130 = $88;
                      label = 64;
                      break L7;
                     }
                    } while(0);
                    $80 = ___cxa_find_matching_catch_2()|0;
                    $81 = tempRet0;
                    __ZN9ExitGames6Common7JStringD2Ev($str$i126);
                    $eh$lpad$body129$index12Z2D = $81;$eh$lpad$body129$indexZ2D = $80;
                   }
                  } while(0);
                  if ((label|0) == 34) {
                   $84 = ___cxa_find_matching_catch_2()|0;
                   $85 = tempRet0;
                   $eh$lpad$body129$index12Z2D = $85;$eh$lpad$body129$indexZ2D = $84;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($0);
                  $eh$lpad$body118$index27Z2D = $eh$lpad$body129$index12Z2D;$eh$lpad$body118$indexZ2D = $eh$lpad$body129$indexZ2D;
                 }
                 $$357$ph = $eh$lpad$body118$index27Z2D;$$368$ph = $eh$lpad$body118$indexZ2D;
                 label = 65;
                 break L7;
                }
               } while(0);
               $118 = ___cxa_find_matching_catch_2()|0;
               $119 = tempRet0;
               $$458$ph = $119;$$469$ph = $118;
               label = 66;
               break L7;
              }
             } while(0);
             $54 = ___cxa_find_matching_catch_2()|0;
             $55 = tempRet0;
             __ZN9ExitGames6Common7JStringD2Ev($str$i122);
             $eh$lpad$body125$index7Z2D = $55;$eh$lpad$body125$indexZ2D = $54;
            }
           } while(0);
           if ((label|0) == 21) {
            $58 = ___cxa_find_matching_catch_2()|0;
            $59 = tempRet0;
            $eh$lpad$body125$index7Z2D = $59;$eh$lpad$body125$indexZ2D = $58;
           }
           __ZN9ExitGames6Common7JStringD2Ev($1);
           $eh$lpad$body115$index22Z2D = $eh$lpad$body125$index7Z2D;$eh$lpad$body115$indexZ2D = $eh$lpad$body125$indexZ2D;
          }
          $$862$ph = $eh$lpad$body115$index22Z2D;$$873$ph = $eh$lpad$body115$indexZ2D;
          label = 70;
          break L7;
         }
        }
        $37 = ___cxa_find_matching_catch_2()|0;
        $38 = tempRet0;
        __ZN9ExitGames6Common7JStringD2Ev($str$i);
        $eh$lpad$body121$index2Z2D = $38;$eh$lpad$body121$indexZ2D = $37;
       }
      } while(0);
      if ((label|0) == 12) {
       $41 = ___cxa_find_matching_catch_2()|0;
       $42 = tempRet0;
       $eh$lpad$body121$index2Z2D = $42;$eh$lpad$body121$indexZ2D = $41;
      }
      __ZN9ExitGames6Common7JStringD2Ev($2);
      $eh$lpad$body$index17Z2D = $eh$lpad$body121$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body121$indexZ2D;
     }
     $$1064$ph = $eh$lpad$body$index17Z2D;$$1075$ph = $eh$lpad$body$indexZ2D;
    }
   } while(0);
   if ((label|0) == 39) {
    __THREW__ = 0;
    invoke_viii(785,($4|0),($5|0),($6|0));
    $92 = __THREW__; __THREW__ = 0;
    $93 = $92&1;
    if ($93) {
     $122 = ___cxa_find_matching_catch_2()|0;
     $123 = tempRet0;
     $$155 = $123;$$166 = $122;
    } else {
     __THREW__ = 0;
     invoke_vii(868,($3|0),($4|0));
     $94 = __THREW__; __THREW__ = 0;
     $95 = $94&1;
     do {
      if ($95) {
       $124 = ___cxa_find_matching_catch_2()|0;
       $125 = tempRet0;
       $$054 = $125;$$065 = $124;
      } else {
       __THREW__ = 0;
       $96 = (invoke_ii(826,($3|0))|0);
       $97 = __THREW__; __THREW__ = 0;
       $98 = $97&1;
       if ($98) {
        $126 = ___cxa_find_matching_catch_2()|0;
        $127 = tempRet0;
        __ZN9ExitGames6Common10UTF8StringD2Ev($3);
        $$054 = $127;$$065 = $126;
        break;
       }
       __ZN9ExitGames6Common10UTF8StringD2Ev($3);
       __ZN9ExitGames6Common7JStringD2Ev($4);
       __ZN9ExitGames6Common7JStringD2Ev($6);
       if ($$044) {
        __ZN9ExitGames6Common7JStringD2Ev($7);
        __ZN9ExitGames6Common7JStringD2Ev($8);
        __ZN9ExitGames6Common7JStringD2Ev($9);
        __ZN9ExitGames6Common7JStringD2Ev($15);
        __ZN9ExitGames6Common7JStringD2Ev($10);
        __ZN9ExitGames6Common7JStringD2Ev($11);
        __ZN9ExitGames6Common7JStringD2Ev($12);
        __ZN9ExitGames6Common7JStringD2Ev($13);
        __ZN9ExitGames6Common7JStringD2Ev($14);
       }
       __ZN9ExitGames6Common7JStringD2Ev($5);
       __ZNK9ExitGames6Common7JString18UTF8RepresentationEv($17,163756);
       __THREW__ = 0;
       $99 = (invoke_ii(826,($17|0))|0);
       $100 = __THREW__; __THREW__ = 0;
       $101 = $100&1;
       if ($101) {
        $130 = ___cxa_find_matching_catch_2()|0;
        $131 = tempRet0;
        __ZN9ExitGames6Common10UTF8StringD2Ev($17);
        $$12 = $131;$$1277 = $130;
        ___resumeException($$1277|0);
        // unreachable;
       } else {
        __ZN9ExitGames6Common10UTF8StringD2Ev($17);
        $102 = _emscripten_asm_const_iii(0, ($99|0), ($96|0))|0;
        $103 = ((($this)) + 48|0);
        ;HEAP32[$103>>2]=0|0;HEAP32[$103+4>>2]=0|0;HEAP32[$103+8>>2]=0|0;HEAP32[$103+12>>2]=0|0;HEAP32[$103+16>>2]=0|0;HEAP32[$103+20>>2]=0|0;HEAP32[$103+24>>2]=0|0;
        HEAP16[$103>>1] = 2;
        HEAP32[$18>>2] = 2;
        $$0 = 1;
        STACKTOP = sp;return ($$0|0);
       }
      }
     } while(0);
     __ZN9ExitGames6Common7JStringD2Ev($4);
     $$155 = $$054;$$166 = $$065;
    }
    __ZN9ExitGames6Common7JStringD2Ev($6);
    if ($$044) {
     $$256131 = $$155;$$267130 = $$166;
     label = 64;
    } else {
     $$11 = $$155;$$1176 = $$166;
     break;
    }
   }
   if ((label|0) == 64) {
    __ZN9ExitGames6Common7JStringD2Ev($7);
    $$357$ph = $$256131;$$368$ph = $$267130;
    label = 65;
   }
   if ((label|0) == 65) {
    __ZN9ExitGames6Common7JStringD2Ev($8);
    $$458$ph = $$357$ph;$$469$ph = $$368$ph;
    label = 66;
   }
   if ((label|0) == 66) {
    __ZN9ExitGames6Common7JStringD2Ev($9);
    $$559$ph = $$458$ph;$$570$ph = $$469$ph;
    label = 67;
   }
   if ((label|0) == 67) {
    __ZN9ExitGames6Common7JStringD2Ev($15);
    $$660$ph = $$559$ph;$$671$ph = $$570$ph;
    label = 68;
   }
   if ((label|0) == 68) {
    __ZN9ExitGames6Common7JStringD2Ev($10);
    $$761$ph = $$660$ph;$$772$ph = $$671$ph;
    label = 69;
   }
   if ((label|0) == 69) {
    __ZN9ExitGames6Common7JStringD2Ev($11);
    $$862$ph = $$761$ph;$$873$ph = $$772$ph;
    label = 70;
   }
   if ((label|0) == 70) {
    __ZN9ExitGames6Common7JStringD2Ev($12);
    $$963$ph = $$862$ph;$$974$ph = $$873$ph;
    label = 71;
   }
   if ((label|0) == 71) {
    __ZN9ExitGames6Common7JStringD2Ev($13);
    $$1064$ph = $$963$ph;$$1075$ph = $$974$ph;
   }
   __ZN9ExitGames6Common7JStringD2Ev($14);
   $$11 = $$1064$ph;$$1176 = $$1075$ph;
  }
 } while(0);
 if ((label|0) == 46) {
  $104 = ___cxa_find_matching_catch_2()|0;
  $105 = tempRet0;
  $$11 = $105;$$1176 = $104;
 }
 __ZN9ExitGames6Common7JStringD2Ev($5);
 $$12 = $$11;$$1277 = $$1176;
 ___resumeException($$1277|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal13SocketConnect14stopConnectionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  return 0;
 }
 HEAP32[$0>>2] = 0;
 $3 = ((($this)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(-1);
 if ($5) {
  return 0;
 }
 (_close($4)|0);
 HEAP32[$3>>2] = -1;
 return 0;
}
function __ZN9ExitGames6Photon8Internal13SocketConnect11sendPackageEPKhi($this,$src,$count) {
 $this = $this|0;
 $src = $src|0;
 $count = $count|0;
 var $$0 = 0, $$pre = 0, $$pre5 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP32[$0>>2] = $count;
 $1 = ((($this)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  $$0 = 10;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = ((($this)) + 24|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(1);
 if ($6) {
  $$0 = 32;
  STACKTOP = sp;return ($$0|0);
 }
 $7 = ($count|0)==(0);
 if ($7) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $8 = ((($this)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($9|0)==(3);
 if ($10) {
  $11 = HEAP32[$this>>2]|0;
  $12 = ((($11)) + 56|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (FUNCTION_TABLE_ii[$13 & 2047]($this)|0);
  if (!($14)) {
   $15 = ((($this)) + 16|0);
   HEAP32[$15>>2] = 540;
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $16 = HEAP32[$this>>2]|0;
 $17 = ((($16)) + 48|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = (FUNCTION_TABLE_iiii[$18 & 2047]($this,$src,$0)|0);
 $20 = ($19|0)==(32);
 $21 = HEAP32[$0>>2]|0;
 $22 = ($21|0)!=(0);
 $or$cond = $20 | $22;
 if (!($or$cond)) {
  switch ($19|0) {
  case 0:  {
   $storemerge = 2;
   break;
  }
  case -1:  {
   $storemerge = 3;
   break;
  }
  default: {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  }
  HEAP32[$4>>2] = $storemerge;
  $$0 = $19;
  STACKTOP = sp;return ($$0|0);
 }
 $23 = ((($this)) + 32|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($21|0)>($24|0);
 $26 = ((($this)) + 28|0);
 $27 = HEAP32[$26>>2]|0;
 if ($25) {
  $28 = ($27|0)==(0|0);
  if ($28) {
   $30 = $21;
  } else {
   $29 = ((($27)) + -8|0);
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($29);
   $$pre = HEAP32[$0>>2]|0;
   $30 = $$pre;
  }
  HEAP32[$23>>2] = $30;
  $31 = (($30) + 8)|0;
  $32 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($31)|0);
  $33 = ((($32)) + 8|0);
  $34 = $32;
  $35 = $34;
  HEAP32[$35>>2] = $30;
  $36 = (($34) + 4)|0;
  $37 = $36;
  HEAP32[$37>>2] = 0;
  $38 = ($30|0)==(0);
  if (!($38)) {
   _memset(($33|0),0,($30|0))|0;
  }
  HEAP32[$26>>2] = $33;
  $$pre5 = HEAP32[$0>>2]|0;
  $39 = $$pre5;$41 = $33;
 } else {
  $39 = $21;$41 = $27;
 }
 $$sum = (($count) - ($39))|0;
 $40 = (($src) + ($$sum)|0);
 _memcpy(($41|0),($40|0),($39|0))|0;
 $42 = HEAP32[$0>>2]|0;
 $43 = ((($this)) + 36|0);
 HEAP32[$43>>2] = $42;
 HEAP32[$4>>2] = 1;
 $$0 = 1;
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal13PhotonConnect6retainEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = (($1) + 1)<<24>>24;
 HEAP8[$0>>0] = $2;
 return ($this|0);
}
function __ZN9ExitGames6Photon8Internal13PhotonConnect7releaseEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = (($1) + -1)<<24>>24;
 HEAP8[$0>>0] = $2;
 $3 = ($2<<24>>24)==(0);
 $4 = ($this|0)!=(0|0);
 $or$cond = $4 & $3;
 if (!($or$cond)) {
  return;
 }
 $5 = HEAP32[$this>>2]|0;
 $6 = ((($5)) + 36|0);
 $7 = HEAP32[$6>>2]|0;
 FUNCTION_TABLE_vi[$7 & 2047]($this);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($this);
 return;
}
function __ZNK9ExitGames6Photon8Internal13SocketConnect11getBytesOutEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal13SocketConnect10getBytesInEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal13SocketConnect12getUsingIPv6Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)!=(0);
 return ($2|0);
}
function __ZN9ExitGames6Photon8Internal13SocketConnect10sendBufferEPcRi($this,$pBuff,$size) {
 $this = $this|0;
 $pBuff = $pBuff|0;
 $size = $size|0;
 var $$0 = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$size>>2]|0;
 $3 = (_send($1,$pBuff,$2,0)|0);
 $4 = ($3|0)<(0);
 L1: do {
  if ($4) {
   $5 = (___errno_location()|0);
   $6 = HEAP32[$5>>2]|0;
   switch ($6|0) {
   case 101: case 100:  {
    $$0 = -1;$$sink = 534;
    break L1;
    break;
   }
   case 115: case 11:  {
    $$0 = 32;$$sink = 0;
    break L1;
    break;
   }
   case 90:  {
    $$0 = -1;$$sink = 539;
    break L1;
    break;
   }
   case 107: case 104:  {
    $$0 = -1;$$sink = 540;
    break L1;
    break;
   }
   default: {
    $$0 = -1;$$sink = -1;
    break L1;
   }
   }
  } else {
   $7 = ((($this)) + 40|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = (($8) + ($3))|0;
   HEAP32[$7>>2] = $9;
   $10 = HEAP32[$size>>2]|0;
   $11 = (($10) - ($3))|0;
   HEAP32[$size>>2] = $11;
   $$0 = 0;$$sink = 0;
  }
 } while(0);
 $12 = ((($this)) + 16|0);
 HEAP32[$12>>2] = $$sink;
 return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal13SocketConnect10recvBufferEPhiPi($this,$pBuff,$iSize,$pBytesRead) {
 $this = $this|0;
 $pBuff = $pBuff|0;
 $iSize = $iSize|0;
 $pBytesRead = $pBytesRead|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (_recv($1,$pBuff,$iSize,0)|0);
 $3 = ($2|0)<(0);
 if (!($3)) {
  $8 = ($2|0)==(0);
  $9 = ((($this)) + 16|0);
  if ($8) {
   HEAP32[$9>>2] = 540;
   $$0 = 540;
   return ($$0|0);
  } else {
   HEAP32[$9>>2] = 0;
   $10 = ((($this)) + 44|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($11) + ($2))|0;
   HEAP32[$10>>2] = $12;
   HEAP32[$pBytesRead>>2] = $2;
   $$0 = 0;
   return ($$0|0);
  }
 }
 HEAP32[$pBytesRead>>2] = 0;
 $4 = (___errno_location()|0);
 $5 = HEAP32[$4>>2]|0;
 switch ($5|0) {
 case 11: case 115:  {
  $6 = ((($this)) + 16|0);
  HEAP32[$6>>2] = 0;
  $$0 = 32;
  return ($$0|0);
  break;
 }
 default: {
 }
 }
 $7 = ((($this)) + 16|0);
 switch ($5|0) {
 case 100: case 111:  {
  $storemerge = 534;
  break;
 }
 default: {
  $storemerge = -1;
 }
 }
 HEAP32[$7>>2] = $storemerge;
 $$0 = -1;
 return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal13SocketConnectD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6236);
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(0);
 if ($2) {
  label = 6;
 } else {
  $3 = ((($this)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==(0);
  if ($5) {
   label = 6;
  } else {
   HEAP32[$3>>2] = 0;
   __THREW__ = 0;
   (invoke_ii(869,($1|0))|0);
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if (!($7)) {
    HEAP32[$0>>2] = -1;
    label = 6;
   }
  }
 }
 do {
  if ((label|0) == 6) {
   $11 = ((($this)) + 28|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0|0);
   if (!($13)) {
    $14 = ((($12)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($14|0));
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if ($16) {
     break;
    }
   }
   $17 = HEAP16[83684]|0;
   $18 = ($17<<16>>16)==(0);
   if ($18) {
    $20 = ((($this)) + 80|0);
    __ZN9ExitGames6Common10UTF8StringD2Ev($20);
    return;
   }
   $19 = (($17) + -1)<<16>>16;
   HEAP16[83684] = $19;
   $20 = ((($this)) + 80|0);
   __ZN9ExitGames6Common10UTF8StringD2Ev($20);
   return;
  }
 } while(0);
 $8 = ___cxa_find_matching_catch_3(0|0)|0;
 $9 = tempRet0;
 $10 = ((($this)) + 80|0);
 __ZN9ExitGames6Common10UTF8StringD2Ev($10);
 ___clang_call_terminate($8);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal13PhotonConnectD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Photon8Internal13SocketConnectD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal13SocketConnectD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal13SocketConnect16addPortToAddressERKNS_6Common7JStringE($agg$result,$address) {
 $agg$result = $agg$result|0;
 $address = $address|0;
 var $$01749 = 0, $$02348 = 0, $$034 = 0, $$118$ph = 0, $$124$ph = 0, $$219$ph = 0, $$225$ph = 0, $$320$ph = 0, $$326$ph = 0, $$421$ph = 0, $$427$ph = 0, $$522$ph = 0, $$528$ph = 0, $$6 = 0, $$629 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0;
 var $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE = 0, $_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(160|0);
 $0 = sp + 144|0;
 $1 = sp + 128|0;
 $2 = sp + 112|0;
 $3 = sp + 96|0;
 $4 = sp + 80|0;
 $5 = sp + 64|0;
 $6 = sp + 48|0;
 $7 = sp + 32|0;
 $8 = sp + 16|0;
 $9 = sp;
 $10 = (__ZNK9ExitGames6Common7JString7indexOfEc($address,46)|0);
 $11 = ($10|0)>(-1);
 if (!($11)) {
  __ZN9ExitGames6Common7JStringC2EPKw($0,6432);
  __THREW__ = 0;
  $12 = (invoke_iii(870,($address|0),($0|0))|0);
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $$6 = $29;$$629 = $30;
   ___resumeException($$6|0);
   // unreachable;
  }
  $15 = ($12|0)>(-1);
  __ZN9ExitGames6Common7JStringD2Ev($0);
  if (!($15)) {
   $39 = (__ZNK9ExitGames6Common7JString7indexOfEc($address,93)|0);
   __ZNK9ExitGames6Common7JString9substringEj($4,$address,$39);
   __THREW__ = 0;
   $40 = (invoke_iii(873,($4|0),58)|0);
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if ($42) {
    $61 = ___cxa_find_matching_catch_2()|0;
    $62 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($4);
    $$6 = $61;$$629 = $62;
    ___resumeException($$6|0);
    // unreachable;
   }
   __ZN9ExitGames6Common7JStringD2Ev($4);
   $43 = ($40|0)==(-1);
   $44 = ($39|0)<(0);
   $45 = $44 | $43;
   if (!($45)) {
    __THREW__ = 0;
    invoke_vii(786,($agg$result|0),($address|0));
    $59 = __THREW__; __THREW__ = 0;
    $60 = $59&1;
    if (!($60)) {
     STACKTOP = sp;return;
    }
    $73 = ___cxa_find_matching_catch_2()|0;
    $74 = tempRet0;
    $$6 = $73;$$629 = $74;
    ___resumeException($$6|0);
    // unreachable;
   }
   __ZN9ExitGames6Common7JStringC2EPKw($7,12344);
   __THREW__ = 0;
   invoke_viii(785,($6|0),($7|0),($address|0));
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    $63 = ___cxa_find_matching_catch_2()|0;
    $64 = tempRet0;
    $$522$ph = $63;$$528$ph = $64;
   } else {
    __THREW__ = 0;
    invoke_vii(801,($8|0),(6496|0));
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $65 = ___cxa_find_matching_catch_2()|0;
     $66 = tempRet0;
     $$421$ph = $65;$$427$ph = $66;
    } else {
     __THREW__ = 0;
     invoke_viii(785,($5|0),($6|0),($8|0));
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     if ($51) {
      $67 = ___cxa_find_matching_catch_2()|0;
      $68 = tempRet0;
      $$320$ph = $67;$$326$ph = $68;
     } else {
      __THREW__ = 0;
      invoke_vii(801,($9|0),(6480|0));
      $52 = __THREW__; __THREW__ = 0;
      $53 = $52&1;
      if ($53) {
       $69 = ___cxa_find_matching_catch_2()|0;
       $70 = tempRet0;
       $$219$ph = $69;$$225$ph = $70;
      } else {
       __THREW__ = 0;
       $54 = (invoke_iii(871,($address|0),($9|0))|0);
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if (!($56)) {
        $_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE1 = $54 ? 22264 : 22266;
        __THREW__ = 0;
        invoke_viii(872,($agg$result|0),($5|0),($_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE1|0));
        $57 = __THREW__; __THREW__ = 0;
        $58 = $57&1;
        if (!($58)) {
         __ZN9ExitGames6Common7JStringD2Ev($9);
         __ZN9ExitGames6Common7JStringD2Ev($5);
         __ZN9ExitGames6Common7JStringD2Ev($8);
         __ZN9ExitGames6Common7JStringD2Ev($6);
         __ZN9ExitGames6Common7JStringD2Ev($7);
         STACKTOP = sp;return;
        }
       }
       $71 = ___cxa_find_matching_catch_2()|0;
       $72 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($9);
       $$219$ph = $71;$$225$ph = $72;
      }
      __ZN9ExitGames6Common7JStringD2Ev($5);
      $$320$ph = $$219$ph;$$326$ph = $$225$ph;
     }
     __ZN9ExitGames6Common7JStringD2Ev($8);
     $$421$ph = $$320$ph;$$427$ph = $$326$ph;
    }
    __ZN9ExitGames6Common7JStringD2Ev($6);
    $$522$ph = $$421$ph;$$528$ph = $$427$ph;
   }
   __ZN9ExitGames6Common7JStringD2Ev($7);
   $$6 = $$522$ph;$$629 = $$528$ph;
   ___resumeException($$6|0);
   // unreachable;
  }
 }
 $16 = (__ZNK9ExitGames6Common7JString7indexOfEc($address,58)|0);
 $17 = ($16|0)<(0);
 do {
  if ($17) {
   __ZN9ExitGames6Common7JStringC2EPKw($2,6472);
   __THREW__ = 0;
   invoke_vii(801,($3|0),(6480|0));
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    $31 = ___cxa_find_matching_catch_2()|0;
    $32 = tempRet0;
    $$118$ph = $31;$$124$ph = $32;
    break;
   }
   __THREW__ = 0;
   $20 = (invoke_iii(871,($address|0),($3|0))|0);
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if (!($22)) {
    $_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE = $20 ? 22264 : 22266;
    __THREW__ = 0;
    invoke_viii(872,($1|0),($2|0),($_ZN9ExitGames6Photon11NetworkPortL3WSSE$_ZN9ExitGames6Photon11NetworkPortL2WSE|0));
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if (!($24)) {
     $$034 = 1;
     label = 9;
     break;
    }
   }
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$01749 = $35;$$02348 = $36;
   label = 17;
  } else {
   __THREW__ = 0;
   invoke_vii(801,($1|0),(163772|0));
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $37 = ___cxa_find_matching_catch_2()|0;
    $38 = tempRet0;
    $$6 = $37;$$629 = $38;
    ___resumeException($$6|0);
    // unreachable;
   } else {
    $$034 = 0;
    label = 9;
   }
  }
 } while(0);
 do {
  if ((label|0) == 9) {
   __THREW__ = 0;
   invoke_viii(785,($agg$result|0),($address|0),($1|0));
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if ($28) {
    $33 = ___cxa_find_matching_catch_2()|0;
    $34 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($1);
    if ($$034) {
     $$01749 = $33;$$02348 = $34;
     label = 17;
     break;
    } else {
     $$6 = $33;$$629 = $34;
    }
    ___resumeException($$6|0);
    // unreachable;
   }
   __ZN9ExitGames6Common7JStringD2Ev($1);
   if (!($$034)) {
    STACKTOP = sp;return;
   }
   __ZN9ExitGames6Common7JStringD2Ev($3);
   __ZN9ExitGames6Common7JStringD2Ev($2);
   STACKTOP = sp;return;
  }
 } while(0);
 if ((label|0) == 17) {
  __ZN9ExitGames6Common7JStringD2Ev($3);
  $$118$ph = $$01749;$$124$ph = $$02348;
 }
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$6 = $$118$ph;$$629 = $$124$ph;
 ___resumeException($$6|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplIhEENS0_7JStringERKS2_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   $3 = HEAP8[$Rsh>>0]|0;
   __THREW__ = 0;
   (invoke_iii(782,($str$i|0),($3|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    $6 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if (!($8)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($6|0));
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $10;$eh$lpad$body$indexZ2D = $9;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 $eh$lpad$body$index2Z2D = $14;$eh$lpad$body$indexZ2D = $13;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase10InitValues22getClientLibIdentifierEv() {
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[167383]|0;
 $1 = ($0<<24>>24)!=(0);
 $2 = (__ZN9ExitGames6Photon8Internal8PeerData10InitValues22getClientLibIdentifierEb($1)|0);
 return ($2|0);
}
function __ZN9ExitGames6Photon8Internal8PeerData10InitValues22getClientLibIdentifierEb($usingObjC) {
 $usingObjC = $usingObjC|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 49;
}
function __ZN9ExitGames6CommonplItEENS0_7JStringERKS2_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   $3 = HEAP16[$Rsh>>1]|0;
   __THREW__ = 0;
   (invoke_iii(874,($str$i|0),($3|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    $6 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if (!($8)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($6|0));
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $10;$eh$lpad$body$indexZ2D = $9;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 $eh$lpad$body$index2Z2D = $14;$eh$lpad$body$indexZ2D = $13;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase14onSendCommandsEi($this,$errorcode) {
 $this = $this|0;
 $errorcode = $errorcode|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 96|0);
 HEAP8[$0>>0] = 0;
 $1 = ((($this)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(1);
 $4 = ($errorcode|0)!=(0);
 $or$cond = $4 & $3;
 $5 = ($errorcode|0)==(534);
 $6 = ($errorcode|0)==(540);
 if ($or$cond) {
  label = 3;
 } else {
  switch ($errorcode|0) {
  case 534: case 540:  {
   label = 3;
   break;
  }
  default: {
  }
  }
 }
 if ((label|0) == 3) {
  HEAP32[$1>>2] = 6;
 }
 if (!($4)) {
  return;
 }
 $7 = ((($this)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($8)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = HEAP32[$1>>2]|0;
 $14 = ($13|0)==(1);
 $15 = $5 | $6;
 $16 = $15 ? 1026 : 1030;
 $17 = $14 ? 1023 : $16;
 FUNCTION_TABLE_vii[$12 & 2047]($9,$17);
 $18 = HEAP32[$1>>2]|0;
 $19 = ($18|0)==(1);
 L9: do {
  if (!($19)) {
   switch ($errorcode|0) {
   case 534: case 540:  {
    break L9;
    break;
   }
   default: {
   }
   }
   return;
  }
 } while(0);
 $20 = HEAP32[$this>>2]|0;
 $21 = ((($20)) + 88|0);
 $22 = HEAP32[$21>>2]|0;
 FUNCTION_TABLE_vi[$22 & 2047]($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase7serviceEb($this,$dispatch) {
 $this = $this|0;
 $dispatch = $dispatch|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = HEAP32[$this>>2]|0;
 $2 = ((($1)) + 32|0);
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_vi[$3 & 2047]($this);
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147264);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($0|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 $9 = $dispatch ? 9084 : 9152;
 __THREW__ = 0;
 invoke_viiiiiiii(765,($5|0),4,(6560|0),($6|0),1,203,($9|0),($vararg_buffer|0));
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 if ($dispatch) {
  while(1) {
   $12 = HEAP32[$this>>2]|0;
   $13 = ((($12)) + 44|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (FUNCTION_TABLE_ii[$14 & 2047]($this)|0);
   if (!($15)) {
    break;
   }
  }
 }
 while(1) {
  $18 = HEAP32[$this>>2]|0;
  $19 = ((($18)) + 36|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (FUNCTION_TABLE_ii[$20 & 2047]($this)|0);
  if (!($21)) {
   break;
  }
 }
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase12serviceBasicEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 FUNCTION_TABLE_vi[$3 & 2047]($1);
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase8opCustomEPKNS0_16OperationRequestEbhbh($this,$operationRequest,$sendReliable,$channelId,$encrypt,$msgType) {
 $this = $this|0;
 $operationRequest = $operationRequest|0;
 $sendReliable = $sendReliable|0;
 $channelId = $channelId|0;
 $encrypt = $encrypt|0;
 $msgType = $msgType|0;
 var $$0 = 0, $$02 = 0, $$03 = 0, $$encrypt = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $or$cond = 0, $serializedOp = 0, $serializedOpSize = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_ptr5 = 0, $vararg_ptr6 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $serializedOp = sp + 84|0;
 $serializedOpSize = sp + 80|0;
 $0 = sp + 64|0;
 $1 = sp + 48|0;
 $2 = sp + 32|0;
 HEAP32[$serializedOp>>2] = 0;
 HEAP32[$serializedOpSize>>2] = 0;
 $3 = ((($this)) + 136|0);
 $4 = HEAP8[$3>>0]|0;
 $not$ = ($4<<24>>24)!=(5);
 $$encrypt = $not$ & $encrypt;
 if ($$encrypt) {
  $5 = ((($this)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 25|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = ($8<<24>>24)==(0);
  if ($9) {
   $10 = ((($this)) + 8|0);
   $11 = HEAP32[$10>>2]|0;
   __ZN9ExitGames6Common7JStringC2EPKc($0,147255);
   __THREW__ = 0;
   $12 = (invoke_ii(764,($0|0))|0);
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if (!($14)) {
    __THREW__ = 0;
    invoke_viiiiiiii(765,($11|0),1,(6560|0),($12|0),1,252,(8516|0),($vararg_buffer|0));
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if (!($16)) {
     __ZN9ExitGames6Common7JStringD2Ev($0);
     $$03 = 0;
     STACKTOP = sp;return ($$03|0);
    }
   }
   $17 = ___cxa_find_matching_catch_2()|0;
   $18 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $$0 = $18;$$02 = $17;
   ___resumeException($$02|0);
   // unreachable;
  }
 }
 $19 = ((($this)) + 36|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20|0)==(3);
 if (!($21)) {
  $22 = ((($this)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  __ZN9ExitGames6Common7JStringC2EPKc($1,147255);
  __THREW__ = 0;
  $24 = (invoke_ii(764,($1|0))|0);
  $25 = __THREW__; __THREW__ = 0;
  $26 = $25&1;
  if (!($26)) {
   __THREW__ = 0;
   invoke_viiiiiiii(765,($23|0),1,(6560|0),($24|0),1,257,(8720|0),($vararg_buffer1|0));
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if (!($28)) {
    __ZN9ExitGames6Common7JStringD2Ev($1);
    $$03 = 0;
    STACKTOP = sp;return ($$03|0);
   }
  }
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  $$0 = $30;$$02 = $29;
  ___resumeException($$02|0);
  // unreachable;
 }
 $31 = $channelId&255;
 $32 = ((($this)) + 4|0);
 $33 = HEAP32[$32>>2]|0;
 $34 = ((($33)) + 4|0);
 $35 = HEAP8[$34>>0]|0;
 $36 = ($35&255)>($channelId&255);
 if (!($36)) {
  $37 = ((($this)) + 8|0);
  $38 = HEAP32[$37>>2]|0;
  __ZN9ExitGames6Common7JStringC2EPKc($2,147255);
  __THREW__ = 0;
  $39 = (invoke_ii(764,($2|0))|0);
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if (!($41)) {
   $42 = HEAP32[$32>>2]|0;
   $43 = ((($42)) + 4|0);
   $44 = HEAP8[$43>>0]|0;
   $45 = $44&255;
   $46 = (($45) + -1)|0;
   __THREW__ = 0;
   HEAP32[$vararg_buffer3>>2] = $31;
   $vararg_ptr5 = ((($vararg_buffer3)) + 4|0);
   HEAP32[$vararg_ptr5>>2] = 0;
   $vararg_ptr6 = ((($vararg_buffer3)) + 8|0);
   HEAP32[$vararg_ptr6>>2] = $46;
   invoke_viiiiiiii(765,($38|0),1,(6560|0),($39|0),1,262,(8844|0),($vararg_buffer3|0));
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if (!($48)) {
    __ZN9ExitGames6Common7JStringD2Ev($2);
    $$03 = 0;
    STACKTOP = sp;return ($$03|0);
   }
  }
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($2);
  $$0 = $50;$$02 = $49;
  ___resumeException($$02|0);
  // unreachable;
 }
 $51 = HEAP32[$this>>2]|0;
 $52 = ((($51)) + 96|0);
 $53 = HEAP32[$52>>2]|0;
 $54 = (FUNCTION_TABLE_iiiiiii[$53 & 1023]($this,$operationRequest,$serializedOp,$serializedOpSize,$$encrypt,$msgType)|0);
 if (!($54)) {
  $$03 = 0;
  STACKTOP = sp;return ($$03|0);
 }
 $55 = HEAP32[$serializedOpSize>>2]|0;
 $56 = ($55|0)>(0);
 $57 = HEAP32[$serializedOp>>2]|0;
 $58 = ($57|0)!=(0|0);
 $or$cond = $56 & $58;
 if (!($or$cond)) {
  $$03 = 0;
  STACKTOP = sp;return ($$03|0);
 }
 $59 = HEAP32[$this>>2]|0;
 $60 = ((($59)) + 92|0);
 $61 = HEAP32[$60>>2]|0;
 $62 = $sendReliable ? 6 : 7;
 FUNCTION_TABLE_viiiii[$61 & 1023]($this,$62,$57,$55,$channelId);
 $63 = HEAP32[$serializedOp>>2]|0;
 $64 = ($63|0)==(0|0);
 if ($64) {
  $$03 = 1;
  STACKTOP = sp;return ($$03|0);
 }
 $65 = ((($63)) + -8|0);
 __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($65);
 $$03 = 1;
 STACKTOP = sp;return ($$03|0);
}
function __ZN9ExitGames6Photon8Internal8PeerBase19establishEncryptionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 120|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (FUNCTION_TABLE_ii[$2 & 2047]($this)|0);
 return ($3|0);
}
function __ZN9ExitGames6Photon8Internal8PeerBase10pingServerERKNS_6Common7JStringEj($this,$address,$pingAttempts) {
 $this = $this|0;
 $address = $address|0;
 $pingAttempts = $pingAttempts|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147244);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if (!($5)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($2|0),1,(6560|0),($3|0),1,288,(8276|0),($vararg_buffer|0));
  $6 = __THREW__; __THREW__ = 0;
  $7 = $6&1;
  if (!($7)) {
   __ZN9ExitGames6Common7JStringD2Ev($0);
   STACKTOP = sp;return;
  }
 }
 $8 = ___cxa_find_matching_catch_2()|0;
 $9 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($8|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase14stopConnectionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 40|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 (FUNCTION_TABLE_ii[$4 & 2047]($1)|0);
 $5 = HEAP32[$this>>2]|0;
 $6 = ((($5)) + 80|0);
 $7 = HEAP32[$6>>2]|0;
 FUNCTION_TABLE_vi[$7 & 2047]($this);
 $8 = (__Z11getTimeUnixv()|0);
 $9 = ((($this)) + 44|0);
 HEAP32[$9>>2] = $8;
 $10 = ((($this)) + 16|0);
 HEAP32[$10>>2] = $8;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase14onDisconnectedEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147223);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($17|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($2|0),4,(6560|0),($3|0),1,442,(163772|0),($vararg_buffer|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($17|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $8 = HEAP32[$this>>2]|0;
  $9 = ((($8)) + 76|0);
  $10 = HEAP32[$9>>2]|0;
  FUNCTION_TABLE_vi[$10 & 2047]($this);
  $11 = ((($this)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($12)|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($14)) + 16|0);
  $16 = HEAP32[$15>>2]|0;
  FUNCTION_TABLE_vii[$16 & 2047]($13,1025);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Photon8Internal8PeerBase18serializeOperationEPKNS0_16OperationRequestEPPhPibh($this,$pOperationRequest,$buffer,$buffSize,$encrypt,$msgType) {
 $this = $this|0;
 $pOperationRequest = $pOperationRequest|0;
 $buffer = $buffer|0;
 $buffSize = $buffSize|0;
 $encrypt = $encrypt|0;
 $msgType = $msgType|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $bytes_num$0 = 0, $bytes_num$0$in = 0, $code = 0;
 var $count = 0, $eh$lpad$body$index10Z2D = 0, $eh$lpad$body$indexZ2D = 0, $enc_bytes = 0, $enc_bytes_num = 0, $i$014 = 0, $operationWriter = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 104|0;
 $1 = sp + 88|0;
 $enc_bytes = sp + 84|0;
 $enc_bytes_num = sp + 80|0;
 $2 = sp + 64|0;
 $operationWriter = sp + 48|0;
 $code = sp + 122|0;
 $count = sp + 120|0;
 $3 = sp + 32|0;
 HEAP32[$enc_bytes>>2] = 0;
 HEAP32[$enc_bytes_num>>2] = 0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($2,147204);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($2|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($5|0),4,(6560|0),($6|0),1,454,(163772|0),($vararg_buffer|0));
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   __ZN9ExitGames6Common7JStringD2Ev($2);
   __ZN9ExitGames6Common7Helpers24SerializerImplementationC2Ev($operationWriter);
   __THREW__ = 0;
   $11 = (invoke_ii(875,($pOperationRequest|0))|0);
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   L4: do {
    if ($13) {
     label = 39;
    } else {
     HEAP8[$code>>0] = $11;
     __THREW__ = 0;
     $14 = (invoke_ii(876,($pOperationRequest|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if ($16) {
      label = 39;
     } else {
      __THREW__ = 0;
      $17 = (invoke_ii(877,($14|0))|0);
      $18 = __THREW__; __THREW__ = 0;
      $19 = $18&1;
      if ($19) {
       label = 39;
      } else {
       $20 = $17&65535;
       HEAP16[$count>>1] = $20;
       __THREW__ = 0;
       invoke_vii(878,($operationWriter|0),1);
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if ($22) {
        label = 39;
       } else {
        __THREW__ = 0;
        invoke_viii(879,($operationWriter|0),($code|0),1);
        $23 = __THREW__; __THREW__ = 0;
        $24 = $23&1;
        if ($24) {
         label = 39;
        } else {
         __THREW__ = 0;
         invoke_vii(878,($operationWriter|0),2);
         $25 = __THREW__; __THREW__ = 0;
         $26 = $25&1;
         if ($26) {
          label = 39;
         } else {
          __THREW__ = 0;
          invoke_viii(879,($operationWriter|0),($count|0),2);
          $27 = __THREW__; __THREW__ = 0;
          $28 = $27&1;
          if ($28) {
           label = 39;
          } else {
           __THREW__ = 0;
           $29 = (invoke_ii(876,($pOperationRequest|0))|0);
           $30 = __THREW__; __THREW__ = 0;
           $31 = $30&1;
           if ($31) {
            label = 39;
           } else {
            __THREW__ = 0;
            $32 = (invoke_ii(767,($29|0))|0);
            $33 = __THREW__; __THREW__ = 0;
            $34 = $33&1;
            if ($34) {
             label = 39;
            } else {
             __THREW__ = 0;
             $35 = (invoke_ii(768,($32|0))|0);
             $36 = __THREW__; __THREW__ = 0;
             $37 = $36&1;
             if ($37) {
              label = 39;
             } else {
              $38 = HEAP16[$count>>1]|0;
              $39 = ($38<<16>>16)>(0);
              L15: do {
               if ($39) {
                $40 = ((($35)) + 4|0);
                $41 = ((($35)) + 16|0);
                $48 = 0;$i$014 = 0;
                while(1) {
                 $46 = HEAP32[$40>>2]|0;
                 $47 = ($46>>>0)>($48>>>0);
                 if (!($47)) {
                  __THREW__ = 0;
                  invoke_vii(739,($1|0),(147471|0));
                  $49 = __THREW__; __THREW__ = 0;
                  $50 = $49&1;
                  if ($50) {
                   label = 38;
                   break;
                  }
                  __THREW__ = 0;
                  $51 = (invoke_ii(764,($1|0))|0);
                  $52 = __THREW__; __THREW__ = 0;
                  $53 = $52&1;
                  if ($53) {
                   label = 21;
                   break;
                  }
                  __THREW__ = 0;
                  invoke_viiiiiiii(765,(162472|0),1,(10856|0),($51|0),1,566,(10968|0),($vararg_buffer1|0));
                  $54 = __THREW__; __THREW__ = 0;
                  $55 = $54&1;
                  if ($55) {
                   label = 21;
                   break;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($1);
                 }
                 $58 = HEAP32[$41>>2]|0;
                 $59 = (($58) + (($48*40)|0)|0);
                 __THREW__ = 0;
                 $60 = (invoke_iiii(880,($operationWriter|0),($59|0),0)|0);
                 $61 = __THREW__; __THREW__ = 0;
                 $62 = $61&1;
                 if ($62) {
                  label = 38;
                  break;
                 }
                 if (!($60)) {
                  label = 33;
                  break;
                 }
                 $63 = HEAP32[$40>>2]|0;
                 $64 = ($63>>>0)>($48>>>0);
                 if (!($64)) {
                  __THREW__ = 0;
                  invoke_vii(739,($0|0),(147471|0));
                  $65 = __THREW__; __THREW__ = 0;
                  $66 = $65&1;
                  if ($66) {
                   label = 38;
                   break;
                  }
                  __THREW__ = 0;
                  $67 = (invoke_ii(764,($0|0))|0);
                  $68 = __THREW__; __THREW__ = 0;
                  $69 = $68&1;
                  if ($69) {
                   label = 29;
                   break;
                  }
                  __THREW__ = 0;
                  invoke_viiiiiiii(765,(162472|0),1,(10856|0),($67|0),1,566,(10968|0),($vararg_buffer3|0));
                  $70 = __THREW__; __THREW__ = 0;
                  $71 = $70&1;
                  if ($71) {
                   label = 29;
                   break;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($0);
                 }
                 $74 = HEAP32[$41>>2]|0;
                 $75 = (($74) + (($48*40)|0)|0);
                 __THREW__ = 0;
                 $76 = (invoke_iii(881,($32|0),($75|0))|0);
                 $77 = __THREW__; __THREW__ = 0;
                 $78 = $77&1;
                 if ($78) {
                  label = 38;
                  break;
                 }
                 __THREW__ = 0;
                 $79 = (invoke_iiii(880,($operationWriter|0),($76|0),1)|0);
                 $80 = __THREW__; __THREW__ = 0;
                 $81 = $80&1;
                 if ($81) {
                  label = 38;
                  break;
                 }
                 $43 = (($i$014) + 1)<<16>>16;
                 if (!($79)) {
                  label = 33;
                  break;
                 }
                 $42 = $43 << 16 >> 16;
                 $44 = HEAP16[$count>>1]|0;
                 $45 = ($43<<16>>16)<($44<<16>>16);
                 if ($45) {
                  $48 = $42;$i$014 = $43;
                 } else {
                  break L15;
                 }
                }
                if ((label|0) == 21) {
                 $56 = ___cxa_find_matching_catch_2()|0;
                 $57 = tempRet0;
                 __ZN9ExitGames6Common7JStringD2Ev($1);
                 $eh$lpad$body$index10Z2D = $57;$eh$lpad$body$indexZ2D = $56;
                 label = 40;
                 break L4;
                }
                else if ((label|0) == 29) {
                 $72 = ___cxa_find_matching_catch_2()|0;
                 $73 = tempRet0;
                 __ZN9ExitGames6Common7JStringD2Ev($0);
                 $eh$lpad$body$index10Z2D = $73;$eh$lpad$body$indexZ2D = $72;
                 label = 40;
                 break L4;
                }
                else if ((label|0) == 33) {
                 $82 = HEAP32[$4>>2]|0;
                 __THREW__ = 0;
                 invoke_vii(739,($3|0),(147204|0));
                 $83 = __THREW__; __THREW__ = 0;
                 $84 = $83&1;
                 if ($84) {
                  label = 39;
                  break L4;
                 }
                 __THREW__ = 0;
                 $85 = (invoke_ii(764,($3|0))|0);
                 $86 = __THREW__; __THREW__ = 0;
                 $87 = $86&1;
                 do {
                  if (!($87)) {
                   __THREW__ = 0;
                   invoke_viiiiiiii(765,($82|0),1,(6560|0),($85|0),1,467,(8064|0),($vararg_buffer5|0));
                   $88 = __THREW__; __THREW__ = 0;
                   $89 = $88&1;
                   if ($89) {
                    break;
                   }
                   __ZN9ExitGames6Common7JStringD2Ev($3);
                   $$0 = 0;
                   __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($operationWriter);
                   STACKTOP = sp;return ($$0|0);
                  }
                 } while(0);
                 $96 = ___cxa_find_matching_catch_2()|0;
                 $97 = tempRet0;
                 __ZN9ExitGames6Common7JStringD2Ev($3);
                 $$01 = $97;$$02 = $96;
                 break L4;
                }
                else if ((label|0) == 38) {
                 $92 = ___cxa_find_matching_catch_2()|0;
                 $93 = tempRet0;
                 $eh$lpad$body$index10Z2D = $93;$eh$lpad$body$indexZ2D = $92;
                 label = 40;
                 break L4;
                }
               }
              } while(0);
              if ($encrypt) {
               $98 = ((($this)) + 4|0);
               $99 = HEAP32[$98>>2]|0;
               $100 = ((($99)) + 25|0);
               $101 = HEAP8[$100>>0]|0;
               $102 = ($101<<24>>24)==(0);
               if ($102) {
                label = 48;
               } else {
                __THREW__ = 0;
                $103 = (invoke_ii(882,($operationWriter|0))|0);
                $104 = __THREW__; __THREW__ = 0;
                $105 = $104&1;
                if ($105) {
                 label = 39;
                 break;
                }
                __THREW__ = 0;
                $106 = (invoke_ii(883,($operationWriter|0))|0);
                $107 = __THREW__; __THREW__ = 0;
                $108 = $107&1;
                if ($108) {
                 label = 39;
                 break;
                }
                $109 = HEAP32[$98>>2]|0;
                $110 = ((($109)) + 20|0);
                $111 = HEAP32[$110>>2]|0;
                __THREW__ = 0;
                (invoke_iiiiii(884,($103|0),($106|0),($111|0),($enc_bytes|0),($enc_bytes_num|0))|0);
                $112 = __THREW__; __THREW__ = 0;
                $113 = $112&1;
                if ($113) {
                 label = 39;
                 break;
                }
                $114 = HEAP32[$enc_bytes_num>>2]|0;
                $bytes_num$0$in = $114;
               }
              } else {
               label = 48;
              }
              if ((label|0) == 48) {
               __THREW__ = 0;
               $115 = (invoke_ii(883,($operationWriter|0))|0);
               $116 = __THREW__; __THREW__ = 0;
               $117 = $116&1;
               if ($117) {
                label = 39;
                break;
               } else {
                $bytes_num$0$in = $115;
               }
              }
              $bytes_num$0 = (($bytes_num$0$in) + 2)|0;
              $118 = (($bytes_num$0$in) + 10)|0;
              __THREW__ = 0;
              $119 = (invoke_ii(775,($118|0))|0);
              $120 = __THREW__; __THREW__ = 0;
              $121 = $120&1;
              if ($121) {
               label = 39;
              } else {
               $122 = ((($119)) + 8|0);
               $123 = $119;
               $124 = $123;
               HEAP32[$124>>2] = $bytes_num$0;
               $125 = (($123) + 4)|0;
               $126 = $125;
               HEAP32[$126>>2] = 0;
               $127 = ($bytes_num$0|0)==(0);
               if (!($127)) {
                _memset(($122|0),0,($bytes_num$0|0))|0;
               }
               HEAP8[$122>>0] = -13;
               $128 = ((($119)) + 9|0);
               HEAP8[$128>>0] = $msgType;
               do {
                if ($encrypt) {
                 $129 = ((($this)) + 4|0);
                 $130 = HEAP32[$129>>2]|0;
                 $131 = ((($130)) + 25|0);
                 $132 = HEAP8[$131>>0]|0;
                 $133 = ($132<<24>>24)==(0);
                 if ($133) {
                  label = 56;
                  break;
                 }
                 $134 = ((($119)) + 10|0);
                 $135 = HEAP32[$enc_bytes>>2]|0;
                 $136 = HEAP32[$enc_bytes_num>>2]|0;
                 _memcpy(($134|0),($135|0),($136|0))|0;
                 $137 = $msgType&255;
                 $138 = $137 | 128;
                 $139 = $138&255;
                 HEAP8[$128>>0] = $139;
                 $140 = HEAP32[$enc_bytes>>2]|0;
                 $141 = ($140|0)==(0|0);
                 if ($141) {
                  break;
                 }
                 $142 = ((($140)) + -8|0);
                 __THREW__ = 0;
                 invoke_vi(750,($142|0));
                 $143 = __THREW__; __THREW__ = 0;
                 $144 = $143&1;
                 if ($144) {
                  label = 39;
                  break L4;
                 }
                } else {
                 label = 56;
                }
               } while(0);
               if ((label|0) == 56) {
                $145 = ((($119)) + 10|0);
                __THREW__ = 0;
                $146 = (invoke_ii(882,($operationWriter|0))|0);
                $147 = __THREW__; __THREW__ = 0;
                $148 = $147&1;
                if ($148) {
                 label = 39;
                 break;
                }
                __THREW__ = 0;
                $149 = (invoke_ii(883,($operationWriter|0))|0);
                $150 = __THREW__; __THREW__ = 0;
                $151 = $150&1;
                if ($151) {
                 label = 39;
                 break;
                }
                _memcpy(($145|0),($146|0),($149|0))|0;
               }
               HEAP32[$buffer>>2] = $122;
               HEAP32[$buffSize>>2] = $bytes_num$0;
               $$0 = 1;
               __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($operationWriter);
               STACKTOP = sp;return ($$0|0);
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    }
   } while(0);
   if ((label|0) == 39) {
    $94 = ___cxa_find_matching_catch_2()|0;
    $95 = tempRet0;
    $eh$lpad$body$index10Z2D = $95;$eh$lpad$body$indexZ2D = $94;
    label = 40;
   }
   if ((label|0) == 40) {
    $$01 = $eh$lpad$body$index10Z2D;$$02 = $eh$lpad$body$indexZ2D;
   }
   __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($operationWriter);
   $$1 = $$01;$$13 = $$02;
   ___resumeException($$13|0);
   // unreachable;
  }
 }
 $90 = ___cxa_find_matching_catch_2()|0;
 $91 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$1 = $91;$$13 = $90;
 ___resumeException($$13|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal8PeerBase20deserializeOperationEPhi($this,$inBuff,$inBuffLength) {
 $this = $this|0;
 $inBuff = $inBuff|0;
 $inBuffLength = $inBuffLength|0;
 var $$0 = 0, $$01 = 0, $$02 = 0, $$lobit = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer6 = 0, $vararg_ptr10 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $vararg_buffer6 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 88|0;
 $1 = sp + 72|0;
 $2 = sp + 56|0;
 $3 = sp + 40|0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147183);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($0|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($5|0),4,(6560|0),($6|0),1,506,(163772|0),($vararg_buffer|0));
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $11 = ($inBuffLength|0)<(2);
   if ($11) {
    $12 = HEAP32[$4>>2]|0;
    __ZN9ExitGames6Common7JStringC2EPKc($1,147183);
    __THREW__ = 0;
    $13 = (invoke_ii(764,($1|0))|0);
    $14 = __THREW__; __THREW__ = 0;
    $15 = $14&1;
    if (!($15)) {
     __THREW__ = 0;
     HEAP32[$vararg_buffer1>>2] = $inBuffLength;
     invoke_viiiiiiii(765,($12|0),1,(6560|0),($13|0),1,510,(7568|0),($vararg_buffer1|0));
     $16 = __THREW__; __THREW__ = 0;
     $17 = $16&1;
     if (!($17)) {
      __ZN9ExitGames6Common7JStringD2Ev($1);
      $$02 = 0;
      STACKTOP = sp;return ($$02|0);
     }
    }
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($1);
    $$0 = $21;$$01 = $20;
    ___resumeException($$01|0);
    // unreachable;
   }
   $22 = HEAP8[$inBuff>>0]|0;
   $23 = ($22<<24>>24)==(-13);
   if (!($23)) {
    $24 = HEAP32[$4>>2]|0;
    __ZN9ExitGames6Common7JStringC2EPKc($2,147183);
    __THREW__ = 0;
    $25 = (invoke_ii(764,($2|0))|0);
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if (!($27)) {
     $28 = HEAP8[$inBuff>>0]|0;
     $29 = $28&255;
     __THREW__ = 0;
     HEAP32[$vararg_buffer3>>2] = $29;
     invoke_viiiiiiii(765,($24|0),1,(6560|0),($25|0),1,517,(7708|0),($vararg_buffer3|0));
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if (!($31)) {
      __ZN9ExitGames6Common7JStringD2Ev($2);
      $$02 = 0;
      STACKTOP = sp;return ($$02|0);
     }
    }
    $32 = ___cxa_find_matching_catch_2()|0;
    $33 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($2);
    $$0 = $33;$$01 = $32;
    ___resumeException($$01|0);
    // unreachable;
   }
   $34 = ((($inBuff)) + 1|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = $35&255;
   $37 = $36 & 127;
   $38 = $37&255;
   $$lobit = ($35&255) >>> 7;
   $39 = HEAP32[$4>>2]|0;
   __ZN9ExitGames6Common7JStringC2EPKc($3,147183);
   __THREW__ = 0;
   $40 = (invoke_ii(764,($3|0))|0);
   $41 = __THREW__; __THREW__ = 0;
   $42 = $41&1;
   if (!($42)) {
    __THREW__ = 0;
    HEAP32[$vararg_buffer6>>2] = $inBuffLength;
    $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
    HEAP32[$vararg_ptr9>>2] = $37;
    $vararg_ptr10 = ((($vararg_buffer6)) + 8|0);
    HEAP32[$vararg_ptr10>>2] = 4;
    invoke_viiiiiiii(765,($39|0),4,(6560|0),($40|0),1,525,(7908|0),($vararg_buffer6|0));
    $43 = __THREW__; __THREW__ = 0;
    $44 = $43&1;
    if (!($44)) {
     __ZN9ExitGames6Common7JStringD2Ev($3);
     switch ($37|0) {
     case 1:  {
      $47 = HEAP32[$this>>2]|0;
      $48 = ((($47)) + 132|0);
      $49 = HEAP32[$48>>2]|0;
      FUNCTION_TABLE_vi[$49 & 2047]($this);
      $$02 = 1;
      STACKTOP = sp;return ($$02|0);
      break;
     }
     case 7: case 3:  {
      $50 = HEAP32[$this>>2]|0;
      $51 = ((($50)) + 104|0);
      $52 = HEAP32[$51>>2]|0;
      $53 = ($$lobit<<24>>24)!=(0);
      $54 = (($inBuffLength) + -2)|0;
      FUNCTION_TABLE_viiiii[$52 & 1023]($this,$inBuff,$53,$54,$38);
      $$02 = 1;
      STACKTOP = sp;return ($$02|0);
      break;
     }
     case 4:  {
      $55 = HEAP32[$this>>2]|0;
      $56 = ((($55)) + 108|0);
      $57 = HEAP32[$56>>2]|0;
      $58 = ($$lobit<<24>>24)!=(0);
      $59 = (($inBuffLength) + -2)|0;
      FUNCTION_TABLE_viiii[$57 & 2047]($this,$inBuff,$58,$59);
      $$02 = 1;
      STACKTOP = sp;return ($$02|0);
      break;
     }
     default: {
      $$02 = 1;
      STACKTOP = sp;return ($$02|0);
     }
     }
    }
   }
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($3);
   $$0 = $46;$$01 = $45;
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = $19;$$01 = $18;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal8PeerBase28deserializeOperationResponseEPhbih($this,$inBuff,$isEncrypted,$inBuffBodyLength,$msgType) {
 $this = $this|0;
 $inBuff = $inBuff|0;
 $isEncrypted = $isEncrypted|0;
 $inBuffBodyLength = $inBuffBodyLength|0;
 $msgType = $msgType|0;
 var $$06 = 0, $$09 = 0, $$1$ph29 = 0, $$12$ph28 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $$pre = 0, $$pre$phi39Z2D = 0, $$pre$phiZ2D = 0, $$pre35 = 0, $$pre38 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $debugMsg = 0, $decryptedBytes = 0, $decryptedResponseData = 0, $eh$lpad$body$index8Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body21$index2Z2D = 0, $eh$lpad$body21$indexZ2D = 0, $i$033 = 0, $operationReader = 0, $operationResponse = 0, $timeBeforeCallback$0 = 0, $timeBeforeCallback$1 = 0, $v = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $vararg_buffer = sp + 160|0;
 $0 = sp + 120|0;
 $decryptedResponseData = sp + 288|0;
 $decryptedBytes = sp + 284|0;
 $operationReader = sp + 272|0;
 $operationResponse = sp + 200|0;
 $debugMsg = sp + 80|0;
 $1 = sp + 184|0;
 $2 = sp + 40|0;
 $v = sp;
 $3 = sp + 168|0;
 HEAP32[$decryptedResponseData>>2] = 0;
 if ($isEncrypted) {
  HEAP32[$decryptedBytes>>2] = 0;
  $4 = ((($inBuff)) + 2|0);
  $5 = ((($this)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ((($6)) + 20|0);
  $8 = HEAP32[$7>>2]|0;
  (__ZN9ExitGames6Photon8Internal10Encryption7decryptEPKhiS4_PPhPi($4,$inBuffBodyLength,$8,$decryptedResponseData,$decryptedBytes)|0);
  $$pre = HEAP32[$decryptedResponseData>>2]|0;
  $$pre$phi39Z2D = $4;$10 = $$pre;
 } else {
  $$pre38 = ((($inBuff)) + 2|0);
  $$pre$phi39Z2D = $$pre38;$10 = 0;
 }
 $9 = ($10|0)!=(0|0);
 $11 = $9 ? $10 : $$pre$phi39Z2D;
 __ZN9ExitGames6Common7Helpers26DeSerializerImplementationC2EPKh($operationReader,$11);
 __THREW__ = 0;
 $12 = (invoke_ii(885,($operationReader|0))|0);
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if (!($14)) {
  __THREW__ = 0;
  $15 = (invoke_ii(886,($operationReader|0))|0);
  $16 = __THREW__; __THREW__ = 0;
  $17 = $16&1;
  if (!($17)) {
   __THREW__ = 0;
   invoke_viii(887,($operationResponse|0),($12|0),($15|0));
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    __THREW__ = 0;
    $20 = (invoke_ii(885,($operationReader|0))|0);
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if ($22) {
     label = 47;
    } else {
     __THREW__ = 0;
     invoke_vi(743,($debugMsg|0));
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if ($24) {
      label = 47;
     } else {
      __THREW__ = 0;
      (invoke_iiii(813,($operationReader|0),($20|0),($debugMsg|0))|0);
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      L12: do {
       if ($26) {
        label = 49;
       } else {
        __THREW__ = 0;
        $27 = (invoke_ii(848,($debugMsg|0))|0);
        $28 = __THREW__; __THREW__ = 0;
        $29 = $28&1;
        if ($29) {
         label = 49;
        } else {
         $30 = ($27<<24>>24)==(115);
         L15: do {
          if ($30) {
           __THREW__ = 0;
           invoke_vi(743,($2|0));
           $31 = __THREW__; __THREW__ = 0;
           $32 = $31&1;
           if ($32) {
            label = 49;
            break L12;
           }
           HEAP32[$2>>2] = (3056);
           __THREW__ = 0;
           $33 = (invoke_ii(848,($debugMsg|0))|0);
           $34 = __THREW__; __THREW__ = 0;
           $35 = $34&1;
           L18: do {
            if ($35) {
             label = 22;
            } else {
             $36 = ($33<<24>>24)==(115);
             if ($36) {
              __THREW__ = 0;
              $37 = (invoke_ii(846,($debugMsg|0))|0);
              $38 = __THREW__; __THREW__ = 0;
              $39 = $38&1;
              if ($39) {
               label = 22;
               break;
              }
              $40 = ($37|0)==(0);
              if ($40) {
               __THREW__ = 0;
               (invoke_iii(239,($2|0),($debugMsg|0))|0);
               $41 = __THREW__; __THREW__ = 0;
               $42 = $41&1;
               if ($42) {
                label = 22;
                break;
               }
              } else {
               label = 18;
              }
             } else {
              label = 18;
             }
             do {
              if ((label|0) == 18) {
               __THREW__ = 0;
               invoke_vi(743,($0|0));
               $43 = __THREW__; __THREW__ = 0;
               $44 = $43&1;
               if ($44) {
                label = 22;
                break L18;
               }
               __THREW__ = 0;
               (invoke_iii(239,($2|0),($0|0))|0);
               $45 = __THREW__; __THREW__ = 0;
               $46 = $45&1;
               if ($46) {
                $47 = ___cxa_find_matching_catch_2()|0;
                $48 = tempRet0;
                __ZN9ExitGames6Common6ObjectD2Ev($0);
                $eh$lpad$body21$index2Z2D = $48;$eh$lpad$body21$indexZ2D = $47;
                break L18;
               } else {
                __ZN9ExitGames6Common6ObjectD2Ev($0);
                break;
               }
              }
             } while(0);
             __THREW__ = 0;
             $51 = (invoke_ii(848,($2|0))|0);
             $52 = __THREW__; __THREW__ = 0;
             $53 = $52&1;
             L31: do {
              if (!($53)) {
               $54 = ($51<<24>>24)==(115);
               do {
                if ($54) {
                 __THREW__ = 0;
                 $55 = (invoke_ii(846,($2|0))|0);
                 $56 = __THREW__; __THREW__ = 0;
                 $57 = $56&1;
                 if ($57) {
                  break L31;
                 }
                 $58 = ($55|0)==(0);
                 if (!($58)) {
                  $62 = 0;
                  break;
                 }
                 __THREW__ = 0;
                 $59 = (invoke_ii(888,($2|0))|0);
                 $60 = __THREW__; __THREW__ = 0;
                 $61 = $60&1;
                 if ($61) {
                  break L31;
                 }
                 $62 = $59;
                } else {
                 $62 = 0;
                }
               } while(0);
               __THREW__ = 0;
               invoke_vii(786,($1|0),($62|0));
               $63 = __THREW__; __THREW__ = 0;
               $64 = $63&1;
               if (!($64)) {
                $$06 = 0;$$09 = 1;
                label = 32;
                break L15;
               }
              }
             } while(0);
             $96 = ___cxa_find_matching_catch_2()|0;
             $97 = tempRet0;
             $$1$ph29 = $96;$$12$ph28 = $97;$$pre$phiZ2D = $2;
             break L15;
            }
           } while(0);
           if ((label|0) == 22) {
            $49 = ___cxa_find_matching_catch_2()|0;
            $50 = tempRet0;
            $eh$lpad$body21$index2Z2D = $50;$eh$lpad$body21$indexZ2D = $49;
           }
           __ZN9ExitGames6Common6ObjectD2Ev($2);
           $eh$lpad$body$index8Z2D = $eh$lpad$body21$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body21$indexZ2D;
           label = 50;
           break L12;
          } else {
           __THREW__ = 0;
           invoke_vii(801,($1|0),(163772|0));
           $65 = __THREW__; __THREW__ = 0;
           $66 = $65&1;
           if ($66) {
            $98 = ___cxa_find_matching_catch_2()|0;
            $99 = tempRet0;
            $$2 = $98;$$23 = $99;
            break L12;
           } else {
            $$06 = 1;$$09 = 0;
            label = 32;
           }
          }
         } while(0);
         if ((label|0) == 32) {
          __THREW__ = 0;
          $67 = (invoke_ii(889,($1|0))|0);
          $68 = __THREW__; __THREW__ = 0;
          $69 = $68&1;
          if (!($69)) {
           __THREW__ = 0;
           invoke_vii(890,($operationResponse|0),($67|0));
           $70 = __THREW__; __THREW__ = 0;
           $71 = $70&1;
           if (!($71)) {
            if ($$06) {
             __ZN9ExitGames6Common7JStringD2Ev($1);
            }
            if ($$09) {
             __ZN9ExitGames6Common7JStringD2Ev($1);
             __ZN9ExitGames6Common6ObjectD2Ev($2);
            }
            __THREW__ = 0;
            $72 = (invoke_ii(886,($operationReader|0))|0);
            $73 = __THREW__; __THREW__ = 0;
            $74 = $73&1;
            if ($74) {
             label = 49;
             break;
            }
            $75 = $72 << 16 >> 16;
            $76 = ($72<<16>>16)>(0);
            L57: do {
             if ($76) {
              $i$033 = 0;
              while(1) {
               __THREW__ = 0;
               $77 = (invoke_ii(885,($operationReader|0))|0);
               $78 = __THREW__; __THREW__ = 0;
               $79 = $78&1;
               if ($79) {
                label = 48;
                break;
               }
               __THREW__ = 0;
               invoke_vi(743,($v|0));
               $80 = __THREW__; __THREW__ = 0;
               $81 = $80&1;
               if ($81) {
                label = 48;
                break;
               }
               __THREW__ = 0;
               (invoke_iii(891,($operationReader|0),($v|0))|0);
               $82 = __THREW__; __THREW__ = 0;
               $83 = $82&1;
               if ($83) {
                label = 58;
                break;
               }
               __THREW__ = 0;
               invoke_viii(892,($operationResponse|0),($77|0),($v|0));
               $84 = __THREW__; __THREW__ = 0;
               $85 = $84&1;
               if ($85) {
                label = 58;
                break;
               }
               __ZN9ExitGames6Common6ObjectD2Ev($v);
               $86 = (($i$033) + 1)|0;
               $87 = ($86|0)<($75|0);
               if ($87) {
                $i$033 = $86;
               } else {
                break L57;
               }
              }
              if ((label|0) == 48) {
               $92 = ___cxa_find_matching_catch_2()|0;
               $93 = tempRet0;
               $eh$lpad$body$index8Z2D = $93;$eh$lpad$body$indexZ2D = $92;
               label = 50;
               break L12;
              }
              else if ((label|0) == 58) {
               $102 = ___cxa_find_matching_catch_2()|0;
               $103 = tempRet0;
               __ZN9ExitGames6Common6ObjectD2Ev($v);
               $$2 = $102;$$23 = $103;
               break L12;
              }
             }
            } while(0);
            $104 = HEAP32[$decryptedResponseData>>2]|0;
            $105 = ($104|0)==(0|0);
            if (!($105)) {
             $106 = ((($104)) + -8|0);
             __THREW__ = 0;
             invoke_vi(750,($106|0));
             $107 = __THREW__; __THREW__ = 0;
             $108 = $107&1;
             if ($108) {
              label = 49;
              break;
             }
            }
            $109 = $msgType&255;
            switch ($109|0) {
            case 3:  {
             $110 = ((($this)) + 4|0);
             $111 = HEAP32[$110>>2]|0;
             $112 = ((($111)) + 44|0);
             $113 = HEAP32[$112>>2]|0;
             $114 = ($113|0)==(0);
             if ($114) {
              $124 = $111;$timeBeforeCallback$0 = 0;
             } else {
              $115 = ((($this)) + 112|0);
              $116 = HEAP32[$115>>2]|0;
              $117 = ((($this)) + 92|0);
              $118 = HEAP32[$117>>2]|0;
              __THREW__ = 0;
              invoke_vii(893,($116|0),($118|0));
              $119 = __THREW__; __THREW__ = 0;
              $120 = $119&1;
              if ($120) {
               label = 49;
               break L12;
              }
              __THREW__ = 0;
              $121 = (invoke_i(823)|0);
              $122 = __THREW__; __THREW__ = 0;
              $123 = $122&1;
              if ($123) {
               label = 49;
               break L12;
              }
              $$pre35 = HEAP32[$110>>2]|0;
              $124 = $$pre35;$timeBeforeCallback$0 = $121;
             }
             __THREW__ = 0;
             $125 = (invoke_ii(894,($124|0))|0);
             $126 = __THREW__; __THREW__ = 0;
             $127 = $126&1;
             if ($127) {
              label = 49;
              break L12;
             }
             $128 = HEAP32[$125>>2]|0;
             $129 = ((($128)) + 12|0);
             $130 = HEAP32[$129>>2]|0;
             __THREW__ = 0;
             invoke_vii($130|0,($125|0),($operationResponse|0));
             $131 = __THREW__; __THREW__ = 0;
             $132 = $131&1;
             if ($132) {
              label = 49;
              break L12;
             }
             $133 = HEAP32[$110>>2]|0;
             $134 = ((($133)) + 44|0);
             $135 = HEAP32[$134>>2]|0;
             $136 = ($135|0)==(0);
             if ($136) {
              __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
              __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
              __ZN9ExitGames6Common4BaseD2Ev($operationReader);
              STACKTOP = sp;return;
             }
             $137 = ((($this)) + 112|0);
             $138 = HEAP32[$137>>2]|0;
             __THREW__ = 0;
             $139 = (invoke_ii(754,($operationResponse|0))|0);
             $140 = __THREW__; __THREW__ = 0;
             $141 = $140&1;
             if ($141) {
              label = 49;
              break L12;
             }
             __THREW__ = 0;
             $142 = (invoke_i(823)|0);
             $143 = __THREW__; __THREW__ = 0;
             $144 = $143&1;
             if ($144) {
              label = 49;
              break L12;
             }
             $145 = (($142) - ($timeBeforeCallback$0))|0;
             __THREW__ = 0;
             invoke_viii(895,($138|0),($139|0),($145|0));
             $146 = __THREW__; __THREW__ = 0;
             $147 = $146&1;
             if ($147) {
              label = 49;
              break L12;
             }
             __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
             __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
             __ZN9ExitGames6Common4BaseD2Ev($operationReader);
             STACKTOP = sp;return;
             break;
            }
            case 7:  {
             $148 = ((($this)) + 4|0);
             $149 = HEAP32[$148>>2]|0;
             $150 = ((($149)) + 44|0);
             $151 = HEAP32[$150>>2]|0;
             $152 = ($151|0)==(0);
             if ($152) {
              $timeBeforeCallback$1 = 0;
             } else {
              $153 = ((($this)) + 112|0);
              $154 = HEAP32[$153>>2]|0;
              $155 = ((($this)) + 92|0);
              $156 = HEAP32[$155>>2]|0;
              __THREW__ = 0;
              invoke_vii(893,($154|0),($156|0));
              $157 = __THREW__; __THREW__ = 0;
              $158 = $157&1;
              if ($158) {
               label = 49;
               break L12;
              }
              __THREW__ = 0;
              $159 = (invoke_i(823)|0);
              $160 = __THREW__; __THREW__ = 0;
              $161 = $160&1;
              if ($161) {
               label = 49;
               break L12;
              } else {
               $timeBeforeCallback$1 = $159;
              }
             }
             __THREW__ = 0;
             $162 = (invoke_ii(754,($operationResponse|0))|0);
             $163 = __THREW__; __THREW__ = 0;
             $164 = $163&1;
             if ($164) {
              label = 49;
              break L12;
             }
             $165 = ($162<<24>>24)==(0);
             do {
              if ($165) {
               $166 = HEAP32[$this>>2]|0;
               $167 = ((($166)) + 124|0);
               $168 = HEAP32[$167>>2]|0;
               __THREW__ = 0;
               invoke_vii($168|0,($this|0),($operationResponse|0));
               $169 = __THREW__; __THREW__ = 0;
               $170 = $169&1;
               if ($170) {
                label = 49;
                break L12;
               }
              } else {
               __THREW__ = 0;
               $171 = (invoke_ii(754,($operationResponse|0))|0);
               $172 = __THREW__; __THREW__ = 0;
               $173 = $172&1;
               if ($173) {
                label = 49;
                break L12;
               }
               $174 = ($171<<24>>24)==(1);
               if (!($174)) {
                break;
               }
               $175 = HEAP32[$this>>2]|0;
               $176 = ((($175)) + 128|0);
               $177 = HEAP32[$176>>2]|0;
               __THREW__ = 0;
               invoke_vii($177|0,($this|0),($operationResponse|0));
               $178 = __THREW__; __THREW__ = 0;
               $179 = $178&1;
               if ($179) {
                label = 49;
                break L12;
               }
              }
             } while(0);
             $180 = HEAP32[$148>>2]|0;
             $181 = ((($180)) + 44|0);
             $182 = HEAP32[$181>>2]|0;
             $183 = ($182|0)==(0);
             if ($183) {
              __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
              __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
              __ZN9ExitGames6Common4BaseD2Ev($operationReader);
              STACKTOP = sp;return;
             }
             $184 = ((($this)) + 112|0);
             $185 = HEAP32[$184>>2]|0;
             __THREW__ = 0;
             $186 = (invoke_ii(754,($operationResponse|0))|0);
             $187 = __THREW__; __THREW__ = 0;
             $188 = $187&1;
             if ($188) {
              label = 49;
              break L12;
             }
             __THREW__ = 0;
             $189 = (invoke_i(823)|0);
             $190 = __THREW__; __THREW__ = 0;
             $191 = $190&1;
             if ($191) {
              label = 49;
              break L12;
             }
             $192 = (($189) - ($timeBeforeCallback$1))|0;
             __THREW__ = 0;
             invoke_viii(895,($185|0),($186|0),($192|0));
             $193 = __THREW__; __THREW__ = 0;
             $194 = $193&1;
             if ($194) {
              label = 49;
              break L12;
             }
             __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
             __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
             __ZN9ExitGames6Common4BaseD2Ev($operationReader);
             STACKTOP = sp;return;
             break;
            }
            default: {
             $195 = ((($this)) + 8|0);
             $196 = HEAP32[$195>>2]|0;
             __THREW__ = 0;
             invoke_vii(739,($3|0),(147154|0));
             $197 = __THREW__; __THREW__ = 0;
             $198 = $197&1;
             if ($198) {
              label = 49;
              break L12;
             }
             __THREW__ = 0;
             $199 = (invoke_ii(764,($3|0))|0);
             $200 = __THREW__; __THREW__ = 0;
             $201 = $200&1;
             do {
              if (!($201)) {
               __THREW__ = 0;
               HEAP32[$vararg_buffer>>2] = $109;
               invoke_viiiiiiii(765,($196|0),1,(6560|0),($199|0),1,614,(7448|0),($vararg_buffer|0));
               $202 = __THREW__; __THREW__ = 0;
               $203 = $202&1;
               if ($203) {
                break;
               }
               __ZN9ExitGames6Common7JStringD2Ev($3);
               __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
               __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
               __ZN9ExitGames6Common4BaseD2Ev($operationReader);
               STACKTOP = sp;return;
              }
             } while(0);
             $204 = ___cxa_find_matching_catch_2()|0;
             $205 = tempRet0;
             __ZN9ExitGames6Common7JStringD2Ev($3);
             $$2 = $204;$$23 = $205;
             break L12;
            }
            }
           }
          }
          $100 = ___cxa_find_matching_catch_2()|0;
          $101 = tempRet0;
          if ($$06) {
           __ZN9ExitGames6Common7JStringD2Ev($1);
           if (!($$09)) {
            $$2 = $100;$$23 = $101;
            break;
           }
          } else {
           if (!($$09)) {
            $$2 = $100;$$23 = $101;
            break;
           }
          }
          __ZN9ExitGames6Common7JStringD2Ev($1);
          $$1$ph29 = $100;$$12$ph28 = $101;$$pre$phiZ2D = $2;
         }
         __ZN9ExitGames6Common6ObjectD2Ev($$pre$phiZ2D);
         $$2 = $$1$ph29;$$23 = $$12$ph28;
        }
       }
      } while(0);
      if ((label|0) == 49) {
       $94 = ___cxa_find_matching_catch_2()|0;
       $95 = tempRet0;
       $eh$lpad$body$index8Z2D = $95;$eh$lpad$body$indexZ2D = $94;
       label = 50;
      }
      if ((label|0) == 50) {
       $$2 = $eh$lpad$body$indexZ2D;$$23 = $eh$lpad$body$index8Z2D;
      }
      __ZN9ExitGames6Common6ObjectD2Ev($debugMsg);
      $$3 = $$2;$$34 = $$23;
     }
    }
    if ((label|0) == 47) {
     $90 = ___cxa_find_matching_catch_2()|0;
     $91 = tempRet0;
     $$3 = $90;$$34 = $91;
    }
    __ZN9ExitGames6Photon17OperationResponseD2Ev($operationResponse);
    $$4 = $$3;$$45 = $$34;
    __ZN9ExitGames6Common4BaseD2Ev($operationReader);
    ___resumeException($$4|0);
    // unreachable;
   }
  }
 }
 $88 = ___cxa_find_matching_catch_2()|0;
 $89 = tempRet0;
 $$4 = $88;$$45 = $89;
 __ZN9ExitGames6Common4BaseD2Ev($operationReader);
 ___resumeException($$4|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase16deserializeEventEPhbi($this,$inBuff,$isEncrypted,$inBuffBodyLength) {
 $this = $this|0;
 $inBuff = $inBuff|0;
 $isEncrypted = $isEncrypted|0;
 $inBuffBodyLength = $inBuffBodyLength|0;
 var $$0 = 0, $$0$i$i = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$pre21 = 0, $$pre23 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $decryptedBytes = 0, $decryptedResponseData = 0, $eh$lpad$body$index9Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body17$index4Z2D = 0, $eh$lpad$body17$indexZ2D = 0, $eventData = 0, $i$020 = 0, $k = 0, $operationReader = 0, $or$cond$i$i = 0, $parameters = 0, $timeBeforeCallback$0 = 0, $v = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $0 = sp + 120|0;
 $decryptedResponseData = sp + 284|0;
 $decryptedBytes = sp + 280|0;
 $operationReader = sp + 268|0;
 $parameters = sp + 216|0;
 $k = sp + 80|0;
 $v = sp + 40|0;
 $1 = sp + 288|0;
 $2 = sp;
 $eventData = sp + 160|0;
 HEAP32[$decryptedResponseData>>2] = 0;
 if ($isEncrypted) {
  HEAP32[$decryptedBytes>>2] = 0;
  $3 = ((($inBuff)) + 2|0);
  $4 = ((($this)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($5)) + 20|0);
  $7 = HEAP32[$6>>2]|0;
  (__ZN9ExitGames6Photon8Internal10Encryption7decryptEPKhiS4_PPhPi($3,$inBuffBodyLength,$7,$decryptedResponseData,$decryptedBytes)|0);
  $$pre = HEAP32[$decryptedResponseData>>2]|0;
  $$pre$phiZ2D = $3;$9 = $$pre;
 } else {
  $$pre23 = ((($inBuff)) + 2|0);
  $$pre$phiZ2D = $$pre23;$9 = 0;
 }
 $8 = ($9|0)!=(0|0);
 $10 = $8 ? $9 : $$pre$phiZ2D;
 __ZN9ExitGames6Common7Helpers26DeSerializerImplementationC2EPKh($operationReader,$10);
 __THREW__ = 0;
 $11 = (invoke_ii(885,($operationReader|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if (!($13)) {
  __THREW__ = 0;
  $14 = (invoke_ii(886,($operationReader|0))|0);
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if (!($16)) {
   $17 = $14 << 16 >> 16;
   __THREW__ = 0;
   invoke_vi(896,($parameters|0));
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if (!($19)) {
    HEAP32[$parameters>>2] = (2720);
    $20 = HEAP8[167373]|0;
    $21 = ($20<<24>>24)==(0);
    if ($21) {
     HEAP8[167374] = 98;
     HEAP8[167375] = 122;
     HEAP32[40417] = 0;
     HEAP8[167373] = 1;
    }
    $22 = ($14<<16>>16)>(0);
    L12: do {
     if ($22) {
      $23 = ((($parameters)) + 4|0);
      $i$020 = 0;
      while(1) {
       __THREW__ = 0;
       invoke_vi(743,($k|0));
       $24 = __THREW__; __THREW__ = 0;
       $25 = $24&1;
       if ($25) {
        label = 36;
        break L12;
       }
       __THREW__ = 0;
       invoke_vi(743,($v|0));
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       if ($27) {
        label = 37;
        break;
       }
       __THREW__ = 0;
       invoke_vii(897,($operationReader|0),($k|0));
       $28 = __THREW__; __THREW__ = 0;
       $29 = $28&1;
       if ($29) {
        label = 38;
        break;
       }
       __THREW__ = 0;
       (invoke_iii(891,($operationReader|0),($v|0))|0);
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        label = 38;
        break;
       }
       __THREW__ = 0;
       invoke_vi(743,($2|0));
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        label = 38;
        break;
       }
       HEAP32[$2>>2] = (2804);
       __THREW__ = 0;
       $34 = (invoke_ii(848,($k|0))|0);
       $35 = __THREW__; __THREW__ = 0;
       $36 = $35&1;
       if ($36) {
        label = 25;
        break;
       }
       $37 = ($34<<24>>24)==(98);
       if ($37) {
        __THREW__ = 0;
        $38 = (invoke_ii(846,($k|0))|0);
        $39 = __THREW__; __THREW__ = 0;
        $40 = $39&1;
        if ($40) {
         label = 25;
         break;
        }
        $41 = ($38|0)==(0);
        if ($41) {
         __THREW__ = 0;
         (invoke_iii(239,($2|0),($k|0))|0);
         $42 = __THREW__; __THREW__ = 0;
         $43 = $42&1;
         if ($43) {
          label = 25;
          break;
         }
        } else {
         label = 21;
        }
       } else {
        label = 21;
       }
       if ((label|0) == 21) {
        label = 0;
        __THREW__ = 0;
        invoke_vi(743,($0|0));
        $44 = __THREW__; __THREW__ = 0;
        $45 = $44&1;
        if ($45) {
         label = 25;
         break;
        }
        __THREW__ = 0;
        (invoke_iii(239,($2|0),($0|0))|0);
        $46 = __THREW__; __THREW__ = 0;
        $47 = $46&1;
        if ($47) {
         label = 24;
         break;
        }
        __ZN9ExitGames6Common6ObjectD2Ev($0);
       }
       __THREW__ = 0;
       $52 = (invoke_ii(843,($2|0))|0);
       $53 = __THREW__; __THREW__ = 0;
       $54 = $53&1;
       if ($54) {
        label = 40;
        break;
       }
       __THREW__ = 0;
       $55 = __THREW__; __THREW__ = 0;
       $56 = $55&1;
       if ($56) {
        label = 40;
        break;
       }
       __THREW__ = 0;
       $57 = (invoke_ii(848,($2|0))|0);
       $58 = __THREW__; __THREW__ = 0;
       $59 = $58&1;
       if ($59) {
        label = 40;
        break;
       }
       __THREW__ = 0;
       $60 = (invoke_ii(846,($2|0))|0);
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if ($62) {
        label = 40;
        break;
       }
       $63 = ($57<<24>>24)!=(98);
       $64 = ($60|0)!=(0);
       $or$cond$i$i = $63 | $64;
       if ($or$cond$i$i) {
        $$0$i$i = 0;
       } else {
        $65 = HEAP8[$52>>0]|0;
        $$0$i$i = $65;
       }
       HEAP8[$1>>0] = $$0$i$i;
       __THREW__ = 0;
       invoke_viii(898,($23|0),($1|0),($v|0));
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        label = 40;
        break;
       }
       __ZN9ExitGames6Common6ObjectD2Ev($2);
       __ZN9ExitGames6Common6ObjectD2Ev($v);
       __ZN9ExitGames6Common6ObjectD2Ev($k);
       $68 = (($i$020) + 1)|0;
       $69 = ($68|0)<($17|0);
       if ($69) {
        $i$020 = $68;
       } else {
        label = 43;
        break L12;
       }
      }
      if ((label|0) == 24) {
       $48 = ___cxa_find_matching_catch_2()|0;
       $49 = tempRet0;
       __ZN9ExitGames6Common6ObjectD2Ev($0);
       $eh$lpad$body17$index4Z2D = $49;$eh$lpad$body17$indexZ2D = $48;
       label = 26;
      }
      else if ((label|0) == 25) {
       $50 = ___cxa_find_matching_catch_2()|0;
       $51 = tempRet0;
       $eh$lpad$body17$index4Z2D = $51;$eh$lpad$body17$indexZ2D = $50;
       label = 26;
      }
      else if ((label|0) == 37) {
       $74 = ___cxa_find_matching_catch_2()|0;
       $75 = tempRet0;
       $$1 = $74;$$12 = $75;
      }
      else if ((label|0) == 38) {
       $76 = ___cxa_find_matching_catch_2()|0;
       $77 = tempRet0;
       $eh$lpad$body$index9Z2D = $77;$eh$lpad$body$indexZ2D = $76;
       label = 39;
      }
      else if ((label|0) == 40) {
       $78 = ___cxa_find_matching_catch_2()|0;
       $79 = tempRet0;
       __ZN9ExitGames6Common6ObjectD2Ev($2);
       $$0 = $78;$$01 = $79;
       label = 41;
      }
      if ((label|0) == 26) {
       __ZN9ExitGames6Common6ObjectD2Ev($2);
       $eh$lpad$body$index9Z2D = $eh$lpad$body17$index4Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body17$indexZ2D;
       label = 39;
      }
      if ((label|0) == 39) {
       $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index9Z2D;
       label = 41;
      }
      if ((label|0) == 41) {
       __ZN9ExitGames6Common6ObjectD2Ev($v);
       $$1 = $$0;$$12 = $$01;
      }
      __ZN9ExitGames6Common6ObjectD2Ev($k);
      $$2 = $$1;$$23 = $$12;
     } else {
      label = 43;
     }
    } while(0);
    do {
     if ((label|0) == 43) {
      $80 = HEAP32[$decryptedResponseData>>2]|0;
      $81 = ($80|0)==(0|0);
      if (!($81)) {
       $82 = ((($80)) + -8|0);
       __THREW__ = 0;
       invoke_vi(750,($82|0));
       $83 = __THREW__; __THREW__ = 0;
       $84 = $83&1;
       if ($84) {
        label = 36;
        break;
       }
      }
      __THREW__ = 0;
      invoke_viii(899,($eventData|0),($11|0),($parameters|0));
      $85 = __THREW__; __THREW__ = 0;
      $86 = $85&1;
      if ($86) {
       label = 36;
      } else {
       $87 = ((($this)) + 4|0);
       $88 = HEAP32[$87>>2]|0;
       $89 = ((($88)) + 44|0);
       $90 = HEAP32[$89>>2]|0;
       $91 = ($90|0)==(0);
       if ($91) {
        $103 = $88;$timeBeforeCallback$0 = 0;
        label = 51;
       } else {
        $92 = ((($this)) + 112|0);
        $93 = HEAP32[$92>>2]|0;
        $94 = ((($this)) + 92|0);
        $95 = HEAP32[$94>>2]|0;
        __THREW__ = 0;
        invoke_vii(900,($93|0),($95|0));
        $96 = __THREW__; __THREW__ = 0;
        $97 = $96&1;
        if (!($97)) {
         __THREW__ = 0;
         $98 = (invoke_i(823)|0);
         $99 = __THREW__; __THREW__ = 0;
         $100 = $99&1;
         if (!($100)) {
          $$pre21 = HEAP32[$87>>2]|0;
          $103 = $$pre21;$timeBeforeCallback$0 = $98;
          label = 51;
         }
        }
       }
       if ((label|0) == 51) {
        __THREW__ = 0;
        $104 = (invoke_ii(894,($103|0))|0);
        $105 = __THREW__; __THREW__ = 0;
        $106 = $105&1;
        if (!($106)) {
         $107 = HEAP32[$104>>2]|0;
         $108 = ((($107)) + 20|0);
         $109 = HEAP32[$108>>2]|0;
         __THREW__ = 0;
         invoke_vii($109|0,($104|0),($eventData|0));
         $110 = __THREW__; __THREW__ = 0;
         $111 = $110&1;
         if (!($111)) {
          $112 = HEAP32[$87>>2]|0;
          $113 = ((($112)) + 44|0);
          $114 = HEAP32[$113>>2]|0;
          $115 = ($114|0)==(0);
          if ($115) {
           __ZN9ExitGames6Photon9EventDataD2Ev($eventData);
           __ZN9ExitGames6Common14DictionaryBaseD2Ev($parameters);
           __ZN9ExitGames6Common4BaseD2Ev($operationReader);
           STACKTOP = sp;return;
          }
          $116 = ((($this)) + 112|0);
          $117 = HEAP32[$116>>2]|0;
          __THREW__ = 0;
          $118 = (invoke_ii(901,($eventData|0))|0);
          $119 = __THREW__; __THREW__ = 0;
          $120 = $119&1;
          if (!($120)) {
           __THREW__ = 0;
           $121 = (invoke_i(823)|0);
           $122 = __THREW__; __THREW__ = 0;
           $123 = $122&1;
           if (!($123)) {
            $124 = (($121) - ($timeBeforeCallback$0))|0;
            __THREW__ = 0;
            invoke_viii(902,($117|0),($118|0),($124|0));
            $125 = __THREW__; __THREW__ = 0;
            $126 = $125&1;
            if (!($126)) {
             __ZN9ExitGames6Photon9EventDataD2Ev($eventData);
             __ZN9ExitGames6Common14DictionaryBaseD2Ev($parameters);
             __ZN9ExitGames6Common4BaseD2Ev($operationReader);
             STACKTOP = sp;return;
            }
           }
          }
         }
        }
       }
       $101 = ___cxa_find_matching_catch_2()|0;
       $102 = tempRet0;
       __ZN9ExitGames6Photon9EventDataD2Ev($eventData);
       $$2 = $101;$$23 = $102;
      }
     }
    } while(0);
    if ((label|0) == 36) {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $$2 = $72;$$23 = $73;
    }
    __ZN9ExitGames6Common14DictionaryBaseD2Ev($parameters);
    $$3 = $$2;$$34 = $$23;
    __ZN9ExitGames6Common4BaseD2Ev($operationReader);
    ___resumeException($$3|0);
    // unreachable;
   }
  }
 }
 $70 = ___cxa_find_matching_catch_2()|0;
 $71 = tempRet0;
 $$3 = $70;$$34 = $71;
 __ZN9ExitGames6Common4BaseD2Ev($operationReader);
 ___resumeException($$3|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase30updateRoundTripTimeAndVarianceEi($this,$time) {
 $this = $this|0;
 $time = $time|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ispos = 0, $neg = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147123);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($2|0),4,(6560|0),($3|0),1,660,(163772|0),($vararg_buffer|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($9|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $8 = ($time|0)<(0);
 if ($8) {
  STACKTOP = sp;return;
 }
 $11 = ((($this)) + 60|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (($time) - ($12))|0;
 $14 = (($13|0) / 8)&-1;
 $15 = (($14) + ($12))|0;
 HEAP32[$11>>2] = $15;
 $16 = (($time) - ($15))|0;
 $ispos = ($16|0)>(-1);
 $neg = (0 - ($16))|0;
 $17 = $ispos ? $16 : $neg;
 $18 = (($17|0) / 4)&-1;
 $19 = ((($this)) + 64|0);
 $20 = HEAP32[$19>>2]|0;
 $21 = ($20*3)|0;
 $22 = (($21|0) / 4)&-1;
 $23 = (($22) + ($18))|0;
 HEAP32[$19>>2] = $23;
 $24 = ((($this)) + 72|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($15|0)<($25|0);
 if ($26) {
  HEAP32[$24>>2] = $15;
 }
 $27 = ((($this)) + 80|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($23|0)>($28|0);
 if (!($29)) {
  STACKTOP = sp;return;
 }
 HEAP32[$27>>2] = $23;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase27opExchangeKeysForEncryptionEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$off = 0, $$off4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $8 = 0, $9 = 0, $OakleyPrime768 = 0, $operationRequest = 0, $wrapDict = 0, dest = 0, label = 0;
 var sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(256|0);
 $OakleyPrime768 = sp + 152|0;
 $wrapDict = sp + 96|0;
 $0 = sp;
 $operationRequest = sp + 40|0;
 $1 = (_EGBN_new()|0);
 $2 = (_EGBN_new()|0);
 $3 = (_EGBN_new()|0);
 $4 = (_EGBN_new()|0);
 dest=$OakleyPrime768; src=146618; stop=dest+96|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $5 = (_EGBN_CTX_new()|0);
 (_EGBN_set_word($3,22)|0);
 (_EGBN_bin2bn($OakleyPrime768,96,$4)|0);
 (_EGBN_rand($1,160,-1,0)|0);
 (_EGBN_mod_exp($2,$3,$1,$4,$5)|0);
 _EGBN_CTX_free($5);
 $6 = (_EGBN_num_bits($1)|0);
 $7 = (($6) + 7)|0;
 $8 = (($7|0) / 8)&-1;
 $9 = ((($this)) + 144|0);
 HEAP32[$9>>2] = $8;
 $10 = (_EGBN_num_bits($2)|0);
 $11 = (($10) + 7)|0;
 $12 = (($11|0) / 8)&-1;
 $13 = (_EGBN_num_bits($1)|0);
 $14 = (($13) + 7)|0;
 $15 = (($14|0) / 8)&-1;
 $16 = (($15) + 8)|0;
 $17 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($16)|0);
 $18 = ((($17)) + 8|0);
 $19 = $17;
 $20 = $19;
 HEAP32[$20>>2] = $15;
 $21 = (($19) + 4)|0;
 $22 = $21;
 HEAP32[$22>>2] = 0;
 $$off = (($13) + 14)|0;
 $23 = ($$off>>>0)<(15);
 if (!($23)) {
  _memset(($18|0),0,($15|0))|0;
 }
 $24 = ((($this)) + 140|0);
 HEAP32[$24>>2] = $18;
 $25 = (_EGBN_num_bits($2)|0);
 $26 = (($25) + 7)|0;
 $27 = (($26|0) / 8)&-1;
 $28 = (($27) + 8)|0;
 $29 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($28)|0);
 $30 = ((($29)) + 8|0);
 $31 = $29;
 $32 = $31;
 HEAP32[$32>>2] = $27;
 $33 = (($31) + 4)|0;
 $34 = $33;
 HEAP32[$34>>2] = 0;
 $$off4 = (($25) + 14)|0;
 $35 = ($$off4>>>0)<(15);
 if (!($35)) {
  _memset(($30|0),0,($27|0))|0;
 }
 $36 = HEAP32[$24>>2]|0;
 (_EGBN_bn2bin($1,$36)|0);
 (_EGBN_bn2bin($2,$30)|0);
 $37 = ((($this)) + 4|0);
 $38 = HEAP32[$37>>2]|0;
 $39 = ((($38)) + 25|0);
 HEAP8[$39>>0] = 0;
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($wrapDict);
 HEAP32[$wrapDict>>2] = (2720);
 $40 = HEAP8[167373]|0;
 $41 = ($40<<24>>24)==(0);
 if ($41) {
  HEAP8[167374] = 98;
  HEAP8[167375] = 122;
  HEAP32[40417] = 0;
  HEAP8[167373] = 1;
 }
 __THREW__ = 0;
 invoke_viiiiii(903,($0|0),($30|0),98,0,($12|0),1);
 $42 = __THREW__; __THREW__ = 0;
 $43 = $42&1;
 if (!($43)) {
  HEAP32[$0>>2] = (6184);
  $44 = ((($wrapDict)) + 4|0);
  __THREW__ = 0;
  invoke_viii(898,($44|0),(147122|0),($0|0));
  $45 = __THREW__; __THREW__ = 0;
  $46 = $45&1;
  if ($46) {
   $67 = ___cxa_find_matching_catch_2()|0;
   $68 = tempRet0;
   __ZN9ExitGames6Common6ObjectD2Ev($0);
   $$0 = $68;$$01 = $67;
   __ZN9ExitGames6Common14DictionaryBaseD2Ev($wrapDict);
   ___resumeException($$01|0);
   // unreachable;
  }
  __ZN9ExitGames6Common6ObjectD2Ev($0);
  __THREW__ = 0;
  invoke_viii(740,($operationRequest|0),0,($wrapDict|0));
  $47 = __THREW__; __THREW__ = 0;
  $48 = $47&1;
  if (!($48)) {
   $49 = HEAP32[$this>>2]|0;
   $50 = ((($49)) + 48|0);
   $51 = HEAP32[$50>>2]|0;
   __THREW__ = 0;
   $52 = (invoke_iiiiiii($51|0,($this|0),($operationRequest|0),1,0,0,6)|0);
   $53 = __THREW__; __THREW__ = 0;
   $54 = $53&1;
   if (!($54)) {
    __THREW__ = 0;
    invoke_vi(750,($29|0));
    $55 = __THREW__; __THREW__ = 0;
    $56 = $55&1;
    if (!($56)) {
     __THREW__ = 0;
     invoke_vi(904,($1|0));
     $57 = __THREW__; __THREW__ = 0;
     $58 = $57&1;
     if (!($58)) {
      __THREW__ = 0;
      invoke_vi(904,($2|0));
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if (!($60)) {
       __THREW__ = 0;
       invoke_vi(904,($3|0));
       $61 = __THREW__; __THREW__ = 0;
       $62 = $61&1;
       if (!($62)) {
        __THREW__ = 0;
        invoke_vi(904,($4|0));
        $63 = __THREW__; __THREW__ = 0;
        $64 = $63&1;
        if (!($64)) {
         __ZN9ExitGames6Photon16OperationRequestD2Ev($operationRequest);
         __ZN9ExitGames6Common14DictionaryBaseD2Ev($wrapDict);
         STACKTOP = sp;return ($52|0);
        }
       }
      }
     }
    }
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __ZN9ExitGames6Photon16OperationRequestD2Ev($operationRequest);
   $$0 = $70;$$01 = $69;
   __ZN9ExitGames6Common14DictionaryBaseD2Ev($wrapDict);
   ___resumeException($$01|0);
   // unreachable;
  }
 }
 $65 = ___cxa_find_matching_catch_2()|0;
 $66 = tempRet0;
 $$0 = $66;$$01 = $65;
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($wrapDict);
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal8PeerBase15deriveSharedKeyEPKNS0_17OperationResponseE($this,$operationResponse) {
 $this = $this|0;
 $operationResponse = $operationResponse|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$off = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $OakleyPrime768 = 0, $eh$lpad$body$index7Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body23$index2Z2D = 0, $eh$lpad$body23$indexZ2D = 0, $phitmp = 0, $phitmp4 = 0, $val = 0, $vararg_buffer = 0;
 var dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $vararg_buffer = sp + 120|0;
 $0 = sp + 80|0;
 $val = sp + 40|0;
 $1 = sp;
 $2 = sp + 128|0;
 $OakleyPrime768 = sp + 144|0;
 $3 = (__ZNK9ExitGames6Photon17OperationResponse13getReturnCodeEv($operationResponse)|0);
 $4 = ($3<<16>>16)==(0);
 if (!($4)) {
  $5 = ((($this)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($6)|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ((($8)) + 16|0);
  $10 = HEAP32[$9>>2]|0;
  FUNCTION_TABLE_vii[$10 & 2047]($7,1049);
  STACKTOP = sp;return;
 }
 __ZNK9ExitGames6Photon17OperationResponse19getParameterForCodeEh($1,$operationResponse,1);
 __THREW__ = 0;
 invoke_vi(743,($val|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $68 = ___cxa_find_matching_catch_2()|0;
  $69 = tempRet0;
  $eh$lpad$body$index7Z2D = $69;$eh$lpad$body$indexZ2D = $68;
 } else {
  HEAP32[$val>>2] = (6184);
  __THREW__ = 0;
  $13 = (invoke_ii(848,($1|0))|0);
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  L8: do {
   if ($15) {
    label = 13;
   } else {
    $16 = ($13<<24>>24)==(98);
    if ($16) {
     __THREW__ = 0;
     $17 = (invoke_ii(846,($1|0))|0);
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      label = 13;
      break;
     }
     $20 = ($17|0)==(1);
     if ($20) {
      __THREW__ = 0;
      (invoke_iii(239,($val|0),($1|0))|0);
      $21 = __THREW__; __THREW__ = 0;
      $22 = $21&1;
      if ($22) {
       label = 13;
       break;
      }
     } else {
      label = 9;
     }
    } else {
     label = 9;
    }
    do {
     if ((label|0) == 9) {
      __THREW__ = 0;
      invoke_vi(743,($0|0));
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      if ($24) {
       label = 13;
       break L8;
      }
      __THREW__ = 0;
      (invoke_iii(239,($val|0),($0|0))|0);
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       $27 = ___cxa_find_matching_catch_2()|0;
       $28 = tempRet0;
       __ZN9ExitGames6Common6ObjectD2Ev($0);
       $eh$lpad$body23$index2Z2D = $28;$eh$lpad$body23$indexZ2D = $27;
       break L8;
      } else {
       __ZN9ExitGames6Common6ObjectD2Ev($0);
       break;
      }
     }
    } while(0);
    __ZN9ExitGames6Common6ObjectD2Ev($1);
    __THREW__ = 0;
    $31 = __THREW__; __THREW__ = 0;
    $32 = $31&1;
    L21: do {
     if ($32) {
      label = 30;
     } else {
      __THREW__ = 0;
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if ($34) {
       label = 30;
      } else {
       __THREW__ = 0;
       $35 = (invoke_ii(888,($val|0))|0);
       $36 = __THREW__; __THREW__ = 0;
       $37 = $36&1;
       if ($37) {
        label = 30;
       } else {
        $38 = HEAP32[$35>>2]|0;
        __THREW__ = 0;
        $39 = (invoke_ii(847,($val|0))|0);
        $40 = __THREW__; __THREW__ = 0;
        $41 = $40&1;
        if ($41) {
         label = 30;
        } else {
         $42 = ($39|0)==(0|0);
         if ($42) {
          $55 = 0;
         } else {
          __THREW__ = 0;
          $43 = (invoke_ii(847,($val|0))|0);
          $44 = __THREW__; __THREW__ = 0;
          $45 = $44&1;
          if ($45) {
           label = 30;
           break;
          }
          $46 = HEAP32[$43>>2]|0;
          $phitmp = $46 << 16;
          $phitmp4 = $phitmp >> 16;
          $55 = $phitmp4;
         }
         $47 = ($38|0)==(0|0);
         do {
          if ($47) {
           $48 = ((($this)) + 8|0);
           $49 = HEAP32[$48>>2]|0;
           __THREW__ = 0;
           invoke_vii(739,($2|0),(146602|0));
           $50 = __THREW__; __THREW__ = 0;
           $51 = $50&1;
           if ($51) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $52 = (invoke_ii(764,($2|0))|0);
           $53 = __THREW__; __THREW__ = 0;
           $54 = $53&1;
           if (!($54)) {
            __THREW__ = 0;
            HEAP32[$vararg_buffer>>2] = $55;
            invoke_viiiiiiii(765,($49|0),1,(6560|0),($52|0),1,744,(6712|0),($vararg_buffer|0));
            $56 = __THREW__; __THREW__ = 0;
            $57 = $56&1;
            if (!($57)) {
             __ZN9ExitGames6Common7JStringD2Ev($2);
             $58 = ((($this)) + 4|0);
             $59 = HEAP32[$58>>2]|0;
             __THREW__ = 0;
             $60 = (invoke_ii(894,($59|0))|0);
             $61 = __THREW__; __THREW__ = 0;
             $62 = $61&1;
             if ($62) {
              label = 30;
              break L21;
             }
             $63 = HEAP32[$60>>2]|0;
             $64 = ((($63)) + 16|0);
             $65 = HEAP32[$64>>2]|0;
             __THREW__ = 0;
             invoke_vii($65|0,($60|0),1049);
             $66 = __THREW__; __THREW__ = 0;
             $67 = $66&1;
             if ($67) {
              label = 30;
              break L21;
             } else {
              break;
             }
            }
           }
           $72 = ___cxa_find_matching_catch_2()|0;
           $73 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($2);
           $$0 = $72;$$01 = $73;
           break L21;
          } else {
           __THREW__ = 0;
           $74 = (invoke_i(905)|0);
           $75 = __THREW__; __THREW__ = 0;
           $76 = $75&1;
           if ($76) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $77 = (invoke_i(905)|0);
           $78 = __THREW__; __THREW__ = 0;
           $79 = $78&1;
           if ($79) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $80 = (invoke_i(905)|0);
           $81 = __THREW__; __THREW__ = 0;
           $82 = $81&1;
           if ($82) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $83 = (invoke_i(905)|0);
           $84 = __THREW__; __THREW__ = 0;
           $85 = $84&1;
           if ($85) {
            label = 30;
            break L21;
           }
           dest=$OakleyPrime768; src=146618; stop=dest+96|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
           __THREW__ = 0;
           $86 = (invoke_i(906)|0);
           $87 = __THREW__; __THREW__ = 0;
           $88 = $87&1;
           if ($88) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           (invoke_iiii(907,($OakleyPrime768|0),96,($77|0))|0);
           $89 = __THREW__; __THREW__ = 0;
           $90 = $89&1;
           if ($90) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           (invoke_iiii(907,($38|0),($55|0),($83|0))|0);
           $91 = __THREW__; __THREW__ = 0;
           $92 = $91&1;
           if ($92) {
            label = 30;
            break L21;
           }
           $93 = ((($this)) + 140|0);
           $94 = HEAP32[$93>>2]|0;
           $95 = ((($this)) + 144|0);
           $96 = HEAP32[$95>>2]|0;
           __THREW__ = 0;
           (invoke_iiii(907,($94|0),($96|0),($74|0))|0);
           $97 = __THREW__; __THREW__ = 0;
           $98 = $97&1;
           if ($98) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           (invoke_iiiiii(908,($80|0),($83|0),($74|0),($77|0),($86|0))|0);
           $99 = __THREW__; __THREW__ = 0;
           $100 = $99&1;
           if ($100) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(909,($86|0));
           $101 = __THREW__; __THREW__ = 0;
           $102 = $101&1;
           if ($102) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $103 = (invoke_ii(910,($80|0))|0);
           $104 = __THREW__; __THREW__ = 0;
           $105 = $104&1;
           if ($105) {
            label = 30;
            break L21;
           }
           $106 = (($103) + 7)|0;
           $107 = (($106|0) / 8)&-1;
           $108 = (($107) + 8)|0;
           __THREW__ = 0;
           $109 = (invoke_ii(775,($108|0))|0);
           $110 = __THREW__; __THREW__ = 0;
           $111 = $110&1;
           if ($111) {
            label = 30;
            break L21;
           }
           $112 = ((($109)) + 8|0);
           $113 = $109;
           $114 = $113;
           HEAP32[$114>>2] = $107;
           $115 = (($113) + 4)|0;
           $116 = $115;
           HEAP32[$116>>2] = 0;
           $$off = (($103) + 14)|0;
           $117 = ($$off>>>0)<(15);
           if (!($117)) {
            _memset(($112|0),0,($107|0))|0;
           }
           __THREW__ = 0;
           (invoke_iii(911,($80|0),($112|0))|0);
           $118 = __THREW__; __THREW__ = 0;
           $119 = $118&1;
           if ($119) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           $120 = (invoke_ii(775,40)|0);
           $121 = __THREW__; __THREW__ = 0;
           $122 = $121&1;
           if ($122) {
            label = 30;
            break L21;
           }
           $123 = ((($120)) + 8|0);
           $124 = $120;
           $125 = $124;
           HEAP32[$125>>2] = 32;
           $126 = (($124) + 4)|0;
           $127 = $126;
           HEAP32[$127>>2] = 0;
           dest=$123; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
           __THREW__ = 0;
           $128 = (invoke_ii(910,($80|0))|0);
           $129 = __THREW__; __THREW__ = 0;
           $130 = $129&1;
           if ($130) {
            label = 30;
            break L21;
           }
           $131 = (($128) + 7)|0;
           $132 = (($131|0) / 8)&-1;
           __THREW__ = 0;
           invoke_viii(912,($112|0),($132|0),($123|0));
           $133 = __THREW__; __THREW__ = 0;
           $134 = $133&1;
           if ($134) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(750,($109|0));
           $135 = __THREW__; __THREW__ = 0;
           $136 = $135&1;
           if ($136) {
            label = 30;
            break L21;
           }
           $137 = HEAP32[$93>>2]|0;
           $138 = ($137|0)==(0|0);
           if (!($138)) {
            $139 = ((($137)) + -8|0);
            __THREW__ = 0;
            invoke_vi(750,($139|0));
            $140 = __THREW__; __THREW__ = 0;
            $141 = $140&1;
            if ($141) {
             label = 30;
             break L21;
            }
           }
           HEAP32[$93>>2] = 0;
           $142 = ((($this)) + 4|0);
           $143 = HEAP32[$142>>2]|0;
           $144 = ((($143)) + 20|0);
           HEAP32[$144>>2] = $123;
           $145 = HEAP32[$142>>2]|0;
           $146 = ((($145)) + 24|0);
           HEAP8[$146>>0] = 1;
           $147 = HEAP32[$142>>2]|0;
           $148 = ((($147)) + 25|0);
           HEAP8[$148>>0] = 1;
           $149 = HEAP32[$142>>2]|0;
           __THREW__ = 0;
           $150 = (invoke_ii(894,($149|0))|0);
           $151 = __THREW__; __THREW__ = 0;
           $152 = $151&1;
           if ($152) {
            label = 30;
            break L21;
           }
           $153 = HEAP32[$150>>2]|0;
           $154 = ((($153)) + 16|0);
           $155 = HEAP32[$154>>2]|0;
           __THREW__ = 0;
           invoke_vii($155|0,($150|0),1048);
           $156 = __THREW__; __THREW__ = 0;
           $157 = $156&1;
           if ($157) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(904,($74|0));
           $158 = __THREW__; __THREW__ = 0;
           $159 = $158&1;
           if ($159) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(904,($77|0));
           $160 = __THREW__; __THREW__ = 0;
           $161 = $160&1;
           if ($161) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(904,($80|0));
           $162 = __THREW__; __THREW__ = 0;
           $163 = $162&1;
           if ($163) {
            label = 30;
            break L21;
           }
           __THREW__ = 0;
           invoke_vi(904,($83|0));
           $164 = __THREW__; __THREW__ = 0;
           $165 = $164&1;
           if ($165) {
            label = 30;
            break L21;
           }
          }
         } while(0);
         __ZN9ExitGames6Common6ObjectD2Ev($val);
         STACKTOP = sp;return;
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 30) {
     $70 = ___cxa_find_matching_catch_2()|0;
     $71 = tempRet0;
     $$0 = $70;$$01 = $71;
    }
    __ZN9ExitGames6Common6ObjectD2Ev($val);
    $$1 = $$0;$$12 = $$01;
    ___resumeException($$1|0);
    // unreachable;
   }
  } while(0);
  if ((label|0) == 13) {
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   $eh$lpad$body23$index2Z2D = $30;$eh$lpad$body23$indexZ2D = $29;
  }
  __ZN9ExitGames6Common6ObjectD2Ev($val);
  $eh$lpad$body$index7Z2D = $eh$lpad$body23$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body23$indexZ2D;
 }
 __ZN9ExitGames6Common6ObjectD2Ev($1);
 $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index7Z2D;
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase18readPingOpResponseEPKNS0_17OperationResponseE($this,$0) {
 $this = $this|0;
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase12initCallbackEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,146589);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($18|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($2|0),4,(6560|0),($3|0),1,803,(6664|0),($vararg_buffer|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $18 = ___cxa_find_matching_catch_2()|0;
  $19 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($18|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $8 = ((($this)) + 97|0);
  HEAP8[$8>>0] = 1;
  $9 = HEAP32[$this>>2]|0;
  $10 = ((($9)) + 52|0);
  $11 = HEAP32[$10>>2]|0;
  FUNCTION_TABLE_vi[$11 & 2047]($this);
  $12 = ((($this)) + 4|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($13)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($15)) + 16|0);
  $17 = HEAP32[$16>>2]|0;
  FUNCTION_TABLE_vii[$17 & 2047]($14,1024);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Photon8Internal13SocketConnectC2ERNS1_21PhotonConnectListenerE($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal13PhotonConnectC2ERNS1_21PhotonConnectListenerE($this,$listener);
 HEAP32[$this>>2] = (6236);
 $0 = ((($this)) + 80|0);
 __THREW__ = 0;
 invoke_vi(913,($0|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___resumeException($5|0);
  // unreachable;
 }
 $3 = HEAP16[83684]|0;
 $4 = ($3<<16>>16)==(-1);
 if ($4) {
  return;
 }
 $7 = (($3) + 1)<<16>>16;
 HEAP16[83684] = $7;
 $8 = ((($this)) + 16|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($this)) + 20|0);
 HEAP32[$9>>2] = -1;
 $10 = ((($this)) + 24|0);
 $11 = ((($this)) + 76|0);
 HEAP8[$11>>0] = 0;
 $12 = ((($this)) + 92|0);
 HEAP16[$12>>1] = 0;
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;HEAP32[$10+12>>2]=0|0;HEAP32[$10+16>>2]=0|0;HEAP32[$10+20>>2]=0|0;
 return;
}
function __ZN9ExitGames6Photon8Internal13PhotonConnectC2ERNS1_21PhotonConnectListenerE($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (6516);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $listener;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 12|0);
 HEAP8[$2>>0] = 1;
 return;
}
function __ZN9ExitGames6Photon8Internal13PhotonConnectD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$this>>2]|0;
 return ($0|0);
}
function __ZNK9ExitGames6Photon17OperationResponse13getReturnCodeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 2|0);
 $1 = HEAP16[$0>>1]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon17OperationResponse19getParameterForCodeEh($agg$result,$this,$parameterCode) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $parameterCode = $parameterCode|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP8[$0>>0] = $parameterCode;
 $1 = ((($this)) + 20|0);
 $2 = (__ZNK9ExitGames6Common14DictionaryBase8getValueIhEEPKNS0_6ObjectERKT_S5_($1,$0,0)|0);
 $3 = ($2|0)==(0|0);
 if ($3) {
  __ZN9ExitGames6Common6ObjectC2Ev($agg$result);
  STACKTOP = sp;return;
 } else {
  __ZN9ExitGames6Common6ObjectC2ERKS1_($agg$result,$2);
  STACKTOP = sp;return;
 }
}
function _EGBN_new() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc(20)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($0)) + 16|0);
 HEAP32[$2>>2] = 1;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;HEAP32[$0+12>>2]=0|0;
 $$0 = $0;
 return ($$0|0);
}
function _EGBN_CTX_new() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc(704)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 _memset(($0|0),0,704)|0;
 $2 = ((($0)) + 644|0);
 HEAP32[$2>>2] = 1;
 $$0 = $0;
 return ($$0|0);
}
function _EGBN_bin2bn($s,$len,$ret) {
 $s = $s|0;
 $len = $len|0;
 $ret = $ret|0;
 var $$01 = 0, $$02 = 0, $$02$ph = 0, $$05 = 0, $$lcssa = 0, $$lcssa20 = 0, $$lcssa21 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ftl$09 = 0, $i$0$ph = 0, $l$0 = 0, $m$0 = 0, $m$0$ph = 0, $n$0 = 0, $n$0$ph = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ret|0)==(0|0);
 do {
  if ($0) {
   $1 = (_malloc(20)|0);
   $2 = ($1|0)==(0|0);
   if ($2) {
    $$01 = 0;
    return ($$01|0);
   } else {
    $3 = ((($1)) + 16|0);
    HEAP32[$3>>2] = 1;
    ;HEAP32[$1>>2]=0|0;HEAP32[$1+4>>2]=0|0;HEAP32[$1+8>>2]=0|0;HEAP32[$1+12>>2]=0|0;
    $$05 = $1;
    break;
   }
  } else {
   $$05 = $ret;
  }
 } while(0);
 $4 = ($len|0)==(0);
 if ($4) {
  $5 = ((($$05)) + 4|0);
  HEAP32[$5>>2] = 0;
  $$01 = $$05;
  return ($$01|0);
 }
 $6 = $len << 3;
 $7 = (($6) + 47)|0;
 $8 = (($7|0) / 32)&-1;
 $9 = ((($$05)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8|0)>($10|0);
 if ($11) {
  $12 = (($6) + 16)|0;
  $13 = (($12|0) / 32)&-1;
  $14 = (($13) + 1)|0;
  $15 = ($10|0)<($14|0);
  if ($15) {
   $16 = (_bn_expand_internal($$05,$14)|0);
   $17 = ($16|0)==(0|0);
   if ($17) {
    $$01 = 0;
    return ($$01|0);
   }
   $18 = HEAP32[$$05>>2]|0;
   $19 = ($18|0)==(0|0);
   if (!($19)) {
    _free($18);
   }
   HEAP32[$$05>>2] = $16;
   HEAP32[$9>>2] = $14;
  }
 }
 $20 = (($len) + -1)|0;
 $21 = $20 >>> 2;
 $22 = (($21) + 1)|0;
 $23 = $20 & 3;
 $24 = ((($$05)) + 4|0);
 HEAP32[$24>>2] = $22;
 $25 = ((($$05)) + 12|0);
 HEAP32[$25>>2] = 0;
 $$02$ph = $s;$i$0$ph = $22;$m$0$ph = $23;$n$0$ph = $len;
 L21: while(1) {
  $$02 = $$02$ph;$l$0 = 0;$m$0 = $m$0$ph;$n$0 = $n$0$ph;
  while(1) {
   $26 = (($n$0) + -1)|0;
   $27 = ($n$0|0)==(0);
   if ($27) {
    break L21;
   }
   $28 = $l$0 << 8;
   $29 = ((($$02)) + 1|0);
   $30 = HEAP8[$$02>>0]|0;
   $31 = $30&255;
   $32 = $31 | $28;
   $33 = (($m$0) + -1)|0;
   $34 = ($m$0|0)==(0);
   if ($34) {
    $$lcssa = $29;$$lcssa20 = $26;$$lcssa21 = $32;
    break;
   } else {
    $$02 = $29;$l$0 = $32;$m$0 = $33;$n$0 = $26;
   }
  }
  $35 = (($i$0$ph) + -1)|0;
  $36 = HEAP32[$$05>>2]|0;
  $37 = (($36) + ($35<<2)|0);
  HEAP32[$37>>2] = $$lcssa21;
  $$02$ph = $$lcssa;$i$0$ph = $35;$m$0$ph = 3;$n$0$ph = $$lcssa20;
 }
 $38 = HEAP32[$$05>>2]|0;
 $39 = (($38) + ($21<<2)|0);
 $44 = $22;$ftl$09 = $39;
 while(1) {
  $40 = HEAP32[$ftl$09>>2]|0;
  $41 = ($40|0)==(0);
  if (!($41)) {
   $$01 = $$05;
   label = 20;
   break;
  }
  $42 = ((($ftl$09)) + -4|0);
  $43 = (($44) + -1)|0;
  HEAP32[$24>>2] = $43;
  $45 = ($44|0)>(1);
  if ($45) {
   $44 = $43;$ftl$09 = $42;
  } else {
   $$01 = $$05;
   label = 20;
   break;
  }
 }
 if ((label|0) == 20) {
  return ($$01|0);
 }
 return (0)|0;
}
function _EGBN_mod_exp($r,$a,$p,$m,$ctx) {
 $r = $r|0;
 $a = $a|0;
 $p = $p|0;
 $m = $m|0;
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($m)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if ($2) {
  $3 = HEAP32[$m>>2]|0;
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 & 1;
  $6 = ($5|0)==(0);
  if (!($6)) {
   $7 = ((($a)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)==(1);
   if ($9) {
    $10 = ((($a)) + 12|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==(0);
    if ($12) {
     $13 = HEAP32[$a>>2]|0;
     $14 = HEAP32[$13>>2]|0;
     $15 = (_EGBN_mod_exp_mont_word($r,$14,$p,$m,$ctx,0)|0);
     $ret$0 = $15;
     return ($ret$0|0);
    }
   }
   $16 = (_EGBN_mod_exp_mont($r,$a,$p,$m,$ctx,0)|0);
   $ret$0 = $16;
   return ($ret$0|0);
  }
 }
 $17 = (_EGBN_mod_exp_recp($r,$a,$p,$m,$ctx)|0);
 $ret$0 = $17;
 return ($ret$0|0);
}
function _EGBN_CTX_free($ctx) {
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ctx|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($ctx)) + 4|0);
 _EGBN_clear_free($1);
 $2 = ((($ctx)) + 24|0);
 _EGBN_clear_free($2);
 $3 = ((($ctx)) + 44|0);
 _EGBN_clear_free($3);
 $4 = ((($ctx)) + 64|0);
 _EGBN_clear_free($4);
 $5 = ((($ctx)) + 84|0);
 _EGBN_clear_free($5);
 $6 = ((($ctx)) + 104|0);
 _EGBN_clear_free($6);
 $7 = ((($ctx)) + 124|0);
 _EGBN_clear_free($7);
 $8 = ((($ctx)) + 144|0);
 _EGBN_clear_free($8);
 $9 = ((($ctx)) + 164|0);
 _EGBN_clear_free($9);
 $10 = ((($ctx)) + 184|0);
 _EGBN_clear_free($10);
 $11 = ((($ctx)) + 204|0);
 _EGBN_clear_free($11);
 $12 = ((($ctx)) + 224|0);
 _EGBN_clear_free($12);
 $13 = ((($ctx)) + 244|0);
 _EGBN_clear_free($13);
 $14 = ((($ctx)) + 264|0);
 _EGBN_clear_free($14);
 $15 = ((($ctx)) + 284|0);
 _EGBN_clear_free($15);
 $16 = ((($ctx)) + 304|0);
 _EGBN_clear_free($16);
 $17 = ((($ctx)) + 324|0);
 _EGBN_clear_free($17);
 $18 = ((($ctx)) + 344|0);
 _EGBN_clear_free($18);
 $19 = ((($ctx)) + 364|0);
 _EGBN_clear_free($19);
 $20 = ((($ctx)) + 384|0);
 _EGBN_clear_free($20);
 $21 = ((($ctx)) + 404|0);
 _EGBN_clear_free($21);
 $22 = ((($ctx)) + 424|0);
 _EGBN_clear_free($22);
 $23 = ((($ctx)) + 444|0);
 _EGBN_clear_free($23);
 $24 = ((($ctx)) + 464|0);
 _EGBN_clear_free($24);
 $25 = ((($ctx)) + 484|0);
 _EGBN_clear_free($25);
 $26 = ((($ctx)) + 504|0);
 _EGBN_clear_free($26);
 $27 = ((($ctx)) + 524|0);
 _EGBN_clear_free($27);
 $28 = ((($ctx)) + 544|0);
 _EGBN_clear_free($28);
 $29 = ((($ctx)) + 564|0);
 _EGBN_clear_free($29);
 $30 = ((($ctx)) + 584|0);
 _EGBN_clear_free($30);
 $31 = ((($ctx)) + 604|0);
 _EGBN_clear_free($31);
 $32 = ((($ctx)) + 624|0);
 _EGBN_clear_free($32);
 $33 = ((($ctx)) + 644|0);
 $34 = HEAP32[$33>>2]|0;
 $35 = $34 & 1;
 $36 = ($35|0)==(0);
 if ($36) {
  return;
 }
 _free($ctx);
 return;
}
function _EGBN_num_bits($a) {
 $a = $a|0;
 var $$0 = 0, $$0$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = (($1) + -1)|0;
 $4 = HEAP32[$a>>2]|0;
 $5 = (($4) + ($3<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3 << 5;
 $8 = ($6>>>0)>(65535);
 do {
  if ($8) {
   $9 = ($6>>>0)>(16777215);
   if ($9) {
    $10 = $6 >>> 24;
    $11 = (146866 + ($10)|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = $12 << 24 >> 24;
    $14 = (($13) + 24)|0;
    $$0$i = $14;
    break;
   } else {
    $15 = $6 >>> 16;
    $16 = (146866 + ($15)|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = $17 << 24 >> 24;
    $19 = (($18) + 16)|0;
    $$0$i = $19;
    break;
   }
  } else {
   $20 = $6 & 65280;
   $21 = ($20|0)==(0);
   if ($21) {
    $27 = (146866 + ($6)|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28 << 24 >> 24;
    $$0$i = $29;
    break;
   } else {
    $22 = $6 >>> 8;
    $23 = (146866 + ($22)|0);
    $24 = HEAP8[$23>>0]|0;
    $25 = $24 << 24 >> 24;
    $26 = (($25) + 8)|0;
    $$0$i = $26;
    break;
   }
  }
 } while(0);
 $30 = (($$0$i) + ($7))|0;
 $$0 = $30;
 return ($$0|0);
}
function _EGBN_bn2bin($a,$to) {
 $a = $a|0;
 $to = $to|0;
 var $$0$i$i = 0, $$0$i3 = 0, $$02 = 0, $$in = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, $phitmp1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0$i3 = 0;
  return ($$0$i3|0);
 }
 $3 = (($1) + -1)|0;
 $4 = HEAP32[$a>>2]|0;
 $5 = (($4) + ($3<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3 << 5;
 $8 = ($6>>>0)>(65535);
 do {
  if ($8) {
   $9 = ($6>>>0)>(16777215);
   if ($9) {
    $10 = $6 >>> 24;
    $11 = (146866 + ($10)|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = $12 << 24 >> 24;
    $14 = (($13) + 24)|0;
    $$0$i$i = $14;
    break;
   } else {
    $15 = $6 >>> 16;
    $16 = (146866 + ($15)|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = $17 << 24 >> 24;
    $19 = (($18) + 16)|0;
    $$0$i$i = $19;
    break;
   }
  } else {
   $20 = $6 & 65280;
   $21 = ($20|0)==(0);
   if ($21) {
    $27 = (146866 + ($6)|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28 << 24 >> 24;
    $$0$i$i = $29;
    break;
   } else {
    $22 = $6 >>> 8;
    $23 = (146866 + ($22)|0);
    $24 = HEAP8[$23>>0]|0;
    $25 = $24 << 24 >> 24;
    $26 = (($25) + 8)|0;
    $$0$i$i = $26;
    break;
   }
  }
 } while(0);
 $30 = $7 | 7;
 $phitmp = (($30) + ($$0$i$i))|0;
 $phitmp1 = (($phitmp|0) / 8)&-1;
 $31 = ($phitmp|0)>(7);
 if (!($31)) {
  $$0$i3 = $phitmp1;
  return ($$0$i3|0);
 }
 $$02 = $to;$$in = $phitmp1;
 while(1) {
  $32 = (($$in) + -1)|0;
  $33 = (($32|0) / 4)&-1;
  $34 = HEAP32[$a>>2]|0;
  $35 = (($34) + ($33<<2)|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = (($32|0) % 4)&-1;
  $38 = $37 << 3;
  $39 = $36 >>> $38;
  $40 = $39&255;
  $41 = ((($$02)) + 1|0);
  HEAP8[$$02>>0] = $40;
  $42 = ($$in|0)>(1);
  if ($42) {
   $$02 = $41;$$in = $32;
  } else {
   $$0$i3 = $phitmp1;
   break;
  }
 }
 return ($$0$i3|0);
}
function __ZN9ExitGames6Photon8Internal10Encryption19calculateSHA256HashEPKhiPh($inputData,$length,$hashedChars) {
 $inputData = $inputData|0;
 $length = $length|0;
 $hashedChars = $hashedChars|0;
 var $ctx256 = 0, $temp = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $temp = sp + 104|0;
 $ctx256 = sp;
 _EGSHA256_Init($ctx256);
 _EGSHA256_Update($ctx256,$inputData,$length);
 (_EGSHA256_End($ctx256,$temp)|0);
 dest=$hashedChars; src=$temp; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 STACKTOP = sp;return;
}
function _EGBN_free($a) {
 $a = $a|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$a>>2]|0;
 $2 = ($1|0)==(0|0);
 $$pre = ((($a)) + 16|0);
 if (!($2)) {
  $3 = HEAP32[$$pre>>2]|0;
  $4 = $3 & 2;
  $5 = ($4|0)==(0);
  if ($5) {
   _free($1);
  }
 }
 $6 = HEAP32[$$pre>>2]|0;
 $7 = $6 | 32768;
 HEAP32[$$pre>>2] = $7;
 $8 = $6 & 1;
 $9 = ($8|0)==(0);
 if ($9) {
  return;
 }
 _free($a);
 return;
}
function _EGSHA256_Init($context) {
 $context = $context|0;
 var $0 = 0, $1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($context|0)==(0|0);
 if ($0) {
  return;
 }
 ;HEAP32[$context>>2]=HEAP32[7392>>2]|0;HEAP32[$context+4>>2]=HEAP32[7392+4>>2]|0;HEAP32[$context+8>>2]=HEAP32[7392+8>>2]|0;HEAP32[$context+12>>2]=HEAP32[7392+12>>2]|0;HEAP32[$context+16>>2]=HEAP32[7392+16>>2]|0;HEAP32[$context+20>>2]=HEAP32[7392+20>>2]|0;HEAP32[$context+24>>2]=HEAP32[7392+24>>2]|0;HEAP32[$context+28>>2]=HEAP32[7392+28>>2]|0;
 $1 = ((($context)) + 32|0);
 dest=$1; stop=dest+72|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _EGSHA256_Update($context,$data,$len) {
 $context = $context|0;
 $data = $data|0;
 $len = $len|0;
 var $$0$lcssa = 0, $$0$ph = 0, $$02$lcssa = 0, $$02$ph = 0, $$023 = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($len|0)==(0);
 if ($0) {
  return;
 }
 $1 = ($context|0)!=(0|0);
 $2 = ($data|0)!=(0|0);
 $or$cond = $1 & $2;
 if (!($or$cond)) {
  ___assert_fail((146799|0),(146740|0),530,(146850|0));
  // unreachable;
 }
 $3 = ((($context)) + 32|0);
 $4 = $3;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) + 4)|0;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 $10 = (_bitshift64Lshr(($6|0),($9|0),3)|0);
 $11 = tempRet0;
 $12 = $10 & 63;
 $13 = ($12|0)==(0);
 do {
  if ($13) {
   $$0$ph = $len;$$02$ph = $data;
  } else {
   $18 = (64 - ($12))|0;
   $19 = ($18>>>0)>($len>>>0);
   $20 = (((($context)) + 40|0) + ($12)|0);
   if (!($19)) {
    _memcpy(($20|0),($data|0),($18|0))|0;
    $21 = $18 << 3;
    $22 = (_i64Add(($6|0),($9|0),($21|0),0)|0);
    $23 = tempRet0;
    $24 = $3;
    $25 = $24;
    HEAP32[$25>>2] = $22;
    $26 = (($24) + 4)|0;
    $27 = $26;
    HEAP32[$27>>2] = $23;
    $28 = (($len) - ($18))|0;
    $29 = (($data) + ($18)|0);
    $30 = ((($context)) + 40|0);
    _SHA256_Transform($context,$30);
    $$0$ph = $28;$$02$ph = $29;
    break;
   }
   _memcpy(($20|0),($data|0),($len|0))|0;
   $31 = $len << 3;
   $32 = (_i64Add(($6|0),($9|0),($31|0),0)|0);
   $33 = tempRet0;
   $34 = $3;
   $35 = $34;
   HEAP32[$35>>2] = $32;
   $36 = (($34) + 4)|0;
   $37 = $36;
   HEAP32[$37>>2] = $33;
   return;
  }
 } while(0);
 $14 = ($$0$ph>>>0)>(63);
 if ($14) {
  $15 = (($$0$ph) + -64)|0;
  $16 = $15 & -64;
  $17 = (($16) + 64)|0;
  $scevgep = (($$02$ph) + ($17)|0);
  $$023 = $$02$ph;$$04 = $$0$ph;
  while(1) {
   _SHA256_Transform($context,$$023);
   $38 = $3;
   $39 = $38;
   $40 = HEAP32[$39>>2]|0;
   $41 = (($38) + 4)|0;
   $42 = $41;
   $43 = HEAP32[$42>>2]|0;
   $44 = (_i64Add(($40|0),($43|0),512,0)|0);
   $45 = tempRet0;
   $46 = $3;
   $47 = $46;
   HEAP32[$47>>2] = $44;
   $48 = (($46) + 4)|0;
   $49 = $48;
   HEAP32[$49>>2] = $45;
   $50 = (($$04) + -64)|0;
   $51 = ((($$023)) + 64|0);
   $52 = ($50>>>0)>(63);
   if ($52) {
    $$023 = $51;$$04 = $50;
   } else {
    break;
   }
  }
  $53 = (($15) - ($16))|0;
  $$0$lcssa = $53;$$02$lcssa = $scevgep;
 } else {
  $$0$lcssa = $$0$ph;$$02$lcssa = $$02$ph;
 }
 $54 = ($$0$lcssa|0)==(0);
 if ($54) {
  return;
 }
 $55 = ((($context)) + 40|0);
 _memcpy(($55|0),($$02$lcssa|0),($$0$lcssa|0))|0;
 $56 = $$0$lcssa << 3;
 $57 = $3;
 $58 = $57;
 $59 = HEAP32[$58>>2]|0;
 $60 = (($57) + 4)|0;
 $61 = $60;
 $62 = HEAP32[$61>>2]|0;
 $63 = (_i64Add(($59|0),($62|0),($56|0),0)|0);
 $64 = tempRet0;
 $65 = $3;
 $66 = $65;
 HEAP32[$66>>2] = $63;
 $67 = (($65) + 4)|0;
 $68 = $67;
 HEAP32[$68>>2] = $64;
 return;
}
function _EGSHA256_End($context,$buffer) {
 $context = $context|0;
 $buffer = $buffer|0;
 var $$1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $digest = 0, $scevgep = 0, dest = 0;
 var label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $digest = sp;
 $0 = ($context|0)==(0|0);
 if ($0) {
  ___assert_fail((146714|0),(146740|0),637,(146771|0));
  // unreachable;
 }
 $1 = ($buffer|0)==(0|0);
 if ($1) {
  dest=$context; stop=dest+104|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  $$1 = 0;
  STACKTOP = sp;return ($$1|0);
 } else {
  _EGSHA256_Final($digest,$context);
  $2 = HEAP8[$digest>>0]|0;
  $3 = ((($buffer)) + 1|0);
  HEAP8[$buffer>>0] = $2;
  $4 = ((($digest)) + 1|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = ((($buffer)) + 2|0);
  HEAP8[$3>>0] = $5;
  $7 = ((($digest)) + 2|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = ((($buffer)) + 3|0);
  HEAP8[$6>>0] = $8;
  $10 = ((($digest)) + 3|0);
  $11 = HEAP8[$10>>0]|0;
  $12 = ((($buffer)) + 4|0);
  HEAP8[$9>>0] = $11;
  $13 = ((($digest)) + 4|0);
  $14 = HEAP8[$13>>0]|0;
  $15 = ((($buffer)) + 5|0);
  HEAP8[$12>>0] = $14;
  $16 = ((($digest)) + 5|0);
  $17 = HEAP8[$16>>0]|0;
  $18 = ((($buffer)) + 6|0);
  HEAP8[$15>>0] = $17;
  $19 = ((($digest)) + 6|0);
  $20 = HEAP8[$19>>0]|0;
  $21 = ((($buffer)) + 7|0);
  HEAP8[$18>>0] = $20;
  $22 = ((($digest)) + 7|0);
  $23 = HEAP8[$22>>0]|0;
  $24 = ((($buffer)) + 8|0);
  HEAP8[$21>>0] = $23;
  $25 = ((($digest)) + 8|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = ((($buffer)) + 9|0);
  HEAP8[$24>>0] = $26;
  $28 = ((($digest)) + 9|0);
  $29 = HEAP8[$28>>0]|0;
  $30 = ((($buffer)) + 10|0);
  HEAP8[$27>>0] = $29;
  $31 = ((($digest)) + 10|0);
  $32 = HEAP8[$31>>0]|0;
  $33 = ((($buffer)) + 11|0);
  HEAP8[$30>>0] = $32;
  $34 = ((($digest)) + 11|0);
  $35 = HEAP8[$34>>0]|0;
  $36 = ((($buffer)) + 12|0);
  HEAP8[$33>>0] = $35;
  $37 = ((($digest)) + 12|0);
  $38 = HEAP8[$37>>0]|0;
  $39 = ((($buffer)) + 13|0);
  HEAP8[$36>>0] = $38;
  $40 = ((($digest)) + 13|0);
  $41 = HEAP8[$40>>0]|0;
  $42 = ((($buffer)) + 14|0);
  HEAP8[$39>>0] = $41;
  $43 = ((($digest)) + 14|0);
  $44 = HEAP8[$43>>0]|0;
  $45 = ((($buffer)) + 15|0);
  HEAP8[$42>>0] = $44;
  $46 = ((($digest)) + 15|0);
  $47 = HEAP8[$46>>0]|0;
  $48 = ((($buffer)) + 16|0);
  HEAP8[$45>>0] = $47;
  $49 = ((($digest)) + 16|0);
  $50 = HEAP8[$49>>0]|0;
  $51 = ((($buffer)) + 17|0);
  HEAP8[$48>>0] = $50;
  $52 = ((($digest)) + 17|0);
  $53 = HEAP8[$52>>0]|0;
  $54 = ((($buffer)) + 18|0);
  HEAP8[$51>>0] = $53;
  $55 = ((($digest)) + 18|0);
  $56 = HEAP8[$55>>0]|0;
  $57 = ((($buffer)) + 19|0);
  HEAP8[$54>>0] = $56;
  $58 = ((($digest)) + 19|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = ((($buffer)) + 20|0);
  HEAP8[$57>>0] = $59;
  $61 = ((($digest)) + 20|0);
  $62 = HEAP8[$61>>0]|0;
  $63 = ((($buffer)) + 21|0);
  HEAP8[$60>>0] = $62;
  $64 = ((($digest)) + 21|0);
  $65 = HEAP8[$64>>0]|0;
  $66 = ((($buffer)) + 22|0);
  HEAP8[$63>>0] = $65;
  $67 = ((($digest)) + 22|0);
  $68 = HEAP8[$67>>0]|0;
  $69 = ((($buffer)) + 23|0);
  HEAP8[$66>>0] = $68;
  $70 = ((($digest)) + 23|0);
  $71 = HEAP8[$70>>0]|0;
  $72 = ((($buffer)) + 24|0);
  HEAP8[$69>>0] = $71;
  $73 = ((($digest)) + 24|0);
  $74 = HEAP8[$73>>0]|0;
  $75 = ((($buffer)) + 25|0);
  HEAP8[$72>>0] = $74;
  $76 = ((($digest)) + 25|0);
  $77 = HEAP8[$76>>0]|0;
  $78 = ((($buffer)) + 26|0);
  HEAP8[$75>>0] = $77;
  $79 = ((($digest)) + 26|0);
  $80 = HEAP8[$79>>0]|0;
  $81 = ((($buffer)) + 27|0);
  HEAP8[$78>>0] = $80;
  $82 = ((($digest)) + 27|0);
  $83 = HEAP8[$82>>0]|0;
  $84 = ((($buffer)) + 28|0);
  HEAP8[$81>>0] = $83;
  $85 = ((($digest)) + 28|0);
  $86 = HEAP8[$85>>0]|0;
  $87 = ((($buffer)) + 29|0);
  HEAP8[$84>>0] = $86;
  $88 = ((($digest)) + 29|0);
  $89 = HEAP8[$88>>0]|0;
  $90 = ((($buffer)) + 30|0);
  HEAP8[$87>>0] = $89;
  $91 = ((($digest)) + 30|0);
  $92 = HEAP8[$91>>0]|0;
  $93 = ((($buffer)) + 31|0);
  HEAP8[$90>>0] = $92;
  $94 = ((($digest)) + 31|0);
  $95 = HEAP8[$94>>0]|0;
  HEAP8[$93>>0] = $95;
  $scevgep = ((($buffer)) + 32|0);
  HEAP8[$scevgep>>0] = 0;
  $$1 = $scevgep;
  STACKTOP = sp;return ($$1|0);
 }
 return (0)|0;
}
function _EGSHA256_Final($digest,$context) {
 $digest = $digest|0;
 $context = $context|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ($context|0)==(0|0);
 if ($0) {
  ___assert_fail((146714|0),(146740|0),574,(146784|0));
  // unreachable;
 }
 $1 = ($digest|0)==(0|0);
 if ($1) {
  dest=$context; stop=dest+104|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  return;
 }
 $2 = ((($context)) + 32|0);
 $3 = $2;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $6 = (($3) + 4)|0;
 $7 = $6;
 $8 = HEAP32[$7>>2]|0;
 $9 = (_bitshift64Lshr(($5|0),($8|0),3)|0);
 $10 = tempRet0;
 $11 = $9 & 63;
 $12 = (_bitshift64Lshr(($8|0),($5|0),8)|0);
 $13 = tempRet0;
 $14 = $12 & 16711935;
 $15 = $13 & 16711935;
 $16 = (_bitshift64Shl(($8|0),($5|0),8)|0);
 $17 = tempRet0;
 $18 = $16 & -16711936;
 $19 = $17 & -16711936;
 $20 = $14 | $18;
 $21 = $15 | $19;
 $22 = (_bitshift64Lshr(($20|0),($21|0),16)|0);
 $23 = tempRet0;
 $24 = $22 & 65535;
 $25 = $23 & 65535;
 $26 = (_bitshift64Shl(($20|0),($21|0),16)|0);
 $27 = tempRet0;
 $28 = $26 & -65536;
 $29 = $27 & -65536;
 $30 = $24 | $28;
 $31 = $25 | $29;
 $32 = $2;
 $33 = $32;
 HEAP32[$33>>2] = $30;
 $34 = (($32) + 4)|0;
 $35 = $34;
 HEAP32[$35>>2] = $31;
 $36 = ($11|0)==(0);
 do {
  if ($36) {
   $52 = ((($context)) + 40|0);
   dest=$52; stop=dest+56|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
   HEAP8[$52>>0] = -128;
   $56 = $30;$59 = $31;
  } else {
   $37 = (($11) + 1)|0;
   $38 = ((($context)) + 40|0);
   $39 = (((($context)) + 40|0) + ($11)|0);
   HEAP8[$39>>0] = -128;
   $40 = ($37>>>0)<(57);
   if ($40) {
    $41 = (((($context)) + 40|0) + ($37)|0);
    $42 = (55 - ($11))|0;
    _memset(($41|0),0,($42|0))|0;
    $56 = $30;$59 = $31;
    break;
   }
   $43 = ($37>>>0)<(64);
   if ($43) {
    $44 = (((($context)) + 40|0) + ($37)|0);
    $45 = $11 ^ 63;
    _memset(($44|0),0,($45|0))|0;
   }
   _SHA256_Transform($context,$38);
   dest=$38; stop=dest+56|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
   $46 = $2;
   $47 = $46;
   $48 = HEAP32[$47>>2]|0;
   $49 = (($46) + 4)|0;
   $50 = $49;
   $51 = HEAP32[$50>>2]|0;
   $56 = $48;$59 = $51;
  }
 } while(0);
 $53 = ((($context)) + 96|0);
 $54 = $53;
 $55 = $54;
 HEAP32[$55>>2] = $56;
 $57 = (($54) + 4)|0;
 $58 = $57;
 HEAP32[$58>>2] = $59;
 $60 = ((($context)) + 40|0);
 _SHA256_Transform($context,$60);
 $61 = HEAP32[$context>>2]|0;
 $62 = $61 >>> 16;
 $63 = $61 << 16;
 $64 = $62 | $63;
 $65 = $64 >>> 8;
 $66 = $65 & 16711935;
 $67 = $64 << 8;
 $68 = $67 & -16711936;
 $69 = $66 | $68;
 HEAP32[$context>>2] = $69;
 $70 = ((($digest)) + 4|0);
 HEAP32[$digest>>2] = $69;
 $71 = ((($context)) + 4|0);
 $72 = HEAP32[$71>>2]|0;
 $73 = $72 >>> 16;
 $74 = $72 << 16;
 $75 = $73 | $74;
 $76 = $75 >>> 8;
 $77 = $76 & 16711935;
 $78 = $75 << 8;
 $79 = $78 & -16711936;
 $80 = $77 | $79;
 HEAP32[$71>>2] = $80;
 $81 = ((($digest)) + 8|0);
 HEAP32[$70>>2] = $80;
 $82 = ((($context)) + 8|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = $83 >>> 16;
 $85 = $83 << 16;
 $86 = $84 | $85;
 $87 = $86 >>> 8;
 $88 = $87 & 16711935;
 $89 = $86 << 8;
 $90 = $89 & -16711936;
 $91 = $88 | $90;
 HEAP32[$82>>2] = $91;
 $92 = ((($digest)) + 12|0);
 HEAP32[$81>>2] = $91;
 $93 = ((($context)) + 12|0);
 $94 = HEAP32[$93>>2]|0;
 $95 = $94 >>> 16;
 $96 = $94 << 16;
 $97 = $95 | $96;
 $98 = $97 >>> 8;
 $99 = $98 & 16711935;
 $100 = $97 << 8;
 $101 = $100 & -16711936;
 $102 = $99 | $101;
 HEAP32[$93>>2] = $102;
 $103 = ((($digest)) + 16|0);
 HEAP32[$92>>2] = $102;
 $104 = ((($context)) + 16|0);
 $105 = HEAP32[$104>>2]|0;
 $106 = $105 >>> 16;
 $107 = $105 << 16;
 $108 = $106 | $107;
 $109 = $108 >>> 8;
 $110 = $109 & 16711935;
 $111 = $108 << 8;
 $112 = $111 & -16711936;
 $113 = $110 | $112;
 HEAP32[$104>>2] = $113;
 $114 = ((($digest)) + 20|0);
 HEAP32[$103>>2] = $113;
 $115 = ((($context)) + 20|0);
 $116 = HEAP32[$115>>2]|0;
 $117 = $116 >>> 16;
 $118 = $116 << 16;
 $119 = $117 | $118;
 $120 = $119 >>> 8;
 $121 = $120 & 16711935;
 $122 = $119 << 8;
 $123 = $122 & -16711936;
 $124 = $121 | $123;
 HEAP32[$115>>2] = $124;
 $125 = ((($digest)) + 24|0);
 HEAP32[$114>>2] = $124;
 $126 = ((($context)) + 24|0);
 $127 = HEAP32[$126>>2]|0;
 $128 = $127 >>> 16;
 $129 = $127 << 16;
 $130 = $128 | $129;
 $131 = $130 >>> 8;
 $132 = $131 & 16711935;
 $133 = $130 << 8;
 $134 = $133 & -16711936;
 $135 = $132 | $134;
 HEAP32[$126>>2] = $135;
 $136 = ((($digest)) + 28|0);
 HEAP32[$125>>2] = $135;
 $137 = ((($context)) + 28|0);
 $138 = HEAP32[$137>>2]|0;
 $139 = $138 >>> 16;
 $140 = $138 << 16;
 $141 = $139 | $140;
 $142 = $141 >>> 8;
 $143 = $142 & 16711935;
 $144 = $141 << 8;
 $145 = $144 & -16711936;
 $146 = $143 | $145;
 HEAP32[$137>>2] = $146;
 HEAP32[$136>>2] = $146;
 dest=$context; stop=dest+104|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function _SHA256_Transform($context,$data) {
 $context = $context|0;
 $data = $data|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa18 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $a$0 = 0, $a$1 = 0, $a$1$lcssa = 0, $b$0 = 0, $b$0$phi = 0, $b$1 = 0, $b$1$lcssa = 0, $b$1$phi = 0, $c$0 = 0, $c$0$phi = 0, $c$1 = 0, $c$1$lcssa = 0, $c$1$phi = 0, $d$0 = 0, $d$0$phi = 0, $d$1 = 0, $d$1$phi = 0, $e$0 = 0;
 var $e$1 = 0, $e$1$lcssa = 0, $exitcond = 0, $f$0 = 0, $f$0$phi = 0, $f$1 = 0, $f$1$lcssa = 0, $f$1$phi = 0, $g$0 = 0, $g$0$phi = 0, $g$1 = 0, $g$1$lcssa = 0, $g$1$phi = 0, $h$0 = 0, $h$0$phi = 0, $h$1 = 0, $h$1$phi = 0, $j$0 = 0, $j$1 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($context)) + 40|0);
 $1 = HEAP32[$context>>2]|0;
 $2 = ((($context)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($context)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($context)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($context)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($context)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($context)) + 24|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($context)) + 28|0);
 $15 = HEAP32[$14>>2]|0;
 $$0 = $data;$a$0 = $1;$b$0 = $3;$c$0 = $5;$d$0 = $7;$e$0 = $9;$f$0 = $11;$g$0 = $13;$h$0 = $15;$j$0 = 0;
 while(1) {
  $16 = ((($$0)) + 4|0);
  $17 = HEAP32[$$0>>2]|0;
  $18 = $17 >>> 16;
  $19 = $17 << 16;
  $20 = $18 | $19;
  $21 = $20 >>> 8;
  $22 = $21 & 16711935;
  $23 = $20 << 8;
  $24 = $23 & -16711936;
  $25 = $22 | $24;
  $26 = (($0) + ($j$0<<2)|0);
  HEAP32[$26>>2] = $25;
  $27 = $e$0 >>> 6;
  $28 = $e$0 << 26;
  $29 = $27 | $28;
  $30 = $e$0 >>> 11;
  $31 = $e$0 << 21;
  $32 = $30 | $31;
  $33 = $29 ^ $32;
  $34 = $e$0 >>> 25;
  $35 = $e$0 << 7;
  $36 = $34 | $35;
  $37 = $33 ^ $36;
  $38 = (($37) + ($h$0))|0;
  $39 = $f$0 & $e$0;
  $40 = $e$0 ^ -1;
  $41 = $g$0 & $40;
  $42 = $41 ^ $39;
  $43 = (($38) + ($42))|0;
  $44 = (7136 + ($j$0<<2)|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($43) + ($45))|0;
  $47 = (($46) + ($25))|0;
  $48 = $a$0 >>> 2;
  $49 = $a$0 << 30;
  $50 = $48 | $49;
  $51 = $a$0 >>> 13;
  $52 = $a$0 << 19;
  $53 = $51 | $52;
  $54 = $50 ^ $53;
  $55 = $a$0 >>> 22;
  $56 = $a$0 << 10;
  $57 = $55 | $56;
  $58 = $54 ^ $57;
  $59 = $b$0 ^ $c$0;
  $60 = $a$0 & $59;
  $61 = $b$0 & $c$0;
  $62 = $60 ^ $61;
  $63 = (($58) + ($62))|0;
  $64 = (($47) + ($d$0))|0;
  $65 = (($63) + ($47))|0;
  $66 = (($j$0) + 1)|0;
  $exitcond = ($66|0)==(16);
  if ($exitcond) {
   $a$1 = $65;$b$1 = $a$0;$c$1 = $b$0;$d$1 = $c$0;$e$1 = $64;$f$1 = $e$0;$g$1 = $f$0;$h$1 = $g$0;$j$1 = 16;
   break;
  } else {
   $h$0$phi = $g$0;$g$0$phi = $f$0;$f$0$phi = $e$0;$d$0$phi = $c$0;$c$0$phi = $b$0;$b$0$phi = $a$0;$$0 = $16;$a$0 = $65;$e$0 = $64;$j$0 = $66;$h$0 = $h$0$phi;$g$0 = $g$0$phi;$f$0 = $f$0$phi;$d$0 = $d$0$phi;$c$0 = $c$0$phi;$b$0 = $b$0$phi;
  }
 }
 while(1) {
  $67 = (($j$1) + 1)|0;
  $68 = $67 & 15;
  $69 = (($0) + ($68<<2)|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = $70 >>> 7;
  $72 = $70 << 25;
  $73 = $71 | $72;
  $74 = $70 >>> 18;
  $75 = $70 << 14;
  $76 = $74 | $75;
  $77 = $70 >>> 3;
  $78 = $76 ^ $77;
  $79 = $78 ^ $73;
  $80 = (($j$1) + 14)|0;
  $81 = $80 & 15;
  $82 = (($0) + ($81<<2)|0);
  $83 = HEAP32[$82>>2]|0;
  $84 = $83 >>> 17;
  $85 = $83 << 15;
  $86 = $84 | $85;
  $87 = $83 >>> 19;
  $88 = $83 << 13;
  $89 = $87 | $88;
  $90 = $83 >>> 10;
  $91 = $89 ^ $90;
  $92 = $91 ^ $86;
  $93 = $e$1 >>> 6;
  $94 = $e$1 << 26;
  $95 = $93 | $94;
  $96 = $e$1 >>> 11;
  $97 = $e$1 << 21;
  $98 = $96 | $97;
  $99 = $95 ^ $98;
  $100 = $e$1 >>> 25;
  $101 = $e$1 << 7;
  $102 = $100 | $101;
  $103 = $99 ^ $102;
  $104 = (($103) + ($h$1))|0;
  $105 = $f$1 & $e$1;
  $106 = $e$1 ^ -1;
  $107 = $g$1 & $106;
  $108 = $107 ^ $105;
  $109 = (($104) + ($108))|0;
  $110 = (7136 + ($j$1<<2)|0);
  $111 = HEAP32[$110>>2]|0;
  $112 = (($109) + ($111))|0;
  $113 = (($j$1) + 9)|0;
  $114 = $113 & 15;
  $115 = (($0) + ($114<<2)|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = $j$1 & 15;
  $118 = (($0) + ($117<<2)|0);
  $119 = HEAP32[$118>>2]|0;
  $120 = (($79) + ($116))|0;
  $121 = (($120) + ($119))|0;
  $122 = (($121) + ($92))|0;
  HEAP32[$118>>2] = $122;
  $123 = (($112) + ($122))|0;
  $124 = $a$1 >>> 2;
  $125 = $a$1 << 30;
  $126 = $124 | $125;
  $127 = $a$1 >>> 13;
  $128 = $a$1 << 19;
  $129 = $127 | $128;
  $130 = $126 ^ $129;
  $131 = $a$1 >>> 22;
  $132 = $a$1 << 10;
  $133 = $131 | $132;
  $134 = $130 ^ $133;
  $135 = $b$1 ^ $c$1;
  $136 = $a$1 & $135;
  $137 = $b$1 & $c$1;
  $138 = $136 ^ $137;
  $139 = (($134) + ($138))|0;
  $140 = (($123) + ($d$1))|0;
  $141 = (($139) + ($123))|0;
  $142 = ($67|0)<(64);
  if ($142) {
   $h$1$phi = $g$1;$g$1$phi = $f$1;$f$1$phi = $e$1;$d$1$phi = $c$1;$c$1$phi = $b$1;$b$1$phi = $a$1;$a$1 = $141;$e$1 = $140;$j$1 = $67;$h$1 = $h$1$phi;$g$1 = $g$1$phi;$f$1 = $f$1$phi;$d$1 = $d$1$phi;$c$1 = $c$1$phi;$b$1 = $b$1$phi;
  } else {
   $$lcssa = $140;$$lcssa18 = $141;$a$1$lcssa = $a$1;$b$1$lcssa = $b$1;$c$1$lcssa = $c$1;$e$1$lcssa = $e$1;$f$1$lcssa = $f$1;$g$1$lcssa = $g$1;
   break;
  }
 }
 $143 = HEAP32[$context>>2]|0;
 $144 = (($143) + ($$lcssa18))|0;
 HEAP32[$context>>2] = $144;
 $145 = HEAP32[$2>>2]|0;
 $146 = (($145) + ($a$1$lcssa))|0;
 HEAP32[$2>>2] = $146;
 $147 = HEAP32[$4>>2]|0;
 $148 = (($147) + ($b$1$lcssa))|0;
 HEAP32[$4>>2] = $148;
 $149 = HEAP32[$6>>2]|0;
 $150 = (($149) + ($c$1$lcssa))|0;
 HEAP32[$6>>2] = $150;
 $151 = HEAP32[$8>>2]|0;
 $152 = (($151) + ($$lcssa))|0;
 HEAP32[$8>>2] = $152;
 $153 = HEAP32[$10>>2]|0;
 $154 = (($153) + ($e$1$lcssa))|0;
 HEAP32[$10>>2] = $154;
 $155 = HEAP32[$12>>2]|0;
 $156 = (($155) + ($f$1$lcssa))|0;
 HEAP32[$12>>2] = $156;
 $157 = HEAP32[$14>>2]|0;
 $158 = (($157) + ($g$1$lcssa))|0;
 HEAP32[$14>>2] = $158;
 return;
}
function _EGBN_clear_free($a) {
 $a = $a|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = HEAP32[$a>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$pre = ((($a)) + 16|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $3 = ((($a)) + 8|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = $4 << 2;
  _memset(($1|0),0,($5|0))|0;
  $6 = ((($a)) + 16|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 & 2;
  $9 = ($8|0)==(0);
  if ($9) {
   $10 = HEAP32[$a>>2]|0;
   _free($10);
   $$pre$phiZ2D = $6;
  } else {
   $$pre$phiZ2D = $6;
  }
 }
 $11 = HEAP32[$$pre$phiZ2D>>2]|0;
 $12 = $11 & 1;
 ;HEAP32[$a>>2]=0|0;HEAP32[$a+4>>2]=0|0;HEAP32[$a+8>>2]=0|0;HEAP32[$a+12>>2]=0|0;HEAP32[$a+16>>2]=0|0;
 $13 = ($12|0)==(0);
 if ($13) {
  return;
 }
 _free($a);
 return;
}
function _EGBN_mod_exp_mont_word($rr,$a,$p,$m,$ctx,$in_mont) {
 $rr = $rr|0;
 $a = $a|0;
 $p = $p|0;
 $m = $m|0;
 $ctx = $ctx|0;
 $in_mont = $in_mont|0;
 var $$0 = 0, $$06 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $8 = 0, $9 = 0, $b$026 = 0, $mont$0 = 0, $mont$1 = 0, $next_w$010 = 0;
 var $next_w$09 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $r$0$lcssa = 0, $r$029 = 0, $r$211 = 0, $r$212 = 0, $r$5 = 0, $r$617 = 0, $r_is_one$0$lcssa = 0, $r_is_one$027 = 0, $r_is_one$28 = 0, $r_is_one$5 = 0, $ret$0 = 0, $ret$021 = 0, $t$0$lcssa = 0, $t$030 = 0, $t$213 = 0, $t$214 = 0;
 var $t$5 = 0, $w$0$lcssa = 0, $w$028 = 0, $w$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($m)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = 0;
  return ($$0|0);
 }
 $3 = HEAP32[$m>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 1;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 0;
  return ($$0|0);
 }
 $7 = ($1|0)==(1);
 if ($7) {
  $8 = (($a>>>0) % ($4>>>0))&-1;
  $$06 = $8;
 } else {
  $$06 = $a;
 }
 $9 = (_EGBN_num_bits($p)|0);
 $10 = ($9|0)==(0);
 if ($10) {
  $11 = (_EGBN_set_word($rr,1)|0);
  $$0 = $11;
  return ($$0|0);
 }
 $12 = ($$06|0)==(0);
 if ($12) {
  $13 = (_EGBN_set_word($rr,0)|0);
  $$0 = $13;
  return ($$0|0);
 }
 _EGBN_CTX_start($ctx);
 $14 = (_EGBN_CTX_get($ctx)|0);
 $15 = (_EGBN_CTX_get($ctx)|0);
 $16 = (_EGBN_CTX_get($ctx)|0);
 $17 = ($14|0)==(0|0);
 $18 = ($15|0)==(0|0);
 $or$cond = $17 | $18;
 $19 = ($16|0)==(0|0);
 $or$cond3 = $or$cond | $19;
 do {
  if ($or$cond3) {
   $ret$021 = 0;
  } else {
   $20 = ($in_mont|0)==(0|0);
   if ($20) {
    $21 = (_EGBN_MONT_CTX_new()|0);
    $22 = ($21|0)==(0|0);
    if ($22) {
     $ret$021 = 0;
     break;
    }
    $23 = (_EGBN_MONT_CTX_set($21,$m,$ctx)|0);
    $24 = ($23|0)==(0);
    if ($24) {
     $mont$1 = $21;$ret$0 = 0;
    } else {
     $mont$0 = $21;
     label = 13;
    }
   } else {
    $mont$0 = $in_mont;
    label = 13;
   }
   L23: do {
    if ((label|0) == 13) {
     $25 = (($9) + -2)|0;
     $26 = ($25|0)>(-1);
     if ($26) {
      $27 = ((($mont$0)) + 4|0);
      $b$026 = $25;$r$029 = $15;$r_is_one$027 = 1;$t$030 = $16;$w$028 = $$06;
      while(1) {
       $28 = Math_imul($w$028, $w$028)|0;
       $29 = (($28>>>0) / ($w$028>>>0))&-1;
       $30 = ($29|0)==($w$028|0);
       $31 = ($r_is_one$027|0)==(0);
       do {
        if ($30) {
         if ($31) {
          $next_w$09 = $28;$r$211 = $r$029;$t$213 = $t$030;
          label = 22;
         } else {
          $73 = 1;$next_w$010 = $28;$r$212 = $r$029;$r_is_one$28 = $r_is_one$027;$t$214 = $t$030;
         }
        } else {
         if ($31) {
          $36 = (_EGBN_mul_word($r$029,$w$028)|0);
          $37 = ($36|0)==(0);
          if ($37) {
           $mont$1 = $mont$0;$ret$0 = 0;
           break L23;
          }
          $38 = (_EGBN_div(0,$t$030,$r$029,$m,$ctx)|0);
          $39 = ($38|0)==(0);
          if ($39) {
           $mont$1 = $mont$0;$ret$0 = 0;
           break L23;
          } else {
           $next_w$09 = 1;$r$211 = $t$030;$t$213 = $r$029;
           label = 22;
           break;
          }
         } else {
          $32 = (_EGBN_set_word($r$029,$w$028)|0);
          $33 = ($32|0)==(0);
          if ($33) {
           $mont$1 = $mont$0;$ret$0 = 0;
           break L23;
          }
          $34 = (_EGBN_mod_mul_montgomery($r$029,$r$029,$27,$mont$0,$ctx)|0);
          $35 = ($34|0)==(0);
          if ($35) {
           $mont$1 = $mont$0;$ret$0 = 0;
           break L23;
          } else {
           $next_w$09 = 1;$r$211 = $r$029;$t$213 = $t$030;
           label = 22;
           break;
          }
         }
        }
       } while(0);
       if ((label|0) == 22) {
        label = 0;
        $40 = (_EGBN_mod_mul_montgomery($r$211,$r$211,$r$211,$mont$0,$ctx)|0);
        $41 = ($40|0)==(0);
        if ($41) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        } else {
         $73 = 0;$next_w$010 = $next_w$09;$r$212 = $r$211;$r_is_one$28 = 0;$t$214 = $t$213;
        }
       }
       $42 = (_EGBN_is_bit_set($p,$b$026)|0);
       $43 = ($42|0)==(0);
       do {
        if ($43) {
         $r$5 = $r$212;$r_is_one$5 = $r_is_one$28;$t$5 = $t$214;$w$1 = $next_w$010;
        } else {
         $44 = Math_imul($next_w$010, $$06)|0;
         $45 = (($44>>>0) / ($$06>>>0))&-1;
         $46 = ($45|0)==($next_w$010|0);
         if ($46) {
          $r$5 = $r$212;$r_is_one$5 = $r_is_one$28;$t$5 = $t$214;$w$1 = $44;
         } else {
          if ($73) {
           $47 = (_EGBN_set_word($r$212,$next_w$010)|0);
           $48 = ($47|0)==(0);
           if ($48) {
            $mont$1 = $mont$0;$ret$0 = 0;
            break L23;
           }
           $49 = (_EGBN_mod_mul_montgomery($r$212,$r$212,$27,$mont$0,$ctx)|0);
           $50 = ($49|0)==(0);
           if ($50) {
            $mont$1 = $mont$0;$ret$0 = 0;
            break L23;
           } else {
            $r$5 = $r$212;$r_is_one$5 = 0;$t$5 = $t$214;$w$1 = $$06;
            break;
           }
          } else {
           $51 = (_EGBN_mul_word($r$212,$next_w$010)|0);
           $52 = ($51|0)==(0);
           if ($52) {
            $mont$1 = $mont$0;$ret$0 = 0;
            break L23;
           }
           $53 = (_EGBN_div(0,$t$214,$r$212,$m,$ctx)|0);
           $54 = ($53|0)==(0);
           if ($54) {
            $mont$1 = $mont$0;$ret$0 = 0;
            break L23;
           } else {
            $r$5 = $t$214;$r_is_one$5 = $r_is_one$28;$t$5 = $r$212;$w$1 = $$06;
            break;
           }
          }
         }
        }
       } while(0);
       $55 = (($b$026) + -1)|0;
       $56 = ($b$026|0)>(0);
       if ($56) {
        $b$026 = $55;$r$029 = $r$5;$r_is_one$027 = $r_is_one$5;$t$030 = $t$5;$w$028 = $w$1;
       } else {
        $r$0$lcssa = $r$5;$r_is_one$0$lcssa = $r_is_one$5;$t$0$lcssa = $t$5;$w$0$lcssa = $w$1;
        break;
       }
      }
     } else {
      $r$0$lcssa = $15;$r_is_one$0$lcssa = 1;$t$0$lcssa = $16;$w$0$lcssa = $$06;
     }
     $57 = ($w$0$lcssa|0)==(1);
     $58 = ($r_is_one$0$lcssa|0)==(0);
     do {
      if ($57) {
       if ($58) {
        $r$617 = $r$0$lcssa;
        label = 39;
       } else {
        $68 = (_EGBN_set_word($rr,1)|0);
        $69 = ($68|0)==(0);
        if ($69) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        }
       }
      } else {
       if ($58) {
        $64 = (_EGBN_mul_word($r$0$lcssa,$w$0$lcssa)|0);
        $65 = ($64|0)==(0);
        if ($65) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        }
        $66 = (_EGBN_div(0,$t$0$lcssa,$r$0$lcssa,$m,$ctx)|0);
        $67 = ($66|0)==(0);
        if ($67) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        } else {
         $r$617 = $t$0$lcssa;
         label = 39;
         break;
        }
       } else {
        $59 = (_EGBN_set_word($r$0$lcssa,$w$0$lcssa)|0);
        $60 = ($59|0)==(0);
        if ($60) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        }
        $61 = ((($mont$0)) + 4|0);
        $62 = (_EGBN_mod_mul_montgomery($r$0$lcssa,$r$0$lcssa,$61,$mont$0,$ctx)|0);
        $63 = ($62|0)==(0);
        if ($63) {
         $mont$1 = $mont$0;$ret$0 = 0;
         break L23;
        } else {
         $r$617 = $r$0$lcssa;
         label = 39;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 39) {
      $70 = (_EGBN_from_montgomery($rr,$r$617,$mont$0,$ctx)|0);
      $71 = ($70|0)==(0);
      if ($71) {
       $mont$1 = $mont$0;$ret$0 = 0;
       break;
      }
     }
     $mont$1 = $mont$0;$ret$0 = 1;
    }
   } while(0);
   $72 = ($mont$1|0)!=(0|0);
   $or$cond5 = $20 & $72;
   if ($or$cond5) {
    _EGBN_MONT_CTX_free($mont$1);
    $ret$021 = $ret$0;
   } else {
    $ret$021 = $ret$0;
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = $ret$021;
 return ($$0|0);
}
function _EGBN_mod_exp_mont($rr,$a,$p,$m,$ctx,$in_mont) {
 $rr = $rr|0;
 $a = $a|0;
 $p = $p|0;
 $m = $m|0;
 $ctx = $ctx|0;
 $in_mont = $in_mont|0;
 var $$ = 0, $$0 = 0, $$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $aa$0 = 0, $exitcond = 0, $i$034 = 0;
 var $i$125 = 0, $i$232 = 0, $i$320 = 0, $mont$0 = 0, $mont$1 = 0, $not$ = 0, $or$cond = 0, $or$cond3 = 0, $or$cond53 = 0, $ret$0 = 0, $start$0$ph = 0, $ts$0 = 0, $ts$1 = 0, $val = 0, $wend$0$lcssa = 0, $wend$023 = 0, $wend$1 = 0, $wstart$0$lcssa = 0, $wstart$0$ph = 0, $wstart$021 = 0;
 var $wstart$021$us = 0, $wvalue$0$lcssa = 0, $wvalue$024 = 0, $wvalue$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 640|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(640|0);
 $val = sp;
 $0 = HEAP32[$m>>2]|0;
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 1;
 $3 = ($2|0)==(0);
 if ($3) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $4 = (_EGBN_num_bits($p)|0);
 $5 = ($4|0)==(0);
 if ($5) {
  $6 = (_EGBN_set_word($rr,1)|0);
  $$0 = $6;
  STACKTOP = sp;return ($$0|0);
 }
 _EGBN_CTX_start($ctx);
 $7 = (_EGBN_CTX_get($ctx)|0);
 $8 = (_EGBN_CTX_get($ctx)|0);
 $9 = ($7|0)==(0|0);
 $10 = ($8|0)==(0|0);
 $or$cond = $9 | $10;
 do {
  if (!($or$cond)) {
   $11 = ($in_mont|0)==(0|0);
   if ($11) {
    $12 = (_EGBN_MONT_CTX_new()|0);
    $13 = ($12|0)==(0|0);
    if ($13) {
     break;
    }
    $14 = (_EGBN_MONT_CTX_set($12,$m,$ctx)|0);
    $15 = ($14|0)==(0);
    if ($15) {
     $mont$1 = $12;$ret$0 = 0;$ts$1 = 0;
    } else {
     $mont$0 = $12;
     label = 8;
    }
   } else {
    $mont$0 = $in_mont;
    label = 8;
   }
   L13: do {
    if ((label|0) == 8) {
     _EGBN_init($val);
     $16 = ((($a)) + 12|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17|0)==(0);
     if ($18) {
      $19 = (_EGBN_ucmp($a,$m)|0);
      $20 = ($19|0)>(-1);
      if ($20) {
       label = 10;
      } else {
       $aa$0 = $a;
      }
     } else {
      label = 10;
     }
     if ((label|0) == 10) {
      $21 = (_EGBN_nnmod($val,$a,$m,$ctx)|0);
      $22 = ($21|0)==(0);
      if ($22) {
       $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = 1;
       break;
      } else {
       $aa$0 = $val;
      }
     }
     $23 = ((($aa$0)) + 4|0);
     $24 = HEAP32[$23>>2]|0;
     switch ($24|0) {
     case 0:  {
      label = 13;
      break;
     }
     case 1:  {
      $25 = HEAP32[$aa$0>>2]|0;
      $26 = HEAP32[$25>>2]|0;
      $27 = ($26|0)==(0);
      if ($27) {
       label = 13;
      }
      break;
     }
     default: {
     }
     }
     if ((label|0) == 13) {
      $28 = (_EGBN_set_word($rr,0)|0);
      $mont$1 = $mont$0;$ret$0 = $28;$ts$1 = 1;
      break;
     }
     $29 = ((($mont$0)) + 4|0);
     $30 = (_EGBN_mod_mul_montgomery($val,$aa$0,$29,$mont$0,$ctx)|0);
     $31 = ($30|0)==(0);
     if ($31) {
      $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = 1;
     } else {
      $32 = ($4|0)>(671);
      if ($32) {
       $$ph = 6;
       label = 19;
      } else {
       $33 = ($4|0)>(239);
       if ($33) {
        $$ph = 5;
        label = 19;
       } else {
        $34 = ($4|0)>(79);
        if ($34) {
         $$ph = 4;
         label = 19;
        } else {
         $35 = ($4|0)>(23);
         if ($35) {
          $$ph = 3;
          label = 19;
         } else {
          $53 = 1;$ts$0 = 1;
         }
        }
       }
      }
      if ((label|0) == 19) {
       $36 = (_EGBN_mod_mul_montgomery($7,$val,$val,$mont$0,$ctx)|0);
       $37 = ($36|0)==(0);
       if ($37) {
        $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = 1;
        break;
       }
       $38 = (($$ph) + -1)|0;
       $39 = 1 << $38;
       $40 = ($39|0)>(1);
       if ($40) {
        $i$034 = 1;
        while(1) {
         $41 = (($val) + (($i$034*20)|0)|0);
         _EGBN_init($41);
         $42 = (($i$034) + -1)|0;
         $43 = (($val) + (($42*20)|0)|0);
         $44 = (_EGBN_mod_mul_montgomery($41,$43,$7,$mont$0,$ctx)|0);
         $45 = ($44|0)==(0);
         if ($45) {
          $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = 1;
          break L13;
         }
         $46 = (($i$034) + 1)|0;
         $47 = ($46|0)<($39|0);
         if ($47) {
          $i$034 = $46;
         } else {
          $53 = $$ph;$ts$0 = $46;
          break;
         }
        }
       } else {
        $53 = $$ph;$ts$0 = 1;
       }
      }
      $48 = (_EGBN_value_one()|0);
      $49 = (_EGBN_mod_mul_montgomery($8,$48,$29,$mont$0,$ctx)|0);
      $50 = ($49|0)==(0);
      if ($50) {
       $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = $ts$0;
      } else {
       $51 = (($4) + -1)|0;
       $52 = ($53|0)>(1);
       $start$0$ph = 1;$wstart$0$ph = $51;
       L38: while(1) {
        $54 = (_EGBN_is_bit_set($p,$wstart$0$ph)|0);
        $55 = ($54|0)==(0);
        L40: do {
         if ($55) {
          $56 = ($start$0$ph|0)==(0);
          if ($56) {
           $wstart$021$us = $wstart$0$ph;
          } else {
           $wstart$021 = $wstart$0$ph;
           while(1) {
            $63 = ($wstart$021|0)==(0);
            if ($63) {
             break L38;
            }
            $64 = (($wstart$021) + -1)|0;
            $65 = (_EGBN_is_bit_set($p,$64)|0);
            $66 = ($65|0)==(0);
            if ($66) {
             $wstart$021 = $64;
            } else {
             $wstart$0$lcssa = $64;
             break L40;
            }
           }
          }
          while(1) {
           $57 = (_EGBN_mod_mul_montgomery($8,$8,$8,$mont$0,$ctx)|0);
           $58 = ($57|0)==(0);
           if ($58) {
            $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = $ts$0;
            break L13;
           }
           $59 = ($wstart$021$us|0)==(0);
           if ($59) {
            break L38;
           }
           $60 = (($wstart$021$us) + -1)|0;
           $61 = (_EGBN_is_bit_set($p,$60)|0);
           $62 = ($61|0)==(0);
           if ($62) {
            $wstart$021$us = $60;
           } else {
            $wstart$0$lcssa = $60;
            break;
           }
          }
         } else {
          $wstart$0$lcssa = $wstart$0$ph;
         }
        } while(0);
        L51: do {
         if ($52) {
          $i$125 = 1;$wend$023 = 0;$wvalue$024 = 1;
          while(1) {
           $67 = (($wstart$0$lcssa) - ($i$125))|0;
           $68 = ($67|0)<(0);
           if ($68) {
            $wend$0$lcssa = $wend$023;$wvalue$0$lcssa = $wvalue$024;
            break L51;
           }
           $69 = (_EGBN_is_bit_set($p,$67)|0);
           $70 = ($69|0)==(0);
           if ($70) {
            $wend$1 = $wend$023;$wvalue$1 = $wvalue$024;
           } else {
            $71 = (($i$125) - ($wend$023))|0;
            $72 = $wvalue$024 << $71;
            $73 = $72 | 1;
            $wend$1 = $i$125;$wvalue$1 = $73;
           }
           $74 = (($i$125) + 1)|0;
           $75 = ($74|0)<($53|0);
           if ($75) {
            $i$125 = $74;$wend$023 = $wend$1;$wvalue$024 = $wvalue$1;
           } else {
            $wend$0$lcssa = $wend$1;$wvalue$0$lcssa = $wvalue$1;
            break;
           }
          }
         } else {
          $wend$0$lcssa = 0;$wvalue$0$lcssa = 1;
         }
        } while(0);
        $76 = (($wend$0$lcssa) + 1)|0;
        $77 = ($start$0$ph|0)==(0);
        $78 = ($wend$0$lcssa|0)>(-1);
        $or$cond53 = $77 & $78;
        if ($or$cond53) {
         $i$232 = 0;
         while(1) {
          $81 = (_EGBN_mod_mul_montgomery($8,$8,$8,$mont$0,$ctx)|0);
          $82 = ($81|0)==(0);
          $80 = (($i$232) + 1)|0;
          if ($82) {
           $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = $ts$0;
           break L13;
          }
          $79 = ($80|0)<($76|0);
          if ($79) {
           $i$232 = $80;
          } else {
           break;
          }
         }
        }
        $83 = $wvalue$0$lcssa >> 1;
        $84 = (($val) + (($83*20)|0)|0);
        $85 = (_EGBN_mod_mul_montgomery($8,$8,$84,$mont$0,$ctx)|0);
        $86 = ($85|0)==(0);
        if ($86) {
         $mont$1 = $mont$0;$ret$0 = 0;$ts$1 = $ts$0;
         break L13;
        }
        $87 = (($wstart$0$lcssa) - ($76))|0;
        $88 = ($87|0)<(0);
        if ($88) {
         break;
        } else {
         $start$0$ph = 0;$wstart$0$ph = $87;
        }
       }
       $89 = (_EGBN_from_montgomery($rr,$8,$mont$0,$ctx)|0);
       $not$ = ($89|0)!=(0);
       $$ = $not$&1;
       $mont$1 = $mont$0;$ret$0 = $$;$ts$1 = $ts$0;
      }
     }
    }
   } while(0);
   $90 = ($mont$1|0)!=(0|0);
   $or$cond3 = $11 & $90;
   if ($or$cond3) {
    _EGBN_MONT_CTX_free($mont$1);
   }
   _EGBN_CTX_end($ctx);
   $91 = ($ts$1|0)>(0);
   if ($91) {
    $i$320 = 0;
   } else {
    $$0 = $ret$0;
    STACKTOP = sp;return ($$0|0);
   }
   while(1) {
    $92 = (($val) + (($i$320*20)|0)|0);
    _EGBN_clear_free($92);
    $93 = (($i$320) + 1)|0;
    $exitcond = ($93|0)==($ts$1|0);
    if ($exitcond) {
     $$0 = $ret$0;
     break;
    } else {
     $i$320 = $93;
    }
   }
   STACKTOP = sp;return ($$0|0);
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_mod_exp_recp($r,$a,$p,$m,$ctx) {
 $r = $r|0;
 $a = $a|0;
 $p = $p|0;
 $m = $m|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$ph = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $9 = 0, $exitcond = 0, $i$026 = 0, $i$115 = 0, $i$224 = 0, $i$311 = 0, $or$cond = 0, $recp = 0, $ret$0 = 0, $ret$0$ph = 0, $ret$036 = 0, $ret$037 = 0, $start$0$ph = 0, $ts$0 = 0, $ts$138 = 0, $val = 0, $wend$0$lcssa = 0;
 var $wend$014 = 0, $wend$1 = 0, $wstart$0$lcssa = 0, $wstart$0$ph = 0, $wstart$012 = 0, $wstart$012$us = 0, $wvalue$0$lcssa = 0, $wvalue$016 = 0, $wvalue$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 704|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(704|0);
 $val = sp + 56|0;
 $recp = sp;
 $0 = (_EGBN_num_bits($p)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = (_EGBN_set_word($r,1)|0);
  $$0 = $2;
  STACKTOP = sp;return ($$0|0);
 }
 _EGBN_CTX_start($ctx);
 $3 = (_EGBN_CTX_get($ctx)|0);
 $4 = ($3|0)==(0|0);
 L5: do {
  if ($4) {
   label = 40;
  } else {
   _EGBN_RECP_CTX_init($recp);
   $5 = ((($m)) + 12|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = ($6|0)==(0);
   if ($7) {
    $13 = (_EGBN_RECP_CTX_set($recp,$m,$ctx)|0);
    $14 = ($13|0)<(1);
    if ($14) {
     label = 40;
     break;
    }
   } else {
    $8 = (_EGBN_copy($3,$m)|0);
    $9 = ($8|0)==(0|0);
    if ($9) {
     label = 40;
     break;
    }
    $10 = ((($3)) + 12|0);
    HEAP32[$10>>2] = 0;
    $11 = (_EGBN_RECP_CTX_set($recp,$3,$ctx)|0);
    $12 = ($11|0)<(1);
    if ($12) {
     label = 40;
     break;
    }
   }
   _EGBN_init($val);
   $15 = (_EGBN_nnmod($val,$a,$m,$ctx)|0);
   $16 = ($15|0)==(0);
   L12: do {
    if ($16) {
     $ret$0$ph = 0;
     label = 39;
    } else {
     $17 = ((($val)) + 4|0);
     $18 = HEAP32[$17>>2]|0;
     switch ($18|0) {
     case 0:  {
      label = 11;
      break;
     }
     case 1:  {
      $19 = HEAP32[$val>>2]|0;
      $20 = HEAP32[$19>>2]|0;
      $21 = ($20|0)==(0);
      if ($21) {
       label = 11;
      }
      break;
     }
     default: {
     }
     }
     if ((label|0) == 11) {
      $22 = (_EGBN_set_word($r,0)|0);
      $ret$0$ph = $22;
      label = 39;
      break;
     }
     $23 = ($0|0)>(671);
     if ($23) {
      $$ph = 6;
      label = 16;
     } else {
      $24 = ($0|0)>(239);
      if ($24) {
       $$ph = 5;
       label = 16;
      } else {
       $25 = ($0|0)>(79);
       if ($25) {
        $$ph = 4;
        label = 16;
       } else {
        $26 = ($0|0)>(23);
        if ($26) {
         $$ph = 3;
         label = 16;
        } else {
         $43 = 1;$ts$0 = 1;
        }
       }
      }
     }
     if ((label|0) == 16) {
      $27 = (_EGBN_mod_mul_reciprocal($3,$val,$val,$recp,$ctx)|0);
      $28 = ($27|0)==(0);
      if ($28) {
       $ret$0$ph = 0;
       label = 39;
       break;
      }
      $29 = (($$ph) + -1)|0;
      $30 = 1 << $29;
      $31 = ($30|0)>(1);
      if ($31) {
       $i$026 = 1;
       while(1) {
        $32 = (($val) + (($i$026*20)|0)|0);
        _EGBN_init($32);
        $33 = (($i$026) + -1)|0;
        $34 = (($val) + (($33*20)|0)|0);
        $35 = (_EGBN_mod_mul_reciprocal($32,$34,$3,$recp,$ctx)|0);
        $36 = ($35|0)==(0);
        if ($36) {
         $ret$0$ph = 0;
         label = 39;
         break L12;
        }
        $37 = (($i$026) + 1)|0;
        $38 = ($37|0)<($30|0);
        if ($38) {
         $i$026 = $37;
        } else {
         $43 = $$ph;$ts$0 = $37;
         break;
        }
       }
      } else {
       $43 = $$ph;$ts$0 = 1;
      }
     }
     $39 = (_EGBN_set_word($r,1)|0);
     $40 = ($39|0)==(0);
     L30: do {
      if ($40) {
       $ret$0 = 0;
      } else {
       $41 = (($0) + -1)|0;
       $42 = ($43|0)>(1);
       $start$0$ph = 1;$wstart$0$ph = $41;
       while(1) {
        $44 = (_EGBN_is_bit_set($p,$wstart$0$ph)|0);
        $45 = ($44|0)==(0);
        L34: do {
         if ($45) {
          $46 = ($start$0$ph|0)==(0);
          if ($46) {
           $wstart$012$us = $wstart$0$ph;
          } else {
           $wstart$012 = $wstart$0$ph;
           while(1) {
            $53 = ($wstart$012|0)==(0);
            if ($53) {
             $ret$0 = 1;
             break L30;
            }
            $54 = (($wstart$012) + -1)|0;
            $55 = (_EGBN_is_bit_set($p,$54)|0);
            $56 = ($55|0)==(0);
            if ($56) {
             $wstart$012 = $54;
            } else {
             $wstart$0$lcssa = $54;
             break L34;
            }
           }
          }
          while(1) {
           $47 = (_EGBN_mod_mul_reciprocal($r,$r,$r,$recp,$ctx)|0);
           $48 = ($47|0)==(0);
           if ($48) {
            $ret$0 = 0;
            break L30;
           }
           $49 = ($wstart$012$us|0)==(0);
           if ($49) {
            $ret$0 = 1;
            break L30;
           }
           $50 = (($wstart$012$us) + -1)|0;
           $51 = (_EGBN_is_bit_set($p,$50)|0);
           $52 = ($51|0)==(0);
           if ($52) {
            $wstart$012$us = $50;
           } else {
            $wstart$0$lcssa = $50;
            break;
           }
          }
         } else {
          $wstart$0$lcssa = $wstart$0$ph;
         }
        } while(0);
        L45: do {
         if ($42) {
          $i$115 = 1;$wend$014 = 0;$wvalue$016 = 1;
          while(1) {
           $57 = (($wstart$0$lcssa) - ($i$115))|0;
           $58 = ($57|0)<(0);
           if ($58) {
            $wend$0$lcssa = $wend$014;$wvalue$0$lcssa = $wvalue$016;
            break L45;
           }
           $59 = (_EGBN_is_bit_set($p,$57)|0);
           $60 = ($59|0)==(0);
           if ($60) {
            $wend$1 = $wend$014;$wvalue$1 = $wvalue$016;
           } else {
            $61 = (($i$115) - ($wend$014))|0;
            $62 = $wvalue$016 << $61;
            $63 = $62 | 1;
            $wend$1 = $i$115;$wvalue$1 = $63;
           }
           $64 = (($i$115) + 1)|0;
           $65 = ($64|0)<($43|0);
           if ($65) {
            $i$115 = $64;$wend$014 = $wend$1;$wvalue$016 = $wvalue$1;
           } else {
            $wend$0$lcssa = $wend$1;$wvalue$0$lcssa = $wvalue$1;
            break;
           }
          }
         } else {
          $wend$0$lcssa = 0;$wvalue$0$lcssa = 1;
         }
        } while(0);
        $66 = (($wend$0$lcssa) + 1)|0;
        $67 = ($start$0$ph|0)==(0);
        $68 = ($wend$0$lcssa|0)>(-1);
        $or$cond = $67 & $68;
        if ($or$cond) {
         $i$224 = 0;
         while(1) {
          $71 = (_EGBN_mod_mul_reciprocal($r,$r,$r,$recp,$ctx)|0);
          $72 = ($71|0)==(0);
          $70 = (($i$224) + 1)|0;
          if ($72) {
           $ret$0 = 0;
           break L30;
          }
          $69 = ($70|0)<($66|0);
          if ($69) {
           $i$224 = $70;
          } else {
           break;
          }
         }
        }
        $73 = $wvalue$0$lcssa >> 1;
        $74 = (($val) + (($73*20)|0)|0);
        $75 = (_EGBN_mod_mul_reciprocal($r,$r,$74,$recp,$ctx)|0);
        $76 = ($75|0)==(0);
        if ($76) {
         $ret$0 = 0;
         break L30;
        }
        $77 = (($wstart$0$lcssa) - ($66))|0;
        $78 = ($77|0)<(0);
        if ($78) {
         $ret$0 = 1;
         break;
        } else {
         $start$0$ph = 0;$wstart$0$ph = $77;
        }
       }
      }
     } while(0);
     _EGBN_CTX_end($ctx);
     $79 = ($ts$0|0)>(0);
     if ($79) {
      $ret$037 = $ret$0;$ts$138 = $ts$0;
     } else {
      $ret$036 = $ret$0;
      break L5;
     }
    }
   } while(0);
   if ((label|0) == 39) {
    _EGBN_CTX_end($ctx);
    $ret$037 = $ret$0$ph;$ts$138 = 1;
   }
   $i$311 = 0;
   while(1) {
    $80 = (($val) + (($i$311*20)|0)|0);
    _EGBN_clear_free($80);
    $81 = (($i$311) + 1)|0;
    $exitcond = ($81|0)==($ts$138|0);
    if ($exitcond) {
     $ret$036 = $ret$037;
     break;
    } else {
     $i$311 = $81;
    }
   }
  }
 } while(0);
 if ((label|0) == 40) {
  _EGBN_CTX_end($ctx);
  $ret$036 = 0;
 }
 _EGBN_RECP_CTX_free($recp);
 $$0 = $ret$036;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_set_word($a,$w) {
 $a = $a|0;
 $w = $w|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(0);
 if ($2) {
  $$pre = HEAP32[$a>>2]|0;
  $9 = $$pre;
 } else {
  $3 = (_bn_expand_internal($a,2)|0);
  $4 = ($3|0)==(0|0);
  if ($4) {
   $$0 = 0;
   return ($$0|0);
  }
  $5 = HEAP32[$a>>2]|0;
  $6 = ($5|0)==(0|0);
  if (!($6)) {
   _free($5);
  }
  HEAP32[$a>>2] = $3;
  HEAP32[$0>>2] = 2;
  $9 = $3;
 }
 $7 = ((($a)) + 12|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($a)) + 4|0);
 HEAP32[$8>>2] = 0;
 HEAP32[$9>>2] = $w;
 $10 = ($w|0)==(0);
 if ($10) {
  $$0 = 1;
  return ($$0|0);
 }
 HEAP32[$8>>2] = 1;
 $$0 = 1;
 return ($$0|0);
}
function _EGBN_CTX_start($ctx) {
 $ctx = $ctx|0;
 var $$pre = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ctx)) + 648|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(12);
 if ($2) {
  $3 = HEAP32[$ctx>>2]|0;
  $4 = (((($ctx)) + 652|0) + ($1<<2)|0);
  HEAP32[$4>>2] = $3;
  $$pre = HEAP32[$0>>2]|0;
  $6 = $$pre;
 } else {
  $6 = $1;
 }
 $5 = (($6) + 1)|0;
 HEAP32[$0>>2] = $5;
 return;
}
function _EGBN_CTX_get($ctx) {
 $ctx = $ctx|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($ctx)) + 648|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)>(12);
 if (!($2)) {
  $3 = HEAP32[$ctx>>2]|0;
  $4 = ($3|0)>(31);
  if (!($4)) {
   $8 = (($3) + 1)|0;
   HEAP32[$ctx>>2] = $8;
   $9 = (((($ctx)) + 4|0) + (($3*20)|0)|0);
   $$0 = $9;
   return ($$0|0);
  }
 }
 $5 = ((($ctx)) + 700|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)==(0);
 if (!($7)) {
  $$0 = 0;
  return ($$0|0);
 }
 HEAP32[$5>>2] = 1;
 $$0 = 0;
 return ($$0|0);
}
function _EGBN_RECP_CTX_init($recp) {
 $recp = $recp|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _EGBN_init($recp);
 $0 = ((($recp)) + 20|0);
 _EGBN_init($0);
 $1 = ((($recp)) + 40|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($recp)) + 48|0);
 HEAP32[$2>>2] = 0;
 return;
}
function _EGBN_copy($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $$pr = 0, $$pre = 0, $$pre9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $A$0$lcssa = 0, $A$03 = 0, $B$0$lcssa = 0, $B$02 = 0, $i$04 = 0, $or$cond = 0, $scevgep = 0, $scevgep7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($a|0)==($b|0);
 if ($0) {
  $$0 = $a;
  return ($$0|0);
 }
 $1 = ((($b)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($a)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)>($4|0);
 if ($5) {
  $6 = (_bn_expand_internal($a,$2)|0);
  $7 = ($6|0)==(0|0);
  if ($7) {
   $$0 = 0;
   return ($$0|0);
  }
  $8 = HEAP32[$a>>2]|0;
  $9 = ($8|0)==(0|0);
  if (!($9)) {
   _free($8);
  }
  HEAP32[$a>>2] = $6;
  HEAP32[$3>>2] = $2;
  $$pre9 = HEAP32[$1>>2]|0;
  $$pr = $$pre9;$28 = $6;
 } else {
  $$pre = HEAP32[$a>>2]|0;
  $$pr = $2;$28 = $$pre;
 }
 $10 = HEAP32[$b>>2]|0;
 $11 = $$pr >> 2;
 $12 = ($11|0)>(0);
 if ($12) {
  $13 = $11 << 2;
  $scevgep = (($10) + ($13<<2)|0);
  $A$03 = $28;$B$02 = $10;$i$04 = $11;
  while(1) {
   $14 = HEAP32[$B$02>>2]|0;
   $15 = ((($B$02)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($B$02)) + 8|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ((($B$02)) + 12|0);
   $20 = HEAP32[$19>>2]|0;
   HEAP32[$A$03>>2] = $14;
   $21 = ((($A$03)) + 4|0);
   HEAP32[$21>>2] = $16;
   $22 = ((($A$03)) + 8|0);
   HEAP32[$22>>2] = $18;
   $23 = ((($A$03)) + 12|0);
   HEAP32[$23>>2] = $20;
   $24 = (($i$04) + -1)|0;
   $25 = ((($A$03)) + 16|0);
   $26 = ((($B$02)) + 16|0);
   $27 = ($i$04|0)>(1);
   if ($27) {
    $A$03 = $25;$B$02 = $26;$i$04 = $24;
   } else {
    break;
   }
  }
  $scevgep7 = (($28) + ($13<<2)|0);
  $A$0$lcssa = $scevgep7;$B$0$lcssa = $scevgep;
 } else {
  $A$0$lcssa = $28;$B$0$lcssa = $10;
 }
 $29 = $$pr & 3;
 switch ($29|0) {
 case 3:  {
  $30 = ((($B$0$lcssa)) + 8|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($A$0$lcssa)) + 8|0);
  HEAP32[$32>>2] = $31;
  label = 14;
  break;
 }
 case 2:  {
  label = 14;
  break;
 }
 case 1:  {
  label = 15;
  break;
 }
 default: {
 }
 }
 if ((label|0) == 14) {
  $33 = ((($B$0$lcssa)) + 4|0);
  $34 = HEAP32[$33>>2]|0;
  $35 = ((($A$0$lcssa)) + 4|0);
  HEAP32[$35>>2] = $34;
  label = 15;
 }
 if ((label|0) == 15) {
  $36 = HEAP32[$B$0$lcssa>>2]|0;
  HEAP32[$A$0$lcssa>>2] = $36;
 }
 $37 = ((($a)) + 4|0);
 HEAP32[$37>>2] = $$pr;
 $38 = ($$pr|0)!=(0);
 $39 = ($28|0)==(0|0);
 $or$cond = $38 | $39;
 if (!($or$cond)) {
  HEAP32[$28>>2] = 0;
 }
 $40 = ((($b)) + 12|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = ((($a)) + 12|0);
 HEAP32[$42>>2] = $41;
 $$0 = $a;
 return ($$0|0);
}
function _EGBN_RECP_CTX_set($recp,$d,$ctx) {
 $recp = $recp|0;
 $d = $d|0;
 $ctx = $ctx|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_EGBN_copy($recp,$d)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($recp)) + 20|0);
 $3 = (_EGBN_set_word($2,0)|0);
 $4 = ($3|0)==(0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = (_EGBN_num_bits($d)|0);
 $6 = ((($recp)) + 40|0);
 HEAP32[$6>>2] = $5;
 $7 = ((($recp)) + 44|0);
 HEAP32[$7>>2] = 0;
 $$0 = 1;
 return ($$0|0);
}
function _EGBN_init($a) {
 $a = $a|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$a>>2]=0|0;HEAP32[$a+4>>2]=0|0;HEAP32[$a+8>>2]=0|0;HEAP32[$a+12>>2]=0|0;HEAP32[$a+16>>2]=0|0;
 return;
}
function _EGBN_nnmod($r,$m,$d,$ctx) {
 $r = $r|0;
 $m = $m|0;
 $d = $d|0;
 $ctx = $ctx|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_EGBN_div(0,$r,$m,$d,$ctx)|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 $2 = ((($r)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0);
 if ($4) {
  $$0 = 1;
  return ($$0|0);
 }
 $5 = ((($d)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(0);
 $8 = $7 ? 915 : 914;
 $9 = (FUNCTION_TABLE_iiii[$8 & 2047]($r,$r,$d)|0);
 $$0 = $9;
 return ($$0|0);
}
function _EGBN_mod_mul_reciprocal($r,$x,$y,$recp,$ctx) {
 $r = $r|0;
 $x = $x|0;
 $y = $y|0;
 $recp = $recp|0;
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $ca$0 = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _EGBN_CTX_start($ctx);
 $0 = (_EGBN_CTX_get($ctx)|0);
 $1 = ($0|0)==(0|0);
 L1: do {
  if ($1) {
   $ret$0 = 0;
  } else {
   $2 = ($y|0)==(0|0);
   do {
    if ($2) {
     $ca$0 = $x;
    } else {
     $3 = ($x|0)==($y|0);
     if ($3) {
      $4 = (_EGBN_sqr($0,$y,$ctx)|0);
      $5 = ($4|0)==(0);
      if ($5) {
       $ret$0 = 0;
       break L1;
      } else {
       $ca$0 = $0;
       break;
      }
     } else {
      $6 = (_EGBN_mul($0,$x,$y,$ctx)|0);
      $7 = ($6|0)==(0);
      if ($7) {
       $ret$0 = 0;
       break L1;
      } else {
       $ca$0 = $0;
       break;
      }
     }
    }
   } while(0);
   $8 = (_EGBN_div_recp(0,$r,$ca$0,$recp,$ctx)|0);
   $ret$0 = $8;
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 return ($ret$0|0);
}
function _EGBN_is_bit_set($a,$n) {
 $a = $a|0;
 $n = $n|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(0);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = (($n|0) / 32)&-1;
 $2 = ((($a)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)>($1|0);
 if (!($4)) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = HEAP32[$a>>2]|0;
 $6 = (($5) + ($1<<2)|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $n & 31;
 $9 = 1 << $8;
 $10 = $7 & $9;
 $11 = ($10|0)!=(0);
 $12 = $11&1;
 $$0 = $12;
 return ($$0|0);
}
function _EGBN_CTX_end($ctx) {
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($ctx|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($ctx)) + 648|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = HEAP32[$ctx>>2]|0;
  $5 = ((($ctx)) + 652|0);
  HEAP32[$5>>2] = $4;
  $6 = ((($ctx)) + 700|0);
  HEAP32[$6>>2] = 0;
  HEAP32[$1>>2] = 0;
  $11 = 0;
 } else {
  $7 = ((($ctx)) + 700|0);
  HEAP32[$7>>2] = 0;
  $8 = (($2) + -1)|0;
  HEAP32[$1>>2] = $8;
  $9 = ($2|0)<(13);
  if ($9) {
   $11 = $8;
  } else {
   return;
  }
 }
 $10 = (((($ctx)) + 652|0) + ($11<<2)|0);
 $12 = HEAP32[$10>>2]|0;
 HEAP32[$ctx>>2] = $12;
 return;
}
function _EGBN_RECP_CTX_free($recp) {
 $recp = $recp|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($recp|0)==(0|0);
 if ($0) {
  return;
 }
 _EGBN_free($recp);
 $1 = ((($recp)) + 20|0);
 _EGBN_free($1);
 $2 = ((($recp)) + 48|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & 1;
 $5 = ($4|0)==(0);
 if ($5) {
  return;
 }
 _free($recp);
 return;
}
function _EGBN_sqr($r,$a,$ctx) {
 $r = $r|0;
 $a = $a|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$pre$phiZ2D = 0, $$pre$pre$phiZZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ret$0 = 0;
 var $t = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $t = sp;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<(1);
 if ($2) {
  $3 = ((($r)) + 4|0);
  HEAP32[$3>>2] = 0;
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 _EGBN_CTX_start($ctx);
 $4 = ($a|0)==($r|0);
 if ($4) {
  $5 = (_EGBN_CTX_get($ctx)|0);
  $11 = $5;
 } else {
  $11 = $r;
 }
 $6 = (_EGBN_CTX_get($ctx)|0);
 $7 = ($6|0)==(0|0);
 L8: do {
  if ($7) {
   $ret$0 = 0;
  } else {
   $8 = $1 << 1;
   $9 = $8 | 1;
   $10 = ((($11)) + 8|0);
   $12 = HEAP32[$10>>2]|0;
   $13 = ($9|0)>($12|0);
   if ($13) {
    $14 = (_egbn_expand2($11,$9)|0);
    $15 = ($14|0)==(0|0);
    if ($15) {
     $ret$0 = 0;
     break;
    }
   }
   L13: do {
    switch ($1|0) {
    case 4:  {
     $16 = HEAP32[$11>>2]|0;
     $17 = HEAP32[$a>>2]|0;
     _egbn_sqr_comba4($16,$17);
     $$pre$pre$phiZZ2D = $11;
     label = 19;
     break;
    }
    case 8:  {
     $18 = HEAP32[$11>>2]|0;
     $19 = HEAP32[$a>>2]|0;
     _egbn_sqr_comba8($18,$19);
     $$pre$pre$phiZZ2D = $11;
     label = 19;
     break;
    }
    default: {
     $20 = ($1|0)<(16);
     if ($20) {
      $44 = HEAP32[$11>>2]|0;
      $45 = HEAP32[$a>>2]|0;
      _egbn_sqr_normal($44,$45,$1,$t);
      $46 = ((($11)) + 4|0);
      HEAP32[$46>>2] = $8;
      $47 = ((($11)) + 12|0);
      HEAP32[$47>>2] = 0;
      $48 = ($1|0)>(0);
      if ($48) {
       $$pre$phiZ2D = $11;$54 = $46;
       label = 21;
       break L13;
      } else {
       break L13;
      }
     }
     $21 = (_EGBN_num_bits_word($1)|0);
     $22 = (($21) + -1)|0;
     $23 = 1 << $22;
     $24 = ($1|0)==($23|0);
     if ($24) {
      $25 = $1 << 2;
      $26 = ((($6)) + 8|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($25|0)>($27|0);
      if ($28) {
       $29 = (_egbn_expand2($6,$25)|0);
       $30 = ($29|0)==(0|0);
       if ($30) {
        $ret$0 = 0;
        break L8;
       }
      }
      $31 = HEAP32[$11>>2]|0;
      $32 = HEAP32[$a>>2]|0;
      $33 = HEAP32[$6>>2]|0;
      _egbn_sqr_recursive($31,$32,$1,$33);
      $$pre$pre$phiZZ2D = $11;
      label = 19;
      break L13;
     } else {
      $34 = ((($6)) + 8|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ($8|0)>($35|0);
      if ($36) {
       $37 = (_egbn_expand2($6,$8)|0);
       $38 = ($37|0)==(0|0);
       if ($38) {
        $ret$0 = 0;
        break L8;
       }
      }
      $39 = HEAP32[$11>>2]|0;
      $40 = HEAP32[$a>>2]|0;
      $41 = HEAP32[$6>>2]|0;
      _egbn_sqr_normal($39,$40,$1,$41);
      $$pre$pre$phiZZ2D = $11;
      label = 19;
      break L13;
     }
    }
    }
   } while(0);
   if ((label|0) == 19) {
    $42 = ((($11)) + 4|0);
    HEAP32[$42>>2] = $8;
    $43 = ((($11)) + 12|0);
    HEAP32[$43>>2] = 0;
    $$pre$phiZ2D = $$pre$pre$phiZZ2D;$54 = $42;
    label = 21;
   }
   if ((label|0) == 21) {
    $49 = (($8) + -1)|0;
    $50 = HEAP32[$$pre$phiZ2D>>2]|0;
    $51 = (($50) + ($49<<2)|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = ($52|0)==(0);
    if ($53) {
     HEAP32[$54>>2] = $49;
    }
   }
   $55 = ($11|0)==($r|0);
   if ($55) {
    $ret$0 = 1;
   } else {
    (_EGBN_copy($r,$11)|0);
    $ret$0 = 1;
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = $ret$0;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_mul($r,$a,$b,$ctx) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $al$011 = 0, $al$1 = 0, $bl$010 = 0;
 var $bl$1 = 0, $ftl$015 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $or$cond6 = 0, $ret$0 = 0, $rr$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==(0);
 $5 = ($3|0)==(0);
 $or$cond = $4 | $5;
 L1: do {
  if ($or$cond) {
   $6 = (_EGBN_set_word($r,0)|0);
   $7 = ($6|0)==(0);
   if ($7) {
    $ret$0 = 0;
   } else {
    $$0 = 1;
    return ($$0|0);
   }
  } else {
   $8 = (($3) + ($1))|0;
   _EGBN_CTX_start($ctx);
   $9 = ($r|0)==($a|0);
   $10 = ($r|0)==($b|0);
   $or$cond6 = $9 | $10;
   if ($or$cond6) {
    $11 = (_EGBN_CTX_get($ctx)|0);
    $12 = ($11|0)==(0|0);
    if ($12) {
     $ret$0 = 0;
     break;
    } else {
     $rr$0 = $11;
    }
   } else {
    $rr$0 = $r;
   }
   $13 = ((($a)) + 12|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ((($b)) + 12|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = $16 ^ $14;
   $18 = ((($rr$0)) + 12|0);
   HEAP32[$18>>2] = $17;
   $19 = (($1) - ($3))|0;
   $20 = ($1|0)==($3|0);
   $21 = ($1|0)==(8);
   $or$cond5 = $21 & $20;
   L8: do {
    if ($or$cond5) {
     $22 = ((($rr$0)) + 8|0);
     $23 = HEAP32[$22>>2]|0;
     $24 = ($23|0)>(15);
     if (!($24)) {
      $25 = (_egbn_expand2($rr$0,16)|0);
      $26 = ($25|0)==(0|0);
      if ($26) {
       $ret$0 = 0;
       break L1;
      }
     }
     $27 = ((($rr$0)) + 4|0);
     HEAP32[$27>>2] = 16;
     $28 = HEAP32[$rr$0>>2]|0;
     $29 = HEAP32[$a>>2]|0;
     $30 = HEAP32[$b>>2]|0;
     _egbn_mul_comba8($28,$29,$30);
     $$pre$phiZ2D = $27;
    } else {
     $31 = ($1|0)>(15);
     $32 = ($3|0)>(15);
     $or$cond3 = $31 & $32;
     L14: do {
      if ($or$cond3) {
       switch ($19|0) {
       case 1:  {
        $33 = ((($b)) + 16|0);
        $34 = HEAP32[$33>>2]|0;
        $35 = $34 & 2;
        $36 = ($35|0)==(0);
        if (!($36)) {
         $al$1 = $1;$bl$1 = $3;
         break L14;
        }
        $37 = ((($b)) + 8|0);
        $38 = HEAP32[$37>>2]|0;
        $39 = ($3|0)<($38|0);
        if (!($39)) {
         $al$1 = $1;$bl$1 = $3;
         break L14;
        }
        $40 = HEAP32[$b>>2]|0;
        $41 = (($40) + ($3<<2)|0);
        HEAP32[$41>>2] = 0;
        $42 = (($3) + 1)|0;
        $al$011 = $1;$bl$010 = $42;
        break;
       }
       case -1:  {
        $43 = ((($a)) + 16|0);
        $44 = HEAP32[$43>>2]|0;
        $45 = $44 & 2;
        $46 = ($45|0)==(0);
        if (!($46)) {
         $al$1 = $1;$bl$1 = $3;
         break L14;
        }
        $47 = ((($a)) + 8|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = ($1|0)<($48|0);
        if (!($49)) {
         $al$1 = $1;$bl$1 = $3;
         break L14;
        }
        $50 = HEAP32[$a>>2]|0;
        $51 = (($50) + ($1<<2)|0);
        HEAP32[$51>>2] = 0;
        $52 = (($1) + 1)|0;
        $al$011 = $52;$bl$010 = $3;
        break;
       }
       case 0:  {
        $al$011 = $1;$bl$010 = $3;
        break;
       }
       default: {
        $al$1 = $1;$bl$1 = $3;
        break L14;
       }
       }
       $53 = (_EGBN_num_bits_word($al$011)|0);
       $54 = (($53) + -1)|0;
       $55 = 1 << $54;
       $56 = (_EGBN_CTX_get($ctx)|0);
       $57 = ($al$011|0)==($55|0);
       if ($57) {
        $58 = $al$011 << 2;
        $59 = ((($56)) + 8|0);
        $60 = HEAP32[$59>>2]|0;
        $61 = ($58|0)>($60|0);
        if ($61) {
         $62 = (_egbn_expand2($56,$58)|0);
         $63 = ($62|0)==(0|0);
         if ($63) {
          $ret$0 = 0;
          break L1;
         }
        }
        $64 = ((($rr$0)) + 8|0);
        $65 = HEAP32[$64>>2]|0;
        $66 = ($58|0)>($65|0);
        if ($66) {
         $67 = (_egbn_expand2($rr$0,$58)|0);
         $68 = ($67|0)==(0|0);
         if ($68) {
          $ret$0 = 0;
          break L1;
         }
        }
        $69 = HEAP32[$rr$0>>2]|0;
        $70 = HEAP32[$a>>2]|0;
        $71 = HEAP32[$b>>2]|0;
        $72 = HEAP32[$56>>2]|0;
        _egbn_mul_recursive($69,$70,$71,$al$011,$72);
        $73 = ((($rr$0)) + 4|0);
        HEAP32[$73>>2] = $8;
        $$pre$phiZ2D = $73;
        break L8;
       } else {
        $al$1 = $al$011;$bl$1 = $bl$010;
       }
      } else {
       $al$1 = $1;$bl$1 = $3;
      }
     } while(0);
     $74 = ((($rr$0)) + 8|0);
     $75 = HEAP32[$74>>2]|0;
     $76 = ($8|0)>($75|0);
     if ($76) {
      $77 = (_egbn_expand2($rr$0,$8)|0);
      $78 = ($77|0)==(0|0);
      if ($78) {
       $ret$0 = 0;
       break L1;
      }
     }
     $79 = ((($rr$0)) + 4|0);
     HEAP32[$79>>2] = $8;
     $80 = HEAP32[$rr$0>>2]|0;
     $81 = HEAP32[$a>>2]|0;
     $82 = HEAP32[$b>>2]|0;
     _egbn_mul_normal($80,$81,$al$1,$82,$bl$1);
     $$pre$phiZ2D = $79;
    }
   } while(0);
   $83 = HEAP32[$$pre$phiZ2D>>2]|0;
   $84 = ($83|0)>(0);
   L36: do {
    if ($84) {
     $85 = HEAP32[$rr$0>>2]|0;
     $86 = (($83) + -1)|0;
     $87 = (($85) + ($86<<2)|0);
     $92 = $83;$ftl$015 = $87;
     while(1) {
      $88 = HEAP32[$ftl$015>>2]|0;
      $89 = ($88|0)==(0);
      if (!($89)) {
       break L36;
      }
      $90 = ((($ftl$015)) + -4|0);
      $91 = (($92) + -1)|0;
      HEAP32[$$pre$phiZ2D>>2] = $91;
      $93 = ($92|0)>(1);
      if ($93) {
       $92 = $91;$ftl$015 = $90;
      } else {
       break;
      }
     }
    }
   } while(0);
   $94 = ($rr$0|0)==($r|0);
   if ($94) {
    $ret$0 = 1;
   } else {
    (_EGBN_copy($r,$rr$0)|0);
    $ret$0 = 1;
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = $ret$0;
 return ($$0|0);
}
function _EGBN_div_recp($dv,$rem,$m,$recp,$ctx) {
 $dv = $dv|0;
 $rem = $rem|0;
 $m = $m|0;
 $recp = $recp|0;
 $ctx = $ctx|0;
 var $$ = 0, $$0 = 0, $$len$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, $d$0 = 0, $j$0 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond5 = 0, $r$0 = 0, $ret$0 = 0;
 var $ret$0$i = 0, $t$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $t$i = sp;
 _EGBN_CTX_start($ctx);
 $0 = (_EGBN_CTX_get($ctx)|0);
 $1 = (_EGBN_CTX_get($ctx)|0);
 $2 = ($dv|0)==(0|0);
 if ($2) {
  $3 = (_EGBN_CTX_get($ctx)|0);
  $d$0 = $3;
 } else {
  $d$0 = $dv;
 }
 $4 = ($rem|0)==(0|0);
 if ($4) {
  $5 = (_EGBN_CTX_get($ctx)|0);
  $r$0 = $5;
 } else {
  $r$0 = $rem;
 }
 $6 = ($0|0)==(0|0);
 $7 = ($1|0)==(0|0);
 $or$cond = $6 | $7;
 $8 = ($d$0|0)==(0|0);
 $or$cond3 = $or$cond | $8;
 $9 = ($r$0|0)==(0|0);
 $or$cond5 = $or$cond3 | $9;
 L7: do {
  if ($or$cond5) {
   $ret$0 = 0;
  } else {
   $10 = (_EGBN_ucmp($m,$recp)|0);
   $11 = ($10|0)<(0);
   if ($11) {
    $12 = (_EGBN_set_word($d$0,0)|0);
    $13 = ($12|0)==(0);
    if ($13) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    $14 = (_EGBN_copy($r$0,$m)|0);
    $15 = ($14|0)==(0|0);
    if ($15) {
     $$0 = 0;
     STACKTOP = sp;return ($$0|0);
    }
    _EGBN_CTX_end($ctx);
    $$0 = 1;
    STACKTOP = sp;return ($$0|0);
   }
   $16 = (_EGBN_num_bits($m)|0);
   $17 = ((($recp)) + 40|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = $18 << 1;
   $20 = ($19|0)>($16|0);
   $$ = $20 ? $19 : $16;
   $21 = ((($recp)) + 44|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($$|0)==($22|0);
   if ($23) {
    $32 = $$;
   } else {
    $24 = ((($recp)) + 20|0);
    _EGBN_init($t$i);
    $25 = (_EGBN_set_word($t$i,0)|0);
    $26 = ($25|0)==(0);
    if ($26) {
     $ret$0$i = -1;
    } else {
     $27 = (_EGBN_set_bit($t$i,$$)|0);
     $28 = ($27|0)==(0);
     if ($28) {
      $ret$0$i = -1;
     } else {
      $29 = (_EGBN_div($24,0,$t$i,$recp,$ctx)|0);
      $30 = ($29|0)==(0);
      $$len$i = $30 ? -1 : $$;
      $ret$0$i = $$len$i;
     }
    }
    _EGBN_free($t$i);
    HEAP32[$21>>2] = $ret$0$i;
    $32 = $ret$0$i;
   }
   $31 = ($32|0)==(-1);
   if ($31) {
    $ret$0 = 0;
   } else {
    $33 = HEAP32[$17>>2]|0;
    $34 = (_EGBN_rshift($0,$m,$33)|0);
    $35 = ($34|0)==(0);
    if ($35) {
     $ret$0 = 0;
    } else {
     $36 = ((($recp)) + 20|0);
     $37 = (_EGBN_mul($1,$0,$36,$ctx)|0);
     $38 = ($37|0)==(0);
     if ($38) {
      $ret$0 = 0;
     } else {
      $39 = HEAP32[$17>>2]|0;
      $40 = (($$) - ($39))|0;
      $41 = (_EGBN_rshift($d$0,$1,$40)|0);
      $42 = ($41|0)==(0);
      if ($42) {
       $ret$0 = 0;
      } else {
       $43 = ((($d$0)) + 12|0);
       HEAP32[$43>>2] = 0;
       $44 = (_EGBN_mul($1,$recp,$d$0,$ctx)|0);
       $45 = ($44|0)==(0);
       if ($45) {
        $ret$0 = 0;
       } else {
        $46 = (_EGBN_usub($r$0,$m,$1)|0);
        $47 = ($46|0)==(0);
        if ($47) {
         $ret$0 = 0;
        } else {
         $48 = ((($r$0)) + 12|0);
         HEAP32[$48>>2] = 0;
         $j$0 = 0;
         while(1) {
          $49 = (_EGBN_ucmp($r$0,$recp)|0);
          $50 = ($49|0)>(-1);
          if (!($50)) {
           break;
          }
          $51 = (($j$0) + 1)|0;
          $52 = ($j$0|0)>(2);
          if ($52) {
           $ret$0 = 0;
           break L7;
          }
          $53 = (_EGBN_usub($r$0,$r$0,$recp)|0);
          $54 = ($53|0)==(0);
          if ($54) {
           $ret$0 = 0;
           break L7;
          }
          $55 = (_EGBN_add_word($d$0,1)|0);
          $56 = ($55|0)==(0);
          if ($56) {
           $ret$0 = 0;
           break L7;
          } else {
           $j$0 = $51;
          }
         }
         $57 = ((($r$0)) + 4|0);
         $58 = HEAP32[$57>>2]|0;
         switch ($58|0) {
         case 0:  {
          $64 = 0;
          break;
         }
         case 1:  {
          $59 = HEAP32[$r$0>>2]|0;
          $60 = HEAP32[$59>>2]|0;
          $61 = ($60|0)==(0);
          if ($61) {
           $64 = 0;
          } else {
           label = 28;
          }
          break;
         }
         default: {
          label = 28;
         }
         }
         if ((label|0) == 28) {
          $62 = ((($m)) + 12|0);
          $63 = HEAP32[$62>>2]|0;
          $64 = $63;
         }
         HEAP32[$48>>2] = $64;
         $65 = ((($m)) + 12|0);
         $66 = HEAP32[$65>>2]|0;
         $67 = ((($recp)) + 12|0);
         $68 = HEAP32[$67>>2]|0;
         $69 = $68 ^ $66;
         HEAP32[$43>>2] = $69;
         $ret$0 = 1;
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = $ret$0;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_ucmp($a,$b) {
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0;
 var $i$0$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if (!($4)) {
  $5 = (($1) - ($3))|0;
  $$0 = $5;
  return ($$0|0);
 }
 $6 = HEAP32[$a>>2]|0;
 $7 = HEAP32[$b>>2]|0;
 $i$0$in = $1;
 while(1) {
  $i$0 = (($i$0$in) + -1)|0;
  $8 = ($i$0$in|0)>(0);
  if (!($8)) {
   $$0 = 0;
   label = 7;
   break;
  }
  $9 = (($6) + ($i$0<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = (($7) + ($i$0<<2)|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ($10|0)==($12|0);
  if ($13) {
   $i$0$in = $i$0;
  } else {
   $$lcssa = $10;$$lcssa6 = $12;
   break;
  }
 }
 if ((label|0) == 7) {
  return ($$0|0);
 }
 $14 = ($$lcssa>>>0)>($$lcssa6>>>0);
 $15 = $14 ? 1 : -1;
 $$0 = $15;
 return ($$0|0);
}
function _EGBN_set_bit($a,$n) {
 $a = $a|0;
 $n = $n|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $scevgep = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($n|0) / 32)&-1;
 $1 = ((($a)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)>($0|0);
 if (!($3)) {
  $4 = (($0) + 1)|0;
  $5 = ((($a)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($0|0)>=($6|0);
  $8 = ($6|0)<($4|0);
  $or$cond = $7 & $8;
  if ($or$cond) {
   $9 = (_bn_expand_internal($a,$4)|0);
   $10 = ($9|0)==(0|0);
   if ($10) {
    $$0 = 0;
    return ($$0|0);
   }
   $11 = HEAP32[$a>>2]|0;
   $12 = ($11|0)==(0|0);
   if (!($12)) {
    _free($11);
   }
   HEAP32[$a>>2] = $9;
   HEAP32[$5>>2] = $4;
   $$pre = HEAP32[$1>>2]|0;
   $14 = $$pre;
  } else {
   $14 = $2;
  }
  $13 = ($0|0)<($14|0);
  if (!($13)) {
   $15 = HEAP32[$a>>2]|0;
   $scevgep = (($15) + ($14<<2)|0);
   $16 = (($4) - ($14))|0;
   $17 = $16 << 2;
   _memset(($scevgep|0),0,($17|0))|0;
  }
  HEAP32[$1>>2] = $4;
 }
 $18 = $n & 31;
 $19 = 1 << $18;
 $20 = HEAP32[$a>>2]|0;
 $21 = (($20) + ($0<<2)|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $22 | $19;
 HEAP32[$21>>2] = $23;
 $$0 = 1;
 return ($$0|0);
}
function _EGBN_div($dv,$rm,$num,$divisor,$ctx) {
 $dv = $dv|0;
 $rm = $rm|0;
 $num = $num|0;
 $divisor = $divisor|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $i$011 = 0, $j$0 = 0, $j$0$lcssa = 0, $or$cond = 0, $q$04 = 0, $q$1 = 0, $q$2 = 0, $rem$05 = 0, $res$0 = 0, $resp$010 = 0, $wnum = 0, $wnump$09 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $wnum = sp;
 $0 = ((($divisor)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 1:  {
  label = 2;
  break;
 }
 case 0:  {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
  break;
 }
 default: {
 }
 }
 if ((label|0) == 2) {
  $2 = HEAP32[$divisor>>2]|0;
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(0);
  if ($4) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $5 = (_EGBN_ucmp($num,$divisor)|0);
 $6 = ($5|0)<(0);
 if ($6) {
  $7 = ($rm|0)==(0|0);
  if (!($7)) {
   $8 = (_EGBN_copy($rm,$num)|0);
   $9 = ($8|0)==(0|0);
   if ($9) {
    $$0 = 0;
    STACKTOP = sp;return ($$0|0);
   }
  }
  $10 = ($dv|0)==(0|0);
  if ($10) {
   $$0 = 1;
   STACKTOP = sp;return ($$0|0);
  }
  (_EGBN_set_word($dv,0)|0);
  $$0 = 1;
  STACKTOP = sp;return ($$0|0);
 }
 _EGBN_CTX_start($ctx);
 $11 = (_EGBN_CTX_get($ctx)|0);
 $12 = (_EGBN_CTX_get($ctx)|0);
 $13 = (_EGBN_CTX_get($ctx)|0);
 $14 = ($dv|0)==(0|0);
 if ($14) {
  $15 = (_EGBN_CTX_get($ctx)|0);
  $res$0 = $15;
 } else {
  $res$0 = $dv;
 }
 $16 = ($13|0)==(0|0);
 $17 = ($res$0|0)==(0|0);
 $or$cond = $16 | $17;
 L21: do {
  if (!($or$cond)) {
   $18 = ((($11)) + 12|0);
   HEAP32[$18>>2] = 0;
   $19 = (_EGBN_num_bits($divisor)|0);
   $20 = (($19|0) % 32)&-1;
   $21 = (32 - ($20))|0;
   $22 = (_EGBN_lshift($13,$divisor,$21)|0);
   $23 = ($22|0)==(0);
   if (!($23)) {
    $24 = ((($13)) + 12|0);
    HEAP32[$24>>2] = 0;
    $25 = (($21) + 32)|0;
    $26 = (_EGBN_lshift($12,$num,$25)|0);
    $27 = ($26|0)==(0);
    if (!($27)) {
     $28 = ((($12)) + 12|0);
     HEAP32[$28>>2] = 0;
     $29 = ((($13)) + 4|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = ((($12)) + 4|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = (($32) - ($30))|0;
     _EGBN_init($wnum);
     $34 = HEAP32[$12>>2]|0;
     $35 = (($34) + ($33<<2)|0);
     HEAP32[$wnum>>2] = $35;
     $36 = ((($wnum)) + 4|0);
     HEAP32[$36>>2] = $30;
     $37 = ((($12)) + 8|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = (($38) + 1)|0;
     $40 = ((($wnum)) + 8|0);
     HEAP32[$40>>2] = $39;
     $41 = (($30) + -1)|0;
     $42 = HEAP32[$13>>2]|0;
     $43 = (($42) + ($41<<2)|0);
     $44 = HEAP32[$43>>2]|0;
     $45 = ($30|0)==(1);
     if ($45) {
      $97 = 0;$98 = 0;
     } else {
      $46 = (($30) + -2)|0;
      $47 = (($42) + ($46<<2)|0);
      $48 = HEAP32[$47>>2]|0;
      $97 = $48;$98 = 0;
     }
     $49 = (($32) + -1)|0;
     $50 = HEAP32[$12>>2]|0;
     $51 = (($50) + ($49<<2)|0);
     $52 = ((($num)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($divisor)) + 12|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = $55 ^ $53;
     $57 = ((($res$0)) + 12|0);
     HEAP32[$57>>2] = $56;
     $58 = ((($res$0)) + 8|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ($33|0)<($59|0);
     if (!($60)) {
      $61 = (($33) + 1)|0;
      $62 = (_egbn_expand2($res$0,$61)|0);
      $63 = ($62|0)==(0|0);
      if ($63) {
       break;
      }
     }
     $64 = ((($res$0)) + 4|0);
     HEAP32[$64>>2] = $33;
     $65 = (($33) + -1)|0;
     $66 = HEAP32[$res$0>>2]|0;
     $67 = (($66) + ($65<<2)|0);
     $68 = (($30) + 1)|0;
     $69 = ((($11)) + 8|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = ($30|0)<($70|0);
     if (!($71)) {
      $72 = (_egbn_expand2($11,$68)|0);
      $73 = ($72|0)==(0|0);
      if ($73) {
       break;
      }
     }
     $74 = (_EGBN_ucmp($wnum,$13)|0);
     $75 = ($74|0)>(-1);
     if ($75) {
      $76 = (_EGBN_usub($wnum,$wnum,$13)|0);
      $77 = ($76|0)==(0);
      if ($77) {
       break;
      }
      HEAP32[$67>>2] = 1;
      $78 = HEAP32[$64>>2]|0;
      $79 = (($78) + -1)|0;
      $80 = HEAP32[$res$0>>2]|0;
      $81 = (($80) + ($79<<2)|0);
      HEAP32[$81>>2] = 1;
      $85 = $78;
     } else {
      $82 = HEAP32[$64>>2]|0;
      $83 = (($82) + -1)|0;
      HEAP32[$64>>2] = $83;
      $85 = $83;
     }
     $84 = ($85|0)==(0);
     if ($84) {
      HEAP32[$57>>2] = 0;
     }
     $86 = ($33|0)>(1);
     if ($86) {
      $$sum = (($33) + -2)|0;
      $87 = (($66) + ($$sum<<2)|0);
      $88 = ((($11)) + 4|0);
      $89 = ((($wnum)) + 12|0);
      $i$011 = 0;$resp$010 = $87;$wnump$09 = $51;
      while(1) {
       $90 = HEAP32[$wnump$09>>2]|0;
       $91 = ((($wnump$09)) + -4|0);
       $92 = HEAP32[$91>>2]|0;
       $93 = ($90|0)==($44|0);
       L46: do {
        if ($93) {
         $q$1 = -1;
        } else {
         $94 = (_egbn_div_words($90,$92,$44)|0);
         $95 = Math_imul($94, $44)|0;
         $96 = (($92) - ($95))|0;
         $99 = (___muldi3(($94|0),0,($97|0),($98|0))|0);
         $100 = tempRet0;
         $101 = ((($wnump$09)) + -8|0);
         $102 = HEAP32[$101>>2]|0;
         $103 = ($100>>>0)>($96>>>0);
         $104 = ($99>>>0)>($102>>>0);
         $105 = ($100|0)==($96|0);
         $106 = $105 & $104;
         $107 = $103 | $106;
         if ($107) {
          $118 = $99;$119 = $100;$q$04 = $94;$rem$05 = $96;
          while(1) {
           $116 = (($q$04) + -1)|0;
           $110 = (($rem$05) + ($44))|0;
           $117 = ($110>>>0)<($44>>>0);
           $112 = (_i64Subtract(($118|0),($119|0),($97|0),($98|0))|0);
           $109 = tempRet0;
           if ($117) {
            $q$1 = $116;
            break L46;
           }
           $108 = ($109>>>0)>($110>>>0);
           $111 = ($112>>>0)>($102>>>0);
           $113 = ($109|0)==($110|0);
           $114 = $113 & $111;
           $115 = $108 | $114;
           if ($115) {
            $118 = $112;$119 = $109;$q$04 = $116;$rem$05 = $110;
           } else {
            $q$1 = $116;
            break;
           }
          }
         } else {
          $q$1 = $94;
         }
        }
       } while(0);
       $120 = HEAP32[$11>>2]|0;
       $121 = HEAP32[$13>>2]|0;
       $122 = (_egbn_mul_words($120,$121,$30,$q$1)|0);
       $123 = HEAP32[$wnum>>2]|0;
       $124 = ((($123)) + -4|0);
       HEAP32[$wnum>>2] = $124;
       $125 = HEAP32[$36>>2]|0;
       $126 = (($125) + 1)|0;
       HEAP32[$36>>2] = $126;
       $127 = HEAP32[$11>>2]|0;
       $128 = (($127) + ($30<<2)|0);
       HEAP32[$128>>2] = $122;
       $j$0 = $68;
       while(1) {
        $129 = ($j$0|0)>(0);
        if (!($129)) {
         $j$0$lcssa = $j$0;
         break;
        }
        $130 = (($j$0) + -1)|0;
        $131 = (($127) + ($130<<2)|0);
        $132 = HEAP32[$131>>2]|0;
        $133 = ($132|0)==(0);
        if ($133) {
         $j$0 = $130;
        } else {
         $j$0$lcssa = $j$0;
         break;
        }
       }
       HEAP32[$88>>2] = $j$0$lcssa;
       $134 = HEAP32[$36>>2]|0;
       $135 = (_EGBN_sub($wnum,$wnum,$11)|0);
       $136 = ($135|0)==(0);
       if ($136) {
        break L21;
       }
       $137 = HEAP32[$31>>2]|0;
       $138 = HEAP32[$36>>2]|0;
       $139 = (($137) - ($134))|0;
       $140 = (($139) + ($138))|0;
       HEAP32[$31>>2] = $140;
       $141 = HEAP32[$89>>2]|0;
       $142 = ($141|0)==(0);
       if ($142) {
        $q$2 = $q$1;
       } else {
        $143 = HEAP32[$36>>2]|0;
        $144 = (_EGBN_add($wnum,$wnum,$13)|0);
        $145 = ($144|0)==(0);
        if ($145) {
         break L21;
        }
        $146 = (($q$1) + -1)|0;
        $147 = HEAP32[$36>>2]|0;
        $148 = (($147) - ($143))|0;
        $149 = HEAP32[$31>>2]|0;
        $150 = (($148) + ($149))|0;
        HEAP32[$31>>2] = $150;
        $q$2 = $146;
       }
       $151 = ((($resp$010)) + -4|0);
       HEAP32[$resp$010>>2] = $q$2;
       $152 = (($i$011) + 1)|0;
       $153 = ($152|0)<($65|0);
       if ($153) {
        $i$011 = $152;$resp$010 = $151;$wnump$09 = $91;
       } else {
        break;
       }
      }
     }
     $154 = ($rm|0)==(0|0);
     L62: do {
      if (!($154)) {
       $155 = HEAP32[$52>>2]|0;
       (_EGBN_rshift($rm,$12,$25)|0);
       $156 = ((($rm)) + 4|0);
       $157 = HEAP32[$156>>2]|0;
       switch ($157|0) {
       case 0:  {
        break L62;
        break;
       }
       case 1:  {
        $158 = HEAP32[$rm>>2]|0;
        $159 = HEAP32[$158>>2]|0;
        $160 = ($159|0)==(0);
        if ($160) {
         break L62;
        }
        break;
       }
       default: {
       }
       }
       $161 = ((($rm)) + 12|0);
       HEAP32[$161>>2] = $155;
      }
     } while(0);
     _EGBN_CTX_end($ctx);
     $$0 = 1;
     STACKTOP = sp;return ($$0|0);
    }
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_rshift($r,$a,$n) {
 $r = $r|0;
 $a = $a|0;
 $n = $n|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa22 = 0, $$pre = 0, $$sum = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond = 0, $f$06 = 0, $f$110 = 0, $ftl$02 = 0, $i$04 = 0, $i$18 = 0, $or$cond = 0, $scevgep = 0, $scevgep17 = 0, $smax = 0, $t$05 = 0, $t$1$lcssa = 0;
 var $t$19 = 0, $t$2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($n|0) / 32)&-1;
 $1 = (($n|0) % 32)&-1;
 $2 = (32 - ($1))|0;
 $3 = ((($a)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)<($0|0);
 $6 = ($4|0)==(0);
 $or$cond = $5 | $6;
 if ($or$cond) {
  (_EGBN_set_word($r,0)|0);
  $$0 = 1;
  return ($$0|0);
 }
 $7 = ($r|0)==($a|0);
 do {
  if ($7) {
   $18 = ($n|0)==(0);
   if ($18) {
    $$0 = 1;
    return ($$0|0);
   } else {
    $23 = $4;
   }
  } else {
   $8 = ((($a)) + 12|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ((($r)) + 12|0);
   HEAP32[$10>>2] = $9;
   $11 = (($4) - ($0))|0;
   $12 = ((($r)) + 8|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($11|0)<($13|0);
   if ($14) {
    $23 = $4;
   } else {
    $15 = (($11) + 1)|0;
    $16 = (_egbn_expand2($r,$15)|0);
    $17 = ($16|0)==(0|0);
    if ($17) {
     $$0 = 0;
     return ($$0|0);
    } else {
     $$pre = HEAP32[$3>>2]|0;
     $23 = $$pre;
     break;
    }
   }
  }
 } while(0);
 $19 = HEAP32[$a>>2]|0;
 $20 = (($19) + ($0<<2)|0);
 $21 = HEAP32[$r>>2]|0;
 $22 = (($23) - ($0))|0;
 $24 = ((($r)) + 4|0);
 HEAP32[$24>>2] = $22;
 $25 = ($1|0)==(0);
 if ($25) {
  $26 = ($22|0)>(-1);
  if (!($26)) {
   HEAP32[$21>>2] = 0;
   $$0 = 1;
   return ($$0|0);
  }
  $27 = (($22) + 1)|0;
  $28 = (($0) + -2)|0;
  $29 = (($28) - ($23))|0;
  $30 = ($29|0)>(-2);
  $smax = $30 ? $29 : -2;
  $31 = (($23) + ($smax))|0;
  $32 = (($31) + 3)|0;
  $33 = (($32) - ($0))|0;
  $f$06 = $20;$i$04 = $27;$t$05 = $21;
  while(1) {
   $34 = ((($f$06)) + 4|0);
   $35 = HEAP32[$f$06>>2]|0;
   $36 = ((($t$05)) + 4|0);
   HEAP32[$t$05>>2] = $35;
   $37 = (($i$04) + -1)|0;
   $38 = ($i$04|0)>(1);
   if ($38) {
    $f$06 = $34;$i$04 = $37;$t$05 = $36;
   } else {
    break;
   }
  }
  $scevgep = (($21) + ($33<<2)|0);
  $t$2 = $scevgep;
 } else {
  $39 = HEAP32[$20>>2]|0;
  $40 = ($22|0)>(1);
  $41 = $39 >>> $1;
  if ($40) {
   $$sum = (($0) + 1)|0;
   $42 = (($19) + ($$sum<<2)|0);
   $43 = (($23) + -1)|0;
   $44 = (($43) - ($0))|0;
   $49 = $41;$f$110 = $42;$i$18 = 1;$t$19 = $21;
   while(1) {
    $45 = ((($f$110)) + 4|0);
    $46 = HEAP32[$f$110>>2]|0;
    $47 = $46 << $2;
    $48 = $47 | $49;
    $50 = ((($t$19)) + 4|0);
    HEAP32[$t$19>>2] = $48;
    $51 = (($i$18) + 1)|0;
    $52 = $46 >>> $1;
    $exitcond = ($51|0)==($22|0);
    if ($exitcond) {
     $$lcssa22 = $52;
     break;
    } else {
     $49 = $52;$f$110 = $45;$i$18 = $51;$t$19 = $50;
    }
   }
   $scevgep17 = (($21) + ($44<<2)|0);
   $$lcssa = $$lcssa22;$t$1$lcssa = $scevgep17;
  } else {
   $$lcssa = $41;$t$1$lcssa = $21;
  }
  $53 = ((($t$1$lcssa)) + 4|0);
  HEAP32[$t$1$lcssa>>2] = $$lcssa;
  $t$2 = $53;
 }
 HEAP32[$t$2>>2] = 0;
 $54 = ($22|0)>(0);
 if (!($54)) {
  $$0 = 1;
  return ($$0|0);
 }
 $55 = (($22) + -1)|0;
 $56 = (($21) + ($55<<2)|0);
 $61 = $22;$ftl$02 = $56;
 while(1) {
  $57 = HEAP32[$ftl$02>>2]|0;
  $58 = ($57|0)==(0);
  if (!($58)) {
   $$0 = 1;
   label = 23;
   break;
  }
  $59 = ((($ftl$02)) + -4|0);
  $60 = (($61) + -1)|0;
  HEAP32[$24>>2] = $60;
  $62 = ($61|0)>(1);
  if ($62) {
   $61 = $60;$ftl$02 = $59;
  } else {
   $$0 = 1;
   label = 23;
   break;
  }
 }
 if ((label|0) == 23) {
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_usub($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $$pn = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $ap$013 = 0, $ap$1 = 0, $ap$2 = 0, $ap$36 = 0, $bp$014 = 0, $carry$017 = 0, $carry$1 = 0, $carry$1$v = 0, $exitcond = 0, $ftl$04 = 0, $i$016 = 0, $i$1 = 0, $i$2 = 0, $i$38 = 0;
 var $not$ = 0, $or$cond = 0, $rp$015 = 0, $rp$1 = 0, $rp$2 = 0, $rp$37 = 0, $scevgep = 0, $scevgep24 = 0, $t1$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)<($3|0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = ((($r)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($1|0)>($6|0);
 if ($7) {
  $8 = (_egbn_expand2($r,$1)|0);
  $9 = ($8|0)==(0|0);
  if ($9) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $10 = HEAP32[$a>>2]|0;
 $11 = HEAP32[$r>>2]|0;
 $12 = ($3|0)>(0);
 L8: do {
  if ($12) {
   $13 = HEAP32[$b>>2]|0;
   $scevgep = (($11) + ($3<<2)|0);
   $ap$013 = $10;$bp$014 = $13;$carry$017 = 0;$i$016 = 0;$rp$015 = $11;
   while(1) {
    $14 = ((($ap$013)) + 4|0);
    $15 = HEAP32[$ap$013>>2]|0;
    $16 = ((($bp$014)) + 4|0);
    $17 = HEAP32[$bp$014>>2]|0;
    $18 = ($carry$017|0)==(0);
    $19 = ($15>>>0)<=($17>>>0);
    $20 = ($15>>>0)<($17>>>0);
    $not$ = $18 ^ 1;
    $21 = $not$ << 31 >> 31;
    $$pn = (($15) + ($21))|0;
    $carry$1$v = $18 ? $20 : $19;
    $carry$1 = $carry$1$v&1;
    $t1$0 = (($$pn) - ($17))|0;
    $22 = ((($rp$015)) + 4|0);
    HEAP32[$rp$015>>2] = $t1$0;
    $23 = (($i$016) + 1)|0;
    $exitcond = ($23|0)==($3|0);
    if ($exitcond) {
     break;
    } else {
     $ap$013 = $14;$bp$014 = $16;$carry$017 = $carry$1;$i$016 = $23;$rp$015 = $22;
    }
   }
   $scevgep24 = (($10) + ($3<<2)|0);
   if ($carry$1$v) {
    $ap$1 = $scevgep24;$i$1 = $3;$rp$1 = $scevgep;
    while(1) {
     $24 = ($i$1|0)<($1|0);
     if (!($24)) {
      $ap$2 = $ap$1;$i$2 = $i$1;$rp$2 = $rp$1;
      break L8;
     }
     $25 = (($i$1) + 1)|0;
     $26 = ((($ap$1)) + 4|0);
     $27 = HEAP32[$ap$1>>2]|0;
     $28 = (($27) + -1)|0;
     $29 = ((($rp$1)) + 4|0);
     HEAP32[$rp$1>>2] = $28;
     $30 = ($27|0)==(0);
     if ($30) {
      $ap$1 = $26;$i$1 = $25;$rp$1 = $29;
     } else {
      $ap$2 = $26;$i$2 = $25;$rp$2 = $29;
      break;
     }
    }
   } else {
    $ap$2 = $scevgep24;$i$2 = $3;$rp$2 = $scevgep;
   }
  } else {
   $ap$2 = $10;$i$2 = 0;$rp$2 = $11;
  }
 } while(0);
 $31 = ($rp$2|0)!=($ap$2|0);
 $32 = ($i$2|0)<($1|0);
 $or$cond = $31 & $32;
 L17: do {
  if ($or$cond) {
   $ap$36 = $ap$2;$i$38 = $i$2;$rp$37 = $rp$2;
   while(1) {
    $33 = (($i$38) + 1)|0;
    $34 = HEAP32[$ap$36>>2]|0;
    HEAP32[$rp$37>>2] = $34;
    $35 = ($33|0)<($1|0);
    if (!($35)) {
     break L17;
    }
    $36 = (($i$38) + 2)|0;
    $37 = ((($ap$36)) + 4|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ((($rp$37)) + 4|0);
    HEAP32[$39>>2] = $38;
    $40 = ($36|0)<($1|0);
    if (!($40)) {
     break L17;
    }
    $41 = (($i$38) + 3)|0;
    $42 = ((($ap$36)) + 8|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = ((($rp$37)) + 8|0);
    HEAP32[$44>>2] = $43;
    $45 = ($41|0)<($1|0);
    if (!($45)) {
     break L17;
    }
    $46 = (($i$38) + 4)|0;
    $47 = ((($ap$36)) + 12|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = ((($rp$37)) + 12|0);
    HEAP32[$49>>2] = $48;
    $50 = ((($rp$37)) + 16|0);
    $51 = ((($ap$36)) + 16|0);
    $52 = ($46|0)<($1|0);
    if ($52) {
     $ap$36 = $51;$i$38 = $46;$rp$37 = $50;
    } else {
     break;
    }
   }
  }
 } while(0);
 $53 = ((($r)) + 4|0);
 HEAP32[$53>>2] = $1;
 $54 = ((($r)) + 12|0);
 HEAP32[$54>>2] = 0;
 $55 = ($1|0)>(0);
 if (!($55)) {
  $$0 = 1;
  return ($$0|0);
 }
 $56 = (($1) + -1)|0;
 $57 = (($11) + ($56<<2)|0);
 $62 = $1;$ftl$04 = $57;
 while(1) {
  $58 = HEAP32[$ftl$04>>2]|0;
  $59 = ($58|0)==(0);
  if (!($59)) {
   $$0 = 1;
   label = 19;
   break;
  }
  $60 = ((($ftl$04)) + -4|0);
  $61 = (($62) + -1)|0;
  HEAP32[$53>>2] = $61;
  $63 = ($62|0)>(1);
  if ($63) {
   $62 = $61;$ftl$04 = $60;
  } else {
   $$0 = 1;
   label = 19;
   break;
  }
 }
 if ((label|0) == 19) {
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_add_word($a,$w) {
 $a = $a|0;
 $w = $w|0;
 var $$0 = 0, $$0$i = 0, $$0$i$lcssa = 0, $$01 = 0, $$01$i = 0, $$01$i10 = 0, $$01$i9 = 0, $$lcssa15 = 0, $$pr2$pre = 0, $$pre = 0, $$pre6 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i$0$i = 0, $i$0$i$lcssa = 0;
 var $l$0 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $27 = ((($a)) + 4|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = ((($a)) + 8|0);
  $30 = HEAP32[$29>>2]|0;
  $31 = ($28|0)<($30|0);
  do {
   if ($31) {
    $37 = $28;
   } else {
    $32 = (($28) + 1)|0;
    $33 = (_egbn_expand2($a,$32)|0);
    $34 = ($33|0)==(0|0);
    if ($34) {
     $$01 = 0;
     return ($$01|0);
    } else {
     $$pre6 = HEAP32[$27>>2]|0;
     $37 = $$pre6;
     break;
    }
   }
  } while(0);
  $35 = HEAP32[$a>>2]|0;
  $$0 = $w;$i$0 = 0;
  while(1) {
   $36 = ($i$0|0)<($37|0);
   $38 = (($35) + ($i$0<<2)|0);
   if ($36) {
    $39 = HEAP32[$38>>2]|0;
    $40 = (($39) + ($$0))|0;
    $l$0 = $40;
   } else {
    $l$0 = $$0;
   }
   HEAP32[$38>>2] = $l$0;
   $41 = ($$0>>>0)>($l$0>>>0);
   $42 = (($i$0) + 1)|0;
   if ($41) {
    $$0 = 1;$i$0 = $42;
   } else {
    break;
   }
  }
  if ($36) {
   $$01 = 1;
   return ($$01|0);
  }
  $43 = (($37) + 1)|0;
  HEAP32[$27>>2] = $43;
  $$01 = 1;
  return ($$01|0);
 }
 HEAP32[$0>>2] = 0;
 $3 = ((($a)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 switch ($4|0) {
 case 0:  {
  label = 6;
  break;
 }
 case 1:  {
  $5 = HEAP32[$a>>2]|0;
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   label = 6;
  } else {
   $9 = ($6>>>0)<($w>>>0);
   if ($9) {
    $10 = (($w) - ($6))|0;
    HEAP32[$5>>2] = $10;
    HEAP32[$0>>2] = 1;
    $$01$i9 = 1;$45 = 1;
    label = 13;
   } else {
    $12 = $5;
    label = 4;
   }
  }
  break;
 }
 default: {
  $$pre = HEAP32[$a>>2]|0;
  $12 = $$pre;
  label = 4;
 }
 }
 if ((label|0) == 4) {
  $$0$i = $w;$i$0$i = 0;
  while(1) {
   $11 = (($12) + ($i$0$i<<2)|0);
   $13 = HEAP32[$11>>2]|0;
   $14 = ($13>>>0)<($$0$i>>>0);
   $15 = (($13) - ($$0$i))|0;
   HEAP32[$11>>2] = $15;
   $16 = (($i$0$i) + 1)|0;
   if ($14) {
    $$0$i = 1;$i$0$i = $16;
   } else {
    $$0$i$lcssa = $$0$i;$$lcssa15 = $13;$i$0$i$lcssa = $i$0$i;
    break;
   }
  }
  $17 = ($$lcssa15|0)==($$0$i$lcssa|0);
  $18 = (($4) + -1)|0;
  $19 = ($i$0$i$lcssa|0)==($18|0);
  $or$cond = $17 & $19;
  if ($or$cond) {
   HEAP32[$3>>2] = $i$0$i$lcssa;
   $$01$i = 1;$20 = $i$0$i$lcssa;$44 = 0;
   label = 12;
  } else {
   $$01$i = 1;$20 = $4;$44 = 0;
   label = 12;
  }
 }
 else if ((label|0) == 6) {
  HEAP32[$0>>2] = 0;
  $8 = (_EGBN_add_word($a,$w)|0);
  HEAP32[$0>>2] = 1;
  $$pr2$pre = HEAP32[$3>>2]|0;
  $$01$i = $8;$20 = $$pr2$pre;$44 = 1;
  label = 12;
 }
 L32: do {
  if ((label|0) == 12) {
   switch ($20|0) {
   case 1:  {
    $$01$i9 = $$01$i;$45 = $44;
    label = 13;
    break L32;
    break;
   }
   case 0:  {
    $$01 = $$01$i;
    break;
   }
   default: {
    $$01$i10 = $$01$i;$25 = $44;
    break L32;
   }
   }
   return ($$01|0);
  }
 } while(0);
 if ((label|0) == 13) {
  $21 = HEAP32[$a>>2]|0;
  $22 = HEAP32[$21>>2]|0;
  $23 = ($22|0)==(0);
  if ($23) {
   $$01 = $$01$i9;
   return ($$01|0);
  } else {
   $$01$i10 = $$01$i9;$25 = $45;
  }
 }
 $24 = ($25|0)==(0);
 $26 = $24&1;
 HEAP32[$0>>2] = $26;
 $$01 = $$01$i10;
 return ($$01|0);
}
function _egbn_expand2($b,$words) {
 $b = $b|0;
 $words = $words|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($b)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<($words|0);
 if (!($2)) {
  $$0 = $b;
  return ($$0|0);
 }
 $3 = (_bn_expand_internal($b,$words)|0);
 $4 = ($3|0)==(0|0);
 if ($4) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = HEAP32[$b>>2]|0;
 $6 = ($5|0)==(0|0);
 if (!($6)) {
  _free($5);
 }
 HEAP32[$b>>2] = $3;
 HEAP32[$0>>2] = $words;
 $$0 = $b;
 return ($$0|0);
}
function _bn_expand_internal($b,$words) {
 $b = $b|0;
 $words = $words|0;
 var $$0 = 0, $$not = 0, $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $A$0$lcssa = 0, $A$08 = 0, $A$1$lcssa = 0, $A$14 = 0, $B$0$lcssa = 0, $B$07 = 0, $i$06 = 0, $i$13 = 0, $scevgep = 0, $scevgep20 = 0, $scevgep21 = 0, $smax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($words|0)>(16777215);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ((($b)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 2;
 $4 = ($3|0)==(0);
 if (!($4)) {
  $$0 = 0;
  return ($$0|0);
 }
 $5 = $words << 2;
 $6 = (($5) + 4)|0;
 $7 = (_malloc($6)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0 = 0;
  return ($$0|0);
 }
 $9 = HEAP32[$b>>2]|0;
 $10 = ($9|0)==(0|0);
 $$phi$trans$insert = ((($b)) + 4|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 L10: do {
  if (!($10)) {
   $11 = $$pre >> 2;
   $12 = ($11|0)>(0);
   if ($12) {
    $13 = (($11) + -2)|0;
    $14 = $13 << 2;
    $15 = (($14) + 8)|0;
    $16 = $13 << 4;
    $17 = (($16) + 32)|0;
    $scevgep21 = (($7) + ($17)|0);
    $A$08 = $7;$B$07 = $9;$i$06 = $11;
    while(1) {
     $18 = HEAP32[$B$07>>2]|0;
     $19 = ((($B$07)) + 4|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = ((($B$07)) + 8|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ((($B$07)) + 12|0);
     $24 = HEAP32[$23>>2]|0;
     HEAP32[$A$08>>2] = $18;
     $25 = ((($A$08)) + 4|0);
     HEAP32[$25>>2] = $20;
     $26 = ((($A$08)) + 8|0);
     HEAP32[$26>>2] = $22;
     $27 = ((($A$08)) + 12|0);
     HEAP32[$27>>2] = $24;
     $28 = (($i$06) + -1)|0;
     $29 = ((($A$08)) + 16|0);
     $30 = ((($B$07)) + 16|0);
     $31 = ($i$06|0)>(1);
     if ($31) {
      $A$08 = $29;$B$07 = $30;$i$06 = $28;
     } else {
      break;
     }
    }
    $scevgep20 = (($9) + ($15<<2)|0);
    $A$0$lcssa = $scevgep21;$B$0$lcssa = $scevgep20;
   } else {
    $A$0$lcssa = $7;$B$0$lcssa = $9;
   }
   $32 = $$pre & 3;
   switch ($32|0) {
   case 3:  {
    $33 = ((($B$0$lcssa)) + 8|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($A$0$lcssa)) + 8|0);
    HEAP32[$35>>2] = $34;
    label = 11;
    break;
   }
   case 2:  {
    label = 11;
    break;
   }
   case 1:  {
    break;
   }
   default: {
    break L10;
   }
   }
   if ((label|0) == 11) {
    $36 = ((($B$0$lcssa)) + 4|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = ((($A$0$lcssa)) + 4|0);
    HEAP32[$38>>2] = $37;
   }
   $39 = HEAP32[$B$0$lcssa>>2]|0;
   HEAP32[$A$0$lcssa>>2] = $39;
  }
 } while(0);
 $40 = (($7) + ($$pre<<2)|0);
 $41 = (($words) - ($$pre))|0;
 $42 = $41 >> 3;
 $43 = ($42|0)>(0);
 if ($43) {
  $44 = $42 << 5;
  $45 = $$pre << 2;
  $46 = (($44) + ($45))|0;
  $scevgep = (($7) + ($46)|0);
  $A$14 = $40;$i$13 = $42;
  while(1) {
   $47 = (($i$13) + -1)|0;
   $48 = ((($A$14)) + 32|0);
   $49 = ($i$13|0)>(1);
   ;HEAP32[$A$14>>2]=0|0;HEAP32[$A$14+4>>2]=0|0;HEAP32[$A$14+8>>2]=0|0;HEAP32[$A$14+12>>2]=0|0;HEAP32[$A$14+16>>2]=0|0;HEAP32[$A$14+20>>2]=0|0;HEAP32[$A$14+24>>2]=0|0;HEAP32[$A$14+28>>2]=0|0;
   if ($49) {
    $A$14 = $48;$i$13 = $47;
   } else {
    break;
   }
  }
  $A$1$lcssa = $scevgep;
 } else {
  $A$1$lcssa = $40;
 }
 $50 = $41 & 7;
 $51 = ($50|0)==(0);
 if ($51) {
  $$0 = $7;
  return ($$0|0);
 }
 $52 = $41 & 7;
 $$not = $41 | -8;
 $53 = $$not ^ 7;
 $54 = ($$not|0)==(-8);
 $smax = $54 ? $53 : 1073741822;
 $55 = (($smax) + ($52))|0;
 $56 = $55 << 2;
 $57 = (($56) + 8)|0;
 _memset(($A$1$lcssa|0),0,($57|0))|0;
 $$0 = $7;
 return ($$0|0);
}
function _EGBN_lshift($r,$a,$n) {
 $r = $r|0;
 $a = $a|0;
 $n = $n|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $ftl$04 = 0, $i$06 = 0, $i$18 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($r)) + 12|0);
 HEAP32[$2>>2] = $1;
 $3 = (($n|0) / 32)&-1;
 $4 = ((($a)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($5) + ($3))|0;
 $7 = ((($r)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6|0)<($8|0);
 do {
  if ($9) {
   $18 = $5;
  } else {
   $10 = (($6) + 1)|0;
   $11 = (_egbn_expand2($r,$10)|0);
   $12 = ($11|0)==(0|0);
   if ($12) {
    $$0 = 0;
    return ($$0|0);
   } else {
    $$pre = HEAP32[$4>>2]|0;
    $18 = $$pre;
    break;
   }
  }
 } while(0);
 $13 = (($n|0) % 32)&-1;
 $14 = (32 - ($13))|0;
 $15 = HEAP32[$a>>2]|0;
 $16 = HEAP32[$r>>2]|0;
 $17 = (($18) + ($3))|0;
 $19 = (($16) + ($17<<2)|0);
 HEAP32[$19>>2] = 0;
 $20 = ($13|0)==(0);
 $21 = (($18) + -1)|0;
 $22 = ($18|0)>(0);
 if ($20) {
  if ($22) {
   $i$06 = $21;
   while(1) {
    $23 = (($15) + ($i$06<<2)|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = (($i$06) + ($3))|0;
    $26 = (($16) + ($25<<2)|0);
    HEAP32[$26>>2] = $24;
    $27 = (($i$06) + -1)|0;
    $28 = ($i$06|0)>(0);
    if ($28) {
     $i$06 = $27;
    } else {
     break;
    }
   }
  }
 } else {
  if ($22) {
   $i$18 = $21;
   while(1) {
    $29 = (($15) + ($i$18<<2)|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = $30 >>> $14;
    $32 = (($i$18) + ($3))|0;
    $33 = (($32) + 1)|0;
    $34 = (($16) + ($33<<2)|0);
    $35 = HEAP32[$34>>2]|0;
    $36 = $35 | $31;
    HEAP32[$34>>2] = $36;
    $37 = $30 << $13;
    $38 = (($16) + ($32<<2)|0);
    HEAP32[$38>>2] = $37;
    $39 = (($i$18) + -1)|0;
    $40 = ($i$18|0)>(0);
    if ($40) {
     $i$18 = $39;
    } else {
     break;
    }
   }
  }
 }
 $41 = $3 << 2;
 _memset(($16|0),0,($41|0))|0;
 $42 = HEAP32[$4>>2]|0;
 $43 = (($42) + ($3))|0;
 $44 = (($43) + 1)|0;
 $45 = ((($r)) + 4|0);
 HEAP32[$45>>2] = $44;
 $46 = ($43|0)>(-1);
 if (!($46)) {
  $$0 = 1;
  return ($$0|0);
 }
 $47 = HEAP32[$r>>2]|0;
 $48 = (($47) + ($43<<2)|0);
 $53 = $44;$ftl$04 = $48;
 while(1) {
  $49 = HEAP32[$ftl$04>>2]|0;
  $50 = ($49|0)==(0);
  if (!($50)) {
   $$0 = 1;
   label = 13;
   break;
  }
  $51 = ((($ftl$04)) + -4|0);
  $52 = (($53) + -1)|0;
  HEAP32[$45>>2] = $52;
  $54 = ($53|0)>(1);
  if ($54) {
   $53 = $52;$ftl$04 = $51;
  } else {
   $$0 = 1;
   label = 13;
   break;
  }
 }
 if ((label|0) == 13) {
  return ($$0|0);
 }
 return (0)|0;
}
function _egbn_div_words($h,$l,$d) {
 $h = $h|0;
 $l = $l|0;
 $d = $d|0;
 var $$ = 0, $$0 = 0, $$02 = 0, $$03 = 0, $$1 = 0, $$14 = 0, $$2 = 0, $$3 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $count$0 = 0, $h$ = 0, $q$0 = 0, $q$1$lcssa = 0;
 var $q$15 = 0, $q$2 = 0, $q$2$lcssa = 0, $ret$0 = 0, $ret$0$lcssa = 0, $th$0$lcssa = 0, $th$06 = 0, $tl$0$lcssa = 0, $tl$07 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($d|0)==(0);
 if ($0) {
  $$0 = -1;
  return ($$0|0);
 }
 $1 = (_EGBN_num_bits_word($d)|0);
 $2 = (32 - ($1))|0;
 $3 = ($h>>>0)<($d>>>0);
 $4 = $3 ? 0 : $d;
 $h$ = (($h) - ($4))|0;
 $5 = ($2|0)==(0);
 if ($5) {
  $$02 = $d;$$03 = $l;$$1 = $h$;
 } else {
  $6 = $d << $2;
  $7 = $h$ << $2;
  $8 = $l >>> $1;
  $9 = $7 | $8;
  $10 = $l << $2;
  $$02 = $6;$$03 = $10;$$1 = $9;
 }
 $11 = $$02 >>> 16;
 $12 = $$02 & 65535;
 $$14 = $$03;$$2 = $$1;$count$0 = 2;$ret$0 = 0;
 while(1) {
  $13 = $$2 >>> 16;
  $14 = ($13|0)==($11|0);
  if ($14) {
   $q$0 = 65535;
  } else {
   $15 = (($$2>>>0) / ($11>>>0))&-1;
   $q$0 = $15;
  }
  $16 = Math_imul($q$0, $11)|0;
  $17 = Math_imul($q$0, $12)|0;
  $18 = (($$2) - ($16))|0;
  $19 = ($18>>>0)>(65535);
  L12: do {
   if ($19) {
    $q$1$lcssa = $q$0;$th$0$lcssa = $16;$tl$0$lcssa = $17;
   } else {
    $20 = $$14 >>> 16;
    $22 = $18;$q$15 = $q$0;$th$06 = $16;$tl$07 = $17;
    while(1) {
     $21 = $22 << 16;
     $23 = $21 | $20;
     $24 = ($tl$07>>>0)>($23>>>0);
     if (!($24)) {
      $q$1$lcssa = $q$15;$th$0$lcssa = $th$06;$tl$0$lcssa = $tl$07;
      break L12;
     }
     $25 = (($q$15) + -1)|0;
     $26 = (($th$06) - ($11))|0;
     $27 = (($tl$07) - ($12))|0;
     $28 = (($$2) - ($26))|0;
     $29 = ($28>>>0)>(65535);
     if ($29) {
      $q$1$lcssa = $25;$th$0$lcssa = $26;$tl$0$lcssa = $27;
      break;
     } else {
      $22 = $28;$q$15 = $25;$th$06 = $26;$tl$07 = $27;
     }
    }
   }
  } while(0);
  $30 = $tl$0$lcssa >>> 16;
  $31 = $tl$0$lcssa << 16;
  $32 = (($30) + ($th$0$lcssa))|0;
  $33 = ($$14>>>0)<($31>>>0);
  $34 = $33&1;
  $$ = (($32) + ($34))|0;
  $35 = (($$14) - ($31))|0;
  $36 = ($$2>>>0)<($$>>>0);
  $37 = $36 << 31 >> 31;
  $q$2 = (($37) + ($q$1$lcssa))|0;
  $38 = (($count$0) + -1)|0;
  $39 = ($38|0)==(0);
  if ($39) {
   $q$2$lcssa = $q$2;$ret$0$lcssa = $ret$0;
   break;
  }
  $40 = $36 ? $$02 : 0;
  $$3 = (($$2) - ($$))|0;
  $41 = (($$3) + ($40))|0;
  $42 = $q$2 << 16;
  $43 = $41 << 16;
  $44 = $35 >>> 16;
  $45 = $43 | $44;
  $46 = $35 << 16;
  $$14 = $46;$$2 = $45;$count$0 = $38;$ret$0 = $42;
 }
 $47 = $q$2$lcssa | $ret$0$lcssa;
 $$0 = $47;
 return ($$0|0);
}
function _egbn_mul_words($rp,$ap,$num,$w) {
 $rp = $rp|0;
 $ap = $ap|0;
 $num = $num|0;
 $w = $w|0;
 var $$0 = 0, $$01$lcssa15 = 0, $$017 = 0, $$02$lcssa16 = 0, $$026 = 0, $$03$lcssa17 = 0, $$035 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c1$0$lcssa18 = 0, $c1$04 = 0, $scevgep = 0, $scevgep14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($num|0)<(1);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ($num>>>0)>(3);
 if ($1) {
  $2 = (($num) + -4)|0;
  $3 = $2 & -4;
  $4 = (($3) + 4)|0;
  $scevgep = (($rp) + ($4<<2)|0);
  $$017 = $rp;$$026 = $ap;$$035 = $num;$c1$04 = 0;
  while(1) {
   $5 = HEAP32[$$026>>2]|0;
   $6 = (___muldi3(($5|0),0,($w|0),0)|0);
   $7 = tempRet0;
   $8 = (_i64Add(($6|0),($7|0),($c1$04|0),0)|0);
   $9 = tempRet0;
   HEAP32[$$017>>2] = $8;
   $10 = ((($$026)) + 4|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (___muldi3(($11|0),0,($w|0),0)|0);
   $13 = tempRet0;
   $14 = (_i64Add(($9|0),0,($12|0),($13|0))|0);
   $15 = tempRet0;
   $16 = ((($$017)) + 4|0);
   HEAP32[$16>>2] = $14;
   $17 = ((($$026)) + 8|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (___muldi3(($18|0),0,($w|0),0)|0);
   $20 = tempRet0;
   $21 = (_i64Add(($15|0),0,($19|0),($20|0))|0);
   $22 = tempRet0;
   $23 = ((($$017)) + 8|0);
   HEAP32[$23>>2] = $21;
   $24 = ((($$026)) + 12|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (___muldi3(($25|0),0,($w|0),0)|0);
   $27 = tempRet0;
   $28 = (_i64Add(($22|0),0,($26|0),($27|0))|0);
   $29 = tempRet0;
   $30 = ((($$017)) + 12|0);
   HEAP32[$30>>2] = $28;
   $31 = ((($$026)) + 16|0);
   $32 = ((($$017)) + 16|0);
   $33 = (($$035) + -4)|0;
   $34 = ($33>>>0)>(3);
   if ($34) {
    $$017 = $32;$$026 = $31;$$035 = $33;$c1$04 = $29;
   } else {
    $$lcssa = $29;
    break;
   }
  }
  $35 = (($2) - ($3))|0;
  $scevgep14 = (($ap) + ($4<<2)|0);
  $36 = ($2|0)==($3|0);
  if ($36) {
   $$0 = $$lcssa;
   return ($$0|0);
  } else {
   $$01$lcssa15 = $scevgep;$$02$lcssa16 = $scevgep14;$$03$lcssa17 = $35;$c1$0$lcssa18 = $$lcssa;
  }
 } else {
  $$01$lcssa15 = $rp;$$02$lcssa16 = $ap;$$03$lcssa17 = $num;$c1$0$lcssa18 = 0;
 }
 $37 = HEAP32[$$02$lcssa16>>2]|0;
 $38 = (___muldi3(($37|0),0,($w|0),0)|0);
 $39 = tempRet0;
 $40 = (_i64Add(($38|0),($39|0),($c1$0$lcssa18|0),0)|0);
 $41 = tempRet0;
 HEAP32[$$01$lcssa15>>2] = $40;
 $42 = ($$03$lcssa17|0)==(1);
 if ($42) {
  $$0 = $41;
  return ($$0|0);
 }
 $43 = ((($$02$lcssa16)) + 4|0);
 $44 = HEAP32[$43>>2]|0;
 $45 = (___muldi3(($44|0),0,($w|0),0)|0);
 $46 = tempRet0;
 $47 = (_i64Add(($45|0),($46|0),($41|0),0)|0);
 $48 = tempRet0;
 $49 = ((($$01$lcssa15)) + 4|0);
 HEAP32[$49>>2] = $47;
 $50 = ($$03$lcssa17|0)==(2);
 if ($50) {
  $$0 = $48;
  return ($$0|0);
 } else {
  $51 = ((($$02$lcssa16)) + 8|0);
  $52 = HEAP32[$51>>2]|0;
  $53 = (___muldi3(($52|0),0,($w|0),0)|0);
  $54 = tempRet0;
  $55 = (_i64Add(($53|0),($54|0),($48|0),0)|0);
  $56 = tempRet0;
  $57 = ((($$01$lcssa15)) + 8|0);
  HEAP32[$57>>2] = $55;
  $$0 = $56;
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_sub($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $$018 = 0, $$019 = 0, $$026 = 0, $$027 = 0, $$3 = 0, $$5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a$b = 0, $b$a = 0, $neg$010 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $3 = ((($b)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(0);
 if ($2) {
  if ($5) {
   $$018 = $a;$$026 = $b;$neg$010 = 0;
   label = 4;
  } else {
   $$019 = $a;$$027 = $b;
  }
 } else {
  $a$b = $5 ? $a : $b;
  $b$a = $5 ? $b : $a;
  $6 = $5&1;
  $$3 = $6 ^ 1;
  if ($5) {
   $$019 = $b$a;$$027 = $a$b;
  } else {
   $$018 = $b$a;$$026 = $a$b;$neg$010 = $$3;
   label = 4;
  }
 }
 if ((label|0) == 4) {
  $7 = (_EGBN_uadd($r,$$018,$$026)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   $$0 = 0;
   return ($$0|0);
  }
  $9 = ((($r)) + 12|0);
  HEAP32[$9>>2] = $neg$010;
  $$0 = 1;
  return ($$0|0);
 }
 $10 = ((($$019)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($$027)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($11|0)>($13|0);
 $$5 = $14 ? $11 : $13;
 $15 = ((($r)) + 8|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ($$5|0)>($16|0);
 if ($17) {
  $18 = (_egbn_expand2($r,$$5)|0);
  $19 = ($18|0)==(0|0);
  if ($19) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $20 = (_EGBN_ucmp($$019,$$027)|0);
 $21 = ($20|0)<(0);
 if ($21) {
  $22 = (_EGBN_usub($r,$$027,$$019)|0);
  $23 = ($22|0)==(0);
  if ($23) {
   $$0 = 0;
   return ($$0|0);
  }
  $24 = ((($r)) + 12|0);
  HEAP32[$24>>2] = 1;
  $$0 = 1;
  return ($$0|0);
 } else {
  $25 = (_EGBN_usub($r,$$019,$$027)|0);
  $26 = ($25|0)==(0);
  if ($26) {
   $$0 = 0;
   return ($$0|0);
  }
  $27 = ((($r)) + 12|0);
  HEAP32[$27>>2] = 0;
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_add($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a$b = 0;
 var $b$a = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if ($4) {
  $14 = (_EGBN_uadd($r,$a,$b)|0);
  $15 = ($14|0)==(0);
  if ($15) {
   $$0 = 0;
   return ($$0|0);
  }
  $16 = ((($r)) + 12|0);
  $not$ = ($1|0)!=(0);
  $$ = $not$&1;
  HEAP32[$16>>2] = $$;
  $$0 = 1;
  return ($$0|0);
 }
 $5 = ($1|0)==(0);
 $b$a = $5 ? $b : $a;
 $a$b = $5 ? $a : $b;
 $6 = (_EGBN_ucmp($a$b,$b$a)|0);
 $7 = ($6|0)<(0);
 if ($7) {
  $8 = (_EGBN_usub($r,$b$a,$a$b)|0);
  $9 = ($8|0)==(0);
  if ($9) {
   $$0 = 0;
   return ($$0|0);
  }
  $10 = ((($r)) + 12|0);
  HEAP32[$10>>2] = 1;
  $$0 = 1;
  return ($$0|0);
 } else {
  $11 = (_EGBN_usub($r,$a$b,$b$a)|0);
  $12 = ($11|0)==(0);
  if ($12) {
   $$0 = 0;
   return ($$0|0);
  }
  $13 = ((($r)) + 12|0);
  HEAP32[$13>>2] = 0;
  $$0 = 1;
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_uadd($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $a$b = 0, $ap$0 = 0, $ap$0$lcssa = 0, $ap$2 = 0, $ap$313 = 0, $b$a = 0, $exitcond = 0, $i$0 = 0, $i$0$lcssa = 0, $i$2 = 0, $i$312 = 0, $or$cond = 0, $rp$0 = 0, $rp$0$lcssa = 0, $rp$2 = 0, $rp$314 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($b)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)<($3|0);
 $a$b = $4 ? $a : $b;
 $b$a = $4 ? $b : $a;
 $5 = ((($b$a)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($a$b)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($r)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($6|0)<($10|0);
 if (!($11)) {
  $12 = (($6) + 1)|0;
  $13 = (_egbn_expand2($r,$12)|0);
  $14 = ($13|0)==(0|0);
  if ($14) {
   $$0 = 0;
   return ($$0|0);
  }
 }
 $15 = ((($r)) + 4|0);
 HEAP32[$15>>2] = $6;
 $16 = HEAP32[$b$a>>2]|0;
 $17 = HEAP32[$a$b>>2]|0;
 $18 = HEAP32[$r>>2]|0;
 $19 = (_egbn_add_words($18,$16,$17,$8)|0);
 $20 = (($18) + ($8<<2)|0);
 $21 = (($16) + ($8<<2)|0);
 $22 = ($19|0)==(0);
 L5: do {
  if ($22) {
   $ap$2 = $21;$i$2 = $8;$rp$2 = $20;
  } else {
   $ap$0 = $21;$i$0 = $8;$rp$0 = $20;
   while(1) {
    $23 = ($i$0|0)<($6|0);
    if (!($23)) {
     $ap$0$lcssa = $ap$0;$i$0$lcssa = $i$0;$rp$0$lcssa = $rp$0;
     break;
    }
    $24 = (($i$0) + 1)|0;
    $25 = ((($ap$0)) + 4|0);
    $26 = HEAP32[$ap$0>>2]|0;
    $27 = (($26) + 1)|0;
    $28 = ((($rp$0)) + 4|0);
    HEAP32[$rp$0>>2] = $27;
    $29 = ($26|0)==(-1);
    if ($29) {
     $ap$0 = $25;$i$0 = $24;$rp$0 = $28;
    } else {
     $ap$2 = $25;$i$2 = $24;$rp$2 = $28;
     break L5;
    }
   }
   $30 = ((($rp$0$lcssa)) + 4|0);
   HEAP32[$rp$0$lcssa>>2] = 1;
   $31 = HEAP32[$15>>2]|0;
   $32 = (($31) + 1)|0;
   HEAP32[$15>>2] = $32;
   $ap$2 = $ap$0$lcssa;$i$2 = $i$0$lcssa;$rp$2 = $30;
  }
 } while(0);
 $33 = ($rp$2|0)!=($ap$2|0);
 $34 = ($i$2|0)<($6|0);
 $or$cond = $33 & $34;
 if ($or$cond) {
  $ap$313 = $ap$2;$i$312 = $i$2;$rp$314 = $rp$2;
  while(1) {
   $35 = ((($ap$313)) + 4|0);
   $36 = HEAP32[$ap$313>>2]|0;
   $37 = ((($rp$314)) + 4|0);
   HEAP32[$rp$314>>2] = $36;
   $38 = (($i$312) + 1)|0;
   $exitcond = ($38|0)==($6|0);
   if ($exitcond) {
    break;
   } else {
    $ap$313 = $35;$i$312 = $38;$rp$314 = $37;
   }
  }
 }
 $39 = ((($r)) + 12|0);
 HEAP32[$39>>2] = 0;
 $$0 = 1;
 return ($$0|0);
}
function _egbn_add_words($r,$a,$b,$n) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$0111 = 0, $$012 = 0, $$0210 = 0, $$039 = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(1);
 if ($0) {
  $$04 = 0;
  return ($$04|0);
 }
 $1 = HEAP32[$a>>2]|0;
 $2 = HEAP32[$b>>2]|0;
 $3 = (_i64Add(($1|0),0,($2|0),0)|0);
 $4 = tempRet0;
 HEAP32[$r>>2] = $3;
 $5 = ($n|0)<(2);
 if ($5) {
  $$04 = $4;
  return ($$04|0);
 } else {
  $$0111 = $b;$$012 = $n;$$0210 = $a;$$039 = $r;$10 = $4;$11 = 0;
 }
 while(1) {
  $6 = ((($$0210)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ((($$0111)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $12 = (_i64Add(($7|0),0,($10|0),($11|0))|0);
  $13 = tempRet0;
  $14 = (_i64Add(($12|0),($13|0),($9|0),0)|0);
  $15 = tempRet0;
  $16 = ((($$039)) + 4|0);
  HEAP32[$16>>2] = $14;
  $17 = (($$012) + -2)|0;
  $18 = ($17|0)<(1);
  if ($18) {
   $51 = $15;$52 = 0;
   label = 9;
   break;
  }
  $19 = ((($$0210)) + 8|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($$0111)) + 8|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (_i64Add(($20|0),0,($15|0),0)|0);
  $24 = tempRet0;
  $25 = (_i64Add(($23|0),($24|0),($22|0),0)|0);
  $26 = tempRet0;
  $27 = ((($$039)) + 8|0);
  HEAP32[$27>>2] = $25;
  $28 = (($$012) + -3)|0;
  $29 = ($28|0)<(1);
  if ($29) {
   $53 = $26;$54 = 0;
   label = 8;
   break;
  }
  $30 = ((($$0210)) + 12|0);
  $31 = HEAP32[$30>>2]|0;
  $32 = ((($$0111)) + 12|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = (_i64Add(($31|0),0,($26|0),0)|0);
  $35 = tempRet0;
  $36 = (_i64Add(($34|0),($35|0),($33|0),0)|0);
  $37 = tempRet0;
  $38 = ((($$039)) + 12|0);
  HEAP32[$38>>2] = $36;
  $39 = (($$012) + -4)|0;
  $40 = ($39|0)<(1);
  if ($40) {
   $55 = $37;$56 = 0;
   label = 7;
   break;
  }
  $41 = ((($$0210)) + 16|0);
  $42 = ((($$0111)) + 16|0);
  $43 = ((($$039)) + 16|0);
  $44 = HEAP32[$41>>2]|0;
  $45 = HEAP32[$42>>2]|0;
  $46 = (_i64Add(($44|0),0,($37|0),0)|0);
  $47 = tempRet0;
  $48 = (_i64Add(($46|0),($47|0),($45|0),0)|0);
  $49 = tempRet0;
  HEAP32[$43>>2] = $48;
  $50 = ($39|0)<(2);
  if ($50) {
   $57 = $49;$58 = 0;
   label = 10;
   break;
  } else {
   $$0111 = $42;$$012 = $39;$$0210 = $41;$$039 = $43;$10 = $49;$11 = 0;
  }
 }
 if ((label|0) == 7) {
  $$04 = $55;
  return ($$04|0);
 }
 else if ((label|0) == 8) {
  $$04 = $53;
  return ($$04|0);
 }
 else if ((label|0) == 9) {
  $$04 = $51;
  return ($$04|0);
 }
 else if ((label|0) == 10) {
  $$04 = $57;
  return ($$04|0);
 }
 return (0)|0;
}
function _EGBN_num_bits_word($l) {
 $l = $l|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($l>>>0)>(65535);
 if ($0) {
  $1 = ($l>>>0)>(16777215);
  if ($1) {
   $2 = $l >>> 24;
   $3 = (146866 + ($2)|0);
   $4 = HEAP8[$3>>0]|0;
   $5 = $4 << 24 >> 24;
   $6 = (($5) + 24)|0;
   $$0 = $6;
   return ($$0|0);
  } else {
   $7 = $l >>> 16;
   $8 = (146866 + ($7)|0);
   $9 = HEAP8[$8>>0]|0;
   $10 = $9 << 24 >> 24;
   $11 = (($10) + 16)|0;
   $$0 = $11;
   return ($$0|0);
  }
 } else {
  $12 = $l & 65280;
  $13 = ($12|0)==(0);
  if ($13) {
   $19 = (146866 + ($l)|0);
   $20 = HEAP8[$19>>0]|0;
   $21 = $20 << 24 >> 24;
   $$0 = $21;
   return ($$0|0);
  } else {
   $14 = $l >>> 8;
   $15 = (146866 + ($14)|0);
   $16 = HEAP8[$15>>0]|0;
   $17 = $16 << 24 >> 24;
   $18 = (($17) + 8)|0;
   $$0 = $18;
   return ($$0|0);
  }
 }
 return (0)|0;
}
function _egbn_mul_comba8($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$14 = 0, $$15 = 0, $$16 = 0, $$17 = 0, $$18 = 0, $$19 = 0, $$2 = 0, $$20 = 0, $$21 = 0, $$22 = 0, $$23 = 0, $$24 = 0, $$25 = 0, $$26 = 0;
 var $$27 = 0, $$28 = 0, $$29 = 0, $$3 = 0, $$30 = 0, $$31 = 0, $$32 = 0, $$33 = 0, $$34 = 0, $$35 = 0, $$36 = 0, $$37 = 0, $$38 = 0, $$39 = 0, $$4 = 0, $$40 = 0, $$41 = 0, $$42 = 0, $$43 = 0, $$44 = 0;
 var $$45 = 0, $$46 = 0, $$47 = 0, $$48 = 0, $$49 = 0, $$5 = 0, $$50 = 0, $$51 = 0, $$52 = 0, $$53 = 0, $$54 = 0, $$55 = 0, $$56 = 0, $$57 = 0, $$58 = 0, $$59 = 0, $$6 = 0, $$60 = 0, $$61 = 0, $$7 = 0;
 var $$8 = 0, $$9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0;
 var $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0;
 var $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0;
 var $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0;
 var $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0;
 var $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0;
 var $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0;
 var $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0;
 var $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0;
 var $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0;
 var $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0;
 var $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0;
 var $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c1$1 = 0, $c1$10 = 0, $c1$11 = 0, $c1$12 = 0, $c1$13 = 0, $c1$14 = 0, $c1$15 = 0, $c1$16 = 0, $c1$17 = 0, $c1$18 = 0, $c1$19 = 0, $c1$2 = 0, $c1$20 = 0;
 var $c1$21 = 0, $c1$3 = 0, $c1$4 = 0, $c1$5 = 0, $c1$6 = 0, $c1$7 = 0, $c1$8 = 0, $c1$9 = 0, $c2$0 = 0, $c2$1 = 0, $c2$10 = 0, $c2$11 = 0, $c2$12 = 0, $c2$13 = 0, $c2$14 = 0, $c2$15 = 0, $c2$16 = 0, $c2$17 = 0, $c2$18 = 0, $c2$19 = 0;
 var $c2$2 = 0, $c2$3 = 0, $c2$4 = 0, $c2$5 = 0, $c2$6 = 0, $c2$7 = 0, $c2$8 = 0, $c2$9 = 0, $c3$1 = 0, $c3$10 = 0, $c3$11 = 0, $c3$12 = 0, $c3$13 = 0, $c3$14 = 0, $c3$15 = 0, $c3$16 = 0, $c3$17 = 0, $c3$18 = 0, $c3$19 = 0, $c3$2 = 0;
 var $c3$20 = 0, $c3$3 = 0, $c3$4 = 0, $c3$5 = 0, $c3$6 = 0, $c3$7 = 0, $c3$8 = 0, $c3$9 = 0, $t2$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = HEAP32[$b>>2]|0;
 $2 = (___muldi3(($1|0),0,($0|0),0)|0);
 $3 = tempRet0;
 HEAP32[$r>>2] = $2;
 $4 = HEAP32[$a>>2]|0;
 $5 = ((($b)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (___muldi3(($6|0),0,($4|0),0)|0);
 $8 = tempRet0;
 $9 = (($7) + ($3))|0;
 $10 = ($9>>>0)<($7>>>0);
 $11 = $10&1;
 $t2$1 = (($11) + ($8))|0;
 $12 = ((($a)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = HEAP32[$b>>2]|0;
 $15 = (___muldi3(($14|0),0,($13|0),0)|0);
 $16 = tempRet0;
 $17 = (($15) + ($9))|0;
 $18 = ($17>>>0)<($15>>>0);
 $19 = $18&1;
 $$ = (($19) + ($16))|0;
 $20 = (($t2$1) + ($$))|0;
 $21 = ($20>>>0)<($$>>>0);
 $c1$1 = $21&1;
 $22 = ((($r)) + 4|0);
 HEAP32[$22>>2] = $17;
 $23 = ((($a)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$b>>2]|0;
 $26 = (___muldi3(($25|0),0,($24|0),0)|0);
 $27 = tempRet0;
 $28 = (($20) + ($26))|0;
 $29 = ($28>>>0)<($26>>>0);
 $30 = $29&1;
 $$1 = (($30) + ($27))|0;
 $31 = (($$1) + ($c1$1))|0;
 $32 = ($31>>>0)<($$1>>>0);
 $c2$0 = $32&1;
 $33 = HEAP32[$12>>2]|0;
 $34 = HEAP32[$5>>2]|0;
 $35 = (___muldi3(($34|0),0,($33|0),0)|0);
 $36 = tempRet0;
 $37 = (($28) + ($35))|0;
 $38 = ($37>>>0)<($35>>>0);
 $39 = $38&1;
 $$2 = (($39) + ($36))|0;
 $40 = (($31) + ($$2))|0;
 $41 = ($40>>>0)<($$2>>>0);
 $42 = $32 ? 2 : 1;
 $c2$1 = $41 ? $42 : $c2$0;
 $43 = HEAP32[$a>>2]|0;
 $44 = ((($b)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (___muldi3(($45|0),0,($43|0),0)|0);
 $47 = tempRet0;
 $48 = (($37) + ($46))|0;
 $49 = ($48>>>0)<($46>>>0);
 $50 = $49&1;
 $$3 = (($50) + ($47))|0;
 $51 = (($40) + ($$3))|0;
 $52 = ($51>>>0)<($$3>>>0);
 $53 = $52&1;
 $c2$2 = (($53) + ($c2$1))|0;
 $54 = ((($r)) + 8|0);
 HEAP32[$54>>2] = $48;
 $55 = HEAP32[$a>>2]|0;
 $56 = ((($b)) + 12|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = (___muldi3(($57|0),0,($55|0),0)|0);
 $59 = tempRet0;
 $60 = (($51) + ($58))|0;
 $61 = ($60>>>0)<($58>>>0);
 $62 = $61&1;
 $$4 = (($62) + ($59))|0;
 $63 = (($c2$2) + ($$4))|0;
 $64 = ($63>>>0)<($$4>>>0);
 $c3$1 = $64&1;
 $65 = HEAP32[$12>>2]|0;
 $66 = HEAP32[$44>>2]|0;
 $67 = (___muldi3(($66|0),0,($65|0),0)|0);
 $68 = tempRet0;
 $69 = (($60) + ($67))|0;
 $70 = ($69>>>0)<($67>>>0);
 $71 = $70&1;
 $$5 = (($71) + ($68))|0;
 $72 = (($63) + ($$5))|0;
 $73 = ($72>>>0)<($$5>>>0);
 $74 = $64 ? 2 : 1;
 $c3$2 = $73 ? $74 : $c3$1;
 $75 = HEAP32[$23>>2]|0;
 $76 = HEAP32[$5>>2]|0;
 $77 = (___muldi3(($76|0),0,($75|0),0)|0);
 $78 = tempRet0;
 $79 = (($69) + ($77))|0;
 $80 = ($79>>>0)<($77>>>0);
 $81 = $80&1;
 $$6 = (($81) + ($78))|0;
 $82 = (($72) + ($$6))|0;
 $83 = ($82>>>0)<($$6>>>0);
 $84 = $83&1;
 $c3$3 = (($84) + ($c3$2))|0;
 $85 = ((($a)) + 12|0);
 $86 = HEAP32[$85>>2]|0;
 $87 = HEAP32[$b>>2]|0;
 $88 = (___muldi3(($87|0),0,($86|0),0)|0);
 $89 = tempRet0;
 $90 = (($79) + ($88))|0;
 $91 = ($90>>>0)<($88>>>0);
 $92 = $91&1;
 $$7 = (($92) + ($89))|0;
 $93 = (($82) + ($$7))|0;
 $94 = ($93>>>0)<($$7>>>0);
 $95 = $94&1;
 $c3$4 = (($c3$3) + ($95))|0;
 $96 = ((($r)) + 12|0);
 HEAP32[$96>>2] = $90;
 $97 = ((($a)) + 16|0);
 $98 = HEAP32[$97>>2]|0;
 $99 = HEAP32[$b>>2]|0;
 $100 = (___muldi3(($99|0),0,($98|0),0)|0);
 $101 = tempRet0;
 $102 = (($93) + ($100))|0;
 $103 = ($102>>>0)<($100>>>0);
 $104 = $103&1;
 $$8 = (($104) + ($101))|0;
 $105 = (($c3$4) + ($$8))|0;
 $106 = ($105>>>0)<($$8>>>0);
 $c1$2 = $106&1;
 $107 = HEAP32[$85>>2]|0;
 $108 = HEAP32[$5>>2]|0;
 $109 = (___muldi3(($108|0),0,($107|0),0)|0);
 $110 = tempRet0;
 $111 = (($102) + ($109))|0;
 $112 = ($111>>>0)<($109>>>0);
 $113 = $112&1;
 $$9 = (($113) + ($110))|0;
 $114 = (($105) + ($$9))|0;
 $115 = ($114>>>0)<($$9>>>0);
 $116 = $106 ? 2 : 1;
 $c1$3 = $115 ? $116 : $c1$2;
 $117 = HEAP32[$23>>2]|0;
 $118 = HEAP32[$44>>2]|0;
 $119 = (___muldi3(($118|0),0,($117|0),0)|0);
 $120 = tempRet0;
 $121 = (($111) + ($119))|0;
 $122 = ($121>>>0)<($119>>>0);
 $123 = $122&1;
 $$10 = (($123) + ($120))|0;
 $124 = (($114) + ($$10))|0;
 $125 = ($124>>>0)<($$10>>>0);
 $126 = $125&1;
 $c1$4 = (($126) + ($c1$3))|0;
 $127 = HEAP32[$12>>2]|0;
 $128 = HEAP32[$56>>2]|0;
 $129 = (___muldi3(($128|0),0,($127|0),0)|0);
 $130 = tempRet0;
 $131 = (($121) + ($129))|0;
 $132 = ($131>>>0)<($129>>>0);
 $133 = $132&1;
 $$11 = (($133) + ($130))|0;
 $134 = (($124) + ($$11))|0;
 $135 = ($134>>>0)<($$11>>>0);
 $136 = $135&1;
 $c1$5 = (($c1$4) + ($136))|0;
 $137 = HEAP32[$a>>2]|0;
 $138 = ((($b)) + 16|0);
 $139 = HEAP32[$138>>2]|0;
 $140 = (___muldi3(($139|0),0,($137|0),0)|0);
 $141 = tempRet0;
 $142 = (($131) + ($140))|0;
 $143 = ($142>>>0)<($140>>>0);
 $144 = $143&1;
 $$12 = (($144) + ($141))|0;
 $145 = (($134) + ($$12))|0;
 $146 = ($145>>>0)<($$12>>>0);
 $147 = $146&1;
 $c1$6 = (($c1$5) + ($147))|0;
 $148 = ((($r)) + 16|0);
 HEAP32[$148>>2] = $142;
 $149 = HEAP32[$a>>2]|0;
 $150 = ((($b)) + 20|0);
 $151 = HEAP32[$150>>2]|0;
 $152 = (___muldi3(($151|0),0,($149|0),0)|0);
 $153 = tempRet0;
 $154 = (($145) + ($152))|0;
 $155 = ($154>>>0)<($152>>>0);
 $156 = $155&1;
 $$13 = (($156) + ($153))|0;
 $157 = (($c1$6) + ($$13))|0;
 $158 = ($157>>>0)<($$13>>>0);
 $c2$3 = $158&1;
 $159 = HEAP32[$12>>2]|0;
 $160 = HEAP32[$138>>2]|0;
 $161 = (___muldi3(($160|0),0,($159|0),0)|0);
 $162 = tempRet0;
 $163 = (($154) + ($161))|0;
 $164 = ($163>>>0)<($161>>>0);
 $165 = $164&1;
 $$14 = (($165) + ($162))|0;
 $166 = (($157) + ($$14))|0;
 $167 = ($166>>>0)<($$14>>>0);
 $168 = $158 ? 2 : 1;
 $c2$4 = $167 ? $168 : $c2$3;
 $169 = HEAP32[$23>>2]|0;
 $170 = HEAP32[$56>>2]|0;
 $171 = (___muldi3(($170|0),0,($169|0),0)|0);
 $172 = tempRet0;
 $173 = (($163) + ($171))|0;
 $174 = ($173>>>0)<($171>>>0);
 $175 = $174&1;
 $$15 = (($175) + ($172))|0;
 $176 = (($166) + ($$15))|0;
 $177 = ($176>>>0)<($$15>>>0);
 $178 = $177&1;
 $c2$5 = (($178) + ($c2$4))|0;
 $179 = HEAP32[$85>>2]|0;
 $180 = HEAP32[$44>>2]|0;
 $181 = (___muldi3(($180|0),0,($179|0),0)|0);
 $182 = tempRet0;
 $183 = (($173) + ($181))|0;
 $184 = ($183>>>0)<($181>>>0);
 $185 = $184&1;
 $$16 = (($185) + ($182))|0;
 $186 = (($176) + ($$16))|0;
 $187 = ($186>>>0)<($$16>>>0);
 $188 = $187&1;
 $c2$6 = (($c2$5) + ($188))|0;
 $189 = HEAP32[$97>>2]|0;
 $190 = HEAP32[$5>>2]|0;
 $191 = (___muldi3(($190|0),0,($189|0),0)|0);
 $192 = tempRet0;
 $193 = (($191) + ($183))|0;
 $194 = ($193>>>0)<($191>>>0);
 $195 = $194&1;
 $$17 = (($195) + ($192))|0;
 $196 = (($$17) + ($186))|0;
 $197 = ($196>>>0)<($$17>>>0);
 $198 = $197&1;
 $c2$7 = (($c2$6) + ($198))|0;
 $199 = ((($a)) + 20|0);
 $200 = HEAP32[$199>>2]|0;
 $201 = HEAP32[$b>>2]|0;
 $202 = (___muldi3(($201|0),0,($200|0),0)|0);
 $203 = tempRet0;
 $204 = (($202) + ($193))|0;
 $205 = ($204>>>0)<($202>>>0);
 $206 = $205&1;
 $$18 = (($206) + ($203))|0;
 $207 = (($$18) + ($196))|0;
 $208 = ($207>>>0)<($$18>>>0);
 $209 = $208&1;
 $c2$8 = (($c2$7) + ($209))|0;
 $210 = ((($r)) + 20|0);
 HEAP32[$210>>2] = $204;
 $211 = ((($a)) + 24|0);
 $212 = HEAP32[$211>>2]|0;
 $213 = HEAP32[$b>>2]|0;
 $214 = (___muldi3(($213|0),0,($212|0),0)|0);
 $215 = tempRet0;
 $216 = (($207) + ($214))|0;
 $217 = ($216>>>0)<($214>>>0);
 $218 = $217&1;
 $$19 = (($218) + ($215))|0;
 $219 = (($c2$8) + ($$19))|0;
 $220 = ($219>>>0)<($$19>>>0);
 $c3$5 = $220&1;
 $221 = HEAP32[$199>>2]|0;
 $222 = HEAP32[$5>>2]|0;
 $223 = (___muldi3(($222|0),0,($221|0),0)|0);
 $224 = tempRet0;
 $225 = (($216) + ($223))|0;
 $226 = ($225>>>0)<($223>>>0);
 $227 = $226&1;
 $$20 = (($227) + ($224))|0;
 $228 = (($219) + ($$20))|0;
 $229 = ($228>>>0)<($$20>>>0);
 $230 = $220 ? 2 : 1;
 $c3$6 = $229 ? $230 : $c3$5;
 $231 = HEAP32[$97>>2]|0;
 $232 = HEAP32[$44>>2]|0;
 $233 = (___muldi3(($232|0),0,($231|0),0)|0);
 $234 = tempRet0;
 $235 = (($225) + ($233))|0;
 $236 = ($235>>>0)<($233>>>0);
 $237 = $236&1;
 $$21 = (($237) + ($234))|0;
 $238 = (($228) + ($$21))|0;
 $239 = ($238>>>0)<($$21>>>0);
 $240 = $239&1;
 $c3$7 = (($240) + ($c3$6))|0;
 $241 = HEAP32[$85>>2]|0;
 $242 = HEAP32[$56>>2]|0;
 $243 = (___muldi3(($242|0),0,($241|0),0)|0);
 $244 = tempRet0;
 $245 = (($235) + ($243))|0;
 $246 = ($245>>>0)<($243>>>0);
 $247 = $246&1;
 $$22 = (($247) + ($244))|0;
 $248 = (($238) + ($$22))|0;
 $249 = ($248>>>0)<($$22>>>0);
 $250 = $249&1;
 $c3$8 = (($c3$7) + ($250))|0;
 $251 = HEAP32[$23>>2]|0;
 $252 = HEAP32[$138>>2]|0;
 $253 = (___muldi3(($252|0),0,($251|0),0)|0);
 $254 = tempRet0;
 $255 = (($253) + ($245))|0;
 $256 = ($255>>>0)<($253>>>0);
 $257 = $256&1;
 $$23 = (($257) + ($254))|0;
 $258 = (($$23) + ($248))|0;
 $259 = ($258>>>0)<($$23>>>0);
 $260 = $259&1;
 $c3$9 = (($c3$8) + ($260))|0;
 $261 = HEAP32[$12>>2]|0;
 $262 = HEAP32[$150>>2]|0;
 $263 = (___muldi3(($262|0),0,($261|0),0)|0);
 $264 = tempRet0;
 $265 = (($263) + ($255))|0;
 $266 = ($265>>>0)<($263>>>0);
 $267 = $266&1;
 $$24 = (($267) + ($264))|0;
 $268 = (($$24) + ($258))|0;
 $269 = ($268>>>0)<($$24>>>0);
 $270 = $269&1;
 $c3$10 = (($c3$9) + ($270))|0;
 $271 = HEAP32[$a>>2]|0;
 $272 = ((($b)) + 24|0);
 $273 = HEAP32[$272>>2]|0;
 $274 = (___muldi3(($273|0),0,($271|0),0)|0);
 $275 = tempRet0;
 $276 = (($274) + ($265))|0;
 $277 = ($276>>>0)<($274>>>0);
 $278 = $277&1;
 $$25 = (($278) + ($275))|0;
 $279 = (($$25) + ($268))|0;
 $280 = ($279>>>0)<($$25>>>0);
 $281 = $280&1;
 $c3$11 = (($c3$10) + ($281))|0;
 $282 = ((($r)) + 24|0);
 HEAP32[$282>>2] = $276;
 $283 = HEAP32[$a>>2]|0;
 $284 = ((($b)) + 28|0);
 $285 = HEAP32[$284>>2]|0;
 $286 = (___muldi3(($285|0),0,($283|0),0)|0);
 $287 = tempRet0;
 $288 = (($279) + ($286))|0;
 $289 = ($288>>>0)<($286>>>0);
 $290 = $289&1;
 $$26 = (($290) + ($287))|0;
 $291 = (($c3$11) + ($$26))|0;
 $292 = ($291>>>0)<($$26>>>0);
 $c1$7 = $292&1;
 $293 = HEAP32[$12>>2]|0;
 $294 = HEAP32[$272>>2]|0;
 $295 = (___muldi3(($294|0),0,($293|0),0)|0);
 $296 = tempRet0;
 $297 = (($288) + ($295))|0;
 $298 = ($297>>>0)<($295>>>0);
 $299 = $298&1;
 $$27 = (($299) + ($296))|0;
 $300 = (($291) + ($$27))|0;
 $301 = ($300>>>0)<($$27>>>0);
 $302 = $292 ? 2 : 1;
 $c1$8 = $301 ? $302 : $c1$7;
 $303 = HEAP32[$23>>2]|0;
 $304 = HEAP32[$150>>2]|0;
 $305 = (___muldi3(($304|0),0,($303|0),0)|0);
 $306 = tempRet0;
 $307 = (($297) + ($305))|0;
 $308 = ($307>>>0)<($305>>>0);
 $309 = $308&1;
 $$28 = (($309) + ($306))|0;
 $310 = (($300) + ($$28))|0;
 $311 = ($310>>>0)<($$28>>>0);
 $312 = $311&1;
 $c1$9 = (($312) + ($c1$8))|0;
 $313 = HEAP32[$85>>2]|0;
 $314 = HEAP32[$138>>2]|0;
 $315 = (___muldi3(($314|0),0,($313|0),0)|0);
 $316 = tempRet0;
 $317 = (($307) + ($315))|0;
 $318 = ($317>>>0)<($315>>>0);
 $319 = $318&1;
 $$29 = (($319) + ($316))|0;
 $320 = (($310) + ($$29))|0;
 $321 = ($320>>>0)<($$29>>>0);
 $322 = $321&1;
 $c1$10 = (($c1$9) + ($322))|0;
 $323 = HEAP32[$97>>2]|0;
 $324 = HEAP32[$56>>2]|0;
 $325 = (___muldi3(($324|0),0,($323|0),0)|0);
 $326 = tempRet0;
 $327 = (($325) + ($317))|0;
 $328 = ($327>>>0)<($325>>>0);
 $329 = $328&1;
 $$30 = (($329) + ($326))|0;
 $330 = (($$30) + ($320))|0;
 $331 = ($330>>>0)<($$30>>>0);
 $332 = $331&1;
 $c1$11 = (($c1$10) + ($332))|0;
 $333 = HEAP32[$199>>2]|0;
 $334 = HEAP32[$44>>2]|0;
 $335 = (___muldi3(($334|0),0,($333|0),0)|0);
 $336 = tempRet0;
 $337 = (($335) + ($327))|0;
 $338 = ($337>>>0)<($335>>>0);
 $339 = $338&1;
 $$31 = (($339) + ($336))|0;
 $340 = (($$31) + ($330))|0;
 $341 = ($340>>>0)<($$31>>>0);
 $342 = $341&1;
 $c1$12 = (($c1$11) + ($342))|0;
 $343 = HEAP32[$211>>2]|0;
 $344 = HEAP32[$5>>2]|0;
 $345 = (___muldi3(($344|0),0,($343|0),0)|0);
 $346 = tempRet0;
 $347 = (($345) + ($337))|0;
 $348 = ($347>>>0)<($345>>>0);
 $349 = $348&1;
 $$32 = (($349) + ($346))|0;
 $350 = (($$32) + ($340))|0;
 $351 = ($350>>>0)<($$32>>>0);
 $352 = $351&1;
 $c1$13 = (($c1$12) + ($352))|0;
 $353 = ((($a)) + 28|0);
 $354 = HEAP32[$353>>2]|0;
 $355 = HEAP32[$b>>2]|0;
 $356 = (___muldi3(($355|0),0,($354|0),0)|0);
 $357 = tempRet0;
 $358 = (($356) + ($347))|0;
 $359 = ($358>>>0)<($356>>>0);
 $360 = $359&1;
 $$33 = (($360) + ($357))|0;
 $361 = (($$33) + ($350))|0;
 $362 = ($361>>>0)<($$33>>>0);
 $363 = $362&1;
 $c1$14 = (($c1$13) + ($363))|0;
 $364 = ((($r)) + 28|0);
 HEAP32[$364>>2] = $358;
 $365 = HEAP32[$353>>2]|0;
 $366 = HEAP32[$5>>2]|0;
 $367 = (___muldi3(($366|0),0,($365|0),0)|0);
 $368 = tempRet0;
 $369 = (($361) + ($367))|0;
 $370 = ($369>>>0)<($367>>>0);
 $371 = $370&1;
 $$34 = (($371) + ($368))|0;
 $372 = (($c1$14) + ($$34))|0;
 $373 = ($372>>>0)<($$34>>>0);
 $c2$9 = $373&1;
 $374 = HEAP32[$211>>2]|0;
 $375 = HEAP32[$44>>2]|0;
 $376 = (___muldi3(($375|0),0,($374|0),0)|0);
 $377 = tempRet0;
 $378 = (($369) + ($376))|0;
 $379 = ($378>>>0)<($376>>>0);
 $380 = $379&1;
 $$35 = (($380) + ($377))|0;
 $381 = (($372) + ($$35))|0;
 $382 = ($381>>>0)<($$35>>>0);
 $383 = $373 ? 2 : 1;
 $c2$10 = $382 ? $383 : $c2$9;
 $384 = HEAP32[$199>>2]|0;
 $385 = HEAP32[$56>>2]|0;
 $386 = (___muldi3(($385|0),0,($384|0),0)|0);
 $387 = tempRet0;
 $388 = (($378) + ($386))|0;
 $389 = ($388>>>0)<($386>>>0);
 $390 = $389&1;
 $$36 = (($390) + ($387))|0;
 $391 = (($381) + ($$36))|0;
 $392 = ($391>>>0)<($$36>>>0);
 $393 = $392&1;
 $c2$11 = (($393) + ($c2$10))|0;
 $394 = HEAP32[$97>>2]|0;
 $395 = HEAP32[$138>>2]|0;
 $396 = (___muldi3(($395|0),0,($394|0),0)|0);
 $397 = tempRet0;
 $398 = (($388) + ($396))|0;
 $399 = ($398>>>0)<($396>>>0);
 $400 = $399&1;
 $$37 = (($400) + ($397))|0;
 $401 = (($391) + ($$37))|0;
 $402 = ($401>>>0)<($$37>>>0);
 $403 = $402&1;
 $c2$12 = (($c2$11) + ($403))|0;
 $404 = HEAP32[$85>>2]|0;
 $405 = HEAP32[$150>>2]|0;
 $406 = (___muldi3(($405|0),0,($404|0),0)|0);
 $407 = tempRet0;
 $408 = (($406) + ($398))|0;
 $409 = ($408>>>0)<($406>>>0);
 $410 = $409&1;
 $$38 = (($410) + ($407))|0;
 $411 = (($$38) + ($401))|0;
 $412 = ($411>>>0)<($$38>>>0);
 $413 = $412&1;
 $c2$13 = (($c2$12) + ($413))|0;
 $414 = HEAP32[$23>>2]|0;
 $415 = HEAP32[$272>>2]|0;
 $416 = (___muldi3(($415|0),0,($414|0),0)|0);
 $417 = tempRet0;
 $418 = (($416) + ($408))|0;
 $419 = ($418>>>0)<($416>>>0);
 $420 = $419&1;
 $$39 = (($420) + ($417))|0;
 $421 = (($$39) + ($411))|0;
 $422 = ($421>>>0)<($$39>>>0);
 $423 = $422&1;
 $c2$14 = (($c2$13) + ($423))|0;
 $424 = HEAP32[$12>>2]|0;
 $425 = HEAP32[$284>>2]|0;
 $426 = (___muldi3(($425|0),0,($424|0),0)|0);
 $427 = tempRet0;
 $428 = (($426) + ($418))|0;
 $429 = ($428>>>0)<($426>>>0);
 $430 = $429&1;
 $$40 = (($430) + ($427))|0;
 $431 = (($$40) + ($421))|0;
 $432 = ($431>>>0)<($$40>>>0);
 $433 = $432&1;
 $c2$15 = (($c2$14) + ($433))|0;
 $434 = ((($r)) + 32|0);
 HEAP32[$434>>2] = $428;
 $435 = HEAP32[$23>>2]|0;
 $436 = HEAP32[$284>>2]|0;
 $437 = (___muldi3(($436|0),0,($435|0),0)|0);
 $438 = tempRet0;
 $439 = (($431) + ($437))|0;
 $440 = ($439>>>0)<($437>>>0);
 $441 = $440&1;
 $$41 = (($441) + ($438))|0;
 $442 = (($c2$15) + ($$41))|0;
 $443 = ($442>>>0)<($$41>>>0);
 $c3$12 = $443&1;
 $444 = HEAP32[$85>>2]|0;
 $445 = HEAP32[$272>>2]|0;
 $446 = (___muldi3(($445|0),0,($444|0),0)|0);
 $447 = tempRet0;
 $448 = (($439) + ($446))|0;
 $449 = ($448>>>0)<($446>>>0);
 $450 = $449&1;
 $$42 = (($450) + ($447))|0;
 $451 = (($442) + ($$42))|0;
 $452 = ($451>>>0)<($$42>>>0);
 $453 = $443 ? 2 : 1;
 $c3$13 = $452 ? $453 : $c3$12;
 $454 = HEAP32[$97>>2]|0;
 $455 = HEAP32[$150>>2]|0;
 $456 = (___muldi3(($455|0),0,($454|0),0)|0);
 $457 = tempRet0;
 $458 = (($448) + ($456))|0;
 $459 = ($458>>>0)<($456>>>0);
 $460 = $459&1;
 $$43 = (($460) + ($457))|0;
 $461 = (($451) + ($$43))|0;
 $462 = ($461>>>0)<($$43>>>0);
 $463 = $462&1;
 $c3$14 = (($463) + ($c3$13))|0;
 $464 = HEAP32[$199>>2]|0;
 $465 = HEAP32[$138>>2]|0;
 $466 = (___muldi3(($465|0),0,($464|0),0)|0);
 $467 = tempRet0;
 $468 = (($458) + ($466))|0;
 $469 = ($468>>>0)<($466>>>0);
 $470 = $469&1;
 $$44 = (($470) + ($467))|0;
 $471 = (($461) + ($$44))|0;
 $472 = ($471>>>0)<($$44>>>0);
 $473 = $472&1;
 $c3$15 = (($c3$14) + ($473))|0;
 $474 = HEAP32[$211>>2]|0;
 $475 = HEAP32[$56>>2]|0;
 $476 = (___muldi3(($475|0),0,($474|0),0)|0);
 $477 = tempRet0;
 $478 = (($476) + ($468))|0;
 $479 = ($478>>>0)<($476>>>0);
 $480 = $479&1;
 $$45 = (($480) + ($477))|0;
 $481 = (($$45) + ($471))|0;
 $482 = ($481>>>0)<($$45>>>0);
 $483 = $482&1;
 $c3$16 = (($c3$15) + ($483))|0;
 $484 = HEAP32[$353>>2]|0;
 $485 = HEAP32[$44>>2]|0;
 $486 = (___muldi3(($485|0),0,($484|0),0)|0);
 $487 = tempRet0;
 $488 = (($486) + ($478))|0;
 $489 = ($488>>>0)<($486>>>0);
 $490 = $489&1;
 $$46 = (($490) + ($487))|0;
 $491 = (($$46) + ($481))|0;
 $492 = ($491>>>0)<($$46>>>0);
 $493 = $492&1;
 $c3$17 = (($c3$16) + ($493))|0;
 $494 = ((($r)) + 36|0);
 HEAP32[$494>>2] = $488;
 $495 = HEAP32[$353>>2]|0;
 $496 = HEAP32[$56>>2]|0;
 $497 = (___muldi3(($496|0),0,($495|0),0)|0);
 $498 = tempRet0;
 $499 = (($491) + ($497))|0;
 $500 = ($499>>>0)<($497>>>0);
 $501 = $500&1;
 $$47 = (($501) + ($498))|0;
 $502 = (($c3$17) + ($$47))|0;
 $503 = ($502>>>0)<($$47>>>0);
 $c1$15 = $503&1;
 $504 = HEAP32[$211>>2]|0;
 $505 = HEAP32[$138>>2]|0;
 $506 = (___muldi3(($505|0),0,($504|0),0)|0);
 $507 = tempRet0;
 $508 = (($499) + ($506))|0;
 $509 = ($508>>>0)<($506>>>0);
 $510 = $509&1;
 $$48 = (($510) + ($507))|0;
 $511 = (($502) + ($$48))|0;
 $512 = ($511>>>0)<($$48>>>0);
 $513 = $503 ? 2 : 1;
 $c1$16 = $512 ? $513 : $c1$15;
 $514 = HEAP32[$199>>2]|0;
 $515 = HEAP32[$150>>2]|0;
 $516 = (___muldi3(($515|0),0,($514|0),0)|0);
 $517 = tempRet0;
 $518 = (($508) + ($516))|0;
 $519 = ($518>>>0)<($516>>>0);
 $520 = $519&1;
 $$49 = (($520) + ($517))|0;
 $521 = (($511) + ($$49))|0;
 $522 = ($521>>>0)<($$49>>>0);
 $523 = $522&1;
 $c1$17 = (($523) + ($c1$16))|0;
 $524 = HEAP32[$97>>2]|0;
 $525 = HEAP32[$272>>2]|0;
 $526 = (___muldi3(($525|0),0,($524|0),0)|0);
 $527 = tempRet0;
 $528 = (($518) + ($526))|0;
 $529 = ($528>>>0)<($526>>>0);
 $530 = $529&1;
 $$50 = (($530) + ($527))|0;
 $531 = (($521) + ($$50))|0;
 $532 = ($531>>>0)<($$50>>>0);
 $533 = $532&1;
 $c1$18 = (($c1$17) + ($533))|0;
 $534 = HEAP32[$85>>2]|0;
 $535 = HEAP32[$284>>2]|0;
 $536 = (___muldi3(($535|0),0,($534|0),0)|0);
 $537 = tempRet0;
 $538 = (($536) + ($528))|0;
 $539 = ($538>>>0)<($536>>>0);
 $540 = $539&1;
 $$51 = (($540) + ($537))|0;
 $541 = (($$51) + ($531))|0;
 $542 = ($541>>>0)<($$51>>>0);
 $543 = $542&1;
 $c1$19 = (($c1$18) + ($543))|0;
 $544 = ((($r)) + 40|0);
 HEAP32[$544>>2] = $538;
 $545 = HEAP32[$97>>2]|0;
 $546 = HEAP32[$284>>2]|0;
 $547 = (___muldi3(($546|0),0,($545|0),0)|0);
 $548 = tempRet0;
 $549 = (($541) + ($547))|0;
 $550 = ($549>>>0)<($547>>>0);
 $551 = $550&1;
 $$52 = (($551) + ($548))|0;
 $552 = (($c1$19) + ($$52))|0;
 $553 = ($552>>>0)<($$52>>>0);
 $c2$16 = $553&1;
 $554 = HEAP32[$199>>2]|0;
 $555 = HEAP32[$272>>2]|0;
 $556 = (___muldi3(($555|0),0,($554|0),0)|0);
 $557 = tempRet0;
 $558 = (($549) + ($556))|0;
 $559 = ($558>>>0)<($556>>>0);
 $560 = $559&1;
 $$53 = (($560) + ($557))|0;
 $561 = (($552) + ($$53))|0;
 $562 = ($561>>>0)<($$53>>>0);
 $563 = $553 ? 2 : 1;
 $c2$17 = $562 ? $563 : $c2$16;
 $564 = HEAP32[$211>>2]|0;
 $565 = HEAP32[$150>>2]|0;
 $566 = (___muldi3(($565|0),0,($564|0),0)|0);
 $567 = tempRet0;
 $568 = (($558) + ($566))|0;
 $569 = ($568>>>0)<($566>>>0);
 $570 = $569&1;
 $$54 = (($570) + ($567))|0;
 $571 = (($561) + ($$54))|0;
 $572 = ($571>>>0)<($$54>>>0);
 $573 = $572&1;
 $c2$18 = (($573) + ($c2$17))|0;
 $574 = HEAP32[$353>>2]|0;
 $575 = HEAP32[$138>>2]|0;
 $576 = (___muldi3(($575|0),0,($574|0),0)|0);
 $577 = tempRet0;
 $578 = (($568) + ($576))|0;
 $579 = ($578>>>0)<($576>>>0);
 $580 = $579&1;
 $$55 = (($580) + ($577))|0;
 $581 = (($571) + ($$55))|0;
 $582 = ($581>>>0)<($$55>>>0);
 $583 = $582&1;
 $c2$19 = (($c2$18) + ($583))|0;
 $584 = ((($r)) + 44|0);
 HEAP32[$584>>2] = $578;
 $585 = HEAP32[$353>>2]|0;
 $586 = HEAP32[$150>>2]|0;
 $587 = (___muldi3(($586|0),0,($585|0),0)|0);
 $588 = tempRet0;
 $589 = (($581) + ($587))|0;
 $590 = ($589>>>0)<($587>>>0);
 $591 = $590&1;
 $$56 = (($591) + ($588))|0;
 $592 = (($c2$19) + ($$56))|0;
 $593 = ($592>>>0)<($$56>>>0);
 $c3$18 = $593&1;
 $594 = HEAP32[$211>>2]|0;
 $595 = HEAP32[$272>>2]|0;
 $596 = (___muldi3(($595|0),0,($594|0),0)|0);
 $597 = tempRet0;
 $598 = (($589) + ($596))|0;
 $599 = ($598>>>0)<($596>>>0);
 $600 = $599&1;
 $$57 = (($600) + ($597))|0;
 $601 = (($592) + ($$57))|0;
 $602 = ($601>>>0)<($$57>>>0);
 $603 = $593 ? 2 : 1;
 $c3$19 = $602 ? $603 : $c3$18;
 $604 = HEAP32[$199>>2]|0;
 $605 = HEAP32[$284>>2]|0;
 $606 = (___muldi3(($605|0),0,($604|0),0)|0);
 $607 = tempRet0;
 $608 = (($598) + ($606))|0;
 $609 = ($608>>>0)<($606>>>0);
 $610 = $609&1;
 $$58 = (($610) + ($607))|0;
 $611 = (($601) + ($$58))|0;
 $612 = ($611>>>0)<($$58>>>0);
 $613 = $612&1;
 $c3$20 = (($613) + ($c3$19))|0;
 $614 = ((($r)) + 48|0);
 HEAP32[$614>>2] = $608;
 $615 = HEAP32[$211>>2]|0;
 $616 = HEAP32[$284>>2]|0;
 $617 = (___muldi3(($616|0),0,($615|0),0)|0);
 $618 = tempRet0;
 $619 = (($611) + ($617))|0;
 $620 = ($619>>>0)<($617>>>0);
 $621 = $620&1;
 $$59 = (($621) + ($618))|0;
 $622 = (($c3$20) + ($$59))|0;
 $623 = ($622>>>0)<($$59>>>0);
 $c1$20 = $623&1;
 $624 = HEAP32[$353>>2]|0;
 $625 = HEAP32[$272>>2]|0;
 $626 = (___muldi3(($625|0),0,($624|0),0)|0);
 $627 = tempRet0;
 $628 = (($619) + ($626))|0;
 $629 = ($628>>>0)<($626>>>0);
 $630 = $629&1;
 $$60 = (($630) + ($627))|0;
 $631 = (($622) + ($$60))|0;
 $632 = ($631>>>0)<($$60>>>0);
 $633 = $623 ? 2 : 1;
 $c1$21 = $632 ? $633 : $c1$20;
 $634 = ((($r)) + 52|0);
 HEAP32[$634>>2] = $628;
 $635 = HEAP32[$353>>2]|0;
 $636 = HEAP32[$284>>2]|0;
 $637 = (___muldi3(($636|0),0,($635|0),0)|0);
 $638 = tempRet0;
 $639 = (($631) + ($637))|0;
 $640 = ($639>>>0)<($637>>>0);
 $641 = $640&1;
 $$61 = (($c1$21) + ($638))|0;
 $642 = (($$61) + ($641))|0;
 $643 = ((($r)) + 56|0);
 HEAP32[$643>>2] = $639;
 $644 = ((($r)) + 60|0);
 HEAP32[$644>>2] = $642;
 return;
}
function _egbn_mul_recursive($r,$a,$b,$n2,$t) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 $n2 = $n2|0;
 $t = $t|0;
 var $$pre$phi1Z2D = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $c1$0 = 0, $neg$0 = 0, $p$0 = 0, $zero$0 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = (($n2|0) / 2)&-1;
 $1 = ($n2|0)==(8);
 if ($1) {
  _egbn_mul_comba8($r,$a,$b);
  return;
 }
 $2 = ($n2|0)<(16);
 if ($2) {
  _egbn_mul_normal($r,$a,$n2,$b,$n2);
  return;
 }
 $3 = (($a) + ($0<<2)|0);
 $4 = (_egbn_cmp_words($a,$3,$0)|0);
 $5 = (($b) + ($0<<2)|0);
 $6 = (_egbn_cmp_words($5,$b,$0)|0);
 $7 = ($4*3)|0;
 $8 = (($7) + ($6))|0;
 switch ($8|0) {
 case -4:  {
  (_egbn_sub_words($t,$3,$a,$0)|0);
  $9 = (($t) + ($0<<2)|0);
  (_egbn_sub_words($9,$b,$5,$0)|0);
  $neg$0 = 0;$zero$0 = 0;
  break;
 }
 case -3:  {
  $neg$0 = 0;$zero$0 = 1;
  break;
 }
 case -2:  {
  (_egbn_sub_words($t,$3,$a,$0)|0);
  $10 = (($t) + ($0<<2)|0);
  (_egbn_sub_words($10,$5,$b,$0)|0);
  $neg$0 = 1;$zero$0 = 0;
  break;
 }
 case 1: case 0: case -1:  {
  $neg$0 = 0;$zero$0 = 1;
  break;
 }
 case 2:  {
  (_egbn_sub_words($t,$a,$3,$0)|0);
  $11 = (($t) + ($0<<2)|0);
  (_egbn_sub_words($11,$b,$5,$0)|0);
  $neg$0 = 1;$zero$0 = 0;
  break;
 }
 case 3:  {
  $neg$0 = 0;$zero$0 = 1;
  break;
 }
 case 4:  {
  (_egbn_sub_words($t,$a,$3,$0)|0);
  $12 = (($t) + ($0<<2)|0);
  (_egbn_sub_words($12,$5,$b,$0)|0);
  $neg$0 = 0;$zero$0 = 0;
  break;
 }
 default: {
  $neg$0 = 0;$zero$0 = 0;
 }
 }
 $13 = $n2 & -2;
 switch ($13|0) {
 case 8:  {
  $14 = ($zero$0|0)==(0);
  $15 = (($t) + ($n2<<2)|0);
  if ($14) {
   $16 = (($t) + ($0<<2)|0);
   _egbn_mul_comba4($15,$t,$16);
  } else {
   ;HEAP32[$15>>2]=0|0;HEAP32[$15+4>>2]=0|0;HEAP32[$15+8>>2]=0|0;HEAP32[$15+12>>2]=0|0;HEAP32[$15+16>>2]=0|0;HEAP32[$15+20>>2]=0|0;HEAP32[$15+24>>2]=0|0;HEAP32[$15+28>>2]=0|0;
  }
  _egbn_mul_comba4($r,$a,$b);
  $17 = (($r) + ($n2<<2)|0);
  _egbn_mul_comba4($17,$3,$5);
  $$pre$phi1Z2D = $15;$$pre$phiZ2D = $17;
  break;
 }
 case 16:  {
  $18 = ($zero$0|0)==(0);
  $19 = (($t) + ($n2<<2)|0);
  if ($18) {
   $20 = (($t) + ($0<<2)|0);
   _egbn_mul_comba8($19,$t,$20);
  } else {
   dest=$19; stop=dest+64|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
  }
  _egbn_mul_comba8($r,$a,$b);
  $21 = (($r) + ($n2<<2)|0);
  _egbn_mul_comba8($21,$3,$5);
  $$pre$phi1Z2D = $19;$$pre$phiZ2D = $21;
  break;
 }
 default: {
  $22 = $n2 << 1;
  $23 = (($t) + ($22<<2)|0);
  $24 = ($zero$0|0)==(0);
  $25 = (($t) + ($n2<<2)|0);
  if ($24) {
   $26 = (($t) + ($0<<2)|0);
   _egbn_mul_recursive($25,$t,$26,$0,$23);
  } else {
   $27 = $n2 << 2;
   _memset(($25|0),0,($27|0))|0;
  }
  _egbn_mul_recursive($r,$a,$b,$0,$23);
  $28 = (($r) + ($n2<<2)|0);
  _egbn_mul_recursive($28,$3,$5,$0,$23);
  $$pre$phi1Z2D = $25;$$pre$phiZ2D = $28;
 }
 }
 $29 = (_egbn_add_words($t,$r,$$pre$phiZ2D,$n2)|0);
 $30 = ($neg$0|0)==(0);
 if ($30) {
  $33 = (_egbn_add_words($$pre$phi1Z2D,$$pre$phi1Z2D,$t,$n2)|0);
  $34 = (($33) + ($29))|0;
  $c1$0 = $34;
 } else {
  $31 = (_egbn_sub_words($$pre$phi1Z2D,$t,$$pre$phi1Z2D,$n2)|0);
  $32 = (($29) - ($31))|0;
  $c1$0 = $32;
 }
 $35 = (($r) + ($0<<2)|0);
 $36 = (_egbn_add_words($35,$35,$$pre$phi1Z2D,$n2)|0);
 $37 = (($36) + ($c1$0))|0;
 $38 = ($37|0)==(0);
 if ($38) {
  return;
 }
 $39 = (($0) + ($n2))|0;
 $40 = (($r) + ($39<<2)|0);
 $41 = HEAP32[$40>>2]|0;
 $42 = (($41) + ($37))|0;
 HEAP32[$40>>2] = $42;
 $43 = ($42>>>0)<($37>>>0);
 if ($43) {
  $p$0 = $40;
 } else {
  return;
 }
 while(1) {
  $44 = ((($p$0)) + 4|0);
  $45 = HEAP32[$44>>2]|0;
  $46 = (($45) + 1)|0;
  HEAP32[$44>>2] = $46;
  $47 = ($46|0)==(0);
  if ($47) {
   $p$0 = $44;
  } else {
   break;
  }
 }
 return;
}
function _egbn_mul_normal($r,$a,$na,$b,$nb) {
 $r = $r|0;
 $a = $a|0;
 $na = $na|0;
 $b = $b|0;
 $nb = $nb|0;
 var $$09 = 0, $$146 = 0, $$17 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $a$b = 0, $b$a = 0, $na$nb = 0, $nb$na = 0, $rr$08 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($na|0)<($nb|0);
 $nb$na = $0 ? $nb : $na;
 $a$b = $0 ? $a : $b;
 $na$nb = $0 ? $na : $nb;
 $b$a = $0 ? $b : $a;
 $1 = (($r) + ($nb$na<<2)|0);
 $2 = HEAP32[$a$b>>2]|0;
 $3 = (_egbn_mul_words($r,$b$a,$nb$na,$2)|0);
 HEAP32[$1>>2] = $3;
 $4 = ($na$nb|0)<(2);
 if ($4) {
  return;
 } else {
  $$09 = $r;$$146 = $a$b;$$17 = $na$nb;$rr$08 = $1;
 }
 while(1) {
  $5 = ((($$09)) + 4|0);
  $6 = ((($$146)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (_egbn_mul_add_words($5,$b$a,$nb$na,$7)|0);
  $9 = ((($rr$08)) + 4|0);
  HEAP32[$9>>2] = $8;
  $10 = (($$17) + -2)|0;
  $11 = ($10|0)<(1);
  if ($11) {
   label = 6;
   break;
  }
  $12 = ((($$09)) + 8|0);
  $13 = ((($$146)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (_egbn_mul_add_words($12,$b$a,$nb$na,$14)|0);
  $16 = ((($rr$08)) + 8|0);
  HEAP32[$16>>2] = $15;
  $17 = (($$17) + -3)|0;
  $18 = ($17|0)<(1);
  if ($18) {
   label = 6;
   break;
  }
  $19 = ((($$09)) + 12|0);
  $20 = ((($$146)) + 12|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (_egbn_mul_add_words($19,$b$a,$nb$na,$21)|0);
  $23 = ((($rr$08)) + 12|0);
  HEAP32[$23>>2] = $22;
  $24 = (($$17) + -4)|0;
  $25 = ($24|0)<(1);
  if ($25) {
   label = 6;
   break;
  }
  $26 = ((($$09)) + 16|0);
  $27 = ((($$146)) + 16|0);
  $28 = HEAP32[$27>>2]|0;
  $29 = (_egbn_mul_add_words($26,$b$a,$nb$na,$28)|0);
  $30 = ((($rr$08)) + 16|0);
  HEAP32[$30>>2] = $29;
  $31 = ($24|0)<(2);
  if ($31) {
   label = 6;
   break;
  } else {
   $$09 = $26;$$146 = $27;$$17 = $24;$rr$08 = $30;
  }
 }
 if ((label|0) == 6) {
  return;
 }
}
function _egbn_mul_add_words($rp,$ap,$num,$w) {
 $rp = $rp|0;
 $ap = $ap|0;
 $num = $num|0;
 $w = $w|0;
 var $$0 = 0, $$01$lcssa15 = 0, $$017 = 0, $$02$lcssa16 = 0, $$026 = 0, $$03$lcssa17 = 0, $$035 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $c1$0$lcssa18 = 0, $c1$04 = 0, $scevgep = 0, $scevgep14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($num|0)<(1);
 if ($0) {
  $$0 = 0;
  return ($$0|0);
 }
 $1 = ($num>>>0)>(3);
 if ($1) {
  $2 = (($num) + -4)|0;
  $3 = $2 & -4;
  $4 = (($3) + 4)|0;
  $scevgep = (($rp) + ($4<<2)|0);
  $$017 = $rp;$$026 = $ap;$$035 = $num;$c1$04 = 0;
  while(1) {
   $5 = HEAP32[$$026>>2]|0;
   $6 = (___muldi3(($5|0),0,($w|0),0)|0);
   $7 = tempRet0;
   $8 = HEAP32[$$017>>2]|0;
   $9 = (_i64Add(($8|0),0,($c1$04|0),0)|0);
   $10 = tempRet0;
   $11 = (_i64Add(($9|0),($10|0),($6|0),($7|0))|0);
   $12 = tempRet0;
   HEAP32[$$017>>2] = $11;
   $13 = ((($$026)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = (___muldi3(($14|0),0,($w|0),0)|0);
   $16 = tempRet0;
   $17 = ((($$017)) + 4|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = (_i64Add(($15|0),($16|0),($18|0),0)|0);
   $20 = tempRet0;
   $21 = (_i64Add(($19|0),($20|0),($12|0),0)|0);
   $22 = tempRet0;
   HEAP32[$17>>2] = $21;
   $23 = ((($$026)) + 8|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = (___muldi3(($24|0),0,($w|0),0)|0);
   $26 = tempRet0;
   $27 = ((($$017)) + 8|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (_i64Add(($25|0),($26|0),($28|0),0)|0);
   $30 = tempRet0;
   $31 = (_i64Add(($29|0),($30|0),($22|0),0)|0);
   $32 = tempRet0;
   HEAP32[$27>>2] = $31;
   $33 = ((($$026)) + 12|0);
   $34 = HEAP32[$33>>2]|0;
   $35 = (___muldi3(($34|0),0,($w|0),0)|0);
   $36 = tempRet0;
   $37 = ((($$017)) + 12|0);
   $38 = HEAP32[$37>>2]|0;
   $39 = (_i64Add(($35|0),($36|0),($38|0),0)|0);
   $40 = tempRet0;
   $41 = (_i64Add(($39|0),($40|0),($32|0),0)|0);
   $42 = tempRet0;
   HEAP32[$37>>2] = $41;
   $43 = ((($$026)) + 16|0);
   $44 = ((($$017)) + 16|0);
   $45 = (($$035) + -4)|0;
   $46 = ($45>>>0)>(3);
   if ($46) {
    $$017 = $44;$$026 = $43;$$035 = $45;$c1$04 = $42;
   } else {
    $$lcssa = $42;
    break;
   }
  }
  $47 = (($2) - ($3))|0;
  $scevgep14 = (($ap) + ($4<<2)|0);
  $48 = ($2|0)==($3|0);
  if ($48) {
   $$0 = $$lcssa;
   return ($$0|0);
  } else {
   $$01$lcssa15 = $scevgep;$$02$lcssa16 = $scevgep14;$$03$lcssa17 = $47;$c1$0$lcssa18 = $$lcssa;
  }
 } else {
  $$01$lcssa15 = $rp;$$02$lcssa16 = $ap;$$03$lcssa17 = $num;$c1$0$lcssa18 = 0;
 }
 $49 = HEAP32[$$02$lcssa16>>2]|0;
 $50 = (___muldi3(($49|0),0,($w|0),0)|0);
 $51 = tempRet0;
 $52 = HEAP32[$$01$lcssa15>>2]|0;
 $53 = (_i64Add(($52|0),0,($c1$0$lcssa18|0),0)|0);
 $54 = tempRet0;
 $55 = (_i64Add(($53|0),($54|0),($50|0),($51|0))|0);
 $56 = tempRet0;
 HEAP32[$$01$lcssa15>>2] = $55;
 $57 = ($$03$lcssa17|0)==(1);
 if ($57) {
  $$0 = $56;
  return ($$0|0);
 }
 $58 = ((($$02$lcssa16)) + 4|0);
 $59 = HEAP32[$58>>2]|0;
 $60 = (___muldi3(($59|0),0,($w|0),0)|0);
 $61 = tempRet0;
 $62 = ((($$01$lcssa15)) + 4|0);
 $63 = HEAP32[$62>>2]|0;
 $64 = (_i64Add(($63|0),0,($56|0),0)|0);
 $65 = tempRet0;
 $66 = (_i64Add(($64|0),($65|0),($60|0),($61|0))|0);
 $67 = tempRet0;
 HEAP32[$62>>2] = $66;
 $68 = ($$03$lcssa17|0)==(2);
 if ($68) {
  $$0 = $67;
  return ($$0|0);
 } else {
  $69 = ((($$02$lcssa16)) + 8|0);
  $70 = HEAP32[$69>>2]|0;
  $71 = (___muldi3(($70|0),0,($w|0),0)|0);
  $72 = tempRet0;
  $73 = ((($$01$lcssa15)) + 8|0);
  $74 = HEAP32[$73>>2]|0;
  $75 = (_i64Add(($74|0),0,($67|0),0)|0);
  $76 = tempRet0;
  $77 = (_i64Add(($75|0),($76|0),($71|0),($72|0))|0);
  $78 = tempRet0;
  HEAP32[$73>>2] = $77;
  $$0 = $78;
  return ($$0|0);
 }
 return (0)|0;
}
function _egbn_cmp_words($a,$b,$n) {
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$0 = 0, $$lcssa = 0, $$lcssa16 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$02 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($n) + -1)|0;
 $1 = (($a) + ($0<<2)|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (($b) + ($0<<2)|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2|0)==($4|0);
 if (!($5)) {
  $6 = ($2>>>0)>($4>>>0);
  $7 = $6 ? 1 : -1;
  $$0 = $7;
  return ($$0|0);
 }
 $8 = (($n) + -2)|0;
 $9 = ($8|0)>(-1);
 if ($9) {
  $i$02 = $8;
 } else {
  $$0 = 0;
  return ($$0|0);
 }
 while(1) {
  $12 = (($a) + ($i$02<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($b) + ($i$02<<2)|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ($13|0)==($15|0);
  if (!($16)) {
   $$lcssa = $13;$$lcssa16 = $15;
   break;
  }
  $10 = (($i$02) + -1)|0;
  $11 = ($i$02|0)>(0);
  if ($11) {
   $i$02 = $10;
  } else {
   $$0 = 0;
   label = 7;
   break;
  }
 }
 if ((label|0) == 7) {
  return ($$0|0);
 }
 $17 = ($$lcssa>>>0)>($$lcssa16>>>0);
 $18 = $17 ? 1 : -1;
 $$0 = $18;
 return ($$0|0);
}
function _egbn_sub_words($r,$a,$b,$n) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 $n = $n|0;
 var $$0112 = 0, $$013 = 0, $$0211 = 0, $$0310 = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $c$1 = 0, $c$114 = 0, $c$19 = 0, $c$2 = 0, $c$3 = 0, $c$4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(1);
 if ($0) {
  $$04 = 0;
  return ($$04|0);
 }
 $1 = HEAP32[$a>>2]|0;
 $2 = HEAP32[$b>>2]|0;
 $3 = (($1) - ($2))|0;
 HEAP32[$r>>2] = $3;
 $4 = ($1|0)==($2|0);
 $5 = ($1>>>0)<($2>>>0);
 $6 = $5&1;
 $c$19 = $4 ? 0 : $6;
 $7 = ($n|0)<(2);
 if ($7) {
  $$04 = $c$19;
  return ($$04|0);
 } else {
  $$0112 = $b;$$013 = $n;$$0211 = $a;$$0310 = $r;$c$114 = $c$19;
 }
 while(1) {
  $8 = ((($$0211)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($$0112)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = (($9) - ($c$114))|0;
  $13 = (($12) - ($11))|0;
  $14 = ((($$0310)) + 4|0);
  HEAP32[$14>>2] = $13;
  $15 = ($9|0)==($11|0);
  $16 = ($9>>>0)<($11>>>0);
  $17 = $16&1;
  $c$2 = $15 ? $c$114 : $17;
  $18 = (($$013) + -2)|0;
  $19 = ($18|0)<(1);
  if ($19) {
   $$04 = $c$2;
   label = 7;
   break;
  }
  $20 = ((($$0211)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = ((($$0112)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = (($21) - ($c$2))|0;
  $25 = (($24) - ($23))|0;
  $26 = ((($$0310)) + 8|0);
  HEAP32[$26>>2] = $25;
  $27 = ($21|0)==($23|0);
  $28 = ($21>>>0)<($23>>>0);
  $29 = $28&1;
  $c$3 = $27 ? $c$2 : $29;
  $30 = (($$013) + -3)|0;
  $31 = ($30|0)<(1);
  if ($31) {
   $$04 = $c$3;
   label = 7;
   break;
  }
  $32 = ((($$0211)) + 12|0);
  $33 = HEAP32[$32>>2]|0;
  $34 = ((($$0112)) + 12|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = (($33) - ($c$3))|0;
  $37 = (($36) - ($35))|0;
  $38 = ((($$0310)) + 12|0);
  HEAP32[$38>>2] = $37;
  $39 = ($33|0)==($35|0);
  $40 = ($33>>>0)<($35>>>0);
  $41 = $40&1;
  $c$4 = $39 ? $c$3 : $41;
  $42 = (($$013) + -4)|0;
  $43 = ($42|0)<(1);
  if ($43) {
   $$04 = $c$4;
   label = 7;
   break;
  }
  $44 = ((($$0211)) + 16|0);
  $45 = ((($$0112)) + 16|0);
  $46 = ((($$0310)) + 16|0);
  $47 = HEAP32[$44>>2]|0;
  $48 = HEAP32[$45>>2]|0;
  $49 = (($47) - ($c$4))|0;
  $50 = (($49) - ($48))|0;
  HEAP32[$46>>2] = $50;
  $51 = ($47|0)==($48|0);
  $52 = ($47>>>0)<($48>>>0);
  $53 = $52&1;
  $c$1 = $51 ? $c$4 : $53;
  $54 = ($42|0)<(2);
  if ($54) {
   $$04 = $c$1;
   label = 7;
   break;
  } else {
   $$0112 = $45;$$013 = $42;$$0211 = $44;$$0310 = $46;$c$114 = $c$1;
  }
 }
 if ((label|0) == 7) {
  return ($$04|0);
 }
 return (0)|0;
}
function _egbn_mul_comba4($r,$a,$b) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 var $$ = 0, $$1 = 0, $$10 = 0, $$11 = 0, $$12 = 0, $$13 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$6 = 0, $$7 = 0, $$8 = 0, $$9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c1$1 = 0, $c1$2 = 0, $c1$3 = 0, $c1$4 = 0, $c2$0 = 0, $c2$1 = 0, $c2$2 = 0, $c2$3 = 0, $c2$4 = 0;
 var $c3$1 = 0, $c3$2 = 0, $c3$3 = 0, $c3$4 = 0, $t2$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = HEAP32[$b>>2]|0;
 $2 = (___muldi3(($1|0),0,($0|0),0)|0);
 $3 = tempRet0;
 HEAP32[$r>>2] = $2;
 $4 = HEAP32[$a>>2]|0;
 $5 = ((($b)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (___muldi3(($6|0),0,($4|0),0)|0);
 $8 = tempRet0;
 $9 = (($7) + ($3))|0;
 $10 = ($9>>>0)<($7>>>0);
 $11 = $10&1;
 $t2$1 = (($11) + ($8))|0;
 $12 = ((($a)) + 4|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = HEAP32[$b>>2]|0;
 $15 = (___muldi3(($14|0),0,($13|0),0)|0);
 $16 = tempRet0;
 $17 = (($15) + ($9))|0;
 $18 = ($17>>>0)<($15>>>0);
 $19 = $18&1;
 $$ = (($19) + ($16))|0;
 $20 = (($t2$1) + ($$))|0;
 $21 = ($20>>>0)<($$>>>0);
 $c1$1 = $21&1;
 $22 = ((($r)) + 4|0);
 HEAP32[$22>>2] = $17;
 $23 = ((($a)) + 8|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = HEAP32[$b>>2]|0;
 $26 = (___muldi3(($25|0),0,($24|0),0)|0);
 $27 = tempRet0;
 $28 = (($20) + ($26))|0;
 $29 = ($28>>>0)<($26>>>0);
 $30 = $29&1;
 $$1 = (($30) + ($27))|0;
 $31 = (($$1) + ($c1$1))|0;
 $32 = ($31>>>0)<($$1>>>0);
 $c2$0 = $32&1;
 $33 = HEAP32[$12>>2]|0;
 $34 = HEAP32[$5>>2]|0;
 $35 = (___muldi3(($34|0),0,($33|0),0)|0);
 $36 = tempRet0;
 $37 = (($28) + ($35))|0;
 $38 = ($37>>>0)<($35>>>0);
 $39 = $38&1;
 $$2 = (($39) + ($36))|0;
 $40 = (($31) + ($$2))|0;
 $41 = ($40>>>0)<($$2>>>0);
 $42 = $32 ? 2 : 1;
 $c2$1 = $41 ? $42 : $c2$0;
 $43 = HEAP32[$a>>2]|0;
 $44 = ((($b)) + 8|0);
 $45 = HEAP32[$44>>2]|0;
 $46 = (___muldi3(($45|0),0,($43|0),0)|0);
 $47 = tempRet0;
 $48 = (($37) + ($46))|0;
 $49 = ($48>>>0)<($46>>>0);
 $50 = $49&1;
 $$3 = (($50) + ($47))|0;
 $51 = (($40) + ($$3))|0;
 $52 = ($51>>>0)<($$3>>>0);
 $53 = $52&1;
 $c2$2 = (($53) + ($c2$1))|0;
 $54 = ((($r)) + 8|0);
 HEAP32[$54>>2] = $48;
 $55 = HEAP32[$a>>2]|0;
 $56 = ((($b)) + 12|0);
 $57 = HEAP32[$56>>2]|0;
 $58 = (___muldi3(($57|0),0,($55|0),0)|0);
 $59 = tempRet0;
 $60 = (($51) + ($58))|0;
 $61 = ($60>>>0)<($58>>>0);
 $62 = $61&1;
 $$4 = (($62) + ($59))|0;
 $63 = (($c2$2) + ($$4))|0;
 $64 = ($63>>>0)<($$4>>>0);
 $c3$1 = $64&1;
 $65 = HEAP32[$12>>2]|0;
 $66 = HEAP32[$44>>2]|0;
 $67 = (___muldi3(($66|0),0,($65|0),0)|0);
 $68 = tempRet0;
 $69 = (($60) + ($67))|0;
 $70 = ($69>>>0)<($67>>>0);
 $71 = $70&1;
 $$5 = (($71) + ($68))|0;
 $72 = (($63) + ($$5))|0;
 $73 = ($72>>>0)<($$5>>>0);
 $74 = $64 ? 2 : 1;
 $c3$2 = $73 ? $74 : $c3$1;
 $75 = HEAP32[$23>>2]|0;
 $76 = HEAP32[$5>>2]|0;
 $77 = (___muldi3(($76|0),0,($75|0),0)|0);
 $78 = tempRet0;
 $79 = (($69) + ($77))|0;
 $80 = ($79>>>0)<($77>>>0);
 $81 = $80&1;
 $$6 = (($81) + ($78))|0;
 $82 = (($72) + ($$6))|0;
 $83 = ($82>>>0)<($$6>>>0);
 $84 = $83&1;
 $c3$3 = (($84) + ($c3$2))|0;
 $85 = ((($a)) + 12|0);
 $86 = HEAP32[$85>>2]|0;
 $87 = HEAP32[$b>>2]|0;
 $88 = (___muldi3(($87|0),0,($86|0),0)|0);
 $89 = tempRet0;
 $90 = (($79) + ($88))|0;
 $91 = ($90>>>0)<($88>>>0);
 $92 = $91&1;
 $$7 = (($92) + ($89))|0;
 $93 = (($82) + ($$7))|0;
 $94 = ($93>>>0)<($$7>>>0);
 $95 = $94&1;
 $c3$4 = (($c3$3) + ($95))|0;
 $96 = ((($r)) + 12|0);
 HEAP32[$96>>2] = $90;
 $97 = HEAP32[$85>>2]|0;
 $98 = HEAP32[$5>>2]|0;
 $99 = (___muldi3(($98|0),0,($97|0),0)|0);
 $100 = tempRet0;
 $101 = (($93) + ($99))|0;
 $102 = ($101>>>0)<($99>>>0);
 $103 = $102&1;
 $$8 = (($103) + ($100))|0;
 $104 = (($c3$4) + ($$8))|0;
 $105 = ($104>>>0)<($$8>>>0);
 $c1$2 = $105&1;
 $106 = HEAP32[$23>>2]|0;
 $107 = HEAP32[$44>>2]|0;
 $108 = (___muldi3(($107|0),0,($106|0),0)|0);
 $109 = tempRet0;
 $110 = (($101) + ($108))|0;
 $111 = ($110>>>0)<($108>>>0);
 $112 = $111&1;
 $$9 = (($112) + ($109))|0;
 $113 = (($104) + ($$9))|0;
 $114 = ($113>>>0)<($$9>>>0);
 $115 = $105 ? 2 : 1;
 $c1$3 = $114 ? $115 : $c1$2;
 $116 = HEAP32[$12>>2]|0;
 $117 = HEAP32[$56>>2]|0;
 $118 = (___muldi3(($117|0),0,($116|0),0)|0);
 $119 = tempRet0;
 $120 = (($110) + ($118))|0;
 $121 = ($120>>>0)<($118>>>0);
 $122 = $121&1;
 $$10 = (($122) + ($119))|0;
 $123 = (($113) + ($$10))|0;
 $124 = ($123>>>0)<($$10>>>0);
 $125 = $124&1;
 $c1$4 = (($125) + ($c1$3))|0;
 $126 = ((($r)) + 16|0);
 HEAP32[$126>>2] = $120;
 $127 = HEAP32[$23>>2]|0;
 $128 = HEAP32[$56>>2]|0;
 $129 = (___muldi3(($128|0),0,($127|0),0)|0);
 $130 = tempRet0;
 $131 = (($123) + ($129))|0;
 $132 = ($131>>>0)<($129>>>0);
 $133 = $132&1;
 $$11 = (($133) + ($130))|0;
 $134 = (($c1$4) + ($$11))|0;
 $135 = ($134>>>0)<($$11>>>0);
 $c2$3 = $135&1;
 $136 = HEAP32[$85>>2]|0;
 $137 = HEAP32[$44>>2]|0;
 $138 = (___muldi3(($137|0),0,($136|0),0)|0);
 $139 = tempRet0;
 $140 = (($131) + ($138))|0;
 $141 = ($140>>>0)<($138>>>0);
 $142 = $141&1;
 $$12 = (($142) + ($139))|0;
 $143 = (($134) + ($$12))|0;
 $144 = ($143>>>0)<($$12>>>0);
 $145 = $135 ? 2 : 1;
 $c2$4 = $144 ? $145 : $c2$3;
 $146 = ((($r)) + 20|0);
 HEAP32[$146>>2] = $140;
 $147 = HEAP32[$85>>2]|0;
 $148 = HEAP32[$56>>2]|0;
 $149 = (___muldi3(($148|0),0,($147|0),0)|0);
 $150 = tempRet0;
 $151 = (($143) + ($149))|0;
 $152 = ($151>>>0)<($149>>>0);
 $153 = $152&1;
 $$13 = (($c2$4) + ($150))|0;
 $154 = (($$13) + ($153))|0;
 $155 = ((($r)) + 24|0);
 HEAP32[$155>>2] = $151;
 $156 = ((($r)) + 28|0);
 HEAP32[$156>>2] = $154;
 return;
}
function _egbn_sqr_comba4($r,$a) {
 $r = $r|0;
 $a = $a|0;
 var $$ = 0, $$$c2$0 = 0, $$1 = 0, $$c1$0 = 0, $$c1$4 = 0, $$c1$5 = 0, $$c2$0 = 0, $$c2$2 = 0, $$c2$4 = 0, $$c2$5 = 0, $$c3$1 = 0, $$c3$2 = 0, $$c3$4 = 0, $$c3$5 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c1$0 = 0, $c1$1 = 0, $c1$4 = 0, $c1$5 = 0;
 var $c2$0 = 0, $c2$2 = 0, $c2$4 = 0, $c2$5 = 0, $c3$1 = 0, $c3$2 = 0, $c3$4 = 0, $c3$5 = 0, $t2$1 = 0, $t2$3 = 0, $t2$4 = 0, $t2$5 = 0, $t2$6 = 0, $t2$7 = 0, $t2$8 = 0, $t2$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = (___muldi3(($0|0),0,($0|0),0)|0);
 $2 = tempRet0;
 HEAP32[$r>>2] = $1;
 $3 = ((($a)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$a>>2]|0;
 $6 = (___muldi3(($5|0),0,($4|0),0)|0);
 $7 = tempRet0;
 $8 = (_bitshift64Shl(($6|0),($7|0),1)|0);
 $9 = tempRet0;
 $10 = ($9>>>0)<($7>>>0);
 $11 = ($8>>>0)<($6>>>0);
 $12 = ($9|0)==($7|0);
 $13 = $12 & $11;
 $14 = $10 | $13;
 $c1$0 = $14&1;
 $15 = (_bitshift64Lshr(($6|0),($7|0),31)|0);
 $16 = tempRet0;
 $17 = (($8) + ($2))|0;
 $18 = ($17>>>0)<($8>>>0);
 if ($18) {
  $19 = (($15) + 1)|0;
  $20 = ($19|0)==(0);
  $21 = $14 ? 2 : 1;
  $$c1$0 = $20 ? $21 : $c1$0;
  $c1$1 = $$c1$0;$t2$1 = $19;
 } else {
  $c1$1 = $c1$0;$t2$1 = $15;
 }
 $22 = ((($r)) + 4|0);
 HEAP32[$22>>2] = $17;
 $23 = HEAP32[$3>>2]|0;
 $24 = (___muldi3(($23|0),0,($23|0),0)|0);
 $25 = tempRet0;
 $26 = (($24) + ($t2$1))|0;
 $27 = ($26>>>0)<($24>>>0);
 $28 = $27&1;
 $$ = (($28) + ($25))|0;
 $29 = (($$) + ($c1$1))|0;
 $30 = ($29>>>0)<($$>>>0);
 $c2$0 = $30&1;
 $31 = ((($a)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = HEAP32[$a>>2]|0;
 $34 = (___muldi3(($33|0),0,($32|0),0)|0);
 $35 = tempRet0;
 $36 = (_bitshift64Shl(($34|0),($35|0),1)|0);
 $37 = tempRet0;
 $38 = ($37>>>0)<($35>>>0);
 $39 = ($36>>>0)<($34>>>0);
 $40 = ($37|0)==($35|0);
 $41 = $40 & $39;
 $42 = $38 | $41;
 $43 = $30 ? 2 : 1;
 $$c2$0 = $42 ? $43 : $c2$0;
 $44 = (_bitshift64Lshr(($34|0),($35|0),31)|0);
 $45 = tempRet0;
 $46 = (($36) + ($26))|0;
 $47 = ($46>>>0)<($36>>>0);
 if ($47) {
  $48 = (($44) + 1)|0;
  $49 = ($48|0)==(0);
  $50 = $49&1;
  $$$c2$0 = (($$c2$0) + ($50))|0;
  $c2$2 = $$$c2$0;$t2$3 = $48;
 } else {
  $c2$2 = $$c2$0;$t2$3 = $44;
 }
 $51 = (($t2$3) + ($29))|0;
 $52 = ($51>>>0)<($t2$3>>>0);
 $53 = $52&1;
 $$c2$2 = (($53) + ($c2$2))|0;
 $54 = ((($r)) + 8|0);
 HEAP32[$54>>2] = $46;
 $55 = ((($a)) + 12|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = HEAP32[$a>>2]|0;
 $58 = (___muldi3(($57|0),0,($56|0),0)|0);
 $59 = tempRet0;
 $60 = (_bitshift64Shl(($58|0),($59|0),1)|0);
 $61 = tempRet0;
 $62 = ($61>>>0)<($59>>>0);
 $63 = ($60>>>0)<($58>>>0);
 $64 = ($61|0)==($59|0);
 $65 = $64 & $63;
 $66 = $62 | $65;
 $c3$1 = $66&1;
 $67 = (_bitshift64Lshr(($58|0),($59|0),31)|0);
 $68 = tempRet0;
 $69 = (($60) + ($51))|0;
 $70 = ($69>>>0)<($60>>>0);
 if ($70) {
  $71 = (($67) + 1)|0;
  $72 = ($71|0)==(0);
  $73 = $66 ? 2 : 1;
  $$c3$1 = $72 ? $73 : $c3$1;
  $c3$2 = $$c3$1;$t2$4 = $71;
 } else {
  $c3$2 = $c3$1;$t2$4 = $67;
 }
 $74 = (($$c2$2) + ($t2$4))|0;
 $75 = ($74>>>0)<($t2$4>>>0);
 $76 = $75&1;
 $$c3$2 = (($76) + ($c3$2))|0;
 $77 = HEAP32[$31>>2]|0;
 $78 = HEAP32[$3>>2]|0;
 $79 = (___muldi3(($78|0),0,($77|0),0)|0);
 $80 = tempRet0;
 $81 = (_bitshift64Shl(($79|0),($80|0),1)|0);
 $82 = tempRet0;
 $83 = ($82>>>0)<($80>>>0);
 $84 = ($81>>>0)<($79>>>0);
 $85 = ($82|0)==($80|0);
 $86 = $85 & $84;
 $87 = $83 | $86;
 $88 = $87&1;
 $c3$4 = (($$c3$2) + ($88))|0;
 $89 = (_bitshift64Lshr(($79|0),($80|0),31)|0);
 $90 = tempRet0;
 $91 = (($81) + ($69))|0;
 $92 = ($91>>>0)<($81>>>0);
 if ($92) {
  $93 = (($89) + 1)|0;
  $94 = ($93|0)==(0);
  $95 = $94&1;
  $$c3$4 = (($95) + ($c3$4))|0;
  $c3$5 = $$c3$4;$t2$5 = $93;
 } else {
  $c3$5 = $c3$4;$t2$5 = $89;
 }
 $96 = (($t2$5) + ($74))|0;
 $97 = ($96>>>0)<($t2$5>>>0);
 $98 = $97&1;
 $$c3$5 = (($98) + ($c3$5))|0;
 $99 = ((($r)) + 12|0);
 HEAP32[$99>>2] = $91;
 $100 = HEAP32[$31>>2]|0;
 $101 = (___muldi3(($100|0),0,($100|0),0)|0);
 $102 = tempRet0;
 $103 = (($101) + ($96))|0;
 $104 = ($103>>>0)<($101>>>0);
 $105 = $104&1;
 $t2$6 = (($105) + ($102))|0;
 $106 = (($$c3$5) + ($t2$6))|0;
 $107 = ($106>>>0)<($t2$6>>>0);
 $$1 = $107&1;
 $108 = HEAP32[$55>>2]|0;
 $109 = HEAP32[$3>>2]|0;
 $110 = (___muldi3(($109|0),0,($108|0),0)|0);
 $111 = tempRet0;
 $112 = (_bitshift64Shl(($110|0),($111|0),1)|0);
 $113 = tempRet0;
 $114 = ($113>>>0)<($111>>>0);
 $115 = ($112>>>0)<($110>>>0);
 $116 = ($113|0)==($111|0);
 $117 = $116 & $115;
 $118 = $114 | $117;
 $119 = $107 ? 2 : 1;
 $c1$4 = $118 ? $119 : $$1;
 $120 = (_bitshift64Lshr(($110|0),($111|0),31)|0);
 $121 = tempRet0;
 $122 = (($112) + ($103))|0;
 $123 = ($122>>>0)<($112>>>0);
 if ($123) {
  $124 = (($120) + 1)|0;
  $125 = ($124|0)==(0);
  $126 = $125&1;
  $$c1$4 = (($c1$4) + ($126))|0;
  $c1$5 = $$c1$4;$t2$7 = $124;
 } else {
  $c1$5 = $c1$4;$t2$7 = $120;
 }
 $127 = (($t2$7) + ($106))|0;
 $128 = ($127>>>0)<($t2$7>>>0);
 $129 = $128&1;
 $$c1$5 = (($129) + ($c1$5))|0;
 $130 = ((($r)) + 16|0);
 HEAP32[$130>>2] = $122;
 $131 = HEAP32[$55>>2]|0;
 $132 = HEAP32[$31>>2]|0;
 $133 = (___muldi3(($132|0),0,($131|0),0)|0);
 $134 = tempRet0;
 $135 = (_bitshift64Shl(($133|0),($134|0),1)|0);
 $136 = tempRet0;
 $137 = ($136>>>0)<($134>>>0);
 $138 = ($135>>>0)<($133>>>0);
 $139 = ($136|0)==($134|0);
 $140 = $139 & $138;
 $141 = $137 | $140;
 $c2$4 = $141&1;
 $142 = (_bitshift64Lshr(($133|0),($134|0),31)|0);
 $143 = tempRet0;
 $144 = (($135) + ($127))|0;
 $145 = ($144>>>0)<($135>>>0);
 if (!($145)) {
  $c2$5 = $c2$4;$t2$8 = $142;
  $149 = (($$c1$5) + ($t2$8))|0;
  $150 = ($149>>>0)<($t2$8>>>0);
  $151 = $150&1;
  $152 = ((($r)) + 20|0);
  HEAP32[$152>>2] = $144;
  $153 = HEAP32[$55>>2]|0;
  $154 = (___muldi3(($153|0),0,($153|0),0)|0);
  $155 = tempRet0;
  $156 = (($154) + ($149))|0;
  $157 = ($156>>>0)<($154>>>0);
  $158 = $157&1;
  $t2$9 = (($151) + ($c2$5))|0;
  $$c2$5 = (($t2$9) + ($155))|0;
  $159 = (($$c2$5) + ($158))|0;
  $160 = ((($r)) + 24|0);
  HEAP32[$160>>2] = $156;
  $161 = ((($r)) + 28|0);
  HEAP32[$161>>2] = $159;
  return;
 }
 $146 = (($142) + 1)|0;
 $147 = ($146|0)==(0);
 $148 = $141 ? 2 : 1;
 $$c2$4 = $147 ? $148 : $c2$4;
 $c2$5 = $$c2$4;$t2$8 = $146;
 $149 = (($$c1$5) + ($t2$8))|0;
 $150 = ($149>>>0)<($t2$8>>>0);
 $151 = $150&1;
 $152 = ((($r)) + 20|0);
 HEAP32[$152>>2] = $144;
 $153 = HEAP32[$55>>2]|0;
 $154 = (___muldi3(($153|0),0,($153|0),0)|0);
 $155 = tempRet0;
 $156 = (($154) + ($149))|0;
 $157 = ($156>>>0)<($154>>>0);
 $158 = $157&1;
 $t2$9 = (($151) + ($c2$5))|0;
 $$c2$5 = (($t2$9) + ($155))|0;
 $159 = (($$c2$5) + ($158))|0;
 $160 = ((($r)) + 24|0);
 HEAP32[$160>>2] = $156;
 $161 = ((($r)) + 28|0);
 HEAP32[$161>>2] = $159;
 return;
}
function _egbn_sqr_comba8($r,$a) {
 $r = $r|0;
 $a = $a|0;
 var $$ = 0, $$$c2$0 = 0, $$1 = 0, $$2 = 0, $$3 = 0, $$4 = 0, $$5 = 0, $$c1$0 = 0, $$c1$10 = 0, $$c1$11 = 0, $$c1$13 = 0, $$c1$14 = 0, $$c1$16 = 0, $$c1$17 = 0, $$c1$19 = 0, $$c1$20 = 0, $$c1$23 = 0, $$c1$24 = 0, $$c1$26 = 0, $$c1$27 = 0;
 var $$c1$29 = 0, $$c1$30 = 0, $$c1$4 = 0, $$c1$5 = 0, $$c1$7 = 0, $$c1$8 = 0, $$c2$0 = 0, $$c2$10 = 0, $$c2$11 = 0, $$c2$14 = 0, $$c2$15 = 0, $$c2$17 = 0, $$c2$18 = 0, $$c2$2 = 0, $$c2$20 = 0, $$c2$21 = 0, $$c2$23 = 0, $$c2$24 = 0, $$c2$26 = 0, $$c2$27 = 0;
 var $$c2$4 = 0, $$c2$5 = 0, $$c2$7 = 0, $$c2$8 = 0, $$c3$1 = 0, $$c3$11 = 0, $$c3$12 = 0, $$c3$14 = 0, $$c3$15 = 0, $$c3$17 = 0, $$c3$18 = 0, $$c3$2 = 0, $$c3$20 = 0, $$c3$21 = 0, $$c3$23 = 0, $$c3$24 = 0, $$c3$27 = 0, $$c3$28 = 0, $$c3$4 = 0, $$c3$5 = 0;
 var $$c3$8 = 0, $$c3$9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0;
 var $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0;
 var $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0;
 var $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0;
 var $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0;
 var $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0;
 var $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0;
 var $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0;
 var $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0;
 var $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0;
 var $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0;
 var $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0;
 var $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0;
 var $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0;
 var $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0;
 var $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0;
 var $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0;
 var $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0;
 var $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0;
 var $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0;
 var $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0;
 var $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0;
 var $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0;
 var $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $c1$0 = 0, $c1$1 = 0, $c1$10 = 0, $c1$11 = 0, $c1$13 = 0, $c1$14 = 0, $c1$16 = 0, $c1$17 = 0, $c1$19 = 0, $c1$20 = 0, $c1$23 = 0, $c1$24 = 0;
 var $c1$26 = 0, $c1$27 = 0, $c1$29 = 0, $c1$30 = 0, $c1$4 = 0, $c1$5 = 0, $c1$7 = 0, $c1$8 = 0, $c2$0 = 0, $c2$10 = 0, $c2$11 = 0, $c2$14 = 0, $c2$15 = 0, $c2$17 = 0, $c2$18 = 0, $c2$2 = 0, $c2$20 = 0, $c2$21 = 0, $c2$23 = 0, $c2$24 = 0;
 var $c2$26 = 0, $c2$27 = 0, $c2$4 = 0, $c2$5 = 0, $c2$7 = 0, $c2$8 = 0, $c3$1 = 0, $c3$11 = 0, $c3$12 = 0, $c3$14 = 0, $c3$15 = 0, $c3$17 = 0, $c3$18 = 0, $c3$2 = 0, $c3$20 = 0, $c3$21 = 0, $c3$23 = 0, $c3$24 = 0, $c3$27 = 0, $c3$28 = 0;
 var $c3$4 = 0, $c3$5 = 0, $c3$8 = 0, $c3$9 = 0, $t2$1 = 0, $t2$10 = 0, $t2$11 = 0, $t2$12 = 0, $t2$13 = 0, $t2$14 = 0, $t2$15 = 0, $t2$16 = 0, $t2$17 = 0, $t2$18 = 0, $t2$19 = 0, $t2$20 = 0, $t2$21 = 0, $t2$22 = 0, $t2$23 = 0, $t2$24 = 0;
 var $t2$25 = 0, $t2$26 = 0, $t2$27 = 0, $t2$28 = 0, $t2$29 = 0, $t2$3 = 0, $t2$30 = 0, $t2$31 = 0, $t2$32 = 0, $t2$33 = 0, $t2$34 = 0, $t2$35 = 0, $t2$4 = 0, $t2$5 = 0, $t2$6 = 0, $t2$7 = 0, $t2$8 = 0, $t2$9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[$a>>2]|0;
 $1 = (___muldi3(($0|0),0,($0|0),0)|0);
 $2 = tempRet0;
 HEAP32[$r>>2] = $1;
 $3 = ((($a)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$a>>2]|0;
 $6 = (___muldi3(($5|0),0,($4|0),0)|0);
 $7 = tempRet0;
 $8 = (_bitshift64Shl(($6|0),($7|0),1)|0);
 $9 = tempRet0;
 $10 = ($9>>>0)<($7>>>0);
 $11 = ($8>>>0)<($6>>>0);
 $12 = ($9|0)==($7|0);
 $13 = $12 & $11;
 $14 = $10 | $13;
 $c1$0 = $14&1;
 $15 = (_bitshift64Lshr(($6|0),($7|0),31)|0);
 $16 = tempRet0;
 $17 = (($8) + ($2))|0;
 $18 = ($17>>>0)<($8>>>0);
 if ($18) {
  $19 = (($15) + 1)|0;
  $20 = ($19|0)==(0);
  $21 = $14 ? 2 : 1;
  $$c1$0 = $20 ? $21 : $c1$0;
  $c1$1 = $$c1$0;$t2$1 = $19;
 } else {
  $c1$1 = $c1$0;$t2$1 = $15;
 }
 $22 = ((($r)) + 4|0);
 HEAP32[$22>>2] = $17;
 $23 = HEAP32[$3>>2]|0;
 $24 = (___muldi3(($23|0),0,($23|0),0)|0);
 $25 = tempRet0;
 $26 = (($24) + ($t2$1))|0;
 $27 = ($26>>>0)<($24>>>0);
 $28 = $27&1;
 $$ = (($28) + ($25))|0;
 $29 = (($$) + ($c1$1))|0;
 $30 = ($29>>>0)<($$>>>0);
 $c2$0 = $30&1;
 $31 = ((($a)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = HEAP32[$a>>2]|0;
 $34 = (___muldi3(($33|0),0,($32|0),0)|0);
 $35 = tempRet0;
 $36 = (_bitshift64Shl(($34|0),($35|0),1)|0);
 $37 = tempRet0;
 $38 = ($37>>>0)<($35>>>0);
 $39 = ($36>>>0)<($34>>>0);
 $40 = ($37|0)==($35|0);
 $41 = $40 & $39;
 $42 = $38 | $41;
 $43 = $30 ? 2 : 1;
 $$c2$0 = $42 ? $43 : $c2$0;
 $44 = (_bitshift64Lshr(($34|0),($35|0),31)|0);
 $45 = tempRet0;
 $46 = (($36) + ($26))|0;
 $47 = ($46>>>0)<($36>>>0);
 if ($47) {
  $48 = (($44) + 1)|0;
  $49 = ($48|0)==(0);
  $50 = $49&1;
  $$$c2$0 = (($$c2$0) + ($50))|0;
  $c2$2 = $$$c2$0;$t2$3 = $48;
 } else {
  $c2$2 = $$c2$0;$t2$3 = $44;
 }
 $51 = (($t2$3) + ($29))|0;
 $52 = ($51>>>0)<($t2$3>>>0);
 $53 = $52&1;
 $$c2$2 = (($53) + ($c2$2))|0;
 $54 = ((($r)) + 8|0);
 HEAP32[$54>>2] = $46;
 $55 = ((($a)) + 12|0);
 $56 = HEAP32[$55>>2]|0;
 $57 = HEAP32[$a>>2]|0;
 $58 = (___muldi3(($57|0),0,($56|0),0)|0);
 $59 = tempRet0;
 $60 = (_bitshift64Shl(($58|0),($59|0),1)|0);
 $61 = tempRet0;
 $62 = ($61>>>0)<($59>>>0);
 $63 = ($60>>>0)<($58>>>0);
 $64 = ($61|0)==($59|0);
 $65 = $64 & $63;
 $66 = $62 | $65;
 $c3$1 = $66&1;
 $67 = (_bitshift64Lshr(($58|0),($59|0),31)|0);
 $68 = tempRet0;
 $69 = (($60) + ($51))|0;
 $70 = ($69>>>0)<($60>>>0);
 if ($70) {
  $71 = (($67) + 1)|0;
  $72 = ($71|0)==(0);
  $73 = $66 ? 2 : 1;
  $$c3$1 = $72 ? $73 : $c3$1;
  $c3$2 = $$c3$1;$t2$4 = $71;
 } else {
  $c3$2 = $c3$1;$t2$4 = $67;
 }
 $74 = (($$c2$2) + ($t2$4))|0;
 $75 = ($74>>>0)<($t2$4>>>0);
 $76 = $75&1;
 $$c3$2 = (($76) + ($c3$2))|0;
 $77 = HEAP32[$31>>2]|0;
 $78 = HEAP32[$3>>2]|0;
 $79 = (___muldi3(($78|0),0,($77|0),0)|0);
 $80 = tempRet0;
 $81 = (_bitshift64Shl(($79|0),($80|0),1)|0);
 $82 = tempRet0;
 $83 = ($82>>>0)<($80>>>0);
 $84 = ($81>>>0)<($79>>>0);
 $85 = ($82|0)==($80|0);
 $86 = $85 & $84;
 $87 = $83 | $86;
 $88 = $87&1;
 $c3$4 = (($$c3$2) + ($88))|0;
 $89 = (_bitshift64Lshr(($79|0),($80|0),31)|0);
 $90 = tempRet0;
 $91 = (($81) + ($69))|0;
 $92 = ($91>>>0)<($81>>>0);
 if ($92) {
  $93 = (($89) + 1)|0;
  $94 = ($93|0)==(0);
  $95 = $94&1;
  $$c3$4 = (($95) + ($c3$4))|0;
  $c3$5 = $$c3$4;$t2$5 = $93;
 } else {
  $c3$5 = $c3$4;$t2$5 = $89;
 }
 $96 = (($t2$5) + ($74))|0;
 $97 = ($96>>>0)<($t2$5>>>0);
 $98 = $97&1;
 $$c3$5 = (($98) + ($c3$5))|0;
 $99 = ((($r)) + 12|0);
 HEAP32[$99>>2] = $91;
 $100 = HEAP32[$31>>2]|0;
 $101 = (___muldi3(($100|0),0,($100|0),0)|0);
 $102 = tempRet0;
 $103 = (($101) + ($96))|0;
 $104 = ($103>>>0)<($101>>>0);
 $105 = $104&1;
 $t2$6 = (($105) + ($102))|0;
 $106 = (($$c3$5) + ($t2$6))|0;
 $107 = ($106>>>0)<($t2$6>>>0);
 $$1 = $107&1;
 $108 = HEAP32[$55>>2]|0;
 $109 = HEAP32[$3>>2]|0;
 $110 = (___muldi3(($109|0),0,($108|0),0)|0);
 $111 = tempRet0;
 $112 = (_bitshift64Shl(($110|0),($111|0),1)|0);
 $113 = tempRet0;
 $114 = ($113>>>0)<($111>>>0);
 $115 = ($112>>>0)<($110>>>0);
 $116 = ($113|0)==($111|0);
 $117 = $116 & $115;
 $118 = $114 | $117;
 $119 = $107 ? 2 : 1;
 $c1$4 = $118 ? $119 : $$1;
 $120 = (_bitshift64Lshr(($110|0),($111|0),31)|0);
 $121 = tempRet0;
 $122 = (($112) + ($103))|0;
 $123 = ($122>>>0)<($112>>>0);
 if ($123) {
  $124 = (($120) + 1)|0;
  $125 = ($124|0)==(0);
  $126 = $125&1;
  $$c1$4 = (($c1$4) + ($126))|0;
  $c1$5 = $$c1$4;$t2$7 = $124;
 } else {
  $c1$5 = $c1$4;$t2$7 = $120;
 }
 $127 = (($t2$7) + ($106))|0;
 $128 = ($127>>>0)<($t2$7>>>0);
 $129 = $128&1;
 $$c1$5 = (($129) + ($c1$5))|0;
 $130 = ((($a)) + 16|0);
 $131 = HEAP32[$130>>2]|0;
 $132 = HEAP32[$a>>2]|0;
 $133 = (___muldi3(($132|0),0,($131|0),0)|0);
 $134 = tempRet0;
 $135 = (_bitshift64Shl(($133|0),($134|0),1)|0);
 $136 = tempRet0;
 $137 = ($136>>>0)<($134>>>0);
 $138 = ($135>>>0)<($133>>>0);
 $139 = ($136|0)==($134|0);
 $140 = $139 & $138;
 $141 = $137 | $140;
 $142 = $141&1;
 $c1$7 = (($$c1$5) + ($142))|0;
 $143 = (_bitshift64Lshr(($133|0),($134|0),31)|0);
 $144 = tempRet0;
 $145 = (($135) + ($122))|0;
 $146 = ($145>>>0)<($135>>>0);
 if ($146) {
  $147 = (($143) + 1)|0;
  $148 = ($147|0)==(0);
  $149 = $148&1;
  $$c1$7 = (($149) + ($c1$7))|0;
  $c1$8 = $$c1$7;$t2$8 = $147;
 } else {
  $c1$8 = $c1$7;$t2$8 = $143;
 }
 $150 = (($t2$8) + ($127))|0;
 $151 = ($150>>>0)<($t2$8>>>0);
 $152 = $151&1;
 $$c1$8 = (($152) + ($c1$8))|0;
 $153 = ((($r)) + 16|0);
 HEAP32[$153>>2] = $145;
 $154 = ((($a)) + 20|0);
 $155 = HEAP32[$154>>2]|0;
 $156 = HEAP32[$a>>2]|0;
 $157 = (___muldi3(($156|0),0,($155|0),0)|0);
 $158 = tempRet0;
 $159 = (_bitshift64Shl(($157|0),($158|0),1)|0);
 $160 = tempRet0;
 $161 = ($160>>>0)<($158>>>0);
 $162 = ($159>>>0)<($157>>>0);
 $163 = ($160|0)==($158|0);
 $164 = $163 & $162;
 $165 = $161 | $164;
 $c2$4 = $165&1;
 $166 = (_bitshift64Lshr(($157|0),($158|0),31)|0);
 $167 = tempRet0;
 $168 = (($159) + ($150))|0;
 $169 = ($168>>>0)<($159>>>0);
 if ($169) {
  $170 = (($166) + 1)|0;
  $171 = ($170|0)==(0);
  $172 = $165 ? 2 : 1;
  $$c2$4 = $171 ? $172 : $c2$4;
  $c2$5 = $$c2$4;$t2$9 = $170;
 } else {
  $c2$5 = $c2$4;$t2$9 = $166;
 }
 $173 = (($$c1$8) + ($t2$9))|0;
 $174 = ($173>>>0)<($t2$9>>>0);
 $175 = $174&1;
 $$c2$5 = (($175) + ($c2$5))|0;
 $176 = HEAP32[$130>>2]|0;
 $177 = HEAP32[$3>>2]|0;
 $178 = (___muldi3(($177|0),0,($176|0),0)|0);
 $179 = tempRet0;
 $180 = (_bitshift64Shl(($178|0),($179|0),1)|0);
 $181 = tempRet0;
 $182 = ($181>>>0)<($179>>>0);
 $183 = ($180>>>0)<($178>>>0);
 $184 = ($181|0)==($179|0);
 $185 = $184 & $183;
 $186 = $182 | $185;
 $187 = $186&1;
 $c2$7 = (($$c2$5) + ($187))|0;
 $188 = (_bitshift64Lshr(($178|0),($179|0),31)|0);
 $189 = tempRet0;
 $190 = (($180) + ($168))|0;
 $191 = ($190>>>0)<($180>>>0);
 if ($191) {
  $192 = (($188) + 1)|0;
  $193 = ($192|0)==(0);
  $194 = $193&1;
  $$c2$7 = (($194) + ($c2$7))|0;
  $c2$8 = $$c2$7;$t2$10 = $192;
 } else {
  $c2$8 = $c2$7;$t2$10 = $188;
 }
 $195 = (($t2$10) + ($173))|0;
 $196 = ($195>>>0)<($t2$10>>>0);
 $197 = $196&1;
 $$c2$8 = (($197) + ($c2$8))|0;
 $198 = HEAP32[$55>>2]|0;
 $199 = HEAP32[$31>>2]|0;
 $200 = (___muldi3(($199|0),0,($198|0),0)|0);
 $201 = tempRet0;
 $202 = (_bitshift64Shl(($200|0),($201|0),1)|0);
 $203 = tempRet0;
 $204 = ($203>>>0)<($201>>>0);
 $205 = ($202>>>0)<($200>>>0);
 $206 = ($203|0)==($201|0);
 $207 = $206 & $205;
 $208 = $204 | $207;
 $209 = $208&1;
 $c2$10 = (($$c2$8) + ($209))|0;
 $210 = (_bitshift64Lshr(($200|0),($201|0),31)|0);
 $211 = tempRet0;
 $212 = (($202) + ($190))|0;
 $213 = ($212>>>0)<($202>>>0);
 if ($213) {
  $214 = (($210) + 1)|0;
  $215 = ($214|0)==(0);
  $216 = $215&1;
  $$c2$10 = (($216) + ($c2$10))|0;
  $c2$11 = $$c2$10;$t2$11 = $214;
 } else {
  $c2$11 = $c2$10;$t2$11 = $210;
 }
 $217 = (($t2$11) + ($195))|0;
 $218 = ($217>>>0)<($t2$11>>>0);
 $219 = $218&1;
 $$c2$11 = (($219) + ($c2$11))|0;
 $220 = ((($r)) + 20|0);
 HEAP32[$220>>2] = $212;
 $221 = HEAP32[$55>>2]|0;
 $222 = (___muldi3(($221|0),0,($221|0),0)|0);
 $223 = tempRet0;
 $224 = (($222) + ($217))|0;
 $225 = ($224>>>0)<($222>>>0);
 $226 = $225&1;
 $t2$12 = (($226) + ($223))|0;
 $227 = (($$c2$11) + ($t2$12))|0;
 $228 = ($227>>>0)<($t2$12>>>0);
 $$2 = $228&1;
 $229 = HEAP32[$130>>2]|0;
 $230 = HEAP32[$31>>2]|0;
 $231 = (___muldi3(($230|0),0,($229|0),0)|0);
 $232 = tempRet0;
 $233 = (_bitshift64Shl(($231|0),($232|0),1)|0);
 $234 = tempRet0;
 $235 = ($234>>>0)<($232>>>0);
 $236 = ($233>>>0)<($231>>>0);
 $237 = ($234|0)==($232|0);
 $238 = $237 & $236;
 $239 = $235 | $238;
 $240 = $228 ? 2 : 1;
 $c3$8 = $239 ? $240 : $$2;
 $241 = (_bitshift64Lshr(($231|0),($232|0),31)|0);
 $242 = tempRet0;
 $243 = (($233) + ($224))|0;
 $244 = ($243>>>0)<($233>>>0);
 if ($244) {
  $245 = (($241) + 1)|0;
  $246 = ($245|0)==(0);
  $247 = $246&1;
  $$c3$8 = (($c3$8) + ($247))|0;
  $c3$9 = $$c3$8;$t2$13 = $245;
 } else {
  $c3$9 = $c3$8;$t2$13 = $241;
 }
 $248 = (($t2$13) + ($227))|0;
 $249 = ($248>>>0)<($t2$13>>>0);
 $250 = $249&1;
 $$c3$9 = (($250) + ($c3$9))|0;
 $251 = HEAP32[$154>>2]|0;
 $252 = HEAP32[$3>>2]|0;
 $253 = (___muldi3(($252|0),0,($251|0),0)|0);
 $254 = tempRet0;
 $255 = (_bitshift64Shl(($253|0),($254|0),1)|0);
 $256 = tempRet0;
 $257 = ($256>>>0)<($254>>>0);
 $258 = ($255>>>0)<($253>>>0);
 $259 = ($256|0)==($254|0);
 $260 = $259 & $258;
 $261 = $257 | $260;
 $262 = $261&1;
 $c3$11 = (($$c3$9) + ($262))|0;
 $263 = (_bitshift64Lshr(($253|0),($254|0),31)|0);
 $264 = tempRet0;
 $265 = (($255) + ($243))|0;
 $266 = ($265>>>0)<($255>>>0);
 if ($266) {
  $267 = (($263) + 1)|0;
  $268 = ($267|0)==(0);
  $269 = $268&1;
  $$c3$11 = (($269) + ($c3$11))|0;
  $c3$12 = $$c3$11;$t2$14 = $267;
 } else {
  $c3$12 = $c3$11;$t2$14 = $263;
 }
 $270 = (($t2$14) + ($248))|0;
 $271 = ($270>>>0)<($t2$14>>>0);
 $272 = $271&1;
 $$c3$12 = (($272) + ($c3$12))|0;
 $273 = ((($a)) + 24|0);
 $274 = HEAP32[$273>>2]|0;
 $275 = HEAP32[$a>>2]|0;
 $276 = (___muldi3(($275|0),0,($274|0),0)|0);
 $277 = tempRet0;
 $278 = (_bitshift64Shl(($276|0),($277|0),1)|0);
 $279 = tempRet0;
 $280 = ($279>>>0)<($277>>>0);
 $281 = ($278>>>0)<($276>>>0);
 $282 = ($279|0)==($277|0);
 $283 = $282 & $281;
 $284 = $280 | $283;
 $285 = $284&1;
 $c3$14 = (($$c3$12) + ($285))|0;
 $286 = (_bitshift64Lshr(($276|0),($277|0),31)|0);
 $287 = tempRet0;
 $288 = (($278) + ($265))|0;
 $289 = ($288>>>0)<($278>>>0);
 if ($289) {
  $290 = (($286) + 1)|0;
  $291 = ($290|0)==(0);
  $292 = $291&1;
  $$c3$14 = (($292) + ($c3$14))|0;
  $c3$15 = $$c3$14;$t2$15 = $290;
 } else {
  $c3$15 = $c3$14;$t2$15 = $286;
 }
 $293 = (($t2$15) + ($270))|0;
 $294 = ($293>>>0)<($t2$15>>>0);
 $295 = $294&1;
 $$c3$15 = (($295) + ($c3$15))|0;
 $296 = ((($r)) + 24|0);
 HEAP32[$296>>2] = $288;
 $297 = ((($a)) + 28|0);
 $298 = HEAP32[$297>>2]|0;
 $299 = HEAP32[$a>>2]|0;
 $300 = (___muldi3(($299|0),0,($298|0),0)|0);
 $301 = tempRet0;
 $302 = (_bitshift64Shl(($300|0),($301|0),1)|0);
 $303 = tempRet0;
 $304 = ($303>>>0)<($301>>>0);
 $305 = ($302>>>0)<($300>>>0);
 $306 = ($303|0)==($301|0);
 $307 = $306 & $305;
 $308 = $304 | $307;
 $c1$10 = $308&1;
 $309 = (_bitshift64Lshr(($300|0),($301|0),31)|0);
 $310 = tempRet0;
 $311 = (($302) + ($293))|0;
 $312 = ($311>>>0)<($302>>>0);
 if ($312) {
  $313 = (($309) + 1)|0;
  $314 = ($313|0)==(0);
  $315 = $308 ? 2 : 1;
  $$c1$10 = $314 ? $315 : $c1$10;
  $c1$11 = $$c1$10;$t2$16 = $313;
 } else {
  $c1$11 = $c1$10;$t2$16 = $309;
 }
 $316 = (($$c3$15) + ($t2$16))|0;
 $317 = ($316>>>0)<($t2$16>>>0);
 $318 = $317&1;
 $$c1$11 = (($318) + ($c1$11))|0;
 $319 = HEAP32[$273>>2]|0;
 $320 = HEAP32[$3>>2]|0;
 $321 = (___muldi3(($320|0),0,($319|0),0)|0);
 $322 = tempRet0;
 $323 = (_bitshift64Shl(($321|0),($322|0),1)|0);
 $324 = tempRet0;
 $325 = ($324>>>0)<($322>>>0);
 $326 = ($323>>>0)<($321>>>0);
 $327 = ($324|0)==($322|0);
 $328 = $327 & $326;
 $329 = $325 | $328;
 $330 = $329&1;
 $c1$13 = (($$c1$11) + ($330))|0;
 $331 = (_bitshift64Lshr(($321|0),($322|0),31)|0);
 $332 = tempRet0;
 $333 = (($323) + ($311))|0;
 $334 = ($333>>>0)<($323>>>0);
 if ($334) {
  $335 = (($331) + 1)|0;
  $336 = ($335|0)==(0);
  $337 = $336&1;
  $$c1$13 = (($337) + ($c1$13))|0;
  $c1$14 = $$c1$13;$t2$17 = $335;
 } else {
  $c1$14 = $c1$13;$t2$17 = $331;
 }
 $338 = (($t2$17) + ($316))|0;
 $339 = ($338>>>0)<($t2$17>>>0);
 $340 = $339&1;
 $$c1$14 = (($340) + ($c1$14))|0;
 $341 = HEAP32[$154>>2]|0;
 $342 = HEAP32[$31>>2]|0;
 $343 = (___muldi3(($342|0),0,($341|0),0)|0);
 $344 = tempRet0;
 $345 = (_bitshift64Shl(($343|0),($344|0),1)|0);
 $346 = tempRet0;
 $347 = ($346>>>0)<($344>>>0);
 $348 = ($345>>>0)<($343>>>0);
 $349 = ($346|0)==($344|0);
 $350 = $349 & $348;
 $351 = $347 | $350;
 $352 = $351&1;
 $c1$16 = (($$c1$14) + ($352))|0;
 $353 = (_bitshift64Lshr(($343|0),($344|0),31)|0);
 $354 = tempRet0;
 $355 = (($345) + ($333))|0;
 $356 = ($355>>>0)<($345>>>0);
 if ($356) {
  $357 = (($353) + 1)|0;
  $358 = ($357|0)==(0);
  $359 = $358&1;
  $$c1$16 = (($359) + ($c1$16))|0;
  $c1$17 = $$c1$16;$t2$18 = $357;
 } else {
  $c1$17 = $c1$16;$t2$18 = $353;
 }
 $360 = (($t2$18) + ($338))|0;
 $361 = ($360>>>0)<($t2$18>>>0);
 $362 = $361&1;
 $$c1$17 = (($362) + ($c1$17))|0;
 $363 = HEAP32[$130>>2]|0;
 $364 = HEAP32[$55>>2]|0;
 $365 = (___muldi3(($364|0),0,($363|0),0)|0);
 $366 = tempRet0;
 $367 = (_bitshift64Shl(($365|0),($366|0),1)|0);
 $368 = tempRet0;
 $369 = ($368>>>0)<($366>>>0);
 $370 = ($367>>>0)<($365>>>0);
 $371 = ($368|0)==($366|0);
 $372 = $371 & $370;
 $373 = $369 | $372;
 $374 = $373&1;
 $c1$19 = (($$c1$17) + ($374))|0;
 $375 = (_bitshift64Lshr(($365|0),($366|0),31)|0);
 $376 = tempRet0;
 $377 = (($367) + ($355))|0;
 $378 = ($377>>>0)<($367>>>0);
 if ($378) {
  $379 = (($375) + 1)|0;
  $380 = ($379|0)==(0);
  $381 = $380&1;
  $$c1$19 = (($381) + ($c1$19))|0;
  $c1$20 = $$c1$19;$t2$19 = $379;
 } else {
  $c1$20 = $c1$19;$t2$19 = $375;
 }
 $382 = (($t2$19) + ($360))|0;
 $383 = ($382>>>0)<($t2$19>>>0);
 $384 = $383&1;
 $$c1$20 = (($384) + ($c1$20))|0;
 $385 = ((($r)) + 28|0);
 HEAP32[$385>>2] = $377;
 $386 = HEAP32[$130>>2]|0;
 $387 = (___muldi3(($386|0),0,($386|0),0)|0);
 $388 = tempRet0;
 $389 = (($387) + ($382))|0;
 $390 = ($389>>>0)<($387>>>0);
 $391 = $390&1;
 $t2$20 = (($391) + ($388))|0;
 $392 = (($$c1$20) + ($t2$20))|0;
 $393 = ($392>>>0)<($t2$20>>>0);
 $$3 = $393&1;
 $394 = HEAP32[$154>>2]|0;
 $395 = HEAP32[$55>>2]|0;
 $396 = (___muldi3(($395|0),0,($394|0),0)|0);
 $397 = tempRet0;
 $398 = (_bitshift64Shl(($396|0),($397|0),1)|0);
 $399 = tempRet0;
 $400 = ($399>>>0)<($397>>>0);
 $401 = ($398>>>0)<($396>>>0);
 $402 = ($399|0)==($397|0);
 $403 = $402 & $401;
 $404 = $400 | $403;
 $405 = $393 ? 2 : 1;
 $c2$14 = $404 ? $405 : $$3;
 $406 = (_bitshift64Lshr(($396|0),($397|0),31)|0);
 $407 = tempRet0;
 $408 = (($398) + ($389))|0;
 $409 = ($408>>>0)<($398>>>0);
 if ($409) {
  $410 = (($406) + 1)|0;
  $411 = ($410|0)==(0);
  $412 = $411&1;
  $$c2$14 = (($c2$14) + ($412))|0;
  $c2$15 = $$c2$14;$t2$21 = $410;
 } else {
  $c2$15 = $c2$14;$t2$21 = $406;
 }
 $413 = (($t2$21) + ($392))|0;
 $414 = ($413>>>0)<($t2$21>>>0);
 $415 = $414&1;
 $$c2$15 = (($415) + ($c2$15))|0;
 $416 = HEAP32[$273>>2]|0;
 $417 = HEAP32[$31>>2]|0;
 $418 = (___muldi3(($417|0),0,($416|0),0)|0);
 $419 = tempRet0;
 $420 = (_bitshift64Shl(($418|0),($419|0),1)|0);
 $421 = tempRet0;
 $422 = ($421>>>0)<($419>>>0);
 $423 = ($420>>>0)<($418>>>0);
 $424 = ($421|0)==($419|0);
 $425 = $424 & $423;
 $426 = $422 | $425;
 $427 = $426&1;
 $c2$17 = (($$c2$15) + ($427))|0;
 $428 = (_bitshift64Lshr(($418|0),($419|0),31)|0);
 $429 = tempRet0;
 $430 = (($420) + ($408))|0;
 $431 = ($430>>>0)<($420>>>0);
 if ($431) {
  $432 = (($428) + 1)|0;
  $433 = ($432|0)==(0);
  $434 = $433&1;
  $$c2$17 = (($434) + ($c2$17))|0;
  $c2$18 = $$c2$17;$t2$22 = $432;
 } else {
  $c2$18 = $c2$17;$t2$22 = $428;
 }
 $435 = (($t2$22) + ($413))|0;
 $436 = ($435>>>0)<($t2$22>>>0);
 $437 = $436&1;
 $$c2$18 = (($437) + ($c2$18))|0;
 $438 = HEAP32[$297>>2]|0;
 $439 = HEAP32[$3>>2]|0;
 $440 = (___muldi3(($439|0),0,($438|0),0)|0);
 $441 = tempRet0;
 $442 = (_bitshift64Shl(($440|0),($441|0),1)|0);
 $443 = tempRet0;
 $444 = ($443>>>0)<($441>>>0);
 $445 = ($442>>>0)<($440>>>0);
 $446 = ($443|0)==($441|0);
 $447 = $446 & $445;
 $448 = $444 | $447;
 $449 = $448&1;
 $c2$20 = (($$c2$18) + ($449))|0;
 $450 = (_bitshift64Lshr(($440|0),($441|0),31)|0);
 $451 = tempRet0;
 $452 = (($442) + ($430))|0;
 $453 = ($452>>>0)<($442>>>0);
 if ($453) {
  $454 = (($450) + 1)|0;
  $455 = ($454|0)==(0);
  $456 = $455&1;
  $$c2$20 = (($456) + ($c2$20))|0;
  $c2$21 = $$c2$20;$t2$23 = $454;
 } else {
  $c2$21 = $c2$20;$t2$23 = $450;
 }
 $457 = (($t2$23) + ($435))|0;
 $458 = ($457>>>0)<($t2$23>>>0);
 $459 = $458&1;
 $$c2$21 = (($459) + ($c2$21))|0;
 $460 = ((($r)) + 32|0);
 HEAP32[$460>>2] = $452;
 $461 = HEAP32[$297>>2]|0;
 $462 = HEAP32[$31>>2]|0;
 $463 = (___muldi3(($462|0),0,($461|0),0)|0);
 $464 = tempRet0;
 $465 = (_bitshift64Shl(($463|0),($464|0),1)|0);
 $466 = tempRet0;
 $467 = ($466>>>0)<($464>>>0);
 $468 = ($465>>>0)<($463>>>0);
 $469 = ($466|0)==($464|0);
 $470 = $469 & $468;
 $471 = $467 | $470;
 $c3$17 = $471&1;
 $472 = (_bitshift64Lshr(($463|0),($464|0),31)|0);
 $473 = tempRet0;
 $474 = (($465) + ($457))|0;
 $475 = ($474>>>0)<($465>>>0);
 if ($475) {
  $476 = (($472) + 1)|0;
  $477 = ($476|0)==(0);
  $478 = $471 ? 2 : 1;
  $$c3$17 = $477 ? $478 : $c3$17;
  $c3$18 = $$c3$17;$t2$24 = $476;
 } else {
  $c3$18 = $c3$17;$t2$24 = $472;
 }
 $479 = (($$c2$21) + ($t2$24))|0;
 $480 = ($479>>>0)<($t2$24>>>0);
 $481 = $480&1;
 $$c3$18 = (($481) + ($c3$18))|0;
 $482 = HEAP32[$273>>2]|0;
 $483 = HEAP32[$55>>2]|0;
 $484 = (___muldi3(($483|0),0,($482|0),0)|0);
 $485 = tempRet0;
 $486 = (_bitshift64Shl(($484|0),($485|0),1)|0);
 $487 = tempRet0;
 $488 = ($487>>>0)<($485>>>0);
 $489 = ($486>>>0)<($484>>>0);
 $490 = ($487|0)==($485|0);
 $491 = $490 & $489;
 $492 = $488 | $491;
 $493 = $492&1;
 $c3$20 = (($$c3$18) + ($493))|0;
 $494 = (_bitshift64Lshr(($484|0),($485|0),31)|0);
 $495 = tempRet0;
 $496 = (($486) + ($474))|0;
 $497 = ($496>>>0)<($486>>>0);
 if ($497) {
  $498 = (($494) + 1)|0;
  $499 = ($498|0)==(0);
  $500 = $499&1;
  $$c3$20 = (($500) + ($c3$20))|0;
  $c3$21 = $$c3$20;$t2$25 = $498;
 } else {
  $c3$21 = $c3$20;$t2$25 = $494;
 }
 $501 = (($t2$25) + ($479))|0;
 $502 = ($501>>>0)<($t2$25>>>0);
 $503 = $502&1;
 $$c3$21 = (($503) + ($c3$21))|0;
 $504 = HEAP32[$154>>2]|0;
 $505 = HEAP32[$130>>2]|0;
 $506 = (___muldi3(($505|0),0,($504|0),0)|0);
 $507 = tempRet0;
 $508 = (_bitshift64Shl(($506|0),($507|0),1)|0);
 $509 = tempRet0;
 $510 = ($509>>>0)<($507>>>0);
 $511 = ($508>>>0)<($506>>>0);
 $512 = ($509|0)==($507|0);
 $513 = $512 & $511;
 $514 = $510 | $513;
 $515 = $514&1;
 $c3$23 = (($$c3$21) + ($515))|0;
 $516 = (_bitshift64Lshr(($506|0),($507|0),31)|0);
 $517 = tempRet0;
 $518 = (($508) + ($496))|0;
 $519 = ($518>>>0)<($508>>>0);
 if ($519) {
  $520 = (($516) + 1)|0;
  $521 = ($520|0)==(0);
  $522 = $521&1;
  $$c3$23 = (($522) + ($c3$23))|0;
  $c3$24 = $$c3$23;$t2$26 = $520;
 } else {
  $c3$24 = $c3$23;$t2$26 = $516;
 }
 $523 = (($t2$26) + ($501))|0;
 $524 = ($523>>>0)<($t2$26>>>0);
 $525 = $524&1;
 $$c3$24 = (($525) + ($c3$24))|0;
 $526 = ((($r)) + 36|0);
 HEAP32[$526>>2] = $518;
 $527 = HEAP32[$154>>2]|0;
 $528 = (___muldi3(($527|0),0,($527|0),0)|0);
 $529 = tempRet0;
 $530 = (($528) + ($523))|0;
 $531 = ($530>>>0)<($528>>>0);
 $532 = $531&1;
 $t2$27 = (($532) + ($529))|0;
 $533 = (($$c3$24) + ($t2$27))|0;
 $534 = ($533>>>0)<($t2$27>>>0);
 $$4 = $534&1;
 $535 = HEAP32[$273>>2]|0;
 $536 = HEAP32[$130>>2]|0;
 $537 = (___muldi3(($536|0),0,($535|0),0)|0);
 $538 = tempRet0;
 $539 = (_bitshift64Shl(($537|0),($538|0),1)|0);
 $540 = tempRet0;
 $541 = ($540>>>0)<($538>>>0);
 $542 = ($539>>>0)<($537>>>0);
 $543 = ($540|0)==($538|0);
 $544 = $543 & $542;
 $545 = $541 | $544;
 $546 = $534 ? 2 : 1;
 $c1$23 = $545 ? $546 : $$4;
 $547 = (_bitshift64Lshr(($537|0),($538|0),31)|0);
 $548 = tempRet0;
 $549 = (($539) + ($530))|0;
 $550 = ($549>>>0)<($539>>>0);
 if ($550) {
  $551 = (($547) + 1)|0;
  $552 = ($551|0)==(0);
  $553 = $552&1;
  $$c1$23 = (($c1$23) + ($553))|0;
  $c1$24 = $$c1$23;$t2$28 = $551;
 } else {
  $c1$24 = $c1$23;$t2$28 = $547;
 }
 $554 = (($t2$28) + ($533))|0;
 $555 = ($554>>>0)<($t2$28>>>0);
 $556 = $555&1;
 $$c1$24 = (($556) + ($c1$24))|0;
 $557 = HEAP32[$297>>2]|0;
 $558 = HEAP32[$55>>2]|0;
 $559 = (___muldi3(($558|0),0,($557|0),0)|0);
 $560 = tempRet0;
 $561 = (_bitshift64Shl(($559|0),($560|0),1)|0);
 $562 = tempRet0;
 $563 = ($562>>>0)<($560>>>0);
 $564 = ($561>>>0)<($559>>>0);
 $565 = ($562|0)==($560|0);
 $566 = $565 & $564;
 $567 = $563 | $566;
 $568 = $567&1;
 $c1$26 = (($$c1$24) + ($568))|0;
 $569 = (_bitshift64Lshr(($559|0),($560|0),31)|0);
 $570 = tempRet0;
 $571 = (($561) + ($549))|0;
 $572 = ($571>>>0)<($561>>>0);
 if ($572) {
  $573 = (($569) + 1)|0;
  $574 = ($573|0)==(0);
  $575 = $574&1;
  $$c1$26 = (($575) + ($c1$26))|0;
  $c1$27 = $$c1$26;$t2$29 = $573;
 } else {
  $c1$27 = $c1$26;$t2$29 = $569;
 }
 $576 = (($t2$29) + ($554))|0;
 $577 = ($576>>>0)<($t2$29>>>0);
 $578 = $577&1;
 $$c1$27 = (($578) + ($c1$27))|0;
 $579 = ((($r)) + 40|0);
 HEAP32[$579>>2] = $571;
 $580 = HEAP32[$297>>2]|0;
 $581 = HEAP32[$130>>2]|0;
 $582 = (___muldi3(($581|0),0,($580|0),0)|0);
 $583 = tempRet0;
 $584 = (_bitshift64Shl(($582|0),($583|0),1)|0);
 $585 = tempRet0;
 $586 = ($585>>>0)<($583>>>0);
 $587 = ($584>>>0)<($582>>>0);
 $588 = ($585|0)==($583|0);
 $589 = $588 & $587;
 $590 = $586 | $589;
 $c2$23 = $590&1;
 $591 = (_bitshift64Lshr(($582|0),($583|0),31)|0);
 $592 = tempRet0;
 $593 = (($584) + ($576))|0;
 $594 = ($593>>>0)<($584>>>0);
 if ($594) {
  $595 = (($591) + 1)|0;
  $596 = ($595|0)==(0);
  $597 = $590 ? 2 : 1;
  $$c2$23 = $596 ? $597 : $c2$23;
  $c2$24 = $$c2$23;$t2$30 = $595;
 } else {
  $c2$24 = $c2$23;$t2$30 = $591;
 }
 $598 = (($$c1$27) + ($t2$30))|0;
 $599 = ($598>>>0)<($t2$30>>>0);
 $600 = $599&1;
 $$c2$24 = (($600) + ($c2$24))|0;
 $601 = HEAP32[$273>>2]|0;
 $602 = HEAP32[$154>>2]|0;
 $603 = (___muldi3(($602|0),0,($601|0),0)|0);
 $604 = tempRet0;
 $605 = (_bitshift64Shl(($603|0),($604|0),1)|0);
 $606 = tempRet0;
 $607 = ($606>>>0)<($604>>>0);
 $608 = ($605>>>0)<($603>>>0);
 $609 = ($606|0)==($604|0);
 $610 = $609 & $608;
 $611 = $607 | $610;
 $612 = $611&1;
 $c2$26 = (($$c2$24) + ($612))|0;
 $613 = (_bitshift64Lshr(($603|0),($604|0),31)|0);
 $614 = tempRet0;
 $615 = (($605) + ($593))|0;
 $616 = ($615>>>0)<($605>>>0);
 if ($616) {
  $617 = (($613) + 1)|0;
  $618 = ($617|0)==(0);
  $619 = $618&1;
  $$c2$26 = (($619) + ($c2$26))|0;
  $c2$27 = $$c2$26;$t2$31 = $617;
 } else {
  $c2$27 = $c2$26;$t2$31 = $613;
 }
 $620 = (($t2$31) + ($598))|0;
 $621 = ($620>>>0)<($t2$31>>>0);
 $622 = $621&1;
 $$c2$27 = (($622) + ($c2$27))|0;
 $623 = ((($r)) + 44|0);
 HEAP32[$623>>2] = $615;
 $624 = HEAP32[$273>>2]|0;
 $625 = (___muldi3(($624|0),0,($624|0),0)|0);
 $626 = tempRet0;
 $627 = (($625) + ($620))|0;
 $628 = ($627>>>0)<($625>>>0);
 $629 = $628&1;
 $t2$32 = (($629) + ($626))|0;
 $630 = (($$c2$27) + ($t2$32))|0;
 $631 = ($630>>>0)<($t2$32>>>0);
 $$5 = $631&1;
 $632 = HEAP32[$297>>2]|0;
 $633 = HEAP32[$154>>2]|0;
 $634 = (___muldi3(($633|0),0,($632|0),0)|0);
 $635 = tempRet0;
 $636 = (_bitshift64Shl(($634|0),($635|0),1)|0);
 $637 = tempRet0;
 $638 = ($637>>>0)<($635>>>0);
 $639 = ($636>>>0)<($634>>>0);
 $640 = ($637|0)==($635|0);
 $641 = $640 & $639;
 $642 = $638 | $641;
 $643 = $631 ? 2 : 1;
 $c3$27 = $642 ? $643 : $$5;
 $644 = (_bitshift64Lshr(($634|0),($635|0),31)|0);
 $645 = tempRet0;
 $646 = (($636) + ($627))|0;
 $647 = ($646>>>0)<($636>>>0);
 if ($647) {
  $648 = (($644) + 1)|0;
  $649 = ($648|0)==(0);
  $650 = $649&1;
  $$c3$27 = (($c3$27) + ($650))|0;
  $c3$28 = $$c3$27;$t2$33 = $648;
 } else {
  $c3$28 = $c3$27;$t2$33 = $644;
 }
 $651 = (($t2$33) + ($630))|0;
 $652 = ($651>>>0)<($t2$33>>>0);
 $653 = $652&1;
 $$c3$28 = (($653) + ($c3$28))|0;
 $654 = ((($r)) + 48|0);
 HEAP32[$654>>2] = $646;
 $655 = HEAP32[$297>>2]|0;
 $656 = HEAP32[$273>>2]|0;
 $657 = (___muldi3(($656|0),0,($655|0),0)|0);
 $658 = tempRet0;
 $659 = (_bitshift64Shl(($657|0),($658|0),1)|0);
 $660 = tempRet0;
 $661 = ($660>>>0)<($658>>>0);
 $662 = ($659>>>0)<($657>>>0);
 $663 = ($660|0)==($658|0);
 $664 = $663 & $662;
 $665 = $661 | $664;
 $c1$29 = $665&1;
 $666 = (_bitshift64Lshr(($657|0),($658|0),31)|0);
 $667 = tempRet0;
 $668 = (($659) + ($651))|0;
 $669 = ($668>>>0)<($659>>>0);
 if (!($669)) {
  $c1$30 = $c1$29;$t2$34 = $666;
  $673 = (($$c3$28) + ($t2$34))|0;
  $674 = ($673>>>0)<($t2$34>>>0);
  $675 = $674&1;
  $676 = ((($r)) + 52|0);
  HEAP32[$676>>2] = $668;
  $677 = HEAP32[$297>>2]|0;
  $678 = (___muldi3(($677|0),0,($677|0),0)|0);
  $679 = tempRet0;
  $680 = (($678) + ($673))|0;
  $681 = ($680>>>0)<($678>>>0);
  $682 = $681&1;
  $t2$35 = (($675) + ($c1$30))|0;
  $$c1$30 = (($t2$35) + ($679))|0;
  $683 = (($$c1$30) + ($682))|0;
  $684 = ((($r)) + 56|0);
  HEAP32[$684>>2] = $680;
  $685 = ((($r)) + 60|0);
  HEAP32[$685>>2] = $683;
  return;
 }
 $670 = (($666) + 1)|0;
 $671 = ($670|0)==(0);
 $672 = $665 ? 2 : 1;
 $$c1$29 = $671 ? $672 : $c1$29;
 $c1$30 = $$c1$29;$t2$34 = $670;
 $673 = (($$c3$28) + ($t2$34))|0;
 $674 = ($673>>>0)<($t2$34>>>0);
 $675 = $674&1;
 $676 = ((($r)) + 52|0);
 HEAP32[$676>>2] = $668;
 $677 = HEAP32[$297>>2]|0;
 $678 = (___muldi3(($677|0),0,($677|0),0)|0);
 $679 = tempRet0;
 $680 = (($678) + ($673))|0;
 $681 = ($680>>>0)<($678>>>0);
 $682 = $681&1;
 $t2$35 = (($675) + ($c1$30))|0;
 $$c1$30 = (($t2$35) + ($679))|0;
 $683 = (($$c1$30) + ($682))|0;
 $684 = ((($r)) + 56|0);
 HEAP32[$684>>2] = $680;
 $685 = ((($r)) + 60|0);
 HEAP32[$685>>2] = $683;
 return;
}
function _egbn_sqr_recursive($r,$a,$n2,$t) {
 $r = $r|0;
 $a = $a|0;
 $n2 = $n2|0;
 $t = $t|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $p$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($n2|0) / 2)&-1;
 switch ($n2|0) {
 case 4:  {
  _egbn_sqr_comba4($r,$a);
  return;
  break;
 }
 case 8:  {
  _egbn_sqr_comba8($r,$a);
  return;
  break;
 }
 default: {
  $1 = ($n2|0)<(16);
  if ($1) {
   _egbn_sqr_normal($r,$a,$n2,$t);
   return;
  }
  $2 = (($a) + ($0<<2)|0);
  $3 = (_egbn_cmp_words($a,$2,$0)|0);
  $4 = ($3|0)>(0);
  do {
   if ($4) {
    (_egbn_sub_words($t,$a,$2,$0)|0);
    label = 10;
   } else {
    $5 = ($3|0)<(0);
    if ($5) {
     (_egbn_sub_words($t,$2,$a,$0)|0);
     label = 10;
     break;
    } else {
     $9 = $n2 << 1;
     $10 = (($t) + ($9<<2)|0);
     $11 = (($t) + ($n2<<2)|0);
     $12 = $n2 << 2;
     _memset(($11|0),0,($12|0))|0;
     $13 = $10;$16 = $11;
     break;
    }
   }
  } while(0);
  if ((label|0) == 10) {
   $6 = $n2 << 1;
   $7 = (($t) + ($6<<2)|0);
   $8 = (($t) + ($n2<<2)|0);
   _egbn_sqr_recursive($8,$t,$0,$7);
   $13 = $7;$16 = $8;
  }
  _egbn_sqr_recursive($r,$a,$0,$13);
  $14 = (($r) + ($n2<<2)|0);
  _egbn_sqr_recursive($14,$2,$0,$13);
  $15 = (_egbn_add_words($t,$r,$14,$n2)|0);
  $17 = (_egbn_sub_words($16,$t,$16,$n2)|0);
  $18 = (($15) - ($17))|0;
  $19 = (($r) + ($0<<2)|0);
  $20 = (_egbn_add_words($19,$19,$16,$n2)|0);
  $21 = (($18) + ($20))|0;
  $22 = ($21|0)==(0);
  if ($22) {
   return;
  }
  $23 = (($0) + ($n2))|0;
  $24 = (($r) + ($23<<2)|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = (($25) + ($21))|0;
  HEAP32[$24>>2] = $26;
  $27 = ($26>>>0)<($21>>>0);
  if ($27) {
   $p$0 = $24;
  } else {
   return;
  }
  while(1) {
   $28 = ((($p$0)) + 4|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = (($29) + 1)|0;
   HEAP32[$28>>2] = $30;
   $31 = ($30|0)==(0);
   if ($31) {
    $p$0 = $28;
   } else {
    break;
   }
  }
  return;
 }
 }
}
function _egbn_sqr_normal($r,$a,$n,$tmp) {
 $r = $r|0;
 $a = $a|0;
 $n = $n|0;
 $tmp = $tmp|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $ap$13 = 0, $i$02 = 0, $j$01 = 0, $rp$14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $n << 1;
 $1 = (($0) + -1)|0;
 $2 = (($r) + ($1<<2)|0);
 HEAP32[$2>>2] = 0;
 HEAP32[$r>>2] = 0;
 $3 = (($n) + -1)|0;
 $4 = ($n|0)>(1);
 if ($4) {
  $5 = ((($r)) + 4|0);
  $6 = ((($a)) + 4|0);
  $7 = HEAP32[$a>>2]|0;
  $8 = (_egbn_mul_words($5,$6,$3,$7)|0);
  $9 = (($r) + ($n<<2)|0);
  HEAP32[$9>>2] = $8;
  $10 = ($n|0)>(2);
  if ($10) {
   $11 = ((($r)) + 12|0);
   $12 = (($n) + -2)|0;
   $ap$13 = $6;$i$02 = $12;$j$01 = $3;$rp$14 = $11;
   while(1) {
    $13 = (($j$01) + -1)|0;
    $14 = ((($ap$13)) + 4|0);
    $15 = HEAP32[$ap$13>>2]|0;
    $16 = (_egbn_mul_add_words($rp$14,$14,$13,$15)|0);
    $17 = (($rp$14) + ($13<<2)|0);
    HEAP32[$17>>2] = $16;
    $18 = ((($rp$14)) + 8|0);
    $19 = (($i$02) + -1)|0;
    $20 = ($i$02|0)>(1);
    if ($20) {
     $ap$13 = $14;$i$02 = $19;$j$01 = $13;$rp$14 = $18;
    } else {
     break;
    }
   }
  }
 }
 (_egbn_add_words($r,$r,$r,$0)|0);
 _egbn_sqr_words($tmp,$a,$n);
 (_egbn_add_words($r,$r,$tmp,$0)|0);
 return;
}
function _egbn_sqr_words($r,$a,$n) {
 $r = $r|0;
 $a = $a|0;
 $n = $n|0;
 var $$0$lcssa11 = 0, $$01$lcssa12 = 0, $$014 = 0, $$02$lcssa13 = 0, $$023 = 0, $$05 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $scevgep = 0, $scevgep10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($n|0)<(1);
 if ($0) {
  return;
 }
 $1 = ($n>>>0)>(3);
 if ($1) {
  $2 = (($n) + -4)|0;
  $3 = $2 >>> 2;
  $4 = $3 << 2;
  $5 = $3 << 3;
  $6 = (($5) + 8)|0;
  $7 = (($4) + 4)|0;
  $$014 = $a;$$023 = $n;$$05 = $r;
  while(1) {
   $8 = HEAP32[$$014>>2]|0;
   $9 = (___muldi3(($8|0),0,($8|0),0)|0);
   $10 = tempRet0;
   HEAP32[$$05>>2] = $9;
   $11 = ((($$05)) + 4|0);
   HEAP32[$11>>2] = $10;
   $12 = ((($$014)) + 4|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = (___muldi3(($13|0),0,($13|0),0)|0);
   $15 = tempRet0;
   $16 = ((($$05)) + 8|0);
   HEAP32[$16>>2] = $14;
   $17 = ((($$05)) + 12|0);
   HEAP32[$17>>2] = $15;
   $18 = ((($$014)) + 8|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = (___muldi3(($19|0),0,($19|0),0)|0);
   $21 = tempRet0;
   $22 = ((($$05)) + 16|0);
   HEAP32[$22>>2] = $20;
   $23 = ((($$05)) + 20|0);
   HEAP32[$23>>2] = $21;
   $24 = ((($$014)) + 12|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = (___muldi3(($25|0),0,($25|0),0)|0);
   $27 = tempRet0;
   $28 = ((($$05)) + 24|0);
   HEAP32[$28>>2] = $26;
   $29 = ((($$05)) + 28|0);
   HEAP32[$29>>2] = $27;
   $30 = ((($$014)) + 16|0);
   $31 = ((($$05)) + 32|0);
   $32 = (($$023) + -4)|0;
   $33 = ($32>>>0)>(3);
   if ($33) {
    $$014 = $30;$$023 = $32;$$05 = $31;
   } else {
    break;
   }
  }
  $34 = (($2) - ($4))|0;
  $scevgep = (($r) + ($6<<2)|0);
  $scevgep10 = (($a) + ($7<<2)|0);
  $35 = ($2|0)==($4|0);
  if ($35) {
   return;
  } else {
   $$0$lcssa11 = $scevgep;$$01$lcssa12 = $scevgep10;$$02$lcssa13 = $34;
  }
 } else {
  $$0$lcssa11 = $r;$$01$lcssa12 = $a;$$02$lcssa13 = $n;
 }
 $36 = HEAP32[$$01$lcssa12>>2]|0;
 $37 = (___muldi3(($36|0),0,($36|0),0)|0);
 $38 = tempRet0;
 HEAP32[$$0$lcssa11>>2] = $37;
 $39 = ((($$0$lcssa11)) + 4|0);
 HEAP32[$39>>2] = $38;
 $40 = ($$02$lcssa13|0)==(1);
 if ($40) {
  return;
 }
 $41 = ((($$01$lcssa12)) + 4|0);
 $42 = HEAP32[$41>>2]|0;
 $43 = (___muldi3(($42|0),0,($42|0),0)|0);
 $44 = tempRet0;
 $45 = ((($$0$lcssa11)) + 8|0);
 HEAP32[$45>>2] = $43;
 $46 = ((($$0$lcssa11)) + 12|0);
 HEAP32[$46>>2] = $44;
 $47 = ($$02$lcssa13|0)==(2);
 if ($47) {
  return;
 }
 $48 = ((($$01$lcssa12)) + 8|0);
 $49 = HEAP32[$48>>2]|0;
 $50 = (___muldi3(($49|0),0,($49|0),0)|0);
 $51 = tempRet0;
 $52 = ((($$0$lcssa11)) + 16|0);
 HEAP32[$52>>2] = $50;
 $53 = ((($$0$lcssa11)) + 20|0);
 HEAP32[$53>>2] = $51;
 return;
}
function _EGBN_MONT_CTX_new() {
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_malloc(72)|0);
 $1 = ($0|0)==(0|0);
 if ($1) {
  $$0 = 0;
  return ($$0|0);
 }
 HEAP32[$0>>2] = 0;
 $2 = ((($0)) + 4|0);
 _EGBN_init($2);
 $3 = ((($0)) + 24|0);
 _EGBN_init($3);
 $4 = ((($0)) + 44|0);
 _EGBN_init($4);
 $5 = ((($0)) + 68|0);
 HEAP32[$5>>2] = 1;
 $$0 = $0;
 return ($$0|0);
}
function _EGBN_MONT_CTX_set($mont,$mod,$ctx) {
 $mont = $mont|0;
 $mod = $mod|0;
 $ctx = $ctx|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $Ri = 0, $buf = 0, $tmod = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $Ri = sp + 28|0;
 $tmod = sp + 8|0;
 $buf = sp;
 _EGBN_init($Ri);
 $0 = ((($mont)) + 4|0);
 $1 = ((($mont)) + 24|0);
 (_EGBN_copy($1,$mod)|0);
 $2 = ((($mont)) + 36|0);
 HEAP32[$2>>2] = 0;
 $3 = (_EGBN_num_bits($mod)|0);
 $4 = (($3) + 31)|0;
 $5 = (($4|0) / 32)&-1;
 $6 = $5 << 5;
 HEAP32[$mont>>2] = $6;
 $7 = (_EGBN_set_word($0,0)|0);
 $8 = ($7|0)==(0);
 do {
  if (!($8)) {
   $9 = (_EGBN_set_bit($0,32)|0);
   $10 = ($9|0)==(0);
   if (!($10)) {
    $11 = HEAP32[$mod>>2]|0;
    $12 = HEAP32[$11>>2]|0;
    HEAP32[$buf>>2] = $12;
    $13 = ((($buf)) + 4|0);
    HEAP32[$13>>2] = 0;
    HEAP32[$tmod>>2] = $buf;
    $14 = ((($tmod)) + 4|0);
    HEAP32[$14>>2] = 1;
    $15 = ((($tmod)) + 8|0);
    HEAP32[$15>>2] = 2;
    $16 = ((($tmod)) + 12|0);
    HEAP32[$16>>2] = 0;
    $17 = (_EGBN_mod_inverse($Ri,$0,$tmod,$ctx)|0);
    $18 = ($17|0)==(0|0);
    if (!($18)) {
     $19 = (_EGBN_lshift($Ri,$Ri,32)|0);
     $20 = ($19|0)==(0);
     if (!($20)) {
      $21 = ((($Ri)) + 4|0);
      $22 = HEAP32[$21>>2]|0;
      switch ($22|0) {
      case 0:  {
       label = 8;
       break;
      }
      case 1:  {
       $23 = HEAP32[$Ri>>2]|0;
       $24 = HEAP32[$23>>2]|0;
       $25 = ($24|0)==(0);
       if ($25) {
        label = 8;
       } else {
        label = 7;
       }
       break;
      }
      default: {
       label = 7;
      }
      }
      if ((label|0) == 7) {
       $26 = (_EGBN_sub_word($Ri,1)|0);
       $27 = ($26|0)==(0);
       if ($27) {
        break;
       }
      }
      else if ((label|0) == 8) {
       $28 = (_EGBN_set_word($Ri,-1)|0);
       $29 = ($28|0)==(0);
       if ($29) {
        break;
       }
      }
      $30 = (_EGBN_div($Ri,0,$Ri,$tmod,$ctx)|0);
      $31 = ($30|0)==(0);
      if (!($31)) {
       $32 = HEAP32[$21>>2]|0;
       $33 = ($32|0)>(0);
       if ($33) {
        $34 = HEAP32[$Ri>>2]|0;
        $35 = HEAP32[$34>>2]|0;
        $37 = $35;
       } else {
        $37 = 0;
       }
       $36 = ((($mont)) + 64|0);
       HEAP32[$36>>2] = $37;
       _EGBN_free($Ri);
       $38 = (_EGBN_set_word($0,0)|0);
       $39 = ($38|0)==(0);
       if (!($39)) {
        $40 = HEAP32[$mont>>2]|0;
        $41 = $40 << 1;
        $42 = (_EGBN_set_bit($0,$41)|0);
        $43 = ($42|0)==(0);
        if (!($43)) {
         $44 = (_EGBN_div(0,$0,$0,$1,$ctx)|0);
         $45 = ($44|0)==(0);
         if (!($45)) {
          $$0 = 1;
          STACKTOP = sp;return ($$0|0);
         }
        }
       }
      }
     }
    }
   }
  }
 } while(0);
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _EGBN_mod_mul_montgomery($r,$a,$b,$mont,$ctx) {
 $r = $r|0;
 $a = $a|0;
 $b = $b|0;
 $mont = $mont|0;
 $ctx = $ctx|0;
 var $$ = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $not$ = 0, $ret$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _EGBN_CTX_start($ctx);
 $0 = (_EGBN_CTX_get($ctx)|0);
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $ret$0 = 0;
  } else {
   $2 = ($a|0)==($b|0);
   if ($2) {
    $3 = (_EGBN_sqr($0,$a,$ctx)|0);
    $4 = ($3|0)==(0);
    if ($4) {
     $ret$0 = 0;
     break;
    }
   } else {
    $5 = (_EGBN_mul($0,$a,$b,$ctx)|0);
    $6 = ($5|0)==(0);
    if ($6) {
     $ret$0 = 0;
     break;
    }
   }
   $7 = (_EGBN_from_montgomery($r,$0,$mont,$ctx)|0);
   $not$ = ($7|0)!=(0);
   $$ = $not$&1;
   $ret$0 = $$;
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 return ($ret$0|0);
}
function _EGBN_value_one() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (7424|0);
}
function _EGBN_from_montgomery($ret,$a,$mont,$ctx) {
 $ret = $ret|0;
 $a = $a|0;
 $mont = $mont|0;
 $ctx = $ctx|0;
 var $$0 = 0, $$off = 0, $$op$op$op = 0, $$op$op32 = 0, $$pr = 0, $$pre = 0, $$sum = 0, $$sum2 = 0, $$sum3 = 0, $$sum4 = 0, $$sum5 = 0, $$sum6 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $al$0 = 0, $exitcond = 0, $ftl$014 = 0, $i$118 = 0, $i$2$lcssa = 0, $i$211 = 0, $i$310 = 0, $nrp$019 = 0, $or$cond = 0, $retn$0 = 0, $rp$020 = 0, $scevgep = 0, $smax = 0, $x$0 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 _EGBN_CTX_start($ctx);
 $0 = (_EGBN_CTX_get($ctx)|0);
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $retn$0 = 0;
  } else {
   $2 = (_EGBN_copy($0,$a)|0);
   $3 = ($2|0)==(0|0);
   if ($3) {
    $retn$0 = 0;
   } else {
    $4 = ((($mont)) + 24|0);
    $5 = HEAP32[$mont>>2]|0;
    $6 = (($5|0) / 32)&-1;
    $7 = ((($mont)) + 28|0);
    $8 = HEAP32[$7>>2]|0;
    $$off = (($5) + 31)|0;
    $9 = ($$off>>>0)<(63);
    $10 = ($8|0)==(0);
    $or$cond = $9 | $10;
    if ($or$cond) {
     $11 = ((($0)) + 4|0);
     HEAP32[$11>>2] = 0;
     $$0 = 1;
     return ($$0|0);
    }
    $12 = (($8) + ($6))|0;
    $13 = (($12) + 1)|0;
    $14 = ((($0)) + 8|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ($12|0)<($15|0);
    if (!($16)) {
     $17 = (_egbn_expand2($0,$13)|0);
     $18 = ($17|0)==(0|0);
     if ($18) {
      $retn$0 = 0;
      break;
     }
    }
    $19 = ((($ret)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($12|0)<($20|0);
    if (!($21)) {
     $22 = (_egbn_expand2($ret,$13)|0);
     $23 = ($22|0)==(0|0);
     if ($23) {
      $retn$0 = 0;
      break;
     }
    }
    $24 = ((($a)) + 12|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ((($mont)) + 36|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = $27 ^ $25;
    $29 = ((($0)) + 12|0);
    HEAP32[$29>>2] = $28;
    $30 = HEAP32[$4>>2]|0;
    $31 = HEAP32[$0>>2]|0;
    $32 = (($31) + ($8<<2)|0);
    $33 = ((($0)) + 4|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ($13|0)>($34|0);
    if ($35) {
     $scevgep = (($31) + ($34<<2)|0);
     $36 = (($13) - ($34))|0;
     $37 = $36 << 2;
     _memset(($scevgep|0),0,($37|0))|0;
    }
    HEAP32[$33>>2] = $13;
    $38 = ((($mont)) + 64|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = ($8|0)>(0);
    if ($40) {
     $i$118 = 0;$nrp$019 = $32;$rp$020 = $31;
     while(1) {
      $41 = HEAP32[$rp$020>>2]|0;
      $42 = Math_imul($41, $39)|0;
      $43 = (_egbn_mul_add_words($rp$020,$30,$8,$42)|0);
      $44 = ((($nrp$019)) + 4|0);
      $45 = ((($rp$020)) + 4|0);
      $46 = HEAP32[$nrp$019>>2]|0;
      $47 = (($46) + ($43))|0;
      HEAP32[$nrp$019>>2] = $47;
      $48 = ($47>>>0)<($43>>>0);
      if ($48) {
       $49 = HEAP32[$44>>2]|0;
       $50 = (($49) + 1)|0;
       HEAP32[$44>>2] = $50;
       $51 = ($50|0)==(0);
       if ($51) {
        $52 = ((($nrp$019)) + 8|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = (($53) + 1)|0;
        HEAP32[$52>>2] = $54;
        $55 = ($54|0)==(0);
        if ($55) {
         $x$0 = 2;
         while(1) {
          $$sum6 = (($x$0) + 1)|0;
          $56 = (($nrp$019) + ($$sum6<<2)|0);
          $57 = HEAP32[$56>>2]|0;
          $58 = (($57) + 1)|0;
          HEAP32[$56>>2] = $58;
          $59 = ($58|0)==(0);
          if ($59) {
           $x$0 = $$sum6;
          } else {
           break;
          }
         }
        }
       }
      }
      $60 = (($i$118) + 1)|0;
      $exitcond = ($60|0)==($8|0);
      if ($exitcond) {
       break;
      } else {
       $i$118 = $60;$nrp$019 = $44;$rp$020 = $45;
      }
     }
     $$pre = HEAP32[$33>>2]|0;
     $$pr = $$pre;
    } else {
     $$pr = $13;
    }
    $61 = ($$pr|0)>(0);
    L28: do {
     if ($61) {
      $62 = HEAP32[$0>>2]|0;
      $63 = (($$pr) + -1)|0;
      $64 = (($62) + ($63<<2)|0);
      $69 = $$pr;$ftl$014 = $64;
      while(1) {
       $65 = HEAP32[$ftl$014>>2]|0;
       $66 = ($65|0)==(0);
       if (!($66)) {
        $76 = $69;
        break L28;
       }
       $67 = ((($ftl$014)) + -4|0);
       $68 = (($69) + -1)|0;
       HEAP32[$33>>2] = $68;
       $70 = ($69|0)>(1);
       if ($70) {
        $69 = $68;$ftl$014 = $67;
       } else {
        $76 = $68;
        break;
       }
      }
     } else {
      $76 = $$pr;
     }
    } while(0);
    $71 = HEAP32[$29>>2]|0;
    $72 = ((($ret)) + 12|0);
    HEAP32[$72>>2] = $71;
    $73 = HEAP32[$ret>>2]|0;
    $74 = HEAP32[$0>>2]|0;
    $75 = ($76|0)<($6|0);
    $77 = (($76) - ($6))|0;
    $al$0 = $75 ? 0 : $77;
    $78 = ((($ret)) + 4|0);
    HEAP32[$78>>2] = $al$0;
    $79 = (($al$0) + -4)|0;
    $80 = ($al$0|0)>(4);
    if ($80) {
     $81 = ($76|0)>($6|0);
     $smax = $81 ? $76 : $6;
     $82 = (($smax) + -4)|0;
     $83 = (($82) - ($6))|0;
     $84 = ($83|0)>(4);
     $i$211 = 0;
     while(1) {
      $$sum2 = (($i$211) + ($6))|0;
      $87 = (($74) + ($$sum2<<2)|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = $i$211 | 1;
      $$sum3 = (($89) + ($6))|0;
      $90 = (($74) + ($$sum3<<2)|0);
      $91 = HEAP32[$90>>2]|0;
      $92 = $i$211 | 2;
      $$sum4 = (($92) + ($6))|0;
      $93 = (($74) + ($$sum4<<2)|0);
      $94 = HEAP32[$93>>2]|0;
      $95 = $i$211 | 3;
      $$sum5 = (($95) + ($6))|0;
      $96 = (($74) + ($$sum5<<2)|0);
      $97 = HEAP32[$96>>2]|0;
      $98 = (($73) + ($i$211<<2)|0);
      HEAP32[$98>>2] = $88;
      $99 = (($73) + ($89<<2)|0);
      HEAP32[$99>>2] = $91;
      $100 = (($73) + ($92<<2)|0);
      HEAP32[$100>>2] = $94;
      $101 = (($73) + ($95<<2)|0);
      HEAP32[$101>>2] = $97;
      $102 = (($i$211) + 4)|0;
      $103 = ($102|0)<($79|0);
      if ($103) {
       $i$211 = $102;
      } else {
       break;
      }
     }
     $$op$op32 = (($83) + 3)|0;
     $$op$op$op = $$op$op32 & -4;
     $85 = $84 ? $$op$op$op : 4;
     $i$2$lcssa = $85;
    } else {
     $i$2$lcssa = 0;
    }
    $86 = ($i$2$lcssa|0)<($al$0|0);
    if ($86) {
     $i$310 = $i$2$lcssa;
     while(1) {
      $$sum = (($i$310) + ($6))|0;
      $104 = (($74) + ($$sum<<2)|0);
      $105 = HEAP32[$104>>2]|0;
      $106 = (($73) + ($i$310<<2)|0);
      HEAP32[$106>>2] = $105;
      $107 = (($i$310) + 1)|0;
      $108 = ($107|0)<($al$0|0);
      if ($108) {
       $i$310 = $107;
      } else {
       break;
      }
     }
    }
    $109 = (_EGBN_ucmp($ret,$4)|0);
    $110 = ($109|0)>(-1);
    if ($110) {
     $111 = (_EGBN_usub($ret,$ret,$4)|0);
     $112 = ($111|0)==(0);
     if ($112) {
      $retn$0 = 0;
      break;
     }
    }
    $retn$0 = 1;
   }
  }
 } while(0);
 _EGBN_CTX_end($ctx);
 $$0 = $retn$0;
 return ($$0|0);
}
function _EGBN_MONT_CTX_free($mont) {
 $mont = $mont|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($mont|0)==(0|0);
 if ($0) {
  return;
 }
 $1 = ((($mont)) + 4|0);
 _EGBN_free($1);
 $2 = ((($mont)) + 24|0);
 _EGBN_free($2);
 $3 = ((($mont)) + 44|0);
 _EGBN_free($3);
 $4 = ((($mont)) + 68|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & 1;
 $7 = ($6|0)==(0);
 if ($7) {
  return;
 }
 _free($mont);
 return;
}
function _EGBN_mod_inverse($in,$a,$n,$ctx) {
 $in = $in|0;
 $a = $a|0;
 $n = $n|0;
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0;
 var $170 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $A$0 = 0, $A$0$lcssa = 0, $A$0$phi = 0, $A$117 = 0, $A$118 = 0, $B$0 = 0, $B$0$phi = 0, $M$0 = 0, $M$0$phi = 0;
 var $R$05 = 0, $R$1 = 0, $X$0 = 0, $X$0$phi = 0, $Y$0 = 0, $Y$0$lcssa = 0, $Y$0$phi = 0, $Y$115 = 0, $Y$116 = 0, $or$cond = 0, $ret$0 = 0, $shift$0 = 0, $shift$0$lcssa = 0, $shift$1 = 0, $shift$1$lcssa = 0, $sign$0 = 0, $sign$0$lcssa = 0, label = 0, sp = 0;
 sp = STACKTOP;
 _EGBN_CTX_start($ctx);
 $0 = (_EGBN_CTX_get($ctx)|0);
 $1 = (_EGBN_CTX_get($ctx)|0);
 $2 = (_EGBN_CTX_get($ctx)|0);
 $3 = (_EGBN_CTX_get($ctx)|0);
 $4 = (_EGBN_CTX_get($ctx)|0);
 $5 = (_EGBN_CTX_get($ctx)|0);
 $6 = (_EGBN_CTX_get($ctx)|0);
 $7 = ($6|0)==(0|0);
 L1: do {
  if ($7) {
   $R$1 = 0;$ret$0 = 0;
  } else {
   $8 = ($in|0)==(0|0);
   if ($8) {
    $9 = (_EGBN_new()|0);
    $10 = ($9|0)==(0|0);
    if ($10) {
     $R$1 = 0;$ret$0 = 0;
     break;
    } else {
     $R$05 = $9;
    }
   } else {
    $R$05 = $in;
   }
   (_EGBN_set_word($2,1)|0);
   (_EGBN_set_word($5,0)|0);
   $11 = (_EGBN_copy($1,$a)|0);
   $12 = ($11|0)==(0|0);
   if ($12) {
    $R$1 = $R$05;$ret$0 = 0;
   } else {
    $13 = (_EGBN_copy($0,$n)|0);
    $14 = ($13|0)==(0|0);
    if ($14) {
     $R$1 = $R$05;$ret$0 = 0;
    } else {
     $15 = ((($0)) + 12|0);
     HEAP32[$15>>2] = 0;
     $16 = ((($1)) + 12|0);
     $17 = HEAP32[$16>>2]|0;
     $18 = ($17|0)==(0);
     if ($18) {
      $19 = (_EGBN_ucmp($1,$0)|0);
      $20 = ($19|0)>(-1);
      if ($20) {
       label = 8;
      }
     } else {
      label = 8;
     }
     if ((label|0) == 8) {
      $21 = (_EGBN_nnmod($1,$1,$0,$ctx)|0);
      $22 = ($21|0)==(0);
      if ($22) {
       $R$1 = $R$05;$ret$0 = 0;
       break;
      }
     }
     $23 = ((($n)) + 4|0);
     $24 = HEAP32[$23>>2]|0;
     $25 = ($24|0)>(0);
     L13: do {
      if ($25) {
       $26 = HEAP32[$n>>2]|0;
       $27 = HEAP32[$26>>2]|0;
       $28 = $27 & 1;
       $29 = ($28|0)==(0);
       if ($29) {
        label = 11;
       } else {
        $32 = (_EGBN_num_bits($n)|0);
        $33 = ($32|0)<(2049);
        if ($33) {
         $34 = ((($1)) + 4|0);
         $35 = ((($2)) + 4|0);
         $36 = ((($5)) + 4|0);
         while(1) {
          $37 = HEAP32[$34>>2]|0;
          switch ($37|0) {
          case 0:  {
           $A$118 = $0;$Y$115 = $5;
           label = 66;
           break L13;
           break;
          }
          case 1:  {
           $38 = HEAP32[$1>>2]|0;
           $39 = HEAP32[$38>>2]|0;
           $40 = ($39|0)==(0);
           if ($40) {
            $A$118 = $0;$Y$115 = $5;
            label = 66;
            break L13;
           } else {
            $shift$0 = 0;
           }
           break;
          }
          default: {
           $shift$0 = 0;
          }
          }
          while(1) {
           $41 = (_EGBN_is_bit_set($1,$shift$0)|0);
           $42 = ($41|0)==(0);
           if (!($42)) {
            $shift$0$lcssa = $shift$0;
            break;
           }
           $43 = (($shift$0) + 1)|0;
           $44 = HEAP32[$35>>2]|0;
           $45 = ($44|0)>(0);
           if ($45) {
            $46 = HEAP32[$2>>2]|0;
            $47 = HEAP32[$46>>2]|0;
            $48 = $47 & 1;
            $49 = ($48|0)==(0);
            if (!($49)) {
             $50 = (_EGBN_uadd($2,$2,$n)|0);
             $51 = ($50|0)==(0);
             if ($51) {
              $R$1 = $R$05;$ret$0 = 0;
              break L1;
             }
            }
           }
           $52 = (_EGBN_rshift1($2,$2)|0);
           $53 = ($52|0)==(0);
           if ($53) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            $shift$0 = $43;
           }
          }
          $54 = ($shift$0$lcssa|0)>(0);
          if ($54) {
           $55 = (_EGBN_rshift($1,$1,$shift$0$lcssa)|0);
           $56 = ($55|0)==(0);
           if ($56) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            $shift$1 = 0;
           }
          } else {
           $shift$1 = 0;
          }
          while(1) {
           $57 = (_EGBN_is_bit_set($0,$shift$1)|0);
           $58 = ($57|0)==(0);
           if (!($58)) {
            $shift$1$lcssa = $shift$1;
            break;
           }
           $59 = (($shift$1) + 1)|0;
           $60 = HEAP32[$36>>2]|0;
           $61 = ($60|0)>(0);
           if ($61) {
            $62 = HEAP32[$5>>2]|0;
            $63 = HEAP32[$62>>2]|0;
            $64 = $63 & 1;
            $65 = ($64|0)==(0);
            if (!($65)) {
             $66 = (_EGBN_uadd($5,$5,$n)|0);
             $67 = ($66|0)==(0);
             if ($67) {
              $R$1 = $R$05;$ret$0 = 0;
              break L1;
             }
            }
           }
           $68 = (_EGBN_rshift1($5,$5)|0);
           $69 = ($68|0)==(0);
           if ($69) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            $shift$1 = $59;
           }
          }
          $70 = ($shift$1$lcssa|0)>(0);
          if ($70) {
           $71 = (_EGBN_rshift($0,$0,$shift$1$lcssa)|0);
           $72 = ($71|0)==(0);
           if ($72) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           }
          }
          $73 = (_EGBN_ucmp($1,$0)|0);
          $74 = ($73|0)>(-1);
          if ($74) {
           $75 = (_EGBN_uadd($2,$2,$5)|0);
           $76 = ($75|0)==(0);
           if ($76) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           }
           $77 = (_EGBN_usub($1,$1,$0)|0);
           $78 = ($77|0)==(0);
           if ($78) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            continue;
           }
          } else {
           $79 = (_EGBN_uadd($5,$5,$2)|0);
           $80 = ($79|0)==(0);
           if ($80) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           }
           $81 = (_EGBN_usub($0,$0,$1)|0);
           $82 = ($81|0)==(0);
           if ($82) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            continue;
           }
          }
         }
        } else {
         label = 11;
        }
       }
      } else {
       label = 11;
      }
     } while(0);
     if ((label|0) == 11) {
      $30 = ((($3)) + 4|0);
      $31 = ((($3)) + 12|0);
      $A$0 = $0;$B$0 = $1;$M$0 = $4;$X$0 = $2;$Y$0 = $5;$sign$0 = -1;
      L49: while(1) {
       $83 = ((($B$0)) + 4|0);
       $84 = HEAP32[$83>>2]|0;
       switch ($84|0) {
       case 0:  {
        $A$0$lcssa = $A$0;$Y$0$lcssa = $Y$0;$sign$0$lcssa = $sign$0;
        break L49;
        break;
       }
       case 1:  {
        $85 = HEAP32[$B$0>>2]|0;
        $86 = HEAP32[$85>>2]|0;
        $87 = ($86|0)==(0);
        if ($87) {
         $A$0$lcssa = $A$0;$Y$0$lcssa = $Y$0;$sign$0$lcssa = $sign$0;
         break L49;
        }
        break;
       }
       default: {
       }
       }
       $88 = (_EGBN_num_bits($A$0)|0);
       $89 = (_EGBN_num_bits($B$0)|0);
       $90 = ($88|0)==($89|0);
       do {
        if ($90) {
         $91 = (_EGBN_set_word($3,1)|0);
         $92 = ($91|0)==(0);
         if ($92) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $93 = (_EGBN_sub($M$0,$A$0,$B$0)|0);
         $94 = ($93|0)==(0);
         if ($94) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
        } else {
         $95 = (_EGBN_num_bits($A$0)|0);
         $96 = (_EGBN_num_bits($B$0)|0);
         $97 = (($96) + 1)|0;
         $98 = ($95|0)==($97|0);
         if (!($98)) {
          $119 = (_EGBN_div($3,$M$0,$A$0,$B$0,$ctx)|0);
          $120 = ($119|0)==(0);
          if ($120) {
           $R$1 = $R$05;$ret$0 = 0;
           break L1;
          } else {
           break;
          }
         }
         $99 = (_EGBN_lshift1($6,$B$0)|0);
         $100 = ($99|0)==(0);
         if ($100) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $101 = (_EGBN_ucmp($A$0,$6)|0);
         $102 = ($101|0)<(0);
         if ($102) {
          $103 = (_EGBN_set_word($3,1)|0);
          $104 = ($103|0)==(0);
          if ($104) {
           $R$1 = $R$05;$ret$0 = 0;
           break L1;
          }
          $105 = (_EGBN_sub($M$0,$A$0,$B$0)|0);
          $106 = ($105|0)==(0);
          if ($106) {
           $R$1 = $R$05;$ret$0 = 0;
           break L1;
          } else {
           break;
          }
         }
         $107 = (_EGBN_sub($M$0,$A$0,$6)|0);
         $108 = ($107|0)==(0);
         if ($108) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $109 = (_EGBN_add($3,$6,$B$0)|0);
         $110 = ($109|0)==(0);
         if ($110) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $111 = (_EGBN_ucmp($A$0,$3)|0);
         $112 = ($111|0)<(0);
         if ($112) {
          $113 = (_EGBN_set_word($3,2)|0);
          $114 = ($113|0)==(0);
          if ($114) {
           $R$1 = $R$05;$ret$0 = 0;
           break L1;
          } else {
           break;
          }
         }
         $115 = (_EGBN_set_word($3,3)|0);
         $116 = ($115|0)==(0);
         if ($116) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $117 = (_EGBN_sub($M$0,$M$0,$B$0)|0);
         $118 = ($117|0)==(0);
         if ($118) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
        }
       } while(0);
       $121 = HEAP32[$30>>2]|0;
       $122 = ($121|0)==(1);
       L73: do {
        if ($122) {
         $123 = HEAP32[$3>>2]|0;
         $124 = HEAP32[$123>>2]|0;
         switch ($124|0) {
         case 1:  {
          $125 = HEAP32[$31>>2]|0;
          $126 = ($125|0)==(0);
          if ($126) {
           $127 = (_EGBN_add($A$0,$X$0,$Y$0)|0);
           $128 = ($127|0)==(0);
           if ($128) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            break L73;
           }
          }
          break;
         }
         case 2:  {
          $129 = HEAP32[$31>>2]|0;
          $130 = ($129|0)==(0);
          if ($130) {
           $131 = (_EGBN_lshift1($A$0,$X$0)|0);
           $132 = ($131|0)==(0);
           if ($132) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            label = 63;
            break L73;
           }
          }
          break;
         }
         case 4:  {
          $133 = HEAP32[$31>>2]|0;
          $134 = ($133|0)==(0);
          if ($134) {
           $135 = (_EGBN_lshift($A$0,$X$0,2)|0);
           $136 = ($135|0)==(0);
           if ($136) {
            $R$1 = $R$05;$ret$0 = 0;
            break L1;
           } else {
            label = 63;
            break L73;
           }
          }
          break;
         }
         default: {
         }
         }
         $137 = (_EGBN_copy($A$0,$X$0)|0);
         $138 = ($137|0)==(0|0);
         if ($138) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         }
         $139 = HEAP32[$3>>2]|0;
         $140 = HEAP32[$139>>2]|0;
         $141 = (_EGBN_mul_word($A$0,$140)|0);
         $142 = ($141|0)==(0);
         if ($142) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         } else {
          label = 63;
         }
        } else {
         $143 = (_EGBN_mul($A$0,$3,$X$0,$ctx)|0);
         $144 = ($143|0)==(0);
         if ($144) {
          $R$1 = $R$05;$ret$0 = 0;
          break L1;
         } else {
          label = 63;
         }
        }
       } while(0);
       if ((label|0) == 63) {
        label = 0;
        $145 = (_EGBN_add($A$0,$A$0,$Y$0)|0);
        $146 = ($145|0)==(0);
        if ($146) {
         $R$1 = $R$05;$ret$0 = 0;
         break L1;
        }
       }
       $147 = (0 - ($sign$0))|0;
       $Y$0$phi = $X$0;$X$0$phi = $A$0;$M$0$phi = $Y$0;$B$0$phi = $M$0;$A$0$phi = $B$0;$sign$0 = $147;$Y$0 = $Y$0$phi;$X$0 = $X$0$phi;$M$0 = $M$0$phi;$B$0 = $B$0$phi;$A$0 = $A$0$phi;
      }
      $148 = ($sign$0$lcssa|0)<(0);
      if ($148) {
       $A$118 = $A$0$lcssa;$Y$115 = $Y$0$lcssa;
       label = 66;
      } else {
       $A$117 = $A$0$lcssa;$Y$116 = $Y$0$lcssa;
      }
     }
     if ((label|0) == 66) {
      $149 = (_EGBN_sub($Y$115,$n,$Y$115)|0);
      $150 = ($149|0)==(0);
      if ($150) {
       $R$1 = $R$05;$ret$0 = 0;
       break;
      } else {
       $A$117 = $A$118;$Y$116 = $Y$115;
      }
     }
     $151 = ((($A$117)) + 4|0);
     $152 = HEAP32[$151>>2]|0;
     $153 = ($152|0)==(1);
     if ($153) {
      $154 = HEAP32[$A$117>>2]|0;
      $155 = HEAP32[$154>>2]|0;
      $156 = ($155|0)==(1);
      if ($156) {
       $157 = ((($A$117)) + 12|0);
       $158 = HEAP32[$157>>2]|0;
       $159 = ($158|0)==(0);
       if ($159) {
        $160 = ((($Y$116)) + 12|0);
        $161 = HEAP32[$160>>2]|0;
        $162 = ($161|0)==(0);
        if ($162) {
         $163 = (_EGBN_ucmp($Y$116,$n)|0);
         $164 = ($163|0)<(0);
         if ($164) {
          $165 = (_EGBN_copy($R$05,$Y$116)|0);
          $166 = ($165|0)==(0|0);
          if ($166) {
           $R$1 = $R$05;$ret$0 = 0;
           break;
          }
         } else {
          label = 73;
         }
        } else {
         label = 73;
        }
        if ((label|0) == 73) {
         $167 = (_EGBN_nnmod($R$05,$Y$116,$n,$ctx)|0);
         $168 = ($167|0)==(0);
         if ($168) {
          $R$1 = $R$05;$ret$0 = 0;
          break;
         }
        }
        $R$1 = $R$05;$ret$0 = $R$05;
       } else {
        $R$1 = $R$05;$ret$0 = 0;
       }
      } else {
       $R$1 = $R$05;$ret$0 = 0;
      }
     } else {
      $R$1 = $R$05;$ret$0 = 0;
     }
    }
   }
  }
 } while(0);
 $169 = ($ret$0|0)==(0|0);
 $170 = ($in|0)==(0|0);
 $or$cond = $170 & $169;
 if (!($or$cond)) {
  _EGBN_CTX_end($ctx);
  return ($ret$0|0);
 }
 _EGBN_free($R$1);
 _EGBN_CTX_end($ctx);
 return ($ret$0|0);
}
function _EGBN_sub_word($a,$w) {
 $a = $a|0;
 $w = $w|0;
 var $$0 = 0, $$0$lcssa = 0, $$01 = 0, $$lcssa = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i$0$lcssa = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 0:  {
  break;
 }
 case 1:  {
  $2 = HEAP32[$a>>2]|0;
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(0);
  if (!($4)) {
   label = 3;
  }
  break;
 }
 default: {
  label = 3;
 }
 }
 if ((label|0) == 3) {
  $5 = ((($a)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   $10 = ($1|0)==(1);
   $11 = HEAP32[$a>>2]|0;
   if ($10) {
    $12 = HEAP32[$11>>2]|0;
    $13 = ($12>>>0)<($w>>>0);
    if ($13) {
     $14 = (($w) - ($12))|0;
     HEAP32[$11>>2] = $14;
     HEAP32[$5>>2] = 1;
     $$01 = 1;
     return ($$01|0);
    } else {
     $$0 = $w;$i$0 = 0;
    }
   } else {
    $$0 = $w;$i$0 = 0;
   }
   while(1) {
    $15 = (($11) + ($i$0<<2)|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16>>>0)<($$0>>>0);
    $18 = (($16) - ($$0))|0;
    HEAP32[$15>>2] = $18;
    $19 = (($i$0) + 1)|0;
    if ($17) {
     $$0 = 1;$i$0 = $19;
    } else {
     $$0$lcssa = $$0;$$lcssa = $16;$i$0$lcssa = $i$0;
     break;
    }
   }
   $20 = ($$lcssa|0)==($$0$lcssa|0);
   $21 = (($1) + -1)|0;
   $22 = ($i$0$lcssa|0)==($21|0);
   $or$cond = $20 & $22;
   if (!($or$cond)) {
    $$01 = 1;
    return ($$01|0);
   }
   HEAP32[$0>>2] = $i$0$lcssa;
   $$01 = 1;
   return ($$01|0);
  }
 }
 $8 = ((($a)) + 12|0);
 HEAP32[$8>>2] = 0;
 $9 = (_EGBN_add_word($a,$w)|0);
 HEAP32[$8>>2] = 1;
 $$01 = $9;
 return ($$01|0);
}
function _EGBN_rshift1($r,$a) {
 $r = $r|0;
 $a = $a|0;
 var $$0 = 0, $$pre = 0, $$pre7 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $c$04 = 0, $ftl$01 = 0, $i$05 = 0, $i$05$in = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 0:  {
  label = 3;
  break;
 }
 case 1:  {
  $2 = HEAP32[$a>>2]|0;
  $3 = HEAP32[$2>>2]|0;
  $4 = ($3|0)==(0);
  if ($4) {
   label = 3;
  }
  break;
 }
 default: {
 }
 }
 if ((label|0) == 3) {
  (_EGBN_set_word($r,0)|0);
  $$0 = 1;
  return ($$0|0);
 }
 $5 = ($a|0)==($r|0);
 if ($5) {
  $19 = $1;
 } else {
  $6 = ((($r)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($1|0)>($7|0);
  do {
   if ($8) {
    $9 = (_egbn_expand2($r,$1)|0);
    $10 = ($9|0)==(0|0);
    if ($10) {
     $$0 = 0;
     return ($$0|0);
    } else {
     $$pre = HEAP32[$0>>2]|0;
     $12 = $$pre;
     break;
    }
   } else {
    $12 = $1;
   }
  } while(0);
  $11 = ((($r)) + 4|0);
  HEAP32[$11>>2] = $12;
  $13 = ((($a)) + 12|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($r)) + 12|0);
  HEAP32[$15>>2] = $14;
  $$pre7 = HEAP32[$0>>2]|0;
  $19 = $$pre7;
 }
 $16 = HEAP32[$a>>2]|0;
 $17 = HEAP32[$r>>2]|0;
 $18 = ($19|0)>(0);
 if ($18) {
  $c$04 = 0;$i$05$in = $19;
  while(1) {
   $i$05 = (($i$05$in) + -1)|0;
   $20 = (($16) + ($i$05<<2)|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = $21 >>> 1;
   $23 = $22 | $c$04;
   $24 = (($17) + ($i$05<<2)|0);
   HEAP32[$24>>2] = $23;
   $25 = $21 << 31;
   $26 = ($i$05$in|0)>(1);
   if ($26) {
    $c$04 = $25;$i$05$in = $i$05;
   } else {
    break;
   }
  }
 }
 $27 = ((($r)) + 4|0);
 $28 = HEAP32[$27>>2]|0;
 $29 = ($28|0)>(0);
 if (!($29)) {
  $$0 = 1;
  return ($$0|0);
 }
 $30 = (($28) + -1)|0;
 $31 = (($17) + ($30<<2)|0);
 $36 = $28;$ftl$01 = $31;
 while(1) {
  $32 = HEAP32[$ftl$01>>2]|0;
  $33 = ($32|0)==(0);
  if (!($33)) {
   $$0 = 1;
   label = 15;
   break;
  }
  $34 = ((($ftl$01)) + -4|0);
  $35 = (($36) + -1)|0;
  HEAP32[$27>>2] = $35;
  $37 = ($36|0)>(1);
  if ($37) {
   $36 = $35;$ftl$01 = $34;
  } else {
   $$0 = 1;
   label = 15;
   break;
  }
 }
 if ((label|0) == 15) {
  return ($$0|0);
 }
 return (0)|0;
}
function _EGBN_lshift1($r,$a) {
 $r = $r|0;
 $a = $a|0;
 var $$0 = 0, $$lobit = 0, $$lobit$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $ap$02 = 0, $c$04 = 0, $i$05 = 0, $phitmp = 0, $rp$03 = 0, $scevgep = 0, $smax = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($r|0)==($a|0);
 if ($0) {
  $14 = ((($r)) + 4|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = ((($r)) + 8|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = ($15|0)<($17|0);
  if (!($18)) {
   $19 = (($15) + 1)|0;
   $20 = (_egbn_expand2($r,$19)|0);
   $21 = ($20|0)==(0|0);
   if ($21) {
    $$0 = 0;
    return ($$0|0);
   }
  }
 } else {
  $1 = ((($a)) + 12|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = ((($r)) + 12|0);
  HEAP32[$3>>2] = $2;
  $4 = ((($a)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ((($r)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($5|0)<($7|0);
  do {
   if ($8) {
    $13 = $5;
   } else {
    $9 = (($5) + 1)|0;
    $10 = (_egbn_expand2($r,$9)|0);
    $11 = ($10|0)==(0|0);
    if ($11) {
     $$0 = 0;
     return ($$0|0);
    } else {
     $$pre = HEAP32[$4>>2]|0;
     $13 = $$pre;
     break;
    }
   }
  } while(0);
  $12 = ((($r)) + 4|0);
  HEAP32[$12>>2] = $13;
 }
 $22 = HEAP32[$r>>2]|0;
 $23 = ((($a)) + 4|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)>(0);
 if (!($25)) {
  $$0 = 1;
  return ($$0|0);
 }
 $26 = HEAP32[$a>>2]|0;
 $27 = ($24|0)>(1);
 $smax = $27 ? $24 : 1;
 $ap$02 = $26;$c$04 = 0;$i$05 = 0;$rp$03 = $22;
 while(1) {
  $28 = ((($ap$02)) + 4|0);
  $29 = HEAP32[$ap$02>>2]|0;
  $30 = $29 << 1;
  $31 = $30 | $c$04;
  $32 = ((($rp$03)) + 4|0);
  HEAP32[$rp$03>>2] = $31;
  $$lobit = $29 >>> 31;
  $33 = (($i$05) + 1)|0;
  $34 = ($33|0)<($24|0);
  if ($34) {
   $ap$02 = $28;$c$04 = $$lobit;$i$05 = $33;$rp$03 = $32;
  } else {
   $$lobit$lcssa = $$lobit;
   break;
  }
 }
 $phitmp = ($$lobit$lcssa|0)==(0);
 if ($phitmp) {
  $$0 = 1;
  return ($$0|0);
 }
 $scevgep = (($22) + ($smax<<2)|0);
 HEAP32[$scevgep>>2] = 1;
 $35 = ((($r)) + 4|0);
 $36 = HEAP32[$35>>2]|0;
 $37 = (($36) + 1)|0;
 HEAP32[$35>>2] = $37;
 $$0 = 1;
 return ($$0|0);
}
function _EGBN_mul_word($a,$w) {
 $a = $a|0;
 $w = $w|0;
 var $$0 = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($a)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if ($2) {
  $$0 = 1;
  return ($$0|0);
 }
 $3 = ($w|0)==(0);
 if ($3) {
  (_EGBN_set_word($a,0)|0);
  $$0 = 1;
  return ($$0|0);
 }
 $4 = HEAP32[$a>>2]|0;
 $5 = (_egbn_mul_words($4,$4,$1,$w)|0);
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 1;
  return ($$0|0);
 }
 $7 = HEAP32[$0>>2]|0;
 $8 = ((($a)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ($7|0)<($9|0);
 do {
  if ($10) {
   $15 = $7;
  } else {
   $11 = (($7) + 1)|0;
   $12 = (_egbn_expand2($a,$11)|0);
   $13 = ($12|0)==(0|0);
   if ($13) {
    $$0 = 0;
    return ($$0|0);
   } else {
    $$pre = HEAP32[$0>>2]|0;
    $15 = $$pre;
    break;
   }
  }
 } while(0);
 $14 = (($15) + 1)|0;
 HEAP32[$0>>2] = $14;
 $16 = HEAP32[$a>>2]|0;
 $17 = (($16) + ($15<<2)|0);
 HEAP32[$17>>2] = $5;
 $$0 = 1;
 return ($$0|0);
}
function __ZNK9ExitGames6Common14DictionaryBase8getValueIhEEPKNS0_6ObjectERKT_S5_($this,$key,$0) {
 $this = $this|0;
 $key = $key|0;
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $1 = sp + 80|0;
 $2 = sp + 40|0;
 $3 = sp;
 $4 = (__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv($this)|0);
 $5 = HEAP8[$key>>0]|0;
 HEAP8[$1>>0] = $5;
 __ZN9ExitGames6Common6ObjectC2EPKvhhb($2,$1,98,0,1);
 HEAP32[$2>>2] = (2768);
 __THREW__ = 0;
 invoke_vii(744,($3|0),($2|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZN9ExitGames6Common6ObjectD2Ev($2);
  ___resumeException($8|0);
  // unreachable;
 }
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 __THREW__ = 0;
 $10 = (invoke_iii(862,($4|0),($3|0))|0);
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZN9ExitGames6Common6ObjectD2Ev($3);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common6ObjectD2Ev($3);
  STACKTOP = sp;return ($10|0);
 }
 return (0)|0;
}
function _EGBN_rand($rnd,$bits,$top,$bottom) {
 $rnd = $rnd|0;
 $bits = $bits|0;
 $top = $top|0;
 $bottom = $bottom|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_bnrand(0,$rnd,$bits,$top,$bottom)|0);
 return ($0|0);
}
function __ZN9ExitGames6Photon16OperationRequestD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($0);
 return;
}
function __ZN9ExitGames6Photon16OperationRequestC2EhRKNS_6Common10DictionaryIhNS2_6ObjectEEE($this,$operationCode,$parameters) {
 $this = $this|0;
 $operationCode = $operationCode|0;
 $parameters = $parameters|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($0);
 HEAP32[$0>>2] = (2720);
 $1 = HEAP8[167373]|0;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  HEAP8[167374] = 98;
  HEAP8[167375] = 122;
  HEAP32[40417] = 0;
  HEAP8[167373] = 1;
 }
 HEAP8[$this>>0] = $operationCode;
 __THREW__ = 0;
 (invoke_iii(249,($0|0),($parameters|0))|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  __ZN9ExitGames6Common14DictionaryBaseD2Ev($0);
  ___resumeException($5|0);
  // unreachable;
 } else {
  return;
 }
}
function _bnrand($pseudorand,$rnd,$bits,$top,$bottom) {
 $pseudorand = $pseudorand|0;
 $rnd = $rnd|0;
 $bits = $bits|0;
 $top = $top|0;
 $bottom = $bottom|0;
 var $$ = 0, $$0 = 0, $$01$i = 0, $$01$i2 = 0, $$off = 0, $$pre = 0, $$sink = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $i$08 = 0, $not$ = 0, $or$cond = 0, $or$cond10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($bits|0)==(0);
 if ($0) {
  (_EGBN_set_word($rnd,0)|0);
  $$0 = 1;
  return ($$0|0);
 }
 $1 = (($bits) + 7)|0;
 $2 = (($1|0) / 8)&-1;
 $3 = (($bits) + -1)|0;
 $4 = (($3|0) % 8)&-1;
 $5 = (($4) + 1)|0;
 $6 = 255 << $5;
 $7 = (_malloc($2)|0);
 $8 = ($7|0)==(0|0);
 if ($8) {
  $$0 = 0;
  return ($$0|0);
 }
 $9 = ($pseudorand|0)==(0);
 $$off = (($bits) + 14)|0;
 $10 = ($$off>>>0)<(15);
 if ($9) {
  if (!($10)) {
   $$01$i2 = $2;
   while(1) {
    $16 = (_rand()|0);
    $17 = $16&255;
    $18 = (($$01$i2) + -1)|0;
    $19 = (($7) + ($18)|0);
    HEAP8[$19>>0] = $17;
    $20 = ($18|0)==(0);
    if ($20) {
     label = 9;
     break;
    } else {
     $$01$i2 = $18;
    }
   }
  }
 } else {
  if ($10) {
   label = 9;
  } else {
   $$01$i = $2;
   while(1) {
    $11 = (_rand()|0);
    $12 = $11&255;
    $13 = (($$01$i) + -1)|0;
    $14 = (($7) + ($13)|0);
    HEAP8[$14>>0] = $12;
    $15 = ($13|0)==(0);
    if ($15) {
     label = 9;
     break;
    } else {
     $$01$i = $13;
    }
   }
  }
 }
 if ((label|0) == 9) {
  $21 = ($pseudorand|0)==(2);
  $22 = ($bits|0)>(0);
  $or$cond10 = $21 & $22;
  if ($or$cond10) {
   $i$08 = 0;
   while(1) {
    $23 = (_rand()|0);
    $24 = $23&255;
    $25 = ($24<<24>>24)<(0);
    $26 = ($i$08|0)>(0);
    $or$cond = $26 & $25;
    if ($or$cond) {
     $27 = (($i$08) + -1)|0;
     $28 = (($7) + ($27)|0);
     $29 = HEAP8[$28>>0]|0;
     $$sink = $29;
     label = 14;
    } else {
     $30 = ($24&255)<(42);
     if ($30) {
      $$sink = 0;
      label = 14;
     } else {
      $31 = ($24&255)<(84);
      if ($31) {
       $$sink = -1;
       label = 14;
      }
     }
    }
    if ((label|0) == 14) {
     label = 0;
     $32 = (($7) + ($i$08)|0);
     HEAP8[$32>>0] = $$sink;
    }
    $33 = (($i$08) + 1)|0;
    $34 = ($33|0)<($2|0);
    if ($34) {
     $i$08 = $33;
    } else {
     break;
    }
   }
  }
 }
 L27: do {
  switch ($top|0) {
  case -1:  {
   $$pre = HEAP8[$7>>0]|0;
   $54 = $$pre;
   break;
  }
  case 0:  {
   $47 = 1 << $4;
   $48 = HEAP8[$7>>0]|0;
   $49 = $48&255;
   $50 = $49 | $47;
   $51 = $50&255;
   HEAP8[$7>>0] = $51;
   $54 = $51;
   break;
  }
  default: {
   $35 = ($4|0)==(0);
   if ($35) {
    HEAP8[$7>>0] = 1;
    $36 = ((($7)) + 1|0);
    $37 = HEAP8[$36>>0]|0;
    $38 = $37&255;
    $39 = $38 | 128;
    $40 = $39&255;
    HEAP8[$36>>0] = $40;
    $54 = 1;
    break L27;
   } else {
    $41 = (($4) + -1)|0;
    $42 = 3 << $41;
    $43 = HEAP8[$7>>0]|0;
    $44 = $43&255;
    $45 = $44 | $42;
    $46 = $45&255;
    HEAP8[$7>>0] = $46;
    $54 = $46;
    break L27;
   }
  }
  }
 } while(0);
 $52 = $6 ^ 255;
 $53 = $54&255;
 $55 = $53 & $52;
 $56 = $55&255;
 HEAP8[$7>>0] = $56;
 $57 = ($bottom|0)==(0);
 if (!($57)) {
  $58 = (($2) + -1)|0;
  $59 = (($7) + ($58)|0);
  $60 = HEAP8[$59>>0]|0;
  $61 = $60&255;
  $62 = $61 | 1;
  $63 = $62&255;
  HEAP8[$59>>0] = $63;
 }
 $64 = (_EGBN_bin2bn($7,$2,$rnd)|0);
 $not$ = ($64|0)!=(0|0);
 $$ = $not$&1;
 _free($7);
 $$0 = $$;
 return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal10Encryption7decryptEPKhiS4_PPhPi($encodedData,$encodedDataSize,$key,$plainData,$plainDataSize) {
 $encodedData = $encodedData|0;
 $encodedDataSize = $encodedDataSize|0;
 $key = $key|0;
 $plainData = $plainData|0;
 $plainDataSize = $plainDataSize|0;
 var $$0$i = 0, $$sum = 0, $$sum3 = 0, $$sum4$1 = 0, $$sum4$10 = 0, $$sum4$11 = 0, $$sum4$12 = 0, $$sum4$13 = 0, $$sum4$14 = 0, $$sum4$15 = 0, $$sum4$2 = 0, $$sum4$3 = 0, $$sum4$4 = 0, $$sum4$5 = 0, $$sum4$6 = 0, $$sum4$7 = 0, $$sum4$8 = 0, $$sum4$9 = 0, $$sum56 = 0, $0 = 0.0;
 var $1 = 0.0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $buf = 0, $ctx = 0, $exitcond = 0, $i$01$i1 = 0, $i$07 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $ctx = sp + 16|0;
 $buf = sp;
 $0 = (+($encodedDataSize|0));
 $1 = $0 * 0.0625;
 $2 = (+Math_ceil((+$1)));
 $3 = (~~(($2)));
 HEAP32[$plainDataSize>>2] = $encodedDataSize;
 $4 = (($encodedDataSize) + 8)|0;
 $5 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($4)|0);
 $6 = ((($5)) + 8|0);
 $7 = $5;
 $8 = $7;
 HEAP32[$8>>2] = $encodedDataSize;
 $9 = (($7) + 4)|0;
 $10 = $9;
 HEAP32[$10>>2] = 0;
 $11 = ($encodedDataSize|0)==(0);
 if (!($11)) {
  _memset(($6|0),0,($encodedDataSize|0))|0;
 }
 _egaes256_init($ctx,$key);
 $12 = ($3|0)>(0);
 if ($12) {
  $13 = ((($buf)) + 1|0);
  $14 = ((($buf)) + 2|0);
  $15 = ((($buf)) + 3|0);
  $16 = ((($buf)) + 4|0);
  $17 = ((($buf)) + 5|0);
  $18 = ((($buf)) + 6|0);
  $19 = ((($buf)) + 7|0);
  $20 = ((($buf)) + 8|0);
  $21 = ((($buf)) + 9|0);
  $22 = ((($buf)) + 10|0);
  $23 = ((($buf)) + 11|0);
  $24 = ((($buf)) + 12|0);
  $25 = ((($buf)) + 13|0);
  $26 = ((($buf)) + 14|0);
  $27 = ((($buf)) + 15|0);
  $i$07 = 0;
  while(1) {
   $28 = $i$07 << 4;
   $29 = (($encodedData) + ($28)|0);
   dest=$buf; src=$29; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
   _egaes256_decrypt_ecb($ctx,$buf);
   $30 = ($i$07|0)==(0);
   if (!($30)) {
    $31 = (($28) + -16)|0;
    $32 = HEAP8[$buf>>0]|0;
    $33 = (($encodedData) + ($31)|0);
    $34 = HEAP8[$33>>0]|0;
    $35 = $34 ^ $32;
    HEAP8[$buf>>0] = $35;
    $36 = HEAP8[$13>>0]|0;
    $$sum4$1 = (($31) + 1)|0;
    $37 = (($encodedData) + ($$sum4$1)|0);
    $38 = HEAP8[$37>>0]|0;
    $39 = $38 ^ $36;
    HEAP8[$13>>0] = $39;
    $40 = HEAP8[$14>>0]|0;
    $$sum4$2 = (($31) + 2)|0;
    $41 = (($encodedData) + ($$sum4$2)|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = $42 ^ $40;
    HEAP8[$14>>0] = $43;
    $44 = HEAP8[$15>>0]|0;
    $$sum4$3 = (($31) + 3)|0;
    $45 = (($encodedData) + ($$sum4$3)|0);
    $46 = HEAP8[$45>>0]|0;
    $47 = $46 ^ $44;
    HEAP8[$15>>0] = $47;
    $48 = HEAP8[$16>>0]|0;
    $$sum4$4 = (($31) + 4)|0;
    $49 = (($encodedData) + ($$sum4$4)|0);
    $50 = HEAP8[$49>>0]|0;
    $51 = $50 ^ $48;
    HEAP8[$16>>0] = $51;
    $52 = HEAP8[$17>>0]|0;
    $$sum4$5 = (($31) + 5)|0;
    $53 = (($encodedData) + ($$sum4$5)|0);
    $54 = HEAP8[$53>>0]|0;
    $55 = $54 ^ $52;
    HEAP8[$17>>0] = $55;
    $56 = HEAP8[$18>>0]|0;
    $$sum4$6 = (($31) + 6)|0;
    $57 = (($encodedData) + ($$sum4$6)|0);
    $58 = HEAP8[$57>>0]|0;
    $59 = $58 ^ $56;
    HEAP8[$18>>0] = $59;
    $60 = HEAP8[$19>>0]|0;
    $$sum4$7 = (($31) + 7)|0;
    $61 = (($encodedData) + ($$sum4$7)|0);
    $62 = HEAP8[$61>>0]|0;
    $63 = $62 ^ $60;
    HEAP8[$19>>0] = $63;
    $64 = HEAP8[$20>>0]|0;
    $$sum4$8 = (($31) + 8)|0;
    $65 = (($encodedData) + ($$sum4$8)|0);
    $66 = HEAP8[$65>>0]|0;
    $67 = $66 ^ $64;
    HEAP8[$20>>0] = $67;
    $68 = HEAP8[$21>>0]|0;
    $$sum4$9 = (($31) + 9)|0;
    $69 = (($encodedData) + ($$sum4$9)|0);
    $70 = HEAP8[$69>>0]|0;
    $71 = $70 ^ $68;
    HEAP8[$21>>0] = $71;
    $72 = HEAP8[$22>>0]|0;
    $$sum4$10 = (($31) + 10)|0;
    $73 = (($encodedData) + ($$sum4$10)|0);
    $74 = HEAP8[$73>>0]|0;
    $75 = $74 ^ $72;
    HEAP8[$22>>0] = $75;
    $76 = HEAP8[$23>>0]|0;
    $$sum4$11 = (($31) + 11)|0;
    $77 = (($encodedData) + ($$sum4$11)|0);
    $78 = HEAP8[$77>>0]|0;
    $79 = $78 ^ $76;
    HEAP8[$23>>0] = $79;
    $80 = HEAP8[$24>>0]|0;
    $$sum4$12 = (($31) + 12)|0;
    $81 = (($encodedData) + ($$sum4$12)|0);
    $82 = HEAP8[$81>>0]|0;
    $83 = $82 ^ $80;
    HEAP8[$24>>0] = $83;
    $84 = HEAP8[$25>>0]|0;
    $$sum4$13 = (($31) + 13)|0;
    $85 = (($encodedData) + ($$sum4$13)|0);
    $86 = HEAP8[$85>>0]|0;
    $87 = $86 ^ $84;
    HEAP8[$25>>0] = $87;
    $88 = HEAP8[$26>>0]|0;
    $$sum4$14 = (($31) + 14)|0;
    $89 = (($encodedData) + ($$sum4$14)|0);
    $90 = HEAP8[$89>>0]|0;
    $91 = $90 ^ $88;
    HEAP8[$26>>0] = $91;
    $92 = HEAP8[$27>>0]|0;
    $$sum4$15 = (($31) + 15)|0;
    $93 = (($encodedData) + ($$sum4$15)|0);
    $94 = HEAP8[$93>>0]|0;
    $95 = $94 ^ $92;
    HEAP8[$27>>0] = $95;
   }
   $$sum56 = $28 | 8;
   $96 = (($5) + ($$sum56)|0);
   dest=$96; src=$buf; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
   $97 = (($i$07) + 1)|0;
   $exitcond = ($97|0)==($3|0);
   if ($exitcond) {
    break;
   } else {
    $i$07 = $97;
   }
  }
 }
 _egaes256_done($ctx);
 HEAP32[$plainData>>2] = $6;
 $98 = HEAP32[$plainDataSize>>2]|0;
 $$sum = (($98) + 7)|0;
 $99 = (($5) + ($$sum)|0);
 $100 = HEAP8[$99>>0]|0;
 $101 = $100&255;
 $102 = ($100&255)>(16);
 if ($102) {
  $$0$i = 0;
  $109 = (($98) - ($$0$i))|0;
  HEAP32[$plainDataSize>>2] = $109;
  STACKTOP = sp;return 0;
 }
 $103 = ($100<<24>>24)==(0);
 if ($103) {
  $$0$i = $101;
  $109 = (($98) - ($$0$i))|0;
  HEAP32[$plainDataSize>>2] = $109;
  STACKTOP = sp;return 0;
 } else {
  $i$01$i1 = 0;
 }
 while(1) {
  $$sum3 = (($$sum) - ($i$01$i1))|0;
  $106 = (($5) + ($$sum3)|0);
  $107 = HEAP8[$106>>0]|0;
  $108 = ($107<<24>>24)==($100<<24>>24);
  $105 = (($i$01$i1) + 1)|0;
  if (!($108)) {
   $$0$i = 0;
   label = 12;
   break;
  }
  $104 = ($105|0)<($101|0);
  if ($104) {
   $i$01$i1 = $105;
  } else {
   $$0$i = $101;
   label = 12;
   break;
  }
 }
 if ((label|0) == 12) {
  $109 = (($98) - ($$0$i))|0;
  HEAP32[$plainDataSize>>2] = $109;
  STACKTOP = sp;return 0;
 }
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter10countEventEi($this,$eventBytes) {
 $this = $this|0;
 $eventBytes = $eventBytes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($eventBytes))|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 HEAP32[$3>>2] = $5;
 return;
}
function __ZNK9ExitGames6Photon9EventData7getCodeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 return ($0|0);
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter20timeForEventCallbackEhi($this,$code,$time) {
 $this = $this|0;
 $code = $code|0;
 $time = $time|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<($time|0);
 if (!($2)) {
  return;
 }
 HEAP32[$0>>2] = $time;
 $3 = ((($this)) + 48|0);
 HEAP8[$3>>0] = $code;
 return;
}
function __ZN9ExitGames6Photon9EventDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($0);
 return;
}
function __ZN9ExitGames6Photon9EventDataC2EhRKNS_6Common10DictionaryIhNS2_6ObjectEEE($this,$code,$params) {
 $this = $this|0;
 $code = $code|0;
 $params = $params|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$this>>0] = $code;
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common14DictionaryBaseC2ERKS1_($0,$params);
 HEAP32[$0>>2] = (2720);
 return;
}
function _egaes256_init($ctx,$k) {
 $ctx = $ctx|0;
 $k = $k|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, $rcon = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $rcon = sp;
 HEAP8[$rcon>>0] = 1;
 $indvars$iv = 0;
 while(1) {
  $1 = (($k) + ($indvars$iv)|0);
  $2 = HEAP8[$1>>0]|0;
  $3 = (((($ctx)) + 64|0) + ($indvars$iv)|0);
  HEAP8[$3>>0] = $2;
  $4 = (((($ctx)) + 32|0) + ($indvars$iv)|0);
  HEAP8[$4>>0] = $2;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $exitcond = ($indvars$iv$next|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $indvars$iv = $indvars$iv$next;
  }
 }
 $0 = ((($ctx)) + 64|0);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 _aes_expandEncKey($0,$rcon);
 STACKTOP = sp;return;
}
function _egaes256_decrypt_ecb($ctx,$buf) {
 $ctx = $ctx|0;
 $buf = $buf|0;
 var $$02$i$i$i$i = 0, $$02$i$i$i$i25 = 0, $$02$i$i$i$i9 = 0, $$02$i$i$i109$i = 0, $$02$i$i$i24$i = 0, $$02$i$i$i41$i = 0, $$02$i$i$i58$i = 0, $$02$i$i$i7$i = 0, $$02$i$i$i75$i = 0, $$02$i$i$i92$i = 0, $$lcssa = 0, $$lcssa82 = 0, $$lcssa83 = 0, $$lcssa84 = 0, $$lcssa85 = 0, $$lcssa86 = 0, $$lcssa87 = 0, $$lcssa88 = 0, $$lcssa89 = 0, $$lcssa90 = 0;
 var $$masked$i$i = 0, $$masked$i$i14 = 0, $$masked$i$i17 = 0, $$masked$i115$i = 0, $$masked$i13$i = 0, $$masked$i30$i = 0, $$masked$i47$i = 0, $$masked$i64$i = 0, $$masked$i81$i = 0, $$masked$i98$i = 0, $$masked1$i$i = 0, $$masked1$i$i15 = 0, $$masked1$i$i18 = 0, $$masked1$i116$i = 0, $$masked1$i14$i = 0, $$masked1$i31$i = 0, $$masked1$i48$i = 0, $$masked1$i65$i = 0, $$masked1$i82$i = 0, $$masked1$i99$i = 0;
 var $$masked2$i$i = 0, $$masked2$i100$i = 0, $$masked2$i117$i = 0, $$masked2$i15$i = 0, $$masked2$i32$i = 0, $$masked2$i49$i = 0, $$masked2$i66$i = 0, $$masked2$i83$i = 0, $$masked4$i$i = 0, $$masked4$i101$i = 0, $$masked4$i118$i = 0, $$masked4$i16$i = 0, $$masked4$i33$i = 0, $$masked4$i50$i = 0, $$masked4$i67$i = 0, $$masked4$i84$i = 0, $$phi$trans$insert = 0, $$phi$trans$insert51 = 0, $$phi$trans$insert53 = 0, $$pre = 0;
 var $$pre52 = 0, $$pre54 = 0, $$sink = 0, $$sink13 = 0, $$sink17 = 0, $$sink21 = 0, $$sink25 = 0, $$sink29 = 0, $$sink33 = 0, $$sink37 = 0, $$sink41 = 0, $$sink45 = 0, $$sink49 = 0, $$sink5 = 0, $$sink53 = 0, $$sink57 = 0, $$sink60 = 0, $$sink9 = 0, $0 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0;
 var $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0;
 var $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0;
 var $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0;
 var $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0;
 var $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0;
 var $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0;
 var $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0;
 var $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0;
 var $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0;
 var $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0;
 var $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0;
 var $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0;
 var $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0;
 var $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0;
 var $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0;
 var $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0;
 var $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0;
 var $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0;
 var $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0;
 var $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0;
 var $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0;
 var $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0;
 var $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0;
 var $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0;
 var $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0;
 var $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0;
 var $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0;
 var $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $atb$0$i$i$i$i = 0, $atb$0$i$i$i$i20 = 0, $atb$0$i$i$i$i4 = 0, $atb$0$i$i$i104$i = 0, $atb$0$i$i$i19$i = 0, $atb$0$i$i$i2$i = 0, $atb$0$i$i$i36$i = 0, $atb$0$i$i$i53$i = 0, $atb$0$i$i$i70$i = 0, $atb$0$i$i$i87$i = 0;
 var $atb$01$i$i$i$i = 0, $atb$01$i$i$i$i10 = 0, $atb$01$i$i$i$i26 = 0, $atb$01$i$i$i110$i = 0, $atb$01$i$i$i25$i = 0, $atb$01$i$i$i42$i = 0, $atb$01$i$i$i59$i = 0, $atb$01$i$i$i76$i = 0, $atb$01$i$i$i8$i = 0, $atb$01$i$i$i93$i = 0, $atb$1$i$i$i$i = 0, $atb$1$i$i$i$i22 = 0, $atb$1$i$i$i$i6 = 0, $atb$1$i$i$i106$i = 0, $atb$1$i$i$i21$i = 0, $atb$1$i$i$i38$i = 0, $atb$1$i$i$i4$i = 0, $atb$1$i$i$i55$i = 0, $atb$1$i$i$i72$i = 0, $atb$1$i$i$i89$i = 0;
 var $atb$1$i1$i$i$i = 0, $atb$1$i1$i$i$i11 = 0, $atb$1$i1$i$i$i27 = 0, $atb$1$i1$i$i111$i = 0, $atb$1$i1$i$i26$i = 0, $atb$1$i1$i$i43$i = 0, $atb$1$i1$i$i60$i = 0, $atb$1$i1$i$i77$i = 0, $atb$1$i1$i$i9$i = 0, $atb$1$i1$i$i94$i = 0, $i$0$i$i$i$i = 0, $i$0$i$i$i$i19 = 0, $i$0$i$i$i$i3 = 0, $i$0$i$i$i1$i = 0, $i$0$i$i$i103$i = 0, $i$0$i$i$i18$i = 0, $i$0$i$i$i35$i = 0, $i$0$i$i$i52$i = 0, $i$0$i$i$i69$i = 0, $i$0$i$i$i86$i = 0;
 var $i$1$i$i$i$i = 0, $i$1$i$i$i$i23 = 0, $i$1$i$i$i$i7 = 0, $i$1$i$i$i107$i = 0, $i$1$i$i$i22$i = 0, $i$1$i$i$i39$i = 0, $i$1$i$i$i5$i = 0, $i$1$i$i$i56$i = 0, $i$1$i$i$i73$i = 0, $i$1$i$i$i90$i = 0, $indvars$iv$i = 0, $indvars$iv$i1 = 0, $indvars$iv$i16 = 0, $indvars$iv$next$i = 0, $indvars$iv$next$i2 = 0, $indvars$iv$next$i30 = 0, $phitmp$i$i = 0, $phitmp$i11$i = 0, $phitmp$i113$i = 0, $phitmp$i28$i = 0;
 var $phitmp$i45$i = 0, $phitmp$i62$i = 0, $phitmp$i79$i = 0, $phitmp$i96$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $indvars$iv$i = 15;
 while(1) {
  $0 = (((($ctx)) + 64|0) + ($indvars$iv$i)|0);
  $1 = HEAP8[$0>>0]|0;
  $2 = (($ctx) + ($indvars$iv$i)|0);
  HEAP8[$2>>0] = $1;
  $3 = (($buf) + ($indvars$iv$i)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4 ^ $1;
  HEAP8[$3>>0] = $5;
  $6 = (($indvars$iv$i) + 16)|0;
  $7 = (((($ctx)) + 64|0) + ($6)|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = (($ctx) + ($6)|0);
  HEAP8[$9>>0] = $8;
  $10 = ($indvars$iv$i|0)==(0);
  $indvars$iv$next$i = (($indvars$iv$i) + -1)|0;
  if ($10) {
   break;
  } else {
   $indvars$iv$i = $indvars$iv$next$i;
  }
 }
 $11 = ((($buf)) + 1|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ((($buf)) + 13|0);
 $14 = HEAP8[$13>>0]|0;
 HEAP8[$11>>0] = $14;
 $15 = ((($buf)) + 9|0);
 $16 = HEAP8[$15>>0]|0;
 HEAP8[$13>>0] = $16;
 $17 = ((($buf)) + 5|0);
 $18 = HEAP8[$17>>0]|0;
 HEAP8[$15>>0] = $18;
 HEAP8[$17>>0] = $12;
 $19 = ((($buf)) + 2|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ((($buf)) + 10|0);
 $22 = HEAP8[$21>>0]|0;
 HEAP8[$19>>0] = $22;
 HEAP8[$21>>0] = $20;
 $23 = ((($buf)) + 3|0);
 $24 = HEAP8[$23>>0]|0;
 $25 = ((($buf)) + 7|0);
 $26 = HEAP8[$25>>0]|0;
 HEAP8[$23>>0] = $26;
 $27 = ((($buf)) + 11|0);
 $28 = HEAP8[$27>>0]|0;
 HEAP8[$25>>0] = $28;
 $29 = ((($buf)) + 15|0);
 $30 = HEAP8[$29>>0]|0;
 HEAP8[$27>>0] = $30;
 HEAP8[$29>>0] = $24;
 $31 = ((($buf)) + 6|0);
 $32 = HEAP8[$31>>0]|0;
 $33 = ((($buf)) + 14|0);
 $34 = HEAP8[$33>>0]|0;
 HEAP8[$31>>0] = $34;
 HEAP8[$33>>0] = $32;
 $37 = $24;$indvars$iv$i16 = 15;
 while(1) {
  $35 = (($buf) + ($indvars$iv$i16)|0);
  $36 = $37&255;
  $38 = $36 << 1;
  $39 = $36 >>> 7;
  $40 = $38 & 254;
  $$masked$i$i17 = $40 | $39;
  $41 = $$masked$i$i17 ^ 198;
  $42 = $41 << 2;
  $43 = $41 >>> 6;
  $$masked1$i$i18 = $42 & 252;
  $44 = $$masked1$i$i18 | $43;
  $45 = $44 << 3;
  $46 = $44 >>> 5;
  $47 = $45 | $46;
  $48 = $44 ^ $41;
  $49 = $48 ^ $47;
  $50 = $49&255;
  $51 = ($50<<24>>24)==(0);
  if ($51) {
   $76 = 0;
  } else {
   $atb$0$i$i$i$i20 = 1;$i$0$i$i$i$i19 = 0;
   while(1) {
    $52 = $atb$0$i$i$i$i20 & 255;
    $53 = $atb$0$i$i$i$i20&255;
    $54 = ($53<<24>>24)==($50<<24>>24);
    if ($54) {
     $i$1$i$i$i$i23 = $i$0$i$i$i$i19;
     break;
    }
    $55 = $52 << 1;
    $56 = $atb$0$i$i$i$i20 & 128;
    $57 = ($56|0)==(0);
    $58 = $55 ^ 27;
    $atb$1$i$i$i$i22 = $57 ? $55 : $58;
    $59 = $atb$1$i$i$i$i22 & 255;
    $60 = $59 ^ $52;
    $61 = (($i$0$i$i$i$i19) + 1)<<24>>24;
    $62 = ($61<<24>>24)==(0);
    if ($62) {
     $i$1$i$i$i$i23 = 0;
     break;
    } else {
     $atb$0$i$i$i$i20 = $60;$i$0$i$i$i$i19 = $61;
    }
   }
   $63 = $i$1$i$i$i$i23&255;
   $64 = $63 ^ 255;
   $65 = $64&255;
   $66 = ($65<<24>>24)==(0);
   if ($66) {
    $76 = 1;
   } else {
    $$02$i$i$i$i25 = $65;$atb$01$i$i$i$i26 = 1;
    while(1) {
     $67 = (($$02$i$i$i$i25) + -1)<<24>>24;
     $68 = $atb$01$i$i$i$i26 & 255;
     $69 = $68 << 1;
     $70 = $atb$01$i$i$i$i26 & 128;
     $71 = ($70|0)==(0);
     $72 = $69 ^ 27;
     $atb$1$i1$i$i$i27 = $71 ? $69 : $72;
     $73 = $atb$1$i1$i$i$i27 ^ $68;
     $74 = ($67<<24>>24)==(0);
     if ($74) {
      $$lcssa90 = $73;
      break;
     } else {
      $$02$i$i$i$i25 = $67;$atb$01$i$i$i$i26 = $73;
     }
    }
    $75 = $$lcssa90&255;
    $76 = $75;
   }
  }
  HEAP8[$35>>0] = $76;
  $77 = ($indvars$iv$i16|0)==(0);
  $indvars$iv$next$i30 = (($indvars$iv$i16) + -1)|0;
  if ($77) {
   break;
  }
  $$phi$trans$insert = (($buf) + ($indvars$iv$next$i30)|0);
  $$pre = HEAP8[$$phi$trans$insert>>0]|0;
  $37 = $$pre;$indvars$iv$i16 = $indvars$iv$next$i30;
 }
 $78 = ((($ctx)) + 15|0);
 $79 = ((($ctx)) + 14|0);
 $80 = ((($ctx)) + 13|0);
 $81 = ((($ctx)) + 12|0);
 $82 = ((($buf)) + 12|0);
 $83 = ((($ctx)) + 11|0);
 $84 = ((($ctx)) + 10|0);
 $85 = ((($ctx)) + 9|0);
 $86 = ((($ctx)) + 8|0);
 $87 = ((($buf)) + 8|0);
 $88 = ((($ctx)) + 7|0);
 $89 = ((($ctx)) + 6|0);
 $90 = ((($ctx)) + 5|0);
 $91 = ((($ctx)) + 4|0);
 $92 = ((($buf)) + 4|0);
 $93 = ((($ctx)) + 3|0);
 $94 = ((($ctx)) + 2|0);
 $95 = ((($ctx)) + 1|0);
 $96 = ((($ctx)) + 24|0);
 $97 = ((($ctx)) + 28|0);
 $98 = ((($ctx)) + 25|0);
 $99 = ((($ctx)) + 29|0);
 $100 = ((($ctx)) + 26|0);
 $101 = ((($ctx)) + 30|0);
 $102 = ((($ctx)) + 27|0);
 $103 = ((($ctx)) + 31|0);
 $104 = ((($ctx)) + 20|0);
 $105 = ((($ctx)) + 21|0);
 $106 = ((($ctx)) + 22|0);
 $107 = ((($ctx)) + 23|0);
 $108 = ((($ctx)) + 16|0);
 $109 = ((($ctx)) + 17|0);
 $110 = ((($ctx)) + 18|0);
 $111 = ((($ctx)) + 19|0);
 $113 = 13;$346 = -128;
 while(1) {
  $115 = $113 & 1;
  $116 = ($115<<24>>24)==(0);
  if ($116) {
   $526 = HEAP8[$78>>0]|0;
   $$sink = $ctx;$$sink13 = $93;$$sink17 = $91;$$sink21 = $90;$$sink25 = $89;$$sink29 = $88;$$sink33 = $86;$$sink37 = $85;$$sink41 = $84;$$sink45 = $83;$$sink49 = $81;$$sink5 = $95;$$sink53 = $80;$$sink57 = $79;$$sink60 = $526;$$sink9 = $94;$771 = $346;
  } else {
   $117 = HEAP8[$96>>0]|0;
   $118 = HEAP8[$97>>0]|0;
   $119 = $118 ^ $117;
   HEAP8[$97>>0] = $119;
   $120 = HEAP8[$98>>0]|0;
   $121 = HEAP8[$99>>0]|0;
   $122 = $121 ^ $120;
   HEAP8[$99>>0] = $122;
   $123 = HEAP8[$100>>0]|0;
   $124 = HEAP8[$101>>0]|0;
   $125 = $124 ^ $123;
   HEAP8[$101>>0] = $125;
   $126 = HEAP8[$102>>0]|0;
   $127 = HEAP8[$103>>0]|0;
   $128 = $127 ^ $126;
   HEAP8[$103>>0] = $128;
   $129 = HEAP8[$104>>0]|0;
   $130 = $129 ^ $117;
   HEAP8[$96>>0] = $130;
   $131 = HEAP8[$105>>0]|0;
   $132 = $131 ^ $120;
   HEAP8[$98>>0] = $132;
   $133 = HEAP8[$106>>0]|0;
   $134 = $133 ^ $123;
   HEAP8[$100>>0] = $134;
   $135 = HEAP8[$107>>0]|0;
   $136 = $135 ^ $126;
   HEAP8[$102>>0] = $136;
   $137 = HEAP8[$108>>0]|0;
   $138 = $137 ^ $129;
   HEAP8[$104>>0] = $138;
   $139 = HEAP8[$109>>0]|0;
   $140 = $139 ^ $131;
   HEAP8[$105>>0] = $140;
   $141 = HEAP8[$110>>0]|0;
   $142 = $141 ^ $133;
   HEAP8[$106>>0] = $142;
   $143 = HEAP8[$111>>0]|0;
   $144 = $143 ^ $135;
   HEAP8[$107>>0] = $144;
   $145 = HEAP8[$81>>0]|0;
   $146 = ($145<<24>>24)==(0);
   if ($146) {
    $171 = 0;
   } else {
    $atb$0$i$i$i$i4 = 1;$i$0$i$i$i$i3 = 0;
    while(1) {
     $147 = $atb$0$i$i$i$i4 & 255;
     $148 = $atb$0$i$i$i$i4&255;
     $149 = ($148<<24>>24)==($145<<24>>24);
     if ($149) {
      $i$1$i$i$i$i7 = $i$0$i$i$i$i3;
      break;
     }
     $150 = $147 << 1;
     $151 = $atb$0$i$i$i$i4 & 128;
     $152 = ($151|0)==(0);
     $153 = $150 ^ 27;
     $atb$1$i$i$i$i6 = $152 ? $150 : $153;
     $154 = $atb$1$i$i$i$i6 & 255;
     $155 = $154 ^ $147;
     $156 = (($i$0$i$i$i$i3) + 1)<<24>>24;
     $157 = ($156<<24>>24)==(0);
     if ($157) {
      $i$1$i$i$i$i7 = 0;
      break;
     } else {
      $atb$0$i$i$i$i4 = $155;$i$0$i$i$i$i3 = $156;
     }
    }
    $158 = $i$1$i$i$i$i7&255;
    $159 = $158 ^ 255;
    $160 = $159&255;
    $161 = ($160<<24>>24)==(0);
    if ($161) {
     $171 = 1;
    } else {
     $$02$i$i$i$i9 = $160;$atb$01$i$i$i$i10 = 1;
     while(1) {
      $162 = (($$02$i$i$i$i9) + -1)<<24>>24;
      $163 = $atb$01$i$i$i$i10 & 255;
      $164 = $163 << 1;
      $165 = $atb$01$i$i$i$i10 & 128;
      $166 = ($165|0)==(0);
      $167 = $164 ^ 27;
      $atb$1$i1$i$i$i11 = $166 ? $164 : $167;
      $168 = $atb$1$i1$i$i$i11 ^ $163;
      $169 = ($162<<24>>24)==(0);
      if ($169) {
       $$lcssa = $168;
       break;
      } else {
       $$02$i$i$i$i9 = $162;$atb$01$i$i$i$i10 = $168;
      }
     }
     $phitmp$i$i = $$lcssa & 255;
     $171 = $phitmp$i$i;
    }
   }
   $170 = $171 << 1;
   $172 = $171 >>> 7;
   $$masked$i$i14 = $170 & 254;
   $173 = $$masked$i$i14 | $172;
   $174 = $173 << 1;
   $175 = $$masked$i$i14 >>> 7;
   $$masked1$i$i15 = $174 & 254;
   $176 = $$masked1$i$i15 | $175;
   $177 = $176 << 1;
   $178 = $$masked1$i$i15 >>> 7;
   $$masked2$i$i = $177 & 254;
   $179 = $$masked2$i$i | $178;
   $180 = $179 << 1;
   $181 = $$masked2$i$i >>> 7;
   $$masked4$i$i = $180 | $181;
   $182 = $171 ^ 99;
   $183 = $182 ^ $173;
   $184 = $183 ^ $176;
   $185 = $184 ^ $179;
   $186 = $185 ^ $$masked4$i$i;
   $187 = $186&255;
   $188 = $187 ^ $137;
   HEAP8[$108>>0] = $188;
   $189 = HEAP8[$80>>0]|0;
   $190 = ($189<<24>>24)==(0);
   if ($190) {
    $215 = 0;
   } else {
    $atb$0$i$i$i104$i = 1;$i$0$i$i$i103$i = 0;
    while(1) {
     $191 = $atb$0$i$i$i104$i & 255;
     $192 = $atb$0$i$i$i104$i&255;
     $193 = ($192<<24>>24)==($189<<24>>24);
     if ($193) {
      $i$1$i$i$i107$i = $i$0$i$i$i103$i;
      break;
     }
     $194 = $191 << 1;
     $195 = $atb$0$i$i$i104$i & 128;
     $196 = ($195|0)==(0);
     $197 = $194 ^ 27;
     $atb$1$i$i$i106$i = $196 ? $194 : $197;
     $198 = $atb$1$i$i$i106$i & 255;
     $199 = $198 ^ $191;
     $200 = (($i$0$i$i$i103$i) + 1)<<24>>24;
     $201 = ($200<<24>>24)==(0);
     if ($201) {
      $i$1$i$i$i107$i = 0;
      break;
     } else {
      $atb$0$i$i$i104$i = $199;$i$0$i$i$i103$i = $200;
     }
    }
    $202 = $i$1$i$i$i107$i&255;
    $203 = $202 ^ 255;
    $204 = $203&255;
    $205 = ($204<<24>>24)==(0);
    if ($205) {
     $215 = 1;
    } else {
     $$02$i$i$i109$i = $204;$atb$01$i$i$i110$i = 1;
     while(1) {
      $206 = (($$02$i$i$i109$i) + -1)<<24>>24;
      $207 = $atb$01$i$i$i110$i & 255;
      $208 = $207 << 1;
      $209 = $atb$01$i$i$i110$i & 128;
      $210 = ($209|0)==(0);
      $211 = $208 ^ 27;
      $atb$1$i1$i$i111$i = $210 ? $208 : $211;
      $212 = $atb$1$i1$i$i111$i ^ $207;
      $213 = ($206<<24>>24)==(0);
      if ($213) {
       $$lcssa82 = $212;
       break;
      } else {
       $$02$i$i$i109$i = $206;$atb$01$i$i$i110$i = $212;
      }
     }
     $phitmp$i113$i = $$lcssa82 & 255;
     $215 = $phitmp$i113$i;
    }
   }
   $214 = $215 << 1;
   $216 = $215 >>> 7;
   $$masked$i115$i = $214 & 254;
   $217 = $$masked$i115$i | $216;
   $218 = $217 << 1;
   $219 = $$masked$i115$i >>> 7;
   $$masked1$i116$i = $218 & 254;
   $220 = $$masked1$i116$i | $219;
   $221 = $220 << 1;
   $222 = $$masked1$i116$i >>> 7;
   $$masked2$i117$i = $221 & 254;
   $223 = $$masked2$i117$i | $222;
   $224 = $223 << 1;
   $225 = $$masked2$i117$i >>> 7;
   $$masked4$i118$i = $224 | $225;
   $226 = $215 ^ 99;
   $227 = $226 ^ $217;
   $228 = $227 ^ $220;
   $229 = $228 ^ $223;
   $230 = $229 ^ $$masked4$i118$i;
   $231 = $230&255;
   $232 = $231 ^ $139;
   HEAP8[$109>>0] = $232;
   $233 = HEAP8[$79>>0]|0;
   $234 = ($233<<24>>24)==(0);
   if ($234) {
    $259 = 0;
   } else {
    $atb$0$i$i$i87$i = 1;$i$0$i$i$i86$i = 0;
    while(1) {
     $235 = $atb$0$i$i$i87$i & 255;
     $236 = $atb$0$i$i$i87$i&255;
     $237 = ($236<<24>>24)==($233<<24>>24);
     if ($237) {
      $i$1$i$i$i90$i = $i$0$i$i$i86$i;
      break;
     }
     $238 = $235 << 1;
     $239 = $atb$0$i$i$i87$i & 128;
     $240 = ($239|0)==(0);
     $241 = $238 ^ 27;
     $atb$1$i$i$i89$i = $240 ? $238 : $241;
     $242 = $atb$1$i$i$i89$i & 255;
     $243 = $242 ^ $235;
     $244 = (($i$0$i$i$i86$i) + 1)<<24>>24;
     $245 = ($244<<24>>24)==(0);
     if ($245) {
      $i$1$i$i$i90$i = 0;
      break;
     } else {
      $atb$0$i$i$i87$i = $243;$i$0$i$i$i86$i = $244;
     }
    }
    $246 = $i$1$i$i$i90$i&255;
    $247 = $246 ^ 255;
    $248 = $247&255;
    $249 = ($248<<24>>24)==(0);
    if ($249) {
     $259 = 1;
    } else {
     $$02$i$i$i92$i = $248;$atb$01$i$i$i93$i = 1;
     while(1) {
      $250 = (($$02$i$i$i92$i) + -1)<<24>>24;
      $251 = $atb$01$i$i$i93$i & 255;
      $252 = $251 << 1;
      $253 = $atb$01$i$i$i93$i & 128;
      $254 = ($253|0)==(0);
      $255 = $252 ^ 27;
      $atb$1$i1$i$i94$i = $254 ? $252 : $255;
      $256 = $atb$1$i1$i$i94$i ^ $251;
      $257 = ($250<<24>>24)==(0);
      if ($257) {
       $$lcssa83 = $256;
       break;
      } else {
       $$02$i$i$i92$i = $250;$atb$01$i$i$i93$i = $256;
      }
     }
     $phitmp$i96$i = $$lcssa83 & 255;
     $259 = $phitmp$i96$i;
    }
   }
   $258 = $259 << 1;
   $260 = $259 >>> 7;
   $$masked$i98$i = $258 & 254;
   $261 = $$masked$i98$i | $260;
   $262 = $261 << 1;
   $263 = $$masked$i98$i >>> 7;
   $$masked1$i99$i = $262 & 254;
   $264 = $$masked1$i99$i | $263;
   $265 = $264 << 1;
   $266 = $$masked1$i99$i >>> 7;
   $$masked2$i100$i = $265 & 254;
   $267 = $$masked2$i100$i | $266;
   $268 = $267 << 1;
   $269 = $$masked2$i100$i >>> 7;
   $$masked4$i101$i = $268 | $269;
   $270 = $259 ^ 99;
   $271 = $270 ^ $261;
   $272 = $271 ^ $264;
   $273 = $272 ^ $267;
   $274 = $273 ^ $$masked4$i101$i;
   $275 = $274&255;
   $276 = $275 ^ $141;
   HEAP8[$110>>0] = $276;
   $277 = HEAP8[$78>>0]|0;
   $278 = ($277<<24>>24)==(0);
   if ($278) {
    $303 = 0;
   } else {
    $atb$0$i$i$i70$i = 1;$i$0$i$i$i69$i = 0;
    while(1) {
     $279 = $atb$0$i$i$i70$i & 255;
     $280 = $atb$0$i$i$i70$i&255;
     $281 = ($280<<24>>24)==($277<<24>>24);
     if ($281) {
      $i$1$i$i$i73$i = $i$0$i$i$i69$i;
      break;
     }
     $282 = $279 << 1;
     $283 = $atb$0$i$i$i70$i & 128;
     $284 = ($283|0)==(0);
     $285 = $282 ^ 27;
     $atb$1$i$i$i72$i = $284 ? $282 : $285;
     $286 = $atb$1$i$i$i72$i & 255;
     $287 = $286 ^ $279;
     $288 = (($i$0$i$i$i69$i) + 1)<<24>>24;
     $289 = ($288<<24>>24)==(0);
     if ($289) {
      $i$1$i$i$i73$i = 0;
      break;
     } else {
      $atb$0$i$i$i70$i = $287;$i$0$i$i$i69$i = $288;
     }
    }
    $290 = $i$1$i$i$i73$i&255;
    $291 = $290 ^ 255;
    $292 = $291&255;
    $293 = ($292<<24>>24)==(0);
    if ($293) {
     $303 = 1;
    } else {
     $$02$i$i$i75$i = $292;$atb$01$i$i$i76$i = 1;
     while(1) {
      $294 = (($$02$i$i$i75$i) + -1)<<24>>24;
      $295 = $atb$01$i$i$i76$i & 255;
      $296 = $295 << 1;
      $297 = $atb$01$i$i$i76$i & 128;
      $298 = ($297|0)==(0);
      $299 = $296 ^ 27;
      $atb$1$i1$i$i77$i = $298 ? $296 : $299;
      $300 = $atb$1$i1$i$i77$i ^ $295;
      $301 = ($294<<24>>24)==(0);
      if ($301) {
       $$lcssa84 = $300;
       break;
      } else {
       $$02$i$i$i75$i = $294;$atb$01$i$i$i76$i = $300;
      }
     }
     $phitmp$i79$i = $$lcssa84 & 255;
     $303 = $phitmp$i79$i;
    }
   }
   $302 = $303 << 1;
   $304 = $303 >>> 7;
   $$masked$i81$i = $302 & 254;
   $305 = $$masked$i81$i | $304;
   $306 = $305 << 1;
   $307 = $$masked$i81$i >>> 7;
   $$masked1$i82$i = $306 & 254;
   $308 = $$masked1$i82$i | $307;
   $309 = $308 << 1;
   $310 = $$masked1$i82$i >>> 7;
   $$masked2$i83$i = $309 & 254;
   $311 = $$masked2$i83$i | $310;
   $312 = $311 << 1;
   $313 = $$masked2$i83$i >>> 7;
   $$masked4$i84$i = $312 | $313;
   $314 = $303 ^ 99;
   $315 = $314 ^ $305;
   $316 = $315 ^ $308;
   $317 = $316 ^ $311;
   $318 = $317 ^ $$masked4$i84$i;
   $319 = $318&255;
   $320 = $319 ^ $143;
   HEAP8[$111>>0] = $320;
   $321 = HEAP8[$86>>0]|0;
   $322 = $321 ^ $145;
   HEAP8[$81>>0] = $322;
   $323 = HEAP8[$85>>0]|0;
   $324 = $323 ^ $189;
   HEAP8[$80>>0] = $324;
   $325 = HEAP8[$84>>0]|0;
   $326 = $325 ^ $233;
   HEAP8[$79>>0] = $326;
   $327 = HEAP8[$83>>0]|0;
   $328 = $327 ^ $277;
   HEAP8[$78>>0] = $328;
   $329 = HEAP8[$91>>0]|0;
   $330 = $329 ^ $321;
   HEAP8[$86>>0] = $330;
   $331 = HEAP8[$90>>0]|0;
   $332 = $331 ^ $323;
   HEAP8[$85>>0] = $332;
   $333 = HEAP8[$89>>0]|0;
   $334 = $333 ^ $325;
   HEAP8[$84>>0] = $334;
   $335 = HEAP8[$88>>0]|0;
   $336 = $335 ^ $327;
   HEAP8[$83>>0] = $336;
   $337 = HEAP8[$ctx>>0]|0;
   $338 = $337 ^ $329;
   HEAP8[$91>>0] = $338;
   $339 = HEAP8[$95>>0]|0;
   $340 = $339 ^ $331;
   HEAP8[$90>>0] = $340;
   $341 = HEAP8[$94>>0]|0;
   $342 = $341 ^ $333;
   HEAP8[$89>>0] = $342;
   $343 = HEAP8[$93>>0]|0;
   $344 = $343 ^ $335;
   HEAP8[$88>>0] = $344;
   $345 = $346&255;
   $347 = $345 >>> 1;
   $348 = $345 & 1;
   $349 = ($348|0)!=(0);
   $350 = $349 ? 141 : 0;
   $351 = $350 ^ $347;
   $352 = $351&255;
   $353 = ($121<<24>>24)==($120<<24>>24);
   if ($353) {
    $378 = 0;
   } else {
    $atb$0$i$i$i53$i = 1;$i$0$i$i$i52$i = 0;
    while(1) {
     $354 = $atb$0$i$i$i53$i & 255;
     $355 = $atb$0$i$i$i53$i&255;
     $356 = ($355<<24>>24)==($122<<24>>24);
     if ($356) {
      $i$1$i$i$i56$i = $i$0$i$i$i52$i;
      break;
     }
     $357 = $354 << 1;
     $358 = $atb$0$i$i$i53$i & 128;
     $359 = ($358|0)==(0);
     $360 = $357 ^ 27;
     $atb$1$i$i$i55$i = $359 ? $357 : $360;
     $361 = $atb$1$i$i$i55$i & 255;
     $362 = $361 ^ $354;
     $363 = (($i$0$i$i$i52$i) + 1)<<24>>24;
     $364 = ($363<<24>>24)==(0);
     if ($364) {
      $i$1$i$i$i56$i = 0;
      break;
     } else {
      $atb$0$i$i$i53$i = $362;$i$0$i$i$i52$i = $363;
     }
    }
    $365 = $i$1$i$i$i56$i&255;
    $366 = $365 ^ 255;
    $367 = $366&255;
    $368 = ($367<<24>>24)==(0);
    if ($368) {
     $378 = 1;
    } else {
     $$02$i$i$i58$i = $367;$atb$01$i$i$i59$i = 1;
     while(1) {
      $369 = (($$02$i$i$i58$i) + -1)<<24>>24;
      $370 = $atb$01$i$i$i59$i & 255;
      $371 = $370 << 1;
      $372 = $atb$01$i$i$i59$i & 128;
      $373 = ($372|0)==(0);
      $374 = $371 ^ 27;
      $atb$1$i1$i$i60$i = $373 ? $371 : $374;
      $375 = $atb$1$i1$i$i60$i ^ $370;
      $376 = ($369<<24>>24)==(0);
      if ($376) {
       $$lcssa85 = $375;
       break;
      } else {
       $$02$i$i$i58$i = $369;$atb$01$i$i$i59$i = $375;
      }
     }
     $phitmp$i62$i = $$lcssa85 & 255;
     $378 = $phitmp$i62$i;
    }
   }
   $377 = $378 << 1;
   $379 = $378 >>> 7;
   $$masked$i64$i = $377 & 254;
   $380 = $$masked$i64$i | $379;
   $381 = $380 << 1;
   $382 = $$masked$i64$i >>> 7;
   $$masked1$i65$i = $381 & 254;
   $383 = $$masked1$i65$i | $382;
   $384 = $383 << 1;
   $385 = $$masked1$i65$i >>> 7;
   $$masked2$i66$i = $384 & 254;
   $386 = $$masked2$i66$i | $385;
   $387 = $386 << 1;
   $388 = $$masked2$i66$i >>> 7;
   $$masked4$i67$i = $387 | $388;
   $389 = $378 ^ 99;
   $390 = $389 ^ $380;
   $391 = $390 ^ $383;
   $392 = $391 ^ $386;
   $393 = $392 ^ $$masked4$i67$i;
   $394 = $393&255;
   $395 = $337 ^ $352;
   $396 = $395 ^ $394;
   HEAP8[$ctx>>0] = $396;
   $397 = ($124<<24>>24)==($123<<24>>24);
   if ($397) {
    $422 = 0;
   } else {
    $atb$0$i$i$i36$i = 1;$i$0$i$i$i35$i = 0;
    while(1) {
     $398 = $atb$0$i$i$i36$i & 255;
     $399 = $atb$0$i$i$i36$i&255;
     $400 = ($399<<24>>24)==($125<<24>>24);
     if ($400) {
      $i$1$i$i$i39$i = $i$0$i$i$i35$i;
      break;
     }
     $401 = $398 << 1;
     $402 = $atb$0$i$i$i36$i & 128;
     $403 = ($402|0)==(0);
     $404 = $401 ^ 27;
     $atb$1$i$i$i38$i = $403 ? $401 : $404;
     $405 = $atb$1$i$i$i38$i & 255;
     $406 = $405 ^ $398;
     $407 = (($i$0$i$i$i35$i) + 1)<<24>>24;
     $408 = ($407<<24>>24)==(0);
     if ($408) {
      $i$1$i$i$i39$i = 0;
      break;
     } else {
      $atb$0$i$i$i36$i = $406;$i$0$i$i$i35$i = $407;
     }
    }
    $409 = $i$1$i$i$i39$i&255;
    $410 = $409 ^ 255;
    $411 = $410&255;
    $412 = ($411<<24>>24)==(0);
    if ($412) {
     $422 = 1;
    } else {
     $$02$i$i$i41$i = $411;$atb$01$i$i$i42$i = 1;
     while(1) {
      $413 = (($$02$i$i$i41$i) + -1)<<24>>24;
      $414 = $atb$01$i$i$i42$i & 255;
      $415 = $414 << 1;
      $416 = $atb$01$i$i$i42$i & 128;
      $417 = ($416|0)==(0);
      $418 = $415 ^ 27;
      $atb$1$i1$i$i43$i = $417 ? $415 : $418;
      $419 = $atb$1$i1$i$i43$i ^ $414;
      $420 = ($413<<24>>24)==(0);
      if ($420) {
       $$lcssa86 = $419;
       break;
      } else {
       $$02$i$i$i41$i = $413;$atb$01$i$i$i42$i = $419;
      }
     }
     $phitmp$i45$i = $$lcssa86 & 255;
     $422 = $phitmp$i45$i;
    }
   }
   $421 = $422 << 1;
   $423 = $422 >>> 7;
   $$masked$i47$i = $421 & 254;
   $424 = $$masked$i47$i | $423;
   $425 = $424 << 1;
   $426 = $$masked$i47$i >>> 7;
   $$masked1$i48$i = $425 & 254;
   $427 = $$masked1$i48$i | $426;
   $428 = $427 << 1;
   $429 = $$masked1$i48$i >>> 7;
   $$masked2$i49$i = $428 & 254;
   $430 = $$masked2$i49$i | $429;
   $431 = $430 << 1;
   $432 = $$masked2$i49$i >>> 7;
   $$masked4$i50$i = $431 | $432;
   $433 = $422 ^ 99;
   $434 = $433 ^ $424;
   $435 = $434 ^ $427;
   $436 = $435 ^ $430;
   $437 = $436 ^ $$masked4$i50$i;
   $438 = $437&255;
   $439 = $438 ^ $339;
   HEAP8[$95>>0] = $439;
   $440 = ($127<<24>>24)==($126<<24>>24);
   if ($440) {
    $465 = 0;
   } else {
    $atb$0$i$i$i19$i = 1;$i$0$i$i$i18$i = 0;
    while(1) {
     $441 = $atb$0$i$i$i19$i & 255;
     $442 = $atb$0$i$i$i19$i&255;
     $443 = ($442<<24>>24)==($128<<24>>24);
     if ($443) {
      $i$1$i$i$i22$i = $i$0$i$i$i18$i;
      break;
     }
     $444 = $441 << 1;
     $445 = $atb$0$i$i$i19$i & 128;
     $446 = ($445|0)==(0);
     $447 = $444 ^ 27;
     $atb$1$i$i$i21$i = $446 ? $444 : $447;
     $448 = $atb$1$i$i$i21$i & 255;
     $449 = $448 ^ $441;
     $450 = (($i$0$i$i$i18$i) + 1)<<24>>24;
     $451 = ($450<<24>>24)==(0);
     if ($451) {
      $i$1$i$i$i22$i = 0;
      break;
     } else {
      $atb$0$i$i$i19$i = $449;$i$0$i$i$i18$i = $450;
     }
    }
    $452 = $i$1$i$i$i22$i&255;
    $453 = $452 ^ 255;
    $454 = $453&255;
    $455 = ($454<<24>>24)==(0);
    if ($455) {
     $465 = 1;
    } else {
     $$02$i$i$i24$i = $454;$atb$01$i$i$i25$i = 1;
     while(1) {
      $456 = (($$02$i$i$i24$i) + -1)<<24>>24;
      $457 = $atb$01$i$i$i25$i & 255;
      $458 = $457 << 1;
      $459 = $atb$01$i$i$i25$i & 128;
      $460 = ($459|0)==(0);
      $461 = $458 ^ 27;
      $atb$1$i1$i$i26$i = $460 ? $458 : $461;
      $462 = $atb$1$i1$i$i26$i ^ $457;
      $463 = ($456<<24>>24)==(0);
      if ($463) {
       $$lcssa87 = $462;
       break;
      } else {
       $$02$i$i$i24$i = $456;$atb$01$i$i$i25$i = $462;
      }
     }
     $phitmp$i28$i = $$lcssa87 & 255;
     $465 = $phitmp$i28$i;
    }
   }
   $464 = $465 << 1;
   $466 = $465 >>> 7;
   $$masked$i30$i = $464 & 254;
   $467 = $$masked$i30$i | $466;
   $468 = $467 << 1;
   $469 = $$masked$i30$i >>> 7;
   $$masked1$i31$i = $468 & 254;
   $470 = $$masked1$i31$i | $469;
   $471 = $470 << 1;
   $472 = $$masked1$i31$i >>> 7;
   $$masked2$i32$i = $471 & 254;
   $473 = $$masked2$i32$i | $472;
   $474 = $473 << 1;
   $475 = $$masked2$i32$i >>> 7;
   $$masked4$i33$i = $474 | $475;
   $476 = $465 ^ 99;
   $477 = $476 ^ $467;
   $478 = $477 ^ $470;
   $479 = $478 ^ $473;
   $480 = $479 ^ $$masked4$i33$i;
   $481 = $480&255;
   $482 = $481 ^ $341;
   HEAP8[$94>>0] = $482;
   $483 = ($118<<24>>24)==($117<<24>>24);
   if ($483) {
    $508 = 0;
   } else {
    $atb$0$i$i$i2$i = 1;$i$0$i$i$i1$i = 0;
    while(1) {
     $484 = $atb$0$i$i$i2$i & 255;
     $485 = $atb$0$i$i$i2$i&255;
     $486 = ($485<<24>>24)==($119<<24>>24);
     if ($486) {
      $i$1$i$i$i5$i = $i$0$i$i$i1$i;
      break;
     }
     $487 = $484 << 1;
     $488 = $atb$0$i$i$i2$i & 128;
     $489 = ($488|0)==(0);
     $490 = $487 ^ 27;
     $atb$1$i$i$i4$i = $489 ? $487 : $490;
     $491 = $atb$1$i$i$i4$i & 255;
     $492 = $491 ^ $484;
     $493 = (($i$0$i$i$i1$i) + 1)<<24>>24;
     $494 = ($493<<24>>24)==(0);
     if ($494) {
      $i$1$i$i$i5$i = 0;
      break;
     } else {
      $atb$0$i$i$i2$i = $492;$i$0$i$i$i1$i = $493;
     }
    }
    $495 = $i$1$i$i$i5$i&255;
    $496 = $495 ^ 255;
    $497 = $496&255;
    $498 = ($497<<24>>24)==(0);
    if ($498) {
     $508 = 1;
    } else {
     $$02$i$i$i7$i = $497;$atb$01$i$i$i8$i = 1;
     while(1) {
      $499 = (($$02$i$i$i7$i) + -1)<<24>>24;
      $500 = $atb$01$i$i$i8$i & 255;
      $501 = $500 << 1;
      $502 = $atb$01$i$i$i8$i & 128;
      $503 = ($502|0)==(0);
      $504 = $501 ^ 27;
      $atb$1$i1$i$i9$i = $503 ? $501 : $504;
      $505 = $atb$1$i1$i$i9$i ^ $500;
      $506 = ($499<<24>>24)==(0);
      if ($506) {
       $$lcssa88 = $505;
       break;
      } else {
       $$02$i$i$i7$i = $499;$atb$01$i$i$i8$i = $505;
      }
     }
     $phitmp$i11$i = $$lcssa88 & 255;
     $508 = $phitmp$i11$i;
    }
   }
   $507 = $508 << 1;
   $509 = $508 >>> 7;
   $$masked$i13$i = $507 & 254;
   $510 = $$masked$i13$i | $509;
   $511 = $510 << 1;
   $512 = $$masked$i13$i >>> 7;
   $$masked1$i14$i = $511 & 254;
   $513 = $$masked1$i14$i | $512;
   $514 = $513 << 1;
   $515 = $$masked1$i14$i >>> 7;
   $$masked2$i15$i = $514 & 254;
   $516 = $$masked2$i15$i | $515;
   $517 = $516 << 1;
   $518 = $$masked2$i15$i >>> 7;
   $$masked4$i16$i = $517 | $518;
   $519 = $508 ^ 99;
   $520 = $519 ^ $510;
   $521 = $520 ^ $513;
   $522 = $521 ^ $516;
   $523 = $522 ^ $$masked4$i16$i;
   $524 = $523&255;
   $525 = $524 ^ $343;
   HEAP8[$93>>0] = $525;
   $$sink = $108;$$sink13 = $111;$$sink17 = $104;$$sink21 = $105;$$sink25 = $106;$$sink29 = $107;$$sink33 = $96;$$sink37 = $98;$$sink41 = $100;$$sink45 = $102;$$sink49 = $97;$$sink5 = $109;$$sink53 = $99;$$sink57 = $101;$$sink60 = $128;$$sink9 = $110;$771 = $352;
  }
  $527 = HEAP8[$29>>0]|0;
  $528 = $527 ^ $$sink60;
  HEAP8[$29>>0] = $528;
  $529 = HEAP8[$$sink57>>0]|0;
  $530 = HEAP8[$33>>0]|0;
  $531 = $530 ^ $529;
  HEAP8[$33>>0] = $531;
  $532 = HEAP8[$$sink53>>0]|0;
  $533 = HEAP8[$13>>0]|0;
  $534 = $533 ^ $532;
  HEAP8[$13>>0] = $534;
  $535 = HEAP8[$$sink49>>0]|0;
  $536 = HEAP8[$82>>0]|0;
  $537 = $536 ^ $535;
  HEAP8[$82>>0] = $537;
  $538 = HEAP8[$$sink45>>0]|0;
  $539 = HEAP8[$27>>0]|0;
  $540 = $539 ^ $538;
  HEAP8[$27>>0] = $540;
  $541 = HEAP8[$$sink41>>0]|0;
  $542 = HEAP8[$21>>0]|0;
  $543 = $542 ^ $541;
  HEAP8[$21>>0] = $543;
  $544 = HEAP8[$$sink37>>0]|0;
  $545 = HEAP8[$15>>0]|0;
  $546 = $545 ^ $544;
  HEAP8[$15>>0] = $546;
  $547 = HEAP8[$$sink33>>0]|0;
  $548 = HEAP8[$87>>0]|0;
  $549 = $548 ^ $547;
  HEAP8[$87>>0] = $549;
  $550 = HEAP8[$$sink29>>0]|0;
  $551 = HEAP8[$25>>0]|0;
  $552 = $551 ^ $550;
  HEAP8[$25>>0] = $552;
  $553 = HEAP8[$$sink25>>0]|0;
  $554 = HEAP8[$31>>0]|0;
  $555 = $554 ^ $553;
  HEAP8[$31>>0] = $555;
  $556 = HEAP8[$$sink21>>0]|0;
  $557 = HEAP8[$17>>0]|0;
  $558 = $557 ^ $556;
  HEAP8[$17>>0] = $558;
  $559 = HEAP8[$$sink17>>0]|0;
  $560 = HEAP8[$92>>0]|0;
  $561 = $560 ^ $559;
  HEAP8[$92>>0] = $561;
  $562 = HEAP8[$$sink13>>0]|0;
  $563 = HEAP8[$23>>0]|0;
  $564 = $563 ^ $562;
  HEAP8[$23>>0] = $564;
  $565 = HEAP8[$$sink9>>0]|0;
  $566 = HEAP8[$19>>0]|0;
  $567 = $566 ^ $565;
  HEAP8[$19>>0] = $567;
  $568 = HEAP8[$$sink5>>0]|0;
  $569 = HEAP8[$11>>0]|0;
  $570 = $569 ^ $568;
  HEAP8[$11>>0] = $570;
  $571 = HEAP8[$$sink>>0]|0;
  $572 = HEAP8[$buf>>0]|0;
  $573 = $572 ^ $571;
  HEAP8[$buf>>0] = $573;
  $575 = 0;$586 = $573;
  while(1) {
   $574 = (($buf) + ($575)|0);
   $576 = (($575) + 1)|0;
   $577 = (($buf) + ($576)|0);
   $578 = HEAP8[$577>>0]|0;
   $579 = (($575) + 2)|0;
   $580 = (($buf) + ($579)|0);
   $581 = HEAP8[$580>>0]|0;
   $582 = (($575) + 3)|0;
   $583 = (($buf) + ($582)|0);
   $584 = HEAP8[$583>>0]|0;
   $585 = $578 ^ $586;
   $587 = $585 ^ $581;
   $588 = $587 ^ $584;
   $589 = $588&255;
   $590 = $589 & 128;
   $591 = ($590|0)!=(0);
   $592 = $589 << 1;
   $593 = $592 ^ 27;
   $594 = $591 ? $593 : $592;
   $595 = $594&255;
   $596 = $581 ^ $586;
   $597 = $596 ^ $595;
   $598 = $597&255;
   $599 = $598 & 128;
   $600 = ($599|0)!=(0);
   $601 = $598 << 1;
   $602 = $601 ^ 27;
   $603 = $600 ? $602 : $601;
   $604 = $603 & 128;
   $605 = ($604|0)!=(0);
   $606 = $603 << 1;
   $607 = $606 ^ 27;
   $608 = $605 ? $607 : $606;
   $609 = $608&255;
   $610 = $609 ^ $588;
   $611 = $584 ^ $578;
   $612 = $611 ^ $595;
   $613 = $612&255;
   $614 = $613 & 128;
   $615 = ($614|0)!=(0);
   $616 = $613 << 1;
   $617 = $616 ^ 27;
   $618 = $615 ? $617 : $616;
   $619 = $618 & 128;
   $620 = ($619|0)!=(0);
   $621 = $618 << 1;
   $622 = $621 ^ 27;
   $623 = $620 ? $622 : $621;
   $624 = $623&255;
   $625 = $624 ^ $588;
   $626 = $585&255;
   $627 = $626 & 128;
   $628 = ($627|0)!=(0);
   $629 = $626 << 1;
   $630 = $629 ^ 27;
   $631 = $628 ? $630 : $629;
   $632 = $631&255;
   $633 = $632 ^ $586;
   $634 = $633 ^ $610;
   HEAP8[$574>>0] = $634;
   $635 = $581 ^ $578;
   $636 = $635&255;
   $637 = $636 & 128;
   $638 = ($637|0)!=(0);
   $639 = $636 << 1;
   $640 = $639 ^ 27;
   $641 = $638 ? $640 : $639;
   $642 = $641&255;
   $643 = $642 ^ $578;
   $644 = $643 ^ $625;
   HEAP8[$577>>0] = $644;
   $645 = $584 ^ $581;
   $646 = $645&255;
   $647 = $646 & 128;
   $648 = ($647|0)!=(0);
   $649 = $646 << 1;
   $650 = $649 ^ 27;
   $651 = $648 ? $650 : $649;
   $652 = $651&255;
   $653 = $652 ^ $581;
   $654 = $653 ^ $610;
   HEAP8[$580>>0] = $654;
   $655 = $584 ^ $586;
   $656 = $655&255;
   $657 = $656 & 128;
   $658 = ($657|0)!=(0);
   $659 = $656 << 1;
   $660 = $659 ^ 27;
   $661 = $658 ? $660 : $659;
   $662 = $661&255;
   $663 = $662 ^ $584;
   $664 = $663 ^ $625;
   HEAP8[$583>>0] = $664;
   $665 = (($575) + 4)|0;
   $666 = $665 & 255;
   $667 = ($666>>>0)<(16);
   if (!($667)) {
    break;
   }
   $$phi$trans$insert51 = (($buf) + ($666)|0);
   $$pre52 = HEAP8[$$phi$trans$insert51>>0]|0;
   $575 = $666;$586 = $$pre52;
  }
  $668 = HEAP8[$11>>0]|0;
  $669 = HEAP8[$13>>0]|0;
  HEAP8[$11>>0] = $669;
  $670 = HEAP8[$15>>0]|0;
  HEAP8[$13>>0] = $670;
  $671 = HEAP8[$17>>0]|0;
  HEAP8[$15>>0] = $671;
  HEAP8[$17>>0] = $668;
  $672 = HEAP8[$19>>0]|0;
  $673 = HEAP8[$21>>0]|0;
  HEAP8[$19>>0] = $673;
  HEAP8[$21>>0] = $672;
  $674 = HEAP8[$23>>0]|0;
  $675 = HEAP8[$25>>0]|0;
  HEAP8[$23>>0] = $675;
  $676 = HEAP8[$27>>0]|0;
  HEAP8[$25>>0] = $676;
  $677 = HEAP8[$29>>0]|0;
  HEAP8[$27>>0] = $677;
  HEAP8[$29>>0] = $674;
  $678 = HEAP8[$31>>0]|0;
  $679 = HEAP8[$33>>0]|0;
  HEAP8[$31>>0] = $679;
  HEAP8[$33>>0] = $678;
  $682 = $674;$indvars$iv$i1 = 15;
  while(1) {
   $680 = (($buf) + ($indvars$iv$i1)|0);
   $681 = $682&255;
   $683 = $681 << 1;
   $684 = $681 >>> 7;
   $685 = $683 & 254;
   $$masked$i$i = $685 | $684;
   $686 = $$masked$i$i ^ 198;
   $687 = $686 << 2;
   $688 = $686 >>> 6;
   $$masked1$i$i = $687 & 252;
   $689 = $$masked1$i$i | $688;
   $690 = $689 << 3;
   $691 = $689 >>> 5;
   $692 = $690 | $691;
   $693 = $689 ^ $686;
   $694 = $693 ^ $692;
   $695 = $694&255;
   $696 = ($695<<24>>24)==(0);
   if ($696) {
    $721 = 0;
   } else {
    $atb$0$i$i$i$i = 1;$i$0$i$i$i$i = 0;
    while(1) {
     $697 = $atb$0$i$i$i$i & 255;
     $698 = $atb$0$i$i$i$i&255;
     $699 = ($698<<24>>24)==($695<<24>>24);
     if ($699) {
      $i$1$i$i$i$i = $i$0$i$i$i$i;
      break;
     }
     $700 = $697 << 1;
     $701 = $atb$0$i$i$i$i & 128;
     $702 = ($701|0)==(0);
     $703 = $700 ^ 27;
     $atb$1$i$i$i$i = $702 ? $700 : $703;
     $704 = $atb$1$i$i$i$i & 255;
     $705 = $704 ^ $697;
     $706 = (($i$0$i$i$i$i) + 1)<<24>>24;
     $707 = ($706<<24>>24)==(0);
     if ($707) {
      $i$1$i$i$i$i = 0;
      break;
     } else {
      $atb$0$i$i$i$i = $705;$i$0$i$i$i$i = $706;
     }
    }
    $708 = $i$1$i$i$i$i&255;
    $709 = $708 ^ 255;
    $710 = $709&255;
    $711 = ($710<<24>>24)==(0);
    if ($711) {
     $721 = 1;
    } else {
     $$02$i$i$i$i = $710;$atb$01$i$i$i$i = 1;
     while(1) {
      $712 = (($$02$i$i$i$i) + -1)<<24>>24;
      $713 = $atb$01$i$i$i$i & 255;
      $714 = $713 << 1;
      $715 = $atb$01$i$i$i$i & 128;
      $716 = ($715|0)==(0);
      $717 = $714 ^ 27;
      $atb$1$i1$i$i$i = $716 ? $714 : $717;
      $718 = $atb$1$i1$i$i$i ^ $713;
      $719 = ($712<<24>>24)==(0);
      if ($719) {
       $$lcssa89 = $718;
       break;
      } else {
       $$02$i$i$i$i = $712;$atb$01$i$i$i$i = $718;
      }
     }
     $720 = $$lcssa89&255;
     $721 = $720;
    }
   }
   HEAP8[$680>>0] = $721;
   $722 = ($indvars$iv$i1|0)==(0);
   $indvars$iv$next$i2 = (($indvars$iv$i1) + -1)|0;
   if ($722) {
    break;
   }
   $$phi$trans$insert53 = (($buf) + ($indvars$iv$next$i2)|0);
   $$pre54 = HEAP8[$$phi$trans$insert53>>0]|0;
   $682 = $$pre54;$indvars$iv$i1 = $indvars$iv$next$i2;
  }
  $112 = (($113) + -1)<<24>>24;
  $114 = ($112<<24>>24)==(0);
  if ($114) {
   break;
  } else {
   $113 = $112;$346 = $771;
  }
 }
 $723 = HEAP8[$78>>0]|0;
 $724 = HEAP8[$29>>0]|0;
 $725 = $724 ^ $723;
 HEAP8[$29>>0] = $725;
 $726 = HEAP8[$79>>0]|0;
 $727 = HEAP8[$33>>0]|0;
 $728 = $727 ^ $726;
 HEAP8[$33>>0] = $728;
 $729 = HEAP8[$80>>0]|0;
 $730 = HEAP8[$13>>0]|0;
 $731 = $730 ^ $729;
 HEAP8[$13>>0] = $731;
 $732 = HEAP8[$81>>0]|0;
 $733 = HEAP8[$82>>0]|0;
 $734 = $733 ^ $732;
 HEAP8[$82>>0] = $734;
 $735 = HEAP8[$83>>0]|0;
 $736 = HEAP8[$27>>0]|0;
 $737 = $736 ^ $735;
 HEAP8[$27>>0] = $737;
 $738 = HEAP8[$84>>0]|0;
 $739 = HEAP8[$21>>0]|0;
 $740 = $739 ^ $738;
 HEAP8[$21>>0] = $740;
 $741 = HEAP8[$85>>0]|0;
 $742 = HEAP8[$15>>0]|0;
 $743 = $742 ^ $741;
 HEAP8[$15>>0] = $743;
 $744 = HEAP8[$86>>0]|0;
 $745 = HEAP8[$87>>0]|0;
 $746 = $745 ^ $744;
 HEAP8[$87>>0] = $746;
 $747 = HEAP8[$88>>0]|0;
 $748 = HEAP8[$25>>0]|0;
 $749 = $748 ^ $747;
 HEAP8[$25>>0] = $749;
 $750 = HEAP8[$89>>0]|0;
 $751 = HEAP8[$31>>0]|0;
 $752 = $751 ^ $750;
 HEAP8[$31>>0] = $752;
 $753 = HEAP8[$90>>0]|0;
 $754 = HEAP8[$17>>0]|0;
 $755 = $754 ^ $753;
 HEAP8[$17>>0] = $755;
 $756 = HEAP8[$91>>0]|0;
 $757 = HEAP8[$92>>0]|0;
 $758 = $757 ^ $756;
 HEAP8[$92>>0] = $758;
 $759 = HEAP8[$93>>0]|0;
 $760 = HEAP8[$23>>0]|0;
 $761 = $760 ^ $759;
 HEAP8[$23>>0] = $761;
 $762 = HEAP8[$94>>0]|0;
 $763 = HEAP8[$19>>0]|0;
 $764 = $763 ^ $762;
 HEAP8[$19>>0] = $764;
 $765 = HEAP8[$95>>0]|0;
 $766 = HEAP8[$11>>0]|0;
 $767 = $766 ^ $765;
 HEAP8[$11>>0] = $767;
 $768 = HEAP8[$ctx>>0]|0;
 $769 = HEAP8[$buf>>0]|0;
 $770 = $769 ^ $768;
 HEAP8[$buf>>0] = $770;
 return;
}
function _egaes256_done($ctx) {
 $ctx = $ctx|0;
 var $0 = 0, $1 = 0, $2 = 0, $exitcond = 0, $indvars$iv = 0, $indvars$iv$next = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $indvars$iv = 0;
 while(1) {
  $0 = (((($ctx)) + 64|0) + ($indvars$iv)|0);
  HEAP8[$0>>0] = 0;
  $1 = (((($ctx)) + 32|0) + ($indvars$iv)|0);
  HEAP8[$1>>0] = 0;
  $2 = (($ctx) + ($indvars$iv)|0);
  HEAP8[$2>>0] = 0;
  $indvars$iv$next = (($indvars$iv) + 1)|0;
  $exitcond = ($indvars$iv$next|0)==(32);
  if ($exitcond) {
   break;
  } else {
   $indvars$iv = $indvars$iv$next;
  }
 }
 return;
}
function _aes_expandEncKey($k,$rc) {
 $k = $k|0;
 $rc = $rc|0;
 var $$02$i$i$i = 0, $$02$i$i$i109 = 0, $$02$i$i$i24 = 0, $$02$i$i$i41 = 0, $$02$i$i$i58 = 0, $$02$i$i$i7 = 0, $$02$i$i$i75 = 0, $$02$i$i$i92 = 0, $$lcssa = 0, $$lcssa155 = 0, $$lcssa156 = 0, $$lcssa157 = 0, $$lcssa158 = 0, $$lcssa159 = 0, $$lcssa160 = 0, $$lcssa161 = 0, $$masked$i = 0, $$masked$i115 = 0, $$masked$i13 = 0, $$masked$i30 = 0;
 var $$masked$i47 = 0, $$masked$i64 = 0, $$masked$i81 = 0, $$masked$i98 = 0, $$masked1$i = 0, $$masked1$i116 = 0, $$masked1$i14 = 0, $$masked1$i31 = 0, $$masked1$i48 = 0, $$masked1$i65 = 0, $$masked1$i82 = 0, $$masked1$i99 = 0, $$masked2$i = 0, $$masked2$i100 = 0, $$masked2$i117 = 0, $$masked2$i15 = 0, $$masked2$i32 = 0, $$masked2$i49 = 0, $$masked2$i66 = 0, $$masked2$i83 = 0;
 var $$masked4$i = 0, $$masked4$i101 = 0, $$masked4$i118 = 0, $$masked4$i16 = 0, $$masked4$i33 = 0, $$masked4$i50 = 0, $$masked4$i67 = 0, $$masked4$i84 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0;
 var $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0;
 var $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0;
 var $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0;
 var $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0;
 var $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0;
 var $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0;
 var $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0;
 var $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0;
 var $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0;
 var $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0;
 var $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0;
 var $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0;
 var $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $atb$0$i$i$i = 0, $atb$0$i$i$i104 = 0, $atb$0$i$i$i19 = 0;
 var $atb$0$i$i$i2 = 0, $atb$0$i$i$i36 = 0, $atb$0$i$i$i53 = 0, $atb$0$i$i$i70 = 0, $atb$0$i$i$i87 = 0, $atb$01$i$i$i = 0, $atb$01$i$i$i110 = 0, $atb$01$i$i$i25 = 0, $atb$01$i$i$i42 = 0, $atb$01$i$i$i59 = 0, $atb$01$i$i$i76 = 0, $atb$01$i$i$i8 = 0, $atb$01$i$i$i93 = 0, $atb$1$i$i$i = 0, $atb$1$i$i$i106 = 0, $atb$1$i$i$i21 = 0, $atb$1$i$i$i38 = 0, $atb$1$i$i$i4 = 0, $atb$1$i$i$i55 = 0, $atb$1$i$i$i72 = 0;
 var $atb$1$i$i$i89 = 0, $atb$1$i1$i$i = 0, $atb$1$i1$i$i111 = 0, $atb$1$i1$i$i26 = 0, $atb$1$i1$i$i43 = 0, $atb$1$i1$i$i60 = 0, $atb$1$i1$i$i77 = 0, $atb$1$i1$i$i9 = 0, $atb$1$i1$i$i94 = 0, $i$0$i$i$i = 0, $i$0$i$i$i1 = 0, $i$0$i$i$i103 = 0, $i$0$i$i$i18 = 0, $i$0$i$i$i35 = 0, $i$0$i$i$i52 = 0, $i$0$i$i$i69 = 0, $i$0$i$i$i86 = 0, $i$1$i$i$i = 0, $i$1$i$i$i107 = 0, $i$1$i$i$i22 = 0;
 var $i$1$i$i$i39 = 0, $i$1$i$i$i5 = 0, $i$1$i$i$i56 = 0, $i$1$i$i$i73 = 0, $i$1$i$i$i90 = 0, $phitmp$i = 0, $phitmp$i11 = 0, $phitmp$i113 = 0, $phitmp$i28 = 0, $phitmp$i45 = 0, $phitmp$i62 = 0, $phitmp$i79 = 0, $phitmp$i96 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($k)) + 29|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 if ($2) {
  $27 = 0;
 } else {
  $atb$0$i$i$i = 1;$i$0$i$i$i = 0;
  while(1) {
   $3 = $atb$0$i$i$i & 255;
   $4 = $atb$0$i$i$i&255;
   $5 = ($4<<24>>24)==($1<<24>>24);
   if ($5) {
    $i$1$i$i$i = $i$0$i$i$i;
    break;
   }
   $6 = $3 << 1;
   $7 = $atb$0$i$i$i & 128;
   $8 = ($7|0)==(0);
   $9 = $6 ^ 27;
   $atb$1$i$i$i = $8 ? $6 : $9;
   $10 = $atb$1$i$i$i & 255;
   $11 = $10 ^ $3;
   $12 = (($i$0$i$i$i) + 1)<<24>>24;
   $13 = ($12<<24>>24)==(0);
   if ($13) {
    $i$1$i$i$i = 0;
    break;
   } else {
    $atb$0$i$i$i = $11;$i$0$i$i$i = $12;
   }
  }
  $14 = $i$1$i$i$i&255;
  $15 = $14 ^ 255;
  $16 = $15&255;
  $17 = ($16<<24>>24)==(0);
  if ($17) {
   $27 = 1;
  } else {
   $$02$i$i$i = $16;$atb$01$i$i$i = 1;
   while(1) {
    $18 = (($$02$i$i$i) + -1)<<24>>24;
    $19 = $atb$01$i$i$i & 255;
    $20 = $19 << 1;
    $21 = $atb$01$i$i$i & 128;
    $22 = ($21|0)==(0);
    $23 = $20 ^ 27;
    $atb$1$i1$i$i = $22 ? $20 : $23;
    $24 = $atb$1$i1$i$i ^ $19;
    $25 = ($18<<24>>24)==(0);
    if ($25) {
     $$lcssa161 = $24;
     break;
    } else {
     $$02$i$i$i = $18;$atb$01$i$i$i = $24;
    }
   }
   $phitmp$i = $$lcssa161 & 255;
   $27 = $phitmp$i;
  }
 }
 $26 = $27 << 1;
 $28 = $27 >>> 7;
 $$masked$i = $26 & 254;
 $29 = $$masked$i | $28;
 $30 = $29 << 1;
 $31 = $$masked$i >>> 7;
 $$masked1$i = $30 & 254;
 $32 = $$masked1$i | $31;
 $33 = $32 << 1;
 $34 = $$masked1$i >>> 7;
 $$masked2$i = $33 & 254;
 $35 = $$masked2$i | $34;
 $36 = $35 << 1;
 $37 = $$masked2$i >>> 7;
 $$masked4$i = $36 | $37;
 $38 = $27 ^ 99;
 $39 = $38 ^ $29;
 $40 = $39 ^ $32;
 $41 = $40 ^ $35;
 $42 = $41 ^ $$masked4$i;
 $43 = $42&255;
 $44 = HEAP8[$rc>>0]|0;
 $45 = HEAP8[$k>>0]|0;
 $46 = $45 ^ $44;
 $47 = $46 ^ $43;
 HEAP8[$k>>0] = $47;
 $48 = ((($k)) + 30|0);
 $49 = HEAP8[$48>>0]|0;
 $50 = ($49<<24>>24)==(0);
 if ($50) {
  $75 = 0;
 } else {
  $atb$0$i$i$i104 = 1;$i$0$i$i$i103 = 0;
  while(1) {
   $51 = $atb$0$i$i$i104 & 255;
   $52 = $atb$0$i$i$i104&255;
   $53 = ($52<<24>>24)==($49<<24>>24);
   if ($53) {
    $i$1$i$i$i107 = $i$0$i$i$i103;
    break;
   }
   $54 = $51 << 1;
   $55 = $atb$0$i$i$i104 & 128;
   $56 = ($55|0)==(0);
   $57 = $54 ^ 27;
   $atb$1$i$i$i106 = $56 ? $54 : $57;
   $58 = $atb$1$i$i$i106 & 255;
   $59 = $58 ^ $51;
   $60 = (($i$0$i$i$i103) + 1)<<24>>24;
   $61 = ($60<<24>>24)==(0);
   if ($61) {
    $i$1$i$i$i107 = 0;
    break;
   } else {
    $atb$0$i$i$i104 = $59;$i$0$i$i$i103 = $60;
   }
  }
  $62 = $i$1$i$i$i107&255;
  $63 = $62 ^ 255;
  $64 = $63&255;
  $65 = ($64<<24>>24)==(0);
  if ($65) {
   $75 = 1;
  } else {
   $$02$i$i$i109 = $64;$atb$01$i$i$i110 = 1;
   while(1) {
    $66 = (($$02$i$i$i109) + -1)<<24>>24;
    $67 = $atb$01$i$i$i110 & 255;
    $68 = $67 << 1;
    $69 = $atb$01$i$i$i110 & 128;
    $70 = ($69|0)==(0);
    $71 = $68 ^ 27;
    $atb$1$i1$i$i111 = $70 ? $68 : $71;
    $72 = $atb$1$i1$i$i111 ^ $67;
    $73 = ($66<<24>>24)==(0);
    if ($73) {
     $$lcssa160 = $72;
     break;
    } else {
     $$02$i$i$i109 = $66;$atb$01$i$i$i110 = $72;
    }
   }
   $phitmp$i113 = $$lcssa160 & 255;
   $75 = $phitmp$i113;
  }
 }
 $74 = $75 << 1;
 $76 = $75 >>> 7;
 $$masked$i115 = $74 & 254;
 $77 = $$masked$i115 | $76;
 $78 = $77 << 1;
 $79 = $$masked$i115 >>> 7;
 $$masked1$i116 = $78 & 254;
 $80 = $$masked1$i116 | $79;
 $81 = $80 << 1;
 $82 = $$masked1$i116 >>> 7;
 $$masked2$i117 = $81 & 254;
 $83 = $$masked2$i117 | $82;
 $84 = $83 << 1;
 $85 = $$masked2$i117 >>> 7;
 $$masked4$i118 = $84 | $85;
 $86 = $75 ^ 99;
 $87 = $86 ^ $77;
 $88 = $87 ^ $80;
 $89 = $88 ^ $83;
 $90 = $89 ^ $$masked4$i118;
 $91 = $90&255;
 $92 = ((($k)) + 1|0);
 $93 = HEAP8[$92>>0]|0;
 $94 = $91 ^ $93;
 HEAP8[$92>>0] = $94;
 $95 = ((($k)) + 31|0);
 $96 = HEAP8[$95>>0]|0;
 $97 = ($96<<24>>24)==(0);
 if ($97) {
  $122 = 0;
 } else {
  $atb$0$i$i$i87 = 1;$i$0$i$i$i86 = 0;
  while(1) {
   $98 = $atb$0$i$i$i87 & 255;
   $99 = $atb$0$i$i$i87&255;
   $100 = ($99<<24>>24)==($96<<24>>24);
   if ($100) {
    $i$1$i$i$i90 = $i$0$i$i$i86;
    break;
   }
   $101 = $98 << 1;
   $102 = $atb$0$i$i$i87 & 128;
   $103 = ($102|0)==(0);
   $104 = $101 ^ 27;
   $atb$1$i$i$i89 = $103 ? $101 : $104;
   $105 = $atb$1$i$i$i89 & 255;
   $106 = $105 ^ $98;
   $107 = (($i$0$i$i$i86) + 1)<<24>>24;
   $108 = ($107<<24>>24)==(0);
   if ($108) {
    $i$1$i$i$i90 = 0;
    break;
   } else {
    $atb$0$i$i$i87 = $106;$i$0$i$i$i86 = $107;
   }
  }
  $109 = $i$1$i$i$i90&255;
  $110 = $109 ^ 255;
  $111 = $110&255;
  $112 = ($111<<24>>24)==(0);
  if ($112) {
   $122 = 1;
  } else {
   $$02$i$i$i92 = $111;$atb$01$i$i$i93 = 1;
   while(1) {
    $113 = (($$02$i$i$i92) + -1)<<24>>24;
    $114 = $atb$01$i$i$i93 & 255;
    $115 = $114 << 1;
    $116 = $atb$01$i$i$i93 & 128;
    $117 = ($116|0)==(0);
    $118 = $115 ^ 27;
    $atb$1$i1$i$i94 = $117 ? $115 : $118;
    $119 = $atb$1$i1$i$i94 ^ $114;
    $120 = ($113<<24>>24)==(0);
    if ($120) {
     $$lcssa159 = $119;
     break;
    } else {
     $$02$i$i$i92 = $113;$atb$01$i$i$i93 = $119;
    }
   }
   $phitmp$i96 = $$lcssa159 & 255;
   $122 = $phitmp$i96;
  }
 }
 $121 = $122 << 1;
 $123 = $122 >>> 7;
 $$masked$i98 = $121 & 254;
 $124 = $$masked$i98 | $123;
 $125 = $124 << 1;
 $126 = $$masked$i98 >>> 7;
 $$masked1$i99 = $125 & 254;
 $127 = $$masked1$i99 | $126;
 $128 = $127 << 1;
 $129 = $$masked1$i99 >>> 7;
 $$masked2$i100 = $128 & 254;
 $130 = $$masked2$i100 | $129;
 $131 = $130 << 1;
 $132 = $$masked2$i100 >>> 7;
 $$masked4$i101 = $131 | $132;
 $133 = $122 ^ 99;
 $134 = $133 ^ $124;
 $135 = $134 ^ $127;
 $136 = $135 ^ $130;
 $137 = $136 ^ $$masked4$i101;
 $138 = $137&255;
 $139 = ((($k)) + 2|0);
 $140 = HEAP8[$139>>0]|0;
 $141 = $138 ^ $140;
 HEAP8[$139>>0] = $141;
 $142 = ((($k)) + 28|0);
 $143 = HEAP8[$142>>0]|0;
 $144 = ($143<<24>>24)==(0);
 if ($144) {
  $169 = 0;
 } else {
  $atb$0$i$i$i70 = 1;$i$0$i$i$i69 = 0;
  while(1) {
   $145 = $atb$0$i$i$i70 & 255;
   $146 = $atb$0$i$i$i70&255;
   $147 = ($146<<24>>24)==($143<<24>>24);
   if ($147) {
    $i$1$i$i$i73 = $i$0$i$i$i69;
    break;
   }
   $148 = $145 << 1;
   $149 = $atb$0$i$i$i70 & 128;
   $150 = ($149|0)==(0);
   $151 = $148 ^ 27;
   $atb$1$i$i$i72 = $150 ? $148 : $151;
   $152 = $atb$1$i$i$i72 & 255;
   $153 = $152 ^ $145;
   $154 = (($i$0$i$i$i69) + 1)<<24>>24;
   $155 = ($154<<24>>24)==(0);
   if ($155) {
    $i$1$i$i$i73 = 0;
    break;
   } else {
    $atb$0$i$i$i70 = $153;$i$0$i$i$i69 = $154;
   }
  }
  $156 = $i$1$i$i$i73&255;
  $157 = $156 ^ 255;
  $158 = $157&255;
  $159 = ($158<<24>>24)==(0);
  if ($159) {
   $169 = 1;
  } else {
   $$02$i$i$i75 = $158;$atb$01$i$i$i76 = 1;
   while(1) {
    $160 = (($$02$i$i$i75) + -1)<<24>>24;
    $161 = $atb$01$i$i$i76 & 255;
    $162 = $161 << 1;
    $163 = $atb$01$i$i$i76 & 128;
    $164 = ($163|0)==(0);
    $165 = $162 ^ 27;
    $atb$1$i1$i$i77 = $164 ? $162 : $165;
    $166 = $atb$1$i1$i$i77 ^ $161;
    $167 = ($160<<24>>24)==(0);
    if ($167) {
     $$lcssa158 = $166;
     break;
    } else {
     $$02$i$i$i75 = $160;$atb$01$i$i$i76 = $166;
    }
   }
   $phitmp$i79 = $$lcssa158 & 255;
   $169 = $phitmp$i79;
  }
 }
 $168 = $169 << 1;
 $170 = $169 >>> 7;
 $$masked$i81 = $168 & 254;
 $171 = $$masked$i81 | $170;
 $172 = $171 << 1;
 $173 = $$masked$i81 >>> 7;
 $$masked1$i82 = $172 & 254;
 $174 = $$masked1$i82 | $173;
 $175 = $174 << 1;
 $176 = $$masked1$i82 >>> 7;
 $$masked2$i83 = $175 & 254;
 $177 = $$masked2$i83 | $176;
 $178 = $177 << 1;
 $179 = $$masked2$i83 >>> 7;
 $$masked4$i84 = $178 | $179;
 $180 = $169 ^ 99;
 $181 = $180 ^ $171;
 $182 = $181 ^ $174;
 $183 = $182 ^ $177;
 $184 = $183 ^ $$masked4$i84;
 $185 = $184&255;
 $186 = ((($k)) + 3|0);
 $187 = HEAP8[$186>>0]|0;
 $188 = $185 ^ $187;
 HEAP8[$186>>0] = $188;
 $189 = HEAP8[$rc>>0]|0;
 $190 = $189&255;
 $191 = $190 << 1;
 $192 = $190 >>> 7;
 $193 = (0 - ($192))|0;
 $194 = $193 & 27;
 $195 = $194 ^ $191;
 $196 = $195&255;
 HEAP8[$rc>>0] = $196;
 $197 = HEAP8[$k>>0]|0;
 $198 = ((($k)) + 4|0);
 $199 = HEAP8[$198>>0]|0;
 $200 = $199 ^ $197;
 HEAP8[$198>>0] = $200;
 $201 = HEAP8[$92>>0]|0;
 $202 = ((($k)) + 5|0);
 $203 = HEAP8[$202>>0]|0;
 $204 = $203 ^ $201;
 HEAP8[$202>>0] = $204;
 $205 = HEAP8[$139>>0]|0;
 $206 = ((($k)) + 6|0);
 $207 = HEAP8[$206>>0]|0;
 $208 = $207 ^ $205;
 HEAP8[$206>>0] = $208;
 $209 = HEAP8[$186>>0]|0;
 $210 = ((($k)) + 7|0);
 $211 = HEAP8[$210>>0]|0;
 $212 = $211 ^ $209;
 HEAP8[$210>>0] = $212;
 $213 = ((($k)) + 8|0);
 $214 = HEAP8[$213>>0]|0;
 $215 = $214 ^ $200;
 HEAP8[$213>>0] = $215;
 $216 = ((($k)) + 9|0);
 $217 = HEAP8[$216>>0]|0;
 $218 = $217 ^ $204;
 HEAP8[$216>>0] = $218;
 $219 = ((($k)) + 10|0);
 $220 = HEAP8[$219>>0]|0;
 $221 = $220 ^ $208;
 HEAP8[$219>>0] = $221;
 $222 = ((($k)) + 11|0);
 $223 = HEAP8[$222>>0]|0;
 $224 = $223 ^ $212;
 HEAP8[$222>>0] = $224;
 $225 = ((($k)) + 12|0);
 $226 = HEAP8[$225>>0]|0;
 $227 = $226 ^ $215;
 HEAP8[$225>>0] = $227;
 $228 = ((($k)) + 13|0);
 $229 = HEAP8[$228>>0]|0;
 $230 = $229 ^ $218;
 HEAP8[$228>>0] = $230;
 $231 = ((($k)) + 14|0);
 $232 = HEAP8[$231>>0]|0;
 $233 = $232 ^ $221;
 HEAP8[$231>>0] = $233;
 $234 = ((($k)) + 15|0);
 $235 = HEAP8[$234>>0]|0;
 $236 = $235 ^ $224;
 HEAP8[$234>>0] = $236;
 $237 = ($226<<24>>24)==($215<<24>>24);
 if ($237) {
  $262 = 0;
 } else {
  $atb$0$i$i$i53 = 1;$i$0$i$i$i52 = 0;
  while(1) {
   $238 = $atb$0$i$i$i53 & 255;
   $239 = $atb$0$i$i$i53&255;
   $240 = ($239<<24>>24)==($227<<24>>24);
   if ($240) {
    $i$1$i$i$i56 = $i$0$i$i$i52;
    break;
   }
   $241 = $238 << 1;
   $242 = $atb$0$i$i$i53 & 128;
   $243 = ($242|0)==(0);
   $244 = $241 ^ 27;
   $atb$1$i$i$i55 = $243 ? $241 : $244;
   $245 = $atb$1$i$i$i55 & 255;
   $246 = $245 ^ $238;
   $247 = (($i$0$i$i$i52) + 1)<<24>>24;
   $248 = ($247<<24>>24)==(0);
   if ($248) {
    $i$1$i$i$i56 = 0;
    break;
   } else {
    $atb$0$i$i$i53 = $246;$i$0$i$i$i52 = $247;
   }
  }
  $249 = $i$1$i$i$i56&255;
  $250 = $249 ^ 255;
  $251 = $250&255;
  $252 = ($251<<24>>24)==(0);
  if ($252) {
   $262 = 1;
  } else {
   $$02$i$i$i58 = $251;$atb$01$i$i$i59 = 1;
   while(1) {
    $253 = (($$02$i$i$i58) + -1)<<24>>24;
    $254 = $atb$01$i$i$i59 & 255;
    $255 = $254 << 1;
    $256 = $atb$01$i$i$i59 & 128;
    $257 = ($256|0)==(0);
    $258 = $255 ^ 27;
    $atb$1$i1$i$i60 = $257 ? $255 : $258;
    $259 = $atb$1$i1$i$i60 ^ $254;
    $260 = ($253<<24>>24)==(0);
    if ($260) {
     $$lcssa157 = $259;
     break;
    } else {
     $$02$i$i$i58 = $253;$atb$01$i$i$i59 = $259;
    }
   }
   $phitmp$i62 = $$lcssa157 & 255;
   $262 = $phitmp$i62;
  }
 }
 $261 = $262 << 1;
 $263 = $262 >>> 7;
 $$masked$i64 = $261 & 254;
 $264 = $$masked$i64 | $263;
 $265 = $264 << 1;
 $266 = $$masked$i64 >>> 7;
 $$masked1$i65 = $265 & 254;
 $267 = $$masked1$i65 | $266;
 $268 = $267 << 1;
 $269 = $$masked1$i65 >>> 7;
 $$masked2$i66 = $268 & 254;
 $270 = $$masked2$i66 | $269;
 $271 = $270 << 1;
 $272 = $$masked2$i66 >>> 7;
 $$masked4$i67 = $271 | $272;
 $273 = $262 ^ 99;
 $274 = $273 ^ $264;
 $275 = $274 ^ $267;
 $276 = $275 ^ $270;
 $277 = $276 ^ $$masked4$i67;
 $278 = $277&255;
 $279 = ((($k)) + 16|0);
 $280 = HEAP8[$279>>0]|0;
 $281 = $278 ^ $280;
 HEAP8[$279>>0] = $281;
 $282 = ($229<<24>>24)==($218<<24>>24);
 if ($282) {
  $307 = 0;
 } else {
  $atb$0$i$i$i36 = 1;$i$0$i$i$i35 = 0;
  while(1) {
   $283 = $atb$0$i$i$i36 & 255;
   $284 = $atb$0$i$i$i36&255;
   $285 = ($284<<24>>24)==($230<<24>>24);
   if ($285) {
    $i$1$i$i$i39 = $i$0$i$i$i35;
    break;
   }
   $286 = $283 << 1;
   $287 = $atb$0$i$i$i36 & 128;
   $288 = ($287|0)==(0);
   $289 = $286 ^ 27;
   $atb$1$i$i$i38 = $288 ? $286 : $289;
   $290 = $atb$1$i$i$i38 & 255;
   $291 = $290 ^ $283;
   $292 = (($i$0$i$i$i35) + 1)<<24>>24;
   $293 = ($292<<24>>24)==(0);
   if ($293) {
    $i$1$i$i$i39 = 0;
    break;
   } else {
    $atb$0$i$i$i36 = $291;$i$0$i$i$i35 = $292;
   }
  }
  $294 = $i$1$i$i$i39&255;
  $295 = $294 ^ 255;
  $296 = $295&255;
  $297 = ($296<<24>>24)==(0);
  if ($297) {
   $307 = 1;
  } else {
   $$02$i$i$i41 = $296;$atb$01$i$i$i42 = 1;
   while(1) {
    $298 = (($$02$i$i$i41) + -1)<<24>>24;
    $299 = $atb$01$i$i$i42 & 255;
    $300 = $299 << 1;
    $301 = $atb$01$i$i$i42 & 128;
    $302 = ($301|0)==(0);
    $303 = $300 ^ 27;
    $atb$1$i1$i$i43 = $302 ? $300 : $303;
    $304 = $atb$1$i1$i$i43 ^ $299;
    $305 = ($298<<24>>24)==(0);
    if ($305) {
     $$lcssa156 = $304;
     break;
    } else {
     $$02$i$i$i41 = $298;$atb$01$i$i$i42 = $304;
    }
   }
   $phitmp$i45 = $$lcssa156 & 255;
   $307 = $phitmp$i45;
  }
 }
 $306 = $307 << 1;
 $308 = $307 >>> 7;
 $$masked$i47 = $306 & 254;
 $309 = $$masked$i47 | $308;
 $310 = $309 << 1;
 $311 = $$masked$i47 >>> 7;
 $$masked1$i48 = $310 & 254;
 $312 = $$masked1$i48 | $311;
 $313 = $312 << 1;
 $314 = $$masked1$i48 >>> 7;
 $$masked2$i49 = $313 & 254;
 $315 = $$masked2$i49 | $314;
 $316 = $315 << 1;
 $317 = $$masked2$i49 >>> 7;
 $$masked4$i50 = $316 | $317;
 $318 = $307 ^ 99;
 $319 = $318 ^ $309;
 $320 = $319 ^ $312;
 $321 = $320 ^ $315;
 $322 = $321 ^ $$masked4$i50;
 $323 = $322&255;
 $324 = ((($k)) + 17|0);
 $325 = HEAP8[$324>>0]|0;
 $326 = $323 ^ $325;
 HEAP8[$324>>0] = $326;
 $327 = ($232<<24>>24)==($221<<24>>24);
 if ($327) {
  $352 = 0;
 } else {
  $atb$0$i$i$i19 = 1;$i$0$i$i$i18 = 0;
  while(1) {
   $328 = $atb$0$i$i$i19 & 255;
   $329 = $atb$0$i$i$i19&255;
   $330 = ($329<<24>>24)==($233<<24>>24);
   if ($330) {
    $i$1$i$i$i22 = $i$0$i$i$i18;
    break;
   }
   $331 = $328 << 1;
   $332 = $atb$0$i$i$i19 & 128;
   $333 = ($332|0)==(0);
   $334 = $331 ^ 27;
   $atb$1$i$i$i21 = $333 ? $331 : $334;
   $335 = $atb$1$i$i$i21 & 255;
   $336 = $335 ^ $328;
   $337 = (($i$0$i$i$i18) + 1)<<24>>24;
   $338 = ($337<<24>>24)==(0);
   if ($338) {
    $i$1$i$i$i22 = 0;
    break;
   } else {
    $atb$0$i$i$i19 = $336;$i$0$i$i$i18 = $337;
   }
  }
  $339 = $i$1$i$i$i22&255;
  $340 = $339 ^ 255;
  $341 = $340&255;
  $342 = ($341<<24>>24)==(0);
  if ($342) {
   $352 = 1;
  } else {
   $$02$i$i$i24 = $341;$atb$01$i$i$i25 = 1;
   while(1) {
    $343 = (($$02$i$i$i24) + -1)<<24>>24;
    $344 = $atb$01$i$i$i25 & 255;
    $345 = $344 << 1;
    $346 = $atb$01$i$i$i25 & 128;
    $347 = ($346|0)==(0);
    $348 = $345 ^ 27;
    $atb$1$i1$i$i26 = $347 ? $345 : $348;
    $349 = $atb$1$i1$i$i26 ^ $344;
    $350 = ($343<<24>>24)==(0);
    if ($350) {
     $$lcssa155 = $349;
     break;
    } else {
     $$02$i$i$i24 = $343;$atb$01$i$i$i25 = $349;
    }
   }
   $phitmp$i28 = $$lcssa155 & 255;
   $352 = $phitmp$i28;
  }
 }
 $351 = $352 << 1;
 $353 = $352 >>> 7;
 $$masked$i30 = $351 & 254;
 $354 = $$masked$i30 | $353;
 $355 = $354 << 1;
 $356 = $$masked$i30 >>> 7;
 $$masked1$i31 = $355 & 254;
 $357 = $$masked1$i31 | $356;
 $358 = $357 << 1;
 $359 = $$masked1$i31 >>> 7;
 $$masked2$i32 = $358 & 254;
 $360 = $$masked2$i32 | $359;
 $361 = $360 << 1;
 $362 = $$masked2$i32 >>> 7;
 $$masked4$i33 = $361 | $362;
 $363 = $352 ^ 99;
 $364 = $363 ^ $354;
 $365 = $364 ^ $357;
 $366 = $365 ^ $360;
 $367 = $366 ^ $$masked4$i33;
 $368 = $367&255;
 $369 = ((($k)) + 18|0);
 $370 = HEAP8[$369>>0]|0;
 $371 = $368 ^ $370;
 HEAP8[$369>>0] = $371;
 $372 = ($235<<24>>24)==($224<<24>>24);
 if ($372) {
  $397 = 0;
 } else {
  $atb$0$i$i$i2 = 1;$i$0$i$i$i1 = 0;
  while(1) {
   $373 = $atb$0$i$i$i2 & 255;
   $374 = $atb$0$i$i$i2&255;
   $375 = ($374<<24>>24)==($236<<24>>24);
   if ($375) {
    $i$1$i$i$i5 = $i$0$i$i$i1;
    break;
   }
   $376 = $373 << 1;
   $377 = $atb$0$i$i$i2 & 128;
   $378 = ($377|0)==(0);
   $379 = $376 ^ 27;
   $atb$1$i$i$i4 = $378 ? $376 : $379;
   $380 = $atb$1$i$i$i4 & 255;
   $381 = $380 ^ $373;
   $382 = (($i$0$i$i$i1) + 1)<<24>>24;
   $383 = ($382<<24>>24)==(0);
   if ($383) {
    $i$1$i$i$i5 = 0;
    break;
   } else {
    $atb$0$i$i$i2 = $381;$i$0$i$i$i1 = $382;
   }
  }
  $384 = $i$1$i$i$i5&255;
  $385 = $384 ^ 255;
  $386 = $385&255;
  $387 = ($386<<24>>24)==(0);
  if ($387) {
   $397 = 1;
  } else {
   $$02$i$i$i7 = $386;$atb$01$i$i$i8 = 1;
   while(1) {
    $388 = (($$02$i$i$i7) + -1)<<24>>24;
    $389 = $atb$01$i$i$i8 & 255;
    $390 = $389 << 1;
    $391 = $atb$01$i$i$i8 & 128;
    $392 = ($391|0)==(0);
    $393 = $390 ^ 27;
    $atb$1$i1$i$i9 = $392 ? $390 : $393;
    $394 = $atb$1$i1$i$i9 ^ $389;
    $395 = ($388<<24>>24)==(0);
    if ($395) {
     $$lcssa = $394;
     break;
    } else {
     $$02$i$i$i7 = $388;$atb$01$i$i$i8 = $394;
    }
   }
   $phitmp$i11 = $$lcssa & 255;
   $397 = $phitmp$i11;
  }
 }
 $396 = $397 << 1;
 $398 = $397 >>> 7;
 $$masked$i13 = $396 & 254;
 $399 = $$masked$i13 | $398;
 $400 = $399 << 1;
 $401 = $$masked$i13 >>> 7;
 $$masked1$i14 = $400 & 254;
 $402 = $$masked1$i14 | $401;
 $403 = $402 << 1;
 $404 = $$masked1$i14 >>> 7;
 $$masked2$i15 = $403 & 254;
 $405 = $$masked2$i15 | $404;
 $406 = $405 << 1;
 $407 = $$masked2$i15 >>> 7;
 $$masked4$i16 = $406 | $407;
 $408 = $397 ^ 99;
 $409 = $408 ^ $399;
 $410 = $409 ^ $402;
 $411 = $410 ^ $405;
 $412 = $411 ^ $$masked4$i16;
 $413 = $412&255;
 $414 = ((($k)) + 19|0);
 $415 = HEAP8[$414>>0]|0;
 $416 = $413 ^ $415;
 HEAP8[$414>>0] = $416;
 $417 = ((($k)) + 20|0);
 $418 = HEAP8[$417>>0]|0;
 $419 = $418 ^ $281;
 HEAP8[$417>>0] = $419;
 $420 = ((($k)) + 21|0);
 $421 = HEAP8[$420>>0]|0;
 $422 = $421 ^ $326;
 HEAP8[$420>>0] = $422;
 $423 = ((($k)) + 22|0);
 $424 = HEAP8[$423>>0]|0;
 $425 = $424 ^ $371;
 HEAP8[$423>>0] = $425;
 $426 = ((($k)) + 23|0);
 $427 = HEAP8[$426>>0]|0;
 $428 = $427 ^ $416;
 HEAP8[$426>>0] = $428;
 $429 = ((($k)) + 24|0);
 $430 = HEAP8[$429>>0]|0;
 $431 = $430 ^ $419;
 HEAP8[$429>>0] = $431;
 $432 = ((($k)) + 25|0);
 $433 = HEAP8[$432>>0]|0;
 $434 = $433 ^ $422;
 HEAP8[$432>>0] = $434;
 $435 = ((($k)) + 26|0);
 $436 = HEAP8[$435>>0]|0;
 $437 = $436 ^ $425;
 HEAP8[$435>>0] = $437;
 $438 = ((($k)) + 27|0);
 $439 = HEAP8[$438>>0]|0;
 $440 = $439 ^ $428;
 HEAP8[$438>>0] = $440;
 $441 = HEAP8[$142>>0]|0;
 $442 = $441 ^ $431;
 HEAP8[$142>>0] = $442;
 $443 = HEAP8[$0>>0]|0;
 $444 = $443 ^ $434;
 HEAP8[$0>>0] = $444;
 $445 = HEAP8[$48>>0]|0;
 $446 = $445 ^ $437;
 HEAP8[$48>>0] = $446;
 $447 = HEAP8[$95>>0]|0;
 $448 = $447 ^ $440;
 HEAP8[$95>>0] = $448;
 return;
}
function __ZN9ExitGames6Photon17OperationResponse15setDebugMessageEPKw($this,$msg) {
 $this = $this|0;
 $msg = $msg|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 (__ZN9ExitGames6Common7JStringaSEPKw($0,$msg)|0);
 return;
}
function __ZN9ExitGames6Photon17OperationResponse12addParameterEhRKNS_6Common6ObjectE($this,$parameterCode,$par) {
 $this = $this|0;
 $parameterCode = $parameterCode|0;
 $par = $par|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $0 = sp;
 HEAP8[$0>>0] = $parameterCode;
 $1 = ((($this)) + 24|0);
 __ZN9ExitGames6Common9Hashtable3putIhNS0_6ObjectEEEvRKT_RKT0_($1,$0,$par);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter11countResultEi($this,$resultBytes) {
 $this = $this|0;
 $resultBytes = $resultBytes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($resultBytes))|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 HEAP32[$3>>2] = $5;
 return;
}
function __ZNK9ExitGames6Photon17OperationResponse16getOperationCodeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 return ($0|0);
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter23timeForResponseCallbackEhi($this,$code,$time) {
 $this = $this|0;
 $code = $code|0;
 $time = $time|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)<($time|0);
 if (!($2)) {
  return;
 }
 HEAP32[$0>>2] = $time;
 $3 = ((($this)) + 40|0);
 HEAP8[$3>>0] = $code;
 return;
}
function __ZN9ExitGames6Photon17OperationResponseD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($0);
 $1 = ((($this)) + 4|0);
 __ZN9ExitGames6Common7JStringD2Ev($1);
 return;
}
function __ZN9ExitGames6Photon17OperationResponseC2Ehs($this,$operationCode,$returnCode) {
 $this = $this|0;
 $operationCode = $operationCode|0;
 $returnCode = $returnCode|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$this>>0] = $operationCode;
 $0 = ((($this)) + 2|0);
 HEAP16[$0>>1] = $returnCode;
 $1 = ((($this)) + 4|0);
 __ZN9ExitGames6Common7JStringC2Ej($1,0);
 $2 = ((($this)) + 20|0);
 __THREW__ = 0;
 invoke_vi(896,($2|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($7|0);
  // unreachable;
 }
 HEAP32[$2>>2] = (2720);
 $5 = HEAP8[167373]|0;
 $6 = ($5<<24>>24)==(0);
 if (!($6)) {
  return;
 }
 HEAP8[167374] = 98;
 HEAP8[167375] = 122;
 HEAP32[40417] = 0;
 HEAP8[167373] = 1;
 return;
}
function __ZNK9ExitGames6Photon16OperationRequest16getOperationCodeEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[$this>>0]|0;
 return ($0|0);
}
function __ZNK9ExitGames6Photon16OperationRequest13getParametersEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZN9ExitGames6Photon8Internal10Encryption7encryptEPKhiS4_PPhPi($plainData,$plainDataSize,$key,$encodedData,$encodedDataSize) {
 $plainData = $plainData|0;
 $plainDataSize = $plainDataSize|0;
 $key = $key|0;
 $encodedData = $encodedData|0;
 $encodedDataSize = $encodedDataSize|0;
 var $$sum = 0, $$sum1 = 0, $$sum2 = 0, $$sum3 = 0, $0 = 0, $1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0.0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0.0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $buf = 0, $ch1$0$i = 0;
 var $ctx = 0, $exitcond = 0, $exitcond$i = 0, $i$01$i = 0, $i$04 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $ctx = sp + 16|0;
 $buf = sp;
 $0 = (($plainDataSize) + 1)|0;
 $1 = (+($0|0));
 $2 = $1 * 0.0625;
 $3 = (+Math_ceil((+$2)));
 $4 = (~~(($3)));
 $5 = $4 << 4;
 HEAP32[$encodedDataSize>>2] = $5;
 $6 = $5 | 8;
 $7 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($6)|0);
 $8 = ((($7)) + 8|0);
 $9 = $7;
 $10 = $9;
 HEAP32[$10>>2] = $5;
 $11 = (($9) + 4)|0;
 $12 = $11;
 HEAP32[$12>>2] = 0;
 $13 = ($4|0)==(0);
 if ($13) {
  _egaes256_init($ctx,$key);
  _egaes256_done($ctx);
  HEAP32[$encodedData>>2] = $8;
  STACKTOP = sp;return 0;
 }
 _memset(($8|0),0,($5|0))|0;
 _egaes256_init($ctx,$key);
 $14 = ($4|0)>(0);
 if (!($14)) {
  _egaes256_done($ctx);
  HEAP32[$encodedData>>2] = $8;
  STACKTOP = sp;return 0;
 }
 $i$04 = 0;
 while(1) {
  $15 = $i$04 << 4;
  $16 = (($plainDataSize) - ($15))|0;
  $17 = ($i$04|0)>(0);
  do {
   if ($17) {
    $$sum = (($15) + -8)|0;
    $18 = (16 - ($16))|0;
    $19 = $18&255;
    $i$01$i = 0;
    while(1) {
     $20 = (($i$01$i) + 1)|0;
     $21 = ($i$01$i|0)<($16|0);
     if ($21) {
      $$sum3 = (($i$01$i) + ($15))|0;
      $22 = (($plainData) + ($$sum3)|0);
      $23 = HEAP8[$22>>0]|0;
      $ch1$0$i = $23;
     } else {
      $ch1$0$i = $19;
     }
     $$sum2 = (($$sum) + ($i$01$i))|0;
     $24 = (($7) + ($$sum2)|0);
     $25 = HEAP8[$24>>0]|0;
     $26 = $25 ^ $ch1$0$i;
     $27 = (($buf) + ($i$01$i)|0);
     HEAP8[$27>>0] = $26;
     $exitcond$i = ($20|0)==(16);
     if ($exitcond$i) {
      break;
     } else {
      $i$01$i = $20;
     }
    }
   } else {
    $28 = ($16|0)<(16);
    if ($28) {
     _memcpy(($buf|0),($plainData|0),($16|0))|0;
     $29 = (($buf) + ($16)|0);
     $30 = (16 - ($16))|0;
     $31 = $30&255;
     _memset(($29|0),($31|0),($30|0))|0;
     break;
    } else {
     dest=$buf; src=$plainData; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
     break;
    }
   }
  } while(0);
  _egaes256_encrypt_ecb($ctx,$buf);
  $$sum1 = $15 | 8;
  $32 = (($7) + ($$sum1)|0);
  dest=$32; src=$buf; stop=dest+16|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
  $33 = (($i$04) + 1)|0;
  $exitcond = ($33|0)==($4|0);
  if ($exitcond) {
   break;
  } else {
   $i$04 = $33;
  }
 }
 _egaes256_done($ctx);
 HEAP32[$encodedData>>2] = $8;
 STACKTOP = sp;return 0;
}
function _egaes256_encrypt_ecb($ctx,$buf) {
 $ctx = $ctx|0;
 $buf = $buf|0;
 var $$02$i$i$i$i = 0, $$lcssa = 0, $$masked$i$i = 0, $$masked1$i$i = 0, $$masked2$i$i = 0, $$masked4$i$i = 0, $$sink = 0, $$sink13 = 0, $$sink17 = 0, $$sink21 = 0, $$sink25 = 0, $$sink29 = 0, $$sink33 = 0, $$sink37 = 0, $$sink41 = 0, $$sink45 = 0, $$sink49 = 0, $$sink5 = 0, $$sink53 = 0, $$sink57 = 0;
 var $$sink61 = 0, $$sink9 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0;
 var $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0;
 var $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $atb$0$i$i$i$i = 0, $atb$01$i$i$i$i = 0, $atb$1$i$i$i$i = 0, $atb$1$i1$i$i$i = 0, $i$0 = 0, $i$0$i$i$i$i = 0, $i$1$i$i$i$i = 0, $indvars$iv$i = 0;
 var $indvars$iv$i1 = 0, $indvars$iv$next$i = 0, $indvars$iv$next$i2 = 0, $phitmp$i$i = 0, $rcon = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $rcon = sp;
 $indvars$iv$i = 15;
 while(1) {
  $0 = (((($ctx)) + 32|0) + ($indvars$iv$i)|0);
  $1 = HEAP8[$0>>0]|0;
  $2 = (($ctx) + ($indvars$iv$i)|0);
  HEAP8[$2>>0] = $1;
  $3 = (($buf) + ($indvars$iv$i)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4 ^ $1;
  HEAP8[$3>>0] = $5;
  $6 = (($indvars$iv$i) + 16)|0;
  $7 = (((($ctx)) + 32|0) + ($6)|0);
  $8 = HEAP8[$7>>0]|0;
  $9 = (($ctx) + ($6)|0);
  HEAP8[$9>>0] = $8;
  $10 = ($indvars$iv$i|0)==(0);
  $indvars$iv$next$i = (($indvars$iv$i) + -1)|0;
  if ($10) {
   break;
  } else {
   $indvars$iv$i = $indvars$iv$next$i;
  }
 }
 HEAP8[$rcon>>0] = 1;
 $11 = ((($buf)) + 1|0);
 $12 = ((($buf)) + 5|0);
 $13 = ((($buf)) + 9|0);
 $14 = ((($buf)) + 13|0);
 $15 = ((($buf)) + 10|0);
 $16 = ((($buf)) + 2|0);
 $17 = ((($buf)) + 3|0);
 $18 = ((($buf)) + 15|0);
 $19 = ((($buf)) + 11|0);
 $20 = ((($buf)) + 7|0);
 $21 = ((($buf)) + 14|0);
 $22 = ((($buf)) + 6|0);
 $23 = ((($ctx)) + 15|0);
 $24 = ((($ctx)) + 14|0);
 $25 = ((($ctx)) + 13|0);
 $26 = ((($ctx)) + 12|0);
 $27 = ((($buf)) + 12|0);
 $28 = ((($ctx)) + 11|0);
 $29 = ((($ctx)) + 10|0);
 $30 = ((($ctx)) + 9|0);
 $31 = ((($ctx)) + 8|0);
 $32 = ((($buf)) + 8|0);
 $33 = ((($ctx)) + 7|0);
 $34 = ((($ctx)) + 6|0);
 $35 = ((($ctx)) + 5|0);
 $36 = ((($ctx)) + 4|0);
 $37 = ((($buf)) + 4|0);
 $38 = ((($ctx)) + 3|0);
 $39 = ((($ctx)) + 2|0);
 $40 = ((($ctx)) + 1|0);
 $41 = ((($ctx)) + 16|0);
 $42 = ((($ctx)) + 31|0);
 $43 = ((($ctx)) + 30|0);
 $44 = ((($ctx)) + 29|0);
 $45 = ((($ctx)) + 28|0);
 $46 = ((($ctx)) + 27|0);
 $47 = ((($ctx)) + 26|0);
 $48 = ((($ctx)) + 25|0);
 $49 = ((($ctx)) + 24|0);
 $50 = ((($ctx)) + 23|0);
 $51 = ((($ctx)) + 22|0);
 $52 = ((($ctx)) + 21|0);
 $53 = ((($ctx)) + 20|0);
 $54 = ((($ctx)) + 19|0);
 $55 = ((($ctx)) + 18|0);
 $56 = ((($ctx)) + 17|0);
 $i$0 = 1;
 while(1) {
  $indvars$iv$i1 = 15;
  while(1) {
   $57 = (($buf) + ($indvars$iv$i1)|0);
   $58 = HEAP8[$57>>0]|0;
   $59 = ($58<<24>>24)==(0);
   if ($59) {
    $84 = 0;
   } else {
    $atb$0$i$i$i$i = 1;$i$0$i$i$i$i = 0;
    while(1) {
     $60 = $atb$0$i$i$i$i & 255;
     $61 = $atb$0$i$i$i$i&255;
     $62 = ($61<<24>>24)==($58<<24>>24);
     if ($62) {
      $i$1$i$i$i$i = $i$0$i$i$i$i;
      break;
     }
     $63 = $60 << 1;
     $64 = $atb$0$i$i$i$i & 128;
     $65 = ($64|0)==(0);
     $66 = $63 ^ 27;
     $atb$1$i$i$i$i = $65 ? $63 : $66;
     $67 = $atb$1$i$i$i$i & 255;
     $68 = $67 ^ $60;
     $69 = (($i$0$i$i$i$i) + 1)<<24>>24;
     $70 = ($69<<24>>24)==(0);
     if ($70) {
      $i$1$i$i$i$i = 0;
      break;
     } else {
      $atb$0$i$i$i$i = $68;$i$0$i$i$i$i = $69;
     }
    }
    $71 = $i$1$i$i$i$i&255;
    $72 = $71 ^ 255;
    $73 = $72&255;
    $74 = ($73<<24>>24)==(0);
    if ($74) {
     $84 = 1;
    } else {
     $$02$i$i$i$i = $73;$atb$01$i$i$i$i = 1;
     while(1) {
      $75 = (($$02$i$i$i$i) + -1)<<24>>24;
      $76 = $atb$01$i$i$i$i & 255;
      $77 = $76 << 1;
      $78 = $atb$01$i$i$i$i & 128;
      $79 = ($78|0)==(0);
      $80 = $77 ^ 27;
      $atb$1$i1$i$i$i = $79 ? $77 : $80;
      $81 = $atb$1$i1$i$i$i ^ $76;
      $82 = ($75<<24>>24)==(0);
      if ($82) {
       $$lcssa = $81;
       break;
      } else {
       $$02$i$i$i$i = $75;$atb$01$i$i$i$i = $81;
      }
     }
     $phitmp$i$i = $$lcssa & 255;
     $84 = $phitmp$i$i;
    }
   }
   $83 = $84 << 1;
   $85 = $84 >>> 7;
   $$masked$i$i = $83 & 254;
   $86 = $$masked$i$i | $85;
   $87 = $86 << 1;
   $88 = $$masked$i$i >>> 7;
   $$masked1$i$i = $87 & 254;
   $89 = $$masked1$i$i | $88;
   $90 = $89 << 1;
   $91 = $$masked1$i$i >>> 7;
   $$masked2$i$i = $90 & 254;
   $92 = $$masked2$i$i | $91;
   $93 = $92 << 1;
   $94 = $$masked2$i$i >>> 7;
   $$masked4$i$i = $93 | $94;
   $95 = $84 ^ 99;
   $96 = $95 ^ $86;
   $97 = $96 ^ $89;
   $98 = $97 ^ $92;
   $99 = $98 ^ $$masked4$i$i;
   $100 = $99&255;
   HEAP8[$57>>0] = $100;
   $101 = ($indvars$iv$i1|0)==(0);
   $indvars$iv$next$i2 = (($indvars$iv$i1) + -1)|0;
   if ($101) {
    break;
   } else {
    $indvars$iv$i1 = $indvars$iv$next$i2;
   }
  }
  $102 = ($i$0&255)<(14);
  $103 = HEAP8[$11>>0]|0;
  $104 = HEAP8[$12>>0]|0;
  HEAP8[$11>>0] = $104;
  $105 = HEAP8[$13>>0]|0;
  HEAP8[$12>>0] = $105;
  $106 = HEAP8[$14>>0]|0;
  HEAP8[$13>>0] = $106;
  HEAP8[$14>>0] = $103;
  $107 = HEAP8[$15>>0]|0;
  $108 = HEAP8[$16>>0]|0;
  HEAP8[$15>>0] = $108;
  HEAP8[$16>>0] = $107;
  $109 = HEAP8[$17>>0]|0;
  $110 = HEAP8[$18>>0]|0;
  HEAP8[$17>>0] = $110;
  $111 = HEAP8[$19>>0]|0;
  HEAP8[$18>>0] = $111;
  $112 = HEAP8[$20>>0]|0;
  HEAP8[$19>>0] = $112;
  HEAP8[$20>>0] = $109;
  $113 = HEAP8[$21>>0]|0;
  $114 = HEAP8[$22>>0]|0;
  HEAP8[$21>>0] = $114;
  HEAP8[$22>>0] = $113;
  if ($102) {
   $116 = 0;
  } else {
   break;
  }
  while(1) {
   $115 = (($buf) + ($116)|0);
   $117 = HEAP8[$115>>0]|0;
   $118 = (($116) + 1)|0;
   $119 = (($buf) + ($118)|0);
   $120 = HEAP8[$119>>0]|0;
   $121 = (($116) + 2)|0;
   $122 = (($buf) + ($121)|0);
   $123 = HEAP8[$122>>0]|0;
   $124 = (($116) + 3)|0;
   $125 = (($buf) + ($124)|0);
   $126 = HEAP8[$125>>0]|0;
   $127 = $120 ^ $117;
   $128 = $127 ^ $123;
   $129 = $128 ^ $126;
   $130 = $127&255;
   $131 = $130 & 128;
   $132 = ($131|0)!=(0);
   $133 = $130 << 1;
   $134 = $133 ^ 27;
   $135 = $132 ? $134 : $133;
   $136 = $135&255;
   $137 = $129 ^ $117;
   $138 = $137 ^ $136;
   HEAP8[$115>>0] = $138;
   $139 = $123 ^ $120;
   $140 = $139&255;
   $141 = $140 & 128;
   $142 = ($141|0)!=(0);
   $143 = $140 << 1;
   $144 = $143 ^ 27;
   $145 = $142 ? $144 : $143;
   $146 = $145&255;
   $147 = $129 ^ $120;
   $148 = $147 ^ $146;
   HEAP8[$119>>0] = $148;
   $149 = $126 ^ $123;
   $150 = $149&255;
   $151 = $150 & 128;
   $152 = ($151|0)!=(0);
   $153 = $150 << 1;
   $154 = $153 ^ 27;
   $155 = $152 ? $154 : $153;
   $156 = $155&255;
   $157 = $126 ^ $127;
   $158 = $157 ^ $156;
   HEAP8[$122>>0] = $158;
   $159 = $126 ^ $117;
   $160 = $159&255;
   $161 = $160 & 128;
   $162 = ($161|0)!=(0);
   $163 = $160 << 1;
   $164 = $163 ^ 27;
   $165 = $162 ? $164 : $163;
   $166 = $165&255;
   $167 = $166 ^ $128;
   HEAP8[$125>>0] = $167;
   $168 = (($116) + 4)|0;
   $169 = $168 & 255;
   $170 = ($169>>>0)<(16);
   if ($170) {
    $116 = $169;
   } else {
    break;
   }
  }
  $171 = $i$0 & 1;
  $172 = ($171<<24>>24)==(0);
  if ($172) {
   _aes_expandEncKey($ctx,$rcon);
   $$sink = $ctx;$$sink13 = $38;$$sink17 = $36;$$sink21 = $35;$$sink25 = $34;$$sink29 = $33;$$sink33 = $31;$$sink37 = $30;$$sink41 = $29;$$sink45 = $28;$$sink49 = $26;$$sink5 = $40;$$sink53 = $25;$$sink57 = $24;$$sink61 = $23;$$sink9 = $39;
  } else {
   $$sink = $41;$$sink13 = $54;$$sink17 = $53;$$sink21 = $52;$$sink25 = $51;$$sink29 = $50;$$sink33 = $49;$$sink37 = $48;$$sink41 = $47;$$sink45 = $46;$$sink49 = $45;$$sink5 = $56;$$sink53 = $44;$$sink57 = $43;$$sink61 = $42;$$sink9 = $55;
  }
  $173 = HEAP8[$$sink61>>0]|0;
  $174 = HEAP8[$18>>0]|0;
  $175 = $174 ^ $173;
  HEAP8[$18>>0] = $175;
  $176 = HEAP8[$$sink57>>0]|0;
  $177 = HEAP8[$21>>0]|0;
  $178 = $177 ^ $176;
  HEAP8[$21>>0] = $178;
  $179 = HEAP8[$$sink53>>0]|0;
  $180 = HEAP8[$14>>0]|0;
  $181 = $180 ^ $179;
  HEAP8[$14>>0] = $181;
  $182 = HEAP8[$$sink49>>0]|0;
  $183 = HEAP8[$27>>0]|0;
  $184 = $183 ^ $182;
  HEAP8[$27>>0] = $184;
  $185 = HEAP8[$$sink45>>0]|0;
  $186 = HEAP8[$19>>0]|0;
  $187 = $186 ^ $185;
  HEAP8[$19>>0] = $187;
  $188 = HEAP8[$$sink41>>0]|0;
  $189 = HEAP8[$15>>0]|0;
  $190 = $189 ^ $188;
  HEAP8[$15>>0] = $190;
  $191 = HEAP8[$$sink37>>0]|0;
  $192 = HEAP8[$13>>0]|0;
  $193 = $192 ^ $191;
  HEAP8[$13>>0] = $193;
  $194 = HEAP8[$$sink33>>0]|0;
  $195 = HEAP8[$32>>0]|0;
  $196 = $195 ^ $194;
  HEAP8[$32>>0] = $196;
  $197 = HEAP8[$$sink29>>0]|0;
  $198 = HEAP8[$20>>0]|0;
  $199 = $198 ^ $197;
  HEAP8[$20>>0] = $199;
  $200 = HEAP8[$$sink25>>0]|0;
  $201 = HEAP8[$22>>0]|0;
  $202 = $201 ^ $200;
  HEAP8[$22>>0] = $202;
  $203 = HEAP8[$$sink21>>0]|0;
  $204 = HEAP8[$12>>0]|0;
  $205 = $204 ^ $203;
  HEAP8[$12>>0] = $205;
  $206 = HEAP8[$$sink17>>0]|0;
  $207 = HEAP8[$37>>0]|0;
  $208 = $207 ^ $206;
  HEAP8[$37>>0] = $208;
  $209 = HEAP8[$$sink13>>0]|0;
  $210 = HEAP8[$17>>0]|0;
  $211 = $210 ^ $209;
  HEAP8[$17>>0] = $211;
  $212 = HEAP8[$$sink9>>0]|0;
  $213 = HEAP8[$16>>0]|0;
  $214 = $213 ^ $212;
  HEAP8[$16>>0] = $214;
  $215 = HEAP8[$$sink5>>0]|0;
  $216 = HEAP8[$11>>0]|0;
  $217 = $216 ^ $215;
  HEAP8[$11>>0] = $217;
  $218 = HEAP8[$$sink>>0]|0;
  $219 = HEAP8[$buf>>0]|0;
  $220 = $219 ^ $218;
  HEAP8[$buf>>0] = $220;
  $221 = (($i$0) + 1)<<24>>24;
  $i$0 = $221;
 }
 _aes_expandEncKey($ctx,$rcon);
 $222 = HEAP8[$23>>0]|0;
 $223 = HEAP8[$18>>0]|0;
 $224 = $223 ^ $222;
 HEAP8[$18>>0] = $224;
 $225 = HEAP8[$24>>0]|0;
 $226 = HEAP8[$21>>0]|0;
 $227 = $226 ^ $225;
 HEAP8[$21>>0] = $227;
 $228 = HEAP8[$25>>0]|0;
 $229 = HEAP8[$14>>0]|0;
 $230 = $229 ^ $228;
 HEAP8[$14>>0] = $230;
 $231 = HEAP8[$26>>0]|0;
 $232 = HEAP8[$27>>0]|0;
 $233 = $232 ^ $231;
 HEAP8[$27>>0] = $233;
 $234 = HEAP8[$28>>0]|0;
 $235 = HEAP8[$19>>0]|0;
 $236 = $235 ^ $234;
 HEAP8[$19>>0] = $236;
 $237 = HEAP8[$29>>0]|0;
 $238 = HEAP8[$15>>0]|0;
 $239 = $238 ^ $237;
 HEAP8[$15>>0] = $239;
 $240 = HEAP8[$30>>0]|0;
 $241 = HEAP8[$13>>0]|0;
 $242 = $241 ^ $240;
 HEAP8[$13>>0] = $242;
 $243 = HEAP8[$31>>0]|0;
 $244 = HEAP8[$32>>0]|0;
 $245 = $244 ^ $243;
 HEAP8[$32>>0] = $245;
 $246 = HEAP8[$33>>0]|0;
 $247 = HEAP8[$20>>0]|0;
 $248 = $247 ^ $246;
 HEAP8[$20>>0] = $248;
 $249 = HEAP8[$34>>0]|0;
 $250 = HEAP8[$22>>0]|0;
 $251 = $250 ^ $249;
 HEAP8[$22>>0] = $251;
 $252 = HEAP8[$35>>0]|0;
 $253 = HEAP8[$12>>0]|0;
 $254 = $253 ^ $252;
 HEAP8[$12>>0] = $254;
 $255 = HEAP8[$36>>0]|0;
 $256 = HEAP8[$37>>0]|0;
 $257 = $256 ^ $255;
 HEAP8[$37>>0] = $257;
 $258 = HEAP8[$38>>0]|0;
 $259 = HEAP8[$17>>0]|0;
 $260 = $259 ^ $258;
 HEAP8[$17>>0] = $260;
 $261 = HEAP8[$39>>0]|0;
 $262 = HEAP8[$16>>0]|0;
 $263 = $262 ^ $261;
 HEAP8[$16>>0] = $263;
 $264 = HEAP8[$40>>0]|0;
 $265 = HEAP8[$11>>0]|0;
 $266 = $265 ^ $264;
 HEAP8[$11>>0] = $266;
 $267 = HEAP8[$ctx>>0]|0;
 $268 = HEAP8[$buf>>0]|0;
 $269 = $268 ^ $267;
 HEAP8[$buf>>0] = $269;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounter22countReliableOpCommandEi($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 32|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($size))|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 HEAP32[$3>>2] = $5;
 return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter14countOperationEi($this,$operationBytes) {
 $this = $this|0;
 $operationBytes = $operationBytes|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 12|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($operationBytes))|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 HEAP32[$3>>2] = $5;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase5resetEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147238);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($19|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($2|0),4,(6560|0),($3|0),1,112,(163772|0),($vararg_buffer|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($19|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $8 = HEAP32[$this>>2]|0;
  $9 = ((($8)) + 80|0);
  $10 = HEAP32[$9>>2]|0;
  FUNCTION_TABLE_vi[$10 & 2047]($this);
  $11 = ((($this)) + 12|0);
  HEAP16[$11>>1] = -1;
  $12 = (__Z11getTimeUnixv()|0);
  $13 = ((($this)) + 44|0);
  HEAP32[$13>>2] = $12;
  $14 = ((($this)) + 16|0);
  HEAP32[$14>>2] = $12;
  $15 = ((($this)) + 60|0);
  HEAP32[$15>>2] = 300;
  $16 = ((($this)) + 96|0);
  HEAP8[$16>>0] = 0;
  $17 = ((($this)) + 56|0);
  HEAP8[$17>>0] = 0;
  $18 = ((($this)) + 52|0);
  HEAP32[$18>>2] = 0;
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Photon8Internal8PeerBase7cleanupEv($this) {
 $this = $this|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 if ($5) {
  return;
 }
 $6 = ((($2)) + 20|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 if ($8) {
  $11 = $2;
 } else {
  $9 = ((($7)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($9);
  $$pre$i = HEAP32[$1>>2]|0;
  $11 = $$pre$i;
 }
 $10 = ((($11)) + 20|0);
 HEAP32[$10>>2] = 0;
 return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter30dispatchIncomingCommandsCalledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = (__Z11getTimeUnixv()|0);
  $4 = HEAP32[$0>>2]|0;
  $5 = (($3) - ($4))|0;
  $6 = ((($this)) + 52|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($5|0)>($7|0);
  if ($8) {
   HEAP32[$6>>2] = $5;
  }
 }
 $9 = ((($this)) + 60|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + 1)|0;
 HEAP32[$9>>2] = $11;
 $12 = (__Z11getTimeUnixv()|0);
 HEAP32[$0>>2] = $12;
 return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter26sendOutgoingCommandsCalledEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 8|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = (__Z11getTimeUnixv()|0);
  $4 = HEAP32[$0>>2]|0;
  $5 = (($3) - ($4))|0;
  $6 = ((($this)) + 56|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($5|0)>($7|0);
  if ($8) {
   HEAP32[$6>>2] = $5;
  }
 }
 $9 = ((($this)) + 64|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (($10) + 1)|0;
 HEAP32[$9>>2] = $11;
 $12 = (__Z11getTimeUnixv()|0);
 HEAP32[$0>>2] = $12;
 return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounter25incrementTotalPacketCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + 1)|0;
 HEAP32[$0>>2] = $2;
 return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounter27countTotalCommandsInPacketsEi($this,$amount) {
 $this = $this|0;
 $amount = $amount|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 28|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($amount))|0;
 HEAP32[$0>>2] = $2;
 return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounter19countControlCommandEi($this,$size) {
 $this = $this|0;
 $size = $size|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (($1) + ($size))|0;
 HEAP32[$0>>2] = $2;
 $3 = ((($this)) + 20|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($4) + 1)|0;
 HEAP32[$3>>2] = $5;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase7connectERKNS_6Common7JStringES6_RKNS3_6ObjectE($this,$ipAddr,$appID,$customData) {
 $this = $this|0;
 $ipAddr = $ipAddr|0;
 $appID = $appID|0;
 $customData = $customData|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$sink = 0, $$01 = 0, $$01$i$i = 0, $$01$i$i$sink = 0, $$03 = 0, $$1 = 0, $$12 = 0, $$14 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $appIdString = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 96|0;
 $1 = sp + 80|0;
 $2 = sp + 64|0;
 $appIdString = sp + 48|0;
 $3 = sp + 32|0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147272);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($0|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  __THREW__ = 0;
  $9 = (invoke_ii(764,($ipAddr|0))|0);
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   __THREW__ = 0;
   HEAP32[$vararg_buffer>>2] = $9;
   invoke_viiiiiiii(765,($5|0),3,(6560|0),($6|0),1,129,(9224|0),($vararg_buffer|0));
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    __ZN9ExitGames6Common7JStringD2Ev($0);
    $14 = (__ZNK9ExitGames6Common7JStringcvPKwEv($ipAddr)|0);
    $15 = ($14|0)==(0|0);
    if ($15) {
     $16 = HEAP32[$4>>2]|0;
     __ZN9ExitGames6Common7JStringC2EPKc($1,147272);
     __THREW__ = 0;
     $17 = (invoke_ii(764,($1|0))|0);
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if (!($19)) {
      __THREW__ = 0;
      invoke_viiiiiiii(765,($16|0),1,(6560|0),($17|0),1,133,(9276|0),($vararg_buffer1|0));
      $20 = __THREW__; __THREW__ = 0;
      $21 = $20&1;
      if (!($21)) {
       __ZN9ExitGames6Common7JStringD2Ev($1);
       $$14 = 0;
       STACKTOP = sp;return ($$14|0);
      }
     }
     $24 = ___cxa_find_matching_catch_2()|0;
     $25 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($1);
     $$1 = $25;$$12 = $24;
     ___resumeException($$12|0);
     // unreachable;
    }
    $26 = ((($this)) + 36|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ($27|0)==(0);
    if (!($28)) {
     $29 = HEAP32[$4>>2]|0;
     __ZN9ExitGames6Common7JStringC2EPKc($2,147272);
     __THREW__ = 0;
     $30 = (invoke_ii(764,($2|0))|0);
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if (!($32)) {
      __THREW__ = 0;
      invoke_viiiiiiii(765,($29|0),1,(6560|0),($30|0),1,139,(9384|0),($vararg_buffer3|0));
      $33 = __THREW__; __THREW__ = 0;
      $34 = $33&1;
      if (!($34)) {
       __ZN9ExitGames6Common7JStringD2Ev($2);
       $$14 = 0;
       STACKTOP = sp;return ($$14|0);
      }
     }
     $35 = ___cxa_find_matching_catch_2()|0;
     $36 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($2);
     $$1 = $36;$$12 = $35;
     ___resumeException($$12|0);
     // unreachable;
    }
    $37 = HEAP32[$this>>2]|0;
    $38 = ((($37)) + 84|0);
    $39 = HEAP32[$38>>2]|0;
    FUNCTION_TABLE_vi[$39 & 2047]($this);
    $40 = ((($this)) + 20|0);
    (__ZN9ExitGames6Common7JStringaSERKS1_($40,$ipAddr)|0);
    $41 = (__ZNK9ExitGames6Common7JString6lengthEv($appID)|0);
    $42 = ($41|0)==(0);
    if ($42) {
     __ZN9ExitGames6Common10ANSIStringC2EPKc($appIdString,147280);
    } else {
     __ZNK9ExitGames6Common7JString18ANSIRepresentationEv($appIdString,$appID);
    }
    $43 = ((($this)) + 4|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = ((($44)) + 61|0);
    __THREW__ = 0;
    $46 = (invoke_ii(826,($appIdString|0))|0);
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    L27: do {
     if ($48) {
      label = 43;
     } else {
      __THREW__ = 0;
      $49 = (invoke_ii(810,($appIdString|0))|0);
      $50 = __THREW__; __THREW__ = 0;
      $51 = $50&1;
      if ($51) {
       label = 43;
      } else {
       $52 = ($49>>>0)<(32);
       if ($52) {
        __THREW__ = 0;
        $53 = (invoke_ii(810,($appIdString|0))|0);
        $54 = __THREW__; __THREW__ = 0;
        $55 = $54&1;
        if ($55) {
         label = 43;
         break;
        } else {
         $56 = $53;
        }
       } else {
        $56 = 32;
       }
       _memcpy(($45|0),($46|0),($56|0))|0;
       $57 = ((($this)) + 148|0);
       $58 = HEAP32[$43>>2]|0;
       __THREW__ = 0;
       $59 = (invoke_ii(775,80)|0);
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        label = 43;
       } else {
        $62 = ((($58)) + 94|0);
        $63 = HEAP8[$62>>0]|0;
        __THREW__ = 0;
        invoke_vii(786,($59|0),($ipAddr|0));
        $64 = __THREW__; __THREW__ = 0;
        $65 = $64&1;
        if ($65) {
         label = 43;
        } else {
         $66 = ((($59)) + 16|0);
         __THREW__ = 0;
         invoke_vii(786,($66|0),($appID|0));
         $67 = __THREW__; __THREW__ = 0;
         $68 = $67&1;
         do {
          if ($68) {
           $72 = ___cxa_find_matching_catch_2()|0;
           $73 = tempRet0;
           $$0$i$i = $73;$$01$i$i = $72;
          } else {
           $69 = ((($59)) + 32|0);
           __THREW__ = 0;
           invoke_vii(744,($69|0),($customData|0));
           $70 = __THREW__; __THREW__ = 0;
           $71 = $70&1;
           if ($71) {
            $74 = ___cxa_find_matching_catch_2()|0;
            $75 = tempRet0;
            __ZN9ExitGames6Common7JStringD2Ev($66);
            $$0$i$i = $75;$$01$i$i = $74;
            break;
           }
           HEAP8[167383] = $63;
           $76 = $59;
           __THREW__ = 0;
           $77 = (invoke_ii(775,8)|0);
           $78 = __THREW__; __THREW__ = 0;
           $79 = $78&1;
           if ($79) {
            label = 43;
            break L27;
           }
           $80 = $77;
           $81 = $80;
           HEAP32[$81>>2] = 1;
           $82 = (($80) + 4)|0;
           $83 = $82;
           HEAP32[$83>>2] = 0;
           $84 = ((($this)) + 152|0);
           $85 = HEAP32[$84>>2]|0;
           $86 = ($85|0)==(0|0);
           L42: do {
            if (!($86)) {
             $87 = $85;
             $88 = $87;
             $89 = HEAP32[$88>>2]|0;
             $90 = (($87) + 4)|0;
             $91 = $90;
             $92 = HEAP32[$91>>2]|0;
             $93 = (_i64Add(($89|0),($92|0),-1,-1)|0);
             $94 = tempRet0;
             $95 = $85;
             $96 = $95;
             HEAP32[$96>>2] = $93;
             $97 = (($95) + 4)|0;
             $98 = $97;
             HEAP32[$98>>2] = $94;
             $99 = ($93|0)==(0);
             $100 = ($94|0)==(0);
             $101 = $99 & $100;
             if ($101) {
              $102 = HEAP32[$57>>2]|0;
              $103 = ($102|0)==(0|0);
              do {
               if ($103) {
                $109 = $85;
                label = 35;
               } else {
                $104 = ((($102)) + 32|0);
                __ZN9ExitGames6Common6ObjectD2Ev($104);
                $105 = ((($102)) + 16|0);
                __ZN9ExitGames6Common7JStringD2Ev($105);
                __ZN9ExitGames6Common7JStringD2Ev($102);
                __THREW__ = 0;
                invoke_vi(750,($102|0));
                $106 = __THREW__; __THREW__ = 0;
                $107 = $106&1;
                if ($107) {
                 break;
                }
                $$pre$i = HEAP32[$84>>2]|0;
                $108 = ($$pre$i|0)==(0|0);
                if ($108) {
                 break L42;
                } else {
                 $109 = $$pre$i;
                 label = 35;
                }
               }
              } while(0);
              if ((label|0) == 35) {
               __THREW__ = 0;
               invoke_vi(750,($109|0));
               $110 = __THREW__; __THREW__ = 0;
               $111 = $110&1;
               if (!($111)) {
                break;
               }
              }
              $138 = ___cxa_find_matching_catch_2()|0;
              $139 = tempRet0;
              $140 = $77;
              $141 = $140;
              $142 = HEAP32[$141>>2]|0;
              $143 = (($140) + 4)|0;
              $144 = $143;
              $145 = HEAP32[$144>>2]|0;
              $146 = (_i64Add(($142|0),($145|0),-1,-1)|0);
              $147 = tempRet0;
              $148 = $77;
              $149 = $148;
              HEAP32[$149>>2] = $146;
              $150 = (($148) + 4)|0;
              $151 = $150;
              HEAP32[$151>>2] = $147;
              $152 = ($146|0)==(0);
              $153 = ($147|0)==(0);
              $154 = $152 & $153;
              if (!($154)) {
               $$0 = $139;$$01 = $138;
               break L27;
              }
              $155 = ($59|0)==(0|0);
              do {
               if (!($155)) {
                __ZN9ExitGames6Common6ObjectD2Ev($69);
                __ZN9ExitGames6Common7JStringD2Ev($66);
                __ZN9ExitGames6Common7JStringD2Ev($59);
                __THREW__ = 0;
                invoke_vi(750,($59|0));
                $156 = __THREW__; __THREW__ = 0;
                $157 = $156&1;
                if (!($157)) {
                 break;
                }
                $160 = ___cxa_find_matching_catch_3(0|0)|0;
                $161 = tempRet0;
                ___clang_call_terminate($160);
                // unreachable;
               }
              } while(0);
              __THREW__ = 0;
              invoke_vi(750,($77|0));
              $158 = __THREW__; __THREW__ = 0;
              $159 = $158&1;
              if (!($159)) {
               $$0 = $139;$$01 = $138;
               break L27;
              }
              $160 = ___cxa_find_matching_catch_3(0|0)|0;
              $161 = tempRet0;
              ___clang_call_terminate($160);
              // unreachable;
             }
            }
           } while(0);
           HEAP32[$57>>2] = $76;
           HEAP32[$84>>2] = $77;
           $112 = $77;
           $113 = $112;
           $114 = HEAP32[$113>>2]|0;
           $115 = (($112) + 4)|0;
           $116 = $115;
           $117 = HEAP32[$116>>2]|0;
           $118 = ($114|0)==(0);
           $119 = ($117|0)==(0);
           $120 = $118 & $119;
           do {
            if ($120) {
             $121 = ($59|0)==(0|0);
             do {
              if (!($121)) {
               __ZN9ExitGames6Common6ObjectD2Ev($69);
               __ZN9ExitGames6Common7JStringD2Ev($66);
               __ZN9ExitGames6Common7JStringD2Ev($59);
               __THREW__ = 0;
               invoke_vi(750,($59|0));
               $122 = __THREW__; __THREW__ = 0;
               $123 = $122&1;
               if (!($123)) {
                break;
               }
               $126 = ___cxa_find_matching_catch_3(0|0)|0;
               $127 = tempRet0;
               ___clang_call_terminate($126);
               // unreachable;
              }
             } while(0);
             __THREW__ = 0;
             invoke_vi(750,($77|0));
             $124 = __THREW__; __THREW__ = 0;
             $125 = $124&1;
             if (!($125)) {
              break;
             }
             $126 = ___cxa_find_matching_catch_3(0|0)|0;
             $127 = tempRet0;
             ___clang_call_terminate($126);
             // unreachable;
            }
           } while(0);
           $128 = ((($this)) + 40|0);
           $129 = HEAP32[$128>>2]|0;
           $130 = HEAP32[$129>>2]|0;
           $131 = ((($130)) + 4|0);
           $132 = HEAP32[$131>>2]|0;
           __THREW__ = 0;
           $133 = (invoke_iiii($132|0,($129|0),($40|0),($appID|0))|0);
           $134 = __THREW__; __THREW__ = 0;
           $135 = $134&1;
           if ($135) {
            label = 43;
            break L27;
           }
           L67: do {
            if ($133) {
             $$03 = 1;
            } else {
             $162 = HEAP32[$4>>2]|0;
             __THREW__ = 0;
             invoke_vii(739,($3|0),(147272|0));
             $163 = __THREW__; __THREW__ = 0;
             $164 = $163&1;
             if ($164) {
              label = 43;
              break L27;
             }
             __THREW__ = 0;
             $165 = (invoke_ii(764,($3|0))|0);
             $166 = __THREW__; __THREW__ = 0;
             $167 = $166&1;
             do {
              if (!($167)) {
               __THREW__ = 0;
               invoke_viiiiiiii(765,($162|0),1,(6560|0),($165|0),1,156,(9524|0),($vararg_buffer5|0));
               $168 = __THREW__; __THREW__ = 0;
               $169 = $168&1;
               if ($169) {
                break;
               }
               __ZN9ExitGames6Common7JStringD2Ev($3);
               $$03 = 0;
               break L67;
              }
             } while(0);
             $170 = ___cxa_find_matching_catch_2()|0;
             $171 = tempRet0;
             __ZN9ExitGames6Common7JStringD2Ev($3);
             $$0 = $171;$$01 = $170;
             break L27;
            }
           } while(0);
           __ZN9ExitGames6Common10ANSIStringD2Ev($appIdString);
           $$14 = $$03;
           STACKTOP = sp;return ($$14|0);
          }
         } while(0);
         __ZN9ExitGames6Common7JStringD2Ev($59);
         $$0$i$i$sink = $$0$i$i;$$01$i$i$sink = $$01$i$i;
         label = 44;
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 43) {
     $136 = ___cxa_find_matching_catch_2()|0;
     $137 = tempRet0;
     $$0$i$i$sink = $137;$$01$i$i$sink = $136;
     label = 44;
    }
    if ((label|0) == 44) {
     $$0 = $$0$i$i$sink;$$01 = $$01$i$i$sink;
    }
    __ZN9ExitGames6Common10ANSIStringD2Ev($appIdString);
    $$1 = $$0;$$12 = $$01;
    ___resumeException($$12|0);
    // unreachable;
   }
  }
 }
 $22 = ___cxa_find_matching_catch_2()|0;
 $23 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$1 = $23;$$12 = $22;
 ___resumeException($$12|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal8PeerBase9onConnectEi($this,$nError) {
 $this = $this|0;
 $nError = $nError|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $storemerge = 0;
 var $storemerge$in = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147285);
 __THREW__ = 0;
 $4 = (invoke_ii(764,($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($3|0),4,(6560|0),($4|0),1,418,(163772|0),($vararg_buffer|0));
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $9 = ($nError|0)==(0);
   do {
    if (!($9)) {
     $10 = HEAP32[$2>>2]|0;
     __ZN9ExitGames6Common7JStringC2EPKc($1,147285);
     __THREW__ = 0;
     $11 = (invoke_ii(764,($1|0))|0);
     $12 = __THREW__; __THREW__ = 0;
     $13 = $12&1;
     if (!($13)) {
      __THREW__ = 0;
      HEAP32[$vararg_buffer1>>2] = $nError;
      invoke_viiiiiiii(765,($10|0),1,(6560|0),($11|0),1,420,(9720|0),($vararg_buffer1|0));
      $14 = __THREW__; __THREW__ = 0;
      $15 = $14&1;
      if (!($15)) {
       __ZN9ExitGames6Common7JStringD2Ev($1);
       break;
      }
     }
     $18 = ___cxa_find_matching_catch_2()|0;
     $19 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($1);
     $$0 = $19;$$01 = $18;
     ___resumeException($$01|0);
     // unreachable;
    }
   } while(0);
   $20 = ((($this)) + 148|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ((($this)) + 40|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = HEAP32[$23>>2]|0;
   $25 = ((($24)) + 32|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = (FUNCTION_TABLE_ii[$26 & 2047]($23)|0);
   $28 = ((($21)) + 72|0);
   $29 = $27&1;
   HEAP8[$28>>0] = $29;
   $30 = ((($this)) + 4|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ((($31)) + 57|0);
   $33 = HEAP8[$32>>0]|0;
   $34 = $33&255;
   $35 = $34 & 127;
   $36 = $34 | 128;
   $storemerge$in = $27 ? $36 : $35;
   $storemerge = $storemerge$in&255;
   HEAP8[$32>>0] = $storemerge;
   __ZN9ExitGames6Photon8Internal8PeerBase16generateInitDataEv($this);
   STACKTOP = sp;return;
  }
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$0 = $17;$$01 = $16;
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBase16generateInitDataEv($this) {
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$10 = 0, $$1019 = 0, $$11 = 0, $$110 = 0, $$1120 = 0, $$12 = 0, $$1221 = 0, $$13 = 0, $$1322 = 0, $$14 = 0, $$1423 = 0, $$15 = 0, $$1524 = 0, $$16 = 0, $$1625 = 0, $$17 = 0, $$1726 = 0;
 var $$18 = 0, $$1827 = 0, $$19 = 0, $$1928 = 0, $$2 = 0, $$211 = 0, $$3 = 0, $$312 = 0, $$4 = 0, $$413 = 0, $$5 = 0, $$514 = 0, $$6 = 0, $$615 = 0, $$7 = 0, $$716 = 0, $$8 = 0, $$817 = 0, $$9 = 0, $$918 = 0;
 var $$pre = 0, $$pre$i = 0, $$pre$i73 = 0, $$pre128 = 0, $$sum = 0, $$sum127 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0;
 var $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0;
 var $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0;
 var $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0;
 var $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0;
 var $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0;
 var $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0;
 var $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0;
 var $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0;
 var $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0;
 var $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0;
 var $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index22Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body102$index12Z2D = 0, $eh$lpad$body102$indexZ2D = 0, $eh$lpad$body106$index42Z2D = 0, $eh$lpad$body106$indexZ2D = 0, $eh$lpad$body110$index47Z2D = 0, $eh$lpad$body110$indexZ2D = 0, $eh$lpad$body114$index52Z2D = 0, $eh$lpad$body114$indexZ2D = 0, $eh$lpad$body118$index57Z2D = 0;
 var $eh$lpad$body118$indexZ2D = 0, $eh$lpad$body122$index62Z2D = 0, $eh$lpad$body122$indexZ2D = 0, $eh$lpad$body126$index67Z2D = 0, $eh$lpad$body126$indexZ2D = 0, $eh$lpad$body38$index27Z2D = 0, $eh$lpad$body38$indexZ2D = 0, $eh$lpad$body42$index32Z2D = 0, $eh$lpad$body42$indexZ2D = 0, $eh$lpad$body45$index37Z2D = 0, $eh$lpad$body45$indexZ2D = 0, $eh$lpad$body50$index17Z2D = 0, $eh$lpad$body50$indexZ2D = 0, $eh$lpad$body53$index72Z2D = 0, $eh$lpad$body53$indexZ2D = 0, $eh$lpad$body56$index77Z2D = 0, $eh$lpad$body56$indexZ2D = 0, $eh$lpad$body59$index82Z2D = 0, $eh$lpad$body59$indexZ2D = 0, $eh$lpad$body62$index87Z2D = 0;
 var $eh$lpad$body62$indexZ2D = 0, $eh$lpad$body65$index92Z2D = 0, $eh$lpad$body65$indexZ2D = 0, $eh$lpad$body68$index97Z2D = 0, $eh$lpad$body68$indexZ2D = 0, $eh$lpad$body76$index102Z2D = 0, $eh$lpad$body76$indexZ2D = 0, $eh$lpad$body94$index2Z2D = 0, $eh$lpad$body94$indexZ2D = 0, $eh$lpad$body98$index7Z2D = 0, $eh$lpad$body98$indexZ2D = 0, $httpHeader = 0, $i$01$i32 = 0, $i$01$i84 = 0, $queryString = 0, $serializer = 0, $str$i = 0, $str$i103 = 0, $str$i107 = 0, $str$i111 = 0;
 var $str$i115 = 0, $str$i119 = 0, $str$i123 = 0, $str$i91 = 0, $str$i95 = 0, $str$i99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 704|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(704|0);
 $str$i123 = sp + 688|0;
 $str$i119 = sp + 672|0;
 $str$i115 = sp + 656|0;
 $str$i111 = sp + 640|0;
 $str$i107 = sp + 624|0;
 $str$i103 = sp + 608|0;
 $str$i99 = sp + 592|0;
 $str$i95 = sp + 576|0;
 $str$i91 = sp + 560|0;
 $0 = sp + 544|0;
 $1 = sp + 528|0;
 $2 = sp + 512|0;
 $3 = sp + 496|0;
 $4 = sp + 480|0;
 $5 = sp + 464|0;
 $str$i = sp + 448|0;
 $6 = sp + 432|0;
 $7 = sp + 416|0;
 $8 = sp + 400|0;
 $9 = sp + 384|0;
 $10 = sp + 360|0;
 $queryString = sp + 344|0;
 $11 = sp + 328|0;
 $12 = sp + 312|0;
 $13 = sp + 296|0;
 $14 = sp + 280|0;
 $15 = sp + 264|0;
 $16 = sp + 248|0;
 $17 = sp + 232|0;
 $18 = sp + 216|0;
 $19 = sp + 200|0;
 $serializer = sp + 180|0;
 $httpHeader = sp + 168|0;
 $20 = sp + 152|0;
 $21 = sp + 136|0;
 $22 = sp + 120|0;
 $23 = sp + 104|0;
 $24 = sp + 88|0;
 $25 = sp + 72|0;
 $26 = sp + 56|0;
 $27 = sp + 40|0;
 $28 = sp + 24|0;
 $29 = sp;
 $30 = ((($this)) + 148|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = ((($31)) + 32|0);
 $33 = (__ZNK9ExitGames6Common6Object7getTypeEv($32)|0);
 $34 = ($33<<24>>24)==(42);
 if ($34) {
  $35 = ((($this)) + 4|0);
  $36 = HEAP32[$35>>2]|0;
  HEAP32[$10>>2] = (3820);
  $37 = ((($10)) + 4|0);
  HEAP32[$37>>2] = 41;
  $38 = ((($10)) + 8|0);
  HEAP32[$38>>2] = 41;
  __THREW__ = 0;
  $39 = (invoke_ii(775,41)|0);
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if ($41) {
   $57 = ___cxa_find_matching_catch_2()|0;
   $58 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($10);
   ___resumeException($57|0);
   // unreachable;
  }
  $42 = ((($10)) + 16|0);
  HEAP32[$42>>2] = $39;
  $43 = ((($10)) + 12|0);
  HEAP32[$43>>2] = 10;
  $44 = HEAP32[$37>>2]|0;
  $45 = ($44|0)==(0);
  if ($45) {
   $441 = 0;
  } else {
   $46 = ((($36)) + 52|0);
   $47 = HEAP8[$46>>0]|0;
   HEAP8[$39>>0] = $47;
   $48 = HEAP32[$37>>2]|0;
   $49 = ($48>>>0)>(1);
   if ($49) {
    $51 = 1;
    while(1) {
     $$pre$i = HEAP32[$42>>2]|0;
     $50 = (($$pre$i) + ($51)|0);
     $52 = (((($36)) + 52|0) + ($51)|0);
     $53 = HEAP8[$52>>0]|0;
     HEAP8[$50>>0] = $53;
     $54 = (($51) + 1)|0;
     $55 = HEAP32[$37>>2]|0;
     $56 = ($54>>>0)<($55>>>0);
     if ($56) {
      $51 = $54;
     } else {
      $441 = $55;
      break;
     }
    }
   } else {
    $441 = $48;
   }
  }
  $59 = ((($this)) + 120|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = ($60|0)==(0);
  $62 = ((($this)) + 124|0);
  if ($61) {
   $63 = HEAP32[$62>>2]|0;
   $64 = HEAP32[$38>>2]|0;
   $65 = ($63>>>0)<($64>>>0);
   if ($65) {
    label = 9;
   } else {
    $74 = $441;
   }
  } else {
   label = 9;
  }
  do {
   if ((label|0) == 9) {
    HEAP32[$59>>2] = 0;
    $66 = ((($this)) + 132|0);
    $67 = HEAP32[$66>>2]|0;
    __THREW__ = 0;
    invoke_vi(750,($67|0));
    $68 = __THREW__; __THREW__ = 0;
    $69 = $68&1;
    if (!($69)) {
     $70 = HEAP32[$38>>2]|0;
     HEAP32[$62>>2] = $70;
     __THREW__ = 0;
     $71 = (invoke_ii(775,($70|0))|0);
     $72 = __THREW__; __THREW__ = 0;
     $73 = $72&1;
     if (!($73)) {
      HEAP32[$66>>2] = $71;
      $$pre128 = HEAP32[$37>>2]|0;
      $74 = $$pre128;
      break;
     }
    }
    $92 = ___cxa_find_matching_catch_2()|0;
    $93 = tempRet0;
    HEAP32[$10>>2] = (3820);
    HEAP32[$37>>2] = 0;
    $94 = HEAP32[$42>>2]|0;
    __THREW__ = 0;
    invoke_vi(750,($94|0));
    $95 = __THREW__; __THREW__ = 0;
    $96 = $95&1;
    if ($96) {
     $97 = ___cxa_find_matching_catch_3(0|0)|0;
     $98 = tempRet0;
     __ZN9ExitGames6Common4BaseD2Ev($10);
     ___clang_call_terminate($97);
     // unreachable;
    }
    __ZN9ExitGames6Common4BaseD2Ev($10);
    $$19 = $92;$$1928 = $93;
    ___resumeException($$19|0);
    // unreachable;
   }
  } while(0);
  HEAP32[$59>>2] = $74;
  $75 = HEAP32[$43>>2]|0;
  $76 = ((($this)) + 128|0);
  HEAP32[$76>>2] = $75;
  $77 = ($74|0)==(0);
  if (!($77)) {
   $78 = ((($this)) + 132|0);
   $i$01$i32 = 0;
   while(1) {
    $79 = HEAP32[$78>>2]|0;
    $80 = (($79) + ($i$01$i32)|0);
    $81 = HEAP32[$42>>2]|0;
    $82 = (($81) + ($i$01$i32)|0);
    $83 = HEAP8[$82>>0]|0;
    HEAP8[$80>>0] = $83;
    $84 = (($i$01$i32) + 1)|0;
    $85 = HEAP32[$59>>2]|0;
    $86 = ($84>>>0)<($85>>>0);
    if ($86) {
     $i$01$i32 = $84;
    } else {
     break;
    }
   }
  }
  HEAP32[$10>>2] = (3820);
  HEAP32[$37>>2] = 0;
  $87 = HEAP32[$42>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($87|0));
  $88 = __THREW__; __THREW__ = 0;
  $89 = $88&1;
  if ($89) {
   $90 = ___cxa_find_matching_catch_3(0|0)|0;
   $91 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($10);
   ___clang_call_terminate($90);
   // unreachable;
  }
  __ZN9ExitGames6Common4BaseD2Ev($10);
  STACKTOP = sp;return;
 }
 __ZN9ExitGames6Common7JStringC2EPKw($19,9732);
 __THREW__ = 0;
 invoke_viii(916,($18|0),($19|0),(9756|0));
 $99 = __THREW__; __THREW__ = 0;
 $100 = $99&1;
 if ($100) {
  $194 = ___cxa_find_matching_catch_2()|0;
  $195 = tempRet0;
  $$7 = $194;$$716 = $195;
 } else {
  $101 = HEAP32[$30>>2]|0;
  $102 = ((($101)) + 16|0);
  __THREW__ = 0;
  invoke_viii(785,($17|0),($18|0),($102|0));
  $103 = __THREW__; __THREW__ = 0;
  $104 = $103&1;
  if ($104) {
   $196 = ___cxa_find_matching_catch_2()|0;
   $197 = tempRet0;
   $$6 = $196;$$615 = $197;
  } else {
   __THREW__ = 0;
   invoke_vii(786,($9|0),($17|0));
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   L40: do {
    if ($106) {
     $198 = ___cxa_find_matching_catch_2()|0;
     $199 = tempRet0;
     $eh$lpad$body$index22Z2D = $199;$eh$lpad$body$indexZ2D = $198;
     label = 66;
    } else {
     __THREW__ = 0;
     invoke_vii(787,($str$i91|0),0);
     $107 = __THREW__; __THREW__ = 0;
     $108 = $107&1;
     do {
      if ($108) {
       label = 29;
      } else {
       __THREW__ = 0;
       (invoke_iii(759,($str$i91|0),(9780|0))|0);
       $109 = __THREW__; __THREW__ = 0;
       $110 = $109&1;
       if (!($110)) {
        __THREW__ = 0;
        $111 = (invoke_iii(760,($9|0),($str$i91|0))|0);
        $112 = __THREW__; __THREW__ = 0;
        $113 = $112&1;
        if (!($113)) {
         __ZN9ExitGames6Common7JStringD2Ev($str$i91);
         __THREW__ = 0;
         invoke_vii(786,($16|0),($111|0));
         $116 = __THREW__; __THREW__ = 0;
         $117 = $116&1;
         if ($117) {
          label = 29;
          break;
         }
         __ZN9ExitGames6Common7JStringD2Ev($9);
         __THREW__ = 0;
         invoke_viii(785,($15|0),($16|0),(163740|0));
         $120 = __THREW__; __THREW__ = 0;
         $121 = $120&1;
         if ($121) {
          $200 = ___cxa_find_matching_catch_2()|0;
          $201 = tempRet0;
          $$4 = $200;$$413 = $201;
         } else {
          __THREW__ = 0;
          invoke_vii(786,($8|0),($15|0));
          $122 = __THREW__; __THREW__ = 0;
          $123 = $122&1;
          L52: do {
           if ($123) {
            $202 = ___cxa_find_matching_catch_2()|0;
            $203 = tempRet0;
            $eh$lpad$body38$index27Z2D = $203;$eh$lpad$body38$indexZ2D = $202;
            label = 69;
           } else {
            __THREW__ = 0;
            invoke_vii(787,($str$i95|0),0);
            $124 = __THREW__; __THREW__ = 0;
            $125 = $124&1;
            L55: do {
             if ($125) {
              label = 38;
             } else {
              __THREW__ = 0;
              (invoke_iii(759,($str$i95|0),(9844|0))|0);
              $126 = __THREW__; __THREW__ = 0;
              $127 = $126&1;
              do {
               if (!($127)) {
                __THREW__ = 0;
                $128 = (invoke_iii(760,($8|0),($str$i95|0))|0);
                $129 = __THREW__; __THREW__ = 0;
                $130 = $129&1;
                if ($130) {
                 break;
                }
                __ZN9ExitGames6Common7JStringD2Ev($str$i95);
                __THREW__ = 0;
                invoke_vii(786,($14|0),($128|0));
                $133 = __THREW__; __THREW__ = 0;
                $134 = $133&1;
                if ($134) {
                 label = 38;
                 break L55;
                }
                __ZN9ExitGames6Common7JStringD2Ev($8);
                __THREW__ = 0;
                invoke_viii(785,($13|0),($14|0),(163756|0));
                $137 = __THREW__; __THREW__ = 0;
                $138 = $137&1;
                if ($138) {
                 $204 = ___cxa_find_matching_catch_2()|0;
                 $205 = tempRet0;
                 $$2 = $204;$$211 = $205;
                } else {
                 __THREW__ = 0;
                 invoke_viii(916,($12|0),($13|0),(9888|0));
                 $139 = __THREW__; __THREW__ = 0;
                 $140 = $139&1;
                 if ($140) {
                  $206 = ___cxa_find_matching_catch_2()|0;
                  $207 = tempRet0;
                  $$1 = $206;$$110 = $207;
                 } else {
                  $141 = HEAP8[167383]|0;
                  $142 = ($141<<24>>24)!=(0);
                  __THREW__ = 0;
                  $143 = (invoke_ii(917,($142|0))|0);
                  $144 = __THREW__; __THREW__ = 0;
                  $145 = $144&1;
                  L67: do {
                   if ($145) {
                    label = 72;
                   } else {
                    __THREW__ = 0;
                    invoke_vii(786,($7|0),($12|0));
                    $146 = __THREW__; __THREW__ = 0;
                    $147 = $146&1;
                    if ($147) {
                     label = 72;
                     break;
                    }
                    __THREW__ = 0;
                    invoke_vii(787,($str$i99|0),0);
                    $148 = __THREW__; __THREW__ = 0;
                    $149 = $148&1;
                    L70: do {
                     if ($149) {
                      label = 49;
                     } else {
                      __THREW__ = 0;
                      (invoke_iii(782,($str$i99|0),($143|0))|0);
                      $150 = __THREW__; __THREW__ = 0;
                      $151 = $150&1;
                      do {
                       if (!($151)) {
                        __THREW__ = 0;
                        $152 = (invoke_iii(760,($7|0),($str$i99|0))|0);
                        $153 = __THREW__; __THREW__ = 0;
                        $154 = $153&1;
                        if ($154) {
                         break;
                        }
                        __ZN9ExitGames6Common7JStringD2Ev($str$i99);
                        __THREW__ = 0;
                        invoke_vii(786,($11|0),($152|0));
                        $157 = __THREW__; __THREW__ = 0;
                        $158 = $157&1;
                        if ($158) {
                         label = 49;
                         break L70;
                        }
                        __ZN9ExitGames6Common7JStringD2Ev($7);
                        $161 = HEAP32[$30>>2]|0;
                        $162 = ((($161)) + 72|0);
                        $163 = HEAP8[$162>>0]|0;
                        $164 = ($163<<24>>24)!=(0);
                        $165 = $164 ? 9912 : 163772;
                        __THREW__ = 0;
                        invoke_vii(786,($6|0),($11|0));
                        $166 = __THREW__; __THREW__ = 0;
                        $167 = $166&1;
                        if ($167) {
                         $210 = ___cxa_find_matching_catch_2()|0;
                         $211 = tempRet0;
                         $eh$lpad$body45$index37Z2D = $211;$eh$lpad$body45$indexZ2D = $210;
                        } else {
                         __THREW__ = 0;
                         invoke_vii(787,($str$i|0),0);
                         $168 = __THREW__; __THREW__ = 0;
                         $169 = $168&1;
                         L79: do {
                          if ($169) {
                           label = 57;
                          } else {
                           __THREW__ = 0;
                           (invoke_iii(759,($str$i|0),($165|0))|0);
                           $170 = __THREW__; __THREW__ = 0;
                           $171 = $170&1;
                           do {
                            if (!($171)) {
                             __THREW__ = 0;
                             $172 = (invoke_iii(760,($6|0),($str$i|0))|0);
                             $173 = __THREW__; __THREW__ = 0;
                             $174 = $173&1;
                             if ($174) {
                              break;
                             }
                             __ZN9ExitGames6Common7JStringD2Ev($str$i);
                             __THREW__ = 0;
                             invoke_vii(786,($queryString|0),($172|0));
                             $177 = __THREW__; __THREW__ = 0;
                             $178 = $177&1;
                             if ($178) {
                              label = 57;
                              break L79;
                             }
                             __ZN9ExitGames6Common7JStringD2Ev($6);
                             __ZN9ExitGames6Common7JStringD2Ev($11);
                             __ZN9ExitGames6Common7JStringD2Ev($12);
                             __ZN9ExitGames6Common7JStringD2Ev($13);
                             __ZN9ExitGames6Common7JStringD2Ev($14);
                             __ZN9ExitGames6Common7JStringD2Ev($15);
                             __ZN9ExitGames6Common7JStringD2Ev($16);
                             __ZN9ExitGames6Common7JStringD2Ev($17);
                             __ZN9ExitGames6Common7JStringD2Ev($18);
                             __ZN9ExitGames6Common7JStringD2Ev($19);
                             HEAP32[$serializer>>2] = (6136);
                             $181 = ((($serializer)) + 4|0);
                             __THREW__ = 0;
                             invoke_vi(918,($181|0));
                             $182 = __THREW__; __THREW__ = 0;
                             $183 = $182&1;
                             if ($183) {
                              $212 = ___cxa_find_matching_catch_2()|0;
                              $213 = tempRet0;
                              __ZN9ExitGames6Common4BaseD2Ev($serializer);
                              $$18 = $212;$$1827 = $213;
                             } else {
                              $184 = HEAP32[$30>>2]|0;
                              $185 = ((($184)) + 32|0);
                              __THREW__ = 0;
                              $186 = (invoke_ii(848,($185|0))|0);
                              $187 = __THREW__; __THREW__ = 0;
                              $188 = $187&1;
                              L88: do {
                               if ($188) {
                                label = 85;
                               } else {
                                $189 = ($186<<24>>24)==(42);
                                if (!($189)) {
                                 $190 = HEAP32[$30>>2]|0;
                                 $191 = ((($190)) + 32|0);
                                 __THREW__ = 0;
                                 (invoke_iii(919,($serializer|0),($191|0))|0);
                                 $192 = __THREW__; __THREW__ = 0;
                                 $193 = $192&1;
                                 if ($193) {
                                  label = 85;
                                  break;
                                 }
                                }
                                __THREW__ = 0;
                                invoke_vii(801,($28|0),(9936|0));
                                $216 = __THREW__; __THREW__ = 0;
                                $217 = $216&1;
                                if ($217) {
                                 label = 85;
                                 break;
                                }
                                __THREW__ = 0;
                                invoke_viii(785,($27|0),($28|0),($queryString|0));
                                $218 = __THREW__; __THREW__ = 0;
                                $219 = $218&1;
                                if ($219) {
                                 $414 = ___cxa_find_matching_catch_2()|0;
                                 $415 = tempRet0;
                                 $$15 = $414;$$1524 = $415;
                                } else {
                                 __THREW__ = 0;
                                 invoke_vii(786,($5|0),($27|0));
                                 $220 = __THREW__; __THREW__ = 0;
                                 $221 = $220&1;
                                 L97: do {
                                  if ($221) {
                                   $416 = ___cxa_find_matching_catch_2()|0;
                                   $417 = tempRet0;
                                   $eh$lpad$body53$index72Z2D = $417;$eh$lpad$body53$indexZ2D = $416;
                                   label = 171;
                                  } else {
                                   __THREW__ = 0;
                                   invoke_vii(787,($str$i103|0),0);
                                   $222 = __THREW__; __THREW__ = 0;
                                   $223 = $222&1;
                                   L100: do {
                                    if ($223) {
                                     label = 94;
                                    } else {
                                     __THREW__ = 0;
                                     (invoke_iii(759,($str$i103|0),(9968|0))|0);
                                     $224 = __THREW__; __THREW__ = 0;
                                     $225 = $224&1;
                                     do {
                                      if (!($225)) {
                                       __THREW__ = 0;
                                       $226 = (invoke_iii(760,($5|0),($str$i103|0))|0);
                                       $227 = __THREW__; __THREW__ = 0;
                                       $228 = $227&1;
                                       if ($228) {
                                        break;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($str$i103);
                                       __THREW__ = 0;
                                       invoke_vii(786,($26|0),($226|0));
                                       $231 = __THREW__; __THREW__ = 0;
                                       $232 = $231&1;
                                       if ($232) {
                                        label = 94;
                                        break L100;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($5);
                                       __THREW__ = 0;
                                       invoke_vii(786,($4|0),($26|0));
                                       $235 = __THREW__; __THREW__ = 0;
                                       $236 = $235&1;
                                       L106: do {
                                        if ($236) {
                                         $418 = ___cxa_find_matching_catch_2()|0;
                                         $419 = tempRet0;
                                         $eh$lpad$body56$index77Z2D = $419;$eh$lpad$body56$indexZ2D = $418;
                                         label = 173;
                                        } else {
                                         __THREW__ = 0;
                                         invoke_vii(787,($str$i107|0),0);
                                         $237 = __THREW__; __THREW__ = 0;
                                         $238 = $237&1;
                                         L109: do {
                                          if ($238) {
                                           label = 102;
                                          } else {
                                           __THREW__ = 0;
                                           (invoke_iii(759,($str$i107|0),(10016|0))|0);
                                           $239 = __THREW__; __THREW__ = 0;
                                           $240 = $239&1;
                                           do {
                                            if (!($240)) {
                                             __THREW__ = 0;
                                             $241 = (invoke_iii(760,($4|0),($str$i107|0))|0);
                                             $242 = __THREW__; __THREW__ = 0;
                                             $243 = $242&1;
                                             if ($243) {
                                              break;
                                             }
                                             __ZN9ExitGames6Common7JStringD2Ev($str$i107);
                                             __THREW__ = 0;
                                             invoke_vii(786,($25|0),($241|0));
                                             $246 = __THREW__; __THREW__ = 0;
                                             $247 = $246&1;
                                             if ($247) {
                                              label = 102;
                                              break L109;
                                             }
                                             __ZN9ExitGames6Common7JStringD2Ev($4);
                                             $250 = HEAP32[$30>>2]|0;
                                             __THREW__ = 0;
                                             invoke_viii(785,($24|0),($25|0),($250|0));
                                             $251 = __THREW__; __THREW__ = 0;
                                             $252 = $251&1;
                                             if ($252) {
                                              $420 = ___cxa_find_matching_catch_2()|0;
                                              $421 = tempRet0;
                                              $$12 = $420;$$1221 = $421;
                                             } else {
                                              __THREW__ = 0;
                                              invoke_vii(786,($3|0),($24|0));
                                              $253 = __THREW__; __THREW__ = 0;
                                              $254 = $253&1;
                                              L118: do {
                                               if ($254) {
                                                $422 = ___cxa_find_matching_catch_2()|0;
                                                $423 = tempRet0;
                                                $eh$lpad$body59$index82Z2D = $423;$eh$lpad$body59$indexZ2D = $422;
                                                label = 176;
                                               } else {
                                                __THREW__ = 0;
                                                invoke_vii(787,($str$i111|0),0);
                                                $255 = __THREW__; __THREW__ = 0;
                                                $256 = $255&1;
                                                L121: do {
                                                 if ($256) {
                                                  label = 111;
                                                 } else {
                                                  __THREW__ = 0;
                                                  (invoke_iii(759,($str$i111|0),(10044|0))|0);
                                                  $257 = __THREW__; __THREW__ = 0;
                                                  $258 = $257&1;
                                                  do {
                                                   if (!($258)) {
                                                    __THREW__ = 0;
                                                    $259 = (invoke_iii(760,($3|0),($str$i111|0))|0);
                                                    $260 = __THREW__; __THREW__ = 0;
                                                    $261 = $260&1;
                                                    if ($261) {
                                                     break;
                                                    }
                                                    __ZN9ExitGames6Common7JStringD2Ev($str$i111);
                                                    __THREW__ = 0;
                                                    invoke_vii(786,($23|0),($259|0));
                                                    $264 = __THREW__; __THREW__ = 0;
                                                    $265 = $264&1;
                                                    if ($265) {
                                                     label = 111;
                                                     break L121;
                                                    }
                                                    __ZN9ExitGames6Common7JStringD2Ev($3);
                                                    __THREW__ = 0;
                                                    invoke_vii(786,($2|0),($23|0));
                                                    $268 = __THREW__; __THREW__ = 0;
                                                    $269 = $268&1;
                                                    L127: do {
                                                     if ($269) {
                                                      $424 = ___cxa_find_matching_catch_2()|0;
                                                      $425 = tempRet0;
                                                      $eh$lpad$body62$index87Z2D = $425;$eh$lpad$body62$indexZ2D = $424;
                                                      label = 178;
                                                     } else {
                                                      __THREW__ = 0;
                                                      invoke_vii(787,($str$i115|0),0);
                                                      $270 = __THREW__; __THREW__ = 0;
                                                      $271 = $270&1;
                                                      L130: do {
                                                       if ($271) {
                                                        label = 119;
                                                       } else {
                                                        __THREW__ = 0;
                                                        (invoke_iii(759,($str$i115|0),(10056|0))|0);
                                                        $272 = __THREW__; __THREW__ = 0;
                                                        $273 = $272&1;
                                                        do {
                                                         if (!($273)) {
                                                          __THREW__ = 0;
                                                          $274 = (invoke_iii(760,($2|0),($str$i115|0))|0);
                                                          $275 = __THREW__; __THREW__ = 0;
                                                          $276 = $275&1;
                                                          if ($276) {
                                                           break;
                                                          }
                                                          __ZN9ExitGames6Common7JStringD2Ev($str$i115);
                                                          __THREW__ = 0;
                                                          invoke_vii(786,($22|0),($274|0));
                                                          $279 = __THREW__; __THREW__ = 0;
                                                          $280 = $279&1;
                                                          if ($280) {
                                                           label = 119;
                                                           break L130;
                                                          }
                                                          __ZN9ExitGames6Common7JStringD2Ev($2);
                                                          __THREW__ = 0;
                                                          $283 = (invoke_ii(920,($serializer|0))|0);
                                                          $284 = __THREW__; __THREW__ = 0;
                                                          $285 = $284&1;
                                                          L136: do {
                                                           if ($285) {
                                                            label = 179;
                                                           } else {
                                                            __THREW__ = 0;
                                                            invoke_vii(786,($1|0),($22|0));
                                                            $286 = __THREW__; __THREW__ = 0;
                                                            $287 = $286&1;
                                                            if ($287) {
                                                             label = 179;
                                                             break;
                                                            }
                                                            __THREW__ = 0;
                                                            invoke_vii(787,($str$i119|0),0);
                                                            $288 = __THREW__; __THREW__ = 0;
                                                            $289 = $288&1;
                                                            L139: do {
                                                             if ($289) {
                                                              label = 128;
                                                             } else {
                                                              __THREW__ = 0;
                                                              (invoke_iii(833,($str$i119|0),($283|0))|0);
                                                              $290 = __THREW__; __THREW__ = 0;
                                                              $291 = $290&1;
                                                              do {
                                                               if (!($291)) {
                                                                __THREW__ = 0;
                                                                $292 = (invoke_iii(760,($1|0),($str$i119|0))|0);
                                                                $293 = __THREW__; __THREW__ = 0;
                                                                $294 = $293&1;
                                                                if ($294) {
                                                                 break;
                                                                }
                                                                __ZN9ExitGames6Common7JStringD2Ev($str$i119);
                                                                __THREW__ = 0;
                                                                invoke_vii(786,($21|0),($292|0));
                                                                $297 = __THREW__; __THREW__ = 0;
                                                                $298 = $297&1;
                                                                if ($298) {
                                                                 label = 128;
                                                                 break L139;
                                                                }
                                                                __ZN9ExitGames6Common7JStringD2Ev($1);
                                                                __THREW__ = 0;
                                                                invoke_vii(786,($0|0),($21|0));
                                                                $301 = __THREW__; __THREW__ = 0;
                                                                $302 = $301&1;
                                                                L145: do {
                                                                 if ($302) {
                                                                  $428 = ___cxa_find_matching_catch_2()|0;
                                                                  $429 = tempRet0;
                                                                  $eh$lpad$body68$index97Z2D = $429;$eh$lpad$body68$indexZ2D = $428;
                                                                  label = 182;
                                                                 } else {
                                                                  __THREW__ = 0;
                                                                  invoke_vii(787,($str$i123|0),0);
                                                                  $303 = __THREW__; __THREW__ = 0;
                                                                  $304 = $303&1;
                                                                  L148: do {
                                                                   if ($304) {
                                                                    label = 136;
                                                                   } else {
                                                                    __THREW__ = 0;
                                                                    (invoke_iii(759,($str$i123|0),(10124|0))|0);
                                                                    $305 = __THREW__; __THREW__ = 0;
                                                                    $306 = $305&1;
                                                                    do {
                                                                     if (!($306)) {
                                                                      __THREW__ = 0;
                                                                      $307 = (invoke_iii(760,($0|0),($str$i123|0))|0);
                                                                      $308 = __THREW__; __THREW__ = 0;
                                                                      $309 = $308&1;
                                                                      if ($309) {
                                                                       break;
                                                                      }
                                                                      __ZN9ExitGames6Common7JStringD2Ev($str$i123);
                                                                      __THREW__ = 0;
                                                                      invoke_vii(786,($20|0),($307|0));
                                                                      $312 = __THREW__; __THREW__ = 0;
                                                                      $313 = $312&1;
                                                                      if ($313) {
                                                                       label = 136;
                                                                       break L148;
                                                                      }
                                                                      __ZN9ExitGames6Common7JStringD2Ev($0);
                                                                      __THREW__ = 0;
                                                                      invoke_vii(921,($httpHeader|0),($20|0));
                                                                      $316 = __THREW__; __THREW__ = 0;
                                                                      $317 = $316&1;
                                                                      if ($317) {
                                                                       $430 = ___cxa_find_matching_catch_2()|0;
                                                                       $431 = tempRet0;
                                                                       __ZN9ExitGames6Common7JStringD2Ev($20);
                                                                       $$8 = $430;$$817 = $431;
                                                                       break L145;
                                                                      }
                                                                      __ZN9ExitGames6Common7JStringD2Ev($20);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($21);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($22);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($23);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($24);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($25);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($26);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($27);
                                                                      __ZN9ExitGames6Common7JStringD2Ev($28);
                                                                      __THREW__ = 0;
                                                                      $318 = (invoke_ii(231,($httpHeader|0))|0);
                                                                      $319 = __THREW__; __THREW__ = 0;
                                                                      $320 = $319&1;
                                                                      L157: do {
                                                                       if ($320) {
                                                                        label = 192;
                                                                       } else {
                                                                        __THREW__ = 0;
                                                                        $321 = (invoke_ii(920,($serializer|0))|0);
                                                                        $322 = __THREW__; __THREW__ = 0;
                                                                        $323 = $322&1;
                                                                        if ($323) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        $324 = (($321) + ($318))|0;
                                                                        $325 = (($324) + 8)|0;
                                                                        __THREW__ = 0;
                                                                        $326 = (invoke_ii(775,($325|0))|0);
                                                                        $327 = __THREW__; __THREW__ = 0;
                                                                        $328 = $327&1;
                                                                        if ($328) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        $329 = ((($326)) + 8|0);
                                                                        $330 = $326;
                                                                        $331 = $330;
                                                                        HEAP32[$331>>2] = $324;
                                                                        $332 = (($330) + 4)|0;
                                                                        $333 = $332;
                                                                        HEAP32[$333>>2] = 0;
                                                                        $334 = ($324|0)==(0);
                                                                        if (!($334)) {
                                                                         _memset(($329|0),0,($324|0))|0;
                                                                        }
                                                                        __THREW__ = 0;
                                                                        $335 = (invoke_ii(826,($httpHeader|0))|0);
                                                                        $336 = __THREW__; __THREW__ = 0;
                                                                        $337 = $336&1;
                                                                        if ($337) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        __THREW__ = 0;
                                                                        $338 = (invoke_ii(231,($httpHeader|0))|0);
                                                                        $339 = __THREW__; __THREW__ = 0;
                                                                        $340 = $339&1;
                                                                        if ($340) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        _memcpy(($329|0),($335|0),($338|0))|0;
                                                                        __THREW__ = 0;
                                                                        $341 = (invoke_ii(231,($httpHeader|0))|0);
                                                                        $342 = __THREW__; __THREW__ = 0;
                                                                        $343 = $342&1;
                                                                        if ($343) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        $$sum = (($341) + 8)|0;
                                                                        $344 = (($326) + ($$sum)|0);
                                                                        __THREW__ = 0;
                                                                        $345 = (invoke_ii(922,($serializer|0))|0);
                                                                        $346 = __THREW__; __THREW__ = 0;
                                                                        $347 = $346&1;
                                                                        if ($347) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        __THREW__ = 0;
                                                                        $348 = (invoke_ii(920,($serializer|0))|0);
                                                                        $349 = __THREW__; __THREW__ = 0;
                                                                        $350 = $349&1;
                                                                        if ($350) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        _memcpy(($344|0),($345|0),($348|0))|0;
                                                                        __THREW__ = 0;
                                                                        $351 = (invoke_ii(231,($httpHeader|0))|0);
                                                                        $352 = __THREW__; __THREW__ = 0;
                                                                        $353 = $352&1;
                                                                        if ($353) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        __THREW__ = 0;
                                                                        $354 = (invoke_ii(920,($serializer|0))|0);
                                                                        $355 = __THREW__; __THREW__ = 0;
                                                                        $356 = $355&1;
                                                                        if ($356) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        $357 = (($354) + ($351))|0;
                                                                        HEAP32[$29>>2] = (3820);
                                                                        $358 = ((($29)) + 4|0);
                                                                        HEAP32[$358>>2] = $357;
                                                                        $359 = ((($29)) + 8|0);
                                                                        HEAP32[$359>>2] = $357;
                                                                        __THREW__ = 0;
                                                                        $360 = (invoke_ii(775,($357|0))|0);
                                                                        $361 = __THREW__; __THREW__ = 0;
                                                                        $362 = $361&1;
                                                                        if ($362) {
                                                                         $377 = ___cxa_find_matching_catch_2()|0;
                                                                         $378 = tempRet0;
                                                                         __ZN9ExitGames6Common4BaseD2Ev($29);
                                                                         $eh$lpad$body76$index102Z2D = $378;$eh$lpad$body76$indexZ2D = $377;
                                                                         label = 193;
                                                                         break;
                                                                        }
                                                                        $363 = ((($29)) + 16|0);
                                                                        HEAP32[$363>>2] = $360;
                                                                        $364 = ((($29)) + 12|0);
                                                                        HEAP32[$364>>2] = 10;
                                                                        $365 = HEAP32[$358>>2]|0;
                                                                        $366 = ($365|0)==(0);
                                                                        do {
                                                                         if ($366) {
                                                                          $442 = 0;
                                                                         } else {
                                                                          $367 = HEAP8[$329>>0]|0;
                                                                          HEAP8[$360>>0] = $367;
                                                                          $368 = HEAP32[$358>>2]|0;
                                                                          $369 = ($368>>>0)>(1);
                                                                          if ($369) {
                                                                           $371 = 1;
                                                                          } else {
                                                                           $442 = $368;
                                                                           break;
                                                                          }
                                                                          while(1) {
                                                                           $$pre$i73 = HEAP32[$363>>2]|0;
                                                                           $370 = (($$pre$i73) + ($371)|0);
                                                                           $$sum127 = (($371) + 8)|0;
                                                                           $372 = (($326) + ($$sum127)|0);
                                                                           $373 = HEAP8[$372>>0]|0;
                                                                           HEAP8[$370>>0] = $373;
                                                                           $374 = (($371) + 1)|0;
                                                                           $375 = HEAP32[$358>>2]|0;
                                                                           $376 = ($374>>>0)<($375>>>0);
                                                                           if ($376) {
                                                                            $371 = $374;
                                                                           } else {
                                                                            $442 = $375;
                                                                            break;
                                                                           }
                                                                          }
                                                                         }
                                                                        } while(0);
                                                                        $379 = ((($this)) + 120|0);
                                                                        $380 = HEAP32[$379>>2]|0;
                                                                        $381 = ($380|0)==(0);
                                                                        $382 = ((($this)) + 124|0);
                                                                        if ($381) {
                                                                         $383 = HEAP32[$382>>2]|0;
                                                                         $384 = HEAP32[$359>>2]|0;
                                                                         $385 = ($383>>>0)<($384>>>0);
                                                                         if ($385) {
                                                                          label = 158;
                                                                         } else {
                                                                          $394 = $442;
                                                                         }
                                                                        } else {
                                                                         label = 158;
                                                                        }
                                                                        L181: do {
                                                                         if ((label|0) == 158) {
                                                                          HEAP32[$379>>2] = 0;
                                                                          $386 = ((($this)) + 132|0);
                                                                          $387 = HEAP32[$386>>2]|0;
                                                                          __THREW__ = 0;
                                                                          invoke_vi(750,($387|0));
                                                                          $388 = __THREW__; __THREW__ = 0;
                                                                          $389 = $388&1;
                                                                          do {
                                                                           if (!($389)) {
                                                                            $390 = HEAP32[$359>>2]|0;
                                                                            HEAP32[$382>>2] = $390;
                                                                            __THREW__ = 0;
                                                                            $391 = (invoke_ii(775,($390|0))|0);
                                                                            $392 = __THREW__; __THREW__ = 0;
                                                                            $393 = $392&1;
                                                                            if ($393) {
                                                                             break;
                                                                            }
                                                                            HEAP32[$386>>2] = $391;
                                                                            $$pre = HEAP32[$358>>2]|0;
                                                                            $394 = $$pre;
                                                                            break L181;
                                                                           }
                                                                          } while(0);
                                                                          $434 = ___cxa_find_matching_catch_2()|0;
                                                                          $435 = tempRet0;
                                                                          HEAP32[$29>>2] = (3820);
                                                                          HEAP32[$358>>2] = 0;
                                                                          $436 = HEAP32[$363>>2]|0;
                                                                          __THREW__ = 0;
                                                                          invoke_vi(750,($436|0));
                                                                          $437 = __THREW__; __THREW__ = 0;
                                                                          $438 = $437&1;
                                                                          if ($438) {
                                                                           $439 = ___cxa_find_matching_catch_3(0|0)|0;
                                                                           $440 = tempRet0;
                                                                           __ZN9ExitGames6Common4BaseD2Ev($29);
                                                                           ___clang_call_terminate($439);
                                                                           // unreachable;
                                                                          } else {
                                                                           __ZN9ExitGames6Common4BaseD2Ev($29);
                                                                           $$16 = $434;$$1625 = $435;
                                                                           break L157;
                                                                          }
                                                                         }
                                                                        } while(0);
                                                                        HEAP32[$379>>2] = $394;
                                                                        $395 = HEAP32[$364>>2]|0;
                                                                        $396 = ((($this)) + 128|0);
                                                                        HEAP32[$396>>2] = $395;
                                                                        $397 = ($394|0)==(0);
                                                                        if (!($397)) {
                                                                         $398 = ((($this)) + 132|0);
                                                                         $i$01$i84 = 0;
                                                                         while(1) {
                                                                          $399 = HEAP32[$398>>2]|0;
                                                                          $400 = (($399) + ($i$01$i84)|0);
                                                                          $401 = HEAP32[$363>>2]|0;
                                                                          $402 = (($401) + ($i$01$i84)|0);
                                                                          $403 = HEAP8[$402>>0]|0;
                                                                          HEAP8[$400>>0] = $403;
                                                                          $404 = (($i$01$i84) + 1)|0;
                                                                          $405 = HEAP32[$379>>2]|0;
                                                                          $406 = ($404>>>0)<($405>>>0);
                                                                          if ($406) {
                                                                           $i$01$i84 = $404;
                                                                          } else {
                                                                           break;
                                                                          }
                                                                         }
                                                                        }
                                                                        HEAP32[$29>>2] = (3820);
                                                                        HEAP32[$358>>2] = 0;
                                                                        $407 = HEAP32[$363>>2]|0;
                                                                        __THREW__ = 0;
                                                                        invoke_vi(750,($407|0));
                                                                        $408 = __THREW__; __THREW__ = 0;
                                                                        $409 = $408&1;
                                                                        if ($409) {
                                                                         $410 = ___cxa_find_matching_catch_3(0|0)|0;
                                                                         $411 = tempRet0;
                                                                         __ZN9ExitGames6Common4BaseD2Ev($29);
                                                                         ___clang_call_terminate($410);
                                                                         // unreachable;
                                                                        }
                                                                        __ZN9ExitGames6Common4BaseD2Ev($29);
                                                                        __THREW__ = 0;
                                                                        invoke_vi(750,($326|0));
                                                                        $412 = __THREW__; __THREW__ = 0;
                                                                        $413 = $412&1;
                                                                        if ($413) {
                                                                         label = 192;
                                                                         break;
                                                                        }
                                                                        __ZN9ExitGames6Common10UTF8StringD2Ev($httpHeader);
                                                                        HEAP32[$serializer>>2] = (6136);
                                                                        __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($181);
                                                                        __ZN9ExitGames6Common4BaseD2Ev($serializer);
                                                                        __ZN9ExitGames6Common7JStringD2Ev($queryString);
                                                                        STACKTOP = sp;return;
                                                                       }
                                                                      } while(0);
                                                                      if ((label|0) == 192) {
                                                                       $432 = ___cxa_find_matching_catch_2()|0;
                                                                       $433 = tempRet0;
                                                                       $eh$lpad$body76$index102Z2D = $433;$eh$lpad$body76$indexZ2D = $432;
                                                                       label = 193;
                                                                      }
                                                                      if ((label|0) == 193) {
                                                                       $$16 = $eh$lpad$body76$indexZ2D;$$1625 = $eh$lpad$body76$index102Z2D;
                                                                      }
                                                                      __ZN9ExitGames6Common10UTF8StringD2Ev($httpHeader);
                                                                      $$17 = $$16;$$1726 = $$1625;
                                                                      break L88;
                                                                     }
                                                                    } while(0);
                                                                    $310 = ___cxa_find_matching_catch_2()|0;
                                                                    $311 = tempRet0;
                                                                    __ZN9ExitGames6Common7JStringD2Ev($str$i123);
                                                                    $eh$lpad$body126$index67Z2D = $311;$eh$lpad$body126$indexZ2D = $310;
                                                                   }
                                                                  } while(0);
                                                                  if ((label|0) == 136) {
                                                                   $314 = ___cxa_find_matching_catch_2()|0;
                                                                   $315 = tempRet0;
                                                                   $eh$lpad$body126$index67Z2D = $315;$eh$lpad$body126$indexZ2D = $314;
                                                                  }
                                                                  __ZN9ExitGames6Common7JStringD2Ev($0);
                                                                  $eh$lpad$body68$index97Z2D = $eh$lpad$body126$index67Z2D;$eh$lpad$body68$indexZ2D = $eh$lpad$body126$indexZ2D;
                                                                  label = 182;
                                                                 }
                                                                } while(0);
                                                                if ((label|0) == 182) {
                                                                 $$8 = $eh$lpad$body68$indexZ2D;$$817 = $eh$lpad$body68$index97Z2D;
                                                                }
                                                                __ZN9ExitGames6Common7JStringD2Ev($21);
                                                                $$9 = $$8;$$918 = $$817;
                                                                break L136;
                                                               }
                                                              } while(0);
                                                              $295 = ___cxa_find_matching_catch_2()|0;
                                                              $296 = tempRet0;
                                                              __ZN9ExitGames6Common7JStringD2Ev($str$i119);
                                                              $eh$lpad$body122$index62Z2D = $296;$eh$lpad$body122$indexZ2D = $295;
                                                             }
                                                            } while(0);
                                                            if ((label|0) == 128) {
                                                             $299 = ___cxa_find_matching_catch_2()|0;
                                                             $300 = tempRet0;
                                                             $eh$lpad$body122$index62Z2D = $300;$eh$lpad$body122$indexZ2D = $299;
                                                            }
                                                            __ZN9ExitGames6Common7JStringD2Ev($1);
                                                            $eh$lpad$body65$index92Z2D = $eh$lpad$body122$index62Z2D;$eh$lpad$body65$indexZ2D = $eh$lpad$body122$indexZ2D;
                                                            label = 180;
                                                           }
                                                          } while(0);
                                                          if ((label|0) == 179) {
                                                           $426 = ___cxa_find_matching_catch_2()|0;
                                                           $427 = tempRet0;
                                                           $eh$lpad$body65$index92Z2D = $427;$eh$lpad$body65$indexZ2D = $426;
                                                           label = 180;
                                                          }
                                                          if ((label|0) == 180) {
                                                           $$9 = $eh$lpad$body65$indexZ2D;$$918 = $eh$lpad$body65$index92Z2D;
                                                          }
                                                          __ZN9ExitGames6Common7JStringD2Ev($22);
                                                          $$10 = $$9;$$1019 = $$918;
                                                          break L127;
                                                         }
                                                        } while(0);
                                                        $277 = ___cxa_find_matching_catch_2()|0;
                                                        $278 = tempRet0;
                                                        __ZN9ExitGames6Common7JStringD2Ev($str$i115);
                                                        $eh$lpad$body118$index57Z2D = $278;$eh$lpad$body118$indexZ2D = $277;
                                                       }
                                                      } while(0);
                                                      if ((label|0) == 119) {
                                                       $281 = ___cxa_find_matching_catch_2()|0;
                                                       $282 = tempRet0;
                                                       $eh$lpad$body118$index57Z2D = $282;$eh$lpad$body118$indexZ2D = $281;
                                                      }
                                                      __ZN9ExitGames6Common7JStringD2Ev($2);
                                                      $eh$lpad$body62$index87Z2D = $eh$lpad$body118$index57Z2D;$eh$lpad$body62$indexZ2D = $eh$lpad$body118$indexZ2D;
                                                      label = 178;
                                                     }
                                                    } while(0);
                                                    if ((label|0) == 178) {
                                                     $$10 = $eh$lpad$body62$indexZ2D;$$1019 = $eh$lpad$body62$index87Z2D;
                                                    }
                                                    __ZN9ExitGames6Common7JStringD2Ev($23);
                                                    $$11 = $$10;$$1120 = $$1019;
                                                    break L118;
                                                   }
                                                  } while(0);
                                                  $262 = ___cxa_find_matching_catch_2()|0;
                                                  $263 = tempRet0;
                                                  __ZN9ExitGames6Common7JStringD2Ev($str$i111);
                                                  $eh$lpad$body114$index52Z2D = $263;$eh$lpad$body114$indexZ2D = $262;
                                                 }
                                                } while(0);
                                                if ((label|0) == 111) {
                                                 $266 = ___cxa_find_matching_catch_2()|0;
                                                 $267 = tempRet0;
                                                 $eh$lpad$body114$index52Z2D = $267;$eh$lpad$body114$indexZ2D = $266;
                                                }
                                                __ZN9ExitGames6Common7JStringD2Ev($3);
                                                $eh$lpad$body59$index82Z2D = $eh$lpad$body114$index52Z2D;$eh$lpad$body59$indexZ2D = $eh$lpad$body114$indexZ2D;
                                                label = 176;
                                               }
                                              } while(0);
                                              if ((label|0) == 176) {
                                               $$11 = $eh$lpad$body59$indexZ2D;$$1120 = $eh$lpad$body59$index82Z2D;
                                              }
                                              __ZN9ExitGames6Common7JStringD2Ev($24);
                                              $$12 = $$11;$$1221 = $$1120;
                                             }
                                             __ZN9ExitGames6Common7JStringD2Ev($25);
                                             $$13 = $$12;$$1322 = $$1221;
                                             break L106;
                                            }
                                           } while(0);
                                           $244 = ___cxa_find_matching_catch_2()|0;
                                           $245 = tempRet0;
                                           __ZN9ExitGames6Common7JStringD2Ev($str$i107);
                                           $eh$lpad$body110$index47Z2D = $245;$eh$lpad$body110$indexZ2D = $244;
                                          }
                                         } while(0);
                                         if ((label|0) == 102) {
                                          $248 = ___cxa_find_matching_catch_2()|0;
                                          $249 = tempRet0;
                                          $eh$lpad$body110$index47Z2D = $249;$eh$lpad$body110$indexZ2D = $248;
                                         }
                                         __ZN9ExitGames6Common7JStringD2Ev($4);
                                         $eh$lpad$body56$index77Z2D = $eh$lpad$body110$index47Z2D;$eh$lpad$body56$indexZ2D = $eh$lpad$body110$indexZ2D;
                                         label = 173;
                                        }
                                       } while(0);
                                       if ((label|0) == 173) {
                                        $$13 = $eh$lpad$body56$indexZ2D;$$1322 = $eh$lpad$body56$index77Z2D;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($26);
                                       $$14 = $$13;$$1423 = $$1322;
                                       break L97;
                                      }
                                     } while(0);
                                     $229 = ___cxa_find_matching_catch_2()|0;
                                     $230 = tempRet0;
                                     __ZN9ExitGames6Common7JStringD2Ev($str$i103);
                                     $eh$lpad$body106$index42Z2D = $230;$eh$lpad$body106$indexZ2D = $229;
                                    }
                                   } while(0);
                                   if ((label|0) == 94) {
                                    $233 = ___cxa_find_matching_catch_2()|0;
                                    $234 = tempRet0;
                                    $eh$lpad$body106$index42Z2D = $234;$eh$lpad$body106$indexZ2D = $233;
                                   }
                                   __ZN9ExitGames6Common7JStringD2Ev($5);
                                   $eh$lpad$body53$index72Z2D = $eh$lpad$body106$index42Z2D;$eh$lpad$body53$indexZ2D = $eh$lpad$body106$indexZ2D;
                                   label = 171;
                                  }
                                 } while(0);
                                 if ((label|0) == 171) {
                                  $$14 = $eh$lpad$body53$indexZ2D;$$1423 = $eh$lpad$body53$index72Z2D;
                                 }
                                 __ZN9ExitGames6Common7JStringD2Ev($27);
                                 $$15 = $$14;$$1524 = $$1423;
                                }
                                __ZN9ExitGames6Common7JStringD2Ev($28);
                                $$17 = $$15;$$1726 = $$1524;
                               }
                              } while(0);
                              if ((label|0) == 85) {
                               $214 = ___cxa_find_matching_catch_2()|0;
                               $215 = tempRet0;
                               $$17 = $214;$$1726 = $215;
                              }
                              HEAP32[$serializer>>2] = (6136);
                              __ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev($181);
                              __ZN9ExitGames6Common4BaseD2Ev($serializer);
                              $$18 = $$17;$$1827 = $$1726;
                             }
                             __ZN9ExitGames6Common7JStringD2Ev($queryString);
                             $$19 = $$18;$$1928 = $$1827;
                             ___resumeException($$19|0);
                             // unreachable;
                            }
                           } while(0);
                           $175 = ___cxa_find_matching_catch_2()|0;
                           $176 = tempRet0;
                           __ZN9ExitGames6Common7JStringD2Ev($str$i);
                           $eh$lpad$body50$index17Z2D = $176;$eh$lpad$body50$indexZ2D = $175;
                          }
                         } while(0);
                         if ((label|0) == 57) {
                          $179 = ___cxa_find_matching_catch_2()|0;
                          $180 = tempRet0;
                          $eh$lpad$body50$index17Z2D = $180;$eh$lpad$body50$indexZ2D = $179;
                         }
                         __ZN9ExitGames6Common7JStringD2Ev($6);
                         $eh$lpad$body45$index37Z2D = $eh$lpad$body50$index17Z2D;$eh$lpad$body45$indexZ2D = $eh$lpad$body50$indexZ2D;
                        }
                        __ZN9ExitGames6Common7JStringD2Ev($11);
                        $$0 = $eh$lpad$body45$indexZ2D;$$01 = $eh$lpad$body45$index37Z2D;
                        break L67;
                       }
                      } while(0);
                      $155 = ___cxa_find_matching_catch_2()|0;
                      $156 = tempRet0;
                      __ZN9ExitGames6Common7JStringD2Ev($str$i99);
                      $eh$lpad$body102$index12Z2D = $156;$eh$lpad$body102$indexZ2D = $155;
                     }
                    } while(0);
                    if ((label|0) == 49) {
                     $159 = ___cxa_find_matching_catch_2()|0;
                     $160 = tempRet0;
                     $eh$lpad$body102$index12Z2D = $160;$eh$lpad$body102$indexZ2D = $159;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($7);
                    $eh$lpad$body42$index32Z2D = $eh$lpad$body102$index12Z2D;$eh$lpad$body42$indexZ2D = $eh$lpad$body102$indexZ2D;
                    label = 73;
                   }
                  } while(0);
                  if ((label|0) == 72) {
                   $208 = ___cxa_find_matching_catch_2()|0;
                   $209 = tempRet0;
                   $eh$lpad$body42$index32Z2D = $209;$eh$lpad$body42$indexZ2D = $208;
                   label = 73;
                  }
                  if ((label|0) == 73) {
                   $$0 = $eh$lpad$body42$indexZ2D;$$01 = $eh$lpad$body42$index32Z2D;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($12);
                  $$1 = $$0;$$110 = $$01;
                 }
                 __ZN9ExitGames6Common7JStringD2Ev($13);
                 $$2 = $$1;$$211 = $$110;
                }
                __ZN9ExitGames6Common7JStringD2Ev($14);
                $$3 = $$2;$$312 = $$211;
                break L52;
               }
              } while(0);
              $131 = ___cxa_find_matching_catch_2()|0;
              $132 = tempRet0;
              __ZN9ExitGames6Common7JStringD2Ev($str$i95);
              $eh$lpad$body98$index7Z2D = $132;$eh$lpad$body98$indexZ2D = $131;
             }
            } while(0);
            if ((label|0) == 38) {
             $135 = ___cxa_find_matching_catch_2()|0;
             $136 = tempRet0;
             $eh$lpad$body98$index7Z2D = $136;$eh$lpad$body98$indexZ2D = $135;
            }
            __ZN9ExitGames6Common7JStringD2Ev($8);
            $eh$lpad$body38$index27Z2D = $eh$lpad$body98$index7Z2D;$eh$lpad$body38$indexZ2D = $eh$lpad$body98$indexZ2D;
            label = 69;
           }
          } while(0);
          if ((label|0) == 69) {
           $$3 = $eh$lpad$body38$indexZ2D;$$312 = $eh$lpad$body38$index27Z2D;
          }
          __ZN9ExitGames6Common7JStringD2Ev($15);
          $$4 = $$3;$$413 = $$312;
         }
         __ZN9ExitGames6Common7JStringD2Ev($16);
         $$5 = $$4;$$514 = $$413;
         break L40;
        }
       }
       $114 = ___cxa_find_matching_catch_2()|0;
       $115 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($str$i91);
       $eh$lpad$body94$index2Z2D = $115;$eh$lpad$body94$indexZ2D = $114;
      }
     } while(0);
     if ((label|0) == 29) {
      $118 = ___cxa_find_matching_catch_2()|0;
      $119 = tempRet0;
      $eh$lpad$body94$index2Z2D = $119;$eh$lpad$body94$indexZ2D = $118;
     }
     __ZN9ExitGames6Common7JStringD2Ev($9);
     $eh$lpad$body$index22Z2D = $eh$lpad$body94$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body94$indexZ2D;
     label = 66;
    }
   } while(0);
   if ((label|0) == 66) {
    $$5 = $eh$lpad$body$indexZ2D;$$514 = $eh$lpad$body$index22Z2D;
   }
   __ZN9ExitGames6Common7JStringD2Ev($17);
   $$6 = $$5;$$615 = $$514;
  }
  __ZN9ExitGames6Common7JStringD2Ev($18);
  $$7 = $$6;$$716 = $$615;
 }
 __ZN9ExitGames6Common7JStringD2Ev($19);
 $$19 = $$7;$$1928 = $$716;
 ___resumeException($$19|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplIA6_wEENS0_7JStringERKS3_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($Rsh|0))|0);
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($5|0));
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 $eh$lpad$body$index2Z2D = $13;$eh$lpad$body$indexZ2D = $12;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Common10Serializer4pushINS0_6ObjectEEEbRKT_($this,$data) {
 $this = $this|0;
 $data = $data|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $forceCompilationToFailForUnsupportedTypes = 0, $obj = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $forceCompilationToFailForUnsupportedTypes = sp + 80|0;
 $obj = sp + 40|0;
 $0 = sp;
 __ZN9ExitGames6Common6ObjectC2Ev($forceCompilationToFailForUnsupportedTypes);
 __THREW__ = 0;
 invoke_vii(744,($0|0),($data|0));
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if ($2) {
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   $$0 = $9;$$01 = $10;
  } else {
   __THREW__ = 0;
   invoke_vii(746,($obj|0),($0|0));
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if ($4) {
    $11 = ___cxa_find_matching_catch_2()|0;
    $12 = tempRet0;
    __ZN9ExitGames6Common6ObjectD2Ev($0);
    $$0 = $11;$$01 = $12;
    break;
   }
   __ZN9ExitGames6Common6ObjectD2Ev($0);
   $5 = ((($this)) + 4|0);
   __THREW__ = 0;
   $6 = (invoke_iiii(880,($5|0),($obj|0),1)|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $13 = ___cxa_find_matching_catch_2()|0;
    $14 = tempRet0;
    __ZN9ExitGames6Common6ObjectD2Ev($obj);
    $$0 = $13;$$01 = $14;
    break;
   } else {
    __ZN9ExitGames6Common6ObjectD2Ev($obj);
    __ZN9ExitGames6Common6ObjectD2Ev($forceCompilationToFailForUnsupportedTypes);
    STACKTOP = sp;return ($6|0);
   }
  }
 } while(0);
 __ZN9ExitGames6Common6ObjectD2Ev($forceCompilationToFailForUnsupportedTypes);
 ___resumeException($$0|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10152);
 $0 = ((($this)) + 36|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 do {
  if ($5) {
   label = 6;
  } else {
   $6 = ((($2)) + 20|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ($7|0)==(0|0);
   if ($8) {
    $13 = $2;
   } else {
    $9 = ((($7)) + -8|0);
    __THREW__ = 0;
    invoke_vi(750,($9|0));
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     break;
    }
    $$pre$i = HEAP32[$1>>2]|0;
    $13 = $$pre$i;
   }
   $12 = ((($13)) + 20|0);
   HEAP32[$12>>2] = 0;
   label = 6;
  }
 } while(0);
 do {
  if ((label|0) == 6) {
   $14 = ((($this)) + 104|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0|0);
   if (!($16)) {
    $17 = HEAP32[$15>>2]|0;
    $18 = HEAP32[$17>>2]|0;
    FUNCTION_TABLE_vi[$18 & 2047]($15);
    __THREW__ = 0;
    invoke_vi(750,($15|0));
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if ($20) {
     break;
    }
   }
   $21 = ((($this)) + 108|0);
   $22 = HEAP32[$21>>2]|0;
   $23 = ($22|0)==(0|0);
   if (!($23)) {
    $24 = HEAP32[$22>>2]|0;
    $25 = HEAP32[$24>>2]|0;
    FUNCTION_TABLE_vi[$25 & 2047]($22);
    __THREW__ = 0;
    invoke_vi(750,($22|0));
    $26 = __THREW__; __THREW__ = 0;
    $27 = $26&1;
    if ($27) {
     break;
    }
   }
   $28 = ((($this)) + 112|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)==(0|0);
   if (!($30)) {
    $31 = HEAP32[$29>>2]|0;
    $32 = HEAP32[$31>>2]|0;
    FUNCTION_TABLE_vi[$32 & 2047]($29);
    __THREW__ = 0;
    invoke_vi(750,($29|0));
    $33 = __THREW__; __THREW__ = 0;
    $34 = $33&1;
    if ($34) {
     break;
    }
   }
   $35 = ((($this)) + 148|0);
   __ZN9ExitGames6Common7Helpers13SharedPointerINS_6Photon8Internal8PeerBase10InitValuesEED2Ev($35);
   $36 = ((($this)) + 116|0);
   HEAP32[$36>>2] = (3820);
   $37 = ((($this)) + 120|0);
   HEAP32[$37>>2] = 0;
   $38 = ((($this)) + 132|0);
   $39 = HEAP32[$38>>2]|0;
   __THREW__ = 0;
   invoke_vi(750,($39|0));
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    $42 = ___cxa_find_matching_catch_3(0|0)|0;
    $43 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($36);
    ___clang_call_terminate($42);
    // unreachable;
   } else {
    __ZN9ExitGames6Common4BaseD2Ev($36);
    $44 = ((($this)) + 20|0);
    __ZN9ExitGames6Common7JStringD2Ev($44);
    return;
   }
  }
 } while(0);
 $45 = ___cxa_find_matching_catch_3(0|0)|0;
 $46 = tempRet0;
 $47 = ((($this)) + 148|0);
 __ZN9ExitGames6Common7Helpers13SharedPointerINS_6Photon8Internal8PeerBase10InitValuesEED2Ev($47);
 $48 = ((($this)) + 116|0);
 HEAP32[$48>>2] = (3820);
 $49 = ((($this)) + 120|0);
 HEAP32[$49>>2] = 0;
 $50 = ((($this)) + 132|0);
 $51 = HEAP32[$50>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($51|0));
 $52 = __THREW__; __THREW__ = 0;
 $53 = $52&1;
 if ($53) {
  $54 = ___cxa_find_matching_catch_3(0|0)|0;
  $55 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($48);
  ___clang_call_terminate($54);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($48);
  $56 = ((($this)) + 20|0);
  __ZN9ExitGames6Common7JStringD2Ev($56);
  ___clang_call_terminate($45);
  // unreachable;
 }
}
function __ZN9ExitGames6Common7Helpers13SharedPointerINS_6Photon8Internal8PeerBase10InitValuesEED2Ev($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 $8 = (_i64Add(($4|0),($7|0),-1,-1)|0);
 $9 = tempRet0;
 $10 = $1;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = ($8|0)==(0);
 $15 = ($9|0)==(0);
 $16 = $14 & $15;
 if (!($16)) {
  return;
 }
 $17 = HEAP32[$this>>2]|0;
 $18 = ($17|0)==(0|0);
 do {
  if ($18) {
   $24 = $1;
  } else {
   $19 = ((($17)) + 32|0);
   __ZN9ExitGames6Common6ObjectD2Ev($19);
   $20 = ((($17)) + 16|0);
   __ZN9ExitGames6Common7JStringD2Ev($20);
   __ZN9ExitGames6Common7JStringD2Ev($17);
   __THREW__ = 0;
   invoke_vi(750,($17|0));
   $21 = __THREW__; __THREW__ = 0;
   $22 = $21&1;
   if ($22) {
    $27 = ___cxa_find_matching_catch_3(0|0)|0;
    $28 = tempRet0;
    ___clang_call_terminate($27);
    // unreachable;
   } else {
    $$pre = HEAP32[$0>>2]|0;
    $24 = $$pre;
    break;
   }
  }
 } while(0);
 $23 = ($24|0)==(0|0);
 if ($23) {
  return;
 }
 __THREW__ = 0;
 invoke_vi(750,($24|0));
 $25 = __THREW__; __THREW__ = 0;
 $26 = $25&1;
 if ($26) {
  $27 = ___cxa_find_matching_catch_3(0|0)|0;
  $28 = tempRet0;
  ___clang_call_terminate($27);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Photon8Internal8PeerBaseD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase12sendAcksOnlyEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZN9ExitGames6Photon8Internal9OperationD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10304);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$0>>2] = 0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  HEAP32[$0>>2] = 0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Photon8Internal9OperationD0Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10304);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  HEAP32[$0>>2] = 0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  HEAP32[$0>>2] = 0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Photon8Internal9Operation8toStringERNS_6Common7JStringEb($this,$retStr,$0) {
 $this = $this|0;
 $retStr = $retStr|0;
 $0 = $0|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index12Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body12$index7Z2D = 0, $eh$lpad$body12$indexZ2D = 0, $eh$lpad$body5$index17Z2D = 0, $eh$lpad$body5$indexZ2D = 0, $eh$lpad$body8$index2Z2D = 0, $eh$lpad$body8$indexZ2D = 0, $str$i = 0, $str$i9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $str$i9 = sp + 96|0;
 $str$i = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp + 48|0;
 $3 = sp + 32|0;
 $4 = sp + 16|0;
 $5 = sp;
 __ZN9ExitGames6Common7JStringC2EPKw($5,10320);
 $6 = ((($this)) + 8|0);
 __THREW__ = 0;
 invoke_vii(786,($2|0),($5|0));
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $41 = ___cxa_find_matching_catch_2()|0;
  $42 = tempRet0;
  $eh$lpad$body$index12Z2D = $42;$eh$lpad$body$indexZ2D = $41;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  do {
   if ($10) {
    label = 7;
   } else {
    $11 = HEAP32[$6>>2]|0;
    __THREW__ = 0;
    (invoke_iii(923,($str$i|0),($11|0))|0);
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     __THREW__ = 0;
     $14 = (invoke_iii(760,($2|0),($str$i|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     if (!($16)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($4|0),($14|0));
      $19 = __THREW__; __THREW__ = 0;
      $20 = $19&1;
      if ($20) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($2);
      __THREW__ = 0;
      invoke_vii(786,($1|0),($4|0));
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      L10: do {
       if ($24) {
        $43 = ___cxa_find_matching_catch_2()|0;
        $44 = tempRet0;
        $eh$lpad$body5$index17Z2D = $44;$eh$lpad$body5$indexZ2D = $43;
        label = 22;
       } else {
        __THREW__ = 0;
        invoke_vii(787,($str$i9|0),0);
        $25 = __THREW__; __THREW__ = 0;
        $26 = $25&1;
        do {
         if ($26) {
          label = 15;
         } else {
          __THREW__ = 0;
          (invoke_iii(759,($str$i9|0),(10392|0))|0);
          $27 = __THREW__; __THREW__ = 0;
          $28 = $27&1;
          if (!($28)) {
           __THREW__ = 0;
           $29 = (invoke_iii(760,($1|0),($str$i9|0))|0);
           $30 = __THREW__; __THREW__ = 0;
           $31 = $30&1;
           if (!($31)) {
            __ZN9ExitGames6Common7JStringD2Ev($str$i9);
            __THREW__ = 0;
            invoke_vii(786,($3|0),($29|0));
            $34 = __THREW__; __THREW__ = 0;
            $35 = $34&1;
            if ($35) {
             label = 15;
             break;
            }
            __ZN9ExitGames6Common7JStringD2Ev($1);
            __THREW__ = 0;
            $38 = (invoke_iii(760,($retStr|0),($3|0))|0);
            $39 = __THREW__; __THREW__ = 0;
            $40 = $39&1;
            if ($40) {
             $45 = ___cxa_find_matching_catch_2()|0;
             $46 = tempRet0;
             __ZN9ExitGames6Common7JStringD2Ev($3);
             $$0 = $46;$$01 = $45;
             break L10;
            } else {
             __ZN9ExitGames6Common7JStringD2Ev($3);
             __ZN9ExitGames6Common7JStringD2Ev($4);
             __ZN9ExitGames6Common7JStringD2Ev($5);
             STACKTOP = sp;return ($38|0);
            }
           }
          }
          $32 = ___cxa_find_matching_catch_2()|0;
          $33 = tempRet0;
          __ZN9ExitGames6Common7JStringD2Ev($str$i9);
          $eh$lpad$body12$index7Z2D = $33;$eh$lpad$body12$indexZ2D = $32;
         }
        } while(0);
        if ((label|0) == 15) {
         $36 = ___cxa_find_matching_catch_2()|0;
         $37 = tempRet0;
         $eh$lpad$body12$index7Z2D = $37;$eh$lpad$body12$indexZ2D = $36;
        }
        __ZN9ExitGames6Common7JStringD2Ev($1);
        $eh$lpad$body5$index17Z2D = $eh$lpad$body12$index7Z2D;$eh$lpad$body5$indexZ2D = $eh$lpad$body12$indexZ2D;
        label = 22;
       }
      } while(0);
      if ((label|0) == 22) {
       $$0 = $eh$lpad$body5$index17Z2D;$$01 = $eh$lpad$body5$indexZ2D;
      }
      __ZN9ExitGames6Common7JStringD2Ev($4);
      $$1 = $$0;$$12 = $$01;
      __ZN9ExitGames6Common7JStringD2Ev($5);
      ___resumeException($$12|0);
      // unreachable;
     }
    }
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body8$index2Z2D = $18;$eh$lpad$body8$indexZ2D = $17;
   }
  } while(0);
  if ((label|0) == 7) {
   $21 = ___cxa_find_matching_catch_2()|0;
   $22 = tempRet0;
   $eh$lpad$body8$index2Z2D = $22;$eh$lpad$body8$indexZ2D = $21;
  }
  __ZN9ExitGames6Common7JStringD2Ev($2);
  $eh$lpad$body$index12Z2D = $eh$lpad$body8$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body8$indexZ2D;
 }
 $$1 = $eh$lpad$body$index12Z2D;$$12 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($5);
 ___resumeException($$12|0);
 // unreachable;
 return (0)|0;
}
function __GLOBAL__sub_I_PeerData_cpp() {
 var $$0$i = 0, $$0$i2 = 0, $$01$i = 0, $$01$i1 = 0, $$1$i = 0, $$12$i = 0, $$2$i = 0, $$23$i = 0, $$3$i = 0, $$34$i = 0, $$4$i = 0, $$45$i = 0, $$5$i = 0, $$56$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $0 = sp + 128|0;
 $1 = sp + 112|0;
 $2 = sp + 96|0;
 $3 = sp + 80|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 32|0;
 $7 = sp + 16|0;
 $8 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($8,0);
 __THREW__ = 0;
 invoke_viii(866,($7|0),($8|0),(147424|0));
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $23 = ___cxa_find_matching_catch_2()|0;
  $24 = tempRet0;
  $$5$i = $24;$$56$i = $23;
  __ZN9ExitGames6Common7JStringD2Ev($8);
  ___resumeException($$56$i|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viii(788,($6|0),($7|0),(13236|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  $$4$i = $26;$$45$i = $25;
 } else {
  __THREW__ = 0;
  invoke_viii(866,($5|0),($6|0),(147425|0));
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $27 = ___cxa_find_matching_catch_2()|0;
   $28 = tempRet0;
   $$3$i = $28;$$34$i = $27;
  } else {
   __THREW__ = 0;
   invoke_viii(788,($4|0),($5|0),(13236|0));
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $29 = ___cxa_find_matching_catch_2()|0;
    $30 = tempRet0;
    $$2$i = $30;$$23$i = $29;
   } else {
    __THREW__ = 0;
    invoke_viii(866,($3|0),($4|0),(147426|0));
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if ($18) {
     $31 = ___cxa_find_matching_catch_2()|0;
     $32 = tempRet0;
     $$1$i = $32;$$12$i = $31;
    } else {
     __THREW__ = 0;
     invoke_viii(788,($2|0),($3|0),(13236|0));
     $19 = __THREW__; __THREW__ = 0;
     $20 = $19&1;
     do {
      if ($20) {
       $33 = ___cxa_find_matching_catch_2()|0;
       $34 = tempRet0;
       $$0$i = $34;$$01$i = $33;
      } else {
       __THREW__ = 0;
       invoke_viii(866,(163740|0),($2|0),(147427|0));
       $21 = __THREW__; __THREW__ = 0;
       $22 = $21&1;
       if ($22) {
        $35 = ___cxa_find_matching_catch_2()|0;
        $36 = tempRet0;
        __ZN9ExitGames6Common7JStringD2Ev($2);
        $$0$i = $36;$$01$i = $35;
        break;
       }
       __ZN9ExitGames6Common7JStringD2Ev($2);
       __ZN9ExitGames6Common7JStringD2Ev($3);
       __ZN9ExitGames6Common7JStringD2Ev($4);
       __ZN9ExitGames6Common7JStringD2Ev($5);
       __ZN9ExitGames6Common7JStringD2Ev($6);
       __ZN9ExitGames6Common7JStringD2Ev($7);
       __ZN9ExitGames6Common7JStringD2Ev($8);
       (___cxa_atexit((211|0),(163740|0),(___dso_handle|0))|0);
       __ZN9ExitGames6Common7JStringC2EPKw($1,10400);
       __THREW__ = 0;
       invoke_viii(866,($0|0),($1|0),(146477|0));
       $37 = __THREW__; __THREW__ = 0;
       $38 = $37&1;
       if ($38) {
        $41 = ___cxa_find_matching_catch_2()|0;
        $42 = tempRet0;
        $$0$i2 = $42;$$01$i1 = $41;
        __ZN9ExitGames6Common7JStringD2Ev($1);
        ___resumeException($$01$i1|0);
        // unreachable;
       }
       __THREW__ = 0;
       invoke_viii(866,(163756|0),($0|0),(146478|0));
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if (!($40)) {
        __ZN9ExitGames6Common7JStringD2Ev($0);
        __ZN9ExitGames6Common7JStringD2Ev($1);
        (___cxa_atexit((211|0),(163756|0),(___dso_handle|0))|0);
        STACKTOP = sp;return;
       }
       $43 = ___cxa_find_matching_catch_2()|0;
       $44 = tempRet0;
       __ZN9ExitGames6Common7JStringD2Ev($0);
       $$0$i2 = $44;$$01$i1 = $43;
       __ZN9ExitGames6Common7JStringD2Ev($1);
       ___resumeException($$01$i1|0);
       // unreachable;
      }
     } while(0);
     __ZN9ExitGames6Common7JStringD2Ev($3);
     $$1$i = $$0$i;$$12$i = $$01$i;
    }
    __ZN9ExitGames6Common7JStringD2Ev($4);
    $$2$i = $$1$i;$$23$i = $$12$i;
   }
   __ZN9ExitGames6Common7JStringD2Ev($5);
   $$3$i = $$2$i;$$34$i = $$23$i;
  }
  __ZN9ExitGames6Common7JStringD2Ev($6);
  $$4$i = $$3$i;$$45$i = $$34$i;
 }
 __ZN9ExitGames6Common7JStringD2Ev($7);
 $$5$i = $$4$i;$$56$i = $$45$i;
 __ZN9ExitGames6Common7JStringD2Ev($8);
 ___resumeException($$56$i|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8TConnectD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal13SocketConnectD2Ev($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8TConnectD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal13SocketConnectD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8TConnect10recvBufferEPPhPi($this,$ppBuff,$pBytesRead) {
 $this = $this|0;
 $ppBuff = $ppBuff|0;
 $pBytesRead = $pBytesRead|0;
 var $0 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $errorCode$0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Photon8Internal13SocketConnect21getRecvAvailableBytesEv($this)|0);
 $1 = ($0|0)>(-1);
 if (!($1)) {
  $errorCode$0 = 0;
  return ($errorCode$0|0);
 }
 $2 = (($0) + 8)|0;
 $3 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($2)|0);
 $4 = ((($3)) + 8|0);
 $5 = $3;
 $6 = $5;
 HEAP32[$6>>2] = $0;
 $7 = (($5) + 4)|0;
 $8 = $7;
 HEAP32[$8>>2] = 0;
 $9 = ($0|0)==(0);
 if (!($9)) {
  _memset(($4|0),0,($0|0))|0;
 }
 HEAP32[$ppBuff>>2] = $4;
 $10 = (__ZN9ExitGames6Photon8Internal13SocketConnect10recvBufferEPhiPi($this,$4,$0,$pBytesRead)|0);
 $errorCode$0 = $10;
 return ($errorCode$0|0);
}
function __ZN9ExitGames6Photon8Internal8TConnect9getSocketEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 76|0);
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)!=(0);
 $3 = $2 ? 10 : 2;
 $4 = (_socket($3,1,6)|0);
 return ($4|0);
}
function __ZN9ExitGames6Photon8Internal8TConnect15checkConnectionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $connected$0$off0 = 0, $errorFlags = 0, $n = 0, $vararg_buffer = 0, $waitd = 0, $writeFlags = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(272|0);
 $vararg_buffer = sp;
 $writeFlags = sp + 144|0;
 $errorFlags = sp + 16|0;
 $waitd = sp + 8|0;
 $n = sp + 4|0;
 dest=$writeFlags; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1 & 31;
 $3 = 1 << $2;
 $4 = $1 >>> 5;
 $5 = (($writeFlags) + ($4<<2)|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $3 | $6;
 HEAP32[$5>>2] = $7;
 dest=$errorFlags; stop=dest+128|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $8 = (($errorFlags) + ($4<<2)|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $3 | $9;
 HEAP32[$8>>2] = $10;
 HEAP32[$waitd>>2] = 0;
 $11 = ((($waitd)) + 4|0);
 HEAP32[$11>>2] = 0;
 $12 = (($1) + 1)|0;
 $13 = (_select($12,$errorFlags,$writeFlags,0,$waitd)|0);
 $14 = ($13|0)==(0);
 if ($14) {
  $connected$0$off0 = 0;
  STACKTOP = sp;return ($connected$0$off0|0);
 }
 $15 = HEAP32[$0>>2]|0;
 $16 = $15 >>> 5;
 $17 = (($errorFlags) + ($16<<2)|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = $15 & 31;
 $20 = 1 << $19;
 $21 = $20 & $18;
 $22 = ($21|0)==(0);
 if ($22) {
  $connected$0$off0 = 1;
  STACKTOP = sp;return ($connected$0$off0|0);
 }
 HEAP32[$n>>2] = 0;
 HEAP32[$vararg_buffer>>2] = $n;
 (_ioctl($15,21531,$vararg_buffer)|0);
 $23 = HEAP32[$n>>2]|0;
 $24 = ($23|0)==(0);
 if (!($24)) {
  $connected$0$off0 = 1;
  STACKTOP = sp;return ($connected$0$off0|0);
 }
 $25 = HEAP32[$this>>2]|0;
 $26 = ((($25)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 (FUNCTION_TABLE_ii[$27 & 2047]($this)|0);
 $28 = ((($this)) + 8|0);
 HEAP32[$28>>2] = 4;
 $connected$0$off0 = 0;
 STACKTOP = sp;return ($connected$0$off0|0);
}
function __ZN9ExitGames6Photon8Internal13SocketConnect21getRecvAvailableBytesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $available = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $available = sp + 4|0;
 HEAP32[$available>>2] = 0;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 HEAP32[$vararg_buffer>>2] = $available;
 (_ioctl($1,21531,$vararg_buffer)|0);
 $2 = HEAP32[$available>>2]|0;
 STACKTOP = sp;return ($2|0);
}
function __ZN9ExitGames6Photon8Internal5TPeerD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i14 = 0, $$pre$i4 = 0, $$pre$i9 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $i$01$i$i = 0, $i$01$i$i1 = 0, $i$01$i$i10 = 0, $i$01$i$i15 = 0, $i$01$i$i5 = 0;
 var $i$01$i2$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10516);
 $0 = ((($this)) + 164|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = ((($this)) + 176|0);
  $i$01$i$i = 0;
  while(1) {
   $4 = HEAP32[$3>>2]|0;
   $5 = (($4) + (($i$01$i$i*12)|0)|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = HEAP32[$6>>2]|0;
   FUNCTION_TABLE_vi[$7 & 2047]($5);
   $8 = (($i$01$i$i) + 1)|0;
   $9 = HEAP32[$0>>2]|0;
   $10 = ($8>>>0)<($9>>>0);
   if ($10) {
    $i$01$i$i = $8;
   } else {
    break;
   }
  }
 }
 HEAP32[$0>>2] = 0;
 $11 = ((($this)) + 184|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0);
 if (!($13)) {
  $14 = ((($this)) + 196|0);
  $i$01$i2$i = 0;
  while(1) {
   $15 = HEAP32[$14>>2]|0;
   $16 = (($15) + (($i$01$i2$i*12)|0)|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = HEAP32[$17>>2]|0;
   FUNCTION_TABLE_vi[$18 & 2047]($16);
   $19 = (($i$01$i2$i) + 1)|0;
   $20 = HEAP32[$11>>2]|0;
   $21 = ($19>>>0)<($20>>>0);
   if ($21) {
    $i$01$i2$i = $19;
   } else {
    break;
   }
  }
 }
 HEAP32[$11>>2] = 0;
 $22 = ((($this)) + 40|0);
 $23 = HEAP32[$22>>2]|0;
 $24 = HEAP32[$23>>2]|0;
 $25 = ((($24)) + 20|0);
 $26 = HEAP32[$25>>2]|0;
 __THREW__ = 0;
 invoke_vi($26|0,($23|0));
 $27 = __THREW__; __THREW__ = 0;
 $28 = $27&1;
 if ($28) {
  $59 = ___cxa_find_matching_catch_3(0|0)|0;
  $60 = tempRet0;
  $61 = ((($this)) + 180|0);
  HEAP32[$61>>2] = (12328);
  $62 = HEAP32[$11>>2]|0;
  $63 = ($62|0)==(0);
  $$pre$i9 = ((($this)) + 196|0);
  if (!($63)) {
   $i$01$i$i10 = 0;
   while(1) {
    $64 = HEAP32[$$pre$i9>>2]|0;
    $65 = (($64) + (($i$01$i$i10*12)|0)|0);
    $66 = HEAP32[$65>>2]|0;
    $67 = HEAP32[$66>>2]|0;
    FUNCTION_TABLE_vi[$67 & 2047]($65);
    $68 = (($i$01$i$i10) + 1)|0;
    $69 = HEAP32[$11>>2]|0;
    $70 = ($68>>>0)<($69>>>0);
    if ($70) {
     $i$01$i$i10 = $68;
    } else {
     break;
    }
   }
  }
  HEAP32[$11>>2] = 0;
  $71 = HEAP32[$$pre$i9>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($71|0));
  $72 = __THREW__; __THREW__ = 0;
  $73 = $72&1;
  if ($73) {
   $74 = ___cxa_find_matching_catch_3(0|0)|0;
   $75 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($61);
   ___clang_call_terminate($74);
   // unreachable;
  }
  __ZN9ExitGames6Common4BaseD2Ev($61);
  $76 = ((($this)) + 160|0);
  HEAP32[$76>>2] = (12328);
  $77 = HEAP32[$0>>2]|0;
  $78 = ($77|0)==(0);
  $$pre$i14 = ((($this)) + 176|0);
  if (!($78)) {
   $i$01$i$i15 = 0;
   while(1) {
    $79 = HEAP32[$$pre$i14>>2]|0;
    $80 = (($79) + (($i$01$i$i15*12)|0)|0);
    $81 = HEAP32[$80>>2]|0;
    $82 = HEAP32[$81>>2]|0;
    FUNCTION_TABLE_vi[$82 & 2047]($80);
    $83 = (($i$01$i$i15) + 1)|0;
    $84 = HEAP32[$0>>2]|0;
    $85 = ($83>>>0)<($84>>>0);
    if ($85) {
     $i$01$i$i15 = $83;
    } else {
     break;
    }
   }
  }
  HEAP32[$0>>2] = 0;
  $86 = HEAP32[$$pre$i14>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($86|0));
  $87 = __THREW__; __THREW__ = 0;
  $88 = $87&1;
  if ($88) {
   $89 = ___cxa_find_matching_catch_3(0|0)|0;
   $90 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($76);
   ___clang_call_terminate($89);
   // unreachable;
  } else {
   __ZN9ExitGames6Common4BaseD2Ev($76);
   __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this);
   ___clang_call_terminate($59);
   // unreachable;
  }
 } else {
  $29 = ((($this)) + 180|0);
  HEAP32[$29>>2] = (12328);
  $30 = HEAP32[$11>>2]|0;
  $31 = ($30|0)==(0);
  $$pre$i = ((($this)) + 196|0);
  if (!($31)) {
   $i$01$i$i1 = 0;
   while(1) {
    $32 = HEAP32[$$pre$i>>2]|0;
    $33 = (($32) + (($i$01$i$i1*12)|0)|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = HEAP32[$34>>2]|0;
    FUNCTION_TABLE_vi[$35 & 2047]($33);
    $36 = (($i$01$i$i1) + 1)|0;
    $37 = HEAP32[$11>>2]|0;
    $38 = ($36>>>0)<($37>>>0);
    if ($38) {
     $i$01$i$i1 = $36;
    } else {
     break;
    }
   }
  }
  HEAP32[$11>>2] = 0;
  $39 = HEAP32[$$pre$i>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($39|0));
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if ($41) {
   $42 = ___cxa_find_matching_catch_3(0|0)|0;
   $43 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($29);
   ___clang_call_terminate($42);
   // unreachable;
  }
  __ZN9ExitGames6Common4BaseD2Ev($29);
  $44 = ((($this)) + 160|0);
  HEAP32[$44>>2] = (12328);
  $45 = HEAP32[$0>>2]|0;
  $46 = ($45|0)==(0);
  $$pre$i4 = ((($this)) + 176|0);
  if (!($46)) {
   $i$01$i$i5 = 0;
   while(1) {
    $47 = HEAP32[$$pre$i4>>2]|0;
    $48 = (($47) + (($i$01$i$i5*12)|0)|0);
    $49 = HEAP32[$48>>2]|0;
    $50 = HEAP32[$49>>2]|0;
    FUNCTION_TABLE_vi[$50 & 2047]($48);
    $51 = (($i$01$i$i5) + 1)|0;
    $52 = HEAP32[$0>>2]|0;
    $53 = ($51>>>0)<($52>>>0);
    if ($53) {
     $i$01$i$i5 = $51;
    } else {
     break;
    }
   }
  }
  HEAP32[$0>>2] = 0;
  $54 = HEAP32[$$pre$i4>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($54|0));
  $55 = __THREW__; __THREW__ = 0;
  $56 = $55&1;
  if ($56) {
   $57 = ___cxa_find_matching_catch_3(0|0)|0;
   $58 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($44);
   ___clang_call_terminate($57);
   // unreachable;
  } else {
   __ZN9ExitGames6Common4BaseD2Ev($44);
   __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this);
   return;
  }
 }
}
function __ZN9ExitGames6Photon8Internal5TPeerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal5TPeerD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer9onConnectEi($this,$nError) {
 $this = $this|0;
 $nError = $nError|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal8PeerBase9onConnectEi($this,$nError);
 $0 = ($nError|0)==(0);
 if (!($0)) {
  $1 = ((($this)) + 4|0);
  $2 = HEAP32[$1>>2]|0;
  $3 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($2)|0);
  $4 = HEAP32[$3>>2]|0;
  $5 = ((($4)) + 16|0);
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vii[$6 & 2047]($3,1023);
  return;
 }
 $7 = ((($this)) + 36|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(3);
 if ($9) {
  return;
 }
 $10 = HEAP32[$this>>2]|0;
 $11 = ((($10)) + 36|0);
 $12 = HEAP32[$11>>2]|0;
 (FUNCTION_TABLE_ii[$12 & 2047]($this)|0);
 HEAP32[$7>>2] = 1;
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer13onReceiveDataEPhii($this,$pBuf,$iCount,$nError) {
 $this = $this|0;
 $pBuf = $pBuf|0;
 $iCount = $iCount|0;
 $nError = $nError|0;
 var $$0 = 0, $$04 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $notlhs = 0, $notrhs = 0, $or$cond$not = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer4 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer4 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 80|0;
 $1 = sp + 72|0;
 $2 = sp + 56|0;
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 HEAP32[$1>>2] = $iCount;
 $5 = ((($this)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($2,147576);
 __THREW__ = 0;
 $7 = (invoke_ii(764,($2|0))|0);
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if (!($9)) {
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $iCount;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $nError;
  invoke_viiiiiiii(765,($6|0),4,(10660|0),($7|0),1,197,(12012|0),($vararg_buffer|0));
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   __ZN9ExitGames6Common7JStringD2Ev($2);
   $12 = ($nError|0)==(0);
   if (!($12)) {
    $13 = ((($this)) + 36|0);
    HEAP32[$13>>2] = 6;
    $14 = ((($this)) + 4|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($15)|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ((($17)) + 16|0);
    $19 = HEAP32[$18>>2]|0;
    FUNCTION_TABLE_vii[$19 & 2047]($16,1039);
    $20 = HEAP32[$this>>2]|0;
    $21 = ((($20)) + 88|0);
    $22 = HEAP32[$21>>2]|0;
    FUNCTION_TABLE_vi[$22 & 2047]($this);
    STACKTOP = sp;return;
   }
   $notlhs = ($iCount|0)==(0);
   $notrhs = ($pBuf|0)==(0|0);
   $or$cond$not = $notrhs | $notlhs;
   if ($or$cond$not) {
    STACKTOP = sp;return;
   }
   $25 = (__Z11getTimeUnixv()|0);
   $26 = ((($this)) + 84|0);
   HEAP32[$26>>2] = $25;
   $27 = ((($this)) + 4|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = ((($28)) + 44|0);
   $30 = HEAP32[$29>>2]|0;
   $31 = ($30|0)==(0);
   if (!($31)) {
    $32 = ((($this)) + 104|0);
    $33 = HEAP32[$32>>2]|0;
    __ZN9ExitGames6Photon8Internal19TrafficStatsCounter25incrementTotalPacketCountEv($33);
    $34 = HEAP32[$32>>2]|0;
    __ZN9ExitGames6Photon8Internal19TrafficStatsCounter27countTotalCommandsInPacketsEi($34,1);
    $35 = HEAP32[$32>>2]|0;
    $36 = HEAP32[$1>>2]|0;
    __ZN9ExitGames6Photon8Internal19TrafficStatsCounter22countReliableOpCommandEi($35,$36);
   }
   $37 = HEAP8[$pBuf>>0]|0;
   $38 = $37&255;
   L14: do {
    switch ($38|0) {
    case 240:  {
     __ZN9ExitGames6Photon8Internal5TPeer18readPingFromBufferEPhPi($this,$pBuf,$1);
     $39 = HEAP32[$27>>2]|0;
     $40 = ((($39)) + 44|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($41|0)==(0);
     if (!($42)) {
      $43 = ((($this)) + 104|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = HEAP32[$1>>2]|0;
      __ZN9ExitGames6Photon8Internal19TrafficStatsCounter19countControlCommandEi($44,$45);
     }
     break;
    }
    case 243:  {
     $46 = ((($this)) + 180|0);
     __ZN9ExitGames6Photon8Internal9OperationC2Ev($3);
     __THREW__ = 0;
     invoke_vii(924,($46|0),($3|0));
     $47 = __THREW__; __THREW__ = 0;
     $48 = $47&1;
     if ($48) {
      $74 = ___cxa_find_matching_catch_2()|0;
      $75 = tempRet0;
      __ZN9ExitGames6Photon8Internal9OperationD2Ev($3);
      $$0 = $75;$$04 = $74;
      ___resumeException($$04|0);
      // unreachable;
     }
     __ZN9ExitGames6Photon8Internal9OperationD2Ev($3);
     $49 = ((($this)) + 184|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = (($50) + -1)|0;
     $52 = ($50|0)==(0);
     do {
      if ($52) {
       __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
       __THREW__ = 0;
       $53 = (invoke_ii(764,($0|0))|0);
       $54 = __THREW__; __THREW__ = 0;
       $55 = $54&1;
       if ($55) {
        $58 = ___cxa_find_matching_catch_2()|0;
        $59 = tempRet0;
        __ZN9ExitGames6Common7JStringD2Ev($0);
        ___resumeException($58|0);
        // unreachable;
       }
       __THREW__ = 0;
       invoke_viiiiiiii(765,(162472|0),1,(10856|0),($53|0),1,566,(10968|0),($vararg_buffer2|0));
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        $58 = ___cxa_find_matching_catch_2()|0;
        $59 = tempRet0;
        __ZN9ExitGames6Common7JStringD2Ev($0);
        ___resumeException($58|0);
        // unreachable;
       } else {
        __ZN9ExitGames6Common7JStringD2Ev($0);
        break;
       }
      }
     } while(0);
     $60 = ((($this)) + 196|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = (($61) + (($51*12)|0)|0);
     $63 = HEAP32[$1>>2]|0;
     __ZN9ExitGames6Photon8Internal9Operation7setDataEPKhj($62,$pBuf,$63);
     $64 = HEAP32[$49>>2]|0;
     $65 = HEAP32[$27>>2]|0;
     $66 = ((($65)) + 32|0);
     $67 = HEAP16[$66>>1]|0;
     $68 = $67&65535;
     $69 = ($64|0)==($68|0);
     if ($69) {
      $70 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($65)|0);
      $71 = HEAP32[$70>>2]|0;
      $72 = ((($71)) + 16|0);
      $73 = HEAP32[$72>>2]|0;
      FUNCTION_TABLE_vii[$73 & 2047]($70,1033);
     }
     break;
    }
    default: {
     $76 = HEAP32[$5>>2]|0;
     __ZN9ExitGames6Common7JStringC2EPKc($4,147576);
     __THREW__ = 0;
     $77 = (invoke_ii(764,($4|0))|0);
     $78 = __THREW__; __THREW__ = 0;
     $79 = $78&1;
     if (!($79)) {
      $80 = HEAP8[$pBuf>>0]|0;
      $81 = $80&255;
      __THREW__ = 0;
      HEAP32[$vararg_buffer4>>2] = $81;
      invoke_viiiiiiii(765,($76|0),1,(10660|0),($77|0),1,231,(12108|0),($vararg_buffer4|0));
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if (!($83)) {
       __ZN9ExitGames6Common7JStringD2Ev($4);
       break L14;
      }
     }
     $84 = ___cxa_find_matching_catch_2()|0;
     $85 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($4);
     $$0 = $85;$$04 = $84;
     ___resumeException($$04|0);
     // unreachable;
    }
    }
   } while(0);
   $86 = (__Z11getTimeUnixv()|0);
   $87 = ((($this)) + 48|0);
   HEAP32[$87>>2] = $86;
   STACKTOP = sp;return;
  }
 }
 $23 = ___cxa_find_matching_catch_2()|0;
 $24 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$0 = $24;$$04 = $23;
 ___resumeException($$04|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal5TPeer7connectERKNS_6Common7JStringES6_RKNS3_6ObjectE($this,$ipAddr,$appID,$customData) {
 $this = $this|0;
 $ipAddr = $ipAddr|0;
 $appID = $appID|0;
 $customData = $customData|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Photon8Internal8PeerBase7connectERKNS_6Common7JStringES6_RKNS3_6ObjectE($this,$ipAddr,$appID,$customData)|0);
 return ($0|0);
}
function __ZN9ExitGames6Photon8Internal5TPeer10disconnectEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 switch ($1|0) {
 case 4: case 0:  {
  return;
  break;
 }
 default: {
 }
 }
 $2 = HEAP32[$this>>2]|0;
 $3 = ((($2)) + 116|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vi[$4 & 2047]($this);
 $5 = HEAP32[$this>>2]|0;
 $6 = ((($5)) + 76|0);
 $7 = HEAP32[$6>>2]|0;
 FUNCTION_TABLE_vi[$7 & 2047]($this);
 $8 = ((($this)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($9)|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($11)) + 16|0);
 $13 = HEAP32[$12>>2]|0;
 FUNCTION_TABLE_vii[$13 & 2047]($10,1025);
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer12serviceBasicEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $timeSinceLastReceive$0 = 0, $vararg_buffer = 0;
 var $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 __ZN9ExitGames6Photon8Internal8PeerBase12serviceBasicEv($this);
 $1 = ((($this)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(3);
 if ($3) {
  $4 = (__Z11getTimeUnixv()|0);
  $5 = ((($this)) + 48|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($4) - ($6))|0;
  $timeSinceLastReceive$0 = $7;
 } else {
  $timeSinceLastReceive$0 = 0;
 }
 $8 = ((($this)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($timeSinceLastReceive$0|0)>($11|0);
 if (!($12)) {
  $35 = ((($9)) + 16|0);
  $36 = HEAP32[$35>>2]|0;
  $37 = ($timeSinceLastReceive$0|0)>($36|0);
  if (!($37)) {
   STACKTOP = sp;return;
  }
  __ZN9ExitGames6Photon8Internal5TPeer8sendPingEv($this);
  STACKTOP = sp;return;
 }
 $13 = ((($this)) + 8|0);
 $14 = HEAP32[$13>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147563);
 __THREW__ = 0;
 $15 = (invoke_ii(764,($0|0))|0);
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($33|0);
  // unreachable;
 }
 __THREW__ = 0;
 $18 = (invoke_i(823)|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($33|0);
  // unreachable;
 }
 $21 = ((($this)) + 48|0);
 $22 = HEAP32[$21>>2]|0;
 __THREW__ = 0;
 HEAP32[$vararg_buffer>>2] = $18;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $22;
 invoke_viiiiiiii(765,($14|0),3,(10660|0),($15|0),1,71,(11748|0),($vararg_buffer|0));
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 if ($24) {
  $33 = ___cxa_find_matching_catch_2()|0;
  $34 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($33|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 HEAP32[$1>>2] = 6;
 $25 = HEAP32[$8>>2]|0;
 $26 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($25)|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = ((($27)) + 16|0);
 $29 = HEAP32[$28>>2]|0;
 FUNCTION_TABLE_vii[$29 & 2047]($26,1040);
 $30 = HEAP32[$this>>2]|0;
 $31 = ((($30)) + 88|0);
 $32 = HEAP32[$31>>2]|0;
 FUNCTION_TABLE_vi[$32 & 2047]($this);
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal5TPeer20sendOutgoingCommandsEv($this) {
 $this = $this|0;
 var $$01 = 0, $$02 = 0, $$pr = 0, $$pre = 0, $$pre5 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, $vararg_buffer3 = 0, $vararg_buffer5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer5 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 80|0;
 $1 = sp + 64|0;
 $2 = sp + 48|0;
 $3 = sp + 32|0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($2,147542);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($2|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($5|0),4,(10660|0),($6|0),1,82,(163772|0),($vararg_buffer|0));
  $9 = __THREW__; __THREW__ = 0;
  $10 = $9&1;
  if (!($10)) {
   __ZN9ExitGames6Common7JStringD2Ev($2);
   $11 = ((($this)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ((($12)) + 44|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)==(0);
   if (!($15)) {
    $16 = ((($this)) + 112|0);
    $17 = HEAP32[$16>>2]|0;
    __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter26sendOutgoingCommandsCalledEv($17);
   }
   $20 = ((($this)) + 36|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($21|0)==(0);
   if ($22) {
    STACKTOP = sp;return 0;
   }
   $23 = ((($this)) + 96|0);
   $24 = HEAP8[$23>>0]|0;
   $25 = ($24<<24>>24)==(0);
   if (!($25)) {
    STACKTOP = sp;return 0;
   }
   $26 = (__Z11getTimeUnixv()|0);
   $27 = HEAP32[$20>>2]|0;
   $28 = ($27|0)==(3);
   $29 = ((($this)) + 164|0);
   do {
    if ($28) {
     $30 = HEAP32[$29>>2]|0;
     $31 = ($30|0)==(0);
     if ($31) {
      $32 = HEAP32[$11>>2]|0;
      $33 = ((($32)) + 16|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = ($34|0)>(0);
      if ($35) {
       $36 = ((($this)) + 156|0);
       $37 = HEAP32[$36>>2]|0;
       $38 = (($26) - ($37))|0;
       $39 = ($38|0)>($34|0);
       if ($39) {
        $40 = HEAP32[$4>>2]|0;
        __ZN9ExitGames6Common7JStringC2EPKc($3,147542);
        __THREW__ = 0;
        $41 = (invoke_ii(764,($3|0))|0);
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        if (!($43)) {
         __THREW__ = 0;
         invoke_viiiiiiii(765,($40|0),4,(10660|0),($41|0),1,98,(11700|0),($vararg_buffer1|0));
         $44 = __THREW__; __THREW__ = 0;
         $45 = $44&1;
         if (!($45)) {
          __ZN9ExitGames6Common7JStringD2Ev($3);
          __ZN9ExitGames6Photon8Internal5TPeer8sendPingEv($this);
          HEAP32[$36>>2] = $26;
          break;
         }
        }
        $46 = ___cxa_find_matching_catch_2()|0;
        $47 = tempRet0;
        __ZN9ExitGames6Common7JStringD2Ev($3);
        $$01 = $46;$$02 = $47;
        ___resumeException($$01|0);
        // unreachable;
       }
      }
     }
    }
   } while(0);
   $48 = ((($this)) + 160|0);
   $49 = HEAP32[$29>>2]|0;
   $50 = ($49|0)==(0);
   if ($50) {
    STACKTOP = sp;return 0;
   }
   HEAP8[$23>>0] = 1;
   $51 = ((($this)) + 156|0);
   HEAP32[$51>>2] = $26;
   $52 = ((($this)) + 40|0);
   $53 = ((($this)) + 176|0);
   $59 = $49;
   while(1) {
    $54 = HEAP32[$52>>2]|0;
    $55 = HEAP32[$54>>2]|0;
    $56 = ((($55)) + 12|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = ($59|0)==(0);
    if ($58) {
     __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
     __THREW__ = 0;
     $60 = (invoke_ii(764,($0|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 22;
      break;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($60|0),1,566,(10968|0),($vararg_buffer3|0));
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      label = 22;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     $$pre = HEAP32[$29>>2]|0;
     $71 = $$pre;
    } else {
     $71 = $59;
    }
    $67 = HEAP32[$53>>2]|0;
    $68 = ((($67)) + 4|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = ($71|0)==(0);
    if ($70) {
     __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
     __THREW__ = 0;
     $72 = (invoke_ii(764,($1|0))|0);
     $73 = __THREW__; __THREW__ = 0;
     $74 = $73&1;
     if ($74) {
      label = 27;
      break;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($72|0),1,566,(10968|0),($vararg_buffer5|0));
     $75 = __THREW__; __THREW__ = 0;
     $76 = $75&1;
     if ($76) {
      label = 27;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($1);
     $$pre5 = HEAP32[$53>>2]|0;
     $80 = $$pre5;
    } else {
     $80 = $67;
    }
    $79 = ((($80)) + 8|0);
    $81 = HEAP32[$79>>2]|0;
    $82 = (FUNCTION_TABLE_iiii[$57 & 2047]($54,$69,$81)|0);
    $83 = ($82|0)==(32);
    if ($83) {
     label = 30;
     break;
    }
    __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE15removeElementAtEj($48,0);
    $$pr = HEAP32[$29>>2]|0;
    $84 = ($$pr|0)==(0);
    if ($84) {
     label = 30;
     break;
    } else {
     $59 = $$pr;
    }
   }
   if ((label|0) == 22) {
    $65 = ___cxa_find_matching_catch_2()|0;
    $66 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($65|0);
    // unreachable;
   }
   else if ((label|0) == 27) {
    $77 = ___cxa_find_matching_catch_2()|0;
    $78 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($1);
    ___resumeException($77|0);
    // unreachable;
   }
   else if ((label|0) == 30) {
    $85 = HEAP32[$11>>2]|0;
    $86 = ((($85)) + 44|0);
    $87 = HEAP32[$86>>2]|0;
    $88 = ($87|0)==(0);
    if ($88) {
     STACKTOP = sp;return 0;
    }
    $89 = HEAP32[$29>>2]|0;
    $90 = ($49|0)==($89|0);
    if ($90) {
     STACKTOP = sp;return 0;
    }
    $91 = ((($this)) + 108|0);
    $92 = HEAP32[$91>>2]|0;
    __ZN9ExitGames6Photon8Internal19TrafficStatsCounter25incrementTotalPacketCountEv($92);
    $93 = HEAP32[$91>>2]|0;
    $94 = HEAP32[$29>>2]|0;
    $95 = (($49) - ($94))|0;
    __ZN9ExitGames6Photon8Internal19TrafficStatsCounter27countTotalCommandsInPacketsEi($93,$95);
    STACKTOP = sp;return 0;
   }
  }
 }
 $18 = ___cxa_find_matching_catch_2()|0;
 $19 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$01 = $18;$$02 = $19;
 ___resumeException($$01|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal5TPeer12sendAcksOnlyEv($this) {
 $this = $this|0;
 var $$01 = 0, $$02 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $0 = sp + 32|0;
 $1 = sp + 16|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147529);
 __THREW__ = 0;
 $4 = (invoke_ii(764,($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if (!($6)) {
  __THREW__ = 0;
  invoke_viiiiiiii(765,($3|0),4,(10660|0),($4|0),1,124,(163772|0),($vararg_buffer|0));
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if (!($8)) {
   __ZN9ExitGames6Common7JStringD2Ev($0);
   $9 = ((($this)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ((($10)) + 44|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==(0);
   if (!($13)) {
    $14 = ((($this)) + 112|0);
    $15 = HEAP32[$14>>2]|0;
    __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter26sendOutgoingCommandsCalledEv($15);
   }
   $18 = ((($this)) + 36|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($19|0)==(0);
   if ($20) {
    STACKTOP = sp;return 0;
   }
   $21 = ((($this)) + 96|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = ($22<<24>>24)==(0);
   if (!($23)) {
    STACKTOP = sp;return 0;
   }
   $24 = (__Z11getTimeUnixv()|0);
   $25 = HEAP32[$18>>2]|0;
   $26 = ($25|0)==(3);
   if (!($26)) {
    STACKTOP = sp;return 0;
   }
   $27 = HEAP32[$9>>2]|0;
   $28 = ((($27)) + 16|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ($29|0)>(0);
   if (!($30)) {
    STACKTOP = sp;return 0;
   }
   $31 = ((($this)) + 156|0);
   $32 = HEAP32[$31>>2]|0;
   $33 = (($24) - ($32))|0;
   $34 = ($33|0)>($29|0);
   if (!($34)) {
    STACKTOP = sp;return 0;
   }
   $35 = HEAP32[$2>>2]|0;
   __ZN9ExitGames6Common7JStringC2EPKc($1,147529);
   __THREW__ = 0;
   $36 = (invoke_ii(764,($1|0))|0);
   $37 = __THREW__; __THREW__ = 0;
   $38 = $37&1;
   if (!($38)) {
    __THREW__ = 0;
    invoke_viiiiiiii(765,($35|0),4,(10660|0),($36|0),1,140,(11700|0),($vararg_buffer1|0));
    $39 = __THREW__; __THREW__ = 0;
    $40 = $39&1;
    if (!($40)) {
     __ZN9ExitGames6Common7JStringD2Ev($1);
     __ZN9ExitGames6Photon8Internal5TPeer8sendPingEv($this);
     HEAP32[$31>>2] = $24;
     STACKTOP = sp;return 0;
    }
   }
   $41 = ___cxa_find_matching_catch_2()|0;
   $42 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($1);
   $$01 = $42;$$02 = $41;
   ___resumeException($$02|0);
   // unreachable;
  }
 }
 $16 = ___cxa_find_matching_catch_2()|0;
 $17 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $$01 = $17;$$02 = $16;
 ___resumeException($$02|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon8Internal5TPeer24dispatchIncomingCommandsEv($this) {
 $this = $this|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre4 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147504);
 __THREW__ = 0;
 $3 = (invoke_ii(764,($0|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($15|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($2|0),4,(10660|0),($3|0),1,150,(163772|0),($vararg_buffer|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($15|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $8 = ((($this)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 44|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ($11|0)==(0);
 if (!($12)) {
  $13 = ((($this)) + 112|0);
  $14 = HEAP32[$13>>2]|0;
  __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter30dispatchIncomingCommandsCalledEv($14);
 }
 $17 = ((($this)) + 180|0);
 $18 = ((($this)) + 184|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = ($19|0)==(0);
 if ($20) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $21 = HEAP32[$this>>2]|0;
 $22 = ((($21)) + 100|0);
 $23 = HEAP32[$22>>2]|0;
 $$phi$trans$insert = ((($this)) + 196|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $24 = ((($$pre)) + 4|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ((($$pre)) + 8|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = (FUNCTION_TABLE_iiii[$23 & 2047]($this,$25,$27)|0);
 $29 = HEAP32[$18>>2]|0;
 $30 = ($29|0)==(0);
 if ($30) {
  $$0 = $28;
  STACKTOP = sp;return ($$0|0);
 }
 $$pre4 = HEAP32[$$phi$trans$insert>>2]|0;
 $31 = ((($$pre4)) + 8|0);
 $32 = HEAP32[$31>>2]|0;
 $33 = ((($this)) + 92|0);
 HEAP32[$33>>2] = $32;
 __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE15removeElementAtEj($17,0);
 $$0 = $28;
 STACKTOP = sp;return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal5TPeer20fetchServerTimestampEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 36|0);
 $2 = HEAP32[$1>>2]|0;
 switch ($2|0) {
 case 4: case 0:  {
  break;
 }
 default: {
  __ZN9ExitGames6Photon8Internal5TPeer8sendPingEv($this);
  STACKTOP = sp;return;
 }
 }
 $3 = ((($this)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147483);
 __THREW__ = 0;
 $5 = (invoke_ii(764,($0|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 __THREW__ = 0;
 invoke_viiiiiiii(765,($4|0),2,(10660|0),($5|0),1,171,(11084|0),($vararg_buffer|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $16 = ___cxa_find_matching_catch_2()|0;
  $17 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($16|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $10 = ((($this)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($11)|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 16|0);
 $15 = HEAP32[$14>>2]|0;
 FUNCTION_TABLE_vii[$15 & 2047]($12,1030);
 STACKTOP = sp;return;
}
function __ZNK9ExitGames6Photon8Internal5TPeer32getIncomingReliableCommandsCountEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal5TPeer25getQueuedIncomingCommandsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 184|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal5TPeer25getQueuedOutgoingCommandsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 164|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Photon8Internal5TPeer7cleanupEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal8PeerBase7cleanupEv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer5resetEv($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal8PeerBase5resetEv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer4sendEhPhjh($this,$cType,$payload,$payloadSize,$0) {
 $this = $this|0;
 $cType = $cType|0;
 $payload = $payload|0;
 $payloadSize = $payloadSize|0;
 $0 = $0|0;
 var $$0 = 0, $$02 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $1 = sp + 40|0;
 $2 = sp + 24|0;
 $3 = sp + 12|0;
 $4 = ((($this)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7JStringC2EPKc($2,147466);
 __THREW__ = 0;
 $6 = (invoke_ii(764,($2|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $9 = $cType&255;
  __THREW__ = 0;
  HEAP32[$vararg_buffer>>2] = $9;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = $payloadSize;
  invoke_viiiiiiii(765,($5|0),4,(10660|0),($6|0),1,255,(10752|0),($vararg_buffer|0));
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   __ZN9ExitGames6Common7JStringD2Ev($2);
   $12 = ($payload|0)==(0|0);
   $13 = ($payloadSize|0)==(0);
   $or$cond = $12 | $13;
   if ($or$cond) {
    STACKTOP = sp;return;
   }
   $16 = ((($this)) + 160|0);
   __ZN9ExitGames6Photon8Internal9OperationC2Ev($3);
   __THREW__ = 0;
   invoke_vii(924,($16|0),($3|0));
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    $44 = ___cxa_find_matching_catch_2()|0;
    $45 = tempRet0;
    __ZN9ExitGames6Photon8Internal9OperationD2Ev($3);
    $$0 = $45;$$02 = $44;
    ___resumeException($$02|0);
    // unreachable;
   }
   __ZN9ExitGames6Photon8Internal9OperationD2Ev($3);
   $19 = ((($this)) + 164|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (($20) + -1)|0;
   $22 = ($20|0)==(0);
   do {
    if ($22) {
     __ZN9ExitGames6Common7JStringC2EPKc($1,147471);
     __THREW__ = 0;
     $23 = (invoke_ii(764,($1|0))|0);
     $24 = __THREW__; __THREW__ = 0;
     $25 = $24&1;
     if ($25) {
      $28 = ___cxa_find_matching_catch_2()|0;
      $29 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($1);
      ___resumeException($28|0);
      // unreachable;
     }
     __THREW__ = 0;
     invoke_viiiiiiii(765,(162472|0),1,(10856|0),($23|0),1,566,(10968|0),($vararg_buffer2|0));
     $26 = __THREW__; __THREW__ = 0;
     $27 = $26&1;
     if ($27) {
      $28 = ___cxa_find_matching_catch_2()|0;
      $29 = tempRet0;
      __ZN9ExitGames6Common7JStringD2Ev($1);
      ___resumeException($28|0);
      // unreachable;
     } else {
      __ZN9ExitGames6Common7JStringD2Ev($1);
      break;
     }
    }
   } while(0);
   $30 = ((($this)) + 176|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = (($31) + (($21*12)|0)|0);
   __ZN9ExitGames6Photon8Internal9Operation7setDataEPKhj($32,$payload,$payloadSize);
   $33 = HEAP32[$19>>2]|0;
   $34 = ((($this)) + 4|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ((($35)) + 26|0);
   $37 = HEAP16[$36>>1]|0;
   $38 = $37&65535;
   $39 = ($33|0)==($38|0);
   if (!($39)) {
    STACKTOP = sp;return;
   }
   $40 = (__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv($35)|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ((($41)) + 16|0);
   $43 = HEAP32[$42>>2]|0;
   FUNCTION_TABLE_vii[$43 & 2047]($40,1027);
   STACKTOP = sp;return;
  }
 }
 $14 = ___cxa_find_matching_catch_2()|0;
 $15 = tempRet0;
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$0 = $15;$$02 = $14;
 ___resumeException($$02|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal5TPeer18serializeOperationEPKNS0_16OperationRequestEPPhPibh($this,$pOperationRequest,$buffer,$buffSize,$encrypt,$msgType) {
 $this = $this|0;
 $pOperationRequest = $pOperationRequest|0;
 $buffer = $buffer|0;
 $buffSize = $buffSize|0;
 $encrypt = $encrypt|0;
 $msgType = $msgType|0;
 var $$0 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Photon8Internal8PeerBase18serializeOperationEPKNS0_16OperationRequestEPPhPibh($this,$pOperationRequest,$buffer,$buffSize,$encrypt,$msgType)|0);
 if (!($0)) {
  $11 = ((($this)) + 88|0);
  HEAP32[$11>>2] = 0;
  $$0 = 0;
  return ($$0|0);
 }
 $1 = HEAP32[$buffSize>>2]|0;
 $2 = ((($this)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + 44|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0);
 if ($6) {
  $$0 = 1;
  return ($$0|0);
 }
 $7 = ((($this)) + 108|0);
 $8 = HEAP32[$7>>2]|0;
 __ZN9ExitGames6Photon8Internal19TrafficStatsCounter22countReliableOpCommandEi($8,$1);
 $9 = ((($this)) + 112|0);
 $10 = HEAP32[$9>>2]|0;
 __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter14countOperationEi($10,$1);
 $$0 = 1;
 return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal5TPeer14clearAllQueuesEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $i$01$i = 0, $i$01$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 164|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 if (!($2)) {
  $3 = ((($this)) + 176|0);
  $i$01$i = 0;
  while(1) {
   $4 = HEAP32[$3>>2]|0;
   $5 = (($4) + (($i$01$i*12)|0)|0);
   $6 = HEAP32[$5>>2]|0;
   $7 = HEAP32[$6>>2]|0;
   FUNCTION_TABLE_vi[$7 & 2047]($5);
   $8 = (($i$01$i) + 1)|0;
   $9 = HEAP32[$0>>2]|0;
   $10 = ($8>>>0)<($9>>>0);
   if ($10) {
    $i$01$i = $8;
   } else {
    break;
   }
  }
 }
 HEAP32[$0>>2] = 0;
 $11 = ((($this)) + 184|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = ($12|0)==(0);
 if ($13) {
  HEAP32[$11>>2] = 0;
  return;
 }
 $14 = ((($this)) + 196|0);
 $i$01$i2 = 0;
 while(1) {
  $15 = HEAP32[$14>>2]|0;
  $16 = (($15) + (($i$01$i2*12)|0)|0);
  $17 = HEAP32[$16>>2]|0;
  $18 = HEAP32[$17>>2]|0;
  FUNCTION_TABLE_vi[$18 & 2047]($16);
  $19 = (($i$01$i2) + 1)|0;
  $20 = HEAP32[$11>>2]|0;
  $21 = ($19>>>0)<($20>>>0);
  if ($21) {
   $i$01$i2 = $19;
  } else {
   break;
  }
 }
 HEAP32[$11>>2] = 0;
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer18readPingOpResponseEPKNS0_17OperationResponseE($this,$operationResponse) {
 $this = $this|0;
 $operationResponse = $operationResponse|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i12 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index12Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body26$index2Z2D = 0, $eh$lpad$body26$indexZ2D = 0, $eh$lpad$body32$index7Z2D = 0, $eh$lpad$body32$indexZ2D = 0, $eh$lpad$body9$index17Z2D = 0, $eh$lpad$body9$indexZ2D = 0, $or$cond$i$i = 0, $or$cond$i$i11 = 0;
 var $val1 = 0, $val2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $0 = sp + 200|0;
 $1 = sp + 160|0;
 $val2 = sp + 120|0;
 $2 = sp + 80|0;
 $val1 = sp + 40|0;
 $3 = sp;
 __ZNK9ExitGames6Photon17OperationResponse19getParameterForCodeEh($2,$operationResponse,2);
 __THREW__ = 0;
 invoke_vi(743,($val2|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $96 = ___cxa_find_matching_catch_2()|0;
  $97 = tempRet0;
  $eh$lpad$body$index12Z2D = $97;$eh$lpad$body$indexZ2D = $96;
 } else {
  HEAP32[$val2>>2] = (2876);
  __THREW__ = 0;
  $6 = (invoke_ii(848,($2|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  L4: do {
   if ($8) {
    label = 11;
   } else {
    $9 = ($6<<24>>24)==(105);
    if ($9) {
     __THREW__ = 0;
     $10 = (invoke_ii(846,($2|0))|0);
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      label = 11;
      break;
     }
     $13 = ($10|0)==(0);
     if ($13) {
      __THREW__ = 0;
      (invoke_iii(239,($val2|0),($2|0))|0);
      $14 = __THREW__; __THREW__ = 0;
      $15 = $14&1;
      if ($15) {
       label = 11;
       break;
      }
     } else {
      label = 7;
     }
    } else {
     label = 7;
    }
    do {
     if ((label|0) == 7) {
      __THREW__ = 0;
      invoke_vi(743,($1|0));
      $16 = __THREW__; __THREW__ = 0;
      $17 = $16&1;
      if ($17) {
       label = 11;
       break L4;
      }
      __THREW__ = 0;
      (invoke_iii(239,($val2|0),($1|0))|0);
      $18 = __THREW__; __THREW__ = 0;
      $19 = $18&1;
      if ($19) {
       $20 = ___cxa_find_matching_catch_2()|0;
       $21 = tempRet0;
       __ZN9ExitGames6Common6ObjectD2Ev($1);
       $eh$lpad$body26$index2Z2D = $21;$eh$lpad$body26$indexZ2D = $20;
       break L4;
      } else {
       __ZN9ExitGames6Common6ObjectD2Ev($1);
       break;
      }
     }
    } while(0);
    __ZN9ExitGames6Common6ObjectD2Ev($2);
    __THREW__ = 0;
    $24 = (invoke_ii(843,($val2|0))|0);
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    L17: do {
     if ($26) {
      label = 48;
     } else {
      __THREW__ = 0;
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       label = 48;
      } else {
       __THREW__ = 0;
       $29 = (invoke_ii(848,($val2|0))|0);
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        label = 48;
       } else {
        __THREW__ = 0;
        $32 = (invoke_ii(846,($val2|0))|0);
        $33 = __THREW__; __THREW__ = 0;
        $34 = $33&1;
        if ($34) {
         label = 48;
        } else {
         $35 = ($29<<24>>24)!=(105);
         $36 = ($32|0)!=(0);
         $or$cond$i$i = $35 | $36;
         if ($or$cond$i$i) {
          $$0$i$i = 0;
         } else {
          $37 = HEAP32[$24>>2]|0;
          $$0$i$i = $37;
         }
         __THREW__ = 0;
         invoke_viii(925,($3|0),($operationResponse|0),1);
         $38 = __THREW__; __THREW__ = 0;
         $39 = $38&1;
         if ($39) {
          label = 48;
         } else {
          __THREW__ = 0;
          invoke_vi(743,($val1|0));
          $40 = __THREW__; __THREW__ = 0;
          $41 = $40&1;
          if ($41) {
           $100 = ___cxa_find_matching_catch_2()|0;
           $101 = tempRet0;
           $eh$lpad$body9$index17Z2D = $101;$eh$lpad$body9$indexZ2D = $100;
          } else {
           HEAP32[$val1>>2] = (2876);
           __THREW__ = 0;
           $42 = (invoke_ii(848,($3|0))|0);
           $43 = __THREW__; __THREW__ = 0;
           $44 = $43&1;
           L29: do {
            if ($44) {
             label = 30;
            } else {
             $45 = ($42<<24>>24)==(105);
             if ($45) {
              __THREW__ = 0;
              $46 = (invoke_ii(846,($3|0))|0);
              $47 = __THREW__; __THREW__ = 0;
              $48 = $47&1;
              if ($48) {
               label = 30;
               break;
              }
              $49 = ($46|0)==(0);
              if ($49) {
               __THREW__ = 0;
               (invoke_iii(239,($val1|0),($3|0))|0);
               $50 = __THREW__; __THREW__ = 0;
               $51 = $50&1;
               if ($51) {
                label = 30;
                break;
               }
              } else {
               label = 26;
              }
             } else {
              label = 26;
             }
             do {
              if ((label|0) == 26) {
               __THREW__ = 0;
               invoke_vi(743,($0|0));
               $52 = __THREW__; __THREW__ = 0;
               $53 = $52&1;
               if ($53) {
                label = 30;
                break L29;
               }
               __THREW__ = 0;
               (invoke_iii(239,($val1|0),($0|0))|0);
               $54 = __THREW__; __THREW__ = 0;
               $55 = $54&1;
               if ($55) {
                $56 = ___cxa_find_matching_catch_2()|0;
                $57 = tempRet0;
                __ZN9ExitGames6Common6ObjectD2Ev($0);
                $eh$lpad$body32$index7Z2D = $57;$eh$lpad$body32$indexZ2D = $56;
                break L29;
               } else {
                __ZN9ExitGames6Common6ObjectD2Ev($0);
                break;
               }
              }
             } while(0);
             __ZN9ExitGames6Common6ObjectD2Ev($3);
             __THREW__ = 0;
             $60 = (invoke_ii(843,($val1|0))|0);
             $61 = __THREW__; __THREW__ = 0;
             $62 = $61&1;
             do {
              if (!($62)) {
               __THREW__ = 0;
               $63 = __THREW__; __THREW__ = 0;
               $64 = $63&1;
               if (!($64)) {
                __THREW__ = 0;
                $65 = (invoke_ii(848,($val1|0))|0);
                $66 = __THREW__; __THREW__ = 0;
                $67 = $66&1;
                if ($67) {
                 break;
                }
                __THREW__ = 0;
                $68 = (invoke_ii(846,($val1|0))|0);
                $69 = __THREW__; __THREW__ = 0;
                $70 = $69&1;
                if ($70) {
                 break;
                }
                $71 = ($65<<24>>24)!=(105);
                $72 = ($68|0)!=(0);
                $or$cond$i$i11 = $71 | $72;
                if ($or$cond$i$i11) {
                 $$0$i$i12 = 0;
                } else {
                 $73 = HEAP32[$60>>2]|0;
                 $$0$i$i12 = $73;
                }
                __THREW__ = 0;
                $74 = (invoke_i(823)|0);
                $75 = __THREW__; __THREW__ = 0;
                $76 = $75&1;
                if ($76) {
                 break;
                }
                $77 = (($74) - ($$0$i$i12))|0;
                $78 = ((($this)) + 56|0);
                $79 = HEAP8[$78>>0]|0;
                $80 = ($79<<24>>24)==(0);
                if ($80) {
                 $81 = ((($this)) + 60|0);
                 HEAP32[$81>>2] = $77;
                }
                $82 = HEAP32[$this>>2]|0;
                $83 = ((($82)) + 112|0);
                $84 = HEAP32[$83>>2]|0;
                __THREW__ = 0;
                invoke_vii($84|0,($this|0),($77|0));
                $85 = __THREW__; __THREW__ = 0;
                $86 = $85&1;
                if ($86) {
                 break;
                }
                $87 = HEAP8[$78>>0]|0;
                $88 = ($87<<24>>24)==(0);
                if (!($88)) {
                 __ZN9ExitGames6Common6ObjectD2Ev($val1);
                 __ZN9ExitGames6Common6ObjectD2Ev($val2);
                 STACKTOP = sp;return;
                }
                __THREW__ = 0;
                $89 = (invoke_i(823)|0);
                $90 = __THREW__; __THREW__ = 0;
                $91 = $90&1;
                if ($91) {
                 break;
                }
                $92 = $77 >> 1;
                $93 = (($92) + ($$0$i$i))|0;
                $94 = (($93) - ($89))|0;
                $95 = ((($this)) + 52|0);
                HEAP32[$95>>2] = $94;
                HEAP8[$78>>0] = 1;
                __ZN9ExitGames6Common6ObjectD2Ev($val1);
                __ZN9ExitGames6Common6ObjectD2Ev($val2);
                STACKTOP = sp;return;
               }
              }
             } while(0);
             $102 = ___cxa_find_matching_catch_2()|0;
             $103 = tempRet0;
             __ZN9ExitGames6Common6ObjectD2Ev($val1);
             $$0 = $102;$$01 = $103;
             break L17;
            }
           } while(0);
           if ((label|0) == 30) {
            $58 = ___cxa_find_matching_catch_2()|0;
            $59 = tempRet0;
            $eh$lpad$body32$index7Z2D = $59;$eh$lpad$body32$indexZ2D = $58;
           }
           __ZN9ExitGames6Common6ObjectD2Ev($val1);
           $eh$lpad$body9$index17Z2D = $eh$lpad$body32$index7Z2D;$eh$lpad$body9$indexZ2D = $eh$lpad$body32$indexZ2D;
          }
          __ZN9ExitGames6Common6ObjectD2Ev($3);
          $$0 = $eh$lpad$body9$indexZ2D;$$01 = $eh$lpad$body9$index17Z2D;
         }
        }
       }
      }
     }
    } while(0);
    if ((label|0) == 48) {
     $98 = ___cxa_find_matching_catch_2()|0;
     $99 = tempRet0;
     $$0 = $98;$$01 = $99;
    }
    __ZN9ExitGames6Common6ObjectD2Ev($val2);
    $$1 = $$0;$$12 = $$01;
    ___resumeException($$1|0);
    // unreachable;
   }
  } while(0);
  if ((label|0) == 11) {
   $22 = ___cxa_find_matching_catch_2()|0;
   $23 = tempRet0;
   $eh$lpad$body26$index2Z2D = $23;$eh$lpad$body26$indexZ2D = $22;
  }
  __ZN9ExitGames6Common6ObjectD2Ev($val2);
  $eh$lpad$body$index12Z2D = $eh$lpad$body26$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body26$indexZ2D;
 }
 __ZN9ExitGames6Common6ObjectD2Ev($2);
 $$1 = $eh$lpad$body$indexZ2D;$$12 = $eh$lpad$body$index12Z2D;
 ___resumeException($$1|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal5TPeer12initCallbackEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(1);
 if ($2) {
  HEAP32[$0>>2] = 3;
 }
 __ZN9ExitGames6Photon8Internal8PeerBase12initCallbackEv($this);
 return;
}
function __ZNK9ExitGames6Photon8Internal5TPeer18allocateConnectionERNS1_21PhotonConnectListenerE($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(108)|0);
 __ZN9ExitGames6Photon8Internal8TConnectC2ERNS1_21PhotonConnectListenerE($0,$listener);
 return ($0|0);
}
function __ZN9ExitGames6Photon8Internal5TPeer17generatePingBytesEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $op = 0, $req = 0, $serializedOp = 0, $serializedOpSize = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $op = sp + 108|0;
 $0 = sp + 160|0;
 $1 = sp;
 $2 = sp + 104|0;
 $req = sp + 48|0;
 $serializedOp = sp + 44|0;
 $serializedOpSize = sp + 40|0;
 __ZN9ExitGames6Common14DictionaryBaseC2Ev($op);
 HEAP32[$op>>2] = (2720);
 $3 = HEAP8[167373]|0;
 $4 = ($3<<24>>24)==(0);
 if ($4) {
  HEAP8[167374] = 98;
  HEAP8[167375] = 122;
  HEAP32[40417] = 0;
  HEAP8[167373] = 1;
 }
 HEAP8[$0>>0] = 1;
 __THREW__ = 0;
 $5 = (invoke_i(823)|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  HEAP32[$2>>2] = $5;
  __THREW__ = 0;
  invoke_viiiii(926,($1|0),($2|0),105,0,1);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if (!($9)) {
   HEAP32[$1>>2] = (2876);
   $10 = ((($op)) + 4|0);
   __THREW__ = 0;
   invoke_viii(898,($10|0),($0|0),($1|0));
   $11 = __THREW__; __THREW__ = 0;
   $12 = $11&1;
   if ($12) {
    $45 = ___cxa_find_matching_catch_2()|0;
    $46 = tempRet0;
    __ZN9ExitGames6Common6ObjectD2Ev($1);
    $$0 = $45;$$01 = $46;
    __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
    ___resumeException($$0|0);
    // unreachable;
   }
   __ZN9ExitGames6Common6ObjectD2Ev($1);
   __THREW__ = 0;
   invoke_viii(740,($req|0),1,($op|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if (!($14)) {
    $15 = HEAP32[$this>>2]|0;
    $16 = ((($15)) + 96|0);
    $17 = HEAP32[$16>>2]|0;
    __THREW__ = 0;
    (invoke_iiiiiii($17|0,($this|0),($req|0),($serializedOp|0),($serializedOpSize|0),0,6)|0);
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    do {
     if ($19) {
      $47 = ___cxa_find_matching_catch_2()|0;
      $48 = tempRet0;
      $eh$lpad$body$index2Z2D = $48;$eh$lpad$body$indexZ2D = $47;
     } else {
      $20 = HEAP32[$serializedOp>>2]|0;
      $21 = HEAP32[$serializedOpSize>>2]|0;
      HEAP32[$agg$result>>2] = (3820);
      $22 = ((($agg$result)) + 4|0);
      HEAP32[$22>>2] = $21;
      $23 = ((($agg$result)) + 8|0);
      HEAP32[$23>>2] = $21;
      __THREW__ = 0;
      $24 = (invoke_ii(775,($21|0))|0);
      $25 = __THREW__; __THREW__ = 0;
      $26 = $25&1;
      if ($26) {
       $41 = ___cxa_find_matching_catch_2()|0;
       $42 = tempRet0;
       __ZN9ExitGames6Common4BaseD2Ev($agg$result);
       $eh$lpad$body$index2Z2D = $42;$eh$lpad$body$indexZ2D = $41;
       break;
      }
      $27 = ((($agg$result)) + 16|0);
      HEAP32[$27>>2] = $24;
      $28 = ((($agg$result)) + 12|0);
      HEAP32[$28>>2] = 10;
      $29 = HEAP32[$22>>2]|0;
      $30 = ($29|0)==(0);
      if ($30) {
       __ZN9ExitGames6Photon16OperationRequestD2Ev($req);
       __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
       STACKTOP = sp;return;
      }
      $31 = HEAP8[$20>>0]|0;
      HEAP8[$24>>0] = $31;
      $32 = HEAP32[$22>>2]|0;
      $33 = ($32>>>0)>(1);
      if ($33) {
       $35 = 1;
      } else {
       __ZN9ExitGames6Photon16OperationRequestD2Ev($req);
       __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
       STACKTOP = sp;return;
      }
      while(1) {
       $$pre$i = HEAP32[$27>>2]|0;
       $34 = (($$pre$i) + ($35)|0);
       $36 = (($20) + ($35)|0);
       $37 = HEAP8[$36>>0]|0;
       HEAP8[$34>>0] = $37;
       $38 = (($35) + 1)|0;
       $39 = HEAP32[$22>>2]|0;
       $40 = ($38>>>0)<($39>>>0);
       if ($40) {
        $35 = $38;
       } else {
        break;
       }
      }
      __ZN9ExitGames6Photon16OperationRequestD2Ev($req);
      __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
      STACKTOP = sp;return;
     }
    } while(0);
    __ZN9ExitGames6Photon16OperationRequestD2Ev($req);
    $$0 = $eh$lpad$body$indexZ2D;$$01 = $eh$lpad$body$index2Z2D;
    __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
    ___resumeException($$0|0);
    // unreachable;
   }
  }
 }
 $43 = ___cxa_find_matching_catch_2()|0;
 $44 = tempRet0;
 $$0 = $43;$$01 = $44;
 __ZN9ExitGames6Common14DictionaryBaseD2Ev($op);
 ___resumeException($$0|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8TConnectC2ERNS1_21PhotonConnectListenerE($this,$listener) {
 $this = $this|0;
 $listener = $listener|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal13SocketConnectC2ERNS1_21PhotonConnectListenerE($this,$listener);
 HEAP32[$this>>2] = (10448);
 $0 = ((($this)) + 96|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 100|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 104|0);
 HEAP32[$2>>2] = 0;
 return;
}
function __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE10addElementERKS4_($this,$elem) {
 $this = $this|0;
 $elem = $elem|0;
 var $$pre = 0, $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($1|0)==($3|0);
 if ($4) {
  $5 = ((($this)) + 12|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (($6) + ($1))|0;
  $8 = ($1>>>0)<($7>>>0);
  if ($8) {
   HEAP32[$2>>2] = $7;
   $9 = ($7*12)|0;
   $10 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($9)|0);
   $11 = HEAP32[$0>>2]|0;
   $12 = ($11|0)==(0);
   $$pre$i = ((($this)) + 16|0);
   if (!($12)) {
    $i$01$i = 0;
    while(1) {
     $13 = (($10) + (($i$01$i*12)|0)|0);
     $14 = HEAP32[$$pre$i>>2]|0;
     $15 = (($14) + (($i$01$i*12)|0)|0);
     __ZN9ExitGames6Photon8Internal9OperationC2ERKS2_($13,$15);
     $16 = HEAP32[$$pre$i>>2]|0;
     $17 = (($16) + (($i$01$i*12)|0)|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = HEAP32[$18>>2]|0;
     FUNCTION_TABLE_vi[$19 & 2047]($17);
     $20 = (($i$01$i) + 1)|0;
     $21 = HEAP32[$0>>2]|0;
     $22 = ($20>>>0)<($21>>>0);
     if ($22) {
      $i$01$i = $20;
     } else {
      break;
     }
    }
   }
   $23 = HEAP32[$$pre$i>>2]|0;
   __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($23);
   HEAP32[$$pre$i>>2] = $10;
   $$pre = HEAP32[$0>>2]|0;
   $27 = $$pre;
  } else {
   $27 = $1;
  }
 } else {
  $27 = $1;
 }
 $24 = ((($this)) + 16|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = (($25) + (($27*12)|0)|0);
 __ZN9ExitGames6Photon8Internal9OperationC2ERKS2_($26,$elem);
 $28 = HEAP32[$0>>2]|0;
 $29 = (($28) + 1)|0;
 HEAP32[$0>>2] = $29;
 return;
}
function __ZN9ExitGames6Photon8Internal9Operation7setDataEPKhj($this,$data,$sz) {
 $this = $this|0;
 $data = $data|0;
 $sz = $sz|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (($sz) + 8)|0;
 $1 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj($0)|0);
 $2 = ((($1)) + 8|0);
 $3 = $1;
 $4 = $3;
 HEAP32[$4>>2] = $sz;
 $5 = (($3) + 4)|0;
 $6 = $5;
 HEAP32[$6>>2] = 0;
 $7 = ($sz|0)==(0);
 if (!($7)) {
  _memset(($2|0),0,($sz|0))|0;
 }
 $8 = ((($this)) + 4|0);
 HEAP32[$8>>2] = $2;
 _memcpy(($2|0),($data|0),($sz|0))|0;
 $9 = ((($this)) + 8|0);
 HEAP32[$9>>2] = $sz;
 return;
}
function __ZN9ExitGames6Photon8Internal9OperationC2ERKS2_($this,$toCopy) {
 $this = $this|0;
 $toCopy = $toCopy|0;
 var $$phi$trans$insert = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10304);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ($this|0)==($toCopy|0);
 if ($1) {
  return;
 }
 $2 = ((($toCopy)) + 8|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($this)) + 8|0);
 HEAP32[$4>>2] = $3;
 $$phi$trans$insert = ((($toCopy)) + 4|0);
 $$pre = HEAP32[$$phi$trans$insert>>2]|0;
 $phitmp = ($$pre|0)==(0|0);
 if ($phitmp) {
  HEAP32[$0>>2] = 0;
  return;
 }
 $5 = (($3) + 8)|0;
 __THREW__ = 0;
 $6 = (invoke_ii(775,($5|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___resumeException($17|0);
  // unreachable;
 }
 $9 = ((($6)) + 8|0);
 $10 = $6;
 $11 = $10;
 HEAP32[$11>>2] = $3;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = 0;
 $14 = ($3|0)==(0);
 if (!($14)) {
  _memset(($9|0),0,($3|0))|0;
 }
 HEAP32[$0>>2] = $9;
 $15 = HEAP32[$$phi$trans$insert>>2]|0;
 $16 = HEAP32[$4>>2]|0;
 _memcpy(($9|0),($15|0),($16|0))|0;
 return;
}
function __ZN9ExitGames6Photon8Internal9OperationC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10304);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = 0;
 return;
}
function __ZN9ExitGames6Photon8Internal5TPeer8sendPingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $pingBytes = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $pingBytes = sp;
 $0 = HEAP32[$this>>2]|0;
 $1 = ((($0)) + 140|0);
 $2 = HEAP32[$1>>2]|0;
 FUNCTION_TABLE_vii[$2 & 2047]($pingBytes,$this);
 $3 = ((($this)) + 40|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$4>>2]|0;
 $6 = ((($5)) + 12|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ((($pingBytes)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($pingBytes)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 __THREW__ = 0;
 (invoke_iiii($7|0,($4|0),($9|0),($11|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $19 = ___cxa_find_matching_catch_2()|0;
  $20 = tempRet0;
  HEAP32[$pingBytes>>2] = (3820);
  HEAP32[$10>>2] = 0;
  $21 = HEAP32[$8>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($21|0));
  $22 = __THREW__; __THREW__ = 0;
  $23 = $22&1;
  if ($23) {
   $24 = ___cxa_find_matching_catch_3(0|0)|0;
   $25 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($pingBytes);
   ___clang_call_terminate($24);
   // unreachable;
  } else {
   __ZN9ExitGames6Common4BaseD2Ev($pingBytes);
   ___resumeException($19|0);
   // unreachable;
  }
 } else {
  HEAP32[$pingBytes>>2] = (3820);
  HEAP32[$10>>2] = 0;
  $14 = HEAP32[$8>>2]|0;
  __THREW__ = 0;
  invoke_vi(750,($14|0));
  $15 = __THREW__; __THREW__ = 0;
  $16 = $15&1;
  if ($16) {
   $17 = ___cxa_find_matching_catch_3(0|0)|0;
   $18 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($pingBytes);
   ___clang_call_terminate($17);
   // unreachable;
  } else {
   __ZN9ExitGames6Common4BaseD2Ev($pingBytes);
   STACKTOP = sp;return;
  }
 }
}
function __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE15removeElementAtEj($this,$index) {
 $this = $this|0;
 $index = $index|0;
 var $$lcssa = 0, $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$0 = 0, $i$0$in2 = 0, $i$0$in2$phi = 0, $i$01 = 0, $i$03 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 $1 = ((($this)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2>>>0)>($index>>>0);
 do {
  if ($3) {
   $12 = $2;
  } else {
   __ZN9ExitGames6Common7JStringC2EPKc($0,147471);
   __THREW__ = 0;
   $4 = (invoke_ii(764,($0|0))|0);
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if ($6) {
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($9|0);
    // unreachable;
   }
   __THREW__ = 0;
   invoke_viiiiiiii(765,(162472|0),1,(10856|0),($4|0),1,566,(10968|0),($vararg_buffer|0));
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $9 = ___cxa_find_matching_catch_2()|0;
    $10 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    ___resumeException($9|0);
    // unreachable;
   } else {
    __ZN9ExitGames6Common7JStringD2Ev($0);
    $$pre = HEAP32[$1>>2]|0;
    $12 = $$pre;
    break;
   }
  }
 } while(0);
 $i$01 = (($index) + 1)|0;
 $11 = ($i$01>>>0)<($12>>>0);
 $13 = ((($this)) + 16|0);
 if ($11) {
  $i$0$in2 = $index;$i$03 = $i$01;
 } else {
  $$lcssa = $12;
  $23 = (($$lcssa) + -1)|0;
  $24 = HEAP32[$13>>2]|0;
  $25 = (($24) + (($23*12)|0)|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = HEAP32[$26>>2]|0;
  FUNCTION_TABLE_vi[$27 & 2047]($25);
  $28 = HEAP32[$1>>2]|0;
  $29 = (($28) + -1)|0;
  HEAP32[$1>>2] = $29;
  STACKTOP = sp;return;
 }
 while(1) {
  $14 = HEAP32[$13>>2]|0;
  $15 = (($14) + (($i$0$in2*12)|0)|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = HEAP32[$16>>2]|0;
  FUNCTION_TABLE_vi[$17 & 2047]($15);
  $18 = HEAP32[$13>>2]|0;
  $19 = (($18) + (($i$03*12)|0)|0);
  $20 = (($18) + (($i$0$in2*12)|0)|0);
  __ZN9ExitGames6Photon8Internal9OperationC2ERKS2_($20,$19);
  $i$0 = (($i$03) + 1)|0;
  $21 = HEAP32[$1>>2]|0;
  $22 = ($i$0>>>0)<($21>>>0);
  if ($22) {
   $i$0$in2$phi = $i$03;$i$03 = $i$0;$i$0$in2 = $i$0$in2$phi;
  } else {
   $$lcssa = $21;
   break;
  }
 }
 $23 = (($$lcssa) + -1)|0;
 $24 = HEAP32[$13>>2]|0;
 $25 = (($24) + (($23*12)|0)|0);
 $26 = HEAP32[$25>>2]|0;
 $27 = HEAP32[$26>>2]|0;
 FUNCTION_TABLE_vi[$27 & 2047]($25);
 $28 = HEAP32[$1>>2]|0;
 $29 = (($28) + -1)|0;
 HEAP32[$1>>2] = $29;
 STACKTOP = sp;return;
}
function __ZN9ExitGames6Photon8Internal5TPeer18readPingFromBufferEPhPi($this,$pBuffer,$nRead) {
 $this = $this|0;
 $pBuffer = $pBuffer|0;
 $nRead = $nRead|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ($nRead|0)!=(0|0);
 if ($0) {
  HEAP32[$nRead>>2] = 0;
 }
 $1 = ($pBuffer|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($pBuffer)) + 1|0);
 $3 = ((($pBuffer)) + 2|0);
 $4 = HEAP8[$2>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = ((($pBuffer)) + 3|0);
 $8 = HEAP8[$3>>0]|0;
 $9 = $8&255;
 $10 = $9 << 16;
 $11 = $10 | $6;
 $12 = ((($pBuffer)) + 4|0);
 $13 = HEAP8[$7>>0]|0;
 $14 = $13&255;
 $15 = $14 << 8;
 $16 = $11 | $15;
 $17 = ((($pBuffer)) + 5|0);
 $18 = HEAP8[$12>>0]|0;
 $19 = $18&255;
 $20 = $16 | $19;
 $21 = ((($pBuffer)) + 6|0);
 $22 = HEAP8[$17>>0]|0;
 $23 = $22&255;
 $24 = $23 << 24;
 $25 = ((($pBuffer)) + 7|0);
 $26 = HEAP8[$21>>0]|0;
 $27 = $26&255;
 $28 = $27 << 16;
 $29 = $28 | $24;
 $30 = ((($pBuffer)) + 8|0);
 $31 = HEAP8[$25>>0]|0;
 $32 = $31&255;
 $33 = $32 << 8;
 $34 = $29 | $33;
 $35 = HEAP8[$30>>0]|0;
 $36 = $35&255;
 $37 = $34 | $36;
 if ($0) {
  HEAP32[$nRead>>2] = 9;
 }
 $38 = (__Z11getTimeUnixv()|0);
 $39 = (($38) - ($37))|0;
 $40 = ((($this)) + 56|0);
 $41 = HEAP8[$40>>0]|0;
 $42 = ($41<<24>>24)==(0);
 if ($42) {
  $43 = ((($this)) + 60|0);
  HEAP32[$43>>2] = $39;
 }
 $44 = HEAP32[$this>>2]|0;
 $45 = ((($44)) + 112|0);
 $46 = HEAP32[$45>>2]|0;
 FUNCTION_TABLE_vii[$46 & 2047]($this,$39);
 $47 = HEAP8[$40>>0]|0;
 $48 = ($47<<24>>24)==(0);
 if (!($48)) {
  return;
 }
 $49 = $39 >> 1;
 $50 = (($49) + ($20))|0;
 $51 = (__Z11getTimeUnixv()|0);
 $52 = (($50) - ($51))|0;
 $53 = ((($this)) + 52|0);
 HEAP32[$53>>2] = $52;
 HEAP8[$40>>0] = 1;
 return;
}
function __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEED2Ev($this) {
 $this = $this|0;
 var $$pre = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12328);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $$pre = ((($this)) + 16|0);
 if (!($2)) {
  $i$01$i = 0;
  while(1) {
   $3 = HEAP32[$$pre>>2]|0;
   $4 = (($3) + (($i$01$i*12)|0)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = HEAP32[$5>>2]|0;
   FUNCTION_TABLE_vi[$6 & 2047]($4);
   $7 = (($i$01$i) + 1)|0;
   $8 = HEAP32[$0>>2]|0;
   $9 = ($7>>>0)<($8>>>0);
   if ($9) {
    $i$01$i = $7;
   } else {
    break;
   }
  }
 }
 HEAP32[$0>>2] = 0;
 $10 = HEAP32[$$pre>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($10|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_3(0|0)|0;
  $14 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($13);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  return;
 }
}
function __ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEED0Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $i$01$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (12328);
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $$pre$i = ((($this)) + 16|0);
 if (!($2)) {
  $i$01$i$i = 0;
  while(1) {
   $3 = HEAP32[$$pre$i>>2]|0;
   $4 = (($3) + (($i$01$i$i*12)|0)|0);
   $5 = HEAP32[$4>>2]|0;
   $6 = HEAP32[$5>>2]|0;
   FUNCTION_TABLE_vi[$6 & 2047]($4);
   $7 = (($i$01$i$i) + 1)|0;
   $8 = HEAP32[$0>>2]|0;
   $9 = ($7>>>0)<($8>>>0);
   if ($9) {
    $i$01$i$i = $7;
   } else {
    break;
   }
  }
 }
 HEAP32[$0>>2] = 0;
 $10 = HEAP32[$$pre$i>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($10|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_3(0|0)|0;
  $14 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($this);
  ___clang_call_terminate($13);
  // unreachable;
 } else {
  __ZN9ExitGames6Common4BaseD2Ev($this);
  __ZdlPv($this);
  return;
 }
}
function __ZNK9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE8toStringERNS0_7JStringEb($this,$retStr,$0) {
 $this = $this|0;
 $retStr = $retStr|0;
 $0 = $0|0;
 var $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = ((($this)) + 16|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($this)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 __ZN9ExitGames6Common7Helpers23ObjectToStringConverterINS_6Photon8Internal9OperationEE8toStringEPS5_jb($1,$3,$5,0);
 __THREW__ = 0;
 $6 = (invoke_iii(760,($retStr|0),($1|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($1);
  ___resumeException($9|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($1);
  STACKTOP = sp;return ($6|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Common7Helpers23ObjectToStringConverterINS_6Photon8Internal9OperationEE8toStringEPS5_jb($agg$result,$data,$size,$withTypes) {
 $agg$result = $agg$result|0;
 $data = $data|0;
 $size = $size|0;
 $withTypes = $withTypes|0;
 var $$0 = 0, $$01 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index4Z2D = 0, $eh$lpad$body$indexZ2D = 0, $i$06 = 0, $str$i = 0, $str$i2 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $str$i2 = sp + 32|0;
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2EPKw($agg$result,12344);
 $1 = ($size|0)==(0);
 L1: do {
  if ($1) {
   label = 16;
  } else {
   $2 = (($size) + -1)|0;
   $i$06 = 0;
   while(1) {
    $3 = (($data) + (($i$06*12)|0)|0);
    __THREW__ = 0;
    invoke_viii(783,($0|0),($3|0),($withTypes|0));
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     label = 11;
     break;
    }
    __THREW__ = 0;
    (invoke_iii(760,($agg$result|0),($0|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if ($7) {
     label = 14;
     break;
    }
    __ZN9ExitGames6Common7JStringD2Ev($0);
    $8 = ($i$06>>>0)<($2>>>0);
    if ($8) {
     __THREW__ = 0;
     invoke_vii(787,($str$i|0),0);
     $9 = __THREW__; __THREW__ = 0;
     $10 = $9&1;
     if ($10) {
      label = 11;
      break;
     }
     __THREW__ = 0;
     (invoke_iii(759,($str$i|0),(12352|0))|0);
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      label = 9;
      break;
     }
     __THREW__ = 0;
     (invoke_iii(760,($agg$result|0),($str$i|0))|0);
     $13 = __THREW__; __THREW__ = 0;
     $14 = $13&1;
     if ($14) {
      label = 9;
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
    }
    $23 = (($i$06) + 1)|0;
    $24 = ($23>>>0)<($size>>>0);
    if ($24) {
     $i$06 = $23;
    } else {
     label = 16;
     break L1;
    }
   }
   if ((label|0) == 9) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body$index4Z2D = $16;$eh$lpad$body$indexZ2D = $15;
    break;
   }
   else if ((label|0) == 11) {
    $17 = ___cxa_find_matching_catch_2()|0;
    $18 = tempRet0;
    $eh$lpad$body$index4Z2D = $18;$eh$lpad$body$indexZ2D = $17;
    break;
   }
   else if ((label|0) == 14) {
    $21 = ___cxa_find_matching_catch_2()|0;
    $22 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($0);
    $$0 = $22;$$01 = $21;
    __ZN9ExitGames6Common7JStringD2Ev($agg$result);
    ___resumeException($$01|0);
    // unreachable;
   }
  }
 } while(0);
 do {
  if ((label|0) == 16) {
   __THREW__ = 0;
   invoke_vii(787,($str$i2|0),0);
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   if ($26) {
    $19 = ___cxa_find_matching_catch_2()|0;
    $20 = tempRet0;
    $eh$lpad$body$index4Z2D = $20;$eh$lpad$body$indexZ2D = $19;
    break;
   }
   __THREW__ = 0;
   (invoke_iii(759,($str$i2|0),(12364|0))|0);
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   if (!($28)) {
    __THREW__ = 0;
    (invoke_iii(760,($agg$result|0),($str$i2|0))|0);
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    if (!($30)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i2);
     STACKTOP = sp;return;
    }
   }
   $31 = ___cxa_find_matching_catch_2()|0;
   $32 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i2);
   $eh$lpad$body$index4Z2D = $32;$eh$lpad$body$indexZ2D = $31;
  }
 } while(0);
 $$0 = $eh$lpad$body$index4Z2D;$$01 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($agg$result);
 ___resumeException($$01|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounterD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon12TrafficStatsD2Ev($this);
 return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounterD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon12TrafficStatsD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Photon12TrafficStats8toStringERNS_6Common7JStringEb($this,$retStr,$0) {
 $this = $this|0;
 $retStr = $retStr|0;
 $0 = $0|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $$5 = 0, $$56 = 0, $$6 = 0, $$67 = 0, $$7 = 0, $$78 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index22Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body11$index27Z2D = 0, $eh$lpad$body11$indexZ2D = 0, $eh$lpad$body14$index32Z2D = 0, $eh$lpad$body14$indexZ2D = 0, $eh$lpad$body17$index37Z2D = 0, $eh$lpad$body17$indexZ2D = 0, $eh$lpad$body20$index2Z2D = 0, $eh$lpad$body20$indexZ2D = 0, $eh$lpad$body24$index7Z2D = 0, $eh$lpad$body24$indexZ2D = 0, $eh$lpad$body28$index12Z2D = 0, $eh$lpad$body28$indexZ2D = 0, $eh$lpad$body32$index17Z2D = 0;
 var $eh$lpad$body32$indexZ2D = 0, $str$i = 0, $str$i21 = 0, $str$i25 = 0, $str$i29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(288|0);
 $str$i29 = sp + 264|0;
 $str$i25 = sp + 248|0;
 $str$i21 = sp + 232|0;
 $str$i = sp + 216|0;
 $1 = sp + 200|0;
 $2 = sp + 184|0;
 $3 = sp + 168|0;
 $4 = sp + 152|0;
 $5 = sp + 136|0;
 $6 = sp + 120|0;
 $7 = sp + 104|0;
 $8 = sp + 88|0;
 $9 = sp + 72|0;
 $10 = sp + 56|0;
 $11 = sp + 40|0;
 $12 = sp + 24|0;
 $13 = sp + 8|0;
 $14 = sp + 4|0;
 $15 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($13,0);
 __THREW__ = 0;
 invoke_vii(786,($4|0),($13|0));
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 if ($17) {
  $112 = ___cxa_find_matching_catch_2()|0;
  $113 = tempRet0;
  $eh$lpad$body$index22Z2D = $113;$eh$lpad$body$indexZ2D = $112;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  do {
   if ($19) {
    label = 7;
   } else {
    __THREW__ = 0;
    (invoke_iii(759,($str$i|0),(12396|0))|0);
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     __THREW__ = 0;
     $22 = (invoke_iii(760,($4|0),($str$i|0))|0);
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if (!($24)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($12|0),($22|0));
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($4);
      $31 = ((($this)) + 32|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ((($this)) + 36|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (($34) + ($32))|0;
      $36 = ((($this)) + 40|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = (($35) + ($37))|0;
      $39 = ((($this)) + 44|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = (($38) + ($40))|0;
      $42 = ((($this)) + 24|0);
      $43 = HEAP32[$42>>2]|0;
      $44 = ((($this)) + 4|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = Math_imul($45, $43)|0;
      $47 = (($41) + ($46))|0;
      HEAP32[$14>>2] = $47;
      __THREW__ = 0;
      invoke_viii(927,($11|0),($12|0),($14|0));
      $48 = __THREW__; __THREW__ = 0;
      $49 = $48&1;
      if ($49) {
       $114 = ___cxa_find_matching_catch_2()|0;
       $115 = tempRet0;
       $$6 = $115;$$67 = $114;
      } else {
       __THREW__ = 0;
       invoke_vii(786,($3|0),($11|0));
       $50 = __THREW__; __THREW__ = 0;
       $51 = $50&1;
       L13: do {
        if ($51) {
         $116 = ___cxa_find_matching_catch_2()|0;
         $117 = tempRet0;
         $eh$lpad$body11$index27Z2D = $117;$eh$lpad$body11$indexZ2D = $116;
         label = 43;
        } else {
         __THREW__ = 0;
         invoke_vii(787,($str$i21|0),0);
         $52 = __THREW__; __THREW__ = 0;
         $53 = $52&1;
         do {
          if ($53) {
           label = 16;
          } else {
           __THREW__ = 0;
           (invoke_iii(759,($str$i21|0),(12472|0))|0);
           $54 = __THREW__; __THREW__ = 0;
           $55 = $54&1;
           if (!($55)) {
            __THREW__ = 0;
            $56 = (invoke_iii(760,($3|0),($str$i21|0))|0);
            $57 = __THREW__; __THREW__ = 0;
            $58 = $57&1;
            if (!($58)) {
             __ZN9ExitGames6Common7JStringD2Ev($str$i21);
             __THREW__ = 0;
             invoke_vii(786,($10|0),($56|0));
             $61 = __THREW__; __THREW__ = 0;
             $62 = $61&1;
             if ($62) {
              label = 16;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($3);
             $65 = HEAP32[$31>>2]|0;
             $66 = HEAP32[$33>>2]|0;
             $67 = (($66) + ($65))|0;
             $68 = HEAP32[$36>>2]|0;
             $69 = (($67) + ($68))|0;
             $70 = HEAP32[$39>>2]|0;
             $71 = (($69) + ($70))|0;
             HEAP32[$15>>2] = $71;
             __THREW__ = 0;
             invoke_viii(927,($9|0),($10|0),($15|0));
             $72 = __THREW__; __THREW__ = 0;
             $73 = $72&1;
             if ($73) {
              $118 = ___cxa_find_matching_catch_2()|0;
              $119 = tempRet0;
              $$4 = $119;$$45 = $118;
             } else {
              __THREW__ = 0;
              invoke_vii(786,($2|0),($9|0));
              $74 = __THREW__; __THREW__ = 0;
              $75 = $74&1;
              L25: do {
               if ($75) {
                $120 = ___cxa_find_matching_catch_2()|0;
                $121 = tempRet0;
                $eh$lpad$body14$index32Z2D = $121;$eh$lpad$body14$indexZ2D = $120;
                label = 46;
               } else {
                __THREW__ = 0;
                invoke_vii(787,($str$i25|0),0);
                $76 = __THREW__; __THREW__ = 0;
                $77 = $76&1;
                L28: do {
                 if ($77) {
                  label = 25;
                 } else {
                  __THREW__ = 0;
                  (invoke_iii(759,($str$i25|0),(12556|0))|0);
                  $78 = __THREW__; __THREW__ = 0;
                  $79 = $78&1;
                  do {
                   if (!($79)) {
                    __THREW__ = 0;
                    $80 = (invoke_iii(760,($2|0),($str$i25|0))|0);
                    $81 = __THREW__; __THREW__ = 0;
                    $82 = $81&1;
                    if ($82) {
                     break;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($str$i25);
                    __THREW__ = 0;
                    invoke_vii(786,($8|0),($80|0));
                    $85 = __THREW__; __THREW__ = 0;
                    $86 = $85&1;
                    if ($86) {
                     label = 25;
                     break L28;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($2);
                    __THREW__ = 0;
                    invoke_viii(927,($7|0),($8|0),($42|0));
                    $89 = __THREW__; __THREW__ = 0;
                    $90 = $89&1;
                    if ($90) {
                     $122 = ___cxa_find_matching_catch_2()|0;
                     $123 = tempRet0;
                     $$2 = $123;$$23 = $122;
                    } else {
                     __THREW__ = 0;
                     invoke_vii(786,($1|0),($7|0));
                     $91 = __THREW__; __THREW__ = 0;
                     $92 = $91&1;
                     L37: do {
                      if ($92) {
                       $124 = ___cxa_find_matching_catch_2()|0;
                       $125 = tempRet0;
                       $eh$lpad$body17$index37Z2D = $125;$eh$lpad$body17$indexZ2D = $124;
                       label = 49;
                      } else {
                       __THREW__ = 0;
                       invoke_vii(787,($str$i29|0),0);
                       $93 = __THREW__; __THREW__ = 0;
                       $94 = $93&1;
                       L40: do {
                        if ($94) {
                         label = 34;
                        } else {
                         __THREW__ = 0;
                         (invoke_iii(759,($str$i29|0),(12636|0))|0);
                         $95 = __THREW__; __THREW__ = 0;
                         $96 = $95&1;
                         do {
                          if (!($96)) {
                           __THREW__ = 0;
                           $97 = (invoke_iii(760,($1|0),($str$i29|0))|0);
                           $98 = __THREW__; __THREW__ = 0;
                           $99 = $98&1;
                           if ($99) {
                            break;
                           }
                           __ZN9ExitGames6Common7JStringD2Ev($str$i29);
                           __THREW__ = 0;
                           invoke_vii(786,($6|0),($97|0));
                           $102 = __THREW__; __THREW__ = 0;
                           $103 = $102&1;
                           if ($103) {
                            label = 34;
                            break L40;
                           }
                           __ZN9ExitGames6Common7JStringD2Ev($1);
                           $106 = ((($this)) + 28|0);
                           __THREW__ = 0;
                           invoke_viii(927,($5|0),($6|0),($106|0));
                           $107 = __THREW__; __THREW__ = 0;
                           $108 = $107&1;
                           do {
                            if ($108) {
                             $126 = ___cxa_find_matching_catch_2()|0;
                             $127 = tempRet0;
                             $$0 = $127;$$01 = $126;
                            } else {
                             __THREW__ = 0;
                             $109 = (invoke_iii(760,($retStr|0),($5|0))|0);
                             $110 = __THREW__; __THREW__ = 0;
                             $111 = $110&1;
                             if ($111) {
                              $128 = ___cxa_find_matching_catch_2()|0;
                              $129 = tempRet0;
                              __ZN9ExitGames6Common7JStringD2Ev($5);
                              $$0 = $129;$$01 = $128;
                              break;
                             } else {
                              __ZN9ExitGames6Common7JStringD2Ev($5);
                              __ZN9ExitGames6Common7JStringD2Ev($6);
                              __ZN9ExitGames6Common7JStringD2Ev($7);
                              __ZN9ExitGames6Common7JStringD2Ev($8);
                              __ZN9ExitGames6Common7JStringD2Ev($9);
                              __ZN9ExitGames6Common7JStringD2Ev($10);
                              __ZN9ExitGames6Common7JStringD2Ev($11);
                              __ZN9ExitGames6Common7JStringD2Ev($12);
                              __ZN9ExitGames6Common7JStringD2Ev($13);
                              STACKTOP = sp;return ($109|0);
                             }
                            }
                           } while(0);
                           __ZN9ExitGames6Common7JStringD2Ev($6);
                           $$1 = $$0;$$12 = $$01;
                           break L37;
                          }
                         } while(0);
                         $100 = ___cxa_find_matching_catch_2()|0;
                         $101 = tempRet0;
                         __ZN9ExitGames6Common7JStringD2Ev($str$i29);
                         $eh$lpad$body32$index17Z2D = $101;$eh$lpad$body32$indexZ2D = $100;
                        }
                       } while(0);
                       if ((label|0) == 34) {
                        $104 = ___cxa_find_matching_catch_2()|0;
                        $105 = tempRet0;
                        $eh$lpad$body32$index17Z2D = $105;$eh$lpad$body32$indexZ2D = $104;
                       }
                       __ZN9ExitGames6Common7JStringD2Ev($1);
                       $eh$lpad$body17$index37Z2D = $eh$lpad$body32$index17Z2D;$eh$lpad$body17$indexZ2D = $eh$lpad$body32$indexZ2D;
                       label = 49;
                      }
                     } while(0);
                     if ((label|0) == 49) {
                      $$1 = $eh$lpad$body17$index37Z2D;$$12 = $eh$lpad$body17$indexZ2D;
                     }
                     __ZN9ExitGames6Common7JStringD2Ev($7);
                     $$2 = $$1;$$23 = $$12;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($8);
                    $$3 = $$2;$$34 = $$23;
                    break L25;
                   }
                  } while(0);
                  $83 = ___cxa_find_matching_catch_2()|0;
                  $84 = tempRet0;
                  __ZN9ExitGames6Common7JStringD2Ev($str$i25);
                  $eh$lpad$body28$index12Z2D = $84;$eh$lpad$body28$indexZ2D = $83;
                 }
                } while(0);
                if ((label|0) == 25) {
                 $87 = ___cxa_find_matching_catch_2()|0;
                 $88 = tempRet0;
                 $eh$lpad$body28$index12Z2D = $88;$eh$lpad$body28$indexZ2D = $87;
                }
                __ZN9ExitGames6Common7JStringD2Ev($2);
                $eh$lpad$body14$index32Z2D = $eh$lpad$body28$index12Z2D;$eh$lpad$body14$indexZ2D = $eh$lpad$body28$indexZ2D;
                label = 46;
               }
              } while(0);
              if ((label|0) == 46) {
               $$3 = $eh$lpad$body14$index32Z2D;$$34 = $eh$lpad$body14$indexZ2D;
              }
              __ZN9ExitGames6Common7JStringD2Ev($9);
              $$4 = $$3;$$45 = $$34;
             }
             __ZN9ExitGames6Common7JStringD2Ev($10);
             $$5 = $$4;$$56 = $$45;
             break L13;
            }
           }
           $59 = ___cxa_find_matching_catch_2()|0;
           $60 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($str$i21);
           $eh$lpad$body24$index7Z2D = $60;$eh$lpad$body24$indexZ2D = $59;
          }
         } while(0);
         if ((label|0) == 16) {
          $63 = ___cxa_find_matching_catch_2()|0;
          $64 = tempRet0;
          $eh$lpad$body24$index7Z2D = $64;$eh$lpad$body24$indexZ2D = $63;
         }
         __ZN9ExitGames6Common7JStringD2Ev($3);
         $eh$lpad$body11$index27Z2D = $eh$lpad$body24$index7Z2D;$eh$lpad$body11$indexZ2D = $eh$lpad$body24$indexZ2D;
         label = 43;
        }
       } while(0);
       if ((label|0) == 43) {
        $$5 = $eh$lpad$body11$index27Z2D;$$56 = $eh$lpad$body11$indexZ2D;
       }
       __ZN9ExitGames6Common7JStringD2Ev($11);
       $$6 = $$5;$$67 = $$56;
      }
      __ZN9ExitGames6Common7JStringD2Ev($12);
      $$7 = $$6;$$78 = $$67;
      __ZN9ExitGames6Common7JStringD2Ev($13);
      ___resumeException($$78|0);
      // unreachable;
     }
    }
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body20$index2Z2D = $26;$eh$lpad$body20$indexZ2D = $25;
   }
  } while(0);
  if ((label|0) == 7) {
   $29 = ___cxa_find_matching_catch_2()|0;
   $30 = tempRet0;
   $eh$lpad$body20$index2Z2D = $30;$eh$lpad$body20$indexZ2D = $29;
  }
  __ZN9ExitGames6Common7JStringD2Ev($4);
  $eh$lpad$body$index22Z2D = $eh$lpad$body20$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body20$indexZ2D;
 }
 $$7 = $eh$lpad$body$index22Z2D;$$78 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($13);
 ___resumeException($$78|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6CommonplIiEENS0_7JStringERKS2_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   $3 = HEAP32[$Rsh>>2]|0;
   __THREW__ = 0;
   (invoke_iii(833,($str$i|0),($3|0))|0);
   $4 = __THREW__; __THREW__ = 0;
   $5 = $4&1;
   if (!($5)) {
    __THREW__ = 0;
    $6 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if (!($8)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($6|0));
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if ($12) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $9 = ___cxa_find_matching_catch_2()|0;
   $10 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $10;$eh$lpad$body$indexZ2D = $9;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $13 = ___cxa_find_matching_catch_2()|0;
 $14 = tempRet0;
 $eh$lpad$body$index2Z2D = $14;$eh$lpad$body$indexZ2D = $13;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Photon12TrafficStatsD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounterD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon21TrafficStatsGameLevelD2Ev($this);
 return;
}
function __ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounterD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon21TrafficStatsGameLevelD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Photon21TrafficStatsGameLevel8toStringERNS_6Common7JStringEb($this,$retStr,$0) {
 $this = $this|0;
 $retStr = $retStr|0;
 $0 = $0|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $$3 = 0, $$34 = 0, $$4 = 0, $$45 = 0, $$5 = 0, $$56 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $eh$lpad$body$index17Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body12$index27Z2D = 0, $eh$lpad$body12$indexZ2D = 0, $eh$lpad$body15$index2Z2D = 0, $eh$lpad$body15$indexZ2D = 0, $eh$lpad$body19$index7Z2D = 0;
 var $eh$lpad$body19$indexZ2D = 0, $eh$lpad$body23$index12Z2D = 0, $eh$lpad$body23$indexZ2D = 0, $eh$lpad$body9$index22Z2D = 0, $eh$lpad$body9$indexZ2D = 0, $str$i = 0, $str$i16 = 0, $str$i20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 208|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(208|0);
 $str$i20 = sp + 192|0;
 $str$i16 = sp + 176|0;
 $str$i = sp + 160|0;
 $1 = sp + 144|0;
 $2 = sp + 128|0;
 $3 = sp + 112|0;
 $4 = sp + 96|0;
 $5 = sp + 80|0;
 $6 = sp + 64|0;
 $7 = sp + 48|0;
 $8 = sp + 32|0;
 $9 = sp + 16|0;
 $10 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($10,0);
 __THREW__ = 0;
 invoke_vii(786,($3|0),($10|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $68 = ___cxa_find_matching_catch_2()|0;
  $69 = tempRet0;
  $eh$lpad$body$index17Z2D = $69;$eh$lpad$body$indexZ2D = $68;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  do {
   if ($14) {
    label = 7;
   } else {
    __THREW__ = 0;
    (invoke_iii(759,($str$i|0),(13244|0))|0);
    $15 = __THREW__; __THREW__ = 0;
    $16 = $15&1;
    if (!($16)) {
     __THREW__ = 0;
     $17 = (invoke_iii(760,($3|0),($str$i|0))|0);
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if (!($19)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($9|0),($17|0));
      $22 = __THREW__; __THREW__ = 0;
      $23 = $22&1;
      if ($23) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($3);
      $26 = ((($this)) + 12|0);
      __THREW__ = 0;
      invoke_viii(927,($8|0),($9|0),($26|0));
      $27 = __THREW__; __THREW__ = 0;
      $28 = $27&1;
      if ($28) {
       $70 = ___cxa_find_matching_catch_2()|0;
       $71 = tempRet0;
       $$4 = $71;$$45 = $70;
      } else {
       __THREW__ = 0;
       invoke_vii(786,($2|0),($8|0));
       $29 = __THREW__; __THREW__ = 0;
       $30 = $29&1;
       L13: do {
        if ($30) {
         $72 = ___cxa_find_matching_catch_2()|0;
         $73 = tempRet0;
         $eh$lpad$body9$index22Z2D = $73;$eh$lpad$body9$indexZ2D = $72;
         label = 34;
        } else {
         __THREW__ = 0;
         invoke_vii(787,($str$i16|0),0);
         $31 = __THREW__; __THREW__ = 0;
         $32 = $31&1;
         do {
          if ($32) {
           label = 16;
          } else {
           __THREW__ = 0;
           (invoke_iii(759,($str$i16|0),(13328|0))|0);
           $33 = __THREW__; __THREW__ = 0;
           $34 = $33&1;
           if (!($34)) {
            __THREW__ = 0;
            $35 = (invoke_iii(760,($2|0),($str$i16|0))|0);
            $36 = __THREW__; __THREW__ = 0;
            $37 = $36&1;
            if (!($37)) {
             __ZN9ExitGames6Common7JStringD2Ev($str$i16);
             __THREW__ = 0;
             invoke_vii(786,($7|0),($35|0));
             $40 = __THREW__; __THREW__ = 0;
             $41 = $40&1;
             if ($41) {
              label = 16;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($2);
             $44 = ((($this)) + 20|0);
             __THREW__ = 0;
             invoke_viii(927,($6|0),($7|0),($44|0));
             $45 = __THREW__; __THREW__ = 0;
             $46 = $45&1;
             if ($46) {
              $74 = ___cxa_find_matching_catch_2()|0;
              $75 = tempRet0;
              $$2 = $75;$$23 = $74;
             } else {
              __THREW__ = 0;
              invoke_vii(786,($1|0),($6|0));
              $47 = __THREW__; __THREW__ = 0;
              $48 = $47&1;
              L25: do {
               if ($48) {
                $76 = ___cxa_find_matching_catch_2()|0;
                $77 = tempRet0;
                $eh$lpad$body12$index27Z2D = $77;$eh$lpad$body12$indexZ2D = $76;
                label = 37;
               } else {
                __THREW__ = 0;
                invoke_vii(787,($str$i20|0),0);
                $49 = __THREW__; __THREW__ = 0;
                $50 = $49&1;
                L28: do {
                 if ($50) {
                  label = 25;
                 } else {
                  __THREW__ = 0;
                  (invoke_iii(759,($str$i20|0),(13404|0))|0);
                  $51 = __THREW__; __THREW__ = 0;
                  $52 = $51&1;
                  do {
                   if (!($52)) {
                    __THREW__ = 0;
                    $53 = (invoke_iii(760,($1|0),($str$i20|0))|0);
                    $54 = __THREW__; __THREW__ = 0;
                    $55 = $54&1;
                    if ($55) {
                     break;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($str$i20);
                    __THREW__ = 0;
                    invoke_vii(786,($5|0),($53|0));
                    $58 = __THREW__; __THREW__ = 0;
                    $59 = $58&1;
                    if ($59) {
                     label = 25;
                     break L28;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($1);
                    $62 = ((($this)) + 28|0);
                    __THREW__ = 0;
                    invoke_viii(927,($4|0),($5|0),($62|0));
                    $63 = __THREW__; __THREW__ = 0;
                    $64 = $63&1;
                    do {
                     if ($64) {
                      $78 = ___cxa_find_matching_catch_2()|0;
                      $79 = tempRet0;
                      $$0 = $79;$$01 = $78;
                     } else {
                      __THREW__ = 0;
                      $65 = (invoke_iii(760,($retStr|0),($4|0))|0);
                      $66 = __THREW__; __THREW__ = 0;
                      $67 = $66&1;
                      if ($67) {
                       $80 = ___cxa_find_matching_catch_2()|0;
                       $81 = tempRet0;
                       __ZN9ExitGames6Common7JStringD2Ev($4);
                       $$0 = $81;$$01 = $80;
                       break;
                      } else {
                       __ZN9ExitGames6Common7JStringD2Ev($4);
                       __ZN9ExitGames6Common7JStringD2Ev($5);
                       __ZN9ExitGames6Common7JStringD2Ev($6);
                       __ZN9ExitGames6Common7JStringD2Ev($7);
                       __ZN9ExitGames6Common7JStringD2Ev($8);
                       __ZN9ExitGames6Common7JStringD2Ev($9);
                       __ZN9ExitGames6Common7JStringD2Ev($10);
                       STACKTOP = sp;return ($65|0);
                      }
                     }
                    } while(0);
                    __ZN9ExitGames6Common7JStringD2Ev($5);
                    $$1 = $$0;$$12 = $$01;
                    break L25;
                   }
                  } while(0);
                  $56 = ___cxa_find_matching_catch_2()|0;
                  $57 = tempRet0;
                  __ZN9ExitGames6Common7JStringD2Ev($str$i20);
                  $eh$lpad$body23$index12Z2D = $57;$eh$lpad$body23$indexZ2D = $56;
                 }
                } while(0);
                if ((label|0) == 25) {
                 $60 = ___cxa_find_matching_catch_2()|0;
                 $61 = tempRet0;
                 $eh$lpad$body23$index12Z2D = $61;$eh$lpad$body23$indexZ2D = $60;
                }
                __ZN9ExitGames6Common7JStringD2Ev($1);
                $eh$lpad$body12$index27Z2D = $eh$lpad$body23$index12Z2D;$eh$lpad$body12$indexZ2D = $eh$lpad$body23$indexZ2D;
                label = 37;
               }
              } while(0);
              if ((label|0) == 37) {
               $$1 = $eh$lpad$body12$index27Z2D;$$12 = $eh$lpad$body12$indexZ2D;
              }
              __ZN9ExitGames6Common7JStringD2Ev($6);
              $$2 = $$1;$$23 = $$12;
             }
             __ZN9ExitGames6Common7JStringD2Ev($7);
             $$3 = $$2;$$34 = $$23;
             break L13;
            }
           }
           $38 = ___cxa_find_matching_catch_2()|0;
           $39 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($str$i16);
           $eh$lpad$body19$index7Z2D = $39;$eh$lpad$body19$indexZ2D = $38;
          }
         } while(0);
         if ((label|0) == 16) {
          $42 = ___cxa_find_matching_catch_2()|0;
          $43 = tempRet0;
          $eh$lpad$body19$index7Z2D = $43;$eh$lpad$body19$indexZ2D = $42;
         }
         __ZN9ExitGames6Common7JStringD2Ev($2);
         $eh$lpad$body9$index22Z2D = $eh$lpad$body19$index7Z2D;$eh$lpad$body9$indexZ2D = $eh$lpad$body19$indexZ2D;
         label = 34;
        }
       } while(0);
       if ((label|0) == 34) {
        $$3 = $eh$lpad$body9$index22Z2D;$$34 = $eh$lpad$body9$indexZ2D;
       }
       __ZN9ExitGames6Common7JStringD2Ev($8);
       $$4 = $$3;$$45 = $$34;
      }
      __ZN9ExitGames6Common7JStringD2Ev($9);
      $$5 = $$4;$$56 = $$45;
      __ZN9ExitGames6Common7JStringD2Ev($10);
      ___resumeException($$56|0);
      // unreachable;
     }
    }
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body15$index2Z2D = $21;$eh$lpad$body15$indexZ2D = $20;
   }
  } while(0);
  if ((label|0) == 7) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   $eh$lpad$body15$index2Z2D = $25;$eh$lpad$body15$indexZ2D = $24;
  }
  __ZN9ExitGames6Common7JStringD2Ev($3);
  $eh$lpad$body$index17Z2D = $eh$lpad$body15$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body15$indexZ2D;
 }
 $$5 = $eh$lpad$body$index17Z2D;$$56 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($10);
 ___resumeException($$56|0);
 // unreachable;
 return (0)|0;
}
function __ZNK9ExitGames6Photon21TrafficStatsGameLevel18toStringVitalStatsEv($agg$result,$this) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$10 = 0, $$1015 = 0, $$11 = 0, $$1116 = 0, $$12 = 0, $$1217 = 0, $$13 = 0, $$1318 = 0, $$14 = 0, $$1419 = 0, $$15 = 0, $$1520 = 0, $$16 = 0, $$2 = 0, $$27 = 0, $$3 = 0, $$38 = 0;
 var $$4 = 0, $$49 = 0, $$5 = 0, $$510 = 0, $$6 = 0, $$611 = 0, $$7 = 0, $$712 = 0, $$8 = 0, $$813 = 0, $$9 = 0, $$914 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $eh$lpad$body$index22Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body23$index27Z2D = 0, $eh$lpad$body23$indexZ2D = 0, $eh$lpad$body26$index32Z2D = 0, $eh$lpad$body26$indexZ2D = 0, $eh$lpad$body29$index37Z2D = 0, $eh$lpad$body29$indexZ2D = 0, $eh$lpad$body32$index2Z2D = 0, $eh$lpad$body32$indexZ2D = 0, $eh$lpad$body36$index7Z2D = 0, $eh$lpad$body36$indexZ2D = 0, $eh$lpad$body40$index12Z2D = 0, $eh$lpad$body40$indexZ2D = 0, $eh$lpad$body44$index17Z2D = 0, $eh$lpad$body44$indexZ2D = 0, $str$i = 0, $str$i33 = 0, $str$i37 = 0;
 var $str$i41 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(400|0);
 $str$i41 = sp + 384|0;
 $str$i37 = sp + 368|0;
 $str$i33 = sp + 352|0;
 $str$i = sp + 336|0;
 $0 = sp + 320|0;
 $1 = sp + 304|0;
 $2 = sp + 288|0;
 $3 = sp + 272|0;
 $4 = sp + 256|0;
 $5 = sp + 240|0;
 $6 = sp + 224|0;
 $7 = sp + 208|0;
 $8 = sp + 192|0;
 $9 = sp + 176|0;
 $10 = sp + 160|0;
 $11 = sp + 144|0;
 $12 = sp + 128|0;
 $13 = sp + 112|0;
 $14 = sp + 96|0;
 $15 = sp + 80|0;
 $16 = sp + 64|0;
 $17 = sp + 48|0;
 $18 = sp + 32|0;
 $19 = sp + 16|0;
 $20 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($20,0);
 __THREW__ = 0;
 invoke_vii(786,($3|0),($20|0));
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $115 = ___cxa_find_matching_catch_2()|0;
  $116 = tempRet0;
  $eh$lpad$body$index22Z2D = $116;$eh$lpad$body$indexZ2D = $115;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  do {
   if ($24) {
    label = 7;
   } else {
    __THREW__ = 0;
    (invoke_iii(759,($str$i|0),(12768|0))|0);
    $25 = __THREW__; __THREW__ = 0;
    $26 = $25&1;
    if (!($26)) {
     __THREW__ = 0;
     $27 = (invoke_iii(760,($3|0),($str$i|0))|0);
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if (!($29)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($19|0),($27|0));
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($3);
      $36 = ((($this)) + 56|0);
      __THREW__ = 0;
      invoke_viii(927,($18|0),($19|0),($36|0));
      $37 = __THREW__; __THREW__ = 0;
      $38 = $37&1;
      if ($38) {
       $117 = ___cxa_find_matching_catch_2()|0;
       $118 = tempRet0;
       $$14 = $118;$$1419 = $117;
      } else {
       __THREW__ = 0;
       invoke_vii(786,($2|0),($18|0));
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       L13: do {
        if ($40) {
         $119 = ___cxa_find_matching_catch_2()|0;
         $120 = tempRet0;
         $eh$lpad$body23$index27Z2D = $120;$eh$lpad$body23$indexZ2D = $119;
         label = 51;
        } else {
         __THREW__ = 0;
         invoke_vii(787,($str$i33|0),0);
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         do {
          if ($42) {
           label = 16;
          } else {
           __THREW__ = 0;
           (invoke_iii(759,($str$i33|0),(12884|0))|0);
           $43 = __THREW__; __THREW__ = 0;
           $44 = $43&1;
           if (!($44)) {
            __THREW__ = 0;
            $45 = (invoke_iii(760,($2|0),($str$i33|0))|0);
            $46 = __THREW__; __THREW__ = 0;
            $47 = $46&1;
            if (!($47)) {
             __ZN9ExitGames6Common7JStringD2Ev($str$i33);
             __THREW__ = 0;
             invoke_vii(786,($17|0),($45|0));
             $50 = __THREW__; __THREW__ = 0;
             $51 = $50&1;
             if ($51) {
              label = 16;
              break;
             }
             __ZN9ExitGames6Common7JStringD2Ev($2);
             $54 = ((($this)) + 52|0);
             __THREW__ = 0;
             invoke_viii(927,($16|0),($17|0),($54|0));
             $55 = __THREW__; __THREW__ = 0;
             $56 = $55&1;
             if ($56) {
              $121 = ___cxa_find_matching_catch_2()|0;
              $122 = tempRet0;
              $$12 = $122;$$1217 = $121;
             } else {
              __THREW__ = 0;
              invoke_vii(786,($1|0),($16|0));
              $57 = __THREW__; __THREW__ = 0;
              $58 = $57&1;
              L25: do {
               if ($58) {
                $123 = ___cxa_find_matching_catch_2()|0;
                $124 = tempRet0;
                $eh$lpad$body26$index32Z2D = $124;$eh$lpad$body26$indexZ2D = $123;
                label = 54;
               } else {
                __THREW__ = 0;
                invoke_vii(787,($str$i37|0),0);
                $59 = __THREW__; __THREW__ = 0;
                $60 = $59&1;
                L28: do {
                 if ($60) {
                  label = 25;
                 } else {
                  __THREW__ = 0;
                  (invoke_iii(759,($str$i37|0),(12940|0))|0);
                  $61 = __THREW__; __THREW__ = 0;
                  $62 = $61&1;
                  do {
                   if (!($62)) {
                    __THREW__ = 0;
                    $63 = (invoke_iii(760,($1|0),($str$i37|0))|0);
                    $64 = __THREW__; __THREW__ = 0;
                    $65 = $64&1;
                    if ($65) {
                     break;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($str$i37);
                    __THREW__ = 0;
                    invoke_vii(786,($15|0),($63|0));
                    $68 = __THREW__; __THREW__ = 0;
                    $69 = $68&1;
                    if ($69) {
                     label = 25;
                     break L28;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($1);
                    $72 = ((($this)) + 48|0);
                    __THREW__ = 0;
                    invoke_viii(866,($14|0),($15|0),($72|0));
                    $73 = __THREW__; __THREW__ = 0;
                    $74 = $73&1;
                    if ($74) {
                     $125 = ___cxa_find_matching_catch_2()|0;
                     $126 = tempRet0;
                     $$10 = $126;$$1015 = $125;
                    } else {
                     __THREW__ = 0;
                     invoke_viii(788,($13|0),($14|0),(13052|0));
                     $75 = __THREW__; __THREW__ = 0;
                     $76 = $75&1;
                     if ($76) {
                      $127 = ___cxa_find_matching_catch_2()|0;
                      $128 = tempRet0;
                      $$9 = $128;$$914 = $127;
                     } else {
                      $77 = ((($this)) + 44|0);
                      __THREW__ = 0;
                      invoke_viii(927,($12|0),($13|0),($77|0));
                      $78 = __THREW__; __THREW__ = 0;
                      $79 = $78&1;
                      if ($79) {
                       $129 = ___cxa_find_matching_catch_2()|0;
                       $130 = tempRet0;
                       $$8 = $130;$$813 = $129;
                      } else {
                       __THREW__ = 0;
                       invoke_viii(928,($11|0),($12|0),(13060|0));
                       $80 = __THREW__; __THREW__ = 0;
                       $81 = $80&1;
                       if ($81) {
                        $131 = ___cxa_find_matching_catch_2()|0;
                        $132 = tempRet0;
                        $$7 = $132;$$712 = $131;
                       } else {
                        $82 = ((($this)) + 40|0);
                        __THREW__ = 0;
                        invoke_viii(866,($10|0),($11|0),($82|0));
                        $83 = __THREW__; __THREW__ = 0;
                        $84 = $83&1;
                        if ($84) {
                         $133 = ___cxa_find_matching_catch_2()|0;
                         $134 = tempRet0;
                         $$6 = $134;$$611 = $133;
                        } else {
                         __THREW__ = 0;
                         invoke_viii(788,($9|0),($10|0),(13052|0));
                         $85 = __THREW__; __THREW__ = 0;
                         $86 = $85&1;
                         if ($86) {
                          $135 = ___cxa_find_matching_catch_2()|0;
                          $136 = tempRet0;
                          $$5 = $136;$$510 = $135;
                         } else {
                          $87 = ((($this)) + 36|0);
                          __THREW__ = 0;
                          invoke_viii(927,($8|0),($9|0),($87|0));
                          $88 = __THREW__; __THREW__ = 0;
                          $89 = $88&1;
                          if ($89) {
                           $137 = ___cxa_find_matching_catch_2()|0;
                           $138 = tempRet0;
                           $$4 = $138;$$49 = $137;
                          } else {
                           __THREW__ = 0;
                           invoke_vii(786,($0|0),($8|0));
                           $90 = __THREW__; __THREW__ = 0;
                           $91 = $90&1;
                           L55: do {
                            if ($91) {
                             $139 = ___cxa_find_matching_catch_2()|0;
                             $140 = tempRet0;
                             $eh$lpad$body29$index37Z2D = $140;$eh$lpad$body29$indexZ2D = $139;
                             label = 63;
                            } else {
                             __THREW__ = 0;
                             invoke_vii(787,($str$i41|0),0);
                             $92 = __THREW__; __THREW__ = 0;
                             $93 = $92&1;
                             L58: do {
                              if ($93) {
                               label = 40;
                              } else {
                               __THREW__ = 0;
                               (invoke_iii(759,($str$i41|0),(13108|0))|0);
                               $94 = __THREW__; __THREW__ = 0;
                               $95 = $94&1;
                               do {
                                if (!($95)) {
                                 __THREW__ = 0;
                                 $96 = (invoke_iii(760,($0|0),($str$i41|0))|0);
                                 $97 = __THREW__; __THREW__ = 0;
                                 $98 = $97&1;
                                 if ($98) {
                                  break;
                                 }
                                 __ZN9ExitGames6Common7JStringD2Ev($str$i41);
                                 __THREW__ = 0;
                                 invoke_vii(786,($7|0),($96|0));
                                 $101 = __THREW__; __THREW__ = 0;
                                 $102 = $101&1;
                                 if ($102) {
                                  label = 40;
                                  break L58;
                                 }
                                 __ZN9ExitGames6Common7JStringD2Ev($0);
                                 $105 = ((($this)) + 64|0);
                                 __THREW__ = 0;
                                 invoke_viii(927,($6|0),($7|0),($105|0));
                                 $106 = __THREW__; __THREW__ = 0;
                                 $107 = $106&1;
                                 if ($107) {
                                  $141 = ___cxa_find_matching_catch_2()|0;
                                  $142 = tempRet0;
                                  $$2 = $142;$$27 = $141;
                                 } else {
                                  __THREW__ = 0;
                                  invoke_viii(928,($5|0),($6|0),(13188|0));
                                  $108 = __THREW__; __THREW__ = 0;
                                  $109 = $108&1;
                                  if ($109) {
                                   $143 = ___cxa_find_matching_catch_2()|0;
                                   $144 = tempRet0;
                                   $$1 = $144;$$16 = $143;
                                  } else {
                                   $110 = ((($this)) + 60|0);
                                   __THREW__ = 0;
                                   invoke_viii(927,($4|0),($5|0),($110|0));
                                   $111 = __THREW__; __THREW__ = 0;
                                   $112 = $111&1;
                                   do {
                                    if ($112) {
                                     $145 = ___cxa_find_matching_catch_2()|0;
                                     $146 = tempRet0;
                                     $$0 = $146;$$01 = $145;
                                    } else {
                                     __THREW__ = 0;
                                     invoke_viii(788,($agg$result|0),($4|0),(13236|0));
                                     $113 = __THREW__; __THREW__ = 0;
                                     $114 = $113&1;
                                     if ($114) {
                                      $147 = ___cxa_find_matching_catch_2()|0;
                                      $148 = tempRet0;
                                      __ZN9ExitGames6Common7JStringD2Ev($4);
                                      $$0 = $148;$$01 = $147;
                                      break;
                                     } else {
                                      __ZN9ExitGames6Common7JStringD2Ev($4);
                                      __ZN9ExitGames6Common7JStringD2Ev($5);
                                      __ZN9ExitGames6Common7JStringD2Ev($6);
                                      __ZN9ExitGames6Common7JStringD2Ev($7);
                                      __ZN9ExitGames6Common7JStringD2Ev($8);
                                      __ZN9ExitGames6Common7JStringD2Ev($9);
                                      __ZN9ExitGames6Common7JStringD2Ev($10);
                                      __ZN9ExitGames6Common7JStringD2Ev($11);
                                      __ZN9ExitGames6Common7JStringD2Ev($12);
                                      __ZN9ExitGames6Common7JStringD2Ev($13);
                                      __ZN9ExitGames6Common7JStringD2Ev($14);
                                      __ZN9ExitGames6Common7JStringD2Ev($15);
                                      __ZN9ExitGames6Common7JStringD2Ev($16);
                                      __ZN9ExitGames6Common7JStringD2Ev($17);
                                      __ZN9ExitGames6Common7JStringD2Ev($18);
                                      __ZN9ExitGames6Common7JStringD2Ev($19);
                                      __ZN9ExitGames6Common7JStringD2Ev($20);
                                      STACKTOP = sp;return;
                                     }
                                    }
                                   } while(0);
                                   __ZN9ExitGames6Common7JStringD2Ev($5);
                                   $$1 = $$0;$$16 = $$01;
                                  }
                                  __ZN9ExitGames6Common7JStringD2Ev($6);
                                  $$2 = $$1;$$27 = $$16;
                                 }
                                 __ZN9ExitGames6Common7JStringD2Ev($7);
                                 $$3 = $$2;$$38 = $$27;
                                 break L55;
                                }
                               } while(0);
                               $99 = ___cxa_find_matching_catch_2()|0;
                               $100 = tempRet0;
                               __ZN9ExitGames6Common7JStringD2Ev($str$i41);
                               $eh$lpad$body44$index17Z2D = $100;$eh$lpad$body44$indexZ2D = $99;
                              }
                             } while(0);
                             if ((label|0) == 40) {
                              $103 = ___cxa_find_matching_catch_2()|0;
                              $104 = tempRet0;
                              $eh$lpad$body44$index17Z2D = $104;$eh$lpad$body44$indexZ2D = $103;
                             }
                             __ZN9ExitGames6Common7JStringD2Ev($0);
                             $eh$lpad$body29$index37Z2D = $eh$lpad$body44$index17Z2D;$eh$lpad$body29$indexZ2D = $eh$lpad$body44$indexZ2D;
                             label = 63;
                            }
                           } while(0);
                           if ((label|0) == 63) {
                            $$3 = $eh$lpad$body29$index37Z2D;$$38 = $eh$lpad$body29$indexZ2D;
                           }
                           __ZN9ExitGames6Common7JStringD2Ev($8);
                           $$4 = $$3;$$49 = $$38;
                          }
                          __ZN9ExitGames6Common7JStringD2Ev($9);
                          $$5 = $$4;$$510 = $$49;
                         }
                         __ZN9ExitGames6Common7JStringD2Ev($10);
                         $$6 = $$5;$$611 = $$510;
                        }
                        __ZN9ExitGames6Common7JStringD2Ev($11);
                        $$7 = $$6;$$712 = $$611;
                       }
                       __ZN9ExitGames6Common7JStringD2Ev($12);
                       $$8 = $$7;$$813 = $$712;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($13);
                      $$9 = $$8;$$914 = $$813;
                     }
                     __ZN9ExitGames6Common7JStringD2Ev($14);
                     $$10 = $$9;$$1015 = $$914;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($15);
                    $$11 = $$10;$$1116 = $$1015;
                    break L25;
                   }
                  } while(0);
                  $66 = ___cxa_find_matching_catch_2()|0;
                  $67 = tempRet0;
                  __ZN9ExitGames6Common7JStringD2Ev($str$i37);
                  $eh$lpad$body40$index12Z2D = $67;$eh$lpad$body40$indexZ2D = $66;
                 }
                } while(0);
                if ((label|0) == 25) {
                 $70 = ___cxa_find_matching_catch_2()|0;
                 $71 = tempRet0;
                 $eh$lpad$body40$index12Z2D = $71;$eh$lpad$body40$indexZ2D = $70;
                }
                __ZN9ExitGames6Common7JStringD2Ev($1);
                $eh$lpad$body26$index32Z2D = $eh$lpad$body40$index12Z2D;$eh$lpad$body26$indexZ2D = $eh$lpad$body40$indexZ2D;
                label = 54;
               }
              } while(0);
              if ((label|0) == 54) {
               $$11 = $eh$lpad$body26$index32Z2D;$$1116 = $eh$lpad$body26$indexZ2D;
              }
              __ZN9ExitGames6Common7JStringD2Ev($16);
              $$12 = $$11;$$1217 = $$1116;
             }
             __ZN9ExitGames6Common7JStringD2Ev($17);
             $$13 = $$12;$$1318 = $$1217;
             break L13;
            }
           }
           $48 = ___cxa_find_matching_catch_2()|0;
           $49 = tempRet0;
           __ZN9ExitGames6Common7JStringD2Ev($str$i33);
           $eh$lpad$body36$index7Z2D = $49;$eh$lpad$body36$indexZ2D = $48;
          }
         } while(0);
         if ((label|0) == 16) {
          $52 = ___cxa_find_matching_catch_2()|0;
          $53 = tempRet0;
          $eh$lpad$body36$index7Z2D = $53;$eh$lpad$body36$indexZ2D = $52;
         }
         __ZN9ExitGames6Common7JStringD2Ev($2);
         $eh$lpad$body23$index27Z2D = $eh$lpad$body36$index7Z2D;$eh$lpad$body23$indexZ2D = $eh$lpad$body36$indexZ2D;
         label = 51;
        }
       } while(0);
       if ((label|0) == 51) {
        $$13 = $eh$lpad$body23$index27Z2D;$$1318 = $eh$lpad$body23$indexZ2D;
       }
       __ZN9ExitGames6Common7JStringD2Ev($18);
       $$14 = $$13;$$1419 = $$1318;
      }
      __ZN9ExitGames6Common7JStringD2Ev($19);
      $$15 = $$14;$$1520 = $$1419;
      __ZN9ExitGames6Common7JStringD2Ev($20);
      ___resumeException($$1520|0);
      // unreachable;
     }
    }
    $30 = ___cxa_find_matching_catch_2()|0;
    $31 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body32$index2Z2D = $31;$eh$lpad$body32$indexZ2D = $30;
   }
  } while(0);
  if ((label|0) == 7) {
   $34 = ___cxa_find_matching_catch_2()|0;
   $35 = tempRet0;
   $eh$lpad$body32$index2Z2D = $35;$eh$lpad$body32$indexZ2D = $34;
  }
  __ZN9ExitGames6Common7JStringD2Ev($3);
  $eh$lpad$body$index22Z2D = $eh$lpad$body32$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body32$indexZ2D;
 }
 $$15 = $eh$lpad$body$index22Z2D;$$1520 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($20);
 ___resumeException($$1520|0);
 // unreachable;
}
function __ZN9ExitGames6CommonplIA12_wEENS0_7JStringERKS3_RKT_($agg$result,$Lsh,$Rsh) {
 $agg$result = $agg$result|0;
 $Lsh = $Lsh|0;
 $Rsh = $Rsh|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, $str$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $str$i = sp + 16|0;
 $0 = sp;
 __ZN9ExitGames6Common7JStringC2ERKS1_($0,$Lsh);
 __THREW__ = 0;
 invoke_vii(787,($str$i|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 do {
  if (!($2)) {
   __THREW__ = 0;
   (invoke_iii(759,($str$i|0),($Rsh|0))|0);
   $3 = __THREW__; __THREW__ = 0;
   $4 = $3&1;
   if (!($4)) {
    __THREW__ = 0;
    $5 = (invoke_iii(760,($0|0),($str$i|0))|0);
    $6 = __THREW__; __THREW__ = 0;
    $7 = $6&1;
    if (!($7)) {
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     __THREW__ = 0;
     invoke_vii(786,($agg$result|0),($5|0));
     $10 = __THREW__; __THREW__ = 0;
     $11 = $10&1;
     if ($11) {
      break;
     }
     __ZN9ExitGames6Common7JStringD2Ev($0);
     STACKTOP = sp;return;
    }
   }
   $8 = ___cxa_find_matching_catch_2()|0;
   $9 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($str$i);
   $eh$lpad$body$index2Z2D = $9;$eh$lpad$body$indexZ2D = $8;
   __ZN9ExitGames6Common7JStringD2Ev($0);
   ___resumeException($eh$lpad$body$indexZ2D|0);
   // unreachable;
  }
 } while(0);
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 $eh$lpad$body$index2Z2D = $13;$eh$lpad$body$indexZ2D = $12;
 __ZN9ExitGames6Common7JStringD2Ev($0);
 ___resumeException($eh$lpad$body$indexZ2D|0);
 // unreachable;
}
function __ZN9ExitGames6Photon21TrafficStatsGameLevelD2Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeer7connectERKNS_6Common7JStringES5_($this,$ipAddr,$appID) {
 $this = $this|0;
 $ipAddr = $ipAddr|0;
 $appID = $appID|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $0 = sp;
 $1 = HEAP32[$this>>2]|0;
 $2 = ((($1)) + 64|0);
 $3 = HEAP32[$2>>2]|0;
 __ZN9ExitGames6Common6ObjectC2Ev($0);
 __THREW__ = 0;
 $4 = (invoke_iiiii($3|0,($this|0),($ipAddr|0),($appID|0),($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $7 = ___cxa_find_matching_catch_2()|0;
  $8 = tempRet0;
  __ZN9ExitGames6Common6ObjectD2Ev($0);
  ___resumeException($7|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common6ObjectD2Ev($0);
  STACKTOP = sp;return ($4|0);
 }
 return (0)|0;
}
function __ZN9ExitGames6Photon10PhotonPeer10disconnectEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $0 = sp + 8|0;
 __ZN9ExitGames6Common7JStringC2EPKc($0,147875);
 __THREW__ = 0;
 $1 = (invoke_ii(764,($0|0))|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($12|0);
  // unreachable;
 }
 $4 = ((($this)) + 4|0);
 __THREW__ = 0;
 invoke_viiiiiiii(765,($4|0),4,(13968|0),($1|0),1,145,(163772|0),($vararg_buffer|0));
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $12 = ___cxa_find_matching_catch_2()|0;
  $13 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  ___resumeException($12|0);
  // unreachable;
 } else {
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $7 = ((($this)) + 20|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 24|0);
  $11 = HEAP32[$10>>2]|0;
  FUNCTION_TABLE_vi[$11 & 2047]($8);
  STACKTOP = sp;return;
 }
}
function __ZN9ExitGames6Photon10PhotonPeer7serviceEb($this,$dispatchIncomingCommands) {
 $this = $this|0;
 $dispatchIncomingCommands = $dispatchIncomingCommands|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($this)) + 24|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) + 4)|0;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 $10 = (_i64Add(($6|0),($9|0),1,0)|0);
 $11 = tempRet0;
 $12 = $3;
 $13 = $12;
 HEAP32[$13>>2] = $10;
 $14 = (($12) + 4)|0;
 $15 = $14;
 HEAP32[$15>>2] = $11;
 $16 = $1;
 $17 = $1;
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($18)) + 28|0);
 $20 = HEAP32[$19>>2]|0;
 __THREW__ = 0;
 invoke_vii($20|0,($16|0),($dispatchIncomingCommands|0));
 $21 = __THREW__; __THREW__ = 0;
 $22 = $21&1;
 if ($22) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  $51 = $3;
  $52 = $51;
  $53 = HEAP32[$52>>2]|0;
  $54 = (($51) + 4)|0;
  $55 = $54;
  $56 = HEAP32[$55>>2]|0;
  $57 = (_i64Add(($53|0),($56|0),-1,-1)|0);
  $58 = tempRet0;
  $59 = $3;
  $60 = $59;
  HEAP32[$60>>2] = $57;
  $61 = (($59) + 4)|0;
  $62 = $61;
  HEAP32[$62>>2] = $58;
  $63 = ($57|0)==(0);
  $64 = ($58|0)==(0);
  $65 = $63 & $64;
  if (!($65)) {
   ___resumeException($49|0);
   // unreachable;
  }
  $66 = ($1|0)==(0);
  if (!($66)) {
   $67 = $1;
   $68 = HEAP32[$67>>2]|0;
   $69 = HEAP32[$68>>2]|0;
   FUNCTION_TABLE_vi[$69 & 2047]($16);
   $70 = $1;
   __THREW__ = 0;
   invoke_vi(750,($70|0));
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $75 = ___cxa_find_matching_catch_3(0|0)|0;
    $76 = tempRet0;
    ___clang_call_terminate($75);
    // unreachable;
   }
  }
  __THREW__ = 0;
  invoke_vi(750,($3|0));
  $73 = __THREW__; __THREW__ = 0;
  $74 = $73&1;
  if ($74) {
   $75 = ___cxa_find_matching_catch_3(0|0)|0;
   $76 = tempRet0;
   ___clang_call_terminate($75);
   // unreachable;
  } else {
   ___resumeException($49|0);
   // unreachable;
  }
 } else {
  $23 = $3;
  $24 = $23;
  $25 = HEAP32[$24>>2]|0;
  $26 = (($23) + 4)|0;
  $27 = $26;
  $28 = HEAP32[$27>>2]|0;
  $29 = (_i64Add(($25|0),($28|0),-1,-1)|0);
  $30 = tempRet0;
  $31 = $3;
  $32 = $31;
  HEAP32[$32>>2] = $29;
  $33 = (($31) + 4)|0;
  $34 = $33;
  HEAP32[$34>>2] = $30;
  $35 = ($29|0)==(0);
  $36 = ($30|0)==(0);
  $37 = $35 & $36;
  if (!($37)) {
   return;
  }
  $38 = ($1|0)==(0);
  if (!($38)) {
   $39 = $1;
   $40 = HEAP32[$39>>2]|0;
   $41 = HEAP32[$40>>2]|0;
   FUNCTION_TABLE_vi[$41 & 2047]($16);
   $42 = $1;
   __THREW__ = 0;
   invoke_vi(750,($42|0));
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if ($44) {
    $47 = ___cxa_find_matching_catch_3(0|0)|0;
    $48 = tempRet0;
    ___clang_call_terminate($47);
    // unreachable;
   }
  }
  __THREW__ = 0;
  invoke_vi(750,($3|0));
  $45 = __THREW__; __THREW__ = 0;
  $46 = $45&1;
  if ($46) {
   $47 = ___cxa_find_matching_catch_3(0|0)|0;
   $48 = tempRet0;
   ___clang_call_terminate($47);
   // unreachable;
  } else {
   return;
  }
 }
}
function __ZN9ExitGames6Photon10PhotonPeer12serviceBasicEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 32|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vi[$4 & 2047]($1);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeer8opCustomERKNS0_16OperationRequestEbhb($this,$operationRequest,$sendReliable,$channelID,$encrypt) {
 $this = $this|0;
 $operationRequest = $operationRequest|0;
 $sendReliable = $sendReliable|0;
 $channelID = $channelID|0;
 $encrypt = $encrypt|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 48|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_iiiiiii[$4 & 1023]($1,$operationRequest,$sendReliable,$channelID,$encrypt,2)|0);
 return ($5|0);
}
function __ZN9ExitGames6Photon10PhotonPeer20sendOutgoingCommandsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 36|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 2047]($1)|0);
 return ($5|0);
}
function __ZN9ExitGames6Photon10PhotonPeer12sendAcksOnlyEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 40|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 2047]($1)|0);
 return ($5|0);
}
function __ZN9ExitGames6Photon10PhotonPeer24dispatchIncomingCommandsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 44|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 2047]($1)|0);
 return ($5|0);
}
function __ZN9ExitGames6Photon10PhotonPeer19establishEncryptionEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 56|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (FUNCTION_TABLE_ii[$4 & 2047]($1)|0);
 return ($5|0);
}
function __ZN9ExitGames6Photon10PhotonPeer20fetchServerTimestampEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($2)) + 52|0);
 $4 = HEAP32[$3>>2]|0;
 FUNCTION_TABLE_vi[$4 & 2047]($1);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeer17resetTrafficStatsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 __ZN9ExitGames6Photon8Internal8PeerBase17resetTrafficStatsEv($1);
 $2 = ((($this)) + 12|0);
 $3 = HEAP32[$2>>2]|0;
 __ZN9ExitGames6Photon8Internal8PeerData22setTrafficStatsEnabledEb($3,1);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeer32resetTrafficStatsMaximumCountersEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 __ZN9ExitGames6Photon8Internal8PeerBase32resetTrafficStatsMaximumCountersEv($1);
 return;
}
function __ZNK9ExitGames6Photon10PhotonPeer18vitalStatsToStringEb($agg$result,$this,$all) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $all = $all|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$10 = 0, $$1016 = 0, $$11 = 0, $$1117 = 0, $$12 = 0, $$1218 = 0, $$13 = 0, $$1319 = 0, $$14 = 0, $$1420 = 0, $$15 = 0, $$1521 = 0, $$16 = 0, $$1622 = 0, $$17 = 0, $$2 = 0, $$28 = 0;
 var $$3 = 0, $$39 = 0, $$4 = 0, $$410 = 0, $$5 = 0, $$511 = 0, $$6 = 0, $$612 = 0, $$7 = 0, $$713 = 0, $$8 = 0, $$814 = 0, $$9 = 0, $$915 = 0, $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0;
 var $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index37Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body26$index42Z2D = 0, $eh$lpad$body26$indexZ2D = 0, $eh$lpad$body30$index47Z2D = 0, $eh$lpad$body30$indexZ2D = 0, $eh$lpad$body34$index52Z2D = 0, $eh$lpad$body34$indexZ2D = 0, $eh$lpad$body38$index57Z2D = 0, $eh$lpad$body38$indexZ2D = 0, $eh$lpad$body42$index62Z2D = 0, $eh$lpad$body42$indexZ2D = 0, $eh$lpad$body46$index67Z2D = 0, $eh$lpad$body46$indexZ2D = 0, $eh$lpad$body50$index2Z2D = 0, $eh$lpad$body50$indexZ2D = 0, $eh$lpad$body54$index7Z2D = 0;
 var $eh$lpad$body54$indexZ2D = 0, $eh$lpad$body58$index12Z2D = 0, $eh$lpad$body58$indexZ2D = 0, $eh$lpad$body62$index17Z2D = 0, $eh$lpad$body62$indexZ2D = 0, $eh$lpad$body66$index22Z2D = 0, $eh$lpad$body66$indexZ2D = 0, $eh$lpad$body70$index27Z2D = 0, $eh$lpad$body70$indexZ2D = 0, $eh$lpad$body74$index32Z2D = 0, $eh$lpad$body74$indexZ2D = 0, $str$i = 0, $str$i51 = 0, $str$i55 = 0, $str$i59 = 0, $str$i63 = 0, $str$i67 = 0, $str$i71 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(512|0);
 $str$i71 = sp + 496|0;
 $str$i67 = sp + 480|0;
 $str$i63 = sp + 464|0;
 $str$i59 = sp + 448|0;
 $str$i55 = sp + 432|0;
 $str$i51 = sp + 416|0;
 $str$i = sp + 400|0;
 $0 = sp + 384|0;
 $1 = sp + 368|0;
 $2 = sp + 352|0;
 $3 = sp + 336|0;
 $4 = sp + 320|0;
 $5 = sp + 304|0;
 $6 = sp + 288|0;
 $7 = sp + 272|0;
 $8 = sp + 256|0;
 $9 = sp + 240|0;
 $10 = sp + 224|0;
 $11 = sp + 208|0;
 $12 = sp + 192|0;
 $13 = sp + 176|0;
 $14 = sp + 160|0;
 $15 = sp + 144|0;
 $16 = sp + 128|0;
 $17 = sp + 124|0;
 $18 = sp + 120|0;
 $19 = sp + 116|0;
 $20 = sp + 112|0;
 $21 = sp + 96|0;
 $22 = sp + 80|0;
 $23 = sp + 64|0;
 $24 = sp + 48|0;
 $25 = sp + 32|0;
 $26 = sp + 16|0;
 $27 = sp;
 __ZN9ExitGames6Common7JStringC2Ej($16,0);
 __THREW__ = 0;
 invoke_vii(786,($6|0),($16|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 L1: do {
  if ($29) {
   $193 = ___cxa_find_matching_catch_2()|0;
   $194 = tempRet0;
   $eh$lpad$body$index37Z2D = $194;$eh$lpad$body$indexZ2D = $193;
   label = 79;
  } else {
   __THREW__ = 0;
   invoke_vii(787,($str$i|0),0);
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   do {
    if ($31) {
     label = 7;
    } else {
     __THREW__ = 0;
     (invoke_iii(759,($str$i|0),(13596|0))|0);
     $32 = __THREW__; __THREW__ = 0;
     $33 = $32&1;
     if (!($33)) {
      __THREW__ = 0;
      $34 = (invoke_iii(760,($6|0),($str$i|0))|0);
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if (!($36)) {
       __ZN9ExitGames6Common7JStringD2Ev($str$i);
       __THREW__ = 0;
       invoke_vii(786,($15|0),($34|0));
       $39 = __THREW__; __THREW__ = 0;
       $40 = $39&1;
       if ($40) {
        label = 7;
        break;
       }
       __ZN9ExitGames6Common7JStringD2Ev($6);
       $43 = ((($this)) + 20|0);
       $44 = HEAP32[$43>>2]|0;
       __THREW__ = 0;
       $45 = (invoke_ii(929,($44|0))|0);
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        label = 80;
       } else {
        HEAP32[$17>>2] = $45;
        __THREW__ = 0;
        invoke_viii(927,($14|0),($15|0),($17|0));
        $48 = __THREW__; __THREW__ = 0;
        $49 = $48&1;
        if ($49) {
         label = 80;
        } else {
         __THREW__ = 0;
         invoke_vii(786,($5|0),($14|0));
         $50 = __THREW__; __THREW__ = 0;
         $51 = $50&1;
         L13: do {
          if ($51) {
           $197 = ___cxa_find_matching_catch_2()|0;
           $198 = tempRet0;
           $eh$lpad$body26$index42Z2D = $198;$eh$lpad$body26$indexZ2D = $197;
           label = 82;
          } else {
           __THREW__ = 0;
           invoke_vii(787,($str$i51|0),0);
           $52 = __THREW__; __THREW__ = 0;
           $53 = $52&1;
           do {
            if ($53) {
             label = 17;
            } else {
             __THREW__ = 0;
             (invoke_iii(759,($str$i51|0),(13660|0))|0);
             $54 = __THREW__; __THREW__ = 0;
             $55 = $54&1;
             if (!($55)) {
              __THREW__ = 0;
              $56 = (invoke_iii(760,($5|0),($str$i51|0))|0);
              $57 = __THREW__; __THREW__ = 0;
              $58 = $57&1;
              if (!($58)) {
               __ZN9ExitGames6Common7JStringD2Ev($str$i51);
               __THREW__ = 0;
               invoke_vii(786,($13|0),($56|0));
               $61 = __THREW__; __THREW__ = 0;
               $62 = $61&1;
               if ($62) {
                label = 17;
                break;
               }
               __ZN9ExitGames6Common7JStringD2Ev($5);
               $65 = HEAP32[$43>>2]|0;
               __THREW__ = 0;
               $66 = (invoke_ii(930,($65|0))|0);
               $67 = __THREW__; __THREW__ = 0;
               $68 = $67&1;
               do {
                if ($68) {
                 label = 83;
                } else {
                 HEAP32[$18>>2] = $66;
                 __THREW__ = 0;
                 invoke_viii(927,($12|0),($13|0),($18|0));
                 $69 = __THREW__; __THREW__ = 0;
                 $70 = $69&1;
                 if ($70) {
                  label = 83;
                  break;
                 }
                 __THREW__ = 0;
                 invoke_vii(786,($4|0),($12|0));
                 $71 = __THREW__; __THREW__ = 0;
                 $72 = $71&1;
                 L25: do {
                  if ($72) {
                   $201 = ___cxa_find_matching_catch_2()|0;
                   $202 = tempRet0;
                   $eh$lpad$body30$index47Z2D = $202;$eh$lpad$body30$indexZ2D = $201;
                   label = 85;
                  } else {
                   __THREW__ = 0;
                   invoke_vii(787,($str$i55|0),0);
                   $73 = __THREW__; __THREW__ = 0;
                   $74 = $73&1;
                   L28: do {
                    if ($74) {
                     label = 27;
                    } else {
                     __THREW__ = 0;
                     (invoke_iii(759,($str$i55|0),(13668|0))|0);
                     $75 = __THREW__; __THREW__ = 0;
                     $76 = $75&1;
                     do {
                      if (!($76)) {
                       __THREW__ = 0;
                       $77 = (invoke_iii(760,($4|0),($str$i55|0))|0);
                       $78 = __THREW__; __THREW__ = 0;
                       $79 = $78&1;
                       if ($79) {
                        break;
                       }
                       __ZN9ExitGames6Common7JStringD2Ev($str$i55);
                       __THREW__ = 0;
                       invoke_vii(786,($11|0),($77|0));
                       $82 = __THREW__; __THREW__ = 0;
                       $83 = $82&1;
                       if ($83) {
                        label = 27;
                        break L28;
                       }
                       __ZN9ExitGames6Common7JStringD2Ev($4);
                       __THREW__ = 0;
                       $86 = (invoke_i(823)|0);
                       $87 = __THREW__; __THREW__ = 0;
                       $88 = $87&1;
                       do {
                        if ($88) {
                         label = 86;
                        } else {
                         $89 = HEAP32[$43>>2]|0;
                         __THREW__ = 0;
                         $90 = (invoke_ii(931,($89|0))|0);
                         $91 = __THREW__; __THREW__ = 0;
                         $92 = $91&1;
                         if ($92) {
                          label = 86;
                          break;
                         }
                         $93 = (($86) - ($90))|0;
                         HEAP32[$19>>2] = $93;
                         __THREW__ = 0;
                         invoke_viii(927,($10|0),($11|0),($19|0));
                         $94 = __THREW__; __THREW__ = 0;
                         $95 = $94&1;
                         if ($95) {
                          label = 86;
                          break;
                         }
                         __THREW__ = 0;
                         invoke_vii(786,($3|0),($10|0));
                         $96 = __THREW__; __THREW__ = 0;
                         $97 = $96&1;
                         L38: do {
                          if ($97) {
                           $205 = ___cxa_find_matching_catch_2()|0;
                           $206 = tempRet0;
                           $eh$lpad$body34$index52Z2D = $206;$eh$lpad$body34$indexZ2D = $205;
                           label = 88;
                          } else {
                           __THREW__ = 0;
                           invoke_vii(787,($str$i59|0),0);
                           $98 = __THREW__; __THREW__ = 0;
                           $99 = $98&1;
                           L41: do {
                            if ($99) {
                             label = 38;
                            } else {
                             __THREW__ = 0;
                             (invoke_iii(759,($str$i59|0),(13776|0))|0);
                             $100 = __THREW__; __THREW__ = 0;
                             $101 = $100&1;
                             do {
                              if (!($101)) {
                               __THREW__ = 0;
                               $102 = (invoke_iii(760,($3|0),($str$i59|0))|0);
                               $103 = __THREW__; __THREW__ = 0;
                               $104 = $103&1;
                               if ($104) {
                                break;
                               }
                               __ZN9ExitGames6Common7JStringD2Ev($str$i59);
                               __THREW__ = 0;
                               invoke_vii(786,($9|0),($102|0));
                               $107 = __THREW__; __THREW__ = 0;
                               $108 = $107&1;
                               if ($108) {
                                label = 38;
                                break L41;
                               }
                               __ZN9ExitGames6Common7JStringD2Ev($3);
                               $111 = ((($this)) + 12|0);
                               $112 = HEAP32[$111>>2]|0;
                               __THREW__ = 0;
                               $113 = (invoke_ii(932,($112|0))|0);
                               $114 = __THREW__; __THREW__ = 0;
                               $115 = $114&1;
                               do {
                                if ($115) {
                                 label = 89;
                                } else {
                                 $116 = (($113|0) / 1000)&-1;
                                 HEAP32[$20>>2] = $116;
                                 __THREW__ = 0;
                                 invoke_viii(927,($8|0),($9|0),($20|0));
                                 $117 = __THREW__; __THREW__ = 0;
                                 $118 = $117&1;
                                 if ($118) {
                                  label = 89;
                                  break;
                                 }
                                 __THREW__ = 0;
                                 invoke_vii(786,($2|0),($8|0));
                                 $119 = __THREW__; __THREW__ = 0;
                                 $120 = $119&1;
                                 L50: do {
                                  if ($120) {
                                   $209 = ___cxa_find_matching_catch_2()|0;
                                   $210 = tempRet0;
                                   $eh$lpad$body38$index57Z2D = $210;$eh$lpad$body38$indexZ2D = $209;
                                   label = 91;
                                  } else {
                                   __THREW__ = 0;
                                   invoke_vii(787,($str$i63|0),0);
                                   $121 = __THREW__; __THREW__ = 0;
                                   $122 = $121&1;
                                   L53: do {
                                    if ($122) {
                                     label = 48;
                                    } else {
                                     __THREW__ = 0;
                                     (invoke_iii(759,($str$i63|0),(13848|0))|0);
                                     $123 = __THREW__; __THREW__ = 0;
                                     $124 = $123&1;
                                     do {
                                      if (!($124)) {
                                       __THREW__ = 0;
                                       $125 = (invoke_iii(760,($2|0),($str$i63|0))|0);
                                       $126 = __THREW__; __THREW__ = 0;
                                       $127 = $126&1;
                                       if ($127) {
                                        break;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($str$i63);
                                       __THREW__ = 0;
                                       invoke_vii(786,($7|0),($125|0));
                                       $130 = __THREW__; __THREW__ = 0;
                                       $131 = $130&1;
                                       if ($131) {
                                        label = 48;
                                        break L53;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($2);
                                       $134 = HEAP32[$43>>2]|0;
                                       __THREW__ = 0;
                                       $135 = (invoke_ii(933,($134|0))|0);
                                       $136 = __THREW__; __THREW__ = 0;
                                       $137 = $136&1;
                                       do {
                                        if ($137) {
                                         label = 92;
                                        } else {
                                         $138 = HEAP32[$135>>2]|0;
                                         $139 = ((($138)) + 16|0);
                                         $140 = HEAP32[$139>>2]|0;
                                         __THREW__ = 0;
                                         invoke_vii($140|0,($21|0),($135|0));
                                         $141 = __THREW__; __THREW__ = 0;
                                         $142 = $141&1;
                                         if ($142) {
                                          label = 92;
                                          break;
                                         }
                                         __THREW__ = 0;
                                         invoke_viii(785,($agg$result|0),($7|0),($21|0));
                                         $143 = __THREW__; __THREW__ = 0;
                                         $144 = $143&1;
                                         if ($144) {
                                          $213 = ___cxa_find_matching_catch_2()|0;
                                          $214 = tempRet0;
                                          __ZN9ExitGames6Common7JStringD2Ev($21);
                                          $$0 = $214;$$01 = $213;
                                          break;
                                         }
                                         __ZN9ExitGames6Common7JStringD2Ev($21);
                                         __ZN9ExitGames6Common7JStringD2Ev($7);
                                         __ZN9ExitGames6Common7JStringD2Ev($8);
                                         __ZN9ExitGames6Common7JStringD2Ev($9);
                                         __ZN9ExitGames6Common7JStringD2Ev($10);
                                         __ZN9ExitGames6Common7JStringD2Ev($11);
                                         __ZN9ExitGames6Common7JStringD2Ev($12);
                                         __ZN9ExitGames6Common7JStringD2Ev($13);
                                         __ZN9ExitGames6Common7JStringD2Ev($14);
                                         __ZN9ExitGames6Common7JStringD2Ev($15);
                                         __ZN9ExitGames6Common7JStringD2Ev($16);
                                         if (!($all)) {
                                          STACKTOP = sp;return;
                                         }
                                         __THREW__ = 0;
                                         invoke_vii(786,($1|0),($agg$result|0));
                                         $145 = __THREW__; __THREW__ = 0;
                                         $146 = $145&1;
                                         L68: do {
                                          if ($146) {
                                           $215 = ___cxa_find_matching_catch_2()|0;
                                           $216 = tempRet0;
                                           $eh$lpad$body42$index62Z2D = $216;$eh$lpad$body42$indexZ2D = $215;
                                           label = 105;
                                          } else {
                                           __THREW__ = 0;
                                           invoke_vii(787,($str$i67|0),0);
                                           $147 = __THREW__; __THREW__ = 0;
                                           $148 = $147&1;
                                           L71: do {
                                            if ($148) {
                                             label = 60;
                                            } else {
                                             __THREW__ = 0;
                                             (invoke_iii(759,($str$i67|0),(13872|0))|0);
                                             $149 = __THREW__; __THREW__ = 0;
                                             $150 = $149&1;
                                             do {
                                              if (!($150)) {
                                               __THREW__ = 0;
                                               $151 = (invoke_iii(760,($1|0),($str$i67|0))|0);
                                               $152 = __THREW__; __THREW__ = 0;
                                               $153 = $152&1;
                                               if ($153) {
                                                break;
                                               }
                                               __ZN9ExitGames6Common7JStringD2Ev($str$i67);
                                               __THREW__ = 0;
                                               invoke_vii(786,($25|0),($151|0));
                                               $156 = __THREW__; __THREW__ = 0;
                                               $157 = $156&1;
                                               if ($157) {
                                                label = 60;
                                                break L71;
                                               }
                                               __ZN9ExitGames6Common7JStringD2Ev($1);
                                               $160 = HEAP32[$43>>2]|0;
                                               __THREW__ = 0;
                                               $161 = (invoke_ii(934,($160|0))|0);
                                               $162 = __THREW__; __THREW__ = 0;
                                               $163 = $162&1;
                                               do {
                                                if ($163) {
                                                 label = 106;
                                                } else {
                                                 __THREW__ = 0;
                                                 invoke_viii(783,($26|0),($161|0),0);
                                                 $164 = __THREW__; __THREW__ = 0;
                                                 $165 = $164&1;
                                                 if ($165) {
                                                  label = 106;
                                                  break;
                                                 }
                                                 __THREW__ = 0;
                                                 invoke_viii(785,($24|0),($25|0),($26|0));
                                                 $166 = __THREW__; __THREW__ = 0;
                                                 $167 = $166&1;
                                                 if ($167) {
                                                  $219 = ___cxa_find_matching_catch_2()|0;
                                                  $220 = tempRet0;
                                                  $$13 = $220;$$1319 = $219;
                                                 } else {
                                                  __THREW__ = 0;
                                                  invoke_vii(786,($0|0),($24|0));
                                                  $168 = __THREW__; __THREW__ = 0;
                                                  $169 = $168&1;
                                                  L83: do {
                                                   if ($169) {
                                                    $221 = ___cxa_find_matching_catch_2()|0;
                                                    $222 = tempRet0;
                                                    $eh$lpad$body46$index67Z2D = $222;$eh$lpad$body46$indexZ2D = $221;
                                                    label = 109;
                                                   } else {
                                                    __THREW__ = 0;
                                                    invoke_vii(787,($str$i71|0),0);
                                                    $170 = __THREW__; __THREW__ = 0;
                                                    $171 = $170&1;
                                                    L86: do {
                                                     if ($171) {
                                                      label = 71;
                                                     } else {
                                                      __THREW__ = 0;
                                                      (invoke_iii(759,($str$i71|0),(13916|0))|0);
                                                      $172 = __THREW__; __THREW__ = 0;
                                                      $173 = $172&1;
                                                      do {
                                                       if (!($173)) {
                                                        __THREW__ = 0;
                                                        $174 = (invoke_iii(760,($0|0),($str$i71|0))|0);
                                                        $175 = __THREW__; __THREW__ = 0;
                                                        $176 = $175&1;
                                                        if ($176) {
                                                         break;
                                                        }
                                                        __ZN9ExitGames6Common7JStringD2Ev($str$i71);
                                                        __THREW__ = 0;
                                                        invoke_vii(786,($23|0),($174|0));
                                                        $179 = __THREW__; __THREW__ = 0;
                                                        $180 = $179&1;
                                                        if ($180) {
                                                         label = 71;
                                                         break L86;
                                                        }
                                                        __ZN9ExitGames6Common7JStringD2Ev($0);
                                                        $183 = HEAP32[$43>>2]|0;
                                                        __THREW__ = 0;
                                                        $184 = (invoke_ii(935,($183|0))|0);
                                                        $185 = __THREW__; __THREW__ = 0;
                                                        $186 = $185&1;
                                                        do {
                                                         if ($186) {
                                                          label = 110;
                                                         } else {
                                                          __THREW__ = 0;
                                                          invoke_viii(783,($27|0),($184|0),0);
                                                          $187 = __THREW__; __THREW__ = 0;
                                                          $188 = $187&1;
                                                          if ($188) {
                                                           label = 110;
                                                           break;
                                                          }
                                                          __THREW__ = 0;
                                                          invoke_viii(785,($22|0),($23|0),($27|0));
                                                          $189 = __THREW__; __THREW__ = 0;
                                                          $190 = $189&1;
                                                          do {
                                                           if ($190) {
                                                            $225 = ___cxa_find_matching_catch_2()|0;
                                                            $226 = tempRet0;
                                                            $$10 = $226;$$1016 = $225;
                                                           } else {
                                                            __THREW__ = 0;
                                                            (invoke_iii(803,($agg$result|0),($22|0))|0);
                                                            $191 = __THREW__; __THREW__ = 0;
                                                            $192 = $191&1;
                                                            if ($192) {
                                                             $227 = ___cxa_find_matching_catch_2()|0;
                                                             $228 = tempRet0;
                                                             __ZN9ExitGames6Common7JStringD2Ev($22);
                                                             $$10 = $228;$$1016 = $227;
                                                             break;
                                                            }
                                                            __ZN9ExitGames6Common7JStringD2Ev($22);
                                                            __ZN9ExitGames6Common7JStringD2Ev($27);
                                                            __ZN9ExitGames6Common7JStringD2Ev($23);
                                                            __ZN9ExitGames6Common7JStringD2Ev($24);
                                                            __ZN9ExitGames6Common7JStringD2Ev($26);
                                                            __ZN9ExitGames6Common7JStringD2Ev($25);
                                                            STACKTOP = sp;return;
                                                           }
                                                          } while(0);
                                                          __ZN9ExitGames6Common7JStringD2Ev($27);
                                                          $$11 = $$10;$$1117 = $$1016;
                                                         }
                                                        } while(0);
                                                        if ((label|0) == 110) {
                                                         $223 = ___cxa_find_matching_catch_2()|0;
                                                         $224 = tempRet0;
                                                         $$11 = $224;$$1117 = $223;
                                                        }
                                                        __ZN9ExitGames6Common7JStringD2Ev($23);
                                                        $$12 = $$11;$$1218 = $$1117;
                                                        break L83;
                                                       }
                                                      } while(0);
                                                      $177 = ___cxa_find_matching_catch_2()|0;
                                                      $178 = tempRet0;
                                                      __ZN9ExitGames6Common7JStringD2Ev($str$i71);
                                                      $eh$lpad$body74$index32Z2D = $178;$eh$lpad$body74$indexZ2D = $177;
                                                     }
                                                    } while(0);
                                                    if ((label|0) == 71) {
                                                     $181 = ___cxa_find_matching_catch_2()|0;
                                                     $182 = tempRet0;
                                                     $eh$lpad$body74$index32Z2D = $182;$eh$lpad$body74$indexZ2D = $181;
                                                    }
                                                    __ZN9ExitGames6Common7JStringD2Ev($0);
                                                    $eh$lpad$body46$index67Z2D = $eh$lpad$body74$index32Z2D;$eh$lpad$body46$indexZ2D = $eh$lpad$body74$indexZ2D;
                                                    label = 109;
                                                   }
                                                  } while(0);
                                                  if ((label|0) == 109) {
                                                   $$12 = $eh$lpad$body46$index67Z2D;$$1218 = $eh$lpad$body46$indexZ2D;
                                                  }
                                                  __ZN9ExitGames6Common7JStringD2Ev($24);
                                                  $$13 = $$12;$$1319 = $$1218;
                                                 }
                                                 __ZN9ExitGames6Common7JStringD2Ev($26);
                                                 $$14 = $$13;$$1420 = $$1319;
                                                }
                                               } while(0);
                                               if ((label|0) == 106) {
                                                $217 = ___cxa_find_matching_catch_2()|0;
                                                $218 = tempRet0;
                                                $$14 = $218;$$1420 = $217;
                                               }
                                               __ZN9ExitGames6Common7JStringD2Ev($25);
                                               $$15 = $$14;$$1521 = $$1420;
                                               break L68;
                                              }
                                             } while(0);
                                             $154 = ___cxa_find_matching_catch_2()|0;
                                             $155 = tempRet0;
                                             __ZN9ExitGames6Common7JStringD2Ev($str$i67);
                                             $eh$lpad$body70$index27Z2D = $155;$eh$lpad$body70$indexZ2D = $154;
                                            }
                                           } while(0);
                                           if ((label|0) == 60) {
                                            $158 = ___cxa_find_matching_catch_2()|0;
                                            $159 = tempRet0;
                                            $eh$lpad$body70$index27Z2D = $159;$eh$lpad$body70$indexZ2D = $158;
                                           }
                                           __ZN9ExitGames6Common7JStringD2Ev($1);
                                           $eh$lpad$body42$index62Z2D = $eh$lpad$body70$index27Z2D;$eh$lpad$body42$indexZ2D = $eh$lpad$body70$indexZ2D;
                                           label = 105;
                                          }
                                         } while(0);
                                         if ((label|0) == 105) {
                                          $$15 = $eh$lpad$body42$index62Z2D;$$1521 = $eh$lpad$body42$indexZ2D;
                                         }
                                         __ZN9ExitGames6Common7JStringD2Ev($agg$result);
                                         $$16 = $$15;$$1622 = $$1521;
                                         ___resumeException($$1622|0);
                                         // unreachable;
                                        }
                                       } while(0);
                                       if ((label|0) == 92) {
                                        $211 = ___cxa_find_matching_catch_2()|0;
                                        $212 = tempRet0;
                                        $$0 = $212;$$01 = $211;
                                       }
                                       __ZN9ExitGames6Common7JStringD2Ev($7);
                                       $$1 = $$0;$$17 = $$01;
                                       break L50;
                                      }
                                     } while(0);
                                     $128 = ___cxa_find_matching_catch_2()|0;
                                     $129 = tempRet0;
                                     __ZN9ExitGames6Common7JStringD2Ev($str$i63);
                                     $eh$lpad$body66$index22Z2D = $129;$eh$lpad$body66$indexZ2D = $128;
                                    }
                                   } while(0);
                                   if ((label|0) == 48) {
                                    $132 = ___cxa_find_matching_catch_2()|0;
                                    $133 = tempRet0;
                                    $eh$lpad$body66$index22Z2D = $133;$eh$lpad$body66$indexZ2D = $132;
                                   }
                                   __ZN9ExitGames6Common7JStringD2Ev($2);
                                   $eh$lpad$body38$index57Z2D = $eh$lpad$body66$index22Z2D;$eh$lpad$body38$indexZ2D = $eh$lpad$body66$indexZ2D;
                                   label = 91;
                                  }
                                 } while(0);
                                 if ((label|0) == 91) {
                                  $$1 = $eh$lpad$body38$index57Z2D;$$17 = $eh$lpad$body38$indexZ2D;
                                 }
                                 __ZN9ExitGames6Common7JStringD2Ev($8);
                                 $$2 = $$1;$$28 = $$17;
                                }
                               } while(0);
                               if ((label|0) == 89) {
                                $207 = ___cxa_find_matching_catch_2()|0;
                                $208 = tempRet0;
                                $$2 = $208;$$28 = $207;
                               }
                               __ZN9ExitGames6Common7JStringD2Ev($9);
                               $$3 = $$2;$$39 = $$28;
                               break L38;
                              }
                             } while(0);
                             $105 = ___cxa_find_matching_catch_2()|0;
                             $106 = tempRet0;
                             __ZN9ExitGames6Common7JStringD2Ev($str$i59);
                             $eh$lpad$body62$index17Z2D = $106;$eh$lpad$body62$indexZ2D = $105;
                            }
                           } while(0);
                           if ((label|0) == 38) {
                            $109 = ___cxa_find_matching_catch_2()|0;
                            $110 = tempRet0;
                            $eh$lpad$body62$index17Z2D = $110;$eh$lpad$body62$indexZ2D = $109;
                           }
                           __ZN9ExitGames6Common7JStringD2Ev($3);
                           $eh$lpad$body34$index52Z2D = $eh$lpad$body62$index17Z2D;$eh$lpad$body34$indexZ2D = $eh$lpad$body62$indexZ2D;
                           label = 88;
                          }
                         } while(0);
                         if ((label|0) == 88) {
                          $$3 = $eh$lpad$body34$index52Z2D;$$39 = $eh$lpad$body34$indexZ2D;
                         }
                         __ZN9ExitGames6Common7JStringD2Ev($10);
                         $$4 = $$3;$$410 = $$39;
                        }
                       } while(0);
                       if ((label|0) == 86) {
                        $203 = ___cxa_find_matching_catch_2()|0;
                        $204 = tempRet0;
                        $$4 = $204;$$410 = $203;
                       }
                       __ZN9ExitGames6Common7JStringD2Ev($11);
                       $$5 = $$4;$$511 = $$410;
                       break L25;
                      }
                     } while(0);
                     $80 = ___cxa_find_matching_catch_2()|0;
                     $81 = tempRet0;
                     __ZN9ExitGames6Common7JStringD2Ev($str$i55);
                     $eh$lpad$body58$index12Z2D = $81;$eh$lpad$body58$indexZ2D = $80;
                    }
                   } while(0);
                   if ((label|0) == 27) {
                    $84 = ___cxa_find_matching_catch_2()|0;
                    $85 = tempRet0;
                    $eh$lpad$body58$index12Z2D = $85;$eh$lpad$body58$indexZ2D = $84;
                   }
                   __ZN9ExitGames6Common7JStringD2Ev($4);
                   $eh$lpad$body30$index47Z2D = $eh$lpad$body58$index12Z2D;$eh$lpad$body30$indexZ2D = $eh$lpad$body58$indexZ2D;
                   label = 85;
                  }
                 } while(0);
                 if ((label|0) == 85) {
                  $$5 = $eh$lpad$body30$index47Z2D;$$511 = $eh$lpad$body30$indexZ2D;
                 }
                 __ZN9ExitGames6Common7JStringD2Ev($12);
                 $$6 = $$5;$$612 = $$511;
                }
               } while(0);
               if ((label|0) == 83) {
                $199 = ___cxa_find_matching_catch_2()|0;
                $200 = tempRet0;
                $$6 = $200;$$612 = $199;
               }
               __ZN9ExitGames6Common7JStringD2Ev($13);
               $$7 = $$6;$$713 = $$612;
               break L13;
              }
             }
             $59 = ___cxa_find_matching_catch_2()|0;
             $60 = tempRet0;
             __ZN9ExitGames6Common7JStringD2Ev($str$i51);
             $eh$lpad$body54$index7Z2D = $60;$eh$lpad$body54$indexZ2D = $59;
            }
           } while(0);
           if ((label|0) == 17) {
            $63 = ___cxa_find_matching_catch_2()|0;
            $64 = tempRet0;
            $eh$lpad$body54$index7Z2D = $64;$eh$lpad$body54$indexZ2D = $63;
           }
           __ZN9ExitGames6Common7JStringD2Ev($5);
           $eh$lpad$body26$index42Z2D = $eh$lpad$body54$index7Z2D;$eh$lpad$body26$indexZ2D = $eh$lpad$body54$indexZ2D;
           label = 82;
          }
         } while(0);
         if ((label|0) == 82) {
          $$7 = $eh$lpad$body26$index42Z2D;$$713 = $eh$lpad$body26$indexZ2D;
         }
         __ZN9ExitGames6Common7JStringD2Ev($14);
         $$8 = $$7;$$814 = $$713;
        }
       }
       if ((label|0) == 80) {
        $195 = ___cxa_find_matching_catch_2()|0;
        $196 = tempRet0;
        $$8 = $196;$$814 = $195;
       }
       __ZN9ExitGames6Common7JStringD2Ev($15);
       $$9 = $$8;$$915 = $$814;
       break L1;
      }
     }
     $37 = ___cxa_find_matching_catch_2()|0;
     $38 = tempRet0;
     __ZN9ExitGames6Common7JStringD2Ev($str$i);
     $eh$lpad$body50$index2Z2D = $38;$eh$lpad$body50$indexZ2D = $37;
    }
   } while(0);
   if ((label|0) == 7) {
    $41 = ___cxa_find_matching_catch_2()|0;
    $42 = tempRet0;
    $eh$lpad$body50$index2Z2D = $42;$eh$lpad$body50$indexZ2D = $41;
   }
   __ZN9ExitGames6Common7JStringD2Ev($6);
   $eh$lpad$body$index37Z2D = $eh$lpad$body50$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body50$indexZ2D;
   label = 79;
  }
 } while(0);
 if ((label|0) == 79) {
  $$9 = $eh$lpad$body$index37Z2D;$$915 = $eh$lpad$body$indexZ2D;
 }
 __ZN9ExitGames6Common7JStringD2Ev($16);
 $$16 = $$9;$$1622 = $$915;
 ___resumeException($$1622|0);
 // unreachable;
}
function __ZN9ExitGames6Photon10PhotonPeer22initUserDataEncryptionERKNS_6Common7JVectorIhEE($this,$secret) {
 $this = $this|0;
 $secret = $secret|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 __ZN9ExitGames6Photon8Internal8PeerBase22initUserDataEncryptionERKNS_6Common7JVectorIhEE($1,$secret);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeer7connectERKNS_6Common7JStringES5_RKNS2_6ObjectE($this,$ipAddr,$appID,$customData) {
 $this = $this|0;
 $ipAddr = $ipAddr|0;
 $appID = $appID|0;
 $customData = $customData|0;
 var $$0 = 0, $$02 = 0, $$1 = 0, $$13 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $ipAddrWithoutProtocol = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $0 = sp + 56|0;
 $ipAddrWithoutProtocol = sp + 40|0;
 $1 = sp + 24|0;
 $2 = sp + 8|0;
 $3 = sp;
 __ZN9ExitGames6Photon10PhotonPeer14createPeerBaseEv($this);
 __ZN9ExitGames6Common7JStringC2EPKc($0,147871);
 __THREW__ = 0;
 $4 = (invoke_iii(870,($ipAddr|0),($0|0))|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  __ZN9ExitGames6Common7JStringD2Ev($0);
  $$1 = $28;$$13 = $29;
  ___resumeException($$1|0);
  // unreachable;
 }
 __ZN9ExitGames6Common7JStringD2Ev($0);
 $7 = ($4|0)==(-1);
 do {
  if ($7) {
   __ZN9ExitGames6Common7JStringC2ERKS1_($ipAddrWithoutProtocol,$ipAddr);
  } else {
   __ZN9ExitGames6Common7JStringC2EPKw($1,13476);
   __THREW__ = 0;
   $8 = (invoke_ii(820,($1|0))|0);
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if (!($10)) {
    $11 = (($8) + ($4))|0;
    __THREW__ = 0;
    invoke_viii(806,($ipAddrWithoutProtocol|0),($ipAddr|0),($11|0));
    $12 = __THREW__; __THREW__ = 0;
    $13 = $12&1;
    if (!($13)) {
     __ZN9ExitGames6Common7JStringD2Ev($1);
     break;
    }
   }
   $30 = ___cxa_find_matching_catch_2()|0;
   $31 = tempRet0;
   __ZN9ExitGames6Common7JStringD2Ev($1);
   $$1 = $30;$$13 = $31;
   ___resumeException($$1|0);
   // unreachable;
  }
 } while(0);
 $14 = ((($this)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($16)) + 20|0);
 $18 = HEAP32[$17>>2]|0;
 $19 = ((($this)) + 28|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)==(5);
 $22 = $21 ? 13492 : 13520;
 HEAP32[$3>>2] = $22;
 __THREW__ = 0;
 invoke_viii(814,($2|0),($3|0),($ipAddrWithoutProtocol|0));
 $23 = __THREW__; __THREW__ = 0;
 $24 = $23&1;
 do {
  if ($24) {
   $32 = ___cxa_find_matching_catch_2()|0;
   $33 = tempRet0;
   $$0 = $32;$$02 = $33;
  } else {
   __THREW__ = 0;
   $25 = (invoke_iiiii($18|0,($15|0),($2|0),($appID|0),($customData|0))|0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $34 = ___cxa_find_matching_catch_2()|0;
    $35 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($2);
    $$0 = $34;$$02 = $35;
    break;
   } else {
    __ZN9ExitGames6Common7JStringD2Ev($2);
    __ZN9ExitGames6Common7JStringD2Ev($ipAddrWithoutProtocol);
    STACKTOP = sp;return ($25|0);
   }
  }
 } while(0);
 __ZN9ExitGames6Common7JStringD2Ev($ipAddrWithoutProtocol);
 $$1 = $$0;$$13 = $$02;
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZN9ExitGames6Photon10PhotonPeer14createPeerBaseEv($this) {
 $this = $this|0;
 var $$pr = 0, $$pre = 0, $$pre$i5 = 0, $$pre$phiZ2D = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$pre = ((($this)) + 28|0);
  $$pre$phiZ2D = $$pre;
 } else {
  $3 = (__ZNK9ExitGames6Photon8Internal8PeerBase21getConnectionProtocolEv($1)|0);
  $4 = ((($this)) + 28|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = ($3<<24>>24)==($5<<24>>24);
  if ($6) {
   return;
  }
  $$pr = HEAP32[$0>>2]|0;
  $7 = ($$pr|0)==(0|0);
  if ($7) {
   $$pre$phiZ2D = $4;
  } else {
   $8 = (__ZNK9ExitGames6Photon8Internal8PeerBase21getConnectionProtocolEv($$pr)|0);
   $9 = (__ZN9ExitGames6Photon18ConnectionProtocol14getIsWebSocketEh($8)|0);
   if ($9) {
    $10 = HEAP8[$4>>0]|0;
    $11 = (__ZN9ExitGames6Photon18ConnectionProtocol14getIsWebSocketEh($10)|0);
    if ($11) {
     return;
    } else {
     $$pre$phiZ2D = $4;
    }
   } else {
    $$pre$phiZ2D = $4;
   }
  }
 }
 $12 = ((($this)) + 12|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($this)) + 4|0);
 $15 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(200)|0);
 $16 = HEAP8[$$pre$phiZ2D>>0]|0;
 __ZN9ExitGames6Photon8Internal5TPeerC2ERNS1_8PeerDataERKNS_6Common6LoggerEh($15,$13,$14,$16);
 $17 = $15;
 $18 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(8)|0);
 $19 = $18;
 $20 = $19;
 HEAP32[$20>>2] = 1;
 $21 = (($19) + 4)|0;
 $22 = $21;
 HEAP32[$22>>2] = 0;
 $23 = ((($this)) + 24|0);
 $24 = HEAP32[$23>>2]|0;
 $25 = ($24|0)==(0|0);
 do {
  if (!($25)) {
   $26 = $24;
   $27 = $26;
   $28 = HEAP32[$27>>2]|0;
   $29 = (($26) + 4)|0;
   $30 = $29;
   $31 = HEAP32[$30>>2]|0;
   $32 = (_i64Add(($28|0),($31|0),-1,-1)|0);
   $33 = tempRet0;
   $34 = $24;
   $35 = $34;
   HEAP32[$35>>2] = $32;
   $36 = (($34) + 4)|0;
   $37 = $36;
   HEAP32[$37>>2] = $33;
   $38 = ($32|0)==(0);
   $39 = ($33|0)==(0);
   $40 = $38 & $39;
   if ($40) {
    $41 = HEAP32[$0>>2]|0;
    $42 = ($41|0)==(0|0);
    if ($42) {
     $48 = $24;
     label = 12;
    } else {
     $43 = HEAP32[$41>>2]|0;
     $44 = HEAP32[$43>>2]|0;
     FUNCTION_TABLE_vi[$44 & 2047]($41);
     __THREW__ = 0;
     invoke_vi(750,($41|0));
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if (!($46)) {
      $$pre$i5 = HEAP32[$23>>2]|0;
      $47 = ($$pre$i5|0)==(0|0);
      if ($47) {
       break;
      } else {
       $48 = $$pre$i5;
       label = 12;
      }
     }
    }
    if ((label|0) == 12) {
     __THREW__ = 0;
     invoke_vi(750,($48|0));
     $49 = __THREW__; __THREW__ = 0;
     $50 = $49&1;
     if (!($50)) {
      break;
     }
    }
    $69 = ___cxa_find_matching_catch_2()|0;
    $70 = tempRet0;
    $71 = $18;
    $72 = $71;
    $73 = HEAP32[$72>>2]|0;
    $74 = (($71) + 4)|0;
    $75 = $74;
    $76 = HEAP32[$75>>2]|0;
    $77 = (_i64Add(($73|0),($76|0),-1,-1)|0);
    $78 = tempRet0;
    $79 = $18;
    $80 = $79;
    HEAP32[$80>>2] = $77;
    $81 = (($79) + 4)|0;
    $82 = $81;
    HEAP32[$82>>2] = $78;
    $83 = ($77|0)==(0);
    $84 = ($78|0)==(0);
    $85 = $83 & $84;
    if (!($85)) {
     ___resumeException($69|0);
     // unreachable;
    }
    $86 = ($15|0)==(0|0);
    if (!($86)) {
     $87 = HEAP32[$15>>2]|0;
     $88 = HEAP32[$87>>2]|0;
     FUNCTION_TABLE_vi[$88 & 2047]($15);
     __THREW__ = 0;
     invoke_vi(750,($15|0));
     $89 = __THREW__; __THREW__ = 0;
     $90 = $89&1;
     if ($90) {
      $93 = ___cxa_find_matching_catch_3(0|0)|0;
      $94 = tempRet0;
      ___clang_call_terminate($93);
      // unreachable;
     }
    }
    __THREW__ = 0;
    invoke_vi(750,($18|0));
    $91 = __THREW__; __THREW__ = 0;
    $92 = $91&1;
    if ($92) {
     $93 = ___cxa_find_matching_catch_3(0|0)|0;
     $94 = tempRet0;
     ___clang_call_terminate($93);
     // unreachable;
    } else {
     ___resumeException($69|0);
     // unreachable;
    }
   }
  }
 } while(0);
 HEAP32[$0>>2] = $17;
 HEAP32[$23>>2] = $18;
 $51 = $18;
 $52 = $51;
 $53 = HEAP32[$52>>2]|0;
 $54 = (($51) + 4)|0;
 $55 = $54;
 $56 = HEAP32[$55>>2]|0;
 $57 = ($53|0)==(0);
 $58 = ($56|0)==(0);
 $59 = $57 & $58;
 if (!($59)) {
  return;
 }
 $60 = ($15|0)==(0|0);
 if (!($60)) {
  $61 = HEAP32[$15>>2]|0;
  $62 = HEAP32[$61>>2]|0;
  FUNCTION_TABLE_vi[$62 & 2047]($15);
  __THREW__ = 0;
  invoke_vi(750,($15|0));
  $63 = __THREW__; __THREW__ = 0;
  $64 = $63&1;
  if ($64) {
   $67 = ___cxa_find_matching_catch_3(0|0)|0;
   $68 = tempRet0;
   ___clang_call_terminate($67);
   // unreachable;
  }
 }
 __THREW__ = 0;
 invoke_vi(750,($18|0));
 $65 = __THREW__; __THREW__ = 0;
 $66 = $65&1;
 if ($66) {
  $67 = ___cxa_find_matching_catch_3(0|0)|0;
  $68 = tempRet0;
  ___clang_call_terminate($67);
  // unreachable;
 } else {
  return;
 }
}
function __ZNK9ExitGames6Photon8Internal8PeerBase21getConnectionProtocolEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 136|0);
 $1 = HEAP8[$0>>0]|0;
 return ($1|0);
}
function __ZN9ExitGames6Photon18ConnectionProtocol14getIsWebSocketEh($connectionProtocol) {
 $connectionProtocol = $connectionProtocol|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = $connectionProtocol & -2;
 $1 = ($0<<24>>24)==(4);
 return ($1|0);
}
function __ZN9ExitGames6Photon8Internal5TPeerC2ERNS1_8PeerDataERKNS_6Common6LoggerEh($this,$peerData,$logger,$connectionProtocol) {
 $this = $this|0;
 $peerData = $peerData|0;
 $logger = $logger|0;
 $connectionProtocol = $connectionProtocol|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $i$01$i$i = 0, $i$01$i$i8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon8Internal8PeerBaseC2ERNS1_8PeerDataERKNS_6Common6LoggerEhi($this,$peerData,$logger,$connectionProtocol,0);
 HEAP32[$this>>2] = (10516);
 $0 = ((($this)) + 160|0);
 HEAP32[$0>>2] = (12328);
 $1 = ((($this)) + 164|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 168|0);
 HEAP32[$2>>2] = 40;
 __THREW__ = 0;
 $3 = (invoke_ii(775,480)|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $22 = ___cxa_find_matching_catch_2()|0;
  $23 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($0);
  $$1 = $23;$$12 = $22;
  __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this);
  ___resumeException($$12|0);
  // unreachable;
 }
 $6 = ((($this)) + 176|0);
 HEAP32[$6>>2] = $3;
 $7 = ((($this)) + 172|0);
 HEAP32[$7>>2] = 10;
 $8 = ((($this)) + 180|0);
 HEAP32[$8>>2] = (12328);
 $9 = ((($this)) + 184|0);
 HEAP32[$9>>2] = 0;
 $10 = ((($this)) + 188|0);
 HEAP32[$10>>2] = 40;
 __THREW__ = 0;
 $11 = (invoke_ii(775,480)|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 do {
  if ($13) {
   $24 = ___cxa_find_matching_catch_2()|0;
   $25 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($8);
   $$0 = $25;$$01 = $24;
  } else {
   $14 = ((($this)) + 196|0);
   HEAP32[$14>>2] = $11;
   $15 = ((($this)) + 192|0);
   HEAP32[$15>>2] = 10;
   __THREW__ = 0;
   $16 = (invoke_ii(775,108)|0);
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if (!($18)) {
    __THREW__ = 0;
    invoke_vii(936,($16|0),($this|0));
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     $21 = ((($this)) + 40|0);
     HEAP32[$21>>2] = $16;
     return;
    }
   }
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   HEAP32[$8>>2] = (12328);
   $28 = HEAP32[$9>>2]|0;
   $29 = ($28|0)==(0);
   if (!($29)) {
    $i$01$i$i = 0;
    while(1) {
     $30 = HEAP32[$14>>2]|0;
     $31 = (($30) + (($i$01$i$i*12)|0)|0);
     $32 = HEAP32[$31>>2]|0;
     $33 = HEAP32[$32>>2]|0;
     FUNCTION_TABLE_vi[$33 & 2047]($31);
     $34 = (($i$01$i$i) + 1)|0;
     $35 = HEAP32[$9>>2]|0;
     $36 = ($34>>>0)<($35>>>0);
     if ($36) {
      $i$01$i$i = $34;
     } else {
      break;
     }
    }
   }
   HEAP32[$9>>2] = 0;
   $37 = HEAP32[$14>>2]|0;
   __THREW__ = 0;
   invoke_vi(750,($37|0));
   $38 = __THREW__; __THREW__ = 0;
   $39 = $38&1;
   if ($39) {
    $40 = ___cxa_find_matching_catch_3(0|0)|0;
    $41 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($8);
    ___clang_call_terminate($40);
    // unreachable;
   } else {
    __ZN9ExitGames6Common4BaseD2Ev($8);
    $$0 = $27;$$01 = $26;
    break;
   }
  }
 } while(0);
 HEAP32[$0>>2] = (12328);
 $42 = HEAP32[$1>>2]|0;
 $43 = ($42|0)==(0);
 if (!($43)) {
  $i$01$i$i8 = 0;
  while(1) {
   $44 = HEAP32[$6>>2]|0;
   $45 = (($44) + (($i$01$i$i8*12)|0)|0);
   $46 = HEAP32[$45>>2]|0;
   $47 = HEAP32[$46>>2]|0;
   FUNCTION_TABLE_vi[$47 & 2047]($45);
   $48 = (($i$01$i$i8) + 1)|0;
   $49 = HEAP32[$1>>2]|0;
   $50 = ($48>>>0)<($49>>>0);
   if ($50) {
    $i$01$i$i8 = $48;
   } else {
    break;
   }
  }
 }
 HEAP32[$1>>2] = 0;
 $51 = HEAP32[$6>>2]|0;
 __THREW__ = 0;
 invoke_vi(750,($51|0));
 $52 = __THREW__; __THREW__ = 0;
 $53 = $52&1;
 if ($53) {
  $54 = ___cxa_find_matching_catch_3(0|0)|0;
  $55 = tempRet0;
  __ZN9ExitGames6Common4BaseD2Ev($0);
  ___clang_call_terminate($54);
  // unreachable;
 }
 __ZN9ExitGames6Common4BaseD2Ev($0);
 $$1 = $$0;$$12 = $$01;
 __ZN9ExitGames6Photon8Internal8PeerBaseD2Ev($this);
 ___resumeException($$12|0);
 // unreachable;
}
function __ZN9ExitGames6Photon8Internal8PeerBaseC2ERNS1_8PeerDataERKNS_6Common6LoggerEhi($this,$peerData,$logger,$connectionProtocol,$packageHeaderSize) {
 $this = $this|0;
 $peerData = $peerData|0;
 $logger = $logger|0;
 $connectionProtocol = $connectionProtocol|0;
 $packageHeaderSize = $packageHeaderSize|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$2 = 0, $$23 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (10152);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $peerData;
 $1 = ((($this)) + 8|0);
 HEAP32[$1>>2] = $logger;
 $2 = ((($this)) + 20|0);
 __THREW__ = 0;
 invoke_vii(787,($2|0),0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $48 = ___cxa_find_matching_catch_2()|0;
  $49 = tempRet0;
  $$2 = $49;$$23 = $48;
  ___resumeException($$23|0);
  // unreachable;
 }
 $5 = ((($this)) + 116|0);
 HEAP32[$5>>2] = (3820);
 $6 = ((($this)) + 120|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($this)) + 124|0);
 HEAP32[$7>>2] = 40;
 __THREW__ = 0;
 $8 = (invoke_ii(775,40)|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 do {
  if ($10) {
   $50 = ___cxa_find_matching_catch_2()|0;
   $51 = tempRet0;
   __ZN9ExitGames6Common4BaseD2Ev($5);
   $$1 = $51;$$12 = $50;
  } else {
   $11 = ((($this)) + 132|0);
   HEAP32[$11>>2] = $8;
   $12 = ((($this)) + 128|0);
   HEAP32[$12>>2] = 10;
   $13 = ((($this)) + 148|0);
   HEAP32[$13>>2] = 0;
   __THREW__ = 0;
   $14 = (invoke_ii(775,8)|0);
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if ($16) {
    $52 = ___cxa_find_matching_catch_2()|0;
    $53 = tempRet0;
    $$0 = $53;$$01 = $52;
   } else {
    $17 = ((($this)) + 152|0);
    $18 = $14;
    $19 = $18;
    HEAP32[$19>>2] = 1;
    $20 = (($18) + 4)|0;
    $21 = $20;
    HEAP32[$21>>2] = 0;
    HEAP32[$17>>2] = $14;
    $22 = ((($this)) + 12|0);
    HEAP16[$22>>1] = -1;
    $23 = ((($this)) + 16|0);
    HEAP32[$23>>2] = 0;
    $24 = ((($this)) + 36|0);
    $25 = ((($this)) + 60|0);
    $26 = ((($this)) + 100|0);
    HEAP32[$26>>2] = 0;
    ;HEAP32[$24>>2]=0|0;HEAP32[$24+4>>2]=0|0;HEAP32[$24+8>>2]=0|0;HEAP32[$24+12>>2]=0|0;HEAP32[$24+16>>2]=0|0;HEAP8[$24+20>>0]=0|0;
    dest=$25; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP16[$25+36>>1]=0|0;HEAP8[$25+38>>0]=0|0;
    __THREW__ = 0;
    $27 = (invoke_ii(775,56)|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     __THREW__ = 0;
     invoke_vii(937,($27|0),($packageHeaderSize|0));
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     if (!($31)) {
      $32 = ((($this)) + 104|0);
      HEAP32[$32>>2] = $27;
      __THREW__ = 0;
      $33 = (invoke_ii(775,56)|0);
      $34 = __THREW__; __THREW__ = 0;
      $35 = $34&1;
      if (!($35)) {
       __THREW__ = 0;
       invoke_vii(937,($33|0),($packageHeaderSize|0));
       $36 = __THREW__; __THREW__ = 0;
       $37 = $36&1;
       if (!($37)) {
        $38 = ((($this)) + 108|0);
        HEAP32[$38>>2] = $33;
        __THREW__ = 0;
        $39 = (invoke_ii(775,68)|0);
        $40 = __THREW__; __THREW__ = 0;
        $41 = $40&1;
        if (!($41)) {
         dest=$39; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
         __THREW__ = 0;
         invoke_vi(938,($39|0));
         $42 = __THREW__; __THREW__ = 0;
         $43 = $42&1;
         if (!($43)) {
          HEAP32[$39>>2] = (12748);
          $44 = ((($this)) + 112|0);
          HEAP32[$44>>2] = $39;
          $45 = ((($this)) + 136|0);
          HEAP8[$45>>0] = $connectionProtocol;
          $46 = ((($this)) + 140|0);
          HEAP32[$46>>2] = 0;
          $47 = ((($this)) + 144|0);
          HEAP32[$47>>2] = 0;
          return;
         }
        }
       }
      }
     }
    }
    $54 = ___cxa_find_matching_catch_2()|0;
    $55 = tempRet0;
    __ZN9ExitGames6Common7Helpers13SharedPointerINS_6Photon8Internal8PeerBase10InitValuesEED2Ev($13);
    $$0 = $55;$$01 = $54;
   }
   HEAP32[$5>>2] = (3820);
   HEAP32[$6>>2] = 0;
   $56 = HEAP32[$11>>2]|0;
   __THREW__ = 0;
   invoke_vi(750,($56|0));
   $57 = __THREW__; __THREW__ = 0;
   $58 = $57&1;
   if ($58) {
    $59 = ___cxa_find_matching_catch_3(0|0)|0;
    $60 = tempRet0;
    __ZN9ExitGames6Common4BaseD2Ev($5);
    ___clang_call_terminate($59);
    // unreachable;
   } else {
    __ZN9ExitGames6Common4BaseD2Ev($5);
    $$1 = $$0;$$12 = $$01;
    break;
   }
  }
 } while(0);
 __ZN9ExitGames6Common7JStringD2Ev($2);
 $$2 = $$1;$$23 = $$12;
 ___resumeException($$23|0);
 // unreachable;
}
function __ZN9ExitGames6Photon21TrafficStatsGameLevelC2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (13552);
 $0 = ((($this)) + 4|0);
 $1 = ((($this)) + 44|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 48|0);
 HEAP8[$2>>0] = 0;
 $3 = ((($this)) + 52|0);
 dest=$0; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP8[$0+36>>0]=0|0;
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;HEAP32[$3+12>>2]=0|0;
 return;
}
function __ZN9ExitGames6Photon21TrafficStatsGameLevelD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal19TrafficStatsCounterC2Ei($this,$packageHeaderSize) {
 $this = $this|0;
 $packageHeaderSize = $packageHeaderSize|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon12TrafficStatsC2Ei($this,$packageHeaderSize);
 HEAP32[$this>>2] = (12380);
 return;
}
function __ZN9ExitGames6Photon12TrafficStatsC2Ei($this,$packageHeaderSize) {
 $this = $this|0;
 $packageHeaderSize = $packageHeaderSize|0;
 var $0 = 0, $1 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (13580);
 $0 = ((($this)) + 4|0);
 HEAP32[$0>>2] = $packageHeaderSize;
 $1 = ((($this)) + 8|0);
 dest=$1; stop=dest+48|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 return;
}
function __ZN9ExitGames6Photon12TrafficStatsD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Common4BaseD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase22initUserDataEncryptionERKNS_6Common7JVectorIhEE($this,$secret) {
 $this = $this|0;
 $secret = $secret|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 20|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = ((($3)) + -8|0);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($5);
 }
 $6 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(40)|0);
 $7 = ((($6)) + 8|0);
 $8 = $6;
 $9 = $8;
 HEAP32[$9>>2] = 32;
 $10 = (($8) + 4)|0;
 $11 = $10;
 HEAP32[$11>>2] = 0;
 dest=$7; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
 $12 = HEAP32[$0>>2]|0;
 $13 = ((($12)) + 20|0);
 HEAP32[$13>>2] = $7;
 $14 = HEAP32[$0>>2]|0;
 $15 = ((($14)) + 20|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = ((($secret)) + 16|0);
 $18 = HEAP32[$17>>2]|0;
 dest=$16; src=$18; stop=dest+32|0; do { HEAP8[dest>>0]=HEAP8[src>>0]|0; dest=dest+1|0; src=src+1|0; } while ((dest|0) < (stop|0));
 $19 = HEAP32[$0>>2]|0;
 $20 = ((($19)) + 24|0);
 HEAP8[$20>>0] = 0;
 $21 = HEAP32[$0>>2]|0;
 $22 = ((($21)) + 25|0);
 HEAP8[$22>>0] = 1;
 return;
}
function __ZNK9ExitGames6Photon8Internal8PeerBase16getRoundTripTimeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 60|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase24getRoundTripTimeVarianceEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 64|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase25getTimestampOfLastReceiveEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 84|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerData26getTrafficStatsEnabledTimeEv($this) {
 $this = $this|0;
 var $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 44|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0);
 $3 = ((($this)) + 40|0);
 $4 = HEAP32[$3>>2]|0;
 if ($2) {
  $$0 = $4;
  return ($$0|0);
 }
 $5 = (__Z11getTimeUnixv()|0);
 $6 = (($5) + ($4))|0;
 $7 = HEAP32[$0>>2]|0;
 $8 = (($6) - ($7))|0;
 $$0 = $8;
 return ($$0|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase24getTrafficStatsGameLevelEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase23getTrafficStatsIncomingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase23getTrafficStatsOutgoingEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 108|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Photon8Internal8PeerBase32resetTrafficStatsMaximumCountersEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 112|0);
 $1 = HEAP32[$0>>2]|0;
 __ZN9ExitGames6Photon21TrafficStatsGameLevel20resetMaximumCountersEv($1);
 return;
}
function __ZN9ExitGames6Photon21TrafficStatsGameLevel20resetMaximumCountersEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 52|0);
 HEAP32[$0>>2] = 0;
 $1 = ((($this)) + 56|0);
 HEAP32[$1>>2] = 0;
 $2 = ((($this)) + 44|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($this)) + 48|0);
 HEAP8[$3>>0] = 0;
 $4 = ((($this)) + 36|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($this)) + 40|0);
 HEAP8[$5>>0] = 0;
 $6 = ((($this)) + 4|0);
 HEAP32[$6>>2] = 0;
 $7 = ((($this)) + 8|0);
 HEAP32[$7>>2] = 0;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerBase17resetTrafficStatsEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 104|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (__ZNK9ExitGames6Photon12TrafficStats20getPackageHeaderSizeEv($1)|0);
 $3 = HEAP32[$0>>2]|0;
 $4 = ($3|0)==(0|0);
 if (!($4)) {
  $5 = HEAP32[$3>>2]|0;
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vi[$6 & 2047]($3);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($3);
 }
 $7 = ((($this)) + 108|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 if (!($9)) {
  $10 = HEAP32[$8>>2]|0;
  $11 = HEAP32[$10>>2]|0;
  FUNCTION_TABLE_vi[$11 & 2047]($8);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($8);
 }
 $12 = ((($this)) + 112|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = ($13|0)==(0|0);
 if (!($14)) {
  $15 = HEAP32[$13>>2]|0;
  $16 = HEAP32[$15>>2]|0;
  FUNCTION_TABLE_vi[$16 & 2047]($13);
  __ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv($13);
 }
 $17 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(56)|0);
 __ZN9ExitGames6Photon8Internal19TrafficStatsCounterC2Ei($17,$2);
 HEAP32[$0>>2] = $17;
 $18 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(56)|0);
 __ZN9ExitGames6Photon8Internal19TrafficStatsCounterC2Ei($18,$2);
 HEAP32[$7>>2] = $18;
 $19 = (__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj(68)|0);
 dest=$19; stop=dest+68|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 __ZN9ExitGames6Photon21TrafficStatsGameLevelC2Ev($19);
 HEAP32[$19>>2] = (12748);
 HEAP32[$12>>2] = $19;
 $20 = ((($this)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = ((($21)) + 40|0);
 HEAP32[$22>>2] = 0;
 $23 = HEAP32[$20>>2]|0;
 $24 = ((($23)) + 44|0);
 HEAP32[$24>>2] = 0;
 return;
}
function __ZN9ExitGames6Photon8Internal8PeerData22setTrafficStatsEnabledEb($this,$value) {
 $this = $this|0;
 $value = $value|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__Z11getTimeUnixv()|0);
 $1 = ((($this)) + 44|0);
 if ($value) {
  $storemerge = $0;
  HEAP32[$1>>2] = $storemerge;
  return;
 }
 $2 = HEAP32[$1>>2]|0;
 $3 = (($0) - ($2))|0;
 $4 = ((($this)) + 40|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($3) + ($5))|0;
 HEAP32[$4>>2] = $6;
 $storemerge = 0;
 HEAP32[$1>>2] = $storemerge;
 return;
}
function __ZNK9ExitGames6Photon12TrafficStats20getPackageHeaderSizeEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZN9ExitGames6Photon10PhotonPeerD2Ev($this) {
 $this = $this|0;
 var $$pre$i = 0, $$pre$i1 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (14052);
 $0 = ((($this)) + 24|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = $1;
 $3 = $2;
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 4)|0;
 $6 = $5;
 $7 = HEAP32[$6>>2]|0;
 $8 = (_i64Add(($4|0),($7|0),-1,-1)|0);
 $9 = tempRet0;
 $10 = $1;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = ($8|0)==(0);
 $15 = ($9|0)==(0);
 $16 = $14 & $15;
 if ($16) {
  $17 = ((($this)) + 20|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = ($18|0)==(0|0);
  do {
   if ($19) {
    $25 = $1;
   } else {
    $20 = HEAP32[$18>>2]|0;
    $21 = HEAP32[$20>>2]|0;
    FUNCTION_TABLE_vi[$21 & 2047]($18);
    __THREW__ = 0;
    invoke_vi(750,($18|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $28 = ___cxa_find_matching_catch_3(0|0)|0;
     $29 = tempRet0;
     ___clang_call_terminate($28);
     // unreachable;
    } else {
     $$pre$i = HEAP32[$0>>2]|0;
     $25 = $$pre$i;
     break;
    }
   }
  } while(0);
  $24 = ($25|0)==(0|0);
  if (!($24)) {
   __THREW__ = 0;
   invoke_vi(750,($25|0));
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_3(0|0)|0;
    $29 = tempRet0;
    ___clang_call_terminate($28);
    // unreachable;
   }
  }
 }
 $30 = ((($this)) + 16|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = $31;
 $33 = $32;
 $34 = HEAP32[$33>>2]|0;
 $35 = (($32) + 4)|0;
 $36 = $35;
 $37 = HEAP32[$36>>2]|0;
 $38 = (_i64Add(($34|0),($37|0),-1,-1)|0);
 $39 = tempRet0;
 $40 = $31;
 $41 = $40;
 HEAP32[$41>>2] = $38;
 $42 = (($40) + 4)|0;
 $43 = $42;
 HEAP32[$43>>2] = $39;
 $44 = ($38|0)==(0);
 $45 = ($39|0)==(0);
 $46 = $44 & $45;
 if (!($46)) {
  $58 = ((($this)) + 4|0);
  __ZN9ExitGames6Common6LoggerD2Ev($58);
  return;
 }
 $47 = ((($this)) + 12|0);
 $48 = HEAP32[$47>>2]|0;
 $49 = ($48|0)==(0|0);
 do {
  if ($49) {
   $53 = $31;
  } else {
   __ZN9ExitGames6Photon8Internal8PeerDataD2Ev($48);
   __THREW__ = 0;
   invoke_vi(750,($48|0));
   $50 = __THREW__; __THREW__ = 0;
   $51 = $50&1;
   if ($51) {
    $56 = ___cxa_find_matching_catch_3(0|0)|0;
    $57 = tempRet0;
    ___clang_call_terminate($56);
    // unreachable;
   } else {
    $$pre$i1 = HEAP32[$30>>2]|0;
    $53 = $$pre$i1;
    break;
   }
  }
 } while(0);
 $52 = ($53|0)==(0|0);
 if ($52) {
  $58 = ((($this)) + 4|0);
  __ZN9ExitGames6Common6LoggerD2Ev($58);
  return;
 }
 __THREW__ = 0;
 invoke_vi(750,($53|0));
 $54 = __THREW__; __THREW__ = 0;
 $55 = $54&1;
 if ($55) {
  $56 = ___cxa_find_matching_catch_3(0|0)|0;
  $57 = tempRet0;
  ___clang_call_terminate($56);
  // unreachable;
 } else {
  $58 = ((($this)) + 4|0);
  __ZN9ExitGames6Common6LoggerD2Ev($58);
  return;
 }
}
function __ZN9ExitGames6Photon8Internal8PeerDataD2Ev($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if ($2) {
  return;
 }
 $3 = ((($1)) + -8|0);
 __THREW__ = 0;
 invoke_vi(750,($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_3(0|0)|0;
  $7 = tempRet0;
  ___clang_call_terminate($6);
  // unreachable;
 } else {
  return;
 }
}
function __ZN9ExitGames6Photon10PhotonPeerD0Ev($this) {
 $this = $this|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN9ExitGames6Photon10PhotonPeerD2Ev($this);
 __ZdlPv($this);
 return;
}
function __ZNK9ExitGames6Photon9EventData13getParametersEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZNK9ExitGames6Photon8Internal8PeerBase12getPeerStateEv($this) {
 $this = $this|0;
 var $$ = 0, $$0 = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 36|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(3);
 if ($2) {
  $3 = ((($this)) + 97|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)==(0);
  if ($5) {
   $$0 = 2;
   return ($$0|0);
  }
 }
 $6 = ($1|0)==(6);
 $$ = $6 ? 4 : $1;
 $$0 = $$;
 return ($$0|0);
}
function __ZN9ExitGames6Photon8Internal8PeerDataC2ERNS0_14PhotonListenerEb($this,$listener,$usingObjC) {
 $this = $this|0;
 $listener = $listener|0;
 $usingObjC = $usingObjC|0;
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = $listener;
 $0 = ((($this)) + 94|0);
 $1 = $usingObjC&1;
 HEAP8[$0>>0] = $1;
 $2 = ((($this)) + 52|0);
 HEAP8[$2>>0] = -13;
 $3 = ((($this)) + 53|0);
 HEAP8[$3>>0] = 0;
 $4 = ((($this)) + 54|0);
 HEAP8[$4>>0] = 1;
 $5 = ((($this)) + 55|0);
 HEAP8[$5>>0] = 6;
 $6 = ((($this)) + 56|0);
 HEAP8[$6>>0] = 49;
 $7 = ((($this)) + 57|0);
 HEAP8[$7>>0] = 65;
 $8 = ((($this)) + 58|0);
 HEAP8[$8>>0] = 9;
 $9 = ((($this)) + 59|0);
 HEAP8[$9>>0] = 2;
 $10 = ((($this)) + 60|0);
 dest=$10; stop=dest+33|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
 $11 = HEAP16[83685]|0;
 $12 = (($11) + 1)<<16>>16;
 HEAP16[83685] = $12;
 $13 = ((($this)) + 4|0);
 HEAP8[$13>>0] = 20;
 $14 = ((($this)) + 8|0);
 HEAP32[$14>>2] = 10000;
 $15 = ((($this)) + 12|0);
 HEAP32[$15>>2] = 5;
 $16 = ((($this)) + 16|0);
 HEAP32[$16>>2] = 2000;
 $17 = ((($this)) + 20|0);
 HEAP32[$17>>2] = 0;
 $18 = ((($this)) + 24|0);
 HEAP8[$18>>0] = 0;
 $19 = ((($this)) + 25|0);
 HEAP8[$19>>0] = 0;
 $20 = ((($this)) + 26|0);
 HEAP16[$20>>1] = 100;
 $21 = ((($this)) + 28|0);
 HEAP16[$21>>1] = 100;
 $22 = ((($this)) + 30|0);
 HEAP16[$22>>1] = 100;
 $23 = ((($this)) + 32|0);
 HEAP16[$23>>1] = 100;
 $24 = ((($this)) + 34|0);
 HEAP16[$24>>1] = 100;
 $25 = ((($this)) + 36|0);
 HEAP16[$25>>1] = 100;
 $26 = ((($this)) + 38|0);
 HEAP8[$26>>0] = 0;
 $27 = ((($this)) + 40|0);
 HEAP32[$27>>2] = 0;
 $28 = ((($this)) + 44|0);
 HEAP32[$28>>2] = 0;
 $29 = ((($this)) + 93|0);
 HEAP8[$29>>0] = 0;
 $30 = ((($this)) + 48|0);
 HEAP32[$30>>2] = 0;
 $31 = (__Z11getTimeUnixv()|0);
 _srand($31);
 return;
}
function __ZN9ExitGames6Photon10PhotonPeerC2ERNS0_14PhotonListenerEh($this,$listener,$connectionProtocol) {
 $this = $this|0;
 $listener = $listener|0;
 $connectionProtocol = $connectionProtocol|0;
 var $$0 = 0, $$01 = 0, $$1 = 0, $$12 = 0, $$pre$i = 0, $$pre$i8 = 0, $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$this>>2] = (14052);
 $0 = ((($this)) + 4|0);
 __ZN9ExitGames6Common6LoggerC2Ei($0,2);
 $1 = ((($this)) + 12|0);
 __THREW__ = 0;
 $2 = (invoke_ii(775,96)|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if (!($4)) {
  __THREW__ = 0;
  invoke_viii(939,($2|0),($listener|0),0);
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if (!($6)) {
   HEAP32[$1>>2] = $2;
   $7 = ((($this)) + 16|0);
   __THREW__ = 0;
   $8 = (invoke_ii(775,8)|0);
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if (!($10)) {
    $11 = $8;
    $12 = $11;
    HEAP32[$12>>2] = 1;
    $13 = (($11) + 4)|0;
    $14 = $13;
    HEAP32[$14>>2] = 0;
    HEAP32[$7>>2] = $8;
    $15 = ((($this)) + 20|0);
    HEAP32[$15>>2] = 0;
    $16 = ((($this)) + 24|0);
    __THREW__ = 0;
    $17 = (invoke_ii(775,8)|0);
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $35 = ___cxa_find_matching_catch_2()|0;
     $36 = tempRet0;
     $$0 = $36;$$01 = $35;
    } else {
     $20 = $17;
     $21 = $20;
     HEAP32[$21>>2] = 1;
     $22 = (($20) + 4)|0;
     $23 = $22;
     HEAP32[$23>>2] = 0;
     HEAP32[$16>>2] = $17;
     $24 = HEAP32[$listener>>2]|0;
     $25 = ((($24)) + -20|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = (($listener) + ($26)|0);
     __THREW__ = 0;
     invoke_vii(940,($0|0),($27|0));
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if (!($29)) {
      $30 = ((($this)) + 28|0);
      HEAP8[$30>>0] = $connectionProtocol;
      __THREW__ = 0;
      invoke_vi(941,($this|0));
      $31 = __THREW__; __THREW__ = 0;
      $32 = $31&1;
      if (!($32)) {
       return;
      }
     }
     $37 = ___cxa_find_matching_catch_2()|0;
     $38 = tempRet0;
     $39 = HEAP32[$16>>2]|0;
     $40 = $39;
     $41 = $40;
     $42 = HEAP32[$41>>2]|0;
     $43 = (($40) + 4)|0;
     $44 = $43;
     $45 = HEAP32[$44>>2]|0;
     $46 = (_i64Add(($42|0),($45|0),-1,-1)|0);
     $47 = tempRet0;
     $48 = $39;
     $49 = $48;
     HEAP32[$49>>2] = $46;
     $50 = (($48) + 4)|0;
     $51 = $50;
     HEAP32[$51>>2] = $47;
     $52 = ($46|0)==(0);
     $53 = ($47|0)==(0);
     $54 = $52 & $53;
     if ($54) {
      $55 = HEAP32[$15>>2]|0;
      $56 = ($55|0)==(0|0);
      do {
       if ($56) {
        $62 = $39;
       } else {
        $57 = HEAP32[$55>>2]|0;
        $58 = HEAP32[$57>>2]|0;
        FUNCTION_TABLE_vi[$58 & 2047]($55);
        __THREW__ = 0;
        invoke_vi(750,($55|0));
        $59 = __THREW__; __THREW__ = 0;
        $60 = $59&1;
        if ($60) {
         $65 = ___cxa_find_matching_catch_3(0|0)|0;
         $66 = tempRet0;
         ___clang_call_terminate($65);
         // unreachable;
        } else {
         $$pre$i = HEAP32[$16>>2]|0;
         $62 = $$pre$i;
         break;
        }
       }
      } while(0);
      $61 = ($62|0)==(0|0);
      if ($61) {
       $$0 = $38;$$01 = $37;
      } else {
       __THREW__ = 0;
       invoke_vi(750,($62|0));
       $63 = __THREW__; __THREW__ = 0;
       $64 = $63&1;
       if ($64) {
        $65 = ___cxa_find_matching_catch_3(0|0)|0;
        $66 = tempRet0;
        ___clang_call_terminate($65);
        // unreachable;
       } else {
        $$0 = $38;$$01 = $37;
       }
      }
     } else {
      $$0 = $38;$$01 = $37;
     }
    }
    $67 = HEAP32[$7>>2]|0;
    $68 = $67;
    $69 = $68;
    $70 = HEAP32[$69>>2]|0;
    $71 = (($68) + 4)|0;
    $72 = $71;
    $73 = HEAP32[$72>>2]|0;
    $74 = (_i64Add(($70|0),($73|0),-1,-1)|0);
    $75 = tempRet0;
    $76 = $67;
    $77 = $76;
    HEAP32[$77>>2] = $74;
    $78 = (($76) + 4)|0;
    $79 = $78;
    HEAP32[$79>>2] = $75;
    $80 = ($74|0)==(0);
    $81 = ($75|0)==(0);
    $82 = $80 & $81;
    if (!($82)) {
     $$1 = $$0;$$12 = $$01;
     __ZN9ExitGames6Common6LoggerD2Ev($0);
     ___resumeException($$12|0);
     // unreachable;
    }
    $83 = HEAP32[$1>>2]|0;
    $84 = ($83|0)==(0|0);
    do {
     if ($84) {
      $88 = $67;
     } else {
      __ZN9ExitGames6Photon8Internal8PeerDataD2Ev($83);
      __THREW__ = 0;
      invoke_vi(750,($83|0));
      $85 = __THREW__; __THREW__ = 0;
      $86 = $85&1;
      if ($86) {
       $91 = ___cxa_find_matching_catch_3(0|0)|0;
       $92 = tempRet0;
       ___clang_call_terminate($91);
       // unreachable;
      } else {
       $$pre$i8 = HEAP32[$7>>2]|0;
       $88 = $$pre$i8;
       break;
      }
     }
    } while(0);
    $87 = ($88|0)==(0|0);
    if ($87) {
     $$1 = $$0;$$12 = $$01;
     __ZN9ExitGames6Common6LoggerD2Ev($0);
     ___resumeException($$12|0);
     // unreachable;
    }
    __THREW__ = 0;
    invoke_vi(750,($88|0));
    $89 = __THREW__; __THREW__ = 0;
    $90 = $89&1;
    if ($90) {
     $91 = ___cxa_find_matching_catch_3(0|0)|0;
     $92 = tempRet0;
     ___clang_call_terminate($91);
     // unreachable;
    } else {
     $$1 = $$0;$$12 = $$01;
     __ZN9ExitGames6Common6LoggerD2Ev($0);
     ___resumeException($$12|0);
     // unreachable;
    }
   }
  }
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 $$1 = $34;$$12 = $33;
 __ZN9ExitGames6Common6LoggerD2Ev($0);
 ___resumeException($$12|0);
 // unreachable;
}
function __ZN9ExitGames6Photon16OperationRequest13getParametersEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZNK9ExitGames6Photon17OperationResponse8toStringEbbb($agg$result,$this,$withDebugMessage,$withParameters,$withParameterTypes) {
 $agg$result = $agg$result|0;
 $this = $this|0;
 $withDebugMessage = $withDebugMessage|0;
 $withParameters = $withParameters|0;
 $withParameterTypes = $withParameterTypes|0;
 var $$0 = 0, $$01144 = 0, $$04 = 0, $$0745 = 0, $$112 = 0, $$18 = 0, $$213 = 0, $$29 = 0, $$31047 = 0, $$31446 = 0, $$4$ph = 0, $$415$ph = 0, $$5 = 0, $$516 = 0, $$6 = 0, $$617 = 0, $$7 = 0, $$718 = 0, $$8 = 0, $$819 = 0;
 var $0 = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $eh$lpad$body$index22Z2D = 0, $eh$lpad$body$indexZ2D = 0, $eh$lpad$body22$index27Z2D = 0, $eh$lpad$body22$indexZ2D = 0, $eh$lpad$body25$index32Z2D = 0, $eh$lpad$body25$indexZ2D = 0, $eh$lpad$body31$index2Z2D = 0, $eh$lpad$body31$indexZ2D = 0, $eh$lpad$body35$index7Z2D = 0, $eh$lpad$body35$indexZ2D = 0, $eh$lpad$body39$index12Z2D = 0, $eh$lpad$body39$indexZ2D = 0, $eh$lpad$body43$index17Z2D = 0, $eh$lpad$body43$indexZ2D = 0, $str$i = 0, $str$i32 = 0, $str$i36 = 0;
 var $str$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(288|0);
 $str$i40 = sp + 272|0;
 $str$i36 = sp + 256|0;
 $str$i32 = sp + 240|0;
 $str$i = sp + 224|0;
 $0 = sp + 208|0;
 $1 = sp + 192|0;
 $2 = sp + 176|0;
 $3 = sp + 160|0;
 $4 = sp + 144|0;
 $5 = sp + 128|0;
 $6 = sp + 112|0;
 $7 = sp + 96|0;
 $8 = sp + 80|0;
 $9 = sp + 64|0;
 $10 = sp + 48|0;
 $11 = sp + 32|0;
 $12 = sp + 16|0;
 $13 = sp;
 __ZN9ExitGames6Common7JStringC2EPKw($8,14120);
 __THREW__ = 0;
 invoke_vii(786,($3|0),($8|0));
 $14 = __THREW__; __THREW__ = 0;
 $15 = $14&1;
 if ($15) {
  $103 = ___cxa_find_matching_catch_2()|0;
  $104 = tempRet0;
  $eh$lpad$body$index22Z2D = $104;$eh$lpad$body$indexZ2D = $103;
 } else {
  __THREW__ = 0;
  invoke_vii(787,($str$i|0),0);
  $16 = __THREW__; __THREW__ = 0;
  $17 = $16&1;
  do {
   if ($17) {
    label = 7;
   } else {
    $18 = HEAP8[$this>>0]|0;
    __THREW__ = 0;
    (invoke_iii(782,($str$i|0),($18|0))|0);
    $19 = __THREW__; __THREW__ = 0;
    $20 = $19&1;
    if (!($20)) {
     __THREW__ = 0;
     $21 = (invoke_iii(760,($3|0),($str$i|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if (!($23)) {
      __ZN9ExitGames6Common7JStringD2Ev($str$i);
      __THREW__ = 0;
      invoke_vii(786,($7|0),($21|0));
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      if ($27) {
       label = 7;
       break;
      }
      __ZN9ExitGames6Common7JStringD2Ev($3);
      __THREW__ = 0;
      invoke_vii(786,($2|0),($7|0));
      $30 = __THREW__; __THREW__ = 0;
      $31 = $30&1;
      L10: do {
       if ($31) {
        $105 = ___cxa_find_matching_catch_2()|0;
        $106 = tempRet0;
        $eh$lpad$body22$index27Z2D = $106;$eh$lpad$body22$indexZ2D = $105;
        label = 56;
       } else {
        __THREW__ = 0;
        invoke_vii(787,($str$i32|0),0);
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        do {
         if ($33) {
          label = 15;
         } else {
          __THREW__ = 0;
          (invoke_iii(759,($str$i32|0),(14264|0))|0);
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if (!($35)) {
           __THREW__ = 0;
           $36 = (invoke_iii(760,($2|0),($str$i32|0))|0);
           $37 = __THREW__; __THREW__ = 0;
           $38 = $37&1;
           if (!($38)) {
            __ZN9ExitGames6Common7JStringD2Ev($str$i32);
            __THREW__ = 0;
            invoke_vii(786,($6|0),($36|0));
            $41 = __THREW__; __THREW__ = 0;
            $42 = $41&1;
            if ($42) {
             label = 15;
             break;
            }
            __ZN9ExitGames6Common7JStringD2Ev($2);
            $45 = ((($this)) + 2|0);
            __THREW__ = 0;
            invoke_vii(786,($1|0),($6|0));
            $46 = __THREW__; __THREW__ = 0;
            $47 = $46&1;
            L19: do {
             if ($47) {
              $107 = ___cxa_find_matching_catch_2()|0;
              $108 = tempRet0;
              $eh$lpad$body25$index32Z2D = $108;$eh$lpad$body25$indexZ2D = $107;
              label = 58;
             } else {
              __THREW__ = 0;
              invoke_vii(787,($str$i36|0),0);
              $48 = __THREW__; __THREW__ = 0;
              $49 = $48&1;
              L22: do {
               if ($49) {
                label = 23;
               } else {
                $50 = HEAP16[$45>>1]|0;
                __THREW__ = 0;
                (invoke_iii(832,($str$i36|0),($50|0))|0);
                $51 = __THREW__; __THREW__ = 0;
                $52 = $51&1;
                do {
                 if (!($52)) {
                  __THREW__ = 0;
                  $53 = (invoke_iii(760,($1|0),($str$i36|0))|0);
                  $54 = __THREW__; __THREW__ = 0;
                  $55 = $54&1;
                  if ($55) {
                   break;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($str$i36);
                  __THREW__ = 0;
                  invoke_vii(786,($5|0),($53|0));
                  $58 = __THREW__; __THREW__ = 0;
                  $59 = $58&1;
                  if ($59) {
                   label = 23;
                   break L22;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($1);
                  L28: do {
                   if ($withDebugMessage) {
                    $62 = ((($this)) + 4|0);
                    __THREW__ = 0;
                    $63 = (invoke_ii(820,($62|0))|0);
                    $64 = __THREW__; __THREW__ = 0;
                    $65 = $64&1;
                    do {
                     if (!($65)) {
                      $66 = ($63|0)==(0);
                      if ($66) {
                       label = 40;
                       break L28;
                      }
                      __THREW__ = 0;
                      invoke_vii(801,($11|0),(14324|0));
                      $67 = __THREW__; __THREW__ = 0;
                      $68 = $67&1;
                      if ($68) {
                       break;
                      }
                      __THREW__ = 0;
                      invoke_viii(785,($10|0),($11|0),($62|0));
                      $69 = __THREW__; __THREW__ = 0;
                      $70 = $69&1;
                      if ($70) {
                       $111 = ___cxa_find_matching_catch_2()|0;
                       $112 = tempRet0;
                       $$4$ph = $112;$$415$ph = $111;
                       label = 70;
                       break L28;
                      }
                      __THREW__ = 0;
                      invoke_vii(786,($0|0),($10|0));
                      $71 = __THREW__; __THREW__ = 0;
                      $72 = $71&1;
                      if ($72) {
                       $73 = ___cxa_find_matching_catch_2()|0;
                       $74 = tempRet0;
                       $$31047 = $74;$$31446 = $73;
                       label = 69;
                       break L28;
                      }
                      __THREW__ = 0;
                      invoke_vii(787,($str$i40|0),0);
                      $75 = __THREW__; __THREW__ = 0;
                      $76 = $75&1;
                      L40: do {
                       if ($76) {
                        label = 37;
                       } else {
                        __THREW__ = 0;
                        (invoke_iii(759,($str$i40|0),(14336|0))|0);
                        $77 = __THREW__; __THREW__ = 0;
                        $78 = $77&1;
                        do {
                         if (!($78)) {
                          __THREW__ = 0;
                          $79 = (invoke_iii(760,($0|0),($str$i40|0))|0);
                          $80 = __THREW__; __THREW__ = 0;
                          $81 = $80&1;
                          if ($81) {
                           break;
                          }
                          __ZN9ExitGames6Common7JStringD2Ev($str$i40);
                          __THREW__ = 0;
                          invoke_vii(786,($9|0),($79|0));
                          $84 = __THREW__; __THREW__ = 0;
                          $85 = $84&1;
                          if ($85) {
                           label = 37;
                           break L40;
                          }
                          __ZN9ExitGames6Common7JStringD2Ev($0);
                          $$04 = 1;
                          label = 41;
                          break L28;
                         }
                        } while(0);
                        $82 = ___cxa_find_matching_catch_2()|0;
                        $83 = tempRet0;
                        __ZN9ExitGames6Common7JStringD2Ev($str$i40);
                        $eh$lpad$body43$index17Z2D = $83;$eh$lpad$body43$indexZ2D = $82;
                       }
                      } while(0);
                      if ((label|0) == 37) {
                       $86 = ___cxa_find_matching_catch_2()|0;
                       $87 = tempRet0;
                       $eh$lpad$body43$index17Z2D = $87;$eh$lpad$body43$indexZ2D = $86;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($0);
                      $$31047 = $eh$lpad$body43$index17Z2D;$$31446 = $eh$lpad$body43$indexZ2D;
                      label = 69;
                      break L28;
                     }
                    } while(0);
                    $109 = ___cxa_find_matching_catch_2()|0;
                    $110 = tempRet0;
                    $$5 = $110;$$516 = $109;
                   } else {
                    label = 40;
                   }
                  } while(0);
                  do {
                   if ((label|0) == 40) {
                    __THREW__ = 0;
                    invoke_vii(801,($9|0),(163772|0));
                    $88 = __THREW__; __THREW__ = 0;
                    $89 = $88&1;
                    if (!($89)) {
                     $$04 = 0;
                     label = 41;
                     break;
                    }
                    $121 = ___cxa_find_matching_catch_2()|0;
                    $122 = tempRet0;
                    $$5 = $122;$$516 = $121;
                   }
                  } while(0);
                  if ((label|0) == 41) {
                   __THREW__ = 0;
                   invoke_viii(785,($4|0),($5|0),($9|0));
                   $90 = __THREW__; __THREW__ = 0;
                   $91 = $90&1;
                   if ($91) {
                    $113 = ___cxa_find_matching_catch_2()|0;
                    $114 = tempRet0;
                    $$213 = $113;$$29 = $114;
                   } else {
                    do {
                     if ($withParameters) {
                      $92 = ((($this)) + 20|0);
                      __THREW__ = 0;
                      invoke_viii(783,($13|0),($92|0),($withParameterTypes|0));
                      $93 = __THREW__; __THREW__ = 0;
                      $94 = $93&1;
                      if ($94) {
                       $115 = ___cxa_find_matching_catch_2()|0;
                       $116 = tempRet0;
                       $$112 = $115;$$18 = $116;
                       break;
                      }
                      __THREW__ = 0;
                      invoke_viii(860,($12|0),(14344|0),($13|0));
                      $95 = __THREW__; __THREW__ = 0;
                      $96 = $95&1;
                      if (!($96)) {
                       $$0 = 1;
                       label = 47;
                       break;
                      }
                      $97 = ___cxa_find_matching_catch_2()|0;
                      $98 = tempRet0;
                      $$01144 = $97;$$0745 = $98;
                      label = 65;
                     } else {
                      __THREW__ = 0;
                      invoke_vii(801,($12|0),(163772|0));
                      $99 = __THREW__; __THREW__ = 0;
                      $100 = $99&1;
                      if (!($100)) {
                       $$0 = 0;
                       label = 47;
                       break;
                      }
                      $119 = ___cxa_find_matching_catch_2()|0;
                      $120 = tempRet0;
                      $$112 = $119;$$18 = $120;
                     }
                    } while(0);
                    do {
                     if ((label|0) == 47) {
                      __THREW__ = 0;
                      invoke_viii(785,($agg$result|0),($4|0),($12|0));
                      $101 = __THREW__; __THREW__ = 0;
                      $102 = $101&1;
                      if ($102) {
                       $117 = ___cxa_find_matching_catch_2()|0;
                       $118 = tempRet0;
                       __ZN9ExitGames6Common7JStringD2Ev($12);
                       if ($$0) {
                        $$01144 = $117;$$0745 = $118;
                        label = 65;
                        break;
                       } else {
                        $$112 = $117;$$18 = $118;
                        break;
                       }
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($12);
                      if ($$0) {
                       __ZN9ExitGames6Common7JStringD2Ev($13);
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($4);
                      __ZN9ExitGames6Common7JStringD2Ev($9);
                      if (!($$04)) {
                       __ZN9ExitGames6Common7JStringD2Ev($5);
                       __ZN9ExitGames6Common7JStringD2Ev($6);
                       __ZN9ExitGames6Common7JStringD2Ev($7);
                       __ZN9ExitGames6Common7JStringD2Ev($8);
                       STACKTOP = sp;return;
                      }
                      __ZN9ExitGames6Common7JStringD2Ev($10);
                      __ZN9ExitGames6Common7JStringD2Ev($11);
                      __ZN9ExitGames6Common7JStringD2Ev($5);
                      __ZN9ExitGames6Common7JStringD2Ev($6);
                      __ZN9ExitGames6Common7JStringD2Ev($7);
                      __ZN9ExitGames6Common7JStringD2Ev($8);
                      STACKTOP = sp;return;
                     }
                    } while(0);
                    if ((label|0) == 65) {
                     __ZN9ExitGames6Common7JStringD2Ev($13);
                     $$112 = $$01144;$$18 = $$0745;
                    }
                    __ZN9ExitGames6Common7JStringD2Ev($4);
                    $$213 = $$112;$$29 = $$18;
                   }
                   __ZN9ExitGames6Common7JStringD2Ev($9);
                   if ($$04) {
                    $$31047 = $$29;$$31446 = $$213;
                    label = 69;
                   } else {
                    $$5 = $$29;$$516 = $$213;
                   }
                  }
                  if ((label|0) == 69) {
                   __ZN9ExitGames6Common7JStringD2Ev($10);
                   $$4$ph = $$31047;$$415$ph = $$31446;
                   label = 70;
                  }
                  if ((label|0) == 70) {
                   __ZN9ExitGames6Common7JStringD2Ev($11);
                   $$5 = $$4$ph;$$516 = $$415$ph;
                  }
                  __ZN9ExitGames6Common7JStringD2Ev($5);
                  $$6 = $$5;$$617 = $$516;
                  break L19;
                 }
                } while(0);
                $56 = ___cxa_find_matching_catch_2()|0;
                $57 = tempRet0;
                __ZN9ExitGames6Common7JStringD2Ev($str$i36);
                $eh$lpad$body39$index12Z2D = $57;$eh$lpad$body39$indexZ2D = $56;
               }
              } while(0);
              if ((label|0) == 23) {
               $60 = ___cxa_find_matching_catch_2()|0;
               $61 = tempRet0;
               $eh$lpad$body39$index12Z2D = $61;$eh$lpad$body39$indexZ2D = $60;
              }
              __ZN9ExitGames6Common7JStringD2Ev($1);
              $eh$lpad$body25$index32Z2D = $eh$lpad$body39$index12Z2D;$eh$lpad$body25$indexZ2D = $eh$lpad$body39$indexZ2D;
              label = 58;
             }
            } while(0);
            if ((label|0) == 58) {
             $$6 = $eh$lpad$body25$index32Z2D;$$617 = $eh$lpad$body25$indexZ2D;
            }
            __ZN9ExitGames6Common7JStringD2Ev($6);
            $$7 = $$6;$$718 = $$617;
            break L10;
           }
          }
          $39 = ___cxa_find_matching_catch_2()|0;
          $40 = tempRet0;
          __ZN9ExitGames6Common7JStringD2Ev($str$i32);
          $eh$lpad$body35$index7Z2D = $40;$eh$lpad$body35$indexZ2D = $39;
         }
        } while(0);
        if ((label|0) == 15) {
         $43 = ___cxa_find_matching_catch_2()|0;
         $44 = tempRet0;
         $eh$lpad$body35$index7Z2D = $44;$eh$lpad$body35$indexZ2D = $43;
        }
        __ZN9ExitGames6Common7JStringD2Ev($2);
        $eh$lpad$body22$index27Z2D = $eh$lpad$body35$index7Z2D;$eh$lpad$body22$indexZ2D = $eh$lpad$body35$indexZ2D;
        label = 56;
       }
      } while(0);
      if ((label|0) == 56) {
       $$7 = $eh$lpad$body22$index27Z2D;$$718 = $eh$lpad$body22$indexZ2D;
      }
      __ZN9ExitGames6Common7JStringD2Ev($7);
      $$8 = $$7;$$819 = $$718;
      __ZN9ExitGames6Common7JStringD2Ev($8);
      ___resumeException($$819|0);
      // unreachable;
     }
    }
    $24 = ___cxa_find_matching_catch_2()|0;
    $25 = tempRet0;
    __ZN9ExitGames6Common7JStringD2Ev($str$i);
    $eh$lpad$body31$index2Z2D = $25;$eh$lpad$body31$indexZ2D = $24;
   }
  } while(0);
  if ((label|0) == 7) {
   $28 = ___cxa_find_matching_catch_2()|0;
   $29 = tempRet0;
   $eh$lpad$body31$index2Z2D = $29;$eh$lpad$body31$indexZ2D = $28;
  }
  __ZN9ExitGames6Common7JStringD2Ev($3);
  $eh$lpad$body$index22Z2D = $eh$lpad$body31$index2Z2D;$eh$lpad$body$indexZ2D = $eh$lpad$body31$indexZ2D;
 }
 $$8 = $eh$lpad$body$index22Z2D;$$819 = $eh$lpad$body$indexZ2D;
 __ZN9ExitGames6Common7JStringD2Ev($8);
 ___resumeException($$819|0);
 // unreachable;
}
function __ZNK9ExitGames6Photon17OperationResponse15getDebugMessageEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 return ($0|0);
}
function __ZNK9ExitGames6Photon17OperationResponse13getParametersEv($this) {
 $this = $this|0;
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 return ($0|0);
}
function __ZNK9ExitGames6Photon10PhotonPeer12getPeerStateEv($this) {
 $this = $this|0;
 var $0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 20|0);
 $1 = HEAP32[$0>>2]|0;
 $2 = (__ZNK9ExitGames6Photon8Internal8PeerBase12getPeerStateEv($1)|0);
 return ($2|0);
}
function __ZN9ExitGames6Photon10PhotonPeer19setDebugOutputLevelEi($this,$debugLevel) {
 $this = $this|0;
 $debugLevel = $debugLevel|0;
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = ((($this)) + 4|0);
 $1 = (__ZN9ExitGames6Common6Logger19setDebugOutputLevelEi($0,$debugLevel)|0);
 return ($1|0);
}
function _malloc($0) {
 $0 = $0|0;
 var $$$0172$i = 0, $$$0173$i = 0, $$$4236$i = 0, $$$4329$i = 0, $$$i = 0, $$0 = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i20$i = 0, $$01$i$i = 0, $$0172$lcssa$i = 0, $$01726$i = 0, $$0173$lcssa$i = 0, $$01735$i = 0, $$0192 = 0, $$0194 = 0, $$0201$i$i = 0, $$0202$i$i = 0, $$0206$i$i = 0;
 var $$0207$i$i = 0, $$024370$i = 0, $$0260$i$i = 0, $$0261$i$i = 0, $$0262$i$i = 0, $$0268$i$i = 0, $$0269$i$i = 0, $$0320$i = 0, $$0322$i = 0, $$0323$i = 0, $$0325$i = 0, $$0331$i = 0, $$0336$i = 0, $$0337$$i = 0, $$0337$i = 0, $$0339$i = 0, $$0340$i = 0, $$0345$i = 0, $$1176$i = 0, $$1178$i = 0;
 var $$124469$i = 0, $$1264$i$i = 0, $$1266$i$i = 0, $$1321$i = 0, $$1326$i = 0, $$1341$i = 0, $$1347$i = 0, $$1351$i = 0, $$2234243136$i = 0, $$2247$ph$i = 0, $$2253$ph$i = 0, $$2333$i = 0, $$3$i = 0, $$3$i$i = 0, $$3$i200 = 0, $$3328$i = 0, $$3349$i = 0, $$4$lcssa$i = 0, $$4$ph$i = 0, $$411$i = 0;
 var $$4236$i = 0, $$4329$lcssa$i = 0, $$432910$i = 0, $$4335$$4$i = 0, $$4335$ph$i = 0, $$43359$i = 0, $$723947$i = 0, $$748$i = 0, $$pre = 0, $$pre$i = 0, $$pre$i$i = 0, $$pre$i17$i = 0, $$pre$i195 = 0, $$pre$i210 = 0, $$pre$phi$i$iZ2D = 0, $$pre$phi$i18$iZ2D = 0, $$pre$phi$i211Z2D = 0, $$pre$phi$iZ2D = 0, $$pre$phiZ2D = 0, $$sink1$i = 0;
 var $$sink1$i$i = 0, $$sink14$i = 0, $$sink2$i = 0, $$sink2$i204 = 0, $$sink3$i = 0, $1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0;
 var $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0;
 var $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0;
 var $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0;
 var $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0;
 var $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0;
 var $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0;
 var $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0;
 var $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0;
 var $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0;
 var $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0;
 var $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0;
 var $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0;
 var $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0;
 var $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0;
 var $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0;
 var $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0;
 var $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0;
 var $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0;
 var $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $98 = 0, $99 = 0, $cond$i = 0, $cond$i$i = 0, $cond$i208 = 0, $exitcond$i$i = 0, $not$$i = 0;
 var $not$$i$i = 0, $not$$i197 = 0, $not$$i209 = 0, $not$1$i = 0, $not$1$i203 = 0, $not$3$i = 0, $not$5$i = 0, $or$cond$i = 0, $or$cond$i201 = 0, $or$cond1$i = 0, $or$cond10$i = 0, $or$cond11$i = 0, $or$cond11$not$i = 0, $or$cond12$i = 0, $or$cond2$i = 0, $or$cond2$i199 = 0, $or$cond49$i = 0, $or$cond5$i = 0, $or$cond50$i = 0, $or$cond7$i = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = ($0>>>0)<(245);
 do {
  if ($2) {
   $3 = ($0>>>0)<(11);
   $4 = (($0) + 11)|0;
   $5 = $4 & -8;
   $6 = $3 ? 16 : $5;
   $7 = $6 >>> 3;
   $8 = HEAP32[40944]|0;
   $9 = $8 >>> $7;
   $10 = $9 & 3;
   $11 = ($10|0)==(0);
   if (!($11)) {
    $12 = $9 & 1;
    $13 = $12 ^ 1;
    $14 = (($13) + ($7))|0;
    $15 = $14 << 1;
    $16 = (163816 + ($15<<2)|0);
    $17 = ((($16)) + 8|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($18)) + 8|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($16|0)==($20|0);
    if ($21) {
     $22 = 1 << $14;
     $23 = $22 ^ -1;
     $24 = $8 & $23;
     HEAP32[40944] = $24;
    } else {
     $25 = ((($20)) + 12|0);
     HEAP32[$25>>2] = $16;
     HEAP32[$17>>2] = $20;
    }
    $26 = $14 << 3;
    $27 = $26 | 3;
    $28 = ((($18)) + 4|0);
    HEAP32[$28>>2] = $27;
    $29 = (($18) + ($26)|0);
    $30 = ((($29)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = $31 | 1;
    HEAP32[$30>>2] = $32;
    $$0 = $19;
    STACKTOP = sp;return ($$0|0);
   }
   $33 = HEAP32[(163784)>>2]|0;
   $34 = ($6>>>0)>($33>>>0);
   if ($34) {
    $35 = ($9|0)==(0);
    if (!($35)) {
     $36 = $9 << $7;
     $37 = 2 << $7;
     $38 = (0 - ($37))|0;
     $39 = $37 | $38;
     $40 = $36 & $39;
     $41 = (0 - ($40))|0;
     $42 = $40 & $41;
     $43 = (($42) + -1)|0;
     $44 = $43 >>> 12;
     $45 = $44 & 16;
     $46 = $43 >>> $45;
     $47 = $46 >>> 5;
     $48 = $47 & 8;
     $49 = $48 | $45;
     $50 = $46 >>> $48;
     $51 = $50 >>> 2;
     $52 = $51 & 4;
     $53 = $49 | $52;
     $54 = $50 >>> $52;
     $55 = $54 >>> 1;
     $56 = $55 & 2;
     $57 = $53 | $56;
     $58 = $54 >>> $56;
     $59 = $58 >>> 1;
     $60 = $59 & 1;
     $61 = $57 | $60;
     $62 = $58 >>> $60;
     $63 = (($61) + ($62))|0;
     $64 = $63 << 1;
     $65 = (163816 + ($64<<2)|0);
     $66 = ((($65)) + 8|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = ((($67)) + 8|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ($65|0)==($69|0);
     if ($70) {
      $71 = 1 << $63;
      $72 = $71 ^ -1;
      $73 = $8 & $72;
      HEAP32[40944] = $73;
      $90 = $73;
     } else {
      $74 = ((($69)) + 12|0);
      HEAP32[$74>>2] = $65;
      HEAP32[$66>>2] = $69;
      $90 = $8;
     }
     $75 = $63 << 3;
     $76 = (($75) - ($6))|0;
     $77 = $6 | 3;
     $78 = ((($67)) + 4|0);
     HEAP32[$78>>2] = $77;
     $79 = (($67) + ($6)|0);
     $80 = $76 | 1;
     $81 = ((($79)) + 4|0);
     HEAP32[$81>>2] = $80;
     $82 = (($79) + ($76)|0);
     HEAP32[$82>>2] = $76;
     $83 = ($33|0)==(0);
     if (!($83)) {
      $84 = HEAP32[(163796)>>2]|0;
      $85 = $33 >>> 3;
      $86 = $85 << 1;
      $87 = (163816 + ($86<<2)|0);
      $88 = 1 << $85;
      $89 = $90 & $88;
      $91 = ($89|0)==(0);
      if ($91) {
       $92 = $90 | $88;
       HEAP32[40944] = $92;
       $$pre = ((($87)) + 8|0);
       $$0194 = $87;$$pre$phiZ2D = $$pre;
      } else {
       $93 = ((($87)) + 8|0);
       $94 = HEAP32[$93>>2]|0;
       $$0194 = $94;$$pre$phiZ2D = $93;
      }
      HEAP32[$$pre$phiZ2D>>2] = $84;
      $95 = ((($$0194)) + 12|0);
      HEAP32[$95>>2] = $84;
      $96 = ((($84)) + 8|0);
      HEAP32[$96>>2] = $$0194;
      $97 = ((($84)) + 12|0);
      HEAP32[$97>>2] = $87;
     }
     HEAP32[(163784)>>2] = $76;
     HEAP32[(163796)>>2] = $79;
     $$0 = $68;
     STACKTOP = sp;return ($$0|0);
    }
    $98 = HEAP32[(163780)>>2]|0;
    $99 = ($98|0)==(0);
    if ($99) {
     $$0192 = $6;
    } else {
     $100 = (0 - ($98))|0;
     $101 = $98 & $100;
     $102 = (($101) + -1)|0;
     $103 = $102 >>> 12;
     $104 = $103 & 16;
     $105 = $102 >>> $104;
     $106 = $105 >>> 5;
     $107 = $106 & 8;
     $108 = $107 | $104;
     $109 = $105 >>> $107;
     $110 = $109 >>> 2;
     $111 = $110 & 4;
     $112 = $108 | $111;
     $113 = $109 >>> $111;
     $114 = $113 >>> 1;
     $115 = $114 & 2;
     $116 = $112 | $115;
     $117 = $113 >>> $115;
     $118 = $117 >>> 1;
     $119 = $118 & 1;
     $120 = $116 | $119;
     $121 = $117 >>> $119;
     $122 = (($120) + ($121))|0;
     $123 = (164080 + ($122<<2)|0);
     $124 = HEAP32[$123>>2]|0;
     $125 = ((($124)) + 4|0);
     $126 = HEAP32[$125>>2]|0;
     $127 = $126 & -8;
     $128 = (($127) - ($6))|0;
     $129 = ((($124)) + 16|0);
     $130 = HEAP32[$129>>2]|0;
     $not$3$i = ($130|0)==(0|0);
     $$sink14$i = $not$3$i&1;
     $131 = (((($124)) + 16|0) + ($$sink14$i<<2)|0);
     $132 = HEAP32[$131>>2]|0;
     $133 = ($132|0)==(0|0);
     if ($133) {
      $$0172$lcssa$i = $124;$$0173$lcssa$i = $128;
     } else {
      $$01726$i = $124;$$01735$i = $128;$135 = $132;
      while(1) {
       $134 = ((($135)) + 4|0);
       $136 = HEAP32[$134>>2]|0;
       $137 = $136 & -8;
       $138 = (($137) - ($6))|0;
       $139 = ($138>>>0)<($$01735$i>>>0);
       $$$0173$i = $139 ? $138 : $$01735$i;
       $$$0172$i = $139 ? $135 : $$01726$i;
       $140 = ((($135)) + 16|0);
       $141 = HEAP32[$140>>2]|0;
       $not$$i = ($141|0)==(0|0);
       $$sink1$i = $not$$i&1;
       $142 = (((($135)) + 16|0) + ($$sink1$i<<2)|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ($143|0)==(0|0);
       if ($144) {
        $$0172$lcssa$i = $$$0172$i;$$0173$lcssa$i = $$$0173$i;
        break;
       } else {
        $$01726$i = $$$0172$i;$$01735$i = $$$0173$i;$135 = $143;
       }
      }
     }
     $145 = (($$0172$lcssa$i) + ($6)|0);
     $146 = ($$0172$lcssa$i>>>0)<($145>>>0);
     if ($146) {
      $147 = ((($$0172$lcssa$i)) + 24|0);
      $148 = HEAP32[$147>>2]|0;
      $149 = ((($$0172$lcssa$i)) + 12|0);
      $150 = HEAP32[$149>>2]|0;
      $151 = ($150|0)==($$0172$lcssa$i|0);
      do {
       if ($151) {
        $156 = ((($$0172$lcssa$i)) + 20|0);
        $157 = HEAP32[$156>>2]|0;
        $158 = ($157|0)==(0|0);
        if ($158) {
         $159 = ((($$0172$lcssa$i)) + 16|0);
         $160 = HEAP32[$159>>2]|0;
         $161 = ($160|0)==(0|0);
         if ($161) {
          $$3$i = 0;
          break;
         } else {
          $$1176$i = $160;$$1178$i = $159;
         }
        } else {
         $$1176$i = $157;$$1178$i = $156;
        }
        while(1) {
         $162 = ((($$1176$i)) + 20|0);
         $163 = HEAP32[$162>>2]|0;
         $164 = ($163|0)==(0|0);
         if (!($164)) {
          $$1176$i = $163;$$1178$i = $162;
          continue;
         }
         $165 = ((($$1176$i)) + 16|0);
         $166 = HEAP32[$165>>2]|0;
         $167 = ($166|0)==(0|0);
         if ($167) {
          break;
         } else {
          $$1176$i = $166;$$1178$i = $165;
         }
        }
        HEAP32[$$1178$i>>2] = 0;
        $$3$i = $$1176$i;
       } else {
        $152 = ((($$0172$lcssa$i)) + 8|0);
        $153 = HEAP32[$152>>2]|0;
        $154 = ((($153)) + 12|0);
        HEAP32[$154>>2] = $150;
        $155 = ((($150)) + 8|0);
        HEAP32[$155>>2] = $153;
        $$3$i = $150;
       }
      } while(0);
      $168 = ($148|0)==(0|0);
      do {
       if (!($168)) {
        $169 = ((($$0172$lcssa$i)) + 28|0);
        $170 = HEAP32[$169>>2]|0;
        $171 = (164080 + ($170<<2)|0);
        $172 = HEAP32[$171>>2]|0;
        $173 = ($$0172$lcssa$i|0)==($172|0);
        if ($173) {
         HEAP32[$171>>2] = $$3$i;
         $cond$i = ($$3$i|0)==(0|0);
         if ($cond$i) {
          $174 = 1 << $170;
          $175 = $174 ^ -1;
          $176 = $98 & $175;
          HEAP32[(163780)>>2] = $176;
          break;
         }
        } else {
         $177 = ((($148)) + 16|0);
         $178 = HEAP32[$177>>2]|0;
         $not$1$i = ($178|0)!=($$0172$lcssa$i|0);
         $$sink2$i = $not$1$i&1;
         $179 = (((($148)) + 16|0) + ($$sink2$i<<2)|0);
         HEAP32[$179>>2] = $$3$i;
         $180 = ($$3$i|0)==(0|0);
         if ($180) {
          break;
         }
        }
        $181 = ((($$3$i)) + 24|0);
        HEAP32[$181>>2] = $148;
        $182 = ((($$0172$lcssa$i)) + 16|0);
        $183 = HEAP32[$182>>2]|0;
        $184 = ($183|0)==(0|0);
        if (!($184)) {
         $185 = ((($$3$i)) + 16|0);
         HEAP32[$185>>2] = $183;
         $186 = ((($183)) + 24|0);
         HEAP32[$186>>2] = $$3$i;
        }
        $187 = ((($$0172$lcssa$i)) + 20|0);
        $188 = HEAP32[$187>>2]|0;
        $189 = ($188|0)==(0|0);
        if (!($189)) {
         $190 = ((($$3$i)) + 20|0);
         HEAP32[$190>>2] = $188;
         $191 = ((($188)) + 24|0);
         HEAP32[$191>>2] = $$3$i;
        }
       }
      } while(0);
      $192 = ($$0173$lcssa$i>>>0)<(16);
      if ($192) {
       $193 = (($$0173$lcssa$i) + ($6))|0;
       $194 = $193 | 3;
       $195 = ((($$0172$lcssa$i)) + 4|0);
       HEAP32[$195>>2] = $194;
       $196 = (($$0172$lcssa$i) + ($193)|0);
       $197 = ((($196)) + 4|0);
       $198 = HEAP32[$197>>2]|0;
       $199 = $198 | 1;
       HEAP32[$197>>2] = $199;
      } else {
       $200 = $6 | 3;
       $201 = ((($$0172$lcssa$i)) + 4|0);
       HEAP32[$201>>2] = $200;
       $202 = $$0173$lcssa$i | 1;
       $203 = ((($145)) + 4|0);
       HEAP32[$203>>2] = $202;
       $204 = (($145) + ($$0173$lcssa$i)|0);
       HEAP32[$204>>2] = $$0173$lcssa$i;
       $205 = ($33|0)==(0);
       if (!($205)) {
        $206 = HEAP32[(163796)>>2]|0;
        $207 = $33 >>> 3;
        $208 = $207 << 1;
        $209 = (163816 + ($208<<2)|0);
        $210 = 1 << $207;
        $211 = $8 & $210;
        $212 = ($211|0)==(0);
        if ($212) {
         $213 = $8 | $210;
         HEAP32[40944] = $213;
         $$pre$i = ((($209)) + 8|0);
         $$0$i = $209;$$pre$phi$iZ2D = $$pre$i;
        } else {
         $214 = ((($209)) + 8|0);
         $215 = HEAP32[$214>>2]|0;
         $$0$i = $215;$$pre$phi$iZ2D = $214;
        }
        HEAP32[$$pre$phi$iZ2D>>2] = $206;
        $216 = ((($$0$i)) + 12|0);
        HEAP32[$216>>2] = $206;
        $217 = ((($206)) + 8|0);
        HEAP32[$217>>2] = $$0$i;
        $218 = ((($206)) + 12|0);
        HEAP32[$218>>2] = $209;
       }
       HEAP32[(163784)>>2] = $$0173$lcssa$i;
       HEAP32[(163796)>>2] = $145;
      }
      $219 = ((($$0172$lcssa$i)) + 8|0);
      $$0 = $219;
      STACKTOP = sp;return ($$0|0);
     } else {
      $$0192 = $6;
     }
    }
   } else {
    $$0192 = $6;
   }
  } else {
   $220 = ($0>>>0)>(4294967231);
   if ($220) {
    $$0192 = -1;
   } else {
    $221 = (($0) + 11)|0;
    $222 = $221 & -8;
    $223 = HEAP32[(163780)>>2]|0;
    $224 = ($223|0)==(0);
    if ($224) {
     $$0192 = $222;
    } else {
     $225 = (0 - ($222))|0;
     $226 = $221 >>> 8;
     $227 = ($226|0)==(0);
     if ($227) {
      $$0336$i = 0;
     } else {
      $228 = ($222>>>0)>(16777215);
      if ($228) {
       $$0336$i = 31;
      } else {
       $229 = (($226) + 1048320)|0;
       $230 = $229 >>> 16;
       $231 = $230 & 8;
       $232 = $226 << $231;
       $233 = (($232) + 520192)|0;
       $234 = $233 >>> 16;
       $235 = $234 & 4;
       $236 = $235 | $231;
       $237 = $232 << $235;
       $238 = (($237) + 245760)|0;
       $239 = $238 >>> 16;
       $240 = $239 & 2;
       $241 = $236 | $240;
       $242 = (14 - ($241))|0;
       $243 = $237 << $240;
       $244 = $243 >>> 15;
       $245 = (($242) + ($244))|0;
       $246 = $245 << 1;
       $247 = (($245) + 7)|0;
       $248 = $222 >>> $247;
       $249 = $248 & 1;
       $250 = $249 | $246;
       $$0336$i = $250;
      }
     }
     $251 = (164080 + ($$0336$i<<2)|0);
     $252 = HEAP32[$251>>2]|0;
     $253 = ($252|0)==(0|0);
     L74: do {
      if ($253) {
       $$2333$i = 0;$$3$i200 = 0;$$3328$i = $225;
       label = 57;
      } else {
       $254 = ($$0336$i|0)==(31);
       $255 = $$0336$i >>> 1;
       $256 = (25 - ($255))|0;
       $257 = $254 ? 0 : $256;
       $258 = $222 << $257;
       $$0320$i = 0;$$0325$i = $225;$$0331$i = $252;$$0337$i = $258;$$0340$i = 0;
       while(1) {
        $259 = ((($$0331$i)) + 4|0);
        $260 = HEAP32[$259>>2]|0;
        $261 = $260 & -8;
        $262 = (($261) - ($222))|0;
        $263 = ($262>>>0)<($$0325$i>>>0);
        if ($263) {
         $264 = ($262|0)==(0);
         if ($264) {
          $$411$i = $$0331$i;$$432910$i = 0;$$43359$i = $$0331$i;
          label = 61;
          break L74;
         } else {
          $$1321$i = $$0331$i;$$1326$i = $262;
         }
        } else {
         $$1321$i = $$0320$i;$$1326$i = $$0325$i;
        }
        $265 = ((($$0331$i)) + 20|0);
        $266 = HEAP32[$265>>2]|0;
        $267 = $$0337$i >>> 31;
        $268 = (((($$0331$i)) + 16|0) + ($267<<2)|0);
        $269 = HEAP32[$268>>2]|0;
        $270 = ($266|0)==(0|0);
        $271 = ($266|0)==($269|0);
        $or$cond2$i199 = $270 | $271;
        $$1341$i = $or$cond2$i199 ? $$0340$i : $266;
        $272 = ($269|0)==(0|0);
        $not$5$i = $272 ^ 1;
        $273 = $not$5$i&1;
        $$0337$$i = $$0337$i << $273;
        if ($272) {
         $$2333$i = $$1341$i;$$3$i200 = $$1321$i;$$3328$i = $$1326$i;
         label = 57;
         break;
        } else {
         $$0320$i = $$1321$i;$$0325$i = $$1326$i;$$0331$i = $269;$$0337$i = $$0337$$i;$$0340$i = $$1341$i;
        }
       }
      }
     } while(0);
     if ((label|0) == 57) {
      $274 = ($$2333$i|0)==(0|0);
      $275 = ($$3$i200|0)==(0|0);
      $or$cond$i201 = $274 & $275;
      if ($or$cond$i201) {
       $276 = 2 << $$0336$i;
       $277 = (0 - ($276))|0;
       $278 = $276 | $277;
       $279 = $223 & $278;
       $280 = ($279|0)==(0);
       if ($280) {
        $$0192 = $222;
        break;
       }
       $281 = (0 - ($279))|0;
       $282 = $279 & $281;
       $283 = (($282) + -1)|0;
       $284 = $283 >>> 12;
       $285 = $284 & 16;
       $286 = $283 >>> $285;
       $287 = $286 >>> 5;
       $288 = $287 & 8;
       $289 = $288 | $285;
       $290 = $286 >>> $288;
       $291 = $290 >>> 2;
       $292 = $291 & 4;
       $293 = $289 | $292;
       $294 = $290 >>> $292;
       $295 = $294 >>> 1;
       $296 = $295 & 2;
       $297 = $293 | $296;
       $298 = $294 >>> $296;
       $299 = $298 >>> 1;
       $300 = $299 & 1;
       $301 = $297 | $300;
       $302 = $298 >>> $300;
       $303 = (($301) + ($302))|0;
       $304 = (164080 + ($303<<2)|0);
       $305 = HEAP32[$304>>2]|0;
       $$4$ph$i = 0;$$4335$ph$i = $305;
      } else {
       $$4$ph$i = $$3$i200;$$4335$ph$i = $$2333$i;
      }
      $306 = ($$4335$ph$i|0)==(0|0);
      if ($306) {
       $$4$lcssa$i = $$4$ph$i;$$4329$lcssa$i = $$3328$i;
      } else {
       $$411$i = $$4$ph$i;$$432910$i = $$3328$i;$$43359$i = $$4335$ph$i;
       label = 61;
      }
     }
     if ((label|0) == 61) {
      while(1) {
       label = 0;
       $307 = ((($$43359$i)) + 4|0);
       $308 = HEAP32[$307>>2]|0;
       $309 = $308 & -8;
       $310 = (($309) - ($222))|0;
       $311 = ($310>>>0)<($$432910$i>>>0);
       $$$4329$i = $311 ? $310 : $$432910$i;
       $$4335$$4$i = $311 ? $$43359$i : $$411$i;
       $312 = ((($$43359$i)) + 16|0);
       $313 = HEAP32[$312>>2]|0;
       $not$1$i203 = ($313|0)==(0|0);
       $$sink2$i204 = $not$1$i203&1;
       $314 = (((($$43359$i)) + 16|0) + ($$sink2$i204<<2)|0);
       $315 = HEAP32[$314>>2]|0;
       $316 = ($315|0)==(0|0);
       if ($316) {
        $$4$lcssa$i = $$4335$$4$i;$$4329$lcssa$i = $$$4329$i;
        break;
       } else {
        $$411$i = $$4335$$4$i;$$432910$i = $$$4329$i;$$43359$i = $315;
        label = 61;
       }
      }
     }
     $317 = ($$4$lcssa$i|0)==(0|0);
     if ($317) {
      $$0192 = $222;
     } else {
      $318 = HEAP32[(163784)>>2]|0;
      $319 = (($318) - ($222))|0;
      $320 = ($$4329$lcssa$i>>>0)<($319>>>0);
      if ($320) {
       $321 = (($$4$lcssa$i) + ($222)|0);
       $322 = ($$4$lcssa$i>>>0)<($321>>>0);
       if (!($322)) {
        $$0 = 0;
        STACKTOP = sp;return ($$0|0);
       }
       $323 = ((($$4$lcssa$i)) + 24|0);
       $324 = HEAP32[$323>>2]|0;
       $325 = ((($$4$lcssa$i)) + 12|0);
       $326 = HEAP32[$325>>2]|0;
       $327 = ($326|0)==($$4$lcssa$i|0);
       do {
        if ($327) {
         $332 = ((($$4$lcssa$i)) + 20|0);
         $333 = HEAP32[$332>>2]|0;
         $334 = ($333|0)==(0|0);
         if ($334) {
          $335 = ((($$4$lcssa$i)) + 16|0);
          $336 = HEAP32[$335>>2]|0;
          $337 = ($336|0)==(0|0);
          if ($337) {
           $$3349$i = 0;
           break;
          } else {
           $$1347$i = $336;$$1351$i = $335;
          }
         } else {
          $$1347$i = $333;$$1351$i = $332;
         }
         while(1) {
          $338 = ((($$1347$i)) + 20|0);
          $339 = HEAP32[$338>>2]|0;
          $340 = ($339|0)==(0|0);
          if (!($340)) {
           $$1347$i = $339;$$1351$i = $338;
           continue;
          }
          $341 = ((($$1347$i)) + 16|0);
          $342 = HEAP32[$341>>2]|0;
          $343 = ($342|0)==(0|0);
          if ($343) {
           break;
          } else {
           $$1347$i = $342;$$1351$i = $341;
          }
         }
         HEAP32[$$1351$i>>2] = 0;
         $$3349$i = $$1347$i;
        } else {
         $328 = ((($$4$lcssa$i)) + 8|0);
         $329 = HEAP32[$328>>2]|0;
         $330 = ((($329)) + 12|0);
         HEAP32[$330>>2] = $326;
         $331 = ((($326)) + 8|0);
         HEAP32[$331>>2] = $329;
         $$3349$i = $326;
        }
       } while(0);
       $344 = ($324|0)==(0|0);
       do {
        if ($344) {
         $426 = $223;
        } else {
         $345 = ((($$4$lcssa$i)) + 28|0);
         $346 = HEAP32[$345>>2]|0;
         $347 = (164080 + ($346<<2)|0);
         $348 = HEAP32[$347>>2]|0;
         $349 = ($$4$lcssa$i|0)==($348|0);
         if ($349) {
          HEAP32[$347>>2] = $$3349$i;
          $cond$i208 = ($$3349$i|0)==(0|0);
          if ($cond$i208) {
           $350 = 1 << $346;
           $351 = $350 ^ -1;
           $352 = $223 & $351;
           HEAP32[(163780)>>2] = $352;
           $426 = $352;
           break;
          }
         } else {
          $353 = ((($324)) + 16|0);
          $354 = HEAP32[$353>>2]|0;
          $not$$i209 = ($354|0)!=($$4$lcssa$i|0);
          $$sink3$i = $not$$i209&1;
          $355 = (((($324)) + 16|0) + ($$sink3$i<<2)|0);
          HEAP32[$355>>2] = $$3349$i;
          $356 = ($$3349$i|0)==(0|0);
          if ($356) {
           $426 = $223;
           break;
          }
         }
         $357 = ((($$3349$i)) + 24|0);
         HEAP32[$357>>2] = $324;
         $358 = ((($$4$lcssa$i)) + 16|0);
         $359 = HEAP32[$358>>2]|0;
         $360 = ($359|0)==(0|0);
         if (!($360)) {
          $361 = ((($$3349$i)) + 16|0);
          HEAP32[$361>>2] = $359;
          $362 = ((($359)) + 24|0);
          HEAP32[$362>>2] = $$3349$i;
         }
         $363 = ((($$4$lcssa$i)) + 20|0);
         $364 = HEAP32[$363>>2]|0;
         $365 = ($364|0)==(0|0);
         if ($365) {
          $426 = $223;
         } else {
          $366 = ((($$3349$i)) + 20|0);
          HEAP32[$366>>2] = $364;
          $367 = ((($364)) + 24|0);
          HEAP32[$367>>2] = $$3349$i;
          $426 = $223;
         }
        }
       } while(0);
       $368 = ($$4329$lcssa$i>>>0)<(16);
       do {
        if ($368) {
         $369 = (($$4329$lcssa$i) + ($222))|0;
         $370 = $369 | 3;
         $371 = ((($$4$lcssa$i)) + 4|0);
         HEAP32[$371>>2] = $370;
         $372 = (($$4$lcssa$i) + ($369)|0);
         $373 = ((($372)) + 4|0);
         $374 = HEAP32[$373>>2]|0;
         $375 = $374 | 1;
         HEAP32[$373>>2] = $375;
        } else {
         $376 = $222 | 3;
         $377 = ((($$4$lcssa$i)) + 4|0);
         HEAP32[$377>>2] = $376;
         $378 = $$4329$lcssa$i | 1;
         $379 = ((($321)) + 4|0);
         HEAP32[$379>>2] = $378;
         $380 = (($321) + ($$4329$lcssa$i)|0);
         HEAP32[$380>>2] = $$4329$lcssa$i;
         $381 = $$4329$lcssa$i >>> 3;
         $382 = ($$4329$lcssa$i>>>0)<(256);
         if ($382) {
          $383 = $381 << 1;
          $384 = (163816 + ($383<<2)|0);
          $385 = HEAP32[40944]|0;
          $386 = 1 << $381;
          $387 = $385 & $386;
          $388 = ($387|0)==(0);
          if ($388) {
           $389 = $385 | $386;
           HEAP32[40944] = $389;
           $$pre$i210 = ((($384)) + 8|0);
           $$0345$i = $384;$$pre$phi$i211Z2D = $$pre$i210;
          } else {
           $390 = ((($384)) + 8|0);
           $391 = HEAP32[$390>>2]|0;
           $$0345$i = $391;$$pre$phi$i211Z2D = $390;
          }
          HEAP32[$$pre$phi$i211Z2D>>2] = $321;
          $392 = ((($$0345$i)) + 12|0);
          HEAP32[$392>>2] = $321;
          $393 = ((($321)) + 8|0);
          HEAP32[$393>>2] = $$0345$i;
          $394 = ((($321)) + 12|0);
          HEAP32[$394>>2] = $384;
          break;
         }
         $395 = $$4329$lcssa$i >>> 8;
         $396 = ($395|0)==(0);
         if ($396) {
          $$0339$i = 0;
         } else {
          $397 = ($$4329$lcssa$i>>>0)>(16777215);
          if ($397) {
           $$0339$i = 31;
          } else {
           $398 = (($395) + 1048320)|0;
           $399 = $398 >>> 16;
           $400 = $399 & 8;
           $401 = $395 << $400;
           $402 = (($401) + 520192)|0;
           $403 = $402 >>> 16;
           $404 = $403 & 4;
           $405 = $404 | $400;
           $406 = $401 << $404;
           $407 = (($406) + 245760)|0;
           $408 = $407 >>> 16;
           $409 = $408 & 2;
           $410 = $405 | $409;
           $411 = (14 - ($410))|0;
           $412 = $406 << $409;
           $413 = $412 >>> 15;
           $414 = (($411) + ($413))|0;
           $415 = $414 << 1;
           $416 = (($414) + 7)|0;
           $417 = $$4329$lcssa$i >>> $416;
           $418 = $417 & 1;
           $419 = $418 | $415;
           $$0339$i = $419;
          }
         }
         $420 = (164080 + ($$0339$i<<2)|0);
         $421 = ((($321)) + 28|0);
         HEAP32[$421>>2] = $$0339$i;
         $422 = ((($321)) + 16|0);
         $423 = ((($422)) + 4|0);
         HEAP32[$423>>2] = 0;
         HEAP32[$422>>2] = 0;
         $424 = 1 << $$0339$i;
         $425 = $426 & $424;
         $427 = ($425|0)==(0);
         if ($427) {
          $428 = $426 | $424;
          HEAP32[(163780)>>2] = $428;
          HEAP32[$420>>2] = $321;
          $429 = ((($321)) + 24|0);
          HEAP32[$429>>2] = $420;
          $430 = ((($321)) + 12|0);
          HEAP32[$430>>2] = $321;
          $431 = ((($321)) + 8|0);
          HEAP32[$431>>2] = $321;
          break;
         }
         $432 = HEAP32[$420>>2]|0;
         $433 = ($$0339$i|0)==(31);
         $434 = $$0339$i >>> 1;
         $435 = (25 - ($434))|0;
         $436 = $433 ? 0 : $435;
         $437 = $$4329$lcssa$i << $436;
         $$0322$i = $437;$$0323$i = $432;
         while(1) {
          $438 = ((($$0323$i)) + 4|0);
          $439 = HEAP32[$438>>2]|0;
          $440 = $439 & -8;
          $441 = ($440|0)==($$4329$lcssa$i|0);
          if ($441) {
           label = 97;
           break;
          }
          $442 = $$0322$i >>> 31;
          $443 = (((($$0323$i)) + 16|0) + ($442<<2)|0);
          $444 = $$0322$i << 1;
          $445 = HEAP32[$443>>2]|0;
          $446 = ($445|0)==(0|0);
          if ($446) {
           label = 96;
           break;
          } else {
           $$0322$i = $444;$$0323$i = $445;
          }
         }
         if ((label|0) == 96) {
          HEAP32[$443>>2] = $321;
          $447 = ((($321)) + 24|0);
          HEAP32[$447>>2] = $$0323$i;
          $448 = ((($321)) + 12|0);
          HEAP32[$448>>2] = $321;
          $449 = ((($321)) + 8|0);
          HEAP32[$449>>2] = $321;
          break;
         }
         else if ((label|0) == 97) {
          $450 = ((($$0323$i)) + 8|0);
          $451 = HEAP32[$450>>2]|0;
          $452 = ((($451)) + 12|0);
          HEAP32[$452>>2] = $321;
          HEAP32[$450>>2] = $321;
          $453 = ((($321)) + 8|0);
          HEAP32[$453>>2] = $451;
          $454 = ((($321)) + 12|0);
          HEAP32[$454>>2] = $$0323$i;
          $455 = ((($321)) + 24|0);
          HEAP32[$455>>2] = 0;
          break;
         }
        }
       } while(0);
       $456 = ((($$4$lcssa$i)) + 8|0);
       $$0 = $456;
       STACKTOP = sp;return ($$0|0);
      } else {
       $$0192 = $222;
      }
     }
    }
   }
  }
 } while(0);
 $457 = HEAP32[(163784)>>2]|0;
 $458 = ($457>>>0)<($$0192>>>0);
 if (!($458)) {
  $459 = (($457) - ($$0192))|0;
  $460 = HEAP32[(163796)>>2]|0;
  $461 = ($459>>>0)>(15);
  if ($461) {
   $462 = (($460) + ($$0192)|0);
   HEAP32[(163796)>>2] = $462;
   HEAP32[(163784)>>2] = $459;
   $463 = $459 | 1;
   $464 = ((($462)) + 4|0);
   HEAP32[$464>>2] = $463;
   $465 = (($462) + ($459)|0);
   HEAP32[$465>>2] = $459;
   $466 = $$0192 | 3;
   $467 = ((($460)) + 4|0);
   HEAP32[$467>>2] = $466;
  } else {
   HEAP32[(163784)>>2] = 0;
   HEAP32[(163796)>>2] = 0;
   $468 = $457 | 3;
   $469 = ((($460)) + 4|0);
   HEAP32[$469>>2] = $468;
   $470 = (($460) + ($457)|0);
   $471 = ((($470)) + 4|0);
   $472 = HEAP32[$471>>2]|0;
   $473 = $472 | 1;
   HEAP32[$471>>2] = $473;
  }
  $474 = ((($460)) + 8|0);
  $$0 = $474;
  STACKTOP = sp;return ($$0|0);
 }
 $475 = HEAP32[(163788)>>2]|0;
 $476 = ($475>>>0)>($$0192>>>0);
 if ($476) {
  $477 = (($475) - ($$0192))|0;
  HEAP32[(163788)>>2] = $477;
  $478 = HEAP32[(163800)>>2]|0;
  $479 = (($478) + ($$0192)|0);
  HEAP32[(163800)>>2] = $479;
  $480 = $477 | 1;
  $481 = ((($479)) + 4|0);
  HEAP32[$481>>2] = $480;
  $482 = $$0192 | 3;
  $483 = ((($478)) + 4|0);
  HEAP32[$483>>2] = $482;
  $484 = ((($478)) + 8|0);
  $$0 = $484;
  STACKTOP = sp;return ($$0|0);
 }
 $485 = HEAP32[41062]|0;
 $486 = ($485|0)==(0);
 if ($486) {
  HEAP32[(164256)>>2] = 4096;
  HEAP32[(164252)>>2] = 4096;
  HEAP32[(164260)>>2] = -1;
  HEAP32[(164264)>>2] = -1;
  HEAP32[(164268)>>2] = 0;
  HEAP32[(164220)>>2] = 0;
  $487 = $1;
  $488 = $487 & -16;
  $489 = $488 ^ 1431655768;
  HEAP32[$1>>2] = $489;
  HEAP32[41062] = $489;
  $493 = 4096;
 } else {
  $$pre$i195 = HEAP32[(164256)>>2]|0;
  $493 = $$pre$i195;
 }
 $490 = (($$0192) + 48)|0;
 $491 = (($$0192) + 47)|0;
 $492 = (($493) + ($491))|0;
 $494 = (0 - ($493))|0;
 $495 = $492 & $494;
 $496 = ($495>>>0)>($$0192>>>0);
 if (!($496)) {
  $$0 = 0;
  STACKTOP = sp;return ($$0|0);
 }
 $497 = HEAP32[(164216)>>2]|0;
 $498 = ($497|0)==(0);
 if (!($498)) {
  $499 = HEAP32[(164208)>>2]|0;
  $500 = (($499) + ($495))|0;
  $501 = ($500>>>0)<=($499>>>0);
  $502 = ($500>>>0)>($497>>>0);
  $or$cond1$i = $501 | $502;
  if ($or$cond1$i) {
   $$0 = 0;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $503 = HEAP32[(164220)>>2]|0;
 $504 = $503 & 4;
 $505 = ($504|0)==(0);
 L167: do {
  if ($505) {
   $506 = HEAP32[(163800)>>2]|0;
   $507 = ($506|0)==(0|0);
   L169: do {
    if ($507) {
     label = 118;
    } else {
     $$0$i20$i = (164224);
     while(1) {
      $508 = HEAP32[$$0$i20$i>>2]|0;
      $509 = ($508>>>0)>($506>>>0);
      if (!($509)) {
       $510 = ((($$0$i20$i)) + 4|0);
       $511 = HEAP32[$510>>2]|0;
       $512 = (($508) + ($511)|0);
       $513 = ($512>>>0)>($506>>>0);
       if ($513) {
        break;
       }
      }
      $514 = ((($$0$i20$i)) + 8|0);
      $515 = HEAP32[$514>>2]|0;
      $516 = ($515|0)==(0|0);
      if ($516) {
       label = 118;
       break L169;
      } else {
       $$0$i20$i = $515;
      }
     }
     $539 = (($492) - ($475))|0;
     $540 = $539 & $494;
     $541 = ($540>>>0)<(2147483647);
     if ($541) {
      $542 = (_sbrk(($540|0))|0);
      $543 = HEAP32[$$0$i20$i>>2]|0;
      $544 = HEAP32[$510>>2]|0;
      $545 = (($543) + ($544)|0);
      $546 = ($542|0)==($545|0);
      if ($546) {
       $547 = ($542|0)==((-1)|0);
       if ($547) {
        $$2234243136$i = $540;
       } else {
        $$723947$i = $540;$$748$i = $542;
        label = 135;
        break L167;
       }
      } else {
       $$2247$ph$i = $542;$$2253$ph$i = $540;
       label = 126;
      }
     } else {
      $$2234243136$i = 0;
     }
    }
   } while(0);
   do {
    if ((label|0) == 118) {
     $517 = (_sbrk(0)|0);
     $518 = ($517|0)==((-1)|0);
     if ($518) {
      $$2234243136$i = 0;
     } else {
      $519 = $517;
      $520 = HEAP32[(164252)>>2]|0;
      $521 = (($520) + -1)|0;
      $522 = $521 & $519;
      $523 = ($522|0)==(0);
      $524 = (($521) + ($519))|0;
      $525 = (0 - ($520))|0;
      $526 = $524 & $525;
      $527 = (($526) - ($519))|0;
      $528 = $523 ? 0 : $527;
      $$$i = (($528) + ($495))|0;
      $529 = HEAP32[(164208)>>2]|0;
      $530 = (($$$i) + ($529))|0;
      $531 = ($$$i>>>0)>($$0192>>>0);
      $532 = ($$$i>>>0)<(2147483647);
      $or$cond$i = $531 & $532;
      if ($or$cond$i) {
       $533 = HEAP32[(164216)>>2]|0;
       $534 = ($533|0)==(0);
       if (!($534)) {
        $535 = ($530>>>0)<=($529>>>0);
        $536 = ($530>>>0)>($533>>>0);
        $or$cond2$i = $535 | $536;
        if ($or$cond2$i) {
         $$2234243136$i = 0;
         break;
        }
       }
       $537 = (_sbrk(($$$i|0))|0);
       $538 = ($537|0)==($517|0);
       if ($538) {
        $$723947$i = $$$i;$$748$i = $517;
        label = 135;
        break L167;
       } else {
        $$2247$ph$i = $537;$$2253$ph$i = $$$i;
        label = 126;
       }
      } else {
       $$2234243136$i = 0;
      }
     }
    }
   } while(0);
   do {
    if ((label|0) == 126) {
     $548 = (0 - ($$2253$ph$i))|0;
     $549 = ($$2247$ph$i|0)!=((-1)|0);
     $550 = ($$2253$ph$i>>>0)<(2147483647);
     $or$cond7$i = $550 & $549;
     $551 = ($490>>>0)>($$2253$ph$i>>>0);
     $or$cond10$i = $551 & $or$cond7$i;
     if (!($or$cond10$i)) {
      $561 = ($$2247$ph$i|0)==((-1)|0);
      if ($561) {
       $$2234243136$i = 0;
       break;
      } else {
       $$723947$i = $$2253$ph$i;$$748$i = $$2247$ph$i;
       label = 135;
       break L167;
      }
     }
     $552 = HEAP32[(164256)>>2]|0;
     $553 = (($491) - ($$2253$ph$i))|0;
     $554 = (($553) + ($552))|0;
     $555 = (0 - ($552))|0;
     $556 = $554 & $555;
     $557 = ($556>>>0)<(2147483647);
     if (!($557)) {
      $$723947$i = $$2253$ph$i;$$748$i = $$2247$ph$i;
      label = 135;
      break L167;
     }
     $558 = (_sbrk(($556|0))|0);
     $559 = ($558|0)==((-1)|0);
     if ($559) {
      (_sbrk(($548|0))|0);
      $$2234243136$i = 0;
      break;
     } else {
      $560 = (($556) + ($$2253$ph$i))|0;
      $$723947$i = $560;$$748$i = $$2247$ph$i;
      label = 135;
      break L167;
     }
    }
   } while(0);
   $562 = HEAP32[(164220)>>2]|0;
   $563 = $562 | 4;
   HEAP32[(164220)>>2] = $563;
   $$4236$i = $$2234243136$i;
   label = 133;
  } else {
   $$4236$i = 0;
   label = 133;
  }
 } while(0);
 if ((label|0) == 133) {
  $564 = ($495>>>0)<(2147483647);
  if ($564) {
   $565 = (_sbrk(($495|0))|0);
   $566 = (_sbrk(0)|0);
   $567 = ($565|0)!=((-1)|0);
   $568 = ($566|0)!=((-1)|0);
   $or$cond5$i = $567 & $568;
   $569 = ($565>>>0)<($566>>>0);
   $or$cond11$i = $569 & $or$cond5$i;
   $570 = $566;
   $571 = $565;
   $572 = (($570) - ($571))|0;
   $573 = (($$0192) + 40)|0;
   $574 = ($572>>>0)>($573>>>0);
   $$$4236$i = $574 ? $572 : $$4236$i;
   $or$cond11$not$i = $or$cond11$i ^ 1;
   $575 = ($565|0)==((-1)|0);
   $not$$i197 = $574 ^ 1;
   $576 = $575 | $not$$i197;
   $or$cond49$i = $576 | $or$cond11$not$i;
   if (!($or$cond49$i)) {
    $$723947$i = $$$4236$i;$$748$i = $565;
    label = 135;
   }
  }
 }
 if ((label|0) == 135) {
  $577 = HEAP32[(164208)>>2]|0;
  $578 = (($577) + ($$723947$i))|0;
  HEAP32[(164208)>>2] = $578;
  $579 = HEAP32[(164212)>>2]|0;
  $580 = ($578>>>0)>($579>>>0);
  if ($580) {
   HEAP32[(164212)>>2] = $578;
  }
  $581 = HEAP32[(163800)>>2]|0;
  $582 = ($581|0)==(0|0);
  do {
   if ($582) {
    $583 = HEAP32[(163792)>>2]|0;
    $584 = ($583|0)==(0|0);
    $585 = ($$748$i>>>0)<($583>>>0);
    $or$cond12$i = $584 | $585;
    if ($or$cond12$i) {
     HEAP32[(163792)>>2] = $$748$i;
    }
    HEAP32[(164224)>>2] = $$748$i;
    HEAP32[(164228)>>2] = $$723947$i;
    HEAP32[(164236)>>2] = 0;
    $586 = HEAP32[41062]|0;
    HEAP32[(163812)>>2] = $586;
    HEAP32[(163808)>>2] = -1;
    $$01$i$i = 0;
    while(1) {
     $587 = $$01$i$i << 1;
     $588 = (163816 + ($587<<2)|0);
     $589 = ((($588)) + 12|0);
     HEAP32[$589>>2] = $588;
     $590 = ((($588)) + 8|0);
     HEAP32[$590>>2] = $588;
     $591 = (($$01$i$i) + 1)|0;
     $exitcond$i$i = ($591|0)==(32);
     if ($exitcond$i$i) {
      break;
     } else {
      $$01$i$i = $591;
     }
    }
    $592 = (($$723947$i) + -40)|0;
    $593 = ((($$748$i)) + 8|0);
    $594 = $593;
    $595 = $594 & 7;
    $596 = ($595|0)==(0);
    $597 = (0 - ($594))|0;
    $598 = $597 & 7;
    $599 = $596 ? 0 : $598;
    $600 = (($$748$i) + ($599)|0);
    $601 = (($592) - ($599))|0;
    HEAP32[(163800)>>2] = $600;
    HEAP32[(163788)>>2] = $601;
    $602 = $601 | 1;
    $603 = ((($600)) + 4|0);
    HEAP32[$603>>2] = $602;
    $604 = (($600) + ($601)|0);
    $605 = ((($604)) + 4|0);
    HEAP32[$605>>2] = 40;
    $606 = HEAP32[(164264)>>2]|0;
    HEAP32[(163804)>>2] = $606;
   } else {
    $$024370$i = (164224);
    while(1) {
     $607 = HEAP32[$$024370$i>>2]|0;
     $608 = ((($$024370$i)) + 4|0);
     $609 = HEAP32[$608>>2]|0;
     $610 = (($607) + ($609)|0);
     $611 = ($$748$i|0)==($610|0);
     if ($611) {
      label = 145;
      break;
     }
     $612 = ((($$024370$i)) + 8|0);
     $613 = HEAP32[$612>>2]|0;
     $614 = ($613|0)==(0|0);
     if ($614) {
      break;
     } else {
      $$024370$i = $613;
     }
    }
    if ((label|0) == 145) {
     $615 = ((($$024370$i)) + 12|0);
     $616 = HEAP32[$615>>2]|0;
     $617 = $616 & 8;
     $618 = ($617|0)==(0);
     if ($618) {
      $619 = ($581>>>0)>=($607>>>0);
      $620 = ($581>>>0)<($$748$i>>>0);
      $or$cond50$i = $620 & $619;
      if ($or$cond50$i) {
       $621 = (($609) + ($$723947$i))|0;
       HEAP32[$608>>2] = $621;
       $622 = HEAP32[(163788)>>2]|0;
       $623 = ((($581)) + 8|0);
       $624 = $623;
       $625 = $624 & 7;
       $626 = ($625|0)==(0);
       $627 = (0 - ($624))|0;
       $628 = $627 & 7;
       $629 = $626 ? 0 : $628;
       $630 = (($581) + ($629)|0);
       $631 = (($$723947$i) - ($629))|0;
       $632 = (($622) + ($631))|0;
       HEAP32[(163800)>>2] = $630;
       HEAP32[(163788)>>2] = $632;
       $633 = $632 | 1;
       $634 = ((($630)) + 4|0);
       HEAP32[$634>>2] = $633;
       $635 = (($630) + ($632)|0);
       $636 = ((($635)) + 4|0);
       HEAP32[$636>>2] = 40;
       $637 = HEAP32[(164264)>>2]|0;
       HEAP32[(163804)>>2] = $637;
       break;
      }
     }
    }
    $638 = HEAP32[(163792)>>2]|0;
    $639 = ($$748$i>>>0)<($638>>>0);
    if ($639) {
     HEAP32[(163792)>>2] = $$748$i;
    }
    $640 = (($$748$i) + ($$723947$i)|0);
    $$124469$i = (164224);
    while(1) {
     $641 = HEAP32[$$124469$i>>2]|0;
     $642 = ($641|0)==($640|0);
     if ($642) {
      label = 153;
      break;
     }
     $643 = ((($$124469$i)) + 8|0);
     $644 = HEAP32[$643>>2]|0;
     $645 = ($644|0)==(0|0);
     if ($645) {
      break;
     } else {
      $$124469$i = $644;
     }
    }
    if ((label|0) == 153) {
     $646 = ((($$124469$i)) + 12|0);
     $647 = HEAP32[$646>>2]|0;
     $648 = $647 & 8;
     $649 = ($648|0)==(0);
     if ($649) {
      HEAP32[$$124469$i>>2] = $$748$i;
      $650 = ((($$124469$i)) + 4|0);
      $651 = HEAP32[$650>>2]|0;
      $652 = (($651) + ($$723947$i))|0;
      HEAP32[$650>>2] = $652;
      $653 = ((($$748$i)) + 8|0);
      $654 = $653;
      $655 = $654 & 7;
      $656 = ($655|0)==(0);
      $657 = (0 - ($654))|0;
      $658 = $657 & 7;
      $659 = $656 ? 0 : $658;
      $660 = (($$748$i) + ($659)|0);
      $661 = ((($640)) + 8|0);
      $662 = $661;
      $663 = $662 & 7;
      $664 = ($663|0)==(0);
      $665 = (0 - ($662))|0;
      $666 = $665 & 7;
      $667 = $664 ? 0 : $666;
      $668 = (($640) + ($667)|0);
      $669 = $668;
      $670 = $660;
      $671 = (($669) - ($670))|0;
      $672 = (($660) + ($$0192)|0);
      $673 = (($671) - ($$0192))|0;
      $674 = $$0192 | 3;
      $675 = ((($660)) + 4|0);
      HEAP32[$675>>2] = $674;
      $676 = ($668|0)==($581|0);
      do {
       if ($676) {
        $677 = HEAP32[(163788)>>2]|0;
        $678 = (($677) + ($673))|0;
        HEAP32[(163788)>>2] = $678;
        HEAP32[(163800)>>2] = $672;
        $679 = $678 | 1;
        $680 = ((($672)) + 4|0);
        HEAP32[$680>>2] = $679;
       } else {
        $681 = HEAP32[(163796)>>2]|0;
        $682 = ($668|0)==($681|0);
        if ($682) {
         $683 = HEAP32[(163784)>>2]|0;
         $684 = (($683) + ($673))|0;
         HEAP32[(163784)>>2] = $684;
         HEAP32[(163796)>>2] = $672;
         $685 = $684 | 1;
         $686 = ((($672)) + 4|0);
         HEAP32[$686>>2] = $685;
         $687 = (($672) + ($684)|0);
         HEAP32[$687>>2] = $684;
         break;
        }
        $688 = ((($668)) + 4|0);
        $689 = HEAP32[$688>>2]|0;
        $690 = $689 & 3;
        $691 = ($690|0)==(1);
        if ($691) {
         $692 = $689 & -8;
         $693 = $689 >>> 3;
         $694 = ($689>>>0)<(256);
         L237: do {
          if ($694) {
           $695 = ((($668)) + 8|0);
           $696 = HEAP32[$695>>2]|0;
           $697 = ((($668)) + 12|0);
           $698 = HEAP32[$697>>2]|0;
           $699 = ($698|0)==($696|0);
           if ($699) {
            $700 = 1 << $693;
            $701 = $700 ^ -1;
            $702 = HEAP32[40944]|0;
            $703 = $702 & $701;
            HEAP32[40944] = $703;
            break;
           } else {
            $704 = ((($696)) + 12|0);
            HEAP32[$704>>2] = $698;
            $705 = ((($698)) + 8|0);
            HEAP32[$705>>2] = $696;
            break;
           }
          } else {
           $706 = ((($668)) + 24|0);
           $707 = HEAP32[$706>>2]|0;
           $708 = ((($668)) + 12|0);
           $709 = HEAP32[$708>>2]|0;
           $710 = ($709|0)==($668|0);
           do {
            if ($710) {
             $715 = ((($668)) + 16|0);
             $716 = ((($715)) + 4|0);
             $717 = HEAP32[$716>>2]|0;
             $718 = ($717|0)==(0|0);
             if ($718) {
              $719 = HEAP32[$715>>2]|0;
              $720 = ($719|0)==(0|0);
              if ($720) {
               $$3$i$i = 0;
               break;
              } else {
               $$1264$i$i = $719;$$1266$i$i = $715;
              }
             } else {
              $$1264$i$i = $717;$$1266$i$i = $716;
             }
             while(1) {
              $721 = ((($$1264$i$i)) + 20|0);
              $722 = HEAP32[$721>>2]|0;
              $723 = ($722|0)==(0|0);
              if (!($723)) {
               $$1264$i$i = $722;$$1266$i$i = $721;
               continue;
              }
              $724 = ((($$1264$i$i)) + 16|0);
              $725 = HEAP32[$724>>2]|0;
              $726 = ($725|0)==(0|0);
              if ($726) {
               break;
              } else {
               $$1264$i$i = $725;$$1266$i$i = $724;
              }
             }
             HEAP32[$$1266$i$i>>2] = 0;
             $$3$i$i = $$1264$i$i;
            } else {
             $711 = ((($668)) + 8|0);
             $712 = HEAP32[$711>>2]|0;
             $713 = ((($712)) + 12|0);
             HEAP32[$713>>2] = $709;
             $714 = ((($709)) + 8|0);
             HEAP32[$714>>2] = $712;
             $$3$i$i = $709;
            }
           } while(0);
           $727 = ($707|0)==(0|0);
           if ($727) {
            break;
           }
           $728 = ((($668)) + 28|0);
           $729 = HEAP32[$728>>2]|0;
           $730 = (164080 + ($729<<2)|0);
           $731 = HEAP32[$730>>2]|0;
           $732 = ($668|0)==($731|0);
           do {
            if ($732) {
             HEAP32[$730>>2] = $$3$i$i;
             $cond$i$i = ($$3$i$i|0)==(0|0);
             if (!($cond$i$i)) {
              break;
             }
             $733 = 1 << $729;
             $734 = $733 ^ -1;
             $735 = HEAP32[(163780)>>2]|0;
             $736 = $735 & $734;
             HEAP32[(163780)>>2] = $736;
             break L237;
            } else {
             $737 = ((($707)) + 16|0);
             $738 = HEAP32[$737>>2]|0;
             $not$$i$i = ($738|0)!=($668|0);
             $$sink1$i$i = $not$$i$i&1;
             $739 = (((($707)) + 16|0) + ($$sink1$i$i<<2)|0);
             HEAP32[$739>>2] = $$3$i$i;
             $740 = ($$3$i$i|0)==(0|0);
             if ($740) {
              break L237;
             }
            }
           } while(0);
           $741 = ((($$3$i$i)) + 24|0);
           HEAP32[$741>>2] = $707;
           $742 = ((($668)) + 16|0);
           $743 = HEAP32[$742>>2]|0;
           $744 = ($743|0)==(0|0);
           if (!($744)) {
            $745 = ((($$3$i$i)) + 16|0);
            HEAP32[$745>>2] = $743;
            $746 = ((($743)) + 24|0);
            HEAP32[$746>>2] = $$3$i$i;
           }
           $747 = ((($742)) + 4|0);
           $748 = HEAP32[$747>>2]|0;
           $749 = ($748|0)==(0|0);
           if ($749) {
            break;
           }
           $750 = ((($$3$i$i)) + 20|0);
           HEAP32[$750>>2] = $748;
           $751 = ((($748)) + 24|0);
           HEAP32[$751>>2] = $$3$i$i;
          }
         } while(0);
         $752 = (($668) + ($692)|0);
         $753 = (($692) + ($673))|0;
         $$0$i$i = $752;$$0260$i$i = $753;
        } else {
         $$0$i$i = $668;$$0260$i$i = $673;
        }
        $754 = ((($$0$i$i)) + 4|0);
        $755 = HEAP32[$754>>2]|0;
        $756 = $755 & -2;
        HEAP32[$754>>2] = $756;
        $757 = $$0260$i$i | 1;
        $758 = ((($672)) + 4|0);
        HEAP32[$758>>2] = $757;
        $759 = (($672) + ($$0260$i$i)|0);
        HEAP32[$759>>2] = $$0260$i$i;
        $760 = $$0260$i$i >>> 3;
        $761 = ($$0260$i$i>>>0)<(256);
        if ($761) {
         $762 = $760 << 1;
         $763 = (163816 + ($762<<2)|0);
         $764 = HEAP32[40944]|0;
         $765 = 1 << $760;
         $766 = $764 & $765;
         $767 = ($766|0)==(0);
         if ($767) {
          $768 = $764 | $765;
          HEAP32[40944] = $768;
          $$pre$i17$i = ((($763)) + 8|0);
          $$0268$i$i = $763;$$pre$phi$i18$iZ2D = $$pre$i17$i;
         } else {
          $769 = ((($763)) + 8|0);
          $770 = HEAP32[$769>>2]|0;
          $$0268$i$i = $770;$$pre$phi$i18$iZ2D = $769;
         }
         HEAP32[$$pre$phi$i18$iZ2D>>2] = $672;
         $771 = ((($$0268$i$i)) + 12|0);
         HEAP32[$771>>2] = $672;
         $772 = ((($672)) + 8|0);
         HEAP32[$772>>2] = $$0268$i$i;
         $773 = ((($672)) + 12|0);
         HEAP32[$773>>2] = $763;
         break;
        }
        $774 = $$0260$i$i >>> 8;
        $775 = ($774|0)==(0);
        do {
         if ($775) {
          $$0269$i$i = 0;
         } else {
          $776 = ($$0260$i$i>>>0)>(16777215);
          if ($776) {
           $$0269$i$i = 31;
           break;
          }
          $777 = (($774) + 1048320)|0;
          $778 = $777 >>> 16;
          $779 = $778 & 8;
          $780 = $774 << $779;
          $781 = (($780) + 520192)|0;
          $782 = $781 >>> 16;
          $783 = $782 & 4;
          $784 = $783 | $779;
          $785 = $780 << $783;
          $786 = (($785) + 245760)|0;
          $787 = $786 >>> 16;
          $788 = $787 & 2;
          $789 = $784 | $788;
          $790 = (14 - ($789))|0;
          $791 = $785 << $788;
          $792 = $791 >>> 15;
          $793 = (($790) + ($792))|0;
          $794 = $793 << 1;
          $795 = (($793) + 7)|0;
          $796 = $$0260$i$i >>> $795;
          $797 = $796 & 1;
          $798 = $797 | $794;
          $$0269$i$i = $798;
         }
        } while(0);
        $799 = (164080 + ($$0269$i$i<<2)|0);
        $800 = ((($672)) + 28|0);
        HEAP32[$800>>2] = $$0269$i$i;
        $801 = ((($672)) + 16|0);
        $802 = ((($801)) + 4|0);
        HEAP32[$802>>2] = 0;
        HEAP32[$801>>2] = 0;
        $803 = HEAP32[(163780)>>2]|0;
        $804 = 1 << $$0269$i$i;
        $805 = $803 & $804;
        $806 = ($805|0)==(0);
        if ($806) {
         $807 = $803 | $804;
         HEAP32[(163780)>>2] = $807;
         HEAP32[$799>>2] = $672;
         $808 = ((($672)) + 24|0);
         HEAP32[$808>>2] = $799;
         $809 = ((($672)) + 12|0);
         HEAP32[$809>>2] = $672;
         $810 = ((($672)) + 8|0);
         HEAP32[$810>>2] = $672;
         break;
        }
        $811 = HEAP32[$799>>2]|0;
        $812 = ($$0269$i$i|0)==(31);
        $813 = $$0269$i$i >>> 1;
        $814 = (25 - ($813))|0;
        $815 = $812 ? 0 : $814;
        $816 = $$0260$i$i << $815;
        $$0261$i$i = $816;$$0262$i$i = $811;
        while(1) {
         $817 = ((($$0262$i$i)) + 4|0);
         $818 = HEAP32[$817>>2]|0;
         $819 = $818 & -8;
         $820 = ($819|0)==($$0260$i$i|0);
         if ($820) {
          label = 194;
          break;
         }
         $821 = $$0261$i$i >>> 31;
         $822 = (((($$0262$i$i)) + 16|0) + ($821<<2)|0);
         $823 = $$0261$i$i << 1;
         $824 = HEAP32[$822>>2]|0;
         $825 = ($824|0)==(0|0);
         if ($825) {
          label = 193;
          break;
         } else {
          $$0261$i$i = $823;$$0262$i$i = $824;
         }
        }
        if ((label|0) == 193) {
         HEAP32[$822>>2] = $672;
         $826 = ((($672)) + 24|0);
         HEAP32[$826>>2] = $$0262$i$i;
         $827 = ((($672)) + 12|0);
         HEAP32[$827>>2] = $672;
         $828 = ((($672)) + 8|0);
         HEAP32[$828>>2] = $672;
         break;
        }
        else if ((label|0) == 194) {
         $829 = ((($$0262$i$i)) + 8|0);
         $830 = HEAP32[$829>>2]|0;
         $831 = ((($830)) + 12|0);
         HEAP32[$831>>2] = $672;
         HEAP32[$829>>2] = $672;
         $832 = ((($672)) + 8|0);
         HEAP32[$832>>2] = $830;
         $833 = ((($672)) + 12|0);
         HEAP32[$833>>2] = $$0262$i$i;
         $834 = ((($672)) + 24|0);
         HEAP32[$834>>2] = 0;
         break;
        }
       }
      } while(0);
      $959 = ((($660)) + 8|0);
      $$0 = $959;
      STACKTOP = sp;return ($$0|0);
     }
    }
    $$0$i$i$i = (164224);
    while(1) {
     $835 = HEAP32[$$0$i$i$i>>2]|0;
     $836 = ($835>>>0)>($581>>>0);
     if (!($836)) {
      $837 = ((($$0$i$i$i)) + 4|0);
      $838 = HEAP32[$837>>2]|0;
      $839 = (($835) + ($838)|0);
      $840 = ($839>>>0)>($581>>>0);
      if ($840) {
       break;
      }
     }
     $841 = ((($$0$i$i$i)) + 8|0);
     $842 = HEAP32[$841>>2]|0;
     $$0$i$i$i = $842;
    }
    $843 = ((($839)) + -47|0);
    $844 = ((($843)) + 8|0);
    $845 = $844;
    $846 = $845 & 7;
    $847 = ($846|0)==(0);
    $848 = (0 - ($845))|0;
    $849 = $848 & 7;
    $850 = $847 ? 0 : $849;
    $851 = (($843) + ($850)|0);
    $852 = ((($581)) + 16|0);
    $853 = ($851>>>0)<($852>>>0);
    $854 = $853 ? $581 : $851;
    $855 = ((($854)) + 8|0);
    $856 = ((($854)) + 24|0);
    $857 = (($$723947$i) + -40)|0;
    $858 = ((($$748$i)) + 8|0);
    $859 = $858;
    $860 = $859 & 7;
    $861 = ($860|0)==(0);
    $862 = (0 - ($859))|0;
    $863 = $862 & 7;
    $864 = $861 ? 0 : $863;
    $865 = (($$748$i) + ($864)|0);
    $866 = (($857) - ($864))|0;
    HEAP32[(163800)>>2] = $865;
    HEAP32[(163788)>>2] = $866;
    $867 = $866 | 1;
    $868 = ((($865)) + 4|0);
    HEAP32[$868>>2] = $867;
    $869 = (($865) + ($866)|0);
    $870 = ((($869)) + 4|0);
    HEAP32[$870>>2] = 40;
    $871 = HEAP32[(164264)>>2]|0;
    HEAP32[(163804)>>2] = $871;
    $872 = ((($854)) + 4|0);
    HEAP32[$872>>2] = 27;
    ;HEAP32[$855>>2]=HEAP32[(164224)>>2]|0;HEAP32[$855+4>>2]=HEAP32[(164224)+4>>2]|0;HEAP32[$855+8>>2]=HEAP32[(164224)+8>>2]|0;HEAP32[$855+12>>2]=HEAP32[(164224)+12>>2]|0;
    HEAP32[(164224)>>2] = $$748$i;
    HEAP32[(164228)>>2] = $$723947$i;
    HEAP32[(164236)>>2] = 0;
    HEAP32[(164232)>>2] = $855;
    $874 = $856;
    while(1) {
     $873 = ((($874)) + 4|0);
     HEAP32[$873>>2] = 7;
     $875 = ((($874)) + 8|0);
     $876 = ($875>>>0)<($839>>>0);
     if ($876) {
      $874 = $873;
     } else {
      break;
     }
    }
    $877 = ($854|0)==($581|0);
    if (!($877)) {
     $878 = $854;
     $879 = $581;
     $880 = (($878) - ($879))|0;
     $881 = HEAP32[$872>>2]|0;
     $882 = $881 & -2;
     HEAP32[$872>>2] = $882;
     $883 = $880 | 1;
     $884 = ((($581)) + 4|0);
     HEAP32[$884>>2] = $883;
     HEAP32[$854>>2] = $880;
     $885 = $880 >>> 3;
     $886 = ($880>>>0)<(256);
     if ($886) {
      $887 = $885 << 1;
      $888 = (163816 + ($887<<2)|0);
      $889 = HEAP32[40944]|0;
      $890 = 1 << $885;
      $891 = $889 & $890;
      $892 = ($891|0)==(0);
      if ($892) {
       $893 = $889 | $890;
       HEAP32[40944] = $893;
       $$pre$i$i = ((($888)) + 8|0);
       $$0206$i$i = $888;$$pre$phi$i$iZ2D = $$pre$i$i;
      } else {
       $894 = ((($888)) + 8|0);
       $895 = HEAP32[$894>>2]|0;
       $$0206$i$i = $895;$$pre$phi$i$iZ2D = $894;
      }
      HEAP32[$$pre$phi$i$iZ2D>>2] = $581;
      $896 = ((($$0206$i$i)) + 12|0);
      HEAP32[$896>>2] = $581;
      $897 = ((($581)) + 8|0);
      HEAP32[$897>>2] = $$0206$i$i;
      $898 = ((($581)) + 12|0);
      HEAP32[$898>>2] = $888;
      break;
     }
     $899 = $880 >>> 8;
     $900 = ($899|0)==(0);
     if ($900) {
      $$0207$i$i = 0;
     } else {
      $901 = ($880>>>0)>(16777215);
      if ($901) {
       $$0207$i$i = 31;
      } else {
       $902 = (($899) + 1048320)|0;
       $903 = $902 >>> 16;
       $904 = $903 & 8;
       $905 = $899 << $904;
       $906 = (($905) + 520192)|0;
       $907 = $906 >>> 16;
       $908 = $907 & 4;
       $909 = $908 | $904;
       $910 = $905 << $908;
       $911 = (($910) + 245760)|0;
       $912 = $911 >>> 16;
       $913 = $912 & 2;
       $914 = $909 | $913;
       $915 = (14 - ($914))|0;
       $916 = $910 << $913;
       $917 = $916 >>> 15;
       $918 = (($915) + ($917))|0;
       $919 = $918 << 1;
       $920 = (($918) + 7)|0;
       $921 = $880 >>> $920;
       $922 = $921 & 1;
       $923 = $922 | $919;
       $$0207$i$i = $923;
      }
     }
     $924 = (164080 + ($$0207$i$i<<2)|0);
     $925 = ((($581)) + 28|0);
     HEAP32[$925>>2] = $$0207$i$i;
     $926 = ((($581)) + 20|0);
     HEAP32[$926>>2] = 0;
     HEAP32[$852>>2] = 0;
     $927 = HEAP32[(163780)>>2]|0;
     $928 = 1 << $$0207$i$i;
     $929 = $927 & $928;
     $930 = ($929|0)==(0);
     if ($930) {
      $931 = $927 | $928;
      HEAP32[(163780)>>2] = $931;
      HEAP32[$924>>2] = $581;
      $932 = ((($581)) + 24|0);
      HEAP32[$932>>2] = $924;
      $933 = ((($581)) + 12|0);
      HEAP32[$933>>2] = $581;
      $934 = ((($581)) + 8|0);
      HEAP32[$934>>2] = $581;
      break;
     }
     $935 = HEAP32[$924>>2]|0;
     $936 = ($$0207$i$i|0)==(31);
     $937 = $$0207$i$i >>> 1;
     $938 = (25 - ($937))|0;
     $939 = $936 ? 0 : $938;
     $940 = $880 << $939;
     $$0201$i$i = $940;$$0202$i$i = $935;
     while(1) {
      $941 = ((($$0202$i$i)) + 4|0);
      $942 = HEAP32[$941>>2]|0;
      $943 = $942 & -8;
      $944 = ($943|0)==($880|0);
      if ($944) {
       label = 216;
       break;
      }
      $945 = $$0201$i$i >>> 31;
      $946 = (((($$0202$i$i)) + 16|0) + ($945<<2)|0);
      $947 = $$0201$i$i << 1;
      $948 = HEAP32[$946>>2]|0;
      $949 = ($948|0)==(0|0);
      if ($949) {
       label = 215;
       break;
      } else {
       $$0201$i$i = $947;$$0202$i$i = $948;
      }
     }
     if ((label|0) == 215) {
      HEAP32[$946>>2] = $581;
      $950 = ((($581)) + 24|0);
      HEAP32[$950>>2] = $$0202$i$i;
      $951 = ((($581)) + 12|0);
      HEAP32[$951>>2] = $581;
      $952 = ((($581)) + 8|0);
      HEAP32[$952>>2] = $581;
      break;
     }
     else if ((label|0) == 216) {
      $953 = ((($$0202$i$i)) + 8|0);
      $954 = HEAP32[$953>>2]|0;
      $955 = ((($954)) + 12|0);
      HEAP32[$955>>2] = $581;
      HEAP32[$953>>2] = $581;
      $956 = ((($581)) + 8|0);
      HEAP32[$956>>2] = $954;
      $957 = ((($581)) + 12|0);
      HEAP32[$957>>2] = $$0202$i$i;
      $958 = ((($581)) + 24|0);
      HEAP32[$958>>2] = 0;
      break;
     }
    }
   }
  } while(0);
  $960 = HEAP32[(163788)>>2]|0;
  $961 = ($960>>>0)>($$0192>>>0);
  if ($961) {
   $962 = (($960) - ($$0192))|0;
   HEAP32[(163788)>>2] = $962;
   $963 = HEAP32[(163800)>>2]|0;
   $964 = (($963) + ($$0192)|0);
   HEAP32[(163800)>>2] = $964;
   $965 = $962 | 1;
   $966 = ((($964)) + 4|0);
   HEAP32[$966>>2] = $965;
   $967 = $$0192 | 3;
   $968 = ((($963)) + 4|0);
   HEAP32[$968>>2] = $967;
   $969 = ((($963)) + 8|0);
   $$0 = $969;
   STACKTOP = sp;return ($$0|0);
  }
 }
 $970 = (___errno_location()|0);
 HEAP32[$970>>2] = 12;
 $$0 = 0;
 STACKTOP = sp;return ($$0|0);
}
function _free($0) {
 $0 = $0|0;
 var $$0195$i = 0, $$0195$in$i = 0, $$0348 = 0, $$0349 = 0, $$0361 = 0, $$0368 = 0, $$1 = 0, $$1347 = 0, $$1352 = 0, $$1355 = 0, $$1363 = 0, $$1367 = 0, $$2 = 0, $$3 = 0, $$3365 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$sink3 = 0, $$sink5 = 0, $1 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0;
 var $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0;
 var $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0;
 var $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond374 = 0, $cond375 = 0, $not$ = 0, $not$370 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0|0);
 if ($1) {
  return;
 }
 $2 = ((($0)) + -8|0);
 $3 = HEAP32[(163792)>>2]|0;
 $4 = ((($0)) + -4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & -8;
 $7 = (($2) + ($6)|0);
 $8 = $5 & 1;
 $9 = ($8|0)==(0);
 do {
  if ($9) {
   $10 = HEAP32[$2>>2]|0;
   $11 = $5 & 3;
   $12 = ($11|0)==(0);
   if ($12) {
    return;
   }
   $13 = (0 - ($10))|0;
   $14 = (($2) + ($13)|0);
   $15 = (($10) + ($6))|0;
   $16 = ($14>>>0)<($3>>>0);
   if ($16) {
    return;
   }
   $17 = HEAP32[(163796)>>2]|0;
   $18 = ($14|0)==($17|0);
   if ($18) {
    $78 = ((($7)) + 4|0);
    $79 = HEAP32[$78>>2]|0;
    $80 = $79 & 3;
    $81 = ($80|0)==(3);
    if (!($81)) {
     $$1 = $14;$$1347 = $15;$87 = $14;
     break;
    }
    $82 = (($14) + ($15)|0);
    $83 = ((($14)) + 4|0);
    $84 = $15 | 1;
    $85 = $79 & -2;
    HEAP32[(163784)>>2] = $15;
    HEAP32[$78>>2] = $85;
    HEAP32[$83>>2] = $84;
    HEAP32[$82>>2] = $15;
    return;
   }
   $19 = $10 >>> 3;
   $20 = ($10>>>0)<(256);
   if ($20) {
    $21 = ((($14)) + 8|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($14)) + 12|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($24|0)==($22|0);
    if ($25) {
     $26 = 1 << $19;
     $27 = $26 ^ -1;
     $28 = HEAP32[40944]|0;
     $29 = $28 & $27;
     HEAP32[40944] = $29;
     $$1 = $14;$$1347 = $15;$87 = $14;
     break;
    } else {
     $30 = ((($22)) + 12|0);
     HEAP32[$30>>2] = $24;
     $31 = ((($24)) + 8|0);
     HEAP32[$31>>2] = $22;
     $$1 = $14;$$1347 = $15;$87 = $14;
     break;
    }
   }
   $32 = ((($14)) + 24|0);
   $33 = HEAP32[$32>>2]|0;
   $34 = ((($14)) + 12|0);
   $35 = HEAP32[$34>>2]|0;
   $36 = ($35|0)==($14|0);
   do {
    if ($36) {
     $41 = ((($14)) + 16|0);
     $42 = ((($41)) + 4|0);
     $43 = HEAP32[$42>>2]|0;
     $44 = ($43|0)==(0|0);
     if ($44) {
      $45 = HEAP32[$41>>2]|0;
      $46 = ($45|0)==(0|0);
      if ($46) {
       $$3 = 0;
       break;
      } else {
       $$1352 = $45;$$1355 = $41;
      }
     } else {
      $$1352 = $43;$$1355 = $42;
     }
     while(1) {
      $47 = ((($$1352)) + 20|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($48|0)==(0|0);
      if (!($49)) {
       $$1352 = $48;$$1355 = $47;
       continue;
      }
      $50 = ((($$1352)) + 16|0);
      $51 = HEAP32[$50>>2]|0;
      $52 = ($51|0)==(0|0);
      if ($52) {
       break;
      } else {
       $$1352 = $51;$$1355 = $50;
      }
     }
     HEAP32[$$1355>>2] = 0;
     $$3 = $$1352;
    } else {
     $37 = ((($14)) + 8|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = ((($38)) + 12|0);
     HEAP32[$39>>2] = $35;
     $40 = ((($35)) + 8|0);
     HEAP32[$40>>2] = $38;
     $$3 = $35;
    }
   } while(0);
   $53 = ($33|0)==(0|0);
   if ($53) {
    $$1 = $14;$$1347 = $15;$87 = $14;
   } else {
    $54 = ((($14)) + 28|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = (164080 + ($55<<2)|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = ($14|0)==($57|0);
    if ($58) {
     HEAP32[$56>>2] = $$3;
     $cond374 = ($$3|0)==(0|0);
     if ($cond374) {
      $59 = 1 << $55;
      $60 = $59 ^ -1;
      $61 = HEAP32[(163780)>>2]|0;
      $62 = $61 & $60;
      HEAP32[(163780)>>2] = $62;
      $$1 = $14;$$1347 = $15;$87 = $14;
      break;
     }
    } else {
     $63 = ((($33)) + 16|0);
     $64 = HEAP32[$63>>2]|0;
     $not$370 = ($64|0)!=($14|0);
     $$sink3 = $not$370&1;
     $65 = (((($33)) + 16|0) + ($$sink3<<2)|0);
     HEAP32[$65>>2] = $$3;
     $66 = ($$3|0)==(0|0);
     if ($66) {
      $$1 = $14;$$1347 = $15;$87 = $14;
      break;
     }
    }
    $67 = ((($$3)) + 24|0);
    HEAP32[$67>>2] = $33;
    $68 = ((($14)) + 16|0);
    $69 = HEAP32[$68>>2]|0;
    $70 = ($69|0)==(0|0);
    if (!($70)) {
     $71 = ((($$3)) + 16|0);
     HEAP32[$71>>2] = $69;
     $72 = ((($69)) + 24|0);
     HEAP32[$72>>2] = $$3;
    }
    $73 = ((($68)) + 4|0);
    $74 = HEAP32[$73>>2]|0;
    $75 = ($74|0)==(0|0);
    if ($75) {
     $$1 = $14;$$1347 = $15;$87 = $14;
    } else {
     $76 = ((($$3)) + 20|0);
     HEAP32[$76>>2] = $74;
     $77 = ((($74)) + 24|0);
     HEAP32[$77>>2] = $$3;
     $$1 = $14;$$1347 = $15;$87 = $14;
    }
   }
  } else {
   $$1 = $2;$$1347 = $6;$87 = $2;
  }
 } while(0);
 $86 = ($87>>>0)<($7>>>0);
 if (!($86)) {
  return;
 }
 $88 = ((($7)) + 4|0);
 $89 = HEAP32[$88>>2]|0;
 $90 = $89 & 1;
 $91 = ($90|0)==(0);
 if ($91) {
  return;
 }
 $92 = $89 & 2;
 $93 = ($92|0)==(0);
 if ($93) {
  $94 = HEAP32[(163800)>>2]|0;
  $95 = ($7|0)==($94|0);
  $96 = HEAP32[(163796)>>2]|0;
  if ($95) {
   $97 = HEAP32[(163788)>>2]|0;
   $98 = (($97) + ($$1347))|0;
   HEAP32[(163788)>>2] = $98;
   HEAP32[(163800)>>2] = $$1;
   $99 = $98 | 1;
   $100 = ((($$1)) + 4|0);
   HEAP32[$100>>2] = $99;
   $101 = ($$1|0)==($96|0);
   if (!($101)) {
    return;
   }
   HEAP32[(163796)>>2] = 0;
   HEAP32[(163784)>>2] = 0;
   return;
  }
  $102 = ($7|0)==($96|0);
  if ($102) {
   $103 = HEAP32[(163784)>>2]|0;
   $104 = (($103) + ($$1347))|0;
   HEAP32[(163784)>>2] = $104;
   HEAP32[(163796)>>2] = $87;
   $105 = $104 | 1;
   $106 = ((($$1)) + 4|0);
   HEAP32[$106>>2] = $105;
   $107 = (($87) + ($104)|0);
   HEAP32[$107>>2] = $104;
   return;
  }
  $108 = $89 & -8;
  $109 = (($108) + ($$1347))|0;
  $110 = $89 >>> 3;
  $111 = ($89>>>0)<(256);
  do {
   if ($111) {
    $112 = ((($7)) + 8|0);
    $113 = HEAP32[$112>>2]|0;
    $114 = ((($7)) + 12|0);
    $115 = HEAP32[$114>>2]|0;
    $116 = ($115|0)==($113|0);
    if ($116) {
     $117 = 1 << $110;
     $118 = $117 ^ -1;
     $119 = HEAP32[40944]|0;
     $120 = $119 & $118;
     HEAP32[40944] = $120;
     break;
    } else {
     $121 = ((($113)) + 12|0);
     HEAP32[$121>>2] = $115;
     $122 = ((($115)) + 8|0);
     HEAP32[$122>>2] = $113;
     break;
    }
   } else {
    $123 = ((($7)) + 24|0);
    $124 = HEAP32[$123>>2]|0;
    $125 = ((($7)) + 12|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = ($126|0)==($7|0);
    do {
     if ($127) {
      $132 = ((($7)) + 16|0);
      $133 = ((($132)) + 4|0);
      $134 = HEAP32[$133>>2]|0;
      $135 = ($134|0)==(0|0);
      if ($135) {
       $136 = HEAP32[$132>>2]|0;
       $137 = ($136|0)==(0|0);
       if ($137) {
        $$3365 = 0;
        break;
       } else {
        $$1363 = $136;$$1367 = $132;
       }
      } else {
       $$1363 = $134;$$1367 = $133;
      }
      while(1) {
       $138 = ((($$1363)) + 20|0);
       $139 = HEAP32[$138>>2]|0;
       $140 = ($139|0)==(0|0);
       if (!($140)) {
        $$1363 = $139;$$1367 = $138;
        continue;
       }
       $141 = ((($$1363)) + 16|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ($142|0)==(0|0);
       if ($143) {
        break;
       } else {
        $$1363 = $142;$$1367 = $141;
       }
      }
      HEAP32[$$1367>>2] = 0;
      $$3365 = $$1363;
     } else {
      $128 = ((($7)) + 8|0);
      $129 = HEAP32[$128>>2]|0;
      $130 = ((($129)) + 12|0);
      HEAP32[$130>>2] = $126;
      $131 = ((($126)) + 8|0);
      HEAP32[$131>>2] = $129;
      $$3365 = $126;
     }
    } while(0);
    $144 = ($124|0)==(0|0);
    if (!($144)) {
     $145 = ((($7)) + 28|0);
     $146 = HEAP32[$145>>2]|0;
     $147 = (164080 + ($146<<2)|0);
     $148 = HEAP32[$147>>2]|0;
     $149 = ($7|0)==($148|0);
     if ($149) {
      HEAP32[$147>>2] = $$3365;
      $cond375 = ($$3365|0)==(0|0);
      if ($cond375) {
       $150 = 1 << $146;
       $151 = $150 ^ -1;
       $152 = HEAP32[(163780)>>2]|0;
       $153 = $152 & $151;
       HEAP32[(163780)>>2] = $153;
       break;
      }
     } else {
      $154 = ((($124)) + 16|0);
      $155 = HEAP32[$154>>2]|0;
      $not$ = ($155|0)!=($7|0);
      $$sink5 = $not$&1;
      $156 = (((($124)) + 16|0) + ($$sink5<<2)|0);
      HEAP32[$156>>2] = $$3365;
      $157 = ($$3365|0)==(0|0);
      if ($157) {
       break;
      }
     }
     $158 = ((($$3365)) + 24|0);
     HEAP32[$158>>2] = $124;
     $159 = ((($7)) + 16|0);
     $160 = HEAP32[$159>>2]|0;
     $161 = ($160|0)==(0|0);
     if (!($161)) {
      $162 = ((($$3365)) + 16|0);
      HEAP32[$162>>2] = $160;
      $163 = ((($160)) + 24|0);
      HEAP32[$163>>2] = $$3365;
     }
     $164 = ((($159)) + 4|0);
     $165 = HEAP32[$164>>2]|0;
     $166 = ($165|0)==(0|0);
     if (!($166)) {
      $167 = ((($$3365)) + 20|0);
      HEAP32[$167>>2] = $165;
      $168 = ((($165)) + 24|0);
      HEAP32[$168>>2] = $$3365;
     }
    }
   }
  } while(0);
  $169 = $109 | 1;
  $170 = ((($$1)) + 4|0);
  HEAP32[$170>>2] = $169;
  $171 = (($87) + ($109)|0);
  HEAP32[$171>>2] = $109;
  $172 = HEAP32[(163796)>>2]|0;
  $173 = ($$1|0)==($172|0);
  if ($173) {
   HEAP32[(163784)>>2] = $109;
   return;
  } else {
   $$2 = $109;
  }
 } else {
  $174 = $89 & -2;
  HEAP32[$88>>2] = $174;
  $175 = $$1347 | 1;
  $176 = ((($$1)) + 4|0);
  HEAP32[$176>>2] = $175;
  $177 = (($87) + ($$1347)|0);
  HEAP32[$177>>2] = $$1347;
  $$2 = $$1347;
 }
 $178 = $$2 >>> 3;
 $179 = ($$2>>>0)<(256);
 if ($179) {
  $180 = $178 << 1;
  $181 = (163816 + ($180<<2)|0);
  $182 = HEAP32[40944]|0;
  $183 = 1 << $178;
  $184 = $182 & $183;
  $185 = ($184|0)==(0);
  if ($185) {
   $186 = $182 | $183;
   HEAP32[40944] = $186;
   $$pre = ((($181)) + 8|0);
   $$0368 = $181;$$pre$phiZ2D = $$pre;
  } else {
   $187 = ((($181)) + 8|0);
   $188 = HEAP32[$187>>2]|0;
   $$0368 = $188;$$pre$phiZ2D = $187;
  }
  HEAP32[$$pre$phiZ2D>>2] = $$1;
  $189 = ((($$0368)) + 12|0);
  HEAP32[$189>>2] = $$1;
  $190 = ((($$1)) + 8|0);
  HEAP32[$190>>2] = $$0368;
  $191 = ((($$1)) + 12|0);
  HEAP32[$191>>2] = $181;
  return;
 }
 $192 = $$2 >>> 8;
 $193 = ($192|0)==(0);
 if ($193) {
  $$0361 = 0;
 } else {
  $194 = ($$2>>>0)>(16777215);
  if ($194) {
   $$0361 = 31;
  } else {
   $195 = (($192) + 1048320)|0;
   $196 = $195 >>> 16;
   $197 = $196 & 8;
   $198 = $192 << $197;
   $199 = (($198) + 520192)|0;
   $200 = $199 >>> 16;
   $201 = $200 & 4;
   $202 = $201 | $197;
   $203 = $198 << $201;
   $204 = (($203) + 245760)|0;
   $205 = $204 >>> 16;
   $206 = $205 & 2;
   $207 = $202 | $206;
   $208 = (14 - ($207))|0;
   $209 = $203 << $206;
   $210 = $209 >>> 15;
   $211 = (($208) + ($210))|0;
   $212 = $211 << 1;
   $213 = (($211) + 7)|0;
   $214 = $$2 >>> $213;
   $215 = $214 & 1;
   $216 = $215 | $212;
   $$0361 = $216;
  }
 }
 $217 = (164080 + ($$0361<<2)|0);
 $218 = ((($$1)) + 28|0);
 HEAP32[$218>>2] = $$0361;
 $219 = ((($$1)) + 16|0);
 $220 = ((($$1)) + 20|0);
 HEAP32[$220>>2] = 0;
 HEAP32[$219>>2] = 0;
 $221 = HEAP32[(163780)>>2]|0;
 $222 = 1 << $$0361;
 $223 = $221 & $222;
 $224 = ($223|0)==(0);
 do {
  if ($224) {
   $225 = $221 | $222;
   HEAP32[(163780)>>2] = $225;
   HEAP32[$217>>2] = $$1;
   $226 = ((($$1)) + 24|0);
   HEAP32[$226>>2] = $217;
   $227 = ((($$1)) + 12|0);
   HEAP32[$227>>2] = $$1;
   $228 = ((($$1)) + 8|0);
   HEAP32[$228>>2] = $$1;
  } else {
   $229 = HEAP32[$217>>2]|0;
   $230 = ($$0361|0)==(31);
   $231 = $$0361 >>> 1;
   $232 = (25 - ($231))|0;
   $233 = $230 ? 0 : $232;
   $234 = $$2 << $233;
   $$0348 = $234;$$0349 = $229;
   while(1) {
    $235 = ((($$0349)) + 4|0);
    $236 = HEAP32[$235>>2]|0;
    $237 = $236 & -8;
    $238 = ($237|0)==($$2|0);
    if ($238) {
     label = 73;
     break;
    }
    $239 = $$0348 >>> 31;
    $240 = (((($$0349)) + 16|0) + ($239<<2)|0);
    $241 = $$0348 << 1;
    $242 = HEAP32[$240>>2]|0;
    $243 = ($242|0)==(0|0);
    if ($243) {
     label = 72;
     break;
    } else {
     $$0348 = $241;$$0349 = $242;
    }
   }
   if ((label|0) == 72) {
    HEAP32[$240>>2] = $$1;
    $244 = ((($$1)) + 24|0);
    HEAP32[$244>>2] = $$0349;
    $245 = ((($$1)) + 12|0);
    HEAP32[$245>>2] = $$1;
    $246 = ((($$1)) + 8|0);
    HEAP32[$246>>2] = $$1;
    break;
   }
   else if ((label|0) == 73) {
    $247 = ((($$0349)) + 8|0);
    $248 = HEAP32[$247>>2]|0;
    $249 = ((($248)) + 12|0);
    HEAP32[$249>>2] = $$1;
    HEAP32[$247>>2] = $$1;
    $250 = ((($$1)) + 8|0);
    HEAP32[$250>>2] = $248;
    $251 = ((($$1)) + 12|0);
    HEAP32[$251>>2] = $$0349;
    $252 = ((($$1)) + 24|0);
    HEAP32[$252>>2] = 0;
    break;
   }
  }
 } while(0);
 $253 = HEAP32[(163808)>>2]|0;
 $254 = (($253) + -1)|0;
 HEAP32[(163808)>>2] = $254;
 $255 = ($254|0)==(0);
 if ($255) {
  $$0195$in$i = (164232);
 } else {
  return;
 }
 while(1) {
  $$0195$i = HEAP32[$$0195$in$i>>2]|0;
  $256 = ($$0195$i|0)==(0|0);
  $257 = ((($$0195$i)) + 8|0);
  if ($256) {
   break;
  } else {
   $$0195$in$i = $257;
  }
 }
 HEAP32[(163808)>>2] = -1;
 return;
}
function _realloc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==(0|0);
 if ($2) {
  $3 = (_malloc($1)|0);
  $$1 = $3;
  return ($$1|0);
 }
 $4 = ($1>>>0)>(4294967231);
 if ($4) {
  $5 = (___errno_location()|0);
  HEAP32[$5>>2] = 12;
  $$1 = 0;
  return ($$1|0);
 }
 $6 = ($1>>>0)<(11);
 $7 = (($1) + 11)|0;
 $8 = $7 & -8;
 $9 = $6 ? 16 : $8;
 $10 = ((($0)) + -8|0);
 $11 = (_try_realloc_chunk($10,$9)|0);
 $12 = ($11|0)==(0|0);
 if (!($12)) {
  $13 = ((($11)) + 8|0);
  $$1 = $13;
  return ($$1|0);
 }
 $14 = (_malloc($1)|0);
 $15 = ($14|0)==(0|0);
 if ($15) {
  $$1 = 0;
  return ($$1|0);
 }
 $16 = ((($0)) + -4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $17 & -8;
 $19 = $17 & 3;
 $20 = ($19|0)==(0);
 $21 = $20 ? 8 : 4;
 $22 = (($18) - ($21))|0;
 $23 = ($22>>>0)<($1>>>0);
 $24 = $23 ? $22 : $1;
 _memcpy(($14|0),($0|0),($24|0))|0;
 _free($0);
 $$1 = $14;
 return ($$1|0);
}
function _try_realloc_chunk($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1246 = 0, $$1249 = 0, $$2 = 0, $$3 = 0, $$sink1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, $not$ = 0, $storemerge = 0, $storemerge1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = $3 & -8;
 $5 = (($0) + ($4)|0);
 $6 = $3 & 3;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ($1>>>0)<(256);
  if ($8) {
   $$2 = 0;
   return ($$2|0);
  }
  $9 = (($1) + 4)|0;
  $10 = ($4>>>0)<($9>>>0);
  if (!($10)) {
   $11 = (($4) - ($1))|0;
   $12 = HEAP32[(164256)>>2]|0;
   $13 = $12 << 1;
   $14 = ($11>>>0)>($13>>>0);
   if (!($14)) {
    $$2 = $0;
    return ($$2|0);
   }
  }
  $$2 = 0;
  return ($$2|0);
 }
 $15 = ($4>>>0)<($1>>>0);
 if (!($15)) {
  $16 = (($4) - ($1))|0;
  $17 = ($16>>>0)>(15);
  if (!($17)) {
   $$2 = $0;
   return ($$2|0);
  }
  $18 = (($0) + ($1)|0);
  $19 = $3 & 1;
  $20 = $19 | $1;
  $21 = $20 | 2;
  HEAP32[$2>>2] = $21;
  $22 = ((($18)) + 4|0);
  $23 = $16 | 3;
  HEAP32[$22>>2] = $23;
  $24 = (($18) + ($16)|0);
  $25 = ((($24)) + 4|0);
  $26 = HEAP32[$25>>2]|0;
  $27 = $26 | 1;
  HEAP32[$25>>2] = $27;
  _dispose_chunk($18,$16);
  $$2 = $0;
  return ($$2|0);
 }
 $28 = HEAP32[(163800)>>2]|0;
 $29 = ($5|0)==($28|0);
 if ($29) {
  $30 = HEAP32[(163788)>>2]|0;
  $31 = (($30) + ($4))|0;
  $32 = ($31>>>0)>($1>>>0);
  $33 = (($31) - ($1))|0;
  $34 = (($0) + ($1)|0);
  if (!($32)) {
   $$2 = 0;
   return ($$2|0);
  }
  $35 = $33 | 1;
  $36 = ((($34)) + 4|0);
  $37 = $3 & 1;
  $38 = $37 | $1;
  $39 = $38 | 2;
  HEAP32[$2>>2] = $39;
  HEAP32[$36>>2] = $35;
  HEAP32[(163800)>>2] = $34;
  HEAP32[(163788)>>2] = $33;
  $$2 = $0;
  return ($$2|0);
 }
 $40 = HEAP32[(163796)>>2]|0;
 $41 = ($5|0)==($40|0);
 if ($41) {
  $42 = HEAP32[(163784)>>2]|0;
  $43 = (($42) + ($4))|0;
  $44 = ($43>>>0)<($1>>>0);
  if ($44) {
   $$2 = 0;
   return ($$2|0);
  }
  $45 = (($43) - ($1))|0;
  $46 = ($45>>>0)>(15);
  $47 = $3 & 1;
  if ($46) {
   $48 = (($0) + ($1)|0);
   $49 = (($48) + ($45)|0);
   $50 = $47 | $1;
   $51 = $50 | 2;
   HEAP32[$2>>2] = $51;
   $52 = ((($48)) + 4|0);
   $53 = $45 | 1;
   HEAP32[$52>>2] = $53;
   HEAP32[$49>>2] = $45;
   $54 = ((($49)) + 4|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = $55 & -2;
   HEAP32[$54>>2] = $56;
   $storemerge = $48;$storemerge1 = $45;
  } else {
   $57 = $47 | $43;
   $58 = $57 | 2;
   HEAP32[$2>>2] = $58;
   $59 = (($0) + ($43)|0);
   $60 = ((($59)) + 4|0);
   $61 = HEAP32[$60>>2]|0;
   $62 = $61 | 1;
   HEAP32[$60>>2] = $62;
   $storemerge = 0;$storemerge1 = 0;
  }
  HEAP32[(163784)>>2] = $storemerge1;
  HEAP32[(163796)>>2] = $storemerge;
  $$2 = $0;
  return ($$2|0);
 }
 $63 = ((($5)) + 4|0);
 $64 = HEAP32[$63>>2]|0;
 $65 = $64 & 2;
 $66 = ($65|0)==(0);
 if (!($66)) {
  $$2 = 0;
  return ($$2|0);
 }
 $67 = $64 & -8;
 $68 = (($67) + ($4))|0;
 $69 = ($68>>>0)<($1>>>0);
 if ($69) {
  $$2 = 0;
  return ($$2|0);
 }
 $70 = (($68) - ($1))|0;
 $71 = $64 >>> 3;
 $72 = ($64>>>0)<(256);
 do {
  if ($72) {
   $73 = ((($5)) + 8|0);
   $74 = HEAP32[$73>>2]|0;
   $75 = ((($5)) + 12|0);
   $76 = HEAP32[$75>>2]|0;
   $77 = ($76|0)==($74|0);
   if ($77) {
    $78 = 1 << $71;
    $79 = $78 ^ -1;
    $80 = HEAP32[40944]|0;
    $81 = $80 & $79;
    HEAP32[40944] = $81;
    break;
   } else {
    $82 = ((($74)) + 12|0);
    HEAP32[$82>>2] = $76;
    $83 = ((($76)) + 8|0);
    HEAP32[$83>>2] = $74;
    break;
   }
  } else {
   $84 = ((($5)) + 24|0);
   $85 = HEAP32[$84>>2]|0;
   $86 = ((($5)) + 12|0);
   $87 = HEAP32[$86>>2]|0;
   $88 = ($87|0)==($5|0);
   do {
    if ($88) {
     $93 = ((($5)) + 16|0);
     $94 = ((($93)) + 4|0);
     $95 = HEAP32[$94>>2]|0;
     $96 = ($95|0)==(0|0);
     if ($96) {
      $97 = HEAP32[$93>>2]|0;
      $98 = ($97|0)==(0|0);
      if ($98) {
       $$3 = 0;
       break;
      } else {
       $$1246 = $97;$$1249 = $93;
      }
     } else {
      $$1246 = $95;$$1249 = $94;
     }
     while(1) {
      $99 = ((($$1246)) + 20|0);
      $100 = HEAP32[$99>>2]|0;
      $101 = ($100|0)==(0|0);
      if (!($101)) {
       $$1246 = $100;$$1249 = $99;
       continue;
      }
      $102 = ((($$1246)) + 16|0);
      $103 = HEAP32[$102>>2]|0;
      $104 = ($103|0)==(0|0);
      if ($104) {
       break;
      } else {
       $$1246 = $103;$$1249 = $102;
      }
     }
     HEAP32[$$1249>>2] = 0;
     $$3 = $$1246;
    } else {
     $89 = ((($5)) + 8|0);
     $90 = HEAP32[$89>>2]|0;
     $91 = ((($90)) + 12|0);
     HEAP32[$91>>2] = $87;
     $92 = ((($87)) + 8|0);
     HEAP32[$92>>2] = $90;
     $$3 = $87;
    }
   } while(0);
   $105 = ($85|0)==(0|0);
   if (!($105)) {
    $106 = ((($5)) + 28|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = (164080 + ($107<<2)|0);
    $109 = HEAP32[$108>>2]|0;
    $110 = ($5|0)==($109|0);
    if ($110) {
     HEAP32[$108>>2] = $$3;
     $cond = ($$3|0)==(0|0);
     if ($cond) {
      $111 = 1 << $107;
      $112 = $111 ^ -1;
      $113 = HEAP32[(163780)>>2]|0;
      $114 = $113 & $112;
      HEAP32[(163780)>>2] = $114;
      break;
     }
    } else {
     $115 = ((($85)) + 16|0);
     $116 = HEAP32[$115>>2]|0;
     $not$ = ($116|0)!=($5|0);
     $$sink1 = $not$&1;
     $117 = (((($85)) + 16|0) + ($$sink1<<2)|0);
     HEAP32[$117>>2] = $$3;
     $118 = ($$3|0)==(0|0);
     if ($118) {
      break;
     }
    }
    $119 = ((($$3)) + 24|0);
    HEAP32[$119>>2] = $85;
    $120 = ((($5)) + 16|0);
    $121 = HEAP32[$120>>2]|0;
    $122 = ($121|0)==(0|0);
    if (!($122)) {
     $123 = ((($$3)) + 16|0);
     HEAP32[$123>>2] = $121;
     $124 = ((($121)) + 24|0);
     HEAP32[$124>>2] = $$3;
    }
    $125 = ((($120)) + 4|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = ($126|0)==(0|0);
    if (!($127)) {
     $128 = ((($$3)) + 20|0);
     HEAP32[$128>>2] = $126;
     $129 = ((($126)) + 24|0);
     HEAP32[$129>>2] = $$3;
    }
   }
  }
 } while(0);
 $130 = ($70>>>0)<(16);
 $131 = $3 & 1;
 if ($130) {
  $132 = $68 | $131;
  $133 = $132 | 2;
  HEAP32[$2>>2] = $133;
  $134 = (($0) + ($68)|0);
  $135 = ((($134)) + 4|0);
  $136 = HEAP32[$135>>2]|0;
  $137 = $136 | 1;
  HEAP32[$135>>2] = $137;
  $$2 = $0;
  return ($$2|0);
 } else {
  $138 = (($0) + ($1)|0);
  $139 = $131 | $1;
  $140 = $139 | 2;
  HEAP32[$2>>2] = $140;
  $141 = ((($138)) + 4|0);
  $142 = $70 | 3;
  HEAP32[$141>>2] = $142;
  $143 = (($138) + ($70)|0);
  $144 = ((($143)) + 4|0);
  $145 = HEAP32[$144>>2]|0;
  $146 = $145 | 1;
  HEAP32[$144>>2] = $146;
  _dispose_chunk($138,$70);
  $$2 = $0;
  return ($$2|0);
 }
 return (0)|0;
}
function _dispose_chunk($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0366 = 0, $$0367 = 0, $$0378 = 0, $$0385 = 0, $$1 = 0, $$1365 = 0, $$1373 = 0, $$1376 = 0, $$1380 = 0, $$1384 = 0, $$2 = 0, $$3 = 0, $$3382 = 0, $$pre = 0, $$pre$phiZ2D = 0, $$sink2 = 0, $$sink4 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0;
 var $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0;
 var $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0;
 var $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $2 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $cond = 0, $cond5 = 0, $not$ = 0, $not$1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (($0) + ($1)|0);
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 1;
 $6 = ($5|0)==(0);
 do {
  if ($6) {
   $7 = HEAP32[$0>>2]|0;
   $8 = $4 & 3;
   $9 = ($8|0)==(0);
   if ($9) {
    return;
   }
   $10 = (0 - ($7))|0;
   $11 = (($0) + ($10)|0);
   $12 = (($7) + ($1))|0;
   $13 = HEAP32[(163796)>>2]|0;
   $14 = ($11|0)==($13|0);
   if ($14) {
    $74 = ((($2)) + 4|0);
    $75 = HEAP32[$74>>2]|0;
    $76 = $75 & 3;
    $77 = ($76|0)==(3);
    if (!($77)) {
     $$1 = $11;$$1365 = $12;
     break;
    }
    $78 = (($11) + ($12)|0);
    $79 = ((($11)) + 4|0);
    $80 = $12 | 1;
    $81 = $75 & -2;
    HEAP32[(163784)>>2] = $12;
    HEAP32[$74>>2] = $81;
    HEAP32[$79>>2] = $80;
    HEAP32[$78>>2] = $12;
    return;
   }
   $15 = $7 >>> 3;
   $16 = ($7>>>0)<(256);
   if ($16) {
    $17 = ((($11)) + 8|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ((($11)) + 12|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ($20|0)==($18|0);
    if ($21) {
     $22 = 1 << $15;
     $23 = $22 ^ -1;
     $24 = HEAP32[40944]|0;
     $25 = $24 & $23;
     HEAP32[40944] = $25;
     $$1 = $11;$$1365 = $12;
     break;
    } else {
     $26 = ((($18)) + 12|0);
     HEAP32[$26>>2] = $20;
     $27 = ((($20)) + 8|0);
     HEAP32[$27>>2] = $18;
     $$1 = $11;$$1365 = $12;
     break;
    }
   }
   $28 = ((($11)) + 24|0);
   $29 = HEAP32[$28>>2]|0;
   $30 = ((($11)) + 12|0);
   $31 = HEAP32[$30>>2]|0;
   $32 = ($31|0)==($11|0);
   do {
    if ($32) {
     $37 = ((($11)) + 16|0);
     $38 = ((($37)) + 4|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($39|0)==(0|0);
     if ($40) {
      $41 = HEAP32[$37>>2]|0;
      $42 = ($41|0)==(0|0);
      if ($42) {
       $$3 = 0;
       break;
      } else {
       $$1373 = $41;$$1376 = $37;
      }
     } else {
      $$1373 = $39;$$1376 = $38;
     }
     while(1) {
      $43 = ((($$1373)) + 20|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = ($44|0)==(0|0);
      if (!($45)) {
       $$1373 = $44;$$1376 = $43;
       continue;
      }
      $46 = ((($$1373)) + 16|0);
      $47 = HEAP32[$46>>2]|0;
      $48 = ($47|0)==(0|0);
      if ($48) {
       break;
      } else {
       $$1373 = $47;$$1376 = $46;
      }
     }
     HEAP32[$$1376>>2] = 0;
     $$3 = $$1373;
    } else {
     $33 = ((($11)) + 8|0);
     $34 = HEAP32[$33>>2]|0;
     $35 = ((($34)) + 12|0);
     HEAP32[$35>>2] = $31;
     $36 = ((($31)) + 8|0);
     HEAP32[$36>>2] = $34;
     $$3 = $31;
    }
   } while(0);
   $49 = ($29|0)==(0|0);
   if ($49) {
    $$1 = $11;$$1365 = $12;
   } else {
    $50 = ((($11)) + 28|0);
    $51 = HEAP32[$50>>2]|0;
    $52 = (164080 + ($51<<2)|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = ($11|0)==($53|0);
    if ($54) {
     HEAP32[$52>>2] = $$3;
     $cond = ($$3|0)==(0|0);
     if ($cond) {
      $55 = 1 << $51;
      $56 = $55 ^ -1;
      $57 = HEAP32[(163780)>>2]|0;
      $58 = $57 & $56;
      HEAP32[(163780)>>2] = $58;
      $$1 = $11;$$1365 = $12;
      break;
     }
    } else {
     $59 = ((($29)) + 16|0);
     $60 = HEAP32[$59>>2]|0;
     $not$1 = ($60|0)!=($11|0);
     $$sink2 = $not$1&1;
     $61 = (((($29)) + 16|0) + ($$sink2<<2)|0);
     HEAP32[$61>>2] = $$3;
     $62 = ($$3|0)==(0|0);
     if ($62) {
      $$1 = $11;$$1365 = $12;
      break;
     }
    }
    $63 = ((($$3)) + 24|0);
    HEAP32[$63>>2] = $29;
    $64 = ((($11)) + 16|0);
    $65 = HEAP32[$64>>2]|0;
    $66 = ($65|0)==(0|0);
    if (!($66)) {
     $67 = ((($$3)) + 16|0);
     HEAP32[$67>>2] = $65;
     $68 = ((($65)) + 24|0);
     HEAP32[$68>>2] = $$3;
    }
    $69 = ((($64)) + 4|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = ($70|0)==(0|0);
    if ($71) {
     $$1 = $11;$$1365 = $12;
    } else {
     $72 = ((($$3)) + 20|0);
     HEAP32[$72>>2] = $70;
     $73 = ((($70)) + 24|0);
     HEAP32[$73>>2] = $$3;
     $$1 = $11;$$1365 = $12;
    }
   }
  } else {
   $$1 = $0;$$1365 = $1;
  }
 } while(0);
 $82 = ((($2)) + 4|0);
 $83 = HEAP32[$82>>2]|0;
 $84 = $83 & 2;
 $85 = ($84|0)==(0);
 if ($85) {
  $86 = HEAP32[(163800)>>2]|0;
  $87 = ($2|0)==($86|0);
  $88 = HEAP32[(163796)>>2]|0;
  if ($87) {
   $89 = HEAP32[(163788)>>2]|0;
   $90 = (($89) + ($$1365))|0;
   HEAP32[(163788)>>2] = $90;
   HEAP32[(163800)>>2] = $$1;
   $91 = $90 | 1;
   $92 = ((($$1)) + 4|0);
   HEAP32[$92>>2] = $91;
   $93 = ($$1|0)==($88|0);
   if (!($93)) {
    return;
   }
   HEAP32[(163796)>>2] = 0;
   HEAP32[(163784)>>2] = 0;
   return;
  }
  $94 = ($2|0)==($88|0);
  if ($94) {
   $95 = HEAP32[(163784)>>2]|0;
   $96 = (($95) + ($$1365))|0;
   HEAP32[(163784)>>2] = $96;
   HEAP32[(163796)>>2] = $$1;
   $97 = $96 | 1;
   $98 = ((($$1)) + 4|0);
   HEAP32[$98>>2] = $97;
   $99 = (($$1) + ($96)|0);
   HEAP32[$99>>2] = $96;
   return;
  }
  $100 = $83 & -8;
  $101 = (($100) + ($$1365))|0;
  $102 = $83 >>> 3;
  $103 = ($83>>>0)<(256);
  do {
   if ($103) {
    $104 = ((($2)) + 8|0);
    $105 = HEAP32[$104>>2]|0;
    $106 = ((($2)) + 12|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = ($107|0)==($105|0);
    if ($108) {
     $109 = 1 << $102;
     $110 = $109 ^ -1;
     $111 = HEAP32[40944]|0;
     $112 = $111 & $110;
     HEAP32[40944] = $112;
     break;
    } else {
     $113 = ((($105)) + 12|0);
     HEAP32[$113>>2] = $107;
     $114 = ((($107)) + 8|0);
     HEAP32[$114>>2] = $105;
     break;
    }
   } else {
    $115 = ((($2)) + 24|0);
    $116 = HEAP32[$115>>2]|0;
    $117 = ((($2)) + 12|0);
    $118 = HEAP32[$117>>2]|0;
    $119 = ($118|0)==($2|0);
    do {
     if ($119) {
      $124 = ((($2)) + 16|0);
      $125 = ((($124)) + 4|0);
      $126 = HEAP32[$125>>2]|0;
      $127 = ($126|0)==(0|0);
      if ($127) {
       $128 = HEAP32[$124>>2]|0;
       $129 = ($128|0)==(0|0);
       if ($129) {
        $$3382 = 0;
        break;
       } else {
        $$1380 = $128;$$1384 = $124;
       }
      } else {
       $$1380 = $126;$$1384 = $125;
      }
      while(1) {
       $130 = ((($$1380)) + 20|0);
       $131 = HEAP32[$130>>2]|0;
       $132 = ($131|0)==(0|0);
       if (!($132)) {
        $$1380 = $131;$$1384 = $130;
        continue;
       }
       $133 = ((($$1380)) + 16|0);
       $134 = HEAP32[$133>>2]|0;
       $135 = ($134|0)==(0|0);
       if ($135) {
        break;
       } else {
        $$1380 = $134;$$1384 = $133;
       }
      }
      HEAP32[$$1384>>2] = 0;
      $$3382 = $$1380;
     } else {
      $120 = ((($2)) + 8|0);
      $121 = HEAP32[$120>>2]|0;
      $122 = ((($121)) + 12|0);
      HEAP32[$122>>2] = $118;
      $123 = ((($118)) + 8|0);
      HEAP32[$123>>2] = $121;
      $$3382 = $118;
     }
    } while(0);
    $136 = ($116|0)==(0|0);
    if (!($136)) {
     $137 = ((($2)) + 28|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = (164080 + ($138<<2)|0);
     $140 = HEAP32[$139>>2]|0;
     $141 = ($2|0)==($140|0);
     if ($141) {
      HEAP32[$139>>2] = $$3382;
      $cond5 = ($$3382|0)==(0|0);
      if ($cond5) {
       $142 = 1 << $138;
       $143 = $142 ^ -1;
       $144 = HEAP32[(163780)>>2]|0;
       $145 = $144 & $143;
       HEAP32[(163780)>>2] = $145;
       break;
      }
     } else {
      $146 = ((($116)) + 16|0);
      $147 = HEAP32[$146>>2]|0;
      $not$ = ($147|0)!=($2|0);
      $$sink4 = $not$&1;
      $148 = (((($116)) + 16|0) + ($$sink4<<2)|0);
      HEAP32[$148>>2] = $$3382;
      $149 = ($$3382|0)==(0|0);
      if ($149) {
       break;
      }
     }
     $150 = ((($$3382)) + 24|0);
     HEAP32[$150>>2] = $116;
     $151 = ((($2)) + 16|0);
     $152 = HEAP32[$151>>2]|0;
     $153 = ($152|0)==(0|0);
     if (!($153)) {
      $154 = ((($$3382)) + 16|0);
      HEAP32[$154>>2] = $152;
      $155 = ((($152)) + 24|0);
      HEAP32[$155>>2] = $$3382;
     }
     $156 = ((($151)) + 4|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = ($157|0)==(0|0);
     if (!($158)) {
      $159 = ((($$3382)) + 20|0);
      HEAP32[$159>>2] = $157;
      $160 = ((($157)) + 24|0);
      HEAP32[$160>>2] = $$3382;
     }
    }
   }
  } while(0);
  $161 = $101 | 1;
  $162 = ((($$1)) + 4|0);
  HEAP32[$162>>2] = $161;
  $163 = (($$1) + ($101)|0);
  HEAP32[$163>>2] = $101;
  $164 = HEAP32[(163796)>>2]|0;
  $165 = ($$1|0)==($164|0);
  if ($165) {
   HEAP32[(163784)>>2] = $101;
   return;
  } else {
   $$2 = $101;
  }
 } else {
  $166 = $83 & -2;
  HEAP32[$82>>2] = $166;
  $167 = $$1365 | 1;
  $168 = ((($$1)) + 4|0);
  HEAP32[$168>>2] = $167;
  $169 = (($$1) + ($$1365)|0);
  HEAP32[$169>>2] = $$1365;
  $$2 = $$1365;
 }
 $170 = $$2 >>> 3;
 $171 = ($$2>>>0)<(256);
 if ($171) {
  $172 = $170 << 1;
  $173 = (163816 + ($172<<2)|0);
  $174 = HEAP32[40944]|0;
  $175 = 1 << $170;
  $176 = $174 & $175;
  $177 = ($176|0)==(0);
  if ($177) {
   $178 = $174 | $175;
   HEAP32[40944] = $178;
   $$pre = ((($173)) + 8|0);
   $$0385 = $173;$$pre$phiZ2D = $$pre;
  } else {
   $179 = ((($173)) + 8|0);
   $180 = HEAP32[$179>>2]|0;
   $$0385 = $180;$$pre$phiZ2D = $179;
  }
  HEAP32[$$pre$phiZ2D>>2] = $$1;
  $181 = ((($$0385)) + 12|0);
  HEAP32[$181>>2] = $$1;
  $182 = ((($$1)) + 8|0);
  HEAP32[$182>>2] = $$0385;
  $183 = ((($$1)) + 12|0);
  HEAP32[$183>>2] = $173;
  return;
 }
 $184 = $$2 >>> 8;
 $185 = ($184|0)==(0);
 if ($185) {
  $$0378 = 0;
 } else {
  $186 = ($$2>>>0)>(16777215);
  if ($186) {
   $$0378 = 31;
  } else {
   $187 = (($184) + 1048320)|0;
   $188 = $187 >>> 16;
   $189 = $188 & 8;
   $190 = $184 << $189;
   $191 = (($190) + 520192)|0;
   $192 = $191 >>> 16;
   $193 = $192 & 4;
   $194 = $193 | $189;
   $195 = $190 << $193;
   $196 = (($195) + 245760)|0;
   $197 = $196 >>> 16;
   $198 = $197 & 2;
   $199 = $194 | $198;
   $200 = (14 - ($199))|0;
   $201 = $195 << $198;
   $202 = $201 >>> 15;
   $203 = (($200) + ($202))|0;
   $204 = $203 << 1;
   $205 = (($203) + 7)|0;
   $206 = $$2 >>> $205;
   $207 = $206 & 1;
   $208 = $207 | $204;
   $$0378 = $208;
  }
 }
 $209 = (164080 + ($$0378<<2)|0);
 $210 = ((($$1)) + 28|0);
 HEAP32[$210>>2] = $$0378;
 $211 = ((($$1)) + 16|0);
 $212 = ((($$1)) + 20|0);
 HEAP32[$212>>2] = 0;
 HEAP32[$211>>2] = 0;
 $213 = HEAP32[(163780)>>2]|0;
 $214 = 1 << $$0378;
 $215 = $213 & $214;
 $216 = ($215|0)==(0);
 if ($216) {
  $217 = $213 | $214;
  HEAP32[(163780)>>2] = $217;
  HEAP32[$209>>2] = $$1;
  $218 = ((($$1)) + 24|0);
  HEAP32[$218>>2] = $209;
  $219 = ((($$1)) + 12|0);
  HEAP32[$219>>2] = $$1;
  $220 = ((($$1)) + 8|0);
  HEAP32[$220>>2] = $$1;
  return;
 }
 $221 = HEAP32[$209>>2]|0;
 $222 = ($$0378|0)==(31);
 $223 = $$0378 >>> 1;
 $224 = (25 - ($223))|0;
 $225 = $222 ? 0 : $224;
 $226 = $$2 << $225;
 $$0366 = $226;$$0367 = $221;
 while(1) {
  $227 = ((($$0367)) + 4|0);
  $228 = HEAP32[$227>>2]|0;
  $229 = $228 & -8;
  $230 = ($229|0)==($$2|0);
  if ($230) {
   label = 69;
   break;
  }
  $231 = $$0366 >>> 31;
  $232 = (((($$0367)) + 16|0) + ($231<<2)|0);
  $233 = $$0366 << 1;
  $234 = HEAP32[$232>>2]|0;
  $235 = ($234|0)==(0|0);
  if ($235) {
   label = 68;
   break;
  } else {
   $$0366 = $233;$$0367 = $234;
  }
 }
 if ((label|0) == 68) {
  HEAP32[$232>>2] = $$1;
  $236 = ((($$1)) + 24|0);
  HEAP32[$236>>2] = $$0367;
  $237 = ((($$1)) + 12|0);
  HEAP32[$237>>2] = $$1;
  $238 = ((($$1)) + 8|0);
  HEAP32[$238>>2] = $$1;
  return;
 }
 else if ((label|0) == 69) {
  $239 = ((($$0367)) + 8|0);
  $240 = HEAP32[$239>>2]|0;
  $241 = ((($240)) + 12|0);
  HEAP32[$241>>2] = $$1;
  HEAP32[$239>>2] = $$1;
  $242 = ((($$1)) + 8|0);
  HEAP32[$242>>2] = $240;
  $243 = ((($$1)) + 12|0);
  HEAP32[$243>>2] = $$0367;
  $244 = ((($$1)) + 24|0);
  HEAP32[$244>>2] = 0;
  return;
 }
}
function _emscripten_get_global_libc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (164272|0);
}
function ___stdio_close($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $1 = ((($0)) + 60|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = (_dummy_738($2)|0);
 HEAP32[$vararg_buffer>>2] = $3;
 $4 = (___syscall6(6,($vararg_buffer|0))|0);
 $5 = (___syscall_ret($4)|0);
 STACKTOP = sp;return ($5|0);
}
function ___stdio_write($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$04756 = 0, $$04855 = 0, $$04954 = 0, $$051 = 0, $$1 = 0, $$150 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer3 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0;
 var $vararg_ptr7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $3 = sp + 32|0;
 $4 = ((($0)) + 28|0);
 $5 = HEAP32[$4>>2]|0;
 HEAP32[$3>>2] = $5;
 $6 = ((($3)) + 4|0);
 $7 = ((($0)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) - ($5))|0;
 HEAP32[$6>>2] = $9;
 $10 = ((($3)) + 8|0);
 HEAP32[$10>>2] = $1;
 $11 = ((($3)) + 12|0);
 HEAP32[$11>>2] = $2;
 $12 = (($9) + ($2))|0;
 $13 = ((($0)) + 60|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $3;
 HEAP32[$vararg_buffer>>2] = $14;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $15;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = 2;
 $16 = (___syscall146(146,($vararg_buffer|0))|0);
 $17 = (___syscall_ret($16)|0);
 $18 = ($12|0)==($17|0);
 L1: do {
  if ($18) {
   label = 3;
  } else {
   $$04756 = 2;$$04855 = $12;$$04954 = $3;$26 = $17;
   while(1) {
    $25 = ($26|0)<(0);
    if ($25) {
     break;
    }
    $34 = (($$04855) - ($26))|0;
    $35 = ((($$04954)) + 4|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($26>>>0)>($36>>>0);
    $38 = ((($$04954)) + 8|0);
    $$150 = $37 ? $38 : $$04954;
    $39 = $37 << 31 >> 31;
    $$1 = (($39) + ($$04756))|0;
    $40 = $37 ? $36 : 0;
    $$0 = (($26) - ($40))|0;
    $41 = HEAP32[$$150>>2]|0;
    $42 = (($41) + ($$0)|0);
    HEAP32[$$150>>2] = $42;
    $43 = ((($$150)) + 4|0);
    $44 = HEAP32[$43>>2]|0;
    $45 = (($44) - ($$0))|0;
    HEAP32[$43>>2] = $45;
    $46 = HEAP32[$13>>2]|0;
    $47 = $$150;
    HEAP32[$vararg_buffer3>>2] = $46;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = $47;
    $vararg_ptr7 = ((($vararg_buffer3)) + 8|0);
    HEAP32[$vararg_ptr7>>2] = $$1;
    $48 = (___syscall146(146,($vararg_buffer3|0))|0);
    $49 = (___syscall_ret($48)|0);
    $50 = ($34|0)==($49|0);
    if ($50) {
     label = 3;
     break L1;
    } else {
     $$04756 = $$1;$$04855 = $34;$$04954 = $$150;$26 = $49;
    }
   }
   $27 = ((($0)) + 16|0);
   HEAP32[$27>>2] = 0;
   HEAP32[$4>>2] = 0;
   HEAP32[$7>>2] = 0;
   $28 = HEAP32[$0>>2]|0;
   $29 = $28 | 32;
   HEAP32[$0>>2] = $29;
   $30 = ($$04756|0)==(2);
   if ($30) {
    $$051 = 0;
   } else {
    $31 = ((($$04954)) + 4|0);
    $32 = HEAP32[$31>>2]|0;
    $33 = (($2) - ($32))|0;
    $$051 = $33;
   }
  }
 } while(0);
 if ((label|0) == 3) {
  $19 = ((($0)) + 44|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($0)) + 48|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + ($22)|0);
  $24 = ((($0)) + 16|0);
  HEAP32[$24>>2] = $23;
  HEAP32[$4>>2] = $20;
  HEAP32[$7>>2] = $20;
  $$051 = $2;
 }
 STACKTOP = sp;return ($$051|0);
}
function ___stdio_seek($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$pre = 0, $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $3 = sp + 20|0;
 $4 = ((($0)) + 60|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $3;
 HEAP32[$vararg_buffer>>2] = $5;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = 0;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $1;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $6;
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
 HEAP32[$vararg_ptr4>>2] = $2;
 $7 = (___syscall140(140,($vararg_buffer|0))|0);
 $8 = (___syscall_ret($7)|0);
 $9 = ($8|0)<(0);
 if ($9) {
  HEAP32[$3>>2] = -1;
  $10 = -1;
 } else {
  $$pre = HEAP32[$3>>2]|0;
  $10 = $$pre;
 }
 STACKTOP = sp;return ($10|0);
}
function ___syscall_ret($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0>>>0)>(4294963200);
 if ($1) {
  $2 = (0 - ($0))|0;
  $3 = (___errno_location()|0);
  HEAP32[$3>>2] = $2;
  $$0 = -1;
 } else {
  $$0 = $0;
 }
 return ($$0|0);
}
function ___errno_location() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___pthread_self_108()|0);
 $1 = ((($0)) + 64|0);
 return ($1|0);
}
function ___pthread_self_108() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function _pthread_self() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (14756|0);
}
function _dummy_738($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($0|0);
}
function ___stdio_read($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $3 = sp + 16|0;
 HEAP32[$3>>2] = $1;
 $4 = ((($3)) + 4|0);
 $5 = ((($0)) + 48|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ($6|0)!=(0);
 $8 = $7&1;
 $9 = (($2) - ($8))|0;
 HEAP32[$4>>2] = $9;
 $10 = ((($3)) + 8|0);
 $11 = ((($0)) + 44|0);
 $12 = HEAP32[$11>>2]|0;
 HEAP32[$10>>2] = $12;
 $13 = ((($3)) + 12|0);
 HEAP32[$13>>2] = $6;
 $14 = ((($0)) + 60|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $3;
 HEAP32[$vararg_buffer>>2] = $15;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $16;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = 2;
 $17 = (___syscall145(145,($vararg_buffer|0))|0);
 $18 = (___syscall_ret($17)|0);
 $19 = ($18|0)<(1);
 if ($19) {
  $20 = $18 & 48;
  $21 = $20 ^ 16;
  $22 = HEAP32[$0>>2]|0;
  $23 = $22 | $21;
  HEAP32[$0>>2] = $23;
  $$0 = $18;
 } else {
  $24 = HEAP32[$4>>2]|0;
  $25 = ($18>>>0)>($24>>>0);
  if ($25) {
   $26 = (($18) - ($24))|0;
   $27 = HEAP32[$11>>2]|0;
   $28 = ((($0)) + 4|0);
   HEAP32[$28>>2] = $27;
   $29 = (($27) + ($26)|0);
   $30 = ((($0)) + 8|0);
   HEAP32[$30>>2] = $29;
   $31 = HEAP32[$5>>2]|0;
   $32 = ($31|0)==(0);
   if ($32) {
    $$0 = $2;
   } else {
    $33 = ((($27)) + 1|0);
    HEAP32[$28>>2] = $33;
    $34 = HEAP8[$27>>0]|0;
    $35 = (($2) + -1)|0;
    $36 = (($1) + ($35)|0);
    HEAP8[$36>>0] = $34;
    $$0 = $2;
   }
  } else {
   $$0 = $18;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function ___stdout_write($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $3 = sp + 16|0;
 $4 = ((($0)) + 36|0);
 HEAP32[$4>>2] = 369;
 $5 = HEAP32[$0>>2]|0;
 $6 = $5 & 64;
 $7 = ($6|0)==(0);
 if ($7) {
  $8 = ((($0)) + 60|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $3;
  HEAP32[$vararg_buffer>>2] = $9;
  $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
  HEAP32[$vararg_ptr1>>2] = 21523;
  $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
  HEAP32[$vararg_ptr2>>2] = $10;
  $11 = (___syscall54(54,($vararg_buffer|0))|0);
  $12 = ($11|0)==(0);
  if (!($12)) {
   $13 = ((($0)) + 75|0);
   HEAP8[$13>>0] = -1;
  }
 }
 $14 = (___stdio_write($0,$1,$2)|0);
 STACKTOP = sp;return ($14|0);
}
function _isupper($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (($0) + -65)|0;
 $2 = ($1>>>0)<(26);
 $3 = $2&1;
 return ($3|0);
}
function _strtoull_l($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = (_strtoull($0,$1,$2)|0);
 $5 = tempRet0;
 tempRet0 = ($5);
 return ($4|0);
}
function _strtoull($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (_strtox_768($0,$1,$2,-1,-1)|0);
 $4 = tempRet0;
 tempRet0 = ($4);
 return ($3|0);
}
function _strtox_768($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$sink = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $5 = sp;
 HEAP32[$5>>2] = 0;
 $6 = ((($5)) + 4|0);
 HEAP32[$6>>2] = $0;
 $7 = ((($5)) + 44|0);
 HEAP32[$7>>2] = $0;
 $8 = ($0|0)<(0|0);
 $9 = ((($0)) + 2147483647|0);
 $$sink = $8 ? (-1) : $9;
 $10 = ((($5)) + 8|0);
 HEAP32[$10>>2] = $$sink;
 $11 = ((($5)) + 76|0);
 HEAP32[$11>>2] = -1;
 ___shlim($5,0);
 $12 = (___intscan($5,$2,1,$3,$4)|0);
 $13 = tempRet0;
 $14 = ($1|0)==(0|0);
 if (!($14)) {
  $15 = ((($5)) + 108|0);
  $16 = HEAP32[$15>>2]|0;
  $17 = HEAP32[$6>>2]|0;
  $18 = HEAP32[$10>>2]|0;
  $19 = (($17) + ($16))|0;
  $20 = (($19) - ($18))|0;
  $21 = (($0) + ($20)|0);
  HEAP32[$1>>2] = $21;
 }
 tempRet0 = ($13);
 STACKTOP = sp;return ($12|0);
}
function ___shlim($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sink = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 104|0);
 HEAP32[$2>>2] = $1;
 $3 = ((($0)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($0)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $4;
 $8 = $6;
 $9 = (($7) - ($8))|0;
 $10 = ((($0)) + 108|0);
 HEAP32[$10>>2] = $9;
 $11 = ($1|0)!=(0);
 $12 = ($9|0)>($1|0);
 $or$cond = $11 & $12;
 $13 = (($6) + ($1)|0);
 $$sink = $or$cond ? $13 : $4;
 $14 = ((($0)) + 100|0);
 HEAP32[$14>>2] = $$sink;
 return;
}
function ___intscan($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0154222 = 0, $$0157 = 0, $$0157$ = 0, $$0159 = 0, $$1155192 = 0, $$1158 = 0, $$1160 = 0, $$1160169 = 0, $$1165 = 0, $$1165167 = 0, $$1165168 = 0, $$166 = 0, $$2156210 = 0, $$2161$be = 0, $$2161$lcssa = 0, $$3162$be = 0, $$3162215 = 0, $$4163$be = 0, $$4163$lcssa = 0, $$5$be = 0;
 var $$6$be = 0, $$6$lcssa = 0, $$7$be = 0, $$7198 = 0, $$8 = 0, $$9$be = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0;
 var $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0;
 var $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0;
 var $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0;
 var $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0;
 var $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0;
 var $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0;
 var $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0;
 var $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0;
 var $294 = 0, $295 = 0, $296 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0;
 var $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond12 = 0;
 var $or$cond187 = 0, $or$cond5 = 0, $or$cond7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ($1>>>0)>(36);
 L1: do {
  if ($5) {
   $8 = (___errno_location()|0);
   HEAP32[$8>>2] = 22;
   $289 = 0;$290 = 0;
  } else {
   $6 = ((($0)) + 4|0);
   $7 = ((($0)) + 100|0);
   while(1) {
    $9 = HEAP32[$6>>2]|0;
    $10 = HEAP32[$7>>2]|0;
    $11 = ($9>>>0)<($10>>>0);
    if ($11) {
     $12 = ((($9)) + 1|0);
     HEAP32[$6>>2] = $12;
     $13 = HEAP8[$9>>0]|0;
     $14 = $13&255;
     $16 = $14;
    } else {
     $15 = (___shgetc($0)|0);
     $16 = $15;
    }
    $17 = (_isspace($16)|0);
    $18 = ($17|0)==(0);
    if ($18) {
     break;
    }
   }
   L11: do {
    switch ($16|0) {
    case 43: case 45:  {
     $19 = ($16|0)==(45);
     $20 = $19 << 31 >> 31;
     $21 = HEAP32[$6>>2]|0;
     $22 = HEAP32[$7>>2]|0;
     $23 = ($21>>>0)<($22>>>0);
     if ($23) {
      $24 = ((($21)) + 1|0);
      HEAP32[$6>>2] = $24;
      $25 = HEAP8[$21>>0]|0;
      $26 = $25&255;
      $$0157 = $20;$$0159 = $26;
      break L11;
     } else {
      $27 = (___shgetc($0)|0);
      $$0157 = $20;$$0159 = $27;
      break L11;
     }
     break;
    }
    default: {
     $$0157 = 0;$$0159 = $16;
    }
    }
   } while(0);
   $28 = ($1|0)==(0);
   $29 = $1 | 16;
   $30 = ($29|0)==(16);
   $31 = ($$0159|0)==(48);
   $or$cond5 = $30 & $31;
   do {
    if ($or$cond5) {
     $32 = HEAP32[$6>>2]|0;
     $33 = HEAP32[$7>>2]|0;
     $34 = ($32>>>0)<($33>>>0);
     if ($34) {
      $35 = ((($32)) + 1|0);
      HEAP32[$6>>2] = $35;
      $36 = HEAP8[$32>>0]|0;
      $37 = $36&255;
      $40 = $37;
     } else {
      $38 = (___shgetc($0)|0);
      $40 = $38;
     }
     $39 = $40 | 32;
     $41 = ($39|0)==(120);
     if (!($41)) {
      if ($28) {
       $$1160169 = $40;$$1165168 = 8;
       label = 46;
       break;
      } else {
       $$1160 = $40;$$1165 = $1;
       label = 32;
       break;
      }
     }
     $42 = HEAP32[$6>>2]|0;
     $43 = HEAP32[$7>>2]|0;
     $44 = ($42>>>0)<($43>>>0);
     if ($44) {
      $45 = ((($42)) + 1|0);
      HEAP32[$6>>2] = $45;
      $46 = HEAP8[$42>>0]|0;
      $47 = $46&255;
      $50 = $47;
     } else {
      $48 = (___shgetc($0)|0);
      $50 = $48;
     }
     $49 = ((147919) + ($50)|0);
     $51 = HEAP8[$49>>0]|0;
     $52 = ($51&255)>(15);
     if ($52) {
      $53 = HEAP32[$7>>2]|0;
      $54 = ($53|0)!=(0|0);
      if ($54) {
       $55 = HEAP32[$6>>2]|0;
       $56 = ((($55)) + -1|0);
       HEAP32[$6>>2] = $56;
      }
      $57 = ($2|0)==(0);
      if ($57) {
       ___shlim($0,0);
       $289 = 0;$290 = 0;
       break L1;
      }
      if (!($54)) {
       $289 = 0;$290 = 0;
       break L1;
      }
      $58 = HEAP32[$6>>2]|0;
      $59 = ((($58)) + -1|0);
      HEAP32[$6>>2] = $59;
      $289 = 0;$290 = 0;
      break L1;
     } else {
      $$1160169 = $50;$$1165168 = 16;
      label = 46;
     }
    } else {
     $$166 = $28 ? 10 : $1;
     $60 = ((147919) + ($$0159)|0);
     $61 = HEAP8[$60>>0]|0;
     $62 = $61&255;
     $63 = ($62>>>0)<($$166>>>0);
     if ($63) {
      $$1160 = $$0159;$$1165 = $$166;
      label = 32;
     } else {
      $64 = HEAP32[$7>>2]|0;
      $65 = ($64|0)==(0|0);
      if (!($65)) {
       $66 = HEAP32[$6>>2]|0;
       $67 = ((($66)) + -1|0);
       HEAP32[$6>>2] = $67;
      }
      ___shlim($0,0);
      $68 = (___errno_location()|0);
      HEAP32[$68>>2] = 22;
      $289 = 0;$290 = 0;
      break L1;
     }
    }
   } while(0);
   L43: do {
    if ((label|0) == 32) {
     $69 = ($$1165|0)==(10);
     if ($69) {
      $70 = (($$1160) + -48)|0;
      $71 = ($70>>>0)<(10);
      if ($71) {
       $$0154222 = 0;$74 = $70;
       while(1) {
        $72 = ($$0154222*10)|0;
        $73 = (($72) + ($74))|0;
        $75 = HEAP32[$6>>2]|0;
        $76 = HEAP32[$7>>2]|0;
        $77 = ($75>>>0)<($76>>>0);
        if ($77) {
         $78 = ((($75)) + 1|0);
         HEAP32[$6>>2] = $78;
         $79 = HEAP8[$75>>0]|0;
         $80 = $79&255;
         $$2161$be = $80;
        } else {
         $81 = (___shgetc($0)|0);
         $$2161$be = $81;
        }
        $82 = (($$2161$be) + -48)|0;
        $83 = ($82>>>0)<(10);
        $84 = ($73>>>0)<(429496729);
        $85 = $83 & $84;
        if ($85) {
         $$0154222 = $73;$74 = $82;
        } else {
         break;
        }
       }
       $$2161$lcssa = $$2161$be;$291 = $73;$292 = 0;
      } else {
       $$2161$lcssa = $$1160;$291 = 0;$292 = 0;
      }
      $86 = (($$2161$lcssa) + -48)|0;
      $87 = ($86>>>0)<(10);
      if ($87) {
       $$3162215 = $$2161$lcssa;$88 = $291;$89 = $292;$93 = $86;
       while(1) {
        $90 = (___muldi3(($88|0),($89|0),10,0)|0);
        $91 = tempRet0;
        $92 = ($93|0)<(0);
        $94 = $92 << 31 >> 31;
        $95 = $93 ^ -1;
        $96 = $94 ^ -1;
        $97 = ($91>>>0)>($96>>>0);
        $98 = ($90>>>0)>($95>>>0);
        $99 = ($91|0)==($96|0);
        $100 = $99 & $98;
        $101 = $97 | $100;
        if ($101) {
         $$1165167 = 10;$$8 = $$3162215;$293 = $88;$294 = $89;
         label = 72;
         break L43;
        }
        $102 = (_i64Add(($90|0),($91|0),($93|0),($94|0))|0);
        $103 = tempRet0;
        $104 = HEAP32[$6>>2]|0;
        $105 = HEAP32[$7>>2]|0;
        $106 = ($104>>>0)<($105>>>0);
        if ($106) {
         $107 = ((($104)) + 1|0);
         HEAP32[$6>>2] = $107;
         $108 = HEAP8[$104>>0]|0;
         $109 = $108&255;
         $$3162$be = $109;
        } else {
         $110 = (___shgetc($0)|0);
         $$3162$be = $110;
        }
        $111 = (($$3162$be) + -48)|0;
        $112 = ($111>>>0)<(10);
        $113 = ($103>>>0)<(429496729);
        $114 = ($102>>>0)<(2576980378);
        $115 = ($103|0)==(429496729);
        $116 = $115 & $114;
        $117 = $113 | $116;
        $or$cond7 = $112 & $117;
        if ($or$cond7) {
         $$3162215 = $$3162$be;$88 = $102;$89 = $103;$93 = $111;
        } else {
         break;
        }
       }
       $118 = ($111>>>0)>(9);
       if ($118) {
        $$1158 = $$0157;$263 = $103;$265 = $102;
       } else {
        $$1165167 = 10;$$8 = $$3162$be;$293 = $102;$294 = $103;
        label = 72;
       }
      } else {
       $$1158 = $$0157;$263 = $292;$265 = $291;
      }
     } else {
      $$1160169 = $$1160;$$1165168 = $$1165;
      label = 46;
     }
    }
   } while(0);
   L63: do {
    if ((label|0) == 46) {
     $119 = (($$1165168) + -1)|0;
     $120 = $119 & $$1165168;
     $121 = ($120|0)==(0);
     if ($121) {
      $126 = ($$1165168*23)|0;
      $127 = $126 >>> 5;
      $128 = $127 & 7;
      $129 = (148175 + ($128)|0);
      $130 = HEAP8[$129>>0]|0;
      $131 = $130 << 24 >> 24;
      $132 = ((147919) + ($$1160169)|0);
      $133 = HEAP8[$132>>0]|0;
      $134 = $133&255;
      $135 = ($134>>>0)<($$1165168>>>0);
      if ($135) {
       $$1155192 = 0;$138 = $134;
       while(1) {
        $136 = $$1155192 << $131;
        $137 = $138 | $136;
        $139 = HEAP32[$6>>2]|0;
        $140 = HEAP32[$7>>2]|0;
        $141 = ($139>>>0)<($140>>>0);
        if ($141) {
         $142 = ((($139)) + 1|0);
         HEAP32[$6>>2] = $142;
         $143 = HEAP8[$139>>0]|0;
         $144 = $143&255;
         $$4163$be = $144;
        } else {
         $145 = (___shgetc($0)|0);
         $$4163$be = $145;
        }
        $146 = ((147919) + ($$4163$be)|0);
        $147 = HEAP8[$146>>0]|0;
        $148 = $147&255;
        $149 = ($148>>>0)<($$1165168>>>0);
        $150 = ($137>>>0)<(134217728);
        $151 = $150 & $149;
        if ($151) {
         $$1155192 = $137;$138 = $148;
        } else {
         break;
        }
       }
       $$4163$lcssa = $$4163$be;$155 = $147;$158 = 0;$160 = $137;
      } else {
       $$4163$lcssa = $$1160169;$155 = $133;$158 = 0;$160 = 0;
      }
      $152 = (_bitshift64Lshr(-1,-1,($131|0))|0);
      $153 = tempRet0;
      $154 = $155&255;
      $156 = ($154>>>0)>=($$1165168>>>0);
      $157 = ($158>>>0)>($153>>>0);
      $159 = ($160>>>0)>($152>>>0);
      $161 = ($158|0)==($153|0);
      $162 = $161 & $159;
      $163 = $157 | $162;
      $or$cond187 = $156 | $163;
      if ($or$cond187) {
       $$1165167 = $$1165168;$$8 = $$4163$lcssa;$293 = $160;$294 = $158;
       label = 72;
       break;
      } else {
       $164 = $160;$165 = $158;$169 = $155;
      }
      while(1) {
       $166 = (_bitshift64Shl(($164|0),($165|0),($131|0))|0);
       $167 = tempRet0;
       $168 = $169&255;
       $170 = $168 | $166;
       $171 = HEAP32[$6>>2]|0;
       $172 = HEAP32[$7>>2]|0;
       $173 = ($171>>>0)<($172>>>0);
       if ($173) {
        $174 = ((($171)) + 1|0);
        HEAP32[$6>>2] = $174;
        $175 = HEAP8[$171>>0]|0;
        $176 = $175&255;
        $$5$be = $176;
       } else {
        $177 = (___shgetc($0)|0);
        $$5$be = $177;
       }
       $178 = ((147919) + ($$5$be)|0);
       $179 = HEAP8[$178>>0]|0;
       $180 = $179&255;
       $181 = ($180>>>0)>=($$1165168>>>0);
       $182 = ($167>>>0)>($153>>>0);
       $183 = ($170>>>0)>($152>>>0);
       $184 = ($167|0)==($153|0);
       $185 = $184 & $183;
       $186 = $182 | $185;
       $or$cond = $181 | $186;
       if ($or$cond) {
        $$1165167 = $$1165168;$$8 = $$5$be;$293 = $170;$294 = $167;
        label = 72;
        break L63;
       } else {
        $164 = $170;$165 = $167;$169 = $179;
       }
      }
     }
     $122 = ((147919) + ($$1160169)|0);
     $123 = HEAP8[$122>>0]|0;
     $124 = $123&255;
     $125 = ($124>>>0)<($$1165168>>>0);
     if ($125) {
      $$2156210 = 0;$189 = $124;
      while(1) {
       $187 = Math_imul($$2156210, $$1165168)|0;
       $188 = (($189) + ($187))|0;
       $190 = HEAP32[$6>>2]|0;
       $191 = HEAP32[$7>>2]|0;
       $192 = ($190>>>0)<($191>>>0);
       if ($192) {
        $193 = ((($190)) + 1|0);
        HEAP32[$6>>2] = $193;
        $194 = HEAP8[$190>>0]|0;
        $195 = $194&255;
        $$6$be = $195;
       } else {
        $196 = (___shgetc($0)|0);
        $$6$be = $196;
       }
       $197 = ((147919) + ($$6$be)|0);
       $198 = HEAP8[$197>>0]|0;
       $199 = $198&255;
       $200 = ($199>>>0)<($$1165168>>>0);
       $201 = ($188>>>0)<(119304647);
       $202 = $201 & $200;
       if ($202) {
        $$2156210 = $188;$189 = $199;
       } else {
        break;
       }
      }
      $$6$lcssa = $$6$be;$204 = $198;$295 = $188;$296 = 0;
     } else {
      $$6$lcssa = $$1160169;$204 = $123;$295 = 0;$296 = 0;
     }
     $203 = $204&255;
     $205 = ($203>>>0)<($$1165168>>>0);
     if ($205) {
      $206 = (___udivdi3(-1,-1,($$1165168|0),0)|0);
      $207 = tempRet0;
      $$7198 = $$6$lcssa;$209 = $296;$211 = $295;$218 = $204;
      while(1) {
       $208 = ($209>>>0)>($207>>>0);
       $210 = ($211>>>0)>($206>>>0);
       $212 = ($209|0)==($207|0);
       $213 = $212 & $210;
       $214 = $208 | $213;
       if ($214) {
        $$1165167 = $$1165168;$$8 = $$7198;$293 = $211;$294 = $209;
        label = 72;
        break L63;
       }
       $215 = (___muldi3(($211|0),($209|0),($$1165168|0),0)|0);
       $216 = tempRet0;
       $217 = $218&255;
       $219 = $217 ^ -1;
       $220 = ($216>>>0)>(4294967295);
       $221 = ($215>>>0)>($219>>>0);
       $222 = ($216|0)==(-1);
       $223 = $222 & $221;
       $224 = $220 | $223;
       if ($224) {
        $$1165167 = $$1165168;$$8 = $$7198;$293 = $211;$294 = $209;
        label = 72;
        break L63;
       }
       $225 = (_i64Add(($217|0),0,($215|0),($216|0))|0);
       $226 = tempRet0;
       $227 = HEAP32[$6>>2]|0;
       $228 = HEAP32[$7>>2]|0;
       $229 = ($227>>>0)<($228>>>0);
       if ($229) {
        $230 = ((($227)) + 1|0);
        HEAP32[$6>>2] = $230;
        $231 = HEAP8[$227>>0]|0;
        $232 = $231&255;
        $$7$be = $232;
       } else {
        $233 = (___shgetc($0)|0);
        $$7$be = $233;
       }
       $234 = ((147919) + ($$7$be)|0);
       $235 = HEAP8[$234>>0]|0;
       $236 = $235&255;
       $237 = ($236>>>0)<($$1165168>>>0);
       if ($237) {
        $$7198 = $$7$be;$209 = $226;$211 = $225;$218 = $235;
       } else {
        $$1165167 = $$1165168;$$8 = $$7$be;$293 = $225;$294 = $226;
        label = 72;
        break;
       }
      }
     } else {
      $$1165167 = $$1165168;$$8 = $$6$lcssa;$293 = $295;$294 = $296;
      label = 72;
     }
    }
   } while(0);
   if ((label|0) == 72) {
    $238 = ((147919) + ($$8)|0);
    $239 = HEAP8[$238>>0]|0;
    $240 = $239&255;
    $241 = ($240>>>0)<($$1165167>>>0);
    if ($241) {
     while(1) {
      $242 = HEAP32[$6>>2]|0;
      $243 = HEAP32[$7>>2]|0;
      $244 = ($242>>>0)<($243>>>0);
      if ($244) {
       $245 = ((($242)) + 1|0);
       HEAP32[$6>>2] = $245;
       $246 = HEAP8[$242>>0]|0;
       $247 = $246&255;
       $$9$be = $247;
      } else {
       $248 = (___shgetc($0)|0);
       $$9$be = $248;
      }
      $249 = ((147919) + ($$9$be)|0);
      $250 = HEAP8[$249>>0]|0;
      $251 = $250&255;
      $252 = ($251>>>0)<($$1165167>>>0);
      if (!($252)) {
       break;
      }
     }
     $253 = (___errno_location()|0);
     HEAP32[$253>>2] = 34;
     $254 = $3 & 1;
     $255 = ($254|0)==(0);
     $256 = (0)==(0);
     $257 = $255 & $256;
     $$0157$ = $257 ? $$0157 : 0;
     $$1158 = $$0157$;$263 = $4;$265 = $3;
    } else {
     $$1158 = $$0157;$263 = $294;$265 = $293;
    }
   }
   $258 = HEAP32[$7>>2]|0;
   $259 = ($258|0)==(0|0);
   if (!($259)) {
    $260 = HEAP32[$6>>2]|0;
    $261 = ((($260)) + -1|0);
    HEAP32[$6>>2] = $261;
   }
   $262 = ($263>>>0)<($4>>>0);
   $264 = ($265>>>0)<($3>>>0);
   $266 = ($263|0)==($4|0);
   $267 = $266 & $264;
   $268 = $262 | $267;
   if (!($268)) {
    $269 = $3 & 1;
    $270 = ($269|0)!=(0);
    $271 = (0)!=(0);
    $272 = $270 | $271;
    $273 = ($$1158|0)!=(0);
    $or$cond12 = $272 | $273;
    if (!($or$cond12)) {
     $274 = (___errno_location()|0);
     HEAP32[$274>>2] = 34;
     $275 = (_i64Add(($3|0),($4|0),-1,-1)|0);
     $276 = tempRet0;
     $289 = $276;$290 = $275;
     break;
    }
    $277 = ($263>>>0)>($4>>>0);
    $278 = ($265>>>0)>($3>>>0);
    $279 = ($263|0)==($4|0);
    $280 = $279 & $278;
    $281 = $277 | $280;
    if ($281) {
     $282 = (___errno_location()|0);
     HEAP32[$282>>2] = 34;
     $289 = $4;$290 = $3;
     break;
    }
   }
   $283 = ($$1158|0)<(0);
   $284 = $283 << 31 >> 31;
   $285 = $265 ^ $$1158;
   $286 = $263 ^ $284;
   $287 = (_i64Subtract(($285|0),($286|0),($$1158|0),($284|0))|0);
   $288 = tempRet0;
   $289 = $288;$290 = $287;
  }
 } while(0);
 tempRet0 = ($289);
 return ($290|0);
}
function ___shgetc($0) {
 $0 = $0|0;
 var $$0 = 0, $$phi$trans$insert = 0, $$phi$trans$insert28$phi$trans$insert = 0, $$pre = 0, $$pre$phi34Z2D = 0, $$pre29$pre = 0, $$pre35 = 0, $$sink = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 104|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  label = 3;
 } else {
  $4 = ((($0)) + 108|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = ($5|0)<($2|0);
  if ($6) {
   label = 3;
  } else {
   label = 4;
  }
 }
 if ((label|0) == 3) {
  $7 = (___uflow($0)|0);
  $8 = ($7|0)<(0);
  if ($8) {
   label = 4;
  } else {
   $10 = HEAP32[$1>>2]|0;
   $11 = ($10|0)==(0);
   $$phi$trans$insert = ((($0)) + 8|0);
   if ($11) {
    $$pre = HEAP32[$$phi$trans$insert>>2]|0;
    $$phi$trans$insert28$phi$trans$insert = ((($0)) + 4|0);
    $$pre29$pre = HEAP32[$$phi$trans$insert28$phi$trans$insert>>2]|0;
    $$pre35 = ((($0)) + 108|0);
    $$pre$phi34Z2D = $$pre35;$$sink = $$pre;$26 = $$pre;$29 = $$pre29$pre;
   } else {
    $12 = HEAP32[$$phi$trans$insert>>2]|0;
    $13 = ((($0)) + 4|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = $14;
    $16 = (($12) - ($15))|0;
    $17 = ((($0)) + 108|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = (($10) - ($18))|0;
    $20 = ($16|0)<($19|0);
    $21 = $12;
    if ($20) {
     $$pre$phi34Z2D = $17;$$sink = $21;$26 = $21;$29 = $14;
    } else {
     $22 = (($19) + -1)|0;
     $23 = (($14) + ($22)|0);
     $$pre$phi34Z2D = $17;$$sink = $23;$26 = $21;$29 = $14;
    }
   }
   $24 = ((($0)) + 100|0);
   HEAP32[$24>>2] = $$sink;
   $25 = ($26|0)==(0|0);
   if (!($25)) {
    $27 = $26;
    $28 = $29;
    $30 = HEAP32[$$pre$phi34Z2D>>2]|0;
    $31 = (($27) + 1)|0;
    $32 = (($31) - ($28))|0;
    $33 = (($32) + ($30))|0;
    HEAP32[$$pre$phi34Z2D>>2] = $33;
   }
   $34 = ((($29)) + -1|0);
   $35 = HEAP8[$34>>0]|0;
   $36 = $35&255;
   $37 = ($36|0)==($7|0);
   if ($37) {
    $$0 = $7;
   } else {
    $38 = $7&255;
    HEAP8[$34>>0] = $38;
    $$0 = $7;
   }
  }
 }
 if ((label|0) == 4) {
  $9 = ((($0)) + 100|0);
  HEAP32[$9>>2] = 0;
  $$0 = -1;
 }
 return ($$0|0);
}
function _isspace($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(32);
 $2 = (($0) + -9)|0;
 $3 = ($2>>>0)<(5);
 $4 = $1 | $3;
 $5 = $4&1;
 return ($5|0);
}
function ___uflow($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = (___toread($0)|0);
 $3 = ($2|0)==(0);
 if ($3) {
  $4 = ((($0)) + 32|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (FUNCTION_TABLE_iiii[$5 & 2047]($0,$1,1)|0);
  $7 = ($6|0)==(1);
  if ($7) {
   $8 = HEAP8[$1>>0]|0;
   $9 = $8&255;
   $$0 = $9;
  } else {
   $$0 = -1;
  }
 } else {
  $$0 = -1;
 }
 STACKTOP = sp;return ($$0|0);
}
function ___toread($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $sext = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 74|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2 << 24 >> 24;
 $4 = (($3) + 255)|0;
 $5 = $4 | $3;
 $6 = $5&255;
 HEAP8[$1>>0] = $6;
 $7 = ((($0)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($0)) + 28|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8>>>0)>($10>>>0);
 if ($11) {
  $12 = ((($0)) + 36|0);
  $13 = HEAP32[$12>>2]|0;
  (FUNCTION_TABLE_iiii[$13 & 2047]($0,0,0)|0);
 }
 $14 = ((($0)) + 16|0);
 HEAP32[$14>>2] = 0;
 HEAP32[$9>>2] = 0;
 HEAP32[$7>>2] = 0;
 $15 = HEAP32[$0>>2]|0;
 $16 = $15 & 4;
 $17 = ($16|0)==(0);
 if ($17) {
  $19 = ((($0)) + 44|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ((($0)) + 48|0);
  $22 = HEAP32[$21>>2]|0;
  $23 = (($20) + ($22)|0);
  $24 = ((($0)) + 8|0);
  HEAP32[$24>>2] = $23;
  $25 = ((($0)) + 4|0);
  HEAP32[$25>>2] = $23;
  $26 = $15 << 27;
  $sext = $26 >> 31;
  $$0 = $sext;
 } else {
  $18 = $15 | 32;
  HEAP32[$0>>2] = $18;
  $$0 = -1;
 }
 return ($$0|0);
}
function _strtoll_l($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = (_strtoll($0,$1,$2)|0);
 $5 = tempRet0;
 tempRet0 = ($5);
 return ($4|0);
}
function _strtoll($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (_strtox_768($0,$1,$2,0,-2147483648)|0);
 $4 = tempRet0;
 tempRet0 = ($4);
 return ($3|0);
}
function _copysign($0,$1) {
 $0 = +$0;
 $1 = +$1;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $0;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 HEAPF64[tempDoublePtr>>3] = $1;$4 = HEAP32[tempDoublePtr>>2]|0;
 $5 = HEAP32[tempDoublePtr+4>>2]|0;
 $6 = $3 & 2147483647;
 $7 = $5 & -2147483648;
 $8 = $7 | $6;
 HEAP32[tempDoublePtr>>2] = $2;HEAP32[tempDoublePtr+4>>2] = $8;$9 = +HEAPF64[tempDoublePtr>>3];
 return (+$9);
}
function _strcmp($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$011 = 0, $$0710 = 0, $$lcssa = 0, $$lcssa8 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $2 = HEAP8[$0>>0]|0;
 $3 = HEAP8[$1>>0]|0;
 $4 = ($2<<24>>24)!=($3<<24>>24);
 $5 = ($2<<24>>24)==(0);
 $or$cond9 = $5 | $4;
 if ($or$cond9) {
  $$lcssa = $3;$$lcssa8 = $2;
 } else {
  $$011 = $1;$$0710 = $0;
  while(1) {
   $6 = ((($$0710)) + 1|0);
   $7 = ((($$011)) + 1|0);
   $8 = HEAP8[$6>>0]|0;
   $9 = HEAP8[$7>>0]|0;
   $10 = ($8<<24>>24)!=($9<<24>>24);
   $11 = ($8<<24>>24)==(0);
   $or$cond = $11 | $10;
   if ($or$cond) {
    $$lcssa = $9;$$lcssa8 = $8;
    break;
   } else {
    $$011 = $7;$$0710 = $6;
   }
  }
 }
 $12 = $$lcssa8&255;
 $13 = $$lcssa&255;
 $14 = (($12) - ($13))|0;
 return ($14|0);
}
function _memcmp($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$01318 = 0, $$01417 = 0, $$019 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $14 = 0;
  } else {
   $$01318 = $0;$$01417 = $2;$$019 = $1;
   while(1) {
    $4 = HEAP8[$$01318>>0]|0;
    $5 = HEAP8[$$019>>0]|0;
    $6 = ($4<<24>>24)==($5<<24>>24);
    if (!($6)) {
     break;
    }
    $7 = (($$01417) + -1)|0;
    $8 = ((($$01318)) + 1|0);
    $9 = ((($$019)) + 1|0);
    $10 = ($7|0)==(0);
    if ($10) {
     $14 = 0;
     break L1;
    } else {
     $$01318 = $8;$$01417 = $7;$$019 = $9;
    }
   }
   $11 = $4&255;
   $12 = $5&255;
   $13 = (($11) - ($12))|0;
   $14 = $13;
  }
 } while(0);
 return ($14|0);
}
function _vsnprintf($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$$015 = 0, $$0 = 0, $$014 = 0, $$015 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, dest = 0, label = 0, sp = 0, src = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $4 = sp + 124|0;
 $5 = sp;
 dest=$5; src=15260; stop=dest+124|0; do { HEAP32[dest>>2]=HEAP32[src>>2]|0; dest=dest+4|0; src=src+4|0; } while ((dest|0) < (stop|0));
 $6 = (($1) + -1)|0;
 $7 = ($6>>>0)>(2147483646);
 if ($7) {
  $8 = ($1|0)==(0);
  if ($8) {
   $$014 = $4;$$015 = 1;
   label = 4;
  } else {
   $9 = (___errno_location()|0);
   HEAP32[$9>>2] = 75;
   $$0 = -1;
  }
 } else {
  $$014 = $0;$$015 = $1;
  label = 4;
 }
 if ((label|0) == 4) {
  $10 = $$014;
  $11 = (-2 - ($10))|0;
  $12 = ($$015>>>0)>($11>>>0);
  $$$015 = $12 ? $11 : $$015;
  $13 = ((($5)) + 48|0);
  HEAP32[$13>>2] = $$$015;
  $14 = ((($5)) + 20|0);
  HEAP32[$14>>2] = $$014;
  $15 = ((($5)) + 44|0);
  HEAP32[$15>>2] = $$014;
  $16 = (($$014) + ($$$015)|0);
  $17 = ((($5)) + 16|0);
  HEAP32[$17>>2] = $16;
  $18 = ((($5)) + 28|0);
  HEAP32[$18>>2] = $16;
  $19 = (_vfprintf($5,$2,$3)|0);
  $20 = ($$$015|0)==(0);
  if ($20) {
   $$0 = $19;
  } else {
   $21 = HEAP32[$14>>2]|0;
   $22 = HEAP32[$17>>2]|0;
   $23 = ($21|0)==($22|0);
   $24 = $23 << 31 >> 31;
   $25 = (($21) + ($24)|0);
   HEAP8[$25>>0] = 0;
   $$0 = $19;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _vfprintf($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0 = 0, $$1 = 0, $$1$ = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $vacopy_currentptr = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $3 = sp + 120|0;
 $4 = sp + 80|0;
 $5 = sp;
 $6 = sp + 136|0;
 dest=$4; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $vacopy_currentptr = HEAP32[$2>>2]|0;
 HEAP32[$3>>2] = $vacopy_currentptr;
 $7 = (_printf_core(0,$1,$3,$5,$4)|0);
 $8 = ($7|0)<(0);
 if ($8) {
  $$0 = -1;
 } else {
  $9 = ((($0)) + 76|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = ($10|0)>(-1);
  if ($11) {
   $12 = (___lockfile($0)|0);
   $40 = $12;
  } else {
   $40 = 0;
  }
  $13 = HEAP32[$0>>2]|0;
  $14 = $13 & 32;
  $15 = ((($0)) + 74|0);
  $16 = HEAP8[$15>>0]|0;
  $17 = ($16<<24>>24)<(1);
  if ($17) {
   $18 = $13 & -33;
   HEAP32[$0>>2] = $18;
  }
  $19 = ((($0)) + 48|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = ($20|0)==(0);
  if ($21) {
   $23 = ((($0)) + 44|0);
   $24 = HEAP32[$23>>2]|0;
   HEAP32[$23>>2] = $6;
   $25 = ((($0)) + 28|0);
   HEAP32[$25>>2] = $6;
   $26 = ((($0)) + 20|0);
   HEAP32[$26>>2] = $6;
   HEAP32[$19>>2] = 80;
   $27 = ((($6)) + 80|0);
   $28 = ((($0)) + 16|0);
   HEAP32[$28>>2] = $27;
   $29 = (_printf_core($0,$1,$3,$5,$4)|0);
   $30 = ($24|0)==(0|0);
   if ($30) {
    $$1 = $29;
   } else {
    $31 = ((($0)) + 36|0);
    $32 = HEAP32[$31>>2]|0;
    (FUNCTION_TABLE_iiii[$32 & 2047]($0,0,0)|0);
    $33 = HEAP32[$26>>2]|0;
    $34 = ($33|0)==(0|0);
    $$ = $34 ? -1 : $29;
    HEAP32[$23>>2] = $24;
    HEAP32[$19>>2] = 0;
    HEAP32[$28>>2] = 0;
    HEAP32[$25>>2] = 0;
    HEAP32[$26>>2] = 0;
    $$1 = $$;
   }
  } else {
   $22 = (_printf_core($0,$1,$3,$5,$4)|0);
   $$1 = $22;
  }
  $35 = HEAP32[$0>>2]|0;
  $36 = $35 & 32;
  $37 = ($36|0)==(0);
  $$1$ = $37 ? $$1 : -1;
  $38 = $35 | $14;
  HEAP32[$0>>2] = $38;
  $39 = ($40|0)==(0);
  if (!($39)) {
   ___unlockfile($0);
  }
  $$0 = $$1$;
 }
 STACKTOP = sp;return ($$0|0);
}
function _printf_core($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$$ = 0, $$$0259 = 0, $$$0262 = 0, $$$0269 = 0, $$$4266 = 0, $$$5 = 0, $$0 = 0, $$0228 = 0, $$0228$ = 0, $$0229322 = 0, $$0232 = 0, $$0235 = 0, $$0237 = 0, $$0240$lcssa = 0, $$0240$lcssa357 = 0, $$0240321 = 0, $$0243 = 0, $$0247 = 0, $$0249$lcssa = 0;
 var $$0249306 = 0, $$0252 = 0, $$0253 = 0, $$0254 = 0, $$0254$$0254$ = 0, $$0259 = 0, $$0262$lcssa = 0, $$0262311 = 0, $$0269 = 0, $$0269$phi = 0, $$1 = 0, $$1230333 = 0, $$1233 = 0, $$1236 = 0, $$1238 = 0, $$1241332 = 0, $$1244320 = 0, $$1248 = 0, $$1250 = 0, $$1255 = 0;
 var $$1260 = 0, $$1263 = 0, $$1263$ = 0, $$1270 = 0, $$2 = 0, $$2234 = 0, $$2239 = 0, $$2242305 = 0, $$2245 = 0, $$2251 = 0, $$2256 = 0, $$2256$ = 0, $$2256$$$2256 = 0, $$2261 = 0, $$2271 = 0, $$284$ = 0, $$289 = 0, $$290 = 0, $$3257 = 0, $$3265 = 0;
 var $$3272 = 0, $$3303 = 0, $$377 = 0, $$4258355 = 0, $$4266 = 0, $$5 = 0, $$6268 = 0, $$lcssa295 = 0, $$pre = 0, $$pre346 = 0, $$pre347 = 0, $$pre347$pre = 0, $$pre349 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0;
 var $306 = 0.0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0;
 var $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $arglist_current = 0, $arglist_current2 = 0, $arglist_next = 0, $arglist_next3 = 0, $expanded = 0, $expanded10 = 0, $expanded11 = 0, $expanded13 = 0, $expanded14 = 0, $expanded15 = 0, $expanded4 = 0, $expanded6 = 0, $expanded7 = 0, $expanded8 = 0, $isdigit = 0, $isdigit275 = 0, $isdigit277 = 0, $isdigittmp = 0, $isdigittmp$ = 0, $isdigittmp274 = 0;
 var $isdigittmp276 = 0, $narrow = 0, $or$cond = 0, $or$cond281 = 0, $or$cond283 = 0, $or$cond286 = 0, $storemerge = 0, $storemerge273310 = 0, $storemerge278 = 0, $trunc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $5 = sp + 16|0;
 $6 = sp;
 $7 = sp + 24|0;
 $8 = sp + 8|0;
 $9 = sp + 20|0;
 HEAP32[$5>>2] = $1;
 $10 = ($0|0)!=(0|0);
 $11 = ((($7)) + 40|0);
 $12 = $11;
 $13 = ((($7)) + 39|0);
 $14 = ((($8)) + 4|0);
 $$0243 = 0;$$0247 = 0;$$0269 = 0;$21 = $1;
 L1: while(1) {
  $15 = ($$0247|0)>(-1);
  do {
   if ($15) {
    $16 = (2147483647 - ($$0247))|0;
    $17 = ($$0243|0)>($16|0);
    if ($17) {
     $18 = (___errno_location()|0);
     HEAP32[$18>>2] = 75;
     $$1248 = -1;
     break;
    } else {
     $19 = (($$0243) + ($$0247))|0;
     $$1248 = $19;
     break;
    }
   } else {
    $$1248 = $$0247;
   }
  } while(0);
  $20 = HEAP8[$21>>0]|0;
  $22 = ($20<<24>>24)==(0);
  if ($22) {
   label = 87;
   break;
  } else {
   $23 = $20;$25 = $21;
  }
  L9: while(1) {
   switch ($23<<24>>24) {
   case 37:  {
    $$0249306 = $25;$27 = $25;
    label = 9;
    break L9;
    break;
   }
   case 0:  {
    $$0249$lcssa = $25;$39 = $25;
    break L9;
    break;
   }
   default: {
   }
   }
   $24 = ((($25)) + 1|0);
   HEAP32[$5>>2] = $24;
   $$pre = HEAP8[$24>>0]|0;
   $23 = $$pre;$25 = $24;
  }
  L12: do {
   if ((label|0) == 9) {
    while(1) {
     label = 0;
     $26 = ((($27)) + 1|0);
     $28 = HEAP8[$26>>0]|0;
     $29 = ($28<<24>>24)==(37);
     if (!($29)) {
      $$0249$lcssa = $$0249306;$39 = $27;
      break L12;
     }
     $30 = ((($$0249306)) + 1|0);
     $31 = ((($27)) + 2|0);
     HEAP32[$5>>2] = $31;
     $32 = HEAP8[$31>>0]|0;
     $33 = ($32<<24>>24)==(37);
     if ($33) {
      $$0249306 = $30;$27 = $31;
      label = 9;
     } else {
      $$0249$lcssa = $30;$39 = $31;
      break;
     }
    }
   }
  } while(0);
  $34 = $$0249$lcssa;
  $35 = $21;
  $36 = (($34) - ($35))|0;
  if ($10) {
   _out($0,$21,$36);
  }
  $37 = ($36|0)==(0);
  if (!($37)) {
   $$0269$phi = $$0269;$$0243 = $36;$$0247 = $$1248;$21 = $39;$$0269 = $$0269$phi;
   continue;
  }
  $38 = ((($39)) + 1|0);
  $40 = HEAP8[$38>>0]|0;
  $41 = $40 << 24 >> 24;
  $isdigittmp = (($41) + -48)|0;
  $isdigit = ($isdigittmp>>>0)<(10);
  if ($isdigit) {
   $42 = ((($39)) + 2|0);
   $43 = HEAP8[$42>>0]|0;
   $44 = ($43<<24>>24)==(36);
   $45 = ((($39)) + 3|0);
   $$377 = $44 ? $45 : $38;
   $$$0269 = $44 ? 1 : $$0269;
   $isdigittmp$ = $44 ? $isdigittmp : -1;
   $$0253 = $isdigittmp$;$$1270 = $$$0269;$storemerge = $$377;
  } else {
   $$0253 = -1;$$1270 = $$0269;$storemerge = $38;
  }
  HEAP32[$5>>2] = $storemerge;
  $46 = HEAP8[$storemerge>>0]|0;
  $47 = $46 << 24 >> 24;
  $48 = (($47) + -32)|0;
  $49 = ($48>>>0)<(32);
  L24: do {
   if ($49) {
    $$0262311 = 0;$329 = $46;$51 = $48;$storemerge273310 = $storemerge;
    while(1) {
     $50 = 1 << $51;
     $52 = $50 & 75913;
     $53 = ($52|0)==(0);
     if ($53) {
      $$0262$lcssa = $$0262311;$$lcssa295 = $329;$62 = $storemerge273310;
      break L24;
     }
     $54 = $50 | $$0262311;
     $55 = ((($storemerge273310)) + 1|0);
     HEAP32[$5>>2] = $55;
     $56 = HEAP8[$55>>0]|0;
     $57 = $56 << 24 >> 24;
     $58 = (($57) + -32)|0;
     $59 = ($58>>>0)<(32);
     if ($59) {
      $$0262311 = $54;$329 = $56;$51 = $58;$storemerge273310 = $55;
     } else {
      $$0262$lcssa = $54;$$lcssa295 = $56;$62 = $55;
      break;
     }
    }
   } else {
    $$0262$lcssa = 0;$$lcssa295 = $46;$62 = $storemerge;
   }
  } while(0);
  $60 = ($$lcssa295<<24>>24)==(42);
  if ($60) {
   $61 = ((($62)) + 1|0);
   $63 = HEAP8[$61>>0]|0;
   $64 = $63 << 24 >> 24;
   $isdigittmp276 = (($64) + -48)|0;
   $isdigit277 = ($isdigittmp276>>>0)<(10);
   if ($isdigit277) {
    $65 = ((($62)) + 2|0);
    $66 = HEAP8[$65>>0]|0;
    $67 = ($66<<24>>24)==(36);
    if ($67) {
     $68 = (($4) + ($isdigittmp276<<2)|0);
     HEAP32[$68>>2] = 10;
     $69 = HEAP8[$61>>0]|0;
     $70 = $69 << 24 >> 24;
     $71 = (($70) + -48)|0;
     $72 = (($3) + ($71<<3)|0);
     $73 = $72;
     $74 = $73;
     $75 = HEAP32[$74>>2]|0;
     $76 = (($73) + 4)|0;
     $77 = $76;
     $78 = HEAP32[$77>>2]|0;
     $79 = ((($62)) + 3|0);
     $$0259 = $75;$$2271 = 1;$storemerge278 = $79;
    } else {
     label = 23;
    }
   } else {
    label = 23;
   }
   if ((label|0) == 23) {
    label = 0;
    $80 = ($$1270|0)==(0);
    if (!($80)) {
     $$0 = -1;
     break;
    }
    if ($10) {
     $arglist_current = HEAP32[$2>>2]|0;
     $81 = $arglist_current;
     $82 = ((0) + 4|0);
     $expanded4 = $82;
     $expanded = (($expanded4) - 1)|0;
     $83 = (($81) + ($expanded))|0;
     $84 = ((0) + 4|0);
     $expanded8 = $84;
     $expanded7 = (($expanded8) - 1)|0;
     $expanded6 = $expanded7 ^ -1;
     $85 = $83 & $expanded6;
     $86 = $85;
     $87 = HEAP32[$86>>2]|0;
     $arglist_next = ((($86)) + 4|0);
     HEAP32[$2>>2] = $arglist_next;
     $$0259 = $87;$$2271 = 0;$storemerge278 = $61;
    } else {
     $$0259 = 0;$$2271 = 0;$storemerge278 = $61;
    }
   }
   HEAP32[$5>>2] = $storemerge278;
   $88 = ($$0259|0)<(0);
   $89 = $$0262$lcssa | 8192;
   $90 = (0 - ($$0259))|0;
   $$$0262 = $88 ? $89 : $$0262$lcssa;
   $$$0259 = $88 ? $90 : $$0259;
   $$1260 = $$$0259;$$1263 = $$$0262;$$3272 = $$2271;$94 = $storemerge278;
  } else {
   $91 = (_getint($5)|0);
   $92 = ($91|0)<(0);
   if ($92) {
    $$0 = -1;
    break;
   }
   $$pre346 = HEAP32[$5>>2]|0;
   $$1260 = $91;$$1263 = $$0262$lcssa;$$3272 = $$1270;$94 = $$pre346;
  }
  $93 = HEAP8[$94>>0]|0;
  $95 = ($93<<24>>24)==(46);
  do {
   if ($95) {
    $96 = ((($94)) + 1|0);
    $97 = HEAP8[$96>>0]|0;
    $98 = ($97<<24>>24)==(42);
    if (!($98)) {
     $125 = ((($94)) + 1|0);
     HEAP32[$5>>2] = $125;
     $126 = (_getint($5)|0);
     $$pre347$pre = HEAP32[$5>>2]|0;
     $$0254 = $126;$$pre347 = $$pre347$pre;
     break;
    }
    $99 = ((($94)) + 2|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = $100 << 24 >> 24;
    $isdigittmp274 = (($101) + -48)|0;
    $isdigit275 = ($isdigittmp274>>>0)<(10);
    if ($isdigit275) {
     $102 = ((($94)) + 3|0);
     $103 = HEAP8[$102>>0]|0;
     $104 = ($103<<24>>24)==(36);
     if ($104) {
      $105 = (($4) + ($isdigittmp274<<2)|0);
      HEAP32[$105>>2] = 10;
      $106 = HEAP8[$99>>0]|0;
      $107 = $106 << 24 >> 24;
      $108 = (($107) + -48)|0;
      $109 = (($3) + ($108<<3)|0);
      $110 = $109;
      $111 = $110;
      $112 = HEAP32[$111>>2]|0;
      $113 = (($110) + 4)|0;
      $114 = $113;
      $115 = HEAP32[$114>>2]|0;
      $116 = ((($94)) + 4|0);
      HEAP32[$5>>2] = $116;
      $$0254 = $112;$$pre347 = $116;
      break;
     }
    }
    $117 = ($$3272|0)==(0);
    if (!($117)) {
     $$0 = -1;
     break L1;
    }
    if ($10) {
     $arglist_current2 = HEAP32[$2>>2]|0;
     $118 = $arglist_current2;
     $119 = ((0) + 4|0);
     $expanded11 = $119;
     $expanded10 = (($expanded11) - 1)|0;
     $120 = (($118) + ($expanded10))|0;
     $121 = ((0) + 4|0);
     $expanded15 = $121;
     $expanded14 = (($expanded15) - 1)|0;
     $expanded13 = $expanded14 ^ -1;
     $122 = $120 & $expanded13;
     $123 = $122;
     $124 = HEAP32[$123>>2]|0;
     $arglist_next3 = ((($123)) + 4|0);
     HEAP32[$2>>2] = $arglist_next3;
     $330 = $124;
    } else {
     $330 = 0;
    }
    HEAP32[$5>>2] = $99;
    $$0254 = $330;$$pre347 = $99;
   } else {
    $$0254 = -1;$$pre347 = $94;
   }
  } while(0);
  $$0252 = 0;$128 = $$pre347;
  while(1) {
   $127 = HEAP8[$128>>0]|0;
   $129 = $127 << 24 >> 24;
   $130 = (($129) + -65)|0;
   $131 = ($130>>>0)>(57);
   if ($131) {
    $$0 = -1;
    break L1;
   }
   $132 = ((($128)) + 1|0);
   HEAP32[$5>>2] = $132;
   $133 = HEAP8[$128>>0]|0;
   $134 = $133 << 24 >> 24;
   $135 = (($134) + -65)|0;
   $136 = ((148184 + (($$0252*58)|0)|0) + ($135)|0);
   $137 = HEAP8[$136>>0]|0;
   $138 = $137&255;
   $139 = (($138) + -1)|0;
   $140 = ($139>>>0)<(8);
   if ($140) {
    $$0252 = $138;$128 = $132;
   } else {
    break;
   }
  }
  $141 = ($137<<24>>24)==(0);
  if ($141) {
   $$0 = -1;
   break;
  }
  $142 = ($137<<24>>24)==(19);
  $143 = ($$0253|0)>(-1);
  do {
   if ($142) {
    if ($143) {
     $$0 = -1;
     break L1;
    } else {
     label = 49;
    }
   } else {
    if ($143) {
     $144 = (($4) + ($$0253<<2)|0);
     HEAP32[$144>>2] = $138;
     $145 = (($3) + ($$0253<<3)|0);
     $146 = $145;
     $147 = $146;
     $148 = HEAP32[$147>>2]|0;
     $149 = (($146) + 4)|0;
     $150 = $149;
     $151 = HEAP32[$150>>2]|0;
     $152 = $6;
     $153 = $152;
     HEAP32[$153>>2] = $148;
     $154 = (($152) + 4)|0;
     $155 = $154;
     HEAP32[$155>>2] = $151;
     label = 49;
     break;
    }
    if (!($10)) {
     $$0 = 0;
     break L1;
    }
    _pop_arg($6,$138,$2);
   }
  } while(0);
  if ((label|0) == 49) {
   label = 0;
   if (!($10)) {
    $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
    continue;
   }
  }
  $156 = HEAP8[$128>>0]|0;
  $157 = $156 << 24 >> 24;
  $158 = ($$0252|0)!=(0);
  $159 = $157 & 15;
  $160 = ($159|0)==(3);
  $or$cond281 = $158 & $160;
  $161 = $157 & -33;
  $$0235 = $or$cond281 ? $161 : $157;
  $162 = $$1263 & 8192;
  $163 = ($162|0)==(0);
  $164 = $$1263 & -65537;
  $$1263$ = $163 ? $$1263 : $164;
  L71: do {
   switch ($$0235|0) {
   case 110:  {
    $trunc = $$0252&255;
    switch ($trunc<<24>>24) {
    case 0:  {
     $171 = HEAP32[$6>>2]|0;
     HEAP32[$171>>2] = $$1248;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 1:  {
     $172 = HEAP32[$6>>2]|0;
     HEAP32[$172>>2] = $$1248;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 2:  {
     $173 = ($$1248|0)<(0);
     $174 = $173 << 31 >> 31;
     $175 = HEAP32[$6>>2]|0;
     $176 = $175;
     $177 = $176;
     HEAP32[$177>>2] = $$1248;
     $178 = (($176) + 4)|0;
     $179 = $178;
     HEAP32[$179>>2] = $174;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 3:  {
     $180 = $$1248&65535;
     $181 = HEAP32[$6>>2]|0;
     HEAP16[$181>>1] = $180;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 4:  {
     $182 = $$1248&255;
     $183 = HEAP32[$6>>2]|0;
     HEAP8[$183>>0] = $182;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 6:  {
     $184 = HEAP32[$6>>2]|0;
     HEAP32[$184>>2] = $$1248;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    case 7:  {
     $185 = ($$1248|0)<(0);
     $186 = $185 << 31 >> 31;
     $187 = HEAP32[$6>>2]|0;
     $188 = $187;
     $189 = $188;
     HEAP32[$189>>2] = $$1248;
     $190 = (($188) + 4)|0;
     $191 = $190;
     HEAP32[$191>>2] = $186;
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
     break;
    }
    default: {
     $$0243 = 0;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
     continue L1;
    }
    }
    break;
   }
   case 112:  {
    $192 = ($$0254>>>0)>(8);
    $193 = $192 ? $$0254 : 8;
    $194 = $$1263$ | 8;
    $$1236 = 120;$$1255 = $193;$$3265 = $194;
    label = 61;
    break;
   }
   case 88: case 120:  {
    $$1236 = $$0235;$$1255 = $$0254;$$3265 = $$1263$;
    label = 61;
    break;
   }
   case 111:  {
    $210 = $6;
    $211 = $210;
    $212 = HEAP32[$211>>2]|0;
    $213 = (($210) + 4)|0;
    $214 = $213;
    $215 = HEAP32[$214>>2]|0;
    $216 = (_fmt_o($212,$215,$11)|0);
    $217 = $$1263$ & 8;
    $218 = ($217|0)==(0);
    $219 = $216;
    $220 = (($12) - ($219))|0;
    $221 = ($$0254|0)>($220|0);
    $222 = (($220) + 1)|0;
    $223 = $218 | $221;
    $$0254$$0254$ = $223 ? $$0254 : $222;
    $$0228 = $216;$$1233 = 0;$$1238 = 148648;$$2256 = $$0254$$0254$;$$4266 = $$1263$;$248 = $212;$250 = $215;
    label = 67;
    break;
   }
   case 105: case 100:  {
    $224 = $6;
    $225 = $224;
    $226 = HEAP32[$225>>2]|0;
    $227 = (($224) + 4)|0;
    $228 = $227;
    $229 = HEAP32[$228>>2]|0;
    $230 = ($229|0)<(0);
    if ($230) {
     $231 = (_i64Subtract(0,0,($226|0),($229|0))|0);
     $232 = tempRet0;
     $233 = $6;
     $234 = $233;
     HEAP32[$234>>2] = $231;
     $235 = (($233) + 4)|0;
     $236 = $235;
     HEAP32[$236>>2] = $232;
     $$0232 = 1;$$0237 = 148648;$242 = $231;$243 = $232;
     label = 66;
     break L71;
    } else {
     $237 = $$1263$ & 2048;
     $238 = ($237|0)==(0);
     $239 = $$1263$ & 1;
     $240 = ($239|0)==(0);
     $$ = $240 ? 148648 : (148650);
     $$$ = $238 ? $$ : (148649);
     $241 = $$1263$ & 2049;
     $narrow = ($241|0)!=(0);
     $$284$ = $narrow&1;
     $$0232 = $$284$;$$0237 = $$$;$242 = $226;$243 = $229;
     label = 66;
     break L71;
    }
    break;
   }
   case 117:  {
    $165 = $6;
    $166 = $165;
    $167 = HEAP32[$166>>2]|0;
    $168 = (($165) + 4)|0;
    $169 = $168;
    $170 = HEAP32[$169>>2]|0;
    $$0232 = 0;$$0237 = 148648;$242 = $167;$243 = $170;
    label = 66;
    break;
   }
   case 99:  {
    $259 = $6;
    $260 = $259;
    $261 = HEAP32[$260>>2]|0;
    $262 = (($259) + 4)|0;
    $263 = $262;
    $264 = HEAP32[$263>>2]|0;
    $265 = $261&255;
    HEAP8[$13>>0] = $265;
    $$2 = $13;$$2234 = 0;$$2239 = 148648;$$2251 = $11;$$5 = 1;$$6268 = $164;
    break;
   }
   case 109:  {
    $266 = (___errno_location()|0);
    $267 = HEAP32[$266>>2]|0;
    $268 = (_strerror($267)|0);
    $$1 = $268;
    label = 71;
    break;
   }
   case 115:  {
    $269 = HEAP32[$6>>2]|0;
    $270 = ($269|0)!=(0|0);
    $271 = $270 ? $269 : 148658;
    $$1 = $271;
    label = 71;
    break;
   }
   case 67:  {
    $278 = $6;
    $279 = $278;
    $280 = HEAP32[$279>>2]|0;
    $281 = (($278) + 4)|0;
    $282 = $281;
    $283 = HEAP32[$282>>2]|0;
    HEAP32[$8>>2] = $280;
    HEAP32[$14>>2] = 0;
    HEAP32[$6>>2] = $8;
    $$4258355 = -1;$331 = $8;
    label = 75;
    break;
   }
   case 83:  {
    $$pre349 = HEAP32[$6>>2]|0;
    $284 = ($$0254|0)==(0);
    if ($284) {
     _pad_674($0,32,$$1260,0,$$1263$);
     $$0240$lcssa357 = 0;
     label = 84;
    } else {
     $$4258355 = $$0254;$331 = $$pre349;
     label = 75;
    }
    break;
   }
   case 65: case 71: case 70: case 69: case 97: case 103: case 102: case 101:  {
    $306 = +HEAPF64[$6>>3];
    $307 = (_fmt_fp($0,$306,$$1260,$$0254,$$1263$,$$0235)|0);
    $$0243 = $307;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
    continue L1;
    break;
   }
   default: {
    $$2 = $21;$$2234 = 0;$$2239 = 148648;$$2251 = $11;$$5 = $$0254;$$6268 = $$1263$;
   }
   }
  } while(0);
  L95: do {
   if ((label|0) == 61) {
    label = 0;
    $195 = $6;
    $196 = $195;
    $197 = HEAP32[$196>>2]|0;
    $198 = (($195) + 4)|0;
    $199 = $198;
    $200 = HEAP32[$199>>2]|0;
    $201 = $$1236 & 32;
    $202 = (_fmt_x($197,$200,$11,$201)|0);
    $203 = ($197|0)==(0);
    $204 = ($200|0)==(0);
    $205 = $203 & $204;
    $206 = $$3265 & 8;
    $207 = ($206|0)==(0);
    $or$cond283 = $207 | $205;
    $208 = $$1236 >> 4;
    $209 = (148648 + ($208)|0);
    $$289 = $or$cond283 ? 148648 : $209;
    $$290 = $or$cond283 ? 0 : 2;
    $$0228 = $202;$$1233 = $$290;$$1238 = $$289;$$2256 = $$1255;$$4266 = $$3265;$248 = $197;$250 = $200;
    label = 67;
   }
   else if ((label|0) == 66) {
    label = 0;
    $244 = (_fmt_u($242,$243,$11)|0);
    $$0228 = $244;$$1233 = $$0232;$$1238 = $$0237;$$2256 = $$0254;$$4266 = $$1263$;$248 = $242;$250 = $243;
    label = 67;
   }
   else if ((label|0) == 71) {
    label = 0;
    $272 = (_memchr($$1,0,$$0254)|0);
    $273 = ($272|0)==(0|0);
    $274 = $272;
    $275 = $$1;
    $276 = (($274) - ($275))|0;
    $277 = (($$1) + ($$0254)|0);
    $$3257 = $273 ? $$0254 : $276;
    $$1250 = $273 ? $277 : $272;
    $$2 = $$1;$$2234 = 0;$$2239 = 148648;$$2251 = $$1250;$$5 = $$3257;$$6268 = $164;
   }
   else if ((label|0) == 75) {
    label = 0;
    $$0229322 = $331;$$0240321 = 0;$$1244320 = 0;
    while(1) {
     $285 = HEAP32[$$0229322>>2]|0;
     $286 = ($285|0)==(0);
     if ($286) {
      $$0240$lcssa = $$0240321;$$2245 = $$1244320;
      break;
     }
     $287 = (_wctomb($9,$285)|0);
     $288 = ($287|0)<(0);
     $289 = (($$4258355) - ($$0240321))|0;
     $290 = ($287>>>0)>($289>>>0);
     $or$cond286 = $288 | $290;
     if ($or$cond286) {
      $$0240$lcssa = $$0240321;$$2245 = $287;
      break;
     }
     $291 = ((($$0229322)) + 4|0);
     $292 = (($287) + ($$0240321))|0;
     $293 = ($$4258355>>>0)>($292>>>0);
     if ($293) {
      $$0229322 = $291;$$0240321 = $292;$$1244320 = $287;
     } else {
      $$0240$lcssa = $292;$$2245 = $287;
      break;
     }
    }
    $294 = ($$2245|0)<(0);
    if ($294) {
     $$0 = -1;
     break L1;
    }
    _pad_674($0,32,$$1260,$$0240$lcssa,$$1263$);
    $295 = ($$0240$lcssa|0)==(0);
    if ($295) {
     $$0240$lcssa357 = 0;
     label = 84;
    } else {
     $$1230333 = $331;$$1241332 = 0;
     while(1) {
      $296 = HEAP32[$$1230333>>2]|0;
      $297 = ($296|0)==(0);
      if ($297) {
       $$0240$lcssa357 = $$0240$lcssa;
       label = 84;
       break L95;
      }
      $298 = (_wctomb($9,$296)|0);
      $299 = (($298) + ($$1241332))|0;
      $300 = ($299|0)>($$0240$lcssa|0);
      if ($300) {
       $$0240$lcssa357 = $$0240$lcssa;
       label = 84;
       break L95;
      }
      $301 = ((($$1230333)) + 4|0);
      _out($0,$9,$298);
      $302 = ($299>>>0)<($$0240$lcssa>>>0);
      if ($302) {
       $$1230333 = $301;$$1241332 = $299;
      } else {
       $$0240$lcssa357 = $$0240$lcssa;
       label = 84;
       break;
      }
     }
    }
   }
  } while(0);
  if ((label|0) == 67) {
   label = 0;
   $245 = ($$2256|0)>(-1);
   $246 = $$4266 & -65537;
   $$$4266 = $245 ? $246 : $$4266;
   $247 = ($248|0)!=(0);
   $249 = ($250|0)!=(0);
   $251 = $247 | $249;
   $252 = ($$2256|0)!=(0);
   $or$cond = $252 | $251;
   $253 = $$0228;
   $254 = (($12) - ($253))|0;
   $255 = $251 ^ 1;
   $256 = $255&1;
   $257 = (($256) + ($254))|0;
   $258 = ($$2256|0)>($257|0);
   $$2256$ = $258 ? $$2256 : $257;
   $$2256$$$2256 = $or$cond ? $$2256$ : $$2256;
   $$0228$ = $or$cond ? $$0228 : $11;
   $$2 = $$0228$;$$2234 = $$1233;$$2239 = $$1238;$$2251 = $11;$$5 = $$2256$$$2256;$$6268 = $$$4266;
  }
  else if ((label|0) == 84) {
   label = 0;
   $303 = $$1263$ ^ 8192;
   _pad_674($0,32,$$1260,$$0240$lcssa357,$303);
   $304 = ($$1260|0)>($$0240$lcssa357|0);
   $305 = $304 ? $$1260 : $$0240$lcssa357;
   $$0243 = $305;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
   continue;
  }
  $308 = $$2251;
  $309 = $$2;
  $310 = (($308) - ($309))|0;
  $311 = ($$5|0)<($310|0);
  $$$5 = $311 ? $310 : $$5;
  $312 = (($$$5) + ($$2234))|0;
  $313 = ($$1260|0)<($312|0);
  $$2261 = $313 ? $312 : $$1260;
  _pad_674($0,32,$$2261,$312,$$6268);
  _out($0,$$2239,$$2234);
  $314 = $$6268 ^ 65536;
  _pad_674($0,48,$$2261,$312,$314);
  _pad_674($0,48,$$$5,$310,0);
  _out($0,$$2,$310);
  $315 = $$6268 ^ 8192;
  _pad_674($0,32,$$2261,$312,$315);
  $$0243 = $$2261;$$0247 = $$1248;$$0269 = $$3272;$21 = $132;
 }
 L114: do {
  if ((label|0) == 87) {
   $316 = ($0|0)==(0|0);
   if ($316) {
    $317 = ($$0269|0)==(0);
    if ($317) {
     $$0 = 0;
    } else {
     $$2242305 = 1;
     while(1) {
      $318 = (($4) + ($$2242305<<2)|0);
      $319 = HEAP32[$318>>2]|0;
      $320 = ($319|0)==(0);
      if ($320) {
       $$3303 = $$2242305;
       break;
      }
      $321 = (($3) + ($$2242305<<3)|0);
      _pop_arg($321,$319,$2);
      $322 = (($$2242305) + 1)|0;
      $323 = ($322|0)<(10);
      if ($323) {
       $$2242305 = $322;
      } else {
       $$0 = 1;
       break L114;
      }
     }
     while(1) {
      $326 = (($4) + ($$3303<<2)|0);
      $327 = HEAP32[$326>>2]|0;
      $328 = ($327|0)==(0);
      $325 = (($$3303) + 1)|0;
      if (!($328)) {
       $$0 = -1;
       break L114;
      }
      $324 = ($325|0)<(10);
      if ($324) {
       $$3303 = $325;
      } else {
       $$0 = 1;
       break;
      }
     }
    }
   } else {
    $$0 = $$1248;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function ___lockfile($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function ___unlockfile($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _out($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = HEAP32[$0>>2]|0;
 $4 = $3 & 32;
 $5 = ($4|0)==(0);
 if ($5) {
  (___fwritex($1,$2,$0)|0);
 }
 return;
}
function _getint($0) {
 $0 = $0|0;
 var $$0$lcssa = 0, $$06 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit5 = 0, $isdigittmp = 0, $isdigittmp4 = 0, $isdigittmp7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP8[$1>>0]|0;
 $3 = $2 << 24 >> 24;
 $isdigittmp4 = (($3) + -48)|0;
 $isdigit5 = ($isdigittmp4>>>0)<(10);
 if ($isdigit5) {
  $$06 = 0;$7 = $1;$isdigittmp7 = $isdigittmp4;
  while(1) {
   $4 = ($$06*10)|0;
   $5 = (($isdigittmp7) + ($4))|0;
   $6 = ((($7)) + 1|0);
   HEAP32[$0>>2] = $6;
   $8 = HEAP8[$6>>0]|0;
   $9 = $8 << 24 >> 24;
   $isdigittmp = (($9) + -48)|0;
   $isdigit = ($isdigittmp>>>0)<(10);
   if ($isdigit) {
    $$06 = $5;$7 = $6;$isdigittmp7 = $isdigittmp;
   } else {
    $$0$lcssa = $5;
    break;
   }
  }
 } else {
  $$0$lcssa = 0;
 }
 return ($$0$lcssa|0);
}
function _pop_arg($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$mask = 0, $$mask31 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0.0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arglist_current = 0, $arglist_current11 = 0, $arglist_current14 = 0, $arglist_current17 = 0;
 var $arglist_current2 = 0, $arglist_current20 = 0, $arglist_current23 = 0, $arglist_current26 = 0, $arglist_current5 = 0, $arglist_current8 = 0, $arglist_next = 0, $arglist_next12 = 0, $arglist_next15 = 0, $arglist_next18 = 0, $arglist_next21 = 0, $arglist_next24 = 0, $arglist_next27 = 0, $arglist_next3 = 0, $arglist_next6 = 0, $arglist_next9 = 0, $expanded = 0, $expanded28 = 0, $expanded30 = 0, $expanded31 = 0;
 var $expanded32 = 0, $expanded34 = 0, $expanded35 = 0, $expanded37 = 0, $expanded38 = 0, $expanded39 = 0, $expanded41 = 0, $expanded42 = 0, $expanded44 = 0, $expanded45 = 0, $expanded46 = 0, $expanded48 = 0, $expanded49 = 0, $expanded51 = 0, $expanded52 = 0, $expanded53 = 0, $expanded55 = 0, $expanded56 = 0, $expanded58 = 0, $expanded59 = 0;
 var $expanded60 = 0, $expanded62 = 0, $expanded63 = 0, $expanded65 = 0, $expanded66 = 0, $expanded67 = 0, $expanded69 = 0, $expanded70 = 0, $expanded72 = 0, $expanded73 = 0, $expanded74 = 0, $expanded76 = 0, $expanded77 = 0, $expanded79 = 0, $expanded80 = 0, $expanded81 = 0, $expanded83 = 0, $expanded84 = 0, $expanded86 = 0, $expanded87 = 0;
 var $expanded88 = 0, $expanded90 = 0, $expanded91 = 0, $expanded93 = 0, $expanded94 = 0, $expanded95 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($1>>>0)>(20);
 L1: do {
  if (!($3)) {
   do {
    switch ($1|0) {
    case 9:  {
     $arglist_current = HEAP32[$2>>2]|0;
     $4 = $arglist_current;
     $5 = ((0) + 4|0);
     $expanded28 = $5;
     $expanded = (($expanded28) - 1)|0;
     $6 = (($4) + ($expanded))|0;
     $7 = ((0) + 4|0);
     $expanded32 = $7;
     $expanded31 = (($expanded32) - 1)|0;
     $expanded30 = $expanded31 ^ -1;
     $8 = $6 & $expanded30;
     $9 = $8;
     $10 = HEAP32[$9>>2]|0;
     $arglist_next = ((($9)) + 4|0);
     HEAP32[$2>>2] = $arglist_next;
     HEAP32[$0>>2] = $10;
     break L1;
     break;
    }
    case 10:  {
     $arglist_current2 = HEAP32[$2>>2]|0;
     $11 = $arglist_current2;
     $12 = ((0) + 4|0);
     $expanded35 = $12;
     $expanded34 = (($expanded35) - 1)|0;
     $13 = (($11) + ($expanded34))|0;
     $14 = ((0) + 4|0);
     $expanded39 = $14;
     $expanded38 = (($expanded39) - 1)|0;
     $expanded37 = $expanded38 ^ -1;
     $15 = $13 & $expanded37;
     $16 = $15;
     $17 = HEAP32[$16>>2]|0;
     $arglist_next3 = ((($16)) + 4|0);
     HEAP32[$2>>2] = $arglist_next3;
     $18 = ($17|0)<(0);
     $19 = $18 << 31 >> 31;
     $20 = $0;
     $21 = $20;
     HEAP32[$21>>2] = $17;
     $22 = (($20) + 4)|0;
     $23 = $22;
     HEAP32[$23>>2] = $19;
     break L1;
     break;
    }
    case 11:  {
     $arglist_current5 = HEAP32[$2>>2]|0;
     $24 = $arglist_current5;
     $25 = ((0) + 4|0);
     $expanded42 = $25;
     $expanded41 = (($expanded42) - 1)|0;
     $26 = (($24) + ($expanded41))|0;
     $27 = ((0) + 4|0);
     $expanded46 = $27;
     $expanded45 = (($expanded46) - 1)|0;
     $expanded44 = $expanded45 ^ -1;
     $28 = $26 & $expanded44;
     $29 = $28;
     $30 = HEAP32[$29>>2]|0;
     $arglist_next6 = ((($29)) + 4|0);
     HEAP32[$2>>2] = $arglist_next6;
     $31 = $0;
     $32 = $31;
     HEAP32[$32>>2] = $30;
     $33 = (($31) + 4)|0;
     $34 = $33;
     HEAP32[$34>>2] = 0;
     break L1;
     break;
    }
    case 12:  {
     $arglist_current8 = HEAP32[$2>>2]|0;
     $35 = $arglist_current8;
     $36 = ((0) + 8|0);
     $expanded49 = $36;
     $expanded48 = (($expanded49) - 1)|0;
     $37 = (($35) + ($expanded48))|0;
     $38 = ((0) + 8|0);
     $expanded53 = $38;
     $expanded52 = (($expanded53) - 1)|0;
     $expanded51 = $expanded52 ^ -1;
     $39 = $37 & $expanded51;
     $40 = $39;
     $41 = $40;
     $42 = $41;
     $43 = HEAP32[$42>>2]|0;
     $44 = (($41) + 4)|0;
     $45 = $44;
     $46 = HEAP32[$45>>2]|0;
     $arglist_next9 = ((($40)) + 8|0);
     HEAP32[$2>>2] = $arglist_next9;
     $47 = $0;
     $48 = $47;
     HEAP32[$48>>2] = $43;
     $49 = (($47) + 4)|0;
     $50 = $49;
     HEAP32[$50>>2] = $46;
     break L1;
     break;
    }
    case 13:  {
     $arglist_current11 = HEAP32[$2>>2]|0;
     $51 = $arglist_current11;
     $52 = ((0) + 4|0);
     $expanded56 = $52;
     $expanded55 = (($expanded56) - 1)|0;
     $53 = (($51) + ($expanded55))|0;
     $54 = ((0) + 4|0);
     $expanded60 = $54;
     $expanded59 = (($expanded60) - 1)|0;
     $expanded58 = $expanded59 ^ -1;
     $55 = $53 & $expanded58;
     $56 = $55;
     $57 = HEAP32[$56>>2]|0;
     $arglist_next12 = ((($56)) + 4|0);
     HEAP32[$2>>2] = $arglist_next12;
     $58 = $57&65535;
     $59 = $58 << 16 >> 16;
     $60 = ($59|0)<(0);
     $61 = $60 << 31 >> 31;
     $62 = $0;
     $63 = $62;
     HEAP32[$63>>2] = $59;
     $64 = (($62) + 4)|0;
     $65 = $64;
     HEAP32[$65>>2] = $61;
     break L1;
     break;
    }
    case 14:  {
     $arglist_current14 = HEAP32[$2>>2]|0;
     $66 = $arglist_current14;
     $67 = ((0) + 4|0);
     $expanded63 = $67;
     $expanded62 = (($expanded63) - 1)|0;
     $68 = (($66) + ($expanded62))|0;
     $69 = ((0) + 4|0);
     $expanded67 = $69;
     $expanded66 = (($expanded67) - 1)|0;
     $expanded65 = $expanded66 ^ -1;
     $70 = $68 & $expanded65;
     $71 = $70;
     $72 = HEAP32[$71>>2]|0;
     $arglist_next15 = ((($71)) + 4|0);
     HEAP32[$2>>2] = $arglist_next15;
     $$mask31 = $72 & 65535;
     $73 = $0;
     $74 = $73;
     HEAP32[$74>>2] = $$mask31;
     $75 = (($73) + 4)|0;
     $76 = $75;
     HEAP32[$76>>2] = 0;
     break L1;
     break;
    }
    case 15:  {
     $arglist_current17 = HEAP32[$2>>2]|0;
     $77 = $arglist_current17;
     $78 = ((0) + 4|0);
     $expanded70 = $78;
     $expanded69 = (($expanded70) - 1)|0;
     $79 = (($77) + ($expanded69))|0;
     $80 = ((0) + 4|0);
     $expanded74 = $80;
     $expanded73 = (($expanded74) - 1)|0;
     $expanded72 = $expanded73 ^ -1;
     $81 = $79 & $expanded72;
     $82 = $81;
     $83 = HEAP32[$82>>2]|0;
     $arglist_next18 = ((($82)) + 4|0);
     HEAP32[$2>>2] = $arglist_next18;
     $84 = $83&255;
     $85 = $84 << 24 >> 24;
     $86 = ($85|0)<(0);
     $87 = $86 << 31 >> 31;
     $88 = $0;
     $89 = $88;
     HEAP32[$89>>2] = $85;
     $90 = (($88) + 4)|0;
     $91 = $90;
     HEAP32[$91>>2] = $87;
     break L1;
     break;
    }
    case 16:  {
     $arglist_current20 = HEAP32[$2>>2]|0;
     $92 = $arglist_current20;
     $93 = ((0) + 4|0);
     $expanded77 = $93;
     $expanded76 = (($expanded77) - 1)|0;
     $94 = (($92) + ($expanded76))|0;
     $95 = ((0) + 4|0);
     $expanded81 = $95;
     $expanded80 = (($expanded81) - 1)|0;
     $expanded79 = $expanded80 ^ -1;
     $96 = $94 & $expanded79;
     $97 = $96;
     $98 = HEAP32[$97>>2]|0;
     $arglist_next21 = ((($97)) + 4|0);
     HEAP32[$2>>2] = $arglist_next21;
     $$mask = $98 & 255;
     $99 = $0;
     $100 = $99;
     HEAP32[$100>>2] = $$mask;
     $101 = (($99) + 4)|0;
     $102 = $101;
     HEAP32[$102>>2] = 0;
     break L1;
     break;
    }
    case 17:  {
     $arglist_current23 = HEAP32[$2>>2]|0;
     $103 = $arglist_current23;
     $104 = ((0) + 8|0);
     $expanded84 = $104;
     $expanded83 = (($expanded84) - 1)|0;
     $105 = (($103) + ($expanded83))|0;
     $106 = ((0) + 8|0);
     $expanded88 = $106;
     $expanded87 = (($expanded88) - 1)|0;
     $expanded86 = $expanded87 ^ -1;
     $107 = $105 & $expanded86;
     $108 = $107;
     $109 = +HEAPF64[$108>>3];
     $arglist_next24 = ((($108)) + 8|0);
     HEAP32[$2>>2] = $arglist_next24;
     HEAPF64[$0>>3] = $109;
     break L1;
     break;
    }
    case 18:  {
     $arglist_current26 = HEAP32[$2>>2]|0;
     $110 = $arglist_current26;
     $111 = ((0) + 8|0);
     $expanded91 = $111;
     $expanded90 = (($expanded91) - 1)|0;
     $112 = (($110) + ($expanded90))|0;
     $113 = ((0) + 8|0);
     $expanded95 = $113;
     $expanded94 = (($expanded95) - 1)|0;
     $expanded93 = $expanded94 ^ -1;
     $114 = $112 & $expanded93;
     $115 = $114;
     $116 = +HEAPF64[$115>>3];
     $arglist_next27 = ((($115)) + 8|0);
     HEAP32[$2>>2] = $arglist_next27;
     HEAPF64[$0>>3] = $116;
     break L1;
     break;
    }
    default: {
     break L1;
    }
    }
   } while(0);
  }
 } while(0);
 return;
}
function _fmt_x($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$05$lcssa = 0, $$056 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $4 = ($0|0)==(0);
 $5 = ($1|0)==(0);
 $6 = $4 & $5;
 if ($6) {
  $$05$lcssa = $2;
 } else {
  $$056 = $2;$15 = $1;$8 = $0;
  while(1) {
   $7 = $8 & 15;
   $9 = (148696 + ($7)|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = $10&255;
   $12 = $11 | $3;
   $13 = $12&255;
   $14 = ((($$056)) + -1|0);
   HEAP8[$14>>0] = $13;
   $16 = (_bitshift64Lshr(($8|0),($15|0),4)|0);
   $17 = tempRet0;
   $18 = ($16|0)==(0);
   $19 = ($17|0)==(0);
   $20 = $18 & $19;
   if ($20) {
    $$05$lcssa = $14;
    break;
   } else {
    $$056 = $14;$15 = $17;$8 = $16;
   }
  }
 }
 return ($$05$lcssa|0);
}
function _fmt_o($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$lcssa = 0, $$06 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==(0);
 $4 = ($1|0)==(0);
 $5 = $3 & $4;
 if ($5) {
  $$0$lcssa = $2;
 } else {
  $$06 = $2;$11 = $1;$7 = $0;
  while(1) {
   $6 = $7&255;
   $8 = $6 & 7;
   $9 = $8 | 48;
   $10 = ((($$06)) + -1|0);
   HEAP8[$10>>0] = $9;
   $12 = (_bitshift64Lshr(($7|0),($11|0),3)|0);
   $13 = tempRet0;
   $14 = ($12|0)==(0);
   $15 = ($13|0)==(0);
   $16 = $14 & $15;
   if ($16) {
    $$0$lcssa = $10;
    break;
   } else {
    $$06 = $10;$11 = $13;$7 = $12;
   }
  }
 }
 return ($$0$lcssa|0);
}
function _fmt_u($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$010$lcssa$off0 = 0, $$012 = 0, $$09$lcssa = 0, $$0914 = 0, $$1$lcssa = 0, $$111 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($1>>>0)>(0);
 $4 = ($0>>>0)>(4294967295);
 $5 = ($1|0)==(0);
 $6 = $5 & $4;
 $7 = $3 | $6;
 if ($7) {
  $$0914 = $2;$8 = $0;$9 = $1;
  while(1) {
   $10 = (___uremdi3(($8|0),($9|0),10,0)|0);
   $11 = tempRet0;
   $12 = $10&255;
   $13 = $12 | 48;
   $14 = ((($$0914)) + -1|0);
   HEAP8[$14>>0] = $13;
   $15 = (___udivdi3(($8|0),($9|0),10,0)|0);
   $16 = tempRet0;
   $17 = ($9>>>0)>(9);
   $18 = ($8>>>0)>(4294967295);
   $19 = ($9|0)==(9);
   $20 = $19 & $18;
   $21 = $17 | $20;
   if ($21) {
    $$0914 = $14;$8 = $15;$9 = $16;
   } else {
    break;
   }
  }
  $$010$lcssa$off0 = $15;$$09$lcssa = $14;
 } else {
  $$010$lcssa$off0 = $0;$$09$lcssa = $2;
 }
 $22 = ($$010$lcssa$off0|0)==(0);
 if ($22) {
  $$1$lcssa = $$09$lcssa;
 } else {
  $$012 = $$010$lcssa$off0;$$111 = $$09$lcssa;
  while(1) {
   $23 = (($$012>>>0) % 10)&-1;
   $24 = $23 | 48;
   $25 = $24&255;
   $26 = ((($$111)) + -1|0);
   HEAP8[$26>>0] = $25;
   $27 = (($$012>>>0) / 10)&-1;
   $28 = ($$012>>>0)<(10);
   if ($28) {
    $$1$lcssa = $26;
    break;
   } else {
    $$012 = $27;$$111 = $26;
   }
  }
 }
 return ($$1$lcssa|0);
}
function _strerror($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___pthread_self_105()|0);
 $2 = ((($1)) + 188|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (___strerror_l($0,$3)|0);
 return ($4|0);
}
function _memchr($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$lcssa = 0, $$035$lcssa = 0, $$035$lcssa65 = 0, $$03555 = 0, $$036$lcssa = 0, $$036$lcssa64 = 0, $$03654 = 0, $$046 = 0, $$137$lcssa = 0, $$13745 = 0, $$140 = 0, $$2 = 0, $$23839 = 0, $$3 = 0, $$lcssa = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0;
 var $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond53 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $1 & 255;
 $4 = $0;
 $5 = $4 & 3;
 $6 = ($5|0)!=(0);
 $7 = ($2|0)!=(0);
 $or$cond53 = $7 & $6;
 L1: do {
  if ($or$cond53) {
   $8 = $1&255;
   $$03555 = $0;$$03654 = $2;
   while(1) {
    $9 = HEAP8[$$03555>>0]|0;
    $10 = ($9<<24>>24)==($8<<24>>24);
    if ($10) {
     $$035$lcssa65 = $$03555;$$036$lcssa64 = $$03654;
     label = 6;
     break L1;
    }
    $11 = ((($$03555)) + 1|0);
    $12 = (($$03654) + -1)|0;
    $13 = $11;
    $14 = $13 & 3;
    $15 = ($14|0)!=(0);
    $16 = ($12|0)!=(0);
    $or$cond = $16 & $15;
    if ($or$cond) {
     $$03555 = $11;$$03654 = $12;
    } else {
     $$035$lcssa = $11;$$036$lcssa = $12;$$lcssa = $16;
     label = 5;
     break;
    }
   }
  } else {
   $$035$lcssa = $0;$$036$lcssa = $2;$$lcssa = $7;
   label = 5;
  }
 } while(0);
 if ((label|0) == 5) {
  if ($$lcssa) {
   $$035$lcssa65 = $$035$lcssa;$$036$lcssa64 = $$036$lcssa;
   label = 6;
  } else {
   $$2 = $$035$lcssa;$$3 = 0;
  }
 }
 L8: do {
  if ((label|0) == 6) {
   $17 = HEAP8[$$035$lcssa65>>0]|0;
   $18 = $1&255;
   $19 = ($17<<24>>24)==($18<<24>>24);
   if ($19) {
    $$2 = $$035$lcssa65;$$3 = $$036$lcssa64;
   } else {
    $20 = Math_imul($3, 16843009)|0;
    $21 = ($$036$lcssa64>>>0)>(3);
    L11: do {
     if ($21) {
      $$046 = $$035$lcssa65;$$13745 = $$036$lcssa64;
      while(1) {
       $22 = HEAP32[$$046>>2]|0;
       $23 = $22 ^ $20;
       $24 = (($23) + -16843009)|0;
       $25 = $23 & -2139062144;
       $26 = $25 ^ -2139062144;
       $27 = $26 & $24;
       $28 = ($27|0)==(0);
       if (!($28)) {
        break;
       }
       $29 = ((($$046)) + 4|0);
       $30 = (($$13745) + -4)|0;
       $31 = ($30>>>0)>(3);
       if ($31) {
        $$046 = $29;$$13745 = $30;
       } else {
        $$0$lcssa = $29;$$137$lcssa = $30;
        label = 11;
        break L11;
       }
      }
      $$140 = $$046;$$23839 = $$13745;
     } else {
      $$0$lcssa = $$035$lcssa65;$$137$lcssa = $$036$lcssa64;
      label = 11;
     }
    } while(0);
    if ((label|0) == 11) {
     $32 = ($$137$lcssa|0)==(0);
     if ($32) {
      $$2 = $$0$lcssa;$$3 = 0;
      break;
     } else {
      $$140 = $$0$lcssa;$$23839 = $$137$lcssa;
     }
    }
    while(1) {
     $33 = HEAP8[$$140>>0]|0;
     $34 = ($33<<24>>24)==($18<<24>>24);
     if ($34) {
      $$2 = $$140;$$3 = $$23839;
      break L8;
     }
     $35 = ((($$140)) + 1|0);
     $36 = (($$23839) + -1)|0;
     $37 = ($36|0)==(0);
     if ($37) {
      $$2 = $35;$$3 = 0;
      break;
     } else {
      $$140 = $35;$$23839 = $36;
     }
    }
   }
  }
 } while(0);
 $38 = ($$3|0)!=(0);
 $39 = $38 ? $$2 : 0;
 return ($39|0);
}
function _pad_674($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0$lcssa = 0, $$011 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(256|0);
 $5 = sp;
 $6 = $4 & 73728;
 $7 = ($6|0)==(0);
 $8 = ($2|0)>($3|0);
 $or$cond = $8 & $7;
 if ($or$cond) {
  $9 = (($2) - ($3))|0;
  $10 = ($9>>>0)<(256);
  $11 = $10 ? $9 : 256;
  _memset(($5|0),($1|0),($11|0))|0;
  $12 = ($9>>>0)>(255);
  if ($12) {
   $13 = (($2) - ($3))|0;
   $$011 = $9;
   while(1) {
    _out($0,$5,256);
    $14 = (($$011) + -256)|0;
    $15 = ($14>>>0)>(255);
    if ($15) {
     $$011 = $14;
    } else {
     break;
    }
   }
   $16 = $13 & 255;
   $$0$lcssa = $16;
  } else {
   $$0$lcssa = $9;
  }
  _out($0,$5,$$0$lcssa);
 }
 STACKTOP = sp;return;
}
function _wctomb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==(0|0);
 if ($2) {
  $$0 = 0;
 } else {
  $3 = (_wcrtomb($0,$1,0)|0);
  $$0 = $3;
 }
 return ($$0|0);
}
function _fmt_fp($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = +$1;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$ = 0, $$$ = 0, $$$$559 = 0.0, $$$3484 = 0, $$$3484691 = 0, $$$3484692 = 0, $$$3501 = 0, $$$4502 = 0, $$$542 = 0.0, $$$559 = 0.0, $$0 = 0, $$0463$lcssa = 0, $$0463584 = 0, $$0464594 = 0, $$0471 = 0.0, $$0479 = 0, $$0487642 = 0, $$0488 = 0, $$0488653 = 0, $$0488655 = 0;
 var $$0496$$9 = 0, $$0497654 = 0, $$0498 = 0, $$0509582 = 0.0, $$0510 = 0, $$0511 = 0, $$0514637 = 0, $$0520 = 0, $$0521 = 0, $$0521$ = 0, $$0523 = 0, $$0525 = 0, $$0527 = 0, $$0527629 = 0, $$0527631 = 0, $$0530636 = 0, $$1465 = 0, $$1467 = 0.0, $$1469 = 0.0, $$1472 = 0.0;
 var $$1480 = 0, $$1482$lcssa = 0, $$1482661 = 0, $$1489641 = 0, $$1499$lcssa = 0, $$1499660 = 0, $$1508583 = 0, $$1512$lcssa = 0, $$1512607 = 0, $$1515 = 0, $$1524 = 0, $$1526 = 0, $$1528614 = 0, $$1531$lcssa = 0, $$1531630 = 0, $$1598 = 0, $$2 = 0, $$2473 = 0.0, $$2476 = 0, $$2476$$547 = 0;
 var $$2476$$549 = 0, $$2483$ph = 0, $$2500 = 0, $$2513 = 0, $$2516618 = 0, $$2529 = 0, $$2532617 = 0, $$3 = 0.0, $$3477 = 0, $$3484$lcssa = 0, $$3484648 = 0, $$3501$lcssa = 0, $$3501647 = 0, $$3533613 = 0, $$4 = 0.0, $$4478$lcssa = 0, $$4478590 = 0, $$4492 = 0, $$4502 = 0, $$4518 = 0;
 var $$5$lcssa = 0, $$534$ = 0, $$539 = 0, $$539$ = 0, $$542 = 0.0, $$546 = 0, $$548 = 0, $$5486$lcssa = 0, $$5486623 = 0, $$5493597 = 0, $$5519$ph = 0, $$555 = 0, $$556 = 0, $$559 = 0.0, $$5602 = 0, $$6 = 0, $$6494589 = 0, $$7495601 = 0, $$7505 = 0, $$7505$ = 0;
 var $$7505$ph = 0, $$8 = 0, $$9$ph = 0, $$lcssa673 = 0, $$neg = 0, $$neg567 = 0, $$pn = 0, $$pn566 = 0, $$pr = 0, $$pr564 = 0, $$pre = 0, $$pre$phi690Z2D = 0, $$pre689 = 0, $$sink545$lcssa = 0, $$sink545622 = 0, $$sink562 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0.0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0.0, $117 = 0.0, $118 = 0.0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0.0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0;
 var $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0;
 var $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0.0, $229 = 0.0, $23 = 0;
 var $230 = 0, $231 = 0.0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0;
 var $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0;
 var $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0;
 var $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0;
 var $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0;
 var $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0;
 var $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0.0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0;
 var $358 = 0, $359 = 0, $36 = 0.0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0;
 var $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0.0, $52 = 0, $53 = 0, $54 = 0, $55 = 0.0, $56 = 0.0, $57 = 0.0, $58 = 0.0, $59 = 0.0, $6 = 0, $60 = 0.0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0.0, $88 = 0.0, $89 = 0.0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond = 0;
 var $narrow = 0, $not$ = 0, $notlhs = 0, $notrhs = 0, $or$cond = 0, $or$cond3$not = 0, $or$cond537 = 0, $or$cond541 = 0, $or$cond544 = 0, $or$cond554 = 0, $or$cond6 = 0, $scevgep684 = 0, $scevgep684685 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 560|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(560|0);
 $6 = sp + 8|0;
 $7 = sp;
 $8 = sp + 524|0;
 $9 = $8;
 $10 = sp + 512|0;
 HEAP32[$7>>2] = 0;
 $11 = ((($10)) + 12|0);
 (___DOUBLE_BITS_675($1)|0);
 $12 = tempRet0;
 $13 = ($12|0)<(0);
 if ($13) {
  $14 = -$1;
  $$0471 = $14;$$0520 = 1;$$0521 = 148665;
 } else {
  $15 = $4 & 2048;
  $16 = ($15|0)==(0);
  $17 = $4 & 1;
  $18 = ($17|0)==(0);
  $$ = $18 ? (148666) : (148671);
  $$$ = $16 ? $$ : (148668);
  $19 = $4 & 2049;
  $narrow = ($19|0)!=(0);
  $$534$ = $narrow&1;
  $$0471 = $1;$$0520 = $$534$;$$0521 = $$$;
 }
 (___DOUBLE_BITS_675($$0471)|0);
 $20 = tempRet0;
 $21 = $20 & 2146435072;
 $22 = ($21>>>0)<(2146435072);
 $23 = (0)<(0);
 $24 = ($21|0)==(2146435072);
 $25 = $24 & $23;
 $26 = $22 | $25;
 do {
  if ($26) {
   $35 = (+_frexpl($$0471,$7));
   $36 = $35 * 2.0;
   $37 = $36 != 0.0;
   if ($37) {
    $38 = HEAP32[$7>>2]|0;
    $39 = (($38) + -1)|0;
    HEAP32[$7>>2] = $39;
   }
   $40 = $5 | 32;
   $41 = ($40|0)==(97);
   if ($41) {
    $42 = $5 & 32;
    $43 = ($42|0)==(0);
    $44 = ((($$0521)) + 9|0);
    $$0521$ = $43 ? $$0521 : $44;
    $45 = $$0520 | 2;
    $46 = ($3>>>0)>(11);
    $47 = (12 - ($3))|0;
    $48 = ($47|0)==(0);
    $49 = $46 | $48;
    do {
     if ($49) {
      $$1472 = $36;
     } else {
      $$0509582 = 8.0;$$1508583 = $47;
      while(1) {
       $50 = (($$1508583) + -1)|0;
       $51 = $$0509582 * 16.0;
       $52 = ($50|0)==(0);
       if ($52) {
        break;
       } else {
        $$0509582 = $51;$$1508583 = $50;
       }
      }
      $53 = HEAP8[$$0521$>>0]|0;
      $54 = ($53<<24>>24)==(45);
      if ($54) {
       $55 = -$36;
       $56 = $55 - $51;
       $57 = $51 + $56;
       $58 = -$57;
       $$1472 = $58;
       break;
      } else {
       $59 = $36 + $51;
       $60 = $59 - $51;
       $$1472 = $60;
       break;
      }
     }
    } while(0);
    $61 = HEAP32[$7>>2]|0;
    $62 = ($61|0)<(0);
    $63 = (0 - ($61))|0;
    $64 = $62 ? $63 : $61;
    $65 = ($64|0)<(0);
    $66 = $65 << 31 >> 31;
    $67 = (_fmt_u($64,$66,$11)|0);
    $68 = ($67|0)==($11|0);
    if ($68) {
     $69 = ((($10)) + 11|0);
     HEAP8[$69>>0] = 48;
     $$0511 = $69;
    } else {
     $$0511 = $67;
    }
    $70 = $61 >> 31;
    $71 = $70 & 2;
    $72 = (($71) + 43)|0;
    $73 = $72&255;
    $74 = ((($$0511)) + -1|0);
    HEAP8[$74>>0] = $73;
    $75 = (($5) + 15)|0;
    $76 = $75&255;
    $77 = ((($$0511)) + -2|0);
    HEAP8[$77>>0] = $76;
    $notrhs = ($3|0)<(1);
    $78 = $4 & 8;
    $79 = ($78|0)==(0);
    $$0523 = $8;$$2473 = $$1472;
    while(1) {
     $80 = (~~(($$2473)));
     $81 = (148696 + ($80)|0);
     $82 = HEAP8[$81>>0]|0;
     $83 = $82&255;
     $84 = $83 | $42;
     $85 = $84&255;
     $86 = ((($$0523)) + 1|0);
     HEAP8[$$0523>>0] = $85;
     $87 = (+($80|0));
     $88 = $$2473 - $87;
     $89 = $88 * 16.0;
     $90 = $86;
     $91 = (($90) - ($9))|0;
     $92 = ($91|0)==(1);
     if ($92) {
      $notlhs = $89 == 0.0;
      $or$cond3$not = $notrhs & $notlhs;
      $or$cond = $79 & $or$cond3$not;
      if ($or$cond) {
       $$1524 = $86;
      } else {
       $93 = ((($$0523)) + 2|0);
       HEAP8[$86>>0] = 46;
       $$1524 = $93;
      }
     } else {
      $$1524 = $86;
     }
     $94 = $89 != 0.0;
     if ($94) {
      $$0523 = $$1524;$$2473 = $89;
     } else {
      break;
     }
    }
    $95 = ($3|0)!=(0);
    $96 = $77;
    $97 = $11;
    $98 = $$1524;
    $99 = (($98) - ($9))|0;
    $100 = (($97) - ($96))|0;
    $101 = (($99) + -2)|0;
    $102 = ($101|0)<($3|0);
    $or$cond537 = $95 & $102;
    $103 = (($3) + 2)|0;
    $$pn = $or$cond537 ? $103 : $99;
    $$0525 = (($100) + ($45))|0;
    $104 = (($$0525) + ($$pn))|0;
    _pad_674($0,32,$2,$104,$4);
    _out($0,$$0521$,$45);
    $105 = $4 ^ 65536;
    _pad_674($0,48,$2,$104,$105);
    _out($0,$8,$99);
    $106 = (($$pn) - ($99))|0;
    _pad_674($0,48,$106,0,0);
    _out($0,$77,$100);
    $107 = $4 ^ 8192;
    _pad_674($0,32,$2,$104,$107);
    $$sink562 = $104;
    break;
   }
   $108 = ($3|0)<(0);
   $$539 = $108 ? 6 : $3;
   if ($37) {
    $109 = $36 * 268435456.0;
    $110 = HEAP32[$7>>2]|0;
    $111 = (($110) + -28)|0;
    HEAP32[$7>>2] = $111;
    $$3 = $109;$$pr = $111;
   } else {
    $$pre = HEAP32[$7>>2]|0;
    $$3 = $36;$$pr = $$pre;
   }
   $112 = ($$pr|0)<(0);
   $113 = ((($6)) + 288|0);
   $$556 = $112 ? $6 : $113;
   $$0498 = $$556;$$4 = $$3;
   while(1) {
    $114 = (~~(($$4))>>>0);
    HEAP32[$$0498>>2] = $114;
    $115 = ((($$0498)) + 4|0);
    $116 = (+($114>>>0));
    $117 = $$4 - $116;
    $118 = $117 * 1.0E+9;
    $119 = $118 != 0.0;
    if ($119) {
     $$0498 = $115;$$4 = $118;
    } else {
     break;
    }
   }
   $120 = ($$pr|0)>(0);
   if ($120) {
    $$1482661 = $$556;$$1499660 = $115;$122 = $$pr;
    while(1) {
     $121 = ($122|0)<(29);
     $123 = $121 ? $122 : 29;
     $$0488653 = ((($$1499660)) + -4|0);
     $124 = ($$0488653>>>0)<($$1482661>>>0);
     if ($124) {
      $$2483$ph = $$1482661;
     } else {
      $$0488655 = $$0488653;$$0497654 = 0;
      while(1) {
       $125 = HEAP32[$$0488655>>2]|0;
       $126 = (_bitshift64Shl(($125|0),0,($123|0))|0);
       $127 = tempRet0;
       $128 = (_i64Add(($126|0),($127|0),($$0497654|0),0)|0);
       $129 = tempRet0;
       $130 = (___uremdi3(($128|0),($129|0),1000000000,0)|0);
       $131 = tempRet0;
       HEAP32[$$0488655>>2] = $130;
       $132 = (___udivdi3(($128|0),($129|0),1000000000,0)|0);
       $133 = tempRet0;
       $$0488 = ((($$0488655)) + -4|0);
       $134 = ($$0488>>>0)<($$1482661>>>0);
       if ($134) {
        break;
       } else {
        $$0488655 = $$0488;$$0497654 = $132;
       }
      }
      $135 = ($132|0)==(0);
      if ($135) {
       $$2483$ph = $$1482661;
      } else {
       $136 = ((($$1482661)) + -4|0);
       HEAP32[$136>>2] = $132;
       $$2483$ph = $136;
      }
     }
     $$2500 = $$1499660;
     while(1) {
      $137 = ($$2500>>>0)>($$2483$ph>>>0);
      if (!($137)) {
       break;
      }
      $138 = ((($$2500)) + -4|0);
      $139 = HEAP32[$138>>2]|0;
      $140 = ($139|0)==(0);
      if ($140) {
       $$2500 = $138;
      } else {
       break;
      }
     }
     $141 = HEAP32[$7>>2]|0;
     $142 = (($141) - ($123))|0;
     HEAP32[$7>>2] = $142;
     $143 = ($142|0)>(0);
     if ($143) {
      $$1482661 = $$2483$ph;$$1499660 = $$2500;$122 = $142;
     } else {
      $$1482$lcssa = $$2483$ph;$$1499$lcssa = $$2500;$$pr564 = $142;
      break;
     }
    }
   } else {
    $$1482$lcssa = $$556;$$1499$lcssa = $115;$$pr564 = $$pr;
   }
   $144 = ($$pr564|0)<(0);
   if ($144) {
    $145 = (($$539) + 25)|0;
    $146 = (($145|0) / 9)&-1;
    $147 = (($146) + 1)|0;
    $148 = ($40|0)==(102);
    $$3484648 = $$1482$lcssa;$$3501647 = $$1499$lcssa;$150 = $$pr564;
    while(1) {
     $149 = (0 - ($150))|0;
     $151 = ($149|0)<(9);
     $152 = $151 ? $149 : 9;
     $153 = ($$3484648>>>0)<($$3501647>>>0);
     if ($153) {
      $157 = 1 << $152;
      $158 = (($157) + -1)|0;
      $159 = 1000000000 >>> $152;
      $$0487642 = 0;$$1489641 = $$3484648;
      while(1) {
       $160 = HEAP32[$$1489641>>2]|0;
       $161 = $160 & $158;
       $162 = $160 >>> $152;
       $163 = (($162) + ($$0487642))|0;
       HEAP32[$$1489641>>2] = $163;
       $164 = Math_imul($161, $159)|0;
       $165 = ((($$1489641)) + 4|0);
       $166 = ($165>>>0)<($$3501647>>>0);
       if ($166) {
        $$0487642 = $164;$$1489641 = $165;
       } else {
        break;
       }
      }
      $167 = HEAP32[$$3484648>>2]|0;
      $168 = ($167|0)==(0);
      $169 = ((($$3484648)) + 4|0);
      $$$3484 = $168 ? $169 : $$3484648;
      $170 = ($164|0)==(0);
      if ($170) {
       $$$3484692 = $$$3484;$$4502 = $$3501647;
      } else {
       $171 = ((($$3501647)) + 4|0);
       HEAP32[$$3501647>>2] = $164;
       $$$3484692 = $$$3484;$$4502 = $171;
      }
     } else {
      $154 = HEAP32[$$3484648>>2]|0;
      $155 = ($154|0)==(0);
      $156 = ((($$3484648)) + 4|0);
      $$$3484691 = $155 ? $156 : $$3484648;
      $$$3484692 = $$$3484691;$$4502 = $$3501647;
     }
     $172 = $148 ? $$556 : $$$3484692;
     $173 = $$4502;
     $174 = $172;
     $175 = (($173) - ($174))|0;
     $176 = $175 >> 2;
     $177 = ($176|0)>($147|0);
     $178 = (($172) + ($147<<2)|0);
     $$$4502 = $177 ? $178 : $$4502;
     $179 = HEAP32[$7>>2]|0;
     $180 = (($179) + ($152))|0;
     HEAP32[$7>>2] = $180;
     $181 = ($180|0)<(0);
     if ($181) {
      $$3484648 = $$$3484692;$$3501647 = $$$4502;$150 = $180;
     } else {
      $$3484$lcssa = $$$3484692;$$3501$lcssa = $$$4502;
      break;
     }
    }
   } else {
    $$3484$lcssa = $$1482$lcssa;$$3501$lcssa = $$1499$lcssa;
   }
   $182 = ($$3484$lcssa>>>0)<($$3501$lcssa>>>0);
   $183 = $$556;
   if ($182) {
    $184 = $$3484$lcssa;
    $185 = (($183) - ($184))|0;
    $186 = $185 >> 2;
    $187 = ($186*9)|0;
    $188 = HEAP32[$$3484$lcssa>>2]|0;
    $189 = ($188>>>0)<(10);
    if ($189) {
     $$1515 = $187;
    } else {
     $$0514637 = $187;$$0530636 = 10;
     while(1) {
      $190 = ($$0530636*10)|0;
      $191 = (($$0514637) + 1)|0;
      $192 = ($188>>>0)<($190>>>0);
      if ($192) {
       $$1515 = $191;
       break;
      } else {
       $$0514637 = $191;$$0530636 = $190;
      }
     }
    }
   } else {
    $$1515 = 0;
   }
   $193 = ($40|0)!=(102);
   $194 = $193 ? $$1515 : 0;
   $195 = (($$539) - ($194))|0;
   $196 = ($40|0)==(103);
   $197 = ($$539|0)!=(0);
   $198 = $197 & $196;
   $$neg = $198 << 31 >> 31;
   $199 = (($195) + ($$neg))|0;
   $200 = $$3501$lcssa;
   $201 = (($200) - ($183))|0;
   $202 = $201 >> 2;
   $203 = ($202*9)|0;
   $204 = (($203) + -9)|0;
   $205 = ($199|0)<($204|0);
   if ($205) {
    $206 = ((($$556)) + 4|0);
    $207 = (($199) + 9216)|0;
    $208 = (($207|0) / 9)&-1;
    $209 = (($208) + -1024)|0;
    $210 = (($206) + ($209<<2)|0);
    $211 = (($207|0) % 9)&-1;
    $$0527629 = (($211) + 1)|0;
    $212 = ($$0527629|0)<(9);
    if ($212) {
     $$0527631 = $$0527629;$$1531630 = 10;
     while(1) {
      $213 = ($$1531630*10)|0;
      $$0527 = (($$0527631) + 1)|0;
      $exitcond = ($$0527|0)==(9);
      if ($exitcond) {
       $$1531$lcssa = $213;
       break;
      } else {
       $$0527631 = $$0527;$$1531630 = $213;
      }
     }
    } else {
     $$1531$lcssa = 10;
    }
    $214 = HEAP32[$210>>2]|0;
    $215 = (($214>>>0) % ($$1531$lcssa>>>0))&-1;
    $216 = ($215|0)==(0);
    $217 = ((($210)) + 4|0);
    $218 = ($217|0)==($$3501$lcssa|0);
    $or$cond541 = $218 & $216;
    if ($or$cond541) {
     $$4492 = $210;$$4518 = $$1515;$$8 = $$3484$lcssa;
    } else {
     $219 = (($214>>>0) / ($$1531$lcssa>>>0))&-1;
     $220 = $219 & 1;
     $221 = ($220|0)==(0);
     $$542 = $221 ? 9007199254740992.0 : 9007199254740994.0;
     $222 = (($$1531$lcssa|0) / 2)&-1;
     $223 = ($215>>>0)<($222>>>0);
     $224 = ($215|0)==($222|0);
     $or$cond544 = $218 & $224;
     $$559 = $or$cond544 ? 1.0 : 1.5;
     $$$559 = $223 ? 0.5 : $$559;
     $225 = ($$0520|0)==(0);
     if ($225) {
      $$1467 = $$$559;$$1469 = $$542;
     } else {
      $226 = HEAP8[$$0521>>0]|0;
      $227 = ($226<<24>>24)==(45);
      $228 = -$$542;
      $229 = -$$$559;
      $$$542 = $227 ? $228 : $$542;
      $$$$559 = $227 ? $229 : $$$559;
      $$1467 = $$$$559;$$1469 = $$$542;
     }
     $230 = (($214) - ($215))|0;
     HEAP32[$210>>2] = $230;
     $231 = $$1469 + $$1467;
     $232 = $231 != $$1469;
     if ($232) {
      $233 = (($230) + ($$1531$lcssa))|0;
      HEAP32[$210>>2] = $233;
      $234 = ($233>>>0)>(999999999);
      if ($234) {
       $$5486623 = $$3484$lcssa;$$sink545622 = $210;
       while(1) {
        $235 = ((($$sink545622)) + -4|0);
        HEAP32[$$sink545622>>2] = 0;
        $236 = ($235>>>0)<($$5486623>>>0);
        if ($236) {
         $237 = ((($$5486623)) + -4|0);
         HEAP32[$237>>2] = 0;
         $$6 = $237;
        } else {
         $$6 = $$5486623;
        }
        $238 = HEAP32[$235>>2]|0;
        $239 = (($238) + 1)|0;
        HEAP32[$235>>2] = $239;
        $240 = ($239>>>0)>(999999999);
        if ($240) {
         $$5486623 = $$6;$$sink545622 = $235;
        } else {
         $$5486$lcssa = $$6;$$sink545$lcssa = $235;
         break;
        }
       }
      } else {
       $$5486$lcssa = $$3484$lcssa;$$sink545$lcssa = $210;
      }
      $241 = $$5486$lcssa;
      $242 = (($183) - ($241))|0;
      $243 = $242 >> 2;
      $244 = ($243*9)|0;
      $245 = HEAP32[$$5486$lcssa>>2]|0;
      $246 = ($245>>>0)<(10);
      if ($246) {
       $$4492 = $$sink545$lcssa;$$4518 = $244;$$8 = $$5486$lcssa;
      } else {
       $$2516618 = $244;$$2532617 = 10;
       while(1) {
        $247 = ($$2532617*10)|0;
        $248 = (($$2516618) + 1)|0;
        $249 = ($245>>>0)<($247>>>0);
        if ($249) {
         $$4492 = $$sink545$lcssa;$$4518 = $248;$$8 = $$5486$lcssa;
         break;
        } else {
         $$2516618 = $248;$$2532617 = $247;
        }
       }
      }
     } else {
      $$4492 = $210;$$4518 = $$1515;$$8 = $$3484$lcssa;
     }
    }
    $250 = ((($$4492)) + 4|0);
    $251 = ($$3501$lcssa>>>0)>($250>>>0);
    $$$3501 = $251 ? $250 : $$3501$lcssa;
    $$5519$ph = $$4518;$$7505$ph = $$$3501;$$9$ph = $$8;
   } else {
    $$5519$ph = $$1515;$$7505$ph = $$3501$lcssa;$$9$ph = $$3484$lcssa;
   }
   $$7505 = $$7505$ph;
   while(1) {
    $252 = ($$7505>>>0)>($$9$ph>>>0);
    if (!($252)) {
     $$lcssa673 = 0;
     break;
    }
    $253 = ((($$7505)) + -4|0);
    $254 = HEAP32[$253>>2]|0;
    $255 = ($254|0)==(0);
    if ($255) {
     $$7505 = $253;
    } else {
     $$lcssa673 = 1;
     break;
    }
   }
   $256 = (0 - ($$5519$ph))|0;
   do {
    if ($196) {
     $not$ = $197 ^ 1;
     $257 = $not$&1;
     $$539$ = (($257) + ($$539))|0;
     $258 = ($$539$|0)>($$5519$ph|0);
     $259 = ($$5519$ph|0)>(-5);
     $or$cond6 = $258 & $259;
     if ($or$cond6) {
      $260 = (($5) + -1)|0;
      $$neg567 = (($$539$) + -1)|0;
      $261 = (($$neg567) - ($$5519$ph))|0;
      $$0479 = $260;$$2476 = $261;
     } else {
      $262 = (($5) + -2)|0;
      $263 = (($$539$) + -1)|0;
      $$0479 = $262;$$2476 = $263;
     }
     $264 = $4 & 8;
     $265 = ($264|0)==(0);
     if ($265) {
      if ($$lcssa673) {
       $266 = ((($$7505)) + -4|0);
       $267 = HEAP32[$266>>2]|0;
       $268 = ($267|0)==(0);
       if ($268) {
        $$2529 = 9;
       } else {
        $269 = (($267>>>0) % 10)&-1;
        $270 = ($269|0)==(0);
        if ($270) {
         $$1528614 = 0;$$3533613 = 10;
         while(1) {
          $271 = ($$3533613*10)|0;
          $272 = (($$1528614) + 1)|0;
          $273 = (($267>>>0) % ($271>>>0))&-1;
          $274 = ($273|0)==(0);
          if ($274) {
           $$1528614 = $272;$$3533613 = $271;
          } else {
           $$2529 = $272;
           break;
          }
         }
        } else {
         $$2529 = 0;
        }
       }
      } else {
       $$2529 = 9;
      }
      $275 = $$0479 | 32;
      $276 = ($275|0)==(102);
      $277 = $$7505;
      $278 = (($277) - ($183))|0;
      $279 = $278 >> 2;
      $280 = ($279*9)|0;
      $281 = (($280) + -9)|0;
      if ($276) {
       $282 = (($281) - ($$2529))|0;
       $283 = ($282|0)>(0);
       $$546 = $283 ? $282 : 0;
       $284 = ($$2476|0)<($$546|0);
       $$2476$$547 = $284 ? $$2476 : $$546;
       $$1480 = $$0479;$$3477 = $$2476$$547;$$pre$phi690Z2D = 0;
       break;
      } else {
       $285 = (($281) + ($$5519$ph))|0;
       $286 = (($285) - ($$2529))|0;
       $287 = ($286|0)>(0);
       $$548 = $287 ? $286 : 0;
       $288 = ($$2476|0)<($$548|0);
       $$2476$$549 = $288 ? $$2476 : $$548;
       $$1480 = $$0479;$$3477 = $$2476$$549;$$pre$phi690Z2D = 0;
       break;
      }
     } else {
      $$1480 = $$0479;$$3477 = $$2476;$$pre$phi690Z2D = $264;
     }
    } else {
     $$pre689 = $4 & 8;
     $$1480 = $5;$$3477 = $$539;$$pre$phi690Z2D = $$pre689;
    }
   } while(0);
   $289 = $$3477 | $$pre$phi690Z2D;
   $290 = ($289|0)!=(0);
   $291 = $290&1;
   $292 = $$1480 | 32;
   $293 = ($292|0)==(102);
   if ($293) {
    $294 = ($$5519$ph|0)>(0);
    $295 = $294 ? $$5519$ph : 0;
    $$2513 = 0;$$pn566 = $295;
   } else {
    $296 = ($$5519$ph|0)<(0);
    $297 = $296 ? $256 : $$5519$ph;
    $298 = ($297|0)<(0);
    $299 = $298 << 31 >> 31;
    $300 = (_fmt_u($297,$299,$11)|0);
    $301 = $11;
    $302 = $300;
    $303 = (($301) - ($302))|0;
    $304 = ($303|0)<(2);
    if ($304) {
     $$1512607 = $300;
     while(1) {
      $305 = ((($$1512607)) + -1|0);
      HEAP8[$305>>0] = 48;
      $306 = $305;
      $307 = (($301) - ($306))|0;
      $308 = ($307|0)<(2);
      if ($308) {
       $$1512607 = $305;
      } else {
       $$1512$lcssa = $305;
       break;
      }
     }
    } else {
     $$1512$lcssa = $300;
    }
    $309 = $$5519$ph >> 31;
    $310 = $309 & 2;
    $311 = (($310) + 43)|0;
    $312 = $311&255;
    $313 = ((($$1512$lcssa)) + -1|0);
    HEAP8[$313>>0] = $312;
    $314 = $$1480&255;
    $315 = ((($$1512$lcssa)) + -2|0);
    HEAP8[$315>>0] = $314;
    $316 = $315;
    $317 = (($301) - ($316))|0;
    $$2513 = $315;$$pn566 = $317;
   }
   $318 = (($$0520) + 1)|0;
   $319 = (($318) + ($$3477))|0;
   $$1526 = (($319) + ($291))|0;
   $320 = (($$1526) + ($$pn566))|0;
   _pad_674($0,32,$2,$320,$4);
   _out($0,$$0521,$$0520);
   $321 = $4 ^ 65536;
   _pad_674($0,48,$2,$320,$321);
   if ($293) {
    $322 = ($$9$ph>>>0)>($$556>>>0);
    $$0496$$9 = $322 ? $$556 : $$9$ph;
    $323 = ((($8)) + 9|0);
    $324 = $323;
    $325 = ((($8)) + 8|0);
    $$5493597 = $$0496$$9;
    while(1) {
     $326 = HEAP32[$$5493597>>2]|0;
     $327 = (_fmt_u($326,0,$323)|0);
     $328 = ($$5493597|0)==($$0496$$9|0);
     if ($328) {
      $334 = ($327|0)==($323|0);
      if ($334) {
       HEAP8[$325>>0] = 48;
       $$1465 = $325;
      } else {
       $$1465 = $327;
      }
     } else {
      $329 = ($327>>>0)>($8>>>0);
      if ($329) {
       $330 = $327;
       $331 = (($330) - ($9))|0;
       _memset(($8|0),48,($331|0))|0;
       $$0464594 = $327;
       while(1) {
        $332 = ((($$0464594)) + -1|0);
        $333 = ($332>>>0)>($8>>>0);
        if ($333) {
         $$0464594 = $332;
        } else {
         $$1465 = $332;
         break;
        }
       }
      } else {
       $$1465 = $327;
      }
     }
     $335 = $$1465;
     $336 = (($324) - ($335))|0;
     _out($0,$$1465,$336);
     $337 = ((($$5493597)) + 4|0);
     $338 = ($337>>>0)>($$556>>>0);
     if ($338) {
      break;
     } else {
      $$5493597 = $337;
     }
    }
    $339 = ($289|0)==(0);
    if (!($339)) {
     _out($0,159372,1);
    }
    $340 = ($337>>>0)<($$7505>>>0);
    $341 = ($$3477|0)>(0);
    $342 = $340 & $341;
    if ($342) {
     $$4478590 = $$3477;$$6494589 = $337;
     while(1) {
      $343 = HEAP32[$$6494589>>2]|0;
      $344 = (_fmt_u($343,0,$323)|0);
      $345 = ($344>>>0)>($8>>>0);
      if ($345) {
       $346 = $344;
       $347 = (($346) - ($9))|0;
       _memset(($8|0),48,($347|0))|0;
       $$0463584 = $344;
       while(1) {
        $348 = ((($$0463584)) + -1|0);
        $349 = ($348>>>0)>($8>>>0);
        if ($349) {
         $$0463584 = $348;
        } else {
         $$0463$lcssa = $348;
         break;
        }
       }
      } else {
       $$0463$lcssa = $344;
      }
      $350 = ($$4478590|0)<(9);
      $351 = $350 ? $$4478590 : 9;
      _out($0,$$0463$lcssa,$351);
      $352 = ((($$6494589)) + 4|0);
      $353 = (($$4478590) + -9)|0;
      $354 = ($352>>>0)<($$7505>>>0);
      $355 = ($$4478590|0)>(9);
      $356 = $354 & $355;
      if ($356) {
       $$4478590 = $353;$$6494589 = $352;
      } else {
       $$4478$lcssa = $353;
       break;
      }
     }
    } else {
     $$4478$lcssa = $$3477;
    }
    $357 = (($$4478$lcssa) + 9)|0;
    _pad_674($0,48,$357,9,0);
   } else {
    $358 = ((($$9$ph)) + 4|0);
    $$7505$ = $$lcssa673 ? $$7505 : $358;
    $359 = ($$3477|0)>(-1);
    if ($359) {
     $360 = ((($8)) + 9|0);
     $361 = ($$pre$phi690Z2D|0)==(0);
     $362 = $360;
     $363 = (0 - ($9))|0;
     $364 = ((($8)) + 8|0);
     $$5602 = $$3477;$$7495601 = $$9$ph;
     while(1) {
      $365 = HEAP32[$$7495601>>2]|0;
      $366 = (_fmt_u($365,0,$360)|0);
      $367 = ($366|0)==($360|0);
      if ($367) {
       HEAP8[$364>>0] = 48;
       $$0 = $364;
      } else {
       $$0 = $366;
      }
      $368 = ($$7495601|0)==($$9$ph|0);
      do {
       if ($368) {
        $372 = ((($$0)) + 1|0);
        _out($0,$$0,1);
        $373 = ($$5602|0)<(1);
        $or$cond554 = $361 & $373;
        if ($or$cond554) {
         $$2 = $372;
         break;
        }
        _out($0,159372,1);
        $$2 = $372;
       } else {
        $369 = ($$0>>>0)>($8>>>0);
        if (!($369)) {
         $$2 = $$0;
         break;
        }
        $scevgep684 = (($$0) + ($363)|0);
        $scevgep684685 = $scevgep684;
        _memset(($8|0),48,($scevgep684685|0))|0;
        $$1598 = $$0;
        while(1) {
         $370 = ((($$1598)) + -1|0);
         $371 = ($370>>>0)>($8>>>0);
         if ($371) {
          $$1598 = $370;
         } else {
          $$2 = $370;
          break;
         }
        }
       }
      } while(0);
      $374 = $$2;
      $375 = (($362) - ($374))|0;
      $376 = ($$5602|0)>($375|0);
      $377 = $376 ? $375 : $$5602;
      _out($0,$$2,$377);
      $378 = (($$5602) - ($375))|0;
      $379 = ((($$7495601)) + 4|0);
      $380 = ($379>>>0)<($$7505$>>>0);
      $381 = ($378|0)>(-1);
      $382 = $380 & $381;
      if ($382) {
       $$5602 = $378;$$7495601 = $379;
      } else {
       $$5$lcssa = $378;
       break;
      }
     }
    } else {
     $$5$lcssa = $$3477;
    }
    $383 = (($$5$lcssa) + 18)|0;
    _pad_674($0,48,$383,18,0);
    $384 = $11;
    $385 = $$2513;
    $386 = (($384) - ($385))|0;
    _out($0,$$2513,$386);
   }
   $387 = $4 ^ 8192;
   _pad_674($0,32,$2,$320,$387);
   $$sink562 = $320;
  } else {
   $27 = $5 & 32;
   $28 = ($27|0)!=(0);
   $29 = $28 ? 148684 : 148688;
   $30 = ($$0471 != $$0471) | (0.0 != 0.0);
   $31 = $28 ? 150613 : 148692;
   $$0510 = $30 ? $31 : $29;
   $32 = (($$0520) + 3)|0;
   $33 = $4 & -65537;
   _pad_674($0,32,$2,$32,$33);
   _out($0,$$0521,$$0520);
   _out($0,$$0510,3);
   $34 = $4 ^ 8192;
   _pad_674($0,32,$2,$32,$34);
   $$sink562 = $32;
  }
 } while(0);
 $388 = ($$sink562|0)<($2|0);
 $$555 = $388 ? $2 : $$sink562;
 STACKTOP = sp;return ($$555|0);
}
function ___DOUBLE_BITS_675($0) {
 $0 = +$0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $0;$1 = HEAP32[tempDoublePtr>>2]|0;
 $2 = HEAP32[tempDoublePtr+4>>2]|0;
 tempRet0 = ($2);
 return ($1|0);
}
function _frexpl($0,$1) {
 $0 = +$0;
 $1 = $1|0;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_frexp($0,$1));
 return (+$2);
}
function _frexp($0,$1) {
 $0 = +$0;
 $1 = $1|0;
 var $$0 = 0.0, $$016 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0.0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0.0, $storemerge = 0, $trunc$clear = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $0;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 $4 = (_bitshift64Lshr(($2|0),($3|0),52)|0);
 $5 = tempRet0;
 $6 = $4&65535;
 $trunc$clear = $6 & 2047;
 switch ($trunc$clear<<16>>16) {
 case 0:  {
  $7 = $0 != 0.0;
  if ($7) {
   $8 = $0 * 1.8446744073709552E+19;
   $9 = (+_frexp($8,$1));
   $10 = HEAP32[$1>>2]|0;
   $11 = (($10) + -64)|0;
   $$016 = $9;$storemerge = $11;
  } else {
   $$016 = $0;$storemerge = 0;
  }
  HEAP32[$1>>2] = $storemerge;
  $$0 = $$016;
  break;
 }
 case 2047:  {
  $$0 = $0;
  break;
 }
 default: {
  $12 = $4 & 2047;
  $13 = (($12) + -1022)|0;
  HEAP32[$1>>2] = $13;
  $14 = $3 & -2146435073;
  $15 = $14 | 1071644672;
  HEAP32[tempDoublePtr>>2] = $2;HEAP32[tempDoublePtr+4>>2] = $15;$16 = +HEAPF64[tempDoublePtr>>3];
  $$0 = $16;
 }
 }
 return (+$$0);
}
function _wcrtomb($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==(0|0);
 do {
  if ($3) {
   $$0 = 1;
  } else {
   $4 = ($1>>>0)<(128);
   if ($4) {
    $5 = $1&255;
    HEAP8[$0>>0] = $5;
    $$0 = 1;
    break;
   }
   $6 = (___pthread_self_448()|0);
   $7 = ((($6)) + 188|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = HEAP32[$8>>2]|0;
   $not$ = ($9|0)==(0|0);
   if ($not$) {
    $10 = $1 & -128;
    $11 = ($10|0)==(57216);
    if ($11) {
     $13 = $1&255;
     HEAP8[$0>>0] = $13;
     $$0 = 1;
     break;
    } else {
     $12 = (___errno_location()|0);
     HEAP32[$12>>2] = 84;
     $$0 = -1;
     break;
    }
   }
   $14 = ($1>>>0)<(2048);
   if ($14) {
    $15 = $1 >>> 6;
    $16 = $15 | 192;
    $17 = $16&255;
    $18 = ((($0)) + 1|0);
    HEAP8[$0>>0] = $17;
    $19 = $1 & 63;
    $20 = $19 | 128;
    $21 = $20&255;
    HEAP8[$18>>0] = $21;
    $$0 = 2;
    break;
   }
   $22 = ($1>>>0)<(55296);
   $23 = $1 & -8192;
   $24 = ($23|0)==(57344);
   $or$cond = $22 | $24;
   if ($or$cond) {
    $25 = $1 >>> 12;
    $26 = $25 | 224;
    $27 = $26&255;
    $28 = ((($0)) + 1|0);
    HEAP8[$0>>0] = $27;
    $29 = $1 >>> 6;
    $30 = $29 & 63;
    $31 = $30 | 128;
    $32 = $31&255;
    $33 = ((($0)) + 2|0);
    HEAP8[$28>>0] = $32;
    $34 = $1 & 63;
    $35 = $34 | 128;
    $36 = $35&255;
    HEAP8[$33>>0] = $36;
    $$0 = 3;
    break;
   }
   $37 = (($1) + -65536)|0;
   $38 = ($37>>>0)<(1048576);
   if ($38) {
    $39 = $1 >>> 18;
    $40 = $39 | 240;
    $41 = $40&255;
    $42 = ((($0)) + 1|0);
    HEAP8[$0>>0] = $41;
    $43 = $1 >>> 12;
    $44 = $43 & 63;
    $45 = $44 | 128;
    $46 = $45&255;
    $47 = ((($0)) + 2|0);
    HEAP8[$42>>0] = $46;
    $48 = $1 >>> 6;
    $49 = $48 & 63;
    $50 = $49 | 128;
    $51 = $50&255;
    $52 = ((($0)) + 3|0);
    HEAP8[$47>>0] = $51;
    $53 = $1 & 63;
    $54 = $53 | 128;
    $55 = $54&255;
    HEAP8[$52>>0] = $55;
    $$0 = 4;
    break;
   } else {
    $56 = (___errno_location()|0);
    HEAP32[$56>>2] = 84;
    $$0 = -1;
    break;
   }
  }
 } while(0);
 return ($$0|0);
}
function ___pthread_self_448() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function ___pthread_self_105() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function ___strerror_l($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$012$lcssa = 0, $$01214 = 0, $$016 = 0, $$113 = 0, $$115 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $$016 = 0;
 while(1) {
  $3 = (148712 + ($$016)|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = $4&255;
  $6 = ($5|0)==($0|0);
  if ($6) {
   label = 2;
   break;
  }
  $7 = (($$016) + 1)|0;
  $8 = ($7|0)==(87);
  if ($8) {
   $$01214 = 148800;$$115 = 87;
   label = 5;
   break;
  } else {
   $$016 = $7;
  }
 }
 if ((label|0) == 2) {
  $2 = ($$016|0)==(0);
  if ($2) {
   $$012$lcssa = 148800;
  } else {
   $$01214 = 148800;$$115 = $$016;
   label = 5;
  }
 }
 if ((label|0) == 5) {
  while(1) {
   label = 0;
   $$113 = $$01214;
   while(1) {
    $9 = HEAP8[$$113>>0]|0;
    $10 = ($9<<24>>24)==(0);
    $11 = ((($$113)) + 1|0);
    if ($10) {
     break;
    } else {
     $$113 = $11;
    }
   }
   $12 = (($$115) + -1)|0;
   $13 = ($12|0)==(0);
   if ($13) {
    $$012$lcssa = $11;
    break;
   } else {
    $$01214 = $11;$$115 = $12;
    label = 5;
   }
  }
 }
 $14 = ((($1)) + 20|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (___lctrans($$012$lcssa,$15)|0);
 return ($16|0);
}
function ___lctrans($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (___lctrans_impl($0,$1)|0);
 return ($2|0);
}
function ___lctrans_impl($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)==(0|0);
 if ($2) {
  $$0 = 0;
 } else {
  $3 = HEAP32[$1>>2]|0;
  $4 = ((($1)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (___mo_lookup($3,$5,$0)|0);
  $$0 = $6;
 }
 $7 = ($$0|0)!=(0|0);
 $8 = $7 ? $$0 : $0;
 return ($8|0);
}
function ___mo_lookup($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$090 = 0, $$094 = 0, $$191 = 0, $$195 = 0, $$4 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond102 = 0, $or$cond104 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = HEAP32[$0>>2]|0;
 $4 = (($3) + 1794895138)|0;
 $5 = ((($0)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_swapc($6,$4)|0);
 $8 = ((($0)) + 12|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (_swapc($9,$4)|0);
 $11 = ((($0)) + 16|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (_swapc($12,$4)|0);
 $14 = $1 >>> 2;
 $15 = ($7>>>0)<($14>>>0);
 L1: do {
  if ($15) {
   $16 = $7 << 2;
   $17 = (($1) - ($16))|0;
   $18 = ($10>>>0)<($17>>>0);
   $19 = ($13>>>0)<($17>>>0);
   $or$cond = $18 & $19;
   if ($or$cond) {
    $20 = $13 | $10;
    $21 = $20 & 3;
    $22 = ($21|0)==(0);
    if ($22) {
     $23 = $10 >>> 2;
     $24 = $13 >>> 2;
     $$090 = 0;$$094 = $7;
     while(1) {
      $25 = $$094 >>> 1;
      $26 = (($$090) + ($25))|0;
      $27 = $26 << 1;
      $28 = (($27) + ($23))|0;
      $29 = (($0) + ($28<<2)|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = (_swapc($30,$4)|0);
      $32 = (($28) + 1)|0;
      $33 = (($0) + ($32<<2)|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (_swapc($34,$4)|0);
      $36 = ($35>>>0)<($1>>>0);
      $37 = (($1) - ($35))|0;
      $38 = ($31>>>0)<($37>>>0);
      $or$cond102 = $36 & $38;
      if (!($or$cond102)) {
       $$4 = 0;
       break L1;
      }
      $39 = (($35) + ($31))|0;
      $40 = (($0) + ($39)|0);
      $41 = HEAP8[$40>>0]|0;
      $42 = ($41<<24>>24)==(0);
      if (!($42)) {
       $$4 = 0;
       break L1;
      }
      $43 = (($0) + ($35)|0);
      $44 = (_strcmp($2,$43)|0);
      $45 = ($44|0)==(0);
      if ($45) {
       break;
      }
      $62 = ($$094|0)==(1);
      $63 = ($44|0)<(0);
      $64 = (($$094) - ($25))|0;
      $$195 = $63 ? $25 : $64;
      $$191 = $63 ? $$090 : $26;
      if ($62) {
       $$4 = 0;
       break L1;
      } else {
       $$090 = $$191;$$094 = $$195;
      }
     }
     $46 = (($27) + ($24))|0;
     $47 = (($0) + ($46<<2)|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = (_swapc($48,$4)|0);
     $50 = (($46) + 1)|0;
     $51 = (($0) + ($50<<2)|0);
     $52 = HEAP32[$51>>2]|0;
     $53 = (_swapc($52,$4)|0);
     $54 = ($53>>>0)<($1>>>0);
     $55 = (($1) - ($53))|0;
     $56 = ($49>>>0)<($55>>>0);
     $or$cond104 = $54 & $56;
     if ($or$cond104) {
      $57 = (($0) + ($53)|0);
      $58 = (($53) + ($49))|0;
      $59 = (($0) + ($58)|0);
      $60 = HEAP8[$59>>0]|0;
      $61 = ($60<<24>>24)==(0);
      $$ = $61 ? $57 : 0;
      $$4 = $$;
     } else {
      $$4 = 0;
     }
    } else {
     $$4 = 0;
    }
   } else {
    $$4 = 0;
   }
  } else {
   $$4 = 0;
  }
 } while(0);
 return ($$4|0);
}
function _swapc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$ = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)==(0);
 $3 = (_llvm_bswap_i32(($0|0))|0);
 $$ = $2 ? $0 : $3;
 return ($$|0);
}
function ___fwritex($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$038 = 0, $$042 = 0, $$1 = 0, $$139 = 0, $$141 = 0, $$143 = 0, $$pre = 0, $$pre47 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($2)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==(0|0);
 if ($5) {
  $7 = (___towrite($2)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   $$pre = HEAP32[$3>>2]|0;
   $12 = $$pre;
   label = 5;
  } else {
   $$1 = 0;
  }
 } else {
  $6 = $4;
  $12 = $6;
  label = 5;
 }
 L5: do {
  if ((label|0) == 5) {
   $9 = ((($2)) + 20|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = (($12) - ($10))|0;
   $13 = ($11>>>0)<($1>>>0);
   $14 = $10;
   if ($13) {
    $15 = ((($2)) + 36|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = (FUNCTION_TABLE_iiii[$16 & 2047]($2,$0,$1)|0);
    $$1 = $17;
    break;
   }
   $18 = ((($2)) + 75|0);
   $19 = HEAP8[$18>>0]|0;
   $20 = ($19<<24>>24)>(-1);
   L10: do {
    if ($20) {
     $$038 = $1;
     while(1) {
      $21 = ($$038|0)==(0);
      if ($21) {
       $$139 = 0;$$141 = $0;$$143 = $1;$31 = $14;
       break L10;
      }
      $22 = (($$038) + -1)|0;
      $23 = (($0) + ($22)|0);
      $24 = HEAP8[$23>>0]|0;
      $25 = ($24<<24>>24)==(10);
      if ($25) {
       break;
      } else {
       $$038 = $22;
      }
     }
     $26 = ((($2)) + 36|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = (FUNCTION_TABLE_iiii[$27 & 2047]($2,$0,$$038)|0);
     $29 = ($28>>>0)<($$038>>>0);
     if ($29) {
      $$1 = $28;
      break L5;
     }
     $30 = (($0) + ($$038)|0);
     $$042 = (($1) - ($$038))|0;
     $$pre47 = HEAP32[$9>>2]|0;
     $$139 = $$038;$$141 = $30;$$143 = $$042;$31 = $$pre47;
    } else {
     $$139 = 0;$$141 = $0;$$143 = $1;$31 = $14;
    }
   } while(0);
   _memcpy(($31|0),($$141|0),($$143|0))|0;
   $32 = HEAP32[$9>>2]|0;
   $33 = (($32) + ($$143)|0);
   HEAP32[$9>>2] = $33;
   $34 = (($$139) + ($$143))|0;
   $$1 = $34;
  }
 } while(0);
 return ($$1|0);
}
function ___towrite($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 74|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = $2 << 24 >> 24;
 $4 = (($3) + 255)|0;
 $5 = $4 | $3;
 $6 = $5&255;
 HEAP8[$1>>0] = $6;
 $7 = HEAP32[$0>>2]|0;
 $8 = $7 & 8;
 $9 = ($8|0)==(0);
 if ($9) {
  $11 = ((($0)) + 8|0);
  HEAP32[$11>>2] = 0;
  $12 = ((($0)) + 4|0);
  HEAP32[$12>>2] = 0;
  $13 = ((($0)) + 44|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = ((($0)) + 28|0);
  HEAP32[$15>>2] = $14;
  $16 = ((($0)) + 20|0);
  HEAP32[$16>>2] = $14;
  $17 = ((($0)) + 48|0);
  $18 = HEAP32[$17>>2]|0;
  $19 = (($14) + ($18)|0);
  $20 = ((($0)) + 16|0);
  HEAP32[$20>>2] = $19;
  $$0 = 0;
 } else {
  $10 = $7 | 32;
  HEAP32[$0>>2] = $10;
  $$0 = -1;
 }
 return ($$0|0);
}
function _sn_write($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $10 = 0, $11 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 16|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($0)) + 20|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6;
 $8 = (($4) - ($7))|0;
 $9 = ($8>>>0)>($2>>>0);
 $$ = $9 ? $2 : $8;
 _memcpy(($6|0),($1|0),($$|0))|0;
 $10 = HEAP32[$5>>2]|0;
 $11 = (($10) + ($$)|0);
 HEAP32[$5>>2] = $11;
 return ($2|0);
}
function _wcschr($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$09 = 0, $$not = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)==(0);
 if ($2) {
  $3 = (_wcslen($0)|0);
  $4 = (($0) + ($3<<2)|0);
  $$0 = $4;
 } else {
  $$09 = $0;
  while(1) {
   $5 = HEAP32[$$09>>2]|0;
   $6 = ($5|0)!=(0);
   $$not = $6 ^ 1;
   $7 = ($5|0)==($1|0);
   $or$cond = $7 | $$not;
   $8 = ((($$09)) + 4|0);
   if ($or$cond) {
    break;
   } else {
    $$09 = $8;
   }
  }
  $9 = $6 ? $$09 : 0;
  $$0 = $9;
 }
 return ($$0|0);
}
function _wcslen($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $0;
 while(1) {
  $1 = HEAP32[$$0>>2]|0;
  $2 = ($1|0)==(0);
  $3 = ((($$0)) + 4|0);
  if ($2) {
   break;
  } else {
   $$0 = $3;
  }
 }
 $4 = $$0;
 $5 = $0;
 $6 = (($4) - ($5))|0;
 $7 = $6 >> 2;
 return ($7|0);
}
function _isxdigit($0) {
 $0 = $0|0;
 var $$ = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $isdigit = 0, $isdigittmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isdigittmp = (($0) + -48)|0;
 $isdigit = ($isdigittmp>>>0)<(10);
 $1 = $0 | 32;
 $2 = (($1) + -97)|0;
 $3 = ($2>>>0)<(6);
 $$ = $isdigit | $3;
 $4 = $$&1;
 return ($4|0);
}
function ___ctype_b_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (15384|0);
}
function ___ctype_get_mb_cur_max() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___pthread_self()|0);
 $1 = ((($0)) + 188|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)!=(0|0);
 $5 = $4 ? 4 : 1;
 return ($5|0);
}
function ___pthread_self() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function ___ctype_tolower_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (15388|0);
}
function ___ctype_toupper_loc() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (16928|0);
}
function _close($0) {
 $0 = $0|0;
 var $$ = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $1 = (_dummy_738($0)|0);
 HEAP32[$vararg_buffer>>2] = $1;
 $2 = (___syscall6(6,($vararg_buffer|0))|0);
 $3 = ($2|0)==(-4);
 $$ = $3 ? 0 : $2;
 $4 = (___syscall_ret($$)|0);
 STACKTOP = sp;return ($4|0);
}
function ___floatscan($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0105$ph = 0, $$0106$ph = 0, $$0107$lcssa = 0, $$0107127 = 0, $$0113 = 0, $$0114 = 0.0, $$1$lcssa = 0, $$1108 = 0, $$1128 = 0, $$2 = 0, $$2109125 = 0, $$3110 = 0, $$3126 = 0, $$4 = 0, $$4111 = 0, $$5 = 0, $$6 = 0, $$in = 0, $$old8 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0.0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0.0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0.0, $54 = 0.0, $55 = 0.0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond5 = 0, $or$cond7 = 0, $or$cond9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 switch ($1|0) {
 case 0:  {
  $$0105$ph = -149;$$0106$ph = 24;
  label = 4;
  break;
 }
 case 1:  {
  $$0105$ph = -1074;$$0106$ph = 53;
  label = 4;
  break;
 }
 case 2:  {
  $$0105$ph = -1074;$$0106$ph = 53;
  label = 4;
  break;
 }
 default: {
  $$0114 = 0.0;
 }
 }
 L4: do {
  if ((label|0) == 4) {
   $3 = ((($0)) + 4|0);
   $4 = ((($0)) + 100|0);
   while(1) {
    $5 = HEAP32[$3>>2]|0;
    $6 = HEAP32[$4>>2]|0;
    $7 = ($5>>>0)<($6>>>0);
    if ($7) {
     $8 = ((($5)) + 1|0);
     HEAP32[$3>>2] = $8;
     $9 = HEAP8[$5>>0]|0;
     $10 = $9&255;
     $12 = $10;
    } else {
     $11 = (___shgetc($0)|0);
     $12 = $11;
    }
    $13 = (_isspace($12)|0);
    $14 = ($13|0)==(0);
    if ($14) {
     break;
    }
   }
   L13: do {
    switch ($12|0) {
    case 43: case 45:  {
     $15 = ($12|0)==(45);
     $16 = $15&1;
     $17 = $16 << 1;
     $18 = (1 - ($17))|0;
     $19 = HEAP32[$3>>2]|0;
     $20 = HEAP32[$4>>2]|0;
     $21 = ($19>>>0)<($20>>>0);
     if ($21) {
      $22 = ((($19)) + 1|0);
      HEAP32[$3>>2] = $22;
      $23 = HEAP8[$19>>0]|0;
      $24 = $23&255;
      $$0 = $24;$$0113 = $18;
      break L13;
     } else {
      $25 = (___shgetc($0)|0);
      $$0 = $25;$$0113 = $18;
      break L13;
     }
     break;
    }
    default: {
     $$0 = $12;$$0113 = 1;
    }
    }
   } while(0);
   $$0107127 = 0;$$1128 = $$0;
   while(1) {
    $26 = $$1128 | 32;
    $27 = (150604 + ($$0107127)|0);
    $28 = HEAP8[$27>>0]|0;
    $29 = $28 << 24 >> 24;
    $30 = ($26|0)==($29|0);
    if (!($30)) {
     $$0107$lcssa = $$0107127;$$1$lcssa = $$1128;
     break;
    }
    $31 = ($$0107127>>>0)<(7);
    do {
     if ($31) {
      $32 = HEAP32[$3>>2]|0;
      $33 = HEAP32[$4>>2]|0;
      $34 = ($32>>>0)<($33>>>0);
      if ($34) {
       $35 = ((($32)) + 1|0);
       HEAP32[$3>>2] = $35;
       $36 = HEAP8[$32>>0]|0;
       $37 = $36&255;
       $$2 = $37;
       break;
      } else {
       $38 = (___shgetc($0)|0);
       $$2 = $38;
       break;
      }
     } else {
      $$2 = $$1128;
     }
    } while(0);
    $39 = (($$0107127) + 1)|0;
    $40 = ($39>>>0)<(8);
    if ($40) {
     $$0107127 = $39;$$1128 = $$2;
    } else {
     $$0107$lcssa = $39;$$1$lcssa = $$2;
     break;
    }
   }
   L29: do {
    switch ($$0107$lcssa|0) {
    case 8:  {
     break;
    }
    case 3:  {
     label = 23;
     break;
    }
    default: {
     $41 = ($$0107$lcssa>>>0)>(3);
     $42 = ($2|0)!=(0);
     $or$cond5 = $42 & $41;
     if ($or$cond5) {
      $43 = ($$0107$lcssa|0)==(8);
      if ($43) {
       break L29;
      } else {
       label = 23;
       break L29;
      }
     }
     $56 = ($$0107$lcssa|0)==(0);
     L34: do {
      if ($56) {
       $$2109125 = 0;$$3126 = $$1$lcssa;
       while(1) {
        $57 = $$3126 | 32;
        $58 = (150613 + ($$2109125)|0);
        $59 = HEAP8[$58>>0]|0;
        $60 = $59 << 24 >> 24;
        $61 = ($57|0)==($60|0);
        if (!($61)) {
         $$3110 = $$2109125;$$5 = $$3126;
         break L34;
        }
        $62 = ($$2109125>>>0)<(2);
        do {
         if ($62) {
          $63 = HEAP32[$3>>2]|0;
          $64 = HEAP32[$4>>2]|0;
          $65 = ($63>>>0)<($64>>>0);
          if ($65) {
           $66 = ((($63)) + 1|0);
           HEAP32[$3>>2] = $66;
           $67 = HEAP8[$63>>0]|0;
           $68 = $67&255;
           $$4 = $68;
           break;
          } else {
           $69 = (___shgetc($0)|0);
           $$4 = $69;
           break;
          }
         } else {
          $$4 = $$3126;
         }
        } while(0);
        $70 = (($$2109125) + 1)|0;
        $71 = ($70>>>0)<(3);
        if ($71) {
         $$2109125 = $70;$$3126 = $$4;
        } else {
         $$3110 = $70;$$5 = $$4;
         break;
        }
       }
      } else {
       $$3110 = $$0107$lcssa;$$5 = $$1$lcssa;
      }
     } while(0);
     switch ($$3110|0) {
     case 3:  {
      $72 = HEAP32[$3>>2]|0;
      $73 = HEAP32[$4>>2]|0;
      $74 = ($72>>>0)<($73>>>0);
      if ($74) {
       $75 = ((($72)) + 1|0);
       HEAP32[$3>>2] = $75;
       $76 = HEAP8[$72>>0]|0;
       $77 = $76&255;
       $80 = $77;
      } else {
       $78 = (___shgetc($0)|0);
       $80 = $78;
      }
      $79 = ($80|0)==(40);
      if ($79) {
       $$4111 = 1;
      } else {
       $81 = HEAP32[$4>>2]|0;
       $82 = ($81|0)==(0|0);
       if ($82) {
        $$0114 = nan;
        break L4;
       }
       $83 = HEAP32[$3>>2]|0;
       $84 = ((($83)) + -1|0);
       HEAP32[$3>>2] = $84;
       $$0114 = nan;
       break L4;
      }
      while(1) {
       $85 = HEAP32[$3>>2]|0;
       $86 = HEAP32[$4>>2]|0;
       $87 = ($85>>>0)<($86>>>0);
       if ($87) {
        $88 = ((($85)) + 1|0);
        HEAP32[$3>>2] = $88;
        $89 = HEAP8[$85>>0]|0;
        $90 = $89&255;
        $93 = $90;
       } else {
        $91 = (___shgetc($0)|0);
        $93 = $91;
       }
       $92 = (($93) + -48)|0;
       $94 = ($92>>>0)<(10);
       $95 = (($93) + -65)|0;
       $96 = ($95>>>0)<(26);
       $or$cond = $94 | $96;
       if (!($or$cond)) {
        $97 = (($93) + -97)|0;
        $98 = ($97>>>0)<(26);
        $99 = ($93|0)==(95);
        $or$cond7 = $99 | $98;
        if (!($or$cond7)) {
         break;
        }
       }
       $111 = (($$4111) + 1)|0;
       $$4111 = $111;
      }
      $100 = ($93|0)==(41);
      if ($100) {
       $$0114 = nan;
       break L4;
      }
      $101 = HEAP32[$4>>2]|0;
      $102 = ($101|0)==(0|0);
      if (!($102)) {
       $103 = HEAP32[$3>>2]|0;
       $104 = ((($103)) + -1|0);
       HEAP32[$3>>2] = $104;
      }
      if (!($42)) {
       $106 = (___errno_location()|0);
       HEAP32[$106>>2] = 22;
       ___shlim($0,0);
       $$0114 = 0.0;
       break L4;
      }
      $105 = ($$4111|0)==(0);
      if ($105) {
       $$0114 = nan;
       break L4;
      } else {
       $$in = $$4111;
      }
      while(1) {
       $107 = (($$in) + -1)|0;
       if (!($102)) {
        $108 = HEAP32[$3>>2]|0;
        $109 = ((($108)) + -1|0);
        HEAP32[$3>>2] = $109;
       }
       $110 = ($107|0)==(0);
       if ($110) {
        $$0114 = nan;
        break L4;
       } else {
        $$in = $107;
       }
      }
      break;
     }
     case 0:  {
      $117 = ($$5|0)==(48);
      if ($117) {
       $118 = HEAP32[$3>>2]|0;
       $119 = HEAP32[$4>>2]|0;
       $120 = ($118>>>0)<($119>>>0);
       if ($120) {
        $121 = ((($118)) + 1|0);
        HEAP32[$3>>2] = $121;
        $122 = HEAP8[$118>>0]|0;
        $123 = $122&255;
        $126 = $123;
       } else {
        $124 = (___shgetc($0)|0);
        $126 = $124;
       }
       $125 = $126 | 32;
       $127 = ($125|0)==(120);
       if ($127) {
        $128 = (+_hexfloat($0,$$0106$ph,$$0105$ph,$$0113,$2));
        $$0114 = $128;
        break L4;
       }
       $129 = HEAP32[$4>>2]|0;
       $130 = ($129|0)==(0|0);
       if ($130) {
        $$6 = 48;
       } else {
        $131 = HEAP32[$3>>2]|0;
        $132 = ((($131)) + -1|0);
        HEAP32[$3>>2] = $132;
        $$6 = 48;
       }
      } else {
       $$6 = $$5;
      }
      $133 = (+_decfloat($0,$$6,$$0106$ph,$$0105$ph,$$0113,$2));
      $$0114 = $133;
      break L4;
      break;
     }
     default: {
      $112 = HEAP32[$4>>2]|0;
      $113 = ($112|0)==(0|0);
      if (!($113)) {
       $114 = HEAP32[$3>>2]|0;
       $115 = ((($114)) + -1|0);
       HEAP32[$3>>2] = $115;
      }
      $116 = (___errno_location()|0);
      HEAP32[$116>>2] = 22;
      ___shlim($0,0);
      $$0114 = 0.0;
      break L4;
     }
     }
    }
    }
   } while(0);
   if ((label|0) == 23) {
    $44 = HEAP32[$4>>2]|0;
    $45 = ($44|0)==(0|0);
    if (!($45)) {
     $46 = HEAP32[$3>>2]|0;
     $47 = ((($46)) + -1|0);
     HEAP32[$3>>2] = $47;
    }
    $48 = ($2|0)!=(0);
    $49 = ($$0107$lcssa>>>0)>(3);
    $or$cond9 = $48 & $49;
    if ($or$cond9) {
     $$1108 = $$0107$lcssa;
     while(1) {
      if (!($45)) {
       $50 = HEAP32[$3>>2]|0;
       $51 = ((($50)) + -1|0);
       HEAP32[$3>>2] = $51;
      }
      $52 = (($$1108) + -1)|0;
      $$old8 = ($52>>>0)>(3);
      if ($$old8) {
       $$1108 = $52;
      } else {
       break;
      }
     }
    }
   }
   $53 = (+($$0113|0));
   $54 = $53 * inf;
   $55 = $54;
   $$0114 = $55;
  }
 } while(0);
 return (+$$0114);
}
function _hexfloat($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$0133 = 0, $$0142 = 0, $$0146 = 0, $$0148 = 0, $$0148$ = 0, $$0151 = 0.0, $$0152 = 0.0, $$0155 = 0.0, $$0155$ = 0.0, $$0159 = 0, $$0165 = 0.0, $$0166 = 0, $$0166169 = 0, $$0166170 = 0, $$1$ph = 0, $$1147 = 0, $$1149 = 0, $$1153 = 0.0, $$1156 = 0.0;
 var $$1160 = 0, $$2 = 0, $$2$lcssa = 0, $$2144 = 0, $$2150 = 0, $$2154 = 0.0, $$2157 = 0.0, $$2161 = 0, $$3145 = 0, $$3158$lcssa = 0.0, $$3158179 = 0.0, $$3162$lcssa = 0, $$3162183 = 0, $$4 = 0.0, $$4163$lcssa = 0, $$4163178 = 0, $$5 = 0.0, $$5164 = 0, $$6 = 0, $$pn = 0.0;
 var $$pre = 0.0, $$pre$phiZ2D = 0.0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0.0, $143 = 0.0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0.0, $154 = 0.0, $155 = 0.0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0.0, $167 = 0.0, $168 = 0.0, $169 = 0, $17 = 0;
 var $170 = 0, $171 = 0.0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0.0, $197 = 0, $198 = 0.0, $199 = 0.0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0.0, $206 = 0.0;
 var $207 = 0.0, $208 = 0.0, $209 = 0.0, $21 = 0, $210 = 0.0, $211 = 0, $212 = 0, $213 = 0.0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0.0, $67 = 0.0;
 var $68 = 0.0, $69 = 0.0, $7 = 0, $70 = 0, $71 = 0, $72 = 0.0, $73 = 0.0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0.0, $96 = 0.0, $97 = 0, $98 = 0, $99 = 0, $not$ = 0, $or$cond = 0, $or$cond168 = 0, $or$cond206 = 0, $or$cond4 = 0;
 var $or$cond6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($0)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = ((($0)) + 100|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($6>>>0)<($8>>>0);
 if ($9) {
  $10 = ((($6)) + 1|0);
  HEAP32[$5>>2] = $10;
  $11 = HEAP8[$6>>0]|0;
  $12 = $11&255;
  $$0 = $12;$$0142 = 0;
 } else {
  $13 = (___shgetc($0)|0);
  $$0 = $13;$$0142 = 0;
 }
 L4: while(1) {
  switch ($$0|0) {
  case 46:  {
   label = 8;
   break L4;
   break;
  }
  case 48:  {
   break;
  }
  default: {
   $$0146 = 0;$$0148 = 0;$$0152 = 1.0;$$0155 = 0.0;$$0159 = 0;$$2 = $$0;$$2144 = $$0142;$101 = 0;$53 = 0;$55 = 0;$99 = 0;
   break L4;
  }
  }
  $14 = HEAP32[$5>>2]|0;
  $15 = HEAP32[$7>>2]|0;
  $16 = ($14>>>0)<($15>>>0);
  if ($16) {
   $17 = ((($14)) + 1|0);
   HEAP32[$5>>2] = $17;
   $18 = HEAP8[$14>>0]|0;
   $19 = $18&255;
   $$0 = $19;$$0142 = 1;
   continue;
  } else {
   $20 = (___shgetc($0)|0);
   $$0 = $20;$$0142 = 1;
   continue;
  }
 }
 if ((label|0) == 8) {
  $21 = HEAP32[$5>>2]|0;
  $22 = HEAP32[$7>>2]|0;
  $23 = ($21>>>0)<($22>>>0);
  if ($23) {
   $24 = ((($21)) + 1|0);
   HEAP32[$5>>2] = $24;
   $25 = HEAP8[$21>>0]|0;
   $26 = $25&255;
   $$1$ph = $26;
  } else {
   $27 = (___shgetc($0)|0);
   $$1$ph = $27;
  }
  $28 = ($$1$ph|0)==(48);
  if ($28) {
   $36 = 0;$37 = 0;
   while(1) {
    $29 = HEAP32[$5>>2]|0;
    $30 = HEAP32[$7>>2]|0;
    $31 = ($29>>>0)<($30>>>0);
    if ($31) {
     $32 = ((($29)) + 1|0);
     HEAP32[$5>>2] = $32;
     $33 = HEAP8[$29>>0]|0;
     $34 = $33&255;
     $41 = $34;
    } else {
     $35 = (___shgetc($0)|0);
     $41 = $35;
    }
    $38 = (_i64Add(($36|0),($37|0),-1,-1)|0);
    $39 = tempRet0;
    $40 = ($41|0)==(48);
    if ($40) {
     $36 = $38;$37 = $39;
    } else {
     $$0146 = 1;$$0148 = 0;$$0152 = 1.0;$$0155 = 0.0;$$0159 = 0;$$2 = $41;$$2144 = 1;$101 = $39;$53 = 0;$55 = 0;$99 = $38;
     break;
    }
   }
  } else {
   $$0146 = 1;$$0148 = 0;$$0152 = 1.0;$$0155 = 0.0;$$0159 = 0;$$2 = $$1$ph;$$2144 = $$0142;$101 = 0;$53 = 0;$55 = 0;$99 = 0;
  }
 }
 while(1) {
  $42 = (($$2) + -48)|0;
  $43 = ($42>>>0)<(10);
  $44 = ($$2|0)==(46);
  if (!($43)) {
   $45 = $$2 | 32;
   $46 = (($45) + -97)|0;
   $47 = ($46>>>0)<(6);
   $or$cond6 = $44 | $47;
   if (!($or$cond6)) {
    $$2$lcssa = $$2;
    break;
   }
  }
  if ($44) {
   $48 = ($$0146|0)==(0);
   if ($48) {
    $$1147 = 1;$$2150 = $$0148;$$2154 = $$0152;$$2157 = $$0155;$$2161 = $$0159;$$3145 = $$2144;$214 = $55;$215 = $53;$216 = $55;$217 = $53;
   } else {
    $$2$lcssa = 46;
    break;
   }
  } else {
   $49 = ($$2|0)>(57);
   $50 = $$2 | 32;
   $51 = (($50) + -87)|0;
   $$0133 = $49 ? $51 : $42;
   $52 = ($53|0)<(0);
   $54 = ($55>>>0)<(8);
   $56 = ($53|0)==(0);
   $57 = $56 & $54;
   $58 = $52 | $57;
   do {
    if ($58) {
     $59 = $$0159 << 4;
     $60 = (($$0133) + ($59))|0;
     $$1149 = $$0148;$$1153 = $$0152;$$1156 = $$0155;$$1160 = $60;
    } else {
     $61 = ($53|0)<(0);
     $62 = ($55>>>0)<(14);
     $63 = ($53|0)==(0);
     $64 = $63 & $62;
     $65 = $61 | $64;
     if ($65) {
      $66 = (+($$0133|0));
      $67 = $$0152 * 0.0625;
      $68 = $67 * $66;
      $69 = $$0155 + $68;
      $$1149 = $$0148;$$1153 = $67;$$1156 = $69;$$1160 = $$0159;
      break;
     } else {
      $70 = ($$0133|0)==(0);
      $71 = ($$0148|0)!=(0);
      $or$cond = $71 | $70;
      $72 = $$0152 * 0.5;
      $73 = $$0155 + $72;
      $$0155$ = $or$cond ? $$0155 : $73;
      $$0148$ = $or$cond ? $$0148 : 1;
      $$1149 = $$0148$;$$1153 = $$0152;$$1156 = $$0155$;$$1160 = $$0159;
      break;
     }
    }
   } while(0);
   $74 = (_i64Add(($55|0),($53|0),1,0)|0);
   $75 = tempRet0;
   $$1147 = $$0146;$$2150 = $$1149;$$2154 = $$1153;$$2157 = $$1156;$$2161 = $$1160;$$3145 = 1;$214 = $99;$215 = $101;$216 = $74;$217 = $75;
  }
  $76 = HEAP32[$5>>2]|0;
  $77 = HEAP32[$7>>2]|0;
  $78 = ($76>>>0)<($77>>>0);
  if ($78) {
   $79 = ((($76)) + 1|0);
   HEAP32[$5>>2] = $79;
   $80 = HEAP8[$76>>0]|0;
   $81 = $80&255;
   $$0146 = $$1147;$$0148 = $$2150;$$0152 = $$2154;$$0155 = $$2157;$$0159 = $$2161;$$2 = $81;$$2144 = $$3145;$101 = $215;$53 = $217;$55 = $216;$99 = $214;
   continue;
  } else {
   $82 = (___shgetc($0)|0);
   $$0146 = $$1147;$$0148 = $$2150;$$0152 = $$2154;$$0155 = $$2157;$$0159 = $$2161;$$2 = $82;$$2144 = $$3145;$101 = $215;$53 = $217;$55 = $216;$99 = $214;
   continue;
  }
 }
 $83 = ($$2144|0)==(0);
 do {
  if ($83) {
   $84 = HEAP32[$7>>2]|0;
   $85 = ($84|0)!=(0|0);
   if ($85) {
    $86 = HEAP32[$5>>2]|0;
    $87 = ((($86)) + -1|0);
    HEAP32[$5>>2] = $87;
   }
   $88 = ($4|0)==(0);
   if ($88) {
    ___shlim($0,0);
   } else {
    if ($85) {
     $89 = HEAP32[$5>>2]|0;
     $90 = ((($89)) + -1|0);
     HEAP32[$5>>2] = $90;
    }
    $91 = ($$0146|0)==(0);
    $92 = ($84|0)==(0|0);
    $or$cond206 = $91 | $92;
    if (!($or$cond206)) {
     $93 = HEAP32[$5>>2]|0;
     $94 = ((($93)) + -1|0);
     HEAP32[$5>>2] = $94;
    }
   }
   $95 = (+($3|0));
   $96 = $95 * 0.0;
   $$0165 = $96;
  } else {
   $97 = ($$0146|0)==(0);
   $98 = $97 ? $55 : $99;
   $100 = $97 ? $53 : $101;
   $102 = ($53|0)<(0);
   $103 = ($55>>>0)<(8);
   $104 = ($53|0)==(0);
   $105 = $104 & $103;
   $106 = $102 | $105;
   if ($106) {
    $$3162183 = $$0159;$108 = $55;$109 = $53;
    while(1) {
     $107 = $$3162183 << 4;
     $110 = (_i64Add(($108|0),($109|0),1,0)|0);
     $111 = tempRet0;
     $112 = ($111|0)<(0);
     $113 = ($110>>>0)<(8);
     $114 = ($111|0)==(0);
     $115 = $114 & $113;
     $116 = $112 | $115;
     if ($116) {
      $$3162183 = $107;$108 = $110;$109 = $111;
     } else {
      $$3162$lcssa = $107;
      break;
     }
    }
   } else {
    $$3162$lcssa = $$0159;
   }
   $117 = $$2$lcssa | 32;
   $118 = ($117|0)==(112);
   if ($118) {
    $119 = (_scanexp($0,$4)|0);
    $120 = tempRet0;
    $121 = ($119|0)==(0);
    $122 = ($120|0)==(-2147483648);
    $123 = $121 & $122;
    if ($123) {
     $124 = ($4|0)==(0);
     if ($124) {
      ___shlim($0,0);
      $$0165 = 0.0;
      break;
     }
     $125 = HEAP32[$7>>2]|0;
     $126 = ($125|0)==(0|0);
     if ($126) {
      $137 = 0;$138 = 0;
     } else {
      $127 = HEAP32[$5>>2]|0;
      $128 = ((($127)) + -1|0);
      HEAP32[$5>>2] = $128;
      $137 = 0;$138 = 0;
     }
    } else {
     $137 = $119;$138 = $120;
    }
   } else {
    $129 = HEAP32[$7>>2]|0;
    $130 = ($129|0)==(0|0);
    if ($130) {
     $137 = 0;$138 = 0;
    } else {
     $131 = HEAP32[$5>>2]|0;
     $132 = ((($131)) + -1|0);
     HEAP32[$5>>2] = $132;
     $137 = 0;$138 = 0;
    }
   }
   $133 = (_bitshift64Shl(($98|0),($100|0),2)|0);
   $134 = tempRet0;
   $135 = (_i64Add(($133|0),($134|0),-32,-1)|0);
   $136 = tempRet0;
   $139 = (_i64Add(($135|0),($136|0),($137|0),($138|0))|0);
   $140 = tempRet0;
   $141 = ($$3162$lcssa|0)==(0);
   if ($141) {
    $142 = (+($3|0));
    $143 = $142 * 0.0;
    $$0165 = $143;
    break;
   }
   $144 = (0 - ($2))|0;
   $145 = ($144|0)<(0);
   $146 = $145 << 31 >> 31;
   $147 = ($140|0)>($146|0);
   $148 = ($139>>>0)>($144>>>0);
   $149 = ($140|0)==($146|0);
   $150 = $149 & $148;
   $151 = $147 | $150;
   if ($151) {
    $152 = (___errno_location()|0);
    HEAP32[$152>>2] = 34;
    $153 = (+($3|0));
    $154 = $153 * 1.7976931348623157E+308;
    $155 = $154 * 1.7976931348623157E+308;
    $$0165 = $155;
    break;
   }
   $156 = (($2) + -106)|0;
   $157 = ($156|0)<(0);
   $158 = $157 << 31 >> 31;
   $159 = ($140|0)<($158|0);
   $160 = ($139>>>0)<($156>>>0);
   $161 = ($140|0)==($158|0);
   $162 = $161 & $160;
   $163 = $159 | $162;
   if ($163) {
    $165 = (___errno_location()|0);
    HEAP32[$165>>2] = 34;
    $166 = (+($3|0));
    $167 = $166 * 2.2250738585072014E-308;
    $168 = $167 * 2.2250738585072014E-308;
    $$0165 = $168;
    break;
   }
   $164 = ($$3162$lcssa|0)>(-1);
   if ($164) {
    $$3158179 = $$0155;$$4163178 = $$3162$lcssa;$173 = $139;$174 = $140;
    while(1) {
     $169 = !($$3158179 >= 0.5);
     $170 = $$4163178 << 1;
     $171 = $$3158179 + -1.0;
     $not$ = $169 ^ 1;
     $172 = $not$&1;
     $$5164 = $170 | $172;
     $$pn = $169 ? $$3158179 : $171;
     $$4 = $$3158179 + $$pn;
     $175 = (_i64Add(($173|0),($174|0),-1,-1)|0);
     $176 = tempRet0;
     $177 = ($$5164|0)>(-1);
     if ($177) {
      $$3158179 = $$4;$$4163178 = $$5164;$173 = $175;$174 = $176;
     } else {
      $$3158$lcssa = $$4;$$4163$lcssa = $$5164;$184 = $175;$185 = $176;
      break;
     }
    }
   } else {
    $$3158$lcssa = $$0155;$$4163$lcssa = $$3162$lcssa;$184 = $139;$185 = $140;
   }
   $178 = ($1|0)<(0);
   $179 = $178 << 31 >> 31;
   $180 = ($2|0)<(0);
   $181 = $180 << 31 >> 31;
   $182 = (_i64Subtract(32,0,($2|0),($181|0))|0);
   $183 = tempRet0;
   $186 = (_i64Add(($182|0),($183|0),($184|0),($185|0))|0);
   $187 = tempRet0;
   $188 = ($179|0)>($187|0);
   $189 = ($1>>>0)>($186>>>0);
   $190 = ($179|0)==($187|0);
   $191 = $190 & $189;
   $192 = $188 | $191;
   if ($192) {
    $193 = ($186|0)>(0);
    if ($193) {
     $$0166 = $186;
     label = 59;
    } else {
     $$0166170 = 0;$197 = 84;
     label = 61;
    }
   } else {
    $$0166 = $1;
    label = 59;
   }
   if ((label|0) == 59) {
    $194 = ($$0166|0)<(53);
    $195 = (84 - ($$0166))|0;
    if ($194) {
     $$0166170 = $$0166;$197 = $195;
     label = 61;
    } else {
     $$pre = (+($3|0));
     $$0151 = 0.0;$$0166169 = $$0166;$$pre$phiZ2D = $$pre;
    }
   }
   if ((label|0) == 61) {
    $196 = (+($3|0));
    $198 = (+_scalbn(1.0,$197));
    $199 = (+_copysignl($198,$196));
    $$0151 = $199;$$0166169 = $$0166170;$$pre$phiZ2D = $196;
   }
   $200 = ($$0166169|0)<(32);
   $201 = $$3158$lcssa != 0.0;
   $or$cond4 = $201 & $200;
   $202 = $$4163$lcssa & 1;
   $203 = ($202|0)==(0);
   $or$cond168 = $203 & $or$cond4;
   $204 = $or$cond168&1;
   $$6 = (($204) + ($$4163$lcssa))|0;
   $$5 = $or$cond168 ? 0.0 : $$3158$lcssa;
   $205 = (+($$6>>>0));
   $206 = $$pre$phiZ2D * $205;
   $207 = $$0151 + $206;
   $208 = $$pre$phiZ2D * $$5;
   $209 = $208 + $207;
   $210 = $209 - $$0151;
   $211 = $210 != 0.0;
   if (!($211)) {
    $212 = (___errno_location()|0);
    HEAP32[$212>>2] = 34;
   }
   $213 = (+_scalbnl($210,$184));
   $$0165 = $213;
  }
 } while(0);
 return (+$$0165);
}
function _decfloat($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$ = 0, $$$0345 = 0, $$$0350 = 0, $$$0385 = 0, $$$0401 = 0, $$$5355 = 0, $$$5390 = 0, $$0329 = 0, $$0332490 = 0, $$0333 = 0, $$0334 = 0, $$0336486 = 0, $$0340496 = 0, $$0341$lcssa = 0, $$0341463 = 0, $$0341464 = 0, $$0341465 = 0, $$0341513 = 0, $$0345$lcssa = 0, $$0345467 = 0;
 var $$0345468 = 0, $$0345469 = 0, $$0345512 = 0, $$0350$lcssa554 = 0, $$0350494 = 0, $$0360 = 0.0, $$0361 = 0.0, $$0365484 = 0.0, $$0372 = 0, $$0380 = 0, $$0380$ph = 0, $$0385$lcssa553 = 0, $$0385493 = 0, $$0393 = 0, $$0396 = 0, $$0401$lcssa = 0, $$0401473 = 0, $$0401474 = 0, $$0401475 = 0, $$0401509 = 0;
 var $$1 = 0.0, $$10 = 0, $$1330$be = 0, $$1330$ph = 0, $$1335 = 0, $$1337 = 0, $$1362 = 0.0, $$1366 = 0.0, $$1373 = 0, $$1373$ph448 = 0, $$1381 = 0, $$1381$ph = 0, $$1381$ph558 = 0, $$1394$lcssa = 0, $$1394511 = 0, $$2 = 0, $$2343 = 0, $$2347 = 0, $$2352$ph449 = 0, $$2367 = 0.0;
 var $$2371$v = 0, $$2374 = 0, $$2387$ph447 = 0, $$2395 = 0, $$2398 = 0, $$2403 = 0, $$3$be = 0, $$3$lcssa = 0, $$3344503 = 0, $$3348 = 0, $$3364 = 0.0, $$3368 = 0.0, $$3375 = 0, $$3383 = 0, $$3399$lcssa = 0, $$3399510 = 0, $$3514 = 0, $$413 = 0, $$425 = 0, $$4349495 = 0;
 var $$4354 = 0, $$4354$ph = 0, $$4354$ph559 = 0, $$4376 = 0, $$4384 = 0, $$4389$ph = 0, $$4389$ph445 = 0, $$4400 = 0, $$4485 = 0, $$5 = 0, $$5$in = 0, $$5355488 = 0, $$5390487 = 0, $$6378$ph = 0, $$6489 = 0, $$9483 = 0, $$neg442 = 0, $$neg443 = 0, $$pre = 0, $$promoted = 0;
 var $$sink = 0, $$sink421$off0 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0.0, $103 = 0.0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0.0, $12 = 0, $120 = 0.0, $121 = 0.0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0.0, $132 = 0.0, $133 = 0.0;
 var $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0.0, $144 = 0.0, $145 = 0.0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0;
 var $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0.0, $159 = 0.0, $16 = 0, $160 = 0.0, $161 = 0, $162 = 0.0, $163 = 0.0, $164 = 0.0, $165 = 0, $166 = 0, $167 = 0, $168 = 0.0, $169 = 0.0, $17 = 0;
 var $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0.0, $177 = 0.0, $178 = 0.0, $179 = 0, $18 = 0, $180 = 0.0, $181 = 0.0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0;
 var $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0;
 var $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0;
 var $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0;
 var $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0;
 var $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0;
 var $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0;
 var $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0.0, $304 = 0, $305 = 0, $306 = 0.0, $307 = 0.0, $308 = 0, $309 = 0.0, $31 = 0, $310 = 0.0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0;
 var $316 = 0, $317 = 0.0, $318 = 0.0, $319 = 0, $32 = 0, $320 = 0.0, $321 = 0.0, $322 = 0.0, $323 = 0.0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0;
 var $334 = 0.0, $335 = 0.0, $336 = 0, $337 = 0.0, $338 = 0.0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0.0, $343 = 0.0, $344 = 0.0, $345 = 0.0, $346 = 0, $347 = 0, $348 = 0.0, $349 = 0, $35 = 0, $350 = 0.0, $351 = 0.0;
 var $352 = 0.0, $353 = 0, $354 = 0, $355 = 0, $356 = 0.0, $357 = 0, $358 = 0.0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0.0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0;
 var $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $39 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $cond = 0, $exitcond = 0, $exitcond551 = 0, $narrow = 0, $not$ = 0, $or$cond = 0, $or$cond11 = 0, $or$cond14 = 0, $or$cond415 = 0, $or$cond417 = 0, $or$cond419 = 0, $or$cond420 = 0, $or$cond422 = 0, $or$cond422$not = 0, $or$cond423 = 0, $or$cond426 = 0, $or$cond5 = 0, $sum = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(512|0);
 $6 = sp;
 $sum = (($3) + ($2))|0;
 $7 = (0 - ($sum))|0;
 $8 = ((($0)) + 4|0);
 $9 = ((($0)) + 100|0);
 $$0329 = $1;$$0396 = 0;
 L1: while(1) {
  switch ($$0329|0) {
  case 46:  {
   label = 6;
   break L1;
   break;
  }
  case 48:  {
   break;
  }
  default: {
   $$0393 = 0;$$2 = $$0329;$$2398 = $$0396;$366 = 0;$367 = 0;
   break L1;
  }
  }
  $10 = HEAP32[$8>>2]|0;
  $11 = HEAP32[$9>>2]|0;
  $12 = ($10>>>0)<($11>>>0);
  if ($12) {
   $13 = ((($10)) + 1|0);
   HEAP32[$8>>2] = $13;
   $14 = HEAP8[$10>>0]|0;
   $15 = $14&255;
   $$0329 = $15;$$0396 = 1;
   continue;
  } else {
   $16 = (___shgetc($0)|0);
   $$0329 = $16;$$0396 = 1;
   continue;
  }
 }
 if ((label|0) == 6) {
  $17 = HEAP32[$8>>2]|0;
  $18 = HEAP32[$9>>2]|0;
  $19 = ($17>>>0)<($18>>>0);
  if ($19) {
   $20 = ((($17)) + 1|0);
   HEAP32[$8>>2] = $20;
   $21 = HEAP8[$17>>0]|0;
   $22 = $21&255;
   $$1330$ph = $22;
  } else {
   $23 = (___shgetc($0)|0);
   $$1330$ph = $23;
  }
  $24 = ($$1330$ph|0)==(48);
  if ($24) {
   $25 = 0;$26 = 0;
   while(1) {
    $27 = (_i64Add(($25|0),($26|0),-1,-1)|0);
    $28 = tempRet0;
    $29 = HEAP32[$8>>2]|0;
    $30 = HEAP32[$9>>2]|0;
    $31 = ($29>>>0)<($30>>>0);
    if ($31) {
     $32 = ((($29)) + 1|0);
     HEAP32[$8>>2] = $32;
     $33 = HEAP8[$29>>0]|0;
     $34 = $33&255;
     $$1330$be = $34;
    } else {
     $35 = (___shgetc($0)|0);
     $$1330$be = $35;
    }
    $36 = ($$1330$be|0)==(48);
    if ($36) {
     $25 = $27;$26 = $28;
    } else {
     $$0393 = 1;$$2 = $$1330$be;$$2398 = 1;$366 = $27;$367 = $28;
     break;
    }
   }
  } else {
   $$0393 = 1;$$2 = $$1330$ph;$$2398 = $$0396;$366 = 0;$367 = 0;
  }
 }
 HEAP32[$6>>2] = 0;
 $37 = (($$2) + -48)|0;
 $38 = ($37>>>0)<(10);
 $39 = ($$2|0)==(46);
 $40 = $39 | $38;
 L20: do {
  if ($40) {
   $41 = ((($6)) + 496|0);
   $$0341513 = 0;$$0345512 = 0;$$0401509 = 0;$$1394511 = $$0393;$$3399510 = $$2398;$$3514 = $$2;$368 = $39;$369 = $37;$370 = $366;$371 = $367;$44 = 0;$45 = 0;
   L22: while(1) {
    do {
     if ($368) {
      $cond = ($$1394511|0)==(0);
      if ($cond) {
       $$2343 = $$0341513;$$2347 = $$0345512;$$2395 = 1;$$2403 = $$0401509;$$4400 = $$3399510;$372 = $44;$373 = $45;$374 = $44;$375 = $45;
      } else {
       break L22;
      }
     } else {
      $43 = ($$0345512|0)<(125);
      $46 = (_i64Add(($44|0),($45|0),1,0)|0);
      $47 = tempRet0;
      $48 = ($$3514|0)!=(48);
      if (!($43)) {
       if (!($48)) {
        $$2343 = $$0341513;$$2347 = $$0345512;$$2395 = $$1394511;$$2403 = $$0401509;$$4400 = $$3399510;$372 = $370;$373 = $371;$374 = $46;$375 = $47;
        break;
       }
       $57 = HEAP32[$41>>2]|0;
       $58 = $57 | 1;
       HEAP32[$41>>2] = $58;
       $$2343 = $$0341513;$$2347 = $$0345512;$$2395 = $$1394511;$$2403 = $$0401509;$$4400 = $$3399510;$372 = $370;$373 = $371;$374 = $46;$375 = $47;
       break;
      }
      $$$0401 = $48 ? $46 : $$0401509;
      $49 = ($$0341513|0)==(0);
      $$pre = (($6) + ($$0345512<<2)|0);
      if ($49) {
       $$sink = $369;
      } else {
       $50 = HEAP32[$$pre>>2]|0;
       $51 = ($50*10)|0;
       $52 = (($$3514) + -48)|0;
       $53 = (($52) + ($51))|0;
       $$sink = $53;
      }
      HEAP32[$$pre>>2] = $$sink;
      $54 = (($$0341513) + 1)|0;
      $55 = ($54|0)==(9);
      $56 = $55&1;
      $$$0345 = (($56) + ($$0345512))|0;
      $$413 = $55 ? 0 : $54;
      $$2343 = $$413;$$2347 = $$$0345;$$2395 = $$1394511;$$2403 = $$$0401;$$4400 = 1;$372 = $370;$373 = $371;$374 = $46;$375 = $47;
     }
    } while(0);
    $59 = HEAP32[$8>>2]|0;
    $60 = HEAP32[$9>>2]|0;
    $61 = ($59>>>0)<($60>>>0);
    if ($61) {
     $62 = ((($59)) + 1|0);
     HEAP32[$8>>2] = $62;
     $63 = HEAP8[$59>>0]|0;
     $64 = $63&255;
     $$3$be = $64;
    } else {
     $65 = (___shgetc($0)|0);
     $$3$be = $65;
    }
    $66 = (($$3$be) + -48)|0;
    $67 = ($66>>>0)<(10);
    $68 = ($$3$be|0)==(46);
    $69 = $68 | $67;
    if ($69) {
     $$0341513 = $$2343;$$0345512 = $$2347;$$0401509 = $$2403;$$1394511 = $$2395;$$3399510 = $$4400;$$3514 = $$3$be;$368 = $68;$369 = $66;$370 = $372;$371 = $373;$44 = $374;$45 = $375;
    } else {
     $$0341$lcssa = $$2343;$$0345$lcssa = $$2347;$$0401$lcssa = $$2403;$$1394$lcssa = $$2395;$$3$lcssa = $$3$be;$$3399$lcssa = $$4400;$72 = $372;$73 = $374;$75 = $373;$76 = $375;
     label = 29;
     break L20;
    }
   }
   $42 = ($$3399510|0)!=(0);
   $$0341465 = $$0341513;$$0345469 = $$0345512;$$0401475 = $$0401509;$376 = $44;$377 = $45;$378 = $370;$379 = $371;$380 = $42;
   label = 37;
  } else {
   $$0341$lcssa = 0;$$0345$lcssa = 0;$$0401$lcssa = 0;$$1394$lcssa = $$0393;$$3$lcssa = $$2;$$3399$lcssa = $$2398;$72 = $366;$73 = 0;$75 = $367;$76 = 0;
   label = 29;
  }
 } while(0);
 do {
  if ((label|0) == 29) {
   $70 = ($$1394$lcssa|0)==(0);
   $71 = $70 ? $73 : $72;
   $74 = $70 ? $76 : $75;
   $77 = ($$3399$lcssa|0)!=(0);
   $78 = $$3$lcssa | 32;
   $79 = ($78|0)==(101);
   $or$cond415 = $77 & $79;
   if (!($or$cond415)) {
    $94 = ($$3$lcssa|0)>(-1);
    if ($94) {
     $$0341465 = $$0341$lcssa;$$0345469 = $$0345$lcssa;$$0401475 = $$0401$lcssa;$376 = $73;$377 = $76;$378 = $71;$379 = $74;$380 = $77;
     label = 37;
     break;
    } else {
     $$0341464 = $$0341$lcssa;$$0345468 = $$0345$lcssa;$$0401474 = $$0401$lcssa;$381 = $73;$382 = $76;$383 = $77;$384 = $71;$385 = $74;
     label = 39;
     break;
    }
   }
   $80 = (_scanexp($0,$5)|0);
   $81 = tempRet0;
   $82 = ($80|0)==(0);
   $83 = ($81|0)==(-2147483648);
   $84 = $82 & $83;
   if ($84) {
    $85 = ($5|0)==(0);
    if ($85) {
     ___shlim($0,0);
     $$1 = 0.0;
     break;
    }
    $86 = HEAP32[$9>>2]|0;
    $87 = ($86|0)==(0|0);
    if ($87) {
     $90 = 0;$91 = 0;
    } else {
     $88 = HEAP32[$8>>2]|0;
     $89 = ((($88)) + -1|0);
     HEAP32[$8>>2] = $89;
     $90 = 0;$91 = 0;
    }
   } else {
    $90 = $80;$91 = $81;
   }
   $92 = (_i64Add(($90|0),($91|0),($71|0),($74|0))|0);
   $93 = tempRet0;
   $$0341463 = $$0341$lcssa;$$0345467 = $$0345$lcssa;$$0401473 = $$0401$lcssa;$105 = $92;$106 = $73;$108 = $93;$109 = $76;
   label = 41;
  }
 } while(0);
 if ((label|0) == 37) {
  $95 = HEAP32[$9>>2]|0;
  $96 = ($95|0)==(0|0);
  if ($96) {
   $$0341464 = $$0341465;$$0345468 = $$0345469;$$0401474 = $$0401475;$381 = $376;$382 = $377;$383 = $380;$384 = $378;$385 = $379;
   label = 39;
  } else {
   $97 = HEAP32[$8>>2]|0;
   $98 = ((($97)) + -1|0);
   HEAP32[$8>>2] = $98;
   if ($380) {
    $$0341463 = $$0341465;$$0345467 = $$0345469;$$0401473 = $$0401475;$105 = $378;$106 = $376;$108 = $379;$109 = $377;
    label = 41;
   } else {
    label = 40;
   }
  }
 }
 if ((label|0) == 39) {
  if ($383) {
   $$0341463 = $$0341464;$$0345467 = $$0345468;$$0401473 = $$0401474;$105 = $384;$106 = $381;$108 = $385;$109 = $382;
   label = 41;
  } else {
   label = 40;
  }
 }
 do {
  if ((label|0) == 40) {
   $99 = (___errno_location()|0);
   HEAP32[$99>>2] = 22;
   ___shlim($0,0);
   $$1 = 0.0;
  }
  else if ((label|0) == 41) {
   $100 = HEAP32[$6>>2]|0;
   $101 = ($100|0)==(0);
   if ($101) {
    $102 = (+($4|0));
    $103 = $102 * 0.0;
    $$1 = $103;
    break;
   }
   $104 = ($105|0)==($106|0);
   $107 = ($108|0)==($109|0);
   $110 = $104 & $107;
   $111 = ($109|0)<(0);
   $112 = ($106>>>0)<(10);
   $113 = ($109|0)==(0);
   $114 = $113 & $112;
   $115 = $111 | $114;
   $or$cond = $115 & $110;
   if ($or$cond) {
    $116 = ($2|0)>(30);
    $117 = $100 >>> $2;
    $118 = ($117|0)==(0);
    $or$cond417 = $116 | $118;
    if ($or$cond417) {
     $119 = (+($4|0));
     $120 = (+($100>>>0));
     $121 = $119 * $120;
     $$1 = $121;
     break;
    }
   }
   $122 = (($3|0) / -2)&-1;
   $123 = ($122|0)<(0);
   $124 = $123 << 31 >> 31;
   $125 = ($108|0)>($124|0);
   $126 = ($105>>>0)>($122>>>0);
   $127 = ($108|0)==($124|0);
   $128 = $127 & $126;
   $129 = $125 | $128;
   if ($129) {
    $130 = (___errno_location()|0);
    HEAP32[$130>>2] = 34;
    $131 = (+($4|0));
    $132 = $131 * 1.7976931348623157E+308;
    $133 = $132 * 1.7976931348623157E+308;
    $$1 = $133;
    break;
   }
   $134 = (($3) + -106)|0;
   $135 = ($134|0)<(0);
   $136 = $135 << 31 >> 31;
   $137 = ($108|0)<($136|0);
   $138 = ($105>>>0)<($134>>>0);
   $139 = ($108|0)==($136|0);
   $140 = $139 & $138;
   $141 = $137 | $140;
   if ($141) {
    $142 = (___errno_location()|0);
    HEAP32[$142>>2] = 34;
    $143 = (+($4|0));
    $144 = $143 * 2.2250738585072014E-308;
    $145 = $144 * 2.2250738585072014E-308;
    $$1 = $145;
    break;
   }
   $146 = ($$0341463|0)==(0);
   if ($146) {
    $$3348 = $$0345467;
   } else {
    $147 = ($$0341463|0)<(9);
    if ($147) {
     $148 = (($6) + ($$0345467<<2)|0);
     $$promoted = HEAP32[$148>>2]|0;
     $$3344503 = $$0341463;$150 = $$promoted;
     while(1) {
      $149 = ($150*10)|0;
      $151 = (($$3344503) + 1)|0;
      $exitcond551 = ($151|0)==(9);
      if ($exitcond551) {
       break;
      } else {
       $$3344503 = $151;$150 = $149;
      }
     }
     HEAP32[$148>>2] = $149;
    }
    $152 = (($$0345467) + 1)|0;
    $$3348 = $152;
   }
   $153 = ($$0401473|0)<(9);
   if ($153) {
    $154 = ($$0401473|0)<=($105|0);
    $155 = ($105|0)<(18);
    $or$cond5 = $154 & $155;
    if ($or$cond5) {
     $156 = ($105|0)==(9);
     $157 = HEAP32[$6>>2]|0;
     if ($156) {
      $158 = (+($4|0));
      $159 = (+($157>>>0));
      $160 = $158 * $159;
      $$1 = $160;
      break;
     }
     $161 = ($105|0)<(9);
     if ($161) {
      $162 = (+($4|0));
      $163 = (+($157>>>0));
      $164 = $162 * $163;
      $165 = (8 - ($105))|0;
      $166 = (18468 + ($165<<2)|0);
      $167 = HEAP32[$166>>2]|0;
      $168 = (+($167|0));
      $169 = $164 / $168;
      $$1 = $169;
      break;
     }
     $$neg442 = Math_imul($105, -3)|0;
     $$neg443 = (($2) + 27)|0;
     $170 = (($$neg443) + ($$neg442))|0;
     $171 = ($170|0)>(30);
     $172 = $157 >>> $170;
     $173 = ($172|0)==(0);
     $or$cond419 = $171 | $173;
     if ($or$cond419) {
      $174 = (($105) + -10)|0;
      $175 = (18468 + ($174<<2)|0);
      $176 = (+($4|0));
      $177 = (+($157>>>0));
      $178 = $176 * $177;
      $179 = HEAP32[$175>>2]|0;
      $180 = (+($179|0));
      $181 = $178 * $180;
      $$1 = $181;
      break;
     }
    }
   }
   $182 = (($105|0) % 9)&-1;
   $183 = ($182|0)==(0);
   if ($183) {
    $$0380$ph = 0;$$1373$ph448 = $$3348;$$2352$ph449 = 0;$$2387$ph447 = $105;
   } else {
    $184 = ($105|0)>(-1);
    $185 = (($182) + 9)|0;
    $186 = $184 ? $182 : $185;
    $187 = (8 - ($186))|0;
    $188 = (18468 + ($187<<2)|0);
    $189 = HEAP32[$188>>2]|0;
    $190 = ($$3348|0)==(0);
    if ($190) {
     $$0350$lcssa554 = 0;$$0372 = 0;$$0385$lcssa553 = $105;
    } else {
     $191 = (1000000000 / ($189|0))&-1;
     $$0340496 = 0;$$0350494 = 0;$$0385493 = $105;$$4349495 = 0;
     while(1) {
      $192 = (($6) + ($$4349495<<2)|0);
      $193 = HEAP32[$192>>2]|0;
      $194 = (($193>>>0) % ($189>>>0))&-1;
      $195 = (($193>>>0) / ($189>>>0))&-1;
      $196 = (($195) + ($$0340496))|0;
      HEAP32[$192>>2] = $196;
      $197 = Math_imul($191, $194)|0;
      $198 = ($$4349495|0)==($$0350494|0);
      $199 = ($196|0)==(0);
      $or$cond420 = $198 & $199;
      $200 = (($$0350494) + 1)|0;
      $201 = $200 & 127;
      $202 = (($$0385493) + -9)|0;
      $$$0385 = $or$cond420 ? $202 : $$0385493;
      $$$0350 = $or$cond420 ? $201 : $$0350494;
      $203 = (($$4349495) + 1)|0;
      $204 = ($203|0)==($$3348|0);
      if ($204) {
       break;
      } else {
       $$0340496 = $197;$$0350494 = $$$0350;$$0385493 = $$$0385;$$4349495 = $203;
      }
     }
     $205 = ($197|0)==(0);
     if ($205) {
      $$0350$lcssa554 = $$$0350;$$0372 = $$3348;$$0385$lcssa553 = $$$0385;
     } else {
      $206 = (($6) + ($$3348<<2)|0);
      $207 = (($$3348) + 1)|0;
      HEAP32[$206>>2] = $197;
      $$0350$lcssa554 = $$$0350;$$0372 = $207;$$0385$lcssa553 = $$$0385;
     }
    }
    $208 = (9 - ($186))|0;
    $209 = (($208) + ($$0385$lcssa553))|0;
    $$0380$ph = 0;$$1373$ph448 = $$0372;$$2352$ph449 = $$0350$lcssa554;$$2387$ph447 = $209;
   }
   L101: while(1) {
    $210 = ($$2387$ph447|0)<(18);
    $211 = ($$2387$ph447|0)==(18);
    $212 = (($6) + ($$2352$ph449<<2)|0);
    $$0380 = $$0380$ph;$$1373 = $$1373$ph448;
    while(1) {
     if (!($210)) {
      if (!($211)) {
       $$1381$ph = $$0380;$$4354$ph = $$2352$ph449;$$4389$ph445 = $$2387$ph447;$$6378$ph = $$1373;
       break L101;
      }
      $213 = HEAP32[$212>>2]|0;
      $214 = ($213>>>0)<(9007199);
      if (!($214)) {
       $$1381$ph = $$0380;$$4354$ph = $$2352$ph449;$$4389$ph445 = 18;$$6378$ph = $$1373;
       break L101;
      }
     }
     $215 = (($$1373) + 127)|0;
     $$0334 = 0;$$2374 = $$1373;$$5$in = $215;
     while(1) {
      $$5 = $$5$in & 127;
      $216 = (($6) + ($$5<<2)|0);
      $217 = HEAP32[$216>>2]|0;
      $218 = (_bitshift64Shl(($217|0),0,29)|0);
      $219 = tempRet0;
      $220 = (_i64Add(($218|0),($219|0),($$0334|0),0)|0);
      $221 = tempRet0;
      $222 = ($221>>>0)>(0);
      $223 = ($220>>>0)>(1000000000);
      $224 = ($221|0)==(0);
      $225 = $224 & $223;
      $226 = $222 | $225;
      if ($226) {
       $227 = (___udivdi3(($220|0),($221|0),1000000000,0)|0);
       $228 = tempRet0;
       $229 = (___uremdi3(($220|0),($221|0),1000000000,0)|0);
       $230 = tempRet0;
       $$1335 = $227;$$sink421$off0 = $229;
      } else {
       $$1335 = 0;$$sink421$off0 = $220;
      }
      HEAP32[$216>>2] = $$sink421$off0;
      $231 = (($$2374) + 127)|0;
      $232 = $231 & 127;
      $233 = ($$5|0)!=($232|0);
      $234 = ($$5|0)==($$2352$ph449|0);
      $or$cond422 = $233 | $234;
      $or$cond422$not = $or$cond422 ^ 1;
      $235 = ($$sink421$off0|0)==(0);
      $or$cond423 = $235 & $or$cond422$not;
      $$3375 = $or$cond423 ? $$5 : $$2374;
      $236 = (($$5) + -1)|0;
      if ($234) {
       break;
      } else {
       $$0334 = $$1335;$$2374 = $$3375;$$5$in = $236;
      }
     }
     $237 = (($$0380) + -29)|0;
     $238 = ($$1335|0)==(0);
     if ($238) {
      $$0380 = $237;$$1373 = $$3375;
     } else {
      break;
     }
    }
    $239 = (($$2387$ph447) + 9)|0;
    $240 = (($$2352$ph449) + 127)|0;
    $241 = $240 & 127;
    $242 = ($241|0)==($$3375|0);
    $243 = (($$3375) + 127)|0;
    $244 = $243 & 127;
    $245 = (($$3375) + 126)|0;
    $246 = $245 & 127;
    $247 = (($6) + ($246<<2)|0);
    if ($242) {
     $248 = (($6) + ($244<<2)|0);
     $249 = HEAP32[$248>>2]|0;
     $250 = HEAP32[$247>>2]|0;
     $251 = $250 | $249;
     HEAP32[$247>>2] = $251;
     $$4376 = $244;
    } else {
     $$4376 = $$3375;
    }
    $252 = (($6) + ($241<<2)|0);
    HEAP32[$252>>2] = $$1335;
    $$0380$ph = $237;$$1373$ph448 = $$4376;$$2352$ph449 = $241;$$2387$ph447 = $239;
   }
   L119: while(1) {
    $289 = (($$6378$ph) + 1)|0;
    $287 = $289 & 127;
    $290 = (($$6378$ph) + 127)|0;
    $291 = $290 & 127;
    $292 = (($6) + ($291<<2)|0);
    $$1381$ph558 = $$1381$ph;$$4354$ph559 = $$4354$ph;$$4389$ph = $$4389$ph445;
    while(1) {
     $265 = ($$4389$ph|0)==(18);
     $293 = ($$4389$ph|0)>(27);
     $$425 = $293 ? 9 : 1;
     $$1381 = $$1381$ph558;$$4354 = $$4354$ph559;
     while(1) {
      $$0336486 = 0;
      while(1) {
       $253 = (($$0336486) + ($$4354))|0;
       $254 = $253 & 127;
       $255 = ($254|0)==($$6378$ph|0);
       if ($255) {
        $$1337 = 2;
        label = 88;
        break;
       }
       $256 = (($6) + ($254<<2)|0);
       $257 = HEAP32[$256>>2]|0;
       $258 = (18500 + ($$0336486<<2)|0);
       $259 = HEAP32[$258>>2]|0;
       $260 = ($257>>>0)<($259>>>0);
       if ($260) {
        $$1337 = 2;
        label = 88;
        break;
       }
       $261 = ($257>>>0)>($259>>>0);
       if ($261) {
        break;
       }
       $262 = (($$0336486) + 1)|0;
       $263 = ($262|0)<(2);
       if ($263) {
        $$0336486 = $262;
       } else {
        $$1337 = $262;
        label = 88;
        break;
       }
      }
      if ((label|0) == 88) {
       label = 0;
       $264 = ($$1337|0)==(2);
       $or$cond11 = $265 & $264;
       if ($or$cond11) {
        $$0365484 = 0.0;$$4485 = 0;$$9483 = $$6378$ph;
        break L119;
       }
      }
      $266 = (($$425) + ($$1381))|0;
      $267 = ($$4354|0)==($$6378$ph|0);
      if ($267) {
       $$1381 = $266;$$4354 = $$6378$ph;
      } else {
       break;
      }
     }
     $268 = 1 << $$425;
     $269 = (($268) + -1)|0;
     $270 = 1000000000 >>> $$425;
     $$0332490 = 0;$$5355488 = $$4354;$$5390487 = $$4389$ph;$$6489 = $$4354;
     while(1) {
      $271 = (($6) + ($$6489<<2)|0);
      $272 = HEAP32[$271>>2]|0;
      $273 = $272 & $269;
      $274 = $272 >>> $$425;
      $275 = (($274) + ($$0332490))|0;
      HEAP32[$271>>2] = $275;
      $276 = Math_imul($273, $270)|0;
      $277 = ($$6489|0)==($$5355488|0);
      $278 = ($275|0)==(0);
      $or$cond426 = $277 & $278;
      $279 = (($$5355488) + 1)|0;
      $280 = $279 & 127;
      $281 = (($$5390487) + -9)|0;
      $$$5390 = $or$cond426 ? $281 : $$5390487;
      $$$5355 = $or$cond426 ? $280 : $$5355488;
      $282 = (($$6489) + 1)|0;
      $283 = $282 & 127;
      $284 = ($283|0)==($$6378$ph|0);
      if ($284) {
       break;
      } else {
       $$0332490 = $276;$$5355488 = $$$5355;$$5390487 = $$$5390;$$6489 = $283;
      }
     }
     $285 = ($276|0)==(0);
     if ($285) {
      $$1381$ph558 = $266;$$4354$ph559 = $$$5355;$$4389$ph = $$$5390;
      continue;
     }
     $286 = ($287|0)==($$$5355|0);
     if (!($286)) {
      break;
     }
     $294 = HEAP32[$292>>2]|0;
     $295 = $294 | 1;
     HEAP32[$292>>2] = $295;
     $$1381$ph558 = $266;$$4354$ph559 = $$$5355;$$4389$ph = $$$5390;
    }
    $288 = (($6) + ($$6378$ph<<2)|0);
    HEAP32[$288>>2] = $276;
    $$1381$ph = $266;$$4354$ph = $$$5355;$$4389$ph445 = $$$5390;$$6378$ph = $287;
   }
   while(1) {
    $296 = (($$4485) + ($$4354))|0;
    $297 = $296 & 127;
    $298 = ($297|0)==($$9483|0);
    $299 = (($$9483) + 1)|0;
    $300 = $299 & 127;
    if ($298) {
     $301 = (($300) + -1)|0;
     $302 = (($6) + ($301<<2)|0);
     HEAP32[$302>>2] = 0;
     $$10 = $300;
    } else {
     $$10 = $$9483;
    }
    $303 = $$0365484 * 1.0E+9;
    $304 = (($6) + ($297<<2)|0);
    $305 = HEAP32[$304>>2]|0;
    $306 = (+($305>>>0));
    $307 = $303 + $306;
    $308 = (($$4485) + 1)|0;
    $exitcond = ($308|0)==(2);
    if ($exitcond) {
     break;
    } else {
     $$0365484 = $307;$$4485 = $308;$$9483 = $$10;
    }
   }
   $309 = (+($4|0));
   $310 = $309 * $307;
   $311 = (($$1381) + 53)|0;
   $312 = (($311) - ($3))|0;
   $313 = ($312|0)<($2|0);
   $314 = ($312|0)>(0);
   $$ = $314 ? $312 : 0;
   $$0333 = $313 ? $$ : $2;
   $315 = ($$0333|0)<(53);
   if ($315) {
    $316 = (105 - ($$0333))|0;
    $317 = (+_scalbn(1.0,$316));
    $318 = (+_copysignl($317,$310));
    $319 = (53 - ($$0333))|0;
    $320 = (+_scalbn(1.0,$319));
    $321 = (+_fmodl($310,$320));
    $322 = $310 - $321;
    $323 = $318 + $322;
    $$0360 = $318;$$0361 = $321;$$1366 = $323;
   } else {
    $$0360 = 0.0;$$0361 = 0.0;$$1366 = $310;
   }
   $324 = (($$4354) + 2)|0;
   $325 = $324 & 127;
   $326 = ($325|0)==($$10|0);
   if ($326) {
    $$3364 = $$0361;
   } else {
    $327 = (($6) + ($325<<2)|0);
    $328 = HEAP32[$327>>2]|0;
    $329 = ($328>>>0)<(500000000);
    do {
     if ($329) {
      $330 = ($328|0)==(0);
      if ($330) {
       $331 = (($$4354) + 3)|0;
       $332 = $331 & 127;
       $333 = ($332|0)==($$10|0);
       if ($333) {
        $$1362 = $$0361;
        break;
       }
      }
      $334 = $309 * 0.25;
      $335 = $334 + $$0361;
      $$1362 = $335;
     } else {
      $336 = ($328|0)==(500000000);
      if (!($336)) {
       $337 = $309 * 0.75;
       $338 = $337 + $$0361;
       $$1362 = $338;
       break;
      }
      $339 = (($$4354) + 3)|0;
      $340 = $339 & 127;
      $341 = ($340|0)==($$10|0);
      if ($341) {
       $342 = $309 * 0.5;
       $343 = $342 + $$0361;
       $$1362 = $343;
       break;
      } else {
       $344 = $309 * 0.75;
       $345 = $344 + $$0361;
       $$1362 = $345;
       break;
      }
     }
    } while(0);
    $346 = (53 - ($$0333))|0;
    $347 = ($346|0)>(1);
    if ($347) {
     $348 = (+_fmodl($$1362,1.0));
     $349 = $348 != 0.0;
     if ($349) {
      $$3364 = $$1362;
     } else {
      $350 = $$1362 + 1.0;
      $$3364 = $350;
     }
    } else {
     $$3364 = $$1362;
    }
   }
   $351 = $$1366 + $$3364;
   $352 = $351 - $$0360;
   $353 = $311 & 2147483647;
   $354 = (-2 - ($sum))|0;
   $355 = ($353|0)>($354|0);
   do {
    if ($355) {
     $356 = (+Math_abs((+$352)));
     $357 = !($356 >= 9007199254740992.0);
     $358 = $352 * 0.5;
     $not$ = $357 ^ 1;
     $359 = $not$&1;
     $$3383 = (($359) + ($$1381))|0;
     $$2367 = $357 ? $352 : $358;
     $360 = (($$3383) + 50)|0;
     $361 = ($360|0)>($7|0);
     if (!($361)) {
      $362 = ($$0333|0)!=($312|0);
      $narrow = $362 | $357;
      $$2371$v = $313 & $narrow;
      $363 = $$3364 != 0.0;
      $or$cond14 = $363 & $$2371$v;
      if (!($or$cond14)) {
       $$3368 = $$2367;$$4384 = $$3383;
       break;
      }
     }
     $364 = (___errno_location()|0);
     HEAP32[$364>>2] = 34;
     $$3368 = $$2367;$$4384 = $$3383;
    } else {
     $$3368 = $352;$$4384 = $$1381;
    }
   } while(0);
   $365 = (+_scalbnl($$3368,$$4384));
   $$1 = $365;
  }
 } while(0);
 STACKTOP = sp;return (+$$1);
}
function _scanexp($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$04861 = 0, $$049 = 0, $$1$be = 0, $$160 = 0, $$2$be = 0, $$2$lcssa = 0, $$254 = 0, $$3$be = 0, $$lcssa = 0, $$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($0)) + 100|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($3>>>0)<($5>>>0);
 if ($6) {
  $7 = ((($3)) + 1|0);
  HEAP32[$2>>2] = $7;
  $8 = HEAP8[$3>>0]|0;
  $9 = $8&255;
  $11 = $9;
 } else {
  $10 = (___shgetc($0)|0);
  $11 = $10;
 }
 switch ($11|0) {
 case 43: case 45:  {
  $12 = ($11|0)==(45);
  $13 = $12&1;
  $14 = HEAP32[$2>>2]|0;
  $15 = HEAP32[$4>>2]|0;
  $16 = ($14>>>0)<($15>>>0);
  if ($16) {
   $17 = ((($14)) + 1|0);
   HEAP32[$2>>2] = $17;
   $18 = HEAP8[$14>>0]|0;
   $19 = $18&255;
   $22 = $19;
  } else {
   $20 = (___shgetc($0)|0);
   $22 = $20;
  }
  $21 = (($22) + -48)|0;
  $23 = ($21>>>0)>(9);
  $24 = ($1|0)!=(0);
  $or$cond3 = $24 & $23;
  if ($or$cond3) {
   $25 = HEAP32[$4>>2]|0;
   $26 = ($25|0)==(0|0);
   if ($26) {
    $$0 = $13;$$049 = $22;
   } else {
    $27 = HEAP32[$2>>2]|0;
    $28 = ((($27)) + -1|0);
    HEAP32[$2>>2] = $28;
    $$0 = $13;$$049 = $22;
   }
  } else {
   $$0 = $13;$$049 = $22;
  }
  break;
 }
 default: {
  $$0 = 0;$$049 = $11;
 }
 }
 $29 = (($$049) + -48)|0;
 $30 = ($29>>>0)>(9);
 if ($30) {
  $31 = HEAP32[$4>>2]|0;
  $32 = ($31|0)==(0|0);
  if ($32) {
   $100 = -2147483648;$101 = 0;
  } else {
   $33 = HEAP32[$2>>2]|0;
   $34 = ((($33)) + -1|0);
   HEAP32[$2>>2] = $34;
   $100 = -2147483648;$101 = 0;
  }
 } else {
  $$04861 = 0;$$160 = $$049;
  while(1) {
   $35 = ($$04861*10)|0;
   $36 = (($$160) + -48)|0;
   $37 = (($36) + ($35))|0;
   $38 = HEAP32[$2>>2]|0;
   $39 = HEAP32[$4>>2]|0;
   $40 = ($38>>>0)<($39>>>0);
   if ($40) {
    $41 = ((($38)) + 1|0);
    HEAP32[$2>>2] = $41;
    $42 = HEAP8[$38>>0]|0;
    $43 = $42&255;
    $$1$be = $43;
   } else {
    $44 = (___shgetc($0)|0);
    $$1$be = $44;
   }
   $45 = (($$1$be) + -48)|0;
   $46 = ($45>>>0)<(10);
   $47 = ($37|0)<(214748364);
   $48 = $46 & $47;
   if ($48) {
    $$04861 = $37;$$160 = $$1$be;
   } else {
    break;
   }
  }
  $49 = ($37|0)<(0);
  $50 = $49 << 31 >> 31;
  $51 = (($$1$be) + -48)|0;
  $52 = ($51>>>0)<(10);
  if ($52) {
   $$254 = $$1$be;$56 = $37;$57 = $50;
   while(1) {
    $58 = (___muldi3(($56|0),($57|0),10,0)|0);
    $59 = tempRet0;
    $60 = ($$254|0)<(0);
    $61 = $60 << 31 >> 31;
    $62 = (_i64Add(($$254|0),($61|0),-48,-1)|0);
    $63 = tempRet0;
    $64 = (_i64Add(($62|0),($63|0),($58|0),($59|0))|0);
    $65 = tempRet0;
    $66 = HEAP32[$2>>2]|0;
    $67 = HEAP32[$4>>2]|0;
    $68 = ($66>>>0)<($67>>>0);
    if ($68) {
     $69 = ((($66)) + 1|0);
     HEAP32[$2>>2] = $69;
     $70 = HEAP8[$66>>0]|0;
     $71 = $70&255;
     $$2$be = $71;
    } else {
     $72 = (___shgetc($0)|0);
     $$2$be = $72;
    }
    $73 = (($$2$be) + -48)|0;
    $74 = ($73>>>0)<(10);
    $75 = ($65|0)<(21474836);
    $76 = ($64>>>0)<(2061584302);
    $77 = ($65|0)==(21474836);
    $78 = $77 & $76;
    $79 = $75 | $78;
    $80 = $74 & $79;
    if ($80) {
     $$254 = $$2$be;$56 = $64;$57 = $65;
    } else {
     $$2$lcssa = $$2$be;$94 = $64;$95 = $65;
     break;
    }
   }
  } else {
   $$2$lcssa = $$1$be;$94 = $37;$95 = $50;
  }
  $53 = (($$2$lcssa) + -48)|0;
  $54 = ($53>>>0)<(10);
  $55 = HEAP32[$4>>2]|0;
  if ($54) {
   $83 = $55;
   while(1) {
    $81 = HEAP32[$2>>2]|0;
    $82 = ($81>>>0)<($83>>>0);
    if ($82) {
     $84 = ((($81)) + 1|0);
     HEAP32[$2>>2] = $84;
     $85 = HEAP8[$81>>0]|0;
     $86 = $85&255;
     $$3$be = $86;$102 = $83;
    } else {
     $87 = (___shgetc($0)|0);
     $$pre = HEAP32[$4>>2]|0;
     $$3$be = $87;$102 = $$pre;
    }
    $88 = (($$3$be) + -48)|0;
    $89 = ($88>>>0)<(10);
    if ($89) {
     $83 = $102;
    } else {
     $$lcssa = $102;
     break;
    }
   }
  } else {
   $$lcssa = $55;
  }
  $90 = ($$lcssa|0)==(0|0);
  if (!($90)) {
   $91 = HEAP32[$2>>2]|0;
   $92 = ((($91)) + -1|0);
   HEAP32[$2>>2] = $92;
  }
  $93 = ($$0|0)!=(0);
  $96 = (_i64Subtract(0,0,($94|0),($95|0))|0);
  $97 = tempRet0;
  $98 = $93 ? $96 : $94;
  $99 = $93 ? $97 : $95;
  $100 = $99;$101 = $98;
 }
 tempRet0 = ($100);
 return ($101|0);
}
function _scalbn($0,$1) {
 $0 = +$0;
 $1 = $1|0;
 var $$ = 0, $$$ = 0, $$0 = 0.0, $$020 = 0, $$1 = 0, $$1$ = 0, $$21 = 0.0, $$22 = 0.0, $10 = 0.0, $11 = 0, $12 = 0, $13 = 0.0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0.0, $2 = 0, $20 = 0.0;
 var $3 = 0.0, $4 = 0, $5 = 0, $6 = 0.0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1|0)>(1023);
 if ($2) {
  $3 = $0 * 8.9884656743115795E+307;
  $4 = (($1) + -1023)|0;
  $5 = ($4|0)>(1023);
  $6 = $3 * 8.9884656743115795E+307;
  $7 = (($1) + -2046)|0;
  $8 = ($7|0)<(1023);
  $$ = $8 ? $7 : 1023;
  $$$ = $5 ? $$ : $4;
  $$21 = $5 ? $6 : $3;
  $$0 = $$21;$$020 = $$$;
 } else {
  $9 = ($1|0)<(-1022);
  if ($9) {
   $10 = $0 * 2.2250738585072014E-308;
   $11 = (($1) + 1022)|0;
   $12 = ($11|0)<(-1022);
   $13 = $10 * 2.2250738585072014E-308;
   $14 = (($1) + 2044)|0;
   $15 = ($14|0)>(-1022);
   $$1 = $15 ? $14 : -1022;
   $$1$ = $12 ? $$1 : $11;
   $$22 = $12 ? $13 : $10;
   $$0 = $$22;$$020 = $$1$;
  } else {
   $$0 = $0;$$020 = $1;
  }
 }
 $16 = (($$020) + 1023)|0;
 $17 = (_bitshift64Shl(($16|0),0,52)|0);
 $18 = tempRet0;
 HEAP32[tempDoublePtr>>2] = $17;HEAP32[tempDoublePtr+4>>2] = $18;$19 = +HEAPF64[tempDoublePtr>>3];
 $20 = $$0 * $19;
 return (+$20);
}
function _copysignl($0,$1) {
 $0 = +$0;
 $1 = +$1;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_copysign($0,$1));
 return (+$2);
}
function _fmodl($0,$1) {
 $0 = +$0;
 $1 = +$1;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_fmod($0,$1));
 return (+$2);
}
function _scalbnl($0,$1) {
 $0 = +$0;
 $1 = $1|0;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_scalbn($0,$1));
 return (+$2);
}
function _fmod($0,$1) {
 $0 = +$0;
 $1 = +$1;
 var $$ = 0.0, $$070 = 0.0, $$071$lcssa = 0, $$07194 = 0, $$073$lcssa = 0, $$073100 = 0, $$172$ph = 0, $$174 = 0, $$275$lcssa = 0, $$27586 = 0, $$376$lcssa = 0, $$37683 = 0, $$lcssa = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0.0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0.0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0.0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0.0, $28 = 0.0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0.0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $0;$2 = HEAP32[tempDoublePtr>>2]|0;
 $3 = HEAP32[tempDoublePtr+4>>2]|0;
 HEAPF64[tempDoublePtr>>3] = $1;$4 = HEAP32[tempDoublePtr>>2]|0;
 $5 = HEAP32[tempDoublePtr+4>>2]|0;
 $6 = (_bitshift64Lshr(($2|0),($3|0),52)|0);
 $7 = tempRet0;
 $8 = $6 & 2047;
 $9 = (_bitshift64Lshr(($4|0),($5|0),52)|0);
 $10 = tempRet0;
 $11 = $9 & 2047;
 $12 = $3 & -2147483648;
 $13 = (_bitshift64Shl(($4|0),($5|0),1)|0);
 $14 = tempRet0;
 $15 = ($13|0)==(0);
 $16 = ($14|0)==(0);
 $17 = $15 & $16;
 L1: do {
  if ($17) {
   label = 3;
  } else {
   $18 = (___DOUBLE_BITS_272($1)|0);
   $19 = tempRet0;
   $20 = $19 & 2147483647;
   $21 = ($20>>>0)>(2146435072);
   $22 = ($18>>>0)>(0);
   $23 = ($20|0)==(2146435072);
   $24 = $23 & $22;
   $25 = $21 | $24;
   $26 = ($8|0)==(2047);
   $or$cond = $26 | $25;
   if ($or$cond) {
    label = 3;
   } else {
    $29 = (_bitshift64Shl(($2|0),($3|0),1)|0);
    $30 = tempRet0;
    $31 = ($30>>>0)>($14>>>0);
    $32 = ($29>>>0)>($13>>>0);
    $33 = ($30|0)==($14|0);
    $34 = $33 & $32;
    $35 = $31 | $34;
    if (!($35)) {
     $36 = ($29|0)==($13|0);
     $37 = ($30|0)==($14|0);
     $38 = $36 & $37;
     $39 = $0 * 0.0;
     $$ = $38 ? $39 : $0;
     return (+$$);
    }
    $40 = ($8|0)==(0);
    if ($40) {
     $41 = (_bitshift64Shl(($2|0),($3|0),12)|0);
     $42 = tempRet0;
     $43 = ($42|0)>(-1);
     $44 = ($41>>>0)>(4294967295);
     $45 = ($42|0)==(-1);
     $46 = $45 & $44;
     $47 = $43 | $46;
     if ($47) {
      $$073100 = 0;$49 = $41;$50 = $42;
      while(1) {
       $48 = (($$073100) + -1)|0;
       $51 = (_bitshift64Shl(($49|0),($50|0),1)|0);
       $52 = tempRet0;
       $53 = ($52|0)>(-1);
       $54 = ($51>>>0)>(4294967295);
       $55 = ($52|0)==(-1);
       $56 = $55 & $54;
       $57 = $53 | $56;
       if ($57) {
        $$073100 = $48;$49 = $51;$50 = $52;
       } else {
        $$073$lcssa = $48;
        break;
       }
      }
     } else {
      $$073$lcssa = 0;
     }
     $58 = (1 - ($$073$lcssa))|0;
     $59 = (_bitshift64Shl(($2|0),($3|0),($58|0))|0);
     $60 = tempRet0;
     $$174 = $$073$lcssa;$87 = $59;$88 = $60;
    } else {
     $61 = $3 & 1048575;
     $62 = $61 | 1048576;
     $$174 = $8;$87 = $2;$88 = $62;
    }
    $63 = ($11|0)==(0);
    if ($63) {
     $64 = (_bitshift64Shl(($4|0),($5|0),12)|0);
     $65 = tempRet0;
     $66 = ($65|0)>(-1);
     $67 = ($64>>>0)>(4294967295);
     $68 = ($65|0)==(-1);
     $69 = $68 & $67;
     $70 = $66 | $69;
     if ($70) {
      $$07194 = 0;$72 = $64;$73 = $65;
      while(1) {
       $71 = (($$07194) + -1)|0;
       $74 = (_bitshift64Shl(($72|0),($73|0),1)|0);
       $75 = tempRet0;
       $76 = ($75|0)>(-1);
       $77 = ($74>>>0)>(4294967295);
       $78 = ($75|0)==(-1);
       $79 = $78 & $77;
       $80 = $76 | $79;
       if ($80) {
        $$07194 = $71;$72 = $74;$73 = $75;
       } else {
        $$071$lcssa = $71;
        break;
       }
      }
     } else {
      $$071$lcssa = 0;
     }
     $81 = (1 - ($$071$lcssa))|0;
     $82 = (_bitshift64Shl(($4|0),($5|0),($81|0))|0);
     $83 = tempRet0;
     $$172$ph = $$071$lcssa;$89 = $82;$90 = $83;
    } else {
     $84 = $5 & 1048575;
     $85 = $84 | 1048576;
     $$172$ph = $11;$89 = $4;$90 = $85;
    }
    $86 = ($$174|0)>($$172$ph|0);
    $91 = (_i64Subtract(($87|0),($88|0),($89|0),($90|0))|0);
    $92 = tempRet0;
    $93 = ($92|0)>(-1);
    $94 = ($91>>>0)>(4294967295);
    $95 = ($92|0)==(-1);
    $96 = $95 & $94;
    $97 = $93 | $96;
    L23: do {
     if ($86) {
      $$27586 = $$174;$101 = $92;$156 = $97;$157 = $87;$158 = $88;$99 = $91;
      while(1) {
       if ($156) {
        $98 = ($99|0)==(0);
        $100 = ($101|0)==(0);
        $102 = $98 & $100;
        if ($102) {
         break;
        } else {
         $104 = $99;$105 = $101;
        }
       } else {
        $104 = $157;$105 = $158;
       }
       $106 = (_bitshift64Shl(($104|0),($105|0),1)|0);
       $107 = tempRet0;
       $108 = (($$27586) + -1)|0;
       $109 = ($108|0)>($$172$ph|0);
       $110 = (_i64Subtract(($106|0),($107|0),($89|0),($90|0))|0);
       $111 = tempRet0;
       $112 = ($111|0)>(-1);
       $113 = ($110>>>0)>(4294967295);
       $114 = ($111|0)==(-1);
       $115 = $114 & $113;
       $116 = $112 | $115;
       if ($109) {
        $$27586 = $108;$101 = $111;$156 = $116;$157 = $106;$158 = $107;$99 = $110;
       } else {
        $$275$lcssa = $108;$$lcssa = $116;$118 = $110;$120 = $111;$159 = $106;$160 = $107;
        break L23;
       }
      }
      $103 = $0 * 0.0;
      $$070 = $103;
      break L1;
     } else {
      $$275$lcssa = $$174;$$lcssa = $97;$118 = $91;$120 = $92;$159 = $87;$160 = $88;
     }
    } while(0);
    if ($$lcssa) {
     $117 = ($118|0)==(0);
     $119 = ($120|0)==(0);
     $121 = $117 & $119;
     if ($121) {
      $129 = $0 * 0.0;
      $$070 = $129;
      break;
     } else {
      $123 = $120;$125 = $118;
     }
    } else {
     $123 = $160;$125 = $159;
    }
    $122 = ($123>>>0)<(1048576);
    $124 = ($125>>>0)<(0);
    $126 = ($123|0)==(1048576);
    $127 = $126 & $124;
    $128 = $122 | $127;
    if ($128) {
     $$37683 = $$275$lcssa;$130 = $125;$131 = $123;
     while(1) {
      $132 = (_bitshift64Shl(($130|0),($131|0),1)|0);
      $133 = tempRet0;
      $134 = (($$37683) + -1)|0;
      $135 = ($133>>>0)<(1048576);
      $136 = ($132>>>0)<(0);
      $137 = ($133|0)==(1048576);
      $138 = $137 & $136;
      $139 = $135 | $138;
      if ($139) {
       $$37683 = $134;$130 = $132;$131 = $133;
      } else {
       $$376$lcssa = $134;$141 = $132;$142 = $133;
       break;
      }
     }
    } else {
     $$376$lcssa = $$275$lcssa;$141 = $125;$142 = $123;
    }
    $140 = ($$376$lcssa|0)>(0);
    if ($140) {
     $143 = (_i64Add(($141|0),($142|0),0,-1048576)|0);
     $144 = tempRet0;
     $145 = (_bitshift64Shl(($$376$lcssa|0),0,52)|0);
     $146 = tempRet0;
     $147 = $143 | $145;
     $148 = $144 | $146;
     $153 = $148;$155 = $147;
    } else {
     $149 = (1 - ($$376$lcssa))|0;
     $150 = (_bitshift64Lshr(($141|0),($142|0),($149|0))|0);
     $151 = tempRet0;
     $153 = $151;$155 = $150;
    }
    $152 = $153 | $12;
    HEAP32[tempDoublePtr>>2] = $155;HEAP32[tempDoublePtr+4>>2] = $152;$154 = +HEAPF64[tempDoublePtr>>3];
    $$070 = $154;
   }
  }
 } while(0);
 if ((label|0) == 3) {
  $27 = $0 * $1;
  $28 = $27 / $27;
  $$070 = $28;
 }
 return (+$$070);
}
function ___DOUBLE_BITS_272($0) {
 $0 = +$0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAPF64[tempDoublePtr>>3] = $0;$1 = HEAP32[tempDoublePtr>>2]|0;
 $2 = HEAP32[tempDoublePtr+4>>2]|0;
 tempRet0 = ($2);
 return ($1|0);
}
function _catclose($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _catgets($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($3|0);
}
function _catopen($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ((-1)|0);
}
function _strlen($0) {
 $0 = $0|0;
 var $$0 = 0, $$015$lcssa = 0, $$01519 = 0, $$1$lcssa = 0, $$pn = 0, $$pre = 0, $$sink = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = $0;
 $2 = $1 & 3;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $$015$lcssa = $0;
   label = 4;
  } else {
   $$01519 = $0;$23 = $1;
   while(1) {
    $4 = HEAP8[$$01519>>0]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $$sink = $23;
     break L1;
    }
    $6 = ((($$01519)) + 1|0);
    $7 = $6;
    $8 = $7 & 3;
    $9 = ($8|0)==(0);
    if ($9) {
     $$015$lcssa = $6;
     label = 4;
     break;
    } else {
     $$01519 = $6;$23 = $7;
    }
   }
  }
 } while(0);
 if ((label|0) == 4) {
  $$0 = $$015$lcssa;
  while(1) {
   $10 = HEAP32[$$0>>2]|0;
   $11 = (($10) + -16843009)|0;
   $12 = $10 & -2139062144;
   $13 = $12 ^ -2139062144;
   $14 = $13 & $11;
   $15 = ($14|0)==(0);
   $16 = ((($$0)) + 4|0);
   if ($15) {
    $$0 = $16;
   } else {
    break;
   }
  }
  $17 = $10&255;
  $18 = ($17<<24>>24)==(0);
  if ($18) {
   $$1$lcssa = $$0;
  } else {
   $$pn = $$0;
   while(1) {
    $19 = ((($$pn)) + 1|0);
    $$pre = HEAP8[$19>>0]|0;
    $20 = ($$pre<<24>>24)==(0);
    if ($20) {
     $$1$lcssa = $19;
     break;
    } else {
     $$pn = $19;
    }
   }
  }
  $21 = $$1$lcssa;
  $$sink = $21;
 }
 $22 = (($$sink) - ($1))|0;
 return ($22|0);
}
function ___munmap($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 _dummy();
 $2 = $0;
 HEAP32[$vararg_buffer>>2] = $2;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $1;
 $3 = (___syscall91(91,($vararg_buffer|0))|0);
 $4 = (___syscall_ret($3)|0);
 STACKTOP = sp;return ($4|0);
}
function _strchr($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (___strchrnul($0,$1)|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = $1&255;
 $5 = ($3<<24>>24)==($4<<24>>24);
 $6 = $5 ? $2 : 0;
 return ($6|0);
}
function ___strchrnul($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$029$lcssa = 0, $$02936 = 0, $$030$lcssa = 0, $$03039 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond33 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1 & 255;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $8 = (_strlen($0)|0);
   $9 = (($0) + ($8)|0);
   $$0 = $9;
  } else {
   $4 = $0;
   $5 = $4 & 3;
   $6 = ($5|0)==(0);
   if ($6) {
    $$030$lcssa = $0;
   } else {
    $7 = $1&255;
    $$03039 = $0;
    while(1) {
     $10 = HEAP8[$$03039>>0]|0;
     $11 = ($10<<24>>24)==(0);
     $12 = ($10<<24>>24)==($7<<24>>24);
     $or$cond = $11 | $12;
     if ($or$cond) {
      $$0 = $$03039;
      break L1;
     }
     $13 = ((($$03039)) + 1|0);
     $14 = $13;
     $15 = $14 & 3;
     $16 = ($15|0)==(0);
     if ($16) {
      $$030$lcssa = $13;
      break;
     } else {
      $$03039 = $13;
     }
    }
   }
   $17 = Math_imul($2, 16843009)|0;
   $18 = HEAP32[$$030$lcssa>>2]|0;
   $19 = (($18) + -16843009)|0;
   $20 = $18 & -2139062144;
   $21 = $20 ^ -2139062144;
   $22 = $21 & $19;
   $23 = ($22|0)==(0);
   L10: do {
    if ($23) {
     $$02936 = $$030$lcssa;$25 = $18;
     while(1) {
      $24 = $25 ^ $17;
      $26 = (($24) + -16843009)|0;
      $27 = $24 & -2139062144;
      $28 = $27 ^ -2139062144;
      $29 = $28 & $26;
      $30 = ($29|0)==(0);
      if (!($30)) {
       $$029$lcssa = $$02936;
       break L10;
      }
      $31 = ((($$02936)) + 4|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = (($32) + -16843009)|0;
      $34 = $32 & -2139062144;
      $35 = $34 ^ -2139062144;
      $36 = $35 & $33;
      $37 = ($36|0)==(0);
      if ($37) {
       $$02936 = $31;$25 = $32;
      } else {
       $$029$lcssa = $31;
       break;
      }
     }
    } else {
     $$029$lcssa = $$030$lcssa;
    }
   } while(0);
   $38 = $1&255;
   $$1 = $$029$lcssa;
   while(1) {
    $39 = HEAP8[$$1>>0]|0;
    $40 = ($39<<24>>24)==(0);
    $41 = ($39<<24>>24)==($38<<24>>24);
    $or$cond33 = $40 | $41;
    $42 = ((($$1)) + 1|0);
    if ($or$cond33) {
     $$0 = $$1;
     break;
    } else {
     $$1 = $42;
    }
   }
  }
 } while(0);
 return ($$0|0);
}
function _dummy() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function _freelocale($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___loc_is_allocated($0)|0);
 $2 = ($1|0)==(0);
 if (!($2)) {
  _free($0);
 }
 return;
}
function ___loc_is_allocated($0) {
 $0 = $0|0;
 var $$ = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)!=(0|0);
 $2 = ($0|0)!=(164336|0);
 $or$cond = $1 & $2;
 $3 = ($0|0)!=(14400|0);
 $$ = $3 & $or$cond;
 $4 = $$&1;
 return ($4|0);
}
function _iconv_open($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_find_charmap($0)|0);
 $3 = ($2|0)==(-1);
 if ($3) {
  label = 4;
 } else {
  $4 = (_find_charmap($1)|0);
  $5 = ($4|0)==(-1);
  if ($5) {
   label = 4;
  } else {
   $6 = (150617 + ($2)|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = ($7&255)>(207);
   if ($8) {
    label = 4;
   } else {
    $10 = $4 << 16;
    $11 = $10 | $2;
    $12 = $11;
    $$0 = $12;
   }
  }
 }
 if ((label|0) == 4) {
  $9 = (___errno_location()|0);
  HEAP32[$9>>2] = 22;
  $$0 = (-1);
 }
 return ($$0|0);
}
function _find_charmap($0) {
 $0 = $0|0;
 var $$ = 0, $$0$be = 0, $$015 = 0, $$018 = 0, $$1$lcssa = 0, $$117 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(0);
 $$ = $2 ? 150617 : $0;
 $$018 = 150617;
 while(1) {
  $3 = (_fuzzycmp($$,$$018)|0);
  $4 = ($3|0)==(0);
  if ($4) {
   label = 3;
   break;
  }
  $15 = (_strlen($$018)|0);
  $16 = (($15) + 1)|0;
  $17 = (($$018) + ($16)|0);
  $18 = HEAP8[$17>>0]|0;
  $19 = ($18<<24>>24)==(0);
  do {
   if ($19) {
    $20 = ((($17)) + 1|0);
    $21 = HEAP8[$20>>0]|0;
    $22 = ($21&255)>(128);
    if ($22) {
     $23 = ((($17)) + 2|0);
     $$0$be = $23;
     break;
    } else {
     $26 = $21&255;
     $27 = (128 - ($26))|0;
     $28 = $27 >>> 2;
     $29 = ($28*5)|0;
     $30 = (($29) + 2)|0;
     $31 = (($17) + ($30)|0);
     $$0$be = $31;
     break;
    }
   } else {
    $$0$be = $17;
   }
  } while(0);
  $24 = HEAP8[$$0$be>>0]|0;
  $25 = ($24<<24>>24)==(0);
  if ($25) {
   $$015 = -1;
   break;
  } else {
   $$018 = $$0$be;
  }
 }
 if ((label|0) == 3) {
  $5 = HEAP8[$$018>>0]|0;
  $6 = ($5<<24>>24)==(0);
  if ($6) {
   $$1$lcssa = $$018;
  } else {
   $$117 = $$018;
   while(1) {
    $7 = (_strlen($$117)|0);
    $8 = (($7) + 1)|0;
    $9 = (($$117) + ($8)|0);
    $10 = HEAP8[$9>>0]|0;
    $11 = ($10<<24>>24)==(0);
    if ($11) {
     $$1$lcssa = $9;
     break;
    } else {
     $$117 = $9;
    }
   }
  }
  $12 = ((($$1$lcssa)) + 1|0);
  $13 = $12;
  $14 = (($13) - (150617))|0;
  $$015 = $14;
 }
 return ($$015|0);
}
function _fuzzycmp($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$lcssa = 0, $$01124 = 0, $$012 = 0, $$025 = 0, $$1$lcssa = 0, $$119 = 0, $$lcssa = 0, $$lcssa16 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP8[$0>>0]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $$0$lcssa = $1;$$lcssa16 = 0;
   label = 8;
  } else {
   $$01124 = $0;$$025 = $1;$7 = $2;
   while(1) {
    $4 = HEAP8[$$025>>0]|0;
    $5 = ($4<<24>>24)==(0);
    if ($5) {
     $$0$lcssa = $$025;$$lcssa16 = $7;
     label = 8;
     break L1;
    }
    $6 = $7&255;
    $$119 = $$01124;$9 = $6;
    while(1) {
     $8 = $9 | 32;
     $10 = (($8) + -97)|0;
     $11 = ($10>>>0)>(26);
     $12 = (($9) + -48)|0;
     $13 = ($12>>>0)>(10);
     $or$cond = $13 & $11;
     if (!($or$cond)) {
      $$1$lcssa = $$119;$$lcssa = $9;
      break;
     }
     $14 = ((($$119)) + 1|0);
     $15 = HEAP8[$14>>0]|0;
     $16 = $15&255;
     $17 = ($15<<24>>24)==(0);
     if ($17) {
      $$1$lcssa = $14;$$lcssa = $16;
      break;
     } else {
      $$119 = $14;$9 = $16;
     }
    }
    $18 = $$lcssa | 32;
    $19 = $4&255;
    $20 = ($18|0)==($19|0);
    if (!($20)) {
     $$012 = 1;
     break L1;
    }
    $21 = ((($$1$lcssa)) + 1|0);
    $22 = ((($$025)) + 1|0);
    $23 = HEAP8[$21>>0]|0;
    $24 = ($23<<24>>24)==(0);
    if ($24) {
     $$0$lcssa = $22;$$lcssa16 = 0;
     label = 8;
     break;
    } else {
     $$01124 = $21;$$025 = $22;$7 = $23;
    }
   }
  }
 } while(0);
 if ((label|0) == 8) {
  $25 = HEAP8[$$0$lcssa>>0]|0;
  $26 = ($$lcssa16<<24>>24)!=($25<<24>>24);
  $27 = $26&1;
  $$012 = $27;
 }
 return ($$012|0);
}
function _iconv_close($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function _iconv($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$$8 = 0, $$0321379 = 0, $$0322383 = 0, $$0323385 = 0, $$0325 = 0, $$0326389 = 0, $$0327 = 0, $$0340 = 0, $$0340$ph = 0, $$0341 = 0, $$0345384 = 0, $$0377 = 0, $$1 = 0, $$10 = 0, $$11388 = 0, $$1328 = 0, $$1334 = 0, $$1346$ = 0, $$1346382 = 0;
 var $$2 = 0, $$2329 = 0, $$2335386 = 0, $$2343 = 0, $$3 = 0, $$3330387 = 0, $$3344 = 0, $$3348 = 0, $$353 = 0, $$354 = 0, $$357 = 0, $$360 = 0, $$4 = 0, $$4331 = 0, $$4349378 = 0, $$5332 = 0, $$5338 = 0, $$5350$ = 0, $$5350376 = 0, $$6339380 = 0;
 var $$6381 = 0, $$7352 = 0, $$8 = 0, $$9 = 0, $$in = 0, $$lobit = 0, $$mask = 0, $$pn = 0, $$pre = 0, $$sink14 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0;
 var $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0;
 var $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $exitcond = 0, $exitcond396 = 0, $exitcond397 = 0, $exitcond398 = 0, $not$ = 0, $not$364 = 0, $or$cond = 0, $or$cond11 = 0, $or$cond3 = 0;
 var $or$cond355 = 0, $or$cond356 = 0, $or$cond359 = 0, $or$cond362 = 0, $or$cond5 = 0, $or$cond9 = 0, $switch$cast = 0, $switch$cast$clear = 0, $switch$downshift = 0, $switch$downshift$clear = 0, $switch$masked = 0, $switch$tableidx = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $5 = sp;
 $6 = sp + 16|0;
 $7 = sp + 12|0;
 $8 = sp + 8|0;
 $9 = sp + 20|0;
 $10 = $0;
 $11 = $10 & 65535;
 $12 = $10 >>> 16;
 $13 = (150617 + ($12)|0);
 $14 = ((($13)) + 1|0);
 $15 = (150617 + ($11)|0);
 $16 = ((($15)) + 1|0);
 $17 = $5;
 $18 = $17;
 HEAP32[$18>>2] = 0;
 $19 = (($17) + 4)|0;
 $20 = $19;
 HEAP32[$20>>2] = 0;
 $21 = HEAP8[$13>>0]|0;
 $22 = HEAP8[$15>>0]|0;
 $23 = (___pthread_self_138()|0);
 $24 = ((($23)) + 188|0);
 $25 = HEAP32[$24>>2]|0;
 $26 = ($1|0)==(0|0);
 do {
  if ($26) {
   $$0325 = 0;
  } else {
   $27 = HEAP32[$1>>2]|0;
   $28 = ($27|0)==(0|0);
   if ($28) {
    $$0325 = 0;
   } else {
    $29 = HEAP32[$2>>2]|0;
    $30 = ($29|0)==(0);
    if ($30) {
     $$0325 = 0;
    } else {
     HEAP32[$24>>2] = 14400;
     $31 = $21&255;
     $32 = (($31) + 128)|0;
     $33 = $31 & 254;
     $34 = ($33|0)==(196);
     $35 = ($21<<24>>24)==(-38);
     $36 = ($21<<24>>24)==(-40);
     $37 = $22&255;
     $38 = (($37) + -192)|0;
     $39 = ($38>>>0)>(8);
     $40 = $11;
     $41 = (154983 + ($38)|0);
     $42 = (($37) + 128)|0;
     $43 = (128 - ($37))|0;
     $44 = ($43|0)==(0);
     $45 = $21 & -2;
     $46 = ($45<<24>>24)==(-60);
     $47 = (($31) + -192)|0;
     $48 = ($47>>>0)<(7);
     $$pre = HEAP32[$1>>2]|0;
     $$0326389 = 0;$$in = $$pre;$52 = $29;
     L5: while(1) {
      $49 = HEAP8[$$in>>0]|0;
      $50 = $49&255;
      $51 = ($49<<24>>24)<(0);
      $brmerge = $51 | $48;
      L7: do {
       if ($brmerge) {
        do {
         switch ($21<<24>>24) {
         case -57:  {
          $$0340$ph = 84;
          label = 114;
          break L5;
          break;
         }
         case -56:  {
          $53 = (_mbrtowc($6,$$in,$52,$5)|0);
          switch ($53|0) {
          case -1:  {
           $$0340$ph = 84;
           label = 114;
           break L5;
           break;
          }
          case -2:  {
           $$0340 = 22;
           break L5;
           break;
          }
          case 0:  {
           $$0341 = 1;
           break;
          }
          default: {
           $$0341 = $53;
          }
          }
          $54 = HEAP32[$6>>2]|0;
          $$2343 = $$0341;$$8 = $54;
          label = 86;
          break L7;
          break;
         }
         case -58:  {
          $55 = ($52>>>0)<(4);
          if ($55) {
           $$0340 = 22;
           break L5;
          }
          $56 = HEAP32[$$in>>2]|0;
          $$0327 = $56;
          label = 14;
          break;
         }
         case -61: case -64:  {
          $57 = ($52>>>0)<(4);
          if ($57) {
           $$0340 = 22;
           break L5;
          }
          $58 = (_get_32($$in,$31)|0);
          $$0327 = $58;
          label = 14;
          break;
         }
         case -63: case -62: case -59: case -60:  {
          $62 = ($52>>>0)<(2);
          if ($62) {
           $$0340 = 22;
           break L5;
          }
          $63 = (_get_16($$in,$31)|0);
          $64 = $63 & -1024;
          switch ($64|0) {
          case 56320:  {
           $$0340$ph = 84;
           label = 114;
           break L5;
           break;
          }
          case 55296:  {
           break;
          }
          default: {
           $$2343 = 2;$$8 = $63;
           label = 86;
           break L7;
          }
          }
          if ($34) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $65 = ($52>>>0)<(4);
          if ($65) {
           $$0340 = 22;
           break L5;
          }
          $66 = ((($$in)) + 2|0);
          $67 = (_get_16($66,$31)|0);
          $68 = (($67) + -56320)|0;
          $69 = ($68>>>0)>(1023);
          if ($69) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $70 = $63 << 10;
          $71 = (($70) + -56557568)|0;
          $72 = (($71) + ($68))|0;
          $$2343 = 4;$$8 = $72;
          label = 86;
          break L7;
          break;
         }
         case -47:  {
          $73 = (($50) + -161)|0;
          $74 = ($73>>>0)<(63);
          if ($74) {
           $75 = (($50) + 65216)|0;
           $$2343 = 1;$$8 = $75;
           label = 86;
           break L7;
          }
          $76 = ($52>>>0)<(2);
          if ($76) {
           $$0340 = 22;
           break L5;
          }
          $77 = ((($$in)) + 1|0);
          $78 = HEAP8[$77>>0]|0;
          $79 = $78&255;
          $80 = (($50) + -129)|0;
          $81 = ($80>>>0)<(31);
          if ($81) {
           $$1328 = $80;
          } else {
           $82 = $50 & 240;
           $83 = ($82|0)==(224);
           if (!($83)) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           }
           $84 = (($50) + -193)|0;
           $$1328 = $84;
          }
          $85 = $$1328 << 1;
          $86 = (($79) + -64)|0;
          $87 = ($86>>>0)<(95);
          if ($87) {
           $88 = ($78<<24>>24)==(127);
           if ($88) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           }
           $$lobit = ($78<<24>>24) >> 7;
           $89 = $$lobit << 24 >> 24;
           $90 = (($86) + ($89))|0;
           $$1334 = $90;$$2329 = $85;
          } else {
           $91 = (($79) + -159)|0;
           $92 = ($91>>>0)<(94);
           $$353 = $92 ? $91 : $79;
           $93 = $92&1;
           $$354 = $85 | $93;
           $$1334 = $$353;$$2329 = $$354;
          }
          $94 = ((23036 + (($$2329*188)|0)|0) + ($$1334<<1)|0);
          $95 = HEAP16[$94>>1]|0;
          $96 = $95&65535;
          $97 = ($95<<16>>16)==(0);
          if ($97) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          } else {
           $$2343 = 2;$$8 = $96;
           label = 86;
           break L7;
          }
          break;
         }
         case -48:  {
          $98 = ($52>>>0)<(2);
          if ($98) {
           $$0340 = 22;
           break L5;
          }
          $99 = ((($$in)) + 1|0);
          $100 = HEAP8[$99>>0]|0;
          $101 = $100&255;
          $102 = ($49<<24>>24)==(-114);
          $103 = (($101) + -161)|0;
          if ($102) {
           $104 = ($103>>>0)>(62);
           if ($104) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           }
           $105 = (($101) + 65216)|0;
           $$2343 = 2;$$8 = $105;
           label = 86;
           break L7;
          } else {
           $106 = (($50) + -161)|0;
           $107 = ($106>>>0)>(83);
           $108 = ($103>>>0)>(93);
           $or$cond3 = $107 | $108;
           if ($or$cond3) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           }
           $109 = ((23036 + (($106*188)|0)|0) + ($103<<1)|0);
           $110 = HEAP16[$109>>1]|0;
           $111 = $110&65535;
           $112 = ($110<<16>>16)==(0);
           if ($112) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           } else {
            $$2343 = 2;$$8 = $111;
            label = 86;
            break L7;
           }
          }
          break;
         }
         case -38:  {
          $113 = ($49&255)<(161);
          if ($113) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          break;
         }
         case -40: case -39:  {
          break;
         }
         case -32:  {
          $160 = ($52>>>0)<(2);
          if ($160) {
           $$0340 = 22;
           break L5;
          }
          $161 = ((($$in)) + 1|0);
          $162 = HEAP8[$161>>0]|0;
          $163 = $162&255;
          $164 = (($163) + -64)|0;
          $165 = ($164>>>0)>(190);
          $166 = (($163) + -127)|0;
          $167 = ($166>>>0)<(34);
          $or$cond359 = $165 | $167;
          if ($or$cond359) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $168 = ($164>>>0)>(62);
          $169 = (($163) + -98)|0;
          $$360 = $168 ? $169 : $164;
          $170 = (($50) + -161)|0;
          $171 = ($170>>>0)>(88);
          if (!($171)) {
           $201 = ((97052 + (($170*314)|0)|0) + ($$360<<1)|0);
           $202 = HEAP16[$201>>1]|0;
           $203 = $202&65535;
           $204 = ($170|0)==(39);
           if ($204) {
            $switch$tableidx = (($$360) + -58)|0;
            $205 = ($switch$tableidx>>>0)<(9);
            if ($205) {
             $switch$cast = $switch$tableidx&65535;
             $switch$cast$clear = $switch$cast & 511;
             $switch$downshift = (261&65535) >>> $switch$cast$clear;
             $switch$downshift$clear = $switch$downshift & 511;
             $switch$masked = $switch$downshift$clear&65535;
             $207 = $switch$masked;
            } else {
             $207 = 0;
            }
           } else {
            $207 = 0;
           }
           $206 = $207 << 17;
           $208 = $206 & 131072;
           $209 = $208 | $203;
           $210 = ($209|0)==(0);
           if ($210) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           } else {
            $$2343 = 2;$$8 = $209;
            label = 86;
            break L7;
           }
          }
          $172 = (($50) + -135)|0;
          $173 = ($172>>>0)>(119);
          if ($173) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $174 = ($49&255)<(161);
          $175 = (($50) + -224)|0;
          $$4331 = $174 ? $172 : $175;
          $176 = ($$4331*157)|0;
          $177 = (($$360) + ($176))|0;
          $178 = $177 >>> 4;
          $179 = (($178) + 4867)|0;
          $180 = (86708 + ($179<<1)|0);
          $181 = HEAP16[$180>>1]|0;
          $182 = $181&65535;
          $183 = $177 & 15;
          $184 = $182 >>> $183;
          $185 = $184 << 17;
          $186 = $185 & 131072;
          $187 = (86708 + ($177<<1)|0);
          $188 = HEAP16[$187>>1]|0;
          $189 = $188&65535;
          $190 = $186 | $189;
          $$mask = $190 & 196352;
          $191 = ($$mask|0)==(56320);
          if (!($191)) {
           $200 = ($190|0)==(0);
           if ($200) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           } else {
            $$2343 = 2;$$8 = $190;
            label = 86;
            break L7;
           }
          }
          if ($39) {
           $$3348 = 2;
          } else {
           $192 = HEAP8[$41>>0]|0;
           $193 = $192 << 24 >> 24;
           $$3348 = $193;
          }
          $194 = HEAP32[$4>>2]|0;
          $195 = ($$3348>>>0)>($194>>>0);
          if ($195) {
           $$0340$ph = 7;
           label = 114;
           break L5;
          }
          $196 = $189 & 255;
          $197 = (154993 + ($196)|0);
          HEAP32[$7>>2] = $197;
          HEAP32[$8>>2] = 4;
          $198 = (_iconv($40,$7,$8,$3,$4)|0);
          $199 = (($198) + ($$0326389))|0;
          $$3344 = 2;$$4 = $199;
          break L7;
          break;
         }
         case -24:  {
          $211 = ($52>>>0)<(2);
          if ($211) {
           $$0340 = 22;
           break L5;
          }
          $212 = ((($$in)) + 1|0);
          $213 = HEAP8[$212>>0]|0;
          $214 = $213&255;
          $215 = (($50) + -161)|0;
          $216 = (($214) + -161)|0;
          $217 = ($215>>>0)>(92);
          $218 = ($216>>>0)>(93);
          $or$cond9 = $217 | $218;
          if (!($or$cond9)) {
           $248 = ((124998 + (($215*188)|0)|0) + ($216<<1)|0);
           $249 = HEAP16[$248>>1]|0;
           $250 = $249&65535;
           $251 = ($249<<16>>16)==(0);
           if ($251) {
            $$0340$ph = 84;
            label = 114;
            break L5;
           } else {
            $$2343 = 2;$$8 = $250;
            label = 86;
            break L7;
           }
          }
          $219 = (($50) + -129)|0;
          $220 = ($219>>>0)>(92);
          if ($220) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $221 = ($219>>>0)>(68);
          $222 = ($213&255)>(82);
          $or$cond11 = $221 & $222;
          if ($or$cond11) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          }
          $223 = (($214) + -65)|0;
          $224 = ($223>>>0)<(26);
          do {
           if ($224) {
            $$5338 = $223;
           } else {
            $225 = (($214) + -97)|0;
            $226 = ($225>>>0)<(26);
            if ($226) {
             $227 = (($214) + -71)|0;
             $$5338 = $227;
             break;
            }
            $228 = (($214) + -129)|0;
            $229 = ($228>>>0)<(126);
            if (!($229)) {
             $$0340$ph = 84;
             label = 114;
             break L5;
            }
            $230 = (($214) + -77)|0;
            $$5338 = $230;
           }
          } while(0);
          $231 = ($219>>>0)<(32);
          $232 = ($219*178)|0;
          $233 = ($215*84)|0;
          $234 = (($233) + 5696)|0;
          $$pn = $231 ? $232 : $234;
          $$5332 = (($$pn) + 44032)|0;
          $235 = (($$5332) + ($$5338))|0;
          $236 = ($235>>>0)<(44032);
          if ($236) {
           $$2343 = 2;$$8 = $235;
           label = 86;
           break L7;
          } else {
           $$6339380 = 44032;$$6381 = $235;
          }
          while(1) {
           $237 = (($$6381) - ($$6339380))|0;
           $$0321379 = 0;$$4349378 = 0;
           while(1) {
            $$0377 = 0;$$5350376 = $$4349378;
            while(1) {
             $242 = ((124998 + (($$0321379*188)|0)|0) + ($$0377<<1)|0);
             $243 = HEAP16[$242>>1]|0;
             $244 = $243&65535;
             $245 = (($244) - ($$6339380))|0;
             $not$ = ($245>>>0)<=($237>>>0);
             $246 = $not$&1;
             $$5350$ = (($246) + ($$5350376))|0;
             $247 = (($$0377) + 1)|0;
             $exitcond = ($247|0)==(94);
             if ($exitcond) {
              break;
             } else {
              $$0377 = $247;$$5350376 = $$5350$;
             }
            }
            $241 = (($$0321379) + 1)|0;
            $exitcond396 = ($241|0)==(93);
            if ($exitcond396) {
             break;
            } else {
             $$0321379 = $241;$$4349378 = $$5350$;
            }
           }
           $238 = (($$6381) + 1)|0;
           $239 = (($$5350$) + ($$6381))|0;
           $240 = ($239>>>0)<($238>>>0);
           if ($240) {
            $$2343 = 2;$$8 = $239;
            label = 86;
            break L7;
           } else {
            $$6339380 = $238;$$6381 = $239;
           }
          }
          break;
         }
         default: {
          $252 = ($50>>>0)<($32>>>0);
          if ($252) {
           $$2343 = 1;$$8 = $50;
           label = 86;
           break L7;
          }
          $253 = (($50) - ($32))|0;
          $254 = ($253*5)|0;
          $255 = $254 >>> 2;
          $256 = (($14) + ($255)|0);
          $257 = HEAP8[$256>>0]|0;
          $258 = $257&255;
          $259 = $253 << 1;
          $260 = $259 & 6;
          $261 = $258 >>> $260;
          $262 = (($255) + 1)|0;
          $263 = (($14) + ($262)|0);
          $264 = HEAP8[$263>>0]|0;
          $265 = $264&255;
          $266 = (8 - ($260))|0;
          $267 = $265 << $266;
          $268 = $267 & 1023;
          $269 = $268 | $261;
          $270 = (142482 + ($269<<1)|0);
          $271 = HEAP16[$270>>1]|0;
          $272 = $271&65535;
          $273 = ($269|0)==(0);
          $$ = $273 ? $50 : $272;
          $274 = ($$|0)==(1);
          if ($274) {
           $$0340$ph = 84;
           label = 114;
           break L5;
          } else {
           $$2343 = 1;$$8 = $$;
           label = 86;
           break L7;
          }
         }
         }
        } while(0);
        if ((label|0) == 14) {
         label = 0;
         $59 = $$0327 & -2048;
         $60 = ($59|0)==(55296);
         $61 = ($$0327>>>0)>(1114111);
         $or$cond = $61 | $60;
         if ($or$cond) {
          $$0340$ph = 84;
          label = 114;
          break L5;
         } else {
          $$2343 = 4;$$8 = $$0327;
          label = 86;
          break;
         }
        }
        $114 = (($50) + -129)|0;
        $115 = ($114>>>0)>(125);
        if ($115) {
         $$0340$ph = 84;
         label = 114;
         break L5;
        }
        $116 = ($52>>>0)<(2);
        if ($116) {
         $$0340 = 22;
         break L5;
        }
        $117 = ((($$in)) + 1|0);
        $118 = HEAP8[$117>>0]|0;
        $119 = $118&255;
        $120 = ($118&255)<(161);
        $or$cond355 = $35 & $120;
        if ($or$cond355) {
         $$0340$ph = 84;
         label = 114;
         break L5;
        }
        $121 = (($119) + -64)|0;
        $122 = ($121>>>0)>(190);
        $123 = ($118<<24>>24)==(127);
        $or$cond5 = $123 | $122;
        if (!($or$cond5)) {
         $155 = ($121>>>0)>(63);
         $156 = (($119) + -65)|0;
         $$357 = $155 ? $156 : $121;
         $157 = ((38828 + (($114*380)|0)|0) + ($$357<<1)|0);
         $158 = HEAP16[$157>>1]|0;
         $159 = $158&65535;
         $$2343 = 2;$$8 = $159;
         label = 86;
         break;
        }
        $124 = (($119) + -48)|0;
        $125 = ($124>>>0)<(10);
        $or$cond356 = $36 & $125;
        if (!($or$cond356)) {
         $$0340$ph = 84;
         label = 114;
         break L5;
        }
        $126 = ($52>>>0)<(4);
        if ($126) {
         $$0340 = 22;
         break L5;
        }
        $127 = ((($$in)) + 2|0);
        $128 = HEAP8[$127>>0]|0;
        $129 = $128&255;
        $130 = (($129) + -129)|0;
        $131 = ($130>>>0)>(126);
        if ($131) {
         $$0340$ph = 84;
         label = 114;
         break L5;
        }
        $132 = ((($$in)) + 3|0);
        $133 = HEAP8[$132>>0]|0;
        $134 = $133&255;
        $135 = (($134) + -48)|0;
        $136 = ($135>>>0)>(9);
        if ($136) {
         $$0340$ph = 84;
         label = 114;
         break L5;
        }
        $137 = ($114*10)|0;
        $138 = (($119) + ($137))|0;
        $139 = ($138*1260)|0;
        $140 = ($130*10)|0;
        $141 = (($139) + -60352)|0;
        $142 = (($141) + ($140))|0;
        $143 = (($142) + ($135))|0;
        $$2335386 = 0;$$3330387 = $143;
        while(1) {
         $144 = (($$3330387) - ($$2335386))|0;
         $$0323385 = 0;$$0345384 = 0;
         while(1) {
          $$0322383 = 0;$$1346382 = $$0345384;
          while(1) {
           $149 = ((38828 + (($$0323385*380)|0)|0) + ($$0322383<<1)|0);
           $150 = HEAP16[$149>>1]|0;
           $151 = $150&65535;
           $152 = (($151) - ($$2335386))|0;
           $not$364 = ($152>>>0)<=($144>>>0);
           $153 = $not$364&1;
           $$1346$ = (($153) + ($$1346382))|0;
           $154 = (($$0322383) + 1)|0;
           $exitcond397 = ($154|0)==(190);
           if ($exitcond397) {
            break;
           } else {
            $$0322383 = $154;$$1346382 = $$1346$;
           }
          }
          $148 = (($$0323385) + 1)|0;
          $exitcond398 = ($148|0)==(126);
          if ($exitcond398) {
           break;
          } else {
           $$0323385 = $148;$$0345384 = $$1346$;
          }
         }
         $145 = (($$3330387) + 1)|0;
         $146 = (($$1346$) + ($$3330387))|0;
         $147 = ($146>>>0)<($145>>>0);
         if ($147) {
          $$2343 = 4;$$8 = $146;
          label = 86;
          break;
         } else {
          $$2335386 = $145;$$3330387 = $146;
         }
        }
       } else {
        $$2343 = 1;$$8 = $50;
        label = 86;
       }
      } while(0);
      if ((label|0) == 86) {
       label = 0;
       L111: do {
        switch ($22<<24>>24) {
        case -58:  {
         $275 = HEAP32[$4>>2]|0;
         $276 = ($275>>>0)<(4);
         if ($276) {
          $$0340$ph = 7;
          label = 114;
          break L5;
         }
         $277 = HEAP32[$3>>2]|0;
         HEAP32[$277>>2] = $$8;
         $278 = ((($277)) + 4|0);
         HEAP32[$3>>2] = $278;
         $$3 = $$0326389;$$sink14 = 4;$339 = $275;
         label = 111;
         break;
        }
        case -56:  {
         $279 = HEAP32[$4>>2]|0;
         $280 = ($279>>>0)<(4);
         if ($280) {
          $281 = (_wctomb($9,$$8)|0);
          $282 = ($279>>>0)<($281>>>0);
          if ($282) {
           label = 92;
           break L5;
          }
          $283 = HEAP32[$3>>2]|0;
          _memcpy(($283|0),($9|0),($281|0))|0;
          $$7352 = $281;$287 = $283;
         } else {
          $284 = HEAP32[$3>>2]|0;
          $285 = (_wctomb($284,$$8)|0);
          $$7352 = $285;$287 = $284;
         }
         $286 = (($287) + ($$7352)|0);
         HEAP32[$3>>2] = $286;
         $$3 = $$0326389;$$sink14 = $$7352;$339 = $279;
         label = 111;
         break;
        }
        case -57:  {
         $288 = ($$8>>>0)>(127);
         if ($288) {
          $$1 = $$0326389;
          label = 96;
         } else {
          $$2 = $$0326389;$$9 = $$8;
          label = 97;
         }
         break;
        }
        case -63: case -62: case -59: case -60:  {
         $319 = ($$8>>>0)<(65536);
         $or$cond362 = $46 | $319;
         $320 = HEAP32[$4>>2]|0;
         if ($or$cond362) {
          $321 = ($320>>>0)<(2);
          if ($321) {
           $$0340$ph = 7;
           label = 114;
           break L5;
          }
          $322 = ($$8>>>0)>(65535);
          $$$8 = $322 ? 65533 : $$8;
          $323 = HEAP32[$3>>2]|0;
          _put_16($323,$$$8,$37);
          $324 = ((($323)) + 2|0);
          HEAP32[$3>>2] = $324;
          $$3 = $$0326389;$$sink14 = 2;$339 = $320;
          label = 111;
          break L111;
         } else {
          $325 = ($320>>>0)<(4);
          if ($325) {
           $$0340$ph = 7;
           label = 114;
           break L5;
          }
          $326 = (($$8) + -65536)|0;
          $327 = HEAP32[$3>>2]|0;
          $328 = $326 >>> 10;
          $329 = $328 | 55296;
          _put_16($327,$329,$37);
          $330 = ((($327)) + 2|0);
          $331 = $326 & 1023;
          $332 = $331 | 56320;
          _put_16($330,$332,$37);
          $333 = ((($327)) + 4|0);
          HEAP32[$3>>2] = $333;
          $$3 = $$0326389;$$sink14 = 4;$339 = $320;
          label = 111;
          break L111;
         }
         break;
        }
        case -61: case -64:  {
         $334 = HEAP32[$4>>2]|0;
         $335 = ($334>>>0)<(4);
         if ($335) {
          $$0340$ph = 7;
          label = 114;
          break L5;
         }
         $336 = HEAP32[$3>>2]|0;
         _put_32($336,$$8,$37);
         $337 = ((($336)) + 4|0);
         HEAP32[$3>>2] = $337;
         $$3 = $$0326389;$$sink14 = 4;$339 = $334;
         label = 111;
         break;
        }
        default: {
         $$2 = $$0326389;$$9 = $$8;
         label = 97;
        }
        }
       } while(0);
       L129: while(1) {
        if ((label|0) == 96) {
         label = 0;
         $289 = (($$1) + 1)|0;
         $$2 = $289;$$9 = 42;
         label = 97;
         continue;
        }
        else if ((label|0) == 97) {
         label = 0;
         $290 = HEAP32[$4>>2]|0;
         $291 = ($290|0)==(0);
         if ($291) {
          $$0340$ph = 7;
          label = 114;
          break L5;
         }
         $292 = ($$9>>>0)<($42>>>0);
         if ($292) {
          $$10 = $$9;
         } else {
          if ($44) {
           $$1 = $$2;
           label = 96;
           continue;
          } else {
           $$11388 = 0;
          }
          while(1) {
           $296 = ($$11388*5)|0;
           $297 = $296 >>> 2;
           $298 = (($16) + ($297)|0);
           $299 = HEAP8[$298>>0]|0;
           $300 = $299&255;
           $301 = $$11388 << 1;
           $302 = $301 & 6;
           $303 = $300 >>> $302;
           $304 = (($297) + 1)|0;
           $305 = (($16) + ($304)|0);
           $306 = HEAP8[$305>>0]|0;
           $307 = $306&255;
           $308 = (8 - ($302))|0;
           $309 = $307 << $308;
           $310 = $309 & 1023;
           $311 = $310 | $303;
           $312 = (142482 + ($311<<1)|0);
           $313 = HEAP16[$312>>1]|0;
           $314 = $313&65535;
           $315 = ($$9|0)==($314|0);
           if ($315) {
            break;
           }
           $317 = (($$11388) + 1)|0;
           $318 = ($317>>>0)<($43>>>0);
           if ($318) {
            $$11388 = $317;
           } else {
            $$1 = $$2;
            label = 96;
            continue L129;
           }
          }
          $316 = (($$11388) + 128)|0;
          $$10 = $316;
         }
         $293 = $$10&255;
         $294 = HEAP32[$3>>2]|0;
         $295 = ((($294)) + 1|0);
         HEAP32[$3>>2] = $295;
         HEAP8[$294>>0] = $293;
         $$3 = $$2;$$sink14 = 1;$339 = $290;
         label = 111;
         continue;
        }
        else if ((label|0) == 111) {
         label = 0;
         $338 = (($339) - ($$sink14))|0;
         HEAP32[$4>>2] = $338;
         $$3344 = $$2343;$$4 = $$3;
         break;
        }
       }
      }
      $340 = HEAP32[$1>>2]|0;
      $341 = (($340) + ($$3344)|0);
      HEAP32[$1>>2] = $341;
      $342 = (($52) - ($$3344))|0;
      HEAP32[$2>>2] = $342;
      $343 = ($342|0)==(0);
      if ($343) {
       label = 113;
       break;
      } else {
       $$0326389 = $$4;$$in = $341;$52 = $342;
      }
     }
     if ((label|0) == 92) {
      $$0340 = 7;
     }
     else if ((label|0) == 113) {
      HEAP32[$24>>2] = $25;
      $$0325 = $$4;
      break;
     }
     else if ((label|0) == 114) {
      $$0340 = $$0340$ph;
     }
     $344 = (___errno_location()|0);
     HEAP32[$344>>2] = $$0340;
     HEAP32[$24>>2] = $25;
     $$0325 = -1;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0325|0);
}
function ___pthread_self_138() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function _mbrtowc($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$ = 0, $$0 = 0, $$03952 = 0, $$04051 = 0, $$04350 = 0, $$1 = 0, $$141 = 0, $$144 = 0, $$2 = 0, $$47 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($3|0)==(0|0);
 $$ = $5 ? 164360 : $3;
 $6 = HEAP32[$$>>2]|0;
 $7 = ($1|0)==(0|0);
 L1: do {
  if ($7) {
   $8 = ($6|0)==(0);
   if ($8) {
    $$0 = 0;
   } else {
    label = 17;
   }
  } else {
   $9 = ($0|0)==(0|0);
   $$47 = $9 ? $4 : $0;
   $10 = ($2|0)==(0);
   if ($10) {
    $$0 = -2;
   } else {
    $11 = ($6|0)==(0);
    if ($11) {
     $12 = HEAP8[$1>>0]|0;
     $13 = ($12<<24>>24)>(-1);
     if ($13) {
      $14 = $12&255;
      HEAP32[$$47>>2] = $14;
      $15 = ($12<<24>>24)!=(0);
      $16 = $15&1;
      $$0 = $16;
      break;
     }
     $17 = (___pthread_self_439()|0);
     $18 = ((($17)) + 188|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = HEAP32[$19>>2]|0;
     $not$ = ($20|0)==(0|0);
     $21 = HEAP8[$1>>0]|0;
     if ($not$) {
      $22 = $21 << 24 >> 24;
      $23 = $22 & 57343;
      HEAP32[$$47>>2] = $23;
      $$0 = 1;
      break;
     }
     $24 = $21&255;
     $25 = (($24) + -194)|0;
     $26 = ($25>>>0)>(50);
     if ($26) {
      label = 17;
      break;
     }
     $27 = ((($1)) + 1|0);
     $28 = (14424 + ($25<<2)|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = (($2) + -1)|0;
     $31 = ($30|0)==(0);
     if ($31) {
      $$2 = $29;
     } else {
      $$03952 = $27;$$04051 = $29;$$04350 = $30;
      label = 11;
     }
    } else {
     $$03952 = $1;$$04051 = $6;$$04350 = $2;
     label = 11;
    }
    L14: do {
     if ((label|0) == 11) {
      $32 = HEAP8[$$03952>>0]|0;
      $33 = $32&255;
      $34 = $33 >>> 3;
      $35 = (($34) + -16)|0;
      $36 = $$04051 >> 26;
      $37 = (($34) + ($36))|0;
      $38 = $35 | $37;
      $39 = ($38>>>0)>(7);
      if ($39) {
       label = 17;
       break L1;
      } else {
       $$1 = $$03952;$$141 = $$04051;$$144 = $$04350;$43 = $32;
      }
      while(1) {
       $40 = $$141 << 6;
       $41 = ((($$1)) + 1|0);
       $42 = $43&255;
       $44 = (($42) + -128)|0;
       $45 = $44 | $40;
       $46 = (($$144) + -1)|0;
       $47 = ($45|0)<(0);
       if (!($47)) {
        break;
       }
       $49 = ($46|0)==(0);
       if ($49) {
        $$2 = $45;
        break L14;
       }
       $50 = HEAP8[$41>>0]|0;
       $51 = $50 & -64;
       $52 = ($51<<24>>24)==(-128);
       if ($52) {
        $$1 = $41;$$141 = $45;$$144 = $46;$43 = $50;
       } else {
        label = 17;
        break L1;
       }
      }
      HEAP32[$$>>2] = 0;
      HEAP32[$$47>>2] = $45;
      $48 = (($2) - ($46))|0;
      $$0 = $48;
      break L1;
     }
    } while(0);
    HEAP32[$$>>2] = $$2;
    $$0 = -2;
   }
  }
 } while(0);
 if ((label|0) == 17) {
  HEAP32[$$>>2] = 0;
  $53 = (___errno_location()|0);
  HEAP32[$53>>2] = 84;
  $$0 = -1;
 }
 STACKTOP = sp;return ($$0|0);
}
function _get_32($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1 & 3;
 $3 = (($0) + ($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&255;
 $6 = $5 << 24;
 $7 = $2 ^ 1;
 $8 = (($0) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $10 << 16;
 $12 = $11 | $6;
 $13 = $2 ^ 2;
 $14 = (($0) + ($13)|0);
 $15 = HEAP8[$14>>0]|0;
 $16 = $15&255;
 $17 = $16 << 8;
 $18 = $12 | $17;
 $19 = $2 ^ 3;
 $20 = (($0) + ($19)|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = $21&255;
 $23 = $18 | $22;
 return ($23|0);
}
function _get_16($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1 & 1;
 $3 = (($0) + ($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = $4&255;
 $6 = $5 << 8;
 $7 = $2 ^ 1;
 $8 = (($0) + ($7)|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = $9&255;
 $11 = $6 | $10;
 return ($11|0);
}
function _put_16($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $2 & 1;
 $4 = $1 >>> 8;
 $5 = $4&255;
 $6 = (($0) + ($3)|0);
 HEAP8[$6>>0] = $5;
 $7 = $1&255;
 $8 = $3 ^ 1;
 $9 = (($0) + ($8)|0);
 HEAP8[$9>>0] = $7;
 return;
}
function _put_32($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $2 & 3;
 $4 = $1 >>> 24;
 $5 = $4&255;
 $6 = (($0) + ($3)|0);
 HEAP8[$6>>0] = $5;
 $7 = $1 >>> 16;
 $8 = $7&255;
 $9 = $3 ^ 1;
 $10 = (($0) + ($9)|0);
 HEAP8[$10>>0] = $8;
 $11 = $1 >>> 8;
 $12 = $11&255;
 $13 = $3 ^ 2;
 $14 = (($0) + ($13)|0);
 HEAP8[$14>>0] = $12;
 $15 = $1&255;
 $16 = $3 ^ 3;
 $17 = (($0) + ($16)|0);
 HEAP8[$17>>0] = $15;
 return;
}
function ___pthread_self_439() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function _isdigit_l($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $isdigit = 0, $isdigittmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $isdigittmp = (($0) + -48)|0;
 $isdigit = ($isdigittmp>>>0)<(10);
 $2 = $isdigit&1;
 return ($2|0);
}
function _isxdigit_l($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_isxdigit($0)|0);
 return ($2|0);
}
function ___get_locale($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$093$lcssa = 0, $$093118 = 0, $$094114 = 0, $$1 = 0, $$1100111 = 0, $$195112 = 0, $$2122 = 0, $$2123 = 0, $$2124 = 0, $$4 = 0, $$neg = 0, $$pre121 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0;
 var $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $__c_dot_utf8$$4 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(272|0);
 $2 = sp + 8|0;
 $3 = sp;
 $4 = HEAP8[$1>>0]|0;
 $5 = ($4<<24>>24)==(0);
 do {
  if ($5) {
   $6 = (_getenv((155010|0))|0);
   $7 = ($6|0)==(0|0);
   if (!($7)) {
    $8 = HEAP8[$6>>0]|0;
    $9 = ($8<<24>>24)==(0);
    if (!($9)) {
     $$1 = $6;
     break;
    }
   }
   $10 = (155017 + (($0*12)|0)|0);
   $11 = (_getenv(($10|0))|0);
   $12 = ($11|0)==(0|0);
   if (!($12)) {
    $13 = HEAP8[$11>>0]|0;
    $14 = ($13<<24>>24)==(0);
    if (!($14)) {
     $$1 = $11;
     break;
    }
   }
   $15 = (_getenv((155089|0))|0);
   $16 = ($15|0)==(0|0);
   if (!($16)) {
    $17 = HEAP8[$15>>0]|0;
    $18 = ($17<<24>>24)==(0);
    if (!($18)) {
     $$1 = $15;
     break;
    }
   }
   $$1 = 155094;
  } else {
   $$1 = $1;
  }
 } while(0);
 $$093118 = 0;
 L13: while(1) {
  $19 = (($$1) + ($$093118)|0);
  $20 = HEAP8[$19>>0]|0;
  switch ($20<<24>>24) {
  case 47: case 0:  {
   $$093$lcssa = $$093118;
   break L13;
   break;
  }
  default: {
  }
  }
  $21 = (($$093118) + 1)|0;
  $22 = ($21>>>0)<(15);
  if ($22) {
   $$093118 = $21;
  } else {
   $$093$lcssa = $21;
   break;
  }
 }
 $23 = HEAP8[$$1>>0]|0;
 $24 = ($23<<24>>24)==(46);
 if ($24) {
  $$2122 = 155094;
  label = 15;
 } else {
  $25 = (($$1) + ($$093$lcssa)|0);
  $26 = HEAP8[$25>>0]|0;
  $27 = ($26<<24>>24)==(0);
  if ($27) {
   $28 = ($23<<24>>24)==(67);
   if ($28) {
    $$2122 = $$1;
    label = 15;
   } else {
    $$2123 = $$1;
    label = 16;
   }
  } else {
   $$2122 = 155094;
   label = 15;
  }
 }
 if ((label|0) == 15) {
  $29 = ((($$2122)) + 1|0);
  $30 = HEAP8[$29>>0]|0;
  $31 = ($30<<24>>24)==(0);
  if ($31) {
   $$2124 = $$2122;
   label = 18;
  } else {
   $$2123 = $$2122;
   label = 16;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $32 = (_strcmp($$2123,155094)|0);
   $33 = ($32|0)==(0);
   if ($33) {
    $$2124 = $$2123;
    label = 18;
   } else {
    $34 = (_strcmp($$2123,155102)|0);
    $35 = ($34|0)==(0);
    if ($35) {
     $$2124 = $$2123;
     label = 18;
    } else {
     $40 = HEAP32[41091]|0;
     $41 = ($40|0)==(0|0);
     if (!($41)) {
      $$094114 = $40;
      while(1) {
       $42 = ((($$094114)) + 8|0);
       $43 = (_strcmp($$2123,$42)|0);
       $44 = ($43|0)==(0);
       if ($44) {
        $$0 = $$094114;
        break L22;
       }
       $45 = ((($$094114)) + 24|0);
       $46 = HEAP32[$45>>2]|0;
       $47 = ($46|0)==(0|0);
       if ($47) {
        break;
       } else {
        $$094114 = $46;
       }
      }
     }
     ___lock((164368|0));
     $48 = HEAP32[41091]|0;
     $49 = ($48|0)==(0|0);
     L32: do {
      if (!($49)) {
       $$195112 = $48;
       while(1) {
        $50 = ((($$195112)) + 8|0);
        $51 = (_strcmp($$2123,$50)|0);
        $52 = ($51|0)==(0);
        if ($52) {
         break;
        }
        $53 = ((($$195112)) + 24|0);
        $54 = HEAP32[$53>>2]|0;
        $55 = ($54|0)==(0|0);
        if ($55) {
         break L32;
        } else {
         $$195112 = $54;
        }
       }
       ___unlock((164368|0));
       $$0 = $$195112;
       break L22;
      }
     } while(0);
     $56 = HEAP32[(164280)>>2]|0;
     $57 = ($56|0)==(0);
     L39: do {
      if ($57) {
       $58 = (_getenv((155108|0))|0);
       $59 = ($58|0)==(0|0);
       if ($59) {
        label = 41;
       } else {
        $60 = HEAP8[$58>>0]|0;
        $61 = ($60<<24>>24)==(0);
        if ($61) {
         label = 41;
        } else {
         $62 = (254 - ($$093$lcssa))|0;
         $63 = (($$093$lcssa) + 1)|0;
         $$1100111 = $58;
         while(1) {
          $64 = (___strchrnul($$1100111,58)|0);
          $65 = $64;
          $66 = $$1100111;
          $67 = (($65) - ($66))|0;
          $68 = HEAP8[$64>>0]|0;
          $69 = ($68<<24>>24)!=(0);
          $$neg = $69 << 31 >> 31;
          $70 = (($$neg) + ($67))|0;
          $71 = ($70>>>0)<($62>>>0);
          if ($71) {
           _memcpy(($2|0),($$1100111|0),($70|0))|0;
           $72 = (($2) + ($70)|0);
           HEAP8[$72>>0] = 47;
           $73 = ((($72)) + 1|0);
           _memcpy(($73|0),($$2123|0),($$093$lcssa|0))|0;
           $74 = (($63) + ($70))|0;
           $75 = (($2) + ($74)|0);
           HEAP8[$75>>0] = 0;
           $76 = (___map_file(($2|0),($3|0))|0);
           $77 = ($76|0)==(0|0);
           if (!($77)) {
            break;
           }
           $$pre121 = HEAP8[$64>>0]|0;
           $87 = $$pre121;
          } else {
           $87 = $68;
          }
          $86 = ($87<<24>>24)!=(0);
          $88 = $86&1;
          $89 = (($64) + ($88)|0);
          $90 = HEAP8[$89>>0]|0;
          $91 = ($90<<24>>24)==(0);
          if ($91) {
           label = 41;
           break L39;
          } else {
           $$1100111 = $89;
          }
         }
         $78 = (_malloc(28)|0);
         $79 = ($78|0)==(0|0);
         $80 = HEAP32[$3>>2]|0;
         if ($79) {
          (___munmap($76,$80)|0);
          label = 41;
          break;
         } else {
          HEAP32[$78>>2] = $76;
          $81 = ((($78)) + 4|0);
          HEAP32[$81>>2] = $80;
          $82 = ((($78)) + 8|0);
          _memcpy(($82|0),($$2123|0),($$093$lcssa|0))|0;
          $83 = (($82) + ($$093$lcssa)|0);
          HEAP8[$83>>0] = 0;
          $84 = HEAP32[41091]|0;
          $85 = ((($78)) + 24|0);
          HEAP32[$85>>2] = $84;
          HEAP32[41091] = $78;
          $$4 = $78;
          break;
         }
        }
       }
      } else {
       label = 41;
      }
     } while(0);
     if ((label|0) == 41) {
      $92 = (_malloc(28)|0);
      $93 = ($92|0)==(0|0);
      if ($93) {
       $$4 = $92;
      } else {
       $94 = HEAP32[3588]|0;
       HEAP32[$92>>2] = $94;
       $95 = HEAP32[(14356)>>2]|0;
       $96 = ((($92)) + 4|0);
       HEAP32[$96>>2] = $95;
       $97 = ((($92)) + 8|0);
       _memcpy(($97|0),($$2123|0),($$093$lcssa|0))|0;
       $98 = (($97) + ($$093$lcssa)|0);
       HEAP8[$98>>0] = 0;
       $99 = HEAP32[41091]|0;
       $100 = ((($92)) + 24|0);
       HEAP32[$100>>2] = $99;
       HEAP32[41091] = $92;
       $$4 = $92;
      }
     }
     $101 = ($$4|0)==(0|0);
     $102 = ($0|0)==(0);
     $or$cond = $102 & $101;
     $__c_dot_utf8$$4 = $or$cond ? 14352 : $$4;
     ___unlock((164368|0));
     $$0 = $__c_dot_utf8$$4;
    }
   }
  }
 } while(0);
 do {
  if ((label|0) == 18) {
   $36 = ($0|0)==(0);
   if ($36) {
    $37 = ((($$2124)) + 1|0);
    $38 = HEAP8[$37>>0]|0;
    $39 = ($38<<24>>24)==(46);
    if ($39) {
     $$0 = 14352;
     break;
    }
   }
   $$0 = 0;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function ___newlocale($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$033$ = 0, $$03339 = 0, $$03440 = 0, $$036 = 0, $$13538 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, $exitcond = 0, $exitcond41 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp;
 $4 = (___loc_is_allocated($2)|0);
 $5 = ($4|0)==(0);
 L1: do {
  if ($5) {
   $6 = ($2|0)==(0|0);
   $$03339 = 0;$$13538 = 0;
   while(1) {
    $13 = 1 << $$13538;
    $14 = $13 & $0;
    $15 = ($14|0)!=(0);
    $brmerge = $6 | $15;
    if ($brmerge) {
     $18 = $15 ? $1 : 169457;
     $19 = (___get_locale($$13538,$18)|0);
     $$0 = $19;
    } else {
     $16 = (($2) + ($$13538<<2)|0);
     $17 = HEAP32[$16>>2]|0;
     $$0 = $17;
    }
    $not$ = ($$0|0)!=(0|0);
    $20 = $not$&1;
    $$033$ = (($20) + ($$03339))|0;
    $21 = (($3) + ($$13538<<2)|0);
    HEAP32[$21>>2] = $$0;
    $22 = (($$13538) + 1)|0;
    $exitcond = ($22|0)==(6);
    if ($exitcond) {
     break;
    } else {
     $$03339 = $$033$;$$13538 = $22;
    }
   }
   switch ($$033$|0) {
   case 0:  {
    $$036 = 164336;
    break L1;
    break;
   }
   case 1:  {
    $23 = HEAP32[$3>>2]|0;
    $24 = ($23|0)==(14352|0);
    if ($24) {
     $$036 = 14400;
     break L1;
    }
    break;
   }
   default: {
   }
   }
   $$036 = $2;
  } else {
   $$03440 = 0;
   while(1) {
    $7 = 1 << $$03440;
    $8 = $7 & $0;
    $9 = ($8|0)==(0);
    if (!($9)) {
     $10 = (($2) + ($$03440<<2)|0);
     $11 = (___get_locale($$03440,$1)|0);
     HEAP32[$10>>2] = $11;
    }
    $12 = (($$03440) + 1)|0;
    $exitcond41 = ($12|0)==(6);
    if ($exitcond41) {
     $$036 = $2;
     break;
    } else {
     $$03440 = $12;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$036|0);
}
function _snprintf($0,$1,$2,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $varargs = $varargs|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 HEAP32[$3>>2] = $varargs;
 $4 = (_vsnprintf($0,$1,$2,$3)|0);
 STACKTOP = sp;return ($4|0);
}
function _strcpy($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (___stpcpy($0,$1)|0);
 return ($0|0);
}
function ___stpcpy($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$lcssa = 0, $$025$lcssa = 0, $$02536 = 0, $$026$lcssa = 0, $$02642 = 0, $$027$lcssa = 0, $$02741 = 0, $$029 = 0, $$037 = 0, $$1$ph = 0, $$128$ph = 0, $$12834 = 0, $$135 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1;
 $3 = $0;
 $4 = $2 ^ $3;
 $5 = $4 & 3;
 $6 = ($5|0)==(0);
 L1: do {
  if ($6) {
   $7 = $2 & 3;
   $8 = ($7|0)==(0);
   if ($8) {
    $$026$lcssa = $1;$$027$lcssa = $0;
   } else {
    $$02642 = $1;$$02741 = $0;
    while(1) {
     $9 = HEAP8[$$02642>>0]|0;
     HEAP8[$$02741>>0] = $9;
     $10 = ($9<<24>>24)==(0);
     if ($10) {
      $$029 = $$02741;
      break L1;
     }
     $11 = ((($$02642)) + 1|0);
     $12 = ((($$02741)) + 1|0);
     $13 = $11;
     $14 = $13 & 3;
     $15 = ($14|0)==(0);
     if ($15) {
      $$026$lcssa = $11;$$027$lcssa = $12;
      break;
     } else {
      $$02642 = $11;$$02741 = $12;
     }
    }
   }
   $16 = HEAP32[$$026$lcssa>>2]|0;
   $17 = (($16) + -16843009)|0;
   $18 = $16 & -2139062144;
   $19 = $18 ^ -2139062144;
   $20 = $19 & $17;
   $21 = ($20|0)==(0);
   if ($21) {
    $$02536 = $$027$lcssa;$$037 = $$026$lcssa;$24 = $16;
    while(1) {
     $22 = ((($$037)) + 4|0);
     $23 = ((($$02536)) + 4|0);
     HEAP32[$$02536>>2] = $24;
     $25 = HEAP32[$22>>2]|0;
     $26 = (($25) + -16843009)|0;
     $27 = $25 & -2139062144;
     $28 = $27 ^ -2139062144;
     $29 = $28 & $26;
     $30 = ($29|0)==(0);
     if ($30) {
      $$02536 = $23;$$037 = $22;$24 = $25;
     } else {
      $$0$lcssa = $22;$$025$lcssa = $23;
      break;
     }
    }
   } else {
    $$0$lcssa = $$026$lcssa;$$025$lcssa = $$027$lcssa;
   }
   $$1$ph = $$0$lcssa;$$128$ph = $$025$lcssa;
   label = 8;
  } else {
   $$1$ph = $1;$$128$ph = $0;
   label = 8;
  }
 } while(0);
 if ((label|0) == 8) {
  $31 = HEAP8[$$1$ph>>0]|0;
  HEAP8[$$128$ph>>0] = $31;
  $32 = ($31<<24>>24)==(0);
  if ($32) {
   $$029 = $$128$ph;
  } else {
   $$12834 = $$128$ph;$$135 = $$1$ph;
   while(1) {
    $33 = ((($$135)) + 1|0);
    $34 = ((($$12834)) + 1|0);
    $35 = HEAP8[$33>>0]|0;
    HEAP8[$34>>0] = $35;
    $36 = ($35<<24>>24)==(0);
    if ($36) {
     $$029 = $34;
     break;
    } else {
     $$12834 = $34;$$135 = $33;
    }
   }
  }
 }
 return ($$029|0);
}
function ___uselocale($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___pthread_self_230()|0);
 $2 = ((($1)) + 188|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($0|0)==(0|0);
 if (!($4)) {
  $5 = ($0|0)==((-1)|0);
  $6 = $5 ? (164312) : $0;
  HEAP32[$2>>2] = $6;
 }
 $7 = ($3|0)==((164312)|0);
 $8 = $7 ? (-1) : $3;
 return ($8|0);
}
function ___pthread_self_230() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function _wcscmp($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$015 = 0, $$0914 = 0, $$lcssa = 0, $$lcssa11 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond10 = 0, $or$cond1013 = 0;
 var $or$cond12 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = HEAP32[$1>>2]|0;
 $4 = ($2|0)!=($3|0);
 $5 = ($2|0)==(0);
 $or$cond12 = $5 | $4;
 $6 = ($3|0)==(0);
 $or$cond1013 = $6 | $or$cond12;
 if ($or$cond1013) {
  $$lcssa = $3;$$lcssa11 = $2;
 } else {
  $$015 = $1;$$0914 = $0;
  while(1) {
   $7 = ((($$0914)) + 4|0);
   $8 = ((($$015)) + 4|0);
   $9 = HEAP32[$7>>2]|0;
   $10 = HEAP32[$8>>2]|0;
   $11 = ($9|0)!=($10|0);
   $12 = ($9|0)==(0);
   $or$cond = $12 | $11;
   $13 = ($10|0)==(0);
   $or$cond10 = $13 | $or$cond;
   if ($or$cond10) {
    $$lcssa = $10;$$lcssa11 = $9;
    break;
   } else {
    $$015 = $8;$$0914 = $7;
   }
  }
 }
 $14 = (($$lcssa11) - ($$lcssa))|0;
 return ($14|0);
}
function _wmemcpy($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$058 = 0, $$067 = 0, $$09 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$058 = $2;$$067 = $1;$$09 = $0;
  while(1) {
   $4 = (($$058) + -1)|0;
   $5 = ((($$067)) + 4|0);
   $6 = HEAP32[$$067>>2]|0;
   $7 = ((($$09)) + 4|0);
   HEAP32[$$09>>2] = $6;
   $8 = ($4|0)==(0);
   if ($8) {
    break;
   } else {
    $$058 = $4;$$067 = $5;$$09 = $7;
   }
  }
 }
 return ($0|0);
}
function _ioctl($0,$1,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $varargs = $varargs|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arglist_current = 0, $arglist_next = 0, $expanded = 0, $expanded2 = 0, $expanded4 = 0, $expanded5 = 0, $expanded6 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0;
 var $vararg_ptr2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $2 = sp + 16|0;
 HEAP32[$2>>2] = $varargs;
 $arglist_current = HEAP32[$2>>2]|0;
 $3 = $arglist_current;
 $4 = ((0) + 4|0);
 $expanded2 = $4;
 $expanded = (($expanded2) - 1)|0;
 $5 = (($3) + ($expanded))|0;
 $6 = ((0) + 4|0);
 $expanded6 = $6;
 $expanded5 = (($expanded6) - 1)|0;
 $expanded4 = $expanded5 ^ -1;
 $7 = $5 & $expanded4;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 $arglist_next = ((($8)) + 4|0);
 HEAP32[$2>>2] = $arglist_next;
 $10 = $9;
 HEAP32[$vararg_buffer>>2] = $0;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $1;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $10;
 $11 = (___syscall54(54,($vararg_buffer|0))|0);
 $12 = (___syscall_ret($11)|0);
 STACKTOP = sp;return ($12|0);
}
function _fwrite($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$ = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = Math_imul($2, $1)|0;
 $5 = ($1|0)==(0);
 $$ = $5 ? 0 : $2;
 $6 = ((($3)) + 76|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)>(-1);
 if ($8) {
  $10 = (___lockfile($3)|0);
  $phitmp = ($10|0)==(0);
  $11 = (___fwritex($0,$4,$3)|0);
  if ($phitmp) {
   $13 = $11;
  } else {
   ___unlockfile($3);
   $13 = $11;
  }
 } else {
  $9 = (___fwritex($0,$4,$3)|0);
  $13 = $9;
 }
 $12 = ($13|0)==($4|0);
 if ($12) {
  $15 = $$;
 } else {
  $14 = (($13>>>0) / ($1>>>0))&-1;
  $15 = $14;
 }
 return ($15|0);
}
function ___overflow($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $3 = $1&255;
 HEAP8[$2>>0] = $3;
 $4 = ((($0)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 if ($6) {
  $7 = (___towrite($0)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   $$pre = HEAP32[$4>>2]|0;
   $12 = $$pre;
   label = 4;
  } else {
   $$0 = -1;
  }
 } else {
  $12 = $5;
  label = 4;
 }
 do {
  if ((label|0) == 4) {
   $9 = ((($0)) + 20|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10>>>0)<($12>>>0);
   if ($11) {
    $13 = $1 & 255;
    $14 = ((($0)) + 75|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = $15 << 24 >> 24;
    $17 = ($13|0)==($16|0);
    if (!($17)) {
     $18 = ((($10)) + 1|0);
     HEAP32[$9>>2] = $18;
     HEAP8[$10>>0] = $3;
     $$0 = $13;
     break;
    }
   }
   $19 = ((($0)) + 36|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = (FUNCTION_TABLE_iiii[$20 & 2047]($0,$2,1)|0);
   $22 = ($21|0)==(1);
   if ($22) {
    $23 = HEAP8[$2>>0]|0;
    $24 = $23&255;
    $$0 = $24;
   } else {
    $$0 = -1;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _mbtowc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($1|0)==(0|0);
 L1: do {
  if ($4) {
   $$0 = 0;
  } else {
   $5 = ($2|0)==(0);
   do {
    if (!($5)) {
     $6 = ($0|0)==(0|0);
     $$ = $6 ? $3 : $0;
     $7 = HEAP8[$1>>0]|0;
     $8 = ($7<<24>>24)>(-1);
     if ($8) {
      $9 = $7&255;
      HEAP32[$$>>2] = $9;
      $10 = ($7<<24>>24)!=(0);
      $11 = $10&1;
      $$0 = $11;
      break L1;
     }
     $12 = (___pthread_self_445()|0);
     $13 = ((($12)) + 188|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = HEAP32[$14>>2]|0;
     $not$ = ($15|0)==(0|0);
     $16 = HEAP8[$1>>0]|0;
     if ($not$) {
      $17 = $16 << 24 >> 24;
      $18 = $17 & 57343;
      HEAP32[$$>>2] = $18;
      $$0 = 1;
      break L1;
     }
     $19 = $16&255;
     $20 = (($19) + -194)|0;
     $21 = ($20>>>0)>(50);
     if (!($21)) {
      $22 = ((($1)) + 1|0);
      $23 = (14424 + ($20<<2)|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ($2>>>0)<(4);
      if ($25) {
       $26 = ($2*6)|0;
       $27 = (($26) + -6)|0;
       $28 = -2147483648 >>> $27;
       $29 = $24 & $28;
       $30 = ($29|0)==(0);
       if (!($30)) {
        break;
       }
      }
      $31 = HEAP8[$22>>0]|0;
      $32 = $31&255;
      $33 = $32 >>> 3;
      $34 = (($33) + -16)|0;
      $35 = $24 >> 26;
      $36 = (($33) + ($35))|0;
      $37 = $34 | $36;
      $38 = ($37>>>0)>(7);
      if (!($38)) {
       $39 = $24 << 6;
       $40 = (($32) + -128)|0;
       $41 = $40 | $39;
       $42 = ($41|0)<(0);
       if (!($42)) {
        HEAP32[$$>>2] = $41;
        $$0 = 2;
        break L1;
       }
       $43 = ((($1)) + 2|0);
       $44 = HEAP8[$43>>0]|0;
       $45 = $44&255;
       $46 = (($45) + -128)|0;
       $47 = ($46>>>0)>(63);
       if (!($47)) {
        $48 = $41 << 6;
        $49 = $46 | $48;
        $50 = ($49|0)<(0);
        if (!($50)) {
         HEAP32[$$>>2] = $49;
         $$0 = 3;
         break L1;
        }
        $51 = ((($1)) + 3|0);
        $52 = HEAP8[$51>>0]|0;
        $53 = $52&255;
        $54 = (($53) + -128)|0;
        $55 = ($54>>>0)>(63);
        if (!($55)) {
         $56 = $49 << 6;
         $57 = $54 | $56;
         HEAP32[$$>>2] = $57;
         $$0 = 4;
         break L1;
        }
       }
      }
     }
    }
   } while(0);
   $58 = (___errno_location()|0);
   HEAP32[$58>>2] = 84;
   $$0 = -1;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function ___pthread_self_445() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function ___ofl_lock() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 ___lock((164376|0));
 return (164384|0);
}
function ___ofl_unlock() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 ___unlock((164376|0));
 return;
}
function _fflush($0) {
 $0 = $0|0;
 var $$0 = 0, $$023 = 0, $$02325 = 0, $$02327 = 0, $$024$lcssa = 0, $$02426 = 0, $$1 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0|0);
 do {
  if ($1) {
   $8 = HEAP32[3814]|0;
   $9 = ($8|0)==(0|0);
   if ($9) {
    $29 = 0;
   } else {
    $10 = HEAP32[3814]|0;
    $11 = (_fflush($10)|0);
    $29 = $11;
   }
   $12 = (___ofl_lock()|0);
   $$02325 = HEAP32[$12>>2]|0;
   $13 = ($$02325|0)==(0|0);
   if ($13) {
    $$024$lcssa = $29;
   } else {
    $$02327 = $$02325;$$02426 = $29;
    while(1) {
     $14 = ((($$02327)) + 76|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)>(-1);
     if ($16) {
      $17 = (___lockfile($$02327)|0);
      $26 = $17;
     } else {
      $26 = 0;
     }
     $18 = ((($$02327)) + 20|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ((($$02327)) + 28|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($19>>>0)>($21>>>0);
     if ($22) {
      $23 = (___fflush_unlocked($$02327)|0);
      $24 = $23 | $$02426;
      $$1 = $24;
     } else {
      $$1 = $$02426;
     }
     $25 = ($26|0)==(0);
     if (!($25)) {
      ___unlockfile($$02327);
     }
     $27 = ((($$02327)) + 56|0);
     $$023 = HEAP32[$27>>2]|0;
     $28 = ($$023|0)==(0|0);
     if ($28) {
      $$024$lcssa = $$1;
      break;
     } else {
      $$02327 = $$023;$$02426 = $$1;
     }
    }
   }
   ___ofl_unlock();
   $$0 = $$024$lcssa;
  } else {
   $2 = ((($0)) + 76|0);
   $3 = HEAP32[$2>>2]|0;
   $4 = ($3|0)>(-1);
   if (!($4)) {
    $5 = (___fflush_unlocked($0)|0);
    $$0 = $5;
    break;
   }
   $6 = (___lockfile($0)|0);
   $phitmp = ($6|0)==(0);
   $7 = (___fflush_unlocked($0)|0);
   if ($phitmp) {
    $$0 = $7;
   } else {
    ___unlockfile($0);
    $$0 = $7;
   }
  }
 } while(0);
 return ($$0|0);
}
function ___fflush_unlocked($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 20|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 28|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($2>>>0)>($4>>>0);
 if ($5) {
  $6 = ((($0)) + 36|0);
  $7 = HEAP32[$6>>2]|0;
  (FUNCTION_TABLE_iiii[$7 & 2047]($0,0,0)|0);
  $8 = HEAP32[$1>>2]|0;
  $9 = ($8|0)==(0|0);
  if ($9) {
   $$0 = -1;
  } else {
   label = 3;
  }
 } else {
  label = 3;
 }
 if ((label|0) == 3) {
  $10 = ((($0)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($0)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($11>>>0)<($13>>>0);
  if ($14) {
   $15 = $11;
   $16 = $13;
   $17 = (($15) - ($16))|0;
   $18 = ((($0)) + 40|0);
   $19 = HEAP32[$18>>2]|0;
   (FUNCTION_TABLE_iiii[$19 & 2047]($0,$17,1)|0);
  }
  $20 = ((($0)) + 16|0);
  HEAP32[$20>>2] = 0;
  HEAP32[$3>>2] = 0;
  HEAP32[$1>>2] = 0;
  HEAP32[$12>>2] = 0;
  HEAP32[$10>>2] = 0;
  $$0 = 0;
 }
 return ($$0|0);
}
function _sscanf($0,$1,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $varargs = $varargs|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 HEAP32[$2>>2] = $varargs;
 $3 = (_vsscanf($0,$1,$2)|0);
 STACKTOP = sp;return ($3|0);
}
function _vsscanf($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $3 = sp;
 dest=$3; stop=dest+124|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $4 = ((($3)) + 32|0);
 HEAP32[$4>>2] = 942;
 $5 = ((($3)) + 44|0);
 HEAP32[$5>>2] = $0;
 $6 = ((($3)) + 76|0);
 HEAP32[$6>>2] = -1;
 $7 = ((($3)) + 84|0);
 HEAP32[$7>>2] = $0;
 $8 = (_vfscanf($3,$1,$2)|0);
 STACKTOP = sp;return ($8|0);
}
function _do_read($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (___string_read($0,$1,$2)|0);
 return ($3|0);
}
function _vfscanf($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$$0266 = 0, $$$0268 = 0, $$$0305 = 0, $$$3 = 0, $$0266$lcssa = 0, $$0266417 = 0, $$0268 = 0, $$0272 = 0, $$0273429 = 0, $$0276$ph = 0, $$0278$ph = 0, $$0278$ph$phi = 0, $$0278$ph336 = 0, $$0283428 = 0, $$0286420 = 0, $$0288$ = 0, $$0288425 = 0, $$0292 = 0, $$0293 = 0;
 var $$0305423 = 0, $$10 = 0, $$11 = 0, $$1267 = 0, $$1271 = 0, $$1274 = 0, $$1277$ph = 0, $$1279 = 0, $$1284 = 0, $$1289 = 0, $$2 = 0, $$2275 = 0, $$2280 = 0, $$2280$ph = 0, $$2280$ph$phi = 0, $$2285 = 0, $$2290 = 0, $$2307$ph = 0, $$3$lcssa = 0, $$319 = 0;
 var $$320 = 0, $$321 = 0, $$322 = 0, $$327 = 0, $$328$le439 = 0, $$328$le441 = 0, $$3281 = 0, $$3291 = 0, $$3416 = 0, $$4282 = 0, $$4309 = 0, $$5 = 0, $$5299 = 0, $$5310 = 0, $$6 = 0, $$6311 = 0, $$7 = 0, $$7$ph = 0, $$7312 = 0, $$8 = 0;
 var $$8313 = 0, $$9 = 0, $$9314 = 0, $$9314$ph = 0, $$lcssa355 = 0, $$not = 0, $$old4 = 0, $$ph = 0, $$ph353 = 0, $$pre = 0, $$pre$phi516Z2D = 0, $$pre507 = 0, $$pre509 = 0, $$pre511 = 0, $$pre512 = 0, $$pre513 = 0, $$pre514 = 0, $$pre515 = 0, $$sink443 = 0, $$sroa$2$0$$sroa_idx13 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0.0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0.0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $arglist_current = 0, $arglist_next = 0, $expanded = 0, $expanded1 = 0, $expanded3 = 0, $expanded4 = 0, $expanded5 = 0, $factor = 0, $factor331 = 0, $isdigit = 0;
 var $isdigit316 = 0, $isdigit316415 = 0, $isdigittmp = 0, $isdigittmp315 = 0, $isdigittmp315414 = 0, $narrow = 0, $narrow469 = 0, $or$cond = 0, $or$cond3 = 0, $or$cond318 = 0, $or$cond5 = 0, $trunc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 288|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(288|0);
 $3 = sp + 8|0;
 $4 = sp + 17|0;
 $5 = sp;
 $6 = sp + 16|0;
 $7 = ((($0)) + 76|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)>(-1);
 if ($9) {
  $10 = (___lockfile($0)|0);
  $306 = $10;
 } else {
  $306 = 0;
 }
 $11 = HEAP8[$1>>0]|0;
 $12 = ($11<<24>>24)==(0);
 L4: do {
  if ($12) {
   $$3291 = 0;
  } else {
   $13 = ((($0)) + 4|0);
   $14 = ((($0)) + 100|0);
   $15 = ((($0)) + 108|0);
   $16 = ((($0)) + 8|0);
   $17 = ((($4)) + 10|0);
   $18 = ((($4)) + 33|0);
   $$sroa$2$0$$sroa_idx13 = ((($3)) + 4|0);
   $19 = ((($4)) + 46|0);
   $20 = ((($4)) + 94|0);
   $21 = ((($4)) + 1|0);
   $22 = ((($4)) + 1|0);
   $$0273429 = $1;$$0283428 = 0;$$0288425 = 0;$$0305423 = 0;$102 = 0;$24 = $11;
   L6: while(1) {
    $23 = $24&255;
    $25 = (_isspace($23)|0);
    $26 = ($25|0)==(0);
    L8: do {
     if ($26) {
      $53 = ($24<<24>>24)==(37);
      L10: do {
       if ($53) {
        $54 = ((($$0273429)) + 1|0);
        $55 = HEAP8[$54>>0]|0;
        L12: do {
         switch ($55<<24>>24) {
         case 37:  {
          break L10;
          break;
         }
         case 42:  {
          $76 = ((($$0273429)) + 2|0);
          $$0293 = 0;$$2275 = $76;
          break;
         }
         default: {
          $77 = $55&255;
          $isdigittmp = (($77) + -48)|0;
          $isdigit = ($isdigittmp>>>0)<(10);
          if ($isdigit) {
           $78 = ((($$0273429)) + 2|0);
           $79 = HEAP8[$78>>0]|0;
           $80 = ($79<<24>>24)==(36);
           if ($80) {
            $81 = (_arg_n($2,$isdigittmp)|0);
            $82 = ((($$0273429)) + 3|0);
            $$0293 = $81;$$2275 = $82;
            break L12;
           }
          }
          $arglist_current = HEAP32[$2>>2]|0;
          $83 = $arglist_current;
          $84 = ((0) + 4|0);
          $expanded1 = $84;
          $expanded = (($expanded1) - 1)|0;
          $85 = (($83) + ($expanded))|0;
          $86 = ((0) + 4|0);
          $expanded5 = $86;
          $expanded4 = (($expanded5) - 1)|0;
          $expanded3 = $expanded4 ^ -1;
          $87 = $85 & $expanded3;
          $88 = $87;
          $89 = HEAP32[$88>>2]|0;
          $arglist_next = ((($88)) + 4|0);
          HEAP32[$2>>2] = $arglist_next;
          $$0293 = $89;$$2275 = $54;
         }
         }
        } while(0);
        $90 = HEAP8[$$2275>>0]|0;
        $91 = $90&255;
        $isdigittmp315414 = (($91) + -48)|0;
        $isdigit316415 = ($isdigittmp315414>>>0)<(10);
        if ($isdigit316415) {
         $$0266417 = 0;$$3416 = $$2275;$95 = $91;
         while(1) {
          $92 = ($$0266417*10)|0;
          $93 = (($92) + -48)|0;
          $94 = (($93) + ($95))|0;
          $96 = ((($$3416)) + 1|0);
          $97 = HEAP8[$96>>0]|0;
          $98 = $97&255;
          $isdigittmp315 = (($98) + -48)|0;
          $isdigit316 = ($isdigittmp315>>>0)<(10);
          if ($isdigit316) {
           $$0266417 = $94;$$3416 = $96;$95 = $98;
          } else {
           $$0266$lcssa = $94;$$3$lcssa = $96;$$lcssa355 = $97;
           break;
          }
         }
        } else {
         $$0266$lcssa = 0;$$3$lcssa = $$2275;$$lcssa355 = $90;
        }
        $99 = ($$lcssa355<<24>>24)==(109);
        $100 = ($$0293|0)!=(0|0);
        $101 = ((($$3$lcssa)) + 1|0);
        $$$0305 = $99 ? 0 : $$0305423;
        $$327 = $99 ? 0 : $102;
        $$$3 = $99 ? $101 : $$3$lcssa;
        $narrow = $100 & $99;
        $103 = ((($$$3)) + 1|0);
        $104 = HEAP8[$$$3>>0]|0;
        switch ($104<<24>>24) {
        case 104:  {
         $105 = HEAP8[$103>>0]|0;
         $106 = ($105<<24>>24)==(104);
         $107 = ((($$$3)) + 2|0);
         $$319 = $106 ? $107 : $103;
         $$320 = $106 ? -2 : -1;
         $$0268 = $$320;$$5 = $$319;
         break;
        }
        case 108:  {
         $108 = HEAP8[$103>>0]|0;
         $109 = ($108<<24>>24)==(108);
         $110 = ((($$$3)) + 2|0);
         $$321 = $109 ? $110 : $103;
         $$322 = $109 ? 3 : 1;
         $$0268 = $$322;$$5 = $$321;
         break;
        }
        case 106:  {
         $$0268 = 3;$$5 = $103;
         break;
        }
        case 116: case 122:  {
         $$0268 = 1;$$5 = $103;
         break;
        }
        case 76:  {
         $$0268 = 2;$$5 = $103;
         break;
        }
        case 110: case 112: case 67: case 83: case 91: case 99: case 115: case 88: case 71: case 70: case 69: case 65: case 103: case 102: case 101: case 97: case 120: case 117: case 111: case 105: case 100:  {
         $$0268 = 0;$$5 = $$$3;
         break;
        }
        default: {
         $$7312 = $$$0305;$309 = $$327;$narrow469 = $narrow;
         label = 137;
         break L6;
        }
        }
        $111 = HEAP8[$$5>>0]|0;
        $112 = $111&255;
        $113 = $112 & 47;
        $114 = ($113|0)==(3);
        $115 = $112 | 32;
        $$ = $114 ? $115 : $112;
        $$$0268 = $114 ? 1 : $$0268;
        $trunc = $$&255;
        switch ($trunc<<24>>24) {
        case 99:  {
         $116 = ($$0266$lcssa|0)>(1);
         $$$0266 = $116 ? $$0266$lcssa : 1;
         $$1267 = $$$0266;$$1284 = $$0283428;
         break;
        }
        case 91:  {
         $$1267 = $$0266$lcssa;$$1284 = $$0283428;
         break;
        }
        case 110:  {
         $117 = ($$0283428|0)<(0);
         $118 = $117 << 31 >> 31;
         _store_int($$0293,$$$0268,$$0283428,$118);
         $$11 = $$5;$$1289 = $$0288425;$$2285 = $$0283428;$$6311 = $$$0305;$307 = $$327;
         break L8;
         break;
        }
        default: {
         ___shlim($0,0);
         while(1) {
          $119 = HEAP32[$13>>2]|0;
          $120 = HEAP32[$14>>2]|0;
          $121 = ($119>>>0)<($120>>>0);
          if ($121) {
           $122 = ((($119)) + 1|0);
           HEAP32[$13>>2] = $122;
           $123 = HEAP8[$119>>0]|0;
           $124 = $123&255;
           $126 = $124;
          } else {
           $125 = (___shgetc($0)|0);
           $126 = $125;
          }
          $127 = (_isspace($126)|0);
          $128 = ($127|0)==(0);
          if ($128) {
           break;
          }
         }
         $129 = HEAP32[$14>>2]|0;
         $130 = ($129|0)==(0|0);
         if ($130) {
          $$pre507 = HEAP32[$13>>2]|0;
          $138 = $$pre507;
         } else {
          $131 = HEAP32[$13>>2]|0;
          $132 = ((($131)) + -1|0);
          HEAP32[$13>>2] = $132;
          $133 = $132;
          $138 = $133;
         }
         $134 = HEAP32[$15>>2]|0;
         $135 = HEAP32[$16>>2]|0;
         $136 = (($134) + ($$0283428))|0;
         $137 = (($136) + ($138))|0;
         $139 = (($137) - ($135))|0;
         $$1267 = $$0266$lcssa;$$1284 = $139;
        }
        }
        ___shlim($0,$$1267);
        $140 = HEAP32[$13>>2]|0;
        $141 = HEAP32[$14>>2]|0;
        $142 = ($140>>>0)<($141>>>0);
        if ($142) {
         $143 = ((($140)) + 1|0);
         HEAP32[$13>>2] = $143;
         $147 = $141;
        } else {
         $144 = (___shgetc($0)|0);
         $145 = ($144|0)<(0);
         if ($145) {
          $$7312 = $$$0305;$309 = $$327;$narrow469 = $narrow;
          label = 137;
          break L6;
         }
         $$pre509 = HEAP32[$14>>2]|0;
         $147 = $$pre509;
        }
        $146 = ($147|0)==(0|0);
        if (!($146)) {
         $148 = HEAP32[$13>>2]|0;
         $149 = ((($148)) + -1|0);
         HEAP32[$13>>2] = $149;
        }
        L55: do {
         switch ($trunc<<24>>24) {
         case 91: case 99: case 115:  {
          $150 = ($$|0)==(99);
          $151 = $$ | 16;
          $152 = ($151|0)==(115);
          L57: do {
           if ($152) {
            $153 = ($$|0)==(115);
            _memset(($21|0),-1,256)|0;
            HEAP8[$4>>0] = 0;
            if ($153) {
             HEAP8[$18>>0] = 0;
             ;HEAP8[$17>>0]=0|0;HEAP8[$17+1>>0]=0|0;HEAP8[$17+2>>0]=0|0;HEAP8[$17+3>>0]=0|0;HEAP8[$17+4>>0]=0|0;
             $$9 = $$5;
            } else {
             $$9 = $$5;
            }
           } else {
            $154 = ((($$5)) + 1|0);
            $155 = HEAP8[$154>>0]|0;
            $156 = ($155<<24>>24)==(94);
            $157 = ((($$5)) + 2|0);
            $$0292 = $156&1;
            $$6 = $156 ? $157 : $154;
            $158 = $156&1;
            _memset(($22|0),($158|0),256)|0;
            HEAP8[$4>>0] = 0;
            $159 = HEAP8[$$6>>0]|0;
            switch ($159<<24>>24) {
            case 45:  {
             $$sink443 = $19;
             label = 64;
             break;
            }
            case 93:  {
             $$sink443 = $20;
             label = 64;
             break;
            }
            default: {
             $$pre514 = $$0292 ^ 1;
             $$pre515 = $$pre514&255;
             $$7$ph = $$6;$$pre$phi516Z2D = $$pre515;
            }
            }
            if ((label|0) == 64) {
             label = 0;
             $160 = ((($$6)) + 1|0);
             $161 = $$0292 ^ 1;
             $162 = $161&255;
             HEAP8[$$sink443>>0] = $162;
             $$7$ph = $160;$$pre$phi516Z2D = $162;
            }
            $$7 = $$7$ph;
            while(1) {
             $163 = HEAP8[$$7>>0]|0;
             L69: do {
              switch ($163<<24>>24) {
              case 0:  {
               $$7312 = $$$0305;$309 = $$327;$narrow469 = $narrow;
               label = 137;
               break L6;
               break;
              }
              case 93:  {
               $$9 = $$7;
               break L57;
               break;
              }
              case 45:  {
               $164 = ((($$7)) + 1|0);
               $165 = HEAP8[$164>>0]|0;
               switch ($165<<24>>24) {
               case 93: case 0:  {
                $$8 = $$7;$176 = 45;
                break L69;
                break;
               }
               default: {
               }
               }
               $166 = ((($$7)) + -1|0);
               $167 = HEAP8[$166>>0]|0;
               $168 = ($167&255)<($165&255);
               if ($168) {
                $169 = $167&255;
                $$0286420 = $169;
                while(1) {
                 $170 = (($$0286420) + 1)|0;
                 $171 = (($4) + ($170)|0);
                 HEAP8[$171>>0] = $$pre$phi516Z2D;
                 $172 = HEAP8[$164>>0]|0;
                 $173 = $172&255;
                 $174 = ($170|0)<($173|0);
                 if ($174) {
                  $$0286420 = $170;
                 } else {
                  $$8 = $164;$176 = $172;
                  break;
                 }
                }
               } else {
                $$8 = $164;$176 = $165;
               }
               break;
              }
              default: {
               $$8 = $$7;$176 = $163;
              }
              }
             } while(0);
             $175 = $176&255;
             $177 = (($175) + 1)|0;
             $178 = (($4) + ($177)|0);
             HEAP8[$178>>0] = $$pre$phi516Z2D;
             $179 = ((($$8)) + 1|0);
             $$7 = $179;
            }
           }
          } while(0);
          $180 = (($$1267) + 1)|0;
          $181 = $150 ? $180 : 31;
          $182 = ($$$0268|0)==(1);
          L77: do {
           if ($182) {
            if ($narrow) {
             $183 = $181 << 2;
             $184 = (_malloc($183)|0);
             $185 = ($184|0)==(0|0);
             if ($185) {
              $$7312 = 0;$309 = 0;$narrow469 = 1;
              label = 137;
              break L6;
             } else {
              $311 = $184;
             }
            } else {
             $311 = $$0293;
            }
            HEAP32[$3>>2] = 0;
            HEAP32[$$sroa$2$0$$sroa_idx13>>2] = 0;
            $$0276$ph = $181;$$0278$ph = 0;$$ph = $311;
            L82: while(1) {
             $186 = ($$ph|0)==(0|0);
             $$0278$ph336 = $$0278$ph;
             while(1) {
              L86: while(1) {
               $187 = HEAP32[$13>>2]|0;
               $188 = HEAP32[$14>>2]|0;
               $189 = ($187>>>0)<($188>>>0);
               if ($189) {
                $190 = ((($187)) + 1|0);
                HEAP32[$13>>2] = $190;
                $191 = HEAP8[$187>>0]|0;
                $192 = $191&255;
                $195 = $192;
               } else {
                $193 = (___shgetc($0)|0);
                $195 = $193;
               }
               $194 = (($195) + 1)|0;
               $196 = (($4) + ($194)|0);
               $197 = HEAP8[$196>>0]|0;
               $198 = ($197<<24>>24)==(0);
               if ($198) {
                break L82;
               }
               $199 = $195&255;
               HEAP8[$6>>0] = $199;
               $200 = (_mbrtowc($5,$6,1,$3)|0);
               switch ($200|0) {
               case -1:  {
                $$7312 = 0;$309 = $$ph;$narrow469 = $narrow;
                label = 137;
                break L6;
                break;
               }
               case -2:  {
                break;
               }
               default: {
                break L86;
               }
               }
              }
              if ($186) {
               $$1279 = $$0278$ph336;
              } else {
               $201 = (($$ph) + ($$0278$ph336<<2)|0);
               $202 = (($$0278$ph336) + 1)|0;
               $203 = HEAP32[$5>>2]|0;
               HEAP32[$201>>2] = $203;
               $$1279 = $202;
              }
              $204 = ($$1279|0)==($$0276$ph|0);
              $or$cond = $narrow & $204;
              if ($or$cond) {
               break;
              } else {
               $$0278$ph336 = $$1279;
              }
             }
             $factor331 = $$0276$ph << 1;
             $205 = $factor331 | 1;
             $206 = $205 << 2;
             $207 = (_realloc($$ph,$206)|0);
             $208 = ($207|0)==(0|0);
             if ($208) {
              $$7312 = 0;$309 = $$ph;$narrow469 = 1;
              label = 137;
              break L6;
             } else {
              $$0278$ph$phi = $$0276$ph;$$0276$ph = $205;$$ph = $207;$$0278$ph = $$0278$ph$phi;
             }
            }
            $209 = (_mbsinit($3)|0);
            $210 = ($209|0)==(0);
            if ($210) {
             $$7312 = 0;$309 = $$ph;$narrow469 = $narrow;
             label = 137;
             break L6;
            } else {
             $$4282 = $$0278$ph336;$$4309 = 0;$$5299 = $$ph;$312 = $$ph;
            }
           } else {
            if ($narrow) {
             $211 = (_malloc($181)|0);
             $212 = ($211|0)==(0|0);
             if ($212) {
              $$7312 = 0;$309 = 0;$narrow469 = 1;
              label = 137;
              break L6;
             } else {
              $$1277$ph = $181;$$2280$ph = 0;$$2307$ph = $211;
             }
             while(1) {
              $$2280 = $$2280$ph;
              while(1) {
               $213 = HEAP32[$13>>2]|0;
               $214 = HEAP32[$14>>2]|0;
               $215 = ($213>>>0)<($214>>>0);
               if ($215) {
                $216 = ((($213)) + 1|0);
                HEAP32[$13>>2] = $216;
                $217 = HEAP8[$213>>0]|0;
                $218 = $217&255;
                $221 = $218;
               } else {
                $219 = (___shgetc($0)|0);
                $221 = $219;
               }
               $220 = (($221) + 1)|0;
               $222 = (($4) + ($220)|0);
               $223 = HEAP8[$222>>0]|0;
               $224 = ($223<<24>>24)==(0);
               if ($224) {
                $$4282 = $$2280;$$4309 = $$2307$ph;$$5299 = 0;$312 = 0;
                break L77;
               }
               $225 = $221&255;
               $226 = (($$2280) + 1)|0;
               $227 = (($$2307$ph) + ($$2280)|0);
               HEAP8[$227>>0] = $225;
               $228 = ($226|0)==($$1277$ph|0);
               if ($228) {
                break;
               } else {
                $$2280 = $226;
               }
              }
              $factor = $$1277$ph << 1;
              $229 = $factor | 1;
              $230 = (_realloc($$2307$ph,$229)|0);
              $231 = ($230|0)==(0|0);
              if ($231) {
               $$7312 = $$2307$ph;$309 = 0;$narrow469 = 1;
               label = 137;
               break L6;
              } else {
               $$2280$ph$phi = $$1277$ph;$$1277$ph = $229;$$2307$ph = $230;$$2280$ph = $$2280$ph$phi;
              }
             }
            }
            $232 = ($$0293|0)==(0|0);
            if ($232) {
             $250 = $147;
             while(1) {
              $248 = HEAP32[$13>>2]|0;
              $249 = ($248>>>0)<($250>>>0);
              if ($249) {
               $251 = ((($248)) + 1|0);
               HEAP32[$13>>2] = $251;
               $252 = HEAP8[$248>>0]|0;
               $253 = $252&255;
               $256 = $253;
              } else {
               $254 = (___shgetc($0)|0);
               $256 = $254;
              }
              $255 = (($256) + 1)|0;
              $257 = (($4) + ($255)|0);
              $258 = HEAP8[$257>>0]|0;
              $259 = ($258<<24>>24)==(0);
              if ($259) {
               $$4282 = 0;$$4309 = 0;$$5299 = 0;$312 = 0;
               break L77;
              }
              $$pre512 = HEAP32[$14>>2]|0;
              $250 = $$pre512;
             }
            } else {
             $$3281 = 0;$235 = $147;
             while(1) {
              $233 = HEAP32[$13>>2]|0;
              $234 = ($233>>>0)<($235>>>0);
              if ($234) {
               $236 = ((($233)) + 1|0);
               HEAP32[$13>>2] = $236;
               $237 = HEAP8[$233>>0]|0;
               $238 = $237&255;
               $241 = $238;
              } else {
               $239 = (___shgetc($0)|0);
               $241 = $239;
              }
              $240 = (($241) + 1)|0;
              $242 = (($4) + ($240)|0);
              $243 = HEAP8[$242>>0]|0;
              $244 = ($243<<24>>24)==(0);
              if ($244) {
               $$4282 = $$3281;$$4309 = $$0293;$$5299 = 0;$312 = 0;
               break L77;
              }
              $245 = $241&255;
              $246 = (($$3281) + 1)|0;
              $247 = (($$0293) + ($$3281)|0);
              HEAP8[$247>>0] = $245;
              $$pre511 = HEAP32[$14>>2]|0;
              $$3281 = $246;$235 = $$pre511;
             }
            }
           }
          } while(0);
          $260 = HEAP32[$14>>2]|0;
          $261 = ($260|0)==(0|0);
          if ($261) {
           $$pre513 = HEAP32[$13>>2]|0;
           $268 = $$pre513;
          } else {
           $262 = HEAP32[$13>>2]|0;
           $263 = ((($262)) + -1|0);
           HEAP32[$13>>2] = $263;
           $264 = $263;
           $268 = $264;
          }
          $265 = HEAP32[$15>>2]|0;
          $266 = HEAP32[$16>>2]|0;
          $267 = (($268) - ($266))|0;
          $269 = (($267) + ($265))|0;
          $270 = ($269|0)==(0);
          if ($270) {
           $$9314$ph = $$4309;$$ph353 = $312;
           label = 139;
           break L6;
          }
          $$not = $150 ^ 1;
          $271 = ($269|0)==($$1267|0);
          $or$cond318 = $271 | $$not;
          if (!($or$cond318)) {
           $$9314$ph = $$4309;$$ph353 = $312;
           label = 139;
           break L6;
          }
          do {
           if ($narrow) {
            if ($182) {
             HEAP32[$$0293>>2] = $$5299;
             break;
            } else {
             HEAP32[$$0293>>2] = $$4309;
             break;
            }
           }
          } while(0);
          if ($150) {
           $$10 = $$9;$$5310 = $$4309;$310 = $312;
          } else {
           $272 = ($$5299|0)==(0|0);
           if (!($272)) {
            $273 = (($$5299) + ($$4282<<2)|0);
            HEAP32[$273>>2] = 0;
           }
           $274 = ($$4309|0)==(0|0);
           if ($274) {
            $$10 = $$9;$$5310 = 0;$310 = $312;
            break L55;
           }
           $275 = (($$4309) + ($$4282)|0);
           HEAP8[$275>>0] = 0;
           $$10 = $$9;$$5310 = $$4309;$310 = $312;
          }
          break;
         }
         case 120: case 88: case 112:  {
          $$0272 = 16;
          label = 125;
          break;
         }
         case 111:  {
          $$0272 = 8;
          label = 125;
          break;
         }
         case 117: case 100:  {
          $$0272 = 10;
          label = 125;
          break;
         }
         case 105:  {
          $$0272 = 0;
          label = 125;
          break;
         }
         case 71: case 103: case 70: case 102: case 69: case 101: case 65: case 97:  {
          $285 = (+___floatscan($0,$$$0268,0));
          $286 = HEAP32[$15>>2]|0;
          $287 = HEAP32[$13>>2]|0;
          $288 = HEAP32[$16>>2]|0;
          $289 = (($288) - ($287))|0;
          $290 = ($286|0)==($289|0);
          if ($290) {
           $$9314$ph = $$$0305;$$ph353 = $$327;
           label = 139;
           break L6;
          }
          $291 = ($$0293|0)==(0|0);
          if ($291) {
           $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
          } else {
           switch ($$$0268|0) {
           case 0:  {
            $292 = $285;
            HEAPF32[$$0293>>2] = $292;
            $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
            break L55;
            break;
           }
           case 1:  {
            HEAPF64[$$0293>>3] = $285;
            $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
            break L55;
            break;
           }
           case 2:  {
            HEAPF64[$$0293>>3] = $285;
            $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
            break L55;
            break;
           }
           default: {
            $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
            break L55;
           }
           }
          }
          break;
         }
         default: {
          $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
         }
         }
        } while(0);
        do {
         if ((label|0) == 125) {
          label = 0;
          $276 = (___intscan($0,$$0272,0,-1,-1)|0);
          $277 = tempRet0;
          $278 = HEAP32[$15>>2]|0;
          $279 = HEAP32[$13>>2]|0;
          $280 = HEAP32[$16>>2]|0;
          $281 = (($280) - ($279))|0;
          $282 = ($278|0)==($281|0);
          if ($282) {
           $$9314$ph = $$$0305;$$ph353 = $$327;
           label = 139;
           break L6;
          }
          $283 = ($$|0)==(112);
          $or$cond3 = $100 & $283;
          if ($or$cond3) {
           $284 = $276;
           HEAP32[$$0293>>2] = $284;
           $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
           break;
          } else {
           _store_int($$0293,$$$0268,$276,$277);
           $$10 = $$5;$$5310 = $$$0305;$310 = $$327;
           break;
          }
         }
        } while(0);
        $293 = HEAP32[$15>>2]|0;
        $294 = HEAP32[$13>>2]|0;
        $295 = HEAP32[$16>>2]|0;
        $296 = (($293) + ($$1284))|0;
        $297 = (($296) + ($294))|0;
        $298 = (($297) - ($295))|0;
        $299 = $100&1;
        $$0288$ = (($299) + ($$0288425))|0;
        $$11 = $$10;$$1289 = $$0288$;$$2285 = $298;$$6311 = $$5310;$307 = $310;
        break L8;
       }
      } while(0);
      $56 = $53&1;
      $57 = (($$0273429) + ($56)|0);
      ___shlim($0,0);
      $58 = HEAP32[$13>>2]|0;
      $59 = HEAP32[$14>>2]|0;
      $60 = ($58>>>0)<($59>>>0);
      if ($60) {
       $61 = ((($58)) + 1|0);
       HEAP32[$13>>2] = $61;
       $62 = HEAP8[$58>>0]|0;
       $63 = $62&255;
       $68 = $63;
      } else {
       $64 = (___shgetc($0)|0);
       $68 = $64;
      }
      $65 = HEAP8[$57>>0]|0;
      $66 = $65&255;
      $67 = ($68|0)==($66|0);
      if (!($67)) {
       label = 22;
       break L6;
      }
      $75 = (($$0283428) + 1)|0;
      $$11 = $57;$$1289 = $$0288425;$$2285 = $75;$$6311 = $$0305423;$307 = $102;
     } else {
      $$1274 = $$0273429;
      while(1) {
       $27 = ((($$1274)) + 1|0);
       $28 = HEAP8[$27>>0]|0;
       $29 = $28&255;
       $30 = (_isspace($29)|0);
       $31 = ($30|0)==(0);
       if ($31) {
        break;
       } else {
        $$1274 = $27;
       }
      }
      ___shlim($0,0);
      while(1) {
       $32 = HEAP32[$13>>2]|0;
       $33 = HEAP32[$14>>2]|0;
       $34 = ($32>>>0)<($33>>>0);
       if ($34) {
        $35 = ((($32)) + 1|0);
        HEAP32[$13>>2] = $35;
        $36 = HEAP8[$32>>0]|0;
        $37 = $36&255;
        $39 = $37;
       } else {
        $38 = (___shgetc($0)|0);
        $39 = $38;
       }
       $40 = (_isspace($39)|0);
       $41 = ($40|0)==(0);
       if ($41) {
        break;
       }
      }
      $42 = HEAP32[$14>>2]|0;
      $43 = ($42|0)==(0|0);
      if ($43) {
       $$pre = HEAP32[$13>>2]|0;
       $51 = $$pre;
      } else {
       $44 = HEAP32[$13>>2]|0;
       $45 = ((($44)) + -1|0);
       HEAP32[$13>>2] = $45;
       $46 = $45;
       $51 = $46;
      }
      $47 = HEAP32[$15>>2]|0;
      $48 = HEAP32[$16>>2]|0;
      $49 = (($47) + ($$0283428))|0;
      $50 = (($49) + ($51))|0;
      $52 = (($50) - ($48))|0;
      $$11 = $$1274;$$1289 = $$0288425;$$2285 = $52;$$6311 = $$0305423;$307 = $102;
     }
    } while(0);
    $300 = ((($$11)) + 1|0);
    $301 = HEAP8[$300>>0]|0;
    $302 = ($301<<24>>24)==(0);
    if ($302) {
     $$3291 = $$1289;
     break L4;
    } else {
     $$0273429 = $300;$$0283428 = $$2285;$$0288425 = $$1289;$$0305423 = $$6311;$102 = $307;$24 = $301;
    }
   }
   if ((label|0) == 22) {
    $69 = HEAP32[$14>>2]|0;
    $70 = ($69|0)==(0|0);
    if (!($70)) {
     $71 = HEAP32[$13>>2]|0;
     $72 = ((($71)) + -1|0);
     HEAP32[$13>>2] = $72;
    }
    $73 = ($68|0)>(-1);
    $74 = ($$0288425|0)!=(0);
    $or$cond5 = $74 | $73;
    if ($or$cond5) {
     $$3291 = $$0288425;
     break;
    } else {
     $$1271 = 0;$$8313 = $$0305423;$308 = $102;
     label = 138;
    }
   }
   else if ((label|0) == 137) {
    $$328$le441 = $narrow469&1;
    $$old4 = ($$0288425|0)==(0);
    if ($$old4) {
     $$1271 = $$328$le441;$$8313 = $$7312;$308 = $309;
     label = 138;
    } else {
     $$2 = $$328$le441;$$2290 = $$0288425;$$9314 = $$7312;$304 = $309;
    }
   }
   else if ((label|0) == 139) {
    $$328$le439 = $narrow&1;
    $$2 = $$328$le439;$$2290 = $$0288425;$$9314 = $$9314$ph;$304 = $$ph353;
   }
   if ((label|0) == 138) {
    $$2 = $$1271;$$2290 = -1;$$9314 = $$8313;$304 = $308;
   }
   $303 = ($$2|0)==(0);
   if ($303) {
    $$3291 = $$2290;
   } else {
    _free($$9314);
    _free($304);
    $$3291 = $$2290;
   }
  }
 } while(0);
 $305 = ($306|0)==(0);
 if (!($305)) {
  ___unlockfile($0);
 }
 STACKTOP = sp;return ($$3291|0);
}
function _arg_n($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $arglist_current = 0, $arglist_next = 0, $expanded = 0, $expanded1 = 0, $expanded3 = 0, $expanded4 = 0, $expanded5 = 0, $vacopy_currentptr = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $vacopy_currentptr = HEAP32[$0>>2]|0;
 HEAP32[$2>>2] = $vacopy_currentptr;
 $$0 = $1;
 while(1) {
  $3 = ($$0>>>0)>(1);
  $arglist_current = HEAP32[$2>>2]|0;
  $4 = $arglist_current;
  $5 = ((0) + 4|0);
  $expanded1 = $5;
  $expanded = (($expanded1) - 1)|0;
  $6 = (($4) + ($expanded))|0;
  $7 = ((0) + 4|0);
  $expanded5 = $7;
  $expanded4 = (($expanded5) - 1)|0;
  $expanded3 = $expanded4 ^ -1;
  $8 = $6 & $expanded3;
  $9 = $8;
  $10 = HEAP32[$9>>2]|0;
  $arglist_next = ((($9)) + 4|0);
  HEAP32[$2>>2] = $arglist_next;
  $11 = (($$0) + -1)|0;
  if ($3) {
   $$0 = $11;
  } else {
   break;
  }
 }
 STACKTOP = sp;return ($10|0);
}
function _store_int($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ($0|0)==(0|0);
 L1: do {
  if (!($4)) {
   switch ($1|0) {
   case -2:  {
    $5 = $2&255;
    HEAP8[$0>>0] = $5;
    break L1;
    break;
   }
   case -1:  {
    $6 = $2&65535;
    HEAP16[$0>>1] = $6;
    break L1;
    break;
   }
   case 0:  {
    HEAP32[$0>>2] = $2;
    break L1;
    break;
   }
   case 1:  {
    HEAP32[$0>>2] = $2;
    break L1;
    break;
   }
   case 3:  {
    $7 = $0;
    $8 = $7;
    HEAP32[$8>>2] = $2;
    $9 = (($7) + 4)|0;
    $10 = $9;
    HEAP32[$10>>2] = $3;
    break L1;
    break;
   }
   default: {
    break L1;
   }
   }
  }
 } while(0);
 return;
}
function _mbsinit($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $5 = 1;
 } else {
  $2 = HEAP32[$0>>2]|0;
  $3 = ($2|0)==(0);
  $5 = $3;
 }
 $4 = $5&1;
 return ($4|0);
}
function ___string_read($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$027 = 0, $$027$ = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 84|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($2) + 256)|0;
 $6 = (_memchr($4,0,$5)|0);
 $7 = ($6|0)==(0|0);
 $8 = $6;
 $9 = $4;
 $10 = (($8) - ($9))|0;
 $$027 = $7 ? $5 : $10;
 $11 = ($$027>>>0)<($2>>>0);
 $$027$ = $11 ? $$027 : $2;
 _memcpy(($1|0),($4|0),($$027$|0))|0;
 $12 = (($4) + ($$027$)|0);
 $13 = ((($0)) + 4|0);
 HEAP32[$13>>2] = $12;
 $14 = (($4) + ($$027)|0);
 $15 = ((($0)) + 8|0);
 HEAP32[$15>>2] = $14;
 HEAP32[$3>>2] = $14;
 return ($$027$|0);
}
function _socket($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$017 = 0, $$old = 0, $$old18 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $vararg_buffer = 0, $vararg_buffer11 = 0, $vararg_buffer2 = 0;
 var $vararg_buffer6 = 0, $vararg_ptr1 = 0, $vararg_ptr10 = 0, $vararg_ptr14 = 0, $vararg_ptr15 = 0, $vararg_ptr5 = 0, $vararg_ptr9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer11 = sp + 32|0;
 $vararg_buffer6 = sp + 16|0;
 $vararg_buffer2 = sp + 8|0;
 $vararg_buffer = sp;
 $3 = sp + 72|0;
 $4 = sp + 48|0;
 HEAP32[$3>>2] = $0;
 $5 = ((($3)) + 4|0);
 HEAP32[$5>>2] = $1;
 $6 = ((($3)) + 8|0);
 HEAP32[$6>>2] = $2;
 $7 = ((($3)) + 12|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($3)) + 16|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($3)) + 20|0);
 HEAP32[$9>>2] = 0;
 $10 = $3;
 HEAP32[$vararg_buffer>>2] = 1;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $10;
 $11 = (___syscall102(102,($vararg_buffer|0))|0);
 $12 = (___syscall_ret($11)|0);
 $13 = ($12|0)<(0);
 do {
  if ($13) {
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(22);
   if ($16) {
    $$old = $1 & 526336;
    $$old18 = ($$old|0)==(0);
    if ($$old18) {
     $$017 = $12;
     break;
    }
   } else {
    $17 = (___errno_location()|0);
    $18 = HEAP32[$17>>2]|0;
    $19 = ($18|0)!=(93);
    $20 = $1 & 526336;
    $21 = ($20|0)==(0);
    $or$cond = $21 | $19;
    if ($or$cond) {
     $$017 = $12;
     break;
    }
   }
   HEAP32[$4>>2] = $0;
   $22 = ((($4)) + 4|0);
   $23 = $1 & -526337;
   HEAP32[$22>>2] = $23;
   $24 = ((($4)) + 8|0);
   HEAP32[$24>>2] = $2;
   $25 = ((($4)) + 12|0);
   HEAP32[$25>>2] = 0;
   $26 = ((($4)) + 16|0);
   HEAP32[$26>>2] = 0;
   $27 = ((($4)) + 20|0);
   HEAP32[$27>>2] = 0;
   $28 = $4;
   HEAP32[$vararg_buffer2>>2] = 1;
   $vararg_ptr5 = ((($vararg_buffer2)) + 4|0);
   HEAP32[$vararg_ptr5>>2] = $28;
   $29 = (___syscall102(102,($vararg_buffer2|0))|0);
   $30 = (___syscall_ret($29)|0);
   $31 = ($30|0)<(0);
   if ($31) {
    $$017 = $30;
   } else {
    $32 = $1 & 524288;
    $33 = ($32|0)==(0);
    if (!($33)) {
     HEAP32[$vararg_buffer6>>2] = $30;
     $vararg_ptr9 = ((($vararg_buffer6)) + 4|0);
     HEAP32[$vararg_ptr9>>2] = 2;
     $vararg_ptr10 = ((($vararg_buffer6)) + 8|0);
     HEAP32[$vararg_ptr10>>2] = 1;
     (___syscall221(221,($vararg_buffer6|0))|0);
    }
    $34 = $1 & 2048;
    $35 = ($34|0)==(0);
    if ($35) {
     $$017 = $30;
    } else {
     HEAP32[$vararg_buffer11>>2] = $30;
     $vararg_ptr14 = ((($vararg_buffer11)) + 4|0);
     HEAP32[$vararg_ptr14>>2] = 4;
     $vararg_ptr15 = ((($vararg_buffer11)) + 8|0);
     HEAP32[$vararg_ptr15>>2] = 2048;
     (___syscall221(221,($vararg_buffer11|0))|0);
     $$017 = $30;
    }
   }
  } else {
   $$017 = $12;
  }
 } while(0);
 STACKTOP = sp;return ($$017|0);
}
function _connect($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $3 = sp + 8|0;
 HEAP32[$3>>2] = $0;
 $4 = ((($3)) + 4|0);
 $5 = $1;
 HEAP32[$4>>2] = $5;
 $6 = ((($3)) + 8|0);
 HEAP32[$6>>2] = $2;
 $7 = ((($3)) + 12|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($3)) + 16|0);
 HEAP32[$8>>2] = 0;
 $9 = ((($3)) + 20|0);
 HEAP32[$9>>2] = 0;
 $10 = $3;
 HEAP32[$vararg_buffer>>2] = 3;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $10;
 $11 = (___syscall102(102,($vararg_buffer|0))|0);
 $12 = (___syscall_ret($11)|0);
 STACKTOP = sp;return ($12|0);
}
function _send($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = (_sendto($0,$1,$2,$3,0,0)|0);
 return ($4|0);
}
function _sendto($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $6 = sp + 8|0;
 HEAP32[$6>>2] = $0;
 $7 = ((($6)) + 4|0);
 $8 = $1;
 HEAP32[$7>>2] = $8;
 $9 = ((($6)) + 8|0);
 HEAP32[$9>>2] = $2;
 $10 = ((($6)) + 12|0);
 HEAP32[$10>>2] = $3;
 $11 = ((($6)) + 16|0);
 $12 = $4;
 HEAP32[$11>>2] = $12;
 $13 = ((($6)) + 20|0);
 HEAP32[$13>>2] = $5;
 $14 = $6;
 HEAP32[$vararg_buffer>>2] = 11;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $14;
 $15 = (___syscall102(102,($vararg_buffer|0))|0);
 $16 = (___syscall_ret($15)|0);
 STACKTOP = sp;return ($16|0);
}
function _mbrlen($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)!=(0|0);
 $4 = $3 ? $2 : 164388;
 $5 = (_mbrtowc(0,$0,$1,$4)|0);
 return ($5|0);
}
function _mbsnrtowcs($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$04975 = 0, $$05274 = 0, $$055$ = 0, $$05573 = 0, $$064 = 0, $$076 = 0, $$1 = 0, $$150 = 0, $$153 = 0, $$156 = 0, $$25170 = 0, $$25469 = 0, $$257 = 0, $$271 = 0, $$3 = 0, $$35868 = 0, $$59 = 0, $$cast = 0, $$pr$pre = 0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $or$cond = 0, $or$cond72 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(1040|0);
 $5 = sp + 8|0;
 $6 = sp;
 $7 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $7;
 $8 = ($0|0)!=(0|0);
 $$ = $8 ? $3 : 256;
 $$59 = $8 ? $0 : $5;
 $$cast = $7;
 $9 = ($7|0)!=(0);
 $10 = ($$|0)!=(0);
 $or$cond72 = $10 & $9;
 L1: do {
  if ($or$cond72) {
   $$04975 = 0;$$05274 = $2;$$05573 = $$;$$076 = $$59;$44 = $$cast;
   while(1) {
    $11 = $$05274 >>> 2;
    $12 = ($11>>>0)>=($$05573>>>0);
    $13 = ($$05274>>>0)>(131);
    $14 = $13 | $12;
    if (!($14)) {
     $$064 = $$076;$$150 = $$04975;$$153 = $$05274;$$257 = $$05573;$26 = $44;
     break L1;
    }
    $$055$ = $12 ? $$05573 : $11;
    $15 = (($$05274) - ($$055$))|0;
    $16 = (_mbsrtowcs($$076,$6,$$055$,$4)|0);
    $17 = ($16|0)==(-1);
    if ($17) {
     break;
    }
    $18 = ($$076|0)==($5|0);
    $19 = (($$076) + ($16<<2)|0);
    $20 = $18 ? 0 : $16;
    $$156 = (($$05573) - ($20))|0;
    $$1 = $18 ? $$076 : $19;
    $21 = (($16) + ($$04975))|0;
    $22 = HEAP32[$6>>2]|0;
    $23 = ($22|0)!=(0|0);
    $24 = ($$156|0)!=(0);
    $or$cond = $24 & $23;
    if ($or$cond) {
     $$04975 = $21;$$05274 = $15;$$05573 = $$156;$$076 = $$1;$44 = $22;
    } else {
     $$064 = $$1;$$150 = $21;$$153 = $15;$$257 = $$156;$26 = $22;
     break L1;
    }
   }
   $$pr$pre = HEAP32[$6>>2]|0;
   $$064 = $$076;$$150 = -1;$$153 = $15;$$257 = 0;$26 = $$pr$pre;
  } else {
   $$064 = $$59;$$150 = 0;$$153 = $2;$$257 = $$;$26 = $$cast;
  }
 } while(0);
 $25 = ($26|0)==(0|0);
 L8: do {
  if ($25) {
   $$3 = $$150;
  } else {
   $27 = ($$257|0)!=(0);
   $28 = ($$153|0)!=(0);
   $29 = $27 & $28;
   if ($29) {
    $$25170 = $$150;$$25469 = $$153;$$271 = $$064;$$35868 = $$257;$30 = $26;
    while(1) {
     $31 = (_mbrtowc($$271,$30,$$25469,$4)|0);
     $32 = (($31) + 2)|0;
     $33 = ($32>>>0)<(3);
     if ($33) {
      break;
     }
     $34 = HEAP32[$6>>2]|0;
     $35 = (($34) + ($31)|0);
     HEAP32[$6>>2] = $35;
     $36 = (($$25469) - ($31))|0;
     $37 = ((($$271)) + 4|0);
     $38 = (($$35868) + -1)|0;
     $39 = (($$25170) + 1)|0;
     $40 = ($38|0)!=(0);
     $41 = ($36|0)!=(0);
     $42 = $40 & $41;
     if ($42) {
      $$25170 = $39;$$25469 = $36;$$271 = $37;$$35868 = $38;$30 = $35;
     } else {
      $$3 = $39;
      break L8;
     }
    }
    switch ($31|0) {
    case -1:  {
     $$3 = -1;
     break L8;
     break;
    }
    case 0:  {
     HEAP32[$6>>2] = 0;
     $$3 = $$25170;
     break L8;
     break;
    }
    default: {
     HEAP32[$4>>2] = 0;
     $$3 = $$25170;
     break L8;
    }
    }
   } else {
    $$3 = $$150;
   }
  }
 } while(0);
 if ($8) {
  $43 = HEAP32[$6>>2]|0;
  HEAP32[$1>>2] = $43;
 }
 STACKTOP = sp;return ($$3|0);
}
function _mbsrtowcs($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0100$lcssa = 0, $$0100136 = 0, $$0107135 = 0, $$0113134 = 0, $$099 = 0, $$10 = 0, $$1101 = 0, $$1101$ph = 0, $$1108$ph = 0, $$1108138 = 0, $$11124 = 0, $$1114 = 0, $$1114$ph = 0, $$11163 = 0, $$2 = 0, $$2102144 = 0, $$2109 = 0, $$2115143 = 0, $$3103 = 0, $$3110 = 0;
 var $$3116 = 0, $$4 = 0, $$4104 = 0, $$4111 = 0, $$4117 = 0, $$5 = 0, $$5105 = 0, $$5112 = 0, $$5112123 = 0, $$5112162 = 0, $$5118$ph = 0, $$5118137 = 0, $$6 = 0, $$6106$lcssa = 0, $$6106$ph = 0, $$6106139 = 0, $$6119 = 0, $$7 = 0, $$7120 = 0, $$8 = 0;
 var $$8121 = 0, $$9 = 0, $$9$sink = 0, $$9122 = 0, $$9122161 = 0, $$old1 = 0, $$pre$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $not$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = HEAP32[$1>>2]|0;
 $5 = ($3|0)==(0|0);
 if ($5) {
  label = 5;
 } else {
  $6 = HEAP32[$3>>2]|0;
  $7 = ($6|0)==(0);
  if ($7) {
   label = 5;
  } else {
   $8 = ($0|0)==(0|0);
   if ($8) {
    $$2 = $6;$$4104 = $4;$$4117 = $2;
    label = 25;
   } else {
    HEAP32[$3>>2] = 0;
    $$4 = $6;$$4111 = $0;$$8121 = $2;$$9 = $4;
    label = 43;
   }
  }
 }
 L5: do {
  if ((label|0) == 5) {
   $9 = (___pthread_self_442()|0);
   $10 = ((($9)) + 188|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = HEAP32[$11>>2]|0;
   $not$ = ($12|0)==(0|0);
   $13 = ($0|0)!=(0|0);
   if (!($not$)) {
    if ($13) {
     $$1108$ph = $0;$$5118$ph = $2;$$6106$ph = $4;
     label = 15;
     break;
    } else {
     $$1101$ph = $4;$$1114$ph = $2;
     label = 14;
     break;
    }
   }
   if (!($13)) {
    $15 = (_strlen($4)|0);
    $$099 = $15;
    label = 59;
    break;
   }
   $14 = ($2|0)==(0);
   L13: do {
    if ($14) {
     $$0100$lcssa = $4;
    } else {
     $$0100136 = $4;$$0107135 = $0;$$0113134 = $2;
     while(1) {
      $16 = HEAP8[$$0100136>>0]|0;
      $17 = ($16<<24>>24)==(0);
      if ($17) {
       break;
      }
      $18 = ((($$0100136)) + 1|0);
      $19 = $16 << 24 >> 24;
      $20 = $19 & 57343;
      $21 = ((($$0107135)) + 4|0);
      HEAP32[$$0107135>>2] = $20;
      $22 = (($$0113134) + -1)|0;
      $23 = ($22|0)==(0);
      if ($23) {
       $$0100$lcssa = $18;
       break L13;
      } else {
       $$0100136 = $18;$$0107135 = $21;$$0113134 = $22;
      }
     }
     HEAP32[$$0107135>>2] = 0;
     HEAP32[$1>>2] = 0;
     $24 = (($2) - ($$0113134))|0;
     $$099 = $24;
     label = 59;
     break L5;
    }
   } while(0);
   HEAP32[$1>>2] = $$0100$lcssa;
   $$099 = $2;
   label = 59;
  }
 } while(0);
 L19: while(1) {
  L20: do {
   if ((label|0) == 14) {
    label = 0;
    $$1101 = $$1101$ph;$$1114 = $$1114$ph;
    while(1) {
     $26 = HEAP8[$$1101>>0]|0;
     $27 = $26&255;
     $28 = (($27) + -1)|0;
     $29 = ($28>>>0)<(127);
     if ($29) {
      $30 = $$1101;
      $31 = $30 & 3;
      $32 = ($31|0)==(0);
      if ($32) {
       $33 = HEAP32[$$1101>>2]|0;
       $34 = (($33) + -16843009)|0;
       $35 = $34 | $33;
       $36 = $35 & -2139062144;
       $37 = ($36|0)==(0);
       $38 = $33&255;
       if ($37) {
        $$2102144 = $$1101;$$2115143 = $$1114;
        while(1) {
         $39 = ((($$2102144)) + 4|0);
         $40 = (($$2115143) + -4)|0;
         $41 = HEAP32[$39>>2]|0;
         $42 = (($41) + -16843009)|0;
         $43 = $42 | $41;
         $44 = $43 & -2139062144;
         $45 = ($44|0)==(0);
         if ($45) {
          $$2102144 = $39;$$2115143 = $40;
         } else {
          break;
         }
        }
        $46 = $41&255;
        $$3103 = $39;$$3116 = $40;$48 = $46;
       } else {
        $$3103 = $$1101;$$3116 = $$1114;$48 = $38;
       }
      } else {
       $$3103 = $$1101;$$3116 = $$1114;$48 = $26;
      }
     } else {
      $$3103 = $$1101;$$3116 = $$1114;$48 = $26;
     }
     $47 = $48&255;
     $49 = (($47) + -1)|0;
     $50 = ($49>>>0)<(127);
     if (!($50)) {
      break;
     }
     $51 = ((($$3103)) + 1|0);
     $52 = (($$3116) + -1)|0;
     $$1101 = $51;$$1114 = $52;
    }
    $53 = (($47) + -194)|0;
    $54 = ($53>>>0)>(50);
    if ($54) {
     $$11163 = $$3103;$$5112162 = $0;$$9122161 = $$3116;
     label = 53;
    } else {
     $55 = ((($$3103)) + 1|0);
     $56 = (14424 + ($53<<2)|0);
     $57 = HEAP32[$56>>2]|0;
     $$2 = $57;$$4104 = $55;$$4117 = $$3116;
     label = 25;
     continue L19;
    }
   }
   else if ((label|0) == 15) {
    label = 0;
    $25 = ($$5118$ph|0)==(0);
    L22: do {
     if ($25) {
      $$6106$lcssa = $$6106$ph;
     } else {
      $$1108138 = $$1108$ph;$$5118137 = $$5118$ph;$$6106139 = $$6106$ph;
      while(1) {
       $80 = HEAP8[$$6106139>>0]|0;
       $81 = $80&255;
       $82 = (($81) + -1)|0;
       $83 = ($82>>>0)<(127);
       do {
        if ($83) {
         $84 = $$6106139;
         $85 = $84 & 3;
         $86 = ($85|0)==(0);
         $87 = ($$5118137>>>0)>(4);
         $or$cond = $87 & $86;
         if ($or$cond) {
          $$2109 = $$1108138;$$6119 = $$5118137;$$7 = $$6106139;
          while(1) {
           $88 = HEAP32[$$7>>2]|0;
           $89 = (($88) + -16843009)|0;
           $90 = $89 | $88;
           $91 = $90 & -2139062144;
           $92 = ($91|0)==(0);
           if (!($92)) {
            label = 38;
            break;
           }
           $93 = ((($$7)) + 1|0);
           $94 = $88 & 255;
           $95 = ((($$2109)) + 4|0);
           HEAP32[$$2109>>2] = $94;
           $96 = ((($$7)) + 2|0);
           $97 = HEAP8[$93>>0]|0;
           $98 = $97&255;
           $99 = ((($$2109)) + 8|0);
           HEAP32[$95>>2] = $98;
           $100 = ((($$7)) + 3|0);
           $101 = HEAP8[$96>>0]|0;
           $102 = $101&255;
           $103 = ((($$2109)) + 12|0);
           HEAP32[$99>>2] = $102;
           $104 = ((($$7)) + 4|0);
           $105 = HEAP8[$100>>0]|0;
           $106 = $105&255;
           $107 = ((($$2109)) + 16|0);
           HEAP32[$103>>2] = $106;
           $108 = (($$6119) + -4)|0;
           $$old1 = ($108>>>0)>(4);
           if ($$old1) {
            $$2109 = $107;$$6119 = $108;$$7 = $104;
           } else {
            label = 37;
            break;
           }
          }
          if ((label|0) == 37) {
           label = 0;
           $$pre$pre = HEAP8[$104>>0]|0;
           $$3110 = $107;$$7120 = $108;$$8 = $104;$111 = $$pre$pre;
           break;
          }
          else if ((label|0) == 38) {
           label = 0;
           $109 = $88&255;
           $$3110 = $$2109;$$7120 = $$6119;$$8 = $$7;$111 = $109;
           break;
          }
         } else {
          $$3110 = $$1108138;$$7120 = $$5118137;$$8 = $$6106139;$111 = $80;
         }
        } else {
         $$3110 = $$1108138;$$7120 = $$5118137;$$8 = $$6106139;$111 = $80;
        }
       } while(0);
       $110 = $111&255;
       $112 = (($110) + -1)|0;
       $113 = ($112>>>0)<(127);
       if (!($113)) {
        break;
       }
       $114 = ((($$8)) + 1|0);
       $115 = ((($$3110)) + 4|0);
       HEAP32[$$3110>>2] = $110;
       $116 = (($$7120) + -1)|0;
       $117 = ($116|0)==(0);
       if ($117) {
        $$6106$lcssa = $114;
        break L22;
       } else {
        $$1108138 = $115;$$5118137 = $116;$$6106139 = $114;
       }
      }
      $118 = (($110) + -194)|0;
      $119 = ($118>>>0)>(50);
      if ($119) {
       $$11163 = $$8;$$5112162 = $$3110;$$9122161 = $$7120;
       label = 53;
       break L20;
      }
      $120 = ((($$8)) + 1|0);
      $121 = (14424 + ($118<<2)|0);
      $122 = HEAP32[$121>>2]|0;
      $$4 = $122;$$4111 = $$3110;$$8121 = $$7120;$$9 = $120;
      label = 43;
      continue L19;
     }
    } while(0);
    HEAP32[$1>>2] = $$6106$lcssa;
    $$099 = $2;
    label = 59;
    continue L19;
   }
   else if ((label|0) == 25) {
    label = 0;
    $58 = HEAP8[$$4104>>0]|0;
    $59 = $58&255;
    $60 = $59 >>> 3;
    $61 = (($60) + -16)|0;
    $62 = $$2 >> 26;
    $63 = (($60) + ($62))|0;
    $64 = $61 | $63;
    $65 = ($64>>>0)>(7);
    if ($65) {
     $$5112 = $0;$$6 = $$2;$$9$sink = $$4104;$$9122 = $$4117;
     label = 52;
    } else {
     $66 = ((($$4104)) + 1|0);
     $67 = $$2 & 33554432;
     $68 = ($67|0)==(0);
     if ($68) {
      $$5105 = $66;
     } else {
      $69 = HEAP8[$66>>0]|0;
      $70 = $69 & -64;
      $71 = ($70<<24>>24)==(-128);
      if (!($71)) {
       $$5112 = $0;$$6 = $$2;$$9$sink = $$4104;$$9122 = $$4117;
       label = 52;
       break;
      }
      $72 = ((($$4104)) + 2|0);
      $73 = $$2 & 524288;
      $74 = ($73|0)==(0);
      if ($74) {
       $$5105 = $72;
      } else {
       $75 = HEAP8[$72>>0]|0;
       $76 = $75 & -64;
       $77 = ($76<<24>>24)==(-128);
       if (!($77)) {
        $$5112 = $0;$$6 = $$2;$$9$sink = $$4104;$$9122 = $$4117;
        label = 52;
        break;
       }
       $78 = ((($$4104)) + 3|0);
       $$5105 = $78;
      }
     }
     $79 = (($$4117) + -1)|0;
     $$1101$ph = $$5105;$$1114$ph = $79;
     label = 14;
     continue L19;
    }
   }
   else if ((label|0) == 43) {
    label = 0;
    $123 = HEAP8[$$9>>0]|0;
    $124 = $123&255;
    $125 = $124 >>> 3;
    $126 = (($125) + -16)|0;
    $127 = $$4 >> 26;
    $128 = (($125) + ($127))|0;
    $129 = $126 | $128;
    $130 = ($129>>>0)>(7);
    if ($130) {
     $$5112 = $$4111;$$6 = $$4;$$9$sink = $$9;$$9122 = $$8121;
     label = 52;
    } else {
     $131 = $$4 << 6;
     $132 = ((($$9)) + 1|0);
     $133 = (($124) + -128)|0;
     $134 = $133 | $131;
     $135 = ($134|0)<(0);
     do {
      if ($135) {
       $136 = HEAP8[$132>>0]|0;
       $137 = $136&255;
       $138 = (($137) + -128)|0;
       $139 = ($138>>>0)>(63);
       if ($139) {
        $140 = ((($$9)) + -1|0);
        $$11124 = $140;$$5112123 = $$4111;
        break L20;
       }
       $141 = $134 << 6;
       $142 = ((($$9)) + 2|0);
       $143 = $138 | $141;
       $144 = ($143|0)<(0);
       if ($144) {
        $145 = HEAP8[$142>>0]|0;
        $146 = $145&255;
        $147 = (($146) + -128)|0;
        $148 = ($147>>>0)>(63);
        if ($148) {
         $149 = ((($$9)) + -1|0);
         $$11124 = $149;$$5112123 = $$4111;
         break L20;
        } else {
         $150 = $143 << 6;
         $151 = ((($$9)) + 3|0);
         $152 = $147 | $150;
         $$10 = $151;$$5 = $152;
         break;
        }
       } else {
        $$10 = $142;$$5 = $143;
       }
      } else {
       $$10 = $132;$$5 = $134;
      }
     } while(0);
     $153 = ((($$4111)) + 4|0);
     HEAP32[$$4111>>2] = $$5;
     $154 = (($$8121) + -1)|0;
     $$1108$ph = $153;$$5118$ph = $154;$$6106$ph = $$10;
     label = 15;
     continue L19;
    }
   }
   else if ((label|0) == 59) {
    label = 0;
    return ($$099|0);
   }
  } while(0);
  if ((label|0) == 52) {
   label = 0;
   $155 = ((($$9$sink)) + -1|0);
   $156 = ($$6|0)==(0);
   if ($156) {
    $$11163 = $155;$$5112162 = $$5112;$$9122161 = $$9122;
    label = 53;
   } else {
    $$11124 = $155;$$5112123 = $$5112;
   }
  }
  if ((label|0) == 53) {
   label = 0;
   $157 = HEAP8[$$11163>>0]|0;
   $158 = ($157<<24>>24)==(0);
   if ($158) {
    $159 = ($$5112162|0)==(0|0);
    if (!($159)) {
     HEAP32[$$5112162>>2] = 0;
     HEAP32[$1>>2] = 0;
    }
    $160 = (($2) - ($$9122161))|0;
    $$099 = $160;
    label = 59;
    continue;
   } else {
    $$11124 = $$11163;$$5112123 = $$5112162;
   }
  }
  $161 = (___errno_location()|0);
  HEAP32[$161>>2] = 84;
  $162 = ($$5112123|0)==(0|0);
  if ($162) {
   $$099 = -1;
   label = 59;
   continue;
  }
  HEAP32[$1>>2] = $$11124;
  $$099 = -1;
  label = 59;
 }
 return (0)|0;
}
function ___pthread_self_442() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (_pthread_self()|0);
 return ($0|0);
}
function _wcsnrtombs($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$04773 = 0, $$05072 = 0, $$053$$050 = 0, $$05371 = 0, $$062 = 0, $$074 = 0, $$1 = 0, $$148 = 0, $$151 = 0, $$154 = 0, $$24968 = 0, $$25267 = 0, $$255 = 0, $$269 = 0, $$3 = 0, $$35666 = 0, $$57 = 0, $$cast = 0, $$pr$pre = 0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $or$cond70 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 272|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(272|0);
 $5 = sp + 8|0;
 $6 = sp;
 $7 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $7;
 $8 = ($0|0)!=(0|0);
 $$ = $8 ? $3 : 256;
 $$57 = $8 ? $0 : $5;
 $$cast = $7;
 $9 = ($7|0)!=(0);
 $10 = ($$|0)!=(0);
 $or$cond70 = $10 & $9;
 L1: do {
  if ($or$cond70) {
   $$04773 = 0;$$05072 = $2;$$05371 = $$;$$074 = $$57;$45 = $$cast;
   while(1) {
    $11 = ($$05072>>>0)>=($$05371>>>0);
    $12 = ($$05072>>>0)>(32);
    $13 = $11 | $12;
    if (!($13)) {
     $$062 = $$074;$$148 = $$04773;$$151 = $$05072;$$255 = $$05371;$25 = $45;
     break L1;
    }
    $$053$$050 = $11 ? $$05371 : $$05072;
    $14 = (($$05072) - ($$053$$050))|0;
    $15 = (_wcsrtombs($$074,$6,$$053$$050,0)|0);
    $16 = ($15|0)==(-1);
    if ($16) {
     break;
    }
    $17 = ($$074|0)==($5|0);
    $18 = (($$074) + ($15)|0);
    $19 = $17 ? 0 : $15;
    $$154 = (($$05371) - ($19))|0;
    $$1 = $17 ? $$074 : $18;
    $20 = (($15) + ($$04773))|0;
    $21 = HEAP32[$6>>2]|0;
    $22 = ($21|0)!=(0|0);
    $23 = ($$154|0)!=(0);
    $or$cond = $23 & $22;
    if ($or$cond) {
     $$04773 = $20;$$05072 = $14;$$05371 = $$154;$$074 = $$1;$45 = $21;
    } else {
     $$062 = $$1;$$148 = $20;$$151 = $14;$$255 = $$154;$25 = $21;
     break L1;
    }
   }
   $$pr$pre = HEAP32[$6>>2]|0;
   $$062 = $$074;$$148 = -1;$$151 = $14;$$255 = 0;$25 = $$pr$pre;
  } else {
   $$062 = $$57;$$148 = 0;$$151 = $2;$$255 = $$;$25 = $$cast;
  }
 } while(0);
 $24 = ($25|0)==(0|0);
 L8: do {
  if ($24) {
   $$3 = $$148;
  } else {
   $26 = ($$255|0)!=(0);
   $27 = ($$151|0)!=(0);
   $28 = $26 & $27;
   if ($28) {
    $$24968 = $$148;$$25267 = $$151;$$269 = $$062;$$35666 = $$255;$30 = $25;
    while(1) {
     $29 = HEAP32[$30>>2]|0;
     $31 = (_wcrtomb($$269,$29,0)|0);
     $32 = (($31) + 1)|0;
     $33 = ($32>>>0)<(2);
     if ($33) {
      break;
     }
     $35 = HEAP32[$6>>2]|0;
     $36 = ((($35)) + 4|0);
     HEAP32[$6>>2] = $36;
     $37 = (($$25267) + -1)|0;
     $38 = (($$269) + ($31)|0);
     $39 = (($$35666) - ($31))|0;
     $40 = (($31) + ($$24968))|0;
     $41 = ($39|0)!=(0);
     $42 = ($37|0)!=(0);
     $43 = $41 & $42;
     if ($43) {
      $$24968 = $40;$$25267 = $37;$$269 = $38;$$35666 = $39;$30 = $36;
     } else {
      $$3 = $40;
      break L8;
     }
    }
    $34 = ($31|0)==(0);
    if ($34) {
     HEAP32[$6>>2] = 0;
     $$3 = $$24968;
    } else {
     $$3 = -1;
    }
   } else {
    $$3 = $$148;
   }
  }
 } while(0);
 if ($8) {
  $44 = HEAP32[$6>>2]|0;
  HEAP32[$1>>2] = $44;
 }
 STACKTOP = sp;return ($$3|0);
}
function _wcsrtombs($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$05675 = 0, $$057$lcssa = 0, $$05785 = 0, $$05874 = 0, $$1 = 0, $$159 = 0, $$260$lcssa = 0, $$26084 = 0, $$279 = 0, $$3 = 0, $$361 = 0, $$478 = 0, $$5 = 0, $$pn = 0, $$pre = 0, $$pre100 = 0, $$pre101 = 0, $$pre102 = 0, $10 = 0;
 var $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==(0|0);
 L1: do {
  if ($5) {
   $7 = HEAP32[$1>>2]|0;
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)==(0);
   if ($9) {
    $$0 = 0;
   } else {
    $$05675 = $7;$$05874 = 0;$11 = $8;
    while(1) {
     $10 = ($11>>>0)>(127);
     if ($10) {
      $12 = (_wcrtomb($4,$11,0)|0);
      $13 = ($12|0)==(-1);
      if ($13) {
       $$0 = -1;
       break L1;
      } else {
       $$pn = $12;
      }
     } else {
      $$pn = 1;
     }
     $$159 = (($$pn) + ($$05874))|0;
     $14 = ((($$05675)) + 4|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==(0);
     if ($16) {
      $$0 = $$159;
      break;
     } else {
      $$05675 = $14;$$05874 = $$159;$11 = $15;
     }
    }
   }
  } else {
   $6 = ($2>>>0)>(3);
   L9: do {
    if ($6) {
     $$pre = HEAP32[$1>>2]|0;
     $$05785 = $0;$$26084 = $2;$19 = $$pre;
     while(1) {
      $18 = HEAP32[$19>>2]|0;
      $20 = (($18) + -1)|0;
      $21 = ($20>>>0)>(126);
      if ($21) {
       $22 = ($18|0)==(0);
       if ($22) {
        break;
       }
       $24 = (_wcrtomb($$05785,$18,0)|0);
       $25 = ($24|0)==(-1);
       if ($25) {
        $$0 = -1;
        break L1;
       }
       $26 = (($$05785) + ($24)|0);
       $27 = (($$26084) - ($24))|0;
       $$1 = $26;$$361 = $27;$32 = $19;
      } else {
       $28 = $18&255;
       $29 = ((($$05785)) + 1|0);
       HEAP8[$$05785>>0] = $28;
       $30 = (($$26084) + -1)|0;
       $$pre100 = HEAP32[$1>>2]|0;
       $$1 = $29;$$361 = $30;$32 = $$pre100;
      }
      $31 = ((($32)) + 4|0);
      HEAP32[$1>>2] = $31;
      $33 = ($$361>>>0)>(3);
      if ($33) {
       $$05785 = $$1;$$26084 = $$361;$19 = $31;
      } else {
       $$057$lcssa = $$1;$$260$lcssa = $$361;
       break L9;
      }
     }
     HEAP8[$$05785>>0] = 0;
     HEAP32[$1>>2] = 0;
     $23 = (($2) - ($$26084))|0;
     $$0 = $23;
     break L1;
    } else {
     $$057$lcssa = $0;$$260$lcssa = $2;
    }
   } while(0);
   $17 = ($$260$lcssa|0)==(0);
   if ($17) {
    $$0 = $2;
   } else {
    $$pre101 = HEAP32[$1>>2]|0;
    $$279 = $$057$lcssa;$$478 = $$260$lcssa;$35 = $$pre101;
    while(1) {
     $34 = HEAP32[$35>>2]|0;
     $36 = (($34) + -1)|0;
     $37 = ($36>>>0)>(126);
     if ($37) {
      $38 = ($34|0)==(0);
      if ($38) {
       label = 19;
       break;
      }
      $40 = (_wcrtomb($4,$34,0)|0);
      $41 = ($40|0)==(-1);
      if ($41) {
       $$0 = -1;
       break L1;
      }
      $42 = ($$478>>>0)<($40>>>0);
      if ($42) {
       label = 22;
       break;
      }
      $44 = HEAP32[$35>>2]|0;
      (_wcrtomb($$279,$44,0)|0);
      $45 = (($$279) + ($40)|0);
      $46 = (($$478) - ($40))|0;
      $$3 = $45;$$5 = $46;$51 = $35;
     } else {
      $47 = $34&255;
      $48 = ((($$279)) + 1|0);
      HEAP8[$$279>>0] = $47;
      $49 = (($$478) + -1)|0;
      $$pre102 = HEAP32[$1>>2]|0;
      $$3 = $48;$$5 = $49;$51 = $$pre102;
     }
     $50 = ((($51)) + 4|0);
     HEAP32[$1>>2] = $50;
     $52 = ($$5|0)==(0);
     if ($52) {
      $$0 = $2;
      break L1;
     } else {
      $$279 = $$3;$$478 = $$5;$35 = $50;
     }
    }
    if ((label|0) == 19) {
     HEAP8[$$279>>0] = 0;
     HEAP32[$1>>2] = 0;
     $39 = (($2) - ($$478))|0;
     $$0 = $39;
     break;
    }
    else if ((label|0) == 22) {
     $43 = (($2) - ($$478))|0;
     $$0 = $43;
     break;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function _recv($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = (_recvfrom($0,$1,$2,$3,0,0)|0);
 return ($4|0);
}
function _recvfrom($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $6 = sp + 8|0;
 HEAP32[$6>>2] = $0;
 $7 = ((($6)) + 4|0);
 $8 = $1;
 HEAP32[$7>>2] = $8;
 $9 = ((($6)) + 8|0);
 HEAP32[$9>>2] = $2;
 $10 = ((($6)) + 12|0);
 HEAP32[$10>>2] = $3;
 $11 = ((($6)) + 16|0);
 $12 = $4;
 HEAP32[$11>>2] = $12;
 $13 = ((($6)) + 20|0);
 $14 = $5;
 HEAP32[$13>>2] = $14;
 $15 = $6;
 HEAP32[$vararg_buffer>>2] = 12;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $15;
 $16 = (___syscall102(102,($vararg_buffer|0))|0);
 $17 = (___syscall_ret($16)|0);
 STACKTOP = sp;return ($17|0);
}
function _htons($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___bswap_16($0)|0);
 return ($1|0);
}
function ___bswap_16($0) {
 $0 = $0|0;
 var $rev = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $rev = (_llvm_bswap_i16(($0|0))|0);
 return ($rev|0);
}
function _htonl($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___bswap_32($0)|0);
 return ($1|0);
}
function ___bswap_32($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (_llvm_bswap_i32(($0|0))|0);
 return ($1|0);
}
function _getc($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 76|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)<(0);
 if ($3) {
  label = 3;
 } else {
  $4 = (___lockfile($0)|0);
  $5 = ($4|0)==(0);
  if ($5) {
   label = 3;
  } else {
   $15 = ((($0)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = ((($0)) + 8|0);
   $18 = HEAP32[$17>>2]|0;
   $19 = ($16>>>0)<($18>>>0);
   if ($19) {
    $20 = ((($16)) + 1|0);
    HEAP32[$15>>2] = $20;
    $21 = HEAP8[$16>>0]|0;
    $22 = $21&255;
    $24 = $22;
   } else {
    $23 = (___uflow($0)|0);
    $24 = $23;
   }
   ___unlockfile($0);
   $$0 = $24;
  }
 }
 do {
  if ((label|0) == 3) {
   $6 = ((($0)) + 4|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = ((($0)) + 8|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = ($7>>>0)<($9>>>0);
   if ($10) {
    $11 = ((($7)) + 1|0);
    HEAP32[$6>>2] = $11;
    $12 = HEAP8[$7>>0]|0;
    $13 = $12&255;
    $$0 = $13;
    break;
   } else {
    $14 = (___uflow($0)|0);
    $$0 = $14;
    break;
   }
  }
 } while(0);
 return ($$0|0);
}
function _ntohs($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___bswap_16_494($0)|0);
 return ($1|0);
}
function ___bswap_16_494($0) {
 $0 = $0|0;
 var $rev = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $rev = (_llvm_bswap_i16(($0|0))|0);
 return ($rev|0);
}
function _srand($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (($0) + -1)|0;
 $2 = 160984;
 $3 = $2;
 HEAP32[$3>>2] = $1;
 $4 = (($2) + 4)|0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 return;
}
function _rand() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = 160984;
 $1 = $0;
 $2 = HEAP32[$1>>2]|0;
 $3 = (($0) + 4)|0;
 $4 = $3;
 $5 = HEAP32[$4>>2]|0;
 $6 = (___muldi3(($2|0),($5|0),1284865837,1481765933)|0);
 $7 = tempRet0;
 $8 = (_i64Add(($6|0),($7|0),1,0)|0);
 $9 = tempRet0;
 $10 = 160984;
 $11 = $10;
 HEAP32[$11>>2] = $8;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 $14 = (_bitshift64Lshr(($8|0),($9|0),33)|0);
 $15 = tempRet0;
 return ($14|0);
}
function _select($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr3 = 0, $vararg_ptr4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $vararg_buffer = sp;
 $5 = $1;
 $6 = $2;
 $7 = $3;
 $8 = $4;
 HEAP32[$vararg_buffer>>2] = $0;
 $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
 HEAP32[$vararg_ptr1>>2] = $5;
 $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
 HEAP32[$vararg_ptr2>>2] = $6;
 $vararg_ptr3 = ((($vararg_buffer)) + 12|0);
 HEAP32[$vararg_ptr3>>2] = $7;
 $vararg_ptr4 = ((($vararg_buffer)) + 16|0);
 HEAP32[$vararg_ptr4>>2] = $8;
 $9 = (___syscall142(142,($vararg_buffer|0))|0);
 $10 = (___syscall_ret($9)|0);
 STACKTOP = sp;return ($10|0);
}
function _vasprintf($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vacopy_currentptr = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $vacopy_currentptr = HEAP32[$2>>2]|0;
 HEAP32[$3>>2] = $vacopy_currentptr;
 $4 = (_vsnprintf(0,0,$1,$3)|0);
 $5 = ($4|0)<(0);
 if ($5) {
  $$0 = -1;
 } else {
  $6 = (($4) + 1)|0;
  $7 = (_malloc($6)|0);
  HEAP32[$0>>2] = $7;
  $8 = ($7|0)==(0|0);
  if ($8) {
   $$0 = -1;
  } else {
   $9 = (_vsnprintf($7,$6,$1,$2)|0);
   $$0 = $9;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function _ungetc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$pr = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==(-1);
 do {
  if ($2) {
   $$1 = -1;
  } else {
   $3 = ((($1)) + 76|0);
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)>(-1);
   if ($5) {
    $6 = (___lockfile($1)|0);
    $11 = $6;
   } else {
    $11 = 0;
   }
   $7 = ((($1)) + 4|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ($8|0)==(0|0);
   if ($9) {
    (___toread($1)|0);
    $$pr = HEAP32[$7>>2]|0;
    $12 = ($$pr|0)==(0|0);
    $13 = ($11|0)!=(0);
    if ($12) {
     $24 = $13;
    } else {
     $18 = $$pr;$23 = $13;
     label = 7;
    }
   } else {
    $10 = ($11|0)!=(0);
    $18 = $8;$23 = $10;
    label = 7;
   }
   if ((label|0) == 7) {
    $14 = ((($1)) + 44|0);
    $15 = HEAP32[$14>>2]|0;
    $16 = ((($15)) + -8|0);
    $17 = ($18>>>0)>($16>>>0);
    if ($17) {
     $19 = $0&255;
     $20 = ((($18)) + -1|0);
     HEAP32[$7>>2] = $20;
     HEAP8[$20>>0] = $19;
     $21 = HEAP32[$1>>2]|0;
     $22 = $21 & -17;
     HEAP32[$1>>2] = $22;
     if (!($23)) {
      $$1 = $0;
      break;
     }
     ___unlockfile($1);
     $$1 = $0;
     break;
    } else {
     $24 = $23;
    }
   }
   if ($24) {
    ___unlockfile($1);
    $$1 = -1;
   } else {
    $$1 = -1;
   }
  }
 } while(0);
 return ($$1|0);
}
function _fputc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 76|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ($3|0)<(0);
 $5 = $0&255;
 $6 = $0 & 255;
 if ($4) {
  label = 3;
 } else {
  $7 = (___lockfile($1)|0);
  $8 = ($7|0)==(0);
  if ($8) {
   label = 3;
  } else {
   $20 = ((($1)) + 75|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = $21 << 24 >> 24;
   $23 = ($6|0)==($22|0);
   if ($23) {
    label = 10;
   } else {
    $24 = ((($1)) + 20|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ((($1)) + 16|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = ($25>>>0)<($27>>>0);
    if ($28) {
     $29 = ((($25)) + 1|0);
     HEAP32[$24>>2] = $29;
     HEAP8[$25>>0] = $5;
     $31 = $6;
    } else {
     label = 10;
    }
   }
   if ((label|0) == 10) {
    $30 = (___overflow($1,$0)|0);
    $31 = $30;
   }
   ___unlockfile($1);
   $$0 = $31;
  }
 }
 do {
  if ((label|0) == 3) {
   $9 = ((($1)) + 75|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = $10 << 24 >> 24;
   $12 = ($6|0)==($11|0);
   if (!($12)) {
    $13 = ((($1)) + 20|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = ((($1)) + 16|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($14>>>0)<($16>>>0);
    if ($17) {
     $18 = ((($14)) + 1|0);
     HEAP32[$13>>2] = $18;
     HEAP8[$14>>0] = $5;
     $$0 = $6;
     break;
    }
   }
   $19 = (___overflow($1,$0)|0);
   $$0 = $19;
  }
 } while(0);
 return ($$0|0);
}
function _wmemchr($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$067 = 0, $$08 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $9 = 0;
  } else {
   $$067 = $0;$$08 = $2;
   while(1) {
    $4 = HEAP32[$$067>>2]|0;
    $5 = ($4|0)==($1|0);
    if ($5) {
     $9 = $$067;
     break L1;
    }
    $6 = (($$08) + -1)|0;
    $7 = ((($$067)) + 4|0);
    $8 = ($6|0)==(0);
    if ($8) {
     $9 = 0;
     break;
    } else {
     $$067 = $7;$$08 = $6;
    }
   }
  }
 } while(0);
 return ($9|0);
}
function _strtod($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_strtox($0,$1,1));
 return (+$2);
}
function _strtox($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0.0, $9 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $3 = sp;
 dest=$3; stop=dest+124|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 $4 = ((($3)) + 4|0);
 HEAP32[$4>>2] = $0;
 $5 = ((($3)) + 8|0);
 HEAP32[$5>>2] = (-1);
 $6 = ((($3)) + 44|0);
 HEAP32[$6>>2] = $0;
 $7 = ((($3)) + 76|0);
 HEAP32[$7>>2] = -1;
 ___shlim($3,0);
 $8 = (+___floatscan($3,$2,1));
 $9 = ((($3)) + 108|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = HEAP32[$4>>2]|0;
 $12 = HEAP32[$5>>2]|0;
 $13 = (($11) - ($12))|0;
 $14 = (($13) + ($10))|0;
 $15 = ($1|0)==(0|0);
 if (!($15)) {
  $16 = ($14|0)!=(0);
  $17 = (($0) + ($14)|0);
  $18 = $16 ? $17 : $0;
  HEAP32[$1>>2] = $18;
 }
 STACKTOP = sp;return (+$8);
}
function _strcspn($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$01824 = 0, $$019$sink = 0, $$01922 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $div = 0;
 var $div20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = HEAP8[$1>>0]|0;
 $4 = ($3<<24>>24)==(0);
 L1: do {
  if ($4) {
   label = 3;
  } else {
   $5 = ((($1)) + 1|0);
   $6 = HEAP8[$5>>0]|0;
   $7 = ($6<<24>>24)==(0);
   if ($7) {
    label = 3;
   } else {
    ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;HEAP32[$2+16>>2]=0|0;HEAP32[$2+20>>2]=0|0;HEAP32[$2+24>>2]=0|0;HEAP32[$2+28>>2]=0|0;
    $$01824 = $1;$13 = $3;
    while(1) {
     $12 = $13 & 31;
     $14 = $12&255;
     $15 = 1 << $14;
     $div20 = ($13&255) >>> 5;
     $16 = $div20&255;
     $17 = (($2) + ($16<<2)|0);
     $18 = HEAP32[$17>>2]|0;
     $19 = $18 | $15;
     HEAP32[$17>>2] = $19;
     $20 = ((($$01824)) + 1|0);
     $21 = HEAP8[$20>>0]|0;
     $22 = ($21<<24>>24)==(0);
     if ($22) {
      break;
     } else {
      $$01824 = $20;$13 = $21;
     }
    }
    $10 = HEAP8[$0>>0]|0;
    $11 = ($10<<24>>24)==(0);
    if ($11) {
     $$019$sink = $0;
    } else {
     $$01922 = $0;$23 = $10;
     while(1) {
      $div = ($23&255) >>> 5;
      $24 = $div&255;
      $25 = (($2) + ($24<<2)|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = $23 & 31;
      $28 = $27&255;
      $29 = 1 << $28;
      $30 = $26 & $29;
      $31 = ($30|0)==(0);
      if (!($31)) {
       $$019$sink = $$01922;
       break L1;
      }
      $32 = ((($$01922)) + 1|0);
      $33 = HEAP8[$32>>0]|0;
      $34 = ($33<<24>>24)==(0);
      if ($34) {
       $$019$sink = $32;
       break;
      } else {
       $$01922 = $32;$23 = $33;
      }
     }
    }
   }
  }
 } while(0);
 if ((label|0) == 3) {
  $8 = $3 << 24 >> 24;
  $9 = (___strchrnul($0,$8)|0);
  $$019$sink = $9;
 }
 $35 = $$019$sink;
 $36 = $0;
 $37 = (($35) - ($36))|0;
 STACKTOP = sp;return ($37|0);
}
function _strtof($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0.0, $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_strtox($0,$1,0));
 $3 = $2;
 return (+$3);
}
function _strtold($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (+_strtox($0,$1,2));
 return (+$2);
}
function _strtof_l($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (+_strtof($0,$1));
 return (+$3);
}
function _strtod_l($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (+_strtod($0,$1));
 return (+$3);
}
function _strtold_l($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (+_strtold($0,$1));
 return (+$3);
}
function _strerror_r($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (_strerror($0)|0);
 $4 = (_strlen($3)|0);
 $5 = ($4>>>0)<($2>>>0);
 if ($5) {
  $9 = (($4) + 1)|0;
  _memcpy(($1|0),($3|0),($9|0))|0;
  $$0 = 0;
 } else {
  $6 = ($2|0)==(0);
  $7 = (($2) + -1)|0;
  if ($6) {
   $$0 = 34;
  } else {
   $8 = (($1) + ($7)|0);
   _memcpy(($1|0),($3|0),($7|0))|0;
   HEAP8[$8>>0] = 0;
   $$0 = 34;
  }
 }
 return ($$0|0);
}
function _strpbrk($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_strcspn($0,$1)|0);
 $3 = (($0) + ($2)|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)!=(0);
 $6 = $5 ? $3 : 0;
 return ($6|0);
}
function _wcscpy($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$05 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $0;$$05 = $1;
 while(1) {
  $2 = ((($$05)) + 4|0);
  $3 = HEAP32[$$05>>2]|0;
  $4 = ((($$0)) + 4|0);
  HEAP32[$$0>>2] = $3;
  $5 = ($3|0)==(0);
  if ($5) {
   break;
  } else {
   $$0 = $4;$$05 = $2;
  }
 }
 return ($0|0);
}
function _wmemset($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$056 = 0, $$07 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if (!($3)) {
  $$056 = $2;$$07 = $0;
  while(1) {
   $4 = (($$056) + -1)|0;
   $5 = ((($$07)) + 4|0);
   HEAP32[$$07>>2] = $1;
   $6 = ($4|0)==(0);
   if ($6) {
    break;
   } else {
    $$056 = $4;$$07 = $5;
   }
  }
 }
 return ($0|0);
}
function _wcscat($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_wcslen($0)|0);
 $3 = (($0) + ($2<<2)|0);
 (_wcscpy($3,$1)|0);
 return ($0|0);
}
function _wcsncmp($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$01218 = 0, $$01317 = 0, $$019 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $14 = 0;
  } else {
   $$01218 = $1;$$01317 = $0;$$019 = $2;
   while(1) {
    $4 = HEAP32[$$01317>>2]|0;
    $5 = HEAP32[$$01218>>2]|0;
    $6 = ($4|0)!=($5|0);
    $7 = ($4|0)==(0);
    $or$cond = $7 | $6;
    $8 = ($5|0)==(0);
    $or$cond14 = $8 | $or$cond;
    if ($or$cond14) {
     break;
    }
    $9 = (($$019) + -1)|0;
    $10 = ((($$01317)) + 4|0);
    $11 = ((($$01218)) + 4|0);
    $12 = ($9|0)==(0);
    if ($12) {
     $14 = 0;
     break L1;
    } else {
     $$01218 = $11;$$01317 = $10;$$019 = $9;
    }
   }
   $13 = (($4) - ($5))|0;
   $14 = $13;
  }
 } while(0);
 return ($14|0);
}
function _wcsrchr($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$010 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_wcslen($0)|0);
 $3 = ($2|0)>(-1);
 L1: do {
  if ($3) {
   $4 = (($0) + ($2<<2)|0);
   $$010 = $4;
   while(1) {
    $5 = HEAP32[$$010>>2]|0;
    $6 = ($5|0)==($1|0);
    if ($6) {
     $9 = $$010;
     break L1;
    }
    $7 = ((($$010)) + -4|0);
    $8 = ($7>>>0)<($0>>>0);
    if ($8) {
     $9 = 0;
     break;
    } else {
     $$010 = $7;
    }
   }
  } else {
   $9 = 0;
  }
 } while(0);
 return ($9|0);
}
function _wcsstr($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 if ($3) {
  $$0 = $0;
 } else {
  $4 = HEAP32[$0>>2]|0;
  $5 = ($4|0)==(0);
  if ($5) {
   $$0 = 0;
  } else {
   $6 = (_wcschr($0,$2)|0);
   $7 = ($6|0)==(0|0);
   if ($7) {
    $$0 = 0;
   } else {
    $8 = ((($1)) + 4|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ($9|0)==(0);
    if ($10) {
     $$0 = $6;
    } else {
     $11 = ((($6)) + 4|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = ($12|0)==(0);
     if ($13) {
      $$0 = 0;
     } else {
      $14 = (_twoway_wcsstr($6,$1)|0);
      $$0 = $14;
     }
    }
   }
  }
 }
 return ($$0|0);
}
function _twoway_wcsstr($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0145 = 0, $$0147 = 0, $$0148 = 0, $$0148$be = 0, $$0149 = 0, $$0154$ph$ph$lcssa194 = 0, $$0154$ph$ph$lcssa194295 = 0, $$0154$ph$ph229 = 0, $$0158217 = 0, $$0162$ph174$ph228 = 0, $$0162$ph174223 = 0, $$0162$ph235 = 0, $$0164$ph$lcssa = 0, $$0164$ph$lcssa294 = 0, $$0164$ph234 = 0, $$0166$lcssa292293 = 0, $$0166238 = 0, $$1155$$0154 = 0, $$1155$ph$ph$lcssa185 = 0, $$1155$ph$ph208 = 0;
 var $$1159197 = 0, $$1163$ph172$ph207 = 0, $$1163$ph172202 = 0, $$1163$ph214 = 0, $$1165$$0164 = 0, $$1165$$0164$ = 0, $$1165$ph$lcssa = 0, $$1165$ph213 = 0, $$2160196 = 0, $$3 = 0, $$3152 = 0, $$3157 = 0, $$3161 = 0, $$sink = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $$0154$ph$ph$lcssa194295 = 1;$$0164$ph$lcssa294 = -1;$$0166$lcssa292293 = 0;$$1155$ph$ph$lcssa185 = 1;$$1165$ph$lcssa = -1;
   label = 26;
  } else {
   $$0166238 = 0;
   while(1) {
    $6 = (($0) + ($$0166238<<2)|0);
    $7 = HEAP32[$6>>2]|0;
    $8 = ($7|0)==(0);
    if ($8) {
     $$3 = 0;
     break L1;
    }
    $5 = (($$0166238) + 1)|0;
    $9 = (($1) + ($5<<2)|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($10|0)==(0);
    if ($11) {
     break;
    } else {
     $$0166238 = $5;
    }
   }
   $4 = ($5>>>0)>(1);
   if ($4) {
    $$0162$ph235 = 0;$$0164$ph234 = -1;$101 = 1;
    L6: while(1) {
     $$0154$ph$ph229 = 1;$$0162$ph174$ph228 = $$0162$ph235;$104 = $101;
     while(1) {
      $$0162$ph174223 = $$0162$ph174$ph228;$103 = $104;
      L10: while(1) {
       $$0158217 = 1;$19 = $103;
       while(1) {
        $15 = (($$0158217) + ($$0164$ph234))|0;
        $16 = (($1) + ($15<<2)|0);
        $17 = HEAP32[$16>>2]|0;
        $18 = (($1) + ($19<<2)|0);
        $20 = HEAP32[$18>>2]|0;
        $21 = ($17|0)==($20|0);
        if (!($21)) {
         break L10;
        }
        $22 = ($$0158217|0)==($$0154$ph$ph229|0);
        $13 = (($$0158217) + 1)|0;
        if ($22) {
         break;
        }
        $12 = (($13) + ($$0162$ph174223))|0;
        $14 = ($12>>>0)<($5>>>0);
        if ($14) {
         $$0158217 = $13;$19 = $12;
        } else {
         $$0154$ph$ph$lcssa194 = $$0154$ph$ph229;$$0164$ph$lcssa = $$0164$ph234;
         break L6;
        }
       }
       $23 = (($$0154$ph$ph229) + ($$0162$ph174223))|0;
       $24 = (($23) + 1)|0;
       $25 = ($24>>>0)<($5>>>0);
       if ($25) {
        $$0162$ph174223 = $23;$103 = $24;
       } else {
        $$0154$ph$ph$lcssa194 = $$0154$ph$ph229;$$0164$ph$lcssa = $$0164$ph234;
        break L6;
       }
      }
      $26 = ($17|0)>($20|0);
      $27 = (($19) - ($$0164$ph234))|0;
      if (!($26)) {
       break;
      }
      $31 = (($19) + 1)|0;
      $32 = ($31>>>0)<($5>>>0);
      if ($32) {
       $$0154$ph$ph229 = $27;$$0162$ph174$ph228 = $19;$104 = $31;
      } else {
       $$0154$ph$ph$lcssa194 = $27;$$0164$ph$lcssa = $$0164$ph234;
       break L6;
      }
     }
     $28 = (($$0162$ph174223) + 1)|0;
     $29 = (($$0162$ph174223) + 2)|0;
     $30 = ($29>>>0)<($5>>>0);
     if ($30) {
      $$0162$ph235 = $28;$$0164$ph234 = $$0162$ph174223;$101 = $29;
     } else {
      $$0154$ph$ph$lcssa194 = 1;$$0164$ph$lcssa = $$0162$ph174223;
      break;
     }
    }
    if ($4) {
     $$1163$ph214 = 0;$$1165$ph213 = -1;$102 = 1;
     while(1) {
      $$1155$ph$ph208 = 1;$$1163$ph172$ph207 = $$1163$ph214;$106 = $102;
      while(1) {
       $$1163$ph172202 = $$1163$ph172$ph207;$105 = $106;
       L25: while(1) {
        $$1159197 = 1;$40 = $105;
        while(1) {
         $36 = (($$1159197) + ($$1165$ph213))|0;
         $37 = (($1) + ($36<<2)|0);
         $38 = HEAP32[$37>>2]|0;
         $39 = (($1) + ($40<<2)|0);
         $41 = HEAP32[$39>>2]|0;
         $42 = ($38|0)==($41|0);
         if (!($42)) {
          break L25;
         }
         $43 = ($$1159197|0)==($$1155$ph$ph208|0);
         $34 = (($$1159197) + 1)|0;
         if ($43) {
          break;
         }
         $33 = (($34) + ($$1163$ph172202))|0;
         $35 = ($33>>>0)<($5>>>0);
         if ($35) {
          $$1159197 = $34;$40 = $33;
         } else {
          $$0154$ph$ph$lcssa194295 = $$0154$ph$ph$lcssa194;$$0164$ph$lcssa294 = $$0164$ph$lcssa;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = $$1155$ph$ph208;$$1165$ph$lcssa = $$1165$ph213;
          label = 26;
          break L1;
         }
        }
        $44 = (($$1155$ph$ph208) + ($$1163$ph172202))|0;
        $45 = (($44) + 1)|0;
        $46 = ($45>>>0)<($5>>>0);
        if ($46) {
         $$1163$ph172202 = $44;$105 = $45;
        } else {
         $$0154$ph$ph$lcssa194295 = $$0154$ph$ph$lcssa194;$$0164$ph$lcssa294 = $$0164$ph$lcssa;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = $$1155$ph$ph208;$$1165$ph$lcssa = $$1165$ph213;
         label = 26;
         break L1;
        }
       }
       $47 = ($38|0)<($41|0);
       $48 = (($40) - ($$1165$ph213))|0;
       if (!($47)) {
        break;
       }
       $52 = (($40) + 1)|0;
       $53 = ($52>>>0)<($5>>>0);
       if ($53) {
        $$1155$ph$ph208 = $48;$$1163$ph172$ph207 = $40;$106 = $52;
       } else {
        $$0154$ph$ph$lcssa194295 = $$0154$ph$ph$lcssa194;$$0164$ph$lcssa294 = $$0164$ph$lcssa;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = $48;$$1165$ph$lcssa = $$1165$ph213;
        label = 26;
        break L1;
       }
      }
      $49 = (($$1163$ph172202) + 1)|0;
      $50 = (($$1163$ph172202) + 2)|0;
      $51 = ($50>>>0)<($5>>>0);
      if ($51) {
       $$1163$ph214 = $49;$$1165$ph213 = $$1163$ph172202;$102 = $50;
      } else {
       $$0154$ph$ph$lcssa194295 = $$0154$ph$ph$lcssa194;$$0164$ph$lcssa294 = $$0164$ph$lcssa;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = 1;$$1165$ph$lcssa = $$1163$ph172202;
       label = 26;
       break;
      }
     }
    } else {
     $$0154$ph$ph$lcssa194295 = $$0154$ph$ph$lcssa194;$$0164$ph$lcssa294 = $$0164$ph$lcssa;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = 1;$$1165$ph$lcssa = -1;
     label = 26;
    }
   } else {
    $$0154$ph$ph$lcssa194295 = 1;$$0164$ph$lcssa294 = -1;$$0166$lcssa292293 = $5;$$1155$ph$ph$lcssa185 = 1;$$1165$ph$lcssa = -1;
    label = 26;
   }
  }
 } while(0);
 L35: do {
  if ((label|0) == 26) {
   $54 = (($$1165$ph$lcssa) + 1)|0;
   $55 = (($$0164$ph$lcssa294) + 1)|0;
   $56 = ($54>>>0)>($55>>>0);
   $$1155$$0154 = $56 ? $$1155$ph$ph$lcssa185 : $$0154$ph$ph$lcssa194295;
   $$1165$$0164 = $56 ? $$1165$ph$lcssa : $$0164$ph$lcssa294;
   $57 = (($1) + ($$1155$$0154<<2)|0);
   $58 = (($$1165$$0164) + 1)|0;
   $59 = (_wmemcmp($1,$57,$58)|0);
   $60 = ($59|0)==(0);
   if ($60) {
    $65 = (($$0166$lcssa292293) - ($$1155$$0154))|0;
    $$0147 = $65;$$3157 = $$1155$$0154;
   } else {
    $61 = (($$0166$lcssa292293) - ($$1165$$0164))|0;
    $62 = (($61) + -1)|0;
    $63 = ($$1165$$0164>>>0)>($62>>>0);
    $$1165$$0164$ = $63 ? $$1165$$0164 : $62;
    $64 = (($$1165$$0164$) + 1)|0;
    $$0147 = 0;$$3157 = $64;
   }
   $66 = $$0166$lcssa292293 | 63;
   $$0145 = $0;$$0148 = 0;$$0149 = $0;
   while(1) {
    $67 = $$0149;
    $68 = $$0145;
    $69 = (($67) - ($68))|0;
    $70 = $69 >> 2;
    $71 = ($70>>>0)<($$0166$lcssa292293>>>0);
    do {
     if ($71) {
      $72 = (_wmemchr($$0149,0,$66)|0);
      $73 = ($72|0)==(0|0);
      if ($73) {
       $78 = (($$0149) + ($66<<2)|0);
       $$3152 = $78;
       break;
      } else {
       $74 = $72;
       $75 = (($74) - ($68))|0;
       $76 = $75 >> 2;
       $77 = ($76>>>0)<($$0166$lcssa292293>>>0);
       if ($77) {
        $$3 = 0;
        break L35;
       } else {
        $$3152 = $72;
        break;
       }
      }
     } else {
      $$3152 = $$0149;
     }
    } while(0);
    $79 = ($58>>>0)>($$0148>>>0);
    $80 = $79 ? $58 : $$0148;
    $81 = (($1) + ($80<<2)|0);
    $82 = HEAP32[$81>>2]|0;
    $83 = ($82|0)==(0);
    L49: do {
     if ($83) {
      $$3161 = $58;
      label = 39;
     } else {
      $$2160196 = $80;$87 = $82;
      while(1) {
       $84 = (($$0145) + ($$2160196<<2)|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ($87|0)==($85|0);
       if (!($86)) {
        break;
       }
       $88 = (($$2160196) + 1)|0;
       $89 = (($1) + ($88<<2)|0);
       $90 = HEAP32[$89>>2]|0;
       $91 = ($90|0)==(0);
       if ($91) {
        $$3161 = $58;
        label = 39;
        break L49;
       } else {
        $$2160196 = $88;$87 = $90;
       }
      }
      $92 = (($$2160196) - ($$1165$$0164))|0;
      $$0148$be = 0;$$sink = $92;
     }
    } while(0);
    if ((label|0) == 39) {
     while(1) {
      label = 0;
      $94 = ($$3161>>>0)>($$0148>>>0);
      if (!($94)) {
       $$3 = $$0145;
       break L35;
      }
      $95 = (($$3161) + -1)|0;
      $96 = (($1) + ($95<<2)|0);
      $97 = HEAP32[$96>>2]|0;
      $98 = (($$0145) + ($95<<2)|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = ($97|0)==($99|0);
      if ($100) {
       $$3161 = $95;
       label = 39;
      } else {
       $$0148$be = $$0147;$$sink = $$3157;
       break;
      }
     }
    }
    $93 = (($$0145) + ($$sink<<2)|0);
    $$0145 = $93;$$0148 = $$0148$be;$$0149 = $$3152;
   }
  }
 } while(0);
 return ($$3|0);
}
function _wmemcmp($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$014 = 0, $$0813 = 0, $$0912 = 0, $10 = 0, $11 = 0, $12 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 L1: do {
  if ($3) {
   $12 = 0;
  } else {
   $$014 = $2;$$0813 = $1;$$0912 = $0;
   while(1) {
    $4 = HEAP32[$$0912>>2]|0;
    $5 = HEAP32[$$0813>>2]|0;
    $6 = ($4|0)==($5|0);
    if (!($6)) {
     break;
    }
    $7 = (($$014) + -1)|0;
    $8 = ((($$0912)) + 4|0);
    $9 = ((($$0813)) + 4|0);
    $10 = ($7|0)==(0);
    if ($10) {
     $12 = 0;
     break L1;
    } else {
     $$014 = $7;$$0813 = $9;$$0912 = $8;
    }
   }
   $11 = (($4) - ($5))|0;
   $12 = $11;
  }
 } while(0);
 return ($12|0);
}
function _wmemmove($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$01417 = 0, $$019 = 0, $$118 = 0, $$in = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $0;
 $4 = $1;
 $5 = (($3) - ($4))|0;
 $6 = $5 >> 2;
 $7 = ($6>>>0)<($2>>>0);
 if ($7) {
  $$in = $2;
  while(1) {
   $9 = (($$in) + -1)|0;
   $10 = (($1) + ($9<<2)|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = (($0) + ($9<<2)|0);
   HEAP32[$12>>2] = $11;
   $13 = ($9|0)==(0);
   if ($13) {
    break;
   } else {
    $$in = $9;
   }
  }
 } else {
  $8 = ($2|0)==(0);
  if (!($8)) {
   $$01417 = $1;$$019 = $0;$$118 = $2;
   while(1) {
    $14 = (($$118) + -1)|0;
    $15 = ((($$01417)) + 4|0);
    $16 = HEAP32[$$01417>>2]|0;
    $17 = ((($$019)) + 4|0);
    HEAP32[$$019>>2] = $16;
    $18 = ($14|0)==(0);
    if ($18) {
     break;
    } else {
     $$01417 = $15;$$019 = $17;$$118 = $14;
    }
   }
  }
 }
 return ($0|0);
}
function __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28ios_baseD2Ev($0);
 return;
}
function __ZNSt3__28ios_baseD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (18516);
 __THREW__ = 0;
 invoke_vii(943,($0|0),0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $12 = ___cxa_find_matching_catch_3(0|0)|0;
  $13 = tempRet0;
  ___clang_call_terminate($12);
  // unreachable;
 } else {
  $3 = ((($0)) + 28|0);
  __ZNSt3__26localeD2Ev($3);
  $4 = ((($0)) + 32|0);
  $5 = HEAP32[$4>>2]|0;
  _free($5);
  $6 = ((($0)) + 36|0);
  $7 = HEAP32[$6>>2]|0;
  _free($7);
  $8 = ((($0)) + 48|0);
  $9 = HEAP32[$8>>2]|0;
  _free($9);
  $10 = ((($0)) + 60|0);
  $11 = HEAP32[$10>>2]|0;
  _free($11);
  return;
 }
}
function __ZNSt3__28ios_base16__call_callbacksENS0_5eventE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 40|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($0)) + 32|0);
 $5 = ((($0)) + 36|0);
 $$0 = $3;
 while(1) {
  $6 = ($$0|0)==(0);
  if ($6) {
   break;
  }
  $7 = (($$0) + -1)|0;
  $8 = HEAP32[$4>>2]|0;
  $9 = (($8) + ($7<<2)|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = HEAP32[$5>>2]|0;
  $12 = (($11) + ($7<<2)|0);
  $13 = HEAP32[$12>>2]|0;
  FUNCTION_TABLE_viii[$10 & 2047]($1,$0,$13);
  $$0 = $7;
 }
 return;
}
function __ZNSt3__28ios_baseD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28ios_baseD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28ios_baseD2Ev($0);
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (18532);
 $1 = ((($0)) + 4|0);
 __ZNSt3__26localeD2Ev($1);
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6setbufEPci($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($0|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE7seekoffExNS_8ios_base7seekdirEj($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = $0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 8|0);
 $11 = $10;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = -1;
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE7seekposENS_4fposI11__mbstate_tEEj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 8|0);
 $9 = $8;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 $11 = (($9) + 4)|0;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE4syncEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9showmanycEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6xsgetnEPci($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$023 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $3 = ((($0)) + 12|0);
 $4 = ((($0)) + 16|0);
 $$0 = $1;$$023 = 0;
 while(1) {
  $5 = ($$023|0)<($2|0);
  if (!($5)) {
   break;
  }
  $6 = HEAP32[$3>>2]|0;
  $7 = HEAP32[$4>>2]|0;
  $8 = ($6>>>0)<($7>>>0);
  if ($8) {
   $9 = $7;
   $10 = $6;
   $11 = (($9) - ($10))|0;
   $12 = (($2) - ($$023))|0;
   $13 = ($12|0)<($11|0);
   $$sroa$speculated = $13 ? $12 : $11;
   (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0,$6,$$sroa$speculated)|0);
   $14 = (($$0) + ($$sroa$speculated)|0);
   $15 = (($$sroa$speculated) + ($$023))|0;
   $16 = HEAP32[$3>>2]|0;
   $17 = (($16) + ($$sroa$speculated)|0);
   HEAP32[$3>>2] = $17;
   $$0 = $14;$$023 = $15;
   continue;
  }
  $18 = HEAP32[$0>>2]|0;
  $19 = ((($18)) + 40|0);
  $20 = HEAP32[$19>>2]|0;
  $21 = (FUNCTION_TABLE_ii[$20 & 2047]($0)|0);
  $22 = ($21|0)==(-1);
  if ($22) {
   break;
  }
  $23 = (__ZNSt3__211char_traitsIcE12to_char_typeEi($21)|0);
  HEAP8[$$0>>0] = $23;
  $24 = ((($$0)) + 1|0);
  $25 = (($$023) + 1)|0;
  $$0 = $24;$$023 = $25;
 }
 return ($$023|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9underflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 return ($1|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE5uflowEv($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 2047]($0)|0);
 $5 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $6 = ($4|0)==($5|0);
 if ($6) {
  $7 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
  $$0 = $7;
 } else {
  $8 = ((($0)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 1|0);
  HEAP32[$8>>2] = $10;
  $11 = HEAP8[$9>>0]|0;
  $12 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($11)|0);
  $$0 = $12;
 }
 return ($$0|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9pbackfailEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 return ($2|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6xsputnEPKci($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$021 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $4 = ((($0)) + 24|0);
 $5 = ((($0)) + 28|0);
 $$0 = $1;$$021 = 0;
 while(1) {
  $6 = ($$021|0)<($2|0);
  if (!($6)) {
   break;
  }
  $7 = HEAP32[$4>>2]|0;
  $8 = HEAP32[$5>>2]|0;
  $9 = ($7>>>0)<($8>>>0);
  if ($9) {
   $19 = $8;
   $20 = $7;
   $21 = (($19) - ($20))|0;
   $22 = (($2) - ($$021))|0;
   $23 = ($22|0)<($21|0);
   $$sroa$speculated = $23 ? $22 : $21;
   (__ZNSt3__211char_traitsIcE4copyEPcPKcj($7,$$0,$$sroa$speculated)|0);
   $24 = HEAP32[$4>>2]|0;
   $25 = (($24) + ($$sroa$speculated)|0);
   HEAP32[$4>>2] = $25;
   $26 = (($$0) + ($$sroa$speculated)|0);
   $27 = (($$sroa$speculated) + ($$021))|0;
   $$0 = $26;$$021 = $27;
   continue;
  }
  $10 = HEAP32[$0>>2]|0;
  $11 = ((($10)) + 52|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP8[$$0>>0]|0;
  $14 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($13)|0);
  $15 = (FUNCTION_TABLE_iii[$12 & 2047]($0,$14)|0);
  $16 = ($15|0)==($3|0);
  if ($16) {
   break;
  }
  $17 = ((($$0)) + 1|0);
  $18 = (($$021) + 1)|0;
  $$0 = $17;$$021 = $18;
 }
 return ($$021|0);
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE8overflowEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 return ($2|0);
}
function __ZNSt3__211char_traitsIcE3eofEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return -1;
}
function __ZNSt3__211char_traitsIcE11to_int_typeEc($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = $0&255;
 return ($1|0);
}
function __ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if (!($3)) {
  _memcpy(($0|0),($1|0),($2|0))|0;
 }
 return ($0|0);
}
function __ZNSt3__211char_traitsIcE12to_char_typeEi($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = $0&255;
 return ($1|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (18596);
 $1 = ((($0)) + 4|0);
 __ZNSt3__26localeD2Ev($1);
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6setbufEPwi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($0|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE7seekoffExNS_8ios_base7seekdirEj($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = $0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = (($6) + 4)|0;
 $9 = $8;
 HEAP32[$9>>2] = 0;
 $10 = ((($0)) + 8|0);
 $11 = $10;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = -1;
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE7seekposENS_4fposI11__mbstate_tEEj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $0;
 $5 = $4;
 HEAP32[$5>>2] = 0;
 $6 = (($4) + 4)|0;
 $7 = $6;
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 8|0);
 $9 = $8;
 $10 = $9;
 HEAP32[$10>>2] = -1;
 $11 = (($9) + 4)|0;
 $12 = $11;
 HEAP32[$12>>2] = -1;
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE4syncEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9showmanycEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6xsgetnEPwi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$023 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $3 = ((($0)) + 12|0);
 $4 = ((($0)) + 16|0);
 $$0 = $1;$$023 = 0;
 while(1) {
  $5 = ($$023|0)<($2|0);
  if (!($5)) {
   break;
  }
  $6 = HEAP32[$3>>2]|0;
  $7 = HEAP32[$4>>2]|0;
  $8 = ($6>>>0)<($7>>>0);
  if ($8) {
   $9 = $7;
   $10 = $6;
   $11 = (($9) - ($10))|0;
   $12 = $11 >> 2;
   $13 = (($2) - ($$023))|0;
   $14 = ($13|0)<($12|0);
   $$sroa$speculated = $14 ? $13 : $12;
   (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0,$6,$$sroa$speculated)|0);
   $15 = (($$0) + ($$sroa$speculated<<2)|0);
   $16 = (($$sroa$speculated) + ($$023))|0;
   $17 = HEAP32[$3>>2]|0;
   $18 = (($17) + ($$sroa$speculated<<2)|0);
   HEAP32[$3>>2] = $18;
   $$0 = $15;$$023 = $16;
   continue;
  }
  $19 = HEAP32[$0>>2]|0;
  $20 = ((($19)) + 40|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = (FUNCTION_TABLE_ii[$21 & 2047]($0)|0);
  $23 = ($22|0)==(-1);
  if ($23) {
   break;
  }
  $24 = (__ZNSt3__211char_traitsIwE12to_char_typeEj($22)|0);
  HEAP32[$$0>>2] = $24;
  $25 = ((($$0)) + 4|0);
  $26 = (($$023) + 1)|0;
  $$0 = $25;$$023 = $26;
 }
 return ($$023|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9underflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 return ($1|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE5uflowEv($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + 36|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (FUNCTION_TABLE_ii[$3 & 2047]($0)|0);
 $5 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $6 = ($4|0)==($5|0);
 if ($6) {
  $7 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
  $$0 = $7;
 } else {
  $8 = ((($0)) + 12|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 4|0);
  HEAP32[$8>>2] = $10;
  $11 = HEAP32[$9>>2]|0;
  $12 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($11)|0);
  $$0 = $12;
 }
 return ($$0|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9pbackfailEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 return ($2|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6xsputnEPKwi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$021 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $4 = ((($0)) + 24|0);
 $5 = ((($0)) + 28|0);
 $$0 = $1;$$021 = 0;
 while(1) {
  $6 = ($$021|0)<($2|0);
  if (!($6)) {
   break;
  }
  $7 = HEAP32[$4>>2]|0;
  $8 = HEAP32[$5>>2]|0;
  $9 = ($7>>>0)<($8>>>0);
  if ($9) {
   $19 = $8;
   $20 = $7;
   $21 = (($19) - ($20))|0;
   $22 = $21 >> 2;
   $23 = (($2) - ($$021))|0;
   $24 = ($23|0)<($22|0);
   $$sroa$speculated = $24 ? $23 : $22;
   (__ZNSt3__211char_traitsIwE4copyEPwPKwj($7,$$0,$$sroa$speculated)|0);
   $25 = HEAP32[$4>>2]|0;
   $26 = (($25) + ($$sroa$speculated<<2)|0);
   HEAP32[$4>>2] = $26;
   $27 = (($$0) + ($$sroa$speculated<<2)|0);
   $28 = (($$sroa$speculated) + ($$021))|0;
   $$0 = $27;$$021 = $28;
   continue;
  }
  $10 = HEAP32[$0>>2]|0;
  $11 = ((($10)) + 52|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$$0>>2]|0;
  $14 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($13)|0);
  $15 = (FUNCTION_TABLE_iii[$12 & 2047]($0,$14)|0);
  $16 = ($15|0)==($3|0);
  if ($16) {
   break;
  }
  $17 = ((($$0)) + 4|0);
  $18 = (($$021) + 1)|0;
  $$0 = $17;$$021 = $18;
 }
 return ($$021|0);
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE8overflowEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 return ($2|0);
}
function __ZNSt3__211char_traitsIwE11to_int_typeEw($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($0|0);
}
function __ZNSt3__211char_traitsIwE4copyEPwPKwj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if ($3) {
  $7 = $0;
 } else {
  __THREW__ = 0;
  $4 = (invoke_iiii(944,($0|0),($1|0),($2|0))|0);
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $8 = ___cxa_find_matching_catch_3(0|0)|0;
   $9 = tempRet0;
   ___clang_call_terminate($8);
   // unreachable;
  } else {
   $7 = $0;
  }
 }
 return ($7|0);
}
function __ZNSt3__211char_traitsIwE12to_char_typeEj($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($0|0);
}
function __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED2Ev($0,18692);
 $1 = ((($0)) + 8|0);
 __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev($1);
 return;
}
function __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev($0);
 __ZdlPv($0);
 return;
}
function __ZTv0_n12_NSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev($4);
 return;
}
function __ZTv0_n12_NSt3__213basic_istreamIcNS_11char_traitsIcEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED0Ev($4);
 return;
}
function __ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED2Ev($0,18740);
 $1 = ((($0)) + 8|0);
 __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev($1);
 return;
}
function __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev($0);
 __ZdlPv($0);
 return;
}
function __ZTv0_n12_NSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev($4);
 return;
}
function __ZTv0_n12_NSt3__213basic_istreamIwNS_11char_traitsIwEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED0Ev($4);
 return;
}
function __ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED2Ev($0,18788);
 $1 = ((($0)) + 4|0);
 __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev($1);
 return;
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev($0);
 __ZdlPv($0);
 return;
}
function __ZTv0_n12_NSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev($4);
 return;
}
function __ZTv0_n12_NSt3__213basic_ostreamIcNS_11char_traitsIcEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED0Ev($4);
 return;
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED2Ev($0,18836);
 $1 = ((($0)) + 4|0);
 __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev($1);
 return;
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev($0);
 __ZdlPv($0);
 return;
}
function __ZTv0_n12_NSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev($4);
 return;
}
function __ZTv0_n12_NSt3__213basic_ostreamIwNS_11char_traitsIwEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ((($1)) + -12|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = (($0) + ($3)|0);
 __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED0Ev($4);
 return;
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED2Ev($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__219__iostream_categoryD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($0);
 return;
}
function __ZNKSt3__219__iostream_category4nameEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (155545|0);
}
function __ZNKSt3__219__iostream_category7messageEi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($2|0)!=(1);
 $5 = ($2|0)<(257);
 $or$cond = $4 & $5;
 if ($or$cond) {
  __ZNKSt3__212__do_message7messageEi($0,$1,$2);
 } else {
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
  $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc(155509)|0);
  $7 = ($6>>>0)>(4294967279);
  if ($7) {
   __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
   // unreachable;
  }
  $8 = ($6>>>0)<(11);
  if ($8) {
   $9 = $6&255;
   $10 = ((($0)) + 11|0);
   HEAP8[$10>>0] = $9;
   $$0$i$i = $0;
  } else {
   $11 = (($6) + 16)|0;
   $12 = $11 & -16;
   $13 = (__Znwj($12)|0);
   HEAP32[$0>>2] = $13;
   $14 = $12 | -2147483648;
   $15 = ((($0)) + 8|0);
   HEAP32[$15>>2] = $14;
   $16 = ((($0)) + 4|0);
   HEAP32[$16>>2] = $6;
   $$0$i$i = $13;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,155509,$6)|0);
  $17 = (($$0$i$i) + ($6)|0);
  HEAP8[$3>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($17,$3);
 }
 STACKTOP = sp;return;
}
function __ZNSt3__211char_traitsIcE6assignERcRKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP8[$1>>0]|0;
 HEAP8[$0>>0] = $2;
 return;
}
function __ZNSt3__28ios_base7failureD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__212system_errorD2Ev($0);
 return;
}
function __ZNSt3__28ios_base7failureD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28ios_base7failureD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__28ios_base5clearEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$index = 0, $$index2 = 0, $$sink = 0, $$sroa_idx2$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $3 = ((($0)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 $not$ = ($4|0)==(0|0);
 $5 = $not$&1;
 $$sink = $5 | $1;
 $6 = ((($0)) + 16|0);
 HEAP32[$6>>2] = $$sink;
 $7 = ((($0)) + 20|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $$sink & $8;
 $10 = ($9|0)==(0);
 if ($10) {
  STACKTOP = sp;return;
 }
 $11 = (___cxa_allocate_exception(16)|0);
 (__ZNSt3__217iostream_categoryEv()|0);
 HEAP32[$2>>2] = 1;
 $$sroa_idx2$i = ((($2)) + 4|0);
 HEAP32[$$sroa_idx2$i>>2] = (18900);
 __THREW__ = 0;
 invoke_viii(945,($11|0),(155610|0),($2|0));
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  ___cxa_free_exception(($11|0));
  $$index = $17;$$index2 = $18;
 } else {
  __THREW__ = 0;
  invoke_viii(946,($11|0),(1168|0),(427|0));
  $14 = __THREW__; __THREW__ = 0;
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  $$index = $15;$$index2 = $16;
 }
 ___resumeException($$index|0);
 // unreachable;
}
function __ZNSt3__217iostream_categoryEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[160992]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(160992)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   (___cxa_atexit((420|0),(18900|0),(___dso_handle|0))|0);
  }
 }
 return (18900|0);
}
function __ZNSt3__28ios_base7failureC2EPKcRKNS_10error_codeE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $tmpcast$byval_copy = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $tmpcast$byval_copy = sp + 8|0;
 $3 = sp;
 $4 = $2;
 $5 = $4;
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) + 4)|0;
 $8 = $7;
 $9 = HEAP32[$8>>2]|0;
 $10 = $3;
 $11 = $10;
 HEAP32[$11>>2] = $6;
 $12 = (($10) + 4)|0;
 $13 = $12;
 HEAP32[$13>>2] = $9;
 ;HEAP32[$tmpcast$byval_copy>>2]=HEAP32[$3>>2]|0;HEAP32[$tmpcast$byval_copy+4>>2]=HEAP32[$3+4>>2]|0;
 __ZNSt3__212system_errorC2ENS_10error_codeEPKc($0,$tmpcast$byval_copy,$1);
 HEAP32[$0>>2] = (18888);
 STACKTOP = sp;return;
}
function __ZNSt3__28ios_base4initEPv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 24|0);
 HEAP32[$2>>2] = $1;
 $not$ = ($1|0)==(0|0);
 $3 = $not$&1;
 $4 = ((($0)) + 16|0);
 HEAP32[$4>>2] = $3;
 $5 = ((($0)) + 20|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 4|0);
 HEAP32[$6>>2] = 4098;
 $7 = ((($0)) + 12|0);
 HEAP32[$7>>2] = 0;
 $8 = ((($0)) + 8|0);
 HEAP32[$8>>2] = 6;
 $9 = ((($0)) + 32|0);
 $10 = ((($0)) + 28|0);
 dest=$9; stop=dest+40|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
 __ZNSt3__26localeC2Ev($10);
 return;
}
function __ZNSt3__211char_traitsIcE11eq_int_typeEii($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==($1|0);
 return ($2|0);
}
function __ZNKSt3__28ios_base6getlocEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 28|0);
 __ZNSt3__26localeC2ERKS0_($0,$2);
 return;
}
function __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEEC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (18532);
 $1 = ((($0)) + 4|0);
 __ZNSt3__26localeC2Ev($1);
 $2 = ((($0)) + 8|0);
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;HEAP32[$2+16>>2]=0|0;HEAP32[$2+20>>2]=0|0;
 return;
}
function __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEEC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (18596);
 $1 = ((($0)) + 4|0);
 __ZNSt3__26localeC2Ev($1);
 $2 = ((($0)) + 8|0);
 ;HEAP32[$2>>2]=0|0;HEAP32[$2+4>>2]=0|0;HEAP32[$2+8>>2]=0|0;HEAP32[$2+12>>2]=0|0;HEAP32[$2+16>>2]=0|0;HEAP32[$2+20>>2]=0|0;
 return;
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE5flushEv($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + -12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($0) + ($4)|0);
 $6 = ((($5)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 L1: do {
  if (!($8)) {
   __THREW__ = 0;
   invoke_vii(947,($1|0),($0|0));
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   L3: do {
    if ($10) {
     $35 = ___cxa_find_matching_catch_3(0|0)|0;
     $36 = tempRet0;
     $$0 = $35;
    } else {
     $11 = HEAP8[$1>>0]|0;
     $12 = ($11<<24>>24)==(0);
     do {
      if (!($12)) {
       $13 = HEAP32[$0>>2]|0;
       $14 = ((($13)) + -12|0);
       $15 = HEAP32[$14>>2]|0;
       $16 = (($0) + ($15)|0);
       $17 = ((($16)) + 24|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = HEAP32[$18>>2]|0;
       $20 = ((($19)) + 24|0);
       $21 = HEAP32[$20>>2]|0;
       __THREW__ = 0;
       $22 = (invoke_ii($21|0,($18|0))|0);
       $23 = __THREW__; __THREW__ = 0;
       $24 = $23&1;
       if (!($24)) {
        $25 = ($22|0)==(-1);
        if (!($25)) {
         break;
        }
        $26 = HEAP32[$0>>2]|0;
        $27 = ((($26)) + -12|0);
        $28 = HEAP32[$27>>2]|0;
        $29 = (($0) + ($28)|0);
        $30 = ((($29)) + 16|0);
        $31 = HEAP32[$30>>2]|0;
        $32 = $31 | 1;
        __THREW__ = 0;
        invoke_vii(736,($29|0),($32|0));
        $33 = __THREW__; __THREW__ = 0;
        $34 = $33&1;
        if (!($34)) {
         break;
        }
       }
       $37 = ___cxa_find_matching_catch_3(0|0)|0;
       $38 = tempRet0;
       __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($1);
       $$0 = $37;
       break L3;
      }
     } while(0);
     __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($1);
     break L1;
    }
   } while(0);
   (___cxa_begin_catch(($$0|0))|0);
   $39 = HEAP32[$0>>2]|0;
   $40 = ((($39)) + -12|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (($0) + ($41)|0);
   __THREW__ = 0;
   invoke_vi(737,($42|0));
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if (!($44)) {
    ___cxa_end_catch();
    break;
   }
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if ($48) {
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    ___clang_call_terminate($49);
    // unreachable;
   } else {
    ___resumeException($45|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE6sentryC2ERS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = $1;
 $3 = HEAP32[$1>>2]|0;
 $4 = ((($3)) + -12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($1) + ($5)|0);
 $7 = ((($6)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  $10 = ((($6)) + 72|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if (!($12)) {
   (__ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE5flushEv($11)|0);
  }
  HEAP8[$0>>0] = 1;
 }
 return;
}
function __ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE6sentryD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + -12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($2) + ($5)|0);
 $7 = ((($6)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 do {
  if (!($9)) {
   $10 = ((($6)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0);
   if ($12) {
    $13 = ((($6)) + 4|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = $14 & 8192;
    $16 = ($15|0)==(0);
    if (!($16)) {
     $17 = (__ZSt18uncaught_exceptionv()|0);
     if (!($17)) {
      $18 = HEAP32[$1>>2]|0;
      $19 = HEAP32[$18>>2]|0;
      $20 = ((($19)) + -12|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($18) + ($21)|0);
      $23 = ((($22)) + 24|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = HEAP32[$24>>2]|0;
      $26 = ((($25)) + 24|0);
      $27 = HEAP32[$26>>2]|0;
      __THREW__ = 0;
      $28 = (invoke_ii($27|0,($24|0))|0);
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if (!($30)) {
       $31 = ($28|0)==(-1);
       if (!($31)) {
        break;
       }
       $32 = HEAP32[$1>>2]|0;
       $33 = HEAP32[$32>>2]|0;
       $34 = ((($33)) + -12|0);
       $35 = HEAP32[$34>>2]|0;
       $36 = (($32) + ($35)|0);
       $37 = ((($36)) + 16|0);
       $38 = HEAP32[$37>>2]|0;
       $39 = $38 | 1;
       __THREW__ = 0;
       invoke_vii(736,($36|0),($39|0));
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if (!($41)) {
        break;
       }
      }
      $42 = ___cxa_find_matching_catch_3(0|0)|0;
      $43 = tempRet0;
      (___cxa_begin_catch(($42|0))|0);
      __THREW__ = 0;
      invoke_v(738);
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       $46 = ___cxa_find_matching_catch_3(0|0)|0;
       $47 = tempRet0;
       ___clang_call_terminate($46);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__28ios_base33__set_badbit_and_consider_rethrowEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 16|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 | 1;
 HEAP32[$1>>2] = $3;
 $4 = ((($0)) + 20|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 & 1;
 $7 = ($6|0)==(0);
 if ($7) {
  return;
 } else {
  ___cxa_rethrow();
  // unreachable;
 }
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE5flushEv($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + -12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = (($0) + ($4)|0);
 $6 = ((($5)) + 24|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==(0|0);
 L1: do {
  if (!($8)) {
   __THREW__ = 0;
   invoke_vii(729,($1|0),($0|0));
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   L3: do {
    if ($10) {
     $35 = ___cxa_find_matching_catch_3(0|0)|0;
     $36 = tempRet0;
     $$0 = $35;
    } else {
     $11 = HEAP8[$1>>0]|0;
     $12 = ($11<<24>>24)==(0);
     do {
      if (!($12)) {
       $13 = HEAP32[$0>>2]|0;
       $14 = ((($13)) + -12|0);
       $15 = HEAP32[$14>>2]|0;
       $16 = (($0) + ($15)|0);
       $17 = ((($16)) + 24|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = HEAP32[$18>>2]|0;
       $20 = ((($19)) + 24|0);
       $21 = HEAP32[$20>>2]|0;
       __THREW__ = 0;
       $22 = (invoke_ii($21|0,($18|0))|0);
       $23 = __THREW__; __THREW__ = 0;
       $24 = $23&1;
       if (!($24)) {
        $25 = ($22|0)==(-1);
        if (!($25)) {
         break;
        }
        $26 = HEAP32[$0>>2]|0;
        $27 = ((($26)) + -12|0);
        $28 = HEAP32[$27>>2]|0;
        $29 = (($0) + ($28)|0);
        $30 = ((($29)) + 16|0);
        $31 = HEAP32[$30>>2]|0;
        $32 = $31 | 1;
        __THREW__ = 0;
        invoke_vii(736,($29|0),($32|0));
        $33 = __THREW__; __THREW__ = 0;
        $34 = $33&1;
        if (!($34)) {
         break;
        }
       }
       $37 = ___cxa_find_matching_catch_3(0|0)|0;
       $38 = tempRet0;
       __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($1);
       $$0 = $37;
       break L3;
      }
     } while(0);
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($1);
     break L1;
    }
   } while(0);
   (___cxa_begin_catch(($$0|0))|0);
   $39 = HEAP32[$0>>2]|0;
   $40 = ((($39)) + -12|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = (($0) + ($41)|0);
   __THREW__ = 0;
   invoke_vi(737,($42|0));
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   if (!($44)) {
    ___cxa_end_catch();
    break;
   }
   $45 = ___cxa_find_matching_catch_2()|0;
   $46 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if ($48) {
    $49 = ___cxa_find_matching_catch_3(0|0)|0;
    $50 = tempRet0;
    ___clang_call_terminate($49);
    // unreachable;
   } else {
    ___resumeException($45|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryC2ERS3_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = $1;
 $3 = HEAP32[$1>>2]|0;
 $4 = ((($3)) + -12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($1) + ($5)|0);
 $7 = ((($6)) + 16|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0);
 if ($9) {
  $10 = ((($6)) + 72|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if (!($12)) {
   (__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE5flushEv($11)|0);
  }
  HEAP8[$0>>0] = 1;
 }
 return;
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$2>>2]|0;
 $4 = ((($3)) + -12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (($2) + ($5)|0);
 $7 = ((($6)) + 24|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ($8|0)==(0|0);
 do {
  if (!($9)) {
   $10 = ((($6)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0);
   if ($12) {
    $13 = ((($6)) + 4|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = $14 & 8192;
    $16 = ($15|0)==(0);
    if (!($16)) {
     $17 = (__ZSt18uncaught_exceptionv()|0);
     if (!($17)) {
      $18 = HEAP32[$1>>2]|0;
      $19 = HEAP32[$18>>2]|0;
      $20 = ((($19)) + -12|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = (($18) + ($21)|0);
      $23 = ((($22)) + 24|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = HEAP32[$24>>2]|0;
      $26 = ((($25)) + 24|0);
      $27 = HEAP32[$26>>2]|0;
      __THREW__ = 0;
      $28 = (invoke_ii($27|0,($24|0))|0);
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if (!($30)) {
       $31 = ($28|0)==(-1);
       if (!($31)) {
        break;
       }
       $32 = HEAP32[$1>>2]|0;
       $33 = HEAP32[$32>>2]|0;
       $34 = ((($33)) + -12|0);
       $35 = HEAP32[$34>>2]|0;
       $36 = (($32) + ($35)|0);
       $37 = ((($36)) + 16|0);
       $38 = HEAP32[$37>>2]|0;
       $39 = $38 | 1;
       __THREW__ = 0;
       invoke_vii(736,($36|0),($39|0));
       $40 = __THREW__; __THREW__ = 0;
       $41 = $40&1;
       if (!($41)) {
        break;
       }
      }
      $42 = ___cxa_find_matching_catch_3(0|0)|0;
      $43 = tempRet0;
      (___cxa_begin_catch(($42|0))|0);
      __THREW__ = 0;
      invoke_v(738);
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       $46 = ___cxa_find_matching_catch_3(0|0)|0;
       $47 = tempRet0;
       ___clang_call_terminate($46);
       // unreachable;
      }
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$2 = 0, $$byval_copy = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = sp + 8|0;
 $5 = sp;
 __THREW__ = 0;
 invoke_vii(729,($4|0),($0|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L1: do {
  if ($7) {
   $55 = ___cxa_find_matching_catch_3(0|0)|0;
   $56 = tempRet0;
   $$2 = $55;
   label = 20;
  } else {
   $8 = HEAP8[$4>>0]|0;
   $9 = ($8<<24>>24)==(0);
   L4: do {
    if (!($9)) {
     $10 = HEAP32[$0>>2]|0;
     $11 = ((($10)) + -12|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = (($0) + ($12)|0);
     __ZNKSt3__28ios_base6getlocEv($5,$13);
     __THREW__ = 0;
     $14 = (invoke_iii(734,($5|0),(165504|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     L6: do {
      if ($16) {
       $57 = ___cxa_find_matching_catch_3(0|0)|0;
       $58 = tempRet0;
       __ZNSt3__26localeD2Ev($5);
       $$1 = $57;
      } else {
       __ZNSt3__26localeD2Ev($5);
       $17 = HEAP32[$0>>2]|0;
       $18 = ((($17)) + -12|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = (($0) + ($19)|0);
       $21 = ((($20)) + 24|0);
       $22 = HEAP32[$21>>2]|0;
       $23 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $24 = ((($20)) + 76|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($23,$25)|0);
       do {
        if ($26) {
         __ZNKSt3__28ios_base6getlocEv($3,$20);
         __THREW__ = 0;
         $27 = (invoke_iii(734,($3|0),(165472|0))|0);
         $28 = __THREW__; __THREW__ = 0;
         $29 = $28&1;
         if (!($29)) {
          $30 = HEAP32[$27>>2]|0;
          $31 = ((($30)) + 44|0);
          $32 = HEAP32[$31>>2]|0;
          __THREW__ = 0;
          $33 = (invoke_iii($32|0,($27|0),32)|0);
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if (!($35)) {
           __ZNSt3__26localeD2Ev($3);
           HEAP32[$24>>2] = $33;
           $$pre$phiZ2D = $20;$41 = $33;
           label = 10;
           break;
          }
         }
         $36 = ___cxa_find_matching_catch_3(0|0)|0;
         $37 = tempRet0;
         __ZNSt3__26localeD2Ev($3);
         $eh$lpad$body$indexZ2D = $36;
        } else {
         $$pre$i = HEAP32[$24>>2]|0;
         $$pre$phiZ2D = $20;$41 = $$pre$i;
         label = 10;
        }
       } while(0);
       do {
        if ((label|0) == 10) {
         $38 = HEAP32[$14>>2]|0;
         $39 = ((($38)) + 12|0);
         $40 = HEAP32[$39>>2]|0;
         HEAP32[$2>>2] = $22;
         __THREW__ = 0;
         ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
         $42 = (invoke_iiiiii($40|0,($14|0),($$byval_copy|0),($$pre$phiZ2D|0),($41|0),($1|0))|0);
         $43 = __THREW__; __THREW__ = 0;
         $44 = $43&1;
         if ($44) {
          $59 = ___cxa_find_matching_catch_3(0|0)|0;
          $60 = tempRet0;
          $eh$lpad$body$indexZ2D = $59;
          break;
         }
         $45 = ($42|0)==(0|0);
         if (!($45)) {
          break L4;
         }
         $46 = HEAP32[$0>>2]|0;
         $47 = ((($46)) + -12|0);
         $48 = HEAP32[$47>>2]|0;
         $49 = (($0) + ($48)|0);
         $50 = ((($49)) + 16|0);
         $51 = HEAP32[$50>>2]|0;
         $52 = $51 | 5;
         __THREW__ = 0;
         invoke_vii(736,($49|0),($52|0));
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          break L4;
         }
         $61 = ___cxa_find_matching_catch_3(0|0)|0;
         $62 = tempRet0;
         $$1 = $61;
         break L6;
        }
       } while(0);
       $$1 = $eh$lpad$body$indexZ2D;
      }
     } while(0);
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
     $$2 = $$1;
     label = 20;
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
  }
 } while(0);
 do {
  if ((label|0) == 20) {
   (___cxa_begin_catch(($$2|0))|0);
   $63 = HEAP32[$0>>2]|0;
   $64 = ((($63)) + -12|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = (($0) + ($65)|0);
   __THREW__ = 0;
   invoke_vi(737,($66|0));
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    ___cxa_end_catch();
    break;
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(0|0)|0;
    $74 = tempRet0;
    ___clang_call_terminate($73);
    // unreachable;
   } else {
    ___resumeException($69|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$2 = 0, $$byval_copy = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = sp + 8|0;
 $5 = sp;
 __THREW__ = 0;
 invoke_vii(729,($4|0),($0|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L1: do {
  if ($7) {
   $55 = ___cxa_find_matching_catch_3(0|0)|0;
   $56 = tempRet0;
   $$2 = $55;
   label = 20;
  } else {
   $8 = HEAP8[$4>>0]|0;
   $9 = ($8<<24>>24)==(0);
   L4: do {
    if (!($9)) {
     $10 = HEAP32[$0>>2]|0;
     $11 = ((($10)) + -12|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = (($0) + ($12)|0);
     __ZNKSt3__28ios_base6getlocEv($5,$13);
     __THREW__ = 0;
     $14 = (invoke_iii(734,($5|0),(165504|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     L6: do {
      if ($16) {
       $57 = ___cxa_find_matching_catch_3(0|0)|0;
       $58 = tempRet0;
       __ZNSt3__26localeD2Ev($5);
       $$1 = $57;
      } else {
       __ZNSt3__26localeD2Ev($5);
       $17 = HEAP32[$0>>2]|0;
       $18 = ((($17)) + -12|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = (($0) + ($19)|0);
       $21 = ((($20)) + 24|0);
       $22 = HEAP32[$21>>2]|0;
       $23 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $24 = ((($20)) + 76|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($23,$25)|0);
       do {
        if ($26) {
         __ZNKSt3__28ios_base6getlocEv($3,$20);
         __THREW__ = 0;
         $27 = (invoke_iii(734,($3|0),(165472|0))|0);
         $28 = __THREW__; __THREW__ = 0;
         $29 = $28&1;
         if (!($29)) {
          $30 = HEAP32[$27>>2]|0;
          $31 = ((($30)) + 44|0);
          $32 = HEAP32[$31>>2]|0;
          __THREW__ = 0;
          $33 = (invoke_iii($32|0,($27|0),32)|0);
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if (!($35)) {
           __ZNSt3__26localeD2Ev($3);
           HEAP32[$24>>2] = $33;
           $$pre$phiZ2D = $20;$41 = $33;
           label = 10;
           break;
          }
         }
         $36 = ___cxa_find_matching_catch_3(0|0)|0;
         $37 = tempRet0;
         __ZNSt3__26localeD2Ev($3);
         $eh$lpad$body$indexZ2D = $36;
        } else {
         $$pre$i = HEAP32[$24>>2]|0;
         $$pre$phiZ2D = $20;$41 = $$pre$i;
         label = 10;
        }
       } while(0);
       do {
        if ((label|0) == 10) {
         $38 = HEAP32[$14>>2]|0;
         $39 = ((($38)) + 16|0);
         $40 = HEAP32[$39>>2]|0;
         HEAP32[$2>>2] = $22;
         __THREW__ = 0;
         ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
         $42 = (invoke_iiiiii($40|0,($14|0),($$byval_copy|0),($$pre$phiZ2D|0),($41|0),($1|0))|0);
         $43 = __THREW__; __THREW__ = 0;
         $44 = $43&1;
         if ($44) {
          $59 = ___cxa_find_matching_catch_3(0|0)|0;
          $60 = tempRet0;
          $eh$lpad$body$indexZ2D = $59;
          break;
         }
         $45 = ($42|0)==(0|0);
         if (!($45)) {
          break L4;
         }
         $46 = HEAP32[$0>>2]|0;
         $47 = ((($46)) + -12|0);
         $48 = HEAP32[$47>>2]|0;
         $49 = (($0) + ($48)|0);
         $50 = ((($49)) + 16|0);
         $51 = HEAP32[$50>>2]|0;
         $52 = $51 | 5;
         __THREW__ = 0;
         invoke_vii(736,($49|0),($52|0));
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          break L4;
         }
         $61 = ___cxa_find_matching_catch_3(0|0)|0;
         $62 = tempRet0;
         $$1 = $61;
         break L6;
        }
       } while(0);
       $$1 = $eh$lpad$body$indexZ2D;
      }
     } while(0);
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
     $$2 = $$1;
     label = 20;
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
  }
 } while(0);
 do {
  if ((label|0) == 20) {
   (___cxa_begin_catch(($$2|0))|0);
   $63 = HEAP32[$0>>2]|0;
   $64 = ((($63)) + -12|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = (($0) + ($65)|0);
   __THREW__ = 0;
   invoke_vi(737,($66|0));
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    ___cxa_end_catch();
    break;
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(0|0)|0;
    $74 = tempRet0;
    ___clang_call_terminate($73);
    // unreachable;
   } else {
    ___resumeException($69|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEElsEPKv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$1 = 0, $$2 = 0, $$byval_copy = 0, $$pre$i = 0, $$pre$phiZ2D = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 16|0;
 $4 = sp + 8|0;
 $5 = sp;
 __THREW__ = 0;
 invoke_vii(729,($4|0),($0|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 L1: do {
  if ($7) {
   $55 = ___cxa_find_matching_catch_3(0|0)|0;
   $56 = tempRet0;
   $$2 = $55;
   label = 20;
  } else {
   $8 = HEAP8[$4>>0]|0;
   $9 = ($8<<24>>24)==(0);
   L4: do {
    if (!($9)) {
     $10 = HEAP32[$0>>2]|0;
     $11 = ((($10)) + -12|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = (($0) + ($12)|0);
     __ZNKSt3__28ios_base6getlocEv($5,$13);
     __THREW__ = 0;
     $14 = (invoke_iii(734,($5|0),(165504|0))|0);
     $15 = __THREW__; __THREW__ = 0;
     $16 = $15&1;
     L6: do {
      if ($16) {
       $57 = ___cxa_find_matching_catch_3(0|0)|0;
       $58 = tempRet0;
       __ZNSt3__26localeD2Ev($5);
       $$1 = $57;
      } else {
       __ZNSt3__26localeD2Ev($5);
       $17 = HEAP32[$0>>2]|0;
       $18 = ((($17)) + -12|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = (($0) + ($19)|0);
       $21 = ((($20)) + 24|0);
       $22 = HEAP32[$21>>2]|0;
       $23 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $24 = ((($20)) + 76|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($23,$25)|0);
       do {
        if ($26) {
         __ZNKSt3__28ios_base6getlocEv($3,$20);
         __THREW__ = 0;
         $27 = (invoke_iii(734,($3|0),(165472|0))|0);
         $28 = __THREW__; __THREW__ = 0;
         $29 = $28&1;
         if (!($29)) {
          $30 = HEAP32[$27>>2]|0;
          $31 = ((($30)) + 44|0);
          $32 = HEAP32[$31>>2]|0;
          __THREW__ = 0;
          $33 = (invoke_iii($32|0,($27|0),32)|0);
          $34 = __THREW__; __THREW__ = 0;
          $35 = $34&1;
          if (!($35)) {
           __ZNSt3__26localeD2Ev($3);
           HEAP32[$24>>2] = $33;
           $$pre$phiZ2D = $20;$41 = $33;
           label = 10;
           break;
          }
         }
         $36 = ___cxa_find_matching_catch_3(0|0)|0;
         $37 = tempRet0;
         __ZNSt3__26localeD2Ev($3);
         $eh$lpad$body$indexZ2D = $36;
        } else {
         $$pre$i = HEAP32[$24>>2]|0;
         $$pre$phiZ2D = $20;$41 = $$pre$i;
         label = 10;
        }
       } while(0);
       do {
        if ((label|0) == 10) {
         $38 = HEAP32[$14>>2]|0;
         $39 = ((($38)) + 40|0);
         $40 = HEAP32[$39>>2]|0;
         HEAP32[$2>>2] = $22;
         __THREW__ = 0;
         ;HEAP32[$$byval_copy>>2]=HEAP32[$2>>2]|0;
         $42 = (invoke_iiiiii($40|0,($14|0),($$byval_copy|0),($$pre$phiZ2D|0),($41|0),($1|0))|0);
         $43 = __THREW__; __THREW__ = 0;
         $44 = $43&1;
         if ($44) {
          $59 = ___cxa_find_matching_catch_3(0|0)|0;
          $60 = tempRet0;
          $eh$lpad$body$indexZ2D = $59;
          break;
         }
         $45 = ($42|0)==(0|0);
         if (!($45)) {
          break L4;
         }
         $46 = HEAP32[$0>>2]|0;
         $47 = ((($46)) + -12|0);
         $48 = HEAP32[$47>>2]|0;
         $49 = (($0) + ($48)|0);
         $50 = ((($49)) + 16|0);
         $51 = HEAP32[$50>>2]|0;
         $52 = $51 | 5;
         __THREW__ = 0;
         invoke_vii(736,($49|0),($52|0));
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if (!($54)) {
          break L4;
         }
         $61 = ___cxa_find_matching_catch_3(0|0)|0;
         $62 = tempRet0;
         $$1 = $61;
         break L6;
        }
       } while(0);
       $$1 = $eh$lpad$body$indexZ2D;
      }
     } while(0);
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
     $$2 = $$1;
     label = 20;
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($4);
  }
 } while(0);
 do {
  if ((label|0) == 20) {
   (___cxa_begin_catch(($$2|0))|0);
   $63 = HEAP32[$0>>2]|0;
   $64 = ((($63)) + -12|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = (($0) + ($65)|0);
   __THREW__ = 0;
   invoke_vi(737,($66|0));
   $67 = __THREW__; __THREW__ = 0;
   $68 = $67&1;
   if (!($68)) {
    ___cxa_end_catch();
    break;
   }
   $69 = ___cxa_find_matching_catch_2()|0;
   $70 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(0|0)|0;
    $74 = tempRet0;
    ___clang_call_terminate($73);
    // unreachable;
   } else {
    ___resumeException($69|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE3putEw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 __THREW__ = 0;
 invoke_vii(729,($2|0),($0|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 L1: do {
  if ($4) {
   $41 = ___cxa_find_matching_catch_3(0|0)|0;
   $42 = tempRet0;
   $$0 = $41;
   label = 12;
  } else {
   $5 = HEAP8[$2>>0]|0;
   $6 = ($5<<24>>24)==(0);
   L4: do {
    if (!($6)) {
     $7 = HEAP32[$0>>2]|0;
     $8 = ((($7)) + -12|0);
     $9 = HEAP32[$8>>2]|0;
     $10 = (($0) + ($9)|0);
     $11 = ((($10)) + 24|0);
     $12 = HEAP32[$11>>2]|0;
     $13 = $12;
     $14 = ($12|0)==(0);
     do {
      if ($14) {
       label = 8;
      } else {
       $15 = ((($13)) + 24|0);
       $16 = HEAP32[$15>>2]|0;
       $17 = ((($13)) + 28|0);
       $18 = HEAP32[$17>>2]|0;
       $19 = ($16|0)==($18|0);
       if ($19) {
        $20 = $12;
        $21 = HEAP32[$20>>2]|0;
        $22 = ((($21)) + 52|0);
        $23 = HEAP32[$22>>2]|0;
        $24 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($1)|0);
        __THREW__ = 0;
        $25 = (invoke_iii($23|0,($13|0),($24|0))|0);
        $26 = __THREW__; __THREW__ = 0;
        $27 = $26&1;
        if ($27) {
         break;
        } else {
         $$0$i$i = $25;
        }
       } else {
        $28 = ((($16)) + 4|0);
        HEAP32[$15>>2] = $28;
        HEAP32[$16>>2] = $1;
        $29 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($1)|0);
        $$0$i$i = $29;
       }
       $30 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $31 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i,$30)|0);
       if ($31) {
        label = 8;
       } else {
        break L4;
       }
      }
     } while(0);
     if ((label|0) == 8) {
      $32 = HEAP32[$0>>2]|0;
      $33 = ((($32)) + -12|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = (($0) + ($34)|0);
      $36 = ((($35)) + 16|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = $37 | 1;
      __THREW__ = 0;
      invoke_vii(736,($35|0),($38|0));
      $39 = __THREW__; __THREW__ = 0;
      $40 = $39&1;
      if (!($40)) {
       break;
      }
     }
     $43 = ___cxa_find_matching_catch_3(0|0)|0;
     $44 = tempRet0;
     __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($2);
     $$0 = $43;
     label = 12;
     break L1;
    }
   } while(0);
   __ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryD2Ev($2);
  }
 } while(0);
 do {
  if ((label|0) == 12) {
   (___cxa_begin_catch(($$0|0))|0);
   $45 = HEAP32[$0>>2]|0;
   $46 = ((($45)) + -12|0);
   $47 = HEAP32[$46>>2]|0;
   $48 = (($0) + ($47)|0);
   __THREW__ = 0;
   invoke_vi(737,($48|0));
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if (!($50)) {
    ___cxa_end_catch();
    break;
   }
   $51 = ___cxa_find_matching_catch_2()|0;
   $52 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $53 = __THREW__; __THREW__ = 0;
   $54 = $53&1;
   if ($54) {
    $55 = ___cxa_find_matching_catch_3(0|0)|0;
    $56 = tempRet0;
    ___clang_call_terminate($55);
    // unreachable;
   } else {
    ___resumeException($51|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __GLOBAL__I_000101() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 ___cxx_global_var_init_288();
 return;
}
function __GLOBAL__sub_I_iostream_cpp() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function ___cxx_global_var_init_288() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28ios_base4InitC2Ev(0);
 (___cxa_atexit((948|0),(169456|0),(___dso_handle|0))|0);
 return;
}
function __ZNSt3__28ios_base4InitC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[3750]|0;
 __ZNSt3__210__stdinbufIcEC2EP8_IO_FILEP11__mbstate_t(165072,$1,165128);
 HEAP32[41098] = (18664);
 HEAP32[(164400)>>2] = (18684);
 HEAP32[(164396)>>2] = 0;
 __THREW__ = 0;
 invoke_vii(949,((164400)|0),(165072|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev((164400));
  ___resumeException($4|0);
  // unreachable;
 }
 HEAP32[(164472)>>2] = 0;
 $6 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 HEAP32[(164476)>>2] = $6;
 __ZNSt3__210__stdinbufIwEC2EP8_IO_FILEP11__mbstate_t(165136,$1,165192);
 HEAP32[41120] = (18712);
 HEAP32[(164488)>>2] = (18732);
 HEAP32[(164484)>>2] = 0;
 __THREW__ = 0;
 invoke_vii(949,((164488)|0),(165136|0));
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $9 = ___cxa_find_matching_catch_2()|0;
  $10 = tempRet0;
  __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev((164488));
  ___resumeException($9|0);
  // unreachable;
 }
 HEAP32[(164560)>>2] = 0;
 $11 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 HEAP32[(164564)>>2] = $11;
 $12 = HEAP32[3782]|0;
 __ZNSt3__211__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t(165200,$12,165248);
 HEAP32[41142] = (18760);
 HEAP32[(164572)>>2] = (18780);
 __THREW__ = 0;
 invoke_vii(949,((164572)|0),(165200|0));
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev((164572));
  ___resumeException($15|0);
  // unreachable;
 }
 HEAP32[(164644)>>2] = 0;
 $17 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 HEAP32[(164648)>>2] = $17;
 __ZNSt3__211__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t(165256,$12,165304);
 HEAP32[41163] = (18808);
 HEAP32[(164656)>>2] = (18828);
 __THREW__ = 0;
 invoke_vii(949,((164656)|0),(165256|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $20 = ___cxa_find_matching_catch_2()|0;
  $21 = tempRet0;
  __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev((164656));
  ___resumeException($20|0);
  // unreachable;
 }
 HEAP32[(164728)>>2] = 0;
 $22 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 HEAP32[(164732)>>2] = $22;
 $23 = HEAP32[3657]|0;
 __ZNSt3__211__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t(165312,$23,165360);
 HEAP32[41184] = (18760);
 HEAP32[(164740)>>2] = (18780);
 __THREW__ = 0;
 invoke_vii(949,((164740)|0),(165312|0));
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_2()|0;
  $27 = tempRet0;
  __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev((164740));
  ___resumeException($26|0);
  // unreachable;
 }
 HEAP32[(164812)>>2] = 0;
 $28 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 HEAP32[(164816)>>2] = $28;
 $29 = HEAP32[41184]|0;
 $30 = ((($29)) + -12|0);
 $31 = HEAP32[$30>>2]|0;
 $32 = (164736 + ($31)|0);
 $33 = ((($32)) + 24|0);
 $34 = HEAP32[$33>>2]|0;
 HEAP32[41226] = (18760);
 HEAP32[(164908)>>2] = (18780);
 __THREW__ = 0;
 invoke_vii(949,((164908)|0),($34|0));
 $35 = __THREW__; __THREW__ = 0;
 $36 = $35&1;
 if ($36) {
  $37 = ___cxa_find_matching_catch_2()|0;
  $38 = tempRet0;
  __ZNSt3__29basic_iosIcNS_11char_traitsIcEEED2Ev((164908));
  ___resumeException($37|0);
  // unreachable;
 }
 HEAP32[(164980)>>2] = 0;
 $39 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 HEAP32[(164984)>>2] = $39;
 __ZNSt3__211__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t(165368,$23,165416);
 HEAP32[41205] = (18808);
 HEAP32[(164824)>>2] = (18828);
 __THREW__ = 0;
 invoke_vii(949,((164824)|0),(165368|0));
 $40 = __THREW__; __THREW__ = 0;
 $41 = $40&1;
 if ($41) {
  $42 = ___cxa_find_matching_catch_2()|0;
  $43 = tempRet0;
  __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev((164824));
  ___resumeException($42|0);
  // unreachable;
 }
 HEAP32[(164896)>>2] = 0;
 $44 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 HEAP32[(164900)>>2] = $44;
 $45 = HEAP32[41205]|0;
 $46 = ((($45)) + -12|0);
 $47 = HEAP32[$46>>2]|0;
 $48 = (164820 + ($47)|0);
 $49 = ((($48)) + 24|0);
 $50 = HEAP32[$49>>2]|0;
 HEAP32[41247] = (18808);
 HEAP32[(164992)>>2] = (18828);
 __THREW__ = 0;
 invoke_vii(949,((164992)|0),($50|0));
 $51 = __THREW__; __THREW__ = 0;
 $52 = $51&1;
 if ($52) {
  $53 = ___cxa_find_matching_catch_2()|0;
  $54 = tempRet0;
  __ZNSt3__29basic_iosIwNS_11char_traitsIwEEED2Ev((164992));
  ___resumeException($53|0);
  // unreachable;
 } else {
  HEAP32[(165064)>>2] = 0;
  $55 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
  HEAP32[(165068)>>2] = $55;
  $56 = HEAP32[41098]|0;
  $57 = ((($56)) + -12|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = (164392 + ($58)|0);
  $60 = ((($59)) + 72|0);
  HEAP32[$60>>2] = 164568;
  $61 = HEAP32[41120]|0;
  $62 = ((($61)) + -12|0);
  $63 = HEAP32[$62>>2]|0;
  $64 = (164480 + ($63)|0);
  $65 = ((($64)) + 72|0);
  HEAP32[$65>>2] = 164652;
  $66 = HEAP32[41184]|0;
  $67 = ((($66)) + -12|0);
  $68 = HEAP32[$67>>2]|0;
  $69 = (164736 + ($68)|0);
  $70 = ((($69)) + 4|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = $71 | 8192;
  HEAP32[$70>>2] = $72;
  $73 = HEAP32[41205]|0;
  $74 = ((($73)) + -12|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = (164820 + ($75)|0);
  $77 = ((($76)) + 4|0);
  $78 = HEAP32[$77>>2]|0;
  $79 = $78 | 8192;
  HEAP32[$77>>2] = $79;
  $80 = HEAP32[41184]|0;
  $81 = ((($80)) + -12|0);
  $82 = HEAP32[$81>>2]|0;
  $83 = (164736 + ($82)|0);
  $84 = ((($83)) + 72|0);
  HEAP32[$84>>2] = 164568;
  $85 = HEAP32[41205]|0;
  $86 = ((($85)) + -12|0);
  $87 = HEAP32[$86>>2]|0;
  $88 = (164820 + ($87)|0);
  $89 = ((($88)) + 72|0);
  HEAP32[$89>>2] = 164652;
  return;
 }
}
function __ZNSt3__28ios_base4InitD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 (invoke_ii(950,(164568|0))|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  __THREW__ = 0;
  (invoke_ii(951,(164652|0))|0);
  $3 = __THREW__; __THREW__ = 0;
  $4 = $3&1;
  if (!($4)) {
   __THREW__ = 0;
   (invoke_ii(950,(164904|0))|0);
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   if (!($6)) {
    __THREW__ = 0;
    (invoke_ii(951,(164988|0))|0);
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if (!($8)) {
     return;
    }
   }
  }
 }
 $9 = ___cxa_find_matching_catch_3(0|0)|0;
 $10 = tempRet0;
 ___clang_call_terminate($9);
 // unreachable;
}
function __ZNSt3__210__stdinbufIcEC2EP8_IO_FILEP11__mbstate_t($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEEC2Ev($0);
 HEAP32[$0>>2] = (19104);
 $4 = ((($0)) + 32|0);
 HEAP32[$4>>2] = $1;
 $5 = ((($0)) + 40|0);
 HEAP32[$5>>2] = $2;
 $6 = ((($0)) + 48|0);
 $7 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 HEAP32[$6>>2] = $7;
 $8 = ((($0)) + 52|0);
 HEAP8[$8>>0] = 0;
 $9 = HEAP32[$0>>2]|0;
 $10 = ((($9)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($0)) + 4|0);
 __ZNSt3__26localeC2ERKS0_($3,$12);
 __THREW__ = 0;
 invoke_vii($11|0,($0|0),($3|0));
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZNSt3__26localeD2Ev($3);
  __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0);
  ___resumeException($15|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZNSt3__210__stdinbufIwEC2EP8_IO_FILEP11__mbstate_t($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEEC2Ev($0);
 HEAP32[$0>>2] = (19040);
 $4 = ((($0)) + 32|0);
 HEAP32[$4>>2] = $1;
 $5 = ((($0)) + 40|0);
 HEAP32[$5>>2] = $2;
 $6 = ((($0)) + 48|0);
 $7 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 HEAP32[$6>>2] = $7;
 $8 = ((($0)) + 52|0);
 HEAP8[$8>>0] = 0;
 $9 = HEAP32[$0>>2]|0;
 $10 = ((($9)) + 8|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = ((($0)) + 4|0);
 __ZNSt3__26localeC2ERKS0_($3,$12);
 __THREW__ = 0;
 invoke_vii($11|0,($0|0),($3|0));
 $13 = __THREW__; __THREW__ = 0;
 $14 = $13&1;
 if ($14) {
  $15 = ___cxa_find_matching_catch_2()|0;
  $16 = tempRet0;
  __ZNSt3__26localeD2Ev($3);
  __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0);
  ___resumeException($15|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZNSt3__211__stdoutbufIcEC2EP8_IO_FILEP11__mbstate_t($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEEC2Ev($0);
 HEAP32[$0>>2] = (18976);
 $4 = ((($0)) + 32|0);
 HEAP32[$4>>2] = $1;
 $5 = ((($0)) + 4|0);
 __ZNSt3__26localeC2ERKS0_($3,$5);
 __THREW__ = 0;
 $6 = (invoke_iii(734,($3|0),(167232|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZNSt3__26localeD2Ev($3);
  __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0);
  ___resumeException($17|0);
  // unreachable;
 } else {
  $9 = ((($0)) + 36|0);
  __ZNSt3__26localeD2Ev($3);
  HEAP32[$9>>2] = $6;
  $10 = ((($0)) + 40|0);
  HEAP32[$10>>2] = $2;
  $11 = ((($0)) + 44|0);
  $12 = HEAP32[$6>>2]|0;
  $13 = ((($12)) + 28|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (FUNCTION_TABLE_ii[$14 & 2047]($6)|0);
  $16 = $15&1;
  HEAP8[$11>>0] = $16;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__211__stdoutbufIwEC2EP8_IO_FILEP11__mbstate_t($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEEC2Ev($0);
 HEAP32[$0>>2] = (18912);
 $4 = ((($0)) + 32|0);
 HEAP32[$4>>2] = $1;
 $5 = ((($0)) + 4|0);
 __ZNSt3__26localeC2ERKS0_($3,$5);
 __THREW__ = 0;
 $6 = (invoke_iii(734,($3|0),(167240|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if ($8) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  __ZNSt3__26localeD2Ev($3);
  __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0);
  ___resumeException($17|0);
  // unreachable;
 } else {
  $9 = ((($0)) + 36|0);
  __ZNSt3__26localeD2Ev($3);
  HEAP32[$9>>2] = $6;
  $10 = ((($0)) + 40|0);
  HEAP32[$10>>2] = $2;
  $11 = ((($0)) + 44|0);
  $12 = HEAP32[$6>>2]|0;
  $13 = ((($12)) + 28|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = (FUNCTION_TABLE_ii[$14 & 2047]($6)|0);
  $16 = $15&1;
  HEAP8[$11>>0] = $16;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__211__stdoutbufIwED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0);
 return;
}
function __ZNSt3__211__stdoutbufIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__211__stdoutbufIwED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__211__stdoutbufIwE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 (FUNCTION_TABLE_ii[$4 & 2047]($0)|0);
 $5 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167240)|0);
 $6 = ((($0)) + 36|0);
 HEAP32[$6>>2] = $5;
 $7 = HEAP32[$5>>2]|0;
 $8 = ((($7)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_ii[$9 & 2047]($5)|0);
 $11 = ((($0)) + 44|0);
 $12 = $10&1;
 HEAP8[$11>>0] = $12;
 return;
}
function __ZNSt3__211__stdoutbufIwE4syncEv($0) {
 $0 = $0|0;
 var $$10 = 0, $$2 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($0)) + 36|0);
 $4 = ((($0)) + 40|0);
 $5 = ((($1)) + 8|0);
 $6 = $1;
 $7 = ((($0)) + 32|0);
 L1: while(1) {
  $8 = HEAP32[$3>>2]|0;
  $9 = HEAP32[$4>>2]|0;
  $10 = HEAP32[$8>>2]|0;
  $11 = ((($10)) + 20|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (FUNCTION_TABLE_iiiiii[$12 & 2047]($8,$9,$1,$5,$2)|0);
  $14 = HEAP32[$2>>2]|0;
  $15 = (($14) - ($6))|0;
  $16 = HEAP32[$7>>2]|0;
  $17 = (_fwrite($1,1,$15,$16)|0);
  $18 = ($17|0)==($15|0);
  if (!($18)) {
   $$2 = -1;
   break;
  }
  switch ($13|0) {
  case 1:  {
   break;
  }
  case 2:  {
   $$2 = -1;
   break L1;
   break;
  }
  default: {
   label = 4;
   break L1;
  }
  }
 }
 if ((label|0) == 4) {
  $19 = HEAP32[$7>>2]|0;
  $20 = (_fflush($19)|0);
  $not$ = ($20|0)!=(0);
  $$10 = $not$ << 31 >> 31;
  $$2 = $$10;
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNSt3__211__stdoutbufIwE6xsputnEPKwi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$010 = 0, $$011 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 44|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 L1: do {
  if ($5) {
   $$0 = 0;$$011 = $1;
   while(1) {
    $9 = ($$0|0)<($2|0);
    if (!($9)) {
     $$010 = $$0;
     break L1;
    }
    $10 = HEAP32[$0>>2]|0;
    $11 = ((($10)) + 52|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = HEAP32[$$011>>2]|0;
    $14 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($13)|0);
    $15 = (FUNCTION_TABLE_iii[$12 & 2047]($0,$14)|0);
    $16 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $17 = ($15|0)==($16|0);
    if ($17) {
     $$010 = $$0;
     break L1;
    }
    $18 = (($$0) + 1)|0;
    $19 = ((($$011)) + 4|0);
    $$0 = $18;$$011 = $19;
   }
  } else {
   $6 = ((($0)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (_fwrite($1,4,$2,$7)|0);
   $$010 = $8;
  }
 } while(0);
 return ($$010|0);
}
function __ZNSt3__211__stdoutbufIwE8overflowEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$$032 = 0, $$032 = 0, $$4 = 0, $$436 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $7 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($1,$6)|0);
 do {
  if ($7) {
   label = 15;
  } else {
   $8 = (__ZNSt3__211char_traitsIwE12to_char_typeEj($1)|0);
   HEAP32[$3>>2] = $8;
   $9 = ((($0)) + 44|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)==(0);
   $12 = ((($0)) + 32|0);
   if (!($11)) {
    $13 = HEAP32[$12>>2]|0;
    $14 = (_fwrite($3,4,1,$13)|0);
    $15 = ($14|0)==(1);
    if ($15) {
     label = 15;
     break;
    }
    $16 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $$4 = $16;
    break;
   }
   HEAP32[$4>>2] = $2;
   $17 = ((($3)) + 4|0);
   $18 = ((($0)) + 36|0);
   $19 = ((($0)) + 40|0);
   $20 = ((($2)) + 8|0);
   $21 = $2;
   $$032 = $3;
   while(1) {
    $22 = HEAP32[$18>>2]|0;
    $23 = HEAP32[$19>>2]|0;
    $24 = HEAP32[$22>>2]|0;
    $25 = ((($24)) + 12|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (FUNCTION_TABLE_iiiiiiiii[$26 & 1023]($22,$23,$$032,$17,$5,$2,$20,$4)|0);
    $28 = HEAP32[$5>>2]|0;
    $29 = ($28|0)==($$032|0);
    $30 = ($27|0)==(1);
    if ($29) {
     label = 14;
     break;
    }
    $31 = ($27|0)==(3);
    if ($31) {
     $32 = HEAP32[$12>>2]|0;
     $33 = (_fwrite($$032,1,1,$32)|0);
     $34 = ($33|0)==(1);
     if ($34) {
      $$436 = $$032;
     } else {
      label = 14;
      break;
     }
    } else {
     $35 = $27 | 1;
     $36 = ($35|0)==(1);
     if (!($36)) {
      label = 14;
      break;
     }
     $37 = HEAP32[$4>>2]|0;
     $38 = (($37) - ($21))|0;
     $39 = HEAP32[$12>>2]|0;
     $40 = (_fwrite($2,1,$38,$39)|0);
     $41 = ($40|0)==($38|0);
     if (!($41)) {
      label = 14;
      break;
     }
     $42 = HEAP32[$5>>2]|0;
     $$$032 = $30 ? $42 : $$032;
     $$436 = $$$032;
    }
    if ($30) {
     $$032 = $$436;
    } else {
     label = 13;
     break;
    }
   }
   if ((label|0) == 13) {
    label = 15;
    break;
   }
   else if ((label|0) == 14) {
    $43 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $$4 = $43;
    break;
   }
  }
 } while(0);
 if ((label|0) == 15) {
  $44 = (__ZNSt3__211char_traitsIwE7not_eofEj($1)|0);
  $$4 = $44;
 }
 STACKTOP = sp;return ($$4|0);
}
function __ZNSt3__211char_traitsIwE7not_eofEj($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $2 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($0,$1)|0);
 if ($2) {
  $3 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
  $4 = $3 ^ -1;
  $5 = $4;
 } else {
  $5 = $0;
 }
 return ($5|0);
}
function __ZNSt3__211__stdoutbufIcED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0);
 return;
}
function __ZNSt3__211__stdoutbufIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__211__stdoutbufIcED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__211__stdoutbufIcE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($2)) + 24|0);
 $4 = HEAP32[$3>>2]|0;
 (FUNCTION_TABLE_ii[$4 & 2047]($0)|0);
 $5 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167232)|0);
 $6 = ((($0)) + 36|0);
 HEAP32[$6>>2] = $5;
 $7 = HEAP32[$5>>2]|0;
 $8 = ((($7)) + 28|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_ii[$9 & 2047]($5)|0);
 $11 = ((($0)) + 44|0);
 $12 = $10&1;
 HEAP8[$11>>0] = $12;
 return;
}
function __ZNSt3__211__stdoutbufIcE4syncEv($0) {
 $0 = $0|0;
 var $$10 = 0, $$2 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp + 8|0;
 $2 = sp;
 $3 = ((($0)) + 36|0);
 $4 = ((($0)) + 40|0);
 $5 = ((($1)) + 8|0);
 $6 = $1;
 $7 = ((($0)) + 32|0);
 L1: while(1) {
  $8 = HEAP32[$3>>2]|0;
  $9 = HEAP32[$4>>2]|0;
  $10 = HEAP32[$8>>2]|0;
  $11 = ((($10)) + 20|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = (FUNCTION_TABLE_iiiiii[$12 & 2047]($8,$9,$1,$5,$2)|0);
  $14 = HEAP32[$2>>2]|0;
  $15 = (($14) - ($6))|0;
  $16 = HEAP32[$7>>2]|0;
  $17 = (_fwrite($1,1,$15,$16)|0);
  $18 = ($17|0)==($15|0);
  if (!($18)) {
   $$2 = -1;
   break;
  }
  switch ($13|0) {
  case 1:  {
   break;
  }
  case 2:  {
   $$2 = -1;
   break L1;
   break;
  }
  default: {
   label = 4;
   break L1;
  }
  }
 }
 if ((label|0) == 4) {
  $19 = HEAP32[$7>>2]|0;
  $20 = (_fflush($19)|0);
  $not$ = ($20|0)!=(0);
  $$10 = $not$ << 31 >> 31;
  $$2 = $$10;
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNSt3__211__stdoutbufIcE6xsputnEPKci($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$010 = 0, $$011 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 44|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)==(0);
 L1: do {
  if ($5) {
   $$0 = 0;$$011 = $1;
   while(1) {
    $9 = ($$0|0)<($2|0);
    if (!($9)) {
     $$010 = $$0;
     break L1;
    }
    $10 = HEAP32[$0>>2]|0;
    $11 = ((($10)) + 52|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = HEAP8[$$011>>0]|0;
    $14 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($13)|0);
    $15 = (FUNCTION_TABLE_iii[$12 & 2047]($0,$14)|0);
    $16 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $17 = ($15|0)==($16|0);
    if ($17) {
     $$010 = $$0;
     break L1;
    }
    $18 = (($$0) + 1)|0;
    $19 = ((($$011)) + 1|0);
    $$0 = $18;$$011 = $19;
   }
  } else {
   $6 = ((($0)) + 32|0);
   $7 = HEAP32[$6>>2]|0;
   $8 = (_fwrite($1,1,$2,$7)|0);
   $$010 = $8;
  }
 } while(0);
 return ($$010|0);
}
function __ZNSt3__211__stdoutbufIcE8overflowEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$$032 = 0, $$032 = 0, $$4 = 0, $$436 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $7 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($1,$6)|0);
 do {
  if ($7) {
   label = 15;
  } else {
   $8 = (__ZNSt3__211char_traitsIcE12to_char_typeEi($1)|0);
   HEAP8[$3>>0] = $8;
   $9 = ((($0)) + 44|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)==(0);
   $12 = ((($0)) + 32|0);
   if (!($11)) {
    $13 = HEAP32[$12>>2]|0;
    $14 = (_fwrite($3,1,1,$13)|0);
    $15 = ($14|0)==(1);
    if ($15) {
     label = 15;
     break;
    }
    $16 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $$4 = $16;
    break;
   }
   HEAP32[$4>>2] = $2;
   $17 = ((($3)) + 1|0);
   $18 = ((($0)) + 36|0);
   $19 = ((($0)) + 40|0);
   $20 = ((($2)) + 8|0);
   $21 = $2;
   $$032 = $3;
   while(1) {
    $22 = HEAP32[$18>>2]|0;
    $23 = HEAP32[$19>>2]|0;
    $24 = HEAP32[$22>>2]|0;
    $25 = ((($24)) + 12|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = (FUNCTION_TABLE_iiiiiiiii[$26 & 1023]($22,$23,$$032,$17,$5,$2,$20,$4)|0);
    $28 = HEAP32[$5>>2]|0;
    $29 = ($28|0)==($$032|0);
    $30 = ($27|0)==(1);
    if ($29) {
     label = 14;
     break;
    }
    $31 = ($27|0)==(3);
    if ($31) {
     $32 = HEAP32[$12>>2]|0;
     $33 = (_fwrite($$032,1,1,$32)|0);
     $34 = ($33|0)==(1);
     if ($34) {
      $$436 = $$032;
     } else {
      label = 14;
      break;
     }
    } else {
     $35 = $27 | 1;
     $36 = ($35|0)==(1);
     if (!($36)) {
      label = 14;
      break;
     }
     $37 = HEAP32[$4>>2]|0;
     $38 = (($37) - ($21))|0;
     $39 = HEAP32[$12>>2]|0;
     $40 = (_fwrite($2,1,$38,$39)|0);
     $41 = ($40|0)==($38|0);
     if (!($41)) {
      label = 14;
      break;
     }
     $42 = HEAP32[$5>>2]|0;
     $$$032 = $30 ? $42 : $$032;
     $$436 = $$$032;
    }
    if ($30) {
     $$032 = $$436;
    } else {
     label = 13;
     break;
    }
   }
   if ((label|0) == 13) {
    label = 15;
    break;
   }
   else if ((label|0) == 14) {
    $43 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $$4 = $43;
    break;
   }
  }
 } while(0);
 if ((label|0) == 15) {
  $44 = (__ZNSt3__211char_traitsIcE7not_eofEi($1)|0);
  $$4 = $44;
 }
 STACKTOP = sp;return ($$4|0);
}
function __ZNSt3__211char_traitsIcE7not_eofEi($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $2 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($0,$1)|0);
 if ($2) {
  $3 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
  $4 = $3 ^ -1;
  $5 = $4;
 } else {
  $5 = $0;
 }
 return ($5|0);
}
function __ZNSt3__210__stdinbufIwED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev($0);
 return;
}
function __ZNSt3__210__stdinbufIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__210__stdinbufIwED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__210__stdinbufIwE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167240)|0);
 $3 = ((($0)) + 36|0);
 HEAP32[$3>>2] = $2;
 $4 = HEAP32[$2>>2]|0;
 $5 = ((($4)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_ii[$6 & 2047]($2)|0);
 $8 = ((($0)) + 44|0);
 HEAP32[$8>>2] = $7;
 $9 = HEAP32[$3>>2]|0;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 28|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_ii[$12 & 2047]($9)|0);
 $14 = ((($0)) + 53|0);
 $15 = $13&1;
 HEAP8[$14>>0] = $15;
 $16 = HEAP32[$8>>2]|0;
 $17 = ($16|0)>(8);
 if ($17) {
  __ZNSt3__221__throw_runtime_errorEPKc(155676);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt3__210__stdinbufIwE9underflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__210__stdinbufIwE9__getcharEb($0,0)|0);
 return ($1|0);
}
function __ZNSt3__210__stdinbufIwE5uflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__210__stdinbufIwE9__getcharEb($0,1)|0);
 return ($1|0);
}
function __ZNSt3__210__stdinbufIwE9pbackfailEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$07 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
 $7 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($1,$6)|0);
 $8 = ((($0)) + 52|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)!=(0);
 $11 = ((($0)) + 48|0);
 do {
  if ($7) {
   if ($10) {
    $$1 = $1;
   } else {
    $12 = HEAP32[$11>>2]|0;
    $13 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $14 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($12,$13)|0);
    $15 = $14 ^ 1;
    $16 = $15&1;
    HEAP8[$8>>0] = $16;
    $$1 = $12;
   }
  } else {
   if ($10) {
    $17 = HEAP32[$11>>2]|0;
    $18 = (__ZNSt3__211char_traitsIwE12to_char_typeEj($17)|0);
    HEAP32[$4>>2] = $18;
    $19 = ((($0)) + 36|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ((($0)) + 40|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($4)) + 4|0);
    $24 = ((($2)) + 8|0);
    $25 = HEAP32[$20>>2]|0;
    $26 = ((($25)) + 12|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (FUNCTION_TABLE_iiiiiiiii[$27 & 1023]($20,$22,$4,$23,$5,$2,$24,$3)|0);
    switch ($28|0) {
    case 1: case 2:  {
     label = 10;
     break;
    }
    case 3:  {
     $29 = HEAP32[$11>>2]|0;
     $30 = $29&255;
     HEAP8[$2>>0] = $30;
     $31 = ((($2)) + 1|0);
     HEAP32[$3>>2] = $31;
     label = 7;
     break;
    }
    default: {
     label = 7;
    }
    }
    L9: do {
     if ((label|0) == 7) {
      $32 = ((($0)) + 32|0);
      while(1) {
       $33 = HEAP32[$3>>2]|0;
       $34 = ($33>>>0)>($2>>>0);
       if (!($34)) {
        $$0 = 1;$$07 = 0;
        break L9;
       }
       $35 = ((($33)) + -1|0);
       HEAP32[$3>>2] = $35;
       $36 = HEAP8[$35>>0]|0;
       $37 = $36 << 24 >> 24;
       $38 = HEAP32[$32>>2]|0;
       $39 = (_ungetc($37,$38)|0);
       $40 = ($39|0)==(-1);
       if ($40) {
        label = 10;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 10) {
     $41 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $$0 = 0;$$07 = $41;
    }
    if (!($$0)) {
     $$1 = $$07;
     break;
    }
   }
   HEAP32[$11>>2] = $1;
   HEAP8[$8>>0] = 1;
   $$1 = $1;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__210__stdinbufIwE9__getcharEb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$048 = 0, $$052 = 0, $$10 = 0, $$11 = 0, $$351 = 0, $$9 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = ((($0)) + 52|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)==(0);
 $9 = ((($0)) + 48|0);
 if ($8) {
  $12 = ((($0)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)>(1);
  $$sroa$speculated = $14 ? $13 : 1;
  $15 = ((($0)) + 32|0);
  $$052 = 0;
  while(1) {
   $16 = ($$052|0)<($$sroa$speculated|0);
   if (!($16)) {
    label = 9;
    break;
   }
   $17 = HEAP32[$15>>2]|0;
   $18 = (_getc($17)|0);
   $19 = ($18|0)==(-1);
   if ($19) {
    label = 8;
    break;
   }
   $20 = $18&255;
   $21 = (($2) + ($$052)|0);
   HEAP8[$21>>0] = $20;
   $22 = (($$052) + 1)|0;
   $$052 = $22;
  }
  if ((label|0) == 8) {
   $23 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $$10 = $23;
  }
  else if ((label|0) == 9) {
   $24 = ((($0)) + 53|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = ($25<<24>>24)==(0);
   do {
    if ($26) {
     $29 = ((($0)) + 40|0);
     $30 = ((($0)) + 36|0);
     $31 = ((($3)) + 4|0);
     $$048 = $$sroa$speculated;
     L11: while(1) {
      $32 = HEAP32[$29>>2]|0;
      $33 = $32;
      $34 = $33;
      $35 = HEAP32[$34>>2]|0;
      $36 = (($33) + 4)|0;
      $37 = $36;
      $38 = HEAP32[$37>>2]|0;
      $39 = HEAP32[$30>>2]|0;
      $40 = (($2) + ($$048)|0);
      $41 = HEAP32[$39>>2]|0;
      $42 = ((($41)) + 16|0);
      $43 = HEAP32[$42>>2]|0;
      $44 = (FUNCTION_TABLE_iiiiiiiii[$43 & 1023]($39,$32,$2,$40,$4,$3,$31,$5)|0);
      switch ($44|0) {
      case 3:  {
       label = 16;
       break L11;
       break;
      }
      case 2:  {
       label = 18;
       break L11;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       break L11;
      }
      }
      $45 = HEAP32[$29>>2]|0;
      $46 = $45;
      $47 = $46;
      HEAP32[$47>>2] = $35;
      $48 = (($46) + 4)|0;
      $49 = $48;
      HEAP32[$49>>2] = $38;
      $50 = ($$048|0)==(8);
      if ($50) {
       label = 18;
       break;
      }
      $51 = HEAP32[$15>>2]|0;
      $52 = (_getc($51)|0);
      $53 = ($52|0)==(-1);
      if ($53) {
       label = 18;
       break;
      }
      $54 = $52&255;
      HEAP8[$40>>0] = $54;
      $55 = (($$048) + 1)|0;
      $$048 = $55;
     }
     if ((label|0) == 16) {
      $56 = HEAP8[$2>>0]|0;
      $57 = $56 << 24 >> 24;
      HEAP32[$3>>2] = $57;
     }
     else if ((label|0) == 18) {
      $58 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $$9 = $58;
      break;
     }
     $$351 = $$048;
     label = 19;
    } else {
     $27 = HEAP8[$2>>0]|0;
     $28 = $27 << 24 >> 24;
     HEAP32[$3>>2] = $28;
     $$351 = $$sroa$speculated;
     label = 19;
    }
   } while(0);
   L21: do {
    if ((label|0) == 19) {
     L23: do {
      if ($1) {
       $69 = HEAP32[$3>>2]|0;
       $70 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($69)|0);
       HEAP32[$9>>2] = $70;
      } else {
       $$0 = $$351;
       while(1) {
        $59 = ($$0|0)>(0);
        if (!($59)) {
         break L23;
        }
        $60 = (($$0) + -1)|0;
        $61 = (($2) + ($60)|0);
        $62 = HEAP8[$61>>0]|0;
        $63 = $62 << 24 >> 24;
        $64 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($63)|0);
        $65 = HEAP32[$15>>2]|0;
        $66 = (_ungetc($64,$65)|0);
        $67 = ($66|0)==(-1);
        if ($67) {
         break;
        } else {
         $$0 = $60;
        }
       }
       $68 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $$9 = $68;
       break L21;
      }
     } while(0);
     $71 = HEAP32[$3>>2]|0;
     $72 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($71)|0);
     $$9 = $72;
    }
   } while(0);
   $$10 = $$9;
  }
  $$11 = $$10;
 } else {
  $10 = HEAP32[$9>>2]|0;
  if ($1) {
   $11 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   HEAP32[$9>>2] = $11;
   HEAP8[$6>>0] = 0;
   $$11 = $10;
  } else {
   $$11 = $10;
  }
 }
 STACKTOP = sp;return ($$11|0);
}
function __ZNSt3__210__stdinbufIcED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev($0);
 return;
}
function __ZNSt3__210__stdinbufIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__210__stdinbufIcED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__210__stdinbufIcE5imbueERKNS_6localeE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167232)|0);
 $3 = ((($0)) + 36|0);
 HEAP32[$3>>2] = $2;
 $4 = HEAP32[$2>>2]|0;
 $5 = ((($4)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (FUNCTION_TABLE_ii[$6 & 2047]($2)|0);
 $8 = ((($0)) + 44|0);
 HEAP32[$8>>2] = $7;
 $9 = HEAP32[$3>>2]|0;
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 28|0);
 $12 = HEAP32[$11>>2]|0;
 $13 = (FUNCTION_TABLE_ii[$12 & 2047]($9)|0);
 $14 = ((($0)) + 53|0);
 $15 = $13&1;
 HEAP8[$14>>0] = $15;
 $16 = HEAP32[$8>>2]|0;
 $17 = ($16|0)>(8);
 if ($17) {
  __ZNSt3__221__throw_runtime_errorEPKc(155676);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt3__210__stdinbufIcE9underflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__210__stdinbufIcE9__getcharEb($0,0)|0);
 return ($1|0);
}
function __ZNSt3__210__stdinbufIcE5uflowEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__210__stdinbufIcE9__getcharEb($0,1)|0);
 return ($1|0);
}
function __ZNSt3__210__stdinbufIcE9pbackfailEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$07 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 4|0;
 $4 = sp + 8|0;
 $5 = sp;
 $6 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
 $7 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($1,$6)|0);
 $8 = ((($0)) + 52|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)!=(0);
 $11 = ((($0)) + 48|0);
 do {
  if ($7) {
   if ($10) {
    $$1 = $1;
   } else {
    $12 = HEAP32[$11>>2]|0;
    $13 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $14 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($12,$13)|0);
    $15 = $14 ^ 1;
    $16 = $15&1;
    HEAP8[$8>>0] = $16;
    $$1 = $12;
   }
  } else {
   if ($10) {
    $17 = HEAP32[$11>>2]|0;
    $18 = (__ZNSt3__211char_traitsIcE12to_char_typeEi($17)|0);
    HEAP8[$4>>0] = $18;
    $19 = ((($0)) + 36|0);
    $20 = HEAP32[$19>>2]|0;
    $21 = ((($0)) + 40|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($4)) + 1|0);
    $24 = ((($2)) + 8|0);
    $25 = HEAP32[$20>>2]|0;
    $26 = ((($25)) + 12|0);
    $27 = HEAP32[$26>>2]|0;
    $28 = (FUNCTION_TABLE_iiiiiiiii[$27 & 1023]($20,$22,$4,$23,$5,$2,$24,$3)|0);
    switch ($28|0) {
    case 1: case 2:  {
     label = 10;
     break;
    }
    case 3:  {
     $29 = HEAP32[$11>>2]|0;
     $30 = $29&255;
     HEAP8[$2>>0] = $30;
     $31 = ((($2)) + 1|0);
     HEAP32[$3>>2] = $31;
     label = 7;
     break;
    }
    default: {
     label = 7;
    }
    }
    L9: do {
     if ((label|0) == 7) {
      $32 = ((($0)) + 32|0);
      while(1) {
       $33 = HEAP32[$3>>2]|0;
       $34 = ($33>>>0)>($2>>>0);
       if (!($34)) {
        $$0 = 1;$$07 = 0;
        break L9;
       }
       $35 = ((($33)) + -1|0);
       HEAP32[$3>>2] = $35;
       $36 = HEAP8[$35>>0]|0;
       $37 = $36 << 24 >> 24;
       $38 = HEAP32[$32>>2]|0;
       $39 = (_ungetc($37,$38)|0);
       $40 = ($39|0)==(-1);
       if ($40) {
        label = 10;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 10) {
     $41 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $$0 = 0;$$07 = $41;
    }
    if (!($$0)) {
     $$1 = $$07;
     break;
    }
   }
   HEAP32[$11>>2] = $1;
   HEAP8[$8>>0] = 1;
   $$1 = $1;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__210__stdinbufIcE9__getcharEb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$048 = 0, $$052 = 0, $$10 = 0, $$11 = 0, $$351 = 0, $$9 = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 16|0;
 $3 = sp + 8|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = ((($0)) + 52|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)==(0);
 $9 = ((($0)) + 48|0);
 if ($8) {
  $12 = ((($0)) + 44|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)>(1);
  $$sroa$speculated = $14 ? $13 : 1;
  $15 = ((($0)) + 32|0);
  $$052 = 0;
  while(1) {
   $16 = ($$052|0)<($$sroa$speculated|0);
   if (!($16)) {
    label = 9;
    break;
   }
   $17 = HEAP32[$15>>2]|0;
   $18 = (_getc($17)|0);
   $19 = ($18|0)==(-1);
   if ($19) {
    label = 8;
    break;
   }
   $20 = $18&255;
   $21 = (($2) + ($$052)|0);
   HEAP8[$21>>0] = $20;
   $22 = (($$052) + 1)|0;
   $$052 = $22;
  }
  if ((label|0) == 8) {
   $23 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $$10 = $23;
  }
  else if ((label|0) == 9) {
   $24 = ((($0)) + 53|0);
   $25 = HEAP8[$24>>0]|0;
   $26 = ($25<<24>>24)==(0);
   do {
    if ($26) {
     $28 = ((($0)) + 40|0);
     $29 = ((($0)) + 36|0);
     $30 = ((($3)) + 1|0);
     $$048 = $$sroa$speculated;
     L11: while(1) {
      $31 = HEAP32[$28>>2]|0;
      $32 = $31;
      $33 = $32;
      $34 = HEAP32[$33>>2]|0;
      $35 = (($32) + 4)|0;
      $36 = $35;
      $37 = HEAP32[$36>>2]|0;
      $38 = HEAP32[$29>>2]|0;
      $39 = (($2) + ($$048)|0);
      $40 = HEAP32[$38>>2]|0;
      $41 = ((($40)) + 16|0);
      $42 = HEAP32[$41>>2]|0;
      $43 = (FUNCTION_TABLE_iiiiiiiii[$42 & 1023]($38,$31,$2,$39,$4,$3,$30,$5)|0);
      switch ($43|0) {
      case 3:  {
       label = 16;
       break L11;
       break;
      }
      case 2:  {
       label = 18;
       break L11;
       break;
      }
      case 1:  {
       break;
      }
      default: {
       break L11;
      }
      }
      $44 = HEAP32[$28>>2]|0;
      $45 = $44;
      $46 = $45;
      HEAP32[$46>>2] = $34;
      $47 = (($45) + 4)|0;
      $48 = $47;
      HEAP32[$48>>2] = $37;
      $49 = ($$048|0)==(8);
      if ($49) {
       label = 18;
       break;
      }
      $50 = HEAP32[$15>>2]|0;
      $51 = (_getc($50)|0);
      $52 = ($51|0)==(-1);
      if ($52) {
       label = 18;
       break;
      }
      $53 = $51&255;
      HEAP8[$39>>0] = $53;
      $54 = (($$048) + 1)|0;
      $$048 = $54;
     }
     if ((label|0) == 16) {
      $55 = HEAP8[$2>>0]|0;
      HEAP8[$3>>0] = $55;
     }
     else if ((label|0) == 18) {
      $56 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $$9 = $56;
      break;
     }
     $$351 = $$048;
     label = 19;
    } else {
     $27 = HEAP8[$2>>0]|0;
     HEAP8[$3>>0] = $27;
     $$351 = $$sroa$speculated;
     label = 19;
    }
   } while(0);
   L21: do {
    if ((label|0) == 19) {
     L23: do {
      if ($1) {
       $66 = HEAP8[$3>>0]|0;
       $67 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($66)|0);
       HEAP32[$9>>2] = $67;
      } else {
       $$0 = $$351;
       while(1) {
        $57 = ($$0|0)>(0);
        if (!($57)) {
         break L23;
        }
        $58 = (($$0) + -1)|0;
        $59 = (($2) + ($58)|0);
        $60 = HEAP8[$59>>0]|0;
        $61 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($60)|0);
        $62 = HEAP32[$15>>2]|0;
        $63 = (_ungetc($61,$62)|0);
        $64 = ($63|0)==(-1);
        if ($64) {
         break;
        } else {
         $$0 = $58;
        }
       }
       $65 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $$9 = $65;
       break L21;
      }
     } while(0);
     $68 = HEAP8[$3>>0]|0;
     $69 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($68)|0);
     $$9 = $69;
    }
   } while(0);
   $$10 = $$9;
  }
  $$11 = $$10;
 } else {
  $10 = HEAP32[$9>>2]|0;
  if ($1) {
   $11 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   HEAP32[$9>>2] = $11;
   HEAP8[$6>>0] = 0;
   $$11 = $10;
  } else {
   $$11 = $10;
  }
 }
 STACKTOP = sp;return ($$11|0);
}
function __ZNSt3__27collateIcED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27collateIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__27collateIcED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__26locale5facet16__on_zero_sharedEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0|0);
 if (!($1)) {
  $2 = HEAP32[$0>>2]|0;
  $3 = ((($2)) + 4|0);
  $4 = HEAP32[$3>>2]|0;
  FUNCTION_TABLE_vi[$4 & 2047]($0);
 }
 return;
}
function __ZNKSt3__27collateIcE10do_compareEPKcS3_S3_S3_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$011 = 0, $$012 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$011 = $3;$$012 = $1;
 while(1) {
  $5 = ($$011|0)==($4|0);
  if ($5) {
   label = 7;
   break;
  }
  $6 = ($$012|0)==($2|0);
  if ($6) {
   $$0 = -1;
   break;
  }
  $7 = HEAP8[$$012>>0]|0;
  $8 = HEAP8[$$011>>0]|0;
  $9 = ($7<<24>>24)<($8<<24>>24);
  if ($9) {
   $$0 = -1;
   break;
  }
  $10 = ($8<<24>>24)<($7<<24>>24);
  if ($10) {
   $$0 = 1;
   break;
  }
  $11 = ((($$012)) + 1|0);
  $12 = ((($$011)) + 1|0);
  $$011 = $12;$$012 = $11;
 }
 if ((label|0) == 7) {
  $13 = ($$012|0)!=($2|0);
  $14 = $13&1;
  $$0 = $14;
 }
 return ($$0|0);
}
function __ZNKSt3__27collateIcE12do_transformEPKcS3_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0$i$i = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $5 = $2;
 $6 = $3;
 $7 = (($6) - ($5))|0;
 $8 = ($7>>>0)>(4294967279);
 if ($8) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $9 = ($7>>>0)<(11);
 if ($9) {
  $10 = $7&255;
  $11 = ((($0)) + 11|0);
  HEAP8[$11>>0] = $10;
  $$021$ph$i$i = $0;
 } else {
  $12 = (($7) + 16)|0;
  $13 = $12 & -16;
  $14 = (__Znwj($13)|0);
  HEAP32[$0>>2] = $14;
  $15 = $13 | -2147483648;
  $16 = ((($0)) + 8|0);
  HEAP32[$16>>2] = $15;
  $17 = ((($0)) + 4|0);
  HEAP32[$17>>2] = $7;
  $$021$ph$i$i = $14;
 }
 $$0$i$i = $2;$$021$i$i = $$021$ph$i$i;
 while(1) {
  $18 = ($$0$i$i|0)==($3|0);
  if ($18) {
   break;
  }
  __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
  $19 = ((($$0$i$i)) + 1|0);
  $20 = ((($$021$i$i)) + 1|0);
  $$0$i$i = $19;$$021$i$i = $20;
 }
 $scevgep23$i$i = (($$021$ph$i$i) + ($7)|0);
 HEAP8[$4>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$4);
 STACKTOP = sp;return;
}
function __ZNKSt3__27collateIcE7do_hashEPKcS3_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$020 = 0, $10 = 0, $11 = 0, $12 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = 0;$$020 = $1;
 while(1) {
  $3 = ($$020|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = $$0 << 4;
  $5 = HEAP8[$$020>>0]|0;
  $6 = $5 << 24 >> 24;
  $7 = (($6) + ($4))|0;
  $8 = $7 & -268435456;
  $9 = $8 >>> 24;
  $10 = $9 | $8;
  $11 = $10 ^ $7;
  $12 = ((($$020)) + 1|0);
  $$0 = $11;$$020 = $12;
 }
 return ($$0|0);
}
function __ZNSt3__26locale5facetD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__27collateIwED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27collateIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__27collateIwED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27collateIwE10do_compareEPKwS3_S3_S3_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$011 = 0, $$012 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$011 = $3;$$012 = $1;
 while(1) {
  $5 = ($$011|0)==($4|0);
  if ($5) {
   label = 7;
   break;
  }
  $6 = ($$012|0)==($2|0);
  if ($6) {
   $$0 = -1;
   break;
  }
  $7 = HEAP32[$$012>>2]|0;
  $8 = HEAP32[$$011>>2]|0;
  $9 = ($7|0)<($8|0);
  if ($9) {
   $$0 = -1;
   break;
  }
  $10 = ($8|0)<($7|0);
  if ($10) {
   $$0 = 1;
   break;
  }
  $11 = ((($$012)) + 4|0);
  $12 = ((($$011)) + 4|0);
  $$011 = $12;$$012 = $11;
 }
 if ((label|0) == 7) {
  $13 = ($$012|0)!=($2|0);
  $14 = $13&1;
  $$0 = $14;
 }
 return ($$0|0);
}
function __ZNKSt3__27collateIwE12do_transformEPKwS3_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0$i$i = 0, $$021$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $5 = $3;
 $6 = $2;
 $7 = (($5) - ($6))|0;
 $8 = $7 >> 2;
 $9 = ($8>>>0)>(1073741807);
 if ($9) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $10 = ($8>>>0)<(2);
 do {
  if ($10) {
   $11 = $8&255;
   $12 = ((($0)) + 8|0);
   $13 = ((($12)) + 3|0);
   HEAP8[$13>>0] = $11;
   $$0$i$i = $2;$$021$i$i = $0;
  } else {
   $14 = (($8) + 4)|0;
   $15 = $14 & -4;
   $16 = ($15>>>0)>(1073741823);
   if (!($16)) {
    $22 = $15 << 2;
    $23 = (__Znwj($22)|0);
    HEAP32[$0>>2] = $23;
    $24 = $15 | -2147483648;
    $25 = ((($0)) + 8|0);
    HEAP32[$25>>2] = $24;
    $26 = ((($0)) + 4|0);
    HEAP32[$26>>2] = $8;
    $$0$i$i = $2;$$021$i$i = $23;
    break;
   }
   $17 = (___cxa_allocate_exception(8)|0);
   __THREW__ = 0;
   invoke_vii(751,($17|0),(158021|0));
   $18 = __THREW__; __THREW__ = 0;
   $19 = $18&1;
   if ($19) {
    $20 = ___cxa_find_matching_catch_2()|0;
    $21 = tempRet0;
    ___cxa_free_exception(($17|0));
    ___resumeException($20|0);
    // unreachable;
   } else {
    HEAP32[$17>>2] = (22172);
    ___cxa_throw(($17|0),(2528|0),(713|0));
    // unreachable;
   }
  }
 } while(0);
 while(1) {
  $27 = ($$0$i$i|0)==($3|0);
  if ($27) {
   break;
  }
  __ZNSt3__211char_traitsIwE6assignERwRKw($$021$i$i,$$0$i$i);
  $28 = ((($$0$i$i)) + 4|0);
  $29 = ((($$021$i$i)) + 4|0);
  $$0$i$i = $28;$$021$i$i = $29;
 }
 HEAP32[$4>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($$021$i$i,$4);
 STACKTOP = sp;return;
}
function __ZNKSt3__27collateIwE7do_hashEPKwS3_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$020 = 0, $10 = 0, $11 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = 0;$$020 = $1;
 while(1) {
  $3 = ($$020|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = $$0 << 4;
  $5 = HEAP32[$$020>>2]|0;
  $6 = (($5) + ($4))|0;
  $7 = $6 & -268435456;
  $8 = $7 >>> 24;
  $9 = $8 | $7;
  $10 = $9 ^ $6;
  $11 = ((($$020)) + 4|0);
  $$0 = $10;$$020 = $11;
 }
 return ($$0|0);
}
function __ZNSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRb($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$2 = 0, $$232 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$ptr = 0, $$sroa$026$0 = 0, $$sroa$026$0$copyload = 0, $$sroa$026$0$copyload28 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $$byval_copy2 = sp + 60|0;
 $$byval_copy1 = sp + 56|0;
 $$byval_copy = sp + 52|0;
 $6 = sp + 48|0;
 $7 = sp + 44|0;
 $8 = sp + 40|0;
 $9 = sp + 36|0;
 $10 = sp + 32|0;
 $11 = sp + 8|0;
 $12 = sp;
 $13 = ((($3)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 1;
 $16 = ($15|0)==(0);
 L1: do {
  if ($16) {
   HEAP32[$6>>2] = -1;
   $17 = HEAP32[$0>>2]|0;
   $18 = ((($17)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = HEAP32[$1>>2]|0;
   HEAP32[$7>>2] = $20;
   $21 = HEAP32[$2>>2]|0;
   HEAP32[$8>>2] = $21;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$8>>2]|0;
   $22 = (FUNCTION_TABLE_iiiiiii[$19 & 1023]($0,$$byval_copy,$$byval_copy1,$3,$4,$6)|0);
   HEAP32[$1>>2] = $22;
   $23 = HEAP32[$6>>2]|0;
   switch ($23|0) {
   case 0:  {
    HEAP8[$5>>0] = 0;
    break;
   }
   case 1:  {
    HEAP8[$5>>0] = 1;
    break;
   }
   default: {
    HEAP8[$5>>0] = 1;
    HEAP32[$4>>2] = 4;
   }
   }
   $$sroa$026$0$copyload = HEAP32[$1>>2]|0;
   $$sroa$026$0 = $$sroa$026$0$copyload;
  } else {
   __ZNKSt3__28ios_base6getlocEv($9,$3);
   __THREW__ = 0;
   $24 = (invoke_iii(734,($9|0),(165440|0))|0);
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   do {
    if ($26) {
     $49 = ___cxa_find_matching_catch_2()|0;
     $50 = tempRet0;
     __ZNSt3__26localeD2Ev($9);
     $$2 = $49;$$232 = $50;
    } else {
     __ZNSt3__26localeD2Ev($9);
     __ZNKSt3__28ios_base6getlocEv($10,$3);
     __THREW__ = 0;
     $27 = (invoke_iii(734,($10|0),(165456|0))|0);
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $51 = ___cxa_find_matching_catch_2()|0;
      $52 = tempRet0;
      __ZNSt3__26localeD2Ev($10);
      $$2 = $51;$$232 = $52;
      break;
     }
     __ZNSt3__26localeD2Ev($10);
     $30 = HEAP32[$27>>2]|0;
     $31 = ((($30)) + 24|0);
     $32 = HEAP32[$31>>2]|0;
     __THREW__ = 0;
     invoke_vii($32|0,($11|0),($27|0));
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      $35 = ___cxa_find_matching_catch_2()|0;
      $36 = tempRet0;
      $$2 = $35;$$232 = $36;
      break;
     }
     $$ptr = ((($11)) + 12|0);
     $37 = HEAP32[$27>>2]|0;
     $38 = ((($37)) + 28|0);
     $39 = HEAP32[$38>>2]|0;
     __THREW__ = 0;
     invoke_vii($39|0,($$ptr|0),($27|0));
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if ($41) {
      $61 = ___cxa_find_matching_catch_2()|0;
      $62 = tempRet0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($11);
      $$2 = $61;$$232 = $62;
      break;
     }
     $42 = HEAP32[$2>>2]|0;
     HEAP32[$12>>2] = $42;
     $43 = ((($11)) + 24|0);
     __THREW__ = 0;
     ;HEAP32[$$byval_copy2>>2]=HEAP32[$12>>2]|0;
     $44 = (invoke_iiiiiiii(952,($1|0),($$byval_copy2|0),($11|0),($43|0),($24|0),($4|0),1)|0);
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      $53 = ___cxa_find_matching_catch_2()|0;
      $54 = tempRet0;
      $59 = $43;
      while(1) {
       $58 = ((($59)) + -12|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($58);
       $60 = ($58|0)==($11|0);
       if ($60) {
        break;
       } else {
        $59 = $58;
       }
      }
      $$2 = $53;$$232 = $54;
      break;
     } else {
      $47 = ($44|0)==($11|0);
      $48 = $47&1;
      HEAP8[$5>>0] = $48;
      $$sroa$026$0$copyload28 = HEAP32[$1>>2]|0;
      $56 = $43;
      while(1) {
       $55 = ((($56)) + -12|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($55);
       $57 = ($55|0)==($11|0);
       if ($57) {
        break;
       } else {
        $56 = $55;
       }
      }
      $$sroa$026$0 = $$sroa$026$0$copyload28;
      break L1;
     }
    }
   } while(0);
   ___resumeException($$2|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$026$0|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRl($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRx($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRt($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRy($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRf($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRd($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRe($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i48 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i54 = 0, $$0$i$i30 = 0, $$0$i$i43 = 0, $$022 = 0, $$025 = 0, $$123 = 0, $$2 = 0, $$224 = 0, $$227 = 0, $$3 = 0, $$328 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i50 = 0, $$pre$i4$i$i56 = 0;
 var $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i31 = 0, $phitmp$i = 0, $phitmp$i38 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $vararg_buffer = sp;
 $6 = sp + 208|0;
 $7 = sp + 196|0;
 $8 = sp + 192|0;
 $9 = sp + 180|0;
 $10 = sp + 176|0;
 $11 = sp + 16|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $14 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$14>>2] = 0;
  $15 = (($$0$i$i) + 1)|0;
  $$0$i$i = $15;
 }
 __THREW__ = 0;
 invoke_vii(733,($8|0),($3|0));
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 do {
  if ($17) {
   $99 = ___cxa_find_matching_catch_2()|0;
   $100 = tempRet0;
   $$0 = $99;$$025 = $100;
   label = 36;
  } else {
   __THREW__ = 0;
   $18 = (invoke_iii(734,($8|0),(165440|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    $21 = HEAP32[$18>>2]|0;
    $22 = ((($21)) + 32|0);
    $23 = HEAP32[$22>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($23|0,($18|0),(155824|0),((155850)|0),($6|0))|0);
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if (!($25)) {
     __ZNSt3__26localeD2Ev($8);
     ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
     $$0$i$i30 = 0;
     while(1) {
      $exitcond$i$i31 = ($$0$i$i30|0)==(3);
      if ($exitcond$i$i31) {
       break;
      }
      $26 = (($9) + ($$0$i$i30<<2)|0);
      HEAP32[$26>>2] = 0;
      $27 = (($$0$i$i30) + 1)|0;
      $$0$i$i30 = $27;
     }
     $28 = ((($9)) + 11|0);
     $29 = HEAP8[$28>>0]|0;
     $30 = ($29<<24>>24)<(0);
     if ($30) {
      $31 = ((($9)) + 8|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = $32 & 2147483647;
      $phitmp$i = (($33) + -1)|0;
      $34 = $phitmp$i;
     } else {
      $34 = 10;
     }
     __THREW__ = 0;
     invoke_viii(953,($9|0),($34|0),0);
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      $$2 = $103;$$227 = $104;
     } else {
      $37 = HEAP8[$28>>0]|0;
      $38 = ($37<<24>>24)<(0);
      $39 = HEAP32[$9>>2]|0;
      $40 = $38 ? $39 : $9;
      HEAP32[$10>>2] = $40;
      HEAP32[$12>>2] = $11;
      HEAP32[$13>>2] = 0;
      $41 = ((($9)) + 4|0);
      $42 = ((($9)) + 8|0);
      $$pre = HEAP32[$1>>2]|0;
      $$022 = $40;$$pre$i$i$i = $$pre;$186 = $$pre;
      L21: while(1) {
       $43 = ($$pre$i$i$i|0)==(0|0);
       if ($43) {
        $$pre$i$i$i50 = 0;$156 = 0;$187 = 1;
       } else {
        $44 = ((($$pre$i$i$i)) + 12|0);
        $45 = HEAP32[$44>>2]|0;
        $46 = ((($$pre$i$i$i)) + 16|0);
        $47 = HEAP32[$46>>2]|0;
        $48 = ($45|0)==($47|0);
        if ($48) {
         $49 = HEAP32[$$pre$i$i$i>>2]|0;
         $50 = ((($49)) + 36|0);
         $51 = HEAP32[$50>>2]|0;
         __THREW__ = 0;
         $52 = (invoke_ii($51|0,($$pre$i$i$i|0))|0);
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if ($54) {
          break;
         } else {
          $$0$i$i$i$i = $52;
         }
        } else {
         $55 = HEAP8[$45>>0]|0;
         $56 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($55)|0);
         $$0$i$i$i$i = $56;
        }
        $57 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $58 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$57)|0);
        if ($58) {
         HEAP32[$1>>2] = 0;
         $$pre$i$i$i50 = 0;$156 = 0;$187 = 1;
        } else {
         $$pre$i$i$i50 = $$pre$i$i$i;$156 = $186;$187 = 0;
        }
       }
       $59 = HEAP32[$2>>2]|0;
       $60 = ($59|0)==(0|0);
       do {
        if ($60) {
         label = 26;
        } else {
         $61 = ((($59)) + 12|0);
         $62 = HEAP32[$61>>2]|0;
         $63 = ((($59)) + 16|0);
         $64 = HEAP32[$63>>2]|0;
         $65 = ($62|0)==($64|0);
         if ($65) {
          $66 = HEAP32[$59>>2]|0;
          $67 = ((($66)) + 36|0);
          $68 = HEAP32[$67>>2]|0;
          __THREW__ = 0;
          $69 = (invoke_ii($68|0,($59|0))|0);
          $70 = __THREW__; __THREW__ = 0;
          $71 = $70&1;
          if ($71) {
           break L21;
          } else {
           $$0$i$i2$i$i = $69;
          }
         } else {
          $72 = HEAP8[$62>>0]|0;
          $73 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($72)|0);
          $$0$i$i2$i$i = $73;
         }
         $74 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $75 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$74)|0);
         if ($75) {
          HEAP32[$2>>2] = 0;
          label = 26;
          break;
         } else {
          if ($187) {
           $188 = $59;
           break;
          } else {
           $$224 = $$022;$$pre$i4$i$i56 = $59;
           label = 46;
           break L21;
          }
         }
        }
       } while(0);
       if ((label|0) == 26) {
        label = 0;
        if ($187) {
         $$224 = $$022;$$pre$i4$i$i56 = 0;
         label = 46;
         break;
        } else {
         $188 = 0;
        }
       }
       $76 = HEAP32[$10>>2]|0;
       $77 = HEAP8[$28>>0]|0;
       $78 = ($77<<24>>24)<(0);
       $79 = HEAP32[$41>>2]|0;
       $80 = $77&255;
       $81 = $78 ? $79 : $80;
       $82 = (($$022) + ($81)|0);
       $83 = ($76|0)==($82|0);
       if ($83) {
        $84 = $81 << 1;
        __THREW__ = 0;
        invoke_viii(953,($9|0),($84|0),0);
        $85 = __THREW__; __THREW__ = 0;
        $86 = $85&1;
        if ($86) {
         break;
        }
        $87 = HEAP8[$28>>0]|0;
        $88 = ($87<<24>>24)<(0);
        if ($88) {
         $89 = HEAP32[$42>>2]|0;
         $90 = $89 & 2147483647;
         $phitmp$i38 = (($90) + -1)|0;
         $91 = $phitmp$i38;
        } else {
         $91 = 10;
        }
        __THREW__ = 0;
        invoke_viii(953,($9|0),($91|0),0);
        $92 = __THREW__; __THREW__ = 0;
        $93 = $92&1;
        if ($93) {
         break;
        }
        $94 = HEAP8[$28>>0]|0;
        $95 = ($94<<24>>24)<(0);
        $96 = HEAP32[$9>>2]|0;
        $97 = $95 ? $96 : $9;
        $98 = (($97) + ($81)|0);
        HEAP32[$10>>2] = $98;
        $$123 = $97;
       } else {
        $$123 = $$022;
       }
       $105 = ((($$pre$i$i$i50)) + 12|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = ((($$pre$i$i$i50)) + 16|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = ($106|0)==($108|0);
       if ($109) {
        $110 = HEAP32[$$pre$i$i$i50>>2]|0;
        $111 = ((($110)) + 36|0);
        $112 = HEAP32[$111>>2]|0;
        __THREW__ = 0;
        $113 = (invoke_ii($112|0,($$pre$i$i$i50|0))|0);
        $114 = __THREW__; __THREW__ = 0;
        $115 = $114&1;
        if ($115) {
         break;
        } else {
         $$0$i$i43 = $113;
        }
       } else {
        $116 = HEAP8[$106>>0]|0;
        $117 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($116)|0);
        $$0$i$i43 = $117;
       }
       $118 = $$0$i$i43&255;
       __THREW__ = 0;
       $119 = (invoke_iiiiiiiiiii(954,($118|0),16,($$123|0),($10|0),($13|0),0,($7|0),($11|0),($12|0),($6|0))|0);
       $120 = __THREW__; __THREW__ = 0;
       $121 = $120&1;
       if ($121) {
        break;
       }
       $122 = ($119|0)==(0);
       if (!($122)) {
        $$224 = $$123;$$pre$i4$i$i56 = $188;
        label = 46;
        break;
       }
       $123 = HEAP32[$105>>2]|0;
       $124 = HEAP32[$107>>2]|0;
       $125 = ($123|0)==($124|0);
       if ($125) {
        $126 = HEAP32[$$pre$i$i$i50>>2]|0;
        $127 = ((($126)) + 40|0);
        $128 = HEAP32[$127>>2]|0;
        __THREW__ = 0;
        (invoke_ii($128|0,($$pre$i$i$i50|0))|0);
        $129 = __THREW__; __THREW__ = 0;
        $130 = $129&1;
        if ($130) {
         break;
        } else {
         $$022 = $$123;$$pre$i$i$i = $$pre$i$i$i50;$186 = $156;
         continue;
        }
       } else {
        $131 = ((($123)) + 1|0);
        HEAP32[$105>>2] = $131;
        $132 = HEAP8[$123>>0]|0;
        (__ZNSt3__211char_traitsIcE11to_int_typeEc($132)|0);
        $$022 = $$123;$$pre$i$i$i = $$pre$i$i$i50;$186 = $156;
        continue;
       }
      }
      L60: do {
       if ((label|0) == 46) {
        $133 = HEAP32[$10>>2]|0;
        $134 = $$224;
        $135 = (($133) - ($134))|0;
        __THREW__ = 0;
        invoke_viii(953,($9|0),($135|0),0);
        $136 = __THREW__; __THREW__ = 0;
        $137 = $136&1;
        if (!($137)) {
         $138 = HEAP8[$28>>0]|0;
         $139 = HEAP32[$9>>2]|0;
         __THREW__ = 0;
         $140 = (invoke_i(955)|0);
         $141 = __THREW__; __THREW__ = 0;
         $142 = $141&1;
         if (!($142)) {
          $143 = ($138<<24>>24)<(0);
          $144 = $143 ? $139 : $9;
          __THREW__ = 0;
          HEAP32[$vararg_buffer>>2] = $5;
          $145 = (invoke_iiiii(956,($144|0),($140|0),(155857|0),($vararg_buffer|0))|0);
          $146 = __THREW__; __THREW__ = 0;
          $147 = $146&1;
          if (!($147)) {
           $148 = ($145|0)==(1);
           if (!($148)) {
            HEAP32[$4>>2] = 4;
           }
           $149 = ($$pre$i$i$i50|0)==(0|0);
           if ($149) {
            $189 = 1;
           } else {
            $150 = ((($$pre$i$i$i50)) + 12|0);
            $151 = HEAP32[$150>>2]|0;
            $152 = ((($$pre$i$i$i50)) + 16|0);
            $153 = HEAP32[$152>>2]|0;
            $154 = ($151|0)==($153|0);
            if ($154) {
             $155 = HEAP32[$156>>2]|0;
             $157 = ((($155)) + 36|0);
             $158 = HEAP32[$157>>2]|0;
             __THREW__ = 0;
             $159 = (invoke_ii($158|0,($$pre$i$i$i50|0))|0);
             $160 = __THREW__; __THREW__ = 0;
             $161 = $160&1;
             if ($161) {
              break;
             } else {
              $$0$i$i$i$i48 = $159;
             }
            } else {
             $162 = HEAP8[$151>>0]|0;
             $163 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($162)|0);
             $$0$i$i$i$i48 = $163;
            }
            $164 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
            $165 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i48,$164)|0);
            if ($165) {
             HEAP32[$1>>2] = 0;
             $189 = 1;
            } else {
             $189 = 0;
            }
           }
           $166 = ($$pre$i4$i$i56|0)==(0|0);
           do {
            if ($166) {
             label = 63;
            } else {
             $167 = ((($$pre$i4$i$i56)) + 12|0);
             $168 = HEAP32[$167>>2]|0;
             $169 = ((($$pre$i4$i$i56)) + 16|0);
             $170 = HEAP32[$169>>2]|0;
             $171 = ($168|0)==($170|0);
             if ($171) {
              $172 = HEAP32[$$pre$i4$i$i56>>2]|0;
              $173 = ((($172)) + 36|0);
              $174 = HEAP32[$173>>2]|0;
              __THREW__ = 0;
              $175 = (invoke_ii($174|0,($$pre$i4$i$i56|0))|0);
              $176 = __THREW__; __THREW__ = 0;
              $177 = $176&1;
              if ($177) {
               break L60;
              } else {
               $$0$i$i2$i$i54 = $175;
              }
             } else {
              $178 = HEAP8[$168>>0]|0;
              $179 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($178)|0);
              $$0$i$i2$i$i54 = $179;
             }
             $180 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
             $181 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i54,$180)|0);
             if ($181) {
              HEAP32[$2>>2] = 0;
              label = 63;
              break;
             } else {
              if ($189) {
               break;
              } else {
               label = 65;
               break;
              }
             }
            }
           } while(0);
           if ((label|0) == 63) {
            if ($189) {
             label = 65;
            }
           }
           if ((label|0) == 65) {
            $182 = HEAP32[$4>>2]|0;
            $183 = $182 | 2;
            HEAP32[$4>>2] = $183;
           }
           $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
           STACKTOP = sp;return ($$sroa$0$0$copyload|0);
          }
         }
        }
       }
      } while(0);
      $184 = ___cxa_find_matching_catch_2()|0;
      $185 = tempRet0;
      $$2 = $184;$$227 = $185;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
     $$3 = $$2;$$328 = $$227;
     break;
    }
   }
   $101 = ___cxa_find_matching_catch_2()|0;
   $102 = tempRet0;
   __ZNSt3__26localeD2Ev($8);
   $$0 = $101;$$025 = $102;
   label = 36;
  }
 } while(0);
 if ((label|0) == 36) {
  $$3 = $$0;$$328 = $$025;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
 ___resumeException($$3|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__26locale9use_facetERNS0_2idE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = (__ZNSt3__26locale2id5__getEv($1)|0);
 $4 = (__ZNKSt3__26locale5__imp9use_facetEl($2,$3)|0);
 return ($4|0);
}
function __ZNSt3__26localeD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 (__ZNSt3__214__shared_count16__release_sharedEv($1)|0);
 return;
}
function __ZNSt3__29__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$add = 0, $$0$i$idx = 0, $$0$i$ptr = 0, $$0$lcssa$i = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $10 = HEAP32[$3>>2]|0;
 $11 = ($10|0)==($2|0);
 do {
  if ($11) {
   $12 = ((($9)) + 24|0);
   $13 = HEAP8[$12>>0]|0;
   $14 = ($13<<24>>24)==($0<<24>>24);
   if (!($14)) {
    $15 = ((($9)) + 25|0);
    $16 = HEAP8[$15>>0]|0;
    $17 = ($16<<24>>24)==($0<<24>>24);
    if (!($17)) {
     label = 5;
     break;
    }
   }
   $18 = $14 ? 43 : 45;
   $19 = ((($2)) + 1|0);
   HEAP32[$3>>2] = $19;
   HEAP8[$2>>0] = $18;
   HEAP32[$4>>2] = 0;
   $$1 = 0;
  } else {
   label = 5;
  }
 } while(0);
 L6: do {
  if ((label|0) == 5) {
   $20 = ((($6)) + 11|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = ($21<<24>>24)<(0);
   $23 = ((($6)) + 4|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = $21&255;
   $26 = $22 ? $24 : $25;
   $27 = ($26|0)!=(0);
   $28 = ($0<<24>>24)==($5<<24>>24);
   $or$cond = $28 & $27;
   if ($or$cond) {
    $29 = HEAP32[$8>>2]|0;
    $30 = $29;
    $31 = $7;
    $32 = (($30) - ($31))|0;
    $33 = ($32|0)<(160);
    if (!($33)) {
     $$1 = 0;
     break;
    }
    $34 = HEAP32[$4>>2]|0;
    $35 = ((($29)) + 4|0);
    HEAP32[$8>>2] = $35;
    HEAP32[$29>>2] = $34;
    HEAP32[$4>>2] = 0;
    $$1 = 0;
    break;
   }
   $36 = ((($9)) + 26|0);
   $$0$i$idx = 0;
   while(1) {
    $$0$i$ptr = (($9) + ($$0$i$idx)|0);
    $37 = ($$0$i$idx|0)==(26);
    if ($37) {
     $$0$lcssa$i = $36;
     break;
    }
    $38 = HEAP8[$$0$i$ptr>>0]|0;
    $39 = ($38<<24>>24)==($0<<24>>24);
    $$0$i$add = (($$0$i$idx) + 1)|0;
    if ($39) {
     $$0$lcssa$i = $$0$i$ptr;
     break;
    } else {
     $$0$i$idx = $$0$i$add;
    }
   }
   $40 = $$0$lcssa$i;
   $41 = $9;
   $42 = (($40) - ($41))|0;
   $43 = ($42|0)>(23);
   if ($43) {
    $$1 = -1;
   } else {
    $44 = (155824 + ($42)|0);
    switch ($1|0) {
    case 10: case 8:  {
     $45 = ($42|0)<($1|0);
     if (!($45)) {
      $$1 = -1;
      break L6;
     }
     break;
    }
    case 16:  {
     $46 = ($42|0)<(22);
     if (!($46)) {
      if ($11) {
       $$1 = -1;
       break L6;
      }
      $47 = $10;
      $48 = $2;
      $49 = (($47) - ($48))|0;
      $50 = ($49|0)<(3);
      if (!($50)) {
       $$1 = -1;
       break L6;
      }
      $51 = ((($10)) + -1|0);
      $52 = HEAP8[$51>>0]|0;
      $53 = ($52<<24>>24)==(48);
      if (!($53)) {
       $$1 = -1;
       break L6;
      }
      HEAP32[$4>>2] = 0;
      $54 = HEAP8[$44>>0]|0;
      $55 = ((($10)) + 1|0);
      HEAP32[$3>>2] = $55;
      HEAP8[$10>>0] = $54;
      $$1 = 0;
      break L6;
     }
     break;
    }
    default: {
    }
    }
    $56 = HEAP8[$44>>0]|0;
    $57 = ((($10)) + 1|0);
    HEAP32[$3>>2] = $57;
    HEAP8[$10>>0] = $56;
    $58 = HEAP32[$4>>2]|0;
    $59 = (($58) + 1)|0;
    HEAP32[$4>>2] = $59;
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZNSt3__26__clocEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161000]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire(161000)|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    $4 = (invoke_iiii(957,2147483647,(155860|0),(0|0))|0);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41362] = $4;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41362]|0;
 return ($7|0);
}
function __ZNSt3__217__libcpp_sscanf_lEPKcP15__locale_structS1_z($0,$1,$2,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $varargs = $varargs|0;
 var $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 HEAP32[$3>>2] = $varargs;
 $4 = (___uselocale($1)|0);
 $5 = (_vsscanf($0,$2,$3)|0);
 $6 = ($4|0)==(0|0);
 if (!($6)) {
  __THREW__ = 0;
  (invoke_ii(958,($4|0))|0);
  $7 = __THREW__; __THREW__ = 0;
  $8 = $7&1;
  if ($8) {
   $9 = ___cxa_find_matching_catch_3(0|0)|0;
   $10 = tempRet0;
   ___clang_call_terminate($9);
   // unreachable;
  }
 }
 STACKTOP = sp;return ($5|0);
}
function __ZNSt3__26locale2id5__getEv($0) {
 $0 = $0|0;
 var $$byval_copy = 0, $$fca$1$gep = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $$byval_copy = sp + 32|0;
 $1 = sp + 24|0;
 $2 = sp + 20|0;
 $3 = sp + 8|0;
 $4 = sp;
 HEAP32[$4>>2] = (959);
 $$fca$1$gep = ((($4)) + 4|0);
 HEAP32[$$fca$1$gep>>2] = 0;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$4>>2]|0;HEAP32[$$byval_copy+4>>2]=HEAP32[$4+4>>2]|0;
 __ZNSt3__212_GLOBAL__N_111__fake_bindC2EMNS_6locale2idEFvvEPS3_($3,$$byval_copy,$0);
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5|0)==(-1);
 if (!($6)) {
  HEAP32[$1>>2] = $3;
  HEAP32[$2>>2] = $1;
  __ZNSt3__211__call_onceERVmPvPFvS2_E($0,$2,960);
 }
 $7 = ((($0)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = (($8) + -1)|0;
 STACKTOP = sp;return ($9|0);
}
function __ZNKSt3__26locale5__imp9use_facetEl($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNKSt3__26locale5__imp9has_facetEl($0,$1)|0);
 if ($2) {
  $4 = ((($0)) + 8|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = (($5) + ($1<<2)|0);
  $7 = HEAP32[$6>>2]|0;
  return ($7|0);
 } else {
  $3 = (___cxa_allocate_exception(4)|0);
  __ZNSt8bad_castC2Ev($3);
  ___cxa_throw(($3|0),(2560|0),(720|0));
  // unreachable;
 }
 return (0)|0;
}
function __ZNKSt3__26locale5__imp9has_facetEl($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 8|0);
 $3 = ((($0)) + 12|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$2>>2]|0;
 $6 = (($4) - ($5))|0;
 $7 = $6 >> 2;
 $8 = ($7>>>0)>($1>>>0);
 if ($8) {
  $9 = $5;
  $10 = (($9) + ($1<<2)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)!=(0|0);
  $13 = $12;
 } else {
  $13 = 0;
 }
 return ($13|0);
}
function __ZNSt3__26locale2id6__initEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[41363]|0;HEAP32[41363] = (($1+1)|0);
 $2 = (($1) + 1)|0;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $2;
 return;
}
function __ZNSt3__212_GLOBAL__N_111__fake_bindC2EMNS_6locale2idEFvvEPS3_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$elt4 = 0, $$repack = 0, $$repack7 = 0, $$unpack = 0, $$unpack5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$unpack = HEAP32[$1>>2]|0;
 $$elt4 = ((($1)) + 4|0);
 $$unpack5 = HEAP32[$$elt4>>2]|0;
 HEAP32[$0>>2] = $2;
 $$repack = ((($0)) + 4|0);
 HEAP32[$$repack>>2] = $$unpack;
 $$repack7 = ((($0)) + 8|0);
 HEAP32[$$repack7>>2] = $$unpack5;
 return;
}
function __ZNSt3__217__call_once_proxyINS_5tupleIJONS_12_GLOBAL__N_111__fake_bindEEEEEEvPv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = HEAP32[$1>>2]|0;
 __ZNKSt3__212_GLOBAL__N_111__fake_bindclEv($2);
 return;
}
function __ZNKSt3__212_GLOBAL__N_111__fake_bindclEv($0) {
 $0 = $0|0;
 var $$elt = 0, $$elt1 = 0, $$unpack = 0, $$unpack2 = 0, $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $$elt = ((($0)) + 4|0);
 $$unpack = HEAP32[$$elt>>2]|0;
 $$elt1 = ((($0)) + 8|0);
 $$unpack2 = HEAP32[$$elt1>>2]|0;
 $2 = $$unpack2 >> 1;
 $3 = (($1) + ($2)|0);
 $4 = $$unpack2 & 1;
 $5 = ($4|0)==(0);
 if ($5) {
  $9 = $$unpack;
  $10 = $9;
 } else {
  $6 = HEAP32[$3>>2]|0;
  $7 = (($6) + ($$unpack)|0);
  $8 = HEAP32[$7>>2]|0;
  $10 = $8;
 }
 FUNCTION_TABLE_vi[$10 & 2047]($3);
 return;
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0;
 var $phitmp$i29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $6 = sp + 200|0;
 $7 = sp + 199|0;
 $8 = sp + 198|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 197|0;
 $16 = sp + 196|0;
 __ZNSt3__29__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$183 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP8[$36>>0]|0;
     $47 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$153 = $183;$184 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP8[$53>>0]|0;
      $64 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($184) {
       $185 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($184) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $185 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP8[$93>>0]|0;
    $104 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = $$0$i$i34&255;
   $106 = HEAP8[$7>>0]|0;
   $107 = HEAP8[$8>>0]|0;
   __THREW__ = 0;
   $108 = (invoke_iiiiiiiiiiiii(961,($105|0),($15|0),($16|0),($$119|0),($11|0),($106|0),($107|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $109 = __THREW__; __THREW__ = 0;
   $110 = $109&1;
   if ($110) {
    break;
   }
   $111 = ($108|0)==(0);
   if (!($111)) {
    $$2 = $$119;$$pre$i4$i$i46 = $185;
    label = 37;
    break;
   }
   $112 = HEAP32[$92>>2]|0;
   $113 = HEAP32[$94>>2]|0;
   $114 = ($112|0)==($113|0);
   if ($114) {
    $115 = HEAP32[$$pre$i$i$i40>>2]|0;
    $116 = ((($115)) + 40|0);
    $117 = HEAP32[$116>>2]|0;
    __THREW__ = 0;
    (invoke_ii($117|0,($$pre$i$i$i40|0))|0);
    $118 = __THREW__; __THREW__ = 0;
    $119 = $118&1;
    if ($119) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
     continue;
    }
   } else {
    $120 = ((($112)) + 1|0);
    HEAP32[$92>>2] = $120;
    $121 = HEAP8[$112>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($121)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $122 = ((($9)) + 11|0);
    $123 = HEAP8[$122>>0]|0;
    $124 = ($123<<24>>24)<(0);
    $125 = ((($9)) + 4|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = $123&255;
    $128 = $124 ? $126 : $127;
    $129 = ($128|0)==(0);
    $130 = HEAP8[$15>>0]|0;
    $131 = ($130<<24>>24)==(0);
    $or$cond = $129 | $131;
    if (!($or$cond)) {
     $132 = HEAP32[$13>>2]|0;
     $133 = $132;
     $134 = $12;
     $135 = (($133) - ($134))|0;
     $136 = ($135|0)<(160);
     if ($136) {
      $137 = HEAP32[$14>>2]|0;
      $138 = ((($132)) + 4|0);
      HEAP32[$13>>2] = $138;
      HEAP32[$132>>2] = $137;
     }
    }
    $139 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $140 = (+invoke_diii(962,($$2|0),($139|0),($4|0)));
    $141 = __THREW__; __THREW__ = 0;
    $142 = $141&1;
    if (!($142)) {
     HEAPF64[$5>>3] = $140;
     $143 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($143|0),($4|0));
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if (!($145)) {
      $146 = ($$pre$i$i$i40|0)==(0|0);
      if ($146) {
       $186 = 1;
      } else {
       $147 = ((($$pre$i$i$i40)) + 12|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ((($$pre$i$i$i40)) + 16|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ($148|0)==($150|0);
       if ($151) {
        $152 = HEAP32[$153>>2]|0;
        $154 = ((($152)) + 36|0);
        $155 = HEAP32[$154>>2]|0;
        __THREW__ = 0;
        $156 = (invoke_ii($155|0,($$pre$i$i$i40|0))|0);
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        } else {
         $$0$i$i$i$i38 = $156;
        }
       } else {
        $159 = HEAP8[$148>>0]|0;
        $160 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($159)|0);
        $$0$i$i$i$i38 = $160;
       }
       $161 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $162 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i38,$161)|0);
       if ($162) {
        HEAP32[$1>>2] = 0;
        $186 = 1;
       } else {
        $186 = 0;
       }
      }
      $163 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($163) {
        label = 54;
       } else {
        $164 = ((($$pre$i4$i$i46)) + 12|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = ((($$pre$i4$i$i46)) + 16|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = ($165|0)==($167|0);
        if ($168) {
         $169 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $170 = ((($169)) + 36|0);
         $171 = HEAP32[$170>>2]|0;
         __THREW__ = 0;
         $172 = (invoke_ii($171|0,($$pre$i4$i$i46|0))|0);
         $173 = __THREW__; __THREW__ = 0;
         $174 = $173&1;
         if ($174) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $172;
         }
        } else {
         $175 = HEAP8[$165>>0]|0;
         $176 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($175)|0);
         $$0$i$i2$i$i44 = $176;
        }
        $177 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $178 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i44,$177)|0);
        if ($178) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($186) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($186) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $179 = HEAP32[$4>>2]|0;
       $180 = $179 | 2;
       HEAP32[$4>>2] = $180;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  $$1 = $181;$$121 = $182;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__29__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $5 = sp;
 __ZNKSt3__28ios_base6getlocEv($5,$1);
 __THREW__ = 0;
 $6 = (invoke_iii(734,($5|0),(165440|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $9 = HEAP32[$6>>2]|0;
  $10 = ((($9)) + 32|0);
  $11 = HEAP32[$10>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($11|0,($6|0),(155824|0),((155856)|0),($2|0))|0);
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   __THREW__ = 0;
   $14 = (invoke_iii(734,($5|0),(165456|0))|0);
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    $17 = HEAP32[$14>>2]|0;
    $18 = ((($17)) + 12|0);
    $19 = HEAP32[$18>>2]|0;
    __THREW__ = 0;
    $20 = (invoke_ii($19|0,($14|0))|0);
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if (!($22)) {
     HEAP8[$3>>0] = $20;
     $23 = HEAP32[$14>>2]|0;
     $24 = ((($23)) + 16|0);
     $25 = HEAP32[$24>>2]|0;
     __THREW__ = 0;
     $26 = (invoke_ii($25|0,($14|0))|0);
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if (!($28)) {
      HEAP8[$4>>0] = $26;
      $29 = HEAP32[$14>>2]|0;
      $30 = ((($29)) + 20|0);
      $31 = HEAP32[$30>>2]|0;
      __THREW__ = 0;
      invoke_vii($31|0,($0|0),($14|0));
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if (!($33)) {
       __ZNSt3__26localeD2Ev($5);
       STACKTOP = sp;return;
      }
     }
    }
   }
  }
 }
 $34 = ___cxa_find_matching_catch_2()|0;
 $35 = tempRet0;
 __ZNSt3__26localeD2Ev($5);
 ___resumeException($34|0);
 // unreachable;
}
function __ZNSt3__29__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 $11 = $11|0;
 var $$0$i$add = 0, $$0$i$idx = 0, $$0$i$ptr = 0, $$0$lcssa$i = 0, $$2 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $12 = ($0<<24>>24)==($5<<24>>24);
 $13 = $8;
 L1: do {
  if ($12) {
   $14 = HEAP8[$1>>0]|0;
   $15 = ($14<<24>>24)==(0);
   if ($15) {
    $$2 = -1;
   } else {
    HEAP8[$1>>0] = 0;
    $16 = HEAP32[$4>>2]|0;
    $17 = ((($16)) + 1|0);
    HEAP32[$4>>2] = $17;
    HEAP8[$16>>0] = 46;
    $18 = ((($7)) + 11|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = ($19<<24>>24)<(0);
    $21 = ((($7)) + 4|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = $19&255;
    $24 = $20 ? $22 : $23;
    $25 = ($24|0)==(0);
    if ($25) {
     $$2 = 0;
    } else {
     $26 = HEAP32[$9>>2]|0;
     $27 = $26;
     $28 = (($27) - ($13))|0;
     $29 = ($28|0)<(160);
     if ($29) {
      $30 = HEAP32[$10>>2]|0;
      $31 = ((($26)) + 4|0);
      HEAP32[$9>>2] = $31;
      HEAP32[$26>>2] = $30;
      $$2 = 0;
     } else {
      $$2 = 0;
     }
    }
   }
  } else {
   $32 = ($0<<24>>24)==($6<<24>>24);
   if ($32) {
    $33 = ((($7)) + 11|0);
    $34 = HEAP8[$33>>0]|0;
    $35 = ($34<<24>>24)<(0);
    $36 = ((($7)) + 4|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = $34&255;
    $39 = $35 ? $37 : $38;
    $40 = ($39|0)==(0);
    if (!($40)) {
     $41 = HEAP8[$1>>0]|0;
     $42 = ($41<<24>>24)==(0);
     if ($42) {
      $$2 = -1;
      break;
     }
     $43 = HEAP32[$9>>2]|0;
     $44 = $43;
     $45 = (($44) - ($13))|0;
     $46 = ($45|0)<(160);
     if (!($46)) {
      $$2 = 0;
      break;
     }
     $47 = HEAP32[$10>>2]|0;
     $48 = ((($43)) + 4|0);
     HEAP32[$9>>2] = $48;
     HEAP32[$43>>2] = $47;
     HEAP32[$10>>2] = 0;
     $$2 = 0;
     break;
    }
   }
   $49 = ((($11)) + 32|0);
   $$0$i$idx = 0;
   while(1) {
    $$0$i$ptr = (($11) + ($$0$i$idx)|0);
    $50 = ($$0$i$idx|0)==(32);
    if ($50) {
     $$0$lcssa$i = $49;
     break;
    }
    $51 = HEAP8[$$0$i$ptr>>0]|0;
    $52 = ($51<<24>>24)==($0<<24>>24);
    $$0$i$add = (($$0$i$idx) + 1)|0;
    if ($52) {
     $$0$lcssa$i = $$0$i$ptr;
     break;
    } else {
     $$0$i$idx = $$0$i$add;
    }
   }
   $53 = $$0$lcssa$i;
   $54 = $11;
   $55 = (($53) - ($54))|0;
   $56 = ($55|0)>(31);
   if ($56) {
    $$2 = -1;
   } else {
    $57 = (155824 + ($55)|0);
    $58 = HEAP8[$57>>0]|0;
    switch ($55|0) {
    case 24: case 25:  {
     $59 = HEAP32[$4>>2]|0;
     $60 = ($59|0)==($3|0);
     if (!($60)) {
      $61 = ((($59)) + -1|0);
      $62 = HEAP8[$61>>0]|0;
      $63 = $62 & 95;
      $64 = HEAP8[$2>>0]|0;
      $65 = $64 & 127;
      $66 = ($63<<24>>24)==($65<<24>>24);
      if (!($66)) {
       $$2 = -1;
       break L1;
      }
     }
     $67 = ((($59)) + 1|0);
     HEAP32[$4>>2] = $67;
     HEAP8[$59>>0] = $58;
     $$2 = 0;
     break L1;
     break;
    }
    case 23: case 22:  {
     HEAP8[$2>>0] = 80;
     $68 = HEAP32[$4>>2]|0;
     $69 = ((($68)) + 1|0);
     HEAP32[$4>>2] = $69;
     HEAP8[$68>>0] = $58;
     $$2 = 0;
     break L1;
     break;
    }
    default: {
     $70 = $58 & 95;
     $71 = $70&255;
     $72 = HEAP8[$2>>0]|0;
     $73 = $72 << 24 >> 24;
     $74 = ($71|0)==($73|0);
     if ($74) {
      $75 = $71 | 128;
      $76 = $75&255;
      HEAP8[$2>>0] = $76;
      $77 = HEAP8[$1>>0]|0;
      $78 = ($77<<24>>24)==(0);
      if (!($78)) {
       HEAP8[$1>>0] = 0;
       $79 = ((($7)) + 11|0);
       $80 = HEAP8[$79>>0]|0;
       $81 = ($80<<24>>24)<(0);
       $82 = ((($7)) + 4|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = $80&255;
       $85 = $81 ? $83 : $84;
       $86 = ($85|0)==(0);
       if (!($86)) {
        $87 = HEAP32[$9>>2]|0;
        $88 = $87;
        $89 = (($88) - ($13))|0;
        $90 = ($89|0)<(160);
        if ($90) {
         $91 = HEAP32[$10>>2]|0;
         $92 = ((($87)) + 4|0);
         HEAP32[$9>>2] = $92;
         HEAP32[$87>>2] = $91;
        }
       }
      }
     }
     $93 = HEAP32[$4>>2]|0;
     $94 = ((($93)) + 1|0);
     HEAP32[$4>>2] = $94;
     HEAP8[$93>>0] = $58;
     $95 = ($55|0)>(21);
     if ($95) {
      $$2 = 0;
      break L1;
     }
     $96 = HEAP32[$10>>2]|0;
     $97 = (($96) + 1)|0;
     HEAP32[$10>>2] = $97;
     $$2 = 0;
     break L1;
    }
    }
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNSt3__215__num_get_floatIeEET_PKcS3_Rj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0.0, $$0$ph = 0.0, $$1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($0|0)==($1|0);
 if ($4) {
  HEAP32[$2>>2] = 4;
  $$1 = 0.0;
 } else {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (___errno_location()|0);
  HEAP32[$7>>2] = 0;
  $8 = (__ZNSt3__26__clocEv()|0);
  $9 = (+_strtold_l($0,$3,$8));
  $10 = (___errno_location()|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = (___errno_location()|0);
   HEAP32[$13>>2] = $6;
  }
  $14 = HEAP32[$3>>2]|0;
  $15 = ($14|0)==($1|0);
  if ($15) {
   $16 = ($11|0)==(34);
   if ($16) {
    $$0$ph = $9;
    label = 6;
   } else {
    $$0 = $9;
   }
  } else {
   $$0$ph = 0.0;
   label = 6;
  }
  if ((label|0) == 6) {
   HEAP32[$2>>2] = 4;
   $$0 = $$0$ph;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNSt3__216__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$$030 = 0, $$0$i$i = 0, $$029 = 0, $$030 = 0, $$07$i$i = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $or$cond32 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 $7 = ((($0)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $5&255;
 $10 = $6 ? $8 : $9;
 $11 = ($10|0)==(0);
 do {
  if (!($11)) {
   $12 = ($1|0)==($2|0);
   if ($12) {
    $21 = $8;
   } else {
    $$0$i$i = $2;$$07$i$i = $1;
    while(1) {
     $13 = ((($$0$i$i)) + -4|0);
     $14 = ($$07$i$i>>>0)<($13>>>0);
     if (!($14)) {
      break;
     }
     $15 = HEAP32[$$07$i$i>>2]|0;
     $16 = HEAP32[$13>>2]|0;
     HEAP32[$$07$i$i>>2] = $16;
     HEAP32[$13>>2] = $15;
     $17 = ((($$07$i$i)) + 4|0);
     $$0$i$i = $13;$$07$i$i = $17;
    }
    $$pre = HEAP32[$7>>2]|0;
    $21 = $$pre;
   }
   $18 = HEAP32[$0>>2]|0;
   $19 = $6 ? $18 : $0;
   $20 = $6 ? $21 : $9;
   $22 = (($19) + ($20)|0);
   $23 = ((($2)) + -4|0);
   $24 = $22;
   $$029 = $1;$$030 = $19;
   while(1) {
    $25 = ($$029>>>0)<($23>>>0);
    $26 = HEAP8[$$030>>0]|0;
    $27 = ($26<<24>>24)>(0);
    $28 = ($26<<24>>24)!=(127);
    $or$cond32 = $27 & $28;
    if (!($25)) {
     break;
    }
    if ($or$cond32) {
     $29 = $26 << 24 >> 24;
     $30 = HEAP32[$$029>>2]|0;
     $31 = ($29|0)==($30|0);
     if (!($31)) {
      label = 10;
      break;
     }
    }
    $32 = $$030;
    $33 = (($24) - ($32))|0;
    $34 = ($33|0)>(1);
    $35 = ((($$030)) + 1|0);
    $$$030 = $34 ? $35 : $$030;
    $36 = ((($$029)) + 4|0);
    $$029 = $36;$$030 = $$$030;
   }
   if ((label|0) == 10) {
    HEAP32[$3>>2] = 4;
    break;
   }
   if ($or$cond32) {
    $37 = $26 << 24 >> 24;
    $38 = HEAP32[$23>>2]|0;
    $39 = (($38) + -1)|0;
    $40 = ($39>>>0)<($37>>>0);
    if (!($40)) {
     HEAP32[$3>>2] = 4;
    }
   }
  }
 } while(0);
 return;
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0;
 var $phitmp$i29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $6 = sp + 200|0;
 $7 = sp + 199|0;
 $8 = sp + 198|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 197|0;
 $16 = sp + 196|0;
 __ZNSt3__29__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$183 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP8[$36>>0]|0;
     $47 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$153 = $183;$184 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP8[$53>>0]|0;
      $64 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($184) {
       $185 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($184) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $185 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP8[$93>>0]|0;
    $104 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = $$0$i$i34&255;
   $106 = HEAP8[$7>>0]|0;
   $107 = HEAP8[$8>>0]|0;
   __THREW__ = 0;
   $108 = (invoke_iiiiiiiiiiiii(961,($105|0),($15|0),($16|0),($$119|0),($11|0),($106|0),($107|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $109 = __THREW__; __THREW__ = 0;
   $110 = $109&1;
   if ($110) {
    break;
   }
   $111 = ($108|0)==(0);
   if (!($111)) {
    $$2 = $$119;$$pre$i4$i$i46 = $185;
    label = 37;
    break;
   }
   $112 = HEAP32[$92>>2]|0;
   $113 = HEAP32[$94>>2]|0;
   $114 = ($112|0)==($113|0);
   if ($114) {
    $115 = HEAP32[$$pre$i$i$i40>>2]|0;
    $116 = ((($115)) + 40|0);
    $117 = HEAP32[$116>>2]|0;
    __THREW__ = 0;
    (invoke_ii($117|0,($$pre$i$i$i40|0))|0);
    $118 = __THREW__; __THREW__ = 0;
    $119 = $118&1;
    if ($119) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
     continue;
    }
   } else {
    $120 = ((($112)) + 1|0);
    HEAP32[$92>>2] = $120;
    $121 = HEAP8[$112>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($121)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $122 = ((($9)) + 11|0);
    $123 = HEAP8[$122>>0]|0;
    $124 = ($123<<24>>24)<(0);
    $125 = ((($9)) + 4|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = $123&255;
    $128 = $124 ? $126 : $127;
    $129 = ($128|0)==(0);
    $130 = HEAP8[$15>>0]|0;
    $131 = ($130<<24>>24)==(0);
    $or$cond = $129 | $131;
    if (!($or$cond)) {
     $132 = HEAP32[$13>>2]|0;
     $133 = $132;
     $134 = $12;
     $135 = (($133) - ($134))|0;
     $136 = ($135|0)<(160);
     if ($136) {
      $137 = HEAP32[$14>>2]|0;
      $138 = ((($132)) + 4|0);
      HEAP32[$13>>2] = $138;
      HEAP32[$132>>2] = $137;
     }
    }
    $139 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $140 = (+invoke_diii(964,($$2|0),($139|0),($4|0)));
    $141 = __THREW__; __THREW__ = 0;
    $142 = $141&1;
    if (!($142)) {
     HEAPF64[$5>>3] = $140;
     $143 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($143|0),($4|0));
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if (!($145)) {
      $146 = ($$pre$i$i$i40|0)==(0|0);
      if ($146) {
       $186 = 1;
      } else {
       $147 = ((($$pre$i$i$i40)) + 12|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ((($$pre$i$i$i40)) + 16|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ($148|0)==($150|0);
       if ($151) {
        $152 = HEAP32[$153>>2]|0;
        $154 = ((($152)) + 36|0);
        $155 = HEAP32[$154>>2]|0;
        __THREW__ = 0;
        $156 = (invoke_ii($155|0,($$pre$i$i$i40|0))|0);
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        } else {
         $$0$i$i$i$i38 = $156;
        }
       } else {
        $159 = HEAP8[$148>>0]|0;
        $160 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($159)|0);
        $$0$i$i$i$i38 = $160;
       }
       $161 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $162 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i38,$161)|0);
       if ($162) {
        HEAP32[$1>>2] = 0;
        $186 = 1;
       } else {
        $186 = 0;
       }
      }
      $163 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($163) {
        label = 54;
       } else {
        $164 = ((($$pre$i4$i$i46)) + 12|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = ((($$pre$i4$i$i46)) + 16|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = ($165|0)==($167|0);
        if ($168) {
         $169 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $170 = ((($169)) + 36|0);
         $171 = HEAP32[$170>>2]|0;
         __THREW__ = 0;
         $172 = (invoke_ii($171|0,($$pre$i4$i$i46|0))|0);
         $173 = __THREW__; __THREW__ = 0;
         $174 = $173&1;
         if ($174) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $172;
         }
        } else {
         $175 = HEAP8[$165>>0]|0;
         $176 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($175)|0);
         $$0$i$i2$i$i44 = $176;
        }
        $177 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $178 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i44,$177)|0);
        if ($178) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($186) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($186) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $179 = HEAP32[$4>>2]|0;
       $180 = $179 | 2;
       HEAP32[$4>>2] = $180;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  $$1 = $181;$$121 = $182;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__215__num_get_floatIdEET_PKcS3_Rj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0.0, $$0$ph = 0.0, $$1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($0|0)==($1|0);
 if ($4) {
  HEAP32[$2>>2] = 4;
  $$1 = 0.0;
 } else {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (___errno_location()|0);
  HEAP32[$7>>2] = 0;
  $8 = (__ZNSt3__26__clocEv()|0);
  $9 = (+_strtod_l($0,$3,$8));
  $10 = (___errno_location()|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = (___errno_location()|0);
   HEAP32[$13>>2] = $6;
  }
  $14 = HEAP32[$3>>2]|0;
  $15 = ($14|0)==($1|0);
  if ($15) {
   $16 = ($11|0)==(34);
   if ($16) {
    $$0$ph = $9;
    label = 6;
   } else {
    $$0 = $9;
   }
  } else {
   $$0$ph = 0.0;
   label = 6;
  }
  if ((label|0) == 6) {
   HEAP32[$2>>2] = 4;
   $$0 = $$0$ph;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0.0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0;
 var $phitmp$i29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $6 = sp + 200|0;
 $7 = sp + 199|0;
 $8 = sp + 198|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 197|0;
 $16 = sp + 196|0;
 __ZNSt3__29__num_getIcE19__stage2_float_prepERNS_8ios_baseEPcRcS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$183 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP8[$36>>0]|0;
     $47 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$153 = 0;$184 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$153 = $183;$184 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP8[$53>>0]|0;
      $64 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($184) {
       $185 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($184) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $185 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP8[$93>>0]|0;
    $104 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = $$0$i$i34&255;
   $106 = HEAP8[$7>>0]|0;
   $107 = HEAP8[$8>>0]|0;
   __THREW__ = 0;
   $108 = (invoke_iiiiiiiiiiiii(961,($105|0),($15|0),($16|0),($$119|0),($11|0),($106|0),($107|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $109 = __THREW__; __THREW__ = 0;
   $110 = $109&1;
   if ($110) {
    break;
   }
   $111 = ($108|0)==(0);
   if (!($111)) {
    $$2 = $$119;$$pre$i4$i$i46 = $185;
    label = 37;
    break;
   }
   $112 = HEAP32[$92>>2]|0;
   $113 = HEAP32[$94>>2]|0;
   $114 = ($112|0)==($113|0);
   if ($114) {
    $115 = HEAP32[$$pre$i$i$i40>>2]|0;
    $116 = ((($115)) + 40|0);
    $117 = HEAP32[$116>>2]|0;
    __THREW__ = 0;
    (invoke_ii($117|0,($$pre$i$i$i40|0))|0);
    $118 = __THREW__; __THREW__ = 0;
    $119 = $118&1;
    if ($119) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
     continue;
    }
   } else {
    $120 = ((($112)) + 1|0);
    HEAP32[$92>>2] = $120;
    $121 = HEAP8[$112>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($121)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$183 = $153;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $122 = ((($9)) + 11|0);
    $123 = HEAP8[$122>>0]|0;
    $124 = ($123<<24>>24)<(0);
    $125 = ((($9)) + 4|0);
    $126 = HEAP32[$125>>2]|0;
    $127 = $123&255;
    $128 = $124 ? $126 : $127;
    $129 = ($128|0)==(0);
    $130 = HEAP8[$15>>0]|0;
    $131 = ($130<<24>>24)==(0);
    $or$cond = $129 | $131;
    if (!($or$cond)) {
     $132 = HEAP32[$13>>2]|0;
     $133 = $132;
     $134 = $12;
     $135 = (($133) - ($134))|0;
     $136 = ($135|0)<(160);
     if ($136) {
      $137 = HEAP32[$14>>2]|0;
      $138 = ((($132)) + 4|0);
      HEAP32[$13>>2] = $138;
      HEAP32[$132>>2] = $137;
     }
    }
    $139 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $140 = (+invoke_diii(965,($$2|0),($139|0),($4|0)));
    $141 = __THREW__; __THREW__ = 0;
    $142 = $141&1;
    if (!($142)) {
     HEAPF32[$5>>2] = $140;
     $143 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($143|0),($4|0));
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if (!($145)) {
      $146 = ($$pre$i$i$i40|0)==(0|0);
      if ($146) {
       $186 = 1;
      } else {
       $147 = ((($$pre$i$i$i40)) + 12|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ((($$pre$i$i$i40)) + 16|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ($148|0)==($150|0);
       if ($151) {
        $152 = HEAP32[$153>>2]|0;
        $154 = ((($152)) + 36|0);
        $155 = HEAP32[$154>>2]|0;
        __THREW__ = 0;
        $156 = (invoke_ii($155|0,($$pre$i$i$i40|0))|0);
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        } else {
         $$0$i$i$i$i38 = $156;
        }
       } else {
        $159 = HEAP8[$148>>0]|0;
        $160 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($159)|0);
        $$0$i$i$i$i38 = $160;
       }
       $161 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $162 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i38,$161)|0);
       if ($162) {
        HEAP32[$1>>2] = 0;
        $186 = 1;
       } else {
        $186 = 0;
       }
      }
      $163 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($163) {
        label = 54;
       } else {
        $164 = ((($$pre$i4$i$i46)) + 12|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = ((($$pre$i4$i$i46)) + 16|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = ($165|0)==($167|0);
        if ($168) {
         $169 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $170 = ((($169)) + 36|0);
         $171 = HEAP32[$170>>2]|0;
         __THREW__ = 0;
         $172 = (invoke_ii($171|0,($$pre$i4$i$i46|0))|0);
         $173 = __THREW__; __THREW__ = 0;
         $174 = $173&1;
         if ($174) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $172;
         }
        } else {
         $175 = HEAP8[$165>>0]|0;
         $176 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($175)|0);
         $$0$i$i2$i$i44 = $176;
        }
        $177 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $178 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i44,$177)|0);
        if ($178) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($186) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($186) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $179 = HEAP32[$4>>2]|0;
       $180 = $179 | 2;
       HEAP32[$4>>2] = $180;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  $$1 = $181;$$121 = $182;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__215__num_get_floatIfEET_PKcS3_Rj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0.0, $$0$ph = 0.0, $$1 = 0.0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0.0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($0|0)==($1|0);
 if ($4) {
  HEAP32[$2>>2] = 4;
  $$1 = 0.0;
 } else {
  $5 = (___errno_location()|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = (___errno_location()|0);
  HEAP32[$7>>2] = 0;
  $8 = (__ZNSt3__26__clocEv()|0);
  $9 = (+_strtof_l($0,$3,$8));
  $10 = (___errno_location()|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $13 = (___errno_location()|0);
   HEAP32[$13>>2] = $6;
  }
  $14 = HEAP32[$3>>2]|0;
  $15 = ($14|0)==($1|0);
  if ($15) {
   $16 = ($11|0)==(34);
   if ($16) {
    $$0$ph = $9;
    label = 6;
   } else {
    $$0 = $9;
   }
  } else {
   $$0$ph = 0.0;
   label = 6;
  }
  if ((label|0) == 6) {
   HEAP32[$2>>2] = 4;
   $$0 = $$0$ph;
  }
  $$1 = $$0;
 }
 STACKTOP = sp;return (+$$1);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$183 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$153 = 0;$184 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$153 = 0;$184 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$153 = $183;$184 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($184) {
       $185 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($184) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $185 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $185;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$183 = $153;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$183 = $153;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(966,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = tempRet0;
    $137 = __THREW__; __THREW__ = 0;
    $138 = $137&1;
    if (!($138)) {
     $139 = $5;
     $140 = $139;
     HEAP32[$140>>2] = $135;
     $141 = (($139) + 4)|0;
     $142 = $141;
     HEAP32[$142>>2] = $136;
     $143 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($143|0),($4|0));
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if (!($145)) {
      $146 = ($$pre$i$i$i45|0)==(0|0);
      if ($146) {
       $186 = 1;
      } else {
       $147 = ((($$pre$i$i$i45)) + 12|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ((($$pre$i$i$i45)) + 16|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ($148|0)==($150|0);
       if ($151) {
        $152 = HEAP32[$153>>2]|0;
        $154 = ((($152)) + 36|0);
        $155 = HEAP32[$154>>2]|0;
        __THREW__ = 0;
        $156 = (invoke_ii($155|0,($$pre$i$i$i45|0))|0);
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        } else {
         $$0$i$i$i$i43 = $156;
        }
       } else {
        $159 = HEAP8[$148>>0]|0;
        $160 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($159)|0);
        $$0$i$i$i$i43 = $160;
       }
       $161 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $162 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$161)|0);
       if ($162) {
        HEAP32[$1>>2] = 0;
        $186 = 1;
       } else {
        $186 = 0;
       }
      }
      $163 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($163) {
        label = 54;
       } else {
        $164 = ((($$pre$i4$i$i51)) + 12|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = ((($$pre$i4$i$i51)) + 16|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = ($165|0)==($167|0);
        if ($168) {
         $169 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $170 = ((($169)) + 36|0);
         $171 = HEAP32[$170>>2]|0;
         __THREW__ = 0;
         $172 = (invoke_ii($171|0,($$pre$i4$i$i51|0))|0);
         $173 = __THREW__; __THREW__ = 0;
         $174 = $173&1;
         if ($174) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $172;
         }
        } else {
         $175 = HEAP8[$165>>0]|0;
         $176 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($175)|0);
         $$0$i$i2$i$i49 = $176;
        }
        $177 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $178 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$177)|0);
        if ($178) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($186) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($186) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $179 = HEAP32[$4>>2]|0;
       $180 = $179 | 2;
       HEAP32[$4>>2] = $180;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  $$1 = $181;$$127 = $182;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, $3 = 0, $trunc = 0, $trunc$clear = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = $2 & 74;
 $trunc = $3&255;
 $trunc$clear = $trunc & 127;
 switch ($trunc$clear<<24>>24) {
 case 64:  {
  $$0 = 8;
  break;
 }
 case 8:  {
  $$0 = 16;
  break;
 }
 case 0:  {
  $$0 = 0;
  break;
 }
 default: {
  $$0 = 10;
 }
 }
 return ($$0|0);
}
function __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 __ZNKSt3__28ios_base6getlocEv($4,$1);
 __THREW__ = 0;
 $5 = (invoke_iii(734,($4|0),(165440|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = HEAP32[$5>>2]|0;
  $9 = ((($8)) + 32|0);
  $10 = HEAP32[$9>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($10|0,($5|0),(155824|0),((155850)|0),($2|0))|0);
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if (!($12)) {
   __THREW__ = 0;
   $13 = (invoke_iii(734,($4|0),(165456|0))|0);
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if (!($15)) {
    $16 = HEAP32[$13>>2]|0;
    $17 = ((($16)) + 16|0);
    $18 = HEAP32[$17>>2]|0;
    __THREW__ = 0;
    $19 = (invoke_ii($18|0,($13|0))|0);
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     HEAP8[$3>>0] = $19;
     $22 = HEAP32[$13>>2]|0;
     $23 = ((($22)) + 20|0);
     $24 = HEAP32[$23>>2]|0;
     __THREW__ = 0;
     invoke_vii($24|0,($0|0),($13|0));
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if (!($26)) {
      __ZNSt3__26localeD2Ev($4);
      STACKTOP = sp;return;
     }
    }
   }
  }
 }
 $27 = ___cxa_find_matching_catch_2()|0;
 $28 = tempRet0;
 __ZNSt3__26localeD2Ev($4);
 ___resumeException($27|0);
 // unreachable;
}
function __ZNSt3__227__num_get_unsigned_integralIyEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 do {
  if ($5) {
   HEAP32[$2>>2] = 4;
   $21 = 0;$22 = 0;
  } else {
   $6 = HEAP8[$0>>0]|0;
   $7 = ($6<<24>>24)==(45);
   if ($7) {
    HEAP32[$2>>2] = 4;
    $21 = 0;$22 = 0;
    break;
   }
   $8 = (___errno_location()|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (___errno_location()|0);
   HEAP32[$10>>2] = 0;
   $11 = (__ZNSt3__26__clocEv()|0);
   $12 = (_strtoull_l($0,$4,$3,$11)|0);
   $13 = tempRet0;
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = (___errno_location()|0);
    HEAP32[$17>>2] = $9;
   }
   $18 = HEAP32[$4>>2]|0;
   $19 = ($18|0)==($1|0);
   if ($19) {
    $20 = ($15|0)==(34);
    if ($20) {
     $23 = -1;$24 = -1;
     label = 8;
    } else {
     $25 = $12;$26 = $13;
    }
   } else {
    $23 = 0;$24 = 0;
    label = 8;
   }
   if ((label|0) == 8) {
    HEAP32[$2>>2] = 4;
    $25 = $23;$26 = $24;
   }
   $21 = $26;$22 = $25;
  }
 } while(0);
 tempRet0 = ($21);
 STACKTOP = sp;return ($22|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$178 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$148 = $178;$179 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($179) {
       $180 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($179) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $180 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $180;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(967,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     HEAP32[$5>>2] = $135;
     $138 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($138|0),($4|0));
     $139 = __THREW__; __THREW__ = 0;
     $140 = $139&1;
     if (!($140)) {
      $141 = ($$pre$i$i$i45|0)==(0|0);
      if ($141) {
       $181 = 1;
      } else {
       $142 = ((($$pre$i$i$i45)) + 12|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ((($$pre$i$i$i45)) + 16|0);
       $145 = HEAP32[$144>>2]|0;
       $146 = ($143|0)==($145|0);
       if ($146) {
        $147 = HEAP32[$148>>2]|0;
        $149 = ((($147)) + 36|0);
        $150 = HEAP32[$149>>2]|0;
        __THREW__ = 0;
        $151 = (invoke_ii($150|0,($$pre$i$i$i45|0))|0);
        $152 = __THREW__; __THREW__ = 0;
        $153 = $152&1;
        if ($153) {
         break;
        } else {
         $$0$i$i$i$i43 = $151;
        }
       } else {
        $154 = HEAP8[$143>>0]|0;
        $155 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($154)|0);
        $$0$i$i$i$i43 = $155;
       }
       $156 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $157 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$156)|0);
       if ($157) {
        HEAP32[$1>>2] = 0;
        $181 = 1;
       } else {
        $181 = 0;
       }
      }
      $158 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($158) {
        label = 54;
       } else {
        $159 = ((($$pre$i4$i$i51)) + 12|0);
        $160 = HEAP32[$159>>2]|0;
        $161 = ((($$pre$i4$i$i51)) + 16|0);
        $162 = HEAP32[$161>>2]|0;
        $163 = ($160|0)==($162|0);
        if ($163) {
         $164 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $165 = ((($164)) + 36|0);
         $166 = HEAP32[$165>>2]|0;
         __THREW__ = 0;
         $167 = (invoke_ii($166|0,($$pre$i4$i$i51|0))|0);
         $168 = __THREW__; __THREW__ = 0;
         $169 = $168&1;
         if ($169) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $167;
         }
        } else {
         $170 = HEAP8[$160>>0]|0;
         $171 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($170)|0);
         $$0$i$i2$i$i49 = $171;
        }
        $172 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $173 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$172)|0);
        if ($173) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($181) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($181) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $174 = HEAP32[$4>>2]|0;
       $175 = $174 | 2;
       HEAP32[$4>>2] = $175;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $176 = ___cxa_find_matching_catch_2()|0;
  $177 = tempRet0;
  $$1 = $176;$$127 = $177;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__227__num_get_unsigned_integralImEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 do {
  if ($5) {
   HEAP32[$2>>2] = 4;
   $$1 = 0;
  } else {
   $6 = HEAP8[$0>>0]|0;
   $7 = ($6<<24>>24)==(45);
   if ($7) {
    HEAP32[$2>>2] = 4;
    $$1 = 0;
    break;
   }
   $8 = (___errno_location()|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (___errno_location()|0);
   HEAP32[$10>>2] = 0;
   $11 = (__ZNSt3__26__clocEv()|0);
   $12 = (_strtoull_l($0,$4,$3,$11)|0);
   $13 = tempRet0;
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = (___errno_location()|0);
    HEAP32[$17>>2] = $9;
   }
   $18 = HEAP32[$4>>2]|0;
   $19 = ($18|0)==($1|0);
   do {
    if ($19) {
     $20 = ($15|0)==(34);
     $21 = ($13>>>0)>(0);
     $22 = ($12>>>0)>(4294967295);
     $23 = ($13|0)==(0);
     $24 = $23 & $22;
     $25 = $21 | $24;
     $or$cond = $25 | $20;
     if ($or$cond) {
      HEAP32[$2>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $$0 = $12;
      break;
     }
    } else {
     HEAP32[$2>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$178 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$148 = $178;$179 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($179) {
       $180 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($179) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $180 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $180;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(968,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     HEAP32[$5>>2] = $135;
     $138 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($138|0),($4|0));
     $139 = __THREW__; __THREW__ = 0;
     $140 = $139&1;
     if (!($140)) {
      $141 = ($$pre$i$i$i45|0)==(0|0);
      if ($141) {
       $181 = 1;
      } else {
       $142 = ((($$pre$i$i$i45)) + 12|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ((($$pre$i$i$i45)) + 16|0);
       $145 = HEAP32[$144>>2]|0;
       $146 = ($143|0)==($145|0);
       if ($146) {
        $147 = HEAP32[$148>>2]|0;
        $149 = ((($147)) + 36|0);
        $150 = HEAP32[$149>>2]|0;
        __THREW__ = 0;
        $151 = (invoke_ii($150|0,($$pre$i$i$i45|0))|0);
        $152 = __THREW__; __THREW__ = 0;
        $153 = $152&1;
        if ($153) {
         break;
        } else {
         $$0$i$i$i$i43 = $151;
        }
       } else {
        $154 = HEAP8[$143>>0]|0;
        $155 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($154)|0);
        $$0$i$i$i$i43 = $155;
       }
       $156 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $157 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$156)|0);
       if ($157) {
        HEAP32[$1>>2] = 0;
        $181 = 1;
       } else {
        $181 = 0;
       }
      }
      $158 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($158) {
        label = 54;
       } else {
        $159 = ((($$pre$i4$i$i51)) + 12|0);
        $160 = HEAP32[$159>>2]|0;
        $161 = ((($$pre$i4$i$i51)) + 16|0);
        $162 = HEAP32[$161>>2]|0;
        $163 = ($160|0)==($162|0);
        if ($163) {
         $164 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $165 = ((($164)) + 36|0);
         $166 = HEAP32[$165>>2]|0;
         __THREW__ = 0;
         $167 = (invoke_ii($166|0,($$pre$i4$i$i51|0))|0);
         $168 = __THREW__; __THREW__ = 0;
         $169 = $168&1;
         if ($169) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $167;
         }
        } else {
         $170 = HEAP8[$160>>0]|0;
         $171 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($170)|0);
         $$0$i$i2$i$i49 = $171;
        }
        $172 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $173 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$172)|0);
        if ($173) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($181) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($181) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $174 = HEAP32[$4>>2]|0;
       $175 = $174 | 2;
       HEAP32[$4>>2] = $175;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $176 = ___cxa_find_matching_catch_2()|0;
  $177 = tempRet0;
  $$1 = $176;$$127 = $177;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__227__num_get_unsigned_integralIjEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 do {
  if ($5) {
   HEAP32[$2>>2] = 4;
   $$1 = 0;
  } else {
   $6 = HEAP8[$0>>0]|0;
   $7 = ($6<<24>>24)==(45);
   if ($7) {
    HEAP32[$2>>2] = 4;
    $$1 = 0;
    break;
   }
   $8 = (___errno_location()|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (___errno_location()|0);
   HEAP32[$10>>2] = 0;
   $11 = (__ZNSt3__26__clocEv()|0);
   $12 = (_strtoull_l($0,$4,$3,$11)|0);
   $13 = tempRet0;
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = (___errno_location()|0);
    HEAP32[$17>>2] = $9;
   }
   $18 = HEAP32[$4>>2]|0;
   $19 = ($18|0)==($1|0);
   do {
    if ($19) {
     $20 = ($15|0)==(34);
     $21 = ($13>>>0)>(0);
     $22 = ($12>>>0)>(4294967295);
     $23 = ($13|0)==(0);
     $24 = $23 & $22;
     $25 = $21 | $24;
     $or$cond = $25 | $20;
     if ($or$cond) {
      HEAP32[$2>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $$0 = $12;
      break;
     }
    } else {
     HEAP32[$2>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$178 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$148 = $178;$179 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($179) {
       $180 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($179) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $180 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $180;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(969,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     HEAP16[$5>>1] = $135;
     $138 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($138|0),($4|0));
     $139 = __THREW__; __THREW__ = 0;
     $140 = $139&1;
     if (!($140)) {
      $141 = ($$pre$i$i$i45|0)==(0|0);
      if ($141) {
       $181 = 1;
      } else {
       $142 = ((($$pre$i$i$i45)) + 12|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ((($$pre$i$i$i45)) + 16|0);
       $145 = HEAP32[$144>>2]|0;
       $146 = ($143|0)==($145|0);
       if ($146) {
        $147 = HEAP32[$148>>2]|0;
        $149 = ((($147)) + 36|0);
        $150 = HEAP32[$149>>2]|0;
        __THREW__ = 0;
        $151 = (invoke_ii($150|0,($$pre$i$i$i45|0))|0);
        $152 = __THREW__; __THREW__ = 0;
        $153 = $152&1;
        if ($153) {
         break;
        } else {
         $$0$i$i$i$i43 = $151;
        }
       } else {
        $154 = HEAP8[$143>>0]|0;
        $155 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($154)|0);
        $$0$i$i$i$i43 = $155;
       }
       $156 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $157 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$156)|0);
       if ($157) {
        HEAP32[$1>>2] = 0;
        $181 = 1;
       } else {
        $181 = 0;
       }
      }
      $158 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($158) {
        label = 54;
       } else {
        $159 = ((($$pre$i4$i$i51)) + 12|0);
        $160 = HEAP32[$159>>2]|0;
        $161 = ((($$pre$i4$i$i51)) + 16|0);
        $162 = HEAP32[$161>>2]|0;
        $163 = ($160|0)==($162|0);
        if ($163) {
         $164 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $165 = ((($164)) + 36|0);
         $166 = HEAP32[$165>>2]|0;
         __THREW__ = 0;
         $167 = (invoke_ii($166|0,($$pre$i4$i$i51|0))|0);
         $168 = __THREW__; __THREW__ = 0;
         $169 = $168&1;
         if ($169) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $167;
         }
        } else {
         $170 = HEAP8[$160>>0]|0;
         $171 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($170)|0);
         $$0$i$i2$i$i49 = $171;
        }
        $172 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $173 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$172)|0);
        if ($173) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($181) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($181) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $174 = HEAP32[$4>>2]|0;
       $175 = $174 | 2;
       HEAP32[$4>>2] = $175;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $176 = ___cxa_find_matching_catch_2()|0;
  $177 = tempRet0;
  $$1 = $176;$$127 = $177;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__227__num_get_unsigned_integralItEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 do {
  if ($5) {
   HEAP32[$2>>2] = 4;
   $$1 = 0;
  } else {
   $6 = HEAP8[$0>>0]|0;
   $7 = ($6<<24>>24)==(45);
   if ($7) {
    HEAP32[$2>>2] = 4;
    $$1 = 0;
    break;
   }
   $8 = (___errno_location()|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (___errno_location()|0);
   HEAP32[$10>>2] = 0;
   $11 = (__ZNSt3__26__clocEv()|0);
   $12 = (_strtoull_l($0,$4,$3,$11)|0);
   $13 = tempRet0;
   $14 = (___errno_location()|0);
   $15 = HEAP32[$14>>2]|0;
   $16 = ($15|0)==(0);
   if ($16) {
    $17 = (___errno_location()|0);
    HEAP32[$17>>2] = $9;
   }
   $18 = HEAP32[$4>>2]|0;
   $19 = ($18|0)==($1|0);
   do {
    if ($19) {
     $20 = ($15|0)==(34);
     $21 = ($13>>>0)>(0);
     $22 = ($12>>>0)>(65535);
     $23 = ($13|0)==(0);
     $24 = $23 & $22;
     $25 = $21 | $24;
     $or$cond = $25 | $20;
     if ($or$cond) {
      HEAP32[$2>>2] = 4;
      $$0 = -1;
      break;
     } else {
      $26 = $12&65535;
      $$0 = $26;
      break;
     }
    } else {
     HEAP32[$2>>2] = 4;
     $$0 = 0;
    }
   } while(0);
   $$1 = $$0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$183 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$153 = 0;$184 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$153 = 0;$184 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$153 = $183;$184 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($184) {
       $185 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($184) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $185 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $185;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$183 = $153;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$183 = $153;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(970,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = tempRet0;
    $137 = __THREW__; __THREW__ = 0;
    $138 = $137&1;
    if (!($138)) {
     $139 = $5;
     $140 = $139;
     HEAP32[$140>>2] = $135;
     $141 = (($139) + 4)|0;
     $142 = $141;
     HEAP32[$142>>2] = $136;
     $143 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($143|0),($4|0));
     $144 = __THREW__; __THREW__ = 0;
     $145 = $144&1;
     if (!($145)) {
      $146 = ($$pre$i$i$i45|0)==(0|0);
      if ($146) {
       $186 = 1;
      } else {
       $147 = ((($$pre$i$i$i45)) + 12|0);
       $148 = HEAP32[$147>>2]|0;
       $149 = ((($$pre$i$i$i45)) + 16|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ($148|0)==($150|0);
       if ($151) {
        $152 = HEAP32[$153>>2]|0;
        $154 = ((($152)) + 36|0);
        $155 = HEAP32[$154>>2]|0;
        __THREW__ = 0;
        $156 = (invoke_ii($155|0,($$pre$i$i$i45|0))|0);
        $157 = __THREW__; __THREW__ = 0;
        $158 = $157&1;
        if ($158) {
         break;
        } else {
         $$0$i$i$i$i43 = $156;
        }
       } else {
        $159 = HEAP8[$148>>0]|0;
        $160 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($159)|0);
        $$0$i$i$i$i43 = $160;
       }
       $161 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $162 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$161)|0);
       if ($162) {
        HEAP32[$1>>2] = 0;
        $186 = 1;
       } else {
        $186 = 0;
       }
      }
      $163 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($163) {
        label = 54;
       } else {
        $164 = ((($$pre$i4$i$i51)) + 12|0);
        $165 = HEAP32[$164>>2]|0;
        $166 = ((($$pre$i4$i$i51)) + 16|0);
        $167 = HEAP32[$166>>2]|0;
        $168 = ($165|0)==($167|0);
        if ($168) {
         $169 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $170 = ((($169)) + 36|0);
         $171 = HEAP32[$170>>2]|0;
         __THREW__ = 0;
         $172 = (invoke_ii($171|0,($$pre$i4$i$i51|0))|0);
         $173 = __THREW__; __THREW__ = 0;
         $174 = $173&1;
         if ($174) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $172;
         }
        } else {
         $175 = HEAP8[$165>>0]|0;
         $176 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($175)|0);
         $$0$i$i2$i$i49 = $176;
        }
        $177 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $178 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$177)|0);
        if ($178) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($186) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($186) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $179 = HEAP32[$4>>2]|0;
       $180 = $179 | 2;
       HEAP32[$4>>2] = $180;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $181 = ___cxa_find_matching_catch_2()|0;
  $182 = tempRet0;
  $$1 = $181;$$127 = $182;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__225__num_get_signed_integralIxEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 if ($5) {
  HEAP32[$2>>2] = 4;
  $26 = 0;$27 = 0;
 } else {
  $6 = (___errno_location()|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (___errno_location()|0);
  HEAP32[$8>>2] = 0;
  $9 = (__ZNSt3__26__clocEv()|0);
  $10 = (_strtoll_l($0,$4,$3,$9)|0);
  $11 = tempRet0;
  $12 = (___errno_location()|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0);
  if ($14) {
   $15 = (___errno_location()|0);
   HEAP32[$15>>2] = $7;
  }
  $16 = HEAP32[$4>>2]|0;
  $17 = ($16|0)==($1|0);
  if ($17) {
   $18 = ($13|0)==(34);
   if ($18) {
    HEAP32[$2>>2] = 4;
    $19 = ($11|0)>(0);
    $20 = ($10>>>0)>(0);
    $21 = ($11|0)==(0);
    $22 = $21 & $20;
    $23 = $19 | $22;
    $24 = $23 ? -1 : 0;
    $25 = $23 ? 2147483647 : -2147483648;
    $28 = $24;$29 = $25;
   } else {
    $28 = $10;$29 = $11;
   }
  } else {
   HEAP32[$2>>2] = 4;
   $28 = 0;$29 = 0;
  }
  $26 = $29;$27 = $28;
 }
 tempRet0 = ($26);
 STACKTOP = sp;return ($27|0);
}
function __ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 198|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIcE17__stage2_int_prepERNS_8ios_baseEPcRc($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$178 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP8[$34>>0]|0;
     $45 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$148 = 0;$179 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$148 = $178;$179 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP8[$51>>0]|0;
      $62 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($179) {
       $180 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($179) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $180 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP8[$91>>0]|0;
    $102 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = $$0$i$i39&255;
   $104 = HEAP8[$7>>0]|0;
   __THREW__ = 0;
   $105 = (invoke_iiiiiiiiiii(954,($103|0),($14|0),($$125|0),($10|0),($13|0),($104|0),($8|0),($11|0),($12|0),($6|0))|0);
   $106 = __THREW__; __THREW__ = 0;
   $107 = $106&1;
   if ($107) {
    break;
   }
   $108 = ($105|0)==(0);
   if (!($108)) {
    $$2 = $$125;$$pre$i4$i$i51 = $180;
    label = 37;
    break;
   }
   $109 = HEAP32[$90>>2]|0;
   $110 = HEAP32[$92>>2]|0;
   $111 = ($109|0)==($110|0);
   if ($111) {
    $112 = HEAP32[$$pre$i$i$i45>>2]|0;
    $113 = ((($112)) + 40|0);
    $114 = HEAP32[$113>>2]|0;
    __THREW__ = 0;
    (invoke_ii($114|0,($$pre$i$i$i45|0))|0);
    $115 = __THREW__; __THREW__ = 0;
    $116 = $115&1;
    if ($116) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
     continue;
    }
   } else {
    $117 = ((($109)) + 1|0);
    HEAP32[$90>>2] = $117;
    $118 = HEAP8[$109>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($118)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$178 = $148;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $119 = ((($8)) + 11|0);
    $120 = HEAP8[$119>>0]|0;
    $121 = ($120<<24>>24)<(0);
    $122 = ((($8)) + 4|0);
    $123 = HEAP32[$122>>2]|0;
    $124 = $120&255;
    $125 = $121 ? $123 : $124;
    $126 = ($125|0)==(0);
    if (!($126)) {
     $127 = HEAP32[$12>>2]|0;
     $128 = $127;
     $129 = $11;
     $130 = (($128) - ($129))|0;
     $131 = ($130|0)<(160);
     if ($131) {
      $132 = HEAP32[$13>>2]|0;
      $133 = ((($127)) + 4|0);
      HEAP32[$12>>2] = $133;
      HEAP32[$127>>2] = $132;
     }
    }
    $134 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $135 = (invoke_iiiii(971,($$2|0),($134|0),($4|0),($14|0))|0);
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     HEAP32[$5>>2] = $135;
     $138 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($138|0),($4|0));
     $139 = __THREW__; __THREW__ = 0;
     $140 = $139&1;
     if (!($140)) {
      $141 = ($$pre$i$i$i45|0)==(0|0);
      if ($141) {
       $181 = 1;
      } else {
       $142 = ((($$pre$i$i$i45)) + 12|0);
       $143 = HEAP32[$142>>2]|0;
       $144 = ((($$pre$i$i$i45)) + 16|0);
       $145 = HEAP32[$144>>2]|0;
       $146 = ($143|0)==($145|0);
       if ($146) {
        $147 = HEAP32[$148>>2]|0;
        $149 = ((($147)) + 36|0);
        $150 = HEAP32[$149>>2]|0;
        __THREW__ = 0;
        $151 = (invoke_ii($150|0,($$pre$i$i$i45|0))|0);
        $152 = __THREW__; __THREW__ = 0;
        $153 = $152&1;
        if ($153) {
         break;
        } else {
         $$0$i$i$i$i43 = $151;
        }
       } else {
        $154 = HEAP8[$143>>0]|0;
        $155 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($154)|0);
        $$0$i$i$i$i43 = $155;
       }
       $156 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $157 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i43,$156)|0);
       if ($157) {
        HEAP32[$1>>2] = 0;
        $181 = 1;
       } else {
        $181 = 0;
       }
      }
      $158 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($158) {
        label = 54;
       } else {
        $159 = ((($$pre$i4$i$i51)) + 12|0);
        $160 = HEAP32[$159>>2]|0;
        $161 = ((($$pre$i4$i$i51)) + 16|0);
        $162 = HEAP32[$161>>2]|0;
        $163 = ($160|0)==($162|0);
        if ($163) {
         $164 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $165 = ((($164)) + 36|0);
         $166 = HEAP32[$165>>2]|0;
         __THREW__ = 0;
         $167 = (invoke_ii($166|0,($$pre$i4$i$i51|0))|0);
         $168 = __THREW__; __THREW__ = 0;
         $169 = $168&1;
         if ($169) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $167;
         }
        } else {
         $170 = HEAP8[$160>>0]|0;
         $171 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($170)|0);
         $$0$i$i2$i$i49 = $171;
        }
        $172 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $173 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i49,$172)|0);
        if ($173) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($181) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($181) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $174 = HEAP32[$4>>2]|0;
       $175 = $174 | 2;
       HEAP32[$4>>2] = $175;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $176 = ___cxa_find_matching_catch_2()|0;
  $177 = tempRet0;
  $$1 = $176;$$127 = $177;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__225__num_get_signed_integralIlEET_PKcS3_Rji($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ($0|0)==($1|0);
 if ($5) {
  HEAP32[$2>>2] = 4;
  $$1 = 0;
 } else {
  $6 = (___errno_location()|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = (___errno_location()|0);
  HEAP32[$8>>2] = 0;
  $9 = (__ZNSt3__26__clocEv()|0);
  $10 = (_strtoll_l($0,$4,$3,$9)|0);
  $11 = tempRet0;
  $12 = (___errno_location()|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = ($13|0)==(0);
  if ($14) {
   $15 = (___errno_location()|0);
   HEAP32[$15>>2] = $7;
  }
  $16 = HEAP32[$4>>2]|0;
  $17 = ($16|0)==($1|0);
  L7: do {
   if ($17) {
    $18 = ($13|0)==(34);
    do {
     if ($18) {
      HEAP32[$2>>2] = 4;
      $29 = ($11|0)>(0);
      $30 = ($10>>>0)>(0);
      $31 = ($11|0)==(0);
      $32 = $31 & $30;
      $33 = $29 | $32;
      if ($33) {
       $$0 = 2147483647;
       break L7;
      }
     } else {
      $19 = ($11|0)<(-1);
      $20 = ($10>>>0)<(2147483648);
      $21 = ($11|0)==(-1);
      $22 = $21 & $20;
      $23 = $19 | $22;
      if ($23) {
       HEAP32[$2>>2] = 4;
       break;
      }
      $24 = ($11|0)>(0);
      $25 = ($10>>>0)>(2147483647);
      $26 = ($11|0)==(0);
      $27 = $26 & $25;
      $28 = $24 | $27;
      if ($28) {
       HEAP32[$2>>2] = 4;
       $$0 = 2147483647;
       break L7;
      } else {
       $$0 = $10;
       break L7;
      }
     }
    } while(0);
    $$0 = -2147483648;
   } else {
    HEAP32[$2>>2] = 4;
    $$0 = 0;
   }
  } while(0);
  $$1 = $$0;
 }
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i137 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i143 = 0, $$0100 = 0, $$0101 = 0, $$0102 = 0, $$0103 = 0, $$0106 = 0, $$0110 = 0, $$0126 = 0, $$0127 = 0, $$0127191 = 0, $$097 = 0, $$098 = 0, $$099$off0 = 0, $$1104 = 0, $$1107 = 0;
 var $$1111 = 0, $$2$off0 = 0, $$2105 = 0, $$2108 = 0, $$2112 = 0, $$3 = 0, $$3109 = 0, $$3113 = 0, $$5 = 0, $$5115 = 0, $$6 = 0, $$7 = 0, $$pre$i$i$i = 0, $$pre$i$i$i139 = 0, $$pre$i4$i$i145 = 0, $$sink134$sink135199$index = 0, $$sink134$sink135199$index5 = 0, $$sroa$0159$0 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i140 = 0, $tmp$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $7 = sp;
 $8 = $3;
 $9 = $2;
 $10 = (($8) - ($9))|0;
 $11 = (($10|0) / 12)&-1;
 $12 = ($11>>>0)>(100);
 do {
  if ($12) {
   $13 = (_malloc($11)|0);
   $14 = ($13|0)==(0|0);
   if ($14) {
    __THREW__ = 0;
    invoke_v(732);
    $15 = __THREW__; __THREW__ = 0;
    $16 = ___cxa_find_matching_catch_2()|0;
    $17 = tempRet0;
    $$sink134$sink135199$index = $16;$$sink134$sink135199$index5 = $17;
    break;
   } else {
    $18 = $13;
    $$0126 = $13;$$sroa$0159$0 = $18;
    label = 5;
    break;
   }
  } else {
   $$0126 = $7;$$sroa$0159$0 = 0;
   label = 5;
  }
 } while(0);
 if ((label|0) == 5) {
  $$0102 = $2;$$0103 = $$0126;$$0106 = 0;$$0110 = $11;
  while(1) {
   $19 = ($$0102|0)==($3|0);
   if ($19) {
    break;
   }
   $20 = ((($$0102)) + 11|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = ($21<<24>>24)<(0);
   if ($22) {
    $23 = ((($$0102)) + 4|0);
    $24 = HEAP32[$23>>2]|0;
    $27 = $24;
   } else {
    $25 = $21&255;
    $27 = $25;
   }
   $26 = ($27|0)==(0);
   if ($26) {
    HEAP8[$$0103>>0] = 2;
    $28 = (($$0110) + -1)|0;
    $29 = (($$0106) + 1)|0;
    $$1107 = $29;$$1111 = $28;
   } else {
    HEAP8[$$0103>>0] = 1;
    $$1107 = $$0106;$$1111 = $$0110;
   }
   $30 = ((($$0102)) + 12|0);
   $31 = ((($$0103)) + 1|0);
   $$0102 = $30;$$0103 = $31;$$0106 = $$1107;$$0110 = $$1111;
  }
  $$0101 = 0;$$2108 = $$0106;$$2112 = $$0110;
  L20: while(1) {
   $32 = HEAP32[$0>>2]|0;
   $33 = ($32|0)==(0|0);
   do {
    if ($33) {
     $66 = 1;
    } else {
     $34 = ((($32)) + 12|0);
     $35 = HEAP32[$34>>2]|0;
     $36 = ((($32)) + 16|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ($35|0)==($37|0);
     if ($38) {
      $39 = HEAP32[$32>>2]|0;
      $40 = ((($39)) + 36|0);
      $41 = HEAP32[$40>>2]|0;
      __THREW__ = 0;
      $42 = (invoke_ii($41|0,($32|0))|0);
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       break L20;
      } else {
       $$0$i$i$i$i = $42;
      }
     } else {
      $45 = HEAP8[$35>>0]|0;
      $46 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($45)|0);
      $$0$i$i$i$i = $46;
     }
     $47 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $48 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$47)|0);
     if ($48) {
      HEAP32[$0>>2] = 0;
      $66 = 1;
      break;
     } else {
      $$pre$i$i$i = HEAP32[$0>>2]|0;
      $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
      $66 = $phitmp$i$i$i;
      break;
     }
    }
   } while(0);
   $49 = HEAP32[$1>>2]|0;
   $50 = ($49|0)==(0|0);
   if ($50) {
    $$pre$i4$i$i145 = 0;$67 = 1;
   } else {
    $51 = ((($49)) + 12|0);
    $52 = HEAP32[$51>>2]|0;
    $53 = ((($49)) + 16|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ($52|0)==($54|0);
    if ($55) {
     $56 = HEAP32[$49>>2]|0;
     $57 = ((($56)) + 36|0);
     $58 = HEAP32[$57>>2]|0;
     __THREW__ = 0;
     $59 = (invoke_ii($58|0,($49|0))|0);
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      break;
     } else {
      $$0$i$i2$i$i = $59;
     }
    } else {
     $62 = HEAP8[$52>>0]|0;
     $63 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($62)|0);
     $$0$i$i2$i$i = $63;
    }
    $64 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $65 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$64)|0);
    if ($65) {
     HEAP32[$1>>2] = 0;
     $$pre$i4$i$i145 = 0;$67 = 1;
    } else {
     $$pre$i4$i$i145 = $49;$67 = 0;
    }
   }
   $tmp$i$i = $66 ^ $67;
   $68 = ($$2112|0)!=(0);
   $69 = $68 & $tmp$i$i;
   $70 = HEAP32[$0>>2]|0;
   if (!($69)) {
    label = 29;
    break;
   }
   $103 = ((($70)) + 12|0);
   $104 = HEAP32[$103>>2]|0;
   $105 = ((($70)) + 16|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = ($104|0)==($106|0);
   if ($107) {
    $108 = HEAP32[$70>>2]|0;
    $109 = ((($108)) + 36|0);
    $110 = HEAP32[$109>>2]|0;
    __THREW__ = 0;
    $111 = (invoke_ii($110|0,($70|0))|0);
    $112 = __THREW__; __THREW__ = 0;
    $113 = $112&1;
    if ($113) {
     break;
    } else {
     $$0$i$i = $111;
    }
   } else {
    $114 = HEAP8[$104>>0]|0;
    $115 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($114)|0);
    $$0$i$i = $115;
   }
   $116 = $$0$i$i&255;
   if ($6) {
    $$0100 = $116;
   } else {
    $117 = HEAP32[$4>>2]|0;
    $118 = ((($117)) + 12|0);
    $119 = HEAP32[$118>>2]|0;
    __THREW__ = 0;
    $120 = (invoke_iii($119|0,($4|0),($116|0))|0);
    $121 = __THREW__; __THREW__ = 0;
    $122 = $121&1;
    if ($122) {
     break;
    } else {
     $$0100 = $120;
    }
   }
   $123 = (($$0101) + 1)|0;
   $$098 = $2;$$099$off0 = 0;$$1104 = $$0126;$$3109 = $$2108;$$3113 = $$2112;
   while(1) {
    $124 = ($$098|0)==($3|0);
    if ($124) {
     break;
    }
    $125 = HEAP8[$$1104>>0]|0;
    $126 = ($125<<24>>24)==(1);
    do {
     if ($126) {
      $127 = ((($$098)) + 11|0);
      $128 = HEAP8[$127>>0]|0;
      $129 = ($128<<24>>24)<(0);
      if ($129) {
       $130 = HEAP32[$$098>>2]|0;
       $132 = $130;
      } else {
       $132 = $$098;
      }
      $131 = (($132) + ($$0101)|0);
      $133 = HEAP8[$131>>0]|0;
      if ($6) {
       $$097 = $133;
      } else {
       $134 = HEAP32[$4>>2]|0;
       $135 = ((($134)) + 12|0);
       $136 = HEAP32[$135>>2]|0;
       __THREW__ = 0;
       $137 = (invoke_iii($136|0,($4|0),($133|0))|0);
       $138 = __THREW__; __THREW__ = 0;
       $139 = $138&1;
       if ($139) {
        break L20;
       } else {
        $$097 = $137;
       }
      }
      $140 = ($$0100<<24>>24)==($$097<<24>>24);
      $141 = (($$3113) + -1)|0;
      if (!($140)) {
       HEAP8[$$1104>>0] = 0;
       $$2$off0 = $$099$off0;$$5 = $$3109;$$5115 = $141;
       break;
      }
      $142 = HEAP8[$127>>0]|0;
      $143 = ($142<<24>>24)<(0);
      if ($143) {
       $144 = ((($$098)) + 4|0);
       $145 = HEAP32[$144>>2]|0;
       $148 = $145;
      } else {
       $146 = $142&255;
       $148 = $146;
      }
      $147 = ($148|0)==($123|0);
      if ($147) {
       $149 = (($$3109) + 1)|0;
       HEAP8[$$1104>>0] = 2;
       $$2$off0 = 1;$$5 = $149;$$5115 = $141;
      } else {
       $$2$off0 = 1;$$5 = $$3109;$$5115 = $$3113;
      }
     } else {
      $$2$off0 = $$099$off0;$$5 = $$3109;$$5115 = $$3113;
     }
    } while(0);
    $150 = ((($$098)) + 12|0);
    $151 = ((($$1104)) + 1|0);
    $$098 = $150;$$099$off0 = $$2$off0;$$1104 = $151;$$3109 = $$5;$$3113 = $$5115;
   }
   if (!($$099$off0)) {
    $$0101 = $123;$$2108 = $$3109;$$2112 = $$3113;
    continue;
   }
   $152 = HEAP32[$0>>2]|0;
   $153 = ((($152)) + 12|0);
   $154 = HEAP32[$153>>2]|0;
   $155 = ((($152)) + 16|0);
   $156 = HEAP32[$155>>2]|0;
   $157 = ($154|0)==($156|0);
   if ($157) {
    $158 = HEAP32[$152>>2]|0;
    $159 = ((($158)) + 40|0);
    $160 = HEAP32[$159>>2]|0;
    __THREW__ = 0;
    (invoke_ii($160|0,($152|0))|0);
    $161 = __THREW__; __THREW__ = 0;
    $162 = $161&1;
    if ($162) {
     break;
    }
   } else {
    $163 = ((($154)) + 1|0);
    HEAP32[$153>>2] = $163;
    $164 = HEAP8[$154>>0]|0;
    (__ZNSt3__211char_traitsIcE11to_int_typeEc($164)|0);
   }
   $165 = (($$3109) + ($$3113))|0;
   $166 = ($165>>>0)>(1);
   if ($166) {
    $$0 = $2;$$2105 = $$0126;$$6 = $$3109;
   } else {
    $$0101 = $123;$$2108 = $$3109;$$2112 = $$3113;
    continue;
   }
   while(1) {
    $167 = ($$0|0)==($3|0);
    if ($167) {
     $$0101 = $123;$$2108 = $$6;$$2112 = $$3113;
     continue L20;
    }
    $168 = HEAP8[$$2105>>0]|0;
    $169 = ($168<<24>>24)==(2);
    if ($169) {
     $170 = ((($$0)) + 11|0);
     $171 = HEAP8[$170>>0]|0;
     $172 = ($171<<24>>24)<(0);
     if ($172) {
      $173 = ((($$0)) + 4|0);
      $174 = HEAP32[$173>>2]|0;
      $177 = $174;
     } else {
      $175 = $171&255;
      $177 = $175;
     }
     $176 = ($177|0)==($123|0);
     if ($176) {
      $$7 = $$6;
     } else {
      $178 = (($$6) + -1)|0;
      HEAP8[$$2105>>0] = 0;
      $$7 = $178;
     }
    } else {
     $$7 = $$6;
    }
    $179 = ((($$0)) + 12|0);
    $180 = ((($$2105)) + 1|0);
    $$0 = $179;$$2105 = $180;$$6 = $$7;
   }
  }
  L86: do {
   if ((label|0) == 29) {
    $71 = ($70|0)==(0|0);
    do {
     if ($71) {
      $196 = 1;
     } else {
      $72 = ((($70)) + 12|0);
      $73 = HEAP32[$72>>2]|0;
      $74 = ((($70)) + 16|0);
      $75 = HEAP32[$74>>2]|0;
      $76 = ($73|0)==($75|0);
      if ($76) {
       $77 = HEAP32[$70>>2]|0;
       $78 = ((($77)) + 36|0);
       $79 = HEAP32[$78>>2]|0;
       __THREW__ = 0;
       $80 = (invoke_ii($79|0,($70|0))|0);
       $81 = __THREW__; __THREW__ = 0;
       $82 = $81&1;
       if ($82) {
        break L86;
       } else {
        $$0$i$i$i$i137 = $80;
       }
      } else {
       $83 = HEAP8[$73>>0]|0;
       $84 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($83)|0);
       $$0$i$i$i$i137 = $84;
      }
      $85 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $86 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i137,$85)|0);
      if ($86) {
       HEAP32[$0>>2] = 0;
       $196 = 1;
       break;
      } else {
       $$pre$i$i$i139 = HEAP32[$0>>2]|0;
       $phitmp$i$i$i140 = ($$pre$i$i$i139|0)==(0|0);
       $196 = $phitmp$i$i$i140;
       break;
      }
     }
    } while(0);
    $87 = ($$pre$i4$i$i145|0)==(0|0);
    do {
     if ($87) {
      label = 76;
     } else {
      $88 = ((($$pre$i4$i$i145)) + 12|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = ((($$pre$i4$i$i145)) + 16|0);
      $91 = HEAP32[$90>>2]|0;
      $92 = ($89|0)==($91|0);
      if ($92) {
       $93 = HEAP32[$$pre$i4$i$i145>>2]|0;
       $94 = ((($93)) + 36|0);
       $95 = HEAP32[$94>>2]|0;
       __THREW__ = 0;
       $96 = (invoke_ii($95|0,($$pre$i4$i$i145|0))|0);
       $97 = __THREW__; __THREW__ = 0;
       $98 = $97&1;
       if ($98) {
        break L86;
       } else {
        $$0$i$i2$i$i143 = $96;
       }
      } else {
       $99 = HEAP8[$89>>0]|0;
       $100 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($99)|0);
       $$0$i$i2$i$i143 = $100;
      }
      $101 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $102 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i143,$101)|0);
      if ($102) {
       HEAP32[$1>>2] = 0;
       label = 76;
       break;
      } else {
       if ($196) {
        $$0127 = $2;$$3 = $$0126;
        break;
       } else {
        label = 78;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 76) {
     if ($196) {
      label = 78;
     } else {
      $$0127 = $2;$$3 = $$0126;
     }
    }
    if ((label|0) == 78) {
     $181 = HEAP32[$5>>2]|0;
     $182 = $181 | 2;
     HEAP32[$5>>2] = $182;
     $$0127 = $2;$$3 = $$0126;
    }
    while(1) {
     $183 = ($$0127|0)==($3|0);
     if ($183) {
      label = 82;
      break;
     }
     $184 = HEAP8[$$3>>0]|0;
     $185 = ($184<<24>>24)==(2);
     if ($185) {
      $$0127191 = $$0127;
      break;
     }
     $186 = ((($$0127)) + 12|0);
     $187 = ((($$3)) + 1|0);
     $$0127 = $186;$$3 = $187;
    }
    if ((label|0) == 82) {
     $188 = HEAP32[$5>>2]|0;
     $189 = $188 | 4;
     HEAP32[$5>>2] = $189;
     $$0127191 = $3;
    }
    $190 = ($$sroa$0159$0|0)==(0);
    if (!($190)) {
     $191 = $$sroa$0159$0;
     _free($191);
    }
    STACKTOP = sp;return ($$0127191|0);
   }
  } while(0);
  $192 = ___cxa_find_matching_catch_2()|0;
  $193 = tempRet0;
  $194 = ($$sroa$0159$0|0)==(0);
  if ($194) {
   $$sink134$sink135199$index = $192;$$sink134$sink135199$index5 = $193;
  } else {
   $195 = $$sroa$0159$0;
   _free($195);
   $$sink134$sink135199$index = $192;$$sink134$sink135199$index5 = $193;
  }
 }
 ___resumeException($$sink134$sink135199$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRb($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$2 = 0, $$232 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy2 = 0, $$ptr = 0, $$sroa$026$0 = 0, $$sroa$026$0$copyload = 0, $$sroa$026$0$copyload28 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $$byval_copy2 = sp + 60|0;
 $$byval_copy1 = sp + 56|0;
 $$byval_copy = sp + 52|0;
 $6 = sp + 48|0;
 $7 = sp + 44|0;
 $8 = sp + 40|0;
 $9 = sp + 36|0;
 $10 = sp + 32|0;
 $11 = sp + 8|0;
 $12 = sp;
 $13 = ((($3)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 $15 = $14 & 1;
 $16 = ($15|0)==(0);
 L1: do {
  if ($16) {
   HEAP32[$6>>2] = -1;
   $17 = HEAP32[$0>>2]|0;
   $18 = ((($17)) + 16|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = HEAP32[$1>>2]|0;
   HEAP32[$7>>2] = $20;
   $21 = HEAP32[$2>>2]|0;
   HEAP32[$8>>2] = $21;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$8>>2]|0;
   $22 = (FUNCTION_TABLE_iiiiiii[$19 & 1023]($0,$$byval_copy,$$byval_copy1,$3,$4,$6)|0);
   HEAP32[$1>>2] = $22;
   $23 = HEAP32[$6>>2]|0;
   switch ($23|0) {
   case 0:  {
    HEAP8[$5>>0] = 0;
    break;
   }
   case 1:  {
    HEAP8[$5>>0] = 1;
    break;
   }
   default: {
    HEAP8[$5>>0] = 1;
    HEAP32[$4>>2] = 4;
   }
   }
   $$sroa$026$0$copyload = HEAP32[$1>>2]|0;
   $$sroa$026$0 = $$sroa$026$0$copyload;
  } else {
   __ZNKSt3__28ios_base6getlocEv($9,$3);
   __THREW__ = 0;
   $24 = (invoke_iii(734,($9|0),(165472|0))|0);
   $25 = __THREW__; __THREW__ = 0;
   $26 = $25&1;
   do {
    if ($26) {
     $49 = ___cxa_find_matching_catch_2()|0;
     $50 = tempRet0;
     __ZNSt3__26localeD2Ev($9);
     $$2 = $49;$$232 = $50;
    } else {
     __ZNSt3__26localeD2Ev($9);
     __ZNKSt3__28ios_base6getlocEv($10,$3);
     __THREW__ = 0;
     $27 = (invoke_iii(734,($10|0),(165480|0))|0);
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $51 = ___cxa_find_matching_catch_2()|0;
      $52 = tempRet0;
      __ZNSt3__26localeD2Ev($10);
      $$2 = $51;$$232 = $52;
      break;
     }
     __ZNSt3__26localeD2Ev($10);
     $30 = HEAP32[$27>>2]|0;
     $31 = ((($30)) + 24|0);
     $32 = HEAP32[$31>>2]|0;
     __THREW__ = 0;
     invoke_vii($32|0,($11|0),($27|0));
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if ($34) {
      $35 = ___cxa_find_matching_catch_2()|0;
      $36 = tempRet0;
      $$2 = $35;$$232 = $36;
      break;
     }
     $$ptr = ((($11)) + 12|0);
     $37 = HEAP32[$27>>2]|0;
     $38 = ((($37)) + 28|0);
     $39 = HEAP32[$38>>2]|0;
     __THREW__ = 0;
     invoke_vii($39|0,($$ptr|0),($27|0));
     $40 = __THREW__; __THREW__ = 0;
     $41 = $40&1;
     if ($41) {
      $61 = ___cxa_find_matching_catch_2()|0;
      $62 = tempRet0;
      __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($11);
      $$2 = $61;$$232 = $62;
      break;
     }
     $42 = HEAP32[$2>>2]|0;
     HEAP32[$12>>2] = $42;
     $43 = ((($11)) + 24|0);
     __THREW__ = 0;
     ;HEAP32[$$byval_copy2>>2]=HEAP32[$12>>2]|0;
     $44 = (invoke_iiiiiiii(972,($1|0),($$byval_copy2|0),($11|0),($43|0),($24|0),($4|0),1)|0);
     $45 = __THREW__; __THREW__ = 0;
     $46 = $45&1;
     if ($46) {
      $53 = ___cxa_find_matching_catch_2()|0;
      $54 = tempRet0;
      $59 = $43;
      while(1) {
       $58 = ((($59)) + -12|0);
       __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($58);
       $60 = ($58|0)==($11|0);
       if ($60) {
        break;
       } else {
        $59 = $58;
       }
      }
      $$2 = $53;$$232 = $54;
      break;
     } else {
      $47 = ($44|0)==($11|0);
      $48 = $47&1;
      HEAP8[$5>>0] = $48;
      $$sroa$026$0$copyload28 = HEAP32[$1>>2]|0;
      $56 = $43;
      while(1) {
       $55 = ((($56)) + -12|0);
       __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($55);
       $57 = ($55|0)==($11|0);
       if ($57) {
        break;
       } else {
        $56 = $55;
       }
      }
      $$sroa$026$0 = $$sroa$026$0$copyload28;
      break L1;
     }
    }
   } while(0);
   ___resumeException($$2|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$026$0|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRl($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRx($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRt($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRy($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRf($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRd($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRe($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($0,$$byval_copy,$$byval_copy1,$3,$4,$5)|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i51 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i57 = 0, $$0$i$i33 = 0, $$0$i$i46 = 0, $$025 = 0, $$028 = 0, $$126 = 0, $$2 = 0, $$227 = 0, $$230 = 0, $$3 = 0, $$331 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i53 = 0, $$pre$i4$i$i59 = 0;
 var $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0;
 var $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0;
 var $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0;
 var $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i34 = 0, $phitmp$i = 0, $phitmp$i41 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 320|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(320|0);
 $vararg_buffer = sp;
 $6 = sp + 208|0;
 $7 = sp + 196|0;
 $8 = sp + 192|0;
 $9 = sp + 180|0;
 $10 = sp + 176|0;
 $11 = sp + 16|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $14 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$14>>2] = 0;
  $15 = (($$0$i$i) + 1)|0;
  $$0$i$i = $15;
 }
 __THREW__ = 0;
 invoke_vii(733,($8|0),($3|0));
 $16 = __THREW__; __THREW__ = 0;
 $17 = $16&1;
 do {
  if ($17) {
   $99 = ___cxa_find_matching_catch_2()|0;
   $100 = tempRet0;
   $$0 = $99;$$028 = $100;
   label = 36;
  } else {
   __THREW__ = 0;
   $18 = (invoke_iii(734,($8|0),(165472|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    $21 = HEAP32[$18>>2]|0;
    $22 = ((($21)) + 48|0);
    $23 = HEAP32[$22>>2]|0;
    __THREW__ = 0;
    (invoke_iiiii($23|0,($18|0),(155824|0),((155850)|0),($6|0))|0);
    $24 = __THREW__; __THREW__ = 0;
    $25 = $24&1;
    if (!($25)) {
     __ZNSt3__26localeD2Ev($8);
     ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
     $$0$i$i33 = 0;
     while(1) {
      $exitcond$i$i34 = ($$0$i$i33|0)==(3);
      if ($exitcond$i$i34) {
       break;
      }
      $26 = (($9) + ($$0$i$i33<<2)|0);
      HEAP32[$26>>2] = 0;
      $27 = (($$0$i$i33) + 1)|0;
      $$0$i$i33 = $27;
     }
     $28 = ((($9)) + 11|0);
     $29 = HEAP8[$28>>0]|0;
     $30 = ($29<<24>>24)<(0);
     if ($30) {
      $31 = ((($9)) + 8|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = $32 & 2147483647;
      $phitmp$i = (($33) + -1)|0;
      $34 = $phitmp$i;
     } else {
      $34 = 10;
     }
     __THREW__ = 0;
     invoke_viii(953,($9|0),($34|0),0);
     $35 = __THREW__; __THREW__ = 0;
     $36 = $35&1;
     if ($36) {
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      $$2 = $103;$$230 = $104;
     } else {
      $37 = HEAP8[$28>>0]|0;
      $38 = ($37<<24>>24)<(0);
      $39 = HEAP32[$9>>2]|0;
      $40 = $38 ? $39 : $9;
      HEAP32[$10>>2] = $40;
      HEAP32[$12>>2] = $11;
      HEAP32[$13>>2] = 0;
      $41 = ((($9)) + 4|0);
      $42 = ((($9)) + 8|0);
      $$pre = HEAP32[$1>>2]|0;
      $$025 = $40;$$pre$i$i$i = $$pre;$185 = $$pre;
      L21: while(1) {
       $43 = ($$pre$i$i$i|0)==(0|0);
       if ($43) {
        $$pre$i$i$i53 = 0;$155 = 0;$186 = 1;
       } else {
        $44 = ((($$pre$i$i$i)) + 12|0);
        $45 = HEAP32[$44>>2]|0;
        $46 = ((($$pre$i$i$i)) + 16|0);
        $47 = HEAP32[$46>>2]|0;
        $48 = ($45|0)==($47|0);
        if ($48) {
         $49 = HEAP32[$$pre$i$i$i>>2]|0;
         $50 = ((($49)) + 36|0);
         $51 = HEAP32[$50>>2]|0;
         __THREW__ = 0;
         $52 = (invoke_ii($51|0,($$pre$i$i$i|0))|0);
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if ($54) {
          break;
         } else {
          $$0$i$i$i$i = $52;
         }
        } else {
         $55 = HEAP32[$45>>2]|0;
         $56 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($55)|0);
         $$0$i$i$i$i = $56;
        }
        $57 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $58 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$57)|0);
        if ($58) {
         HEAP32[$1>>2] = 0;
         $$pre$i$i$i53 = 0;$155 = 0;$186 = 1;
        } else {
         $$pre$i$i$i53 = $$pre$i$i$i;$155 = $185;$186 = 0;
        }
       }
       $59 = HEAP32[$2>>2]|0;
       $60 = ($59|0)==(0|0);
       do {
        if ($60) {
         label = 26;
        } else {
         $61 = ((($59)) + 12|0);
         $62 = HEAP32[$61>>2]|0;
         $63 = ((($59)) + 16|0);
         $64 = HEAP32[$63>>2]|0;
         $65 = ($62|0)==($64|0);
         if ($65) {
          $66 = HEAP32[$59>>2]|0;
          $67 = ((($66)) + 36|0);
          $68 = HEAP32[$67>>2]|0;
          __THREW__ = 0;
          $69 = (invoke_ii($68|0,($59|0))|0);
          $70 = __THREW__; __THREW__ = 0;
          $71 = $70&1;
          if ($71) {
           break L21;
          } else {
           $$0$i$i2$i$i = $69;
          }
         } else {
          $72 = HEAP32[$62>>2]|0;
          $73 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($72)|0);
          $$0$i$i2$i$i = $73;
         }
         $74 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $75 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$74)|0);
         if ($75) {
          HEAP32[$2>>2] = 0;
          label = 26;
          break;
         } else {
          if ($186) {
           $187 = $59;
           break;
          } else {
           $$227 = $$025;$$pre$i4$i$i59 = $59;
           label = 46;
           break L21;
          }
         }
        }
       } while(0);
       if ((label|0) == 26) {
        label = 0;
        if ($186) {
         $$227 = $$025;$$pre$i4$i$i59 = 0;
         label = 46;
         break;
        } else {
         $187 = 0;
        }
       }
       $76 = HEAP32[$10>>2]|0;
       $77 = HEAP8[$28>>0]|0;
       $78 = ($77<<24>>24)<(0);
       $79 = HEAP32[$41>>2]|0;
       $80 = $77&255;
       $81 = $78 ? $79 : $80;
       $82 = (($$025) + ($81)|0);
       $83 = ($76|0)==($82|0);
       if ($83) {
        $84 = $81 << 1;
        __THREW__ = 0;
        invoke_viii(953,($9|0),($84|0),0);
        $85 = __THREW__; __THREW__ = 0;
        $86 = $85&1;
        if ($86) {
         break;
        }
        $87 = HEAP8[$28>>0]|0;
        $88 = ($87<<24>>24)<(0);
        if ($88) {
         $89 = HEAP32[$42>>2]|0;
         $90 = $89 & 2147483647;
         $phitmp$i41 = (($90) + -1)|0;
         $91 = $phitmp$i41;
        } else {
         $91 = 10;
        }
        __THREW__ = 0;
        invoke_viii(953,($9|0),($91|0),0);
        $92 = __THREW__; __THREW__ = 0;
        $93 = $92&1;
        if ($93) {
         break;
        }
        $94 = HEAP8[$28>>0]|0;
        $95 = ($94<<24>>24)<(0);
        $96 = HEAP32[$9>>2]|0;
        $97 = $95 ? $96 : $9;
        $98 = (($97) + ($81)|0);
        HEAP32[$10>>2] = $98;
        $$126 = $97;
       } else {
        $$126 = $$025;
       }
       $105 = ((($$pre$i$i$i53)) + 12|0);
       $106 = HEAP32[$105>>2]|0;
       $107 = ((($$pre$i$i$i53)) + 16|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = ($106|0)==($108|0);
       if ($109) {
        $110 = HEAP32[$$pre$i$i$i53>>2]|0;
        $111 = ((($110)) + 36|0);
        $112 = HEAP32[$111>>2]|0;
        __THREW__ = 0;
        $113 = (invoke_ii($112|0,($$pre$i$i$i53|0))|0);
        $114 = __THREW__; __THREW__ = 0;
        $115 = $114&1;
        if ($115) {
         break;
        } else {
         $$0$i$i46 = $113;
        }
       } else {
        $116 = HEAP32[$106>>2]|0;
        $117 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($116)|0);
        $$0$i$i46 = $117;
       }
       __THREW__ = 0;
       $118 = (invoke_iiiiiiiiiii(973,($$0$i$i46|0),16,($$126|0),($10|0),($13|0),0,($7|0),($11|0),($12|0),($6|0))|0);
       $119 = __THREW__; __THREW__ = 0;
       $120 = $119&1;
       if ($120) {
        break;
       }
       $121 = ($118|0)==(0);
       if (!($121)) {
        $$227 = $$126;$$pre$i4$i$i59 = $187;
        label = 46;
        break;
       }
       $122 = HEAP32[$105>>2]|0;
       $123 = HEAP32[$107>>2]|0;
       $124 = ($122|0)==($123|0);
       if ($124) {
        $125 = HEAP32[$$pre$i$i$i53>>2]|0;
        $126 = ((($125)) + 40|0);
        $127 = HEAP32[$126>>2]|0;
        __THREW__ = 0;
        (invoke_ii($127|0,($$pre$i$i$i53|0))|0);
        $128 = __THREW__; __THREW__ = 0;
        $129 = $128&1;
        if ($129) {
         break;
        } else {
         $$025 = $$126;$$pre$i$i$i = $$pre$i$i$i53;$185 = $155;
         continue;
        }
       } else {
        $130 = ((($122)) + 4|0);
        HEAP32[$105>>2] = $130;
        $131 = HEAP32[$122>>2]|0;
        (__ZNSt3__211char_traitsIwE11to_int_typeEw($131)|0);
        $$025 = $$126;$$pre$i$i$i = $$pre$i$i$i53;$185 = $155;
        continue;
       }
      }
      L60: do {
       if ((label|0) == 46) {
        $132 = HEAP32[$10>>2]|0;
        $133 = $$227;
        $134 = (($132) - ($133))|0;
        __THREW__ = 0;
        invoke_viii(953,($9|0),($134|0),0);
        $135 = __THREW__; __THREW__ = 0;
        $136 = $135&1;
        if (!($136)) {
         $137 = HEAP8[$28>>0]|0;
         $138 = HEAP32[$9>>2]|0;
         __THREW__ = 0;
         $139 = (invoke_i(955)|0);
         $140 = __THREW__; __THREW__ = 0;
         $141 = $140&1;
         if (!($141)) {
          $142 = ($137<<24>>24)<(0);
          $143 = $142 ? $138 : $9;
          __THREW__ = 0;
          HEAP32[$vararg_buffer>>2] = $5;
          $144 = (invoke_iiiii(956,($143|0),($139|0),(155857|0),($vararg_buffer|0))|0);
          $145 = __THREW__; __THREW__ = 0;
          $146 = $145&1;
          if (!($146)) {
           $147 = ($144|0)==(1);
           if (!($147)) {
            HEAP32[$4>>2] = 4;
           }
           $148 = ($$pre$i$i$i53|0)==(0|0);
           if ($148) {
            $188 = 1;
           } else {
            $149 = ((($$pre$i$i$i53)) + 12|0);
            $150 = HEAP32[$149>>2]|0;
            $151 = ((($$pre$i$i$i53)) + 16|0);
            $152 = HEAP32[$151>>2]|0;
            $153 = ($150|0)==($152|0);
            if ($153) {
             $154 = HEAP32[$155>>2]|0;
             $156 = ((($154)) + 36|0);
             $157 = HEAP32[$156>>2]|0;
             __THREW__ = 0;
             $158 = (invoke_ii($157|0,($$pre$i$i$i53|0))|0);
             $159 = __THREW__; __THREW__ = 0;
             $160 = $159&1;
             if ($160) {
              break;
             } else {
              $$0$i$i$i$i51 = $158;
             }
            } else {
             $161 = HEAP32[$150>>2]|0;
             $162 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($161)|0);
             $$0$i$i$i$i51 = $162;
            }
            $163 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
            $164 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i51,$163)|0);
            if ($164) {
             HEAP32[$1>>2] = 0;
             $188 = 1;
            } else {
             $188 = 0;
            }
           }
           $165 = ($$pre$i4$i$i59|0)==(0|0);
           do {
            if ($165) {
             label = 63;
            } else {
             $166 = ((($$pre$i4$i$i59)) + 12|0);
             $167 = HEAP32[$166>>2]|0;
             $168 = ((($$pre$i4$i$i59)) + 16|0);
             $169 = HEAP32[$168>>2]|0;
             $170 = ($167|0)==($169|0);
             if ($170) {
              $171 = HEAP32[$$pre$i4$i$i59>>2]|0;
              $172 = ((($171)) + 36|0);
              $173 = HEAP32[$172>>2]|0;
              __THREW__ = 0;
              $174 = (invoke_ii($173|0,($$pre$i4$i$i59|0))|0);
              $175 = __THREW__; __THREW__ = 0;
              $176 = $175&1;
              if ($176) {
               break L60;
              } else {
               $$0$i$i2$i$i57 = $174;
              }
             } else {
              $177 = HEAP32[$167>>2]|0;
              $178 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($177)|0);
              $$0$i$i2$i$i57 = $178;
             }
             $179 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
             $180 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i57,$179)|0);
             if ($180) {
              HEAP32[$2>>2] = 0;
              label = 63;
              break;
             } else {
              if ($188) {
               break;
              } else {
               label = 65;
               break;
              }
             }
            }
           } while(0);
           if ((label|0) == 63) {
            if ($188) {
             label = 65;
            }
           }
           if ((label|0) == 65) {
            $181 = HEAP32[$4>>2]|0;
            $182 = $181 | 2;
            HEAP32[$4>>2] = $182;
           }
           $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
           STACKTOP = sp;return ($$sroa$0$0$copyload|0);
          }
         }
        }
       }
      } while(0);
      $183 = ___cxa_find_matching_catch_2()|0;
      $184 = tempRet0;
      $$2 = $183;$$230 = $184;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
     $$3 = $$2;$$331 = $$230;
     break;
    }
   }
   $101 = ___cxa_find_matching_catch_2()|0;
   $102 = tempRet0;
   __ZNSt3__26localeD2Ev($8);
   $$0 = $101;$$028 = $102;
   label = 36;
  }
 } while(0);
 if ((label|0) == 36) {
  $$3 = $$0;$$331 = $$028;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
 ___resumeException($$3|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__29__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$add = 0, $$0$i$idx = 0, $$0$i$ptr = 0, $$0$lcssa$i = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $10 = HEAP32[$3>>2]|0;
 $11 = ($10|0)==($2|0);
 do {
  if ($11) {
   $12 = ((($9)) + 96|0);
   $13 = HEAP32[$12>>2]|0;
   $14 = ($13|0)==($0|0);
   if (!($14)) {
    $15 = ((($9)) + 100|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==($0|0);
    if (!($17)) {
     label = 5;
     break;
    }
   }
   $18 = $14 ? 43 : 45;
   $19 = ((($2)) + 1|0);
   HEAP32[$3>>2] = $19;
   HEAP8[$2>>0] = $18;
   HEAP32[$4>>2] = 0;
   $$1 = 0;
  } else {
   label = 5;
  }
 } while(0);
 L6: do {
  if ((label|0) == 5) {
   $20 = ((($6)) + 11|0);
   $21 = HEAP8[$20>>0]|0;
   $22 = ($21<<24>>24)<(0);
   $23 = ((($6)) + 4|0);
   $24 = HEAP32[$23>>2]|0;
   $25 = $21&255;
   $26 = $22 ? $24 : $25;
   $27 = ($26|0)!=(0);
   $28 = ($0|0)==($5|0);
   $or$cond = $28 & $27;
   if ($or$cond) {
    $29 = HEAP32[$8>>2]|0;
    $30 = $29;
    $31 = $7;
    $32 = (($30) - ($31))|0;
    $33 = ($32|0)<(160);
    if (!($33)) {
     $$1 = 0;
     break;
    }
    $34 = HEAP32[$4>>2]|0;
    $35 = ((($29)) + 4|0);
    HEAP32[$8>>2] = $35;
    HEAP32[$29>>2] = $34;
    HEAP32[$4>>2] = 0;
    $$1 = 0;
    break;
   }
   $36 = ((($9)) + 104|0);
   $$0$i$idx = 0;
   while(1) {
    $$0$i$ptr = (($9) + ($$0$i$idx<<2)|0);
    $37 = ($$0$i$idx|0)==(26);
    if ($37) {
     $$0$lcssa$i = $36;
     break;
    }
    $38 = HEAP32[$$0$i$ptr>>2]|0;
    $39 = ($38|0)==($0|0);
    $$0$i$add = (($$0$i$idx) + 1)|0;
    if ($39) {
     $$0$lcssa$i = $$0$i$ptr;
     break;
    } else {
     $$0$i$idx = $$0$i$add;
    }
   }
   $40 = $$0$lcssa$i;
   $41 = $9;
   $42 = (($40) - ($41))|0;
   $43 = $42 >> 2;
   $44 = ($42|0)>(92);
   if ($44) {
    $$1 = -1;
   } else {
    $45 = (155824 + ($43)|0);
    switch ($1|0) {
    case 10: case 8:  {
     $46 = ($43|0)<($1|0);
     if (!($46)) {
      $$1 = -1;
      break L6;
     }
     break;
    }
    case 16:  {
     $47 = ($42|0)<(88);
     if (!($47)) {
      if ($11) {
       $$1 = -1;
       break L6;
      }
      $48 = $10;
      $49 = $2;
      $50 = (($48) - ($49))|0;
      $51 = ($50|0)<(3);
      if (!($51)) {
       $$1 = -1;
       break L6;
      }
      $52 = ((($10)) + -1|0);
      $53 = HEAP8[$52>>0]|0;
      $54 = ($53<<24>>24)==(48);
      if (!($54)) {
       $$1 = -1;
       break L6;
      }
      HEAP32[$4>>2] = 0;
      $55 = HEAP8[$45>>0]|0;
      $56 = ((($10)) + 1|0);
      HEAP32[$3>>2] = $56;
      HEAP8[$10>>0] = $55;
      $$1 = 0;
      break L6;
     }
     break;
    }
    default: {
    }
    }
    $57 = HEAP8[$45>>0]|0;
    $58 = ((($10)) + 1|0);
    HEAP32[$3>>2] = $58;
    HEAP8[$10>>0] = $57;
    $59 = HEAP32[$4>>2]|0;
    $60 = (($59) + 1)|0;
    HEAP32[$4>>2] = $60;
    $$1 = 0;
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIeEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0.0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i29 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $6 = sp + 208|0;
 $7 = sp + 200|0;
 $8 = sp + 196|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 337|0;
 $16 = sp + 336|0;
 __ZNSt3__29__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$182 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP32[$36>>2]|0;
     $47 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$152 = $182;$183 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP32[$53>>2]|0;
      $64 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($183) {
       $184 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($183) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $184 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP32[$93>>2]|0;
    $104 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = HEAP32[$7>>2]|0;
   $106 = HEAP32[$8>>2]|0;
   __THREW__ = 0;
   $107 = (invoke_iiiiiiiiiiiii(974,($$0$i$i34|0),($15|0),($16|0),($$119|0),($11|0),($105|0),($106|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $108 = __THREW__; __THREW__ = 0;
   $109 = $108&1;
   if ($109) {
    break;
   }
   $110 = ($107|0)==(0);
   if (!($110)) {
    $$2 = $$119;$$pre$i4$i$i46 = $184;
    label = 37;
    break;
   }
   $111 = HEAP32[$92>>2]|0;
   $112 = HEAP32[$94>>2]|0;
   $113 = ($111|0)==($112|0);
   if ($113) {
    $114 = HEAP32[$$pre$i$i$i40>>2]|0;
    $115 = ((($114)) + 40|0);
    $116 = HEAP32[$115>>2]|0;
    __THREW__ = 0;
    (invoke_ii($116|0,($$pre$i$i$i40|0))|0);
    $117 = __THREW__; __THREW__ = 0;
    $118 = $117&1;
    if ($118) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
     continue;
    }
   } else {
    $119 = ((($111)) + 4|0);
    HEAP32[$92>>2] = $119;
    $120 = HEAP32[$111>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($120)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $121 = ((($9)) + 11|0);
    $122 = HEAP8[$121>>0]|0;
    $123 = ($122<<24>>24)<(0);
    $124 = ((($9)) + 4|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = $122&255;
    $127 = $123 ? $125 : $126;
    $128 = ($127|0)==(0);
    $129 = HEAP8[$15>>0]|0;
    $130 = ($129<<24>>24)==(0);
    $or$cond = $128 | $130;
    if (!($or$cond)) {
     $131 = HEAP32[$13>>2]|0;
     $132 = $131;
     $133 = $12;
     $134 = (($132) - ($133))|0;
     $135 = ($134|0)<(160);
     if ($135) {
      $136 = HEAP32[$14>>2]|0;
      $137 = ((($131)) + 4|0);
      HEAP32[$13>>2] = $137;
      HEAP32[$131>>2] = $136;
     }
    }
    $138 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $139 = (+invoke_diii(962,($$2|0),($138|0),($4|0)));
    $140 = __THREW__; __THREW__ = 0;
    $141 = $140&1;
    if (!($141)) {
     HEAPF64[$5>>3] = $139;
     $142 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($142|0),($4|0));
     $143 = __THREW__; __THREW__ = 0;
     $144 = $143&1;
     if (!($144)) {
      $145 = ($$pre$i$i$i40|0)==(0|0);
      if ($145) {
       $185 = 1;
      } else {
       $146 = ((($$pre$i$i$i40)) + 12|0);
       $147 = HEAP32[$146>>2]|0;
       $148 = ((($$pre$i$i$i40)) + 16|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = ($147|0)==($149|0);
       if ($150) {
        $151 = HEAP32[$152>>2]|0;
        $153 = ((($151)) + 36|0);
        $154 = HEAP32[$153>>2]|0;
        __THREW__ = 0;
        $155 = (invoke_ii($154|0,($$pre$i$i$i40|0))|0);
        $156 = __THREW__; __THREW__ = 0;
        $157 = $156&1;
        if ($157) {
         break;
        } else {
         $$0$i$i$i$i38 = $155;
        }
       } else {
        $158 = HEAP32[$147>>2]|0;
        $159 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($158)|0);
        $$0$i$i$i$i38 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i38,$160)|0);
       if ($161) {
        HEAP32[$1>>2] = 0;
        $185 = 1;
       } else {
        $185 = 0;
       }
      }
      $162 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($162) {
        label = 54;
       } else {
        $163 = ((($$pre$i4$i$i46)) + 12|0);
        $164 = HEAP32[$163>>2]|0;
        $165 = ((($$pre$i4$i$i46)) + 16|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $168 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $169 = ((($168)) + 36|0);
         $170 = HEAP32[$169>>2]|0;
         __THREW__ = 0;
         $171 = (invoke_ii($170|0,($$pre$i4$i$i46|0))|0);
         $172 = __THREW__; __THREW__ = 0;
         $173 = $172&1;
         if ($173) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $171;
         }
        } else {
         $174 = HEAP32[$164>>2]|0;
         $175 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($174)|0);
         $$0$i$i2$i$i44 = $175;
        }
        $176 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $177 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i44,$176)|0);
        if ($177) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($185) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($185) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $178 = HEAP32[$4>>2]|0;
       $179 = $178 | 2;
       HEAP32[$4>>2] = $179;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $180 = ___cxa_find_matching_catch_2()|0;
  $181 = tempRet0;
  $$1 = $180;$$121 = $181;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__29__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $5 = sp;
 __ZNKSt3__28ios_base6getlocEv($5,$1);
 __THREW__ = 0;
 $6 = (invoke_iii(734,($5|0),(165472|0))|0);
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $9 = HEAP32[$6>>2]|0;
  $10 = ((($9)) + 48|0);
  $11 = HEAP32[$10>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($11|0,($6|0),(155824|0),((155856)|0),($2|0))|0);
  $12 = __THREW__; __THREW__ = 0;
  $13 = $12&1;
  if (!($13)) {
   __THREW__ = 0;
   $14 = (invoke_iii(734,($5|0),(165480|0))|0);
   $15 = __THREW__; __THREW__ = 0;
   $16 = $15&1;
   if (!($16)) {
    $17 = HEAP32[$14>>2]|0;
    $18 = ((($17)) + 12|0);
    $19 = HEAP32[$18>>2]|0;
    __THREW__ = 0;
    $20 = (invoke_ii($19|0,($14|0))|0);
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if (!($22)) {
     HEAP32[$3>>2] = $20;
     $23 = HEAP32[$14>>2]|0;
     $24 = ((($23)) + 16|0);
     $25 = HEAP32[$24>>2]|0;
     __THREW__ = 0;
     $26 = (invoke_ii($25|0,($14|0))|0);
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if (!($28)) {
      HEAP32[$4>>2] = $26;
      $29 = HEAP32[$14>>2]|0;
      $30 = ((($29)) + 20|0);
      $31 = HEAP32[$30>>2]|0;
      __THREW__ = 0;
      invoke_vii($31|0,($0|0),($14|0));
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if (!($33)) {
       __ZNSt3__26localeD2Ev($5);
       STACKTOP = sp;return;
      }
     }
    }
   }
  }
 }
 $34 = ___cxa_find_matching_catch_2()|0;
 $35 = tempRet0;
 __ZNSt3__26localeD2Ev($5);
 ___resumeException($34|0);
 // unreachable;
}
function __ZNSt3__29__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 $11 = $11|0;
 var $$0$i$add = 0, $$0$i$idx = 0, $$0$i$ptr = 0, $$0$lcssa$i = 0, $$2 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0;
 var $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0;
 var $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $12 = ($0|0)==($5|0);
 $13 = $8;
 L1: do {
  if ($12) {
   $14 = HEAP8[$1>>0]|0;
   $15 = ($14<<24>>24)==(0);
   if ($15) {
    $$2 = -1;
   } else {
    HEAP8[$1>>0] = 0;
    $16 = HEAP32[$4>>2]|0;
    $17 = ((($16)) + 1|0);
    HEAP32[$4>>2] = $17;
    HEAP8[$16>>0] = 46;
    $18 = ((($7)) + 11|0);
    $19 = HEAP8[$18>>0]|0;
    $20 = ($19<<24>>24)<(0);
    $21 = ((($7)) + 4|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = $19&255;
    $24 = $20 ? $22 : $23;
    $25 = ($24|0)==(0);
    if ($25) {
     $$2 = 0;
    } else {
     $26 = HEAP32[$9>>2]|0;
     $27 = $26;
     $28 = (($27) - ($13))|0;
     $29 = ($28|0)<(160);
     if ($29) {
      $30 = HEAP32[$10>>2]|0;
      $31 = ((($26)) + 4|0);
      HEAP32[$9>>2] = $31;
      HEAP32[$26>>2] = $30;
      $$2 = 0;
     } else {
      $$2 = 0;
     }
    }
   }
  } else {
   $32 = ($0|0)==($6|0);
   if ($32) {
    $33 = ((($7)) + 11|0);
    $34 = HEAP8[$33>>0]|0;
    $35 = ($34<<24>>24)<(0);
    $36 = ((($7)) + 4|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = $34&255;
    $39 = $35 ? $37 : $38;
    $40 = ($39|0)==(0);
    if (!($40)) {
     $41 = HEAP8[$1>>0]|0;
     $42 = ($41<<24>>24)==(0);
     if ($42) {
      $$2 = -1;
      break;
     }
     $43 = HEAP32[$9>>2]|0;
     $44 = $43;
     $45 = (($44) - ($13))|0;
     $46 = ($45|0)<(160);
     if (!($46)) {
      $$2 = 0;
      break;
     }
     $47 = HEAP32[$10>>2]|0;
     $48 = ((($43)) + 4|0);
     HEAP32[$9>>2] = $48;
     HEAP32[$43>>2] = $47;
     HEAP32[$10>>2] = 0;
     $$2 = 0;
     break;
    }
   }
   $49 = ((($11)) + 128|0);
   $$0$i$idx = 0;
   while(1) {
    $$0$i$ptr = (($11) + ($$0$i$idx<<2)|0);
    $50 = ($$0$i$idx|0)==(32);
    if ($50) {
     $$0$lcssa$i = $49;
     break;
    }
    $51 = HEAP32[$$0$i$ptr>>2]|0;
    $52 = ($51|0)==($0|0);
    $$0$i$add = (($$0$i$idx) + 1)|0;
    if ($52) {
     $$0$lcssa$i = $$0$i$ptr;
     break;
    } else {
     $$0$i$idx = $$0$i$add;
    }
   }
   $53 = $$0$lcssa$i;
   $54 = $11;
   $55 = (($53) - ($54))|0;
   $56 = $55 >> 2;
   $57 = ($55|0)>(124);
   if ($57) {
    $$2 = -1;
   } else {
    $58 = (155824 + ($56)|0);
    $59 = HEAP8[$58>>0]|0;
    switch ($56|0) {
    case 24: case 25:  {
     $60 = HEAP32[$4>>2]|0;
     $61 = ($60|0)==($3|0);
     if (!($61)) {
      $62 = ((($60)) + -1|0);
      $63 = HEAP8[$62>>0]|0;
      $64 = $63 & 95;
      $65 = HEAP8[$2>>0]|0;
      $66 = $65 & 127;
      $67 = ($64<<24>>24)==($66<<24>>24);
      if (!($67)) {
       $$2 = -1;
       break L1;
      }
     }
     $68 = ((($60)) + 1|0);
     HEAP32[$4>>2] = $68;
     HEAP8[$60>>0] = $59;
     $$2 = 0;
     break L1;
     break;
    }
    case 23: case 22:  {
     HEAP8[$2>>0] = 80;
     break;
    }
    default: {
     $69 = $59 & 95;
     $70 = $69&255;
     $71 = HEAP8[$2>>0]|0;
     $72 = $71 << 24 >> 24;
     $73 = ($70|0)==($72|0);
     if ($73) {
      $74 = $70 | 128;
      $75 = $74&255;
      HEAP8[$2>>0] = $75;
      $76 = HEAP8[$1>>0]|0;
      $77 = ($76<<24>>24)==(0);
      if (!($77)) {
       HEAP8[$1>>0] = 0;
       $78 = ((($7)) + 11|0);
       $79 = HEAP8[$78>>0]|0;
       $80 = ($79<<24>>24)<(0);
       $81 = ((($7)) + 4|0);
       $82 = HEAP32[$81>>2]|0;
       $83 = $79&255;
       $84 = $80 ? $82 : $83;
       $85 = ($84|0)==(0);
       if (!($85)) {
        $86 = HEAP32[$9>>2]|0;
        $87 = $86;
        $88 = (($87) - ($13))|0;
        $89 = ($88|0)<(160);
        if ($89) {
         $90 = HEAP32[$10>>2]|0;
         $91 = ((($86)) + 4|0);
         HEAP32[$9>>2] = $91;
         HEAP32[$86>>2] = $90;
        }
       }
      }
     }
    }
    }
    $92 = HEAP32[$4>>2]|0;
    $93 = ((($92)) + 1|0);
    HEAP32[$4>>2] = $93;
    HEAP8[$92>>0] = $59;
    $94 = ($55|0)>(84);
    if ($94) {
     $$2 = 0;
    } else {
     $95 = HEAP32[$10>>2]|0;
     $96 = (($95) + 1)|0;
     HEAP32[$10>>2] = $96;
     $$2 = 0;
    }
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIdEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0.0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i29 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $6 = sp + 208|0;
 $7 = sp + 200|0;
 $8 = sp + 196|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 337|0;
 $16 = sp + 336|0;
 __ZNSt3__29__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$182 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP32[$36>>2]|0;
     $47 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$152 = $182;$183 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP32[$53>>2]|0;
      $64 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($183) {
       $184 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($183) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $184 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP32[$93>>2]|0;
    $104 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = HEAP32[$7>>2]|0;
   $106 = HEAP32[$8>>2]|0;
   __THREW__ = 0;
   $107 = (invoke_iiiiiiiiiiiii(974,($$0$i$i34|0),($15|0),($16|0),($$119|0),($11|0),($105|0),($106|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $108 = __THREW__; __THREW__ = 0;
   $109 = $108&1;
   if ($109) {
    break;
   }
   $110 = ($107|0)==(0);
   if (!($110)) {
    $$2 = $$119;$$pre$i4$i$i46 = $184;
    label = 37;
    break;
   }
   $111 = HEAP32[$92>>2]|0;
   $112 = HEAP32[$94>>2]|0;
   $113 = ($111|0)==($112|0);
   if ($113) {
    $114 = HEAP32[$$pre$i$i$i40>>2]|0;
    $115 = ((($114)) + 40|0);
    $116 = HEAP32[$115>>2]|0;
    __THREW__ = 0;
    (invoke_ii($116|0,($$pre$i$i$i40|0))|0);
    $117 = __THREW__; __THREW__ = 0;
    $118 = $117&1;
    if ($118) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
     continue;
    }
   } else {
    $119 = ((($111)) + 4|0);
    HEAP32[$92>>2] = $119;
    $120 = HEAP32[$111>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($120)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $121 = ((($9)) + 11|0);
    $122 = HEAP8[$121>>0]|0;
    $123 = ($122<<24>>24)<(0);
    $124 = ((($9)) + 4|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = $122&255;
    $127 = $123 ? $125 : $126;
    $128 = ($127|0)==(0);
    $129 = HEAP8[$15>>0]|0;
    $130 = ($129<<24>>24)==(0);
    $or$cond = $128 | $130;
    if (!($or$cond)) {
     $131 = HEAP32[$13>>2]|0;
     $132 = $131;
     $133 = $12;
     $134 = (($132) - ($133))|0;
     $135 = ($134|0)<(160);
     if ($135) {
      $136 = HEAP32[$14>>2]|0;
      $137 = ((($131)) + 4|0);
      HEAP32[$13>>2] = $137;
      HEAP32[$131>>2] = $136;
     }
    }
    $138 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $139 = (+invoke_diii(964,($$2|0),($138|0),($4|0)));
    $140 = __THREW__; __THREW__ = 0;
    $141 = $140&1;
    if (!($141)) {
     HEAPF64[$5>>3] = $139;
     $142 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($142|0),($4|0));
     $143 = __THREW__; __THREW__ = 0;
     $144 = $143&1;
     if (!($144)) {
      $145 = ($$pre$i$i$i40|0)==(0|0);
      if ($145) {
       $185 = 1;
      } else {
       $146 = ((($$pre$i$i$i40)) + 12|0);
       $147 = HEAP32[$146>>2]|0;
       $148 = ((($$pre$i$i$i40)) + 16|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = ($147|0)==($149|0);
       if ($150) {
        $151 = HEAP32[$152>>2]|0;
        $153 = ((($151)) + 36|0);
        $154 = HEAP32[$153>>2]|0;
        __THREW__ = 0;
        $155 = (invoke_ii($154|0,($$pre$i$i$i40|0))|0);
        $156 = __THREW__; __THREW__ = 0;
        $157 = $156&1;
        if ($157) {
         break;
        } else {
         $$0$i$i$i$i38 = $155;
        }
       } else {
        $158 = HEAP32[$147>>2]|0;
        $159 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($158)|0);
        $$0$i$i$i$i38 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i38,$160)|0);
       if ($161) {
        HEAP32[$1>>2] = 0;
        $185 = 1;
       } else {
        $185 = 0;
       }
      }
      $162 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($162) {
        label = 54;
       } else {
        $163 = ((($$pre$i4$i$i46)) + 12|0);
        $164 = HEAP32[$163>>2]|0;
        $165 = ((($$pre$i4$i$i46)) + 16|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $168 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $169 = ((($168)) + 36|0);
         $170 = HEAP32[$169>>2]|0;
         __THREW__ = 0;
         $171 = (invoke_ii($170|0,($$pre$i4$i$i46|0))|0);
         $172 = __THREW__; __THREW__ = 0;
         $173 = $172&1;
         if ($173) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $171;
         }
        } else {
         $174 = HEAP32[$164>>2]|0;
         $175 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($174)|0);
         $$0$i$i2$i$i44 = $175;
        }
        $176 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $177 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i44,$176)|0);
        if ($177) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($185) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($185) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $178 = HEAP32[$4>>2]|0;
       $179 = $178 | 2;
       HEAP32[$4>>2] = $179;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $180 = ___cxa_find_matching_catch_2()|0;
  $181 = tempRet0;
  $$1 = $180;$$121 = $181;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE23__do_get_floating_pointIfEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i38 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i44 = 0, $$0$i$i34 = 0, $$018 = 0, $$1 = 0, $$119 = 0, $$121 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i40 = 0, $$pre$i4$i$i46 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0.0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i29 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $6 = sp + 208|0;
 $7 = sp + 200|0;
 $8 = sp + 196|0;
 $9 = sp + 184|0;
 $10 = sp + 172|0;
 $11 = sp + 168|0;
 $12 = sp + 8|0;
 $13 = sp + 4|0;
 $14 = sp;
 $15 = sp + 337|0;
 $16 = sp + 336|0;
 __ZNSt3__29__num_getIwE19__stage2_float_prepERNS_8ios_baseEPwRwS5_($9,$3,$6,$7,$8);
 ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($10) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($10)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($10)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  $24 = $23 & 2147483647;
  $phitmp$i = (($24) + -1)|0;
  $25 = $phitmp$i;
 } else {
  $25 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($10|0),($25|0),0);
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $90 = ___cxa_find_matching_catch_2()|0;
  $91 = tempRet0;
  $$1 = $90;$$121 = $91;
 } else {
  $28 = HEAP8[$19>>0]|0;
  $29 = ($28<<24>>24)<(0);
  $30 = HEAP32[$10>>2]|0;
  $31 = $29 ? $30 : $10;
  HEAP32[$11>>2] = $31;
  HEAP32[$13>>2] = $12;
  HEAP32[$14>>2] = 0;
  HEAP8[$15>>0] = 1;
  HEAP8[$16>>0] = 69;
  $32 = ((($10)) + 4|0);
  $33 = ((($10)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$018 = $31;$$pre$i$i$i = $$pre;$182 = $$pre;
  L11: while(1) {
   $34 = ($$pre$i$i$i|0)==(0|0);
   if ($34) {
    $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
   } else {
    $35 = ((($$pre$i$i$i)) + 12|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ((($$pre$i$i$i)) + 16|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $40 = HEAP32[$$pre$i$i$i>>2]|0;
     $41 = ((($40)) + 36|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_ii($42|0,($$pre$i$i$i|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      break;
     } else {
      $$0$i$i$i$i = $43;
     }
    } else {
     $46 = HEAP32[$36>>2]|0;
     $47 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($46)|0);
     $$0$i$i$i$i = $47;
    }
    $48 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $49 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$48)|0);
    if ($49) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i40 = 0;$152 = 0;$183 = 1;
    } else {
     $$pre$i$i$i40 = $$pre$i$i$i;$152 = $182;$183 = 0;
    }
   }
   $50 = HEAP32[$2>>2]|0;
   $51 = ($50|0)==(0|0);
   do {
    if ($51) {
     label = 20;
    } else {
     $52 = ((($50)) + 12|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ((($50)) + 16|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = ($53|0)==($55|0);
     if ($56) {
      $57 = HEAP32[$50>>2]|0;
      $58 = ((($57)) + 36|0);
      $59 = HEAP32[$58>>2]|0;
      __THREW__ = 0;
      $60 = (invoke_ii($59|0,($50|0))|0);
      $61 = __THREW__; __THREW__ = 0;
      $62 = $61&1;
      if ($62) {
       break L11;
      } else {
       $$0$i$i2$i$i = $60;
      }
     } else {
      $63 = HEAP32[$53>>2]|0;
      $64 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($63)|0);
      $$0$i$i2$i$i = $64;
     }
     $65 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $66 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$65)|0);
     if ($66) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($183) {
       $184 = $50;
       break;
      } else {
       $$2 = $$018;$$pre$i4$i$i46 = $50;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($183) {
     $$2 = $$018;$$pre$i4$i$i46 = 0;
     label = 37;
     break;
    } else {
     $184 = 0;
    }
   }
   $67 = HEAP32[$11>>2]|0;
   $68 = HEAP8[$19>>0]|0;
   $69 = ($68<<24>>24)<(0);
   $70 = HEAP32[$32>>2]|0;
   $71 = $68&255;
   $72 = $69 ? $70 : $71;
   $73 = (($$018) + ($72)|0);
   $74 = ($67|0)==($73|0);
   if ($74) {
    $75 = $72 << 1;
    __THREW__ = 0;
    invoke_viii(953,($10|0),($75|0),0);
    $76 = __THREW__; __THREW__ = 0;
    $77 = $76&1;
    if ($77) {
     break;
    }
    $78 = HEAP8[$19>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = HEAP32[$33>>2]|0;
     $81 = $80 & 2147483647;
     $phitmp$i29 = (($81) + -1)|0;
     $82 = $phitmp$i29;
    } else {
     $82 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($10|0),($82|0),0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     break;
    }
    $85 = HEAP8[$19>>0]|0;
    $86 = ($85<<24>>24)<(0);
    $87 = HEAP32[$10>>2]|0;
    $88 = $86 ? $87 : $10;
    $89 = (($88) + ($72)|0);
    HEAP32[$11>>2] = $89;
    $$119 = $88;
   } else {
    $$119 = $$018;
   }
   $92 = ((($$pre$i$i$i40)) + 12|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ((($$pre$i$i$i40)) + 16|0);
   $95 = HEAP32[$94>>2]|0;
   $96 = ($93|0)==($95|0);
   if ($96) {
    $97 = HEAP32[$$pre$i$i$i40>>2]|0;
    $98 = ((($97)) + 36|0);
    $99 = HEAP32[$98>>2]|0;
    __THREW__ = 0;
    $100 = (invoke_ii($99|0,($$pre$i$i$i40|0))|0);
    $101 = __THREW__; __THREW__ = 0;
    $102 = $101&1;
    if ($102) {
     break;
    } else {
     $$0$i$i34 = $100;
    }
   } else {
    $103 = HEAP32[$93>>2]|0;
    $104 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($103)|0);
    $$0$i$i34 = $104;
   }
   $105 = HEAP32[$7>>2]|0;
   $106 = HEAP32[$8>>2]|0;
   __THREW__ = 0;
   $107 = (invoke_iiiiiiiiiiiii(974,($$0$i$i34|0),($15|0),($16|0),($$119|0),($11|0),($105|0),($106|0),($9|0),($12|0),($13|0),($14|0),($6|0))|0);
   $108 = __THREW__; __THREW__ = 0;
   $109 = $108&1;
   if ($109) {
    break;
   }
   $110 = ($107|0)==(0);
   if (!($110)) {
    $$2 = $$119;$$pre$i4$i$i46 = $184;
    label = 37;
    break;
   }
   $111 = HEAP32[$92>>2]|0;
   $112 = HEAP32[$94>>2]|0;
   $113 = ($111|0)==($112|0);
   if ($113) {
    $114 = HEAP32[$$pre$i$i$i40>>2]|0;
    $115 = ((($114)) + 40|0);
    $116 = HEAP32[$115>>2]|0;
    __THREW__ = 0;
    (invoke_ii($116|0,($$pre$i$i$i40|0))|0);
    $117 = __THREW__; __THREW__ = 0;
    $118 = $117&1;
    if ($118) {
     break;
    } else {
     $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
     continue;
    }
   } else {
    $119 = ((($111)) + 4|0);
    HEAP32[$92>>2] = $119;
    $120 = HEAP32[$111>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($120)|0);
    $$018 = $$119;$$pre$i$i$i = $$pre$i$i$i40;$182 = $152;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $121 = ((($9)) + 11|0);
    $122 = HEAP8[$121>>0]|0;
    $123 = ($122<<24>>24)<(0);
    $124 = ((($9)) + 4|0);
    $125 = HEAP32[$124>>2]|0;
    $126 = $122&255;
    $127 = $123 ? $125 : $126;
    $128 = ($127|0)==(0);
    $129 = HEAP8[$15>>0]|0;
    $130 = ($129<<24>>24)==(0);
    $or$cond = $128 | $130;
    if (!($or$cond)) {
     $131 = HEAP32[$13>>2]|0;
     $132 = $131;
     $133 = $12;
     $134 = (($132) - ($133))|0;
     $135 = ($134|0)<(160);
     if ($135) {
      $136 = HEAP32[$14>>2]|0;
      $137 = ((($131)) + 4|0);
      HEAP32[$13>>2] = $137;
      HEAP32[$131>>2] = $136;
     }
    }
    $138 = HEAP32[$11>>2]|0;
    __THREW__ = 0;
    $139 = (+invoke_diii(965,($$2|0),($138|0),($4|0)));
    $140 = __THREW__; __THREW__ = 0;
    $141 = $140&1;
    if (!($141)) {
     HEAPF32[$5>>2] = $139;
     $142 = HEAP32[$13>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($9|0),($12|0),($142|0),($4|0));
     $143 = __THREW__; __THREW__ = 0;
     $144 = $143&1;
     if (!($144)) {
      $145 = ($$pre$i$i$i40|0)==(0|0);
      if ($145) {
       $185 = 1;
      } else {
       $146 = ((($$pre$i$i$i40)) + 12|0);
       $147 = HEAP32[$146>>2]|0;
       $148 = ((($$pre$i$i$i40)) + 16|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = ($147|0)==($149|0);
       if ($150) {
        $151 = HEAP32[$152>>2]|0;
        $153 = ((($151)) + 36|0);
        $154 = HEAP32[$153>>2]|0;
        __THREW__ = 0;
        $155 = (invoke_ii($154|0,($$pre$i$i$i40|0))|0);
        $156 = __THREW__; __THREW__ = 0;
        $157 = $156&1;
        if ($157) {
         break;
        } else {
         $$0$i$i$i$i38 = $155;
        }
       } else {
        $158 = HEAP32[$147>>2]|0;
        $159 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($158)|0);
        $$0$i$i$i$i38 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i38,$160)|0);
       if ($161) {
        HEAP32[$1>>2] = 0;
        $185 = 1;
       } else {
        $185 = 0;
       }
      }
      $162 = ($$pre$i4$i$i46|0)==(0|0);
      do {
       if ($162) {
        label = 54;
       } else {
        $163 = ((($$pre$i4$i$i46)) + 12|0);
        $164 = HEAP32[$163>>2]|0;
        $165 = ((($$pre$i4$i$i46)) + 16|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $168 = HEAP32[$$pre$i4$i$i46>>2]|0;
         $169 = ((($168)) + 36|0);
         $170 = HEAP32[$169>>2]|0;
         __THREW__ = 0;
         $171 = (invoke_ii($170|0,($$pre$i4$i$i46|0))|0);
         $172 = __THREW__; __THREW__ = 0;
         $173 = $172&1;
         if ($173) {
          break L50;
         } else {
          $$0$i$i2$i$i44 = $171;
         }
        } else {
         $174 = HEAP32[$164>>2]|0;
         $175 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($174)|0);
         $$0$i$i2$i$i44 = $175;
        }
        $176 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $177 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i44,$176)|0);
        if ($177) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($185) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($185) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $178 = HEAP32[$4>>2]|0;
       $179 = $178 | 2;
       HEAP32[$4>>2] = $179;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $180 = ___cxa_find_matching_catch_2()|0;
  $181 = tempRet0;
  $$1 = $180;$$121 = $181;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIyEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$182 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$152 = 0;$183 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$152 = 0;$183 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$152 = $182;$183 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($183) {
       $184 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($183) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $184 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $184;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$182 = $152;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$182 = $152;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(966,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = tempRet0;
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     $138 = $5;
     $139 = $138;
     HEAP32[$139>>2] = $134;
     $140 = (($138) + 4)|0;
     $141 = $140;
     HEAP32[$141>>2] = $135;
     $142 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($142|0),($4|0));
     $143 = __THREW__; __THREW__ = 0;
     $144 = $143&1;
     if (!($144)) {
      $145 = ($$pre$i$i$i45|0)==(0|0);
      if ($145) {
       $185 = 1;
      } else {
       $146 = ((($$pre$i$i$i45)) + 12|0);
       $147 = HEAP32[$146>>2]|0;
       $148 = ((($$pre$i$i$i45)) + 16|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = ($147|0)==($149|0);
       if ($150) {
        $151 = HEAP32[$152>>2]|0;
        $153 = ((($151)) + 36|0);
        $154 = HEAP32[$153>>2]|0;
        __THREW__ = 0;
        $155 = (invoke_ii($154|0,($$pre$i$i$i45|0))|0);
        $156 = __THREW__; __THREW__ = 0;
        $157 = $156&1;
        if ($157) {
         break;
        } else {
         $$0$i$i$i$i43 = $155;
        }
       } else {
        $158 = HEAP32[$147>>2]|0;
        $159 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($158)|0);
        $$0$i$i$i$i43 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$160)|0);
       if ($161) {
        HEAP32[$1>>2] = 0;
        $185 = 1;
       } else {
        $185 = 0;
       }
      }
      $162 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($162) {
        label = 54;
       } else {
        $163 = ((($$pre$i4$i$i51)) + 12|0);
        $164 = HEAP32[$163>>2]|0;
        $165 = ((($$pre$i4$i$i51)) + 16|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $168 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $169 = ((($168)) + 36|0);
         $170 = HEAP32[$169>>2]|0;
         __THREW__ = 0;
         $171 = (invoke_ii($170|0,($$pre$i4$i$i51|0))|0);
         $172 = __THREW__; __THREW__ = 0;
         $173 = $172&1;
         if ($173) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $171;
         }
        } else {
         $174 = HEAP32[$164>>2]|0;
         $175 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($174)|0);
         $$0$i$i2$i$i49 = $175;
        }
        $176 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $177 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$176)|0);
        if ($177) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($185) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($185) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $178 = HEAP32[$4>>2]|0;
       $179 = $178 | 2;
       HEAP32[$4>>2] = $179;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $180 = ___cxa_find_matching_catch_2()|0;
  $181 = tempRet0;
  $$1 = $180;$$127 = $181;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 __ZNKSt3__28ios_base6getlocEv($4,$1);
 __THREW__ = 0;
 $5 = (invoke_iii(734,($4|0),(165472|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if (!($7)) {
  $8 = HEAP32[$5>>2]|0;
  $9 = ((($8)) + 48|0);
  $10 = HEAP32[$9>>2]|0;
  __THREW__ = 0;
  (invoke_iiiii($10|0,($5|0),(155824|0),((155850)|0),($2|0))|0);
  $11 = __THREW__; __THREW__ = 0;
  $12 = $11&1;
  if (!($12)) {
   __THREW__ = 0;
   $13 = (invoke_iii(734,($4|0),(165480|0))|0);
   $14 = __THREW__; __THREW__ = 0;
   $15 = $14&1;
   if (!($15)) {
    $16 = HEAP32[$13>>2]|0;
    $17 = ((($16)) + 16|0);
    $18 = HEAP32[$17>>2]|0;
    __THREW__ = 0;
    $19 = (invoke_ii($18|0,($13|0))|0);
    $20 = __THREW__; __THREW__ = 0;
    $21 = $20&1;
    if (!($21)) {
     HEAP32[$3>>2] = $19;
     $22 = HEAP32[$13>>2]|0;
     $23 = ((($22)) + 20|0);
     $24 = HEAP32[$23>>2]|0;
     __THREW__ = 0;
     invoke_vii($24|0,($0|0),($13|0));
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if (!($26)) {
      __ZNSt3__26localeD2Ev($4);
      STACKTOP = sp;return;
     }
    }
   }
  }
 }
 $27 = ___cxa_find_matching_catch_2()|0;
 $28 = tempRet0;
 __ZNSt3__26localeD2Ev($4);
 ___resumeException($27|0);
 // unreachable;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedImEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$177 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$147 = $177;$178 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($178) {
       $179 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($178) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $179 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $179;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(967,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = __THREW__; __THREW__ = 0;
    $136 = $135&1;
    if (!($136)) {
     HEAP32[$5>>2] = $134;
     $137 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($137|0),($4|0));
     $138 = __THREW__; __THREW__ = 0;
     $139 = $138&1;
     if (!($139)) {
      $140 = ($$pre$i$i$i45|0)==(0|0);
      if ($140) {
       $180 = 1;
      } else {
       $141 = ((($$pre$i$i$i45)) + 12|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ((($$pre$i$i$i45)) + 16|0);
       $144 = HEAP32[$143>>2]|0;
       $145 = ($142|0)==($144|0);
       if ($145) {
        $146 = HEAP32[$147>>2]|0;
        $148 = ((($146)) + 36|0);
        $149 = HEAP32[$148>>2]|0;
        __THREW__ = 0;
        $150 = (invoke_ii($149|0,($$pre$i$i$i45|0))|0);
        $151 = __THREW__; __THREW__ = 0;
        $152 = $151&1;
        if ($152) {
         break;
        } else {
         $$0$i$i$i$i43 = $150;
        }
       } else {
        $153 = HEAP32[$142>>2]|0;
        $154 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($153)|0);
        $$0$i$i$i$i43 = $154;
       }
       $155 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $156 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$155)|0);
       if ($156) {
        HEAP32[$1>>2] = 0;
        $180 = 1;
       } else {
        $180 = 0;
       }
      }
      $157 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($157) {
        label = 54;
       } else {
        $158 = ((($$pre$i4$i$i51)) + 12|0);
        $159 = HEAP32[$158>>2]|0;
        $160 = ((($$pre$i4$i$i51)) + 16|0);
        $161 = HEAP32[$160>>2]|0;
        $162 = ($159|0)==($161|0);
        if ($162) {
         $163 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $164 = ((($163)) + 36|0);
         $165 = HEAP32[$164>>2]|0;
         __THREW__ = 0;
         $166 = (invoke_ii($165|0,($$pre$i4$i$i51|0))|0);
         $167 = __THREW__; __THREW__ = 0;
         $168 = $167&1;
         if ($168) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $166;
         }
        } else {
         $169 = HEAP32[$159>>2]|0;
         $170 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($169)|0);
         $$0$i$i2$i$i49 = $170;
        }
        $171 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $172 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$171)|0);
        if ($172) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($180) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($180) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $173 = HEAP32[$4>>2]|0;
       $174 = $173 | 2;
       HEAP32[$4>>2] = $174;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $175 = ___cxa_find_matching_catch_2()|0;
  $176 = tempRet0;
  $$1 = $175;$$127 = $176;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedIjEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$177 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$147 = $177;$178 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($178) {
       $179 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($178) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $179 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $179;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(968,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = __THREW__; __THREW__ = 0;
    $136 = $135&1;
    if (!($136)) {
     HEAP32[$5>>2] = $134;
     $137 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($137|0),($4|0));
     $138 = __THREW__; __THREW__ = 0;
     $139 = $138&1;
     if (!($139)) {
      $140 = ($$pre$i$i$i45|0)==(0|0);
      if ($140) {
       $180 = 1;
      } else {
       $141 = ((($$pre$i$i$i45)) + 12|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ((($$pre$i$i$i45)) + 16|0);
       $144 = HEAP32[$143>>2]|0;
       $145 = ($142|0)==($144|0);
       if ($145) {
        $146 = HEAP32[$147>>2]|0;
        $148 = ((($146)) + 36|0);
        $149 = HEAP32[$148>>2]|0;
        __THREW__ = 0;
        $150 = (invoke_ii($149|0,($$pre$i$i$i45|0))|0);
        $151 = __THREW__; __THREW__ = 0;
        $152 = $151&1;
        if ($152) {
         break;
        } else {
         $$0$i$i$i$i43 = $150;
        }
       } else {
        $153 = HEAP32[$142>>2]|0;
        $154 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($153)|0);
        $$0$i$i$i$i43 = $154;
       }
       $155 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $156 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$155)|0);
       if ($156) {
        HEAP32[$1>>2] = 0;
        $180 = 1;
       } else {
        $180 = 0;
       }
      }
      $157 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($157) {
        label = 54;
       } else {
        $158 = ((($$pre$i4$i$i51)) + 12|0);
        $159 = HEAP32[$158>>2]|0;
        $160 = ((($$pre$i4$i$i51)) + 16|0);
        $161 = HEAP32[$160>>2]|0;
        $162 = ($159|0)==($161|0);
        if ($162) {
         $163 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $164 = ((($163)) + 36|0);
         $165 = HEAP32[$164>>2]|0;
         __THREW__ = 0;
         $166 = (invoke_ii($165|0,($$pre$i4$i$i51|0))|0);
         $167 = __THREW__; __THREW__ = 0;
         $168 = $167&1;
         if ($168) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $166;
         }
        } else {
         $169 = HEAP32[$159>>2]|0;
         $170 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($169)|0);
         $$0$i$i2$i$i49 = $170;
        }
        $171 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $172 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$171)|0);
        if ($172) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($180) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($180) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $173 = HEAP32[$4>>2]|0;
       $174 = $173 | 2;
       HEAP32[$4>>2] = $174;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $175 = ___cxa_find_matching_catch_2()|0;
  $176 = tempRet0;
  $$1 = $175;$$127 = $176;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__do_get_unsignedItEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$177 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$147 = $177;$178 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($178) {
       $179 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($178) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $179 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $179;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(969,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = __THREW__; __THREW__ = 0;
    $136 = $135&1;
    if (!($136)) {
     HEAP16[$5>>1] = $134;
     $137 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($137|0),($4|0));
     $138 = __THREW__; __THREW__ = 0;
     $139 = $138&1;
     if (!($139)) {
      $140 = ($$pre$i$i$i45|0)==(0|0);
      if ($140) {
       $180 = 1;
      } else {
       $141 = ((($$pre$i$i$i45)) + 12|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ((($$pre$i$i$i45)) + 16|0);
       $144 = HEAP32[$143>>2]|0;
       $145 = ($142|0)==($144|0);
       if ($145) {
        $146 = HEAP32[$147>>2]|0;
        $148 = ((($146)) + 36|0);
        $149 = HEAP32[$148>>2]|0;
        __THREW__ = 0;
        $150 = (invoke_ii($149|0,($$pre$i$i$i45|0))|0);
        $151 = __THREW__; __THREW__ = 0;
        $152 = $151&1;
        if ($152) {
         break;
        } else {
         $$0$i$i$i$i43 = $150;
        }
       } else {
        $153 = HEAP32[$142>>2]|0;
        $154 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($153)|0);
        $$0$i$i$i$i43 = $154;
       }
       $155 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $156 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$155)|0);
       if ($156) {
        HEAP32[$1>>2] = 0;
        $180 = 1;
       } else {
        $180 = 0;
       }
      }
      $157 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($157) {
        label = 54;
       } else {
        $158 = ((($$pre$i4$i$i51)) + 12|0);
        $159 = HEAP32[$158>>2]|0;
        $160 = ((($$pre$i4$i$i51)) + 16|0);
        $161 = HEAP32[$160>>2]|0;
        $162 = ($159|0)==($161|0);
        if ($162) {
         $163 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $164 = ((($163)) + 36|0);
         $165 = HEAP32[$164>>2]|0;
         __THREW__ = 0;
         $166 = (invoke_ii($165|0,($$pre$i4$i$i51|0))|0);
         $167 = __THREW__; __THREW__ = 0;
         $168 = $167&1;
         if ($168) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $166;
         }
        } else {
         $169 = HEAP32[$159>>2]|0;
         $170 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($169)|0);
         $$0$i$i2$i$i49 = $170;
        }
        $171 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $172 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$171)|0);
        if ($172) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($180) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($180) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $173 = HEAP32[$4>>2]|0;
       $174 = $173 | 2;
       HEAP32[$4>>2] = $174;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $175 = ___cxa_find_matching_catch_2()|0;
  $176 = tempRet0;
  $$1 = $175;$$127 = $176;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIxEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0;
 var $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$182 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$152 = 0;$183 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$152 = 0;$183 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$152 = $182;$183 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($183) {
       $184 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($183) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $184 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $184;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$182 = $152;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$182 = $152;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(970,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = tempRet0;
    $136 = __THREW__; __THREW__ = 0;
    $137 = $136&1;
    if (!($137)) {
     $138 = $5;
     $139 = $138;
     HEAP32[$139>>2] = $134;
     $140 = (($138) + 4)|0;
     $141 = $140;
     HEAP32[$141>>2] = $135;
     $142 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($142|0),($4|0));
     $143 = __THREW__; __THREW__ = 0;
     $144 = $143&1;
     if (!($144)) {
      $145 = ($$pre$i$i$i45|0)==(0|0);
      if ($145) {
       $185 = 1;
      } else {
       $146 = ((($$pre$i$i$i45)) + 12|0);
       $147 = HEAP32[$146>>2]|0;
       $148 = ((($$pre$i$i$i45)) + 16|0);
       $149 = HEAP32[$148>>2]|0;
       $150 = ($147|0)==($149|0);
       if ($150) {
        $151 = HEAP32[$152>>2]|0;
        $153 = ((($151)) + 36|0);
        $154 = HEAP32[$153>>2]|0;
        __THREW__ = 0;
        $155 = (invoke_ii($154|0,($$pre$i$i$i45|0))|0);
        $156 = __THREW__; __THREW__ = 0;
        $157 = $156&1;
        if ($157) {
         break;
        } else {
         $$0$i$i$i$i43 = $155;
        }
       } else {
        $158 = HEAP32[$147>>2]|0;
        $159 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($158)|0);
        $$0$i$i$i$i43 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$160)|0);
       if ($161) {
        HEAP32[$1>>2] = 0;
        $185 = 1;
       } else {
        $185 = 0;
       }
      }
      $162 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($162) {
        label = 54;
       } else {
        $163 = ((($$pre$i4$i$i51)) + 12|0);
        $164 = HEAP32[$163>>2]|0;
        $165 = ((($$pre$i4$i$i51)) + 16|0);
        $166 = HEAP32[$165>>2]|0;
        $167 = ($164|0)==($166|0);
        if ($167) {
         $168 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $169 = ((($168)) + 36|0);
         $170 = HEAP32[$169>>2]|0;
         __THREW__ = 0;
         $171 = (invoke_ii($170|0,($$pre$i4$i$i51|0))|0);
         $172 = __THREW__; __THREW__ = 0;
         $173 = $172&1;
         if ($173) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $171;
         }
        } else {
         $174 = HEAP32[$164>>2]|0;
         $175 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($174)|0);
         $$0$i$i2$i$i49 = $175;
        }
        $176 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $177 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$176)|0);
        if ($177) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($185) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($185) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $178 = HEAP32[$4>>2]|0;
       $179 = $178 | 2;
       HEAP32[$4>>2] = $179;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $180 = ___cxa_find_matching_catch_2()|0;
  $181 = tempRet0;
  $$1 = $180;$$127 = $181;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__do_get_signedIlEES4_S4_S4_RNS_8ios_baseERjRT_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i43 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i49 = 0, $$0$i$i39 = 0, $$024 = 0, $$1 = 0, $$125 = 0, $$127 = 0, $$2 = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i45 = 0, $$pre$i4$i$i51 = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0;
 var $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0;
 var $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0;
 var $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0;
 var $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $phitmp$i = 0, $phitmp$i34 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 304|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(304|0);
 $6 = sp + 200|0;
 $7 = sp + 196|0;
 $8 = sp + 184|0;
 $9 = sp + 172|0;
 $10 = sp + 168|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = (__ZNSt3__214__num_get_base10__get_baseERNS_8ios_baseE($3)|0);
 __ZNSt3__29__num_getIwE17__stage2_int_prepERNS_8ios_baseEPwRw($8,$3,$6,$7);
 ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $15 = (($9) + ($$0$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i) + 1)|0;
  $$0$i$i = $16;
 }
 $17 = ((($9)) + 11|0);
 $18 = HEAP8[$17>>0]|0;
 $19 = ($18<<24>>24)<(0);
 if ($19) {
  $20 = ((($9)) + 8|0);
  $21 = HEAP32[$20>>2]|0;
  $22 = $21 & 2147483647;
  $phitmp$i = (($22) + -1)|0;
  $23 = $phitmp$i;
 } else {
  $23 = 10;
 }
 __THREW__ = 0;
 invoke_viii(953,($9|0),($23|0),0);
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $88 = ___cxa_find_matching_catch_2()|0;
  $89 = tempRet0;
  $$1 = $88;$$127 = $89;
 } else {
  $26 = HEAP8[$17>>0]|0;
  $27 = ($26<<24>>24)<(0);
  $28 = HEAP32[$9>>2]|0;
  $29 = $27 ? $28 : $9;
  HEAP32[$10>>2] = $29;
  HEAP32[$12>>2] = $11;
  HEAP32[$13>>2] = 0;
  $30 = ((($9)) + 4|0);
  $31 = ((($9)) + 8|0);
  $$pre = HEAP32[$1>>2]|0;
  $$024 = $29;$$pre$i$i$i = $$pre;$177 = $$pre;
  L11: while(1) {
   $32 = ($$pre$i$i$i|0)==(0|0);
   if ($32) {
    $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
   } else {
    $33 = ((($$pre$i$i$i)) + 12|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = ((($$pre$i$i$i)) + 16|0);
    $36 = HEAP32[$35>>2]|0;
    $37 = ($34|0)==($36|0);
    if ($37) {
     $38 = HEAP32[$$pre$i$i$i>>2]|0;
     $39 = ((($38)) + 36|0);
     $40 = HEAP32[$39>>2]|0;
     __THREW__ = 0;
     $41 = (invoke_ii($40|0,($$pre$i$i$i|0))|0);
     $42 = __THREW__; __THREW__ = 0;
     $43 = $42&1;
     if ($43) {
      break;
     } else {
      $$0$i$i$i$i = $41;
     }
    } else {
     $44 = HEAP32[$34>>2]|0;
     $45 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($44)|0);
     $$0$i$i$i$i = $45;
    }
    $46 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $47 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$46)|0);
    if ($47) {
     HEAP32[$1>>2] = 0;
     $$pre$i$i$i45 = 0;$147 = 0;$178 = 1;
    } else {
     $$pre$i$i$i45 = $$pre$i$i$i;$147 = $177;$178 = 0;
    }
   }
   $48 = HEAP32[$2>>2]|0;
   $49 = ($48|0)==(0|0);
   do {
    if ($49) {
     label = 20;
    } else {
     $50 = ((($48)) + 12|0);
     $51 = HEAP32[$50>>2]|0;
     $52 = ((($48)) + 16|0);
     $53 = HEAP32[$52>>2]|0;
     $54 = ($51|0)==($53|0);
     if ($54) {
      $55 = HEAP32[$48>>2]|0;
      $56 = ((($55)) + 36|0);
      $57 = HEAP32[$56>>2]|0;
      __THREW__ = 0;
      $58 = (invoke_ii($57|0,($48|0))|0);
      $59 = __THREW__; __THREW__ = 0;
      $60 = $59&1;
      if ($60) {
       break L11;
      } else {
       $$0$i$i2$i$i = $58;
      }
     } else {
      $61 = HEAP32[$51>>2]|0;
      $62 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
      $$0$i$i2$i$i = $62;
     }
     $63 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $64 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$63)|0);
     if ($64) {
      HEAP32[$2>>2] = 0;
      label = 20;
      break;
     } else {
      if ($178) {
       $179 = $48;
       break;
      } else {
       $$2 = $$024;$$pre$i4$i$i51 = $48;
       label = 37;
       break L11;
      }
     }
    }
   } while(0);
   if ((label|0) == 20) {
    label = 0;
    if ($178) {
     $$2 = $$024;$$pre$i4$i$i51 = 0;
     label = 37;
     break;
    } else {
     $179 = 0;
    }
   }
   $65 = HEAP32[$10>>2]|0;
   $66 = HEAP8[$17>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$30>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = (($$024) + ($70)|0);
   $72 = ($65|0)==($71|0);
   if ($72) {
    $73 = $70 << 1;
    __THREW__ = 0;
    invoke_viii(953,($9|0),($73|0),0);
    $74 = __THREW__; __THREW__ = 0;
    $75 = $74&1;
    if ($75) {
     break;
    }
    $76 = HEAP8[$17>>0]|0;
    $77 = ($76<<24>>24)<(0);
    if ($77) {
     $78 = HEAP32[$31>>2]|0;
     $79 = $78 & 2147483647;
     $phitmp$i34 = (($79) + -1)|0;
     $80 = $phitmp$i34;
    } else {
     $80 = 10;
    }
    __THREW__ = 0;
    invoke_viii(953,($9|0),($80|0),0);
    $81 = __THREW__; __THREW__ = 0;
    $82 = $81&1;
    if ($82) {
     break;
    }
    $83 = HEAP8[$17>>0]|0;
    $84 = ($83<<24>>24)<(0);
    $85 = HEAP32[$9>>2]|0;
    $86 = $84 ? $85 : $9;
    $87 = (($86) + ($70)|0);
    HEAP32[$10>>2] = $87;
    $$125 = $86;
   } else {
    $$125 = $$024;
   }
   $90 = ((($$pre$i$i$i45)) + 12|0);
   $91 = HEAP32[$90>>2]|0;
   $92 = ((($$pre$i$i$i45)) + 16|0);
   $93 = HEAP32[$92>>2]|0;
   $94 = ($91|0)==($93|0);
   if ($94) {
    $95 = HEAP32[$$pre$i$i$i45>>2]|0;
    $96 = ((($95)) + 36|0);
    $97 = HEAP32[$96>>2]|0;
    __THREW__ = 0;
    $98 = (invoke_ii($97|0,($$pre$i$i$i45|0))|0);
    $99 = __THREW__; __THREW__ = 0;
    $100 = $99&1;
    if ($100) {
     break;
    } else {
     $$0$i$i39 = $98;
    }
   } else {
    $101 = HEAP32[$91>>2]|0;
    $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
    $$0$i$i39 = $102;
   }
   $103 = HEAP32[$7>>2]|0;
   __THREW__ = 0;
   $104 = (invoke_iiiiiiiiiii(973,($$0$i$i39|0),($14|0),($$125|0),($10|0),($13|0),($103|0),($8|0),($11|0),($12|0),($6|0))|0);
   $105 = __THREW__; __THREW__ = 0;
   $106 = $105&1;
   if ($106) {
    break;
   }
   $107 = ($104|0)==(0);
   if (!($107)) {
    $$2 = $$125;$$pre$i4$i$i51 = $179;
    label = 37;
    break;
   }
   $108 = HEAP32[$90>>2]|0;
   $109 = HEAP32[$92>>2]|0;
   $110 = ($108|0)==($109|0);
   if ($110) {
    $111 = HEAP32[$$pre$i$i$i45>>2]|0;
    $112 = ((($111)) + 40|0);
    $113 = HEAP32[$112>>2]|0;
    __THREW__ = 0;
    (invoke_ii($113|0,($$pre$i$i$i45|0))|0);
    $114 = __THREW__; __THREW__ = 0;
    $115 = $114&1;
    if ($115) {
     break;
    } else {
     $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
     continue;
    }
   } else {
    $116 = ((($108)) + 4|0);
    HEAP32[$90>>2] = $116;
    $117 = HEAP32[$108>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($117)|0);
    $$024 = $$125;$$pre$i$i$i = $$pre$i$i$i45;$177 = $147;
    continue;
   }
  }
  L50: do {
   if ((label|0) == 37) {
    $118 = ((($8)) + 11|0);
    $119 = HEAP8[$118>>0]|0;
    $120 = ($119<<24>>24)<(0);
    $121 = ((($8)) + 4|0);
    $122 = HEAP32[$121>>2]|0;
    $123 = $119&255;
    $124 = $120 ? $122 : $123;
    $125 = ($124|0)==(0);
    if (!($125)) {
     $126 = HEAP32[$12>>2]|0;
     $127 = $126;
     $128 = $11;
     $129 = (($127) - ($128))|0;
     $130 = ($129|0)<(160);
     if ($130) {
      $131 = HEAP32[$13>>2]|0;
      $132 = ((($126)) + 4|0);
      HEAP32[$12>>2] = $132;
      HEAP32[$126>>2] = $131;
     }
    }
    $133 = HEAP32[$10>>2]|0;
    __THREW__ = 0;
    $134 = (invoke_iiiii(971,($$2|0),($133|0),($4|0),($14|0))|0);
    $135 = __THREW__; __THREW__ = 0;
    $136 = $135&1;
    if (!($136)) {
     HEAP32[$5>>2] = $134;
     $137 = HEAP32[$12>>2]|0;
     __THREW__ = 0;
     invoke_viiii(963,($8|0),($11|0),($137|0),($4|0));
     $138 = __THREW__; __THREW__ = 0;
     $139 = $138&1;
     if (!($139)) {
      $140 = ($$pre$i$i$i45|0)==(0|0);
      if ($140) {
       $180 = 1;
      } else {
       $141 = ((($$pre$i$i$i45)) + 12|0);
       $142 = HEAP32[$141>>2]|0;
       $143 = ((($$pre$i$i$i45)) + 16|0);
       $144 = HEAP32[$143>>2]|0;
       $145 = ($142|0)==($144|0);
       if ($145) {
        $146 = HEAP32[$147>>2]|0;
        $148 = ((($146)) + 36|0);
        $149 = HEAP32[$148>>2]|0;
        __THREW__ = 0;
        $150 = (invoke_ii($149|0,($$pre$i$i$i45|0))|0);
        $151 = __THREW__; __THREW__ = 0;
        $152 = $151&1;
        if ($152) {
         break;
        } else {
         $$0$i$i$i$i43 = $150;
        }
       } else {
        $153 = HEAP32[$142>>2]|0;
        $154 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($153)|0);
        $$0$i$i$i$i43 = $154;
       }
       $155 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $156 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i43,$155)|0);
       if ($156) {
        HEAP32[$1>>2] = 0;
        $180 = 1;
       } else {
        $180 = 0;
       }
      }
      $157 = ($$pre$i4$i$i51|0)==(0|0);
      do {
       if ($157) {
        label = 54;
       } else {
        $158 = ((($$pre$i4$i$i51)) + 12|0);
        $159 = HEAP32[$158>>2]|0;
        $160 = ((($$pre$i4$i$i51)) + 16|0);
        $161 = HEAP32[$160>>2]|0;
        $162 = ($159|0)==($161|0);
        if ($162) {
         $163 = HEAP32[$$pre$i4$i$i51>>2]|0;
         $164 = ((($163)) + 36|0);
         $165 = HEAP32[$164>>2]|0;
         __THREW__ = 0;
         $166 = (invoke_ii($165|0,($$pre$i4$i$i51|0))|0);
         $167 = __THREW__; __THREW__ = 0;
         $168 = $167&1;
         if ($168) {
          break L50;
         } else {
          $$0$i$i2$i$i49 = $166;
         }
        } else {
         $169 = HEAP32[$159>>2]|0;
         $170 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($169)|0);
         $$0$i$i2$i$i49 = $170;
        }
        $171 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
        $172 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i49,$171)|0);
        if ($172) {
         HEAP32[$2>>2] = 0;
         label = 54;
         break;
        } else {
         if ($180) {
          break;
         } else {
          label = 56;
          break;
         }
        }
       }
      } while(0);
      if ((label|0) == 54) {
       if ($180) {
        label = 56;
       }
      }
      if ((label|0) == 56) {
       $173 = HEAP32[$4>>2]|0;
       $174 = $173 | 2;
       HEAP32[$4>>2] = $174;
      }
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
      STACKTOP = sp;return ($$sroa$0$0$copyload|0);
     }
    }
   }
  } while(0);
  $175 = ___cxa_find_matching_catch_2()|0;
  $176 = tempRet0;
  $$1 = $175;$$127 = $176;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($9);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 ___resumeException($$1|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i143 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i149 = 0, $$0103 = 0, $$0104 = 0, $$0105$off0 = 0, $$0106 = 0, $$0107 = 0, $$0108 = 0, $$0109 = 0, $$0112 = 0, $$0116 = 0, $$0132 = 0, $$0133 = 0, $$0133197 = 0, $$1110 = 0, $$1113 = 0;
 var $$1117 = 0, $$2$off0 = 0, $$2111 = 0, $$2114 = 0, $$2118 = 0, $$3 = 0, $$3115 = 0, $$3119 = 0, $$5 = 0, $$5121 = 0, $$6 = 0, $$7 = 0, $$pre$i$i$i = 0, $$pre$i$i$i145 = 0, $$pre$i4$i$i151 = 0, $$sink140$sink141205$index = 0, $$sink140$sink141205$index5 = 0, $$sroa$0165$0 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i146 = 0, $tmp$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $7 = sp;
 $8 = $3;
 $9 = $2;
 $10 = (($8) - ($9))|0;
 $11 = (($10|0) / 12)&-1;
 $12 = ($11>>>0)>(100);
 do {
  if ($12) {
   $13 = (_malloc($11)|0);
   $14 = ($13|0)==(0|0);
   if ($14) {
    __THREW__ = 0;
    invoke_v(732);
    $15 = __THREW__; __THREW__ = 0;
    $16 = ___cxa_find_matching_catch_2()|0;
    $17 = tempRet0;
    $$sink140$sink141205$index = $16;$$sink140$sink141205$index5 = $17;
    break;
   } else {
    $18 = $13;
    $$0132 = $13;$$sroa$0165$0 = $18;
    label = 5;
    break;
   }
  } else {
   $$0132 = $7;$$sroa$0165$0 = 0;
   label = 5;
  }
 } while(0);
 if ((label|0) == 5) {
  $$0108 = $2;$$0109 = $$0132;$$0112 = 0;$$0116 = $11;
  while(1) {
   $19 = ($$0108|0)==($3|0);
   if ($19) {
    break;
   }
   $20 = ((($$0108)) + 8|0);
   $21 = ((($20)) + 3|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = ($22<<24>>24)<(0);
   if ($23) {
    $24 = ((($$0108)) + 4|0);
    $25 = HEAP32[$24>>2]|0;
    $28 = $25;
   } else {
    $26 = $22&255;
    $28 = $26;
   }
   $27 = ($28|0)==(0);
   if ($27) {
    HEAP8[$$0109>>0] = 2;
    $29 = (($$0116) + -1)|0;
    $30 = (($$0112) + 1)|0;
    $$1113 = $30;$$1117 = $29;
   } else {
    HEAP8[$$0109>>0] = 1;
    $$1113 = $$0112;$$1117 = $$0116;
   }
   $31 = ((($$0108)) + 12|0);
   $32 = ((($$0109)) + 1|0);
   $$0108 = $31;$$0109 = $32;$$0112 = $$1113;$$0116 = $$1117;
  }
  $$0107 = 0;$$2114 = $$0112;$$2118 = $$0116;
  L20: while(1) {
   $33 = HEAP32[$0>>2]|0;
   $34 = ($33|0)==(0|0);
   do {
    if ($34) {
     $67 = 1;
    } else {
     $35 = ((($33)) + 12|0);
     $36 = HEAP32[$35>>2]|0;
     $37 = ((($33)) + 16|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = ($36|0)==($38|0);
     if ($39) {
      $40 = HEAP32[$33>>2]|0;
      $41 = ((($40)) + 36|0);
      $42 = HEAP32[$41>>2]|0;
      __THREW__ = 0;
      $43 = (invoke_ii($42|0,($33|0))|0);
      $44 = __THREW__; __THREW__ = 0;
      $45 = $44&1;
      if ($45) {
       break L20;
      } else {
       $$0$i$i$i$i = $43;
      }
     } else {
      $46 = HEAP32[$36>>2]|0;
      $47 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($46)|0);
      $$0$i$i$i$i = $47;
     }
     $48 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $49 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$48)|0);
     if ($49) {
      HEAP32[$0>>2] = 0;
      $67 = 1;
      break;
     } else {
      $$pre$i$i$i = HEAP32[$0>>2]|0;
      $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
      $67 = $phitmp$i$i$i;
      break;
     }
    }
   } while(0);
   $50 = HEAP32[$1>>2]|0;
   $51 = ($50|0)==(0|0);
   if ($51) {
    $$pre$i4$i$i151 = 0;$68 = 1;
   } else {
    $52 = ((($50)) + 12|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = ((($50)) + 16|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ($53|0)==($55|0);
    if ($56) {
     $57 = HEAP32[$50>>2]|0;
     $58 = ((($57)) + 36|0);
     $59 = HEAP32[$58>>2]|0;
     __THREW__ = 0;
     $60 = (invoke_ii($59|0,($50|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      break;
     } else {
      $$0$i$i2$i$i = $60;
     }
    } else {
     $63 = HEAP32[$53>>2]|0;
     $64 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($63)|0);
     $$0$i$i2$i$i = $64;
    }
    $65 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $66 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$65)|0);
    if ($66) {
     HEAP32[$1>>2] = 0;
     $$pre$i4$i$i151 = 0;$68 = 1;
    } else {
     $$pre$i4$i$i151 = $50;$68 = 0;
    }
   }
   $tmp$i$i = $67 ^ $68;
   $69 = ($$2118|0)!=(0);
   $70 = $69 & $tmp$i$i;
   $71 = HEAP32[$0>>2]|0;
   if (!($70)) {
    label = 29;
    break;
   }
   $104 = ((($71)) + 12|0);
   $105 = HEAP32[$104>>2]|0;
   $106 = ((($71)) + 16|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = ($105|0)==($107|0);
   if ($108) {
    $109 = HEAP32[$71>>2]|0;
    $110 = ((($109)) + 36|0);
    $111 = HEAP32[$110>>2]|0;
    __THREW__ = 0;
    $112 = (invoke_ii($111|0,($71|0))|0);
    $113 = __THREW__; __THREW__ = 0;
    $114 = $113&1;
    if ($114) {
     break;
    } else {
     $$0$i$i = $112;
    }
   } else {
    $115 = HEAP32[$105>>2]|0;
    $116 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($115)|0);
    $$0$i$i = $116;
   }
   if ($6) {
    $$0106 = $$0$i$i;
   } else {
    $117 = HEAP32[$4>>2]|0;
    $118 = ((($117)) + 28|0);
    $119 = HEAP32[$118>>2]|0;
    __THREW__ = 0;
    $120 = (invoke_iii($119|0,($4|0),($$0$i$i|0))|0);
    $121 = __THREW__; __THREW__ = 0;
    $122 = $121&1;
    if ($122) {
     break;
    } else {
     $$0106 = $120;
    }
   }
   $123 = (($$0107) + 1)|0;
   $$0104 = $2;$$0105$off0 = 0;$$1110 = $$0132;$$3115 = $$2114;$$3119 = $$2118;
   while(1) {
    $124 = ($$0104|0)==($3|0);
    if ($124) {
     break;
    }
    $125 = HEAP8[$$1110>>0]|0;
    $126 = ($125<<24>>24)==(1);
    do {
     if ($126) {
      $127 = ((($$0104)) + 8|0);
      $128 = ((($127)) + 3|0);
      $129 = HEAP8[$128>>0]|0;
      $130 = ($129<<24>>24)<(0);
      if ($130) {
       $131 = HEAP32[$$0104>>2]|0;
       $133 = $131;
      } else {
       $133 = $$0104;
      }
      $132 = (($133) + ($$0107<<2)|0);
      $134 = HEAP32[$132>>2]|0;
      if ($6) {
       $$0103 = $134;
      } else {
       $135 = HEAP32[$4>>2]|0;
       $136 = ((($135)) + 28|0);
       $137 = HEAP32[$136>>2]|0;
       __THREW__ = 0;
       $138 = (invoke_iii($137|0,($4|0),($134|0))|0);
       $139 = __THREW__; __THREW__ = 0;
       $140 = $139&1;
       if ($140) {
        break L20;
       } else {
        $$0103 = $138;
       }
      }
      $141 = ($$0106|0)==($$0103|0);
      $142 = (($$3119) + -1)|0;
      if (!($141)) {
       HEAP8[$$1110>>0] = 0;
       $$2$off0 = $$0105$off0;$$5 = $$3115;$$5121 = $142;
       break;
      }
      $143 = HEAP8[$128>>0]|0;
      $144 = ($143<<24>>24)<(0);
      if ($144) {
       $145 = ((($$0104)) + 4|0);
       $146 = HEAP32[$145>>2]|0;
       $149 = $146;
      } else {
       $147 = $143&255;
       $149 = $147;
      }
      $148 = ($149|0)==($123|0);
      if ($148) {
       $150 = (($$3115) + 1)|0;
       HEAP8[$$1110>>0] = 2;
       $$2$off0 = 1;$$5 = $150;$$5121 = $142;
      } else {
       $$2$off0 = 1;$$5 = $$3115;$$5121 = $$3119;
      }
     } else {
      $$2$off0 = $$0105$off0;$$5 = $$3115;$$5121 = $$3119;
     }
    } while(0);
    $151 = ((($$0104)) + 12|0);
    $152 = ((($$1110)) + 1|0);
    $$0104 = $151;$$0105$off0 = $$2$off0;$$1110 = $152;$$3115 = $$5;$$3119 = $$5121;
   }
   if (!($$0105$off0)) {
    $$0107 = $123;$$2114 = $$3115;$$2118 = $$3119;
    continue;
   }
   $153 = HEAP32[$0>>2]|0;
   $154 = ((($153)) + 12|0);
   $155 = HEAP32[$154>>2]|0;
   $156 = ((($153)) + 16|0);
   $157 = HEAP32[$156>>2]|0;
   $158 = ($155|0)==($157|0);
   if ($158) {
    $159 = HEAP32[$153>>2]|0;
    $160 = ((($159)) + 40|0);
    $161 = HEAP32[$160>>2]|0;
    __THREW__ = 0;
    (invoke_ii($161|0,($153|0))|0);
    $162 = __THREW__; __THREW__ = 0;
    $163 = $162&1;
    if ($163) {
     break;
    }
   } else {
    $164 = ((($155)) + 4|0);
    HEAP32[$154>>2] = $164;
    $165 = HEAP32[$155>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($165)|0);
   }
   $166 = (($$3115) + ($$3119))|0;
   $167 = ($166>>>0)>(1);
   if ($167) {
    $$0 = $2;$$2111 = $$0132;$$6 = $$3115;
   } else {
    $$0107 = $123;$$2114 = $$3115;$$2118 = $$3119;
    continue;
   }
   while(1) {
    $168 = ($$0|0)==($3|0);
    if ($168) {
     $$0107 = $123;$$2114 = $$6;$$2118 = $$3119;
     continue L20;
    }
    $169 = HEAP8[$$2111>>0]|0;
    $170 = ($169<<24>>24)==(2);
    if ($170) {
     $171 = ((($$0)) + 8|0);
     $172 = ((($171)) + 3|0);
     $173 = HEAP8[$172>>0]|0;
     $174 = ($173<<24>>24)<(0);
     if ($174) {
      $175 = ((($$0)) + 4|0);
      $176 = HEAP32[$175>>2]|0;
      $179 = $176;
     } else {
      $177 = $173&255;
      $179 = $177;
     }
     $178 = ($179|0)==($123|0);
     if ($178) {
      $$7 = $$6;
     } else {
      $180 = (($$6) + -1)|0;
      HEAP8[$$2111>>0] = 0;
      $$7 = $180;
     }
    } else {
     $$7 = $$6;
    }
    $181 = ((($$0)) + 12|0);
    $182 = ((($$2111)) + 1|0);
    $$0 = $181;$$2111 = $182;$$6 = $$7;
   }
  }
  L86: do {
   if ((label|0) == 29) {
    $72 = ($71|0)==(0|0);
    do {
     if ($72) {
      $198 = 1;
     } else {
      $73 = ((($71)) + 12|0);
      $74 = HEAP32[$73>>2]|0;
      $75 = ((($71)) + 16|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ($74|0)==($76|0);
      if ($77) {
       $78 = HEAP32[$71>>2]|0;
       $79 = ((($78)) + 36|0);
       $80 = HEAP32[$79>>2]|0;
       __THREW__ = 0;
       $81 = (invoke_ii($80|0,($71|0))|0);
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        break L86;
       } else {
        $$0$i$i$i$i143 = $81;
       }
      } else {
       $84 = HEAP32[$74>>2]|0;
       $85 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($84)|0);
       $$0$i$i$i$i143 = $85;
      }
      $86 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $87 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i143,$86)|0);
      if ($87) {
       HEAP32[$0>>2] = 0;
       $198 = 1;
       break;
      } else {
       $$pre$i$i$i145 = HEAP32[$0>>2]|0;
       $phitmp$i$i$i146 = ($$pre$i$i$i145|0)==(0|0);
       $198 = $phitmp$i$i$i146;
       break;
      }
     }
    } while(0);
    $88 = ($$pre$i4$i$i151|0)==(0|0);
    do {
     if ($88) {
      label = 76;
     } else {
      $89 = ((($$pre$i4$i$i151)) + 12|0);
      $90 = HEAP32[$89>>2]|0;
      $91 = ((($$pre$i4$i$i151)) + 16|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = ($90|0)==($92|0);
      if ($93) {
       $94 = HEAP32[$$pre$i4$i$i151>>2]|0;
       $95 = ((($94)) + 36|0);
       $96 = HEAP32[$95>>2]|0;
       __THREW__ = 0;
       $97 = (invoke_ii($96|0,($$pre$i4$i$i151|0))|0);
       $98 = __THREW__; __THREW__ = 0;
       $99 = $98&1;
       if ($99) {
        break L86;
       } else {
        $$0$i$i2$i$i149 = $97;
       }
      } else {
       $100 = HEAP32[$90>>2]|0;
       $101 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($100)|0);
       $$0$i$i2$i$i149 = $101;
      }
      $102 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $103 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i149,$102)|0);
      if ($103) {
       HEAP32[$1>>2] = 0;
       label = 76;
       break;
      } else {
       if ($198) {
        $$0133 = $2;$$3 = $$0132;
        break;
       } else {
        label = 78;
        break;
       }
      }
     }
    } while(0);
    if ((label|0) == 76) {
     if ($198) {
      label = 78;
     } else {
      $$0133 = $2;$$3 = $$0132;
     }
    }
    if ((label|0) == 78) {
     $183 = HEAP32[$5>>2]|0;
     $184 = $183 | 2;
     HEAP32[$5>>2] = $184;
     $$0133 = $2;$$3 = $$0132;
    }
    while(1) {
     $185 = ($$0133|0)==($3|0);
     if ($185) {
      label = 82;
      break;
     }
     $186 = HEAP8[$$3>>0]|0;
     $187 = ($186<<24>>24)==(2);
     if ($187) {
      $$0133197 = $$0133;
      break;
     }
     $188 = ((($$0133)) + 12|0);
     $189 = ((($$3)) + 1|0);
     $$0133 = $188;$$3 = $189;
    }
    if ((label|0) == 82) {
     $190 = HEAP32[$5>>2]|0;
     $191 = $190 | 4;
     HEAP32[$5>>2] = $191;
     $$0133197 = $3;
    }
    $192 = ($$sroa$0165$0|0)==(0);
    if (!($192)) {
     $193 = $$sroa$0165$0;
     _free($193);
    }
    STACKTOP = sp;return ($$0133197|0);
   }
  } while(0);
  $194 = ___cxa_find_matching_catch_2()|0;
  $195 = tempRet0;
  $196 = ($$sroa$0165$0|0)==(0);
  if ($196) {
   $$sink140$sink141205$index = $194;$$sink140$sink141205$index5 = $195;
  } else {
   $197 = $$sroa$0165$0;
   _free($197);
   $$sink140$sink141205$index = $194;$$sink140$sink141205$index5 = $195;
  }
 }
 ___resumeException($$sink140$sink141205$index|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcb($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$0$i$i = 0, $$011 = 0, $$byval_copy = 0, $$pr = 0, $$pr$pre = 0, $$sroa$0$0 = 0, $$sroa$0$0$copyload = 0, $$sroa$013$0 = 0, $$sroa$013$0$ph = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 20|0;
 $5 = sp + 16|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = ((($2)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 & 1;
 $11 = ($10|0)==(0);
 L1: do {
  if ($11) {
   $12 = HEAP32[$0>>2]|0;
   $13 = ((($12)) + 24|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$1>>2]|0;
   HEAP32[$5>>2] = $15;
   $16 = $4&1;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$5>>2]|0;
   $17 = (FUNCTION_TABLE_iiiiii[$14 & 2047]($0,$$byval_copy,$2,$3,$16)|0);
   $$sroa$0$0 = $17;
  } else {
   __ZNKSt3__28ios_base6getlocEv($6,$2);
   __THREW__ = 0;
   $18 = (invoke_iii(734,($6|0),(165456|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   do {
    if ($20) {
     $40 = ___cxa_find_matching_catch_2()|0;
     $41 = tempRet0;
     __ZNSt3__26localeD2Ev($6);
     $$0 = $41;$$011 = $40;
    } else {
     __ZNSt3__26localeD2Ev($6);
     $21 = HEAP32[$18>>2]|0;
     if ($4) {
      $22 = ((($21)) + 24|0);
      $23 = HEAP32[$22>>2]|0;
      FUNCTION_TABLE_vii[$23 & 2047]($7,$18);
     } else {
      $24 = ((($21)) + 28|0);
      $25 = HEAP32[$24>>2]|0;
      FUNCTION_TABLE_vii[$25 & 2047]($7,$18);
     }
     $26 = ((($7)) + 11|0);
     $27 = HEAP8[$26>>0]|0;
     $28 = ($27<<24>>24)<(0);
     if ($28) {
      $29 = HEAP32[$7>>2]|0;
      $$pr = $27;$$sroa$013$0$ph = $29;
      label = 9;
     } else {
      $$sroa$013$0 = $7;$31 = $27;
     }
     while(1) {
      if ((label|0) == 9) {
       label = 0;
       $$sroa$013$0 = $$sroa$013$0$ph;$31 = $$pr;
      }
      $30 = ($31<<24>>24)<(0);
      $32 = HEAP32[$7>>2]|0;
      $33 = ((($7)) + 4|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = $31&255;
      $36 = $30 ? $32 : $7;
      $37 = $30 ? $34 : $35;
      $38 = (($36) + ($37)|0);
      $39 = ($$sroa$013$0|0)==($38|0);
      if ($39) {
       label = 11;
       break;
      }
      $42 = HEAP8[$$sroa$013$0>>0]|0;
      $43 = HEAP32[$1>>2]|0;
      $44 = ($43|0)==(0|0);
      if (!($44)) {
       $45 = ((($43)) + 24|0);
       $46 = HEAP32[$45>>2]|0;
       $47 = ((($43)) + 28|0);
       $48 = HEAP32[$47>>2]|0;
       $49 = ($46|0)==($48|0);
       if ($49) {
        $50 = HEAP32[$43>>2]|0;
        $51 = ((($50)) + 52|0);
        $52 = HEAP32[$51>>2]|0;
        $53 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($42)|0);
        __THREW__ = 0;
        $54 = (invoke_iii($52|0,($43|0),($53|0))|0);
        $55 = __THREW__; __THREW__ = 0;
        $56 = $55&1;
        if ($56) {
         label = 20;
         break;
        } else {
         $$0$i$i = $54;
        }
       } else {
        $57 = ((($46)) + 1|0);
        HEAP32[$45>>2] = $57;
        HEAP8[$46>>0] = $42;
        $58 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($42)|0);
        $$0$i$i = $58;
       }
       $59 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $60 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i,$59)|0);
       if ($60) {
        HEAP32[$1>>2] = 0;
       }
      }
      $61 = ((($$sroa$013$0)) + 1|0);
      $$pr$pre = HEAP8[$26>>0]|0;
      $$pr = $$pr$pre;$$sroa$013$0$ph = $61;
      label = 9;
     }
     if ((label|0) == 11) {
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
      $$sroa$0$0 = $$sroa$0$0$copyload;
      break L1;
     }
     else if ((label|0) == 20) {
      $62 = ___cxa_find_matching_catch_2()|0;
      $63 = tempRet0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
      $$0 = $63;$$011 = $62;
      break;
     }
    }
   } while(0);
   ___resumeException($$011|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$0|0);
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcl($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $$byval_copy = sp + 20|0;
 $vararg_buffer = sp;
 $5 = sp + 60|0;
 $6 = sp + 47|0;
 $7 = sp + 24|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156075>>0]|0;HEAP8[$5+1>>0]=HEAP8[156075+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156075+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156075+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156075+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156075+5>>0]|0;
 $12 = ((($5)) + 1|0);
 $13 = ((($2)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($12,160348,1,$14);
 $15 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $16 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,13,$15,$5,$vararg_buffer)|0);
 $17 = (($6) + ($16)|0);
 $18 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$17,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($10,$2);
 __THREW__ = 0;
 invoke_viiiiiii(975,($6|0),($18|0),($17|0),($7|0),($8|0),($9|0),($10|0));
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZNSt3__26localeD2Ev($10);
  ___resumeException($25|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($10);
  $21 = HEAP32[$1>>2]|0;
  HEAP32[$11>>2] = $21;
  $22 = HEAP32[$8>>2]|0;
  $23 = HEAP32[$9>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$11>>2]|0;
  $24 = (__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$22,$23,$2,$3)|0);
  STACKTOP = sp;return ($24|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcx($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $$byval_copy = sp + 32|0;
 $vararg_buffer = sp + 8|0;
 $6 = sp;
 $7 = sp + 79|0;
 $8 = sp + 36|0;
 $9 = sp + 28|0;
 $10 = sp + 24|0;
 $11 = sp + 20|0;
 $12 = sp + 16|0;
 $13 = $6;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($6)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($17,160353,1,$19);
 $20 = (__ZNSt3__26__clocEv()|0);
 $21 = $vararg_buffer;
 $22 = $21;
 HEAP32[$22>>2] = $4;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = $5;
 $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($7,23,$20,$6,$vararg_buffer)|0);
 $26 = (($7) + ($25)|0);
 $27 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($7,$26,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($11,$2);
 __THREW__ = 0;
 invoke_viiiiiii(975,($7|0),($27|0),($26|0),($8|0),($9|0),($10|0),($11|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  __ZNSt3__26localeD2Ev($11);
  ___resumeException($34|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($11);
  $30 = HEAP32[$1>>2]|0;
  HEAP32[$12>>2] = $30;
  $31 = HEAP32[$9>>2]|0;
  $32 = HEAP32[$10>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
  $33 = (__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$8,$31,$32,$2,$3)|0);
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcm($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $$byval_copy = sp + 20|0;
 $vararg_buffer = sp;
 $5 = sp + 60|0;
 $6 = sp + 48|0;
 $7 = sp + 24|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156075>>0]|0;HEAP8[$5+1>>0]=HEAP8[156075+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156075+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156075+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156075+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156075+5>>0]|0;
 $12 = ((($5)) + 1|0);
 $13 = ((($2)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($12,160348,0,$14);
 $15 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $16 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,12,$15,$5,$vararg_buffer)|0);
 $17 = (($6) + ($16)|0);
 $18 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$17,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($10,$2);
 __THREW__ = 0;
 invoke_viiiiiii(975,($6|0),($18|0),($17|0),($7|0),($8|0),($9|0),($10|0));
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZNSt3__26localeD2Ev($10);
  ___resumeException($25|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($10);
  $21 = HEAP32[$1>>2]|0;
  HEAP32[$11>>2] = $21;
  $22 = HEAP32[$8>>2]|0;
  $23 = HEAP32[$9>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$11>>2]|0;
  $24 = (__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$22,$23,$2,$3)|0);
  STACKTOP = sp;return ($24|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcy($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $$byval_copy = sp + 32|0;
 $vararg_buffer = sp + 8|0;
 $6 = sp;
 $7 = sp + 79|0;
 $8 = sp + 36|0;
 $9 = sp + 28|0;
 $10 = sp + 24|0;
 $11 = sp + 20|0;
 $12 = sp + 16|0;
 $13 = $6;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($6)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($17,160353,0,$19);
 $20 = (__ZNSt3__26__clocEv()|0);
 $21 = $vararg_buffer;
 $22 = $21;
 HEAP32[$22>>2] = $4;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = $5;
 $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($7,23,$20,$6,$vararg_buffer)|0);
 $26 = (($7) + ($25)|0);
 $27 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($7,$26,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($11,$2);
 __THREW__ = 0;
 invoke_viiiiiii(975,($7|0),($27|0),($26|0),($8|0),($9|0),($10|0),($11|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  __ZNSt3__26localeD2Ev($11);
  ___resumeException($34|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($11);
  $30 = HEAP32[$1>>2]|0;
  HEAP32[$12>>2] = $30;
  $31 = HEAP32[$9>>2]|0;
  $32 = HEAP32[$10>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
  $33 = (__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$8,$31,$32,$2,$3)|0);
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcd($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = +$4;
 var $$0 = 0, $$046 = 0, $$047 = 0, $$050 = 0, $$1 = 0, $$148 = 0, $$151 = 0, $$2 = 0, $$24982 = 0, $$25281 = 0, $$45484 = 0, $$485 = 0, $$byval_copy = 0, $$pre = 0, $$sroa$060$0 = 0, $$sroa$071$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr8 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $$byval_copy = sp + 76|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $5 = sp;
 $6 = sp + 138|0;
 $7 = sp + 72|0;
 $8 = sp + 80|0;
 $9 = sp + 68|0;
 $10 = sp + 64|0;
 $11 = sp + 60|0;
 $12 = sp + 56|0;
 $13 = $5;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($5)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (__ZNSt3__214__num_put_base14__format_floatEPcPKcj($17,169457,$19)|0);
 HEAP32[$7>>2] = $6;
 $21 = (__ZNSt3__26__clocEv()|0);
 if ($20) {
  $22 = ((($2)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  HEAP32[$vararg_buffer>>2] = $23;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $4;
  $24 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer)|0);
  $$0 = $24;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $4;
  $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer2)|0);
  $$0 = $25;
 }
 $26 = ($$0|0)>(29);
 L5: do {
  if ($26) {
   if ($20) {
    __THREW__ = 0;
    $27 = (invoke_i(955)|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     $30 = ((($2)) + 8|0);
     $31 = HEAP32[$30>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $31;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $4;
     $32 = (invoke_iiiii(976,($7|0),($27|0),($5|0),($vararg_buffer5|0))|0);
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if (!($34)) {
      $$1 = $32;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $37 = (invoke_i(955)|0);
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $4;
     $40 = (invoke_iiiii(976,($7|0),($37|0),($5|0),($vararg_buffer9|0))|0);
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if (!($42)) {
      $$1 = $40;
      label = 12;
     }
    }
   }
   do {
    if ((label|0) == 12) {
     $43 = HEAP32[$7>>2]|0;
     $44 = ($43|0)==(0|0);
     if ($44) {
      __THREW__ = 0;
      invoke_v(732);
      $45 = __THREW__; __THREW__ = 0;
      break;
     } else {
      $46 = $43;
      $$2 = $$1;$$sroa$071$0 = $46;$48 = $43;
      label = 15;
      break L5;
     }
    }
   } while(0);
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$45484 = $36;$$485 = $35;
  } else {
   $$pre = HEAP32[$7>>2]|0;
   $$2 = $$0;$$sroa$071$0 = 0;$48 = $$pre;
   label = 15;
  }
 } while(0);
 if ((label|0) == 15) {
  $47 = (($48) + ($$2)|0);
  $49 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($48,$47,$2)|0);
  $50 = ($48|0)==($6|0);
  do {
   if ($50) {
    $$046 = $8;$$sroa$060$0 = 0;
    label = 19;
   } else {
    $51 = $$2 << 1;
    $52 = (_malloc($51)|0);
    $53 = ($52|0)==(0|0);
    if ($53) {
     __THREW__ = 0;
     invoke_v(732);
     $54 = __THREW__; __THREW__ = 0;
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $$24982 = $55;$$25281 = $56;
     break;
    } else {
     $57 = $52;
     $$046 = $52;$$sroa$060$0 = $57;
     label = 19;
     break;
    }
   }
  } while(0);
  if ((label|0) == 19) {
   __THREW__ = 0;
   invoke_vii(733,($11|0),($2|0));
   $58 = __THREW__; __THREW__ = 0;
   $59 = $58&1;
   do {
    if ($59) {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $$047 = $72;$$050 = $73;
     label = 29;
    } else {
     __THREW__ = 0;
     invoke_viiiiiii(977,($48|0),($49|0),($47|0),($$046|0),($9|0),($10|0),($11|0));
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      $74 = ___cxa_find_matching_catch_2()|0;
      $75 = tempRet0;
      __ZNSt3__26localeD2Ev($11);
      $$047 = $74;$$050 = $75;
      label = 29;
      break;
     }
     __ZNSt3__26localeD2Ev($11);
     $62 = HEAP32[$1>>2]|0;
     HEAP32[$12>>2] = $62;
     $63 = HEAP32[$9>>2]|0;
     $64 = HEAP32[$10>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
     $65 = (invoke_iiiiiii(978,($$byval_copy|0),($$046|0),($63|0),($64|0),($2|0),($3|0))|0);
     $66 = __THREW__; __THREW__ = 0;
     $67 = $66&1;
     if ($67) {
      $76 = ___cxa_find_matching_catch_2()|0;
      $77 = tempRet0;
      $$148 = $76;$$151 = $77;
      break;
     }
     HEAP32[$1>>2] = $65;
     $68 = ($$sroa$060$0|0)==(0);
     if (!($68)) {
      $69 = $$sroa$060$0;
      _free($69);
     }
     $70 = ($$sroa$071$0|0)==(0);
     if (!($70)) {
      $71 = $$sroa$071$0;
      _free($71);
     }
     STACKTOP = sp;return ($65|0);
    }
   } while(0);
   if ((label|0) == 29) {
    $$148 = $$047;$$151 = $$050;
   }
   $78 = ($$sroa$060$0|0)==(0);
   if ($78) {
    $$24982 = $$148;$$25281 = $$151;
   } else {
    $79 = $$sroa$060$0;
    _free($79);
    $$24982 = $$148;$$25281 = $$151;
   }
  }
  $80 = ($$sroa$071$0|0)==(0);
  if ($80) {
   $$45484 = $$25281;$$485 = $$24982;
  } else {
   $81 = $$sroa$071$0;
   _free($81);
   $$45484 = $$25281;$$485 = $$24982;
  }
 }
 ___resumeException($$485|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEce($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = +$4;
 var $$0 = 0, $$046 = 0, $$047 = 0, $$050 = 0, $$1 = 0, $$148 = 0, $$151 = 0, $$2 = 0, $$24982 = 0, $$25281 = 0, $$45484 = 0, $$485 = 0, $$byval_copy = 0, $$pre = 0, $$sroa$060$0 = 0, $$sroa$071$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr8 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $$byval_copy = sp + 76|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $5 = sp;
 $6 = sp + 138|0;
 $7 = sp + 72|0;
 $8 = sp + 80|0;
 $9 = sp + 68|0;
 $10 = sp + 64|0;
 $11 = sp + 60|0;
 $12 = sp + 56|0;
 $13 = $5;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($5)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (__ZNSt3__214__num_put_base14__format_floatEPcPKcj($17,156073,$19)|0);
 HEAP32[$7>>2] = $6;
 $21 = (__ZNSt3__26__clocEv()|0);
 if ($20) {
  $22 = ((($2)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  HEAP32[$vararg_buffer>>2] = $23;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $4;
  $24 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer)|0);
  $$0 = $24;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $4;
  $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer2)|0);
  $$0 = $25;
 }
 $26 = ($$0|0)>(29);
 L5: do {
  if ($26) {
   if ($20) {
    __THREW__ = 0;
    $27 = (invoke_i(955)|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     $30 = ((($2)) + 8|0);
     $31 = HEAP32[$30>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $31;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $4;
     $32 = (invoke_iiiii(976,($7|0),($27|0),($5|0),($vararg_buffer5|0))|0);
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if (!($34)) {
      $$1 = $32;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $37 = (invoke_i(955)|0);
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $4;
     $40 = (invoke_iiiii(976,($7|0),($37|0),($5|0),($vararg_buffer9|0))|0);
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if (!($42)) {
      $$1 = $40;
      label = 12;
     }
    }
   }
   do {
    if ((label|0) == 12) {
     $43 = HEAP32[$7>>2]|0;
     $44 = ($43|0)==(0|0);
     if ($44) {
      __THREW__ = 0;
      invoke_v(732);
      $45 = __THREW__; __THREW__ = 0;
      break;
     } else {
      $46 = $43;
      $$2 = $$1;$$sroa$071$0 = $46;$48 = $43;
      label = 15;
      break L5;
     }
    }
   } while(0);
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$45484 = $36;$$485 = $35;
  } else {
   $$pre = HEAP32[$7>>2]|0;
   $$2 = $$0;$$sroa$071$0 = 0;$48 = $$pre;
   label = 15;
  }
 } while(0);
 if ((label|0) == 15) {
  $47 = (($48) + ($$2)|0);
  $49 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($48,$47,$2)|0);
  $50 = ($48|0)==($6|0);
  do {
   if ($50) {
    $$046 = $8;$$sroa$060$0 = 0;
    label = 19;
   } else {
    $51 = $$2 << 1;
    $52 = (_malloc($51)|0);
    $53 = ($52|0)==(0|0);
    if ($53) {
     __THREW__ = 0;
     invoke_v(732);
     $54 = __THREW__; __THREW__ = 0;
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $$24982 = $55;$$25281 = $56;
     break;
    } else {
     $57 = $52;
     $$046 = $52;$$sroa$060$0 = $57;
     label = 19;
     break;
    }
   }
  } while(0);
  if ((label|0) == 19) {
   __THREW__ = 0;
   invoke_vii(733,($11|0),($2|0));
   $58 = __THREW__; __THREW__ = 0;
   $59 = $58&1;
   do {
    if ($59) {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $$047 = $72;$$050 = $73;
     label = 29;
    } else {
     __THREW__ = 0;
     invoke_viiiiiii(977,($48|0),($49|0),($47|0),($$046|0),($9|0),($10|0),($11|0));
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      $74 = ___cxa_find_matching_catch_2()|0;
      $75 = tempRet0;
      __ZNSt3__26localeD2Ev($11);
      $$047 = $74;$$050 = $75;
      label = 29;
      break;
     }
     __ZNSt3__26localeD2Ev($11);
     $62 = HEAP32[$1>>2]|0;
     HEAP32[$12>>2] = $62;
     $63 = HEAP32[$9>>2]|0;
     $64 = HEAP32[$10>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
     $65 = (invoke_iiiiiii(978,($$byval_copy|0),($$046|0),($63|0),($64|0),($2|0),($3|0))|0);
     $66 = __THREW__; __THREW__ = 0;
     $67 = $66&1;
     if ($67) {
      $76 = ___cxa_find_matching_catch_2()|0;
      $77 = tempRet0;
      $$148 = $76;$$151 = $77;
      break;
     }
     HEAP32[$1>>2] = $65;
     $68 = ($$sroa$060$0|0)==(0);
     if (!($68)) {
      $69 = $$sroa$060$0;
      _free($69);
     }
     $70 = ($$sroa$071$0|0)==(0);
     if (!($70)) {
      $71 = $$sroa$071$0;
      _free($71);
     }
     STACKTOP = sp;return ($65|0);
    }
   } while(0);
   if ((label|0) == 29) {
    $$148 = $$047;$$151 = $$050;
   }
   $78 = ($$sroa$060$0|0)==(0);
   if ($78) {
    $$24982 = $$148;$$25281 = $$151;
   } else {
    $79 = $$sroa$060$0;
    _free($79);
    $$24982 = $$148;$$25281 = $$151;
   }
  }
  $80 = ($$sroa$071$0|0)==(0);
  if ($80) {
   $$45484 = $$25281;$$485 = $$24982;
  } else {
   $81 = $$sroa$071$0;
   _free($81);
   $$45484 = $$25281;$$485 = $$24982;
  }
 }
 ___resumeException($$485|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPKv($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $$byval_copy = sp + 12|0;
 $vararg_buffer = sp;
 $5 = sp + 76|0;
 $6 = sp + 56|0;
 $7 = sp + 16|0;
 $8 = sp + 8|0;
 $9 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156067>>0]|0;HEAP8[$5+1>>0]=HEAP8[156067+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156067+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156067+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156067+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156067+5>>0]|0;
 $10 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $11 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,20,$10,$5,$vararg_buffer)|0);
 $12 = (($6) + ($11)|0);
 $13 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$12,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($8,$2);
 __THREW__ = 0;
 $14 = (invoke_iii(734,($8|0),(165440|0))|0);
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($28|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($8);
  $17 = HEAP32[$14>>2]|0;
  $18 = ((($17)) + 32|0);
  $19 = HEAP32[$18>>2]|0;
  (FUNCTION_TABLE_iiiii[$19 & 2047]($14,$6,$12,$7)|0);
  $20 = (($7) + ($11)|0);
  $21 = ($13|0)==($12|0);
  $22 = $6;
  $23 = $13;
  $24 = (($23) - ($22))|0;
  $25 = (($7) + ($24)|0);
  $$0 = $21 ? $20 : $25;
  $26 = HEAP32[$1>>2]|0;
  HEAP32[$9>>2] = $26;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
  $27 = (__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$$0,$20,$2,$3)|0);
  STACKTOP = sp;return ($27|0);
 }
 return (0)|0;
}
function __ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($0,$1,$2,$3,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $varargs = $varargs|0;
 var $10 = 0, $11 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 HEAP32[$4>>2] = $varargs;
 $5 = (___uselocale($2)|0);
 $6 = (_vsnprintf($0,$1,$3,$4)|0);
 $7 = ($5|0)==(0|0);
 if (!($7)) {
  __THREW__ = 0;
  (invoke_ii(958,($5|0))|0);
  $8 = __THREW__; __THREW__ = 0;
  $9 = $8&1;
  if ($9) {
   $10 = ___cxa_find_matching_catch_3(0|0)|0;
   $11 = tempRet0;
   ___clang_call_terminate($10);
   // unreachable;
  }
 }
 STACKTOP = sp;return ($6|0);
}
function __ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $trunc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($2)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = $4 & 176;
 $trunc = $5&255;
 L1: do {
  switch ($trunc<<24>>24) {
  case 16:  {
   $6 = HEAP8[$0>>0]|0;
   $7 = ((($0)) + 1|0);
   switch ($6<<24>>24) {
   case 43: case 45:  {
    $$0 = $7;
    break L1;
    break;
   }
   default: {
   }
   }
   $8 = $1;
   $9 = $0;
   $10 = (($8) - ($9))|0;
   $11 = ($10|0)>(1);
   $12 = ($6<<24>>24)==(48);
   $or$cond = $11 & $12;
   if ($or$cond) {
    $13 = HEAP8[$7>>0]|0;
    switch ($13<<24>>24) {
    case 88: case 120:  {
     break;
    }
    default: {
     label = 6;
     break L1;
    }
    }
    $14 = ((($0)) + 2|0);
    $$0 = $14;
   } else {
    label = 6;
   }
   break;
  }
  case 32:  {
   $$0 = $1;
   break;
  }
  default: {
   label = 6;
  }
  }
 } while(0);
 if ((label|0) == 6) {
  $$0 = $0;
 }
 return ($$0|0);
}
function __ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $$0$i$i = 0, $$pre$phi45Z2D = 0, $$pre$phi47Z2D = 0, $$pre44 = 0, $$sroa$0$2 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $6 = sp + 12|0;
 $7 = sp;
 $8 = HEAP32[$0>>2]|0;
 $9 = ($8|0)==(0|0);
 L1: do {
  if ($9) {
   $$sroa$0$2 = 0;
  } else {
   $10 = $3;
   $11 = $1;
   $12 = (($10) - ($11))|0;
   $13 = ((($4)) + 12|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($14|0)>($12|0);
   $16 = (($14) - ($12))|0;
   $$0 = $15 ? $16 : 0;
   $17 = $2;
   $18 = (($17) - ($11))|0;
   $19 = ($18|0)>(0);
   if ($19) {
    $20 = HEAP32[$8>>2]|0;
    $21 = ((($20)) + 48|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = (FUNCTION_TABLE_iiii[$22 & 2047]($8,$1,$18)|0);
    $24 = ($23|0)==($18|0);
    if (!($24)) {
     HEAP32[$0>>2] = 0;
     $$sroa$0$2 = 0;
     break;
    }
   }
   $25 = ($$0|0)>(0);
   do {
    if ($25) {
     ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
     $26 = ($$0>>>0)<(11);
     if ($26) {
      $27 = $$0&255;
      $28 = ((($7)) + 11|0);
      HEAP8[$28>>0] = $27;
      $$0$i$i = $7;$$pre$phi45Z2D = $28;$$pre$phi47Z2D = $7;
     } else {
      $29 = (($$0) + 16)|0;
      $30 = $29 & -16;
      $31 = (__Znwj($30)|0);
      HEAP32[$7>>2] = $31;
      $32 = $30 | -2147483648;
      $33 = ((($7)) + 8|0);
      HEAP32[$33>>2] = $32;
      $34 = ((($7)) + 4|0);
      HEAP32[$34>>2] = $$0;
      $$pre44 = ((($7)) + 11|0);
      $$0$i$i = $31;$$pre$phi45Z2D = $$pre44;$$pre$phi47Z2D = $7;
     }
     (__ZNSt3__211char_traitsIcE6assignEPcjc($$0$i$i,$$0,$5)|0);
     $35 = (($$0$i$i) + ($$0)|0);
     HEAP8[$6>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($35,$6);
     $36 = HEAP8[$$pre$phi45Z2D>>0]|0;
     $37 = ($36<<24>>24)<(0);
     $38 = HEAP32[$$pre$phi47Z2D>>2]|0;
     $39 = $37 ? $38 : $7;
     $40 = HEAP32[$8>>2]|0;
     $41 = ((($40)) + 48|0);
     $42 = HEAP32[$41>>2]|0;
     __THREW__ = 0;
     $43 = (invoke_iiii($42|0,($8|0),($39|0),($$0|0))|0);
     $44 = __THREW__; __THREW__ = 0;
     $45 = $44&1;
     if ($45) {
      $47 = ___cxa_find_matching_catch_2()|0;
      $48 = tempRet0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
      ___resumeException($47|0);
      // unreachable;
     }
     $46 = ($43|0)==($$0|0);
     if ($46) {
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
      break;
     } else {
      HEAP32[$0>>2] = 0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
      $$sroa$0$2 = 0;
      break L1;
     }
    }
   } while(0);
   $49 = (($10) - ($17))|0;
   $50 = ($49|0)>(0);
   if ($50) {
    $51 = HEAP32[$8>>2]|0;
    $52 = ((($51)) + 48|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = (FUNCTION_TABLE_iiii[$53 & 2047]($8,$2,$49)|0);
    $55 = ($54|0)==($49|0);
    if (!($55)) {
     HEAP32[$0>>2] = 0;
     $$sroa$0$2 = 0;
     break;
    }
   }
   HEAP32[$13>>2] = 0;
   $$sroa$0$2 = $8;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$2|0);
}
function __ZNSt3__211char_traitsIcE6assignEPcjc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($1|0)==(0);
 if (!($3)) {
  $4 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($2)|0);
  $5 = $4&255;
  _memset(($0|0),($5|0),($1|0))|0;
 }
 return ($0|0);
}
function __ZNSt3__214__num_put_base14__format_floatEPcPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$026 = 0, $$027$off0$ph = 0, $$1 = 0, $$2 = 0, $$2$ph = 0, $$sink = 0, $$sink1 = 0, $$sink1$sink = 0, $$sink2 = 0, $$sink3 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $trunc = 0, $trunc$clear = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $2 & 2048;
 $4 = ($3|0)==(0);
 if ($4) {
  $$0 = $0;
 } else {
  $5 = ((($0)) + 1|0);
  HEAP8[$0>>0] = 43;
  $$0 = $5;
 }
 $6 = $2 & 1024;
 $7 = ($6|0)==(0);
 if ($7) {
  $$1 = $$0;
 } else {
  $8 = ((($$0)) + 1|0);
  HEAP8[$$0>>0] = 35;
  $$1 = $8;
 }
 $9 = $2 & 260;
 $10 = $2 & 16384;
 $11 = ($10|0)!=(0);
 $12 = ($9|0)==(260);
 if ($12) {
  $$027$off0$ph = 0;$$2$ph = $$1;
 } else {
  $13 = ((($$1)) + 1|0);
  HEAP8[$$1>>0] = 46;
  $14 = ((($$1)) + 2|0);
  HEAP8[$13>>0] = 42;
  $$027$off0$ph = 1;$$2$ph = $14;
 }
 $$026 = $1;$$2 = $$2$ph;
 while(1) {
  $15 = HEAP8[$$026>>0]|0;
  $16 = ($15<<24>>24)==(0);
  if ($16) {
   break;
  }
  $17 = ((($$026)) + 1|0);
  $18 = ((($$2)) + 1|0);
  HEAP8[$$2>>0] = $15;
  $$026 = $17;$$2 = $18;
 }
 $trunc = $9&65535;
 $trunc$clear = $trunc & 511;
 L14: do {
  switch ($trunc$clear<<16>>16) {
  case 4:  {
   $$sink = $11 ? 70 : 102;
   $$sink1$sink = $$sink;
   break;
  }
  case 256:  {
   $$sink1 = $11 ? 69 : 101;
   $$sink1$sink = $$sink1;
   break;
  }
  default: {
   if ($12) {
    $$sink2 = $11 ? 65 : 97;
    $$sink1$sink = $$sink2;
    break L14;
   } else {
    $$sink3 = $11 ? 71 : 103;
    $$sink1$sink = $$sink3;
    break L14;
   }
  }
  }
 } while(0);
 HEAP8[$$2>>0] = $$sink1$sink;
 return ($$027$off0$ph|0);
}
function __ZNSt3__219__libcpp_asprintf_lEPPcP15__locale_structPKcz($0,$1,$2,$varargs) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $varargs = $varargs|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 HEAP32[$3>>2] = $varargs;
 $4 = (___uselocale($1)|0);
 __THREW__ = 0;
 $5 = (invoke_iiii(979,($0|0),($2|0),($3|0))|0);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  $15 = ($4|0)==(0|0);
  if (!($15)) {
   __THREW__ = 0;
   (invoke_ii(958,($4|0))|0);
   $16 = __THREW__; __THREW__ = 0;
   $17 = $16&1;
   if ($17) {
    $18 = ___cxa_find_matching_catch_3(0|0)|0;
    $19 = tempRet0;
    ___clang_call_terminate($18);
    // unreachable;
   }
  }
  ___resumeException($13|0);
  // unreachable;
 } else {
  $8 = ($4|0)==(0|0);
  if (!($8)) {
   __THREW__ = 0;
   (invoke_ii(958,($4|0))|0);
   $9 = __THREW__; __THREW__ = 0;
   $10 = $9&1;
   if ($10) {
    $11 = ___cxa_find_matching_catch_3(0|0)|0;
    $12 = tempRet0;
    ___clang_call_terminate($11);
    // unreachable;
   }
  }
  STACKTOP = sp;return ($5|0);
 }
 return (0)|0;
}
function __ZNSt3__29__num_putIcE23__widen_and_group_floatEPcS2_S2_S2_RS2_S3_RKNS_6localeE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$0108 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i128 = 0, $$0108 = 0, $$0109 = 0, $$0111 = 0, $$0113 = 0, $$07$i$i = 0, $$07$i$i127 = 0, $$1 = 0, $$1110 = 0, $$1112 = 0, $$1114 = 0, $$2 = 0, $$2115 = 0, $$3 = 0, $$pre$phiZ2D = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $7 = sp;
 $8 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165440)|0);
 $9 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165456)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 FUNCTION_TABLE_vii[$12 & 2047]($7,$9);
 HEAP32[$5>>2] = $3;
 $13 = HEAP8[$0>>0]|0;
 switch ($13<<24>>24) {
 case 43: case 45:  {
  $14 = HEAP32[$8>>2]|0;
  $15 = ((($14)) + 28|0);
  $16 = HEAP32[$15>>2]|0;
  __THREW__ = 0;
  $17 = (invoke_iii($16|0,($8|0),($13|0))|0);
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   $20 = ((($0)) + 1|0);
   $21 = HEAP32[$5>>2]|0;
   $22 = ((($21)) + 1|0);
   HEAP32[$5>>2] = $22;
   HEAP8[$21>>0] = $17;
   $$0111 = $20;
   label = 4;
  }
  break;
 }
 default: {
  $$0111 = $0;
  label = 4;
 }
 }
 L4: do {
  if ((label|0) == 4) {
   $23 = $2;
   $24 = $$0111;
   $25 = (($23) - ($24))|0;
   $26 = ($25|0)>(1);
   L6: do {
    if ($26) {
     $27 = HEAP8[$$0111>>0]|0;
     $28 = ($27<<24>>24)==(48);
     if ($28) {
      $29 = ((($$0111)) + 1|0);
      $30 = HEAP8[$29>>0]|0;
      switch ($30<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$1114 = $$0111;
       label = 15;
       break L6;
      }
      }
      $31 = HEAP32[$8>>2]|0;
      $32 = ((($31)) + 28|0);
      $33 = HEAP32[$32>>2]|0;
      __THREW__ = 0;
      $34 = (invoke_iii($33|0,($8|0),48)|0);
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break L4;
      }
      $37 = HEAP32[$5>>2]|0;
      $38 = ((($37)) + 1|0);
      HEAP32[$5>>2] = $38;
      HEAP8[$37>>0] = $34;
      $39 = ((($$0111)) + 2|0);
      $40 = HEAP8[$29>>0]|0;
      $41 = HEAP32[$8>>2]|0;
      $42 = ((($41)) + 28|0);
      $43 = HEAP32[$42>>2]|0;
      __THREW__ = 0;
      $44 = (invoke_iii($43|0,($8|0),($40|0))|0);
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       break L4;
      }
      $47 = HEAP32[$5>>2]|0;
      $48 = ((($47)) + 1|0);
      HEAP32[$5>>2] = $48;
      HEAP8[$47>>0] = $44;
      $$0113 = $39;
      while(1) {
       $49 = ($$0113>>>0)<($2>>>0);
       if (!($49)) {
        $$1112 = $39;$$2115 = $$0113;
        break L6;
       }
       $50 = HEAP8[$$0113>>0]|0;
       __THREW__ = 0;
       $51 = (invoke_i(955)|0);
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        break L4;
       }
       $54 = $50 << 24 >> 24;
       __THREW__ = 0;
       $55 = (invoke_iii(980,($54|0),($51|0))|0);
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        break L4;
       }
       $58 = ($55|0)==(0);
       if ($58) {
        $$1112 = $39;$$2115 = $$0113;
        break L6;
       }
       $59 = ((($$0113)) + 1|0);
       $$0113 = $59;
      }
     } else {
      $$1114 = $$0111;
      label = 15;
     }
    } else {
     $$1114 = $$0111;
     label = 15;
    }
   } while(0);
   L18: do {
    if ((label|0) == 15) {
     while(1) {
      label = 0;
      $60 = ($$1114>>>0)<($2>>>0);
      if (!($60)) {
       $$1112 = $$0111;$$2115 = $$1114;
       break L18;
      }
      $61 = HEAP8[$$1114>>0]|0;
      __THREW__ = 0;
      $62 = (invoke_i(955)|0);
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       break L4;
      }
      $65 = $61 << 24 >> 24;
      __THREW__ = 0;
      $66 = (invoke_iii(981,($65|0),($62|0))|0);
      $67 = __THREW__; __THREW__ = 0;
      $68 = $67&1;
      if ($68) {
       break L4;
      }
      $69 = ($66|0)==(0);
      if ($69) {
       $$1112 = $$0111;$$2115 = $$1114;
       break L18;
      }
      $70 = ((($$1114)) + 1|0);
      $$1114 = $70;
      label = 15;
     }
    }
   } while(0);
   $71 = ((($7)) + 11|0);
   $72 = HEAP8[$71>>0]|0;
   $73 = ($72<<24>>24)<(0);
   $74 = ((($7)) + 4|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $72&255;
   $77 = $73 ? $75 : $76;
   $78 = ($77|0)==(0);
   $79 = $$1112;
   $80 = $0;
   L26: do {
    if ($78) {
     $81 = HEAP32[$5>>2]|0;
     $82 = HEAP32[$8>>2]|0;
     $83 = ((($82)) + 32|0);
     $84 = HEAP32[$83>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($84|0,($8|0),($$1112|0),($$2115|0),($81|0))|0);
     $85 = __THREW__; __THREW__ = 0;
     $86 = $85&1;
     if ($86) {
      break L4;
     }
     $87 = $$2115;
     $88 = (($87) - ($79))|0;
     $89 = HEAP32[$5>>2]|0;
     $90 = (($89) + ($88)|0);
     HEAP32[$5>>2] = $90;
     $$pre$phiZ2D = $8;
    } else {
     $91 = ($$1112|0)==($$2115|0);
     L30: do {
      if (!($91)) {
       $$0$i$i = $$2115;$$07$i$i = $$1112;
       while(1) {
        $92 = ((($$0$i$i)) + -1|0);
        $93 = ($$07$i$i>>>0)<($92>>>0);
        if (!($93)) {
         break L30;
        }
        $94 = HEAP8[$$07$i$i>>0]|0;
        $95 = HEAP8[$92>>0]|0;
        HEAP8[$$07$i$i>>0] = $95;
        HEAP8[$92>>0] = $94;
        $96 = ((($$07$i$i)) + 1|0);
        $$0$i$i = $92;$$07$i$i = $96;
       }
      }
     } while(0);
     $97 = HEAP32[$9>>2]|0;
     $98 = ((($97)) + 16|0);
     $99 = HEAP32[$98>>2]|0;
     __THREW__ = 0;
     $100 = (invoke_ii($99|0,($9|0))|0);
     $101 = __THREW__; __THREW__ = 0;
     $102 = $101&1;
     if ($102) {
      break L4;
     }
     $$0 = $$1112;$$0108 = 0;$$0109 = 0;
     while(1) {
      $103 = ($$0>>>0)<($$2115>>>0);
      if (!($103)) {
       break;
      }
      $113 = HEAP8[$71>>0]|0;
      $114 = ($113<<24>>24)<(0);
      $115 = HEAP32[$7>>2]|0;
      $116 = $114 ? $115 : $7;
      $117 = (($116) + ($$0108)|0);
      $118 = HEAP8[$117>>0]|0;
      $119 = ($118<<24>>24)>(0);
      $120 = $118 << 24 >> 24;
      $121 = ($$0109|0)==($120|0);
      $or$cond = $119 & $121;
      if ($or$cond) {
       $122 = HEAP32[$5>>2]|0;
       $123 = ((($122)) + 1|0);
       HEAP32[$5>>2] = $123;
       HEAP8[$122>>0] = $100;
       $124 = HEAP8[$71>>0]|0;
       $125 = ($124<<24>>24)<(0);
       $126 = HEAP32[$74>>2]|0;
       $127 = $124&255;
       $128 = $125 ? $126 : $127;
       $129 = (($128) + -1)|0;
       $130 = ($$0108>>>0)<($129>>>0);
       $131 = $130&1;
       $$$0108 = (($131) + ($$0108))|0;
       $$1 = $$$0108;$$1110 = 0;
      } else {
       $$1 = $$0108;$$1110 = $$0109;
      }
      $132 = HEAP8[$$0>>0]|0;
      $133 = HEAP32[$8>>2]|0;
      $134 = ((($133)) + 28|0);
      $135 = HEAP32[$134>>2]|0;
      __THREW__ = 0;
      $136 = (invoke_iii($135|0,($8|0),($132|0))|0);
      $137 = __THREW__; __THREW__ = 0;
      $138 = $137&1;
      if ($138) {
       break L4;
      }
      $139 = HEAP32[$5>>2]|0;
      $140 = ((($139)) + 1|0);
      HEAP32[$5>>2] = $140;
      HEAP8[$139>>0] = $136;
      $141 = (($$1110) + 1)|0;
      $142 = ((($$0)) + 1|0);
      $$0 = $142;$$0108 = $$1;$$0109 = $141;
     }
     $104 = (($79) - ($80))|0;
     $105 = (($3) + ($104)|0);
     $106 = HEAP32[$5>>2]|0;
     $107 = ($105|0)==($106|0);
     if ($107) {
      $$pre$phiZ2D = $8;
     } else {
      $$0$i$i128 = $106;$$07$i$i127 = $105;
      while(1) {
       $108 = ((($$0$i$i128)) + -1|0);
       $109 = ($$07$i$i127>>>0)<($108>>>0);
       if (!($109)) {
        $$pre$phiZ2D = $8;
        break L26;
       }
       $110 = HEAP8[$$07$i$i127>>0]|0;
       $111 = HEAP8[$108>>0]|0;
       HEAP8[$$07$i$i127>>0] = $111;
       HEAP8[$108>>0] = $110;
       $112 = ((($$07$i$i127)) + 1|0);
       $$0$i$i128 = $108;$$07$i$i127 = $112;
      }
     }
    }
   } while(0);
   $$2 = $$2115;
   while(1) {
    $143 = ($$2>>>0)<($2>>>0);
    if (!($143)) {
     $$3 = $$2;
     break;
    }
    $144 = HEAP8[$$2>>0]|0;
    $145 = ($144<<24>>24)==(46);
    $146 = ((($$2)) + 1|0);
    if ($145) {
     label = 39;
     break;
    }
    $155 = HEAP32[$$pre$phiZ2D>>2]|0;
    $156 = ((($155)) + 28|0);
    $157 = HEAP32[$156>>2]|0;
    __THREW__ = 0;
    $158 = (invoke_iii($157|0,($8|0),($144|0))|0);
    $159 = __THREW__; __THREW__ = 0;
    $160 = $159&1;
    if ($160) {
     break L4;
    }
    $161 = HEAP32[$5>>2]|0;
    $162 = ((($161)) + 1|0);
    HEAP32[$5>>2] = $162;
    HEAP8[$161>>0] = $158;
    $$2 = $146;
   }
   if ((label|0) == 39) {
    $147 = HEAP32[$9>>2]|0;
    $148 = ((($147)) + 12|0);
    $149 = HEAP32[$148>>2]|0;
    __THREW__ = 0;
    $150 = (invoke_ii($149|0,($9|0))|0);
    $151 = __THREW__; __THREW__ = 0;
    $152 = $151&1;
    if ($152) {
     break;
    }
    $153 = HEAP32[$5>>2]|0;
    $154 = ((($153)) + 1|0);
    HEAP32[$5>>2] = $154;
    HEAP8[$153>>0] = $150;
    $$3 = $146;
   }
   $163 = HEAP32[$5>>2]|0;
   $164 = HEAP32[$8>>2]|0;
   $165 = ((($164)) + 32|0);
   $166 = HEAP32[$165>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($166|0,($8|0),($$3|0),($2|0),($163|0))|0);
   $167 = __THREW__; __THREW__ = 0;
   $168 = $167&1;
   if (!($168)) {
    $169 = $$3;
    $170 = (($23) - ($169))|0;
    $171 = HEAP32[$5>>2]|0;
    $172 = (($171) + ($170)|0);
    HEAP32[$5>>2] = $172;
    $173 = ($1|0)==($2|0);
    $174 = $1;
    $175 = (($174) - ($80))|0;
    $176 = (($3) + ($175)|0);
    $storemerge = $173 ? $172 : $176;
    HEAP32[$4>>2] = $storemerge;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 $177 = ___cxa_find_matching_catch_2()|0;
 $178 = tempRet0;
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
 ___resumeException($177|0);
 // unreachable;
}
function __ZNSt3__214__num_put_base12__format_intEPcPKcbj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$014 = 0, $$1 = 0, $$sink = 0, $$sink$sink = 0, $$sink1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $trunc = 0;
 var $trunc$clear = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $3 & 2048;
 $5 = ($4|0)==(0);
 if ($5) {
  $$0 = $0;
 } else {
  $6 = ((($0)) + 1|0);
  HEAP8[$0>>0] = 43;
  $$0 = $6;
 }
 $7 = $3 & 512;
 $8 = ($7|0)==(0);
 if ($8) {
  $$014 = $1;$$1 = $$0;
 } else {
  $9 = ((($$0)) + 1|0);
  HEAP8[$$0>>0] = 35;
  $$014 = $1;$$1 = $9;
 }
 while(1) {
  $10 = HEAP8[$$014>>0]|0;
  $11 = ($10<<24>>24)==(0);
  if ($11) {
   break;
  }
  $12 = ((($$014)) + 1|0);
  $13 = ((($$1)) + 1|0);
  HEAP8[$$1>>0] = $10;
  $$014 = $12;$$1 = $13;
 }
 $14 = $3 & 74;
 $trunc = $14&255;
 $trunc$clear = $trunc & 127;
 switch ($trunc$clear<<24>>24) {
 case 64:  {
  $$sink$sink = 111;
  break;
 }
 case 8:  {
  $15 = $3 & 16384;
  $16 = ($15|0)!=(0);
  $$sink = $16 ? 88 : 120;
  $$sink$sink = $$sink;
  break;
 }
 default: {
  $$sink1 = $2 ? 100 : 117;
  $$sink$sink = $$sink1;
 }
 }
 HEAP8[$$1>>0] = $$sink$sink;
 return;
}
function __ZNSt3__29__num_putIcE21__widen_and_group_intEPcS2_S2_S2_RS2_S3_RKNS_6localeE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$078 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i94 = 0, $$07$i$i = 0, $$07$i$i93 = 0, $$078 = 0, $$079 = 0, $$083 = 0, $$1 = 0, $$180 = 0, $$184 = 0, $$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0;
 var $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0;
 var $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $7 = sp;
 $8 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165440)|0);
 $9 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165456)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 FUNCTION_TABLE_vii[$12 & 2047]($7,$9);
 $13 = ((($7)) + 11|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = ($14<<24>>24)<(0);
 $16 = ((($7)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $14&255;
 $19 = $15 ? $17 : $18;
 $20 = ($19|0)==(0);
 $21 = $2;
 $22 = $0;
 L1: do {
  if ($20) {
   $23 = HEAP32[$8>>2]|0;
   $24 = ((($23)) + 32|0);
   $25 = HEAP32[$24>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($25|0,($8|0),($0|0),($2|0),($3|0))|0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    label = 28;
   } else {
    $28 = (($21) - ($22))|0;
    $29 = (($3) + ($28)|0);
    HEAP32[$5>>2] = $29;
    $122 = $29;
    label = 27;
   }
  } else {
   HEAP32[$5>>2] = $3;
   $30 = HEAP8[$0>>0]|0;
   switch ($30<<24>>24) {
   case 43: case 45:  {
    $31 = HEAP32[$8>>2]|0;
    $32 = ((($31)) + 28|0);
    $33 = HEAP32[$32>>2]|0;
    __THREW__ = 0;
    $34 = (invoke_iii($33|0,($8|0),($30|0))|0);
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if ($36) {
     label = 28;
     break L1;
    }
    $37 = ((($0)) + 1|0);
    $38 = HEAP32[$5>>2]|0;
    $39 = ((($38)) + 1|0);
    HEAP32[$5>>2] = $39;
    HEAP8[$38>>0] = $34;
    $$083 = $37;
    break;
   }
   default: {
    $$083 = $0;
   }
   }
   $40 = $$083;
   $41 = (($21) - ($40))|0;
   $42 = ($41|0)>(1);
   L9: do {
    if ($42) {
     $43 = HEAP8[$$083>>0]|0;
     $44 = ($43<<24>>24)==(48);
     if ($44) {
      $45 = ((($$083)) + 1|0);
      $46 = HEAP8[$45>>0]|0;
      switch ($46<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$184 = $$083;
       break L9;
      }
      }
      $47 = HEAP32[$8>>2]|0;
      $48 = ((($47)) + 28|0);
      $49 = HEAP32[$48>>2]|0;
      __THREW__ = 0;
      $50 = (invoke_iii($49|0,($8|0),48)|0);
      $51 = __THREW__; __THREW__ = 0;
      $52 = $51&1;
      if ($52) {
       label = 28;
       break L1;
      }
      $53 = HEAP32[$5>>2]|0;
      $54 = ((($53)) + 1|0);
      HEAP32[$5>>2] = $54;
      HEAP8[$53>>0] = $50;
      $55 = HEAP8[$45>>0]|0;
      $56 = HEAP32[$8>>2]|0;
      $57 = ((($56)) + 28|0);
      $58 = HEAP32[$57>>2]|0;
      __THREW__ = 0;
      $59 = (invoke_iii($58|0,($8|0),($55|0))|0);
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if ($61) {
       label = 28;
       break L1;
      }
      $62 = ((($$083)) + 2|0);
      $63 = HEAP32[$5>>2]|0;
      $64 = ((($63)) + 1|0);
      HEAP32[$5>>2] = $64;
      HEAP8[$63>>0] = $59;
      $$184 = $62;
     } else {
      $$184 = $$083;
     }
    } else {
     $$184 = $$083;
    }
   } while(0);
   $65 = ($$184|0)==($2|0);
   L16: do {
    if (!($65)) {
     $$0$i$i = $2;$$07$i$i = $$184;
     while(1) {
      $66 = ((($$0$i$i)) + -1|0);
      $67 = ($$07$i$i>>>0)<($66>>>0);
      if (!($67)) {
       break L16;
      }
      $68 = HEAP8[$$07$i$i>>0]|0;
      $69 = HEAP8[$66>>0]|0;
      HEAP8[$$07$i$i>>0] = $69;
      HEAP8[$66>>0] = $68;
      $70 = ((($$07$i$i)) + 1|0);
      $$0$i$i = $66;$$07$i$i = $70;
     }
    }
   } while(0);
   $71 = HEAP32[$9>>2]|0;
   $72 = ((($71)) + 16|0);
   $73 = HEAP32[$72>>2]|0;
   __THREW__ = 0;
   $74 = (invoke_ii($73|0,($9|0))|0);
   $75 = __THREW__; __THREW__ = 0;
   $76 = $75&1;
   if ($76) {
    label = 28;
   } else {
    $$0 = $$184;$$078 = 0;$$079 = 0;
    while(1) {
     $77 = ($$0>>>0)<($2>>>0);
     if (!($77)) {
      break;
     }
     $88 = HEAP8[$13>>0]|0;
     $89 = ($88<<24>>24)<(0);
     $90 = HEAP32[$7>>2]|0;
     $91 = $89 ? $90 : $7;
     $92 = (($91) + ($$078)|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = ($93<<24>>24)!=(0);
     $95 = $93 << 24 >> 24;
     $96 = ($$079|0)==($95|0);
     $or$cond = $94 & $96;
     if ($or$cond) {
      $97 = HEAP32[$5>>2]|0;
      $98 = ((($97)) + 1|0);
      HEAP32[$5>>2] = $98;
      HEAP8[$97>>0] = $74;
      $99 = HEAP8[$13>>0]|0;
      $100 = ($99<<24>>24)<(0);
      $101 = HEAP32[$16>>2]|0;
      $102 = $99&255;
      $103 = $100 ? $101 : $102;
      $104 = (($103) + -1)|0;
      $105 = ($$078>>>0)<($104>>>0);
      $106 = $105&1;
      $$$078 = (($106) + ($$078))|0;
      $$1 = $$$078;$$180 = 0;
     } else {
      $$1 = $$078;$$180 = $$079;
     }
     $107 = HEAP8[$$0>>0]|0;
     $108 = HEAP32[$8>>2]|0;
     $109 = ((($108)) + 28|0);
     $110 = HEAP32[$109>>2]|0;
     __THREW__ = 0;
     $111 = (invoke_iii($110|0,($8|0),($107|0))|0);
     $112 = __THREW__; __THREW__ = 0;
     $113 = $112&1;
     if ($113) {
      label = 28;
      break L1;
     }
     $114 = HEAP32[$5>>2]|0;
     $115 = ((($114)) + 1|0);
     HEAP32[$5>>2] = $115;
     HEAP8[$114>>0] = $111;
     $116 = (($$180) + 1)|0;
     $117 = ((($$0)) + 1|0);
     $$0 = $117;$$078 = $$1;$$079 = $116;
    }
    $78 = $$184;
    $79 = (($78) - ($22))|0;
    $80 = (($3) + ($79)|0);
    $81 = HEAP32[$5>>2]|0;
    $82 = ($80|0)==($81|0);
    if ($82) {
     $122 = $80;
     label = 27;
    } else {
     $$0$i$i94 = $81;$$07$i$i93 = $80;
     while(1) {
      $83 = ((($$0$i$i94)) + -1|0);
      $84 = ($$07$i$i93>>>0)<($83>>>0);
      if (!($84)) {
       break;
      }
      $85 = HEAP8[$$07$i$i93>>0]|0;
      $86 = HEAP8[$83>>0]|0;
      HEAP8[$$07$i$i93>>0] = $86;
      HEAP8[$83>>0] = $85;
      $87 = ((($$07$i$i93)) + 1|0);
      $$0$i$i94 = $83;$$07$i$i93 = $87;
     }
     $$pre = HEAP32[$5>>2]|0;
     $122 = $$pre;
     label = 27;
    }
   }
  }
 } while(0);
 if ((label|0) == 27) {
  $118 = ($1|0)==($2|0);
  $119 = $1;
  $120 = (($119) - ($22))|0;
  $121 = (($3) + ($120)|0);
  $storemerge = $118 ? $122 : $121;
  HEAP32[$4>>2] = $storemerge;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 28) {
  $123 = ___cxa_find_matching_catch_2()|0;
  $124 = tempRet0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
  ___resumeException($123|0);
  // unreachable;
 }
}
function __ZNSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwb($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$0$i$i = 0, $$011 = 0, $$byval_copy = 0, $$pre = 0, $$pre18 = 0, $$sroa$0$0 = 0, $$sroa$0$0$copyload = 0, $$sroa$013$0 = 0, $$sroa$013$0$ph = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 20|0;
 $5 = sp + 16|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = ((($2)) + 4|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $9 & 1;
 $11 = ($10|0)==(0);
 L1: do {
  if ($11) {
   $12 = HEAP32[$0>>2]|0;
   $13 = ((($12)) + 24|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = HEAP32[$1>>2]|0;
   HEAP32[$5>>2] = $15;
   $16 = $4&1;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$5>>2]|0;
   $17 = (FUNCTION_TABLE_iiiiii[$14 & 2047]($0,$$byval_copy,$2,$3,$16)|0);
   $$sroa$0$0 = $17;
  } else {
   __ZNKSt3__28ios_base6getlocEv($6,$2);
   __THREW__ = 0;
   $18 = (invoke_iii(734,($6|0),(165480|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   do {
    if ($20) {
     $41 = ___cxa_find_matching_catch_2()|0;
     $42 = tempRet0;
     __ZNSt3__26localeD2Ev($6);
     $$0 = $42;$$011 = $41;
    } else {
     __ZNSt3__26localeD2Ev($6);
     $21 = HEAP32[$18>>2]|0;
     if ($4) {
      $22 = ((($21)) + 24|0);
      $23 = HEAP32[$22>>2]|0;
      FUNCTION_TABLE_vii[$23 & 2047]($7,$18);
     } else {
      $24 = ((($21)) + 28|0);
      $25 = HEAP32[$24>>2]|0;
      FUNCTION_TABLE_vii[$25 & 2047]($7,$18);
     }
     $26 = ((($7)) + 8|0);
     $27 = ((($26)) + 3|0);
     $28 = HEAP8[$27>>0]|0;
     $29 = ($28<<24>>24)<(0);
     $30 = HEAP32[$7>>2]|0;
     $$sroa$013$0$ph = $29 ? $30 : $7;
     $31 = ((($7)) + 4|0);
     $$sroa$013$0 = $$sroa$013$0$ph;$33 = $28;$37 = $30;
     while(1) {
      $32 = ($33<<24>>24)<(0);
      $34 = HEAP32[$31>>2]|0;
      $35 = $33&255;
      $36 = $32 ? $37 : $7;
      $38 = $32 ? $34 : $35;
      $39 = (($36) + ($38<<2)|0);
      $40 = ($$sroa$013$0|0)==($39|0);
      if ($40) {
       label = 9;
       break;
      }
      $43 = HEAP32[$$sroa$013$0>>2]|0;
      $44 = HEAP32[$1>>2]|0;
      $45 = ($44|0)==(0|0);
      if (!($45)) {
       $46 = ((($44)) + 24|0);
       $47 = HEAP32[$46>>2]|0;
       $48 = ((($44)) + 28|0);
       $49 = HEAP32[$48>>2]|0;
       $50 = ($47|0)==($49|0);
       if ($50) {
        $51 = HEAP32[$44>>2]|0;
        $52 = ((($51)) + 52|0);
        $53 = HEAP32[$52>>2]|0;
        $54 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($43)|0);
        __THREW__ = 0;
        $55 = (invoke_iii($53|0,($44|0),($54|0))|0);
        $56 = __THREW__; __THREW__ = 0;
        $57 = $56&1;
        if ($57) {
         label = 18;
         break;
        } else {
         $$0$i$i = $55;
        }
       } else {
        $58 = ((($47)) + 4|0);
        HEAP32[$46>>2] = $58;
        HEAP32[$47>>2] = $43;
        $59 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($43)|0);
        $$0$i$i = $59;
       }
       $60 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $61 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i,$60)|0);
       if ($61) {
        HEAP32[$1>>2] = 0;
       }
      }
      $62 = ((($$sroa$013$0)) + 4|0);
      $$pre = HEAP8[$27>>0]|0;
      $$pre18 = HEAP32[$7>>2]|0;
      $$sroa$013$0 = $62;$33 = $$pre;$37 = $$pre18;
     }
     if ((label|0) == 9) {
      $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
      __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($7);
      $$sroa$0$0 = $$sroa$0$0$copyload;
      break L1;
     }
     else if ((label|0) == 18) {
      $63 = ___cxa_find_matching_catch_2()|0;
      $64 = tempRet0;
      __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($7);
      $$0 = $64;$$011 = $63;
      break;
     }
    }
   } while(0);
   ___resumeException($$011|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$sroa$0$0|0);
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwl($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $$byval_copy = sp + 112|0;
 $vararg_buffer = sp;
 $5 = sp + 130|0;
 $6 = sp + 116|0;
 $7 = sp + 20|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156075>>0]|0;HEAP8[$5+1>>0]=HEAP8[156075+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156075+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156075+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156075+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156075+5>>0]|0;
 $12 = ((($5)) + 1|0);
 $13 = ((($2)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($12,160348,1,$14);
 $15 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $16 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,13,$15,$5,$vararg_buffer)|0);
 $17 = (($6) + ($16)|0);
 $18 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$17,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($10,$2);
 __THREW__ = 0;
 invoke_viiiiiii(982,($6|0),($18|0),($17|0),($7|0),($8|0),($9|0),($10|0));
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZNSt3__26localeD2Ev($10);
  ___resumeException($25|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($10);
  $21 = HEAP32[$1>>2]|0;
  HEAP32[$11>>2] = $21;
  $22 = HEAP32[$8>>2]|0;
  $23 = HEAP32[$9>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$11>>2]|0;
  $24 = (__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$22,$23,$2,$3)|0);
  STACKTOP = sp;return ($24|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwx($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $$byval_copy = sp + 204|0;
 $vararg_buffer = sp + 8|0;
 $6 = sp;
 $7 = sp + 208|0;
 $8 = sp + 32|0;
 $9 = sp + 28|0;
 $10 = sp + 24|0;
 $11 = sp + 20|0;
 $12 = sp + 16|0;
 $13 = $6;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($6)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($17,160353,1,$19);
 $20 = (__ZNSt3__26__clocEv()|0);
 $21 = $vararg_buffer;
 $22 = $21;
 HEAP32[$22>>2] = $4;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = $5;
 $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($7,23,$20,$6,$vararg_buffer)|0);
 $26 = (($7) + ($25)|0);
 $27 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($7,$26,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($11,$2);
 __THREW__ = 0;
 invoke_viiiiiii(982,($7|0),($27|0),($26|0),($8|0),($9|0),($10|0),($11|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  __ZNSt3__26localeD2Ev($11);
  ___resumeException($34|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($11);
  $30 = HEAP32[$1>>2]|0;
  HEAP32[$12>>2] = $30;
  $31 = HEAP32[$9>>2]|0;
  $32 = HEAP32[$10>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
  $33 = (__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$8,$31,$32,$2,$3)|0);
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwm($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $$byval_copy = sp + 104|0;
 $vararg_buffer = sp;
 $5 = sp + 120|0;
 $6 = sp + 108|0;
 $7 = sp + 20|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156075>>0]|0;HEAP8[$5+1>>0]=HEAP8[156075+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156075+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156075+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156075+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156075+5>>0]|0;
 $12 = ((($5)) + 1|0);
 $13 = ((($2)) + 4|0);
 $14 = HEAP32[$13>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($12,160348,0,$14);
 $15 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $16 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,12,$15,$5,$vararg_buffer)|0);
 $17 = (($6) + ($16)|0);
 $18 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$17,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($10,$2);
 __THREW__ = 0;
 invoke_viiiiiii(982,($6|0),($18|0),($17|0),($7|0),($8|0),($9|0),($10|0));
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 if ($20) {
  $25 = ___cxa_find_matching_catch_2()|0;
  $26 = tempRet0;
  __ZNSt3__26localeD2Ev($10);
  ___resumeException($25|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($10);
  $21 = HEAP32[$1>>2]|0;
  HEAP32[$11>>2] = $21;
  $22 = HEAP32[$8>>2]|0;
  $23 = HEAP32[$9>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$11>>2]|0;
  $24 = (__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$22,$23,$2,$3)|0);
  STACKTOP = sp;return ($24|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwy($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $$byval_copy = sp + 204|0;
 $vararg_buffer = sp + 8|0;
 $6 = sp;
 $7 = sp + 208|0;
 $8 = sp + 32|0;
 $9 = sp + 28|0;
 $10 = sp + 24|0;
 $11 = sp + 20|0;
 $12 = sp + 16|0;
 $13 = $6;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($6)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 __ZNSt3__214__num_put_base12__format_intEPcPKcbj($17,160353,0,$19);
 $20 = (__ZNSt3__26__clocEv()|0);
 $21 = $vararg_buffer;
 $22 = $21;
 HEAP32[$22>>2] = $4;
 $23 = (($21) + 4)|0;
 $24 = $23;
 HEAP32[$24>>2] = $5;
 $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($7,23,$20,$6,$vararg_buffer)|0);
 $26 = (($7) + ($25)|0);
 $27 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($7,$26,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($11,$2);
 __THREW__ = 0;
 invoke_viiiiiii(982,($7|0),($27|0),($26|0),($8|0),($9|0),($10|0),($11|0));
 $28 = __THREW__; __THREW__ = 0;
 $29 = $28&1;
 if ($29) {
  $34 = ___cxa_find_matching_catch_2()|0;
  $35 = tempRet0;
  __ZNSt3__26localeD2Ev($11);
  ___resumeException($34|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($11);
  $30 = HEAP32[$1>>2]|0;
  HEAP32[$12>>2] = $30;
  $31 = HEAP32[$9>>2]|0;
  $32 = HEAP32[$10>>2]|0;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
  $33 = (__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$8,$31,$32,$2,$3)|0);
  STACKTOP = sp;return ($33|0);
 }
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwd($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = +$4;
 var $$0 = 0, $$046 = 0, $$047 = 0, $$050 = 0, $$1 = 0, $$148 = 0, $$151 = 0, $$2 = 0, $$24980 = 0, $$25279 = 0, $$45482 = 0, $$483 = 0, $$byval_copy = 0, $$pre = 0, $$sroa$058$0 = 0, $$sroa$069$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr8 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $$byval_copy = sp + 304|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $5 = sp;
 $6 = sp + 308|0;
 $7 = sp + 300|0;
 $8 = sp + 72|0;
 $9 = sp + 68|0;
 $10 = sp + 64|0;
 $11 = sp + 60|0;
 $12 = sp + 56|0;
 $13 = $5;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($5)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (__ZNSt3__214__num_put_base14__format_floatEPcPKcj($17,169457,$19)|0);
 HEAP32[$7>>2] = $6;
 $21 = (__ZNSt3__26__clocEv()|0);
 if ($20) {
  $22 = ((($2)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  HEAP32[$vararg_buffer>>2] = $23;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $4;
  $24 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer)|0);
  $$0 = $24;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $4;
  $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer2)|0);
  $$0 = $25;
 }
 $26 = ($$0|0)>(29);
 L5: do {
  if ($26) {
   if ($20) {
    __THREW__ = 0;
    $27 = (invoke_i(955)|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     $30 = ((($2)) + 8|0);
     $31 = HEAP32[$30>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $31;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $4;
     $32 = (invoke_iiiii(976,($7|0),($27|0),($5|0),($vararg_buffer5|0))|0);
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if (!($34)) {
      $$1 = $32;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $37 = (invoke_i(955)|0);
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $4;
     $40 = (invoke_iiiii(976,($7|0),($37|0),($5|0),($vararg_buffer9|0))|0);
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if (!($42)) {
      $$1 = $40;
      label = 12;
     }
    }
   }
   do {
    if ((label|0) == 12) {
     $43 = HEAP32[$7>>2]|0;
     $44 = ($43|0)==(0|0);
     if ($44) {
      __THREW__ = 0;
      invoke_v(732);
      $45 = __THREW__; __THREW__ = 0;
      break;
     } else {
      $46 = $43;
      $$2 = $$1;$$sroa$069$0 = $46;$48 = $43;
      label = 15;
      break L5;
     }
    }
   } while(0);
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$45482 = $36;$$483 = $35;
  } else {
   $$pre = HEAP32[$7>>2]|0;
   $$2 = $$0;$$sroa$069$0 = 0;$48 = $$pre;
   label = 15;
  }
 } while(0);
 if ((label|0) == 15) {
  $47 = (($48) + ($$2)|0);
  $49 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($48,$47,$2)|0);
  $50 = ($48|0)==($6|0);
  do {
   if ($50) {
    $$046 = $8;$$sroa$058$0 = 0;
    label = 19;
   } else {
    $51 = $$2 << 3;
    $52 = (_malloc($51)|0);
    $53 = ($52|0)==(0|0);
    if ($53) {
     __THREW__ = 0;
     invoke_v(732);
     $54 = __THREW__; __THREW__ = 0;
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $$24980 = $55;$$25279 = $56;
     break;
    } else {
     $57 = $52;
     $$046 = $52;$$sroa$058$0 = $57;
     label = 19;
     break;
    }
   }
  } while(0);
  if ((label|0) == 19) {
   __THREW__ = 0;
   invoke_vii(733,($11|0),($2|0));
   $58 = __THREW__; __THREW__ = 0;
   $59 = $58&1;
   do {
    if ($59) {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $$047 = $72;$$050 = $73;
     label = 29;
    } else {
     __THREW__ = 0;
     invoke_viiiiiii(983,($48|0),($49|0),($47|0),($$046|0),($9|0),($10|0),($11|0));
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      $74 = ___cxa_find_matching_catch_2()|0;
      $75 = tempRet0;
      __ZNSt3__26localeD2Ev($11);
      $$047 = $74;$$050 = $75;
      label = 29;
      break;
     }
     __ZNSt3__26localeD2Ev($11);
     $62 = HEAP32[$1>>2]|0;
     HEAP32[$12>>2] = $62;
     $63 = HEAP32[$9>>2]|0;
     $64 = HEAP32[$10>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
     $65 = (invoke_iiiiiii(735,($$byval_copy|0),($$046|0),($63|0),($64|0),($2|0),($3|0))|0);
     $66 = __THREW__; __THREW__ = 0;
     $67 = $66&1;
     if ($67) {
      $76 = ___cxa_find_matching_catch_2()|0;
      $77 = tempRet0;
      $$148 = $76;$$151 = $77;
      break;
     }
     HEAP32[$1>>2] = $65;
     $68 = ($$sroa$058$0|0)==(0);
     if (!($68)) {
      $69 = $$sroa$058$0;
      _free($69);
     }
     $70 = ($$sroa$069$0|0)==(0);
     if (!($70)) {
      $71 = $$sroa$069$0;
      _free($71);
     }
     STACKTOP = sp;return ($65|0);
    }
   } while(0);
   if ((label|0) == 29) {
    $$148 = $$047;$$151 = $$050;
   }
   $78 = ($$sroa$058$0|0)==(0);
   if ($78) {
    $$24980 = $$148;$$25279 = $$151;
   } else {
    $79 = $$sroa$058$0;
    _free($79);
    $$24980 = $$148;$$25279 = $$151;
   }
  }
  $80 = ($$sroa$069$0|0)==(0);
  if ($80) {
   $$45482 = $$25279;$$483 = $$24980;
  } else {
   $81 = $$sroa$069$0;
   _free($81);
   $$45482 = $$25279;$$483 = $$24980;
  }
 }
 ___resumeException($$483|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwe($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = +$4;
 var $$0 = 0, $$046 = 0, $$047 = 0, $$050 = 0, $$1 = 0, $$148 = 0, $$151 = 0, $$2 = 0, $$24980 = 0, $$25279 = 0, $$45482 = 0, $$483 = 0, $$byval_copy = 0, $$pre = 0, $$sroa$058$0 = 0, $$sroa$069$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer2 = 0, $vararg_buffer5 = 0, $vararg_buffer9 = 0, $vararg_ptr1 = 0, $vararg_ptr8 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 352|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(352|0);
 $$byval_copy = sp + 304|0;
 $vararg_buffer9 = sp + 48|0;
 $vararg_buffer5 = sp + 32|0;
 $vararg_buffer2 = sp + 24|0;
 $vararg_buffer = sp + 8|0;
 $5 = sp;
 $6 = sp + 308|0;
 $7 = sp + 300|0;
 $8 = sp + 72|0;
 $9 = sp + 68|0;
 $10 = sp + 64|0;
 $11 = sp + 60|0;
 $12 = sp + 56|0;
 $13 = $5;
 $14 = $13;
 HEAP32[$14>>2] = 37;
 $15 = (($13) + 4)|0;
 $16 = $15;
 HEAP32[$16>>2] = 0;
 $17 = ((($5)) + 1|0);
 $18 = ((($2)) + 4|0);
 $19 = HEAP32[$18>>2]|0;
 $20 = (__ZNSt3__214__num_put_base14__format_floatEPcPKcj($17,156073,$19)|0);
 HEAP32[$7>>2] = $6;
 $21 = (__ZNSt3__26__clocEv()|0);
 if ($20) {
  $22 = ((($2)) + 8|0);
  $23 = HEAP32[$22>>2]|0;
  HEAP32[$vararg_buffer>>2] = $23;
  $vararg_ptr1 = ((($vararg_buffer)) + 8|0);
  HEAPF64[$vararg_ptr1>>3] = $4;
  $24 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer)|0);
  $$0 = $24;
 } else {
  HEAPF64[$vararg_buffer2>>3] = $4;
  $25 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,30,$21,$5,$vararg_buffer2)|0);
  $$0 = $25;
 }
 $26 = ($$0|0)>(29);
 L5: do {
  if ($26) {
   if ($20) {
    __THREW__ = 0;
    $27 = (invoke_i(955)|0);
    $28 = __THREW__; __THREW__ = 0;
    $29 = $28&1;
    if (!($29)) {
     $30 = ((($2)) + 8|0);
     $31 = HEAP32[$30>>2]|0;
     __THREW__ = 0;
     HEAP32[$vararg_buffer5>>2] = $31;
     $vararg_ptr8 = ((($vararg_buffer5)) + 8|0);
     HEAPF64[$vararg_ptr8>>3] = $4;
     $32 = (invoke_iiiii(976,($7|0),($27|0),($5|0),($vararg_buffer5|0))|0);
     $33 = __THREW__; __THREW__ = 0;
     $34 = $33&1;
     if (!($34)) {
      $$1 = $32;
      label = 12;
     }
    }
   } else {
    __THREW__ = 0;
    $37 = (invoke_i(955)|0);
    $38 = __THREW__; __THREW__ = 0;
    $39 = $38&1;
    if (!($39)) {
     __THREW__ = 0;
     HEAPF64[$vararg_buffer9>>3] = $4;
     $40 = (invoke_iiiii(976,($7|0),($37|0),($5|0),($vararg_buffer9|0))|0);
     $41 = __THREW__; __THREW__ = 0;
     $42 = $41&1;
     if (!($42)) {
      $$1 = $40;
      label = 12;
     }
    }
   }
   do {
    if ((label|0) == 12) {
     $43 = HEAP32[$7>>2]|0;
     $44 = ($43|0)==(0|0);
     if ($44) {
      __THREW__ = 0;
      invoke_v(732);
      $45 = __THREW__; __THREW__ = 0;
      break;
     } else {
      $46 = $43;
      $$2 = $$1;$$sroa$069$0 = $46;$48 = $43;
      label = 15;
      break L5;
     }
    }
   } while(0);
   $35 = ___cxa_find_matching_catch_2()|0;
   $36 = tempRet0;
   $$45482 = $36;$$483 = $35;
  } else {
   $$pre = HEAP32[$7>>2]|0;
   $$2 = $$0;$$sroa$069$0 = 0;$48 = $$pre;
   label = 15;
  }
 } while(0);
 if ((label|0) == 15) {
  $47 = (($48) + ($$2)|0);
  $49 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($48,$47,$2)|0);
  $50 = ($48|0)==($6|0);
  do {
   if ($50) {
    $$046 = $8;$$sroa$058$0 = 0;
    label = 19;
   } else {
    $51 = $$2 << 3;
    $52 = (_malloc($51)|0);
    $53 = ($52|0)==(0|0);
    if ($53) {
     __THREW__ = 0;
     invoke_v(732);
     $54 = __THREW__; __THREW__ = 0;
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     $$24980 = $55;$$25279 = $56;
     break;
    } else {
     $57 = $52;
     $$046 = $52;$$sroa$058$0 = $57;
     label = 19;
     break;
    }
   }
  } while(0);
  if ((label|0) == 19) {
   __THREW__ = 0;
   invoke_vii(733,($11|0),($2|0));
   $58 = __THREW__; __THREW__ = 0;
   $59 = $58&1;
   do {
    if ($59) {
     $72 = ___cxa_find_matching_catch_2()|0;
     $73 = tempRet0;
     $$047 = $72;$$050 = $73;
     label = 29;
    } else {
     __THREW__ = 0;
     invoke_viiiiiii(983,($48|0),($49|0),($47|0),($$046|0),($9|0),($10|0),($11|0));
     $60 = __THREW__; __THREW__ = 0;
     $61 = $60&1;
     if ($61) {
      $74 = ___cxa_find_matching_catch_2()|0;
      $75 = tempRet0;
      __ZNSt3__26localeD2Ev($11);
      $$047 = $74;$$050 = $75;
      label = 29;
      break;
     }
     __ZNSt3__26localeD2Ev($11);
     $62 = HEAP32[$1>>2]|0;
     HEAP32[$12>>2] = $62;
     $63 = HEAP32[$9>>2]|0;
     $64 = HEAP32[$10>>2]|0;
     __THREW__ = 0;
     ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
     $65 = (invoke_iiiiiii(735,($$byval_copy|0),($$046|0),($63|0),($64|0),($2|0),($3|0))|0);
     $66 = __THREW__; __THREW__ = 0;
     $67 = $66&1;
     if ($67) {
      $76 = ___cxa_find_matching_catch_2()|0;
      $77 = tempRet0;
      $$148 = $76;$$151 = $77;
      break;
     }
     HEAP32[$1>>2] = $65;
     $68 = ($$sroa$058$0|0)==(0);
     if (!($68)) {
      $69 = $$sroa$058$0;
      _free($69);
     }
     $70 = ($$sroa$069$0|0)==(0);
     if (!($70)) {
      $71 = $$sroa$069$0;
      _free($71);
     }
     STACKTOP = sp;return ($65|0);
    }
   } while(0);
   if ((label|0) == 29) {
    $$148 = $$047;$$151 = $$050;
   }
   $78 = ($$sroa$058$0|0)==(0);
   if ($78) {
    $$24980 = $$148;$$25279 = $$151;
   } else {
    $79 = $$sroa$058$0;
    _free($79);
    $$24980 = $$148;$$25279 = $$151;
   }
  }
  $80 = ($$sroa$069$0|0)==(0);
  if ($80) {
   $$45482 = $$25279;$$483 = $$24980;
  } else {
   $81 = $$sroa$069$0;
   _free($81);
   $$45482 = $$25279;$$483 = $$24980;
  }
 }
 ___resumeException($$483|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPKv($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(192|0);
 $$byval_copy = sp + 160|0;
 $vararg_buffer = sp;
 $5 = sp + 184|0;
 $6 = sp + 164|0;
 $7 = sp + 12|0;
 $8 = sp + 8|0;
 $9 = sp + 4|0;
 ;HEAP8[$5>>0]=HEAP8[156067>>0]|0;HEAP8[$5+1>>0]=HEAP8[156067+1>>0]|0;HEAP8[$5+2>>0]=HEAP8[156067+2>>0]|0;HEAP8[$5+3>>0]=HEAP8[156067+3>>0]|0;HEAP8[$5+4>>0]=HEAP8[156067+4>>0]|0;HEAP8[$5+5>>0]=HEAP8[156067+5>>0]|0;
 $10 = (__ZNSt3__26__clocEv()|0);
 HEAP32[$vararg_buffer>>2] = $4;
 $11 = (__ZNSt3__219__libcpp_snprintf_lEPcjP15__locale_structPKcz($6,20,$10,$5,$vararg_buffer)|0);
 $12 = (($6) + ($11)|0);
 $13 = (__ZNSt3__214__num_put_base18__identify_paddingEPcS1_RKNS_8ios_baseE($6,$12,$2)|0);
 __ZNKSt3__28ios_base6getlocEv($8,$2);
 __THREW__ = 0;
 $14 = (invoke_iii(734,($8|0),(165472|0))|0);
 $15 = __THREW__; __THREW__ = 0;
 $16 = $15&1;
 if ($16) {
  $28 = ___cxa_find_matching_catch_2()|0;
  $29 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($28|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($8);
  $17 = HEAP32[$14>>2]|0;
  $18 = ((($17)) + 48|0);
  $19 = HEAP32[$18>>2]|0;
  (FUNCTION_TABLE_iiiii[$19 & 2047]($14,$6,$12,$7)|0);
  $20 = (($7) + ($11<<2)|0);
  $21 = ($13|0)==($12|0);
  $22 = $6;
  $23 = $13;
  $24 = (($23) - ($22))|0;
  $25 = (($7) + ($24<<2)|0);
  $$0 = $21 ? $20 : $25;
  $26 = HEAP32[$1>>2]|0;
  HEAP32[$9>>2] = $26;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
  $27 = (__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_($$byval_copy,$7,$$0,$20,$2,$3)|0);
  STACKTOP = sp;return ($27|0);
 }
 return (0)|0;
}
function __ZNSt3__29__num_putIwE23__widen_and_group_floatEPcS2_S2_PwRS3_S4_RKNS_6localeE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$0111 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i131 = 0, $$0111 = 0, $$0112 = 0, $$0114 = 0, $$0116 = 0, $$07$i$i = 0, $$07$i$i130 = 0, $$1 = 0, $$1113 = 0, $$1115 = 0, $$1117 = 0, $$2 = 0, $$2118 = 0, $$3 = 0, $$pre$phiZ2D = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $7 = sp;
 $8 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165472)|0);
 $9 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165480)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 FUNCTION_TABLE_vii[$12 & 2047]($7,$9);
 HEAP32[$5>>2] = $3;
 $13 = HEAP8[$0>>0]|0;
 switch ($13<<24>>24) {
 case 43: case 45:  {
  $14 = HEAP32[$8>>2]|0;
  $15 = ((($14)) + 44|0);
  $16 = HEAP32[$15>>2]|0;
  __THREW__ = 0;
  $17 = (invoke_iii($16|0,($8|0),($13|0))|0);
  $18 = __THREW__; __THREW__ = 0;
  $19 = $18&1;
  if (!($19)) {
   $20 = ((($0)) + 1|0);
   $21 = HEAP32[$5>>2]|0;
   $22 = ((($21)) + 4|0);
   HEAP32[$5>>2] = $22;
   HEAP32[$21>>2] = $17;
   $$0114 = $20;
   label = 4;
  }
  break;
 }
 default: {
  $$0114 = $0;
  label = 4;
 }
 }
 L4: do {
  if ((label|0) == 4) {
   $23 = $2;
   $24 = $$0114;
   $25 = (($23) - ($24))|0;
   $26 = ($25|0)>(1);
   L6: do {
    if ($26) {
     $27 = HEAP8[$$0114>>0]|0;
     $28 = ($27<<24>>24)==(48);
     if ($28) {
      $29 = ((($$0114)) + 1|0);
      $30 = HEAP8[$29>>0]|0;
      switch ($30<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$1117 = $$0114;
       label = 15;
       break L6;
      }
      }
      $31 = HEAP32[$8>>2]|0;
      $32 = ((($31)) + 44|0);
      $33 = HEAP32[$32>>2]|0;
      __THREW__ = 0;
      $34 = (invoke_iii($33|0,($8|0),48)|0);
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      if ($36) {
       break L4;
      }
      $37 = HEAP32[$5>>2]|0;
      $38 = ((($37)) + 4|0);
      HEAP32[$5>>2] = $38;
      HEAP32[$37>>2] = $34;
      $39 = ((($$0114)) + 2|0);
      $40 = HEAP8[$29>>0]|0;
      $41 = HEAP32[$8>>2]|0;
      $42 = ((($41)) + 44|0);
      $43 = HEAP32[$42>>2]|0;
      __THREW__ = 0;
      $44 = (invoke_iii($43|0,($8|0),($40|0))|0);
      $45 = __THREW__; __THREW__ = 0;
      $46 = $45&1;
      if ($46) {
       break L4;
      }
      $47 = HEAP32[$5>>2]|0;
      $48 = ((($47)) + 4|0);
      HEAP32[$5>>2] = $48;
      HEAP32[$47>>2] = $44;
      $$0116 = $39;
      while(1) {
       $49 = ($$0116>>>0)<($2>>>0);
       if (!($49)) {
        $$1115 = $39;$$2118 = $$0116;
        break L6;
       }
       $50 = HEAP8[$$0116>>0]|0;
       __THREW__ = 0;
       $51 = (invoke_i(955)|0);
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        break L4;
       }
       $54 = $50 << 24 >> 24;
       __THREW__ = 0;
       $55 = (invoke_iii(980,($54|0),($51|0))|0);
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        break L4;
       }
       $58 = ($55|0)==(0);
       if ($58) {
        $$1115 = $39;$$2118 = $$0116;
        break L6;
       }
       $59 = ((($$0116)) + 1|0);
       $$0116 = $59;
      }
     } else {
      $$1117 = $$0114;
      label = 15;
     }
    } else {
     $$1117 = $$0114;
     label = 15;
    }
   } while(0);
   L18: do {
    if ((label|0) == 15) {
     while(1) {
      label = 0;
      $60 = ($$1117>>>0)<($2>>>0);
      if (!($60)) {
       $$1115 = $$0114;$$2118 = $$1117;
       break L18;
      }
      $61 = HEAP8[$$1117>>0]|0;
      __THREW__ = 0;
      $62 = (invoke_i(955)|0);
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       break L4;
      }
      $65 = $61 << 24 >> 24;
      __THREW__ = 0;
      $66 = (invoke_iii(981,($65|0),($62|0))|0);
      $67 = __THREW__; __THREW__ = 0;
      $68 = $67&1;
      if ($68) {
       break L4;
      }
      $69 = ($66|0)==(0);
      if ($69) {
       $$1115 = $$0114;$$2118 = $$1117;
       break L18;
      }
      $70 = ((($$1117)) + 1|0);
      $$1117 = $70;
      label = 15;
     }
    }
   } while(0);
   $71 = ((($7)) + 11|0);
   $72 = HEAP8[$71>>0]|0;
   $73 = ($72<<24>>24)<(0);
   $74 = ((($7)) + 4|0);
   $75 = HEAP32[$74>>2]|0;
   $76 = $72&255;
   $77 = $73 ? $75 : $76;
   $78 = ($77|0)==(0);
   $79 = $$1115;
   $80 = $0;
   L26: do {
    if ($78) {
     $81 = HEAP32[$5>>2]|0;
     $82 = HEAP32[$8>>2]|0;
     $83 = ((($82)) + 48|0);
     $84 = HEAP32[$83>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($84|0,($8|0),($$1115|0),($$2118|0),($81|0))|0);
     $85 = __THREW__; __THREW__ = 0;
     $86 = $85&1;
     if ($86) {
      break L4;
     }
     $87 = $$2118;
     $88 = (($87) - ($79))|0;
     $89 = HEAP32[$5>>2]|0;
     $90 = (($89) + ($88<<2)|0);
     HEAP32[$5>>2] = $90;
     $$pre$phiZ2D = $8;$177 = $90;
    } else {
     $91 = ($$1115|0)==($$2118|0);
     L30: do {
      if (!($91)) {
       $$0$i$i = $$2118;$$07$i$i = $$1115;
       while(1) {
        $92 = ((($$0$i$i)) + -1|0);
        $93 = ($$07$i$i>>>0)<($92>>>0);
        if (!($93)) {
         break L30;
        }
        $94 = HEAP8[$$07$i$i>>0]|0;
        $95 = HEAP8[$92>>0]|0;
        HEAP8[$$07$i$i>>0] = $95;
        HEAP8[$92>>0] = $94;
        $96 = ((($$07$i$i)) + 1|0);
        $$0$i$i = $92;$$07$i$i = $96;
       }
      }
     } while(0);
     $97 = HEAP32[$9>>2]|0;
     $98 = ((($97)) + 16|0);
     $99 = HEAP32[$98>>2]|0;
     __THREW__ = 0;
     $100 = (invoke_ii($99|0,($9|0))|0);
     $101 = __THREW__; __THREW__ = 0;
     $102 = $101&1;
     if ($102) {
      break L4;
     }
     $$0 = $$1115;$$0111 = 0;$$0112 = 0;
     while(1) {
      $103 = ($$0>>>0)<($$2118>>>0);
      if (!($103)) {
       break;
      }
      $113 = HEAP8[$71>>0]|0;
      $114 = ($113<<24>>24)<(0);
      $115 = HEAP32[$7>>2]|0;
      $116 = $114 ? $115 : $7;
      $117 = (($116) + ($$0111)|0);
      $118 = HEAP8[$117>>0]|0;
      $119 = ($118<<24>>24)>(0);
      $120 = $118 << 24 >> 24;
      $121 = ($$0112|0)==($120|0);
      $or$cond = $119 & $121;
      if ($or$cond) {
       $122 = HEAP32[$5>>2]|0;
       $123 = ((($122)) + 4|0);
       HEAP32[$5>>2] = $123;
       HEAP32[$122>>2] = $100;
       $124 = HEAP32[$74>>2]|0;
       $125 = $113&255;
       $126 = $114 ? $124 : $125;
       $127 = (($126) + -1)|0;
       $128 = ($$0111>>>0)<($127>>>0);
       $129 = $128&1;
       $$$0111 = (($129) + ($$0111))|0;
       $$1 = $$$0111;$$1113 = 0;
      } else {
       $$1 = $$0111;$$1113 = $$0112;
      }
      $130 = HEAP8[$$0>>0]|0;
      $131 = HEAP32[$8>>2]|0;
      $132 = ((($131)) + 44|0);
      $133 = HEAP32[$132>>2]|0;
      __THREW__ = 0;
      $134 = (invoke_iii($133|0,($8|0),($130|0))|0);
      $135 = __THREW__; __THREW__ = 0;
      $136 = $135&1;
      if ($136) {
       break L4;
      }
      $137 = HEAP32[$5>>2]|0;
      $138 = ((($137)) + 4|0);
      HEAP32[$5>>2] = $138;
      HEAP32[$137>>2] = $134;
      $139 = (($$1113) + 1)|0;
      $140 = ((($$0)) + 1|0);
      $$0 = $140;$$0111 = $$1;$$0112 = $139;
     }
     $104 = (($79) - ($80))|0;
     $105 = (($3) + ($104<<2)|0);
     $106 = HEAP32[$5>>2]|0;
     $107 = ($105|0)==($106|0);
     if ($107) {
      $$pre$phiZ2D = $8;$177 = $105;
     } else {
      $$0$i$i131 = $106;$$07$i$i130 = $105;
      while(1) {
       $108 = ((($$0$i$i131)) + -4|0);
       $109 = ($$07$i$i130>>>0)<($108>>>0);
       if (!($109)) {
        $$pre$phiZ2D = $8;$177 = $106;
        break L26;
       }
       $110 = HEAP32[$$07$i$i130>>2]|0;
       $111 = HEAP32[$108>>2]|0;
       HEAP32[$$07$i$i130>>2] = $111;
       HEAP32[$108>>2] = $110;
       $112 = ((($$07$i$i130)) + 4|0);
       $$0$i$i131 = $108;$$07$i$i130 = $112;
      }
     }
    }
   } while(0);
   $$2 = $$2118;$178 = $177;
   while(1) {
    $141 = ($$2>>>0)<($2>>>0);
    if (!($141)) {
     $$3 = $$2;$164 = $178;
     break;
    }
    $142 = HEAP8[$$2>>0]|0;
    $143 = ($142<<24>>24)==(46);
    $144 = ((($$2)) + 1|0);
    if ($143) {
     label = 39;
     break;
    }
    $153 = HEAP32[$$pre$phiZ2D>>2]|0;
    $154 = ((($153)) + 44|0);
    $155 = HEAP32[$154>>2]|0;
    __THREW__ = 0;
    $156 = (invoke_iii($155|0,($8|0),($142|0))|0);
    $157 = __THREW__; __THREW__ = 0;
    $158 = $157&1;
    if ($158) {
     break L4;
    }
    $159 = HEAP32[$5>>2]|0;
    $160 = ((($159)) + 4|0);
    HEAP32[$5>>2] = $160;
    HEAP32[$159>>2] = $156;
    $$2 = $144;$178 = $160;
   }
   if ((label|0) == 39) {
    $145 = HEAP32[$9>>2]|0;
    $146 = ((($145)) + 12|0);
    $147 = HEAP32[$146>>2]|0;
    __THREW__ = 0;
    $148 = (invoke_ii($147|0,($9|0))|0);
    $149 = __THREW__; __THREW__ = 0;
    $150 = $149&1;
    if ($150) {
     break;
    }
    $151 = HEAP32[$5>>2]|0;
    $152 = ((($151)) + 4|0);
    HEAP32[$5>>2] = $152;
    HEAP32[$151>>2] = $148;
    $$3 = $144;$164 = $152;
   }
   $161 = HEAP32[$8>>2]|0;
   $162 = ((($161)) + 48|0);
   $163 = HEAP32[$162>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($163|0,($8|0),($$3|0),($2|0),($164|0))|0);
   $165 = __THREW__; __THREW__ = 0;
   $166 = $165&1;
   if (!($166)) {
    $167 = $$3;
    $168 = (($23) - ($167))|0;
    $169 = HEAP32[$5>>2]|0;
    $170 = (($169) + ($168<<2)|0);
    HEAP32[$5>>2] = $170;
    $171 = ($1|0)==($2|0);
    $172 = $1;
    $173 = (($172) - ($80))|0;
    $174 = (($3) + ($173<<2)|0);
    $storemerge = $171 ? $170 : $174;
    HEAP32[$4>>2] = $storemerge;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
    STACKTOP = sp;return;
   }
  }
 } while(0);
 $175 = ___cxa_find_matching_catch_2()|0;
 $176 = tempRet0;
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
 ___resumeException($175|0);
 // unreachable;
}
function __ZNSt3__29__num_putIwE21__widen_and_group_intEPcS2_S2_PwRS3_S4_RKNS_6localeE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$081 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i97 = 0, $$07$i$i = 0, $$07$i$i96 = 0, $$081 = 0, $$082 = 0, $$086 = 0, $$1 = 0, $$183 = 0, $$187 = 0, $$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $7 = sp;
 $8 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165472)|0);
 $9 = (__ZNKSt3__26locale9use_facetERNS0_2idE($6,165480)|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ((($10)) + 20|0);
 $12 = HEAP32[$11>>2]|0;
 FUNCTION_TABLE_vii[$12 & 2047]($7,$9);
 $13 = ((($7)) + 11|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = ($14<<24>>24)<(0);
 $16 = ((($7)) + 4|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = $14&255;
 $19 = $15 ? $17 : $18;
 $20 = ($19|0)==(0);
 $21 = $2;
 $22 = $0;
 L1: do {
  if ($20) {
   $23 = HEAP32[$8>>2]|0;
   $24 = ((($23)) + 48|0);
   $25 = HEAP32[$24>>2]|0;
   __THREW__ = 0;
   (invoke_iiiii($25|0,($8|0),($0|0),($2|0),($3|0))|0);
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    label = 28;
   } else {
    $28 = (($21) - ($22))|0;
    $29 = (($3) + ($28<<2)|0);
    HEAP32[$5>>2] = $29;
    $120 = $29;
    label = 27;
   }
  } else {
   HEAP32[$5>>2] = $3;
   $30 = HEAP8[$0>>0]|0;
   switch ($30<<24>>24) {
   case 43: case 45:  {
    $31 = HEAP32[$8>>2]|0;
    $32 = ((($31)) + 44|0);
    $33 = HEAP32[$32>>2]|0;
    __THREW__ = 0;
    $34 = (invoke_iii($33|0,($8|0),($30|0))|0);
    $35 = __THREW__; __THREW__ = 0;
    $36 = $35&1;
    if ($36) {
     label = 28;
     break L1;
    }
    $37 = ((($0)) + 1|0);
    $38 = HEAP32[$5>>2]|0;
    $39 = ((($38)) + 4|0);
    HEAP32[$5>>2] = $39;
    HEAP32[$38>>2] = $34;
    $$086 = $37;
    break;
   }
   default: {
    $$086 = $0;
   }
   }
   $40 = $$086;
   $41 = (($21) - ($40))|0;
   $42 = ($41|0)>(1);
   L9: do {
    if ($42) {
     $43 = HEAP8[$$086>>0]|0;
     $44 = ($43<<24>>24)==(48);
     if ($44) {
      $45 = ((($$086)) + 1|0);
      $46 = HEAP8[$45>>0]|0;
      switch ($46<<24>>24) {
      case 88: case 120:  {
       break;
      }
      default: {
       $$187 = $$086;
       break L9;
      }
      }
      $47 = HEAP32[$8>>2]|0;
      $48 = ((($47)) + 44|0);
      $49 = HEAP32[$48>>2]|0;
      __THREW__ = 0;
      $50 = (invoke_iii($49|0,($8|0),48)|0);
      $51 = __THREW__; __THREW__ = 0;
      $52 = $51&1;
      if ($52) {
       label = 28;
       break L1;
      }
      $53 = HEAP32[$5>>2]|0;
      $54 = ((($53)) + 4|0);
      HEAP32[$5>>2] = $54;
      HEAP32[$53>>2] = $50;
      $55 = HEAP8[$45>>0]|0;
      $56 = HEAP32[$8>>2]|0;
      $57 = ((($56)) + 44|0);
      $58 = HEAP32[$57>>2]|0;
      __THREW__ = 0;
      $59 = (invoke_iii($58|0,($8|0),($55|0))|0);
      $60 = __THREW__; __THREW__ = 0;
      $61 = $60&1;
      if ($61) {
       label = 28;
       break L1;
      }
      $62 = ((($$086)) + 2|0);
      $63 = HEAP32[$5>>2]|0;
      $64 = ((($63)) + 4|0);
      HEAP32[$5>>2] = $64;
      HEAP32[$63>>2] = $59;
      $$187 = $62;
     } else {
      $$187 = $$086;
     }
    } else {
     $$187 = $$086;
    }
   } while(0);
   $65 = ($$187|0)==($2|0);
   L16: do {
    if (!($65)) {
     $$0$i$i = $2;$$07$i$i = $$187;
     while(1) {
      $66 = ((($$0$i$i)) + -1|0);
      $67 = ($$07$i$i>>>0)<($66>>>0);
      if (!($67)) {
       break L16;
      }
      $68 = HEAP8[$$07$i$i>>0]|0;
      $69 = HEAP8[$66>>0]|0;
      HEAP8[$$07$i$i>>0] = $69;
      HEAP8[$66>>0] = $68;
      $70 = ((($$07$i$i)) + 1|0);
      $$0$i$i = $66;$$07$i$i = $70;
     }
    }
   } while(0);
   $71 = HEAP32[$9>>2]|0;
   $72 = ((($71)) + 16|0);
   $73 = HEAP32[$72>>2]|0;
   __THREW__ = 0;
   $74 = (invoke_ii($73|0,($9|0))|0);
   $75 = __THREW__; __THREW__ = 0;
   $76 = $75&1;
   if ($76) {
    label = 28;
   } else {
    $$0 = $$187;$$081 = 0;$$082 = 0;
    while(1) {
     $77 = ($$0>>>0)<($2>>>0);
     if (!($77)) {
      break;
     }
     $88 = HEAP8[$13>>0]|0;
     $89 = ($88<<24>>24)<(0);
     $90 = HEAP32[$7>>2]|0;
     $91 = $89 ? $90 : $7;
     $92 = (($91) + ($$081)|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = ($93<<24>>24)!=(0);
     $95 = $93 << 24 >> 24;
     $96 = ($$082|0)==($95|0);
     $or$cond = $94 & $96;
     if ($or$cond) {
      $97 = HEAP32[$5>>2]|0;
      $98 = ((($97)) + 4|0);
      HEAP32[$5>>2] = $98;
      HEAP32[$97>>2] = $74;
      $99 = HEAP32[$16>>2]|0;
      $100 = $88&255;
      $101 = $89 ? $99 : $100;
      $102 = (($101) + -1)|0;
      $103 = ($$081>>>0)<($102>>>0);
      $104 = $103&1;
      $$$081 = (($104) + ($$081))|0;
      $$1 = $$$081;$$183 = 0;
     } else {
      $$1 = $$081;$$183 = $$082;
     }
     $105 = HEAP8[$$0>>0]|0;
     $106 = HEAP32[$8>>2]|0;
     $107 = ((($106)) + 44|0);
     $108 = HEAP32[$107>>2]|0;
     __THREW__ = 0;
     $109 = (invoke_iii($108|0,($8|0),($105|0))|0);
     $110 = __THREW__; __THREW__ = 0;
     $111 = $110&1;
     if ($111) {
      label = 28;
      break L1;
     }
     $112 = HEAP32[$5>>2]|0;
     $113 = ((($112)) + 4|0);
     HEAP32[$5>>2] = $113;
     HEAP32[$112>>2] = $109;
     $114 = (($$183) + 1)|0;
     $115 = ((($$0)) + 1|0);
     $$0 = $115;$$081 = $$1;$$082 = $114;
    }
    $78 = $$187;
    $79 = (($78) - ($22))|0;
    $80 = (($3) + ($79<<2)|0);
    $81 = HEAP32[$5>>2]|0;
    $82 = ($80|0)==($81|0);
    if ($82) {
     $120 = $80;
     label = 27;
    } else {
     $$0$i$i97 = $81;$$07$i$i96 = $80;
     while(1) {
      $83 = ((($$0$i$i97)) + -4|0);
      $84 = ($$07$i$i96>>>0)<($83>>>0);
      if (!($84)) {
       break;
      }
      $85 = HEAP32[$$07$i$i96>>2]|0;
      $86 = HEAP32[$83>>2]|0;
      HEAP32[$$07$i$i96>>2] = $86;
      HEAP32[$83>>2] = $85;
      $87 = ((($$07$i$i96)) + 4|0);
      $$0$i$i97 = $83;$$07$i$i96 = $87;
     }
     $$pre = HEAP32[$5>>2]|0;
     $120 = $$pre;
     label = 27;
    }
   }
  }
 } while(0);
 if ((label|0) == 27) {
  $116 = ($1|0)==($2|0);
  $117 = $1;
  $118 = (($117) - ($22))|0;
  $119 = (($3) + ($118<<2)|0);
  $storemerge = $116 ? $120 : $119;
  HEAP32[$4>>2] = $storemerge;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
  STACKTOP = sp;return;
 }
 else if ((label|0) == 28) {
  $121 = ___cxa_find_matching_catch_2()|0;
  $122 = tempRet0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($7);
  ___resumeException($121|0);
  // unreachable;
 }
}
function __ZNSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13do_date_orderEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy,$$byval_copy1,$3,$4,$5,156598,(156606))|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = ((($0)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (FUNCTION_TABLE_ii[$11 & 2047]($8)|0);
 $13 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $13;
 $14 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $14;
 $15 = ((($12)) + 11|0);
 $16 = HEAP8[$15>>0]|0;
 $17 = ($16<<24>>24)<(0);
 $18 = HEAP32[$12>>2]|0;
 $19 = $17 ? $18 : $12;
 $20 = ((($12)) + 4|0);
 $21 = HEAP32[$20>>2]|0;
 $22 = $16&255;
 $23 = $17 ? $21 : $22;
 $24 = (($19) + ($23)|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $25 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy,$$byval_copy1,$3,$4,$5,$19,$24)|0);
 STACKTOP = sp;return ($25|0);
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165440|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 24|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165440|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 16|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165440|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 20|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy10 = 0, $$byval_copy11 = 0, $$byval_copy12 = 0, $$byval_copy13 = 0, $$byval_copy14 = 0, $$byval_copy15 = 0, $$byval_copy16 = 0, $$byval_copy17 = 0, $$byval_copy18 = 0, $$byval_copy19 = 0, $$byval_copy2 = 0, $$byval_copy20 = 0, $$byval_copy21 = 0, $$byval_copy22 = 0, $$byval_copy23 = 0, $$byval_copy24 = 0, $$byval_copy25 = 0, $$byval_copy26 = 0;
 var $$byval_copy27 = 0, $$byval_copy28 = 0, $$byval_copy29 = 0, $$byval_copy3 = 0, $$byval_copy30 = 0, $$byval_copy4 = 0, $$byval_copy5 = 0, $$byval_copy6 = 0, $$byval_copy7 = 0, $$byval_copy8 = 0, $$byval_copy9 = 0, $$sroa$097$0 = 0, $$sroa$097$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0;
 var $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(256|0);
 $$byval_copy30 = sp + 248|0;
 $$byval_copy29 = sp + 244|0;
 $$byval_copy28 = sp + 240|0;
 $$byval_copy27 = sp + 236|0;
 $$byval_copy26 = sp + 232|0;
 $$byval_copy25 = sp + 228|0;
 $$byval_copy24 = sp + 224|0;
 $$byval_copy23 = sp + 220|0;
 $$byval_copy22 = sp + 216|0;
 $$byval_copy21 = sp + 212|0;
 $$byval_copy20 = sp + 208|0;
 $$byval_copy19 = sp + 204|0;
 $$byval_copy18 = sp + 200|0;
 $$byval_copy17 = sp + 196|0;
 $$byval_copy16 = sp + 192|0;
 $$byval_copy15 = sp + 188|0;
 $$byval_copy14 = sp + 184|0;
 $$byval_copy13 = sp + 180|0;
 $$byval_copy12 = sp + 176|0;
 $$byval_copy11 = sp + 172|0;
 $$byval_copy10 = sp + 168|0;
 $$byval_copy9 = sp + 164|0;
 $$byval_copy8 = sp + 160|0;
 $$byval_copy7 = sp + 156|0;
 $$byval_copy6 = sp + 152|0;
 $$byval_copy5 = sp + 148|0;
 $$byval_copy4 = sp + 144|0;
 $$byval_copy3 = sp + 140|0;
 $$byval_copy2 = sp + 136|0;
 $$byval_copy1 = sp + 132|0;
 $$byval_copy = sp + 128|0;
 $8 = sp + 124|0;
 $9 = sp + 120|0;
 $10 = sp + 116|0;
 $11 = sp + 112|0;
 $12 = sp + 108|0;
 $13 = sp + 104|0;
 $14 = sp + 100|0;
 $15 = sp + 96|0;
 $16 = sp + 92|0;
 $17 = sp + 88|0;
 $18 = sp + 84|0;
 $19 = sp + 80|0;
 $20 = sp + 76|0;
 $21 = sp + 72|0;
 $22 = sp + 68|0;
 $23 = sp + 64|0;
 $24 = sp + 60|0;
 $25 = sp + 56|0;
 $26 = sp + 52|0;
 $27 = sp + 48|0;
 $28 = sp + 44|0;
 $29 = sp + 40|0;
 $30 = sp + 36|0;
 $31 = sp + 32|0;
 $32 = sp + 28|0;
 $33 = sp + 24|0;
 $34 = sp + 20|0;
 $35 = sp + 16|0;
 $36 = sp + 12|0;
 $37 = sp + 8|0;
 $38 = sp + 4|0;
 $39 = sp;
 HEAP32[$4>>2] = 0;
 __ZNKSt3__28ios_base6getlocEv($8,$3);
 __THREW__ = 0;
 $40 = (invoke_iii(734,($8|0),(165440|0))|0);
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if ($42) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($49|0);
  // unreachable;
 }
 __ZNSt3__26localeD2Ev($8);
 $43 = $6 << 24 >> 24;
 $44 = ((($5)) + 16|0);
 $45 = ((($5)) + 8|0);
 $46 = ((($5)) + 20|0);
 $47 = ((($5)) + 24|0);
 $48 = ((($0)) + 8|0);
 do {
  switch ($43|0) {
  case 65: case 97:  {
   $51 = HEAP32[$2>>2]|0;
   HEAP32[$9>>2] = $51;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$47,$1,$$byval_copy,$4,$40);
   label = 28;
   break;
  }
  case 104: case 66: case 98:  {
   $52 = HEAP32[$2>>2]|0;
   HEAP32[$10>>2] = $52;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$10>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$44,$1,$$byval_copy1,$4,$40);
   label = 28;
   break;
  }
  case 99:  {
   $53 = HEAP32[$48>>2]|0;
   $54 = ((($53)) + 12|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (FUNCTION_TABLE_ii[$55 & 2047]($48)|0);
   $57 = HEAP32[$1>>2]|0;
   HEAP32[$11>>2] = $57;
   $58 = HEAP32[$2>>2]|0;
   HEAP32[$12>>2] = $58;
   $59 = ((($56)) + 11|0);
   $60 = HEAP8[$59>>0]|0;
   $61 = ($60<<24>>24)<(0);
   $62 = HEAP32[$56>>2]|0;
   $63 = $61 ? $62 : $56;
   $64 = ((($56)) + 4|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = $60&255;
   $67 = $61 ? $65 : $66;
   $68 = (($63) + ($67)|0);
   ;HEAP32[$$byval_copy2>>2]=HEAP32[$11>>2]|0;
   ;HEAP32[$$byval_copy3>>2]=HEAP32[$12>>2]|0;
   $69 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy2,$$byval_copy3,$3,$4,$5,$63,$68)|0);
   HEAP32[$1>>2] = $69;
   label = 28;
   break;
  }
  case 101: case 100:  {
   $70 = ((($5)) + 12|0);
   $71 = HEAP32[$2>>2]|0;
   HEAP32[$13>>2] = $71;
   ;HEAP32[$$byval_copy4>>2]=HEAP32[$13>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIcEE($0,$70,$1,$$byval_copy4,$4,$40);
   label = 28;
   break;
  }
  case 68:  {
   $72 = HEAP32[$1>>2]|0;
   HEAP32[$14>>2] = $72;
   $73 = HEAP32[$2>>2]|0;
   HEAP32[$15>>2] = $73;
   ;HEAP32[$$byval_copy5>>2]=HEAP32[$14>>2]|0;
   ;HEAP32[$$byval_copy6>>2]=HEAP32[$15>>2]|0;
   $74 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy5,$$byval_copy6,$3,$4,$5,156558,(156566))|0);
   HEAP32[$1>>2] = $74;
   label = 28;
   break;
  }
  case 70:  {
   $75 = HEAP32[$1>>2]|0;
   HEAP32[$16>>2] = $75;
   $76 = HEAP32[$2>>2]|0;
   HEAP32[$17>>2] = $76;
   ;HEAP32[$$byval_copy7>>2]=HEAP32[$16>>2]|0;
   ;HEAP32[$$byval_copy8>>2]=HEAP32[$17>>2]|0;
   $77 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy7,$$byval_copy8,$3,$4,$5,156566,(156574))|0);
   HEAP32[$1>>2] = $77;
   label = 28;
   break;
  }
  case 72:  {
   $78 = HEAP32[$2>>2]|0;
   HEAP32[$18>>2] = $78;
   ;HEAP32[$$byval_copy9>>2]=HEAP32[$18>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIcEE($0,$45,$1,$$byval_copy9,$4,$40);
   label = 28;
   break;
  }
  case 73:  {
   $79 = HEAP32[$2>>2]|0;
   HEAP32[$19>>2] = $79;
   ;HEAP32[$$byval_copy10>>2]=HEAP32[$19>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIcEE($0,$45,$1,$$byval_copy10,$4,$40);
   label = 28;
   break;
  }
  case 106:  {
   $80 = ((($5)) + 28|0);
   $81 = HEAP32[$2>>2]|0;
   HEAP32[$20>>2] = $81;
   ;HEAP32[$$byval_copy11>>2]=HEAP32[$20>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIcEE($0,$80,$1,$$byval_copy11,$4,$40);
   label = 28;
   break;
  }
  case 109:  {
   $82 = HEAP32[$2>>2]|0;
   HEAP32[$21>>2] = $82;
   ;HEAP32[$$byval_copy12>>2]=HEAP32[$21>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIcEE($0,$44,$1,$$byval_copy12,$4,$40);
   label = 28;
   break;
  }
  case 77:  {
   $83 = ((($5)) + 4|0);
   $84 = HEAP32[$2>>2]|0;
   HEAP32[$22>>2] = $84;
   ;HEAP32[$$byval_copy13>>2]=HEAP32[$22>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIcEE($0,$83,$1,$$byval_copy13,$4,$40);
   label = 28;
   break;
  }
  case 116: case 110:  {
   $85 = HEAP32[$2>>2]|0;
   HEAP32[$23>>2] = $85;
   ;HEAP32[$$byval_copy14>>2]=HEAP32[$23>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIcEE($0,$1,$$byval_copy14,$4,$40);
   label = 28;
   break;
  }
  case 112:  {
   $86 = HEAP32[$2>>2]|0;
   HEAP32[$24>>2] = $86;
   ;HEAP32[$$byval_copy15>>2]=HEAP32[$24>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIcEE($0,$45,$1,$$byval_copy15,$4,$40);
   label = 28;
   break;
  }
  case 114:  {
   $87 = HEAP32[$1>>2]|0;
   HEAP32[$25>>2] = $87;
   $88 = HEAP32[$2>>2]|0;
   HEAP32[$26>>2] = $88;
   ;HEAP32[$$byval_copy16>>2]=HEAP32[$25>>2]|0;
   ;HEAP32[$$byval_copy17>>2]=HEAP32[$26>>2]|0;
   $89 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy16,$$byval_copy17,$3,$4,$5,156574,(156585))|0);
   HEAP32[$1>>2] = $89;
   label = 28;
   break;
  }
  case 82:  {
   $90 = HEAP32[$1>>2]|0;
   HEAP32[$27>>2] = $90;
   $91 = HEAP32[$2>>2]|0;
   HEAP32[$28>>2] = $91;
   ;HEAP32[$$byval_copy18>>2]=HEAP32[$27>>2]|0;
   ;HEAP32[$$byval_copy19>>2]=HEAP32[$28>>2]|0;
   $92 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy18,$$byval_copy19,$3,$4,$5,156585,(156590))|0);
   HEAP32[$1>>2] = $92;
   label = 28;
   break;
  }
  case 83:  {
   $93 = HEAP32[$2>>2]|0;
   HEAP32[$29>>2] = $93;
   ;HEAP32[$$byval_copy20>>2]=HEAP32[$29>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIcEE($0,$5,$1,$$byval_copy20,$4,$40);
   label = 28;
   break;
  }
  case 84:  {
   $94 = HEAP32[$1>>2]|0;
   HEAP32[$30>>2] = $94;
   $95 = HEAP32[$2>>2]|0;
   HEAP32[$31>>2] = $95;
   ;HEAP32[$$byval_copy21>>2]=HEAP32[$30>>2]|0;
   ;HEAP32[$$byval_copy22>>2]=HEAP32[$31>>2]|0;
   $96 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy21,$$byval_copy22,$3,$4,$5,156590,(156598))|0);
   HEAP32[$1>>2] = $96;
   label = 28;
   break;
  }
  case 119:  {
   $97 = HEAP32[$2>>2]|0;
   HEAP32[$32>>2] = $97;
   ;HEAP32[$$byval_copy23>>2]=HEAP32[$32>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIcEE($0,$47,$1,$$byval_copy23,$4,$40);
   label = 28;
   break;
  }
  case 120:  {
   $98 = HEAP32[$0>>2]|0;
   $99 = ((($98)) + 20|0);
   $100 = HEAP32[$99>>2]|0;
   $101 = HEAP32[$1>>2]|0;
   HEAP32[$33>>2] = $101;
   $102 = HEAP32[$2>>2]|0;
   HEAP32[$34>>2] = $102;
   ;HEAP32[$$byval_copy24>>2]=HEAP32[$33>>2]|0;
   ;HEAP32[$$byval_copy25>>2]=HEAP32[$34>>2]|0;
   $103 = (FUNCTION_TABLE_iiiiiii[$100 & 1023]($0,$$byval_copy24,$$byval_copy25,$3,$4,$5)|0);
   $$sroa$097$0 = $103;
   break;
  }
  case 88:  {
   $104 = HEAP32[$48>>2]|0;
   $105 = ((($104)) + 24|0);
   $106 = HEAP32[$105>>2]|0;
   $107 = (FUNCTION_TABLE_ii[$106 & 2047]($48)|0);
   $108 = HEAP32[$1>>2]|0;
   HEAP32[$35>>2] = $108;
   $109 = HEAP32[$2>>2]|0;
   HEAP32[$36>>2] = $109;
   $110 = ((($107)) + 11|0);
   $111 = HEAP8[$110>>0]|0;
   $112 = ($111<<24>>24)<(0);
   $113 = HEAP32[$107>>2]|0;
   $114 = $112 ? $113 : $107;
   $115 = ((($107)) + 4|0);
   $116 = HEAP32[$115>>2]|0;
   $117 = $111&255;
   $118 = $112 ? $116 : $117;
   $119 = (($114) + ($118)|0);
   ;HEAP32[$$byval_copy26>>2]=HEAP32[$35>>2]|0;
   ;HEAP32[$$byval_copy27>>2]=HEAP32[$36>>2]|0;
   $120 = (__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$$byval_copy26,$$byval_copy27,$3,$4,$5,$114,$119)|0);
   HEAP32[$1>>2] = $120;
   label = 28;
   break;
  }
  case 121:  {
   $121 = HEAP32[$2>>2]|0;
   HEAP32[$37>>2] = $121;
   ;HEAP32[$$byval_copy28>>2]=HEAP32[$37>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($0,$46,$1,$$byval_copy28,$4,$40);
   label = 28;
   break;
  }
  case 89:  {
   $122 = HEAP32[$2>>2]|0;
   HEAP32[$38>>2] = $122;
   ;HEAP32[$$byval_copy29>>2]=HEAP32[$38>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIcEE($0,$46,$1,$$byval_copy29,$4,$40);
   label = 28;
   break;
  }
  case 37:  {
   $123 = HEAP32[$2>>2]|0;
   HEAP32[$39>>2] = $123;
   ;HEAP32[$$byval_copy30>>2]=HEAP32[$39>>2]|0;
   __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIcEE($0,$1,$$byval_copy30,$4,$40);
   label = 28;
   break;
  }
  default: {
   $124 = HEAP32[$4>>2]|0;
   $125 = $124 | 4;
   HEAP32[$4>>2] = $125;
   label = 28;
  }
  }
 } while(0);
 if ((label|0) == 28) {
  $$sroa$097$0$copyload = HEAP32[$1>>2]|0;
  $$sroa$097$0 = $$sroa$097$0$copyload;
 }
 STACKTOP = sp;return ($$sroa$097$0|0);
}
function __ZNKSt3__220__time_get_c_storageIcE7__weeksEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161072]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161072)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(984);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41578] = 166144;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41578]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIcE8__monthsEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161056]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161056)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(985);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41535] = 165852;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41535]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIcE7__am_pmEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161040]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161040)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(986);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41462] = 165560;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41462]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIcE3__cEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161032]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161032)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[165548>>2]=0|0;HEAP32[165548+4>>2]=0|0;HEAP32[165548+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc(156316)|0);
    $7 = ($6>>>0)>(4294967279);
    do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(165548|0));
      $8 = __THREW__; __THREW__ = 0;
     } else {
      $9 = ($6>>>0)<(11);
      if ($9) {
       $10 = $6&255;
       HEAP8[(165559)>>0] = $10;
       $$0$i$i = 165548;
      } else {
       $11 = (($6) + 16)|0;
       $12 = $11 & -16;
       __THREW__ = 0;
       $13 = (invoke_ii(988,($12|0))|0);
       $14 = __THREW__; __THREW__ = 0;
       $15 = $14&1;
       if ($15) {
        break;
       }
       HEAP32[41387] = $13;
       $16 = $12 | -2147483648;
       HEAP32[(165556)>>2] = $16;
       HEAP32[(165552)>>2] = $6;
       $$0$i$i = $13;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,156316,$6)|0);
      $17 = (($$0$i$i) + ($6)|0);
      HEAP8[$1>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($17,$1);
      (___cxa_atexit((989|0),(165548|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (165548|0);
}
function __ZNKSt3__220__time_get_c_storageIcE3__rEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161024]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161024)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[165536>>2]=0|0;HEAP32[165536+4>>2]=0|0;HEAP32[165536+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc(156304)|0);
    $7 = ($6>>>0)>(4294967279);
    do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(165536|0));
      $8 = __THREW__; __THREW__ = 0;
     } else {
      $9 = ($6>>>0)<(11);
      if ($9) {
       $10 = $6&255;
       HEAP8[(165547)>>0] = $10;
       $$0$i$i = 165536;
      } else {
       $11 = (($6) + 16)|0;
       $12 = $11 & -16;
       __THREW__ = 0;
       $13 = (invoke_ii(988,($12|0))|0);
       $14 = __THREW__; __THREW__ = 0;
       $15 = $14&1;
       if ($15) {
        break;
       }
       HEAP32[41384] = $13;
       $16 = $12 | -2147483648;
       HEAP32[(165544)>>2] = $16;
       HEAP32[(165540)>>2] = $6;
       $$0$i$i = $13;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,156304,$6)|0);
      $17 = (($$0$i$i) + ($6)|0);
      HEAP8[$1>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($17,$1);
      (___cxa_atexit((989|0),(165536|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (165536|0);
}
function __ZNKSt3__220__time_get_c_storageIcE3__xEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161016]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161016)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[165524>>2]=0|0;HEAP32[165524+4>>2]=0|0;HEAP32[165524+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc(156295)|0);
    $7 = ($6>>>0)>(4294967279);
    do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(165524|0));
      $8 = __THREW__; __THREW__ = 0;
     } else {
      $9 = ($6>>>0)<(11);
      if ($9) {
       $10 = $6&255;
       HEAP8[(165535)>>0] = $10;
       $$0$i$i = 165524;
      } else {
       $11 = (($6) + 16)|0;
       $12 = $11 & -16;
       __THREW__ = 0;
       $13 = (invoke_ii(988,($12|0))|0);
       $14 = __THREW__; __THREW__ = 0;
       $15 = $14&1;
       if ($15) {
        break;
       }
       HEAP32[41381] = $13;
       $16 = $12 | -2147483648;
       HEAP32[(165532)>>2] = $16;
       HEAP32[(165528)>>2] = $6;
       $$0$i$i = $13;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,156295,$6)|0);
      $17 = (($$0$i$i) + ($6)|0);
      HEAP8[$1>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($17,$1);
      (___cxa_atexit((989|0),(165524|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (165524|0);
}
function __ZNKSt3__220__time_get_c_storageIcE3__XEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161008]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161008)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[165512>>2]=0|0;HEAP32[165512+4>>2]=0|0;HEAP32[165512+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc(156286)|0);
    $7 = ($6>>>0)>(4294967279);
    do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(165512|0));
      $8 = __THREW__; __THREW__ = 0;
     } else {
      $9 = ($6>>>0)<(11);
      if ($9) {
       $10 = $6&255;
       HEAP8[(165523)>>0] = $10;
       $$0$i$i = 165512;
      } else {
       $11 = (($6) + 16)|0;
       $12 = $11 & -16;
       __THREW__ = 0;
       $13 = (invoke_ii(988,($12|0))|0);
       $14 = __THREW__; __THREW__ = 0;
       $15 = $14&1;
       if ($15) {
        break;
       }
       HEAP32[41378] = $13;
       $16 = $12 | -2147483648;
       HEAP32[(165520)>>2] = $16;
       HEAP32[(165516)>>2] = $6;
       $$0$i$i = $13;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,156286,$6)|0);
      $17 = (($$0$i$i) + ($6)|0);
      HEAP8[$1>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($17,$1);
      (___cxa_atexit((989|0),(165512|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    $18 = ___cxa_find_matching_catch_2()|0;
    $19 = tempRet0;
    ___resumeException($18|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (165512|0);
}
function __ZNSt3__2L10init_am_pmEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161048]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161048)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 165560;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((165848)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((990|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc(165560,156337)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165572),156340)|0);
 return;
}
function ___cxx_global_array_dtor_118($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (165848);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
  $3 = ($1|0)==(165560|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNSt3__2L11init_monthsEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161064]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161064)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 165852;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((166140)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((991|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc(165852,156343)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165864),156351)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165876),156360)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165888),156366)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165900),156372)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165912),156376)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165924),156381)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165936),156386)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165948),156393)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165960),156403)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165972),156411)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165984),156420)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((165996),156429)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166008),156433)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166020),156437)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166032),156441)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166044),156372)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166056),156445)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166068),156449)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166080),156453)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166092),156457)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166104),156461)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166116),156465)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166128),156469)|0);
 return;
}
function ___cxx_global_array_dtor_70($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (166140);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
  $3 = ($1|0)==(165852|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNSt3__2L10init_weeksEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161080]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161080)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 166144;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((166312)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((992|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc(166144,156473)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166156),156480)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166168),156487)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166180),156495)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166192),156505)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166204),156514)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166216),156521)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166228),156530)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166240),156534)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166252),156538)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166264),156542)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166276),156546)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166288),156550)|0);
 (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc((166300),156554)|0);
 return;
}
function ___cxx_global_array_dtor($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (166312);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
  $3 = ($1|0)==(166144|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_ii[$9 & 2047]($7)|0);
 $11 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $11;
 $12 = ((($10)) + 168|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $13 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$10,$12,$5,$4,0)|0);
 $14 = $13;
 $15 = $10;
 $16 = (($14) - ($15))|0;
 $17 = ($16|0)<(168);
 if ($17) {
  $18 = (($16|0) / 12)&-1;
  $19 = (($18|0) % 7)&-1;
  HEAP32[$1>>2] = $19;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 2047]($7)|0);
 $12 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $12;
 $13 = ((($11)) + 288|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $14 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$11,$13,$5,$4,0)|0);
 $15 = $14;
 $16 = $11;
 $17 = (($15) - ($16))|0;
 $18 = ($17|0)<(288);
 if ($18) {
  $19 = (($17|0) / 12)&-1;
  $20 = (($19|0) % 12)&-1;
  HEAP32[$1>>2] = $20;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKcSC_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i55 = 0, $$0$i$i$i$i71 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i61 = 0, $$0$i$i2$i$i77 = 0, $$0$i$i68 = 0, $$0$ph = 0, $$0$ph$be = 0, $$0$ph$sink = 0, $$051 = 0, $$052 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$lcssa = 0, $$ph = 0, $$pre = 0, $$pre$i$i$i57 = 0, $$pre$i$i$i73 = 0;
 var $$pre$i4$i$i63 = 0, $$pre$i4$i$i63100 = 0, $$pre$i4$i$i63101 = 0, $$pre$pre = 0, $$sroa$049$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0;
 var $phitmp$i5$i$i64 = 0, $sext = 0, $tmp$i$i66 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy1 = sp + 16|0;
 $$byval_copy = sp + 12|0;
 $8 = sp + 8|0;
 $9 = sp + 4|0;
 $10 = sp;
 __ZNKSt3__28ios_base6getlocEv($8,$3);
 __THREW__ = 0;
 $11 = (invoke_iii(734,($8|0),(165440|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $50 = ___cxa_find_matching_catch_2()|0;
  $51 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($50|0);
  // unreachable;
 }
 __ZNSt3__26localeD2Ev($8);
 HEAP32[$4>>2] = 0;
 $14 = ((($11)) + 8|0);
 $$0$ph = $6;$$pre = 0;
 L4: while(1) {
  $15 = ($$0$ph|0)!=($7|0);
  $17 = $$pre;
  while(1) {
   $16 = ($17|0)==(0);
   $or$cond = $15 & $16;
   $18 = HEAP32[$1>>2]|0;
   if (!($or$cond)) {
    $$pre$i$i$i73 = $18;
    break L4;
   }
   $19 = ($18|0)==(0|0);
   $20 = $18;
   if ($19) {
    $156 = 0;$218 = 1;$75 = $20;
   } else {
    $21 = ((($18)) + 12|0);
    $22 = HEAP32[$21>>2]|0;
    $23 = ((($18)) + 16|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($22|0)==($24|0);
    if ($25) {
     $26 = HEAP32[$18>>2]|0;
     $27 = ((($26)) + 36|0);
     $28 = HEAP32[$27>>2]|0;
     $29 = (FUNCTION_TABLE_ii[$28 & 2047]($18)|0);
     $$0$i$i$i$i = $29;
    } else {
     $30 = HEAP8[$22>>0]|0;
     $31 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($30)|0);
     $$0$i$i$i$i = $31;
    }
    $32 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $33 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$32)|0);
    if ($33) {
     HEAP32[$1>>2] = 0;
     $156 = 0;$218 = 1;$75 = 0;
    } else {
     $156 = $18;$218 = 0;$75 = $20;
    }
   }
   $34 = HEAP32[$2>>2]|0;
   $35 = ($34|0)==(0|0);
   $36 = $34;
   do {
    if ($35) {
     $$ph = $36;
     label = 17;
    } else {
     $37 = ((($34)) + 12|0);
     $38 = HEAP32[$37>>2]|0;
     $39 = ((($34)) + 16|0);
     $40 = HEAP32[$39>>2]|0;
     $41 = ($38|0)==($40|0);
     if ($41) {
      $42 = HEAP32[$34>>2]|0;
      $43 = ((($42)) + 36|0);
      $44 = HEAP32[$43>>2]|0;
      $45 = (FUNCTION_TABLE_ii[$44 & 2047]($34)|0);
      $$0$i$i2$i$i = $45;
     } else {
      $46 = HEAP8[$38>>0]|0;
      $47 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($46)|0);
      $$0$i$i2$i$i = $47;
     }
     $48 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
     $49 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$48)|0);
     if ($49) {
      HEAP32[$2>>2] = 0;
      $$ph = 0;
      label = 17;
      break;
     } else {
      if ($218) {
       $219 = $34;$76 = $36;
       break;
      } else {
       label = 63;
       break L4;
      }
     }
    }
   } while(0);
   if ((label|0) == 17) {
    label = 0;
    if ($218) {
     label = 63;
     break L4;
    } else {
     $219 = 0;$76 = $$ph;
    }
   }
   $52 = HEAP8[$$0$ph>>0]|0;
   $53 = HEAP32[$11>>2]|0;
   $54 = ((($53)) + 36|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (FUNCTION_TABLE_iiii[$55 & 2047]($11,$52,0)|0);
   $57 = ($56<<24>>24)==(37);
   if ($57) {
    label = 21;
    break;
   }
   $80 = HEAP8[$$0$ph>>0]|0;
   $81 = ($80<<24>>24)>(-1);
   if ($81) {
    $82 = $80 << 24 >> 24;
    $83 = HEAP32[$14>>2]|0;
    $84 = (($83) + ($82<<1)|0);
    $85 = HEAP16[$84>>1]|0;
    $86 = $85 & 8192;
    $87 = ($86<<16>>16)==(0);
    if (!($87)) {
     $$0$ph$sink = $$0$ph;
     label = 29;
     break;
    }
   }
   $155 = ((($156)) + 12|0);
   $157 = HEAP32[$155>>2]|0;
   $158 = ((($156)) + 16|0);
   $159 = HEAP32[$158>>2]|0;
   $160 = ($157|0)==($159|0);
   if ($160) {
    $161 = HEAP32[$156>>2]|0;
    $162 = ((($161)) + 36|0);
    $163 = HEAP32[$162>>2]|0;
    $164 = (FUNCTION_TABLE_ii[$163 & 2047]($156)|0);
    $$0$i$i68 = $164;
   } else {
    $165 = HEAP8[$157>>0]|0;
    $166 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($165)|0);
    $$0$i$i68 = $166;
   }
   $167 = $$0$i$i68&255;
   $168 = HEAP32[$11>>2]|0;
   $169 = ((($168)) + 12|0);
   $170 = HEAP32[$169>>2]|0;
   $171 = (FUNCTION_TABLE_iii[$170 & 2047]($11,$167)|0);
   $172 = HEAP8[$$0$ph>>0]|0;
   $173 = HEAP32[$11>>2]|0;
   $174 = ((($173)) + 12|0);
   $175 = HEAP32[$174>>2]|0;
   $176 = (FUNCTION_TABLE_iii[$175 & 2047]($11,$172)|0);
   $177 = ($171<<24>>24)==($176<<24>>24);
   if ($177) {
    label = 59;
    break;
   }
   HEAP32[$4>>2] = 4;
   $17 = 4;
  }
  L38: do {
   if ((label|0) == 21) {
    label = 0;
    $58 = ((($$0$ph)) + 1|0);
    $59 = ($58|0)==($7|0);
    if ($59) {
     label = 63;
     break L4;
    }
    $60 = HEAP8[$58>>0]|0;
    $61 = HEAP32[$11>>2]|0;
    $62 = ((($61)) + 36|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = (FUNCTION_TABLE_iiii[$63 & 2047]($11,$60,0)|0);
    switch ($64<<24>>24) {
    case 48: case 69:  {
     $65 = ((($$0$ph)) + 2|0);
     $66 = ($65|0)==($7|0);
     if ($66) {
      label = 63;
      break L4;
     }
     $67 = HEAP8[$65>>0]|0;
     $68 = HEAP32[$11>>2]|0;
     $69 = ((($68)) + 36|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = (FUNCTION_TABLE_iiii[$70 & 2047]($11,$67,0)|0);
     $$051 = $64;$$052 = $71;$79 = $58;
     break;
    }
    default: {
     $$051 = 0;$$052 = $64;$79 = $$0$ph;
    }
    }
    $72 = HEAP32[$0>>2]|0;
    $73 = ((($72)) + 36|0);
    $74 = HEAP32[$73>>2]|0;
    HEAP32[$9>>2] = $75;
    HEAP32[$10>>2] = $76;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
    ;HEAP32[$$byval_copy1>>2]=HEAP32[$10>>2]|0;
    $77 = (FUNCTION_TABLE_iiiiiiiii[$74 & 1023]($0,$$byval_copy,$$byval_copy1,$3,$4,$5,$$052,$$051)|0);
    HEAP32[$1>>2] = $77;
    $78 = ((($79)) + 2|0);
    $$0$ph$be = $78;
   }
   else if ((label|0) == 29) {
    while(1) {
     label = 0;
     $88 = ((($$0$ph$sink)) + 1|0);
     $89 = ($88|0)==($7|0);
     if ($89) {
      $$lcssa = $7;
      break;
     }
     $90 = HEAP8[$88>>0]|0;
     $91 = ($90<<24>>24)>(-1);
     if (!($91)) {
      $$lcssa = $88;
      break;
     }
     $92 = $90 << 24 >> 24;
     $93 = (($83) + ($92<<1)|0);
     $94 = HEAP16[$93>>1]|0;
     $95 = $94 & 8192;
     $96 = ($95<<16>>16)==(0);
     if ($96) {
      $$lcssa = $88;
      break;
     } else {
      $$0$ph$sink = $88;
      label = 29;
     }
    }
    $$pre$i$i$i57 = $156;$$pre$i4$i$i63 = $219;$112 = $219;
    while(1) {
     $97 = ($$pre$i$i$i57|0)==(0|0);
     if ($97) {
      $126 = 1;$128 = 0;
     } else {
      $98 = ((($$pre$i$i$i57)) + 12|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = ((($$pre$i$i$i57)) + 16|0);
      $101 = HEAP32[$100>>2]|0;
      $102 = ($99|0)==($101|0);
      if ($102) {
       $103 = HEAP32[$$pre$i$i$i57>>2]|0;
       $104 = ((($103)) + 36|0);
       $105 = HEAP32[$104>>2]|0;
       $106 = (FUNCTION_TABLE_ii[$105 & 2047]($$pre$i$i$i57)|0);
       $$0$i$i$i$i55 = $106;
      } else {
       $107 = HEAP8[$99>>0]|0;
       $108 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($107)|0);
       $$0$i$i$i$i55 = $108;
      }
      $109 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $110 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i55,$109)|0);
      if ($110) {
       HEAP32[$1>>2] = 0;
       $126 = 1;$128 = 0;
      } else {
       $126 = 0;$128 = $$pre$i$i$i57;
      }
     }
     $111 = ($112|0)==(0|0);
     do {
      if ($111) {
       $$pre$i4$i$i63101 = $$pre$i4$i$i63;
       label = 45;
      } else {
       $113 = ((($112)) + 12|0);
       $114 = HEAP32[$113>>2]|0;
       $115 = ((($112)) + 16|0);
       $116 = HEAP32[$115>>2]|0;
       $117 = ($114|0)==($116|0);
       if ($117) {
        $118 = HEAP32[$112>>2]|0;
        $119 = ((($118)) + 36|0);
        $120 = HEAP32[$119>>2]|0;
        $121 = (FUNCTION_TABLE_ii[$120 & 2047]($112)|0);
        $$0$i$i2$i$i61 = $121;
       } else {
        $122 = HEAP8[$114>>0]|0;
        $123 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($122)|0);
        $$0$i$i2$i$i61 = $123;
       }
       $124 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $125 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i61,$124)|0);
       if ($125) {
        HEAP32[$2>>2] = 0;
        $$pre$i4$i$i63101 = 0;
        label = 45;
        break;
       } else {
        $phitmp$i5$i$i64 = ($$pre$i4$i$i63|0)==(0|0);
        $tmp$i$i66 = $126 ^ $phitmp$i5$i$i64;
        if ($tmp$i$i66) {
         $$pre$i4$i$i63100 = $$pre$i4$i$i63;$220 = $$pre$i4$i$i63;
         break;
        } else {
         $$0$ph$be = $$lcssa;
         break L38;
        }
       }
      }
     } while(0);
     if ((label|0) == 45) {
      label = 0;
      if ($126) {
       $$0$ph$be = $$lcssa;
       break L38;
      } else {
       $$pre$i4$i$i63100 = $$pre$i4$i$i63101;$220 = 0;
      }
     }
     $127 = ((($128)) + 12|0);
     $129 = HEAP32[$127>>2]|0;
     $130 = ((($128)) + 16|0);
     $131 = HEAP32[$130>>2]|0;
     $132 = ($129|0)==($131|0);
     if ($132) {
      $133 = HEAP32[$128>>2]|0;
      $134 = ((($133)) + 36|0);
      $135 = HEAP32[$134>>2]|0;
      $136 = (FUNCTION_TABLE_ii[$135 & 2047]($128)|0);
      $$0$i$i = $136;
     } else {
      $137 = HEAP8[$129>>0]|0;
      $138 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($137)|0);
      $$0$i$i = $138;
     }
     $139 = $$0$i$i&255;
     $140 = ($139<<24>>24)>(-1);
     if (!($140)) {
      $$0$ph$be = $$lcssa;
      break L38;
     }
     $sext = $$0$i$i << 24;
     $141 = $sext >> 24;
     $142 = HEAP32[$14>>2]|0;
     $143 = (($142) + ($141<<1)|0);
     $144 = HEAP16[$143>>1]|0;
     $145 = $144 & 8192;
     $146 = ($145<<16>>16)==(0);
     if ($146) {
      $$0$ph$be = $$lcssa;
      break L38;
     }
     $147 = HEAP32[$127>>2]|0;
     $148 = HEAP32[$130>>2]|0;
     $149 = ($147|0)==($148|0);
     if ($149) {
      $150 = HEAP32[$128>>2]|0;
      $151 = ((($150)) + 40|0);
      $152 = HEAP32[$151>>2]|0;
      (FUNCTION_TABLE_ii[$152 & 2047]($128)|0);
      $$pre$i$i$i57 = $128;$$pre$i4$i$i63 = $$pre$i4$i$i63100;$112 = $220;
      continue;
     } else {
      $153 = ((($147)) + 1|0);
      HEAP32[$127>>2] = $153;
      $154 = HEAP8[$147>>0]|0;
      (__ZNSt3__211char_traitsIcE11to_int_typeEc($154)|0);
      $$pre$i$i$i57 = $128;$$pre$i4$i$i63 = $$pre$i4$i$i63100;$112 = $220;
      continue;
     }
    }
   }
   else if ((label|0) == 59) {
    label = 0;
    $178 = ((($$0$ph)) + 1|0);
    $179 = HEAP32[$155>>2]|0;
    $180 = HEAP32[$158>>2]|0;
    $181 = ($179|0)==($180|0);
    if ($181) {
     $182 = HEAP32[$156>>2]|0;
     $183 = ((($182)) + 40|0);
     $184 = HEAP32[$183>>2]|0;
     (FUNCTION_TABLE_ii[$184 & 2047]($156)|0);
     $$0$ph$be = $178;
     break;
    } else {
     $185 = ((($179)) + 1|0);
     HEAP32[$155>>2] = $185;
     $186 = HEAP8[$179>>0]|0;
     (__ZNSt3__211char_traitsIcE11to_int_typeEc($186)|0);
     $$0$ph$be = $178;
     break;
    }
   }
  } while(0);
  $$pre$pre = HEAP32[$4>>2]|0;
  $$0$ph = $$0$ph$be;$$pre = $$pre$pre;
 }
 if ((label|0) == 63) {
  HEAP32[$4>>2] = 4;
  $$pre$i$i$i73 = $156;
 }
 $187 = ($$pre$i$i$i73|0)==(0|0);
 if ($187) {
  $$sroa$049$0$copyload = 0;$221 = 1;
 } else {
  $188 = ((($$pre$i$i$i73)) + 12|0);
  $189 = HEAP32[$188>>2]|0;
  $190 = ((($$pre$i$i$i73)) + 16|0);
  $191 = HEAP32[$190>>2]|0;
  $192 = ($189|0)==($191|0);
  if ($192) {
   $193 = HEAP32[$$pre$i$i$i73>>2]|0;
   $194 = ((($193)) + 36|0);
   $195 = HEAP32[$194>>2]|0;
   $196 = (FUNCTION_TABLE_ii[$195 & 2047]($$pre$i$i$i73)|0);
   $$0$i$i$i$i71 = $196;
  } else {
   $197 = HEAP8[$189>>0]|0;
   $198 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($197)|0);
   $$0$i$i$i$i71 = $198;
  }
  $199 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
  $200 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i71,$199)|0);
  if ($200) {
   HEAP32[$1>>2] = 0;
   $$sroa$049$0$copyload = 0;$221 = 1;
  } else {
   $$sroa$049$0$copyload = $$pre$i$i$i73;$221 = 0;
  }
 }
 $201 = HEAP32[$2>>2]|0;
 $202 = ($201|0)==(0|0);
 do {
  if ($202) {
   label = 76;
  } else {
   $203 = ((($201)) + 12|0);
   $204 = HEAP32[$203>>2]|0;
   $205 = ((($201)) + 16|0);
   $206 = HEAP32[$205>>2]|0;
   $207 = ($204|0)==($206|0);
   if ($207) {
    $208 = HEAP32[$201>>2]|0;
    $209 = ((($208)) + 36|0);
    $210 = HEAP32[$209>>2]|0;
    $211 = (FUNCTION_TABLE_ii[$210 & 2047]($201)|0);
    $$0$i$i2$i$i77 = $211;
   } else {
    $212 = HEAP8[$204>>0]|0;
    $213 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($212)|0);
    $$0$i$i2$i$i77 = $213;
   }
   $214 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $215 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i77,$214)|0);
   if ($215) {
    HEAP32[$2>>2] = 0;
    label = 76;
    break;
   } else {
    if ($221) {
     break;
    } else {
     label = 78;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 76) {
  if ($221) {
   label = 78;
  }
 }
 if ((label|0) == 78) {
  $216 = HEAP32[$4>>2]|0;
  $217 = $216 | 2;
  HEAP32[$4>>2] = $217;
 }
 STACKTOP = sp;return ($$sroa$049$0$copyload|0);
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $$off = (($8) + -1)|0;
 $12 = ($$off>>>0)<(31);
 $13 = $12 & $11;
 if ($13) {
  HEAP32[$1>>2] = $8;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(24);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $$off = (($8) + -1)|0;
 $12 = ($$off>>>0)<(12);
 $13 = $12 & $11;
 if ($13) {
  HEAP32[$1>>2] = $8;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,3)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(366);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(13);
 $or$cond = $12 & $11;
 if ($or$cond) {
  $13 = (($8) + -1)|0;
  HEAP32[$1>>2] = $13;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(60);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i7 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i13 = 0, $$pre$i$i$i = 0, $$pre$i$i$i9 = 0, $$pre$i4$i$i15 = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i10 = 0, $sext = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($4)) + 8|0);
 L1: while(1) {
  $6 = HEAP32[$1>>2]|0;
  $7 = ($6|0)==(0|0);
  do {
   if ($7) {
    $98 = 1;
   } else {
    $8 = ((($6)) + 12|0);
    $9 = HEAP32[$8>>2]|0;
    $10 = ((($6)) + 16|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($9|0)==($11|0);
    if ($12) {
     $13 = HEAP32[$6>>2]|0;
     $14 = ((($13)) + 36|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = (FUNCTION_TABLE_ii[$15 & 2047]($6)|0);
     $$0$i$i$i$i = $16;
    } else {
     $17 = HEAP8[$9>>0]|0;
     $18 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($17)|0);
     $$0$i$i$i$i = $18;
    }
    $19 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $20 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$19)|0);
    if ($20) {
     HEAP32[$1>>2] = 0;
     $98 = 1;
     break;
    } else {
     $$pre$i$i$i = HEAP32[$1>>2]|0;
     $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
     $98 = $phitmp$i$i$i;
     break;
    }
   }
  } while(0);
  $21 = HEAP32[$2>>2]|0;
  $22 = ($21|0)==(0|0);
  do {
   if ($22) {
    label = 15;
   } else {
    $23 = ((($21)) + 12|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ((($21)) + 16|0);
    $26 = HEAP32[$25>>2]|0;
    $27 = ($24|0)==($26|0);
    if ($27) {
     $28 = HEAP32[$21>>2]|0;
     $29 = ((($28)) + 36|0);
     $30 = HEAP32[$29>>2]|0;
     $31 = (FUNCTION_TABLE_ii[$30 & 2047]($21)|0);
     $$0$i$i2$i$i = $31;
    } else {
     $32 = HEAP8[$24>>0]|0;
     $33 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($32)|0);
     $$0$i$i2$i$i = $33;
    }
    $34 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
    $35 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$34)|0);
    if ($35) {
     HEAP32[$2>>2] = 0;
     label = 15;
     break;
    } else {
     if ($98) {
      $99 = $21;
      break;
     } else {
      $$pre$i4$i$i15 = $21;
      break L1;
     }
    }
   }
  } while(0);
  if ((label|0) == 15) {
   label = 0;
   if ($98) {
    $$pre$i4$i$i15 = 0;
    break;
   } else {
    $99 = 0;
   }
  }
  $36 = HEAP32[$1>>2]|0;
  $37 = ((($36)) + 12|0);
  $38 = HEAP32[$37>>2]|0;
  $39 = ((($36)) + 16|0);
  $40 = HEAP32[$39>>2]|0;
  $41 = ($38|0)==($40|0);
  if ($41) {
   $42 = HEAP32[$36>>2]|0;
   $43 = ((($42)) + 36|0);
   $44 = HEAP32[$43>>2]|0;
   $45 = (FUNCTION_TABLE_ii[$44 & 2047]($36)|0);
   $$0$i$i = $45;
  } else {
   $46 = HEAP8[$38>>0]|0;
   $47 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($46)|0);
   $$0$i$i = $47;
  }
  $48 = $$0$i$i&255;
  $49 = ($48<<24>>24)>(-1);
  if (!($49)) {
   $$pre$i4$i$i15 = $99;
   break;
  }
  $sext = $$0$i$i << 24;
  $50 = $sext >> 24;
  $51 = HEAP32[$5>>2]|0;
  $52 = (($51) + ($50<<1)|0);
  $53 = HEAP16[$52>>1]|0;
  $54 = $53 & 8192;
  $55 = ($54<<16>>16)==(0);
  if ($55) {
   $$pre$i4$i$i15 = $99;
   break;
  }
  $56 = HEAP32[$1>>2]|0;
  $57 = ((($56)) + 12|0);
  $58 = HEAP32[$57>>2]|0;
  $59 = ((($56)) + 16|0);
  $60 = HEAP32[$59>>2]|0;
  $61 = ($58|0)==($60|0);
  if ($61) {
   $62 = HEAP32[$56>>2]|0;
   $63 = ((($62)) + 40|0);
   $64 = HEAP32[$63>>2]|0;
   (FUNCTION_TABLE_ii[$64 & 2047]($56)|0);
   continue;
  } else {
   $65 = ((($58)) + 1|0);
   HEAP32[$57>>2] = $65;
   $66 = HEAP8[$58>>0]|0;
   (__ZNSt3__211char_traitsIcE11to_int_typeEc($66)|0);
   continue;
  }
 }
 $67 = HEAP32[$1>>2]|0;
 $68 = ($67|0)==(0|0);
 do {
  if ($68) {
   $100 = 1;
  } else {
   $69 = ((($67)) + 12|0);
   $70 = HEAP32[$69>>2]|0;
   $71 = ((($67)) + 16|0);
   $72 = HEAP32[$71>>2]|0;
   $73 = ($70|0)==($72|0);
   if ($73) {
    $74 = HEAP32[$67>>2]|0;
    $75 = ((($74)) + 36|0);
    $76 = HEAP32[$75>>2]|0;
    $77 = (FUNCTION_TABLE_ii[$76 & 2047]($67)|0);
    $$0$i$i$i$i7 = $77;
   } else {
    $78 = HEAP8[$70>>0]|0;
    $79 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($78)|0);
    $$0$i$i$i$i7 = $79;
   }
   $80 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $81 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i7,$80)|0);
   if ($81) {
    HEAP32[$1>>2] = 0;
    $100 = 1;
    break;
   } else {
    $$pre$i$i$i9 = HEAP32[$1>>2]|0;
    $phitmp$i$i$i10 = ($$pre$i$i$i9|0)==(0|0);
    $100 = $phitmp$i$i$i10;
    break;
   }
  }
 } while(0);
 $82 = ($$pre$i4$i$i15|0)==(0|0);
 do {
  if ($82) {
   label = 38;
  } else {
   $83 = ((($$pre$i4$i$i15)) + 12|0);
   $84 = HEAP32[$83>>2]|0;
   $85 = ((($$pre$i4$i$i15)) + 16|0);
   $86 = HEAP32[$85>>2]|0;
   $87 = ($84|0)==($86|0);
   if ($87) {
    $88 = HEAP32[$$pre$i4$i$i15>>2]|0;
    $89 = ((($88)) + 36|0);
    $90 = HEAP32[$89>>2]|0;
    $91 = (FUNCTION_TABLE_ii[$90 & 2047]($$pre$i4$i$i15)|0);
    $$0$i$i2$i$i13 = $91;
   } else {
    $92 = HEAP8[$84>>0]|0;
    $93 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($92)|0);
    $$0$i$i2$i$i13 = $93;
   }
   $94 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $95 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i13,$94)|0);
   if ($95) {
    HEAP32[$2>>2] = 0;
    label = 38;
    break;
   } else {
    if ($100) {
     break;
    } else {
     label = 40;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 38) {
  if ($100) {
   label = 40;
  }
 }
 if ((label|0) == 40) {
  $96 = HEAP32[$3>>2]|0;
  $97 = $96 | 2;
  HEAP32[$3>>2] = $97;
 }
 return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond20 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 2047]($7)|0);
 $12 = ((($11)) + 11|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = ($13<<24>>24)<(0);
 if ($14) {
  $15 = ((($11)) + 4|0);
  $16 = HEAP32[$15>>2]|0;
  $28 = $16;
 } else {
  $17 = $13&255;
  $28 = $17;
 }
 $18 = ((($11)) + 12|0);
 $19 = ((($18)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = ((($11)) + 16|0);
  $23 = HEAP32[$22>>2]|0;
  $26 = $23;
 } else {
  $24 = $20&255;
  $26 = $24;
 }
 $25 = (0 - ($26))|0;
 $27 = ($28|0)==($25|0);
 do {
  if ($27) {
   $29 = HEAP32[$4>>2]|0;
   $30 = $29 | 4;
   HEAP32[$4>>2] = $30;
  } else {
   $31 = HEAP32[$3>>2]|0;
   HEAP32[$6>>2] = $31;
   $32 = ((($11)) + 24|0);
   ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
   $33 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$11,$32,$5,$4,0)|0);
   $34 = $33;
   $35 = $11;
   $36 = (($34) - ($35))|0;
   $37 = ($36|0)==(0);
   $38 = HEAP32[$1>>2]|0;
   $39 = ($38|0)==(12);
   $or$cond = $39 & $37;
   if ($or$cond) {
    HEAP32[$1>>2] = 0;
    break;
   }
   $40 = ($36|0)==(12);
   $41 = ($38|0)<(12);
   $or$cond20 = $41 & $40;
   if ($or$cond20) {
    $42 = (($38) + 12)|0;
    HEAP32[$1>>2] = $42;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(61);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,1)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(7);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$ = 0, $$0 = 0, $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,4)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = ($8|0)<(69);
  if ($12) {
   $13 = (($8) + 2000)|0;
   $$0 = $13;
  } else {
   $14 = ($8|0)<(100);
   $15 = (($8) + 1900)|0;
   $$ = $14 ? $15 : $8;
   $$0 = $$;
  }
  $16 = (($$0) + -1900)|0;
  HEAP32[$1>>2] = $16;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,4)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = (($8) + -1900)|0;
  HEAP32[$1>>2] = $12;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIcEE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i11 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i17 = 0, $$pre$i$i$i = 0, $$pre$i$i$i13 = 0, $$pre$i4$i$i1925 = 0, $$sink3 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($5|0)==(0|0);
 do {
  if ($6) {
   $95 = 1;
  } else {
   $7 = ((($5)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    $12 = HEAP32[$5>>2]|0;
    $13 = ((($12)) + 36|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (FUNCTION_TABLE_ii[$14 & 2047]($5)|0);
    $$0$i$i$i$i = $15;
   } else {
    $16 = HEAP8[$8>>0]|0;
    $17 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($16)|0);
    $$0$i$i$i$i = $17;
   }
   $18 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $19 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$18)|0);
   if ($19) {
    HEAP32[$1>>2] = 0;
    $95 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$1>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $95 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $20 = HEAP32[$2>>2]|0;
 $21 = ($20|0)==(0|0);
 do {
  if ($21) {
   label = 14;
  } else {
   $22 = ((($20)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($20)) + 16|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($23|0)==($25|0);
   if ($26) {
    $27 = HEAP32[$20>>2]|0;
    $28 = ((($27)) + 36|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (FUNCTION_TABLE_ii[$29 & 2047]($20)|0);
    $$0$i$i2$i$i = $30;
   } else {
    $31 = HEAP8[$23>>0]|0;
    $32 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($31)|0);
    $$0$i$i2$i$i = $32;
   }
   $33 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $34 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$33)|0);
   if ($34) {
    HEAP32[$2>>2] = 0;
    label = 14;
    break;
   } else {
    if ($95) {
     $$pre$i4$i$i1925 = $20;
     label = 16;
     break;
    } else {
     $$sink3 = 6;
     label = 38;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($95) {
   $$sink3 = 6;
   label = 38;
  } else {
   $$pre$i4$i$i1925 = 0;
   label = 16;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $35 = HEAP32[$1>>2]|0;
   $36 = ((($35)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ((($35)) + 16|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($37|0)==($39|0);
   if ($40) {
    $41 = HEAP32[$35>>2]|0;
    $42 = ((($41)) + 36|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = (FUNCTION_TABLE_ii[$43 & 2047]($35)|0);
    $$0$i$i = $44;
   } else {
    $45 = HEAP8[$37>>0]|0;
    $46 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($45)|0);
    $$0$i$i = $46;
   }
   $47 = $$0$i$i&255;
   $48 = HEAP32[$4>>2]|0;
   $49 = ((($48)) + 36|0);
   $50 = HEAP32[$49>>2]|0;
   $51 = (FUNCTION_TABLE_iiii[$50 & 2047]($4,$47,0)|0);
   $52 = ($51<<24>>24)==(37);
   if ($52) {
    $53 = HEAP32[$1>>2]|0;
    $54 = ((($53)) + 12|0);
    $55 = HEAP32[$54>>2]|0;
    $56 = ((($53)) + 16|0);
    $57 = HEAP32[$56>>2]|0;
    $58 = ($55|0)==($57|0);
    if ($58) {
     $59 = HEAP32[$53>>2]|0;
     $60 = ((($59)) + 40|0);
     $61 = HEAP32[$60>>2]|0;
     (FUNCTION_TABLE_ii[$61 & 2047]($53)|0);
    } else {
     $62 = ((($55)) + 1|0);
     HEAP32[$54>>2] = $62;
     $63 = HEAP8[$55>>0]|0;
     (__ZNSt3__211char_traitsIcE11to_int_typeEc($63)|0);
    }
    $64 = HEAP32[$1>>2]|0;
    $65 = ($64|0)==(0|0);
    do {
     if ($65) {
      $96 = 1;
     } else {
      $66 = ((($64)) + 12|0);
      $67 = HEAP32[$66>>2]|0;
      $68 = ((($64)) + 16|0);
      $69 = HEAP32[$68>>2]|0;
      $70 = ($67|0)==($69|0);
      if ($70) {
       $71 = HEAP32[$64>>2]|0;
       $72 = ((($71)) + 36|0);
       $73 = HEAP32[$72>>2]|0;
       $74 = (FUNCTION_TABLE_ii[$73 & 2047]($64)|0);
       $$0$i$i$i$i11 = $74;
      } else {
       $75 = HEAP8[$67>>0]|0;
       $76 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($75)|0);
       $$0$i$i$i$i11 = $76;
      }
      $77 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $78 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i11,$77)|0);
      if ($78) {
       HEAP32[$1>>2] = 0;
       $96 = 1;
       break;
      } else {
       $$pre$i$i$i13 = HEAP32[$1>>2]|0;
       $phitmp$i$i$i14 = ($$pre$i$i$i13|0)==(0|0);
       $96 = $phitmp$i$i$i14;
       break;
      }
     }
    } while(0);
    $79 = ($$pre$i4$i$i1925|0)==(0|0);
    do {
     if (!($79)) {
      $80 = ((($$pre$i4$i$i1925)) + 12|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = ((($$pre$i4$i$i1925)) + 16|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = ($81|0)==($83|0);
      if ($84) {
       $85 = HEAP32[$$pre$i4$i$i1925>>2]|0;
       $86 = ((($85)) + 36|0);
       $87 = HEAP32[$86>>2]|0;
       $88 = (FUNCTION_TABLE_ii[$87 & 2047]($$pre$i4$i$i1925)|0);
       $$0$i$i2$i$i17 = $88;
      } else {
       $89 = HEAP8[$81>>0]|0;
       $90 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($89)|0);
       $$0$i$i2$i$i17 = $90;
      }
      $91 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $92 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i17,$91)|0);
      if ($92) {
       HEAP32[$2>>2] = 0;
       break;
      } else {
       if ($96) {
        break L22;
       } else {
        $$sink3 = 2;
        label = 38;
        break L22;
       }
      }
     }
    } while(0);
    if ($96) {
     $$sink3 = 2;
     label = 38;
    }
   } else {
    $$sink3 = 4;
    label = 38;
   }
  }
 } while(0);
 if ((label|0) == 38) {
  $93 = HEAP32[$3>>2]|0;
  $94 = $93 | $$sink3;
  HEAP32[$3>>2] = $94;
 }
 return;
}
function __ZNSt3__220__get_up_to_n_digitsIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i25 = 0, $$0$i$i$i$i41 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i31 = 0, $$0$i$i2$i$i47 = 0, $$0$i$i37 = 0, $$0$in = 0, $$023 = 0, $$023$in = 0, $$2 = 0, $$pre$i$i$i = 0, $$pre$i$i$i27 = 0, $$pre$i$i$i43 = 0, $$pre$i4$i$i33 = 0, $$pre$i4$i$i49 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i28 = 0, $phitmp$i$i$i44 = 0, $phitmp$i5$i$i34 = 0, $sext = 0;
 var $sext54 = 0, $tmp$i$i36 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5|0)==(0|0);
 do {
  if ($6) {
   $178 = 1;
  } else {
   $7 = ((($5)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    $12 = HEAP32[$5>>2]|0;
    $13 = ((($12)) + 36|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (FUNCTION_TABLE_ii[$14 & 2047]($5)|0);
    $$0$i$i$i$i = $15;
   } else {
    $16 = HEAP8[$8>>0]|0;
    $17 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($16)|0);
    $$0$i$i$i$i = $17;
   }
   $18 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $19 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$18)|0);
   if ($19) {
    HEAP32[$0>>2] = 0;
    $178 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$0>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $178 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $20 = HEAP32[$1>>2]|0;
 $21 = ($20|0)==(0|0);
 do {
  if ($21) {
   label = 14;
  } else {
   $22 = ((($20)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($20)) + 16|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($23|0)==($25|0);
   if ($26) {
    $27 = HEAP32[$20>>2]|0;
    $28 = ((($27)) + 36|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (FUNCTION_TABLE_ii[$29 & 2047]($20)|0);
    $$0$i$i2$i$i = $30;
   } else {
    $31 = HEAP8[$23>>0]|0;
    $32 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($31)|0);
    $$0$i$i2$i$i = $32;
   }
   $33 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $34 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$33)|0);
   if ($34) {
    HEAP32[$1>>2] = 0;
    label = 14;
    break;
   } else {
    if ($178) {
     $179 = $20;
     label = 17;
     break;
    } else {
     label = 16;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($178) {
   label = 16;
  } else {
   $179 = 0;
   label = 17;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $35 = HEAP32[$2>>2]|0;
   $36 = $35 | 6;
   HEAP32[$2>>2] = $36;
   $$2 = 0;
  }
  else if ((label|0) == 17) {
   $37 = HEAP32[$0>>2]|0;
   $38 = ((($37)) + 12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($37)) + 16|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ($39|0)==($41|0);
   if ($42) {
    $43 = HEAP32[$37>>2]|0;
    $44 = ((($43)) + 36|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = (FUNCTION_TABLE_ii[$45 & 2047]($37)|0);
    $$0$i$i = $46;
   } else {
    $47 = HEAP8[$39>>0]|0;
    $48 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($47)|0);
    $$0$i$i = $48;
   }
   $49 = $$0$i$i&255;
   $50 = ($49<<24>>24)>(-1);
   if ($50) {
    $sext = $$0$i$i << 24;
    $51 = $sext >> 24;
    $52 = ((($3)) + 8|0);
    $53 = HEAP32[$52>>2]|0;
    $54 = (($53) + ($51<<1)|0);
    $55 = HEAP16[$54>>1]|0;
    $56 = $55 & 2048;
    $57 = ($56<<16>>16)==(0);
    if (!($57)) {
     $60 = HEAP32[$3>>2]|0;
     $61 = ((($60)) + 36|0);
     $62 = HEAP32[$61>>2]|0;
     $63 = (FUNCTION_TABLE_iiii[$62 & 2047]($3,$49,0)|0);
     $64 = $63 << 24 >> 24;
     $65 = HEAP32[$0>>2]|0;
     $66 = ((($65)) + 12|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = ((($65)) + 16|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = ($67|0)==($69|0);
     if ($70) {
      $71 = HEAP32[$65>>2]|0;
      $72 = ((($71)) + 40|0);
      $73 = HEAP32[$72>>2]|0;
      (FUNCTION_TABLE_ii[$73 & 2047]($65)|0);
      $$0$in = $64;$$023$in = $4;$$pre$i4$i$i33 = $179;$92 = $179;
     } else {
      $74 = ((($67)) + 1|0);
      HEAP32[$66>>2] = $74;
      $75 = HEAP8[$67>>0]|0;
      (__ZNSt3__211char_traitsIcE11to_int_typeEc($75)|0);
      $$0$in = $64;$$023$in = $4;$$pre$i4$i$i33 = $179;$92 = $179;
     }
     while(1) {
      $$0 = (($$0$in) + -48)|0;
      $$023 = (($$023$in) + -1)|0;
      $76 = HEAP32[$0>>2]|0;
      $77 = ($76|0)==(0|0);
      do {
       if ($77) {
        $106 = 1;
       } else {
        $78 = ((($76)) + 12|0);
        $79 = HEAP32[$78>>2]|0;
        $80 = ((($76)) + 16|0);
        $81 = HEAP32[$80>>2]|0;
        $82 = ($79|0)==($81|0);
        if ($82) {
         $83 = HEAP32[$76>>2]|0;
         $84 = ((($83)) + 36|0);
         $85 = HEAP32[$84>>2]|0;
         $86 = (FUNCTION_TABLE_ii[$85 & 2047]($76)|0);
         $$0$i$i$i$i25 = $86;
        } else {
         $87 = HEAP8[$79>>0]|0;
         $88 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($87)|0);
         $$0$i$i$i$i25 = $88;
        }
        $89 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $90 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i25,$89)|0);
        if ($90) {
         HEAP32[$0>>2] = 0;
         $106 = 1;
         break;
        } else {
         $$pre$i$i$i27 = HEAP32[$0>>2]|0;
         $phitmp$i$i$i28 = ($$pre$i$i$i27|0)==(0|0);
         $106 = $phitmp$i$i$i28;
         break;
        }
       }
      } while(0);
      $91 = ($92|0)==(0|0);
      do {
       if ($91) {
        $$pre$i4$i$i49 = $$pre$i4$i$i33;$107 = 1;$180 = 0;
       } else {
        $93 = ((($92)) + 12|0);
        $94 = HEAP32[$93>>2]|0;
        $95 = ((($92)) + 16|0);
        $96 = HEAP32[$95>>2]|0;
        $97 = ($94|0)==($96|0);
        if ($97) {
         $98 = HEAP32[$92>>2]|0;
         $99 = ((($98)) + 36|0);
         $100 = HEAP32[$99>>2]|0;
         $101 = (FUNCTION_TABLE_ii[$100 & 2047]($92)|0);
         $$0$i$i2$i$i31 = $101;
        } else {
         $102 = HEAP8[$94>>0]|0;
         $103 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($102)|0);
         $$0$i$i2$i$i31 = $103;
        }
        $104 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
        $105 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i31,$104)|0);
        if ($105) {
         HEAP32[$1>>2] = 0;
         $$pre$i4$i$i49 = 0;$107 = 1;$180 = 0;
         break;
        } else {
         $phitmp$i5$i$i34 = ($$pre$i4$i$i33|0)==(0|0);
         $$pre$i4$i$i49 = $$pre$i4$i$i33;$107 = $phitmp$i5$i$i34;$180 = $$pre$i4$i$i33;
         break;
        }
       }
      } while(0);
      $tmp$i$i36 = $106 ^ $107;
      $108 = ($$023$in|0)>(1);
      $109 = $108 & $tmp$i$i36;
      $110 = HEAP32[$0>>2]|0;
      if (!($109)) {
       break;
      }
      $111 = ((($110)) + 12|0);
      $112 = HEAP32[$111>>2]|0;
      $113 = ((($110)) + 16|0);
      $114 = HEAP32[$113>>2]|0;
      $115 = ($112|0)==($114|0);
      if ($115) {
       $116 = HEAP32[$110>>2]|0;
       $117 = ((($116)) + 36|0);
       $118 = HEAP32[$117>>2]|0;
       $119 = (FUNCTION_TABLE_ii[$118 & 2047]($110)|0);
       $$0$i$i37 = $119;
      } else {
       $120 = HEAP8[$112>>0]|0;
       $121 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($120)|0);
       $$0$i$i37 = $121;
      }
      $122 = $$0$i$i37&255;
      $123 = ($122<<24>>24)>(-1);
      if (!($123)) {
       $$2 = $$0;
       break L22;
      }
      $sext54 = $$0$i$i37 << 24;
      $124 = $sext54 >> 24;
      $125 = HEAP32[$52>>2]|0;
      $126 = (($125) + ($124<<1)|0);
      $127 = HEAP16[$126>>1]|0;
      $128 = $127 & 2048;
      $129 = ($128<<16>>16)==(0);
      if ($129) {
       $$2 = $$0;
       break L22;
      }
      $130 = ($$0*10)|0;
      $131 = HEAP32[$3>>2]|0;
      $132 = ((($131)) + 36|0);
      $133 = HEAP32[$132>>2]|0;
      $134 = (FUNCTION_TABLE_iiii[$133 & 2047]($3,$122,0)|0);
      $135 = $134 << 24 >> 24;
      $136 = (($135) + ($130))|0;
      $137 = HEAP32[$0>>2]|0;
      $138 = ((($137)) + 12|0);
      $139 = HEAP32[$138>>2]|0;
      $140 = ((($137)) + 16|0);
      $141 = HEAP32[$140>>2]|0;
      $142 = ($139|0)==($141|0);
      if ($142) {
       $143 = HEAP32[$137>>2]|0;
       $144 = ((($143)) + 40|0);
       $145 = HEAP32[$144>>2]|0;
       (FUNCTION_TABLE_ii[$145 & 2047]($137)|0);
       $$0$in = $136;$$023$in = $$023;$$pre$i4$i$i33 = $$pre$i4$i$i49;$92 = $180;
       continue;
      } else {
       $146 = ((($139)) + 1|0);
       HEAP32[$138>>2] = $146;
       $147 = HEAP8[$139>>0]|0;
       (__ZNSt3__211char_traitsIcE11to_int_typeEc($147)|0);
       $$0$in = $136;$$023$in = $$023;$$pre$i4$i$i33 = $$pre$i4$i$i49;$92 = $180;
       continue;
      }
     }
     $148 = ($110|0)==(0|0);
     do {
      if ($148) {
       $181 = 1;
      } else {
       $149 = ((($110)) + 12|0);
       $150 = HEAP32[$149>>2]|0;
       $151 = ((($110)) + 16|0);
       $152 = HEAP32[$151>>2]|0;
       $153 = ($150|0)==($152|0);
       if ($153) {
        $154 = HEAP32[$110>>2]|0;
        $155 = ((($154)) + 36|0);
        $156 = HEAP32[$155>>2]|0;
        $157 = (FUNCTION_TABLE_ii[$156 & 2047]($110)|0);
        $$0$i$i$i$i41 = $157;
       } else {
        $158 = HEAP8[$150>>0]|0;
        $159 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($158)|0);
        $$0$i$i$i$i41 = $159;
       }
       $160 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $161 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i41,$160)|0);
       if ($161) {
        HEAP32[$0>>2] = 0;
        $181 = 1;
        break;
       } else {
        $$pre$i$i$i43 = HEAP32[$0>>2]|0;
        $phitmp$i$i$i44 = ($$pre$i$i$i43|0)==(0|0);
        $181 = $phitmp$i$i$i44;
        break;
       }
      }
     } while(0);
     $162 = ($$pre$i4$i$i49|0)==(0|0);
     do {
      if ($162) {
       label = 62;
      } else {
       $163 = ((($$pre$i4$i$i49)) + 12|0);
       $164 = HEAP32[$163>>2]|0;
       $165 = ((($$pre$i4$i$i49)) + 16|0);
       $166 = HEAP32[$165>>2]|0;
       $167 = ($164|0)==($166|0);
       if ($167) {
        $168 = HEAP32[$$pre$i4$i$i49>>2]|0;
        $169 = ((($168)) + 36|0);
        $170 = HEAP32[$169>>2]|0;
        $171 = (FUNCTION_TABLE_ii[$170 & 2047]($$pre$i4$i$i49)|0);
        $$0$i$i2$i$i47 = $171;
       } else {
        $172 = HEAP8[$164>>0]|0;
        $173 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($172)|0);
        $$0$i$i2$i$i47 = $173;
       }
       $174 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $175 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i47,$174)|0);
       if ($175) {
        HEAP32[$1>>2] = 0;
        label = 62;
        break;
       } else {
        if ($181) {
         $$2 = $$0;
         break L22;
        } else {
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 62) {
      if (!($181)) {
       $$2 = $$0;
       break;
      }
     }
     $176 = HEAP32[$2>>2]|0;
     $177 = $176 | 2;
     HEAP32[$2>>2] = $177;
     $$2 = $$0;
     break;
    }
   }
   $58 = HEAP32[$2>>2]|0;
   $59 = $58 | 4;
   HEAP32[$2>>2] = $59;
   $$2 = 0;
  }
 } while(0);
 return ($$2|0);
}
function __ZNSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13do_date_orderEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $8;
 $9 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $9;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $10 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy,$$byval_copy1,$3,$4,$5,20872,(20904))|0);
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy1 = sp + 12|0;
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 $8 = ((($0)) + 8|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = ((($9)) + 20|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = (FUNCTION_TABLE_ii[$11 & 2047]($8)|0);
 $13 = HEAP32[$1>>2]|0;
 HEAP32[$6>>2] = $13;
 $14 = HEAP32[$2>>2]|0;
 HEAP32[$7>>2] = $14;
 $15 = ((($12)) + 8|0);
 $16 = ((($15)) + 3|0);
 $17 = HEAP8[$16>>0]|0;
 $18 = ($17<<24>>24)<(0);
 $19 = HEAP32[$12>>2]|0;
 $20 = $18 ? $19 : $12;
 $21 = ((($12)) + 4|0);
 $22 = HEAP32[$21>>2]|0;
 $23 = $17&255;
 $24 = $18 ? $22 : $23;
 $25 = (($20) + ($24<<2)|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 ;HEAP32[$$byval_copy1>>2]=HEAP32[$7>>2]|0;
 $26 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy,$$byval_copy1,$3,$4,$5,$20,$25)|0);
 STACKTOP = sp;return ($26|0);
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165472|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 24|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165472|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 16|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$sroa$0$0$copyload = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 8|0;
 $6 = sp + 4|0;
 $7 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $8 = (invoke_iii(734,($6|0),(165472|0))|0);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $13 = ___cxa_find_matching_catch_2()|0;
  $14 = tempRet0;
  __ZNSt3__26localeD2Ev($6);
  ___resumeException($13|0);
  // unreachable;
 } else {
  __ZNSt3__26localeD2Ev($6);
  $11 = ((($5)) + 20|0);
  $12 = HEAP32[$2>>2]|0;
  HEAP32[$7>>2] = $12;
  ;HEAP32[$$byval_copy>>2]=HEAP32[$7>>2]|0;
  __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($0,$11,$1,$$byval_copy,$4,$8);
  $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
  STACKTOP = sp;return ($$sroa$0$0$copyload|0);
 }
 return (0)|0;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$byval_copy = 0, $$byval_copy1 = 0, $$byval_copy10 = 0, $$byval_copy11 = 0, $$byval_copy12 = 0, $$byval_copy13 = 0, $$byval_copy14 = 0, $$byval_copy15 = 0, $$byval_copy16 = 0, $$byval_copy17 = 0, $$byval_copy18 = 0, $$byval_copy19 = 0, $$byval_copy2 = 0, $$byval_copy20 = 0, $$byval_copy21 = 0, $$byval_copy22 = 0, $$byval_copy23 = 0, $$byval_copy24 = 0, $$byval_copy25 = 0, $$byval_copy26 = 0;
 var $$byval_copy27 = 0, $$byval_copy28 = 0, $$byval_copy29 = 0, $$byval_copy3 = 0, $$byval_copy30 = 0, $$byval_copy4 = 0, $$byval_copy5 = 0, $$byval_copy6 = 0, $$byval_copy7 = 0, $$byval_copy8 = 0, $$byval_copy9 = 0, $$sroa$097$0 = 0, $$sroa$097$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0;
 var $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 256|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(256|0);
 $$byval_copy30 = sp + 248|0;
 $$byval_copy29 = sp + 244|0;
 $$byval_copy28 = sp + 240|0;
 $$byval_copy27 = sp + 236|0;
 $$byval_copy26 = sp + 232|0;
 $$byval_copy25 = sp + 228|0;
 $$byval_copy24 = sp + 224|0;
 $$byval_copy23 = sp + 220|0;
 $$byval_copy22 = sp + 216|0;
 $$byval_copy21 = sp + 212|0;
 $$byval_copy20 = sp + 208|0;
 $$byval_copy19 = sp + 204|0;
 $$byval_copy18 = sp + 200|0;
 $$byval_copy17 = sp + 196|0;
 $$byval_copy16 = sp + 192|0;
 $$byval_copy15 = sp + 188|0;
 $$byval_copy14 = sp + 184|0;
 $$byval_copy13 = sp + 180|0;
 $$byval_copy12 = sp + 176|0;
 $$byval_copy11 = sp + 172|0;
 $$byval_copy10 = sp + 168|0;
 $$byval_copy9 = sp + 164|0;
 $$byval_copy8 = sp + 160|0;
 $$byval_copy7 = sp + 156|0;
 $$byval_copy6 = sp + 152|0;
 $$byval_copy5 = sp + 148|0;
 $$byval_copy4 = sp + 144|0;
 $$byval_copy3 = sp + 140|0;
 $$byval_copy2 = sp + 136|0;
 $$byval_copy1 = sp + 132|0;
 $$byval_copy = sp + 128|0;
 $8 = sp + 124|0;
 $9 = sp + 120|0;
 $10 = sp + 116|0;
 $11 = sp + 112|0;
 $12 = sp + 108|0;
 $13 = sp + 104|0;
 $14 = sp + 100|0;
 $15 = sp + 96|0;
 $16 = sp + 92|0;
 $17 = sp + 88|0;
 $18 = sp + 84|0;
 $19 = sp + 80|0;
 $20 = sp + 76|0;
 $21 = sp + 72|0;
 $22 = sp + 68|0;
 $23 = sp + 64|0;
 $24 = sp + 60|0;
 $25 = sp + 56|0;
 $26 = sp + 52|0;
 $27 = sp + 48|0;
 $28 = sp + 44|0;
 $29 = sp + 40|0;
 $30 = sp + 36|0;
 $31 = sp + 32|0;
 $32 = sp + 28|0;
 $33 = sp + 24|0;
 $34 = sp + 20|0;
 $35 = sp + 16|0;
 $36 = sp + 12|0;
 $37 = sp + 8|0;
 $38 = sp + 4|0;
 $39 = sp;
 HEAP32[$4>>2] = 0;
 __ZNKSt3__28ios_base6getlocEv($8,$3);
 __THREW__ = 0;
 $40 = (invoke_iii(734,($8|0),(165472|0))|0);
 $41 = __THREW__; __THREW__ = 0;
 $42 = $41&1;
 if ($42) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($49|0);
  // unreachable;
 }
 __ZNSt3__26localeD2Ev($8);
 $43 = $6 << 24 >> 24;
 $44 = ((($5)) + 16|0);
 $45 = ((($5)) + 8|0);
 $46 = ((($5)) + 20|0);
 $47 = ((($5)) + 24|0);
 $48 = ((($0)) + 8|0);
 do {
  switch ($43|0) {
  case 65: case 97:  {
   $51 = HEAP32[$2>>2]|0;
   HEAP32[$9>>2] = $51;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$47,$1,$$byval_copy,$4,$40);
   label = 28;
   break;
  }
  case 104: case 66: case 98:  {
   $52 = HEAP32[$2>>2]|0;
   HEAP32[$10>>2] = $52;
   ;HEAP32[$$byval_copy1>>2]=HEAP32[$10>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$44,$1,$$byval_copy1,$4,$40);
   label = 28;
   break;
  }
  case 99:  {
   $53 = HEAP32[$48>>2]|0;
   $54 = ((($53)) + 12|0);
   $55 = HEAP32[$54>>2]|0;
   $56 = (FUNCTION_TABLE_ii[$55 & 2047]($48)|0);
   $57 = HEAP32[$1>>2]|0;
   HEAP32[$11>>2] = $57;
   $58 = HEAP32[$2>>2]|0;
   HEAP32[$12>>2] = $58;
   $59 = ((($56)) + 8|0);
   $60 = ((($59)) + 3|0);
   $61 = HEAP8[$60>>0]|0;
   $62 = ($61<<24>>24)<(0);
   $63 = HEAP32[$56>>2]|0;
   $64 = $62 ? $63 : $56;
   $65 = ((($56)) + 4|0);
   $66 = HEAP32[$65>>2]|0;
   $67 = $61&255;
   $68 = $62 ? $66 : $67;
   $69 = (($64) + ($68<<2)|0);
   ;HEAP32[$$byval_copy2>>2]=HEAP32[$11>>2]|0;
   ;HEAP32[$$byval_copy3>>2]=HEAP32[$12>>2]|0;
   $70 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy2,$$byval_copy3,$3,$4,$5,$64,$69)|0);
   HEAP32[$1>>2] = $70;
   label = 28;
   break;
  }
  case 101: case 100:  {
   $71 = ((($5)) + 12|0);
   $72 = HEAP32[$2>>2]|0;
   HEAP32[$13>>2] = $72;
   ;HEAP32[$$byval_copy4>>2]=HEAP32[$13>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIwEE($0,$71,$1,$$byval_copy4,$4,$40);
   label = 28;
   break;
  }
  case 68:  {
   $73 = HEAP32[$1>>2]|0;
   HEAP32[$14>>2] = $73;
   $74 = HEAP32[$2>>2]|0;
   HEAP32[$15>>2] = $74;
   ;HEAP32[$$byval_copy5>>2]=HEAP32[$14>>2]|0;
   ;HEAP32[$$byval_copy6>>2]=HEAP32[$15>>2]|0;
   $75 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy5,$$byval_copy6,$3,$4,$5,20712,(20744))|0);
   HEAP32[$1>>2] = $75;
   label = 28;
   break;
  }
  case 70:  {
   $76 = HEAP32[$1>>2]|0;
   HEAP32[$16>>2] = $76;
   $77 = HEAP32[$2>>2]|0;
   HEAP32[$17>>2] = $77;
   ;HEAP32[$$byval_copy7>>2]=HEAP32[$16>>2]|0;
   ;HEAP32[$$byval_copy8>>2]=HEAP32[$17>>2]|0;
   $78 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy7,$$byval_copy8,$3,$4,$5,20744,(20776))|0);
   HEAP32[$1>>2] = $78;
   label = 28;
   break;
  }
  case 72:  {
   $79 = HEAP32[$2>>2]|0;
   HEAP32[$18>>2] = $79;
   ;HEAP32[$$byval_copy9>>2]=HEAP32[$18>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIwEE($0,$45,$1,$$byval_copy9,$4,$40);
   label = 28;
   break;
  }
  case 73:  {
   $80 = HEAP32[$2>>2]|0;
   HEAP32[$19>>2] = $80;
   ;HEAP32[$$byval_copy10>>2]=HEAP32[$19>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIwEE($0,$45,$1,$$byval_copy10,$4,$40);
   label = 28;
   break;
  }
  case 106:  {
   $81 = ((($5)) + 28|0);
   $82 = HEAP32[$2>>2]|0;
   HEAP32[$20>>2] = $82;
   ;HEAP32[$$byval_copy11>>2]=HEAP32[$20>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIwEE($0,$81,$1,$$byval_copy11,$4,$40);
   label = 28;
   break;
  }
  case 109:  {
   $83 = HEAP32[$2>>2]|0;
   HEAP32[$21>>2] = $83;
   ;HEAP32[$$byval_copy12>>2]=HEAP32[$21>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIwEE($0,$44,$1,$$byval_copy12,$4,$40);
   label = 28;
   break;
  }
  case 77:  {
   $84 = ((($5)) + 4|0);
   $85 = HEAP32[$2>>2]|0;
   HEAP32[$22>>2] = $85;
   ;HEAP32[$$byval_copy13>>2]=HEAP32[$22>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIwEE($0,$84,$1,$$byval_copy13,$4,$40);
   label = 28;
   break;
  }
  case 116: case 110:  {
   $86 = HEAP32[$2>>2]|0;
   HEAP32[$23>>2] = $86;
   ;HEAP32[$$byval_copy14>>2]=HEAP32[$23>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIwEE($0,$1,$$byval_copy14,$4,$40);
   label = 28;
   break;
  }
  case 112:  {
   $87 = HEAP32[$2>>2]|0;
   HEAP32[$24>>2] = $87;
   ;HEAP32[$$byval_copy15>>2]=HEAP32[$24>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIwEE($0,$45,$1,$$byval_copy15,$4,$40);
   label = 28;
   break;
  }
  case 114:  {
   $88 = HEAP32[$1>>2]|0;
   HEAP32[$25>>2] = $88;
   $89 = HEAP32[$2>>2]|0;
   HEAP32[$26>>2] = $89;
   ;HEAP32[$$byval_copy16>>2]=HEAP32[$25>>2]|0;
   ;HEAP32[$$byval_copy17>>2]=HEAP32[$26>>2]|0;
   $90 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy16,$$byval_copy17,$3,$4,$5,20776,(20820))|0);
   HEAP32[$1>>2] = $90;
   label = 28;
   break;
  }
  case 82:  {
   $91 = HEAP32[$1>>2]|0;
   HEAP32[$27>>2] = $91;
   $92 = HEAP32[$2>>2]|0;
   HEAP32[$28>>2] = $92;
   ;HEAP32[$$byval_copy18>>2]=HEAP32[$27>>2]|0;
   ;HEAP32[$$byval_copy19>>2]=HEAP32[$28>>2]|0;
   $93 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy18,$$byval_copy19,$3,$4,$5,20820,(20840))|0);
   HEAP32[$1>>2] = $93;
   label = 28;
   break;
  }
  case 83:  {
   $94 = HEAP32[$2>>2]|0;
   HEAP32[$29>>2] = $94;
   ;HEAP32[$$byval_copy20>>2]=HEAP32[$29>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIwEE($0,$5,$1,$$byval_copy20,$4,$40);
   label = 28;
   break;
  }
  case 84:  {
   $95 = HEAP32[$1>>2]|0;
   HEAP32[$30>>2] = $95;
   $96 = HEAP32[$2>>2]|0;
   HEAP32[$31>>2] = $96;
   ;HEAP32[$$byval_copy21>>2]=HEAP32[$30>>2]|0;
   ;HEAP32[$$byval_copy22>>2]=HEAP32[$31>>2]|0;
   $97 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy21,$$byval_copy22,$3,$4,$5,20840,(20872))|0);
   HEAP32[$1>>2] = $97;
   label = 28;
   break;
  }
  case 119:  {
   $98 = HEAP32[$2>>2]|0;
   HEAP32[$32>>2] = $98;
   ;HEAP32[$$byval_copy23>>2]=HEAP32[$32>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIwEE($0,$47,$1,$$byval_copy23,$4,$40);
   label = 28;
   break;
  }
  case 120:  {
   $99 = HEAP32[$0>>2]|0;
   $100 = ((($99)) + 20|0);
   $101 = HEAP32[$100>>2]|0;
   $102 = HEAP32[$1>>2]|0;
   HEAP32[$33>>2] = $102;
   $103 = HEAP32[$2>>2]|0;
   HEAP32[$34>>2] = $103;
   ;HEAP32[$$byval_copy24>>2]=HEAP32[$33>>2]|0;
   ;HEAP32[$$byval_copy25>>2]=HEAP32[$34>>2]|0;
   $104 = (FUNCTION_TABLE_iiiiiii[$101 & 1023]($0,$$byval_copy24,$$byval_copy25,$3,$4,$5)|0);
   $$sroa$097$0 = $104;
   break;
  }
  case 88:  {
   $105 = HEAP32[$48>>2]|0;
   $106 = ((($105)) + 24|0);
   $107 = HEAP32[$106>>2]|0;
   $108 = (FUNCTION_TABLE_ii[$107 & 2047]($48)|0);
   $109 = HEAP32[$1>>2]|0;
   HEAP32[$35>>2] = $109;
   $110 = HEAP32[$2>>2]|0;
   HEAP32[$36>>2] = $110;
   $111 = ((($108)) + 8|0);
   $112 = ((($111)) + 3|0);
   $113 = HEAP8[$112>>0]|0;
   $114 = ($113<<24>>24)<(0);
   $115 = HEAP32[$108>>2]|0;
   $116 = $114 ? $115 : $108;
   $117 = ((($108)) + 4|0);
   $118 = HEAP32[$117>>2]|0;
   $119 = $113&255;
   $120 = $114 ? $118 : $119;
   $121 = (($116) + ($120<<2)|0);
   ;HEAP32[$$byval_copy26>>2]=HEAP32[$35>>2]|0;
   ;HEAP32[$$byval_copy27>>2]=HEAP32[$36>>2]|0;
   $122 = (__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$$byval_copy26,$$byval_copy27,$3,$4,$5,$116,$121)|0);
   HEAP32[$1>>2] = $122;
   label = 28;
   break;
  }
  case 121:  {
   $123 = HEAP32[$2>>2]|0;
   HEAP32[$37>>2] = $123;
   ;HEAP32[$$byval_copy28>>2]=HEAP32[$37>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($0,$46,$1,$$byval_copy28,$4,$40);
   label = 28;
   break;
  }
  case 89:  {
   $124 = HEAP32[$2>>2]|0;
   HEAP32[$38>>2] = $124;
   ;HEAP32[$$byval_copy29>>2]=HEAP32[$38>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIwEE($0,$46,$1,$$byval_copy29,$4,$40);
   label = 28;
   break;
  }
  case 37:  {
   $125 = HEAP32[$2>>2]|0;
   HEAP32[$39>>2] = $125;
   ;HEAP32[$$byval_copy30>>2]=HEAP32[$39>>2]|0;
   __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIwEE($0,$1,$$byval_copy30,$4,$40);
   label = 28;
   break;
  }
  default: {
   $126 = HEAP32[$4>>2]|0;
   $127 = $126 | 4;
   HEAP32[$4>>2] = $127;
   label = 28;
  }
  }
 } while(0);
 if ((label|0) == 28) {
  $$sroa$097$0$copyload = HEAP32[$1>>2]|0;
  $$sroa$097$0 = $$sroa$097$0$copyload;
 }
 STACKTOP = sp;return ($$sroa$097$0|0);
}
function __ZNKSt3__220__time_get_c_storageIwE7__weeksEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161152]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161152)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(993);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41781] = 166956;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41781]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIwE8__monthsEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161136]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161136)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(994);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41738] = 166664;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41738]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIwE7__am_pmEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[161120]|0;
 $2 = ($1<<24>>24)==(0);
 do {
  if ($2) {
   $3 = (___cxa_guard_acquire(161120)|0);
   $4 = ($3|0)==(0);
   if (!($4)) {
    __THREW__ = 0;
    invoke_v(995);
    $5 = __THREW__; __THREW__ = 0;
    $6 = $5&1;
    if ($6) {
     $8 = ___cxa_find_matching_catch_2()|0;
     $9 = tempRet0;
     ___resumeException($8|0);
     // unreachable;
    } else {
     HEAP32[41665] = 166372;
     break;
    }
   }
  }
 } while(0);
 $7 = HEAP32[41665]|0;
 return ($7|0);
}
function __ZNKSt3__220__time_get_c_storageIwE3__cEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161112]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161112)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[166360>>2]=0|0;HEAP32[166360+4>>2]=0|0;HEAP32[166360+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIwE6lengthEPKw(19744)|0);
    $7 = ($6>>>0)>(1073741807);
    L4: do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(166360|0));
      $8 = __THREW__; __THREW__ = 0;
      label = 15;
     } else {
      $9 = ($6>>>0)<(2);
      do {
       if ($9) {
        $10 = $6&255;
        HEAP8[((((166368)) + 3|0))>>0] = $10;
        $$0$i$i = 166360;
       } else {
        $11 = (($6) + 4)|0;
        $12 = $11 & -4;
        $13 = ($12>>>0)>(1073741823);
        if (!($13)) {
         $20 = $12 << 2;
         __THREW__ = 0;
         $21 = (invoke_ii(988,($20|0))|0);
         $22 = __THREW__; __THREW__ = 0;
         $23 = $22&1;
         if ($23) {
          label = 15;
          break L4;
         }
         HEAP32[41590] = $21;
         $24 = $12 | -2147483648;
         HEAP32[(166368)>>2] = $24;
         HEAP32[(166364)>>2] = $6;
         $$0$i$i = $21;
         break;
        }
        $14 = (___cxa_allocate_exception(8)|0);
        __THREW__ = 0;
        invoke_vii(751,($14|0),(158021|0));
        $15 = __THREW__; __THREW__ = 0;
        $16 = $15&1;
        if ($16) {
         $18 = ___cxa_find_matching_catch_2()|0;
         $19 = tempRet0;
         ___cxa_free_exception(($14|0));
         $eh$lpad$body$index2Z2D = $19;$eh$lpad$body$indexZ2D = $18;
         break L4;
        } else {
         HEAP32[$14>>2] = (22172);
         __THREW__ = 0;
         invoke_viii(946,($14|0),(2528|0),(713|0));
         $17 = __THREW__; __THREW__ = 0;
         label = 15;
         break L4;
        }
       }
      } while(0);
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,19744,$6)|0);
      $25 = (($$0$i$i) + ($6<<2)|0);
      HEAP32[$1>>2] = 0;
      __ZNSt3__211char_traitsIwE6assignERwRKw($25,$1);
      (___cxa_atexit((996|0),(166360|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    if ((label|0) == 15) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $eh$lpad$body$index2Z2D = $27;$eh$lpad$body$indexZ2D = $26;
    }
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (166360|0);
}
function __ZNKSt3__220__time_get_c_storageIwE3__rEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161104]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161104)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[166348>>2]=0|0;HEAP32[166348+4>>2]=0|0;HEAP32[166348+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIwE6lengthEPKw(19696)|0);
    $7 = ($6>>>0)>(1073741807);
    L4: do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(166348|0));
      $8 = __THREW__; __THREW__ = 0;
      label = 15;
     } else {
      $9 = ($6>>>0)<(2);
      do {
       if ($9) {
        $10 = $6&255;
        HEAP8[((((166356)) + 3|0))>>0] = $10;
        $$0$i$i = 166348;
       } else {
        $11 = (($6) + 4)|0;
        $12 = $11 & -4;
        $13 = ($12>>>0)>(1073741823);
        if (!($13)) {
         $20 = $12 << 2;
         __THREW__ = 0;
         $21 = (invoke_ii(988,($20|0))|0);
         $22 = __THREW__; __THREW__ = 0;
         $23 = $22&1;
         if ($23) {
          label = 15;
          break L4;
         }
         HEAP32[41587] = $21;
         $24 = $12 | -2147483648;
         HEAP32[(166356)>>2] = $24;
         HEAP32[(166352)>>2] = $6;
         $$0$i$i = $21;
         break;
        }
        $14 = (___cxa_allocate_exception(8)|0);
        __THREW__ = 0;
        invoke_vii(751,($14|0),(158021|0));
        $15 = __THREW__; __THREW__ = 0;
        $16 = $15&1;
        if ($16) {
         $18 = ___cxa_find_matching_catch_2()|0;
         $19 = tempRet0;
         ___cxa_free_exception(($14|0));
         $eh$lpad$body$index2Z2D = $19;$eh$lpad$body$indexZ2D = $18;
         break L4;
        } else {
         HEAP32[$14>>2] = (22172);
         __THREW__ = 0;
         invoke_viii(946,($14|0),(2528|0),(713|0));
         $17 = __THREW__; __THREW__ = 0;
         label = 15;
         break L4;
        }
       }
      } while(0);
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,19696,$6)|0);
      $25 = (($$0$i$i) + ($6<<2)|0);
      HEAP32[$1>>2] = 0;
      __ZNSt3__211char_traitsIwE6assignERwRKw($25,$1);
      (___cxa_atexit((996|0),(166348|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    if ((label|0) == 15) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $eh$lpad$body$index2Z2D = $27;$eh$lpad$body$indexZ2D = $26;
    }
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (166348|0);
}
function __ZNKSt3__220__time_get_c_storageIwE3__xEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161096]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161096)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[166336>>2]=0|0;HEAP32[166336+4>>2]=0|0;HEAP32[166336+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIwE6lengthEPKw(19660)|0);
    $7 = ($6>>>0)>(1073741807);
    L4: do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(166336|0));
      $8 = __THREW__; __THREW__ = 0;
      label = 15;
     } else {
      $9 = ($6>>>0)<(2);
      do {
       if ($9) {
        $10 = $6&255;
        HEAP8[((((166344)) + 3|0))>>0] = $10;
        $$0$i$i = 166336;
       } else {
        $11 = (($6) + 4)|0;
        $12 = $11 & -4;
        $13 = ($12>>>0)>(1073741823);
        if (!($13)) {
         $20 = $12 << 2;
         __THREW__ = 0;
         $21 = (invoke_ii(988,($20|0))|0);
         $22 = __THREW__; __THREW__ = 0;
         $23 = $22&1;
         if ($23) {
          label = 15;
          break L4;
         }
         HEAP32[41584] = $21;
         $24 = $12 | -2147483648;
         HEAP32[(166344)>>2] = $24;
         HEAP32[(166340)>>2] = $6;
         $$0$i$i = $21;
         break;
        }
        $14 = (___cxa_allocate_exception(8)|0);
        __THREW__ = 0;
        invoke_vii(751,($14|0),(158021|0));
        $15 = __THREW__; __THREW__ = 0;
        $16 = $15&1;
        if ($16) {
         $18 = ___cxa_find_matching_catch_2()|0;
         $19 = tempRet0;
         ___cxa_free_exception(($14|0));
         $eh$lpad$body$index2Z2D = $19;$eh$lpad$body$indexZ2D = $18;
         break L4;
        } else {
         HEAP32[$14>>2] = (22172);
         __THREW__ = 0;
         invoke_viii(946,($14|0),(2528|0),(713|0));
         $17 = __THREW__; __THREW__ = 0;
         label = 15;
         break L4;
        }
       }
      } while(0);
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,19660,$6)|0);
      $25 = (($$0$i$i) + ($6<<2)|0);
      HEAP32[$1>>2] = 0;
      __ZNSt3__211char_traitsIwE6assignERwRKw($25,$1);
      (___cxa_atexit((996|0),(166336|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    if ((label|0) == 15) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $eh$lpad$body$index2Z2D = $27;$eh$lpad$body$indexZ2D = $26;
    }
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (166336|0);
}
function __ZNKSt3__220__time_get_c_storageIwE3__XEv($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$index2Z2D = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = HEAP8[161088]|0;
 $3 = ($2<<24>>24)==(0);
 L1: do {
  if ($3) {
   $4 = (___cxa_guard_acquire(161088)|0);
   $5 = ($4|0)==(0);
   if (!($5)) {
    ;HEAP32[166324>>2]=0|0;HEAP32[166324+4>>2]=0|0;HEAP32[166324+8>>2]=0|0;
    $6 = (__ZNSt3__211char_traitsIwE6lengthEPKw(19624)|0);
    $7 = ($6>>>0)>(1073741807);
    L4: do {
     if ($7) {
      __THREW__ = 0;
      invoke_vi(987,(166324|0));
      $8 = __THREW__; __THREW__ = 0;
      label = 15;
     } else {
      $9 = ($6>>>0)<(2);
      do {
       if ($9) {
        $10 = $6&255;
        HEAP8[((((166332)) + 3|0))>>0] = $10;
        $$0$i$i = 166324;
       } else {
        $11 = (($6) + 4)|0;
        $12 = $11 & -4;
        $13 = ($12>>>0)>(1073741823);
        if (!($13)) {
         $20 = $12 << 2;
         __THREW__ = 0;
         $21 = (invoke_ii(988,($20|0))|0);
         $22 = __THREW__; __THREW__ = 0;
         $23 = $22&1;
         if ($23) {
          label = 15;
          break L4;
         }
         HEAP32[41581] = $21;
         $24 = $12 | -2147483648;
         HEAP32[(166332)>>2] = $24;
         HEAP32[(166328)>>2] = $6;
         $$0$i$i = $21;
         break;
        }
        $14 = (___cxa_allocate_exception(8)|0);
        __THREW__ = 0;
        invoke_vii(751,($14|0),(158021|0));
        $15 = __THREW__; __THREW__ = 0;
        $16 = $15&1;
        if ($16) {
         $18 = ___cxa_find_matching_catch_2()|0;
         $19 = tempRet0;
         ___cxa_free_exception(($14|0));
         $eh$lpad$body$index2Z2D = $19;$eh$lpad$body$indexZ2D = $18;
         break L4;
        } else {
         HEAP32[$14>>2] = (22172);
         __THREW__ = 0;
         invoke_viii(946,($14|0),(2528|0),(713|0));
         $17 = __THREW__; __THREW__ = 0;
         label = 15;
         break L4;
        }
       }
      } while(0);
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,19624,$6)|0);
      $25 = (($$0$i$i) + ($6<<2)|0);
      HEAP32[$1>>2] = 0;
      __ZNSt3__211char_traitsIwE6assignERwRKw($25,$1);
      (___cxa_atexit((996|0),(166324|0),(___dso_handle|0))|0);
      break L1;
     }
    } while(0);
    if ((label|0) == 15) {
     $26 = ___cxa_find_matching_catch_2()|0;
     $27 = tempRet0;
     $eh$lpad$body$index2Z2D = $27;$eh$lpad$body$indexZ2D = $26;
    }
    ___resumeException($eh$lpad$body$indexZ2D|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return (166324|0);
}
function __ZNSt3__2L11init_wam_pmEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161128]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161128)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 166372;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((166660)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((997|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw(166372,19828)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166384),19840)|0);
 return;
}
function ___cxx_global_array_dtor_121($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (166660);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($1);
  $3 = ($1|0)==(166372|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNSt3__2L12init_wmonthsEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161144]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161144)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 166664;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((166952)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((998|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw(166664,19852)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166676),19884)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166688),19920)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166700),19944)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166712),19968)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166724),19984)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166736),20004)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166748),20024)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166760),20052)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166772),20092)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166784),20124)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166796),20160)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166808),20196)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166820),20212)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166832),20228)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166844),20244)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166856),19968)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166868),20260)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166880),20276)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166892),20292)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166904),20308)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166916),20324)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166928),20340)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166940),20356)|0);
 return;
}
function ___cxx_global_array_dtor_94($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (166952);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($1);
  $3 = ($1|0)==(166664|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNSt3__2L11init_wweeksEv() {
 var $$0$i$i = 0, $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161160]|0;
 $1 = ($0<<24>>24)==(0);
 if ($1) {
  $2 = (___cxa_guard_acquire(161160)|0);
  $3 = ($2|0)==(0);
  if (!($3)) {
   $4 = 166956;
   while(1) {
    ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
    $$0$i$i = 0;
    while(1) {
     $exitcond$i$i = ($$0$i$i|0)==(3);
     if ($exitcond$i$i) {
      break;
     }
     $5 = (($4) + ($$0$i$i<<2)|0);
     HEAP32[$5>>2] = 0;
     $6 = (($$0$i$i) + 1)|0;
     $$0$i$i = $6;
    }
    $7 = ((($4)) + 12|0);
    $8 = ($7|0)==((167124)|0);
    if ($8) {
     break;
    } else {
     $4 = $7;
    }
   }
   (___cxa_atexit((999|0),(0|0),(___dso_handle|0))|0);
  }
 }
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw(166956,20372)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166968),20400)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166980),20428)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((166992),20460)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167004),20500)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167016),20536)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167028),20564)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167040),20600)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167052),20616)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167064),20632)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167076),20648)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167088),20664)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167100),20680)|0);
 (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw((167112),20696)|0);
 return;
}
function ___cxx_global_array_dtor_55($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (167124);
 while(1) {
  $1 = ((($2)) + -12|0);
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($1);
  $3 = ($1|0)==(166956|0);
  if ($3) {
   break;
  } else {
   $2 = $1;
  }
 }
 return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_weekdaynameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = HEAP32[$8>>2]|0;
 $10 = (FUNCTION_TABLE_ii[$9 & 2047]($7)|0);
 $11 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $11;
 $12 = ((($10)) + 168|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $13 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$10,$12,$5,$4,0)|0);
 $14 = $13;
 $15 = $10;
 $16 = (($14) - ($15))|0;
 $17 = ($16|0)<(168);
 if ($17) {
  $18 = (($16|0) / 12)&-1;
  $19 = (($18|0) % 7)&-1;
  HEAP32[$1>>2] = $19;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE15__get_monthnameERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 2047]($7)|0);
 $12 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $12;
 $13 = ((($11)) + 288|0);
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $14 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$11,$13,$5,$4,0)|0);
 $15 = $14;
 $16 = $11;
 $17 = (($15) - ($16))|0;
 $18 = ($17|0)<(288);
 if ($18) {
  $19 = (($17|0) / 12)&-1;
  $20 = (($19|0) % 12)&-1;
  HEAP32[$1>>2] = $20;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE3getES4_S4_RNS_8ios_baseERjP2tmPKwSC_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i54 = 0, $$0$i$i$i$i69 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i60 = 0, $$0$i$i2$i$i75 = 0, $$0$i$i66 = 0, $$0$ph = 0, $$0$ph$be = 0, $$0$ph$sink = 0, $$051 = 0, $$052 = 0, $$byval_copy = 0, $$byval_copy1 = 0, $$lcssa = 0, $$ph = 0, $$pre = 0, $$pre$i$i$i56 = 0, $$pre$i$i$i71 = 0;
 var $$pre$i4$i$i62 = 0, $$pre$i4$i$i62100 = 0, $$pre$i4$i$i62101 = 0, $$pre$pre = 0, $$sroa$049$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $phitmp$i5$i$i63 = 0, $tmp$i$i65 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy1 = sp + 16|0;
 $$byval_copy = sp + 12|0;
 $8 = sp + 8|0;
 $9 = sp + 4|0;
 $10 = sp;
 __ZNKSt3__28ios_base6getlocEv($8,$3);
 __THREW__ = 0;
 $11 = (invoke_iii(734,($8|0),(165472|0))|0);
 $12 = __THREW__; __THREW__ = 0;
 $13 = $12&1;
 if ($13) {
  $49 = ___cxa_find_matching_catch_2()|0;
  $50 = tempRet0;
  __ZNSt3__26localeD2Ev($8);
  ___resumeException($49|0);
  // unreachable;
 }
 __ZNSt3__26localeD2Ev($8);
 HEAP32[$4>>2] = 0;
 $$0$ph = $6;$$pre = 0;
 L4: while(1) {
  $14 = ($$0$ph|0)!=($7|0);
  $16 = $$pre;
  while(1) {
   $15 = ($16|0)==(0);
   $or$cond = $14 & $15;
   $17 = HEAP32[$1>>2]|0;
   if (!($or$cond)) {
    $$pre$i$i$i71 = $17;
    break L4;
   }
   $18 = ($17|0)==(0|0);
   $19 = $17;
   if ($18) {
    $146 = 0;$207 = 1;$74 = $19;
   } else {
    $20 = ((($17)) + 12|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ((($17)) + 16|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ($21|0)==($23|0);
    if ($24) {
     $25 = HEAP32[$17>>2]|0;
     $26 = ((($25)) + 36|0);
     $27 = HEAP32[$26>>2]|0;
     $28 = (FUNCTION_TABLE_ii[$27 & 2047]($17)|0);
     $$0$i$i$i$i = $28;
    } else {
     $29 = HEAP32[$21>>2]|0;
     $30 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($29)|0);
     $$0$i$i$i$i = $30;
    }
    $31 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $32 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$31)|0);
    if ($32) {
     HEAP32[$1>>2] = 0;
     $146 = 0;$207 = 1;$74 = 0;
    } else {
     $146 = $17;$207 = 0;$74 = $19;
    }
   }
   $33 = HEAP32[$2>>2]|0;
   $34 = ($33|0)==(0|0);
   $35 = $33;
   do {
    if ($34) {
     $$ph = $35;
     label = 17;
    } else {
     $36 = ((($33)) + 12|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ((($33)) + 16|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($37|0)==($39|0);
     if ($40) {
      $41 = HEAP32[$33>>2]|0;
      $42 = ((($41)) + 36|0);
      $43 = HEAP32[$42>>2]|0;
      $44 = (FUNCTION_TABLE_ii[$43 & 2047]($33)|0);
      $$0$i$i2$i$i = $44;
     } else {
      $45 = HEAP32[$37>>2]|0;
      $46 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($45)|0);
      $$0$i$i2$i$i = $46;
     }
     $47 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $48 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$47)|0);
     if ($48) {
      HEAP32[$2>>2] = 0;
      $$ph = 0;
      label = 17;
      break;
     } else {
      if ($207) {
       $208 = $33;$75 = $35;
       break;
      } else {
       label = 60;
       break L4;
      }
     }
    }
   } while(0);
   if ((label|0) == 17) {
    label = 0;
    if ($207) {
     label = 60;
     break L4;
    } else {
     $208 = 0;$75 = $$ph;
    }
   }
   $51 = HEAP32[$$0$ph>>2]|0;
   $52 = HEAP32[$11>>2]|0;
   $53 = ((($52)) + 52|0);
   $54 = HEAP32[$53>>2]|0;
   $55 = (FUNCTION_TABLE_iiii[$54 & 2047]($11,$51,0)|0);
   $56 = ($55<<24>>24)==(37);
   if ($56) {
    label = 21;
    break;
   }
   $79 = HEAP32[$$0$ph>>2]|0;
   $80 = HEAP32[$11>>2]|0;
   $81 = ((($80)) + 12|0);
   $82 = HEAP32[$81>>2]|0;
   $83 = (FUNCTION_TABLE_iiii[$82 & 2047]($11,8192,$79)|0);
   if ($83) {
    $$0$ph$sink = $$0$ph;
    label = 28;
    break;
   }
   $145 = ((($146)) + 12|0);
   $147 = HEAP32[$145>>2]|0;
   $148 = ((($146)) + 16|0);
   $149 = HEAP32[$148>>2]|0;
   $150 = ($147|0)==($149|0);
   if ($150) {
    $151 = HEAP32[$146>>2]|0;
    $152 = ((($151)) + 36|0);
    $153 = HEAP32[$152>>2]|0;
    $154 = (FUNCTION_TABLE_ii[$153 & 2047]($146)|0);
    $$0$i$i66 = $154;
   } else {
    $155 = HEAP32[$147>>2]|0;
    $156 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($155)|0);
    $$0$i$i66 = $156;
   }
   $157 = HEAP32[$11>>2]|0;
   $158 = ((($157)) + 28|0);
   $159 = HEAP32[$158>>2]|0;
   $160 = (FUNCTION_TABLE_iii[$159 & 2047]($11,$$0$i$i66)|0);
   $161 = HEAP32[$$0$ph>>2]|0;
   $162 = HEAP32[$11>>2]|0;
   $163 = ((($162)) + 28|0);
   $164 = HEAP32[$163>>2]|0;
   $165 = (FUNCTION_TABLE_iii[$164 & 2047]($11,$161)|0);
   $166 = ($160|0)==($165|0);
   if ($166) {
    label = 56;
    break;
   }
   HEAP32[$4>>2] = 4;
   $16 = 4;
  }
  L36: do {
   if ((label|0) == 21) {
    label = 0;
    $57 = ((($$0$ph)) + 4|0);
    $58 = ($57|0)==($7|0);
    if ($58) {
     label = 60;
     break L4;
    }
    $59 = HEAP32[$57>>2]|0;
    $60 = HEAP32[$11>>2]|0;
    $61 = ((($60)) + 52|0);
    $62 = HEAP32[$61>>2]|0;
    $63 = (FUNCTION_TABLE_iiii[$62 & 2047]($11,$59,0)|0);
    switch ($63<<24>>24) {
    case 48: case 69:  {
     $64 = ((($$0$ph)) + 8|0);
     $65 = ($64|0)==($7|0);
     if ($65) {
      label = 60;
      break L4;
     }
     $66 = HEAP32[$64>>2]|0;
     $67 = HEAP32[$11>>2]|0;
     $68 = ((($67)) + 52|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = (FUNCTION_TABLE_iiii[$69 & 2047]($11,$66,0)|0);
     $$051 = $63;$$052 = $70;$78 = $57;
     break;
    }
    default: {
     $$051 = 0;$$052 = $63;$78 = $$0$ph;
    }
    }
    $71 = HEAP32[$0>>2]|0;
    $72 = ((($71)) + 36|0);
    $73 = HEAP32[$72>>2]|0;
    HEAP32[$9>>2] = $74;
    HEAP32[$10>>2] = $75;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$9>>2]|0;
    ;HEAP32[$$byval_copy1>>2]=HEAP32[$10>>2]|0;
    $76 = (FUNCTION_TABLE_iiiiiiiii[$73 & 1023]($0,$$byval_copy,$$byval_copy1,$3,$4,$5,$$052,$$051)|0);
    HEAP32[$1>>2] = $76;
    $77 = ((($78)) + 8|0);
    $$0$ph$be = $77;
   }
   else if ((label|0) == 28) {
    while(1) {
     label = 0;
     $84 = ((($$0$ph$sink)) + 4|0);
     $85 = ($84|0)==($7|0);
     if ($85) {
      $$lcssa = $7;
      break;
     }
     $86 = HEAP32[$84>>2]|0;
     $87 = HEAP32[$11>>2]|0;
     $88 = ((($87)) + 12|0);
     $89 = HEAP32[$88>>2]|0;
     $90 = (FUNCTION_TABLE_iiii[$89 & 2047]($11,8192,$86)|0);
     if ($90) {
      $$0$ph$sink = $84;
      label = 28;
     } else {
      $$lcssa = $84;
      break;
     }
    }
    $$pre$i$i$i56 = $146;$$pre$i4$i$i62 = $208;$106 = $208;
    while(1) {
     $91 = ($$pre$i$i$i56|0)==(0|0);
     if ($91) {
      $120 = 1;$122 = 0;
     } else {
      $92 = ((($$pre$i$i$i56)) + 12|0);
      $93 = HEAP32[$92>>2]|0;
      $94 = ((($$pre$i$i$i56)) + 16|0);
      $95 = HEAP32[$94>>2]|0;
      $96 = ($93|0)==($95|0);
      if ($96) {
       $97 = HEAP32[$$pre$i$i$i56>>2]|0;
       $98 = ((($97)) + 36|0);
       $99 = HEAP32[$98>>2]|0;
       $100 = (FUNCTION_TABLE_ii[$99 & 2047]($$pre$i$i$i56)|0);
       $$0$i$i$i$i54 = $100;
      } else {
       $101 = HEAP32[$93>>2]|0;
       $102 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($101)|0);
       $$0$i$i$i$i54 = $102;
      }
      $103 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $104 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i54,$103)|0);
      if ($104) {
       HEAP32[$1>>2] = 0;
       $120 = 1;$122 = 0;
      } else {
       $120 = 0;$122 = $$pre$i$i$i56;
      }
     }
     $105 = ($106|0)==(0|0);
     do {
      if ($105) {
       $$pre$i4$i$i62101 = $$pre$i4$i$i62;
       label = 43;
      } else {
       $107 = ((($106)) + 12|0);
       $108 = HEAP32[$107>>2]|0;
       $109 = ((($106)) + 16|0);
       $110 = HEAP32[$109>>2]|0;
       $111 = ($108|0)==($110|0);
       if ($111) {
        $112 = HEAP32[$106>>2]|0;
        $113 = ((($112)) + 36|0);
        $114 = HEAP32[$113>>2]|0;
        $115 = (FUNCTION_TABLE_ii[$114 & 2047]($106)|0);
        $$0$i$i2$i$i60 = $115;
       } else {
        $116 = HEAP32[$108>>2]|0;
        $117 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($116)|0);
        $$0$i$i2$i$i60 = $117;
       }
       $118 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $119 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i60,$118)|0);
       if ($119) {
        HEAP32[$2>>2] = 0;
        $$pre$i4$i$i62101 = 0;
        label = 43;
        break;
       } else {
        $phitmp$i5$i$i63 = ($$pre$i4$i$i62|0)==(0|0);
        $tmp$i$i65 = $120 ^ $phitmp$i5$i$i63;
        if ($tmp$i$i65) {
         $$pre$i4$i$i62100 = $$pre$i4$i$i62;$209 = $$pre$i4$i$i62;
         break;
        } else {
         $$0$ph$be = $$lcssa;
         break L36;
        }
       }
      }
     } while(0);
     if ((label|0) == 43) {
      label = 0;
      if ($120) {
       $$0$ph$be = $$lcssa;
       break L36;
      } else {
       $$pre$i4$i$i62100 = $$pre$i4$i$i62101;$209 = 0;
      }
     }
     $121 = ((($122)) + 12|0);
     $123 = HEAP32[$121>>2]|0;
     $124 = ((($122)) + 16|0);
     $125 = HEAP32[$124>>2]|0;
     $126 = ($123|0)==($125|0);
     if ($126) {
      $127 = HEAP32[$122>>2]|0;
      $128 = ((($127)) + 36|0);
      $129 = HEAP32[$128>>2]|0;
      $130 = (FUNCTION_TABLE_ii[$129 & 2047]($122)|0);
      $$0$i$i = $130;
     } else {
      $131 = HEAP32[$123>>2]|0;
      $132 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($131)|0);
      $$0$i$i = $132;
     }
     $133 = HEAP32[$11>>2]|0;
     $134 = ((($133)) + 12|0);
     $135 = HEAP32[$134>>2]|0;
     $136 = (FUNCTION_TABLE_iiii[$135 & 2047]($11,8192,$$0$i$i)|0);
     if (!($136)) {
      $$0$ph$be = $$lcssa;
      break L36;
     }
     $137 = HEAP32[$121>>2]|0;
     $138 = HEAP32[$124>>2]|0;
     $139 = ($137|0)==($138|0);
     if ($139) {
      $140 = HEAP32[$122>>2]|0;
      $141 = ((($140)) + 40|0);
      $142 = HEAP32[$141>>2]|0;
      (FUNCTION_TABLE_ii[$142 & 2047]($122)|0);
      $$pre$i$i$i56 = $122;$$pre$i4$i$i62 = $$pre$i4$i$i62100;$106 = $209;
      continue;
     } else {
      $143 = ((($137)) + 4|0);
      HEAP32[$121>>2] = $143;
      $144 = HEAP32[$137>>2]|0;
      (__ZNSt3__211char_traitsIwE11to_int_typeEw($144)|0);
      $$pre$i$i$i56 = $122;$$pre$i4$i$i62 = $$pre$i4$i$i62100;$106 = $209;
      continue;
     }
    }
   }
   else if ((label|0) == 56) {
    label = 0;
    $167 = ((($$0$ph)) + 4|0);
    $168 = HEAP32[$145>>2]|0;
    $169 = HEAP32[$148>>2]|0;
    $170 = ($168|0)==($169|0);
    if ($170) {
     $171 = HEAP32[$146>>2]|0;
     $172 = ((($171)) + 40|0);
     $173 = HEAP32[$172>>2]|0;
     (FUNCTION_TABLE_ii[$173 & 2047]($146)|0);
     $$0$ph$be = $167;
     break;
    } else {
     $174 = ((($168)) + 4|0);
     HEAP32[$145>>2] = $174;
     $175 = HEAP32[$168>>2]|0;
     (__ZNSt3__211char_traitsIwE11to_int_typeEw($175)|0);
     $$0$ph$be = $167;
     break;
    }
   }
  } while(0);
  $$pre$pre = HEAP32[$4>>2]|0;
  $$0$ph = $$0$ph$be;$$pre = $$pre$pre;
 }
 if ((label|0) == 60) {
  HEAP32[$4>>2] = 4;
  $$pre$i$i$i71 = $146;
 }
 $176 = ($$pre$i$i$i71|0)==(0|0);
 if ($176) {
  $$sroa$049$0$copyload = 0;$210 = 1;
 } else {
  $177 = ((($$pre$i$i$i71)) + 12|0);
  $178 = HEAP32[$177>>2]|0;
  $179 = ((($$pre$i$i$i71)) + 16|0);
  $180 = HEAP32[$179>>2]|0;
  $181 = ($178|0)==($180|0);
  if ($181) {
   $182 = HEAP32[$$pre$i$i$i71>>2]|0;
   $183 = ((($182)) + 36|0);
   $184 = HEAP32[$183>>2]|0;
   $185 = (FUNCTION_TABLE_ii[$184 & 2047]($$pre$i$i$i71)|0);
   $$0$i$i$i$i69 = $185;
  } else {
   $186 = HEAP32[$178>>2]|0;
   $187 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($186)|0);
   $$0$i$i$i$i69 = $187;
  }
  $188 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
  $189 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i69,$188)|0);
  if ($189) {
   HEAP32[$1>>2] = 0;
   $$sroa$049$0$copyload = 0;$210 = 1;
  } else {
   $$sroa$049$0$copyload = $$pre$i$i$i71;$210 = 0;
  }
 }
 $190 = HEAP32[$2>>2]|0;
 $191 = ($190|0)==(0|0);
 do {
  if ($191) {
   label = 73;
  } else {
   $192 = ((($190)) + 12|0);
   $193 = HEAP32[$192>>2]|0;
   $194 = ((($190)) + 16|0);
   $195 = HEAP32[$194>>2]|0;
   $196 = ($193|0)==($195|0);
   if ($196) {
    $197 = HEAP32[$190>>2]|0;
    $198 = ((($197)) + 36|0);
    $199 = HEAP32[$198>>2]|0;
    $200 = (FUNCTION_TABLE_ii[$199 & 2047]($190)|0);
    $$0$i$i2$i$i75 = $200;
   } else {
    $201 = HEAP32[$193>>2]|0;
    $202 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($201)|0);
    $$0$i$i2$i$i75 = $202;
   }
   $203 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $204 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i75,$203)|0);
   if ($204) {
    HEAP32[$2>>2] = 0;
    label = 73;
    break;
   } else {
    if ($210) {
     break;
    } else {
     label = 75;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 73) {
  if ($210) {
   label = 75;
  }
 }
 if ((label|0) == 75) {
  $205 = HEAP32[$4>>2]|0;
  $206 = $205 | 2;
  HEAP32[$4>>2] = $206;
 }
 STACKTOP = sp;return ($$sroa$049$0$copyload|0);
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE9__get_dayERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $$off = (($8) + -1)|0;
 $12 = ($$off>>>0)<(31);
 $13 = $12 & $11;
 if ($13) {
  HEAP32[$1>>2] = $8;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_hourERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(24);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_12_hourERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $$off = (($8) + -1)|0;
 $12 = ($$off>>>0)<(12);
 $13 = $12 & $11;
 if ($13) {
  HEAP32[$1>>2] = $8;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE18__get_day_year_numERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,3)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(366);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_monthERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(13);
 $or$cond = $12 & $11;
 if ($or$cond) {
  $13 = (($8) + -1)|0;
  HEAP32[$1>>2] = $13;
 } else {
  $14 = $9 | 4;
  HEAP32[$4>>2] = $14;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_minuteERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(60);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE17__get_white_spaceERS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i7 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i13 = 0, $$pre$i$i$i = 0, $$pre$i$i$i9 = 0, $$pre$i4$i$i15 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $phitmp$i$i$i = 0;
 var $phitmp$i$i$i10 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 L1: while(1) {
  $5 = HEAP32[$1>>2]|0;
  $6 = ($5|0)==(0|0);
  do {
   if ($6) {
    $93 = 1;
   } else {
    $7 = ((($5)) + 12|0);
    $8 = HEAP32[$7>>2]|0;
    $9 = ((($5)) + 16|0);
    $10 = HEAP32[$9>>2]|0;
    $11 = ($8|0)==($10|0);
    if ($11) {
     $12 = HEAP32[$5>>2]|0;
     $13 = ((($12)) + 36|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = (FUNCTION_TABLE_ii[$14 & 2047]($5)|0);
     $$0$i$i$i$i = $15;
    } else {
     $16 = HEAP32[$8>>2]|0;
     $17 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($16)|0);
     $$0$i$i$i$i = $17;
    }
    $18 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $19 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$18)|0);
    if ($19) {
     HEAP32[$1>>2] = 0;
     $93 = 1;
     break;
    } else {
     $$pre$i$i$i = HEAP32[$1>>2]|0;
     $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
     $93 = $phitmp$i$i$i;
     break;
    }
   }
  } while(0);
  $20 = HEAP32[$2>>2]|0;
  $21 = ($20|0)==(0|0);
  do {
   if ($21) {
    label = 15;
   } else {
    $22 = ((($20)) + 12|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = ((($20)) + 16|0);
    $25 = HEAP32[$24>>2]|0;
    $26 = ($23|0)==($25|0);
    if ($26) {
     $27 = HEAP32[$20>>2]|0;
     $28 = ((($27)) + 36|0);
     $29 = HEAP32[$28>>2]|0;
     $30 = (FUNCTION_TABLE_ii[$29 & 2047]($20)|0);
     $$0$i$i2$i$i = $30;
    } else {
     $31 = HEAP32[$23>>2]|0;
     $32 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($31)|0);
     $$0$i$i2$i$i = $32;
    }
    $33 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
    $34 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$33)|0);
    if ($34) {
     HEAP32[$2>>2] = 0;
     label = 15;
     break;
    } else {
     if ($93) {
      $94 = $20;
      break;
     } else {
      $$pre$i4$i$i15 = $20;
      break L1;
     }
    }
   }
  } while(0);
  if ((label|0) == 15) {
   label = 0;
   if ($93) {
    $$pre$i4$i$i15 = 0;
    break;
   } else {
    $94 = 0;
   }
  }
  $35 = HEAP32[$1>>2]|0;
  $36 = ((($35)) + 12|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($35)) + 16|0);
  $39 = HEAP32[$38>>2]|0;
  $40 = ($37|0)==($39|0);
  if ($40) {
   $41 = HEAP32[$35>>2]|0;
   $42 = ((($41)) + 36|0);
   $43 = HEAP32[$42>>2]|0;
   $44 = (FUNCTION_TABLE_ii[$43 & 2047]($35)|0);
   $$0$i$i = $44;
  } else {
   $45 = HEAP32[$37>>2]|0;
   $46 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($45)|0);
   $$0$i$i = $46;
  }
  $47 = HEAP32[$4>>2]|0;
  $48 = ((($47)) + 12|0);
  $49 = HEAP32[$48>>2]|0;
  $50 = (FUNCTION_TABLE_iiii[$49 & 2047]($4,8192,$$0$i$i)|0);
  if (!($50)) {
   $$pre$i4$i$i15 = $94;
   break;
  }
  $51 = HEAP32[$1>>2]|0;
  $52 = ((($51)) + 12|0);
  $53 = HEAP32[$52>>2]|0;
  $54 = ((($51)) + 16|0);
  $55 = HEAP32[$54>>2]|0;
  $56 = ($53|0)==($55|0);
  if ($56) {
   $57 = HEAP32[$51>>2]|0;
   $58 = ((($57)) + 40|0);
   $59 = HEAP32[$58>>2]|0;
   (FUNCTION_TABLE_ii[$59 & 2047]($51)|0);
   continue;
  } else {
   $60 = ((($53)) + 4|0);
   HEAP32[$52>>2] = $60;
   $61 = HEAP32[$53>>2]|0;
   (__ZNSt3__211char_traitsIwE11to_int_typeEw($61)|0);
   continue;
  }
 }
 $62 = HEAP32[$1>>2]|0;
 $63 = ($62|0)==(0|0);
 do {
  if ($63) {
   $95 = 1;
  } else {
   $64 = ((($62)) + 12|0);
   $65 = HEAP32[$64>>2]|0;
   $66 = ((($62)) + 16|0);
   $67 = HEAP32[$66>>2]|0;
   $68 = ($65|0)==($67|0);
   if ($68) {
    $69 = HEAP32[$62>>2]|0;
    $70 = ((($69)) + 36|0);
    $71 = HEAP32[$70>>2]|0;
    $72 = (FUNCTION_TABLE_ii[$71 & 2047]($62)|0);
    $$0$i$i$i$i7 = $72;
   } else {
    $73 = HEAP32[$65>>2]|0;
    $74 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($73)|0);
    $$0$i$i$i$i7 = $74;
   }
   $75 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $76 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i7,$75)|0);
   if ($76) {
    HEAP32[$1>>2] = 0;
    $95 = 1;
    break;
   } else {
    $$pre$i$i$i9 = HEAP32[$1>>2]|0;
    $phitmp$i$i$i10 = ($$pre$i$i$i9|0)==(0|0);
    $95 = $phitmp$i$i$i10;
    break;
   }
  }
 } while(0);
 $77 = ($$pre$i4$i$i15|0)==(0|0);
 do {
  if ($77) {
   label = 37;
  } else {
   $78 = ((($$pre$i4$i$i15)) + 12|0);
   $79 = HEAP32[$78>>2]|0;
   $80 = ((($$pre$i4$i$i15)) + 16|0);
   $81 = HEAP32[$80>>2]|0;
   $82 = ($79|0)==($81|0);
   if ($82) {
    $83 = HEAP32[$$pre$i4$i$i15>>2]|0;
    $84 = ((($83)) + 36|0);
    $85 = HEAP32[$84>>2]|0;
    $86 = (FUNCTION_TABLE_ii[$85 & 2047]($$pre$i4$i$i15)|0);
    $$0$i$i2$i$i13 = $86;
   } else {
    $87 = HEAP32[$79>>2]|0;
    $88 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($87)|0);
    $$0$i$i2$i$i13 = $88;
   }
   $89 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $90 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i13,$89)|0);
   if ($90) {
    HEAP32[$2>>2] = 0;
    label = 37;
    break;
   } else {
    if ($95) {
     break;
    } else {
     label = 39;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 37) {
  if ($95) {
   label = 39;
  }
 }
 if ((label|0) == 39) {
  $91 = HEAP32[$3>>2]|0;
  $92 = $91 | 2;
  HEAP32[$3>>2] = $92;
 }
 return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_am_pmERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0;
 var $or$cond20 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = ((($0)) + 8|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($8)) + 8|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = (FUNCTION_TABLE_ii[$10 & 2047]($7)|0);
 $12 = ((($11)) + 8|0);
 $13 = ((($12)) + 3|0);
 $14 = HEAP8[$13>>0]|0;
 $15 = ($14<<24>>24)<(0);
 if ($15) {
  $16 = ((($11)) + 4|0);
  $17 = HEAP32[$16>>2]|0;
  $29 = $17;
 } else {
  $18 = $14&255;
  $29 = $18;
 }
 $19 = ((($11)) + 20|0);
 $20 = ((($19)) + 3|0);
 $21 = HEAP8[$20>>0]|0;
 $22 = ($21<<24>>24)<(0);
 if ($22) {
  $23 = ((($11)) + 16|0);
  $24 = HEAP32[$23>>2]|0;
  $27 = $24;
 } else {
  $25 = $21&255;
  $27 = $25;
 }
 $26 = (0 - ($27))|0;
 $28 = ($29|0)==($26|0);
 do {
  if ($28) {
   $30 = HEAP32[$4>>2]|0;
   $31 = $30 | 4;
   HEAP32[$4>>2] = $31;
  } else {
   $32 = HEAP32[$3>>2]|0;
   HEAP32[$6>>2] = $32;
   $33 = ((($11)) + 24|0);
   ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
   $34 = (__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb($2,$$byval_copy,$11,$33,$5,$4,0)|0);
   $35 = $34;
   $36 = $11;
   $37 = (($35) - ($36))|0;
   $38 = ($37|0)==(0);
   $39 = HEAP32[$1>>2]|0;
   $40 = ($39|0)==(12);
   $or$cond = $40 & $38;
   if ($or$cond) {
    HEAP32[$1>>2] = 0;
    break;
   }
   $41 = ($37|0)==(12);
   $42 = ($39|0)<(12);
   $or$cond20 = $42 & $41;
   if ($or$cond20) {
    $43 = (($39) + 12)|0;
    HEAP32[$1>>2] = $43;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE12__get_secondERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,2)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(61);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_weekdayERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,1)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 $12 = ($8|0)<(7);
 $or$cond = $12 & $11;
 if ($or$cond) {
  HEAP32[$1>>2] = $8;
 } else {
  $13 = $9 | 4;
  HEAP32[$4>>2] = $13;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE10__get_yearERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$ = 0, $$0 = 0, $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,4)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = ($8|0)<(69);
  if ($12) {
   $13 = (($8) + 2000)|0;
   $$0 = $13;
  } else {
   $14 = ($8|0)<(100);
   $15 = (($8) + 1900)|0;
   $$ = $14 ? $15 : $8;
   $$0 = $$;
  }
  $16 = (($$0) + -1900)|0;
  HEAP32[$1>>2] = $16;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11__get_year4ERiRS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$byval_copy = 0, $10 = 0, $11 = 0, $12 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $$byval_copy = sp + 4|0;
 $6 = sp;
 $7 = HEAP32[$3>>2]|0;
 HEAP32[$6>>2] = $7;
 ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
 $8 = (__ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($2,$$byval_copy,$4,$5,4)|0);
 $9 = HEAP32[$4>>2]|0;
 $10 = $9 & 4;
 $11 = ($10|0)==(0);
 if ($11) {
  $12 = (($8) + -1900)|0;
  HEAP32[$1>>2] = $12;
 }
 STACKTOP = sp;return;
}
function __ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13__get_percentERS4_S4_RjRKNS_5ctypeIwEE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i11 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i17 = 0, $$pre$i$i$i = 0, $$pre$i$i$i13 = 0, $$pre$i4$i$i1925 = 0, $$sink3 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $phitmp$i$i$i = 0, $phitmp$i$i$i14 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = HEAP32[$1>>2]|0;
 $6 = ($5|0)==(0|0);
 do {
  if ($6) {
   $94 = 1;
  } else {
   $7 = ((($5)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    $12 = HEAP32[$5>>2]|0;
    $13 = ((($12)) + 36|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (FUNCTION_TABLE_ii[$14 & 2047]($5)|0);
    $$0$i$i$i$i = $15;
   } else {
    $16 = HEAP32[$8>>2]|0;
    $17 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($16)|0);
    $$0$i$i$i$i = $17;
   }
   $18 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $19 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$18)|0);
   if ($19) {
    HEAP32[$1>>2] = 0;
    $94 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$1>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $94 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $20 = HEAP32[$2>>2]|0;
 $21 = ($20|0)==(0|0);
 do {
  if ($21) {
   label = 14;
  } else {
   $22 = ((($20)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($20)) + 16|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($23|0)==($25|0);
   if ($26) {
    $27 = HEAP32[$20>>2]|0;
    $28 = ((($27)) + 36|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (FUNCTION_TABLE_ii[$29 & 2047]($20)|0);
    $$0$i$i2$i$i = $30;
   } else {
    $31 = HEAP32[$23>>2]|0;
    $32 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($31)|0);
    $$0$i$i2$i$i = $32;
   }
   $33 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $34 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$33)|0);
   if ($34) {
    HEAP32[$2>>2] = 0;
    label = 14;
    break;
   } else {
    if ($94) {
     $$pre$i4$i$i1925 = $20;
     label = 16;
     break;
    } else {
     $$sink3 = 6;
     label = 38;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($94) {
   $$sink3 = 6;
   label = 38;
  } else {
   $$pre$i4$i$i1925 = 0;
   label = 16;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $35 = HEAP32[$1>>2]|0;
   $36 = ((($35)) + 12|0);
   $37 = HEAP32[$36>>2]|0;
   $38 = ((($35)) + 16|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ($37|0)==($39|0);
   if ($40) {
    $41 = HEAP32[$35>>2]|0;
    $42 = ((($41)) + 36|0);
    $43 = HEAP32[$42>>2]|0;
    $44 = (FUNCTION_TABLE_ii[$43 & 2047]($35)|0);
    $$0$i$i = $44;
   } else {
    $45 = HEAP32[$37>>2]|0;
    $46 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($45)|0);
    $$0$i$i = $46;
   }
   $47 = HEAP32[$4>>2]|0;
   $48 = ((($47)) + 52|0);
   $49 = HEAP32[$48>>2]|0;
   $50 = (FUNCTION_TABLE_iiii[$49 & 2047]($4,$$0$i$i,0)|0);
   $51 = ($50<<24>>24)==(37);
   if ($51) {
    $52 = HEAP32[$1>>2]|0;
    $53 = ((($52)) + 12|0);
    $54 = HEAP32[$53>>2]|0;
    $55 = ((($52)) + 16|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ($54|0)==($56|0);
    if ($57) {
     $58 = HEAP32[$52>>2]|0;
     $59 = ((($58)) + 40|0);
     $60 = HEAP32[$59>>2]|0;
     (FUNCTION_TABLE_ii[$60 & 2047]($52)|0);
    } else {
     $61 = ((($54)) + 4|0);
     HEAP32[$53>>2] = $61;
     $62 = HEAP32[$54>>2]|0;
     (__ZNSt3__211char_traitsIwE11to_int_typeEw($62)|0);
    }
    $63 = HEAP32[$1>>2]|0;
    $64 = ($63|0)==(0|0);
    do {
     if ($64) {
      $95 = 1;
     } else {
      $65 = ((($63)) + 12|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = ((($63)) + 16|0);
      $68 = HEAP32[$67>>2]|0;
      $69 = ($66|0)==($68|0);
      if ($69) {
       $70 = HEAP32[$63>>2]|0;
       $71 = ((($70)) + 36|0);
       $72 = HEAP32[$71>>2]|0;
       $73 = (FUNCTION_TABLE_ii[$72 & 2047]($63)|0);
       $$0$i$i$i$i11 = $73;
      } else {
       $74 = HEAP32[$66>>2]|0;
       $75 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($74)|0);
       $$0$i$i$i$i11 = $75;
      }
      $76 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $77 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i11,$76)|0);
      if ($77) {
       HEAP32[$1>>2] = 0;
       $95 = 1;
       break;
      } else {
       $$pre$i$i$i13 = HEAP32[$1>>2]|0;
       $phitmp$i$i$i14 = ($$pre$i$i$i13|0)==(0|0);
       $95 = $phitmp$i$i$i14;
       break;
      }
     }
    } while(0);
    $78 = ($$pre$i4$i$i1925|0)==(0|0);
    do {
     if (!($78)) {
      $79 = ((($$pre$i4$i$i1925)) + 12|0);
      $80 = HEAP32[$79>>2]|0;
      $81 = ((($$pre$i4$i$i1925)) + 16|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = ($80|0)==($82|0);
      if ($83) {
       $84 = HEAP32[$$pre$i4$i$i1925>>2]|0;
       $85 = ((($84)) + 36|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = (FUNCTION_TABLE_ii[$86 & 2047]($$pre$i4$i$i1925)|0);
       $$0$i$i2$i$i17 = $87;
      } else {
       $88 = HEAP32[$80>>2]|0;
       $89 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($88)|0);
       $$0$i$i2$i$i17 = $89;
      }
      $90 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $91 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i17,$90)|0);
      if ($91) {
       HEAP32[$2>>2] = 0;
       break;
      } else {
       if ($95) {
        break L22;
       } else {
        $$sink3 = 2;
        label = 38;
        break L22;
       }
      }
     }
    } while(0);
    if ($95) {
     $$sink3 = 2;
     label = 38;
    }
   } else {
    $$sink3 = 4;
    label = 38;
   }
  }
 } while(0);
 if ((label|0) == 38) {
  $92 = HEAP32[$3>>2]|0;
  $93 = $92 | $$sink3;
  HEAP32[$3>>2] = $93;
 }
 return;
}
function __ZNSt3__220__get_up_to_n_digitsIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEiRT0_S5_RjRKNS_5ctypeIT_EEi($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i27 = 0, $$0$i$i$i$i42 = 0, $$0$i$i2$i$i = 0, $$0$i$i2$i$i33 = 0, $$0$i$i2$i$i48 = 0, $$0$i$i39 = 0, $$0$in = 0, $$025 = 0, $$025$in = 0, $$2 = 0, $$pre$i$i$i = 0, $$pre$i$i$i29 = 0, $$pre$i$i$i44 = 0, $$pre$i4$i$i35 = 0, $$pre$i4$i$i50 = 0, $10 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $18 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i30 = 0, $phitmp$i$i$i45 = 0, $phitmp$i5$i$i36 = 0, $tmp$i$i38 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5|0)==(0|0);
 do {
  if ($6) {
   $169 = 1;
  } else {
   $7 = ((($5)) + 12|0);
   $8 = HEAP32[$7>>2]|0;
   $9 = ((($5)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    $12 = HEAP32[$5>>2]|0;
    $13 = ((($12)) + 36|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = (FUNCTION_TABLE_ii[$14 & 2047]($5)|0);
    $$0$i$i$i$i = $15;
   } else {
    $16 = HEAP32[$8>>2]|0;
    $17 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($16)|0);
    $$0$i$i$i$i = $17;
   }
   $18 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $19 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$18)|0);
   if ($19) {
    HEAP32[$0>>2] = 0;
    $169 = 1;
    break;
   } else {
    $$pre$i$i$i = HEAP32[$0>>2]|0;
    $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
    $169 = $phitmp$i$i$i;
    break;
   }
  }
 } while(0);
 $20 = HEAP32[$1>>2]|0;
 $21 = ($20|0)==(0|0);
 do {
  if ($21) {
   label = 14;
  } else {
   $22 = ((($20)) + 12|0);
   $23 = HEAP32[$22>>2]|0;
   $24 = ((($20)) + 16|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = ($23|0)==($25|0);
   if ($26) {
    $27 = HEAP32[$20>>2]|0;
    $28 = ((($27)) + 36|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = (FUNCTION_TABLE_ii[$29 & 2047]($20)|0);
    $$0$i$i2$i$i = $30;
   } else {
    $31 = HEAP32[$23>>2]|0;
    $32 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($31)|0);
    $$0$i$i2$i$i = $32;
   }
   $33 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $34 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$33)|0);
   if ($34) {
    HEAP32[$1>>2] = 0;
    label = 14;
    break;
   } else {
    if ($169) {
     $170 = $20;
     label = 17;
     break;
    } else {
     label = 16;
     break;
    }
   }
  }
 } while(0);
 if ((label|0) == 14) {
  if ($169) {
   label = 16;
  } else {
   $170 = 0;
   label = 17;
  }
 }
 L22: do {
  if ((label|0) == 16) {
   $35 = HEAP32[$2>>2]|0;
   $36 = $35 | 6;
   HEAP32[$2>>2] = $36;
   $$2 = 0;
  }
  else if ((label|0) == 17) {
   $37 = HEAP32[$0>>2]|0;
   $38 = ((($37)) + 12|0);
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($37)) + 16|0);
   $41 = HEAP32[$40>>2]|0;
   $42 = ($39|0)==($41|0);
   if ($42) {
    $43 = HEAP32[$37>>2]|0;
    $44 = ((($43)) + 36|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = (FUNCTION_TABLE_ii[$45 & 2047]($37)|0);
    $$0$i$i = $46;
   } else {
    $47 = HEAP32[$39>>2]|0;
    $48 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($47)|0);
    $$0$i$i = $48;
   }
   $49 = HEAP32[$3>>2]|0;
   $50 = ((($49)) + 12|0);
   $51 = HEAP32[$50>>2]|0;
   $52 = (FUNCTION_TABLE_iiii[$51 & 2047]($3,2048,$$0$i$i)|0);
   if (!($52)) {
    $53 = HEAP32[$2>>2]|0;
    $54 = $53 | 4;
    HEAP32[$2>>2] = $54;
    $$2 = 0;
    break;
   }
   $55 = HEAP32[$3>>2]|0;
   $56 = ((($55)) + 52|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = (FUNCTION_TABLE_iiii[$57 & 2047]($3,$$0$i$i,0)|0);
   $59 = $58 << 24 >> 24;
   $60 = HEAP32[$0>>2]|0;
   $61 = ((($60)) + 12|0);
   $62 = HEAP32[$61>>2]|0;
   $63 = ((($60)) + 16|0);
   $64 = HEAP32[$63>>2]|0;
   $65 = ($62|0)==($64|0);
   if ($65) {
    $66 = HEAP32[$60>>2]|0;
    $67 = ((($66)) + 40|0);
    $68 = HEAP32[$67>>2]|0;
    (FUNCTION_TABLE_ii[$68 & 2047]($60)|0);
    $$0$in = $59;$$025$in = $4;$$pre$i4$i$i35 = $170;$87 = $170;
   } else {
    $69 = ((($62)) + 4|0);
    HEAP32[$61>>2] = $69;
    $70 = HEAP32[$62>>2]|0;
    (__ZNSt3__211char_traitsIwE11to_int_typeEw($70)|0);
    $$0$in = $59;$$025$in = $4;$$pre$i4$i$i35 = $170;$87 = $170;
   }
   while(1) {
    $$0 = (($$0$in) + -48)|0;
    $$025 = (($$025$in) + -1)|0;
    $71 = HEAP32[$0>>2]|0;
    $72 = ($71|0)==(0|0);
    do {
     if ($72) {
      $101 = 1;
     } else {
      $73 = ((($71)) + 12|0);
      $74 = HEAP32[$73>>2]|0;
      $75 = ((($71)) + 16|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ($74|0)==($76|0);
      if ($77) {
       $78 = HEAP32[$71>>2]|0;
       $79 = ((($78)) + 36|0);
       $80 = HEAP32[$79>>2]|0;
       $81 = (FUNCTION_TABLE_ii[$80 & 2047]($71)|0);
       $$0$i$i$i$i27 = $81;
      } else {
       $82 = HEAP32[$74>>2]|0;
       $83 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($82)|0);
       $$0$i$i$i$i27 = $83;
      }
      $84 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $85 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i27,$84)|0);
      if ($85) {
       HEAP32[$0>>2] = 0;
       $101 = 1;
       break;
      } else {
       $$pre$i$i$i29 = HEAP32[$0>>2]|0;
       $phitmp$i$i$i30 = ($$pre$i$i$i29|0)==(0|0);
       $101 = $phitmp$i$i$i30;
       break;
      }
     }
    } while(0);
    $86 = ($87|0)==(0|0);
    do {
     if ($86) {
      $$pre$i4$i$i50 = $$pre$i4$i$i35;$102 = 1;$171 = 0;
     } else {
      $88 = ((($87)) + 12|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = ((($87)) + 16|0);
      $91 = HEAP32[$90>>2]|0;
      $92 = ($89|0)==($91|0);
      if ($92) {
       $93 = HEAP32[$87>>2]|0;
       $94 = ((($93)) + 36|0);
       $95 = HEAP32[$94>>2]|0;
       $96 = (FUNCTION_TABLE_ii[$95 & 2047]($87)|0);
       $$0$i$i2$i$i33 = $96;
      } else {
       $97 = HEAP32[$89>>2]|0;
       $98 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($97)|0);
       $$0$i$i2$i$i33 = $98;
      }
      $99 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $100 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i33,$99)|0);
      if ($100) {
       HEAP32[$1>>2] = 0;
       $$pre$i4$i$i50 = 0;$102 = 1;$171 = 0;
       break;
      } else {
       $phitmp$i5$i$i36 = ($$pre$i4$i$i35|0)==(0|0);
       $$pre$i4$i$i50 = $$pre$i4$i$i35;$102 = $phitmp$i5$i$i36;$171 = $$pre$i4$i$i35;
       break;
      }
     }
    } while(0);
    $tmp$i$i38 = $101 ^ $102;
    $103 = ($$025$in|0)>(1);
    $104 = $103 & $tmp$i$i38;
    $105 = HEAP32[$0>>2]|0;
    if (!($104)) {
     break;
    }
    $106 = ((($105)) + 12|0);
    $107 = HEAP32[$106>>2]|0;
    $108 = ((($105)) + 16|0);
    $109 = HEAP32[$108>>2]|0;
    $110 = ($107|0)==($109|0);
    if ($110) {
     $111 = HEAP32[$105>>2]|0;
     $112 = ((($111)) + 36|0);
     $113 = HEAP32[$112>>2]|0;
     $114 = (FUNCTION_TABLE_ii[$113 & 2047]($105)|0);
     $$0$i$i39 = $114;
    } else {
     $115 = HEAP32[$107>>2]|0;
     $116 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($115)|0);
     $$0$i$i39 = $116;
    }
    $117 = HEAP32[$3>>2]|0;
    $118 = ((($117)) + 12|0);
    $119 = HEAP32[$118>>2]|0;
    $120 = (FUNCTION_TABLE_iiii[$119 & 2047]($3,2048,$$0$i$i39)|0);
    if (!($120)) {
     $$2 = $$0;
     break L22;
    }
    $121 = ($$0*10)|0;
    $122 = HEAP32[$3>>2]|0;
    $123 = ((($122)) + 52|0);
    $124 = HEAP32[$123>>2]|0;
    $125 = (FUNCTION_TABLE_iiii[$124 & 2047]($3,$$0$i$i39,0)|0);
    $126 = $125 << 24 >> 24;
    $127 = (($126) + ($121))|0;
    $128 = HEAP32[$0>>2]|0;
    $129 = ((($128)) + 12|0);
    $130 = HEAP32[$129>>2]|0;
    $131 = ((($128)) + 16|0);
    $132 = HEAP32[$131>>2]|0;
    $133 = ($130|0)==($132|0);
    if ($133) {
     $134 = HEAP32[$128>>2]|0;
     $135 = ((($134)) + 40|0);
     $136 = HEAP32[$135>>2]|0;
     (FUNCTION_TABLE_ii[$136 & 2047]($128)|0);
     $$0$in = $127;$$025$in = $$025;$$pre$i4$i$i35 = $$pre$i4$i$i50;$87 = $171;
     continue;
    } else {
     $137 = ((($130)) + 4|0);
     HEAP32[$129>>2] = $137;
     $138 = HEAP32[$130>>2]|0;
     (__ZNSt3__211char_traitsIwE11to_int_typeEw($138)|0);
     $$0$in = $127;$$025$in = $$025;$$pre$i4$i$i35 = $$pre$i4$i$i50;$87 = $171;
     continue;
    }
   }
   $139 = ($105|0)==(0|0);
   do {
    if ($139) {
     $172 = 1;
    } else {
     $140 = ((($105)) + 12|0);
     $141 = HEAP32[$140>>2]|0;
     $142 = ((($105)) + 16|0);
     $143 = HEAP32[$142>>2]|0;
     $144 = ($141|0)==($143|0);
     if ($144) {
      $145 = HEAP32[$105>>2]|0;
      $146 = ((($145)) + 36|0);
      $147 = HEAP32[$146>>2]|0;
      $148 = (FUNCTION_TABLE_ii[$147 & 2047]($105)|0);
      $$0$i$i$i$i42 = $148;
     } else {
      $149 = HEAP32[$141>>2]|0;
      $150 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($149)|0);
      $$0$i$i$i$i42 = $150;
     }
     $151 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $152 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i42,$151)|0);
     if ($152) {
      HEAP32[$0>>2] = 0;
      $172 = 1;
      break;
     } else {
      $$pre$i$i$i44 = HEAP32[$0>>2]|0;
      $phitmp$i$i$i45 = ($$pre$i$i$i44|0)==(0|0);
      $172 = $phitmp$i$i$i45;
      break;
     }
    }
   } while(0);
   $153 = ($$pre$i4$i$i50|0)==(0|0);
   do {
    if ($153) {
     label = 60;
    } else {
     $154 = ((($$pre$i4$i$i50)) + 12|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = ((($$pre$i4$i$i50)) + 16|0);
     $157 = HEAP32[$156>>2]|0;
     $158 = ($155|0)==($157|0);
     if ($158) {
      $159 = HEAP32[$$pre$i4$i$i50>>2]|0;
      $160 = ((($159)) + 36|0);
      $161 = HEAP32[$160>>2]|0;
      $162 = (FUNCTION_TABLE_ii[$161 & 2047]($$pre$i4$i$i50)|0);
      $$0$i$i2$i$i48 = $162;
     } else {
      $163 = HEAP32[$155>>2]|0;
      $164 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($163)|0);
      $$0$i$i2$i$i48 = $164;
     }
     $165 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
     $166 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i48,$165)|0);
     if ($166) {
      HEAP32[$1>>2] = 0;
      label = 60;
      break;
     } else {
      if ($172) {
       $$2 = $$0;
       break L22;
      } else {
       break;
      }
     }
    }
   } while(0);
   if ((label|0) == 60) {
    if (!($172)) {
     $$2 = $$0;
     break;
    }
   }
   $167 = HEAP32[$2>>2]|0;
   $168 = $167 | 2;
   HEAP32[$2>>2] = $168;
   $$2 = $$0;
  }
 } while(0);
 return ($$2|0);
}
function __ZNSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 __ZNSt3__210__time_putD2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 __ZNSt3__210__time_putD2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPK2tmcc($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$sroa$04$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$sroa$04$0$i = 0, $$sroa$04$1$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $7 = sp + 4|0;
 $8 = sp;
 $9 = ((($7)) + 100|0);
 HEAP32[$8>>2] = $9;
 $10 = ((($0)) + 8|0);
 __ZNKSt3__210__time_put8__do_putEPcRS1_PK2tmcc($10,$7,$8,$4,$5,$6);
 $11 = HEAP32[$8>>2]|0;
 $12 = HEAP32[$1>>2]|0;
 $$0$i$i = $7;$$sroa$04$0$i = $12;
 while(1) {
  $13 = ($$0$i$i|0)==($11|0);
  if ($13) {
   break;
  }
  $14 = HEAP8[$$0$i$i>>0]|0;
  $15 = ($$sroa$04$0$i|0)==(0|0);
  if ($15) {
   $$sroa$04$1$i = 0;
  } else {
   $16 = ((($$sroa$04$0$i)) + 24|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($$sroa$04$0$i)) + 28|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($17|0)==($19|0);
   if ($20) {
    $21 = HEAP32[$$sroa$04$0$i>>2]|0;
    $22 = ((($21)) + 52|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($14)|0);
    $25 = (FUNCTION_TABLE_iii[$23 & 2047]($$sroa$04$0$i,$24)|0);
    $$0$i$i$i$i = $25;
   } else {
    $26 = ((($17)) + 1|0);
    HEAP32[$16>>2] = $26;
    HEAP8[$17>>0] = $14;
    $27 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($14)|0);
    $$0$i$i$i$i = $27;
   }
   $28 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
   $29 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$28)|0);
   $$$sroa$04$0$i = $29 ? 0 : $$sroa$04$0$i;
   $$sroa$04$1$i = $$$sroa$04$0$i;
  }
  $30 = ((($$0$i$i)) + 1|0);
  $$0$i$i = $30;$$sroa$04$0$i = $$sroa$04$1$i;
 }
 STACKTOP = sp;return ($$sroa$04$0$i|0);
}
function __ZNKSt3__210__time_put8__do_putEPcRS1_PK2tmcc($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $6 = sp;
 HEAP8[$6>>0] = 37;
 $7 = ((($6)) + 1|0);
 HEAP8[$7>>0] = $4;
 $8 = ((($6)) + 2|0);
 HEAP8[$8>>0] = $5;
 $9 = ((($6)) + 3|0);
 HEAP8[$9>>0] = 0;
 $10 = ($5<<24>>24)==(0);
 if (!($10)) {
  HEAP8[$7>>0] = $5;
  HEAP8[$8>>0] = $4;
 }
 $11 = HEAP32[$2>>2]|0;
 $12 = (__ZNSt3__212_GLOBAL__N_17countofIcEEjPKT_S4_($1,$11)|0);
 $13 = HEAP32[$0>>2]|0;
 $14 = (_strftime_l(($1|0),($12|0),($6|0),($3|0),($13|0))|0);
 $15 = (($1) + ($14)|0);
 HEAP32[$2>>2] = $15;
 STACKTOP = sp;return;
}
function __ZNSt3__212_GLOBAL__N_17countofIcEEjPKT_S4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1;
 $3 = $0;
 $4 = (($2) - ($3))|0;
 return ($4|0);
}
function __ZNSt3__210__time_putD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 __THREW__ = 0;
 $2 = (invoke_i(955)|0);
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 do {
  if (!($4)) {
   $5 = ($1|0)==($2|0);
   if (!($5)) {
    $6 = HEAP32[$0>>2]|0;
    __THREW__ = 0;
    invoke_vi(1000,($6|0));
    $7 = __THREW__; __THREW__ = 0;
    $8 = $7&1;
    if ($8) {
     break;
    }
   }
   return;
  }
 } while(0);
 $9 = ___cxa_find_matching_catch_3(0|0)|0;
 $10 = tempRet0;
 ___clang_call_terminate($9);
 // unreachable;
}
function __ZNSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 __ZNSt3__210__time_putD2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 __ZNSt3__210__time_putD2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPK2tmcc($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$$sroa$04$0$i = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$sroa$04$0$i = 0, $$sroa$04$1$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 416|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(416|0);
 $7 = sp + 8|0;
 $8 = sp;
 $9 = ((($7)) + 400|0);
 HEAP32[$8>>2] = $9;
 $10 = ((($0)) + 8|0);
 __ZNKSt3__210__time_put8__do_putEPwRS1_PK2tmcc($10,$7,$8,$4,$5,$6);
 $11 = HEAP32[$8>>2]|0;
 $12 = HEAP32[$1>>2]|0;
 $$0$i$i = $7;$$sroa$04$0$i = $12;
 while(1) {
  $13 = ($$0$i$i|0)==($11|0);
  if ($13) {
   break;
  }
  $14 = HEAP32[$$0$i$i>>2]|0;
  $15 = ($$sroa$04$0$i|0)==(0|0);
  if ($15) {
   $$sroa$04$1$i = 0;
  } else {
   $16 = ((($$sroa$04$0$i)) + 24|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ((($$sroa$04$0$i)) + 28|0);
   $19 = HEAP32[$18>>2]|0;
   $20 = ($17|0)==($19|0);
   if ($20) {
    $21 = HEAP32[$$sroa$04$0$i>>2]|0;
    $22 = ((($21)) + 52|0);
    $23 = HEAP32[$22>>2]|0;
    $24 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($14)|0);
    $25 = (FUNCTION_TABLE_iii[$23 & 2047]($$sroa$04$0$i,$24)|0);
    $$0$i$i$i$i = $25;
   } else {
    $26 = ((($17)) + 4|0);
    HEAP32[$16>>2] = $26;
    HEAP32[$17>>2] = $14;
    $27 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($14)|0);
    $$0$i$i$i$i = $27;
   }
   $28 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
   $29 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$28)|0);
   $$$sroa$04$0$i = $29 ? 0 : $$sroa$04$0$i;
   $$sroa$04$1$i = $$$sroa$04$0$i;
  }
  $30 = ((($$0$i$i)) + 4|0);
  $$0$i$i = $30;$$sroa$04$0$i = $$sroa$04$1$i;
 }
 STACKTOP = sp;return ($$sroa$04$0$i|0);
}
function __ZNKSt3__210__time_put8__do_putEPwRS1_PK2tmcc($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $6 = sp + 16|0;
 $7 = sp + 12|0;
 $8 = sp;
 $9 = sp + 8|0;
 $10 = ((($6)) + 100|0);
 HEAP32[$7>>2] = $10;
 __ZNKSt3__210__time_put8__do_putEPcRS1_PK2tmcc($0,$6,$7,$3,$4,$5);
 $11 = $8;
 $12 = $11;
 HEAP32[$12>>2] = 0;
 $13 = (($11) + 4)|0;
 $14 = $13;
 HEAP32[$14>>2] = 0;
 HEAP32[$9>>2] = $6;
 $15 = HEAP32[$2>>2]|0;
 $16 = (__ZNSt3__212_GLOBAL__N_17countofIwEEjPKT_S4_($1,$15)|0);
 $17 = HEAP32[$0>>2]|0;
 $18 = (___uselocale($17)|0);
 __THREW__ = 0;
 $19 = (invoke_iiiii(1001,($1|0),($9|0),($16|0),($8|0))|0);
 $20 = __THREW__; __THREW__ = 0;
 $21 = $20&1;
 if ($21) {
  $27 = ___cxa_find_matching_catch_2()|0;
  $28 = tempRet0;
  $29 = ($18|0)==(0|0);
  if (!($29)) {
   __THREW__ = 0;
   (invoke_ii(958,($18|0))|0);
   $30 = __THREW__; __THREW__ = 0;
   $31 = $30&1;
   if ($31) {
    $32 = ___cxa_find_matching_catch_3(0|0)|0;
    $33 = tempRet0;
    ___clang_call_terminate($32);
    // unreachable;
   }
  }
  ___resumeException($27|0);
  // unreachable;
 }
 $22 = ($18|0)==(0|0);
 if (!($22)) {
  __THREW__ = 0;
  (invoke_ii(958,($18|0))|0);
  $23 = __THREW__; __THREW__ = 0;
  $24 = $23&1;
  if ($24) {
   $25 = ___cxa_find_matching_catch_3(0|0)|0;
   $26 = tempRet0;
   ___clang_call_terminate($25);
   // unreachable;
  }
 }
 $34 = ($19|0)==(-1);
 if ($34) {
  __ZNSt3__221__throw_runtime_errorEPKc(156921);
  // unreachable;
 } else {
  $35 = (($1) + ($19<<2)|0);
  HEAP32[$2>>2] = $35;
  STACKTOP = sp;return;
 }
}
function __ZNSt3__212_GLOBAL__N_17countofIwEEjPKT_S4_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1;
 $3 = $0;
 $4 = (($2) - ($3))|0;
 $5 = $4 >> 2;
 return ($5|0);
}
function __ZNSt3__221__throw_runtime_errorEPKc($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(1002,($1|0),($0|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0));
  ___resumeException($4|0);
  // unreachable;
 } else {
  ___cxa_throw(($1|0),(2512|0),(716|0));
  // unreachable;
 }
}
function __ZNSt3__210moneypunctIcLb0EED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__210moneypunctIcLb0EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__210moneypunctIcLb0EE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__210moneypunctIcLb0EE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__210moneypunctIcLb0EE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb0EE14do_curr_symbolEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb0EE16do_positive_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb0EE16do_negative_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 1;
 (__ZNSt3__211char_traitsIcE6assignEPcjc($0,1,45)|0);
 $4 = ((($0)) + 1|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__210moneypunctIcLb0EE14do_frac_digitsEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__210moneypunctIcLb0EE13do_pos_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__210moneypunctIcLb0EE13do_neg_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__210moneypunctIcLb1EED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__210moneypunctIcLb1EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__210moneypunctIcLb1EE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__210moneypunctIcLb1EE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 127;
}
function __ZNKSt3__210moneypunctIcLb1EE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb1EE14do_curr_symbolEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb1EE16do_positive_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIcLb1EE16do_negative_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 1;
 (__ZNSt3__211char_traitsIcE6assignEPcjc($0,1,45)|0);
 $4 = ((($0)) + 1|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__210moneypunctIcLb1EE14do_frac_digitsEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__210moneypunctIcLb1EE13do_pos_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__210moneypunctIcLb1EE13do_neg_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__210moneypunctIwLb0EED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__210moneypunctIwLb0EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__210moneypunctIwLb0EE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__210moneypunctIwLb0EE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__210moneypunctIwLb0EE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb0EE14do_curr_symbolEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb0EE16do_positive_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb0EE16do_negative_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 8|0);
 $4 = ((($3)) + 3|0);
 HEAP8[$4>>0] = 1;
 (__ZNSt3__211char_traitsIwE6assignEPwjw($0,1,45)|0);
 $5 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($5,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__210moneypunctIwLb0EE14do_frac_digitsEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__210moneypunctIwLb0EE13do_pos_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__210moneypunctIwLb0EE13do_neg_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__210moneypunctIwLb1EED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__210moneypunctIwLb1EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__210moneypunctIwLb1EE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__210moneypunctIwLb1EE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 2147483647;
}
function __ZNKSt3__210moneypunctIwLb1EE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb1EE14do_curr_symbolEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb1EE16do_positive_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 return;
}
function __ZNKSt3__210moneypunctIwLb1EE16do_negative_signEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 8|0);
 $4 = ((($3)) + 3|0);
 HEAP8[$4>>0] = 1;
 (__ZNSt3__211char_traitsIwE6assignEPwjw($0,1,45)|0);
 $5 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($5,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__210moneypunctIwLb1EE14do_frac_digitsEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__210moneypunctIwLb1EE13do_pos_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNKSt3__210moneypunctIwLb1EE13do_neg_formatEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP8[$0>>0]=67109634&255;HEAP8[$0+1>>0]=(67109634>>8)&255;HEAP8[$0+2>>0]=(67109634>>16)&255;HEAP8[$0+3>>0]=67109634>>24;
 return;
}
function __ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i = 0, $$0$i$i$i$i = 0, $$0$i$i2$i$i = 0, $$0$lcssa$i = 0, $$031 = 0, $$032 = 0, $$035 = 0, $$1 = 0, $$133 = 0, $$136 = 0, $$2 = 0, $$234 = 0, $$237 = 0, $$33860 = 0, $$33861 = 0, $$362 = 0, $$363 = 0, $$byval_copy = 0, $$pre = 0;
 var $$pre$i$i$i = 0, $$pre57 = 0, $$pre64 = 0, $$sroa$0$0$copyload = 0, $$sroa$044$0 = 0, $$sroa$044$1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $$byval_copy = sp + 24|0;
 $vararg_buffer = sp;
 $7 = sp + 140|0;
 $8 = sp + 16|0;
 $9 = sp + 12|0;
 $10 = sp + 8|0;
 $11 = sp + 138|0;
 $12 = sp + 4|0;
 $13 = sp + 128|0;
 $14 = sp + 28|0;
 $15 = $7;
 HEAP32[$8>>2] = $15;
 $16 = ((($8)) + 4|0);
 HEAP32[$16>>2] = (1003);
 $17 = ((($7)) + 100|0);
 __THREW__ = 0;
 invoke_vii(733,($10|0),($4|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  HEAP32[$8>>2] = 0;
  $$33860 = $46;$$362 = $45;$124 = 1003;$125 = $7;
  label = 54;
 } else {
  __THREW__ = 0;
  $20 = (invoke_iii(734,($10|0),(165440|0))|0);
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $47 = ___cxa_find_matching_catch_2()|0;
   $48 = tempRet0;
   $$234 = $47;$$237 = $48;
  } else {
   HEAP8[$11>>0] = 0;
   $23 = HEAP32[$2>>2]|0;
   HEAP32[$12>>2] = $23;
   $24 = ((($4)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   __THREW__ = 0;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
   $26 = (invoke_iiiiiiiiiiii(1004,($1|0),($$byval_copy|0),($3|0),($10|0),($25|0),($5|0),($11|0),($20|0),($8|0),($9|0),($17|0))|0);
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   L7: do {
    if ($28) {
     label = 11;
    } else {
     L9: do {
      if ($26) {
       $29 = HEAP32[$20>>2]|0;
       $30 = ((($29)) + 32|0);
       $31 = HEAP32[$30>>2]|0;
       __THREW__ = 0;
       (invoke_iiiii($31|0,($20|0),(157144|0),((157154)|0),($13|0))|0);
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        $51 = ___cxa_find_matching_catch_2()|0;
        $52 = tempRet0;
        $$032 = $51;$$035 = $52;
       } else {
        $34 = HEAP32[$9>>2]|0;
        $35 = HEAP32[$8>>2]|0;
        $36 = (($34) - ($35))|0;
        $37 = ($36|0)>(98);
        $38 = $35;
        $39 = $34;
        if ($37) {
         $40 = (($36) + 2)|0;
         $41 = (_malloc($40)|0);
         $42 = $41;
         $43 = ($41|0)==(0|0);
         if ($43) {
          __THREW__ = 0;
          invoke_v(732);
          $44 = __THREW__; __THREW__ = 0;
          $$sroa$044$0 = $42;
         } else {
          $$031 = $41;$$sroa$044$1 = $42;
          label = 16;
         }
        } else {
         $$031 = $14;$$sroa$044$1 = 0;
         label = 16;
        }
        do {
         if ((label|0) == 16) {
          $57 = HEAP8[$11>>0]|0;
          $58 = ($57<<24>>24)==(0);
          if ($58) {
           $$1 = $$031;
          } else {
           $59 = ((($$031)) + 1|0);
           HEAP8[$$031>>0] = 45;
           $$1 = $59;
          }
          $60 = ((($13)) + 10|0);
          $61 = $13;
          $$0 = $38;$$2 = $$1;$63 = $39;
          while(1) {
           $62 = ($$0>>>0)<($63>>>0);
           if (!($62)) {
            break;
           }
           $66 = HEAP8[$$0>>0]|0;
           $$0$i = $13;
           while(1) {
            $67 = ($$0$i|0)==($60|0);
            if ($67) {
             $$0$lcssa$i = $60;
             break;
            }
            $68 = HEAP8[$$0$i>>0]|0;
            $69 = ($68<<24>>24)==($66<<24>>24);
            if ($69) {
             $$0$lcssa$i = $$0$i;
             break;
            }
            $70 = ((($$0$i)) + 1|0);
            $$0$i = $70;
           }
           $71 = $$0$lcssa$i;
           $72 = (($71) - ($61))|0;
           $73 = (157144 + ($72)|0);
           $74 = HEAP8[$73>>0]|0;
           HEAP8[$$2>>0] = $74;
           $75 = ((($$0)) + 1|0);
           $76 = ((($$2)) + 1|0);
           $$pre = HEAP32[$9>>2]|0;
           $$0 = $75;$$2 = $76;$63 = $$pre;
          }
          HEAP8[$$2>>0] = 0;
          HEAP32[$vararg_buffer>>2] = $6;
          $64 = (_sscanf($14,157155,$vararg_buffer)|0);
          $65 = ($64|0)==(1);
          if (!($65)) {
           __THREW__ = 0;
           invoke_vi(1005,(157159|0));
           $77 = __THREW__; __THREW__ = 0;
           $$sroa$044$0 = $$sroa$044$1;
           break;
          }
          $78 = ($$sroa$044$1|0)==(0);
          if (!($78)) {
           $79 = $$sroa$044$1;
           _free($79);
          }
          break L9;
         }
        } while(0);
        $53 = ___cxa_find_matching_catch_2()|0;
        $54 = tempRet0;
        $55 = ($$sroa$044$0|0)==(0);
        if (!($55)) {
         $56 = $$sroa$044$0;
         _free($56);
        }
        $$032 = $53;$$035 = $54;
       }
       $$133 = $$032;$$136 = $$035;
       break L7;
      }
     } while(0);
     $80 = HEAP32[$1>>2]|0;
     $81 = ($80|0)==(0|0);
     do {
      if ($81) {
       $130 = 1;
      } else {
       $82 = ((($80)) + 12|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = ((($80)) + 16|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ($83|0)==($85|0);
       if ($86) {
        $87 = HEAP32[$80>>2]|0;
        $88 = ((($87)) + 36|0);
        $89 = HEAP32[$88>>2]|0;
        __THREW__ = 0;
        $90 = (invoke_ii($89|0,($80|0))|0);
        $91 = __THREW__; __THREW__ = 0;
        $92 = $91&1;
        if ($92) {
         label = 11;
         break L7;
        } else {
         $$0$i$i$i$i = $90;
        }
       } else {
        $93 = HEAP8[$83>>0]|0;
        $94 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($93)|0);
        $$0$i$i$i$i = $94;
       }
       $95 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $96 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$95)|0);
       if ($96) {
        HEAP32[$1>>2] = 0;
        $130 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$1>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $130 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $97 = HEAP32[$2>>2]|0;
     $98 = ($97|0)==(0|0);
     do {
      if ($98) {
       label = 44;
      } else {
       $99 = ((($97)) + 12|0);
       $100 = HEAP32[$99>>2]|0;
       $101 = ((($97)) + 16|0);
       $102 = HEAP32[$101>>2]|0;
       $103 = ($100|0)==($102|0);
       if ($103) {
        $104 = HEAP32[$97>>2]|0;
        $105 = ((($104)) + 36|0);
        $106 = HEAP32[$105>>2]|0;
        __THREW__ = 0;
        $107 = (invoke_ii($106|0,($97|0))|0);
        $108 = __THREW__; __THREW__ = 0;
        $109 = $108&1;
        if ($109) {
         label = 11;
         break L7;
        } else {
         $$0$i$i2$i$i = $107;
        }
       } else {
        $110 = HEAP8[$100>>0]|0;
        $111 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($110)|0);
        $$0$i$i2$i$i = $111;
       }
       $112 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $113 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$112)|0);
       if ($113) {
        HEAP32[$2>>2] = 0;
        label = 44;
        break;
       } else {
        if ($130) {
         break;
        } else {
         label = 46;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 44) {
      if ($130) {
       label = 46;
      }
     }
     if ((label|0) == 46) {
      $114 = HEAP32[$5>>2]|0;
      $115 = $114 | 2;
      HEAP32[$5>>2] = $115;
     }
     $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
     __ZNSt3__26localeD2Ev($10);
     $116 = HEAP32[$8>>2]|0;
     HEAP32[$8>>2] = 0;
     $117 = ($116|0)==(0|0);
     if (!($117)) {
      $118 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      invoke_vi($118|0,($116|0));
      $119 = __THREW__; __THREW__ = 0;
      $120 = $119&1;
      if ($120) {
       $121 = ___cxa_find_matching_catch_3(0|0)|0;
       $122 = tempRet0;
       ___clang_call_terminate($121);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($$sroa$0$0$copyload|0);
    }
   } while(0);
   if ((label|0) == 11) {
    $49 = ___cxa_find_matching_catch_2()|0;
    $50 = tempRet0;
    $$133 = $49;$$136 = $50;
   }
   $$234 = $$133;$$237 = $$136;
  }
  __ZNSt3__26localeD2Ev($10);
  $$pre57 = HEAP32[$8>>2]|0;
  HEAP32[$8>>2] = 0;
  $123 = ($$pre57|0)==(0|0);
  if ($123) {
   $$33861 = $$237;$$363 = $$234;
  } else {
   $$pre64 = HEAP32[$16>>2]|0;
   $$33860 = $$237;$$362 = $$234;$124 = $$pre64;$125 = $$pre57;
   label = 54;
  }
 }
 if ((label|0) == 54) {
  __THREW__ = 0;
  invoke_vi($124|0,($125|0));
  $126 = __THREW__; __THREW__ = 0;
  $127 = $126&1;
  if ($127) {
   $128 = ___cxa_find_matching_catch_3(0|0)|0;
   $129 = tempRet0;
   ___clang_call_terminate($128);
   // unreachable;
  } else {
   $$33861 = $$33860;$$363 = $$362;
  }
 }
 ___resumeException($$363|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIcS3_NS_9allocatorIcEEEE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i$i$i$i = 0, $$0$i$i2$i$i = 0, $$1 = 0, $$132 = 0, $$23343 = 0, $$23344 = 0, $$245 = 0, $$246 = 0, $$byval_copy = 0, $$pre = 0, $$pre$i$i$i = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $$byval_copy = sp + 24|0;
 $7 = sp + 133|0;
 $8 = sp + 132|0;
 $9 = sp + 32|0;
 $10 = sp + 16|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp + 28|0;
 $14 = sp;
 $15 = $9;
 HEAP32[$10>>2] = $15;
 $16 = ((($10)) + 4|0);
 HEAP32[$16>>2] = (1003);
 $17 = ((($9)) + 100|0);
 __THREW__ = 0;
 invoke_vii(733,($12|0),($4|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $45 = ___cxa_find_matching_catch_2()|0;
  $46 = tempRet0;
  HEAP32[$10>>2] = 0;
  $$23343 = $46;$$245 = $45;$111 = $9;
  label = 41;
 } else {
  __THREW__ = 0;
  $20 = (invoke_iii(734,($12|0),(165440|0))|0);
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $47 = ___cxa_find_matching_catch_2()|0;
   $48 = tempRet0;
   $$1 = $47;$$132 = $48;
  } else {
   HEAP8[$13>>0] = 0;
   $23 = HEAP32[$2>>2]|0;
   HEAP32[$14>>2] = $23;
   $24 = ((($4)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $23;
   __THREW__ = 0;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$14>>2]|0;
   $27 = (invoke_iiiiiiiiiiii(1004,($1|0),($$byval_copy|0),($3|0),($12|0),($25|0),($5|0),($13|0),($20|0),($10|0),($11|0),($17|0))|0);
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   L7: do {
    if (!($29)) {
     if ($27) {
      $30 = ((($6)) + 11|0);
      $31 = HEAP8[$30>>0]|0;
      $32 = ($31<<24>>24)<(0);
      if ($32) {
       $33 = HEAP32[$6>>2]|0;
       HEAP8[$7>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($33,$7);
       $34 = ((($6)) + 4|0);
       HEAP32[$34>>2] = 0;
      } else {
       HEAP8[$8>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($6,$8);
       HEAP8[$30>>0] = 0;
      }
      $35 = HEAP8[$13>>0]|0;
      $36 = ($35<<24>>24)==(0);
      if (!($36)) {
       $37 = HEAP32[$20>>2]|0;
       $38 = ((($37)) + 28|0);
       $39 = HEAP32[$38>>2]|0;
       __THREW__ = 0;
       $40 = (invoke_iii($39|0,($20|0),45)|0);
       $41 = __THREW__; __THREW__ = 0;
       $42 = $41&1;
       if ($42) {
        break;
       }
       __THREW__ = 0;
       invoke_vii(1006,($6|0),($40|0));
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        break;
       }
      }
      $49 = HEAP32[$20>>2]|0;
      $50 = ((($49)) + 28|0);
      $51 = HEAP32[$50>>2]|0;
      __THREW__ = 0;
      $52 = (invoke_iii($51|0,($20|0),48)|0);
      $53 = __THREW__; __THREW__ = 0;
      $54 = $53&1;
      if ($54) {
       break;
      }
      $55 = HEAP32[$10>>2]|0;
      $56 = HEAP32[$11>>2]|0;
      $57 = ((($56)) + -1|0);
      $$0 = $55;
      while(1) {
       $58 = ($$0>>>0)<($57>>>0);
       if (!($58)) {
        break;
       }
       $59 = HEAP8[$$0>>0]|0;
       $60 = ($59<<24>>24)==($52<<24>>24);
       if (!($60)) {
        break;
       }
       $61 = ((($$0)) + 1|0);
       $$0 = $61;
      }
      __THREW__ = 0;
      (invoke_iiii(1007,($6|0),($$0|0),($56|0))|0);
      $62 = __THREW__; __THREW__ = 0;
      $63 = $62&1;
      if ($63) {
       break;
      }
     }
     $64 = HEAP32[$1>>2]|0;
     $65 = ($64|0)==(0|0);
     do {
      if ($65) {
       $116 = 1;
      } else {
       $66 = ((($64)) + 12|0);
       $67 = HEAP32[$66>>2]|0;
       $68 = ((($64)) + 16|0);
       $69 = HEAP32[$68>>2]|0;
       $70 = ($67|0)==($69|0);
       if ($70) {
        $71 = HEAP32[$64>>2]|0;
        $72 = ((($71)) + 36|0);
        $73 = HEAP32[$72>>2]|0;
        __THREW__ = 0;
        $74 = (invoke_ii($73|0,($64|0))|0);
        $75 = __THREW__; __THREW__ = 0;
        $76 = $75&1;
        if ($76) {
         break L7;
        } else {
         $$0$i$i$i$i = $74;
        }
       } else {
        $77 = HEAP8[$67>>0]|0;
        $78 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($77)|0);
        $$0$i$i$i$i = $78;
       }
       $79 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $80 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$79)|0);
       if ($80) {
        HEAP32[$1>>2] = 0;
        $116 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$1>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $116 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $81 = ($23|0)==(0);
     do {
      if ($81) {
       label = 32;
      } else {
       $82 = ((($26)) + 12|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = ((($26)) + 16|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ($83|0)==($85|0);
       if ($86) {
        $87 = $23;
        $88 = HEAP32[$87>>2]|0;
        $89 = ((($88)) + 36|0);
        $90 = HEAP32[$89>>2]|0;
        __THREW__ = 0;
        $91 = (invoke_ii($90|0,($26|0))|0);
        $92 = __THREW__; __THREW__ = 0;
        $93 = $92&1;
        if ($93) {
         break L7;
        } else {
         $$0$i$i2$i$i = $91;
        }
       } else {
        $94 = HEAP8[$83>>0]|0;
        $95 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($94)|0);
        $$0$i$i2$i$i = $95;
       }
       $96 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
       $97 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$96)|0);
       if ($97) {
        HEAP32[$2>>2] = 0;
        label = 32;
        break;
       } else {
        if ($116) {
         break;
        } else {
         label = 34;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 32) {
      if ($116) {
       label = 34;
      }
     }
     if ((label|0) == 34) {
      $98 = HEAP32[$5>>2]|0;
      $99 = $98 | 2;
      HEAP32[$5>>2] = $99;
     }
     $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
     __ZNSt3__26localeD2Ev($12);
     $100 = HEAP32[$10>>2]|0;
     HEAP32[$10>>2] = 0;
     $101 = ($100|0)==(0|0);
     if (!($101)) {
      $102 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      invoke_vi($102|0,($100|0));
      $103 = __THREW__; __THREW__ = 0;
      $104 = $103&1;
      if ($104) {
       $105 = ___cxa_find_matching_catch_3(0|0)|0;
       $106 = tempRet0;
       ___clang_call_terminate($105);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($$sroa$0$0$copyload|0);
    }
   } while(0);
   $107 = ___cxa_find_matching_catch_2()|0;
   $108 = tempRet0;
   $$1 = $107;$$132 = $108;
  }
  __ZNSt3__26localeD2Ev($12);
  $$pre = HEAP32[$10>>2]|0;
  HEAP32[$10>>2] = 0;
  $109 = ($$pre|0)==(0|0);
  if ($109) {
   $$23344 = $$132;$$246 = $$1;
  } else {
   $$23343 = $$132;$$245 = $$1;$111 = $$pre;
   label = 41;
  }
 }
 if ((label|0) == 41) {
  $110 = HEAP32[$16>>2]|0;
  __THREW__ = 0;
  invoke_vi($110|0,($111|0));
  $112 = __THREW__; __THREW__ = 0;
  $113 = $112&1;
  if ($113) {
   $114 = ___cxa_find_matching_catch_3(0|0)|0;
   $115 = tempRet0;
   ___clang_call_terminate($114);
   // unreachable;
  } else {
   $$23344 = $$23343;$$246 = $$245;
  }
 }
 ___resumeException($$246|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__212__do_nothingEPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIcEERNS_10unique_ptrIcPFvPvEEERPcSM_($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 var $$$0149 = 0, $$$0149159 = 0, $$$0149160 = 0, $$$0149161 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i186 = 0, $$0$i$i$i$i252 = 0, $$0$i$i$i$i273 = 0, $$0$i$i$i$i295 = 0, $$0$i$i$i$i314 = 0, $$0$i$i$i$i339 = 0, $$0$i$i168 = 0, $$0$i$i171 = 0, $$0$i$i174 = 0, $$0$i$i177 = 0, $$0$i$i182 = 0, $$0$i$i184 = 0, $$0$i$i2$i$i = 0;
 var $$0$i$i2$i$i192 = 0, $$0$i$i2$i$i258 = 0, $$0$i$i2$i$i279 = 0, $$0$i$i2$i$i301 = 0, $$0$i$i2$i$i320 = 0, $$0$i$i2$i$i345 = 0, $$0$i$i201 = 0, $$0$i$i205 = 0, $$0$i$i212 = 0, $$0$i$i217 = 0, $$0$i$i224 = 0, $$0$i$i231 = 0, $$0$i$i249 = 0, $$0$i$i267 = 0, $$0$i$i288 = 0, $$0$i$i309 = 0, $$0$i$i329 = 0, $$0$i$i333 = 0, $$0132 = 0, $$0149 = 0;
 var $$0158 = 0, $$10 = 0, $$1150 = 0, $$2134$ph = 0, $$5142 = 0, $$5148 = 0, $$lcssa = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i188 = 0, $$pre$i$i$i254 = 0, $$pre$i$i$i275 = 0, $$pre$i$i$i297 = 0, $$pre$i$i$i316 = 0, $$pre$i$i$i341 = 0, $$pre$i4$i$i194 = 0, $$pre$i4$i$i194512 = 0, $$pre$i4$i$i194514 = 0, $$pre$i4$i$i260 = 0, $$pre$i4$i$i260509 = 0;
 var $$pre$i4$i$i260510 = 0, $$pre$i4$i$i281 = 0, $$pre$i4$i$i281498 = 0, $$pre$i4$i$i281500 = 0, $$pre$i4$i$i303 = 0, $$pre$i4$i$i322 = 0, $$pre$i4$i$i322505$ph = 0, $$pre$i4$i$i322505518 = 0, $$pre502 = 0, $$pre504 = 0, $$pre507 = 0, $$pre508 = 0, $$sink = 0, $$sroa$0360$0 = 0, $$sroa$0381$0 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0;
 var $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0;
 var $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0;
 var $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0;
 var $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0;
 var $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0;
 var $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0;
 var $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0;
 var $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0;
 var $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0;
 var $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0;
 var $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0;
 var $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0;
 var $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0;
 var $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0;
 var $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0;
 var $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0;
 var $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0;
 var $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0;
 var $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0;
 var $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0;
 var $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0;
 var $596 = 0, $597 = 0, $598 = 0, $599 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0;
 var $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0;
 var $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0;
 var $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0;
 var $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0;
 var $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0;
 var $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0;
 var $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0;
 var $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0;
 var $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0;
 var $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0;
 var $796 = 0, $797 = 0, $798 = 0, $799 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0;
 var $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0;
 var $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0;
 var $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0;
 var $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0;
 var $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0;
 var $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $exitcond$i$i = 0, $exitcond$i$i169 = 0, $exitcond$i$i172 = 0, $exitcond$i$i175 = 0;
 var $exitcond$i$i178 = 0, $or$cond = 0, $or$cond13 = 0, $or$cond15 = 0, $or$cond164 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i189 = 0, $phitmp$i$i$i255 = 0, $phitmp$i$i$i276 = 0, $phitmp$i$i$i298 = 0, $phitmp$i$i$i317 = 0, $phitmp$i$i$i342 = 0, $phitmp$i5$i$i195 = 0, $phitmp$i5$i$i261 = 0, $phitmp$i5$i$i282 = 0, $phitmp$i5$i$i323 = 0, $sext = 0, $sext421 = 0, $sext422 = 0, $sext423 = 0;
 var $switch = 0, $tmp$i$i197 = 0, $tmp$i$i263 = 0, $tmp$i$i284 = 0, $tmp$i$i325 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(512|0);
 $11 = sp + 488|0;
 $12 = sp + 88|0;
 $13 = sp + 80|0;
 $14 = sp + 72|0;
 $15 = sp + 68|0;
 $16 = sp + 496|0;
 $17 = sp + 493|0;
 $18 = sp + 492|0;
 $19 = sp + 56|0;
 $20 = sp + 44|0;
 $21 = sp + 32|0;
 $22 = sp + 20|0;
 $23 = sp + 8|0;
 $24 = sp + 4|0;
 $25 = sp;
 HEAP32[$11>>2] = $10;
 $26 = $12;
 HEAP32[$13>>2] = $26;
 $27 = ((($13)) + 4|0);
 HEAP32[$27>>2] = (1003);
 HEAP32[$14>>2] = $12;
 $28 = ((($12)) + 400|0);
 HEAP32[$15>>2] = $28;
 ;HEAP32[$19>>2]=0|0;HEAP32[$19+4>>2]=0|0;HEAP32[$19+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $29 = (($19) + ($$0$i$i<<2)|0);
  HEAP32[$29>>2] = 0;
  $30 = (($$0$i$i) + 1)|0;
  $$0$i$i = $30;
 }
 ;HEAP32[$20>>2]=0|0;HEAP32[$20+4>>2]=0|0;HEAP32[$20+8>>2]=0|0;
 $$0$i$i168 = 0;
 while(1) {
  $exitcond$i$i169 = ($$0$i$i168|0)==(3);
  if ($exitcond$i$i169) {
   break;
  }
  $31 = (($20) + ($$0$i$i168<<2)|0);
  HEAP32[$31>>2] = 0;
  $32 = (($$0$i$i168) + 1)|0;
  $$0$i$i168 = $32;
 }
 ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;
 $$0$i$i171 = 0;
 while(1) {
  $exitcond$i$i172 = ($$0$i$i171|0)==(3);
  if ($exitcond$i$i172) {
   break;
  }
  $33 = (($21) + ($$0$i$i171<<2)|0);
  HEAP32[$33>>2] = 0;
  $34 = (($$0$i$i171) + 1)|0;
  $$0$i$i171 = $34;
 }
 ;HEAP32[$22>>2]=0|0;HEAP32[$22+4>>2]=0|0;HEAP32[$22+8>>2]=0|0;
 $$0$i$i174 = 0;
 while(1) {
  $exitcond$i$i175 = ($$0$i$i174|0)==(3);
  if ($exitcond$i$i175) {
   break;
  }
  $35 = (($22) + ($$0$i$i174<<2)|0);
  HEAP32[$35>>2] = 0;
  $36 = (($$0$i$i174) + 1)|0;
  $$0$i$i174 = $36;
 }
 ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP32[$23+8>>2]=0|0;
 $$0$i$i177 = 0;
 while(1) {
  $exitcond$i$i178 = ($$0$i$i177|0)==(3);
  if ($exitcond$i$i178) {
   break;
  }
  $37 = (($23) + ($$0$i$i177<<2)|0);
  HEAP32[$37>>2] = 0;
  $38 = (($$0$i$i177) + 1)|0;
  $$0$i$i177 = $38;
 }
 __THREW__ = 0;
 invoke_viiiiiiiiii(1008,($2|0),($3|0),($16|0),($17|0),($18|0),($19|0),($20|0),($21|0),($22|0),($24|0));
 $39 = __THREW__; __THREW__ = 0;
 $40 = $39&1;
 L21: do {
  if ($40) {
   $92 = ___cxa_find_matching_catch_2()|0;
   $93 = tempRet0;
   $$5142 = $92;$$5148 = $93;
  } else {
   $41 = HEAP32[$8>>2]|0;
   HEAP32[$9>>2] = $41;
   $42 = ((($7)) + 8|0);
   $43 = ((($21)) + 11|0);
   $44 = ((($21)) + 4|0);
   $45 = ((($22)) + 11|0);
   $46 = ((($22)) + 4|0);
   $47 = $4 & 512;
   $48 = ($47|0)!=(0);
   $49 = ((($20)) + 11|0);
   $50 = ((($20)) + 4|0);
   $51 = ((($23)) + 11|0);
   $52 = ((($23)) + 4|0);
   $53 = ((($16)) + 3|0);
   $54 = ((($19)) + 11|0);
   $55 = ((($19)) + 4|0);
   $$0149 = 0;$$0158 = 0;
   L24: while(1) {
    $56 = ($$0158>>>0)<(4);
    $57 = ($$0149|0)!=(0|0);
    if (!($56)) {
     label = 246;
     break;
    }
    $58 = HEAP32[$0>>2]|0;
    $59 = ($58|0)==(0|0);
    do {
     if ($59) {
      $902 = 1;
     } else {
      $60 = ((($58)) + 12|0);
      $61 = HEAP32[$60>>2]|0;
      $62 = ((($58)) + 16|0);
      $63 = HEAP32[$62>>2]|0;
      $64 = ($61|0)==($63|0);
      if ($64) {
       $65 = HEAP32[$58>>2]|0;
       $66 = ((($65)) + 36|0);
       $67 = HEAP32[$66>>2]|0;
       __THREW__ = 0;
       $68 = (invoke_ii($67|0,($58|0))|0);
       $69 = __THREW__; __THREW__ = 0;
       $70 = $69&1;
       if ($70) {
        label = 35;
        break L24;
       } else {
        $$0$i$i$i$i = $68;
       }
      } else {
       $71 = HEAP8[$61>>0]|0;
       $72 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($71)|0);
       $$0$i$i$i$i = $72;
      }
      $73 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $74 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i,$73)|0);
      if ($74) {
       HEAP32[$0>>2] = 0;
       $902 = 1;
       break;
      } else {
       $$pre$i$i$i = HEAP32[$0>>2]|0;
       $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
       $902 = $phitmp$i$i$i;
       break;
      }
     }
    } while(0);
    $75 = HEAP32[$1>>2]|0;
    $76 = ($75|0)==(0|0);
    do {
     if ($76) {
      label = 32;
     } else {
      $77 = ((($75)) + 12|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ((($75)) + 16|0);
      $80 = HEAP32[$79>>2]|0;
      $81 = ($78|0)==($80|0);
      if ($81) {
       $82 = HEAP32[$75>>2]|0;
       $83 = ((($82)) + 36|0);
       $84 = HEAP32[$83>>2]|0;
       __THREW__ = 0;
       $85 = (invoke_ii($84|0,($75|0))|0);
       $86 = __THREW__; __THREW__ = 0;
       $87 = $86&1;
       if ($87) {
        label = 35;
        break L24;
       } else {
        $$0$i$i2$i$i = $85;
       }
      } else {
       $88 = HEAP8[$78>>0]|0;
       $89 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($88)|0);
       $$0$i$i2$i$i = $89;
      }
      $90 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
      $91 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i,$90)|0);
      if ($91) {
       HEAP32[$1>>2] = 0;
       label = 32;
       break;
      } else {
       if ($902) {
        $903 = $75;
        break;
       } else {
        label = 246;
        break L24;
       }
      }
     }
    } while(0);
    if ((label|0) == 32) {
     label = 0;
     if ($902) {
      label = 246;
      break;
     } else {
      $903 = 0;
     }
    }
    $96 = (($16) + ($$0158)|0);
    $97 = HEAP8[$96>>0]|0;
    $98 = $97 << 24 >> 24;
    $99 = ($$0158|0)!=(3);
    L49: do {
     switch ($98|0) {
     case 1:  {
      if ($99) {
       $100 = HEAP32[$0>>2]|0;
       $101 = ((($100)) + 12|0);
       $102 = HEAP32[$101>>2]|0;
       $103 = ((($100)) + 16|0);
       $104 = HEAP32[$103>>2]|0;
       $105 = ($102|0)==($104|0);
       if ($105) {
        $106 = HEAP32[$100>>2]|0;
        $107 = ((($106)) + 36|0);
        $108 = HEAP32[$107>>2]|0;
        __THREW__ = 0;
        $109 = (invoke_ii($108|0,($100|0))|0);
        $110 = __THREW__; __THREW__ = 0;
        $111 = $110&1;
        if ($111) {
         label = 35;
         break L24;
        } else {
         $$0$i$i182 = $109;
        }
       } else {
        $112 = HEAP8[$102>>0]|0;
        $113 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($112)|0);
        $$0$i$i182 = $113;
       }
       $114 = $$0$i$i182&255;
       $115 = ($114<<24>>24)>(-1);
       if (!($115)) {
        label = 47;
        break L24;
       }
       $sext423 = $$0$i$i182 << 24;
       $116 = $sext423 >> 24;
       $117 = HEAP32[$42>>2]|0;
       $118 = (($117) + ($116<<1)|0);
       $119 = HEAP16[$118>>1]|0;
       $120 = $119 & 8192;
       $121 = ($120<<16>>16)==(0);
       if ($121) {
        label = 47;
        break L24;
       }
       $122 = HEAP32[$0>>2]|0;
       $123 = ((($122)) + 12|0);
       $124 = HEAP32[$123>>2]|0;
       $125 = ((($122)) + 16|0);
       $126 = HEAP32[$125>>2]|0;
       $127 = ($124|0)==($126|0);
       if ($127) {
        $128 = HEAP32[$122>>2]|0;
        $129 = ((($128)) + 40|0);
        $130 = HEAP32[$129>>2]|0;
        __THREW__ = 0;
        $131 = (invoke_ii($130|0,($122|0))|0);
        $132 = __THREW__; __THREW__ = 0;
        $133 = $132&1;
        if ($133) {
         label = 35;
         break L24;
        } else {
         $$0$i$i184 = $131;
        }
       } else {
        $134 = ((($124)) + 1|0);
        HEAP32[$123>>2] = $134;
        $135 = HEAP8[$124>>0]|0;
        $136 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($135)|0);
        $$0$i$i184 = $136;
       }
       $137 = $$0$i$i184&255;
       __THREW__ = 0;
       invoke_vii(1006,($23|0),($137|0));
       $138 = __THREW__; __THREW__ = 0;
       $139 = $138&1;
       if ($139) {
        label = 35;
        break L24;
       } else {
        $$pre$i4$i$i194 = $903;$160 = $903;
        label = 49;
       }
      } else {
       $$1150 = $$0149;
      }
      break;
     }
     case 0:  {
      if ($99) {
       $$pre$i4$i$i194 = $903;$160 = $903;
       label = 49;
      } else {
       $$1150 = $$0149;
      }
      break;
     }
     case 3:  {
      $217 = HEAP8[$43>>0]|0;
      $218 = ($217<<24>>24)<(0);
      $219 = HEAP32[$44>>2]|0;
      $220 = $217&255;
      $221 = $218 ? $219 : $220;
      $222 = HEAP8[$45>>0]|0;
      $223 = ($222<<24>>24)<(0);
      $224 = HEAP32[$46>>2]|0;
      $225 = $222&255;
      $226 = $223 ? $224 : $225;
      $227 = (0 - ($226))|0;
      $228 = ($221|0)==($227|0);
      if ($228) {
       $$1150 = $$0149;
      } else {
       $229 = ($221|0)==(0);
       if ($229) {
        $271 = HEAP32[$0>>2]|0;
        $272 = ((($271)) + 12|0);
        $273 = HEAP32[$272>>2]|0;
        $274 = ((($271)) + 16|0);
        $275 = HEAP32[$274>>2]|0;
        $276 = ($273|0)==($275|0);
        if ($276) {
         $277 = HEAP32[$271>>2]|0;
         $278 = ((($277)) + 36|0);
         $279 = HEAP32[$278>>2]|0;
         __THREW__ = 0;
         $280 = (invoke_ii($279|0,($271|0))|0);
         $281 = __THREW__; __THREW__ = 0;
         $282 = $281&1;
         if ($282) {
          label = 35;
          break L24;
         } else {
          $$0$i$i217 = $280;
         }
        } else {
         $283 = HEAP8[$273>>0]|0;
         $284 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($283)|0);
         $$0$i$i217 = $284;
        }
        $285 = $$0$i$i217&255;
        $286 = HEAP8[$45>>0]|0;
        $287 = ($286<<24>>24)<(0);
        $288 = HEAP32[$22>>2]|0;
        $289 = $287 ? $288 : $22;
        $290 = HEAP8[$289>>0]|0;
        $291 = ($285<<24>>24)==($290<<24>>24);
        if (!($291)) {
         $$1150 = $$0149;
         break L49;
        }
        $292 = HEAP32[$0>>2]|0;
        $293 = ((($292)) + 12|0);
        $294 = HEAP32[$293>>2]|0;
        $295 = ((($292)) + 16|0);
        $296 = HEAP32[$295>>2]|0;
        $297 = ($294|0)==($296|0);
        if ($297) {
         $298 = HEAP32[$292>>2]|0;
         $299 = ((($298)) + 40|0);
         $300 = HEAP32[$299>>2]|0;
         __THREW__ = 0;
         (invoke_ii($300|0,($292|0))|0);
         $301 = __THREW__; __THREW__ = 0;
         $302 = $301&1;
         if ($302) {
          label = 35;
          break L24;
         }
        } else {
         $303 = ((($294)) + 1|0);
         HEAP32[$293>>2] = $303;
         $304 = HEAP8[$294>>0]|0;
         (__ZNSt3__211char_traitsIcE11to_int_typeEc($304)|0);
        }
        HEAP8[$6>>0] = 1;
        $305 = HEAP8[$45>>0]|0;
        $306 = ($305<<24>>24)<(0);
        $307 = HEAP32[$46>>2]|0;
        $308 = $305&255;
        $309 = $306 ? $307 : $308;
        $310 = ($309>>>0)>(1);
        $$$0149159 = $310 ? $22 : $$0149;
        $$1150 = $$$0149159;
        break L49;
       }
       $230 = ($226|0)==(0);
       $231 = HEAP32[$0>>2]|0;
       $232 = ((($231)) + 12|0);
       $233 = HEAP32[$232>>2]|0;
       $234 = ((($231)) + 16|0);
       $235 = HEAP32[$234>>2]|0;
       $236 = ($233|0)==($235|0);
       if ($230) {
        if ($236) {
         $237 = HEAP32[$231>>2]|0;
         $238 = ((($237)) + 36|0);
         $239 = HEAP32[$238>>2]|0;
         __THREW__ = 0;
         $240 = (invoke_ii($239|0,($231|0))|0);
         $241 = __THREW__; __THREW__ = 0;
         $242 = $241&1;
         if ($242) {
          label = 35;
          break L24;
         } else {
          $$0$i$i212 = $240;
         }
        } else {
         $243 = HEAP8[$233>>0]|0;
         $244 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($243)|0);
         $$0$i$i212 = $244;
        }
        $245 = $$0$i$i212&255;
        $246 = HEAP8[$43>>0]|0;
        $247 = ($246<<24>>24)<(0);
        $248 = HEAP32[$21>>2]|0;
        $249 = $247 ? $248 : $21;
        $250 = HEAP8[$249>>0]|0;
        $251 = ($245<<24>>24)==($250<<24>>24);
        if (!($251)) {
         HEAP8[$6>>0] = 1;
         $$1150 = $$0149;
         break L49;
        }
        $252 = HEAP32[$0>>2]|0;
        $253 = ((($252)) + 12|0);
        $254 = HEAP32[$253>>2]|0;
        $255 = ((($252)) + 16|0);
        $256 = HEAP32[$255>>2]|0;
        $257 = ($254|0)==($256|0);
        if ($257) {
         $258 = HEAP32[$252>>2]|0;
         $259 = ((($258)) + 40|0);
         $260 = HEAP32[$259>>2]|0;
         __THREW__ = 0;
         (invoke_ii($260|0,($252|0))|0);
         $261 = __THREW__; __THREW__ = 0;
         $262 = $261&1;
         if ($262) {
          label = 35;
          break L24;
         }
        } else {
         $263 = ((($254)) + 1|0);
         HEAP32[$253>>2] = $263;
         $264 = HEAP8[$254>>0]|0;
         (__ZNSt3__211char_traitsIcE11to_int_typeEc($264)|0);
        }
        $265 = HEAP8[$43>>0]|0;
        $266 = ($265<<24>>24)<(0);
        $267 = HEAP32[$44>>2]|0;
        $268 = $265&255;
        $269 = $266 ? $267 : $268;
        $270 = ($269>>>0)>(1);
        $$$0149 = $270 ? $21 : $$0149;
        $$1150 = $$$0149;
        break L49;
       }
       if ($236) {
        $311 = HEAP32[$231>>2]|0;
        $312 = ((($311)) + 36|0);
        $313 = HEAP32[$312>>2]|0;
        __THREW__ = 0;
        $314 = (invoke_ii($313|0,($231|0))|0);
        $315 = __THREW__; __THREW__ = 0;
        $316 = $315&1;
        if ($316) {
         label = 35;
         break L24;
        } else {
         $$0$i$i224 = $314;
        }
       } else {
        $317 = HEAP8[$233>>0]|0;
        $318 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($317)|0);
        $$0$i$i224 = $318;
       }
       $319 = $$0$i$i224&255;
       $320 = HEAP8[$43>>0]|0;
       $321 = ($320<<24>>24)<(0);
       $322 = HEAP32[$21>>2]|0;
       $323 = $321 ? $322 : $21;
       $324 = HEAP8[$323>>0]|0;
       $325 = ($319<<24>>24)==($324<<24>>24);
       $326 = HEAP32[$0>>2]|0;
       $327 = ((($326)) + 12|0);
       $328 = HEAP32[$327>>2]|0;
       $329 = ((($326)) + 16|0);
       $330 = HEAP32[$329>>2]|0;
       $331 = ($328|0)==($330|0);
       if ($325) {
        if ($331) {
         $332 = HEAP32[$326>>2]|0;
         $333 = ((($332)) + 40|0);
         $334 = HEAP32[$333>>2]|0;
         __THREW__ = 0;
         (invoke_ii($334|0,($326|0))|0);
         $335 = __THREW__; __THREW__ = 0;
         $336 = $335&1;
         if ($336) {
          label = 35;
          break L24;
         }
        } else {
         $337 = ((($328)) + 1|0);
         HEAP32[$327>>2] = $337;
         $338 = HEAP8[$328>>0]|0;
         (__ZNSt3__211char_traitsIcE11to_int_typeEc($338)|0);
        }
        $339 = HEAP8[$43>>0]|0;
        $340 = ($339<<24>>24)<(0);
        $341 = HEAP32[$44>>2]|0;
        $342 = $339&255;
        $343 = $340 ? $341 : $342;
        $344 = ($343>>>0)>(1);
        $$$0149160 = $344 ? $21 : $$0149;
        $$1150 = $$$0149160;
        break L49;
       }
       if ($331) {
        $345 = HEAP32[$326>>2]|0;
        $346 = ((($345)) + 36|0);
        $347 = HEAP32[$346>>2]|0;
        __THREW__ = 0;
        $348 = (invoke_ii($347|0,($326|0))|0);
        $349 = __THREW__; __THREW__ = 0;
        $350 = $349&1;
        if ($350) {
         label = 35;
         break L24;
        } else {
         $$0$i$i231 = $348;
        }
       } else {
        $351 = HEAP8[$328>>0]|0;
        $352 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($351)|0);
        $$0$i$i231 = $352;
       }
       $353 = $$0$i$i231&255;
       $354 = HEAP8[$45>>0]|0;
       $355 = ($354<<24>>24)<(0);
       $356 = HEAP32[$22>>2]|0;
       $357 = $355 ? $356 : $22;
       $358 = HEAP8[$357>>0]|0;
       $359 = ($353<<24>>24)==($358<<24>>24);
       if (!($359)) {
        label = 109;
        break L24;
       }
       $360 = HEAP32[$0>>2]|0;
       $361 = ((($360)) + 12|0);
       $362 = HEAP32[$361>>2]|0;
       $363 = ((($360)) + 16|0);
       $364 = HEAP32[$363>>2]|0;
       $365 = ($362|0)==($364|0);
       if ($365) {
        $366 = HEAP32[$360>>2]|0;
        $367 = ((($366)) + 40|0);
        $368 = HEAP32[$367>>2]|0;
        __THREW__ = 0;
        (invoke_ii($368|0,($360|0))|0);
        $369 = __THREW__; __THREW__ = 0;
        $370 = $369&1;
        if ($370) {
         label = 35;
         break L24;
        }
       } else {
        $371 = ((($362)) + 1|0);
        HEAP32[$361>>2] = $371;
        $372 = HEAP8[$362>>0]|0;
        (__ZNSt3__211char_traitsIcE11to_int_typeEc($372)|0);
       }
       HEAP8[$6>>0] = 1;
       $373 = HEAP8[$45>>0]|0;
       $374 = ($373<<24>>24)<(0);
       $375 = HEAP32[$46>>2]|0;
       $376 = $373&255;
       $377 = $374 ? $375 : $376;
       $378 = ($377>>>0)>(1);
       $$$0149161 = $378 ? $22 : $$0149;
       $$1150 = $$$0149161;
      }
      break;
     }
     case 2:  {
      $381 = ($$0158>>>0)<(2);
      $or$cond = $381 | $57;
      if (!($or$cond)) {
       $382 = ($$0158|0)==(2);
       $383 = HEAP8[$53>>0]|0;
       $384 = ($383<<24>>24)!=(0);
       $385 = $382 & $384;
       $brmerge = $48 | $385;
       if (!($brmerge)) {
        $$1150 = 0;
        break L49;
       }
      }
      $386 = HEAP8[$49>>0]|0;
      $387 = ($386<<24>>24)<(0);
      $388 = HEAP32[$20>>2]|0;
      $389 = $387 ? $388 : $20;
      $390 = $389;
      $391 = ($$0158|0)==(0);
      L115: do {
       if ($391) {
        $$pre$i4$i$i260 = $903;$$sroa$0360$0 = $390;$431 = $386;$435 = $388;$458 = $903;
       } else {
        $392 = (($$0158) + -1)|0;
        $393 = (($16) + ($392)|0);
        $394 = HEAP8[$393>>0]|0;
        $switch = ($394&255)<(2);
        if ($switch) {
         $395 = HEAP32[$50>>2]|0;
         $396 = $386&255;
         $397 = $387 ? $395 : $396;
         $398 = (($389) + ($397)|0);
         $$sroa$0381$0 = $390;
         while(1) {
          $399 = $$sroa$0381$0;
          $400 = ($399|0)==($398|0);
          if ($400) {
           break;
          }
          $401 = HEAP8[$399>>0]|0;
          $402 = ($401<<24>>24)>(-1);
          if (!($402)) {
           break;
          }
          $403 = $401 << 24 >> 24;
          $404 = HEAP32[$42>>2]|0;
          $405 = (($404) + ($403<<1)|0);
          $406 = HEAP16[$405>>1]|0;
          $407 = $406 & 8192;
          $408 = ($407<<16>>16)==(0);
          if ($408) {
           break;
          }
          $409 = ((($399)) + 1|0);
          $410 = $409;
          $$sroa$0381$0 = $410;
         }
         $411 = (($$sroa$0381$0) - ($390))|0;
         $412 = HEAP8[$51>>0]|0;
         $413 = ($412<<24>>24)<(0);
         $414 = HEAP32[$52>>2]|0;
         $415 = $412&255;
         $416 = $413 ? $414 : $415;
         $417 = ($411>>>0)>($416>>>0);
         if ($417) {
          $$pre$i4$i$i260 = $903;$$sroa$0360$0 = $390;$431 = $386;$435 = $388;$458 = $903;
         } else {
          $418 = HEAP32[$23>>2]|0;
          $419 = $413 ? $418 : $23;
          $420 = (($419) + ($416)|0);
          $421 = (0 - ($411))|0;
          $422 = (($420) + ($421)|0);
          $$sink = $389;$424 = $422;
          while(1) {
           $423 = ($424|0)==($420|0);
           if ($423) {
            $$pre$i4$i$i260 = $903;$$sroa$0360$0 = $$sroa$0381$0;$431 = $386;$435 = $388;$458 = $903;
            break L115;
           }
           $425 = HEAP8[$424>>0]|0;
           $426 = HEAP8[$$sink>>0]|0;
           $427 = ($425<<24>>24)==($426<<24>>24);
           if (!($427)) {
            $$pre$i4$i$i260 = $903;$$sroa$0360$0 = $390;$431 = $386;$435 = $388;$458 = $903;
            break L115;
           }
           $428 = ((($424)) + 1|0);
           $429 = ((($$sink)) + 1|0);
           $$sink = $429;$424 = $428;
          }
         }
        } else {
         $$pre$i4$i$i260 = $903;$$sroa$0360$0 = $390;$431 = $386;$435 = $388;$458 = $903;
        }
       }
      } while(0);
      L129: while(1) {
       $430 = ($431<<24>>24)<(0);
       $432 = HEAP32[$50>>2]|0;
       $433 = $431&255;
       $434 = $430 ? $435 : $20;
       $436 = $430 ? $432 : $433;
       $437 = (($434) + ($436)|0);
       $438 = $$sroa$0360$0;
       $439 = ($438|0)==($437|0);
       if ($439) {
        $$lcssa = $437;
        break;
       }
       $440 = HEAP32[$0>>2]|0;
       $441 = ($440|0)==(0|0);
       do {
        if ($441) {
         $474 = 1;
        } else {
         $442 = ((($440)) + 12|0);
         $443 = HEAP32[$442>>2]|0;
         $444 = ((($440)) + 16|0);
         $445 = HEAP32[$444>>2]|0;
         $446 = ($443|0)==($445|0);
         if ($446) {
          $447 = HEAP32[$440>>2]|0;
          $448 = ((($447)) + 36|0);
          $449 = HEAP32[$448>>2]|0;
          __THREW__ = 0;
          $450 = (invoke_ii($449|0,($440|0))|0);
          $451 = __THREW__; __THREW__ = 0;
          $452 = $451&1;
          if ($452) {
           label = 148;
           break L24;
          } else {
           $$0$i$i$i$i252 = $450;
          }
         } else {
          $453 = HEAP8[$443>>0]|0;
          $454 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($453)|0);
          $$0$i$i$i$i252 = $454;
         }
         $455 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $456 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i252,$455)|0);
         if ($456) {
          HEAP32[$0>>2] = 0;
          $474 = 1;
          break;
         } else {
          $$pre$i$i$i254 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i255 = ($$pre$i$i$i254|0)==(0|0);
          $474 = $phitmp$i$i$i255;
          break;
         }
        }
       } while(0);
       $457 = ($458|0)==(0|0);
       do {
        if ($457) {
         $$pre$i4$i$i260510 = $$pre$i4$i$i260;
         label = 138;
        } else {
         $459 = ((($458)) + 12|0);
         $460 = HEAP32[$459>>2]|0;
         $461 = ((($458)) + 16|0);
         $462 = HEAP32[$461>>2]|0;
         $463 = ($460|0)==($462|0);
         if ($463) {
          $464 = HEAP32[$458>>2]|0;
          $465 = ((($464)) + 36|0);
          $466 = HEAP32[$465>>2]|0;
          __THREW__ = 0;
          $467 = (invoke_ii($466|0,($458|0))|0);
          $468 = __THREW__; __THREW__ = 0;
          $469 = $468&1;
          if ($469) {
           label = 148;
           break L24;
          } else {
           $$0$i$i2$i$i258 = $467;
          }
         } else {
          $470 = HEAP8[$460>>0]|0;
          $471 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($470)|0);
          $$0$i$i2$i$i258 = $471;
         }
         $472 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $473 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i258,$472)|0);
         if ($473) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i260510 = 0;
          label = 138;
          break;
         } else {
          $phitmp$i5$i$i261 = ($$pre$i4$i$i260|0)==(0|0);
          $tmp$i$i263 = $474 ^ $phitmp$i5$i$i261;
          if ($tmp$i$i263) {
           $$pre$i4$i$i260509 = $$pre$i4$i$i260;$905 = $$pre$i4$i$i260;
           break;
          } else {
           $$lcssa = $438;
           break L129;
          }
         }
        }
       } while(0);
       if ((label|0) == 138) {
        label = 0;
        if ($474) {
         $$lcssa = $438;
         break;
        } else {
         $$pre$i4$i$i260509 = $$pre$i4$i$i260510;$905 = 0;
        }
       }
       $475 = HEAP32[$0>>2]|0;
       $476 = ((($475)) + 12|0);
       $477 = HEAP32[$476>>2]|0;
       $478 = ((($475)) + 16|0);
       $479 = HEAP32[$478>>2]|0;
       $480 = ($477|0)==($479|0);
       if ($480) {
        $481 = HEAP32[$475>>2]|0;
        $482 = ((($481)) + 36|0);
        $483 = HEAP32[$482>>2]|0;
        __THREW__ = 0;
        $484 = (invoke_ii($483|0,($475|0))|0);
        $485 = __THREW__; __THREW__ = 0;
        $486 = $485&1;
        if ($486) {
         label = 148;
         break L24;
        } else {
         $$0$i$i267 = $484;
        }
       } else {
        $487 = HEAP8[$477>>0]|0;
        $488 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($487)|0);
        $$0$i$i267 = $488;
       }
       $489 = $$0$i$i267&255;
       $490 = HEAP8[$438>>0]|0;
       $491 = ($489<<24>>24)==($490<<24>>24);
       if (!($491)) {
        $$lcssa = $438;
        break;
       }
       $492 = HEAP32[$0>>2]|0;
       $493 = ((($492)) + 12|0);
       $494 = HEAP32[$493>>2]|0;
       $495 = ((($492)) + 16|0);
       $496 = HEAP32[$495>>2]|0;
       $497 = ($494|0)==($496|0);
       if ($497) {
        $498 = HEAP32[$492>>2]|0;
        $499 = ((($498)) + 40|0);
        $500 = HEAP32[$499>>2]|0;
        __THREW__ = 0;
        (invoke_ii($500|0,($492|0))|0);
        $501 = __THREW__; __THREW__ = 0;
        $502 = $501&1;
        if ($502) {
         label = 149;
         break L24;
        }
       } else {
        $503 = ((($494)) + 1|0);
        HEAP32[$493>>2] = $503;
        $504 = HEAP8[$494>>0]|0;
        (__ZNSt3__211char_traitsIcE11to_int_typeEc($504)|0);
       }
       $505 = ((($438)) + 1|0);
       $506 = $505;
       $$pre507 = HEAP8[$49>>0]|0;
       $$pre508 = HEAP32[$20>>2]|0;
       $$pre$i4$i$i260 = $$pre$i4$i$i260509;$$sroa$0360$0 = $506;$431 = $$pre507;$435 = $$pre508;$458 = $905;
      }
      if ($48) {
       $511 = HEAP8[$49>>0]|0;
       $512 = ($511<<24>>24)<(0);
       $513 = HEAP32[$20>>2]|0;
       $514 = HEAP32[$50>>2]|0;
       $515 = $511&255;
       $516 = $512 ? $513 : $20;
       $517 = $512 ? $514 : $515;
       $518 = (($516) + ($517)|0);
       $519 = ($$lcssa|0)==($518|0);
       if ($519) {
        $$1150 = $$0149;
       } else {
        label = 152;
        break L24;
       }
      } else {
       $$1150 = $$0149;
      }
      break;
     }
     case 4:  {
      $$0132 = 0;$$pre$i4$i$i281 = $903;$540 = $903;
      L165: while(1) {
       $522 = HEAP32[$0>>2]|0;
       $523 = ($522|0)==(0|0);
       do {
        if ($523) {
         $556 = 1;
        } else {
         $524 = ((($522)) + 12|0);
         $525 = HEAP32[$524>>2]|0;
         $526 = ((($522)) + 16|0);
         $527 = HEAP32[$526>>2]|0;
         $528 = ($525|0)==($527|0);
         if ($528) {
          $529 = HEAP32[$522>>2]|0;
          $530 = ((($529)) + 36|0);
          $531 = HEAP32[$530>>2]|0;
          __THREW__ = 0;
          $532 = (invoke_ii($531|0,($522|0))|0);
          $533 = __THREW__; __THREW__ = 0;
          $534 = $533&1;
          if ($534) {
           label = 244;
           break L24;
          } else {
           $$0$i$i$i$i273 = $532;
          }
         } else {
          $535 = HEAP8[$525>>0]|0;
          $536 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($535)|0);
          $$0$i$i$i$i273 = $536;
         }
         $537 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $538 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i273,$537)|0);
         if ($538) {
          HEAP32[$0>>2] = 0;
          $556 = 1;
          break;
         } else {
          $$pre$i$i$i275 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i276 = ($$pre$i$i$i275|0)==(0|0);
          $556 = $phitmp$i$i$i276;
          break;
         }
        }
       } while(0);
       $539 = ($540|0)==(0|0);
       do {
        if ($539) {
         $$pre$i4$i$i281498 = $$pre$i4$i$i281;
         label = 166;
        } else {
         $541 = ((($540)) + 12|0);
         $542 = HEAP32[$541>>2]|0;
         $543 = ((($540)) + 16|0);
         $544 = HEAP32[$543>>2]|0;
         $545 = ($542|0)==($544|0);
         if ($545) {
          $546 = HEAP32[$540>>2]|0;
          $547 = ((($546)) + 36|0);
          $548 = HEAP32[$547>>2]|0;
          __THREW__ = 0;
          $549 = (invoke_ii($548|0,($540|0))|0);
          $550 = __THREW__; __THREW__ = 0;
          $551 = $550&1;
          if ($551) {
           label = 244;
           break L24;
          } else {
           $$0$i$i2$i$i279 = $549;
          }
         } else {
          $552 = HEAP8[$542>>0]|0;
          $553 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($552)|0);
          $$0$i$i2$i$i279 = $553;
         }
         $554 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $555 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i279,$554)|0);
         if ($555) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i281498 = 0;
          label = 166;
          break;
         } else {
          $phitmp$i5$i$i282 = ($$pre$i4$i$i281|0)==(0|0);
          $tmp$i$i284 = $556 ^ $phitmp$i5$i$i282;
          if ($tmp$i$i284) {
           $$pre$i4$i$i281500 = $$pre$i4$i$i281;$906 = $$pre$i4$i$i281;
           break;
          } else {
           $$pre$i4$i$i303 = $$pre$i4$i$i281;
           break L165;
          }
         }
        }
       } while(0);
       if ((label|0) == 166) {
        label = 0;
        if ($556) {
         $$pre$i4$i$i303 = $$pre$i4$i$i281498;
         break;
        } else {
         $$pre$i4$i$i281500 = $$pre$i4$i$i281498;$906 = 0;
        }
       }
       $557 = HEAP32[$0>>2]|0;
       $558 = ((($557)) + 12|0);
       $559 = HEAP32[$558>>2]|0;
       $560 = ((($557)) + 16|0);
       $561 = HEAP32[$560>>2]|0;
       $562 = ($559|0)==($561|0);
       if ($562) {
        $563 = HEAP32[$557>>2]|0;
        $564 = ((($563)) + 36|0);
        $565 = HEAP32[$564>>2]|0;
        __THREW__ = 0;
        $566 = (invoke_ii($565|0,($557|0))|0);
        $567 = __THREW__; __THREW__ = 0;
        $568 = $567&1;
        if ($568) {
         label = 244;
         break L24;
        } else {
         $$0$i$i288 = $566;
        }
       } else {
        $569 = HEAP8[$559>>0]|0;
        $570 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($569)|0);
        $$0$i$i288 = $570;
       }
       $571 = $$0$i$i288&255;
       $572 = ($571<<24>>24)>(-1);
       if ($572) {
        $sext421 = $$0$i$i288 << 24;
        $573 = $sext421 >> 24;
        $574 = HEAP32[$42>>2]|0;
        $575 = (($574) + ($573<<1)|0);
        $576 = HEAP16[$575>>1]|0;
        $577 = $576 & 2048;
        $578 = ($577<<16>>16)==(0);
        if ($578) {
         label = 177;
        } else {
         $579 = HEAP32[$9>>2]|0;
         $580 = HEAP32[$11>>2]|0;
         $581 = ($579|0)==($580|0);
         if ($581) {
          __THREW__ = 0;
          invoke_viii(1009,($8|0),($9|0),($11|0));
          $582 = __THREW__; __THREW__ = 0;
          $583 = $582&1;
          if ($583) {
           label = 244;
           break L24;
          }
          $$pre = HEAP32[$9>>2]|0;
          $585 = $$pre;
         } else {
          $585 = $579;
         }
         $584 = ((($585)) + 1|0);
         HEAP32[$9>>2] = $584;
         HEAP8[$585>>0] = $571;
         $586 = (($$0132) + 1)|0;
         $$2134$ph = $586;
        }
       } else {
        label = 177;
       }
       if ((label|0) == 177) {
        label = 0;
        $587 = HEAP8[$54>>0]|0;
        $588 = ($587<<24>>24)<(0);
        $589 = HEAP32[$55>>2]|0;
        $590 = $587&255;
        $591 = $588 ? $589 : $590;
        $592 = ($591|0)!=(0);
        $593 = ($$0132|0)!=(0);
        $or$cond13 = $593 & $592;
        $594 = HEAP8[$18>>0]|0;
        $595 = ($571<<24>>24)==($594<<24>>24);
        $or$cond164 = $595 & $or$cond13;
        if (!($or$cond164)) {
         $$pre$i4$i$i303 = $$pre$i4$i$i281500;
         break;
        }
        $596 = HEAP32[$14>>2]|0;
        $597 = HEAP32[$15>>2]|0;
        $598 = ($596|0)==($597|0);
        if ($598) {
         __THREW__ = 0;
         invoke_viii(1010,($13|0),($14|0),($15|0));
         $599 = __THREW__; __THREW__ = 0;
         $600 = $599&1;
         if ($600) {
          label = 244;
          break L24;
         }
         $$pre502 = HEAP32[$14>>2]|0;
         $602 = $$pre502;
        } else {
         $602 = $596;
        }
        $601 = ((($602)) + 4|0);
        HEAP32[$14>>2] = $601;
        HEAP32[$602>>2] = $$0132;
        $$2134$ph = 0;
       }
       $603 = HEAP32[$0>>2]|0;
       $604 = ((($603)) + 12|0);
       $605 = HEAP32[$604>>2]|0;
       $606 = ((($603)) + 16|0);
       $607 = HEAP32[$606>>2]|0;
       $608 = ($605|0)==($607|0);
       if ($608) {
        $609 = HEAP32[$603>>2]|0;
        $610 = ((($609)) + 40|0);
        $611 = HEAP32[$610>>2]|0;
        __THREW__ = 0;
        (invoke_ii($611|0,($603|0))|0);
        $612 = __THREW__; __THREW__ = 0;
        $613 = $612&1;
        if ($613) {
         label = 244;
         break L24;
        } else {
         $$0132 = $$2134$ph;$$pre$i4$i$i281 = $$pre$i4$i$i281500;$540 = $906;
         continue;
        }
       } else {
        $614 = ((($605)) + 1|0);
        HEAP32[$604>>2] = $614;
        $615 = HEAP8[$605>>0]|0;
        (__ZNSt3__211char_traitsIcE11to_int_typeEc($615)|0);
        $$0132 = $$2134$ph;$$pre$i4$i$i281 = $$pre$i4$i$i281500;$540 = $906;
        continue;
       }
      }
      $616 = HEAP32[$13>>2]|0;
      $617 = HEAP32[$14>>2]|0;
      $618 = ($616|0)!=($617|0);
      $619 = ($$0132|0)!=(0);
      $or$cond15 = $619 & $618;
      if ($or$cond15) {
       $620 = HEAP32[$15>>2]|0;
       $621 = ($617|0)==($620|0);
       if ($621) {
        __THREW__ = 0;
        invoke_viii(1010,($13|0),($14|0),($15|0));
        $622 = __THREW__; __THREW__ = 0;
        $623 = $622&1;
        if ($623) {
         label = 244;
         break L24;
        }
        $$pre504 = HEAP32[$14>>2]|0;
        $625 = $$pre504;
       } else {
        $625 = $617;
       }
       $624 = ((($625)) + 4|0);
       HEAP32[$14>>2] = $624;
       HEAP32[$625>>2] = $$0132;
      }
      $626 = HEAP32[$24>>2]|0;
      $627 = ($626|0)>(0);
      L219: do {
       if ($627) {
        $628 = HEAP32[$0>>2]|0;
        $629 = ($628|0)==(0|0);
        do {
         if ($629) {
          $907 = 1;
         } else {
          $630 = ((($628)) + 12|0);
          $631 = HEAP32[$630>>2]|0;
          $632 = ((($628)) + 16|0);
          $633 = HEAP32[$632>>2]|0;
          $634 = ($631|0)==($633|0);
          if ($634) {
           $635 = HEAP32[$628>>2]|0;
           $636 = ((($635)) + 36|0);
           $637 = HEAP32[$636>>2]|0;
           __THREW__ = 0;
           $638 = (invoke_ii($637|0,($628|0))|0);
           $639 = __THREW__; __THREW__ = 0;
           $640 = $639&1;
           if ($640) {
            label = 244;
            break L24;
           } else {
            $$0$i$i$i$i295 = $638;
           }
          } else {
           $641 = HEAP8[$631>>0]|0;
           $642 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($641)|0);
           $$0$i$i$i$i295 = $642;
          }
          $643 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
          $644 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i295,$643)|0);
          if ($644) {
           HEAP32[$0>>2] = 0;
           $907 = 1;
           break;
          } else {
           $$pre$i$i$i297 = HEAP32[$0>>2]|0;
           $phitmp$i$i$i298 = ($$pre$i$i$i297|0)==(0|0);
           $907 = $phitmp$i$i$i298;
           break;
          }
         }
        } while(0);
        $645 = ($$pre$i4$i$i303|0)==(0|0);
        do {
         if ($645) {
          label = 204;
         } else {
          $646 = ((($$pre$i4$i$i303)) + 12|0);
          $647 = HEAP32[$646>>2]|0;
          $648 = ((($$pre$i4$i$i303)) + 16|0);
          $649 = HEAP32[$648>>2]|0;
          $650 = ($647|0)==($649|0);
          if ($650) {
           $651 = HEAP32[$$pre$i4$i$i303>>2]|0;
           $652 = ((($651)) + 36|0);
           $653 = HEAP32[$652>>2]|0;
           __THREW__ = 0;
           $654 = (invoke_ii($653|0,($$pre$i4$i$i303|0))|0);
           $655 = __THREW__; __THREW__ = 0;
           $656 = $655&1;
           if ($656) {
            label = 244;
            break L24;
           } else {
            $$0$i$i2$i$i301 = $654;
           }
          } else {
           $657 = HEAP8[$647>>0]|0;
           $658 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($657)|0);
           $$0$i$i2$i$i301 = $658;
          }
          $659 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
          $660 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i301,$659)|0);
          if ($660) {
           HEAP32[$1>>2] = 0;
           label = 204;
           break;
          } else {
           if ($907) {
            $908 = $$pre$i4$i$i303;
            break;
           } else {
            label = 243;
            break L24;
           }
          }
         }
        } while(0);
        if ((label|0) == 204) {
         label = 0;
         if ($907) {
          label = 243;
          break L24;
         } else {
          $908 = 0;
         }
        }
        $661 = HEAP32[$0>>2]|0;
        $662 = ((($661)) + 12|0);
        $663 = HEAP32[$662>>2]|0;
        $664 = ((($661)) + 16|0);
        $665 = HEAP32[$664>>2]|0;
        $666 = ($663|0)==($665|0);
        if ($666) {
         $667 = HEAP32[$661>>2]|0;
         $668 = ((($667)) + 36|0);
         $669 = HEAP32[$668>>2]|0;
         __THREW__ = 0;
         $670 = (invoke_ii($669|0,($661|0))|0);
         $671 = __THREW__; __THREW__ = 0;
         $672 = $671&1;
         if ($672) {
          label = 244;
          break L24;
         } else {
          $$0$i$i309 = $670;
         }
        } else {
         $673 = HEAP8[$663>>0]|0;
         $674 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($673)|0);
         $$0$i$i309 = $674;
        }
        $675 = $$0$i$i309&255;
        $676 = HEAP8[$17>>0]|0;
        $677 = ($675<<24>>24)==($676<<24>>24);
        if (!($677)) {
         label = 243;
         break L24;
        }
        $678 = HEAP32[$0>>2]|0;
        $679 = ((($678)) + 12|0);
        $680 = HEAP32[$679>>2]|0;
        $681 = ((($678)) + 16|0);
        $682 = HEAP32[$681>>2]|0;
        $683 = ($680|0)==($682|0);
        if ($683) {
         $684 = HEAP32[$678>>2]|0;
         $685 = ((($684)) + 40|0);
         $686 = HEAP32[$685>>2]|0;
         __THREW__ = 0;
         (invoke_ii($686|0,($678|0))|0);
         $687 = __THREW__; __THREW__ = 0;
         $688 = $687&1;
         if ($688) {
          label = 244;
          break L24;
         } else {
          $$pre$i4$i$i322 = $908;$711 = $908;
         }
        } else {
         $689 = ((($680)) + 1|0);
         HEAP32[$679>>2] = $689;
         $690 = HEAP8[$680>>0]|0;
         (__ZNSt3__211char_traitsIcE11to_int_typeEc($690)|0);
         $$pre$i4$i$i322 = $908;$711 = $908;
        }
        while(1) {
         $691 = HEAP32[$24>>2]|0;
         $692 = ($691|0)>(0);
         if (!($692)) {
          break L219;
         }
         $693 = HEAP32[$0>>2]|0;
         $694 = ($693|0)==(0|0);
         do {
          if ($694) {
           $727 = 1;
          } else {
           $695 = ((($693)) + 12|0);
           $696 = HEAP32[$695>>2]|0;
           $697 = ((($693)) + 16|0);
           $698 = HEAP32[$697>>2]|0;
           $699 = ($696|0)==($698|0);
           if ($699) {
            $700 = HEAP32[$693>>2]|0;
            $701 = ((($700)) + 36|0);
            $702 = HEAP32[$701>>2]|0;
            __THREW__ = 0;
            $703 = (invoke_ii($702|0,($693|0))|0);
            $704 = __THREW__; __THREW__ = 0;
            $705 = $704&1;
            if ($705) {
             label = 244;
             break L24;
            } else {
             $$0$i$i$i$i314 = $703;
            }
           } else {
            $706 = HEAP8[$696>>0]|0;
            $707 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($706)|0);
            $$0$i$i$i$i314 = $707;
           }
           $708 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
           $709 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i314,$708)|0);
           if ($709) {
            HEAP32[$0>>2] = 0;
            $727 = 1;
            break;
           } else {
            $$pre$i$i$i316 = HEAP32[$0>>2]|0;
            $phitmp$i$i$i317 = ($$pre$i$i$i316|0)==(0|0);
            $727 = $phitmp$i$i$i317;
            break;
           }
          }
         } while(0);
         $710 = ($711|0)==(0|0);
         do {
          if ($710) {
           $$pre$i4$i$i322505$ph = $$pre$i4$i$i322;
           label = 227;
          } else {
           $712 = ((($711)) + 12|0);
           $713 = HEAP32[$712>>2]|0;
           $714 = ((($711)) + 16|0);
           $715 = HEAP32[$714>>2]|0;
           $716 = ($713|0)==($715|0);
           if ($716) {
            $717 = HEAP32[$711>>2]|0;
            $718 = ((($717)) + 36|0);
            $719 = HEAP32[$718>>2]|0;
            __THREW__ = 0;
            $720 = (invoke_ii($719|0,($711|0))|0);
            $721 = __THREW__; __THREW__ = 0;
            $722 = $721&1;
            if ($722) {
             label = 244;
             break L24;
            } else {
             $$0$i$i2$i$i320 = $720;
            }
           } else {
            $723 = HEAP8[$713>>0]|0;
            $724 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($723)|0);
            $$0$i$i2$i$i320 = $724;
           }
           $725 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
           $726 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i320,$725)|0);
           if ($726) {
            HEAP32[$1>>2] = 0;
            $$pre$i4$i$i322505$ph = 0;
            label = 227;
            break;
           } else {
            $phitmp$i5$i$i323 = ($$pre$i4$i$i322|0)==(0|0);
            $tmp$i$i325 = $727 ^ $phitmp$i5$i$i323;
            if ($tmp$i$i325) {
             $$pre$i4$i$i322505518 = $$pre$i4$i$i322;$909 = $$pre$i4$i$i322;
             break;
            } else {
             label = 243;
             break L24;
            }
           }
          }
         } while(0);
         if ((label|0) == 227) {
          label = 0;
          if ($727) {
           label = 243;
           break L24;
          } else {
           $$pre$i4$i$i322505518 = $$pre$i4$i$i322505$ph;$909 = 0;
          }
         }
         $728 = HEAP32[$0>>2]|0;
         $729 = ((($728)) + 12|0);
         $730 = HEAP32[$729>>2]|0;
         $731 = ((($728)) + 16|0);
         $732 = HEAP32[$731>>2]|0;
         $733 = ($730|0)==($732|0);
         if ($733) {
          $734 = HEAP32[$728>>2]|0;
          $735 = ((($734)) + 36|0);
          $736 = HEAP32[$735>>2]|0;
          __THREW__ = 0;
          $737 = (invoke_ii($736|0,($728|0))|0);
          $738 = __THREW__; __THREW__ = 0;
          $739 = $738&1;
          if ($739) {
           label = 244;
           break L24;
          } else {
           $$0$i$i329 = $737;
          }
         } else {
          $740 = HEAP8[$730>>0]|0;
          $741 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($740)|0);
          $$0$i$i329 = $741;
         }
         $742 = $$0$i$i329&255;
         $743 = ($742<<24>>24)>(-1);
         if (!($743)) {
          label = 243;
          break L24;
         }
         $sext = $$0$i$i329 << 24;
         $744 = $sext >> 24;
         $745 = HEAP32[$42>>2]|0;
         $746 = (($745) + ($744<<1)|0);
         $747 = HEAP16[$746>>1]|0;
         $748 = $747 & 2048;
         $749 = ($748<<16>>16)==(0);
         if ($749) {
          label = 243;
          break L24;
         }
         $750 = HEAP32[$9>>2]|0;
         $751 = HEAP32[$11>>2]|0;
         $752 = ($750|0)==($751|0);
         if ($752) {
          __THREW__ = 0;
          invoke_viii(1009,($8|0),($9|0),($11|0));
          $753 = __THREW__; __THREW__ = 0;
          $754 = $753&1;
          if ($754) {
           label = 244;
           break L24;
          }
         }
         $755 = HEAP32[$0>>2]|0;
         $756 = ((($755)) + 12|0);
         $757 = HEAP32[$756>>2]|0;
         $758 = ((($755)) + 16|0);
         $759 = HEAP32[$758>>2]|0;
         $760 = ($757|0)==($759|0);
         if ($760) {
          $761 = HEAP32[$755>>2]|0;
          $762 = ((($761)) + 36|0);
          $763 = HEAP32[$762>>2]|0;
          __THREW__ = 0;
          $764 = (invoke_ii($763|0,($755|0))|0);
          $765 = __THREW__; __THREW__ = 0;
          $766 = $765&1;
          if ($766) {
           label = 244;
           break L24;
          } else {
           $$0$i$i333 = $764;
          }
         } else {
          $767 = HEAP8[$757>>0]|0;
          $768 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($767)|0);
          $$0$i$i333 = $768;
         }
         $769 = $$0$i$i333&255;
         $770 = HEAP32[$9>>2]|0;
         $771 = ((($770)) + 1|0);
         HEAP32[$9>>2] = $771;
         HEAP8[$770>>0] = $769;
         $772 = HEAP32[$24>>2]|0;
         $773 = (($772) + -1)|0;
         HEAP32[$24>>2] = $773;
         $774 = HEAP32[$0>>2]|0;
         $775 = ((($774)) + 12|0);
         $776 = HEAP32[$775>>2]|0;
         $777 = ((($774)) + 16|0);
         $778 = HEAP32[$777>>2]|0;
         $779 = ($776|0)==($778|0);
         if ($779) {
          $780 = HEAP32[$774>>2]|0;
          $781 = ((($780)) + 40|0);
          $782 = HEAP32[$781>>2]|0;
          __THREW__ = 0;
          (invoke_ii($782|0,($774|0))|0);
          $783 = __THREW__; __THREW__ = 0;
          $784 = $783&1;
          if ($784) {
           label = 244;
           break L24;
          } else {
           $$pre$i4$i$i322 = $$pre$i4$i$i322505518;$711 = $909;
           continue;
          }
         } else {
          $785 = ((($776)) + 1|0);
          HEAP32[$775>>2] = $785;
          $786 = HEAP8[$776>>0]|0;
          (__ZNSt3__211char_traitsIcE11to_int_typeEc($786)|0);
          $$pre$i4$i$i322 = $$pre$i4$i$i322505518;$711 = $909;
          continue;
         }
        }
       }
      } while(0);
      $787 = HEAP32[$9>>2]|0;
      $788 = HEAP32[$8>>2]|0;
      $789 = ($787|0)==($788|0);
      if ($789) {
       label = 243;
       break L24;
      } else {
       $$1150 = $$0149;
      }
      break;
     }
     default: {
      $$1150 = $$0149;
     }
     }
    } while(0);
    L293: do {
     if ((label|0) == 49) {
      while(1) {
       label = 0;
       $142 = HEAP32[$0>>2]|0;
       $143 = ($142|0)==(0|0);
       do {
        if ($143) {
         $176 = 1;
        } else {
         $144 = ((($142)) + 12|0);
         $145 = HEAP32[$144>>2]|0;
         $146 = ((($142)) + 16|0);
         $147 = HEAP32[$146>>2]|0;
         $148 = ($145|0)==($147|0);
         if ($148) {
          $149 = HEAP32[$142>>2]|0;
          $150 = ((($149)) + 36|0);
          $151 = HEAP32[$150>>2]|0;
          __THREW__ = 0;
          $152 = (invoke_ii($151|0,($142|0))|0);
          $153 = __THREW__; __THREW__ = 0;
          $154 = $153&1;
          if ($154) {
           label = 35;
           break L24;
          } else {
           $$0$i$i$i$i186 = $152;
          }
         } else {
          $155 = HEAP8[$145>>0]|0;
          $156 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($155)|0);
          $$0$i$i$i$i186 = $156;
         }
         $157 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $158 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i186,$157)|0);
         if ($158) {
          HEAP32[$0>>2] = 0;
          $176 = 1;
          break;
         } else {
          $$pre$i$i$i188 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i189 = ($$pre$i$i$i188|0)==(0|0);
          $176 = $phitmp$i$i$i189;
          break;
         }
        }
       } while(0);
       $159 = ($160|0)==(0|0);
       do {
        if ($159) {
         $$pre$i4$i$i194512 = $$pre$i4$i$i194;
         label = 62;
        } else {
         $161 = ((($160)) + 12|0);
         $162 = HEAP32[$161>>2]|0;
         $163 = ((($160)) + 16|0);
         $164 = HEAP32[$163>>2]|0;
         $165 = ($162|0)==($164|0);
         if ($165) {
          $166 = HEAP32[$160>>2]|0;
          $167 = ((($166)) + 36|0);
          $168 = HEAP32[$167>>2]|0;
          __THREW__ = 0;
          $169 = (invoke_ii($168|0,($160|0))|0);
          $170 = __THREW__; __THREW__ = 0;
          $171 = $170&1;
          if ($171) {
           label = 35;
           break L24;
          } else {
           $$0$i$i2$i$i192 = $169;
          }
         } else {
          $172 = HEAP8[$162>>0]|0;
          $173 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($172)|0);
          $$0$i$i2$i$i192 = $173;
         }
         $174 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
         $175 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i192,$174)|0);
         if ($175) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i194512 = 0;
          label = 62;
          break;
         } else {
          $phitmp$i5$i$i195 = ($$pre$i4$i$i194|0)==(0|0);
          $tmp$i$i197 = $176 ^ $phitmp$i5$i$i195;
          if ($tmp$i$i197) {
           $$pre$i4$i$i194514 = $$pre$i4$i$i194;$904 = $$pre$i4$i$i194;
           break;
          } else {
           $$1150 = $$0149;
           break L293;
          }
         }
        }
       } while(0);
       if ((label|0) == 62) {
        label = 0;
        if ($176) {
         $$1150 = $$0149;
         break L293;
        } else {
         $$pre$i4$i$i194514 = $$pre$i4$i$i194512;$904 = 0;
        }
       }
       $177 = HEAP32[$0>>2]|0;
       $178 = ((($177)) + 12|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = ((($177)) + 16|0);
       $181 = HEAP32[$180>>2]|0;
       $182 = ($179|0)==($181|0);
       if ($182) {
        $183 = HEAP32[$177>>2]|0;
        $184 = ((($183)) + 36|0);
        $185 = HEAP32[$184>>2]|0;
        __THREW__ = 0;
        $186 = (invoke_ii($185|0,($177|0))|0);
        $187 = __THREW__; __THREW__ = 0;
        $188 = $187&1;
        if ($188) {
         label = 35;
         break L24;
        } else {
         $$0$i$i201 = $186;
        }
       } else {
        $189 = HEAP8[$179>>0]|0;
        $190 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($189)|0);
        $$0$i$i201 = $190;
       }
       $191 = $$0$i$i201&255;
       $192 = ($191<<24>>24)>(-1);
       if (!($192)) {
        $$1150 = $$0149;
        break L293;
       }
       $sext422 = $$0$i$i201 << 24;
       $193 = $sext422 >> 24;
       $194 = HEAP32[$42>>2]|0;
       $195 = (($194) + ($193<<1)|0);
       $196 = HEAP16[$195>>1]|0;
       $197 = $196 & 8192;
       $198 = ($197<<16>>16)==(0);
       if ($198) {
        $$1150 = $$0149;
        break L293;
       }
       $199 = HEAP32[$0>>2]|0;
       $200 = ((($199)) + 12|0);
       $201 = HEAP32[$200>>2]|0;
       $202 = ((($199)) + 16|0);
       $203 = HEAP32[$202>>2]|0;
       $204 = ($201|0)==($203|0);
       if ($204) {
        $205 = HEAP32[$199>>2]|0;
        $206 = ((($205)) + 40|0);
        $207 = HEAP32[$206>>2]|0;
        __THREW__ = 0;
        $208 = (invoke_ii($207|0,($199|0))|0);
        $209 = __THREW__; __THREW__ = 0;
        $210 = $209&1;
        if ($210) {
         label = 35;
         break L24;
        } else {
         $$0$i$i205 = $208;
        }
       } else {
        $211 = ((($201)) + 1|0);
        HEAP32[$200>>2] = $211;
        $212 = HEAP8[$201>>0]|0;
        $213 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($212)|0);
        $$0$i$i205 = $213;
       }
       $214 = $$0$i$i205&255;
       __THREW__ = 0;
       invoke_vii(1006,($23|0),($214|0));
       $215 = __THREW__; __THREW__ = 0;
       $216 = $215&1;
       if ($216) {
        label = 35;
        break L24;
       } else {
        $$pre$i4$i$i194 = $$pre$i4$i$i194514;$160 = $904;
        label = 49;
       }
      }
     }
    } while(0);
    $794 = (($$0158) + 1)|0;
    $$0149 = $$1150;$$0158 = $794;
   }
   L329: do {
    if ((label|0) == 35) {
     $94 = ___cxa_find_matching_catch_2()|0;
     $95 = tempRet0;
     $$5142 = $94;$$5148 = $95;
     break L21;
    }
    else if ((label|0) == 47) {
     $140 = HEAP32[$5>>2]|0;
     $141 = $140 | 4;
     HEAP32[$5>>2] = $141;
     $$10 = 0;
    }
    else if ((label|0) == 109) {
     $379 = HEAP32[$5>>2]|0;
     $380 = $379 | 4;
     HEAP32[$5>>2] = $380;
     $$10 = 0;
    }
    else if ((label|0) == 148) {
     $507 = ___cxa_find_matching_catch_2()|0;
     $508 = tempRet0;
     $$5142 = $507;$$5148 = $508;
     break L21;
    }
    else if ((label|0) == 149) {
     $509 = ___cxa_find_matching_catch_2()|0;
     $510 = tempRet0;
     $$5142 = $509;$$5148 = $510;
     break L21;
    }
    else if ((label|0) == 152) {
     $520 = HEAP32[$5>>2]|0;
     $521 = $520 | 4;
     HEAP32[$5>>2] = $521;
     $$10 = 0;
    }
    else if ((label|0) == 243) {
     $790 = HEAP32[$5>>2]|0;
     $791 = $790 | 4;
     HEAP32[$5>>2] = $791;
     $$10 = 0;
    }
    else if ((label|0) == 244) {
     $792 = ___cxa_find_matching_catch_2()|0;
     $793 = tempRet0;
     $$5142 = $792;$$5148 = $793;
     break L21;
    }
    else if ((label|0) == 246) {
     L331: do {
      if ($57) {
       $795 = ((($$0149)) + 11|0);
       $796 = ((($$0149)) + 4|0);
       $$0 = 1;
       L333: while(1) {
        $797 = HEAP8[$795>>0]|0;
        $798 = ($797<<24>>24)<(0);
        if ($798) {
         $799 = HEAP32[$796>>2]|0;
         $802 = $799;
        } else {
         $800 = $797&255;
         $802 = $800;
        }
        $801 = ($$0>>>0)<($802>>>0);
        if (!($801)) {
         break L331;
        }
        $803 = HEAP32[$0>>2]|0;
        $804 = ($803|0)==(0|0);
        do {
         if ($804) {
          $910 = 1;
         } else {
          $805 = ((($803)) + 12|0);
          $806 = HEAP32[$805>>2]|0;
          $807 = ((($803)) + 16|0);
          $808 = HEAP32[$807>>2]|0;
          $809 = ($806|0)==($808|0);
          if ($809) {
           $810 = HEAP32[$803>>2]|0;
           $811 = ((($810)) + 36|0);
           $812 = HEAP32[$811>>2]|0;
           __THREW__ = 0;
           $813 = (invoke_ii($812|0,($803|0))|0);
           $814 = __THREW__; __THREW__ = 0;
           $815 = $814&1;
           if ($815) {
            label = 274;
            break L333;
           } else {
            $$0$i$i$i$i339 = $813;
           }
          } else {
           $816 = HEAP8[$806>>0]|0;
           $817 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($816)|0);
           $$0$i$i$i$i339 = $817;
          }
          $818 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
          $819 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i$i$i339,$818)|0);
          if ($819) {
           HEAP32[$0>>2] = 0;
           $910 = 1;
           break;
          } else {
           $$pre$i$i$i341 = HEAP32[$0>>2]|0;
           $phitmp$i$i$i342 = ($$pre$i$i$i341|0)==(0|0);
           $910 = $phitmp$i$i$i342;
           break;
          }
         }
        } while(0);
        $820 = HEAP32[$1>>2]|0;
        $821 = ($820|0)==(0|0);
        do {
         if ($821) {
          label = 265;
         } else {
          $822 = ((($820)) + 12|0);
          $823 = HEAP32[$822>>2]|0;
          $824 = ((($820)) + 16|0);
          $825 = HEAP32[$824>>2]|0;
          $826 = ($823|0)==($825|0);
          if ($826) {
           $827 = HEAP32[$820>>2]|0;
           $828 = ((($827)) + 36|0);
           $829 = HEAP32[$828>>2]|0;
           __THREW__ = 0;
           $830 = (invoke_ii($829|0,($820|0))|0);
           $831 = __THREW__; __THREW__ = 0;
           $832 = $831&1;
           if ($832) {
            label = 274;
            break L333;
           } else {
            $$0$i$i2$i$i345 = $830;
           }
          } else {
           $833 = HEAP8[$823>>0]|0;
           $834 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($833)|0);
           $$0$i$i2$i$i345 = $834;
          }
          $835 = (__ZNSt3__211char_traitsIcE3eofEv()|0);
          $836 = (__ZNSt3__211char_traitsIcE11eq_int_typeEii($$0$i$i2$i$i345,$835)|0);
          if ($836) {
           HEAP32[$1>>2] = 0;
           label = 265;
           break;
          } else {
           if ($910) {
            break;
           } else {
            label = 273;
            break L333;
           }
          }
         }
        } while(0);
        if ((label|0) == 265) {
         label = 0;
         if ($910) {
          label = 273;
          break;
         }
        }
        $837 = HEAP32[$0>>2]|0;
        $838 = ((($837)) + 12|0);
        $839 = HEAP32[$838>>2]|0;
        $840 = ((($837)) + 16|0);
        $841 = HEAP32[$840>>2]|0;
        $842 = ($839|0)==($841|0);
        if ($842) {
         $843 = HEAP32[$837>>2]|0;
         $844 = ((($843)) + 36|0);
         $845 = HEAP32[$844>>2]|0;
         __THREW__ = 0;
         $846 = (invoke_ii($845|0,($837|0))|0);
         $847 = __THREW__; __THREW__ = 0;
         $848 = $847&1;
         if ($848) {
          label = 274;
          break;
         } else {
          $$0$i$i249 = $846;
         }
        } else {
         $849 = HEAP8[$839>>0]|0;
         $850 = (__ZNSt3__211char_traitsIcE11to_int_typeEc($849)|0);
         $$0$i$i249 = $850;
        }
        $851 = $$0$i$i249&255;
        $852 = HEAP8[$795>>0]|0;
        $853 = ($852<<24>>24)<(0);
        if ($853) {
         $854 = HEAP32[$$0149>>2]|0;
         $856 = $854;
        } else {
         $856 = $$0149;
        }
        $855 = (($856) + ($$0)|0);
        $857 = HEAP8[$855>>0]|0;
        $858 = ($851<<24>>24)==($857<<24>>24);
        if (!($858)) {
         label = 273;
         break;
        }
        $863 = (($$0) + 1)|0;
        $864 = HEAP32[$0>>2]|0;
        $865 = ((($864)) + 12|0);
        $866 = HEAP32[$865>>2]|0;
        $867 = ((($864)) + 16|0);
        $868 = HEAP32[$867>>2]|0;
        $869 = ($866|0)==($868|0);
        if ($869) {
         $870 = HEAP32[$864>>2]|0;
         $871 = ((($870)) + 40|0);
         $872 = HEAP32[$871>>2]|0;
         __THREW__ = 0;
         (invoke_ii($872|0,($864|0))|0);
         $873 = __THREW__; __THREW__ = 0;
         $874 = $873&1;
         if ($874) {
          label = 274;
          break;
         } else {
          $$0 = $863;
          continue;
         }
        } else {
         $875 = ((($866)) + 1|0);
         HEAP32[$865>>2] = $875;
         $876 = HEAP8[$866>>0]|0;
         (__ZNSt3__211char_traitsIcE11to_int_typeEc($876)|0);
         $$0 = $863;
         continue;
        }
       }
       if ((label|0) == 273) {
        $859 = HEAP32[$5>>2]|0;
        $860 = $859 | 4;
        HEAP32[$5>>2] = $860;
        $$10 = 0;
        break L329;
       }
       else if ((label|0) == 274) {
        $861 = ___cxa_find_matching_catch_2()|0;
        $862 = tempRet0;
        $$5142 = $861;$$5148 = $862;
        break L21;
       }
      }
     } while(0);
     $877 = HEAP32[$13>>2]|0;
     $878 = HEAP32[$14>>2]|0;
     $879 = ($877|0)==($878|0);
     if ($879) {
      $$10 = 1;
     } else {
      HEAP32[$25>>2] = 0;
      __THREW__ = 0;
      invoke_viiii(963,($19|0),($877|0),($878|0),($25|0));
      $880 = __THREW__; __THREW__ = 0;
      $881 = $880&1;
      if ($881) {
       $886 = ___cxa_find_matching_catch_2()|0;
       $887 = tempRet0;
       $$5142 = $886;$$5148 = $887;
       break L21;
      }
      $882 = HEAP32[$25>>2]|0;
      $883 = ($882|0)==(0);
      if ($883) {
       $$10 = 1;
       break;
      } else {
       $884 = HEAP32[$5>>2]|0;
       $885 = $884 | 4;
       HEAP32[$5>>2] = $885;
       $$10 = 0;
       break;
      }
     }
    }
   } while(0);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($22);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19);
   $888 = HEAP32[$13>>2]|0;
   HEAP32[$13>>2] = 0;
   $889 = ($888|0)==(0|0);
   if (!($889)) {
    $890 = HEAP32[$27>>2]|0;
    __THREW__ = 0;
    invoke_vi($890|0,($888|0));
    $891 = __THREW__; __THREW__ = 0;
    $892 = $891&1;
    if ($892) {
     $893 = ___cxa_find_matching_catch_3(0|0)|0;
     $894 = tempRet0;
     ___clang_call_terminate($893);
     // unreachable;
    }
   }
   STACKTOP = sp;return ($$10|0);
  }
 } while(0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($23);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($22);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($21);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($20);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19);
 $895 = HEAP32[$13>>2]|0;
 HEAP32[$13>>2] = 0;
 $896 = ($895|0)==(0|0);
 if (!($896)) {
  $897 = HEAP32[$27>>2]|0;
  __THREW__ = 0;
  invoke_vi($897|0,($895|0));
  $898 = __THREW__; __THREW__ = 0;
  $899 = $898&1;
  if ($899) {
   $900 = ___cxa_find_matching_catch_3(0|0)|0;
   $901 = tempRet0;
   ___clang_call_terminate($900);
   // unreachable;
  }
 }
 ___resumeException($$5142|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE23__append_forward_unsafeIPcEERS5_T_S9_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$040 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, $scevgep = 0, $scevgep23$i$i = 0, $scevgep45 = 0, $scevgep46 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $1;
 $4 = sp + 13|0;
 $5 = sp;
 $6 = sp + 12|0;
 $7 = ((($0)) + 11|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)<(0);
 if ($9) {
  $11 = ((($0)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($0)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = $14 & 2147483647;
  $phitmp$i = (($15) + -1)|0;
  $55 = $12;$56 = $phitmp$i;
 } else {
  $10 = $8&255;
  $55 = $10;$56 = 10;
 }
 $16 = $2;
 $17 = (($16) - ($3))|0;
 $18 = ($17|0)==(0);
 do {
  if (!($18)) {
   if ($9) {
    $19 = HEAP32[$0>>2]|0;
    $20 = ((($0)) + 4|0);
    $21 = HEAP32[$20>>2]|0;
    $24 = $19;$25 = $21;$26 = $19;
   } else {
    $22 = $8&255;
    $24 = $0;$25 = $22;$26 = $0;
   }
   $23 = (($24) + ($25)|0);
   $27 = (__ZNSt3__214__ptr_in_rangeIcEEbPKT_S3_S3_($1,$26,$23)|0);
   if ($27) {
    ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
    $28 = ($17>>>0)>(4294967279);
    if ($28) {
     __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($5);
     // unreachable;
    }
    $29 = ($17>>>0)<(11);
    if ($29) {
     $30 = $17&255;
     $31 = ((($5)) + 11|0);
     HEAP8[$31>>0] = $30;
     $$021$ph$i$i = $5;
    } else {
     $32 = (($17) + 16)|0;
     $33 = $32 & -16;
     $34 = (__Znwj($33)|0);
     HEAP32[$5>>2] = $34;
     $35 = $33 | -2147483648;
     $36 = ((($5)) + 8|0);
     HEAP32[$36>>2] = $35;
     $37 = ((($5)) + 4|0);
     HEAP32[$37>>2] = $17;
     $$021$ph$i$i = $34;
    }
    $$0$i$i = $1;$$021$i$i = $$021$ph$i$i;
    while(1) {
     $38 = ($$0$i$i|0)==($2|0);
     if ($38) {
      break;
     }
     __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
     $39 = ((($$0$i$i)) + 1|0);
     $40 = ((($$021$i$i)) + 1|0);
     $$0$i$i = $39;$$021$i$i = $40;
    }
    $scevgep23$i$i = (($$021$ph$i$i) + ($17)|0);
    HEAP8[$4>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$4);
    $41 = ((($5)) + 11|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = ($42<<24>>24)<(0);
    $44 = HEAP32[$5>>2]|0;
    $45 = ((($5)) + 4|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = $42&255;
    $48 = $43 ? $44 : $5;
    $49 = $43 ? $46 : $47;
    __THREW__ = 0;
    (invoke_iiii(1011,($0|0),($48|0),($49|0))|0);
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     $52 = ___cxa_find_matching_catch_2()|0;
     $53 = tempRet0;
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
     ___resumeException($52|0);
     // unreachable;
    } else {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
     break;
    }
   } else {
    $54 = (($56) - ($55))|0;
    $57 = ($54>>>0)<($17>>>0);
    $58 = (($55) + ($17))|0;
    if ($57) {
     $59 = (($58) - ($56))|0;
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($0,$56,$59,$55,$55,0,0);
    }
    $60 = HEAP8[$7>>0]|0;
    $61 = ($60<<24>>24)<(0);
    if ($61) {
     $62 = HEAP32[$0>>2]|0;
     $64 = $62;
    } else {
     $64 = $0;
    }
    $63 = (($64) + ($55)|0);
    $65 = (($55) - ($3))|0;
    $scevgep = (($2) + ($65)|0);
    $scevgep45 = $scevgep;
    $$0 = $63;$$040 = $1;
    while(1) {
     $66 = ($$040|0)==($2|0);
     if ($66) {
      break;
     }
     __ZNSt3__211char_traitsIcE6assignERcRKc($$0,$$040);
     $67 = ((($$0)) + 1|0);
     $68 = ((($$040)) + 1|0);
     $$0 = $67;$$040 = $68;
    }
    $scevgep46 = (($64) + ($scevgep45)|0);
    HEAP8[$6>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep46,$6);
    $69 = HEAP8[$7>>0]|0;
    $70 = ($69<<24>>24)<(0);
    if ($70) {
     $71 = ((($0)) + 4|0);
     HEAP32[$71>>2] = $58;
     break;
    } else {
     $72 = $58&255;
     HEAP8[$7>>0] = $72;
     break;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__214__ptr_in_rangeIcEEbPKT_S3_S3_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $3 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0>>>0)<($2>>>0);
 $not$ = ($1>>>0)<=($0>>>0);
 $$ = $not$ & $3;
 return ($$|0);
}
function __ZNSt3__211__money_getIcE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_SF_Ri($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$i$i = 0, $$0$i$i$i40 = 0, $$0$i$i$i45 = 0, $$0$i$i$i50 = 0, $$0$i$i$i55 = 0, $$0$i$i$i60 = 0, $$0$i$i$i65 = 0, $$0$i$i$i70 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $exitcond$i$i$i = 0, $exitcond$i$i$i41 = 0, $exitcond$i$i$i46 = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i56 = 0, $exitcond$i$i$i61 = 0, $exitcond$i$i$i66 = 0, $exitcond$i$i$i71 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $10 = sp + 119|0;
 $11 = sp + 118|0;
 $12 = sp + 117|0;
 $13 = sp + 116|0;
 $14 = sp + 115|0;
 $15 = sp + 114|0;
 $16 = sp + 113|0;
 $17 = sp + 112|0;
 $18 = sp + 111|0;
 $19 = sp + 110|0;
 $20 = sp + 109|0;
 $21 = sp + 108|0;
 $22 = sp + 107|0;
 $23 = sp + 106|0;
 $24 = sp + 105|0;
 $25 = sp + 104|0;
 $26 = sp + 100|0;
 $27 = sp + 88|0;
 $28 = sp + 76|0;
 $29 = sp + 64|0;
 $30 = sp + 52|0;
 $31 = sp + 48|0;
 $32 = sp + 36|0;
 $33 = sp + 24|0;
 $34 = sp + 12|0;
 $35 = sp;
 if ($0) {
  $36 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167160)|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($37)) + 44|0);
  $39 = HEAP32[$38>>2]|0;
  FUNCTION_TABLE_vii[$39 & 2047]($26,$36);
  $40 = HEAP32[$26>>2]|0;
  HEAP8[$2>>0]=$40&255;HEAP8[$2+1>>0]=($40>>8)&255;HEAP8[$2+2>>0]=($40>>16)&255;HEAP8[$2+3>>0]=$40>>24;
  $41 = HEAP32[$36>>2]|0;
  $42 = ((($41)) + 32|0);
  $43 = HEAP32[$42>>2]|0;
  FUNCTION_TABLE_vii[$43 & 2047]($27,$36);
  $44 = ((($8)) + 11|0);
  $45 = HEAP8[$44>>0]|0;
  $46 = ($45<<24>>24)<(0);
  if ($46) {
   $47 = HEAP32[$8>>2]|0;
   HEAP8[$24>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($47,$24);
   $48 = ((($8)) + 4|0);
   HEAP32[$48>>2] = 0;
  } else {
   HEAP8[$25>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($8,$25);
   HEAP8[$44>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($8|0),0);
  $49 = __THREW__; __THREW__ = 0;
  $50 = $49&1;
  if ($50) {
   $51 = ___cxa_find_matching_catch_3(0|0)|0;
   $52 = tempRet0;
   ___clang_call_terminate($51);
   // unreachable;
  }
  ;HEAP32[$8>>2]=HEAP32[$27>>2]|0;HEAP32[$8+4>>2]=HEAP32[$27+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$27+8>>2]|0;
  $$0$i$i$i = 0;
  while(1) {
   $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
   if ($exitcond$i$i$i) {
    break;
   }
   $53 = (($27) + ($$0$i$i$i<<2)|0);
   HEAP32[$53>>2] = 0;
   $54 = (($$0$i$i$i) + 1)|0;
   $$0$i$i$i = $54;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27);
  $55 = HEAP32[$36>>2]|0;
  $56 = ((($55)) + 28|0);
  $57 = HEAP32[$56>>2]|0;
  FUNCTION_TABLE_vii[$57 & 2047]($28,$36);
  $58 = ((($7)) + 11|0);
  $59 = HEAP8[$58>>0]|0;
  $60 = ($59<<24>>24)<(0);
  if ($60) {
   $61 = HEAP32[$7>>2]|0;
   HEAP8[$22>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($61,$22);
   $62 = ((($7)) + 4|0);
   HEAP32[$62>>2] = 0;
  } else {
   HEAP8[$23>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($7,$23);
   HEAP8[$58>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($7|0),0);
  $63 = __THREW__; __THREW__ = 0;
  $64 = $63&1;
  if ($64) {
   $65 = ___cxa_find_matching_catch_3(0|0)|0;
   $66 = tempRet0;
   ___clang_call_terminate($65);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$28>>2]|0;HEAP32[$7+4>>2]=HEAP32[$28+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$28+8>>2]|0;
  $$0$i$i$i40 = 0;
  while(1) {
   $exitcond$i$i$i41 = ($$0$i$i$i40|0)==(3);
   if ($exitcond$i$i$i41) {
    break;
   }
   $67 = (($28) + ($$0$i$i$i40<<2)|0);
   HEAP32[$67>>2] = 0;
   $68 = (($$0$i$i$i40) + 1)|0;
   $$0$i$i$i40 = $68;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($28);
  $69 = HEAP32[$36>>2]|0;
  $70 = ((($69)) + 12|0);
  $71 = HEAP32[$70>>2]|0;
  $72 = (FUNCTION_TABLE_ii[$71 & 2047]($36)|0);
  HEAP8[$3>>0] = $72;
  $73 = HEAP32[$36>>2]|0;
  $74 = ((($73)) + 16|0);
  $75 = HEAP32[$74>>2]|0;
  $76 = (FUNCTION_TABLE_ii[$75 & 2047]($36)|0);
  HEAP8[$4>>0] = $76;
  $77 = HEAP32[$36>>2]|0;
  $78 = ((($77)) + 20|0);
  $79 = HEAP32[$78>>2]|0;
  FUNCTION_TABLE_vii[$79 & 2047]($29,$36);
  $80 = ((($5)) + 11|0);
  $81 = HEAP8[$80>>0]|0;
  $82 = ($81<<24>>24)<(0);
  if ($82) {
   $83 = HEAP32[$5>>2]|0;
   HEAP8[$20>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($83,$20);
   $84 = ((($5)) + 4|0);
   HEAP32[$84>>2] = 0;
  } else {
   HEAP8[$21>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($5,$21);
   HEAP8[$80>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($5|0),0);
  $85 = __THREW__; __THREW__ = 0;
  $86 = $85&1;
  if ($86) {
   $87 = ___cxa_find_matching_catch_3(0|0)|0;
   $88 = tempRet0;
   ___clang_call_terminate($87);
   // unreachable;
  }
  ;HEAP32[$5>>2]=HEAP32[$29>>2]|0;HEAP32[$5+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$29+8>>2]|0;
  $$0$i$i$i45 = 0;
  while(1) {
   $exitcond$i$i$i46 = ($$0$i$i$i45|0)==(3);
   if ($exitcond$i$i$i46) {
    break;
   }
   $89 = (($29) + ($$0$i$i$i45<<2)|0);
   HEAP32[$89>>2] = 0;
   $90 = (($$0$i$i$i45) + 1)|0;
   $$0$i$i$i45 = $90;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($29);
  $91 = HEAP32[$36>>2]|0;
  $92 = ((($91)) + 24|0);
  $93 = HEAP32[$92>>2]|0;
  FUNCTION_TABLE_vii[$93 & 2047]($30,$36);
  $94 = ((($6)) + 11|0);
  $95 = HEAP8[$94>>0]|0;
  $96 = ($95<<24>>24)<(0);
  if ($96) {
   $97 = HEAP32[$6>>2]|0;
   HEAP8[$18>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($97,$18);
   $98 = ((($6)) + 4|0);
   HEAP32[$98>>2] = 0;
  } else {
   HEAP8[$19>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$19);
   HEAP8[$94>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $99 = __THREW__; __THREW__ = 0;
  $100 = $99&1;
  if ($100) {
   $101 = ___cxa_find_matching_catch_3(0|0)|0;
   $102 = tempRet0;
   ___clang_call_terminate($101);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$30>>2]|0;HEAP32[$6+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$30+8>>2]|0;
  $$0$i$i$i50 = 0;
  while(1) {
   $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
   if ($exitcond$i$i$i51) {
    break;
   }
   $103 = (($30) + ($$0$i$i$i50<<2)|0);
   HEAP32[$103>>2] = 0;
   $104 = (($$0$i$i$i50) + 1)|0;
   $$0$i$i$i50 = $104;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30);
  $105 = HEAP32[$36>>2]|0;
  $106 = ((($105)) + 36|0);
  $107 = HEAP32[$106>>2]|0;
  $108 = (FUNCTION_TABLE_ii[$107 & 2047]($36)|0);
  $storemerge = $108;
 } else {
  $109 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167152)|0);
  $110 = HEAP32[$109>>2]|0;
  $111 = ((($110)) + 44|0);
  $112 = HEAP32[$111>>2]|0;
  FUNCTION_TABLE_vii[$112 & 2047]($31,$109);
  $113 = HEAP32[$31>>2]|0;
  HEAP8[$2>>0]=$113&255;HEAP8[$2+1>>0]=($113>>8)&255;HEAP8[$2+2>>0]=($113>>16)&255;HEAP8[$2+3>>0]=$113>>24;
  $114 = HEAP32[$109>>2]|0;
  $115 = ((($114)) + 32|0);
  $116 = HEAP32[$115>>2]|0;
  FUNCTION_TABLE_vii[$116 & 2047]($32,$109);
  $117 = ((($8)) + 11|0);
  $118 = HEAP8[$117>>0]|0;
  $119 = ($118<<24>>24)<(0);
  if ($119) {
   $120 = HEAP32[$8>>2]|0;
   HEAP8[$16>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($120,$16);
   $121 = ((($8)) + 4|0);
   HEAP32[$121>>2] = 0;
  } else {
   HEAP8[$17>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($8,$17);
   HEAP8[$117>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($8|0),0);
  $122 = __THREW__; __THREW__ = 0;
  $123 = $122&1;
  if ($123) {
   $124 = ___cxa_find_matching_catch_3(0|0)|0;
   $125 = tempRet0;
   ___clang_call_terminate($124);
   // unreachable;
  }
  ;HEAP32[$8>>2]=HEAP32[$32>>2]|0;HEAP32[$8+4>>2]=HEAP32[$32+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$32+8>>2]|0;
  $$0$i$i$i55 = 0;
  while(1) {
   $exitcond$i$i$i56 = ($$0$i$i$i55|0)==(3);
   if ($exitcond$i$i$i56) {
    break;
   }
   $126 = (($32) + ($$0$i$i$i55<<2)|0);
   HEAP32[$126>>2] = 0;
   $127 = (($$0$i$i$i55) + 1)|0;
   $$0$i$i$i55 = $127;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($32);
  $128 = HEAP32[$109>>2]|0;
  $129 = ((($128)) + 28|0);
  $130 = HEAP32[$129>>2]|0;
  FUNCTION_TABLE_vii[$130 & 2047]($33,$109);
  $131 = ((($7)) + 11|0);
  $132 = HEAP8[$131>>0]|0;
  $133 = ($132<<24>>24)<(0);
  if ($133) {
   $134 = HEAP32[$7>>2]|0;
   HEAP8[$14>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($134,$14);
   $135 = ((($7)) + 4|0);
   HEAP32[$135>>2] = 0;
  } else {
   HEAP8[$15>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($7,$15);
   HEAP8[$131>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($7|0),0);
  $136 = __THREW__; __THREW__ = 0;
  $137 = $136&1;
  if ($137) {
   $138 = ___cxa_find_matching_catch_3(0|0)|0;
   $139 = tempRet0;
   ___clang_call_terminate($138);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$33>>2]|0;HEAP32[$7+4>>2]=HEAP32[$33+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$33+8>>2]|0;
  $$0$i$i$i60 = 0;
  while(1) {
   $exitcond$i$i$i61 = ($$0$i$i$i60|0)==(3);
   if ($exitcond$i$i$i61) {
    break;
   }
   $140 = (($33) + ($$0$i$i$i60<<2)|0);
   HEAP32[$140>>2] = 0;
   $141 = (($$0$i$i$i60) + 1)|0;
   $$0$i$i$i60 = $141;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33);
  $142 = HEAP32[$109>>2]|0;
  $143 = ((($142)) + 12|0);
  $144 = HEAP32[$143>>2]|0;
  $145 = (FUNCTION_TABLE_ii[$144 & 2047]($109)|0);
  HEAP8[$3>>0] = $145;
  $146 = HEAP32[$109>>2]|0;
  $147 = ((($146)) + 16|0);
  $148 = HEAP32[$147>>2]|0;
  $149 = (FUNCTION_TABLE_ii[$148 & 2047]($109)|0);
  HEAP8[$4>>0] = $149;
  $150 = HEAP32[$109>>2]|0;
  $151 = ((($150)) + 20|0);
  $152 = HEAP32[$151>>2]|0;
  FUNCTION_TABLE_vii[$152 & 2047]($34,$109);
  $153 = ((($5)) + 11|0);
  $154 = HEAP8[$153>>0]|0;
  $155 = ($154<<24>>24)<(0);
  if ($155) {
   $156 = HEAP32[$5>>2]|0;
   HEAP8[$12>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($156,$12);
   $157 = ((($5)) + 4|0);
   HEAP32[$157>>2] = 0;
  } else {
   HEAP8[$13>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($5,$13);
   HEAP8[$153>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($5|0),0);
  $158 = __THREW__; __THREW__ = 0;
  $159 = $158&1;
  if ($159) {
   $160 = ___cxa_find_matching_catch_3(0|0)|0;
   $161 = tempRet0;
   ___clang_call_terminate($160);
   // unreachable;
  }
  ;HEAP32[$5>>2]=HEAP32[$34>>2]|0;HEAP32[$5+4>>2]=HEAP32[$34+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$34+8>>2]|0;
  $$0$i$i$i65 = 0;
  while(1) {
   $exitcond$i$i$i66 = ($$0$i$i$i65|0)==(3);
   if ($exitcond$i$i$i66) {
    break;
   }
   $162 = (($34) + ($$0$i$i$i65<<2)|0);
   HEAP32[$162>>2] = 0;
   $163 = (($$0$i$i$i65) + 1)|0;
   $$0$i$i$i65 = $163;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34);
  $164 = HEAP32[$109>>2]|0;
  $165 = ((($164)) + 24|0);
  $166 = HEAP32[$165>>2]|0;
  FUNCTION_TABLE_vii[$166 & 2047]($35,$109);
  $167 = ((($6)) + 11|0);
  $168 = HEAP8[$167>>0]|0;
  $169 = ($168<<24>>24)<(0);
  if ($169) {
   $170 = HEAP32[$6>>2]|0;
   HEAP8[$10>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($170,$10);
   $171 = ((($6)) + 4|0);
   HEAP32[$171>>2] = 0;
  } else {
   HEAP8[$11>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$11);
   HEAP8[$167>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $172 = __THREW__; __THREW__ = 0;
  $173 = $172&1;
  if ($173) {
   $174 = ___cxa_find_matching_catch_3(0|0)|0;
   $175 = tempRet0;
   ___clang_call_terminate($174);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$35>>2]|0;HEAP32[$6+4>>2]=HEAP32[$35+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$35+8>>2]|0;
  $$0$i$i$i70 = 0;
  while(1) {
   $exitcond$i$i$i71 = ($$0$i$i$i70|0)==(3);
   if ($exitcond$i$i$i71) {
    break;
   }
   $176 = (($35) + ($$0$i$i$i70<<2)|0);
   HEAP32[$176>>2] = 0;
   $177 = (($$0$i$i$i70) + 1)|0;
   $$0$i$i$i70 = $177;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35);
  $178 = HEAP32[$109>>2]|0;
  $179 = ((($178)) + 36|0);
  $180 = HEAP32[$179>>2]|0;
  $181 = (FUNCTION_TABLE_ii[$180 & 2047]($109)|0);
  $storemerge = $181;
 }
 HEAP32[$9>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__219__double_or_nothingIcEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$34 = 0, $$pr = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(1003|0);
 $6 = HEAP32[$2>>2]|0;
 $7 = HEAP32[$0>>2]|0;
 $8 = (($6) - ($7))|0;
 $9 = ($8>>>0)<(2147483647);
 $10 = $8 << 1;
 $$34 = $9 ? $10 : -1;
 $11 = ($$34|0)==(0);
 $$ = $11 ? 1 : $$34;
 $12 = HEAP32[$1>>2]|0;
 $13 = $7;
 $14 = (($12) - ($7))|0;
 $15 = $5 ? $13 : 0;
 $16 = (_realloc($15,$$)|0);
 $17 = ($16|0)==(0|0);
 if ($17) {
  __ZSt17__throw_bad_allocv();
  // unreachable;
 }
 do {
  if ($5) {
   $18 = $16;
   HEAP32[$0>>2] = $18;
   $27 = $16;
  } else {
   $$pr = HEAP32[$0>>2]|0;
   $19 = $16;
   HEAP32[$0>>2] = $19;
   $20 = ($$pr|0)==(0|0);
   if ($20) {
    $27 = $16;
   } else {
    $21 = HEAP32[$3>>2]|0;
    __THREW__ = 0;
    invoke_vi($21|0,($$pr|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $24 = ___cxa_find_matching_catch_3(0|0)|0;
     $25 = tempRet0;
     ___clang_call_terminate($24);
     // unreachable;
    } else {
     $$pre = HEAP32[$0>>2]|0;
     $27 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$3>>2] = (730);
 $26 = (($27) + ($14)|0);
 HEAP32[$1>>2] = $26;
 $28 = HEAP32[$0>>2]|0;
 $29 = (($28) + ($$)|0);
 HEAP32[$2>>2] = $29;
 return;
}
function __ZNSt3__219__double_or_nothingIjEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$34 = 0, $$pr = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(1003|0);
 $6 = HEAP32[$2>>2]|0;
 $7 = HEAP32[$0>>2]|0;
 $8 = (($6) - ($7))|0;
 $9 = ($8>>>0)<(2147483647);
 $10 = $8 << 1;
 $$34 = $9 ? $10 : -1;
 $11 = ($$34|0)==(0);
 $$ = $11 ? 4 : $$34;
 $12 = HEAP32[$1>>2]|0;
 $13 = (($12) - ($7))|0;
 $14 = $13 >> 2;
 $phitmp = $7;
 $15 = $5 ? $phitmp : 0;
 $16 = (_realloc($15,$$)|0);
 $17 = ($16|0)==(0|0);
 if ($17) {
  __ZSt17__throw_bad_allocv();
  // unreachable;
 }
 do {
  if ($5) {
   $18 = $16;
   HEAP32[$0>>2] = $18;
   $28 = $16;
  } else {
   $$pr = HEAP32[$0>>2]|0;
   $19 = $16;
   HEAP32[$0>>2] = $19;
   $20 = ($$pr|0)==(0|0);
   if ($20) {
    $28 = $16;
   } else {
    $21 = HEAP32[$3>>2]|0;
    __THREW__ = 0;
    invoke_vi($21|0,($$pr|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $24 = ___cxa_find_matching_catch_3(0|0)|0;
     $25 = tempRet0;
     ___clang_call_terminate($24);
     // unreachable;
    } else {
     $$pre = HEAP32[$0>>2]|0;
     $28 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$3>>2] = (730);
 $26 = $$ >>> 2;
 $27 = (($28) + ($14<<2)|0);
 HEAP32[$1>>2] = $27;
 $29 = HEAP32[$0>>2]|0;
 $30 = (($29) + ($26<<2)|0);
 HEAP32[$2>>2] = $30;
 return;
}
function __ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i = 0, $$0$i$i$i$i = 0, $$0$i$i2$i$i = 0, $$0$lcssa$i = 0, $$031 = 0, $$032 = 0, $$035 = 0, $$1 = 0, $$133 = 0, $$136 = 0, $$2 = 0, $$234 = 0, $$237 = 0, $$33859 = 0, $$33860 = 0, $$361 = 0, $$362 = 0, $$byval_copy = 0, $$pre = 0;
 var $$pre$i$i$i = 0, $$pre56 = 0, $$pre63 = 0, $$sroa$0$0$copyload = 0, $$sroa$043$0 = 0, $$sroa$043$1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0;
 var $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0;
 var $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(592|0);
 $$byval_copy = sp + 472|0;
 $vararg_buffer = sp;
 $7 = sp + 72|0;
 $8 = sp + 64|0;
 $9 = sp + 56|0;
 $10 = sp + 52|0;
 $11 = sp + 576|0;
 $12 = sp + 48|0;
 $13 = sp + 8|0;
 $14 = sp + 476|0;
 $15 = $7;
 HEAP32[$8>>2] = $15;
 $16 = ((($8)) + 4|0);
 HEAP32[$16>>2] = (1003);
 $17 = ((($7)) + 400|0);
 __THREW__ = 0;
 invoke_vii(733,($10|0),($4|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $46 = ___cxa_find_matching_catch_2()|0;
  $47 = tempRet0;
  HEAP32[$8>>2] = 0;
  $$33859 = $47;$$361 = $46;$126 = 1003;$127 = $7;
  label = 54;
 } else {
  __THREW__ = 0;
  $20 = (invoke_iii(734,($10|0),(165472|0))|0);
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $48 = ___cxa_find_matching_catch_2()|0;
   $49 = tempRet0;
   $$234 = $48;$$237 = $49;
  } else {
   HEAP8[$11>>0] = 0;
   $23 = HEAP32[$2>>2]|0;
   HEAP32[$12>>2] = $23;
   $24 = ((($4)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   __THREW__ = 0;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$12>>2]|0;
   $26 = (invoke_iiiiiiiiiiii(1013,($1|0),($$byval_copy|0),($3|0),($10|0),($25|0),($5|0),($11|0),($20|0),($8|0),($9|0),($17|0))|0);
   $27 = __THREW__; __THREW__ = 0;
   $28 = $27&1;
   L7: do {
    if ($28) {
     label = 11;
    } else {
     L9: do {
      if ($26) {
       $29 = HEAP32[$20>>2]|0;
       $30 = ((($29)) + 48|0);
       $31 = HEAP32[$30>>2]|0;
       __THREW__ = 0;
       (invoke_iiiii($31|0,($20|0),(157270|0),((157280)|0),($13|0))|0);
       $32 = __THREW__; __THREW__ = 0;
       $33 = $32&1;
       if ($33) {
        $52 = ___cxa_find_matching_catch_2()|0;
        $53 = tempRet0;
        $$032 = $52;$$035 = $53;
       } else {
        $34 = HEAP32[$9>>2]|0;
        $35 = HEAP32[$8>>2]|0;
        $36 = (($34) - ($35))|0;
        $37 = ($36|0)>(392);
        $38 = $35;
        $39 = $34;
        if ($37) {
         $40 = $36 >>> 2;
         $41 = (($40) + 2)|0;
         $42 = (_malloc($41)|0);
         $43 = $42;
         $44 = ($42|0)==(0|0);
         if ($44) {
          __THREW__ = 0;
          invoke_v(732);
          $45 = __THREW__; __THREW__ = 0;
          $$sroa$043$0 = $43;
         } else {
          $$031 = $42;$$sroa$043$1 = $43;
          label = 16;
         }
        } else {
         $$031 = $14;$$sroa$043$1 = 0;
         label = 16;
        }
        do {
         if ((label|0) == 16) {
          $58 = HEAP8[$11>>0]|0;
          $59 = ($58<<24>>24)==(0);
          if ($59) {
           $$1 = $$031;
          } else {
           $60 = ((($$031)) + 1|0);
           HEAP8[$$031>>0] = 45;
           $$1 = $60;
          }
          $61 = ((($13)) + 40|0);
          $62 = $13;
          $$0 = $38;$$2 = $$1;$64 = $39;
          while(1) {
           $63 = ($$0>>>0)<($64>>>0);
           if (!($63)) {
            break;
           }
           $67 = HEAP32[$$0>>2]|0;
           $$0$i = $13;
           while(1) {
            $68 = ($$0$i|0)==($61|0);
            if ($68) {
             $$0$lcssa$i = $61;
             break;
            }
            $69 = HEAP32[$$0$i>>2]|0;
            $70 = ($69|0)==($67|0);
            if ($70) {
             $$0$lcssa$i = $$0$i;
             break;
            }
            $71 = ((($$0$i)) + 4|0);
            $$0$i = $71;
           }
           $72 = $$0$lcssa$i;
           $73 = (($72) - ($62))|0;
           $74 = $73 >> 2;
           $75 = (157270 + ($74)|0);
           $76 = HEAP8[$75>>0]|0;
           HEAP8[$$2>>0] = $76;
           $77 = ((($$0)) + 4|0);
           $78 = ((($$2)) + 1|0);
           $$pre = HEAP32[$9>>2]|0;
           $$0 = $77;$$2 = $78;$64 = $$pre;
          }
          HEAP8[$$2>>0] = 0;
          HEAP32[$vararg_buffer>>2] = $6;
          $65 = (_sscanf($14,157155,$vararg_buffer)|0);
          $66 = ($65|0)==(1);
          if (!($66)) {
           __THREW__ = 0;
           invoke_vi(1005,(157159|0));
           $79 = __THREW__; __THREW__ = 0;
           $$sroa$043$0 = $$sroa$043$1;
           break;
          }
          $80 = ($$sroa$043$1|0)==(0);
          if (!($80)) {
           $81 = $$sroa$043$1;
           _free($81);
          }
          break L9;
         }
        } while(0);
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        $56 = ($$sroa$043$0|0)==(0);
        if (!($56)) {
         $57 = $$sroa$043$0;
         _free($57);
        }
        $$032 = $54;$$035 = $55;
       }
       $$133 = $$032;$$136 = $$035;
       break L7;
      }
     } while(0);
     $82 = HEAP32[$1>>2]|0;
     $83 = ($82|0)==(0|0);
     do {
      if ($83) {
       $132 = 1;
      } else {
       $84 = ((($82)) + 12|0);
       $85 = HEAP32[$84>>2]|0;
       $86 = ((($82)) + 16|0);
       $87 = HEAP32[$86>>2]|0;
       $88 = ($85|0)==($87|0);
       if ($88) {
        $89 = HEAP32[$82>>2]|0;
        $90 = ((($89)) + 36|0);
        $91 = HEAP32[$90>>2]|0;
        __THREW__ = 0;
        $92 = (invoke_ii($91|0,($82|0))|0);
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         label = 11;
         break L7;
        } else {
         $$0$i$i$i$i = $92;
        }
       } else {
        $95 = HEAP32[$85>>2]|0;
        $96 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($95)|0);
        $$0$i$i$i$i = $96;
       }
       $97 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $98 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$97)|0);
       if ($98) {
        HEAP32[$1>>2] = 0;
        $132 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$1>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $132 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $99 = HEAP32[$2>>2]|0;
     $100 = ($99|0)==(0|0);
     do {
      if ($100) {
       label = 44;
      } else {
       $101 = ((($99)) + 12|0);
       $102 = HEAP32[$101>>2]|0;
       $103 = ((($99)) + 16|0);
       $104 = HEAP32[$103>>2]|0;
       $105 = ($102|0)==($104|0);
       if ($105) {
        $106 = HEAP32[$99>>2]|0;
        $107 = ((($106)) + 36|0);
        $108 = HEAP32[$107>>2]|0;
        __THREW__ = 0;
        $109 = (invoke_ii($108|0,($99|0))|0);
        $110 = __THREW__; __THREW__ = 0;
        $111 = $110&1;
        if ($111) {
         label = 11;
         break L7;
        } else {
         $$0$i$i2$i$i = $109;
        }
       } else {
        $112 = HEAP32[$102>>2]|0;
        $113 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($112)|0);
        $$0$i$i2$i$i = $113;
       }
       $114 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $115 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$114)|0);
       if ($115) {
        HEAP32[$2>>2] = 0;
        label = 44;
        break;
       } else {
        if ($132) {
         break;
        } else {
         label = 46;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 44) {
      if ($132) {
       label = 46;
      }
     }
     if ((label|0) == 46) {
      $116 = HEAP32[$5>>2]|0;
      $117 = $116 | 2;
      HEAP32[$5>>2] = $117;
     }
     $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
     __ZNSt3__26localeD2Ev($10);
     $118 = HEAP32[$8>>2]|0;
     HEAP32[$8>>2] = 0;
     $119 = ($118|0)==(0|0);
     if (!($119)) {
      $120 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      invoke_vi($120|0,($118|0));
      $121 = __THREW__; __THREW__ = 0;
      $122 = $121&1;
      if ($122) {
       $123 = ___cxa_find_matching_catch_3(0|0)|0;
       $124 = tempRet0;
       ___clang_call_terminate($123);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($$sroa$0$0$copyload|0);
    }
   } while(0);
   if ((label|0) == 11) {
    $50 = ___cxa_find_matching_catch_2()|0;
    $51 = tempRet0;
    $$133 = $50;$$136 = $51;
   }
   $$234 = $$133;$$237 = $$136;
  }
  __ZNSt3__26localeD2Ev($10);
  $$pre56 = HEAP32[$8>>2]|0;
  HEAP32[$8>>2] = 0;
  $125 = ($$pre56|0)==(0|0);
  if ($125) {
   $$33860 = $$237;$$362 = $$234;
  } else {
   $$pre63 = HEAP32[$16>>2]|0;
   $$33859 = $$237;$$361 = $$234;$126 = $$pre63;$127 = $$pre56;
   label = 54;
  }
 }
 if ((label|0) == 54) {
  __THREW__ = 0;
  invoke_vi($126|0,($127|0));
  $128 = __THREW__; __THREW__ = 0;
  $129 = $128&1;
  if ($129) {
   $130 = ___cxa_find_matching_catch_3(0|0)|0;
   $131 = tempRet0;
   ___clang_call_terminate($130);
   // unreachable;
  } else {
   $$33860 = $$33859;$$362 = $$361;
  }
 }
 ___resumeException($$362|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIwS3_NS_9allocatorIwEEEE($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$0 = 0, $$0$i$i$i$i = 0, $$0$i$i2$i$i = 0, $$1 = 0, $$135 = 0, $$23646 = 0, $$23647 = 0, $$248 = 0, $$249 = 0, $$byval_copy = 0, $$pre = 0, $$pre$i$i$i = 0, $$sroa$0$0$copyload = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $phitmp$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 448|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(448|0);
 $$byval_copy = sp + 432|0;
 $7 = sp + 428|0;
 $8 = sp + 424|0;
 $9 = sp + 24|0;
 $10 = sp + 16|0;
 $11 = sp + 8|0;
 $12 = sp + 4|0;
 $13 = sp + 436|0;
 $14 = sp;
 $15 = $9;
 HEAP32[$10>>2] = $15;
 $16 = ((($10)) + 4|0);
 HEAP32[$16>>2] = (1003);
 $17 = ((($9)) + 400|0);
 __THREW__ = 0;
 invoke_vii(733,($12|0),($4|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 if ($19) {
  $46 = ___cxa_find_matching_catch_2()|0;
  $47 = tempRet0;
  HEAP32[$10>>2] = 0;
  $$23646 = $47;$$248 = $46;$112 = $9;
  label = 41;
 } else {
  __THREW__ = 0;
  $20 = (invoke_iii(734,($12|0),(165472|0))|0);
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   $48 = ___cxa_find_matching_catch_2()|0;
   $49 = tempRet0;
   $$1 = $48;$$135 = $49;
  } else {
   HEAP8[$13>>0] = 0;
   $23 = HEAP32[$2>>2]|0;
   HEAP32[$14>>2] = $23;
   $24 = ((($4)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $23;
   __THREW__ = 0;
   ;HEAP32[$$byval_copy>>2]=HEAP32[$14>>2]|0;
   $27 = (invoke_iiiiiiiiiiii(1013,($1|0),($$byval_copy|0),($3|0),($12|0),($25|0),($5|0),($13|0),($20|0),($10|0),($11|0),($17|0))|0);
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   L7: do {
    if (!($29)) {
     if ($27) {
      $30 = ((($6)) + 8|0);
      $31 = ((($30)) + 3|0);
      $32 = HEAP8[$31>>0]|0;
      $33 = ($32<<24>>24)<(0);
      if ($33) {
       $34 = HEAP32[$6>>2]|0;
       HEAP32[$7>>2] = 0;
       __ZNSt3__211char_traitsIwE6assignERwRKw($34,$7);
       $35 = ((($6)) + 4|0);
       HEAP32[$35>>2] = 0;
      } else {
       HEAP32[$8>>2] = 0;
       __ZNSt3__211char_traitsIwE6assignERwRKw($6,$8);
       HEAP8[$31>>0] = 0;
      }
      $36 = HEAP8[$13>>0]|0;
      $37 = ($36<<24>>24)==(0);
      if (!($37)) {
       $38 = HEAP32[$20>>2]|0;
       $39 = ((($38)) + 44|0);
       $40 = HEAP32[$39>>2]|0;
       __THREW__ = 0;
       $41 = (invoke_iii($40|0,($20|0),45)|0);
       $42 = __THREW__; __THREW__ = 0;
       $43 = $42&1;
       if ($43) {
        break;
       }
       __THREW__ = 0;
       invoke_vii(1014,($6|0),($41|0));
       $44 = __THREW__; __THREW__ = 0;
       $45 = $44&1;
       if ($45) {
        break;
       }
      }
      $50 = HEAP32[$20>>2]|0;
      $51 = ((($50)) + 44|0);
      $52 = HEAP32[$51>>2]|0;
      __THREW__ = 0;
      $53 = (invoke_iii($52|0,($20|0),48)|0);
      $54 = __THREW__; __THREW__ = 0;
      $55 = $54&1;
      if ($55) {
       break;
      }
      $56 = HEAP32[$10>>2]|0;
      $57 = HEAP32[$11>>2]|0;
      $58 = ((($57)) + -4|0);
      $$0 = $56;
      while(1) {
       $59 = ($$0>>>0)<($58>>>0);
       if (!($59)) {
        break;
       }
       $60 = HEAP32[$$0>>2]|0;
       $61 = ($60|0)==($53|0);
       if (!($61)) {
        break;
       }
       $62 = ((($$0)) + 4|0);
       $$0 = $62;
      }
      __THREW__ = 0;
      (invoke_iiii(1015,($6|0),($$0|0),($57|0))|0);
      $63 = __THREW__; __THREW__ = 0;
      $64 = $63&1;
      if ($64) {
       break;
      }
     }
     $65 = HEAP32[$1>>2]|0;
     $66 = ($65|0)==(0|0);
     do {
      if ($66) {
       $117 = 1;
      } else {
       $67 = ((($65)) + 12|0);
       $68 = HEAP32[$67>>2]|0;
       $69 = ((($65)) + 16|0);
       $70 = HEAP32[$69>>2]|0;
       $71 = ($68|0)==($70|0);
       if ($71) {
        $72 = HEAP32[$65>>2]|0;
        $73 = ((($72)) + 36|0);
        $74 = HEAP32[$73>>2]|0;
        __THREW__ = 0;
        $75 = (invoke_ii($74|0,($65|0))|0);
        $76 = __THREW__; __THREW__ = 0;
        $77 = $76&1;
        if ($77) {
         break L7;
        } else {
         $$0$i$i$i$i = $75;
        }
       } else {
        $78 = HEAP32[$68>>2]|0;
        $79 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($78)|0);
        $$0$i$i$i$i = $79;
       }
       $80 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $81 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$80)|0);
       if ($81) {
        HEAP32[$1>>2] = 0;
        $117 = 1;
        break;
       } else {
        $$pre$i$i$i = HEAP32[$1>>2]|0;
        $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
        $117 = $phitmp$i$i$i;
        break;
       }
      }
     } while(0);
     $82 = ($23|0)==(0);
     do {
      if ($82) {
       label = 32;
      } else {
       $83 = ((($26)) + 12|0);
       $84 = HEAP32[$83>>2]|0;
       $85 = ((($26)) + 16|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = ($84|0)==($86|0);
       if ($87) {
        $88 = $23;
        $89 = HEAP32[$88>>2]|0;
        $90 = ((($89)) + 36|0);
        $91 = HEAP32[$90>>2]|0;
        __THREW__ = 0;
        $92 = (invoke_ii($91|0,($26|0))|0);
        $93 = __THREW__; __THREW__ = 0;
        $94 = $93&1;
        if ($94) {
         break L7;
        } else {
         $$0$i$i2$i$i = $92;
        }
       } else {
        $95 = HEAP32[$84>>2]|0;
        $96 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($95)|0);
        $$0$i$i2$i$i = $96;
       }
       $97 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
       $98 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$97)|0);
       if ($98) {
        HEAP32[$2>>2] = 0;
        label = 32;
        break;
       } else {
        if ($117) {
         break;
        } else {
         label = 34;
         break;
        }
       }
      }
     } while(0);
     if ((label|0) == 32) {
      if ($117) {
       label = 34;
      }
     }
     if ((label|0) == 34) {
      $99 = HEAP32[$5>>2]|0;
      $100 = $99 | 2;
      HEAP32[$5>>2] = $100;
     }
     $$sroa$0$0$copyload = HEAP32[$1>>2]|0;
     __ZNSt3__26localeD2Ev($12);
     $101 = HEAP32[$10>>2]|0;
     HEAP32[$10>>2] = 0;
     $102 = ($101|0)==(0|0);
     if (!($102)) {
      $103 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      invoke_vi($103|0,($101|0));
      $104 = __THREW__; __THREW__ = 0;
      $105 = $104&1;
      if ($105) {
       $106 = ___cxa_find_matching_catch_3(0|0)|0;
       $107 = tempRet0;
       ___clang_call_terminate($106);
       // unreachable;
      }
     }
     STACKTOP = sp;return ($$sroa$0$0$copyload|0);
    }
   } while(0);
   $108 = ___cxa_find_matching_catch_2()|0;
   $109 = tempRet0;
   $$1 = $108;$$135 = $109;
  }
  __ZNSt3__26localeD2Ev($12);
  $$pre = HEAP32[$10>>2]|0;
  HEAP32[$10>>2] = 0;
  $110 = ($$pre|0)==(0|0);
  if ($110) {
   $$23647 = $$135;$$249 = $$1;
  } else {
   $$23646 = $$135;$$248 = $$1;$112 = $$pre;
   label = 41;
  }
 }
 if ((label|0) == 41) {
  $111 = HEAP32[$16>>2]|0;
  __THREW__ = 0;
  invoke_vi($111|0,($112|0));
  $113 = __THREW__; __THREW__ = 0;
  $114 = $113&1;
  if ($114) {
   $115 = ___cxa_find_matching_catch_3(0|0)|0;
   $116 = tempRet0;
   ___clang_call_terminate($115);
   // unreachable;
  } else {
   $$23647 = $$23646;$$249 = $$248;
  }
 }
 ___resumeException($$249|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIwEERNS_10unique_ptrIwPFvPvEEERPwSM_($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 var $$$0152 = 0, $$$0152162 = 0, $$$0152163 = 0, $$$0152164 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i189 = 0, $$0$i$i$i$i252 = 0, $$0$i$i$i$i273 = 0, $$0$i$i$i$i295 = 0, $$0$i$i$i$i314 = 0, $$0$i$i$i$i340 = 0, $$0$i$i171 = 0, $$0$i$i173 = 0, $$0$i$i176 = 0, $$0$i$i179 = 0, $$0$i$i184 = 0, $$0$i$i187 = 0, $$0$i$i2$i$i = 0;
 var $$0$i$i2$i$i195 = 0, $$0$i$i2$i$i258 = 0, $$0$i$i2$i$i279 = 0, $$0$i$i2$i$i301 = 0, $$0$i$i2$i$i320 = 0, $$0$i$i2$i$i346 = 0, $$0$i$i204 = 0, $$0$i$i209 = 0, $$0$i$i216 = 0, $$0$i$i221 = 0, $$0$i$i228 = 0, $$0$i$i235 = 0, $$0$i$i267 = 0, $$0$i$i288 = 0, $$0$i$i309 = 0, $$0$i$i329 = 0, $$0$i$i334 = 0, $$0$i$i355 = 0, $$0135 = 0, $$0152 = 0;
 var $$0161 = 0, $$10 = 0, $$1153 = 0, $$2137$ph = 0, $$5145 = 0, $$5151 = 0, $$lcssa = 0, $$pre = 0, $$pre$i$i$i = 0, $$pre$i$i$i191 = 0, $$pre$i$i$i254 = 0, $$pre$i$i$i275 = 0, $$pre$i$i$i297 = 0, $$pre$i$i$i316 = 0, $$pre$i$i$i342 = 0, $$pre$i4$i$i197 = 0, $$pre$i4$i$i197527 = 0, $$pre$i4$i$i197529 = 0, $$pre$i4$i$i260 = 0, $$pre$i4$i$i260524 = 0;
 var $$pre$i4$i$i260525 = 0, $$pre$i4$i$i281 = 0, $$pre$i4$i$i281509 = 0, $$pre$i4$i$i281511 = 0, $$pre$i4$i$i303 = 0, $$pre$i4$i$i322 = 0, $$pre$i4$i$i322516$ph = 0, $$pre$i4$i$i322516533 = 0, $$pre513 = 0, $$pre515 = 0, $$pre518 = 0, $$pre519 = 0, $$pre520 = 0, $$pre521 = 0, $$pre522 = 0, $$pre523 = 0, $$sink = 0, $$sroa$0366$0 = 0, $$sroa$0388$0 = 0, $100 = 0;
 var $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0;
 var $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0;
 var $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0;
 var $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0;
 var $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0;
 var $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0;
 var $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0;
 var $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0;
 var $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0;
 var $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0;
 var $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $30 = 0, $300 = 0;
 var $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0;
 var $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0;
 var $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0;
 var $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0;
 var $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0;
 var $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0;
 var $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0;
 var $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0;
 var $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0;
 var $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0;
 var $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $50 = 0, $500 = 0;
 var $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0;
 var $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0;
 var $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0;
 var $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0;
 var $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0;
 var $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0;
 var $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0;
 var $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0;
 var $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0;
 var $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0;
 var $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0;
 var $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0;
 var $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0;
 var $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0;
 var $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0;
 var $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $90 = 0, $900 = 0;
 var $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0, $exitcond$i$i = 0, $exitcond$i$i172 = 0, $exitcond$i$i174 = 0;
 var $exitcond$i$i177 = 0, $exitcond$i$i180 = 0, $or$cond = 0, $or$cond13 = 0, $or$cond15 = 0, $or$cond167 = 0, $phitmp$i$i$i = 0, $phitmp$i$i$i192 = 0, $phitmp$i$i$i255 = 0, $phitmp$i$i$i276 = 0, $phitmp$i$i$i298 = 0, $phitmp$i$i$i317 = 0, $phitmp$i$i$i343 = 0, $phitmp$i5$i$i198 = 0, $phitmp$i5$i$i261 = 0, $phitmp$i5$i$i282 = 0, $phitmp$i5$i$i323 = 0, $switch = 0, $tmp$i$i200 = 0, $tmp$i$i263 = 0;
 var $tmp$i$i284 = 0, $tmp$i$i325 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 512|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(512|0);
 $11 = sp + 496|0;
 $12 = sp + 96|0;
 $13 = sp + 88|0;
 $14 = sp + 80|0;
 $15 = sp + 76|0;
 $16 = sp + 500|0;
 $17 = sp + 72|0;
 $18 = sp + 68|0;
 $19 = sp + 56|0;
 $20 = sp + 44|0;
 $21 = sp + 32|0;
 $22 = sp + 20|0;
 $23 = sp + 8|0;
 $24 = sp + 4|0;
 $25 = sp;
 HEAP32[$11>>2] = $10;
 $26 = $12;
 HEAP32[$13>>2] = $26;
 $27 = ((($13)) + 4|0);
 HEAP32[$27>>2] = (1003);
 HEAP32[$14>>2] = $12;
 $28 = ((($12)) + 400|0);
 HEAP32[$15>>2] = $28;
 ;HEAP32[$19>>2]=0|0;HEAP32[$19+4>>2]=0|0;HEAP32[$19+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $29 = (($19) + ($$0$i$i<<2)|0);
  HEAP32[$29>>2] = 0;
  $30 = (($$0$i$i) + 1)|0;
  $$0$i$i = $30;
 }
 ;HEAP32[$20>>2]=0|0;HEAP32[$20+4>>2]=0|0;HEAP32[$20+8>>2]=0|0;
 $$0$i$i171 = 0;
 while(1) {
  $exitcond$i$i172 = ($$0$i$i171|0)==(3);
  if ($exitcond$i$i172) {
   break;
  }
  $31 = (($20) + ($$0$i$i171<<2)|0);
  HEAP32[$31>>2] = 0;
  $32 = (($$0$i$i171) + 1)|0;
  $$0$i$i171 = $32;
 }
 ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;
 $$0$i$i173 = 0;
 while(1) {
  $exitcond$i$i174 = ($$0$i$i173|0)==(3);
  if ($exitcond$i$i174) {
   break;
  }
  $33 = (($21) + ($$0$i$i173<<2)|0);
  HEAP32[$33>>2] = 0;
  $34 = (($$0$i$i173) + 1)|0;
  $$0$i$i173 = $34;
 }
 ;HEAP32[$22>>2]=0|0;HEAP32[$22+4>>2]=0|0;HEAP32[$22+8>>2]=0|0;
 $$0$i$i176 = 0;
 while(1) {
  $exitcond$i$i177 = ($$0$i$i176|0)==(3);
  if ($exitcond$i$i177) {
   break;
  }
  $35 = (($22) + ($$0$i$i176<<2)|0);
  HEAP32[$35>>2] = 0;
  $36 = (($$0$i$i176) + 1)|0;
  $$0$i$i176 = $36;
 }
 ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP32[$23+8>>2]=0|0;
 $$0$i$i179 = 0;
 while(1) {
  $exitcond$i$i180 = ($$0$i$i179|0)==(3);
  if ($exitcond$i$i180) {
   break;
  }
  $37 = (($23) + ($$0$i$i179<<2)|0);
  HEAP32[$37>>2] = 0;
  $38 = (($$0$i$i179) + 1)|0;
  $$0$i$i179 = $38;
 }
 __THREW__ = 0;
 invoke_viiiiiiiiii(1016,($2|0),($3|0),($16|0),($17|0),($18|0),($19|0),($20|0),($21|0),($22|0),($24|0));
 $39 = __THREW__; __THREW__ = 0;
 $40 = $39&1;
 L21: do {
  if ($40) {
   $95 = ___cxa_find_matching_catch_2()|0;
   $96 = tempRet0;
   $$5145 = $95;$$5151 = $96;
  } else {
   $41 = HEAP32[$8>>2]|0;
   HEAP32[$9>>2] = $41;
   $42 = ((($21)) + 8|0);
   $43 = ((($42)) + 3|0);
   $44 = ((($21)) + 4|0);
   $45 = ((($22)) + 8|0);
   $46 = ((($45)) + 3|0);
   $47 = ((($22)) + 4|0);
   $48 = $4 & 512;
   $49 = ($48|0)!=(0);
   $50 = ((($20)) + 8|0);
   $51 = ((($50)) + 3|0);
   $52 = ((($20)) + 4|0);
   $53 = ((($23)) + 8|0);
   $54 = ((($53)) + 3|0);
   $55 = ((($23)) + 4|0);
   $56 = ((($16)) + 3|0);
   $57 = ((($19)) + 11|0);
   $58 = ((($19)) + 4|0);
   $$0152 = 0;$$0161 = 0;
   L24: while(1) {
    $59 = ($$0161>>>0)<(4);
    $60 = ($$0152|0)!=(0|0);
    if (!($59)) {
     label = 247;
     break;
    }
    $61 = HEAP32[$0>>2]|0;
    $62 = ($61|0)==(0|0);
    do {
     if ($62) {
      $899 = 1;
     } else {
      $63 = ((($61)) + 12|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = ((($61)) + 16|0);
      $66 = HEAP32[$65>>2]|0;
      $67 = ($64|0)==($66|0);
      if ($67) {
       $68 = HEAP32[$61>>2]|0;
       $69 = ((($68)) + 36|0);
       $70 = HEAP32[$69>>2]|0;
       __THREW__ = 0;
       $71 = (invoke_ii($70|0,($61|0))|0);
       $72 = __THREW__; __THREW__ = 0;
       $73 = $72&1;
       if ($73) {
        label = 35;
        break L24;
       } else {
        $$0$i$i$i$i = $71;
       }
      } else {
       $74 = HEAP32[$64>>2]|0;
       $75 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($74)|0);
       $$0$i$i$i$i = $75;
      }
      $76 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $77 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i,$76)|0);
      if ($77) {
       HEAP32[$0>>2] = 0;
       $899 = 1;
       break;
      } else {
       $$pre$i$i$i = HEAP32[$0>>2]|0;
       $phitmp$i$i$i = ($$pre$i$i$i|0)==(0|0);
       $899 = $phitmp$i$i$i;
       break;
      }
     }
    } while(0);
    $78 = HEAP32[$1>>2]|0;
    $79 = ($78|0)==(0|0);
    do {
     if ($79) {
      label = 32;
     } else {
      $80 = ((($78)) + 12|0);
      $81 = HEAP32[$80>>2]|0;
      $82 = ((($78)) + 16|0);
      $83 = HEAP32[$82>>2]|0;
      $84 = ($81|0)==($83|0);
      if ($84) {
       $85 = HEAP32[$78>>2]|0;
       $86 = ((($85)) + 36|0);
       $87 = HEAP32[$86>>2]|0;
       __THREW__ = 0;
       $88 = (invoke_ii($87|0,($78|0))|0);
       $89 = __THREW__; __THREW__ = 0;
       $90 = $89&1;
       if ($90) {
        label = 35;
        break L24;
       } else {
        $$0$i$i2$i$i = $88;
       }
      } else {
       $91 = HEAP32[$81>>2]|0;
       $92 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($91)|0);
       $$0$i$i2$i$i = $92;
      }
      $93 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
      $94 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i,$93)|0);
      if ($94) {
       HEAP32[$1>>2] = 0;
       label = 32;
       break;
      } else {
       if ($899) {
        $900 = $78;
        break;
       } else {
        label = 247;
        break L24;
       }
      }
     }
    } while(0);
    if ((label|0) == 32) {
     label = 0;
     if ($899) {
      label = 247;
      break;
     } else {
      $900 = 0;
     }
    }
    $99 = (($16) + ($$0161)|0);
    $100 = HEAP8[$99>>0]|0;
    $101 = $100 << 24 >> 24;
    $102 = ($$0161|0)!=(3);
    L49: do {
     switch ($101|0) {
     case 1:  {
      if ($102) {
       $103 = HEAP32[$0>>2]|0;
       $104 = ((($103)) + 12|0);
       $105 = HEAP32[$104>>2]|0;
       $106 = ((($103)) + 16|0);
       $107 = HEAP32[$106>>2]|0;
       $108 = ($105|0)==($107|0);
       if ($108) {
        $109 = HEAP32[$103>>2]|0;
        $110 = ((($109)) + 36|0);
        $111 = HEAP32[$110>>2]|0;
        __THREW__ = 0;
        $112 = (invoke_ii($111|0,($103|0))|0);
        $113 = __THREW__; __THREW__ = 0;
        $114 = $113&1;
        if ($114) {
         label = 35;
         break L24;
        } else {
         $$0$i$i184 = $112;
        }
       } else {
        $115 = HEAP32[$105>>2]|0;
        $116 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($115)|0);
        $$0$i$i184 = $116;
       }
       $117 = HEAP32[$7>>2]|0;
       $118 = ((($117)) + 12|0);
       $119 = HEAP32[$118>>2]|0;
       __THREW__ = 0;
       $120 = (invoke_iiii($119|0,($7|0),8192,($$0$i$i184|0))|0);
       $121 = __THREW__; __THREW__ = 0;
       $122 = $121&1;
       if ($122) {
        label = 35;
        break L24;
       }
       if (!($120)) {
        label = 47;
        break L24;
       }
       $123 = HEAP32[$0>>2]|0;
       $124 = ((($123)) + 12|0);
       $125 = HEAP32[$124>>2]|0;
       $126 = ((($123)) + 16|0);
       $127 = HEAP32[$126>>2]|0;
       $128 = ($125|0)==($127|0);
       if ($128) {
        $129 = HEAP32[$123>>2]|0;
        $130 = ((($129)) + 40|0);
        $131 = HEAP32[$130>>2]|0;
        __THREW__ = 0;
        $132 = (invoke_ii($131|0,($123|0))|0);
        $133 = __THREW__; __THREW__ = 0;
        $134 = $133&1;
        if ($134) {
         label = 35;
         break L24;
        } else {
         $$0$i$i187 = $132;
        }
       } else {
        $135 = ((($125)) + 4|0);
        HEAP32[$124>>2] = $135;
        $136 = HEAP32[$125>>2]|0;
        $137 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($136)|0);
        $$0$i$i187 = $137;
       }
       __THREW__ = 0;
       invoke_vii(1014,($23|0),($$0$i$i187|0));
       $138 = __THREW__; __THREW__ = 0;
       $139 = $138&1;
       if ($139) {
        label = 35;
        break L24;
       } else {
        $$pre$i4$i$i197 = $900;$160 = $900;
        label = 49;
       }
      } else {
       $$1153 = $$0152;
      }
      break;
     }
     case 0:  {
      if ($102) {
       $$pre$i4$i$i197 = $900;$160 = $900;
       label = 49;
      } else {
       $$1153 = $$0152;
      }
      break;
     }
     case 3:  {
      $214 = HEAP8[$43>>0]|0;
      $215 = ($214<<24>>24)<(0);
      $216 = HEAP32[$44>>2]|0;
      $217 = $214&255;
      $218 = $215 ? $216 : $217;
      $219 = HEAP8[$46>>0]|0;
      $220 = ($219<<24>>24)<(0);
      $221 = HEAP32[$47>>2]|0;
      $222 = $219&255;
      $223 = $220 ? $221 : $222;
      $224 = (0 - ($223))|0;
      $225 = ($218|0)==($224|0);
      if ($225) {
       $$1153 = $$0152;
      } else {
       $226 = ($218|0)==(0);
       if ($226) {
        $267 = HEAP32[$0>>2]|0;
        $268 = ((($267)) + 12|0);
        $269 = HEAP32[$268>>2]|0;
        $270 = ((($267)) + 16|0);
        $271 = HEAP32[$270>>2]|0;
        $272 = ($269|0)==($271|0);
        if ($272) {
         $273 = HEAP32[$267>>2]|0;
         $274 = ((($273)) + 36|0);
         $275 = HEAP32[$274>>2]|0;
         __THREW__ = 0;
         $276 = (invoke_ii($275|0,($267|0))|0);
         $277 = __THREW__; __THREW__ = 0;
         $278 = $277&1;
         if ($278) {
          label = 35;
          break L24;
         } else {
          $$0$i$i221 = $276;
         }
        } else {
         $279 = HEAP32[$269>>2]|0;
         $280 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($279)|0);
         $$0$i$i221 = $280;
        }
        $281 = HEAP8[$46>>0]|0;
        $282 = ($281<<24>>24)<(0);
        $283 = HEAP32[$22>>2]|0;
        $284 = $282 ? $283 : $22;
        $285 = HEAP32[$284>>2]|0;
        $286 = ($$0$i$i221|0)==($285|0);
        if (!($286)) {
         $$1153 = $$0152;
         break L49;
        }
        $287 = HEAP32[$0>>2]|0;
        $288 = ((($287)) + 12|0);
        $289 = HEAP32[$288>>2]|0;
        $290 = ((($287)) + 16|0);
        $291 = HEAP32[$290>>2]|0;
        $292 = ($289|0)==($291|0);
        if ($292) {
         $293 = HEAP32[$287>>2]|0;
         $294 = ((($293)) + 40|0);
         $295 = HEAP32[$294>>2]|0;
         __THREW__ = 0;
         (invoke_ii($295|0,($287|0))|0);
         $296 = __THREW__; __THREW__ = 0;
         $297 = $296&1;
         if ($297) {
          label = 35;
          break L24;
         }
        } else {
         $298 = ((($289)) + 4|0);
         HEAP32[$288>>2] = $298;
         $299 = HEAP32[$289>>2]|0;
         (__ZNSt3__211char_traitsIwE11to_int_typeEw($299)|0);
        }
        HEAP8[$6>>0] = 1;
        $300 = HEAP8[$46>>0]|0;
        $301 = ($300<<24>>24)<(0);
        $302 = HEAP32[$47>>2]|0;
        $303 = $300&255;
        $304 = $301 ? $302 : $303;
        $305 = ($304>>>0)>(1);
        $$$0152162 = $305 ? $22 : $$0152;
        $$1153 = $$$0152162;
        break L49;
       }
       $227 = ($223|0)==(0);
       $228 = HEAP32[$0>>2]|0;
       $229 = ((($228)) + 12|0);
       $230 = HEAP32[$229>>2]|0;
       $231 = ((($228)) + 16|0);
       $232 = HEAP32[$231>>2]|0;
       $233 = ($230|0)==($232|0);
       if ($227) {
        if ($233) {
         $234 = HEAP32[$228>>2]|0;
         $235 = ((($234)) + 36|0);
         $236 = HEAP32[$235>>2]|0;
         __THREW__ = 0;
         $237 = (invoke_ii($236|0,($228|0))|0);
         $238 = __THREW__; __THREW__ = 0;
         $239 = $238&1;
         if ($239) {
          label = 35;
          break L24;
         } else {
          $$0$i$i216 = $237;
         }
        } else {
         $240 = HEAP32[$230>>2]|0;
         $241 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($240)|0);
         $$0$i$i216 = $241;
        }
        $242 = HEAP8[$43>>0]|0;
        $243 = ($242<<24>>24)<(0);
        $244 = HEAP32[$21>>2]|0;
        $245 = $243 ? $244 : $21;
        $246 = HEAP32[$245>>2]|0;
        $247 = ($$0$i$i216|0)==($246|0);
        if (!($247)) {
         HEAP8[$6>>0] = 1;
         $$1153 = $$0152;
         break L49;
        }
        $248 = HEAP32[$0>>2]|0;
        $249 = ((($248)) + 12|0);
        $250 = HEAP32[$249>>2]|0;
        $251 = ((($248)) + 16|0);
        $252 = HEAP32[$251>>2]|0;
        $253 = ($250|0)==($252|0);
        if ($253) {
         $254 = HEAP32[$248>>2]|0;
         $255 = ((($254)) + 40|0);
         $256 = HEAP32[$255>>2]|0;
         __THREW__ = 0;
         (invoke_ii($256|0,($248|0))|0);
         $257 = __THREW__; __THREW__ = 0;
         $258 = $257&1;
         if ($258) {
          label = 35;
          break L24;
         }
        } else {
         $259 = ((($250)) + 4|0);
         HEAP32[$249>>2] = $259;
         $260 = HEAP32[$250>>2]|0;
         (__ZNSt3__211char_traitsIwE11to_int_typeEw($260)|0);
        }
        $261 = HEAP8[$43>>0]|0;
        $262 = ($261<<24>>24)<(0);
        $263 = HEAP32[$44>>2]|0;
        $264 = $261&255;
        $265 = $262 ? $263 : $264;
        $266 = ($265>>>0)>(1);
        $$$0152 = $266 ? $21 : $$0152;
        $$1153 = $$$0152;
        break L49;
       }
       if ($233) {
        $306 = HEAP32[$228>>2]|0;
        $307 = ((($306)) + 36|0);
        $308 = HEAP32[$307>>2]|0;
        __THREW__ = 0;
        $309 = (invoke_ii($308|0,($228|0))|0);
        $310 = __THREW__; __THREW__ = 0;
        $311 = $310&1;
        if ($311) {
         label = 35;
         break L24;
        } else {
         $$0$i$i228 = $309;
        }
       } else {
        $312 = HEAP32[$230>>2]|0;
        $313 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($312)|0);
        $$0$i$i228 = $313;
       }
       $314 = HEAP8[$43>>0]|0;
       $315 = ($314<<24>>24)<(0);
       $316 = HEAP32[$21>>2]|0;
       $317 = $315 ? $316 : $21;
       $318 = HEAP32[$317>>2]|0;
       $319 = ($$0$i$i228|0)==($318|0);
       $320 = HEAP32[$0>>2]|0;
       $321 = ((($320)) + 12|0);
       $322 = HEAP32[$321>>2]|0;
       $323 = ((($320)) + 16|0);
       $324 = HEAP32[$323>>2]|0;
       $325 = ($322|0)==($324|0);
       if ($319) {
        if ($325) {
         $326 = HEAP32[$320>>2]|0;
         $327 = ((($326)) + 40|0);
         $328 = HEAP32[$327>>2]|0;
         __THREW__ = 0;
         (invoke_ii($328|0,($320|0))|0);
         $329 = __THREW__; __THREW__ = 0;
         $330 = $329&1;
         if ($330) {
          label = 35;
          break L24;
         }
        } else {
         $331 = ((($322)) + 4|0);
         HEAP32[$321>>2] = $331;
         $332 = HEAP32[$322>>2]|0;
         (__ZNSt3__211char_traitsIwE11to_int_typeEw($332)|0);
        }
        $333 = HEAP8[$43>>0]|0;
        $334 = ($333<<24>>24)<(0);
        $335 = HEAP32[$44>>2]|0;
        $336 = $333&255;
        $337 = $334 ? $335 : $336;
        $338 = ($337>>>0)>(1);
        $$$0152163 = $338 ? $21 : $$0152;
        $$1153 = $$$0152163;
        break L49;
       }
       if ($325) {
        $339 = HEAP32[$320>>2]|0;
        $340 = ((($339)) + 36|0);
        $341 = HEAP32[$340>>2]|0;
        __THREW__ = 0;
        $342 = (invoke_ii($341|0,($320|0))|0);
        $343 = __THREW__; __THREW__ = 0;
        $344 = $343&1;
        if ($344) {
         label = 35;
         break L24;
        } else {
         $$0$i$i235 = $342;
        }
       } else {
        $345 = HEAP32[$322>>2]|0;
        $346 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($345)|0);
        $$0$i$i235 = $346;
       }
       $347 = HEAP8[$46>>0]|0;
       $348 = ($347<<24>>24)<(0);
       $349 = HEAP32[$22>>2]|0;
       $350 = $348 ? $349 : $22;
       $351 = HEAP32[$350>>2]|0;
       $352 = ($$0$i$i235|0)==($351|0);
       if (!($352)) {
        label = 109;
        break L24;
       }
       $353 = HEAP32[$0>>2]|0;
       $354 = ((($353)) + 12|0);
       $355 = HEAP32[$354>>2]|0;
       $356 = ((($353)) + 16|0);
       $357 = HEAP32[$356>>2]|0;
       $358 = ($355|0)==($357|0);
       if ($358) {
        $359 = HEAP32[$353>>2]|0;
        $360 = ((($359)) + 40|0);
        $361 = HEAP32[$360>>2]|0;
        __THREW__ = 0;
        (invoke_ii($361|0,($353|0))|0);
        $362 = __THREW__; __THREW__ = 0;
        $363 = $362&1;
        if ($363) {
         label = 35;
         break L24;
        }
       } else {
        $364 = ((($355)) + 4|0);
        HEAP32[$354>>2] = $364;
        $365 = HEAP32[$355>>2]|0;
        (__ZNSt3__211char_traitsIwE11to_int_typeEw($365)|0);
       }
       HEAP8[$6>>0] = 1;
       $366 = HEAP8[$46>>0]|0;
       $367 = ($366<<24>>24)<(0);
       $368 = HEAP32[$47>>2]|0;
       $369 = $366&255;
       $370 = $367 ? $368 : $369;
       $371 = ($370>>>0)>(1);
       $$$0152164 = $371 ? $22 : $$0152;
       $$1153 = $$$0152164;
      }
      break;
     }
     case 2:  {
      $374 = ($$0161>>>0)<(2);
      $or$cond = $374 | $60;
      if (!($or$cond)) {
       $375 = ($$0161|0)==(2);
       $376 = HEAP8[$56>>0]|0;
       $377 = ($376<<24>>24)!=(0);
       $378 = $375 & $377;
       $brmerge = $49 | $378;
       if (!($brmerge)) {
        $$1153 = 0;
        break L49;
       }
      }
      $379 = HEAP8[$51>>0]|0;
      $380 = ($379<<24>>24)<(0);
      $381 = HEAP32[$20>>2]|0;
      $382 = $380 ? $381 : $20;
      $383 = $382;
      $384 = ($$0161|0)==(0);
      L115: do {
       if ($384) {
        $$pre$i4$i$i260 = $900;$$sroa$0366$0 = $383;$435 = $379;$439 = $381;$462 = $900;
       } else {
        $385 = (($$0161) + -1)|0;
        $386 = (($16) + ($385)|0);
        $387 = HEAP8[$386>>0]|0;
        $switch = ($387&255)<(2);
        if ($switch) {
         $$sroa$0388$0 = $383;$389 = $379;$393 = $381;
         while(1) {
          $388 = ($389<<24>>24)<(0);
          $390 = HEAP32[$52>>2]|0;
          $391 = $389&255;
          $392 = $388 ? $393 : $20;
          $394 = $388 ? $390 : $391;
          $395 = (($392) + ($394<<2)|0);
          $396 = $$sroa$0388$0;
          $397 = ($396|0)==($395|0);
          if ($397) {
           $410 = $389;$412 = $393;
           break;
          }
          $398 = HEAP32[$396>>2]|0;
          $399 = HEAP32[$7>>2]|0;
          $400 = ((($399)) + 12|0);
          $401 = HEAP32[$400>>2]|0;
          __THREW__ = 0;
          $402 = (invoke_iiii($401|0,($7|0),8192,($398|0))|0);
          $403 = __THREW__; __THREW__ = 0;
          $404 = $403&1;
          if ($404) {
           label = 119;
           break L24;
          }
          if (!($402)) {
           label = 117;
           break;
          }
          $405 = ((($396)) + 4|0);
          $406 = $405;
          $$pre518 = HEAP8[$51>>0]|0;
          $$pre519 = HEAP32[$20>>2]|0;
          $$sroa$0388$0 = $406;$389 = $$pre518;$393 = $$pre519;
         }
         if ((label|0) == 117) {
          label = 0;
          $$pre520 = HEAP8[$51>>0]|0;
          $$pre521 = HEAP32[$20>>2]|0;
          $410 = $$pre520;$412 = $$pre521;
         }
         $409 = ($410<<24>>24)<(0);
         $411 = $409 ? $412 : $20;
         $413 = $411;
         $414 = (($$sroa$0388$0) - ($413))|0;
         $415 = $414 >> 2;
         $416 = HEAP8[$54>>0]|0;
         $417 = ($416<<24>>24)<(0);
         $418 = HEAP32[$55>>2]|0;
         $419 = $416&255;
         $420 = $417 ? $418 : $419;
         $421 = ($415>>>0)>($420>>>0);
         if ($421) {
          $$pre$i4$i$i260 = $900;$$sroa$0366$0 = $413;$435 = $410;$439 = $412;$462 = $900;
         } else {
          $422 = HEAP32[$23>>2]|0;
          $423 = $417 ? $422 : $23;
          $424 = (($423) + ($420<<2)|0);
          $425 = (0 - ($415))|0;
          $426 = (($424) + ($425<<2)|0);
          $$sink = $411;$428 = $426;
          while(1) {
           $427 = ($428|0)==($424|0);
           if ($427) {
            $$pre$i4$i$i260 = $900;$$sroa$0366$0 = $$sroa$0388$0;$435 = $410;$439 = $412;$462 = $900;
            break L115;
           }
           $429 = HEAP32[$428>>2]|0;
           $430 = HEAP32[$$sink>>2]|0;
           $431 = ($429|0)==($430|0);
           if (!($431)) {
            $$pre$i4$i$i260 = $900;$$sroa$0366$0 = $413;$435 = $410;$439 = $412;$462 = $900;
            break L115;
           }
           $432 = ((($428)) + 4|0);
           $433 = ((($$sink)) + 4|0);
           $$sink = $433;$428 = $432;
          }
         }
        } else {
         $$pre$i4$i$i260 = $900;$$sroa$0366$0 = $383;$435 = $379;$439 = $381;$462 = $900;
        }
       }
      } while(0);
      L130: while(1) {
       $434 = ($435<<24>>24)<(0);
       $436 = HEAP32[$52>>2]|0;
       $437 = $435&255;
       $438 = $434 ? $439 : $20;
       $440 = $434 ? $436 : $437;
       $441 = (($438) + ($440<<2)|0);
       $442 = $$sroa$0366$0;
       $443 = ($442|0)==($441|0);
       if ($443) {
        $$lcssa = $441;
        break;
       }
       $444 = HEAP32[$0>>2]|0;
       $445 = ($444|0)==(0|0);
       do {
        if ($445) {
         $478 = 1;
        } else {
         $446 = ((($444)) + 12|0);
         $447 = HEAP32[$446>>2]|0;
         $448 = ((($444)) + 16|0);
         $449 = HEAP32[$448>>2]|0;
         $450 = ($447|0)==($449|0);
         if ($450) {
          $451 = HEAP32[$444>>2]|0;
          $452 = ((($451)) + 36|0);
          $453 = HEAP32[$452>>2]|0;
          __THREW__ = 0;
          $454 = (invoke_ii($453|0,($444|0))|0);
          $455 = __THREW__; __THREW__ = 0;
          $456 = $455&1;
          if ($456) {
           label = 149;
           break L24;
          } else {
           $$0$i$i$i$i252 = $454;
          }
         } else {
          $457 = HEAP32[$447>>2]|0;
          $458 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($457)|0);
          $$0$i$i$i$i252 = $458;
         }
         $459 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $460 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i252,$459)|0);
         if ($460) {
          HEAP32[$0>>2] = 0;
          $478 = 1;
          break;
         } else {
          $$pre$i$i$i254 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i255 = ($$pre$i$i$i254|0)==(0|0);
          $478 = $phitmp$i$i$i255;
          break;
         }
        }
       } while(0);
       $461 = ($462|0)==(0|0);
       do {
        if ($461) {
         $$pre$i4$i$i260525 = $$pre$i4$i$i260;
         label = 139;
        } else {
         $463 = ((($462)) + 12|0);
         $464 = HEAP32[$463>>2]|0;
         $465 = ((($462)) + 16|0);
         $466 = HEAP32[$465>>2]|0;
         $467 = ($464|0)==($466|0);
         if ($467) {
          $468 = HEAP32[$462>>2]|0;
          $469 = ((($468)) + 36|0);
          $470 = HEAP32[$469>>2]|0;
          __THREW__ = 0;
          $471 = (invoke_ii($470|0,($462|0))|0);
          $472 = __THREW__; __THREW__ = 0;
          $473 = $472&1;
          if ($473) {
           label = 149;
           break L24;
          } else {
           $$0$i$i2$i$i258 = $471;
          }
         } else {
          $474 = HEAP32[$464>>2]|0;
          $475 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($474)|0);
          $$0$i$i2$i$i258 = $475;
         }
         $476 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $477 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i258,$476)|0);
         if ($477) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i260525 = 0;
          label = 139;
          break;
         } else {
          $phitmp$i5$i$i261 = ($$pre$i4$i$i260|0)==(0|0);
          $tmp$i$i263 = $478 ^ $phitmp$i5$i$i261;
          if ($tmp$i$i263) {
           $$pre$i4$i$i260524 = $$pre$i4$i$i260;$902 = $$pre$i4$i$i260;
           break;
          } else {
           $$lcssa = $442;
           break L130;
          }
         }
        }
       } while(0);
       if ((label|0) == 139) {
        label = 0;
        if ($478) {
         $$lcssa = $442;
         break;
        } else {
         $$pre$i4$i$i260524 = $$pre$i4$i$i260525;$902 = 0;
        }
       }
       $479 = HEAP32[$0>>2]|0;
       $480 = ((($479)) + 12|0);
       $481 = HEAP32[$480>>2]|0;
       $482 = ((($479)) + 16|0);
       $483 = HEAP32[$482>>2]|0;
       $484 = ($481|0)==($483|0);
       if ($484) {
        $485 = HEAP32[$479>>2]|0;
        $486 = ((($485)) + 36|0);
        $487 = HEAP32[$486>>2]|0;
        __THREW__ = 0;
        $488 = (invoke_ii($487|0,($479|0))|0);
        $489 = __THREW__; __THREW__ = 0;
        $490 = $489&1;
        if ($490) {
         label = 149;
         break L24;
        } else {
         $$0$i$i267 = $488;
        }
       } else {
        $491 = HEAP32[$481>>2]|0;
        $492 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($491)|0);
        $$0$i$i267 = $492;
       }
       $493 = HEAP32[$442>>2]|0;
       $494 = ($$0$i$i267|0)==($493|0);
       if (!($494)) {
        $$lcssa = $442;
        break;
       }
       $495 = HEAP32[$0>>2]|0;
       $496 = ((($495)) + 12|0);
       $497 = HEAP32[$496>>2]|0;
       $498 = ((($495)) + 16|0);
       $499 = HEAP32[$498>>2]|0;
       $500 = ($497|0)==($499|0);
       if ($500) {
        $501 = HEAP32[$495>>2]|0;
        $502 = ((($501)) + 40|0);
        $503 = HEAP32[$502>>2]|0;
        __THREW__ = 0;
        (invoke_ii($503|0,($495|0))|0);
        $504 = __THREW__; __THREW__ = 0;
        $505 = $504&1;
        if ($505) {
         label = 150;
         break L24;
        }
       } else {
        $506 = ((($497)) + 4|0);
        HEAP32[$496>>2] = $506;
        $507 = HEAP32[$497>>2]|0;
        (__ZNSt3__211char_traitsIwE11to_int_typeEw($507)|0);
       }
       $508 = ((($442)) + 4|0);
       $509 = $508;
       $$pre522 = HEAP8[$51>>0]|0;
       $$pre523 = HEAP32[$20>>2]|0;
       $$pre$i4$i$i260 = $$pre$i4$i$i260524;$$sroa$0366$0 = $509;$435 = $$pre522;$439 = $$pre523;$462 = $902;
      }
      if ($49) {
       $514 = HEAP8[$51>>0]|0;
       $515 = ($514<<24>>24)<(0);
       $516 = HEAP32[$20>>2]|0;
       $517 = HEAP32[$52>>2]|0;
       $518 = $514&255;
       $519 = $515 ? $516 : $20;
       $520 = $515 ? $517 : $518;
       $521 = (($519) + ($520<<2)|0);
       $522 = ($$lcssa|0)==($521|0);
       if ($522) {
        $$1153 = $$0152;
       } else {
        label = 153;
        break L24;
       }
      } else {
       $$1153 = $$0152;
      }
      break;
     }
     case 4:  {
      $$0135 = 0;$$pre$i4$i$i281 = $900;$543 = $900;
      L166: while(1) {
       $525 = HEAP32[$0>>2]|0;
       $526 = ($525|0)==(0|0);
       do {
        if ($526) {
         $559 = 1;
        } else {
         $527 = ((($525)) + 12|0);
         $528 = HEAP32[$527>>2]|0;
         $529 = ((($525)) + 16|0);
         $530 = HEAP32[$529>>2]|0;
         $531 = ($528|0)==($530|0);
         if ($531) {
          $532 = HEAP32[$525>>2]|0;
          $533 = ((($532)) + 36|0);
          $534 = HEAP32[$533>>2]|0;
          __THREW__ = 0;
          $535 = (invoke_ii($534|0,($525|0))|0);
          $536 = __THREW__; __THREW__ = 0;
          $537 = $536&1;
          if ($537) {
           label = 245;
           break L24;
          } else {
           $$0$i$i$i$i273 = $535;
          }
         } else {
          $538 = HEAP32[$528>>2]|0;
          $539 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($538)|0);
          $$0$i$i$i$i273 = $539;
         }
         $540 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $541 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i273,$540)|0);
         if ($541) {
          HEAP32[$0>>2] = 0;
          $559 = 1;
          break;
         } else {
          $$pre$i$i$i275 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i276 = ($$pre$i$i$i275|0)==(0|0);
          $559 = $phitmp$i$i$i276;
          break;
         }
        }
       } while(0);
       $542 = ($543|0)==(0|0);
       do {
        if ($542) {
         $$pre$i4$i$i281509 = $$pre$i4$i$i281;
         label = 167;
        } else {
         $544 = ((($543)) + 12|0);
         $545 = HEAP32[$544>>2]|0;
         $546 = ((($543)) + 16|0);
         $547 = HEAP32[$546>>2]|0;
         $548 = ($545|0)==($547|0);
         if ($548) {
          $549 = HEAP32[$543>>2]|0;
          $550 = ((($549)) + 36|0);
          $551 = HEAP32[$550>>2]|0;
          __THREW__ = 0;
          $552 = (invoke_ii($551|0,($543|0))|0);
          $553 = __THREW__; __THREW__ = 0;
          $554 = $553&1;
          if ($554) {
           label = 245;
           break L24;
          } else {
           $$0$i$i2$i$i279 = $552;
          }
         } else {
          $555 = HEAP32[$545>>2]|0;
          $556 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($555)|0);
          $$0$i$i2$i$i279 = $556;
         }
         $557 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $558 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i279,$557)|0);
         if ($558) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i281509 = 0;
          label = 167;
          break;
         } else {
          $phitmp$i5$i$i282 = ($$pre$i4$i$i281|0)==(0|0);
          $tmp$i$i284 = $559 ^ $phitmp$i5$i$i282;
          if ($tmp$i$i284) {
           $$pre$i4$i$i281511 = $$pre$i4$i$i281;$903 = $$pre$i4$i$i281;
           break;
          } else {
           $$pre$i4$i$i303 = $$pre$i4$i$i281;
           break L166;
          }
         }
        }
       } while(0);
       if ((label|0) == 167) {
        label = 0;
        if ($559) {
         $$pre$i4$i$i303 = $$pre$i4$i$i281509;
         break;
        } else {
         $$pre$i4$i$i281511 = $$pre$i4$i$i281509;$903 = 0;
        }
       }
       $560 = HEAP32[$0>>2]|0;
       $561 = ((($560)) + 12|0);
       $562 = HEAP32[$561>>2]|0;
       $563 = ((($560)) + 16|0);
       $564 = HEAP32[$563>>2]|0;
       $565 = ($562|0)==($564|0);
       if ($565) {
        $566 = HEAP32[$560>>2]|0;
        $567 = ((($566)) + 36|0);
        $568 = HEAP32[$567>>2]|0;
        __THREW__ = 0;
        $569 = (invoke_ii($568|0,($560|0))|0);
        $570 = __THREW__; __THREW__ = 0;
        $571 = $570&1;
        if ($571) {
         label = 245;
         break L24;
        } else {
         $$0$i$i288 = $569;
        }
       } else {
        $572 = HEAP32[$562>>2]|0;
        $573 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($572)|0);
        $$0$i$i288 = $573;
       }
       $574 = HEAP32[$7>>2]|0;
       $575 = ((($574)) + 12|0);
       $576 = HEAP32[$575>>2]|0;
       __THREW__ = 0;
       $577 = (invoke_iiii($576|0,($7|0),2048,($$0$i$i288|0))|0);
       $578 = __THREW__; __THREW__ = 0;
       $579 = $578&1;
       if ($579) {
        label = 245;
        break L24;
       }
       if ($577) {
        $580 = HEAP32[$9>>2]|0;
        $581 = HEAP32[$11>>2]|0;
        $582 = ($580|0)==($581|0);
        if ($582) {
         __THREW__ = 0;
         invoke_viii(1017,($8|0),($9|0),($11|0));
         $583 = __THREW__; __THREW__ = 0;
         $584 = $583&1;
         if ($584) {
          label = 245;
          break L24;
         }
         $$pre513 = HEAP32[$9>>2]|0;
         $586 = $$pre513;
        } else {
         $586 = $580;
        }
        $585 = ((($586)) + 4|0);
        HEAP32[$9>>2] = $585;
        HEAP32[$586>>2] = $$0$i$i288;
        $587 = (($$0135) + 1)|0;
        $$2137$ph = $587;
       } else {
        $588 = HEAP8[$57>>0]|0;
        $589 = ($588<<24>>24)<(0);
        $590 = HEAP32[$58>>2]|0;
        $591 = $588&255;
        $592 = $589 ? $590 : $591;
        $593 = ($592|0)!=(0);
        $594 = ($$0135|0)!=(0);
        $or$cond13 = $594 & $593;
        $595 = HEAP32[$18>>2]|0;
        $596 = ($$0$i$i288|0)==($595|0);
        $or$cond167 = $596 & $or$cond13;
        if (!($or$cond167)) {
         $$pre$i4$i$i303 = $$pre$i4$i$i281511;
         break;
        }
        $597 = HEAP32[$14>>2]|0;
        $598 = HEAP32[$15>>2]|0;
        $599 = ($597|0)==($598|0);
        if ($599) {
         __THREW__ = 0;
         invoke_viii(1010,($13|0),($14|0),($15|0));
         $600 = __THREW__; __THREW__ = 0;
         $601 = $600&1;
         if ($601) {
          label = 245;
          break L24;
         }
         $$pre = HEAP32[$14>>2]|0;
         $603 = $$pre;
        } else {
         $603 = $597;
        }
        $602 = ((($603)) + 4|0);
        HEAP32[$14>>2] = $602;
        HEAP32[$603>>2] = $$0135;
        $$2137$ph = 0;
       }
       $604 = HEAP32[$0>>2]|0;
       $605 = ((($604)) + 12|0);
       $606 = HEAP32[$605>>2]|0;
       $607 = ((($604)) + 16|0);
       $608 = HEAP32[$607>>2]|0;
       $609 = ($606|0)==($608|0);
       if ($609) {
        $610 = HEAP32[$604>>2]|0;
        $611 = ((($610)) + 40|0);
        $612 = HEAP32[$611>>2]|0;
        __THREW__ = 0;
        (invoke_ii($612|0,($604|0))|0);
        $613 = __THREW__; __THREW__ = 0;
        $614 = $613&1;
        if ($614) {
         label = 245;
         break L24;
        } else {
         $$0135 = $$2137$ph;$$pre$i4$i$i281 = $$pre$i4$i$i281511;$543 = $903;
         continue;
        }
       } else {
        $615 = ((($606)) + 4|0);
        HEAP32[$605>>2] = $615;
        $616 = HEAP32[$606>>2]|0;
        (__ZNSt3__211char_traitsIwE11to_int_typeEw($616)|0);
        $$0135 = $$2137$ph;$$pre$i4$i$i281 = $$pre$i4$i$i281511;$543 = $903;
        continue;
       }
      }
      $617 = HEAP32[$13>>2]|0;
      $618 = HEAP32[$14>>2]|0;
      $619 = ($617|0)!=($618|0);
      $620 = ($$0135|0)!=(0);
      $or$cond15 = $620 & $619;
      if ($or$cond15) {
       $621 = HEAP32[$15>>2]|0;
       $622 = ($618|0)==($621|0);
       if ($622) {
        __THREW__ = 0;
        invoke_viii(1010,($13|0),($14|0),($15|0));
        $623 = __THREW__; __THREW__ = 0;
        $624 = $623&1;
        if ($624) {
         label = 245;
         break L24;
        }
        $$pre515 = HEAP32[$14>>2]|0;
        $626 = $$pre515;
       } else {
        $626 = $618;
       }
       $625 = ((($626)) + 4|0);
       HEAP32[$14>>2] = $625;
       HEAP32[$626>>2] = $$0135;
      }
      $627 = HEAP32[$24>>2]|0;
      $628 = ($627|0)>(0);
      L219: do {
       if ($628) {
        $629 = HEAP32[$0>>2]|0;
        $630 = ($629|0)==(0|0);
        do {
         if ($630) {
          $904 = 1;
         } else {
          $631 = ((($629)) + 12|0);
          $632 = HEAP32[$631>>2]|0;
          $633 = ((($629)) + 16|0);
          $634 = HEAP32[$633>>2]|0;
          $635 = ($632|0)==($634|0);
          if ($635) {
           $636 = HEAP32[$629>>2]|0;
           $637 = ((($636)) + 36|0);
           $638 = HEAP32[$637>>2]|0;
           __THREW__ = 0;
           $639 = (invoke_ii($638|0,($629|0))|0);
           $640 = __THREW__; __THREW__ = 0;
           $641 = $640&1;
           if ($641) {
            label = 245;
            break L24;
           } else {
            $$0$i$i$i$i295 = $639;
           }
          } else {
           $642 = HEAP32[$632>>2]|0;
           $643 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($642)|0);
           $$0$i$i$i$i295 = $643;
          }
          $644 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
          $645 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i295,$644)|0);
          if ($645) {
           HEAP32[$0>>2] = 0;
           $904 = 1;
           break;
          } else {
           $$pre$i$i$i297 = HEAP32[$0>>2]|0;
           $phitmp$i$i$i298 = ($$pre$i$i$i297|0)==(0|0);
           $904 = $phitmp$i$i$i298;
           break;
          }
         }
        } while(0);
        $646 = ($$pre$i4$i$i303|0)==(0|0);
        do {
         if ($646) {
          label = 205;
         } else {
          $647 = ((($$pre$i4$i$i303)) + 12|0);
          $648 = HEAP32[$647>>2]|0;
          $649 = ((($$pre$i4$i$i303)) + 16|0);
          $650 = HEAP32[$649>>2]|0;
          $651 = ($648|0)==($650|0);
          if ($651) {
           $652 = HEAP32[$$pre$i4$i$i303>>2]|0;
           $653 = ((($652)) + 36|0);
           $654 = HEAP32[$653>>2]|0;
           __THREW__ = 0;
           $655 = (invoke_ii($654|0,($$pre$i4$i$i303|0))|0);
           $656 = __THREW__; __THREW__ = 0;
           $657 = $656&1;
           if ($657) {
            label = 245;
            break L24;
           } else {
            $$0$i$i2$i$i301 = $655;
           }
          } else {
           $658 = HEAP32[$648>>2]|0;
           $659 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($658)|0);
           $$0$i$i2$i$i301 = $659;
          }
          $660 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
          $661 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i301,$660)|0);
          if ($661) {
           HEAP32[$1>>2] = 0;
           label = 205;
           break;
          } else {
           if ($904) {
            $905 = $$pre$i4$i$i303;
            break;
           } else {
            label = 244;
            break L24;
           }
          }
         }
        } while(0);
        if ((label|0) == 205) {
         label = 0;
         if ($904) {
          label = 244;
          break L24;
         } else {
          $905 = 0;
         }
        }
        $662 = HEAP32[$0>>2]|0;
        $663 = ((($662)) + 12|0);
        $664 = HEAP32[$663>>2]|0;
        $665 = ((($662)) + 16|0);
        $666 = HEAP32[$665>>2]|0;
        $667 = ($664|0)==($666|0);
        if ($667) {
         $668 = HEAP32[$662>>2]|0;
         $669 = ((($668)) + 36|0);
         $670 = HEAP32[$669>>2]|0;
         __THREW__ = 0;
         $671 = (invoke_ii($670|0,($662|0))|0);
         $672 = __THREW__; __THREW__ = 0;
         $673 = $672&1;
         if ($673) {
          label = 245;
          break L24;
         } else {
          $$0$i$i309 = $671;
         }
        } else {
         $674 = HEAP32[$664>>2]|0;
         $675 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($674)|0);
         $$0$i$i309 = $675;
        }
        $676 = HEAP32[$17>>2]|0;
        $677 = ($$0$i$i309|0)==($676|0);
        if (!($677)) {
         label = 244;
         break L24;
        }
        $678 = HEAP32[$0>>2]|0;
        $679 = ((($678)) + 12|0);
        $680 = HEAP32[$679>>2]|0;
        $681 = ((($678)) + 16|0);
        $682 = HEAP32[$681>>2]|0;
        $683 = ($680|0)==($682|0);
        if ($683) {
         $684 = HEAP32[$678>>2]|0;
         $685 = ((($684)) + 40|0);
         $686 = HEAP32[$685>>2]|0;
         __THREW__ = 0;
         (invoke_ii($686|0,($678|0))|0);
         $687 = __THREW__; __THREW__ = 0;
         $688 = $687&1;
         if ($688) {
          label = 245;
          break L24;
         } else {
          $$pre$i4$i$i322 = $905;$711 = $905;
         }
        } else {
         $689 = ((($680)) + 4|0);
         HEAP32[$679>>2] = $689;
         $690 = HEAP32[$680>>2]|0;
         (__ZNSt3__211char_traitsIwE11to_int_typeEw($690)|0);
         $$pre$i4$i$i322 = $905;$711 = $905;
        }
        while(1) {
         $691 = HEAP32[$24>>2]|0;
         $692 = ($691|0)>(0);
         if (!($692)) {
          break L219;
         }
         $693 = HEAP32[$0>>2]|0;
         $694 = ($693|0)==(0|0);
         do {
          if ($694) {
           $727 = 1;
          } else {
           $695 = ((($693)) + 12|0);
           $696 = HEAP32[$695>>2]|0;
           $697 = ((($693)) + 16|0);
           $698 = HEAP32[$697>>2]|0;
           $699 = ($696|0)==($698|0);
           if ($699) {
            $700 = HEAP32[$693>>2]|0;
            $701 = ((($700)) + 36|0);
            $702 = HEAP32[$701>>2]|0;
            __THREW__ = 0;
            $703 = (invoke_ii($702|0,($693|0))|0);
            $704 = __THREW__; __THREW__ = 0;
            $705 = $704&1;
            if ($705) {
             label = 245;
             break L24;
            } else {
             $$0$i$i$i$i314 = $703;
            }
           } else {
            $706 = HEAP32[$696>>2]|0;
            $707 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($706)|0);
            $$0$i$i$i$i314 = $707;
           }
           $708 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
           $709 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i314,$708)|0);
           if ($709) {
            HEAP32[$0>>2] = 0;
            $727 = 1;
            break;
           } else {
            $$pre$i$i$i316 = HEAP32[$0>>2]|0;
            $phitmp$i$i$i317 = ($$pre$i$i$i316|0)==(0|0);
            $727 = $phitmp$i$i$i317;
            break;
           }
          }
         } while(0);
         $710 = ($711|0)==(0|0);
         do {
          if ($710) {
           $$pre$i4$i$i322516$ph = $$pre$i4$i$i322;
           label = 228;
          } else {
           $712 = ((($711)) + 12|0);
           $713 = HEAP32[$712>>2]|0;
           $714 = ((($711)) + 16|0);
           $715 = HEAP32[$714>>2]|0;
           $716 = ($713|0)==($715|0);
           if ($716) {
            $717 = HEAP32[$711>>2]|0;
            $718 = ((($717)) + 36|0);
            $719 = HEAP32[$718>>2]|0;
            __THREW__ = 0;
            $720 = (invoke_ii($719|0,($711|0))|0);
            $721 = __THREW__; __THREW__ = 0;
            $722 = $721&1;
            if ($722) {
             label = 245;
             break L24;
            } else {
             $$0$i$i2$i$i320 = $720;
            }
           } else {
            $723 = HEAP32[$713>>2]|0;
            $724 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($723)|0);
            $$0$i$i2$i$i320 = $724;
           }
           $725 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
           $726 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i320,$725)|0);
           if ($726) {
            HEAP32[$1>>2] = 0;
            $$pre$i4$i$i322516$ph = 0;
            label = 228;
            break;
           } else {
            $phitmp$i5$i$i323 = ($$pre$i4$i$i322|0)==(0|0);
            $tmp$i$i325 = $727 ^ $phitmp$i5$i$i323;
            if ($tmp$i$i325) {
             $$pre$i4$i$i322516533 = $$pre$i4$i$i322;$906 = $$pre$i4$i$i322;
             break;
            } else {
             label = 244;
             break L24;
            }
           }
          }
         } while(0);
         if ((label|0) == 228) {
          label = 0;
          if ($727) {
           label = 244;
           break L24;
          } else {
           $$pre$i4$i$i322516533 = $$pre$i4$i$i322516$ph;$906 = 0;
          }
         }
         $728 = HEAP32[$0>>2]|0;
         $729 = ((($728)) + 12|0);
         $730 = HEAP32[$729>>2]|0;
         $731 = ((($728)) + 16|0);
         $732 = HEAP32[$731>>2]|0;
         $733 = ($730|0)==($732|0);
         if ($733) {
          $734 = HEAP32[$728>>2]|0;
          $735 = ((($734)) + 36|0);
          $736 = HEAP32[$735>>2]|0;
          __THREW__ = 0;
          $737 = (invoke_ii($736|0,($728|0))|0);
          $738 = __THREW__; __THREW__ = 0;
          $739 = $738&1;
          if ($739) {
           label = 245;
           break L24;
          } else {
           $$0$i$i329 = $737;
          }
         } else {
          $740 = HEAP32[$730>>2]|0;
          $741 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($740)|0);
          $$0$i$i329 = $741;
         }
         $742 = HEAP32[$7>>2]|0;
         $743 = ((($742)) + 12|0);
         $744 = HEAP32[$743>>2]|0;
         __THREW__ = 0;
         $745 = (invoke_iiii($744|0,($7|0),2048,($$0$i$i329|0))|0);
         $746 = __THREW__; __THREW__ = 0;
         $747 = $746&1;
         if ($747) {
          label = 245;
          break L24;
         }
         if (!($745)) {
          label = 244;
          break L24;
         }
         $748 = HEAP32[$9>>2]|0;
         $749 = HEAP32[$11>>2]|0;
         $750 = ($748|0)==($749|0);
         if ($750) {
          __THREW__ = 0;
          invoke_viii(1017,($8|0),($9|0),($11|0));
          $751 = __THREW__; __THREW__ = 0;
          $752 = $751&1;
          if ($752) {
           label = 245;
           break L24;
          }
         }
         $753 = HEAP32[$0>>2]|0;
         $754 = ((($753)) + 12|0);
         $755 = HEAP32[$754>>2]|0;
         $756 = ((($753)) + 16|0);
         $757 = HEAP32[$756>>2]|0;
         $758 = ($755|0)==($757|0);
         if ($758) {
          $759 = HEAP32[$753>>2]|0;
          $760 = ((($759)) + 36|0);
          $761 = HEAP32[$760>>2]|0;
          __THREW__ = 0;
          $762 = (invoke_ii($761|0,($753|0))|0);
          $763 = __THREW__; __THREW__ = 0;
          $764 = $763&1;
          if ($764) {
           label = 245;
           break L24;
          } else {
           $$0$i$i334 = $762;
          }
         } else {
          $765 = HEAP32[$755>>2]|0;
          $766 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($765)|0);
          $$0$i$i334 = $766;
         }
         $767 = HEAP32[$9>>2]|0;
         $768 = ((($767)) + 4|0);
         HEAP32[$9>>2] = $768;
         HEAP32[$767>>2] = $$0$i$i334;
         $769 = HEAP32[$24>>2]|0;
         $770 = (($769) + -1)|0;
         HEAP32[$24>>2] = $770;
         $771 = HEAP32[$0>>2]|0;
         $772 = ((($771)) + 12|0);
         $773 = HEAP32[$772>>2]|0;
         $774 = ((($771)) + 16|0);
         $775 = HEAP32[$774>>2]|0;
         $776 = ($773|0)==($775|0);
         if ($776) {
          $777 = HEAP32[$771>>2]|0;
          $778 = ((($777)) + 40|0);
          $779 = HEAP32[$778>>2]|0;
          __THREW__ = 0;
          (invoke_ii($779|0,($771|0))|0);
          $780 = __THREW__; __THREW__ = 0;
          $781 = $780&1;
          if ($781) {
           label = 245;
           break L24;
          } else {
           $$pre$i4$i$i322 = $$pre$i4$i$i322516533;$711 = $906;
           continue;
          }
         } else {
          $782 = ((($773)) + 4|0);
          HEAP32[$772>>2] = $782;
          $783 = HEAP32[$773>>2]|0;
          (__ZNSt3__211char_traitsIwE11to_int_typeEw($783)|0);
          $$pre$i4$i$i322 = $$pre$i4$i$i322516533;$711 = $906;
          continue;
         }
        }
       }
      } while(0);
      $784 = HEAP32[$9>>2]|0;
      $785 = HEAP32[$8>>2]|0;
      $786 = ($784|0)==($785|0);
      if ($786) {
       label = 244;
       break L24;
      } else {
       $$1153 = $$0152;
      }
      break;
     }
     default: {
      $$1153 = $$0152;
     }
     }
    } while(0);
    L293: do {
     if ((label|0) == 49) {
      while(1) {
       label = 0;
       $142 = HEAP32[$0>>2]|0;
       $143 = ($142|0)==(0|0);
       do {
        if ($143) {
         $176 = 1;
        } else {
         $144 = ((($142)) + 12|0);
         $145 = HEAP32[$144>>2]|0;
         $146 = ((($142)) + 16|0);
         $147 = HEAP32[$146>>2]|0;
         $148 = ($145|0)==($147|0);
         if ($148) {
          $149 = HEAP32[$142>>2]|0;
          $150 = ((($149)) + 36|0);
          $151 = HEAP32[$150>>2]|0;
          __THREW__ = 0;
          $152 = (invoke_ii($151|0,($142|0))|0);
          $153 = __THREW__; __THREW__ = 0;
          $154 = $153&1;
          if ($154) {
           label = 35;
           break L24;
          } else {
           $$0$i$i$i$i189 = $152;
          }
         } else {
          $155 = HEAP32[$145>>2]|0;
          $156 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($155)|0);
          $$0$i$i$i$i189 = $156;
         }
         $157 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $158 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i189,$157)|0);
         if ($158) {
          HEAP32[$0>>2] = 0;
          $176 = 1;
          break;
         } else {
          $$pre$i$i$i191 = HEAP32[$0>>2]|0;
          $phitmp$i$i$i192 = ($$pre$i$i$i191|0)==(0|0);
          $176 = $phitmp$i$i$i192;
          break;
         }
        }
       } while(0);
       $159 = ($160|0)==(0|0);
       do {
        if ($159) {
         $$pre$i4$i$i197527 = $$pre$i4$i$i197;
         label = 62;
        } else {
         $161 = ((($160)) + 12|0);
         $162 = HEAP32[$161>>2]|0;
         $163 = ((($160)) + 16|0);
         $164 = HEAP32[$163>>2]|0;
         $165 = ($162|0)==($164|0);
         if ($165) {
          $166 = HEAP32[$160>>2]|0;
          $167 = ((($166)) + 36|0);
          $168 = HEAP32[$167>>2]|0;
          __THREW__ = 0;
          $169 = (invoke_ii($168|0,($160|0))|0);
          $170 = __THREW__; __THREW__ = 0;
          $171 = $170&1;
          if ($171) {
           label = 35;
           break L24;
          } else {
           $$0$i$i2$i$i195 = $169;
          }
         } else {
          $172 = HEAP32[$162>>2]|0;
          $173 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($172)|0);
          $$0$i$i2$i$i195 = $173;
         }
         $174 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
         $175 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i195,$174)|0);
         if ($175) {
          HEAP32[$1>>2] = 0;
          $$pre$i4$i$i197527 = 0;
          label = 62;
          break;
         } else {
          $phitmp$i5$i$i198 = ($$pre$i4$i$i197|0)==(0|0);
          $tmp$i$i200 = $176 ^ $phitmp$i5$i$i198;
          if ($tmp$i$i200) {
           $$pre$i4$i$i197529 = $$pre$i4$i$i197;$901 = $$pre$i4$i$i197;
           break;
          } else {
           $$1153 = $$0152;
           break L293;
          }
         }
        }
       } while(0);
       if ((label|0) == 62) {
        label = 0;
        if ($176) {
         $$1153 = $$0152;
         break L293;
        } else {
         $$pre$i4$i$i197529 = $$pre$i4$i$i197527;$901 = 0;
        }
       }
       $177 = HEAP32[$0>>2]|0;
       $178 = ((($177)) + 12|0);
       $179 = HEAP32[$178>>2]|0;
       $180 = ((($177)) + 16|0);
       $181 = HEAP32[$180>>2]|0;
       $182 = ($179|0)==($181|0);
       if ($182) {
        $183 = HEAP32[$177>>2]|0;
        $184 = ((($183)) + 36|0);
        $185 = HEAP32[$184>>2]|0;
        __THREW__ = 0;
        $186 = (invoke_ii($185|0,($177|0))|0);
        $187 = __THREW__; __THREW__ = 0;
        $188 = $187&1;
        if ($188) {
         label = 35;
         break L24;
        } else {
         $$0$i$i204 = $186;
        }
       } else {
        $189 = HEAP32[$179>>2]|0;
        $190 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($189)|0);
        $$0$i$i204 = $190;
       }
       $191 = HEAP32[$7>>2]|0;
       $192 = ((($191)) + 12|0);
       $193 = HEAP32[$192>>2]|0;
       __THREW__ = 0;
       $194 = (invoke_iiii($193|0,($7|0),8192,($$0$i$i204|0))|0);
       $195 = __THREW__; __THREW__ = 0;
       $196 = $195&1;
       if ($196) {
        label = 35;
        break L24;
       }
       if (!($194)) {
        $$1153 = $$0152;
        break L293;
       }
       $197 = HEAP32[$0>>2]|0;
       $198 = ((($197)) + 12|0);
       $199 = HEAP32[$198>>2]|0;
       $200 = ((($197)) + 16|0);
       $201 = HEAP32[$200>>2]|0;
       $202 = ($199|0)==($201|0);
       if ($202) {
        $203 = HEAP32[$197>>2]|0;
        $204 = ((($203)) + 40|0);
        $205 = HEAP32[$204>>2]|0;
        __THREW__ = 0;
        $206 = (invoke_ii($205|0,($197|0))|0);
        $207 = __THREW__; __THREW__ = 0;
        $208 = $207&1;
        if ($208) {
         label = 35;
         break L24;
        } else {
         $$0$i$i209 = $206;
        }
       } else {
        $209 = ((($199)) + 4|0);
        HEAP32[$198>>2] = $209;
        $210 = HEAP32[$199>>2]|0;
        $211 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($210)|0);
        $$0$i$i209 = $211;
       }
       __THREW__ = 0;
       invoke_vii(1014,($23|0),($$0$i$i209|0));
       $212 = __THREW__; __THREW__ = 0;
       $213 = $212&1;
       if ($213) {
        label = 35;
        break L24;
       } else {
        $$pre$i4$i$i197 = $$pre$i4$i$i197529;$160 = $901;
        label = 49;
       }
      }
     }
    } while(0);
    $791 = (($$0161) + 1)|0;
    $$0152 = $$1153;$$0161 = $791;
   }
   L329:    switch (label|0) {
    case 35: {
     $97 = ___cxa_find_matching_catch_2()|0;
     $98 = tempRet0;
     $$5145 = $97;$$5151 = $98;
     break L21;
     break;
    }
    case 47: {
     $140 = HEAP32[$5>>2]|0;
     $141 = $140 | 4;
     HEAP32[$5>>2] = $141;
     $$10 = 0;
     break;
    }
    case 109: {
     $372 = HEAP32[$5>>2]|0;
     $373 = $372 | 4;
     HEAP32[$5>>2] = $373;
     $$10 = 0;
     break;
    }
    case 119: {
     $407 = ___cxa_find_matching_catch_2()|0;
     $408 = tempRet0;
     $$5145 = $407;$$5151 = $408;
     break L21;
     break;
    }
    case 149: {
     $510 = ___cxa_find_matching_catch_2()|0;
     $511 = tempRet0;
     $$5145 = $510;$$5151 = $511;
     break L21;
     break;
    }
    case 150: {
     $512 = ___cxa_find_matching_catch_2()|0;
     $513 = tempRet0;
     $$5145 = $512;$$5151 = $513;
     break L21;
     break;
    }
    case 153: {
     $523 = HEAP32[$5>>2]|0;
     $524 = $523 | 4;
     HEAP32[$5>>2] = $524;
     $$10 = 0;
     break;
    }
    case 244: {
     $787 = HEAP32[$5>>2]|0;
     $788 = $787 | 4;
     HEAP32[$5>>2] = $788;
     $$10 = 0;
     break;
    }
    case 245: {
     $789 = ___cxa_find_matching_catch_2()|0;
     $790 = tempRet0;
     $$5145 = $789;$$5151 = $790;
     break L21;
     break;
    }
    case 247: {
     L331: do {
      if ($60) {
       $792 = ((($$0152)) + 8|0);
       $793 = ((($792)) + 3|0);
       $794 = ((($$0152)) + 4|0);
       $$0 = 1;
       L333: while(1) {
        $795 = HEAP8[$793>>0]|0;
        $796 = ($795<<24>>24)<(0);
        if ($796) {
         $797 = HEAP32[$794>>2]|0;
         $800 = $797;
        } else {
         $798 = $795&255;
         $800 = $798;
        }
        $799 = ($$0>>>0)<($800>>>0);
        if (!($799)) {
         break L331;
        }
        $801 = HEAP32[$0>>2]|0;
        $802 = ($801|0)==(0|0);
        do {
         if ($802) {
          $907 = 1;
         } else {
          $803 = ((($801)) + 12|0);
          $804 = HEAP32[$803>>2]|0;
          $805 = ((($801)) + 16|0);
          $806 = HEAP32[$805>>2]|0;
          $807 = ($804|0)==($806|0);
          if ($807) {
           $808 = HEAP32[$801>>2]|0;
           $809 = ((($808)) + 36|0);
           $810 = HEAP32[$809>>2]|0;
           __THREW__ = 0;
           $811 = (invoke_ii($810|0,($801|0))|0);
           $812 = __THREW__; __THREW__ = 0;
           $813 = $812&1;
           if ($813) {
            label = 275;
            break L333;
           } else {
            $$0$i$i$i$i340 = $811;
           }
          } else {
           $814 = HEAP32[$804>>2]|0;
           $815 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($814)|0);
           $$0$i$i$i$i340 = $815;
          }
          $816 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
          $817 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i$i$i340,$816)|0);
          if ($817) {
           HEAP32[$0>>2] = 0;
           $907 = 1;
           break;
          } else {
           $$pre$i$i$i342 = HEAP32[$0>>2]|0;
           $phitmp$i$i$i343 = ($$pre$i$i$i342|0)==(0|0);
           $907 = $phitmp$i$i$i343;
           break;
          }
         }
        } while(0);
        $818 = HEAP32[$1>>2]|0;
        $819 = ($818|0)==(0|0);
        do {
         if ($819) {
          label = 266;
         } else {
          $820 = ((($818)) + 12|0);
          $821 = HEAP32[$820>>2]|0;
          $822 = ((($818)) + 16|0);
          $823 = HEAP32[$822>>2]|0;
          $824 = ($821|0)==($823|0);
          if ($824) {
           $825 = HEAP32[$818>>2]|0;
           $826 = ((($825)) + 36|0);
           $827 = HEAP32[$826>>2]|0;
           __THREW__ = 0;
           $828 = (invoke_ii($827|0,($818|0))|0);
           $829 = __THREW__; __THREW__ = 0;
           $830 = $829&1;
           if ($830) {
            label = 275;
            break L333;
           } else {
            $$0$i$i2$i$i346 = $828;
           }
          } else {
           $831 = HEAP32[$821>>2]|0;
           $832 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($831)|0);
           $$0$i$i2$i$i346 = $832;
          }
          $833 = (__ZNSt3__211char_traitsIwE3eofEv()|0);
          $834 = (__ZNSt3__211char_traitsIwE11eq_int_typeEjj($$0$i$i2$i$i346,$833)|0);
          if ($834) {
           HEAP32[$1>>2] = 0;
           label = 266;
           break;
          } else {
           if ($907) {
            break;
           } else {
            label = 274;
            break L333;
           }
          }
         }
        } while(0);
        if ((label|0) == 266) {
         label = 0;
         if ($907) {
          label = 274;
          break;
         }
        }
        $835 = HEAP32[$0>>2]|0;
        $836 = ((($835)) + 12|0);
        $837 = HEAP32[$836>>2]|0;
        $838 = ((($835)) + 16|0);
        $839 = HEAP32[$838>>2]|0;
        $840 = ($837|0)==($839|0);
        if ($840) {
         $841 = HEAP32[$835>>2]|0;
         $842 = ((($841)) + 36|0);
         $843 = HEAP32[$842>>2]|0;
         __THREW__ = 0;
         $844 = (invoke_ii($843|0,($835|0))|0);
         $845 = __THREW__; __THREW__ = 0;
         $846 = $845&1;
         if ($846) {
          label = 275;
          break;
         } else {
          $$0$i$i355 = $844;
         }
        } else {
         $847 = HEAP32[$837>>2]|0;
         $848 = (__ZNSt3__211char_traitsIwE11to_int_typeEw($847)|0);
         $$0$i$i355 = $848;
        }
        $849 = HEAP8[$793>>0]|0;
        $850 = ($849<<24>>24)<(0);
        if ($850) {
         $851 = HEAP32[$$0152>>2]|0;
         $853 = $851;
        } else {
         $853 = $$0152;
        }
        $852 = (($853) + ($$0<<2)|0);
        $854 = HEAP32[$852>>2]|0;
        $855 = ($$0$i$i355|0)==($854|0);
        if (!($855)) {
         label = 274;
         break;
        }
        $860 = (($$0) + 1)|0;
        $861 = HEAP32[$0>>2]|0;
        $862 = ((($861)) + 12|0);
        $863 = HEAP32[$862>>2]|0;
        $864 = ((($861)) + 16|0);
        $865 = HEAP32[$864>>2]|0;
        $866 = ($863|0)==($865|0);
        if ($866) {
         $867 = HEAP32[$861>>2]|0;
         $868 = ((($867)) + 40|0);
         $869 = HEAP32[$868>>2]|0;
         __THREW__ = 0;
         (invoke_ii($869|0,($861|0))|0);
         $870 = __THREW__; __THREW__ = 0;
         $871 = $870&1;
         if ($871) {
          label = 275;
          break;
         } else {
          $$0 = $860;
          continue;
         }
        } else {
         $872 = ((($863)) + 4|0);
         HEAP32[$862>>2] = $872;
         $873 = HEAP32[$863>>2]|0;
         (__ZNSt3__211char_traitsIwE11to_int_typeEw($873)|0);
         $$0 = $860;
         continue;
        }
       }
       if ((label|0) == 274) {
        $856 = HEAP32[$5>>2]|0;
        $857 = $856 | 4;
        HEAP32[$5>>2] = $857;
        $$10 = 0;
        break L329;
       }
       else if ((label|0) == 275) {
        $858 = ___cxa_find_matching_catch_2()|0;
        $859 = tempRet0;
        $$5145 = $858;$$5151 = $859;
        break L21;
       }
      }
     } while(0);
     $874 = HEAP32[$13>>2]|0;
     $875 = HEAP32[$14>>2]|0;
     $876 = ($874|0)==($875|0);
     if ($876) {
      $$10 = 1;
     } else {
      HEAP32[$25>>2] = 0;
      __THREW__ = 0;
      invoke_viiii(963,($19|0),($874|0),($875|0),($25|0));
      $877 = __THREW__; __THREW__ = 0;
      $878 = $877&1;
      if ($878) {
       $883 = ___cxa_find_matching_catch_2()|0;
       $884 = tempRet0;
       $$5145 = $883;$$5151 = $884;
       break L21;
      }
      $879 = HEAP32[$25>>2]|0;
      $880 = ($879|0)==(0);
      if ($880) {
       $$10 = 1;
       break;
      } else {
       $881 = HEAP32[$5>>2]|0;
       $882 = $881 | 4;
       HEAP32[$5>>2] = $882;
       $$10 = 0;
       break;
      }
     }
     break;
    }
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($23);
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($22);
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($21);
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($20);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19);
   $885 = HEAP32[$13>>2]|0;
   HEAP32[$13>>2] = 0;
   $886 = ($885|0)==(0|0);
   if (!($886)) {
    $887 = HEAP32[$27>>2]|0;
    __THREW__ = 0;
    invoke_vi($887|0,($885|0));
    $888 = __THREW__; __THREW__ = 0;
    $889 = $888&1;
    if ($889) {
     $890 = ___cxa_find_matching_catch_3(0|0)|0;
     $891 = tempRet0;
     ___clang_call_terminate($890);
     // unreachable;
    }
   }
   STACKTOP = sp;return ($$10|0);
  }
 } while(0);
 __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($23);
 __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($22);
 __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($21);
 __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($20);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($19);
 $892 = HEAP32[$13>>2]|0;
 HEAP32[$13>>2] = 0;
 $893 = ($892|0)==(0|0);
 if (!($893)) {
  $894 = HEAP32[$27>>2]|0;
  __THREW__ = 0;
  invoke_vi($894|0,($892|0));
  $895 = __THREW__; __THREW__ = 0;
  $896 = $895&1;
  if ($896) {
   $897 = ___cxa_find_matching_catch_3(0|0)|0;
   $898 = tempRet0;
   ___clang_call_terminate($897);
   // unreachable;
  }
 }
 ___resumeException($$5145|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE23__append_forward_unsafeIPwEERS5_T_S9_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$021$i$i = 0, $$040 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 16|0;
 $4 = sp + 4|0;
 $5 = sp;
 $6 = ((($0)) + 8|0);
 $7 = ((($6)) + 3|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)<(0);
 if ($9) {
  $11 = ((($0)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$6>>2]|0;
  $14 = $13 & 2147483647;
  $phitmp$i = (($14) + -1)|0;
  $65 = $12;$66 = $phitmp$i;
 } else {
  $10 = $8&255;
  $65 = $10;$66 = 1;
 }
 $15 = $2;
 $16 = $1;
 $17 = (($15) - ($16))|0;
 $18 = $17 >> 2;
 $19 = ($18|0)==(0);
 do {
  if (!($19)) {
   if ($9) {
    $20 = HEAP32[$0>>2]|0;
    $21 = ((($0)) + 4|0);
    $22 = HEAP32[$21>>2]|0;
    $25 = $20;$26 = $22;$27 = $20;
   } else {
    $23 = $8&255;
    $25 = $0;$26 = $23;$27 = $0;
   }
   $24 = (($25) + ($26<<2)|0);
   $28 = (__ZNSt3__214__ptr_in_rangeIwEEbPKT_S3_S3_($1,$27,$24)|0);
   if (!($28)) {
    $64 = (($66) - ($65))|0;
    $67 = ($64>>>0)<($18>>>0);
    $68 = (($65) + ($18))|0;
    if ($67) {
     $69 = (($68) - ($66))|0;
     __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($0,$66,$69,$65,$65,0,0);
    }
    $70 = HEAP8[$7>>0]|0;
    $71 = ($70<<24>>24)<(0);
    if ($71) {
     $72 = HEAP32[$0>>2]|0;
     $74 = $72;
    } else {
     $74 = $0;
    }
    $73 = (($74) + ($65<<2)|0);
    $$0 = $73;$$040 = $1;
    while(1) {
     $75 = ($$040|0)==($2|0);
     if ($75) {
      break;
     }
     __ZNSt3__211char_traitsIwE6assignERwRKw($$0,$$040);
     $76 = ((($$0)) + 4|0);
     $77 = ((($$040)) + 4|0);
     $$0 = $76;$$040 = $77;
    }
    HEAP32[$5>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($$0,$5);
    $78 = HEAP8[$7>>0]|0;
    $79 = ($78<<24>>24)<(0);
    if ($79) {
     $80 = ((($0)) + 4|0);
     HEAP32[$80>>2] = $68;
     break;
    } else {
     $81 = $68&255;
     HEAP8[$7>>0] = $81;
     break;
    }
   }
   ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
   $29 = ($18>>>0)>(1073741807);
   if ($29) {
    __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($4);
    // unreachable;
   }
   $30 = ($18>>>0)<(2);
   do {
    if ($30) {
     $31 = $18&255;
     $32 = ((($4)) + 8|0);
     $33 = ((($32)) + 3|0);
     HEAP8[$33>>0] = $31;
     $$0$i$i = $1;$$021$i$i = $4;
    } else {
     $34 = (($18) + 4)|0;
     $35 = $34 & -4;
     $36 = ($35>>>0)>(1073741823);
     if (!($36)) {
      $42 = $35 << 2;
      $43 = (__Znwj($42)|0);
      HEAP32[$4>>2] = $43;
      $44 = $35 | -2147483648;
      $45 = ((($4)) + 8|0);
      HEAP32[$45>>2] = $44;
      $46 = ((($4)) + 4|0);
      HEAP32[$46>>2] = $18;
      $$0$i$i = $1;$$021$i$i = $43;
      break;
     }
     $37 = (___cxa_allocate_exception(8)|0);
     __THREW__ = 0;
     invoke_vii(751,($37|0),(158021|0));
     $38 = __THREW__; __THREW__ = 0;
     $39 = $38&1;
     if ($39) {
      $40 = ___cxa_find_matching_catch_2()|0;
      $41 = tempRet0;
      ___cxa_free_exception(($37|0));
      ___resumeException($40|0);
      // unreachable;
     } else {
      HEAP32[$37>>2] = (22172);
      ___cxa_throw(($37|0),(2528|0),(713|0));
      // unreachable;
     }
    }
   } while(0);
   while(1) {
    $47 = ($$0$i$i|0)==($2|0);
    if ($47) {
     break;
    }
    __ZNSt3__211char_traitsIwE6assignERwRKw($$021$i$i,$$0$i$i);
    $48 = ((($$0$i$i)) + 4|0);
    $49 = ((($$021$i$i)) + 4|0);
    $$0$i$i = $48;$$021$i$i = $49;
   }
   HEAP32[$3>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($$021$i$i,$3);
   $50 = ((($4)) + 8|0);
   $51 = ((($50)) + 3|0);
   $52 = HEAP8[$51>>0]|0;
   $53 = ($52<<24>>24)<(0);
   $54 = HEAP32[$4>>2]|0;
   $55 = $53 ? $54 : $4;
   $56 = ((($4)) + 4|0);
   $57 = HEAP32[$56>>2]|0;
   $58 = $52&255;
   $59 = $53 ? $57 : $58;
   __THREW__ = 0;
   (invoke_iiii(1018,($0|0),($55|0),($59|0))|0);
   $60 = __THREW__; __THREW__ = 0;
   $61 = $60&1;
   if ($61) {
    $62 = ___cxa_find_matching_catch_2()|0;
    $63 = tempRet0;
    __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($4);
    ___resumeException($62|0);
    // unreachable;
   } else {
    __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($4);
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__214__ptr_in_rangeIwEEbPKT_S3_S3_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $3 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0>>>0)<($2>>>0);
 $not$ = ($1>>>0)<=($0>>>0);
 $$ = $not$ & $3;
 return ($$|0);
}
function __ZNSt3__211__money_getIwE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_SJ_Ri($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$i$i = 0, $$0$i$i$i40 = 0, $$0$i$i$i43 = 0, $$0$i$i$i47 = 0, $$0$i$i$i52 = 0, $$0$i$i$i57 = 0, $$0$i$i$i62 = 0, $$0$i$i$i67 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0;
 var $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i41 = 0, $exitcond$i$i$i44 = 0, $exitcond$i$i$i48 = 0, $exitcond$i$i$i53 = 0, $exitcond$i$i$i58 = 0, $exitcond$i$i$i63 = 0, $exitcond$i$i$i68 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(160|0);
 $10 = sp + 148|0;
 $11 = sp + 144|0;
 $12 = sp + 155|0;
 $13 = sp + 154|0;
 $14 = sp + 140|0;
 $15 = sp + 136|0;
 $16 = sp + 132|0;
 $17 = sp + 128|0;
 $18 = sp + 124|0;
 $19 = sp + 120|0;
 $20 = sp + 153|0;
 $21 = sp + 152|0;
 $22 = sp + 116|0;
 $23 = sp + 112|0;
 $24 = sp + 108|0;
 $25 = sp + 104|0;
 $26 = sp + 100|0;
 $27 = sp + 88|0;
 $28 = sp + 76|0;
 $29 = sp + 64|0;
 $30 = sp + 52|0;
 $31 = sp + 48|0;
 $32 = sp + 36|0;
 $33 = sp + 24|0;
 $34 = sp + 12|0;
 $35 = sp;
 if ($0) {
  $36 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167176)|0);
  $37 = HEAP32[$36>>2]|0;
  $38 = ((($37)) + 44|0);
  $39 = HEAP32[$38>>2]|0;
  FUNCTION_TABLE_vii[$39 & 2047]($26,$36);
  $40 = HEAP32[$26>>2]|0;
  HEAP8[$2>>0]=$40&255;HEAP8[$2+1>>0]=($40>>8)&255;HEAP8[$2+2>>0]=($40>>16)&255;HEAP8[$2+3>>0]=$40>>24;
  $41 = HEAP32[$36>>2]|0;
  $42 = ((($41)) + 32|0);
  $43 = HEAP32[$42>>2]|0;
  FUNCTION_TABLE_vii[$43 & 2047]($27,$36);
  $44 = ((($8)) + 8|0);
  $45 = ((($44)) + 3|0);
  $46 = HEAP8[$45>>0]|0;
  $47 = ($46<<24>>24)<(0);
  if ($47) {
   $48 = HEAP32[$8>>2]|0;
   HEAP32[$24>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($48,$24);
   $49 = ((($8)) + 4|0);
   HEAP32[$49>>2] = 0;
  } else {
   HEAP32[$25>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($8,$25);
   HEAP8[$45>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($8|0),0);
  $50 = __THREW__; __THREW__ = 0;
  $51 = $50&1;
  if ($51) {
   $52 = ___cxa_find_matching_catch_3(0|0)|0;
   $53 = tempRet0;
   ___clang_call_terminate($52);
   // unreachable;
  }
  ;HEAP32[$8>>2]=HEAP32[$27>>2]|0;HEAP32[$8+4>>2]=HEAP32[$27+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$27+8>>2]|0;
  $$0$i$i$i = 0;
  while(1) {
   $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
   if ($exitcond$i$i$i) {
    break;
   }
   $54 = (($27) + ($$0$i$i$i<<2)|0);
   HEAP32[$54>>2] = 0;
   $55 = (($$0$i$i$i) + 1)|0;
   $$0$i$i$i = $55;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($27);
  $56 = HEAP32[$36>>2]|0;
  $57 = ((($56)) + 28|0);
  $58 = HEAP32[$57>>2]|0;
  FUNCTION_TABLE_vii[$58 & 2047]($28,$36);
  $59 = ((($7)) + 8|0);
  $60 = ((($59)) + 3|0);
  $61 = HEAP8[$60>>0]|0;
  $62 = ($61<<24>>24)<(0);
  if ($62) {
   $63 = HEAP32[$7>>2]|0;
   HEAP32[$22>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($63,$22);
   $64 = ((($7)) + 4|0);
   HEAP32[$64>>2] = 0;
  } else {
   HEAP32[$23>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($7,$23);
   HEAP8[$60>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($7|0),0);
  $65 = __THREW__; __THREW__ = 0;
  $66 = $65&1;
  if ($66) {
   $67 = ___cxa_find_matching_catch_3(0|0)|0;
   $68 = tempRet0;
   ___clang_call_terminate($67);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$28>>2]|0;HEAP32[$7+4>>2]=HEAP32[$28+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$28+8>>2]|0;
  $$0$i$i$i40 = 0;
  while(1) {
   $exitcond$i$i$i41 = ($$0$i$i$i40|0)==(3);
   if ($exitcond$i$i$i41) {
    break;
   }
   $69 = (($28) + ($$0$i$i$i40<<2)|0);
   HEAP32[$69>>2] = 0;
   $70 = (($$0$i$i$i40) + 1)|0;
   $$0$i$i$i40 = $70;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($28);
  $71 = HEAP32[$36>>2]|0;
  $72 = ((($71)) + 12|0);
  $73 = HEAP32[$72>>2]|0;
  $74 = (FUNCTION_TABLE_ii[$73 & 2047]($36)|0);
  HEAP32[$3>>2] = $74;
  $75 = HEAP32[$36>>2]|0;
  $76 = ((($75)) + 16|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = (FUNCTION_TABLE_ii[$77 & 2047]($36)|0);
  HEAP32[$4>>2] = $78;
  $79 = HEAP32[$36>>2]|0;
  $80 = ((($79)) + 20|0);
  $81 = HEAP32[$80>>2]|0;
  FUNCTION_TABLE_vii[$81 & 2047]($29,$36);
  $82 = ((($5)) + 11|0);
  $83 = HEAP8[$82>>0]|0;
  $84 = ($83<<24>>24)<(0);
  if ($84) {
   $85 = HEAP32[$5>>2]|0;
   HEAP8[$20>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($85,$20);
   $86 = ((($5)) + 4|0);
   HEAP32[$86>>2] = 0;
  } else {
   HEAP8[$21>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($5,$21);
   HEAP8[$82>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($5|0),0);
  $87 = __THREW__; __THREW__ = 0;
  $88 = $87&1;
  if ($88) {
   $89 = ___cxa_find_matching_catch_3(0|0)|0;
   $90 = tempRet0;
   ___clang_call_terminate($89);
   // unreachable;
  }
  ;HEAP32[$5>>2]=HEAP32[$29>>2]|0;HEAP32[$5+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$29+8>>2]|0;
  $$0$i$i$i43 = 0;
  while(1) {
   $exitcond$i$i$i44 = ($$0$i$i$i43|0)==(3);
   if ($exitcond$i$i$i44) {
    break;
   }
   $91 = (($29) + ($$0$i$i$i43<<2)|0);
   HEAP32[$91>>2] = 0;
   $92 = (($$0$i$i$i43) + 1)|0;
   $$0$i$i$i43 = $92;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($29);
  $93 = HEAP32[$36>>2]|0;
  $94 = ((($93)) + 24|0);
  $95 = HEAP32[$94>>2]|0;
  FUNCTION_TABLE_vii[$95 & 2047]($30,$36);
  $96 = ((($6)) + 8|0);
  $97 = ((($96)) + 3|0);
  $98 = HEAP8[$97>>0]|0;
  $99 = ($98<<24>>24)<(0);
  if ($99) {
   $100 = HEAP32[$6>>2]|0;
   HEAP32[$18>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($100,$18);
   $101 = ((($6)) + 4|0);
   HEAP32[$101>>2] = 0;
  } else {
   HEAP32[$19>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($6,$19);
   HEAP8[$97>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($6|0),0);
  $102 = __THREW__; __THREW__ = 0;
  $103 = $102&1;
  if ($103) {
   $104 = ___cxa_find_matching_catch_3(0|0)|0;
   $105 = tempRet0;
   ___clang_call_terminate($104);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$30>>2]|0;HEAP32[$6+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$30+8>>2]|0;
  $$0$i$i$i47 = 0;
  while(1) {
   $exitcond$i$i$i48 = ($$0$i$i$i47|0)==(3);
   if ($exitcond$i$i$i48) {
    break;
   }
   $106 = (($30) + ($$0$i$i$i47<<2)|0);
   HEAP32[$106>>2] = 0;
   $107 = (($$0$i$i$i47) + 1)|0;
   $$0$i$i$i47 = $107;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($30);
  $108 = HEAP32[$36>>2]|0;
  $109 = ((($108)) + 36|0);
  $110 = HEAP32[$109>>2]|0;
  $111 = (FUNCTION_TABLE_ii[$110 & 2047]($36)|0);
  $storemerge = $111;
 } else {
  $112 = (__ZNKSt3__26locale9use_facetERNS0_2idE($1,167168)|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = ((($113)) + 44|0);
  $115 = HEAP32[$114>>2]|0;
  FUNCTION_TABLE_vii[$115 & 2047]($31,$112);
  $116 = HEAP32[$31>>2]|0;
  HEAP8[$2>>0]=$116&255;HEAP8[$2+1>>0]=($116>>8)&255;HEAP8[$2+2>>0]=($116>>16)&255;HEAP8[$2+3>>0]=$116>>24;
  $117 = HEAP32[$112>>2]|0;
  $118 = ((($117)) + 32|0);
  $119 = HEAP32[$118>>2]|0;
  FUNCTION_TABLE_vii[$119 & 2047]($32,$112);
  $120 = ((($8)) + 8|0);
  $121 = ((($120)) + 3|0);
  $122 = HEAP8[$121>>0]|0;
  $123 = ($122<<24>>24)<(0);
  if ($123) {
   $124 = HEAP32[$8>>2]|0;
   HEAP32[$16>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($124,$16);
   $125 = ((($8)) + 4|0);
   HEAP32[$125>>2] = 0;
  } else {
   HEAP32[$17>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($8,$17);
   HEAP8[$121>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($8|0),0);
  $126 = __THREW__; __THREW__ = 0;
  $127 = $126&1;
  if ($127) {
   $128 = ___cxa_find_matching_catch_3(0|0)|0;
   $129 = tempRet0;
   ___clang_call_terminate($128);
   // unreachable;
  }
  ;HEAP32[$8>>2]=HEAP32[$32>>2]|0;HEAP32[$8+4>>2]=HEAP32[$32+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$32+8>>2]|0;
  $$0$i$i$i52 = 0;
  while(1) {
   $exitcond$i$i$i53 = ($$0$i$i$i52|0)==(3);
   if ($exitcond$i$i$i53) {
    break;
   }
   $130 = (($32) + ($$0$i$i$i52<<2)|0);
   HEAP32[$130>>2] = 0;
   $131 = (($$0$i$i$i52) + 1)|0;
   $$0$i$i$i52 = $131;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($32);
  $132 = HEAP32[$112>>2]|0;
  $133 = ((($132)) + 28|0);
  $134 = HEAP32[$133>>2]|0;
  FUNCTION_TABLE_vii[$134 & 2047]($33,$112);
  $135 = ((($7)) + 8|0);
  $136 = ((($135)) + 3|0);
  $137 = HEAP8[$136>>0]|0;
  $138 = ($137<<24>>24)<(0);
  if ($138) {
   $139 = HEAP32[$7>>2]|0;
   HEAP32[$14>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($139,$14);
   $140 = ((($7)) + 4|0);
   HEAP32[$140>>2] = 0;
  } else {
   HEAP32[$15>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($7,$15);
   HEAP8[$136>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($7|0),0);
  $141 = __THREW__; __THREW__ = 0;
  $142 = $141&1;
  if ($142) {
   $143 = ___cxa_find_matching_catch_3(0|0)|0;
   $144 = tempRet0;
   ___clang_call_terminate($143);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$33>>2]|0;HEAP32[$7+4>>2]=HEAP32[$33+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$33+8>>2]|0;
  $$0$i$i$i57 = 0;
  while(1) {
   $exitcond$i$i$i58 = ($$0$i$i$i57|0)==(3);
   if ($exitcond$i$i$i58) {
    break;
   }
   $145 = (($33) + ($$0$i$i$i57<<2)|0);
   HEAP32[$145>>2] = 0;
   $146 = (($$0$i$i$i57) + 1)|0;
   $$0$i$i$i57 = $146;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($33);
  $147 = HEAP32[$112>>2]|0;
  $148 = ((($147)) + 12|0);
  $149 = HEAP32[$148>>2]|0;
  $150 = (FUNCTION_TABLE_ii[$149 & 2047]($112)|0);
  HEAP32[$3>>2] = $150;
  $151 = HEAP32[$112>>2]|0;
  $152 = ((($151)) + 16|0);
  $153 = HEAP32[$152>>2]|0;
  $154 = (FUNCTION_TABLE_ii[$153 & 2047]($112)|0);
  HEAP32[$4>>2] = $154;
  $155 = HEAP32[$112>>2]|0;
  $156 = ((($155)) + 20|0);
  $157 = HEAP32[$156>>2]|0;
  FUNCTION_TABLE_vii[$157 & 2047]($34,$112);
  $158 = ((($5)) + 11|0);
  $159 = HEAP8[$158>>0]|0;
  $160 = ($159<<24>>24)<(0);
  if ($160) {
   $161 = HEAP32[$5>>2]|0;
   HEAP8[$12>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($161,$12);
   $162 = ((($5)) + 4|0);
   HEAP32[$162>>2] = 0;
  } else {
   HEAP8[$13>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($5,$13);
   HEAP8[$158>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($5|0),0);
  $163 = __THREW__; __THREW__ = 0;
  $164 = $163&1;
  if ($164) {
   $165 = ___cxa_find_matching_catch_3(0|0)|0;
   $166 = tempRet0;
   ___clang_call_terminate($165);
   // unreachable;
  }
  ;HEAP32[$5>>2]=HEAP32[$34>>2]|0;HEAP32[$5+4>>2]=HEAP32[$34+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$34+8>>2]|0;
  $$0$i$i$i62 = 0;
  while(1) {
   $exitcond$i$i$i63 = ($$0$i$i$i62|0)==(3);
   if ($exitcond$i$i$i63) {
    break;
   }
   $167 = (($34) + ($$0$i$i$i62<<2)|0);
   HEAP32[$167>>2] = 0;
   $168 = (($$0$i$i$i62) + 1)|0;
   $$0$i$i$i62 = $168;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($34);
  $169 = HEAP32[$112>>2]|0;
  $170 = ((($169)) + 24|0);
  $171 = HEAP32[$170>>2]|0;
  FUNCTION_TABLE_vii[$171 & 2047]($35,$112);
  $172 = ((($6)) + 8|0);
  $173 = ((($172)) + 3|0);
  $174 = HEAP8[$173>>0]|0;
  $175 = ($174<<24>>24)<(0);
  if ($175) {
   $176 = HEAP32[$6>>2]|0;
   HEAP32[$10>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($176,$10);
   $177 = ((($6)) + 4|0);
   HEAP32[$177>>2] = 0;
  } else {
   HEAP32[$11>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($6,$11);
   HEAP8[$173>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($6|0),0);
  $178 = __THREW__; __THREW__ = 0;
  $179 = $178&1;
  if ($179) {
   $180 = ___cxa_find_matching_catch_3(0|0)|0;
   $181 = tempRet0;
   ___clang_call_terminate($180);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$35>>2]|0;HEAP32[$6+4>>2]=HEAP32[$35+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$35+8>>2]|0;
  $$0$i$i$i67 = 0;
  while(1) {
   $exitcond$i$i$i68 = ($$0$i$i$i67|0)==(3);
   if ($exitcond$i$i$i68) {
    break;
   }
   $182 = (($35) + ($$0$i$i$i67<<2)|0);
   HEAP32[$182>>2] = 0;
   $183 = (($$0$i$i$i67) + 1)|0;
   $$0$i$i$i67 = $183;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($35);
  $184 = HEAP32[$112>>2]|0;
  $185 = ((($184)) + 36|0);
  $186 = HEAP32[$185>>2]|0;
  $187 = (FUNCTION_TABLE_ii[$186 & 2047]($112)|0);
  $storemerge = $187;
 }
 HEAP32[$9>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__219__double_or_nothingIwEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$34 = 0, $$pr = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)!=(1003|0);
 $6 = HEAP32[$2>>2]|0;
 $7 = HEAP32[$0>>2]|0;
 $8 = (($6) - ($7))|0;
 $9 = ($8>>>0)<(2147483647);
 $10 = $8 << 1;
 $$34 = $9 ? $10 : -1;
 $11 = ($$34|0)==(0);
 $$ = $11 ? 4 : $$34;
 $12 = HEAP32[$1>>2]|0;
 $13 = (($12) - ($7))|0;
 $14 = $13 >> 2;
 $phitmp = $7;
 $15 = $5 ? $phitmp : 0;
 $16 = (_realloc($15,$$)|0);
 $17 = ($16|0)==(0|0);
 if ($17) {
  __ZSt17__throw_bad_allocv();
  // unreachable;
 }
 do {
  if ($5) {
   $18 = $16;
   HEAP32[$0>>2] = $18;
   $28 = $16;
  } else {
   $$pr = HEAP32[$0>>2]|0;
   $19 = $16;
   HEAP32[$0>>2] = $19;
   $20 = ($$pr|0)==(0|0);
   if ($20) {
    $28 = $16;
   } else {
    $21 = HEAP32[$3>>2]|0;
    __THREW__ = 0;
    invoke_vi($21|0,($$pr|0));
    $22 = __THREW__; __THREW__ = 0;
    $23 = $22&1;
    if ($23) {
     $24 = ___cxa_find_matching_catch_3(0|0)|0;
     $25 = tempRet0;
     ___clang_call_terminate($24);
     // unreachable;
    } else {
     $$pre = HEAP32[$0>>2]|0;
     $28 = $$pre;
     break;
    }
   }
  }
 } while(0);
 HEAP32[$3>>2] = (730);
 $26 = $$ >>> 2;
 $27 = (($28) + ($14<<2)|0);
 HEAP32[$1>>2] = $27;
 $29 = HEAP32[$0>>2]|0;
 $30 = (($29) + ($26<<2)|0);
 HEAP32[$2>>2] = $30;
 return;
}
function __ZNSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEce($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = +$5;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i56 = 0, $$0$i$i59 = 0, $$046 = 0, $$047 = 0, $$048 = 0, $$049 = 0, $$1 = 0, $$150 = 0, $$3 = 0, $$352 = 0, $$4 = 0, $$453 = 0, $$byval_copy = 0, $$sink = 0, $$sink1 = 0, $$sink105 = 0, $$sink106 = 0, $$sroa$071$0 = 0;
 var $$sroa$071$1 = 0, $$sroa$083$0 = 0, $$sroa$083$1 = 0, $$sroa$083$2 = 0, $$sroa$096$0 = 0, $$sroa$096$1 = 0, $$sroa$096$2 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0;
 var $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0;
 var $exitcond$i$i = 0, $exitcond$i$i57 = 0, $exitcond$i$i60 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 400|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(400|0);
 $$byval_copy = sp + 76|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $6 = sp + 288|0;
 $7 = sp + 72|0;
 $8 = sp + 188|0;
 $9 = sp + 68|0;
 $10 = sp + 184|0;
 $11 = sp + 181|0;
 $12 = sp + 180|0;
 $13 = sp + 56|0;
 $14 = sp + 44|0;
 $15 = sp + 32|0;
 $16 = sp + 28|0;
 $17 = sp + 80|0;
 $18 = sp + 24|0;
 $19 = sp + 20|0;
 $20 = sp + 16|0;
 HEAP32[$7>>2] = $6;
 HEAPF64[$vararg_buffer>>3] = $5;
 $21 = (_snprintf($6,100,157376,$vararg_buffer)|0);
 $22 = ($21>>>0)>(99);
 do {
  if ($22) {
   __THREW__ = 0;
   $23 = (invoke_i(955)|0);
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $$sroa$083$0 = 0;$$sroa$096$0 = 0;
    label = 6;
   } else {
    __THREW__ = 0;
    HEAPF64[$vararg_buffer1>>3] = $5;
    $26 = (invoke_iiiii(976,($7|0),($23|0),(157376|0),($vararg_buffer1|0))|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     $$sroa$083$0 = 0;$$sroa$096$0 = 0;
     label = 6;
    } else {
     $29 = HEAP32[$7>>2]|0;
     $30 = ($29|0)==(0|0);
     if ($30) {
      __THREW__ = 0;
      invoke_v(732);
      $31 = __THREW__; __THREW__ = 0;
      $$sroa$083$0 = 0;$$sroa$096$0 = 0;
      label = 6;
      break;
     }
     $34 = $29;
     $35 = (_malloc($26)|0);
     $36 = $35;
     $37 = ($35|0)==(0|0);
     if ($37) {
      __THREW__ = 0;
      invoke_v(732);
      $38 = __THREW__; __THREW__ = 0;
      $$sroa$083$0 = $36;$$sroa$096$0 = $34;
      label = 6;
     } else {
      $$0 = $35;$$047 = $26;$$sroa$083$1 = $36;$$sroa$096$1 = $34;
      label = 9;
     }
    }
   }
  } else {
   $$0 = $8;$$047 = $21;$$sroa$083$1 = 0;$$sroa$096$1 = 0;
   label = 9;
  }
 } while(0);
 if ((label|0) == 9) {
  __THREW__ = 0;
  invoke_vii(733,($9|0),($3|0));
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if ($40) {
   $$sroa$083$0 = $$sroa$083$1;$$sroa$096$0 = $$sroa$096$1;
   label = 6;
  } else {
   __THREW__ = 0;
   $41 = (invoke_iii(734,($9|0),(165440|0))|0);
   $42 = __THREW__; __THREW__ = 0;
   $43 = $42&1;
   do {
    if ($43) {
     $106 = ___cxa_find_matching_catch_2()|0;
     $107 = tempRet0;
     $$3 = $106;$$352 = $107;
    } else {
     $44 = HEAP32[$7>>2]|0;
     $45 = (($44) + ($$047)|0);
     $46 = HEAP32[$41>>2]|0;
     $47 = ((($46)) + 32|0);
     $48 = HEAP32[$47>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($48|0,($41|0),($44|0),($45|0),($$0|0))|0);
     $49 = __THREW__; __THREW__ = 0;
     $50 = $49&1;
     if ($50) {
      $108 = ___cxa_find_matching_catch_2()|0;
      $109 = tempRet0;
      $$3 = $108;$$352 = $109;
      break;
     }
     $51 = ($$047|0)==(0);
     if ($51) {
      $61 = 0;
     } else {
      $52 = HEAP32[$7>>2]|0;
      $53 = HEAP8[$52>>0]|0;
      $54 = ($53<<24>>24)==(45);
      $61 = $54;
     }
     ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;
     $$0$i$i = 0;
     while(1) {
      $exitcond$i$i = ($$0$i$i|0)==(3);
      if ($exitcond$i$i) {
       break;
      }
      $55 = (($13) + ($$0$i$i<<2)|0);
      HEAP32[$55>>2] = 0;
      $56 = (($$0$i$i) + 1)|0;
      $$0$i$i = $56;
     }
     ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;
     $$0$i$i56 = 0;
     while(1) {
      $exitcond$i$i57 = ($$0$i$i56|0)==(3);
      if ($exitcond$i$i57) {
       break;
      }
      $57 = (($14) + ($$0$i$i56<<2)|0);
      HEAP32[$57>>2] = 0;
      $58 = (($$0$i$i56) + 1)|0;
      $$0$i$i56 = $58;
     }
     ;HEAP32[$15>>2]=0|0;HEAP32[$15+4>>2]=0|0;HEAP32[$15+8>>2]=0|0;
     $$0$i$i59 = 0;
     while(1) {
      $exitcond$i$i60 = ($$0$i$i59|0)==(3);
      if ($exitcond$i$i60) {
       break;
      }
      $59 = (($15) + ($$0$i$i59<<2)|0);
      HEAP32[$59>>2] = 0;
      $60 = (($$0$i$i59) + 1)|0;
      $$0$i$i59 = $60;
     }
     __THREW__ = 0;
     invoke_viiiiiiiiii(1020,($2|0),($61|0),($9|0),($10|0),($11|0),($12|0),($13|0),($14|0),($15|0),($16|0));
     $62 = __THREW__; __THREW__ = 0;
     $63 = $62&1;
     if ($63) {
      $110 = ___cxa_find_matching_catch_2()|0;
      $111 = tempRet0;
      $$1 = $110;$$150 = $111;
     } else {
      $64 = HEAP32[$16>>2]|0;
      $65 = ($$047|0)>($64|0);
      if ($65) {
       $66 = (($$047) - ($64))|0;
       $67 = $66 << 1;
       $68 = ((($15)) + 11|0);
       $69 = HEAP8[$68>>0]|0;
       $70 = ($69<<24>>24)<(0);
       $71 = ((($15)) + 4|0);
       $72 = HEAP32[$71>>2]|0;
       $73 = $69&255;
       $74 = $70 ? $72 : $73;
       $75 = (($74) + ($67))|0;
       $76 = ((($14)) + 11|0);
       $77 = HEAP8[$76>>0]|0;
       $78 = ($77<<24>>24)<(0);
       $79 = ((($14)) + 4|0);
       $80 = HEAP32[$79>>2]|0;
       $81 = $77&255;
       $82 = $78 ? $80 : $81;
       $$sink1 = 1;$$sink105 = $75;$$sink106 = $82;
      } else {
       $83 = ((($15)) + 11|0);
       $84 = HEAP8[$83>>0]|0;
       $85 = ($84<<24>>24)<(0);
       $86 = ((($15)) + 4|0);
       $87 = HEAP32[$86>>2]|0;
       $88 = $84&255;
       $89 = $85 ? $87 : $88;
       $90 = ((($14)) + 11|0);
       $91 = HEAP8[$90>>0]|0;
       $92 = ($91<<24>>24)<(0);
       $93 = ((($14)) + 4|0);
       $94 = HEAP32[$93>>2]|0;
       $95 = $91&255;
       $96 = $92 ? $94 : $95;
       $$sink1 = 2;$$sink105 = $96;$$sink106 = $89;
      }
      $97 = (($$sink106) + ($64))|0;
      $$sink = (($97) + ($$sink105))|0;
      $98 = (($$sink) + ($$sink1))|0;
      $99 = ($98>>>0)>(100);
      if ($99) {
       $100 = (_malloc($98)|0);
       $101 = $100;
       $102 = ($100|0)==(0|0);
       if ($102) {
        __THREW__ = 0;
        invoke_v(732);
        $103 = __THREW__; __THREW__ = 0;
        $104 = ___cxa_find_matching_catch_2()|0;
        $105 = tempRet0;
        $$048 = $104;$$049 = $105;$$sroa$071$1 = $101;
       } else {
        $$046 = $100;$$sroa$071$0 = $101;
        label = 33;
       }
      } else {
       $$046 = $17;$$sroa$071$0 = 0;
       label = 33;
      }
      if ((label|0) == 33) {
       $112 = ((($3)) + 4|0);
       $113 = HEAP32[$112>>2]|0;
       $114 = (($$0) + ($$047)|0);
       $115 = HEAP8[$11>>0]|0;
       $116 = HEAP8[$12>>0]|0;
       __THREW__ = 0;
       invoke_viiiiiiiiiiiiiii(1021,($$046|0),($18|0),($19|0),($113|0),($$0|0),($114|0),($41|0),($61|0),($10|0),($115|0),($116|0),($13|0),($14|0),($15|0),($64|0));
       $117 = __THREW__; __THREW__ = 0;
       $118 = $117&1;
       if (!($118)) {
        $119 = HEAP32[$1>>2]|0;
        HEAP32[$20>>2] = $119;
        $120 = HEAP32[$18>>2]|0;
        $121 = HEAP32[$19>>2]|0;
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$20>>2]|0;
        $122 = (invoke_iiiiiii(978,($$byval_copy|0),($$046|0),($120|0),($121|0),($3|0),($4|0))|0);
        $123 = __THREW__; __THREW__ = 0;
        $124 = $123&1;
        if (!($124)) {
         $125 = ($$sroa$071$0|0)==(0);
         if (!($125)) {
          $126 = $$sroa$071$0;
          _free($126);
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($15);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13);
         __ZNSt3__26localeD2Ev($9);
         $127 = ($$sroa$083$1|0)==(0);
         if (!($127)) {
          $128 = $$sroa$083$1;
          _free($128);
         }
         $129 = ($$sroa$096$1|0)==(0);
         if (!($129)) {
          $130 = $$sroa$096$1;
          _free($130);
         }
         STACKTOP = sp;return ($122|0);
        }
       }
       $131 = ___cxa_find_matching_catch_2()|0;
       $132 = tempRet0;
       $$048 = $131;$$049 = $132;$$sroa$071$1 = $$sroa$071$0;
      }
      $133 = ($$sroa$071$1|0)==(0);
      if (!($133)) {
       $134 = $$sroa$071$1;
       _free($134);
      }
      $$1 = $$048;$$150 = $$049;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($15);
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14);
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13);
     $$3 = $$1;$$352 = $$150;
    }
   } while(0);
   __ZNSt3__26localeD2Ev($9);
   $$4 = $$3;$$453 = $$352;$$sroa$083$2 = $$sroa$083$1;$$sroa$096$2 = $$sroa$096$1;
  }
 }
 if ((label|0) == 6) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  $$4 = $32;$$453 = $33;$$sroa$083$2 = $$sroa$083$0;$$sroa$096$2 = $$sroa$096$0;
 }
 $135 = ($$sroa$083$2|0)==(0);
 if (!($135)) {
  $136 = $$sroa$083$2;
  _free($136);
 }
 $137 = ($$sroa$096$2|0)==(0);
 if (!($137)) {
  $138 = $$sroa$096$2;
  _free($138);
 }
 ___resumeException($$4|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEcRKNS_12basic_stringIcS3_NS_9allocatorIcEEEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i37 = 0, $$0$i$i40 = 0, $$032 = 0, $$033 = 0, $$1 = 0, $$134 = 0, $$3 = 0, $$336 = 0, $$byval_copy = 0, $$sink = 0, $$sink1 = 0, $$sink63 = 0, $$sink64 = 0, $$sroa$053$0 = 0, $$sroa$053$1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i38 = 0, $exitcond$i$i41 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $$byval_copy = sp + 56|0;
 $6 = sp + 52|0;
 $7 = sp + 164|0;
 $8 = sp + 161|0;
 $9 = sp + 160|0;
 $10 = sp + 40|0;
 $11 = sp + 28|0;
 $12 = sp + 16|0;
 $13 = sp + 12|0;
 $14 = sp + 60|0;
 $15 = sp + 8|0;
 $16 = sp + 4|0;
 $17 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $18 = (invoke_iii(734,($6|0),(165440|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 L1: do {
  if ($20) {
   $95 = ___cxa_find_matching_catch_2()|0;
   $96 = tempRet0;
   $$3 = $95;$$336 = $96;
  } else {
   $21 = ((($5)) + 11|0);
   $22 = HEAP8[$21>>0]|0;
   $23 = ($22<<24>>24)<(0);
   $24 = ((($5)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = $22&255;
   $27 = $23 ? $25 : $26;
   $28 = ($27|0)==(0);
   do {
    if ($28) {
     $45 = 0;
    } else {
     $29 = HEAP32[$5>>2]|0;
     $30 = $23 ? $29 : $5;
     $31 = HEAP8[$30>>0]|0;
     $32 = HEAP32[$18>>2]|0;
     $33 = ((($32)) + 28|0);
     $34 = HEAP32[$33>>2]|0;
     __THREW__ = 0;
     $35 = (invoke_iii($34|0,($18|0),45)|0);
     $36 = __THREW__; __THREW__ = 0;
     $37 = $36&1;
     if ($37) {
      $97 = ___cxa_find_matching_catch_2()|0;
      $98 = tempRet0;
      $$3 = $97;$$336 = $98;
      break L1;
     } else {
      $38 = ($31<<24>>24)==($35<<24>>24);
      $45 = $38;
      break;
     }
    }
   } while(0);
   ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
   $$0$i$i = 0;
   while(1) {
    $exitcond$i$i = ($$0$i$i|0)==(3);
    if ($exitcond$i$i) {
     break;
    }
    $39 = (($10) + ($$0$i$i<<2)|0);
    HEAP32[$39>>2] = 0;
    $40 = (($$0$i$i) + 1)|0;
    $$0$i$i = $40;
   }
   ;HEAP32[$11>>2]=0|0;HEAP32[$11+4>>2]=0|0;HEAP32[$11+8>>2]=0|0;
   $$0$i$i37 = 0;
   while(1) {
    $exitcond$i$i38 = ($$0$i$i37|0)==(3);
    if ($exitcond$i$i38) {
     break;
    }
    $41 = (($11) + ($$0$i$i37<<2)|0);
    HEAP32[$41>>2] = 0;
    $42 = (($$0$i$i37) + 1)|0;
    $$0$i$i37 = $42;
   }
   ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;
   $$0$i$i40 = 0;
   while(1) {
    $exitcond$i$i41 = ($$0$i$i40|0)==(3);
    if ($exitcond$i$i41) {
     break;
    }
    $43 = (($12) + ($$0$i$i40<<2)|0);
    HEAP32[$43>>2] = 0;
    $44 = (($$0$i$i40) + 1)|0;
    $$0$i$i40 = $44;
   }
   __THREW__ = 0;
   invoke_viiiiiiiiii(1020,($2|0),($45|0),($6|0),($7|0),($8|0),($9|0),($10|0),($11|0),($12|0),($13|0));
   $46 = __THREW__; __THREW__ = 0;
   $47 = $46&1;
   if ($47) {
    $99 = ___cxa_find_matching_catch_2()|0;
    $100 = tempRet0;
    $$1 = $99;$$134 = $100;
   } else {
    $48 = HEAP8[$21>>0]|0;
    $49 = ($48<<24>>24)<(0);
    $50 = HEAP32[$24>>2]|0;
    $51 = $48&255;
    $52 = $49 ? $50 : $51;
    $53 = HEAP32[$13>>2]|0;
    $54 = ($52|0)>($53|0);
    if ($54) {
     $55 = (($52) - ($53))|0;
     $56 = $55 << 1;
     $57 = ((($12)) + 11|0);
     $58 = HEAP8[$57>>0]|0;
     $59 = ($58<<24>>24)<(0);
     $60 = ((($12)) + 4|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = $58&255;
     $63 = $59 ? $61 : $62;
     $64 = (($63) + ($56))|0;
     $65 = ((($11)) + 11|0);
     $66 = HEAP8[$65>>0]|0;
     $67 = ($66<<24>>24)<(0);
     $68 = ((($11)) + 4|0);
     $69 = HEAP32[$68>>2]|0;
     $70 = $66&255;
     $71 = $67 ? $69 : $70;
     $$sink1 = 1;$$sink63 = $64;$$sink64 = $71;
    } else {
     $72 = ((($12)) + 11|0);
     $73 = HEAP8[$72>>0]|0;
     $74 = ($73<<24>>24)<(0);
     $75 = ((($12)) + 4|0);
     $76 = HEAP32[$75>>2]|0;
     $77 = $73&255;
     $78 = $74 ? $76 : $77;
     $79 = ((($11)) + 11|0);
     $80 = HEAP8[$79>>0]|0;
     $81 = ($80<<24>>24)<(0);
     $82 = ((($11)) + 4|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = $80&255;
     $85 = $81 ? $83 : $84;
     $$sink1 = 2;$$sink63 = $85;$$sink64 = $78;
    }
    $86 = (($$sink64) + ($53))|0;
    $$sink = (($86) + ($$sink63))|0;
    $87 = (($$sink) + ($$sink1))|0;
    $88 = ($87>>>0)>(100);
    if ($88) {
     $89 = (_malloc($87)|0);
     $90 = $89;
     $91 = ($89|0)==(0|0);
     if ($91) {
      __THREW__ = 0;
      invoke_v(732);
      $92 = __THREW__; __THREW__ = 0;
      $93 = ___cxa_find_matching_catch_2()|0;
      $94 = tempRet0;
      $$0 = $93;$$033 = $94;$$sroa$053$1 = $90;
     } else {
      $$032 = $89;$$sroa$053$0 = $90;
      label = 24;
     }
    } else {
     $$032 = $14;$$sroa$053$0 = 0;
     label = 24;
    }
    if ((label|0) == 24) {
     $101 = ((($3)) + 4|0);
     $102 = HEAP32[$101>>2]|0;
     $103 = HEAP32[$5>>2]|0;
     $104 = $49 ? $103 : $5;
     $105 = (($104) + ($52)|0);
     $106 = HEAP8[$8>>0]|0;
     $107 = HEAP8[$9>>0]|0;
     __THREW__ = 0;
     invoke_viiiiiiiiiiiiiii(1021,($$032|0),($15|0),($16|0),($102|0),($104|0),($105|0),($18|0),($45|0),($7|0),($106|0),($107|0),($10|0),($11|0),($12|0),($53|0));
     $108 = __THREW__; __THREW__ = 0;
     $109 = $108&1;
     if (!($109)) {
      $110 = HEAP32[$1>>2]|0;
      HEAP32[$17>>2] = $110;
      $111 = HEAP32[$15>>2]|0;
      $112 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$17>>2]|0;
      $113 = (invoke_iiiiiii(978,($$byval_copy|0),($$032|0),($111|0),($112|0),($3|0),($4|0))|0);
      $114 = __THREW__; __THREW__ = 0;
      $115 = $114&1;
      if (!($115)) {
       $116 = ($$sroa$053$0|0)==(0);
       if (!($116)) {
        $117 = $$sroa$053$0;
        _free($117);
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($11);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
       __ZNSt3__26localeD2Ev($6);
       STACKTOP = sp;return ($113|0);
      }
     }
     $118 = ___cxa_find_matching_catch_2()|0;
     $119 = tempRet0;
     $$0 = $118;$$033 = $119;$$sroa$053$1 = $$sroa$053$0;
    }
    $120 = ($$sroa$053$1|0)==(0);
    if (!($120)) {
     $121 = $$sroa$053$1;
     _free($121);
    }
    $$1 = $$0;$$134 = $$033;
   }
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($12);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($11);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
   $$3 = $$1;$$336 = $$134;
  }
 } while(0);
 __ZNSt3__26localeD2Ev($6);
 ___resumeException($$3|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__211__money_putIcE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_Ri($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$i$i = 0, $$0$i$i$i48 = 0, $$0$i$i$i53 = 0, $$0$i$i$i58 = 0, $$0$i$i$i63 = 0, $$0$i$i$i68 = 0, $$0$i$i$i73 = 0, $$0$i$i$i78 = 0, $$pre$phi81Z2D = 0, $$pre$phiZ2D = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i49 = 0, $exitcond$i$i$i54 = 0, $exitcond$i$i$i59 = 0, $exitcond$i$i$i64 = 0, $exitcond$i$i$i69 = 0, $exitcond$i$i$i74 = 0, $exitcond$i$i$i79 = 0;
 var $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 128|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(128|0);
 $10 = sp + 127|0;
 $11 = sp + 126|0;
 $12 = sp + 125|0;
 $13 = sp + 124|0;
 $14 = sp + 123|0;
 $15 = sp + 122|0;
 $16 = sp + 121|0;
 $17 = sp + 120|0;
 $18 = sp + 119|0;
 $19 = sp + 118|0;
 $20 = sp + 117|0;
 $21 = sp + 116|0;
 $22 = sp + 115|0;
 $23 = sp + 114|0;
 $24 = sp + 113|0;
 $25 = sp + 112|0;
 $26 = sp + 108|0;
 $27 = sp + 96|0;
 $28 = sp + 92|0;
 $29 = sp + 80|0;
 $30 = sp + 68|0;
 $31 = sp + 56|0;
 $32 = sp + 52|0;
 $33 = sp + 40|0;
 $34 = sp + 36|0;
 $35 = sp + 24|0;
 $36 = sp + 12|0;
 $37 = sp;
 if ($0) {
  $38 = (__ZNKSt3__26locale9use_facetERNS0_2idE($2,167160)|0);
  if ($1) {
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($39)) + 44|0);
   $41 = HEAP32[$40>>2]|0;
   FUNCTION_TABLE_vii[$41 & 2047]($26,$38);
   $42 = HEAP32[$26>>2]|0;
   HEAP8[$3>>0]=$42&255;HEAP8[$3+1>>0]=($42>>8)&255;HEAP8[$3+2>>0]=($42>>16)&255;HEAP8[$3+3>>0]=$42>>24;
   $43 = HEAP32[$38>>2]|0;
   $44 = ((($43)) + 32|0);
   $45 = HEAP32[$44>>2]|0;
   FUNCTION_TABLE_vii[$45 & 2047]($27,$38);
   $46 = ((($8)) + 11|0);
   $47 = HEAP8[$46>>0]|0;
   $48 = ($47<<24>>24)<(0);
   if ($48) {
    $49 = HEAP32[$8>>2]|0;
    HEAP8[$24>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($49,$24);
    $50 = ((($8)) + 4|0);
    HEAP32[$50>>2] = 0;
   } else {
    HEAP8[$25>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($8,$25);
    HEAP8[$46>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1012,($8|0),0);
   $51 = __THREW__; __THREW__ = 0;
   $52 = $51&1;
   if ($52) {
    $53 = ___cxa_find_matching_catch_3(0|0)|0;
    $54 = tempRet0;
    ___clang_call_terminate($53);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$27>>2]|0;HEAP32[$8+4>>2]=HEAP32[$27+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$27+8>>2]|0;
   $$0$i$i$i = 0;
   while(1) {
    $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
    if ($exitcond$i$i$i) {
     break;
    }
    $55 = (($27) + ($$0$i$i$i<<2)|0);
    HEAP32[$55>>2] = 0;
    $56 = (($$0$i$i$i) + 1)|0;
    $$0$i$i$i = $56;
   }
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($27);
   $$pre$phiZ2D = $38;
  } else {
   $57 = HEAP32[$38>>2]|0;
   $58 = ((($57)) + 40|0);
   $59 = HEAP32[$58>>2]|0;
   FUNCTION_TABLE_vii[$59 & 2047]($28,$38);
   $60 = HEAP32[$28>>2]|0;
   HEAP8[$3>>0]=$60&255;HEAP8[$3+1>>0]=($60>>8)&255;HEAP8[$3+2>>0]=($60>>16)&255;HEAP8[$3+3>>0]=$60>>24;
   $61 = HEAP32[$38>>2]|0;
   $62 = ((($61)) + 28|0);
   $63 = HEAP32[$62>>2]|0;
   FUNCTION_TABLE_vii[$63 & 2047]($29,$38);
   $64 = ((($8)) + 11|0);
   $65 = HEAP8[$64>>0]|0;
   $66 = ($65<<24>>24)<(0);
   if ($66) {
    $67 = HEAP32[$8>>2]|0;
    HEAP8[$22>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($67,$22);
    $68 = ((($8)) + 4|0);
    HEAP32[$68>>2] = 0;
   } else {
    HEAP8[$23>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($8,$23);
    HEAP8[$64>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1012,($8|0),0);
   $69 = __THREW__; __THREW__ = 0;
   $70 = $69&1;
   if ($70) {
    $71 = ___cxa_find_matching_catch_3(0|0)|0;
    $72 = tempRet0;
    ___clang_call_terminate($71);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$29>>2]|0;HEAP32[$8+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$29+8>>2]|0;
   $$0$i$i$i48 = 0;
   while(1) {
    $exitcond$i$i$i49 = ($$0$i$i$i48|0)==(3);
    if ($exitcond$i$i$i49) {
     break;
    }
    $73 = (($29) + ($$0$i$i$i48<<2)|0);
    HEAP32[$73>>2] = 0;
    $74 = (($$0$i$i$i48) + 1)|0;
    $$0$i$i$i48 = $74;
   }
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($29);
   $$pre$phiZ2D = $38;
  }
  $75 = HEAP32[$38>>2]|0;
  $76 = ((($75)) + 12|0);
  $77 = HEAP32[$76>>2]|0;
  $78 = (FUNCTION_TABLE_ii[$77 & 2047]($38)|0);
  HEAP8[$4>>0] = $78;
  $79 = HEAP32[$38>>2]|0;
  $80 = ((($79)) + 16|0);
  $81 = HEAP32[$80>>2]|0;
  $82 = (FUNCTION_TABLE_ii[$81 & 2047]($38)|0);
  HEAP8[$5>>0] = $82;
  $83 = HEAP32[$$pre$phiZ2D>>2]|0;
  $84 = ((($83)) + 20|0);
  $85 = HEAP32[$84>>2]|0;
  FUNCTION_TABLE_vii[$85 & 2047]($30,$38);
  $86 = ((($6)) + 11|0);
  $87 = HEAP8[$86>>0]|0;
  $88 = ($87<<24>>24)<(0);
  if ($88) {
   $89 = HEAP32[$6>>2]|0;
   HEAP8[$20>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($89,$20);
   $90 = ((($6)) + 4|0);
   HEAP32[$90>>2] = 0;
  } else {
   HEAP8[$21>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$21);
   HEAP8[$86>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $91 = __THREW__; __THREW__ = 0;
  $92 = $91&1;
  if ($92) {
   $93 = ___cxa_find_matching_catch_3(0|0)|0;
   $94 = tempRet0;
   ___clang_call_terminate($93);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$30>>2]|0;HEAP32[$6+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$30+8>>2]|0;
  $$0$i$i$i53 = 0;
  while(1) {
   $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
   if ($exitcond$i$i$i54) {
    break;
   }
   $95 = (($30) + ($$0$i$i$i53<<2)|0);
   HEAP32[$95>>2] = 0;
   $96 = (($$0$i$i$i53) + 1)|0;
   $$0$i$i$i53 = $96;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30);
  $97 = HEAP32[$$pre$phiZ2D>>2]|0;
  $98 = ((($97)) + 24|0);
  $99 = HEAP32[$98>>2]|0;
  FUNCTION_TABLE_vii[$99 & 2047]($31,$38);
  $100 = ((($7)) + 11|0);
  $101 = HEAP8[$100>>0]|0;
  $102 = ($101<<24>>24)<(0);
  if ($102) {
   $103 = HEAP32[$7>>2]|0;
   HEAP8[$18>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($103,$18);
   $104 = ((($7)) + 4|0);
   HEAP32[$104>>2] = 0;
  } else {
   HEAP8[$19>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($7,$19);
   HEAP8[$100>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($7|0),0);
  $105 = __THREW__; __THREW__ = 0;
  $106 = $105&1;
  if ($106) {
   $107 = ___cxa_find_matching_catch_3(0|0)|0;
   $108 = tempRet0;
   ___clang_call_terminate($107);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$31>>2]|0;HEAP32[$7+4>>2]=HEAP32[$31+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$31+8>>2]|0;
  $$0$i$i$i58 = 0;
  while(1) {
   $exitcond$i$i$i59 = ($$0$i$i$i58|0)==(3);
   if ($exitcond$i$i$i59) {
    break;
   }
   $109 = (($31) + ($$0$i$i$i58<<2)|0);
   HEAP32[$109>>2] = 0;
   $110 = (($$0$i$i$i58) + 1)|0;
   $$0$i$i$i58 = $110;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($31);
  $111 = HEAP32[$38>>2]|0;
  $112 = ((($111)) + 36|0);
  $113 = HEAP32[$112>>2]|0;
  $114 = (FUNCTION_TABLE_ii[$113 & 2047]($38)|0);
  $storemerge = $114;
 } else {
  $115 = (__ZNKSt3__26locale9use_facetERNS0_2idE($2,167152)|0);
  if ($1) {
   $116 = HEAP32[$115>>2]|0;
   $117 = ((($116)) + 44|0);
   $118 = HEAP32[$117>>2]|0;
   FUNCTION_TABLE_vii[$118 & 2047]($32,$115);
   $119 = HEAP32[$32>>2]|0;
   HEAP8[$3>>0]=$119&255;HEAP8[$3+1>>0]=($119>>8)&255;HEAP8[$3+2>>0]=($119>>16)&255;HEAP8[$3+3>>0]=$119>>24;
   $120 = HEAP32[$115>>2]|0;
   $121 = ((($120)) + 32|0);
   $122 = HEAP32[$121>>2]|0;
   FUNCTION_TABLE_vii[$122 & 2047]($33,$115);
   $123 = ((($8)) + 11|0);
   $124 = HEAP8[$123>>0]|0;
   $125 = ($124<<24>>24)<(0);
   if ($125) {
    $126 = HEAP32[$8>>2]|0;
    HEAP8[$16>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($126,$16);
    $127 = ((($8)) + 4|0);
    HEAP32[$127>>2] = 0;
   } else {
    HEAP8[$17>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($8,$17);
    HEAP8[$123>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1012,($8|0),0);
   $128 = __THREW__; __THREW__ = 0;
   $129 = $128&1;
   if ($129) {
    $130 = ___cxa_find_matching_catch_3(0|0)|0;
    $131 = tempRet0;
    ___clang_call_terminate($130);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$33>>2]|0;HEAP32[$8+4>>2]=HEAP32[$33+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$33+8>>2]|0;
   $$0$i$i$i63 = 0;
   while(1) {
    $exitcond$i$i$i64 = ($$0$i$i$i63|0)==(3);
    if ($exitcond$i$i$i64) {
     break;
    }
    $132 = (($33) + ($$0$i$i$i63<<2)|0);
    HEAP32[$132>>2] = 0;
    $133 = (($$0$i$i$i63) + 1)|0;
    $$0$i$i$i63 = $133;
   }
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($33);
   $$pre$phi81Z2D = $115;
  } else {
   $134 = HEAP32[$115>>2]|0;
   $135 = ((($134)) + 40|0);
   $136 = HEAP32[$135>>2]|0;
   FUNCTION_TABLE_vii[$136 & 2047]($34,$115);
   $137 = HEAP32[$34>>2]|0;
   HEAP8[$3>>0]=$137&255;HEAP8[$3+1>>0]=($137>>8)&255;HEAP8[$3+2>>0]=($137>>16)&255;HEAP8[$3+3>>0]=$137>>24;
   $138 = HEAP32[$115>>2]|0;
   $139 = ((($138)) + 28|0);
   $140 = HEAP32[$139>>2]|0;
   FUNCTION_TABLE_vii[$140 & 2047]($35,$115);
   $141 = ((($8)) + 11|0);
   $142 = HEAP8[$141>>0]|0;
   $143 = ($142<<24>>24)<(0);
   if ($143) {
    $144 = HEAP32[$8>>2]|0;
    HEAP8[$14>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($144,$14);
    $145 = ((($8)) + 4|0);
    HEAP32[$145>>2] = 0;
   } else {
    HEAP8[$15>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($8,$15);
    HEAP8[$141>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1012,($8|0),0);
   $146 = __THREW__; __THREW__ = 0;
   $147 = $146&1;
   if ($147) {
    $148 = ___cxa_find_matching_catch_3(0|0)|0;
    $149 = tempRet0;
    ___clang_call_terminate($148);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$35>>2]|0;HEAP32[$8+4>>2]=HEAP32[$35+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$35+8>>2]|0;
   $$0$i$i$i68 = 0;
   while(1) {
    $exitcond$i$i$i69 = ($$0$i$i$i68|0)==(3);
    if ($exitcond$i$i$i69) {
     break;
    }
    $150 = (($35) + ($$0$i$i$i68<<2)|0);
    HEAP32[$150>>2] = 0;
    $151 = (($$0$i$i$i68) + 1)|0;
    $$0$i$i$i68 = $151;
   }
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($35);
   $$pre$phi81Z2D = $115;
  }
  $152 = HEAP32[$115>>2]|0;
  $153 = ((($152)) + 12|0);
  $154 = HEAP32[$153>>2]|0;
  $155 = (FUNCTION_TABLE_ii[$154 & 2047]($115)|0);
  HEAP8[$4>>0] = $155;
  $156 = HEAP32[$115>>2]|0;
  $157 = ((($156)) + 16|0);
  $158 = HEAP32[$157>>2]|0;
  $159 = (FUNCTION_TABLE_ii[$158 & 2047]($115)|0);
  HEAP8[$5>>0] = $159;
  $160 = HEAP32[$$pre$phi81Z2D>>2]|0;
  $161 = ((($160)) + 20|0);
  $162 = HEAP32[$161>>2]|0;
  FUNCTION_TABLE_vii[$162 & 2047]($36,$115);
  $163 = ((($6)) + 11|0);
  $164 = HEAP8[$163>>0]|0;
  $165 = ($164<<24>>24)<(0);
  if ($165) {
   $166 = HEAP32[$6>>2]|0;
   HEAP8[$12>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($166,$12);
   $167 = ((($6)) + 4|0);
   HEAP32[$167>>2] = 0;
  } else {
   HEAP8[$13>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$13);
   HEAP8[$163>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $168 = __THREW__; __THREW__ = 0;
  $169 = $168&1;
  if ($169) {
   $170 = ___cxa_find_matching_catch_3(0|0)|0;
   $171 = tempRet0;
   ___clang_call_terminate($170);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$36>>2]|0;HEAP32[$6+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$36+8>>2]|0;
  $$0$i$i$i73 = 0;
  while(1) {
   $exitcond$i$i$i74 = ($$0$i$i$i73|0)==(3);
   if ($exitcond$i$i$i74) {
    break;
   }
   $172 = (($36) + ($$0$i$i$i73<<2)|0);
   HEAP32[$172>>2] = 0;
   $173 = (($$0$i$i$i73) + 1)|0;
   $$0$i$i$i73 = $173;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($36);
  $174 = HEAP32[$$pre$phi81Z2D>>2]|0;
  $175 = ((($174)) + 24|0);
  $176 = HEAP32[$175>>2]|0;
  FUNCTION_TABLE_vii[$176 & 2047]($37,$115);
  $177 = ((($7)) + 11|0);
  $178 = HEAP8[$177>>0]|0;
  $179 = ($178<<24>>24)<(0);
  if ($179) {
   $180 = HEAP32[$7>>2]|0;
   HEAP8[$10>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($180,$10);
   $181 = ((($7)) + 4|0);
   HEAP32[$181>>2] = 0;
  } else {
   HEAP8[$11>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($7,$11);
   HEAP8[$177>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($7|0),0);
  $182 = __THREW__; __THREW__ = 0;
  $183 = $182&1;
  if ($183) {
   $184 = ___cxa_find_matching_catch_3(0|0)|0;
   $185 = tempRet0;
   ___clang_call_terminate($184);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$37>>2]|0;HEAP32[$7+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$37+8>>2]|0;
  $$0$i$i$i78 = 0;
  while(1) {
   $exitcond$i$i$i79 = ($$0$i$i$i78|0)==(3);
   if ($exitcond$i$i$i79) {
    break;
   }
   $186 = (($37) + ($$0$i$i$i78<<2)|0);
   HEAP32[$186>>2] = 0;
   $187 = (($$0$i$i$i78) + 1)|0;
   $$0$i$i$i78 = $187;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($37);
  $188 = HEAP32[$115>>2]|0;
  $189 = ((($188)) + 36|0);
  $190 = HEAP32[$189>>2]|0;
  $191 = (FUNCTION_TABLE_ii[$190 & 2047]($115)|0);
  $storemerge = $191;
 }
 HEAP32[$9>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__211__money_putIcE8__formatEPcRS2_S3_jPKcS5_RKNS_5ctypeIcEEbRKNS_10money_base7patternEccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESL_SL_i($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11,$12,$13,$14) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 $11 = $11|0;
 $12 = $12|0;
 $13 = $13|0;
 $14 = $14|0;
 var $$ = 0, $$$097 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i106 = 0, $$0$i$i113 = 0, $$0101 = 0, $$0104 = 0, $$07$i$i = 0, $$093 = 0, $$095 = 0, $$097 = 0, $$099 = 0, $$1 = 0, $$1100 = 0, $$1102 = 0, $$194 = 0, $$196 = 0, $$2 = 0, $$2103 = 0;
 var $$3 = 0, $$pn = 0, $$sroa$08$0$i = 0, $$sroa$08$0$i112 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $exitcond = 0, $or$cond = 0, $trunc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 $15 = ((($13)) + 11|0);
 $16 = ((($13)) + 4|0);
 $17 = ((($12)) + 11|0);
 $18 = ((($12)) + 4|0);
 $19 = $3 & 512;
 $20 = ($19|0)==(0);
 $21 = ((($6)) + 8|0);
 $22 = ($14|0)>(0);
 $23 = ((($11)) + 11|0);
 $24 = ((($11)) + 4|0);
 $$0104 = 0;$$097 = $4;
 while(1) {
  $exitcond = ($$0104|0)==(4);
  if ($exitcond) {
   break;
  }
  $31 = (($8) + ($$0104)|0);
  $32 = HEAP8[$31>>0]|0;
  $33 = $32 << 24 >> 24;
  L4: do {
   switch ($33|0) {
   case 0:  {
    $34 = HEAP32[$2>>2]|0;
    HEAP32[$1>>2] = $34;
    $$2 = $$097;
    break;
   }
   case 1:  {
    $35 = HEAP32[$2>>2]|0;
    HEAP32[$1>>2] = $35;
    $36 = HEAP32[$6>>2]|0;
    $37 = ((($36)) + 28|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = (FUNCTION_TABLE_iii[$38 & 2047]($6,32)|0);
    $40 = HEAP32[$2>>2]|0;
    $41 = ((($40)) + 1|0);
    HEAP32[$2>>2] = $41;
    HEAP8[$40>>0] = $39;
    $$2 = $$097;
    break;
   }
   case 3:  {
    $42 = HEAP8[$15>>0]|0;
    $43 = ($42<<24>>24)<(0);
    $44 = HEAP32[$16>>2]|0;
    $45 = $42&255;
    $46 = $43 ? $44 : $45;
    $47 = ($46|0)==(0);
    if ($47) {
     $$2 = $$097;
    } else {
     $48 = HEAP32[$13>>2]|0;
     $49 = $43 ? $48 : $13;
     $50 = HEAP8[$49>>0]|0;
     $51 = HEAP32[$2>>2]|0;
     $52 = ((($51)) + 1|0);
     HEAP32[$2>>2] = $52;
     HEAP8[$51>>0] = $50;
     $$2 = $$097;
    }
    break;
   }
   case 2:  {
    $53 = HEAP8[$17>>0]|0;
    $54 = ($53<<24>>24)<(0);
    $55 = HEAP32[$18>>2]|0;
    $56 = $53&255;
    $57 = $54 ? $55 : $56;
    $58 = ($57|0)==(0);
    $or$cond = $20 | $58;
    if ($or$cond) {
     $$2 = $$097;
    } else {
     $59 = HEAP32[$12>>2]|0;
     $60 = $54 ? $59 : $12;
     $61 = (($60) + ($57)|0);
     $62 = HEAP32[$2>>2]|0;
     $$0$i$i113 = $62;$$sroa$08$0$i112 = $60;
     while(1) {
      $63 = ($$sroa$08$0$i112|0)==($61|0);
      if ($63) {
       break;
      }
      $64 = HEAP8[$$sroa$08$0$i112>>0]|0;
      HEAP8[$$0$i$i113>>0] = $64;
      $65 = ((($$sroa$08$0$i112)) + 1|0);
      $66 = ((($$0$i$i113)) + 1|0);
      $$0$i$i113 = $66;$$sroa$08$0$i112 = $65;
     }
     HEAP32[$2>>2] = $$0$i$i113;
     $$2 = $$097;
    }
    break;
   }
   case 4:  {
    $67 = HEAP32[$2>>2]|0;
    $68 = ((($$097)) + 1|0);
    $$$097 = $7 ? $68 : $$097;
    $$0101 = $$$097;
    while(1) {
     $69 = ($$0101>>>0)<($5>>>0);
     if (!($69)) {
      break;
     }
     $70 = HEAP8[$$0101>>0]|0;
     $71 = ($70<<24>>24)>(-1);
     if (!($71)) {
      break;
     }
     $72 = $70 << 24 >> 24;
     $73 = HEAP32[$21>>2]|0;
     $74 = (($73) + ($72<<1)|0);
     $75 = HEAP16[$74>>1]|0;
     $76 = $75 & 2048;
     $77 = ($76<<16>>16)==(0);
     if ($77) {
      break;
     }
     $78 = ((($$0101)) + 1|0);
     $$0101 = $78;
    }
    if ($22) {
     $$099 = $14;$$1102 = $$0101;
     while(1) {
      $79 = ($$1102>>>0)>($$$097>>>0);
      $80 = ($$099|0)>(0);
      $81 = $79 & $80;
      if (!($81)) {
       break;
      }
      $82 = ((($$1102)) + -1|0);
      $83 = HEAP8[$82>>0]|0;
      $84 = HEAP32[$2>>2]|0;
      $85 = ((($84)) + 1|0);
      HEAP32[$2>>2] = $85;
      HEAP8[$84>>0] = $83;
      $86 = (($$099) + -1)|0;
      $$099 = $86;$$1102 = $82;
     }
     if ($80) {
      $87 = HEAP32[$6>>2]|0;
      $88 = ((($87)) + 28|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = (FUNCTION_TABLE_iii[$89 & 2047]($6,48)|0);
      $94 = $90;
     } else {
      $94 = 0;
     }
     $$1100 = $$099;
     while(1) {
      $91 = ($$1100|0)>(0);
      $92 = HEAP32[$2>>2]|0;
      $93 = ((($92)) + 1|0);
      HEAP32[$2>>2] = $93;
      if (!($91)) {
       break;
      }
      HEAP8[$92>>0] = $94;
      $95 = (($$1100) + -1)|0;
      $$1100 = $95;
     }
     HEAP8[$92>>0] = $9;
     $$2103 = $$1102;
    } else {
     $$2103 = $$0101;
    }
    $96 = ($$2103|0)==($$$097|0);
    L35: do {
     if ($96) {
      $97 = HEAP32[$6>>2]|0;
      $98 = ((($97)) + 28|0);
      $99 = HEAP32[$98>>2]|0;
      $100 = (FUNCTION_TABLE_iii[$99 & 2047]($6,48)|0);
      $101 = HEAP32[$2>>2]|0;
      $102 = ((($101)) + 1|0);
      HEAP32[$2>>2] = $102;
      HEAP8[$101>>0] = $100;
     } else {
      $103 = HEAP8[$23>>0]|0;
      $104 = ($103<<24>>24)<(0);
      $105 = HEAP32[$24>>2]|0;
      $106 = $103&255;
      $107 = $104 ? $105 : $106;
      $108 = ($107|0)==(0);
      if ($108) {
       $$0 = -1;$$093 = 0;$$095 = 0;$$3 = $$2103;
      } else {
       $109 = HEAP32[$11>>2]|0;
       $110 = $104 ? $109 : $11;
       $111 = HEAP8[$110>>0]|0;
       $112 = $111 << 24 >> 24;
       $$0 = $112;$$093 = 0;$$095 = 0;$$3 = $$2103;
      }
      while(1) {
       $113 = ($$3|0)==($$$097|0);
       if ($113) {
        break L35;
       }
       $114 = ($$095|0)==($$0|0);
       if ($114) {
        $115 = HEAP32[$2>>2]|0;
        $116 = ((($115)) + 1|0);
        HEAP32[$2>>2] = $116;
        HEAP8[$115>>0] = $10;
        $117 = (($$093) + 1)|0;
        $118 = HEAP8[$23>>0]|0;
        $119 = ($118<<24>>24)<(0);
        $120 = HEAP32[$24>>2]|0;
        $121 = $118&255;
        $122 = $119 ? $120 : $121;
        $123 = ($117>>>0)<($122>>>0);
        if ($123) {
         $124 = HEAP32[$11>>2]|0;
         $125 = $119 ? $124 : $11;
         $126 = (($125) + ($117)|0);
         $127 = HEAP8[$126>>0]|0;
         $128 = ($127<<24>>24)==(127);
         $129 = $127 << 24 >> 24;
         $$ = $128 ? -1 : $129;
         $$1 = $$;$$194 = $117;$$196 = 0;
        } else {
         $$1 = $$095;$$194 = $117;$$196 = 0;
        }
       } else {
        $$1 = $$0;$$194 = $$093;$$196 = $$095;
       }
       $130 = ((($$3)) + -1|0);
       $131 = HEAP8[$130>>0]|0;
       $132 = HEAP32[$2>>2]|0;
       $133 = ((($132)) + 1|0);
       HEAP32[$2>>2] = $133;
       HEAP8[$132>>0] = $131;
       $134 = (($$196) + 1)|0;
       $$0 = $$1;$$093 = $$194;$$095 = $134;$$3 = $130;
      }
     }
    } while(0);
    $135 = HEAP32[$2>>2]|0;
    $136 = ($67|0)==($135|0);
    if ($136) {
     $$2 = $$$097;
    } else {
     $$0$i$i106 = $135;$$07$i$i = $67;
     while(1) {
      $137 = ((($$0$i$i106)) + -1|0);
      $138 = ($$07$i$i>>>0)<($137>>>0);
      if (!($138)) {
       $$2 = $$$097;
       break L4;
      }
      $139 = HEAP8[$$07$i$i>>0]|0;
      $140 = HEAP8[$137>>0]|0;
      HEAP8[$$07$i$i>>0] = $140;
      HEAP8[$137>>0] = $139;
      $141 = ((($$07$i$i)) + 1|0);
      $$0$i$i106 = $137;$$07$i$i = $141;
     }
    }
    break;
   }
   default: {
    $$2 = $$097;
   }
   }
  } while(0);
  $142 = (($$0104) + 1)|0;
  $$0104 = $142;$$097 = $$2;
 }
 $25 = HEAP8[$15>>0]|0;
 $26 = ($25<<24>>24)<(0);
 $27 = HEAP32[$16>>2]|0;
 $28 = $25&255;
 $29 = $26 ? $27 : $28;
 $30 = ($29>>>0)>(1);
 if ($30) {
  $143 = HEAP32[$13>>2]|0;
  $144 = $26 ? $143 : $13;
  $145 = (($144) + ($29)|0);
  $146 = HEAP32[$2>>2]|0;
  $$0$i$i = $146;$$pn = $144;
  while(1) {
   $$sroa$08$0$i = ((($$pn)) + 1|0);
   $147 = ($$sroa$08$0$i|0)==($145|0);
   if ($147) {
    break;
   }
   $148 = HEAP8[$$sroa$08$0$i>>0]|0;
   HEAP8[$$0$i$i>>0] = $148;
   $149 = ((($$0$i$i)) + 1|0);
   $$0$i$i = $149;$$pn = $$sroa$08$0$i;
  }
  HEAP32[$2>>2] = $$0$i$i;
 }
 $150 = $3 & 176;
 $trunc = $150&255;
 switch ($trunc<<24>>24) {
 case 32:  {
  $151 = HEAP32[$2>>2]|0;
  HEAP32[$1>>2] = $151;
  break;
 }
 case 16:  {
  break;
 }
 default: {
  HEAP32[$1>>2] = $0;
 }
 }
 return;
}
function __ZNSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwe($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = +$5;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i54 = 0, $$0$i$i56 = 0, $$046 = 0, $$047 = 0, $$048 = 0, $$049 = 0, $$1 = 0, $$150 = 0, $$3 = 0, $$352 = 0, $$4 = 0, $$453 = 0, $$byval_copy = 0, $$sink = 0, $$sink1 = 0, $$sink102 = 0, $$sink103 = 0, $$sroa$068$0 = 0;
 var $$sroa$068$1 = 0, $$sroa$080$0 = 0, $$sroa$080$1 = 0, $$sroa$080$2 = 0, $$sroa$093$0 = 0, $$sroa$093$1 = 0, $$sroa$093$2 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i55 = 0, $exitcond$i$i57 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1008|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(1008|0);
 $$byval_copy = sp + 892|0;
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 $6 = sp + 900|0;
 $7 = sp + 888|0;
 $8 = sp + 488|0;
 $9 = sp + 480|0;
 $10 = sp + 896|0;
 $11 = sp + 476|0;
 $12 = sp + 472|0;
 $13 = sp + 460|0;
 $14 = sp + 448|0;
 $15 = sp + 436|0;
 $16 = sp + 432|0;
 $17 = sp + 32|0;
 $18 = sp + 24|0;
 $19 = sp + 20|0;
 $20 = sp + 16|0;
 HEAP32[$7>>2] = $6;
 HEAPF64[$vararg_buffer>>3] = $5;
 $21 = (_snprintf($6,100,157376,$vararg_buffer)|0);
 $22 = ($21>>>0)>(99);
 do {
  if ($22) {
   __THREW__ = 0;
   $23 = (invoke_i(955)|0);
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $$sroa$080$0 = 0;$$sroa$093$0 = 0;
    label = 6;
   } else {
    __THREW__ = 0;
    HEAPF64[$vararg_buffer1>>3] = $5;
    $26 = (invoke_iiiii(976,($7|0),($23|0),(157376|0),($vararg_buffer1|0))|0);
    $27 = __THREW__; __THREW__ = 0;
    $28 = $27&1;
    if ($28) {
     $$sroa$080$0 = 0;$$sroa$093$0 = 0;
     label = 6;
    } else {
     $29 = HEAP32[$7>>2]|0;
     $30 = ($29|0)==(0|0);
     if ($30) {
      __THREW__ = 0;
      invoke_v(732);
      $31 = __THREW__; __THREW__ = 0;
      $$sroa$080$0 = 0;$$sroa$093$0 = 0;
      label = 6;
      break;
     }
     $34 = $29;
     $35 = $26 << 2;
     $36 = (_malloc($35)|0);
     $37 = $36;
     $38 = ($36|0)==(0|0);
     if ($38) {
      __THREW__ = 0;
      invoke_v(732);
      $39 = __THREW__; __THREW__ = 0;
      $$sroa$080$0 = $37;$$sroa$093$0 = $34;
      label = 6;
     } else {
      $$0 = $36;$$047 = $26;$$sroa$080$1 = $37;$$sroa$093$1 = $34;
      label = 9;
     }
    }
   }
  } else {
   $$0 = $8;$$047 = $21;$$sroa$080$1 = 0;$$sroa$093$1 = 0;
   label = 9;
  }
 } while(0);
 if ((label|0) == 9) {
  __THREW__ = 0;
  invoke_vii(733,($9|0),($3|0));
  $40 = __THREW__; __THREW__ = 0;
  $41 = $40&1;
  if ($41) {
   $$sroa$080$0 = $$sroa$080$1;$$sroa$093$0 = $$sroa$093$1;
   label = 6;
  } else {
   __THREW__ = 0;
   $42 = (invoke_iii(734,($9|0),(165472|0))|0);
   $43 = __THREW__; __THREW__ = 0;
   $44 = $43&1;
   do {
    if ($44) {
     $112 = ___cxa_find_matching_catch_2()|0;
     $113 = tempRet0;
     $$3 = $112;$$352 = $113;
    } else {
     $45 = HEAP32[$7>>2]|0;
     $46 = (($45) + ($$047)|0);
     $47 = HEAP32[$42>>2]|0;
     $48 = ((($47)) + 48|0);
     $49 = HEAP32[$48>>2]|0;
     __THREW__ = 0;
     (invoke_iiiii($49|0,($42|0),($45|0),($46|0),($$0|0))|0);
     $50 = __THREW__; __THREW__ = 0;
     $51 = $50&1;
     if ($51) {
      $114 = ___cxa_find_matching_catch_2()|0;
      $115 = tempRet0;
      $$3 = $114;$$352 = $115;
      break;
     }
     $52 = ($$047|0)==(0);
     if ($52) {
      $62 = 0;
     } else {
      $53 = HEAP32[$7>>2]|0;
      $54 = HEAP8[$53>>0]|0;
      $55 = ($54<<24>>24)==(45);
      $62 = $55;
     }
     ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;
     $$0$i$i = 0;
     while(1) {
      $exitcond$i$i = ($$0$i$i|0)==(3);
      if ($exitcond$i$i) {
       break;
      }
      $56 = (($13) + ($$0$i$i<<2)|0);
      HEAP32[$56>>2] = 0;
      $57 = (($$0$i$i) + 1)|0;
      $$0$i$i = $57;
     }
     ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;
     $$0$i$i54 = 0;
     while(1) {
      $exitcond$i$i55 = ($$0$i$i54|0)==(3);
      if ($exitcond$i$i55) {
       break;
      }
      $58 = (($14) + ($$0$i$i54<<2)|0);
      HEAP32[$58>>2] = 0;
      $59 = (($$0$i$i54) + 1)|0;
      $$0$i$i54 = $59;
     }
     ;HEAP32[$15>>2]=0|0;HEAP32[$15+4>>2]=0|0;HEAP32[$15+8>>2]=0|0;
     $$0$i$i56 = 0;
     while(1) {
      $exitcond$i$i57 = ($$0$i$i56|0)==(3);
      if ($exitcond$i$i57) {
       break;
      }
      $60 = (($15) + ($$0$i$i56<<2)|0);
      HEAP32[$60>>2] = 0;
      $61 = (($$0$i$i56) + 1)|0;
      $$0$i$i56 = $61;
     }
     __THREW__ = 0;
     invoke_viiiiiiiiii(1022,($2|0),($62|0),($9|0),($10|0),($11|0),($12|0),($13|0),($14|0),($15|0),($16|0));
     $63 = __THREW__; __THREW__ = 0;
     $64 = $63&1;
     if ($64) {
      $116 = ___cxa_find_matching_catch_2()|0;
      $117 = tempRet0;
      $$1 = $116;$$150 = $117;
     } else {
      $65 = HEAP32[$16>>2]|0;
      $66 = ($$047|0)>($65|0);
      if ($66) {
       $67 = (($$047) - ($65))|0;
       $68 = $67 << 1;
       $69 = ((($15)) + 8|0);
       $70 = ((($69)) + 3|0);
       $71 = HEAP8[$70>>0]|0;
       $72 = ($71<<24>>24)<(0);
       $73 = ((($15)) + 4|0);
       $74 = HEAP32[$73>>2]|0;
       $75 = $71&255;
       $76 = $72 ? $74 : $75;
       $77 = (($76) + ($68))|0;
       $78 = ((($14)) + 8|0);
       $79 = ((($78)) + 3|0);
       $80 = HEAP8[$79>>0]|0;
       $81 = ($80<<24>>24)<(0);
       $82 = ((($14)) + 4|0);
       $83 = HEAP32[$82>>2]|0;
       $84 = $80&255;
       $85 = $81 ? $83 : $84;
       $$sink1 = 1;$$sink102 = $77;$$sink103 = $85;
      } else {
       $86 = ((($15)) + 8|0);
       $87 = ((($86)) + 3|0);
       $88 = HEAP8[$87>>0]|0;
       $89 = ($88<<24>>24)<(0);
       $90 = ((($15)) + 4|0);
       $91 = HEAP32[$90>>2]|0;
       $92 = $88&255;
       $93 = $89 ? $91 : $92;
       $94 = ((($14)) + 8|0);
       $95 = ((($94)) + 3|0);
       $96 = HEAP8[$95>>0]|0;
       $97 = ($96<<24>>24)<(0);
       $98 = ((($14)) + 4|0);
       $99 = HEAP32[$98>>2]|0;
       $100 = $96&255;
       $101 = $97 ? $99 : $100;
       $$sink1 = 2;$$sink102 = $101;$$sink103 = $93;
      }
      $102 = (($$sink103) + ($65))|0;
      $$sink = (($102) + ($$sink102))|0;
      $103 = (($$sink) + ($$sink1))|0;
      $104 = ($103>>>0)>(100);
      if ($104) {
       $105 = $103 << 2;
       $106 = (_malloc($105)|0);
       $107 = $106;
       $108 = ($106|0)==(0|0);
       if ($108) {
        __THREW__ = 0;
        invoke_v(732);
        $109 = __THREW__; __THREW__ = 0;
        $110 = ___cxa_find_matching_catch_2()|0;
        $111 = tempRet0;
        $$048 = $110;$$049 = $111;$$sroa$068$1 = $107;
       } else {
        $$046 = $106;$$sroa$068$0 = $107;
        label = 33;
       }
      } else {
       $$046 = $17;$$sroa$068$0 = 0;
       label = 33;
      }
      if ((label|0) == 33) {
       $118 = ((($3)) + 4|0);
       $119 = HEAP32[$118>>2]|0;
       $120 = (($$0) + ($$047<<2)|0);
       $121 = HEAP32[$11>>2]|0;
       $122 = HEAP32[$12>>2]|0;
       __THREW__ = 0;
       invoke_viiiiiiiiiiiiiii(1023,($$046|0),($18|0),($19|0),($119|0),($$0|0),($120|0),($42|0),($62|0),($10|0),($121|0),($122|0),($13|0),($14|0),($15|0),($65|0));
       $123 = __THREW__; __THREW__ = 0;
       $124 = $123&1;
       if (!($124)) {
        $125 = HEAP32[$1>>2]|0;
        HEAP32[$20>>2] = $125;
        $126 = HEAP32[$18>>2]|0;
        $127 = HEAP32[$19>>2]|0;
        __THREW__ = 0;
        ;HEAP32[$$byval_copy>>2]=HEAP32[$20>>2]|0;
        $128 = (invoke_iiiiiii(735,($$byval_copy|0),($$046|0),($126|0),($127|0),($3|0),($4|0))|0);
        $129 = __THREW__; __THREW__ = 0;
        $130 = $129&1;
        if (!($130)) {
         $131 = ($$sroa$068$0|0)==(0);
         if (!($131)) {
          $132 = $$sroa$068$0;
          _free($132);
         }
         __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($15);
         __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($14);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13);
         __ZNSt3__26localeD2Ev($9);
         $133 = ($$sroa$080$1|0)==(0);
         if (!($133)) {
          $134 = $$sroa$080$1;
          _free($134);
         }
         $135 = ($$sroa$093$1|0)==(0);
         if (!($135)) {
          $136 = $$sroa$093$1;
          _free($136);
         }
         STACKTOP = sp;return ($128|0);
        }
       }
       $137 = ___cxa_find_matching_catch_2()|0;
       $138 = tempRet0;
       $$048 = $137;$$049 = $138;$$sroa$068$1 = $$sroa$068$0;
      }
      $139 = ($$sroa$068$1|0)==(0);
      if (!($139)) {
       $140 = $$sroa$068$1;
       _free($140);
      }
      $$1 = $$048;$$150 = $$049;
     }
     __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($15);
     __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($14);
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($13);
     $$3 = $$1;$$352 = $$150;
    }
   } while(0);
   __ZNSt3__26localeD2Ev($9);
   $$4 = $$3;$$453 = $$352;$$sroa$080$2 = $$sroa$080$1;$$sroa$093$2 = $$sroa$093$1;
  }
 }
 if ((label|0) == 6) {
  $32 = ___cxa_find_matching_catch_2()|0;
  $33 = tempRet0;
  $$4 = $32;$$453 = $33;$$sroa$080$2 = $$sroa$080$0;$$sroa$093$2 = $$sroa$093$0;
 }
 $141 = ($$sroa$080$2|0)==(0);
 if (!($141)) {
  $142 = $$sroa$080$2;
  _free($142);
 }
 $143 = ($$sroa$093$2|0)==(0);
 if (!($143)) {
  $144 = $$sroa$093$2;
  _free($144);
 }
 ___resumeException($$4|0);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwRKNS_12basic_stringIwS3_NS_9allocatorIwEEEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i37 = 0, $$0$i$i39 = 0, $$032 = 0, $$033 = 0, $$1 = 0, $$134 = 0, $$3 = 0, $$336 = 0, $$byval_copy = 0, $$sink = 0, $$sink1 = 0, $$sink62 = 0, $$sink63 = 0, $$sroa$052$0 = 0, $$sroa$052$1 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0;
 var $exitcond$i$i38 = 0, $exitcond$i$i40 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 480|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(480|0);
 $$byval_copy = sp + 468|0;
 $6 = sp + 464|0;
 $7 = sp + 472|0;
 $8 = sp + 460|0;
 $9 = sp + 456|0;
 $10 = sp + 444|0;
 $11 = sp + 432|0;
 $12 = sp + 420|0;
 $13 = sp + 416|0;
 $14 = sp + 16|0;
 $15 = sp + 8|0;
 $16 = sp + 4|0;
 $17 = sp;
 __ZNKSt3__28ios_base6getlocEv($6,$3);
 __THREW__ = 0;
 $18 = (invoke_iii(734,($6|0),(165472|0))|0);
 $19 = __THREW__; __THREW__ = 0;
 $20 = $19&1;
 L1: do {
  if ($20) {
   $101 = ___cxa_find_matching_catch_2()|0;
   $102 = tempRet0;
   $$3 = $101;$$336 = $102;
  } else {
   $21 = ((($5)) + 8|0);
   $22 = ((($21)) + 3|0);
   $23 = HEAP8[$22>>0]|0;
   $24 = ($23<<24>>24)<(0);
   $25 = ((($5)) + 4|0);
   $26 = HEAP32[$25>>2]|0;
   $27 = $23&255;
   $28 = $24 ? $26 : $27;
   $29 = ($28|0)==(0);
   do {
    if ($29) {
     $46 = 0;
    } else {
     $30 = HEAP32[$5>>2]|0;
     $31 = $24 ? $30 : $5;
     $32 = HEAP32[$31>>2]|0;
     $33 = HEAP32[$18>>2]|0;
     $34 = ((($33)) + 44|0);
     $35 = HEAP32[$34>>2]|0;
     __THREW__ = 0;
     $36 = (invoke_iii($35|0,($18|0),45)|0);
     $37 = __THREW__; __THREW__ = 0;
     $38 = $37&1;
     if ($38) {
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      $$3 = $103;$$336 = $104;
      break L1;
     } else {
      $39 = ($32|0)==($36|0);
      $46 = $39;
      break;
     }
    }
   } while(0);
   ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
   $$0$i$i = 0;
   while(1) {
    $exitcond$i$i = ($$0$i$i|0)==(3);
    if ($exitcond$i$i) {
     break;
    }
    $40 = (($10) + ($$0$i$i<<2)|0);
    HEAP32[$40>>2] = 0;
    $41 = (($$0$i$i) + 1)|0;
    $$0$i$i = $41;
   }
   ;HEAP32[$11>>2]=0|0;HEAP32[$11+4>>2]=0|0;HEAP32[$11+8>>2]=0|0;
   $$0$i$i37 = 0;
   while(1) {
    $exitcond$i$i38 = ($$0$i$i37|0)==(3);
    if ($exitcond$i$i38) {
     break;
    }
    $42 = (($11) + ($$0$i$i37<<2)|0);
    HEAP32[$42>>2] = 0;
    $43 = (($$0$i$i37) + 1)|0;
    $$0$i$i37 = $43;
   }
   ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;
   $$0$i$i39 = 0;
   while(1) {
    $exitcond$i$i40 = ($$0$i$i39|0)==(3);
    if ($exitcond$i$i40) {
     break;
    }
    $44 = (($12) + ($$0$i$i39<<2)|0);
    HEAP32[$44>>2] = 0;
    $45 = (($$0$i$i39) + 1)|0;
    $$0$i$i39 = $45;
   }
   __THREW__ = 0;
   invoke_viiiiiiiiii(1022,($2|0),($46|0),($6|0),($7|0),($8|0),($9|0),($10|0),($11|0),($12|0),($13|0));
   $47 = __THREW__; __THREW__ = 0;
   $48 = $47&1;
   if ($48) {
    $105 = ___cxa_find_matching_catch_2()|0;
    $106 = tempRet0;
    $$1 = $105;$$134 = $106;
   } else {
    $49 = HEAP8[$22>>0]|0;
    $50 = ($49<<24>>24)<(0);
    $51 = HEAP32[$25>>2]|0;
    $52 = $49&255;
    $53 = $50 ? $51 : $52;
    $54 = HEAP32[$13>>2]|0;
    $55 = ($53|0)>($54|0);
    if ($55) {
     $56 = (($53) - ($54))|0;
     $57 = $56 << 1;
     $58 = ((($12)) + 8|0);
     $59 = ((($58)) + 3|0);
     $60 = HEAP8[$59>>0]|0;
     $61 = ($60<<24>>24)<(0);
     $62 = ((($12)) + 4|0);
     $63 = HEAP32[$62>>2]|0;
     $64 = $60&255;
     $65 = $61 ? $63 : $64;
     $66 = (($65) + ($57))|0;
     $67 = ((($11)) + 8|0);
     $68 = ((($67)) + 3|0);
     $69 = HEAP8[$68>>0]|0;
     $70 = ($69<<24>>24)<(0);
     $71 = ((($11)) + 4|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = $69&255;
     $74 = $70 ? $72 : $73;
     $$sink1 = 1;$$sink62 = $66;$$sink63 = $74;
    } else {
     $75 = ((($12)) + 8|0);
     $76 = ((($75)) + 3|0);
     $77 = HEAP8[$76>>0]|0;
     $78 = ($77<<24>>24)<(0);
     $79 = ((($12)) + 4|0);
     $80 = HEAP32[$79>>2]|0;
     $81 = $77&255;
     $82 = $78 ? $80 : $81;
     $83 = ((($11)) + 8|0);
     $84 = ((($83)) + 3|0);
     $85 = HEAP8[$84>>0]|0;
     $86 = ($85<<24>>24)<(0);
     $87 = ((($11)) + 4|0);
     $88 = HEAP32[$87>>2]|0;
     $89 = $85&255;
     $90 = $86 ? $88 : $89;
     $$sink1 = 2;$$sink62 = $90;$$sink63 = $82;
    }
    $91 = (($$sink63) + ($54))|0;
    $$sink = (($91) + ($$sink62))|0;
    $92 = (($$sink) + ($$sink1))|0;
    $93 = ($92>>>0)>(100);
    if ($93) {
     $94 = $92 << 2;
     $95 = (_malloc($94)|0);
     $96 = $95;
     $97 = ($95|0)==(0|0);
     if ($97) {
      __THREW__ = 0;
      invoke_v(732);
      $98 = __THREW__; __THREW__ = 0;
      $99 = ___cxa_find_matching_catch_2()|0;
      $100 = tempRet0;
      $$0 = $99;$$033 = $100;$$sroa$052$1 = $96;
     } else {
      $$032 = $95;$$sroa$052$0 = $96;
      label = 24;
     }
    } else {
     $$032 = $14;$$sroa$052$0 = 0;
     label = 24;
    }
    if ((label|0) == 24) {
     $107 = ((($3)) + 4|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = HEAP32[$5>>2]|0;
     $110 = $50 ? $109 : $5;
     $111 = (($110) + ($53<<2)|0);
     $112 = HEAP32[$8>>2]|0;
     $113 = HEAP32[$9>>2]|0;
     __THREW__ = 0;
     invoke_viiiiiiiiiiiiiii(1023,($$032|0),($15|0),($16|0),($108|0),($110|0),($111|0),($18|0),($46|0),($7|0),($112|0),($113|0),($10|0),($11|0),($12|0),($54|0));
     $114 = __THREW__; __THREW__ = 0;
     $115 = $114&1;
     if (!($115)) {
      $116 = HEAP32[$1>>2]|0;
      HEAP32[$17>>2] = $116;
      $117 = HEAP32[$15>>2]|0;
      $118 = HEAP32[$16>>2]|0;
      __THREW__ = 0;
      ;HEAP32[$$byval_copy>>2]=HEAP32[$17>>2]|0;
      $119 = (invoke_iiiiiii(735,($$byval_copy|0),($$032|0),($117|0),($118|0),($3|0),($4|0))|0);
      $120 = __THREW__; __THREW__ = 0;
      $121 = $120&1;
      if (!($121)) {
       $122 = ($$sroa$052$0|0)==(0);
       if (!($122)) {
        $123 = $$sroa$052$0;
        _free($123);
       }
       __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($12);
       __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($11);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
       __ZNSt3__26localeD2Ev($6);
       STACKTOP = sp;return ($119|0);
      }
     }
     $124 = ___cxa_find_matching_catch_2()|0;
     $125 = tempRet0;
     $$0 = $124;$$033 = $125;$$sroa$052$1 = $$sroa$052$0;
    }
    $126 = ($$sroa$052$1|0)==(0);
    if (!($126)) {
     $127 = $$sroa$052$1;
     _free($127);
    }
    $$1 = $$0;$$134 = $$033;
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($12);
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($11);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($10);
   $$3 = $$1;$$336 = $$134;
  }
 } while(0);
 __ZNSt3__26localeD2Ev($6);
 ___resumeException($$3|0);
 // unreachable;
 return (0)|0;
}
function __ZNSt3__211__money_putIwE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_Ri($0,$1,$2,$3,$4,$5,$6,$7,$8,$9) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 var $$0$i$i$i = 0, $$0$i$i$i48 = 0, $$0$i$i$i51 = 0, $$0$i$i$i55 = 0, $$0$i$i$i60 = 0, $$0$i$i$i65 = 0, $$0$i$i$i70 = 0, $$0$i$i$i75 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i49 = 0, $exitcond$i$i$i52 = 0, $exitcond$i$i$i56 = 0;
 var $exitcond$i$i$i61 = 0, $exitcond$i$i$i66 = 0, $exitcond$i$i$i71 = 0, $exitcond$i$i$i76 = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $10 = sp + 156|0;
 $11 = sp + 152|0;
 $12 = sp + 163|0;
 $13 = sp + 162|0;
 $14 = sp + 148|0;
 $15 = sp + 144|0;
 $16 = sp + 140|0;
 $17 = sp + 136|0;
 $18 = sp + 132|0;
 $19 = sp + 128|0;
 $20 = sp + 161|0;
 $21 = sp + 160|0;
 $22 = sp + 124|0;
 $23 = sp + 120|0;
 $24 = sp + 116|0;
 $25 = sp + 112|0;
 $26 = sp + 108|0;
 $27 = sp + 96|0;
 $28 = sp + 92|0;
 $29 = sp + 80|0;
 $30 = sp + 68|0;
 $31 = sp + 56|0;
 $32 = sp + 52|0;
 $33 = sp + 40|0;
 $34 = sp + 36|0;
 $35 = sp + 24|0;
 $36 = sp + 12|0;
 $37 = sp;
 if ($0) {
  $38 = (__ZNKSt3__26locale9use_facetERNS0_2idE($2,167176)|0);
  if ($1) {
   $39 = HEAP32[$38>>2]|0;
   $40 = ((($39)) + 44|0);
   $41 = HEAP32[$40>>2]|0;
   FUNCTION_TABLE_vii[$41 & 2047]($26,$38);
   $42 = HEAP32[$26>>2]|0;
   HEAP8[$3>>0]=$42&255;HEAP8[$3+1>>0]=($42>>8)&255;HEAP8[$3+2>>0]=($42>>16)&255;HEAP8[$3+3>>0]=$42>>24;
   $43 = HEAP32[$38>>2]|0;
   $44 = ((($43)) + 32|0);
   $45 = HEAP32[$44>>2]|0;
   FUNCTION_TABLE_vii[$45 & 2047]($27,$38);
   $46 = ((($8)) + 8|0);
   $47 = ((($46)) + 3|0);
   $48 = HEAP8[$47>>0]|0;
   $49 = ($48<<24>>24)<(0);
   if ($49) {
    $50 = HEAP32[$8>>2]|0;
    HEAP32[$24>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($50,$24);
    $51 = ((($8)) + 4|0);
    HEAP32[$51>>2] = 0;
   } else {
    HEAP32[$25>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($8,$25);
    HEAP8[$47>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1019,($8|0),0);
   $52 = __THREW__; __THREW__ = 0;
   $53 = $52&1;
   if ($53) {
    $54 = ___cxa_find_matching_catch_3(0|0)|0;
    $55 = tempRet0;
    ___clang_call_terminate($54);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$27>>2]|0;HEAP32[$8+4>>2]=HEAP32[$27+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$27+8>>2]|0;
   $$0$i$i$i = 0;
   while(1) {
    $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
    if ($exitcond$i$i$i) {
     break;
    }
    $56 = (($27) + ($$0$i$i$i<<2)|0);
    HEAP32[$56>>2] = 0;
    $57 = (($$0$i$i$i) + 1)|0;
    $$0$i$i$i = $57;
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($27);
  } else {
   $58 = HEAP32[$38>>2]|0;
   $59 = ((($58)) + 40|0);
   $60 = HEAP32[$59>>2]|0;
   FUNCTION_TABLE_vii[$60 & 2047]($28,$38);
   $61 = HEAP32[$28>>2]|0;
   HEAP8[$3>>0]=$61&255;HEAP8[$3+1>>0]=($61>>8)&255;HEAP8[$3+2>>0]=($61>>16)&255;HEAP8[$3+3>>0]=$61>>24;
   $62 = HEAP32[$38>>2]|0;
   $63 = ((($62)) + 28|0);
   $64 = HEAP32[$63>>2]|0;
   FUNCTION_TABLE_vii[$64 & 2047]($29,$38);
   $65 = ((($8)) + 8|0);
   $66 = ((($65)) + 3|0);
   $67 = HEAP8[$66>>0]|0;
   $68 = ($67<<24>>24)<(0);
   if ($68) {
    $69 = HEAP32[$8>>2]|0;
    HEAP32[$22>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($69,$22);
    $70 = ((($8)) + 4|0);
    HEAP32[$70>>2] = 0;
   } else {
    HEAP32[$23>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($8,$23);
    HEAP8[$66>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1019,($8|0),0);
   $71 = __THREW__; __THREW__ = 0;
   $72 = $71&1;
   if ($72) {
    $73 = ___cxa_find_matching_catch_3(0|0)|0;
    $74 = tempRet0;
    ___clang_call_terminate($73);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$29>>2]|0;HEAP32[$8+4>>2]=HEAP32[$29+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$29+8>>2]|0;
   $$0$i$i$i48 = 0;
   while(1) {
    $exitcond$i$i$i49 = ($$0$i$i$i48|0)==(3);
    if ($exitcond$i$i$i49) {
     break;
    }
    $75 = (($29) + ($$0$i$i$i48<<2)|0);
    HEAP32[$75>>2] = 0;
    $76 = (($$0$i$i$i48) + 1)|0;
    $$0$i$i$i48 = $76;
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($29);
  }
  $77 = HEAP32[$38>>2]|0;
  $78 = ((($77)) + 12|0);
  $79 = HEAP32[$78>>2]|0;
  $80 = (FUNCTION_TABLE_ii[$79 & 2047]($38)|0);
  HEAP32[$4>>2] = $80;
  $81 = HEAP32[$38>>2]|0;
  $82 = ((($81)) + 16|0);
  $83 = HEAP32[$82>>2]|0;
  $84 = (FUNCTION_TABLE_ii[$83 & 2047]($38)|0);
  HEAP32[$5>>2] = $84;
  $85 = HEAP32[$38>>2]|0;
  $86 = ((($85)) + 20|0);
  $87 = HEAP32[$86>>2]|0;
  FUNCTION_TABLE_vii[$87 & 2047]($30,$38);
  $88 = ((($6)) + 11|0);
  $89 = HEAP8[$88>>0]|0;
  $90 = ($89<<24>>24)<(0);
  if ($90) {
   $91 = HEAP32[$6>>2]|0;
   HEAP8[$20>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($91,$20);
   $92 = ((($6)) + 4|0);
   HEAP32[$92>>2] = 0;
  } else {
   HEAP8[$21>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$21);
   HEAP8[$88>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $93 = __THREW__; __THREW__ = 0;
  $94 = $93&1;
  if ($94) {
   $95 = ___cxa_find_matching_catch_3(0|0)|0;
   $96 = tempRet0;
   ___clang_call_terminate($95);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$30>>2]|0;HEAP32[$6+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$30+8>>2]|0;
  $$0$i$i$i51 = 0;
  while(1) {
   $exitcond$i$i$i52 = ($$0$i$i$i51|0)==(3);
   if ($exitcond$i$i$i52) {
    break;
   }
   $97 = (($30) + ($$0$i$i$i51<<2)|0);
   HEAP32[$97>>2] = 0;
   $98 = (($$0$i$i$i51) + 1)|0;
   $$0$i$i$i51 = $98;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($30);
  $99 = HEAP32[$38>>2]|0;
  $100 = ((($99)) + 24|0);
  $101 = HEAP32[$100>>2]|0;
  FUNCTION_TABLE_vii[$101 & 2047]($31,$38);
  $102 = ((($7)) + 8|0);
  $103 = ((($102)) + 3|0);
  $104 = HEAP8[$103>>0]|0;
  $105 = ($104<<24>>24)<(0);
  if ($105) {
   $106 = HEAP32[$7>>2]|0;
   HEAP32[$18>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($106,$18);
   $107 = ((($7)) + 4|0);
   HEAP32[$107>>2] = 0;
  } else {
   HEAP32[$19>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($7,$19);
   HEAP8[$103>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($7|0),0);
  $108 = __THREW__; __THREW__ = 0;
  $109 = $108&1;
  if ($109) {
   $110 = ___cxa_find_matching_catch_3(0|0)|0;
   $111 = tempRet0;
   ___clang_call_terminate($110);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$31>>2]|0;HEAP32[$7+4>>2]=HEAP32[$31+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$31+8>>2]|0;
  $$0$i$i$i55 = 0;
  while(1) {
   $exitcond$i$i$i56 = ($$0$i$i$i55|0)==(3);
   if ($exitcond$i$i$i56) {
    break;
   }
   $112 = (($31) + ($$0$i$i$i55<<2)|0);
   HEAP32[$112>>2] = 0;
   $113 = (($$0$i$i$i55) + 1)|0;
   $$0$i$i$i55 = $113;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($31);
  $114 = HEAP32[$38>>2]|0;
  $115 = ((($114)) + 36|0);
  $116 = HEAP32[$115>>2]|0;
  $117 = (FUNCTION_TABLE_ii[$116 & 2047]($38)|0);
  $storemerge = $117;
 } else {
  $118 = (__ZNKSt3__26locale9use_facetERNS0_2idE($2,167168)|0);
  if ($1) {
   $119 = HEAP32[$118>>2]|0;
   $120 = ((($119)) + 44|0);
   $121 = HEAP32[$120>>2]|0;
   FUNCTION_TABLE_vii[$121 & 2047]($32,$118);
   $122 = HEAP32[$32>>2]|0;
   HEAP8[$3>>0]=$122&255;HEAP8[$3+1>>0]=($122>>8)&255;HEAP8[$3+2>>0]=($122>>16)&255;HEAP8[$3+3>>0]=$122>>24;
   $123 = HEAP32[$118>>2]|0;
   $124 = ((($123)) + 32|0);
   $125 = HEAP32[$124>>2]|0;
   FUNCTION_TABLE_vii[$125 & 2047]($33,$118);
   $126 = ((($8)) + 8|0);
   $127 = ((($126)) + 3|0);
   $128 = HEAP8[$127>>0]|0;
   $129 = ($128<<24>>24)<(0);
   if ($129) {
    $130 = HEAP32[$8>>2]|0;
    HEAP32[$16>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($130,$16);
    $131 = ((($8)) + 4|0);
    HEAP32[$131>>2] = 0;
   } else {
    HEAP32[$17>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($8,$17);
    HEAP8[$127>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1019,($8|0),0);
   $132 = __THREW__; __THREW__ = 0;
   $133 = $132&1;
   if ($133) {
    $134 = ___cxa_find_matching_catch_3(0|0)|0;
    $135 = tempRet0;
    ___clang_call_terminate($134);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$33>>2]|0;HEAP32[$8+4>>2]=HEAP32[$33+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$33+8>>2]|0;
   $$0$i$i$i60 = 0;
   while(1) {
    $exitcond$i$i$i61 = ($$0$i$i$i60|0)==(3);
    if ($exitcond$i$i$i61) {
     break;
    }
    $136 = (($33) + ($$0$i$i$i60<<2)|0);
    HEAP32[$136>>2] = 0;
    $137 = (($$0$i$i$i60) + 1)|0;
    $$0$i$i$i60 = $137;
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($33);
  } else {
   $138 = HEAP32[$118>>2]|0;
   $139 = ((($138)) + 40|0);
   $140 = HEAP32[$139>>2]|0;
   FUNCTION_TABLE_vii[$140 & 2047]($34,$118);
   $141 = HEAP32[$34>>2]|0;
   HEAP8[$3>>0]=$141&255;HEAP8[$3+1>>0]=($141>>8)&255;HEAP8[$3+2>>0]=($141>>16)&255;HEAP8[$3+3>>0]=$141>>24;
   $142 = HEAP32[$118>>2]|0;
   $143 = ((($142)) + 28|0);
   $144 = HEAP32[$143>>2]|0;
   FUNCTION_TABLE_vii[$144 & 2047]($35,$118);
   $145 = ((($8)) + 8|0);
   $146 = ((($145)) + 3|0);
   $147 = HEAP8[$146>>0]|0;
   $148 = ($147<<24>>24)<(0);
   if ($148) {
    $149 = HEAP32[$8>>2]|0;
    HEAP32[$14>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($149,$14);
    $150 = ((($8)) + 4|0);
    HEAP32[$150>>2] = 0;
   } else {
    HEAP32[$15>>2] = 0;
    __ZNSt3__211char_traitsIwE6assignERwRKw($8,$15);
    HEAP8[$146>>0] = 0;
   }
   __THREW__ = 0;
   invoke_vii(1019,($8|0),0);
   $151 = __THREW__; __THREW__ = 0;
   $152 = $151&1;
   if ($152) {
    $153 = ___cxa_find_matching_catch_3(0|0)|0;
    $154 = tempRet0;
    ___clang_call_terminate($153);
    // unreachable;
   }
   ;HEAP32[$8>>2]=HEAP32[$35>>2]|0;HEAP32[$8+4>>2]=HEAP32[$35+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$35+8>>2]|0;
   $$0$i$i$i65 = 0;
   while(1) {
    $exitcond$i$i$i66 = ($$0$i$i$i65|0)==(3);
    if ($exitcond$i$i$i66) {
     break;
    }
    $155 = (($35) + ($$0$i$i$i65<<2)|0);
    HEAP32[$155>>2] = 0;
    $156 = (($$0$i$i$i65) + 1)|0;
    $$0$i$i$i65 = $156;
   }
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($35);
  }
  $157 = HEAP32[$118>>2]|0;
  $158 = ((($157)) + 12|0);
  $159 = HEAP32[$158>>2]|0;
  $160 = (FUNCTION_TABLE_ii[$159 & 2047]($118)|0);
  HEAP32[$4>>2] = $160;
  $161 = HEAP32[$118>>2]|0;
  $162 = ((($161)) + 16|0);
  $163 = HEAP32[$162>>2]|0;
  $164 = (FUNCTION_TABLE_ii[$163 & 2047]($118)|0);
  HEAP32[$5>>2] = $164;
  $165 = HEAP32[$118>>2]|0;
  $166 = ((($165)) + 20|0);
  $167 = HEAP32[$166>>2]|0;
  FUNCTION_TABLE_vii[$167 & 2047]($36,$118);
  $168 = ((($6)) + 11|0);
  $169 = HEAP8[$168>>0]|0;
  $170 = ($169<<24>>24)<(0);
  if ($170) {
   $171 = HEAP32[$6>>2]|0;
   HEAP8[$12>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($171,$12);
   $172 = ((($6)) + 4|0);
   HEAP32[$172>>2] = 0;
  } else {
   HEAP8[$13>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($6,$13);
   HEAP8[$168>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1012,($6|0),0);
  $173 = __THREW__; __THREW__ = 0;
  $174 = $173&1;
  if ($174) {
   $175 = ___cxa_find_matching_catch_3(0|0)|0;
   $176 = tempRet0;
   ___clang_call_terminate($175);
   // unreachable;
  }
  ;HEAP32[$6>>2]=HEAP32[$36>>2]|0;HEAP32[$6+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$36+8>>2]|0;
  $$0$i$i$i70 = 0;
  while(1) {
   $exitcond$i$i$i71 = ($$0$i$i$i70|0)==(3);
   if ($exitcond$i$i$i71) {
    break;
   }
   $177 = (($36) + ($$0$i$i$i70<<2)|0);
   HEAP32[$177>>2] = 0;
   $178 = (($$0$i$i$i70) + 1)|0;
   $$0$i$i$i70 = $178;
  }
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($36);
  $179 = HEAP32[$118>>2]|0;
  $180 = ((($179)) + 24|0);
  $181 = HEAP32[$180>>2]|0;
  FUNCTION_TABLE_vii[$181 & 2047]($37,$118);
  $182 = ((($7)) + 8|0);
  $183 = ((($182)) + 3|0);
  $184 = HEAP8[$183>>0]|0;
  $185 = ($184<<24>>24)<(0);
  if ($185) {
   $186 = HEAP32[$7>>2]|0;
   HEAP32[$10>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($186,$10);
   $187 = ((($7)) + 4|0);
   HEAP32[$187>>2] = 0;
  } else {
   HEAP32[$11>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($7,$11);
   HEAP8[$183>>0] = 0;
  }
  __THREW__ = 0;
  invoke_vii(1019,($7|0),0);
  $188 = __THREW__; __THREW__ = 0;
  $189 = $188&1;
  if ($189) {
   $190 = ___cxa_find_matching_catch_3(0|0)|0;
   $191 = tempRet0;
   ___clang_call_terminate($190);
   // unreachable;
  }
  ;HEAP32[$7>>2]=HEAP32[$37>>2]|0;HEAP32[$7+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$37+8>>2]|0;
  $$0$i$i$i75 = 0;
  while(1) {
   $exitcond$i$i$i76 = ($$0$i$i$i75|0)==(3);
   if ($exitcond$i$i$i76) {
    break;
   }
   $192 = (($37) + ($$0$i$i$i75<<2)|0);
   HEAP32[$192>>2] = 0;
   $193 = (($$0$i$i$i75) + 1)|0;
   $$0$i$i$i75 = $193;
  }
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($37);
  $194 = HEAP32[$118>>2]|0;
  $195 = ((($194)) + 36|0);
  $196 = HEAP32[$195>>2]|0;
  $197 = (FUNCTION_TABLE_ii[$196 & 2047]($118)|0);
  $storemerge = $197;
 }
 HEAP32[$9>>2] = $storemerge;
 STACKTOP = sp;return;
}
function __ZNSt3__211__money_putIwE8__formatEPwRS2_S3_jPKwS5_RKNS_5ctypeIwEEbRKNS_10money_base7patternEwwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNSE_IwNSF_IwEENSH_IwEEEESQ_i($0,$1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11,$12,$13,$14) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 $8 = $8|0;
 $9 = $9|0;
 $10 = $10|0;
 $11 = $11|0;
 $12 = $12|0;
 $13 = $13|0;
 $14 = $14|0;
 var $$ = 0, $$$099 = 0, $$0 = 0, $$0$i$i = 0, $$0$i$i108 = 0, $$0$i$i112 = 0, $$0101 = 0, $$0103 = 0, $$0106 = 0, $$07$i$i = 0, $$095 = 0, $$097 = 0, $$099 = 0, $$1 = 0, $$1102 = 0, $$1104 = 0, $$196 = 0, $$198 = 0, $$2 = 0, $$2105 = 0;
 var $$3 = 0, $$pre137 = 0, $$promoted = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0;
 var $117 = 0, $118 = 0, $119 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond = 0, $or$cond = 0, $scevgep$i = 0, $scevgep$i113 = 0, $trunc = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 $15 = ((($13)) + 8|0);
 $16 = ((($15)) + 3|0);
 $17 = ((($13)) + 4|0);
 $18 = ((($12)) + 8|0);
 $19 = ((($18)) + 3|0);
 $20 = ((($12)) + 4|0);
 $21 = $3 & 512;
 $22 = ($21|0)==(0);
 $23 = ($14|0)>(0);
 $24 = ((($11)) + 11|0);
 $25 = ((($11)) + 4|0);
 $$0106 = 0;$$099 = $4;
 while(1) {
  $exitcond = ($$0106|0)==(4);
  if ($exitcond) {
   break;
  }
  $32 = (($8) + ($$0106)|0);
  $33 = HEAP8[$32>>0]|0;
  $34 = $33 << 24 >> 24;
  L4: do {
   switch ($34|0) {
   case 0:  {
    $35 = HEAP32[$2>>2]|0;
    HEAP32[$1>>2] = $35;
    $$2 = $$099;
    break;
   }
   case 1:  {
    $36 = HEAP32[$2>>2]|0;
    HEAP32[$1>>2] = $36;
    $37 = HEAP32[$6>>2]|0;
    $38 = ((($37)) + 44|0);
    $39 = HEAP32[$38>>2]|0;
    $40 = (FUNCTION_TABLE_iii[$39 & 2047]($6,32)|0);
    $41 = HEAP32[$2>>2]|0;
    $42 = ((($41)) + 4|0);
    HEAP32[$2>>2] = $42;
    HEAP32[$41>>2] = $40;
    $$2 = $$099;
    break;
   }
   case 3:  {
    $43 = HEAP8[$16>>0]|0;
    $44 = ($43<<24>>24)<(0);
    $45 = HEAP32[$17>>2]|0;
    $46 = $43&255;
    $47 = $44 ? $45 : $46;
    $48 = ($47|0)==(0);
    if ($48) {
     $$2 = $$099;
    } else {
     $49 = HEAP32[$13>>2]|0;
     $50 = $44 ? $49 : $13;
     $51 = HEAP32[$50>>2]|0;
     $52 = HEAP32[$2>>2]|0;
     $53 = ((($52)) + 4|0);
     HEAP32[$2>>2] = $53;
     HEAP32[$52>>2] = $51;
     $$2 = $$099;
    }
    break;
   }
   case 2:  {
    $54 = HEAP8[$19>>0]|0;
    $55 = ($54<<24>>24)<(0);
    $56 = HEAP32[$20>>2]|0;
    $57 = $54&255;
    $58 = $55 ? $56 : $57;
    $59 = ($58|0)==(0);
    $or$cond = $22 | $59;
    if ($or$cond) {
     $$2 = $$099;
    } else {
     $60 = HEAP32[$12>>2]|0;
     $61 = $55 ? $60 : $12;
     $62 = (($61) + ($58<<2)|0);
     $63 = HEAP32[$2>>2]|0;
     $$0$i$i112 = $63;$65 = $61;
     while(1) {
      $64 = ($65|0)==($62|0);
      if ($64) {
       break;
      }
      $66 = HEAP32[$65>>2]|0;
      HEAP32[$$0$i$i112>>2] = $66;
      $67 = ((($65)) + 4|0);
      $68 = ((($$0$i$i112)) + 4|0);
      $$0$i$i112 = $68;$65 = $67;
     }
     $scevgep$i113 = (($63) + ($58<<2)|0);
     HEAP32[$2>>2] = $scevgep$i113;
     $$2 = $$099;
    }
    break;
   }
   case 4:  {
    $69 = HEAP32[$2>>2]|0;
    $70 = ((($$099)) + 4|0);
    $$$099 = $7 ? $70 : $$099;
    $$0103 = $$$099;
    while(1) {
     $71 = ($$0103>>>0)<($5>>>0);
     if (!($71)) {
      break;
     }
     $72 = HEAP32[$$0103>>2]|0;
     $73 = HEAP32[$6>>2]|0;
     $74 = ((($73)) + 12|0);
     $75 = HEAP32[$74>>2]|0;
     $76 = (FUNCTION_TABLE_iiii[$75 & 2047]($6,2048,$72)|0);
     if (!($76)) {
      break;
     }
     $77 = ((($$0103)) + 4|0);
     $$0103 = $77;
    }
    if ($23) {
     $$0101 = $14;$$1104 = $$0103;
     while(1) {
      $78 = ($$1104>>>0)>($$$099>>>0);
      $79 = ($$0101|0)>(0);
      $80 = $78 & $79;
      if (!($80)) {
       break;
      }
      $81 = ((($$1104)) + -4|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = HEAP32[$2>>2]|0;
      $84 = ((($83)) + 4|0);
      HEAP32[$2>>2] = $84;
      HEAP32[$83>>2] = $82;
      $85 = (($$0101) + -1)|0;
      $$0101 = $85;$$1104 = $81;
     }
     if ($79) {
      $86 = HEAP32[$6>>2]|0;
      $87 = ((($86)) + 44|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = (FUNCTION_TABLE_iii[$88 & 2047]($6,48)|0);
      $93 = $89;
     } else {
      $93 = 0;
     }
     $$promoted = HEAP32[$2>>2]|0;
     $$1102 = $$0101;$92 = $$promoted;
     while(1) {
      $90 = ($$1102|0)>(0);
      $91 = ((($92)) + 4|0);
      if (!($90)) {
       break;
      }
      HEAP32[$92>>2] = $93;
      $94 = (($$1102) + -1)|0;
      $$1102 = $94;$92 = $91;
     }
     HEAP32[$2>>2] = $91;
     HEAP32[$92>>2] = $9;
     $$2105 = $$1104;
    } else {
     $$2105 = $$0103;
    }
    $95 = ($$2105|0)==($$$099|0);
    if ($95) {
     $96 = HEAP32[$6>>2]|0;
     $97 = ((($96)) + 44|0);
     $98 = HEAP32[$97>>2]|0;
     $99 = (FUNCTION_TABLE_iii[$98 & 2047]($6,48)|0);
     $100 = HEAP32[$2>>2]|0;
     $101 = ((($100)) + 4|0);
     HEAP32[$2>>2] = $101;
     HEAP32[$100>>2] = $99;
     $132 = $101;
    } else {
     $102 = HEAP8[$24>>0]|0;
     $103 = ($102<<24>>24)<(0);
     $104 = HEAP32[$25>>2]|0;
     $105 = $102&255;
     $106 = $103 ? $104 : $105;
     $107 = ($106|0)==(0);
     if ($107) {
      $$0 = -1;$$095 = 0;$$097 = 0;$$3 = $$2105;
     } else {
      $108 = HEAP32[$11>>2]|0;
      $109 = $103 ? $108 : $11;
      $110 = HEAP8[$109>>0]|0;
      $111 = $110 << 24 >> 24;
      $$0 = $111;$$095 = 0;$$097 = 0;$$3 = $$2105;
     }
     while(1) {
      $112 = ($$3|0)==($$$099|0);
      if ($112) {
       break;
      }
      $113 = ($$097|0)==($$0|0);
      $114 = HEAP32[$2>>2]|0;
      if ($113) {
       $115 = ((($114)) + 4|0);
       HEAP32[$2>>2] = $115;
       HEAP32[$114>>2] = $10;
       $116 = (($$095) + 1)|0;
       $117 = HEAP32[$25>>2]|0;
       $118 = $103 ? $117 : $105;
       $119 = ($116>>>0)<($118>>>0);
       if ($119) {
        $120 = HEAP32[$11>>2]|0;
        $121 = $103 ? $120 : $11;
        $122 = (($121) + ($116)|0);
        $123 = HEAP8[$122>>0]|0;
        $124 = ($123<<24>>24)==(127);
        $125 = $123 << 24 >> 24;
        $$ = $124 ? -1 : $125;
        $$1 = $$;$$196 = $116;$$198 = 0;$129 = $115;
       } else {
        $$1 = $$097;$$196 = $116;$$198 = 0;$129 = $115;
       }
      } else {
       $$1 = $$0;$$196 = $$095;$$198 = $$097;$129 = $114;
      }
      $126 = ((($$3)) + -4|0);
      $127 = HEAP32[$126>>2]|0;
      $128 = ((($129)) + 4|0);
      HEAP32[$2>>2] = $128;
      HEAP32[$129>>2] = $127;
      $130 = (($$198) + 1)|0;
      $$0 = $$1;$$095 = $$196;$$097 = $130;$$3 = $126;
     }
     $$pre137 = HEAP32[$2>>2]|0;
     $132 = $$pre137;
    }
    $131 = ($69|0)==($132|0);
    if ($131) {
     $$2 = $$$099;
    } else {
     $$0$i$i108 = $132;$$07$i$i = $69;
     while(1) {
      $133 = ((($$0$i$i108)) + -4|0);
      $134 = ($$07$i$i>>>0)<($133>>>0);
      if (!($134)) {
       $$2 = $$$099;
       break L4;
      }
      $135 = HEAP32[$$07$i$i>>2]|0;
      $136 = HEAP32[$133>>2]|0;
      HEAP32[$$07$i$i>>2] = $136;
      HEAP32[$133>>2] = $135;
      $137 = ((($$07$i$i)) + 4|0);
      $$0$i$i108 = $133;$$07$i$i = $137;
     }
    }
    break;
   }
   default: {
    $$2 = $$099;
   }
   }
  } while(0);
  $138 = (($$0106) + 1)|0;
  $$0106 = $138;$$099 = $$2;
 }
 $26 = HEAP8[$16>>0]|0;
 $27 = ($26<<24>>24)<(0);
 $28 = HEAP32[$17>>2]|0;
 $29 = $26&255;
 $30 = $27 ? $28 : $29;
 $31 = ($30>>>0)>(1);
 if ($31) {
  $139 = HEAP32[$13>>2]|0;
  $140 = $27 ? $139 : $13;
  $141 = ((($140)) + 4|0);
  $142 = (($140) + ($30<<2)|0);
  $143 = HEAP32[$2>>2]|0;
  $144 = $141;
  $145 = $142;
  $146 = (($145) - ($144))|0;
  $$0$i$i = $143;$148 = $141;
  while(1) {
   $147 = ($148|0)==($142|0);
   if ($147) {
    break;
   }
   $149 = HEAP32[$148>>2]|0;
   HEAP32[$$0$i$i>>2] = $149;
   $150 = ((($148)) + 4|0);
   $151 = ((($$0$i$i)) + 4|0);
   $$0$i$i = $151;$148 = $150;
  }
  $152 = $146 >>> 2;
  $scevgep$i = (($143) + ($152<<2)|0);
  HEAP32[$2>>2] = $scevgep$i;
 }
 $153 = $3 & 176;
 $trunc = $153&255;
 switch ($trunc<<24>>24) {
 case 32:  {
  $154 = HEAP32[$2>>2]|0;
  HEAP32[$1>>2] = $154;
  break;
 }
 case 16:  {
  break;
 }
 default: {
  HEAP32[$1>>2] = $0;
 }
 }
 return;
}
function __ZNSt3__28messagesIcED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28messagesIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28messagesIcE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($1)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 $6 = HEAP32[$1>>2]|0;
 $7 = $5 ? $6 : $1;
 $8 = (_catopen($7,1)|0);
 $9 = $8;
 $not$ = ($8|0)!=((-1)|0);
 $10 = $not$&1;
 $$ = $9 >>> $10;
 return ($$|0);
}
function __ZNKSt3__28messagesIcE6do_getEiiiRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i25 = 0, $$0$i30 = 0, $$1 = 0, $$123 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, $exitcond$i$i26 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $6 = sp;
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 $9 = ((($5)) + 11|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = ($10<<24>>24)<(0);
 $12 = HEAP32[$5>>2]|0;
 $13 = $11 ? $12 : $5;
 $14 = ((($5)) + 4|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = $10&255;
 $17 = $11 ? $15 : $16;
 $18 = (($13) + ($17)|0);
 $$0$i30 = $13;
 while(1) {
  $19 = ($$0$i30>>>0)<($18>>>0);
  if (!($19)) {
   label = 8;
   break;
  }
  $20 = HEAP8[$$0$i30>>0]|0;
  __THREW__ = 0;
  invoke_vii(1006,($6|0),($20|0));
  $21 = __THREW__; __THREW__ = 0;
  $22 = $21&1;
  if ($22) {
   label = 9;
   break;
  }
  $23 = ((($$0$i30)) + 1|0);
  $$0$i30 = $23;
 }
 do {
  if ((label|0) == 8) {
   $24 = ($2|0)==(-1);
   $25 = $2 << 1;
   $26 = $25;
   $27 = $24 ? (-1) : $26;
   $28 = ((($6)) + 11|0);
   $29 = HEAP8[$28>>0]|0;
   $30 = ($29<<24>>24)<(0);
   $31 = HEAP32[$6>>2]|0;
   $32 = $30 ? $31 : $6;
   __THREW__ = 0;
   $33 = (invoke_iiiii(1024,($27|0),($3|0),($4|0),($32|0))|0);
   $34 = __THREW__; __THREW__ = 0;
   $35 = $34&1;
   if ($35) {
    $47 = ___cxa_find_matching_catch_2()|0;
    $48 = tempRet0;
    $$1 = $47;$$123 = $48;
    break;
   }
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
   $$0$i$i25 = 0;
   while(1) {
    $exitcond$i$i26 = ($$0$i$i25|0)==(3);
    if ($exitcond$i$i26) {
     break;
    }
    $38 = (($0) + ($$0$i$i25<<2)|0);
    HEAP32[$38>>2] = 0;
    $39 = (($$0$i$i25) + 1)|0;
    $$0$i$i25 = $39;
   }
   $40 = (_strlen($33)|0);
   $41 = (($32) + ($40)|0);
   $$0$i = $32;
   while(1) {
    $42 = ($$0$i>>>0)<($41>>>0);
    if (!($42)) {
     label = 19;
     break;
    }
    $43 = HEAP8[$$0$i>>0]|0;
    __THREW__ = 0;
    invoke_vii(1006,($0|0),($43|0));
    $44 = __THREW__; __THREW__ = 0;
    $45 = $44&1;
    if ($45) {
     label = 18;
     break;
    }
    $46 = ((($$0$i)) + 1|0);
    $$0$i = $46;
   }
   if ((label|0) == 18) {
    $49 = ___cxa_find_matching_catch_2()|0;
    $50 = tempRet0;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0);
    $$1 = $49;$$123 = $50;
    break;
   }
   else if ((label|0) == 19) {
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6);
    STACKTOP = sp;return;
   }
  }
  else if ((label|0) == 9) {
   $36 = ___cxa_find_matching_catch_2()|0;
   $37 = tempRet0;
   $$1 = $36;$$123 = $37;
  }
 } while(0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($6);
 ___resumeException($$1|0);
 // unreachable;
}
function __ZNKSt3__28messagesIcE8do_closeEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__28messagesIwED2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28messagesIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28messagesIwE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($1)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 $6 = HEAP32[$1>>2]|0;
 $7 = $5 ? $6 : $1;
 $8 = (_catopen($7,1)|0);
 $9 = $8;
 $not$ = ($8|0)!=((-1)|0);
 $10 = $not$&1;
 $$ = $9 >>> $10;
 return ($$|0);
}
function __ZNKSt3__28messagesIwE6do_getEiiiRKNS_12basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEEE($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0$i = 0, $$0$i$i = 0, $$0$i$i34 = 0, $$0$i29 = 0, $$016$i = 0, $$017$i = 0, $$019$i = 0, $$020$i = 0, $$2 = 0, $$224 = 0, $10 = 0, $100 = 0, $101 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i35 = 0, $or$cond$i = 0, $or$cond$i28 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 224|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(224|0);
 $6 = sp + 184|0;
 $7 = sp + 192|0;
 $8 = sp + 180|0;
 $9 = sp + 176|0;
 $10 = sp + 168|0;
 $11 = sp + 40|0;
 $12 = sp + 32|0;
 $13 = sp + 28|0;
 $14 = sp + 16|0;
 $15 = sp + 8|0;
 $16 = sp;
 ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $17 = (($14) + ($$0$i$i<<2)|0);
  HEAP32[$17>>2] = 0;
  $18 = (($$0$i$i) + 1)|0;
  $$0$i$i = $18;
 }
 $19 = ((($15)) + 4|0);
 HEAP32[$19>>2] = 0;
 HEAP32[$15>>2] = (21360);
 $20 = ((($5)) + 8|0);
 $21 = ((($20)) + 3|0);
 $22 = HEAP8[$21>>0]|0;
 $23 = ($22<<24>>24)<(0);
 $24 = HEAP32[$5>>2]|0;
 $25 = $23 ? $24 : $5;
 $26 = ((($5)) + 4|0);
 $27 = HEAP32[$26>>2]|0;
 $28 = $22&255;
 $29 = $23 ? $27 : $28;
 $30 = (($25) + ($29<<2)|0);
 $31 = ((($7)) + 32|0);
 $$016$i = $25;$$017$i = 0;
 L5: while(1) {
  $32 = ($$016$i>>>0)<($30>>>0);
  $33 = ($$017$i|0)!=(2);
  $34 = $33 & $32;
  if (!($34)) {
   label = 13;
   break;
  }
  HEAP32[$9>>2] = $$016$i;
  $35 = HEAP32[$15>>2]|0;
  $36 = ((($35)) + 12|0);
  $37 = HEAP32[$36>>2]|0;
  __THREW__ = 0;
  $38 = (invoke_iiiiiiiii($37|0,($15|0),($6|0),($$016$i|0),($30|0),($9|0),($7|0),($31|0),($8|0))|0);
  $39 = __THREW__; __THREW__ = 0;
  $40 = $39&1;
  if ($40) {
   label = 14;
   break;
  }
  $41 = ($38|0)==(2);
  $42 = HEAP32[$9>>2]|0;
  $43 = ($42|0)==($$016$i|0);
  $or$cond$i28 = $41 | $43;
  if ($or$cond$i28) {
   label = 8;
   break;
  } else {
   $$0$i29 = $7;
  }
  while(1) {
   $45 = HEAP32[$8>>2]|0;
   $46 = ($$0$i29>>>0)<($45>>>0);
   if (!($46)) {
    break;
   }
   $48 = HEAP8[$$0$i29>>0]|0;
   __THREW__ = 0;
   invoke_vii(1006,($14|0),($48|0));
   $49 = __THREW__; __THREW__ = 0;
   $50 = $49&1;
   if ($50) {
    label = 14;
    break L5;
   }
   $51 = ((($$0$i29)) + 1|0);
   $$0$i29 = $51;
  }
  $47 = HEAP32[$9>>2]|0;
  $$016$i = $47;$$017$i = $38;
 }
 do {
  if ((label|0) == 8) {
   __THREW__ = 0;
   invoke_vi(1005,(156921|0));
   $44 = __THREW__; __THREW__ = 0;
   label = 14;
  }
  else if ((label|0) == 13) {
   __ZNSt3__26locale5facetD2Ev($15);
   $52 = ($2|0)==(-1);
   $53 = $2 << 1;
   $54 = $53;
   $55 = $52 ? (-1) : $54;
   $56 = ((($14)) + 11|0);
   $57 = HEAP8[$56>>0]|0;
   $58 = ($57<<24>>24)<(0);
   $59 = HEAP32[$14>>2]|0;
   $60 = $58 ? $59 : $14;
   __THREW__ = 0;
   $61 = (invoke_iiiii(1024,($55|0),($3|0),($4|0),($60|0))|0);
   $62 = __THREW__; __THREW__ = 0;
   $63 = $62&1;
   if ($63) {
    $98 = ___cxa_find_matching_catch_2()|0;
    $99 = tempRet0;
    $$2 = $98;$$224 = $99;
    break;
   }
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
   $$0$i$i34 = 0;
   while(1) {
    $exitcond$i$i35 = ($$0$i$i34|0)==(3);
    if ($exitcond$i$i35) {
     break;
    }
    $66 = (($0) + ($$0$i$i34<<2)|0);
    HEAP32[$66>>2] = 0;
    $67 = (($$0$i$i34) + 1)|0;
    $$0$i$i34 = $67;
   }
   $68 = ((($16)) + 4|0);
   HEAP32[$68>>2] = 0;
   HEAP32[$16>>2] = (21408);
   $69 = (_strlen($61)|0);
   $70 = (($60) + ($69)|0);
   $71 = $70;
   $72 = ((($11)) + 128|0);
   $$019$i = $60;$$020$i = 0;
   L23: while(1) {
    $73 = ($$019$i>>>0)<($70>>>0);
    $74 = ($$020$i|0)!=(2);
    $75 = $74 & $73;
    if (!($75)) {
     label = 27;
     break;
    }
    HEAP32[$13>>2] = $$019$i;
    $76 = HEAP32[$16>>2]|0;
    $77 = ((($76)) + 16|0);
    $78 = HEAP32[$77>>2]|0;
    $79 = $$019$i;
    $80 = (($71) - ($79))|0;
    $81 = ($80|0)>(32);
    $82 = ((($$019$i)) + 32|0);
    $83 = $81 ? $82 : $70;
    __THREW__ = 0;
    $84 = (invoke_iiiiiiiii($78|0,($16|0),($10|0),($$019$i|0),($83|0),($13|0),($11|0),($72|0),($12|0))|0);
    $85 = __THREW__; __THREW__ = 0;
    $86 = $85&1;
    if ($86) {
     break;
    }
    $87 = ($84|0)==(2);
    $88 = HEAP32[$13>>2]|0;
    $89 = ($88|0)==($$019$i|0);
    $or$cond$i = $87 | $89;
    if ($or$cond$i) {
     label = 22;
     break;
    } else {
     $$0$i = $11;
    }
    while(1) {
     $91 = HEAP32[$12>>2]|0;
     $92 = ($$0$i>>>0)<($91>>>0);
     if (!($92)) {
      break;
     }
     $94 = HEAP32[$$0$i>>2]|0;
     __THREW__ = 0;
     invoke_vii(1014,($0|0),($94|0));
     $95 = __THREW__; __THREW__ = 0;
     $96 = $95&1;
     if ($96) {
      break L23;
     }
     $97 = ((($$0$i)) + 4|0);
     $$0$i = $97;
    }
    $93 = HEAP32[$13>>2]|0;
    $$019$i = $93;$$020$i = $84;
   }
   if ((label|0) == 22) {
    __THREW__ = 0;
    invoke_vi(1005,(156921|0));
    $90 = __THREW__; __THREW__ = 0;
   }
   else if ((label|0) == 27) {
    __ZNSt3__26locale5facetD2Ev($16);
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14);
    STACKTOP = sp;return;
   }
   $100 = ___cxa_find_matching_catch_2()|0;
   $101 = tempRet0;
   __ZNSt3__26locale5facetD2Ev($16);
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($0);
   $$2 = $100;$$224 = $101;
  }
 } while(0);
 if ((label|0) == 14) {
  $64 = ___cxa_find_matching_catch_2()|0;
  $65 = tempRet0;
  __ZNSt3__26locale5facetD2Ev($15);
  $$2 = $64;$$224 = $65;
 }
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($14);
 ___resumeException($$2|0);
 // unreachable;
}
function __ZNKSt3__28messagesIwE8do_closeEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__217__widen_from_utf8ILj32EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE6do_outERS1_PKDiS5_RS5_PcS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $10 = 0, $11 = 0, $12 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp + 4|0;
 $9 = sp;
 HEAP32[$8>>2] = $2;
 HEAP32[$9>>2] = $5;
 $10 = (__ZNSt3__2L12ucs4_to_utf8EPKjS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($2,$3,$8,$5,$6,$9,1114111,0)|0);
 $11 = HEAP32[$8>>2]|0;
 HEAP32[$4>>2] = $11;
 $12 = HEAP32[$9>>2]|0;
 HEAP32[$7>>2] = $12;
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE5do_inERS1_PKcS5_RS5_PDiS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $10 = 0, $11 = 0, $12 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp + 4|0;
 $9 = sp;
 HEAP32[$8>>2] = $2;
 HEAP32[$9>>2] = $5;
 $10 = (__ZNSt3__2L12utf8_to_ucs4EPKhS1_RS1_PjS3_RS3_mNS_12codecvt_modeE($2,$3,$8,$5,$6,$9,1114111,0)|0);
 $11 = HEAP32[$8>>2]|0;
 HEAP32[$4>>2] = $11;
 $12 = HEAP32[$9>>2]|0;
 HEAP32[$7>>2] = $12;
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$4>>2] = $2;
 return 3;
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE11do_encodingEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE16do_always_noconvEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE9do_lengthERS1_PKcS5_j($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = (__ZNSt3__2L19utf8_to_ucs4_lengthEPKhS1_jmNS_12codecvt_modeE($2,$3,$4,1114111,0)|0);
 return ($5|0);
}
function __ZNKSt3__27codecvtIDic11__mbstate_tE13do_max_lengthEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 4;
}
function __ZNSt3__2L19utf8_to_ucs4_lengthEPKhS1_jmNS_12codecvt_modeE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$091 = 0, $$185 = 0, $$488 = 0, $$lcssa112$sink = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $or$cond = 0, $or$cond100 = 0, $or$cond97 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 $5 = $4 & 4;
 $6 = ($5|0)!=(0);
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 $or$cond = $10 & $6;
 if ($or$cond) {
  $11 = HEAP8[$0>>0]|0;
  $12 = ($11<<24>>24)==(-17);
  if ($12) {
   $13 = ((($0)) + 1|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = ($14<<24>>24)==(-69);
   if ($15) {
    $16 = ((($0)) + 2|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(-65);
    $19 = ((($0)) + 3|0);
    $$ = $18 ? $19 : $0;
    $$091 = 0;$$185 = $$;
   } else {
    $$091 = 0;$$185 = $0;
   }
  } else {
   $$091 = 0;$$185 = $0;
  }
 } else {
  $$091 = 0;$$185 = $0;
 }
 L5: while(1) {
  $20 = ($$185>>>0)<($1>>>0);
  $21 = ($$091>>>0)<($2>>>0);
  $or$cond97 = $21 & $20;
  if (!($or$cond97)) {
   label = 30;
   break;
  }
  $22 = HEAP8[$$185>>0]|0;
  $23 = $22&255;
  $24 = ($22<<24>>24)>(-1);
  $25 = ((($$185)) + 1|0);
  do {
   if ($24) {
    $26 = ($23>>>0)>($3>>>0);
    if ($26) {
     label = 30;
     break L5;
    } else {
     $$488 = $25;
    }
   } else {
    $27 = ($22&255)<(194);
    if ($27) {
     label = 30;
     break L5;
    }
    $28 = ($22&255)<(224);
    $29 = ((($$185)) + 2|0);
    $30 = $$185;
    $31 = (($7) - ($30))|0;
    if ($28) {
     $32 = ($31|0)<(2);
     if ($32) {
      label = 30;
      break L5;
     }
     $33 = HEAP8[$25>>0]|0;
     $34 = $33&255;
     $35 = $34 & 192;
     $36 = ($35|0)==(128);
     if (!($36)) {
      label = 30;
      break L5;
     }
     $37 = $23 << 6;
     $38 = $37 & 1984;
     $39 = $34 & 63;
     $40 = $39 | $38;
     $41 = ($40>>>0)>($3>>>0);
     if ($41) {
      label = 30;
      break L5;
     } else {
      $$488 = $29;
      break;
     }
    }
    $42 = ($22&255)<(240);
    $43 = ((($$185)) + 3|0);
    if ($42) {
     $44 = ($31|0)<(3);
     if ($44) {
      label = 30;
      break L5;
     }
     $45 = HEAP8[$25>>0]|0;
     $46 = HEAP8[$29>>0]|0;
     $47 = $45&255;
     $48 = $47 & 224;
     switch ($22<<24>>24) {
     case -32:  {
      $49 = ($48|0)==(160);
      if (!($49)) {
       $$lcssa112$sink = $30;
       break L5;
      }
      break;
     }
     case -19:  {
      $50 = ($48|0)==(128);
      if (!($50)) {
       $$lcssa112$sink = $30;
       break L5;
      }
      break;
     }
     default: {
      $51 = $47 & 192;
      $52 = ($51|0)==(128);
      if (!($52)) {
       $$lcssa112$sink = $30;
       break L5;
      }
     }
     }
     $53 = $46&255;
     $54 = $53 & 192;
     $55 = ($54|0)==(128);
     if (!($55)) {
      label = 30;
      break L5;
     }
     $56 = $23 << 12;
     $57 = $56 & 61440;
     $58 = $47 << 6;
     $59 = $58 & 4032;
     $60 = $59 | $57;
     $61 = $53 & 63;
     $62 = $60 | $61;
     $63 = ($62>>>0)>($3>>>0);
     if ($63) {
      label = 30;
      break L5;
     } else {
      $$488 = $43;
      break;
     }
    }
    $64 = ($22&255)>(244);
    $65 = ($31|0)<(4);
    $or$cond100 = $65 | $64;
    if ($or$cond100) {
     label = 30;
     break L5;
    }
    $66 = HEAP8[$25>>0]|0;
    $67 = HEAP8[$29>>0]|0;
    $68 = HEAP8[$43>>0]|0;
    $69 = $66&255;
    switch ($22<<24>>24) {
    case -16:  {
     $$off = (($66) + 112)<<24>>24;
     $70 = ($$off&255)<(48);
     if (!($70)) {
      $$lcssa112$sink = $30;
      break L5;
     }
     break;
    }
    case -12:  {
     $71 = $69 & 240;
     $72 = ($71|0)==(128);
     if (!($72)) {
      $$lcssa112$sink = $30;
      break L5;
     }
     break;
    }
    default: {
     $73 = $69 & 192;
     $74 = ($73|0)==(128);
     if (!($74)) {
      $$lcssa112$sink = $30;
      break L5;
     }
    }
    }
    $75 = $67&255;
    $76 = $75 & 192;
    $77 = ($76|0)==(128);
    if (!($77)) {
     label = 30;
     break L5;
    }
    $78 = $68&255;
    $79 = $78 & 192;
    $80 = ($79|0)==(128);
    if (!($80)) {
     label = 30;
     break L5;
    }
    $81 = $23 << 18;
    $82 = $81 & 1835008;
    $83 = $69 << 12;
    $84 = $83 & 258048;
    $85 = $84 | $82;
    $86 = $75 << 6;
    $87 = $86 & 4032;
    $88 = $85 | $87;
    $89 = $78 & 63;
    $90 = $88 | $89;
    $91 = ($90>>>0)>($3>>>0);
    $92 = ((($$185)) + 4|0);
    if ($91) {
     label = 30;
     break L5;
    } else {
     $$488 = $92;
    }
   }
  } while(0);
  $93 = (($$091) + 1)|0;
  $$091 = $93;$$185 = $$488;
 }
 if ((label|0) == 30) {
  $94 = $$185;
  $$lcssa112$sink = $94;
 }
 $95 = (($$lcssa112$sink) - ($8))|0;
 return ($95|0);
}
function __ZNSt3__2L12utf8_to_ucs4EPKhS1_RS1_PjS3_RS3_mNS_12codecvt_modeE($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$9 = 0, $$off = 0, $$sink = 0, $10 = 0, $100 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0;
 var $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 HEAP32[$5>>2] = $3;
 $8 = $7 & 4;
 $9 = ($8|0)==(0);
 $10 = $1;
 if (!($9)) {
  $11 = HEAP32[$2>>2]|0;
  $12 = $11;
  $13 = (($10) - ($12))|0;
  $14 = ($13|0)>(2);
  if ($14) {
   $15 = HEAP8[$11>>0]|0;
   $16 = ($15<<24>>24)==(-17);
   if ($16) {
    $17 = ((($11)) + 1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = ($18<<24>>24)==(-69);
    if ($19) {
     $20 = ((($11)) + 2|0);
     $21 = HEAP8[$20>>0]|0;
     $22 = ($21<<24>>24)==(-65);
     if ($22) {
      $23 = ((($11)) + 3|0);
      HEAP32[$2>>2] = $23;
     }
    }
   }
  }
 }
 L7: while(1) {
  $24 = HEAP32[$2>>2]|0;
  $25 = ($24>>>0)<($1>>>0);
  if (!($25)) {
   $$9 = 0;
   break;
  }
  $26 = HEAP32[$5>>2]|0;
  $27 = ($26>>>0)<($4>>>0);
  if (!($27)) {
   $$9 = 1;
   break;
  }
  $28 = HEAP8[$24>>0]|0;
  $29 = $28&255;
  $30 = ($28<<24>>24)>(-1);
  $31 = ((($24)) + 1|0);
  do {
   if ($30) {
    $32 = ($29>>>0)>($6>>>0);
    if ($32) {
     $$9 = 2;
     break L7;
    }
    HEAP32[$26>>2] = $29;
    $$sink = $31;
   } else {
    $33 = ($28&255)<(194);
    if ($33) {
     $$9 = 2;
     break L7;
    }
    $34 = ($28&255)<(224);
    $35 = $24;
    $36 = ((($24)) + 2|0);
    $37 = (($10) - ($35))|0;
    if ($34) {
     $38 = ($37|0)<(2);
     if ($38) {
      $$9 = 1;
      break L7;
     }
     $39 = HEAP8[$31>>0]|0;
     $40 = $39&255;
     $41 = $40 & 192;
     $42 = ($41|0)==(128);
     if (!($42)) {
      $$9 = 2;
      break L7;
     }
     $43 = $29 << 6;
     $44 = $43 & 1984;
     $45 = $40 & 63;
     $46 = $45 | $44;
     $47 = ($46>>>0)>($6>>>0);
     if ($47) {
      $$9 = 2;
      break L7;
     }
     HEAP32[$26>>2] = $46;
     $$sink = $36;
     break;
    }
    $48 = ($28&255)<(240);
    $49 = ((($24)) + 3|0);
    if ($48) {
     $50 = ($37|0)<(3);
     if ($50) {
      $$9 = 1;
      break L7;
     }
     $51 = HEAP8[$31>>0]|0;
     $52 = HEAP8[$36>>0]|0;
     $53 = $51&255;
     $54 = $53 & 224;
     switch ($28<<24>>24) {
     case -32:  {
      $55 = ($54|0)==(160);
      if (!($55)) {
       $$9 = 2;
       break L7;
      }
      break;
     }
     case -19:  {
      $56 = ($54|0)==(128);
      if (!($56)) {
       $$9 = 2;
       break L7;
      }
      break;
     }
     default: {
      $57 = $53 & 192;
      $58 = ($57|0)==(128);
      if (!($58)) {
       $$9 = 2;
       break L7;
      }
     }
     }
     $59 = $52&255;
     $60 = $59 & 192;
     $61 = ($60|0)==(128);
     if (!($61)) {
      $$9 = 2;
      break L7;
     }
     $62 = $29 << 12;
     $63 = $62 & 61440;
     $64 = $53 << 6;
     $65 = $64 & 4032;
     $66 = $65 | $63;
     $67 = $59 & 63;
     $68 = $66 | $67;
     $69 = ($68>>>0)>($6>>>0);
     if ($69) {
      $$9 = 2;
      break L7;
     }
     HEAP32[$26>>2] = $68;
     $$sink = $49;
     break;
    }
    $70 = ($28&255)<(245);
    if (!($70)) {
     $$9 = 2;
     break L7;
    }
    $71 = ($37|0)<(4);
    if ($71) {
     $$9 = 1;
     break L7;
    }
    $72 = HEAP8[$31>>0]|0;
    $73 = HEAP8[$36>>0]|0;
    $74 = HEAP8[$49>>0]|0;
    $75 = $72&255;
    switch ($28<<24>>24) {
    case -16:  {
     $$off = (($72) + 112)<<24>>24;
     $76 = ($$off&255)<(48);
     if (!($76)) {
      $$9 = 2;
      break L7;
     }
     break;
    }
    case -12:  {
     $77 = $75 & 240;
     $78 = ($77|0)==(128);
     if (!($78)) {
      $$9 = 2;
      break L7;
     }
     break;
    }
    default: {
     $79 = $75 & 192;
     $80 = ($79|0)==(128);
     if (!($80)) {
      $$9 = 2;
      break L7;
     }
    }
    }
    $81 = $73&255;
    $82 = $81 & 192;
    $83 = ($82|0)==(128);
    if (!($83)) {
     $$9 = 2;
     break L7;
    }
    $84 = $74&255;
    $85 = $84 & 192;
    $86 = ($85|0)==(128);
    if (!($86)) {
     $$9 = 2;
     break L7;
    }
    $87 = $29 << 18;
    $88 = $87 & 1835008;
    $89 = $75 << 12;
    $90 = $89 & 258048;
    $91 = $90 | $88;
    $92 = $81 << 6;
    $93 = $92 & 4032;
    $94 = $91 | $93;
    $95 = $84 & 63;
    $96 = $94 | $95;
    $97 = ($96>>>0)>($6>>>0);
    if ($97) {
     $$9 = 2;
     break L7;
    }
    HEAP32[$26>>2] = $96;
    $98 = ((($24)) + 4|0);
    $$sink = $98;
   }
  } while(0);
  HEAP32[$2>>2] = $$sink;
  $99 = HEAP32[$5>>2]|0;
  $100 = ((($99)) + 4|0);
  HEAP32[$5>>2] = $100;
 }
 return ($$9|0);
}
function __ZNSt3__2L12ucs4_to_utf8EPKjS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$2 = 0, $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 HEAP32[$5>>2] = $3;
 $8 = $7 & 2;
 $9 = ($8|0)==(0);
 $10 = $4;
 if ($9) {
  label = 4;
 } else {
  $11 = $3;
  $12 = (($10) - ($11))|0;
  $13 = ($12|0)<(3);
  if ($13) {
   $$2 = 1;
  } else {
   $14 = ((($3)) + 1|0);
   HEAP32[$5>>2] = $14;
   HEAP8[$3>>0] = -17;
   $15 = HEAP32[$5>>2]|0;
   $16 = ((($15)) + 1|0);
   HEAP32[$5>>2] = $16;
   HEAP8[$15>>0] = -69;
   $17 = HEAP32[$5>>2]|0;
   $18 = ((($17)) + 1|0);
   HEAP32[$5>>2] = $18;
   HEAP8[$17>>0] = -65;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $$pre = HEAP32[$2>>2]|0;
   $20 = $$pre;
   while(1) {
    $19 = ($20>>>0)<($1>>>0);
    if (!($19)) {
     $$2 = 0;
     break L4;
    }
    $21 = HEAP32[$20>>2]|0;
    $22 = $21 & -2048;
    $23 = ($22|0)==(55296);
    $24 = ($21>>>0)>($6>>>0);
    $or$cond = $24 | $23;
    if ($or$cond) {
     $$2 = 2;
     break L4;
    }
    $25 = ($21>>>0)<(128);
    do {
     if ($25) {
      $26 = HEAP32[$5>>2]|0;
      $27 = $26;
      $28 = (($10) - ($27))|0;
      $29 = ($28|0)<(1);
      if ($29) {
       $$2 = 1;
       break L4;
      }
      $30 = $21&255;
      $31 = ((($26)) + 1|0);
      HEAP32[$5>>2] = $31;
      HEAP8[$26>>0] = $30;
     } else {
      $32 = ($21>>>0)<(2048);
      $33 = $21 & 63;
      $34 = $33 | 128;
      $35 = $34&255;
      if ($32) {
       $36 = HEAP32[$5>>2]|0;
       $37 = $36;
       $38 = (($10) - ($37))|0;
       $39 = ($38|0)<(2);
       if ($39) {
        $$2 = 1;
        break L4;
       }
       $40 = $21 >>> 6;
       $41 = $40 | 192;
       $42 = $41&255;
       $43 = ((($36)) + 1|0);
       HEAP32[$5>>2] = $43;
       HEAP8[$36>>0] = $42;
       $44 = HEAP32[$5>>2]|0;
       $45 = ((($44)) + 1|0);
       HEAP32[$5>>2] = $45;
       HEAP8[$44>>0] = $35;
       break;
      }
      $46 = ($21>>>0)<(65536);
      $47 = HEAP32[$5>>2]|0;
      $48 = $47;
      $49 = (($10) - ($48))|0;
      $50 = $21 >>> 6;
      $51 = $50 & 63;
      $52 = $51 | 128;
      $53 = ((($47)) + 1|0);
      $54 = $52&255;
      if ($46) {
       $55 = ($49|0)<(3);
       if ($55) {
        $$2 = 1;
        break L4;
       }
       $56 = $21 >>> 12;
       $57 = $56 | 224;
       $58 = $57&255;
       HEAP32[$5>>2] = $53;
       HEAP8[$47>>0] = $58;
       $59 = HEAP32[$5>>2]|0;
       $60 = ((($59)) + 1|0);
       HEAP32[$5>>2] = $60;
       HEAP8[$59>>0] = $54;
       $61 = HEAP32[$5>>2]|0;
       $62 = ((($61)) + 1|0);
       HEAP32[$5>>2] = $62;
       HEAP8[$61>>0] = $35;
       break;
      } else {
       $63 = ($49|0)<(4);
       if ($63) {
        $$2 = 1;
        break L4;
       }
       $64 = $21 >>> 18;
       $65 = $64 | 240;
       $66 = $65&255;
       HEAP32[$5>>2] = $53;
       HEAP8[$47>>0] = $66;
       $67 = $21 >>> 12;
       $68 = $67 & 63;
       $69 = $68 | 128;
       $70 = $69&255;
       $71 = HEAP32[$5>>2]|0;
       $72 = ((($71)) + 1|0);
       HEAP32[$5>>2] = $72;
       HEAP8[$71>>0] = $70;
       $73 = HEAP32[$5>>2]|0;
       $74 = ((($73)) + 1|0);
       HEAP32[$5>>2] = $74;
       HEAP8[$73>>0] = $54;
       $75 = HEAP32[$5>>2]|0;
       $76 = ((($75)) + 1|0);
       HEAP32[$5>>2] = $76;
       HEAP8[$75>>0] = $35;
       break;
      }
     }
    } while(0);
    $77 = HEAP32[$2>>2]|0;
    $78 = ((($77)) + 4|0);
    HEAP32[$2>>2] = $78;
    $20 = $78;
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZNSt3__216__narrow_to_utf8ILj32EED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE6do_outERS1_PKcS5_RS5_PcS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$4>>2] = $2;
 HEAP32[$7>>2] = $5;
 return 3;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE5do_inERS1_PKcS5_RS5_PcS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$4>>2] = $2;
 HEAP32[$7>>2] = $5;
 return 3;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$4>>2] = $2;
 return 3;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE11do_encodingEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE16do_always_noconvEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE9do_lengthERS1_PKcS5_j($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$sroa$speculated = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = $3;
 $6 = $2;
 $7 = (($5) - ($6))|0;
 $8 = ($7>>>0)<($4>>>0);
 $$sroa$speculated = $8 ? $7 : $4;
 return ($$sroa$speculated|0);
}
function __ZNKSt3__27codecvtIcc11__mbstate_tE13do_max_lengthEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE6do_outERS1_PKwS5_RS5_PcS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$0 = 0, $$069 = 0, $$070 = 0, $$070$lcssa = 0, $$077 = 0, $$079 = 0, $$171 = 0, $$173 = 0, $$178 = 0, $$2 = 0, $$2$lcssa = 0, $$375$ph = 0, $$476 = 0, $$5$ph87 = 0, $$pre = 0, $$pre170 = 0, $$pre171 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0;
 var $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0;
 var $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $not$ = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp;
 $9 = sp + 8|0;
 $$070 = $2;
 while(1) {
  $10 = ($$070|0)==($3|0);
  if ($10) {
   $$070$lcssa = $3;
   break;
  }
  $11 = HEAP32[$$070>>2]|0;
  $12 = ($11|0)==(0);
  if ($12) {
   $$070$lcssa = $$070;
   break;
  }
  $13 = ((($$070)) + 4|0);
  $$070 = $13;
 }
 HEAP32[$7>>2] = $5;
 HEAP32[$4>>2] = $2;
 $14 = $6;
 $15 = ((($0)) + 8|0);
 $$077 = $2;$$079 = $5;$$171 = $$070$lcssa;
 L6: while(1) {
  $16 = ($$077|0)==($3|0);
  $17 = ($$079|0)==($6|0);
  $or$cond = $17 | $16;
  if ($or$cond) {
   $113 = $$077;
   label = 53;
   break;
  }
  $18 = $1;
  $19 = $18;
  $20 = HEAP32[$19>>2]|0;
  $21 = (($18) + 4)|0;
  $22 = $21;
  $23 = HEAP32[$22>>2]|0;
  $24 = $8;
  $25 = $24;
  HEAP32[$25>>2] = $20;
  $26 = (($24) + 4)|0;
  $27 = $26;
  HEAP32[$27>>2] = $23;
  $28 = $$171;
  $29 = $$077;
  $30 = (($28) - ($29))|0;
  $31 = $30 >> 2;
  $32 = $$079;
  $33 = (($14) - ($32))|0;
  $34 = HEAP32[$15>>2]|0;
  $35 = (___uselocale($34)|0);
  __THREW__ = 0;
  $36 = (invoke_iiiiii(1025,($$079|0),($4|0),($31|0),($33|0),($1|0))|0);
  $37 = __THREW__; __THREW__ = 0;
  $38 = $37&1;
  if ($38) {
   label = 11;
   break;
  }
  $39 = ($35|0)==(0|0);
  if (!($39)) {
   __THREW__ = 0;
   (invoke_ii(958,($35|0))|0);
   $40 = __THREW__; __THREW__ = 0;
   $41 = $40&1;
   if ($41) {
    label = 10;
    break;
   }
  }
  switch ($36|0) {
  case -1:  {
   label = 16;
   break L6;
   break;
  }
  case 0:  {
   $$375$ph = 1;
   label = 50;
   break L6;
   break;
  }
  default: {
  }
  }
  $76 = HEAP32[$7>>2]|0;
  $77 = (($76) + ($36)|0);
  HEAP32[$7>>2] = $77;
  $78 = ($77|0)==($6|0);
  if ($78) {
   label = 51;
   break;
  }
  $79 = ($$171|0)==($3|0);
  if ($79) {
   $$pre = HEAP32[$4>>2]|0;
   $$5$ph87 = $3;$115 = $77;$116 = $$pre;
  } else {
   $80 = HEAP32[$15>>2]|0;
   $81 = (___uselocale($80)|0);
   __THREW__ = 0;
   $82 = (invoke_iiii(1026,($9|0),0,($1|0))|0);
   $83 = __THREW__; __THREW__ = 0;
   $84 = $83&1;
   if ($84) {
    label = 36;
    break;
   }
   $85 = ($81|0)==(0|0);
   if (!($85)) {
    __THREW__ = 0;
    (invoke_ii(958,($81|0))|0);
    $86 = __THREW__; __THREW__ = 0;
    $87 = $86&1;
    if ($87) {
     label = 35;
     break;
    }
   }
   $97 = ($82|0)==(-1);
   if ($97) {
    $$173 = 2;
    label = 49;
    break;
   }
   $98 = HEAP32[$7>>2]|0;
   $99 = (($14) - ($98))|0;
   $100 = ($82>>>0)>($99>>>0);
   if ($100) {
    $$173 = 1;
    label = 49;
    break;
   } else {
    $$0 = $9;$$069 = $82;
   }
   while(1) {
    $101 = ($$069|0)==(0);
    if ($101) {
     break;
    }
    $104 = ((($$0)) + 1|0);
    $105 = HEAP8[$$0>>0]|0;
    $106 = HEAP32[$7>>2]|0;
    $107 = ((($106)) + 1|0);
    HEAP32[$7>>2] = $107;
    HEAP8[$106>>0] = $105;
    $108 = (($$069) + -1)|0;
    $$0 = $104;$$069 = $108;
   }
   $102 = HEAP32[$4>>2]|0;
   $103 = ((($102)) + 4|0);
   HEAP32[$4>>2] = $103;
   $$2 = $103;
   while(1) {
    $109 = ($$2|0)==($3|0);
    if ($109) {
     $$2$lcssa = $3;
     break;
    }
    $110 = HEAP32[$$2>>2]|0;
    $111 = ($110|0)==(0);
    if ($111) {
     $$2$lcssa = $$2;
     break;
    }
    $112 = ((($$2)) + 4|0);
    $$2 = $112;
   }
   $$pre170 = HEAP32[$7>>2]|0;
   $$5$ph87 = $$2$lcssa;$115 = $$pre170;$116 = $103;
  }
  $$077 = $116;$$079 = $115;$$171 = $$5$ph87;
 }
 do {
  if ((label|0) == 10) {
   $42 = ___cxa_find_matching_catch_3(0|0)|0;
   $43 = tempRet0;
   ___clang_call_terminate($42);
   // unreachable;
  }
  else if ((label|0) == 11) {
   $44 = ___cxa_find_matching_catch_2()|0;
   $45 = tempRet0;
   $46 = ($35|0)==(0|0);
   if (!($46)) {
    __THREW__ = 0;
    (invoke_ii(958,($35|0))|0);
    $47 = __THREW__; __THREW__ = 0;
    $48 = $47&1;
    if ($48) {
     $49 = ___cxa_find_matching_catch_3(0|0)|0;
     $50 = tempRet0;
     ___clang_call_terminate($49);
     // unreachable;
    }
   }
   ___resumeException($44|0);
   // unreachable;
  }
  else if ((label|0) == 16) {
   HEAP32[$7>>2] = $$079;
   $$178 = $$077;$56 = $$079;
   while(1) {
    $51 = HEAP32[$4>>2]|0;
    $52 = ($$178|0)==($51|0);
    if ($52) {
     label = 28;
     break;
    }
    $53 = HEAP32[$$178>>2]|0;
    $54 = HEAP32[$15>>2]|0;
    $55 = (___uselocale($54)|0);
    __THREW__ = 0;
    $57 = (invoke_iiii(1026,($56|0),($53|0),($8|0))|0);
    $58 = __THREW__; __THREW__ = 0;
    $59 = $58&1;
    if ($59) {
     label = 22;
     break;
    }
    $60 = ($55|0)==(0|0);
    if (!($60)) {
     __THREW__ = 0;
     (invoke_ii(958,($55|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 21;
      break;
     }
    }
    $72 = ($57|0)==(-1);
    if ($72) {
     label = 28;
     break;
    }
    $73 = HEAP32[$7>>2]|0;
    $74 = (($73) + ($57)|0);
    HEAP32[$7>>2] = $74;
    $75 = ((($$178)) + 4|0);
    $$178 = $75;$56 = $74;
   }
   if ((label|0) == 21) {
    $63 = ___cxa_find_matching_catch_3(0|0)|0;
    $64 = tempRet0;
    ___clang_call_terminate($63);
    // unreachable;
   }
   else if ((label|0) == 22) {
    $65 = ___cxa_find_matching_catch_2()|0;
    $66 = tempRet0;
    $67 = ($55|0)==(0|0);
    if (!($67)) {
     __THREW__ = 0;
     (invoke_ii(958,($55|0))|0);
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      $70 = ___cxa_find_matching_catch_3(0|0)|0;
      $71 = tempRet0;
      ___clang_call_terminate($70);
      // unreachable;
     }
    }
    ___resumeException($65|0);
    // unreachable;
   }
   else if ((label|0) == 28) {
    HEAP32[$4>>2] = $$178;
    $$375$ph = 2;
    label = 50;
    break;
   }
  }
  else if ((label|0) == 35) {
   $88 = ___cxa_find_matching_catch_3(0|0)|0;
   $89 = tempRet0;
   ___clang_call_terminate($88);
   // unreachable;
  }
  else if ((label|0) == 36) {
   $90 = ___cxa_find_matching_catch_2()|0;
   $91 = tempRet0;
   $92 = ($81|0)==(0|0);
   if (!($92)) {
    __THREW__ = 0;
    (invoke_ii(958,($81|0))|0);
    $93 = __THREW__; __THREW__ = 0;
    $94 = $93&1;
    if ($94) {
     $95 = ___cxa_find_matching_catch_3(0|0)|0;
     $96 = tempRet0;
     ___clang_call_terminate($95);
     // unreachable;
    }
   }
   ___resumeException($90|0);
   // unreachable;
  }
  else if ((label|0) == 49) {
   $$375$ph = $$173;
   label = 50;
  }
  else if ((label|0) == 51) {
   $$pre171 = HEAP32[$4>>2]|0;
   $113 = $$pre171;
   label = 53;
  }
 } while(0);
 if ((label|0) == 50) {
  $$476 = $$375$ph;
 }
 else if ((label|0) == 53) {
  $not$ = ($113|0)!=($3|0);
  $114 = $not$&1;
  $$476 = $114;
 }
 STACKTOP = sp;return ($$476|0);
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE5do_inERS1_PKcS5_RS5_PwS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$070 = 0, $$070$lcssa = 0, $$074 = 0, $$077 = 0, $$1 = 0, $$172$ph = 0, $$175 = 0, $$2 = 0, $$2$lcssa = 0, $$273 = 0, $$4$ph = 0, $$pre = 0, $$pre176 = 0, $$pre177 = 0, $$sink = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0;
 var $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0;
 var $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $not$ = 0, $not$78 = 0, $or$cond = 0;
 var $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp;
 $$070 = $2;
 while(1) {
  $9 = ($$070|0)==($3|0);
  if ($9) {
   $$070$lcssa = $3;
   break;
  }
  $10 = HEAP8[$$070>>0]|0;
  $11 = ($10<<24>>24)==(0);
  if ($11) {
   $$070$lcssa = $$070;
   break;
  }
  $12 = ((($$070)) + 1|0);
  $$070 = $12;
 }
 HEAP32[$7>>2] = $5;
 HEAP32[$4>>2] = $2;
 $13 = $6;
 $14 = ((($0)) + 8|0);
 $$074 = $2;$$077 = $5;$$1 = $$070$lcssa;
 while(1) {
  $15 = ($$074|0)==($3|0);
  $16 = ($$077|0)==($6|0);
  $or$cond = $16 | $15;
  if ($or$cond) {
   $106 = $$074;
   label = 50;
   break;
  }
  $17 = $1;
  $18 = $17;
  $19 = HEAP32[$18>>2]|0;
  $20 = (($17) + 4)|0;
  $21 = $20;
  $22 = HEAP32[$21>>2]|0;
  $23 = $8;
  $24 = $23;
  HEAP32[$24>>2] = $19;
  $25 = (($23) + 4)|0;
  $26 = $25;
  HEAP32[$26>>2] = $22;
  $27 = $$1;
  $28 = $$074;
  $29 = (($27) - ($28))|0;
  $30 = $$077;
  $31 = (($13) - ($30))|0;
  $32 = $31 >> 2;
  $33 = HEAP32[$14>>2]|0;
  $34 = (___uselocale($33)|0);
  __THREW__ = 0;
  $35 = (invoke_iiiiii(1027,($$077|0),($4|0),($29|0),($32|0),($1|0))|0);
  $36 = __THREW__; __THREW__ = 0;
  $37 = $36&1;
  if ($37) {
   label = 11;
   break;
  }
  $38 = ($34|0)==(0|0);
  if (!($38)) {
   __THREW__ = 0;
   (invoke_ii(958,($34|0))|0);
   $39 = __THREW__; __THREW__ = 0;
   $40 = $39&1;
   if ($40) {
    label = 10;
    break;
   }
  }
  $50 = ($35|0)==(-1);
  if ($50) {
   $$175 = $$074;$storemerge = $$077;
   label = 16;
   break;
  }
  $76 = HEAP32[$7>>2]|0;
  $77 = (($76) + ($35<<2)|0);
  HEAP32[$7>>2] = $77;
  $78 = ($77|0)==($6|0);
  if ($78) {
   label = 47;
   break;
  }
  $79 = ($$1|0)==($3|0);
  $$pre = HEAP32[$4>>2]|0;
  if ($79) {
   $$4$ph = $3;$108 = $77;$109 = $$pre;
  } else {
   $80 = HEAP32[$14>>2]|0;
   $81 = (___uselocale($80)|0);
   __THREW__ = 0;
   $82 = (invoke_iiiii(1028,($77|0),($$pre|0),1,($1|0))|0);
   $83 = __THREW__; __THREW__ = 0;
   $84 = $83&1;
   if ($84) {
    label = 37;
    break;
   }
   $85 = ($81|0)==(0|0);
   if (!($85)) {
    __THREW__ = 0;
    (invoke_ii(958,($81|0))|0);
    $86 = __THREW__; __THREW__ = 0;
    $87 = $86&1;
    if ($87) {
     label = 36;
     break;
    }
   }
   $97 = ($82|0)==(0);
   if (!($97)) {
    $$172$ph = 2;
    label = 46;
    break;
   }
   $98 = HEAP32[$7>>2]|0;
   $99 = ((($98)) + 4|0);
   HEAP32[$7>>2] = $99;
   $100 = HEAP32[$4>>2]|0;
   $101 = ((($100)) + 1|0);
   HEAP32[$4>>2] = $101;
   $$2 = $101;
   while(1) {
    $102 = ($$2|0)==($3|0);
    if ($102) {
     $$2$lcssa = $3;
     break;
    }
    $103 = HEAP8[$$2>>0]|0;
    $104 = ($103<<24>>24)==(0);
    if ($104) {
     $$2$lcssa = $$2;
     break;
    }
    $105 = ((($$2)) + 1|0);
    $$2 = $105;
   }
   $$pre176 = HEAP32[$7>>2]|0;
   $$4$ph = $$2$lcssa;$108 = $$pre176;$109 = $101;
  }
  $$074 = $109;$$077 = $108;$$1 = $$4$ph;
 }
 do {
  if ((label|0) == 10) {
   $41 = ___cxa_find_matching_catch_3(0|0)|0;
   $42 = tempRet0;
   ___clang_call_terminate($41);
   // unreachable;
  }
  else if ((label|0) == 11) {
   $43 = ___cxa_find_matching_catch_2()|0;
   $44 = tempRet0;
   $45 = ($34|0)==(0|0);
   if (!($45)) {
    __THREW__ = 0;
    (invoke_ii(958,($34|0))|0);
    $46 = __THREW__; __THREW__ = 0;
    $47 = $46&1;
    if ($47) {
     $48 = ___cxa_find_matching_catch_3(0|0)|0;
     $49 = tempRet0;
     ___clang_call_terminate($48);
     // unreachable;
    }
   }
   ___resumeException($43|0);
   // unreachable;
  }
  else if ((label|0) == 16) {
   L30: while(1) {
    label = 0;
    HEAP32[$7>>2] = $storemerge;
    $51 = HEAP32[$4>>2]|0;
    $52 = ($$175|0)==($51|0);
    if ($52) {
     label = 30;
     break;
    }
    $53 = $$175;
    $54 = (($27) - ($53))|0;
    $55 = HEAP32[$14>>2]|0;
    $56 = (___uselocale($55)|0);
    __THREW__ = 0;
    $57 = (invoke_iiiii(1028,($storemerge|0),($$175|0),($54|0),($8|0))|0);
    $58 = __THREW__; __THREW__ = 0;
    $59 = $58&1;
    if ($59) {
     label = 21;
     break;
    }
    $60 = ($56|0)==(0|0);
    if (!($60)) {
     __THREW__ = 0;
     (invoke_ii(958,($56|0))|0);
     $61 = __THREW__; __THREW__ = 0;
     $62 = $61&1;
     if ($62) {
      label = 20;
      break;
     }
    }
    switch ($57|0) {
    case -1:  {
     label = 27;
     break L30;
     break;
    }
    case -2:  {
     label = 28;
     break L30;
     break;
    }
    case 0:  {
     $$sink = 1;
     break;
    }
    default: {
     $$sink = $57;
    }
    }
    $72 = (($$175) + ($$sink)|0);
    $73 = HEAP32[$7>>2]|0;
    $74 = ((($73)) + 4|0);
    $$175 = $72;$storemerge = $74;
    label = 16;
   }
   if ((label|0) == 20) {
    $63 = ___cxa_find_matching_catch_3(0|0)|0;
    $64 = tempRet0;
    ___clang_call_terminate($63);
    // unreachable;
   }
   else if ((label|0) == 21) {
    $65 = ___cxa_find_matching_catch_2()|0;
    $66 = tempRet0;
    $67 = ($56|0)==(0|0);
    if (!($67)) {
     __THREW__ = 0;
     (invoke_ii(958,($56|0))|0);
     $68 = __THREW__; __THREW__ = 0;
     $69 = $68&1;
     if ($69) {
      $70 = ___cxa_find_matching_catch_3(0|0)|0;
      $71 = tempRet0;
      ___clang_call_terminate($70);
      // unreachable;
     }
    }
    ___resumeException($65|0);
    // unreachable;
   }
   else if ((label|0) == 27) {
    HEAP32[$4>>2] = $$175;
    $$172$ph = 2;
    label = 46;
    break;
   }
   else if ((label|0) == 28) {
    HEAP32[$4>>2] = $$175;
    $$172$ph = 1;
    label = 46;
    break;
   }
   else if ((label|0) == 30) {
    HEAP32[$4>>2] = $$175;
    $not$78 = ($$175|0)!=($3|0);
    $75 = $not$78&1;
    $$172$ph = $75;
    label = 46;
    break;
   }
  }
  else if ((label|0) == 36) {
   $88 = ___cxa_find_matching_catch_3(0|0)|0;
   $89 = tempRet0;
   ___clang_call_terminate($88);
   // unreachable;
  }
  else if ((label|0) == 37) {
   $90 = ___cxa_find_matching_catch_2()|0;
   $91 = tempRet0;
   $92 = ($81|0)==(0|0);
   if (!($92)) {
    __THREW__ = 0;
    (invoke_ii(958,($81|0))|0);
    $93 = __THREW__; __THREW__ = 0;
    $94 = $93&1;
    if ($94) {
     $95 = ___cxa_find_matching_catch_3(0|0)|0;
     $96 = tempRet0;
     ___clang_call_terminate($95);
     // unreachable;
    }
   }
   ___resumeException($90|0);
   // unreachable;
  }
  else if ((label|0) == 47) {
   $$pre177 = HEAP32[$4>>2]|0;
   $106 = $$pre177;
   label = 50;
  }
 } while(0);
 if ((label|0) == 46) {
  $$273 = $$172$ph;
 }
 else if ((label|0) == 50) {
  $not$ = ($106|0)!=($3|0);
  $107 = $not$&1;
  $$273 = $107;
 }
 STACKTOP = sp;return ($$273|0);
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$018 = 0, $$019 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $5 = sp;
 HEAP32[$4>>2] = $2;
 $6 = ((($0)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (___uselocale($7)|0);
 __THREW__ = 0;
 $9 = (invoke_iiii(1026,($5|0),0,($1|0))|0);
 $10 = __THREW__; __THREW__ = 0;
 $11 = $10&1;
 if ($11) {
  $17 = ___cxa_find_matching_catch_2()|0;
  $18 = tempRet0;
  $19 = ($8|0)==(0|0);
  if (!($19)) {
   __THREW__ = 0;
   (invoke_ii(958,($8|0))|0);
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $22 = ___cxa_find_matching_catch_3(0|0)|0;
    $23 = tempRet0;
    ___clang_call_terminate($22);
    // unreachable;
   }
  }
  ___resumeException($17|0);
  // unreachable;
 }
 $12 = ($8|0)==(0|0);
 if (!($12)) {
  __THREW__ = 0;
  (invoke_ii(958,($8|0))|0);
  $13 = __THREW__; __THREW__ = 0;
  $14 = $13&1;
  if ($14) {
   $15 = ___cxa_find_matching_catch_3(0|0)|0;
   $16 = tempRet0;
   ___clang_call_terminate($15);
   // unreachable;
  }
 }
 L12: do {
  switch ($9|0) {
  case 0: case -1:  {
   $$019 = 2;
   break;
  }
  default: {
   $24 = (($9) + -1)|0;
   $25 = HEAP32[$4>>2]|0;
   $26 = $3;
   $27 = (($26) - ($25))|0;
   $28 = ($24>>>0)>($27>>>0);
   if ($28) {
    $$019 = 1;
   } else {
    $$0 = $5;$$018 = $24;
    while(1) {
     $29 = ($$018|0)==(0);
     if ($29) {
      $$019 = 0;
      break L12;
     }
     $30 = ((($$0)) + 1|0);
     $31 = HEAP8[$$0>>0]|0;
     $32 = HEAP32[$4>>2]|0;
     $33 = ((($32)) + 1|0);
     HEAP32[$4>>2] = $33;
     HEAP8[$32>>0] = $31;
     $34 = (($$018) + -1)|0;
     $$0 = $30;$$018 = $34;
    }
   }
  }
  }
 } while(0);
 STACKTOP = sp;return ($$019|0);
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE11do_encodingEv($0) {
 $0 = $0|0;
 var $$ = 0, $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __THREW__ = 0;
 $3 = (invoke_ii(958,($2|0))|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if ($5) {
   label = 21;
  } else {
   __THREW__ = 0;
   $6 = (invoke_iiii(1029,(0|0),(0|0),4)|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if ($8) {
    $14 = ___cxa_find_matching_catch_3(0|0)|0;
    $15 = tempRet0;
    $16 = ($3|0)==(0|0);
    if ($16) {
     $eh$lpad$body$indexZ2D = $14;
     break;
    }
    __THREW__ = 0;
    (invoke_ii(958,($3|0))|0);
    $17 = __THREW__; __THREW__ = 0;
    $18 = $17&1;
    if (!($18)) {
     $eh$lpad$body$indexZ2D = $14;
     break;
    }
    $19 = ___cxa_find_matching_catch_3(0|0)|0;
    $20 = tempRet0;
    ___clang_call_terminate($19);
    // unreachable;
   }
   $9 = ($3|0)==(0|0);
   if (!($9)) {
    __THREW__ = 0;
    (invoke_ii(958,($3|0))|0);
    $10 = __THREW__; __THREW__ = 0;
    $11 = $10&1;
    if ($11) {
     $12 = ___cxa_find_matching_catch_3(0|0)|0;
     $13 = tempRet0;
     ___clang_call_terminate($12);
     // unreachable;
    }
   }
   $21 = ($6|0)==(0);
   if ($21) {
    $22 = HEAP32[$1>>2]|0;
    $23 = ($22|0)==(0|0);
    if ($23) {
     $$0 = 1;
    } else {
     __THREW__ = 0;
     $24 = (invoke_ii(958,($22|0))|0);
     $25 = __THREW__; __THREW__ = 0;
     $26 = $25&1;
     if ($26) {
      label = 21;
      break;
     }
     __THREW__ = 0;
     $27 = (invoke_i(1030)|0);
     $28 = __THREW__; __THREW__ = 0;
     $29 = $28&1;
     if ($29) {
      $35 = ___cxa_find_matching_catch_3(0|0)|0;
      $36 = tempRet0;
      $37 = ($24|0)==(0|0);
      if ($37) {
       $eh$lpad$body$indexZ2D = $35;
       break;
      }
      __THREW__ = 0;
      (invoke_ii(958,($24|0))|0);
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      if (!($39)) {
       $eh$lpad$body$indexZ2D = $35;
       break;
      }
      $40 = ___cxa_find_matching_catch_3(0|0)|0;
      $41 = tempRet0;
      ___clang_call_terminate($40);
      // unreachable;
     }
     $30 = ($24|0)==(0|0);
     if (!($30)) {
      __THREW__ = 0;
      (invoke_ii(958,($24|0))|0);
      $31 = __THREW__; __THREW__ = 0;
      $32 = $31&1;
      if ($32) {
       $33 = ___cxa_find_matching_catch_3(0|0)|0;
       $34 = tempRet0;
       ___clang_call_terminate($33);
       // unreachable;
      }
     }
     $42 = ($27|0)==(1);
     $$ = $42&1;
     return ($$|0);
    }
   } else {
    $$0 = -1;
   }
   return ($$0|0);
  }
 } while(0);
 if ((label|0) == 21) {
  $43 = ___cxa_find_matching_catch_3(0|0)|0;
  $44 = tempRet0;
  $eh$lpad$body$indexZ2D = $43;
 }
 ___clang_call_terminate($eh$lpad$body$indexZ2D);
 // unreachable;
 return (0)|0;
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE16do_always_noconvEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE9do_lengthERS1_PKcS5_j($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$029 = 0, $$030 = 0, $$034 = 0, $$131 = 0, $$sink = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = $3;
 $6 = ((($0)) + 8|0);
 $$029 = 0;$$030 = 0;$$034 = $2;
 L1: while(1) {
  $7 = ($$029>>>0)>=($4>>>0);
  $8 = ($$034|0)==($3|0);
  $or$cond = $8 | $7;
  if ($or$cond) {
   label = 14;
   break;
  }
  $9 = $$034;
  $10 = (($5) - ($9))|0;
  $11 = HEAP32[$6>>2]|0;
  $12 = (___uselocale($11)|0);
  __THREW__ = 0;
  $13 = (invoke_iiii(1031,($$034|0),($10|0),($1|0))|0);
  $14 = __THREW__; __THREW__ = 0;
  $15 = $14&1;
  if ($15) {
   label = 7;
   break;
  }
  $16 = ($12|0)==(0|0);
  if (!($16)) {
   __THREW__ = 0;
   (invoke_ii(958,($12|0))|0);
   $17 = __THREW__; __THREW__ = 0;
   $18 = $17&1;
   if ($18) {
    label = 6;
    break;
   }
  }
  switch ($13|0) {
  case -2: case -1:  {
   label = 14;
   break L1;
   break;
  }
  case 0:  {
   $$sink = 1;
   break;
  }
  default: {
   $$sink = $13;
  }
  }
  $28 = (($$034) + ($$sink)|0);
  $$131 = (($$sink) + ($$030))|0;
  $29 = (($$029) + 1)|0;
  $$029 = $29;$$030 = $$131;$$034 = $28;
 }
 if ((label|0) == 6) {
  $19 = ___cxa_find_matching_catch_3(0|0)|0;
  $20 = tempRet0;
  ___clang_call_terminate($19);
  // unreachable;
 }
 else if ((label|0) == 7) {
  $21 = ___cxa_find_matching_catch_2()|0;
  $22 = tempRet0;
  $23 = ($12|0)==(0|0);
  if (!($23)) {
   __THREW__ = 0;
   (invoke_ii(958,($12|0))|0);
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $26 = ___cxa_find_matching_catch_3(0|0)|0;
    $27 = tempRet0;
    ___clang_call_terminate($26);
    // unreachable;
   }
  }
  ___resumeException($21|0);
  // unreachable;
 }
 else if ((label|0) == 14) {
  return ($$030|0);
 }
 return (0)|0;
}
function __ZNKSt3__27codecvtIwc11__mbstate_tE13do_max_lengthEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 L1: do {
  if ($3) {
   $22 = 1;
  } else {
   __THREW__ = 0;
   $4 = (invoke_ii(958,($2|0))|0);
   $5 = __THREW__; __THREW__ = 0;
   $6 = $5&1;
   do {
    if ($6) {
     $23 = ___cxa_find_matching_catch_3(0|0)|0;
     $24 = tempRet0;
     $eh$lpad$body$indexZ2D = $23;
    } else {
     __THREW__ = 0;
     $7 = (invoke_i(1030)|0);
     $8 = __THREW__; __THREW__ = 0;
     $9 = $8&1;
     if ($9) {
      $15 = ___cxa_find_matching_catch_3(0|0)|0;
      $16 = tempRet0;
      $17 = ($4|0)==(0|0);
      if ($17) {
       $eh$lpad$body$indexZ2D = $15;
       break;
      }
      __THREW__ = 0;
      (invoke_ii(958,($4|0))|0);
      $18 = __THREW__; __THREW__ = 0;
      $19 = $18&1;
      if (!($19)) {
       $eh$lpad$body$indexZ2D = $15;
       break;
      }
      $20 = ___cxa_find_matching_catch_3(0|0)|0;
      $21 = tempRet0;
      ___clang_call_terminate($20);
      // unreachable;
     } else {
      $10 = ($4|0)==(0|0);
      if ($10) {
       $22 = $7;
       break L1;
      }
      __THREW__ = 0;
      (invoke_ii(958,($4|0))|0);
      $11 = __THREW__; __THREW__ = 0;
      $12 = $11&1;
      if (!($12)) {
       $22 = $7;
       break L1;
      }
      $13 = ___cxa_find_matching_catch_3(0|0)|0;
      $14 = tempRet0;
      ___clang_call_terminate($13);
      // unreachable;
     }
    }
   } while(0);
   ___clang_call_terminate($eh$lpad$body$indexZ2D);
   // unreachable;
  }
 } while(0);
 return ($22|0);
}
function __ZNSt3__27codecvtIwc11__mbstate_tED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21456);
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 __THREW__ = 0;
 $3 = (invoke_i(955)|0);
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 do {
  if (!($5)) {
   $6 = ($2|0)==($3|0);
   if (!($6)) {
    $7 = HEAP32[$1>>2]|0;
    __THREW__ = 0;
    invoke_vi(1000,($7|0));
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if ($9) {
     break;
    }
   }
   __ZNSt3__26locale5facetD2Ev($0);
   return;
  }
 } while(0);
 $10 = ___cxa_find_matching_catch_3(0|0)|0;
 $11 = tempRet0;
 __ZNSt3__26locale5facetD2Ev($0);
 ___clang_call_terminate($10);
 // unreachable;
}
function __ZNSt3__27codecvtIwc11__mbstate_tED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__27codecvtIwc11__mbstate_tED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE6do_outERS1_PKDsS5_RS5_PcS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $10 = 0, $11 = 0, $12 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp + 4|0;
 $9 = sp;
 HEAP32[$8>>2] = $2;
 HEAP32[$9>>2] = $5;
 $10 = (__ZNSt3__2L13utf16_to_utf8EPKtS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($2,$3,$8,$5,$6,$9,1114111,0)|0);
 $11 = HEAP32[$8>>2]|0;
 HEAP32[$4>>2] = $11;
 $12 = HEAP32[$9>>2]|0;
 HEAP32[$7>>2] = $12;
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE5do_inERS1_PKcS5_RS5_PDsS7_RS7_($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $10 = 0, $11 = 0, $12 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp + 4|0;
 $9 = sp;
 HEAP32[$8>>2] = $2;
 HEAP32[$9>>2] = $5;
 $10 = (__ZNSt3__2L13utf8_to_utf16EPKhS1_RS1_PtS3_RS3_mNS_12codecvt_modeE($2,$3,$8,$5,$6,$9,1114111,0)|0);
 $11 = HEAP32[$8>>2]|0;
 HEAP32[$4>>2] = $11;
 $12 = HEAP32[$9>>2]|0;
 HEAP32[$7>>2] = $12;
 STACKTOP = sp;return ($10|0);
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$4>>2] = $2;
 return 3;
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE11do_encodingEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE16do_always_noconvEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 0;
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE9do_lengthERS1_PKcS5_j($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = (__ZNSt3__2L20utf8_to_utf16_lengthEPKhS1_jmNS_12codecvt_modeE($2,$3,$4,1114111,0)|0);
 return ($5|0);
}
function __ZNKSt3__27codecvtIDsc11__mbstate_tE13do_max_lengthEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 4;
}
function __ZNSt3__2L20utf8_to_utf16_lengthEPKhS1_jmNS_12codecvt_modeE($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$0100 = 0, $$194 = 0, $$2102 = 0, $$598 = 0, $$lcssa125$sink = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $or$cond = 0, $or$cond110 = 0, $or$cond115 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = $4 & 4;
 $6 = ($5|0)!=(0);
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 $or$cond = $10 & $6;
 if ($or$cond) {
  $11 = HEAP8[$0>>0]|0;
  $12 = ($11<<24>>24)==(-17);
  if ($12) {
   $13 = ((($0)) + 1|0);
   $14 = HEAP8[$13>>0]|0;
   $15 = ($14<<24>>24)==(-69);
   if ($15) {
    $16 = ((($0)) + 2|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(-65);
    $19 = ((($0)) + 3|0);
    $$ = $18 ? $19 : $0;
    $$0100 = 0;$$194 = $$;
   } else {
    $$0100 = 0;$$194 = $0;
   }
  } else {
   $$0100 = 0;$$194 = $0;
  }
 } else {
  $$0100 = 0;$$194 = $0;
 }
 L5: while(1) {
  $20 = ($$194>>>0)<($1>>>0);
  $21 = ($$0100>>>0)<($2>>>0);
  $or$cond110 = $21 & $20;
  if (!($or$cond110)) {
   label = 31;
   break;
  }
  $22 = HEAP8[$$194>>0]|0;
  $23 = $22&255;
  $24 = ($23>>>0)>($3>>>0);
  if ($24) {
   label = 31;
   break;
  }
  $25 = ($22<<24>>24)>(-1);
  $26 = ((($$194)) + 1|0);
  do {
   if ($25) {
    $$2102 = $$0100;$$598 = $26;
   } else {
    $27 = ($22&255)<(194);
    if ($27) {
     label = 31;
     break L5;
    }
    $28 = ($22&255)<(224);
    $29 = ((($$194)) + 2|0);
    $30 = $$194;
    $31 = (($7) - ($30))|0;
    if ($28) {
     $32 = ($31|0)<(2);
     if ($32) {
      label = 31;
      break L5;
     }
     $33 = HEAP8[$26>>0]|0;
     $34 = $33&255;
     $35 = $34 & 192;
     $36 = ($35|0)==(128);
     if (!($36)) {
      label = 31;
      break L5;
     }
     $37 = $23 << 6;
     $38 = $37 & 1984;
     $39 = $34 & 63;
     $40 = $39 | $38;
     $41 = ($40>>>0)>($3>>>0);
     if ($41) {
      label = 31;
      break L5;
     } else {
      $$2102 = $$0100;$$598 = $29;
      break;
     }
    }
    $42 = ($22&255)<(240);
    $43 = ((($$194)) + 3|0);
    if ($42) {
     $44 = ($31|0)<(3);
     if ($44) {
      label = 31;
      break L5;
     }
     $45 = HEAP8[$26>>0]|0;
     $46 = HEAP8[$29>>0]|0;
     $47 = $45&255;
     $48 = $47 & 224;
     switch ($22<<24>>24) {
     case -32:  {
      $49 = ($48|0)==(160);
      if (!($49)) {
       $$lcssa125$sink = $30;
       break L5;
      }
      break;
     }
     case -19:  {
      $50 = ($48|0)==(128);
      if (!($50)) {
       $$lcssa125$sink = $30;
       break L5;
      }
      break;
     }
     default: {
      $51 = $47 & 192;
      $52 = ($51|0)==(128);
      if (!($52)) {
       $$lcssa125$sink = $30;
       break L5;
      }
     }
     }
     $53 = $46&255;
     $54 = $53 & 192;
     $55 = ($54|0)==(128);
     if (!($55)) {
      label = 31;
      break L5;
     }
     $56 = $23 << 12;
     $57 = $56 & 61440;
     $58 = $47 << 6;
     $59 = $58 & 4032;
     $60 = $59 | $57;
     $61 = $53 & 63;
     $62 = $60 | $61;
     $63 = ($62>>>0)>($3>>>0);
     if ($63) {
      label = 31;
      break L5;
     } else {
      $$2102 = $$0100;$$598 = $43;
      break;
     }
    }
    $64 = ($22&255)<(245);
    if (!($64)) {
     label = 31;
     break L5;
    }
    $65 = ($31|0)<(4);
    $66 = (($2) - ($$0100))|0;
    $67 = ($66>>>0)<(2);
    $or$cond115 = $67 | $65;
    if ($or$cond115) {
     label = 31;
     break L5;
    }
    $68 = HEAP8[$26>>0]|0;
    $69 = HEAP8[$29>>0]|0;
    $70 = HEAP8[$43>>0]|0;
    $71 = $68&255;
    switch ($22<<24>>24) {
    case -16:  {
     $$off = (($68) + 112)<<24>>24;
     $72 = ($$off&255)<(48);
     if (!($72)) {
      $$lcssa125$sink = $30;
      break L5;
     }
     break;
    }
    case -12:  {
     $73 = $71 & 240;
     $74 = ($73|0)==(128);
     if (!($74)) {
      $$lcssa125$sink = $30;
      break L5;
     }
     break;
    }
    default: {
     $75 = $71 & 192;
     $76 = ($75|0)==(128);
     if (!($76)) {
      $$lcssa125$sink = $30;
      break L5;
     }
    }
    }
    $77 = $69&255;
    $78 = $77 & 192;
    $79 = ($78|0)==(128);
    if (!($79)) {
     label = 31;
     break L5;
    }
    $80 = $70&255;
    $81 = $80 & 192;
    $82 = ($81|0)==(128);
    if (!($82)) {
     label = 31;
     break L5;
    }
    $83 = $23 << 18;
    $84 = $83 & 1835008;
    $85 = $71 << 12;
    $86 = $85 & 258048;
    $87 = $86 | $84;
    $88 = $77 << 6;
    $89 = $88 & 4032;
    $90 = $87 | $89;
    $91 = $80 & 63;
    $92 = $90 | $91;
    $93 = ($92>>>0)>($3>>>0);
    $94 = (($$0100) + 1)|0;
    $95 = ((($$194)) + 4|0);
    if ($93) {
     label = 31;
     break L5;
    } else {
     $$2102 = $94;$$598 = $95;
    }
   }
  } while(0);
  $96 = (($$2102) + 1)|0;
  $$0100 = $96;$$194 = $$598;
 }
 if ((label|0) == 31) {
  $97 = $$194;
  $$lcssa125$sink = $97;
 }
 $98 = (($$lcssa125$sink) - ($8))|0;
 return ($98|0);
}
function __ZNSt3__2L13utf8_to_utf16EPKhS1_RS1_PtS3_RS3_mNS_12codecvt_modeE($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$8 = 0, $$off = 0, $$sink = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0;
 var $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 HEAP32[$5>>2] = $3;
 $8 = $7 & 4;
 $9 = ($8|0)==(0);
 $10 = $1;
 if (!($9)) {
  $11 = HEAP32[$2>>2]|0;
  $12 = $11;
  $13 = (($10) - ($12))|0;
  $14 = ($13|0)>(2);
  if ($14) {
   $15 = HEAP8[$11>>0]|0;
   $16 = ($15<<24>>24)==(-17);
   if ($16) {
    $17 = ((($11)) + 1|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = ($18<<24>>24)==(-69);
    if ($19) {
     $20 = ((($11)) + 2|0);
     $21 = HEAP8[$20>>0]|0;
     $22 = ($21<<24>>24)==(-65);
     if ($22) {
      $23 = ((($11)) + 3|0);
      HEAP32[$2>>2] = $23;
     }
    }
   }
  }
 }
 $24 = $4;
 L8: while(1) {
  $25 = HEAP32[$2>>2]|0;
  $26 = ($25>>>0)<($1>>>0);
  if (!($26)) {
   label = 40;
   break;
  }
  $27 = HEAP32[$5>>2]|0;
  $28 = ($27>>>0)<($4>>>0);
  if (!($28)) {
   label = 40;
   break;
  }
  $29 = HEAP8[$25>>0]|0;
  $30 = $29&255;
  $31 = ($30>>>0)>($6>>>0);
  if ($31) {
   $$8 = 2;
   break;
  }
  $32 = ($29<<24>>24)>(-1);
  $33 = ((($25)) + 1|0);
  do {
   if ($32) {
    $34 = $29&255;
    HEAP16[$27>>1] = $34;
    $$sink = $33;
   } else {
    $35 = ($29&255)<(194);
    if ($35) {
     $$8 = 2;
     break L8;
    }
    $36 = ($29&255)<(224);
    $37 = $25;
    $38 = (($10) - ($37))|0;
    $39 = ((($25)) + 2|0);
    if ($36) {
     $40 = ($38|0)<(2);
     if ($40) {
      $$8 = 1;
      break L8;
     }
     $41 = HEAP8[$33>>0]|0;
     $42 = $41&255;
     $43 = $42 & 192;
     $44 = ($43|0)==(128);
     if (!($44)) {
      $$8 = 2;
      break L8;
     }
     $45 = $30 << 6;
     $46 = $45 & 1984;
     $47 = $42 & 63;
     $48 = $47 | $46;
     $49 = ($48>>>0)>($6>>>0);
     if ($49) {
      $$8 = 2;
      break L8;
     }
     $50 = $48&65535;
     HEAP16[$27>>1] = $50;
     $$sink = $39;
     break;
    }
    $51 = ($29&255)<(240);
    $52 = ((($25)) + 3|0);
    if ($51) {
     $53 = ($38|0)<(3);
     if ($53) {
      $$8 = 1;
      break L8;
     }
     $54 = HEAP8[$33>>0]|0;
     $55 = HEAP8[$39>>0]|0;
     $56 = $54&255;
     $57 = $56 & 224;
     switch ($29<<24>>24) {
     case -32:  {
      $58 = ($57|0)==(160);
      if (!($58)) {
       $$8 = 2;
       break L8;
      }
      break;
     }
     case -19:  {
      $59 = ($57|0)==(128);
      if (!($59)) {
       $$8 = 2;
       break L8;
      }
      break;
     }
     default: {
      $60 = $56 & 192;
      $61 = ($60|0)==(128);
      if (!($61)) {
       $$8 = 2;
       break L8;
      }
     }
     }
     $62 = $55&255;
     $63 = $62 & 192;
     $64 = ($63|0)==(128);
     if (!($64)) {
      $$8 = 2;
      break L8;
     }
     $65 = $30 << 12;
     $66 = $56 << 6;
     $67 = $66 & 4032;
     $68 = $67 | $65;
     $69 = $62 & 63;
     $70 = $68 | $69;
     $71 = $70 & 65535;
     $72 = ($71>>>0)>($6>>>0);
     if ($72) {
      $$8 = 2;
      break L8;
     }
     $73 = $70&65535;
     HEAP16[$27>>1] = $73;
     $$sink = $52;
     break;
    }
    $74 = ($29&255)<(245);
    if (!($74)) {
     $$8 = 2;
     break L8;
    }
    $75 = ($38|0)<(4);
    if ($75) {
     $$8 = 1;
     break L8;
    }
    $76 = HEAP8[$33>>0]|0;
    $77 = HEAP8[$39>>0]|0;
    $78 = HEAP8[$52>>0]|0;
    $79 = $76&255;
    switch ($29<<24>>24) {
    case -16:  {
     $$off = (($76) + 112)<<24>>24;
     $80 = ($$off&255)<(48);
     if (!($80)) {
      $$8 = 2;
      break L8;
     }
     break;
    }
    case -12:  {
     $81 = $79 & 240;
     $82 = ($81|0)==(128);
     if (!($82)) {
      $$8 = 2;
      break L8;
     }
     break;
    }
    default: {
     $83 = $79 & 192;
     $84 = ($83|0)==(128);
     if (!($84)) {
      $$8 = 2;
      break L8;
     }
    }
    }
    $85 = $77&255;
    $86 = $85 & 192;
    $87 = ($86|0)==(128);
    if (!($87)) {
     $$8 = 2;
     break L8;
    }
    $88 = $78&255;
    $89 = $88 & 192;
    $90 = ($89|0)==(128);
    if (!($90)) {
     $$8 = 2;
     break L8;
    }
    $91 = $27;
    $92 = (($24) - ($91))|0;
    $93 = ($92|0)<(4);
    if ($93) {
     $$8 = 1;
     break L8;
    }
    $94 = $30 & 7;
    $95 = $94 << 18;
    $96 = $79 << 12;
    $97 = $96 & 258048;
    $98 = $97 | $95;
    $99 = $85 << 6;
    $100 = $99 & 4032;
    $101 = $98 | $100;
    $102 = $88 & 63;
    $103 = $101 | $102;
    $104 = ($103>>>0)>($6>>>0);
    if ($104) {
     $$8 = 2;
     break L8;
    }
    $105 = $94 << 2;
    $106 = $79 >>> 4;
    $107 = $106 & 3;
    $108 = $107 | $105;
    $109 = $108 << 6;
    $110 = (($109) + 16320)|0;
    $111 = $79 << 2;
    $112 = $111 & 60;
    $113 = $85 >>> 4;
    $114 = $113 & 3;
    $115 = $112 | $114;
    $116 = $115 | $110;
    $117 = $116 | 55296;
    $118 = $117&65535;
    HEAP16[$27>>1] = $118;
    $119 = $99 & 960;
    $120 = $102 | $119;
    $121 = $120 | 56320;
    $122 = $121&65535;
    $123 = ((($27)) + 2|0);
    HEAP32[$5>>2] = $123;
    HEAP16[$123>>1] = $122;
    $124 = HEAP32[$2>>2]|0;
    $125 = ((($124)) + 4|0);
    $$sink = $125;
   }
  } while(0);
  HEAP32[$2>>2] = $$sink;
  $126 = HEAP32[$5>>2]|0;
  $127 = ((($126)) + 2|0);
  HEAP32[$5>>2] = $127;
 }
 if ((label|0) == 40) {
  $128 = $26&1;
  $$8 = $128;
 }
 return ($$8|0);
}
function __ZNSt3__2L13utf16_to_utf8EPKtS1_RS1_PhS3_RS3_mNS_12codecvt_modeE($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$4 = 0, $$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $60 = 0, $61 = 0, $62 = 0;
 var $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = $0;
 HEAP32[$5>>2] = $3;
 $8 = $7 & 2;
 $9 = ($8|0)==(0);
 $10 = $4;
 if ($9) {
  label = 4;
 } else {
  $11 = $3;
  $12 = (($10) - ($11))|0;
  $13 = ($12|0)<(3);
  if ($13) {
   $$4 = 1;
  } else {
   $14 = ((($3)) + 1|0);
   HEAP32[$5>>2] = $14;
   HEAP8[$3>>0] = -17;
   $15 = HEAP32[$5>>2]|0;
   $16 = ((($15)) + 1|0);
   HEAP32[$5>>2] = $16;
   HEAP8[$15>>0] = -69;
   $17 = HEAP32[$5>>2]|0;
   $18 = ((($17)) + 1|0);
   HEAP32[$5>>2] = $18;
   HEAP8[$17>>0] = -65;
   label = 4;
  }
 }
 L4: do {
  if ((label|0) == 4) {
   $19 = $1;
   $$pre = HEAP32[$2>>2]|0;
   $21 = $$pre;
   while(1) {
    $20 = ($21>>>0)<($1>>>0);
    if (!($20)) {
     $$4 = 0;
     break L4;
    }
    $22 = HEAP16[$21>>1]|0;
    $23 = $22&65535;
    $24 = ($23>>>0)>($6>>>0);
    if ($24) {
     $$4 = 2;
     break L4;
    }
    $25 = ($22&65535)<(128);
    do {
     if ($25) {
      $26 = HEAP32[$5>>2]|0;
      $27 = $26;
      $28 = (($10) - ($27))|0;
      $29 = ($28|0)<(1);
      if ($29) {
       $$4 = 1;
       break L4;
      }
      $30 = $22&255;
      $31 = ((($26)) + 1|0);
      HEAP32[$5>>2] = $31;
      HEAP8[$26>>0] = $30;
     } else {
      $32 = ($22&65535)<(2048);
      $33 = $23 & 63;
      $34 = $33 | 128;
      $35 = $34&255;
      if ($32) {
       $36 = HEAP32[$5>>2]|0;
       $37 = $36;
       $38 = (($10) - ($37))|0;
       $39 = ($38|0)<(2);
       if ($39) {
        $$4 = 1;
        break L4;
       }
       $40 = $23 >>> 6;
       $41 = $40 | 192;
       $42 = $41&255;
       $43 = ((($36)) + 1|0);
       HEAP32[$5>>2] = $43;
       HEAP8[$36>>0] = $42;
       $44 = HEAP32[$5>>2]|0;
       $45 = ((($44)) + 1|0);
       HEAP32[$5>>2] = $45;
       HEAP8[$44>>0] = $35;
       break;
      }
      $46 = ($22&65535)<(55296);
      $47 = $23 >>> 12;
      $48 = $47 | 224;
      $49 = $48&255;
      $50 = $23 >>> 6;
      $51 = $50 & 63;
      $52 = $51 | 128;
      $53 = $52&255;
      if ($46) {
       $54 = HEAP32[$5>>2]|0;
       $55 = $54;
       $56 = (($10) - ($55))|0;
       $57 = ($56|0)<(3);
       if ($57) {
        $$4 = 1;
        break L4;
       }
       $58 = ((($54)) + 1|0);
       HEAP32[$5>>2] = $58;
       HEAP8[$54>>0] = $49;
       $59 = HEAP32[$5>>2]|0;
       $60 = ((($59)) + 1|0);
       HEAP32[$5>>2] = $60;
       HEAP8[$59>>0] = $53;
       $61 = HEAP32[$5>>2]|0;
       $62 = ((($61)) + 1|0);
       HEAP32[$5>>2] = $62;
       HEAP8[$61>>0] = $35;
       break;
      }
      $63 = ($22&65535)<(56320);
      if (!($63)) {
       $113 = ($22&65535)<(57344);
       if ($113) {
        $$4 = 2;
        break L4;
       }
       $114 = HEAP32[$5>>2]|0;
       $115 = $114;
       $116 = (($10) - ($115))|0;
       $117 = ($116|0)<(3);
       if ($117) {
        $$4 = 1;
        break L4;
       }
       $118 = ((($114)) + 1|0);
       HEAP32[$5>>2] = $118;
       HEAP8[$114>>0] = $49;
       $119 = HEAP32[$5>>2]|0;
       $120 = ((($119)) + 1|0);
       HEAP32[$5>>2] = $120;
       HEAP8[$119>>0] = $53;
       $121 = HEAP32[$5>>2]|0;
       $122 = ((($121)) + 1|0);
       HEAP32[$5>>2] = $122;
       HEAP8[$121>>0] = $35;
       break;
      }
      $64 = $21;
      $65 = (($19) - ($64))|0;
      $66 = ($65|0)<(4);
      if ($66) {
       $$4 = 1;
       break L4;
      }
      $67 = ((($21)) + 2|0);
      $68 = HEAP16[$67>>1]|0;
      $69 = $68&65535;
      $70 = $69 & 64512;
      $71 = ($70|0)==(56320);
      if (!($71)) {
       $$4 = 2;
       break L4;
      }
      $72 = HEAP32[$5>>2]|0;
      $73 = (($10) - ($72))|0;
      $74 = ($73|0)<(4);
      if ($74) {
       $$4 = 1;
       break L4;
      }
      $75 = $23 & 960;
      $76 = $75 << 10;
      $77 = (($76) + 65536)|0;
      $78 = $33 << 10;
      $79 = $77 | $78;
      $80 = $69 & 1023;
      $81 = $79 | $80;
      $82 = ($81>>>0)>($6>>>0);
      if ($82) {
       $$4 = 2;
       break L4;
      }
      HEAP32[$2>>2] = $67;
      $83 = $75 >>> 6;
      $84 = (($83) + 1)|0;
      $85 = $84 >>> 2;
      $86 = $85 | 240;
      $87 = $86&255;
      $88 = HEAP32[$5>>2]|0;
      $89 = ((($88)) + 1|0);
      HEAP32[$5>>2] = $89;
      HEAP8[$88>>0] = $87;
      $90 = $84 << 4;
      $91 = $90 & 48;
      $92 = $23 >>> 2;
      $93 = $92 & 15;
      $94 = $93 | $91;
      $95 = $94 | 128;
      $96 = $95&255;
      $97 = HEAP32[$5>>2]|0;
      $98 = ((($97)) + 1|0);
      HEAP32[$5>>2] = $98;
      HEAP8[$97>>0] = $96;
      $99 = $23 << 4;
      $100 = $99 & 48;
      $101 = $69 >>> 6;
      $102 = $101 & 15;
      $103 = $100 | $102;
      $104 = $103 | 128;
      $105 = $104&255;
      $106 = HEAP32[$5>>2]|0;
      $107 = ((($106)) + 1|0);
      HEAP32[$5>>2] = $107;
      HEAP8[$106>>0] = $105;
      $108 = $69 & 63;
      $109 = $108 | 128;
      $110 = $109&255;
      $111 = HEAP32[$5>>2]|0;
      $112 = ((($111)) + 1|0);
      HEAP32[$5>>2] = $112;
      HEAP8[$111>>0] = $110;
     }
    } while(0);
    $123 = HEAP32[$2>>2]|0;
    $124 = ((($123)) + 2|0);
    HEAP32[$2>>2] = $124;
    $21 = $124;
   }
  }
 } while(0);
 return ($$4|0);
}
function __ZNSt3__26locale5__impD2Ev($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21504);
 $1 = ((($0)) + 8|0);
 $2 = ((($0)) + 12|0);
 $$0 = 0;
 while(1) {
  $3 = HEAP32[$2>>2]|0;
  $4 = HEAP32[$1>>2]|0;
  $5 = (($3) - ($4))|0;
  $6 = $5 >> 2;
  $7 = ($$0>>>0)<($6>>>0);
  if (!($7)) {
   break;
  }
  $9 = $4;
  $10 = (($9) + ($$0<<2)|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ($11|0)==(0|0);
  if (!($12)) {
   (__ZNSt3__214__shared_count16__release_sharedEv($11)|0);
  }
  $13 = (($$0) + 1)|0;
  $$0 = $13;
 }
 $8 = ((($0)) + 144|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
 __ZNSt3__213__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__26locale5__impD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5__impD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__213__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($0) {
 $0 = $0|0;
 var $$pre$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 do {
  if (!($2)) {
   $3 = ((($0)) + 4|0);
   $$pre$i$i = HEAP32[$3>>2]|0;
   $5 = $$pre$i$i;
   while(1) {
    $4 = ($5|0)==($1|0);
    if ($4) {
     break;
    }
    $6 = ((($5)) + -4|0);
    HEAP32[$3>>2] = $6;
    $5 = $6;
   }
   $7 = ((($0)) + 16|0);
   $8 = ($7|0)==($1|0);
   if ($8) {
    $9 = ((($0)) + 128|0);
    HEAP8[$9>>0] = 0;
    break;
   } else {
    __ZdlPv($1);
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__25ctypeIcED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21524);
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ($2|0)==(0|0);
 if (!($3)) {
  $4 = ((($0)) + 12|0);
  $5 = HEAP8[$4>>0]|0;
  $6 = ($5<<24>>24)==(0);
  if (!($6)) {
   __ZdaPv($2);
  }
 }
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__25ctypeIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__25ctypeIcED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__25ctypeIcE10do_toupperEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1<<24>>24)>(-1);
 if ($2) {
  $3 = $1&255;
  $4 = (__ZNSt3__25ctypeIcE21__classic_upper_tableEv()|0);
  $5 = (($4) + ($3<<2)|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = $6&255;
  $8 = $7;
 } else {
  $8 = $1;
 }
 return ($8|0);
}
function __ZNKSt3__25ctypeIcE10do_toupperEPcPKc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $1;
 while(1) {
  $3 = ($$0|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = HEAP8[$$0>>0]|0;
  $5 = ($4<<24>>24)>(-1);
  if ($5) {
   $6 = (__ZNSt3__25ctypeIcE21__classic_upper_tableEv()|0);
   $7 = HEAP8[$$0>>0]|0;
   $8 = $7 << 24 >> 24;
   $9 = (($6) + ($8<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = $10&255;
   $12 = $11;
  } else {
   $12 = $4;
  }
  HEAP8[$$0>>0] = $12;
  $13 = ((($$0)) + 1|0);
  $$0 = $13;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIcE10do_tolowerEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1<<24>>24)>(-1);
 if ($2) {
  $3 = $1 << 24 >> 24;
  $4 = (__ZNSt3__25ctypeIcE21__classic_lower_tableEv()|0);
  $5 = (($4) + ($3<<2)|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = $6&255;
  $8 = $7;
 } else {
  $8 = $1;
 }
 return ($8|0);
}
function __ZNKSt3__25ctypeIcE10do_tolowerEPcPKc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $1;
 while(1) {
  $3 = ($$0|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = HEAP8[$$0>>0]|0;
  $5 = ($4<<24>>24)>(-1);
  if ($5) {
   $6 = (__ZNSt3__25ctypeIcE21__classic_lower_tableEv()|0);
   $7 = HEAP8[$$0>>0]|0;
   $8 = $7 << 24 >> 24;
   $9 = (($6) + ($8<<2)|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = $10&255;
   $12 = $11;
  } else {
   $12 = $4;
  }
  HEAP8[$$0>>0] = $12;
  $13 = ((($$0)) + 1|0);
  $$0 = $13;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIcE8do_widenEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return ($1|0);
}
function __ZNKSt3__25ctypeIcE8do_widenEPKcS3_Pc($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$07 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $3;$$07 = $1;
 while(1) {
  $4 = ($$07|0)==($2|0);
  if ($4) {
   break;
  }
  $5 = HEAP8[$$07>>0]|0;
  HEAP8[$$0>>0] = $5;
  $6 = ((($$07)) + 1|0);
  $7 = ((($$0)) + 1|0);
  $$0 = $7;$$07 = $6;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIcE9do_narrowEcc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($1<<24>>24)>(-1);
 $$ = $3 ? $1 : $2;
 return ($$|0);
}
function __ZNKSt3__25ctypeIcE9do_narrowEPKcS3_cPc($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$ = 0, $$0 = 0, $$09 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $4;$$09 = $1;
 while(1) {
  $5 = ($$09|0)==($2|0);
  if ($5) {
   break;
  }
  $6 = HEAP8[$$09>>0]|0;
  $7 = ($6<<24>>24)>(-1);
  $$ = $7 ? $6 : $3;
  HEAP8[$$0>>0] = $$;
  $8 = ((($$09)) + 1|0);
  $9 = ((($$0)) + 1|0);
  $$0 = $9;$$09 = $8;
 }
 return ($2|0);
}
function __ZNSt3__25ctypeIcE21__classic_lower_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(1032)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNSt3__25ctypeIcE21__classic_upper_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(1033)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNSt3__25ctypeIcE13classic_tableEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(1034)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $4 = ___cxa_find_matching_catch_3(0|0)|0;
  $5 = tempRet0;
  ___clang_call_terminate($4);
  // unreachable;
 } else {
  $3 = HEAP32[$0>>2]|0;
  return ($3|0);
 }
 return (0)|0;
}
function __ZNSt3__28numpunctIcED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21576);
 $1 = ((($0)) + 12|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28numpunctIcED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28numpunctIcED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28numpunctIcE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 $2 = HEAP8[$1>>0]|0;
 return ($2|0);
}
function __ZNKSt3__28numpunctIcE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 9|0);
 $2 = HEAP8[$1>>0]|0;
 return ($2|0);
}
function __ZNKSt3__28numpunctIcE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 12|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($0,$2);
 return;
}
function __ZNKSt3__28numpunctIcE11do_truenameEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160289)|0);
 $4 = ($3>>>0)>(4294967279);
 if ($4) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $5 = ($3>>>0)<(11);
 if ($5) {
  $6 = $3&255;
  $7 = ((($0)) + 11|0);
  HEAP8[$7>>0] = $6;
  $$0$i$i = $0;
 } else {
  $8 = (($3) + 16)|0;
  $9 = $8 & -16;
  $10 = (__Znwj($9)|0);
  HEAP32[$0>>2] = $10;
  $11 = $9 | -2147483648;
  $12 = ((($0)) + 8|0);
  HEAP32[$12>>2] = $11;
  $13 = ((($0)) + 4|0);
  HEAP32[$13>>2] = $3;
  $$0$i$i = $10;
 }
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,160289,$3)|0);
 $14 = (($$0$i$i) + ($3)|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($14,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__28numpunctIcE12do_falsenameEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160283)|0);
 $4 = ($3>>>0)>(4294967279);
 if ($4) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $5 = ($3>>>0)<(11);
 if ($5) {
  $6 = $3&255;
  $7 = ((($0)) + 11|0);
  HEAP8[$7>>0] = $6;
  $$0$i$i = $0;
 } else {
  $8 = (($3) + 16)|0;
  $9 = $8 & -16;
  $10 = (__Znwj($9)|0);
  HEAP32[$0>>2] = $10;
  $11 = $9 | -2147483648;
  $12 = ((($0)) + 8|0);
  HEAP32[$12>>2] = $11;
  $13 = ((($0)) + 4|0);
  HEAP32[$13>>2] = $3;
  $$0$i$i = $10;
 }
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,160283,$3)|0);
 $14 = (($$0$i$i) + ($3)|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($14,$2);
 STACKTOP = sp;return;
}
function __ZNSt3__28numpunctIwED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21616);
 $1 = ((($0)) + 16|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($1);
 __ZNSt3__26locale5facetD2Ev($0);
 return;
}
function __ZNSt3__28numpunctIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28numpunctIwED2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__28numpunctIwE16do_decimal_pointEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 $2 = HEAP32[$1>>2]|0;
 return ($2|0);
}
function __ZNKSt3__28numpunctIwE16do_thousands_sepEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 12|0);
 $2 = HEAP32[$1>>2]|0;
 return ($2|0);
}
function __ZNKSt3__28numpunctIwE11do_groupingEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 16|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($0,$2);
 return;
}
function __ZNKSt3__28numpunctIwE11do_truenameEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZNSt3__211char_traitsIwE6lengthEPKw(21672)|0);
 $4 = ($3>>>0)>(1073741807);
 if ($4) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $5 = ($3>>>0)<(2);
 do {
  if ($5) {
   $6 = $3&255;
   $7 = ((($0)) + 8|0);
   $8 = ((($7)) + 3|0);
   HEAP8[$8>>0] = $6;
   $$0$i$i = $0;
  } else {
   $9 = (($3) + 4)|0;
   $10 = $9 & -4;
   $11 = ($10>>>0)>(1073741823);
   if (!($11)) {
    $17 = $10 << 2;
    $18 = (__Znwj($17)|0);
    HEAP32[$0>>2] = $18;
    $19 = $10 | -2147483648;
    $20 = ((($0)) + 8|0);
    HEAP32[$20>>2] = $19;
    $21 = ((($0)) + 4|0);
    HEAP32[$21>>2] = $3;
    $$0$i$i = $18;
    break;
   }
   $12 = (___cxa_allocate_exception(8)|0);
   __THREW__ = 0;
   invoke_vii(751,($12|0),(158021|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    ___cxa_free_exception(($12|0));
    ___resumeException($15|0);
    // unreachable;
   } else {
    HEAP32[$12>>2] = (22172);
    ___cxa_throw(($12|0),(2528|0),(713|0));
    // unreachable;
   }
  }
 } while(0);
 (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,21672,$3)|0);
 $22 = (($$0$i$i) + ($3<<2)|0);
 HEAP32[$2>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($22,$2);
 STACKTOP = sp;return;
}
function __ZNKSt3__28numpunctIwE12do_falsenameEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZNSt3__211char_traitsIwE6lengthEPKw(21648)|0);
 $4 = ($3>>>0)>(1073741807);
 if ($4) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $5 = ($3>>>0)<(2);
 do {
  if ($5) {
   $6 = $3&255;
   $7 = ((($0)) + 8|0);
   $8 = ((($7)) + 3|0);
   HEAP8[$8>>0] = $6;
   $$0$i$i = $0;
  } else {
   $9 = (($3) + 4)|0;
   $10 = $9 & -4;
   $11 = ($10>>>0)>(1073741823);
   if (!($11)) {
    $17 = $10 << 2;
    $18 = (__Znwj($17)|0);
    HEAP32[$0>>2] = $18;
    $19 = $10 | -2147483648;
    $20 = ((($0)) + 8|0);
    HEAP32[$20>>2] = $19;
    $21 = ((($0)) + 4|0);
    HEAP32[$21>>2] = $3;
    $$0$i$i = $18;
    break;
   }
   $12 = (___cxa_allocate_exception(8)|0);
   __THREW__ = 0;
   invoke_vii(751,($12|0),(158021|0));
   $13 = __THREW__; __THREW__ = 0;
   $14 = $13&1;
   if ($14) {
    $15 = ___cxa_find_matching_catch_2()|0;
    $16 = tempRet0;
    ___cxa_free_exception(($12|0));
    ___resumeException($15|0);
    // unreachable;
   } else {
    HEAP32[$12>>2] = (22172);
    ___cxa_throw(($12|0),(2528|0),(713|0));
    // unreachable;
   }
  }
 } while(0);
 (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$0$i$i,21648,$3)|0);
 $22 = (($$0$i$i) + ($3<<2)|0);
 HEAP32[$2>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($22,$2);
 STACKTOP = sp;return;
}
function __ZNSt3__26locale5facetD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__25ctypeIwED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__25ctypeIwE5do_isEtw($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2>>>0)<(128);
 if ($3) {
  $4 = (__ZNSt3__25ctypeIcE13classic_tableEv()|0);
  $5 = (($4) + ($2<<1)|0);
  $6 = HEAP16[$5>>1]|0;
  $7 = $6 & $1;
  $8 = ($7<<16>>16)!=(0);
  $9 = $8;
 } else {
  $9 = 0;
 }
 return ($9|0);
}
function __ZNKSt3__25ctypeIwE5do_isEPKwS3_Pt($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$08 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $3;$$08 = $1;
 while(1) {
  $4 = ($$08|0)==($2|0);
  if ($4) {
   break;
  }
  $5 = HEAP32[$$08>>2]|0;
  $6 = ($5>>>0)<(128);
  if ($6) {
   $7 = (__ZNSt3__25ctypeIcE13classic_tableEv()|0);
   $8 = HEAP32[$$08>>2]|0;
   $9 = (($7) + ($8<<1)|0);
   $10 = HEAP16[$9>>1]|0;
   $11 = $10&65535;
   $13 = $11;
  } else {
   $13 = 0;
  }
  $12 = $13&65535;
  HEAP16[$$0>>1] = $12;
  $14 = ((($$08)) + 4|0);
  $15 = ((($$0)) + 2|0);
  $$0 = $15;$$08 = $14;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIwE10do_scan_isEtPKwS3_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$0$lcssa = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $2;
 while(1) {
  $4 = ($$0|0)==($3|0);
  if ($4) {
   $$0$lcssa = $3;
   break;
  }
  $5 = HEAP32[$$0>>2]|0;
  $6 = ($5>>>0)<(128);
  if ($6) {
   $7 = (__ZNSt3__25ctypeIcE13classic_tableEv()|0);
   $8 = HEAP32[$$0>>2]|0;
   $9 = (($7) + ($8<<1)|0);
   $10 = HEAP16[$9>>1]|0;
   $11 = $10 & $1;
   $12 = ($11<<16>>16)==(0);
   if (!($12)) {
    $$0$lcssa = $$0;
    break;
   }
  }
  $13 = ((($$0)) + 4|0);
  $$0 = $13;
 }
 return ($$0$lcssa|0);
}
function __ZNKSt3__25ctypeIwE11do_scan_notEtPKwS3_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$0$lcssa = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $2;
 while(1) {
  $4 = ($$0|0)==($3|0);
  if ($4) {
   $$0$lcssa = $3;
   break;
  }
  $5 = HEAP32[$$0>>2]|0;
  $6 = ($5>>>0)<(128);
  if (!($6)) {
   $$0$lcssa = $$0;
   break;
  }
  $7 = (__ZNSt3__25ctypeIcE13classic_tableEv()|0);
  $8 = HEAP32[$$0>>2]|0;
  $9 = (($7) + ($8<<1)|0);
  $10 = HEAP16[$9>>1]|0;
  $11 = $10 & $1;
  $12 = ($11<<16>>16)==(0);
  if ($12) {
   $$0$lcssa = $$0;
   break;
  }
  $13 = ((($$0)) + 4|0);
  $$0 = $13;
 }
 return ($$0$lcssa|0);
}
function __ZNKSt3__25ctypeIwE10do_toupperEw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1>>>0)<(128);
 if ($2) {
  $3 = (__ZNSt3__25ctypeIcE21__classic_upper_tableEv()|0);
  $4 = (($3) + ($1<<2)|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $5;
 } else {
  $6 = $1;
 }
 return ($6|0);
}
function __ZNKSt3__25ctypeIwE10do_toupperEPwPKw($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$pre = 0, $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $1;
 while(1) {
  $3 = ($$0|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$0>>2]|0;
  $5 = ($4>>>0)<(128);
  if ($5) {
   $6 = (__ZNSt3__25ctypeIcE21__classic_upper_tableEv()|0);
   $7 = HEAP32[$$0>>2]|0;
   $8 = (($6) + ($7<<2)|0);
   $$pre = HEAP32[$8>>2]|0;
   $9 = $$pre;
  } else {
   $9 = $4;
  }
  HEAP32[$$0>>2] = $9;
  $10 = ((($$0)) + 4|0);
  $$0 = $10;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIwE10do_tolowerEw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1>>>0)<(128);
 if ($2) {
  $3 = (__ZNSt3__25ctypeIcE21__classic_lower_tableEv()|0);
  $4 = (($3) + ($1<<2)|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $5;
 } else {
  $6 = $1;
 }
 return ($6|0);
}
function __ZNKSt3__25ctypeIwE10do_tolowerEPwPKw($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$pre = 0, $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $1;
 while(1) {
  $3 = ($$0|0)==($2|0);
  if ($3) {
   break;
  }
  $4 = HEAP32[$$0>>2]|0;
  $5 = ($4>>>0)<(128);
  if ($5) {
   $6 = (__ZNSt3__25ctypeIcE21__classic_lower_tableEv()|0);
   $7 = HEAP32[$$0>>2]|0;
   $8 = (($6) + ($7<<2)|0);
   $$pre = HEAP32[$8>>2]|0;
   $9 = $$pre;
  } else {
   $9 = $4;
  }
  HEAP32[$$0>>2] = $9;
  $10 = ((($$0)) + 4|0);
  $$0 = $10;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIwE8do_widenEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = $1 << 24 >> 24;
 return ($2|0);
}
function __ZNKSt3__25ctypeIwE8do_widenEPKcS3_Pw($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$07 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $$0 = $3;$$07 = $1;
 while(1) {
  $4 = ($$07|0)==($2|0);
  if ($4) {
   break;
  }
  $5 = HEAP8[$$07>>0]|0;
  $6 = $5 << 24 >> 24;
  HEAP32[$$0>>2] = $6;
  $7 = ((($$07)) + 1|0);
  $8 = ((($$0)) + 4|0);
  $$0 = $8;$$07 = $7;
 }
 return ($2|0);
}
function __ZNKSt3__25ctypeIwE9do_narrowEwc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($1>>>0)<(128);
 $4 = $1&255;
 $$0 = $3 ? $4 : $2;
 return ($$0|0);
}
function __ZNKSt3__25ctypeIwE9do_narrowEPKwS3_cPc($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$09 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $scevgep = 0, $storemerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = $1;
 $6 = $2;
 $7 = (($6) - ($5))|0;
 $8 = $7 >>> 2;
 $$0 = $4;$$09 = $1;
 while(1) {
  $9 = ($$09|0)==($2|0);
  if ($9) {
   break;
  }
  $10 = HEAP32[$$09>>2]|0;
  $11 = ($10>>>0)<(128);
  $12 = $10&255;
  $storemerge = $11 ? $12 : $3;
  HEAP8[$$0>>0] = $storemerge;
  $13 = ((($$09)) + 4|0);
  $14 = ((($$0)) + 1|0);
  $$0 = $14;$$09 = $13;
 }
 $scevgep = (($1) + ($8<<2)|0);
 return ($scevgep|0);
}
function __ZNSt3__27codecvtIcc11__mbstate_tED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__27codecvtIDsc11__mbstate_tED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__27codecvtIDic11__mbstate_tED0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5facetD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt3__220__time_get_c_storageIcEC2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21932);
 return;
}
function __ZNSt3__220__time_get_c_storageIwEC2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (21968);
 return;
}
function __ZNSt3__25ctypeIcEC2EPKtbj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = $2&1;
 $5 = (($3) + -1)|0;
 $6 = ((($0)) + 4|0);
 HEAP32[$6>>2] = $5;
 HEAP32[$0>>2] = (21524);
 $7 = ((($0)) + 8|0);
 HEAP32[$7>>2] = $1;
 $8 = ((($0)) + 12|0);
 HEAP8[$8>>0] = $4;
 $9 = ($1|0)==(0|0);
 if ($9) {
  $10 = (__ZNSt3__25ctypeIcE13classic_tableEv()|0);
  HEAP32[$7>>2] = $10;
 }
 return;
}
function __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(751,($1|0),(157989|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0));
  ___resumeException($4|0);
  // unreachable;
 } else {
  HEAP32[$1>>2] = (22172);
  ___cxa_throw(($1|0),(2528|0),(713|0));
  // unreachable;
 }
}
function __ZNSt3__26locale5__impC2Ej($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$0$i$i = 0, $$03 = 0, $$1 = 0, $$14 = 0, $$pre$i$i$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0;
 var $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0;
 var $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0;
 var $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $3 = (($1) + -1)|0;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = $3;
 HEAP32[$0>>2] = (21504);
 $5 = ((($0)) + 8|0);
 __THREW__ = 0;
 invoke_vii(1035,($5|0),28);
 $6 = __THREW__; __THREW__ = 0;
 $7 = $6&1;
 if ($7) {
  $135 = ___cxa_find_matching_catch_2()|0;
  $136 = tempRet0;
  $$1 = $136;$$14 = $135;
 } else {
  $8 = ((($0)) + 144|0);
  ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;
  $9 = (__ZNSt3__211char_traitsIcE6lengthEPKc(155860)|0);
  $10 = ($9>>>0)>(4294967279);
  do {
   if ($10) {
    __THREW__ = 0;
    invoke_vi(987,($8|0));
    $11 = __THREW__; __THREW__ = 0;
    label = 66;
   } else {
    $12 = ($9>>>0)<(11);
    if ($12) {
     $13 = $9&255;
     $14 = ((($8)) + 11|0);
     HEAP8[$14>>0] = $13;
     $$0$i$i = $8;
    } else {
     $15 = (($9) + 16)|0;
     $16 = $15 & -16;
     __THREW__ = 0;
     $17 = (invoke_ii(988,($16|0))|0);
     $18 = __THREW__; __THREW__ = 0;
     $19 = $18&1;
     if ($19) {
      label = 66;
      break;
     }
     HEAP32[$8>>2] = $17;
     $20 = $16 | -2147483648;
     $21 = ((($0)) + 152|0);
     HEAP32[$21>>2] = $20;
     $22 = ((($0)) + 148|0);
     HEAP32[$22>>2] = $9;
     $$0$i$i = $17;
    }
    (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,155860,$9)|0);
    $23 = (($$0$i$i) + ($9)|0);
    HEAP8[$2>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($23,$2);
    $24 = HEAP32[$5>>2]|0;
    $25 = ((($0)) + 12|0);
    $$pre$i$i$i = HEAP32[$25>>2]|0;
    $27 = $$pre$i$i$i;
    while(1) {
     $26 = ($27|0)==($24|0);
     if ($26) {
      break;
     }
     $28 = ((($27)) + -4|0);
     HEAP32[$25>>2] = $28;
     $27 = $28;
    }
    __THREW__ = 0;
    invoke_v(1036);
    $29 = __THREW__; __THREW__ = 0;
    $30 = $29&1;
    do {
     if (!($30)) {
      __THREW__ = 0;
      invoke_vii(1037,($0|0),(161168|0));
      $31 = __THREW__; __THREW__ = 0;
      $32 = $31&1;
      if (!($32)) {
       __THREW__ = 0;
       invoke_v(1038);
       $33 = __THREW__; __THREW__ = 0;
       $34 = $33&1;
       if (!($34)) {
        __THREW__ = 0;
        invoke_vii(1039,($0|0),(161176|0));
        $35 = __THREW__; __THREW__ = 0;
        $36 = $35&1;
        if (!($36)) {
         __ZNSt3__212_GLOBAL__N_14makeINS_5ctypeIcEEDnbjEERT_T0_T1_T2_();
         __THREW__ = 0;
         invoke_vii(1040,($0|0),(161184|0));
         $37 = __THREW__; __THREW__ = 0;
         $38 = $37&1;
         if (!($38)) {
          __THREW__ = 0;
          invoke_v(1041);
          $39 = __THREW__; __THREW__ = 0;
          $40 = $39&1;
          if (!($40)) {
           __THREW__ = 0;
           invoke_vii(1042,($0|0),(161200|0));
           $41 = __THREW__; __THREW__ = 0;
           $42 = $41&1;
           if (!($42)) {
            __THREW__ = 0;
            invoke_v(1043);
            $43 = __THREW__; __THREW__ = 0;
            $44 = $43&1;
            if (!($44)) {
             __THREW__ = 0;
             invoke_vii(1044,($0|0),(161208|0));
             $45 = __THREW__; __THREW__ = 0;
             $46 = $45&1;
             if (!($46)) {
              __THREW__ = 0;
              invoke_v(1045);
              $47 = __THREW__; __THREW__ = 0;
              $48 = $47&1;
              if (!($48)) {
               __THREW__ = 0;
               invoke_vii(1046,($0|0),(161216|0));
               $49 = __THREW__; __THREW__ = 0;
               $50 = $49&1;
               if (!($50)) {
                __THREW__ = 0;
                invoke_v(1047);
                $51 = __THREW__; __THREW__ = 0;
                $52 = $51&1;
                if (!($52)) {
                 __THREW__ = 0;
                 invoke_vii(1048,($0|0),(161232|0));
                 $53 = __THREW__; __THREW__ = 0;
                 $54 = $53&1;
                 if (!($54)) {
                  __THREW__ = 0;
                  invoke_v(1049);
                  $55 = __THREW__; __THREW__ = 0;
                  $56 = $55&1;
                  if (!($56)) {
                   __THREW__ = 0;
                   invoke_vii(1050,($0|0),(161240|0));
                   $57 = __THREW__; __THREW__ = 0;
                   $58 = $57&1;
                   if ($58) {
                    break;
                   }
                   __ZNSt3__212_GLOBAL__N_14makeINS_8numpunctIcEEjEERT_T0_();
                   __THREW__ = 0;
                   invoke_vii(1051,($0|0),(161248|0));
                   $59 = __THREW__; __THREW__ = 0;
                   $60 = $59&1;
                   if ($60) {
                    break;
                   }
                   __ZNSt3__212_GLOBAL__N_14makeINS_8numpunctIwEEjEERT_T0_();
                   __THREW__ = 0;
                   invoke_vii(1052,($0|0),(161272|0));
                   $61 = __THREW__; __THREW__ = 0;
                   $62 = $61&1;
                   if ($62) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1053);
                   $63 = __THREW__; __THREW__ = 0;
                   $64 = $63&1;
                   if ($64) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1054,($0|0),(161304|0));
                   $65 = __THREW__; __THREW__ = 0;
                   $66 = $65&1;
                   if ($66) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1055);
                   $67 = __THREW__; __THREW__ = 0;
                   $68 = $67&1;
                   if ($68) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1056,($0|0),(161312|0));
                   $69 = __THREW__; __THREW__ = 0;
                   $70 = $69&1;
                   if ($70) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1057);
                   $71 = __THREW__; __THREW__ = 0;
                   $72 = $71&1;
                   if ($72) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1058,($0|0),(161320|0));
                   $73 = __THREW__; __THREW__ = 0;
                   $74 = $73&1;
                   if ($74) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1059);
                   $75 = __THREW__; __THREW__ = 0;
                   $76 = $75&1;
                   if ($76) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1060,($0|0),(161328|0));
                   $77 = __THREW__; __THREW__ = 0;
                   $78 = $77&1;
                   if ($78) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1061);
                   $79 = __THREW__; __THREW__ = 0;
                   $80 = $79&1;
                   if ($80) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1062,($0|0),(161336|0));
                   $81 = __THREW__; __THREW__ = 0;
                   $82 = $81&1;
                   if ($82) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1063);
                   $83 = __THREW__; __THREW__ = 0;
                   $84 = $83&1;
                   if ($84) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1064,($0|0),(161344|0));
                   $85 = __THREW__; __THREW__ = 0;
                   $86 = $85&1;
                   if ($86) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1065);
                   $87 = __THREW__; __THREW__ = 0;
                   $88 = $87&1;
                   if ($88) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1066,($0|0),(161352|0));
                   $89 = __THREW__; __THREW__ = 0;
                   $90 = $89&1;
                   if ($90) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1067);
                   $91 = __THREW__; __THREW__ = 0;
                   $92 = $91&1;
                   if ($92) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1068,($0|0),(161360|0));
                   $93 = __THREW__; __THREW__ = 0;
                   $94 = $93&1;
                   if ($94) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1069);
                   $95 = __THREW__; __THREW__ = 0;
                   $96 = $95&1;
                   if ($96) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1070,($0|0),(161368|0));
                   $97 = __THREW__; __THREW__ = 0;
                   $98 = $97&1;
                   if ($98) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1071);
                   $99 = __THREW__; __THREW__ = 0;
                   $100 = $99&1;
                   if ($100) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1072,($0|0),(161376|0));
                   $101 = __THREW__; __THREW__ = 0;
                   $102 = $101&1;
                   if ($102) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1073);
                   $103 = __THREW__; __THREW__ = 0;
                   $104 = $103&1;
                   if ($104) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1074,($0|0),(161384|0));
                   $105 = __THREW__; __THREW__ = 0;
                   $106 = $105&1;
                   if ($106) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1075);
                   $107 = __THREW__; __THREW__ = 0;
                   $108 = $107&1;
                   if ($108) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1076,($0|0),(161392|0));
                   $109 = __THREW__; __THREW__ = 0;
                   $110 = $109&1;
                   if ($110) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1077);
                   $111 = __THREW__; __THREW__ = 0;
                   $112 = $111&1;
                   if ($112) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1078,($0|0),(161400|0));
                   $113 = __THREW__; __THREW__ = 0;
                   $114 = $113&1;
                   if ($114) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1079);
                   $115 = __THREW__; __THREW__ = 0;
                   $116 = $115&1;
                   if ($116) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1080,($0|0),(161416|0));
                   $117 = __THREW__; __THREW__ = 0;
                   $118 = $117&1;
                   if ($118) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1081);
                   $119 = __THREW__; __THREW__ = 0;
                   $120 = $119&1;
                   if ($120) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1082,($0|0),(161432|0));
                   $121 = __THREW__; __THREW__ = 0;
                   $122 = $121&1;
                   if ($122) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1083);
                   $123 = __THREW__; __THREW__ = 0;
                   $124 = $123&1;
                   if ($124) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1084,($0|0),(161448|0));
                   $125 = __THREW__; __THREW__ = 0;
                   $126 = $125&1;
                   if ($126) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1085);
                   $127 = __THREW__; __THREW__ = 0;
                   $128 = $127&1;
                   if ($128) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1086,($0|0),(161464|0));
                   $129 = __THREW__; __THREW__ = 0;
                   $130 = $129&1;
                   if ($130) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_v(1087);
                   $131 = __THREW__; __THREW__ = 0;
                   $132 = $131&1;
                   if ($132) {
                    break;
                   }
                   __THREW__ = 0;
                   invoke_vii(1088,($0|0),(161472|0));
                   $133 = __THREW__; __THREW__ = 0;
                   $134 = $133&1;
                   if ($134) {
                    break;
                   }
                   STACKTOP = sp;return;
                  }
                 }
                }
               }
              }
             }
            }
           }
          }
         }
        }
       }
      }
     }
    } while(0);
    $139 = ___cxa_find_matching_catch_2()|0;
    $140 = tempRet0;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($8);
    $$0 = $140;$$03 = $139;
   }
  } while(0);
  if ((label|0) == 66) {
   $137 = ___cxa_find_matching_catch_2()|0;
   $138 = tempRet0;
   $$0 = $138;$$03 = $137;
  }
  __ZNSt3__213__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($5);
  $$1 = $$0;$$14 = $$03;
 }
 __ZNSt3__26locale5facetD2Ev($0);
 ___resumeException($$14|0);
 // unreachable;
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEEC2Ej($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($0)) + 128|0);
 HEAP8[$4>>0] = 0;
 $5 = ($1|0)==(0);
 do {
  if (!($5)) {
   __THREW__ = 0;
   invoke_vii(1089,($0|0),($1|0));
   $6 = __THREW__; __THREW__ = 0;
   $7 = $6&1;
   if (!($7)) {
    __THREW__ = 0;
    invoke_vii(1090,($0|0),($1|0));
    $8 = __THREW__; __THREW__ = 0;
    $9 = $8&1;
    if (!($9)) {
     break;
    }
   }
   $10 = ___cxa_find_matching_catch_2()|0;
   $11 = tempRet0;
   __ZNSt3__213__vector_baseIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEED2Ev($0);
   ___resumeException($10|0);
   // unreachable;
  }
 } while(0);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7collateIcEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161172)>>2] = 0;
 HEAP32[40292] = (19168);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7collateIcEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165424)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7collateIwEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161180)>>2] = 0;
 HEAP32[40294] = (19200);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7collateIwEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165432)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_5ctypeIcEEDnbjEERT_T0_T1_T2_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__25ctypeIcEC2EPKtbj(161184,0,0,1);
 return;
}
function __ZNSt3__26locale5__imp7installINS_5ctypeIcEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165440)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_5ctypeIwEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161204)>>2] = 0;
 HEAP32[40300] = (21720);
 return;
}
function __ZNSt3__26locale5__imp7installINS_5ctypeIwEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165472)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIcc11__mbstate_tEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161212)>>2] = 0;
 HEAP32[40302] = (21788);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7codecvtIcc11__mbstate_tEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167232)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIwc11__mbstate_tEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__27codecvtIwc11__mbstate_tEC2Ej(161216,1);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7codecvtIwc11__mbstate_tEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167240)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIDsc11__mbstate_tEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161236)>>2] = 0;
 HEAP32[40308] = (21836);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7codecvtIDsc11__mbstate_tEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167248)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIDic11__mbstate_tEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161244)>>2] = 0;
 HEAP32[40310] = (21884);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7codecvtIDic11__mbstate_tEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167256)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8numpunctIcEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28numpunctIcEC2Ej(161248,1);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8numpunctIcEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165456)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8numpunctIwEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__28numpunctIwEC2Ej(161272,1);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8numpunctIwEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165480)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161308)>>2] = 0;
 HEAP32[40326] = (19232);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165464)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161316)>>2] = 0;
 HEAP32[40328] = (19296);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165488)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161324)>>2] = 0;
 HEAP32[40330] = (19360);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165496)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161332)>>2] = 0;
 HEAP32[40332] = (19412);
 return;
}
function __ZNSt3__26locale5__imp7installINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(165504)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIcLb0EEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161340)>>2] = 0;
 HEAP32[40334] = (20960);
 return;
}
function __ZNSt3__26locale5__imp7installINS_10moneypunctIcLb0EEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167152)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIcLb1EEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161348)>>2] = 0;
 HEAP32[40336] = (21016);
 return;
}
function __ZNSt3__26locale5__imp7installINS_10moneypunctIcLb1EEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167160)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIwLb0EEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161356)>>2] = 0;
 HEAP32[40338] = (21072);
 return;
}
function __ZNSt3__26locale5__imp7installINS_10moneypunctIwLb0EEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167168)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIwLb1EEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161364)>>2] = 0;
 HEAP32[40340] = (21128);
 return;
}
function __ZNSt3__26locale5__imp7installINS_10moneypunctIwLb1EEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167176)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161372)>>2] = 0;
 HEAP32[40342] = (21184);
 return;
}
function __ZNSt3__26locale5__imp7installINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167184)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161380)>>2] = 0;
 HEAP32[40344] = (21212);
 return;
}
function __ZNSt3__26locale5__imp7installINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167192)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161388)>>2] = 0;
 HEAP32[40346] = (21240);
 return;
}
function __ZNSt3__26locale5__imp7installINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167200)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161396)>>2] = 0;
 HEAP32[40348] = (21268);
 return;
}
function __ZNSt3__26locale5__imp7installINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167208)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161404)>>2] = 0;
 HEAP32[40350] = (21700);
 __ZNSt3__220__time_get_c_storageIcEC2Ev((161408));
 HEAP32[40350] = (19464);
 HEAP32[(161408)>>2] = (19512);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(166316)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161420)>>2] = 0;
 HEAP32[40354] = (21700);
 __ZNSt3__220__time_get_c_storageIwEC2Ev((161424));
 HEAP32[40354] = (19548);
 HEAP32[(161424)>>2] = (19596);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167128)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161436)>>2] = 0;
 HEAP32[40358] = (21700);
 __THREW__ = 0;
 $0 = (invoke_i(955)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZNSt3__26locale5facetD2Ev(161432);
  ___resumeException($3|0);
  // unreachable;
 } else {
  HEAP32[(161440)>>2] = $0;
  HEAP32[40358] = (20912);
  return;
 }
}
function __ZNSt3__26locale5__imp7installINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167136)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161452)>>2] = 0;
 HEAP32[40362] = (21700);
 __THREW__ = 0;
 $0 = (invoke_i(955)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $3 = ___cxa_find_matching_catch_2()|0;
  $4 = tempRet0;
  __ZNSt3__26locale5facetD2Ev(161448);
  ___resumeException($3|0);
  // unreachable;
 } else {
  HEAP32[(161456)>>2] = $0;
  HEAP32[40362] = (20936);
  return;
 }
}
function __ZNSt3__26locale5__imp7installINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167144)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8messagesIcEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161468)>>2] = 0;
 HEAP32[40366] = (21296);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8messagesIcEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167216)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_14makeINS_8messagesIwEEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[(161476)>>2] = 0;
 HEAP32[40368] = (21328);
 return;
}
function __ZNSt3__26locale5__imp7installINS_8messagesIwEEEEvPT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__26locale2id5__getEv(167224)|0);
 __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2);
 return;
}
function __ZNSt3__26locale5__imp7installEPNS0_5facetEl($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $$pre14 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__214__shared_count12__add_sharedEv($1);
 $3 = ((($0)) + 8|0);
 $4 = ((($0)) + 12|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$3>>2]|0;
 $7 = (($5) - ($6))|0;
 $8 = $7 >> 2;
 $9 = ($8>>>0)>($2>>>0);
 do {
  if ($9) {
   $10 = $6;
   $$pre$phiZ2D = $3;$18 = $10;
  } else {
   $11 = (($2) + 1)|0;
   __THREW__ = 0;
   invoke_vii(1091,($3|0),($11|0));
   $12 = __THREW__; __THREW__ = 0;
   $13 = $12&1;
   if (!($13)) {
    $$pre = HEAP32[$3>>2]|0;
    $$pre$phiZ2D = $3;$18 = $$pre;
    break;
   }
   $14 = ___cxa_find_matching_catch_2()|0;
   $15 = tempRet0;
   $16 = ($1|0)==(0|0);
   if (!($16)) {
    __ZNSt3__212_GLOBAL__N_17releaseclEPNS_6locale5facetE($1);
   }
   ___resumeException($14|0);
   // unreachable;
  }
 } while(0);
 $17 = (($18) + ($2<<2)|0);
 $19 = HEAP32[$17>>2]|0;
 $20 = ($19|0)==(0|0);
 if ($20) {
  $22 = $18;
 } else {
  (__ZNSt3__214__shared_count16__release_sharedEv($19)|0);
  $$pre14 = HEAP32[$$pre$phiZ2D>>2]|0;
  $22 = $$pre14;
 }
 $21 = (($22) + ($2<<2)|0);
 HEAP32[$21>>2] = $1;
 return;
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE6resizeEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$0>>2]|0;
 $5 = (($3) - ($4))|0;
 $6 = $5 >> 2;
 $7 = ($6>>>0)<($1>>>0);
 $8 = $4;
 $9 = $3;
 L1: do {
  if ($7) {
   $10 = (($1) - ($6))|0;
   __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8__appendEj($0,$10);
  } else {
   $11 = ($6>>>0)>($1>>>0);
   if ($11) {
    $12 = (($8) + ($1<<2)|0);
    $14 = $9;
    while(1) {
     $13 = ($14|0)==($12|0);
     if ($13) {
      break L1;
     }
     $15 = ((($14)) + -4|0);
     HEAP32[$2>>2] = $15;
     $14 = $15;
    }
   }
  }
 } while(0);
 return;
}
function __ZNSt3__212_GLOBAL__N_17releaseclEPNS_6locale5facetE($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 (__ZNSt3__214__shared_count16__release_sharedEv($0)|0);
 return;
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8__appendEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = ((($0)) + 8|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($0)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (($4) - ($6))|0;
 $8 = $7 >> 2;
 $9 = ($8>>>0)<($1>>>0);
 do {
  if ($9) {
   $10 = HEAP32[$0>>2]|0;
   $11 = (($6) - ($10))|0;
   $12 = $11 >> 2;
   $13 = (($12) + ($1))|0;
   $14 = (__ZNKSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8max_sizeEv($0)|0);
   $15 = ($14>>>0)<($13>>>0);
   if ($15) {
    __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
    // unreachable;
   }
   $16 = ((($0)) + 16|0);
   $17 = HEAP32[$3>>2]|0;
   $18 = HEAP32[$0>>2]|0;
   $19 = (($17) - ($18))|0;
   $20 = $19 >> 2;
   $21 = $14 >>> 1;
   $22 = ($20>>>0)<($21>>>0);
   $23 = $19 >> 1;
   $24 = ($23>>>0)<($13>>>0);
   $$sroa$speculated$i = $24 ? $13 : $23;
   $$sroa$speculated$$i = $22 ? $$sroa$speculated$i : $14;
   $25 = HEAP32[$5>>2]|0;
   $26 = (($25) - ($18))|0;
   $27 = $26 >> 2;
   __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEEC2EjjS6_($2,$$sroa$speculated$$i,$27,$16);
   __THREW__ = 0;
   invoke_vii(1092,($2|0),($1|0));
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if (!($29)) {
    __THREW__ = 0;
    invoke_vii(1093,($0|0),($2|0));
    $30 = __THREW__; __THREW__ = 0;
    $31 = $30&1;
    if (!($31)) {
     __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEED2Ev($2);
     break;
    }
   }
   $32 = ___cxa_find_matching_catch_2()|0;
   $33 = tempRet0;
   __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEED2Ev($2);
   ___resumeException($32|0);
   // unreachable;
  } else {
   __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE18__construct_at_endEj($0,$1);
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE18__construct_at_endEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$pre = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 4|0);
 $$pre = HEAP32[$2>>2]|0;
 $$0 = $1;$3 = $$pre;
 while(1) {
  HEAP32[$3>>2] = 0;
  $4 = HEAP32[$2>>2]|0;
  $5 = ((($4)) + 4|0);
  HEAP32[$2>>2] = $5;
  $6 = (($$0) + -1)|0;
  $7 = ($6|0)==(0);
  if ($7) {
   break;
  } else {
   $$0 = $6;$3 = $5;
  }
 }
 return;
}
function __ZNKSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8max_sizeEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1073741823;
}
function __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEEC2EjjS6_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$not$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = $3;
 $6 = ($1|0)==(0);
 do {
  if ($6) {
   $12 = 0;
  } else {
   $7 = ((($3)) + 112|0);
   $8 = HEAP8[$7>>0]|0;
   $$not$i$i = ($8<<24>>24)==(0);
   $9 = ($1>>>0)<(29);
   $or$cond$i$i = $9 & $$not$i$i;
   if ($or$cond$i$i) {
    HEAP8[$7>>0] = 1;
    $12 = $3;
    break;
   } else {
    $10 = $1 << 2;
    $11 = (__Znwj($10)|0);
    $12 = $11;
    break;
   }
  }
 } while(0);
 HEAP32[$0>>2] = $12;
 $13 = (($12) + ($2<<2)|0);
 $14 = ((($0)) + 8|0);
 HEAP32[$14>>2] = $13;
 $15 = ((($0)) + 4|0);
 HEAP32[$15>>2] = $13;
 $16 = (($12) + ($1<<2)|0);
 $17 = ((($0)) + 12|0);
 HEAP32[$17>>2] = $16;
 return;
}
function __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEE18__construct_at_endEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$pre = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($0)) + 8|0);
 $$pre = HEAP32[$2>>2]|0;
 $$0 = $1;$3 = $$pre;
 while(1) {
  HEAP32[$3>>2] = 0;
  $4 = HEAP32[$2>>2]|0;
  $5 = ((($4)) + 4|0);
  HEAP32[$2>>2] = $5;
  $6 = (($$0) + -1)|0;
  $7 = ($6|0)==(0);
  if ($7) {
   break;
  } else {
   $$0 = $6;$3 = $5;
  }
 }
 return;
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$pre = 0, $$pre$phiZ2D = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 4|0);
 $6 = $2;
 $7 = (($4) - ($6))|0;
 $8 = $7 >> 2;
 $9 = HEAP32[$5>>2]|0;
 $10 = (0 - ($8))|0;
 $11 = (($9) + ($10<<2)|0);
 HEAP32[$5>>2] = $11;
 $12 = ($7|0)>(0);
 if ($12) {
  _memcpy(($11|0),($2|0),($7|0))|0;
  $$pre = HEAP32[$5>>2]|0;
  $$pre$phiZ2D = $5;$15 = $$pre;
 } else {
  $13 = $11;
  $$pre$phiZ2D = $5;$15 = $13;
 }
 $14 = HEAP32[$0>>2]|0;
 HEAP32[$0>>2] = $15;
 HEAP32[$$pre$phiZ2D>>2] = $14;
 $16 = ((($1)) + 8|0);
 $17 = HEAP32[$3>>2]|0;
 $18 = HEAP32[$16>>2]|0;
 HEAP32[$3>>2] = $18;
 HEAP32[$16>>2] = $17;
 $19 = ((($0)) + 8|0);
 $20 = ((($1)) + 12|0);
 $21 = HEAP32[$19>>2]|0;
 $22 = HEAP32[$20>>2]|0;
 HEAP32[$19>>2] = $22;
 HEAP32[$20>>2] = $21;
 $23 = HEAP32[$$pre$phiZ2D>>2]|0;
 HEAP32[$1>>2] = $23;
 return;
}
function __ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEED2Ev($0) {
 $0 = $0|0;
 var $$pre$i$i$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 8|0);
 $$pre$i$i$i = HEAP32[$3>>2]|0;
 $5 = $$pre$i$i$i;
 while(1) {
  $4 = ($5|0)==($2|0);
  if ($4) {
   break;
  }
  $6 = ((($5)) + -4|0);
  HEAP32[$3>>2] = $6;
  $5 = $6;
 }
 $7 = HEAP32[$0>>2]|0;
 $8 = ($7|0)==(0|0);
 do {
  if (!($8)) {
   $9 = ((($0)) + 16|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($10|0)==($7|0);
   if ($11) {
    $12 = ((($10)) + 112|0);
    HEAP8[$12>>0] = 0;
    break;
   } else {
    __ZdlPv($7);
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__28numpunctIwEC2Ej($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (($1) + -1)|0;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $2;
 HEAP32[$0>>2] = (21616);
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = 46;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = 44;
 $6 = ((($0)) + 16|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 return;
}
function __ZNSt3__28numpunctIcEC2Ej($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (($1) + -1)|0;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $2;
 HEAP32[$0>>2] = (21576);
 $4 = ((($0)) + 8|0);
 HEAP8[$4>>0] = 46;
 $5 = ((($0)) + 9|0);
 HEAP8[$5>>0] = 44;
 $6 = ((($0)) + 12|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 return;
}
function __ZNSt3__27codecvtIwc11__mbstate_tEC2Ej($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (($1) + -1)|0;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $2;
 HEAP32[$0>>2] = (21456);
 __THREW__ = 0;
 $4 = (invoke_i(955)|0);
 $5 = __THREW__; __THREW__ = 0;
 $6 = $5&1;
 if ($6) {
  $8 = ___cxa_find_matching_catch_2()|0;
  $9 = tempRet0;
  __ZNSt3__26locale5facetD2Ev($0);
  ___resumeException($8|0);
  // unreachable;
 } else {
  $7 = ((($0)) + 8|0);
  HEAP32[$7>>2] = $4;
  return;
 }
}
function __ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8allocateEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $$not$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNKSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8max_sizeEv($0)|0);
 $3 = ($2>>>0)<($1>>>0);
 if ($3) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $4 = ((($0)) + 128|0);
 $5 = HEAP8[$4>>0]|0;
 $$not$i$i = ($5<<24>>24)==(0);
 $6 = ($1>>>0)<(29);
 $or$cond$i$i = $6 & $$not$i$i;
 if ($or$cond$i$i) {
  $7 = ((($0)) + 16|0);
  HEAP8[$4>>0] = 1;
  $$0$i$i = $7;
 } else {
  $8 = $1 << 2;
  $9 = (__Znwj($8)|0);
  $$0$i$i = $9;
 }
 $10 = ((($0)) + 4|0);
 HEAP32[$10>>2] = $$0$i$i;
 HEAP32[$0>>2] = $$0$i$i;
 $11 = (($$0$i$i) + ($1<<2)|0);
 $12 = ((($0)) + 8|0);
 HEAP32[$12>>2] = $11;
 return;
}
function __ZNSt3__26locale7classicEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161480]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire(161480)|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    (invoke_i(1094)|0);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $7 = ___cxa_find_matching_catch_2()|0;
     $8 = tempRet0;
     ___resumeException($7|0);
     // unreachable;
    } else {
     HEAP32[41817] = 167264;
     break;
    }
   }
  }
 } while(0);
 $6 = HEAP32[41817]|0;
 return ($6|0);
}
function __ZNSt3__26locale5__imp12make_classicEv() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__212_GLOBAL__N_14makeINS_6locale5__impEjEERT_T0_();
 HEAP32[41816] = 161488;
 return (167264|0);
}
function __ZNSt3__212_GLOBAL__N_14makeINS_6locale5__impEjEERT_T0_() {
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__26locale5__impC2Ej(161488,1);
 return;
}
function __ZNSt3__26locale5__imp11make_globalEv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (__ZNSt3__26locale7classicEv()|0);
 __ZNSt3__26localeC2ERKS0_(167272,$0);
 return (167272|0);
}
function __ZNSt3__26localeC2ERKS0_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$0>>2] = $2;
 __ZNSt3__214__shared_count12__add_sharedEv($2);
 return;
}
function __ZNSt3__26locale8__globalEv() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP8[161648]|0;
 $1 = ($0<<24>>24)==(0);
 do {
  if ($1) {
   $2 = (___cxa_guard_acquire(161648)|0);
   $3 = ($2|0)==(0);
   if (!($3)) {
    __THREW__ = 0;
    (invoke_i(1095)|0);
    $4 = __THREW__; __THREW__ = 0;
    $5 = $4&1;
    if ($5) {
     $7 = ___cxa_find_matching_catch_2()|0;
     $8 = tempRet0;
     ___resumeException($7|0);
     // unreachable;
    } else {
     HEAP32[41819] = 167272;
     break;
    }
   }
  }
 } while(0);
 $6 = HEAP32[41819]|0;
 return ($6|0);
}
function __ZNSt3__26localeC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $1 = (invoke_i(1096)|0);
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $5 = ___cxa_find_matching_catch_3(0|0)|0;
  $6 = tempRet0;
  ___clang_call_terminate($5);
  // unreachable;
 } else {
  $4 = HEAP32[$1>>2]|0;
  HEAP32[$0>>2] = $4;
  __ZNSt3__214__shared_count12__add_sharedEv($4);
  return;
 }
}
function __ZNSt3__211char_traitsIcE7compareEPKcS3_j($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if ($3) {
  $$0 = 0;
 } else {
  $4 = (_memcmp($0,$1,$2)|0);
  $$0 = $4;
 }
 return ($$0|0);
}
function __ZNSt3__214__shared_countD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt3__214__shared_count12__add_sharedEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 __ZNSt3__212_GLOBAL__N_19incrementIlEET_RS2_($1);
 return;
}
function __ZNSt3__212_GLOBAL__N_19incrementIlEET_RS2_($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1+1)|0);
 return;
}
function __ZNSt3__214__shared_count16__release_sharedEv($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = (__ZNSt3__212_GLOBAL__N_19decrementIlEET_RS2_($1)|0);
 $3 = ($2|0)==(-1);
 if ($3) {
  $4 = HEAP32[$0>>2]|0;
  $5 = ((($4)) + 8|0);
  $6 = HEAP32[$5>>2]|0;
  FUNCTION_TABLE_vi[$6 & 2047]($0);
  $$0 = 1;
 } else {
  $$0 = 0;
 }
 return ($$0|0);
}
function __ZNSt3__212_GLOBAL__N_19decrementIlEET_RS2_($0) {
 $0 = $0|0;
 var $$0$i = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;HEAP32[$0>>2] = (($1+-1)|0);
 $$0$i = (($1) + -1)|0;
 return ($$0$i|0);
}
function __ZNSt3__211__call_onceERVmPvPFvS2_E($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0;
 var $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 (_pthread_mutex_lock((167280|0))|0);
 while(1) {
  $3 = HEAP32[$0>>2]|0;
  $4 = ($3|0)==(1);
  if (!($4)) {
   break;
  }
  (_pthread_cond_wait((167308|0),(167280|0))|0);
 }
 $5 = HEAP32[$0>>2]|0;
 $6 = ($5|0)==(0);
 do {
  if ($6) {
   HEAP32[$0>>2] = 1;
   __THREW__ = 0;
   (invoke_ii(1097,(167280|0))|0);
   $7 = __THREW__; __THREW__ = 0;
   $8 = $7&1;
   if (!($8)) {
    __THREW__ = 0;
    invoke_vi($2|0,($1|0));
    $9 = __THREW__; __THREW__ = 0;
    $10 = $9&1;
    if (!($10)) {
     __THREW__ = 0;
     (invoke_ii(1098,(167280|0))|0);
     $11 = __THREW__; __THREW__ = 0;
     $12 = $11&1;
     if (!($12)) {
      HEAP32[$0>>2] = -1;
      __THREW__ = 0;
      (invoke_ii(1097,(167280|0))|0);
      $13 = __THREW__; __THREW__ = 0;
      $14 = $13&1;
      if (!($14)) {
       __THREW__ = 0;
       (invoke_ii(1099,(167308|0))|0);
       $15 = __THREW__; __THREW__ = 0;
       $16 = $15&1;
       if (!($16)) {
        break;
       }
      }
     }
    }
   }
   $17 = ___cxa_find_matching_catch_3(0|0)|0;
   $18 = tempRet0;
   (___cxa_begin_catch(($17|0))|0);
   __THREW__ = 0;
   (invoke_ii(1098,(167280|0))|0);
   $19 = __THREW__; __THREW__ = 0;
   $20 = $19&1;
   if (!($20)) {
    HEAP32[$0>>2] = 0;
    __THREW__ = 0;
    (invoke_ii(1097,(167280|0))|0);
    $21 = __THREW__; __THREW__ = 0;
    $22 = $21&1;
    if (!($22)) {
     __THREW__ = 0;
     (invoke_ii(1099,(167308|0))|0);
     $23 = __THREW__; __THREW__ = 0;
     $24 = $23&1;
     if (!($24)) {
      __THREW__ = 0;
      invoke_v(1100);
      $25 = __THREW__; __THREW__ = 0;
     }
    }
   }
   $26 = ___cxa_find_matching_catch_2()|0;
   $27 = tempRet0;
   __THREW__ = 0;
   invoke_v(738);
   $28 = __THREW__; __THREW__ = 0;
   $29 = $28&1;
   if ($29) {
    $30 = ___cxa_find_matching_catch_3(0|0)|0;
    $31 = tempRet0;
    ___clang_call_terminate($30);
    // unreachable;
   } else {
    ___resumeException($26|0);
    // unreachable;
   }
  } else {
   (_pthread_mutex_unlock((167280|0))|0);
  }
 } while(0);
 return;
}
function __Znwj($0) {
 $0 = $0|0;
 var $$ = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0);
 $$ = $1 ? 1 : $0;
 while(1) {
  $2 = (_malloc($$)|0);
  $3 = ($2|0)==(0|0);
  if (!($3)) {
   label = 6;
   break;
  }
  $4 = (__ZSt15get_new_handlerv()|0);
  $5 = ($4|0)==(0|0);
  if ($5) {
   label = 5;
   break;
  }
  FUNCTION_TABLE_v[$4 & 2047]();
 }
 if ((label|0) == 5) {
  $6 = (___cxa_allocate_exception(4)|0);
  __ZNSt9bad_allocC2Ev($6);
  ___cxa_throw(($6|0),(2480|0),(710|0));
  // unreachable;
 }
 else if ((label|0) == 6) {
  return ($2|0);
 }
 return (0)|0;
}
function __Znaj($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__Znwj($0)|0);
 return ($1|0);
}
function __ZdlPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($0);
 return;
}
function __ZdaPv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZdlPv($0);
 return;
}
function __ZSt17__throw_bad_allocv() {
 var $0 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = (___cxa_allocate_exception(4)|0);
 __ZNSt9bad_allocC2Ev($0);
 ___cxa_throw(($0|0),(2480|0),(710|0));
 // unreachable;
}
function __ZNSt3__218__libcpp_refstringC2EPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (_strlen($1)|0);
 $3 = (($2) + 13)|0;
 $4 = (__Znwj($3)|0);
 HEAP32[$4>>2] = $2;
 $5 = ((($4)) + 4|0);
 HEAP32[$5>>2] = $2;
 $6 = ((($4)) + 8|0);
 HEAP32[$6>>2] = 0;
 $7 = (__ZNSt3__215__refstring_imp12_GLOBAL__N_113data_from_repEPNS1_9_Rep_baseE($4)|0);
 $8 = (($2) + 1)|0;
 _memcpy(($7|0),($1|0),($8|0))|0;
 HEAP32[$0>>2] = $7;
 return;
}
function __ZNSt3__215__refstring_imp12_GLOBAL__N_113data_from_repEPNS1_9_Rep_baseE($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 12|0);
 return ($1|0);
}
function __ZNSt11logic_errorC2EPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22132);
 $2 = ((($0)) + 4|0);
 __THREW__ = 0;
 invoke_vii(1101,($2|0),($1|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___resumeException($5|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNKSt3__218__libcpp_refstring15__uses_refcountEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 1;
}
function __ZNSt13runtime_errorC2ERKNSt3__212basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22152);
 $2 = ((($0)) + 4|0);
 $3 = ((($1)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 $6 = HEAP32[$1>>2]|0;
 $7 = $5 ? $6 : $1;
 __THREW__ = 0;
 invoke_vii(1101,($2|0),($7|0));
 $8 = __THREW__; __THREW__ = 0;
 $9 = $8&1;
 if ($9) {
  $10 = ___cxa_find_matching_catch_2()|0;
  $11 = tempRet0;
  ___resumeException($10|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt13runtime_errorC2EPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22152);
 $2 = ((($0)) + 4|0);
 __THREW__ = 0;
 invoke_vii(1101,($2|0),($1|0));
 $3 = __THREW__; __THREW__ = 0;
 $4 = $3&1;
 if ($4) {
  $5 = ___cxa_find_matching_catch_2()|0;
  $6 = tempRet0;
  ___resumeException($5|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(751,($1|0),(159986|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0));
  ___resumeException($4|0);
  // unreachable;
 } else {
  HEAP32[$1>>2] = (22172);
  ___cxa_throw(($1|0),(2528|0),(713|0));
  // unreachable;
 }
}
function __ZNKSt3__221__basic_string_commonILb1EE20__throw_out_of_rangeEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (___cxa_allocate_exception(8)|0);
 __THREW__ = 0;
 invoke_vii(751,($1|0),(159986|0));
 $2 = __THREW__; __THREW__ = 0;
 $3 = $2&1;
 if ($3) {
  $4 = ___cxa_find_matching_catch_2()|0;
  $5 = tempRet0;
  ___cxa_free_exception(($1|0));
  ___resumeException($4|0);
  // unreachable;
 } else {
  HEAP32[$1>>2] = (22192);
  ___cxa_throw(($1|0),(2544|0),(713|0));
  // unreachable;
 }
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEC2ERKS5_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($1)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 if ($5) {
  $6 = HEAP32[$1>>2]|0;
  $7 = ((($1)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8>>>0)>(4294967279);
  if ($9) {
   __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
   // unreachable;
  }
  $10 = ($8>>>0)<(11);
  if ($10) {
   $11 = $8&255;
   $12 = ((($0)) + 11|0);
   HEAP8[$12>>0] = $11;
   $$0$i = $0;
  } else {
   $13 = (($8) + 16)|0;
   $14 = $13 & -16;
   $15 = (__Znwj($14)|0);
   HEAP32[$0>>2] = $15;
   $16 = $14 | -2147483648;
   $17 = ((($0)) + 8|0);
   HEAP32[$17>>2] = $16;
   $18 = ((($0)) + 4|0);
   HEAP32[$18>>2] = $8;
   $$0$i = $15;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i,$6,$8)|0);
  $19 = (($$0$i) + ($8)|0);
  HEAP8[$2>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($19,$2);
 } else {
  ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 11|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)<(0);
 if ($3) {
  $4 = HEAP32[$0>>2]|0;
  __ZdlPv($4);
 }
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $7 = ((($0)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $11 = $phitmp$i;
 } else {
  $11 = 10;
 }
 $10 = ($11>>>0)<($2>>>0);
 do {
  if ($10) {
   if ($6) {
    $19 = ((($0)) + 4|0);
    $20 = HEAP32[$19>>2]|0;
    $23 = $20;
   } else {
    $21 = $5&255;
    $23 = $21;
   }
   $22 = (($2) - ($11))|0;
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$11,$22,$23,0,$23,$2,$1);
  } else {
   if ($6) {
    $12 = HEAP32[$0>>2]|0;
    $13 = $12;
   } else {
    $13 = $0;
   }
   (__ZNSt3__211char_traitsIcE4moveEPcPKcj($13,$1,$2)|0);
   $14 = (($13) + ($2)|0);
   HEAP8[$3>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($14,$3);
   $15 = HEAP8[$4>>0]|0;
   $16 = ($15<<24>>24)<(0);
   if ($16) {
    $17 = ((($0)) + 4|0);
    HEAP32[$17>>2] = $2;
    break;
   } else {
    $18 = $2&255;
    HEAP8[$4>>0] = $18;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__211char_traitsIcE4moveEPcPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if (!($3)) {
  _memmove(($0|0),($1|0),($2|0))|0;
 }
 return ($0|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp;
 $9 = (-18 - ($1))|0;
 $10 = ($9>>>0)<($2>>>0);
 if ($10) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $11 = ((($0)) + 11|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ($12<<24>>24)<(0);
 if ($13) {
  $14 = HEAP32[$0>>2]|0;
  $25 = $14;
 } else {
  $25 = $0;
 }
 $15 = ($1>>>0)<(2147483623);
 if ($15) {
  $16 = (($2) + ($1))|0;
  $17 = $1 << 1;
  $18 = ($16>>>0)<($17>>>0);
  $$sroa$speculated = $18 ? $17 : $16;
  $19 = ($$sroa$speculated>>>0)<(11);
  $20 = (($$sroa$speculated) + 16)|0;
  $21 = $20 & -16;
  $phitmp = $19 ? 11 : $21;
  $22 = $phitmp;
 } else {
  $22 = -17;
 }
 $23 = (__Znwj($22)|0);
 $24 = ($4|0)==(0);
 if (!($24)) {
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($23,$25,$4)|0);
 }
 $26 = ($6|0)==(0);
 if (!($26)) {
  $27 = (($23) + ($4)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($27,$7,$6)|0);
 }
 $28 = (($3) - ($5))|0;
 $29 = (($28) - ($4))|0;
 $30 = ($29|0)==(0);
 if (!($30)) {
  $31 = (($23) + ($4)|0);
  $32 = (($31) + ($6)|0);
  $33 = (($25) + ($4)|0);
  $34 = (($33) + ($5)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($32,$34,$29)|0);
 }
 $35 = ($1|0)==(10);
 if (!($35)) {
  __ZdlPv($25);
 }
 HEAP32[$0>>2] = $23;
 $36 = $22 | -2147483648;
 $37 = ((($0)) + 8|0);
 HEAP32[$37>>2] = $36;
 $38 = (($28) + ($6))|0;
 $39 = ((($0)) + 4|0);
 HEAP32[$39>>2] = $38;
 $40 = (($23) + ($38)|0);
 HEAP8[$8>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($40,$8);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7reserveEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$016 = 0, $$01739 = 0, $$018$off037 = 0, $$1404753 = 0, $$141 = 0, $$phitmp$i = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, $phitmp$i21 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1>>>0)>(4294967279);
 if ($2) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $3 = ((($0)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 if ($5) {
  $6 = ((($0)) + 8|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = $7 & 2147483647;
  $phitmp$i = (($8) + -1)|0;
  $9 = ((($0)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $13 = $10;$18 = $phitmp$i;
 } else {
  $11 = $4&255;
  $13 = $11;$18 = 10;
 }
 $12 = ($13>>>0)>($1>>>0);
 $$sroa$speculated = $12 ? $13 : $1;
 $14 = ($$sroa$speculated>>>0)<(11);
 $15 = (($$sroa$speculated) + 16)|0;
 $16 = $15 & -16;
 $phitmp$i21 = (($16) + -1)|0;
 $$phitmp$i = $14 ? 10 : $phitmp$i21;
 $17 = ($$phitmp$i|0)==($18|0);
 L8: do {
  if (!($17)) {
   do {
    if ($14) {
     $28 = HEAP32[$0>>2]|0;
     if ($5) {
      $$01739 = $28;$$018$off037 = 0;$$141 = $0;
      label = 17;
     } else {
      $29 = $4&255;
      $30 = (($29) + 1)|0;
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$28,$30)|0);
      __ZdlPv($28);
      label = 19;
     }
    } else {
     $19 = ($$phitmp$i>>>0)>($18>>>0);
     $20 = (($$phitmp$i) + 1)|0;
     if ($19) {
      $21 = (__Znwj($20)|0);
      $$016 = $21;
     } else {
      __THREW__ = 0;
      $22 = (invoke_ii(988,($20|0))|0);
      $23 = __THREW__; __THREW__ = 0;
      $24 = $23&1;
      if ($24) {
       $25 = ___cxa_find_matching_catch_3(0|0)|0;
       $26 = tempRet0;
       (___cxa_begin_catch(($25|0))|0);
       ___cxa_end_catch();
       break L8;
      } else {
       $$016 = $22;
      }
     }
     if ($5) {
      $27 = HEAP32[$0>>2]|0;
      $$01739 = $27;$$018$off037 = 1;$$141 = $$016;
      label = 17;
      break;
     } else {
      $31 = $4&255;
      $32 = (($31) + 1)|0;
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$016,$0,$32)|0);
      $$1404753 = $$016;
      label = 18;
      break;
     }
    }
   } while(0);
   if ((label|0) == 17) {
    $33 = ((($0)) + 4|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (($34) + 1)|0;
    (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$141,$$01739,$35)|0);
    __ZdlPv($$01739);
    if ($$018$off037) {
     $$1404753 = $$141;
     label = 18;
    } else {
     label = 19;
    }
   }
   if ((label|0) == 18) {
    $36 = (($$phitmp$i) + 1)|0;
    $37 = $36 | -2147483648;
    $38 = ((($0)) + 8|0);
    HEAP32[$38>>2] = $37;
    $39 = ((($0)) + 4|0);
    HEAP32[$39>>2] = $13;
    HEAP32[$0>>2] = $$1404753;
    break;
   }
   else if ((label|0) == 19) {
    $40 = $13&255;
    HEAP8[$3>>0] = $40;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE6lengthEPKc($1)|0);
 $3 = (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6assignEPKcj($0,$1,$2)|0);
 return ($3|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6resizeEjc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp + 1|0;
 $4 = sp;
 $5 = ((($0)) + 11|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)<(0);
 if ($7) {
  $8 = ((($0)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $12 = $9;
 } else {
  $10 = $6&255;
  $12 = $10;
 }
 $11 = ($12>>>0)<($1>>>0);
 do {
  if ($11) {
   $13 = (($1) - ($12))|0;
   (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEjc($0,$13,$2)|0);
  } else {
   if ($7) {
    $14 = HEAP32[$0>>2]|0;
    $15 = (($14) + ($1)|0);
    HEAP8[$3>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($15,$3);
    $16 = ((($0)) + 4|0);
    HEAP32[$16>>2] = $1;
    break;
   } else {
    $17 = (($0) + ($1)|0);
    HEAP8[$4>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($17,$4);
    $18 = $1&255;
    HEAP8[$5>>0] = $18;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEjc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$pre = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($1|0)==(0);
 if (!($4)) {
  $5 = ((($0)) + 11|0);
  $6 = HEAP8[$5>>0]|0;
  $7 = ($6<<24>>24)<(0);
  if ($7) {
   $8 = ((($0)) + 8|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = $9 & 2147483647;
   $phitmp$i = (($10) + -1)|0;
   $11 = ((($0)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   $15 = $12;$16 = $phitmp$i;
  } else {
   $13 = $6&255;
   $15 = $13;$16 = 10;
  }
  $14 = (($16) - ($15))|0;
  $17 = ($14>>>0)<($1>>>0);
  $18 = (($15) + ($1))|0;
  if ($17) {
   $19 = (($18) - ($16))|0;
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($0,$16,$19,$15,$15,0,0);
   $$pre = HEAP8[$5>>0]|0;
   $21 = $$pre;
  } else {
   $21 = $6;
  }
  $20 = ($21<<24>>24)<(0);
  if ($20) {
   $22 = HEAP32[$0>>2]|0;
   $24 = $22;
  } else {
   $24 = $0;
  }
  $23 = (($24) + ($15)|0);
  (__ZNSt3__211char_traitsIcE6assignEPcjc($23,$1,$2)|0);
  $25 = HEAP8[$5>>0]|0;
  $26 = ($25<<24>>24)<(0);
  if ($26) {
   $27 = ((($0)) + 4|0);
   HEAP32[$27>>2] = $18;
  } else {
   $28 = $18&255;
   HEAP8[$5>>0] = $28;
  }
  $29 = (($24) + ($18)|0);
  HEAP8[$3>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($29,$3);
 }
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $7 = (-17 - ($1))|0;
 $8 = ($7>>>0)<($2>>>0);
 if ($8) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $9 = ((($0)) + 11|0);
 $10 = HEAP8[$9>>0]|0;
 $11 = ($10<<24>>24)<(0);
 if ($11) {
  $12 = HEAP32[$0>>2]|0;
  $23 = $12;
 } else {
  $23 = $0;
 }
 $13 = ($1>>>0)<(2147483623);
 if ($13) {
  $14 = (($2) + ($1))|0;
  $15 = $1 << 1;
  $16 = ($14>>>0)<($15>>>0);
  $$sroa$speculated = $16 ? $15 : $14;
  $17 = ($$sroa$speculated>>>0)<(11);
  $18 = (($$sroa$speculated) + 16)|0;
  $19 = $18 & -16;
  $phitmp = $17 ? 11 : $19;
  $20 = $phitmp;
 } else {
  $20 = -17;
 }
 $21 = (__Znwj($20)|0);
 $22 = ($4|0)==(0);
 if (!($22)) {
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($21,$23,$4)|0);
 }
 $24 = (($3) - ($5))|0;
 $25 = (($24) - ($4))|0;
 $26 = ($25|0)==(0);
 if (!($26)) {
  $27 = (($21) + ($4)|0);
  $28 = (($27) + ($6)|0);
  $29 = (($23) + ($4)|0);
  $30 = (($29) + ($5)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($28,$30,$25)|0);
 }
 $31 = ($1|0)==(10);
 if (!($31)) {
  __ZdlPv($23);
 }
 HEAP32[$0>>2] = $21;
 $32 = $20 | -2147483648;
 $33 = ((($0)) + 8|0);
 HEAP32[$33>>2] = $32;
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $7 = ((($0)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $10 = ((($0)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $14 = $11;$15 = $phitmp$i;
 } else {
  $12 = $5&255;
  $14 = $12;$15 = 10;
 }
 $13 = (($15) - ($14))|0;
 $16 = ($13>>>0)<($2>>>0);
 $17 = (($14) + ($2))|0;
 if ($16) {
  $27 = (($17) - ($15))|0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$15,$27,$14,$14,0,$2,$1);
 } else {
  $18 = ($2|0)==(0);
  if (!($18)) {
   if ($6) {
    $19 = HEAP32[$0>>2]|0;
    $21 = $19;
   } else {
    $21 = $0;
   }
   $20 = (($21) + ($14)|0);
   (__ZNSt3__211char_traitsIcE4copyEPcPKcj($20,$1,$2)|0);
   $22 = HEAP8[$4>>0]|0;
   $23 = ($22<<24>>24)<(0);
   if ($23) {
    $24 = ((($0)) + 4|0);
    HEAP32[$24>>2] = $17;
   } else {
    $25 = $17&255;
    HEAP8[$4>>0] = $25;
   }
   $26 = (($21) + ($17)|0);
   HEAP8[$3>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($26,$3);
  }
 }
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE6lengthEPKc($1)|0);
 $3 = (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj($0,$1,$2)|0);
 return ($3|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9push_backEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$018 = 0, $$019 = 0, $$pn = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 1|0;
 $3 = sp;
 HEAP8[$2>>0] = $1;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $8 = ((($0)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 2147483647;
  $11 = (($10) + -1)|0;
  $12 = ((($0)) + 4|0);
  $13 = HEAP32[$12>>2]|0;
  $$018 = $13;$$019 = $11;
 } else {
  $7 = $5&255;
  $$018 = $7;$$019 = 10;
 }
 $14 = ($$018|0)==($$019|0);
 if ($14) {
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9__grow_byEjjjjjj($0,$$019,1,$$019,$$019,0,0);
  $15 = HEAP8[$4>>0]|0;
  $16 = ($15<<24>>24)<(0);
  $17 = (($$019) + 1)|0;
  if ($16) {
   $23 = $17;
   label = 8;
  } else {
   $20 = $17;
   label = 7;
  }
 } else {
  $18 = (($$018) + 1)|0;
  if ($6) {
   $23 = $18;
   label = 8;
  } else {
   $20 = $18;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $19 = $20&255;
  HEAP8[$4>>0] = $19;
  $$pn = $0;
 }
 else if ((label|0) == 8) {
  $21 = HEAP32[$0>>2]|0;
  $22 = ((($0)) + 4|0);
  HEAP32[$22>>2] = $23;
  $$pn = $21;
 }
 $$0 = (($$pn) + ($$018)|0);
 __ZNSt3__211char_traitsIcE6assignERcRKc($$0,$2);
 $24 = ((($$0)) + 1|0);
 HEAP8[$3>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($24,$3);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 8|0);
 $2 = ((($1)) + 3|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)<(0);
 if ($4) {
  $5 = HEAP32[$0>>2]|0;
  __ZdlPv($5);
 }
 return;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKwj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 8|0);
 $5 = ((($4)) + 3|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)<(0);
 if ($7) {
  $8 = HEAP32[$4>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $11 = $phitmp$i;
 } else {
  $11 = 1;
 }
 $10 = ($11>>>0)<($2>>>0);
 do {
  if ($10) {
   if ($7) {
    $19 = ((($0)) + 4|0);
    $20 = HEAP32[$19>>2]|0;
    $23 = $20;
   } else {
    $21 = $6&255;
    $23 = $21;
   }
   $22 = (($2) - ($11))|0;
   __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE21__grow_by_and_replaceEjjjjjjPKw($0,$11,$22,$23,0,$23,$2,$1);
  } else {
   if ($7) {
    $12 = HEAP32[$0>>2]|0;
    $13 = $12;
   } else {
    $13 = $0;
   }
   (__ZNSt3__211char_traitsIwE4moveEPwPKwj($13,$1,$2)|0);
   $14 = (($13) + ($2<<2)|0);
   HEAP32[$3>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($14,$3);
   $15 = HEAP8[$5>>0]|0;
   $16 = ($15<<24>>24)<(0);
   if ($16) {
    $17 = ((($0)) + 4|0);
    HEAP32[$17>>2] = $2;
    break;
   } else {
    $18 = $2&255;
    HEAP8[$5>>0] = $18;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__211char_traitsIwE4moveEPwPKwj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($2|0)==(0);
 if ($3) {
  $7 = $0;
 } else {
  __THREW__ = 0;
  $4 = (invoke_iiii(1102,($0|0),($1|0),($2|0))|0);
  $5 = __THREW__; __THREW__ = 0;
  $6 = $5&1;
  if ($6) {
   $8 = ___cxa_find_matching_catch_3(0|0)|0;
   $9 = tempRet0;
   ___clang_call_terminate($8);
   // unreachable;
  } else {
   $7 = $0;
  }
 }
 return ($7|0);
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE21__grow_by_and_replaceEjjjjjjPKw($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp;
 $9 = (1073741806 - ($1))|0;
 $10 = ($9>>>0)<($2>>>0);
 if ($10) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $11 = ((($0)) + 8|0);
 $12 = ((($11)) + 3|0);
 $13 = HEAP8[$12>>0]|0;
 $14 = ($13<<24>>24)<(0);
 if ($14) {
  $15 = HEAP32[$0>>2]|0;
  $34 = $15;
 } else {
  $34 = $0;
 }
 $16 = ($1>>>0)<(536870887);
 if ($16) {
  $17 = (($2) + ($1))|0;
  $18 = $1 << 1;
  $19 = ($17>>>0)<($18>>>0);
  $$sroa$speculated = $19 ? $18 : $17;
  $20 = ($$sroa$speculated>>>0)<(2);
  $21 = (($$sroa$speculated) + 4)|0;
  $22 = $21 & -4;
  $23 = $20 ? 2 : $22;
  $24 = ($23>>>0)>(1073741823);
  if ($24) {
   $25 = (___cxa_allocate_exception(8)|0);
   __THREW__ = 0;
   invoke_vii(751,($25|0),(158021|0));
   $26 = __THREW__; __THREW__ = 0;
   $27 = $26&1;
   if ($27) {
    $28 = ___cxa_find_matching_catch_2()|0;
    $29 = tempRet0;
    ___cxa_free_exception(($25|0));
    ___resumeException($28|0);
    // unreachable;
   } else {
    HEAP32[$25>>2] = (22172);
    ___cxa_throw(($25|0),(2528|0),(713|0));
    // unreachable;
   }
  } else {
   $31 = $23;
  }
 } else {
  $31 = 1073741807;
 }
 $30 = $31 << 2;
 $32 = (__Znwj($30)|0);
 $33 = ($4|0)==(0);
 if (!($33)) {
  (__ZNSt3__211char_traitsIwE4copyEPwPKwj($32,$34,$4)|0);
 }
 $35 = ($6|0)==(0);
 if (!($35)) {
  $36 = (($32) + ($4<<2)|0);
  (__ZNSt3__211char_traitsIwE4copyEPwPKwj($36,$7,$6)|0);
 }
 $37 = (($3) - ($5))|0;
 $38 = (($37) - ($4))|0;
 $39 = ($38|0)==(0);
 if (!($39)) {
  $40 = (($34) + ($4<<2)|0);
  $41 = (($40) + ($5<<2)|0);
  $42 = (($32) + ($4<<2)|0);
  $43 = (($42) + ($6<<2)|0);
  (__ZNSt3__211char_traitsIwE4copyEPwPKwj($43,$41,$38)|0);
 }
 $44 = ($1|0)==(1);
 if (!($44)) {
  __ZdlPv($34);
 }
 HEAP32[$0>>2] = $32;
 $45 = $31 | -2147483648;
 HEAP32[$11>>2] = $45;
 $46 = (($37) + ($6))|0;
 $47 = ((($0)) + 4|0);
 HEAP32[$47>>2] = $46;
 $48 = (($32) + ($46<<2)|0);
 HEAP32[$8>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($48,$8);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE7reserveEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$01742 = 0, $$018$off040 = 0, $$1435056 = 0, $$144 = 0, $$phitmp$i = 0, $$sink = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $eh$lpad$body$indexZ2D = 0;
 var $phitmp$i = 0, $phitmp$i22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($1>>>0)>(1073741807);
 if ($2) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $3 = ((($0)) + 8|0);
 $4 = ((($3)) + 3|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $7 = HEAP32[$3>>2]|0;
  $8 = $7 & 2147483647;
  $phitmp$i = (($8) + -1)|0;
  $9 = ((($0)) + 4|0);
  $10 = HEAP32[$9>>2]|0;
  $13 = $10;$18 = $phitmp$i;
 } else {
  $11 = $5&255;
  $13 = $11;$18 = 1;
 }
 $12 = ($13>>>0)>($1>>>0);
 $$sroa$speculated = $12 ? $13 : $1;
 $14 = ($$sroa$speculated>>>0)<(2);
 $15 = (($$sroa$speculated) + 4)|0;
 $16 = $15 & -4;
 $phitmp$i22 = (($16) + -1)|0;
 $$phitmp$i = $14 ? 1 : $phitmp$i22;
 $17 = ($$phitmp$i|0)==($18|0);
 L8: do {
  if (!($17)) {
   do {
    if ($14) {
     $42 = HEAP32[$0>>2]|0;
     if ($6) {
      $$01742 = $42;$$018$off040 = 0;$$144 = $0;
      label = 26;
     } else {
      $43 = $5&255;
      $44 = (($43) + 1)|0;
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($0,$42,$44)|0);
      __ZdlPv($42);
      label = 28;
     }
    } else {
     $19 = ($$phitmp$i>>>0)>($18>>>0);
     $20 = (($$phitmp$i) + 1)|0;
     $21 = ($20>>>0)>(1073741823);
     L14: do {
      if ($19) {
       if (!($21)) {
        $27 = $20 << 2;
        $28 = (__Znwj($27)|0);
        $$sink = $28;
        break;
       }
       $22 = (___cxa_allocate_exception(8)|0);
       __THREW__ = 0;
       invoke_vii(751,($22|0),(158021|0));
       $23 = __THREW__; __THREW__ = 0;
       $24 = $23&1;
       if ($24) {
        $25 = ___cxa_find_matching_catch_2()|0;
        $26 = tempRet0;
        ___cxa_free_exception(($22|0));
        ___resumeException($25|0);
        // unreachable;
       } else {
        HEAP32[$22>>2] = (22172);
        ___cxa_throw(($22|0),(2528|0),(713|0));
        // unreachable;
       }
      } else {
       do {
        if ($21) {
         $29 = (___cxa_allocate_exception(8)|0);
         __THREW__ = 0;
         invoke_vii(751,($29|0),(158021|0));
         $30 = __THREW__; __THREW__ = 0;
         $31 = $30&1;
         if ($31) {
          $33 = ___cxa_find_matching_catch_3(0|0)|0;
          $34 = tempRet0;
          ___cxa_free_exception(($29|0));
          $eh$lpad$body$indexZ2D = $33;
          break;
         } else {
          HEAP32[$29>>2] = (22172);
          __THREW__ = 0;
          invoke_viii(946,($29|0),(2528|0),(713|0));
          $32 = __THREW__; __THREW__ = 0;
          label = 19;
          break;
         }
        } else {
         $35 = $20 << 2;
         __THREW__ = 0;
         $36 = (invoke_ii(988,($35|0))|0);
         $37 = __THREW__; __THREW__ = 0;
         $38 = $37&1;
         if ($38) {
          label = 19;
         } else {
          $$sink = $36;
          break L14;
         }
        }
       } while(0);
       if ((label|0) == 19) {
        $39 = ___cxa_find_matching_catch_3(0|0)|0;
        $40 = tempRet0;
        $eh$lpad$body$indexZ2D = $39;
       }
       (___cxa_begin_catch(($eh$lpad$body$indexZ2D|0))|0);
       ___cxa_end_catch();
       break L8;
      }
     } while(0);
     if ($6) {
      $41 = HEAP32[$0>>2]|0;
      $$01742 = $41;$$018$off040 = 1;$$144 = $$sink;
      label = 26;
      break;
     } else {
      $45 = $5&255;
      $46 = (($45) + 1)|0;
      (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$sink,$0,$46)|0);
      $$1435056 = $$sink;
      label = 27;
      break;
     }
    }
   } while(0);
   if ((label|0) == 26) {
    $47 = ((($0)) + 4|0);
    $48 = HEAP32[$47>>2]|0;
    $49 = (($48) + 1)|0;
    (__ZNSt3__211char_traitsIwE4copyEPwPKwj($$144,$$01742,$49)|0);
    __ZdlPv($$01742);
    if ($$018$off040) {
     $$1435056 = $$144;
     label = 27;
    } else {
     label = 28;
    }
   }
   if ((label|0) == 27) {
    $50 = (($$phitmp$i) + 1)|0;
    $51 = $50 | -2147483648;
    HEAP32[$3>>2] = $51;
    $52 = ((($0)) + 4|0);
    HEAP32[$52>>2] = $13;
    HEAP32[$0>>2] = $$1435056;
    break;
   }
   else if ((label|0) == 28) {
    $53 = $13&255;
    HEAP8[$4>>0] = $53;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIwE6lengthEPKw($1)|0);
 $3 = (__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6assignEPKwj($0,$1,$2)|0);
 return ($3|0);
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($0,$1,$2,$3,$4,$5,$6) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $7 = (1073741807 - ($1))|0;
 $8 = ($7>>>0)<($2>>>0);
 if ($8) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $9 = ((($0)) + 8|0);
 $10 = ((($9)) + 3|0);
 $11 = HEAP8[$10>>0]|0;
 $12 = ($11<<24>>24)<(0);
 if ($12) {
  $13 = HEAP32[$0>>2]|0;
  $32 = $13;
 } else {
  $32 = $0;
 }
 $14 = ($1>>>0)<(536870887);
 if ($14) {
  $15 = (($2) + ($1))|0;
  $16 = $1 << 1;
  $17 = ($15>>>0)<($16>>>0);
  $$sroa$speculated = $17 ? $16 : $15;
  $18 = ($$sroa$speculated>>>0)<(2);
  $19 = (($$sroa$speculated) + 4)|0;
  $20 = $19 & -4;
  $21 = $18 ? 2 : $20;
  $22 = ($21>>>0)>(1073741823);
  if ($22) {
   $23 = (___cxa_allocate_exception(8)|0);
   __THREW__ = 0;
   invoke_vii(751,($23|0),(158021|0));
   $24 = __THREW__; __THREW__ = 0;
   $25 = $24&1;
   if ($25) {
    $26 = ___cxa_find_matching_catch_2()|0;
    $27 = tempRet0;
    ___cxa_free_exception(($23|0));
    ___resumeException($26|0);
    // unreachable;
   } else {
    HEAP32[$23>>2] = (22172);
    ___cxa_throw(($23|0),(2528|0),(713|0));
    // unreachable;
   }
  } else {
   $29 = $21;
  }
 } else {
  $29 = 1073741807;
 }
 $28 = $29 << 2;
 $30 = (__Znwj($28)|0);
 $31 = ($4|0)==(0);
 if (!($31)) {
  (__ZNSt3__211char_traitsIwE4copyEPwPKwj($30,$32,$4)|0);
 }
 $33 = (($3) - ($5))|0;
 $34 = (($33) - ($4))|0;
 $35 = ($34|0)==(0);
 if (!($35)) {
  $36 = (($32) + ($4<<2)|0);
  $37 = (($36) + ($5<<2)|0);
  $38 = (($30) + ($4<<2)|0);
  $39 = (($38) + ($6<<2)|0);
  (__ZNSt3__211char_traitsIwE4copyEPwPKwj($39,$37,$34)|0);
 }
 $40 = ($1|0)==(1);
 if (!($40)) {
  __ZdlPv($32);
 }
 HEAP32[$0>>2] = $30;
 $41 = $29 | -2147483648;
 HEAP32[$9>>2] = $41;
 return;
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6appendEPKwj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 8|0);
 $5 = ((($4)) + 3|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)<(0);
 if ($7) {
  $8 = HEAP32[$4>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $10 = ((($0)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $14 = $11;$15 = $phitmp$i;
 } else {
  $12 = $6&255;
  $14 = $12;$15 = 1;
 }
 $13 = (($15) - ($14))|0;
 $16 = ($13>>>0)<($2>>>0);
 $17 = (($14) + ($2))|0;
 if ($16) {
  $27 = (($17) - ($15))|0;
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE21__grow_by_and_replaceEjjjjjjPKw($0,$15,$27,$14,$14,0,$2,$1);
 } else {
  $18 = ($2|0)==(0);
  if (!($18)) {
   if ($7) {
    $19 = HEAP32[$0>>2]|0;
    $21 = $19;
   } else {
    $21 = $0;
   }
   $20 = (($21) + ($14<<2)|0);
   (__ZNSt3__211char_traitsIwE4copyEPwPKwj($20,$1,$2)|0);
   $22 = HEAP8[$5>>0]|0;
   $23 = ($22<<24>>24)<(0);
   if ($23) {
    $24 = ((($0)) + 4|0);
    HEAP32[$24>>2] = $17;
   } else {
    $25 = $17&255;
    HEAP8[$5>>0] = $25;
   }
   $26 = (($21) + ($17<<2)|0);
   HEAP32[$3>>2] = 0;
   __ZNSt3__211char_traitsIwE6assignERwRKw($26,$3);
  }
 }
 STACKTOP = sp;return ($0|0);
}
function __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9push_backEw($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$018 = 0, $$019 = 0, $$pn = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $3 = sp;
 HEAP32[$2>>2] = $1;
 $4 = ((($0)) + 8|0);
 $5 = ((($4)) + 3|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)<(0);
 if ($7) {
  $9 = HEAP32[$4>>2]|0;
  $10 = $9 & 2147483647;
  $11 = (($10) + -1)|0;
  $12 = ((($0)) + 4|0);
  $13 = HEAP32[$12>>2]|0;
  $$018 = $13;$$019 = $11;
 } else {
  $8 = $6&255;
  $$018 = $8;$$019 = 1;
 }
 $14 = ($$018|0)==($$019|0);
 if ($14) {
  __ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9__grow_byEjjjjjj($0,$$019,1,$$019,$$019,0,0);
  $15 = HEAP8[$5>>0]|0;
  $16 = ($15<<24>>24)<(0);
  $17 = (($$019) + 1)|0;
  if ($16) {
   $23 = $17;
   label = 8;
  } else {
   $20 = $17;
   label = 7;
  }
 } else {
  $18 = (($$018) + 1)|0;
  if ($7) {
   $23 = $18;
   label = 8;
  } else {
   $20 = $18;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $19 = $20&255;
  HEAP8[$5>>0] = $19;
  $$pn = $0;
 }
 else if ((label|0) == 8) {
  $21 = HEAP32[$0>>2]|0;
  $22 = ((($0)) + 4|0);
  HEAP32[$22>>2] = $23;
  $$pn = $21;
 }
 $$0 = (($$pn) + ($$018<<2)|0);
 __ZNSt3__211char_traitsIwE6assignERwRKw($$0,$2);
 $24 = ((($$0)) + 4|0);
 HEAP32[$3>>2] = 0;
 __ZNSt3__211char_traitsIwE6assignERwRKw($24,$3);
 STACKTOP = sp;return;
}
function __ZNSt3__214error_categoryD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNKSt3__214error_category23default_error_conditionEi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $2;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = $1;
 return;
}
function __ZNKSt3__214error_category10equivalentEiRKNS_15error_conditionE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = HEAP32[$0>>2]|0;
 $5 = ((($4)) + 12|0);
 $6 = HEAP32[$5>>2]|0;
 FUNCTION_TABLE_viii[$6 & 2047]($3,$0,$1);
 $7 = ((($3)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = ((($2)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = ($8|0)==($10|0);
 if ($11) {
  $12 = HEAP32[$3>>2]|0;
  $13 = HEAP32[$2>>2]|0;
  $14 = ($12|0)==($13|0);
  $15 = $14;
 } else {
  $15 = 0;
 }
 STACKTOP = sp;return ($15|0);
}
function __ZNKSt3__214error_category10equivalentERKNS_10error_codeEi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($1)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ($4|0)==($0|0);
 $6 = HEAP32[$1>>2]|0;
 $7 = ($6|0)==($2|0);
 $8 = $5 & $7;
 return ($8|0);
}
function __ZNKSt3__212__do_message7messageEi($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__212_GLOBAL__N_113do_strerror_rEi($0,$2);
 return;
}
function __ZNSt3__212_GLOBAL__N_113do_strerror_rEi($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $$0$i$i12 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1040|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(1040|0);
 $vararg_buffer = sp;
 $2 = sp + 1033|0;
 $3 = sp + 1032|0;
 $4 = sp + 8|0;
 $5 = (___errno_location()|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (_strerror_r($1,$4,1024)|0);
 switch ($7|0) {
 case 0:  {
  ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
  $24 = (__ZNSt3__211char_traitsIcE6lengthEPKc($4)|0);
  $25 = ($24>>>0)>(4294967279);
  if ($25) {
   __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
   // unreachable;
  }
  $26 = ($24>>>0)<(11);
  if ($26) {
   $27 = $24&255;
   $28 = ((($0)) + 11|0);
   HEAP8[$28>>0] = $27;
   $$0$i$i12 = $0;
  } else {
   $29 = (($24) + 16)|0;
   $30 = $29 & -16;
   $31 = (__Znwj($30)|0);
   HEAP32[$0>>2] = $31;
   $32 = $30 | -2147483648;
   $33 = ((($0)) + 8|0);
   HEAP32[$33>>2] = $32;
   $34 = ((($0)) + 4|0);
   HEAP32[$34>>2] = $24;
   $$0$i$i12 = $31;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i12,$4,$24)|0);
  $35 = (($$0$i$i12) + ($24)|0);
  HEAP8[$2>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($35,$2);
  break;
 }
 case -1:  {
  $8 = (___errno_location()|0);
  $9 = HEAP32[$8>>2]|0;
  $11 = $9;
  label = 3;
  break;
 }
 default: {
  $11 = $7;
  label = 3;
 }
 }
 L11: do {
  if ((label|0) == 3) {
   $10 = (___errno_location()|0);
   HEAP32[$10>>2] = $6;
   switch ($11|0) {
   case 22:  {
    HEAP32[$vararg_buffer>>2] = $1;
    (_snprintf($4,1024,158089,$vararg_buffer)|0);
    ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
    $12 = (__ZNSt3__211char_traitsIcE6lengthEPKc($4)|0);
    $13 = ($12>>>0)>(4294967279);
    if ($13) {
     __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
     // unreachable;
    }
    $14 = ($12>>>0)<(11);
    if ($14) {
     $15 = $12&255;
     $16 = ((($0)) + 11|0);
     HEAP8[$16>>0] = $15;
     $$0$i$i = $0;
    } else {
     $17 = (($12) + 16)|0;
     $18 = $17 & -16;
     $19 = (__Znwj($18)|0);
     HEAP32[$0>>2] = $19;
     $20 = $18 | -2147483648;
     $21 = ((($0)) + 8|0);
     HEAP32[$21>>2] = $20;
     $22 = ((($0)) + 4|0);
     HEAP32[$22>>2] = $12;
     $$0$i$i = $19;
    }
    (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,$4,$12)|0);
    $23 = (($$0$i$i) + ($12)|0);
    HEAP8[$3>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($23,$3);
    break L11;
    break;
   }
   case 34:  {
    _abort();
    // unreachable;
    break;
   }
   default: {
    ___assert_fail((158106|0),(158126|0),99,(158193|0));
    // unreachable;
   }
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNSt3__212system_errorD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($0);
 return;
}
function __ZNSt3__212system_errorD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__212system_errorD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt3__210error_code7messageEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($4)) + 24|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = HEAP32[$1>>2]|0;
 FUNCTION_TABLE_viii[$6 & 2047]($0,$3,$7);
 return;
}
function __ZNSt3__212system_error6__initERKNS_10error_codeENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = HEAP32[$1>>2]|0;
 $5 = ($4|0)==(0);
 do {
  if (!($5)) {
   $6 = ((($2)) + 11|0);
   $7 = HEAP8[$6>>0]|0;
   $8 = ($7<<24>>24)<(0);
   if ($8) {
    $9 = ((($2)) + 4|0);
    $10 = HEAP32[$9>>2]|0;
    $13 = $10;
   } else {
    $11 = $7&255;
    $13 = $11;
   }
   $12 = ($13|0)==(0);
   if (!($12)) {
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKc($2,158278)|0);
   }
   __ZNKSt3__210error_code7messageEv($3,$1);
   $14 = ((($3)) + 11|0);
   $15 = HEAP8[$14>>0]|0;
   $16 = ($15<<24>>24)<(0);
   $17 = HEAP32[$3>>2]|0;
   $18 = $16 ? $17 : $3;
   $19 = ((($3)) + 4|0);
   $20 = HEAP32[$19>>2]|0;
   $21 = $15&255;
   $22 = $16 ? $20 : $21;
   __THREW__ = 0;
   (invoke_iiii(1011,($2|0),($18|0),($22|0))|0);
   $23 = __THREW__; __THREW__ = 0;
   $24 = $23&1;
   if ($24) {
    $25 = ___cxa_find_matching_catch_2()|0;
    $26 = tempRet0;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3);
    ___resumeException($25|0);
    // unreachable;
   } else {
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($3);
    break;
   }
  }
 } while(0);
 ;HEAP32[$0>>2]=HEAP32[$2>>2]|0;HEAP32[$0+4>>2]=HEAP32[$2+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$2+8>>2]|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $27 = (($2) + ($$0$i$i<<2)|0);
  HEAP32[$27>>2] = 0;
  $28 = (($$0$i$i) + 1)|0;
  $$0$i$i = $28;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__212system_errorC2ENS_10error_codeEPKc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$03 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 24|0;
 $4 = sp + 12|0;
 $5 = sp;
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc($2)|0);
 $7 = ($6>>>0)>(4294967279);
 if ($7) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($5);
  // unreachable;
 }
 $8 = ($6>>>0)<(11);
 if ($8) {
  $9 = $6&255;
  $10 = ((($5)) + 11|0);
  HEAP8[$10>>0] = $9;
  $$0$i$i = $5;
 } else {
  $11 = (($6) + 16)|0;
  $12 = $11 & -16;
  $13 = (__Znwj($12)|0);
  HEAP32[$5>>2] = $13;
  $14 = $12 | -2147483648;
  $15 = ((($5)) + 8|0);
  HEAP32[$15>>2] = $14;
  $16 = ((($5)) + 4|0);
  HEAP32[$16>>2] = $6;
  $$0$i$i = $13;
 }
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,$2,$6)|0);
 $17 = (($$0$i$i) + ($6)|0);
 HEAP8[$3>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($17,$3);
 __THREW__ = 0;
 invoke_viii(1103,($4|0),($1|0),($5|0));
 $18 = __THREW__; __THREW__ = 0;
 $19 = $18&1;
 do {
  if ($19) {
   $33 = ___cxa_find_matching_catch_2()|0;
   $34 = tempRet0;
   $$0 = $34;$$03 = $33;
  } else {
   __THREW__ = 0;
   invoke_vii(1104,($0|0),($4|0));
   $20 = __THREW__; __THREW__ = 0;
   $21 = $20&1;
   if ($21) {
    $35 = ___cxa_find_matching_catch_2()|0;
    $36 = tempRet0;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4);
    $$0 = $36;$$03 = $35;
    break;
   } else {
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($4);
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
    HEAP32[$0>>2] = (22004);
    $22 = ((($0)) + 8|0);
    $23 = $1;
    $24 = $23;
    $25 = HEAP32[$24>>2]|0;
    $26 = (($23) + 4)|0;
    $27 = $26;
    $28 = HEAP32[$27>>2]|0;
    $29 = $22;
    $30 = $29;
    HEAP32[$30>>2] = $25;
    $31 = (($29) + 4)|0;
    $32 = $31;
    HEAP32[$32>>2] = $28;
    STACKTOP = sp;return;
   }
  }
 } while(0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev($5);
 ___resumeException($$03|0);
 // unreachable;
}
function __ZL25default_terminate_handlerv() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer10 = 0, $vararg_buffer3 = 0;
 var $vararg_buffer7 = 0, $vararg_ptr1 = 0, $vararg_ptr2 = 0, $vararg_ptr6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $vararg_buffer10 = sp + 32|0;
 $vararg_buffer7 = sp + 24|0;
 $vararg_buffer3 = sp + 16|0;
 $vararg_buffer = sp;
 $0 = sp + 36|0;
 $1 = (___cxa_get_globals_fast()|0);
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = HEAP32[$1>>2]|0;
  $4 = ($3|0)==(0|0);
  if (!($4)) {
   $5 = ((($3)) + 80|0);
   $6 = ((($3)) + 48|0);
   $7 = $6;
   $8 = $7;
   $9 = HEAP32[$8>>2]|0;
   $10 = (($7) + 4)|0;
   $11 = $10;
   $12 = HEAP32[$11>>2]|0;
   $13 = $9 & -256;
   $14 = ($13|0)==(1126902528);
   $15 = ($12|0)==(1129074247);
   $16 = $14 & $15;
   if (!($16)) {
    $36 = HEAP32[5505]|0;
    HEAP32[$vararg_buffer7>>2] = $36;
    _abort_message(158367,$vararg_buffer7);
    // unreachable;
   }
   $17 = ($9|0)==(1126902529);
   $18 = ($12|0)==(1129074247);
   $19 = $17 & $18;
   if ($19) {
    $20 = ((($3)) + 44|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = $21;
   } else {
    $22 = $5;
   }
   HEAP32[$0>>2] = $22;
   $23 = HEAP32[$3>>2]|0;
   $24 = ((($23)) + 4|0);
   $25 = HEAP32[$24>>2]|0;
   $26 = HEAP32[604]|0;
   $27 = ((($26)) + 16|0);
   $28 = HEAP32[$27>>2]|0;
   $29 = (FUNCTION_TABLE_iiii[$28 & 2047](2416,$23,$0)|0);
   $30 = HEAP32[5505]|0;
   if ($29) {
    $31 = HEAP32[$0>>2]|0;
    $32 = HEAP32[$31>>2]|0;
    $33 = ((($32)) + 8|0);
    $34 = HEAP32[$33>>2]|0;
    $35 = (FUNCTION_TABLE_ii[$34 & 2047]($31)|0);
    HEAP32[$vararg_buffer>>2] = $30;
    $vararg_ptr1 = ((($vararg_buffer)) + 4|0);
    HEAP32[$vararg_ptr1>>2] = $25;
    $vararg_ptr2 = ((($vararg_buffer)) + 8|0);
    HEAP32[$vararg_ptr2>>2] = $35;
    _abort_message(158281,$vararg_buffer);
    // unreachable;
   } else {
    HEAP32[$vararg_buffer3>>2] = $30;
    $vararg_ptr6 = ((($vararg_buffer3)) + 4|0);
    HEAP32[$vararg_ptr6>>2] = $25;
    _abort_message(158326,$vararg_buffer3);
    // unreachable;
   }
  }
 }
 _abort_message(158405,$vararg_buffer10);
 // unreachable;
}
function ___cxa_get_globals_fast() {
 var $0 = 0, $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = (_pthread_once((167356|0),(1105|0))|0);
 $1 = ($0|0)==(0);
 if ($1) {
  $2 = HEAP32[41840]|0;
  $3 = (_pthread_getspecific(($2|0))|0);
  STACKTOP = sp;return ($3|0);
 } else {
  _abort_message(158556,$vararg_buffer);
  // unreachable;
 }
 return (0)|0;
}
function _abort_message($0,$varargs) {
 $0 = $0|0;
 $varargs = $varargs|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 HEAP32[$1>>2] = $varargs;
 $2 = HEAP32[3657]|0;
 (_vfprintf($2,$0,$1)|0);
 (_fputc(10,$2)|0);
 _abort();
 // unreachable;
}
function __ZN10__cxxabiv116__shim_type_infoD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN10__cxxabiv117__class_type_infoD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN10__cxxabiv116__shim_type_infoD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNK10__cxxabiv116__shim_type_info5noop1Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNK10__cxxabiv116__shim_type_info5noop2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNK10__cxxabiv117__class_type_info9can_catchEPKNS_16__shim_type_infoERPv($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$2 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp;
 $4 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$1,0)|0);
 if ($4) {
  $$2 = 1;
 } else {
  $5 = ($1|0)==(0|0);
  if ($5) {
   $$2 = 0;
  } else {
   $6 = (___dynamic_cast($1,2440,2424,0)|0);
   $7 = ($6|0)==(0|0);
   if ($7) {
    $$2 = 0;
   } else {
    $8 = ((($3)) + 4|0);
    dest=$8; stop=dest+52|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));
    HEAP32[$3>>2] = $6;
    $9 = ((($3)) + 8|0);
    HEAP32[$9>>2] = $0;
    $10 = ((($3)) + 12|0);
    HEAP32[$10>>2] = -1;
    $11 = ((($3)) + 48|0);
    HEAP32[$11>>2] = 1;
    $12 = HEAP32[$6>>2]|0;
    $13 = ((($12)) + 28|0);
    $14 = HEAP32[$13>>2]|0;
    $15 = HEAP32[$2>>2]|0;
    FUNCTION_TABLE_viiii[$14 & 2047]($6,$3,$15,1);
    $16 = ((($3)) + 24|0);
    $17 = HEAP32[$16>>2]|0;
    $18 = ($17|0)==(1);
    if ($18) {
     $19 = ((($3)) + 16|0);
     $20 = HEAP32[$19>>2]|0;
     HEAP32[$2>>2] = $20;
     $$0 = 1;
    } else {
     $$0 = 0;
    }
    $$2 = $$0;
   }
  }
 }
 STACKTOP = sp;return ($$2|0);
}
function __ZNK10__cxxabiv117__class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$7,$5)|0);
 if ($8) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$1,$2,$3,$4);
 }
 return;
}
function __ZNK10__cxxabiv117__class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($1)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$6,$4)|0);
 do {
  if ($7) {
   __ZNK10__cxxabiv117__class_type_info29process_static_type_below_dstEPNS_19__dynamic_cast_infoEPKvi(0,$1,$2,$3);
  } else {
   $8 = HEAP32[$1>>2]|0;
   $9 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$8,$4)|0);
   if ($9) {
    $10 = ((($1)) + 16|0);
    $11 = HEAP32[$10>>2]|0;
    $12 = ($11|0)==($2|0);
    $13 = ((($1)) + 32|0);
    if (!($12)) {
     $14 = ((($1)) + 20|0);
     $15 = HEAP32[$14>>2]|0;
     $16 = ($15|0)==($2|0);
     if (!($16)) {
      HEAP32[$13>>2] = $3;
      HEAP32[$14>>2] = $2;
      $18 = ((($1)) + 40|0);
      $19 = HEAP32[$18>>2]|0;
      $20 = (($19) + 1)|0;
      HEAP32[$18>>2] = $20;
      $21 = ((($1)) + 36|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($22|0)==(1);
      if ($23) {
       $24 = ((($1)) + 24|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = ($25|0)==(2);
       if ($26) {
        $27 = ((($1)) + 54|0);
        HEAP8[$27>>0] = 1;
       }
      }
      $28 = ((($1)) + 44|0);
      HEAP32[$28>>2] = 4;
      break;
     }
    }
    $17 = ($3|0)==(1);
    if ($17) {
     HEAP32[$13>>2] = 1;
    }
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv117__class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, $6 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$5,0)|0);
 if ($6) {
  __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$1,$2,$3);
 }
 return;
}
function __ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==($1|0);
 return ($3|0);
}
function __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($1)) + 16|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==(0|0);
 $7 = ((($1)) + 36|0);
 $8 = ((($1)) + 24|0);
 do {
  if ($6) {
   HEAP32[$4>>2] = $2;
   HEAP32[$8>>2] = $3;
   HEAP32[$7>>2] = 1;
  } else {
   $9 = ($5|0)==($2|0);
   if (!($9)) {
    $12 = HEAP32[$7>>2]|0;
    $13 = (($12) + 1)|0;
    HEAP32[$7>>2] = $13;
    HEAP32[$8>>2] = 2;
    $14 = ((($1)) + 54|0);
    HEAP8[$14>>0] = 1;
    break;
   }
   $10 = HEAP32[$8>>2]|0;
   $11 = ($10|0)==(2);
   if ($11) {
    HEAP32[$8>>2] = $3;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv117__class_type_info29process_static_type_below_dstEPNS_19__dynamic_cast_infoEPKvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($1)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = ($5|0)==($2|0);
 if ($6) {
  $7 = ((($1)) + 28|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8|0)==(1);
  if (!($9)) {
   HEAP32[$7>>2] = $3;
  }
 }
 return;
}
function __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond22 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($1)) + 53|0);
 HEAP8[$5>>0] = 1;
 $6 = ((($1)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = ($7|0)==($3|0);
 do {
  if ($8) {
   $9 = ((($1)) + 52|0);
   HEAP8[$9>>0] = 1;
   $10 = ((($1)) + 16|0);
   $11 = HEAP32[$10>>2]|0;
   $12 = ($11|0)==(0|0);
   $13 = ((($1)) + 54|0);
   $14 = ((($1)) + 48|0);
   $15 = ((($1)) + 24|0);
   $16 = ((($1)) + 36|0);
   if ($12) {
    HEAP32[$10>>2] = $2;
    HEAP32[$15>>2] = $4;
    HEAP32[$16>>2] = 1;
    $17 = HEAP32[$14>>2]|0;
    $18 = ($17|0)==(1);
    $19 = ($4|0)==(1);
    $or$cond = $18 & $19;
    if (!($or$cond)) {
     break;
    }
    HEAP8[$13>>0] = 1;
    break;
   }
   $20 = ($11|0)==($2|0);
   if (!($20)) {
    $27 = HEAP32[$16>>2]|0;
    $28 = (($27) + 1)|0;
    HEAP32[$16>>2] = $28;
    HEAP8[$13>>0] = 1;
    break;
   }
   $21 = HEAP32[$15>>2]|0;
   $22 = ($21|0)==(2);
   if ($22) {
    HEAP32[$15>>2] = $4;
    $26 = $4;
   } else {
    $26 = $21;
   }
   $23 = HEAP32[$14>>2]|0;
   $24 = ($23|0)==(1);
   $25 = ($26|0)==(1);
   $or$cond22 = $24 & $25;
   if ($or$cond22) {
    HEAP8[$13>>0] = 1;
   }
  }
 } while(0);
 return;
}
function ___dynamic_cast($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$ = 0, $$0 = 0, $$33 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond28 = 0, $or$cond30 = 0, $or$cond32 = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $4 = sp;
 $5 = HEAP32[$0>>2]|0;
 $6 = ((($5)) + -8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (($0) + ($7)|0);
 $9 = ((($5)) + -4|0);
 $10 = HEAP32[$9>>2]|0;
 HEAP32[$4>>2] = $2;
 $11 = ((($4)) + 4|0);
 HEAP32[$11>>2] = $0;
 $12 = ((($4)) + 8|0);
 HEAP32[$12>>2] = $1;
 $13 = ((($4)) + 12|0);
 HEAP32[$13>>2] = $3;
 $14 = ((($4)) + 16|0);
 $15 = ((($4)) + 20|0);
 $16 = ((($4)) + 24|0);
 $17 = ((($4)) + 28|0);
 $18 = ((($4)) + 32|0);
 $19 = ((($4)) + 40|0);
 dest=$14; stop=dest+36|0; do { HEAP32[dest>>2]=0|0; dest=dest+4|0; } while ((dest|0) < (stop|0));HEAP16[$14+36>>1]=0|0;HEAP8[$14+38>>0]=0|0;
 $20 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($10,$2,0)|0);
 L1: do {
  if ($20) {
   $21 = ((($4)) + 48|0);
   HEAP32[$21>>2] = 1;
   $22 = HEAP32[$10>>2]|0;
   $23 = ((($22)) + 20|0);
   $24 = HEAP32[$23>>2]|0;
   FUNCTION_TABLE_viiiiii[$24 & 1023]($10,$4,$8,$8,1,0);
   $25 = HEAP32[$16>>2]|0;
   $26 = ($25|0)==(1);
   $$ = $26 ? $8 : 0;
   $$0 = $$;
  } else {
   $27 = ((($4)) + 36|0);
   $28 = HEAP32[$10>>2]|0;
   $29 = ((($28)) + 24|0);
   $30 = HEAP32[$29>>2]|0;
   FUNCTION_TABLE_viiiii[$30 & 1023]($10,$4,$8,1,0);
   $31 = HEAP32[$27>>2]|0;
   switch ($31|0) {
   case 0:  {
    $32 = HEAP32[$19>>2]|0;
    $33 = ($32|0)==(1);
    $34 = HEAP32[$17>>2]|0;
    $35 = ($34|0)==(1);
    $or$cond = $33 & $35;
    $36 = HEAP32[$18>>2]|0;
    $37 = ($36|0)==(1);
    $or$cond28 = $or$cond & $37;
    $38 = HEAP32[$15>>2]|0;
    $$33 = $or$cond28 ? $38 : 0;
    $$0 = $$33;
    break L1;
    break;
   }
   case 1:  {
    break;
   }
   default: {
    $$0 = 0;
    break L1;
   }
   }
   $39 = HEAP32[$16>>2]|0;
   $40 = ($39|0)==(1);
   if (!($40)) {
    $41 = HEAP32[$19>>2]|0;
    $42 = ($41|0)==(0);
    $43 = HEAP32[$17>>2]|0;
    $44 = ($43|0)==(1);
    $or$cond30 = $42 & $44;
    $45 = HEAP32[$18>>2]|0;
    $46 = ($45|0)==(1);
    $or$cond32 = $or$cond30 & $46;
    if (!($or$cond32)) {
     $$0 = 0;
     break;
    }
   }
   $47 = HEAP32[$14>>2]|0;
   $$0 = $47;
  }
 } while(0);
 STACKTOP = sp;return ($$0|0);
}
function __ZN10__cxxabiv120__si_class_type_infoD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN10__cxxabiv116__shim_type_infoD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNK10__cxxabiv120__si_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$7,$5)|0);
 if ($8) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$1,$2,$3,$4);
 } else {
  $9 = ((($0)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($11)) + 20|0);
  $13 = HEAP32[$12>>2]|0;
  FUNCTION_TABLE_viiiiii[$13 & 1023]($10,$1,$2,$3,$4,$5);
 }
 return;
}
function __ZNK10__cxxabiv120__si_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$037$off038 = 0, $$037$off039 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($1)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$6,$4)|0);
 do {
  if ($7) {
   __ZNK10__cxxabiv117__class_type_info29process_static_type_below_dstEPNS_19__dynamic_cast_infoEPKvi(0,$1,$2,$3);
  } else {
   $8 = HEAP32[$1>>2]|0;
   $9 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$8,$4)|0);
   $10 = ((($0)) + 8|0);
   if (!($9)) {
    $41 = HEAP32[$10>>2]|0;
    $42 = HEAP32[$41>>2]|0;
    $43 = ((($42)) + 24|0);
    $44 = HEAP32[$43>>2]|0;
    FUNCTION_TABLE_viiiii[$44 & 1023]($41,$1,$2,$3,$4);
    break;
   }
   $11 = ((($1)) + 16|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($12|0)==($2|0);
   $14 = ((($1)) + 32|0);
   if (!($13)) {
    $15 = ((($1)) + 20|0);
    $16 = HEAP32[$15>>2]|0;
    $17 = ($16|0)==($2|0);
    if (!($17)) {
     HEAP32[$14>>2] = $3;
     $19 = ((($1)) + 44|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = ($20|0)==(4);
     if ($21) {
      break;
     }
     $22 = ((($1)) + 52|0);
     HEAP8[$22>>0] = 0;
     $23 = ((($1)) + 53|0);
     HEAP8[$23>>0] = 0;
     $24 = HEAP32[$10>>2]|0;
     $25 = HEAP32[$24>>2]|0;
     $26 = ((($25)) + 20|0);
     $27 = HEAP32[$26>>2]|0;
     FUNCTION_TABLE_viiiiii[$27 & 1023]($24,$1,$2,$2,1,$4);
     $28 = HEAP8[$23>>0]|0;
     $29 = ($28<<24>>24)==(0);
     if ($29) {
      $$037$off038 = 4;
      label = 11;
     } else {
      $30 = HEAP8[$22>>0]|0;
      $not$ = ($30<<24>>24)==(0);
      if ($not$) {
       $$037$off038 = 3;
       label = 11;
      } else {
       $$037$off039 = 3;
      }
     }
     if ((label|0) == 11) {
      HEAP32[$15>>2] = $2;
      $31 = ((($1)) + 40|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = (($32) + 1)|0;
      HEAP32[$31>>2] = $33;
      $34 = ((($1)) + 36|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ($35|0)==(1);
      if ($36) {
       $37 = ((($1)) + 24|0);
       $38 = HEAP32[$37>>2]|0;
       $39 = ($38|0)==(2);
       if ($39) {
        $40 = ((($1)) + 54|0);
        HEAP8[$40>>0] = 1;
        $$037$off039 = $$037$off038;
       } else {
        $$037$off039 = $$037$off038;
       }
      } else {
       $$037$off039 = $$037$off038;
      }
     }
     HEAP32[$19>>2] = $$037$off039;
     break;
    }
   }
   $18 = ($3|0)==(1);
   if ($18) {
    HEAP32[$14>>2] = 1;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv120__si_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$5,0)|0);
 if ($6) {
  __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$1,$2,$3);
 } else {
  $7 = ((($0)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = HEAP32[$8>>2]|0;
  $10 = ((($9)) + 28|0);
  $11 = HEAP32[$10>>2]|0;
  FUNCTION_TABLE_viiii[$11 & 2047]($8,$1,$2,$3);
 }
 return;
}
function __ZNSt9type_infoD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_110construct_Ev() {
 var $0 = 0, $1 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 $0 = (_pthread_key_create((167360|0),(1106|0))|0);
 $1 = ($0|0)==(0);
 if ($1) {
  STACKTOP = sp;return;
 } else {
  _abort_message(158605,$vararg_buffer);
  // unreachable;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_19destruct_EPv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $vararg_buffer = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer = sp;
 _free($0);
 $1 = HEAP32[41840]|0;
 $2 = (_pthread_setspecific(($1|0),(0|0))|0);
 $3 = ($2|0)==(0);
 if ($3) {
  STACKTOP = sp;return;
 } else {
  _abort_message(158655,$vararg_buffer);
  // unreachable;
 }
}
function __ZSt9terminatev() {
 var $0 = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __THREW__ = 0;
 $0 = (invoke_i(1107)|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if ($2) {
  $20 = ___cxa_find_matching_catch_3(0|0)|0;
  $21 = tempRet0;
  ___clang_call_terminate($20);
  // unreachable;
 }
 $3 = ($0|0)==(0|0);
 if (!($3)) {
  $4 = HEAP32[$0>>2]|0;
  $5 = ($4|0)==(0|0);
  if (!($5)) {
   $6 = ((($4)) + 48|0);
   $7 = $6;
   $8 = $7;
   $9 = HEAP32[$8>>2]|0;
   $10 = (($7) + 4)|0;
   $11 = $10;
   $12 = HEAP32[$11>>2]|0;
   $13 = $9 & -256;
   $14 = ($13|0)==(1126902528);
   $15 = ($12|0)==(1129074247);
   $16 = $14 & $15;
   if ($16) {
    $17 = ((($4)) + 12|0);
    $18 = HEAP32[$17>>2]|0;
    __ZSt11__terminatePFvvE($18);
    // unreachable;
   }
  }
 }
 $19 = (__ZSt13get_terminatev()|0);
 __ZSt11__terminatePFvvE($19);
 // unreachable;
}
function __ZSt11__terminatePFvvE($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $vararg_buffer = 0, $vararg_buffer1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $vararg_buffer1 = sp + 8|0;
 $vararg_buffer = sp;
 __THREW__ = 0;
 invoke_v($0|0);
 $1 = __THREW__; __THREW__ = 0;
 $2 = $1&1;
 if (!($2)) {
  __THREW__ = 0;
  invoke_vii(1108,(158708|0),($vararg_buffer|0));
  $3 = __THREW__; __THREW__ = 0;
 }
 $4 = ___cxa_find_matching_catch_3(0|0)|0;
 $5 = tempRet0;
 (___cxa_begin_catch(($4|0))|0);
 __THREW__ = 0;
 invoke_vii(1108,(158748|0),($vararg_buffer1|0));
 $6 = __THREW__; __THREW__ = 0;
 $7 = ___cxa_find_matching_catch_3(0|0)|0;
 $8 = tempRet0;
 __THREW__ = 0;
 invoke_v(738);
 $9 = __THREW__; __THREW__ = 0;
 $10 = $9&1;
 if ($10) {
  $11 = ___cxa_find_matching_catch_3(0|0)|0;
  $12 = tempRet0;
  ___clang_call_terminate($11);
  // unreachable;
 } else {
  ___clang_call_terminate($7);
  // unreachable;
 }
}
function __ZSt13get_terminatev() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[5504]|0;HEAP32[5504] = (($0+0)|0);
 $1 = $0;
 return ($1|0);
}
function __ZNSt9bad_allocD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt9bad_allocD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt9bad_allocD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt9bad_alloc4whatEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (158798|0);
}
function __ZNSt9exceptionD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt11logic_errorD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22132);
 $1 = ((($0)) + 4|0);
 __ZNSt3__218__libcpp_refstringD2Ev($1);
 return;
}
function __ZNSt11logic_errorD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt11logic_error4whatEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = (__ZNKSt3__218__libcpp_refstring5c_strEv($1)|0);
 return ($2|0);
}
function __ZNKSt3__218__libcpp_refstring5c_strEv($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 return ($1|0);
}
function __ZNSt3__218__libcpp_refstringD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNKSt3__218__libcpp_refstring15__uses_refcountEv($0)|0);
 if ($1) {
  $2 = HEAP32[$0>>2]|0;
  $3 = (__ZNSt3__215__refstring_imp12_GLOBAL__N_113rep_from_dataEPKc_550($2)|0);
  $4 = ((($3)) + 8|0);
  $5 = HEAP32[$4>>2]|0;HEAP32[$4>>2] = (($5+-1)|0);
  $6 = (($5) + -1)|0;
  $7 = ($6|0)<(0);
  if ($7) {
   __ZdlPv($3);
  }
 }
 return;
}
function __ZNSt3__215__refstring_imp12_GLOBAL__N_113rep_from_dataEPKc_550($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + -12|0);
 return ($1|0);
}
function __ZNSt13runtime_errorD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22152);
 $1 = ((($0)) + 4|0);
 __ZNSt3__218__libcpp_refstringD2Ev($1);
 return;
}
function __ZNSt13runtime_errorD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt13runtime_errorD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt13runtime_error4whatEv($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = (__ZNKSt3__218__libcpp_refstring5c_strEv($1)|0);
 return ($2|0);
}
function __ZNSt12length_errorD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt12out_of_rangeD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt11logic_errorD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNSt8bad_castD2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt8bad_castD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt8bad_castD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNKSt8bad_cast4whatEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return (158894|0);
}
function __ZN10__cxxabiv121__vmi_class_type_infoD0Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 __ZN10__cxxabiv116__shim_type_infoD2Ev($0);
 __ZdlPv($0);
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = ((($1)) + 8|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$7,$5)|0);
 if ($8) {
  __ZNK10__cxxabiv117__class_type_info29process_static_type_above_dstEPNS_19__dynamic_cast_infoEPKvS4_i(0,$1,$2,$3,$4);
 } else {
  $9 = ((($1)) + 52|0);
  $10 = HEAP8[$9>>0]|0;
  $11 = ((($1)) + 53|0);
  $12 = HEAP8[$11>>0]|0;
  $13 = ((($0)) + 16|0);
  $14 = ((($0)) + 12|0);
  $15 = HEAP32[$14>>2]|0;
  $16 = (((($0)) + 16|0) + ($15<<3)|0);
  HEAP8[$9>>0] = 0;
  HEAP8[$11>>0] = 0;
  __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($13,$1,$2,$3,$4,$5);
  $17 = ($15|0)>(1);
  L4: do {
   if ($17) {
    $18 = ((($0)) + 24|0);
    $19 = ((($1)) + 24|0);
    $20 = ((($1)) + 54|0);
    $21 = ((($0)) + 8|0);
    $$0 = $18;
    while(1) {
     $22 = HEAP8[$20>>0]|0;
     $23 = ($22<<24>>24)==(0);
     if (!($23)) {
      break L4;
     }
     $24 = HEAP8[$9>>0]|0;
     $25 = ($24<<24>>24)==(0);
     if ($25) {
      $31 = HEAP8[$11>>0]|0;
      $32 = ($31<<24>>24)==(0);
      if (!($32)) {
       $33 = HEAP32[$21>>2]|0;
       $34 = $33 & 1;
       $35 = ($34|0)==(0);
       if ($35) {
        break L4;
       }
      }
     } else {
      $26 = HEAP32[$19>>2]|0;
      $27 = ($26|0)==(1);
      if ($27) {
       break L4;
      }
      $28 = HEAP32[$21>>2]|0;
      $29 = $28 & 2;
      $30 = ($29|0)==(0);
      if ($30) {
       break L4;
      }
     }
     HEAP8[$9>>0] = 0;
     HEAP8[$11>>0] = 0;
     __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($$0,$1,$2,$3,$4,$5);
     $36 = ((($$0)) + 8|0);
     $37 = ($36>>>0)<($16>>>0);
     if ($37) {
      $$0 = $36;
     } else {
      break;
     }
    }
   }
  } while(0);
  HEAP8[$9>>0] = $10;
  HEAP8[$11>>0] = $12;
 }
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $$081$off0 = 0, $$084 = 0, $$085$off0 = 0, $$1 = 0, $$182$off0 = 0, $$186$off0 = 0, $$2 = 0, $$283$off0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($1)) + 8|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$6,$4)|0);
 L1: do {
  if ($7) {
   __ZNK10__cxxabiv117__class_type_info29process_static_type_below_dstEPNS_19__dynamic_cast_infoEPKvi(0,$1,$2,$3);
  } else {
   $8 = HEAP32[$1>>2]|0;
   $9 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$8,$4)|0);
   $10 = ((($0)) + 12|0);
   $11 = ((($1)) + 24|0);
   $12 = ((($1)) + 36|0);
   $13 = ((($1)) + 54|0);
   $14 = ((($0)) + 8|0);
   $15 = ((($0)) + 16|0);
   if (!($9)) {
    $55 = HEAP32[$10>>2]|0;
    $56 = (((($0)) + 16|0) + ($55<<3)|0);
    __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($15,$1,$2,$3,$4);
    $57 = ((($0)) + 24|0);
    $58 = ($55|0)>(1);
    if (!($58)) {
     break;
    }
    $59 = HEAP32[$14>>2]|0;
    $60 = $59 & 2;
    $61 = ($60|0)==(0);
    if ($61) {
     $62 = HEAP32[$12>>2]|0;
     $63 = ($62|0)==(1);
     if ($63) {
      $$0 = $57;
     } else {
      $68 = $59 & 1;
      $69 = ($68|0)==(0);
      if ($69) {
       $$2 = $57;
       while(1) {
        $78 = HEAP8[$13>>0]|0;
        $79 = ($78<<24>>24)==(0);
        if (!($79)) {
         break L1;
        }
        $80 = HEAP32[$12>>2]|0;
        $81 = ($80|0)==(1);
        if ($81) {
         break L1;
        }
        __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($$2,$1,$2,$3,$4);
        $82 = ((($$2)) + 8|0);
        $83 = ($82>>>0)<($56>>>0);
        if ($83) {
         $$2 = $82;
        } else {
         break L1;
        }
       }
      } else {
       $$1 = $57;
      }
      while(1) {
       $70 = HEAP8[$13>>0]|0;
       $71 = ($70<<24>>24)==(0);
       if (!($71)) {
        break L1;
       }
       $72 = HEAP32[$12>>2]|0;
       $73 = ($72|0)==(1);
       if ($73) {
        $74 = HEAP32[$11>>2]|0;
        $75 = ($74|0)==(1);
        if ($75) {
         break L1;
        }
       }
       __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($$1,$1,$2,$3,$4);
       $76 = ((($$1)) + 8|0);
       $77 = ($76>>>0)<($56>>>0);
       if ($77) {
        $$1 = $76;
       } else {
        break L1;
       }
      }
     }
    } else {
     $$0 = $57;
    }
    while(1) {
     $64 = HEAP8[$13>>0]|0;
     $65 = ($64<<24>>24)==(0);
     if (!($65)) {
      break L1;
     }
     __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($$0,$1,$2,$3,$4);
     $66 = ((($$0)) + 8|0);
     $67 = ($66>>>0)<($56>>>0);
     if ($67) {
      $$0 = $66;
     } else {
      break L1;
     }
    }
   }
   $16 = ((($1)) + 16|0);
   $17 = HEAP32[$16>>2]|0;
   $18 = ($17|0)==($2|0);
   $19 = ((($1)) + 32|0);
   if (!($18)) {
    $20 = ((($1)) + 20|0);
    $21 = HEAP32[$20>>2]|0;
    $22 = ($21|0)==($2|0);
    if (!($22)) {
     HEAP32[$19>>2] = $3;
     $24 = ((($1)) + 44|0);
     $25 = HEAP32[$24>>2]|0;
     $26 = ($25|0)==(4);
     if ($26) {
      break;
     }
     $27 = HEAP32[$10>>2]|0;
     $28 = (((($0)) + 16|0) + ($27<<3)|0);
     $29 = ((($1)) + 52|0);
     $30 = ((($1)) + 53|0);
     $$081$off0 = 0;$$084 = $15;$$085$off0 = 0;
     L29: while(1) {
      $31 = ($$084>>>0)<($28>>>0);
      if (!($31)) {
       $$283$off0 = $$081$off0;
       label = 18;
       break;
      }
      HEAP8[$29>>0] = 0;
      HEAP8[$30>>0] = 0;
      __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($$084,$1,$2,$2,1,$4);
      $32 = HEAP8[$13>>0]|0;
      $33 = ($32<<24>>24)==(0);
      if (!($33)) {
       $$283$off0 = $$081$off0;
       label = 18;
       break;
      }
      $34 = HEAP8[$30>>0]|0;
      $35 = ($34<<24>>24)==(0);
      do {
       if ($35) {
        $$182$off0 = $$081$off0;$$186$off0 = $$085$off0;
       } else {
        $36 = HEAP8[$29>>0]|0;
        $37 = ($36<<24>>24)==(0);
        if ($37) {
         $43 = HEAP32[$14>>2]|0;
         $44 = $43 & 1;
         $45 = ($44|0)==(0);
         if ($45) {
          $$283$off0 = 1;
          label = 18;
          break L29;
         } else {
          $$182$off0 = 1;$$186$off0 = $$085$off0;
          break;
         }
        }
        $38 = HEAP32[$11>>2]|0;
        $39 = ($38|0)==(1);
        if ($39) {
         label = 23;
         break L29;
        }
        $40 = HEAP32[$14>>2]|0;
        $41 = $40 & 2;
        $42 = ($41|0)==(0);
        if ($42) {
         label = 23;
         break L29;
        } else {
         $$182$off0 = 1;$$186$off0 = 1;
        }
       }
      } while(0);
      $46 = ((($$084)) + 8|0);
      $$081$off0 = $$182$off0;$$084 = $46;$$085$off0 = $$186$off0;
     }
     do {
      if ((label|0) == 18) {
       if (!($$085$off0)) {
        HEAP32[$20>>2] = $2;
        $47 = ((($1)) + 40|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = (($48) + 1)|0;
        HEAP32[$47>>2] = $49;
        $50 = HEAP32[$12>>2]|0;
        $51 = ($50|0)==(1);
        if ($51) {
         $52 = HEAP32[$11>>2]|0;
         $53 = ($52|0)==(2);
         if ($53) {
          HEAP8[$13>>0] = 1;
          if ($$283$off0) {
           label = 23;
           break;
          } else {
           $54 = 4;
           break;
          }
         }
        }
       }
       if ($$283$off0) {
        label = 23;
       } else {
        $54 = 4;
       }
      }
     } while(0);
     if ((label|0) == 23) {
      $54 = 3;
     }
     HEAP32[$24>>2] = $54;
     break;
    }
   }
   $23 = ($3|0)==(1);
   if ($23) {
    HEAP32[$19>>2] = 1;
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv121__vmi_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($1)) + 8|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = (__ZN10__cxxabiv18is_equalEPKSt9type_infoS2_b($0,$5,0)|0);
 L1: do {
  if ($6) {
   __ZNK10__cxxabiv117__class_type_info24process_found_base_classEPNS_19__dynamic_cast_infoEPvi(0,$1,$2,$3);
  } else {
   $7 = ((($0)) + 16|0);
   $8 = ((($0)) + 12|0);
   $9 = HEAP32[$8>>2]|0;
   $10 = (((($0)) + 16|0) + ($9<<3)|0);
   __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($7,$1,$2,$3);
   $11 = ($9|0)>(1);
   if ($11) {
    $12 = ((($0)) + 24|0);
    $13 = ((($1)) + 54|0);
    $$0 = $12;
    while(1) {
     __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($$0,$1,$2,$3);
     $14 = HEAP8[$13>>0]|0;
     $15 = ($14<<24>>24)==(0);
     if (!($15)) {
      break L1;
     }
     $16 = ((($$0)) + 8|0);
     $17 = ($16>>>0)<($10>>>0);
     if ($17) {
      $$0 = $16;
     } else {
      break;
     }
    }
   }
  }
 } while(0);
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $5 >> 8;
 $7 = $5 & 1;
 $8 = ($7|0)==(0);
 if ($8) {
  $$0 = $6;
 } else {
  $9 = HEAP32[$2>>2]|0;
  $10 = (($9) + ($6)|0);
  $11 = HEAP32[$10>>2]|0;
  $$0 = $11;
 }
 $12 = HEAP32[$0>>2]|0;
 $13 = HEAP32[$12>>2]|0;
 $14 = ((($13)) + 28|0);
 $15 = HEAP32[$14>>2]|0;
 $16 = (($2) + ($$0)|0);
 $17 = $5 & 2;
 $18 = ($17|0)!=(0);
 $19 = $18 ? $3 : 2;
 FUNCTION_TABLE_viiii[$15 & 2047]($12,$1,$16,$19);
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = ((($0)) + 4|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = $7 >> 8;
 $9 = $7 & 1;
 $10 = ($9|0)==(0);
 if ($10) {
  $$0 = $8;
 } else {
  $11 = HEAP32[$3>>2]|0;
  $12 = (($11) + ($8)|0);
  $13 = HEAP32[$12>>2]|0;
  $$0 = $13;
 }
 $14 = HEAP32[$0>>2]|0;
 $15 = HEAP32[$14>>2]|0;
 $16 = ((($15)) + 20|0);
 $17 = HEAP32[$16>>2]|0;
 $18 = (($3) + ($$0)|0);
 $19 = $7 & 2;
 $20 = ($19|0)!=(0);
 $21 = $20 ? $4 : 2;
 FUNCTION_TABLE_viiiiii[$17 & 1023]($14,$1,$2,$18,$21,$5);
 return;
}
function __ZNK10__cxxabiv122__base_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib($0,$1,$2,$3,$4) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 var $$0 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $5 = ((($0)) + 4|0);
 $6 = HEAP32[$5>>2]|0;
 $7 = $6 >> 8;
 $8 = $6 & 1;
 $9 = ($8|0)==(0);
 if ($9) {
  $$0 = $7;
 } else {
  $10 = HEAP32[$2>>2]|0;
  $11 = (($10) + ($7)|0);
  $12 = HEAP32[$11>>2]|0;
  $$0 = $12;
 }
 $13 = HEAP32[$0>>2]|0;
 $14 = HEAP32[$13>>2]|0;
 $15 = ((($14)) + 24|0);
 $16 = HEAP32[$15>>2]|0;
 $17 = (($2) + ($$0)|0);
 $18 = $6 & 2;
 $19 = ($18|0)!=(0);
 $20 = $19 ? $3 : 2;
 FUNCTION_TABLE_viiiii[$16 & 1023]($13,$1,$17,$20,$4);
 return;
}
function __ZNSt8bad_castC2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22212);
 return;
}
function ___cxa_demangle($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$157 = 0, $$2 = 0, $$247 = 0, $$250 = 0, $$3 = 0, $$351 = 0, $$pr = 0, $$pr$pre = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0;
 var $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond53 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 4192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(4192|0);
 $4 = sp + 4184|0;
 $5 = sp;
 $6 = sp + 4120|0;
 $7 = sp + 4112|0;
 $8 = ($0|0)==(0|0);
 $9 = ($3|0)!=(0|0);
 L1: do {
  if ($8) {
   label = 3;
  } else {
   $10 = ($1|0)!=(0|0);
   $11 = ($2|0)==(0|0);
   $or$cond = $10 & $11;
   if ($or$cond) {
    label = 3;
   } else {
    if ($10) {
     $12 = HEAP32[$2>>2]|0;
     $74 = $12;
    } else {
     $74 = 0;
    }
    __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EEC2Ev($5);
    __THREW__ = 0;
    invoke_vii(1109,($6|0),($5|0));
    $13 = __THREW__; __THREW__ = 0;
    $14 = $13&1;
    if ($14) {
     $63 = ___cxa_find_matching_catch_2()|0;
     $64 = tempRet0;
     $$3 = $63;$$351 = $64;
    } else {
     $15 = ((($6)) + 48|0);
     $16 = ((($6)) + 61|0);
     ;HEAP32[$15>>2]=0|0;HEAP32[$15+4>>2]=0|0;HEAP32[$15+8>>2]=0|0;HEAP8[$15+12>>0]=0|0;
     HEAP8[$16>>0] = 1;
     $17 = ((($6)) + 32|0);
     $18 = ((($6)) + 36|0);
     $19 = HEAP32[$18>>2]|0;
     $20 = ((($6)) + 40|0);
     $21 = HEAP32[$20>>2]|0;
     $22 = ($19>>>0)<($21>>>0);
     if ($22) {
      __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($4,$5);
      HEAP32[$19>>2] = 0;
      $23 = ((($19)) + 4|0);
      HEAP32[$23>>2] = 0;
      $24 = ((($19)) + 8|0);
      $25 = HEAP32[$4>>2]|0;
      HEAP32[$24>>2] = 0;
      $26 = ((($19)) + 12|0);
      HEAP32[$26>>2] = $25;
      $27 = HEAP32[$18>>2]|0;
      $28 = ((($27)) + 16|0);
      HEAP32[$18>>2] = $28;
      label = 11;
     } else {
      __THREW__ = 0;
      invoke_vii(1110,($17|0),($5|0));
      $29 = __THREW__; __THREW__ = 0;
      $30 = $29&1;
      if ($30) {
       $65 = ___cxa_find_matching_catch_2()|0;
       $66 = tempRet0;
       $$247 = $65;$$250 = $66;
      } else {
       label = 11;
      }
     }
     if ((label|0) == 11) {
      $31 = ((($6)) + 62|0);
      HEAP8[$31>>0] = 0;
      $32 = ((($6)) + 63|0);
      HEAP8[$32>>0] = 1;
      HEAP32[$7>>2] = 0;
      $33 = (_strlen($0)|0);
      $34 = (($0) + ($33)|0);
      __THREW__ = 0;
      invoke_viiii(1111,($0|0),($34|0),($6|0),($7|0));
      $35 = __THREW__; __THREW__ = 0;
      $36 = $35&1;
      L16: do {
       if (!($36)) {
        $37 = HEAP32[$7>>2]|0;
        $38 = ($37|0)!=(0);
        $39 = HEAP8[$31>>0]|0;
        $40 = ($39<<24>>24)==(0);
        $or$cond53 = $38 | $40;
        do {
         if ($or$cond53) {
          $$pr = $37;
          label = 27;
         } else {
          $41 = HEAP32[$17>>2]|0;
          $42 = HEAP32[$18>>2]|0;
          $43 = ($41|0)==($42|0);
          if ($43) {
           label = 28;
          } else {
           $44 = HEAP32[$41>>2]|0;
           $45 = ((($41)) + 4|0);
           $46 = HEAP32[$45>>2]|0;
           $47 = ($44|0)==($46|0);
           if ($47) {
            label = 28;
           } else {
            HEAP8[$31>>0] = 0;
            HEAP8[$16>>0] = 0;
            $48 = HEAP32[$6>>2]|0;
            $49 = ((($6)) + 4|0);
            while(1) {
             $50 = HEAP32[$49>>2]|0;
             $51 = ($50|0)==($48|0);
             if ($51) {
              break;
             }
             $52 = ((($50)) + -24|0);
             HEAP32[$49>>2] = $52;
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($52);
            }
            $53 = ((($6)) + 16|0);
            $54 = HEAP32[$53>>2]|0;
            $55 = ((($6)) + 20|0);
            while(1) {
             $56 = HEAP32[$55>>2]|0;
             $57 = ($56|0)==($54|0);
             if ($57) {
              break;
             }
             $58 = ((($56)) + -16|0);
             HEAP32[$55>>2] = $58;
             __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($58);
            }
            __THREW__ = 0;
            invoke_viiii(1111,($0|0),($34|0),($6|0),($7|0));
            $59 = __THREW__; __THREW__ = 0;
            $60 = $59&1;
            if ($60) {
             break L16;
            }
            $61 = HEAP8[$31>>0]|0;
            $62 = ($61<<24>>24)==(0);
            if ($62) {
             $$pr$pre = HEAP32[$7>>2]|0;
             $$pr = $$pr$pre;
             label = 27;
             break;
            } else {
             HEAP32[$7>>2] = -2;
             $$2 = 0;$100 = -2;
             break;
            }
           }
          }
         }
        } while(0);
        if ((label|0) == 27) {
         $67 = ($$pr|0)==(0);
         if ($67) {
          label = 28;
         } else {
          $$2 = 0;$100 = $$pr;
         }
        }
        do {
         if ((label|0) == 28) {
          $68 = ((($6)) + 4|0);
          $69 = HEAP32[$68>>2]|0;
          $70 = ((($69)) + -24|0);
          $71 = (__ZNK10__cxxabiv112_GLOBAL__N_111string_pair4sizeEv($70)|0);
          $72 = (($71) + 1)|0;
          $73 = ($72>>>0)>($74>>>0);
          if ($73) {
           $75 = (_realloc($1,$72)|0);
           $76 = ($75|0)==(0|0);
           if ($76) {
            HEAP32[$7>>2] = -1;
            $$2 = 0;$100 = -1;
            break;
           }
           if ($11) {
            $$157 = $75;
           } else {
            HEAP32[$2>>2] = $72;
            $$157 = $75;
           }
          } else {
           $77 = ($1|0)==(0|0);
           if ($77) {
            $$2 = 0;$100 = 0;
            break;
           } else {
            $$157 = $1;
           }
          }
          $78 = HEAP32[$68>>2]|0;
          $79 = ((($78)) + -12|0);
          $80 = ((($78)) + -24|0);
          $81 = ((($79)) + 11|0);
          $82 = HEAP8[$81>>0]|0;
          $83 = ($82<<24>>24)<(0);
          $84 = HEAP32[$79>>2]|0;
          $85 = $83 ? $84 : $79;
          $86 = ((($78)) + -8|0);
          $87 = HEAP32[$86>>2]|0;
          $88 = $82&255;
          $89 = $83 ? $87 : $88;
          __THREW__ = 0;
          (invoke_iiii(1112,($80|0),($85|0),($89|0))|0);
          $90 = __THREW__; __THREW__ = 0;
          $91 = $90&1;
          if ($91) {
           break L16;
          }
          $92 = HEAP32[$68>>2]|0;
          $93 = ((($92)) + -24|0);
          $94 = ((($93)) + 11|0);
          $95 = HEAP8[$94>>0]|0;
          $96 = ($95<<24>>24)<(0);
          if ($96) {
           $97 = HEAP32[$93>>2]|0;
           $98 = $97;
          } else {
           $98 = $93;
          }
          _memcpy(($$157|0),($98|0),($71|0))|0;
          $99 = (($$157) + ($71)|0);
          HEAP8[$99>>0] = 0;
          $$2 = $$157;$100 = 0;
         }
        } while(0);
        if ($9) {
         HEAP32[$3>>2] = $100;
        }
        __ZN10__cxxabiv112_GLOBAL__N_12DbD2Ev($6);
        __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EED2Ev($5);
        $$0 = $$2;
        break L1;
       }
      } while(0);
      $101 = ___cxa_find_matching_catch_2()|0;
      $102 = tempRet0;
      $$247 = $101;$$250 = $102;
     }
     __ZN10__cxxabiv112_GLOBAL__N_12DbD2Ev($6);
     $$3 = $$247;$$351 = $$250;
    }
    __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EED2Ev($5);
    ___resumeException($$3|0);
    // unreachable;
   }
  }
 } while(0);
 if ((label|0) == 3) {
  if ($9) {
   HEAP32[$3>>2] = -3;
   $$0 = 0;
  } else {
   $$0 = 0;
  }
 }
 STACKTOP = sp;return ($$0|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EEC2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4096|0);
 HEAP32[$1>>2] = $0;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_12DbC2ILj4096EEERNS0_5arenaIXT_EEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 8|0;
 $3 = sp + 4|0;
 $4 = sp;
 __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EEC2ERNS0_5arenaILj4096EEE($2,$1);
 HEAP32[$0>>2] = 0;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 0;
 $6 = ((($0)) + 8|0);
 $7 = HEAP32[$2>>2]|0;
 HEAP32[$6>>2] = 0;
 $8 = ((($0)) + 12|0);
 HEAP32[$8>>2] = $7;
 $9 = ((($0)) + 16|0);
 __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($3,$1);
 __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEC2EjRKS6_RKS7_($9,$3);
 $10 = ((($0)) + 32|0);
 __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($4,$1);
 __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEEC2EjRKS8_RKS9_($10,$4);
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $1;
 return;
}
function __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE24__emplace_back_slow_pathIJRNS2_5arenaILj4096EEEEEEvDpOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 20|0;
 $3 = sp;
 $4 = ((($0)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$0>>2]|0;
 $7 = (($5) - ($6))|0;
 $8 = $7 >> 4;
 $9 = (($8) + 1)|0;
 $10 = (__ZNKSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE8max_sizeEv($0)|0);
 $11 = ($10>>>0)<($9>>>0);
 if ($11) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $12 = ((($0)) + 12|0);
  $13 = ((($0)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$0>>2]|0;
  $16 = (($14) - ($15))|0;
  $17 = $16 >> 4;
  $18 = $10 >>> 1;
  $19 = ($17>>>0)<($18>>>0);
  $20 = $16 >> 3;
  $21 = ($20>>>0)<($9>>>0);
  $$sroa$speculated$i = $21 ? $9 : $20;
  $$sroa$speculated$$i = $19 ? $$sroa$speculated$i : $10;
  $22 = HEAP32[$4>>2]|0;
  $23 = (($22) - ($15))|0;
  $24 = $23 >> 4;
  __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEEC2EjjSB_($3,$$sroa$speculated$$i,$24,$12);
  $25 = ((($3)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($2,$1);
  HEAP32[$26>>2] = 0;
  $27 = ((($26)) + 4|0);
  HEAP32[$27>>2] = 0;
  $28 = ((($26)) + 8|0);
  $29 = HEAP32[$2>>2]|0;
  HEAP32[$28>>2] = 0;
  $30 = ((($26)) + 12|0);
  HEAP32[$30>>2] = $29;
  $31 = ((($26)) + 16|0);
  HEAP32[$25>>2] = $31;
  __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS8_RS9_EE($0,$3);
  __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEED2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_18demangleINS0_2DbEEEvPKcS4_RT_Ri($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$pr = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond56 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $4 = ($0>>>0)<($1>>>0);
 L1: do {
  if ($4) {
   $5 = HEAP8[$0>>0]|0;
   $6 = ($5<<24>>24)==(95);
   L3: do {
    if ($6) {
     $7 = $1;
     $8 = $0;
     $9 = (($7) - ($8))|0;
     $10 = ($9|0)>(3);
     if (!($10)) {
      HEAP32[$3>>2] = -2;
      break L1;
     }
     $11 = ((($0)) + 1|0);
     $12 = HEAP8[$11>>0]|0;
     $13 = ((($0)) + 2|0);
     switch ($12<<24>>24) {
     case 90:  {
      $14 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($13,$1,$2)|0);
      $15 = ($14|0)==($13|0);
      $16 = ($14|0)==($1|0);
      $or$cond = $15 | $16;
      if ($or$cond) {
       $$0 = $14;
      } else {
       $17 = HEAP8[$14>>0]|0;
       $18 = ($17<<24>>24)==(46);
       if ($18) {
        $19 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_dot_suffixINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
        $$0 = $19;
       } else {
        $$0 = $14;
       }
      }
      $20 = ($$0|0)==($1|0);
      if ($20) {
       break L3;
      }
      HEAP32[$3>>2] = -2;
      break L1;
      break;
     }
     case 95:  {
      $21 = HEAP8[$13>>0]|0;
      $22 = ($21<<24>>24)==(95);
      if ($22) {
       $23 = ((($0)) + 3|0);
       $24 = HEAP8[$23>>0]|0;
       $25 = ($24<<24>>24)==(90);
       if ($25) {
        $26 = ((($0)) + 4|0);
        $27 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($26,$1,$2)|0);
        $28 = ($27|0)==($26|0);
        $29 = ($27|0)==($1|0);
        $or$cond56 = $28 | $29;
        if (!($or$cond56)) {
         $30 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_block_invokeINS0_2DbEEEPKcS4_S4_RT_($27,$1,$2)|0);
         $31 = ($30|0)==($1|0);
         if ($31) {
          break L3;
         }
        }
        HEAP32[$3>>2] = -2;
        break L1;
       }
      }
      break;
     }
     default: {
     }
     }
     HEAP32[$3>>2] = -2;
     break L1;
    } else {
     $32 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $33 = ($32|0)==($1|0);
     if (!($33)) {
      HEAP32[$3>>2] = -2;
      break L1;
     }
    }
   } while(0);
   $$pr = HEAP32[$3>>2]|0;
   $34 = ($$pr|0)==(0);
   if ($34) {
    $35 = HEAP32[$2>>2]|0;
    $36 = ((($2)) + 4|0);
    $37 = HEAP32[$36>>2]|0;
    $38 = ($35|0)==($37|0);
    if ($38) {
     HEAP32[$3>>2] = -2;
    }
   }
  } else {
   HEAP32[$3>>2] = -2;
  }
 } while(0);
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 12|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($1);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($0);
 return;
}
function __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $$cast = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($0)) + 4|0);
  while(1) {
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($1|0);
   if ($5) {
    break;
   }
   $6 = ((($4)) + -24|0);
   HEAP32[$3>>2] = $6;
   __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
  }
  $7 = ((($0)) + 12|0);
  $8 = HEAP32[$0>>2]|0;
  $9 = ((($0)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $$cast = $8;
  $11 = (($10) - ($$cast))|0;
  $12 = (($11|0) / 24)&-1;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE10deallocateEPS2_j($7,$8,$12);
 }
 return;
}
function __ZNK10__cxxabiv112_GLOBAL__N_111string_pair4sizeEv($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 11|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)<(0);
 if ($3) {
  $4 = ((($0)) + 4|0);
  $5 = HEAP32[$4>>2]|0;
  $15 = $5;
 } else {
  $6 = $2&255;
  $15 = $6;
 }
 $7 = ((($0)) + 12|0);
 $8 = ((($7)) + 11|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)<(0);
 if ($10) {
  $11 = ((($0)) + 16|0);
  $12 = HEAP32[$11>>2]|0;
  $16 = $12;
 } else {
  $13 = $9&255;
  $16 = $13;
 }
 $14 = (($16) + ($15))|0;
 return ($14|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $7 = ((($0)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $10 = ((($0)) + 4|0);
  $11 = HEAP32[$10>>2]|0;
  $14 = $11;$15 = $phitmp$i;
 } else {
  $12 = $5&255;
  $14 = $12;$15 = 10;
 }
 $13 = (($15) - ($14))|0;
 $16 = ($13>>>0)<($2>>>0);
 $17 = (($14) + ($2))|0;
 if ($16) {
  $27 = (($17) - ($15))|0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$15,$27,$14,$14,0,$2,$1);
 } else {
  $18 = ($2|0)==(0);
  if (!($18)) {
   if ($6) {
    $19 = HEAP32[$0>>2]|0;
    $21 = $19;
   } else {
    $21 = $0;
   }
   $20 = (($21) + ($14)|0);
   (__ZNSt3__211char_traitsIcE4copyEPcPKcj($20,$1,$2)|0);
   $22 = HEAP8[$4>>0]|0;
   $23 = ($22<<24>>24)<(0);
   if ($23) {
    $24 = ((($0)) + 4|0);
    HEAP32[$24>>2] = $17;
   } else {
    $25 = $17&255;
    HEAP8[$4>>0] = $25;
   }
   $26 = (($21) + ($17)|0);
   HEAP8[$3>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($26,$3);
  }
 }
 STACKTOP = sp;return ($0|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_12DbD2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 32|0);
 __ZNSt3__213__vector_baseINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEENS5_IS9_Lj4096EEEED2Ev($1);
 $2 = ((($0)) + 16|0);
 __ZNSt3__213__vector_baseINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEED2Ev($2);
 __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($0);
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4096|0);
 HEAP32[$1>>2] = 0;
 return;
}
function __ZNSt3__213__vector_baseINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEENS5_IS9_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $$cast = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($0)) + 4|0);
  while(1) {
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($1|0);
   if ($5) {
    break;
   }
   $6 = ((($4)) + -16|0);
   HEAP32[$3>>2] = $6;
   __ZNSt3__213__vector_baseINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEED2Ev($6);
  }
  $7 = ((($0)) + 12|0);
  $8 = HEAP32[$0>>2]|0;
  $9 = ((($0)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $$cast = $8;
  $11 = (($10) - ($$cast))|0;
  $12 = $11 >> 4;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EE10deallocateEPS8_j($7,$8,$12);
 }
 return;
}
function __ZNSt3__213__vector_baseINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $$cast = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP32[$0>>2]|0;
 $2 = ($1|0)==(0|0);
 if (!($2)) {
  $3 = ((($0)) + 4|0);
  while(1) {
   $4 = HEAP32[$3>>2]|0;
   $5 = ($4|0)==($1|0);
   if ($5) {
    break;
   }
   $6 = ((($4)) + -16|0);
   HEAP32[$3>>2] = $6;
   __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($6);
  }
  $7 = ((($0)) + 12|0);
  $8 = HEAP32[$0>>2]|0;
  $9 = ((($0)) + 8|0);
  $10 = HEAP32[$9>>2]|0;
  $$cast = $8;
  $11 = (($10) - ($$cast))|0;
  $12 = $11 >> 4;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EE10deallocateEPS6_j($7,$8,$12);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EE10deallocateEPS6_j($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = HEAP32[$0>>2]|0;
 $4 = $2 << 4;
 __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE10deallocateEPcj($3,$1,$4);
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE10deallocateEPcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE17pointer_in_bufferEPc($0,$1)|0);
 if ($3) {
  $4 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8align_upEj($2)|0);
  $5 = (($1) + ($4)|0);
  $6 = ((($0)) + 4096|0);
  $7 = HEAP32[$6>>2]|0;
  $8 = ($5|0)==($7|0);
  if ($8) {
   HEAP32[$6>>2] = $1;
  }
 } else {
  _free($1);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE17pointer_in_bufferEPc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0>>>0)<=($1>>>0);
 $3 = ((($0)) + 4096|0);
 $4 = ($3>>>0)>=($1>>>0);
 $5 = $2 & $4;
 return ($5|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8align_upEj($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (($0) + 15)|0;
 $2 = $1 & -16;
 return ($2|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EE10deallocateEPS8_j($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = HEAP32[$0>>2]|0;
 $4 = $2 << 4;
 __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE10deallocateEPcj($3,$1,$4);
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$1,$2,$3,$4,$5,$6,$7) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 $6 = $6|0;
 $7 = $7|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $40 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $8 = sp;
 $9 = (-18 - ($1))|0;
 $10 = ($9>>>0)<($2>>>0);
 if ($10) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $11 = ((($0)) + 11|0);
 $12 = HEAP8[$11>>0]|0;
 $13 = ($12<<24>>24)<(0);
 if ($13) {
  $14 = HEAP32[$0>>2]|0;
  $25 = $14;
 } else {
  $25 = $0;
 }
 $15 = ($1>>>0)<(2147483623);
 if ($15) {
  $16 = (($2) + ($1))|0;
  $17 = $1 << 1;
  $18 = ($16>>>0)<($17>>>0);
  $$sroa$speculated = $18 ? $17 : $16;
  $19 = ($$sroa$speculated>>>0)<(11);
  $20 = (($$sroa$speculated) + 16)|0;
  $21 = $20 & -16;
  $phitmp = $19 ? 11 : $21;
  $22 = $phitmp;
 } else {
  $22 = -17;
 }
 $23 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($22)|0);
 $24 = ($4|0)==(0);
 if (!($24)) {
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($23,$25,$4)|0);
 }
 $26 = ($6|0)==(0);
 if (!($26)) {
  $27 = (($23) + ($4)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($27,$7,$6)|0);
 }
 $28 = (($3) - ($5))|0;
 $29 = (($28) - ($4))|0;
 $30 = ($29|0)==(0);
 if (!($30)) {
  $31 = (($23) + ($4)|0);
  $32 = (($31) + ($6)|0);
  $33 = (($25) + ($4)|0);
  $34 = (($33) + ($5)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($32,$34,$29)|0);
 }
 $35 = ($1|0)==(10);
 if (!($35)) {
  __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($25);
 }
 HEAP32[$0>>2] = $23;
 $36 = $22 | -2147483648;
 $37 = ((($0)) + 8|0);
 HEAP32[$37>>2] = $36;
 $38 = (($28) + ($6))|0;
 $39 = ((($0)) + 4|0);
 HEAP32[$39>>2] = $38;
 $40 = (($23) + ($38)|0);
 HEAP8[$8>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($40,$8);
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (_malloc($0)|0);
 return ($1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 _free($0);
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE10deallocateEPS2_j($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = HEAP32[$0>>2]|0;
 $4 = ($2*24)|0;
 __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE10deallocateEPcj($3,$1,$4);
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 11|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)<(0);
 if ($3) {
  $4 = HEAP32[$0>>2]|0;
  __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($4);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i189 = 0, $$0$i$i194 = 0, $$0132 = 0, $$0141$off0$ph = 0, $$0147 = 0, $$0160 = 0, $$0166 = 0, $$11 = 0, $$1135 = 0, $$1167 = 0, $$1167$ph = 0, $$2136 = 0, $$2149 = 0, $$2162 = 0, $$3144$off0$ph = 0, $$3169 = 0, $$4151 = 0;
 var $$4157 = 0, $$4164 = 0, $$5139 = 0, $$5152 = 0, $$5165 = 0, $$6140 = 0, $$6159 = 0, $$in = 0, $$in256 = 0, $$pre = 0, $$pre$i = 0, $$pre$i185 = 0, $$pre237 = 0, $$pre238 = 0, $$pre239 = 0, $$pre240 = 0, $$pre241 = 0, $$pre242 = 0, $$pre243 = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0;
 var $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0;
 var $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0;
 var $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0;
 var $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0;
 var $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0;
 var $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0;
 var $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0;
 var $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0;
 var $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i190 = 0, $exitcond$i$i195 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $3 = sp + 73|0;
 $4 = sp + 72|0;
 $5 = sp + 64|0;
 $6 = sp + 56|0;
 $7 = sp + 48|0;
 $8 = sp + 36|0;
 $9 = sp + 24|0;
 $10 = sp + 12|0;
 $11 = sp;
 $12 = ($0|0)==($1|0);
 do {
  if ($12) {
   $$6140 = $0;
   label = 111;
  } else {
   $13 = ((($2)) + 56|0);
   __ZN10__cxxabiv112_GLOBAL__N_110save_valueIjEC2ERj($5,$13);
   $14 = HEAP32[$13>>2]|0;
   $15 = (($14) + 1)|0;
   HEAP32[$13>>2] = $15;
   $16 = ((($2)) + 61|0);
   __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbEC2ERb($6,$16);
   $17 = HEAP32[$13>>2]|0;
   $18 = ($17>>>0)>(1);
   if ($18) {
    HEAP8[$16>>0] = 1;
   }
   $19 = HEAP8[$0>>0]|0;
   $20 = $19 << 24 >> 24;
   L6: do {
    switch ($20|0) {
    case 84: case 71:  {
     __THREW__ = 0;
     $21 = (invoke_iiii(1113,($0|0),($1|0),($2|0))|0);
     $22 = __THREW__; __THREW__ = 0;
     $23 = $22&1;
     if ($23) {
      label = 6;
     } else {
      $$5139 = $21;$$6159 = 1;
      label = 109;
     }
     break;
    }
    default: {
     __THREW__ = 0;
     $26 = (invoke_iiii(1114,($0|0),($1|0),($2|0))|0);
     $27 = __THREW__; __THREW__ = 0;
     $28 = $27&1;
     if ($28) {
      label = 6;
     } else {
      $29 = ((($2)) + 48|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ((($2)) + 52|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ($26|0)==($0|0);
      L9: do {
       if ($33) {
        $$2136 = $0;
       } else {
        $34 = ($26|0)==($1|0);
        if ($34) {
         $$2136 = $1;
        } else {
         $35 = HEAP8[$26>>0]|0;
         switch ($35<<24>>24) {
         case 46: case 69:  {
          $$2136 = $26;
          break L9;
          break;
         }
         default: {
         }
         }
         __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbEC2ERb($7,$16);
         HEAP8[$16>>0] = 0;
         ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;
         $$0$i$i194 = 0;
         while(1) {
          $exitcond$i$i195 = ($$0$i$i194|0)==(3);
          if ($exitcond$i$i195) {
           break;
          }
          $36 = (($8) + ($$0$i$i194<<2)|0);
          HEAP32[$36>>2] = 0;
          $37 = (($$0$i$i194) + 1)|0;
          $$0$i$i194 = $37;
         }
         $38 = HEAP32[$2>>2]|0;
         $39 = ((($2)) + 4|0);
         $40 = HEAP32[$39>>2]|0;
         $41 = ($38|0)==($40|0);
         L17: do {
          if ($41) {
           $$1135 = $0;$$4157 = 0;
          } else {
           $42 = ((($40)) + -24|0);
           $43 = ((($42)) + 11|0);
           $44 = HEAP8[$43>>0]|0;
           $45 = ($44<<24>>24)<(0);
           if ($45) {
            $46 = ((($40)) + -20|0);
            $47 = HEAP32[$46>>2]|0;
            $50 = $47;
           } else {
            $48 = $44&255;
            $50 = $48;
           }
           $49 = ($50|0)==(0);
           if ($49) {
            $$1135 = $0;$$4157 = 0;
           } else {
            $51 = ((($2)) + 60|0);
            $52 = HEAP8[$51>>0]|0;
            $53 = ($52<<24>>24)==(0);
            L24: do {
             if ($53) {
              if ($45) {
               $54 = HEAP32[$42>>2]|0;
               $55 = ((($40)) + -20|0);
               $56 = HEAP32[$55>>2]|0;
               $59 = $54;$60 = $56;
              } else {
               $57 = $44&255;
               $59 = $42;$60 = $57;
              }
              $58 = (($59) + ($60)|0);
              $61 = ((($58)) + -1|0);
              $62 = HEAP8[$61>>0]|0;
              $63 = ($62<<24>>24)==(62);
              if ($63) {
               if ($45) {
                $64 = ((($40)) + -20|0);
                $65 = HEAP32[$64>>2]|0;
                $66 = HEAP32[$42>>2]|0;
                $$in = $65;$70 = $66;
               } else {
                $67 = $44&255;
                $$in = $67;$70 = $42;
               }
               $68 = (($$in) + -2)|0;
               $69 = (($70) + ($68)|0);
               $71 = HEAP8[$69>>0]|0;
               $72 = ($71<<24>>24)==(45);
               if ($72) {
                $$0166 = $26;$138 = $40;
                label = 55;
               } else {
                if ($45) {
                 $73 = ((($40)) + -20|0);
                 $74 = HEAP32[$73>>2]|0;
                 $75 = HEAP32[$42>>2]|0;
                 $$in256 = $74;$79 = $75;
                } else {
                 $76 = $44&255;
                 $$in256 = $76;$79 = $42;
                }
                $77 = (($$in256) + -2)|0;
                $78 = (($79) + ($77)|0);
                $80 = HEAP8[$78>>0]|0;
                $81 = ($80<<24>>24)==(62);
                if ($81) {
                 $$0166 = $26;$138 = $40;
                 label = 55;
                } else {
                 __THREW__ = 0;
                 $82 = (invoke_iiii(1115,($26|0),($1|0),($2|0))|0);
                 $83 = __THREW__; __THREW__ = 0;
                 $84 = $83&1;
                 if ($84) {
                  label = 34;
                  break;
                 }
                 $85 = ($82|0)==($26|0);
                 if ($85) {
                  $$1135 = $0;$$4157 = 0;
                  break L17;
                 }
                 $88 = HEAP32[$39>>2]|0;
                 $89 = HEAP32[$2>>2]|0;
                 $90 = (($88) - ($89))|0;
                 $91 = (($90|0) / 24)&-1;
                 $92 = ($91>>>0)<(2);
                 $93 = $88;
                 if ($92) {
                  $$1135 = $0;$$4157 = 0;
                  break L17;
                 }
                 $94 = ((($93)) + -24|0);
                 ;HEAP32[$9>>2]=HEAP32[$94>>2]|0;HEAP32[$9+4>>2]=HEAP32[$94+4>>2]|0;HEAP32[$9+8>>2]=HEAP32[$94+8>>2]|0;
                 $$0$i$i189 = 0;
                 while(1) {
                  $exitcond$i$i190 = ($$0$i$i189|0)==(3);
                  if ($exitcond$i$i190) {
                   break;
                  }
                  $95 = (($94) + ($$0$i$i189<<2)|0);
                  HEAP32[$95>>2] = 0;
                  $96 = (($$0$i$i189) + 1)|0;
                  $$0$i$i189 = $96;
                 }
                 $97 = ((($93)) + -12|0);
                 $98 = ((($8)) + 11|0);
                 $99 = HEAP8[$98>>0]|0;
                 $100 = ($99<<24>>24)<(0);
                 if ($100) {
                  $101 = HEAP32[$8>>2]|0;
                  HEAP8[$3>>0] = 0;
                  __ZNSt3__211char_traitsIcE6assignERcRKc($101,$3);
                  $102 = ((($8)) + 4|0);
                  HEAP32[$102>>2] = 0;
                 } else {
                  HEAP8[$4>>0] = 0;
                  __ZNSt3__211char_traitsIcE6assignERcRKc($8,$4);
                  HEAP8[$98>>0] = 0;
                 }
                 __THREW__ = 0;
                 invoke_vi(1116,($8|0));
                 $103 = __THREW__; __THREW__ = 0;
                 $104 = $103&1;
                 if ($104) {
                  $105 = ___cxa_find_matching_catch_3(0|0)|0;
                  $106 = tempRet0;
                  ___clang_call_terminate($105);
                  // unreachable;
                 }
                 ;HEAP32[$8>>2]=HEAP32[$97>>2]|0;HEAP32[$8+4>>2]=HEAP32[$97+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$97+8>>2]|0;
                 $$0$i$i$i$i = 0;
                 while(1) {
                  $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
                  if ($exitcond$i$i$i$i) {
                   break;
                  }
                  $107 = (($97) + ($$0$i$i$i$i<<2)|0);
                  HEAP32[$107>>2] = 0;
                  $108 = (($$0$i$i$i$i) + 1)|0;
                  $$0$i$i$i$i = $108;
                 }
                 $109 = HEAP8[$98>>0]|0;
                 $110 = ($109<<24>>24)<(0);
                 $111 = ((($8)) + 4|0);
                 $112 = HEAP32[$111>>2]|0;
                 $113 = $109&255;
                 $114 = $110 ? $112 : $113;
                 $115 = ($114|0)==(0);
                 if ($115) {
                  __THREW__ = 0;
                  invoke_vii(1117,($9|0),32);
                  $116 = __THREW__; __THREW__ = 0;
                  $117 = $116&1;
                  if (!($117)) {
                   label = 50;
                  }
                 } else {
                  label = 50;
                 }
                 do {
                  if ((label|0) == 50) {
                   $120 = HEAP32[$39>>2]|0;
                   $121 = ((($120)) + -24|0);
                   $123 = $120;
                   while(1) {
                    $122 = ($123|0)==($121|0);
                    if ($122) {
                     break;
                    }
                    $124 = ((($123)) + -24|0);
                    HEAP32[$39>>2] = $124;
                    __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($124);
                    $$pre$i185 = HEAP32[$39>>2]|0;
                    $123 = $$pre$i185;
                   }
                   $125 = ((($120)) + -48|0);
                   $126 = ((($9)) + 11|0);
                   $127 = HEAP8[$126>>0]|0;
                   $128 = ($127<<24>>24)<(0);
                   $129 = HEAP32[$9>>2]|0;
                   $130 = $128 ? $129 : $9;
                   $131 = ((($9)) + 4|0);
                   $132 = HEAP32[$131>>2]|0;
                   $133 = $127&255;
                   $134 = $128 ? $132 : $133;
                   __THREW__ = 0;
                   (invoke_iiiii(1118,($125|0),0,($130|0),($134|0))|0);
                   $135 = __THREW__; __THREW__ = 0;
                   $136 = $135&1;
                   if ($136) {
                    break;
                   }
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
                   $$pre = HEAP32[$39>>2]|0;
                   $$0166 = $82;$138 = $$pre;
                   label = 55;
                   break L24;
                  }
                 } while(0);
                 $118 = ___cxa_find_matching_catch_2()|0;
                 $119 = tempRet0;
                 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
                 $$4151 = $118;$$4164 = $119;
                }
               }
              } else {
               $$0166 = $26;$138 = $40;
               label = 55;
              }
             } else {
              $$0166 = $26;$138 = $40;
              label = 55;
             }
            } while(0);
            L69: do {
             if ((label|0) == 55) {
              $137 = ((($138)) + -24|0);
              __THREW__ = 0;
              invoke_vii(1117,($137|0),40);
              $139 = __THREW__; __THREW__ = 0;
              $140 = $139&1;
              if ($140) {
               label = 34;
              } else {
               $141 = ($$0166|0)==($1|0);
               if ($141) {
                label = 58;
               } else {
                $142 = HEAP8[$$0166>>0]|0;
                $143 = ($142<<24>>24)==(118);
                if ($143) {
                 $148 = ((($$0166)) + 1|0);
                 $$pre242 = HEAP32[$2>>2]|0;
                 $$pre243 = HEAP32[$39>>2]|0;
                 $$3169 = $148;$230 = $$pre242;$231 = $$pre243;
                } else {
                 label = 58;
                }
               }
               do {
                if ((label|0) == 58) {
                 $144 = ((($10)) + 11|0);
                 $145 = ((($10)) + 4|0);
                 $146 = ((($11)) + 11|0);
                 $147 = ((($11)) + 4|0);
                 $$0141$off0$ph = 1;$$1167$ph = $$0166;
                 L77: while(1) {
                  $$pre237 = HEAP32[$39>>2]|0;
                  $$pre238 = HEAP32[$2>>2]|0;
                  $$1167 = $$1167$ph;$150 = $$pre238;$151 = $$pre237;
                  while(1) {
                   $149 = (($151) - ($150))|0;
                   $152 = (($149|0) / 24)&-1;
                   __THREW__ = 0;
                   $153 = (invoke_iiii(1115,($$1167|0),($1|0),($2|0))|0);
                   $154 = __THREW__; __THREW__ = 0;
                   $155 = $154&1;
                   if ($155) {
                    label = 62;
                    break L77;
                   }
                   $156 = HEAP32[$39>>2]|0;
                   $157 = HEAP32[$2>>2]|0;
                   $158 = (($156) - ($157))|0;
                   $159 = (($158|0) / 24)&-1;
                   $160 = ($153|0)==($$1167|0);
                   if ($160) {
                    label = 92;
                    break L77;
                   }
                   $163 = ($159>>>0)>($152>>>0);
                   if ($163) {
                    break;
                   } else {
                    $$1167 = $153;$150 = $157;$151 = $156;
                   }
                  }
                  ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
                  $$0$i$i = 0;
                  while(1) {
                   $exitcond$i$i = ($$0$i$i|0)==(3);
                   if ($exitcond$i$i) {
                    $$0132 = $152;
                    break;
                   }
                   $164 = (($10) + ($$0$i$i<<2)|0);
                   HEAP32[$164>>2] = 0;
                   $165 = (($$0$i$i) + 1)|0;
                   $$0$i$i = $165;
                  }
                  while(1) {
                   $166 = ($$0132>>>0)<($159>>>0);
                   if (!($166)) {
                    $$0 = $152;
                    break;
                   }
                   $167 = HEAP8[$144>>0]|0;
                   $168 = ($167<<24>>24)<(0);
                   $169 = HEAP32[$145>>2]|0;
                   $170 = $167&255;
                   $171 = $168 ? $169 : $170;
                   $172 = ($171|0)==(0);
                   if (!($172)) {
                    __THREW__ = 0;
                    (invoke_iii(1119,($10|0),(159305|0))|0);
                    $173 = __THREW__; __THREW__ = 0;
                    $174 = $173&1;
                    if ($174) {
                     label = 70;
                     break L77;
                    }
                   }
                   $177 = HEAP32[$2>>2]|0;
                   $178 = (($177) + (($$0132*24)|0)|0);
                   __THREW__ = 0;
                   invoke_vii(1120,($11|0),($178|0));
                   $179 = __THREW__; __THREW__ = 0;
                   $180 = $179&1;
                   if ($180) {
                    label = 74;
                    break L77;
                   }
                   $181 = HEAP8[$146>>0]|0;
                   $182 = ($181<<24>>24)<(0);
                   $183 = HEAP32[$11>>2]|0;
                   $184 = $182 ? $183 : $11;
                   $185 = HEAP32[$147>>2]|0;
                   $186 = $181&255;
                   $187 = $182 ? $185 : $186;
                   __THREW__ = 0;
                   (invoke_iiii(1112,($10|0),($184|0),($187|0))|0);
                   $188 = __THREW__; __THREW__ = 0;
                   $189 = $188&1;
                   if ($189) {
                    label = 75;
                    break L77;
                   }
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
                   $190 = (($$0132) + 1)|0;
                   $$0132 = $190;
                  }
                  while(1) {
                   $195 = ($$0>>>0)<($159>>>0);
                   if (!($195)) {
                    break;
                   }
                   $202 = HEAP32[$39>>2]|0;
                   $203 = ((($202)) + -24|0);
                   $205 = $202;
                   while(1) {
                    $204 = ($205|0)==($203|0);
                    if ($204) {
                     break;
                    }
                    $206 = ((($205)) + -24|0);
                    HEAP32[$39>>2] = $206;
                    __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($206);
                    $$pre$i = HEAP32[$39>>2]|0;
                    $205 = $$pre$i;
                   }
                   $207 = (($$0) + 1)|0;
                   $$0 = $207;
                  }
                  $196 = HEAP8[$144>>0]|0;
                  $197 = ($196<<24>>24)<(0);
                  $198 = HEAP32[$145>>2]|0;
                  $199 = $196&255;
                  $200 = $197 ? $198 : $199;
                  $201 = ($200|0)==(0);
                  if ($201) {
                   $$3144$off0$ph = $$0141$off0$ph;
                  } else {
                   $208 = HEAP32[$2>>2]|0;
                   $209 = HEAP32[$39>>2]|0;
                   $210 = ($208|0)==($209|0);
                   if ($210) {
                    label = 105;
                    break;
                   }
                   if ($$0141$off0$ph) {
                    $217 = $209;$219 = $196;$224 = $198;
                   } else {
                    $211 = ((($209)) + -24|0);
                    __THREW__ = 0;
                    (invoke_iii(1119,($211|0),(159305|0))|0);
                    $212 = __THREW__; __THREW__ = 0;
                    $213 = $212&1;
                    if ($213) {
                     label = 87;
                     break;
                    }
                    $$pre239 = HEAP32[$39>>2]|0;
                    $$pre240 = HEAP8[$144>>0]|0;
                    $$pre241 = HEAP32[$145>>2]|0;
                    $217 = $$pre239;$219 = $$pre240;$224 = $$pre241;
                   }
                   $216 = ((($217)) + -24|0);
                   $218 = ($219<<24>>24)<(0);
                   $220 = HEAP32[$10>>2]|0;
                   $221 = $218 ? $220 : $10;
                   $222 = $219&255;
                   $223 = $218 ? $224 : $222;
                   __THREW__ = 0;
                   (invoke_iiii(1112,($216|0),($221|0),($223|0))|0);
                   $225 = __THREW__; __THREW__ = 0;
                   $226 = $225&1;
                   if ($226) {
                    label = 87;
                    break;
                   } else {
                    $$3144$off0$ph = 0;
                   }
                  }
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
                  $$0141$off0$ph = $$3144$off0$ph;$$1167$ph = $153;
                 }
                 if ((label|0) == 62) {
                  $161 = ___cxa_find_matching_catch_2()|0;
                  $162 = tempRet0;
                  $$4151 = $161;$$4164 = $162;
                  break L69;
                 }
                 else if ((label|0) == 70) {
                  $175 = ___cxa_find_matching_catch_2()|0;
                  $176 = tempRet0;
                  $$2149 = $175;$$2162 = $176;
                 }
                 else if ((label|0) == 74) {
                  $191 = ___cxa_find_matching_catch_2()|0;
                  $192 = tempRet0;
                  $$0147 = $191;$$0160 = $192;
                  label = 76;
                 }
                 else if ((label|0) == 75) {
                  $193 = ___cxa_find_matching_catch_2()|0;
                  $194 = tempRet0;
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
                  $$0147 = $193;$$0160 = $194;
                  label = 76;
                 }
                 else if ((label|0) == 87) {
                  $214 = ___cxa_find_matching_catch_2()|0;
                  $215 = tempRet0;
                  $$2149 = $214;$$2162 = $215;
                 }
                 else if ((label|0) == 92) {
                  $227 = $156;
                  $228 = $157;
                  $$3169 = $$1167;$230 = $228;$231 = $227;
                  break;
                 }
                 else if ((label|0) == 105) {
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
                  $$1135 = $0;$$4157 = 0;
                  break L17;
                 }
                 if ((label|0) == 76) {
                  $$2149 = $$0147;$$2162 = $$0160;
                 }
                 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
                 $$4151 = $$2149;$$4164 = $$2162;
                 break L69;
                }
               } while(0);
               $229 = ($230|0)==($231|0);
               if ($229) {
                $$1135 = $0;$$4157 = 0;
                break L17;
               }
               $232 = ((($231)) + -24|0);
               __THREW__ = 0;
               invoke_vii(1117,($232|0),41);
               $233 = __THREW__; __THREW__ = 0;
               $234 = $233&1;
               if ($234) {
                label = 34;
               } else {
                $235 = $30 & 1;
                $236 = ($235|0)==(0);
                if (!($236)) {
                 $237 = HEAP32[$39>>2]|0;
                 $238 = ((($237)) + -24|0);
                 __THREW__ = 0;
                 (invoke_iii(1119,($238|0),(159028|0))|0);
                 $239 = __THREW__; __THREW__ = 0;
                 $240 = $239&1;
                 if ($240) {
                  label = 34;
                  break;
                 }
                }
                $241 = $30 & 2;
                $242 = ($241|0)==(0);
                if (!($242)) {
                 $243 = HEAP32[$39>>2]|0;
                 $244 = ((($243)) + -24|0);
                 __THREW__ = 0;
                 (invoke_iii(1119,($244|0),(159035|0))|0);
                 $245 = __THREW__; __THREW__ = 0;
                 $246 = $245&1;
                 if ($246) {
                  label = 34;
                  break;
                 }
                }
                $247 = $30 & 4;
                $248 = ($247|0)==(0);
                if (!($248)) {
                 $249 = HEAP32[$39>>2]|0;
                 $250 = ((($249)) + -24|0);
                 __THREW__ = 0;
                 (invoke_iii(1119,($250|0),(159045|0))|0);
                 $251 = __THREW__; __THREW__ = 0;
                 $252 = $251&1;
                 if ($252) {
                  label = 34;
                  break;
                 }
                }
                switch ($32|0) {
                case 1:  {
                 $253 = HEAP32[$39>>2]|0;
                 $254 = ((($253)) + -24|0);
                 __THREW__ = 0;
                 (invoke_iii(1119,($254|0),(160509|0))|0);
                 $255 = __THREW__; __THREW__ = 0;
                 $256 = $255&1;
                 if ($256) {
                  label = 34;
                  break L69;
                 }
                 break;
                }
                case 2:  {
                 $257 = HEAP32[$39>>2]|0;
                 $258 = ((($257)) + -24|0);
                 __THREW__ = 0;
                 (invoke_iii(1119,($258|0),(160512|0))|0);
                 $259 = __THREW__; __THREW__ = 0;
                 $260 = $259&1;
                 if ($260) {
                  label = 34;
                  break L69;
                 }
                 break;
                }
                default: {
                }
                }
                $261 = HEAP32[$39>>2]|0;
                $262 = ((($261)) + -24|0);
                $263 = ((($8)) + 11|0);
                $264 = HEAP8[$263>>0]|0;
                $265 = ($264<<24>>24)<(0);
                $266 = HEAP32[$8>>2]|0;
                $267 = $265 ? $266 : $8;
                $268 = ((($8)) + 4|0);
                $269 = HEAP32[$268>>2]|0;
                $270 = $264&255;
                $271 = $265 ? $269 : $270;
                __THREW__ = 0;
                (invoke_iiii(1112,($262|0),($267|0),($271|0))|0);
                $272 = __THREW__; __THREW__ = 0;
                $273 = $272&1;
                if ($273) {
                 label = 34;
                } else {
                 $$1135 = $$3169;$$4157 = 1;
                 break L17;
                }
               }
              }
             }
            } while(0);
            if ((label|0) == 34) {
             $86 = ___cxa_find_matching_catch_2()|0;
             $87 = tempRet0;
             $$4151 = $86;$$4164 = $87;
            }
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbED2Ev($7);
            $$5152 = $$4151;$$5165 = $$4164;
            break L6;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbED2Ev($7);
         if ($$4157) {
          $$2136 = $$1135;
         } else {
          $$5139 = $$1135;$$6159 = 0;
          label = 109;
          break L6;
         }
        }
       }
      } while(0);
      $$5139 = $$2136;$$6159 = 1;
      label = 109;
     }
    }
    }
   } while(0);
   if ((label|0) == 6) {
    $24 = ___cxa_find_matching_catch_2()|0;
    $25 = tempRet0;
    $$5152 = $24;$$5165 = $25;
   }
   else if ((label|0) == 109) {
    __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbED2Ev($6);
    __ZN10__cxxabiv112_GLOBAL__N_110save_valueIjED2Ev($5);
    if ($$6159) {
     $$6140 = $$5139;
     label = 111;
     break;
    } else {
     $$11 = $0;
     break;
    }
   }
   __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbED2Ev($6);
   __ZN10__cxxabiv112_GLOBAL__N_110save_valueIjED2Ev($5);
   ___resumeException($$5152|0);
   // unreachable;
  }
 } while(0);
 if ((label|0) == 111) {
  $$11 = $$6140;
 }
 STACKTOP = sp;return ($$11|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_dot_suffixINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i16 = 0, $$012 = 0, $$014 = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$1 = 0, $$113 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0;
 var $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0;
 var $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0;
 var $57 = 0, $58 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i17 = 0, $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 36|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = ($0|0)==($1|0);
 L1: do {
  if ($7) {
   $$014 = $0;
  } else {
   $8 = HEAP8[$0>>0]|0;
   $9 = ($8<<24>>24)==(46);
   if ($9) {
    $10 = HEAP32[$2>>2]|0;
    $11 = ((($2)) + 4|0);
    $12 = HEAP32[$11>>2]|0;
    $13 = ($10|0)==($12|0);
    if ($13) {
     $$014 = $0;
    } else {
     ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
     $14 = $0;
     $15 = $1;
     $16 = (($15) - ($14))|0;
     $17 = ($16>>>0)>(4294967279);
     if ($17) {
      __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($6);
      // unreachable;
     }
     $18 = ($16>>>0)<(11);
     if ($18) {
      $19 = $16&255;
      $20 = ((($6)) + 11|0);
      HEAP8[$20>>0] = $19;
      $$021$ph$i$i = $6;
     } else {
      $21 = (($16) + 16)|0;
      $22 = $21 & -16;
      $23 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($22)|0);
      HEAP32[$6>>2] = $23;
      $24 = $22 | -2147483648;
      $25 = ((($6)) + 8|0);
      HEAP32[$25>>2] = $24;
      $26 = ((($6)) + 4|0);
      HEAP32[$26>>2] = $16;
      $$021$ph$i$i = $23;
     }
     $$0$i$i = $0;$$021$i$i = $$021$ph$i$i;
     while(1) {
      $27 = ($$0$i$i|0)==($1|0);
      if ($27) {
       break;
      }
      __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
      $28 = ((($$0$i$i)) + 1|0);
      $29 = ((($$021$i$i)) + 1|0);
      $$0$i$i = $28;$$021$i$i = $29;
     }
     $scevgep23$i$i = (($$021$ph$i$i) + ($16)|0);
     HEAP8[$3>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$3);
     __THREW__ = 0;
     $30 = (invoke_iiii(1121,($6|0),0,(159078|0))|0);
     $31 = __THREW__; __THREW__ = 0;
     $32 = $31&1;
     if ($32) {
      $53 = ___cxa_find_matching_catch_2()|0;
      $54 = tempRet0;
      $$1 = $54;$$113 = $53;
     } else {
      ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
      $$0$i$i$i = 0;
      while(1) {
       $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
       if ($exitcond$i$i$i) {
        break;
       }
       $33 = (($6) + ($$0$i$i$i<<2)|0);
       HEAP32[$33>>2] = 0;
       $34 = (($$0$i$i$i) + 1)|0;
       $$0$i$i$i = $34;
      }
      __THREW__ = 0;
      $35 = (invoke_iii(1119,($5|0),(159081|0))|0);
      $36 = __THREW__; __THREW__ = 0;
      $37 = $36&1;
      do {
       if ($37) {
        $55 = ___cxa_find_matching_catch_2()|0;
        $56 = tempRet0;
        $$0 = $56;$$012 = $55;
       } else {
        ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
        $$0$i$i$i16 = 0;
        while(1) {
         $exitcond$i$i$i17 = ($$0$i$i$i16|0)==(3);
         if ($exitcond$i$i$i17) {
          break;
         }
         $38 = (($5) + ($$0$i$i$i16<<2)|0);
         HEAP32[$38>>2] = 0;
         $39 = (($$0$i$i$i16) + 1)|0;
         $$0$i$i$i16 = $39;
        }
        $40 = HEAP32[$11>>2]|0;
        $41 = ((($40)) + -24|0);
        $42 = ((($4)) + 11|0);
        $43 = HEAP8[$42>>0]|0;
        $44 = ($43<<24>>24)<(0);
        $45 = HEAP32[$4>>2]|0;
        $46 = $44 ? $45 : $4;
        $47 = ((($4)) + 4|0);
        $48 = HEAP32[$47>>2]|0;
        $49 = $43&255;
        $50 = $44 ? $48 : $49;
        __THREW__ = 0;
        (invoke_iiii(1112,($41|0),($46|0),($50|0))|0);
        $51 = __THREW__; __THREW__ = 0;
        $52 = $51&1;
        if ($52) {
         $57 = ___cxa_find_matching_catch_2()|0;
         $58 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         $$0 = $58;$$012 = $57;
         break;
        } else {
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         $$014 = $1;
         break L1;
        }
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
      $$1 = $$0;$$113 = $$012;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
     ___resumeException($$113|0);
     // unreachable;
    }
   } else {
    $$014 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$014|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_block_invokeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$031 = 0, $$032 = 0, $$133 = 0, $$2 = 0, $$236 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit40 = 0, $isdigittmp = 0, $isdigittmp39 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $1;
 $4 = $0;
 $5 = (($3) - ($4))|0;
 $6 = ($5|0)>(12);
 L1: do {
  if ($6) {
   $$031 = 0;$$032 = $0;
   while(1) {
    $7 = ($$031|0)<(13);
    $8 = ((($$032)) + 1|0);
    if (!($7)) {
     break;
    }
    $9 = HEAP8[$$032>>0]|0;
    $10 = (160714 + ($$031)|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = ($9<<24>>24)==($11<<24>>24);
    $13 = (($$031) + 1)|0;
    if ($12) {
     $$031 = $13;$$032 = $8;
    } else {
     $$236 = $0;
     break L1;
    }
   }
   $14 = ($$032|0)==($1|0);
   L6: do {
    if ($14) {
     $$2 = $1;
    } else {
     $15 = HEAP8[$$032>>0]|0;
     $16 = ($15<<24>>24)==(95);
     if ($16) {
      $17 = ($8|0)==($1|0);
      if ($17) {
       $$236 = $0;
       break L1;
      }
      $18 = HEAP8[$8>>0]|0;
      $19 = $18 << 24 >> 24;
      $isdigittmp39 = (($19) + -48)|0;
      $isdigit40 = ($isdigittmp39>>>0)<(10);
      if (!($isdigit40)) {
       $$236 = $0;
       break L1;
      }
      $20 = ((($$032)) + 2|0);
      $$133 = $20;
     } else {
      $$133 = $$032;
     }
     while(1) {
      $21 = ($$133|0)==($1|0);
      if ($21) {
       $$2 = $1;
       break L6;
      }
      $22 = HEAP8[$$133>>0]|0;
      $23 = $22 << 24 >> 24;
      $isdigittmp = (($23) + -48)|0;
      $isdigit = ($isdigittmp>>>0)<(10);
      if (!($isdigit)) {
       $$2 = $$133;
       break L6;
      }
      $24 = ((($$133)) + 1|0);
      $$133 = $24;
     }
    }
   } while(0);
   $25 = HEAP32[$2>>2]|0;
   $26 = ((($2)) + 4|0);
   $27 = HEAP32[$26>>2]|0;
   $28 = ($25|0)==($27|0);
   if ($28) {
    $$236 = $0;
   } else {
    $29 = ((($27)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($29,0,160728)|0);
    $$236 = $$2;
   }
  } else {
   $$236 = $0;
  }
 } while(0);
 return ($$236|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0 = 0, $$0$i = 0, $$0$i$i$i = 0, $$0$i$i$i568 = 0, $$0$i$i$i572 = 0, $$0$i$i$i575 = 0, $$0$i531 = 0, $$0$i538 = 0, $$0434 = 0, $$0435 = 0, $$0438 = 0, $$0444 = 0, $$0445 = 0, $$0458 = 0, $$0477 = 0, $$0478 = 0, $$10455 = 0, $$10468 = 0, $$11456 = 0;
 var $$11469 = 0, $$12457 = 0, $$12470 = 0, $$13 = 0, $$13471 = 0, $$14 = 0, $$1429 = 0, $$1436 = 0, $$1446 = 0, $$14472 = 0, $$1459 = 0, $$15 = 0, $$15473 = 0, $$16 = 0, $$16474 = 0, $$2437 = 0, $$3448 = 0, $$3461 = 0, $$4449 = 0, $$4462 = 0;
 var $$5 = 0, $$6451 = 0, $$6464 = 0, $$7452 = 0, $$7465 = 0, $$8453 = 0, $$8466 = 0, $$9454 = 0, $$9467 = 0, $$in = 0, $$ph = 0, $$ph581 = 0, $$ph581615 = 0, $$pre$i = 0, $$pre$i495 = 0, $$pre$i546 = 0, $$pre$i556 = 0, $$pre$i563 = 0, $$sroa$0$0$$sroa_idx$i = 0, $$sroa$0$0$$sroa_idx$i482 = 0;
 var $$sroa$0$0$$sroa_idx$i486 = 0, $$sroa$0$0$$sroa_idx$i488 = 0, $$sroa$0$0$$sroa_idx$i492 = 0, $$sroa$0$0$$sroa_idx$i496 = 0, $$sroa$0$0$$sroa_idx$i502 = 0, $$sroa$0$0$$sroa_idx$i506 = 0, $$sroa$0$0$$sroa_idx$i510 = 0, $$sroa$0$0$$sroa_idx$i512 = 0, $$sroa$0$0$$sroa_idx$i516 = 0, $$sroa$0$0$$sroa_idx$i520 = 0, $$sroa$0$0$$sroa_idx$i524 = 0, $$sroa$0$0$$sroa_idx$i528 = 0, $$sroa$0$0$$sroa_idx$i535 = 0, $$sroa$0$0$$sroa_idx$i542 = 0, $$sroa$0$0$$sroa_idx$i579 = 0, $$sroa$0$0$copyload$i483592 = 0, $$sroa$0$0$copyload$i487593 = 0, $$sroa$0$0$copyload$i489594 = 0, $$sroa$0$0$copyload$i493595 = 0, $$sroa$0$0$copyload$i497590 = 0;
 var $$sroa$0$0$copyload$i503607 = 0, $$sroa$0$0$copyload$i507606 = 0, $$sroa$0$0$copyload$i511596 = 0, $$sroa$0$0$copyload$i513605 = 0, $$sroa$0$0$copyload$i517604 = 0, $$sroa$0$0$copyload$i521603 = 0, $$sroa$0$0$copyload$i525602 = 0, $$sroa$0$0$copyload$i529601 = 0, $$sroa$0$0$copyload$i536600 = 0, $$sroa$0$0$copyload$i543598 = 0, $$sroa$0$0$copyload$i551599 = 0, $$sroa$0$0$copyload$i580597 = 0, $$sroa$0$0$copyload$i591 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0;
 var $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0, $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0;
 var $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0;
 var $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0;
 var $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0;
 var $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0;
 var $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0;
 var $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0;
 var $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0;
 var $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0;
 var $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0;
 var $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0;
 var $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0;
 var $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0;
 var $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0;
 var $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0;
 var $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0;
 var $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0;
 var $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0;
 var $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0;
 var $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0;
 var $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0;
 var $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0;
 var $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0;
 var $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0;
 var $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0;
 var $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0;
 var $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0;
 var $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0;
 var $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0;
 var $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0, $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0;
 var $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0, $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0;
 var $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0, $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0;
 var $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0, $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0;
 var $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0, $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0;
 var $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0, $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0;
 var $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0, $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0;
 var $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0, $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0;
 var $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0, $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0;
 var $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0, $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0;
 var $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0, $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0;
 var $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0, $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0;
 var $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0, $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0;
 var $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0, $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0;
 var $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0, $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0;
 var $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0, $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0;
 var $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0, $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0;
 var $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0, $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0;
 var $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0, $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0;
 var $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0, $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0;
 var $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0, $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0;
 var $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0, $996 = 0, $997 = 0, $998 = 0, $999 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i569 = 0, $exitcond$i$i$i573 = 0, $exitcond$i$i$i576 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i559 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 592|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(592|0);
 $3 = sp + 588|0;
 $4 = sp + 584|0;
 $5 = sp + 568|0;
 $6 = sp + 560|0;
 $7 = sp + 544|0;
 $8 = sp + 536|0;
 $9 = sp + 520|0;
 $10 = sp + 512|0;
 $11 = sp + 496|0;
 $12 = sp + 488|0;
 $13 = sp + 472|0;
 $14 = sp + 468|0;
 $15 = sp + 464|0;
 $16 = sp + 452|0;
 $17 = sp + 448|0;
 $18 = sp + 436|0;
 $19 = sp + 424|0;
 $20 = sp + 420|0;
 $21 = sp + 408|0;
 $22 = sp + 404|0;
 $23 = sp + 392|0;
 $24 = sp + 376|0;
 $25 = sp + 368|0;
 $26 = sp + 356|0;
 $27 = sp + 344|0;
 $28 = sp + 320|0;
 $29 = sp + 308|0;
 $30 = sp + 296|0;
 $31 = sp + 284|0;
 $32 = sp + 272|0;
 $33 = sp + 248|0;
 $34 = sp + 236|0;
 $35 = sp + 224|0;
 $36 = sp + 212|0;
 $37 = sp + 200|0;
 $38 = sp + 176|0;
 $39 = sp + 164|0;
 $40 = sp + 152|0;
 $41 = sp + 136|0;
 $42 = sp + 128|0;
 $43 = sp + 112|0;
 $44 = sp + 108|0;
 $45 = sp + 96|0;
 $46 = sp + 80|0;
 $47 = sp + 76|0;
 $48 = sp + 72|0;
 $49 = sp + 56|0;
 $50 = sp + 48|0;
 $51 = sp + 32|0;
 $52 = sp + 24|0;
 $53 = sp + 8|0;
 $54 = sp;
 $55 = ($0|0)==($1|0);
 L1: do {
  if ($55) {
   $$5 = $0;
  } else {
   $56 = HEAP8[$0>>0]|0;
   $57 = $56 << 24 >> 24;
   $58 = ((($2)) + 16|0);
   switch ($57|0) {
   case 75: case 86: case 114:  {
    HEAP32[$3>>2] = 0;
    $59 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_cv_qualifiersEPKcS2_Rj($0,$1,$3)|0);
    $60 = ($59|0)==($0|0);
    L5: do {
     if ($60) {
      $$1429 = $0;
     } else {
      $61 = HEAP8[$59>>0]|0;
      $62 = ($61<<24>>24)==(70);
      $63 = ((($2)) + 4|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = HEAP32[$2>>2]|0;
      $66 = (($64) - ($65))|0;
      $67 = (($66|0) / 24)&-1;
      $68 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($59,$1,$2)|0);
      $69 = HEAP32[$63>>2]|0;
      $70 = HEAP32[$2>>2]|0;
      $71 = (($69) - ($70))|0;
      $72 = (($71|0) / 24)&-1;
      $73 = ($68|0)==($59|0);
      if ($73) {
       $$1429 = $0;
      } else {
       $74 = ((($2)) + 20|0);
       $75 = HEAP32[$74>>2]|0;
       L8: do {
        if ($62) {
         $76 = ((($75)) + -16|0);
         $78 = $75;
         while(1) {
          $77 = ($78|0)==($76|0);
          if ($77) {
           $83 = $76;
           break L8;
          }
          $79 = ((($78)) + -16|0);
          HEAP32[$74>>2] = $79;
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($79);
          $$pre$i495 = HEAP32[$74>>2]|0;
          $78 = $$pre$i495;
         }
        } else {
         $83 = $75;
        }
       } while(0);
       $$sroa$0$0$$sroa_idx$i496 = ((($2)) + 12|0);
       $$sroa$0$0$copyload$i497590 = HEAP32[$$sroa$0$0$$sroa_idx$i496>>2]|0;
       HEAP32[$4>>2] = $$sroa$0$0$copyload$i497590;
       $80 = ((($2)) + 24|0);
       $81 = HEAP32[$80>>2]|0;
       $82 = ($83>>>0)<($81>>>0);
       if ($82) {
        HEAP32[$83>>2] = 0;
        $84 = ((($83)) + 4|0);
        HEAP32[$84>>2] = 0;
        $85 = ((($83)) + 8|0);
        $86 = HEAP32[$4>>2]|0;
        HEAP32[$85>>2] = 0;
        $87 = ((($83)) + 12|0);
        HEAP32[$87>>2] = $86;
        $88 = HEAP32[$74>>2]|0;
        $89 = ((($88)) + 16|0);
        HEAP32[$74>>2] = $89;
       } else {
        __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$4);
       }
       $90 = HEAP32[$3>>2]|0;
       $91 = $90 & 1;
       $92 = ($91|0)==(0);
       $93 = $90 & 2;
       $94 = ($93|0)==(0);
       $95 = $90 & 4;
       $96 = ($95|0)==(0);
       $$0434 = $67;
       while(1) {
        $97 = ($$0434>>>0)<($72>>>0);
        if (!($97)) {
         $$1429 = $68;
         break L5;
        }
        if ($62) {
         $98 = HEAP32[$2>>2]|0;
         $99 = (((($98) + (($$0434*24)|0)|0)) + 12|0);
         $100 = ((($99)) + 11|0);
         $101 = HEAP8[$100>>0]|0;
         $102 = ($101<<24>>24)<(0);
         if ($102) {
          $103 = (((($98) + (($$0434*24)|0)|0)) + 16|0);
          $104 = HEAP32[$103>>2]|0;
          $105 = HEAP32[$99>>2]|0;
          $$in = $104;$109 = $105;
         } else {
          $106 = $101&255;
          $$in = $106;$109 = $99;
         }
         $107 = (($$in) + -2)|0;
         $108 = (($109) + ($107)|0);
         $110 = HEAP8[$108>>0]|0;
         $111 = ($110<<24>>24)==(38);
         if ($111) {
          $112 = (($$in) + -3)|0;
          $$0435 = $112;
         } else {
          if ($102) {
           $113 = HEAP32[$99>>2]|0;
           $114 = (((($98) + (($$0434*24)|0)|0)) + 16|0);
           $115 = HEAP32[$114>>2]|0;
           $118 = $113;$119 = $115;
          } else {
           $116 = $101&255;
           $118 = $99;$119 = $116;
          }
          $117 = (($118) + ($119)|0);
          $120 = ((($117)) + -1|0);
          $121 = HEAP8[$120>>0]|0;
          $122 = ($121<<24>>24)==(38);
          $$ = $122 ? $107 : $$in;
          $$0435 = $$;
         }
         if ($92) {
          $$1436 = $$0435;
         } else {
          $123 = (($$0435) + 6)|0;
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($99,$$0435,159028)|0);
          $$1436 = $123;
         }
         if ($94) {
          $$2437 = $$1436;
         } else {
          $124 = (($$1436) + 9)|0;
          $125 = HEAP32[$2>>2]|0;
          $126 = (((($125) + (($$0434*24)|0)|0)) + 12|0);
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($126,$$1436,159035)|0);
          $$2437 = $124;
         }
         if (!($96)) {
          $127 = HEAP32[$2>>2]|0;
          $128 = (((($127) + (($$0434*24)|0)|0)) + 12|0);
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($128,$$2437,159045)|0);
         }
        } else {
         if (!($92)) {
          $129 = HEAP32[$2>>2]|0;
          $130 = (($129) + (($$0434*24)|0)|0);
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($130,159028)|0);
         }
         if (!($94)) {
          $131 = HEAP32[$2>>2]|0;
          $132 = (($131) + (($$0434*24)|0)|0);
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($132,159035)|0);
         }
         if (!($96)) {
          $133 = HEAP32[$2>>2]|0;
          $134 = (($133) + (($$0434*24)|0)|0);
          (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($134,159045)|0);
         }
        }
        $135 = HEAP32[$74>>2]|0;
        $136 = HEAP32[$2>>2]|0;
        $137 = (($136) + (($$0434*24)|0)|0);
        $138 = ((($135)) + -12|0);
        $139 = HEAP32[$138>>2]|0;
        $140 = ((($135)) + -8|0);
        $141 = HEAP32[$140>>2]|0;
        $142 = ($139|0)==($141|0);
        if ($142) {
         $145 = ((($135)) + -16|0);
         __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($145,$137);
        } else {
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($139,$137);
         $143 = HEAP32[$138>>2]|0;
         $144 = ((($143)) + 24|0);
         HEAP32[$138>>2] = $144;
        }
        $146 = (($$0434) + 1)|0;
        $$0434 = $146;
       }
      }
     }
    } while(0);
    $$5 = $$1429;
    break L1;
    break;
   }
   default: {
   }
   }
   $147 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_builtin_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
   $148 = ($147|0)==($0|0);
   if ($148) {
    $149 = HEAP8[$0>>0]|0;
    $150 = $149 << 24 >> 24;
    $151 = ((($0)) + 1|0);
    $152 = ($151|0)!=($1|0);
    L58: do {
     switch ($150|0) {
     case 65:  {
      $153 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_array_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $154 = ($153|0)==($0|0);
      if ($154) {
       $$5 = $0;
       break L1;
      }
      $155 = HEAP32[$2>>2]|0;
      $156 = ((($2)) + 4|0);
      $157 = HEAP32[$156>>2]|0;
      $158 = ($155|0)==($157|0);
      if ($158) {
       $$5 = $0;
       break L1;
      }
      $159 = ((($157)) + -24|0);
      $$sroa$0$0$$sroa_idx$i502 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i503607 = HEAP32[$$sroa$0$0$$sroa_idx$i502>>2]|0;
      HEAP32[$6>>2] = $$sroa$0$0$copyload$i503607;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($5,$159,$6);
      $160 = ((($2)) + 20|0);
      $161 = HEAP32[$160>>2]|0;
      $162 = ((($2)) + 24|0);
      $163 = HEAP32[$162>>2]|0;
      $164 = ($161>>>0)<($163>>>0);
      if ($164) {
       $165 = ((($5)) + 12|0);
       HEAP32[$161>>2] = 0;
       $166 = ((($161)) + 4|0);
       HEAP32[$166>>2] = 0;
       $167 = ((($161)) + 8|0);
       $168 = HEAP32[$165>>2]|0;
       HEAP32[$167>>2] = 0;
       $169 = ((($161)) + 12|0);
       HEAP32[$169>>2] = $168;
       $170 = HEAP32[$5>>2]|0;
       HEAP32[$161>>2] = $170;
       $171 = ((($5)) + 4|0);
       $172 = HEAP32[$171>>2]|0;
       HEAP32[$166>>2] = $172;
       $173 = ((($5)) + 8|0);
       $174 = HEAP32[$173>>2]|0;
       $175 = ((($161)) + 8|0);
       HEAP32[$175>>2] = $174;
       HEAP32[$173>>2] = 0;
       HEAP32[$171>>2] = 0;
       HEAP32[$5>>2] = 0;
       $176 = HEAP32[$160>>2]|0;
       $177 = ((($176)) + 16|0);
       HEAP32[$160>>2] = $177;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($5|0));
       $178 = __THREW__; __THREW__ = 0;
       $179 = $178&1;
       if ($179) {
        $180 = ___cxa_find_matching_catch_2()|0;
        $181 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($5);
        $$16 = $180;$$16474 = $181;
        break L58;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($5);
      $$5 = $153;
      break L1;
      break;
     }
     case 67:  {
      $182 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $183 = ($182|0)==($151|0);
      if ($183) {
       $$5 = $0;
       break L1;
      }
      $184 = HEAP32[$2>>2]|0;
      $185 = ((($2)) + 4|0);
      $186 = HEAP32[$185>>2]|0;
      $187 = ($184|0)==($186|0);
      if ($187) {
       $$5 = $0;
       break L1;
      }
      $188 = ((($186)) + -24|0);
      (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($188,159055)|0);
      $189 = HEAP32[$185>>2]|0;
      $190 = ((($189)) + -24|0);
      $$sroa$0$0$$sroa_idx$i506 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i507606 = HEAP32[$$sroa$0$0$$sroa_idx$i506>>2]|0;
      HEAP32[$8>>2] = $$sroa$0$0$copyload$i507606;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($7,$190,$8);
      $191 = ((($2)) + 20|0);
      $192 = HEAP32[$191>>2]|0;
      $193 = ((($2)) + 24|0);
      $194 = HEAP32[$193>>2]|0;
      $195 = ($192>>>0)<($194>>>0);
      if ($195) {
       $196 = ((($7)) + 12|0);
       HEAP32[$192>>2] = 0;
       $197 = ((($192)) + 4|0);
       HEAP32[$197>>2] = 0;
       $198 = ((($192)) + 8|0);
       $199 = HEAP32[$196>>2]|0;
       HEAP32[$198>>2] = 0;
       $200 = ((($192)) + 12|0);
       HEAP32[$200>>2] = $199;
       $201 = HEAP32[$7>>2]|0;
       HEAP32[$192>>2] = $201;
       $202 = ((($7)) + 4|0);
       $203 = HEAP32[$202>>2]|0;
       HEAP32[$197>>2] = $203;
       $204 = ((($7)) + 8|0);
       $205 = HEAP32[$204>>2]|0;
       $206 = ((($192)) + 8|0);
       HEAP32[$206>>2] = $205;
       HEAP32[$204>>2] = 0;
       HEAP32[$202>>2] = 0;
       HEAP32[$7>>2] = 0;
       $207 = HEAP32[$191>>2]|0;
       $208 = ((($207)) + 16|0);
       HEAP32[$191>>2] = $208;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($7|0));
       $209 = __THREW__; __THREW__ = 0;
       $210 = $209&1;
       if ($210) {
        $211 = ___cxa_find_matching_catch_2()|0;
        $212 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($7);
        $$16 = $211;$$16474 = $212;
        break L58;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($7);
      $$5 = $182;
      break L1;
      break;
     }
     case 70:  {
      $213 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_function_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $214 = ($213|0)==($0|0);
      if ($214) {
       $$5 = $0;
       break L1;
      }
      $215 = HEAP32[$2>>2]|0;
      $216 = ((($2)) + 4|0);
      $217 = HEAP32[$216>>2]|0;
      $218 = ($215|0)==($217|0);
      if ($218) {
       $$5 = $0;
       break L1;
      }
      $219 = ((($217)) + -24|0);
      $$sroa$0$0$$sroa_idx$i512 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i513605 = HEAP32[$$sroa$0$0$$sroa_idx$i512>>2]|0;
      HEAP32[$10>>2] = $$sroa$0$0$copyload$i513605;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($9,$219,$10);
      $220 = ((($2)) + 20|0);
      $221 = HEAP32[$220>>2]|0;
      $222 = ((($2)) + 24|0);
      $223 = HEAP32[$222>>2]|0;
      $224 = ($221>>>0)<($223>>>0);
      if ($224) {
       $225 = ((($9)) + 12|0);
       HEAP32[$221>>2] = 0;
       $226 = ((($221)) + 4|0);
       HEAP32[$226>>2] = 0;
       $227 = ((($221)) + 8|0);
       $228 = HEAP32[$225>>2]|0;
       HEAP32[$227>>2] = 0;
       $229 = ((($221)) + 12|0);
       HEAP32[$229>>2] = $228;
       $230 = HEAP32[$9>>2]|0;
       HEAP32[$221>>2] = $230;
       $231 = ((($9)) + 4|0);
       $232 = HEAP32[$231>>2]|0;
       HEAP32[$226>>2] = $232;
       $233 = ((($9)) + 8|0);
       $234 = HEAP32[$233>>2]|0;
       $235 = ((($221)) + 8|0);
       HEAP32[$235>>2] = $234;
       HEAP32[$233>>2] = 0;
       HEAP32[$231>>2] = 0;
       HEAP32[$9>>2] = 0;
       $236 = HEAP32[$220>>2]|0;
       $237 = ((($236)) + 16|0);
       HEAP32[$220>>2] = $237;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($9|0));
       $238 = __THREW__; __THREW__ = 0;
       $239 = $238&1;
       if ($239) {
        $240 = ___cxa_find_matching_catch_2()|0;
        $241 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($9);
        $$16 = $240;$$16474 = $241;
        break L58;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($9);
      $$5 = $213;
      break L1;
      break;
     }
     case 71:  {
      $242 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $243 = ($242|0)==($151|0);
      if ($243) {
       $$5 = $0;
       break L1;
      }
      $244 = HEAP32[$2>>2]|0;
      $245 = ((($2)) + 4|0);
      $246 = HEAP32[$245>>2]|0;
      $247 = ($244|0)==($246|0);
      if ($247) {
       $$5 = $0;
       break L1;
      }
      $248 = ((($246)) + -24|0);
      (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($248,159064)|0);
      $249 = HEAP32[$245>>2]|0;
      $250 = ((($249)) + -24|0);
      $$sroa$0$0$$sroa_idx$i516 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i517604 = HEAP32[$$sroa$0$0$$sroa_idx$i516>>2]|0;
      HEAP32[$12>>2] = $$sroa$0$0$copyload$i517604;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($11,$250,$12);
      $251 = ((($2)) + 20|0);
      $252 = HEAP32[$251>>2]|0;
      $253 = ((($2)) + 24|0);
      $254 = HEAP32[$253>>2]|0;
      $255 = ($252>>>0)<($254>>>0);
      if ($255) {
       $256 = ((($11)) + 12|0);
       HEAP32[$252>>2] = 0;
       $257 = ((($252)) + 4|0);
       HEAP32[$257>>2] = 0;
       $258 = ((($252)) + 8|0);
       $259 = HEAP32[$256>>2]|0;
       HEAP32[$258>>2] = 0;
       $260 = ((($252)) + 12|0);
       HEAP32[$260>>2] = $259;
       $261 = HEAP32[$11>>2]|0;
       HEAP32[$252>>2] = $261;
       $262 = ((($11)) + 4|0);
       $263 = HEAP32[$262>>2]|0;
       HEAP32[$257>>2] = $263;
       $264 = ((($11)) + 8|0);
       $265 = HEAP32[$264>>2]|0;
       $266 = ((($252)) + 8|0);
       HEAP32[$266>>2] = $265;
       HEAP32[$264>>2] = 0;
       HEAP32[$262>>2] = 0;
       HEAP32[$11>>2] = 0;
       $267 = HEAP32[$251>>2]|0;
       $268 = ((($267)) + 16|0);
       HEAP32[$251>>2] = $268;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($11|0));
       $269 = __THREW__; __THREW__ = 0;
       $270 = $269&1;
       if ($270) {
        $271 = ___cxa_find_matching_catch_2()|0;
        $272 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($11);
        $$16 = $271;$$16474 = $272;
        break L58;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($11);
      $$5 = $242;
      break L1;
      break;
     }
     case 77:  {
      $273 = (__ZN10__cxxabiv112_GLOBAL__N_128parse_pointer_to_member_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $274 = ($273|0)==($0|0);
      if ($274) {
       $$5 = $0;
       break L1;
      }
      $275 = HEAP32[$2>>2]|0;
      $276 = ((($2)) + 4|0);
      $277 = HEAP32[$276>>2]|0;
      $278 = ($275|0)==($277|0);
      if ($278) {
       $$5 = $0;
       break L1;
      }
      $279 = ((($277)) + -24|0);
      $$sroa$0$0$$sroa_idx$i520 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i521603 = HEAP32[$$sroa$0$0$$sroa_idx$i520>>2]|0;
      HEAP32[$14>>2] = $$sroa$0$0$copyload$i521603;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($13,$279,$14);
      $280 = ((($2)) + 20|0);
      $281 = HEAP32[$280>>2]|0;
      $282 = ((($2)) + 24|0);
      $283 = HEAP32[$282>>2]|0;
      $284 = ($281>>>0)<($283>>>0);
      if ($284) {
       $285 = ((($13)) + 12|0);
       HEAP32[$281>>2] = 0;
       $286 = ((($281)) + 4|0);
       HEAP32[$286>>2] = 0;
       $287 = ((($281)) + 8|0);
       $288 = HEAP32[$285>>2]|0;
       HEAP32[$287>>2] = 0;
       $289 = ((($281)) + 12|0);
       HEAP32[$289>>2] = $288;
       $290 = HEAP32[$13>>2]|0;
       HEAP32[$281>>2] = $290;
       $291 = ((($13)) + 4|0);
       $292 = HEAP32[$291>>2]|0;
       HEAP32[$286>>2] = $292;
       $293 = ((($13)) + 8|0);
       $294 = HEAP32[$293>>2]|0;
       $295 = ((($281)) + 8|0);
       HEAP32[$295>>2] = $294;
       HEAP32[$293>>2] = 0;
       HEAP32[$291>>2] = 0;
       HEAP32[$13>>2] = 0;
       $296 = HEAP32[$280>>2]|0;
       $297 = ((($296)) + 16|0);
       HEAP32[$280>>2] = $297;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($13|0));
       $298 = __THREW__; __THREW__ = 0;
       $299 = $298&1;
       if ($299) {
        $300 = ___cxa_find_matching_catch_2()|0;
        $301 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($13);
        $$16 = $300;$$16474 = $301;
        break L58;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($13);
      $$5 = $273;
      break L1;
      break;
     }
     case 79:  {
      $302 = ((($2)) + 4|0);
      $303 = HEAP32[$302>>2]|0;
      $304 = HEAP32[$2>>2]|0;
      $305 = (($303) - ($304))|0;
      $306 = (($305|0) / 24)&-1;
      $307 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $308 = HEAP32[$302>>2]|0;
      $309 = HEAP32[$2>>2]|0;
      $310 = (($308) - ($309))|0;
      $311 = (($310|0) / 24)&-1;
      $312 = ($307|0)==($151|0);
      if ($312) {
       $$5 = $0;
       break L1;
      }
      $$sroa$0$0$$sroa_idx$i524 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i525602 = HEAP32[$$sroa$0$0$$sroa_idx$i524>>2]|0;
      HEAP32[$15>>2] = $$sroa$0$0$copyload$i525602;
      $313 = ((($2)) + 20|0);
      $314 = HEAP32[$313>>2]|0;
      $315 = ((($2)) + 24|0);
      $316 = HEAP32[$315>>2]|0;
      $317 = ($314>>>0)<($316>>>0);
      if ($317) {
       HEAP32[$314>>2] = 0;
       $318 = ((($314)) + 4|0);
       HEAP32[$318>>2] = 0;
       $319 = ((($314)) + 8|0);
       HEAP32[$319>>2] = 0;
       $320 = ((($314)) + 12|0);
       HEAP32[$320>>2] = $$sroa$0$0$copyload$i525602;
       $321 = HEAP32[$313>>2]|0;
       $322 = ((($321)) + 16|0);
       HEAP32[$313>>2] = $322;
      } else {
       __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$15);
      }
      $323 = ((($16)) + 11|0);
      $324 = ((($16)) + 4|0);
      $$0478 = $306;
      while(1) {
       $325 = ($$0478>>>0)<($311>>>0);
       if (!($325)) {
        $$5 = $307;
        break L1;
       }
       $326 = HEAP32[$2>>2]|0;
       $327 = (((($326) + (($$0478*24)|0)|0)) + 12|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($16,$327,2);
       $328 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
       $329 = HEAP8[$323>>0]|0;
       $330 = ($329<<24>>24)<(0);
       $331 = HEAP32[$324>>2]|0;
       $332 = $329&255;
       $333 = $330 ? $331 : $332;
       $334 = ($328|0)==($333|0);
       if ($334) {
        __THREW__ = 0;
        $335 = (invoke_iiii(1123,($16|0),(159075|0),($328|0))|0);
        $336 = __THREW__; __THREW__ = 0;
        $337 = $336&1;
        if ($337) {
         break;
        }
        $338 = ($335|0)==(0);
        $$0$i = $338;
       } else {
        $$0$i = 0;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
       $341 = HEAP32[$2>>2]|0;
       if ($$0$i) {
        $342 = (($341) + (($$0478*24)|0)|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($342,159078)|0);
        $343 = HEAP32[$2>>2]|0;
        $344 = (((($343) + (($$0478*24)|0)|0)) + 12|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($344,0,159081)|0);
       } else {
        $345 = (((($341) + (($$0478*24)|0)|0)) + 12|0);
        $346 = ((($345)) + 11|0);
        $347 = HEAP8[$346>>0]|0;
        $348 = ($347<<24>>24)<(0);
        if ($348) {
         $349 = HEAP32[$345>>2]|0;
         $351 = $349;
        } else {
         $351 = $345;
        }
        $350 = HEAP8[$351>>0]|0;
        $352 = ($350<<24>>24)==(40);
        if ($352) {
         $353 = (($341) + (($$0478*24)|0)|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($353,159083)|0);
         $354 = HEAP32[$2>>2]|0;
         $355 = (((($354) + (($$0478*24)|0)|0)) + 12|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($355,0,159081)|0);
        }
       }
       $356 = HEAP32[$2>>2]|0;
       $357 = (($356) + (($$0478*24)|0)|0);
       (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($357,159085)|0);
       $358 = HEAP32[$313>>2]|0;
       $359 = HEAP32[$2>>2]|0;
       $360 = (($359) + (($$0478*24)|0)|0);
       $361 = ((($358)) + -12|0);
       $362 = HEAP32[$361>>2]|0;
       $363 = ((($358)) + -8|0);
       $364 = HEAP32[$363>>2]|0;
       $365 = ($362|0)==($364|0);
       if ($365) {
        $368 = ((($358)) + -16|0);
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($368,$360);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($362,$360);
        $366 = HEAP32[$361>>2]|0;
        $367 = ((($366)) + 24|0);
        HEAP32[$361>>2] = $367;
       }
       $369 = (($$0478) + 1)|0;
       $$0478 = $369;
      }
      $339 = ___cxa_find_matching_catch_3(0|0)|0;
      $340 = tempRet0;
      ___clang_call_terminate($339);
      // unreachable;
      break;
     }
     case 80:  {
      $370 = ((($2)) + 4|0);
      $371 = HEAP32[$370>>2]|0;
      $372 = HEAP32[$2>>2]|0;
      $373 = (($371) - ($372))|0;
      $374 = (($373|0) / 24)&-1;
      $375 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $376 = HEAP32[$370>>2]|0;
      $377 = HEAP32[$2>>2]|0;
      $378 = (($376) - ($377))|0;
      $379 = (($378|0) / 24)&-1;
      $380 = ($375|0)==($151|0);
      if ($380) {
       $$5 = $0;
       break L1;
      }
      $$sroa$0$0$$sroa_idx$i528 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i529601 = HEAP32[$$sroa$0$0$$sroa_idx$i528>>2]|0;
      HEAP32[$17>>2] = $$sroa$0$0$copyload$i529601;
      $381 = ((($2)) + 20|0);
      $382 = HEAP32[$381>>2]|0;
      $383 = ((($2)) + 24|0);
      $384 = HEAP32[$383>>2]|0;
      $385 = ($382>>>0)<($384>>>0);
      if ($385) {
       HEAP32[$382>>2] = 0;
       $386 = ((($382)) + 4|0);
       HEAP32[$386>>2] = 0;
       $387 = ((($382)) + 8|0);
       HEAP32[$387>>2] = 0;
       $388 = ((($382)) + 12|0);
       HEAP32[$388>>2] = $$sroa$0$0$copyload$i529601;
       $389 = HEAP32[$381>>2]|0;
       $390 = ((($389)) + 16|0);
       HEAP32[$381>>2] = $390;
      } else {
       __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$17);
      }
      $391 = ((($18)) + 11|0);
      $392 = ((($18)) + 4|0);
      $393 = ((($19)) + 11|0);
      $394 = ((($19)) + 4|0);
      $$0477 = $374;
      while(1) {
       $395 = ($$0477>>>0)<($379>>>0);
       if (!($395)) {
        $$5 = $375;
        break L1;
       }
       $396 = HEAP32[$2>>2]|0;
       $397 = (((($396) + (($$0477*24)|0)|0)) + 12|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($18,$397,2);
       $398 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
       $399 = HEAP8[$391>>0]|0;
       $400 = ($399<<24>>24)<(0);
       $401 = HEAP32[$392>>2]|0;
       $402 = $399&255;
       $403 = $400 ? $401 : $402;
       $404 = ($398|0)==($403|0);
       if ($404) {
        __THREW__ = 0;
        $405 = (invoke_iiii(1123,($18|0),(159075|0),($398|0))|0);
        $406 = __THREW__; __THREW__ = 0;
        $407 = $406&1;
        if ($407) {
         label = 108;
         break;
        }
        $408 = ($405|0)==(0);
        $$0$i531 = $408;
       } else {
        $$0$i531 = 0;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
       $411 = HEAP32[$2>>2]|0;
       if ($$0$i531) {
        $412 = (($411) + (($$0477*24)|0)|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($412,159078)|0);
        $413 = HEAP32[$2>>2]|0;
        $414 = (((($413) + (($$0477*24)|0)|0)) + 12|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($414,0,159081)|0);
       } else {
        $415 = (((($411) + (($$0477*24)|0)|0)) + 12|0);
        $416 = ((($415)) + 11|0);
        $417 = HEAP8[$416>>0]|0;
        $418 = ($417<<24>>24)<(0);
        if ($418) {
         $419 = HEAP32[$415>>2]|0;
         $421 = $419;
        } else {
         $421 = $415;
        }
        $420 = HEAP8[$421>>0]|0;
        $422 = ($420<<24>>24)==(40);
        if ($422) {
         $423 = (($411) + (($$0477*24)|0)|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($423,159083)|0);
         $424 = HEAP32[$2>>2]|0;
         $425 = (((($424) + (($$0477*24)|0)|0)) + 12|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($425,0,159081)|0);
        }
       }
       $426 = HEAP8[$151>>0]|0;
       $427 = ($426<<24>>24)==(85);
       if ($427) {
        $430 = HEAP32[$2>>2]|0;
        $431 = (($430) + (($$0477*24)|0)|0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($19,$431,12);
        $432 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159088)|0);
        $433 = HEAP8[$393>>0]|0;
        $434 = ($433<<24>>24)<(0);
        $435 = HEAP32[$394>>2]|0;
        $436 = $433&255;
        $437 = $434 ? $435 : $436;
        $438 = ($432|0)==($437|0);
        if ($438) {
         __THREW__ = 0;
         $439 = (invoke_iiii(1123,($19|0),(159088|0),($432|0))|0);
         $440 = __THREW__; __THREW__ = 0;
         $441 = $440&1;
         if ($441) {
          label = 121;
          break;
         }
         $phitmp$i = ($439|0)!=(0);
         $$ph = $phitmp$i;
        } else {
         $$ph = 1;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
        $444 = HEAP32[$2>>2]|0;
        $445 = (($444) + (($$0477*24)|0)|0);
        if ($$ph) {
         $446 = $445;
         label = 123;
        } else {
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7replaceEjjPKc($445);
        }
       } else {
        $428 = HEAP32[$2>>2]|0;
        $429 = (($428) + (($$0477*24)|0)|0);
        $446 = $429;
        label = 123;
       }
       if ((label|0) == 123) {
        label = 0;
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($446,159101)|0);
       }
       $447 = HEAP32[$381>>2]|0;
       $448 = HEAP32[$2>>2]|0;
       $449 = (($448) + (($$0477*24)|0)|0);
       $450 = ((($447)) + -12|0);
       $451 = HEAP32[$450>>2]|0;
       $452 = ((($447)) + -8|0);
       $453 = HEAP32[$452>>2]|0;
       $454 = ($451|0)==($453|0);
       if ($454) {
        $457 = ((($447)) + -16|0);
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($457,$449);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($451,$449);
        $455 = HEAP32[$450>>2]|0;
        $456 = ((($455)) + 24|0);
        HEAP32[$450>>2] = $456;
       }
       $458 = (($$0477) + 1)|0;
       $$0477 = $458;
      }
      if ((label|0) == 108) {
       $409 = ___cxa_find_matching_catch_3(0|0)|0;
       $410 = tempRet0;
       ___clang_call_terminate($409);
       // unreachable;
      }
      else if ((label|0) == 121) {
       $442 = ___cxa_find_matching_catch_3(0|0)|0;
       $443 = tempRet0;
       ___clang_call_terminate($442);
       // unreachable;
      }
      break;
     }
     case 82:  {
      $459 = ((($2)) + 4|0);
      $460 = HEAP32[$459>>2]|0;
      $461 = HEAP32[$2>>2]|0;
      $462 = (($460) - ($461))|0;
      $463 = (($462|0) / 24)&-1;
      $464 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $465 = HEAP32[$459>>2]|0;
      $466 = HEAP32[$2>>2]|0;
      $467 = (($465) - ($466))|0;
      $468 = (($467|0) / 24)&-1;
      $469 = ($464|0)==($151|0);
      if ($469) {
       $$5 = $0;
       break L1;
      }
      $$sroa$0$0$$sroa_idx$i535 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i536600 = HEAP32[$$sroa$0$0$$sroa_idx$i535>>2]|0;
      HEAP32[$20>>2] = $$sroa$0$0$copyload$i536600;
      $470 = ((($2)) + 20|0);
      $471 = HEAP32[$470>>2]|0;
      $472 = ((($2)) + 24|0);
      $473 = HEAP32[$472>>2]|0;
      $474 = ($471>>>0)<($473>>>0);
      if ($474) {
       HEAP32[$471>>2] = 0;
       $475 = ((($471)) + 4|0);
       HEAP32[$475>>2] = 0;
       $476 = ((($471)) + 8|0);
       HEAP32[$476>>2] = 0;
       $477 = ((($471)) + 12|0);
       HEAP32[$477>>2] = $$sroa$0$0$copyload$i536600;
       $478 = HEAP32[$470>>2]|0;
       $479 = ((($478)) + 16|0);
       HEAP32[$470>>2] = $479;
      } else {
       __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$20);
      }
      $480 = ((($21)) + 11|0);
      $481 = ((($21)) + 4|0);
      $$0444 = $463;
      while(1) {
       $482 = ($$0444>>>0)<($468>>>0);
       if (!($482)) {
        $$5 = $464;
        break L1;
       }
       $483 = HEAP32[$2>>2]|0;
       $484 = (((($483) + (($$0444*24)|0)|0)) + 12|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($21,$484,2);
       $485 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
       $486 = HEAP8[$480>>0]|0;
       $487 = ($486<<24>>24)<(0);
       $488 = HEAP32[$481>>2]|0;
       $489 = $486&255;
       $490 = $487 ? $488 : $489;
       $491 = ($485|0)==($490|0);
       if ($491) {
        __THREW__ = 0;
        $492 = (invoke_iiii(1123,($21|0),(159075|0),($485|0))|0);
        $493 = __THREW__; __THREW__ = 0;
        $494 = $493&1;
        if ($494) {
         break;
        }
        $495 = ($492|0)==(0);
        $$0$i538 = $495;
       } else {
        $$0$i538 = 0;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($21);
       $498 = HEAP32[$2>>2]|0;
       if ($$0$i538) {
        $499 = (($498) + (($$0444*24)|0)|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($499,159078)|0);
        $500 = HEAP32[$2>>2]|0;
        $501 = (((($500) + (($$0444*24)|0)|0)) + 12|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($501,0,159081)|0);
       } else {
        $502 = (((($498) + (($$0444*24)|0)|0)) + 12|0);
        $503 = ((($502)) + 11|0);
        $504 = HEAP8[$503>>0]|0;
        $505 = ($504<<24>>24)<(0);
        if ($505) {
         $506 = HEAP32[$502>>2]|0;
         $508 = $506;
        } else {
         $508 = $502;
        }
        $507 = HEAP8[$508>>0]|0;
        $509 = ($507<<24>>24)==(40);
        if ($509) {
         $510 = (($498) + (($$0444*24)|0)|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($510,159083)|0);
         $511 = HEAP32[$2>>2]|0;
         $512 = (((($511) + (($$0444*24)|0)|0)) + 12|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($512,0,159081)|0);
        }
       }
       $513 = HEAP32[$2>>2]|0;
       $514 = (($513) + (($$0444*24)|0)|0);
       (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($514,159103)|0);
       $515 = HEAP32[$470>>2]|0;
       $516 = HEAP32[$2>>2]|0;
       $517 = (($516) + (($$0444*24)|0)|0);
       $518 = ((($515)) + -12|0);
       $519 = HEAP32[$518>>2]|0;
       $520 = ((($515)) + -8|0);
       $521 = HEAP32[$520>>2]|0;
       $522 = ($519|0)==($521|0);
       if ($522) {
        $525 = ((($515)) + -16|0);
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($525,$517);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($519,$517);
        $523 = HEAP32[$518>>2]|0;
        $524 = ((($523)) + 24|0);
        HEAP32[$518>>2] = $524;
       }
       $526 = (($$0444) + 1)|0;
       $$0444 = $526;
      }
      $496 = ___cxa_find_matching_catch_3(0|0)|0;
      $497 = tempRet0;
      ___clang_call_terminate($496);
      // unreachable;
      break;
     }
     case 84:  {
      $527 = ((($2)) + 4|0);
      $528 = HEAP32[$527>>2]|0;
      $529 = HEAP32[$2>>2]|0;
      $530 = (($528) - ($529))|0;
      $531 = (($530|0) / 24)&-1;
      $532 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $533 = HEAP32[$527>>2]|0;
      $534 = HEAP32[$2>>2]|0;
      $535 = (($533) - ($534))|0;
      $536 = (($535|0) / 24)&-1;
      $537 = ($532|0)==($0|0);
      if ($537) {
       $$5 = $0;
       break L1;
      }
      $$sroa$0$0$$sroa_idx$i542 = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i543598 = HEAP32[$$sroa$0$0$$sroa_idx$i542>>2]|0;
      HEAP32[$22>>2] = $$sroa$0$0$copyload$i543598;
      $538 = ((($2)) + 20|0);
      $539 = HEAP32[$538>>2]|0;
      $540 = ((($2)) + 24|0);
      $541 = HEAP32[$540>>2]|0;
      $542 = ($539>>>0)<($541>>>0);
      if ($542) {
       HEAP32[$539>>2] = 0;
       $543 = ((($539)) + 4|0);
       HEAP32[$543>>2] = 0;
       $544 = ((($539)) + 8|0);
       HEAP32[$544>>2] = 0;
       $545 = ((($539)) + 12|0);
       HEAP32[$545>>2] = $$sroa$0$0$copyload$i543598;
       $546 = HEAP32[$538>>2]|0;
       $547 = ((($546)) + 16|0);
       HEAP32[$538>>2] = $547;
      } else {
       __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$22);
      }
      $$0438 = $531;
      while(1) {
       $548 = ($$0438>>>0)<($536>>>0);
       if (!($548)) {
        break;
       }
       $554 = HEAP32[$538>>2]|0;
       $555 = HEAP32[$2>>2]|0;
       $556 = (($555) + (($$0438*24)|0)|0);
       $557 = ((($554)) + -12|0);
       $558 = HEAP32[$557>>2]|0;
       $559 = ((($554)) + -8|0);
       $560 = HEAP32[$559>>2]|0;
       $561 = ($558|0)==($560|0);
       if ($561) {
        $564 = ((($554)) + -16|0);
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($564,$556);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($558,$556);
        $562 = HEAP32[$557>>2]|0;
        $563 = ((($562)) + 24|0);
        HEAP32[$557>>2] = $563;
       }
       $565 = (($$0438) + 1)|0;
       $$0438 = $565;
      }
      $549 = ((($2)) + 63|0);
      $550 = HEAP8[$549>>0]|0;
      $551 = ($550<<24>>24)!=(0);
      $552 = (($531) + 1)|0;
      $553 = ($536|0)==($552|0);
      $or$cond = $553 & $551;
      if (!($or$cond)) {
       $$5 = $532;
       break L1;
      }
      $566 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($532,$1,$2)|0);
      $567 = ($566|0)==($532|0);
      if ($567) {
       $$5 = $532;
       break L1;
      }
      $568 = HEAP32[$527>>2]|0;
      $569 = ((($568)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($23,$569);
      $570 = HEAP32[$527>>2]|0;
      $571 = ((($570)) + -24|0);
      $573 = $570;
      while(1) {
       $572 = ($573|0)==($571|0);
       if ($572) {
        break;
       }
       $574 = ((($573)) + -24|0);
       HEAP32[$527>>2] = $574;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($574);
       $$pre$i546 = HEAP32[$527>>2]|0;
       $573 = $$pre$i546;
      }
      $575 = ((($570)) + -48|0);
      $576 = ((($23)) + 11|0);
      $577 = HEAP8[$576>>0]|0;
      $578 = ($577<<24>>24)<(0);
      $579 = HEAP32[$23>>2]|0;
      $580 = $578 ? $579 : $23;
      $581 = ((($23)) + 4|0);
      $582 = HEAP32[$581>>2]|0;
      $583 = $577&255;
      $584 = $578 ? $582 : $583;
      __THREW__ = 0;
      (invoke_iiii(1112,($575|0),($580|0),($584|0))|0);
      $585 = __THREW__; __THREW__ = 0;
      $586 = $585&1;
      if ($586) {
       $609 = ___cxa_find_matching_catch_2()|0;
       $610 = tempRet0;
       $$1446 = $609;$$1459 = $610;
      } else {
       $587 = HEAP32[$527>>2]|0;
       $588 = ((($587)) + -24|0);
       $$sroa$0$0$copyload$i551599 = HEAP32[$$sroa$0$0$$sroa_idx$i542>>2]|0;
       HEAP32[$25>>2] = $$sroa$0$0$copyload$i551599;
       __THREW__ = 0;
       invoke_viii(1124,($24|0),($588|0),($25|0));
       $589 = __THREW__; __THREW__ = 0;
       $590 = $589&1;
       do {
        if ($590) {
         $611 = ___cxa_find_matching_catch_2()|0;
         $612 = tempRet0;
         $$0445 = $611;$$0458 = $612;
        } else {
         $591 = HEAP32[$538>>2]|0;
         $592 = HEAP32[$540>>2]|0;
         $593 = ($591>>>0)<($592>>>0);
         if ($593) {
          $594 = ((($24)) + 12|0);
          HEAP32[$591>>2] = 0;
          $595 = ((($591)) + 4|0);
          HEAP32[$595>>2] = 0;
          $596 = ((($591)) + 8|0);
          $597 = HEAP32[$594>>2]|0;
          HEAP32[$596>>2] = 0;
          $598 = ((($591)) + 12|0);
          HEAP32[$598>>2] = $597;
          $599 = HEAP32[$24>>2]|0;
          HEAP32[$591>>2] = $599;
          $600 = ((($24)) + 4|0);
          $601 = HEAP32[$600>>2]|0;
          HEAP32[$595>>2] = $601;
          $602 = ((($24)) + 8|0);
          $603 = HEAP32[$602>>2]|0;
          $604 = ((($591)) + 8|0);
          HEAP32[$604>>2] = $603;
          HEAP32[$602>>2] = 0;
          HEAP32[$600>>2] = 0;
          HEAP32[$24>>2] = 0;
          $605 = HEAP32[$538>>2]|0;
          $606 = ((($605)) + 16|0);
          HEAP32[$538>>2] = $606;
         } else {
          __THREW__ = 0;
          invoke_vii(1122,($58|0),($24|0));
          $607 = __THREW__; __THREW__ = 0;
          $608 = $607&1;
          if ($608) {
           $613 = ___cxa_find_matching_catch_2()|0;
           $614 = tempRet0;
           __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($24);
           $$0445 = $613;$$0458 = $614;
           break;
          }
         }
         __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($24);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($23);
         $$5 = $566;
         break L1;
        }
       } while(0);
       $$1446 = $$0445;$$1459 = $$0458;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($23);
      $$16 = $$1446;$$16474 = $$1459;
      break;
     }
     case 85:  {
      if (!($152)) {
       $$5 = $0;
       break L1;
      }
      $615 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($151,$1,$2)|0);
      $616 = ($615|0)==($151|0);
      if ($616) {
       $$5 = $0;
       break L1;
      }
      $617 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($615,$1,$2)|0);
      $618 = ($617|0)==($615|0);
      if ($618) {
       $$5 = $0;
       break L1;
      }
      $619 = ((($2)) + 4|0);
      $620 = HEAP32[$619>>2]|0;
      $621 = HEAP32[$2>>2]|0;
      $622 = (($620) - ($621))|0;
      $623 = (($622|0) / 24)&-1;
      $624 = ($623>>>0)<(2);
      if ($624) {
       $$5 = $0;
       break L1;
      }
      $625 = $620;
      $626 = ((($625)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($26,$626);
      $627 = HEAP32[$619>>2]|0;
      $628 = ((($627)) + -24|0);
      $630 = $627;
      while(1) {
       $629 = ($630|0)==($628|0);
       if ($629) {
        break;
       }
       $631 = ((($630)) + -24|0);
       HEAP32[$619>>2] = $631;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($631);
       $$pre$i556 = HEAP32[$619>>2]|0;
       $630 = $$pre$i556;
      }
      $632 = ((($627)) + -48|0);
      __THREW__ = 0;
      invoke_viii(1125,($27|0),($632|0),9);
      $633 = __THREW__; __THREW__ = 0;
      $634 = $633&1;
      L235: do {
       if ($634) {
        $673 = ___cxa_find_matching_catch_2()|0;
        $674 = tempRet0;
        $$13 = $673;$$13471 = $674;
       } else {
        $635 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159105)|0);
        $636 = ((($27)) + 11|0);
        $637 = HEAP8[$636>>0]|0;
        $638 = ($637<<24>>24)<(0);
        $639 = ((($27)) + 4|0);
        $640 = HEAP32[$639>>2]|0;
        $641 = $637&255;
        $642 = $638 ? $640 : $641;
        $643 = ($635|0)==($642|0);
        L238: do {
         if ($643) {
          __THREW__ = 0;
          $644 = (invoke_iiii(1123,($27|0),(159105|0),($635|0))|0);
          $645 = __THREW__; __THREW__ = 0;
          $646 = $645&1;
          if ($646) {
           $647 = ___cxa_find_matching_catch_3(0|0)|0;
           $648 = tempRet0;
           ___clang_call_terminate($647);
           // unreachable;
          }
          $phitmp$i559 = ($644|0)==(0);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($27);
          if ($phitmp$i559) {
           $681 = HEAP32[$619>>2]|0;
           $682 = ((($681)) + -24|0);
           __THREW__ = 0;
           invoke_vii(1120,($32|0),($682|0));
           $683 = __THREW__; __THREW__ = 0;
           $684 = $683&1;
           if ($684) {
            $671 = ___cxa_find_matching_catch_2()|0;
            $672 = tempRet0;
            $$13 = $671;$$13471 = $672;
            break L235;
           }
           $685 = HEAP32[$619>>2]|0;
           $686 = ((($685)) + -24|0);
           $688 = $685;
           while(1) {
            $687 = ($688|0)==($686|0);
            if ($687) {
             break;
            }
            $689 = ((($688)) + -24|0);
            HEAP32[$619>>2] = $689;
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($689);
            $$pre$i563 = HEAP32[$619>>2]|0;
            $688 = $$pre$i563;
           }
           $690 = ((($32)) + 11|0);
           $691 = HEAP8[$690>>0]|0;
           $692 = ($691<<24>>24)<(0);
           $693 = HEAP32[$32>>2]|0;
           $$ph581615 = ((($693)) + 9|0);
           $694 = ((($32)) + 4|0);
           $695 = HEAP32[$694>>2]|0;
           $$ph581 = ((($32)) + 9|0);
           $696 = $691&255;
           $697 = $692 ? $693 : $32;
           $698 = $692 ? $$ph581615 : $$ph581;
           $699 = $692 ? $695 : $696;
           $700 = (($697) + ($699)|0);
           __THREW__ = 0;
           $701 = (invoke_iiii(1127,($698|0),($700|0),($2|0))|0);
           $702 = __THREW__; __THREW__ = 0;
           $703 = $702&1;
           L251: do {
            if ($703) {
             $737 = ___cxa_find_matching_catch_2()|0;
             $738 = tempRet0;
             $$11456 = $737;$$11469 = $738;
            } else {
             $704 = HEAP8[$690>>0]|0;
             $705 = ($704<<24>>24)<(0);
             $706 = HEAP32[$32>>2]|0;
             $707 = $705 ? $706 : $32;
             $708 = ((($707)) + 9|0);
             $709 = ($701|0)==($708|0);
             L254: do {
              if ($709) {
               __THREW__ = 0;
               invoke_viii(1126,($40|0),($26|0),(159115|0));
               $747 = __THREW__; __THREW__ = 0;
               $748 = $747&1;
               if ($748) {
                $770 = ___cxa_find_matching_catch_2()|0;
                $771 = tempRet0;
                $$10455 = $770;$$10468 = $771;
               } else {
                $749 = HEAP8[$690>>0]|0;
                $750 = ($749<<24>>24)<(0);
                $751 = HEAP32[$32>>2]|0;
                $752 = $750 ? $751 : $32;
                $753 = ((($32)) + 4|0);
                $754 = HEAP32[$753>>2]|0;
                $755 = $749&255;
                $756 = $750 ? $754 : $755;
                __THREW__ = 0;
                $757 = (invoke_iiii(1112,($40|0),($752|0),($756|0))|0);
                $758 = __THREW__; __THREW__ = 0;
                $759 = $758&1;
                L259: do {
                 if ($759) {
                  $772 = ___cxa_find_matching_catch_2()|0;
                  $773 = tempRet0;
                  $$9454 = $772;$$9467 = $773;
                 } else {
                  ;HEAP32[$39>>2]=HEAP32[$40>>2]|0;HEAP32[$39+4>>2]=HEAP32[$40+4>>2]|0;HEAP32[$39+8>>2]=HEAP32[$40+8>>2]|0;
                  $$0$i$i$i575 = 0;
                  while(1) {
                   $exitcond$i$i$i576 = ($$0$i$i$i575|0)==(3);
                   if ($exitcond$i$i$i576) {
                    break;
                   }
                   $760 = (($40) + ($$0$i$i$i575<<2)|0);
                   HEAP32[$760>>2] = 0;
                   $761 = (($$0$i$i$i575) + 1)|0;
                   $$0$i$i$i575 = $761;
                  }
                  __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($38,$39);
                  $762 = HEAP32[$619>>2]|0;
                  $763 = ((($2)) + 8|0);
                  $764 = HEAP32[$763>>2]|0;
                  $765 = ($762>>>0)<($764>>>0);
                  do {
                   if ($765) {
                    __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($762,$38);
                    $766 = HEAP32[$619>>2]|0;
                    $767 = ((($766)) + 24|0);
                    HEAP32[$619>>2] = $767;
                   } else {
                    __THREW__ = 0;
                    invoke_vii(1128,($2|0),($38|0));
                    $768 = __THREW__; __THREW__ = 0;
                    $769 = $768&1;
                    if (!($769)) {
                     break;
                    }
                    $774 = ___cxa_find_matching_catch_2()|0;
                    $775 = tempRet0;
                    __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($38);
                    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($39);
                    $$9454 = $774;$$9467 = $775;
                    break L259;
                   }
                  } while(0);
                  __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($38);
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($39);
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($40);
                  break L254;
                 }
                } while(0);
                __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($40);
                $$10455 = $$9454;$$10468 = $$9467;
               }
               $$11456 = $$10455;$$11469 = $$10468;
               break L251;
              } else {
               __THREW__ = 0;
               invoke_viii(1126,($36|0),($26|0),(159117|0));
               $710 = __THREW__; __THREW__ = 0;
               $711 = $710&1;
               if ($711) {
                $739 = ___cxa_find_matching_catch_2()|0;
                $740 = tempRet0;
                $$8453 = $739;$$8466 = $740;
               } else {
                $712 = HEAP32[$619>>2]|0;
                $713 = ((($712)) + -24|0);
                __THREW__ = 0;
                invoke_vii(1120,($37|0),($713|0));
                $714 = __THREW__; __THREW__ = 0;
                $715 = $714&1;
                if ($715) {
                 $741 = ___cxa_find_matching_catch_2()|0;
                 $742 = tempRet0;
                 $$7452 = $741;$$7465 = $742;
                } else {
                 $716 = ((($37)) + 11|0);
                 $717 = HEAP8[$716>>0]|0;
                 $718 = ($717<<24>>24)<(0);
                 $719 = HEAP32[$37>>2]|0;
                 $720 = $718 ? $719 : $37;
                 $721 = ((($37)) + 4|0);
                 $722 = HEAP32[$721>>2]|0;
                 $723 = $717&255;
                 $724 = $718 ? $722 : $723;
                 __THREW__ = 0;
                 $725 = (invoke_iiii(1112,($36|0),($720|0),($724|0))|0);
                 $726 = __THREW__; __THREW__ = 0;
                 $727 = $726&1;
                 do {
                  if ($727) {
                   $743 = ___cxa_find_matching_catch_2()|0;
                   $744 = tempRet0;
                   $$6451 = $743;$$6464 = $744;
                  } else {
                   ;HEAP32[$35>>2]=HEAP32[$36>>2]|0;HEAP32[$35+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$35+8>>2]=HEAP32[$36+8>>2]|0;
                   $$0$i$i$i568 = 0;
                   while(1) {
                    $exitcond$i$i$i569 = ($$0$i$i$i568|0)==(3);
                    if ($exitcond$i$i$i569) {
                     break;
                    }
                    $728 = (($36) + ($$0$i$i$i568<<2)|0);
                    HEAP32[$728>>2] = 0;
                    $729 = (($$0$i$i$i568) + 1)|0;
                    $$0$i$i$i568 = $729;
                   }
                   __THREW__ = 0;
                   $730 = (invoke_iii(1119,($35|0),(159119|0))|0);
                   $731 = __THREW__; __THREW__ = 0;
                   $732 = $731&1;
                   if ($732) {
                    $745 = ___cxa_find_matching_catch_2()|0;
                    $746 = tempRet0;
                    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($35);
                    $$6451 = $745;$$6464 = $746;
                    break;
                   }
                   ;HEAP32[$34>>2]=HEAP32[$35>>2]|0;HEAP32[$34+4>>2]=HEAP32[$35+4>>2]|0;HEAP32[$34+8>>2]=HEAP32[$35+8>>2]|0;
                   $$0$i$i$i572 = 0;
                   while(1) {
                    $exitcond$i$i$i573 = ($$0$i$i$i572|0)==(3);
                    if ($exitcond$i$i$i573) {
                     break;
                    }
                    $733 = (($35) + ($$0$i$i$i572<<2)|0);
                    HEAP32[$733>>2] = 0;
                    $734 = (($$0$i$i$i572) + 1)|0;
                    $$0$i$i$i572 = $734;
                   }
                   __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($33,$34);
                   $735 = HEAP32[$619>>2]|0;
                   $736 = ((($735)) + -24|0);
                   __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($736,$33);
                   __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($33);
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($34);
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($35);
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($37);
                   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($36);
                   break L254;
                  }
                 } while(0);
                 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($37);
                 $$7452 = $$6451;$$7465 = $$6464;
                }
                __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($36);
                $$8453 = $$7452;$$8466 = $$7465;
               }
               $$11456 = $$8453;$$11469 = $$8466;
               break L251;
              }
             } while(0);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($32);
             break L238;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($32);
           $$13 = $$11456;$$13471 = $$11469;
           break L235;
          } else {
           label = 189;
          }
         } else {
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($27);
          label = 189;
         }
        } while(0);
        L300: do {
         if ((label|0) == 189) {
          __THREW__ = 0;
          invoke_viii(1126,($30|0),($26|0),(159115|0));
          $649 = __THREW__; __THREW__ = 0;
          $650 = $649&1;
          if ($650) {
           $675 = ___cxa_find_matching_catch_2()|0;
           $676 = tempRet0;
           $$4449 = $675;$$4462 = $676;
          } else {
           $651 = HEAP32[$619>>2]|0;
           $652 = ((($651)) + -24|0);
           __THREW__ = 0;
           invoke_vii(1120,($31|0),($652|0));
           $653 = __THREW__; __THREW__ = 0;
           $654 = $653&1;
           do {
            if ($654) {
             $677 = ___cxa_find_matching_catch_2()|0;
             $678 = tempRet0;
             $$3448 = $677;$$3461 = $678;
            } else {
             $655 = ((($31)) + 11|0);
             $656 = HEAP8[$655>>0]|0;
             $657 = ($656<<24>>24)<(0);
             $658 = HEAP32[$31>>2]|0;
             $659 = $657 ? $658 : $31;
             $660 = ((($31)) + 4|0);
             $661 = HEAP32[$660>>2]|0;
             $662 = $656&255;
             $663 = $657 ? $661 : $662;
             __THREW__ = 0;
             $664 = (invoke_iiii(1112,($30|0),($659|0),($663|0))|0);
             $665 = __THREW__; __THREW__ = 0;
             $666 = $665&1;
             if ($666) {
              $679 = ___cxa_find_matching_catch_2()|0;
              $680 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($31);
              $$3448 = $679;$$3461 = $680;
              break;
             }
             ;HEAP32[$29>>2]=HEAP32[$30>>2]|0;HEAP32[$29+4>>2]=HEAP32[$30+4>>2]|0;HEAP32[$29+8>>2]=HEAP32[$30+8>>2]|0;
             $$0$i$i$i = 0;
             while(1) {
              $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
              if ($exitcond$i$i$i) {
               break;
              }
              $667 = (($30) + ($$0$i$i$i<<2)|0);
              HEAP32[$667>>2] = 0;
              $668 = (($$0$i$i$i) + 1)|0;
              $$0$i$i$i = $668;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($28,$29);
             $669 = HEAP32[$619>>2]|0;
             $670 = ((($669)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($670,$28);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($28);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($29);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($31);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($30);
             break L300;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($30);
           $$4449 = $$3448;$$4462 = $$3461;
          }
          $$13 = $$4449;$$13471 = $$4462;
          break L235;
         }
        } while(0);
        $776 = HEAP32[$619>>2]|0;
        $777 = ((($776)) + -24|0);
        $$sroa$0$0$$sroa_idx$i579 = ((($2)) + 12|0);
        $$sroa$0$0$copyload$i580597 = HEAP32[$$sroa$0$0$$sroa_idx$i579>>2]|0;
        HEAP32[$42>>2] = $$sroa$0$0$copyload$i580597;
        __THREW__ = 0;
        invoke_viii(1124,($41|0),($777|0),($42|0));
        $778 = __THREW__; __THREW__ = 0;
        $779 = $778&1;
        do {
         if ($779) {
          $800 = ___cxa_find_matching_catch_2()|0;
          $801 = tempRet0;
          $$12457 = $800;$$12470 = $801;
         } else {
          $780 = ((($2)) + 20|0);
          $781 = HEAP32[$780>>2]|0;
          $782 = ((($2)) + 24|0);
          $783 = HEAP32[$782>>2]|0;
          $784 = ($781>>>0)<($783>>>0);
          if ($784) {
           $785 = ((($41)) + 12|0);
           HEAP32[$781>>2] = 0;
           $786 = ((($781)) + 4|0);
           HEAP32[$786>>2] = 0;
           $787 = ((($781)) + 8|0);
           $788 = HEAP32[$785>>2]|0;
           HEAP32[$787>>2] = 0;
           $789 = ((($781)) + 12|0);
           HEAP32[$789>>2] = $788;
           $790 = HEAP32[$41>>2]|0;
           HEAP32[$781>>2] = $790;
           $791 = ((($41)) + 4|0);
           $792 = HEAP32[$791>>2]|0;
           HEAP32[$786>>2] = $792;
           $793 = ((($41)) + 8|0);
           $794 = HEAP32[$793>>2]|0;
           $795 = ((($781)) + 8|0);
           HEAP32[$795>>2] = $794;
           HEAP32[$793>>2] = 0;
           HEAP32[$791>>2] = 0;
           HEAP32[$41>>2] = 0;
           $796 = HEAP32[$780>>2]|0;
           $797 = ((($796)) + 16|0);
           HEAP32[$780>>2] = $797;
          } else {
           __THREW__ = 0;
           invoke_vii(1122,($58|0),($41|0));
           $798 = __THREW__; __THREW__ = 0;
           $799 = $798&1;
           if ($799) {
            $802 = ___cxa_find_matching_catch_2()|0;
            $803 = tempRet0;
            __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($41);
            $$12457 = $802;$$12470 = $803;
            break;
           }
          }
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($41);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($26);
          $$5 = $617;
          break L1;
         }
        } while(0);
        $$13 = $$12457;$$13471 = $$12470;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($26);
      $$16 = $$13;$$16474 = $$13471;
      break;
     }
     case 83:  {
      if ($152) {
       $804 = HEAP8[$151>>0]|0;
       $805 = ($804<<24>>24)==(116);
       if ($805) {
        $806 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
        $807 = ($806|0)==($0|0);
        if ($807) {
         $$5 = $0;
         break L1;
        }
        $808 = HEAP32[$2>>2]|0;
        $809 = ((($2)) + 4|0);
        $810 = HEAP32[$809>>2]|0;
        $811 = ($808|0)==($810|0);
        if ($811) {
         $$5 = $0;
         break L1;
        }
        $812 = ((($810)) + -24|0);
        $$sroa$0$0$$sroa_idx$i510 = ((($2)) + 12|0);
        $$sroa$0$0$copyload$i511596 = HEAP32[$$sroa$0$0$$sroa_idx$i510>>2]|0;
        HEAP32[$44>>2] = $$sroa$0$0$copyload$i511596;
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($43,$812,$44);
        $813 = ((($2)) + 20|0);
        $814 = HEAP32[$813>>2]|0;
        $815 = ((($2)) + 24|0);
        $816 = HEAP32[$815>>2]|0;
        $817 = ($814>>>0)<($816>>>0);
        if ($817) {
         $818 = ((($43)) + 12|0);
         HEAP32[$814>>2] = 0;
         $819 = ((($814)) + 4|0);
         HEAP32[$819>>2] = 0;
         $820 = ((($814)) + 8|0);
         $821 = HEAP32[$818>>2]|0;
         HEAP32[$820>>2] = 0;
         $822 = ((($814)) + 12|0);
         HEAP32[$822>>2] = $821;
         $823 = HEAP32[$43>>2]|0;
         HEAP32[$814>>2] = $823;
         $824 = ((($43)) + 4|0);
         $825 = HEAP32[$824>>2]|0;
         HEAP32[$819>>2] = $825;
         $826 = ((($43)) + 8|0);
         $827 = HEAP32[$826>>2]|0;
         $828 = ((($814)) + 8|0);
         HEAP32[$828>>2] = $827;
         HEAP32[$826>>2] = 0;
         HEAP32[$824>>2] = 0;
         HEAP32[$43>>2] = 0;
         $829 = HEAP32[$813>>2]|0;
         $830 = ((($829)) + 16|0);
         HEAP32[$813>>2] = $830;
        } else {
         __THREW__ = 0;
         invoke_vii(1122,($58|0),($43|0));
         $831 = __THREW__; __THREW__ = 0;
         $832 = $831&1;
         if ($832) {
          $833 = ___cxa_find_matching_catch_2()|0;
          $834 = tempRet0;
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($43);
          $$16 = $833;$$16474 = $834;
          break L58;
         }
        }
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($43);
        $$5 = $806;
        break L1;
       }
      }
      $835 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_substitutionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $836 = ($835|0)==($0|0);
      if ($836) {
       $$5 = $0;
       break L1;
      }
      $837 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($835,$1,$2)|0);
      $838 = ($837|0)==($835|0);
      if ($838) {
       $$5 = $835;
       break L1;
      }
      $839 = ((($2)) + 4|0);
      $840 = HEAP32[$839>>2]|0;
      $841 = HEAP32[$2>>2]|0;
      $842 = (($840) - ($841))|0;
      $843 = (($842|0) / 24)&-1;
      $844 = ($843>>>0)<(2);
      if ($844) {
       $$5 = $835;
       break L1;
      }
      $845 = $840;
      $846 = ((($845)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($45,$846);
      $847 = HEAP32[$839>>2]|0;
      $848 = ((($847)) + -24|0);
      $850 = $847;
      while(1) {
       $849 = ($850|0)==($848|0);
       if ($849) {
        break;
       }
       $851 = ((($850)) + -24|0);
       HEAP32[$839>>2] = $851;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($851);
       $$pre$i = HEAP32[$839>>2]|0;
       $850 = $$pre$i;
      }
      $852 = ((($847)) + -48|0);
      $853 = ((($45)) + 11|0);
      $854 = HEAP8[$853>>0]|0;
      $855 = ($854<<24>>24)<(0);
      $856 = HEAP32[$45>>2]|0;
      $857 = $855 ? $856 : $45;
      $858 = ((($45)) + 4|0);
      $859 = HEAP32[$858>>2]|0;
      $860 = $854&255;
      $861 = $855 ? $859 : $860;
      __THREW__ = 0;
      (invoke_iiii(1112,($852|0),($857|0),($861|0))|0);
      $862 = __THREW__; __THREW__ = 0;
      $863 = $862&1;
      if ($863) {
       $888 = ___cxa_find_matching_catch_2()|0;
       $889 = tempRet0;
       $$15 = $888;$$15473 = $889;
      } else {
       $864 = HEAP32[$839>>2]|0;
       $865 = ((($864)) + -24|0);
       $$sroa$0$0$$sroa_idx$i492 = ((($2)) + 12|0);
       $$sroa$0$0$copyload$i493595 = HEAP32[$$sroa$0$0$$sroa_idx$i492>>2]|0;
       HEAP32[$47>>2] = $$sroa$0$0$copyload$i493595;
       __THREW__ = 0;
       invoke_viii(1124,($46|0),($865|0),($47|0));
       $866 = __THREW__; __THREW__ = 0;
       $867 = $866&1;
       do {
        if ($867) {
         $890 = ___cxa_find_matching_catch_2()|0;
         $891 = tempRet0;
         $$14 = $890;$$14472 = $891;
        } else {
         $868 = ((($2)) + 20|0);
         $869 = HEAP32[$868>>2]|0;
         $870 = ((($2)) + 24|0);
         $871 = HEAP32[$870>>2]|0;
         $872 = ($869>>>0)<($871>>>0);
         if ($872) {
          $873 = ((($46)) + 12|0);
          HEAP32[$869>>2] = 0;
          $874 = ((($869)) + 4|0);
          HEAP32[$874>>2] = 0;
          $875 = ((($869)) + 8|0);
          $876 = HEAP32[$873>>2]|0;
          HEAP32[$875>>2] = 0;
          $877 = ((($869)) + 12|0);
          HEAP32[$877>>2] = $876;
          $878 = HEAP32[$46>>2]|0;
          HEAP32[$869>>2] = $878;
          $879 = ((($46)) + 4|0);
          $880 = HEAP32[$879>>2]|0;
          HEAP32[$874>>2] = $880;
          $881 = ((($46)) + 8|0);
          $882 = HEAP32[$881>>2]|0;
          $883 = ((($869)) + 8|0);
          HEAP32[$883>>2] = $882;
          HEAP32[$881>>2] = 0;
          HEAP32[$879>>2] = 0;
          HEAP32[$46>>2] = 0;
          $884 = HEAP32[$868>>2]|0;
          $885 = ((($884)) + 16|0);
          HEAP32[$868>>2] = $885;
         } else {
          __THREW__ = 0;
          invoke_vii(1122,($58|0),($46|0));
          $886 = __THREW__; __THREW__ = 0;
          $887 = $886&1;
          if ($887) {
           $892 = ___cxa_find_matching_catch_2()|0;
           $893 = tempRet0;
           __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($46);
           $$14 = $892;$$14472 = $893;
           break;
          }
         }
         __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($46);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($45);
         $$5 = $837;
         break L1;
        }
       } while(0);
       $$15 = $$14;$$15473 = $$14472;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($45);
      $$16 = $$15;$$16474 = $$15473;
      break;
     }
     case 68:  {
      if ($152) {
       $894 = HEAP8[$151>>0]|0;
       $895 = $894 << 24 >> 24;
       switch ($895|0) {
       case 112:  {
        $896 = ((($2)) + 4|0);
        $897 = HEAP32[$896>>2]|0;
        $898 = HEAP32[$2>>2]|0;
        $899 = (($897) - ($898))|0;
        $900 = (($899|0) / 24)&-1;
        $901 = ((($0)) + 2|0);
        $902 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($901,$1,$2)|0);
        $903 = HEAP32[$896>>2]|0;
        $904 = HEAP32[$2>>2]|0;
        $905 = (($903) - ($904))|0;
        $906 = (($905|0) / 24)&-1;
        $907 = ($902|0)==($901|0);
        if ($907) {
         label = 305;
         break L58;
        }
        $$sroa$0$0$$sroa_idx$i488 = ((($2)) + 12|0);
        $$sroa$0$0$copyload$i489594 = HEAP32[$$sroa$0$0$$sroa_idx$i488>>2]|0;
        HEAP32[$48>>2] = $$sroa$0$0$copyload$i489594;
        $908 = ((($2)) + 20|0);
        $909 = HEAP32[$908>>2]|0;
        $910 = ((($2)) + 24|0);
        $911 = HEAP32[$910>>2]|0;
        $912 = ($909>>>0)<($911>>>0);
        if ($912) {
         HEAP32[$909>>2] = 0;
         $913 = ((($909)) + 4|0);
         HEAP32[$913>>2] = 0;
         $914 = ((($909)) + 8|0);
         HEAP32[$914>>2] = 0;
         $915 = ((($909)) + 12|0);
         HEAP32[$915>>2] = $$sroa$0$0$copyload$i489594;
         $916 = HEAP32[$908>>2]|0;
         $917 = ((($916)) + 16|0);
         HEAP32[$908>>2] = $917;
        } else {
         __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($58,$48);
        }
        $$0 = $900;
        while(1) {
         $918 = ($$0>>>0)<($906>>>0);
         if (!($918)) {
          $$5 = $902;
          break L1;
         }
         $919 = HEAP32[$908>>2]|0;
         $920 = HEAP32[$2>>2]|0;
         $921 = (($920) + (($$0*24)|0)|0);
         $922 = ((($919)) + -12|0);
         $923 = HEAP32[$922>>2]|0;
         $924 = ((($919)) + -8|0);
         $925 = HEAP32[$924>>2]|0;
         $926 = ($923|0)==($925|0);
         if ($926) {
          $929 = ((($919)) + -16|0);
          __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($929,$921);
         } else {
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($923,$921);
          $927 = HEAP32[$922>>2]|0;
          $928 = ((($927)) + 24|0);
          HEAP32[$922>>2] = $928;
         }
         $930 = (($$0) + 1)|0;
         $$0 = $930;
        }
        break;
       }
       case 84: case 116:  {
        $931 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_decltypeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
        $932 = ($931|0)==($0|0);
        if ($932) {
         label = 305;
         break L58;
        }
        $933 = HEAP32[$2>>2]|0;
        $934 = ((($2)) + 4|0);
        $935 = HEAP32[$934>>2]|0;
        $936 = ($933|0)==($935|0);
        if ($936) {
         $$5 = $0;
         break L1;
        }
        $937 = ((($935)) + -24|0);
        $$sroa$0$0$$sroa_idx$i486 = ((($2)) + 12|0);
        $$sroa$0$0$copyload$i487593 = HEAP32[$$sroa$0$0$$sroa_idx$i486>>2]|0;
        HEAP32[$50>>2] = $$sroa$0$0$copyload$i487593;
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($49,$937,$50);
        $938 = ((($2)) + 20|0);
        $939 = HEAP32[$938>>2]|0;
        $940 = ((($2)) + 24|0);
        $941 = HEAP32[$940>>2]|0;
        $942 = ($939>>>0)<($941>>>0);
        if ($942) {
         $943 = ((($49)) + 12|0);
         HEAP32[$939>>2] = 0;
         $944 = ((($939)) + 4|0);
         HEAP32[$944>>2] = 0;
         $945 = ((($939)) + 8|0);
         $946 = HEAP32[$943>>2]|0;
         HEAP32[$945>>2] = 0;
         $947 = ((($939)) + 12|0);
         HEAP32[$947>>2] = $946;
         $948 = HEAP32[$49>>2]|0;
         HEAP32[$939>>2] = $948;
         $949 = ((($49)) + 4|0);
         $950 = HEAP32[$949>>2]|0;
         HEAP32[$944>>2] = $950;
         $951 = ((($49)) + 8|0);
         $952 = HEAP32[$951>>2]|0;
         $953 = ((($939)) + 8|0);
         HEAP32[$953>>2] = $952;
         HEAP32[$951>>2] = 0;
         HEAP32[$949>>2] = 0;
         HEAP32[$49>>2] = 0;
         $954 = HEAP32[$938>>2]|0;
         $955 = ((($954)) + 16|0);
         HEAP32[$938>>2] = $955;
        } else {
         __THREW__ = 0;
         invoke_vii(1122,($58|0),($49|0));
         $956 = __THREW__; __THREW__ = 0;
         $957 = $956&1;
         if ($957) {
          $958 = ___cxa_find_matching_catch_2()|0;
          $959 = tempRet0;
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($49);
          $$16 = $958;$$16474 = $959;
          break L58;
         }
        }
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($49);
        $$5 = $931;
        break L1;
        break;
       }
       case 118:  {
        $960 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_vector_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
        $961 = ($960|0)==($0|0);
        if ($961) {
         label = 305;
         break L58;
        }
        $962 = HEAP32[$2>>2]|0;
        $963 = ((($2)) + 4|0);
        $964 = HEAP32[$963>>2]|0;
        $965 = ($962|0)==($964|0);
        if ($965) {
         $$5 = $0;
         break L1;
        }
        $966 = ((($964)) + -24|0);
        $$sroa$0$0$$sroa_idx$i482 = ((($2)) + 12|0);
        $$sroa$0$0$copyload$i483592 = HEAP32[$$sroa$0$0$$sroa_idx$i482>>2]|0;
        HEAP32[$52>>2] = $$sroa$0$0$copyload$i483592;
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($51,$966,$52);
        $967 = ((($2)) + 20|0);
        $968 = HEAP32[$967>>2]|0;
        $969 = ((($2)) + 24|0);
        $970 = HEAP32[$969>>2]|0;
        $971 = ($968>>>0)<($970>>>0);
        if ($971) {
         $972 = ((($51)) + 12|0);
         HEAP32[$968>>2] = 0;
         $973 = ((($968)) + 4|0);
         HEAP32[$973>>2] = 0;
         $974 = ((($968)) + 8|0);
         $975 = HEAP32[$972>>2]|0;
         HEAP32[$974>>2] = 0;
         $976 = ((($968)) + 12|0);
         HEAP32[$976>>2] = $975;
         $977 = HEAP32[$51>>2]|0;
         HEAP32[$968>>2] = $977;
         $978 = ((($51)) + 4|0);
         $979 = HEAP32[$978>>2]|0;
         HEAP32[$973>>2] = $979;
         $980 = ((($51)) + 8|0);
         $981 = HEAP32[$980>>2]|0;
         $982 = ((($968)) + 8|0);
         HEAP32[$982>>2] = $981;
         HEAP32[$980>>2] = 0;
         HEAP32[$978>>2] = 0;
         HEAP32[$51>>2] = 0;
         $983 = HEAP32[$967>>2]|0;
         $984 = ((($983)) + 16|0);
         HEAP32[$967>>2] = $984;
        } else {
         __THREW__ = 0;
         invoke_vii(1122,($58|0),($51|0));
         $985 = __THREW__; __THREW__ = 0;
         $986 = $985&1;
         if ($986) {
          $987 = ___cxa_find_matching_catch_2()|0;
          $988 = tempRet0;
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($51);
          $$16 = $987;$$16474 = $988;
          break L58;
         }
        }
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($51);
        $$5 = $960;
        break L1;
        break;
       }
       default: {
        label = 305;
        break L58;
       }
       }
      } else {
       label = 305;
      }
      break;
     }
     default: {
      label = 305;
     }
     }
    } while(0);
    do {
     if ((label|0) == 305) {
      $989 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_builtin_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $990 = ($989|0)==($0|0);
      if (!($990)) {
       $$5 = $989;
       break L1;
      }
      $991 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $992 = ($991|0)==($0|0);
      if ($992) {
       $$5 = $0;
       break L1;
      }
      $993 = HEAP32[$2>>2]|0;
      $994 = ((($2)) + 4|0);
      $995 = HEAP32[$994>>2]|0;
      $996 = ($993|0)==($995|0);
      if ($996) {
       $$5 = $0;
       break L1;
      }
      $997 = ((($995)) + -24|0);
      $$sroa$0$0$$sroa_idx$i = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i591 = HEAP32[$$sroa$0$0$$sroa_idx$i>>2]|0;
      HEAP32[$54>>2] = $$sroa$0$0$copyload$i591;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($53,$997,$54);
      $998 = ((($2)) + 20|0);
      $999 = HEAP32[$998>>2]|0;
      $1000 = ((($2)) + 24|0);
      $1001 = HEAP32[$1000>>2]|0;
      $1002 = ($999>>>0)<($1001>>>0);
      if ($1002) {
       $1003 = ((($53)) + 12|0);
       HEAP32[$999>>2] = 0;
       $1004 = ((($999)) + 4|0);
       HEAP32[$1004>>2] = 0;
       $1005 = ((($999)) + 8|0);
       $1006 = HEAP32[$1003>>2]|0;
       HEAP32[$1005>>2] = 0;
       $1007 = ((($999)) + 12|0);
       HEAP32[$1007>>2] = $1006;
       $1008 = HEAP32[$53>>2]|0;
       HEAP32[$999>>2] = $1008;
       $1009 = ((($53)) + 4|0);
       $1010 = HEAP32[$1009>>2]|0;
       HEAP32[$1004>>2] = $1010;
       $1011 = ((($53)) + 8|0);
       $1012 = HEAP32[$1011>>2]|0;
       $1013 = ((($999)) + 8|0);
       HEAP32[$1013>>2] = $1012;
       HEAP32[$1011>>2] = 0;
       HEAP32[$1009>>2] = 0;
       HEAP32[$53>>2] = 0;
       $1014 = HEAP32[$998>>2]|0;
       $1015 = ((($1014)) + 16|0);
       HEAP32[$998>>2] = $1015;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($58|0),($53|0));
       $1016 = __THREW__; __THREW__ = 0;
       $1017 = $1016&1;
       if ($1017) {
        $1018 = ___cxa_find_matching_catch_2()|0;
        $1019 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($53);
        $$16 = $1018;$$16474 = $1019;
        break;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($53);
      $$5 = $991;
      break L1;
     }
    } while(0);
    ___resumeException($$16|0);
    // unreachable;
   } else {
    $$5 = $147;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$5|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_cv_qualifiersEPKcS2_Rj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$1 = 0, $$2 = 0, $$pre = 0, $$pre13 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 HEAP32[$2>>2] = 0;
 $3 = ($0|0)==($1|0);
 if ($3) {
  $$2 = $0;
 } else {
  $4 = HEAP8[$0>>0]|0;
  $5 = ($4<<24>>24)==(114);
  if ($5) {
   HEAP32[$2>>2] = 4;
   $6 = ((($0)) + 1|0);
   $$pre = HEAP8[$6>>0]|0;
   $$0 = $6;$10 = 4;$8 = $$pre;
  } else {
   $$0 = $0;$10 = 0;$8 = $4;
  }
  $7 = ($8<<24>>24)==(86);
  if ($7) {
   $9 = $10 | 2;
   HEAP32[$2>>2] = $9;
   $11 = ((($$0)) + 1|0);
   $$pre13 = HEAP8[$11>>0]|0;
   $$1 = $11;$13 = $$pre13;$15 = $9;
  } else {
   $$1 = $$0;$13 = $8;$15 = $10;
  }
  $12 = ($13<<24>>24)==(75);
  if ($12) {
   $14 = $15 | 1;
   HEAP32[$2>>2] = $14;
   $16 = ((($$1)) + 1|0);
   $$2 = $16;
  } else {
   $$2 = $$1;
  }
 }
 return ($$2|0);
}
function __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$0>>2]|0;
 $6 = (($4) - ($5))|0;
 $7 = $6 >> 4;
 $8 = (($7) + 1)|0;
 $9 = (__ZNKSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE8max_sizeEv($0)|0);
 $10 = ($9>>>0)<($8>>>0);
 if ($10) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $11 = ((($0)) + 12|0);
  $12 = ((($0)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$0>>2]|0;
  $15 = (($13) - ($14))|0;
  $16 = $15 >> 4;
  $17 = $9 >>> 1;
  $18 = ($16>>>0)<($17>>>0);
  $19 = $15 >> 3;
  $20 = ($19>>>0)<($8>>>0);
  $$sroa$speculated$i = $20 ? $8 : $19;
  $$sroa$speculated$$i = $18 ? $$sroa$speculated$i : $9;
  $21 = HEAP32[$3>>2]|0;
  $22 = (($21) - ($14))|0;
  $23 = $22 >> 4;
  __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEEC2EjjS9_($2,$$sroa$speculated$$i,$23,$11);
  $24 = ((($2)) + 8|0);
  $25 = HEAP32[$24>>2]|0;
  HEAP32[$25>>2] = 0;
  $26 = ((($25)) + 4|0);
  HEAP32[$26>>2] = 0;
  $27 = ((($25)) + 8|0);
  $28 = HEAP32[$1>>2]|0;
  HEAP32[$27>>2] = 0;
  $29 = ((($25)) + 12|0);
  HEAP32[$29>>2] = $28;
  $30 = ((($25)) + 16|0);
  HEAP32[$24>>2] = $30;
  __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS6_RS7_EE($0,$2);
  __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEED2Ev($2);
  STACKTOP = sp;return;
 }
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (__ZNSt3__211char_traitsIcE6lengthEPKc($2)|0);
 $4 = (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKcj($0,$1,$2,$3)|0);
 return ($4|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE6lengthEPKc($1)|0);
 $3 = (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj($0,$1,$2)|0);
 return ($3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_($0,$1);
 $2 = ((($0)) + 12|0);
 $3 = ((($1)) + 12|0);
 __THREW__ = 0;
 invoke_vii(1129,($2|0),($3|0));
 $4 = __THREW__; __THREW__ = 0;
 $5 = $4&1;
 if ($5) {
  $6 = ___cxa_find_matching_catch_2()|0;
  $7 = tempRet0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($0);
  ___resumeException($6|0);
  // unreachable;
 } else {
  return;
 }
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$0>>2]|0;
 $6 = (($4) - ($5))|0;
 $7 = (($6|0) / 24)&-1;
 $8 = (($7) + 1)|0;
 $9 = (__ZNKSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8max_sizeEv($0)|0);
 $10 = ($9>>>0)<($8>>>0);
 if ($10) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $11 = ((($0)) + 12|0);
 $12 = ((($0)) + 8|0);
 $13 = HEAP32[$12>>2]|0;
 $14 = HEAP32[$0>>2]|0;
 $15 = (($13) - ($14))|0;
 $16 = (($15|0) / 24)&-1;
 $17 = $9 >>> 1;
 $18 = ($16>>>0)<($17>>>0);
 $19 = $16 << 1;
 $20 = ($19>>>0)<($8>>>0);
 $$sroa$speculated$i = $20 ? $8 : $19;
 $$sroa$speculated$$i = $18 ? $$sroa$speculated$i : $9;
 $21 = HEAP32[$3>>2]|0;
 $22 = (($21) - ($14))|0;
 $23 = (($22|0) / 24)&-1;
 __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEEC2EjjS6_($2,$$sroa$speculated$$i,$23,$11);
 $24 = ((($2)) + 8|0);
 $25 = HEAP32[$24>>2]|0;
 __THREW__ = 0;
 invoke_vii(1130,($25|0),($1|0));
 $26 = __THREW__; __THREW__ = 0;
 $27 = $26&1;
 if ($27) {
  $29 = ___cxa_find_matching_catch_2()|0;
  $30 = tempRet0;
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEED2Ev($2);
  ___resumeException($29|0);
  // unreachable;
 } else {
  $28 = ((($25)) + 24|0);
  HEAP32[$24>>2] = $28;
  __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($0,$2);
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEED2Ev($2);
  STACKTOP = sp;return;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_builtin_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$073 = 0, $$074 = 0, $$1 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0;
 var $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0;
 var $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0;
 var $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0;
 var $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0;
 var $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0;
 var $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0;
 var $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0;
 var $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 704|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(704|0);
 $3 = sp + 672|0;
 $4 = sp + 648|0;
 $5 = sp + 624|0;
 $6 = sp + 600|0;
 $7 = sp + 576|0;
 $8 = sp + 552|0;
 $9 = sp + 528|0;
 $10 = sp + 504|0;
 $11 = sp + 480|0;
 $12 = sp + 456|0;
 $13 = sp + 432|0;
 $14 = sp + 408|0;
 $15 = sp + 384|0;
 $16 = sp + 360|0;
 $17 = sp + 336|0;
 $18 = sp + 312|0;
 $19 = sp + 288|0;
 $20 = sp + 264|0;
 $21 = sp + 240|0;
 $22 = sp + 216|0;
 $23 = sp + 192|0;
 $24 = sp + 168|0;
 $25 = sp + 144|0;
 $26 = sp + 120|0;
 $27 = sp + 96|0;
 $28 = sp + 72|0;
 $29 = sp + 48|0;
 $30 = sp + 24|0;
 $31 = sp;
 $32 = ($0|0)==($1|0);
 L1: do {
  if ($32) {
   $$1 = $0;
  } else {
   $33 = HEAP8[$0>>0]|0;
   $34 = $33 << 24 >> 24;
   $35 = ((($0)) + 1|0);
   L3: do {
    switch ($34|0) {
    case 118:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($3,160520);
     $36 = ((($2)) + 4|0);
     $37 = HEAP32[$36>>2]|0;
     $38 = ((($2)) + 8|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ($37>>>0)<($39>>>0);
     if ($40) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($37,$3);
      $41 = HEAP32[$36>>2]|0;
      $42 = ((($41)) + 24|0);
      HEAP32[$36>>2] = $42;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($3|0));
      $43 = __THREW__; __THREW__ = 0;
      $44 = $43&1;
      if ($44) {
       $45 = ___cxa_find_matching_catch_2()|0;
       $46 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
       $$073 = $46;$$074 = $45;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
     $$1 = $35;
     break L1;
     break;
    }
    case 119:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj8EEERAT__Kc($4);
     $47 = ((($2)) + 4|0);
     $48 = HEAP32[$47>>2]|0;
     $49 = ((($2)) + 8|0);
     $50 = HEAP32[$49>>2]|0;
     $51 = ($48>>>0)<($50>>>0);
     if ($51) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($48,$4);
      $52 = HEAP32[$47>>2]|0;
      $53 = ((($52)) + 24|0);
      HEAP32[$47>>2] = $53;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($4|0));
      $54 = __THREW__; __THREW__ = 0;
      $55 = $54&1;
      if ($55) {
       $56 = ___cxa_find_matching_catch_2()|0;
       $57 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
       $$073 = $57;$$074 = $56;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
     $$1 = $35;
     break L1;
     break;
    }
    case 98:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($5,160525);
     $58 = ((($2)) + 4|0);
     $59 = HEAP32[$58>>2]|0;
     $60 = ((($2)) + 8|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = ($59>>>0)<($61>>>0);
     if ($62) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($59,$5);
      $63 = HEAP32[$58>>2]|0;
      $64 = ((($63)) + 24|0);
      HEAP32[$58>>2] = $64;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($5|0));
      $65 = __THREW__; __THREW__ = 0;
      $66 = $65&1;
      if ($66) {
       $67 = ___cxa_find_matching_catch_2()|0;
       $68 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
       $$073 = $68;$$074 = $67;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
     $$1 = $35;
     break L1;
     break;
    }
    case 99:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($6,160294);
     $69 = ((($2)) + 4|0);
     $70 = HEAP32[$69>>2]|0;
     $71 = ((($2)) + 8|0);
     $72 = HEAP32[$71>>2]|0;
     $73 = ($70>>>0)<($72>>>0);
     if ($73) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($70,$6);
      $74 = HEAP32[$69>>2]|0;
      $75 = ((($74)) + 24|0);
      HEAP32[$69>>2] = $75;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($6|0));
      $76 = __THREW__; __THREW__ = 0;
      $77 = $76&1;
      if ($77) {
       $78 = ___cxa_find_matching_catch_2()|0;
       $79 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
       $$073 = $79;$$074 = $78;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
     $$1 = $35;
     break L1;
     break;
    }
    case 97:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($7,160299);
     $80 = ((($2)) + 4|0);
     $81 = HEAP32[$80>>2]|0;
     $82 = ((($2)) + 8|0);
     $83 = HEAP32[$82>>2]|0;
     $84 = ($81>>>0)<($83>>>0);
     if ($84) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($81,$7);
      $85 = HEAP32[$80>>2]|0;
      $86 = ((($85)) + 24|0);
      HEAP32[$80>>2] = $86;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($7|0));
      $87 = __THREW__; __THREW__ = 0;
      $88 = $87&1;
      if ($88) {
       $89 = ___cxa_find_matching_catch_2()|0;
       $90 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
       $$073 = $90;$$074 = $89;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
     $$1 = $35;
     break L1;
     break;
    }
    case 104:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj14EEERAT__Kc($8,160311);
     $91 = ((($2)) + 4|0);
     $92 = HEAP32[$91>>2]|0;
     $93 = ((($2)) + 8|0);
     $94 = HEAP32[$93>>2]|0;
     $95 = ($92>>>0)<($94>>>0);
     if ($95) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($92,$8);
      $96 = HEAP32[$91>>2]|0;
      $97 = ((($96)) + 24|0);
      HEAP32[$91>>2] = $97;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($8|0));
      $98 = __THREW__; __THREW__ = 0;
      $99 = $98&1;
      if ($99) {
       $100 = ___cxa_find_matching_catch_2()|0;
       $101 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
       $$073 = $101;$$074 = $100;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
     $$1 = $35;
     break L1;
     break;
    }
    case 115:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj6EEERAT__Kc($9,160325);
     $102 = ((($2)) + 4|0);
     $103 = HEAP32[$102>>2]|0;
     $104 = ((($2)) + 8|0);
     $105 = HEAP32[$104>>2]|0;
     $106 = ($103>>>0)<($105>>>0);
     if ($106) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($103,$9);
      $107 = HEAP32[$102>>2]|0;
      $108 = ((($107)) + 24|0);
      HEAP32[$102>>2] = $108;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($9|0));
      $109 = __THREW__; __THREW__ = 0;
      $110 = $109&1;
      if ($110) {
       $111 = ___cxa_find_matching_catch_2()|0;
       $112 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
       $$073 = $112;$$074 = $111;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
     $$1 = $35;
     break L1;
     break;
    }
    case 116:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj15EEERAT__Kc($10,160331);
     $113 = ((($2)) + 4|0);
     $114 = HEAP32[$113>>2]|0;
     $115 = ((($2)) + 8|0);
     $116 = HEAP32[$115>>2]|0;
     $117 = ($114>>>0)<($116>>>0);
     if ($117) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($114,$10);
      $118 = HEAP32[$113>>2]|0;
      $119 = ((($118)) + 24|0);
      HEAP32[$113>>2] = $119;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($10|0));
      $120 = __THREW__; __THREW__ = 0;
      $121 = $120&1;
      if ($121) {
       $122 = ___cxa_find_matching_catch_2()|0;
       $123 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
       $$073 = $123;$$074 = $122;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
     $$1 = $35;
     break L1;
     break;
    }
    case 105:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj4EEERAT__Kc($11,160530);
     $124 = ((($2)) + 4|0);
     $125 = HEAP32[$124>>2]|0;
     $126 = ((($2)) + 8|0);
     $127 = HEAP32[$126>>2]|0;
     $128 = ($125>>>0)<($127>>>0);
     if ($128) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($125,$11);
      $129 = HEAP32[$124>>2]|0;
      $130 = ((($129)) + 24|0);
      HEAP32[$124>>2] = $130;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($11|0));
      $131 = __THREW__; __THREW__ = 0;
      $132 = $131&1;
      if ($132) {
       $133 = ___cxa_find_matching_catch_2()|0;
       $134 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
       $$073 = $134;$$074 = $133;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
     $$1 = $35;
     break L1;
     break;
    }
    case 106:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj13EEERAT__Kc($12,160534);
     $135 = ((($2)) + 4|0);
     $136 = HEAP32[$135>>2]|0;
     $137 = ((($2)) + 8|0);
     $138 = HEAP32[$137>>2]|0;
     $139 = ($136>>>0)<($138>>>0);
     if ($139) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($136,$12);
      $140 = HEAP32[$135>>2]|0;
      $141 = ((($140)) + 24|0);
      HEAP32[$135>>2] = $141;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($12|0));
      $142 = __THREW__; __THREW__ = 0;
      $143 = $142&1;
      if ($143) {
       $144 = ___cxa_find_matching_catch_2()|0;
       $145 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
       $$073 = $145;$$074 = $144;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
     $$1 = $35;
     break L1;
     break;
    }
    case 108:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($13,160547);
     $146 = ((($2)) + 4|0);
     $147 = HEAP32[$146>>2]|0;
     $148 = ((($2)) + 8|0);
     $149 = HEAP32[$148>>2]|0;
     $150 = ($147>>>0)<($149>>>0);
     if ($150) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($147,$13);
      $151 = HEAP32[$146>>2]|0;
      $152 = ((($151)) + 24|0);
      HEAP32[$146>>2] = $152;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($13|0));
      $153 = __THREW__; __THREW__ = 0;
      $154 = $153&1;
      if ($154) {
       $155 = ___cxa_find_matching_catch_2()|0;
       $156 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
       $$073 = $156;$$074 = $155;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
     $$1 = $35;
     break L1;
     break;
    }
    case 109:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj14EEERAT__Kc($14,160552);
     $157 = ((($2)) + 4|0);
     $158 = HEAP32[$157>>2]|0;
     $159 = ((($2)) + 8|0);
     $160 = HEAP32[$159>>2]|0;
     $161 = ($158>>>0)<($160>>>0);
     if ($161) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($158,$14);
      $162 = HEAP32[$157>>2]|0;
      $163 = ((($162)) + 24|0);
      HEAP32[$157>>2] = $163;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($14|0));
      $164 = __THREW__; __THREW__ = 0;
      $165 = $164&1;
      if ($165) {
       $166 = ___cxa_find_matching_catch_2()|0;
       $167 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
       $$073 = $167;$$074 = $166;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
     $$1 = $35;
     break L1;
     break;
    }
    case 120:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($15,160566);
     $168 = ((($2)) + 4|0);
     $169 = HEAP32[$168>>2]|0;
     $170 = ((($2)) + 8|0);
     $171 = HEAP32[$170>>2]|0;
     $172 = ($169>>>0)<($171>>>0);
     if ($172) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($169,$15);
      $173 = HEAP32[$168>>2]|0;
      $174 = ((($173)) + 24|0);
      HEAP32[$168>>2] = $174;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($15|0));
      $175 = __THREW__; __THREW__ = 0;
      $176 = $175&1;
      if ($176) {
       $177 = ___cxa_find_matching_catch_2()|0;
       $178 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($15);
       $$073 = $178;$$074 = $177;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($15);
     $$1 = $35;
     break L1;
     break;
    }
    case 121:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj19EEERAT__Kc($16);
     $179 = ((($2)) + 4|0);
     $180 = HEAP32[$179>>2]|0;
     $181 = ((($2)) + 8|0);
     $182 = HEAP32[$181>>2]|0;
     $183 = ($180>>>0)<($182>>>0);
     if ($183) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($180,$16);
      $184 = HEAP32[$179>>2]|0;
      $185 = ((($184)) + 24|0);
      HEAP32[$179>>2] = $185;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($16|0));
      $186 = __THREW__; __THREW__ = 0;
      $187 = $186&1;
      if ($187) {
       $188 = ___cxa_find_matching_catch_2()|0;
       $189 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($16);
       $$073 = $189;$$074 = $188;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($16);
     $$1 = $35;
     break L1;
     break;
    }
    case 110:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj9EEERAT__Kc($17,160360);
     $190 = ((($2)) + 4|0);
     $191 = HEAP32[$190>>2]|0;
     $192 = ((($2)) + 8|0);
     $193 = HEAP32[$192>>2]|0;
     $194 = ($191>>>0)<($193>>>0);
     if ($194) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($191,$17);
      $195 = HEAP32[$190>>2]|0;
      $196 = ((($195)) + 24|0);
      HEAP32[$190>>2] = $196;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($17|0));
      $197 = __THREW__; __THREW__ = 0;
      $198 = $197&1;
      if ($198) {
       $199 = ___cxa_find_matching_catch_2()|0;
       $200 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
       $$073 = $200;$$074 = $199;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
     $$1 = $35;
     break L1;
     break;
    }
    case 111:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj18EEERAT__Kc($18,160369);
     $201 = ((($2)) + 4|0);
     $202 = HEAP32[$201>>2]|0;
     $203 = ((($2)) + 8|0);
     $204 = HEAP32[$203>>2]|0;
     $205 = ($202>>>0)<($204>>>0);
     if ($205) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($202,$18);
      $206 = HEAP32[$201>>2]|0;
      $207 = ((($206)) + 24|0);
      HEAP32[$201>>2] = $207;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($18|0));
      $208 = __THREW__; __THREW__ = 0;
      $209 = $208&1;
      if ($209) {
       $210 = ___cxa_find_matching_catch_2()|0;
       $211 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($18);
       $$073 = $211;$$074 = $210;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($18);
     $$1 = $35;
     break L1;
     break;
    }
    case 102:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj6EEERAT__Kc($19,160576);
     $212 = ((($2)) + 4|0);
     $213 = HEAP32[$212>>2]|0;
     $214 = ((($2)) + 8|0);
     $215 = HEAP32[$214>>2]|0;
     $216 = ($213>>>0)<($215>>>0);
     if ($216) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($213,$19);
      $217 = HEAP32[$212>>2]|0;
      $218 = ((($217)) + 24|0);
      HEAP32[$212>>2] = $218;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($19|0));
      $219 = __THREW__; __THREW__ = 0;
      $220 = $219&1;
      if ($220) {
       $221 = ___cxa_find_matching_catch_2()|0;
       $222 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
       $$073 = $222;$$074 = $221;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
     $$1 = $35;
     break L1;
     break;
    }
    case 100:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj7EEERAT__Kc($20);
     $223 = ((($2)) + 4|0);
     $224 = HEAP32[$223>>2]|0;
     $225 = ((($2)) + 8|0);
     $226 = HEAP32[$225>>2]|0;
     $227 = ($224>>>0)<($226>>>0);
     if ($227) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($224,$20);
      $228 = HEAP32[$223>>2]|0;
      $229 = ((($228)) + 24|0);
      HEAP32[$223>>2] = $229;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($20|0));
      $230 = __THREW__; __THREW__ = 0;
      $231 = $230&1;
      if ($231) {
       $232 = ___cxa_find_matching_catch_2()|0;
       $233 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
       $$073 = $233;$$074 = $232;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
     $$1 = $35;
     break L1;
     break;
    }
    case 101:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($21,160582);
     $234 = ((($2)) + 4|0);
     $235 = HEAP32[$234>>2]|0;
     $236 = ((($2)) + 8|0);
     $237 = HEAP32[$236>>2]|0;
     $238 = ($235>>>0)<($237>>>0);
     if ($238) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($235,$21);
      $239 = HEAP32[$234>>2]|0;
      $240 = ((($239)) + 24|0);
      HEAP32[$234>>2] = $240;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($21|0));
      $241 = __THREW__; __THREW__ = 0;
      $242 = $241&1;
      if ($242) {
       $243 = ___cxa_find_matching_catch_2()|0;
       $244 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($21);
       $$073 = $244;$$074 = $243;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($21);
     $$1 = $35;
     break L1;
     break;
    }
    case 103:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($22,160594);
     $245 = ((($2)) + 4|0);
     $246 = HEAP32[$245>>2]|0;
     $247 = ((($2)) + 8|0);
     $248 = HEAP32[$247>>2]|0;
     $249 = ($246>>>0)<($248>>>0);
     if ($249) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($246,$22);
      $250 = HEAP32[$245>>2]|0;
      $251 = ((($250)) + 24|0);
      HEAP32[$245>>2] = $251;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($22|0));
      $252 = __THREW__; __THREW__ = 0;
      $253 = $252&1;
      if ($253) {
       $254 = ___cxa_find_matching_catch_2()|0;
       $255 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($22);
       $$073 = $255;$$074 = $254;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($22);
     $$1 = $35;
     break L1;
     break;
    }
    case 122:  {
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj4EEERAT__Kc($23,160605);
     $256 = ((($2)) + 4|0);
     $257 = HEAP32[$256>>2]|0;
     $258 = ((($2)) + 8|0);
     $259 = HEAP32[$258>>2]|0;
     $260 = ($257>>>0)<($259>>>0);
     if ($260) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($257,$23);
      $261 = HEAP32[$256>>2]|0;
      $262 = ((($261)) + 24|0);
      HEAP32[$256>>2] = $262;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($23|0));
      $263 = __THREW__; __THREW__ = 0;
      $264 = $263&1;
      if ($264) {
       $265 = ___cxa_find_matching_catch_2()|0;
       $266 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($23);
       $$073 = $266;$$074 = $265;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($23);
     $$1 = $35;
     break L1;
     break;
    }
    case 117:  {
     $267 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($35,$1,$2)|0);
     $268 = ($267|0)==($35|0);
     $$ = $268 ? $0 : $267;
     $$1 = $$;
     break L1;
     break;
    }
    case 68:  {
     $269 = ($35|0)==($1|0);
     if ($269) {
      $$1 = $0;
      break L1;
     }
     $270 = HEAP8[$35>>0]|0;
     $271 = $270 << 24 >> 24;
     $272 = ((($0)) + 2|0);
     switch ($271|0) {
     case 100:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($24,160609);
      $273 = ((($2)) + 4|0);
      $274 = HEAP32[$273>>2]|0;
      $275 = ((($2)) + 8|0);
      $276 = HEAP32[$275>>2]|0;
      $277 = ($274>>>0)<($276>>>0);
      if ($277) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($274,$24);
       $278 = HEAP32[$273>>2]|0;
       $279 = ((($278)) + 24|0);
       HEAP32[$273>>2] = $279;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($24|0));
       $280 = __THREW__; __THREW__ = 0;
       $281 = $280&1;
       if ($281) {
        $282 = ___cxa_find_matching_catch_2()|0;
        $283 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($24);
        $$073 = $283;$$074 = $282;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($24);
      $$1 = $272;
      break L1;
      break;
     }
     case 101:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($25,160619);
      $284 = ((($2)) + 4|0);
      $285 = HEAP32[$284>>2]|0;
      $286 = ((($2)) + 8|0);
      $287 = HEAP32[$286>>2]|0;
      $288 = ($285>>>0)<($287>>>0);
      if ($288) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($285,$25);
       $289 = HEAP32[$284>>2]|0;
       $290 = ((($289)) + 24|0);
       HEAP32[$284>>2] = $290;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($25|0));
       $291 = __THREW__; __THREW__ = 0;
       $292 = $291&1;
       if ($292) {
        $293 = ___cxa_find_matching_catch_2()|0;
        $294 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($25);
        $$073 = $294;$$074 = $293;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($25);
      $$1 = $272;
      break L1;
      break;
     }
     case 102:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($26,160630);
      $295 = ((($2)) + 4|0);
      $296 = HEAP32[$295>>2]|0;
      $297 = ((($2)) + 8|0);
      $298 = HEAP32[$297>>2]|0;
      $299 = ($296>>>0)<($298>>>0);
      if ($299) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($296,$26);
       $300 = HEAP32[$295>>2]|0;
       $301 = ((($300)) + 24|0);
       HEAP32[$295>>2] = $301;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($26|0));
       $302 = __THREW__; __THREW__ = 0;
       $303 = $302&1;
       if ($303) {
        $304 = ___cxa_find_matching_catch_2()|0;
        $305 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($26);
        $$073 = $305;$$074 = $304;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($26);
      $$1 = $272;
      break L1;
      break;
     }
     case 104:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($27,160640);
      $306 = ((($2)) + 4|0);
      $307 = HEAP32[$306>>2]|0;
      $308 = ((($2)) + 8|0);
      $309 = HEAP32[$308>>2]|0;
      $310 = ($307>>>0)<($309>>>0);
      if ($310) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($307,$27);
       $311 = HEAP32[$306>>2]|0;
       $312 = ((($311)) + 24|0);
       HEAP32[$306>>2] = $312;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($27|0));
       $313 = __THREW__; __THREW__ = 0;
       $314 = $313&1;
       if ($314) {
        $315 = ___cxa_find_matching_catch_2()|0;
        $316 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($27);
        $$073 = $316;$$074 = $315;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($27);
      $$1 = $272;
      break L1;
      break;
     }
     case 105:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj9EEERAT__Kc($28,160650);
      $317 = ((($2)) + 4|0);
      $318 = HEAP32[$317>>2]|0;
      $319 = ((($2)) + 8|0);
      $320 = HEAP32[$319>>2]|0;
      $321 = ($318>>>0)<($320>>>0);
      if ($321) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($318,$28);
       $322 = HEAP32[$317>>2]|0;
       $323 = ((($322)) + 24|0);
       HEAP32[$317>>2] = $323;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($28|0));
       $324 = __THREW__; __THREW__ = 0;
       $325 = $324&1;
       if ($325) {
        $326 = ___cxa_find_matching_catch_2()|0;
        $327 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($28);
        $$073 = $327;$$074 = $326;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($28);
      $$1 = $272;
      break L1;
      break;
     }
     case 115:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj9EEERAT__Kc($29,160659);
      $328 = ((($2)) + 4|0);
      $329 = HEAP32[$328>>2]|0;
      $330 = ((($2)) + 8|0);
      $331 = HEAP32[$330>>2]|0;
      $332 = ($329>>>0)<($331>>>0);
      if ($332) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($329,$29);
       $333 = HEAP32[$328>>2]|0;
       $334 = ((($333)) + 24|0);
       HEAP32[$328>>2] = $334;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($29|0));
       $335 = __THREW__; __THREW__ = 0;
       $336 = $335&1;
       if ($336) {
        $337 = ___cxa_find_matching_catch_2()|0;
        $338 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($29);
        $$073 = $338;$$074 = $337;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($29);
      $$1 = $272;
      break L1;
      break;
     }
     case 97:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($30,160668);
      $339 = ((($2)) + 4|0);
      $340 = HEAP32[$339>>2]|0;
      $341 = ((($2)) + 8|0);
      $342 = HEAP32[$341>>2]|0;
      $343 = ($340>>>0)<($342>>>0);
      if ($343) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($340,$30);
       $344 = HEAP32[$339>>2]|0;
       $345 = ((($344)) + 24|0);
       HEAP32[$339>>2] = $345;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($30|0));
       $346 = __THREW__; __THREW__ = 0;
       $347 = $346&1;
       if ($347) {
        $348 = ___cxa_find_matching_catch_2()|0;
        $349 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
        $$073 = $349;$$074 = $348;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
      $$1 = $272;
      break L1;
      break;
     }
     case 110:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj15EEERAT__Kc($31,160673);
      $350 = ((($2)) + 4|0);
      $351 = HEAP32[$350>>2]|0;
      $352 = ((($2)) + 8|0);
      $353 = HEAP32[$352>>2]|0;
      $354 = ($351>>>0)<($353>>>0);
      if ($354) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($351,$31);
       $355 = HEAP32[$350>>2]|0;
       $356 = ((($355)) + 24|0);
       HEAP32[$350>>2] = $356;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($31|0));
       $357 = __THREW__; __THREW__ = 0;
       $358 = $357&1;
       if ($358) {
        $359 = ___cxa_find_matching_catch_2()|0;
        $360 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
        $$073 = $360;$$074 = $359;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
      $$1 = $272;
      break L1;
      break;
     }
     default: {
      $$1 = $0;
      break L1;
     }
     }
     break;
    }
    default: {
     $$1 = $0;
     break L1;
    }
    }
   } while(0);
   ___resumeException($$074|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_array_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i132 = 0, $$0$i$i$i112 = 0, $$0$i$i$i119 = 0, $$0$i$i$i123 = 0, $$0$i$i$i127 = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$082 = 0, $$089 = 0, $$183 = 0, $$190 = 0, $$284 = 0, $$291 = 0, $$385 = 0, $$392 = 0, $$486 = 0, $$493 = 0;
 var $$587 = 0, $$594 = 0, $$688 = 0, $$695 = 0, $$7 = 0, $$off = 0, $$pre$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0;
 var $exitcond$i$i$i$i = 0, $exitcond$i$i$i$i133 = 0, $exitcond$i$i$i113 = 0, $exitcond$i$i$i120 = 0, $exitcond$i$i$i124 = 0, $exitcond$i$i$i128 = 0, $or$cond109 = 0, $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 192|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(192|0);
 $3 = sp + 176|0;
 $4 = sp + 175|0;
 $5 = sp + 174|0;
 $6 = sp + 173|0;
 $7 = sp + 172|0;
 $8 = sp + 160|0;
 $9 = sp + 148|0;
 $10 = sp + 136|0;
 $11 = sp + 124|0;
 $12 = sp + 112|0;
 $13 = sp + 88|0;
 $14 = sp + 64|0;
 $15 = sp + 48|0;
 $16 = sp + 36|0;
 $17 = sp + 24|0;
 $18 = sp + 12|0;
 $19 = sp;
 $20 = ($0|0)==($1|0);
 L1: do {
  if ($20) {
   $$7 = $0;
  } else {
   $21 = HEAP8[$0>>0]|0;
   $22 = ($21<<24>>24)==(65);
   if ($22) {
    $23 = ((($0)) + 1|0);
    $24 = ($23|0)==($1|0);
    if ($24) {
     $$7 = $0;
    } else {
     $25 = HEAP8[$23>>0]|0;
     $26 = ($25<<24>>24)==(95);
     if ($26) {
      $27 = ((($0)) + 2|0);
      $28 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($27,$1,$2)|0);
      $29 = ($28|0)==($27|0);
      if ($29) {
       $$7 = $0;
       break;
      }
      $30 = HEAP32[$2>>2]|0;
      $31 = ((($2)) + 4|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ($30|0)==($32|0);
      if ($33) {
       $$7 = $0;
       break;
      }
      $34 = ((($32)) + -12|0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($8,$34,2);
      $35 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
      $36 = ((($8)) + 11|0);
      $37 = HEAP8[$36>>0]|0;
      $38 = ($37<<24>>24)<(0);
      $39 = ((($8)) + 4|0);
      $40 = HEAP32[$39>>2]|0;
      $41 = $37&255;
      $42 = $38 ? $40 : $41;
      $43 = ($35|0)==($42|0);
      if ($43) {
       __THREW__ = 0;
       $44 = (invoke_iiii(1123,($8|0),(159075|0),($35|0))|0);
       $45 = __THREW__; __THREW__ = 0;
       $46 = $45&1;
       if ($46) {
        $47 = ___cxa_find_matching_catch_3(0|0)|0;
        $48 = tempRet0;
        ___clang_call_terminate($47);
        // unreachable;
       }
       $49 = ($44|0)==(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
       if ($49) {
        $50 = HEAP32[$31>>2]|0;
        $51 = ((($50)) + -12|0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE5eraseEjj($51);
       }
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
      }
      $52 = HEAP32[$31>>2]|0;
      $53 = ((($52)) + -12|0);
      (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($53,0,160516)|0);
      $$7 = $28;
      break;
     }
     $$off = (($25) + -49)<<24>>24;
     $54 = ($$off&255)<(9);
     if ($54) {
      $55 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($23,$1)|0);
      $56 = ($55|0)==($1|0);
      if ($56) {
       $$7 = $0;
       break;
      }
      $57 = HEAP8[$55>>0]|0;
      $58 = ($57<<24>>24)==(95);
      if (!($58)) {
       $$7 = $0;
       break;
      }
      $59 = ((($55)) + 1|0);
      $60 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($59,$1,$2)|0);
      $61 = ($60|0)==($59|0);
      if ($61) {
       $$7 = $0;
       break;
      }
      $62 = HEAP32[$2>>2]|0;
      $63 = ((($2)) + 4|0);
      $64 = HEAP32[$63>>2]|0;
      $65 = ($62|0)==($64|0);
      if ($65) {
       $$7 = $0;
       break;
      }
      $66 = ((($64)) + -12|0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($9,$66,2);
      $67 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
      $68 = ((($9)) + 11|0);
      $69 = HEAP8[$68>>0]|0;
      $70 = ($69<<24>>24)<(0);
      $71 = ((($9)) + 4|0);
      $72 = HEAP32[$71>>2]|0;
      $73 = $69&255;
      $74 = $70 ? $72 : $73;
      $75 = ($67|0)==($74|0);
      if ($75) {
       __THREW__ = 0;
       $76 = (invoke_iiii(1123,($9|0),(159075|0),($67|0))|0);
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        $79 = ___cxa_find_matching_catch_3(0|0)|0;
        $80 = tempRet0;
        ___clang_call_terminate($79);
        // unreachable;
       }
       $81 = ($76|0)==(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
       if ($81) {
        $82 = HEAP32[$63>>2]|0;
        $83 = ((($82)) + -12|0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE5eraseEjj($83);
       }
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
      }
      $84 = HEAP32[$63>>2]|0;
      $85 = ((($84)) + -12|0);
      ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;
      $86 = $23;
      $87 = $55;
      $88 = (($87) - ($86))|0;
      $89 = ($88>>>0)>(4294967279);
      if ($89) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($12);
       // unreachable;
      }
      $90 = ($88>>>0)<(11);
      if ($90) {
       $91 = $88&255;
       $92 = ((($12)) + 11|0);
       HEAP8[$92>>0] = $91;
       $$021$ph$i$i = $12;
      } else {
       $93 = (($88) + 16)|0;
       $94 = $93 & -16;
       $95 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($94)|0);
       HEAP32[$12>>2] = $95;
       $96 = $94 | -2147483648;
       $97 = ((($12)) + 8|0);
       HEAP32[$97>>2] = $96;
       $98 = ((($12)) + 4|0);
       HEAP32[$98>>2] = $88;
       $$021$ph$i$i = $95;
      }
      $$0$i$i = $23;$$021$i$i = $$021$ph$i$i;
      while(1) {
       $99 = ($$0$i$i|0)==($55|0);
       if ($99) {
        break;
       }
       __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
       $100 = ((($$0$i$i)) + 1|0);
       $101 = ((($$021$i$i)) + 1|0);
       $$0$i$i = $100;$$021$i$i = $101;
      }
      $scevgep23$i$i = (($$021$ph$i$i) + ($88)|0);
      HEAP8[$7>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$7);
      __THREW__ = 0;
      $102 = (invoke_iiii(1121,($12|0),0,(159075|0))|0);
      $103 = __THREW__; __THREW__ = 0;
      $104 = $103&1;
      if ($104) {
       $123 = ___cxa_find_matching_catch_2()|0;
       $124 = tempRet0;
       $$183 = $124;$$190 = $123;
      } else {
       ;HEAP32[$11>>2]=HEAP32[$12>>2]|0;HEAP32[$11+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$12+8>>2]|0;
       $$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $105 = (($12) + ($$0$i$i$i<<2)|0);
        HEAP32[$105>>2] = 0;
        $106 = (($$0$i$i$i) + 1)|0;
        $$0$i$i$i = $106;
       }
       __THREW__ = 0;
       $107 = (invoke_iii(1119,($11|0),(159130|0))|0);
       $108 = __THREW__; __THREW__ = 0;
       $109 = $108&1;
       do {
        if ($109) {
         $125 = ___cxa_find_matching_catch_2()|0;
         $126 = tempRet0;
         $$082 = $126;$$089 = $125;
        } else {
         ;HEAP32[$10>>2]=HEAP32[$11>>2]|0;HEAP32[$10+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$11+8>>2]|0;
         $$0$i$i$i112 = 0;
         while(1) {
          $exitcond$i$i$i113 = ($$0$i$i$i112|0)==(3);
          if ($exitcond$i$i$i113) {
           break;
          }
          $110 = (($11) + ($$0$i$i$i112<<2)|0);
          HEAP32[$110>>2] = 0;
          $111 = (($$0$i$i$i112) + 1)|0;
          $$0$i$i$i112 = $111;
         }
         $112 = ((($10)) + 11|0);
         $113 = HEAP8[$112>>0]|0;
         $114 = ($113<<24>>24)<(0);
         $115 = HEAP32[$10>>2]|0;
         $116 = $114 ? $115 : $10;
         $117 = ((($10)) + 4|0);
         $118 = HEAP32[$117>>2]|0;
         $119 = $113&255;
         $120 = $114 ? $118 : $119;
         __THREW__ = 0;
         (invoke_iiiii(1118,($85|0),0,($116|0),($120|0))|0);
         $121 = __THREW__; __THREW__ = 0;
         $122 = $121&1;
         if ($122) {
          $127 = ___cxa_find_matching_catch_2()|0;
          $128 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
          $$082 = $128;$$089 = $127;
          break;
         } else {
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
          $$7 = $60;
          break L1;
         }
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
       $$183 = $$082;$$190 = $$089;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
      $$688 = $$183;$$695 = $$190;
     } else {
      $129 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($23,$1,$2)|0);
      $130 = ($129|0)==($23|0);
      $131 = ($129|0)==($1|0);
      $or$cond109 = $130 | $131;
      if ($or$cond109) {
       $$7 = $0;
       break;
      }
      $132 = HEAP8[$129>>0]|0;
      $133 = ($132<<24>>24)==(95);
      if (!($133)) {
       $$7 = $0;
       break;
      }
      $134 = ((($129)) + 1|0);
      $135 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($134,$1,$2)|0);
      $136 = ($135|0)==($134|0);
      if ($136) {
       $$7 = $0;
       break;
      }
      $137 = ((($2)) + 4|0);
      $138 = HEAP32[$137>>2]|0;
      $139 = HEAP32[$2>>2]|0;
      $140 = (($138) - ($139))|0;
      $141 = (($140|0) / 24)&-1;
      $142 = ($141>>>0)<(2);
      if ($142) {
       $$7 = $0;
       break;
      }
      $143 = $138;
      $144 = ((($143)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($13,$144);
      $145 = HEAP32[$137>>2]|0;
      $146 = ((($145)) + -24|0);
      $148 = $145;
      while(1) {
       $147 = ($148|0)==($146|0);
       if ($147) {
        break;
       }
       $149 = ((($148)) + -24|0);
       HEAP32[$137>>2] = $149;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($149);
       $$pre$i = HEAP32[$137>>2]|0;
       $148 = $$pre$i;
      }
      $150 = ((($145)) + -48|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($14,$150);
      $151 = HEAP32[$137>>2]|0;
      $152 = ((($151)) + -24|0);
      $153 = ((($152)) + 11|0);
      $154 = HEAP8[$153>>0]|0;
      $155 = ($154<<24>>24)<(0);
      if ($155) {
       $156 = HEAP32[$152>>2]|0;
       HEAP8[$5>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($156,$5);
       $157 = ((($151)) + -20|0);
       HEAP32[$157>>2] = 0;
      } else {
       HEAP8[$6>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($152,$6);
       HEAP8[$153>>0] = 0;
      }
      __THREW__ = 0;
      invoke_vi(1116,($152|0));
      $158 = __THREW__; __THREW__ = 0;
      $159 = $158&1;
      if ($159) {
       $160 = ___cxa_find_matching_catch_3(0|0)|0;
       $161 = tempRet0;
       ___clang_call_terminate($160);
       // unreachable;
      }
      ;HEAP32[$152>>2]=HEAP32[$13>>2]|0;HEAP32[$152+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$152+8>>2]=HEAP32[$13+8>>2]|0;
      $$0$i$i$i$i = 0;
      while(1) {
       $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
       if ($exitcond$i$i$i$i) {
        break;
       }
       $162 = (($13) + ($$0$i$i$i$i<<2)|0);
       HEAP32[$162>>2] = 0;
       $163 = (($$0$i$i$i$i) + 1)|0;
       $$0$i$i$i$i = $163;
      }
      $164 = ((($13)) + 12|0);
      __THREW__ = 0;
      invoke_viii(1125,($15|0),($164|0),2);
      $165 = __THREW__; __THREW__ = 0;
      $166 = $165&1;
      do {
       if ($166) {
        $186 = ___cxa_find_matching_catch_2()|0;
        $187 = tempRet0;
        $$587 = $187;$$594 = $186;
       } else {
        $167 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159075)|0);
        $168 = ((($15)) + 11|0);
        $169 = HEAP8[$168>>0]|0;
        $170 = ($169<<24>>24)<(0);
        $171 = ((($15)) + 4|0);
        $172 = HEAP32[$171>>2]|0;
        $173 = $169&255;
        $174 = $170 ? $172 : $173;
        $175 = ($167|0)==($174|0);
        if ($175) {
         __THREW__ = 0;
         $176 = (invoke_iiii(1123,($15|0),(159075|0),($167|0))|0);
         $177 = __THREW__; __THREW__ = 0;
         $178 = $177&1;
         if ($178) {
          $179 = ___cxa_find_matching_catch_3(0|0)|0;
          $180 = tempRet0;
          ___clang_call_terminate($179);
          // unreachable;
         }
         $181 = ($176|0)==(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
         if ($181) {
          __THREW__ = 0;
          invoke_vi(1131,($164|0));
          $182 = __THREW__; __THREW__ = 0;
          $183 = $182&1;
          if ($183) {
           $184 = ___cxa_find_matching_catch_2()|0;
           $185 = tempRet0;
           $$587 = $185;$$594 = $184;
           break;
          }
         }
        } else {
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
        }
        __THREW__ = 0;
        invoke_vii(1120,($19|0),($14|0));
        $188 = __THREW__; __THREW__ = 0;
        $189 = $188&1;
        if ($189) {
         $227 = ___cxa_find_matching_catch_2()|0;
         $228 = tempRet0;
         $$486 = $228;$$493 = $227;
        } else {
         __THREW__ = 0;
         $190 = (invoke_iiii(1121,($19|0),0,(159075|0))|0);
         $191 = __THREW__; __THREW__ = 0;
         $192 = $191&1;
         if ($192) {
          $229 = ___cxa_find_matching_catch_2()|0;
          $230 = tempRet0;
          $$385 = $230;$$392 = $229;
         } else {
          ;HEAP32[$18>>2]=HEAP32[$19>>2]|0;HEAP32[$18+4>>2]=HEAP32[$19+4>>2]|0;HEAP32[$18+8>>2]=HEAP32[$19+8>>2]|0;
          $$0$i$i$i119 = 0;
          while(1) {
           $exitcond$i$i$i120 = ($$0$i$i$i119|0)==(3);
           if ($exitcond$i$i$i120) {
            break;
           }
           $193 = (($19) + ($$0$i$i$i119<<2)|0);
           HEAP32[$193>>2] = 0;
           $194 = (($$0$i$i$i119) + 1)|0;
           $$0$i$i$i119 = $194;
          }
          __THREW__ = 0;
          $195 = (invoke_iii(1119,($18|0),(159130|0))|0);
          $196 = __THREW__; __THREW__ = 0;
          $197 = $196&1;
          do {
           if ($197) {
            $231 = ___cxa_find_matching_catch_2()|0;
            $232 = tempRet0;
            $$284 = $232;$$291 = $231;
           } else {
            ;HEAP32[$17>>2]=HEAP32[$18>>2]|0;HEAP32[$17+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$17+8>>2]=HEAP32[$18+8>>2]|0;
            $$0$i$i$i123 = 0;
            while(1) {
             $exitcond$i$i$i124 = ($$0$i$i$i123|0)==(3);
             if ($exitcond$i$i$i124) {
              break;
             }
             $198 = (($18) + ($$0$i$i$i123<<2)|0);
             HEAP32[$198>>2] = 0;
             $199 = (($$0$i$i$i123) + 1)|0;
             $$0$i$i$i123 = $199;
            }
            $200 = ((($164)) + 11|0);
            $201 = HEAP8[$200>>0]|0;
            $202 = ($201<<24>>24)<(0);
            $203 = HEAP32[$164>>2]|0;
            $204 = $202 ? $203 : $164;
            $205 = ((($13)) + 16|0);
            $206 = HEAP32[$205>>2]|0;
            $207 = $201&255;
            $208 = $202 ? $206 : $207;
            __THREW__ = 0;
            $209 = (invoke_iiii(1112,($17|0),($204|0),($208|0))|0);
            $210 = __THREW__; __THREW__ = 0;
            $211 = $210&1;
            if ($211) {
             $233 = ___cxa_find_matching_catch_2()|0;
             $234 = tempRet0;
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($17);
             $$284 = $234;$$291 = $233;
             break;
            }
            ;HEAP32[$16>>2]=HEAP32[$17>>2]|0;HEAP32[$16+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$16+8>>2]=HEAP32[$17+8>>2]|0;
            $$0$i$i$i127 = 0;
            while(1) {
             $exitcond$i$i$i128 = ($$0$i$i$i127|0)==(3);
             if ($exitcond$i$i$i128) {
              break;
             }
             $212 = (($17) + ($$0$i$i$i127<<2)|0);
             HEAP32[$212>>2] = 0;
             $213 = (($$0$i$i$i127) + 1)|0;
             $$0$i$i$i127 = $213;
            }
            $214 = HEAP32[$137>>2]|0;
            $215 = ((($214)) + -12|0);
            $216 = ((($215)) + 11|0);
            $217 = HEAP8[$216>>0]|0;
            $218 = ($217<<24>>24)<(0);
            if ($218) {
             $219 = HEAP32[$215>>2]|0;
             HEAP8[$3>>0] = 0;
             __ZNSt3__211char_traitsIcE6assignERcRKc($219,$3);
             $220 = ((($214)) + -8|0);
             HEAP32[$220>>2] = 0;
            } else {
             HEAP8[$4>>0] = 0;
             __ZNSt3__211char_traitsIcE6assignERcRKc($215,$4);
             HEAP8[$216>>0] = 0;
            }
            __THREW__ = 0;
            invoke_vi(1116,($215|0));
            $221 = __THREW__; __THREW__ = 0;
            $222 = $221&1;
            if ($222) {
             $223 = ___cxa_find_matching_catch_3(0|0)|0;
             $224 = tempRet0;
             ___clang_call_terminate($223);
             // unreachable;
            }
            ;HEAP32[$215>>2]=HEAP32[$16>>2]|0;HEAP32[$215+4>>2]=HEAP32[$16+4>>2]|0;HEAP32[$215+8>>2]=HEAP32[$16+8>>2]|0;
            $$0$i$i$i$i132 = 0;
            while(1) {
             $exitcond$i$i$i$i133 = ($$0$i$i$i$i132|0)==(3);
             if ($exitcond$i$i$i$i133) {
              break;
             }
             $225 = (($16) + ($$0$i$i$i$i132<<2)|0);
             HEAP32[$225>>2] = 0;
             $226 = (($$0$i$i$i$i132) + 1)|0;
             $$0$i$i$i$i132 = $226;
            }
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($17);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
            $$7 = $135;
            break L1;
           }
          } while(0);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
          $$385 = $$284;$$392 = $$291;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
         $$486 = $$385;$$493 = $$392;
        }
        $$587 = $$486;$$594 = $$493;
       }
      } while(0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
      $$688 = $$587;$$695 = $$594;
     }
     ___resumeException($$695|0);
     // unreachable;
    }
   } else {
    $$7 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$7|0);
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = 0;
 $3 = ((($0)) + 4|0);
 HEAP32[$3>>2] = 0;
 $4 = ((($0)) + 8|0);
 $5 = HEAP32[$2>>2]|0;
 HEAP32[$4>>2] = 0;
 $6 = ((($0)) + 12|0);
 HEAP32[$6>>2] = $5;
 __THREW__ = 0;
 invoke_vi(1132,($0|0));
 $7 = __THREW__; __THREW__ = 0;
 $8 = $7&1;
 if (!($8)) {
  $9 = HEAP32[$3>>2]|0;
  __THREW__ = 0;
  invoke_vii(1130,($9|0),($1|0));
  $10 = __THREW__; __THREW__ = 0;
  $11 = $10&1;
  if (!($11)) {
   $14 = HEAP32[$3>>2]|0;
   $15 = ((($14)) + 24|0);
   HEAP32[$3>>2] = $15;
   return;
  }
 }
 $12 = ___cxa_find_matching_catch_2()|0;
 $13 = tempRet0;
 __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($0);
 ___resumeException($12|0);
 // unreachable;
}
function __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE21__push_back_slow_pathIS6_EEvOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$0>>2]|0;
 $6 = (($4) - ($5))|0;
 $7 = $6 >> 4;
 $8 = (($7) + 1)|0;
 $9 = (__ZNKSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE8max_sizeEv($0)|0);
 $10 = ($9>>>0)<($8>>>0);
 if ($10) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $11 = ((($0)) + 12|0);
  $12 = ((($0)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$0>>2]|0;
  $15 = (($13) - ($14))|0;
  $16 = $15 >> 4;
  $17 = $9 >>> 1;
  $18 = ($16>>>0)<($17>>>0);
  $19 = $15 >> 3;
  $20 = ($19>>>0)<($8>>>0);
  $$sroa$speculated$i = $20 ? $8 : $19;
  $$sroa$speculated$$i = $18 ? $$sroa$speculated$i : $9;
  $21 = HEAP32[$3>>2]|0;
  $22 = (($21) - ($14))|0;
  $23 = $22 >> 4;
  __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEEC2EjjS9_($2,$$sroa$speculated$$i,$23,$11);
  $24 = ((($2)) + 8|0);
  $25 = HEAP32[$24>>2]|0;
  $26 = ((($1)) + 12|0);
  HEAP32[$25>>2] = 0;
  $27 = ((($25)) + 4|0);
  HEAP32[$27>>2] = 0;
  $28 = ((($25)) + 8|0);
  $29 = HEAP32[$26>>2]|0;
  HEAP32[$28>>2] = 0;
  $30 = ((($25)) + 12|0);
  HEAP32[$30>>2] = $29;
  $31 = HEAP32[$1>>2]|0;
  HEAP32[$25>>2] = $31;
  $32 = ((($1)) + 4|0);
  $33 = HEAP32[$32>>2]|0;
  HEAP32[$27>>2] = $33;
  $34 = ((($1)) + 8|0);
  $35 = HEAP32[$34>>2]|0;
  $36 = ((($25)) + 8|0);
  HEAP32[$36>>2] = $35;
  HEAP32[$34>>2] = 0;
  HEAP32[$32>>2] = 0;
  HEAP32[$1>>2] = 0;
  $37 = ((($25)) + 16|0);
  HEAP32[$24>>2] = $37;
  __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS6_RS7_EE($0,$2);
  __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEED2Ev($2);
  STACKTOP = sp;return;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_function_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0101 = 0, $$0105$ph = 0, $$0105$ph$clear = 0, $$0111 = 0, $$088 = 0, $$097 = 0, $$1112 = 0, $$1112$ph = 0, $$3100 = 0, $$3104 = 0, $$8 = 0, $$not = 0, $$not114 = 0, $$not115 = 0, $$pre$i = 0, $$pre$i119 = 0, $$pre$phi160Z2D = 0, $$pre$phi162Z2D = 0;
 var $$pre159 = 0, $$pre161 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0;
 var $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $brmerge = 0;
 var $brmerge117 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 24|0;
 $4 = sp + 12|0;
 $5 = sp;
 $6 = ($0|0)==($1|0);
 L1: do {
  if ($6) {
   $$8 = $0;
  } else {
   $7 = HEAP8[$0>>0]|0;
   $8 = ($7<<24>>24)==(70);
   if ($8) {
    $9 = ((($0)) + 1|0);
    $10 = ($9|0)==($1|0);
    if ($10) {
     $$8 = $0;
    } else {
     $11 = HEAP8[$9>>0]|0;
     $12 = ($11<<24>>24)==(89);
     if ($12) {
      $13 = ((($0)) + 2|0);
      $14 = ($13|0)==($1|0);
      if ($14) {
       $$8 = $0;
       break;
      } else {
       $$0111 = $13;
      }
     } else {
      $$0111 = $9;
     }
     $15 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($$0111,$1,$2)|0);
     $16 = ($15|0)==($$0111|0);
     if ($16) {
      $$8 = $0;
     } else {
      ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
      $17 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159083)|0);
      $18 = ($17>>>0)>(4294967279);
      if ($18) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($4);
       // unreachable;
      }
      $19 = ($17>>>0)<(11);
      if ($19) {
       $20 = $17&255;
       $21 = ((($4)) + 11|0);
       HEAP8[$21>>0] = $20;
       $$pre161 = ((($4)) + 4|0);
       $$0$i$i = $4;$$pre$phi160Z2D = $21;$$pre$phi162Z2D = $$pre161;
      } else {
       $22 = (($17) + 16)|0;
       $23 = $22 & -16;
       $24 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($23)|0);
       HEAP32[$4>>2] = $24;
       $25 = $23 | -2147483648;
       $26 = ((($4)) + 8|0);
       HEAP32[$26>>2] = $25;
       $27 = ((($4)) + 4|0);
       HEAP32[$27>>2] = $17;
       $$pre159 = ((($4)) + 11|0);
       $$0$i$i = $24;$$pre$phi160Z2D = $$pre159;$$pre$phi162Z2D = $27;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,159083,$17)|0);
      $28 = (($$0$i$i) + ($17)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($28,$3);
      $29 = ((($2)) + 4|0);
      $30 = ((($5)) + 11|0);
      $31 = ((($5)) + 4|0);
      $$0105$ph = 0;$$1112$ph = $15;
      L16: while(1) {
       $$1112 = $$1112$ph;
       L18: while(1) {
        $32 = ($$1112|0)==($1|0);
        if ($32) {
         label = 15;
         break L16;
        }
        $40 = HEAP8[$$1112>>0]|0;
        $41 = ((($$1112)) + 1|0);
        switch ($40<<24>>24) {
        case 69:  {
         label = 20;
         break L16;
         break;
        }
        case 118:  {
         $$1112 = $41;
         continue L18;
         break;
        }
        default: {
        }
        }
        $$not = ($40<<24>>24)!=(82);
        $$not114 = ($41|0)==($1|0);
        $brmerge = $$not114 | $$not;
        if (!($brmerge)) {
         $44 = HEAP8[$41>>0]|0;
         $45 = ($44<<24>>24)==(69);
         if ($45) {
          $$0105$ph = 1;$$1112$ph = $41;
          continue L16;
         }
        }
        $$not115 = ($40<<24>>24)!=(79);
        $brmerge117 = $$not114 | $$not115;
        if (!($brmerge117)) {
         $46 = HEAP8[$41>>0]|0;
         $47 = ($46<<24>>24)==(69);
         if ($47) {
          $$0105$ph = 2;$$1112$ph = $41;
          continue L16;
         }
        }
        $48 = HEAP32[$29>>2]|0;
        $49 = HEAP32[$2>>2]|0;
        $50 = (($48) - ($49))|0;
        $51 = (($50|0) / 24)&-1;
        __THREW__ = 0;
        $52 = (invoke_iiii(1115,($$1112|0),($1|0),($2|0))|0);
        $53 = __THREW__; __THREW__ = 0;
        $54 = $53&1;
        if ($54) {
         label = 27;
         break L16;
        }
        $55 = HEAP32[$29>>2]|0;
        $56 = HEAP32[$2>>2]|0;
        $57 = (($55) - ($56))|0;
        $58 = (($57|0) / 24)&-1;
        $59 = ($52|0)==($$1112|0);
        $60 = ($52|0)==($1|0);
        $or$cond = $59 | $60;
        if ($or$cond) {
         label = 49;
         break L16;
        } else {
         $$088 = $51;
        }
        while(1) {
         $63 = ($$088>>>0)<($58>>>0);
         if (!($63)) {
          $$0 = $51;
          break;
         }
         $64 = HEAP8[$$pre$phi160Z2D>>0]|0;
         $65 = ($64<<24>>24)<(0);
         $66 = HEAP32[$$pre$phi162Z2D>>2]|0;
         $67 = $64&255;
         $68 = $65 ? $66 : $67;
         $69 = ($68>>>0)>(1);
         if ($69) {
          __THREW__ = 0;
          (invoke_iii(1119,($4|0),(159305|0))|0);
          $70 = __THREW__; __THREW__ = 0;
          $71 = $70&1;
          if ($71) {
           label = 31;
           break L16;
          }
         }
         $74 = HEAP32[$2>>2]|0;
         $75 = (($74) + (($$088*24)|0)|0);
         __THREW__ = 0;
         invoke_vii(1120,($5|0),($75|0));
         $76 = __THREW__; __THREW__ = 0;
         $77 = $76&1;
         if ($77) {
          label = 35;
          break L16;
         }
         $78 = HEAP8[$30>>0]|0;
         $79 = ($78<<24>>24)<(0);
         $80 = HEAP32[$5>>2]|0;
         $81 = $79 ? $80 : $5;
         $82 = HEAP32[$31>>2]|0;
         $83 = $78&255;
         $84 = $79 ? $82 : $83;
         __THREW__ = 0;
         (invoke_iiii(1112,($4|0),($81|0),($84|0))|0);
         $85 = __THREW__; __THREW__ = 0;
         $86 = $85&1;
         if ($86) {
          label = 36;
          break L16;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         $87 = (($$088) + 1)|0;
         $$088 = $87;
        }
        while(1) {
         $92 = ($$0>>>0)<($58>>>0);
         if (!($92)) {
          $$1112 = $52;
          continue L18;
         }
         $93 = HEAP32[$29>>2]|0;
         $94 = ((($93)) + -24|0);
         $96 = $93;
         while(1) {
          $95 = ($96|0)==($94|0);
          if ($95) {
           break;
          }
          $97 = ((($96)) + -24|0);
          HEAP32[$29>>2] = $97;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($97);
          $$pre$i119 = HEAP32[$29>>2]|0;
          $96 = $$pre$i119;
         }
         $98 = (($$0) + 1)|0;
         $$0 = $98;
        }
       }
      }
      L44: do {
       if ((label|0) == 15) {
        $33 = HEAP32[$29>>2]|0;
        $34 = ((($33)) + -24|0);
        $36 = $33;
        while(1) {
         $35 = ($36|0)==($34|0);
         if ($35) {
          label = 49;
          break L44;
         }
         $37 = ((($36)) + -24|0);
         HEAP32[$29>>2] = $37;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($37);
         $$pre$i = HEAP32[$29>>2]|0;
         $36 = $$pre$i;
        }
       }
       else if ((label|0) == 20) {
        __THREW__ = 0;
        (invoke_iii(1119,($4|0),(159081|0))|0);
        $42 = __THREW__; __THREW__ = 0;
        $43 = $42&1;
        L46: do {
         if (!($43)) {
          $$0105$ph$clear = $$0105$ph & 3;
          switch ($$0105$ph$clear<<24>>24) {
          case 1:  {
           __THREW__ = 0;
           (invoke_iii(1119,($4|0),(160509|0))|0);
           $99 = __THREW__; __THREW__ = 0;
           $100 = $99&1;
           if ($100) {
            break L46;
           }
           break;
          }
          case 2:  {
           __THREW__ = 0;
           (invoke_iii(1119,($4|0),(160512|0))|0);
           $101 = __THREW__; __THREW__ = 0;
           $102 = $101&1;
           if ($102) {
            break L46;
           }
           break;
          }
          default: {
          }
          }
          $103 = HEAP32[$2>>2]|0;
          $104 = HEAP32[$29>>2]|0;
          $105 = ($103|0)==($104|0);
          if ($105) {
           label = 49;
           break L44;
          }
          $106 = ((($104)) + -24|0);
          __THREW__ = 0;
          (invoke_iii(1119,($106|0),(159115|0))|0);
          $107 = __THREW__; __THREW__ = 0;
          $108 = $107&1;
          if (!($108)) {
           $109 = HEAP32[$29>>2]|0;
           $110 = ((($109)) + -12|0);
           $111 = HEAP8[$$pre$phi160Z2D>>0]|0;
           $112 = ($111<<24>>24)<(0);
           $113 = HEAP32[$4>>2]|0;
           $114 = $112 ? $113 : $4;
           $115 = HEAP32[$$pre$phi162Z2D>>2]|0;
           $116 = $111&255;
           $117 = $112 ? $115 : $116;
           __THREW__ = 0;
           (invoke_iiiii(1118,($110|0),0,($114|0),($117|0))|0);
           $118 = __THREW__; __THREW__ = 0;
           $119 = $118&1;
           if (!($119)) {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
            $$8 = $41;
            break L1;
           }
          }
         }
        } while(0);
        $38 = ___cxa_find_matching_catch_2()|0;
        $39 = tempRet0;
        $$3100 = $39;$$3104 = $38;
       }
       else if ((label|0) == 27) {
        $61 = ___cxa_find_matching_catch_2()|0;
        $62 = tempRet0;
        $$3100 = $62;$$3104 = $61;
       }
       else if ((label|0) == 31) {
        $72 = ___cxa_find_matching_catch_2()|0;
        $73 = tempRet0;
        $$3100 = $73;$$3104 = $72;
       }
       else if ((label|0) == 35) {
        $88 = ___cxa_find_matching_catch_2()|0;
        $89 = tempRet0;
        $$0101 = $88;$$097 = $89;
        label = 37;
       }
       else if ((label|0) == 36) {
        $90 = ___cxa_find_matching_catch_2()|0;
        $91 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
        $$0101 = $90;$$097 = $91;
        label = 37;
       }
      } while(0);
      if ((label|0) == 37) {
       $$3100 = $$097;$$3104 = $$0101;
      }
      else if ((label|0) == 49) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
       $$8 = $0;
       break;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      ___resumeException($$3104|0);
      // unreachable;
     }
    }
   } else {
    $$8 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$8|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_128parse_pointer_to_member_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i63 = 0, $$0$i$i$i$i80 = 0, $$0$i$i$i$i85 = 0, $$0$i$i$i51 = 0, $$0$i$i$i54 = 0, $$0$i$i$i58 = 0, $$0$i$i$i66 = 0, $$0$i$i$i70 = 0, $$0$i$i$i74 = 0, $$033 = 0, $$1 = 0, $$134 = 0, $$2 = 0, $$235 = 0, $$3 = 0, $$336 = 0, $$345 = 0;
 var $$4 = 0, $$437 = 0, $$5 = 0, $$538 = 0, $$6 = 0, $$639 = 0, $$pre$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0;
 var $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0;
 var $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i$i64 = 0, $exitcond$i$i$i$i81 = 0, $exitcond$i$i$i$i86 = 0, $exitcond$i$i$i52 = 0, $exitcond$i$i$i55 = 0, $exitcond$i$i$i59 = 0, $exitcond$i$i$i67 = 0, $exitcond$i$i$i71 = 0, $exitcond$i$i$i75 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $3 = sp + 167|0;
 $4 = sp + 166|0;
 $5 = sp + 165|0;
 $6 = sp + 164|0;
 $7 = sp + 163|0;
 $8 = sp + 162|0;
 $9 = sp + 161|0;
 $10 = sp + 160|0;
 $11 = sp + 136|0;
 $12 = sp + 112|0;
 $13 = sp + 96|0;
 $14 = sp + 84|0;
 $15 = sp + 72|0;
 $16 = sp + 60|0;
 $17 = sp + 48|0;
 $18 = sp + 36|0;
 $19 = sp + 24|0;
 $20 = sp + 12|0;
 $21 = sp;
 $22 = ($0|0)==($1|0);
 do {
  if ($22) {
   $$345 = $0;
  } else {
   $23 = HEAP8[$0>>0]|0;
   $24 = ($23<<24>>24)==(77);
   if ($24) {
    $25 = ((($0)) + 1|0);
    $26 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($25,$1,$2)|0);
    $27 = ($26|0)==($25|0);
    if ($27) {
     $$345 = $0;
    } else {
     $28 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($26,$1,$2)|0);
     $29 = ($28|0)==($26|0);
     if ($29) {
      $$345 = $0;
     } else {
      $30 = ((($2)) + 4|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = HEAP32[$2>>2]|0;
      $33 = (($31) - ($32))|0;
      $34 = (($33|0) / 24)&-1;
      $35 = ($34>>>0)<(2);
      if ($35) {
       $$345 = $0;
      } else {
       $36 = $31;
       $37 = ((($36)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($11,$37);
       $38 = HEAP32[$30>>2]|0;
       $39 = ((($38)) + -24|0);
       $41 = $38;
       while(1) {
        $40 = ($41|0)==($39|0);
        if ($40) {
         break;
        }
        $42 = ((($41)) + -24|0);
        HEAP32[$30>>2] = $42;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($42);
        $$pre$i = HEAP32[$30>>2]|0;
        $41 = $$pre$i;
       }
       $43 = ((($38)) + -48|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($12,$43);
       $44 = ((($11)) + 12|0);
       $45 = ((($44)) + 11|0);
       $46 = HEAP8[$45>>0]|0;
       $47 = ($46<<24>>24)<(0);
       $48 = HEAP32[$44>>2]|0;
       $49 = $47 ? $48 : $44;
       $50 = HEAP8[$49>>0]|0;
       $51 = ($50<<24>>24)==(40);
       L11: do {
        if ($51) {
         __THREW__ = 0;
         $52 = (invoke_iii(1119,($11|0),(159083|0))|0);
         $53 = __THREW__; __THREW__ = 0;
         $54 = $53&1;
         if ($54) {
          $109 = ___cxa_find_matching_catch_2()|0;
          $110 = tempRet0;
          $$2 = $110;$$235 = $109;
         } else {
          ;HEAP32[$15>>2]=HEAP32[$11>>2]|0;HEAP32[$15+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$11+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $55 = (($11) + ($$0$i$i$i<<2)|0);
           HEAP32[$55>>2] = 0;
           $56 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $56;
          }
          __THREW__ = 0;
          invoke_vii(1120,($16|0),($12|0));
          $57 = __THREW__; __THREW__ = 0;
          $58 = $57&1;
          if ($58) {
           $111 = ___cxa_find_matching_catch_2()|0;
           $112 = tempRet0;
           $$1 = $112;$$134 = $111;
          } else {
           $59 = ((($16)) + 11|0);
           $60 = HEAP8[$59>>0]|0;
           $61 = ($60<<24>>24)<(0);
           $62 = HEAP32[$16>>2]|0;
           $63 = $61 ? $62 : $16;
           $64 = ((($16)) + 4|0);
           $65 = HEAP32[$64>>2]|0;
           $66 = $60&255;
           $67 = $61 ? $65 : $66;
           __THREW__ = 0;
           $68 = (invoke_iiii(1112,($15|0),($63|0),($67|0))|0);
           $69 = __THREW__; __THREW__ = 0;
           $70 = $69&1;
           do {
            if ($70) {
             $113 = ___cxa_find_matching_catch_2()|0;
             $114 = tempRet0;
             $$0 = $114;$$033 = $113;
            } else {
             ;HEAP32[$14>>2]=HEAP32[$15>>2]|0;HEAP32[$14+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$15+8>>2]|0;
             $$0$i$i$i51 = 0;
             while(1) {
              $exitcond$i$i$i52 = ($$0$i$i$i51|0)==(3);
              if ($exitcond$i$i$i52) {
               break;
              }
              $71 = (($15) + ($$0$i$i$i51<<2)|0);
              HEAP32[$71>>2] = 0;
              $72 = (($$0$i$i$i51) + 1)|0;
              $$0$i$i$i51 = $72;
             }
             __THREW__ = 0;
             $73 = (invoke_iii(1119,($14|0),(160505|0))|0);
             $74 = __THREW__; __THREW__ = 0;
             $75 = $74&1;
             if ($75) {
              $115 = ___cxa_find_matching_catch_2()|0;
              $116 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
              $$0 = $116;$$033 = $115;
              break;
             }
             ;HEAP32[$13>>2]=HEAP32[$14>>2]|0;HEAP32[$13+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$13+8>>2]=HEAP32[$14+8>>2]|0;
             $$0$i$i$i54 = 0;
             while(1) {
              $exitcond$i$i$i55 = ($$0$i$i$i54|0)==(3);
              if ($exitcond$i$i$i55) {
               break;
              }
              $76 = (($14) + ($$0$i$i$i54<<2)|0);
              HEAP32[$76>>2] = 0;
              $77 = (($$0$i$i$i54) + 1)|0;
              $$0$i$i$i54 = $77;
             }
             $78 = HEAP32[$30>>2]|0;
             $79 = ((($78)) + -24|0);
             $80 = ((($79)) + 11|0);
             $81 = HEAP8[$80>>0]|0;
             $82 = ($81<<24>>24)<(0);
             if ($82) {
              $83 = HEAP32[$79>>2]|0;
              HEAP8[$9>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($83,$9);
              $84 = ((($78)) + -20|0);
              HEAP32[$84>>2] = 0;
             } else {
              HEAP8[$10>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($79,$10);
              HEAP8[$80>>0] = 0;
             }
             __THREW__ = 0;
             invoke_vi(1116,($79|0));
             $85 = __THREW__; __THREW__ = 0;
             $86 = $85&1;
             if ($86) {
              $87 = ___cxa_find_matching_catch_3(0|0)|0;
              $88 = tempRet0;
              ___clang_call_terminate($87);
              // unreachable;
             }
             ;HEAP32[$79>>2]=HEAP32[$13>>2]|0;HEAP32[$79+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$79+8>>2]=HEAP32[$13+8>>2]|0;
             $$0$i$i$i$i = 0;
             while(1) {
              $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
              if ($exitcond$i$i$i$i) {
               break;
              }
              $89 = (($13) + ($$0$i$i$i$i<<2)|0);
              HEAP32[$89>>2] = 0;
              $90 = (($$0$i$i$i$i) + 1)|0;
              $$0$i$i$i$i = $90;
             }
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
             __THREW__ = 0;
             $91 = (invoke_iiii(1121,($44|0),0,(159081|0))|0);
             $92 = __THREW__; __THREW__ = 0;
             $93 = $92&1;
             if ($93) {
              $117 = ___cxa_find_matching_catch_2()|0;
              $118 = tempRet0;
              $$6 = $118;$$639 = $117;
              label = 89;
              break L11;
             }
             ;HEAP32[$17>>2]=HEAP32[$44>>2]|0;HEAP32[$17+4>>2]=HEAP32[$44+4>>2]|0;HEAP32[$17+8>>2]=HEAP32[$44+8>>2]|0;
             $$0$i$i$i58 = 0;
             while(1) {
              $exitcond$i$i$i59 = ($$0$i$i$i58|0)==(3);
              if ($exitcond$i$i$i59) {
               break;
              }
              $94 = (($44) + ($$0$i$i$i58<<2)|0);
              HEAP32[$94>>2] = 0;
              $95 = (($$0$i$i$i58) + 1)|0;
              $$0$i$i$i58 = $95;
             }
             $96 = HEAP32[$30>>2]|0;
             $97 = ((($96)) + -12|0);
             $98 = ((($97)) + 11|0);
             $99 = HEAP8[$98>>0]|0;
             $100 = ($99<<24>>24)<(0);
             if ($100) {
              $101 = HEAP32[$97>>2]|0;
              HEAP8[$7>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($101,$7);
              $102 = ((($96)) + -8|0);
              HEAP32[$102>>2] = 0;
             } else {
              HEAP8[$8>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($97,$8);
              HEAP8[$98>>0] = 0;
             }
             __THREW__ = 0;
             invoke_vi(1116,($97|0));
             $103 = __THREW__; __THREW__ = 0;
             $104 = $103&1;
             if ($104) {
              $105 = ___cxa_find_matching_catch_3(0|0)|0;
              $106 = tempRet0;
              ___clang_call_terminate($105);
              // unreachable;
             }
             ;HEAP32[$97>>2]=HEAP32[$17>>2]|0;HEAP32[$97+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$97+8>>2]=HEAP32[$17+8>>2]|0;
             $$0$i$i$i$i63 = 0;
             while(1) {
              $exitcond$i$i$i$i64 = ($$0$i$i$i$i63|0)==(3);
              if ($exitcond$i$i$i$i64) {
               break;
              }
              $107 = (($17) + ($$0$i$i$i$i63<<2)|0);
              HEAP32[$107>>2] = 0;
              $108 = (($$0$i$i$i$i63) + 1)|0;
              $$0$i$i$i$i63 = $108;
             }
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($17);
             label = 88;
             break L11;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
           $$1 = $$0;$$134 = $$033;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
          $$2 = $$1;$$235 = $$134;
         }
         $$6 = $$2;$$639 = $$235;
         label = 89;
        } else {
         __THREW__ = 0;
         $119 = (invoke_iii(1119,($11|0),(159115|0))|0);
         $120 = __THREW__; __THREW__ = 0;
         $121 = $120&1;
         if ($121) {
          $171 = ___cxa_find_matching_catch_2()|0;
          $172 = tempRet0;
          $$5 = $172;$$538 = $171;
         } else {
          ;HEAP32[$20>>2]=HEAP32[$11>>2]|0;HEAP32[$20+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$20+8>>2]=HEAP32[$11+8>>2]|0;
          $$0$i$i$i66 = 0;
          while(1) {
           $exitcond$i$i$i67 = ($$0$i$i$i66|0)==(3);
           if ($exitcond$i$i$i67) {
            break;
           }
           $122 = (($11) + ($$0$i$i$i66<<2)|0);
           HEAP32[$122>>2] = 0;
           $123 = (($$0$i$i$i66) + 1)|0;
           $$0$i$i$i66 = $123;
          }
          __THREW__ = 0;
          invoke_vii(1120,($21|0),($12|0));
          $124 = __THREW__; __THREW__ = 0;
          $125 = $124&1;
          if ($125) {
           $173 = ___cxa_find_matching_catch_2()|0;
           $174 = tempRet0;
           $$4 = $174;$$437 = $173;
          } else {
           $126 = ((($21)) + 11|0);
           $127 = HEAP8[$126>>0]|0;
           $128 = ($127<<24>>24)<(0);
           $129 = HEAP32[$21>>2]|0;
           $130 = $128 ? $129 : $21;
           $131 = ((($21)) + 4|0);
           $132 = HEAP32[$131>>2]|0;
           $133 = $127&255;
           $134 = $128 ? $132 : $133;
           __THREW__ = 0;
           $135 = (invoke_iiii(1112,($20|0),($130|0),($134|0))|0);
           $136 = __THREW__; __THREW__ = 0;
           $137 = $136&1;
           do {
            if ($137) {
             $175 = ___cxa_find_matching_catch_2()|0;
             $176 = tempRet0;
             $$3 = $176;$$336 = $175;
            } else {
             ;HEAP32[$19>>2]=HEAP32[$20>>2]|0;HEAP32[$19+4>>2]=HEAP32[$20+4>>2]|0;HEAP32[$19+8>>2]=HEAP32[$20+8>>2]|0;
             $$0$i$i$i70 = 0;
             while(1) {
              $exitcond$i$i$i71 = ($$0$i$i$i70|0)==(3);
              if ($exitcond$i$i$i71) {
               break;
              }
              $138 = (($20) + ($$0$i$i$i70<<2)|0);
              HEAP32[$138>>2] = 0;
              $139 = (($$0$i$i$i70) + 1)|0;
              $$0$i$i$i70 = $139;
             }
             __THREW__ = 0;
             $140 = (invoke_iii(1119,($19|0),(160505|0))|0);
             $141 = __THREW__; __THREW__ = 0;
             $142 = $141&1;
             if ($142) {
              $177 = ___cxa_find_matching_catch_2()|0;
              $178 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
              $$3 = $178;$$336 = $177;
              break;
             }
             ;HEAP32[$18>>2]=HEAP32[$19>>2]|0;HEAP32[$18+4>>2]=HEAP32[$19+4>>2]|0;HEAP32[$18+8>>2]=HEAP32[$19+8>>2]|0;
             $$0$i$i$i74 = 0;
             while(1) {
              $exitcond$i$i$i75 = ($$0$i$i$i74|0)==(3);
              if ($exitcond$i$i$i75) {
               break;
              }
              $143 = (($19) + ($$0$i$i$i74<<2)|0);
              HEAP32[$143>>2] = 0;
              $144 = (($$0$i$i$i74) + 1)|0;
              $$0$i$i$i74 = $144;
             }
             $145 = HEAP32[$30>>2]|0;
             $146 = ((($145)) + -24|0);
             $147 = ((($146)) + 11|0);
             $148 = HEAP8[$147>>0]|0;
             $149 = ($148<<24>>24)<(0);
             if ($149) {
              $150 = HEAP32[$146>>2]|0;
              HEAP8[$5>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($150,$5);
              $151 = ((($145)) + -20|0);
              HEAP32[$151>>2] = 0;
             } else {
              HEAP8[$6>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($146,$6);
              HEAP8[$147>>0] = 0;
             }
             __THREW__ = 0;
             invoke_vi(1116,($146|0));
             $152 = __THREW__; __THREW__ = 0;
             $153 = $152&1;
             if ($153) {
              $154 = ___cxa_find_matching_catch_3(0|0)|0;
              $155 = tempRet0;
              ___clang_call_terminate($154);
              // unreachable;
             }
             ;HEAP32[$146>>2]=HEAP32[$18>>2]|0;HEAP32[$146+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$146+8>>2]=HEAP32[$18+8>>2]|0;
             $$0$i$i$i$i80 = 0;
             while(1) {
              $exitcond$i$i$i$i81 = ($$0$i$i$i$i80|0)==(3);
              if ($exitcond$i$i$i$i81) {
               break;
              }
              $156 = (($18) + ($$0$i$i$i$i80<<2)|0);
              HEAP32[$156>>2] = 0;
              $157 = (($$0$i$i$i$i80) + 1)|0;
              $$0$i$i$i$i80 = $157;
             }
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($21);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($20);
             $158 = HEAP32[$30>>2]|0;
             $159 = ((($158)) + -12|0);
             $160 = ((($159)) + 11|0);
             $161 = HEAP8[$160>>0]|0;
             $162 = ($161<<24>>24)<(0);
             if ($162) {
              $163 = HEAP32[$159>>2]|0;
              HEAP8[$3>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($163,$3);
              $164 = ((($158)) + -8|0);
              HEAP32[$164>>2] = 0;
             } else {
              HEAP8[$4>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($159,$4);
              HEAP8[$160>>0] = 0;
             }
             __THREW__ = 0;
             invoke_vi(1116,($159|0));
             $165 = __THREW__; __THREW__ = 0;
             $166 = $165&1;
             if ($166) {
              $167 = ___cxa_find_matching_catch_3(0|0)|0;
              $168 = tempRet0;
              ___clang_call_terminate($167);
              // unreachable;
             }
             ;HEAP32[$159>>2]=HEAP32[$44>>2]|0;HEAP32[$159+4>>2]=HEAP32[$44+4>>2]|0;HEAP32[$159+8>>2]=HEAP32[$44+8>>2]|0;
             $$0$i$i$i$i85 = 0;
             while(1) {
              $exitcond$i$i$i$i86 = ($$0$i$i$i$i85|0)==(3);
              if ($exitcond$i$i$i$i86) {
               label = 88;
               break L11;
              }
              $169 = (($44) + ($$0$i$i$i$i85<<2)|0);
              HEAP32[$169>>2] = 0;
              $170 = (($$0$i$i$i$i85) + 1)|0;
              $$0$i$i$i$i85 = $170;
             }
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($21);
           $$4 = $$3;$$437 = $$336;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($20);
          $$5 = $$4;$$538 = $$437;
         }
         $$6 = $$5;$$639 = $$538;
         label = 89;
        }
       } while(0);
       if ((label|0) == 88) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
        $$345 = $28;
        break;
       }
       else if ((label|0) == 89) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
        ___resumeException($$639|0);
        // unreachable;
       }
      }
     }
    }
   } else {
    $$345 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$345|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i = 0, $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $3 = 0, $4 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $4 = ((($1)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 $7 = ((($1)) + 4|0);
 $8 = HEAP32[$7>>2]|0;
 $9 = $5&255;
 $10 = $6 ? $8 : $9;
 $11 = HEAP32[$1>>2]|0;
 $12 = $6 ? $11 : $1;
 $13 = ($10>>>0)<($2>>>0);
 $$sroa$speculated = $13 ? $10 : $2;
 $14 = ($$sroa$speculated>>>0)>(4294967279);
 if ($14) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $15 = ($$sroa$speculated>>>0)<(11);
 if ($15) {
  $16 = $$sroa$speculated&255;
  $17 = ((($0)) + 11|0);
  HEAP8[$17>>0] = $16;
  $$0$i = $0;
 } else {
  $18 = (($$sroa$speculated) + 16)|0;
  $19 = $18 & -16;
  $20 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($19)|0);
  HEAP32[$0>>2] = $20;
  $21 = $19 | -2147483648;
  $22 = ((($0)) + 8|0);
  HEAP32[$22>>2] = $21;
  $23 = ((($0)) + 4|0);
  HEAP32[$23>>2] = $$sroa$speculated;
  $$0$i = $20;
 }
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i,$12,$$sroa$speculated)|0);
 $24 = (($$0$i) + ($$sroa$speculated)|0);
 HEAP8[$3>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($24,$3);
 STACKTOP = sp;return;
}
function __ZNKSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7compareEjjPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$$ = 0, $$sroa$speculated8 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ((($0)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 if ($5) {
  $6 = ((($0)) + 4|0);
  $7 = HEAP32[$6>>2]|0;
  $12 = $7;
 } else {
  $8 = $4&255;
  $12 = $8;
 }
 $9 = ($2|0)==(-1);
 if ($9) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_out_of_rangeEv($0);
  // unreachable;
 }
 if ($5) {
  $10 = HEAP32[$0>>2]|0;
  $13 = $10;
 } else {
  $13 = $0;
 }
 $11 = ($12>>>0)>($2>>>0);
 $$sroa$speculated8 = $11 ? $2 : $12;
 $14 = (__ZNSt3__211char_traitsIcE7compareEPKcS3_j($13,$1,$$sroa$speculated8)|0);
 $15 = ($14|0)==(0);
 if ($15) {
  $16 = ($12>>>0)<($2>>>0);
  $$ = $11&1;
  $$$ = $16 ? -1 : $$;
  return ($$$|0);
 } else {
  return ($14|0);
 }
 return (0)|0;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7replaceEjjPKc($0) {
 $0 = $0|0;
 var $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160502)|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7replaceEjjPKcj($0,$1);
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$067 = 0, $$068 = 0, $$069 = 0, $$070 = 0, $$1 = 0, $$sroa$077$0 = 0, $$sroa$083$0 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $isdigit = 0;
 var $isdigit72 = 0, $isdigittmp = 0, $isdigittmp71 = 0, $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $3 = sp + 64|0;
 $4 = sp + 40|0;
 $5 = sp + 16|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(1);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(84);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = ($14<<24>>24)==(95);
    $16 = ((($2)) + 62|0);
    $17 = ((($2)) + 32|0);
    do {
     if ($15) {
      $18 = HEAP32[$17>>2]|0;
      $19 = ((($2)) + 36|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = ($18|0)==($20|0);
      if ($21) {
       $$1 = $0;
       break L1;
      }
      $22 = ((($20)) + -16|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = ((($20)) + -12|0);
      $25 = HEAP32[$24>>2]|0;
      $26 = ($23|0)==($25|0);
      $27 = ((($0)) + 2|0);
      if ($26) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj3EEERAT__Kc($4);
       $40 = ((($2)) + 4|0);
       $41 = HEAP32[$40>>2]|0;
       $42 = ((($2)) + 8|0);
       $43 = HEAP32[$42>>2]|0;
       $44 = ($41>>>0)<($43>>>0);
       if ($44) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($41,$4);
        $45 = HEAP32[$40>>2]|0;
        $46 = ((($45)) + 24|0);
        HEAP32[$40>>2] = $46;
       } else {
        __THREW__ = 0;
        invoke_vii(1128,($2|0),($4|0));
        $47 = __THREW__; __THREW__ = 0;
        $48 = $47&1;
        if ($48) {
         $49 = ___cxa_find_matching_catch_2()|0;
         $50 = tempRet0;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
         $$069 = $50;$$070 = $49;
         break;
        }
       }
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
       HEAP8[$16>>0] = 1;
       $$1 = $27;
       break L1;
      }
      $28 = HEAP32[$23>>2]|0;
      $29 = ((($23)) + 4|0);
      $30 = HEAP32[$29>>2]|0;
      $31 = ((($2)) + 4|0);
      $32 = ((($2)) + 8|0);
      $$sroa$083$0 = $28;
      while(1) {
       $33 = ($$sroa$083$0|0)==($30|0);
       if ($33) {
        $$1 = $27;
        break L1;
       }
       $34 = HEAP32[$31>>2]|0;
       $35 = HEAP32[$32>>2]|0;
       $36 = ($34|0)==($35|0);
       if ($36) {
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($2,$$sroa$083$0);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($34,$$sroa$083$0);
        $37 = HEAP32[$31>>2]|0;
        $38 = ((($37)) + 24|0);
        HEAP32[$31>>2] = $38;
       }
       $39 = ((($$sroa$083$0)) + 24|0);
       $$sroa$083$0 = $39;
      }
     } else {
      $51 = $14 << 24 >> 24;
      $isdigittmp = (($51) + -48)|0;
      $isdigit = ($isdigittmp>>>0)<(10);
      if (!($isdigit)) {
       $$1 = $0;
       break L1;
      }
      $52 = ((($0)) + 2|0);
      $$067 = $isdigittmp;$$068 = $52;
      while(1) {
       $53 = ($$068|0)==($1|0);
       $54 = ((($$068)) + 1|0);
       if ($53) {
        $$1 = $0;
        break L1;
       }
       $55 = HEAP8[$$068>>0]|0;
       $56 = $55 << 24 >> 24;
       $isdigittmp71 = (($56) + -48)|0;
       $isdigit72 = ($isdigittmp71>>>0)<(10);
       if (!($isdigit72)) {
        break;
       }
       $57 = ($$067*10)|0;
       $58 = (($isdigittmp71) + ($57))|0;
       $$067 = $58;$$068 = $54;
      }
      $59 = ($55<<24>>24)==(95);
      if (!($59)) {
       $$1 = $0;
       break L1;
      }
      $60 = HEAP32[$17>>2]|0;
      $61 = ((($2)) + 36|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ($60|0)==($62|0);
      if ($63) {
       $$1 = $0;
       break L1;
      }
      $64 = (($$067) + 1)|0;
      $65 = ((($62)) + -16|0);
      $66 = ((($62)) + -12|0);
      $67 = HEAP32[$66>>2]|0;
      $68 = HEAP32[$65>>2]|0;
      $69 = (($67) - ($68))|0;
      $70 = $69 >> 4;
      $71 = ($64>>>0)<($70>>>0);
      $72 = $68;
      if ($71) {
       $73 = (($72) + ($64<<4)|0);
       $74 = HEAP32[$73>>2]|0;
       $75 = (((($72) + ($64<<4)|0)) + 4|0);
       $76 = HEAP32[$75>>2]|0;
       $77 = ((($2)) + 4|0);
       $78 = ((($2)) + 8|0);
       $$sroa$077$0 = $74;
       while(1) {
        $79 = ($$sroa$077$0|0)==($76|0);
        if ($79) {
         $$1 = $54;
         break L1;
        }
        $80 = HEAP32[$77>>2]|0;
        $81 = HEAP32[$78>>2]|0;
        $82 = ($80|0)==($81|0);
        if ($82) {
         __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($2,$$sroa$077$0);
        } else {
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($80,$$sroa$077$0);
         $83 = HEAP32[$77>>2]|0;
         $84 = ((($83)) + 24|0);
         HEAP32[$77>>2] = $84;
        }
        $85 = ((($$sroa$077$0)) + 24|0);
        $$sroa$077$0 = $85;
       }
      }
      ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
      $86 = $54;
      $87 = (($86) - ($8))|0;
      $88 = ($87>>>0)>(4294967279);
      if ($88) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($6);
       // unreachable;
      }
      $89 = ($87>>>0)<(11);
      if ($89) {
       $90 = $87&255;
       $91 = ((($6)) + 11|0);
       HEAP8[$91>>0] = $90;
       $$021$ph$i$i = $6;
      } else {
       $92 = (($87) + 16)|0;
       $93 = $92 & -16;
       $94 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($93)|0);
       HEAP32[$6>>2] = $94;
       $95 = $93 | -2147483648;
       $96 = ((($6)) + 8|0);
       HEAP32[$96>>2] = $95;
       $97 = ((($6)) + 4|0);
       HEAP32[$97>>2] = $87;
       $$021$ph$i$i = $94;
      }
      $$0$i$i = $0;$$021$i$i = $$021$ph$i$i;
      while(1) {
       $98 = ($$0$i$i|0)==($54|0);
       if ($98) {
        break;
       }
       __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
       $99 = ((($$0$i$i)) + 1|0);
       $100 = ((($$021$i$i)) + 1|0);
       $$0$i$i = $99;$$021$i$i = $100;
      }
      $scevgep23$i$i = (($$021$ph$i$i) + ($87)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$3);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($5,$6);
      $101 = ((($2)) + 4|0);
      $102 = HEAP32[$101>>2]|0;
      $103 = ((($2)) + 8|0);
      $104 = HEAP32[$103>>2]|0;
      $105 = ($102>>>0)<($104>>>0);
      if ($105) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($102,$5);
       $106 = HEAP32[$101>>2]|0;
       $107 = ((($106)) + 24|0);
       HEAP32[$101>>2] = $107;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($5|0));
       $108 = __THREW__; __THREW__ = 0;
       $109 = $108&1;
       if ($109) {
        $110 = ___cxa_find_matching_catch_2()|0;
        $111 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        $$069 = $111;$$070 = $110;
        break;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
      HEAP8[$16>>0] = 1;
      $$1 = $54;
      break L1;
     }
    } while(0);
    ___resumeException($$070|0);
    // unreachable;
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i97 = 0, $$067 = 0, $$069 = 0, $$074 = 0, $$078 = 0, $$083 = 0, $$3 = 0, $$4 = 0, $$482 = 0, $$pre$i = 0, $$pre$i92 = 0, $$pre$phi133Z2D = 0, $$pre$phi135Z2D = 0, $$pre132 = 0, $$pre134 = 0, $$sroa$0$0$$sroa_idx$i85 = 0, $$sroa$0$0$copyload$i100 = 0, $$sroa$0$0$copyload$i86101 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $3 = sp + 72|0;
 $4 = sp + 76|0;
 $5 = sp + 60|0;
 $6 = sp + 56|0;
 $7 = sp + 52|0;
 $8 = sp + 40|0;
 $9 = sp + 16|0;
 $10 = sp;
 $11 = $1;
 $12 = $0;
 $13 = (($11) - ($12))|0;
 $14 = ($13|0)>(1);
 L1: do {
  if ($14) {
   $15 = HEAP8[$0>>0]|0;
   $16 = ($15<<24>>24)==(73);
   if ($16) {
    $17 = ((($2)) + 61|0);
    $18 = HEAP8[$17>>0]|0;
    $19 = ($18<<24>>24)==(0);
    $20 = ((($2)) + 32|0);
    L4: do {
     if (!($19)) {
      $21 = ((($2)) + 36|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ((($22)) + -16|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ((($22)) + -12|0);
      while(1) {
       $26 = HEAP32[$25>>2]|0;
       $27 = ($26|0)==($24|0);
       if ($27) {
        break L4;
       }
       $28 = ((($26)) + -16|0);
       HEAP32[$25>>2] = $28;
       __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($28);
      }
     }
    } while(0);
    $29 = ((($0)) + 1|0);
    ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
    $30 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159117)|0);
    $31 = ($30>>>0)>(4294967279);
    if ($31) {
     __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($5);
     // unreachable;
    }
    $32 = ($30>>>0)<(11);
    if ($32) {
     $33 = $30&255;
     $34 = ((($5)) + 11|0);
     HEAP8[$34>>0] = $33;
     $$pre134 = ((($5)) + 4|0);
     $$0$i$i = $5;$$pre$phi133Z2D = $34;$$pre$phi135Z2D = $$pre134;
    } else {
     $35 = (($30) + 16)|0;
     $36 = $35 & -16;
     $37 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($36)|0);
     HEAP32[$5>>2] = $37;
     $38 = $36 | -2147483648;
     $39 = ((($5)) + 8|0);
     HEAP32[$39>>2] = $38;
     $40 = ((($5)) + 4|0);
     HEAP32[$40>>2] = $30;
     $$pre132 = ((($5)) + 11|0);
     $$0$i$i = $37;$$pre$phi133Z2D = $$pre132;$$pre$phi135Z2D = $40;
    }
    (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,159117,$30)|0);
    $41 = (($$0$i$i) + ($30)|0);
    HEAP8[$4>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($41,$4);
    $42 = ((($2)) + 4|0);
    $43 = ((($2)) + 36|0);
    $$sroa$0$0$$sroa_idx$i85 = ((($2)) + 12|0);
    $44 = ((($8)) + 11|0);
    $45 = ((($8)) + 4|0);
    $46 = ((($2)) + 40|0);
    $$083 = $29;
    L17: while(1) {
     $47 = HEAP8[$$083>>0]|0;
     $48 = ($47<<24>>24)==(69);
     if ($48) {
      label = 55;
      break;
     }
     $49 = HEAP8[$17>>0]|0;
     $50 = ($49<<24>>24)==(0);
     if (!($50)) {
      $$sroa$0$0$copyload$i100 = HEAP32[$$sroa$0$0$$sroa_idx$i85>>2]|0;
      HEAP32[$6>>2] = $$sroa$0$0$copyload$i100;
      $51 = HEAP32[$43>>2]|0;
      $52 = HEAP32[$46>>2]|0;
      $53 = ($51>>>0)<($52>>>0);
      if ($53) {
       __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2IS4_EERKNS1_IT_Lj4096EEE($3,$6);
       HEAP32[$51>>2] = 0;
       $54 = ((($51)) + 4|0);
       HEAP32[$54>>2] = 0;
       $55 = ((($51)) + 8|0);
       $56 = HEAP32[$3>>2]|0;
       HEAP32[$55>>2] = 0;
       $57 = ((($51)) + 12|0);
       HEAP32[$57>>2] = $56;
       $58 = HEAP32[$43>>2]|0;
       $59 = ((($58)) + 16|0);
       HEAP32[$43>>2] = $59;
      } else {
       __THREW__ = 0;
       invoke_vii(1133,($20|0),($6|0));
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        label = 19;
        break;
       }
      }
     }
     $64 = HEAP32[$42>>2]|0;
     $65 = HEAP32[$2>>2]|0;
     $66 = (($64) - ($65))|0;
     $67 = (($66|0) / 24)&-1;
     __THREW__ = 0;
     $68 = (invoke_iiii(1134,($$083|0),($1|0),($2|0))|0);
     $69 = __THREW__; __THREW__ = 0;
     $70 = $69&1;
     if ($70) {
      label = 25;
      break;
     }
     $71 = HEAP32[$42>>2]|0;
     $72 = HEAP32[$2>>2]|0;
     $73 = (($71) - ($72))|0;
     $74 = (($73|0) / 24)&-1;
     $75 = HEAP8[$17>>0]|0;
     $76 = ($75<<24>>24)==(0);
     L28: do {
      if (!($76)) {
       $77 = HEAP32[$43>>2]|0;
       $78 = ((($77)) + -16|0);
       $80 = $77;
       while(1) {
        $79 = ($80|0)==($78|0);
        if ($79) {
         break L28;
        }
        $81 = ((($80)) + -16|0);
        HEAP32[$43>>2] = $81;
        __ZNSt3__213__vector_baseINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEED2Ev($81);
        $$pre$i = HEAP32[$43>>2]|0;
        $80 = $$pre$i;
       }
      }
     } while(0);
     $84 = ($68|0)==($$083|0);
     $85 = ($68|0)==($1|0);
     $or$cond = $84 | $85;
     if ($or$cond) {
      label = 66;
      break;
     }
     $86 = HEAP8[$17>>0]|0;
     $87 = ($86<<24>>24)==(0);
     L35: do {
      if ($87) {
       $$0 = $67;
      } else {
       $88 = HEAP32[$43>>2]|0;
       $$sroa$0$0$copyload$i86101 = HEAP32[$$sroa$0$0$$sroa_idx$i85>>2]|0;
       HEAP32[$7>>2] = $$sroa$0$0$copyload$i86101;
       $89 = ((($88)) + -12|0);
       $90 = HEAP32[$89>>2]|0;
       $91 = ((($88)) + -8|0);
       $92 = HEAP32[$91>>2]|0;
       $93 = ($90>>>0)<($92>>>0);
       if ($93) {
        HEAP32[$90>>2] = 0;
        $94 = ((($90)) + 4|0);
        HEAP32[$94>>2] = 0;
        $95 = ((($90)) + 8|0);
        HEAP32[$95>>2] = 0;
        $96 = ((($90)) + 12|0);
        HEAP32[$96>>2] = $$sroa$0$0$copyload$i86101;
        $97 = HEAP32[$89>>2]|0;
        $98 = ((($97)) + 16|0);
        HEAP32[$89>>2] = $98;
       } else {
        $99 = ((($88)) + -16|0);
        __THREW__ = 0;
        invoke_vii(1135,($99|0),($7|0));
        $100 = __THREW__; __THREW__ = 0;
        $101 = $100&1;
        if ($101) {
         label = 33;
         break L17;
        }
       }
       $$067 = $67;
       while(1) {
        $102 = ($$067>>>0)<($74>>>0);
        if (!($102)) {
         $$0 = $67;
         break L35;
        }
        $105 = HEAP32[$43>>2]|0;
        $106 = ((($105)) + -12|0);
        $107 = HEAP32[$106>>2]|0;
        $108 = HEAP32[$2>>2]|0;
        $109 = (($108) + (($$067*24)|0)|0);
        $110 = ((($107)) + -12|0);
        $111 = HEAP32[$110>>2]|0;
        $112 = ((($107)) + -8|0);
        $113 = HEAP32[$112>>2]|0;
        $114 = ($111|0)==($113|0);
        if ($114) {
         $119 = ((($107)) + -16|0);
         __THREW__ = 0;
         invoke_vii(1136,($119|0),($109|0));
         $120 = __THREW__; __THREW__ = 0;
         $121 = $120&1;
         if ($121) {
          label = 39;
          break L17;
         }
        } else {
         __THREW__ = 0;
         invoke_vii(1130,($111|0),($109|0));
         $115 = __THREW__; __THREW__ = 0;
         $116 = $115&1;
         if ($116) {
          label = 39;
          break L17;
         }
         $117 = HEAP32[$110>>2]|0;
         $118 = ((($117)) + 24|0);
         HEAP32[$110>>2] = $118;
        }
        $122 = (($$067) + 1)|0;
        $$067 = $122;
       }
      }
     } while(0);
     while(1) {
      $125 = ($$0>>>0)<($74>>>0);
      if (!($125)) {
       $$069 = $74;
       break;
      }
      $126 = HEAP8[$$pre$phi133Z2D>>0]|0;
      $127 = ($126<<24>>24)<(0);
      $128 = HEAP32[$$pre$phi135Z2D>>2]|0;
      $129 = $126&255;
      $130 = $127 ? $128 : $129;
      $131 = ($130>>>0)>(1);
      if ($131) {
       __THREW__ = 0;
       (invoke_iii(1119,($5|0),(159305|0))|0);
       $132 = __THREW__; __THREW__ = 0;
       $133 = $132&1;
       if ($133) {
        label = 43;
        break L17;
       }
      }
      $136 = HEAP32[$2>>2]|0;
      $137 = (($136) + (($$0*24)|0)|0);
      __THREW__ = 0;
      invoke_vii(1120,($8|0),($137|0));
      $138 = __THREW__; __THREW__ = 0;
      $139 = $138&1;
      if ($139) {
       label = 47;
       break L17;
      }
      $140 = HEAP8[$44>>0]|0;
      $141 = ($140<<24>>24)<(0);
      $142 = HEAP32[$8>>2]|0;
      $143 = $141 ? $142 : $8;
      $144 = HEAP32[$45>>2]|0;
      $145 = $140&255;
      $146 = $141 ? $144 : $145;
      __THREW__ = 0;
      (invoke_iiii(1112,($5|0),($143|0),($146|0))|0);
      $147 = __THREW__; __THREW__ = 0;
      $148 = $147&1;
      if ($148) {
       label = 48;
       break L17;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
      $149 = (($$0) + 1)|0;
      $$0 = $149;
     }
     while(1) {
      $154 = ($$069|0)==($67|0);
      if ($154) {
       $$083 = $68;
       continue L17;
      }
      $155 = HEAP32[$42>>2]|0;
      $156 = ((($155)) + -24|0);
      $158 = $155;
      while(1) {
       $157 = ($158|0)==($156|0);
       if ($157) {
        break;
       }
       $159 = ((($158)) + -24|0);
       HEAP32[$42>>2] = $159;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($159);
       $$pre$i92 = HEAP32[$42>>2]|0;
       $158 = $$pre$i92;
      }
      $160 = (($$069) + -1)|0;
      $$069 = $160;
     }
    }
    do {
     if ((label|0) == 19) {
      $62 = ___cxa_find_matching_catch_2()|0;
      $63 = tempRet0;
      $$4 = $63;$$482 = $62;
     }
     else if ((label|0) == 25) {
      $82 = ___cxa_find_matching_catch_2()|0;
      $83 = tempRet0;
      $$4 = $83;$$482 = $82;
     }
     else if ((label|0) == 33) {
      $103 = ___cxa_find_matching_catch_2()|0;
      $104 = tempRet0;
      $$4 = $104;$$482 = $103;
     }
     else if ((label|0) == 39) {
      $123 = ___cxa_find_matching_catch_2()|0;
      $124 = tempRet0;
      $$4 = $124;$$482 = $123;
     }
     else if ((label|0) == 43) {
      $134 = ___cxa_find_matching_catch_2()|0;
      $135 = tempRet0;
      $$4 = $135;$$482 = $134;
     }
     else if ((label|0) == 47) {
      $150 = ___cxa_find_matching_catch_2()|0;
      $151 = tempRet0;
      $$074 = $151;$$078 = $150;
      label = 49;
     }
     else if ((label|0) == 48) {
      $152 = ___cxa_find_matching_catch_2()|0;
      $153 = tempRet0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
      $$074 = $153;$$078 = $152;
      label = 49;
     }
     else if ((label|0) == 55) {
      $161 = ((($$083)) + 1|0);
      $162 = HEAP8[$$pre$phi133Z2D>>0]|0;
      $163 = ($162<<24>>24)<(0);
      $164 = HEAP32[$5>>2]|0;
      $165 = HEAP32[$$pre$phi135Z2D>>2]|0;
      $166 = $162&255;
      $167 = $163 ? $164 : $5;
      $168 = $163 ? $165 : $166;
      $169 = (($167) + ($168)|0);
      $170 = ((($169)) + -1|0);
      $171 = HEAP8[$170>>0]|0;
      $172 = ($171<<24>>24)==(62);
      if ($172) {
       __THREW__ = 0;
       (invoke_iii(1119,($5|0),(160496|0))|0);
       $177 = __THREW__; __THREW__ = 0;
       $178 = $177&1;
       if ($178) {
        label = 57;
       }
      } else {
       __THREW__ = 0;
       (invoke_iii(1119,($5|0),(159119|0))|0);
       $173 = __THREW__; __THREW__ = 0;
       $174 = $173&1;
       if ($174) {
        label = 57;
       }
      }
      if ((label|0) == 57) {
       $175 = ___cxa_find_matching_catch_2()|0;
       $176 = tempRet0;
       $$4 = $176;$$482 = $175;
       break;
      }
      ;HEAP32[$10>>2]=HEAP32[$5>>2]|0;HEAP32[$10+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$5+8>>2]|0;
      $$0$i$i97 = 0;
      while(1) {
       $exitcond$i$i = ($$0$i$i97|0)==(3);
       if ($exitcond$i$i) {
        break;
       }
       $179 = (($5) + ($$0$i$i97<<2)|0);
       HEAP32[$179>>2] = 0;
       $180 = (($$0$i$i97) + 1)|0;
       $$0$i$i97 = $180;
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($9,$10);
      $181 = HEAP32[$42>>2]|0;
      $182 = ((($2)) + 8|0);
      $183 = HEAP32[$182>>2]|0;
      $184 = ($181>>>0)<($183>>>0);
      if ($184) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($181,$9);
       $185 = HEAP32[$42>>2]|0;
       $186 = ((($185)) + 24|0);
       HEAP32[$42>>2] = $186;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($9|0));
       $187 = __THREW__; __THREW__ = 0;
       $188 = $187&1;
       if ($188) {
        $189 = ___cxa_find_matching_catch_2()|0;
        $190 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
        $$4 = $190;$$482 = $189;
        break;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
      $$3 = $161;
      break L1;
     }
     else if ((label|0) == 66) {
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
      $$3 = $0;
      break L1;
     }
    } while(0);
    if ((label|0) == 49) {
     $$4 = $$074;$$482 = $$078;
    }
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
    ___resumeException($$482|0);
    // unreachable;
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ((($1)) + 12|0);
 $3 = ((($2)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 $6 = HEAP32[$2>>2]|0;
 $7 = $5 ? $6 : $2;
 $8 = ((($1)) + 16|0);
 $9 = HEAP32[$8>>2]|0;
 $10 = $4&255;
 $11 = $5 ? $9 : $10;
 $12 = (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj($1,$7,$11)|0);
 ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 $$0$i$i$i = 0;
 while(1) {
  $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
  if ($exitcond$i$i$i) {
   break;
  }
  $13 = (($1) + ($$0$i$i$i<<2)|0);
  HEAP32[$13>>2] = 0;
  $14 = (($$0$i$i$i) + 1)|0;
  $$0$i$i$i = $14;
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i = 0, $$0$i$i = 0, $$0$i$i49 = 0, $$034 = 0, $$038 = 0, $$040 = 0, $$040$in = 0, $$1 = 0, $$135 = 0, $$3 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $9 = 0, $exitcond$i$i = 0, $isdigit = 0, $isdigit48 = 0, $isdigittmp = 0, $isdigittmp47 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $3 = sp + 88|0;
 $4 = sp + 76|0;
 $5 = sp + 64|0;
 $6 = sp + 40|0;
 $7 = sp + 16|0;
 $8 = sp;
 $9 = ($0|0)==($1|0);
 L1: do {
  if ($9) {
   $$3 = $0;
  } else {
   $10 = HEAP8[$0>>0]|0;
   $11 = $10 << 24 >> 24;
   $isdigittmp = (($11) + -48)|0;
   $isdigit = ($isdigittmp>>>0)<(10);
   if ($isdigit) {
    $12 = ((($0)) + 1|0);
    $13 = ($12|0)==($1|0);
    if ($13) {
     $$3 = $0;
    } else {
     $$038 = $isdigittmp;$$040$in = $12;
     while(1) {
      $$040 = HEAP8[$$040$in>>0]|0;
      $14 = $$040 << 24 >> 24;
      $isdigittmp47 = (($14) + -48)|0;
      $isdigit48 = ($isdigittmp47>>>0)<(10);
      if (!($isdigit48)) {
       break;
      }
      $15 = ((($$040$in)) + 1|0);
      $16 = ($15|0)==($1|0);
      if ($16) {
       $$3 = $0;
       break L1;
      }
      $17 = ($$038*10)|0;
      $18 = (($isdigittmp47) + ($17))|0;
      $$038 = $18;$$040$in = $15;
     }
     $19 = $1;
     $20 = $$040$in;
     $21 = (($19) - ($20))|0;
     $22 = ($21>>>0)<($$038>>>0);
     if ($22) {
      $$3 = $0;
     } else {
      ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
      $23 = ($$038>>>0)>(4294967279);
      if ($23) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($4);
       // unreachable;
      }
      $24 = ($$038>>>0)<(11);
      if ($24) {
       $25 = $$038&255;
       $26 = ((($4)) + 11|0);
       HEAP8[$26>>0] = $25;
       $$0$i$i = $4;
      } else {
       $27 = (($$038) + 16)|0;
       $28 = $27 & -16;
       $29 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($28)|0);
       HEAP32[$4>>2] = $29;
       $30 = $28 | -2147483648;
       $31 = ((($4)) + 8|0);
       HEAP32[$31>>2] = $30;
       $32 = ((($4)) + 4|0);
       HEAP32[$32>>2] = $$038;
       $$0$i$i = $29;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,$$040$in,$$038)|0);
      $33 = (($$0$i$i) + ($$038)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($33,$3);
      __THREW__ = 0;
      invoke_viii(1125,($5|0),($4|0),10);
      $34 = __THREW__; __THREW__ = 0;
      $35 = $34&1;
      L17: do {
       if ($35) {
        $62 = ___cxa_find_matching_catch_2()|0;
        $63 = tempRet0;
        $$1 = $63;$$135 = $62;
       } else {
        $36 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160463)|0);
        $37 = ((($5)) + 11|0);
        $38 = HEAP8[$37>>0]|0;
        $39 = ($38<<24>>24)<(0);
        $40 = ((($5)) + 4|0);
        $41 = HEAP32[$40>>2]|0;
        $42 = $38&255;
        $43 = $39 ? $41 : $42;
        $44 = ($36|0)==($43|0);
        do {
         if ($44) {
          __THREW__ = 0;
          $45 = (invoke_iiii(1123,($5|0),(160463|0),($36|0))|0);
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           $49 = ___cxa_find_matching_catch_3(0|0)|0;
           $50 = tempRet0;
           ___clang_call_terminate($49);
           // unreachable;
          } else {
           $48 = ($45|0)==(0);
           $$0$i = $48;
           break;
          }
         } else {
          $$0$i = 0;
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
        L26: do {
         if ($$0$i) {
          __THREW__ = 0;
          invoke_vi(1137,($6|0));
          $51 = __THREW__; __THREW__ = 0;
          $52 = $51&1;
          do {
           if ($52) {
            $64 = ___cxa_find_matching_catch_2()|0;
            $65 = tempRet0;
            $$0 = $65;$$034 = $64;
           } else {
            $53 = ((($2)) + 4|0);
            $54 = HEAP32[$53>>2]|0;
            $55 = ((($2)) + 8|0);
            $56 = HEAP32[$55>>2]|0;
            $57 = ($54>>>0)<($56>>>0);
            if ($57) {
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($54,$6);
             $58 = HEAP32[$53>>2]|0;
             $59 = ((($58)) + 24|0);
             HEAP32[$53>>2] = $59;
            } else {
             __THREW__ = 0;
             invoke_vii(1128,($2|0),($6|0));
             $60 = __THREW__; __THREW__ = 0;
             $61 = $60&1;
             if ($61) {
              $66 = ___cxa_find_matching_catch_2()|0;
              $67 = tempRet0;
              __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
              $$0 = $67;$$034 = $66;
              break;
             }
            }
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
            break L26;
           }
          } while(0);
          $$1 = $$0;$$135 = $$034;
          break L17;
         } else {
          ;HEAP32[$8>>2]=HEAP32[$4>>2]|0;HEAP32[$8+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$4+8>>2]|0;
          $$0$i$i49 = 0;
          while(1) {
           $exitcond$i$i = ($$0$i$i49|0)==(3);
           if ($exitcond$i$i) {
            break;
           }
           $68 = (($4) + ($$0$i$i49<<2)|0);
           HEAP32[$68>>2] = 0;
           $69 = (($$0$i$i49) + 1)|0;
           $$0$i$i49 = $69;
          }
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($7,$8);
          $70 = ((($2)) + 4|0);
          $71 = HEAP32[$70>>2]|0;
          $72 = ((($2)) + 8|0);
          $73 = HEAP32[$72>>2]|0;
          $74 = ($71>>>0)<($73>>>0);
          if ($74) {
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($71,$7);
           $75 = HEAP32[$70>>2]|0;
           $76 = ((($75)) + 24|0);
           HEAP32[$70>>2] = $76;
          } else {
           __THREW__ = 0;
           invoke_vii(1128,($2|0),($7|0));
           $77 = __THREW__; __THREW__ = 0;
           $78 = $77&1;
           if ($78) {
            $79 = ___cxa_find_matching_catch_2()|0;
            $80 = tempRet0;
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            $$1 = $80;$$135 = $79;
            break L17;
           }
          }
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         }
        } while(0);
        $81 = (($$040$in) + ($$038)|0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
        $$3 = $81;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      ___resumeException($$135|0);
      // unreachable;
     }
    }
   } else {
    $$3 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EERKSB_PKS8_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i = 0, $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $4 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$4>>2] = 0;
  $5 = (($$0$i$i) + 1)|0;
  $$0$i$i = $5;
 }
 $6 = ((($1)) + 11|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)<(0);
 $9 = ((($1)) + 4|0);
 $10 = HEAP32[$9>>2]|0;
 $11 = $7&255;
 $12 = $8 ? $10 : $11;
 $13 = (__ZNSt3__211char_traitsIcE6lengthEPKc($2)|0);
 $14 = HEAP8[$6>>0]|0;
 $15 = ($14<<24>>24)<(0);
 $16 = HEAP32[$1>>2]|0;
 $17 = $15 ? $16 : $1;
 $18 = (($12) + ($13))|0;
 $19 = ($18>>>0)>(4294967279);
 if ($19) {
  __THREW__ = 0;
  invoke_vi(987,($0|0));
  $20 = __THREW__; __THREW__ = 0;
 } else {
  $21 = ($18>>>0)<(11);
  if ($21) {
   $22 = $12&255;
   $23 = ((($0)) + 11|0);
   HEAP8[$23>>0] = $22;
   $$0$i = $0;
  } else {
   $24 = (($18) + 16)|0;
   $25 = $24 & -16;
   $26 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($25)|0);
   HEAP32[$0>>2] = $26;
   $27 = $25 | -2147483648;
   $28 = ((($0)) + 8|0);
   HEAP32[$28>>2] = $27;
   $29 = ((($0)) + 4|0);
   HEAP32[$29>>2] = $12;
   $$0$i = $26;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i,$17,$12)|0);
  $30 = (($$0$i) + ($12)|0);
  HEAP8[$3>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($30,$3);
  __THREW__ = 0;
  (invoke_iiii(1112,($0|0),($2|0),($13|0))|0);
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if (!($32)) {
   STACKTOP = sp;return;
  }
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($0);
 ___resumeException($33|0);
 // unreachable;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $$0$i$i1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond$i$i = 0, $exitcond$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 $$0$i$i1 = 0;
 while(1) {
  $exitcond$i$i2 = ($$0$i$i1|0)==(3);
  if ($exitcond$i$i2) {
   break;
  }
  $2 = (($1) + ($$0$i$i1<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i1) + 1)|0;
  $$0$i$i1 = $3;
 }
 $4 = ((($0)) + 12|0);
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($$0$i$i) + 1)|0;
  $$0$i$i = $6;
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i$i$i = 0, $$0$i$i$i$i3 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 3|0;
 $3 = sp + 2|0;
 $4 = sp + 1|0;
 $5 = sp;
 $6 = ((($0)) + 11|0);
 $7 = HEAP8[$6>>0]|0;
 $8 = ($7<<24>>24)<(0);
 if ($8) {
  $9 = HEAP32[$0>>2]|0;
  HEAP8[$4>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($9,$4);
  $10 = ((($0)) + 4|0);
  HEAP32[$10>>2] = 0;
 } else {
  HEAP8[$5>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($0,$5);
  HEAP8[$6>>0] = 0;
 }
 __THREW__ = 0;
 invoke_vi(1116,($0|0));
 $11 = __THREW__; __THREW__ = 0;
 $12 = $11&1;
 if ($12) {
  $13 = ___cxa_find_matching_catch_3(0|0)|0;
  $14 = tempRet0;
  ___clang_call_terminate($13);
  // unreachable;
 }
 ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 $$0$i$i$i$i = 0;
 while(1) {
  $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
  if ($exitcond$i$i$i$i) {
   break;
  }
  $15 = (($1) + ($$0$i$i$i$i<<2)|0);
  HEAP32[$15>>2] = 0;
  $16 = (($$0$i$i$i$i) + 1)|0;
  $$0$i$i$i$i = $16;
 }
 $17 = ((($0)) + 12|0);
 $18 = ((($1)) + 12|0);
 $19 = ((($17)) + 11|0);
 $20 = HEAP8[$19>>0]|0;
 $21 = ($20<<24>>24)<(0);
 if ($21) {
  $22 = HEAP32[$17>>2]|0;
  HEAP8[$2>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($22,$2);
  $23 = ((($0)) + 16|0);
  HEAP32[$23>>2] = 0;
 } else {
  HEAP8[$3>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($17,$3);
  HEAP8[$19>>0] = 0;
 }
 __THREW__ = 0;
 invoke_vi(1116,($17|0));
 $24 = __THREW__; __THREW__ = 0;
 $25 = $24&1;
 if ($25) {
  $26 = ___cxa_find_matching_catch_3(0|0)|0;
  $27 = tempRet0;
  ___clang_call_terminate($26);
  // unreachable;
 }
 ;HEAP32[$17>>2]=HEAP32[$18>>2]|0;HEAP32[$17+4>>2]=HEAP32[$18+4>>2]|0;HEAP32[$17+8>>2]=HEAP32[$18+8>>2]|0;
 $$0$i$i$i$i3 = 0;
 while(1) {
  $exitcond$i$i$i$i4 = ($$0$i$i$i$i3|0)==(3);
  if ($exitcond$i$i$i$i4) {
   break;
  }
  $28 = (($18) + ($$0$i$i$i$i3<<2)|0);
  HEAP32[$28>>2] = 0;
  $29 = (($$0$i$i$i$i3) + 1)|0;
  $$0$i$i$i$i3 = $29;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $$0$i$i3 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, $exitcond$i$i4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $2 = (($1) + ($$0$i$i<<2)|0);
  HEAP32[$2>>2] = 0;
  $3 = (($$0$i$i) + 1)|0;
  $$0$i$i = $3;
 }
 $4 = ((($0)) + 12|0);
 $5 = ((($1)) + 12|0);
 ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
 $$0$i$i3 = 0;
 while(1) {
  $exitcond$i$i4 = ($$0$i$i3|0)==(3);
  if ($exitcond$i$i4) {
   break;
  }
  $6 = (($5) + ($$0$i$i3<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i3) + 1)|0;
  $$0$i$i3 = $7;
 }
 return;
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIS3_EEvOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = HEAP32[$0>>2]|0;
 $6 = (($4) - ($5))|0;
 $7 = (($6|0) / 24)&-1;
 $8 = (($7) + 1)|0;
 $9 = (__ZNKSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8max_sizeEv($0)|0);
 $10 = ($9>>>0)<($8>>>0);
 if ($10) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $11 = ((($0)) + 12|0);
  $12 = ((($0)) + 8|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = HEAP32[$0>>2]|0;
  $15 = (($13) - ($14))|0;
  $16 = (($15|0) / 24)&-1;
  $17 = $9 >>> 1;
  $18 = ($16>>>0)<($17>>>0);
  $19 = $16 << 1;
  $20 = ($19>>>0)<($8>>>0);
  $$sroa$speculated$i = $20 ? $8 : $19;
  $$sroa$speculated$$i = $18 ? $$sroa$speculated$i : $9;
  $21 = HEAP32[$3>>2]|0;
  $22 = (($21) - ($14))|0;
  $23 = (($22|0) / 24)&-1;
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEEC2EjjS6_($2,$$sroa$speculated$$i,$23,$11);
  $24 = ((($2)) + 8|0);
  $25 = HEAP32[$24>>2]|0;
  __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($25,$1);
  $26 = ((($25)) + 24|0);
  HEAP32[$24>>2] = $26;
  __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($0,$2);
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEED2Ev($2);
  STACKTOP = sp;return;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0 = 0, $$078 = 0, $$3 = 0, $$87 = 0, $$88 = 0, $$pre$i = 0, $$pre$i90 = 0, $$sroa$0$0$$sroa_idx$i = 0, $$sroa$0$0$copyload$i96 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 32|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(1);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(76);
   $13 = ((($0)) + 1|0);
   $$ = $12 ? $13 : $0;
   $14 = HEAP8[$$>>0]|0;
   $15 = $14 << 24 >> 24;
   switch ($15|0) {
   case 78:  {
    $16 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_nested_nameINS0_2DbEEEPKcS4_S4_RT_($$,$1,$2)|0);
    $17 = ($16|0)==($$|0);
    $$87 = $17 ? $0 : $16;
    $$3 = $$87;
    break L1;
    break;
   }
   case 90:  {
    $18 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_local_nameINS0_2DbEEEPKcS4_S4_RT_($$,$1,$2)|0);
    $19 = ($18|0)==($$|0);
    $$88 = $19 ? $0 : $18;
    $$3 = $$88;
    break L1;
    break;
   }
   default: {
    $20 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_unscoped_nameINS0_2DbEEEPKcS4_S4_RT_($$,$1,$2)|0);
    $21 = ($20|0)==($$|0);
    do {
     if ($21) {
      $81 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_substitutionINS0_2DbEEEPKcS4_S4_RT_($$,$1,$2)|0);
      $82 = ($81|0)==($$|0);
      $83 = ($81|0)==($1|0);
      $or$cond = $82 | $83;
      if ($or$cond) {
       $$3 = $0;
       break L1;
      }
      $84 = HEAP8[$81>>0]|0;
      $85 = ($84<<24>>24)==(73);
      if (!($85)) {
       $$3 = $0;
       break L1;
      }
      $86 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($81,$1,$2)|0);
      $87 = ($86|0)==($81|0);
      if ($87) {
       $$3 = $0;
       break L1;
      }
      $88 = ((($2)) + 4|0);
      $89 = HEAP32[$88>>2]|0;
      $90 = HEAP32[$2>>2]|0;
      $91 = (($89) - ($90))|0;
      $92 = (($91|0) / 24)&-1;
      $93 = ($92>>>0)<(2);
      if ($93) {
       $$3 = $0;
       break L1;
      }
      $94 = $89;
      $95 = ((($94)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$95);
      $96 = HEAP32[$88>>2]|0;
      $97 = ((($96)) + -24|0);
      $99 = $96;
      while(1) {
       $98 = ($99|0)==($97|0);
       if ($98) {
        break;
       }
       $100 = ((($99)) + -24|0);
       HEAP32[$88>>2] = $100;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($100);
       $$pre$i90 = HEAP32[$88>>2]|0;
       $99 = $$pre$i90;
      }
      $101 = ((($96)) + -48|0);
      $102 = ((($6)) + 11|0);
      $103 = HEAP8[$102>>0]|0;
      $104 = ($103<<24>>24)<(0);
      $105 = HEAP32[$6>>2]|0;
      $106 = $104 ? $105 : $6;
      $107 = ((($6)) + 4|0);
      $108 = HEAP32[$107>>2]|0;
      $109 = $103&255;
      $110 = $104 ? $108 : $109;
      __THREW__ = 0;
      (invoke_iiii(1112,($101|0),($106|0),($110|0))|0);
      $111 = __THREW__; __THREW__ = 0;
      $112 = $111&1;
      if ($112) {
       $113 = ___cxa_find_matching_catch_2()|0;
       $114 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       $$0 = $114;$$078 = $113;
       break;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       $$3 = $86;
       break L1;
      }
     } else {
      $22 = ($20|0)==($1|0);
      if ($22) {
       $$3 = $1;
       break L1;
      }
      $23 = HEAP8[$20>>0]|0;
      $24 = ($23<<24>>24)==(73);
      if (!($24)) {
       $$3 = $20;
       break L1;
      }
      $25 = HEAP32[$2>>2]|0;
      $26 = ((($2)) + 4|0);
      $27 = HEAP32[$26>>2]|0;
      $28 = ($25|0)==($27|0);
      if ($28) {
       $$3 = $0;
       break L1;
      }
      $29 = ((($27)) + -24|0);
      $$sroa$0$0$$sroa_idx$i = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i96 = HEAP32[$$sroa$0$0$$sroa_idx$i>>2]|0;
      HEAP32[$4>>2] = $$sroa$0$0$copyload$i96;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($3,$29,$4);
      $30 = ((($2)) + 20|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = ((($2)) + 24|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = ($31>>>0)<($33>>>0);
      if ($34) {
       $35 = ((($3)) + 12|0);
       HEAP32[$31>>2] = 0;
       $36 = ((($31)) + 4|0);
       HEAP32[$36>>2] = 0;
       $37 = ((($31)) + 8|0);
       $38 = HEAP32[$35>>2]|0;
       HEAP32[$37>>2] = 0;
       $39 = ((($31)) + 12|0);
       HEAP32[$39>>2] = $38;
       $40 = HEAP32[$3>>2]|0;
       HEAP32[$31>>2] = $40;
       $41 = ((($3)) + 4|0);
       $42 = HEAP32[$41>>2]|0;
       HEAP32[$36>>2] = $42;
       $43 = ((($3)) + 8|0);
       $44 = HEAP32[$43>>2]|0;
       $45 = ((($31)) + 8|0);
       HEAP32[$45>>2] = $44;
       HEAP32[$43>>2] = 0;
       HEAP32[$41>>2] = 0;
       HEAP32[$3>>2] = 0;
       $46 = HEAP32[$30>>2]|0;
       $47 = ((($46)) + 16|0);
       HEAP32[$30>>2] = $47;
      } else {
       $48 = ((($2)) + 16|0);
       __THREW__ = 0;
       invoke_vii(1122,($48|0),($3|0));
       $49 = __THREW__; __THREW__ = 0;
       $50 = $49&1;
       if ($50) {
        $58 = ___cxa_find_matching_catch_2()|0;
        $59 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($3);
        $$0 = $59;$$078 = $58;
        break;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($3);
      $51 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($20,$1,$2)|0);
      $52 = ($51|0)==($20|0);
      if ($52) {
       $$3 = $0;
       break L1;
      }
      $53 = HEAP32[$26>>2]|0;
      $54 = HEAP32[$2>>2]|0;
      $55 = (($53) - ($54))|0;
      $56 = (($55|0) / 24)&-1;
      $57 = ($56>>>0)<(2);
      if ($57) {
       $$3 = $0;
       break L1;
      }
      $60 = $53;
      $61 = ((($60)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($5,$61);
      $62 = HEAP32[$26>>2]|0;
      $63 = ((($62)) + -24|0);
      $65 = $62;
      while(1) {
       $64 = ($65|0)==($63|0);
       if ($64) {
        break;
       }
       $66 = ((($65)) + -24|0);
       HEAP32[$26>>2] = $66;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($66);
       $$pre$i = HEAP32[$26>>2]|0;
       $65 = $$pre$i;
      }
      $67 = ((($62)) + -48|0);
      $68 = ((($5)) + 11|0);
      $69 = HEAP8[$68>>0]|0;
      $70 = ($69<<24>>24)<(0);
      $71 = HEAP32[$5>>2]|0;
      $72 = $70 ? $71 : $5;
      $73 = ((($5)) + 4|0);
      $74 = HEAP32[$73>>2]|0;
      $75 = $69&255;
      $76 = $70 ? $74 : $75;
      __THREW__ = 0;
      (invoke_iiii(1112,($67|0),($72|0),($76|0))|0);
      $77 = __THREW__; __THREW__ = 0;
      $78 = $77&1;
      if ($78) {
       $79 = ___cxa_find_matching_catch_2()|0;
       $80 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
       $$0 = $80;$$078 = $79;
       break;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
       $$3 = $51;
       break L1;
      }
     }
    } while(0);
    ___resumeException($$078|0);
    // unreachable;
   }
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_substitutionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$072 = 0, $$076 = 0, $$077 = 0, $$1 = 0, $$174 = 0, $$275 = 0, $$sink = 0, $$sink84 = 0, $$sroa$0103$0 = 0, $$sroa$097$0 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0;
 var $98 = 0, $99 = 0, $isdigit = 0, $isdigit81 = 0, $isdigittmp = 0, $isdigittmp80 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $3 = sp + 120|0;
 $4 = sp + 96|0;
 $5 = sp + 72|0;
 $6 = sp + 48|0;
 $7 = sp + 24|0;
 $8 = sp;
 $9 = $1;
 $10 = $0;
 $11 = (($9) - ($10))|0;
 $12 = ($11|0)>(1);
 L1: do {
  if ($12) {
   $13 = HEAP8[$0>>0]|0;
   $14 = ($13<<24>>24)==(83);
   if ($14) {
    $15 = ((($0)) + 1|0);
    $16 = HEAP8[$15>>0]|0;
    $17 = $16 << 24 >> 24;
    $18 = ((($0)) + 2|0);
    $19 = ((($2)) + 16|0);
    L4: do {
     switch ($17|0) {
     case 97:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj15EEERAT__Kc($3,160409);
      $20 = ((($2)) + 4|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ((($2)) + 8|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = ($21>>>0)<($23>>>0);
      if ($24) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($21,$3);
       $25 = HEAP32[$20>>2]|0;
       $26 = ((($25)) + 24|0);
       HEAP32[$20>>2] = $26;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($3|0));
       $27 = __THREW__; __THREW__ = 0;
       $28 = $27&1;
       if ($28) {
        $29 = ___cxa_find_matching_catch_2()|0;
        $30 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
        $$076 = $29;$$077 = $30;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
      $$1 = $18;
      break L1;
      break;
     }
     case 98:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj18EEERAT__Kc($4,160424);
      $31 = ((($2)) + 4|0);
      $32 = HEAP32[$31>>2]|0;
      $33 = ((($2)) + 8|0);
      $34 = HEAP32[$33>>2]|0;
      $35 = ($32>>>0)<($34>>>0);
      if ($35) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($32,$4);
       $36 = HEAP32[$31>>2]|0;
       $37 = ((($36)) + 24|0);
       HEAP32[$31>>2] = $37;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($4|0));
       $38 = __THREW__; __THREW__ = 0;
       $39 = $38&1;
       if ($39) {
        $40 = ___cxa_find_matching_catch_2()|0;
        $41 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
        $$076 = $40;$$077 = $41;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
      $$1 = $18;
      break L1;
      break;
     }
     case 115:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($5,159903);
      $42 = ((($2)) + 4|0);
      $43 = HEAP32[$42>>2]|0;
      $44 = ((($2)) + 8|0);
      $45 = HEAP32[$44>>2]|0;
      $46 = ($43>>>0)<($45>>>0);
      if ($46) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($43,$5);
       $47 = HEAP32[$42>>2]|0;
       $48 = ((($47)) + 24|0);
       HEAP32[$42>>2] = $48;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($5|0));
       $49 = __THREW__; __THREW__ = 0;
       $50 = $49&1;
       if ($50) {
        $51 = ___cxa_find_matching_catch_2()|0;
        $52 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
        $$076 = $51;$$077 = $52;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
      $$1 = $18;
      break L1;
      break;
     }
     case 105:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj13EEERAT__Kc($6,159999);
      $53 = ((($2)) + 4|0);
      $54 = HEAP32[$53>>2]|0;
      $55 = ((($2)) + 8|0);
      $56 = HEAP32[$55>>2]|0;
      $57 = ($54>>>0)<($56>>>0);
      if ($57) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($54,$6);
       $58 = HEAP32[$53>>2]|0;
       $59 = ((($58)) + 24|0);
       HEAP32[$53>>2] = $59;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($6|0));
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        $62 = ___cxa_find_matching_catch_2()|0;
        $63 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
        $$076 = $62;$$077 = $63;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
      $$1 = $18;
      break L1;
      break;
     }
     case 111:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj13EEERAT__Kc($7,160076);
      $64 = ((($2)) + 4|0);
      $65 = HEAP32[$64>>2]|0;
      $66 = ((($2)) + 8|0);
      $67 = HEAP32[$66>>2]|0;
      $68 = ($65>>>0)<($67>>>0);
      if ($68) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($65,$7);
       $69 = HEAP32[$64>>2]|0;
       $70 = ((($69)) + 24|0);
       HEAP32[$64>>2] = $70;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($7|0));
       $71 = __THREW__; __THREW__ = 0;
       $72 = $71&1;
       if ($72) {
        $73 = ___cxa_find_matching_catch_2()|0;
        $74 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
        $$076 = $73;$$077 = $74;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
      $$1 = $18;
      break L1;
      break;
     }
     case 100:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj14EEERAT__Kc($8,160153);
      $75 = ((($2)) + 4|0);
      $76 = HEAP32[$75>>2]|0;
      $77 = ((($2)) + 8|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ($76>>>0)<($78>>>0);
      if ($79) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($76,$8);
       $80 = HEAP32[$75>>2]|0;
       $81 = ((($80)) + 24|0);
       HEAP32[$75>>2] = $81;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($8|0));
       $82 = __THREW__; __THREW__ = 0;
       $83 = $82&1;
       if ($83) {
        $84 = ___cxa_find_matching_catch_2()|0;
        $85 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
        $$076 = $84;$$077 = $85;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
      $$1 = $18;
      break L1;
      break;
     }
     case 95:  {
      $86 = HEAP32[$19>>2]|0;
      $87 = ((($2)) + 20|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = ($86|0)==($88|0);
      if ($89) {
       $$1 = $0;
       break L1;
      }
      $90 = HEAP32[$86>>2]|0;
      $91 = ((($86)) + 4|0);
      $92 = HEAP32[$91>>2]|0;
      $93 = ((($2)) + 4|0);
      $94 = ((($2)) + 8|0);
      $$sroa$0103$0 = $90;
      while(1) {
       $95 = ($$sroa$0103$0|0)==($92|0);
       if ($95) {
        $$1 = $18;
        break L1;
       }
       $96 = HEAP32[$93>>2]|0;
       $97 = HEAP32[$94>>2]|0;
       $98 = ($96|0)==($97|0);
       if ($98) {
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($2,$$sroa$0103$0);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($96,$$sroa$0103$0);
        $99 = HEAP32[$93>>2]|0;
        $100 = ((($99)) + 24|0);
        HEAP32[$93>>2] = $100;
       }
       $101 = ((($$sroa$0103$0)) + 24|0);
       $$sroa$0103$0 = $101;
      }
      break;
     }
     default: {
      $isdigittmp = (($17) + -48)|0;
      $isdigit = ($isdigittmp>>>0)<(10);
      if (!($isdigit)) {
       $102 = (_isupper($17)|0);
       $103 = ($102|0)==(0);
       if ($103) {
        $$1 = $0;
        break L1;
       }
      }
      $$sink = $isdigit ? -48 : -55;
      $104 = (($$sink) + ($17))|0;
      $$072 = $18;$$174 = $104;
      while(1) {
       $105 = ($$072|0)==($1|0);
       $106 = ((($$072)) + 1|0);
       if ($105) {
        $$1 = $0;
        break L1;
       }
       $107 = HEAP8[$$072>>0]|0;
       $108 = $107 << 24 >> 24;
       $isdigittmp80 = (($108) + -48)|0;
       $isdigit81 = ($isdigittmp80>>>0)<(10);
       if (!($isdigit81)) {
        $109 = (_isupper($108)|0);
        $110 = ($109|0)==(0);
        if ($110) {
         break;
        }
       }
       $111 = ($$174*36)|0;
       $$sink84 = $isdigit81 ? -48 : -55;
       $112 = (($108) + ($111))|0;
       $$275 = (($112) + ($$sink84))|0;
       $$072 = $106;$$174 = $$275;
      }
      $113 = ($107<<24>>24)==(95);
      if (!($113)) {
       $$1 = $0;
       break L1;
      }
      $114 = (($$174) + 1)|0;
      $115 = ((($2)) + 20|0);
      $116 = HEAP32[$115>>2]|0;
      $117 = HEAP32[$19>>2]|0;
      $118 = (($116) - ($117))|0;
      $119 = $118 >> 4;
      $120 = ($114>>>0)<($119>>>0);
      $121 = $117;
      if (!($120)) {
       $$1 = $0;
       break L1;
      }
      $122 = (($121) + ($114<<4)|0);
      $123 = HEAP32[$122>>2]|0;
      $124 = (((($121) + ($114<<4)|0)) + 4|0);
      $125 = HEAP32[$124>>2]|0;
      $126 = ((($2)) + 4|0);
      $127 = ((($2)) + 8|0);
      $$sroa$097$0 = $123;
      while(1) {
       $128 = ($$sroa$097$0|0)==($125|0);
       if ($128) {
        $$1 = $106;
        break L1;
       }
       $129 = HEAP32[$126>>2]|0;
       $130 = HEAP32[$127>>2]|0;
       $131 = ($129|0)==($130|0);
       if ($131) {
        __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_($2,$$sroa$097$0);
       } else {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_($129,$$sroa$097$0);
        $132 = HEAP32[$126>>2]|0;
        $133 = ((($132)) + 24|0);
        HEAP32[$126>>2] = $133;
       }
       $134 = ((($$sroa$097$0)) + 24|0);
       $$sroa$097$0 = $134;
      }
     }
     }
    } while(0);
    ___resumeException($$076|0);
    // unreachable;
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_114parse_decltypeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i30 = 0, $$1 = 0, $$124 = 0, $$127 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i31 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(3);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(68);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = $14 << 24 >> 24;
    switch ($15|0) {
    case 84: case 116:  {
     break;
    }
    default: {
     $$127 = $0;
     break L1;
    }
    }
    $16 = ((($0)) + 2|0);
    $17 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $18 = ($17|0)==($16|0);
    $19 = ($17|0)==($1|0);
    $or$cond = $18 | $19;
    if ($or$cond) {
     $$127 = $0;
    } else {
     $20 = HEAP8[$17>>0]|0;
     $21 = ($20<<24>>24)==(69);
     if ($21) {
      $22 = HEAP32[$2>>2]|0;
      $23 = ((($2)) + 4|0);
      $24 = HEAP32[$23>>2]|0;
      $25 = ($22|0)==($24|0);
      if ($25) {
       $$127 = $0;
      } else {
       $26 = ((($24)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$26);
       __THREW__ = 0;
       $27 = (invoke_iiii(1121,($6|0),0,(160399|0))|0);
       $28 = __THREW__; __THREW__ = 0;
       $29 = $28&1;
       do {
        if ($29) {
         $40 = ___cxa_find_matching_catch_2()|0;
         $41 = tempRet0;
         $$1 = $41;$$124 = $40;
        } else {
         ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $30 = (($6) + ($$0$i$i$i<<2)|0);
          HEAP32[$30>>2] = 0;
          $31 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $31;
         }
         __THREW__ = 0;
         $32 = (invoke_iii(1119,($5|0),(159081|0))|0);
         $33 = __THREW__; __THREW__ = 0;
         $34 = $33&1;
         if ($34) {
          $42 = ___cxa_find_matching_catch_2()|0;
          $43 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
          $$1 = $43;$$124 = $42;
          break;
         }
         ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
         $$0$i$i$i30 = 0;
         while(1) {
          $exitcond$i$i$i31 = ($$0$i$i$i30|0)==(3);
          if ($exitcond$i$i$i31) {
           break;
          }
          $35 = (($5) + ($$0$i$i$i30<<2)|0);
          HEAP32[$35>>2] = 0;
          $36 = (($$0$i$i$i30) + 1)|0;
          $$0$i$i$i30 = $36;
         }
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
         $37 = HEAP32[$23>>2]|0;
         $38 = ((($37)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($38,$3);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         $39 = ((($17)) + 1|0);
         $$127 = $39;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       ___resumeException($$124|0);
       // unreachable;
      }
     } else {
      $$127 = $0;
     }
    }
   } else {
    $$127 = $0;
   }
  } else {
   $$127 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$127|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_117parse_vector_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i128 = 0, $$0$i$i$i134 = 0, $$0$i$i$i138 = 0, $$0$i$i$i144 = 0, $$0$i$i132 = 0, $$0$i$i143 = 0, $$0104 = 0, $$0113 = 0, $$1105 = 0, $$1114 = 0, $$2103 = 0, $$2106 = 0, $$2115 = 0, $$3107 = 0, $$3116 = 0, $$5109 = 0, $$5118 = 0;
 var $$6110 = 0, $$6119 = 0, $$7111 = 0, $$7120 = 0, $$8112 = 0, $$8121 = 0, $$899$ph = 0, $$9 = 0, $$off = 0, $$pre$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0;
 var $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0;
 var $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0;
 var $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i129 = 0, $exitcond$i$i$i135 = 0, $exitcond$i$i$i139 = 0, $exitcond$i$i$i145 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 160|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(160|0);
 $3 = sp + 151|0;
 $4 = sp + 150|0;
 $5 = sp + 149|0;
 $6 = sp + 148|0;
 $7 = sp + 136|0;
 $8 = sp + 124|0;
 $9 = sp + 112|0;
 $10 = sp + 88|0;
 $11 = sp + 72|0;
 $12 = sp + 60|0;
 $13 = sp + 48|0;
 $14 = sp + 36|0;
 $15 = sp + 24|0;
 $16 = sp + 12|0;
 $17 = sp;
 $18 = $1;
 $19 = $0;
 $20 = (($18) - ($19))|0;
 $21 = ($20|0)>(3);
 L1: do {
  if ($21) {
   $22 = HEAP8[$0>>0]|0;
   $23 = ($22<<24>>24)==(68);
   if ($23) {
    $24 = ((($0)) + 1|0);
    $25 = HEAP8[$24>>0]|0;
    $26 = ($25<<24>>24)==(118);
    if ($26) {
     $27 = ((($0)) + 2|0);
     $28 = HEAP8[$27>>0]|0;
     $$off = (($28) + -49)<<24>>24;
     $29 = ($$off&255)<(9);
     do {
      if ($29) {
       $30 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($27,$1)|0);
       $31 = ($30|0)==($1|0);
       if ($31) {
        $$9 = $0;
        break L1;
       }
       $32 = HEAP8[$30>>0]|0;
       $33 = ($32<<24>>24)==(95);
       if (!($33)) {
        $$9 = $0;
        break L1;
       }
       $34 = $30;
       $35 = $27;
       $36 = (($34) - ($35))|0;
       $37 = ((($30)) + 1|0);
       $38 = ($37|0)==($1|0);
       if ($38) {
        $$9 = $0;
        break L1;
       }
       $39 = HEAP8[$37>>0]|0;
       $40 = ($39<<24>>24)==(112);
       if ($40) {
        $87 = ((($30)) + 2|0);
        ;HEAP32[$13>>2]=0|0;HEAP32[$13+4>>2]=0|0;HEAP32[$13+8>>2]=0|0;
        $88 = ($36>>>0)>(4294967279);
        if ($88) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($13);
         // unreachable;
        }
        $89 = ($36>>>0)<(11);
        if ($89) {
         $90 = $36&255;
         $91 = ((($13)) + 11|0);
         HEAP8[$91>>0] = $90;
         $$0$i$i132 = $13;
        } else {
         $92 = (($36) + 16)|0;
         $93 = $92 & -16;
         $94 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($93)|0);
         HEAP32[$13>>2] = $94;
         $95 = $93 | -2147483648;
         $96 = ((($13)) + 8|0);
         HEAP32[$96>>2] = $95;
         $97 = ((($13)) + 4|0);
         HEAP32[$97>>2] = $36;
         $$0$i$i132 = $94;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i132,$27,$36)|0);
        $98 = (($$0$i$i132) + ($36)|0);
        HEAP8[$5>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($98,$5);
        __THREW__ = 0;
        $99 = (invoke_iiii(1121,($13|0),0,(159132|0))|0);
        $100 = __THREW__; __THREW__ = 0;
        $101 = $100&1;
        if ($101) {
         $118 = ___cxa_find_matching_catch_2()|0;
         $119 = tempRet0;
         $$3107 = $119;$$3116 = $118;
        } else {
         ;HEAP32[$12>>2]=HEAP32[$13>>2]|0;HEAP32[$12+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$13+8>>2]|0;
         $$0$i$i$i134 = 0;
         while(1) {
          $exitcond$i$i$i135 = ($$0$i$i$i134|0)==(3);
          if ($exitcond$i$i$i135) {
           break;
          }
          $102 = (($13) + ($$0$i$i$i134<<2)|0);
          HEAP32[$102>>2] = 0;
          $103 = (($$0$i$i$i134) + 1)|0;
          $$0$i$i$i134 = $103;
         }
         __THREW__ = 0;
         $104 = (invoke_iii(1119,($12|0),(159130|0))|0);
         $105 = __THREW__; __THREW__ = 0;
         $106 = $105&1;
         do {
          if ($106) {
           $120 = ___cxa_find_matching_catch_2()|0;
           $121 = tempRet0;
           $$2106 = $121;$$2115 = $120;
          } else {
           ;HEAP32[$11>>2]=HEAP32[$12>>2]|0;HEAP32[$11+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$12+8>>2]|0;
           $$0$i$i$i138 = 0;
           while(1) {
            $exitcond$i$i$i139 = ($$0$i$i$i138|0)==(3);
            if ($exitcond$i$i$i139) {
             break;
            }
            $107 = (($12) + ($$0$i$i$i138<<2)|0);
            HEAP32[$107>>2] = 0;
            $108 = (($$0$i$i$i138) + 1)|0;
            $$0$i$i$i138 = $108;
           }
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($10,$11);
           $109 = ((($2)) + 4|0);
           $110 = HEAP32[$109>>2]|0;
           $111 = ((($2)) + 8|0);
           $112 = HEAP32[$111>>2]|0;
           $113 = ($110>>>0)<($112>>>0);
           if ($113) {
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($110,$10);
            $114 = HEAP32[$109>>2]|0;
            $115 = ((($114)) + 24|0);
            HEAP32[$109>>2] = $115;
           } else {
            __THREW__ = 0;
            invoke_vii(1128,($2|0),($10|0));
            $116 = __THREW__; __THREW__ = 0;
            $117 = $116&1;
            if ($117) {
             $122 = ___cxa_find_matching_catch_2()|0;
             $123 = tempRet0;
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
             $$2106 = $123;$$2115 = $122;
             break;
            }
           }
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
           $$9 = $87;
           break L1;
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
         $$3107 = $$2106;$$3116 = $$2115;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
        $$8112 = $$3107;$$8121 = $$3116;
        break;
       } else {
        $41 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($37,$1,$2)|0);
        $42 = ($41|0)==($37|0);
        if ($42) {
         $$9 = $0;
         break L1;
        }
        $43 = HEAP32[$2>>2]|0;
        $44 = ((($2)) + 4|0);
        $45 = HEAP32[$44>>2]|0;
        $46 = ($43|0)==($45|0);
        if ($46) {
         $$9 = $0;
         break L1;
        }
        ;HEAP32[$9>>2]=0|0;HEAP32[$9+4>>2]=0|0;HEAP32[$9+8>>2]=0|0;
        $47 = ($36>>>0)>(4294967279);
        if ($47) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($9);
         // unreachable;
        }
        $48 = ($36>>>0)<(11);
        if ($48) {
         $49 = $36&255;
         $50 = ((($9)) + 11|0);
         HEAP8[$50>>0] = $49;
         $$0$i$i = $9;
        } else {
         $51 = (($36) + 16)|0;
         $52 = $51 & -16;
         $53 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($52)|0);
         HEAP32[$9>>2] = $53;
         $54 = $52 | -2147483648;
         $55 = ((($9)) + 8|0);
         HEAP32[$55>>2] = $54;
         $56 = ((($9)) + 4|0);
         HEAP32[$56>>2] = $36;
         $$0$i$i = $53;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,$27,$36)|0);
        $57 = (($$0$i$i) + ($36)|0);
        HEAP8[$6>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($57,$6);
        __THREW__ = 0;
        $58 = (invoke_iiii(1121,($9|0),0,(159121|0))|0);
        $59 = __THREW__; __THREW__ = 0;
        $60 = $59&1;
        if ($60) {
         $81 = ___cxa_find_matching_catch_2()|0;
         $82 = tempRet0;
         $$1105 = $82;$$1114 = $81;
        } else {
         ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $61 = (($9) + ($$0$i$i$i<<2)|0);
          HEAP32[$61>>2] = 0;
          $62 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $62;
         }
         __THREW__ = 0;
         $63 = (invoke_iii(1119,($8|0),(159130|0))|0);
         $64 = __THREW__; __THREW__ = 0;
         $65 = $64&1;
         do {
          if ($65) {
           $83 = ___cxa_find_matching_catch_2()|0;
           $84 = tempRet0;
           $$0104 = $84;$$0113 = $83;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i128 = 0;
           while(1) {
            $exitcond$i$i$i129 = ($$0$i$i$i128|0)==(3);
            if ($exitcond$i$i$i129) {
             break;
            }
            $66 = (($8) + ($$0$i$i$i128<<2)|0);
            HEAP32[$66>>2] = 0;
            $67 = (($$0$i$i$i128) + 1)|0;
            $$0$i$i$i128 = $67;
           }
           $68 = HEAP32[$44>>2]|0;
           $69 = ((($68)) + -24|0);
           $70 = ((($7)) + 11|0);
           $71 = HEAP8[$70>>0]|0;
           $72 = ($71<<24>>24)<(0);
           $73 = HEAP32[$7>>2]|0;
           $74 = $72 ? $73 : $7;
           $75 = ((($7)) + 4|0);
           $76 = HEAP32[$75>>2]|0;
           $77 = $71&255;
           $78 = $72 ? $76 : $77;
           __THREW__ = 0;
           (invoke_iiii(1112,($69|0),($74|0),($78|0))|0);
           $79 = __THREW__; __THREW__ = 0;
           $80 = $79&1;
           if ($80) {
            $85 = ___cxa_find_matching_catch_2()|0;
            $86 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            $$0104 = $86;$$0113 = $85;
            break;
           } else {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
            $$9 = $41;
            break L1;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         $$1105 = $$0104;$$1114 = $$0113;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
        $$8112 = $$1105;$$8121 = $$1114;
        break;
       }
      } else {
       ;HEAP32[$14>>2]=0|0;HEAP32[$14+4>>2]=0|0;HEAP32[$14+8>>2]=0|0;
       $$0$i$i143 = 0;
       while(1) {
        $exitcond$i$i = ($$0$i$i143|0)==(3);
        if ($exitcond$i$i) {
         break;
        }
        $124 = (($14) + ($$0$i$i143<<2)|0);
        HEAP32[$124>>2] = 0;
        $125 = (($$0$i$i143) + 1)|0;
        $$0$i$i143 = $125;
       }
       $126 = ($28<<24>>24)==(95);
       L74: do {
        if ($126) {
         $$2103 = $27;
         label = 74;
        } else {
         __THREW__ = 0;
         $127 = (invoke_iiii(1138,($27|0),($1|0),($2|0))|0);
         $128 = __THREW__; __THREW__ = 0;
         $129 = $128&1;
         if ($129) {
          label = 60;
         } else {
          $130 = ($127|0)==($27|0);
          if ($130) {
           $$2103 = $27;
           label = 74;
          } else {
           $131 = HEAP32[$2>>2]|0;
           $132 = ((($2)) + 4|0);
           $133 = HEAP32[$132>>2]|0;
           $134 = ($131|0)==($133|0);
           if ($134) {
            label = 93;
           } else {
            $137 = ((($133)) + -24|0);
            __THREW__ = 0;
            invoke_vii(1120,($15|0),($137|0));
            $138 = __THREW__; __THREW__ = 0;
            $139 = $138&1;
            if ($139) {
             $156 = ___cxa_find_matching_catch_2()|0;
             $157 = tempRet0;
             $$7111 = $157;$$7120 = $156;
             break;
            }
            $140 = ((($14)) + 11|0);
            $141 = HEAP8[$140>>0]|0;
            $142 = ($141<<24>>24)<(0);
            if ($142) {
             $143 = HEAP32[$14>>2]|0;
             HEAP8[$3>>0] = 0;
             __ZNSt3__211char_traitsIcE6assignERcRKc($143,$3);
             $144 = ((($14)) + 4|0);
             HEAP32[$144>>2] = 0;
            } else {
             HEAP8[$4>>0] = 0;
             __ZNSt3__211char_traitsIcE6assignERcRKc($14,$4);
             HEAP8[$140>>0] = 0;
            }
            __THREW__ = 0;
            invoke_vi(1116,($14|0));
            $145 = __THREW__; __THREW__ = 0;
            $146 = $145&1;
            if ($146) {
             $147 = ___cxa_find_matching_catch_3(0|0)|0;
             $148 = tempRet0;
             ___clang_call_terminate($147);
             // unreachable;
            }
            ;HEAP32[$14>>2]=HEAP32[$15>>2]|0;HEAP32[$14+4>>2]=HEAP32[$15+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$15+8>>2]|0;
            $$0$i$i$i$i = 0;
            while(1) {
             $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
             if ($exitcond$i$i$i$i) {
              break;
             }
             $149 = (($15) + ($$0$i$i$i$i<<2)|0);
             HEAP32[$149>>2] = 0;
             $150 = (($$0$i$i$i$i) + 1)|0;
             $$0$i$i$i$i = $150;
            }
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
            $151 = HEAP32[$132>>2]|0;
            $152 = ((($151)) + -24|0);
            $154 = $151;
            while(1) {
             $153 = ($154|0)==($152|0);
             if ($153) {
              $$2103 = $127;
              label = 74;
              break L74;
             }
             $155 = ((($154)) + -24|0);
             HEAP32[$132>>2] = $155;
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($155);
             $$pre$i = HEAP32[$132>>2]|0;
             $154 = $$pre$i;
            }
           }
          }
         }
        }
       } while(0);
       L96: do {
        if ((label|0) == 74) {
         $158 = ($$2103|0)==($1|0);
         L98: do {
          if ($158) {
           $$899$ph = $0;
          } else {
           $159 = HEAP8[$$2103>>0]|0;
           $160 = ($159<<24>>24)==(95);
           if ($160) {
            $161 = ((($$2103)) + 1|0);
            $162 = ($161|0)==($1|0);
            if ($162) {
             $$899$ph = $0;
            } else {
             __THREW__ = 0;
             $163 = (invoke_iiii(1115,($161|0),($1|0),($2|0))|0);
             $164 = __THREW__; __THREW__ = 0;
             $165 = $164&1;
             if ($165) {
              label = 60;
              break L96;
             }
             $166 = ($163|0)==($161|0);
             if ($166) {
              $$899$ph = $0;
             } else {
              $167 = HEAP32[$2>>2]|0;
              $168 = ((($2)) + 4|0);
              $169 = HEAP32[$168>>2]|0;
              $170 = ($167|0)==($169|0);
              if ($170) {
               label = 93;
               break L96;
              }
              __THREW__ = 0;
              invoke_viii(1139,($17|0),(159121|0),($14|0));
              $171 = __THREW__; __THREW__ = 0;
              $172 = $171&1;
              if ($172) {
               $191 = ___cxa_find_matching_catch_2()|0;
               $192 = tempRet0;
               $$6110 = $192;$$6119 = $191;
              } else {
               __THREW__ = 0;
               $173 = (invoke_iii(1119,($17|0),(159130|0))|0);
               $174 = __THREW__; __THREW__ = 0;
               $175 = $174&1;
               do {
                if ($175) {
                 $193 = ___cxa_find_matching_catch_2()|0;
                 $194 = tempRet0;
                 $$5109 = $194;$$5118 = $193;
                } else {
                 ;HEAP32[$16>>2]=HEAP32[$17>>2]|0;HEAP32[$16+4>>2]=HEAP32[$17+4>>2]|0;HEAP32[$16+8>>2]=HEAP32[$17+8>>2]|0;
                 $$0$i$i$i144 = 0;
                 while(1) {
                  $exitcond$i$i$i145 = ($$0$i$i$i144|0)==(3);
                  if ($exitcond$i$i$i145) {
                   break;
                  }
                  $176 = (($17) + ($$0$i$i$i144<<2)|0);
                  HEAP32[$176>>2] = 0;
                  $177 = (($$0$i$i$i144) + 1)|0;
                  $$0$i$i$i144 = $177;
                 }
                 $178 = HEAP32[$168>>2]|0;
                 $179 = ((($178)) + -24|0);
                 $180 = ((($16)) + 11|0);
                 $181 = HEAP8[$180>>0]|0;
                 $182 = ($181<<24>>24)<(0);
                 $183 = HEAP32[$16>>2]|0;
                 $184 = $182 ? $183 : $16;
                 $185 = ((($16)) + 4|0);
                 $186 = HEAP32[$185>>2]|0;
                 $187 = $181&255;
                 $188 = $182 ? $186 : $187;
                 __THREW__ = 0;
                 (invoke_iiii(1112,($179|0),($184|0),($188|0))|0);
                 $189 = __THREW__; __THREW__ = 0;
                 $190 = $189&1;
                 if ($190) {
                  $195 = ___cxa_find_matching_catch_2()|0;
                  $196 = tempRet0;
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
                  $$5109 = $196;$$5118 = $195;
                  break;
                 } else {
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($17);
                  $$899$ph = $163;
                  break L98;
                 }
                }
               } while(0);
               __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($17);
               $$6110 = $$5109;$$6119 = $$5118;
              }
              $$7111 = $$6110;$$7120 = $$6119;
              break L96;
             }
            }
           } else {
            $$899$ph = $0;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
         $$9 = $$899$ph;
         break L1;
        }
       } while(0);
       if ((label|0) == 60) {
        $135 = ___cxa_find_matching_catch_2()|0;
        $136 = tempRet0;
        $$7111 = $136;$$7120 = $135;
       }
       else if ((label|0) == 93) {
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
        $$9 = $0;
        break L1;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
       $$8112 = $$7111;$$8121 = $$7120;
      }
     } while(0);
     ___resumeException($$8121|0);
     // unreachable;
    } else {
     $$9 = $0;
    }
   } else {
    $$9 = $0;
   }
  } else {
   $$9 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$9|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$ = 0, $$017 = 0, $$2 = 0, $$off = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigittmp = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==($1|0);
 L1: do {
  if ($2) {
   $$2 = $0;
  } else {
   $3 = HEAP8[$0>>0]|0;
   $4 = ($3<<24>>24)==(110);
   $5 = ((($0)) + 1|0);
   $$ = $4 ? $5 : $0;
   $6 = ($$|0)==($1|0);
   if ($6) {
    $$2 = $0;
   } else {
    $7 = HEAP8[$$>>0]|0;
    $8 = ($7<<24>>24)==(48);
    $9 = ((($$)) + 1|0);
    if ($8) {
     $$2 = $9;
    } else {
     $$off = (($7) + -49)<<24>>24;
     $10 = ($$off&255)<(9);
     if ($10) {
      $$017 = $9;
      while(1) {
       $11 = ($$017|0)==($1|0);
       if ($11) {
        $$2 = $1;
        break L1;
       }
       $12 = HEAP8[$$017>>0]|0;
       $13 = $12 << 24 >> 24;
       $isdigittmp = (($13) + -48)|0;
       $isdigit = ($isdigittmp>>>0)<(10);
       if (!($isdigit)) {
        $$2 = $$017;
        break L1;
       }
       $14 = ((($$017)) + 1|0);
       $$017 = $14;
      }
     } else {
      $$2 = $0;
     }
    }
   }
  }
 } while(0);
 return ($$2|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i634 = 0, $$0$i$i$i614 = 0, $$0$i$i$i624 = 0, $$0$i$i$i628 = 0, $$0$i$i$i651 = 0, $$0$i$i$i655 = 0, $$0$i$i$i658 = 0, $$0$i$i$i682 = 0, $$0$i$i$i685 = 0, $$0$i$i$i709 = 0, $$0$i$i$i713 = 0, $$0$i$i$i723 = 0, $$0$i$i$i727 = 0, $$0$i$i$i731 = 0, $$0$i$i$i735 = 0, $$0$i$i$i739 = 0, $$0$i$i599 = 0;
 var $$0$i$i601 = 0, $$0$i$i603 = 0, $$0$i$i605 = 0, $$0$i$i607 = 0, $$0$i$i609 = 0, $$0$i$i611 = 0, $$0$i$i613 = 0, $$0$i$i617 = 0, $$0$i$i619 = 0, $$0$i$i621 = 0, $$0$i$i637 = 0, $$0$i$i639 = 0, $$0$i$i641 = 0, $$0$i$i643 = 0, $$0$i$i645 = 0, $$0$i$i647 = 0, $$0$i$i649 = 0, $$0$i$i664 = 0, $$0$i$i666 = 0, $$0$i$i668 = 0;
 var $$0$i$i670 = 0, $$0$i$i672 = 0, $$0$i$i674 = 0, $$0$i$i676 = 0, $$0$i$i678 = 0, $$0$i$i680 = 0, $$0$i$i689 = 0, $$0$i$i691 = 0, $$0$i$i693 = 0, $$0$i$i695 = 0, $$0$i$i697 = 0, $$0$i$i699 = 0, $$0$i$i701 = 0, $$0$i$i703 = 0, $$0$i$i705 = 0, $$0$i$i707 = 0, $$0$i$i717 = 0, $$0$i$i749 = 0, $$0$i$i751 = 0, $$0$i$i753 = 0;
 var $$0$i$i755 = 0, $$0503 = 0, $$0504$off0 = 0, $$0505 = 0, $$0526 = 0, $$10515 = 0, $$10536 = 0, $$12517 = 0, $$12538 = 0, $$14 = 0, $$14519 = 0, $$14540 = 0, $$1506 = 0, $$1527 = 0, $$15520 = 0, $$15541 = 0, $$16521 = 0, $$16542 = 0, $$17522 = 0, $$17543 = 0;
 var $$18523 = 0, $$18544 = 0, $$19524 = 0, $$19545 = 0, $$20525 = 0, $$20546 = 0, $$21 = 0, $$21547 = 0, $$2507 = 0, $$2528 = 0, $$3508 = 0, $$3529 = 0, $$5510 = 0, $$5531 = 0, $$558 = 0, $$559 = 0, $$560 = 0, $$561 = 0, $$562 = 0, $$563 = 0;
 var $$564 = 0, $$565 = 0, $$566 = 0, $$567 = 0, $$568 = 0, $$569 = 0, $$570 = 0, $$571 = 0, $$572 = 0, $$573 = 0, $$574 = 0, $$575 = 0, $$576 = 0, $$577 = 0, $$578 = 0, $$579 = 0, $$580 = 0, $$581 = 0, $$582 = 0, $$583 = 0;
 var $$584 = 0, $$585 = 0, $$586 = 0, $$587 = 0, $$588 = 0, $$589 = 0, $$590 = 0, $$591 = 0, $$592 = 0, $$593 = 0, $$594 = 0, $$595 = 0, $$596 = 0, $$597 = 0, $$598 = 0, $$6511 = 0, $$6532 = 0, $$7512 = 0, $$7533 = 0, $$8513 = 0;
 var $$8534 = 0, $$pre$i = 0, $$pre$i662 = 0, $$pre$i719 = 0, $$pre$i721 = 0, $$pre$i743 = 0, $$pre$i745 = 0, $$pre$i747 = 0, $10 = 0, $100 = 0, $1000 = 0, $1001 = 0, $1002 = 0, $1003 = 0, $1004 = 0, $1005 = 0, $1006 = 0, $1007 = 0, $1008 = 0, $1009 = 0;
 var $101 = 0, $1010 = 0, $1011 = 0, $1012 = 0, $1013 = 0, $1014 = 0, $1015 = 0, $1016 = 0, $1017 = 0, $1018 = 0, $1019 = 0, $102 = 0, $1020 = 0, $1021 = 0, $1022 = 0, $1023 = 0, $1024 = 0, $1025 = 0, $1026 = 0, $1027 = 0;
 var $1028 = 0, $1029 = 0, $103 = 0, $1030 = 0, $1031 = 0, $1032 = 0, $1033 = 0, $1034 = 0, $1035 = 0, $1036 = 0, $1037 = 0, $1038 = 0, $1039 = 0, $104 = 0, $1040 = 0, $1041 = 0, $1042 = 0, $1043 = 0, $1044 = 0, $1045 = 0;
 var $1046 = 0, $1047 = 0, $1048 = 0, $1049 = 0, $105 = 0, $1050 = 0, $1051 = 0, $1052 = 0, $1053 = 0, $1054 = 0, $1055 = 0, $1056 = 0, $1057 = 0, $1058 = 0, $1059 = 0, $106 = 0, $1060 = 0, $1061 = 0, $1062 = 0, $1063 = 0;
 var $1064 = 0, $1065 = 0, $1066 = 0, $1067 = 0, $1068 = 0, $1069 = 0, $107 = 0, $1070 = 0, $1071 = 0, $1072 = 0, $1073 = 0, $1074 = 0, $1075 = 0, $1076 = 0, $1077 = 0, $1078 = 0, $1079 = 0, $108 = 0, $1080 = 0, $1081 = 0;
 var $1082 = 0, $1083 = 0, $1084 = 0, $1085 = 0, $1086 = 0, $1087 = 0, $1088 = 0, $1089 = 0, $109 = 0, $1090 = 0, $1091 = 0, $1092 = 0, $1093 = 0, $1094 = 0, $1095 = 0, $1096 = 0, $1097 = 0, $1098 = 0, $1099 = 0, $11 = 0;
 var $110 = 0, $1100 = 0, $1101 = 0, $1102 = 0, $1103 = 0, $1104 = 0, $1105 = 0, $1106 = 0, $1107 = 0, $1108 = 0, $1109 = 0, $111 = 0, $1110 = 0, $1111 = 0, $1112 = 0, $1113 = 0, $1114 = 0, $1115 = 0, $1116 = 0, $1117 = 0;
 var $1118 = 0, $1119 = 0, $112 = 0, $1120 = 0, $1121 = 0, $1122 = 0, $1123 = 0, $1124 = 0, $1125 = 0, $1126 = 0, $1127 = 0, $1128 = 0, $1129 = 0, $113 = 0, $1130 = 0, $1131 = 0, $1132 = 0, $1133 = 0, $1134 = 0, $1135 = 0;
 var $1136 = 0, $1137 = 0, $1138 = 0, $1139 = 0, $114 = 0, $1140 = 0, $1141 = 0, $1142 = 0, $1143 = 0, $1144 = 0, $1145 = 0, $1146 = 0, $1147 = 0, $1148 = 0, $1149 = 0, $115 = 0, $1150 = 0, $1151 = 0, $1152 = 0, $1153 = 0;
 var $1154 = 0, $1155 = 0, $1156 = 0, $1157 = 0, $1158 = 0, $1159 = 0, $116 = 0, $1160 = 0, $1161 = 0, $1162 = 0, $1163 = 0, $1164 = 0, $1165 = 0, $1166 = 0, $1167 = 0, $1168 = 0, $1169 = 0, $117 = 0, $1170 = 0, $1171 = 0;
 var $1172 = 0, $1173 = 0, $1174 = 0, $1175 = 0, $1176 = 0, $1177 = 0, $1178 = 0, $1179 = 0, $118 = 0, $1180 = 0, $1181 = 0, $1182 = 0, $1183 = 0, $1184 = 0, $1185 = 0, $1186 = 0, $1187 = 0, $1188 = 0, $1189 = 0, $119 = 0;
 var $1190 = 0, $1191 = 0, $1192 = 0, $1193 = 0, $1194 = 0, $1195 = 0, $1196 = 0, $1197 = 0, $1198 = 0, $1199 = 0, $12 = 0, $120 = 0, $1200 = 0, $1201 = 0, $1202 = 0, $1203 = 0, $1204 = 0, $1205 = 0, $1206 = 0, $1207 = 0;
 var $1208 = 0, $1209 = 0, $121 = 0, $1210 = 0, $1211 = 0, $1212 = 0, $1213 = 0, $1214 = 0, $1215 = 0, $1216 = 0, $1217 = 0, $1218 = 0, $1219 = 0, $122 = 0, $1220 = 0, $1221 = 0, $1222 = 0, $1223 = 0, $1224 = 0, $1225 = 0;
 var $1226 = 0, $1227 = 0, $1228 = 0, $1229 = 0, $123 = 0, $1230 = 0, $1231 = 0, $1232 = 0, $1233 = 0, $1234 = 0, $1235 = 0, $1236 = 0, $1237 = 0, $1238 = 0, $1239 = 0, $124 = 0, $1240 = 0, $1241 = 0, $1242 = 0, $1243 = 0;
 var $1244 = 0, $1245 = 0, $1246 = 0, $1247 = 0, $1248 = 0, $1249 = 0, $125 = 0, $1250 = 0, $1251 = 0, $1252 = 0, $1253 = 0, $1254 = 0, $1255 = 0, $1256 = 0, $1257 = 0, $1258 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0;
 var $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0;
 var $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0;
 var $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0;
 var $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0;
 var $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0;
 var $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0;
 var $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0;
 var $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0;
 var $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0;
 var $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0;
 var $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0;
 var $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0;
 var $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0;
 var $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0;
 var $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0;
 var $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0;
 var $419 = 0, $42 = 0, $420 = 0, $421 = 0, $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0;
 var $437 = 0, $438 = 0, $439 = 0, $44 = 0, $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0;
 var $455 = 0, $456 = 0, $457 = 0, $458 = 0, $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0;
 var $473 = 0, $474 = 0, $475 = 0, $476 = 0, $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0;
 var $491 = 0, $492 = 0, $493 = 0, $494 = 0, $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0;
 var $509 = 0, $51 = 0, $510 = 0, $511 = 0, $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0;
 var $527 = 0, $528 = 0, $529 = 0, $53 = 0, $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0;
 var $545 = 0, $546 = 0, $547 = 0, $548 = 0, $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0;
 var $563 = 0, $564 = 0, $565 = 0, $566 = 0, $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0;
 var $581 = 0, $582 = 0, $583 = 0, $584 = 0, $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0;
 var $6 = 0, $60 = 0, $600 = 0, $601 = 0, $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $615 = 0, $616 = 0;
 var $617 = 0, $618 = 0, $619 = 0, $62 = 0, $620 = 0, $621 = 0, $622 = 0, $623 = 0, $624 = 0, $625 = 0, $626 = 0, $627 = 0, $628 = 0, $629 = 0, $63 = 0, $630 = 0, $631 = 0, $632 = 0, $633 = 0, $634 = 0;
 var $635 = 0, $636 = 0, $637 = 0, $638 = 0, $639 = 0, $64 = 0, $640 = 0, $641 = 0, $642 = 0, $643 = 0, $644 = 0, $645 = 0, $646 = 0, $647 = 0, $648 = 0, $649 = 0, $65 = 0, $650 = 0, $651 = 0, $652 = 0;
 var $653 = 0, $654 = 0, $655 = 0, $656 = 0, $657 = 0, $658 = 0, $659 = 0, $66 = 0, $660 = 0, $661 = 0, $662 = 0, $663 = 0, $664 = 0, $665 = 0, $666 = 0, $667 = 0, $668 = 0, $669 = 0, $67 = 0, $670 = 0;
 var $671 = 0, $672 = 0, $673 = 0, $674 = 0, $675 = 0, $676 = 0, $677 = 0, $678 = 0, $679 = 0, $68 = 0, $680 = 0, $681 = 0, $682 = 0, $683 = 0, $684 = 0, $685 = 0, $686 = 0, $687 = 0, $688 = 0, $689 = 0;
 var $69 = 0, $690 = 0, $691 = 0, $692 = 0, $693 = 0, $694 = 0, $695 = 0, $696 = 0, $697 = 0, $698 = 0, $699 = 0, $7 = 0, $70 = 0, $700 = 0, $701 = 0, $702 = 0, $703 = 0, $704 = 0, $705 = 0, $706 = 0;
 var $707 = 0, $708 = 0, $709 = 0, $71 = 0, $710 = 0, $711 = 0, $712 = 0, $713 = 0, $714 = 0, $715 = 0, $716 = 0, $717 = 0, $718 = 0, $719 = 0, $72 = 0, $720 = 0, $721 = 0, $722 = 0, $723 = 0, $724 = 0;
 var $725 = 0, $726 = 0, $727 = 0, $728 = 0, $729 = 0, $73 = 0, $730 = 0, $731 = 0, $732 = 0, $733 = 0, $734 = 0, $735 = 0, $736 = 0, $737 = 0, $738 = 0, $739 = 0, $74 = 0, $740 = 0, $741 = 0, $742 = 0;
 var $743 = 0, $744 = 0, $745 = 0, $746 = 0, $747 = 0, $748 = 0, $749 = 0, $75 = 0, $750 = 0, $751 = 0, $752 = 0, $753 = 0, $754 = 0, $755 = 0, $756 = 0, $757 = 0, $758 = 0, $759 = 0, $76 = 0, $760 = 0;
 var $761 = 0, $762 = 0, $763 = 0, $764 = 0, $765 = 0, $766 = 0, $767 = 0, $768 = 0, $769 = 0, $77 = 0, $770 = 0, $771 = 0, $772 = 0, $773 = 0, $774 = 0, $775 = 0, $776 = 0, $777 = 0, $778 = 0, $779 = 0;
 var $78 = 0, $780 = 0, $781 = 0, $782 = 0, $783 = 0, $784 = 0, $785 = 0, $786 = 0, $787 = 0, $788 = 0, $789 = 0, $79 = 0, $790 = 0, $791 = 0, $792 = 0, $793 = 0, $794 = 0, $795 = 0, $796 = 0, $797 = 0;
 var $798 = 0, $799 = 0, $8 = 0, $80 = 0, $800 = 0, $801 = 0, $802 = 0, $803 = 0, $804 = 0, $805 = 0, $806 = 0, $807 = 0, $808 = 0, $809 = 0, $81 = 0, $810 = 0, $811 = 0, $812 = 0, $813 = 0, $814 = 0;
 var $815 = 0, $816 = 0, $817 = 0, $818 = 0, $819 = 0, $82 = 0, $820 = 0, $821 = 0, $822 = 0, $823 = 0, $824 = 0, $825 = 0, $826 = 0, $827 = 0, $828 = 0, $829 = 0, $83 = 0, $830 = 0, $831 = 0, $832 = 0;
 var $833 = 0, $834 = 0, $835 = 0, $836 = 0, $837 = 0, $838 = 0, $839 = 0, $84 = 0, $840 = 0, $841 = 0, $842 = 0, $843 = 0, $844 = 0, $845 = 0, $846 = 0, $847 = 0, $848 = 0, $849 = 0, $85 = 0, $850 = 0;
 var $851 = 0, $852 = 0, $853 = 0, $854 = 0, $855 = 0, $856 = 0, $857 = 0, $858 = 0, $859 = 0, $86 = 0, $860 = 0, $861 = 0, $862 = 0, $863 = 0, $864 = 0, $865 = 0, $866 = 0, $867 = 0, $868 = 0, $869 = 0;
 var $87 = 0, $870 = 0, $871 = 0, $872 = 0, $873 = 0, $874 = 0, $875 = 0, $876 = 0, $877 = 0, $878 = 0, $879 = 0, $88 = 0, $880 = 0, $881 = 0, $882 = 0, $883 = 0, $884 = 0, $885 = 0, $886 = 0, $887 = 0;
 var $888 = 0, $889 = 0, $89 = 0, $890 = 0, $891 = 0, $892 = 0, $893 = 0, $894 = 0, $895 = 0, $896 = 0, $897 = 0, $898 = 0, $899 = 0, $9 = 0, $90 = 0, $900 = 0, $901 = 0, $902 = 0, $903 = 0, $904 = 0;
 var $905 = 0, $906 = 0, $907 = 0, $908 = 0, $909 = 0, $91 = 0, $910 = 0, $911 = 0, $912 = 0, $913 = 0, $914 = 0, $915 = 0, $916 = 0, $917 = 0, $918 = 0, $919 = 0, $92 = 0, $920 = 0, $921 = 0, $922 = 0;
 var $923 = 0, $924 = 0, $925 = 0, $926 = 0, $927 = 0, $928 = 0, $929 = 0, $93 = 0, $930 = 0, $931 = 0, $932 = 0, $933 = 0, $934 = 0, $935 = 0, $936 = 0, $937 = 0, $938 = 0, $939 = 0, $94 = 0, $940 = 0;
 var $941 = 0, $942 = 0, $943 = 0, $944 = 0, $945 = 0, $946 = 0, $947 = 0, $948 = 0, $949 = 0, $95 = 0, $950 = 0, $951 = 0, $952 = 0, $953 = 0, $954 = 0, $955 = 0, $956 = 0, $957 = 0, $958 = 0, $959 = 0;
 var $96 = 0, $960 = 0, $961 = 0, $962 = 0, $963 = 0, $964 = 0, $965 = 0, $966 = 0, $967 = 0, $968 = 0, $969 = 0, $97 = 0, $970 = 0, $971 = 0, $972 = 0, $973 = 0, $974 = 0, $975 = 0, $976 = 0, $977 = 0;
 var $978 = 0, $979 = 0, $98 = 0, $980 = 0, $981 = 0, $982 = 0, $983 = 0, $984 = 0, $985 = 0, $986 = 0, $987 = 0, $988 = 0, $989 = 0, $99 = 0, $990 = 0, $991 = 0, $992 = 0, $993 = 0, $994 = 0, $995 = 0;
 var $996 = 0, $997 = 0, $998 = 0, $999 = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i$i635 = 0, $exitcond$i$i$i615 = 0, $exitcond$i$i$i625 = 0, $exitcond$i$i$i629 = 0, $exitcond$i$i$i652 = 0, $exitcond$i$i$i656 = 0, $exitcond$i$i$i659 = 0, $exitcond$i$i$i683 = 0, $exitcond$i$i$i686 = 0, $exitcond$i$i$i710 = 0, $exitcond$i$i$i714 = 0, $exitcond$i$i$i724 = 0, $exitcond$i$i$i728 = 0;
 var $exitcond$i$i$i732 = 0, $exitcond$i$i$i736 = 0, $exitcond$i$i$i740 = 0, $exitcond$i$i622 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1008|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(1008|0);
 $3 = sp + 992|0;
 $4 = sp + 991|0;
 $5 = sp + 990|0;
 $6 = sp + 989|0;
 $7 = sp + 988|0;
 $8 = sp + 987|0;
 $9 = sp + 986|0;
 $10 = sp + 985|0;
 $11 = sp + 984|0;
 $12 = sp + 983|0;
 $13 = sp + 982|0;
 $14 = sp + 981|0;
 $15 = sp + 980|0;
 $16 = sp + 979|0;
 $17 = sp + 978|0;
 $18 = sp + 977|0;
 $19 = sp + 976|0;
 $20 = sp + 975|0;
 $21 = sp + 974|0;
 $22 = sp + 973|0;
 $23 = sp + 972|0;
 $24 = sp + 971|0;
 $25 = sp + 970|0;
 $26 = sp + 969|0;
 $27 = sp + 968|0;
 $28 = sp + 967|0;
 $29 = sp + 966|0;
 $30 = sp + 965|0;
 $31 = sp + 964|0;
 $32 = sp + 963|0;
 $33 = sp + 962|0;
 $34 = sp + 961|0;
 $35 = sp + 960|0;
 $36 = sp + 959|0;
 $37 = sp + 958|0;
 $38 = sp + 957|0;
 $39 = sp + 956|0;
 $40 = sp + 955|0;
 $41 = sp + 954|0;
 $42 = sp + 953|0;
 $43 = sp + 952|0;
 $44 = sp + 951|0;
 $45 = sp + 950|0;
 $46 = sp + 949|0;
 $47 = sp + 948|0;
 $48 = sp + 936|0;
 $49 = sp + 924|0;
 $50 = sp + 912|0;
 $51 = sp + 900|0;
 $52 = sp + 888|0;
 $53 = sp + 876|0;
 $54 = sp + 864|0;
 $55 = sp + 852|0;
 $56 = sp + 840|0;
 $57 = sp + 828|0;
 $58 = sp + 816|0;
 $59 = sp + 804|0;
 $60 = sp + 792|0;
 $61 = sp + 780|0;
 $62 = sp + 768|0;
 $63 = sp + 756|0;
 $64 = sp + 744|0;
 $65 = sp + 732|0;
 $66 = sp + 720|0;
 $67 = sp + 708|0;
 $68 = sp + 696|0;
 $69 = sp + 684|0;
 $70 = sp + 672|0;
 $71 = sp + 660|0;
 $72 = sp + 648|0;
 $73 = sp + 624|0;
 $74 = sp + 608|0;
 $75 = sp + 596|0;
 $76 = sp + 584|0;
 $77 = sp + 572|0;
 $78 = sp + 560|0;
 $79 = sp + 548|0;
 $80 = sp + 536|0;
 $81 = sp + 524|0;
 $82 = sp + 512|0;
 $83 = sp + 500|0;
 $84 = sp + 488|0;
 $85 = sp + 476|0;
 $86 = sp + 464|0;
 $87 = sp + 440|0;
 $88 = sp + 424|0;
 $89 = sp + 412|0;
 $90 = sp + 400|0;
 $91 = sp + 388|0;
 $92 = sp + 376|0;
 $93 = sp + 364|0;
 $94 = sp + 352|0;
 $95 = sp + 340|0;
 $96 = sp + 328|0;
 $97 = sp + 316|0;
 $98 = sp + 304|0;
 $99 = sp + 292|0;
 $100 = sp + 280|0;
 $101 = sp + 256|0;
 $102 = sp + 240|0;
 $103 = sp + 228|0;
 $104 = sp + 216|0;
 $105 = sp + 204|0;
 $106 = sp + 192|0;
 $107 = sp + 180|0;
 $108 = sp + 168|0;
 $109 = sp + 144|0;
 $110 = sp + 132|0;
 $111 = sp + 120|0;
 $112 = sp + 108|0;
 $113 = sp + 96|0;
 $114 = sp + 84|0;
 $115 = sp + 72|0;
 $116 = sp + 60|0;
 $117 = sp + 48|0;
 $118 = sp + 36|0;
 $119 = sp + 24|0;
 $120 = sp;
 $121 = $1;
 $122 = $0;
 $123 = (($121) - ($122))|0;
 $124 = ($123|0)>(1);
 L1: do {
  if ($124) {
   $125 = ($123|0)>(3);
   $126 = ((($0)) + 2|0);
   if ($125) {
    $127 = HEAP8[$0>>0]|0;
    $128 = ($127<<24>>24)==(103);
    if ($128) {
     $129 = ((($0)) + 1|0);
     $130 = HEAP8[$129>>0]|0;
     $131 = ($130<<24>>24)==(115);
     $$558 = $131 ? $126 : $0;
     $$0503 = $$558;$$0504$off0 = $131;
    } else {
     $$0503 = $0;$$0504$off0 = 0;
    }
   } else {
    $$0503 = $0;$$0504$off0 = 0;
   }
   $132 = HEAP8[$$0503>>0]|0;
   $133 = $132 << 24 >> 24;
   $134 = ((($$0503)) + 1|0);
   $135 = ((($0)) + 3|0);
   $136 = ($126|0)!=($1|0);
   $137 = ((($$0503)) + 2|0);
   L7: do {
    switch ($133|0) {
    case 76:  {
     $138 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_expr_primaryINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $$14 = $138;
     break L1;
     break;
    }
    case 84:  {
     $139 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $$14 = $139;
     break L1;
     break;
    }
    case 102:  {
     $140 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_function_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $$14 = $140;
     break L1;
     break;
    }
    case 97:  {
     $141 = HEAP8[$134>>0]|0;
     $142 = $141 << 24 >> 24;
     switch ($142|0) {
     case 97:  {
      ;HEAP32[$48>>2]=0|0;HEAP32[$48+4>>2]=0|0;HEAP32[$48+8>>2]=0|0;
      $143 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159085)|0);
      $144 = ($143>>>0)>(4294967279);
      if ($144) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($48);
       // unreachable;
      }
      $145 = ($143>>>0)<(11);
      if ($145) {
       $146 = $143&255;
       $147 = ((($48)) + 11|0);
       HEAP8[$147>>0] = $146;
       $$0$i$i = $48;
      } else {
       $148 = (($143) + 16)|0;
       $149 = $148 & -16;
       $150 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($149)|0);
       HEAP32[$48>>2] = $150;
       $151 = $149 | -2147483648;
       $152 = ((($48)) + 8|0);
       HEAP32[$152>>2] = $151;
       $153 = ((($48)) + 4|0);
       HEAP32[$153>>2] = $143;
       $$0$i$i = $150;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,159085,$143)|0);
      $154 = (($$0$i$i) + ($143)|0);
      HEAP8[$47>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($154,$47);
      __THREW__ = 0;
      $155 = (invoke_iiiii(1140,($126|0),($1|0),($48|0),($2|0))|0);
      $156 = __THREW__; __THREW__ = 0;
      $157 = $156&1;
      if ($157) {
       $159 = ___cxa_find_matching_catch_2()|0;
       $160 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($48);
       $$21 = $159;$$21547 = $160;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($48);
       $158 = ($155|0)==($126|0);
       $$559 = $158 ? $0 : $155;
       $$14 = $$559;
       break L1;
      }
      break;
     }
     case 100:  {
      ;HEAP32[$49>>2]=0|0;HEAP32[$49+4>>2]=0|0;HEAP32[$49+8>>2]=0|0;
      $161 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159103)|0);
      $162 = ($161>>>0)>(4294967279);
      if ($162) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($49);
       // unreachable;
      }
      $163 = ($161>>>0)<(11);
      if ($163) {
       $164 = $161&255;
       $165 = ((($49)) + 11|0);
       HEAP8[$165>>0] = $164;
       $$0$i$i599 = $49;
      } else {
       $166 = (($161) + 16)|0;
       $167 = $166 & -16;
       $168 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($167)|0);
       HEAP32[$49>>2] = $168;
       $169 = $167 | -2147483648;
       $170 = ((($49)) + 8|0);
       HEAP32[$170>>2] = $169;
       $171 = ((($49)) + 4|0);
       HEAP32[$171>>2] = $161;
       $$0$i$i599 = $168;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i599,159103,$161)|0);
      $172 = (($$0$i$i599) + ($161)|0);
      HEAP8[$46>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($172,$46);
      __THREW__ = 0;
      $173 = (invoke_iiiii(1141,($126|0),($1|0),($49|0),($2|0))|0);
      $174 = __THREW__; __THREW__ = 0;
      $175 = $174&1;
      if ($175) {
       $177 = ___cxa_find_matching_catch_2()|0;
       $178 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($49);
       $$21 = $177;$$21547 = $178;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($49);
       $176 = ($173|0)==($126|0);
       $$560 = $176 ? $0 : $173;
       $$14 = $$560;
       break L1;
      }
      break;
     }
     case 110:  {
      ;HEAP32[$50>>2]=0|0;HEAP32[$50+4>>2]=0|0;HEAP32[$50+8>>2]=0|0;
      $179 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159103)|0);
      $180 = ($179>>>0)>(4294967279);
      if ($180) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($50);
       // unreachable;
      }
      $181 = ($179>>>0)<(11);
      if ($181) {
       $182 = $179&255;
       $183 = ((($50)) + 11|0);
       HEAP8[$183>>0] = $182;
       $$0$i$i601 = $50;
      } else {
       $184 = (($179) + 16)|0;
       $185 = $184 & -16;
       $186 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($185)|0);
       HEAP32[$50>>2] = $186;
       $187 = $185 | -2147483648;
       $188 = ((($50)) + 8|0);
       HEAP32[$188>>2] = $187;
       $189 = ((($50)) + 4|0);
       HEAP32[$189>>2] = $179;
       $$0$i$i601 = $186;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i601,159103,$179)|0);
      $190 = (($$0$i$i601) + ($179)|0);
      HEAP8[$45>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($190,$45);
      __THREW__ = 0;
      $191 = (invoke_iiiii(1140,($126|0),($1|0),($50|0),($2|0))|0);
      $192 = __THREW__; __THREW__ = 0;
      $193 = $192&1;
      if ($193) {
       $195 = ___cxa_find_matching_catch_2()|0;
       $196 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($50);
       $$21 = $195;$$21547 = $196;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($50);
       $194 = ($191|0)==($126|0);
       $$561 = $194 ? $0 : $191;
       $$14 = $$561;
       break L1;
      }
      break;
     }
     case 78:  {
      ;HEAP32[$51>>2]=0|0;HEAP32[$51+4>>2]=0|0;HEAP32[$51+8>>2]=0|0;
      $197 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159146)|0);
      $198 = ($197>>>0)>(4294967279);
      if ($198) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($51);
       // unreachable;
      }
      $199 = ($197>>>0)<(11);
      if ($199) {
       $200 = $197&255;
       $201 = ((($51)) + 11|0);
       HEAP8[$201>>0] = $200;
       $$0$i$i603 = $51;
      } else {
       $202 = (($197) + 16)|0;
       $203 = $202 & -16;
       $204 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($203)|0);
       HEAP32[$51>>2] = $204;
       $205 = $203 | -2147483648;
       $206 = ((($51)) + 8|0);
       HEAP32[$206>>2] = $205;
       $207 = ((($51)) + 4|0);
       HEAP32[$207>>2] = $197;
       $$0$i$i603 = $204;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i603,159146,$197)|0);
      $208 = (($$0$i$i603) + ($197)|0);
      HEAP8[$44>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($208,$44);
      __THREW__ = 0;
      $209 = (invoke_iiiii(1140,($126|0),($1|0),($51|0),($2|0))|0);
      $210 = __THREW__; __THREW__ = 0;
      $211 = $210&1;
      if ($211) {
       $213 = ___cxa_find_matching_catch_2()|0;
       $214 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($51);
       $$21 = $213;$$21547 = $214;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($51);
       $212 = ($209|0)==($126|0);
       $$562 = $212 ? $0 : $209;
       $$14 = $$562;
       break L1;
      }
      break;
     }
     case 83:  {
      ;HEAP32[$52>>2]=0|0;HEAP32[$52+4>>2]=0|0;HEAP32[$52+8>>2]=0|0;
      $215 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159149)|0);
      $216 = ($215>>>0)>(4294967279);
      if ($216) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($52);
       // unreachable;
      }
      $217 = ($215>>>0)<(11);
      if ($217) {
       $218 = $215&255;
       $219 = ((($52)) + 11|0);
       HEAP8[$219>>0] = $218;
       $$0$i$i605 = $52;
      } else {
       $220 = (($215) + 16)|0;
       $221 = $220 & -16;
       $222 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($221)|0);
       HEAP32[$52>>2] = $222;
       $223 = $221 | -2147483648;
       $224 = ((($52)) + 8|0);
       HEAP32[$224>>2] = $223;
       $225 = ((($52)) + 4|0);
       HEAP32[$225>>2] = $215;
       $$0$i$i605 = $222;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i605,159149,$215)|0);
      $226 = (($$0$i$i605) + ($215)|0);
      HEAP8[$43>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($226,$43);
      __THREW__ = 0;
      $227 = (invoke_iiiii(1140,($126|0),($1|0),($52|0),($2|0))|0);
      $228 = __THREW__; __THREW__ = 0;
      $229 = $228&1;
      if ($229) {
       $231 = ___cxa_find_matching_catch_2()|0;
       $232 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($52);
       $$21 = $231;$$21547 = $232;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($52);
       $230 = ($227|0)==($126|0);
       $$563 = $230 ? $0 : $227;
       $$14 = $$563;
       break L1;
      }
      break;
     }
     case 116:  {
      $233 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_alignof_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $233;
      break L1;
      break;
     }
     case 122:  {
      $234 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_alignof_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $234;
      break L1;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 99:  {
     $235 = HEAP8[$134>>0]|0;
     $236 = $235 << 24 >> 24;
     switch ($236|0) {
     case 99:  {
      $237 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_const_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $237;
      break L1;
      break;
     }
     case 108:  {
      $238 = (__ZN10__cxxabiv112_GLOBAL__N_115parse_call_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $238;
      break L1;
      break;
     }
     case 109:  {
      ;HEAP32[$53>>2]=0|0;HEAP32[$53+4>>2]=0|0;HEAP32[$53+8>>2]=0|0;
      $239 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159151)|0);
      $240 = ($239>>>0)>(4294967279);
      if ($240) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($53);
       // unreachable;
      }
      $241 = ($239>>>0)<(11);
      if ($241) {
       $242 = $239&255;
       $243 = ((($53)) + 11|0);
       HEAP8[$243>>0] = $242;
       $$0$i$i607 = $53;
      } else {
       $244 = (($239) + 16)|0;
       $245 = $244 & -16;
       $246 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($245)|0);
       HEAP32[$53>>2] = $246;
       $247 = $245 | -2147483648;
       $248 = ((($53)) + 8|0);
       HEAP32[$248>>2] = $247;
       $249 = ((($53)) + 4|0);
       HEAP32[$249>>2] = $239;
       $$0$i$i607 = $246;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i607,159151,$239)|0);
      $250 = (($$0$i$i607) + ($239)|0);
      HEAP8[$42>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($250,$42);
      __THREW__ = 0;
      $251 = (invoke_iiiii(1140,($126|0),($1|0),($53|0),($2|0))|0);
      $252 = __THREW__; __THREW__ = 0;
      $253 = $252&1;
      if ($253) {
       $255 = ___cxa_find_matching_catch_2()|0;
       $256 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($53);
       $$21 = $255;$$21547 = $256;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($53);
       $254 = ($251|0)==($126|0);
       $$564 = $254 ? $0 : $251;
       $$14 = $$564;
       break L1;
      }
      break;
     }
     case 111:  {
      ;HEAP32[$54>>2]=0|0;HEAP32[$54+4>>2]=0|0;HEAP32[$54+8>>2]=0|0;
      $257 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159153)|0);
      $258 = ($257>>>0)>(4294967279);
      if ($258) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($54);
       // unreachable;
      }
      $259 = ($257>>>0)<(11);
      if ($259) {
       $260 = $257&255;
       $261 = ((($54)) + 11|0);
       HEAP8[$261>>0] = $260;
       $$0$i$i609 = $54;
      } else {
       $262 = (($257) + 16)|0;
       $263 = $262 & -16;
       $264 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($263)|0);
       HEAP32[$54>>2] = $264;
       $265 = $263 | -2147483648;
       $266 = ((($54)) + 8|0);
       HEAP32[$266>>2] = $265;
       $267 = ((($54)) + 4|0);
       HEAP32[$267>>2] = $257;
       $$0$i$i609 = $264;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i609,159153,$257)|0);
      $268 = (($$0$i$i609) + ($257)|0);
      HEAP8[$41>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($268,$41);
      __THREW__ = 0;
      $269 = (invoke_iiiii(1141,($126|0),($1|0),($54|0),($2|0))|0);
      $270 = __THREW__; __THREW__ = 0;
      $271 = $270&1;
      if ($271) {
       $273 = ___cxa_find_matching_catch_2()|0;
       $274 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($54);
       $$21 = $273;$$21547 = $274;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($54);
       $272 = ($269|0)==($126|0);
       $$565 = $272 ? $0 : $269;
       $$14 = $$565;
       break L1;
      }
      break;
     }
     case 118:  {
      $275 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_conversion_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $275;
      break L1;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 100:  {
     $276 = HEAP8[$134>>0]|0;
     $277 = $276 << 24 >> 24;
     switch ($277|0) {
     case 97:  {
      $278 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($137,$1,$2)|0);
      $279 = ($278|0)==($137|0);
      if ($279) {
       $$14 = $0;
       break L1;
      }
      $280 = HEAP32[$2>>2]|0;
      $281 = ((($2)) + 4|0);
      $282 = HEAP32[$281>>2]|0;
      $283 = ($280|0)==($282|0);
      if ($283) {
       $$14 = $0;
       break L1;
      }
      ;HEAP32[$57>>2]=0|0;HEAP32[$57+4>>2]=0|0;HEAP32[$57+8>>2]=0|0;
      L102: do {
       if ($$0504$off0) {
        $284 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159155)|0);
        $285 = ($284>>>0)>(4294967279);
        if ($285) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($57);
         // unreachable;
        }
        $286 = ($284>>>0)<(11);
        if ($286) {
         $287 = $284&255;
         $288 = ((($57)) + 11|0);
         HEAP8[$288>>0] = $287;
         $$0$i$i611 = $57;
        } else {
         $289 = (($284) + 16)|0;
         $290 = $289 & -16;
         $291 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($290)|0);
         HEAP32[$57>>2] = $291;
         $292 = $290 | -2147483648;
         $293 = ((($57)) + 8|0);
         HEAP32[$293>>2] = $292;
         $294 = ((($57)) + 4|0);
         HEAP32[$294>>2] = $284;
         $$0$i$i611 = $291;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i611,159155,$284)|0);
        $295 = (($$0$i$i611) + ($284)|0);
        HEAP8[$40>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($295,$40);
       } else {
        $$0$i$i613 = 0;
        while(1) {
         $exitcond$i$i = ($$0$i$i613|0)==(3);
         if ($exitcond$i$i) {
          break L102;
         }
         $296 = (($57) + ($$0$i$i613<<2)|0);
         HEAP32[$296>>2] = 0;
         $297 = (($$0$i$i613) + 1)|0;
         $$0$i$i613 = $297;
        }
       }
      } while(0);
      __THREW__ = 0;
      $298 = (invoke_iii(1119,($57|0),(159158|0))|0);
      $299 = __THREW__; __THREW__ = 0;
      $300 = $299&1;
      if ($300) {
       $334 = ___cxa_find_matching_catch_2()|0;
       $335 = tempRet0;
       $$1506 = $334;$$1527 = $335;
      } else {
       ;HEAP32[$56>>2]=HEAP32[$57>>2]|0;HEAP32[$56+4>>2]=HEAP32[$57+4>>2]|0;HEAP32[$56+8>>2]=HEAP32[$57+8>>2]|0;
       $$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $301 = (($57) + ($$0$i$i$i<<2)|0);
        HEAP32[$301>>2] = 0;
        $302 = (($$0$i$i$i) + 1)|0;
        $$0$i$i$i = $302;
       }
       $303 = HEAP32[$281>>2]|0;
       $304 = ((($303)) + -24|0);
       __THREW__ = 0;
       invoke_vii(1120,($58|0),($304|0));
       $305 = __THREW__; __THREW__ = 0;
       $306 = $305&1;
       do {
        if ($306) {
         $336 = ___cxa_find_matching_catch_2()|0;
         $337 = tempRet0;
         $$0505 = $336;$$0526 = $337;
        } else {
         $307 = ((($58)) + 11|0);
         $308 = HEAP8[$307>>0]|0;
         $309 = ($308<<24>>24)<(0);
         $310 = HEAP32[$58>>2]|0;
         $311 = $309 ? $310 : $58;
         $312 = ((($58)) + 4|0);
         $313 = HEAP32[$312>>2]|0;
         $314 = $308&255;
         $315 = $309 ? $313 : $314;
         __THREW__ = 0;
         $316 = (invoke_iiii(1112,($56|0),($311|0),($315|0))|0);
         $317 = __THREW__; __THREW__ = 0;
         $318 = $317&1;
         if ($318) {
          $338 = ___cxa_find_matching_catch_2()|0;
          $339 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($58);
          $$0505 = $338;$$0526 = $339;
          break;
         }
         ;HEAP32[$55>>2]=HEAP32[$56>>2]|0;HEAP32[$55+4>>2]=HEAP32[$56+4>>2]|0;HEAP32[$55+8>>2]=HEAP32[$56+8>>2]|0;
         $$0$i$i$i614 = 0;
         while(1) {
          $exitcond$i$i$i615 = ($$0$i$i$i614|0)==(3);
          if ($exitcond$i$i$i615) {
           break;
          }
          $319 = (($56) + ($$0$i$i$i614<<2)|0);
          HEAP32[$319>>2] = 0;
          $320 = (($$0$i$i$i614) + 1)|0;
          $$0$i$i$i614 = $320;
         }
         $321 = HEAP32[$281>>2]|0;
         $322 = ((($321)) + -24|0);
         $323 = ((($322)) + 11|0);
         $324 = HEAP8[$323>>0]|0;
         $325 = ($324<<24>>24)<(0);
         if ($325) {
          $326 = HEAP32[$322>>2]|0;
          HEAP8[$38>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($326,$38);
          $327 = ((($321)) + -20|0);
          HEAP32[$327>>2] = 0;
         } else {
          HEAP8[$39>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($322,$39);
          HEAP8[$323>>0] = 0;
         }
         __THREW__ = 0;
         invoke_vi(1116,($322|0));
         $328 = __THREW__; __THREW__ = 0;
         $329 = $328&1;
         if ($329) {
          $330 = ___cxa_find_matching_catch_3(0|0)|0;
          $331 = tempRet0;
          ___clang_call_terminate($330);
          // unreachable;
         }
         ;HEAP32[$322>>2]=HEAP32[$55>>2]|0;HEAP32[$322+4>>2]=HEAP32[$55+4>>2]|0;HEAP32[$322+8>>2]=HEAP32[$55+8>>2]|0;
         $$0$i$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
          if ($exitcond$i$i$i$i) {
           break;
          }
          $332 = (($55) + ($$0$i$i$i$i<<2)|0);
          HEAP32[$332>>2] = 0;
          $333 = (($$0$i$i$i$i) + 1)|0;
          $$0$i$i$i$i = $333;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($55);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($58);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($56);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($57);
         $$14 = $278;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($56);
       $$1506 = $$0505;$$1527 = $$0526;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($57);
      $$21 = $$1506;$$21547 = $$1527;
      break L7;
      break;
     }
     case 99:  {
      $340 = (__ZN10__cxxabiv112_GLOBAL__N_123parse_dynamic_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $340;
      break L1;
      break;
     }
     case 101:  {
      ;HEAP32[$59>>2]=0|0;HEAP32[$59+4>>2]=0|0;HEAP32[$59+8>>2]=0|0;
      $341 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159101)|0);
      $342 = ($341>>>0)>(4294967279);
      if ($342) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($59);
       // unreachable;
      }
      $343 = ($341>>>0)<(11);
      if ($343) {
       $344 = $341&255;
       $345 = ((($59)) + 11|0);
       HEAP8[$345>>0] = $344;
       $$0$i$i617 = $59;
      } else {
       $346 = (($341) + 16)|0;
       $347 = $346 & -16;
       $348 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($347)|0);
       HEAP32[$59>>2] = $348;
       $349 = $347 | -2147483648;
       $350 = ((($59)) + 8|0);
       HEAP32[$350>>2] = $349;
       $351 = ((($59)) + 4|0);
       HEAP32[$351>>2] = $341;
       $$0$i$i617 = $348;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i617,159101,$341)|0);
      $352 = (($$0$i$i617) + ($341)|0);
      HEAP8[$37>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($352,$37);
      __THREW__ = 0;
      $353 = (invoke_iiiii(1141,($126|0),($1|0),($59|0),($2|0))|0);
      $354 = __THREW__; __THREW__ = 0;
      $355 = $354&1;
      if ($355) {
       $357 = ___cxa_find_matching_catch_2()|0;
       $358 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($59);
       $$21 = $357;$$21547 = $358;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($59);
       $356 = ($353|0)==($126|0);
       $$566 = $356 ? $0 : $353;
       $$14 = $$566;
       break L1;
      }
      break;
     }
     case 108:  {
      $359 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($137,$1,$2)|0);
      $360 = ($359|0)==($137|0);
      if ($360) {
       $$14 = $0;
       break L1;
      }
      $361 = HEAP32[$2>>2]|0;
      $362 = ((($2)) + 4|0);
      $363 = HEAP32[$362>>2]|0;
      $364 = ($361|0)==($363|0);
      if ($364) {
       $$14 = $0;
       break L1;
      }
      ;HEAP32[$62>>2]=0|0;HEAP32[$62+4>>2]=0|0;HEAP32[$62+8>>2]=0|0;
      L161: do {
       if ($$0504$off0) {
        $365 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159155)|0);
        $366 = ($365>>>0)>(4294967279);
        if ($366) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($62);
         // unreachable;
        }
        $367 = ($365>>>0)<(11);
        if ($367) {
         $368 = $365&255;
         $369 = ((($62)) + 11|0);
         HEAP8[$369>>0] = $368;
         $$0$i$i619 = $62;
        } else {
         $370 = (($365) + 16)|0;
         $371 = $370 & -16;
         $372 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($371)|0);
         HEAP32[$62>>2] = $372;
         $373 = $371 | -2147483648;
         $374 = ((($62)) + 8|0);
         HEAP32[$374>>2] = $373;
         $375 = ((($62)) + 4|0);
         HEAP32[$375>>2] = $365;
         $$0$i$i619 = $372;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i619,159155,$365)|0);
        $376 = (($$0$i$i619) + ($365)|0);
        HEAP8[$36>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($376,$36);
       } else {
        $$0$i$i621 = 0;
        while(1) {
         $exitcond$i$i622 = ($$0$i$i621|0)==(3);
         if ($exitcond$i$i622) {
          break L161;
         }
         $377 = (($62) + ($$0$i$i621<<2)|0);
         HEAP32[$377>>2] = 0;
         $378 = (($$0$i$i621) + 1)|0;
         $$0$i$i621 = $378;
        }
       }
      } while(0);
      __THREW__ = 0;
      $379 = (invoke_iii(1119,($62|0),(159168|0))|0);
      $380 = __THREW__; __THREW__ = 0;
      $381 = $380&1;
      if ($381) {
       $415 = ___cxa_find_matching_catch_2()|0;
       $416 = tempRet0;
       $$3508 = $415;$$3529 = $416;
      } else {
       ;HEAP32[$61>>2]=HEAP32[$62>>2]|0;HEAP32[$61+4>>2]=HEAP32[$62+4>>2]|0;HEAP32[$61+8>>2]=HEAP32[$62+8>>2]|0;
       $$0$i$i$i624 = 0;
       while(1) {
        $exitcond$i$i$i625 = ($$0$i$i$i624|0)==(3);
        if ($exitcond$i$i$i625) {
         break;
        }
        $382 = (($62) + ($$0$i$i$i624<<2)|0);
        HEAP32[$382>>2] = 0;
        $383 = (($$0$i$i$i624) + 1)|0;
        $$0$i$i$i624 = $383;
       }
       $384 = HEAP32[$362>>2]|0;
       $385 = ((($384)) + -24|0);
       __THREW__ = 0;
       invoke_vii(1120,($63|0),($385|0));
       $386 = __THREW__; __THREW__ = 0;
       $387 = $386&1;
       do {
        if ($387) {
         $417 = ___cxa_find_matching_catch_2()|0;
         $418 = tempRet0;
         $$2507 = $417;$$2528 = $418;
        } else {
         $388 = ((($63)) + 11|0);
         $389 = HEAP8[$388>>0]|0;
         $390 = ($389<<24>>24)<(0);
         $391 = HEAP32[$63>>2]|0;
         $392 = $390 ? $391 : $63;
         $393 = ((($63)) + 4|0);
         $394 = HEAP32[$393>>2]|0;
         $395 = $389&255;
         $396 = $390 ? $394 : $395;
         __THREW__ = 0;
         $397 = (invoke_iiii(1112,($61|0),($392|0),($396|0))|0);
         $398 = __THREW__; __THREW__ = 0;
         $399 = $398&1;
         if ($399) {
          $419 = ___cxa_find_matching_catch_2()|0;
          $420 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($63);
          $$2507 = $419;$$2528 = $420;
          break;
         }
         ;HEAP32[$60>>2]=HEAP32[$61>>2]|0;HEAP32[$60+4>>2]=HEAP32[$61+4>>2]|0;HEAP32[$60+8>>2]=HEAP32[$61+8>>2]|0;
         $$0$i$i$i628 = 0;
         while(1) {
          $exitcond$i$i$i629 = ($$0$i$i$i628|0)==(3);
          if ($exitcond$i$i$i629) {
           break;
          }
          $400 = (($61) + ($$0$i$i$i628<<2)|0);
          HEAP32[$400>>2] = 0;
          $401 = (($$0$i$i$i628) + 1)|0;
          $$0$i$i$i628 = $401;
         }
         $402 = HEAP32[$362>>2]|0;
         $403 = ((($402)) + -24|0);
         $404 = ((($403)) + 11|0);
         $405 = HEAP8[$404>>0]|0;
         $406 = ($405<<24>>24)<(0);
         if ($406) {
          $407 = HEAP32[$403>>2]|0;
          HEAP8[$34>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($407,$34);
          $408 = ((($402)) + -20|0);
          HEAP32[$408>>2] = 0;
         } else {
          HEAP8[$35>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($403,$35);
          HEAP8[$404>>0] = 0;
         }
         __THREW__ = 0;
         invoke_vi(1116,($403|0));
         $409 = __THREW__; __THREW__ = 0;
         $410 = $409&1;
         if ($410) {
          $411 = ___cxa_find_matching_catch_3(0|0)|0;
          $412 = tempRet0;
          ___clang_call_terminate($411);
          // unreachable;
         }
         ;HEAP32[$403>>2]=HEAP32[$60>>2]|0;HEAP32[$403+4>>2]=HEAP32[$60+4>>2]|0;HEAP32[$403+8>>2]=HEAP32[$60+8>>2]|0;
         $$0$i$i$i$i634 = 0;
         while(1) {
          $exitcond$i$i$i$i635 = ($$0$i$i$i$i634|0)==(3);
          if ($exitcond$i$i$i$i635) {
           break;
          }
          $413 = (($60) + ($$0$i$i$i$i634<<2)|0);
          HEAP32[$413>>2] = 0;
          $414 = (($$0$i$i$i$i634) + 1)|0;
          $$0$i$i$i$i634 = $414;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($60);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($63);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($61);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($62);
         $$14 = $359;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($61);
       $$3508 = $$2507;$$3529 = $$2528;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($62);
      $$21 = $$3508;$$21547 = $$3529;
      break L7;
      break;
     }
     case 110:  {
      $421 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $421;
      break L1;
      break;
     }
     case 115:  {
      $422 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_dot_star_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $422;
      break L1;
      break;
     }
     case 116:  {
      $423 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_dot_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $423;
      break L1;
      break;
     }
     case 118:  {
      ;HEAP32[$64>>2]=0|0;HEAP32[$64+4>>2]=0|0;HEAP32[$64+8>>2]=0|0;
      $424 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159176)|0);
      $425 = ($424>>>0)>(4294967279);
      if ($425) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($64);
       // unreachable;
      }
      $426 = ($424>>>0)<(11);
      if ($426) {
       $427 = $424&255;
       $428 = ((($64)) + 11|0);
       HEAP8[$428>>0] = $427;
       $$0$i$i637 = $64;
      } else {
       $429 = (($424) + 16)|0;
       $430 = $429 & -16;
       $431 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($430)|0);
       HEAP32[$64>>2] = $431;
       $432 = $430 | -2147483648;
       $433 = ((($64)) + 8|0);
       HEAP32[$433>>2] = $432;
       $434 = ((($64)) + 4|0);
       HEAP32[$434>>2] = $424;
       $$0$i$i637 = $431;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i637,159176,$424)|0);
      $435 = (($$0$i$i637) + ($424)|0);
      HEAP8[$33>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($435,$33);
      __THREW__ = 0;
      $436 = (invoke_iiiii(1140,($126|0),($1|0),($64|0),($2|0))|0);
      $437 = __THREW__; __THREW__ = 0;
      $438 = $437&1;
      if ($438) {
       $440 = ___cxa_find_matching_catch_2()|0;
       $441 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($64);
       $$21 = $440;$$21547 = $441;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($64);
       $439 = ($436|0)==($126|0);
       $$567 = $439 ? $0 : $436;
       $$14 = $$567;
       break L1;
      }
      break;
     }
     case 86:  {
      ;HEAP32[$65>>2]=0|0;HEAP32[$65+4>>2]=0|0;HEAP32[$65+8>>2]=0|0;
      $442 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159178)|0);
      $443 = ($442>>>0)>(4294967279);
      if ($443) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($65);
       // unreachable;
      }
      $444 = ($442>>>0)<(11);
      if ($444) {
       $445 = $442&255;
       $446 = ((($65)) + 11|0);
       HEAP8[$446>>0] = $445;
       $$0$i$i639 = $65;
      } else {
       $447 = (($442) + 16)|0;
       $448 = $447 & -16;
       $449 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($448)|0);
       HEAP32[$65>>2] = $449;
       $450 = $448 | -2147483648;
       $451 = ((($65)) + 8|0);
       HEAP32[$451>>2] = $450;
       $452 = ((($65)) + 4|0);
       HEAP32[$452>>2] = $442;
       $$0$i$i639 = $449;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i639,159178,$442)|0);
      $453 = (($$0$i$i639) + ($442)|0);
      HEAP8[$32>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($453,$32);
      __THREW__ = 0;
      $454 = (invoke_iiiii(1140,($126|0),($1|0),($65|0),($2|0))|0);
      $455 = __THREW__; __THREW__ = 0;
      $456 = $455&1;
      if ($456) {
       $458 = ___cxa_find_matching_catch_2()|0;
       $459 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($65);
       $$21 = $458;$$21547 = $459;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($65);
       $457 = ($454|0)==($126|0);
       $$568 = $457 ? $0 : $454;
       $$14 = $$568;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 101:  {
     $460 = HEAP8[$134>>0]|0;
     $461 = $460 << 24 >> 24;
     switch ($461|0) {
     case 111:  {
      ;HEAP32[$66>>2]=0|0;HEAP32[$66+4>>2]=0|0;HEAP32[$66+8>>2]=0|0;
      $462 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159181)|0);
      $463 = ($462>>>0)>(4294967279);
      if ($463) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($66);
       // unreachable;
      }
      $464 = ($462>>>0)<(11);
      if ($464) {
       $465 = $462&255;
       $466 = ((($66)) + 11|0);
       HEAP8[$466>>0] = $465;
       $$0$i$i641 = $66;
      } else {
       $467 = (($462) + 16)|0;
       $468 = $467 & -16;
       $469 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($468)|0);
       HEAP32[$66>>2] = $469;
       $470 = $468 | -2147483648;
       $471 = ((($66)) + 8|0);
       HEAP32[$471>>2] = $470;
       $472 = ((($66)) + 4|0);
       HEAP32[$472>>2] = $462;
       $$0$i$i641 = $469;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i641,159181,$462)|0);
      $473 = (($$0$i$i641) + ($462)|0);
      HEAP8[$31>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($473,$31);
      __THREW__ = 0;
      $474 = (invoke_iiiii(1140,($126|0),($1|0),($66|0),($2|0))|0);
      $475 = __THREW__; __THREW__ = 0;
      $476 = $475&1;
      if ($476) {
       $478 = ___cxa_find_matching_catch_2()|0;
       $479 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($66);
       $$21 = $478;$$21547 = $479;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($66);
       $477 = ($474|0)==($126|0);
       $$569 = $477 ? $0 : $474;
       $$14 = $$569;
       break L1;
      }
      break;
     }
     case 79:  {
      ;HEAP32[$67>>2]=0|0;HEAP32[$67+4>>2]=0|0;HEAP32[$67+8>>2]=0|0;
      $480 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159183)|0);
      $481 = ($480>>>0)>(4294967279);
      if ($481) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($67);
       // unreachable;
      }
      $482 = ($480>>>0)<(11);
      if ($482) {
       $483 = $480&255;
       $484 = ((($67)) + 11|0);
       HEAP8[$484>>0] = $483;
       $$0$i$i643 = $67;
      } else {
       $485 = (($480) + 16)|0;
       $486 = $485 & -16;
       $487 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($486)|0);
       HEAP32[$67>>2] = $487;
       $488 = $486 | -2147483648;
       $489 = ((($67)) + 8|0);
       HEAP32[$489>>2] = $488;
       $490 = ((($67)) + 4|0);
       HEAP32[$490>>2] = $480;
       $$0$i$i643 = $487;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i643,159183,$480)|0);
      $491 = (($$0$i$i643) + ($480)|0);
      HEAP8[$30>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($491,$30);
      __THREW__ = 0;
      $492 = (invoke_iiiii(1140,($126|0),($1|0),($67|0),($2|0))|0);
      $493 = __THREW__; __THREW__ = 0;
      $494 = $493&1;
      if ($494) {
       $496 = ___cxa_find_matching_catch_2()|0;
       $497 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($67);
       $$21 = $496;$$21547 = $497;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($67);
       $495 = ($492|0)==($126|0);
       $$570 = $495 ? $0 : $492;
       $$14 = $$570;
       break L1;
      }
      break;
     }
     case 113:  {
      ;HEAP32[$68>>2]=0|0;HEAP32[$68+4>>2]=0|0;HEAP32[$68+8>>2]=0|0;
      $498 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159186)|0);
      $499 = ($498>>>0)>(4294967279);
      if ($499) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($68);
       // unreachable;
      }
      $500 = ($498>>>0)<(11);
      if ($500) {
       $501 = $498&255;
       $502 = ((($68)) + 11|0);
       HEAP8[$502>>0] = $501;
       $$0$i$i645 = $68;
      } else {
       $503 = (($498) + 16)|0;
       $504 = $503 & -16;
       $505 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($504)|0);
       HEAP32[$68>>2] = $505;
       $506 = $504 | -2147483648;
       $507 = ((($68)) + 8|0);
       HEAP32[$507>>2] = $506;
       $508 = ((($68)) + 4|0);
       HEAP32[$508>>2] = $498;
       $$0$i$i645 = $505;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i645,159186,$498)|0);
      $509 = (($$0$i$i645) + ($498)|0);
      HEAP8[$29>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($509,$29);
      __THREW__ = 0;
      $510 = (invoke_iiiii(1140,($126|0),($1|0),($68|0),($2|0))|0);
      $511 = __THREW__; __THREW__ = 0;
      $512 = $511&1;
      if ($512) {
       $514 = ___cxa_find_matching_catch_2()|0;
       $515 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($68);
       $$21 = $514;$$21547 = $515;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($68);
       $513 = ($510|0)==($126|0);
       $$571 = $513 ? $0 : $510;
       $$14 = $$571;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 103:  {
     $516 = HEAP8[$134>>0]|0;
     $517 = $516 << 24 >> 24;
     switch ($517|0) {
     case 101:  {
      ;HEAP32[$69>>2]=0|0;HEAP32[$69+4>>2]=0|0;HEAP32[$69+8>>2]=0|0;
      $518 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159189)|0);
      $519 = ($518>>>0)>(4294967279);
      if ($519) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($69);
       // unreachable;
      }
      $520 = ($518>>>0)<(11);
      if ($520) {
       $521 = $518&255;
       $522 = ((($69)) + 11|0);
       HEAP8[$522>>0] = $521;
       $$0$i$i647 = $69;
      } else {
       $523 = (($518) + 16)|0;
       $524 = $523 & -16;
       $525 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($524)|0);
       HEAP32[$69>>2] = $525;
       $526 = $524 | -2147483648;
       $527 = ((($69)) + 8|0);
       HEAP32[$527>>2] = $526;
       $528 = ((($69)) + 4|0);
       HEAP32[$528>>2] = $518;
       $$0$i$i647 = $525;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i647,159189,$518)|0);
      $529 = (($$0$i$i647) + ($518)|0);
      HEAP8[$28>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($529,$28);
      __THREW__ = 0;
      $530 = (invoke_iiiii(1140,($126|0),($1|0),($69|0),($2|0))|0);
      $531 = __THREW__; __THREW__ = 0;
      $532 = $531&1;
      if ($532) {
       $534 = ___cxa_find_matching_catch_2()|0;
       $535 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($69);
       $$21 = $534;$$21547 = $535;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($69);
       $533 = ($530|0)==($126|0);
       $$572 = $533 ? $0 : $530;
       $$14 = $$572;
       break L1;
      }
      break;
     }
     case 116:  {
      ;HEAP32[$70>>2]=0|0;HEAP32[$70+4>>2]=0|0;HEAP32[$70+8>>2]=0|0;
      $536 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159119)|0);
      $537 = ($536>>>0)>(4294967279);
      if ($537) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($70);
       // unreachable;
      }
      $538 = ($536>>>0)<(11);
      if ($538) {
       $539 = $536&255;
       $540 = ((($70)) + 11|0);
       HEAP8[$540>>0] = $539;
       $$0$i$i649 = $70;
      } else {
       $541 = (($536) + 16)|0;
       $542 = $541 & -16;
       $543 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($542)|0);
       HEAP32[$70>>2] = $543;
       $544 = $542 | -2147483648;
       $545 = ((($70)) + 8|0);
       HEAP32[$545>>2] = $544;
       $546 = ((($70)) + 4|0);
       HEAP32[$546>>2] = $536;
       $$0$i$i649 = $543;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i649,159119,$536)|0);
      $547 = (($$0$i$i649) + ($536)|0);
      HEAP8[$27>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($547,$27);
      __THREW__ = 0;
      $548 = (invoke_iiiii(1140,($126|0),($1|0),($70|0),($2|0))|0);
      $549 = __THREW__; __THREW__ = 0;
      $550 = $549&1;
      if ($550) {
       $552 = ___cxa_find_matching_catch_2()|0;
       $553 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($70);
       $$21 = $552;$$21547 = $553;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($70);
       $551 = ($548|0)==($126|0);
       $$573 = $551 ? $0 : $548;
       $$14 = $$573;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 105:  {
     $554 = HEAP8[$134>>0]|0;
     $555 = ($554<<24>>24)==(120);
     if (!($555)) {
      $$14 = $0;
      break L1;
     }
     $556 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($126,$1,$2)|0);
     $557 = ($556|0)==($126|0);
     if ($557) {
      $$14 = $0;
      break L1;
     }
     $558 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($556,$1,$2)|0);
     $559 = ($558|0)==($556|0);
     $560 = ((($2)) + 4|0);
     if ($559) {
      $614 = HEAP32[$560>>2]|0;
      $615 = ((($614)) + -24|0);
      $617 = $614;
      while(1) {
       $616 = ($617|0)==($615|0);
       if ($616) {
        $$14 = $0;
        break L1;
       }
       $618 = ((($617)) + -24|0);
       HEAP32[$560>>2] = $618;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($618);
       $$pre$i662 = HEAP32[$560>>2]|0;
       $617 = $$pre$i662;
      }
     }
     $561 = HEAP32[$560>>2]|0;
     $562 = HEAP32[$2>>2]|0;
     $563 = (($561) - ($562))|0;
     $564 = (($563|0) / 24)&-1;
     $565 = ($564>>>0)<(2);
     if ($565) {
      $$14 = $0;
      break L1;
     }
     $566 = $561;
     $567 = ((($566)) + -24|0);
     __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($71,$567);
     $568 = HEAP32[$560>>2]|0;
     $569 = ((($568)) + -24|0);
     $571 = $568;
     while(1) {
      $570 = ($571|0)==($569|0);
      if ($570) {
       break;
      }
      $572 = ((($571)) + -24|0);
      HEAP32[$560>>2] = $572;
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($572);
      $$pre$i = HEAP32[$560>>2]|0;
      $571 = $$pre$i;
     }
     $573 = ((($568)) + -48|0);
     __THREW__ = 0;
     invoke_vii(1120,($72|0),($573|0));
     $574 = __THREW__; __THREW__ = 0;
     $575 = $574&1;
     if ($575) {
      $604 = ___cxa_find_matching_catch_2()|0;
      $605 = tempRet0;
      $$8513 = $604;$$8534 = $605;
     } else {
      __THREW__ = 0;
      invoke_viii(1139,($77|0),(159083|0),($72|0));
      $576 = __THREW__; __THREW__ = 0;
      $577 = $576&1;
      if ($577) {
       $606 = ___cxa_find_matching_catch_2()|0;
       $607 = tempRet0;
       $$7512 = $606;$$7533 = $607;
      } else {
       __THREW__ = 0;
       $578 = (invoke_iii(1119,($77|0),(159192|0))|0);
       $579 = __THREW__; __THREW__ = 0;
       $580 = $579&1;
       if ($580) {
        $608 = ___cxa_find_matching_catch_2()|0;
        $609 = tempRet0;
        $$6511 = $608;$$6532 = $609;
       } else {
        ;HEAP32[$76>>2]=HEAP32[$77>>2]|0;HEAP32[$76+4>>2]=HEAP32[$77+4>>2]|0;HEAP32[$76+8>>2]=HEAP32[$77+8>>2]|0;
        $$0$i$i$i651 = 0;
        while(1) {
         $exitcond$i$i$i652 = ($$0$i$i$i651|0)==(3);
         if ($exitcond$i$i$i652) {
          break;
         }
         $581 = (($77) + ($$0$i$i$i651<<2)|0);
         HEAP32[$581>>2] = 0;
         $582 = (($$0$i$i$i651) + 1)|0;
         $$0$i$i$i651 = $582;
        }
        $583 = ((($71)) + 11|0);
        $584 = HEAP8[$583>>0]|0;
        $585 = ($584<<24>>24)<(0);
        $586 = HEAP32[$71>>2]|0;
        $587 = $585 ? $586 : $71;
        $588 = ((($71)) + 4|0);
        $589 = HEAP32[$588>>2]|0;
        $590 = $584&255;
        $591 = $585 ? $589 : $590;
        __THREW__ = 0;
        $592 = (invoke_iiii(1112,($76|0),($587|0),($591|0))|0);
        $593 = __THREW__; __THREW__ = 0;
        $594 = $593&1;
        do {
         if ($594) {
          $610 = ___cxa_find_matching_catch_2()|0;
          $611 = tempRet0;
          $$5510 = $610;$$5531 = $611;
         } else {
          ;HEAP32[$75>>2]=HEAP32[$76>>2]|0;HEAP32[$75+4>>2]=HEAP32[$76+4>>2]|0;HEAP32[$75+8>>2]=HEAP32[$76+8>>2]|0;
          $$0$i$i$i655 = 0;
          while(1) {
           $exitcond$i$i$i656 = ($$0$i$i$i655|0)==(3);
           if ($exitcond$i$i$i656) {
            break;
           }
           $595 = (($76) + ($$0$i$i$i655<<2)|0);
           HEAP32[$595>>2] = 0;
           $596 = (($$0$i$i$i655) + 1)|0;
           $$0$i$i$i655 = $596;
          }
          __THREW__ = 0;
          $597 = (invoke_iii(1119,($75|0),(159130|0))|0);
          $598 = __THREW__; __THREW__ = 0;
          $599 = $598&1;
          if ($599) {
           $612 = ___cxa_find_matching_catch_2()|0;
           $613 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($75);
           $$5510 = $612;$$5531 = $613;
           break;
          }
          ;HEAP32[$74>>2]=HEAP32[$75>>2]|0;HEAP32[$74+4>>2]=HEAP32[$75+4>>2]|0;HEAP32[$74+8>>2]=HEAP32[$75+8>>2]|0;
          $$0$i$i$i658 = 0;
          while(1) {
           $exitcond$i$i$i659 = ($$0$i$i$i658|0)==(3);
           if ($exitcond$i$i$i659) {
            break;
           }
           $600 = (($75) + ($$0$i$i$i658<<2)|0);
           HEAP32[$600>>2] = 0;
           $601 = (($$0$i$i$i658) + 1)|0;
           $$0$i$i$i658 = $601;
          }
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($73,$74);
          $602 = HEAP32[$560>>2]|0;
          $603 = ((($602)) + -24|0);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($603,$73);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($73);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($74);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($75);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($76);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($77);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($72);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($71);
          $$14 = $558;
          break L1;
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($76);
        $$6511 = $$5510;$$6532 = $$5531;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($77);
       $$7512 = $$6511;$$7533 = $$6532;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($72);
      $$8513 = $$7512;$$8534 = $$7533;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($71);
     $$21 = $$8513;$$21547 = $$8534;
     break;
    }
    case 108:  {
     $619 = HEAP8[$134>>0]|0;
     $620 = $619 << 24 >> 24;
     switch ($620|0) {
     case 101:  {
      ;HEAP32[$78>>2]=0|0;HEAP32[$78+4>>2]=0|0;HEAP32[$78+8>>2]=0|0;
      $621 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159195)|0);
      $622 = ($621>>>0)>(4294967279);
      if ($622) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($78);
       // unreachable;
      }
      $623 = ($621>>>0)<(11);
      if ($623) {
       $624 = $621&255;
       $625 = ((($78)) + 11|0);
       HEAP8[$625>>0] = $624;
       $$0$i$i664 = $78;
      } else {
       $626 = (($621) + 16)|0;
       $627 = $626 & -16;
       $628 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($627)|0);
       HEAP32[$78>>2] = $628;
       $629 = $627 | -2147483648;
       $630 = ((($78)) + 8|0);
       HEAP32[$630>>2] = $629;
       $631 = ((($78)) + 4|0);
       HEAP32[$631>>2] = $621;
       $$0$i$i664 = $628;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i664,159195,$621)|0);
      $632 = (($$0$i$i664) + ($621)|0);
      HEAP8[$26>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($632,$26);
      __THREW__ = 0;
      $633 = (invoke_iiiii(1140,($126|0),($1|0),($78|0),($2|0))|0);
      $634 = __THREW__; __THREW__ = 0;
      $635 = $634&1;
      if ($635) {
       $637 = ___cxa_find_matching_catch_2()|0;
       $638 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($78);
       $$21 = $637;$$21547 = $638;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($78);
       $636 = ($633|0)==($126|0);
       $$574 = $636 ? $0 : $633;
       $$14 = $$574;
       break L1;
      }
      break;
     }
     case 115:  {
      ;HEAP32[$79>>2]=0|0;HEAP32[$79+4>>2]=0|0;HEAP32[$79+8>>2]=0|0;
      $639 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159198)|0);
      $640 = ($639>>>0)>(4294967279);
      if ($640) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($79);
       // unreachable;
      }
      $641 = ($639>>>0)<(11);
      if ($641) {
       $642 = $639&255;
       $643 = ((($79)) + 11|0);
       HEAP8[$643>>0] = $642;
       $$0$i$i666 = $79;
      } else {
       $644 = (($639) + 16)|0;
       $645 = $644 & -16;
       $646 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($645)|0);
       HEAP32[$79>>2] = $646;
       $647 = $645 | -2147483648;
       $648 = ((($79)) + 8|0);
       HEAP32[$648>>2] = $647;
       $649 = ((($79)) + 4|0);
       HEAP32[$649>>2] = $639;
       $$0$i$i666 = $646;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i666,159198,$639)|0);
      $650 = (($$0$i$i666) + ($639)|0);
      HEAP8[$25>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($650,$25);
      __THREW__ = 0;
      $651 = (invoke_iiiii(1140,($126|0),($1|0),($79|0),($2|0))|0);
      $652 = __THREW__; __THREW__ = 0;
      $653 = $652&1;
      if ($653) {
       $655 = ___cxa_find_matching_catch_2()|0;
       $656 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($79);
       $$21 = $655;$$21547 = $656;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($79);
       $654 = ($651|0)==($126|0);
       $$575 = $654 ? $0 : $651;
       $$14 = $$575;
       break L1;
      }
      break;
     }
     case 83:  {
      ;HEAP32[$80>>2]=0|0;HEAP32[$80+4>>2]=0|0;HEAP32[$80+8>>2]=0|0;
      $657 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159201)|0);
      $658 = ($657>>>0)>(4294967279);
      if ($658) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($80);
       // unreachable;
      }
      $659 = ($657>>>0)<(11);
      if ($659) {
       $660 = $657&255;
       $661 = ((($80)) + 11|0);
       HEAP8[$661>>0] = $660;
       $$0$i$i668 = $80;
      } else {
       $662 = (($657) + 16)|0;
       $663 = $662 & -16;
       $664 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($663)|0);
       HEAP32[$80>>2] = $664;
       $665 = $663 | -2147483648;
       $666 = ((($80)) + 8|0);
       HEAP32[$666>>2] = $665;
       $667 = ((($80)) + 4|0);
       HEAP32[$667>>2] = $657;
       $$0$i$i668 = $664;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i668,159201,$657)|0);
      $668 = (($$0$i$i668) + ($657)|0);
      HEAP8[$24>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($668,$24);
      __THREW__ = 0;
      $669 = (invoke_iiiii(1140,($126|0),($1|0),($80|0),($2|0))|0);
      $670 = __THREW__; __THREW__ = 0;
      $671 = $670&1;
      if ($671) {
       $673 = ___cxa_find_matching_catch_2()|0;
       $674 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($80);
       $$21 = $673;$$21547 = $674;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($80);
       $672 = ($669|0)==($126|0);
       $$576 = $672 ? $0 : $669;
       $$14 = $$576;
       break L1;
      }
      break;
     }
     case 116:  {
      ;HEAP32[$81>>2]=0|0;HEAP32[$81+4>>2]=0|0;HEAP32[$81+8>>2]=0|0;
      $675 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159117)|0);
      $676 = ($675>>>0)>(4294967279);
      if ($676) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($81);
       // unreachable;
      }
      $677 = ($675>>>0)<(11);
      if ($677) {
       $678 = $675&255;
       $679 = ((($81)) + 11|0);
       HEAP8[$679>>0] = $678;
       $$0$i$i670 = $81;
      } else {
       $680 = (($675) + 16)|0;
       $681 = $680 & -16;
       $682 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($681)|0);
       HEAP32[$81>>2] = $682;
       $683 = $681 | -2147483648;
       $684 = ((($81)) + 8|0);
       HEAP32[$684>>2] = $683;
       $685 = ((($81)) + 4|0);
       HEAP32[$685>>2] = $675;
       $$0$i$i670 = $682;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i670,159117,$675)|0);
      $686 = (($$0$i$i670) + ($675)|0);
      HEAP8[$23>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($686,$23);
      __THREW__ = 0;
      $687 = (invoke_iiiii(1140,($126|0),($1|0),($81|0),($2|0))|0);
      $688 = __THREW__; __THREW__ = 0;
      $689 = $688&1;
      if ($689) {
       $691 = ___cxa_find_matching_catch_2()|0;
       $692 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($81);
       $$21 = $691;$$21547 = $692;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($81);
       $690 = ($687|0)==($126|0);
       $$577 = $690 ? $0 : $687;
       $$14 = $$577;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 109:  {
     $693 = HEAP8[$134>>0]|0;
     $694 = $693 << 24 >> 24;
     switch ($694|0) {
     case 105:  {
      ;HEAP32[$82>>2]=0|0;HEAP32[$82+4>>2]=0|0;HEAP32[$82+8>>2]=0|0;
      $695 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159205)|0);
      $696 = ($695>>>0)>(4294967279);
      if ($696) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($82);
       // unreachable;
      }
      $697 = ($695>>>0)<(11);
      if ($697) {
       $698 = $695&255;
       $699 = ((($82)) + 11|0);
       HEAP8[$699>>0] = $698;
       $$0$i$i672 = $82;
      } else {
       $700 = (($695) + 16)|0;
       $701 = $700 & -16;
       $702 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($701)|0);
       HEAP32[$82>>2] = $702;
       $703 = $701 | -2147483648;
       $704 = ((($82)) + 8|0);
       HEAP32[$704>>2] = $703;
       $705 = ((($82)) + 4|0);
       HEAP32[$705>>2] = $695;
       $$0$i$i672 = $702;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i672,159205,$695)|0);
      $706 = (($$0$i$i672) + ($695)|0);
      HEAP8[$22>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($706,$22);
      __THREW__ = 0;
      $707 = (invoke_iiiii(1140,($126|0),($1|0),($82|0),($2|0))|0);
      $708 = __THREW__; __THREW__ = 0;
      $709 = $708&1;
      if ($709) {
       $711 = ___cxa_find_matching_catch_2()|0;
       $712 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($82);
       $$21 = $711;$$21547 = $712;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($82);
       $710 = ($707|0)==($126|0);
       $$578 = $710 ? $0 : $707;
       $$14 = $$578;
       break L1;
      }
      break;
     }
     case 73:  {
      ;HEAP32[$83>>2]=0|0;HEAP32[$83+4>>2]=0|0;HEAP32[$83+8>>2]=0|0;
      $713 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159207)|0);
      $714 = ($713>>>0)>(4294967279);
      if ($714) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($83);
       // unreachable;
      }
      $715 = ($713>>>0)<(11);
      if ($715) {
       $716 = $713&255;
       $717 = ((($83)) + 11|0);
       HEAP8[$717>>0] = $716;
       $$0$i$i674 = $83;
      } else {
       $718 = (($713) + 16)|0;
       $719 = $718 & -16;
       $720 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($719)|0);
       HEAP32[$83>>2] = $720;
       $721 = $719 | -2147483648;
       $722 = ((($83)) + 8|0);
       HEAP32[$722>>2] = $721;
       $723 = ((($83)) + 4|0);
       HEAP32[$723>>2] = $713;
       $$0$i$i674 = $720;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i674,159207,$713)|0);
      $724 = (($$0$i$i674) + ($713)|0);
      HEAP8[$21>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($724,$21);
      __THREW__ = 0;
      $725 = (invoke_iiiii(1140,($126|0),($1|0),($83|0),($2|0))|0);
      $726 = __THREW__; __THREW__ = 0;
      $727 = $726&1;
      if ($727) {
       $729 = ___cxa_find_matching_catch_2()|0;
       $730 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($83);
       $$21 = $729;$$21547 = $730;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($83);
       $728 = ($725|0)==($126|0);
       $$579 = $728 ? $0 : $725;
       $$14 = $$579;
       break L1;
      }
      break;
     }
     case 108:  {
      ;HEAP32[$84>>2]=0|0;HEAP32[$84+4>>2]=0|0;HEAP32[$84+8>>2]=0|0;
      $731 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159101)|0);
      $732 = ($731>>>0)>(4294967279);
      if ($732) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($84);
       // unreachable;
      }
      $733 = ($731>>>0)<(11);
      if ($733) {
       $734 = $731&255;
       $735 = ((($84)) + 11|0);
       HEAP8[$735>>0] = $734;
       $$0$i$i676 = $84;
      } else {
       $736 = (($731) + 16)|0;
       $737 = $736 & -16;
       $738 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($737)|0);
       HEAP32[$84>>2] = $738;
       $739 = $737 | -2147483648;
       $740 = ((($84)) + 8|0);
       HEAP32[$740>>2] = $739;
       $741 = ((($84)) + 4|0);
       HEAP32[$741>>2] = $731;
       $$0$i$i676 = $738;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i676,159101,$731)|0);
      $742 = (($$0$i$i676) + ($731)|0);
      HEAP8[$20>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($742,$20);
      __THREW__ = 0;
      $743 = (invoke_iiiii(1140,($126|0),($1|0),($84|0),($2|0))|0);
      $744 = __THREW__; __THREW__ = 0;
      $745 = $744&1;
      if ($745) {
       $747 = ___cxa_find_matching_catch_2()|0;
       $748 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($84);
       $$21 = $747;$$21547 = $748;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($84);
       $746 = ($743|0)==($126|0);
       $$580 = $746 ? $0 : $743;
       $$14 = $$580;
       break L1;
      }
      break;
     }
     case 76:  {
      ;HEAP32[$85>>2]=0|0;HEAP32[$85+4>>2]=0|0;HEAP32[$85+8>>2]=0|0;
      $749 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159210)|0);
      $750 = ($749>>>0)>(4294967279);
      if ($750) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($85);
       // unreachable;
      }
      $751 = ($749>>>0)<(11);
      if ($751) {
       $752 = $749&255;
       $753 = ((($85)) + 11|0);
       HEAP8[$753>>0] = $752;
       $$0$i$i678 = $85;
      } else {
       $754 = (($749) + 16)|0;
       $755 = $754 & -16;
       $756 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($755)|0);
       HEAP32[$85>>2] = $756;
       $757 = $755 | -2147483648;
       $758 = ((($85)) + 8|0);
       HEAP32[$758>>2] = $757;
       $759 = ((($85)) + 4|0);
       HEAP32[$759>>2] = $749;
       $$0$i$i678 = $756;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i678,159210,$749)|0);
      $760 = (($$0$i$i678) + ($749)|0);
      HEAP8[$19>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($760,$19);
      __THREW__ = 0;
      $761 = (invoke_iiiii(1140,($126|0),($1|0),($85|0),($2|0))|0);
      $762 = __THREW__; __THREW__ = 0;
      $763 = $762&1;
      if ($763) {
       $765 = ___cxa_find_matching_catch_2()|0;
       $766 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($85);
       $$21 = $765;$$21547 = $766;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($85);
       $764 = ($761|0)==($126|0);
       $$581 = $764 ? $0 : $761;
       $$14 = $$581;
       break L1;
      }
      break;
     }
     case 109:  {
      if ($136) {
       $767 = HEAP8[$126>>0]|0;
       $768 = ($767<<24>>24)==(95);
       if ($768) {
        ;HEAP32[$86>>2]=0|0;HEAP32[$86+4>>2]=0|0;HEAP32[$86+8>>2]=0|0;
        $769 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159213)|0);
        $770 = ($769>>>0)>(4294967279);
        if ($770) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($86);
         // unreachable;
        }
        $771 = ($769>>>0)<(11);
        if ($771) {
         $772 = $769&255;
         $773 = ((($86)) + 11|0);
         HEAP8[$773>>0] = $772;
         $$0$i$i680 = $86;
        } else {
         $774 = (($769) + 16)|0;
         $775 = $774 & -16;
         $776 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($775)|0);
         HEAP32[$86>>2] = $776;
         $777 = $775 | -2147483648;
         $778 = ((($86)) + 8|0);
         HEAP32[$778>>2] = $777;
         $779 = ((($86)) + 4|0);
         HEAP32[$779>>2] = $769;
         $$0$i$i680 = $776;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i680,159213,$769)|0);
        $780 = (($$0$i$i680) + ($769)|0);
        HEAP8[$18>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($780,$18);
        __THREW__ = 0;
        $781 = (invoke_iiiii(1141,($135|0),($1|0),($86|0),($2|0))|0);
        $782 = __THREW__; __THREW__ = 0;
        $783 = $782&1;
        if ($783) {
         $785 = ___cxa_find_matching_catch_2()|0;
         $786 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($86);
         $$21 = $785;$$21547 = $786;
         break L7;
        } else {
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($86);
         $784 = ($781|0)==($135|0);
         $$582 = $784 ? $0 : $781;
         $$14 = $$582;
         break L1;
        }
       }
      }
      $787 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($126,$1,$2)|0);
      $788 = ($787|0)==($126|0);
      if ($788) {
       $$14 = $0;
       break L1;
      }
      $789 = HEAP32[$2>>2]|0;
      $790 = ((($2)) + 4|0);
      $791 = HEAP32[$790>>2]|0;
      $792 = ($789|0)==($791|0);
      if ($792) {
       $$14 = $0;
       break L1;
      }
      $793 = ((($791)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($90,$793);
      __THREW__ = 0;
      $794 = (invoke_iiii(1121,($90|0),0,(159083|0))|0);
      $795 = __THREW__; __THREW__ = 0;
      $796 = $795&1;
      do {
       if ($796) {
        $806 = ___cxa_find_matching_catch_2()|0;
        $807 = tempRet0;
        $$10515 = $806;$$10536 = $807;
       } else {
        ;HEAP32[$89>>2]=HEAP32[$90>>2]|0;HEAP32[$89+4>>2]=HEAP32[$90+4>>2]|0;HEAP32[$89+8>>2]=HEAP32[$90+8>>2]|0;
        $$0$i$i$i682 = 0;
        while(1) {
         $exitcond$i$i$i683 = ($$0$i$i$i682|0)==(3);
         if ($exitcond$i$i$i683) {
          break;
         }
         $797 = (($90) + ($$0$i$i$i682<<2)|0);
         HEAP32[$797>>2] = 0;
         $798 = (($$0$i$i$i682) + 1)|0;
         $$0$i$i$i682 = $798;
        }
        __THREW__ = 0;
        $799 = (invoke_iii(1119,($89|0),(159216|0))|0);
        $800 = __THREW__; __THREW__ = 0;
        $801 = $800&1;
        if ($801) {
         $808 = ___cxa_find_matching_catch_2()|0;
         $809 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($89);
         $$10515 = $808;$$10536 = $809;
         break;
        }
        ;HEAP32[$88>>2]=HEAP32[$89>>2]|0;HEAP32[$88+4>>2]=HEAP32[$89+4>>2]|0;HEAP32[$88+8>>2]=HEAP32[$89+8>>2]|0;
        $$0$i$i$i685 = 0;
        while(1) {
         $exitcond$i$i$i686 = ($$0$i$i$i685|0)==(3);
         if ($exitcond$i$i$i686) {
          break;
         }
         $802 = (($89) + ($$0$i$i$i685<<2)|0);
         HEAP32[$802>>2] = 0;
         $803 = (($$0$i$i$i685) + 1)|0;
         $$0$i$i$i685 = $803;
        }
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($87,$88);
        $804 = HEAP32[$790>>2]|0;
        $805 = ((($804)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($805,$87);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($87);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($88);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($89);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($90);
        $$14 = $787;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($90);
      $$21 = $$10515;$$21547 = $$10536;
      break L7;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 110:  {
     $810 = HEAP8[$134>>0]|0;
     $811 = $810 << 24 >> 24;
     switch ($811|0) {
     case 119: case 97:  {
      $812 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_new_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $812;
      break L1;
      break;
     }
     case 101:  {
      ;HEAP32[$91>>2]=0|0;HEAP32[$91+4>>2]=0|0;HEAP32[$91+8>>2]=0|0;
      $813 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159220)|0);
      $814 = ($813>>>0)>(4294967279);
      if ($814) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($91);
       // unreachable;
      }
      $815 = ($813>>>0)<(11);
      if ($815) {
       $816 = $813&255;
       $817 = ((($91)) + 11|0);
       HEAP8[$817>>0] = $816;
       $$0$i$i689 = $91;
      } else {
       $818 = (($813) + 16)|0;
       $819 = $818 & -16;
       $820 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($819)|0);
       HEAP32[$91>>2] = $820;
       $821 = $819 | -2147483648;
       $822 = ((($91)) + 8|0);
       HEAP32[$822>>2] = $821;
       $823 = ((($91)) + 4|0);
       HEAP32[$823>>2] = $813;
       $$0$i$i689 = $820;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i689,159220,$813)|0);
      $824 = (($$0$i$i689) + ($813)|0);
      HEAP8[$17>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($824,$17);
      __THREW__ = 0;
      $825 = (invoke_iiiii(1140,($126|0),($1|0),($91|0),($2|0))|0);
      $826 = __THREW__; __THREW__ = 0;
      $827 = $826&1;
      if ($827) {
       $829 = ___cxa_find_matching_catch_2()|0;
       $830 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($91);
       $$21 = $829;$$21547 = $830;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($91);
       $828 = ($825|0)==($126|0);
       $$583 = $828 ? $0 : $825;
       $$14 = $$583;
       break L1;
      }
      break;
     }
     case 103:  {
      ;HEAP32[$92>>2]=0|0;HEAP32[$92+4>>2]=0|0;HEAP32[$92+8>>2]=0|0;
      $831 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159205)|0);
      $832 = ($831>>>0)>(4294967279);
      if ($832) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($92);
       // unreachable;
      }
      $833 = ($831>>>0)<(11);
      if ($833) {
       $834 = $831&255;
       $835 = ((($92)) + 11|0);
       HEAP8[$835>>0] = $834;
       $$0$i$i691 = $92;
      } else {
       $836 = (($831) + 16)|0;
       $837 = $836 & -16;
       $838 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($837)|0);
       HEAP32[$92>>2] = $838;
       $839 = $837 | -2147483648;
       $840 = ((($92)) + 8|0);
       HEAP32[$840>>2] = $839;
       $841 = ((($92)) + 4|0);
       HEAP32[$841>>2] = $831;
       $$0$i$i691 = $838;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i691,159205,$831)|0);
      $842 = (($$0$i$i691) + ($831)|0);
      HEAP8[$16>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($842,$16);
      __THREW__ = 0;
      $843 = (invoke_iiiii(1141,($126|0),($1|0),($92|0),($2|0))|0);
      $844 = __THREW__; __THREW__ = 0;
      $845 = $844&1;
      if ($845) {
       $847 = ___cxa_find_matching_catch_2()|0;
       $848 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($92);
       $$21 = $847;$$21547 = $848;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($92);
       $846 = ($843|0)==($126|0);
       $$584 = $846 ? $0 : $843;
       $$14 = $$584;
       break L1;
      }
      break;
     }
     case 116:  {
      ;HEAP32[$93>>2]=0|0;HEAP32[$93+4>>2]=0|0;HEAP32[$93+8>>2]=0|0;
      $849 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159223)|0);
      $850 = ($849>>>0)>(4294967279);
      if ($850) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($93);
       // unreachable;
      }
      $851 = ($849>>>0)<(11);
      if ($851) {
       $852 = $849&255;
       $853 = ((($93)) + 11|0);
       HEAP8[$853>>0] = $852;
       $$0$i$i693 = $93;
      } else {
       $854 = (($849) + 16)|0;
       $855 = $854 & -16;
       $856 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($855)|0);
       HEAP32[$93>>2] = $856;
       $857 = $855 | -2147483648;
       $858 = ((($93)) + 8|0);
       HEAP32[$858>>2] = $857;
       $859 = ((($93)) + 4|0);
       HEAP32[$859>>2] = $849;
       $$0$i$i693 = $856;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i693,159223,$849)|0);
      $860 = (($$0$i$i693) + ($849)|0);
      HEAP8[$15>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($860,$15);
      __THREW__ = 0;
      $861 = (invoke_iiiii(1141,($126|0),($1|0),($93|0),($2|0))|0);
      $862 = __THREW__; __THREW__ = 0;
      $863 = $862&1;
      if ($863) {
       $865 = ___cxa_find_matching_catch_2()|0;
       $866 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($93);
       $$21 = $865;$$21547 = $866;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($93);
       $864 = ($861|0)==($126|0);
       $$585 = $864 ? $0 : $861;
       $$14 = $$585;
       break L1;
      }
      break;
     }
     case 120:  {
      $867 = (__ZN10__cxxabiv112_GLOBAL__N_125parse_noexcept_expressionINS0_2DbEEEPKcS4_S4_RT_($126,$1,$2)|0);
      $868 = ($867|0)==($126|0);
      $$586 = $868 ? $0 : $867;
      $$14 = $$586;
      break L1;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 111:  {
     $869 = HEAP8[$134>>0]|0;
     $870 = $869 << 24 >> 24;
     switch ($870|0) {
     case 110:  {
      $871 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $871;
      break L1;
      break;
     }
     case 111:  {
      ;HEAP32[$94>>2]=0|0;HEAP32[$94+4>>2]=0|0;HEAP32[$94+8>>2]=0|0;
      $872 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159225)|0);
      $873 = ($872>>>0)>(4294967279);
      if ($873) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($94);
       // unreachable;
      }
      $874 = ($872>>>0)<(11);
      if ($874) {
       $875 = $872&255;
       $876 = ((($94)) + 11|0);
       HEAP8[$876>>0] = $875;
       $$0$i$i695 = $94;
      } else {
       $877 = (($872) + 16)|0;
       $878 = $877 & -16;
       $879 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($878)|0);
       HEAP32[$94>>2] = $879;
       $880 = $878 | -2147483648;
       $881 = ((($94)) + 8|0);
       HEAP32[$881>>2] = $880;
       $882 = ((($94)) + 4|0);
       HEAP32[$882>>2] = $872;
       $$0$i$i695 = $879;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i695,159225,$872)|0);
      $883 = (($$0$i$i695) + ($872)|0);
      HEAP8[$14>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($883,$14);
      __THREW__ = 0;
      $884 = (invoke_iiiii(1140,($126|0),($1|0),($94|0),($2|0))|0);
      $885 = __THREW__; __THREW__ = 0;
      $886 = $885&1;
      if ($886) {
       $888 = ___cxa_find_matching_catch_2()|0;
       $889 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($94);
       $$21 = $888;$$21547 = $889;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($94);
       $887 = ($884|0)==($126|0);
       $$587 = $887 ? $0 : $884;
       $$14 = $$587;
       break L1;
      }
      break;
     }
     case 114:  {
      ;HEAP32[$95>>2]=0|0;HEAP32[$95+4>>2]=0|0;HEAP32[$95+8>>2]=0|0;
      $890 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159228)|0);
      $891 = ($890>>>0)>(4294967279);
      if ($891) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($95);
       // unreachable;
      }
      $892 = ($890>>>0)<(11);
      if ($892) {
       $893 = $890&255;
       $894 = ((($95)) + 11|0);
       HEAP8[$894>>0] = $893;
       $$0$i$i697 = $95;
      } else {
       $895 = (($890) + 16)|0;
       $896 = $895 & -16;
       $897 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($896)|0);
       HEAP32[$95>>2] = $897;
       $898 = $896 | -2147483648;
       $899 = ((($95)) + 8|0);
       HEAP32[$899>>2] = $898;
       $900 = ((($95)) + 4|0);
       HEAP32[$900>>2] = $890;
       $$0$i$i697 = $897;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i697,159228,$890)|0);
      $901 = (($$0$i$i697) + ($890)|0);
      HEAP8[$13>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($901,$13);
      __THREW__ = 0;
      $902 = (invoke_iiiii(1140,($126|0),($1|0),($95|0),($2|0))|0);
      $903 = __THREW__; __THREW__ = 0;
      $904 = $903&1;
      if ($904) {
       $906 = ___cxa_find_matching_catch_2()|0;
       $907 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($95);
       $$21 = $906;$$21547 = $907;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($95);
       $905 = ($902|0)==($126|0);
       $$588 = $905 ? $0 : $902;
       $$14 = $$588;
       break L1;
      }
      break;
     }
     case 82:  {
      ;HEAP32[$96>>2]=0|0;HEAP32[$96+4>>2]=0|0;HEAP32[$96+8>>2]=0|0;
      $908 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159230)|0);
      $909 = ($908>>>0)>(4294967279);
      if ($909) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($96);
       // unreachable;
      }
      $910 = ($908>>>0)<(11);
      if ($910) {
       $911 = $908&255;
       $912 = ((($96)) + 11|0);
       HEAP8[$912>>0] = $911;
       $$0$i$i699 = $96;
      } else {
       $913 = (($908) + 16)|0;
       $914 = $913 & -16;
       $915 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($914)|0);
       HEAP32[$96>>2] = $915;
       $916 = $914 | -2147483648;
       $917 = ((($96)) + 8|0);
       HEAP32[$917>>2] = $916;
       $918 = ((($96)) + 4|0);
       HEAP32[$918>>2] = $908;
       $$0$i$i699 = $915;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i699,159230,$908)|0);
      $919 = (($$0$i$i699) + ($908)|0);
      HEAP8[$12>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($919,$12);
      __THREW__ = 0;
      $920 = (invoke_iiiii(1140,($126|0),($1|0),($96|0),($2|0))|0);
      $921 = __THREW__; __THREW__ = 0;
      $922 = $921&1;
      if ($922) {
       $924 = ___cxa_find_matching_catch_2()|0;
       $925 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($96);
       $$21 = $924;$$21547 = $925;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($96);
       $923 = ($920|0)==($126|0);
       $$589 = $923 ? $0 : $920;
       $$14 = $$589;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 112:  {
     $926 = HEAP8[$134>>0]|0;
     $927 = $926 << 24 >> 24;
     switch ($927|0) {
     case 109:  {
      ;HEAP32[$97>>2]=0|0;HEAP32[$97+4>>2]=0|0;HEAP32[$97+8>>2]=0|0;
      $928 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159233)|0);
      $929 = ($928>>>0)>(4294967279);
      if ($929) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($97);
       // unreachable;
      }
      $930 = ($928>>>0)<(11);
      if ($930) {
       $931 = $928&255;
       $932 = ((($97)) + 11|0);
       HEAP8[$932>>0] = $931;
       $$0$i$i701 = $97;
      } else {
       $933 = (($928) + 16)|0;
       $934 = $933 & -16;
       $935 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($934)|0);
       HEAP32[$97>>2] = $935;
       $936 = $934 | -2147483648;
       $937 = ((($97)) + 8|0);
       HEAP32[$937>>2] = $936;
       $938 = ((($97)) + 4|0);
       HEAP32[$938>>2] = $928;
       $$0$i$i701 = $935;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i701,159233,$928)|0);
      $939 = (($$0$i$i701) + ($928)|0);
      HEAP8[$11>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($939,$11);
      __THREW__ = 0;
      $940 = (invoke_iiiii(1140,($126|0),($1|0),($97|0),($2|0))|0);
      $941 = __THREW__; __THREW__ = 0;
      $942 = $941&1;
      if ($942) {
       $944 = ___cxa_find_matching_catch_2()|0;
       $945 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($97);
       $$21 = $944;$$21547 = $945;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($97);
       $943 = ($940|0)==($126|0);
       $$590 = $943 ? $0 : $940;
       $$14 = $$590;
       break L1;
      }
      break;
     }
     case 108:  {
      ;HEAP32[$98>>2]=0|0;HEAP32[$98+4>>2]=0|0;HEAP32[$98+8>>2]=0|0;
      $946 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159237)|0);
      $947 = ($946>>>0)>(4294967279);
      if ($947) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($98);
       // unreachable;
      }
      $948 = ($946>>>0)<(11);
      if ($948) {
       $949 = $946&255;
       $950 = ((($98)) + 11|0);
       HEAP8[$950>>0] = $949;
       $$0$i$i703 = $98;
      } else {
       $951 = (($946) + 16)|0;
       $952 = $951 & -16;
       $953 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($952)|0);
       HEAP32[$98>>2] = $953;
       $954 = $952 | -2147483648;
       $955 = ((($98)) + 8|0);
       HEAP32[$955>>2] = $954;
       $956 = ((($98)) + 4|0);
       HEAP32[$956>>2] = $946;
       $$0$i$i703 = $953;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i703,159237,$946)|0);
      $957 = (($$0$i$i703) + ($946)|0);
      HEAP8[$10>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($957,$10);
      __THREW__ = 0;
      $958 = (invoke_iiiii(1140,($126|0),($1|0),($98|0),($2|0))|0);
      $959 = __THREW__; __THREW__ = 0;
      $960 = $959&1;
      if ($960) {
       $962 = ___cxa_find_matching_catch_2()|0;
       $963 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($98);
       $$21 = $962;$$21547 = $963;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($98);
       $961 = ($958|0)==($126|0);
       $$591 = $961 ? $0 : $958;
       $$14 = $$591;
       break L1;
      }
      break;
     }
     case 76:  {
      ;HEAP32[$99>>2]=0|0;HEAP32[$99+4>>2]=0|0;HEAP32[$99+8>>2]=0|0;
      $964 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159239)|0);
      $965 = ($964>>>0)>(4294967279);
      if ($965) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($99);
       // unreachable;
      }
      $966 = ($964>>>0)<(11);
      if ($966) {
       $967 = $964&255;
       $968 = ((($99)) + 11|0);
       HEAP8[$968>>0] = $967;
       $$0$i$i705 = $99;
      } else {
       $969 = (($964) + 16)|0;
       $970 = $969 & -16;
       $971 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($970)|0);
       HEAP32[$99>>2] = $971;
       $972 = $970 | -2147483648;
       $973 = ((($99)) + 8|0);
       HEAP32[$973>>2] = $972;
       $974 = ((($99)) + 4|0);
       HEAP32[$974>>2] = $964;
       $$0$i$i705 = $971;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i705,159239,$964)|0);
      $975 = (($$0$i$i705) + ($964)|0);
      HEAP8[$9>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($975,$9);
      __THREW__ = 0;
      $976 = (invoke_iiiii(1140,($126|0),($1|0),($99|0),($2|0))|0);
      $977 = __THREW__; __THREW__ = 0;
      $978 = $977&1;
      if ($978) {
       $980 = ___cxa_find_matching_catch_2()|0;
       $981 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($99);
       $$21 = $980;$$21547 = $981;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($99);
       $979 = ($976|0)==($126|0);
       $$592 = $979 ? $0 : $976;
       $$14 = $$592;
       break L1;
      }
      break;
     }
     case 112:  {
      if ($136) {
       $982 = HEAP8[$126>>0]|0;
       $983 = ($982<<24>>24)==(95);
       if ($983) {
        ;HEAP32[$100>>2]=0|0;HEAP32[$100+4>>2]=0|0;HEAP32[$100+8>>2]=0|0;
        $984 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159242)|0);
        $985 = ($984>>>0)>(4294967279);
        if ($985) {
         __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($100);
         // unreachable;
        }
        $986 = ($984>>>0)<(11);
        if ($986) {
         $987 = $984&255;
         $988 = ((($100)) + 11|0);
         HEAP8[$988>>0] = $987;
         $$0$i$i707 = $100;
        } else {
         $989 = (($984) + 16)|0;
         $990 = $989 & -16;
         $991 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($990)|0);
         HEAP32[$100>>2] = $991;
         $992 = $990 | -2147483648;
         $993 = ((($100)) + 8|0);
         HEAP32[$993>>2] = $992;
         $994 = ((($100)) + 4|0);
         HEAP32[$994>>2] = $984;
         $$0$i$i707 = $991;
        }
        (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i707,159242,$984)|0);
        $995 = (($$0$i$i707) + ($984)|0);
        HEAP8[$8>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($995,$8);
        __THREW__ = 0;
        $996 = (invoke_iiiii(1141,($135|0),($1|0),($100|0),($2|0))|0);
        $997 = __THREW__; __THREW__ = 0;
        $998 = $997&1;
        if ($998) {
         $1000 = ___cxa_find_matching_catch_2()|0;
         $1001 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($100);
         $$21 = $1000;$$21547 = $1001;
         break L7;
        } else {
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($100);
         $999 = ($996|0)==($135|0);
         $$593 = $999 ? $0 : $996;
         $$14 = $$593;
         break L1;
        }
       }
      }
      $1002 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($126,$1,$2)|0);
      $1003 = ($1002|0)==($126|0);
      if ($1003) {
       $$14 = $0;
       break L1;
      }
      $1004 = HEAP32[$2>>2]|0;
      $1005 = ((($2)) + 4|0);
      $1006 = HEAP32[$1005>>2]|0;
      $1007 = ($1004|0)==($1006|0);
      if ($1007) {
       $$14 = $0;
       break L1;
      }
      $1008 = ((($1006)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($104,$1008);
      __THREW__ = 0;
      $1009 = (invoke_iiii(1121,($104|0),0,(159083|0))|0);
      $1010 = __THREW__; __THREW__ = 0;
      $1011 = $1010&1;
      do {
       if ($1011) {
        $1021 = ___cxa_find_matching_catch_2()|0;
        $1022 = tempRet0;
        $$12517 = $1021;$$12538 = $1022;
       } else {
        ;HEAP32[$103>>2]=HEAP32[$104>>2]|0;HEAP32[$103+4>>2]=HEAP32[$104+4>>2]|0;HEAP32[$103+8>>2]=HEAP32[$104+8>>2]|0;
        $$0$i$i$i709 = 0;
        while(1) {
         $exitcond$i$i$i710 = ($$0$i$i$i709|0)==(3);
         if ($exitcond$i$i$i710) {
          break;
         }
         $1012 = (($104) + ($$0$i$i$i709<<2)|0);
         HEAP32[$1012>>2] = 0;
         $1013 = (($$0$i$i$i709) + 1)|0;
         $$0$i$i$i709 = $1013;
        }
        __THREW__ = 0;
        $1014 = (invoke_iii(1119,($103|0),(159245|0))|0);
        $1015 = __THREW__; __THREW__ = 0;
        $1016 = $1015&1;
        if ($1016) {
         $1023 = ___cxa_find_matching_catch_2()|0;
         $1024 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($103);
         $$12517 = $1023;$$12538 = $1024;
         break;
        }
        ;HEAP32[$102>>2]=HEAP32[$103>>2]|0;HEAP32[$102+4>>2]=HEAP32[$103+4>>2]|0;HEAP32[$102+8>>2]=HEAP32[$103+8>>2]|0;
        $$0$i$i$i713 = 0;
        while(1) {
         $exitcond$i$i$i714 = ($$0$i$i$i713|0)==(3);
         if ($exitcond$i$i$i714) {
          break;
         }
         $1017 = (($103) + ($$0$i$i$i713<<2)|0);
         HEAP32[$1017>>2] = 0;
         $1018 = (($$0$i$i$i713) + 1)|0;
         $$0$i$i$i713 = $1018;
        }
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($101,$102);
        $1019 = HEAP32[$1005>>2]|0;
        $1020 = ((($1019)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($1020,$101);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($101);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($102);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($103);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($104);
        $$14 = $1002;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($104);
      $$21 = $$12517;$$21547 = $$12538;
      break L7;
      break;
     }
     case 115:  {
      ;HEAP32[$105>>2]=0|0;HEAP32[$105+4>>2]=0|0;HEAP32[$105+8>>2]=0|0;
      $1025 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159237)|0);
      $1026 = ($1025>>>0)>(4294967279);
      if ($1026) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($105);
       // unreachable;
      }
      $1027 = ($1025>>>0)<(11);
      if ($1027) {
       $1028 = $1025&255;
       $1029 = ((($105)) + 11|0);
       HEAP8[$1029>>0] = $1028;
       $$0$i$i717 = $105;
      } else {
       $1030 = (($1025) + 16)|0;
       $1031 = $1030 & -16;
       $1032 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($1031)|0);
       HEAP32[$105>>2] = $1032;
       $1033 = $1031 | -2147483648;
       $1034 = ((($105)) + 8|0);
       HEAP32[$1034>>2] = $1033;
       $1035 = ((($105)) + 4|0);
       HEAP32[$1035>>2] = $1025;
       $$0$i$i717 = $1032;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i717,159237,$1025)|0);
      $1036 = (($$0$i$i717) + ($1025)|0);
      HEAP8[$7>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($1036,$7);
      __THREW__ = 0;
      $1037 = (invoke_iiiii(1141,($126|0),($1|0),($105|0),($2|0))|0);
      $1038 = __THREW__; __THREW__ = 0;
      $1039 = $1038&1;
      if ($1039) {
       $1041 = ___cxa_find_matching_catch_2()|0;
       $1042 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($105);
       $$21 = $1041;$$21547 = $1042;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($105);
       $1040 = ($1037|0)==($126|0);
       $$594 = $1040 ? $0 : $1037;
       $$14 = $$594;
       break L1;
      }
      break;
     }
     case 116:  {
      $1043 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_arrow_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1043;
      break L1;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 113:  {
     $1044 = HEAP8[$134>>0]|0;
     $1045 = ($1044<<24>>24)==(117);
     if (!($1045)) {
      $$14 = $0;
      break L1;
     }
     $1046 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($126,$1,$2)|0);
     $1047 = ($1046|0)==($126|0);
     if ($1047) {
      $$14 = $0;
      break L1;
     }
     $1048 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($1046,$1,$2)|0);
     $1049 = ($1048|0)==($1046|0);
     if ($1049) {
      $1148 = ((($2)) + 4|0);
      $1149 = HEAP32[$1148>>2]|0;
      $1150 = ((($1149)) + -24|0);
      $1152 = $1149;
      while(1) {
       $1151 = ($1152|0)==($1150|0);
       if ($1151) {
        $$14 = $0;
        break L1;
       }
       $1153 = ((($1152)) + -24|0);
       HEAP32[$1148>>2] = $1153;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($1153);
       $$pre$i747 = HEAP32[$1148>>2]|0;
       $1152 = $$pre$i747;
      }
     }
     $1050 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($1048,$1,$2)|0);
     $1051 = ($1050|0)==($1048|0);
     $1052 = ((($2)) + 4|0);
     if ($1051) {
      $1139 = HEAP32[$1052>>2]|0;
      $1140 = ((($1139)) + -24|0);
      $1142 = $1139;
      while(1) {
       $1141 = ($1142|0)==($1140|0);
       if ($1141) {
        break;
       }
       $1143 = ((($1142)) + -24|0);
       HEAP32[$1052>>2] = $1143;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($1143);
       $$pre$i743 = HEAP32[$1052>>2]|0;
       $1142 = $$pre$i743;
      }
      $1144 = ((($1139)) + -48|0);
      $1146 = $1140;
      while(1) {
       $1145 = ($1146|0)==($1144|0);
       if ($1145) {
        $$14 = $0;
        break L1;
       }
       $1147 = ((($1146)) + -24|0);
       HEAP32[$1052>>2] = $1147;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($1147);
       $$pre$i745 = HEAP32[$1052>>2]|0;
       $1146 = $$pre$i745;
      }
     }
     $1053 = HEAP32[$1052>>2]|0;
     $1054 = HEAP32[$2>>2]|0;
     $1055 = (($1053) - ($1054))|0;
     $1056 = (($1055|0) / 24)&-1;
     $1057 = ($1056>>>0)<(3);
     if ($1057) {
      $$14 = $0;
      break L1;
     }
     $1058 = $1053;
     $1059 = ((($1058)) + -24|0);
     __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($106,$1059);
     $1060 = HEAP32[$1052>>2]|0;
     $1061 = ((($1060)) + -24|0);
     $1063 = $1060;
     while(1) {
      $1062 = ($1063|0)==($1061|0);
      if ($1062) {
       break;
      }
      $1064 = ((($1063)) + -24|0);
      HEAP32[$1052>>2] = $1064;
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($1064);
      $$pre$i719 = HEAP32[$1052>>2]|0;
      $1063 = $$pre$i719;
     }
     $1065 = ((($1060)) + -48|0);
     __THREW__ = 0;
     invoke_vii(1120,($107|0),($1065|0));
     $1066 = __THREW__; __THREW__ = 0;
     $1067 = $1066&1;
     if ($1067) {
      $1123 = ___cxa_find_matching_catch_2()|0;
      $1124 = tempRet0;
      $$20525 = $1123;$$20546 = $1124;
     } else {
      $1068 = HEAP32[$1052>>2]|0;
      $1069 = ((($1068)) + -24|0);
      $1071 = $1068;
      while(1) {
       $1070 = ($1071|0)==($1069|0);
       if ($1070) {
        break;
       }
       $1072 = ((($1071)) + -24|0);
       HEAP32[$1052>>2] = $1072;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($1072);
       $$pre$i721 = HEAP32[$1052>>2]|0;
       $1071 = $$pre$i721;
      }
      $1073 = ((($1068)) + -48|0);
      __THREW__ = 0;
      invoke_vii(1120,($108|0),($1073|0));
      $1074 = __THREW__; __THREW__ = 0;
      $1075 = $1074&1;
      if ($1075) {
       $1125 = ___cxa_find_matching_catch_2()|0;
       $1126 = tempRet0;
       $$19524 = $1125;$$19545 = $1126;
      } else {
       __THREW__ = 0;
       invoke_viii(1139,($115|0),(159083|0),($108|0));
       $1076 = __THREW__; __THREW__ = 0;
       $1077 = $1076&1;
       if ($1077) {
        $1127 = ___cxa_find_matching_catch_2()|0;
        $1128 = tempRet0;
        $$18523 = $1127;$$18544 = $1128;
       } else {
        __THREW__ = 0;
        $1078 = (invoke_iii(1119,($115|0),(159249|0))|0);
        $1079 = __THREW__; __THREW__ = 0;
        $1080 = $1079&1;
        if ($1080) {
         $1129 = ___cxa_find_matching_catch_2()|0;
         $1130 = tempRet0;
         $$17522 = $1129;$$17543 = $1130;
        } else {
         ;HEAP32[$114>>2]=HEAP32[$115>>2]|0;HEAP32[$114+4>>2]=HEAP32[$115+4>>2]|0;HEAP32[$114+8>>2]=HEAP32[$115+8>>2]|0;
         $$0$i$i$i723 = 0;
         while(1) {
          $exitcond$i$i$i724 = ($$0$i$i$i723|0)==(3);
          if ($exitcond$i$i$i724) {
           break;
          }
          $1081 = (($115) + ($$0$i$i$i723<<2)|0);
          HEAP32[$1081>>2] = 0;
          $1082 = (($$0$i$i$i723) + 1)|0;
          $$0$i$i$i723 = $1082;
         }
         $1083 = ((($107)) + 11|0);
         $1084 = HEAP8[$1083>>0]|0;
         $1085 = ($1084<<24>>24)<(0);
         $1086 = HEAP32[$107>>2]|0;
         $1087 = $1085 ? $1086 : $107;
         $1088 = ((($107)) + 4|0);
         $1089 = HEAP32[$1088>>2]|0;
         $1090 = $1084&255;
         $1091 = $1085 ? $1089 : $1090;
         __THREW__ = 0;
         $1092 = (invoke_iiii(1112,($114|0),($1087|0),($1091|0))|0);
         $1093 = __THREW__; __THREW__ = 0;
         $1094 = $1093&1;
         if ($1094) {
          $1131 = ___cxa_find_matching_catch_2()|0;
          $1132 = tempRet0;
          $$16521 = $1131;$$16542 = $1132;
         } else {
          ;HEAP32[$113>>2]=HEAP32[$114>>2]|0;HEAP32[$113+4>>2]=HEAP32[$114+4>>2]|0;HEAP32[$113+8>>2]=HEAP32[$114+8>>2]|0;
          $$0$i$i$i727 = 0;
          while(1) {
           $exitcond$i$i$i728 = ($$0$i$i$i727|0)==(3);
           if ($exitcond$i$i$i728) {
            break;
           }
           $1095 = (($114) + ($$0$i$i$i727<<2)|0);
           HEAP32[$1095>>2] = 0;
           $1096 = (($$0$i$i$i727) + 1)|0;
           $$0$i$i$i727 = $1096;
          }
          __THREW__ = 0;
          $1097 = (invoke_iii(1119,($113|0),(159255|0))|0);
          $1098 = __THREW__; __THREW__ = 0;
          $1099 = $1098&1;
          if ($1099) {
           $1133 = ___cxa_find_matching_catch_2()|0;
           $1134 = tempRet0;
           $$15520 = $1133;$$15541 = $1134;
          } else {
           ;HEAP32[$112>>2]=HEAP32[$113>>2]|0;HEAP32[$112+4>>2]=HEAP32[$113+4>>2]|0;HEAP32[$112+8>>2]=HEAP32[$113+8>>2]|0;
           $$0$i$i$i731 = 0;
           while(1) {
            $exitcond$i$i$i732 = ($$0$i$i$i731|0)==(3);
            if ($exitcond$i$i$i732) {
             break;
            }
            $1100 = (($113) + ($$0$i$i$i731<<2)|0);
            HEAP32[$1100>>2] = 0;
            $1101 = (($$0$i$i$i731) + 1)|0;
            $$0$i$i$i731 = $1101;
           }
           $1102 = ((($106)) + 11|0);
           $1103 = HEAP8[$1102>>0]|0;
           $1104 = ($1103<<24>>24)<(0);
           $1105 = HEAP32[$106>>2]|0;
           $1106 = $1104 ? $1105 : $106;
           $1107 = ((($106)) + 4|0);
           $1108 = HEAP32[$1107>>2]|0;
           $1109 = $1103&255;
           $1110 = $1104 ? $1108 : $1109;
           __THREW__ = 0;
           $1111 = (invoke_iiii(1112,($112|0),($1106|0),($1110|0))|0);
           $1112 = __THREW__; __THREW__ = 0;
           $1113 = $1112&1;
           do {
            if ($1113) {
             $1135 = ___cxa_find_matching_catch_2()|0;
             $1136 = tempRet0;
             $$14519 = $1135;$$14540 = $1136;
            } else {
             ;HEAP32[$111>>2]=HEAP32[$112>>2]|0;HEAP32[$111+4>>2]=HEAP32[$112+4>>2]|0;HEAP32[$111+8>>2]=HEAP32[$112+8>>2]|0;
             $$0$i$i$i735 = 0;
             while(1) {
              $exitcond$i$i$i736 = ($$0$i$i$i735|0)==(3);
              if ($exitcond$i$i$i736) {
               break;
              }
              $1114 = (($112) + ($$0$i$i$i735<<2)|0);
              HEAP32[$1114>>2] = 0;
              $1115 = (($$0$i$i$i735) + 1)|0;
              $$0$i$i$i735 = $1115;
             }
             __THREW__ = 0;
             $1116 = (invoke_iii(1119,($111|0),(159081|0))|0);
             $1117 = __THREW__; __THREW__ = 0;
             $1118 = $1117&1;
             if ($1118) {
              $1137 = ___cxa_find_matching_catch_2()|0;
              $1138 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($111);
              $$14519 = $1137;$$14540 = $1138;
              break;
             }
             ;HEAP32[$110>>2]=HEAP32[$111>>2]|0;HEAP32[$110+4>>2]=HEAP32[$111+4>>2]|0;HEAP32[$110+8>>2]=HEAP32[$111+8>>2]|0;
             $$0$i$i$i739 = 0;
             while(1) {
              $exitcond$i$i$i740 = ($$0$i$i$i739|0)==(3);
              if ($exitcond$i$i$i740) {
               break;
              }
              $1119 = (($111) + ($$0$i$i$i739<<2)|0);
              HEAP32[$1119>>2] = 0;
              $1120 = (($$0$i$i$i739) + 1)|0;
              $$0$i$i$i739 = $1120;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($109,$110);
             $1121 = HEAP32[$1052>>2]|0;
             $1122 = ((($1121)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($1122,$109);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($109);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($110);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($111);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($112);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($113);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($114);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($115);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($108);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($107);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($106);
             $$14 = $1050;
             break L1;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($112);
           $$15520 = $$14519;$$15541 = $$14540;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($113);
          $$16521 = $$15520;$$16542 = $$15541;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($114);
         $$17522 = $$16521;$$17543 = $$16542;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($115);
        $$18523 = $$17522;$$18544 = $$17543;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($108);
       $$19524 = $$18523;$$19545 = $$18544;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($107);
      $$20525 = $$19524;$$20546 = $$19545;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($106);
     $$21 = $$20525;$$21547 = $$20546;
     break;
    }
    case 114:  {
     $1154 = HEAP8[$134>>0]|0;
     $1155 = $1154 << 24 >> 24;
     switch ($1155|0) {
     case 99:  {
      $1156 = (__ZN10__cxxabiv112_GLOBAL__N_127parse_reinterpret_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1156;
      break L1;
      break;
     }
     case 109:  {
      ;HEAP32[$116>>2]=0|0;HEAP32[$116+4>>2]=0|0;HEAP32[$116+8>>2]=0|0;
      $1157 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159261)|0);
      $1158 = ($1157>>>0)>(4294967279);
      if ($1158) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($116);
       // unreachable;
      }
      $1159 = ($1157>>>0)<(11);
      if ($1159) {
       $1160 = $1157&255;
       $1161 = ((($116)) + 11|0);
       HEAP8[$1161>>0] = $1160;
       $$0$i$i749 = $116;
      } else {
       $1162 = (($1157) + 16)|0;
       $1163 = $1162 & -16;
       $1164 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($1163)|0);
       HEAP32[$116>>2] = $1164;
       $1165 = $1163 | -2147483648;
       $1166 = ((($116)) + 8|0);
       HEAP32[$1166>>2] = $1165;
       $1167 = ((($116)) + 4|0);
       HEAP32[$1167>>2] = $1157;
       $$0$i$i749 = $1164;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i749,159261,$1157)|0);
      $1168 = (($$0$i$i749) + ($1157)|0);
      HEAP8[$6>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($1168,$6);
      __THREW__ = 0;
      $1169 = (invoke_iiiii(1140,($126|0),($1|0),($116|0),($2|0))|0);
      $1170 = __THREW__; __THREW__ = 0;
      $1171 = $1170&1;
      if ($1171) {
       $1173 = ___cxa_find_matching_catch_2()|0;
       $1174 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($116);
       $$21 = $1173;$$21547 = $1174;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($116);
       $1172 = ($1169|0)==($126|0);
       $$595 = $1172 ? $0 : $1169;
       $$14 = $$595;
       break L1;
      }
      break;
     }
     case 77:  {
      ;HEAP32[$117>>2]=0|0;HEAP32[$117+4>>2]=0|0;HEAP32[$117+8>>2]=0|0;
      $1175 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159263)|0);
      $1176 = ($1175>>>0)>(4294967279);
      if ($1176) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($117);
       // unreachable;
      }
      $1177 = ($1175>>>0)<(11);
      if ($1177) {
       $1178 = $1175&255;
       $1179 = ((($117)) + 11|0);
       HEAP8[$1179>>0] = $1178;
       $$0$i$i751 = $117;
      } else {
       $1180 = (($1175) + 16)|0;
       $1181 = $1180 & -16;
       $1182 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($1181)|0);
       HEAP32[$117>>2] = $1182;
       $1183 = $1181 | -2147483648;
       $1184 = ((($117)) + 8|0);
       HEAP32[$1184>>2] = $1183;
       $1185 = ((($117)) + 4|0);
       HEAP32[$1185>>2] = $1175;
       $$0$i$i751 = $1182;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i751,159263,$1175)|0);
      $1186 = (($$0$i$i751) + ($1175)|0);
      HEAP8[$5>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($1186,$5);
      __THREW__ = 0;
      $1187 = (invoke_iiiii(1140,($126|0),($1|0),($117|0),($2|0))|0);
      $1188 = __THREW__; __THREW__ = 0;
      $1189 = $1188&1;
      if ($1189) {
       $1191 = ___cxa_find_matching_catch_2()|0;
       $1192 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($117);
       $$21 = $1191;$$21547 = $1192;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($117);
       $1190 = ($1187|0)==($126|0);
       $$596 = $1190 ? $0 : $1187;
       $$14 = $$596;
       break L1;
      }
      break;
     }
     case 115:  {
      ;HEAP32[$118>>2]=0|0;HEAP32[$118+4>>2]=0|0;HEAP32[$118+8>>2]=0|0;
      $1193 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159266)|0);
      $1194 = ($1193>>>0)>(4294967279);
      if ($1194) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($118);
       // unreachable;
      }
      $1195 = ($1193>>>0)<(11);
      if ($1195) {
       $1196 = $1193&255;
       $1197 = ((($118)) + 11|0);
       HEAP8[$1197>>0] = $1196;
       $$0$i$i753 = $118;
      } else {
       $1198 = (($1193) + 16)|0;
       $1199 = $1198 & -16;
       $1200 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($1199)|0);
       HEAP32[$118>>2] = $1200;
       $1201 = $1199 | -2147483648;
       $1202 = ((($118)) + 8|0);
       HEAP32[$1202>>2] = $1201;
       $1203 = ((($118)) + 4|0);
       HEAP32[$1203>>2] = $1193;
       $$0$i$i753 = $1200;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i753,159266,$1193)|0);
      $1204 = (($$0$i$i753) + ($1193)|0);
      HEAP8[$4>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($1204,$4);
      __THREW__ = 0;
      $1205 = (invoke_iiiii(1140,($126|0),($1|0),($118|0),($2|0))|0);
      $1206 = __THREW__; __THREW__ = 0;
      $1207 = $1206&1;
      if ($1207) {
       $1209 = ___cxa_find_matching_catch_2()|0;
       $1210 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($118);
       $$21 = $1209;$$21547 = $1210;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($118);
       $1208 = ($1205|0)==($126|0);
       $$597 = $1208 ? $0 : $1205;
       $$14 = $$597;
       break L1;
      }
      break;
     }
     case 83:  {
      ;HEAP32[$119>>2]=0|0;HEAP32[$119+4>>2]=0|0;HEAP32[$119+8>>2]=0|0;
      $1211 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159269)|0);
      $1212 = ($1211>>>0)>(4294967279);
      if ($1212) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($119);
       // unreachable;
      }
      $1213 = ($1211>>>0)<(11);
      if ($1213) {
       $1214 = $1211&255;
       $1215 = ((($119)) + 11|0);
       HEAP8[$1215>>0] = $1214;
       $$0$i$i755 = $119;
      } else {
       $1216 = (($1211) + 16)|0;
       $1217 = $1216 & -16;
       $1218 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($1217)|0);
       HEAP32[$119>>2] = $1218;
       $1219 = $1217 | -2147483648;
       $1220 = ((($119)) + 8|0);
       HEAP32[$1220>>2] = $1219;
       $1221 = ((($119)) + 4|0);
       HEAP32[$1221>>2] = $1211;
       $$0$i$i755 = $1218;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i755,159269,$1211)|0);
      $1222 = (($$0$i$i755) + ($1211)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($1222,$3);
      __THREW__ = 0;
      $1223 = (invoke_iiiii(1140,($126|0),($1|0),($119|0),($2|0))|0);
      $1224 = __THREW__; __THREW__ = 0;
      $1225 = $1224&1;
      if ($1225) {
       $1227 = ___cxa_find_matching_catch_2()|0;
       $1228 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($119);
       $$21 = $1227;$$21547 = $1228;
       break L7;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($119);
       $1226 = ($1223|0)==($126|0);
       $$598 = $1226 ? $0 : $1223;
       $$14 = $$598;
       break L1;
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 115:  {
     $1229 = HEAP8[$134>>0]|0;
     $1230 = $1229 << 24 >> 24;
     switch ($1230|0) {
     case 99:  {
      $1231 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_static_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1231;
      break L1;
      break;
     }
     case 112:  {
      $1232 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_pack_expansionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1232;
      break L1;
      break;
     }
     case 114:  {
      $1233 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1233;
      break L1;
      break;
     }
     case 116:  {
      $1234 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_sizeof_type_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1234;
      break L1;
      break;
     }
     case 122:  {
      $1235 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_sizeof_expr_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1235;
      break L1;
      break;
     }
     case 90:  {
      $1236 = $$0503;
      $1237 = (($121) - ($1236))|0;
      $1238 = ($1237|0)>(2);
      if (!($1238)) {
       $$14 = $0;
       break L1;
      }
      $1239 = HEAP8[$137>>0]|0;
      $1240 = $1239 << 24 >> 24;
      switch ($1240|0) {
      case 84:  {
       $1241 = (__ZN10__cxxabiv112_GLOBAL__N_128parse_sizeof_param_pack_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
       $$14 = $1241;
       break L1;
       break;
      }
      case 102:  {
       $1242 = (__ZN10__cxxabiv112_GLOBAL__N_137parse_sizeof_function_param_pack_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
       $$14 = $1242;
       break L1;
       break;
      }
      default: {
       $$14 = $0;
       break L1;
      }
      }
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 116:  {
     $1243 = HEAP8[$134>>0]|0;
     $1244 = $1243 << 24 >> 24;
     switch ($1244|0) {
     case 105: case 101:  {
      $1245 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_typeid_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1245;
      break L1;
      break;
     }
     case 114:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj6EEERAT__Kc($120,159273);
      $1246 = ((($2)) + 4|0);
      $1247 = HEAP32[$1246>>2]|0;
      $1248 = ((($2)) + 8|0);
      $1249 = HEAP32[$1248>>2]|0;
      $1250 = ($1247>>>0)<($1249>>>0);
      if ($1250) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($1247,$120);
       $1251 = HEAP32[$1246>>2]|0;
       $1252 = ((($1251)) + 24|0);
       HEAP32[$1246>>2] = $1252;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($120|0));
       $1253 = __THREW__; __THREW__ = 0;
       $1254 = $1253&1;
       if ($1254) {
        $1255 = ___cxa_find_matching_catch_2()|0;
        $1256 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($120);
        $$21 = $1255;$$21547 = $1256;
        break L7;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($120);
      $$14 = $126;
      break L1;
      break;
     }
     case 119:  {
      $1257 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_throw_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
      $$14 = $1257;
      break L1;
      break;
     }
     default: {
      $$14 = $0;
      break L1;
     }
     }
     break;
    }
    case 57: case 56: case 55: case 54: case 53: case 52: case 51: case 50: case 49:  {
     $1258 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $$14 = $1258;
     break L1;
     break;
    }
    default: {
     $$14 = $0;
     break L1;
    }
    }
   } while(0);
   ___resumeException($$21|0);
   // unreachable;
  } else {
   $$14 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$14|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7reserveEj($0) {
 $0 = $0|0;
 var $$017$off022 = 0, $$01918 = 0, $$020153134 = 0, $$02016 = 0, $$phitmp$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0;
 var $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $phitmp$i = 0, $phitmp$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 11|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = ($2<<24>>24)<(0);
 if ($3) {
  $4 = ((($0)) + 8|0);
  $5 = HEAP32[$4>>2]|0;
  $6 = $5 & 2147483647;
  $phitmp$i = (($6) + -1)|0;
  $7 = ((($0)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $11 = $8;$15 = $phitmp$i;
 } else {
  $9 = $2&255;
  $11 = $9;$15 = 10;
 }
 $10 = ($11>>>0)<(11);
 $12 = (($11) + 16)|0;
 $13 = $12 & -16;
 $phitmp$i2 = (($13) + -1)|0;
 $$phitmp$i = $10 ? 10 : $phitmp$i2;
 $14 = ($$phitmp$i|0)==($15|0);
 L5: do {
  if (!($14)) {
   do {
    if ($10) {
     $23 = HEAP32[$0>>2]|0;
     $24 = ($2<<24>>24)<(0);
     if ($24) {
      $$017$off022 = 0;$$01918 = $23;$$02016 = $0;
      label = 12;
     } else {
      $25 = $2&255;
      $26 = (($25) + 1)|0;
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$23,$26)|0);
      __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($23);
      label = 14;
     }
    } else {
     $16 = ($$phitmp$i>>>0)<=($15>>>0);
     $17 = (($$phitmp$i) + 1)|0;
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($17)|0);
     $19 = ($18|0)==(0|0);
     $or$cond = $16 & $19;
     if ($or$cond) {
      break L5;
     }
     $20 = HEAP8[$1>>0]|0;
     $21 = ($20<<24>>24)<(0);
     if ($21) {
      $22 = HEAP32[$0>>2]|0;
      $$017$off022 = 1;$$01918 = $22;$$02016 = $18;
      label = 12;
      break;
     } else {
      $27 = $20&255;
      $28 = (($27) + 1)|0;
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($18,$0,$28)|0);
      $$020153134 = $18;
      label = 13;
      break;
     }
    }
   } while(0);
   if ((label|0) == 12) {
    $29 = ((($0)) + 4|0);
    $30 = HEAP32[$29>>2]|0;
    $31 = (($30) + 1)|0;
    (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$02016,$$01918,$31)|0);
    __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($$01918);
    if ($$017$off022) {
     $$020153134 = $$02016;
     label = 13;
    } else {
     label = 14;
    }
   }
   if ((label|0) == 13) {
    $32 = (($$phitmp$i) + 1)|0;
    $33 = $32 | -2147483648;
    $34 = ((($0)) + 8|0);
    HEAP32[$34>>2] = $33;
    $35 = ((($0)) + 4|0);
    HEAP32[$35>>2] = $11;
    HEAP32[$0>>2] = $$020153134;
    break;
   }
   else if ((label|0) == 14) {
    $36 = $11&255;
    HEAP8[$1>>0] = $36;
    break;
   }
  }
 } while(0);
 return;
}
function __ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EEPKS8_RKSB_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i = 0, $$0$i$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $4 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$4>>2] = 0;
  $5 = (($$0$i$i) + 1)|0;
  $$0$i$i = $5;
 }
 $6 = (__ZNSt3__211char_traitsIcE6lengthEPKc($1)|0);
 $7 = ((($2)) + 11|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)<(0);
 $10 = ((($2)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $8&255;
 $13 = $9 ? $11 : $12;
 $14 = (($13) + ($6))|0;
 $15 = ($14>>>0)>(4294967279);
 if ($15) {
  __THREW__ = 0;
  invoke_vi(987,($0|0));
  $16 = __THREW__; __THREW__ = 0;
 } else {
  $17 = ($14>>>0)<(11);
  if ($17) {
   $18 = $6&255;
   $19 = ((($0)) + 11|0);
   HEAP8[$19>>0] = $18;
   $$0$i = $0;
  } else {
   $20 = (($14) + 16)|0;
   $21 = $20 & -16;
   $22 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($21)|0);
   HEAP32[$0>>2] = $22;
   $23 = $21 | -2147483648;
   $24 = ((($0)) + 8|0);
   HEAP32[$24>>2] = $23;
   $25 = ((($0)) + 4|0);
   HEAP32[$25>>2] = $6;
   $$0$i = $22;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i,$1,$6)|0);
  $26 = (($$0$i) + ($6)|0);
  HEAP8[$3>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($26,$3);
  $27 = HEAP8[$7>>0]|0;
  $28 = ($27<<24>>24)<(0);
  $29 = HEAP32[$2>>2]|0;
  $30 = $28 ? $29 : $2;
  __THREW__ = 0;
  (invoke_iiii(1112,($0|0),($30|0),($13|0))|0);
  $31 = __THREW__; __THREW__ = 0;
  $32 = $31&1;
  if (!($32)) {
   STACKTOP = sp;return;
  }
 }
 $33 = ___cxa_find_matching_catch_2()|0;
 $34 = tempRet0;
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($0);
 ___resumeException($33|0);
 // unreachable;
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_expr_primaryINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i258 = 0, $$0$i$i$i263 = 0, $$0$i$i231 = 0, $$0$i$i233 = 0, $$0$i$i235 = 0, $$0$i$i237 = 0, $$0$i$i239 = 0, $$0$i$i241 = 0, $$0$i$i243 = 0, $$0$i$i245 = 0, $$0$i$i247 = 0, $$0$i$i249 = 0, $$0$i$i251 = 0, $$0$i$i253 = 0, $$0$i$i255 = 0, $$0$i$i261 = 0, $$0194 = 0;
 var $$021$i$i = 0, $$021$ph$i$i = 0, $$1202 = 0, $$1207 = 0, $$211 = 0, $$212 = 0, $$213 = 0, $$214 = 0, $$215 = 0, $$216 = 0, $$217 = 0, $$218 = 0, $$219 = 0, $$220 = 0, $$2203 = 0, $$2208 = 0, $$221 = 0, $$222 = 0, $$223 = 0, $$224 = 0;
 var $$225 = 0, $$226 = 0, $$227 = 0, $$3 = 0, $$3204 = 0, $$3209 = 0, $$4205 = 0, $$4210 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0;
 var $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0;
 var $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0;
 var $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0;
 var $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0;
 var $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0;
 var $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0;
 var $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0;
 var $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0;
 var $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0;
 var $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0;
 var $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0;
 var $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0;
 var $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0;
 var $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0;
 var $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0;
 var $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0;
 var $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0, $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0;
 var $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0;
 var $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0;
 var $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i259 = 0, $exitcond$i$i$i264 = 0;
 var $isdigit = 0, $isdigittmp = 0, $or$cond = 0, $or$cond228 = 0, $or$cond283 = 0, $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 336|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(336|0);
 $3 = sp + 322|0;
 $4 = sp + 321|0;
 $5 = sp + 320|0;
 $6 = sp + 319|0;
 $7 = sp + 318|0;
 $8 = sp + 317|0;
 $9 = sp + 316|0;
 $10 = sp + 315|0;
 $11 = sp + 314|0;
 $12 = sp + 313|0;
 $13 = sp + 312|0;
 $14 = sp + 311|0;
 $15 = sp + 310|0;
 $16 = sp + 309|0;
 $17 = sp + 308|0;
 $18 = sp + 296|0;
 $19 = sp + 272|0;
 $20 = sp + 248|0;
 $21 = sp + 232|0;
 $22 = sp + 220|0;
 $23 = sp + 208|0;
 $24 = sp + 196|0;
 $25 = sp + 184|0;
 $26 = sp + 172|0;
 $27 = sp + 160|0;
 $28 = sp + 148|0;
 $29 = sp + 136|0;
 $30 = sp + 124|0;
 $31 = sp + 112|0;
 $32 = sp + 100|0;
 $33 = sp + 88|0;
 $34 = sp + 64|0;
 $35 = sp + 48|0;
 $36 = sp + 36|0;
 $37 = sp + 24|0;
 $38 = sp + 12|0;
 $39 = sp;
 $40 = $1;
 $41 = $0;
 $42 = (($40) - ($41))|0;
 $43 = ($42|0)>(3);
 L1: do {
  if ($43) {
   $44 = HEAP8[$0>>0]|0;
   $45 = ($44<<24>>24)==(76);
   if ($45) {
    $46 = ((($0)) + 1|0);
    $47 = HEAP8[$46>>0]|0;
    $48 = $47 << 24 >> 24;
    $49 = ((($0)) + 3|0);
    $50 = ((($0)) + 2|0);
    L4: do {
     switch ($48|0) {
     case 84:  {
      $$3 = $0;
      break L1;
      break;
     }
     case 119:  {
      ;HEAP32[$18>>2]=0|0;HEAP32[$18+4>>2]=0|0;HEAP32[$18+8>>2]=0|0;
      $51 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160275)|0);
      $52 = ($51>>>0)>(4294967279);
      if ($52) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($18);
       // unreachable;
      }
      $53 = ($51>>>0)<(11);
      if ($53) {
       $54 = $51&255;
       $55 = ((($18)) + 11|0);
       HEAP8[$55>>0] = $54;
       $$0$i$i = $18;
      } else {
       $56 = (($51) + 16)|0;
       $57 = $56 & -16;
       $58 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($57)|0);
       HEAP32[$18>>2] = $58;
       $59 = $57 | -2147483648;
       $60 = ((($18)) + 8|0);
       HEAP32[$60>>2] = $59;
       $61 = ((($18)) + 4|0);
       HEAP32[$61>>2] = $51;
       $$0$i$i = $58;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,160275,$51)|0);
      $62 = (($$0$i$i) + ($51)|0);
      HEAP8[$17>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($62,$17);
      __THREW__ = 0;
      $63 = (invoke_iiiii(1142,($50|0),($1|0),($18|0),($2|0))|0);
      $64 = __THREW__; __THREW__ = 0;
      $65 = $64&1;
      if ($65) {
       $67 = ___cxa_find_matching_catch_2()|0;
       $68 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
       $$4205 = $67;$$4210 = $68;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
       $66 = ($63|0)==($50|0);
       $$ = $66 ? $0 : $63;
       $$3 = $$;
       break L1;
      }
      break;
     }
     case 98:  {
      $69 = HEAP8[$49>>0]|0;
      $70 = ($69<<24>>24)==(69);
      if (!($70)) {
       $$3 = $0;
       break L1;
      }
      $71 = HEAP8[$50>>0]|0;
      $72 = $71 << 24 >> 24;
      $73 = ((($0)) + 4|0);
      switch ($72|0) {
      case 48:  {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj6EEERAT__Kc($19,160283);
       $74 = ((($2)) + 4|0);
       $75 = HEAP32[$74>>2]|0;
       $76 = ((($2)) + 8|0);
       $77 = HEAP32[$76>>2]|0;
       $78 = ($75>>>0)<($77>>>0);
       if ($78) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($75,$19);
        $79 = HEAP32[$74>>2]|0;
        $80 = ((($79)) + 24|0);
        HEAP32[$74>>2] = $80;
       } else {
        __THREW__ = 0;
        invoke_vii(1128,($2|0),($19|0));
        $81 = __THREW__; __THREW__ = 0;
        $82 = $81&1;
        if ($82) {
         $83 = ___cxa_find_matching_catch_2()|0;
         $84 = tempRet0;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
         $$4205 = $83;$$4210 = $84;
         break L4;
        }
       }
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
       $$3 = $73;
       break L1;
       break;
      }
      case 49:  {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($20,160289);
       $85 = ((($2)) + 4|0);
       $86 = HEAP32[$85>>2]|0;
       $87 = ((($2)) + 8|0);
       $88 = HEAP32[$87>>2]|0;
       $89 = ($86>>>0)<($88>>>0);
       if ($89) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($86,$20);
        $90 = HEAP32[$85>>2]|0;
        $91 = ((($90)) + 24|0);
        HEAP32[$85>>2] = $91;
       } else {
        __THREW__ = 0;
        invoke_vii(1128,($2|0),($20|0));
        $92 = __THREW__; __THREW__ = 0;
        $93 = $92&1;
        if ($93) {
         $94 = ___cxa_find_matching_catch_2()|0;
         $95 = tempRet0;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
         $$4205 = $94;$$4210 = $95;
         break L4;
        }
       }
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
       $$3 = $73;
       break L1;
       break;
      }
      default: {
       $$3 = $0;
       break L1;
      }
      }
      break;
     }
     case 99:  {
      ;HEAP32[$21>>2]=0|0;HEAP32[$21+4>>2]=0|0;HEAP32[$21+8>>2]=0|0;
      $96 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160294)|0);
      $97 = ($96>>>0)>(4294967279);
      if ($97) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($21);
       // unreachable;
      }
      $98 = ($96>>>0)<(11);
      if ($98) {
       $99 = $96&255;
       $100 = ((($21)) + 11|0);
       HEAP8[$100>>0] = $99;
       $$0$i$i231 = $21;
      } else {
       $101 = (($96) + 16)|0;
       $102 = $101 & -16;
       $103 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($102)|0);
       HEAP32[$21>>2] = $103;
       $104 = $102 | -2147483648;
       $105 = ((($21)) + 8|0);
       HEAP32[$105>>2] = $104;
       $106 = ((($21)) + 4|0);
       HEAP32[$106>>2] = $96;
       $$0$i$i231 = $103;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i231,160294,$96)|0);
      $107 = (($$0$i$i231) + ($96)|0);
      HEAP8[$16>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($107,$16);
      __THREW__ = 0;
      $108 = (invoke_iiiii(1142,($50|0),($1|0),($21|0),($2|0))|0);
      $109 = __THREW__; __THREW__ = 0;
      $110 = $109&1;
      if ($110) {
       $112 = ___cxa_find_matching_catch_2()|0;
       $113 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($21);
       $$4205 = $112;$$4210 = $113;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($21);
       $111 = ($108|0)==($50|0);
       $$211 = $111 ? $0 : $108;
       $$3 = $$211;
       break L1;
      }
      break;
     }
     case 97:  {
      ;HEAP32[$22>>2]=0|0;HEAP32[$22+4>>2]=0|0;HEAP32[$22+8>>2]=0|0;
      $114 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160299)|0);
      $115 = ($114>>>0)>(4294967279);
      if ($115) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($22);
       // unreachable;
      }
      $116 = ($114>>>0)<(11);
      if ($116) {
       $117 = $114&255;
       $118 = ((($22)) + 11|0);
       HEAP8[$118>>0] = $117;
       $$0$i$i233 = $22;
      } else {
       $119 = (($114) + 16)|0;
       $120 = $119 & -16;
       $121 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($120)|0);
       HEAP32[$22>>2] = $121;
       $122 = $120 | -2147483648;
       $123 = ((($22)) + 8|0);
       HEAP32[$123>>2] = $122;
       $124 = ((($22)) + 4|0);
       HEAP32[$124>>2] = $114;
       $$0$i$i233 = $121;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i233,160299,$114)|0);
      $125 = (($$0$i$i233) + ($114)|0);
      HEAP8[$15>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($125,$15);
      __THREW__ = 0;
      $126 = (invoke_iiiii(1142,($50|0),($1|0),($22|0),($2|0))|0);
      $127 = __THREW__; __THREW__ = 0;
      $128 = $127&1;
      if ($128) {
       $130 = ___cxa_find_matching_catch_2()|0;
       $131 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($22);
       $$4205 = $130;$$4210 = $131;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($22);
       $129 = ($126|0)==($50|0);
       $$212 = $129 ? $0 : $126;
       $$3 = $$212;
       break L1;
      }
      break;
     }
     case 104:  {
      ;HEAP32[$23>>2]=0|0;HEAP32[$23+4>>2]=0|0;HEAP32[$23+8>>2]=0|0;
      $132 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160311)|0);
      $133 = ($132>>>0)>(4294967279);
      if ($133) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($23);
       // unreachable;
      }
      $134 = ($132>>>0)<(11);
      if ($134) {
       $135 = $132&255;
       $136 = ((($23)) + 11|0);
       HEAP8[$136>>0] = $135;
       $$0$i$i235 = $23;
      } else {
       $137 = (($132) + 16)|0;
       $138 = $137 & -16;
       $139 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($138)|0);
       HEAP32[$23>>2] = $139;
       $140 = $138 | -2147483648;
       $141 = ((($23)) + 8|0);
       HEAP32[$141>>2] = $140;
       $142 = ((($23)) + 4|0);
       HEAP32[$142>>2] = $132;
       $$0$i$i235 = $139;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i235,160311,$132)|0);
      $143 = (($$0$i$i235) + ($132)|0);
      HEAP8[$14>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($143,$14);
      __THREW__ = 0;
      $144 = (invoke_iiiii(1142,($50|0),($1|0),($23|0),($2|0))|0);
      $145 = __THREW__; __THREW__ = 0;
      $146 = $145&1;
      if ($146) {
       $148 = ___cxa_find_matching_catch_2()|0;
       $149 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($23);
       $$4205 = $148;$$4210 = $149;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($23);
       $147 = ($144|0)==($50|0);
       $$213 = $147 ? $0 : $144;
       $$3 = $$213;
       break L1;
      }
      break;
     }
     case 115:  {
      ;HEAP32[$24>>2]=0|0;HEAP32[$24+4>>2]=0|0;HEAP32[$24+8>>2]=0|0;
      $150 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160325)|0);
      $151 = ($150>>>0)>(4294967279);
      if ($151) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($24);
       // unreachable;
      }
      $152 = ($150>>>0)<(11);
      if ($152) {
       $153 = $150&255;
       $154 = ((($24)) + 11|0);
       HEAP8[$154>>0] = $153;
       $$0$i$i237 = $24;
      } else {
       $155 = (($150) + 16)|0;
       $156 = $155 & -16;
       $157 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($156)|0);
       HEAP32[$24>>2] = $157;
       $158 = $156 | -2147483648;
       $159 = ((($24)) + 8|0);
       HEAP32[$159>>2] = $158;
       $160 = ((($24)) + 4|0);
       HEAP32[$160>>2] = $150;
       $$0$i$i237 = $157;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i237,160325,$150)|0);
      $161 = (($$0$i$i237) + ($150)|0);
      HEAP8[$13>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($161,$13);
      __THREW__ = 0;
      $162 = (invoke_iiiii(1142,($50|0),($1|0),($24|0),($2|0))|0);
      $163 = __THREW__; __THREW__ = 0;
      $164 = $163&1;
      if ($164) {
       $166 = ___cxa_find_matching_catch_2()|0;
       $167 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($24);
       $$4205 = $166;$$4210 = $167;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($24);
       $165 = ($162|0)==($50|0);
       $$214 = $165 ? $0 : $162;
       $$3 = $$214;
       break L1;
      }
      break;
     }
     case 116:  {
      ;HEAP32[$25>>2]=0|0;HEAP32[$25+4>>2]=0|0;HEAP32[$25+8>>2]=0|0;
      $168 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160331)|0);
      $169 = ($168>>>0)>(4294967279);
      if ($169) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($25);
       // unreachable;
      }
      $170 = ($168>>>0)<(11);
      if ($170) {
       $171 = $168&255;
       $172 = ((($25)) + 11|0);
       HEAP8[$172>>0] = $171;
       $$0$i$i239 = $25;
      } else {
       $173 = (($168) + 16)|0;
       $174 = $173 & -16;
       $175 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($174)|0);
       HEAP32[$25>>2] = $175;
       $176 = $174 | -2147483648;
       $177 = ((($25)) + 8|0);
       HEAP32[$177>>2] = $176;
       $178 = ((($25)) + 4|0);
       HEAP32[$178>>2] = $168;
       $$0$i$i239 = $175;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i239,160331,$168)|0);
      $179 = (($$0$i$i239) + ($168)|0);
      HEAP8[$12>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($179,$12);
      __THREW__ = 0;
      $180 = (invoke_iiiii(1142,($50|0),($1|0),($25|0),($2|0))|0);
      $181 = __THREW__; __THREW__ = 0;
      $182 = $181&1;
      if ($182) {
       $184 = ___cxa_find_matching_catch_2()|0;
       $185 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($25);
       $$4205 = $184;$$4210 = $185;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($25);
       $183 = ($180|0)==($50|0);
       $$215 = $183 ? $0 : $180;
       $$3 = $$215;
       break L1;
      }
      break;
     }
     case 105:  {
      ;HEAP32[$26>>2]=0|0;HEAP32[$26+4>>2]=0|0;HEAP32[$26+8>>2]=0|0;
      $186 = (__ZNSt3__211char_traitsIcE6lengthEPKc(169457)|0);
      $187 = ($186>>>0)>(4294967279);
      if ($187) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($26);
       // unreachable;
      }
      $188 = ($186>>>0)<(11);
      if ($188) {
       $189 = $186&255;
       $190 = ((($26)) + 11|0);
       HEAP8[$190>>0] = $189;
       $$0$i$i241 = $26;
      } else {
       $191 = (($186) + 16)|0;
       $192 = $191 & -16;
       $193 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($192)|0);
       HEAP32[$26>>2] = $193;
       $194 = $192 | -2147483648;
       $195 = ((($26)) + 8|0);
       HEAP32[$195>>2] = $194;
       $196 = ((($26)) + 4|0);
       HEAP32[$196>>2] = $186;
       $$0$i$i241 = $193;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i241,169457,$186)|0);
      $197 = (($$0$i$i241) + ($186)|0);
      HEAP8[$11>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($197,$11);
      __THREW__ = 0;
      $198 = (invoke_iiiii(1142,($50|0),($1|0),($26|0),($2|0))|0);
      $199 = __THREW__; __THREW__ = 0;
      $200 = $199&1;
      if ($200) {
       $202 = ___cxa_find_matching_catch_2()|0;
       $203 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($26);
       $$4205 = $202;$$4210 = $203;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($26);
       $201 = ($198|0)==($50|0);
       $$216 = $201 ? $0 : $198;
       $$3 = $$216;
       break L1;
      }
      break;
     }
     case 106:  {
      ;HEAP32[$27>>2]=0|0;HEAP32[$27+4>>2]=0|0;HEAP32[$27+8>>2]=0|0;
      $204 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160346)|0);
      $205 = ($204>>>0)>(4294967279);
      if ($205) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($27);
       // unreachable;
      }
      $206 = ($204>>>0)<(11);
      if ($206) {
       $207 = $204&255;
       $208 = ((($27)) + 11|0);
       HEAP8[$208>>0] = $207;
       $$0$i$i243 = $27;
      } else {
       $209 = (($204) + 16)|0;
       $210 = $209 & -16;
       $211 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($210)|0);
       HEAP32[$27>>2] = $211;
       $212 = $210 | -2147483648;
       $213 = ((($27)) + 8|0);
       HEAP32[$213>>2] = $212;
       $214 = ((($27)) + 4|0);
       HEAP32[$214>>2] = $204;
       $$0$i$i243 = $211;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i243,160346,$204)|0);
      $215 = (($$0$i$i243) + ($204)|0);
      HEAP8[$10>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($215,$10);
      __THREW__ = 0;
      $216 = (invoke_iiiii(1142,($50|0),($1|0),($27|0),($2|0))|0);
      $217 = __THREW__; __THREW__ = 0;
      $218 = $217&1;
      if ($218) {
       $220 = ___cxa_find_matching_catch_2()|0;
       $221 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($27);
       $$4205 = $220;$$4210 = $221;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($27);
       $219 = ($216|0)==($50|0);
       $$217 = $219 ? $0 : $216;
       $$3 = $$217;
       break L1;
      }
      break;
     }
     case 108:  {
      ;HEAP32[$28>>2]=0|0;HEAP32[$28+4>>2]=0|0;HEAP32[$28+8>>2]=0|0;
      $222 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160348)|0);
      $223 = ($222>>>0)>(4294967279);
      if ($223) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($28);
       // unreachable;
      }
      $224 = ($222>>>0)<(11);
      if ($224) {
       $225 = $222&255;
       $226 = ((($28)) + 11|0);
       HEAP8[$226>>0] = $225;
       $$0$i$i245 = $28;
      } else {
       $227 = (($222) + 16)|0;
       $228 = $227 & -16;
       $229 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($228)|0);
       HEAP32[$28>>2] = $229;
       $230 = $228 | -2147483648;
       $231 = ((($28)) + 8|0);
       HEAP32[$231>>2] = $230;
       $232 = ((($28)) + 4|0);
       HEAP32[$232>>2] = $222;
       $$0$i$i245 = $229;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i245,160348,$222)|0);
      $233 = (($$0$i$i245) + ($222)|0);
      HEAP8[$9>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($233,$9);
      __THREW__ = 0;
      $234 = (invoke_iiiii(1142,($50|0),($1|0),($28|0),($2|0))|0);
      $235 = __THREW__; __THREW__ = 0;
      $236 = $235&1;
      if ($236) {
       $238 = ___cxa_find_matching_catch_2()|0;
       $239 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($28);
       $$4205 = $238;$$4210 = $239;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($28);
       $237 = ($234|0)==($50|0);
       $$218 = $237 ? $0 : $234;
       $$3 = $$218;
       break L1;
      }
      break;
     }
     case 109:  {
      ;HEAP32[$29>>2]=0|0;HEAP32[$29+4>>2]=0|0;HEAP32[$29+8>>2]=0|0;
      $240 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160350)|0);
      $241 = ($240>>>0)>(4294967279);
      if ($241) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($29);
       // unreachable;
      }
      $242 = ($240>>>0)<(11);
      if ($242) {
       $243 = $240&255;
       $244 = ((($29)) + 11|0);
       HEAP8[$244>>0] = $243;
       $$0$i$i247 = $29;
      } else {
       $245 = (($240) + 16)|0;
       $246 = $245 & -16;
       $247 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($246)|0);
       HEAP32[$29>>2] = $247;
       $248 = $246 | -2147483648;
       $249 = ((($29)) + 8|0);
       HEAP32[$249>>2] = $248;
       $250 = ((($29)) + 4|0);
       HEAP32[$250>>2] = $240;
       $$0$i$i247 = $247;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i247,160350,$240)|0);
      $251 = (($$0$i$i247) + ($240)|0);
      HEAP8[$8>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($251,$8);
      __THREW__ = 0;
      $252 = (invoke_iiiii(1142,($50|0),($1|0),($29|0),($2|0))|0);
      $253 = __THREW__; __THREW__ = 0;
      $254 = $253&1;
      if ($254) {
       $256 = ___cxa_find_matching_catch_2()|0;
       $257 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($29);
       $$4205 = $256;$$4210 = $257;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($29);
       $255 = ($252|0)==($50|0);
       $$219 = $255 ? $0 : $252;
       $$3 = $$219;
       break L1;
      }
      break;
     }
     case 120:  {
      ;HEAP32[$30>>2]=0|0;HEAP32[$30+4>>2]=0|0;HEAP32[$30+8>>2]=0|0;
      $258 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160353)|0);
      $259 = ($258>>>0)>(4294967279);
      if ($259) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($30);
       // unreachable;
      }
      $260 = ($258>>>0)<(11);
      if ($260) {
       $261 = $258&255;
       $262 = ((($30)) + 11|0);
       HEAP8[$262>>0] = $261;
       $$0$i$i249 = $30;
      } else {
       $263 = (($258) + 16)|0;
       $264 = $263 & -16;
       $265 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($264)|0);
       HEAP32[$30>>2] = $265;
       $266 = $264 | -2147483648;
       $267 = ((($30)) + 8|0);
       HEAP32[$267>>2] = $266;
       $268 = ((($30)) + 4|0);
       HEAP32[$268>>2] = $258;
       $$0$i$i249 = $265;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i249,160353,$258)|0);
      $269 = (($$0$i$i249) + ($258)|0);
      HEAP8[$7>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($269,$7);
      __THREW__ = 0;
      $270 = (invoke_iiiii(1142,($50|0),($1|0),($30|0),($2|0))|0);
      $271 = __THREW__; __THREW__ = 0;
      $272 = $271&1;
      if ($272) {
       $274 = ___cxa_find_matching_catch_2()|0;
       $275 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($30);
       $$4205 = $274;$$4210 = $275;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($30);
       $273 = ($270|0)==($50|0);
       $$220 = $273 ? $0 : $270;
       $$3 = $$220;
       break L1;
      }
      break;
     }
     case 121:  {
      ;HEAP32[$31>>2]=0|0;HEAP32[$31+4>>2]=0|0;HEAP32[$31+8>>2]=0|0;
      $276 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160356)|0);
      $277 = ($276>>>0)>(4294967279);
      if ($277) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($31);
       // unreachable;
      }
      $278 = ($276>>>0)<(11);
      if ($278) {
       $279 = $276&255;
       $280 = ((($31)) + 11|0);
       HEAP8[$280>>0] = $279;
       $$0$i$i251 = $31;
      } else {
       $281 = (($276) + 16)|0;
       $282 = $281 & -16;
       $283 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($282)|0);
       HEAP32[$31>>2] = $283;
       $284 = $282 | -2147483648;
       $285 = ((($31)) + 8|0);
       HEAP32[$285>>2] = $284;
       $286 = ((($31)) + 4|0);
       HEAP32[$286>>2] = $276;
       $$0$i$i251 = $283;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i251,160356,$276)|0);
      $287 = (($$0$i$i251) + ($276)|0);
      HEAP8[$6>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($287,$6);
      __THREW__ = 0;
      $288 = (invoke_iiiii(1142,($50|0),($1|0),($31|0),($2|0))|0);
      $289 = __THREW__; __THREW__ = 0;
      $290 = $289&1;
      if ($290) {
       $292 = ___cxa_find_matching_catch_2()|0;
       $293 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($31);
       $$4205 = $292;$$4210 = $293;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($31);
       $291 = ($288|0)==($50|0);
       $$221 = $291 ? $0 : $288;
       $$3 = $$221;
       break L1;
      }
      break;
     }
     case 110:  {
      ;HEAP32[$32>>2]=0|0;HEAP32[$32+4>>2]=0|0;HEAP32[$32+8>>2]=0|0;
      $294 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160360)|0);
      $295 = ($294>>>0)>(4294967279);
      if ($295) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($32);
       // unreachable;
      }
      $296 = ($294>>>0)<(11);
      if ($296) {
       $297 = $294&255;
       $298 = ((($32)) + 11|0);
       HEAP8[$298>>0] = $297;
       $$0$i$i253 = $32;
      } else {
       $299 = (($294) + 16)|0;
       $300 = $299 & -16;
       $301 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($300)|0);
       HEAP32[$32>>2] = $301;
       $302 = $300 | -2147483648;
       $303 = ((($32)) + 8|0);
       HEAP32[$303>>2] = $302;
       $304 = ((($32)) + 4|0);
       HEAP32[$304>>2] = $294;
       $$0$i$i253 = $301;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i253,160360,$294)|0);
      $305 = (($$0$i$i253) + ($294)|0);
      HEAP8[$5>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($305,$5);
      __THREW__ = 0;
      $306 = (invoke_iiiii(1142,($50|0),($1|0),($32|0),($2|0))|0);
      $307 = __THREW__; __THREW__ = 0;
      $308 = $307&1;
      if ($308) {
       $310 = ___cxa_find_matching_catch_2()|0;
       $311 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($32);
       $$4205 = $310;$$4210 = $311;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($32);
       $309 = ($306|0)==($50|0);
       $$222 = $309 ? $0 : $306;
       $$3 = $$222;
       break L1;
      }
      break;
     }
     case 111:  {
      ;HEAP32[$33>>2]=0|0;HEAP32[$33+4>>2]=0|0;HEAP32[$33+8>>2]=0|0;
      $312 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160369)|0);
      $313 = ($312>>>0)>(4294967279);
      if ($313) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($33);
       // unreachable;
      }
      $314 = ($312>>>0)<(11);
      if ($314) {
       $315 = $312&255;
       $316 = ((($33)) + 11|0);
       HEAP8[$316>>0] = $315;
       $$0$i$i255 = $33;
      } else {
       $317 = (($312) + 16)|0;
       $318 = $317 & -16;
       $319 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($318)|0);
       HEAP32[$33>>2] = $319;
       $320 = $318 | -2147483648;
       $321 = ((($33)) + 8|0);
       HEAP32[$321>>2] = $320;
       $322 = ((($33)) + 4|0);
       HEAP32[$322>>2] = $312;
       $$0$i$i255 = $319;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i255,160369,$312)|0);
      $323 = (($$0$i$i255) + ($312)|0);
      HEAP8[$4>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($323,$4);
      __THREW__ = 0;
      $324 = (invoke_iiiii(1142,($50|0),($1|0),($33|0),($2|0))|0);
      $325 = __THREW__; __THREW__ = 0;
      $326 = $325&1;
      if ($326) {
       $328 = ___cxa_find_matching_catch_2()|0;
       $329 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($33);
       $$4205 = $328;$$4210 = $329;
       break L4;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($33);
       $327 = ($324|0)==($50|0);
       $$223 = $327 ? $0 : $324;
       $$3 = $$223;
       break L1;
      }
      break;
     }
     case 102:  {
      $330 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIfNS0_2DbEEEPKcS4_S4_RT0_($50,$1,$2)|0);
      $331 = ($330|0)==($50|0);
      $$224 = $331 ? $0 : $330;
      $$3 = $$224;
      break L1;
      break;
     }
     case 100:  {
      $332 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIdNS0_2DbEEEPKcS4_S4_RT0_($50,$1,$2)|0);
      $333 = ($332|0)==($50|0);
      $$225 = $333 ? $0 : $332;
      $$3 = $$225;
      break L1;
      break;
     }
     case 101:  {
      $334 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIeNS0_2DbEEEPKcS4_S4_RT0_($50,$1,$2)|0);
      $335 = ($334|0)==($50|0);
      $$226 = $335 ? $0 : $334;
      $$3 = $$226;
      break L1;
      break;
     }
     case 95:  {
      $336 = HEAP8[$50>>0]|0;
      $337 = ($336<<24>>24)==(90);
      if (!($337)) {
       $$3 = $0;
       break L1;
      }
      $338 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($49,$1,$2)|0);
      $339 = ($338|0)==($49|0);
      $340 = ($338|0)==($1|0);
      $or$cond = $339 | $340;
      if ($or$cond) {
       $$3 = $0;
       break L1;
      }
      $341 = HEAP8[$338>>0]|0;
      $342 = ($341<<24>>24)==(69);
      $343 = ((($338)) + 1|0);
      $$227 = $342 ? $343 : $0;
      $$3 = $$227;
      break L1;
      break;
     }
     default: {
      $344 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($46,$1,$2)|0);
      $345 = ($344|0)==($46|0);
      $346 = ($344|0)==($1|0);
      $or$cond228 = $345 | $346;
      if ($or$cond228) {
       $$3 = $0;
       break L1;
      }
      $347 = HEAP8[$344>>0]|0;
      $348 = ($347<<24>>24)==(69);
      if ($348) {
       $411 = ((($344)) + 1|0);
       $$3 = $411;
       break L1;
      } else {
       $$0194 = $344;
      }
      while(1) {
       $349 = ($$0194|0)==($1|0);
       $350 = ((($$0194)) + 1|0);
       if ($349) {
        $$3 = $0;
        break L1;
       }
       $351 = HEAP8[$$0194>>0]|0;
       $352 = $351 << 24 >> 24;
       $isdigittmp = (($352) + -48)|0;
       $isdigit = ($isdigittmp>>>0)<(10);
       if ($isdigit) {
        $$0194 = $350;
       } else {
        break;
       }
      }
      $353 = ($$0194|0)!=($344|0);
      $354 = ($351<<24>>24)==(69);
      $or$cond283 = $353 & $354;
      if (!($or$cond283)) {
       $$3 = $0;
       break L1;
      }
      $355 = HEAP32[$2>>2]|0;
      $356 = ((($2)) + 4|0);
      $357 = HEAP32[$356>>2]|0;
      $358 = ($355|0)==($357|0);
      if ($358) {
       $$3 = $0;
       break L1;
      }
      $359 = ((($357)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($38,$359);
      __THREW__ = 0;
      $360 = (invoke_iiii(1121,($38|0),0,(159083|0))|0);
      $361 = __THREW__; __THREW__ = 0;
      $362 = $361&1;
      if ($362) {
       $405 = ___cxa_find_matching_catch_2()|0;
       $406 = tempRet0;
       $$3204 = $405;$$3209 = $406;
      } else {
       ;HEAP32[$37>>2]=HEAP32[$38>>2]|0;HEAP32[$37+4>>2]=HEAP32[$38+4>>2]|0;HEAP32[$37+8>>2]=HEAP32[$38+8>>2]|0;
       $$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $363 = (($38) + ($$0$i$i$i<<2)|0);
        HEAP32[$363>>2] = 0;
        $364 = (($$0$i$i$i) + 1)|0;
        $$0$i$i$i = $364;
       }
       __THREW__ = 0;
       $365 = (invoke_iii(1119,($37|0),(159081|0))|0);
       $366 = __THREW__; __THREW__ = 0;
       $367 = $366&1;
       if ($367) {
        $407 = ___cxa_find_matching_catch_2()|0;
        $408 = tempRet0;
        $$2203 = $407;$$2208 = $408;
       } else {
        ;HEAP32[$36>>2]=HEAP32[$37>>2]|0;HEAP32[$36+4>>2]=HEAP32[$37+4>>2]|0;HEAP32[$36+8>>2]=HEAP32[$37+8>>2]|0;
        $$0$i$i$i258 = 0;
        while(1) {
         $exitcond$i$i$i259 = ($$0$i$i$i258|0)==(3);
         if ($exitcond$i$i$i259) {
          break;
         }
         $368 = (($37) + ($$0$i$i$i258<<2)|0);
         HEAP32[$368>>2] = 0;
         $369 = (($$0$i$i$i258) + 1)|0;
         $$0$i$i$i258 = $369;
        }
        ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0;
        $370 = $344;
        $371 = $$0194;
        $372 = (($371) - ($370))|0;
        $373 = ($372>>>0)>(4294967279);
        do {
         if ($373) {
          __THREW__ = 0;
          invoke_vi(987,($39|0));
          $374 = __THREW__; __THREW__ = 0;
          $375 = ___cxa_find_matching_catch_2()|0;
          $376 = tempRet0;
          $$1202 = $375;$$1207 = $376;
         } else {
          $377 = ($372>>>0)<(11);
          if ($377) {
           $378 = $372&255;
           $379 = ((($39)) + 11|0);
           HEAP8[$379>>0] = $378;
           $$021$ph$i$i = $39;
          } else {
           $380 = (($372) + 16)|0;
           $381 = $380 & -16;
           $382 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($381)|0);
           HEAP32[$39>>2] = $382;
           $383 = $381 | -2147483648;
           $384 = ((($39)) + 8|0);
           HEAP32[$384>>2] = $383;
           $385 = ((($39)) + 4|0);
           HEAP32[$385>>2] = $372;
           $$021$ph$i$i = $382;
          }
          $$0$i$i261 = $344;$$021$i$i = $$021$ph$i$i;
          while(1) {
           $386 = ($$0$i$i261|0)==($$0194|0);
           if ($386) {
            break;
           }
           __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i261);
           $387 = ((($$0$i$i261)) + 1|0);
           $388 = ((($$021$i$i)) + 1|0);
           $$0$i$i261 = $387;$$021$i$i = $388;
          }
          $scevgep23$i$i = (($$021$ph$i$i) + ($372)|0);
          HEAP8[$3>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$3);
          $389 = ((($39)) + 11|0);
          $390 = HEAP8[$389>>0]|0;
          $391 = ($390<<24>>24)<(0);
          $392 = HEAP32[$39>>2]|0;
          $393 = $391 ? $392 : $39;
          $394 = ((($39)) + 4|0);
          $395 = HEAP32[$394>>2]|0;
          $396 = $390&255;
          $397 = $391 ? $395 : $396;
          __THREW__ = 0;
          $398 = (invoke_iiii(1112,($36|0),($393|0),($397|0))|0);
          $399 = __THREW__; __THREW__ = 0;
          $400 = $399&1;
          if ($400) {
           $409 = ___cxa_find_matching_catch_2()|0;
           $410 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($39);
           $$1202 = $409;$$1207 = $410;
           break;
          }
          ;HEAP32[$35>>2]=HEAP32[$36>>2]|0;HEAP32[$35+4>>2]=HEAP32[$36+4>>2]|0;HEAP32[$35+8>>2]=HEAP32[$36+8>>2]|0;
          $$0$i$i$i263 = 0;
          while(1) {
           $exitcond$i$i$i264 = ($$0$i$i$i263|0)==(3);
           if ($exitcond$i$i$i264) {
            break;
           }
           $401 = (($36) + ($$0$i$i$i263<<2)|0);
           HEAP32[$401>>2] = 0;
           $402 = (($$0$i$i$i263) + 1)|0;
           $$0$i$i$i263 = $402;
          }
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($34,$35);
          $403 = HEAP32[$356>>2]|0;
          $404 = ((($403)) + -24|0);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($404,$34);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($34);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($35);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($39);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($36);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($37);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($38);
          $$3 = $350;
          break L1;
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($36);
        $$2203 = $$1202;$$2208 = $$1207;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($37);
       $$3204 = $$2203;$$3209 = $$2208;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($38);
      $$4205 = $$3204;$$4210 = $$3209;
     }
     }
    } while(0);
    ___resumeException($$4205|0);
    // unreachable;
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_120parse_function_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i63 = 0, $$0$i$i59 = 0, $$021$i$i = 0, $$021$i$i58 = 0, $$021$ph$i$i = 0, $$021$ph$i$i56 = 0, $$049 = 0, $$052 = 0, $$150 = 0, $$153 = 0, $$2 = 0, $$251 = 0, $$254 = 0, $$3 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0;
 var $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0;
 var $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i64 = 0, $scevgep23$i$i = 0, $scevgep23$i$i61 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 109|0;
 $4 = sp + 108|0;
 $5 = sp + 104|0;
 $6 = sp + 80|0;
 $7 = sp + 64|0;
 $8 = sp + 52|0;
 $9 = sp + 48|0;
 $10 = sp + 24|0;
 $11 = sp + 12|0;
 $12 = sp;
 $13 = $1;
 $14 = $0;
 $15 = (($13) - ($14))|0;
 $16 = ($15|0)>(2);
 L1: do {
  if ($16) {
   $17 = HEAP8[$0>>0]|0;
   $18 = ($17<<24>>24)==(102);
   if ($18) {
    $19 = ((($0)) + 1|0);
    $20 = HEAP8[$19>>0]|0;
    $21 = ((($0)) + 2|0);
    L4: do {
     switch ($20<<24>>24) {
     case 112:  {
      $22 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_cv_qualifiersEPKcS2_Rj($21,$1,$5)|0);
      $23 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($22,$1)|0);
      $24 = ($23|0)==($1|0);
      L6: do {
       if ($24) {
        $$0 = $0;
       } else {
        $25 = HEAP8[$23>>0]|0;
        $26 = ($25<<24>>24)==(95);
        if ($26) {
         ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;
         $27 = $22;
         $28 = $23;
         $29 = (($28) - ($27))|0;
         $30 = ($29>>>0)>(4294967279);
         if ($30) {
          __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($8);
          // unreachable;
         }
         $31 = ($29>>>0)<(11);
         if ($31) {
          $32 = $29&255;
          $33 = ((($8)) + 11|0);
          HEAP8[$33>>0] = $32;
          $$021$ph$i$i = $8;
         } else {
          $34 = (($29) + 16)|0;
          $35 = $34 & -16;
          $36 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($35)|0);
          HEAP32[$8>>2] = $36;
          $37 = $35 | -2147483648;
          $38 = ((($8)) + 8|0);
          HEAP32[$38>>2] = $37;
          $39 = ((($8)) + 4|0);
          HEAP32[$39>>2] = $29;
          $$021$ph$i$i = $36;
         }
         $$0$i$i = $22;$$021$i$i = $$021$ph$i$i;
         while(1) {
          $40 = ($$0$i$i|0)==($23|0);
          if ($40) {
           break;
          }
          __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
          $41 = ((($$0$i$i)) + 1|0);
          $42 = ((($$021$i$i)) + 1|0);
          $$0$i$i = $41;$$021$i$i = $42;
         }
         $scevgep23$i$i = (($$021$ph$i$i) + ($29)|0);
         HEAP8[$4>>0] = 0;
         __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$4);
         __THREW__ = 0;
         $43 = (invoke_iiii(1121,($8|0),0,(160272|0))|0);
         $44 = __THREW__; __THREW__ = 0;
         $45 = $44&1;
         do {
          if ($45) {
           $58 = ___cxa_find_matching_catch_2()|0;
           $59 = tempRet0;
           $$049 = $58;$$052 = $59;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i = 0;
           while(1) {
            $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
            if ($exitcond$i$i$i) {
             break;
            }
            $46 = (($8) + ($$0$i$i$i<<2)|0);
            HEAP32[$46>>2] = 0;
            $47 = (($$0$i$i$i) + 1)|0;
            $$0$i$i$i = $47;
           }
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($6,$7);
           $48 = ((($2)) + 4|0);
           $49 = HEAP32[$48>>2]|0;
           $50 = ((($2)) + 8|0);
           $51 = HEAP32[$50>>2]|0;
           $52 = ($49>>>0)<($51>>>0);
           if ($52) {
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($49,$6);
            $53 = HEAP32[$48>>2]|0;
            $54 = ((($53)) + 24|0);
            HEAP32[$48>>2] = $54;
           } else {
            __THREW__ = 0;
            invoke_vii(1128,($2|0),($6|0));
            $55 = __THREW__; __THREW__ = 0;
            $56 = $55&1;
            if ($56) {
             $60 = ___cxa_find_matching_catch_2()|0;
             $61 = tempRet0;
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
             $$049 = $60;$$052 = $61;
             break;
            }
           }
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
           $57 = ((($23)) + 1|0);
           $$0 = $57;
           break L6;
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         $$251 = $$049;$$254 = $$052;
         break L4;
        } else {
         $$0 = $0;
        }
       }
      } while(0);
      $$3 = $$0;
      break L1;
      break;
     }
     case 76:  {
      $62 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($21,$1)|0);
      $63 = ($62|0)==($1|0);
      L35: do {
       if ($63) {
        $$2 = $0;
       } else {
        $64 = HEAP8[$62>>0]|0;
        $65 = ($64<<24>>24)==(112);
        if ($65) {
         $66 = ((($62)) + 1|0);
         $67 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_cv_qualifiersEPKcS2_Rj($66,$1,$9)|0);
         $68 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($67,$1)|0);
         $69 = ($68|0)==($1|0);
         if ($69) {
          $$2 = $0;
         } else {
          $70 = HEAP8[$68>>0]|0;
          $71 = ($70<<24>>24)==(95);
          if ($71) {
           ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;
           $72 = $67;
           $73 = $68;
           $74 = (($73) - ($72))|0;
           $75 = ($74>>>0)>(4294967279);
           if ($75) {
            __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($12);
            // unreachable;
           }
           $76 = ($74>>>0)<(11);
           if ($76) {
            $77 = $74&255;
            $78 = ((($12)) + 11|0);
            HEAP8[$78>>0] = $77;
            $$021$ph$i$i56 = $12;
           } else {
            $79 = (($74) + 16)|0;
            $80 = $79 & -16;
            $81 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($80)|0);
            HEAP32[$12>>2] = $81;
            $82 = $80 | -2147483648;
            $83 = ((($12)) + 8|0);
            HEAP32[$83>>2] = $82;
            $84 = ((($12)) + 4|0);
            HEAP32[$84>>2] = $74;
            $$021$ph$i$i56 = $81;
           }
           $$0$i$i59 = $67;$$021$i$i58 = $$021$ph$i$i56;
           while(1) {
            $85 = ($$0$i$i59|0)==($68|0);
            if ($85) {
             break;
            }
            __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i58,$$0$i$i59);
            $86 = ((($$0$i$i59)) + 1|0);
            $87 = ((($$021$i$i58)) + 1|0);
            $$0$i$i59 = $86;$$021$i$i58 = $87;
           }
           $scevgep23$i$i61 = (($$021$ph$i$i56) + ($74)|0);
           HEAP8[$3>>0] = 0;
           __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i61,$3);
           __THREW__ = 0;
           $88 = (invoke_iiii(1121,($12|0),0,(160272|0))|0);
           $89 = __THREW__; __THREW__ = 0;
           $90 = $89&1;
           do {
            if ($90) {
             $103 = ___cxa_find_matching_catch_2()|0;
             $104 = tempRet0;
             $$150 = $103;$$153 = $104;
            } else {
             ;HEAP32[$11>>2]=HEAP32[$12>>2]|0;HEAP32[$11+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$12+8>>2]|0;
             $$0$i$i$i63 = 0;
             while(1) {
              $exitcond$i$i$i64 = ($$0$i$i$i63|0)==(3);
              if ($exitcond$i$i$i64) {
               break;
              }
              $91 = (($12) + ($$0$i$i$i63<<2)|0);
              HEAP32[$91>>2] = 0;
              $92 = (($$0$i$i$i63) + 1)|0;
              $$0$i$i$i63 = $92;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($10,$11);
             $93 = ((($2)) + 4|0);
             $94 = HEAP32[$93>>2]|0;
             $95 = ((($2)) + 8|0);
             $96 = HEAP32[$95>>2]|0;
             $97 = ($94>>>0)<($96>>>0);
             if ($97) {
              __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($94,$10);
              $98 = HEAP32[$93>>2]|0;
              $99 = ((($98)) + 24|0);
              HEAP32[$93>>2] = $99;
             } else {
              __THREW__ = 0;
              invoke_vii(1128,($2|0),($10|0));
              $100 = __THREW__; __THREW__ = 0;
              $101 = $100&1;
              if ($101) {
               $105 = ___cxa_find_matching_catch_2()|0;
               $106 = tempRet0;
               __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
               __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
               $$150 = $105;$$153 = $106;
               break;
              }
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
             $102 = ((($68)) + 1|0);
             $$2 = $102;
             break L35;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
           $$251 = $$150;$$254 = $$153;
           break L4;
          } else {
           $$2 = $0;
          }
         }
        } else {
         $$2 = $0;
        }
       }
      } while(0);
      $$3 = $$2;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
    } while(0);
    ___resumeException($$251|0);
    // unreachable;
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_123parse_binary_expressionINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0$i$i$i = 0, $$0$i$i$i50 = 0, $$0$i$i$i53 = 0, $$0$i$i$i57 = 0, $$0$i$i$i61 = 0, $$036 = 0, $$038 = 0, $$1 = 0, $$137 = 0, $$139 = 0, $$2 = 0, $$240 = 0, $$3 = 0, $$341 = 0, $$4 = 0, $$442 = 0, $$5 = 0, $$543 = 0, $$6 = 0, $$644 = 0;
 var $$7 = 0, $$745 = 0, $$pre$i = 0, $$pre$i70 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i54 = 0, $exitcond$i$i$i58 = 0, $exitcond$i$i$i62 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $4 = sp + 97|0;
 $5 = sp + 96|0;
 $6 = sp + 84|0;
 $7 = sp + 72|0;
 $8 = sp + 60|0;
 $9 = sp + 48|0;
 $10 = sp + 36|0;
 $11 = sp + 24|0;
 $12 = sp + 12|0;
 $13 = sp;
 $14 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$3)|0);
 $15 = ($14|0)==($0|0);
 L1: do {
  if ($15) {
   $$1 = $0;
  } else {
   $16 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($14,$1,$3)|0);
   $17 = ($16|0)==($14|0);
   $18 = ((($3)) + 4|0);
   if ($17) {
    $145 = HEAP32[$18>>2]|0;
    $146 = ((($145)) + -24|0);
    $148 = $145;
    while(1) {
     $147 = ($148|0)==($146|0);
     if ($147) {
      $$1 = $0;
      break L1;
     }
     $149 = ((($148)) + -24|0);
     HEAP32[$18>>2] = $149;
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($149);
     $$pre$i70 = HEAP32[$18>>2]|0;
     $148 = $$pre$i70;
    }
   }
   $19 = HEAP32[$18>>2]|0;
   $20 = HEAP32[$3>>2]|0;
   $21 = (($19) - ($20))|0;
   $22 = (($21|0) / 24)&-1;
   $23 = ($22>>>0)<(2);
   if ($23) {
    $$1 = $0;
   } else {
    $24 = $19;
    $25 = ((($24)) + -24|0);
    __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$25);
    $26 = HEAP32[$18>>2]|0;
    $27 = ((($26)) + -24|0);
    $29 = $26;
    while(1) {
     $28 = ($29|0)==($27|0);
     if ($28) {
      break;
     }
     $30 = ((($29)) + -24|0);
     HEAP32[$18>>2] = $30;
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
     $$pre$i = HEAP32[$18>>2]|0;
     $29 = $$pre$i;
    }
    $31 = ((($26)) + -48|0);
    __THREW__ = 0;
    invoke_vii(1120,($7|0),($31|0));
    $32 = __THREW__; __THREW__ = 0;
    $33 = $32&1;
    if ($33) {
     $58 = ___cxa_find_matching_catch_2()|0;
     $59 = tempRet0;
     $$7 = $59;$$745 = $58;
    } else {
     $34 = HEAP32[$18>>2]|0;
     $35 = ((($34)) + -24|0);
     $36 = ((($35)) + 11|0);
     $37 = HEAP8[$36>>0]|0;
     $38 = ($37<<24>>24)<(0);
     if ($38) {
      $39 = HEAP32[$35>>2]|0;
      HEAP8[$4>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($39,$4);
      $40 = ((($34)) + -20|0);
      HEAP32[$40>>2] = 0;
     } else {
      HEAP8[$5>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($35,$5);
      HEAP8[$36>>0] = 0;
     }
     $41 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159119)|0);
     $42 = ((($2)) + 11|0);
     $43 = HEAP8[$42>>0]|0;
     $44 = ($43<<24>>24)<(0);
     $45 = ((($2)) + 4|0);
     $46 = HEAP32[$45>>2]|0;
     $47 = $43&255;
     $48 = $44 ? $46 : $47;
     $49 = ($41|0)==($48|0);
     if ($49) {
      __THREW__ = 0;
      $50 = (invoke_iiii(1123,($2|0),(159119|0),($41|0))|0);
      $51 = __THREW__; __THREW__ = 0;
      $52 = $51&1;
      if ($52) {
       $53 = ___cxa_find_matching_catch_3(0|0)|0;
       $54 = tempRet0;
       ___clang_call_terminate($53);
       // unreachable;
      }
      $55 = ($50|0)==(0);
      if ($55) {
       __THREW__ = 0;
       invoke_vii(1117,($35|0),40);
       $56 = __THREW__; __THREW__ = 0;
       $57 = $56&1;
       if ($57) {
        label = 17;
       } else {
        label = 18;
       }
      } else {
       label = 18;
      }
     } else {
      label = 18;
     }
     L27: do {
      if ((label|0) == 18) {
       __THREW__ = 0;
       invoke_viii(1139,($13|0),(159083|0),($7|0));
       $62 = __THREW__; __THREW__ = 0;
       $63 = $62&1;
       if ($63) {
        $131 = ___cxa_find_matching_catch_2()|0;
        $132 = tempRet0;
        $$5 = $132;$$543 = $131;
       } else {
        __THREW__ = 0;
        $64 = (invoke_iii(1119,($13|0),(159369|0))|0);
        $65 = __THREW__; __THREW__ = 0;
        $66 = $65&1;
        if ($66) {
         $133 = ___cxa_find_matching_catch_2()|0;
         $134 = tempRet0;
         $$4 = $134;$$442 = $133;
        } else {
         ;HEAP32[$12>>2]=HEAP32[$13>>2]|0;HEAP32[$12+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$13+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $67 = (($13) + ($$0$i$i$i<<2)|0);
          HEAP32[$67>>2] = 0;
          $68 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $68;
         }
         $69 = HEAP8[$42>>0]|0;
         $70 = ($69<<24>>24)<(0);
         $71 = HEAP32[$2>>2]|0;
         $72 = $70 ? $71 : $2;
         $73 = HEAP32[$45>>2]|0;
         $74 = $69&255;
         $75 = $70 ? $73 : $74;
         __THREW__ = 0;
         $76 = (invoke_iiii(1112,($12|0),($72|0),($75|0))|0);
         $77 = __THREW__; __THREW__ = 0;
         $78 = $77&1;
         if ($78) {
          $135 = ___cxa_find_matching_catch_2()|0;
          $136 = tempRet0;
          $$3 = $136;$$341 = $135;
         } else {
          ;HEAP32[$11>>2]=HEAP32[$12>>2]|0;HEAP32[$11+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$12+8>>2]|0;
          $$0$i$i$i50 = 0;
          while(1) {
           $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
           if ($exitcond$i$i$i51) {
            break;
           }
           $79 = (($12) + ($$0$i$i$i50<<2)|0);
           HEAP32[$79>>2] = 0;
           $80 = (($$0$i$i$i50) + 1)|0;
           $$0$i$i$i50 = $80;
          }
          __THREW__ = 0;
          $81 = (invoke_iii(1119,($11|0),(159078|0))|0);
          $82 = __THREW__; __THREW__ = 0;
          $83 = $82&1;
          if ($83) {
           $137 = ___cxa_find_matching_catch_2()|0;
           $138 = tempRet0;
           $$2 = $138;$$240 = $137;
          } else {
           ;HEAP32[$10>>2]=HEAP32[$11>>2]|0;HEAP32[$10+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$11+8>>2]|0;
           $$0$i$i$i53 = 0;
           while(1) {
            $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
            if ($exitcond$i$i$i54) {
             break;
            }
            $84 = (($11) + ($$0$i$i$i53<<2)|0);
            HEAP32[$84>>2] = 0;
            $85 = (($$0$i$i$i53) + 1)|0;
            $$0$i$i$i53 = $85;
           }
           $86 = ((($6)) + 11|0);
           $87 = HEAP8[$86>>0]|0;
           $88 = ($87<<24>>24)<(0);
           $89 = HEAP32[$6>>2]|0;
           $90 = $88 ? $89 : $6;
           $91 = ((($6)) + 4|0);
           $92 = HEAP32[$91>>2]|0;
           $93 = $87&255;
           $94 = $88 ? $92 : $93;
           __THREW__ = 0;
           $95 = (invoke_iiii(1112,($10|0),($90|0),($94|0))|0);
           $96 = __THREW__; __THREW__ = 0;
           $97 = $96&1;
           if ($97) {
            $139 = ___cxa_find_matching_catch_2()|0;
            $140 = tempRet0;
            $$137 = $140;$$139 = $139;
           } else {
            ;HEAP32[$9>>2]=HEAP32[$10>>2]|0;HEAP32[$9+4>>2]=HEAP32[$10+4>>2]|0;HEAP32[$9+8>>2]=HEAP32[$10+8>>2]|0;
            $$0$i$i$i57 = 0;
            while(1) {
             $exitcond$i$i$i58 = ($$0$i$i$i57|0)==(3);
             if ($exitcond$i$i$i58) {
              break;
             }
             $98 = (($10) + ($$0$i$i$i57<<2)|0);
             HEAP32[$98>>2] = 0;
             $99 = (($$0$i$i$i57) + 1)|0;
             $$0$i$i$i57 = $99;
            }
            __THREW__ = 0;
            $100 = (invoke_iii(1119,($9|0),(159081|0))|0);
            $101 = __THREW__; __THREW__ = 0;
            $102 = $101&1;
            do {
             if ($102) {
              $141 = ___cxa_find_matching_catch_2()|0;
              $142 = tempRet0;
              $$036 = $142;$$038 = $141;
             } else {
              ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
              $$0$i$i$i61 = 0;
              while(1) {
               $exitcond$i$i$i62 = ($$0$i$i$i61|0)==(3);
               if ($exitcond$i$i$i62) {
                break;
               }
               $103 = (($9) + ($$0$i$i$i61<<2)|0);
               HEAP32[$103>>2] = 0;
               $104 = (($$0$i$i$i61) + 1)|0;
               $$0$i$i$i61 = $104;
              }
              $105 = ((($8)) + 11|0);
              $106 = HEAP8[$105>>0]|0;
              $107 = ($106<<24>>24)<(0);
              $108 = HEAP32[$8>>2]|0;
              $109 = $107 ? $108 : $8;
              $110 = ((($8)) + 4|0);
              $111 = HEAP32[$110>>2]|0;
              $112 = $106&255;
              $113 = $107 ? $111 : $112;
              __THREW__ = 0;
              (invoke_iiii(1112,($35|0),($109|0),($113|0))|0);
              $114 = __THREW__; __THREW__ = 0;
              $115 = $114&1;
              if ($115) {
               $143 = ___cxa_find_matching_catch_2()|0;
               $144 = tempRet0;
               __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
               $$036 = $144;$$038 = $143;
               break;
              }
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
              $116 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159119)|0);
              $117 = HEAP8[$42>>0]|0;
              $118 = ($117<<24>>24)<(0);
              $119 = HEAP32[$45>>2]|0;
              $120 = $117&255;
              $121 = $118 ? $119 : $120;
              $122 = ($116|0)==($121|0);
              do {
               if ($122) {
                __THREW__ = 0;
                $123 = (invoke_iiii(1123,($2|0),(159119|0),($116|0))|0);
                $124 = __THREW__; __THREW__ = 0;
                $125 = $124&1;
                if ($125) {
                 $126 = ___cxa_find_matching_catch_3(0|0)|0;
                 $127 = tempRet0;
                 ___clang_call_terminate($126);
                 // unreachable;
                }
                $128 = ($123|0)==(0);
                if (!($128)) {
                 break;
                }
                __THREW__ = 0;
                invoke_vii(1117,($35|0),41);
                $129 = __THREW__; __THREW__ = 0;
                $130 = $129&1;
                if ($130) {
                 label = 17;
                 break L27;
                }
               }
              } while(0);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
              $$1 = $16;
              break L1;
             }
            } while(0);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
            $$137 = $$036;$$139 = $$038;
           }
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
           $$2 = $$137;$$240 = $$139;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
          $$3 = $$2;$$341 = $$240;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
         $$4 = $$3;$$442 = $$341;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
        $$5 = $$4;$$543 = $$442;
       }
       $$6 = $$5;$$644 = $$543;
      }
     } while(0);
     if ((label|0) == 17) {
      $60 = ___cxa_find_matching_catch_2()|0;
      $61 = tempRet0;
      $$6 = $61;$$644 = $60;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
     $$7 = $$6;$$745 = $$644;
    }
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
    ___resumeException($$745|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_123parse_prefix_expressionINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i20 = 0, $$016 = 0, $$018 = 0, $$1 = 0, $$117 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0;
 var $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0;
 var $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i21 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $4 = sp + 49|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$3)|0);
 $11 = ($10|0)==($0|0);
 L1: do {
  if ($11) {
   $$018 = $0;
  } else {
   $12 = HEAP32[$3>>2]|0;
   $13 = ((($3)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($12|0)==($14|0);
   if ($15) {
    $$018 = $0;
   } else {
    __ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EERKSB_PKS8_($8,$2,159083);
    $16 = HEAP32[$13>>2]|0;
    $17 = ((($16)) + -24|0);
    __THREW__ = 0;
    invoke_vii(1120,($9|0),($17|0));
    $18 = __THREW__; __THREW__ = 0;
    $19 = $18&1;
    if ($19) {
     $52 = ___cxa_find_matching_catch_2()|0;
     $53 = tempRet0;
     $$1 = $53;$$117 = $52;
    } else {
     $20 = ((($9)) + 11|0);
     $21 = HEAP8[$20>>0]|0;
     $22 = ($21<<24>>24)<(0);
     $23 = HEAP32[$9>>2]|0;
     $24 = $22 ? $23 : $9;
     $25 = ((($9)) + 4|0);
     $26 = HEAP32[$25>>2]|0;
     $27 = $21&255;
     $28 = $22 ? $26 : $27;
     __THREW__ = 0;
     $29 = (invoke_iiii(1112,($8|0),($24|0),($28|0))|0);
     $30 = __THREW__; __THREW__ = 0;
     $31 = $30&1;
     do {
      if ($31) {
       $54 = ___cxa_find_matching_catch_2()|0;
       $55 = tempRet0;
       $$0 = $55;$$016 = $54;
      } else {
       ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
       $$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $32 = (($8) + ($$0$i$i$i<<2)|0);
        HEAP32[$32>>2] = 0;
        $33 = (($$0$i$i$i) + 1)|0;
        $$0$i$i$i = $33;
       }
       __THREW__ = 0;
       $34 = (invoke_iii(1119,($7|0),(159081|0))|0);
       $35 = __THREW__; __THREW__ = 0;
       $36 = $35&1;
       if ($36) {
        $56 = ___cxa_find_matching_catch_2()|0;
        $57 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        $$0 = $57;$$016 = $56;
        break;
       }
       ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
       $$0$i$i$i20 = 0;
       while(1) {
        $exitcond$i$i$i21 = ($$0$i$i$i20|0)==(3);
        if ($exitcond$i$i$i21) {
         break;
        }
        $37 = (($7) + ($$0$i$i$i20<<2)|0);
        HEAP32[$37>>2] = 0;
        $38 = (($$0$i$i$i20) + 1)|0;
        $$0$i$i$i20 = $38;
       }
       $39 = HEAP32[$13>>2]|0;
       $40 = ((($39)) + -24|0);
       $41 = ((($40)) + 11|0);
       $42 = HEAP8[$41>>0]|0;
       $43 = ($42<<24>>24)<(0);
       if ($43) {
        $44 = HEAP32[$40>>2]|0;
        HEAP8[$4>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($44,$4);
        $45 = ((($39)) + -20|0);
        HEAP32[$45>>2] = 0;
       } else {
        HEAP8[$5>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($40,$5);
        HEAP8[$41>>0] = 0;
       }
       __THREW__ = 0;
       invoke_vi(1116,($40|0));
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        $48 = ___cxa_find_matching_catch_3(0|0)|0;
        $49 = tempRet0;
        ___clang_call_terminate($48);
        // unreachable;
       }
       ;HEAP32[$40>>2]=HEAP32[$6>>2]|0;HEAP32[$40+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$40+8>>2]=HEAP32[$6+8>>2]|0;
       $$0$i$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
        if ($exitcond$i$i$i$i) {
         break;
        }
        $50 = (($6) + ($$0$i$i$i$i<<2)|0);
        HEAP32[$50>>2] = 0;
        $51 = (($$0$i$i$i$i) + 1)|0;
        $$0$i$i$i$i = $51;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
       $$018 = $10;
       break L1;
      }
     } while(0);
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
     $$1 = $$0;$$117 = $$016;
    }
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
    ___resumeException($$117|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$018|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_alignof_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i25 = 0, $$020 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i26 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 37|0;
 $4 = sp + 36|0;
 $5 = sp + 24|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10|0)>(2);
 L1: do {
  if ($11) {
   $12 = HEAP8[$0>>0]|0;
   $13 = ($12<<24>>24)==(97);
   if ($13) {
    $14 = ((($0)) + 1|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = ($15<<24>>24)==(116);
    if ($16) {
     $17 = ((($0)) + 2|0);
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($17,$1,$2)|0);
     $19 = ($18|0)==($17|0);
     if ($19) {
      $$1 = $0;
     } else {
      $20 = HEAP32[$2>>2]|0;
      $21 = ((($2)) + 4|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($20|0)==($22|0);
      if ($23) {
       $$1 = $0;
      } else {
       $24 = ((($22)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($7,$24);
       __THREW__ = 0;
       $25 = (invoke_iiii(1121,($7|0),0,(160262|0))|0);
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       do {
        if ($27) {
         $48 = ___cxa_find_matching_catch_2()|0;
         $49 = tempRet0;
         $$0 = $49;$$020 = $48;
        } else {
         ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $28 = (($7) + ($$0$i$i$i<<2)|0);
          HEAP32[$28>>2] = 0;
          $29 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $29;
         }
         __THREW__ = 0;
         $30 = (invoke_iii(1119,($6|0),(159081|0))|0);
         $31 = __THREW__; __THREW__ = 0;
         $32 = $31&1;
         if ($32) {
          $50 = ___cxa_find_matching_catch_2()|0;
          $51 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
          $$0 = $51;$$020 = $50;
          break;
         }
         ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
         $$0$i$i$i25 = 0;
         while(1) {
          $exitcond$i$i$i26 = ($$0$i$i$i25|0)==(3);
          if ($exitcond$i$i$i26) {
           break;
          }
          $33 = (($6) + ($$0$i$i$i25<<2)|0);
          HEAP32[$33>>2] = 0;
          $34 = (($$0$i$i$i25) + 1)|0;
          $$0$i$i$i25 = $34;
         }
         $35 = HEAP32[$21>>2]|0;
         $36 = ((($35)) + -24|0);
         $37 = ((($36)) + 11|0);
         $38 = HEAP8[$37>>0]|0;
         $39 = ($38<<24>>24)<(0);
         if ($39) {
          $40 = HEAP32[$36>>2]|0;
          HEAP8[$3>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($40,$3);
          $41 = ((($35)) + -20|0);
          HEAP32[$41>>2] = 0;
         } else {
          HEAP8[$4>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($36,$4);
          HEAP8[$37>>0] = 0;
         }
         __THREW__ = 0;
         invoke_vi(1116,($36|0));
         $42 = __THREW__; __THREW__ = 0;
         $43 = $42&1;
         if ($43) {
          $44 = ___cxa_find_matching_catch_3(0|0)|0;
          $45 = tempRet0;
          ___clang_call_terminate($44);
          // unreachable;
         }
         ;HEAP32[$36>>2]=HEAP32[$5>>2]|0;HEAP32[$36+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$36+8>>2]=HEAP32[$5+8>>2]|0;
         $$0$i$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
          if ($exitcond$i$i$i$i) {
           break;
          }
          $46 = (($5) + ($$0$i$i$i$i<<2)|0);
          HEAP32[$46>>2] = 0;
          $47 = (($$0$i$i$i$i) + 1)|0;
          $$0$i$i$i$i = $47;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
         $$1 = $18;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
       ___resumeException($$020|0);
       // unreachable;
      }
     }
    } else {
     $$1 = $0;
    }
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_alignof_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i25 = 0, $$020 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i26 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 37|0;
 $4 = sp + 36|0;
 $5 = sp + 24|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10|0)>(2);
 L1: do {
  if ($11) {
   $12 = HEAP8[$0>>0]|0;
   $13 = ($12<<24>>24)==(97);
   if ($13) {
    $14 = ((($0)) + 1|0);
    $15 = HEAP8[$14>>0]|0;
    $16 = ($15<<24>>24)==(122);
    if ($16) {
     $17 = ((($0)) + 2|0);
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($17,$1,$2)|0);
     $19 = ($18|0)==($17|0);
     if ($19) {
      $$1 = $0;
     } else {
      $20 = HEAP32[$2>>2]|0;
      $21 = ((($2)) + 4|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($20|0)==($22|0);
      if ($23) {
       $$1 = $0;
      } else {
       $24 = ((($22)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($7,$24);
       __THREW__ = 0;
       $25 = (invoke_iiii(1121,($7|0),0,(160262|0))|0);
       $26 = __THREW__; __THREW__ = 0;
       $27 = $26&1;
       do {
        if ($27) {
         $48 = ___cxa_find_matching_catch_2()|0;
         $49 = tempRet0;
         $$0 = $49;$$020 = $48;
        } else {
         ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $28 = (($7) + ($$0$i$i$i<<2)|0);
          HEAP32[$28>>2] = 0;
          $29 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $29;
         }
         __THREW__ = 0;
         $30 = (invoke_iii(1119,($6|0),(159081|0))|0);
         $31 = __THREW__; __THREW__ = 0;
         $32 = $31&1;
         if ($32) {
          $50 = ___cxa_find_matching_catch_2()|0;
          $51 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
          $$0 = $51;$$020 = $50;
          break;
         }
         ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
         $$0$i$i$i25 = 0;
         while(1) {
          $exitcond$i$i$i26 = ($$0$i$i$i25|0)==(3);
          if ($exitcond$i$i$i26) {
           break;
          }
          $33 = (($6) + ($$0$i$i$i25<<2)|0);
          HEAP32[$33>>2] = 0;
          $34 = (($$0$i$i$i25) + 1)|0;
          $$0$i$i$i25 = $34;
         }
         $35 = HEAP32[$21>>2]|0;
         $36 = ((($35)) + -24|0);
         $37 = ((($36)) + 11|0);
         $38 = HEAP8[$37>>0]|0;
         $39 = ($38<<24>>24)<(0);
         if ($39) {
          $40 = HEAP32[$36>>2]|0;
          HEAP8[$3>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($40,$3);
          $41 = ((($35)) + -20|0);
          HEAP32[$41>>2] = 0;
         } else {
          HEAP8[$4>>0] = 0;
          __ZNSt3__211char_traitsIcE6assignERcRKc($36,$4);
          HEAP8[$37>>0] = 0;
         }
         __THREW__ = 0;
         invoke_vi(1116,($36|0));
         $42 = __THREW__; __THREW__ = 0;
         $43 = $42&1;
         if ($43) {
          $44 = ___cxa_find_matching_catch_3(0|0)|0;
          $45 = tempRet0;
          ___clang_call_terminate($44);
          // unreachable;
         }
         ;HEAP32[$36>>2]=HEAP32[$5>>2]|0;HEAP32[$36+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$36+8>>2]=HEAP32[$5+8>>2]|0;
         $$0$i$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
          if ($exitcond$i$i$i$i) {
           break;
          }
          $46 = (($5) + ($$0$i$i$i$i<<2)|0);
          HEAP32[$46>>2] = 0;
          $47 = (($$0$i$i$i$i) + 1)|0;
          $$0$i$i$i$i = $47;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
         $$1 = $18;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
       ___resumeException($$020|0);
       // unreachable;
      }
     }
    } else {
     $$1 = $0;
    }
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_const_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i47 = 0, $$0$i$i$i50 = 0, $$0$i$i$i53 = 0, $$1 = 0, $$132 = 0, $$2 = 0, $$233 = 0, $$3 = 0, $$334 = 0, $$341 = 0, $$4 = 0, $$435 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i48 = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i54 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 88|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(99);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(99);
    if ($18) {
     $19 = ((($0)) + 2|0);
     $20 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($19,$1,$2)|0);
     $21 = ($20|0)==($19|0);
     if ($21) {
      $$341 = $0;
     } else {
      $22 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($20,$1,$2)|0);
      $23 = ($22|0)==($20|0);
      if ($23) {
       $$341 = $0;
      } else {
       $24 = ((($2)) + 4|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = HEAP32[$2>>2]|0;
       $27 = (($25) - ($26))|0;
       $28 = (($27|0) / 24)&-1;
       $29 = ($28>>>0)<(2);
       if ($29) {
        $$341 = $0;
       } else {
        $30 = $25;
        $31 = ((($30)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$31);
        $32 = HEAP32[$24>>2]|0;
        $33 = ((($32)) + -24|0);
        $35 = $32;
        while(1) {
         $34 = ($35|0)==($33|0);
         if ($34) {
          break;
         }
         $36 = ((($35)) + -24|0);
         HEAP32[$24>>2] = $36;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
         $$pre$i = HEAP32[$24>>2]|0;
         $35 = $$pre$i;
        }
        $37 = ((($32)) + -48|0);
        __THREW__ = 0;
        invoke_vii(1120,($9|0),($37|0));
        $38 = __THREW__; __THREW__ = 0;
        $39 = $38&1;
        if ($39) {
         $71 = ___cxa_find_matching_catch_2()|0;
         $72 = tempRet0;
         $$4 = $72;$$435 = $71;
        } else {
         __THREW__ = 0;
         $40 = (invoke_iiii(1121,($9|0),0,(160250|0))|0);
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         if ($42) {
          $73 = ___cxa_find_matching_catch_2()|0;
          $74 = tempRet0;
          $$3 = $74;$$334 = $73;
         } else {
          ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $43 = (($9) + ($$0$i$i$i<<2)|0);
           HEAP32[$43>>2] = 0;
           $44 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $44;
          }
          __THREW__ = 0;
          $45 = (invoke_iii(1119,($8|0),(159330|0))|0);
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           $75 = ___cxa_find_matching_catch_2()|0;
           $76 = tempRet0;
           $$2 = $76;$$233 = $75;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i47 = 0;
           while(1) {
            $exitcond$i$i$i48 = ($$0$i$i$i47|0)==(3);
            if ($exitcond$i$i$i48) {
             break;
            }
            $48 = (($8) + ($$0$i$i$i47<<2)|0);
            HEAP32[$48>>2] = 0;
            $49 = (($$0$i$i$i47) + 1)|0;
            $$0$i$i$i47 = $49;
           }
           $50 = ((($3)) + 11|0);
           $51 = HEAP8[$50>>0]|0;
           $52 = ($51<<24>>24)<(0);
           $53 = HEAP32[$3>>2]|0;
           $54 = $52 ? $53 : $3;
           $55 = ((($3)) + 4|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $51&255;
           $58 = $52 ? $56 : $57;
           __THREW__ = 0;
           $59 = (invoke_iiii(1112,($7|0),($54|0),($58|0))|0);
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           do {
            if ($61) {
             $77 = ___cxa_find_matching_catch_2()|0;
             $78 = tempRet0;
             $$1 = $78;$$132 = $77;
            } else {
             ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
             $$0$i$i$i50 = 0;
             while(1) {
              $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
              if ($exitcond$i$i$i51) {
               break;
              }
              $62 = (($7) + ($$0$i$i$i50<<2)|0);
              HEAP32[$62>>2] = 0;
              $63 = (($$0$i$i$i50) + 1)|0;
              $$0$i$i$i50 = $63;
             }
             __THREW__ = 0;
             $64 = (invoke_iii(1119,($6|0),(159081|0))|0);
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              $79 = ___cxa_find_matching_catch_2()|0;
              $80 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
              $$1 = $80;$$132 = $79;
              break;
             }
             ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
             $$0$i$i$i53 = 0;
             while(1) {
              $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
              if ($exitcond$i$i$i54) {
               break;
              }
              $67 = (($6) + ($$0$i$i$i53<<2)|0);
              HEAP32[$67>>2] = 0;
              $68 = (($$0$i$i$i53) + 1)|0;
              $$0$i$i$i53 = $68;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
             $69 = HEAP32[$24>>2]|0;
             $70 = ((($69)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($70,$4);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
             $$341 = $22;
             break L1;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
           $$2 = $$1;$$233 = $$132;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
          $$3 = $$2;$$334 = $$233;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$4 = $$3;$$435 = $$334;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$435|0);
        // unreachable;
       }
      }
     }
    } else {
     $$341 = $0;
    }
   } else {
    $$341 = $0;
   }
  } else {
   $$341 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$341|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_115parse_call_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$063 = 0, $$6 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0;
 var $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $9 = 0, $exitcond$i$i = 0, $exitcond$i$i$i$i = 0, $or$cond = 0, $or$cond77 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 25|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(3);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(99);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = ($14<<24>>24)==(108);
    if ($15) {
     $16 = ((($0)) + 2|0);
     $17 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $18 = ($17|0)==($16|0);
     $19 = ($17|0)==($1|0);
     $or$cond77 = $18 | $19;
     if ($or$cond77) {
      $$6 = $0;
     } else {
      $20 = HEAP32[$2>>2]|0;
      $21 = ((($2)) + 4|0);
      $22 = HEAP32[$21>>2]|0;
      $23 = ($20|0)==($22|0);
      if ($23) {
       $$6 = $0;
      } else {
       $24 = ((($22)) + -12|0);
       $25 = ((($22)) + -24|0);
       $26 = ((($24)) + 11|0);
       $27 = HEAP8[$26>>0]|0;
       $28 = ($27<<24>>24)<(0);
       $29 = HEAP32[$24>>2]|0;
       $30 = $28 ? $29 : $24;
       $31 = ((($22)) + -8|0);
       $32 = HEAP32[$31>>2]|0;
       $33 = $27&255;
       $34 = $28 ? $32 : $33;
       (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj($25,$30,$34)|0);
       ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
       $$0$i$i = 0;
       while(1) {
        $exitcond$i$i = ($$0$i$i|0)==(3);
        if ($exitcond$i$i) {
         break;
        }
        $35 = (($5) + ($$0$i$i<<2)|0);
        HEAP32[$35>>2] = 0;
        $36 = (($$0$i$i) + 1)|0;
        $$0$i$i = $36;
       }
       $37 = HEAP32[$21>>2]|0;
       $38 = ((($37)) + -12|0);
       $39 = ((($38)) + 11|0);
       $40 = HEAP8[$39>>0]|0;
       $41 = ($40<<24>>24)<(0);
       if ($41) {
        $42 = HEAP32[$38>>2]|0;
        HEAP8[$3>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($42,$3);
        $43 = ((($37)) + -8|0);
        HEAP32[$43>>2] = 0;
       } else {
        HEAP8[$4>>0] = 0;
        __ZNSt3__211char_traitsIcE6assignERcRKc($38,$4);
        HEAP8[$39>>0] = 0;
       }
       __THREW__ = 0;
       invoke_vi(1116,($38|0));
       $44 = __THREW__; __THREW__ = 0;
       $45 = $44&1;
       if ($45) {
        $46 = ___cxa_find_matching_catch_3(0|0)|0;
        $47 = tempRet0;
        ___clang_call_terminate($46);
        // unreachable;
       }
       ;HEAP32[$38>>2]=HEAP32[$5>>2]|0;HEAP32[$38+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$38+8>>2]=HEAP32[$5+8>>2]|0;
       $$0$i$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
        if ($exitcond$i$i$i$i) {
         break;
        }
        $48 = (($5) + ($$0$i$i$i$i<<2)|0);
        HEAP32[$48>>2] = 0;
        $49 = (($$0$i$i$i$i) + 1)|0;
        $$0$i$i$i$i = $49;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
       $50 = HEAP32[$21>>2]|0;
       $51 = ((($50)) + -24|0);
       (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($51,159083)|0);
       $52 = ((($6)) + 11|0);
       $53 = ((($6)) + 4|0);
       $$063 = $17;
       while(1) {
        $54 = HEAP8[$$063>>0]|0;
        $55 = ($54<<24>>24)==(69);
        if ($55) {
         label = 30;
         break;
        }
        $56 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($$063,$1,$2)|0);
        $57 = ($56|0)==($$063|0);
        $58 = ($56|0)==($1|0);
        $or$cond = $57 | $58;
        if ($or$cond) {
         $$6 = $0;
         break L1;
        }
        $59 = HEAP32[$2>>2]|0;
        $60 = HEAP32[$21>>2]|0;
        $61 = ($59|0)==($60|0);
        if ($61) {
         $$6 = $0;
         break L1;
        }
        $62 = ((($60)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$62);
        $63 = HEAP32[$21>>2]|0;
        $64 = ((($63)) + -24|0);
        $66 = $63;
        while(1) {
         $65 = ($66|0)==($64|0);
         if ($65) {
          break;
         }
         $67 = ((($66)) + -24|0);
         HEAP32[$21>>2] = $67;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($67);
         $$pre$i = HEAP32[$21>>2]|0;
         $66 = $$pre$i;
        }
        $68 = HEAP8[$52>>0]|0;
        $69 = ($68<<24>>24)<(0);
        $70 = HEAP32[$53>>2]|0;
        $71 = $68&255;
        $72 = $69 ? $70 : $71;
        $73 = ($72|0)==(0);
        if (!($73)) {
         $74 = HEAP32[$2>>2]|0;
         $75 = ($74|0)==($64|0);
         if ($75) {
          label = 29;
          break;
         }
         $78 = ((($63)) + -48|0);
         $79 = HEAP32[$6>>2]|0;
         $80 = $69 ? $79 : $6;
         __THREW__ = 0;
         (invoke_iiii(1112,($78|0),($80|0),($72|0))|0);
         $81 = __THREW__; __THREW__ = 0;
         $82 = $81&1;
         if ($82) {
          label = 26;
          break;
         }
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        $$063 = $56;
       }
       if ((label|0) == 26) {
        $76 = ___cxa_find_matching_catch_2()|0;
        $77 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        ___resumeException($76|0);
        // unreachable;
       }
       else if ((label|0) == 29) {
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        $$6 = $0;
        break;
       }
       else if ((label|0) == 30) {
        $83 = HEAP32[$2>>2]|0;
        $84 = HEAP32[$21>>2]|0;
        $85 = ($83|0)==($84|0);
        if ($85) {
         $$6 = $0;
         break;
        }
        $86 = ((($$063)) + 1|0);
        $87 = ((($84)) + -24|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($87,159081)|0);
        $$6 = $86;
        break;
       }
      }
     }
    } else {
     $$6 = $0;
    }
   } else {
    $$6 = $0;
   }
  } else {
   $$6 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$6|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_conversion_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i107 = 0, $$0$i$i$i110 = 0, $$0$i$i$i113 = 0, $$1 = 0, $$10 = 0, $$1100 = 0, $$175 = 0, $$2 = 0, $$2101 = 0, $$276 = 0, $$3 = 0, $$377 = 0, $$4 = 0, $$4103 = 0, $$478 = 0, $$5104 = 0, $$pre = 0, $$pre$i = 0, $10 = 0;
 var $100 = 0, $101 = 0, $102 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0;
 var $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0;
 var $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0;
 var $exitcond$i$i$i108 = 0, $exitcond$i$i$i111 = 0, $exitcond$i$i$i114 = 0, $or$cond = 0, $or$cond106 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 88|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(99);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(118);
    if ($18) {
     $19 = ((($2)) + 63|0);
     $20 = HEAP8[$19>>0]|0;
     HEAP8[$19>>0] = 0;
     $21 = ((($0)) + 2|0);
     $22 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($21,$1,$2)|0);
     HEAP8[$19>>0] = $20;
     $23 = ($22|0)==($21|0);
     $24 = ($22|0)==($1|0);
     $or$cond = $23 | $24;
     if ($or$cond) {
      $$10 = $0;
     } else {
      $25 = HEAP8[$22>>0]|0;
      $26 = ($25<<24>>24)==(95);
      if ($26) {
       $29 = ((($22)) + 1|0);
       $30 = ($29|0)==($1|0);
       if ($30) {
        $$10 = $0;
        break;
       }
       $31 = HEAP8[$29>>0]|0;
       $32 = ($31<<24>>24)==(69);
       L9: do {
        if ($32) {
         $33 = ((($2)) + 4|0);
         $34 = HEAP32[$33>>2]|0;
         $35 = ((($2)) + 8|0);
         $36 = HEAP32[$35>>2]|0;
         $37 = ($34>>>0)<($36>>>0);
         if ($37) {
          ;HEAP32[$34>>2]=0|0;HEAP32[$34+4>>2]=0|0;HEAP32[$34+8>>2]=0|0;HEAP32[$34+12>>2]=0|0;HEAP32[$34+16>>2]=0|0;HEAP32[$34+20>>2]=0|0;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2Ev($34);
          $38 = HEAP32[$33>>2]|0;
          $39 = ((($38)) + 24|0);
          HEAP32[$33>>2] = $39;
          $$4103 = $29;
          break;
         } else {
          __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE24__emplace_back_slow_pathIJEEEvDpOT_($2);
          $$4103 = $29;
          break;
         }
        } else {
         $$1100 = $29;$41 = $31;
         while(1) {
          $40 = ($41<<24>>24)==(69);
          if ($40) {
           $$4103 = $$1100;
           break L9;
          }
          $42 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($$1100,$1,$2)|0);
          $43 = ($42|0)==($$1100|0);
          $44 = ($42|0)==($1|0);
          $or$cond106 = $43 | $44;
          if ($or$cond106) {
           $$10 = $0;
           break L1;
          }
          $$2101 = $or$cond106 ? $$1100 : $42;
          $$pre = HEAP8[$$2101>>0]|0;
          $$1100 = $42;$41 = $$pre;
         }
        }
       } while(0);
       $45 = ((($$4103)) + 1|0);
       $$5104 = $45;
      } else {
       $27 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($22,$1,$2)|0);
       $28 = ($27|0)==($22|0);
       if ($28) {
        $$10 = $0;
        break;
       } else {
        $$5104 = $27;
       }
      }
      $46 = ((($2)) + 4|0);
      $47 = HEAP32[$46>>2]|0;
      $48 = HEAP32[$2>>2]|0;
      $49 = (($47) - ($48))|0;
      $50 = (($49|0) / 24)&-1;
      $51 = ($50>>>0)<(2);
      if ($51) {
       $$10 = $0;
      } else {
       $52 = $47;
       $53 = ((($52)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$53);
       $54 = HEAP32[$46>>2]|0;
       $55 = ((($54)) + -24|0);
       $57 = $54;
       while(1) {
        $56 = ($57|0)==($55|0);
        if ($56) {
         break;
        }
        $58 = ((($57)) + -24|0);
        HEAP32[$46>>2] = $58;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($58);
        $$pre$i = HEAP32[$46>>2]|0;
        $57 = $$pre$i;
       }
       $59 = ((($54)) + -48|0);
       __THREW__ = 0;
       invoke_vii(1120,($9|0),($59|0));
       $60 = __THREW__; __THREW__ = 0;
       $61 = $60&1;
       if ($61) {
        $93 = ___cxa_find_matching_catch_2()|0;
        $94 = tempRet0;
        $$4 = $94;$$478 = $93;
       } else {
        __THREW__ = 0;
        $62 = (invoke_iiii(1121,($9|0),0,(159083|0))|0);
        $63 = __THREW__; __THREW__ = 0;
        $64 = $63&1;
        if ($64) {
         $95 = ___cxa_find_matching_catch_2()|0;
         $96 = tempRet0;
         $$3 = $96;$$377 = $95;
        } else {
         ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $65 = (($9) + ($$0$i$i$i<<2)|0);
          HEAP32[$65>>2] = 0;
          $66 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $66;
         }
         __THREW__ = 0;
         $67 = (invoke_iii(1119,($8|0),(160247|0))|0);
         $68 = __THREW__; __THREW__ = 0;
         $69 = $68&1;
         if ($69) {
          $97 = ___cxa_find_matching_catch_2()|0;
          $98 = tempRet0;
          $$2 = $98;$$276 = $97;
         } else {
          ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
          $$0$i$i$i107 = 0;
          while(1) {
           $exitcond$i$i$i108 = ($$0$i$i$i107|0)==(3);
           if ($exitcond$i$i$i108) {
            break;
           }
           $70 = (($8) + ($$0$i$i$i107<<2)|0);
           HEAP32[$70>>2] = 0;
           $71 = (($$0$i$i$i107) + 1)|0;
           $$0$i$i$i107 = $71;
          }
          $72 = ((($3)) + 11|0);
          $73 = HEAP8[$72>>0]|0;
          $74 = ($73<<24>>24)<(0);
          $75 = HEAP32[$3>>2]|0;
          $76 = $74 ? $75 : $3;
          $77 = ((($3)) + 4|0);
          $78 = HEAP32[$77>>2]|0;
          $79 = $73&255;
          $80 = $74 ? $78 : $79;
          __THREW__ = 0;
          $81 = (invoke_iiii(1112,($7|0),($76|0),($80|0))|0);
          $82 = __THREW__; __THREW__ = 0;
          $83 = $82&1;
          do {
           if ($83) {
            $99 = ___cxa_find_matching_catch_2()|0;
            $100 = tempRet0;
            $$1 = $100;$$175 = $99;
           } else {
            ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
            $$0$i$i$i110 = 0;
            while(1) {
             $exitcond$i$i$i111 = ($$0$i$i$i110|0)==(3);
             if ($exitcond$i$i$i111) {
              break;
             }
             $84 = (($7) + ($$0$i$i$i110<<2)|0);
             HEAP32[$84>>2] = 0;
             $85 = (($$0$i$i$i110) + 1)|0;
             $$0$i$i$i110 = $85;
            }
            __THREW__ = 0;
            $86 = (invoke_iii(1119,($6|0),(159081|0))|0);
            $87 = __THREW__; __THREW__ = 0;
            $88 = $87&1;
            if ($88) {
             $101 = ___cxa_find_matching_catch_2()|0;
             $102 = tempRet0;
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
             $$1 = $102;$$175 = $101;
             break;
            }
            ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
            $$0$i$i$i113 = 0;
            while(1) {
             $exitcond$i$i$i114 = ($$0$i$i$i113|0)==(3);
             if ($exitcond$i$i$i114) {
              break;
             }
             $89 = (($6) + ($$0$i$i$i113<<2)|0);
             HEAP32[$89>>2] = 0;
             $90 = (($$0$i$i$i113) + 1)|0;
             $$0$i$i$i113 = $90;
            }
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
            $91 = HEAP32[$46>>2]|0;
            $92 = ((($91)) + -24|0);
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($92,$4);
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
            $$10 = $$5104;
            break L1;
           }
          } while(0);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
          $$2 = $$1;$$276 = $$175;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         $$3 = $$2;$$377 = $$276;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
        $$4 = $$3;$$478 = $$377;
       }
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
       ___resumeException($$478|0);
       // unreachable;
      }
     }
    } else {
     $$10 = $0;
    }
   } else {
    $$10 = $0;
   }
  } else {
   $$10 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$10|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_123parse_dynamic_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i47 = 0, $$0$i$i$i50 = 0, $$0$i$i$i53 = 0, $$1 = 0, $$132 = 0, $$2 = 0, $$233 = 0, $$3 = 0, $$334 = 0, $$341 = 0, $$4 = 0, $$435 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i48 = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i54 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 88|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(100);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(99);
    if ($18) {
     $19 = ((($0)) + 2|0);
     $20 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($19,$1,$2)|0);
     $21 = ($20|0)==($19|0);
     if ($21) {
      $$341 = $0;
     } else {
      $22 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($20,$1,$2)|0);
      $23 = ($22|0)==($20|0);
      if ($23) {
       $$341 = $0;
      } else {
       $24 = ((($2)) + 4|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = HEAP32[$2>>2]|0;
       $27 = (($25) - ($26))|0;
       $28 = (($27|0) / 24)&-1;
       $29 = ($28>>>0)<(2);
       if ($29) {
        $$341 = $0;
       } else {
        $30 = $25;
        $31 = ((($30)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$31);
        $32 = HEAP32[$24>>2]|0;
        $33 = ((($32)) + -24|0);
        $35 = $32;
        while(1) {
         $34 = ($35|0)==($33|0);
         if ($34) {
          break;
         }
         $36 = ((($35)) + -24|0);
         HEAP32[$24>>2] = $36;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
         $$pre$i = HEAP32[$24>>2]|0;
         $35 = $$pre$i;
        }
        $37 = ((($32)) + -48|0);
        __THREW__ = 0;
        invoke_vii(1120,($9|0),($37|0));
        $38 = __THREW__; __THREW__ = 0;
        $39 = $38&1;
        if ($39) {
         $71 = ___cxa_find_matching_catch_2()|0;
         $72 = tempRet0;
         $$4 = $72;$$435 = $71;
        } else {
         __THREW__ = 0;
         $40 = (invoke_iiii(1121,($9|0),0,(160233|0))|0);
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         if ($42) {
          $73 = ___cxa_find_matching_catch_2()|0;
          $74 = tempRet0;
          $$3 = $74;$$334 = $73;
         } else {
          ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $43 = (($9) + ($$0$i$i$i<<2)|0);
           HEAP32[$43>>2] = 0;
           $44 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $44;
          }
          __THREW__ = 0;
          $45 = (invoke_iii(1119,($8|0),(159330|0))|0);
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           $75 = ___cxa_find_matching_catch_2()|0;
           $76 = tempRet0;
           $$2 = $76;$$233 = $75;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i47 = 0;
           while(1) {
            $exitcond$i$i$i48 = ($$0$i$i$i47|0)==(3);
            if ($exitcond$i$i$i48) {
             break;
            }
            $48 = (($8) + ($$0$i$i$i47<<2)|0);
            HEAP32[$48>>2] = 0;
            $49 = (($$0$i$i$i47) + 1)|0;
            $$0$i$i$i47 = $49;
           }
           $50 = ((($3)) + 11|0);
           $51 = HEAP8[$50>>0]|0;
           $52 = ($51<<24>>24)<(0);
           $53 = HEAP32[$3>>2]|0;
           $54 = $52 ? $53 : $3;
           $55 = ((($3)) + 4|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $51&255;
           $58 = $52 ? $56 : $57;
           __THREW__ = 0;
           $59 = (invoke_iiii(1112,($7|0),($54|0),($58|0))|0);
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           do {
            if ($61) {
             $77 = ___cxa_find_matching_catch_2()|0;
             $78 = tempRet0;
             $$1 = $78;$$132 = $77;
            } else {
             ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
             $$0$i$i$i50 = 0;
             while(1) {
              $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
              if ($exitcond$i$i$i51) {
               break;
              }
              $62 = (($7) + ($$0$i$i$i50<<2)|0);
              HEAP32[$62>>2] = 0;
              $63 = (($$0$i$i$i50) + 1)|0;
              $$0$i$i$i50 = $63;
             }
             __THREW__ = 0;
             $64 = (invoke_iii(1119,($6|0),(159081|0))|0);
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              $79 = ___cxa_find_matching_catch_2()|0;
              $80 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
              $$1 = $80;$$132 = $79;
              break;
             }
             ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
             $$0$i$i$i53 = 0;
             while(1) {
              $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
              if ($exitcond$i$i$i54) {
               break;
              }
              $67 = (($6) + ($$0$i$i$i53<<2)|0);
              HEAP32[$67>>2] = 0;
              $68 = (($$0$i$i$i53) + 1)|0;
              $$0$i$i$i53 = $68;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
             $69 = HEAP32[$24>>2]|0;
             $70 = ((($69)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($70,$4);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
             $$341 = $22;
             break L1;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
           $$2 = $$1;$$233 = $$132;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
          $$3 = $$2;$$334 = $$233;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$4 = $$3;$$435 = $$334;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$435|0);
        // unreachable;
       }
      }
     }
    } else {
     $$341 = $0;
    }
   } else {
    $$341 = $0;
   }
  } else {
   $$341 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$341|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i207 = 0, $$0$i$i$i221 = 0, $$0$i$i$i229 = 0, $$0$i$i$i239 = 0, $$0163 = 0, $$0182$off0 = 0, $$0183 = 0, $$1 = 0, $$1164 = 0, $$1184 = 0, $$1184$ph = 0, $$2185 = 0, $$3 = 0, $$3166 = 0, $$3186 = 0, $$4 = 0, $$4167 = 0;
 var $$5 = 0, $$5168 = 0, $$6181 = 0, $$7 = 0, $$7170 = 0, $$pre = 0, $$pre$i = 0, $$pre$i197 = 0, $$pre$i199 = 0, $$pre$i203 = 0, $$pre$i205 = 0, $$pre$i213 = 0, $$pre$i217 = 0, $$pre$i219 = 0, $$pre$i227 = 0, $$pre$i235 = 0, $$pre$i237 = 0, $$pre$phi281Z2D = 0, $$pre$phi283Z2D = 0, $$pre282 = 0;
 var $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0;
 var $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0;
 var $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0;
 var $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0;
 var $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0;
 var $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0;
 var $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0;
 var $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0;
 var $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0;
 var $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0;
 var $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0;
 var $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0;
 var $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i208 = 0, $exitcond$i$i$i222 = 0, $exitcond$i$i$i230 = 0, $exitcond$i$i$i240 = 0, $or$cond = 0, $or$cond194 = 0, $or$cond195 = 0, $or$cond196 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 144|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(144|0);
 $3 = sp + 132|0;
 $4 = sp + 120|0;
 $5 = sp + 108|0;
 $6 = sp + 96|0;
 $7 = sp + 84|0;
 $8 = sp + 72|0;
 $9 = sp + 60|0;
 $10 = sp + 48|0;
 $11 = sp + 36|0;
 $12 = sp + 24|0;
 $13 = sp + 12|0;
 $14 = sp;
 $15 = $1;
 $16 = $0;
 $17 = (($15) - ($16))|0;
 $18 = ($17|0)>(2);
 L1: do {
  if ($18) {
   $19 = HEAP8[$0>>0]|0;
   $20 = ($19<<24>>24)==(103);
   if ($20) {
    $21 = ((($0)) + 1|0);
    $22 = HEAP8[$21>>0]|0;
    $23 = ($22<<24>>24)==(115);
    $24 = ((($0)) + 2|0);
    $$ = $23 ? $24 : $0;
    $$0182$off0 = $23;$$0183 = $$;
   } else {
    $$0182$off0 = 0;$$0183 = $0;
   }
   $25 = (__ZN10__cxxabiv112_GLOBAL__N_126parse_base_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($$0183,$1,$2)|0);
   $26 = ($25|0)==($$0183|0);
   if (!($26)) {
    if (!($$0182$off0)) {
     $$6181 = $25;
     break;
    }
    $27 = HEAP32[$2>>2]|0;
    $28 = ((($2)) + 4|0);
    $29 = HEAP32[$28>>2]|0;
    $30 = ($27|0)==($29|0);
    if ($30) {
     $$6181 = $0;
     break;
    }
    $31 = ((($29)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($31,0,159155)|0);
    $$6181 = $25;
    break;
   }
   $32 = $$0183;
   $33 = (($15) - ($32))|0;
   $34 = ($33|0)>(2);
   if ($34) {
    $35 = HEAP8[$$0183>>0]|0;
    $36 = ($35<<24>>24)==(115);
    if ($36) {
     $37 = ((($$0183)) + 1|0);
     $38 = HEAP8[$37>>0]|0;
     $39 = ($38<<24>>24)==(114);
     if ($39) {
      $40 = ((($$0183)) + 2|0);
      $41 = HEAP8[$40>>0]|0;
      $42 = ($41<<24>>24)==(78);
      L14: do {
       if ($42) {
        $43 = ((($$0183)) + 3|0);
        $44 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_typeINS0_2DbEEEPKcS4_S4_RT_($43,$1,$2)|0);
        $45 = ($44|0)==($43|0);
        $46 = ($44|0)==($1|0);
        $or$cond = $45 | $46;
        if ($or$cond) {
         $$6181 = $0;
         break L1;
        }
        $47 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($44,$1,$2)|0);
        $48 = ($47|0)==($44|0);
        $$pre = ((($2)) + 4|0);
        do {
         if ($48) {
          $$1184$ph = $44;$$pre$phi281Z2D = $2;
         } else {
          $49 = HEAP32[$$pre>>2]|0;
          $50 = HEAP32[$2>>2]|0;
          $51 = (($49) - ($50))|0;
          $52 = (($51|0) / 24)&-1;
          $53 = ($52>>>0)<(2);
          if ($53) {
           $$6181 = $0;
           break L1;
          }
          $54 = $49;
          $55 = ((($54)) + -24|0);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$55);
          $56 = HEAP32[$$pre>>2]|0;
          $57 = ((($56)) + -24|0);
          $59 = $56;
          while(1) {
           $58 = ($59|0)==($57|0);
           if ($58) {
            break;
           }
           $60 = ((($59)) + -24|0);
           HEAP32[$$pre>>2] = $60;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($60);
           $$pre$i = HEAP32[$$pre>>2]|0;
           $59 = $$pre$i;
          }
          $61 = ((($56)) + -48|0);
          $62 = ((($3)) + 11|0);
          $63 = HEAP8[$62>>0]|0;
          $64 = ($63<<24>>24)<(0);
          $65 = HEAP32[$3>>2]|0;
          $66 = $64 ? $65 : $3;
          $67 = ((($3)) + 4|0);
          $68 = HEAP32[$67>>2]|0;
          $69 = $63&255;
          $70 = $64 ? $68 : $69;
          __THREW__ = 0;
          (invoke_iiii(1112,($61|0),($66|0),($70|0))|0);
          $71 = __THREW__; __THREW__ = 0;
          $72 = $71&1;
          if ($72) {
           $79 = ___cxa_find_matching_catch_2()|0;
           $80 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
           $$7 = $80;$$7170 = $79;
           break L14;
          }
          $73 = ($47|0)==($1|0);
          if (!($73)) {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
           $$1184$ph = $47;$$pre$phi281Z2D = $2;
           break;
          }
          $74 = HEAP32[$$pre>>2]|0;
          $75 = ((($74)) + -24|0);
          $77 = $74;
          while(1) {
           $76 = ($77|0)==($75|0);
           if ($76) {
            break;
           }
           $78 = ((($77)) + -24|0);
           HEAP32[$$pre>>2] = $78;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($78);
           $$pre$i197 = HEAP32[$$pre>>2]|0;
           $77 = $$pre$i197;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
          $$6181 = $0;
          break L1;
         }
        } while(0);
        $81 = ((($5)) + 11|0);
        $82 = ((($5)) + 4|0);
        $$1184 = $$1184$ph;
        while(1) {
         $83 = HEAP8[$$1184>>0]|0;
         $84 = ($83<<24>>24)==(69);
         if ($84) {
          label = 43;
          break;
         }
         $85 = (__ZN10__cxxabiv112_GLOBAL__N_132parse_unresolved_qualifier_levelINS0_2DbEEEPKcS4_S4_RT_($$1184,$1,$2)|0);
         $86 = ($85|0)==($$1184|0);
         $87 = ($85|0)==($1|0);
         $or$cond194 = $86 | $87;
         if ($or$cond194) {
          $$6181 = $0;
          break L1;
         }
         $88 = HEAP32[$$pre>>2]|0;
         $89 = HEAP32[$$pre$phi281Z2D>>2]|0;
         $90 = (($88) - ($89))|0;
         $91 = (($90|0) / 24)&-1;
         $92 = ($91>>>0)<(2);
         if ($92) {
          $$6181 = $0;
          break L1;
         }
         $93 = $88;
         $94 = ((($93)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($4,$94);
         $95 = HEAP32[$$pre>>2]|0;
         $96 = ((($95)) + -24|0);
         $98 = $95;
         while(1) {
          $97 = ($98|0)==($96|0);
          if ($97) {
           break;
          }
          $99 = ((($98)) + -24|0);
          HEAP32[$$pre>>2] = $99;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($99);
          $$pre$i199 = HEAP32[$$pre>>2]|0;
          $98 = $$pre$i199;
         }
         __THREW__ = 0;
         $100 = (invoke_iiii(1121,($4|0),0,(159155|0))|0);
         $101 = __THREW__; __THREW__ = 0;
         $102 = $101&1;
         if ($102) {
          label = 40;
          break;
         }
         ;HEAP32[$5>>2]=HEAP32[$4>>2]|0;HEAP32[$5+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$4+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $103 = (($4) + ($$0$i$i$i<<2)|0);
          HEAP32[$103>>2] = 0;
          $104 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $104;
         }
         $105 = HEAP32[$$pre>>2]|0;
         $106 = ((($105)) + -24|0);
         $107 = HEAP8[$81>>0]|0;
         $108 = ($107<<24>>24)<(0);
         $109 = HEAP32[$5>>2]|0;
         $110 = $108 ? $109 : $5;
         $111 = HEAP32[$82>>2]|0;
         $112 = $107&255;
         $113 = $108 ? $111 : $112;
         __THREW__ = 0;
         (invoke_iiii(1112,($106|0),($110|0),($113|0))|0);
         $114 = __THREW__; __THREW__ = 0;
         $115 = $114&1;
         if ($115) {
          label = 41;
          break;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         $$1184 = $85;
        }
        if ((label|0) == 40) {
         $116 = ___cxa_find_matching_catch_2()|0;
         $117 = tempRet0;
         $$0 = $117;$$0163 = $116;
        }
        else if ((label|0) == 41) {
         $118 = ___cxa_find_matching_catch_2()|0;
         $119 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         $$0 = $119;$$0163 = $118;
        }
        else if ((label|0) == 43) {
         $120 = ((($$1184)) + 1|0);
         $121 = (__ZN10__cxxabiv112_GLOBAL__N_126parse_base_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($120,$1,$2)|0);
         $122 = ($121|0)==($120|0);
         if ($122) {
          $123 = HEAP32[$2>>2]|0;
          $124 = HEAP32[$$pre>>2]|0;
          $125 = ($123|0)==($124|0);
          if ($125) {
           $$6181 = $0;
           break L1;
          }
          $126 = ((($124)) + -24|0);
          $128 = $124;
          while(1) {
           $127 = ($128|0)==($126|0);
           if ($127) {
            $$6181 = $0;
            break L1;
           }
           $129 = ((($128)) + -24|0);
           HEAP32[$$pre>>2] = $129;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($129);
           $$pre$i203 = HEAP32[$$pre>>2]|0;
           $128 = $$pre$i203;
          }
         }
         $130 = HEAP32[$$pre>>2]|0;
         $131 = HEAP32[$$pre$phi281Z2D>>2]|0;
         $132 = (($130) - ($131))|0;
         $133 = (($132|0) / 24)&-1;
         $134 = ($133>>>0)<(2);
         if ($134) {
          $$6181 = $0;
          break L1;
         }
         $135 = $130;
         $136 = ((($135)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$136);
         $137 = HEAP32[$$pre>>2]|0;
         $138 = ((($137)) + -24|0);
         $140 = $137;
         while(1) {
          $139 = ($140|0)==($138|0);
          if ($139) {
           break;
          }
          $141 = ((($140)) + -24|0);
          HEAP32[$$pre>>2] = $141;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($141);
          $$pre$i205 = HEAP32[$$pre>>2]|0;
          $140 = $$pre$i205;
         }
         __THREW__ = 0;
         $142 = (invoke_iiii(1121,($6|0),0,(159155|0))|0);
         $143 = __THREW__; __THREW__ = 0;
         $144 = $143&1;
         do {
          if ($144) {
           $160 = ___cxa_find_matching_catch_2()|0;
           $161 = tempRet0;
           $$1 = $161;$$1164 = $160;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$6>>2]|0;HEAP32[$7+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$6+8>>2]|0;
           $$0$i$i$i207 = 0;
           while(1) {
            $exitcond$i$i$i208 = ($$0$i$i$i207|0)==(3);
            if ($exitcond$i$i$i208) {
             break;
            }
            $145 = (($6) + ($$0$i$i$i207<<2)|0);
            HEAP32[$145>>2] = 0;
            $146 = (($$0$i$i$i207) + 1)|0;
            $$0$i$i$i207 = $146;
           }
           $147 = HEAP32[$$pre>>2]|0;
           $148 = ((($147)) + -24|0);
           $149 = ((($7)) + 11|0);
           $150 = HEAP8[$149>>0]|0;
           $151 = ($150<<24>>24)<(0);
           $152 = HEAP32[$7>>2]|0;
           $153 = $151 ? $152 : $7;
           $154 = ((($7)) + 4|0);
           $155 = HEAP32[$154>>2]|0;
           $156 = $150&255;
           $157 = $151 ? $155 : $156;
           __THREW__ = 0;
           (invoke_iiii(1112,($148|0),($153|0),($157|0))|0);
           $158 = __THREW__; __THREW__ = 0;
           $159 = $158&1;
           if ($159) {
            $162 = ___cxa_find_matching_catch_2()|0;
            $163 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            $$1 = $163;$$1164 = $162;
            break;
           } else {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
            $$6181 = $121;
            break L1;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         $$7 = $$1;$$7170 = $$1164;
         break;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
        $$7 = $$0;$$7170 = $$0163;
       } else {
        $164 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_typeINS0_2DbEEEPKcS4_S4_RT_($40,$1,$2)|0);
        $165 = ($164|0)==($40|0);
        if (!($165)) {
         $166 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($164,$1,$2)|0);
         $167 = ($166|0)==($164|0);
         do {
          if ($167) {
           $$2185 = $164;
          } else {
           $168 = ((($2)) + 4|0);
           $169 = HEAP32[$168>>2]|0;
           $170 = HEAP32[$2>>2]|0;
           $171 = (($169) - ($170))|0;
           $172 = (($171|0) / 24)&-1;
           $173 = ($172>>>0)<(2);
           if ($173) {
            $$6181 = $0;
            break L1;
           }
           $174 = $169;
           $175 = ((($174)) + -24|0);
           __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($8,$175);
           $176 = HEAP32[$168>>2]|0;
           $177 = ((($176)) + -24|0);
           $179 = $176;
           while(1) {
            $178 = ($179|0)==($177|0);
            if ($178) {
             break;
            }
            $180 = ((($179)) + -24|0);
            HEAP32[$168>>2] = $180;
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($180);
            $$pre$i213 = HEAP32[$168>>2]|0;
            $179 = $$pre$i213;
           }
           $181 = ((($176)) + -48|0);
           $182 = ((($8)) + 11|0);
           $183 = HEAP8[$182>>0]|0;
           $184 = ($183<<24>>24)<(0);
           $185 = HEAP32[$8>>2]|0;
           $186 = $184 ? $185 : $8;
           $187 = ((($8)) + 4|0);
           $188 = HEAP32[$187>>2]|0;
           $189 = $183&255;
           $190 = $184 ? $188 : $189;
           __THREW__ = 0;
           (invoke_iiii(1112,($181|0),($186|0),($190|0))|0);
           $191 = __THREW__; __THREW__ = 0;
           $192 = $191&1;
           if ($192) {
            $193 = ___cxa_find_matching_catch_2()|0;
            $194 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            $$7 = $194;$$7170 = $193;
            break L14;
           } else {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
            $$2185 = $166;
            break;
           }
          }
         } while(0);
         $195 = (__ZN10__cxxabiv112_GLOBAL__N_126parse_base_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($$2185,$1,$2)|0);
         $196 = ($195|0)==($$2185|0);
         if ($196) {
          $197 = HEAP32[$2>>2]|0;
          $198 = ((($2)) + 4|0);
          $199 = HEAP32[$198>>2]|0;
          $200 = ($197|0)==($199|0);
          if ($200) {
           $$6181 = $0;
           break L1;
          }
          $201 = ((($199)) + -24|0);
          $203 = $199;
          while(1) {
           $202 = ($203|0)==($201|0);
           if ($202) {
            $$6181 = $0;
            break L1;
           }
           $204 = ((($203)) + -24|0);
           HEAP32[$198>>2] = $204;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($204);
           $$pre$i217 = HEAP32[$198>>2]|0;
           $203 = $$pre$i217;
          }
         }
         $205 = ((($2)) + 4|0);
         $206 = HEAP32[$205>>2]|0;
         $207 = HEAP32[$2>>2]|0;
         $208 = (($206) - ($207))|0;
         $209 = (($208|0) / 24)&-1;
         $210 = ($209>>>0)<(2);
         if ($210) {
          $$6181 = $0;
          break L1;
         }
         $211 = $206;
         $212 = ((($211)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($9,$212);
         $213 = HEAP32[$205>>2]|0;
         $214 = ((($213)) + -24|0);
         $216 = $213;
         while(1) {
          $215 = ($216|0)==($214|0);
          if ($215) {
           break;
          }
          $217 = ((($216)) + -24|0);
          HEAP32[$205>>2] = $217;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($217);
          $$pre$i219 = HEAP32[$205>>2]|0;
          $216 = $$pre$i219;
         }
         __THREW__ = 0;
         $218 = (invoke_iiii(1121,($9|0),0,(159155|0))|0);
         $219 = __THREW__; __THREW__ = 0;
         $220 = $219&1;
         do {
          if ($220) {
           $236 = ___cxa_find_matching_catch_2()|0;
           $237 = tempRet0;
           $$3 = $237;$$3166 = $236;
          } else {
           ;HEAP32[$10>>2]=HEAP32[$9>>2]|0;HEAP32[$10+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$9+8>>2]|0;
           $$0$i$i$i221 = 0;
           while(1) {
            $exitcond$i$i$i222 = ($$0$i$i$i221|0)==(3);
            if ($exitcond$i$i$i222) {
             break;
            }
            $221 = (($9) + ($$0$i$i$i221<<2)|0);
            HEAP32[$221>>2] = 0;
            $222 = (($$0$i$i$i221) + 1)|0;
            $$0$i$i$i221 = $222;
           }
           $223 = HEAP32[$205>>2]|0;
           $224 = ((($223)) + -24|0);
           $225 = ((($10)) + 11|0);
           $226 = HEAP8[$225>>0]|0;
           $227 = ($226<<24>>24)<(0);
           $228 = HEAP32[$10>>2]|0;
           $229 = $227 ? $228 : $10;
           $230 = ((($10)) + 4|0);
           $231 = HEAP32[$230>>2]|0;
           $232 = $226&255;
           $233 = $227 ? $231 : $232;
           __THREW__ = 0;
           (invoke_iiii(1112,($224|0),($229|0),($233|0))|0);
           $234 = __THREW__; __THREW__ = 0;
           $235 = $234&1;
           if ($235) {
            $238 = ___cxa_find_matching_catch_2()|0;
            $239 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
            $$3 = $239;$$3166 = $238;
            break;
           } else {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
            $$6181 = $195;
            break L1;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$7 = $$3;$$7170 = $$3166;
         break;
        }
        $240 = (__ZN10__cxxabiv112_GLOBAL__N_132parse_unresolved_qualifier_levelINS0_2DbEEEPKcS4_S4_RT_($40,$1,$2)|0);
        $241 = ($240|0)==($40|0);
        $242 = ($240|0)==($1|0);
        $or$cond195 = $241 | $242;
        if ($or$cond195) {
         $$6181 = $0;
         break L1;
        }
        if ($$0182$off0) {
         $243 = HEAP32[$2>>2]|0;
         $244 = ((($2)) + 4|0);
         $245 = HEAP32[$244>>2]|0;
         $246 = ($243|0)==($245|0);
         if ($246) {
          $$6181 = $0;
          break L1;
         }
         $247 = ((($245)) + -24|0);
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($247,0,159155)|0);
         $$pre$phi283Z2D = $244;
        } else {
         $$pre282 = ((($2)) + 4|0);
         $$pre$phi283Z2D = $$pre282;
        }
        $248 = ((($12)) + 11|0);
        $249 = ((($12)) + 4|0);
        $$3186 = $240;
        while(1) {
         $250 = HEAP8[$$3186>>0]|0;
         $251 = ($250<<24>>24)==(69);
         if ($251) {
          label = 109;
          break;
         }
         $252 = (__ZN10__cxxabiv112_GLOBAL__N_132parse_unresolved_qualifier_levelINS0_2DbEEEPKcS4_S4_RT_($$3186,$1,$2)|0);
         $253 = ($252|0)==($$3186|0);
         $254 = ($252|0)==($1|0);
         $or$cond196 = $253 | $254;
         if ($or$cond196) {
          $$6181 = $0;
          break L1;
         }
         $255 = HEAP32[$$pre$phi283Z2D>>2]|0;
         $256 = HEAP32[$2>>2]|0;
         $257 = (($255) - ($256))|0;
         $258 = (($257|0) / 24)&-1;
         $259 = ($258>>>0)<(2);
         if ($259) {
          $$6181 = $0;
          break L1;
         }
         $260 = $255;
         $261 = ((($260)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($11,$261);
         $262 = HEAP32[$$pre$phi283Z2D>>2]|0;
         $263 = ((($262)) + -24|0);
         $265 = $262;
         while(1) {
          $264 = ($265|0)==($263|0);
          if ($264) {
           break;
          }
          $266 = ((($265)) + -24|0);
          HEAP32[$$pre$phi283Z2D>>2] = $266;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($266);
          $$pre$i227 = HEAP32[$$pre$phi283Z2D>>2]|0;
          $265 = $$pre$i227;
         }
         __THREW__ = 0;
         $267 = (invoke_iiii(1121,($11|0),0,(159155|0))|0);
         $268 = __THREW__; __THREW__ = 0;
         $269 = $268&1;
         if ($269) {
          label = 106;
          break;
         }
         ;HEAP32[$12>>2]=HEAP32[$11>>2]|0;HEAP32[$12+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$12+8>>2]=HEAP32[$11+8>>2]|0;
         $$0$i$i$i229 = 0;
         while(1) {
          $exitcond$i$i$i230 = ($$0$i$i$i229|0)==(3);
          if ($exitcond$i$i$i230) {
           break;
          }
          $270 = (($11) + ($$0$i$i$i229<<2)|0);
          HEAP32[$270>>2] = 0;
          $271 = (($$0$i$i$i229) + 1)|0;
          $$0$i$i$i229 = $271;
         }
         $272 = HEAP32[$$pre$phi283Z2D>>2]|0;
         $273 = ((($272)) + -24|0);
         $274 = HEAP8[$248>>0]|0;
         $275 = ($274<<24>>24)<(0);
         $276 = HEAP32[$12>>2]|0;
         $277 = $275 ? $276 : $12;
         $278 = HEAP32[$249>>2]|0;
         $279 = $274&255;
         $280 = $275 ? $278 : $279;
         __THREW__ = 0;
         (invoke_iiii(1112,($273|0),($277|0),($280|0))|0);
         $281 = __THREW__; __THREW__ = 0;
         $282 = $281&1;
         if ($282) {
          label = 107;
          break;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
         $$3186 = $252;
        }
        if ((label|0) == 106) {
         $283 = ___cxa_find_matching_catch_2()|0;
         $284 = tempRet0;
         $$4 = $284;$$4167 = $283;
        }
        else if ((label|0) == 107) {
         $285 = ___cxa_find_matching_catch_2()|0;
         $286 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
         $$4 = $286;$$4167 = $285;
        }
        else if ((label|0) == 109) {
         $287 = ((($$3186)) + 1|0);
         $288 = (__ZN10__cxxabiv112_GLOBAL__N_126parse_base_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($287,$1,$2)|0);
         $289 = ($288|0)==($287|0);
         if ($289) {
          $290 = HEAP32[$2>>2]|0;
          $291 = HEAP32[$$pre$phi283Z2D>>2]|0;
          $292 = ($290|0)==($291|0);
          if ($292) {
           $$6181 = $0;
           break L1;
          }
          $293 = ((($291)) + -24|0);
          $295 = $291;
          while(1) {
           $294 = ($295|0)==($293|0);
           if ($294) {
            $$6181 = $0;
            break L1;
           }
           $296 = ((($295)) + -24|0);
           HEAP32[$$pre$phi283Z2D>>2] = $296;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($296);
           $$pre$i235 = HEAP32[$$pre$phi283Z2D>>2]|0;
           $295 = $$pre$i235;
          }
         }
         $297 = HEAP32[$$pre$phi283Z2D>>2]|0;
         $298 = HEAP32[$2>>2]|0;
         $299 = (($297) - ($298))|0;
         $300 = (($299|0) / 24)&-1;
         $301 = ($300>>>0)<(2);
         if ($301) {
          $$6181 = $0;
          break L1;
         }
         $302 = $297;
         $303 = ((($302)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($13,$303);
         $304 = HEAP32[$$pre$phi283Z2D>>2]|0;
         $305 = ((($304)) + -24|0);
         $307 = $304;
         while(1) {
          $306 = ($307|0)==($305|0);
          if ($306) {
           break;
          }
          $308 = ((($307)) + -24|0);
          HEAP32[$$pre$phi283Z2D>>2] = $308;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($308);
          $$pre$i237 = HEAP32[$$pre$phi283Z2D>>2]|0;
          $307 = $$pre$i237;
         }
         __THREW__ = 0;
         $309 = (invoke_iiii(1121,($13|0),0,(159155|0))|0);
         $310 = __THREW__; __THREW__ = 0;
         $311 = $310&1;
         do {
          if ($311) {
           $327 = ___cxa_find_matching_catch_2()|0;
           $328 = tempRet0;
           $$5 = $328;$$5168 = $327;
          } else {
           ;HEAP32[$14>>2]=HEAP32[$13>>2]|0;HEAP32[$14+4>>2]=HEAP32[$13+4>>2]|0;HEAP32[$14+8>>2]=HEAP32[$13+8>>2]|0;
           $$0$i$i$i239 = 0;
           while(1) {
            $exitcond$i$i$i240 = ($$0$i$i$i239|0)==(3);
            if ($exitcond$i$i$i240) {
             break;
            }
            $312 = (($13) + ($$0$i$i$i239<<2)|0);
            HEAP32[$312>>2] = 0;
            $313 = (($$0$i$i$i239) + 1)|0;
            $$0$i$i$i239 = $313;
           }
           $314 = HEAP32[$$pre$phi283Z2D>>2]|0;
           $315 = ((($314)) + -24|0);
           $316 = ((($14)) + 11|0);
           $317 = HEAP8[$316>>0]|0;
           $318 = ($317<<24>>24)<(0);
           $319 = HEAP32[$14>>2]|0;
           $320 = $318 ? $319 : $14;
           $321 = ((($14)) + 4|0);
           $322 = HEAP32[$321>>2]|0;
           $323 = $317&255;
           $324 = $318 ? $322 : $323;
           __THREW__ = 0;
           (invoke_iiii(1112,($315|0),($320|0),($324|0))|0);
           $325 = __THREW__; __THREW__ = 0;
           $326 = $325&1;
           if ($326) {
            $329 = ___cxa_find_matching_catch_2()|0;
            $330 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
            $$5 = $330;$$5168 = $329;
            break;
           } else {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
            $$6181 = $288;
            break L1;
           }
          }
         } while(0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
         $$7 = $$5;$$7170 = $$5168;
         break;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
        $$7 = $$4;$$7170 = $$4167;
       }
      } while(0);
      ___resumeException($$7170|0);
      // unreachable;
     } else {
      $$6181 = $0;
     }
    } else {
     $$6181 = $0;
    }
   } else {
    $$6181 = $0;
   }
  } else {
   $$6181 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$6181|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_dot_star_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$030 = 0, $$3 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 12|0;
 $4 = sp;
 $5 = $1;
 $6 = $0;
 $7 = (($5) - ($6))|0;
 $8 = ($7|0)>(2);
 L1: do {
  if ($8) {
   $9 = HEAP8[$0>>0]|0;
   $10 = ($9<<24>>24)==(100);
   if ($10) {
    $11 = ((($0)) + 1|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = ($12<<24>>24)==(115);
    if ($13) {
     $14 = ((($0)) + 2|0);
     $15 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
     $16 = ($15|0)==($14|0);
     if ($16) {
      $$3 = $0;
     } else {
      $17 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($15,$1,$2)|0);
      $18 = ($17|0)==($15|0);
      if ($18) {
       $$3 = $0;
      } else {
       $19 = ((($2)) + 4|0);
       $20 = HEAP32[$19>>2]|0;
       $21 = HEAP32[$2>>2]|0;
       $22 = (($20) - ($21))|0;
       $23 = (($22|0) / 24)&-1;
       $24 = ($23>>>0)<(2);
       if ($24) {
        $$3 = $0;
       } else {
        $25 = $20;
        $26 = ((($25)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$26);
        $27 = HEAP32[$19>>2]|0;
        $28 = ((($27)) + -24|0);
        $30 = $27;
        while(1) {
         $29 = ($30|0)==($28|0);
         if ($29) {
          break;
         }
         $31 = ((($30)) + -24|0);
         HEAP32[$19>>2] = $31;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
         $$pre$i = HEAP32[$19>>2]|0;
         $30 = $$pre$i;
        }
        __THREW__ = 0;
        invoke_viii(1139,($4|0),(159374|0),($3|0));
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        do {
         if ($33) {
          $47 = ___cxa_find_matching_catch_2()|0;
          $48 = tempRet0;
          $$0 = $48;$$030 = $47;
         } else {
          $34 = HEAP32[$19>>2]|0;
          $35 = ((($34)) + -24|0);
          $36 = ((($4)) + 11|0);
          $37 = HEAP8[$36>>0]|0;
          $38 = ($37<<24>>24)<(0);
          $39 = HEAP32[$4>>2]|0;
          $40 = $38 ? $39 : $4;
          $41 = ((($4)) + 4|0);
          $42 = HEAP32[$41>>2]|0;
          $43 = $37&255;
          $44 = $38 ? $42 : $43;
          __THREW__ = 0;
          (invoke_iiii(1112,($35|0),($40|0),($44|0))|0);
          $45 = __THREW__; __THREW__ = 0;
          $46 = $45&1;
          if ($46) {
           $49 = ___cxa_find_matching_catch_2()|0;
           $50 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
           $$0 = $50;$$030 = $49;
           break;
          } else {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
           $$3 = $17;
           break L1;
          }
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$030|0);
        // unreachable;
       }
      }
     }
    } else {
     $$3 = $0;
    }
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_114parse_dot_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$030 = 0, $$3 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 12|0;
 $4 = sp;
 $5 = $1;
 $6 = $0;
 $7 = (($5) - ($6))|0;
 $8 = ($7|0)>(2);
 L1: do {
  if ($8) {
   $9 = HEAP8[$0>>0]|0;
   $10 = ($9<<24>>24)==(100);
   if ($10) {
    $11 = ((($0)) + 1|0);
    $12 = HEAP8[$11>>0]|0;
    $13 = ($12<<24>>24)==(116);
    if ($13) {
     $14 = ((($0)) + 2|0);
     $15 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
     $16 = ($15|0)==($14|0);
     if ($16) {
      $$3 = $0;
     } else {
      $17 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($15,$1,$2)|0);
      $18 = ($17|0)==($15|0);
      if ($18) {
       $$3 = $0;
      } else {
       $19 = ((($2)) + 4|0);
       $20 = HEAP32[$19>>2]|0;
       $21 = HEAP32[$2>>2]|0;
       $22 = (($20) - ($21))|0;
       $23 = (($22|0) / 24)&-1;
       $24 = ($23>>>0)<(2);
       if ($24) {
        $$3 = $0;
       } else {
        $25 = $20;
        $26 = ((($25)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$26);
        $27 = HEAP32[$19>>2]|0;
        $28 = ((($27)) + -24|0);
        $30 = $27;
        while(1) {
         $29 = ($30|0)==($28|0);
         if ($29) {
          break;
         }
         $31 = ((($30)) + -24|0);
         HEAP32[$19>>2] = $31;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
         $$pre$i = HEAP32[$19>>2]|0;
         $30 = $$pre$i;
        }
        __THREW__ = 0;
        invoke_viii(1139,($4|0),(159372|0),($3|0));
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        do {
         if ($33) {
          $47 = ___cxa_find_matching_catch_2()|0;
          $48 = tempRet0;
          $$0 = $48;$$030 = $47;
         } else {
          $34 = HEAP32[$19>>2]|0;
          $35 = ((($34)) + -24|0);
          $36 = ((($4)) + 11|0);
          $37 = HEAP8[$36>>0]|0;
          $38 = ($37<<24>>24)<(0);
          $39 = HEAP32[$4>>2]|0;
          $40 = $38 ? $39 : $4;
          $41 = ((($4)) + 4|0);
          $42 = HEAP32[$41>>2]|0;
          $43 = $37&255;
          $44 = $38 ? $42 : $43;
          __THREW__ = 0;
          (invoke_iiii(1112,($35|0),($40|0),($44|0))|0);
          $45 = __THREW__; __THREW__ = 0;
          $46 = $45&1;
          if ($46) {
           $49 = ___cxa_find_matching_catch_2()|0;
           $50 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
           $$0 = $50;$$030 = $49;
           break;
          } else {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
           $$3 = $17;
           break L1;
          }
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$030|0);
        // unreachable;
       }
      }
     }
    } else {
     $$3 = $0;
    }
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_114parse_new_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i$i175 = 0, $$0$i$i$i187 = 0, $$0$i$i170 = 0, $$0$i$i180 = 0, $$0$i$i193 = 0, $$0$off0216 = 0, $$0108 = 0, $$0109 = 0, $$0136$off0 = 0, $$0136$off0$ = 0, $$0148 = 0, $$0157$off0 = 0, $$0158 = 0, $$1 = 0, $$1110 = 0, $$1149 = 0;
 var $$1159 = 0, $$16 = 0, $$2 = 0, $$2111 = 0, $$3 = 0, $$3112 = 0, $$3141 = 0, $$3161 = 0, $$4 = 0, $$4113 = 0, $$4142 = 0, $$4162215 = 0, $$5 = 0, $$5114 = 0, $$6 = 0, $$6115 = 0, $$phi$trans$insert = 0, $$pre = 0, $$pre$i = 0, $$pre$i168 = 0;
 var $$pre$i178 = 0, $$pre$phi214Z2D = 0, $$pre$phiZ2D = 0, $$pre212 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0;
 var $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0;
 var $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0;
 var $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0;
 var $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0;
 var $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0;
 var $205 = 0, $206 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $cond = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i$i176 = 0, $exitcond$i$i$i188 = 0, $exitcond$i$i171 = 0, $exitcond$i$i181 = 0, $exitcond$i$i194 = 0, $not$or$cond = 0, $not$or$cond$demorgan = 0, $or$cond = 0;
 var $or$cond165 = 0, $or$cond166 = 0, $or$cond222 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 176|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(176|0);
 $3 = sp + 163|0;
 $4 = sp + 162|0;
 $5 = sp + 161|0;
 $6 = sp + 160|0;
 $7 = sp + 148|0;
 $8 = sp + 136|0;
 $9 = sp + 124|0;
 $10 = sp + 112|0;
 $11 = sp + 100|0;
 $12 = sp + 88|0;
 $13 = sp + 76|0;
 $14 = sp + 64|0;
 $15 = sp + 52|0;
 $16 = sp + 40|0;
 $17 = sp + 16|0;
 $18 = sp;
 $19 = $1;
 $20 = $0;
 $21 = (($19) - ($20))|0;
 $22 = ($21|0)>(3);
 L1: do {
  if ($22) {
   $23 = HEAP8[$0>>0]|0;
   $24 = ($23<<24>>24)==(103);
   if ($24) {
    $25 = ((($0)) + 1|0);
    $26 = HEAP8[$25>>0]|0;
    $27 = ($26<<24>>24)==(115);
    $28 = ((($0)) + 2|0);
    $$ = $27 ? $28 : $0;
    $$pre = HEAP8[$$>>0]|0;
    $$0157$off0 = $27;$$0158 = $$;$30 = $$pre;
   } else {
    $$0157$off0 = 0;$$0158 = $0;$30 = $23;
   }
   $29 = ($30<<24>>24)==(110);
   if ($29) {
    $31 = ((($$0158)) + 1|0);
    $32 = HEAP8[$31>>0]|0;
    $33 = ($32<<24>>24)==(97);
    switch ($32<<24>>24) {
    case 97: case 119:  {
     break;
    }
    default: {
     $$16 = $0;
     break L1;
    }
    }
    $34 = ((($$0158)) + 2|0);
    $35 = ($34|0)==($1|0);
    L8: do {
     if (!($35)) {
      $$0136$off0 = 0;$$1159 = $34;
      while(1) {
       $36 = HEAP8[$$1159>>0]|0;
       $37 = ($36<<24>>24)==(95);
       if ($37) {
        break;
       }
       $38 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($$1159,$1,$2)|0);
       $39 = ($38|0)==($$1159|0);
       $40 = ($38|0)==($1|0);
       $or$cond = $39 | $40;
       $not$or$cond$demorgan = $39 | $40;
       $not$or$cond = $not$or$cond$demorgan ^ 1;
       $$0136$off0$ = $$0136$off0 | $not$or$cond;
       if ($or$cond) {
        break L8;
       } else {
        $$0136$off0 = $$0136$off0$;$$1159 = $38;
       }
      }
      $41 = ((($$1159)) + 1|0);
      $42 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($41,$1,$2)|0);
      $43 = ($42|0)==($41|0);
      $44 = ($42|0)==($1|0);
      $or$cond165 = $43 | $44;
      if (!($or$cond165)) {
       $45 = $42;
       $46 = (($19) - ($45))|0;
       $47 = ($46|0)>(2);
       $48 = HEAP8[$42>>0]|0;
       $49 = ($48<<24>>24)==(112);
       $or$cond222 = $47 & $49;
       L14: do {
        if ($or$cond222) {
         $50 = ((($42)) + 1|0);
         $51 = HEAP8[$50>>0]|0;
         $52 = ($51<<24>>24)==(105);
         if (!($52)) {
          break L8;
         }
         $53 = ((($42)) + 2|0);
         $$3161 = $53;
         while(1) {
          $54 = HEAP8[$$3161>>0]|0;
          $55 = ($54<<24>>24)==(69);
          if ($55) {
           $$0$off0216 = 1;$$4162215 = $$3161;
           break L14;
          }
          $56 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($$3161,$1,$2)|0);
          $57 = ($56|0)==($$3161|0);
          $58 = ($56|0)==($1|0);
          $or$cond166 = $57 | $58;
          if ($or$cond166) {
           break L8;
          } else {
           $$3161 = $56;
          }
         }
        } else {
         $59 = ($48<<24>>24)==(69);
         if ($59) {
          $$0$off0216 = 0;$$4162215 = $42;
         } else {
          break L8;
         }
        }
       } while(0);
       ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
       $$0$i$i = 0;
       while(1) {
        $exitcond$i$i = ($$0$i$i|0)==(3);
        if ($exitcond$i$i) {
         break;
        }
        $60 = (($7) + ($$0$i$i<<2)|0);
        HEAP32[$60>>2] = 0;
        $61 = (($$0$i$i) + 1)|0;
        $$0$i$i = $61;
       }
       L26: do {
        if ($$0$off0216) {
         $62 = HEAP32[$2>>2]|0;
         $63 = ((($2)) + 4|0);
         $64 = HEAP32[$63>>2]|0;
         $65 = ($62|0)==($64|0);
         if ($65) {
          $$1149 = $0;$$4142 = 1;
          label = 103;
         } else {
          $66 = ((($64)) + -24|0);
          __THREW__ = 0;
          invoke_vii(1120,($8|0),($66|0));
          $67 = __THREW__; __THREW__ = 0;
          $68 = $67&1;
          if ($68) {
           $85 = ___cxa_find_matching_catch_2()|0;
           $86 = tempRet0;
           $$6 = $86;$$6115 = $85;
           label = 107;
           break;
          }
          $69 = ((($7)) + 11|0);
          $70 = HEAP8[$69>>0]|0;
          $71 = ($70<<24>>24)<(0);
          if ($71) {
           $72 = HEAP32[$7>>2]|0;
           HEAP8[$5>>0] = 0;
           __ZNSt3__211char_traitsIcE6assignERcRKc($72,$5);
           $73 = ((($7)) + 4|0);
           HEAP32[$73>>2] = 0;
          } else {
           HEAP8[$6>>0] = 0;
           __ZNSt3__211char_traitsIcE6assignERcRKc($7,$6);
           HEAP8[$69>>0] = 0;
          }
          __THREW__ = 0;
          invoke_vi(1116,($7|0));
          $74 = __THREW__; __THREW__ = 0;
          $75 = $74&1;
          if ($75) {
           $76 = ___cxa_find_matching_catch_3(0|0)|0;
           $77 = tempRet0;
           ___clang_call_terminate($76);
           // unreachable;
          }
          ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
          $$0$i$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
           if ($exitcond$i$i$i$i) {
            break;
           }
           $78 = (($8) + ($$0$i$i$i$i<<2)|0);
           HEAP32[$78>>2] = 0;
           $79 = (($$0$i$i$i$i) + 1)|0;
           $$0$i$i$i$i = $79;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
          $80 = HEAP32[$63>>2]|0;
          $81 = ((($80)) + -24|0);
          $83 = $80;
          while(1) {
           $82 = ($83|0)==($81|0);
           if ($82) {
            $$pre$phi214Z2D = $63;$$pre$phiZ2D = $2;$91 = $81;
            label = 36;
            break L26;
           }
           $84 = ((($83)) + -24|0);
           HEAP32[$63>>2] = $84;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($84);
           $$pre$i = HEAP32[$63>>2]|0;
           $83 = $$pre$i;
          }
         }
        } else {
         $$phi$trans$insert = ((($2)) + 4|0);
         $$pre212 = HEAP32[$$phi$trans$insert>>2]|0;
         $$pre$phi214Z2D = $$phi$trans$insert;$$pre$phiZ2D = $2;$91 = $$pre212;
         label = 36;
        }
       } while(0);
       do {
        if ((label|0) == 36) {
         $89 = HEAP32[$$pre$phiZ2D>>2]|0;
         $90 = ($89|0)==($91|0);
         if ($90) {
          $$1149 = $0;$$4142 = 1;
          label = 103;
         } else {
          $92 = ((($91)) + -24|0);
          __THREW__ = 0;
          invoke_vii(1120,($9|0),($92|0));
          $93 = __THREW__; __THREW__ = 0;
          $94 = $93&1;
          if ($94) {
           $87 = ___cxa_find_matching_catch_2()|0;
           $88 = tempRet0;
           $$6 = $88;$$6115 = $87;
           label = 107;
           break;
          }
          $95 = HEAP32[$$pre$phi214Z2D>>2]|0;
          $96 = ((($95)) + -24|0);
          $98 = $95;
          while(1) {
           $97 = ($98|0)==($96|0);
           if ($97) {
            break;
           }
           $99 = ((($98)) + -24|0);
           HEAP32[$$pre$phi214Z2D>>2] = $99;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($99);
           $$pre$i168 = HEAP32[$$pre$phi214Z2D>>2]|0;
           $98 = $$pre$i168;
          }
          ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
          $$0$i$i170 = 0;
          while(1) {
           $exitcond$i$i171 = ($$0$i$i170|0)==(3);
           if ($exitcond$i$i171) {
            break;
           }
           $100 = (($10) + ($$0$i$i170<<2)|0);
           HEAP32[$100>>2] = 0;
           $101 = (($$0$i$i170) + 1)|0;
           $$0$i$i170 = $101;
          }
          L61: do {
           if ($$0136$off0) {
            $102 = HEAP32[$$pre$phiZ2D>>2]|0;
            $103 = ($102|0)==($96|0);
            if ($103) {
             $$0148 = $0;$$3141 = 1;
             label = 102;
            } else {
             $104 = ((($95)) + -48|0);
             __THREW__ = 0;
             invoke_vii(1120,($11|0),($104|0));
             $105 = __THREW__; __THREW__ = 0;
             $106 = $105&1;
             if ($106) {
              $123 = ___cxa_find_matching_catch_2()|0;
              $124 = tempRet0;
              $$5 = $124;$$5114 = $123;
              label = 106;
              break;
             }
             $107 = ((($10)) + 11|0);
             $108 = HEAP8[$107>>0]|0;
             $109 = ($108<<24>>24)<(0);
             if ($109) {
              $110 = HEAP32[$10>>2]|0;
              HEAP8[$3>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($110,$3);
              $111 = ((($10)) + 4|0);
              HEAP32[$111>>2] = 0;
             } else {
              HEAP8[$4>>0] = 0;
              __ZNSt3__211char_traitsIcE6assignERcRKc($10,$4);
              HEAP8[$107>>0] = 0;
             }
             __THREW__ = 0;
             invoke_vi(1116,($10|0));
             $112 = __THREW__; __THREW__ = 0;
             $113 = $112&1;
             if ($113) {
              $114 = ___cxa_find_matching_catch_3(0|0)|0;
              $115 = tempRet0;
              ___clang_call_terminate($114);
              // unreachable;
             }
             ;HEAP32[$10>>2]=HEAP32[$11>>2]|0;HEAP32[$10+4>>2]=HEAP32[$11+4>>2]|0;HEAP32[$10+8>>2]=HEAP32[$11+8>>2]|0;
             $$0$i$i$i$i175 = 0;
             while(1) {
              $exitcond$i$i$i$i176 = ($$0$i$i$i$i175|0)==(3);
              if ($exitcond$i$i$i$i176) {
               break;
              }
              $116 = (($11) + ($$0$i$i$i$i175<<2)|0);
              HEAP32[$116>>2] = 0;
              $117 = (($$0$i$i$i$i175) + 1)|0;
              $$0$i$i$i$i175 = $117;
             }
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
             $118 = HEAP32[$$pre$phi214Z2D>>2]|0;
             $119 = ((($118)) + -24|0);
             $121 = $118;
             while(1) {
              $120 = ($121|0)==($119|0);
              if ($120) {
               label = 59;
               break L61;
              }
              $122 = ((($121)) + -24|0);
              HEAP32[$$pre$phi214Z2D>>2] = $122;
              __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($122);
              $$pre$i178 = HEAP32[$$pre$phi214Z2D>>2]|0;
              $121 = $$pre$i178;
             }
            }
           } else {
            label = 59;
           }
          } while(0);
          L81: do {
           if ((label|0) == 59) {
            ;HEAP32[$12>>2]=0|0;HEAP32[$12+4>>2]=0|0;HEAP32[$12+8>>2]=0|0;
            $$0$i$i180 = 0;
            while(1) {
             $exitcond$i$i181 = ($$0$i$i180|0)==(3);
             if ($exitcond$i$i181) {
              break;
             }
             $125 = (($12) + ($$0$i$i180<<2)|0);
             HEAP32[$125>>2] = 0;
             $126 = (($$0$i$i180) + 1)|0;
             $$0$i$i180 = $126;
            }
            if ($$0157$off0) {
             __THREW__ = 0;
             invoke_vii(1143,($12|0),(159155|0));
             $127 = __THREW__; __THREW__ = 0;
             $128 = $127&1;
             if ($128) {
              label = 64;
             } else {
              label = 65;
             }
            } else {
             label = 65;
            }
            L89: do {
             if ((label|0) == 65) {
              if ($33) {
               __THREW__ = 0;
               (invoke_iii(1119,($12|0),(159365|0))|0);
               $131 = __THREW__; __THREW__ = 0;
               $132 = $131&1;
               if ($132) {
                label = 64;
                break;
               }
              } else {
               __THREW__ = 0;
               (invoke_iii(1119,($12|0),(159115|0))|0);
               $133 = __THREW__; __THREW__ = 0;
               $134 = $133&1;
               if ($134) {
                label = 64;
                break;
               }
              }
              L95: do {
               if ($$0136$off0) {
                __THREW__ = 0;
                invoke_viii(1139,($14|0),(159083|0),($10|0));
                $135 = __THREW__; __THREW__ = 0;
                $136 = $135&1;
                if ($136) {
                 $153 = ___cxa_find_matching_catch_2()|0;
                 $154 = tempRet0;
                 $$1 = $154;$$1110 = $153;
                } else {
                 __THREW__ = 0;
                 $137 = (invoke_iii(1119,($14|0),(159369|0))|0);
                 $138 = __THREW__; __THREW__ = 0;
                 $139 = $138&1;
                 do {
                  if ($139) {
                   $155 = ___cxa_find_matching_catch_2()|0;
                   $156 = tempRet0;
                   $$0108 = $156;$$0109 = $155;
                  } else {
                   ;HEAP32[$13>>2]=HEAP32[$14>>2]|0;HEAP32[$13+4>>2]=HEAP32[$14+4>>2]|0;HEAP32[$13+8>>2]=HEAP32[$14+8>>2]|0;
                   $$0$i$i$i = 0;
                   while(1) {
                    $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
                    if ($exitcond$i$i$i) {
                     break;
                    }
                    $140 = (($14) + ($$0$i$i$i<<2)|0);
                    HEAP32[$140>>2] = 0;
                    $141 = (($$0$i$i$i) + 1)|0;
                    $$0$i$i$i = $141;
                   }
                   $142 = ((($13)) + 11|0);
                   $143 = HEAP8[$142>>0]|0;
                   $144 = ($143<<24>>24)<(0);
                   $145 = HEAP32[$13>>2]|0;
                   $146 = $144 ? $145 : $13;
                   $147 = ((($13)) + 4|0);
                   $148 = HEAP32[$147>>2]|0;
                   $149 = $143&255;
                   $150 = $144 ? $148 : $149;
                   __THREW__ = 0;
                   (invoke_iiii(1112,($12|0),($146|0),($150|0))|0);
                   $151 = __THREW__; __THREW__ = 0;
                   $152 = $151&1;
                   if ($152) {
                    $157 = ___cxa_find_matching_catch_2()|0;
                    $158 = tempRet0;
                    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
                    $$0108 = $158;$$0109 = $157;
                    break;
                   } else {
                    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
                    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
                    break L95;
                   }
                  }
                 } while(0);
                 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($14);
                 $$1 = $$0108;$$1110 = $$0109;
                }
                $$4 = $$1;$$4113 = $$1110;
                break L89;
               }
              } while(0);
              $159 = ((($9)) + 11|0);
              $160 = HEAP8[$159>>0]|0;
              $161 = ($160<<24>>24)<(0);
              $162 = HEAP32[$9>>2]|0;
              $163 = $161 ? $162 : $9;
              $164 = ((($9)) + 4|0);
              $165 = HEAP32[$164>>2]|0;
              $166 = $160&255;
              $167 = $161 ? $165 : $166;
              __THREW__ = 0;
              (invoke_iiii(1112,($12|0),($163|0),($167|0))|0);
              $168 = __THREW__; __THREW__ = 0;
              $169 = $168&1;
              if ($169) {
               label = 64;
              } else {
               L114: do {
                if ($$0$off0216) {
                 __THREW__ = 0;
                 invoke_viii(1139,($16|0),(159078|0),($7|0));
                 $170 = __THREW__; __THREW__ = 0;
                 $171 = $170&1;
                 if ($171) {
                  $188 = ___cxa_find_matching_catch_2()|0;
                  $189 = tempRet0;
                  $$3 = $189;$$3112 = $188;
                 } else {
                  __THREW__ = 0;
                  $172 = (invoke_iii(1119,($16|0),(159081|0))|0);
                  $173 = __THREW__; __THREW__ = 0;
                  $174 = $173&1;
                  do {
                   if ($174) {
                    $190 = ___cxa_find_matching_catch_2()|0;
                    $191 = tempRet0;
                    $$2 = $191;$$2111 = $190;
                   } else {
                    ;HEAP32[$15>>2]=HEAP32[$16>>2]|0;HEAP32[$15+4>>2]=HEAP32[$16+4>>2]|0;HEAP32[$15+8>>2]=HEAP32[$16+8>>2]|0;
                    $$0$i$i$i187 = 0;
                    while(1) {
                     $exitcond$i$i$i188 = ($$0$i$i$i187|0)==(3);
                     if ($exitcond$i$i$i188) {
                      break;
                     }
                     $175 = (($16) + ($$0$i$i$i187<<2)|0);
                     HEAP32[$175>>2] = 0;
                     $176 = (($$0$i$i$i187) + 1)|0;
                     $$0$i$i$i187 = $176;
                    }
                    $177 = ((($15)) + 11|0);
                    $178 = HEAP8[$177>>0]|0;
                    $179 = ($178<<24>>24)<(0);
                    $180 = HEAP32[$15>>2]|0;
                    $181 = $179 ? $180 : $15;
                    $182 = ((($15)) + 4|0);
                    $183 = HEAP32[$182>>2]|0;
                    $184 = $178&255;
                    $185 = $179 ? $183 : $184;
                    __THREW__ = 0;
                    (invoke_iiii(1112,($12|0),($181|0),($185|0))|0);
                    $186 = __THREW__; __THREW__ = 0;
                    $187 = $186&1;
                    if ($187) {
                     $192 = ___cxa_find_matching_catch_2()|0;
                     $193 = tempRet0;
                     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
                     $$2 = $193;$$2111 = $192;
                     break;
                    } else {
                     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($15);
                     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
                     break L114;
                    }
                   }
                  } while(0);
                  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
                  $$3 = $$2;$$3112 = $$2111;
                 }
                 $$4 = $$3;$$4113 = $$3112;
                 break L89;
                }
               } while(0);
               ;HEAP32[$18>>2]=HEAP32[$12>>2]|0;HEAP32[$18+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$18+8>>2]=HEAP32[$12+8>>2]|0;
               $$0$i$i193 = 0;
               while(1) {
                $exitcond$i$i194 = ($$0$i$i193|0)==(3);
                if ($exitcond$i$i194) {
                 break;
                }
                $194 = (($12) + ($$0$i$i193<<2)|0);
                HEAP32[$194>>2] = 0;
                $195 = (($$0$i$i193) + 1)|0;
                $$0$i$i193 = $195;
               }
               __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($17,$18);
               $196 = HEAP32[$$pre$phi214Z2D>>2]|0;
               $197 = ((($2)) + 8|0);
               $198 = HEAP32[$197>>2]|0;
               $199 = ($196>>>0)<($198>>>0);
               do {
                if ($199) {
                 __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($196,$17);
                 $200 = HEAP32[$$pre$phi214Z2D>>2]|0;
                 $201 = ((($200)) + 24|0);
                 HEAP32[$$pre$phi214Z2D>>2] = $201;
                } else {
                 __THREW__ = 0;
                 invoke_vii(1128,($2|0),($17|0));
                 $202 = __THREW__; __THREW__ = 0;
                 $203 = $202&1;
                 if (!($203)) {
                  break;
                 }
                 $205 = ___cxa_find_matching_catch_2()|0;
                 $206 = tempRet0;
                 __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
                 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
                 $$4 = $206;$$4113 = $205;
                 break L89;
                }
               } while(0);
               __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
               __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($18);
               $204 = ((($$4162215)) + 1|0);
               __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
               $$0148 = $204;$$3141 = 0;
               label = 102;
               break L81;
              }
             }
            } while(0);
            if ((label|0) == 64) {
             $129 = ___cxa_find_matching_catch_2()|0;
             $130 = tempRet0;
             $$4 = $130;$$4113 = $129;
            }
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
            $$5 = $$4;$$5114 = $$4113;
            label = 106;
           }
          } while(0);
          if ((label|0) == 102) {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
           $$1149 = $$0148;$$4142 = $$3141;
           label = 103;
           break;
          }
          else if ((label|0) == 106) {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
           $$6 = $$5;$$6115 = $$5114;
           label = 107;
           break;
          }
         }
        }
       } while(0);
       if ((label|0) == 103) {
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        $cond = ($$4142|0)==(0);
        if ($cond) {
         $$16 = $$1149;
         break L1;
        } else {
         break;
        }
       }
       else if ((label|0) == 107) {
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        ___resumeException($$6115|0);
        // unreachable;
       }
      }
     }
    } while(0);
    $$16 = $0;
   } else {
    $$16 = $0;
   }
  } else {
   $$16 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$16|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_125parse_noexcept_expressionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i18 = 0, $$015 = 0, $$016 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i19 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 37|0;
 $4 = sp + 36|0;
 $5 = sp + 24|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
 $9 = ($8|0)==($0|0);
 L1: do {
  if ($9) {
   $$016 = $0;
  } else {
   $10 = HEAP32[$2>>2]|0;
   $11 = ((($2)) + 4|0);
   $12 = HEAP32[$11>>2]|0;
   $13 = ($10|0)==($12|0);
   if ($13) {
    $$016 = $0;
   } else {
    $14 = ((($12)) + -24|0);
    __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($7,$14);
    __THREW__ = 0;
    $15 = (invoke_iiii(1121,($7|0),0,(159354|0))|0);
    $16 = __THREW__; __THREW__ = 0;
    $17 = $16&1;
    do {
     if ($17) {
      $38 = ___cxa_find_matching_catch_2()|0;
      $39 = tempRet0;
      $$0 = $39;$$015 = $38;
     } else {
      ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
      $$0$i$i$i = 0;
      while(1) {
       $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
       if ($exitcond$i$i$i) {
        break;
       }
       $18 = (($7) + ($$0$i$i$i<<2)|0);
       HEAP32[$18>>2] = 0;
       $19 = (($$0$i$i$i) + 1)|0;
       $$0$i$i$i = $19;
      }
      __THREW__ = 0;
      $20 = (invoke_iii(1119,($6|0),(159081|0))|0);
      $21 = __THREW__; __THREW__ = 0;
      $22 = $21&1;
      if ($22) {
       $40 = ___cxa_find_matching_catch_2()|0;
       $41 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       $$0 = $41;$$015 = $40;
       break;
      }
      ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
      $$0$i$i$i18 = 0;
      while(1) {
       $exitcond$i$i$i19 = ($$0$i$i$i18|0)==(3);
       if ($exitcond$i$i$i19) {
        break;
       }
       $23 = (($6) + ($$0$i$i$i18<<2)|0);
       HEAP32[$23>>2] = 0;
       $24 = (($$0$i$i$i18) + 1)|0;
       $$0$i$i$i18 = $24;
      }
      $25 = HEAP32[$11>>2]|0;
      $26 = ((($25)) + -24|0);
      $27 = ((($26)) + 11|0);
      $28 = HEAP8[$27>>0]|0;
      $29 = ($28<<24>>24)<(0);
      if ($29) {
       $30 = HEAP32[$26>>2]|0;
       HEAP8[$3>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($30,$3);
       $31 = ((($25)) + -20|0);
       HEAP32[$31>>2] = 0;
      } else {
       HEAP8[$4>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($26,$4);
       HEAP8[$27>>0] = 0;
      }
      __THREW__ = 0;
      invoke_vi(1116,($26|0));
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       $34 = ___cxa_find_matching_catch_3(0|0)|0;
       $35 = tempRet0;
       ___clang_call_terminate($34);
       // unreachable;
      }
      ;HEAP32[$26>>2]=HEAP32[$5>>2]|0;HEAP32[$26+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$26+8>>2]=HEAP32[$5+8>>2]|0;
      $$0$i$i$i$i = 0;
      while(1) {
       $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
       if ($exitcond$i$i$i$i) {
        break;
       }
       $36 = (($5) + ($$0$i$i$i$i<<2)|0);
       HEAP32[$36>>2] = 0;
       $37 = (($$0$i$i$i$i) + 1)|0;
       $$0$i$i$i$i = $37;
      }
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$016 = $8;
      break L1;
     }
    } while(0);
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
    ___resumeException($$015|0);
    // unreachable;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$016|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_arrow_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$3 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0;
 var $46 = 0, $47 = 0, $48 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = $1;
 $5 = $0;
 $6 = (($4) - ($5))|0;
 $7 = ($6|0)>(2);
 do {
  if ($7) {
   $8 = HEAP8[$0>>0]|0;
   $9 = ($8<<24>>24)==(112);
   if ($9) {
    $10 = ((($0)) + 1|0);
    $11 = HEAP8[$10>>0]|0;
    $12 = ($11<<24>>24)==(116);
    if ($12) {
     $13 = ((($0)) + 2|0);
     $14 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($13,$1,$2)|0);
     $15 = ($14|0)==($13|0);
     if ($15) {
      $$3 = $0;
     } else {
      $16 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
      $17 = ($16|0)==($14|0);
      if ($17) {
       $$3 = $0;
      } else {
       $18 = ((($2)) + 4|0);
       $19 = HEAP32[$18>>2]|0;
       $20 = HEAP32[$2>>2]|0;
       $21 = (($19) - ($20))|0;
       $22 = (($21|0) / 24)&-1;
       $23 = ($22>>>0)<(2);
       if ($23) {
        $$3 = $0;
       } else {
        $24 = $19;
        $25 = ((($24)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$25);
        $26 = HEAP32[$18>>2]|0;
        $27 = ((($26)) + -24|0);
        $29 = $26;
        while(1) {
         $28 = ($29|0)==($27|0);
         if ($28) {
          break;
         }
         $30 = ((($29)) + -24|0);
         HEAP32[$18>>2] = $30;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
         $$pre$i = HEAP32[$18>>2]|0;
         $29 = $$pre$i;
        }
        $31 = ((($26)) + -48|0);
        __THREW__ = 0;
        (invoke_iii(1119,($31|0),(159351|0))|0);
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        if (!($33)) {
         $34 = HEAP32[$18>>2]|0;
         $35 = ((($34)) + -24|0);
         $36 = ((($3)) + 11|0);
         $37 = HEAP8[$36>>0]|0;
         $38 = ($37<<24>>24)<(0);
         $39 = HEAP32[$3>>2]|0;
         $40 = $38 ? $39 : $3;
         $41 = ((($3)) + 4|0);
         $42 = HEAP32[$41>>2]|0;
         $43 = $37&255;
         $44 = $38 ? $42 : $43;
         __THREW__ = 0;
         (invoke_iiii(1112,($35|0),($40|0),($44|0))|0);
         $45 = __THREW__; __THREW__ = 0;
         $46 = $45&1;
         if (!($46)) {
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
          $$3 = $16;
          break;
         }
        }
        $47 = ___cxa_find_matching_catch_2()|0;
        $48 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($47|0);
        // unreachable;
       }
      }
     }
    } else {
     $$3 = $0;
    }
   } else {
    $$3 = $0;
   }
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_127parse_reinterpret_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i47 = 0, $$0$i$i$i50 = 0, $$0$i$i$i53 = 0, $$1 = 0, $$132 = 0, $$2 = 0, $$233 = 0, $$3 = 0, $$334 = 0, $$341 = 0, $$4 = 0, $$435 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i48 = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i54 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 88|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(114);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(99);
    if ($18) {
     $19 = ((($0)) + 2|0);
     $20 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($19,$1,$2)|0);
     $21 = ($20|0)==($19|0);
     if ($21) {
      $$341 = $0;
     } else {
      $22 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($20,$1,$2)|0);
      $23 = ($22|0)==($20|0);
      if ($23) {
       $$341 = $0;
      } else {
       $24 = ((($2)) + 4|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = HEAP32[$2>>2]|0;
       $27 = (($25) - ($26))|0;
       $28 = (($27|0) / 24)&-1;
       $29 = ($28>>>0)<(2);
       if ($29) {
        $$341 = $0;
       } else {
        $30 = $25;
        $31 = ((($30)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$31);
        $32 = HEAP32[$24>>2]|0;
        $33 = ((($32)) + -24|0);
        $35 = $32;
        while(1) {
         $34 = ($35|0)==($33|0);
         if ($34) {
          break;
         }
         $36 = ((($35)) + -24|0);
         HEAP32[$24>>2] = $36;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
         $$pre$i = HEAP32[$24>>2]|0;
         $35 = $$pre$i;
        }
        $37 = ((($32)) + -48|0);
        __THREW__ = 0;
        invoke_vii(1120,($9|0),($37|0));
        $38 = __THREW__; __THREW__ = 0;
        $39 = $38&1;
        if ($39) {
         $71 = ___cxa_find_matching_catch_2()|0;
         $72 = tempRet0;
         $$4 = $72;$$435 = $71;
        } else {
         __THREW__ = 0;
         $40 = (invoke_iiii(1121,($9|0),0,(159333|0))|0);
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         if ($42) {
          $73 = ___cxa_find_matching_catch_2()|0;
          $74 = tempRet0;
          $$3 = $74;$$334 = $73;
         } else {
          ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $43 = (($9) + ($$0$i$i$i<<2)|0);
           HEAP32[$43>>2] = 0;
           $44 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $44;
          }
          __THREW__ = 0;
          $45 = (invoke_iii(1119,($8|0),(159330|0))|0);
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           $75 = ___cxa_find_matching_catch_2()|0;
           $76 = tempRet0;
           $$2 = $76;$$233 = $75;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i47 = 0;
           while(1) {
            $exitcond$i$i$i48 = ($$0$i$i$i47|0)==(3);
            if ($exitcond$i$i$i48) {
             break;
            }
            $48 = (($8) + ($$0$i$i$i47<<2)|0);
            HEAP32[$48>>2] = 0;
            $49 = (($$0$i$i$i47) + 1)|0;
            $$0$i$i$i47 = $49;
           }
           $50 = ((($3)) + 11|0);
           $51 = HEAP8[$50>>0]|0;
           $52 = ($51<<24>>24)<(0);
           $53 = HEAP32[$3>>2]|0;
           $54 = $52 ? $53 : $3;
           $55 = ((($3)) + 4|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $51&255;
           $58 = $52 ? $56 : $57;
           __THREW__ = 0;
           $59 = (invoke_iiii(1112,($7|0),($54|0),($58|0))|0);
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           do {
            if ($61) {
             $77 = ___cxa_find_matching_catch_2()|0;
             $78 = tempRet0;
             $$1 = $78;$$132 = $77;
            } else {
             ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
             $$0$i$i$i50 = 0;
             while(1) {
              $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
              if ($exitcond$i$i$i51) {
               break;
              }
              $62 = (($7) + ($$0$i$i$i50<<2)|0);
              HEAP32[$62>>2] = 0;
              $63 = (($$0$i$i$i50) + 1)|0;
              $$0$i$i$i50 = $63;
             }
             __THREW__ = 0;
             $64 = (invoke_iii(1119,($6|0),(159081|0))|0);
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              $79 = ___cxa_find_matching_catch_2()|0;
              $80 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
              $$1 = $80;$$132 = $79;
              break;
             }
             ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
             $$0$i$i$i53 = 0;
             while(1) {
              $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
              if ($exitcond$i$i$i54) {
               break;
              }
              $67 = (($6) + ($$0$i$i$i53<<2)|0);
              HEAP32[$67>>2] = 0;
              $68 = (($$0$i$i$i53) + 1)|0;
              $$0$i$i$i53 = $68;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
             $69 = HEAP32[$24>>2]|0;
             $70 = ((($69)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($70,$4);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
             $$341 = $22;
             break L1;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
           $$2 = $$1;$$233 = $$132;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
          $$3 = $$2;$$334 = $$233;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$4 = $$3;$$435 = $$334;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$435|0);
        // unreachable;
       }
      }
     }
    } else {
     $$341 = $0;
    }
   } else {
    $$341 = $0;
   }
  } else {
   $$341 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$341|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_122parse_static_cast_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i47 = 0, $$0$i$i$i50 = 0, $$0$i$i$i53 = 0, $$1 = 0, $$132 = 0, $$2 = 0, $$233 = 0, $$3 = 0, $$334 = 0, $$341 = 0, $$4 = 0, $$435 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0;
 var $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0;
 var $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0;
 var $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i48 = 0, $exitcond$i$i$i51 = 0, $exitcond$i$i$i54 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $3 = sp + 88|0;
 $4 = sp + 64|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(115);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(99);
    if ($18) {
     $19 = ((($0)) + 2|0);
     $20 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($19,$1,$2)|0);
     $21 = ($20|0)==($19|0);
     if ($21) {
      $$341 = $0;
     } else {
      $22 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($20,$1,$2)|0);
      $23 = ($22|0)==($20|0);
      if ($23) {
       $$341 = $0;
      } else {
       $24 = ((($2)) + 4|0);
       $25 = HEAP32[$24>>2]|0;
       $26 = HEAP32[$2>>2]|0;
       $27 = (($25) - ($26))|0;
       $28 = (($27|0) / 24)&-1;
       $29 = ($28>>>0)<(2);
       if ($29) {
        $$341 = $0;
       } else {
        $30 = $25;
        $31 = ((($30)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$31);
        $32 = HEAP32[$24>>2]|0;
        $33 = ((($32)) + -24|0);
        $35 = $32;
        while(1) {
         $34 = ($35|0)==($33|0);
         if ($34) {
          break;
         }
         $36 = ((($35)) + -24|0);
         HEAP32[$24>>2] = $36;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
         $$pre$i = HEAP32[$24>>2]|0;
         $35 = $$pre$i;
        }
        $37 = ((($32)) + -48|0);
        __THREW__ = 0;
        invoke_vii(1120,($9|0),($37|0));
        $38 = __THREW__; __THREW__ = 0;
        $39 = $38&1;
        if ($39) {
         $71 = ___cxa_find_matching_catch_2()|0;
         $72 = tempRet0;
         $$4 = $72;$$435 = $71;
        } else {
         __THREW__ = 0;
         $40 = (invoke_iiii(1121,($9|0),0,(159317|0))|0);
         $41 = __THREW__; __THREW__ = 0;
         $42 = $41&1;
         if ($42) {
          $73 = ___cxa_find_matching_catch_2()|0;
          $74 = tempRet0;
          $$3 = $74;$$334 = $73;
         } else {
          ;HEAP32[$8>>2]=HEAP32[$9>>2]|0;HEAP32[$8+4>>2]=HEAP32[$9+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$9+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $43 = (($9) + ($$0$i$i$i<<2)|0);
           HEAP32[$43>>2] = 0;
           $44 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $44;
          }
          __THREW__ = 0;
          $45 = (invoke_iii(1119,($8|0),(159330|0))|0);
          $46 = __THREW__; __THREW__ = 0;
          $47 = $46&1;
          if ($47) {
           $75 = ___cxa_find_matching_catch_2()|0;
           $76 = tempRet0;
           $$2 = $76;$$233 = $75;
          } else {
           ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
           $$0$i$i$i47 = 0;
           while(1) {
            $exitcond$i$i$i48 = ($$0$i$i$i47|0)==(3);
            if ($exitcond$i$i$i48) {
             break;
            }
            $48 = (($8) + ($$0$i$i$i47<<2)|0);
            HEAP32[$48>>2] = 0;
            $49 = (($$0$i$i$i47) + 1)|0;
            $$0$i$i$i47 = $49;
           }
           $50 = ((($3)) + 11|0);
           $51 = HEAP8[$50>>0]|0;
           $52 = ($51<<24>>24)<(0);
           $53 = HEAP32[$3>>2]|0;
           $54 = $52 ? $53 : $3;
           $55 = ((($3)) + 4|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $51&255;
           $58 = $52 ? $56 : $57;
           __THREW__ = 0;
           $59 = (invoke_iiii(1112,($7|0),($54|0),($58|0))|0);
           $60 = __THREW__; __THREW__ = 0;
           $61 = $60&1;
           do {
            if ($61) {
             $77 = ___cxa_find_matching_catch_2()|0;
             $78 = tempRet0;
             $$1 = $78;$$132 = $77;
            } else {
             ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
             $$0$i$i$i50 = 0;
             while(1) {
              $exitcond$i$i$i51 = ($$0$i$i$i50|0)==(3);
              if ($exitcond$i$i$i51) {
               break;
              }
              $62 = (($7) + ($$0$i$i$i50<<2)|0);
              HEAP32[$62>>2] = 0;
              $63 = (($$0$i$i$i50) + 1)|0;
              $$0$i$i$i50 = $63;
             }
             __THREW__ = 0;
             $64 = (invoke_iii(1119,($6|0),(159081|0))|0);
             $65 = __THREW__; __THREW__ = 0;
             $66 = $65&1;
             if ($66) {
              $79 = ___cxa_find_matching_catch_2()|0;
              $80 = tempRet0;
              __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
              $$1 = $80;$$132 = $79;
              break;
             }
             ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
             $$0$i$i$i53 = 0;
             while(1) {
              $exitcond$i$i$i54 = ($$0$i$i$i53|0)==(3);
              if ($exitcond$i$i$i54) {
               break;
              }
              $67 = (($6) + ($$0$i$i$i53<<2)|0);
              HEAP32[$67>>2] = 0;
              $68 = (($$0$i$i$i53) + 1)|0;
              $$0$i$i$i53 = $68;
             }
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
             $69 = HEAP32[$24>>2]|0;
             $70 = ((($69)) + -24|0);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($70,$4);
             __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
             __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
             $$341 = $22;
             break L1;
            }
           } while(0);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
           $$2 = $$1;$$233 = $$132;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
          $$3 = $$2;$$334 = $$233;
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$4 = $$3;$$435 = $$334;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        ___resumeException($$435|0);
        // unreachable;
       }
      }
     }
    } else {
     $$341 = $0;
    }
   } else {
    $$341 = $0;
   }
  } else {
   $$341 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$341|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_120parse_pack_expansionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $1;
 $4 = $0;
 $5 = (($3) - ($4))|0;
 $6 = ($5|0)>(2);
 if ($6) {
  $7 = HEAP8[$0>>0]|0;
  $8 = ($7<<24>>24)==(115);
  if ($8) {
   $9 = ((($0)) + 1|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)==(112);
   if ($11) {
    $12 = ((($0)) + 2|0);
    $13 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($12,$1,$2)|0);
    $14 = ($13|0)==($12|0);
    $$ = $14 ? $0 : $13;
    $$1 = $$;
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } else {
  $$1 = $0;
 }
 return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_122parse_sizeof_type_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i27 = 0, $$1 = 0, $$121 = 0, $$124 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i28 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(115);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = ($14<<24>>24)==(116);
    if ($15) {
     $16 = ((($0)) + 2|0);
     $17 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $18 = ($17|0)==($16|0);
     if ($18) {
      $$124 = $0;
     } else {
      $19 = HEAP32[$2>>2]|0;
      $20 = ((($2)) + 4|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ($19|0)==($21|0);
      if ($22) {
       $$124 = $0;
      } else {
       $23 = ((($21)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$23);
       __THREW__ = 0;
       $24 = (invoke_iiii(1121,($6|0),0,(159308|0))|0);
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       do {
        if ($26) {
         $36 = ___cxa_find_matching_catch_2()|0;
         $37 = tempRet0;
         $$1 = $37;$$121 = $36;
        } else {
         ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $27 = (($6) + ($$0$i$i$i<<2)|0);
          HEAP32[$27>>2] = 0;
          $28 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $28;
         }
         __THREW__ = 0;
         $29 = (invoke_iii(1119,($5|0),(159081|0))|0);
         $30 = __THREW__; __THREW__ = 0;
         $31 = $30&1;
         if ($31) {
          $38 = ___cxa_find_matching_catch_2()|0;
          $39 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
          $$1 = $39;$$121 = $38;
          break;
         }
         ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
         $$0$i$i$i27 = 0;
         while(1) {
          $exitcond$i$i$i28 = ($$0$i$i$i27|0)==(3);
          if ($exitcond$i$i$i28) {
           break;
          }
          $32 = (($5) + ($$0$i$i$i27<<2)|0);
          HEAP32[$32>>2] = 0;
          $33 = (($$0$i$i$i27) + 1)|0;
          $$0$i$i$i27 = $33;
         }
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
         $34 = HEAP32[$20>>2]|0;
         $35 = ((($34)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($35,$3);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         $$124 = $17;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       ___resumeException($$121|0);
       // unreachable;
      }
     }
    } else {
     $$124 = $0;
    }
   } else {
    $$124 = $0;
   }
  } else {
   $$124 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$124|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_122parse_sizeof_expr_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i27 = 0, $$1 = 0, $$121 = 0, $$124 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0;
 var $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i28 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(115);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = ($14<<24>>24)==(122);
    if ($15) {
     $16 = ((($0)) + 2|0);
     $17 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $18 = ($17|0)==($16|0);
     if ($18) {
      $$124 = $0;
     } else {
      $19 = HEAP32[$2>>2]|0;
      $20 = ((($2)) + 4|0);
      $21 = HEAP32[$20>>2]|0;
      $22 = ($19|0)==($21|0);
      if ($22) {
       $$124 = $0;
      } else {
       $23 = ((($21)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$23);
       __THREW__ = 0;
       $24 = (invoke_iiii(1121,($6|0),0,(159308|0))|0);
       $25 = __THREW__; __THREW__ = 0;
       $26 = $25&1;
       do {
        if ($26) {
         $36 = ___cxa_find_matching_catch_2()|0;
         $37 = tempRet0;
         $$1 = $37;$$121 = $36;
        } else {
         ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
         $$0$i$i$i = 0;
         while(1) {
          $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
          if ($exitcond$i$i$i) {
           break;
          }
          $27 = (($6) + ($$0$i$i$i<<2)|0);
          HEAP32[$27>>2] = 0;
          $28 = (($$0$i$i$i) + 1)|0;
          $$0$i$i$i = $28;
         }
         __THREW__ = 0;
         $29 = (invoke_iii(1119,($5|0),(159081|0))|0);
         $30 = __THREW__; __THREW__ = 0;
         $31 = $30&1;
         if ($31) {
          $38 = ___cxa_find_matching_catch_2()|0;
          $39 = tempRet0;
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
          $$1 = $39;$$121 = $38;
          break;
         }
         ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
         $$0$i$i$i27 = 0;
         while(1) {
          $exitcond$i$i$i28 = ($$0$i$i$i27|0)==(3);
          if ($exitcond$i$i$i28) {
           break;
          }
          $32 = (($5) + ($$0$i$i$i27<<2)|0);
          HEAP32[$32>>2] = 0;
          $33 = (($$0$i$i$i27) + 1)|0;
          $$0$i$i$i27 = $33;
         }
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
         $34 = HEAP32[$20>>2]|0;
         $35 = ((($34)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($35,$3);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
         $$124 = $17;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       ___resumeException($$121|0);
       // unreachable;
      }
     }
    } else {
     $$124 = $0;
    }
   } else {
    $$124 = $0;
   }
  } else {
   $$124 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$124|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_128parse_sizeof_param_pack_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i$i = 0, $$0$i$i52 = 0, $$041 = 0, $$045 = 0, $$045$in = 0, $$046 = 0, $$1 = 0, $$142 = 0, $$148 = 0, $$2 = 0, $$243 = 0, $$3 = 0, $$344 = 0, $$pre$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0;
 var $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0;
 var $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $3 = sp + 88|0;
 $4 = sp + 76|0;
 $5 = sp + 64|0;
 $6 = sp + 52|0;
 $7 = sp + 40|0;
 $8 = sp + 16|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = ($14<<24>>24)==(115);
   if ($15) {
    $16 = ((($0)) + 1|0);
    $17 = HEAP8[$16>>0]|0;
    $18 = ($17<<24>>24)==(90);
    if ($18) {
     $19 = ((($0)) + 2|0);
     $20 = HEAP8[$19>>0]|0;
     $21 = ($20<<24>>24)==(84);
     if ($21) {
      $22 = ((($2)) + 4|0);
      $23 = HEAP32[$22>>2]|0;
      $24 = HEAP32[$2>>2]|0;
      $25 = (($23) - ($24))|0;
      $26 = (($25|0) / 24)&-1;
      $27 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($19,$1,$2)|0);
      $28 = HEAP32[$22>>2]|0;
      $29 = HEAP32[$2>>2]|0;
      $30 = (($28) - ($29))|0;
      $31 = (($30|0) / 24)&-1;
      $32 = ($27|0)==($19|0);
      if ($32) {
       $$148 = $0;
      } else {
       ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
       $33 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159294)|0);
       $34 = ($33>>>0)>(4294967279);
       if ($34) {
        __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($4);
        // unreachable;
       }
       $35 = ($33>>>0)<(11);
       if ($35) {
        $36 = $33&255;
        $37 = ((($4)) + 11|0);
        HEAP8[$37>>0] = $36;
        $$0$i$i = $4;
       } else {
        $38 = (($33) + 16)|0;
        $39 = $38 & -16;
        $40 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($39)|0);
        HEAP32[$4>>2] = $40;
        $41 = $39 | -2147483648;
        $42 = ((($4)) + 8|0);
        HEAP32[$42>>2] = $41;
        $43 = ((($4)) + 4|0);
        HEAP32[$43>>2] = $33;
        $$0$i$i = $40;
       }
       (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,159294,$33)|0);
       $44 = (($$0$i$i) + ($33)|0);
       HEAP8[$3>>0] = 0;
       __ZNSt3__211char_traitsIcE6assignERcRKc($44,$3);
       $45 = ($26|0)==($31|0);
       L14: do {
        if ($45) {
         label = 31;
        } else {
         $46 = HEAP32[$2>>2]|0;
         $47 = (($46) + (($26*24)|0)|0);
         __THREW__ = 0;
         invoke_vii(1120,($5|0),($47|0));
         $48 = __THREW__; __THREW__ = 0;
         $49 = $48&1;
         do {
          if ($49) {
           $82 = ___cxa_find_matching_catch_2()|0;
           $83 = tempRet0;
           $$0 = $83;$$041 = $82;
          } else {
           $50 = ((($5)) + 11|0);
           $51 = HEAP8[$50>>0]|0;
           $52 = ($51<<24>>24)<(0);
           $53 = HEAP32[$5>>2]|0;
           $54 = $52 ? $53 : $5;
           $55 = ((($5)) + 4|0);
           $56 = HEAP32[$55>>2]|0;
           $57 = $51&255;
           $58 = $52 ? $56 : $57;
           __THREW__ = 0;
           (invoke_iiii(1112,($4|0),($54|0),($58|0))|0);
           $59 = __THREW__; __THREW__ = 0;
           $60 = $59&1;
           if ($60) {
            $84 = ___cxa_find_matching_catch_2()|0;
            $85 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
            $$0 = $85;$$041 = $84;
            break;
           }
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
           $61 = ((($6)) + 11|0);
           $62 = ((($6)) + 4|0);
           $$045$in = $26;
           while(1) {
            $$045 = (($$045$in) + 1)|0;
            $63 = ($$045|0)==($31|0);
            if ($63) {
             label = 31;
             break L14;
            }
            $64 = HEAP32[$2>>2]|0;
            $65 = (($64) + (($$045*24)|0)|0);
            __THREW__ = 0;
            invoke_vii(1120,($7|0),($65|0));
            $66 = __THREW__; __THREW__ = 0;
            $67 = $66&1;
            if ($67) {
             label = 26;
             break;
            }
            __THREW__ = 0;
            $68 = (invoke_iiii(1121,($7|0),0,(159305|0))|0);
            $69 = __THREW__; __THREW__ = 0;
            $70 = $69&1;
            if ($70) {
             label = 27;
             break;
            }
            ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
            $$0$i$i$i = 0;
            while(1) {
             $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
             if ($exitcond$i$i$i) {
              break;
             }
             $71 = (($7) + ($$0$i$i$i<<2)|0);
             HEAP32[$71>>2] = 0;
             $72 = (($$0$i$i$i) + 1)|0;
             $$0$i$i$i = $72;
            }
            $73 = HEAP8[$61>>0]|0;
            $74 = ($73<<24>>24)<(0);
            $75 = HEAP32[$6>>2]|0;
            $76 = $74 ? $75 : $6;
            $77 = HEAP32[$62>>2]|0;
            $78 = $73&255;
            $79 = $74 ? $77 : $78;
            __THREW__ = 0;
            (invoke_iiii(1112,($4|0),($76|0),($79|0))|0);
            $80 = __THREW__; __THREW__ = 0;
            $81 = $80&1;
            if ($81) {
             label = 28;
             break;
            }
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            $$045$in = $$045;
           }
           if ((label|0) == 26) {
            $86 = ___cxa_find_matching_catch_2()|0;
            $87 = tempRet0;
            $$2 = $87;$$243 = $86;
           }
           else if ((label|0) == 27) {
            $88 = ___cxa_find_matching_catch_2()|0;
            $89 = tempRet0;
            $$1 = $89;$$142 = $88;
            label = 29;
           }
           else if ((label|0) == 28) {
            $90 = ___cxa_find_matching_catch_2()|0;
            $91 = tempRet0;
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
            $$1 = $91;$$142 = $90;
            label = 29;
           }
           if ((label|0) == 29) {
            __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
            $$2 = $$1;$$243 = $$142;
           }
           $$3 = $$2;$$344 = $$243;
           break L14;
          }
         } while(0);
         $$3 = $$0;$$344 = $$041;
        }
       } while(0);
       do {
        if ((label|0) == 31) {
         __THREW__ = 0;
         (invoke_iii(1119,($4|0),(159081|0))|0);
         $92 = __THREW__; __THREW__ = 0;
         $93 = $92&1;
         if ($93) {
          $101 = ___cxa_find_matching_catch_2()|0;
          $102 = tempRet0;
          $$3 = $102;$$344 = $101;
          break;
         } else {
          $$046 = $31;
         }
         while(1) {
          $94 = ($$046|0)==($26|0);
          if ($94) {
           break;
          }
          $95 = HEAP32[$22>>2]|0;
          $96 = ((($95)) + -24|0);
          $98 = $95;
          while(1) {
           $97 = ($98|0)==($96|0);
           if ($97) {
            break;
           }
           $99 = ((($98)) + -24|0);
           HEAP32[$22>>2] = $99;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($99);
           $$pre$i = HEAP32[$22>>2]|0;
           $98 = $$pre$i;
          }
          $100 = (($$046) + -1)|0;
          $$046 = $100;
         }
         ;HEAP32[$9>>2]=HEAP32[$4>>2]|0;HEAP32[$9+4>>2]=HEAP32[$4+4>>2]|0;HEAP32[$9+8>>2]=HEAP32[$4+8>>2]|0;
         $$0$i$i52 = 0;
         while(1) {
          $exitcond$i$i = ($$0$i$i52|0)==(3);
          if ($exitcond$i$i) {
           break;
          }
          $103 = (($4) + ($$0$i$i52<<2)|0);
          HEAP32[$103>>2] = 0;
          $104 = (($$0$i$i52) + 1)|0;
          $$0$i$i52 = $104;
         }
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($8,$9);
         $105 = HEAP32[$22>>2]|0;
         $106 = ((($2)) + 8|0);
         $107 = HEAP32[$106>>2]|0;
         $108 = ($105>>>0)<($107>>>0);
         if ($108) {
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($105,$8);
          $109 = HEAP32[$22>>2]|0;
          $110 = ((($109)) + 24|0);
          HEAP32[$22>>2] = $110;
         } else {
          __THREW__ = 0;
          invoke_vii(1128,($2|0),($8|0));
          $111 = __THREW__; __THREW__ = 0;
          $112 = $111&1;
          if ($112) {
           $113 = ___cxa_find_matching_catch_2()|0;
           $114 = tempRet0;
           __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
           $$3 = $114;$$344 = $113;
           break;
          }
         }
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         $$148 = $27;
         break L1;
        }
       } while(0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
       ___resumeException($$344|0);
       // unreachable;
      }
     } else {
      $$148 = $0;
     }
    } else {
     $$148 = $0;
    }
   } else {
    $$148 = $0;
   }
  } else {
   $$148 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$148|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_137parse_sizeof_function_param_pack_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i28 = 0, $$1 = 0, $$122 = 0, $$125 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i29 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(115);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    $15 = ($14<<24>>24)==(90);
    if ($15) {
     $16 = ((($0)) + 2|0);
     $17 = HEAP8[$16>>0]|0;
     $18 = ($17<<24>>24)==(102);
     if ($18) {
      $19 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_function_paramINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
      $20 = ($19|0)==($16|0);
      if ($20) {
       $$125 = $0;
      } else {
       $21 = HEAP32[$2>>2]|0;
       $22 = ((($2)) + 4|0);
       $23 = HEAP32[$22>>2]|0;
       $24 = ($21|0)==($23|0);
       if ($24) {
        $$125 = $0;
       } else {
        $25 = ((($23)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$25);
        __THREW__ = 0;
        $26 = (invoke_iiii(1121,($6|0),0,(159294|0))|0);
        $27 = __THREW__; __THREW__ = 0;
        $28 = $27&1;
        do {
         if ($28) {
          $38 = ___cxa_find_matching_catch_2()|0;
          $39 = tempRet0;
          $$1 = $39;$$122 = $38;
         } else {
          ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
          $$0$i$i$i = 0;
          while(1) {
           $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
           if ($exitcond$i$i$i) {
            break;
           }
           $29 = (($6) + ($$0$i$i$i<<2)|0);
           HEAP32[$29>>2] = 0;
           $30 = (($$0$i$i$i) + 1)|0;
           $$0$i$i$i = $30;
          }
          __THREW__ = 0;
          $31 = (invoke_iii(1119,($5|0),(159081|0))|0);
          $32 = __THREW__; __THREW__ = 0;
          $33 = $32&1;
          if ($33) {
           $40 = ___cxa_find_matching_catch_2()|0;
           $41 = tempRet0;
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
           $$1 = $41;$$122 = $40;
           break;
          }
          ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
          $$0$i$i$i28 = 0;
          while(1) {
           $exitcond$i$i$i29 = ($$0$i$i$i28|0)==(3);
           if ($exitcond$i$i$i29) {
            break;
           }
           $34 = (($5) + ($$0$i$i$i28<<2)|0);
           HEAP32[$34>>2] = 0;
           $35 = (($$0$i$i$i28) + 1)|0;
           $$0$i$i$i28 = $35;
          }
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
          $36 = HEAP32[$22>>2]|0;
          $37 = ((($36)) + -24|0);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($37,$3);
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
          $$125 = $19;
          break L1;
         }
        } while(0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        ___resumeException($$122|0);
        // unreachable;
       }
      }
     } else {
      $$125 = $0;
     }
    } else {
     $$125 = $0;
    }
   } else {
    $$125 = $0;
   }
  } else {
   $$125 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$125|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_117parse_typeid_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i30 = 0, $$025 = 0, $$1 = 0, $$123 = 0, $$127 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $5 = 0;
 var $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i31 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 40|0;
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = $1;
 $8 = $0;
 $9 = (($7) - ($8))|0;
 $10 = ($9|0)>(2);
 L1: do {
  if ($10) {
   $11 = HEAP8[$0>>0]|0;
   $12 = ($11<<24>>24)==(116);
   if ($12) {
    $13 = ((($0)) + 1|0);
    $14 = HEAP8[$13>>0]|0;
    switch ($14<<24>>24) {
    case 101: case 105:  {
     break;
    }
    default: {
     $$127 = $0;
     break L1;
    }
    }
    $15 = ($14<<24>>24)==(101);
    $16 = ((($0)) + 2|0);
    if ($15) {
     $17 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $$025 = $17;
    } else {
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $$025 = $18;
    }
    $19 = ($$025|0)==($16|0);
    if ($19) {
     $$127 = $0;
    } else {
     $20 = HEAP32[$2>>2]|0;
     $21 = ((($2)) + 4|0);
     $22 = HEAP32[$21>>2]|0;
     $23 = ($20|0)==($22|0);
     if ($23) {
      $$127 = $0;
     } else {
      $24 = ((($22)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($6,$24);
      __THREW__ = 0;
      $25 = (invoke_iiii(1121,($6|0),0,(159286|0))|0);
      $26 = __THREW__; __THREW__ = 0;
      $27 = $26&1;
      do {
       if ($27) {
        $37 = ___cxa_find_matching_catch_2()|0;
        $38 = tempRet0;
        $$1 = $38;$$123 = $37;
       } else {
        ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
        $$0$i$i$i = 0;
        while(1) {
         $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
         if ($exitcond$i$i$i) {
          break;
         }
         $28 = (($6) + ($$0$i$i$i<<2)|0);
         HEAP32[$28>>2] = 0;
         $29 = (($$0$i$i$i) + 1)|0;
         $$0$i$i$i = $29;
        }
        __THREW__ = 0;
        $30 = (invoke_iii(1119,($5|0),(159081|0))|0);
        $31 = __THREW__; __THREW__ = 0;
        $32 = $31&1;
        if ($32) {
         $39 = ___cxa_find_matching_catch_2()|0;
         $40 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         $$1 = $40;$$123 = $39;
         break;
        }
        ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
        $$0$i$i$i30 = 0;
        while(1) {
         $exitcond$i$i$i31 = ($$0$i$i$i30|0)==(3);
         if ($exitcond$i$i$i31) {
          break;
         }
         $33 = (($5) + ($$0$i$i$i30<<2)|0);
         HEAP32[$33>>2] = 0;
         $34 = (($$0$i$i$i30) + 1)|0;
         $$0$i$i$i30 = $34;
        }
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
        $35 = HEAP32[$21>>2]|0;
        $36 = ((($35)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($36,$3);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        $$127 = $$025;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
      ___resumeException($$123|0);
      // unreachable;
     }
    }
   } else {
    $$127 = $0;
   }
  } else {
   $$127 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$127|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj6EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 5;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,5)|0);
 $4 = ((($0)) + 5|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_throw_exprINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $3 = sp + 24|0;
 $4 = sp + 12|0;
 $5 = sp;
 $6 = $1;
 $7 = $0;
 $8 = (($6) - ($7))|0;
 $9 = ($8|0)>(2);
 if ($9) {
  $10 = HEAP8[$0>>0]|0;
  $11 = ($10<<24>>24)==(116);
  if ($11) {
   $12 = ((($0)) + 1|0);
   $13 = HEAP8[$12>>0]|0;
   $14 = ($13<<24>>24)==(119);
   if ($14) {
    $15 = ((($0)) + 2|0);
    $16 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($15,$1,$2)|0);
    $17 = ($16|0)==($15|0);
    if ($17) {
     $$1 = $0;
    } else {
     $18 = HEAP32[$2>>2]|0;
     $19 = ((($2)) + 4|0);
     $20 = HEAP32[$19>>2]|0;
     $21 = ($18|0)==($20|0);
     if ($21) {
      $$1 = $0;
     } else {
      $22 = ((($20)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($5,$22);
      __THREW__ = 0;
      $23 = (invoke_iiii(1121,($5|0),0,(159279|0))|0);
      $24 = __THREW__; __THREW__ = 0;
      $25 = $24&1;
      if ($25) {
       $30 = ___cxa_find_matching_catch_2()|0;
       $31 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
       ___resumeException($30|0);
       // unreachable;
      }
      ;HEAP32[$4>>2]=HEAP32[$5>>2]|0;HEAP32[$4+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$4+8>>2]=HEAP32[$5+8>>2]|0;
      $$0$i$i$i = 0;
      while(1) {
       $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
       if ($exitcond$i$i$i) {
        break;
       }
       $26 = (($5) + ($$0$i$i$i<<2)|0);
       HEAP32[$26>>2] = 0;
       $27 = (($$0$i$i$i) + 1)|0;
       $$0$i$i$i = $27;
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
      $28 = HEAP32[$19>>2]|0;
      $29 = ((($28)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairaSEOS1_($29,$3);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
      $$1 = $16;
     }
    }
   } else {
    $$1 = $0;
   }
  } else {
   $$1 = $0;
  }
 } else {
  $$1 = $0;
 }
 STACKTOP = sp;return ($$1|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZNSt3__211char_traitsIcE6lengthEPKc($1)|0);
 __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKcj($0,$1,$2);
 return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKcj($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $7 = ((($0)) + 8|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = $8 & 2147483647;
  $phitmp$i = (($9) + -1)|0;
  $11 = $phitmp$i;
 } else {
  $11 = 10;
 }
 $10 = ($11>>>0)<($2>>>0);
 do {
  if ($10) {
   if ($6) {
    $19 = ((($0)) + 4|0);
    $20 = HEAP32[$19>>2]|0;
    $23 = $20;
   } else {
    $21 = $5&255;
    $23 = $21;
   }
   $22 = (($2) - ($11))|0;
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$11,$22,$23,0,$23,$2,$1);
  } else {
   if ($6) {
    $12 = HEAP32[$0>>2]|0;
    $13 = $12;
   } else {
    $13 = $0;
   }
   (__ZNSt3__211char_traitsIcE4moveEPcPKcj($13,$1,$2)|0);
   $14 = (($13) + ($2)|0);
   HEAP8[$3>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($14,$3);
   $15 = HEAP8[$4>>0]|0;
   $16 = ($15<<24>>24)<(0);
   if ($16) {
    $17 = ((($0)) + 4|0);
    HEAP32[$17>>2] = $2;
    break;
   } else {
    $18 = $2&255;
    HEAP8[$4>>0] = $18;
    break;
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_126parse_base_unresolved_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$061 = 0, $$062 = 0, $$2 = 0, $$pre$i = 0, $$pre$i68 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0;
 var $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0;
 var $79 = 0, $8 = 0, $80 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 12|0;
 $4 = sp;
 $5 = $1;
 $6 = $0;
 $7 = (($5) - ($6))|0;
 $8 = ($7|0)>(1);
 L1: do {
  if ($8) {
   $9 = HEAP8[$0>>0]|0;
   $10 = ($9<<24>>24)==(111);
   L3: do {
    switch ($9<<24>>24) {
    case 100: case 111:  {
     $11 = ((($0)) + 1|0);
     $12 = HEAP8[$11>>0]|0;
     $13 = ($12<<24>>24)==(110);
     if ($13) {
      $14 = ((($0)) + 2|0);
      if (!($10)) {
       $46 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_destructor_nameINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
       $47 = ($46|0)==($14|0);
       $$ = $47 ? $0 : $46;
       $$2 = $$;
       break L1;
      }
      $15 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_operator_nameINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
      $16 = ($15|0)==($14|0);
      if ($16) {
       $$2 = $0;
       break L1;
      }
      $17 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($15,$1,$2)|0);
      $18 = ($17|0)==($15|0);
      if ($18) {
       $$2 = $15;
       break L1;
      }
      $19 = ((($2)) + 4|0);
      $20 = HEAP32[$19>>2]|0;
      $21 = HEAP32[$2>>2]|0;
      $22 = (($20) - ($21))|0;
      $23 = (($22|0) / 24)&-1;
      $24 = ($23>>>0)<(2);
      if ($24) {
       $$2 = $17;
       break L1;
      }
      $25 = $20;
      $26 = ((($25)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$26);
      $27 = HEAP32[$19>>2]|0;
      $28 = ((($27)) + -24|0);
      $30 = $27;
      while(1) {
       $29 = ($30|0)==($28|0);
       if ($29) {
        break;
       }
       $31 = ((($30)) + -24|0);
       HEAP32[$19>>2] = $31;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
       $$pre$i = HEAP32[$19>>2]|0;
       $30 = $$pre$i;
      }
      $32 = ((($27)) + -48|0);
      $33 = ((($3)) + 11|0);
      $34 = HEAP8[$33>>0]|0;
      $35 = ($34<<24>>24)<(0);
      $36 = HEAP32[$3>>2]|0;
      $37 = $35 ? $36 : $3;
      $38 = ((($3)) + 4|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = $34&255;
      $41 = $35 ? $39 : $40;
      __THREW__ = 0;
      (invoke_iiii(1112,($32|0),($37|0),($41|0))|0);
      $42 = __THREW__; __THREW__ = 0;
      $43 = $42&1;
      if ($43) {
       $44 = ___cxa_find_matching_catch_2()|0;
       $45 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
       $$061 = $45;$$062 = $44;
       break L3;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
       $$2 = $17;
       break L1;
      }
     } else {
      label = 15;
     }
     break;
    }
    default: {
     label = 15;
    }
    }
   } while(0);
   do {
    if ((label|0) == 15) {
     $48 = (__ZN10__cxxabiv112_GLOBAL__N_115parse_simple_idINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $49 = ($48|0)==($0|0);
     if (!($49)) {
      $$2 = $48;
      break L1;
     }
     $50 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_operator_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $51 = ($50|0)==($0|0);
     if ($51) {
      $$2 = $0;
      break L1;
     }
     $52 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($50,$1,$2)|0);
     $53 = ($52|0)==($50|0);
     if ($53) {
      $$2 = $50;
      break L1;
     }
     $54 = ((($2)) + 4|0);
     $55 = HEAP32[$54>>2]|0;
     $56 = HEAP32[$2>>2]|0;
     $57 = (($55) - ($56))|0;
     $58 = (($57|0) / 24)&-1;
     $59 = ($58>>>0)<(2);
     if ($59) {
      $$2 = $52;
      break L1;
     }
     $60 = $55;
     $61 = ((($60)) + -24|0);
     __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($4,$61);
     $62 = HEAP32[$54>>2]|0;
     $63 = ((($62)) + -24|0);
     $65 = $62;
     while(1) {
      $64 = ($65|0)==($63|0);
      if ($64) {
       break;
      }
      $66 = ((($65)) + -24|0);
      HEAP32[$54>>2] = $66;
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($66);
      $$pre$i68 = HEAP32[$54>>2]|0;
      $65 = $$pre$i68;
     }
     $67 = ((($62)) + -48|0);
     $68 = ((($4)) + 11|0);
     $69 = HEAP8[$68>>0]|0;
     $70 = ($69<<24>>24)<(0);
     $71 = HEAP32[$4>>2]|0;
     $72 = $70 ? $71 : $4;
     $73 = ((($4)) + 4|0);
     $74 = HEAP32[$73>>2]|0;
     $75 = $69&255;
     $76 = $70 ? $74 : $75;
     __THREW__ = 0;
     (invoke_iiii(1112,($67|0),($72|0),($76|0))|0);
     $77 = __THREW__; __THREW__ = 0;
     $78 = $77&1;
     if ($78) {
      $79 = ___cxa_find_matching_catch_2()|0;
      $80 = tempRet0;
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      $$061 = $80;$$062 = $79;
      break;
     } else {
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      $$2 = $52;
      break L1;
     }
    }
   } while(0);
   ___resumeException($$062|0);
   // unreachable;
  } else {
   $$2 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$2|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$065 = 0, $$066 = 0, $$067 = 0, $$1 = 0, $$pre$i = 0, $$sroa$0$0$$sroa_idx$i = 0, $$sroa$0$0$$sroa_idx$i72 = 0, $$sroa$0$0$$sroa_idx$i76 = 0, $$sroa$0$0$copyload$i7381 = 0, $$sroa$0$0$copyload$i7780 = 0, $$sroa$0$0$copyload$i82 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0;
 var $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0;
 var $126 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0;
 var $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0;
 var $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 80|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(80|0);
 $3 = sp + 56|0;
 $4 = sp + 48|0;
 $5 = sp + 32|0;
 $6 = sp + 24|0;
 $7 = sp + 8|0;
 $8 = sp;
 $9 = ($0|0)==($1|0);
 L1: do {
  if ($9) {
   $$1 = $0;
  } else {
   $10 = HEAP8[$0>>0]|0;
   $11 = $10 << 24 >> 24;
   $12 = ((($2)) + 16|0);
   L3: do {
    switch ($11|0) {
    case 84:  {
     $13 = ((($2)) + 4|0);
     $14 = HEAP32[$13>>2]|0;
     $15 = HEAP32[$2>>2]|0;
     $16 = (($14) - ($15))|0;
     $17 = (($16|0) / 24)&-1;
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $19 = HEAP32[$13>>2]|0;
     $20 = HEAP32[$2>>2]|0;
     $21 = (($19) - ($20))|0;
     $22 = (($21|0) / 24)&-1;
     $23 = ($18|0)!=($0|0);
     $24 = (($17) + 1)|0;
     $25 = ($22|0)==($24|0);
     $or$cond = $23 & $25;
     $26 = $19;
     if ($or$cond) {
      $27 = ((($26)) + -24|0);
      $$sroa$0$0$$sroa_idx$i = ((($2)) + 12|0);
      $$sroa$0$0$copyload$i82 = HEAP32[$$sroa$0$0$$sroa_idx$i>>2]|0;
      HEAP32[$4>>2] = $$sroa$0$0$copyload$i82;
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($3,$27,$4);
      $28 = ((($2)) + 20|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = ((($2)) + 24|0);
      $31 = HEAP32[$30>>2]|0;
      $32 = ($29>>>0)<($31>>>0);
      if ($32) {
       $33 = ((($3)) + 12|0);
       HEAP32[$29>>2] = 0;
       $34 = ((($29)) + 4|0);
       HEAP32[$34>>2] = 0;
       $35 = ((($29)) + 8|0);
       $36 = HEAP32[$33>>2]|0;
       HEAP32[$35>>2] = 0;
       $37 = ((($29)) + 12|0);
       HEAP32[$37>>2] = $36;
       $38 = HEAP32[$3>>2]|0;
       HEAP32[$29>>2] = $38;
       $39 = ((($3)) + 4|0);
       $40 = HEAP32[$39>>2]|0;
       HEAP32[$34>>2] = $40;
       $41 = ((($3)) + 8|0);
       $42 = HEAP32[$41>>2]|0;
       $43 = ((($29)) + 8|0);
       HEAP32[$43>>2] = $42;
       HEAP32[$41>>2] = 0;
       HEAP32[$39>>2] = 0;
       HEAP32[$3>>2] = 0;
       $44 = HEAP32[$28>>2]|0;
       $45 = ((($44)) + 16|0);
       HEAP32[$28>>2] = $45;
      } else {
       __THREW__ = 0;
       invoke_vii(1122,($12|0),($3|0));
       $46 = __THREW__; __THREW__ = 0;
       $47 = $46&1;
       if ($47) {
        $48 = ___cxa_find_matching_catch_2()|0;
        $49 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($3);
        $$065 = $49;$$066 = $48;
        break L3;
       }
      }
      __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($3);
      $$1 = $18;
      break L1;
     } else {
      $$067 = $22;$52 = $26;
      while(1) {
       $50 = ($$067|0)==($17|0);
       if ($50) {
        $$1 = $0;
        break L1;
       }
       $51 = ((($52)) + -24|0);
       $54 = $52;
       while(1) {
        $53 = ($54|0)==($51|0);
        if ($53) {
         break;
        }
        $55 = ((($54)) + -24|0);
        HEAP32[$13>>2] = $55;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($55);
        $$pre$i = HEAP32[$13>>2]|0;
        $54 = $$pre$i;
       }
       $56 = (($$067) + -1)|0;
       $$067 = $56;$52 = $51;
      }
     }
     break;
    }
    case 68:  {
     $57 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_decltypeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $58 = ($57|0)==($0|0);
     if ($58) {
      $$1 = $0;
      break L1;
     }
     $59 = HEAP32[$2>>2]|0;
     $60 = ((($2)) + 4|0);
     $61 = HEAP32[$60>>2]|0;
     $62 = ($59|0)==($61|0);
     if ($62) {
      $$1 = $0;
      break L1;
     }
     $63 = ((($61)) + -24|0);
     $$sroa$0$0$$sroa_idx$i72 = ((($2)) + 12|0);
     $$sroa$0$0$copyload$i7381 = HEAP32[$$sroa$0$0$$sroa_idx$i72>>2]|0;
     HEAP32[$6>>2] = $$sroa$0$0$copyload$i7381;
     __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($5,$63,$6);
     $64 = ((($2)) + 20|0);
     $65 = HEAP32[$64>>2]|0;
     $66 = ((($2)) + 24|0);
     $67 = HEAP32[$66>>2]|0;
     $68 = ($65>>>0)<($67>>>0);
     if ($68) {
      $69 = ((($5)) + 12|0);
      HEAP32[$65>>2] = 0;
      $70 = ((($65)) + 4|0);
      HEAP32[$70>>2] = 0;
      $71 = ((($65)) + 8|0);
      $72 = HEAP32[$69>>2]|0;
      HEAP32[$71>>2] = 0;
      $73 = ((($65)) + 12|0);
      HEAP32[$73>>2] = $72;
      $74 = HEAP32[$5>>2]|0;
      HEAP32[$65>>2] = $74;
      $75 = ((($5)) + 4|0);
      $76 = HEAP32[$75>>2]|0;
      HEAP32[$70>>2] = $76;
      $77 = ((($5)) + 8|0);
      $78 = HEAP32[$77>>2]|0;
      $79 = ((($65)) + 8|0);
      HEAP32[$79>>2] = $78;
      HEAP32[$77>>2] = 0;
      HEAP32[$75>>2] = 0;
      HEAP32[$5>>2] = 0;
      $80 = HEAP32[$64>>2]|0;
      $81 = ((($80)) + 16|0);
      HEAP32[$64>>2] = $81;
     } else {
      __THREW__ = 0;
      invoke_vii(1122,($12|0),($5|0));
      $82 = __THREW__; __THREW__ = 0;
      $83 = $82&1;
      if ($83) {
       $84 = ___cxa_find_matching_catch_2()|0;
       $85 = tempRet0;
       __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($5);
       $$065 = $85;$$066 = $84;
       break L3;
      }
     }
     __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($5);
     $$1 = $57;
     break L1;
     break;
    }
    case 83:  {
     $86 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_substitutionINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
     $87 = ($86|0)==($0|0);
     if (!($87)) {
      $$1 = $86;
      break L1;
     }
     $88 = $1;
     $89 = $0;
     $90 = (($88) - ($89))|0;
     $91 = ($90|0)>(2);
     if (!($91)) {
      $$1 = $0;
      break L1;
     }
     $92 = ((($0)) + 1|0);
     $93 = HEAP8[$92>>0]|0;
     $94 = ($93<<24>>24)==(116);
     if (!($94)) {
      $$1 = $0;
      break L1;
     }
     $95 = ((($0)) + 2|0);
     $96 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_unqualified_nameINS0_2DbEEEPKcS4_S4_RT_($95,$1,$2)|0);
     $97 = ($96|0)==($95|0);
     if ($97) {
      $$1 = $0;
      break L1;
     }
     $98 = HEAP32[$2>>2]|0;
     $99 = ((($2)) + 4|0);
     $100 = HEAP32[$99>>2]|0;
     $101 = ($98|0)==($100|0);
     if ($101) {
      $$1 = $0;
      break L1;
     }
     $102 = ((($100)) + -24|0);
     (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($102,0,159377)|0);
     $103 = HEAP32[$99>>2]|0;
     $104 = ((($103)) + -24|0);
     $$sroa$0$0$$sroa_idx$i76 = ((($2)) + 12|0);
     $$sroa$0$0$copyload$i7780 = HEAP32[$$sroa$0$0$$sroa_idx$i76>>2]|0;
     HEAP32[$8>>2] = $$sroa$0$0$copyload$i7780;
     __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_($7,$104,$8);
     $105 = ((($2)) + 20|0);
     $106 = HEAP32[$105>>2]|0;
     $107 = ((($2)) + 24|0);
     $108 = HEAP32[$107>>2]|0;
     $109 = ($106>>>0)<($108>>>0);
     if ($109) {
      $110 = ((($7)) + 12|0);
      HEAP32[$106>>2] = 0;
      $111 = ((($106)) + 4|0);
      HEAP32[$111>>2] = 0;
      $112 = ((($106)) + 8|0);
      $113 = HEAP32[$110>>2]|0;
      HEAP32[$112>>2] = 0;
      $114 = ((($106)) + 12|0);
      HEAP32[$114>>2] = $113;
      $115 = HEAP32[$7>>2]|0;
      HEAP32[$106>>2] = $115;
      $116 = ((($7)) + 4|0);
      $117 = HEAP32[$116>>2]|0;
      HEAP32[$111>>2] = $117;
      $118 = ((($7)) + 8|0);
      $119 = HEAP32[$118>>2]|0;
      $120 = ((($106)) + 8|0);
      HEAP32[$120>>2] = $119;
      HEAP32[$118>>2] = 0;
      HEAP32[$116>>2] = 0;
      HEAP32[$7>>2] = 0;
      $121 = HEAP32[$105>>2]|0;
      $122 = ((($121)) + 16|0);
      HEAP32[$105>>2] = $122;
     } else {
      __THREW__ = 0;
      invoke_vii(1122,($12|0),($7|0));
      $123 = __THREW__; __THREW__ = 0;
      $124 = $123&1;
      if ($124) {
       $125 = ___cxa_find_matching_catch_2()|0;
       $126 = tempRet0;
       __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($7);
       $$065 = $126;$$066 = $125;
       break L3;
      }
     }
     __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($7);
     $$1 = $96;
     break L1;
     break;
    }
    default: {
     $$1 = $0;
     break L1;
    }
    }
   } while(0);
   ___resumeException($$066|0);
   // unreachable;
  }
 } while(0);
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_132parse_unresolved_qualifier_levelINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_115parse_simple_idINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
 return ($3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_115parse_simple_idINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$3 = 0, $$pre$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = ($0|0)==($1|0);
 do {
  if ($4) {
   $$3 = $0;
  } else {
   $5 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
   $6 = ($5|0)==($0|0);
   if ($6) {
    $$3 = $0;
   } else {
    $7 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($5,$1,$2)|0);
    $8 = ($7|0)==($5|0);
    if ($8) {
     $$3 = $5;
    } else {
     $9 = ((($2)) + 4|0);
     $10 = HEAP32[$9>>2]|0;
     $11 = HEAP32[$2>>2]|0;
     $12 = (($10) - ($11))|0;
     $13 = (($12|0) / 24)&-1;
     $14 = ($13>>>0)<(2);
     if ($14) {
      $$3 = $0;
     } else {
      $15 = $10;
      $16 = ((($15)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$16);
      $17 = HEAP32[$9>>2]|0;
      $18 = ((($17)) + -24|0);
      $20 = $17;
      while(1) {
       $19 = ($20|0)==($18|0);
       if ($19) {
        break;
       }
       $21 = ((($20)) + -24|0);
       HEAP32[$9>>2] = $21;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($21);
       $$pre$i = HEAP32[$9>>2]|0;
       $20 = $$pre$i;
      }
      $22 = ((($17)) + -48|0);
      $23 = ((($3)) + 11|0);
      $24 = HEAP8[$23>>0]|0;
      $25 = ($24<<24>>24)<(0);
      $26 = HEAP32[$3>>2]|0;
      $27 = $25 ? $26 : $3;
      $28 = ((($3)) + 4|0);
      $29 = HEAP32[$28>>2]|0;
      $30 = $24&255;
      $31 = $25 ? $29 : $30;
      __THREW__ = 0;
      (invoke_iiii(1112,($22|0),($27|0),($31|0))|0);
      $32 = __THREW__; __THREW__ = 0;
      $33 = $32&1;
      if ($33) {
       $34 = ___cxa_find_matching_catch_2()|0;
       $35 = tempRet0;
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
       ___resumeException($34|0);
       // unreachable;
      } else {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
       $$3 = $7;
       break;
      }
     }
    }
   }
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_122parse_unqualified_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$1 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==($1|0);
 L1: do {
  if ($3) {
   $$1 = $0;
  } else {
   $4 = HEAP8[$0>>0]|0;
   $5 = $4 << 24 >> 24;
   switch ($5|0) {
   case 68: case 67:  {
    $6 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_ctor_dtor_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    $$1 = $6;
    break L1;
    break;
   }
   case 85:  {
    $7 = (__ZN10__cxxabiv112_GLOBAL__N_123parse_unnamed_type_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    $$1 = $7;
    break L1;
    break;
   }
   case 57: case 56: case 55: case 54: case 53: case 52: case 51: case 50: case 49:  {
    $8 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    $$1 = $8;
    break L1;
    break;
   }
   default: {
    $9 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_operator_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    return ($9|0);
   }
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_120parse_ctor_dtor_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i$i = 0, $$021 = 0, $$023 = 0, $$1 = 0, $$122 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0;
 var $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0;
 var $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $3 = sp + 64|0;
 $4 = sp + 48|0;
 $5 = sp + 24|0;
 $6 = sp + 12|0;
 $7 = sp;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10|0)>(1);
 L1: do {
  if ($11) {
   $12 = HEAP32[$2>>2]|0;
   $13 = ((($2)) + 4|0);
   $14 = HEAP32[$13>>2]|0;
   $15 = ($12|0)==($14|0);
   if ($15) {
    $$023 = $0;
   } else {
    $16 = HEAP8[$0>>0]|0;
    $17 = $16 << 24 >> 24;
    $18 = ((($0)) + 2|0);
    $19 = ((($2)) + 60|0);
    $20 = ((($0)) + 1|0);
    L4: do {
     switch ($17|0) {
     case 67:  {
      $21 = HEAP8[$20>>0]|0;
      $22 = $21 << 24 >> 24;
      switch ($22|0) {
      case 53: case 51: case 50: case 49:  {
       break;
      }
      default: {
       $$023 = $0;
       break L1;
      }
      }
      $23 = ((($14)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_19base_nameINSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEEEET_RS9_($4,$23);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($3,$4);
      $24 = HEAP32[$13>>2]|0;
      $25 = ((($2)) + 8|0);
      $26 = HEAP32[$25>>2]|0;
      $27 = ($24>>>0)<($26>>>0);
      if ($27) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($24,$3);
       $28 = HEAP32[$13>>2]|0;
       $29 = ((($28)) + 24|0);
       HEAP32[$13>>2] = $29;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($3|0));
       $30 = __THREW__; __THREW__ = 0;
       $31 = $30&1;
       if ($31) {
        $32 = ___cxa_find_matching_catch_2()|0;
        $33 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
        $$1 = $33;$$122 = $32;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
      HEAP8[$19>>0] = 1;
      $$023 = $18;
      break L1;
      break;
     }
     case 68:  {
      $34 = HEAP8[$20>>0]|0;
      $35 = $34 << 24 >> 24;
      switch ($35|0) {
      case 53: case 50: case 49: case 48:  {
       break;
      }
      default: {
       $$023 = $0;
       break L1;
      }
      }
      $36 = ((($14)) + -24|0);
      __ZN10__cxxabiv112_GLOBAL__N_19base_nameINSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEEEET_RS9_($7,$36);
      __THREW__ = 0;
      $37 = (invoke_iiii(1121,($7|0),0,(159153|0))|0);
      $38 = __THREW__; __THREW__ = 0;
      $39 = $38&1;
      do {
       if ($39) {
        $50 = ___cxa_find_matching_catch_2()|0;
        $51 = tempRet0;
        $$0 = $51;$$021 = $50;
       } else {
        ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
        $$0$i$i$i = 0;
        while(1) {
         $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
         if ($exitcond$i$i$i) {
          break;
         }
         $40 = (($7) + ($$0$i$i$i<<2)|0);
         HEAP32[$40>>2] = 0;
         $41 = (($$0$i$i$i) + 1)|0;
         $$0$i$i$i = $41;
        }
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($5,$6);
        $42 = HEAP32[$13>>2]|0;
        $43 = ((($2)) + 8|0);
        $44 = HEAP32[$43>>2]|0;
        $45 = ($42>>>0)<($44>>>0);
        if ($45) {
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($42,$5);
         $46 = HEAP32[$13>>2]|0;
         $47 = ((($46)) + 24|0);
         HEAP32[$13>>2] = $47;
        } else {
         __THREW__ = 0;
         invoke_vii(1128,($2|0),($5|0));
         $48 = __THREW__; __THREW__ = 0;
         $49 = $48&1;
         if ($49) {
          $52 = ___cxa_find_matching_catch_2()|0;
          $53 = tempRet0;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
          $$0 = $53;$$021 = $52;
          break;
         }
        }
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        HEAP8[$19>>0] = 1;
        $$023 = $18;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$1 = $$0;$$122 = $$021;
      break;
     }
     default: {
      $$023 = $0;
      break L1;
     }
     }
    } while(0);
    ___resumeException($$122|0);
    // unreachable;
   }
  } else {
   $$023 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$023|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_123parse_unnamed_type_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i134 = 0, $$0105 = 0, $$0105$lcssa = 0, $$0106 = 0, $$0110 = 0, $$0110$lcssa = 0, $$0116 = 0, $$0117 = 0, $$0119 = 0, $$1118 = 0, $$1120 = 0, $$2108 = 0, $$3109 = 0, $$7 = 0, $$byval_copy = 0, $$pre$i = 0, $$pre$i132 = 0, $$pre$i138 = 0;
 var $$pre$i140 = 0, $$pre$i143 = 0, $$pre$i147 = 0, $$pre$i149 = 0, $$pre$i151 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0;
 var $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0;
 var $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0;
 var $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0;
 var $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0;
 var $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0;
 var $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0;
 var $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i$i = 0, $isdigit = 0, $isdigit127 = 0, $isdigit129 = 0, $isdigit131 = 0, $isdigittmp = 0, $isdigittmp126 = 0, $isdigittmp128 = 0, $isdigittmp130 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $$byval_copy = sp + 104|0;
 $3 = sp + 111|0;
 $4 = sp + 110|0;
 $5 = sp + 109|0;
 $6 = sp + 108|0;
 $7 = sp + 80|0;
 $8 = sp + 64|0;
 $9 = sp + 40|0;
 $10 = sp + 28|0;
 $11 = sp + 16|0;
 $12 = sp + 4|0;
 $13 = sp;
 $14 = $1;
 $15 = $0;
 $16 = (($14) - ($15))|0;
 $17 = ($16|0)>(2);
 L1: do {
  if ($17) {
   $18 = HEAP8[$0>>0]|0;
   $19 = ($18<<24>>24)==(85);
   if ($19) {
    $20 = ((($0)) + 1|0);
    $21 = HEAP8[$20>>0]|0;
    $22 = $21 << 24 >> 24;
    $23 = ((($0)) + 2|0);
    L4: do {
     switch ($22|0) {
     case 116:  {
      ;HEAP32[$8>>2]=0|0;HEAP32[$8+4>>2]=0|0;HEAP32[$8+8>>2]=0|0;
      $24 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159884)|0);
      $25 = ($24>>>0)>(4294967279);
      if ($25) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($8);
       // unreachable;
      }
      $26 = ($24>>>0)<(11);
      if ($26) {
       $27 = $24&255;
       $28 = ((($8)) + 11|0);
       HEAP8[$28>>0] = $27;
       $$0$i$i = $8;
      } else {
       $29 = (($24) + 16)|0;
       $30 = $29 & -16;
       $31 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($30)|0);
       HEAP32[$8>>2] = $31;
       $32 = $30 | -2147483648;
       $33 = ((($8)) + 8|0);
       HEAP32[$33>>2] = $32;
       $34 = ((($8)) + 4|0);
       HEAP32[$34>>2] = $24;
       $$0$i$i = $31;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i,159884,$24)|0);
      $35 = (($$0$i$i) + ($24)|0);
      HEAP8[$6>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($35,$6);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($7,$8);
      $36 = ((($2)) + 4|0);
      $37 = HEAP32[$36>>2]|0;
      $38 = ((($2)) + 8|0);
      $39 = HEAP32[$38>>2]|0;
      $40 = ($37>>>0)<($39>>>0);
      if ($40) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($37,$7);
       $41 = HEAP32[$36>>2]|0;
       $42 = ((($41)) + 24|0);
       HEAP32[$36>>2] = $42;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($7|0));
       $43 = __THREW__; __THREW__ = 0;
       $44 = $43&1;
       if ($44) {
        $51 = ___cxa_find_matching_catch_2()|0;
        $52 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
        $$1118 = $52;$$1120 = $51;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
      $45 = ($23|0)==($1|0);
      if ($45) {
       $46 = HEAP32[$36>>2]|0;
       $47 = ((($46)) + -24|0);
       $49 = $46;
       while(1) {
        $48 = ($49|0)==($47|0);
        if ($48) {
         $$7 = $0;
         break L1;
        }
        $50 = ((($49)) + -24|0);
        HEAP32[$36>>2] = $50;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($50);
        $$pre$i = HEAP32[$36>>2]|0;
        $49 = $$pre$i;
       }
      }
      $53 = HEAP8[$23>>0]|0;
      $54 = $53 << 24 >> 24;
      $isdigittmp128 = (($54) + -48)|0;
      $isdigit129 = ($isdigittmp128>>>0)<(10);
      if ($isdigit129) {
       $55 = ((($0)) + 3|0);
       $$0110 = $55;
       while(1) {
        $56 = ($$0110|0)==($1|0);
        if ($56) {
         $$0110$lcssa = $1;
         break;
        }
        $57 = HEAP8[$$0110>>0]|0;
        $58 = $57 << 24 >> 24;
        $isdigittmp130 = (($58) + -48)|0;
        $isdigit131 = ($isdigittmp130>>>0)<(10);
        if (!($isdigit131)) {
         $$0110$lcssa = $$0110;
         break;
        }
        $59 = ((($$0110)) + 1|0);
        $$0110 = $59;
       }
       $60 = HEAP32[$36>>2]|0;
       $61 = ((($60)) + -24|0);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE23__append_forward_unsafeIPKcEERS7_T_SC_($61,$23,$$0110$lcssa);
       $$0116 = $$0110$lcssa;
      } else {
       $$0116 = $23;
      }
      $62 = HEAP32[$36>>2]|0;
      $63 = ((($62)) + -24|0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9push_backEc($63,39);
      $64 = ($$0116|0)==($1|0);
      if (!($64)) {
       $65 = HEAP8[$$0116>>0]|0;
       $66 = ($65<<24>>24)==(95);
       if ($66) {
        $72 = ((($$0116)) + 1|0);
        $$7 = $72;
        break L1;
       }
      }
      $67 = HEAP32[$36>>2]|0;
      $68 = ((($67)) + -24|0);
      $70 = $67;
      while(1) {
       $69 = ($70|0)==($68|0);
       if ($69) {
        $$7 = $0;
        break L1;
       }
       $71 = ((($70)) + -24|0);
       HEAP32[$36>>2] = $71;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($71);
       $$pre$i132 = HEAP32[$36>>2]|0;
       $70 = $$pre$i132;
      }
      break;
     }
     case 108:  {
      ;HEAP32[$10>>2]=0|0;HEAP32[$10+4>>2]=0|0;HEAP32[$10+8>>2]=0|0;
      $73 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159893)|0);
      $74 = ($73>>>0)>(4294967279);
      if ($74) {
       __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($10);
       // unreachable;
      }
      $75 = ($73>>>0)<(11);
      if ($75) {
       $76 = $73&255;
       $77 = ((($10)) + 11|0);
       HEAP8[$77>>0] = $76;
       $$0$i$i134 = $10;
      } else {
       $78 = (($73) + 16)|0;
       $79 = $78 & -16;
       $80 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($79)|0);
       HEAP32[$10>>2] = $80;
       $81 = $79 | -2147483648;
       $82 = ((($10)) + 8|0);
       HEAP32[$82>>2] = $81;
       $83 = ((($10)) + 4|0);
       HEAP32[$83>>2] = $73;
       $$0$i$i134 = $80;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i134,159893,$73)|0);
      $84 = (($$0$i$i134) + ($73)|0);
      HEAP8[$5>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($84,$5);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($9,$10);
      $85 = ((($2)) + 4|0);
      $86 = HEAP32[$85>>2]|0;
      $87 = ((($2)) + 8|0);
      $88 = HEAP32[$87>>2]|0;
      $89 = ($86>>>0)<($88>>>0);
      if ($89) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($86,$9);
       $90 = HEAP32[$85>>2]|0;
       $91 = ((($90)) + 24|0);
       HEAP32[$85>>2] = $91;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($9|0));
       $92 = __THREW__; __THREW__ = 0;
       $93 = $92&1;
       if ($93) {
        $99 = ___cxa_find_matching_catch_2()|0;
        $100 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
        $$1118 = $100;$$1120 = $99;
        break L4;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($10);
      $94 = HEAP8[$23>>0]|0;
      $95 = ($94<<24>>24)==(118);
      L52: do {
       if ($95) {
        $96 = HEAP32[$85>>2]|0;
        $97 = ((($96)) + -24|0);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9push_backEc($97,41);
        $98 = ((($0)) + 3|0);
        $$2108 = $98;
       } else {
        $101 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($23,$1,$2)|0);
        $102 = ($101|0)==($23|0);
        if ($102) {
         $103 = HEAP32[$85>>2]|0;
         $104 = ((($103)) + -24|0);
         $106 = $103;
         while(1) {
          $105 = ($106|0)==($104|0);
          if ($105) {
           $$7 = $0;
           break L1;
          }
          $107 = ((($106)) + -24|0);
          HEAP32[$85>>2] = $107;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($107);
          $$pre$i138 = HEAP32[$85>>2]|0;
          $106 = $$pre$i138;
         }
        }
        $108 = HEAP32[$85>>2]|0;
        $109 = HEAP32[$2>>2]|0;
        $110 = (($108) - ($109))|0;
        $111 = (($110|0) / 24)&-1;
        $112 = ($111>>>0)<(2);
        if ($112) {
         $$7 = $0;
         break L1;
        }
        $113 = $108;
        $114 = ((($113)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($11,$114);
        $115 = HEAP32[$85>>2]|0;
        $116 = ((($115)) + -24|0);
        $118 = $115;
        while(1) {
         $117 = ($118|0)==($116|0);
         if ($117) {
          break;
         }
         $119 = ((($118)) + -24|0);
         HEAP32[$85>>2] = $119;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($119);
         $$pre$i140 = HEAP32[$85>>2]|0;
         $118 = $$pre$i140;
        }
        $120 = ((($115)) + -48|0);
        $121 = ((($11)) + 11|0);
        $122 = HEAP8[$121>>0]|0;
        $123 = ($122<<24>>24)<(0);
        $124 = HEAP32[$11>>2]|0;
        $125 = $123 ? $124 : $11;
        $126 = ((($11)) + 4|0);
        $127 = HEAP32[$126>>2]|0;
        $128 = $122&255;
        $129 = $123 ? $127 : $128;
        __THREW__ = 0;
        (invoke_iiii(1112,($120|0),($125|0),($129|0))|0);
        $130 = __THREW__; __THREW__ = 0;
        $131 = $130&1;
        L66: do {
         if ($131) {
          label = 52;
         } else {
          $$0106 = $101;
          while(1) {
           __THREW__ = 0;
           $132 = (invoke_iiii(1115,($$0106|0),($1|0),($2|0))|0);
           $133 = __THREW__; __THREW__ = 0;
           $134 = $133&1;
           if ($134) {
            label = 52;
            break L66;
           }
           $135 = ($132|0)==($$0106|0);
           if ($135) {
            label = 70;
            break;
           }
           $138 = HEAP32[$85>>2]|0;
           $139 = HEAP32[$2>>2]|0;
           $140 = (($138) - ($139))|0;
           $141 = (($140|0) / 24)&-1;
           $142 = ($141>>>0)<(2);
           if ($142) {
            label = 98;
            break;
           }
           $143 = $138;
           $144 = ((($143)) + -24|0);
           __THREW__ = 0;
           invoke_vii(1120,($12|0),($144|0));
           $145 = __THREW__; __THREW__ = 0;
           $146 = $145&1;
           if ($146) {
            label = 69;
            break;
           }
           $147 = HEAP8[$121>>0]|0;
           $148 = ($147<<24>>24)<(0);
           if ($148) {
            $149 = HEAP32[$11>>2]|0;
            HEAP8[$3>>0] = 0;
            __ZNSt3__211char_traitsIcE6assignERcRKc($149,$3);
            HEAP32[$126>>2] = 0;
           } else {
            HEAP8[$4>>0] = 0;
            __ZNSt3__211char_traitsIcE6assignERcRKc($11,$4);
            HEAP8[$121>>0] = 0;
           }
           __THREW__ = 0;
           invoke_vi(1116,($11|0));
           $150 = __THREW__; __THREW__ = 0;
           $151 = $150&1;
           if ($151) {
            label = 59;
            break;
           }
           ;HEAP32[$11>>2]=HEAP32[$12>>2]|0;HEAP32[$11+4>>2]=HEAP32[$12+4>>2]|0;HEAP32[$11+8>>2]=HEAP32[$12+8>>2]|0;
           $$0$i$i$i$i = 0;
           while(1) {
            $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
            if ($exitcond$i$i$i$i) {
             break;
            }
            $154 = (($12) + ($$0$i$i$i$i<<2)|0);
            HEAP32[$154>>2] = 0;
            $155 = (($$0$i$i$i$i) + 1)|0;
            $$0$i$i$i$i = $155;
           }
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
           $156 = HEAP32[$85>>2]|0;
           $157 = ((($156)) + -24|0);
           $159 = $156;
           while(1) {
            $158 = ($159|0)==($157|0);
            if ($158) {
             break;
            }
            $160 = ((($159)) + -24|0);
            HEAP32[$85>>2] = $160;
            __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($160);
            $$pre$i143 = HEAP32[$85>>2]|0;
            $159 = $$pre$i143;
           }
           $161 = HEAP8[$121>>0]|0;
           $162 = ($161<<24>>24)<(0);
           $163 = HEAP32[$126>>2]|0;
           $164 = $161&255;
           $165 = $162 ? $163 : $164;
           $166 = ($165|0)==(0);
           if ($166) {
            $$0106 = $132;
            continue;
           }
           $167 = ((($156)) + -48|0);
           __THREW__ = 0;
           (invoke_iii(1119,($167|0),(159305|0))|0);
           $168 = __THREW__; __THREW__ = 0;
           $169 = $168&1;
           if ($169) {
            label = 52;
            break L66;
           }
           $170 = HEAP32[$85>>2]|0;
           $171 = ((($170)) + -24|0);
           $172 = HEAP8[$121>>0]|0;
           $173 = ($172<<24>>24)<(0);
           $174 = HEAP32[$11>>2]|0;
           $175 = $173 ? $174 : $11;
           $176 = HEAP32[$126>>2]|0;
           $177 = $172&255;
           $178 = $173 ? $176 : $177;
           __THREW__ = 0;
           (invoke_iiii(1112,($171|0),($175|0),($178|0))|0);
           $179 = __THREW__; __THREW__ = 0;
           $180 = $179&1;
           if ($180) {
            label = 52;
            break L66;
           } else {
            $$0106 = $132;
           }
          }
          if ((label|0) == 59) {
           $152 = ___cxa_find_matching_catch_3(0|0)|0;
           $153 = tempRet0;
           ___clang_call_terminate($152);
           // unreachable;
          }
          else if ((label|0) == 69) {
           $181 = ___cxa_find_matching_catch_2()|0;
           $182 = tempRet0;
           $$0117 = $182;$$0119 = $181;
           break;
          }
          else if ((label|0) == 70) {
           $183 = HEAP32[$85>>2]|0;
           $184 = ((($183)) + -24|0);
           __THREW__ = 0;
           (invoke_iii(1119,($184|0),(159081|0))|0);
           $185 = __THREW__; __THREW__ = 0;
           $186 = $185&1;
           if ($186) {
            label = 52;
            break;
           }
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
           $$2108 = $$0106;
           break L52;
          }
          else if ((label|0) == 98) {
           __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
           $$7 = $0;
           break L1;
          }
         }
        } while(0);
        if ((label|0) == 52) {
         $136 = ___cxa_find_matching_catch_2()|0;
         $137 = tempRet0;
         $$0117 = $137;$$0119 = $136;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($11);
        $$1118 = $$0117;$$1120 = $$0119;
        break L4;
       }
      } while(0);
      $187 = ($$2108|0)==($1|0);
      if (!($187)) {
       $188 = HEAP8[$$2108>>0]|0;
       $189 = ($188<<24>>24)==(69);
       if ($189) {
        $195 = ((($$2108)) + 1|0);
        $196 = ($195|0)==($1|0);
        if ($196) {
         $197 = HEAP32[$85>>2]|0;
         $198 = ((($197)) + -24|0);
         $200 = $197;
         while(1) {
          $199 = ($200|0)==($198|0);
          if ($199) {
           $$7 = $0;
           break L1;
          }
          $201 = ((($200)) + -24|0);
          HEAP32[$85>>2] = $201;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($201);
          $$pre$i149 = HEAP32[$85>>2]|0;
          $200 = $$pre$i149;
         }
        }
        $202 = HEAP8[$195>>0]|0;
        $203 = $202 << 24 >> 24;
        $isdigittmp = (($203) + -48)|0;
        $isdigit = ($isdigittmp>>>0)<(10);
        if ($isdigit) {
         $204 = ((($$2108)) + 2|0);
         $$0105 = $204;
         while(1) {
          $205 = ($$0105|0)==($1|0);
          if ($205) {
           $$0105$lcssa = $1;
           break;
          }
          $206 = HEAP8[$$0105>>0]|0;
          $207 = $206 << 24 >> 24;
          $isdigittmp126 = (($207) + -48)|0;
          $isdigit127 = ($isdigittmp126>>>0)<(10);
          if (!($isdigit127)) {
           $$0105$lcssa = $$0105;
           break;
          }
          $208 = ((($$0105)) + 1|0);
          $$0105 = $208;
         }
         $209 = HEAP32[$85>>2]|0;
         $210 = ((($209)) + -24|0);
         $211 = ((($210)) + 11|0);
         $212 = HEAP8[$211>>0]|0;
         $213 = ($212<<24>>24)<(0);
         if ($213) {
          $214 = HEAP32[$210>>2]|0;
          $216 = $214;
         } else {
          $216 = $210;
         }
         $215 = ((($216)) + 7|0);
         $217 = $215;
         HEAP32[$13>>2] = $217;
         ;HEAP32[$$byval_copy>>2]=HEAP32[$13>>2]|0;
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertIPKcEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr38__libcpp_string_gets_noexcept_iteratorISC_EE5valueENS_11__wrap_iterIPcEEE4typeENSD_ISA_EESC_SC_($210,$$byval_copy,$195,$$0105$lcssa)|0);
         $$3109 = $$0105$lcssa;
        } else {
         $$3109 = $195;
        }
        $218 = ($$3109|0)==($1|0);
        if (!($218)) {
         $219 = HEAP8[$$3109>>0]|0;
         $220 = ($219<<24>>24)==(95);
         if ($220) {
          $226 = ((($$3109)) + 1|0);
          $$7 = $226;
          break L1;
         }
        }
        $221 = HEAP32[$85>>2]|0;
        $222 = ((($221)) + -24|0);
        $224 = $221;
        while(1) {
         $223 = ($224|0)==($222|0);
         if ($223) {
          $$7 = $0;
          break L1;
         }
         $225 = ((($224)) + -24|0);
         HEAP32[$85>>2] = $225;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($225);
         $$pre$i151 = HEAP32[$85>>2]|0;
         $224 = $$pre$i151;
        }
       }
      }
      $190 = HEAP32[$85>>2]|0;
      $191 = ((($190)) + -24|0);
      $193 = $190;
      while(1) {
       $192 = ($193|0)==($191|0);
       if ($192) {
        $$7 = $0;
        break L1;
       }
       $194 = ((($193)) + -24|0);
       HEAP32[$85>>2] = $194;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($194);
       $$pre$i147 = HEAP32[$85>>2]|0;
       $193 = $$pre$i147;
      }
      break;
     }
     default: {
      $$7 = $0;
      break L1;
     }
     }
    } while(0);
    ___resumeException($$1120|0);
    // unreachable;
   } else {
    $$7 = $0;
   }
  } else {
   $$7 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$7|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_operator_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0161 = 0, $$0162 = 0, $$3 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0;
 var $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0;
 var $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0;
 var $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0, $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0;
 var $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0, $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0;
 var $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0, $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0;
 var $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0, $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0;
 var $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0, $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0;
 var $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0, $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0;
 var $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0, $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0;
 var $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0, $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0;
 var $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0, $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0;
 var $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0, $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0;
 var $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0, $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0;
 var $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0, $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0;
 var $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0, $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0;
 var $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0, $399 = 0, $4 = 0, $40 = 0, $400 = 0, $401 = 0, $402 = 0, $403 = 0;
 var $404 = 0, $405 = 0, $406 = 0, $407 = 0, $408 = 0, $409 = 0, $41 = 0, $410 = 0, $411 = 0, $412 = 0, $413 = 0, $414 = 0, $415 = 0, $416 = 0, $417 = 0, $418 = 0, $419 = 0, $42 = 0, $420 = 0, $421 = 0;
 var $422 = 0, $423 = 0, $424 = 0, $425 = 0, $426 = 0, $427 = 0, $428 = 0, $429 = 0, $43 = 0, $430 = 0, $431 = 0, $432 = 0, $433 = 0, $434 = 0, $435 = 0, $436 = 0, $437 = 0, $438 = 0, $439 = 0, $44 = 0;
 var $440 = 0, $441 = 0, $442 = 0, $443 = 0, $444 = 0, $445 = 0, $446 = 0, $447 = 0, $448 = 0, $449 = 0, $45 = 0, $450 = 0, $451 = 0, $452 = 0, $453 = 0, $454 = 0, $455 = 0, $456 = 0, $457 = 0, $458 = 0;
 var $459 = 0, $46 = 0, $460 = 0, $461 = 0, $462 = 0, $463 = 0, $464 = 0, $465 = 0, $466 = 0, $467 = 0, $468 = 0, $469 = 0, $47 = 0, $470 = 0, $471 = 0, $472 = 0, $473 = 0, $474 = 0, $475 = 0, $476 = 0;
 var $477 = 0, $478 = 0, $479 = 0, $48 = 0, $480 = 0, $481 = 0, $482 = 0, $483 = 0, $484 = 0, $485 = 0, $486 = 0, $487 = 0, $488 = 0, $489 = 0, $49 = 0, $490 = 0, $491 = 0, $492 = 0, $493 = 0, $494 = 0;
 var $495 = 0, $496 = 0, $497 = 0, $498 = 0, $499 = 0, $5 = 0, $50 = 0, $500 = 0, $501 = 0, $502 = 0, $503 = 0, $504 = 0, $505 = 0, $506 = 0, $507 = 0, $508 = 0, $509 = 0, $51 = 0, $510 = 0, $511 = 0;
 var $512 = 0, $513 = 0, $514 = 0, $515 = 0, $516 = 0, $517 = 0, $518 = 0, $519 = 0, $52 = 0, $520 = 0, $521 = 0, $522 = 0, $523 = 0, $524 = 0, $525 = 0, $526 = 0, $527 = 0, $528 = 0, $529 = 0, $53 = 0;
 var $530 = 0, $531 = 0, $532 = 0, $533 = 0, $534 = 0, $535 = 0, $536 = 0, $537 = 0, $538 = 0, $539 = 0, $54 = 0, $540 = 0, $541 = 0, $542 = 0, $543 = 0, $544 = 0, $545 = 0, $546 = 0, $547 = 0, $548 = 0;
 var $549 = 0, $55 = 0, $550 = 0, $551 = 0, $552 = 0, $553 = 0, $554 = 0, $555 = 0, $556 = 0, $557 = 0, $558 = 0, $559 = 0, $56 = 0, $560 = 0, $561 = 0, $562 = 0, $563 = 0, $564 = 0, $565 = 0, $566 = 0;
 var $567 = 0, $568 = 0, $569 = 0, $57 = 0, $570 = 0, $571 = 0, $572 = 0, $573 = 0, $574 = 0, $575 = 0, $576 = 0, $577 = 0, $578 = 0, $579 = 0, $58 = 0, $580 = 0, $581 = 0, $582 = 0, $583 = 0, $584 = 0;
 var $585 = 0, $586 = 0, $587 = 0, $588 = 0, $589 = 0, $59 = 0, $590 = 0, $591 = 0, $592 = 0, $593 = 0, $594 = 0, $595 = 0, $596 = 0, $597 = 0, $598 = 0, $599 = 0, $6 = 0, $60 = 0, $600 = 0, $601 = 0;
 var $602 = 0, $603 = 0, $604 = 0, $605 = 0, $606 = 0, $607 = 0, $608 = 0, $609 = 0, $61 = 0, $610 = 0, $611 = 0, $612 = 0, $613 = 0, $614 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0;
 var $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0;
 var $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $isdigit = 0, $isdigittmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 1104|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(1104|0);
 $3 = sp + 1080|0;
 $4 = sp + 1056|0;
 $5 = sp + 1032|0;
 $6 = sp + 1008|0;
 $7 = sp + 984|0;
 $8 = sp + 960|0;
 $9 = sp + 936|0;
 $10 = sp + 912|0;
 $11 = sp + 888|0;
 $12 = sp + 864|0;
 $13 = sp + 840|0;
 $14 = sp + 816|0;
 $15 = sp + 792|0;
 $16 = sp + 768|0;
 $17 = sp + 744|0;
 $18 = sp + 720|0;
 $19 = sp + 696|0;
 $20 = sp + 672|0;
 $21 = sp + 648|0;
 $22 = sp + 624|0;
 $23 = sp + 600|0;
 $24 = sp + 576|0;
 $25 = sp + 552|0;
 $26 = sp + 528|0;
 $27 = sp + 504|0;
 $28 = sp + 480|0;
 $29 = sp + 456|0;
 $30 = sp + 432|0;
 $31 = sp + 408|0;
 $32 = sp + 384|0;
 $33 = sp + 360|0;
 $34 = sp + 336|0;
 $35 = sp + 312|0;
 $36 = sp + 288|0;
 $37 = sp + 264|0;
 $38 = sp + 240|0;
 $39 = sp + 216|0;
 $40 = sp + 192|0;
 $41 = sp + 168|0;
 $42 = sp + 144|0;
 $43 = sp + 120|0;
 $44 = sp + 96|0;
 $45 = sp + 72|0;
 $46 = sp + 48|0;
 $47 = sp + 24|0;
 $48 = sp;
 $49 = $1;
 $50 = $0;
 $51 = (($49) - ($50))|0;
 $52 = ($51|0)>(1);
 L1: do {
  if ($52) {
   $53 = HEAP8[$0>>0]|0;
   $54 = $53 << 24 >> 24;
   $55 = ((($0)) + 2|0);
   $56 = ((($0)) + 1|0);
   L3: do {
    switch ($54|0) {
    case 97:  {
     $57 = HEAP8[$56>>0]|0;
     $58 = $57 << 24 >> 24;
     switch ($58|0) {
     case 97:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($3,159383);
      $59 = ((($2)) + 4|0);
      $60 = HEAP32[$59>>2]|0;
      $61 = ((($2)) + 8|0);
      $62 = HEAP32[$61>>2]|0;
      $63 = ($60>>>0)<($62>>>0);
      if ($63) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($60,$3);
       $64 = HEAP32[$59>>2]|0;
       $65 = ((($64)) + 24|0);
       HEAP32[$59>>2] = $65;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($3|0));
       $66 = __THREW__; __THREW__ = 0;
       $67 = $66&1;
       if ($67) {
        $68 = ___cxa_find_matching_catch_2()|0;
        $69 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
        $$0161 = $68;$$0162 = $69;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($3);
      $$3 = $55;
      break L1;
      break;
     }
     case 110: case 100:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($4,159394);
      $70 = ((($2)) + 4|0);
      $71 = HEAP32[$70>>2]|0;
      $72 = ((($2)) + 8|0);
      $73 = HEAP32[$72>>2]|0;
      $74 = ($71>>>0)<($73>>>0);
      if ($74) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($71,$4);
       $75 = HEAP32[$70>>2]|0;
       $76 = ((($75)) + 24|0);
       HEAP32[$70>>2] = $76;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($4|0));
       $77 = __THREW__; __THREW__ = 0;
       $78 = $77&1;
       if ($78) {
        $79 = ___cxa_find_matching_catch_2()|0;
        $80 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
        $$0161 = $79;$$0162 = $80;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
      $$3 = $55;
      break L1;
      break;
     }
     case 78:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($5,159404);
      $81 = ((($2)) + 4|0);
      $82 = HEAP32[$81>>2]|0;
      $83 = ((($2)) + 8|0);
      $84 = HEAP32[$83>>2]|0;
      $85 = ($82>>>0)<($84>>>0);
      if ($85) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($82,$5);
       $86 = HEAP32[$81>>2]|0;
       $87 = ((($86)) + 24|0);
       HEAP32[$81>>2] = $87;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($5|0));
       $88 = __THREW__; __THREW__ = 0;
       $89 = $88&1;
       if ($89) {
        $90 = ___cxa_find_matching_catch_2()|0;
        $91 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
        $$0161 = $90;$$0162 = $91;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($5);
      $$3 = $55;
      break L1;
      break;
     }
     case 83:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($6,159415);
      $92 = ((($2)) + 4|0);
      $93 = HEAP32[$92>>2]|0;
      $94 = ((($2)) + 8|0);
      $95 = HEAP32[$94>>2]|0;
      $96 = ($93>>>0)<($95>>>0);
      if ($96) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($93,$6);
       $97 = HEAP32[$92>>2]|0;
       $98 = ((($97)) + 24|0);
       HEAP32[$92>>2] = $98;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($6|0));
       $99 = __THREW__; __THREW__ = 0;
       $100 = $99&1;
       if ($100) {
        $101 = ___cxa_find_matching_catch_2()|0;
        $102 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
        $$0161 = $101;$$0162 = $102;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 99:  {
     $103 = HEAP8[$56>>0]|0;
     $104 = $103 << 24 >> 24;
     switch ($104|0) {
     case 108:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($7,159425);
      $105 = ((($2)) + 4|0);
      $106 = HEAP32[$105>>2]|0;
      $107 = ((($2)) + 8|0);
      $108 = HEAP32[$107>>2]|0;
      $109 = ($106>>>0)<($108>>>0);
      if ($109) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($106,$7);
       $110 = HEAP32[$105>>2]|0;
       $111 = ((($110)) + 24|0);
       HEAP32[$105>>2] = $111;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($7|0));
       $112 = __THREW__; __THREW__ = 0;
       $113 = $112&1;
       if ($113) {
        $114 = ___cxa_find_matching_catch_2()|0;
        $115 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
        $$0161 = $114;$$0162 = $115;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($7);
      $$3 = $55;
      break L1;
      break;
     }
     case 109:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($8,159436);
      $116 = ((($2)) + 4|0);
      $117 = HEAP32[$116>>2]|0;
      $118 = ((($2)) + 8|0);
      $119 = HEAP32[$118>>2]|0;
      $120 = ($117>>>0)<($119>>>0);
      if ($120) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($117,$8);
       $121 = HEAP32[$116>>2]|0;
       $122 = ((($121)) + 24|0);
       HEAP32[$116>>2] = $122;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($8|0));
       $123 = __THREW__; __THREW__ = 0;
       $124 = $123&1;
       if ($124) {
        $125 = ___cxa_find_matching_catch_2()|0;
        $126 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
        $$0161 = $125;$$0162 = $126;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($8);
      $$3 = $55;
      break L1;
      break;
     }
     case 111:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($9,159446);
      $127 = ((($2)) + 4|0);
      $128 = HEAP32[$127>>2]|0;
      $129 = ((($2)) + 8|0);
      $130 = HEAP32[$129>>2]|0;
      $131 = ($128>>>0)<($130>>>0);
      if ($131) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($128,$9);
       $132 = HEAP32[$127>>2]|0;
       $133 = ((($132)) + 24|0);
       HEAP32[$127>>2] = $133;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($9|0));
       $134 = __THREW__; __THREW__ = 0;
       $135 = $134&1;
       if ($135) {
        $136 = ___cxa_find_matching_catch_2()|0;
        $137 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
        $$0161 = $136;$$0162 = $137;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($9);
      $$3 = $55;
      break L1;
      break;
     }
     case 118:  {
      $138 = ((($2)) + 63|0);
      $139 = HEAP8[$138>>0]|0;
      HEAP8[$138>>0] = 0;
      $140 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($55,$1,$2)|0);
      HEAP8[$138>>0] = $139;
      $141 = ($140|0)==($55|0);
      if ($141) {
       $$3 = $0;
       break L1;
      }
      $142 = HEAP32[$2>>2]|0;
      $143 = ((($2)) + 4|0);
      $144 = HEAP32[$143>>2]|0;
      $145 = ($142|0)==($144|0);
      if ($145) {
       $$3 = $0;
       break L1;
      }
      $146 = ((($144)) + -24|0);
      (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($146,0,159456)|0);
      $147 = ((($2)) + 60|0);
      HEAP8[$147>>0] = 1;
      $$3 = $140;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 100:  {
     $148 = HEAP8[$56>>0]|0;
     $149 = $148 << 24 >> 24;
     switch ($149|0) {
     case 97:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj18EEERAT__Kc($10,159466);
      $150 = ((($2)) + 4|0);
      $151 = HEAP32[$150>>2]|0;
      $152 = ((($2)) + 8|0);
      $153 = HEAP32[$152>>2]|0;
      $154 = ($151>>>0)<($153>>>0);
      if ($154) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($151,$10);
       $155 = HEAP32[$150>>2]|0;
       $156 = ((($155)) + 24|0);
       HEAP32[$150>>2] = $156;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($10|0));
       $157 = __THREW__; __THREW__ = 0;
       $158 = $157&1;
       if ($158) {
        $159 = ___cxa_find_matching_catch_2()|0;
        $160 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
        $$0161 = $159;$$0162 = $160;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($10);
      $$3 = $55;
      break L1;
      break;
     }
     case 101:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($11,159484);
      $161 = ((($2)) + 4|0);
      $162 = HEAP32[$161>>2]|0;
      $163 = ((($2)) + 8|0);
      $164 = HEAP32[$163>>2]|0;
      $165 = ($162>>>0)<($164>>>0);
      if ($165) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($162,$11);
       $166 = HEAP32[$161>>2]|0;
       $167 = ((($166)) + 24|0);
       HEAP32[$161>>2] = $167;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($11|0));
       $168 = __THREW__; __THREW__ = 0;
       $169 = $168&1;
       if ($169) {
        $170 = ___cxa_find_matching_catch_2()|0;
        $171 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
        $$0161 = $170;$$0162 = $171;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($11);
      $$3 = $55;
      break L1;
      break;
     }
     case 108:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj16EEERAT__Kc($12);
      $172 = ((($2)) + 4|0);
      $173 = HEAP32[$172>>2]|0;
      $174 = ((($2)) + 8|0);
      $175 = HEAP32[$174>>2]|0;
      $176 = ($173>>>0)<($175>>>0);
      if ($176) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($173,$12);
       $177 = HEAP32[$172>>2]|0;
       $178 = ((($177)) + 24|0);
       HEAP32[$172>>2] = $178;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($12|0));
       $179 = __THREW__; __THREW__ = 0;
       $180 = $179&1;
       if ($180) {
        $181 = ___cxa_find_matching_catch_2()|0;
        $182 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
        $$0161 = $181;$$0162 = $182;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($12);
      $$3 = $55;
      break L1;
      break;
     }
     case 118:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($13,159494);
      $183 = ((($2)) + 4|0);
      $184 = HEAP32[$183>>2]|0;
      $185 = ((($2)) + 8|0);
      $186 = HEAP32[$185>>2]|0;
      $187 = ($184>>>0)<($186>>>0);
      if ($187) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($184,$13);
       $188 = HEAP32[$183>>2]|0;
       $189 = ((($188)) + 24|0);
       HEAP32[$183>>2] = $189;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($13|0));
       $190 = __THREW__; __THREW__ = 0;
       $191 = $190&1;
       if ($191) {
        $192 = ___cxa_find_matching_catch_2()|0;
        $193 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
        $$0161 = $192;$$0162 = $193;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($13);
      $$3 = $55;
      break L1;
      break;
     }
     case 86:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($14,159504);
      $194 = ((($2)) + 4|0);
      $195 = HEAP32[$194>>2]|0;
      $196 = ((($2)) + 8|0);
      $197 = HEAP32[$196>>2]|0;
      $198 = ($195>>>0)<($197>>>0);
      if ($198) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($195,$14);
       $199 = HEAP32[$194>>2]|0;
       $200 = ((($199)) + 24|0);
       HEAP32[$194>>2] = $200;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($14|0));
       $201 = __THREW__; __THREW__ = 0;
       $202 = $201&1;
       if ($202) {
        $203 = ___cxa_find_matching_catch_2()|0;
        $204 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
        $$0161 = $203;$$0162 = $204;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($14);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 101:  {
     $205 = HEAP8[$56>>0]|0;
     $206 = $205 << 24 >> 24;
     switch ($206|0) {
     case 111:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($15,159515);
      $207 = ((($2)) + 4|0);
      $208 = HEAP32[$207>>2]|0;
      $209 = ((($2)) + 8|0);
      $210 = HEAP32[$209>>2]|0;
      $211 = ($208>>>0)<($210>>>0);
      if ($211) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($208,$15);
       $212 = HEAP32[$207>>2]|0;
       $213 = ((($212)) + 24|0);
       HEAP32[$207>>2] = $213;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($15|0));
       $214 = __THREW__; __THREW__ = 0;
       $215 = $214&1;
       if ($215) {
        $216 = ___cxa_find_matching_catch_2()|0;
        $217 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($15);
        $$0161 = $216;$$0162 = $217;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($15);
      $$3 = $55;
      break L1;
      break;
     }
     case 79:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($16,159525);
      $218 = ((($2)) + 4|0);
      $219 = HEAP32[$218>>2]|0;
      $220 = ((($2)) + 8|0);
      $221 = HEAP32[$220>>2]|0;
      $222 = ($219>>>0)<($221>>>0);
      if ($222) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($219,$16);
       $223 = HEAP32[$218>>2]|0;
       $224 = ((($223)) + 24|0);
       HEAP32[$218>>2] = $224;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($16|0));
       $225 = __THREW__; __THREW__ = 0;
       $226 = $225&1;
       if ($226) {
        $227 = ___cxa_find_matching_catch_2()|0;
        $228 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($16);
        $$0161 = $227;$$0162 = $228;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($16);
      $$3 = $55;
      break L1;
      break;
     }
     case 113:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($17,159536);
      $229 = ((($2)) + 4|0);
      $230 = HEAP32[$229>>2]|0;
      $231 = ((($2)) + 8|0);
      $232 = HEAP32[$231>>2]|0;
      $233 = ($230>>>0)<($232>>>0);
      if ($233) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($230,$17);
       $234 = HEAP32[$229>>2]|0;
       $235 = ((($234)) + 24|0);
       HEAP32[$229>>2] = $235;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($17|0));
       $236 = __THREW__; __THREW__ = 0;
       $237 = $236&1;
       if ($237) {
        $238 = ___cxa_find_matching_catch_2()|0;
        $239 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
        $$0161 = $238;$$0162 = $239;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($17);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 103:  {
     $240 = HEAP8[$56>>0]|0;
     $241 = $240 << 24 >> 24;
     switch ($241|0) {
     case 101:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($18,159547);
      $242 = ((($2)) + 4|0);
      $243 = HEAP32[$242>>2]|0;
      $244 = ((($2)) + 8|0);
      $245 = HEAP32[$244>>2]|0;
      $246 = ($243>>>0)<($245>>>0);
      if ($246) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($243,$18);
       $247 = HEAP32[$242>>2]|0;
       $248 = ((($247)) + 24|0);
       HEAP32[$242>>2] = $248;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($18|0));
       $249 = __THREW__; __THREW__ = 0;
       $250 = $249&1;
       if ($250) {
        $251 = ___cxa_find_matching_catch_2()|0;
        $252 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($18);
        $$0161 = $251;$$0162 = $252;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($18);
      $$3 = $55;
      break L1;
      break;
     }
     case 116:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($19,159558);
      $253 = ((($2)) + 4|0);
      $254 = HEAP32[$253>>2]|0;
      $255 = ((($2)) + 8|0);
      $256 = HEAP32[$255>>2]|0;
      $257 = ($254>>>0)<($256>>>0);
      if ($257) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($254,$19);
       $258 = HEAP32[$253>>2]|0;
       $259 = ((($258)) + 24|0);
       HEAP32[$253>>2] = $259;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($19|0));
       $260 = __THREW__; __THREW__ = 0;
       $261 = $260&1;
       if ($261) {
        $262 = ___cxa_find_matching_catch_2()|0;
        $263 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
        $$0161 = $262;$$0162 = $263;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($19);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 105:  {
     $264 = HEAP8[$56>>0]|0;
     $265 = ($264<<24>>24)==(120);
     if (!($265)) {
      $$3 = $0;
      break L1;
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($20,159568);
     $266 = ((($2)) + 4|0);
     $267 = HEAP32[$266>>2]|0;
     $268 = ((($2)) + 8|0);
     $269 = HEAP32[$268>>2]|0;
     $270 = ($267>>>0)<($269>>>0);
     if ($270) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($267,$20);
      $271 = HEAP32[$266>>2]|0;
      $272 = ((($271)) + 24|0);
      HEAP32[$266>>2] = $272;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($20|0));
      $273 = __THREW__; __THREW__ = 0;
      $274 = $273&1;
      if ($274) {
       $275 = ___cxa_find_matching_catch_2()|0;
       $276 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
       $$0161 = $275;$$0162 = $276;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($20);
     $$3 = $55;
     break L1;
     break;
    }
    case 108:  {
     $277 = HEAP8[$56>>0]|0;
     $278 = $277 << 24 >> 24;
     switch ($278|0) {
     case 101:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($21,159579);
      $279 = ((($2)) + 4|0);
      $280 = HEAP32[$279>>2]|0;
      $281 = ((($2)) + 8|0);
      $282 = HEAP32[$281>>2]|0;
      $283 = ($280>>>0)<($282>>>0);
      if ($283) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($280,$21);
       $284 = HEAP32[$279>>2]|0;
       $285 = ((($284)) + 24|0);
       HEAP32[$279>>2] = $285;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($21|0));
       $286 = __THREW__; __THREW__ = 0;
       $287 = $286&1;
       if ($287) {
        $288 = ___cxa_find_matching_catch_2()|0;
        $289 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($21);
        $$0161 = $288;$$0162 = $289;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($21);
      $$3 = $55;
      break L1;
      break;
     }
     case 105:  {
      $290 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($55,$1,$2)|0);
      $291 = ($290|0)==($55|0);
      if ($291) {
       $$3 = $0;
       break L1;
      }
      $292 = HEAP32[$2>>2]|0;
      $293 = ((($2)) + 4|0);
      $294 = HEAP32[$293>>2]|0;
      $295 = ($292|0)==($294|0);
      if ($295) {
       $$3 = $0;
       break L1;
      }
      $296 = ((($294)) + -24|0);
      (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($296,0,159590)|0);
      $$3 = $290;
      break L1;
      break;
     }
     case 115:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($22,159602);
      $297 = ((($2)) + 4|0);
      $298 = HEAP32[$297>>2]|0;
      $299 = ((($2)) + 8|0);
      $300 = HEAP32[$299>>2]|0;
      $301 = ($298>>>0)<($300>>>0);
      if ($301) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($298,$22);
       $302 = HEAP32[$297>>2]|0;
       $303 = ((($302)) + 24|0);
       HEAP32[$297>>2] = $303;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($22|0));
       $304 = __THREW__; __THREW__ = 0;
       $305 = $304&1;
       if ($305) {
        $306 = ___cxa_find_matching_catch_2()|0;
        $307 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($22);
        $$0161 = $306;$$0162 = $307;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($22);
      $$3 = $55;
      break L1;
      break;
     }
     case 83:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($23,159613);
      $308 = ((($2)) + 4|0);
      $309 = HEAP32[$308>>2]|0;
      $310 = ((($2)) + 8|0);
      $311 = HEAP32[$310>>2]|0;
      $312 = ($309>>>0)<($311>>>0);
      if ($312) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($309,$23);
       $313 = HEAP32[$308>>2]|0;
       $314 = ((($313)) + 24|0);
       HEAP32[$308>>2] = $314;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($23|0));
       $315 = __THREW__; __THREW__ = 0;
       $316 = $315&1;
       if ($316) {
        $317 = ___cxa_find_matching_catch_2()|0;
        $318 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($23);
        $$0161 = $317;$$0162 = $318;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($23);
      $$3 = $55;
      break L1;
      break;
     }
     case 116:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($24,159625);
      $319 = ((($2)) + 4|0);
      $320 = HEAP32[$319>>2]|0;
      $321 = ((($2)) + 8|0);
      $322 = HEAP32[$321>>2]|0;
      $323 = ($320>>>0)<($322>>>0);
      if ($323) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($320,$24);
       $324 = HEAP32[$319>>2]|0;
       $325 = ((($324)) + 24|0);
       HEAP32[$319>>2] = $325;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($24|0));
       $326 = __THREW__; __THREW__ = 0;
       $327 = $326&1;
       if ($327) {
        $328 = ___cxa_find_matching_catch_2()|0;
        $329 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($24);
        $$0161 = $328;$$0162 = $329;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($24);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 109:  {
     $330 = HEAP8[$56>>0]|0;
     $331 = $330 << 24 >> 24;
     switch ($331|0) {
     case 105:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($25,159635);
      $332 = ((($2)) + 4|0);
      $333 = HEAP32[$332>>2]|0;
      $334 = ((($2)) + 8|0);
      $335 = HEAP32[$334>>2]|0;
      $336 = ($333>>>0)<($335>>>0);
      if ($336) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($333,$25);
       $337 = HEAP32[$332>>2]|0;
       $338 = ((($337)) + 24|0);
       HEAP32[$332>>2] = $338;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($25|0));
       $339 = __THREW__; __THREW__ = 0;
       $340 = $339&1;
       if ($340) {
        $341 = ___cxa_find_matching_catch_2()|0;
        $342 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($25);
        $$0161 = $341;$$0162 = $342;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($25);
      $$3 = $55;
      break L1;
      break;
     }
     case 73:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($26,159645);
      $343 = ((($2)) + 4|0);
      $344 = HEAP32[$343>>2]|0;
      $345 = ((($2)) + 8|0);
      $346 = HEAP32[$345>>2]|0;
      $347 = ($344>>>0)<($346>>>0);
      if ($347) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($344,$26);
       $348 = HEAP32[$343>>2]|0;
       $349 = ((($348)) + 24|0);
       HEAP32[$343>>2] = $349;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($26|0));
       $350 = __THREW__; __THREW__ = 0;
       $351 = $350&1;
       if ($351) {
        $352 = ___cxa_find_matching_catch_2()|0;
        $353 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($26);
        $$0161 = $352;$$0162 = $353;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($26);
      $$3 = $55;
      break L1;
      break;
     }
     case 108:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($27,159484);
      $354 = ((($2)) + 4|0);
      $355 = HEAP32[$354>>2]|0;
      $356 = ((($2)) + 8|0);
      $357 = HEAP32[$356>>2]|0;
      $358 = ($355>>>0)<($357>>>0);
      if ($358) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($355,$27);
       $359 = HEAP32[$354>>2]|0;
       $360 = ((($359)) + 24|0);
       HEAP32[$354>>2] = $360;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($27|0));
       $361 = __THREW__; __THREW__ = 0;
       $362 = $361&1;
       if ($362) {
        $363 = ___cxa_find_matching_catch_2()|0;
        $364 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($27);
        $$0161 = $363;$$0162 = $364;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($27);
      $$3 = $55;
      break L1;
      break;
     }
     case 76:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($28,159656);
      $365 = ((($2)) + 4|0);
      $366 = HEAP32[$365>>2]|0;
      $367 = ((($2)) + 8|0);
      $368 = HEAP32[$367>>2]|0;
      $369 = ($366>>>0)<($368>>>0);
      if ($369) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($366,$28);
       $370 = HEAP32[$365>>2]|0;
       $371 = ((($370)) + 24|0);
       HEAP32[$365>>2] = $371;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($28|0));
       $372 = __THREW__; __THREW__ = 0;
       $373 = $372&1;
       if ($373) {
        $374 = ___cxa_find_matching_catch_2()|0;
        $375 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($28);
        $$0161 = $374;$$0162 = $375;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($28);
      $$3 = $55;
      break L1;
      break;
     }
     case 109:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($29,159667);
      $376 = ((($2)) + 4|0);
      $377 = HEAP32[$376>>2]|0;
      $378 = ((($2)) + 8|0);
      $379 = HEAP32[$378>>2]|0;
      $380 = ($377>>>0)<($379>>>0);
      if ($380) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($377,$29);
       $381 = HEAP32[$376>>2]|0;
       $382 = ((($381)) + 24|0);
       HEAP32[$376>>2] = $382;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($29|0));
       $383 = __THREW__; __THREW__ = 0;
       $384 = $383&1;
       if ($384) {
        $385 = ___cxa_find_matching_catch_2()|0;
        $386 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($29);
        $$0161 = $385;$$0162 = $386;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($29);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 110:  {
     $387 = HEAP8[$56>>0]|0;
     $388 = $387 << 24 >> 24;
     switch ($388|0) {
     case 97:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj15EEERAT__Kc($30,159678);
      $389 = ((($2)) + 4|0);
      $390 = HEAP32[$389>>2]|0;
      $391 = ((($2)) + 8|0);
      $392 = HEAP32[$391>>2]|0;
      $393 = ($390>>>0)<($392>>>0);
      if ($393) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($390,$30);
       $394 = HEAP32[$389>>2]|0;
       $395 = ((($394)) + 24|0);
       HEAP32[$389>>2] = $395;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($30|0));
       $396 = __THREW__; __THREW__ = 0;
       $397 = $396&1;
       if ($397) {
        $398 = ___cxa_find_matching_catch_2()|0;
        $399 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
        $$0161 = $398;$$0162 = $399;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($30);
      $$3 = $55;
      break L1;
      break;
     }
     case 101:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($31,159693);
      $400 = ((($2)) + 4|0);
      $401 = HEAP32[$400>>2]|0;
      $402 = ((($2)) + 8|0);
      $403 = HEAP32[$402>>2]|0;
      $404 = ($401>>>0)<($403>>>0);
      if ($404) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($401,$31);
       $405 = HEAP32[$400>>2]|0;
       $406 = ((($405)) + 24|0);
       HEAP32[$400>>2] = $406;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($31|0));
       $407 = __THREW__; __THREW__ = 0;
       $408 = $407&1;
       if ($408) {
        $409 = ___cxa_find_matching_catch_2()|0;
        $410 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
        $$0161 = $409;$$0162 = $410;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($31);
      $$3 = $55;
      break L1;
      break;
     }
     case 103:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($32,159635);
      $411 = ((($2)) + 4|0);
      $412 = HEAP32[$411>>2]|0;
      $413 = ((($2)) + 8|0);
      $414 = HEAP32[$413>>2]|0;
      $415 = ($412>>>0)<($414>>>0);
      if ($415) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($412,$32);
       $416 = HEAP32[$411>>2]|0;
       $417 = ((($416)) + 24|0);
       HEAP32[$411>>2] = $417;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($32|0));
       $418 = __THREW__; __THREW__ = 0;
       $419 = $418&1;
       if ($419) {
        $420 = ___cxa_find_matching_catch_2()|0;
        $421 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($32);
        $$0161 = $420;$$0162 = $421;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($32);
      $$3 = $55;
      break L1;
      break;
     }
     case 116:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($33,159704);
      $422 = ((($2)) + 4|0);
      $423 = HEAP32[$422>>2]|0;
      $424 = ((($2)) + 8|0);
      $425 = HEAP32[$424>>2]|0;
      $426 = ($423>>>0)<($425>>>0);
      if ($426) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($423,$33);
       $427 = HEAP32[$422>>2]|0;
       $428 = ((($427)) + 24|0);
       HEAP32[$422>>2] = $428;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($33|0));
       $429 = __THREW__; __THREW__ = 0;
       $430 = $429&1;
       if ($430) {
        $431 = ___cxa_find_matching_catch_2()|0;
        $432 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($33);
        $$0161 = $431;$$0162 = $432;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($33);
      $$3 = $55;
      break L1;
      break;
     }
     case 119:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj13EEERAT__Kc($34,159714);
      $433 = ((($2)) + 4|0);
      $434 = HEAP32[$433>>2]|0;
      $435 = ((($2)) + 8|0);
      $436 = HEAP32[$435>>2]|0;
      $437 = ($434>>>0)<($436>>>0);
      if ($437) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($434,$34);
       $438 = HEAP32[$433>>2]|0;
       $439 = ((($438)) + 24|0);
       HEAP32[$433>>2] = $439;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($34|0));
       $440 = __THREW__; __THREW__ = 0;
       $441 = $440&1;
       if ($441) {
        $442 = ___cxa_find_matching_catch_2()|0;
        $443 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($34);
        $$0161 = $442;$$0162 = $443;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($34);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 111:  {
     $444 = HEAP8[$56>>0]|0;
     $445 = $444 << 24 >> 24;
     switch ($445|0) {
     case 111:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($35,159727);
      $446 = ((($2)) + 4|0);
      $447 = HEAP32[$446>>2]|0;
      $448 = ((($2)) + 8|0);
      $449 = HEAP32[$448>>2]|0;
      $450 = ($447>>>0)<($449>>>0);
      if ($450) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($447,$35);
       $451 = HEAP32[$446>>2]|0;
       $452 = ((($451)) + 24|0);
       HEAP32[$446>>2] = $452;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($35|0));
       $453 = __THREW__; __THREW__ = 0;
       $454 = $453&1;
       if ($454) {
        $455 = ___cxa_find_matching_catch_2()|0;
        $456 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($35);
        $$0161 = $455;$$0162 = $456;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($35);
      $$3 = $55;
      break L1;
      break;
     }
     case 114:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($36,159738);
      $457 = ((($2)) + 4|0);
      $458 = HEAP32[$457>>2]|0;
      $459 = ((($2)) + 8|0);
      $460 = HEAP32[$459>>2]|0;
      $461 = ($458>>>0)<($460>>>0);
      if ($461) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($458,$36);
       $462 = HEAP32[$457>>2]|0;
       $463 = ((($462)) + 24|0);
       HEAP32[$457>>2] = $463;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($36|0));
       $464 = __THREW__; __THREW__ = 0;
       $465 = $464&1;
       if ($465) {
        $466 = ___cxa_find_matching_catch_2()|0;
        $467 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
        $$0161 = $466;$$0162 = $467;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($36);
      $$3 = $55;
      break L1;
      break;
     }
     case 82:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($37,159748);
      $468 = ((($2)) + 4|0);
      $469 = HEAP32[$468>>2]|0;
      $470 = ((($2)) + 8|0);
      $471 = HEAP32[$470>>2]|0;
      $472 = ($469>>>0)<($471>>>0);
      if ($472) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($469,$37);
       $473 = HEAP32[$468>>2]|0;
       $474 = ((($473)) + 24|0);
       HEAP32[$468>>2] = $474;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($37|0));
       $475 = __THREW__; __THREW__ = 0;
       $476 = $475&1;
       if ($476) {
        $477 = ___cxa_find_matching_catch_2()|0;
        $478 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($37);
        $$0161 = $477;$$0162 = $478;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($37);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 112:  {
     $479 = HEAP8[$56>>0]|0;
     $480 = $479 << 24 >> 24;
     switch ($480|0) {
     case 109:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($38,159759);
      $481 = ((($2)) + 4|0);
      $482 = HEAP32[$481>>2]|0;
      $483 = ((($2)) + 8|0);
      $484 = HEAP32[$483>>2]|0;
      $485 = ($482>>>0)<($484>>>0);
      if ($485) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($482,$38);
       $486 = HEAP32[$481>>2]|0;
       $487 = ((($486)) + 24|0);
       HEAP32[$481>>2] = $487;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($38|0));
       $488 = __THREW__; __THREW__ = 0;
       $489 = $488&1;
       if ($489) {
        $490 = ___cxa_find_matching_catch_2()|0;
        $491 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($38);
        $$0161 = $490;$$0162 = $491;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($38);
      $$3 = $55;
      break L1;
      break;
     }
     case 108:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($39,159771);
      $492 = ((($2)) + 4|0);
      $493 = HEAP32[$492>>2]|0;
      $494 = ((($2)) + 8|0);
      $495 = HEAP32[$494>>2]|0;
      $496 = ($493>>>0)<($495>>>0);
      if ($496) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($493,$39);
       $497 = HEAP32[$492>>2]|0;
       $498 = ((($497)) + 24|0);
       HEAP32[$492>>2] = $498;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($39|0));
       $499 = __THREW__; __THREW__ = 0;
       $500 = $499&1;
       if ($500) {
        $501 = ___cxa_find_matching_catch_2()|0;
        $502 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($39);
        $$0161 = $501;$$0162 = $502;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($39);
      $$3 = $55;
      break L1;
      break;
     }
     case 76:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($40,159781);
      $503 = ((($2)) + 4|0);
      $504 = HEAP32[$503>>2]|0;
      $505 = ((($2)) + 8|0);
      $506 = HEAP32[$505>>2]|0;
      $507 = ($504>>>0)<($506>>>0);
      if ($507) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($504,$40);
       $508 = HEAP32[$503>>2]|0;
       $509 = ((($508)) + 24|0);
       HEAP32[$503>>2] = $509;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($40|0));
       $510 = __THREW__; __THREW__ = 0;
       $511 = $510&1;
       if ($511) {
        $512 = ___cxa_find_matching_catch_2()|0;
        $513 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($40);
        $$0161 = $512;$$0162 = $513;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($40);
      $$3 = $55;
      break L1;
      break;
     }
     case 112:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($41,159792);
      $514 = ((($2)) + 4|0);
      $515 = HEAP32[$514>>2]|0;
      $516 = ((($2)) + 8|0);
      $517 = HEAP32[$516>>2]|0;
      $518 = ($515>>>0)<($517>>>0);
      if ($518) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($515,$41);
       $519 = HEAP32[$514>>2]|0;
       $520 = ((($519)) + 24|0);
       HEAP32[$514>>2] = $520;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($41|0));
       $521 = __THREW__; __THREW__ = 0;
       $522 = $521&1;
       if ($522) {
        $523 = ___cxa_find_matching_catch_2()|0;
        $524 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($41);
        $$0161 = $523;$$0162 = $524;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($41);
      $$3 = $55;
      break L1;
      break;
     }
     case 115:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($42,159771);
      $525 = ((($2)) + 4|0);
      $526 = HEAP32[$525>>2]|0;
      $527 = ((($2)) + 8|0);
      $528 = HEAP32[$527>>2]|0;
      $529 = ($526>>>0)<($528>>>0);
      if ($529) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($526,$42);
       $530 = HEAP32[$525>>2]|0;
       $531 = ((($530)) + 24|0);
       HEAP32[$525>>2] = $531;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($42|0));
       $532 = __THREW__; __THREW__ = 0;
       $533 = $532&1;
       if ($533) {
        $534 = ___cxa_find_matching_catch_2()|0;
        $535 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($42);
        $$0161 = $534;$$0162 = $535;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($42);
      $$3 = $55;
      break L1;
      break;
     }
     case 116:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($43,159803);
      $536 = ((($2)) + 4|0);
      $537 = HEAP32[$536>>2]|0;
      $538 = ((($2)) + 8|0);
      $539 = HEAP32[$538>>2]|0;
      $540 = ($537>>>0)<($539>>>0);
      if ($540) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($537,$43);
       $541 = HEAP32[$536>>2]|0;
       $542 = ((($541)) + 24|0);
       HEAP32[$536>>2] = $542;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($43|0));
       $543 = __THREW__; __THREW__ = 0;
       $544 = $543&1;
       if ($544) {
        $545 = ___cxa_find_matching_catch_2()|0;
        $546 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($43);
        $$0161 = $545;$$0162 = $546;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($43);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 113:  {
     $547 = HEAP8[$56>>0]|0;
     $548 = ($547<<24>>24)==(117);
     if (!($548)) {
      $$3 = $0;
      break L1;
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($44,159814);
     $549 = ((($2)) + 4|0);
     $550 = HEAP32[$549>>2]|0;
     $551 = ((($2)) + 8|0);
     $552 = HEAP32[$551>>2]|0;
     $553 = ($550>>>0)<($552>>>0);
     if ($553) {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($550,$44);
      $554 = HEAP32[$549>>2]|0;
      $555 = ((($554)) + 24|0);
      HEAP32[$549>>2] = $555;
     } else {
      __THREW__ = 0;
      invoke_vii(1128,($2|0),($44|0));
      $556 = __THREW__; __THREW__ = 0;
      $557 = $556&1;
      if ($557) {
       $558 = ___cxa_find_matching_catch_2()|0;
       $559 = tempRet0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($44);
       $$0161 = $558;$$0162 = $559;
       break L3;
      }
     }
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($44);
     $$3 = $55;
     break L1;
     break;
    }
    case 114:  {
     $560 = HEAP8[$56>>0]|0;
     $561 = $560 << 24 >> 24;
     switch ($561|0) {
     case 109:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($45,159824);
      $562 = ((($2)) + 4|0);
      $563 = HEAP32[$562>>2]|0;
      $564 = ((($2)) + 8|0);
      $565 = HEAP32[$564>>2]|0;
      $566 = ($563>>>0)<($565>>>0);
      if ($566) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($563,$45);
       $567 = HEAP32[$562>>2]|0;
       $568 = ((($567)) + 24|0);
       HEAP32[$562>>2] = $568;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($45|0));
       $569 = __THREW__; __THREW__ = 0;
       $570 = $569&1;
       if ($570) {
        $571 = ___cxa_find_matching_catch_2()|0;
        $572 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($45);
        $$0161 = $571;$$0162 = $572;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($45);
      $$3 = $55;
      break L1;
      break;
     }
     case 77:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($46,159834);
      $573 = ((($2)) + 4|0);
      $574 = HEAP32[$573>>2]|0;
      $575 = ((($2)) + 8|0);
      $576 = HEAP32[$575>>2]|0;
      $577 = ($574>>>0)<($576>>>0);
      if ($577) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($574,$46);
       $578 = HEAP32[$573>>2]|0;
       $579 = ((($578)) + 24|0);
       HEAP32[$573>>2] = $579;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($46|0));
       $580 = __THREW__; __THREW__ = 0;
       $581 = $580&1;
       if ($581) {
        $582 = ___cxa_find_matching_catch_2()|0;
        $583 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($46);
        $$0161 = $582;$$0162 = $583;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($46);
      $$3 = $55;
      break L1;
      break;
     }
     case 115:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($47,159845);
      $584 = ((($2)) + 4|0);
      $585 = HEAP32[$584>>2]|0;
      $586 = ((($2)) + 8|0);
      $587 = HEAP32[$586>>2]|0;
      $588 = ($585>>>0)<($587>>>0);
      if ($588) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($585,$47);
       $589 = HEAP32[$584>>2]|0;
       $590 = ((($589)) + 24|0);
       HEAP32[$584>>2] = $590;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($47|0));
       $591 = __THREW__; __THREW__ = 0;
       $592 = $591&1;
       if ($592) {
        $593 = ___cxa_find_matching_catch_2()|0;
        $594 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($47);
        $$0161 = $593;$$0162 = $594;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($47);
      $$3 = $55;
      break L1;
      break;
     }
     case 83:  {
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($48,159856);
      $595 = ((($2)) + 4|0);
      $596 = HEAP32[$595>>2]|0;
      $597 = ((($2)) + 8|0);
      $598 = HEAP32[$597>>2]|0;
      $599 = ($596>>>0)<($598>>>0);
      if ($599) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($596,$48);
       $600 = HEAP32[$595>>2]|0;
       $601 = ((($600)) + 24|0);
       HEAP32[$595>>2] = $601;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($48|0));
       $602 = __THREW__; __THREW__ = 0;
       $603 = $602&1;
       if ($603) {
        $604 = ___cxa_find_matching_catch_2()|0;
        $605 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($48);
        $$0161 = $604;$$0162 = $605;
        break L3;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($48);
      $$3 = $55;
      break L1;
      break;
     }
     default: {
      $$3 = $0;
      break L1;
     }
     }
     break;
    }
    case 118:  {
     $606 = HEAP8[$56>>0]|0;
     $607 = $606 << 24 >> 24;
     $isdigittmp = (($607) + -48)|0;
     $isdigit = ($isdigittmp>>>0)<(10);
     if (!($isdigit)) {
      $$3 = $0;
      break L1;
     }
     $608 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_($55,$1,$2)|0);
     $609 = ($608|0)==($55|0);
     if ($609) {
      $$3 = $0;
      break L1;
     }
     $610 = HEAP32[$2>>2]|0;
     $611 = ((($2)) + 4|0);
     $612 = HEAP32[$611>>2]|0;
     $613 = ($610|0)==($612|0);
     if ($613) {
      $$3 = $0;
      break L1;
     }
     $614 = ((($612)) + -24|0);
     (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($614,0,159456)|0);
     $$3 = $608;
     break L1;
     break;
    }
    default: {
     $$3 = $0;
     break L1;
    }
    }
   } while(0);
   ___resumeException($$0161|0);
   // unreachable;
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj11EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 10;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,10)|0);
 $4 = ((($0)) + 10|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj10EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 9;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,9)|0);
 $4 = ((($0)) + 9|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj18EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(32)|0);
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = -2147483616;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 17;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($3,$1,17)|0);
 $6 = ((($3)) + 17|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($6,$2);
 $7 = ((($0)) + 12|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $8 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($$0$i$i) + 1)|0;
  $$0$i$i = $9;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj16EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(16)|0);
 HEAP32[$0>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = -2147483632;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 15;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($2,159868,15)|0);
 $5 = ((($2)) + 15|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($5,$1);
 $6 = ((($0)) + 12|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj12EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(16)|0);
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = -2147483632;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 11;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($3,$1,11)|0);
 $6 = ((($3)) + 11|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($6,$2);
 $7 = ((($0)) + 12|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $8 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($$0$i$i) + 1)|0;
  $$0$i$i = $9;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj15EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(16)|0);
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = -2147483632;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 14;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($3,$1,14)|0);
 $6 = ((($3)) + 14|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($6,$2);
 $7 = ((($0)) + 12|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $8 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($$0$i$i) + 1)|0;
  $$0$i$i = $9;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj13EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(16)|0);
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = -2147483632;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 12;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($3,$1,12)|0);
 $6 = ((($3)) + 12|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($6,$2);
 $7 = ((($0)) + 12|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $8 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($$0$i$i) + 1)|0;
  $$0$i$i = $9;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE23__append_forward_unsafeIPKcEERS7_T_SC_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0$i$i = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$040 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0;
 var $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0;
 var $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $8 = 0, $9 = 0, $phitmp$i = 0, $scevgep = 0, $scevgep23$i$i = 0, $scevgep5 = 0, $scevgep6 = 0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = $1;
 $4 = sp + 13|0;
 $5 = sp;
 $6 = sp + 12|0;
 $7 = ((($0)) + 11|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)<(0);
 if ($9) {
  $11 = ((($0)) + 4|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = ((($0)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = $14 & 2147483647;
  $phitmp$i = (($15) + -1)|0;
  $55 = $12;$56 = $phitmp$i;
 } else {
  $10 = $8&255;
  $55 = $10;$56 = 10;
 }
 $16 = $2;
 $17 = (($16) - ($3))|0;
 $18 = ($17|0)==(0);
 do {
  if (!($18)) {
   if ($9) {
    $19 = HEAP32[$0>>2]|0;
    $20 = ((($0)) + 4|0);
    $21 = HEAP32[$20>>2]|0;
    $24 = $19;$25 = $21;$26 = $19;
   } else {
    $22 = $8&255;
    $24 = $0;$25 = $22;$26 = $0;
   }
   $23 = (($24) + ($25)|0);
   $27 = (__ZNSt3__214__ptr_in_rangeIcEEbPKT_S3_S3_($1,$26,$23)|0);
   if ($27) {
    ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
    $28 = ($17>>>0)>(4294967279);
    if ($28) {
     __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($5);
     // unreachable;
    }
    $29 = ($17>>>0)<(11);
    if ($29) {
     $30 = $17&255;
     $31 = ((($5)) + 11|0);
     HEAP8[$31>>0] = $30;
     $$021$ph$i$i = $5;
    } else {
     $32 = (($17) + 16)|0;
     $33 = $32 & -16;
     $34 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($33)|0);
     HEAP32[$5>>2] = $34;
     $35 = $33 | -2147483648;
     $36 = ((($5)) + 8|0);
     HEAP32[$36>>2] = $35;
     $37 = ((($5)) + 4|0);
     HEAP32[$37>>2] = $17;
     $$021$ph$i$i = $34;
    }
    $$0$i$i = $1;$$021$i$i = $$021$ph$i$i;
    while(1) {
     $38 = ($$0$i$i|0)==($2|0);
     if ($38) {
      break;
     }
     __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
     $39 = ((($$0$i$i)) + 1|0);
     $40 = ((($$021$i$i)) + 1|0);
     $$0$i$i = $39;$$021$i$i = $40;
    }
    $scevgep23$i$i = (($$021$ph$i$i) + ($17)|0);
    HEAP8[$4>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$4);
    $41 = ((($5)) + 11|0);
    $42 = HEAP8[$41>>0]|0;
    $43 = ($42<<24>>24)<(0);
    $44 = HEAP32[$5>>2]|0;
    $45 = ((($5)) + 4|0);
    $46 = HEAP32[$45>>2]|0;
    $47 = $42&255;
    $48 = $43 ? $44 : $5;
    $49 = $43 ? $46 : $47;
    __THREW__ = 0;
    (invoke_iiii(1112,($0|0),($48|0),($49|0))|0);
    $50 = __THREW__; __THREW__ = 0;
    $51 = $50&1;
    if ($51) {
     $52 = ___cxa_find_matching_catch_2()|0;
     $53 = tempRet0;
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
     ___resumeException($52|0);
     // unreachable;
    } else {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
     break;
    }
   } else {
    $54 = (($56) - ($55))|0;
    $57 = ($54>>>0)<($17>>>0);
    $58 = (($55) + ($17))|0;
    if ($57) {
     $59 = (($58) - ($56))|0;
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9__grow_byEjjjjjj($0,$56,$59,$55,$55,0);
    }
    $60 = HEAP8[$7>>0]|0;
    $61 = ($60<<24>>24)<(0);
    if ($61) {
     $62 = HEAP32[$0>>2]|0;
     $64 = $62;
    } else {
     $64 = $0;
    }
    $63 = (($64) + ($55)|0);
    $65 = (($55) - ($3))|0;
    $scevgep = (($2) + ($65)|0);
    $scevgep5 = $scevgep;
    $$0 = $63;$$040 = $1;
    while(1) {
     $66 = ($$040|0)==($2|0);
     if ($66) {
      break;
     }
     __ZNSt3__211char_traitsIcE6assignERcRKc($$0,$$040);
     $67 = ((($$0)) + 1|0);
     $68 = ((($$040)) + 1|0);
     $$0 = $67;$$040 = $68;
    }
    $scevgep6 = (($64) + ($scevgep5)|0);
    HEAP8[$6>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep6,$6);
    $69 = HEAP8[$7>>0]|0;
    $70 = ($69<<24>>24)<(0);
    if ($70) {
     $71 = ((($0)) + 4|0);
     HEAP32[$71>>2] = $58;
     break;
    } else {
     $72 = $58&255;
     HEAP8[$7>>0] = $72;
     break;
    }
   }
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9push_backEc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$018 = 0, $$019 = 0, $$pn = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 1|0;
 $3 = sp;
 HEAP8[$2>>0] = $1;
 $4 = ((($0)) + 11|0);
 $5 = HEAP8[$4>>0]|0;
 $6 = ($5<<24>>24)<(0);
 if ($6) {
  $8 = ((($0)) + 8|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $9 & 2147483647;
  $11 = (($10) + -1)|0;
  $12 = ((($0)) + 4|0);
  $13 = HEAP32[$12>>2]|0;
  $$018 = $13;$$019 = $11;
 } else {
  $7 = $5&255;
  $$018 = $7;$$019 = 10;
 }
 $14 = ($$018|0)==($$019|0);
 if ($14) {
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9__grow_byEjjjjjj($0,$$019,1,$$019,$$019,0);
  $15 = HEAP8[$4>>0]|0;
  $16 = ($15<<24>>24)<(0);
  $17 = (($$019) + 1)|0;
  if ($16) {
   $23 = $17;
   label = 8;
  } else {
   $20 = $17;
   label = 7;
  }
 } else {
  $18 = (($$018) + 1)|0;
  if ($6) {
   $23 = $18;
   label = 8;
  } else {
   $20 = $18;
   label = 7;
  }
 }
 if ((label|0) == 7) {
  $19 = $20&255;
  HEAP8[$4>>0] = $19;
  $$pn = $0;
 }
 else if ((label|0) == 8) {
  $21 = HEAP32[$0>>2]|0;
  $22 = ((($0)) + 4|0);
  HEAP32[$22>>2] = $23;
  $$pn = $21;
 }
 $$0 = (($$pn) + ($$018)|0);
 __ZNSt3__211char_traitsIcE6assignERcRKc($$0,$2);
 $24 = ((($$0)) + 1|0);
 HEAP8[$3>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($24,$3);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertIPKcEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr38__libcpp_string_gets_noexcept_iteratorISC_EE5valueENS_11__wrap_iterIPcEEE4typeENSD_ISA_EESC_SC_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$0$i$i = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$059 = 0, $$1 = 0, $$byval_copy = 0, $$pre = 0, $$sroa$0$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0;
 var $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0;
 var $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0;
 var $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0;
 var $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $phitmp$i = 0, $scevgep23$i$i = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $$byval_copy = sp + 16|0;
 $4 = sp + 21|0;
 $5 = sp + 4|0;
 $6 = sp;
 $7 = sp + 20|0;
 $8 = ((($0)) + 11|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)<(0);
 if ($10) {
  $11 = HEAP32[$0>>2]|0;
  $13 = $11;
 } else {
  $13 = $0;
 }
 $12 = $13;
 $14 = HEAP32[$1>>2]|0;
 $15 = (($14) - ($12))|0;
 $16 = $3;
 $17 = $2;
 $18 = (($16) - ($17))|0;
 $19 = ($18|0)==(0);
 do {
  if ($19) {
   $88 = $9;
   label = 37;
  } else {
   if ($10) {
    $20 = HEAP32[$0>>2]|0;
    $21 = ((($0)) + 4|0);
    $22 = HEAP32[$21>>2]|0;
    $25 = $20;$26 = $22;$27 = $20;
   } else {
    $23 = $9&255;
    $25 = $0;$26 = $23;$27 = $0;
   }
   $24 = (($25) + ($26)|0);
   $28 = (__ZNSt3__214__ptr_in_rangeIcEEbPKT_S3_S3_($2,$27,$24)|0);
   if ($28) {
    ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
    $29 = ($18>>>0)>(4294967279);
    if ($29) {
     __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($5);
     // unreachable;
    }
    $30 = ($18>>>0)<(11);
    if ($30) {
     $31 = $18&255;
     $32 = ((($5)) + 11|0);
     HEAP8[$32>>0] = $31;
     $$021$ph$i$i = $5;
    } else {
     $33 = (($18) + 16)|0;
     $34 = $33 & -16;
     $35 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($34)|0);
     HEAP32[$5>>2] = $35;
     $36 = $34 | -2147483648;
     $37 = ((($5)) + 8|0);
     HEAP32[$37>>2] = $36;
     $38 = ((($5)) + 4|0);
     HEAP32[$38>>2] = $18;
     $$021$ph$i$i = $35;
    }
    $$0$i$i = $2;$$021$i$i = $$021$ph$i$i;
    while(1) {
     $39 = ($$0$i$i|0)==($3|0);
     if ($39) {
      break;
     }
     __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
     $40 = ((($$0$i$i)) + 1|0);
     $41 = ((($$021$i$i)) + 1|0);
     $$0$i$i = $40;$$021$i$i = $41;
    }
    $scevgep23$i$i = (($$021$ph$i$i) + ($18)|0);
    HEAP8[$4>>0] = 0;
    __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$4);
    HEAP32[$6>>2] = $14;
    $42 = ((($5)) + 11|0);
    $43 = HEAP8[$42>>0]|0;
    $44 = ($43<<24>>24)<(0);
    $45 = HEAP32[$5>>2]|0;
    $46 = ((($5)) + 4|0);
    $47 = HEAP32[$46>>2]|0;
    $48 = $43&255;
    $49 = $44 ? $45 : $5;
    $50 = $44 ? $47 : $48;
    $51 = (($49) + ($50)|0);
    __THREW__ = 0;
    ;HEAP32[$$byval_copy>>2]=HEAP32[$6>>2]|0;
    $52 = (invoke_iiiii(1144,($0|0),($$byval_copy|0),($49|0),($51|0))|0);
    $53 = __THREW__; __THREW__ = 0;
    $54 = $53&1;
    if ($54) {
     $55 = ___cxa_find_matching_catch_2()|0;
     $56 = tempRet0;
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
     ___resumeException($55|0);
     // unreachable;
    } else {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
     $$sroa$0$1 = $52;
     break;
    }
   }
   $57 = HEAP8[$8>>0]|0;
   $58 = ($57<<24>>24)<(0);
   if ($58) {
    $60 = ((($0)) + 4|0);
    $61 = HEAP32[$60>>2]|0;
    $62 = ((($0)) + 8|0);
    $63 = HEAP32[$62>>2]|0;
    $64 = $63 & 2147483647;
    $phitmp$i = (($64) + -1)|0;
    $66 = $61;$67 = $phitmp$i;
   } else {
    $59 = $57&255;
    $66 = $59;$67 = 10;
   }
   $65 = (($67) - ($66))|0;
   $68 = ($65>>>0)<($18>>>0);
   $69 = (($66) + ($18))|0;
   if ($68) {
    $76 = (($69) - ($67))|0;
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9__grow_byEjjjjjj($0,$67,$76,$66,$15,$18);
    $77 = HEAP32[$0>>2]|0;
    $$0 = $77;
   } else {
    if ($58) {
     $70 = HEAP32[$0>>2]|0;
     $74 = $70;
    } else {
     $74 = $0;
    }
    $71 = (($66) - ($15))|0;
    $72 = ($71|0)==(0);
    if ($72) {
     $$0 = $74;
    } else {
     $73 = (($74) + ($15)|0);
     $75 = (($73) + ($18)|0);
     (__ZNSt3__211char_traitsIcE4moveEPcPKcj($75,$73,$71)|0);
     $$0 = $74;
    }
   }
   $78 = HEAP8[$8>>0]|0;
   $79 = ($78<<24>>24)<(0);
   if ($79) {
    $80 = ((($0)) + 4|0);
    HEAP32[$80>>2] = $69;
   } else {
    $81 = $69&255;
    HEAP8[$8>>0] = $81;
   }
   $82 = (($$0) + ($69)|0);
   HEAP8[$7>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($82,$7);
   $83 = (($$0) + ($15)|0);
   $$059 = $2;$$1 = $83;
   while(1) {
    $84 = ($$059|0)==($3|0);
    if ($84) {
     break;
    }
    __ZNSt3__211char_traitsIcE6assignERcRKc($$1,$$059);
    $85 = ((($$1)) + 1|0);
    $86 = ((($$059)) + 1|0);
    $$059 = $86;$$1 = $85;
   }
   $$pre = HEAP8[$8>>0]|0;
   $88 = $$pre;
   label = 37;
  }
 } while(0);
 if ((label|0) == 37) {
  $87 = ($88<<24>>24)<(0);
  if ($87) {
   $89 = HEAP32[$0>>2]|0;
   $91 = $89;
  } else {
   $91 = $0;
  }
  $90 = (($91) + ($15)|0);
  $$sroa$0$1 = $90;
 }
 STACKTOP = sp;return ($$sroa$0$1|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9__grow_byEjjjjjj($0,$1,$2,$3,$4,$5) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 $4 = $4|0;
 $5 = $5|0;
 var $$sroa$speculated = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0;
 var $29 = 0, $30 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $6 = (-17 - ($1))|0;
 $7 = ($6>>>0)<($2>>>0);
 if ($7) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 }
 $8 = ((($0)) + 11|0);
 $9 = HEAP8[$8>>0]|0;
 $10 = ($9<<24>>24)<(0);
 if ($10) {
  $11 = HEAP32[$0>>2]|0;
  $22 = $11;
 } else {
  $22 = $0;
 }
 $12 = ($1>>>0)<(2147483623);
 if ($12) {
  $13 = (($2) + ($1))|0;
  $14 = $1 << 1;
  $15 = ($13>>>0)<($14>>>0);
  $$sroa$speculated = $15 ? $14 : $13;
  $16 = ($$sroa$speculated>>>0)<(11);
  $17 = (($$sroa$speculated) + 16)|0;
  $18 = $17 & -16;
  $phitmp = $16 ? 11 : $18;
  $19 = $phitmp;
 } else {
  $19 = -17;
 }
 $20 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($19)|0);
 $21 = ($4|0)==(0);
 if (!($21)) {
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($20,$22,$4)|0);
 }
 $23 = (($3) - ($4))|0;
 $24 = ($23|0)==(0);
 if (!($24)) {
  $25 = (($20) + ($4)|0);
  $26 = (($25) + ($5)|0);
  $27 = (($22) + ($4)|0);
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($26,$27,$23)|0);
 }
 $28 = ($1|0)==(10);
 if (!($28)) {
  __ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE10deallocateEPcj($22);
 }
 HEAP32[$0>>2] = $20;
 $29 = $19 | -2147483648;
 $30 = ((($0)) + 8|0);
 HEAP32[$30>>2] = $29;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_19base_nameINSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEEEET_RS9_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $$0$i$i = 0, $$0$i$i35 = 0, $$0$i$i38 = 0, $$0$i$i42 = 0, $$0$i$i46 = 0, $$0$i$i48 = 0, $$021$i$i = 0, $$021$ph$i$i = 0, $$032$ph = 0, $$033 = 0, $$033$ph = 0, $$1 = 0, $$2 = 0, $$2$pn = 0, $$pre = 0, $$pre57 = 0, $10 = 0, $100 = 0, $101 = 0;
 var $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0;
 var $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0;
 var $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $16 = 0, $17 = 0, $18 = 0;
 var $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0;
 var $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0;
 var $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0;
 var $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0;
 var $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i = 0, $scevgep23$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp + 4|0;
 $3 = sp + 3|0;
 $4 = sp + 2|0;
 $5 = sp + 1|0;
 $6 = sp;
 $7 = ((($1)) + 11|0);
 $8 = HEAP8[$7>>0]|0;
 $9 = ($8<<24>>24)<(0);
 $10 = ((($1)) + 4|0);
 $11 = HEAP32[$10>>2]|0;
 $12 = $8&255;
 $13 = $9 ? $11 : $12;
 $14 = ($13|0)==(0);
 L1: do {
  if ($14) {
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_($0,$1);
  } else {
   $15 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159903)|0);
   $16 = HEAP8[$7>>0]|0;
   $17 = ($16<<24>>24)<(0);
   $18 = HEAP32[$10>>2]|0;
   $19 = $16&255;
   $20 = $17 ? $18 : $19;
   $21 = ($15|0)==($20|0);
   if ($21) {
    __THREW__ = 0;
    $22 = (invoke_iiii(1123,($1|0),(159903|0),($15|0))|0);
    $23 = __THREW__; __THREW__ = 0;
    $24 = $23&1;
    if ($24) {
     $25 = ___cxa_find_matching_catch_3(0|0)|0;
     $26 = tempRet0;
     ___clang_call_terminate($25);
     // unreachable;
    }
    $27 = ($22|0)==(0);
    if ($27) {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($1,159915);
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
     $28 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159986)|0);
     $29 = ($28>>>0)>(4294967279);
     if ($29) {
      __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
      // unreachable;
     }
     $30 = ($28>>>0)<(11);
     if ($30) {
      $31 = $28&255;
      $32 = ((($0)) + 11|0);
      HEAP8[$32>>0] = $31;
      $$0$i$i35 = $0;
     } else {
      $33 = (($28) + 16)|0;
      $34 = $33 & -16;
      $35 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($34)|0);
      HEAP32[$0>>2] = $35;
      $36 = $34 | -2147483648;
      $37 = ((($0)) + 8|0);
      HEAP32[$37>>2] = $36;
      $38 = ((($0)) + 4|0);
      HEAP32[$38>>2] = $28;
      $$0$i$i35 = $35;
     }
     (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i35,159986,$28)|0);
     $39 = (($$0$i$i35) + ($28)|0);
     HEAP8[$5>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($39,$5);
     break;
    }
   }
   $40 = (__ZNSt3__211char_traitsIcE6lengthEPKc(159999)|0);
   $41 = HEAP8[$7>>0]|0;
   $42 = ($41<<24>>24)<(0);
   $43 = HEAP32[$10>>2]|0;
   $44 = $41&255;
   $45 = $42 ? $43 : $44;
   $46 = ($40|0)==($45|0);
   if ($46) {
    __THREW__ = 0;
    $47 = (invoke_iiii(1123,($1|0),(159999|0),($40|0))|0);
    $48 = __THREW__; __THREW__ = 0;
    $49 = $48&1;
    if ($49) {
     $50 = ___cxa_find_matching_catch_3(0|0)|0;
     $51 = tempRet0;
     ___clang_call_terminate($50);
     // unreachable;
    }
    $52 = ($47|0)==(0);
    if ($52) {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($1,160012);
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
     $53 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160062)|0);
     $54 = ($53>>>0)>(4294967279);
     if ($54) {
      __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
      // unreachable;
     }
     $55 = ($53>>>0)<(11);
     if ($55) {
      $56 = $53&255;
      $57 = ((($0)) + 11|0);
      HEAP8[$57>>0] = $56;
      $$0$i$i38 = $0;
     } else {
      $58 = (($53) + 16)|0;
      $59 = $58 & -16;
      $60 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($59)|0);
      HEAP32[$0>>2] = $60;
      $61 = $59 | -2147483648;
      $62 = ((($0)) + 8|0);
      HEAP32[$62>>2] = $61;
      $63 = ((($0)) + 4|0);
      HEAP32[$63>>2] = $53;
      $$0$i$i38 = $60;
     }
     (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i38,160062,$53)|0);
     $64 = (($$0$i$i38) + ($53)|0);
     HEAP8[$4>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($64,$4);
     break;
    }
   }
   $65 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160076)|0);
   $66 = HEAP8[$7>>0]|0;
   $67 = ($66<<24>>24)<(0);
   $68 = HEAP32[$10>>2]|0;
   $69 = $66&255;
   $70 = $67 ? $68 : $69;
   $71 = ($65|0)==($70|0);
   if ($71) {
    __THREW__ = 0;
    $72 = (invoke_iiii(1123,($1|0),(160076|0),($65|0))|0);
    $73 = __THREW__; __THREW__ = 0;
    $74 = $73&1;
    if ($74) {
     $75 = ___cxa_find_matching_catch_3(0|0)|0;
     $76 = tempRet0;
     ___clang_call_terminate($75);
     // unreachable;
    }
    $77 = ($72|0)==(0);
    if ($77) {
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($1,160089);
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
     $78 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160139)|0);
     $79 = ($78>>>0)>(4294967279);
     if ($79) {
      __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
      // unreachable;
     }
     $80 = ($78>>>0)<(11);
     if ($80) {
      $81 = $78&255;
      $82 = ((($0)) + 11|0);
      HEAP8[$82>>0] = $81;
      $$0$i$i42 = $0;
     } else {
      $83 = (($78) + 16)|0;
      $84 = $83 & -16;
      $85 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($84)|0);
      HEAP32[$0>>2] = $85;
      $86 = $84 | -2147483648;
      $87 = ((($0)) + 8|0);
      HEAP32[$87>>2] = $86;
      $88 = ((($0)) + 4|0);
      HEAP32[$88>>2] = $78;
      $$0$i$i42 = $85;
     }
     (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i42,160139,$78)|0);
     $89 = (($$0$i$i42) + ($78)|0);
     HEAP8[$3>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($89,$3);
     break;
    }
   }
   $90 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160153)|0);
   $91 = HEAP8[$7>>0]|0;
   $92 = ($91<<24>>24)<(0);
   $93 = HEAP32[$10>>2]|0;
   $94 = $91&255;
   $95 = $92 ? $93 : $94;
   $96 = ($90|0)==($95|0);
   do {
    if ($96) {
     __THREW__ = 0;
     $97 = (invoke_iiii(1123,($1|0),(160153|0),($90|0))|0);
     $98 = __THREW__; __THREW__ = 0;
     $99 = $98&1;
     if ($99) {
      $100 = ___cxa_find_matching_catch_3(0|0)|0;
      $101 = tempRet0;
      ___clang_call_terminate($100);
      // unreachable;
     }
     $102 = ($97|0)==(0);
     if (!($102)) {
      $$pre = HEAP8[$7>>0]|0;
      $$pre57 = HEAP32[$10>>2]|0;
      $116 = $$pre;$121 = $$pre57;
      break;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($1,160167);
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
     $103 = (__ZNSt3__211char_traitsIcE6lengthEPKc(160218)|0);
     $104 = ($103>>>0)>(4294967279);
     if ($104) {
      __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
      // unreachable;
     }
     $105 = ($103>>>0)<(11);
     if ($105) {
      $106 = $103&255;
      $107 = ((($0)) + 11|0);
      HEAP8[$107>>0] = $106;
      $$0$i$i46 = $0;
     } else {
      $108 = (($103) + 16)|0;
      $109 = $108 & -16;
      $110 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($109)|0);
      HEAP32[$0>>2] = $110;
      $111 = $109 | -2147483648;
      $112 = ((($0)) + 8|0);
      HEAP32[$112>>2] = $111;
      $113 = ((($0)) + 4|0);
      HEAP32[$113>>2] = $103;
      $$0$i$i46 = $110;
     }
     (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i46,160218,$103)|0);
     $114 = (($$0$i$i46) + ($103)|0);
     HEAP8[$2>>0] = 0;
     __ZNSt3__211char_traitsIcE6assignERcRKc($114,$2);
     break L1;
    } else {
     $116 = $91;$121 = $93;
    }
   } while(0);
   $115 = ($116<<24>>24)<(0);
   $117 = HEAP32[$1>>2]|0;
   $118 = $115 ? $117 : $1;
   $119 = $116&255;
   $120 = $115 ? $121 : $119;
   $122 = (($118) + ($120)|0);
   $123 = ((($122)) + -1|0);
   $124 = HEAP8[$123>>0]|0;
   $125 = ($124<<24>>24)==(62);
   L62: do {
    if ($125) {
     $$032$ph = 1;$$033$ph = $122;
     L63: while(1) {
      $$033 = $$033$ph;
      L65: while(1) {
       $126 = ((($$033)) + -1|0);
       $127 = ($126|0)==($118|0);
       if ($127) {
        break L63;
       }
       $130 = ((($$033)) + -2|0);
       $131 = HEAP8[$130>>0]|0;
       switch ($131<<24>>24) {
       case 60:  {
        label = 51;
        break L65;
        break;
       }
       case 62:  {
        label = 52;
        break L65;
        break;
       }
       default: {
        $$033 = $126;
       }
       }
      }
      if ((label|0) == 51) {
       label = 0;
       $132 = (($$032$ph) + -1)|0;
       $133 = ($132|0)==(0);
       if ($133) {
        $$2 = $130;
        break L62;
       } else {
        $$032$ph = $132;$$033$ph = $126;
        continue;
       }
      }
      else if ((label|0) == 52) {
       label = 0;
       $134 = (($$032$ph) + 1)|0;
       $$032$ph = $134;$$033$ph = $126;
       continue;
      }
     }
     ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
     $$0$i$i48 = 0;
     while(1) {
      $exitcond$i$i = ($$0$i$i48|0)==(3);
      if ($exitcond$i$i) {
       break L1;
      }
      $128 = (($0) + ($$0$i$i48<<2)|0);
      HEAP32[$128>>2] = 0;
      $129 = (($$0$i$i48) + 1)|0;
      $$0$i$i48 = $129;
     }
    } else {
     $$2 = $122;
    }
   } while(0);
   $$2$pn = $$2;
   while(1) {
    $$0 = ((($$2$pn)) + -1|0);
    $135 = ($$0|0)==($118|0);
    if ($135) {
     $$1 = $118;
     break;
    }
    $136 = HEAP8[$$0>>0]|0;
    $137 = ($136<<24>>24)==(58);
    if ($137) {
     $$1 = $$2$pn;
     break;
    } else {
     $$2$pn = $$0;
    }
   }
   ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
   $138 = $$1;
   $139 = $$2;
   $140 = (($139) - ($138))|0;
   $141 = ($140>>>0)>(4294967279);
   if ($141) {
    __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
    // unreachable;
   }
   $142 = ($140>>>0)<(11);
   if ($142) {
    $143 = $140&255;
    $144 = ((($0)) + 11|0);
    HEAP8[$144>>0] = $143;
    $$021$ph$i$i = $0;
   } else {
    $145 = (($140) + 16)|0;
    $146 = $145 & -16;
    $147 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($146)|0);
    HEAP32[$0>>2] = $147;
    $148 = $146 | -2147483648;
    $149 = ((($0)) + 8|0);
    HEAP32[$149>>2] = $148;
    $150 = ((($0)) + 4|0);
    HEAP32[$150>>2] = $140;
    $$021$ph$i$i = $147;
   }
   $$0$i$i = $$1;$$021$i$i = $$021$ph$i$i;
   while(1) {
    $151 = ($$0$i$i|0)==($$2|0);
    if ($151) {
     break;
    }
    __ZNSt3__211char_traitsIcE6assignERcRKc($$021$i$i,$$0$i$i);
    $152 = ((($$0$i$i)) + 1|0);
    $153 = ((($$021$i$i)) + 1|0);
    $$0$i$i = $152;$$021$i$i = $153;
   }
   $scevgep23$i$i = (($$021$ph$i$i) + ($140)|0);
   HEAP8[$6>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($scevgep23$i$i,$6);
  }
 } while(0);
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0;
 var sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($1)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 if ($5) {
  $6 = HEAP32[$1>>2]|0;
  $7 = ((($1)) + 4|0);
  $8 = HEAP32[$7>>2]|0;
  $9 = ($8>>>0)>(4294967279);
  if ($9) {
   __ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv($0);
   // unreachable;
  }
  $10 = ($8>>>0)<(11);
  if ($10) {
   $11 = $8&255;
   $12 = ((($0)) + 11|0);
   HEAP8[$12>>0] = $11;
   $$0$i = $0;
  } else {
   $13 = (($8) + 16)|0;
   $14 = $13 & -16;
   $15 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($14)|0);
   HEAP32[$0>>2] = $15;
   $16 = $14 | -2147483648;
   $17 = ((($0)) + 8|0);
   HEAP32[$17>>2] = $16;
   $18 = ((($0)) + 4|0);
   HEAP32[$18>>2] = $8;
   $$0$i = $15;
  }
  (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i,$6,$8)|0);
  $19 = (($$0$i) + ($8)|0);
  HEAP8[$2>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($19,$2);
 } else {
  ;HEAP32[$0>>2]=HEAP32[$1>>2]|0;HEAP32[$0+4>>2]=HEAP32[$1+4>>2]|0;HEAP32[$0+8>>2]=HEAP32[$1+8>>2]|0;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_destructor_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$019 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==($1|0);
 if ($3) {
  $$1 = $0;
 } else {
  $4 = (__ZN10__cxxabiv112_GLOBAL__N_121parse_unresolved_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
  $5 = ($4|0)==($0|0);
  if ($5) {
   $6 = (__ZN10__cxxabiv112_GLOBAL__N_115parse_simple_idINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
   $$019 = $6;
  } else {
   $$019 = $4;
  }
  $7 = ($$019|0)==($0|0);
  if ($7) {
   $$1 = $0;
  } else {
   $8 = HEAP32[$2>>2]|0;
   $9 = ((($2)) + 4|0);
   $10 = HEAP32[$9>>2]|0;
   $11 = ($8|0)==($10|0);
   if ($11) {
    $$1 = $0;
   } else {
    $12 = ((($10)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($12,0,159153)|0);
    $$1 = $$019;
   }
  }
 }
 return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2Ev($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $$0$i$i1 = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $exitcond$i$i = 0, $exitcond$i$i2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $1 = (($0) + ($$0$i$i<<2)|0);
  HEAP32[$1>>2] = 0;
  $2 = (($$0$i$i) + 1)|0;
  $$0$i$i = $2;
 }
 $3 = ((($0)) + 12|0);
 ;HEAP32[$3>>2]=0|0;HEAP32[$3+4>>2]=0|0;HEAP32[$3+8>>2]=0|0;
 $$0$i$i1 = 0;
 while(1) {
  $exitcond$i$i2 = ($$0$i$i1|0)==(3);
  if ($exitcond$i$i2) {
   break;
  }
  $4 = (($3) + ($$0$i$i1<<2)|0);
  HEAP32[$4>>2] = 0;
  $5 = (($$0$i$i1) + 1)|0;
  $$0$i$i1 = $5;
 }
 return;
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE24__emplace_back_slow_pathIJEEEvDpOT_($0) {
 $0 = $0|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $1 = sp;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$2>>2]|0;
 $4 = HEAP32[$0>>2]|0;
 $5 = (($3) - ($4))|0;
 $6 = (($5|0) / 24)&-1;
 $7 = (($6) + 1)|0;
 $8 = (__ZNKSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8max_sizeEv($0)|0);
 $9 = ($8>>>0)<($7>>>0);
 if ($9) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $10 = ((($0)) + 12|0);
  $11 = ((($0)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = HEAP32[$0>>2]|0;
  $14 = (($12) - ($13))|0;
  $15 = (($14|0) / 24)&-1;
  $16 = $8 >>> 1;
  $17 = ($15>>>0)<($16>>>0);
  $18 = $15 << 1;
  $19 = ($18>>>0)<($7>>>0);
  $$sroa$speculated$i = $19 ? $7 : $18;
  $$sroa$speculated$$i = $17 ? $$sroa$speculated$i : $8;
  $20 = HEAP32[$2>>2]|0;
  $21 = (($20) - ($13))|0;
  $22 = (($21|0) / 24)&-1;
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEEC2EjjS6_($1,$$sroa$speculated$$i,$22,$10);
  $23 = ((($1)) + 8|0);
  $24 = HEAP32[$23>>2]|0;
  ;HEAP32[$24>>2]=0|0;HEAP32[$24+4>>2]=0|0;HEAP32[$24+8>>2]=0|0;HEAP32[$24+12>>2]=0|0;HEAP32[$24+16>>2]=0|0;HEAP32[$24+20>>2]=0|0;
  __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2Ev($24);
  $25 = ((($24)) + 24|0);
  HEAP32[$23>>2] = $25;
  __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($0,$1);
  __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEED2Ev($1);
  STACKTOP = sp;return;
 }
}
function __ZNKSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8max_sizeEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 178956970;
}
function __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEEC2EjjS6_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = $3;
 $6 = ($1|0)==(0);
 if ($6) {
  $8 = 0;
 } else {
  $7 = (__ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE8allocateEj($3,$1)|0);
  $8 = $7;
 }
 HEAP32[$0>>2] = $8;
 $9 = (($8) + (($2*24)|0)|0);
 $10 = ((($0)) + 8|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($0)) + 4|0);
 HEAP32[$11>>2] = $9;
 $12 = (($8) + (($1*24)|0)|0);
 $13 = ((($0)) + 12|0);
 HEAP32[$13>>2] = $12;
 return;
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 4|0);
 $$0$i = $4;
 while(1) {
  $6 = ($$0$i|0)==($2|0);
  if ($6) {
   break;
  }
  $7 = HEAP32[$5>>2]|0;
  $8 = ((($7)) + -24|0);
  $9 = ((($$0$i)) + -24|0);
  __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($8,$9);
  $10 = HEAP32[$5>>2]|0;
  $11 = ((($10)) + -24|0);
  HEAP32[$5>>2] = $11;
  $$0$i = $9;
 }
 $12 = HEAP32[$0>>2]|0;
 $13 = HEAP32[$5>>2]|0;
 HEAP32[$0>>2] = $13;
 HEAP32[$5>>2] = $12;
 $14 = ((($1)) + 8|0);
 $15 = HEAP32[$3>>2]|0;
 $16 = HEAP32[$14>>2]|0;
 HEAP32[$3>>2] = $16;
 HEAP32[$14>>2] = $15;
 $17 = ((($0)) + 8|0);
 $18 = ((($1)) + 12|0);
 $19 = HEAP32[$17>>2]|0;
 $20 = HEAP32[$18>>2]|0;
 HEAP32[$17>>2] = $20;
 HEAP32[$18>>2] = $19;
 $21 = HEAP32[$5>>2]|0;
 HEAP32[$1>>2] = $21;
 return;
}
function __ZNSt3__214__split_bufferIN10__cxxabiv112_GLOBAL__N_111string_pairERNS2_11short_allocIS3_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 8|0);
 while(1) {
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==($2|0);
  if ($5) {
   break;
  }
  $6 = ((($4)) + -24|0);
  HEAP32[$3>>2] = $6;
  __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
 }
 $7 = HEAP32[$0>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  $9 = $7;
  $10 = ((($0)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($0)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) - ($9))|0;
  $15 = (($14|0) / 24)&-1;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE10deallocateEPS2_j($11,$7,$15);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE8allocateEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ($1*24)|0;
 $4 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8allocateEj($2,$3)|0);
 return ($4|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8allocateEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0 = 0, $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8align_upEj($1)|0);
 $3 = ((($0)) + 4096|0);
 $4 = ((($0)) + 4096|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = $3;
 $7 = $5;
 $8 = (($6) - ($7))|0;
 $9 = ($8>>>0)<($2>>>0);
 if ($9) {
  $11 = (_malloc($2)|0);
  $$0 = $11;
 } else {
  $10 = (($5) + ($2)|0);
  HEAP32[$4>>2] = $10;
  $$0 = $5;
 }
 return ($$0|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_integer_literalINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$0$i$i$i = 0, $$024 = 0, $$025 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0;
 var $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0;
 var $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0;
 var $62 = 0, $63 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i$i = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 48|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(48|0);
 $4 = sp + 24|0;
 $5 = sp + 12|0;
 $6 = sp;
 $7 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($0,$1)|0);
 $8 = ($7|0)==($0|0);
 $9 = ($7|0)==($1|0);
 $or$cond = $8 | $9;
 if ($or$cond) {
  $$1 = $0;
 } else {
  $10 = HEAP8[$7>>0]|0;
  $11 = ($10<<24>>24)==(69);
  if ($11) {
   $12 = ((($2)) + 11|0);
   $13 = HEAP8[$12>>0]|0;
   $14 = ($13<<24>>24)<(0);
   $15 = ((($2)) + 4|0);
   $16 = HEAP32[$15>>2]|0;
   $17 = $13&255;
   $18 = $14 ? $16 : $17;
   $19 = ($18>>>0)>(3);
   L4: do {
    if ($19) {
     __ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EEPKS8_RKSB_($6,159083,$2);
     __THREW__ = 0;
     $20 = (invoke_iii(1119,($6|0),(159081|0))|0);
     $21 = __THREW__; __THREW__ = 0;
     $22 = $21&1;
     do {
      if ($22) {
       $34 = ___cxa_find_matching_catch_2()|0;
       $35 = tempRet0;
       $$0 = $35;$$024 = $34;
      } else {
       ;HEAP32[$5>>2]=HEAP32[$6>>2]|0;HEAP32[$5+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$5+8>>2]=HEAP32[$6+8>>2]|0;
       $$0$i$i$i = 0;
       while(1) {
        $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
        if ($exitcond$i$i$i) {
         break;
        }
        $23 = (($6) + ($$0$i$i$i<<2)|0);
        HEAP32[$23>>2] = 0;
        $24 = (($$0$i$i$i) + 1)|0;
        $$0$i$i$i = $24;
       }
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($4,$5);
       $25 = ((($3)) + 4|0);
       $26 = HEAP32[$25>>2]|0;
       $27 = ((($3)) + 8|0);
       $28 = HEAP32[$27>>2]|0;
       $29 = ($26>>>0)<($28>>>0);
       if ($29) {
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($26,$4);
        $30 = HEAP32[$25>>2]|0;
        $31 = ((($30)) + 24|0);
        HEAP32[$25>>2] = $31;
       } else {
        __THREW__ = 0;
        invoke_vii(1128,($3|0),($4|0));
        $32 = __THREW__; __THREW__ = 0;
        $33 = $32&1;
        if ($33) {
         $36 = ___cxa_find_matching_catch_2()|0;
         $37 = tempRet0;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
         $$0 = $37;$$024 = $36;
         break;
        }
       }
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($4);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
       break L4;
      }
     } while(0);
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
     ___resumeException($$024|0);
     // unreachable;
    } else {
     $38 = ((($3)) + 4|0);
     $39 = HEAP32[$38>>2]|0;
     $40 = ((($3)) + 8|0);
     $41 = HEAP32[$40>>2]|0;
     $42 = ($39>>>0)<($41>>>0);
     if ($42) {
      ;HEAP32[$39>>2]=0|0;HEAP32[$39+4>>2]=0|0;HEAP32[$39+8>>2]=0|0;HEAP32[$39+12>>2]=0|0;HEAP32[$39+16>>2]=0|0;HEAP32[$39+20>>2]=0|0;
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2Ev($39);
      $43 = HEAP32[$38>>2]|0;
      $44 = ((($43)) + 24|0);
      HEAP32[$38>>2] = $44;
      break;
     } else {
      __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE24__emplace_back_slow_pathIJEEEvDpOT_($3);
      break;
     }
    }
   } while(0);
   $45 = HEAP8[$0>>0]|0;
   $46 = ($45<<24>>24)==(110);
   $47 = ((($3)) + 4|0);
   if ($46) {
    $48 = HEAP32[$47>>2]|0;
    $49 = ((($48)) + -24|0);
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9push_backEc($49,45);
    $50 = ((($0)) + 1|0);
    $$025 = $50;
   } else {
    $$025 = $0;
   }
   $51 = HEAP32[$47>>2]|0;
   $52 = ((($51)) + -24|0);
   __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE23__append_forward_unsafeIPKcEERS7_T_SC_($52,$$025,$7);
   $53 = HEAP8[$12>>0]|0;
   $54 = ($53<<24>>24)<(0);
   $55 = HEAP32[$15>>2]|0;
   $56 = $53&255;
   $57 = $54 ? $55 : $56;
   $58 = ($57>>>0)<(4);
   if ($58) {
    $59 = HEAP32[$47>>2]|0;
    $60 = ((($59)) + -24|0);
    $61 = HEAP32[$2>>2]|0;
    $62 = $54 ? $61 : $2;
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj($60,$62,$57)|0);
   }
   $63 = ((($7)) + 1|0);
   $$1 = $63;
  } else {
   $$1 = $0;
  }
 }
 STACKTOP = sp;return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj5EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 4;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,4)|0);
 $4 = ((($0)) + 4|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIfNS0_2DbEEEPKcS4_S4_RT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i57 = 0, $$049$add = 0, $$049$idx = 0, $$049$ptr = 0, $$050 = 0, $$07$i$i = 0, $$247$ph = 0, $$3 = 0, $$add = 0, $$ptr = 0, $$v = 0, $$v56 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0.0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit55 = 0, $isdigittmp = 0, $isdigittmp54 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer = sp;
 $3 = sp + 80|0;
 $4 = sp + 48|0;
 $5 = sp + 56|0;
 $6 = sp + 24|0;
 $7 = sp + 8|0;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10>>>0)>(8);
 L1: do {
  if ($11) {
   $$049$idx = 0;$$050 = $4;
   while(1) {
    $$049$ptr = (($0) + ($$049$idx)|0);
    $12 = ($$049$idx|0)==(8);
    $13 = HEAP8[$$049$ptr>>0]|0;
    $14 = $13 << 24 >> 24;
    $$049$add = $$049$idx | 1;
    $$ptr = (($0) + ($$049$add)|0);
    if ($12) {
     label = 6;
     break;
    }
    $15 = (_isxdigit($14)|0);
    $16 = ($15|0)==(0);
    if ($16) {
     break;
    }
    $isdigittmp = (($14) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    $$v = $isdigit ? 0 : 9;
    $17 = (($$v) + ($14))|0;
    $18 = HEAP8[$$ptr>>0]|0;
    $19 = $18 << 24 >> 24;
    $isdigittmp54 = (($19) + -48)|0;
    $isdigit55 = ($isdigittmp54>>>0)<(10);
    $$v56 = $isdigit55 ? 208 : 169;
    $20 = $17 << 4;
    $21 = (($19) + ($20))|0;
    $22 = (($21) + ($$v56))|0;
    $23 = $22&255;
    HEAP8[$$050>>0] = $23;
    $$add = (($$049$idx) + 2)|0;
    $24 = ((($$050)) + 1|0);
    $$049$idx = $$add;$$050 = $24;
   }
   do {
    if ((label|0) == 6) {
     $25 = ($13<<24>>24)==(69);
     if ($25) {
      $26 = ($4|0)==($$050|0);
      L11: do {
       if (!($26)) {
        $$0$i$i = $$050;$$07$i$i = $4;
        while(1) {
         $27 = ((($$0$i$i)) + -1|0);
         $28 = ($$07$i$i>>>0)<($27>>>0);
         if (!($28)) {
          break L11;
         }
         $29 = HEAP8[$$07$i$i>>0]|0;
         $30 = HEAP8[$27>>0]|0;
         HEAP8[$$07$i$i>>0] = $30;
         HEAP8[$27>>0] = $29;
         $31 = ((($$07$i$i)) + 1|0);
         $$0$i$i = $27;$$07$i$i = $31;
        }
       }
      } while(0);
      dest=$5; stop=dest+24|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
      $32 = +HEAPF32[$4>>2];
      $33 = $32;
      HEAPF64[$vararg_buffer>>3] = $33;
      $34 = (_snprintf($5,24,160395,$vararg_buffer)|0);
      $35 = ($34>>>0)>(23);
      if ($35) {
       break;
      }
      ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
      $36 = ($34>>>0)<(11);
      if ($36) {
       $37 = $34&255;
       $38 = ((($7)) + 11|0);
       HEAP8[$38>>0] = $37;
       $$0$i$i57 = $7;
      } else {
       $39 = (($34) + 16)|0;
       $40 = $39 & -16;
       $41 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($40)|0);
       HEAP32[$7>>2] = $41;
       $42 = $40 | -2147483648;
       $43 = ((($7)) + 8|0);
       HEAP32[$43>>2] = $42;
       $44 = ((($7)) + 4|0);
       HEAP32[$44>>2] = $34;
       $$0$i$i57 = $41;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i57,$5,$34)|0);
      $45 = (($$0$i$i57) + ($34)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($45,$3);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($6,$7);
      $46 = ((($2)) + 4|0);
      $47 = HEAP32[$46>>2]|0;
      $48 = ((($2)) + 8|0);
      $49 = HEAP32[$48>>2]|0;
      $50 = ($47>>>0)<($49>>>0);
      if ($50) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($47,$6);
       $51 = HEAP32[$46>>2]|0;
       $52 = ((($51)) + 24|0);
       HEAP32[$46>>2] = $52;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($6|0));
       $53 = __THREW__; __THREW__ = 0;
       $54 = $53&1;
       if ($54) {
        $55 = ___cxa_find_matching_catch_2()|0;
        $56 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        ___resumeException($55|0);
        // unreachable;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$247$ph = $$ptr;
     } else {
      $$247$ph = $0;
     }
     $$3 = $$247$ph;
     break L1;
    }
   } while(0);
   $$3 = $0;
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIdNS0_2DbEEEPKcS4_S4_RT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i57 = 0, $$049$add = 0, $$049$idx = 0, $$049$ptr = 0, $$050 = 0, $$07$i$i = 0, $$247$ph = 0, $$3 = 0, $$add = 0, $$ptr = 0, $$v = 0, $$v56 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit55 = 0, $isdigittmp = 0, $isdigittmp54 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 96|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(96|0);
 $vararg_buffer = sp + 8|0;
 $3 = sp + 88|0;
 $4 = sp;
 $5 = sp + 56|0;
 $6 = sp + 32|0;
 $7 = sp + 16|0;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10>>>0)>(16);
 L1: do {
  if ($11) {
   $$049$idx = 0;$$050 = $4;
   while(1) {
    $$049$ptr = (($0) + ($$049$idx)|0);
    $12 = ($$049$idx|0)==(16);
    $13 = HEAP8[$$049$ptr>>0]|0;
    $14 = $13 << 24 >> 24;
    $$049$add = $$049$idx | 1;
    $$ptr = (($0) + ($$049$add)|0);
    if ($12) {
     label = 6;
     break;
    }
    $15 = (_isxdigit($14)|0);
    $16 = ($15|0)==(0);
    if ($16) {
     break;
    }
    $isdigittmp = (($14) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    $$v = $isdigit ? 0 : 9;
    $17 = (($$v) + ($14))|0;
    $18 = HEAP8[$$ptr>>0]|0;
    $19 = $18 << 24 >> 24;
    $isdigittmp54 = (($19) + -48)|0;
    $isdigit55 = ($isdigittmp54>>>0)<(10);
    $$v56 = $isdigit55 ? 208 : 169;
    $20 = $17 << 4;
    $21 = (($19) + ($20))|0;
    $22 = (($21) + ($$v56))|0;
    $23 = $22&255;
    HEAP8[$$050>>0] = $23;
    $$add = (($$049$idx) + 2)|0;
    $24 = ((($$050)) + 1|0);
    $$049$idx = $$add;$$050 = $24;
   }
   do {
    if ((label|0) == 6) {
     $25 = ($13<<24>>24)==(69);
     if ($25) {
      $26 = ($4|0)==($$050|0);
      L11: do {
       if (!($26)) {
        $$0$i$i = $$050;$$07$i$i = $4;
        while(1) {
         $27 = ((($$0$i$i)) + -1|0);
         $28 = ($$07$i$i>>>0)<($27>>>0);
         if (!($28)) {
          break L11;
         }
         $29 = HEAP8[$$07$i$i>>0]|0;
         $30 = HEAP8[$27>>0]|0;
         HEAP8[$$07$i$i>>0] = $30;
         HEAP8[$27>>0] = $29;
         $31 = ((($$07$i$i)) + 1|0);
         $$0$i$i = $27;$$07$i$i = $31;
        }
       }
      } while(0);
      dest=$5; stop=dest+32|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
      $32 = +HEAPF64[$4>>3];
      HEAPF64[$vararg_buffer>>3] = $32;
      $33 = (_snprintf($5,32,160392,$vararg_buffer)|0);
      $34 = ($33>>>0)>(31);
      if ($34) {
       break;
      }
      ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
      $35 = ($33>>>0)<(11);
      if ($35) {
       $36 = $33&255;
       $37 = ((($7)) + 11|0);
       HEAP8[$37>>0] = $36;
       $$0$i$i57 = $7;
      } else {
       $38 = (($33) + 16)|0;
       $39 = $38 & -16;
       $40 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($39)|0);
       HEAP32[$7>>2] = $40;
       $41 = $39 | -2147483648;
       $42 = ((($7)) + 8|0);
       HEAP32[$42>>2] = $41;
       $43 = ((($7)) + 4|0);
       HEAP32[$43>>2] = $33;
       $$0$i$i57 = $40;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i57,$5,$33)|0);
      $44 = (($$0$i$i57) + ($33)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($44,$3);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($6,$7);
      $45 = ((($2)) + 4|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = ((($2)) + 8|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($46>>>0)<($48>>>0);
      if ($49) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($46,$6);
       $50 = HEAP32[$45>>2]|0;
       $51 = ((($50)) + 24|0);
       HEAP32[$45>>2] = $51;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($6|0));
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        ___resumeException($54|0);
        // unreachable;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$247$ph = $$ptr;
     } else {
      $$247$ph = $0;
     }
     $$3 = $$247$ph;
     break L1;
    }
   } while(0);
   $$3 = $0;
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_121parse_floating_numberIeNS0_2DbEEEPKcS4_S4_RT0_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i = 0, $$0$i$i57 = 0, $$049$add = 0, $$049$idx = 0, $$049$ptr = 0, $$050 = 0, $$07$i$i = 0, $$247$ph = 0, $$3 = 0, $$add = 0, $$ptr = 0, $$v = 0, $$v56 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0.0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit55 = 0, $isdigittmp = 0, $isdigittmp54 = 0, $vararg_buffer = 0, dest = 0, label = 0, sp = 0, stop = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 112|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(112|0);
 $vararg_buffer = sp + 8|0;
 $3 = sp + 96|0;
 $4 = sp;
 $5 = sp + 56|0;
 $6 = sp + 32|0;
 $7 = sp + 16|0;
 $8 = $1;
 $9 = $0;
 $10 = (($8) - ($9))|0;
 $11 = ($10>>>0)>(20);
 L1: do {
  if ($11) {
   $$049$idx = 0;$$050 = $4;
   while(1) {
    $$049$ptr = (($0) + ($$049$idx)|0);
    $12 = ($$049$idx|0)==(20);
    $13 = HEAP8[$$049$ptr>>0]|0;
    $14 = $13 << 24 >> 24;
    $$049$add = $$049$idx | 1;
    $$ptr = (($0) + ($$049$add)|0);
    if ($12) {
     label = 6;
     break;
    }
    $15 = (_isxdigit($14)|0);
    $16 = ($15|0)==(0);
    if ($16) {
     break;
    }
    $isdigittmp = (($14) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    $$v = $isdigit ? 0 : 9;
    $17 = (($$v) + ($14))|0;
    $18 = HEAP8[$$ptr>>0]|0;
    $19 = $18 << 24 >> 24;
    $isdigittmp54 = (($19) + -48)|0;
    $isdigit55 = ($isdigittmp54>>>0)<(10);
    $$v56 = $isdigit55 ? 208 : 169;
    $20 = $17 << 4;
    $21 = (($19) + ($20))|0;
    $22 = (($21) + ($$v56))|0;
    $23 = $22&255;
    HEAP8[$$050>>0] = $23;
    $$add = (($$049$idx) + 2)|0;
    $24 = ((($$050)) + 1|0);
    $$049$idx = $$add;$$050 = $24;
   }
   do {
    if ((label|0) == 6) {
     $25 = ($13<<24>>24)==(69);
     if ($25) {
      $26 = ($4|0)==($$050|0);
      L11: do {
       if (!($26)) {
        $$0$i$i = $$050;$$07$i$i = $4;
        while(1) {
         $27 = ((($$0$i$i)) + -1|0);
         $28 = ($$07$i$i>>>0)<($27>>>0);
         if (!($28)) {
          break L11;
         }
         $29 = HEAP8[$$07$i$i>>0]|0;
         $30 = HEAP8[$27>>0]|0;
         HEAP8[$$07$i$i>>0] = $30;
         HEAP8[$27>>0] = $29;
         $31 = ((($$07$i$i)) + 1|0);
         $$0$i$i = $27;$$07$i$i = $31;
        }
       }
      } while(0);
      dest=$5; stop=dest+40|0; do { HEAP8[dest>>0]=0|0; dest=dest+1|0; } while ((dest|0) < (stop|0));
      $32 = +HEAPF64[$4>>3];
      HEAPF64[$vararg_buffer>>3] = $32;
      $33 = (_snprintf($5,40,160387,$vararg_buffer)|0);
      $34 = ($33>>>0)>(39);
      if ($34) {
       break;
      }
      ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
      $35 = ($33>>>0)<(11);
      if ($35) {
       $36 = $33&255;
       $37 = ((($7)) + 11|0);
       HEAP8[$37>>0] = $36;
       $$0$i$i57 = $7;
      } else {
       $38 = (($33) + 16)|0;
       $39 = $38 & -16;
       $40 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj($39)|0);
       HEAP32[$7>>2] = $40;
       $41 = $39 | -2147483648;
       $42 = ((($7)) + 8|0);
       HEAP32[$42>>2] = $41;
       $43 = ((($7)) + 4|0);
       HEAP32[$43>>2] = $33;
       $$0$i$i57 = $40;
      }
      (__ZNSt3__211char_traitsIcE4copyEPcPKcj($$0$i$i57,$5,$33)|0);
      $44 = (($$0$i$i57) + ($33)|0);
      HEAP8[$3>>0] = 0;
      __ZNSt3__211char_traitsIcE6assignERcRKc($44,$3);
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ENSt3__212basic_stringIcNS2_11char_traitsIcEENS0_12malloc_allocIcEEEE($6,$7);
      $45 = ((($2)) + 4|0);
      $46 = HEAP32[$45>>2]|0;
      $47 = ((($2)) + 8|0);
      $48 = HEAP32[$47>>2]|0;
      $49 = ($46>>>0)<($48>>>0);
      if ($49) {
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2EOS1_($46,$6);
       $50 = HEAP32[$45>>2]|0;
       $51 = ((($50)) + 24|0);
       HEAP32[$45>>2] = $51;
      } else {
       __THREW__ = 0;
       invoke_vii(1128,($2|0),($6|0));
       $52 = __THREW__; __THREW__ = 0;
       $53 = $52&1;
       if ($53) {
        $54 = ___cxa_find_matching_catch_2()|0;
        $55 = tempRet0;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        ___resumeException($54|0);
        // unreachable;
       }
      }
      __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($6);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$247$ph = $$ptr;
     } else {
      $$247$ph = $0;
     }
     $$3 = $$247$ph;
     break L1;
    }
   } while(0);
   $$3 = $0;
  } else {
   $$3 = $0;
  }
 } while(0);
 STACKTOP = sp;return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj14EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(16)|0);
 HEAP32[$0>>2] = $3;
 $4 = ((($0)) + 8|0);
 HEAP32[$4>>2] = -2147483632;
 $5 = ((($0)) + 4|0);
 HEAP32[$5>>2] = 13;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($3,$1,13)|0);
 $6 = ((($3)) + 13|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($6,$2);
 $7 = ((($0)) + 12|0);
 ;HEAP32[$7>>2]=0|0;HEAP32[$7+4>>2]=0|0;HEAP32[$7+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $8 = (($7) + ($$0$i$i<<2)|0);
  HEAP32[$8>>2] = 0;
  $9 = (($$0$i$i) + 1)|0;
  $$0$i$i = $9;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_117parse_nested_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0 = 0, $$0136 = 0, $$0151$off0$ph = 0, $$0157 = 0, $$1 = 0, $$10 = 0, $$10146 = 0, $$11 = 0, $$11147 = 0, $$1137 = 0, $$1158 = 0, $$12 = 0, $$12148 = 0, $$13 = 0, $$13149 = 0, $$14 = 0, $$14150 = 0, $$2 = 0, $$2138 = 0, $$2159 = 0;
 var $$2159$ph = 0, $$3 = 0, $$3139 = 0, $$4 = 0, $$4140 = 0, $$4166 = 0, $$5 = 0, $$5141 = 0, $$6 = 0, $$6142 = 0, $$7 = 0, $$7143 = 0, $$8 = 0, $$8144 = 0, $$9 = 0, $$9145 = 0, $$pre$i = 0, $$pre$i183 = 0, $$pre$i185 = 0, $$pre$i194 = 0;
 var $$pre$i203 = 0, $$pre$i211 = 0, $$pre$i220 = 0, $$sink = 0, $$sroa$0$0$$sroa_idx$i207 = 0, $$sroa$0$0$copyload$i191226 = 0, $$sroa$0$0$copyload$i200224 = 0, $$sroa$0$0$copyload$i208223 = 0, $$sroa$0$0$copyload$i217225 = 0, $$sroa$0$0$copyload$i227 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0;
 var $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0, $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0;
 var $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0, $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0;
 var $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $158 = 0, $159 = 0, $16 = 0, $160 = 0, $161 = 0, $162 = 0;
 var $163 = 0, $164 = 0, $165 = 0, $166 = 0, $167 = 0, $168 = 0, $169 = 0, $17 = 0, $170 = 0, $171 = 0, $172 = 0, $173 = 0, $174 = 0, $175 = 0, $176 = 0, $177 = 0, $178 = 0, $179 = 0, $18 = 0, $180 = 0;
 var $181 = 0, $182 = 0, $183 = 0, $184 = 0, $185 = 0, $186 = 0, $187 = 0, $188 = 0, $189 = 0, $19 = 0, $190 = 0, $191 = 0, $192 = 0, $193 = 0, $194 = 0, $195 = 0, $196 = 0, $197 = 0, $198 = 0, $199 = 0;
 var $20 = 0, $200 = 0, $201 = 0, $202 = 0, $203 = 0, $204 = 0, $205 = 0, $206 = 0, $207 = 0, $208 = 0, $209 = 0, $21 = 0, $210 = 0, $211 = 0, $212 = 0, $213 = 0, $214 = 0, $215 = 0, $216 = 0, $217 = 0;
 var $218 = 0, $219 = 0, $22 = 0, $220 = 0, $221 = 0, $222 = 0, $223 = 0, $224 = 0, $225 = 0, $226 = 0, $227 = 0, $228 = 0, $229 = 0, $23 = 0, $230 = 0, $231 = 0, $232 = 0, $233 = 0, $234 = 0, $235 = 0;
 var $236 = 0, $237 = 0, $238 = 0, $239 = 0, $24 = 0, $240 = 0, $241 = 0, $242 = 0, $243 = 0, $244 = 0, $245 = 0, $246 = 0, $247 = 0, $248 = 0, $249 = 0, $25 = 0, $250 = 0, $251 = 0, $252 = 0, $253 = 0;
 var $254 = 0, $255 = 0, $256 = 0, $257 = 0, $258 = 0, $259 = 0, $26 = 0, $260 = 0, $261 = 0, $262 = 0, $263 = 0, $264 = 0, $265 = 0, $266 = 0, $267 = 0, $268 = 0, $269 = 0, $27 = 0, $270 = 0, $271 = 0;
 var $272 = 0, $273 = 0, $274 = 0, $275 = 0, $276 = 0, $277 = 0, $278 = 0, $279 = 0, $28 = 0, $280 = 0, $281 = 0, $282 = 0, $283 = 0, $284 = 0, $285 = 0, $286 = 0, $287 = 0, $288 = 0, $289 = 0, $29 = 0;
 var $290 = 0, $291 = 0, $292 = 0, $293 = 0, $294 = 0, $295 = 0, $296 = 0, $297 = 0, $298 = 0, $299 = 0, $3 = 0, $30 = 0, $300 = 0, $301 = 0, $302 = 0, $303 = 0, $304 = 0, $305 = 0, $306 = 0, $307 = 0;
 var $308 = 0, $309 = 0, $31 = 0, $310 = 0, $311 = 0, $312 = 0, $313 = 0, $314 = 0, $315 = 0, $316 = 0, $317 = 0, $318 = 0, $319 = 0, $32 = 0, $320 = 0, $321 = 0, $322 = 0, $323 = 0, $324 = 0, $325 = 0;
 var $326 = 0, $327 = 0, $328 = 0, $329 = 0, $33 = 0, $330 = 0, $331 = 0, $332 = 0, $333 = 0, $334 = 0, $335 = 0, $336 = 0, $337 = 0, $338 = 0, $339 = 0, $34 = 0, $340 = 0, $341 = 0, $342 = 0, $343 = 0;
 var $344 = 0, $345 = 0, $346 = 0, $347 = 0, $348 = 0, $349 = 0, $35 = 0, $350 = 0, $351 = 0, $352 = 0, $353 = 0, $354 = 0, $355 = 0, $356 = 0, $357 = 0, $358 = 0, $359 = 0, $36 = 0, $360 = 0, $361 = 0;
 var $362 = 0, $363 = 0, $364 = 0, $365 = 0, $366 = 0, $367 = 0, $368 = 0, $369 = 0, $37 = 0, $370 = 0, $371 = 0, $372 = 0, $373 = 0, $374 = 0, $375 = 0, $376 = 0, $377 = 0, $378 = 0, $379 = 0, $38 = 0;
 var $380 = 0, $381 = 0, $382 = 0, $383 = 0, $384 = 0, $385 = 0, $386 = 0, $387 = 0, $388 = 0, $389 = 0, $39 = 0, $390 = 0, $391 = 0, $392 = 0, $393 = 0, $394 = 0, $395 = 0, $396 = 0, $397 = 0, $398 = 0;
 var $399 = 0, $4 = 0, $40 = 0, $400 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0;
 var $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0;
 var $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0;
 var $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, $or$cond179 = 0, $or$cond180 = 0, $or$cond181 = 0, $or$cond182 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 240|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(240|0);
 $3 = sp + 224|0;
 $4 = sp + 212|0;
 $5 = sp + 200|0;
 $6 = sp + 184|0;
 $7 = sp + 176|0;
 $8 = sp + 164|0;
 $9 = sp + 152|0;
 $10 = sp + 136|0;
 $11 = sp + 128|0;
 $12 = sp + 116|0;
 $13 = sp + 104|0;
 $14 = sp + 88|0;
 $15 = sp + 84|0;
 $16 = sp + 72|0;
 $17 = sp + 56|0;
 $18 = sp + 48|0;
 $19 = sp + 36|0;
 $20 = sp + 24|0;
 $21 = sp + 8|0;
 $22 = sp;
 $23 = ($0|0)==($1|0);
 L1: do {
  if ($23) {
   $$4166 = $0;
  } else {
   $24 = HEAP8[$0>>0]|0;
   $25 = ($24<<24>>24)==(78);
   if ($25) {
    $26 = ((($0)) + 1|0);
    $27 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_cv_qualifiersEPKcS2_Rj($26,$1,$3)|0);
    $28 = ($27|0)==($1|0);
    L4: do {
     if (!($28)) {
      $29 = ((($2)) + 52|0);
      HEAP32[$29>>2] = 0;
      $30 = HEAP8[$27>>0]|0;
      $31 = ((($27)) + 1|0);
      switch ($30<<24>>24) {
      case 82:  {
       $$sink = 1;
       label = 6;
       break;
      }
      case 79:  {
       $$sink = 2;
       label = 6;
       break;
      }
      default: {
       $$0157 = $27;
      }
      }
      if ((label|0) == 6) {
       HEAP32[$29>>2] = $$sink;
       $$0157 = $31;
      }
      $32 = ((($2)) + 4|0);
      $33 = HEAP32[$32>>2]|0;
      $34 = ((($2)) + 8|0);
      $35 = HEAP32[$34>>2]|0;
      $36 = ($33>>>0)<($35>>>0);
      if ($36) {
       ;HEAP32[$33>>2]=0|0;HEAP32[$33+4>>2]=0|0;HEAP32[$33+8>>2]=0|0;HEAP32[$33+12>>2]=0|0;HEAP32[$33+16>>2]=0|0;HEAP32[$33+20>>2]=0|0;
       __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2Ev($33);
       $37 = HEAP32[$32>>2]|0;
       $38 = ((($37)) + 24|0);
       HEAP32[$32>>2] = $38;
      } else {
       __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE24__emplace_back_slow_pathIJEEEvDpOT_($2);
      }
      $39 = $1;
      $40 = $$0157;
      $41 = (($39) - ($40))|0;
      $42 = ($41|0)>(1);
      if ($42) {
       $43 = HEAP8[$$0157>>0]|0;
       $44 = ($43<<24>>24)==(83);
       if ($44) {
        $45 = ((($$0157)) + 1|0);
        $46 = HEAP8[$45>>0]|0;
        $47 = ($46<<24>>24)==(116);
        if ($47) {
         $48 = ((($$0157)) + 2|0);
         $49 = HEAP32[$32>>2]|0;
         $50 = ((($49)) + -24|0);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc($50,160459);
         $$1158 = $48;
        } else {
         $$1158 = $$0157;
        }
       } else {
        $$1158 = $$0157;
       }
      } else {
       $$1158 = $$0157;
      }
      $51 = ($$1158|0)==($1|0);
      if ($51) {
       $80 = HEAP32[$32>>2]|0;
       $81 = ((($80)) + -24|0);
       $83 = $80;
       while(1) {
        $82 = ($83|0)==($81|0);
        if ($82) {
         break L4;
        }
        $84 = ((($83)) + -24|0);
        HEAP32[$32>>2] = $84;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($84);
        $$pre$i = HEAP32[$32>>2]|0;
        $83 = $$pre$i;
       }
      }
      $52 = ((($2)) + 16|0);
      $53 = ((($16)) + 11|0);
      $54 = ((($16)) + 4|0);
      $$sroa$0$0$$sroa_idx$i207 = ((($2)) + 12|0);
      $55 = ((($2)) + 20|0);
      $56 = ((($2)) + 24|0);
      $57 = ((($17)) + 12|0);
      $58 = ((($17)) + 4|0);
      $59 = ((($17)) + 8|0);
      $60 = ((($21)) + 12|0);
      $61 = ((($21)) + 4|0);
      $62 = ((($21)) + 8|0);
      $63 = ((($20)) + 11|0);
      $64 = ((($20)) + 4|0);
      $65 = ((($5)) + 11|0);
      $66 = ((($5)) + 4|0);
      $67 = ((($6)) + 12|0);
      $68 = ((($6)) + 4|0);
      $69 = ((($6)) + 8|0);
      $70 = ((($10)) + 12|0);
      $71 = ((($10)) + 4|0);
      $72 = ((($10)) + 8|0);
      $73 = ((($9)) + 11|0);
      $74 = ((($9)) + 4|0);
      $75 = ((($14)) + 12|0);
      $76 = ((($14)) + 4|0);
      $77 = ((($14)) + 8|0);
      $78 = ((($13)) + 11|0);
      $79 = ((($13)) + 4|0);
      $$0151$off0$ph = 0;$$2159$ph = $$1158;
      L26: while(1) {
       $$2159 = $$2159$ph;
       L28: while(1) {
        $85 = HEAP8[$$2159>>0]|0;
        $86 = ($85<<24>>24)==(69);
        $87 = ((($$2159)) + 1|0);
        if ($86) {
         label = 143;
         break L26;
        }
        $88 = $85 << 24 >> 24;
        switch ($88|0) {
        case 83:  {
         label = 21;
         break L28;
         break;
        }
        case 84:  {
         label = 48;
         break L28;
         break;
        }
        case 68:  {
         label = 73;
         break L28;
         break;
        }
        case 73:  {
         break;
        }
        case 76:  {
         $328 = ($87|0)==($1|0);
         if ($328) {
          break L4;
         } else {
          $$2159 = $87;
          continue L28;
         }
         break;
        }
        default: {
         break L28;
        }
        }
        $283 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_template_argsINS0_2DbEEEPKcS4_S4_RT_($$2159,$1,$2)|0);
        $284 = ($283|0)==($$2159|0);
        $285 = ($283|0)==($1|0);
        $or$cond181 = $284 | $285;
        if ($or$cond181) {
         break L4;
        }
        $286 = HEAP32[$32>>2]|0;
        $287 = ((($286)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($16,$287);
        $288 = HEAP32[$32>>2]|0;
        $289 = ((($288)) + -24|0);
        $291 = $288;
        while(1) {
         $290 = ($291|0)==($289|0);
         if ($290) {
          break;
         }
         $292 = ((($291)) + -24|0);
         HEAP32[$32>>2] = $292;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($292);
         $$pre$i203 = HEAP32[$32>>2]|0;
         $291 = $$pre$i203;
        }
        $293 = ((($288)) + -48|0);
        $294 = HEAP8[$53>>0]|0;
        $295 = ($294<<24>>24)<(0);
        $296 = HEAP32[$16>>2]|0;
        $297 = $295 ? $296 : $16;
        $298 = HEAP32[$54>>2]|0;
        $299 = $294&255;
        $300 = $295 ? $298 : $299;
        __THREW__ = 0;
        (invoke_iiii(1112,($293|0),($297|0),($300|0))|0);
        $301 = __THREW__; __THREW__ = 0;
        $302 = $301&1;
        if ($302) {
         label = 110;
         break L26;
        }
        $303 = HEAP32[$32>>2]|0;
        $304 = ((($303)) + -24|0);
        $$sroa$0$0$copyload$i208223 = HEAP32[$$sroa$0$0$$sroa_idx$i207>>2]|0;
        HEAP32[$18>>2] = $$sroa$0$0$copyload$i208223;
        __THREW__ = 0;
        invoke_viii(1124,($17|0),($304|0),($18|0));
        $305 = __THREW__; __THREW__ = 0;
        $306 = $305&1;
        if ($306) {
         label = 111;
         break L26;
        }
        $307 = HEAP32[$55>>2]|0;
        $308 = HEAP32[$56>>2]|0;
        $309 = ($307>>>0)<($308>>>0);
        if ($309) {
         HEAP32[$307>>2] = 0;
         $310 = ((($307)) + 4|0);
         HEAP32[$310>>2] = 0;
         $311 = ((($307)) + 8|0);
         $312 = HEAP32[$57>>2]|0;
         HEAP32[$311>>2] = 0;
         $313 = ((($307)) + 12|0);
         HEAP32[$313>>2] = $312;
         $314 = HEAP32[$17>>2]|0;
         HEAP32[$307>>2] = $314;
         $315 = HEAP32[$58>>2]|0;
         HEAP32[$310>>2] = $315;
         $316 = HEAP32[$59>>2]|0;
         $317 = ((($307)) + 8|0);
         HEAP32[$317>>2] = $316;
         HEAP32[$59>>2] = 0;
         HEAP32[$58>>2] = 0;
         HEAP32[$17>>2] = 0;
         $318 = HEAP32[$55>>2]|0;
         $319 = ((($318)) + 16|0);
         HEAP32[$55>>2] = $319;
        } else {
         __THREW__ = 0;
         invoke_vii(1122,($52|0),($17|0));
         $320 = __THREW__; __THREW__ = 0;
         $321 = $320&1;
         if ($321) {
          label = 112;
          break L26;
         }
        }
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($17);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
        $$2159 = $283;
       }
       L45: do {
        if ((label|0) == 21) {
         label = 0;
         $89 = ($87|0)==($1|0);
         if (!($89)) {
          $90 = HEAP8[$87>>0]|0;
          $91 = ($90<<24>>24)==(116);
          if ($91) {
           break;
          }
         }
         $92 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_substitutionINS0_2DbEEEPKcS4_S4_RT_($$2159,$1,$2)|0);
         $93 = ($92|0)==($$2159|0);
         $94 = ($92|0)==($1|0);
         $or$cond = $93 | $94;
         if ($or$cond) {
          break L4;
         }
         $95 = HEAP32[$32>>2]|0;
         $96 = ((($95)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($4,$96);
         $97 = HEAP32[$32>>2]|0;
         $98 = ((($97)) + -24|0);
         $100 = $97;
         while(1) {
          $99 = ($100|0)==($98|0);
          if ($99) {
           break;
          }
          $101 = ((($100)) + -24|0);
          HEAP32[$32>>2] = $101;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($101);
          $$pre$i183 = HEAP32[$32>>2]|0;
          $100 = $$pre$i183;
         }
         $102 = ((($97)) + -48|0);
         $103 = ((($102)) + 11|0);
         $104 = HEAP8[$103>>0]|0;
         $105 = ($104<<24>>24)<(0);
         if ($105) {
          $106 = ((($97)) + -44|0);
          $107 = HEAP32[$106>>2]|0;
          $110 = $107;
         } else {
          $108 = $104&255;
          $110 = $108;
         }
         $109 = ($110|0)==(0);
         if ($109) {
          __THREW__ = 0;
          invoke_vii(1145,($102|0),($4|0));
          $153 = __THREW__; __THREW__ = 0;
          $154 = $153&1;
          if ($154) {
           label = 38;
           break L26;
          }
         } else {
          __THREW__ = 0;
          invoke_viii(1139,($5|0),(159155|0),($4|0));
          $111 = __THREW__; __THREW__ = 0;
          $112 = $111&1;
          if ($112) {
           label = 39;
           break L26;
          }
          $113 = HEAP32[$32>>2]|0;
          $114 = ((($113)) + -24|0);
          $115 = HEAP8[$65>>0]|0;
          $116 = ($115<<24>>24)<(0);
          $117 = HEAP32[$5>>2]|0;
          $118 = $116 ? $117 : $5;
          $119 = HEAP32[$66>>2]|0;
          $120 = $115&255;
          $121 = $116 ? $119 : $120;
          __THREW__ = 0;
          (invoke_iiii(1112,($114|0),($118|0),($121|0))|0);
          $122 = __THREW__; __THREW__ = 0;
          $123 = $122&1;
          if ($123) {
           label = 40;
           break L26;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
          $124 = HEAP32[$32>>2]|0;
          $125 = ((($124)) + -24|0);
          $$sroa$0$0$copyload$i227 = HEAP32[$$sroa$0$0$$sroa_idx$i207>>2]|0;
          HEAP32[$7>>2] = $$sroa$0$0$copyload$i227;
          __THREW__ = 0;
          invoke_viii(1124,($6|0),($125|0),($7|0));
          $126 = __THREW__; __THREW__ = 0;
          $127 = $126&1;
          if ($127) {
           label = 42;
           break L26;
          }
          $128 = HEAP32[$55>>2]|0;
          $129 = HEAP32[$56>>2]|0;
          $130 = ($128>>>0)<($129>>>0);
          if ($130) {
           HEAP32[$128>>2] = 0;
           $131 = ((($128)) + 4|0);
           HEAP32[$131>>2] = 0;
           $132 = ((($128)) + 8|0);
           $133 = HEAP32[$67>>2]|0;
           HEAP32[$132>>2] = 0;
           $134 = ((($128)) + 12|0);
           HEAP32[$134>>2] = $133;
           $135 = HEAP32[$6>>2]|0;
           HEAP32[$128>>2] = $135;
           $136 = HEAP32[$68>>2]|0;
           HEAP32[$131>>2] = $136;
           $137 = HEAP32[$69>>2]|0;
           $138 = ((($128)) + 8|0);
           HEAP32[$138>>2] = $137;
           HEAP32[$69>>2] = 0;
           HEAP32[$68>>2] = 0;
           HEAP32[$6>>2] = 0;
           $139 = HEAP32[$55>>2]|0;
           $140 = ((($139)) + 16|0);
           HEAP32[$55>>2] = $140;
          } else {
           __THREW__ = 0;
           invoke_vii(1122,($52|0),($6|0));
           $141 = __THREW__; __THREW__ = 0;
           $142 = $141&1;
           if ($142) {
            label = 43;
            break L26;
           }
          }
          __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($6);
         }
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
         $$0151$off0$ph = 1;$$2159$ph = $92;
         continue L26;
        }
        else if ((label|0) == 48) {
         label = 0;
         $155 = (__ZN10__cxxabiv112_GLOBAL__N_120parse_template_paramINS0_2DbEEEPKcS4_S4_RT_($$2159,$1,$2)|0);
         $156 = ($155|0)==($$2159|0);
         $157 = ($155|0)==($1|0);
         $or$cond179 = $156 | $157;
         if ($or$cond179) {
          break L4;
         }
         $158 = HEAP32[$32>>2]|0;
         $159 = ((($158)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($8,$159);
         $160 = HEAP32[$32>>2]|0;
         $161 = ((($160)) + -24|0);
         $163 = $160;
         while(1) {
          $162 = ($163|0)==($161|0);
          if ($162) {
           break;
          }
          $164 = ((($163)) + -24|0);
          HEAP32[$32>>2] = $164;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($164);
          $$pre$i185 = HEAP32[$32>>2]|0;
          $163 = $$pre$i185;
         }
         $165 = ((($160)) + -48|0);
         $166 = ((($165)) + 11|0);
         $167 = HEAP8[$166>>0]|0;
         $168 = ($167<<24>>24)<(0);
         if ($168) {
          $169 = ((($160)) + -44|0);
          $170 = HEAP32[$169>>2]|0;
          $173 = $170;
         } else {
          $171 = $167&255;
          $173 = $171;
         }
         $172 = ($173|0)==(0);
         if ($172) {
          __THREW__ = 0;
          invoke_vii(1145,($165|0),($8|0));
          $193 = __THREW__; __THREW__ = 0;
          $194 = $193&1;
          if ($194) {
           label = 59;
           break L26;
          }
         } else {
          __THREW__ = 0;
          invoke_viii(1139,($9|0),(159155|0),($8|0));
          $174 = __THREW__; __THREW__ = 0;
          $175 = $174&1;
          if ($175) {
           label = 60;
           break L26;
          }
          $176 = HEAP32[$32>>2]|0;
          $177 = ((($176)) + -24|0);
          $178 = HEAP8[$73>>0]|0;
          $179 = ($178<<24>>24)<(0);
          $180 = HEAP32[$9>>2]|0;
          $181 = $179 ? $180 : $9;
          $182 = HEAP32[$74>>2]|0;
          $183 = $178&255;
          $184 = $179 ? $182 : $183;
          __THREW__ = 0;
          (invoke_iiii(1112,($177|0),($181|0),($184|0))|0);
          $185 = __THREW__; __THREW__ = 0;
          $186 = $185&1;
          if ($186) {
           label = 61;
           break L26;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         }
         $195 = HEAP32[$32>>2]|0;
         $196 = ((($195)) + -24|0);
         $$sroa$0$0$copyload$i191226 = HEAP32[$$sroa$0$0$$sroa_idx$i207>>2]|0;
         HEAP32[$11>>2] = $$sroa$0$0$copyload$i191226;
         __THREW__ = 0;
         invoke_viii(1124,($10|0),($196|0),($11|0));
         $197 = __THREW__; __THREW__ = 0;
         $198 = $197&1;
         if ($198) {
          label = 69;
          break L26;
         }
         $199 = HEAP32[$55>>2]|0;
         $200 = HEAP32[$56>>2]|0;
         $201 = ($199>>>0)<($200>>>0);
         if ($201) {
          HEAP32[$199>>2] = 0;
          $202 = ((($199)) + 4|0);
          HEAP32[$202>>2] = 0;
          $203 = ((($199)) + 8|0);
          $204 = HEAP32[$70>>2]|0;
          HEAP32[$203>>2] = 0;
          $205 = ((($199)) + 12|0);
          HEAP32[$205>>2] = $204;
          $206 = HEAP32[$10>>2]|0;
          HEAP32[$199>>2] = $206;
          $207 = HEAP32[$71>>2]|0;
          HEAP32[$202>>2] = $207;
          $208 = HEAP32[$72>>2]|0;
          $209 = ((($199)) + 8|0);
          HEAP32[$209>>2] = $208;
          HEAP32[$72>>2] = 0;
          HEAP32[$71>>2] = 0;
          HEAP32[$10>>2] = 0;
          $210 = HEAP32[$55>>2]|0;
          $211 = ((($210)) + 16|0);
          HEAP32[$55>>2] = $211;
         } else {
          __THREW__ = 0;
          invoke_vii(1122,($52|0),($10|0));
          $212 = __THREW__; __THREW__ = 0;
          $213 = $212&1;
          if ($213) {
           label = 70;
           break L26;
          }
         }
         __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($10);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
         $$0151$off0$ph = 1;$$2159$ph = $155;
         continue L26;
        }
        else if ((label|0) == 73) {
         label = 0;
         $218 = ($87|0)==($1|0);
         if (!($218)) {
          $219 = HEAP8[$87>>0]|0;
          switch ($219<<24>>24) {
          case 84: case 116:  {
           break;
          }
          default: {
           break L45;
          }
          }
         }
         $220 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_decltypeINS0_2DbEEEPKcS4_S4_RT_($$2159,$1,$2)|0);
         $221 = ($220|0)==($$2159|0);
         $222 = ($220|0)==($1|0);
         $or$cond180 = $221 | $222;
         if ($or$cond180) {
          break L4;
         }
         $223 = HEAP32[$32>>2]|0;
         $224 = ((($223)) + -24|0);
         __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($12,$224);
         $225 = HEAP32[$32>>2]|0;
         $226 = ((($225)) + -24|0);
         $228 = $225;
         while(1) {
          $227 = ($228|0)==($226|0);
          if ($227) {
           break;
          }
          $229 = ((($228)) + -24|0);
          HEAP32[$32>>2] = $229;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($229);
          $$pre$i194 = HEAP32[$32>>2]|0;
          $228 = $$pre$i194;
         }
         $230 = ((($225)) + -48|0);
         $231 = ((($230)) + 11|0);
         $232 = HEAP8[$231>>0]|0;
         $233 = ($232<<24>>24)<(0);
         if ($233) {
          $234 = ((($225)) + -44|0);
          $235 = HEAP32[$234>>2]|0;
          $238 = $235;
         } else {
          $236 = $232&255;
          $238 = $236;
         }
         $237 = ($238|0)==(0);
         if ($237) {
          __THREW__ = 0;
          invoke_vii(1145,($230|0),($12|0));
          $258 = __THREW__; __THREW__ = 0;
          $259 = $258&1;
          if ($259) {
           label = 86;
           break L26;
          }
         } else {
          __THREW__ = 0;
          invoke_viii(1139,($13|0),(159155|0),($12|0));
          $239 = __THREW__; __THREW__ = 0;
          $240 = $239&1;
          if ($240) {
           label = 87;
           break L26;
          }
          $241 = HEAP32[$32>>2]|0;
          $242 = ((($241)) + -24|0);
          $243 = HEAP8[$78>>0]|0;
          $244 = ($243<<24>>24)<(0);
          $245 = HEAP32[$13>>2]|0;
          $246 = $244 ? $245 : $13;
          $247 = HEAP32[$79>>2]|0;
          $248 = $243&255;
          $249 = $244 ? $247 : $248;
          __THREW__ = 0;
          (invoke_iiii(1112,($242|0),($246|0),($249|0))|0);
          $250 = __THREW__; __THREW__ = 0;
          $251 = $250&1;
          if ($251) {
           label = 88;
           break L26;
          }
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
         }
         $260 = HEAP32[$32>>2]|0;
         $261 = ((($260)) + -24|0);
         $$sroa$0$0$copyload$i200224 = HEAP32[$$sroa$0$0$$sroa_idx$i207>>2]|0;
         HEAP32[$15>>2] = $$sroa$0$0$copyload$i200224;
         __THREW__ = 0;
         invoke_viii(1124,($14|0),($261|0),($15|0));
         $262 = __THREW__; __THREW__ = 0;
         $263 = $262&1;
         if ($263) {
          label = 96;
          break L26;
         }
         $264 = HEAP32[$55>>2]|0;
         $265 = HEAP32[$56>>2]|0;
         $266 = ($264>>>0)<($265>>>0);
         if ($266) {
          HEAP32[$264>>2] = 0;
          $267 = ((($264)) + 4|0);
          HEAP32[$267>>2] = 0;
          $268 = ((($264)) + 8|0);
          $269 = HEAP32[$75>>2]|0;
          HEAP32[$268>>2] = 0;
          $270 = ((($264)) + 12|0);
          HEAP32[$270>>2] = $269;
          $271 = HEAP32[$14>>2]|0;
          HEAP32[$264>>2] = $271;
          $272 = HEAP32[$76>>2]|0;
          HEAP32[$267>>2] = $272;
          $273 = HEAP32[$77>>2]|0;
          $274 = ((($264)) + 8|0);
          HEAP32[$274>>2] = $273;
          HEAP32[$77>>2] = 0;
          HEAP32[$76>>2] = 0;
          HEAP32[$14>>2] = 0;
          $275 = HEAP32[$55>>2]|0;
          $276 = ((($275)) + 16|0);
          HEAP32[$55>>2] = $276;
         } else {
          __THREW__ = 0;
          invoke_vii(1122,($52|0),($14|0));
          $277 = __THREW__; __THREW__ = 0;
          $278 = $277&1;
          if ($278) {
           label = 97;
           break L26;
          }
         }
         __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($14);
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
         $$0151$off0$ph = 1;$$2159$ph = $220;
         continue L26;
        }
       } while(0);
       $329 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_unqualified_nameINS0_2DbEEEPKcS4_S4_RT_($$2159,$1,$2)|0);
       $330 = ($329|0)==($$2159|0);
       $331 = ($329|0)==($1|0);
       $or$cond182 = $330 | $331;
       if ($or$cond182) {
        break L4;
       }
       $332 = HEAP32[$32>>2]|0;
       $333 = ((($332)) + -24|0);
       __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($19,$333);
       $334 = HEAP32[$32>>2]|0;
       $335 = ((($334)) + -24|0);
       $337 = $334;
       while(1) {
        $336 = ($337|0)==($335|0);
        if ($336) {
         break;
        }
        $338 = ((($337)) + -24|0);
        HEAP32[$32>>2] = $338;
        __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($338);
        $$pre$i211 = HEAP32[$32>>2]|0;
        $337 = $$pre$i211;
       }
       $339 = ((($334)) + -48|0);
       $340 = ((($339)) + 11|0);
       $341 = HEAP8[$340>>0]|0;
       $342 = ($341<<24>>24)<(0);
       if ($342) {
        $343 = ((($334)) + -44|0);
        $344 = HEAP32[$343>>2]|0;
        $347 = $344;
       } else {
        $345 = $341&255;
        $347 = $345;
       }
       $346 = ($347|0)==(0);
       if ($346) {
        __THREW__ = 0;
        invoke_vii(1145,($339|0),($19|0));
        $367 = __THREW__; __THREW__ = 0;
        $368 = $367&1;
        if ($368) {
         label = 127;
         break;
        }
       } else {
        __THREW__ = 0;
        invoke_viii(1139,($20|0),(159155|0),($19|0));
        $348 = __THREW__; __THREW__ = 0;
        $349 = $348&1;
        if ($349) {
         label = 128;
         break;
        }
        $350 = HEAP32[$32>>2]|0;
        $351 = ((($350)) + -24|0);
        $352 = HEAP8[$63>>0]|0;
        $353 = ($352<<24>>24)<(0);
        $354 = HEAP32[$20>>2]|0;
        $355 = $353 ? $354 : $20;
        $356 = HEAP32[$64>>2]|0;
        $357 = $352&255;
        $358 = $353 ? $356 : $357;
        __THREW__ = 0;
        (invoke_iiii(1112,($351|0),($355|0),($358|0))|0);
        $359 = __THREW__; __THREW__ = 0;
        $360 = $359&1;
        if ($360) {
         label = 129;
         break;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($20);
       }
       $369 = HEAP32[$32>>2]|0;
       $370 = ((($369)) + -24|0);
       $$sroa$0$0$copyload$i217225 = HEAP32[$$sroa$0$0$$sroa_idx$i207>>2]|0;
       HEAP32[$22>>2] = $$sroa$0$0$copyload$i217225;
       __THREW__ = 0;
       invoke_viii(1124,($21|0),($370|0),($22|0));
       $371 = __THREW__; __THREW__ = 0;
       $372 = $371&1;
       if ($372) {
        label = 138;
        break;
       }
       $373 = HEAP32[$55>>2]|0;
       $374 = HEAP32[$56>>2]|0;
       $375 = ($373>>>0)<($374>>>0);
       if ($375) {
        HEAP32[$373>>2] = 0;
        $376 = ((($373)) + 4|0);
        HEAP32[$376>>2] = 0;
        $377 = ((($373)) + 8|0);
        $378 = HEAP32[$60>>2]|0;
        HEAP32[$377>>2] = 0;
        $379 = ((($373)) + 12|0);
        HEAP32[$379>>2] = $378;
        $380 = HEAP32[$21>>2]|0;
        HEAP32[$373>>2] = $380;
        $381 = HEAP32[$61>>2]|0;
        HEAP32[$376>>2] = $381;
        $382 = HEAP32[$62>>2]|0;
        $383 = ((($373)) + 8|0);
        HEAP32[$383>>2] = $382;
        HEAP32[$62>>2] = 0;
        HEAP32[$61>>2] = 0;
        HEAP32[$21>>2] = 0;
        $384 = HEAP32[$55>>2]|0;
        $385 = ((($384)) + 16|0);
        HEAP32[$55>>2] = $385;
       } else {
        __THREW__ = 0;
        invoke_vii(1122,($52|0),($21|0));
        $386 = __THREW__; __THREW__ = 0;
        $387 = $386&1;
        if ($387) {
         label = 139;
         break;
        }
       }
       __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($21);
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
       $$0151$off0$ph = 1;$$2159$ph = $329;
      }
      switch (label|0) {
       case 38: {
        $143 = ___cxa_find_matching_catch_2()|0;
        $144 = tempRet0;
        $$2 = $144;$$2138 = $143;
        label = 47;
        break;
       }
       case 39: {
        $145 = ___cxa_find_matching_catch_2()|0;
        $146 = tempRet0;
        $$0 = $146;$$0136 = $145;
        label = 41;
        break;
       }
       case 40: {
        $147 = ___cxa_find_matching_catch_2()|0;
        $148 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
        $$0 = $148;$$0136 = $147;
        label = 41;
        break;
       }
       case 42: {
        $149 = ___cxa_find_matching_catch_2()|0;
        $150 = tempRet0;
        $$1 = $150;$$1137 = $149;
        label = 44;
        break;
       }
       case 43: {
        $151 = ___cxa_find_matching_catch_2()|0;
        $152 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($6);
        $$1 = $152;$$1137 = $151;
        label = 44;
        break;
       }
       case 59: {
        $187 = ___cxa_find_matching_catch_2()|0;
        $188 = tempRet0;
        $$5 = $188;$$5141 = $187;
        label = 72;
        break;
       }
       case 60: {
        $189 = ___cxa_find_matching_catch_2()|0;
        $190 = tempRet0;
        $$3 = $190;$$3139 = $189;
        label = 62;
        break;
       }
       case 61: {
        $191 = ___cxa_find_matching_catch_2()|0;
        $192 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
        $$3 = $192;$$3139 = $191;
        label = 62;
        break;
       }
       case 69: {
        $214 = ___cxa_find_matching_catch_2()|0;
        $215 = tempRet0;
        $$4 = $215;$$4140 = $214;
        label = 71;
        break;
       }
       case 70: {
        $216 = ___cxa_find_matching_catch_2()|0;
        $217 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($10);
        $$4 = $217;$$4140 = $216;
        label = 71;
        break;
       }
       case 86: {
        $252 = ___cxa_find_matching_catch_2()|0;
        $253 = tempRet0;
        $$8 = $253;$$8144 = $252;
        label = 99;
        break;
       }
       case 87: {
        $254 = ___cxa_find_matching_catch_2()|0;
        $255 = tempRet0;
        $$6 = $255;$$6142 = $254;
        label = 89;
        break;
       }
       case 88: {
        $256 = ___cxa_find_matching_catch_2()|0;
        $257 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($13);
        $$6 = $257;$$6142 = $256;
        label = 89;
        break;
       }
       case 96: {
        $279 = ___cxa_find_matching_catch_2()|0;
        $280 = tempRet0;
        $$7 = $280;$$7143 = $279;
        label = 98;
        break;
       }
       case 97: {
        $281 = ___cxa_find_matching_catch_2()|0;
        $282 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($14);
        $$7 = $282;$$7143 = $281;
        label = 98;
        break;
       }
       case 110: {
        $322 = ___cxa_find_matching_catch_2()|0;
        $323 = tempRet0;
        $$10 = $323;$$10146 = $322;
        label = 114;
        break;
       }
       case 111: {
        $324 = ___cxa_find_matching_catch_2()|0;
        $325 = tempRet0;
        $$9 = $325;$$9145 = $324;
        label = 113;
        break;
       }
       case 112: {
        $326 = ___cxa_find_matching_catch_2()|0;
        $327 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($17);
        $$9 = $327;$$9145 = $326;
        label = 113;
        break;
       }
       case 127: {
        $361 = ___cxa_find_matching_catch_2()|0;
        $362 = tempRet0;
        $$13 = $362;$$13149 = $361;
        label = 141;
        break;
       }
       case 128: {
        $363 = ___cxa_find_matching_catch_2()|0;
        $364 = tempRet0;
        $$11 = $364;$$11147 = $363;
        label = 130;
        break;
       }
       case 129: {
        $365 = ___cxa_find_matching_catch_2()|0;
        $366 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($20);
        $$11 = $366;$$11147 = $365;
        label = 130;
        break;
       }
       case 138: {
        $388 = ___cxa_find_matching_catch_2()|0;
        $389 = tempRet0;
        $$12 = $389;$$12148 = $388;
        label = 140;
        break;
       }
       case 139: {
        $390 = ___cxa_find_matching_catch_2()|0;
        $391 = tempRet0;
        __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($21);
        $$12 = $391;$$12148 = $390;
        label = 140;
        break;
       }
       case 143: {
        $392 = HEAP32[$3>>2]|0;
        $393 = ((($2)) + 48|0);
        HEAP32[$393>>2] = $392;
        L158: do {
         if ($$0151$off0$ph) {
          $394 = HEAP32[$52>>2]|0;
          $395 = HEAP32[$55>>2]|0;
          $396 = ($394|0)==($395|0);
          if (!($396)) {
           $397 = ((($395)) + -16|0);
           $399 = $395;
           while(1) {
            $398 = ($399|0)==($397|0);
            if ($398) {
             break L158;
            }
            $400 = ((($399)) + -16|0);
            HEAP32[$55>>2] = $400;
            __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($400);
            $$pre$i220 = HEAP32[$55>>2]|0;
            $399 = $$pre$i220;
           }
          }
         }
        } while(0);
        $$4166 = $87;
        break L1;
        break;
       }
      }
      if ((label|0) == 41) {
       $$2 = $$0;$$2138 = $$0136;
       label = 47;
      }
      else if ((label|0) == 44) {
       $$2 = $$1;$$2138 = $$1137;
       label = 47;
      }
      else if ((label|0) == 62) {
       $$5 = $$3;$$5141 = $$3139;
       label = 72;
      }
      else if ((label|0) == 71) {
       $$5 = $$4;$$5141 = $$4140;
       label = 72;
      }
      else if ((label|0) == 89) {
       $$8 = $$6;$$8144 = $$6142;
       label = 99;
      }
      else if ((label|0) == 98) {
       $$8 = $$7;$$8144 = $$7143;
       label = 99;
      }
      else if ((label|0) == 113) {
       $$10 = $$9;$$10146 = $$9145;
       label = 114;
      }
      else if ((label|0) == 130) {
       $$13 = $$11;$$13149 = $$11147;
       label = 141;
      }
      else if ((label|0) == 140) {
       $$13 = $$12;$$13149 = $$12148;
       label = 141;
      }
      if ((label|0) == 47) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
       $$14 = $$2;$$14150 = $$2138;
      }
      else if ((label|0) == 72) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
       $$14 = $$5;$$14150 = $$5141;
      }
      else if ((label|0) == 99) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($12);
       $$14 = $$8;$$14150 = $$8144;
      }
      else if ((label|0) == 114) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($16);
       $$14 = $$10;$$14150 = $$10146;
      }
      else if ((label|0) == 141) {
       __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($19);
       $$14 = $$13;$$14150 = $$13149;
      }
      ___resumeException($$14150|0);
      // unreachable;
     }
    } while(0);
    $$4166 = $0;
   } else {
    $$4166 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$4166|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_116parse_local_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$068 = 0, $$069 = 0, $$4 = 0, $$pre$i = 0, $$pre$i78 = 0, $$pre$i80 = 0, $$pre$i83 = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0, $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $11 = 0, $12 = 0, $13 = 0;
 var $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0;
 var $33 = 0, $34 = 0, $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0;
 var $51 = 0, $52 = 0, $53 = 0, $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0;
 var $7 = 0, $70 = 0, $71 = 0, $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0;
 var $88 = 0, $89 = 0, $9 = 0, $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $or$cond = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $3 = sp + 12|0;
 $4 = sp;
 $5 = ($0|0)==($1|0);
 L1: do {
  if ($5) {
   $$4 = $0;
  } else {
   $6 = HEAP8[$0>>0]|0;
   $7 = ($6<<24>>24)==(90);
   if ($7) {
    $8 = ((($0)) + 1|0);
    $9 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($8,$1,$2)|0);
    $10 = ($9|0)==($8|0);
    $11 = ($9|0)==($1|0);
    $or$cond = $10 | $11;
    if ($or$cond) {
     $$4 = $0;
    } else {
     $12 = HEAP8[$9>>0]|0;
     $13 = ($12<<24>>24)==(69);
     if ($13) {
      $14 = ((($9)) + 1|0);
      $15 = ($14|0)==($1|0);
      if ($15) {
       $$4 = $0;
      } else {
       $16 = HEAP8[$14>>0]|0;
       $17 = $16 << 24 >> 24;
       $18 = ((($9)) + 2|0);
       switch ($17|0) {
       case 115:  {
        $19 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_discriminatorEPKcS2_($18,$1)|0);
        $20 = HEAP32[$2>>2]|0;
        $21 = ((($2)) + 4|0);
        $22 = HEAP32[$21>>2]|0;
        $23 = ($20|0)==($22|0);
        if ($23) {
         $$4 = $19;
         break L1;
        }
        $24 = ((($22)) + -24|0);
        (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc($24,160442)|0);
        $$4 = $19;
        break L1;
        break;
       }
       case 100:  {
        $25 = ($18|0)==($1|0);
        if ($25) {
         $$4 = $0;
         break L1;
        }
        $26 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($18,$1)|0);
        $27 = ($26|0)==($1|0);
        if ($27) {
         $$4 = $0;
         break L1;
        }
        $28 = HEAP8[$26>>0]|0;
        $29 = ($28<<24>>24)==(95);
        if (!($29)) {
         $$4 = $0;
         break L1;
        }
        $30 = ((($26)) + 1|0);
        $31 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($30,$1,$2)|0);
        $32 = ($31|0)==($30|0);
        $33 = ((($2)) + 4|0);
        if ($32) {
         $64 = HEAP32[$33>>2]|0;
         $65 = ((($64)) + -24|0);
         $67 = $64;
         while(1) {
          $66 = ($67|0)==($65|0);
          if ($66) {
           $$4 = $0;
           break L1;
          }
          $68 = ((($67)) + -24|0);
          HEAP32[$33>>2] = $68;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($68);
          $$pre$i78 = HEAP32[$33>>2]|0;
          $67 = $$pre$i78;
         }
        }
        $34 = HEAP32[$33>>2]|0;
        $35 = HEAP32[$2>>2]|0;
        $36 = (($34) - ($35))|0;
        $37 = (($36|0) / 24)&-1;
        $38 = ($37>>>0)<(2);
        if ($38) {
         $$4 = $0;
         break L1;
        }
        $39 = $34;
        $40 = ((($39)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($3,$40);
        $41 = HEAP32[$33>>2]|0;
        $42 = ((($41)) + -24|0);
        $44 = $41;
        while(1) {
         $43 = ($44|0)==($42|0);
         if ($43) {
          break;
         }
         $45 = ((($44)) + -24|0);
         HEAP32[$33>>2] = $45;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($45);
         $$pre$i = HEAP32[$33>>2]|0;
         $44 = $$pre$i;
        }
        $46 = ((($41)) + -48|0);
        __THREW__ = 0;
        (invoke_iii(1119,($46|0),(159155|0))|0);
        $47 = __THREW__; __THREW__ = 0;
        $48 = $47&1;
        if (!($48)) {
         $49 = HEAP32[$33>>2]|0;
         $50 = ((($49)) + -24|0);
         $51 = ((($3)) + 11|0);
         $52 = HEAP8[$51>>0]|0;
         $53 = ($52<<24>>24)<(0);
         $54 = HEAP32[$3>>2]|0;
         $55 = $53 ? $54 : $3;
         $56 = ((($3)) + 4|0);
         $57 = HEAP32[$56>>2]|0;
         $58 = $52&255;
         $59 = $53 ? $57 : $58;
         __THREW__ = 0;
         (invoke_iiii(1112,($50|0),($55|0),($59|0))|0);
         $60 = __THREW__; __THREW__ = 0;
         $61 = $60&1;
         if (!($61)) {
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
          $$4 = $31;
          break L1;
         }
        }
        $62 = ___cxa_find_matching_catch_2()|0;
        $63 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($3);
        $$068 = $63;$$069 = $62;
        break;
       }
       default: {
        $69 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($14,$1,$2)|0);
        $70 = ($69|0)==($14|0);
        if ($70) {
         $103 = ((($2)) + 4|0);
         $104 = HEAP32[$103>>2]|0;
         $105 = ((($104)) + -24|0);
         $107 = $104;
         while(1) {
          $106 = ($107|0)==($105|0);
          if ($106) {
           $$4 = $0;
           break L1;
          }
          $108 = ((($107)) + -24|0);
          HEAP32[$103>>2] = $108;
          __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($108);
          $$pre$i83 = HEAP32[$103>>2]|0;
          $107 = $$pre$i83;
         }
        }
        $71 = (__ZN10__cxxabiv112_GLOBAL__N_119parse_discriminatorEPKcS2_($69,$1)|0);
        $72 = ((($2)) + 4|0);
        $73 = HEAP32[$72>>2]|0;
        $74 = HEAP32[$2>>2]|0;
        $75 = (($73) - ($74))|0;
        $76 = (($75|0) / 24)&-1;
        $77 = ($76>>>0)<(2);
        if ($77) {
         $$4 = $71;
         break L1;
        }
        $78 = $73;
        $79 = ((($78)) + -24|0);
        __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($4,$79);
        $80 = HEAP32[$72>>2]|0;
        $81 = ((($80)) + -24|0);
        $83 = $80;
        while(1) {
         $82 = ($83|0)==($81|0);
         if ($82) {
          break;
         }
         $84 = ((($83)) + -24|0);
         HEAP32[$72>>2] = $84;
         __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($84);
         $$pre$i80 = HEAP32[$72>>2]|0;
         $83 = $$pre$i80;
        }
        $85 = ((($80)) + -48|0);
        __THREW__ = 0;
        (invoke_iii(1119,($85|0),(159155|0))|0);
        $86 = __THREW__; __THREW__ = 0;
        $87 = $86&1;
        if (!($87)) {
         $88 = HEAP32[$72>>2]|0;
         $89 = ((($88)) + -24|0);
         $90 = ((($4)) + 11|0);
         $91 = HEAP8[$90>>0]|0;
         $92 = ($91<<24>>24)<(0);
         $93 = HEAP32[$4>>2]|0;
         $94 = $92 ? $93 : $4;
         $95 = ((($4)) + 4|0);
         $96 = HEAP32[$95>>2]|0;
         $97 = $91&255;
         $98 = $92 ? $96 : $97;
         __THREW__ = 0;
         (invoke_iiii(1112,($89|0),($94|0),($98|0))|0);
         $99 = __THREW__; __THREW__ = 0;
         $100 = $99&1;
         if (!($100)) {
          __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
          $$4 = $71;
          break L1;
         }
        }
        $101 = ___cxa_find_matching_catch_2()|0;
        $102 = tempRet0;
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($4);
        $$068 = $102;$$069 = $101;
       }
       }
       ___resumeException($$069|0);
       // unreachable;
      }
     } else {
      $$4 = $0;
     }
    }
   } else {
    $$4 = $0;
   }
  }
 } while(0);
 STACKTOP = sp;return ($$4|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_unscoped_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$026$off0 = 0, $$027 = 0, $$1 = 0, $$mux = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0;
 var $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $brmerge = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = $1;
 $4 = $0;
 $5 = (($3) - ($4))|0;
 $6 = ($5|0)>(1);
 if ($6) {
  $7 = HEAP8[$0>>0]|0;
  $8 = ($7<<24>>24)==(83);
  if ($8) {
   $9 = ((($0)) + 1|0);
   $10 = HEAP8[$9>>0]|0;
   $11 = ($10<<24>>24)==(116);
   if ($11) {
    $12 = ((($0)) + 2|0);
    $13 = ($12|0)==($1|0);
    if ($13) {
     $$026$off0 = 0;$$027 = $1;
    } else {
     $14 = HEAP8[$12>>0]|0;
     $15 = ($14<<24>>24)==(76);
     $16 = ((($0)) + 3|0);
     $$ = $15 ? $16 : $12;
     $$026$off0 = 0;$$027 = $$;
    }
   } else {
    $$026$off0 = 1;$$027 = $0;
   }
  } else {
   $$026$off0 = 1;$$027 = $0;
  }
  $17 = (__ZN10__cxxabiv112_GLOBAL__N_122parse_unqualified_nameINS0_2DbEEEPKcS4_S4_RT_($$027,$1,$2)|0);
  $18 = ($17|0)==($$027|0);
  $brmerge = $$026$off0 | $18;
  $$mux = $18 ? $0 : $17;
  if ($brmerge) {
   $$1 = $$mux;
  } else {
   $19 = HEAP32[$2>>2]|0;
   $20 = ((($2)) + 4|0);
   $21 = HEAP32[$20>>2]|0;
   $22 = ($19|0)==($21|0);
   if ($22) {
    $$1 = $0;
   } else {
    $23 = ((($21)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($23,0,159377)|0);
    $$1 = $17;
   }
  }
 } else {
  $$1 = $0;
 }
 return ($$1|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_119parse_discriminatorEPKcS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$ = 0, $$0 = 0, $$029 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0;
 var $7 = 0, $8 = 0, $9 = 0, $isdigit = 0, $isdigit32 = 0, $isdigit34 = 0, $isdigit36 = 0, $isdigittmp = 0, $isdigittmp31 = 0, $isdigittmp33 = 0, $isdigittmp35 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==($1|0);
 L1: do {
  if ($2) {
   $$1 = $0;
  } else {
   $3 = HEAP8[$0>>0]|0;
   $4 = ($3<<24>>24)==(95);
   $5 = ((($0)) + 1|0);
   if (!($4)) {
    $16 = $3 << 24 >> 24;
    $isdigittmp = (($16) + -48)|0;
    $isdigit = ($isdigittmp>>>0)<(10);
    if ($isdigit) {
     $$0 = $5;
    } else {
     $$1 = $0;
     break;
    }
    while(1) {
     $17 = ($$0|0)==($1|0);
     if ($17) {
      $$1 = $1;
      break L1;
     }
     $18 = HEAP8[$$0>>0]|0;
     $19 = $18 << 24 >> 24;
     $isdigittmp31 = (($19) + -48)|0;
     $isdigit32 = ($isdigittmp31>>>0)<(10);
     if (!($isdigit32)) {
      $$1 = $$0;
      break L1;
     }
     $20 = ((($$0)) + 1|0);
     $$0 = $20;
    }
   }
   $6 = ($5|0)==($1|0);
   if ($6) {
    $$1 = $0;
   } else {
    $7 = HEAP8[$5>>0]|0;
    $8 = $7 << 24 >> 24;
    $isdigittmp33 = (($8) + -48)|0;
    $isdigit34 = ($isdigittmp33>>>0)<(10);
    $9 = ((($0)) + 2|0);
    if ($isdigit34) {
     $$1 = $9;
    } else {
     $10 = ($7<<24>>24)==(95);
     if ($10) {
      $$029 = $9;
      while(1) {
       $11 = ($$029|0)==($1|0);
       $12 = ((($$029)) + 1|0);
       if ($11) {
        $$1 = $0;
        break L1;
       }
       $13 = HEAP8[$$029>>0]|0;
       $14 = $13 << 24 >> 24;
       $isdigittmp35 = (($14) + -48)|0;
       $isdigit36 = ($isdigittmp35>>>0)<(10);
       if ($isdigit36) {
        $$029 = $12;
       } else {
        break;
       }
      }
      $15 = ($13<<24>>24)==(95);
      $$ = $15 ? $12 : $0;
      return ($$|0);
     } else {
      $$1 = $0;
     }
    }
   }
  }
 } while(0);
 return ($$1|0);
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEaSERKS7_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $10 = 0, $11 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==($1|0);
 if (!($2)) {
  $3 = ((($1)) + 11|0);
  $4 = HEAP8[$3>>0]|0;
  $5 = ($4<<24>>24)<(0);
  $6 = HEAP32[$1>>2]|0;
  $7 = $5 ? $6 : $1;
  $8 = ((($1)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = $4&255;
  $11 = $5 ? $9 : $10;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKcj($0,$7,$11);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj22EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(32)|0);
 HEAP32[$0>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = -2147483616;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 21;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($2,160474,21)|0);
 $5 = ((($2)) + 21|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($5,$1);
 $6 = ((($0)) + 12|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2IS4_EERKNS1_IT_Lj4096EEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$1>>2]|0;
 HEAP32[$0>>2] = $2;
 return;
}
function __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$sroa$speculated$$i = 0, $$sroa$speculated$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0;
 var $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 32|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(32|0);
 $2 = sp + 20|0;
 $3 = sp;
 $4 = ((($0)) + 4|0);
 $5 = HEAP32[$4>>2]|0;
 $6 = HEAP32[$0>>2]|0;
 $7 = (($5) - ($6))|0;
 $8 = $7 >> 4;
 $9 = (($8) + 1)|0;
 $10 = (__ZNKSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE8max_sizeEv($0)|0);
 $11 = ($10>>>0)<($9>>>0);
 if ($11) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $12 = ((($0)) + 12|0);
  $13 = ((($0)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = HEAP32[$0>>2]|0;
  $16 = (($14) - ($15))|0;
  $17 = $16 >> 4;
  $18 = $10 >>> 1;
  $19 = ($17>>>0)<($18>>>0);
  $20 = $16 >> 3;
  $21 = ($20>>>0)<($9>>>0);
  $$sroa$speculated$i = $21 ? $9 : $20;
  $$sroa$speculated$$i = $19 ? $$sroa$speculated$i : $10;
  $22 = HEAP32[$4>>2]|0;
  $23 = (($22) - ($15))|0;
  $24 = $23 >> 4;
  __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEEC2EjjSB_($3,$$sroa$speculated$$i,$24,$12);
  $25 = ((($3)) + 8|0);
  $26 = HEAP32[$25>>2]|0;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EEC2IS4_EERKNS1_IT_Lj4096EEE($2,$1);
  HEAP32[$26>>2] = 0;
  $27 = ((($26)) + 4|0);
  HEAP32[$27>>2] = 0;
  $28 = ((($26)) + 8|0);
  $29 = HEAP32[$2>>2]|0;
  HEAP32[$28>>2] = 0;
  $30 = ((($26)) + 12|0);
  HEAP32[$30>>2] = $29;
  $31 = ((($26)) + 16|0);
  HEAP32[$25>>2] = $31;
  __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS8_RS9_EE($0,$3);
  __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEED2Ev($3);
  STACKTOP = sp;return;
 }
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_template_argINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$ = 0, $$051 = 0, $$4 = 0, $$58 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0;
 var $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond57 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $3 = ($0|0)==($1|0);
 L1: do {
  if ($3) {
   $$4 = $0;
  } else {
   $4 = HEAP8[$0>>0]|0;
   $5 = $4 << 24 >> 24;
   $6 = ((($0)) + 1|0);
   switch ($5|0) {
   case 88:  {
    $7 = (__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_($6,$1,$2)|0);
    $8 = ($7|0)==($6|0);
    $9 = ($7|0)==($1|0);
    $or$cond = $8 | $9;
    if ($or$cond) {
     $$4 = $0;
     break L1;
    }
    $10 = HEAP8[$7>>0]|0;
    $11 = ($10<<24>>24)==(69);
    $12 = ((($7)) + 1|0);
    $$ = $11 ? $12 : $0;
    $$4 = $$;
    break L1;
    break;
   }
   case 74:  {
    $13 = ($6|0)==($1|0);
    if ($13) {
     $$4 = $0;
     break L1;
    } else {
     $$051 = $6;
    }
    while(1) {
     $14 = HEAP8[$$051>>0]|0;
     $15 = ($14<<24>>24)==(69);
     if ($15) {
      break;
     }
     $16 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_template_argINS0_2DbEEEPKcS4_S4_RT_($$051,$1,$2)|0);
     $17 = ($16|0)==($$051|0);
     if ($17) {
      $$4 = $0;
      break L1;
     } else {
      $$051 = $16;
     }
    }
    $18 = ((($$051)) + 1|0);
    $$4 = $18;
    break L1;
    break;
   }
   case 76:  {
    $19 = ($6|0)==($1|0);
    if (!($19)) {
     $20 = HEAP8[$6>>0]|0;
     $21 = ($20<<24>>24)==(90);
     if ($21) {
      $22 = ((($0)) + 2|0);
      $23 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($22,$1,$2)|0);
      $24 = ($23|0)==($22|0);
      $25 = ($23|0)==($1|0);
      $or$cond57 = $24 | $25;
      if ($or$cond57) {
       $$4 = $0;
       break L1;
      }
      $26 = HEAP8[$23>>0]|0;
      $27 = ($26<<24>>24)==(69);
      $28 = ((($23)) + 1|0);
      $$58 = $27 ? $28 : $0;
      $$4 = $$58;
      break L1;
     }
    }
    $29 = (__ZN10__cxxabiv112_GLOBAL__N_118parse_expr_primaryINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    $$4 = $29;
    break L1;
    break;
   }
   default: {
    $30 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2)|0);
    $$4 = $30;
    break L1;
   }
   }
  }
 } while(0);
 return ($$4|0);
}
function __ZNKSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE8max_sizeEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 268435455;
}
function __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEEC2EjjSB_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = $3;
 $6 = ($1|0)==(0);
 if ($6) {
  $8 = 0;
 } else {
  $7 = (__ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EE8allocateEj($3,$1)|0);
  $8 = $7;
 }
 HEAP32[$0>>2] = $8;
 $9 = (($8) + ($2<<4)|0);
 $10 = ((($0)) + 8|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($0)) + 4|0);
 HEAP32[$11>>2] = $9;
 $12 = (($8) + ($1<<4)|0);
 $13 = ((($0)) + 12|0);
 HEAP32[$13>>2] = $12;
 return;
}
function __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS8_RS9_EE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 4|0);
 $$0$i = $4;
 while(1) {
  $6 = ($$0$i|0)==($2|0);
  if ($6) {
   break;
  }
  $7 = HEAP32[$5>>2]|0;
  $8 = ((($7)) + -16|0);
  $9 = ((($$0$i)) + -16|0);
  $10 = ((($$0$i)) + -4|0);
  HEAP32[$8>>2] = 0;
  $11 = ((($7)) + -12|0);
  HEAP32[$11>>2] = 0;
  $12 = ((($7)) + -8|0);
  $13 = HEAP32[$10>>2]|0;
  HEAP32[$12>>2] = 0;
  $14 = ((($7)) + -4|0);
  HEAP32[$14>>2] = $13;
  $15 = HEAP32[$9>>2]|0;
  HEAP32[$8>>2] = $15;
  $16 = ((($$0$i)) + -12|0);
  $17 = HEAP32[$16>>2]|0;
  HEAP32[$11>>2] = $17;
  $18 = ((($$0$i)) + -8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($7)) + -8|0);
  HEAP32[$20>>2] = $19;
  HEAP32[$18>>2] = 0;
  HEAP32[$16>>2] = 0;
  HEAP32[$9>>2] = 0;
  $21 = HEAP32[$5>>2]|0;
  $22 = ((($21)) + -16|0);
  HEAP32[$5>>2] = $22;
  $$0$i = $9;
 }
 $23 = HEAP32[$0>>2]|0;
 $24 = HEAP32[$5>>2]|0;
 HEAP32[$0>>2] = $24;
 HEAP32[$5>>2] = $23;
 $25 = ((($1)) + 8|0);
 $26 = HEAP32[$3>>2]|0;
 $27 = HEAP32[$25>>2]|0;
 HEAP32[$3>>2] = $27;
 HEAP32[$25>>2] = $26;
 $28 = ((($0)) + 8|0);
 $29 = ((($1)) + 12|0);
 $30 = HEAP32[$28>>2]|0;
 $31 = HEAP32[$29>>2]|0;
 HEAP32[$28>>2] = $31;
 HEAP32[$29>>2] = $30;
 $32 = HEAP32[$5>>2]|0;
 HEAP32[$1>>2] = $32;
 return;
}
function __ZNSt3__214__split_bufferINS_6vectorINS1_IN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEEERNS5_IS9_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 8|0);
 while(1) {
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==($2|0);
  if ($5) {
   break;
  }
  $6 = ((($4)) + -16|0);
  HEAP32[$3>>2] = $6;
  __ZNSt3__213__vector_baseINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEENS5_IS7_Lj4096EEEED2Ev($6);
 }
 $7 = HEAP32[$0>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  $9 = $7;
  $10 = ((($0)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($0)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) - ($9))|0;
  $15 = $14 >> 4;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EE10deallocateEPS8_j($11,$7,$15);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EE8allocateEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = $1 << 4;
 $4 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8allocateEj($2,$3)|0);
 return ($4|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj3EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = ((($0)) + 11|0);
 HEAP8[$2>>0] = 2;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,160499,2)|0);
 $3 = ((($0)) + 2|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($3,$1);
 $4 = ((($0)) + 12|0);
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($$0$i$i) + 1)|0;
  $$0$i$i = $6;
 }
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7replaceEjjPKcj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$014 = 0, $$069 = 0, $$070 = 0, $$074 = 0, $$215 = 0, $$3 = 0, $$316 = 0, $$373 = 0, $$377 = 0, $$4 = 0, $$sroa$speculated = 0, $$sroa$speculated18 = 0, $$sroa$speculated19 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0;
 var $35 = 0, $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0;
 var $9 = 0, $or$cond = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 $3 = ((($0)) + 11|0);
 $4 = HEAP8[$3>>0]|0;
 $5 = ($4<<24>>24)<(0);
 if ($5) {
  $8 = ((($0)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $10 = ($9>>>0)<(11);
  $$sroa$speculated18 = $10 ? $9 : 11;
  $11 = ((($0)) + 8|0);
  $12 = HEAP32[$11>>2]|0;
  $13 = $12 & 2147483647;
  $phitmp$i = (($13) + -1)|0;
  $$sroa$speculated19 = $$sroa$speculated18;$15 = $9;$17 = $phitmp$i;
 } else {
  $6 = $4&255;
  $7 = ($6>>>0)<(11);
  $$sroa$speculated = $7 ? $6 : 11;
  $$sroa$speculated19 = $$sroa$speculated;$15 = $6;$17 = 10;
 }
 $14 = (($$sroa$speculated19) - ($15))|0;
 $16 = (($14) + ($17))|0;
 $18 = ($16>>>0)<($1>>>0);
 if ($18) {
  $47 = (($15) + ($1))|0;
  $48 = (($47) - ($$sroa$speculated19))|0;
  $49 = (($48) - ($17))|0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$17,$49,$15,0,$$sroa$speculated19,$1,160502);
 } else {
  if ($5) {
   $19 = HEAP32[$0>>2]|0;
   $25 = $19;
  } else {
   $25 = $0;
  }
  $20 = ($$sroa$speculated19|0)==($1|0);
  do {
   if ($20) {
    $$215 = $1;$$3 = 0;$$373 = 160502;$$377 = $1;
    label = 17;
   } else {
    $21 = (($15) - ($$sroa$speculated19))|0;
    $22 = ($21|0)==(0);
    if ($22) {
     $$215 = $$sroa$speculated19;$$3 = 0;$$373 = 160502;$$377 = $1;
     label = 17;
    } else {
     $23 = ($$sroa$speculated19>>>0)>($1>>>0);
     if ($23) {
      (__ZNSt3__211char_traitsIcE4moveEPcPKcj($25,160502,$1)|0);
      $37 = (($25) + ($1)|0);
      $38 = (($25) + ($$sroa$speculated19)|0);
      (__ZNSt3__211char_traitsIcE4moveEPcPKcj($37,$38,$21)|0);
      $$316 = $$sroa$speculated19;$$4 = $1;
      break;
     }
     $24 = ($25>>>0)<(160502>>>0);
     $26 = (($25) + ($15)|0);
     $27 = ($26>>>0)>(160502>>>0);
     $or$cond = $24 & $27;
     do {
      if ($or$cond) {
       $28 = (($25) + ($$sroa$speculated19)|0);
       $29 = ($28>>>0)>(160502>>>0);
       if ($29) {
        (__ZNSt3__211char_traitsIcE4moveEPcPKcj($25,160502,$$sroa$speculated19)|0);
        $32 = (160502 + ($1)|0);
        $33 = (($1) - ($$sroa$speculated19))|0;
        $$014 = 0;$$069 = $$sroa$speculated19;$$070 = $32;$$074 = $33;
        break;
       } else {
        $30 = (($1) - ($$sroa$speculated19))|0;
        $31 = (160502 + ($30)|0);
        $$014 = $$sroa$speculated19;$$069 = 0;$$070 = $31;$$074 = $1;
        break;
       }
      } else {
       $$014 = $$sroa$speculated19;$$069 = 0;$$070 = 160502;$$074 = $1;
      }
     } while(0);
     $34 = (($25) + ($$069)|0);
     $35 = (($34) + ($$074)|0);
     $36 = (($34) + ($$014)|0);
     (__ZNSt3__211char_traitsIcE4moveEPcPKcj($35,$36,$21)|0);
     $$215 = $$014;$$3 = $$069;$$373 = $$070;$$377 = $$074;
     label = 17;
    }
   }
  } while(0);
  if ((label|0) == 17) {
   $39 = (($25) + ($$3)|0);
   (__ZNSt3__211char_traitsIcE4moveEPcPKcj($39,$$373,$$377)|0);
   $$316 = $$215;$$4 = $$377;
  }
  $40 = (($$4) - ($$316))|0;
  $41 = (($40) + ($15))|0;
  $42 = HEAP8[$3>>0]|0;
  $43 = ($42<<24>>24)<(0);
  if ($43) {
   $44 = ((($0)) + 4|0);
   HEAP32[$44>>2] = $41;
  } else {
   $45 = $41&255;
   HEAP8[$3>>0] = $45;
  }
  $46 = (($25) + ($41)|0);
  HEAP8[$2>>0] = 0;
  __ZNSt3__211char_traitsIcE6assignERcRKc($46,$2);
 }
 STACKTOP = sp;return;
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKcj($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $$0 = 0, $$1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0, $36 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $phitmp$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $4 = sp;
 $5 = ((($0)) + 11|0);
 $6 = HEAP8[$5>>0]|0;
 $7 = ($6<<24>>24)<(0);
 if ($7) {
  $8 = ((($0)) + 4|0);
  $9 = HEAP32[$8>>2]|0;
  $12 = $9;
 } else {
  $10 = $6&255;
  $12 = $10;
 }
 $11 = ($12>>>0)<($1>>>0);
 if ($11) {
  __ZNKSt3__221__basic_string_commonILb1EE20__throw_out_of_rangeEv($0);
  // unreachable;
 }
 if ($7) {
  $13 = ((($0)) + 8|0);
  $14 = HEAP32[$13>>2]|0;
  $15 = $14 & 2147483647;
  $phitmp$i = (($15) + -1)|0;
  $17 = $phitmp$i;
 } else {
  $17 = 10;
 }
 $16 = (($17) - ($12))|0;
 $18 = ($16>>>0)<($3>>>0);
 $19 = (($12) + ($3))|0;
 if ($18) {
  $36 = (($19) - ($17))|0;
  __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE21__grow_by_and_replaceEjjjjjjPKc($0,$17,$36,$12,$1,0,$3,$2);
 } else {
  $20 = ($3|0)==(0);
  if (!($20)) {
   if ($7) {
    $21 = HEAP32[$0>>2]|0;
    $25 = $21;
   } else {
    $25 = $0;
   }
   $22 = (($12) - ($1))|0;
   $23 = ($22|0)==(0);
   $24 = (($25) + ($1)|0);
   if ($23) {
    $$1 = $2;
   } else {
    $26 = (($24) + ($3)|0);
    $27 = ($24>>>0)<=($2>>>0);
    $28 = (($25) + ($12)|0);
    $29 = ($28>>>0)>($2>>>0);
    $or$cond = $27 & $29;
    $30 = (($2) + ($3)|0);
    $$0 = $or$cond ? $30 : $2;
    (__ZNSt3__211char_traitsIcE4moveEPcPKcj($26,$24,$22)|0);
    $$1 = $$0;
   }
   (__ZNSt3__211char_traitsIcE4moveEPcPKcj($24,$$1,$3)|0);
   $31 = HEAP8[$5>>0]|0;
   $32 = ($31<<24>>24)<(0);
   if ($32) {
    $33 = ((($0)) + 4|0);
    HEAP32[$33>>2] = $19;
   } else {
    $34 = $19&255;
    HEAP8[$5>>0] = $34;
   }
   $35 = (($25) + ($19)|0);
   HEAP8[$4>>0] = 0;
   __ZNSt3__211char_traitsIcE6assignERcRKc($35,$4);
  }
 }
 STACKTOP = sp;return ($0|0);
}
function __ZNKSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE8max_sizeEv($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return 268435455;
}
function __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEEC2EjjS9_($0,$1,$2,$3) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 $3 = $3|0;
 var $10 = 0, $11 = 0, $12 = 0, $13 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $4 = ((($0)) + 12|0);
 HEAP32[$4>>2] = 0;
 $5 = ((($0)) + 16|0);
 HEAP32[$5>>2] = $3;
 $6 = ($1|0)==(0);
 if ($6) {
  $8 = 0;
 } else {
  $7 = (__ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EE8allocateEj($3,$1)|0);
  $8 = $7;
 }
 HEAP32[$0>>2] = $8;
 $9 = (($8) + ($2<<4)|0);
 $10 = ((($0)) + 8|0);
 HEAP32[$10>>2] = $9;
 $11 = ((($0)) + 4|0);
 HEAP32[$11>>2] = $9;
 $12 = (($8) + ($1<<4)|0);
 $13 = ((($0)) + 12|0);
 HEAP32[$13>>2] = $12;
 return;
}
function __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS6_RS7_EE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0;
 var $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = ((($0)) + 4|0);
 $4 = HEAP32[$3>>2]|0;
 $5 = ((($1)) + 4|0);
 $$0$i = $4;
 while(1) {
  $6 = ($$0$i|0)==($2|0);
  if ($6) {
   break;
  }
  $7 = HEAP32[$5>>2]|0;
  $8 = ((($7)) + -16|0);
  $9 = ((($$0$i)) + -16|0);
  $10 = ((($$0$i)) + -4|0);
  HEAP32[$8>>2] = 0;
  $11 = ((($7)) + -12|0);
  HEAP32[$11>>2] = 0;
  $12 = ((($7)) + -8|0);
  $13 = HEAP32[$10>>2]|0;
  HEAP32[$12>>2] = 0;
  $14 = ((($7)) + -4|0);
  HEAP32[$14>>2] = $13;
  $15 = HEAP32[$9>>2]|0;
  HEAP32[$8>>2] = $15;
  $16 = ((($$0$i)) + -12|0);
  $17 = HEAP32[$16>>2]|0;
  HEAP32[$11>>2] = $17;
  $18 = ((($$0$i)) + -8|0);
  $19 = HEAP32[$18>>2]|0;
  $20 = ((($7)) + -8|0);
  HEAP32[$20>>2] = $19;
  HEAP32[$18>>2] = 0;
  HEAP32[$16>>2] = 0;
  HEAP32[$9>>2] = 0;
  $21 = HEAP32[$5>>2]|0;
  $22 = ((($21)) + -16|0);
  HEAP32[$5>>2] = $22;
  $$0$i = $9;
 }
 $23 = HEAP32[$0>>2]|0;
 $24 = HEAP32[$5>>2]|0;
 HEAP32[$0>>2] = $24;
 HEAP32[$5>>2] = $23;
 $25 = ((($1)) + 8|0);
 $26 = HEAP32[$3>>2]|0;
 $27 = HEAP32[$25>>2]|0;
 HEAP32[$3>>2] = $27;
 HEAP32[$25>>2] = $26;
 $28 = ((($0)) + 8|0);
 $29 = ((($1)) + 12|0);
 $30 = HEAP32[$28>>2]|0;
 $31 = HEAP32[$29>>2]|0;
 HEAP32[$28>>2] = $31;
 HEAP32[$29>>2] = $30;
 $32 = HEAP32[$5>>2]|0;
 HEAP32[$1>>2] = $32;
 return;
}
function __ZNSt3__214__split_bufferINS_6vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS3_11short_allocIS4_Lj4096EEEEERNS5_IS7_Lj4096EEEED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = ((($0)) + 8|0);
 while(1) {
  $4 = HEAP32[$3>>2]|0;
  $5 = ($4|0)==($2|0);
  if ($5) {
   break;
  }
  $6 = ((($4)) + -16|0);
  HEAP32[$3>>2] = $6;
  __ZNSt3__213__vector_baseIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEED2Ev($6);
 }
 $7 = HEAP32[$0>>2]|0;
 $8 = ($7|0)==(0|0);
 if (!($8)) {
  $9 = $7;
  $10 = ((($0)) + 16|0);
  $11 = HEAP32[$10>>2]|0;
  $12 = ((($0)) + 12|0);
  $13 = HEAP32[$12>>2]|0;
  $14 = (($13) - ($9))|0;
  $15 = $14 >> 4;
  __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EE10deallocateEPS6_j($11,$7,$15);
 }
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS0_11string_pairENS1_IS4_Lj4096EEEEELj4096EE8allocateEj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = HEAP32[$0>>2]|0;
 $3 = $1 << 4;
 $4 = (__ZN10__cxxabiv112_GLOBAL__N_15arenaILj4096EE8allocateEj($2,$3)|0);
 return ($4|0);
}
function __ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8allocateEj($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = (__ZNKSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8max_sizeEv($0)|0);
 $2 = ($1|0)==(0);
 if ($2) {
  __ZNKSt3__220__vector_base_commonILb1EE20__throw_length_errorEv($0);
  // unreachable;
 } else {
  $3 = ((($0)) + 12|0);
  $4 = (__ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EE8allocateEj($3,1)|0);
  $5 = ((($0)) + 4|0);
  HEAP32[$5>>2] = $4;
  HEAP32[$0>>2] = $4;
  $6 = ((($4)) + 24|0);
  $7 = ((($0)) + 8|0);
  HEAP32[$7>>2] = $6;
  return;
 }
}
function __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE5eraseEjj($0) {
 $0 = $0|0;
 var $$pre = 0, $$sroa$speculated = 0, $1 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0;
 var $not$ = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 $2 = ((($0)) + 11|0);
 $3 = HEAP8[$2>>0]|0;
 $4 = ($3<<24>>24)<(0);
 if ($4) {
  $5 = ((($0)) + 4|0);
  $6 = HEAP32[$5>>2]|0;
  $7 = HEAP32[$0>>2]|0;
  $13 = $7;$9 = $6;
 } else {
  $8 = $3&255;
  $13 = $0;$9 = $8;
 }
 $not$ = ($9|0)!=(0);
 $$sroa$speculated = $not$&1;
 $10 = (($9) - ($$sroa$speculated))|0;
 $11 = ($10|0)==(0);
 if ($11) {
  $15 = $3;
 } else {
  $12 = (($13) + ($$sroa$speculated)|0);
  (__ZNSt3__211char_traitsIcE4moveEPcPKcj($13,$12,$10)|0);
  $$pre = HEAP8[$2>>0]|0;
  $15 = $$pre;
 }
 $14 = ($15<<24>>24)<(0);
 if ($14) {
  $16 = ((($0)) + 4|0);
  HEAP32[$16>>2] = $10;
 } else {
  $17 = $10&255;
  HEAP8[$2>>0] = $17;
 }
 $18 = (($13) + ($10)|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($18,$1);
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj8EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = ((($0)) + 11|0);
 HEAP8[$2>>0] = 7;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,160275,7)|0);
 $3 = ((($0)) + 7|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($3,$1);
 $4 = ((($0)) + 12|0);
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($$0$i$i) + 1)|0;
  $$0$i$i = $6;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj4EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 3;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,3)|0);
 $4 = ((($0)) + 3|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj19EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = (__ZN10__cxxabiv112_GLOBAL__N_112malloc_allocIcE8allocateEj(32)|0);
 HEAP32[$0>>2] = $2;
 $3 = ((($0)) + 8|0);
 HEAP32[$3>>2] = -2147483616;
 $4 = ((($0)) + 4|0);
 HEAP32[$4>>2] = 18;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($2,160695,18)|0);
 $5 = ((($2)) + 18|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($5,$1);
 $6 = ((($0)) + 12|0);
 ;HEAP32[$6>>2]=0|0;HEAP32[$6+4>>2]=0|0;HEAP32[$6+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $7 = (($6) + ($$0$i$i<<2)|0);
  HEAP32[$7>>2] = 0;
  $8 = (($$0$i$i) + 1)|0;
  $$0$i$i = $8;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj9EEERAT__Kc($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$0$i$i = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $2 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $3 = ((($0)) + 11|0);
 HEAP8[$3>>0] = 8;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,$1,8)|0);
 $4 = ((($0)) + 8|0);
 HEAP8[$2>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($4,$2);
 $5 = ((($0)) + 12|0);
 ;HEAP32[$5>>2]=0|0;HEAP32[$5+4>>2]=0|0;HEAP32[$5+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $6 = (($5) + ($$0$i$i<<2)|0);
  HEAP32[$6>>2] = 0;
  $7 = (($$0$i$i) + 1)|0;
  $$0$i$i = $7;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj7EEERAT__Kc($0) {
 $0 = $0|0;
 var $$0$i$i = 0, $1 = 0, $2 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $exitcond$i$i = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $1 = sp;
 ;HEAP32[$0>>2]=0|0;HEAP32[$0+4>>2]=0|0;HEAP32[$0+8>>2]=0|0;
 $2 = ((($0)) + 11|0);
 HEAP8[$2>>0] = 6;
 (__ZNSt3__211char_traitsIcE4copyEPcPKcj($0,160688,6)|0);
 $3 = ((($0)) + 6|0);
 HEAP8[$1>>0] = 0;
 __ZNSt3__211char_traitsIcE6assignERcRKc($3,$1);
 $4 = ((($0)) + 12|0);
 ;HEAP32[$4>>2]=0|0;HEAP32[$4+4>>2]=0|0;HEAP32[$4+8>>2]=0|0;
 $$0$i$i = 0;
 while(1) {
  $exitcond$i$i = ($$0$i$i|0)==(3);
  if ($exitcond$i$i) {
   break;
  }
  $5 = (($4) + ($$0$i$i<<2)|0);
  HEAP32[$5>>2] = 0;
  $6 = (($$0$i$i) + 1)|0;
  $$0$i$i = $6;
 }
 STACKTOP = sp;return;
}
function __ZN10__cxxabiv112_GLOBAL__N_110save_valueIjEC2ERj($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $1;
 $2 = ((($0)) + 4|0);
 $3 = HEAP32[$1>>2]|0;
 HEAP32[$2>>2] = $3;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbEC2ERb($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $1;
 $2 = ((($0)) + 4|0);
 $3 = HEAP8[$1>>0]|0;
 HEAP8[$2>>0] = $3;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_118parse_special_nameINS0_2DbEEEPKcS4_S4_RT_($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $$0$i$i$i = 0, $$0$i$i$i$i = 0, $$0$i$i$i158 = 0, $$0$i$i$i161 = 0, $$0137 = 0, $$0140 = 0, $$11 = 0, $$1138 = 0, $$1141 = 0, $$2139 = 0, $$2142 = 0, $$4156 = 0, $$8 = 0, $$8151 = 0, $$pre$i = 0, $10 = 0, $100 = 0, $101 = 0, $102 = 0, $103 = 0;
 var $104 = 0, $105 = 0, $106 = 0, $107 = 0, $108 = 0, $109 = 0, $11 = 0, $110 = 0, $111 = 0, $112 = 0, $113 = 0, $114 = 0, $115 = 0, $116 = 0, $117 = 0, $118 = 0, $119 = 0, $12 = 0, $120 = 0, $121 = 0;
 var $122 = 0, $123 = 0, $124 = 0, $125 = 0, $126 = 0, $127 = 0, $128 = 0, $129 = 0, $13 = 0, $130 = 0, $131 = 0, $132 = 0, $133 = 0, $134 = 0, $135 = 0, $136 = 0, $137 = 0, $138 = 0, $139 = 0, $14 = 0;
 var $140 = 0, $141 = 0, $142 = 0, $143 = 0, $144 = 0, $145 = 0, $146 = 0, $147 = 0, $148 = 0, $149 = 0, $15 = 0, $150 = 0, $151 = 0, $152 = 0, $153 = 0, $154 = 0, $155 = 0, $156 = 0, $157 = 0, $16 = 0;
 var $17 = 0, $18 = 0, $19 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $24 = 0, $25 = 0, $26 = 0, $27 = 0, $28 = 0, $29 = 0, $3 = 0, $30 = 0, $31 = 0, $32 = 0, $33 = 0, $34 = 0, $35 = 0;
 var $36 = 0, $37 = 0, $38 = 0, $39 = 0, $4 = 0, $40 = 0, $41 = 0, $42 = 0, $43 = 0, $44 = 0, $45 = 0, $46 = 0, $47 = 0, $48 = 0, $49 = 0, $5 = 0, $50 = 0, $51 = 0, $52 = 0, $53 = 0;
 var $54 = 0, $55 = 0, $56 = 0, $57 = 0, $58 = 0, $59 = 0, $6 = 0, $60 = 0, $61 = 0, $62 = 0, $63 = 0, $64 = 0, $65 = 0, $66 = 0, $67 = 0, $68 = 0, $69 = 0, $7 = 0, $70 = 0, $71 = 0;
 var $72 = 0, $73 = 0, $74 = 0, $75 = 0, $76 = 0, $77 = 0, $78 = 0, $79 = 0, $8 = 0, $80 = 0, $81 = 0, $82 = 0, $83 = 0, $84 = 0, $85 = 0, $86 = 0, $87 = 0, $88 = 0, $89 = 0, $9 = 0;
 var $90 = 0, $91 = 0, $92 = 0, $93 = 0, $94 = 0, $95 = 0, $96 = 0, $97 = 0, $98 = 0, $99 = 0, $exitcond$i$i$i = 0, $exitcond$i$i$i$i = 0, $exitcond$i$i$i159 = 0, $exitcond$i$i$i162 = 0, $or$cond = 0, $switch = 0, $trunc = 0, $trunc$clear = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 64|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(64|0);
 $3 = sp + 61|0;
 $4 = sp + 60|0;
 $5 = sp + 48|0;
 $6 = sp + 36|0;
 $7 = sp + 24|0;
 $8 = sp + 12|0;
 $9 = sp;
 $10 = $1;
 $11 = $0;
 $12 = (($10) - ($11))|0;
 $13 = ($12|0)>(2);
 L1: do {
  if ($13) {
   $14 = HEAP8[$0>>0]|0;
   $15 = $14 << 24 >> 24;
   $16 = ((($0)) + 2|0);
   $17 = ((($0)) + 1|0);
   switch ($15|0) {
   case 84:  {
    break;
   }
   case 71:  {
    $142 = HEAP8[$17>>0]|0;
    $143 = $142 << 24 >> 24;
    switch ($143|0) {
    case 86:  {
     $144 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $145 = ($144|0)==($16|0);
     if ($145) {
      $$11 = $0;
      label = 73;
      break L1;
     }
     $146 = HEAP32[$2>>2]|0;
     $147 = ((($2)) + 4|0);
     $148 = HEAP32[$147>>2]|0;
     $149 = ($146|0)==($148|0);
     if ($149) {
      $$8 = $0;
      break L1;
     }
     $150 = ((($148)) + -24|0);
     (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($150,0,160913)|0);
     $$11 = $144;
     label = 73;
     break L1;
     break;
    }
    case 82:  {
     $151 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
     $152 = ($151|0)==($16|0);
     if ($152) {
      $$11 = $0;
      label = 73;
      break L1;
     }
     $153 = HEAP32[$2>>2]|0;
     $154 = ((($2)) + 4|0);
     $155 = HEAP32[$154>>2]|0;
     $156 = ($153|0)==($155|0);
     if ($156) {
      $$8 = $0;
      break L1;
     }
     $157 = ((($155)) + -24|0);
     (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($157,0,160933)|0);
     $$11 = $151;
     label = 73;
     break L1;
     break;
    }
    default: {
     $$11 = $0;
     label = 73;
     break L1;
    }
    }
    break;
   }
   default: {
    $$11 = $0;
    label = 73;
    break L1;
   }
   }
   $18 = HEAP8[$17>>0]|0;
   $19 = $18 << 24 >> 24;
   switch ($19|0) {
   case 86:  {
    $20 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $21 = ($20|0)==($16|0);
    if ($21) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $22 = HEAP32[$2>>2]|0;
    $23 = ((($2)) + 4|0);
    $24 = HEAP32[$23>>2]|0;
    $25 = ($22|0)==($24|0);
    if ($25) {
     $$8 = $0;
     break L1;
    }
    $26 = ((($24)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($26,0,160762)|0);
    $$11 = $20;
    label = 73;
    break L1;
    break;
   }
   case 84:  {
    $27 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $28 = ($27|0)==($16|0);
    if ($28) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $29 = HEAP32[$2>>2]|0;
    $30 = ((($2)) + 4|0);
    $31 = HEAP32[$30>>2]|0;
    $32 = ($29|0)==($31|0);
    if ($32) {
     $$8 = $0;
     break L1;
    }
    $33 = ((($31)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($33,0,160774)|0);
    $$11 = $27;
    label = 73;
    break L1;
    break;
   }
   case 73:  {
    $34 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $35 = ($34|0)==($16|0);
    if ($35) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $36 = HEAP32[$2>>2]|0;
    $37 = ((($2)) + 4|0);
    $38 = HEAP32[$37>>2]|0;
    $39 = ($36|0)==($38|0);
    if ($39) {
     $$8 = $0;
     break L1;
    }
    $40 = ((($38)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($40,0,160783)|0);
    $$11 = $34;
    label = 73;
    break L1;
    break;
   }
   case 83:  {
    $41 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $42 = ($41|0)==($16|0);
    if ($42) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $43 = HEAP32[$2>>2]|0;
    $44 = ((($2)) + 4|0);
    $45 = HEAP32[$44>>2]|0;
    $46 = ($43|0)==($45|0);
    if ($46) {
     $$8 = $0;
     break L1;
    }
    $47 = ((($45)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($47,0,160797)|0);
    $$11 = $41;
    label = 73;
    break L1;
    break;
   }
   case 99:  {
    $48 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_call_offsetEPKcS2_($16,$1)|0);
    $49 = ($48|0)==($16|0);
    if ($49) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $50 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_call_offsetEPKcS2_($48,$1)|0);
    $51 = ($50|0)==($48|0);
    if ($51) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $52 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($50,$1,$2)|0);
    $53 = ($52|0)==($50|0);
    if ($53) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $54 = HEAP32[$2>>2]|0;
    $55 = ((($2)) + 4|0);
    $56 = HEAP32[$55>>2]|0;
    $57 = ($54|0)==($56|0);
    if ($57) {
     $$8 = $0;
     break L1;
    }
    $58 = ((($56)) + -24|0);
    (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($58,0,160816)|0);
    $$11 = $52;
    label = 73;
    break L1;
    break;
   }
   case 67:  {
    $59 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($16,$1,$2)|0);
    $60 = ($59|0)==($16|0);
    if ($60) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $61 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($59,$1)|0);
    $62 = ($61|0)==($59|0);
    $63 = ($61|0)==($1|0);
    $or$cond = $62 | $63;
    if ($or$cond) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $64 = HEAP8[$61>>0]|0;
    $65 = ($64<<24>>24)==(95);
    if (!($65)) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $66 = ((($61)) + 1|0);
    $67 = (__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_($66,$1,$2)|0);
    $68 = ($67|0)==($66|0);
    if ($68) {
     $$11 = $0;
     label = 73;
     break L1;
    }
    $69 = ((($2)) + 4|0);
    $70 = HEAP32[$69>>2]|0;
    $71 = HEAP32[$2>>2]|0;
    $72 = (($70) - ($71))|0;
    $73 = (($72|0) / 24)&-1;
    $74 = ($73>>>0)<(2);
    if ($74) {
     $$8 = $0;
     break L1;
    }
    $75 = $70;
    $76 = ((($75)) + -24|0);
    __ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv($5,$76);
    $77 = HEAP32[$69>>2]|0;
    $78 = ((($77)) + -24|0);
    $80 = $77;
    while(1) {
     $79 = ($80|0)==($78|0);
     if ($79) {
      break;
     }
     $81 = ((($80)) + -24|0);
     HEAP32[$69>>2] = $81;
     __ZN10__cxxabiv112_GLOBAL__N_111string_pairD2Ev($81);
     $$pre$i = HEAP32[$69>>2]|0;
     $80 = $$pre$i;
    }
    __THREW__ = 0;
    $82 = (invoke_iiii(1121,($5|0),0,(160843|0))|0);
    $83 = __THREW__; __THREW__ = 0;
    $84 = $83&1;
    if ($84) {
     $123 = ___cxa_find_matching_catch_2()|0;
     $124 = tempRet0;
     $$2139 = $124;$$2142 = $123;
    } else {
     ;HEAP32[$8>>2]=HEAP32[$5>>2]|0;HEAP32[$8+4>>2]=HEAP32[$5+4>>2]|0;HEAP32[$8+8>>2]=HEAP32[$5+8>>2]|0;
     $$0$i$i$i161 = 0;
     while(1) {
      $exitcond$i$i$i162 = ($$0$i$i$i161|0)==(3);
      if ($exitcond$i$i$i162) {
       break;
      }
      $85 = (($5) + ($$0$i$i$i161<<2)|0);
      HEAP32[$85>>2] = 0;
      $86 = (($$0$i$i$i161) + 1)|0;
      $$0$i$i$i161 = $86;
     }
     __THREW__ = 0;
     $87 = (invoke_iii(1119,($8|0),(160868|0))|0);
     $88 = __THREW__; __THREW__ = 0;
     $89 = $88&1;
     if ($89) {
      $125 = ___cxa_find_matching_catch_2()|0;
      $126 = tempRet0;
      $$1138 = $126;$$1141 = $125;
     } else {
      ;HEAP32[$7>>2]=HEAP32[$8>>2]|0;HEAP32[$7+4>>2]=HEAP32[$8+4>>2]|0;HEAP32[$7+8>>2]=HEAP32[$8+8>>2]|0;
      $$0$i$i$i158 = 0;
      while(1) {
       $exitcond$i$i$i159 = ($$0$i$i$i158|0)==(3);
       if ($exitcond$i$i$i159) {
        break;
       }
       $90 = (($8) + ($$0$i$i$i158<<2)|0);
       HEAP32[$90>>2] = 0;
       $91 = (($$0$i$i$i158) + 1)|0;
       $$0$i$i$i158 = $91;
      }
      $92 = HEAP32[$69>>2]|0;
      $93 = ((($92)) + -24|0);
      __THREW__ = 0;
      invoke_vii(1120,($9|0),($93|0));
      $94 = __THREW__; __THREW__ = 0;
      $95 = $94&1;
      do {
       if ($95) {
        $127 = ___cxa_find_matching_catch_2()|0;
        $128 = tempRet0;
        $$0137 = $128;$$0140 = $127;
       } else {
        $96 = ((($9)) + 11|0);
        $97 = HEAP8[$96>>0]|0;
        $98 = ($97<<24>>24)<(0);
        $99 = HEAP32[$9>>2]|0;
        $100 = $98 ? $99 : $9;
        $101 = ((($9)) + 4|0);
        $102 = HEAP32[$101>>2]|0;
        $103 = $97&255;
        $104 = $98 ? $102 : $103;
        __THREW__ = 0;
        $105 = (invoke_iiii(1112,($7|0),($100|0),($104|0))|0);
        $106 = __THREW__; __THREW__ = 0;
        $107 = $106&1;
        if ($107) {
         $129 = ___cxa_find_matching_catch_2()|0;
         $130 = tempRet0;
         __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
         $$0137 = $130;$$0140 = $129;
         break;
        }
        ;HEAP32[$6>>2]=HEAP32[$7>>2]|0;HEAP32[$6+4>>2]=HEAP32[$7+4>>2]|0;HEAP32[$6+8>>2]=HEAP32[$7+8>>2]|0;
        $$0$i$i$i = 0;
        while(1) {
         $exitcond$i$i$i = ($$0$i$i$i|0)==(3);
         if ($exitcond$i$i$i) {
          break;
         }
         $108 = (($7) + ($$0$i$i$i<<2)|0);
         HEAP32[$108>>2] = 0;
         $109 = (($$0$i$i$i) + 1)|0;
         $$0$i$i$i = $109;
        }
        $110 = HEAP32[$69>>2]|0;
        $111 = ((($110)) + -24|0);
        $112 = ((($111)) + 11|0);
        $113 = HEAP8[$112>>0]|0;
        $114 = ($113<<24>>24)<(0);
        if ($114) {
         $115 = HEAP32[$111>>2]|0;
         HEAP8[$3>>0] = 0;
         __ZNSt3__211char_traitsIcE6assignERcRKc($115,$3);
         $116 = ((($110)) + -20|0);
         HEAP32[$116>>2] = 0;
        } else {
         HEAP8[$4>>0] = 0;
         __ZNSt3__211char_traitsIcE6assignERcRKc($111,$4);
         HEAP8[$112>>0] = 0;
        }
        __THREW__ = 0;
        invoke_vi(1116,($111|0));
        $117 = __THREW__; __THREW__ = 0;
        $118 = $117&1;
        if ($118) {
         $119 = ___cxa_find_matching_catch_3(0|0)|0;
         $120 = tempRet0;
         ___clang_call_terminate($119);
         // unreachable;
        }
        ;HEAP32[$111>>2]=HEAP32[$6>>2]|0;HEAP32[$111+4>>2]=HEAP32[$6+4>>2]|0;HEAP32[$111+8>>2]=HEAP32[$6+8>>2]|0;
        $$0$i$i$i$i = 0;
        while(1) {
         $exitcond$i$i$i$i = ($$0$i$i$i$i|0)==(3);
         if ($exitcond$i$i$i$i) {
          break;
         }
         $121 = (($6) + ($$0$i$i$i$i<<2)|0);
         HEAP32[$121>>2] = 0;
         $122 = (($$0$i$i$i$i) + 1)|0;
         $$0$i$i$i$i = $122;
        }
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($6);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($9);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
        __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
        $$11 = $67;
        label = 73;
        break L1;
       }
      } while(0);
      __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($7);
      $$1138 = $$0137;$$1141 = $$0140;
     }
     __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($8);
     $$2139 = $$1138;$$2142 = $$1141;
    }
    __ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEED2Ev($5);
    ___resumeException($$2142|0);
    // unreachable;
    break;
   }
   default: {
    $131 = (__ZN10__cxxabiv112_GLOBAL__N_117parse_call_offsetEPKcS2_($17,$1)|0);
    $132 = ($131|0)==($17|0);
    do {
     if ($132) {
      $$4156 = 3;$$8151 = $0;
     } else {
      $133 = (__ZN10__cxxabiv112_GLOBAL__N_114parse_encodingINS0_2DbEEEPKcS4_S4_RT_($131,$1,$2)|0);
      $134 = ($133|0)==($131|0);
      if ($134) {
       $$4156 = 0;$$8151 = $0;
      } else {
       $135 = HEAP32[$2>>2]|0;
       $136 = ((($2)) + 4|0);
       $137 = HEAP32[$136>>2]|0;
       $138 = ($135|0)==($137|0);
       if ($138) {
        $$4156 = 1;$$8151 = $0;
       } else {
        $139 = HEAP8[$16>>0]|0;
        $140 = ($139<<24>>24)==(118);
        $141 = ((($137)) + -24|0);
        if ($140) {
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($141,0,160873)|0);
         $$4156 = 0;$$8151 = $133;
         break;
        } else {
         (__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc($141,0,160891)|0);
         $$4156 = 0;$$8151 = $133;
         break;
        }
       }
      }
     }
    } while(0);
    $trunc = $$4156&255;
    $trunc$clear = $trunc & 3;
    switch ($trunc$clear<<24>>24) {
    case 3: case 0:  {
     $$11 = $$8151;
     label = 73;
     break L1;
     break;
    }
    default: {
    }
    }
    $switch = ($$4156|0)==(0);
    if ($switch) {
     $$11 = $$8151;
     label = 73;
     break L1;
    } else {
     $$8 = $0;
     break L1;
    }
   }
   }
  } else {
   $$11 = $0;
   label = 73;
  }
 } while(0);
 if ((label|0) == 73) {
  $$8 = $$11;
 }
 STACKTOP = sp;return ($$8|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_110save_valueIbED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP8[$1>>0]|0;
 $3 = HEAP32[$0>>2]|0;
 HEAP8[$3>>0] = $2;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_110save_valueIjED2Ev($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ((($0)) + 4|0);
 $2 = HEAP32[$1>>2]|0;
 $3 = HEAP32[$0>>2]|0;
 HEAP32[$3>>2] = $2;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_117parse_call_offsetEPKcS2_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $$ = 0, $$3 = 0, $$34 = 0, $10 = 0, $11 = 0, $12 = 0, $13 = 0, $14 = 0, $15 = 0, $16 = 0, $17 = 0, $18 = 0, $19 = 0, $2 = 0, $20 = 0, $21 = 0, $22 = 0, $23 = 0, $3 = 0, $4 = 0;
 var $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, $or$cond = 0, $or$cond32 = 0, $or$cond33 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $2 = ($0|0)==($1|0);
 L1: do {
  if ($2) {
   $$3 = $0;
  } else {
   $3 = HEAP8[$0>>0]|0;
   $4 = $3 << 24 >> 24;
   $5 = ((($0)) + 1|0);
   switch ($4|0) {
   case 104:  {
    $6 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($5,$1)|0);
    $7 = ($6|0)==($5|0);
    $8 = ($6|0)==($1|0);
    $or$cond = $7 | $8;
    if ($or$cond) {
     $$3 = $0;
     break L1;
    }
    $9 = HEAP8[$6>>0]|0;
    $10 = ($9<<24>>24)==(95);
    $11 = ((($6)) + 1|0);
    $$ = $10 ? $11 : $0;
    return ($$|0);
    break;
   }
   case 118:  {
    break;
   }
   default: {
    $$3 = $0;
    break L1;
   }
   }
   $12 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($5,$1)|0);
   $13 = ($12|0)==($5|0);
   $14 = ($12|0)==($1|0);
   $or$cond32 = $13 | $14;
   if ($or$cond32) {
    $$3 = $0;
   } else {
    $15 = HEAP8[$12>>0]|0;
    $16 = ($15<<24>>24)==(95);
    if ($16) {
     $17 = ((($12)) + 1|0);
     $18 = (__ZN10__cxxabiv112_GLOBAL__N_112parse_numberEPKcS2_($17,$1)|0);
     $19 = ($18|0)==($17|0);
     $20 = ($18|0)==($1|0);
     $or$cond33 = $19 | $20;
     if ($or$cond33) {
      $$3 = $0;
     } else {
      $21 = HEAP8[$18>>0]|0;
      $22 = ($21<<24>>24)==(95);
      $23 = ((($18)) + 1|0);
      $$34 = $22 ? $23 : $0;
      $$3 = $$34;
     }
    } else {
     $$3 = $0;
    }
   }
  }
 } while(0);
 return ($$3|0);
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINS0_11string_pairELj4096EEC2ERNS0_5arenaILj4096EEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $1;
 return;
}
function __ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEC2EjRKS6_RKS7_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 8|0);
 $4 = HEAP32[$1>>2]|0;
 HEAP32[$3>>2] = 0;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = $4;
 return;
}
function __ZN10__cxxabiv112_GLOBAL__N_111short_allocINSt3__26vectorINS3_INS0_11string_pairENS1_IS4_Lj4096EEEEENS1_IS6_Lj4096EEEEELj4096EEC2ERNS0_5arenaILj4096EEE($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = $1;
 return;
}
function __ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEEC2EjRKS8_RKS9_($0,$1) {
 $0 = $0|0;
 $1 = $1|0;
 var $2 = 0, $3 = 0, $4 = 0, $5 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = 0;
 $2 = ((($0)) + 4|0);
 HEAP32[$2>>2] = 0;
 $3 = ((($0)) + 8|0);
 $4 = HEAP32[$1>>2]|0;
 HEAP32[$3>>2] = 0;
 $5 = ((($0)) + 12|0);
 HEAP32[$5>>2] = $4;
 return;
}
function ___cxa_guard_acquire($0) {
 $0 = $0|0;
 var $$0 = 0, $1 = 0, $2 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = HEAP8[$0>>0]|0;
 $2 = ($1<<24>>24)==(1);
 if ($2) {
  $$0 = 0;
 } else {
  HEAP8[$0>>0] = 1;
  $$0 = 1;
 }
 return ($$0|0);
}
function ___cxa_guard_release($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function ___cxa_guard_abort($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 return;
}
function __ZNSt9bad_allocC2Ev($0) {
 $0 = $0|0;
 var label = 0, sp = 0;
 sp = STACKTOP;
 HEAP32[$0>>2] = (22112);
 return;
}
function __ZSt15get_new_handlerv() {
 var $0 = 0, $1 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $0 = HEAP32[41841]|0;HEAP32[41841] = (($0+0)|0);
 $1 = $0;
 return ($1|0);
}
function ___cxa_can_catch($0,$1,$2) {
 $0 = $0|0;
 $1 = $1|0;
 $2 = $2|0;
 var $10 = 0, $3 = 0, $4 = 0, $5 = 0, $6 = 0, $7 = 0, $8 = 0, $9 = 0, label = 0, sp = 0;
 sp = STACKTOP;
 STACKTOP = STACKTOP + 16|0; if ((STACKTOP|0) >= (STACK_MAX|0)) abortStackOverflow(16|0);
 $3 = sp;
 $4 = HEAP32[$2>>2]|0;
 HEAP32[$3>>2] = $4;
 $5 = HEAP32[$0>>2]|0;
 $6 = ((($5)) + 16|0);
 $7 = HEAP32[$6>>2]|0;
 $8 = (FUNCTION_TABLE_iiii[$7 & 2047]($0,$1,$3)|0);
 $9 = $8&1;
 if ($8) {
  $10 = HEAP32[$3>>2]|0;
  HEAP32[$2>>2] = $10;
 }
 STACKTOP = sp;return ($9|0);
}
function ___cxa_is_pointer_type($0) {
 $0 = $0|0;
 var $1 = 0, $2 = 0, $3 = 0, $4 = 0, $phitmp = 0, label = 0, sp = 0;
 sp = STACKTOP;
 $1 = ($0|0)==(0|0);
 if ($1) {
  $4 = 0;
 } else {
  $2 = (___dynamic_cast($0,2440,2592,0)|0);
  $phitmp = ($2|0)!=(0|0);
  $4 = $phitmp;
 }
 $3 = $4&1;
 return ($3|0);
}
function runPostSets() {
}
function _i64Subtract(a, b, c, d) {
    a = a|0; b = b|0; c = c|0; d = d|0;
    var l = 0, h = 0;
    l = (a - c)>>>0;
    h = (b - d)>>>0;
    h = (b - d - (((c>>>0) > (a>>>0))|0))>>>0; // Borrow one from high word to low word on underflow.
    return ((tempRet0 = h,l|0)|0);
}
function _memset(ptr, value, num) {
    ptr = ptr|0; value = value|0; num = num|0;
    var end = 0, aligned_end = 0, block_aligned_end = 0, value4 = 0;
    end = (ptr + num)|0;

    value = value & 0xff;
    if ((num|0) >= 67 /* 64 bytes for an unrolled loop + 3 bytes for unaligned head*/) {
      while ((ptr&3) != 0) {
        HEAP8[((ptr)>>0)]=value;
        ptr = (ptr+1)|0;
      }

      aligned_end = (end & -4)|0;
      block_aligned_end = (aligned_end - 64)|0;
      value4 = value | (value << 8) | (value << 16) | (value << 24);

      while((ptr|0) <= (block_aligned_end|0)) {
        HEAP32[((ptr)>>2)]=value4;
        HEAP32[(((ptr)+(4))>>2)]=value4;
        HEAP32[(((ptr)+(8))>>2)]=value4;
        HEAP32[(((ptr)+(12))>>2)]=value4;
        HEAP32[(((ptr)+(16))>>2)]=value4;
        HEAP32[(((ptr)+(20))>>2)]=value4;
        HEAP32[(((ptr)+(24))>>2)]=value4;
        HEAP32[(((ptr)+(28))>>2)]=value4;
        HEAP32[(((ptr)+(32))>>2)]=value4;
        HEAP32[(((ptr)+(36))>>2)]=value4;
        HEAP32[(((ptr)+(40))>>2)]=value4;
        HEAP32[(((ptr)+(44))>>2)]=value4;
        HEAP32[(((ptr)+(48))>>2)]=value4;
        HEAP32[(((ptr)+(52))>>2)]=value4;
        HEAP32[(((ptr)+(56))>>2)]=value4;
        HEAP32[(((ptr)+(60))>>2)]=value4;
        ptr = (ptr + 64)|0;
      }

      while ((ptr|0) < (aligned_end|0) ) {
        HEAP32[((ptr)>>2)]=value4;
        ptr = (ptr+4)|0;
      }
    }
    // The remaining bytes.
    while ((ptr|0) < (end|0)) {
      HEAP8[((ptr)>>0)]=value;
      ptr = (ptr+1)|0;
    }
    return (end-num)|0;
}
function _pthread_mutex_lock(x) {
    x = x | 0;
    return 0;
}
function _i64Add(a, b, c, d) {
    /*
      x = a + b*2^32
      y = c + d*2^32
      result = l + h*2^32
    */
    a = a|0; b = b|0; c = c|0; d = d|0;
    var l = 0, h = 0;
    l = (a + c)>>>0;
    h = (b + d + (((l>>>0) < (a>>>0))|0))>>>0; // Add carry from low word to high word on overflow.
    return ((tempRet0 = h,l|0)|0);
}
function _bitshift64Ashr(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = high >> bits;
      return (low >>> bits) | ((high&ander) << (32 - bits));
    }
    tempRet0 = (high|0) < 0 ? -1 : 0;
    return (high >> (bits - 32))|0;
}
function _bitshift64Lshr(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = high >>> bits;
      return (low >>> bits) | ((high&ander) << (32 - bits));
    }
    tempRet0 = 0;
    return (high >>> (bits - 32))|0;
}
function _pthread_cond_broadcast(x) {
    x = x | 0;
    return 0;
}
function _memcpy(dest, src, num) {
    dest = dest|0; src = src|0; num = num|0;
    var ret = 0;
    var aligned_dest_end = 0;
    var block_aligned_dest_end = 0;
    var dest_end = 0;
    // Test against a benchmarked cutoff limit for when HEAPU8.set() becomes faster to use.
    if ((num|0) >=
      8192
    ) {
      return _emscripten_memcpy_big(dest|0, src|0, num|0)|0;
    }

    ret = dest|0;
    dest_end = (dest + num)|0;
    if ((dest&3) == (src&3)) {
      // The initial unaligned < 4-byte front.
      while (dest & 3) {
        if ((num|0) == 0) return ret|0;
        HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
        dest = (dest+1)|0;
        src = (src+1)|0;
        num = (num-1)|0;
      }
      aligned_dest_end = (dest_end & -4)|0;
      block_aligned_dest_end = (aligned_dest_end - 64)|0;
      while ((dest|0) <= (block_aligned_dest_end|0) ) {
        HEAP32[((dest)>>2)]=((HEAP32[((src)>>2)])|0);
        HEAP32[(((dest)+(4))>>2)]=((HEAP32[(((src)+(4))>>2)])|0);
        HEAP32[(((dest)+(8))>>2)]=((HEAP32[(((src)+(8))>>2)])|0);
        HEAP32[(((dest)+(12))>>2)]=((HEAP32[(((src)+(12))>>2)])|0);
        HEAP32[(((dest)+(16))>>2)]=((HEAP32[(((src)+(16))>>2)])|0);
        HEAP32[(((dest)+(20))>>2)]=((HEAP32[(((src)+(20))>>2)])|0);
        HEAP32[(((dest)+(24))>>2)]=((HEAP32[(((src)+(24))>>2)])|0);
        HEAP32[(((dest)+(28))>>2)]=((HEAP32[(((src)+(28))>>2)])|0);
        HEAP32[(((dest)+(32))>>2)]=((HEAP32[(((src)+(32))>>2)])|0);
        HEAP32[(((dest)+(36))>>2)]=((HEAP32[(((src)+(36))>>2)])|0);
        HEAP32[(((dest)+(40))>>2)]=((HEAP32[(((src)+(40))>>2)])|0);
        HEAP32[(((dest)+(44))>>2)]=((HEAP32[(((src)+(44))>>2)])|0);
        HEAP32[(((dest)+(48))>>2)]=((HEAP32[(((src)+(48))>>2)])|0);
        HEAP32[(((dest)+(52))>>2)]=((HEAP32[(((src)+(52))>>2)])|0);
        HEAP32[(((dest)+(56))>>2)]=((HEAP32[(((src)+(56))>>2)])|0);
        HEAP32[(((dest)+(60))>>2)]=((HEAP32[(((src)+(60))>>2)])|0);
        dest = (dest+64)|0;
        src = (src+64)|0;
      }
      while ((dest|0) < (aligned_dest_end|0) ) {
        HEAP32[((dest)>>2)]=((HEAP32[((src)>>2)])|0);
        dest = (dest+4)|0;
        src = (src+4)|0;
      }
    } else {
      // In the unaligned copy case, unroll a bit as well.
      aligned_dest_end = (dest_end - 4)|0;
      while ((dest|0) < (aligned_dest_end|0) ) {
        HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
        HEAP8[(((dest)+(1))>>0)]=((HEAP8[(((src)+(1))>>0)])|0);
        HEAP8[(((dest)+(2))>>0)]=((HEAP8[(((src)+(2))>>0)])|0);
        HEAP8[(((dest)+(3))>>0)]=((HEAP8[(((src)+(3))>>0)])|0);
        dest = (dest+4)|0;
        src = (src+4)|0;
      }
    }
    // The remaining unaligned < 4 byte tail.
    while ((dest|0) < (dest_end|0)) {
      HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
      dest = (dest+1)|0;
      src = (src+1)|0;
    }
    return ret|0;
}
function _llvm_cttz_i32(x) {
    x = x|0;
    var ret = 0;
    ret = ((HEAP8[(((cttz_i8)+(x & 0xff))>>0)])|0);
    if ((ret|0) < 8) return ret|0;
    ret = ((HEAP8[(((cttz_i8)+((x >> 8)&0xff))>>0)])|0);
    if ((ret|0) < 8) return (ret + 8)|0;
    ret = ((HEAP8[(((cttz_i8)+((x >> 16)&0xff))>>0)])|0);
    if ((ret|0) < 8) return (ret + 16)|0;
    return (((HEAP8[(((cttz_i8)+(x >>> 24))>>0)])|0) + 24)|0;
}
function ___udivmoddi4($a$0, $a$1, $b$0, $b$1, $rem) {
    $a$0 = $a$0 | 0;
    $a$1 = $a$1 | 0;
    $b$0 = $b$0 | 0;
    $b$1 = $b$1 | 0;
    $rem = $rem | 0;
    var $n_sroa_0_0_extract_trunc = 0, $n_sroa_1_4_extract_shift$0 = 0, $n_sroa_1_4_extract_trunc = 0, $d_sroa_0_0_extract_trunc = 0, $d_sroa_1_4_extract_shift$0 = 0, $d_sroa_1_4_extract_trunc = 0, $4 = 0, $17 = 0, $37 = 0, $49 = 0, $51 = 0, $57 = 0, $58 = 0, $66 = 0, $78 = 0, $86 = 0, $88 = 0, $89 = 0, $91 = 0, $92 = 0, $95 = 0, $105 = 0, $117 = 0, $119 = 0, $125 = 0, $126 = 0, $130 = 0, $q_sroa_1_1_ph = 0, $q_sroa_0_1_ph = 0, $r_sroa_1_1_ph = 0, $r_sroa_0_1_ph = 0, $sr_1_ph = 0, $d_sroa_0_0_insert_insert99$0 = 0, $d_sroa_0_0_insert_insert99$1 = 0, $137$0 = 0, $137$1 = 0, $carry_0203 = 0, $sr_1202 = 0, $r_sroa_0_1201 = 0, $r_sroa_1_1200 = 0, $q_sroa_0_1199 = 0, $q_sroa_1_1198 = 0, $147 = 0, $149 = 0, $r_sroa_0_0_insert_insert42$0 = 0, $r_sroa_0_0_insert_insert42$1 = 0, $150$1 = 0, $151$0 = 0, $152 = 0, $154$0 = 0, $r_sroa_0_0_extract_trunc = 0, $r_sroa_1_4_extract_trunc = 0, $155 = 0, $carry_0_lcssa$0 = 0, $carry_0_lcssa$1 = 0, $r_sroa_0_1_lcssa = 0, $r_sroa_1_1_lcssa = 0, $q_sroa_0_1_lcssa = 0, $q_sroa_1_1_lcssa = 0, $q_sroa_0_0_insert_ext75$0 = 0, $q_sroa_0_0_insert_ext75$1 = 0, $q_sroa_0_0_insert_insert77$1 = 0, $_0$0 = 0, $_0$1 = 0;
    $n_sroa_0_0_extract_trunc = $a$0;
    $n_sroa_1_4_extract_shift$0 = $a$1;
    $n_sroa_1_4_extract_trunc = $n_sroa_1_4_extract_shift$0;
    $d_sroa_0_0_extract_trunc = $b$0;
    $d_sroa_1_4_extract_shift$0 = $b$1;
    $d_sroa_1_4_extract_trunc = $d_sroa_1_4_extract_shift$0;
    if (($n_sroa_1_4_extract_trunc | 0) == 0) {
      $4 = ($rem | 0) != 0;
      if (($d_sroa_1_4_extract_trunc | 0) == 0) {
        if ($4) {
          HEAP32[$rem >> 2] = ($n_sroa_0_0_extract_trunc >>> 0) % ($d_sroa_0_0_extract_trunc >>> 0);
          HEAP32[$rem + 4 >> 2] = 0;
        }
        $_0$1 = 0;
        $_0$0 = ($n_sroa_0_0_extract_trunc >>> 0) / ($d_sroa_0_0_extract_trunc >>> 0) >>> 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      } else {
        if (!$4) {
          $_0$1 = 0;
          $_0$0 = 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        HEAP32[$rem >> 2] = $a$0 & -1;
        HEAP32[$rem + 4 >> 2] = $a$1 & 0;
        $_0$1 = 0;
        $_0$0 = 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      }
    }
    $17 = ($d_sroa_1_4_extract_trunc | 0) == 0;
    do {
      if (($d_sroa_0_0_extract_trunc | 0) == 0) {
        if ($17) {
          if (($rem | 0) != 0) {
            HEAP32[$rem >> 2] = ($n_sroa_1_4_extract_trunc >>> 0) % ($d_sroa_0_0_extract_trunc >>> 0);
            HEAP32[$rem + 4 >> 2] = 0;
          }
          $_0$1 = 0;
          $_0$0 = ($n_sroa_1_4_extract_trunc >>> 0) / ($d_sroa_0_0_extract_trunc >>> 0) >>> 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        if (($n_sroa_0_0_extract_trunc | 0) == 0) {
          if (($rem | 0) != 0) {
            HEAP32[$rem >> 2] = 0;
            HEAP32[$rem + 4 >> 2] = ($n_sroa_1_4_extract_trunc >>> 0) % ($d_sroa_1_4_extract_trunc >>> 0);
          }
          $_0$1 = 0;
          $_0$0 = ($n_sroa_1_4_extract_trunc >>> 0) / ($d_sroa_1_4_extract_trunc >>> 0) >>> 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        $37 = $d_sroa_1_4_extract_trunc - 1 | 0;
        if (($37 & $d_sroa_1_4_extract_trunc | 0) == 0) {
          if (($rem | 0) != 0) {
            HEAP32[$rem >> 2] = 0 | $a$0 & -1;
            HEAP32[$rem + 4 >> 2] = $37 & $n_sroa_1_4_extract_trunc | $a$1 & 0;
          }
          $_0$1 = 0;
          $_0$0 = $n_sroa_1_4_extract_trunc >>> ((_llvm_cttz_i32($d_sroa_1_4_extract_trunc | 0) | 0) >>> 0);
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        $49 = Math_clz32($d_sroa_1_4_extract_trunc | 0) | 0;
        $51 = $49 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
        if ($51 >>> 0 <= 30) {
          $57 = $51 + 1 | 0;
          $58 = 31 - $51 | 0;
          $sr_1_ph = $57;
          $r_sroa_0_1_ph = $n_sroa_1_4_extract_trunc << $58 | $n_sroa_0_0_extract_trunc >>> ($57 >>> 0);
          $r_sroa_1_1_ph = $n_sroa_1_4_extract_trunc >>> ($57 >>> 0);
          $q_sroa_0_1_ph = 0;
          $q_sroa_1_1_ph = $n_sroa_0_0_extract_trunc << $58;
          break;
        }
        if (($rem | 0) == 0) {
          $_0$1 = 0;
          $_0$0 = 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        HEAP32[$rem >> 2] = 0 | $a$0 & -1;
        HEAP32[$rem + 4 >> 2] = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
        $_0$1 = 0;
        $_0$0 = 0;
        return (tempRet0 = $_0$1, $_0$0) | 0;
      } else {
        if (!$17) {
          $117 = Math_clz32($d_sroa_1_4_extract_trunc | 0) | 0;
          $119 = $117 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
          if ($119 >>> 0 <= 31) {
            $125 = $119 + 1 | 0;
            $126 = 31 - $119 | 0;
            $130 = $119 - 31 >> 31;
            $sr_1_ph = $125;
            $r_sroa_0_1_ph = $n_sroa_0_0_extract_trunc >>> ($125 >>> 0) & $130 | $n_sroa_1_4_extract_trunc << $126;
            $r_sroa_1_1_ph = $n_sroa_1_4_extract_trunc >>> ($125 >>> 0) & $130;
            $q_sroa_0_1_ph = 0;
            $q_sroa_1_1_ph = $n_sroa_0_0_extract_trunc << $126;
            break;
          }
          if (($rem | 0) == 0) {
            $_0$1 = 0;
            $_0$0 = 0;
            return (tempRet0 = $_0$1, $_0$0) | 0;
          }
          HEAP32[$rem >> 2] = 0 | $a$0 & -1;
          HEAP32[$rem + 4 >> 2] = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
          $_0$1 = 0;
          $_0$0 = 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
        $66 = $d_sroa_0_0_extract_trunc - 1 | 0;
        if (($66 & $d_sroa_0_0_extract_trunc | 0) != 0) {
          $86 = (Math_clz32($d_sroa_0_0_extract_trunc | 0) | 0) + 33 | 0;
          $88 = $86 - (Math_clz32($n_sroa_1_4_extract_trunc | 0) | 0) | 0;
          $89 = 64 - $88 | 0;
          $91 = 32 - $88 | 0;
          $92 = $91 >> 31;
          $95 = $88 - 32 | 0;
          $105 = $95 >> 31;
          $sr_1_ph = $88;
          $r_sroa_0_1_ph = $91 - 1 >> 31 & $n_sroa_1_4_extract_trunc >>> ($95 >>> 0) | ($n_sroa_1_4_extract_trunc << $91 | $n_sroa_0_0_extract_trunc >>> ($88 >>> 0)) & $105;
          $r_sroa_1_1_ph = $105 & $n_sroa_1_4_extract_trunc >>> ($88 >>> 0);
          $q_sroa_0_1_ph = $n_sroa_0_0_extract_trunc << $89 & $92;
          $q_sroa_1_1_ph = ($n_sroa_1_4_extract_trunc << $89 | $n_sroa_0_0_extract_trunc >>> ($95 >>> 0)) & $92 | $n_sroa_0_0_extract_trunc << $91 & $88 - 33 >> 31;
          break;
        }
        if (($rem | 0) != 0) {
          HEAP32[$rem >> 2] = $66 & $n_sroa_0_0_extract_trunc;
          HEAP32[$rem + 4 >> 2] = 0;
        }
        if (($d_sroa_0_0_extract_trunc | 0) == 1) {
          $_0$1 = $n_sroa_1_4_extract_shift$0 | $a$1 & 0;
          $_0$0 = 0 | $a$0 & -1;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        } else {
          $78 = _llvm_cttz_i32($d_sroa_0_0_extract_trunc | 0) | 0;
          $_0$1 = 0 | $n_sroa_1_4_extract_trunc >>> ($78 >>> 0);
          $_0$0 = $n_sroa_1_4_extract_trunc << 32 - $78 | $n_sroa_0_0_extract_trunc >>> ($78 >>> 0) | 0;
          return (tempRet0 = $_0$1, $_0$0) | 0;
        }
      }
    } while (0);
    if (($sr_1_ph | 0) == 0) {
      $q_sroa_1_1_lcssa = $q_sroa_1_1_ph;
      $q_sroa_0_1_lcssa = $q_sroa_0_1_ph;
      $r_sroa_1_1_lcssa = $r_sroa_1_1_ph;
      $r_sroa_0_1_lcssa = $r_sroa_0_1_ph;
      $carry_0_lcssa$1 = 0;
      $carry_0_lcssa$0 = 0;
    } else {
      $d_sroa_0_0_insert_insert99$0 = 0 | $b$0 & -1;
      $d_sroa_0_0_insert_insert99$1 = $d_sroa_1_4_extract_shift$0 | $b$1 & 0;
      $137$0 = _i64Add($d_sroa_0_0_insert_insert99$0 | 0, $d_sroa_0_0_insert_insert99$1 | 0, -1, -1) | 0;
      $137$1 = tempRet0;
      $q_sroa_1_1198 = $q_sroa_1_1_ph;
      $q_sroa_0_1199 = $q_sroa_0_1_ph;
      $r_sroa_1_1200 = $r_sroa_1_1_ph;
      $r_sroa_0_1201 = $r_sroa_0_1_ph;
      $sr_1202 = $sr_1_ph;
      $carry_0203 = 0;
      while (1) {
        $147 = $q_sroa_0_1199 >>> 31 | $q_sroa_1_1198 << 1;
        $149 = $carry_0203 | $q_sroa_0_1199 << 1;
        $r_sroa_0_0_insert_insert42$0 = 0 | ($r_sroa_0_1201 << 1 | $q_sroa_1_1198 >>> 31);
        $r_sroa_0_0_insert_insert42$1 = $r_sroa_0_1201 >>> 31 | $r_sroa_1_1200 << 1 | 0;
        _i64Subtract($137$0 | 0, $137$1 | 0, $r_sroa_0_0_insert_insert42$0 | 0, $r_sroa_0_0_insert_insert42$1 | 0) | 0;
        $150$1 = tempRet0;
        $151$0 = $150$1 >> 31 | (($150$1 | 0) < 0 ? -1 : 0) << 1;
        $152 = $151$0 & 1;
        $154$0 = _i64Subtract($r_sroa_0_0_insert_insert42$0 | 0, $r_sroa_0_0_insert_insert42$1 | 0, $151$0 & $d_sroa_0_0_insert_insert99$0 | 0, ((($150$1 | 0) < 0 ? -1 : 0) >> 31 | (($150$1 | 0) < 0 ? -1 : 0) << 1) & $d_sroa_0_0_insert_insert99$1 | 0) | 0;
        $r_sroa_0_0_extract_trunc = $154$0;
        $r_sroa_1_4_extract_trunc = tempRet0;
        $155 = $sr_1202 - 1 | 0;
        if (($155 | 0) == 0) {
          break;
        } else {
          $q_sroa_1_1198 = $147;
          $q_sroa_0_1199 = $149;
          $r_sroa_1_1200 = $r_sroa_1_4_extract_trunc;
          $r_sroa_0_1201 = $r_sroa_0_0_extract_trunc;
          $sr_1202 = $155;
          $carry_0203 = $152;
        }
      }
      $q_sroa_1_1_lcssa = $147;
      $q_sroa_0_1_lcssa = $149;
      $r_sroa_1_1_lcssa = $r_sroa_1_4_extract_trunc;
      $r_sroa_0_1_lcssa = $r_sroa_0_0_extract_trunc;
      $carry_0_lcssa$1 = 0;
      $carry_0_lcssa$0 = $152;
    }
    $q_sroa_0_0_insert_ext75$0 = $q_sroa_0_1_lcssa;
    $q_sroa_0_0_insert_ext75$1 = 0;
    $q_sroa_0_0_insert_insert77$1 = $q_sroa_1_1_lcssa | $q_sroa_0_0_insert_ext75$1;
    if (($rem | 0) != 0) {
      HEAP32[$rem >> 2] = 0 | $r_sroa_0_1_lcssa;
      HEAP32[$rem + 4 >> 2] = $r_sroa_1_1_lcssa | 0;
    }
    $_0$1 = (0 | $q_sroa_0_0_insert_ext75$0) >>> 31 | $q_sroa_0_0_insert_insert77$1 << 1 | ($q_sroa_0_0_insert_ext75$1 << 1 | $q_sroa_0_0_insert_ext75$0 >>> 31) & 0 | $carry_0_lcssa$1;
    $_0$0 = ($q_sroa_0_0_insert_ext75$0 << 1 | 0 >>> 31) & -2 | $carry_0_lcssa$0;
    return (tempRet0 = $_0$1, $_0$0) | 0;
}
function ___udivdi3($a$0, $a$1, $b$0, $b$1) {
    $a$0 = $a$0 | 0;
    $a$1 = $a$1 | 0;
    $b$0 = $b$0 | 0;
    $b$1 = $b$1 | 0;
    var $1$0 = 0;
    $1$0 = ___udivmoddi4($a$0, $a$1, $b$0, $b$1, 0) | 0;
    return $1$0 | 0;
}
function ___muldsi3($a, $b) {
    $a = $a | 0;
    $b = $b | 0;
    var $1 = 0, $2 = 0, $3 = 0, $6 = 0, $8 = 0, $11 = 0, $12 = 0;
    $1 = $a & 65535;
    $2 = $b & 65535;
    $3 = Math_imul($2, $1) | 0;
    $6 = $a >>> 16;
    $8 = ($3 >>> 16) + (Math_imul($2, $6) | 0) | 0;
    $11 = $b >>> 16;
    $12 = Math_imul($11, $1) | 0;
    return (tempRet0 = (($8 >>> 16) + (Math_imul($11, $6) | 0) | 0) + ((($8 & 65535) + $12 | 0) >>> 16) | 0, 0 | ($8 + $12 << 16 | $3 & 65535)) | 0;
}
function ___muldi3($a$0, $a$1, $b$0, $b$1) {
    $a$0 = $a$0 | 0;
    $a$1 = $a$1 | 0;
    $b$0 = $b$0 | 0;
    $b$1 = $b$1 | 0;
    var $x_sroa_0_0_extract_trunc = 0, $y_sroa_0_0_extract_trunc = 0, $1$0 = 0, $1$1 = 0, $2 = 0;
    $x_sroa_0_0_extract_trunc = $a$0;
    $y_sroa_0_0_extract_trunc = $b$0;
    $1$0 = ___muldsi3($x_sroa_0_0_extract_trunc, $y_sroa_0_0_extract_trunc) | 0;
    $1$1 = tempRet0;
    $2 = Math_imul($a$1, $y_sroa_0_0_extract_trunc) | 0;
    return (tempRet0 = ((Math_imul($b$1, $x_sroa_0_0_extract_trunc) | 0) + $2 | 0) + $1$1 | $1$1 & 0, 0 | $1$0 & -1) | 0;
}
function _sbrk(increment) {
    increment = increment|0;
    var oldDynamicTop = 0;
    var oldDynamicTopOnChange = 0;
    var newDynamicTop = 0;
    var totalMemory = 0;
    increment = ((increment + 15) & -16)|0;
    oldDynamicTop = HEAP32[DYNAMICTOP_PTR>>2]|0;
    newDynamicTop = oldDynamicTop + increment | 0;

    if (((increment|0) > 0 & (newDynamicTop|0) < (oldDynamicTop|0)) // Detect and fail if we would wrap around signed 32-bit int.
      | (newDynamicTop|0) < 0) { // Also underflow, sbrk() should be able to be used to subtract.
      abortOnCannotGrowMemory()|0;
      ___setErrNo(12);
      return -1;
    }

    HEAP32[DYNAMICTOP_PTR>>2] = newDynamicTop;
    totalMemory = getTotalMemory()|0;
    if ((newDynamicTop|0) > (totalMemory|0)) {
      if ((enlargeMemory()|0) == 0) {
        HEAP32[DYNAMICTOP_PTR>>2] = oldDynamicTop;
        ___setErrNo(12);
        return -1;
      }
    }
    return oldDynamicTop|0;
}
function _bitshift64Shl(low, high, bits) {
    low = low|0; high = high|0; bits = bits|0;
    var ander = 0;
    if ((bits|0) < 32) {
      ander = ((1 << bits) - 1)|0;
      tempRet0 = (high << bits) | ((low&(ander << (32 - bits))) >>> (32 - bits));
      return low << bits;
    }
    tempRet0 = low << (bits - 32);
    return 0;
}
function _memmove(dest, src, num) {
    dest = dest|0; src = src|0; num = num|0;
    var ret = 0;
    if (((src|0) < (dest|0)) & ((dest|0) < ((src + num)|0))) {
      // Unlikely case: Copy backwards in a safe manner
      ret = dest;
      src = (src + num)|0;
      dest = (dest + num)|0;
      while ((num|0) > 0) {
        dest = (dest - 1)|0;
        src = (src - 1)|0;
        num = (num - 1)|0;
        HEAP8[((dest)>>0)]=((HEAP8[((src)>>0)])|0);
      }
      dest = ret;
    } else {
      _memcpy(dest, src, num) | 0;
    }
    return dest | 0;
}
function ___uremdi3($a$0, $a$1, $b$0, $b$1) {
    $a$0 = $a$0 | 0;
    $a$1 = $a$1 | 0;
    $b$0 = $b$0 | 0;
    $b$1 = $b$1 | 0;
    var $rem = 0, __stackBase__ = 0;
    __stackBase__ = STACKTOP;
    STACKTOP = STACKTOP + 16 | 0;
    $rem = __stackBase__ | 0;
    ___udivmoddi4($a$0, $a$1, $b$0, $b$1, $rem) | 0;
    STACKTOP = __stackBase__;
    return (tempRet0 = HEAP32[$rem + 4 >> 2] | 0, HEAP32[$rem >> 2] | 0) | 0;
}
function _pthread_mutex_unlock(x) {
    x = x | 0;
    return 0;
}
function _llvm_bswap_i32(x) {
    x = x|0;
    return (((x&0xff)<<24) | (((x>>8)&0xff)<<16) | (((x>>16)&0xff)<<8) | (x>>>24))|0;
}
function _llvm_bswap_i16(x) {
    x = x|0;
    return (((x&0xff)<<8) | ((x>>8)&0xff))|0;
}

  
function dynCall_iiiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return FUNCTION_TABLE_iiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}


function jsCall_iiiiiiii_0(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_1(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_2(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_3(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_4(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_5(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_6(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_7(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_8(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_9(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_10(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_11(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_12(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_13(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_14(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_15(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_16(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_17(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_18(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function jsCall_iiiiiiii_19(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  return jsCall_iiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0)|0;
}



function dynCall_viiiii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  FUNCTION_TABLE_viiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0);
}


function jsCall_viiiii_0(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_1(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_2(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_3(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_4(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_5(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_6(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_7(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_8(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_9(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_10(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_11(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_12(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_13(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_14(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_15(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_16(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_17(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_18(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function jsCall_viiiii_19(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  jsCall_viiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0);
}



function dynCall_iiiiiid(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return FUNCTION_TABLE_iiiiiid[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}


function jsCall_iiiiiid_0(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(0,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_1(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(1,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_2(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(2,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_3(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(3,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_4(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(4,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_5(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(5,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_6(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(6,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_7(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(7,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_8(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(8,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_9(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(9,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_10(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(10,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_11(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(11,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_12(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(12,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_13(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(13,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_14(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(14,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_15(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(15,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_16(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(16,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_17(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(17,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_18(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(18,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function jsCall_iiiiiid_19(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=+a6;
  return jsCall_iiiiiid(19,a1|0,a2|0,a3|0,a4|0,a5|0,+a6)|0;
}



function dynCall_vi(index,a1) {
  index = index|0;
  a1=a1|0;
  FUNCTION_TABLE_vi[index&2047](a1|0);
}


function jsCall_vi_0(a1) {
  a1=a1|0;
  jsCall_vi(0,a1|0);
}



function jsCall_vi_1(a1) {
  a1=a1|0;
  jsCall_vi(1,a1|0);
}



function jsCall_vi_2(a1) {
  a1=a1|0;
  jsCall_vi(2,a1|0);
}



function jsCall_vi_3(a1) {
  a1=a1|0;
  jsCall_vi(3,a1|0);
}



function jsCall_vi_4(a1) {
  a1=a1|0;
  jsCall_vi(4,a1|0);
}



function jsCall_vi_5(a1) {
  a1=a1|0;
  jsCall_vi(5,a1|0);
}



function jsCall_vi_6(a1) {
  a1=a1|0;
  jsCall_vi(6,a1|0);
}



function jsCall_vi_7(a1) {
  a1=a1|0;
  jsCall_vi(7,a1|0);
}



function jsCall_vi_8(a1) {
  a1=a1|0;
  jsCall_vi(8,a1|0);
}



function jsCall_vi_9(a1) {
  a1=a1|0;
  jsCall_vi(9,a1|0);
}



function jsCall_vi_10(a1) {
  a1=a1|0;
  jsCall_vi(10,a1|0);
}



function jsCall_vi_11(a1) {
  a1=a1|0;
  jsCall_vi(11,a1|0);
}



function jsCall_vi_12(a1) {
  a1=a1|0;
  jsCall_vi(12,a1|0);
}



function jsCall_vi_13(a1) {
  a1=a1|0;
  jsCall_vi(13,a1|0);
}



function jsCall_vi_14(a1) {
  a1=a1|0;
  jsCall_vi(14,a1|0);
}



function jsCall_vi_15(a1) {
  a1=a1|0;
  jsCall_vi(15,a1|0);
}



function jsCall_vi_16(a1) {
  a1=a1|0;
  jsCall_vi(16,a1|0);
}



function jsCall_vi_17(a1) {
  a1=a1|0;
  jsCall_vi(17,a1|0);
}



function jsCall_vi_18(a1) {
  a1=a1|0;
  jsCall_vi(18,a1|0);
}



function jsCall_vi_19(a1) {
  a1=a1|0;
  jsCall_vi(19,a1|0);
}



function dynCall_vii(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=a2|0;
  FUNCTION_TABLE_vii[index&2047](a1|0,a2|0);
}


function jsCall_vii_0(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(0,a1|0,a2|0);
}



function jsCall_vii_1(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(1,a1|0,a2|0);
}



function jsCall_vii_2(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(2,a1|0,a2|0);
}



function jsCall_vii_3(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(3,a1|0,a2|0);
}



function jsCall_vii_4(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(4,a1|0,a2|0);
}



function jsCall_vii_5(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(5,a1|0,a2|0);
}



function jsCall_vii_6(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(6,a1|0,a2|0);
}



function jsCall_vii_7(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(7,a1|0,a2|0);
}



function jsCall_vii_8(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(8,a1|0,a2|0);
}



function jsCall_vii_9(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(9,a1|0,a2|0);
}



function jsCall_vii_10(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(10,a1|0,a2|0);
}



function jsCall_vii_11(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(11,a1|0,a2|0);
}



function jsCall_vii_12(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(12,a1|0,a2|0);
}



function jsCall_vii_13(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(13,a1|0,a2|0);
}



function jsCall_vii_14(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(14,a1|0,a2|0);
}



function jsCall_vii_15(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(15,a1|0,a2|0);
}



function jsCall_vii_16(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(16,a1|0,a2|0);
}



function jsCall_vii_17(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(17,a1|0,a2|0);
}



function jsCall_vii_18(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(18,a1|0,a2|0);
}



function jsCall_vii_19(a1,a2) {
  a1=a1|0; a2=a2|0;
  jsCall_vii(19,a1|0,a2|0);
}



function dynCall_iiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return FUNCTION_TABLE_iiiiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}


function jsCall_iiiiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function jsCall_iiiiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  return jsCall_iiiiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0)|0;
}



function dynCall_ii(index,a1) {
  index = index|0;
  a1=a1|0;
  return FUNCTION_TABLE_ii[index&2047](a1|0)|0;
}


function jsCall_ii_0(a1) {
  a1=a1|0;
  return jsCall_ii(0,a1|0)|0;
}



function jsCall_ii_1(a1) {
  a1=a1|0;
  return jsCall_ii(1,a1|0)|0;
}



function jsCall_ii_2(a1) {
  a1=a1|0;
  return jsCall_ii(2,a1|0)|0;
}



function jsCall_ii_3(a1) {
  a1=a1|0;
  return jsCall_ii(3,a1|0)|0;
}



function jsCall_ii_4(a1) {
  a1=a1|0;
  return jsCall_ii(4,a1|0)|0;
}



function jsCall_ii_5(a1) {
  a1=a1|0;
  return jsCall_ii(5,a1|0)|0;
}



function jsCall_ii_6(a1) {
  a1=a1|0;
  return jsCall_ii(6,a1|0)|0;
}



function jsCall_ii_7(a1) {
  a1=a1|0;
  return jsCall_ii(7,a1|0)|0;
}



function jsCall_ii_8(a1) {
  a1=a1|0;
  return jsCall_ii(8,a1|0)|0;
}



function jsCall_ii_9(a1) {
  a1=a1|0;
  return jsCall_ii(9,a1|0)|0;
}



function jsCall_ii_10(a1) {
  a1=a1|0;
  return jsCall_ii(10,a1|0)|0;
}



function jsCall_ii_11(a1) {
  a1=a1|0;
  return jsCall_ii(11,a1|0)|0;
}



function jsCall_ii_12(a1) {
  a1=a1|0;
  return jsCall_ii(12,a1|0)|0;
}



function jsCall_ii_13(a1) {
  a1=a1|0;
  return jsCall_ii(13,a1|0)|0;
}



function jsCall_ii_14(a1) {
  a1=a1|0;
  return jsCall_ii(14,a1|0)|0;
}



function jsCall_ii_15(a1) {
  a1=a1|0;
  return jsCall_ii(15,a1|0)|0;
}



function jsCall_ii_16(a1) {
  a1=a1|0;
  return jsCall_ii(16,a1|0)|0;
}



function jsCall_ii_17(a1) {
  a1=a1|0;
  return jsCall_ii(17,a1|0)|0;
}



function jsCall_ii_18(a1) {
  a1=a1|0;
  return jsCall_ii(18,a1|0)|0;
}



function jsCall_ii_19(a1) {
  a1=a1|0;
  return jsCall_ii(19,a1|0)|0;
}



function dynCall_iiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return FUNCTION_TABLE_iiiiiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}


function jsCall_iiiiiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function jsCall_iiiiiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0;
  return jsCall_iiiiiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0)|0;
}



function dynCall_iiii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  return FUNCTION_TABLE_iiii[index&2047](a1|0,a2|0,a3|0)|0;
}


function jsCall_iiii_0(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(0,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_1(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(1,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_2(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(2,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_3(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(3,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_4(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(4,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_5(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(5,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_6(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(6,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_7(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(7,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_8(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(8,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_9(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(9,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_10(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(10,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_11(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(11,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_12(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(12,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_13(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(13,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_14(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(14,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_15(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(15,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_16(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(16,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_17(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(17,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_18(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(18,a1|0,a2|0,a3|0)|0;
}



function jsCall_iiii_19(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return jsCall_iiii(19,a1|0,a2|0,a3|0)|0;
}



function dynCall_viiiiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  FUNCTION_TABLE_viiiiiiiiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}


function jsCall_viiiiiiiiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function jsCall_viiiiiiiiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0; a13=a13|0; a14=a14|0; a15=a15|0;
  jsCall_viiiiiiiiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0,a13|0,a14|0,a15|0);
}



function dynCall_viiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  FUNCTION_TABLE_viiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}


function jsCall_viiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function jsCall_viiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  jsCall_viiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0);
}



function dynCall_viiiiii(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  FUNCTION_TABLE_viiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}


function jsCall_viiiiii_0(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_1(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_2(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_3(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_4(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_5(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_6(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_7(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_8(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_9(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_10(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_11(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_12(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_13(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_14(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_15(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_16(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_17(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_18(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function jsCall_viiiiii_19(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  jsCall_viiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0);
}



function dynCall_iiiiiii(index,a1,a2,a3,a4,a5,a6) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return FUNCTION_TABLE_iiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}


function jsCall_iiiiiii_0(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_1(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_2(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_3(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_4(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_5(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_6(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_7(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_8(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_9(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_10(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_11(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_12(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_13(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_14(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_15(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_16(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_17(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_18(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function jsCall_iiiiiii_19(a1,a2,a3,a4,a5,a6) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0;
  return jsCall_iiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0)|0;
}



function dynCall_iiiiiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return FUNCTION_TABLE_iiiiiiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}


function jsCall_iiiiiiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function jsCall_iiiiiiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0; a11=a11|0; a12=a12|0;
  return jsCall_iiiiiiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0,a11|0,a12|0)|0;
}



function dynCall_iid(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=+a2;
  return FUNCTION_TABLE_iid[index&1023](a1|0,+a2)|0;
}


function jsCall_iid_0(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(0,a1|0,+a2)|0;
}



function jsCall_iid_1(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(1,a1|0,+a2)|0;
}



function jsCall_iid_2(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(2,a1|0,+a2)|0;
}



function jsCall_iid_3(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(3,a1|0,+a2)|0;
}



function jsCall_iid_4(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(4,a1|0,+a2)|0;
}



function jsCall_iid_5(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(5,a1|0,+a2)|0;
}



function jsCall_iid_6(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(6,a1|0,+a2)|0;
}



function jsCall_iid_7(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(7,a1|0,+a2)|0;
}



function jsCall_iid_8(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(8,a1|0,+a2)|0;
}



function jsCall_iid_9(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(9,a1|0,+a2)|0;
}



function jsCall_iid_10(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(10,a1|0,+a2)|0;
}



function jsCall_iid_11(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(11,a1|0,+a2)|0;
}



function jsCall_iid_12(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(12,a1|0,+a2)|0;
}



function jsCall_iid_13(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(13,a1|0,+a2)|0;
}



function jsCall_iid_14(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(14,a1|0,+a2)|0;
}



function jsCall_iid_15(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(15,a1|0,+a2)|0;
}



function jsCall_iid_16(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(16,a1|0,+a2)|0;
}



function jsCall_iid_17(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(17,a1|0,+a2)|0;
}



function jsCall_iid_18(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(18,a1|0,+a2)|0;
}



function jsCall_iid_19(a1,a2) {
  a1=a1|0; a2=+a2;
  return jsCall_iid(19,a1|0,+a2)|0;
}



function dynCall_viiiiiii(index,a1,a2,a3,a4,a5,a6,a7) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  FUNCTION_TABLE_viiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}


function jsCall_viiiiiii_0(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_1(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_2(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_3(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_4(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_5(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_6(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_7(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_8(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_9(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_10(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_11(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_12(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_13(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_14(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_15(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_16(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_17(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_18(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function jsCall_viiiiiii_19(a1,a2,a3,a4,a5,a6,a7) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0;
  jsCall_viiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0);
}



function dynCall_viiiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  FUNCTION_TABLE_viiiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}


function jsCall_viiiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function jsCall_viiiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0; a9=a9|0; a10=a10|0;
  jsCall_viiiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0,a9|0,a10|0);
}



function dynCall_iii(index,a1,a2) {
  index = index|0;
  a1=a1|0; a2=a2|0;
  return FUNCTION_TABLE_iii[index&2047](a1|0,a2|0)|0;
}


function jsCall_iii_0(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(0,a1|0,a2|0)|0;
}



function jsCall_iii_1(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(1,a1|0,a2|0)|0;
}



function jsCall_iii_2(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(2,a1|0,a2|0)|0;
}



function jsCall_iii_3(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(3,a1|0,a2|0)|0;
}



function jsCall_iii_4(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(4,a1|0,a2|0)|0;
}



function jsCall_iii_5(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(5,a1|0,a2|0)|0;
}



function jsCall_iii_6(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(6,a1|0,a2|0)|0;
}



function jsCall_iii_7(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(7,a1|0,a2|0)|0;
}



function jsCall_iii_8(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(8,a1|0,a2|0)|0;
}



function jsCall_iii_9(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(9,a1|0,a2|0)|0;
}



function jsCall_iii_10(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(10,a1|0,a2|0)|0;
}



function jsCall_iii_11(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(11,a1|0,a2|0)|0;
}



function jsCall_iii_12(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(12,a1|0,a2|0)|0;
}



function jsCall_iii_13(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(13,a1|0,a2|0)|0;
}



function jsCall_iii_14(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(14,a1|0,a2|0)|0;
}



function jsCall_iii_15(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(15,a1|0,a2|0)|0;
}



function jsCall_iii_16(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(16,a1|0,a2|0)|0;
}



function jsCall_iii_17(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(17,a1|0,a2|0)|0;
}



function jsCall_iii_18(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(18,a1|0,a2|0)|0;
}



function jsCall_iii_19(a1,a2) {
  a1=a1|0; a2=a2|0;
  return jsCall_iii(19,a1|0,a2|0)|0;
}



function dynCall_iiiiii(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return FUNCTION_TABLE_iiiiii[index&2047](a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}


function jsCall_iiiiii_0(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_1(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_2(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_3(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_4(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_5(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_6(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_7(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_8(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_9(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_10(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_11(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_12(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_13(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_14(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_15(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_16(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_17(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_18(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function jsCall_iiiiii_19(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0;
  return jsCall_iiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0)|0;
}



function dynCall_diii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +FUNCTION_TABLE_diii[index&1023](a1|0,a2|0,a3|0);
}


function jsCall_diii_0(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(0,a1|0,a2|0,a3|0);
}



function jsCall_diii_1(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(1,a1|0,a2|0,a3|0);
}



function jsCall_diii_2(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(2,a1|0,a2|0,a3|0);
}



function jsCall_diii_3(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(3,a1|0,a2|0,a3|0);
}



function jsCall_diii_4(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(4,a1|0,a2|0,a3|0);
}



function jsCall_diii_5(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(5,a1|0,a2|0,a3|0);
}



function jsCall_diii_6(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(6,a1|0,a2|0,a3|0);
}



function jsCall_diii_7(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(7,a1|0,a2|0,a3|0);
}



function jsCall_diii_8(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(8,a1|0,a2|0,a3|0);
}



function jsCall_diii_9(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(9,a1|0,a2|0,a3|0);
}



function jsCall_diii_10(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(10,a1|0,a2|0,a3|0);
}



function jsCall_diii_11(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(11,a1|0,a2|0,a3|0);
}



function jsCall_diii_12(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(12,a1|0,a2|0,a3|0);
}



function jsCall_diii_13(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(13,a1|0,a2|0,a3|0);
}



function jsCall_diii_14(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(14,a1|0,a2|0,a3|0);
}



function jsCall_diii_15(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(15,a1|0,a2|0,a3|0);
}



function jsCall_diii_16(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(16,a1|0,a2|0,a3|0);
}



function jsCall_diii_17(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(17,a1|0,a2|0,a3|0);
}



function jsCall_diii_18(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(18,a1|0,a2|0,a3|0);
}



function jsCall_diii_19(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  return +jsCall_diii(19,a1|0,a2|0,a3|0);
}



function dynCall_i(index) {
  index = index|0;
  
  return FUNCTION_TABLE_i[index&2047]()|0;
}


function jsCall_i_0() {
  
  return jsCall_i(0)|0;
}



function jsCall_i_1() {
  
  return jsCall_i(1)|0;
}



function jsCall_i_2() {
  
  return jsCall_i(2)|0;
}



function jsCall_i_3() {
  
  return jsCall_i(3)|0;
}



function jsCall_i_4() {
  
  return jsCall_i(4)|0;
}



function jsCall_i_5() {
  
  return jsCall_i(5)|0;
}



function jsCall_i_6() {
  
  return jsCall_i(6)|0;
}



function jsCall_i_7() {
  
  return jsCall_i(7)|0;
}



function jsCall_i_8() {
  
  return jsCall_i(8)|0;
}



function jsCall_i_9() {
  
  return jsCall_i(9)|0;
}



function jsCall_i_10() {
  
  return jsCall_i(10)|0;
}



function jsCall_i_11() {
  
  return jsCall_i(11)|0;
}



function jsCall_i_12() {
  
  return jsCall_i(12)|0;
}



function jsCall_i_13() {
  
  return jsCall_i(13)|0;
}



function jsCall_i_14() {
  
  return jsCall_i(14)|0;
}



function jsCall_i_15() {
  
  return jsCall_i(15)|0;
}



function jsCall_i_16() {
  
  return jsCall_i(16)|0;
}



function jsCall_i_17() {
  
  return jsCall_i(17)|0;
}



function jsCall_i_18() {
  
  return jsCall_i(18)|0;
}



function jsCall_i_19() {
  
  return jsCall_i(19)|0;
}



function dynCall_iiiii(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return FUNCTION_TABLE_iiiii[index&2047](a1|0,a2|0,a3|0,a4|0)|0;
}


function jsCall_iiiii_0(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(0,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_1(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(1,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_2(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(2,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_3(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(3,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_4(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(4,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_5(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(5,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_6(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(6,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_7(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(7,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_8(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(8,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_9(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(9,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_10(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(10,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_11(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(11,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_12(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(12,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_13(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(13,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_14(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(14,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_15(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(15,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_16(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(16,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_17(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(17,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_18(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(18,a1|0,a2|0,a3|0,a4|0)|0;
}



function jsCall_iiiii_19(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  return jsCall_iiiii(19,a1|0,a2|0,a3|0,a4|0)|0;
}



function dynCall_viii(index,a1,a2,a3) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0;
  FUNCTION_TABLE_viii[index&2047](a1|0,a2|0,a3|0);
}


function jsCall_viii_0(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(0,a1|0,a2|0,a3|0);
}



function jsCall_viii_1(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(1,a1|0,a2|0,a3|0);
}



function jsCall_viii_2(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(2,a1|0,a2|0,a3|0);
}



function jsCall_viii_3(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(3,a1|0,a2|0,a3|0);
}



function jsCall_viii_4(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(4,a1|0,a2|0,a3|0);
}



function jsCall_viii_5(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(5,a1|0,a2|0,a3|0);
}



function jsCall_viii_6(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(6,a1|0,a2|0,a3|0);
}



function jsCall_viii_7(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(7,a1|0,a2|0,a3|0);
}



function jsCall_viii_8(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(8,a1|0,a2|0,a3|0);
}



function jsCall_viii_9(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(9,a1|0,a2|0,a3|0);
}



function jsCall_viii_10(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(10,a1|0,a2|0,a3|0);
}



function jsCall_viii_11(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(11,a1|0,a2|0,a3|0);
}



function jsCall_viii_12(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(12,a1|0,a2|0,a3|0);
}



function jsCall_viii_13(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(13,a1|0,a2|0,a3|0);
}



function jsCall_viii_14(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(14,a1|0,a2|0,a3|0);
}



function jsCall_viii_15(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(15,a1|0,a2|0,a3|0);
}



function jsCall_viii_16(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(16,a1|0,a2|0,a3|0);
}



function jsCall_viii_17(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(17,a1|0,a2|0,a3|0);
}



function jsCall_viii_18(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(18,a1|0,a2|0,a3|0);
}



function jsCall_viii_19(a1,a2,a3) {
  a1=a1|0; a2=a2|0; a3=a3|0;
  jsCall_viii(19,a1|0,a2|0,a3|0);
}



function dynCall_v(index) {
  index = index|0;
  
  FUNCTION_TABLE_v[index&2047]();
}


function jsCall_v_0() {
  
  jsCall_v(0);
}



function jsCall_v_1() {
  
  jsCall_v(1);
}



function jsCall_v_2() {
  
  jsCall_v(2);
}



function jsCall_v_3() {
  
  jsCall_v(3);
}



function jsCall_v_4() {
  
  jsCall_v(4);
}



function jsCall_v_5() {
  
  jsCall_v(5);
}



function jsCall_v_6() {
  
  jsCall_v(6);
}



function jsCall_v_7() {
  
  jsCall_v(7);
}



function jsCall_v_8() {
  
  jsCall_v(8);
}



function jsCall_v_9() {
  
  jsCall_v(9);
}



function jsCall_v_10() {
  
  jsCall_v(10);
}



function jsCall_v_11() {
  
  jsCall_v(11);
}



function jsCall_v_12() {
  
  jsCall_v(12);
}



function jsCall_v_13() {
  
  jsCall_v(13);
}



function jsCall_v_14() {
  
  jsCall_v(14);
}



function jsCall_v_15() {
  
  jsCall_v(15);
}



function jsCall_v_16() {
  
  jsCall_v(16);
}



function jsCall_v_17() {
  
  jsCall_v(17);
}



function jsCall_v_18() {
  
  jsCall_v(18);
}



function jsCall_v_19() {
  
  jsCall_v(19);
}



function dynCall_iiiiiiiii(index,a1,a2,a3,a4,a5,a6,a7,a8) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return FUNCTION_TABLE_iiiiiiiii[index&1023](a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}


function jsCall_iiiiiiiii_0(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(0,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_1(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(1,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_2(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(2,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_3(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(3,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_4(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(4,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_5(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(5,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_6(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(6,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_7(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(7,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_8(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(8,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_9(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(9,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_10(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(10,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_11(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(11,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_12(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(12,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_13(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(13,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_14(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(14,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_15(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(15,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_16(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(16,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_17(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(17,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_18(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(18,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function jsCall_iiiiiiiii_19(a1,a2,a3,a4,a5,a6,a7,a8) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=a5|0; a6=a6|0; a7=a7|0; a8=a8|0;
  return jsCall_iiiiiiiii(19,a1|0,a2|0,a3|0,a4|0,a5|0,a6|0,a7|0,a8|0)|0;
}



function dynCall_iiiiid(index,a1,a2,a3,a4,a5) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return FUNCTION_TABLE_iiiiid[index&511](a1|0,a2|0,a3|0,a4|0,+a5)|0;
}


function jsCall_iiiiid_0(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(0,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_1(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(1,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_2(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(2,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_3(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(3,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_4(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(4,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_5(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(5,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_6(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(6,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_7(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(7,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_8(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(8,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_9(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(9,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_10(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(10,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_11(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(11,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_12(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(12,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_13(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(13,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_14(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(14,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_15(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(15,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_16(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(16,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_17(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(17,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_18(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(18,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function jsCall_iiiiid_19(a1,a2,a3,a4,a5) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0; a5=+a5;
  return jsCall_iiiiid(19,a1|0,a2|0,a3|0,a4|0,+a5)|0;
}



function dynCall_viiii(index,a1,a2,a3,a4) {
  index = index|0;
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  FUNCTION_TABLE_viiii[index&2047](a1|0,a2|0,a3|0,a4|0);
}


function jsCall_viiii_0(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(0,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_1(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(1,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_2(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(2,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_3(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(3,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_4(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(4,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_5(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(5,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_6(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(6,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_7(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(7,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_8(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(8,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_9(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(9,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_10(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(10,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_11(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(11,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_12(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(12,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_13(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(13,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_14(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(14,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_15(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(15,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_16(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(16,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_17(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(17,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_18(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(18,a1|0,a2|0,a3|0,a4|0);
}



function jsCall_viiii_19(a1,a2,a3,a4) {
  a1=a1|0; a2=a2|0; a3=a3|0; a4=a4|0;
  jsCall_viiii(19,a1|0,a2|0,a3|0,a4|0);
}


function b0(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; nullFunc_iiiiiiii(0);return 0;
}
function b1(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; nullFunc_viiiii(1);
}
function b2(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = +p5; nullFunc_iiiiiid(2);return 0;
}
function b3(p0) {
 p0 = p0|0; nullFunc_vi(3);
}
function b4(p0,p1) {
 p0 = p0|0;p1 = p1|0; nullFunc_vii(4);
}
function b5(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0; nullFunc_iiiiiiiiiii(5);return 0;
}
function b6(p0) {
 p0 = p0|0; nullFunc_ii(6);return 0;
}
function _time__wrapper(p0) {
 p0 = p0|0; return _time(p0|0)|0;
}
function _localtime__wrapper(p0) {
 p0 = p0|0; return _localtime(p0|0)|0;
}
function _pthread_mutex_unlock__wrapper(p0) {
 p0 = p0|0; return _pthread_mutex_unlock(p0|0)|0;
}
function _pthread_mutex_lock__wrapper(p0) {
 p0 = p0|0; return _pthread_mutex_lock(p0|0)|0;
}
function _pthread_cond_broadcast__wrapper(p0) {
 p0 = p0|0; return _pthread_cond_broadcast(p0|0)|0;
}
function b7(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0; nullFunc_iiiiiiiiiiii(7);return 0;
}
function b8(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_iiii(8);return 0;
}
function b9(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0;p11 = p11|0;p12 = p12|0;p13 = p13|0;p14 = p14|0; nullFunc_viiiiiiiiiiiiiii(9);
}
function b10(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; nullFunc_viiiiiiii(10);
}
function b11(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; nullFunc_viiiiii(11);
}
function b12(p0,p1,p2,p3,p4,p5) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0; nullFunc_iiiiiii(12);return 0;
}
function b13(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0;p10 = p10|0;p11 = p11|0; nullFunc_iiiiiiiiiiiii(13);return 0;
}
function b14(p0,p1) {
 p0 = p0|0;p1 = +p1; nullFunc_iid(14);return 0;
}
function b15(p0,p1,p2,p3,p4,p5,p6) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0; nullFunc_viiiiiii(15);
}
function b16(p0,p1,p2,p3,p4,p5,p6,p7,p8,p9) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0;p8 = p8|0;p9 = p9|0; nullFunc_viiiiiiiiii(16);
}
function b17(p0,p1) {
 p0 = p0|0;p1 = p1|0; nullFunc_iii(17);return 0;
}
function b18(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0; nullFunc_iiiiii(18);return 0;
}
function b19(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_diii(19);return +0;
}
function b20() {
 ; nullFunc_i(20);return 0;
}
function b21(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; nullFunc_iiiii(21);return 0;
}
function _strftime__wrapper(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; return _strftime(p0|0,p1|0,p2|0,p3|0)|0;
}
function b22(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; nullFunc_viii(22);
}
function ___cxa_throw__wrapper(p0,p1,p2) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0; ___cxa_throw(p0|0,p1|0,p2|0);
}
function b23() {
 ; nullFunc_v(23);
}
function ___cxa_pure_virtual__wrapper() {
 ; ___cxa_pure_virtual();
}
function ___cxa_end_catch__wrapper() {
 ; ___cxa_end_catch();
}
function ___cxa_rethrow__wrapper() {
 ; ___cxa_rethrow();
}
function b24(p0,p1,p2,p3,p4,p5,p6,p7) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = p4|0;p5 = p5|0;p6 = p6|0;p7 = p7|0; nullFunc_iiiiiiiii(24);return 0;
}
function b25(p0,p1,p2,p3,p4) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0;p4 = +p4; nullFunc_iiiiid(25);return 0;
}
function b26(p0,p1,p2,p3) {
 p0 = p0|0;p1 = p1|0;p2 = p2|0;p3 = p3|0; nullFunc_viiii(26);
}

// EMSCRIPTEN_END_FUNCS
var FUNCTION_TABLE_iiiiiiii = [b0,b0,jsCall_iiiiiiii_0,b0,jsCall_iiiiiiii_1,b0,jsCall_iiiiiiii_2,b0,jsCall_iiiiiiii_3,b0,jsCall_iiiiiiii_4,b0,jsCall_iiiiiiii_5,b0,jsCall_iiiiiiii_6,b0,jsCall_iiiiiiii_7,b0,jsCall_iiiiiiii_8,b0,jsCall_iiiiiiii_9,b0,jsCall_iiiiiiii_10,b0,jsCall_iiiiiiii_11,b0,jsCall_iiiiiiii_12,b0,jsCall_iiiiiiii_13
,b0,jsCall_iiiiiiii_14,b0,jsCall_iiiiiiii_15,b0,jsCall_iiiiiiii_16,b0,jsCall_iiiiiiii_17,b0,jsCall_iiiiiiii_18,b0,jsCall_iiiiiiii_19,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,__ZNKSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPK2tmcc,b0,b0,__ZNKSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPK2tmcc,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,__ZNKSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe,__ZNKSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIcS3_NS_9allocatorIcEEEE,b0,b0
,__ZNKSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRe,__ZNKSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_bRNS_8ios_baseERjRNS_12basic_stringIwS3_NS_9allocatorIwEEEE,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEPKNS_12basic_stringIcS3_NS_9allocatorIcEEEENS_5ctypeIcEEEET0_RT_SE_SD_SD_RKT1_Rjb,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,__ZNSt3__214__scan_keywordINS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEPKNS_12basic_stringIwS3_NS_9allocatorIwEEEENS_5ctypeIwEEEET0_RT_SE_SD_SD_RKT1_Rjb,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0,b0
,b0,b0,b0,b0,b0];
var FUNCTION_TABLE_viiiii = [b1,b1,jsCall_viiiii_0,b1,jsCall_viiiii_1,b1,jsCall_viiiii_2,b1,jsCall_viiiii_3,b1,jsCall_viiiii_4,b1,jsCall_viiiii_5,b1,jsCall_viiiii_6,b1,jsCall_viiiii_7,b1,jsCall_viiiii_8,b1,jsCall_viiiii_9,b1,jsCall_viiiii_10,b1,jsCall_viiiii_11,b1,jsCall_viiiii_12,b1,jsCall_viiiii_13
,b1,jsCall_viiiii_14,b1,jsCall_viiiii_15,b1,jsCall_viiiii_16,b1,jsCall_viiiii_17,b1,jsCall_viiiii_18,b1,jsCall_viiiii_19,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,__ZN9ExitGames6Photon8Internal8PeerBase28deserializeOperationResponseEPhbih,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,__ZN9ExitGames6Photon8Internal5TPeer4sendEhPhjh,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,__ZNK10__cxxabiv117__class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b1,b1,b1,__ZNK10__cxxabiv120__si_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,__ZNK10__cxxabiv121__vmi_class_type_info16search_below_dstEPNS_19__dynamic_cast_infoEPKvib,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,__ZN9ExitGames6Common6Object3setEPKvhhb,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popCustomHelperERNS0_6ObjectEhsi,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,__ZN9ExitGames6Common6ObjectC2EPKvhhb,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1,b1
,b1,b1,b1,b1,b1];
var FUNCTION_TABLE_iiiiiid = [b2,b2,jsCall_iiiiiid_0,b2,jsCall_iiiiiid_1,b2,jsCall_iiiiiid_2,b2,jsCall_iiiiiid_3,b2,jsCall_iiiiiid_4,b2,jsCall_iiiiiid_5,b2,jsCall_iiiiiid_6,b2,jsCall_iiiiiid_7,b2,jsCall_iiiiiid_8,b2,jsCall_iiiiiid_9,b2,jsCall_iiiiiid_10,b2,jsCall_iiiiiid_11,b2,jsCall_iiiiiid_12,b2,jsCall_iiiiiid_13
,b2,jsCall_iiiiiid_14,b2,jsCall_iiiiiid_15,b2,jsCall_iiiiiid_16,b2,jsCall_iiiiiid_17,b2,jsCall_iiiiiid_18,b2,jsCall_iiiiiid_19,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,__ZNKSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEce,b2,b2,b2,__ZNKSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwe,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2,b2
,b2,b2,b2,b2,b2];
var FUNCTION_TABLE_vi = [b3,b3,jsCall_vi_0,b3,jsCall_vi_1,b3,jsCall_vi_2,b3,jsCall_vi_3,b3,jsCall_vi_4,b3,jsCall_vi_5,b3,jsCall_vi_6,b3,jsCall_vi_7,b3,jsCall_vi_8,b3,jsCall_vi_9,b3,jsCall_vi_10,b3,jsCall_vi_11,b3,jsCall_vi_12,b3,jsCall_vi_13
,b3,jsCall_vi_14,b3,jsCall_vi_15,b3,jsCall_vi_16,b3,jsCall_vi_17,b3,jsCall_vi_18,b3,jsCall_vi_19,b3,__ZN16PhotonListenerEMD1Ev,__ZN16PhotonListenerEMD0Ev,b3,b3,b3,b3,__ZN9ExitGames6Common12BaseListenerD2Ev,__ZN9ExitGames6Common12BaseListenerD0Ev,b3,__ZN9ExitGames6Photon14PhotonListenerD1Ev,__ZN9ExitGames6Photon14PhotonListenerD0Ev,__ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED2Ev,__ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEED0Ev,b3,b3,b3,b3
,b3,b3,b3,b3,__ZN9ExitGames6Common9KeyObjectIhED2Ev,__ZN9ExitGames6Common9KeyObjectIhED0Ev,b3,b3,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIhED2Ev,__ZN9ExitGames6Common11ValueObjectIhED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIsED2Ev,__ZN9ExitGames6Common11ValueObjectIsED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIiED2Ev,__ZN9ExitGames6Common11ValueObjectIiED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIxED2Ev,__ZN9ExitGames6Common11ValueObjectIxED0Ev,b3,b3
,b3,__ZN9ExitGames6Common11ValueObjectIfED2Ev,__ZN9ExitGames6Common11ValueObjectIfED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIdED2Ev,__ZN9ExitGames6Common11ValueObjectIdED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIbED2Ev,__ZN9ExitGames6Common11ValueObjectIbED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectINS0_7JStringEED2Ev,__ZN9ExitGames6Common11ValueObjectINS0_7JStringEED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectINS0_9HashtableEED2Ev,__ZN9ExitGames6Common11ValueObjectINS0_9HashtableEED0Ev,b3,b3,b3,__ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEED2Ev,__ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEED0Ev,b3,b3
,b3,b3,b3,b3,b3,__ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_ED2Ev,__ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_ED0Ev,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEED2Ev,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEED2Ev,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEED2Ev,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPKhED2Ev
,__ZN9ExitGames6Common11ValueObjectIPKhED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPsED2Ev,__ZN9ExitGames6Common11ValueObjectIPsED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPiED2Ev,__ZN9ExitGames6Common11ValueObjectIPiED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPxED2Ev,__ZN9ExitGames6Common11ValueObjectIPxED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPfED2Ev,__ZN9ExitGames6Common11ValueObjectIPfED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPdED2Ev,__ZN9ExitGames6Common11ValueObjectIPdED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEED2Ev
,__ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPbED2Ev,__ZN9ExitGames6Common11ValueObjectIPbED0Ev,b3,b3,b3,__ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEED2Ev,__ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEED0Ev,b3,b3,b3,__ZN9ExitGames6Common9KeyObjectIiED2Ev,__ZN9ExitGames6Common9KeyObjectIiED0Ev,b3,b3,b3,__ZN9ExitGames6Common7JVectorIhED2Ev,__ZN9ExitGames6Common7JVectorIhED0Ev,b3,__ZN9ExitGames6Common7JVectorINS0_6ObjectEED2Ev,__ZN9ExitGames6Common7JVectorINS0_6ObjectEED0Ev,b3,__ZN9ExitGames6Common10ANSIStringD2Ev,__ZN9ExitGames6Common10ANSIStringD0Ev,b3,b3,b3
,b3,b3,__ZN9ExitGames6Common7JStringD2Ev,__ZN9ExitGames6Common7JStringD0Ev,b3,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocatorD2Ev,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocatorD0Ev,b3,b3,b3,b3,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerD2Ev,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerD0Ev,b3,b3,b3,b3,__ZN9ExitGames6Common10UTF8StringD2Ev,__ZN9ExitGames6Common10UTF8StringD0Ev,b3,b3,b3,b3,__ZN9ExitGames6Common14BaseCharStringD2Ev,__ZN9ExitGames6Common14BaseCharStringD0Ev,__ZN9ExitGames6Common4BaseD2Ev,__ZN9ExitGames6Common4BaseD0Ev,__ZN9ExitGames6Common6ObjectD2Ev,__ZN9ExitGames6Common6ObjectD0Ev,b3
,b3,__ZN9ExitGames6Common9HashtableD2Ev,__ZN9ExitGames6Common9HashtableD0Ev,b3,__ZN9ExitGames6Common14DictionaryBaseD2Ev,__ZN9ExitGames6Common14DictionaryBaseD0Ev,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common14DictionaryBase8TypeInfoD2Ev,__ZN9ExitGames6Common14DictionaryBase8TypeInfoD0Ev,b3,__ZN9ExitGames6Common7Helpers26DeSerializerImplementationD0Ev,b3,__ZN9ExitGames6Common7Helpers24SerializerImplementationD2Ev,__ZN9ExitGames6Common7Helpers24SerializerImplementationD0Ev,b3,__ZN9ExitGames6Common10SerializerD2Ev,__ZN9ExitGames6Common10SerializerD0Ev,b3,__ZN9ExitGames6Common8ToStringD2Ev,__ZN9ExitGames6Common8ToStringD0Ev,__ZN9ExitGames6Common11ValueObjectIPhED0Ev,b3,b3,b3
,__ZN9ExitGames6Photon8Internal13SocketConnect7serviceEv,b3,b3,b3,b3,__ZN9ExitGames6Photon8Internal13PhotonConnect7releaseEv,b3,b3,b3,__ZN9ExitGames6Photon8Internal13SocketConnectD2Ev,__ZN9ExitGames6Photon8Internal13SocketConnectD0Ev,b3,__ZN9ExitGames6Photon8Internal13PhotonConnectD2Ev,__ZN9ExitGames6Photon8Internal13PhotonConnectD0Ev,__ZN9ExitGames6Photon8Internal8PeerBaseD2Ev,__ZN9ExitGames6Photon8Internal8PeerBaseD0Ev,b3,b3,b3,__ZN9ExitGames6Photon8Internal8PeerBase12serviceBasicEv,b3,b3,b3,b3,__ZN9ExitGames6Photon8Internal8PeerBase14stopConnectionEv,__ZN9ExitGames6Photon8Internal8PeerBase7cleanupEv,__ZN9ExitGames6Photon8Internal8PeerBase5resetEv,__ZN9ExitGames6Photon8Internal8PeerBase14onDisconnectedEv,b3,b3
,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Photon8Internal8PeerBase12initCallbackEv,__ZN9ExitGames6Photon8Internal9OperationD2Ev,__ZN9ExitGames6Photon8Internal9OperationD0Ev,b3,__ZN9ExitGames6Photon8Internal8TConnectD2Ev,__ZN9ExitGames6Photon8Internal8TConnectD0Ev,b3,b3,b3,__ZN9ExitGames6Photon8Internal5TPeerD2Ev,__ZN9ExitGames6Photon8Internal5TPeerD0Ev,b3,b3,b3,__ZN9ExitGames6Photon8Internal5TPeer10disconnectEv,__ZN9ExitGames6Photon8Internal5TPeer12serviceBasicEv,b3,b3,b3,__ZN9ExitGames6Photon8Internal5TPeer20fetchServerTimestampEv,b3,b3,b3,__ZN9ExitGames6Photon8Internal5TPeer7cleanupEv
,__ZN9ExitGames6Photon8Internal5TPeer5resetEv,b3,b3,__ZN9ExitGames6Photon8Internal5TPeer14clearAllQueuesEv,b3,__ZN9ExitGames6Photon8Internal5TPeer12initCallbackEv,b3,b3,__ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEED2Ev,__ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEED0Ev,b3,__ZN9ExitGames6Photon8Internal19TrafficStatsCounterD2Ev,__ZN9ExitGames6Photon8Internal19TrafficStatsCounterD0Ev,b3,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounterD2Ev,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounterD0Ev,b3,b3,__ZN9ExitGames6Photon21TrafficStatsGameLevelD2Ev,__ZN9ExitGames6Photon21TrafficStatsGameLevelD0Ev,__ZN9ExitGames6Photon12TrafficStatsD2Ev,__ZN9ExitGames6Photon12TrafficStatsD0Ev,__ZN9ExitGames6Photon10PhotonPeerD2Ev,__ZN9ExitGames6Photon10PhotonPeerD0Ev,b3,__ZN9ExitGames6Photon10PhotonPeer10disconnectEv,b3,__ZN9ExitGames6Photon10PhotonPeer12serviceBasicEv,b3,b3
,b3,b3,b3,__ZN9ExitGames6Photon10PhotonPeer20fetchServerTimestampEv,__ZN9ExitGames6Photon10PhotonPeer17resetTrafficStatsEv,__ZN9ExitGames6Photon10PhotonPeer32resetTrafficStatsMaximumCountersEv,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__28ios_baseD2Ev,__ZNSt3__28ios_baseD0Ev,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED2Ev,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED2Ev,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev,__ZNSt3__213basic_istreamIcNS_11char_traitsIcEEED0Ev,__ZTv0_n12_NSt3__213basic_istreamIcNS_11char_traitsIcEEED1Ev,__ZTv0_n12_NSt3__213basic_istreamIcNS_11char_traitsIcEEED0Ev,__ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev,__ZNSt3__213basic_istreamIwNS_11char_traitsIwEEED0Ev,__ZTv0_n12_NSt3__213basic_istreamIwNS_11char_traitsIwEEED1Ev,__ZTv0_n12_NSt3__213basic_istreamIwNS_11char_traitsIwEEED0Ev,__ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev,__ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEED0Ev,__ZTv0_n12_NSt3__213basic_ostreamIcNS_11char_traitsIcEEED1Ev,__ZTv0_n12_NSt3__213basic_ostreamIcNS_11char_traitsIcEEED0Ev,__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev,__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEED0Ev,__ZTv0_n12_NSt3__213basic_ostreamIwNS_11char_traitsIwEEED1Ev
,__ZTv0_n12_NSt3__213basic_ostreamIwNS_11char_traitsIwEEED0Ev,__ZNSt3__214error_categoryD2Ev,__ZNSt3__219__iostream_categoryD0Ev,b3,b3,b3,b3,b3,__ZNSt3__28ios_base7failureD2Ev,__ZNSt3__28ios_base7failureD0Ev,b3,__ZNSt3__211__stdoutbufIwED2Ev,__ZNSt3__211__stdoutbufIwED0Ev,b3,b3,b3,b3,__ZNSt3__211__stdoutbufIcED2Ev,__ZNSt3__211__stdoutbufIcED0Ev,b3,b3,b3,b3,__ZNSt3__210__stdinbufIwED2Ev,__ZNSt3__210__stdinbufIwED0Ev,b3,b3,b3,b3,__ZNSt3__210__stdinbufIcED2Ev
,__ZNSt3__210__stdinbufIcED0Ev,b3,b3,b3,b3,__ZNSt3__27collateIcED2Ev,__ZNSt3__27collateIcED0Ev,__ZNSt3__26locale5facet16__on_zero_sharedEv,b3,b3,b3,__ZNSt3__27collateIwED2Ev,__ZNSt3__27collateIwED0Ev,b3,b3,b3,__ZNSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev
,__ZNSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b3,b3,b3,b3,b3,b3
,b3,b3,__ZNSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,__ZNSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__28time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,__ZNSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__28time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b3,__ZNSt3__210moneypunctIcLb0EED2Ev,__ZNSt3__210moneypunctIcLb0EED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__210moneypunctIcLb1EED2Ev,__ZNSt3__210moneypunctIcLb1EED0Ev,b3,b3,b3,b3,b3,b3,b3
,b3,b3,__ZNSt3__210moneypunctIwLb0EED2Ev,__ZNSt3__210moneypunctIwLb0EED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__210moneypunctIwLb1EED2Ev,__ZNSt3__210moneypunctIwLb1EED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,b3,__ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev
,b3,b3,__ZNSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED2Ev,__ZNSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEED0Ev,b3,b3,__ZNSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED2Ev,__ZNSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEED0Ev,b3,b3,__ZNSt3__28messagesIcED2Ev,__ZNSt3__28messagesIcED0Ev,b3,b3,b3,__ZNSt3__28messagesIwED2Ev,__ZNSt3__28messagesIwED0Ev,b3,b3,b3,__ZNSt3__26locale5facetD2Ev,__ZNSt3__216__narrow_to_utf8ILj32EED0Ev,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__217__widen_from_utf8ILj32EED0Ev
,__ZNSt3__27codecvtIwc11__mbstate_tED2Ev,__ZNSt3__27codecvtIwc11__mbstate_tED0Ev,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__26locale5__impD2Ev,__ZNSt3__26locale5__impD0Ev,__ZNSt3__25ctypeIcED2Ev,__ZNSt3__25ctypeIcED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__28numpunctIcED2Ev,__ZNSt3__28numpunctIcED0Ev,b3,b3,b3,b3,b3,__ZNSt3__28numpunctIwED2Ev,__ZNSt3__28numpunctIwED0Ev
,b3,b3,b3,b3,b3,__ZNSt3__26locale5facetD0Ev,__ZNSt3__25ctypeIwED0Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__27codecvtIcc11__mbstate_tED0Ev,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__27codecvtIDsc11__mbstate_tED0Ev,b3,b3
,b3,b3,b3,b3,b3,__ZNSt3__27codecvtIDic11__mbstate_tED0Ev,__ZNSt3__212system_errorD2Ev,__ZNSt3__212system_errorD0Ev,b3,__ZN10__cxxabiv116__shim_type_infoD2Ev,__ZN10__cxxabiv117__class_type_infoD0Ev,__ZNK10__cxxabiv116__shim_type_info5noop1Ev,__ZNK10__cxxabiv116__shim_type_info5noop2Ev,b3,b3,b3,b3,__ZN10__cxxabiv120__si_class_type_infoD0Ev,b3,b3,b3,__ZNSt9bad_allocD2Ev,__ZNSt9bad_allocD0Ev,b3,__ZNSt11logic_errorD2Ev,__ZNSt11logic_errorD0Ev,b3,__ZNSt13runtime_errorD2Ev,__ZNSt13runtime_errorD0Ev,__ZNSt12length_errorD0Ev
,__ZNSt12out_of_rangeD0Ev,__ZNSt8bad_castD2Ev,__ZNSt8bad_castD0Ev,b3,__ZN10__cxxabiv121__vmi_class_type_infoD0Ev,b3,b3,b3,__ZNSt3__23mapIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMNS_4lessIS4_EENS_9allocatorINS_4pairIKS4_S6_EEEEED2Ev,b3,b3,_free,b3,b3,b3,b3,b3,b3,__ZNSt3__28ios_base33__set_badbit_and_consider_rethrowEv,b3,b3,b3,b3,b3,__ZN9ExitGames6Common6ObjectC2Ev,b3,b3,b3,b3,b3
,__ZN9ExitGames6Common7JVectorIhE17removeAllElementsEv,__ZN9ExitGames6Common16MemoryManagement8Internal9Interface4freeEPv,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common7JVectorINS0_6ObjectEE17removeAllElementsEv
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManagerC2Ev,__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPoolC2Ev,b3,__ZN9ExitGames6Common7Helpers8SpinlockC2Ev,b3,b3,b3,b3,__ZN9ExitGames6Common7Helpers8Spinlock6unlockEv,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,__ZN9ExitGames6Common6LoggerD2Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common9HashtableC2Ev
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Common14DictionaryBaseC2Ev,b3,b3
,b3,b3,b3,b3,b3,_EGBN_free,b3,b3,b3,b3,_EGBN_CTX_free,b3,b3,b3,__ZN9ExitGames6Common10UTF8StringC2Ev,b3,b3,b3,b3,__ZN9ExitGames6Common7Helpers24SerializerImplementationC2Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN9ExitGames6Photon21TrafficStatsGameLevelC2Ev,b3,b3,__ZN9ExitGames6Photon10PhotonPeer14createPeerBaseEv,b3,b3,b3,b3,b3,b3,__ZNSt3__28ios_base4InitD2Ev,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,__ZNSt3__26locale2id6__initEv,__ZNSt3__217__call_once_proxyINS_5tupleIJONS_12_GLOBAL__N_111__fake_bindEEEEEEvPv,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNKSt3__221__basic_string_commonILb1EE20__throw_length_errorEv,b3
,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEED2Ev,___cxx_global_array_dtor_118,___cxx_global_array_dtor_70,___cxx_global_array_dtor,b3,b3,b3,__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEED2Ev,___cxx_global_array_dtor_121,___cxx_global_array_dtor_94,___cxx_global_array_dtor_55,_freelocale,b3,b3,__ZNSt3__212__do_nothingEPv,b3,__ZNSt3__221__throw_runtime_errorEPKc,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZN10__cxxabiv112_GLOBAL__N_19destruct_EPv,b3,b3
,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7reserveEj,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE5eraseEjj,__ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE8allocateEj,b3,b3,b3,b3,__ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ILj22EEERAT__Kc,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3,b3
,b3,b3,b3,b3,b3,b3,b3,b3,b3];
var FUNCTION_TABLE_vii = [b4,b4,jsCall_vii_0,b4,jsCall_vii_1,b4,jsCall_vii_2,b4,jsCall_vii_3,b4,jsCall_vii_4,b4,jsCall_vii_5,b4,jsCall_vii_6,b4,jsCall_vii_7,b4,jsCall_vii_8,b4,jsCall_vii_9,b4,jsCall_vii_10,b4,jsCall_vii_11,b4,jsCall_vii_12,b4,jsCall_vii_13
,b4,jsCall_vii_14,b4,jsCall_vii_15,b4,jsCall_vii_16,b4,jsCall_vii_17,b4,jsCall_vii_18,b4,jsCall_vii_19,b4,b4,b4,b4,__ZN16PhotonListenerEM19onOperationResponseERKN9ExitGames6Photon17OperationResponseE,__ZN16PhotonListenerEM15onStatusChangedEi,__ZN16PhotonListenerEM7onEventERKN9ExitGames6Photon9EventDataE,b4,b4,b4,b4,b4,b4,b4,__ZNK9ExitGames6Common14DictionaryBase12typeToStringEv,b4,b4,b4
,b4,b4,b4,b4,b4,b4,__ZNK9ExitGames6Common8ToString12typeToStringEv,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNK9ExitGames6Common10ANSIStringcvNS0_7JStringEEv
,__ZNK9ExitGames6Common10ANSIString21JStringRepresentationEv,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator15setMaxAllocSizeEj,b4,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator7deallocEPv,b4,b4,b4,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager15setMaxAllocSizeEj,b4,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager7deallocEPv,b4,b4,b4,b4,__ZNK9ExitGames6Common10UTF8StringcvNS0_7JStringEEv,__ZNK9ExitGames6Common10UTF8String21JStringRepresentationEv,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon8Internal8PeerBase9onConnectEi,__ZN9ExitGames6Photon8Internal8PeerBase14onSendCommandsEi,__ZN9ExitGames6Photon8Internal8PeerBase7serviceEb,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,__ZN9ExitGames6Photon8Internal8PeerBase30updateRoundTripTimeAndVarianceEi,b4,__ZN9ExitGames6Photon8Internal8PeerBase15deriveSharedKeyEPKNS0_17OperationResponseE,__ZN9ExitGames6Photon8Internal8PeerBase18readPingOpResponseEPKNS0_17OperationResponseE,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon8Internal5TPeer9onConnectEi,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,__ZN9ExitGames6Photon8Internal5TPeer18readPingOpResponseEPKNS0_17OperationResponseE,b4,b4,__ZN9ExitGames6Photon8Internal5TPeer17generatePingBytesEv,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNK9ExitGames6Photon21TrafficStatsGameLevel18toStringVitalStatsEv,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon10PhotonPeer7serviceEb,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon10PhotonPeer22initUserDataEncryptionERKNS_6Common7JVectorIhEE,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE5imbueERKNS_6localeE,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE5imbueERKNS_6localeE,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNSt3__211__stdoutbufIwE5imbueERKNS_6localeE,b4,b4,b4,b4,b4,__ZNSt3__211__stdoutbufIcE5imbueERKNS_6localeE,b4,b4,b4,b4,b4,__ZNSt3__210__stdinbufIwE5imbueERKNS_6localeE,b4,b4,b4,b4
,b4,__ZNSt3__210__stdinbufIcE5imbueERKNS_6localeE,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNKSt3__210moneypunctIcLb0EE11do_groupingEv,__ZNKSt3__210moneypunctIcLb0EE14do_curr_symbolEv,__ZNKSt3__210moneypunctIcLb0EE16do_positive_signEv,__ZNKSt3__210moneypunctIcLb0EE16do_negative_signEv,b4,__ZNKSt3__210moneypunctIcLb0EE13do_pos_formatEv,__ZNKSt3__210moneypunctIcLb0EE13do_neg_formatEv,b4,b4,b4,b4,__ZNKSt3__210moneypunctIcLb1EE11do_groupingEv,__ZNKSt3__210moneypunctIcLb1EE14do_curr_symbolEv,__ZNKSt3__210moneypunctIcLb1EE16do_positive_signEv,__ZNKSt3__210moneypunctIcLb1EE16do_negative_signEv,b4
,__ZNKSt3__210moneypunctIcLb1EE13do_pos_formatEv,__ZNKSt3__210moneypunctIcLb1EE13do_neg_formatEv,b4,b4,b4,b4,__ZNKSt3__210moneypunctIwLb0EE11do_groupingEv,__ZNKSt3__210moneypunctIwLb0EE14do_curr_symbolEv,__ZNKSt3__210moneypunctIwLb0EE16do_positive_signEv,__ZNKSt3__210moneypunctIwLb0EE16do_negative_signEv,b4,__ZNKSt3__210moneypunctIwLb0EE13do_pos_formatEv,__ZNKSt3__210moneypunctIwLb0EE13do_neg_formatEv,b4,b4,b4,b4,__ZNKSt3__210moneypunctIwLb1EE11do_groupingEv,__ZNKSt3__210moneypunctIwLb1EE14do_curr_symbolEv,__ZNKSt3__210moneypunctIwLb1EE16do_positive_signEv,__ZNKSt3__210moneypunctIwLb1EE16do_negative_signEv,b4,__ZNKSt3__210moneypunctIwLb1EE13do_pos_formatEv,__ZNKSt3__210moneypunctIwLb1EE13do_neg_formatEv,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNKSt3__28messagesIcE8do_closeEi,b4,b4,b4,b4,__ZNKSt3__28messagesIwE8do_closeEi,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNKSt3__28numpunctIcE11do_groupingEv,__ZNKSt3__28numpunctIcE11do_truenameEv,__ZNKSt3__28numpunctIcE12do_falsenameEv,b4,b4
,b4,b4,__ZNKSt3__28numpunctIwE11do_groupingEv,__ZNKSt3__28numpunctIwE11do_truenameEv,__ZNKSt3__28numpunctIwE12do_falsenameEv,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE6sentryC2ERS3_,b4,b4,b4,__ZNKSt3__28ios_base6getlocEv,b4,b4,__ZNSt3__28ios_base5clearEj,b4,b4,__ZN9ExitGames6Common7JStringC2EPKc,b4,b4,b4,b4,__ZN9ExitGames6Common6ObjectC2ERKS1_,__ZN9ExitGames6Common7Helpers11KeyToObject3getENS0_6ObjectE,__ZN9ExitGames6Common7Helpers13ValueToObject3getENS0_6ObjectE,b4,b4
,b4,b4,__ZNSt11logic_errorC2EPKc,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers11KeyToObject3getIhEENS0_6ObjectET_,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers11KeyToObject3getIiEENS0_6ObjectET_,b4,b4,__ZN9ExitGames6Common7JVectorINS0_6ObjectEEC2ERKS3_,b4,b4,__ZN9ExitGames6Common9KeyObjectIhEC2ERKNS0_6ObjectE,b4,__ZN9ExitGames6Common7JVectorIhE10addElementERKh,b4,b4,b4,b4
,b4,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers8TypeName3cutEPKc,b4,__ZN9ExitGames6Common7JStringC2ERKS1_,__ZN9ExitGames6Common7JStringC2Ej,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool7deallocEPv,b4,b4,__ZN9ExitGames6Common7JStringC2EPKw,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common10UTF8StringC2EPKc
,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation15popStringHelperERNS0_7JStringE,b4,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation18popHashTableHelperERNS0_9HashtableE,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common10UTF8StringC2EPKw,b4,b4,__ZN9ExitGames6Common6Object7setDataEPKv,b4,b4,b4,b4,b4,b4,b4,b4
,b4,__ZN9ExitGames6Common14CustomTypeBase10freeObjectEPKvh,b4,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation8popArrayERNS0_6ObjectE,b4,__ZN9ExitGames6Common6Object13setDataNoCopyEPv,__ZN9ExitGames6Common6Object7setTypeEh,b4,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation12popByteArrayERNS0_6ObjectE,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation14popObjectArrayERNS0_6ObjectE,b4,b4,b4,b4,b4,__ZN9ExitGames6Common14DictionaryBase8TypeInfo12typeToStringEh,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNK9ExitGames6Common7JString18UTF8RepresentationEv
,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common7Helpers24SerializerImplementation20extendInternalBufferEi,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon17OperationResponse15setDebugMessageEPKw,b4,b4,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter11countResultEi,b4,b4,b4,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation7popByteERNS0_6ObjectE,b4
,b4,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter10countEventEi,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Common10UTF8StringC2ERKNS0_7JStringE,b4,b4,__ZN9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE10addElementERKS4_,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,__ZN9ExitGames6Photon8Internal8TConnectC2ERNS1_21PhotonConnectListenerE,__ZN9ExitGames6Photon8Internal19TrafficStatsCounterC2Ei,b4,b4,__ZN9ExitGames6Common6Logger11setListenerERKNS0_12BaseListenerE,b4,b4,__ZNSt3__28ios_base16__call_callbacksENS0_5eventE,b4,b4,b4,__ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE6sentryC2ERS3_,b4,__ZNSt3__28ios_base4initEPv,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNSt13runtime_errorC2EPKc,b4,b4,b4,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE9push_backEc,b4,b4,b4,b4,b4,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE7reserveEj,b4,__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE9push_backEw,b4,b4,b4,b4
,__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE7reserveEj,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,__ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEEC2Ej,b4,__ZNSt3__26locale5__imp7installINS_7collateIcEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7collateIwEEEEvPT_,__ZNSt3__26locale5__imp7installINS_5ctypeIcEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_5ctypeIwEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7codecvtIcc11__mbstate_tEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7codecvtIwc11__mbstate_tEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7codecvtIDsc11__mbstate_tEEEEvPT_
,b4,__ZNSt3__26locale5__imp7installINS_7codecvtIDic11__mbstate_tEEEEvPT_,__ZNSt3__26locale5__imp7installINS_8numpunctIcEEEEvPT_,__ZNSt3__26locale5__imp7installINS_8numpunctIwEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_10moneypunctIcLb0EEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_10moneypunctIcLb1EEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_10moneypunctIwLb0EEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_10moneypunctIwLb1EEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_
,b4,__ZNSt3__26locale5__imp7installINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_8messagesIcEEEEvPT_,b4,__ZNSt3__26locale5__imp7installINS_8messagesIwEEEEvPT_,__ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE8allocateEj,__ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE18__construct_at_endEj,__ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE6resizeEj,__ZNSt3__214__split_bufferIPNS_6locale5facetERNS_15__sso_allocatorIS3_Lj28EEEE18__construct_at_endEj,__ZNSt3__26vectorIPNS_6locale5facetENS_15__sso_allocatorIS3_Lj28EEEE26__swap_out_circular_bufferERNS_14__split_bufferIS3_RS5_EE,b4,b4,b4,b4,b4,b4,b4,__ZNSt3__218__libcpp_refstringC2EPKc,b4,b4,__ZNSt13runtime_errorC2ERKNSt3__212basic_stringIcNS0_11char_traitsIcEENS0_9allocatorIcEEEE,b4,b4,b4,_abort_message
,__ZN10__cxxabiv112_GLOBAL__N_12DbC2ILj4096EEERNS0_5arenaIXT_EEE,__ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE24__emplace_back_slow_pathIJRNS2_5arenaILj4096EEEEEEvDpOT_,b4,b4,b4,b4,b4,b4,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE9push_backEc,b4,b4,__ZN10__cxxabiv112_GLOBAL__N_111string_pair9move_fullEv,b4,__ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE21__push_back_slow_pathIS6_EEvOT_,b4,b4,b4,b4,b4,__ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIS3_EEvOT_,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_,__ZN10__cxxabiv112_GLOBAL__N_111string_pairC2ERKS1_,b4,b4,__ZNSt3__26vectorINS0_INS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEEENS4_IS8_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_,b4,__ZNSt3__26vectorINS0_IN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEENS4_IS6_Lj4096EEEE24__emplace_back_slow_pathIJS5_EEEvDpOT_,__ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEE21__push_back_slow_pathIRKS3_EEvOT_,b4,b4
,b4,b4,b4,b4,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6assignEPKc,b4,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEaSERKS7_,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4,b4
,b4,b4,b4,b4,b4,b4,b4,b4,b4];
var FUNCTION_TABLE_iiiiiiiiiii = [b5,b5,jsCall_iiiiiiiiiii_0,b5,jsCall_iiiiiiiiiii_1,b5,jsCall_iiiiiiiiiii_2,b5,jsCall_iiiiiiiiiii_3,b5,jsCall_iiiiiiiiiii_4,b5,jsCall_iiiiiiiiiii_5,b5,jsCall_iiiiiiiiiii_6,b5,jsCall_iiiiiiiiiii_7,b5,jsCall_iiiiiiiiiii_8,b5,jsCall_iiiiiiiiiii_9,b5,jsCall_iiiiiiiiiii_10,b5,jsCall_iiiiiiiiiii_11,b5,jsCall_iiiiiiiiiii_12,b5,jsCall_iiiiiiiiiii_13
,b5,jsCall_iiiiiiiiiii_14,b5,jsCall_iiiiiiiiiii_15,b5,jsCall_iiiiiiiiiii_16,b5,jsCall_iiiiiiiiiii_17,b5,jsCall_iiiiiiiiiii_18,b5,jsCall_iiiiiiiiiii_19,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,__ZNSt3__29__num_getIcE17__stage2_int_loopEciPcRS2_RjcRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_S2_,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,__ZNSt3__29__num_getIwE17__stage2_int_loopEwiPcRS2_RjwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSD_Pw,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5,b5
,b5,b5,b5,b5,b5];
var FUNCTION_TABLE_ii = [b6,b6,jsCall_ii_0,b6,jsCall_ii_1,b6,jsCall_ii_2,b6,jsCall_ii_3,b6,jsCall_ii_4,b6,jsCall_ii_5,b6,jsCall_ii_6,b6,jsCall_ii_7,b6,jsCall_ii_8,b6,jsCall_ii_9,b6,jsCall_ii_10,b6,jsCall_ii_11,b6,jsCall_ii_12,b6,jsCall_ii_13
,b6,jsCall_ii_14,b6,jsCall_ii_15,b6,jsCall_ii_16,b6,jsCall_ii_17,b6,jsCall_ii_18,b6,jsCall_ii_19,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE11getKeyTypesEv,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE13getValueTypesEv
,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE18getValueDimensionsEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE11getKeyTypesEv
,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE13getValueTypesEv,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE18getValueDimensionsEv,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E11getKeyTypesEv,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E13getValueTypesEv,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E18getValueDimensionsEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common10ANSIStringcvPKcEv,b6
,b6,__ZNK9ExitGames6Common10ANSIString4sizeEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common10UTF8StringcvPKcEv,b6,b6,__ZNK9ExitGames6Common10UTF8String4sizeEv,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common14DictionaryBase11getKeyTypesEv,__ZNK9ExitGames6Common14DictionaryBase13getValueTypesEv,__ZNK9ExitGames6Common14DictionaryBase18getValueDimensionsEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,__ZN9ExitGames6Photon8Internal13SocketConnect14stopConnectionEv,b6,__ZN9ExitGames6Photon8Internal13PhotonConnect6retainEv,b6,__ZNK9ExitGames6Photon8Internal13SocketConnect11getBytesOutEv,__ZNK9ExitGames6Photon8Internal13SocketConnect10getBytesInEv,__ZNK9ExitGames6Photon8Internal13SocketConnect12getUsingIPv6Ev,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Photon8Internal8PeerBase12sendAcksOnlyEv,b6,__ZN9ExitGames6Photon8Internal8PeerBase19establishEncryptionEv,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,__ZN9ExitGames6Photon8Internal8PeerBase27opExchangeKeysForEncryptionEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Photon8Internal8TConnect9getSocketEv,__ZN9ExitGames6Photon8Internal8TConnect15checkConnectionEv,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Photon8Internal5TPeer20sendOutgoingCommandsEv,__ZN9ExitGames6Photon8Internal5TPeer12sendAcksOnlyEv,__ZN9ExitGames6Photon8Internal5TPeer24dispatchIncomingCommandsEv,b6,__ZNK9ExitGames6Photon8Internal5TPeer32getIncomingReliableCommandsCountEv,__ZNK9ExitGames6Photon8Internal5TPeer25getQueuedIncomingCommandsEv,__ZNK9ExitGames6Photon8Internal5TPeer25getQueuedOutgoingCommandsEv,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Photon10PhotonPeer20sendOutgoingCommandsEv
,__ZN9ExitGames6Photon10PhotonPeer12sendAcksOnlyEv,__ZN9ExitGames6Photon10PhotonPeer24dispatchIncomingCommandsEv,__ZN9ExitGames6Photon10PhotonPeer19establishEncryptionEv,b6,b6,b6,b6,b6,b6,___stdio_close,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE4syncEv,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9showmanycEv,b6,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9underflowEv,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE5uflowEv,b6,b6
,b6,b6,b6,b6,b6,b6,b6,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE4syncEv,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9showmanycEv,b6,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9underflowEv,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE5uflowEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,__ZNKSt3__219__iostream_category4nameEv,b6,b6,b6,b6,b6,b6,__ZNKSt13runtime_error4whatEv,b6,b6,b6,__ZNSt3__211__stdoutbufIwE4syncEv,b6,b6,b6,b6,b6,__ZNSt3__211__stdoutbufIcE4syncEv,b6,b6,b6,b6,b6,__ZNSt3__210__stdinbufIwE9underflowEv,__ZNSt3__210__stdinbufIwE5uflowEv,b6,b6
,b6,b6,__ZNSt3__210__stdinbufIcE9underflowEv,__ZNSt3__210__stdinbufIcE5uflowEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE13do_date_orderEv,b6,b6,b6,b6,b6,b6,__ZNKSt3__220__time_get_c_storageIcE7__weeksEv,__ZNKSt3__220__time_get_c_storageIcE8__monthsEv,__ZNKSt3__220__time_get_c_storageIcE7__am_pmEv,__ZNKSt3__220__time_get_c_storageIcE3__cEv,__ZNKSt3__220__time_get_c_storageIcE3__rEv,__ZNKSt3__220__time_get_c_storageIcE3__xEv,__ZNKSt3__220__time_get_c_storageIcE3__XEv,b6,b6,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE13do_date_orderEv,b6,b6,b6,b6,b6,b6,__ZNKSt3__220__time_get_c_storageIwE7__weeksEv,__ZNKSt3__220__time_get_c_storageIwE8__monthsEv,__ZNKSt3__220__time_get_c_storageIwE7__am_pmEv
,__ZNKSt3__220__time_get_c_storageIwE3__cEv,__ZNKSt3__220__time_get_c_storageIwE3__rEv,__ZNKSt3__220__time_get_c_storageIwE3__xEv,__ZNKSt3__220__time_get_c_storageIwE3__XEv,b6,b6,b6,b6,b6,b6,b6,b6,__ZNKSt3__210moneypunctIcLb0EE16do_decimal_pointEv,__ZNKSt3__210moneypunctIcLb0EE16do_thousands_sepEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIcLb0EE14do_frac_digitsEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIcLb1EE16do_decimal_pointEv,__ZNKSt3__210moneypunctIcLb1EE16do_thousands_sepEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIcLb1EE14do_frac_digitsEv
,b6,b6,b6,b6,__ZNKSt3__210moneypunctIwLb0EE16do_decimal_pointEv,__ZNKSt3__210moneypunctIwLb0EE16do_thousands_sepEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIwLb0EE14do_frac_digitsEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIwLb1EE16do_decimal_pointEv,__ZNKSt3__210moneypunctIwLb1EE16do_thousands_sepEv,b6,b6,b6,b6,__ZNKSt3__210moneypunctIwLb1EE14do_frac_digitsEv,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNKSt3__27codecvtIDic11__mbstate_tE11do_encodingEv,__ZNKSt3__27codecvtIDic11__mbstate_tE16do_always_noconvEv,b6,__ZNKSt3__27codecvtIDic11__mbstate_tE13do_max_lengthEv,b6
,b6,b6,b6,b6,b6,__ZNKSt3__27codecvtIwc11__mbstate_tE11do_encodingEv,__ZNKSt3__27codecvtIwc11__mbstate_tE16do_always_noconvEv,b6,__ZNKSt3__27codecvtIwc11__mbstate_tE13do_max_lengthEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNKSt3__28numpunctIcE16do_decimal_pointEv,__ZNKSt3__28numpunctIcE16do_thousands_sepEv,b6,b6,b6,b6,b6
,__ZNKSt3__28numpunctIwE16do_decimal_pointEv,__ZNKSt3__28numpunctIwE16do_thousands_sepEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNKSt3__27codecvtIcc11__mbstate_tE11do_encodingEv,__ZNKSt3__27codecvtIcc11__mbstate_tE16do_always_noconvEv,b6,__ZNKSt3__27codecvtIcc11__mbstate_tE13do_max_lengthEv,b6,b6,b6
,b6,__ZNKSt3__27codecvtIDsc11__mbstate_tE11do_encodingEv,__ZNKSt3__27codecvtIDsc11__mbstate_tE16do_always_noconvEv,b6,__ZNKSt3__27codecvtIDsc11__mbstate_tE13do_max_lengthEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNKSt9bad_alloc4whatEv,b6,b6,__ZNKSt11logic_error4whatEv,b6,b6,b6
,b6,b6,b6,__ZNKSt8bad_cast4whatEv,b6,b6,b6,b6,b6,__ZNSt3__24endlIwNS_11char_traitsIwEEEERNS_13basic_ostreamIT_T0_EES7_,b6,b6,_malloc,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_7JStringEEES4_EEPT_RKT0_,__ZN9ExitGames6Common16MemoryManagement8allocateINS0_11ValueObjectINS0_9HashtableEEES4_EEPT_RKT0_,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,__ZNK9ExitGames6Photon17OperationResponse16getOperationCodeEv,__ZNK9ExitGames6Photon17OperationResponse13getParametersEv,_wcslen,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common7JString4cstrEv,b6,b6,__ZNK9ExitGames6Common14DictionaryBase12getHashtableEv,__ZNK9ExitGames6Common9Hashtable7getKeysEv,b6,__ZNK9ExitGames6Common7JVectorINS0_6ObjectEE7getSizeEv,b6,b6,__ZNK9ExitGames6Common9KeyObjectIhE11getDataCopyEv,b6,__ZN9ExitGames6Common16MemoryManagement8Internal9Interface6mallocEj,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Common7Helpers10MathHelper7getLog2Ej,__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool5allocEv,b6,__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool12getBlockSizeEv,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,__ZNK9ExitGames6Common14BaseCharString6lengthEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNK9ExitGames6Common7JString6lengthEv,_time__wrapper,_localtime__wrapper,b6,b6,b6,__ZNK9ExitGames6Common14BaseCharString4cstrEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,__ZNK9ExitGames6Common6Object7getDataEv,b6,b6,__ZNK9ExitGames6Common6Object13getDimensionsEv,__ZNK9ExitGames6Common6Object8getSizesEv,__ZNK9ExitGames6Common6Object7getTypeEv,__ZNK9ExitGames6Common6Object13getCustomTypeEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,_close,b6,b6,b6,b6,b6,__ZNK9ExitGames6Photon16OperationRequest16getOperationCodeEv,__ZNK9ExitGames6Photon16OperationRequest13getParametersEv,__ZNK9ExitGames6Common14DictionaryBase7getSizeEv,b6,b6,b6,b6,__ZNK9ExitGames6Common7Helpers24SerializerImplementation7getDataEv,__ZNK9ExitGames6Common7Helpers24SerializerImplementation13getDataOffsetEv,b6,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation8readByteEv,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation9readShortEv,b6,__ZNK9ExitGames6Common6Object14getDataPointerEv,__ZNK9ExitGames6Common7JStringcvPKwEv,b6,b6,b6,b6,__ZN9ExitGames6Photon8Internal8PeerData11getListenerEv,b6,b6,b6,b6
,b6,b6,__ZNK9ExitGames6Photon9EventData7getCodeEv,b6,b6,b6,b6,b6,b6,b6,b6,_EGBN_num_bits,b6,b6,b6,b6,b6,b6,__ZN9ExitGames6Photon8Internal8PeerData10InitValues22getClientLibIdentifierEb,b6,b6,__ZNK9ExitGames6Common10Serializer7getSizeEv,b6,__ZNK9ExitGames6Common10Serializer7getDataEv,b6,b6,b6,b6,b6,b6
,__ZNK9ExitGames6Photon8Internal8PeerBase16getRoundTripTimeEv,__ZNK9ExitGames6Photon8Internal8PeerBase24getRoundTripTimeVarianceEv,__ZNK9ExitGames6Photon8Internal8PeerBase25getTimestampOfLastReceiveEv,__ZNK9ExitGames6Photon8Internal8PeerData26getTrafficStatsEnabledTimeEv,__ZNK9ExitGames6Photon8Internal8PeerBase24getTrafficStatsGameLevelEv,__ZNK9ExitGames6Photon8Internal8PeerBase23getTrafficStatsIncomingEv,__ZNK9ExitGames6Photon8Internal8PeerBase23getTrafficStatsOutgoingEv,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__ZNSt3__213basic_ostreamIcNS_11char_traitsIcEEE5flushEv,__ZNSt3__213basic_ostreamIwNS_11char_traitsIwEEE5flushEv,b6,b6,b6,b6,b6,b6,___uselocale
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,__Znwj
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,_pthread_mutex_unlock__wrapper,_pthread_mutex_lock__wrapper,_pthread_cond_broadcast__wrapper,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6,b6
,b6,b6,b6,b6,b6,b6,b6,b6,b6];
var FUNCTION_TABLE_iiiiiiiiiiii = [b7,b7,jsCall_iiiiiiiiiiii_0,b7,jsCall_iiiiiiiiiiii_1,b7,jsCall_iiiiiiiiiiii_2,b7,jsCall_iiiiiiiiiiii_3,b7,jsCall_iiiiiiiiiiii_4,b7,jsCall_iiiiiiiiiiii_5,b7,jsCall_iiiiiiiiiiii_6,b7,jsCall_iiiiiiiiiiii_7,b7,jsCall_iiiiiiiiiiii_8,b7,jsCall_iiiiiiiiiiii_9,b7,jsCall_iiiiiiiiiiii_10,b7,jsCall_iiiiiiiiiiii_11,b7,jsCall_iiiiiiiiiiii_12,b7,jsCall_iiiiiiiiiiii_13
,b7,jsCall_iiiiiiiiiiii_14,b7,jsCall_iiiiiiiiiiii_15,b7,jsCall_iiiiiiiiiiii_16,b7,jsCall_iiiiiiiiiiii_17,b7,jsCall_iiiiiiiiiiii_18,b7,jsCall_iiiiiiiiiiii_19,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,b7,__ZNSt3__29money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIcEERNS_10unique_ptrIcPFvPvEEERPcSM_,b7,b7,b7,b7,b7,b7,b7,b7,__ZNSt3__29money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE8__do_getERS4_S4_bRKNS_6localeEjRjRbRKNS_5ctypeIwEERNS_10unique_ptrIwPFvPvEEERPwSM_,b7,b7,b7,b7,b7
,b7,b7,b7,b7,b7];
var FUNCTION_TABLE_iiii = [b8,b8,jsCall_iiii_0,b8,jsCall_iiii_1,b8,jsCall_iiii_2,b8,jsCall_iiii_3,b8,jsCall_iiii_4,b8,jsCall_iiii_5,b8,jsCall_iiii_6,b8,jsCall_iiii_7,b8,jsCall_iiii_8,b8,jsCall_iiii_9,b8,jsCall_iiii_10,b8,jsCall_iiii_11,b8,jsCall_iiii_12,b8,jsCall_iiii_13
,b8,jsCall_iiii_14,b8,jsCall_iiii_15,b8,jsCall_iiii_16,b8,jsCall_iiii_17,b8,jsCall_iiii_18,b8,jsCall_iiii_19,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE8toStringERNS0_7JStringEb,b8,b8
,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common6Object8toStringERNS0_7JStringEb,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE8toStringERNS0_7JStringEb,b8
,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E8toStringERNS0_7JStringEb,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common7JVectorIhE8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Common7JVectorINS0_6ObjectEE8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Common14BaseCharString8toStringERNS0_7JStringEb,b8,b8
,b8,b8,b8,b8,__ZNK9ExitGames6Common7JString8toStringERS1_b,b8,b8,b8,b8,b8,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator6resizeEPvj,b8,b8,b8,b8,b8,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager6resizeEPvj,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,__ZNK9ExitGames6Common9Hashtable8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Common14DictionaryBase8toStringERNS0_7JStringEb,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common14DictionaryBase8TypeInfo8toStringERNS0_7JStringEb,b8,__ZNK9ExitGames6Common7Helpers26DeSerializerImplementation8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Common7Helpers24SerializerImplementation8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Common10Serializer8toStringERNS0_7JStringEb,b8,b8,b8,b8,b8,b8
,b8,__ZN9ExitGames6Photon8Internal13SocketConnect15startConnectionERKNS_6Common7JStringES6_,b8,__ZN9ExitGames6Photon8Internal13SocketConnect11sendPackageEPKhi,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Photon8Internal13SocketConnect10sendBufferEPcRi,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Photon8Internal8PeerBase20deserializeOperationEPhi
,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Photon8Internal9Operation8toStringERNS_6Common7JStringEb,b8,b8,__ZN9ExitGames6Photon8Internal8TConnect10recvBufferEPPhPi,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK9ExitGames6Common7JVectorINS_6Photon8Internal9OperationEE8toStringERNS0_7JStringEb,b8,b8,__ZNK9ExitGames6Photon12TrafficStats8toStringERNS_6Common7JStringEb,b8,b8,__ZNK9ExitGames6Photon21TrafficStatsGameLevel8toStringERNS_6Common7JStringEb,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Photon10PhotonPeer7connectERKNS_6Common7JStringES5_,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,___stdio_write,___stdio_seek,___stdio_read,___stdout_write,_sn_write,b8,b8,b8,b8,b8,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6setbufEPci,b8,b8,b8,b8,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6xsgetnEPci,b8,b8,b8,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE6xsputnEPKci
,b8,b8,b8,b8,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6setbufEPwi,b8,b8,b8,b8,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6xsgetnEPwi,b8,b8,b8,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE6xsputnEPKwi,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,__ZNKSt3__214error_category10equivalentEiRKNS_15error_conditionE,__ZNKSt3__214error_category10equivalentERKNS_10error_codeEi,b8,b8,b8,b8,b8,b8,b8,b8,__ZNSt3__211__stdoutbufIwE6xsputnEPKwi,b8,b8,b8,b8,b8,__ZNSt3__211__stdoutbufIcE6xsputnEPKci,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNKSt3__27collateIcE7do_hashEPKcS3_,b8,b8,b8,b8,__ZNKSt3__27collateIwE7do_hashEPKwS3_,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNKSt3__28messagesIcE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE,b8,b8,b8,b8,__ZNKSt3__28messagesIwE7do_openERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNS_6localeE,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNKSt3__25ctypeIcE10do_toupperEPcPKc,b8,__ZNKSt3__25ctypeIcE10do_tolowerEPcPKc,b8,b8,__ZNKSt3__25ctypeIcE9do_narrowEcc,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,__ZNKSt3__25ctypeIwE5do_isEtw,b8,b8,b8,b8,__ZNKSt3__25ctypeIwE10do_toupperEPwPKw,b8,__ZNKSt3__25ctypeIwE10do_tolowerEPwPKw,b8,b8,__ZNKSt3__25ctypeIwE9do_narrowEwc,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNK10__cxxabiv117__class_type_info9can_catchEPKNS_16__shim_type_infoERPv,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,_wmemset,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation7popTypeEhRNS0_6ObjectE,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Common7JStringaSEx,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Common10ANSIString13ANSIConverter12Unicode2ANSIEPKwPcj,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZN9ExitGames6Common7Helpers24SerializerImplementation10pushObjectEPKNS0_6ObjectEb,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,_EGBN_bin2bn,b8,b8,b8,b8,b8,b8,_EGBN_add,_EGBN_sub,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,_do_read,b8,_wmemcpy,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,___newlocale,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,_vasprintf,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE23__append_forward_unsafeIPcEERS5_T_S9_,b8,b8,b8,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6appendEPKcj,b8,b8,b8,__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE23__append_forward_unsafeIPwEERS5_T_S9_,b8,b8,__ZNSt3__212basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEE6appendEPKwj
,b8,b8,b8,b8,b8,b8,b8,_wcrtomb,b8,b8,_mbtowc,b8,_mbrlen,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,_wmemmove,b8,b8,b8,b8,b8,b8
,b8,b8,b8,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKcj,__ZN10__cxxabiv112_GLOBAL__N_118parse_special_nameINS0_2DbEEEPKcS4_S4_RT_,__ZN10__cxxabiv112_GLOBAL__N_110parse_nameINS0_2DbEEEPKcS4_S4_RT_,__ZN10__cxxabiv112_GLOBAL__N_110parse_typeINS0_2DbEEEPKcS4_S4_RT_,b8,b8,b8,b8,b8,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKc,b8,__ZNKSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE7compareEjjPKcj,b8,b8,b8,__ZN10__cxxabiv112_GLOBAL__N_117parse_source_nameINS0_2DbEEEPKcS4_S4_RT_,b8,b8,b8,b8,b8,b8,__ZN10__cxxabiv112_GLOBAL__N_118parse_template_argINS0_2DbEEEPKcS4_S4_RT_,b8,b8,b8,__ZN10__cxxabiv112_GLOBAL__N_116parse_expressionINS0_2DbEEEPKcS4_S4_RT_
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8,b8
,b8,b8,b8,b8,b8,b8,b8,b8,b8];
var FUNCTION_TABLE_viiiiiiiiiiiiiii = [b9,b9,jsCall_viiiiiiiiiiiiiii_0,b9,jsCall_viiiiiiiiiiiiiii_1,b9,jsCall_viiiiiiiiiiiiiii_2,b9,jsCall_viiiiiiiiiiiiiii_3,b9,jsCall_viiiiiiiiiiiiiii_4,b9,jsCall_viiiiiiiiiiiiiii_5,b9,jsCall_viiiiiiiiiiiiiii_6,b9,jsCall_viiiiiiiiiiiiiii_7,b9,jsCall_viiiiiiiiiiiiiii_8,b9,jsCall_viiiiiiiiiiiiiii_9,b9,jsCall_viiiiiiiiiiiiiii_10,b9,jsCall_viiiiiiiiiiiiiii_11,b9,jsCall_viiiiiiiiiiiiiii_12,b9,jsCall_viiiiiiiiiiiiiii_13
,b9,jsCall_viiiiiiiiiiiiiii_14,b9,jsCall_viiiiiiiiiiiiiii_15,b9,jsCall_viiiiiiiiiiiiiii_16,b9,jsCall_viiiiiiiiiiiiiii_17,b9,jsCall_viiiiiiiiiiiiiii_18,b9,jsCall_viiiiiiiiiiiiiii_19,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9,b9
,b9,b9,__ZNSt3__211__money_putIcE8__formatEPcRS2_S3_jPKcS5_RKNS_5ctypeIcEEbRKNS_10money_base7patternEccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESL_SL_i,b9,__ZNSt3__211__money_putIwE8__formatEPwRS2_S3_jPKwS5_RKNS_5ctypeIwEEbRKNS_10money_base7patternEwwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERKNSE_IwNSF_IwEENSH_IwEEEESQ_i];
var FUNCTION_TABLE_viiiiiiii = [b10,b10,jsCall_viiiiiiii_0,b10,jsCall_viiiiiiii_1,b10,jsCall_viiiiiiii_2,b10,jsCall_viiiiiiii_3,b10,jsCall_viiiiiiii_4,b10,jsCall_viiiiiiii_5,b10,jsCall_viiiiiiii_6,b10,jsCall_viiiiiiii_7,b10,jsCall_viiiiiiii_8,b10,jsCall_viiiiiiii_9,b10,jsCall_viiiiiiii_10,b10,jsCall_viiiiiiii_11,b10,jsCall_viiiiiiii_12,b10,jsCall_viiiiiiii_13
,b10,jsCall_viiiiiiii_14,b10,jsCall_viiiiiiii_15,b10,jsCall_viiiiiiii_16,b10,jsCall_viiiiiiii_17,b10,jsCall_viiiiiiii_18,b10,jsCall_viiiiiiii_19,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,__ZNK9ExitGames6Common6Logger3logEiPKwS3_bjS3_z,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10,b10
,b10,b10,b10,b10,b10];
var FUNCTION_TABLE_viiiiii = [b11,b11,jsCall_viiiiii_0,b11,jsCall_viiiiii_1,b11,jsCall_viiiiii_2,b11,jsCall_viiiiii_3,b11,jsCall_viiiiii_4,b11,jsCall_viiiiii_5,b11,jsCall_viiiiii_6,b11,jsCall_viiiiii_7,b11,jsCall_viiiiii_8,b11,jsCall_viiiiii_9,b11,jsCall_viiiiii_10,b11,jsCall_viiiiii_11,b11,jsCall_viiiiii_12,b11,jsCall_viiiiii_13
,b11,jsCall_viiiiii_14,b11,jsCall_viiiiii_15,b11,jsCall_viiiiii_16,b11,jsCall_viiiiii_17,b11,jsCall_viiiiii_18,b11,jsCall_viiiiii_19,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE7seekoffExNS_8ios_base7seekdirEj,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE7seekoffExNS_8ios_base7seekdirEj,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,__ZNKSt3__28messagesIcE6do_getEiiiRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE,b11,b11,b11,b11,__ZNKSt3__28messagesIwE6do_getEiiiRKNS_12basic_stringIwNS_11char_traitsIwEENS_9allocatorIwEEEE,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,__ZNK10__cxxabiv117__class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b11,b11,b11,__ZNK10__cxxabiv120__si_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,__ZNK10__cxxabiv121__vmi_class_type_info16search_above_dstEPNS_19__dynamic_cast_infoEPKvS4_ib,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,__ZN9ExitGames6Common6ObjectC2EPKvhhib,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11,b11
,b11,b11,b11,b11,b11];
var FUNCTION_TABLE_iiiiiii = [b12,b12,jsCall_iiiiiii_0,b12,jsCall_iiiiiii_1,b12,jsCall_iiiiiii_2,b12,jsCall_iiiiiii_3,b12,jsCall_iiiiiii_4,b12,jsCall_iiiiiii_5,b12,jsCall_iiiiiii_6,b12,jsCall_iiiiiii_7,b12,jsCall_iiiiiii_8,b12,jsCall_iiiiiii_9,b12,jsCall_iiiiiii_10,b12,jsCall_iiiiiii_11,b12,jsCall_iiiiiii_12,b12,jsCall_iiiiiii_13
,b12,jsCall_iiiiiii_14,b12,jsCall_iiiiiii_15,b12,jsCall_iiiiiii_16,b12,jsCall_iiiiiii_17,b12,jsCall_iiiiiii_18,b12,jsCall_iiiiiii_19,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZN9ExitGames6Photon8Internal8PeerBase8opCustomEPKNS0_16OperationRequestEbhbh,b12,b12,b12,b12,b12,b12,__ZN9ExitGames6Photon8Internal8PeerBase18serializeOperationEPKNS0_16OperationRequestEPPhPibh,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,__ZN9ExitGames6Photon8Internal5TPeer18serializeOperationEPKNS0_16OperationRequestEPPhPibh,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRb,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRl,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRx,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRt,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRm,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRy,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRf,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRd,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRe,__ZNKSt3__27num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv,b12
,b12,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRb,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRl,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRx,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRt,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjS8_,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRm,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRy,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRf,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRd,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRe,__ZNKSt3__27num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjRPv,b12,b12,b12,b12,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcx,b12,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcy,b12,b12,b12,b12,b12,b12,b12,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwx,b12,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwy,b12
,b12,b12,b12,b12,b12,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_timeES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_dateES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE14do_get_weekdayES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE16do_get_monthnameES4_S4_RNS_8ios_baseERjP2tm,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE11do_get_yearES4_S4_RNS_8ios_baseERjP2tm,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,__ZNKSt3__29money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_bRNS_8ios_baseEcRKNS_12basic_stringIcS3_NS_9allocatorIcEEEE,b12,b12,b12,__ZNKSt3__29money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_bRNS_8ios_baseEwRKNS_12basic_stringIwS3_NS_9allocatorIwEEEE,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZNSt3__216__pad_and_outputIwNS_11char_traitsIwEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,__ZNSt3__216__pad_and_outputIcNS_11char_traitsIcEEEENS_19ostreambuf_iteratorIT_T0_EES6_PKS4_S8_S8_RNS_8ios_baseES4_,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12,b12
,b12,b12,b12,b12,b12];
var FUNCTION_TABLE_iiiiiiiiiiiii = [b13,b13,jsCall_iiiiiiiiiiiii_0,b13,jsCall_iiiiiiiiiiiii_1,b13,jsCall_iiiiiiiiiiiii_2,b13,jsCall_iiiiiiiiiiiii_3,b13,jsCall_iiiiiiiiiiiii_4,b13,jsCall_iiiiiiiiiiiii_5,b13,jsCall_iiiiiiiiiiiii_6,b13,jsCall_iiiiiiiiiiiii_7,b13,jsCall_iiiiiiiiiiiii_8,b13,jsCall_iiiiiiiiiiiii_9,b13,jsCall_iiiiiiiiiiiii_10,b13,jsCall_iiiiiiiiiiiii_11,b13,jsCall_iiiiiiiiiiiii_12,b13,jsCall_iiiiiiiiiiiii_13
,b13,jsCall_iiiiiiiiiiiii_14,b13,jsCall_iiiiiiiiiiiii_15,b13,jsCall_iiiiiiiiiiiii_16,b13,jsCall_iiiiiiiiiiiii_17,b13,jsCall_iiiiiiiiiiiii_18,b13,jsCall_iiiiiiiiiiiii_19,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,__ZNSt3__29__num_getIcE19__stage2_float_loopEcRbRcPcRS4_ccRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjS4_,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,__ZNSt3__29__num_getIwE19__stage2_float_loopEwRbRcPcRS4_wwRKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjRSE_RjPw,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13,b13
,b13,b13,b13,b13,b13];
var FUNCTION_TABLE_iid = [b14,b14,jsCall_iid_0,b14,jsCall_iid_1,b14,jsCall_iid_2,b14,jsCall_iid_3,b14,jsCall_iid_4,b14,jsCall_iid_5,b14,jsCall_iid_6,b14,jsCall_iid_7,b14,jsCall_iid_8,b14,jsCall_iid_9,b14,jsCall_iid_10,b14,jsCall_iid_11,b14,jsCall_iid_12,b14,jsCall_iid_13
,b14,jsCall_iid_14,b14,jsCall_iid_15,b14,jsCall_iid_16,b14,jsCall_iid_17,b14,jsCall_iid_18,b14,jsCall_iid_19,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,__ZN9ExitGames6Common7JStringaSEf,__ZN9ExitGames6Common7JStringaSEd,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14,b14
,b14,b14,b14,b14,b14];
var FUNCTION_TABLE_viiiiiii = [b15,b15,jsCall_viiiiiii_0,b15,jsCall_viiiiiii_1,b15,jsCall_viiiiiii_2,b15,jsCall_viiiiiii_3,b15,jsCall_viiiiiii_4,b15,jsCall_viiiiiii_5,b15,jsCall_viiiiiii_6,b15,jsCall_viiiiiii_7,b15,jsCall_viiiiiii_8,b15,jsCall_viiiiiii_9,b15,jsCall_viiiiiii_10,b15,jsCall_viiiiiii_11,b15,jsCall_viiiiiii_12,b15,jsCall_viiiiiii_13
,b15,jsCall_viiiiiii_14,b15,jsCall_viiiiiii_15,b15,jsCall_viiiiiii_16,b15,jsCall_viiiiiii_17,b15,jsCall_viiiiiii_18,b15,jsCall_viiiiiii_19,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,__ZN9ExitGames6Common6Object3setEPKvhhjPKsb,b15,b15,b15,b15,__ZN9ExitGames6Common6Object17setWithoutCleanupEPKvhhjPKsb,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,__ZNSt3__29__num_putIcE21__widen_and_group_intEPcS2_S2_S2_RS2_S3_RKNS_6localeE,b15,__ZNSt3__29__num_putIcE23__widen_and_group_floatEPcS2_S2_S2_RS2_S3_RKNS_6localeE,b15,b15,b15,b15,__ZNSt3__29__num_putIwE21__widen_and_group_intEPcS2_S2_PwRS3_S4_RKNS_6localeE,__ZNSt3__29__num_putIwE23__widen_and_group_floatEPcS2_S2_PwRS3_S4_RKNS_6localeE,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15,b15
,b15,b15,b15,b15,b15];
var FUNCTION_TABLE_viiiiiiiiii = [b16,b16,jsCall_viiiiiiiiii_0,b16,jsCall_viiiiiiiiii_1,b16,jsCall_viiiiiiiiii_2,b16,jsCall_viiiiiiiiii_3,b16,jsCall_viiiiiiiiii_4,b16,jsCall_viiiiiiiiii_5,b16,jsCall_viiiiiiiiii_6,b16,jsCall_viiiiiiiiii_7,b16,jsCall_viiiiiiiiii_8,b16,jsCall_viiiiiiiiii_9,b16,jsCall_viiiiiiiiii_10,b16,jsCall_viiiiiiiiii_11,b16,jsCall_viiiiiiiiii_12,b16,jsCall_viiiiiiiiii_13
,b16,jsCall_viiiiiiiiii_14,b16,jsCall_viiiiiiiiii_15,b16,jsCall_viiiiiiiiii_16,b16,jsCall_viiiiiiiiii_17,b16,jsCall_viiiiiiiiii_18,b16,jsCall_viiiiiiiiii_19,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16
,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,b16,__ZNSt3__211__money_getIcE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_SF_Ri,b16,b16,b16,b16,b16,b16,b16,__ZNSt3__211__money_getIwE13__gather_infoEbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_SJ_Ri,b16,b16
,b16,__ZNSt3__211__money_putIcE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERcS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEESF_SF_Ri,b16,__ZNSt3__211__money_putIwE13__gather_infoEbbRKNS_6localeERNS_10money_base7patternERwS8_RNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEERNS9_IwNSA_IwEENSC_IwEEEESJ_Ri,b16];
var FUNCTION_TABLE_iii = [b17,b17,jsCall_iii_0,b17,jsCall_iii_1,b17,jsCall_iii_2,b17,jsCall_iii_3,b17,jsCall_iii_4,b17,jsCall_iii_5,b17,jsCall_iii_6,b17,jsCall_iii_7,b17,jsCall_iii_8,b17,jsCall_iii_9,b17,jsCall_iii_10,b17,jsCall_iii_11,b17,jsCall_iii_12,b17,jsCall_iii_13
,b17,jsCall_iii_14,b17,jsCall_iii_15,b17,jsCall_iii_16,b17,jsCall_iii_17,b17,jsCall_iii_18,b17,jsCall_iii_19,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,__ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE6assignERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE7compareERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryIhNS0_6ObjectEE4copyEs,b17,b17,b17,b17,__ZN9ExitGames6Common9KeyObjectIhEaSERKNS0_6ObjectE,__ZN9ExitGames6Common9KeyObjectIhE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common9KeyObjectIhEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIhEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIhE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIhEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIsEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIsE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIsEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIiEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIiE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIiEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIxEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIxE6assignERKNS0_6ObjectE
,__ZN9ExitGames6Common11ValueObjectIxEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIfEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIfE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIfEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIdEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIdE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIdEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectIbEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIbE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIbEaSERKS2_,b17,b17,__ZN9ExitGames6Common11ValueObjectINS0_7JStringEEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectINS0_7JStringEE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectINS0_7JStringEEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectINS0_9HashtableEE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectINS0_9HashtableEEaSERKS3_,b17,b17,b17,b17
,b17,b17,__ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE6assignERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE7compareERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryIiNS0_6ObjectEE4copyEs,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E6assignERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E7compareERKNS0_14DictionaryBaseE,__ZNK9ExitGames6Common10DictionaryINS0_6ObjectES2_E4copyEs,b17,b17,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEaSERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEE6assignERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIhNS0_6ObjectEEEEaSERKS5_,b17,b17,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEaSERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEE6assignERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryIiNS0_6ObjectEEEEaSERKS5_,b17,b17,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEaSERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEE6assignERKS3_,__ZN9ExitGames6Common11ValueObjectINS0_10DictionaryINS0_6ObjectES3_EEEaSERKS5_,b17
,b17,__ZN9ExitGames6Common11ValueObjectIPKhEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPKhE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPKhEaSERKS4_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPsEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPsE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPsEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPiEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPiE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPiEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPxEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPxE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPxEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPfEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPfE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPfEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPdEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPdE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPdEaSERKS3_,b17
,b17,__ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPNS0_7JStringEEaSERKS4_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPbEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPbE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPbEaSERKS3_,b17,b17,__ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEaSERKS2_,__ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEE6assignERKS2_,__ZN9ExitGames6Common11ValueObjectIPNS0_6ObjectEEaSERKS4_,b17,b17,__ZN9ExitGames6Common9KeyObjectIiEaSERKNS0_6ObjectE,__ZN9ExitGames6Common9KeyObjectIiE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common9KeyObjectIiEaSERKS2_,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common16MemoryManagement8Internal15DirectAllocator5allocEj,b17,b17,b17,b17,b17,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager5allocEj,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common6ObjectaSERKS1_
,__ZN9ExitGames6Common6Object6assignERKS1_,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common14DictionaryBase6assignERKS1_,__ZNK9ExitGames6Common14DictionaryBase7compareERKS1_,__ZNK9ExitGames6Common14DictionaryBase4copyEs,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common11ValueObjectIPhEaSERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPhE6assignERKNS0_6ObjectE,__ZN9ExitGames6Common11ValueObjectIPhEaSERKS3_
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,__ZNK9ExitGames6Photon8Internal5TPeer18allocateConnectionERNS1_21PhotonConnectListenerE,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE9pbackfailEi,b17
,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE8overflowEi,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE9pbackfailEj,b17,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE8overflowEj,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNSt3__211__stdoutbufIwE8overflowEj,b17,b17,b17,b17,b17,__ZNSt3__211__stdoutbufIcE8overflowEi,b17,b17,b17,b17,b17,__ZNSt3__210__stdinbufIwE9pbackfailEj,b17
,b17,b17,b17,b17,__ZNSt3__210__stdinbufIcE9pbackfailEi,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNKSt3__25ctypeIcE10do_toupperEc,b17,__ZNKSt3__25ctypeIcE10do_tolowerEc,b17,__ZNKSt3__25ctypeIcE8do_widenEc,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNKSt3__25ctypeIwE10do_toupperEw,b17,__ZNKSt3__25ctypeIwE10do_tolowerEw,b17,__ZNKSt3__25ctypeIwE8do_widenEc,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNKSt3__26locale9use_facetERNS0_2idE,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7JVectorIhEixEj
,b17,b17,b17,b17,__ZN9ExitGames6CommonlsIwNSt3__211char_traitsIwEEEERNS2_13basic_ostreamIT_T0_EES9_RKNS0_7JStringE,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7JStringaSEPKw,__ZN9ExitGames6Common7JStringpLERKS1_,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7JVectorINS0_6ObjectEEixEj,b17,b17,b17,b17,__ZN9ExitGames6Common7JVectorINS0_6ObjectEEaSERKS3_,b17,b17
,__ZN9ExitGames6Common7JStringpLIhEERS1_RKT_,__ZN9ExitGames6Common7JStringpLIA3_wEERS1_RKT_,__ZN9ExitGames6Common7JStringpLIA2_wEERS1_RKT_,__ZN9ExitGames6Common7JStringaSEh,b17,b17,b17,b17,b17,b17,_wcscpy,_wcscat,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7JStringaSERKS1_,__ZNK9ExitGames6Common7JString11lastIndexOfEw,b17,b17,_wcsstr,b17
,__ZN9ExitGames6Common10UTF8StringaSEPKc,b17,__ZN9ExitGames6Common10ANSIStringaSEPKc,b17,b17,b17,b17,b17,b17,b17,__ZNK9ExitGames6Common7JString11lastIndexOfEc,b17,b17,b17,b17,b17,b17,b17,b17,_wcsrchr,__ZN9ExitGames6Common10UTF8StringaSERKNS0_7JStringE,b17,b17,__ZN9ExitGames6Common7JStringaSEs,__ZN9ExitGames6Common7JStringaSEi,b17,b17,b17,__ZN9ExitGames6Common7JStringaSEb,b17
,__ZN9ExitGames6Common14CustomTypeBase11allocObjectEsh,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7JStringaSEPKc,__ZNK9ExitGames6Common9Hashtable22getValueImplementationERKNS0_6ObjectE,b17,__ZN9ExitGames6Common9HashtableaSERKS1_,_wcschr,b17,b17,b17
,b17,__ZNK9ExitGames6Common7JString7indexOfERKS1_,__ZNK9ExitGames6Common7JString10startsWithERKS1_,b17,__ZNK9ExitGames6Common7JString7indexOfEc,__ZN9ExitGames6Common7JStringaSEt,b17,b17,b17,b17,b17,b17,__ZNK9ExitGames6Common9Hashtable8getValueINS0_6ObjectEEEPKS3_RKT_,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation3popERNS0_6ObjectE,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,_EGBN_bn2bin,b17,b17,b17,b17,b17,b17,b17,__ZN9ExitGames6Common10Serializer4pushINS0_6ObjectEEEbRKT_,b17,b17,b17,__ZN9ExitGames6Common7JStringaSEj,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,_isxdigit_l,_isdigit_l,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6appendEPKc,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17,b17
,b17,b17,b17,b17,b17,b17,b17,b17,b17];
var FUNCTION_TABLE_iiiiii = [b18,b18,jsCall_iiiiii_0,b18,jsCall_iiiiii_1,b18,jsCall_iiiiii_2,b18,jsCall_iiiiii_3,b18,jsCall_iiiiii_4,b18,jsCall_iiiiii_5,b18,jsCall_iiiiii_6,b18,jsCall_iiiiii_7,b18,jsCall_iiiiii_8,b18,jsCall_iiiiii_9,b18,jsCall_iiiiii_10,b18,jsCall_iiiiii_11,b18,jsCall_iiiiii_12,b18,jsCall_iiiiii_13
,b18,jsCall_iiiiii_14,b18,jsCall_iiiiii_15,b18,jsCall_iiiiii_16,b18,jsCall_iiiiii_17,b18,jsCall_iiiiii_18,b18,jsCall_iiiiii_19,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZN9ExitGames6Photon10PhotonPeer8opCustomERKNS0_16OperationRequestEbhb,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,__ZNKSt3__27collateIcE10do_compareEPKcS3_S3_S3_,b18,b18,b18,b18,__ZNKSt3__27collateIwE10do_compareEPKwS3_S3_S3_,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcb,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcl,b18,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcm,b18,b18,b18,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcPKv,b18,b18,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwb,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwl,b18,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwm,b18,b18
,b18,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwPKv,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZNKSt3__27codecvtIDic11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b18,b18,__ZNKSt3__27codecvtIDic11__mbstate_tE9do_lengthERS1_PKcS5_j,b18,b18
,b18,b18,b18,b18,__ZNKSt3__27codecvtIwc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b18,b18,__ZNKSt3__27codecvtIwc11__mbstate_tE9do_lengthERS1_PKcS5_j,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZNKSt3__25ctypeIcE9do_narrowEPKcS3_cPc,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZNKSt3__25ctypeIwE9do_narrowEPKwS3_cPc,b18,b18,b18,__ZNKSt3__27codecvtIcc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b18,b18,__ZNKSt3__27codecvtIcc11__mbstate_tE9do_lengthERS1_PKcS5_j,b18,b18,b18,b18
,__ZNKSt3__27codecvtIDsc11__mbstate_tE10do_unshiftERS1_PcS4_RS4_,b18,b18,__ZNKSt3__27codecvtIDsc11__mbstate_tE9do_lengthERS1_PKcS5_j,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,__ZN9ExitGames6Photon8Internal10Encryption7encryptEPKhiS4_PPhPi,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,_EGBN_mod_exp,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,_wcsnrtombs,b18,_mbsnrtowcs,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18,b18
,b18,b18,b18,b18,b18,b18,b18,b18,b18];
var FUNCTION_TABLE_diii = [b19,b19,jsCall_diii_0,b19,jsCall_diii_1,b19,jsCall_diii_2,b19,jsCall_diii_3,b19,jsCall_diii_4,b19,jsCall_diii_5,b19,jsCall_diii_6,b19,jsCall_diii_7,b19,jsCall_diii_8,b19,jsCall_diii_9,b19,jsCall_diii_10,b19,jsCall_diii_11,b19,jsCall_diii_12,b19,jsCall_diii_13
,b19,jsCall_diii_14,b19,jsCall_diii_15,b19,jsCall_diii_16,b19,jsCall_diii_17,b19,jsCall_diii_18,b19,jsCall_diii_19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,__ZNSt3__215__num_get_floatIeEET_PKcS3_Rj,b19,__ZNSt3__215__num_get_floatIdEET_PKcS3_Rj,__ZNSt3__215__num_get_floatIfEET_PKcS3_Rj,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19,b19
,b19,b19,b19,b19,b19];
var FUNCTION_TABLE_i = [b20,b20,jsCall_i_0,b20,jsCall_i_1,b20,jsCall_i_2,b20,jsCall_i_3,b20,jsCall_i_4,b20,jsCall_i_5,b20,jsCall_i_6,b20,jsCall_i_7,b20,jsCall_i_8,b20,jsCall_i_9,b20,jsCall_i_10,b20,jsCall_i_11,b20,jsCall_i_12,b20,jsCall_i_13
,b20,jsCall_i_14,b20,jsCall_i_15,b20,jsCall_i_16,b20,jsCall_i_17,b20,jsCall_i_18,b20,jsCall_i_19,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,__ZN9ExitGames6Common16MemoryManagement8Internal17MemoryPoolManager3getEv,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,__Z11getTimeUnixv,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,__ZN9ExitGames6Photon8Internal8PeerBase10InitValues22getClientLibIdentifierEv,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,_EGBN_new,_EGBN_CTX_new,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,__ZNSt3__26__clocEv,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,___ctype_get_mb_cur_max,b20,___ctype_tolower_loc,___ctype_toupper_loc,___ctype_b_loc,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,__ZNSt3__26locale5__imp12make_classicEv,__ZNSt3__26locale5__imp11make_globalEv,__ZNSt3__26locale8__globalEv,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,___cxa_get_globals_fast,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20,b20
,b20,b20,b20,b20,b20,b20,b20,b20,b20];
var FUNCTION_TABLE_iiiii = [b21,b21,jsCall_iiiii_0,b21,jsCall_iiiii_1,b21,jsCall_iiiii_2,b21,jsCall_iiiii_3,b21,jsCall_iiiii_4,b21,jsCall_iiiii_5,b21,jsCall_iiiii_6,b21,jsCall_iiiii_7,b21,jsCall_iiiii_8,b21,jsCall_iiiii_9,b21,jsCall_iiiii_10,b21,jsCall_iiiii_11,b21,jsCall_iiiii_12,b21,jsCall_iiiii_13
,b21,jsCall_iiiii_14,b21,jsCall_iiiii_15,b21,jsCall_iiiii_16,b21,jsCall_iiiii_17,b21,jsCall_iiiii_18,b21,jsCall_iiiii_19,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,__ZN9ExitGames6Photon8Internal5TPeer7connectERKNS_6Common7JStringES6_RKNS3_6ObjectE,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,__ZN9ExitGames6Photon10PhotonPeer7connectERKNS_6Common7JStringES5_RKNS2_6ObjectE,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,__ZNKSt3__25ctypeIcE8do_widenEPKcS3_Pc,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,__ZNKSt3__25ctypeIwE5do_isEPKwS3_Pt,__ZNKSt3__25ctypeIwE10do_scan_isEtPKwS3_,__ZNKSt3__25ctypeIwE11do_scan_notEtPKwS3_,b21,b21,b21,b21,b21,__ZNKSt3__25ctypeIwE8do_widenEPKcS3_Pw,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,_strftime__wrapper,__Z11EG_swprintfPwjPKwz,b21,b21,b21,b21,b21,__ZN9ExitGames6Common6Object14toStringHelperERKS1_RNS0_7JStringEbb,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,__ZNSt3__217__libcpp_sscanf_lEPKcP15__locale_structS1_z,b21,b21
,b21,b21,b21,b21,b21,b21,b21,__ZNSt3__227__num_get_unsigned_integralIyEET_PKcS3_Rji,__ZNSt3__227__num_get_unsigned_integralImEET_PKcS3_Rji,__ZNSt3__227__num_get_unsigned_integralIjEET_PKcS3_Rji,__ZNSt3__227__num_get_unsigned_integralItEET_PKcS3_Rji,__ZNSt3__225__num_get_signed_integralIxEET_PKcS3_Rji,__ZNSt3__225__num_get_signed_integralIlEET_PKcS3_Rji,b21,b21,b21,b21,__ZNSt3__219__libcpp_asprintf_lEPPcP15__locale_structPKcz,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,_mbsrtowcs,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,_catgets,b21,b21,b21,_mbrtowc,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertEjPKcj,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,__ZN10__cxxabiv112_GLOBAL__N_123parse_binary_expressionINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_,__ZN10__cxxabiv112_GLOBAL__N_123parse_prefix_expressionINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_,__ZN10__cxxabiv112_GLOBAL__N_121parse_integer_literalINS0_2DbEEEPKcS4_S4_RKNT_6StringERS5_,b21,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEE6insertIPKcEENS_9enable_ifIXaasr21__is_forward_iteratorIT_EE5valuesr38__libcpp_string_gets_noexcept_iteratorISC_EE5valueENS_11__wrap_iterIPcEEE4typeENSD_ISA_EESC_SC_,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21,b21
,b21,b21,b21,b21,b21,b21,b21,b21,b21];
var FUNCTION_TABLE_viii = [b22,b22,jsCall_viii_0,b22,jsCall_viii_1,b22,jsCall_viii_2,b22,jsCall_viii_3,b22,jsCall_viii_4,b22,jsCall_viii_5,b22,jsCall_viii_6,b22,jsCall_viii_7,b22,jsCall_viii_8,b22,jsCall_viii_9,b22,jsCall_viii_10,b22,jsCall_viii_11,b22,jsCall_viii_12,b22,jsCall_viii_13
,b22,jsCall_viii_14,b22,jsCall_viii_15,b22,jsCall_viii_16,b22,jsCall_viii_17,b22,jsCall_viii_18,b22,jsCall_viii_19,b22,b22,b22,__ZN16PhotonListenerEM11debugReturnEiRKN9ExitGames6Common7JStringE,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Photon8Internal8PeerBase10pingServerERKNS_6Common7JStringEj,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,__ZNK9ExitGames6Photon10PhotonPeer18vitalStatsToStringEb,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,__ZNKSt3__214error_category23default_error_conditionEi,b22,b22,__ZNKSt3__219__iostream_category7messageEi,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Photon16OperationRequestC2EhRKNS_6Common10DictionaryIhNS2_6ObjectEEE,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Common9Hashtable17putImplementationERKNS0_6ObjectES4_,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Common9KeyObjectIhE7convertEPKNS0_6ObjectEh,b22
,b22,b22,b22,b22,__ZNK9ExitGames6Common8ToString8toStringEb,b22,__ZN9ExitGames6CommonplERKNS0_7JStringES3_,b22,b22,__ZN9ExitGames6CommonplIA2_wEENS0_7JStringERKS3_RKT_,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZNK9ExitGames6Common7JString9substringEj,b22,b22
,b22,b22,b22,b22,b22,__ZN9ExitGames6CommonplIPKwEENS0_7JStringERKT_RKS4_,b22,b22,b22,__ZN9ExitGames6Common6Object7cleanupEPKvj,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,__ZN9ExitGames6Common9Hashtable3putINS0_6ObjectES3_EEvRKT_RKT0_,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6CommonplIA3_wEENS0_7JStringERKS3_RKT_,b22,b22,__ZN9ExitGames6CommonplIA2_wEENS0_7JStringERKT_RKS3_,b22,b22,b22,b22,b22,__ZN9ExitGames6CommonplIhEENS0_7JStringERKS2_RKT_,b22,b22
,b22,b22,b22,__ZN9ExitGames6CommonplItEENS0_7JStringERKS2_RKT_,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Common7Helpers24SerializerImplementation17writeInvertedDataEPKvi,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Photon17OperationResponseC2Ehs,b22,b22,b22,b22,__ZN9ExitGames6Photon17OperationResponse12addParameterEhRKNS_6Common6ObjectE,b22,b22,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter23timeForResponseCallbackEhi,b22,b22,__ZN9ExitGames6Common9Hashtable3putIhNS0_6ObjectEEEvRKT_RKT0_
,__ZN9ExitGames6Photon9EventDataC2EhRKNS_6Common10DictionaryIhNS2_6ObjectEEE,b22,b22,__ZN9ExitGames6Photon8Internal28TrafficStatsGameLevelCounter20timeForEventCallbackEhi,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Photon8Internal10Encryption19calculateSHA256HashEPKhiPh,b22,b22,b22,__ZN9ExitGames6CommonplIA6_wEENS0_7JStringERKS3_RKT_,b22,b22,b22,b22,b22,b22,b22,b22,__ZNK9ExitGames6Photon17OperationResponse19getParameterForCodeEh,b22,__ZN9ExitGames6CommonplIiEENS0_7JStringERKS2_RKT_,__ZN9ExitGames6CommonplIA12_wEENS0_7JStringERKS3_RKT_
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZN9ExitGames6Photon8Internal8PeerDataC2ERNS0_14PhotonListenerEb,b22,b22,b22,b22,b22,__ZNSt3__28ios_base7failureC2EPKcRKNS_10error_codeE,___cxa_throw__wrapper,b22,b22,b22,b22,b22,b22,__ZNSt3__212basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEE6resizeEjc,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZNSt3__219__double_or_nothingIcEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,__ZNSt3__219__double_or_nothingIjEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,b22,b22,b22,b22,b22,b22,__ZNSt3__219__double_or_nothingIwEEvRNS_10unique_ptrIT_PFvPvEEERPS2_S9_,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZNSt3__212system_error6__initERKNS_10error_codeENS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEE,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,__ZNSt3__26vectorIN10__cxxabiv112_GLOBAL__N_111string_pairENS2_11short_allocIS3_Lj4096EEEEC2EjRKS3_RKS5_,__ZNSt3__212basic_stringIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEC2ERKS7_jjRKS6_,__ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EERKSB_PKS8_,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,__ZNSt3__2plIcNS_11char_traitsIcEEN10__cxxabiv112_GLOBAL__N_112malloc_allocIcEEEENS_12basic_stringIT_T0_T1_EEPKS8_RKSB_,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22,b22
,b22,b22,b22,b22,b22,b22,b22,b22,b22];
var FUNCTION_TABLE_v = [b23,b23,jsCall_v_0,b23,jsCall_v_1,b23,jsCall_v_2,b23,jsCall_v_3,b23,jsCall_v_4,b23,jsCall_v_5,b23,jsCall_v_6,b23,jsCall_v_7,b23,jsCall_v_8,b23,jsCall_v_9,b23,jsCall_v_10,b23,jsCall_v_11,b23,jsCall_v_12,b23,jsCall_v_13
,b23,jsCall_v_14,b23,jsCall_v_15,b23,jsCall_v_16,b23,jsCall_v_17,b23,jsCall_v_18,b23,jsCall_v_19,b23,b23,b23,b23,b23,b23,b23,b23,b23,___cxa_pure_virtual__wrapper,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,__ZL25default_terminate_handlerv,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,__ZSt17__throw_bad_allocv,b23,b23,b23,b23,b23,___cxa_end_catch__wrapper,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,__ZNSt3__2L10init_weeksEv,__ZNSt3__2L11init_monthsEv,__ZNSt3__2L10init_am_pmEv,b23,b23
,b23,b23,b23,b23,__ZNSt3__2L11init_wweeksEv,__ZNSt3__2L12init_wmonthsEv,__ZNSt3__2L11init_wam_pmEv,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7collateIcEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7collateIwEEjEERT_T0_,b23,b23,__ZNSt3__212_GLOBAL__N_14makeINS_5ctypeIwEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIcc11__mbstate_tEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIwc11__mbstate_tEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIDsc11__mbstate_tEEjEERT_T0_,b23
,__ZNSt3__212_GLOBAL__N_14makeINS_7codecvtIDic11__mbstate_tEEjEERT_T0_,b23,b23,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7num_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7num_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_7num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIcLb0EEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIcLb1EEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIwLb0EEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_10moneypunctIwLb1EEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_9money_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_9money_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_9money_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_9money_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_8time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23
,__ZNSt3__212_GLOBAL__N_14makeINS_8time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_8time_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_8time_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_8messagesIcEEjEERT_T0_,b23,__ZNSt3__212_GLOBAL__N_14makeINS_8messagesIwEEjEERT_T0_,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,___cxa_rethrow__wrapper,b23,b23,b23,b23,__ZN10__cxxabiv112_GLOBAL__N_110construct_Ev,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23,b23
,b23,b23,b23,b23,b23,b23,b23,b23,b23];
var FUNCTION_TABLE_iiiiiiiii = [b24,b24,jsCall_iiiiiiiii_0,b24,jsCall_iiiiiiiii_1,b24,jsCall_iiiiiiiii_2,b24,jsCall_iiiiiiiii_3,b24,jsCall_iiiiiiiii_4,b24,jsCall_iiiiiiiii_5,b24,jsCall_iiiiiiiii_6,b24,jsCall_iiiiiiiii_7,b24,jsCall_iiiiiiiii_8,b24,jsCall_iiiiiiiii_9,b24,jsCall_iiiiiiiii_10,b24,jsCall_iiiiiiiii_11,b24,jsCall_iiiiiiiii_12,b24,jsCall_iiiiiiiii_13
,b24,jsCall_iiiiiiiii_14,b24,jsCall_iiiiiiiii_15,b24,jsCall_iiiiiiiii_16,b24,jsCall_iiiiiiiii_17,b24,jsCall_iiiiiiiii_18,b24,jsCall_iiiiiiiii_19,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,__ZNKSt3__28time_getIcNS_19istreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,__ZNKSt3__28time_getIwNS_19istreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_getES4_S4_RNS_8ios_baseERjP2tmcc,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,__ZNKSt3__27codecvtIDic11__mbstate_tE6do_outERS1_PKDiS5_RS5_PcS7_RS7_,__ZNKSt3__27codecvtIDic11__mbstate_tE5do_inERS1_PKcS5_RS5_PDiS7_RS7_,b24,b24,b24,b24,b24,b24
,b24,b24,__ZNKSt3__27codecvtIwc11__mbstate_tE6do_outERS1_PKwS5_RS5_PcS7_RS7_,__ZNKSt3__27codecvtIwc11__mbstate_tE5do_inERS1_PKcS5_RS5_PwS7_RS7_,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,__ZNKSt3__27codecvtIcc11__mbstate_tE6do_outERS1_PKcS5_RS5_PcS7_RS7_,__ZNKSt3__27codecvtIcc11__mbstate_tE5do_inERS1_PKcS5_RS5_PcS7_RS7_,b24,b24,b24,b24,b24,b24,__ZNKSt3__27codecvtIDsc11__mbstate_tE6do_outERS1_PKDsS5_RS5_PcS7_RS7_,__ZNKSt3__27codecvtIDsc11__mbstate_tE5do_inERS1_PKcS5_RS5_PDsS7_RS7_
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24,b24
,b24,b24,b24,b24,b24];
var FUNCTION_TABLE_iiiiid = [b25,b25,jsCall_iiiiid_0,b25,jsCall_iiiiid_1,b25,jsCall_iiiiid_2,b25,jsCall_iiiiid_3,b25,jsCall_iiiiid_4,b25,jsCall_iiiiid_5,b25,jsCall_iiiiid_6,b25,jsCall_iiiiid_7,b25,jsCall_iiiiid_8,b25,jsCall_iiiiid_9,b25,jsCall_iiiiid_10,b25,jsCall_iiiiid_11,b25,jsCall_iiiiid_12,b25,jsCall_iiiiid_13
,b25,jsCall_iiiiid_14,b25,jsCall_iiiiid_15,b25,jsCall_iiiiid_16,b25,jsCall_iiiiid_17,b25,jsCall_iiiiid_18,b25,jsCall_iiiiid_19,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25
,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,b25,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEcd,__ZNKSt3__27num_putIcNS_19ostreambuf_iteratorIcNS_11char_traitsIcEEEEE6do_putES4_RNS_8ios_baseEce,b25,b25,b25,b25,b25,b25,b25,b25,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwd
,__ZNKSt3__27num_putIwNS_19ostreambuf_iteratorIwNS_11char_traitsIwEEEEE6do_putES4_RNS_8ios_baseEwe,b25,b25];
var FUNCTION_TABLE_viiii = [b26,b26,jsCall_viiii_0,b26,jsCall_viiii_1,b26,jsCall_viiii_2,b26,jsCall_viiii_3,b26,jsCall_viiii_4,b26,jsCall_viiii_5,b26,jsCall_viiii_6,b26,jsCall_viiii_7,b26,jsCall_viiii_8,b26,jsCall_viiii_9,b26,jsCall_viiii_10,b26,jsCall_viiii_11,b26,jsCall_viiii_12,b26,jsCall_viiii_13
,b26,jsCall_viiii_14,b26,jsCall_viiii_15,b26,jsCall_viiii_16,b26,jsCall_viiii_17,b26,jsCall_viiii_18,b26,jsCall_viiii_19,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,__ZN9ExitGames6Photon8Internal8PeerBase16deserializeEventEPhbi,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZN9ExitGames6Photon8Internal5TPeer13onReceiveDataEPhii,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZNSt3__215basic_streambufIcNS_11char_traitsIcEEE7seekposENS_4fposI11__mbstate_tEEj,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,__ZNSt3__215basic_streambufIwNS_11char_traitsIwEEE7seekposENS_4fposI11__mbstate_tEEj,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZNKSt3__27collateIcE12do_transformEPKcS3_,b26,b26,b26,b26,__ZNKSt3__27collateIwE12do_transformEPKwS3_,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZNK10__cxxabiv117__class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b26,b26,b26,__ZNK10__cxxabiv120__si_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,__ZNK10__cxxabiv121__vmi_class_type_info27has_unambiguous_public_baseEPNS_19__dynamic_cast_infoEPvi,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,__ZNSt3__26__treeINS_12__value_typeIPN9ExitGames6Photon10PhotonPeerEP16PhotonListenerEMEENS_19__map_value_compareIS5_S8_NS_4lessIS5_EELb1EEENS_9allocatorIS8_EEE16__insert_node_atEPNS_15__tree_end_nodeIPNS_16__tree_node_baseIPvEEEERSK_SK_,__ZN9ExitGames6Common10DictionaryIhNS0_6ObjectEE11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj,b26,b26,b26,__ZN9ExitGames6Common10DictionaryIiNS0_6ObjectEE11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj,__ZN9ExitGames6Common10DictionaryINS0_6ObjectES2_E11NestedInfosIS3_S3_E10fillArraysEPhS6_Pjj,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZN9ExitGames6Common16MemoryManagement8Internal10MemoryPool4initEjjj,b26,b26,b26,b26,b26,b26,b26,__ZNK9ExitGames6Common7JString7replaceERKS1_S3_,b26,b26,__ZNK9ExitGames6Common7JString9substringEjj,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,__ZN9ExitGames6Common7Helpers26DeSerializerImplementation22popArrayOfDictionariesERNS0_6ObjectEsi,b26,__ZN9ExitGames6Common14DictionaryBase8TypeInfo4initEPKhS4_PKj,b26,b26,__ZNK9ExitGames6Common14DictionaryBase8TypeInfo16toStringRecursorEbj,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,__ZNSt3__216__check_groupingERKNS_12basic_stringIcNS_11char_traitsIcEENS_9allocatorIcEEEEPjS8_Rj,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,__ZN10__cxxabiv112_GLOBAL__N_18demangleINS0_2DbEEEvPKcS4_RT_Ri,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26,b26
,b26,b26,b26,b26,b26,b26,b26,b26,b26];

  return { _PhotonTypesTable: _PhotonTypesTable, _PhotonPeerService: _PhotonPeerService, _PhotonTypesBoolAt: _PhotonTypesBoolAt, _bitshift64Lshr: _bitshift64Lshr, dynCall_iiiiii: dynCall_iiiiii, _bitshift64Ashr: _bitshift64Ashr, __GLOBAL__sub_I_PeerData_cpp: __GLOBAL__sub_I_PeerData_cpp, _sbrk: _sbrk, _memcpy: _memcpy, _PhotonTypesDouble: _PhotonTypesDouble, _PhotonTypesArray: _PhotonTypesArray, _PhotonTypesTablePut: _PhotonTypesTablePut, dynCall_vi: dynCall_vi, _PhotonPeerState: _PhotonPeerState, _PhotonTypesShort: _PhotonTypesShort, _PhotonTypesTableKeyAt: _PhotonTypesTableKeyAt, dynCall_viiiiiiii: dynCall_viiiiiiii, dynCall_iiii: dynCall_iiii, __GLOBAL__I_000101: __GLOBAL__I_000101, dynCall_iiiiid: dynCall_iiiiid, _emscripten_get_global_libc: _emscripten_get_global_libc, dynCall_i: dynCall_i, dynCall_iiiiiiiiiiii: dynCall_iiiiiiiiiiii, _PhotonTypesString: _PhotonTypesString, stackSave: stackSave, dynCall_viiiii: dynCall_viiiii, ___cxa_can_catch: ___cxa_can_catch, _PhotonTypesDictBaseFromDict: _PhotonTypesDictBaseFromDict, _free: _free, runPostSets: runPostSets, _PhotonPeerDestroy: _PhotonPeerDestroy, _PhotonTypesGetBool: _PhotonTypesGetBool, _PhotonPeerOpCustom: _PhotonPeerOpCustom, _PhotonTypesGetInt: _PhotonTypesGetInt, _PhotonTypesLongLowAt: _PhotonTypesLongLowAt, _PhotonPeerConnect: _PhotonPeerConnect, _PhotonPeerSetDebugOutputLevel: _PhotonPeerSetDebugOutputLevel, _PhotonTypesByte: _PhotonTypesByte, _PhotonPeerCreate: _PhotonPeerCreate, _PhotonTypesDictValueAt: _PhotonTypesDictValueAt, _PhotonTypesGetLongLow: _PhotonTypesGetLongLow, dynCall_iiiiiiiiiiiii: dynCall_iiiiiiiiiiiii, ___cxa_is_pointer_type: ___cxa_is_pointer_type, _PhotonTypesDictBasePut: _PhotonTypesDictBasePut, dynCall_diii: dynCall_diii, stackAlloc: stackAlloc, dynCall_viiiiii: dynCall_viiiiii, getTempRet0: getTempRet0, _ntohs: _ntohs, _PhotonTypesIsArray: _PhotonTypesIsArray, dynCall_viiii: dynCall_viiii, _llvm_bswap_i32: _llvm_bswap_i32, establishStackSpace: establishStackSpace, _PhotonPeerDisconnect: _PhotonPeerDisconnect, _pthread_mutex_lock: _pthread_mutex_lock, _PhotonTypesByteAt: _PhotonTypesByteAt, _PhotonTypesDict: _PhotonTypesDict, _PhotonTypesShortAt: _PhotonTypesShortAt, _PhotonTypesGetFloat: _PhotonTypesGetFloat, dynCall_viiiiiiiiii: dynCall_viiiiiiiiii, dynCall_iiiiiii: dynCall_iiiiiii, dynCall_viii: dynCall_viii, _PhotonTypesInt: _PhotonTypesInt, _PhotonTypesFloatAt: _PhotonTypesFloatAt, _PhotonTypesBool: _PhotonTypesBool, dynCall_iiiiiid: dynCall_iiiiiid, _PhotonTypesTableSize: _PhotonTypesTableSize, dynCall_iiiiiiiii: dynCall_iiiiiiiii, dynCall_iid: dynCall_iid, dynCall_iii: dynCall_iii, _memset: _memset, _PhotonTypesGetDouble: _PhotonTypesGetDouble, ___errno_location: ___errno_location, dynCall_viiiiiii: dynCall_viiiiiii, _i64Subtract: _i64Subtract, _PhotonTypesArraySize: _PhotonTypesArraySize, setTempRet0: setTempRet0, _PhotonTypesDestroy: _PhotonTypesDestroy, _llvm_bswap_i16: _llvm_bswap_i16, __GLOBAL__sub_I_iostream_cpp: __GLOBAL__sub_I_iostream_cpp, _PhotonTypesByteDictSize: _PhotonTypesByteDictSize, _pthread_cond_broadcast: _pthread_cond_broadcast, _PhotonTypesGetShort: _PhotonTypesGetShort, _PhotonOperationRequestSetParam: _PhotonOperationRequestSetParam, _PhotonTypesDoubleAt: _PhotonTypesDoubleAt, dynCall_iiiiiiii: dynCall_iiiiiiii, _PhotonTypesDictBase: _PhotonTypesDictBase, setThrew: setThrew, _malloc: _malloc, _emscripten_replace_memory: _emscripten_replace_memory, __GLOBAL__sub_I_Base_cpp: __GLOBAL__sub_I_Base_cpp, dynCall_v: dynCall_v, _PhotonTypesGetTable: _PhotonTypesGetTable, _PhotonTypesByteDictPut: _PhotonTypesByteDictPut, ___udivdi3: ___udivdi3, _PhotonTypesGetByte: _PhotonTypesGetByte, _PhotonOperationRequestDestroy: _PhotonOperationRequestDestroy, _bitshift64Shl: _bitshift64Shl, _PhotonTypesStringAt: _PhotonTypesStringAt, _PhotonTypesGetLongHigh: _PhotonTypesGetLongHigh, __GLOBAL__sub_I_PhotonPeer_cpp: __GLOBAL__sub_I_PhotonPeer_cpp, _PhotonTypesGetString: _PhotonTypesGetString, ___cxa_demangle: ___cxa_demangle, _PhotonTypesGetType: _PhotonTypesGetType, ___muldi3: ___muldi3, _PhotonTypesByteDictValueAt: _PhotonTypesByteDictValueAt, _PhotonTypesFloat: _PhotonTypesFloat, ___uremdi3: ___uremdi3, dynCall_vii: dynCall_vii, _htonl: _htonl, _i64Add: _i64Add, _PhotonTypesLongHighAt: _PhotonTypesLongHighAt, _pthread_mutex_unlock: _pthread_mutex_unlock, stackRestore: stackRestore, dynCall_iiiii: dynCall_iiiii, dynCall_iiiiiiiiiii: dynCall_iiiiiiiiiii, _htons: _htons, _fflush: _fflush, _PhotonTypesByteDictKeyAt: _PhotonTypesByteDictKeyAt, _PhotonTypesLong: _PhotonTypesLong, dynCall_viiiiiiiiiiiiiii: dynCall_viiiiiiiiiiiiiii, _PhotonTypesDictSize: _PhotonTypesDictSize, dynCall_ii: dynCall_ii, __GLOBAL__sub_I_ToString_cpp: __GLOBAL__sub_I_ToString_cpp, _memmove: _memmove, _PhotonOperationRequestCreate: _PhotonOperationRequestCreate, _PhotonTypesDictKeyAt: _PhotonTypesDictKeyAt, _PhotonTypesTableAt: _PhotonTypesTableAt, _PhotonTypesIntAt: _PhotonTypesIntAt, _PhotonTypesTableValueAt: _PhotonTypesTableValueAt, _PhotonTypesObjectAt: _PhotonTypesObjectAt };
})
// EMSCRIPTEN_END_ASM
(Module.asmGlobalArg, Module.asmLibraryArg, buffer);

var real__PhotonTypesTable = asm["_PhotonTypesTable"]; asm["_PhotonTypesTable"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTable.apply(null, arguments);
};

var real__PhotonPeerService = asm["_PhotonPeerService"]; asm["_PhotonPeerService"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerService.apply(null, arguments);
};

var real__PhotonTypesBoolAt = asm["_PhotonTypesBoolAt"]; asm["_PhotonTypesBoolAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesBoolAt.apply(null, arguments);
};

var real__bitshift64Lshr = asm["_bitshift64Lshr"]; asm["_bitshift64Lshr"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__bitshift64Lshr.apply(null, arguments);
};

var real__bitshift64Ashr = asm["_bitshift64Ashr"]; asm["_bitshift64Ashr"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__bitshift64Ashr.apply(null, arguments);
};

var real___GLOBAL__sub_I_PeerData_cpp = asm["__GLOBAL__sub_I_PeerData_cpp"]; asm["__GLOBAL__sub_I_PeerData_cpp"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__sub_I_PeerData_cpp.apply(null, arguments);
};

var real__sbrk = asm["_sbrk"]; asm["_sbrk"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__sbrk.apply(null, arguments);
};

var real__PhotonTypesDouble = asm["_PhotonTypesDouble"]; asm["_PhotonTypesDouble"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDouble.apply(null, arguments);
};

var real__PhotonTypesGetShort = asm["_PhotonTypesGetShort"]; asm["_PhotonTypesGetShort"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetShort.apply(null, arguments);
};

var real__PhotonOperationRequestSetParam = asm["_PhotonOperationRequestSetParam"]; asm["_PhotonOperationRequestSetParam"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonOperationRequestSetParam.apply(null, arguments);
};

var real__PhotonPeerState = asm["_PhotonPeerState"]; asm["_PhotonPeerState"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerState.apply(null, arguments);
};

var real__PhotonTypesShort = asm["_PhotonTypesShort"]; asm["_PhotonTypesShort"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesShort.apply(null, arguments);
};

var real__PhotonTypesTableKeyAt = asm["_PhotonTypesTableKeyAt"]; asm["_PhotonTypesTableKeyAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTableKeyAt.apply(null, arguments);
};

var real___GLOBAL__I_000101 = asm["__GLOBAL__I_000101"]; asm["__GLOBAL__I_000101"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__I_000101.apply(null, arguments);
};

var real__emscripten_get_global_libc = asm["_emscripten_get_global_libc"]; asm["_emscripten_get_global_libc"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__emscripten_get_global_libc.apply(null, arguments);
};

var real__PhotonTypesString = asm["_PhotonTypesString"]; asm["_PhotonTypesString"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesString.apply(null, arguments);
};

var real_stackSave = asm["stackSave"]; asm["stackSave"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_stackSave.apply(null, arguments);
};

var real____cxa_can_catch = asm["___cxa_can_catch"]; asm["___cxa_can_catch"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____cxa_can_catch.apply(null, arguments);
};

var real__PhotonTypesDictBaseFromDict = asm["_PhotonTypesDictBaseFromDict"]; asm["_PhotonTypesDictBaseFromDict"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictBaseFromDict.apply(null, arguments);
};

var real__free = asm["_free"]; asm["_free"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__free.apply(null, arguments);
};

var real__PhotonPeerDestroy = asm["_PhotonPeerDestroy"]; asm["_PhotonPeerDestroy"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerDestroy.apply(null, arguments);
};

var real__PhotonTypesGetBool = asm["_PhotonTypesGetBool"]; asm["_PhotonTypesGetBool"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetBool.apply(null, arguments);
};

var real__PhotonPeerOpCustom = asm["_PhotonPeerOpCustom"]; asm["_PhotonPeerOpCustom"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerOpCustom.apply(null, arguments);
};

var real__PhotonTypesGetInt = asm["_PhotonTypesGetInt"]; asm["_PhotonTypesGetInt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetInt.apply(null, arguments);
};

var real__PhotonTypesLongLowAt = asm["_PhotonTypesLongLowAt"]; asm["_PhotonTypesLongLowAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesLongLowAt.apply(null, arguments);
};

var real__PhotonPeerConnect = asm["_PhotonPeerConnect"]; asm["_PhotonPeerConnect"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerConnect.apply(null, arguments);
};

var real__PhotonPeerSetDebugOutputLevel = asm["_PhotonPeerSetDebugOutputLevel"]; asm["_PhotonPeerSetDebugOutputLevel"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerSetDebugOutputLevel.apply(null, arguments);
};

var real__PhotonTypesByte = asm["_PhotonTypesByte"]; asm["_PhotonTypesByte"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByte.apply(null, arguments);
};

var real__PhotonPeerCreate = asm["_PhotonPeerCreate"]; asm["_PhotonPeerCreate"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerCreate.apply(null, arguments);
};

var real__PhotonTypesDictValueAt = asm["_PhotonTypesDictValueAt"]; asm["_PhotonTypesDictValueAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictValueAt.apply(null, arguments);
};

var real__PhotonTypesGetLongLow = asm["_PhotonTypesGetLongLow"]; asm["_PhotonTypesGetLongLow"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetLongLow.apply(null, arguments);
};

var real____cxa_is_pointer_type = asm["___cxa_is_pointer_type"]; asm["___cxa_is_pointer_type"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____cxa_is_pointer_type.apply(null, arguments);
};

var real__PhotonTypesDictBasePut = asm["_PhotonTypesDictBasePut"]; asm["_PhotonTypesDictBasePut"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictBasePut.apply(null, arguments);
};

var real_stackAlloc = asm["stackAlloc"]; asm["stackAlloc"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_stackAlloc.apply(null, arguments);
};

var real_getTempRet0 = asm["getTempRet0"]; asm["getTempRet0"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_getTempRet0.apply(null, arguments);
};

var real__ntohs = asm["_ntohs"]; asm["_ntohs"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__ntohs.apply(null, arguments);
};

var real__PhotonTypesIsArray = asm["_PhotonTypesIsArray"]; asm["_PhotonTypesIsArray"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesIsArray.apply(null, arguments);
};

var real__llvm_bswap_i32 = asm["_llvm_bswap_i32"]; asm["_llvm_bswap_i32"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__llvm_bswap_i32.apply(null, arguments);
};

var real_establishStackSpace = asm["establishStackSpace"]; asm["establishStackSpace"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_establishStackSpace.apply(null, arguments);
};

var real_stackRestore = asm["stackRestore"]; asm["stackRestore"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_stackRestore.apply(null, arguments);
};

var real__pthread_mutex_lock = asm["_pthread_mutex_lock"]; asm["_pthread_mutex_lock"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__pthread_mutex_lock.apply(null, arguments);
};

var real__PhotonTypesGetType = asm["_PhotonTypesGetType"]; asm["_PhotonTypesGetType"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetType.apply(null, arguments);
};

var real__PhotonTypesDict = asm["_PhotonTypesDict"]; asm["_PhotonTypesDict"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDict.apply(null, arguments);
};

var real__PhotonTypesShortAt = asm["_PhotonTypesShortAt"]; asm["_PhotonTypesShortAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesShortAt.apply(null, arguments);
};

var real__PhotonTypesGetFloat = asm["_PhotonTypesGetFloat"]; asm["_PhotonTypesGetFloat"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetFloat.apply(null, arguments);
};

var real__PhotonTypesInt = asm["_PhotonTypesInt"]; asm["_PhotonTypesInt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesInt.apply(null, arguments);
};

var real__PhotonTypesFloatAt = asm["_PhotonTypesFloatAt"]; asm["_PhotonTypesFloatAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesFloatAt.apply(null, arguments);
};

var real__PhotonTypesBool = asm["_PhotonTypesBool"]; asm["_PhotonTypesBool"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesBool.apply(null, arguments);
};

var real__PhotonTypesTableSize = asm["_PhotonTypesTableSize"]; asm["_PhotonTypesTableSize"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTableSize.apply(null, arguments);
};

var real__PhotonTypesGetDouble = asm["_PhotonTypesGetDouble"]; asm["_PhotonTypesGetDouble"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetDouble.apply(null, arguments);
};

var real__PhotonTypesLong = asm["_PhotonTypesLong"]; asm["_PhotonTypesLong"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesLong.apply(null, arguments);
};

var real__i64Subtract = asm["_i64Subtract"]; asm["_i64Subtract"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__i64Subtract.apply(null, arguments);
};

var real__PhotonTypesArraySize = asm["_PhotonTypesArraySize"]; asm["_PhotonTypesArraySize"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesArraySize.apply(null, arguments);
};

var real_setTempRet0 = asm["setTempRet0"]; asm["setTempRet0"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_setTempRet0.apply(null, arguments);
};

var real__PhotonTypesDestroy = asm["_PhotonTypesDestroy"]; asm["_PhotonTypesDestroy"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDestroy.apply(null, arguments);
};

var real__llvm_bswap_i16 = asm["_llvm_bswap_i16"]; asm["_llvm_bswap_i16"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__llvm_bswap_i16.apply(null, arguments);
};

var real___GLOBAL__sub_I_iostream_cpp = asm["__GLOBAL__sub_I_iostream_cpp"]; asm["__GLOBAL__sub_I_iostream_cpp"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__sub_I_iostream_cpp.apply(null, arguments);
};

var real__PhotonTypesByteDictSize = asm["_PhotonTypesByteDictSize"]; asm["_PhotonTypesByteDictSize"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByteDictSize.apply(null, arguments);
};

var real__pthread_cond_broadcast = asm["_pthread_cond_broadcast"]; asm["_pthread_cond_broadcast"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__pthread_cond_broadcast.apply(null, arguments);
};

var real__PhotonTypesArray = asm["_PhotonTypesArray"]; asm["_PhotonTypesArray"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesArray.apply(null, arguments);
};

var real__PhotonTypesTablePut = asm["_PhotonTypesTablePut"]; asm["_PhotonTypesTablePut"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTablePut.apply(null, arguments);
};

var real__PhotonTypesDoubleAt = asm["_PhotonTypesDoubleAt"]; asm["_PhotonTypesDoubleAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDoubleAt.apply(null, arguments);
};

var real__PhotonTypesDictBase = asm["_PhotonTypesDictBase"]; asm["_PhotonTypesDictBase"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictBase.apply(null, arguments);
};

var real_setThrew = asm["setThrew"]; asm["setThrew"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real_setThrew.apply(null, arguments);
};

var real__malloc = asm["_malloc"]; asm["_malloc"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__malloc.apply(null, arguments);
};

var real___GLOBAL__sub_I_Base_cpp = asm["__GLOBAL__sub_I_Base_cpp"]; asm["__GLOBAL__sub_I_Base_cpp"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__sub_I_Base_cpp.apply(null, arguments);
};

var real__PhotonTypesGetTable = asm["_PhotonTypesGetTable"]; asm["_PhotonTypesGetTable"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetTable.apply(null, arguments);
};

var real__PhotonTypesByteDictPut = asm["_PhotonTypesByteDictPut"]; asm["_PhotonTypesByteDictPut"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByteDictPut.apply(null, arguments);
};

var real____udivdi3 = asm["___udivdi3"]; asm["___udivdi3"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____udivdi3.apply(null, arguments);
};

var real__PhotonTypesGetByte = asm["_PhotonTypesGetByte"]; asm["_PhotonTypesGetByte"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetByte.apply(null, arguments);
};

var real__PhotonOperationRequestDestroy = asm["_PhotonOperationRequestDestroy"]; asm["_PhotonOperationRequestDestroy"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonOperationRequestDestroy.apply(null, arguments);
};

var real__bitshift64Shl = asm["_bitshift64Shl"]; asm["_bitshift64Shl"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__bitshift64Shl.apply(null, arguments);
};

var real__PhotonTypesStringAt = asm["_PhotonTypesStringAt"]; asm["_PhotonTypesStringAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesStringAt.apply(null, arguments);
};

var real__PhotonTypesGetLongHigh = asm["_PhotonTypesGetLongHigh"]; asm["_PhotonTypesGetLongHigh"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetLongHigh.apply(null, arguments);
};

var real___GLOBAL__sub_I_PhotonPeer_cpp = asm["__GLOBAL__sub_I_PhotonPeer_cpp"]; asm["__GLOBAL__sub_I_PhotonPeer_cpp"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__sub_I_PhotonPeer_cpp.apply(null, arguments);
};

var real__PhotonTypesGetString = asm["_PhotonTypesGetString"]; asm["_PhotonTypesGetString"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesGetString.apply(null, arguments);
};

var real____cxa_demangle = asm["___cxa_demangle"]; asm["___cxa_demangle"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____cxa_demangle.apply(null, arguments);
};

var real__PhotonTypesByteAt = asm["_PhotonTypesByteAt"]; asm["_PhotonTypesByteAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByteAt.apply(null, arguments);
};

var real____muldi3 = asm["___muldi3"]; asm["___muldi3"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____muldi3.apply(null, arguments);
};

var real__PhotonTypesByteDictValueAt = asm["_PhotonTypesByteDictValueAt"]; asm["_PhotonTypesByteDictValueAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByteDictValueAt.apply(null, arguments);
};

var real__PhotonTypesFloat = asm["_PhotonTypesFloat"]; asm["_PhotonTypesFloat"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesFloat.apply(null, arguments);
};

var real____uremdi3 = asm["___uremdi3"]; asm["___uremdi3"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____uremdi3.apply(null, arguments);
};

var real__htonl = asm["_htonl"]; asm["_htonl"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__htonl.apply(null, arguments);
};

var real__i64Add = asm["_i64Add"]; asm["_i64Add"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__i64Add.apply(null, arguments);
};

var real__PhotonTypesLongHighAt = asm["_PhotonTypesLongHighAt"]; asm["_PhotonTypesLongHighAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesLongHighAt.apply(null, arguments);
};

var real__pthread_mutex_unlock = asm["_pthread_mutex_unlock"]; asm["_pthread_mutex_unlock"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__pthread_mutex_unlock.apply(null, arguments);
};

var real__PhotonPeerDisconnect = asm["_PhotonPeerDisconnect"]; asm["_PhotonPeerDisconnect"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonPeerDisconnect.apply(null, arguments);
};

var real__htons = asm["_htons"]; asm["_htons"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__htons.apply(null, arguments);
};

var real__fflush = asm["_fflush"]; asm["_fflush"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__fflush.apply(null, arguments);
};

var real__PhotonTypesByteDictKeyAt = asm["_PhotonTypesByteDictKeyAt"]; asm["_PhotonTypesByteDictKeyAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesByteDictKeyAt.apply(null, arguments);
};

var real____errno_location = asm["___errno_location"]; asm["___errno_location"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real____errno_location.apply(null, arguments);
};

var real__PhotonTypesDictSize = asm["_PhotonTypesDictSize"]; asm["_PhotonTypesDictSize"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictSize.apply(null, arguments);
};

var real___GLOBAL__sub_I_ToString_cpp = asm["__GLOBAL__sub_I_ToString_cpp"]; asm["__GLOBAL__sub_I_ToString_cpp"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real___GLOBAL__sub_I_ToString_cpp.apply(null, arguments);
};

var real__memmove = asm["_memmove"]; asm["_memmove"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__memmove.apply(null, arguments);
};

var real__PhotonOperationRequestCreate = asm["_PhotonOperationRequestCreate"]; asm["_PhotonOperationRequestCreate"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonOperationRequestCreate.apply(null, arguments);
};

var real__PhotonTypesDictKeyAt = asm["_PhotonTypesDictKeyAt"]; asm["_PhotonTypesDictKeyAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesDictKeyAt.apply(null, arguments);
};

var real__PhotonTypesTableAt = asm["_PhotonTypesTableAt"]; asm["_PhotonTypesTableAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTableAt.apply(null, arguments);
};

var real__PhotonTypesIntAt = asm["_PhotonTypesIntAt"]; asm["_PhotonTypesIntAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesIntAt.apply(null, arguments);
};

var real__PhotonTypesTableValueAt = asm["_PhotonTypesTableValueAt"]; asm["_PhotonTypesTableValueAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesTableValueAt.apply(null, arguments);
};

var real__PhotonTypesObjectAt = asm["_PhotonTypesObjectAt"]; asm["_PhotonTypesObjectAt"] = function() {
  assert(runtimeInitialized, 'you need to wait for the runtime to be ready (e.g. wait for main() to be called)');
  assert(!runtimeExited, 'the runtime was exited (use NO_EXIT_RUNTIME to keep it alive after main() exits)');
  return real__PhotonTypesObjectAt.apply(null, arguments);
};
var _PhotonTypesTable = Module["_PhotonTypesTable"] = asm["_PhotonTypesTable"];
var _PhotonPeerService = Module["_PhotonPeerService"] = asm["_PhotonPeerService"];
var _PhotonTypesBoolAt = Module["_PhotonTypesBoolAt"] = asm["_PhotonTypesBoolAt"];
var _bitshift64Lshr = Module["_bitshift64Lshr"] = asm["_bitshift64Lshr"];
var _bitshift64Ashr = Module["_bitshift64Ashr"] = asm["_bitshift64Ashr"];
var __GLOBAL__sub_I_PeerData_cpp = Module["__GLOBAL__sub_I_PeerData_cpp"] = asm["__GLOBAL__sub_I_PeerData_cpp"];
var _sbrk = Module["_sbrk"] = asm["_sbrk"];
var _memcpy = Module["_memcpy"] = asm["_memcpy"];
var _PhotonTypesDouble = Module["_PhotonTypesDouble"] = asm["_PhotonTypesDouble"];
var _PhotonTypesGetShort = Module["_PhotonTypesGetShort"] = asm["_PhotonTypesGetShort"];
var _PhotonOperationRequestSetParam = Module["_PhotonOperationRequestSetParam"] = asm["_PhotonOperationRequestSetParam"];
var _PhotonPeerState = Module["_PhotonPeerState"] = asm["_PhotonPeerState"];
var _PhotonTypesShort = Module["_PhotonTypesShort"] = asm["_PhotonTypesShort"];
var _PhotonTypesTableKeyAt = Module["_PhotonTypesTableKeyAt"] = asm["_PhotonTypesTableKeyAt"];
var __GLOBAL__I_000101 = Module["__GLOBAL__I_000101"] = asm["__GLOBAL__I_000101"];
var _emscripten_get_global_libc = Module["_emscripten_get_global_libc"] = asm["_emscripten_get_global_libc"];
var _PhotonTypesString = Module["_PhotonTypesString"] = asm["_PhotonTypesString"];
var stackSave = Module["stackSave"] = asm["stackSave"];
var ___cxa_can_catch = Module["___cxa_can_catch"] = asm["___cxa_can_catch"];
var _PhotonTypesDictBaseFromDict = Module["_PhotonTypesDictBaseFromDict"] = asm["_PhotonTypesDictBaseFromDict"];
var _free = Module["_free"] = asm["_free"];
var runPostSets = Module["runPostSets"] = asm["runPostSets"];
var _PhotonPeerDestroy = Module["_PhotonPeerDestroy"] = asm["_PhotonPeerDestroy"];
var _PhotonTypesGetBool = Module["_PhotonTypesGetBool"] = asm["_PhotonTypesGetBool"];
var _PhotonPeerOpCustom = Module["_PhotonPeerOpCustom"] = asm["_PhotonPeerOpCustom"];
var _PhotonTypesGetInt = Module["_PhotonTypesGetInt"] = asm["_PhotonTypesGetInt"];
var _PhotonTypesLongLowAt = Module["_PhotonTypesLongLowAt"] = asm["_PhotonTypesLongLowAt"];
var _PhotonPeerConnect = Module["_PhotonPeerConnect"] = asm["_PhotonPeerConnect"];
var _PhotonPeerSetDebugOutputLevel = Module["_PhotonPeerSetDebugOutputLevel"] = asm["_PhotonPeerSetDebugOutputLevel"];
var _PhotonTypesByte = Module["_PhotonTypesByte"] = asm["_PhotonTypesByte"];
var _PhotonPeerCreate = Module["_PhotonPeerCreate"] = asm["_PhotonPeerCreate"];
var _PhotonTypesDictValueAt = Module["_PhotonTypesDictValueAt"] = asm["_PhotonTypesDictValueAt"];
var _PhotonTypesGetLongLow = Module["_PhotonTypesGetLongLow"] = asm["_PhotonTypesGetLongLow"];
var ___cxa_is_pointer_type = Module["___cxa_is_pointer_type"] = asm["___cxa_is_pointer_type"];
var _PhotonTypesDictBasePut = Module["_PhotonTypesDictBasePut"] = asm["_PhotonTypesDictBasePut"];
var stackAlloc = Module["stackAlloc"] = asm["stackAlloc"];
var getTempRet0 = Module["getTempRet0"] = asm["getTempRet0"];
var _ntohs = Module["_ntohs"] = asm["_ntohs"];
var _PhotonTypesIsArray = Module["_PhotonTypesIsArray"] = asm["_PhotonTypesIsArray"];
var _llvm_bswap_i32 = Module["_llvm_bswap_i32"] = asm["_llvm_bswap_i32"];
var establishStackSpace = Module["establishStackSpace"] = asm["establishStackSpace"];
var stackRestore = Module["stackRestore"] = asm["stackRestore"];
var _pthread_mutex_lock = Module["_pthread_mutex_lock"] = asm["_pthread_mutex_lock"];
var _PhotonTypesGetType = Module["_PhotonTypesGetType"] = asm["_PhotonTypesGetType"];
var _PhotonTypesDict = Module["_PhotonTypesDict"] = asm["_PhotonTypesDict"];
var _PhotonTypesShortAt = Module["_PhotonTypesShortAt"] = asm["_PhotonTypesShortAt"];
var _PhotonTypesGetFloat = Module["_PhotonTypesGetFloat"] = asm["_PhotonTypesGetFloat"];
var _PhotonTypesInt = Module["_PhotonTypesInt"] = asm["_PhotonTypesInt"];
var _PhotonTypesFloatAt = Module["_PhotonTypesFloatAt"] = asm["_PhotonTypesFloatAt"];
var _PhotonTypesBool = Module["_PhotonTypesBool"] = asm["_PhotonTypesBool"];
var _PhotonTypesTableSize = Module["_PhotonTypesTableSize"] = asm["_PhotonTypesTableSize"];
var _memset = Module["_memset"] = asm["_memset"];
var _PhotonTypesGetDouble = Module["_PhotonTypesGetDouble"] = asm["_PhotonTypesGetDouble"];
var _PhotonTypesLong = Module["_PhotonTypesLong"] = asm["_PhotonTypesLong"];
var _i64Subtract = Module["_i64Subtract"] = asm["_i64Subtract"];
var _PhotonTypesArraySize = Module["_PhotonTypesArraySize"] = asm["_PhotonTypesArraySize"];
var setTempRet0 = Module["setTempRet0"] = asm["setTempRet0"];
var _PhotonTypesDestroy = Module["_PhotonTypesDestroy"] = asm["_PhotonTypesDestroy"];
var _llvm_bswap_i16 = Module["_llvm_bswap_i16"] = asm["_llvm_bswap_i16"];
var __GLOBAL__sub_I_iostream_cpp = Module["__GLOBAL__sub_I_iostream_cpp"] = asm["__GLOBAL__sub_I_iostream_cpp"];
var _PhotonTypesByteDictSize = Module["_PhotonTypesByteDictSize"] = asm["_PhotonTypesByteDictSize"];
var _pthread_cond_broadcast = Module["_pthread_cond_broadcast"] = asm["_pthread_cond_broadcast"];
var _PhotonTypesArray = Module["_PhotonTypesArray"] = asm["_PhotonTypesArray"];
var _PhotonTypesTablePut = Module["_PhotonTypesTablePut"] = asm["_PhotonTypesTablePut"];
var _PhotonTypesDoubleAt = Module["_PhotonTypesDoubleAt"] = asm["_PhotonTypesDoubleAt"];
var _PhotonTypesDictBase = Module["_PhotonTypesDictBase"] = asm["_PhotonTypesDictBase"];
var setThrew = Module["setThrew"] = asm["setThrew"];
var _malloc = Module["_malloc"] = asm["_malloc"];
var _emscripten_replace_memory = Module["_emscripten_replace_memory"] = asm["_emscripten_replace_memory"];
var __GLOBAL__sub_I_Base_cpp = Module["__GLOBAL__sub_I_Base_cpp"] = asm["__GLOBAL__sub_I_Base_cpp"];
var _PhotonTypesGetTable = Module["_PhotonTypesGetTable"] = asm["_PhotonTypesGetTable"];
var _PhotonTypesByteDictPut = Module["_PhotonTypesByteDictPut"] = asm["_PhotonTypesByteDictPut"];
var ___udivdi3 = Module["___udivdi3"] = asm["___udivdi3"];
var _PhotonTypesGetByte = Module["_PhotonTypesGetByte"] = asm["_PhotonTypesGetByte"];
var _PhotonOperationRequestDestroy = Module["_PhotonOperationRequestDestroy"] = asm["_PhotonOperationRequestDestroy"];
var _bitshift64Shl = Module["_bitshift64Shl"] = asm["_bitshift64Shl"];
var _PhotonTypesStringAt = Module["_PhotonTypesStringAt"] = asm["_PhotonTypesStringAt"];
var _PhotonTypesGetLongHigh = Module["_PhotonTypesGetLongHigh"] = asm["_PhotonTypesGetLongHigh"];
var __GLOBAL__sub_I_PhotonPeer_cpp = Module["__GLOBAL__sub_I_PhotonPeer_cpp"] = asm["__GLOBAL__sub_I_PhotonPeer_cpp"];
var _PhotonTypesGetString = Module["_PhotonTypesGetString"] = asm["_PhotonTypesGetString"];
var ___cxa_demangle = Module["___cxa_demangle"] = asm["___cxa_demangle"];
var _PhotonTypesByteAt = Module["_PhotonTypesByteAt"] = asm["_PhotonTypesByteAt"];
var ___muldi3 = Module["___muldi3"] = asm["___muldi3"];
var _PhotonTypesByteDictValueAt = Module["_PhotonTypesByteDictValueAt"] = asm["_PhotonTypesByteDictValueAt"];
var _PhotonTypesFloat = Module["_PhotonTypesFloat"] = asm["_PhotonTypesFloat"];
var ___uremdi3 = Module["___uremdi3"] = asm["___uremdi3"];
var _htonl = Module["_htonl"] = asm["_htonl"];
var _i64Add = Module["_i64Add"] = asm["_i64Add"];
var _PhotonTypesLongHighAt = Module["_PhotonTypesLongHighAt"] = asm["_PhotonTypesLongHighAt"];
var _pthread_mutex_unlock = Module["_pthread_mutex_unlock"] = asm["_pthread_mutex_unlock"];
var _PhotonPeerDisconnect = Module["_PhotonPeerDisconnect"] = asm["_PhotonPeerDisconnect"];
var _htons = Module["_htons"] = asm["_htons"];
var _fflush = Module["_fflush"] = asm["_fflush"];
var _PhotonTypesByteDictKeyAt = Module["_PhotonTypesByteDictKeyAt"] = asm["_PhotonTypesByteDictKeyAt"];
var ___errno_location = Module["___errno_location"] = asm["___errno_location"];
var _PhotonTypesDictSize = Module["_PhotonTypesDictSize"] = asm["_PhotonTypesDictSize"];
var __GLOBAL__sub_I_ToString_cpp = Module["__GLOBAL__sub_I_ToString_cpp"] = asm["__GLOBAL__sub_I_ToString_cpp"];
var _memmove = Module["_memmove"] = asm["_memmove"];
var _PhotonOperationRequestCreate = Module["_PhotonOperationRequestCreate"] = asm["_PhotonOperationRequestCreate"];
var _PhotonTypesDictKeyAt = Module["_PhotonTypesDictKeyAt"] = asm["_PhotonTypesDictKeyAt"];
var _PhotonTypesTableAt = Module["_PhotonTypesTableAt"] = asm["_PhotonTypesTableAt"];
var _PhotonTypesIntAt = Module["_PhotonTypesIntAt"] = asm["_PhotonTypesIntAt"];
var _PhotonTypesTableValueAt = Module["_PhotonTypesTableValueAt"] = asm["_PhotonTypesTableValueAt"];
var _PhotonTypesObjectAt = Module["_PhotonTypesObjectAt"] = asm["_PhotonTypesObjectAt"];
var dynCall_iiiiiiii = Module["dynCall_iiiiiiii"] = asm["dynCall_iiiiiiii"];
var dynCall_viiiii = Module["dynCall_viiiii"] = asm["dynCall_viiiii"];
var dynCall_iiiiiid = Module["dynCall_iiiiiid"] = asm["dynCall_iiiiiid"];
var dynCall_vi = Module["dynCall_vi"] = asm["dynCall_vi"];
var dynCall_vii = Module["dynCall_vii"] = asm["dynCall_vii"];
var dynCall_iiiiiiiiiii = Module["dynCall_iiiiiiiiiii"] = asm["dynCall_iiiiiiiiiii"];
var dynCall_ii = Module["dynCall_ii"] = asm["dynCall_ii"];
var dynCall_iiiiiiiiiiii = Module["dynCall_iiiiiiiiiiii"] = asm["dynCall_iiiiiiiiiiii"];
var dynCall_iiii = Module["dynCall_iiii"] = asm["dynCall_iiii"];
var dynCall_viiiiiiiiiiiiiii = Module["dynCall_viiiiiiiiiiiiiii"] = asm["dynCall_viiiiiiiiiiiiiii"];
var dynCall_viiiiiiii = Module["dynCall_viiiiiiii"] = asm["dynCall_viiiiiiii"];
var dynCall_viiiiii = Module["dynCall_viiiiii"] = asm["dynCall_viiiiii"];
var dynCall_iiiiiii = Module["dynCall_iiiiiii"] = asm["dynCall_iiiiiii"];
var dynCall_iiiiiiiiiiiii = Module["dynCall_iiiiiiiiiiiii"] = asm["dynCall_iiiiiiiiiiiii"];
var dynCall_iid = Module["dynCall_iid"] = asm["dynCall_iid"];
var dynCall_viiiiiii = Module["dynCall_viiiiiii"] = asm["dynCall_viiiiiii"];
var dynCall_viiiiiiiiii = Module["dynCall_viiiiiiiiii"] = asm["dynCall_viiiiiiiiii"];
var dynCall_iii = Module["dynCall_iii"] = asm["dynCall_iii"];
var dynCall_iiiiii = Module["dynCall_iiiiii"] = asm["dynCall_iiiiii"];
var dynCall_diii = Module["dynCall_diii"] = asm["dynCall_diii"];
var dynCall_i = Module["dynCall_i"] = asm["dynCall_i"];
var dynCall_iiiii = Module["dynCall_iiiii"] = asm["dynCall_iiiii"];
var dynCall_viii = Module["dynCall_viii"] = asm["dynCall_viii"];
var dynCall_v = Module["dynCall_v"] = asm["dynCall_v"];
var dynCall_iiiiiiiii = Module["dynCall_iiiiiiiii"] = asm["dynCall_iiiiiiiii"];
var dynCall_iiiiid = Module["dynCall_iiiiid"] = asm["dynCall_iiiiid"];
var dynCall_viiii = Module["dynCall_viiii"] = asm["dynCall_viiii"];
;
Runtime.stackAlloc = Module['stackAlloc'];
Runtime.stackSave = Module['stackSave'];
Runtime.stackRestore = Module['stackRestore'];
Runtime.establishStackSpace = Module['establishStackSpace'];
Runtime.setTempRet0 = Module['setTempRet0'];
Runtime.getTempRet0 = Module['getTempRet0'];


// === Auto-generated postamble setup entry stuff ===

Module['asm'] = asm;






/**
 * @constructor
 * @extends {Error}
 */
function ExitStatus(status) {
  this.name = "ExitStatus";
  this.message = "Program terminated with exit(" + status + ")";
  this.status = status;
};
ExitStatus.prototype = new Error();
ExitStatus.prototype.constructor = ExitStatus;

var initialStackTop;
var preloadStartTime = null;
var calledMain = false;

dependenciesFulfilled = function runCaller() {
  // If run has never been called, and we should call run (INVOKE_RUN is true, and Module.noInitialRun is not false)
  if (!Module['calledRun']) run();
  if (!Module['calledRun']) dependenciesFulfilled = runCaller; // try this again later, after new deps are fulfilled
}

Module['callMain'] = Module.callMain = function callMain(args) {
  assert(runDependencies == 0, 'cannot call main when async dependencies remain! (listen on __ATMAIN__)');
  assert(__ATPRERUN__.length == 0, 'cannot call main when preRun functions remain to be called');

  args = args || [];

  ensureInitRuntime();

  var argc = args.length+1;
  function pad() {
    for (var i = 0; i < 4-1; i++) {
      argv.push(0);
    }
  }
  var argv = [allocate(intArrayFromString(Module['thisProgram']), 'i8', ALLOC_NORMAL) ];
  pad();
  for (var i = 0; i < argc-1; i = i + 1) {
    argv.push(allocate(intArrayFromString(args[i]), 'i8', ALLOC_NORMAL));
    pad();
  }
  argv.push(0);
  argv = allocate(argv, 'i32', ALLOC_NORMAL);


  try {

    var ret = Module['_main'](argc, argv, 0);


    // if we're not running an evented main loop, it's time to exit
    exit(ret, /* implicit = */ true);
  }
  catch(e) {
    if (e instanceof ExitStatus) {
      // exit() throws this once it's done to make sure execution
      // has been stopped completely
      return;
    } else if (e == 'SimulateInfiniteLoop') {
      // running an evented main loop, don't immediately exit
      Module['noExitRuntime'] = true;
      return;
    } else {
      var toLog = e;
      if (e && typeof e === 'object' && e.stack) {
        toLog = [e, e.stack];
      }
      Module.printErr('exception thrown: ' + toLog);
      Module['quit'](1, e);
    }
  } finally {
    calledMain = true;
  }
}




/** @type {function(Array=)} */
function run(args) {
  args = args || Module['arguments'];

  if (preloadStartTime === null) preloadStartTime = Date.now();

  if (runDependencies > 0) {
    return;
  }

  writeStackCookie();

  preRun();

  if (runDependencies > 0) return; // a preRun added a dependency, run will be called later
  if (Module['calledRun']) return; // run may have just been called through dependencies being fulfilled just in this very frame

  function doRun() {
    if (Module['calledRun']) return; // run may have just been called while the async setStatus time below was happening
    Module['calledRun'] = true;

    if (ABORT) return;

    ensureInitRuntime();

    preMain();

    if (ENVIRONMENT_IS_WEB && preloadStartTime !== null) {
      Module.printErr('pre-main prep time: ' + (Date.now() - preloadStartTime) + ' ms');
    }

    if (Module['onRuntimeInitialized']) Module['onRuntimeInitialized']();

    if (Module['_main'] && shouldRunNow) Module['callMain'](args);

    postRun();
  }

  if (Module['setStatus']) {
    Module['setStatus']('Running...');
    setTimeout(function() {
      setTimeout(function() {
        Module['setStatus']('');
      }, 1);
      doRun();
    }, 1);
  } else {
    doRun();
  }
  checkStackCookie();
}
Module['run'] = Module.run = run;

function exit(status, implicit) {
  if (implicit && Module['noExitRuntime']) {
    Module.printErr('exit(' + status + ') implicitly called by end of main(), but noExitRuntime, so not exiting the runtime (you can use emscripten_force_exit, if you want to force a true shutdown)');
    return;
  }

  if (Module['noExitRuntime']) {
    Module.printErr('exit(' + status + ') called, but noExitRuntime, so halting execution but not exiting the runtime or preventing further async execution (you can use emscripten_force_exit, if you want to force a true shutdown)');
  } else {

    ABORT = true;
    EXITSTATUS = status;
    STACKTOP = initialStackTop;

    exitRuntime();

    if (Module['onExit']) Module['onExit'](status);
  }

  if (ENVIRONMENT_IS_NODE) {
    process['exit'](status);
  }
  Module['quit'](status, new ExitStatus(status));
}
Module['exit'] = Module.exit = exit;

var abortDecorators = [];

function abort(what) {
  if (Module['onAbort']) {
    Module['onAbort'](what);
  }

  if (what !== undefined) {
    Module.print(what);
    Module.printErr(what);
    what = JSON.stringify(what)
  } else {
    what = '';
  }

  ABORT = true;
  EXITSTATUS = 1;

  var extra = '';

  var output = 'abort(' + what + ') at ' + stackTrace() + extra;
  if (abortDecorators) {
    abortDecorators.forEach(function(decorator) {
      output = decorator(output, what);
    });
  }
  throw output;
}
Module['abort'] = Module.abort = abort;

// {{PRE_RUN_ADDITIONS}}

if (Module['preInit']) {
  if (typeof Module['preInit'] == 'function') Module['preInit'] = [Module['preInit']];
  while (Module['preInit'].length > 0) {
    Module['preInit'].pop()();
  }
}

// shouldRunNow refers to calling main(), not run().
var shouldRunNow = true;
if (Module['noInitialRun']) {
  shouldRunNow = false;
}


run();

// {{POST_RUN_ADDITIONS}}





// {{MODULE_ADDITIONS}}



var Photon;
(function (Photon) {
    var EmPhotonPeerLib;
	var callbacks = {};
    (function (EmPhotonPeerLib) {
        EmPhotonPeerLib.Create = function (wss, xf, yf, zf, tf) {			
			var x = Runtime.addFunction(xf);
			var y = Runtime.addFunction(yf);
			var z = Runtime.addFunction(zf);
			var t = Runtime.addFunction(tf);
            var peer = Module.ccall('PhotonPeerCreate', 'number', ['number', 'number', 'number', 'number', 'number'], [wss, x, y, z, t]);
			callbacks[peer] = [x, y, z, t];
			return peer;
        };
        EmPhotonPeerLib.Destroy = function(peer) {
			var c = callbacks[peer];
			for (var i in c) {
				Runtime.removeFunction(c[i]);
			}
			Module.ccall('PhotonPeerDestroy', ['number'], [peer]);
		}

        EmPhotonPeerLib.Connect = Module.cwrap('PhotonPeerConnect', 'number', ['number', 'string', 'string']);
        EmPhotonPeerLib.Disconnect = Module.cwrap('PhotonPeerDisconnect', ['number']);
        EmPhotonPeerLib.State = Module.cwrap('PhotonPeerState', 'number', ['number']);
        EmPhotonPeerLib.Service = Module.cwrap('PhotonPeerService', ['number']);
		EmPhotonPeerLib.SetDebugOutputLevel = Module.cwrap('PhotonPeerSetDebugOutputLevel', 'number', ['number', 'number']);
		
        EmPhotonPeerLib.OpCustom = Module.cwrap('PhotonPeerOpCustom', 'number', ['number', 'number', 'number', 'number', 'number']);	
        EmPhotonPeerLib.OperationRequestCreate = Module.cwrap('PhotonOperationRequestCreate', 'number', ['number']);
        EmPhotonPeerLib.OperationRequestDestroy = Module.cwrap('PhotonOperationRequestDestroy', ['number']);
        EmPhotonPeerLib.OperationRequestSetParam = Module.cwrap('PhotonOperationRequestSetParam', ['number', 'number', 'number']);
    })(EmPhotonPeerLib = Photon.EmPhotonPeerLib || (Photon.EmPhotonPeerLib = {}));

    var EmTypesLib;
    (function (EmTypesLib) {
        var TypeCode;
        (function (TypeCode) {
            TypeCode["BYTE"] = 'b'.charCodeAt(0);
            TypeCode["SHORT"] = 'k'.charCodeAt(0);
            TypeCode["INTEGER"] = 'i'.charCodeAt(0);
            TypeCode["LONG"] = 'l'.charCodeAt(0);
            TypeCode["FLOAT"] = 'f'.charCodeAt(0);
            TypeCode["DOUBLE"] = 'd'.charCodeAt(0);
            TypeCode["BOOLEAN"] = 'o'.charCodeAt(0);
            TypeCode["STRING"] = 's'.charCodeAt(0);
            TypeCode["HASHTABLE"] = 'h'.charCodeAt(0);
            TypeCode["DICTIONARY"] = 'D'.charCodeAt(0);
            TypeCode["OBJECT"] = 'z'.charCodeAt(0);
            TypeCode["ARRAY"] = 'y'.charCodeAt(0);
            TypeCode["BYTEARRAY"] = 'x'.charCodeAt(0);
            TypeCode["PHOTON_COMMAND"] = 'p'.charCodeAt(0);
            TypeCode["EG_NULL"] = '*'.charCodeAt(0);
            TypeCode["CUSTOM"] = 'c'.charCodeAt(0);
            TypeCode["UNKNOWN"] = 0;
        })(TypeCode = Photon.EmTypesLib.TypeCode || (Photon.EmTypesLib.TypeCode = {}));

        EmTypesLib.Pointer_stringify = Module.Pointer_stringify;
        EmTypesLib.UTF8ToString = Module.UTF8ToString;
        EmTypesLib.stringToUTF8 = Module.stringToUTF8;

        EmTypesLib.Byte = Module.cwrap('PhotonTypesByte', 'number', ['number']);
        EmTypesLib.Short = Module.cwrap('PhotonTypesShort', 'number', ['number']);
        EmTypesLib.Int = Module.cwrap('PhotonTypesInt', 'number', ['number']);
        EmTypesLib.Long = Module.cwrap('PhotonTypesLong', 'number', ['number', 'number']);
        EmTypesLib.Float = Module.cwrap('PhotonTypesFloat', 'number', ['number']);
        EmTypesLib.Double = Module.cwrap('PhotonTypesDouble', 'number', ['number']);
        EmTypesLib.String = Module.cwrap('PhotonTypesString', 'number', ['string']);
        EmTypesLib.Bool = Module.cwrap('PhotonTypesBool', 'number', ['number']);
        EmTypesLib.Table = Module.cwrap('PhotonTypesTable', 'number');
		// DictionaryBase from key and value types
		EmTypesLib.DictBase = Module.cwrap('PhotonTypesDictBase', 'number', ['number', 'number']);
		// DictionaryBase from Object
		EmTypesLib.DictBaseFromDict = Module.cwrap('PhotonTypesDictBaseFromDict', 'number', ['number']);
		// DictionaryBase to Object
		EmTypesLib.Dict = Module.cwrap('PhotonTypesDict', 'number', ['number']);
		
		function arrayToHeap(typedArray) {
            var numBytes = typedArray.length * typedArray.BYTES_PER_ELEMENT;
            var ptr = Module._malloc(numBytes);
            var heapBytes = new Uint8Array(Module.HEAPU8.buffer, ptr, numBytes);
            heapBytes.set(new Uint8Array(typedArray.buffer));
            return ptr;
        }		
		
        EmTypesLib.Array = function (typeCode, arr) {
            var typeSize = 8;
            var offset = 0;
			var allocs = [];
            switch (typeCode) {
                case TypeCode.BYTE:
                    var offset = arrayToHeap(new Int8Array(arr));
                    break;
                case TypeCode.SHORT:
                    var offset = arrayToHeap(new Int16Array(arr));
                    break;
                case TypeCode.INTEGER:
                    var offset = arrayToHeap(new Int32Array(arr));
                    break;
				case TypeCode.LONG: // 2 ints per long, low bytes first
                    var offset = arrayToHeap(new Int32Array(arr));
                    break;
                case TypeCode.FLOAT:
                    var offset = arrayToHeap(new Float32Array(arr));
                    break;
                case TypeCode.DOUBLE:
                    var offset = arrayToHeap(new Float64Array(arr));
                    break;
                case TypeCode.STRING:
                    // array of JString*
                    var offset = Module._malloc(arr.length * 4);
					allocs.push(offset)
                    for (var i in arr) {                  
						var len = Module.lengthBytesUTF8(arr[i])
						var strPtr = Module._malloc(len + 1);
						allocs.push(strPtr)
						Module.stringToUTF8(arr[i], strPtr, len + 1)
                        Module.setValue(offset + i * 4, strPtr, '*');
                    }
                    break;
				case TypeCode.BOOLEAN:
					offset = Module._malloc(arr.length);
					allocs.push(offset);
					for (var i = 0; i < arr.length; i++) {
						Module.setValue(offset + i, arr[i], 'i8');
					}
					break;
				case TypeCode.OBJECT:
					offset = Module._malloc(arr.length * 4);
					allocs.push(offset);
					for (var i = 0; i < arr.length; i++) {
						Module.setValue(offset + i * 4, arr[i], '*');
					}
					break;
            }
            if (offset) {
                var count = typeCode == TypeCode.LONG ? arr.length / 2 : arr.length;
                var res = Module.ccall('PhotonTypesArray', 'number', ['number', 'number', 'number'], [typeCode, offset, count]);
				for (var i in allocs) {
					Module._free(allocs[i])
				}
				return res
            } else {
                return 0;
            }
        };            

        EmTypesLib.Destroy = Module.cwrap('PhotonTypesDestroy', ['number']);

        EmTypesLib.GetType = Module.cwrap('PhotonTypesGetType', 'number', ['number']);
        EmTypesLib.GetByte = Module.cwrap('PhotonTypesGetByte', 'number', ['number']);
        EmTypesLib.GetShort = Module.cwrap('PhotonTypesGetShort', 'number', ['number']);
        EmTypesLib.GetInt = Module.cwrap('PhotonTypesGetInt', 'number', ['number']);
        EmTypesLib.GetLongLow = Module.cwrap('PhotonTypesGetLongLow', 'number', ['number']);
        EmTypesLib.GetLongHigh = Module.cwrap('PhotonTypesGetLongHigh', 'number', ['number']);
        EmTypesLib.GetFloat = Module.cwrap('PhotonTypesGetFloat', 'number', ['number']);
        EmTypesLib.GetDouble = Module.cwrap('PhotonTypesGetDouble', 'number', ['number']);
        EmTypesLib.GetString = Module.cwrap('PhotonTypesGetString', 'string', ['number']);
        EmTypesLib.GetBool = function (x) {
            return Module.ccall('PhotonTypesGetBool', 'number', ['number'], [x]) != 0;
        }
        EmTypesLib.GetTable = Module.cwrap('PhotonTypesGetTable', 'number', ['number']);

        EmTypesLib.TableSize = Module.cwrap('PhotonTypesTableSize', 'number', ['number']);
        EmTypesLib.TablePut = Module.cwrap('PhotonTypesTablePut', ['number', 'number', 'number']);
        EmTypesLib.TableKeyAt = Module.cwrap('PhotonTypesTableKeyAt', 'number', ['number', 'number']);
        EmTypesLib.TableValueAt = Module.cwrap('PhotonTypesTableValueAt', 'number', ['number', 'number']);
		EmTypesLib.DictSize = Module.cwrap('PhotonTypesDictSize', 'number', ['number']);
        EmTypesLib.DictBasePut = Module.cwrap('PhotonTypesDictBasePut', ['number', 'number', 'number']);
        EmTypesLib.DictKeyAt = Module.cwrap('PhotonTypesDictKeyAt', 'number', ['number', 'number']);
        EmTypesLib.DictValueAt = Module.cwrap('PhotonTypesDictValueAt', 'number', ['number', 'number']);
        EmTypesLib.ByteDictSize = Module.cwrap('PhotonTypesByteDictSize', 'number', ['number']);
        EmTypesLib.ByteDictPut = Module.cwrap('PhotonTypesByteDictPut', ['number', 'number', 'number']);
        EmTypesLib.ByteDictKeyAt = Module.cwrap('PhotonTypesByteDictKeyAt', 'number', ['number', 'number']);
        EmTypesLib.ByteDictValueAt = Module.cwrap('PhotonTypesByteDictValueAt', 'number', ['number', 'number']);

        EmTypesLib.IsArray = function (x) {
            return Module.ccall('PhotonTypesIsArray', 'number', ['number'], [x]) != 0;
        }
        EmTypesLib.ArraySize = Module.cwrap('PhotonTypesArraySize', 'number', ['number']);
        
        EmTypesLib.ByteAt = Module.cwrap('PhotonTypesByteAt', 'number', ['number', 'number']);
        EmTypesLib.ShortAt = Module.cwrap('PhotonTypesShortAt', 'number', ['number', 'number']);
        EmTypesLib.IntAt = Module.cwrap('PhotonTypesIntAt', 'number', ['number', 'number']);
        EmTypesLib.LongLowAt = Module.cwrap('PhotonTypesLongLowAt', 'number', ['number', 'number']);
        EmTypesLib.LongHighAt = Module.cwrap('PhotonTypesLongHighAt', 'number', ['number', 'number']);
        EmTypesLib.FloatAt = Module.cwrap('PhotonTypesFloatAt', 'number', ['number', 'number']);
        EmTypesLib.DoubleAt = Module.cwrap('PhotonTypesDoubleAt', 'number', ['number', 'number']);
        EmTypesLib.StringAt = Module.cwrap('PhotonTypesStringAt', 'string', ['number', 'number']);
        EmTypesLib.BoolAt = function(x, i) {
			return Module.ccall('PhotonTypesBoolAt', 'number', ['number', 'number'], [x, i]) != 0;
		}
        EmTypesLib.TableAt = Module.cwrap('PhotonTypesTableAt', 'number', ['number', 'number']);
        EmTypesLib.ObjectAt = Module.cwrap('PhotonTypesObjectAt', 'number', ['number', 'number']);
    })(EmTypesLib = Photon.EmTypesLib || (Photon.EmTypesLib = {}));

})(Photon || (Photon = {}));
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
/// --------------------------------------------------------------------------------------------------------------------------------------------------------------
/// ------------------- Exitgames.Common
/// --------------------------------------------------------------------------------------------------------------------------------------------------------------
/**
    Exitgames
    @namespace Exitgames
*/
/**
    Exitgames utilities
    @namespace Exitgames.Common
*/
var Exitgames;
(function (Exitgames) {
    var Common;
    (function (Common) {
        var Logger = /** @class */ (function () {
            /**
                @classdesc Logger with ability to control logging level.
                Prints messages to browser console.
                Each logging method perfoms toString() calls and default formatting of arguments only after it checks logging level. Therefore disabled level logging method call with plain arguments doesn't involves much overhead.
                But if one prefer custom formatting or some calculation for logging methods arguments he should check logging level before doing this to avoid unnecessary operations:
                if(logger.isLevelEnabled(Logger.Level.DEBUG)) {
                    logger.debug("", someCall(x, y), x + "," + y);
                }
                @constructor Exitgames.Common.Logger
                @param {string} [prefix=""] All log messages will be prefixed with that.
                @param {Exitgames.Common.Logger.Level} [level=Level.INFO] Initial logging level.
            */
            function Logger(prefix, level) {
                if (prefix === void 0) { prefix = ""; }
                if (level === void 0) { level = Logger.Level.INFO; }
                this.prefix = prefix;
                this.level = level;
            }
            /**
                @summary Sets logger prefix.
                @method Exitgames.Common.Logger#setPrefix
                @param {stirng} prefix New prefix.
            */
            Logger.prototype.setPrefix = function (prefix) {
                this.prefix = prefix;
            };
            /**
                @summary Gets logger prefix.
                @method Exitgames.Common.Logger#getPrefix
                @returns {string} Prefix.
            */
            Logger.prototype.getPrefix = function () {
                return this.prefix;
            };
            /**
                @summary Changes current logging level.
                @method Exitgames.Common.Logger#setLevel
                @param {Exitgames.Common.Logger.Level} level New logging level.
            */
            Logger.prototype.setLevel = function (level) {
                level = Math.max(level, Logger.Level.DEBUG);
                level = Math.min(level, Logger.Level.OFF);
                this.level = level;
            };
            /**
                @summary Sets global method to be called on logger.exception call.
                @method Exitgames.Common.Logger#setExceptionHandler
                @param {(string) => boolean} handler Exception handler. Return true to cancel throwing.
            */
            Logger.setExceptionHandler = function (handler) {
                this.exceptionHandler = handler;
            };
            /**
                @summary Checks if logging level active.
                @method Exitgames.Common.Logger#isLevelEnabled
                @param {Exitgames.Common.Logger.Level} level Level to check.
                @returns {boolean} True if level active.
            */
            Logger.prototype.isLevelEnabled = function (level) { return level >= this.level; };
            /**
                @summary Returns current logging level.
                @method Exitgames.Common.Logger#getLevel
                @returns {Exitgames.Common.Logger.Level} Current logging level.
            */
            Logger.prototype.getLevel = function () { return this.level; };
            /**
                @summary Logs message if logging level = DEBUG, INFO, WARN, ERROR
                @method Exitgames.Common.Logger#debug
                @param {string} mess Message to log.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of log message after space character.
            */
            Logger.prototype.debug = function (mess) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                this.log(Logger.Level.DEBUG, mess, optionalParams);
            };
            /**
                @summary Logs message if logging level = INFO, WARN, ERROR
                @method Exitgames.Common.Logger#info
                @param {string} mess Message to log.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of log message after space character.
            */
            Logger.prototype.info = function (mess) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                this.log(Logger.Level.INFO, mess, optionalParams);
            };
            /**
                @summary Logs message if logging level = WARN, ERROR
                @method Exitgames.Common.Logger#warn
                @param {string} mess Message to log.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of log message after space character.
            */
            Logger.prototype.warn = function (mess) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                this.log(Logger.Level.WARN, mess, optionalParams);
            };
            /**
                @summary Logs message if logging level = ERROR
                @method Exitgames.Common.Logger#error
                @param {string} mess Message to log.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of log message after space character.
            */
            Logger.prototype.error = function (mess) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                this.log(Logger.Level.ERROR, mess, optionalParams);
            };
            /**
                @summary Throws an Error or executes exception handler if set.
                @method Exitgames.Common.Logger#exception
                @param {string} mess Message passed to Error or exception handler.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of log message after space character.
            */
            Logger.prototype.exception = function (code, mess) {
                var optionalParams = [];
                for (var _i = 2; _i < arguments.length; _i++) {
                    optionalParams[_i - 2] = arguments[_i];
                }
                if (Logger.exceptionHandler && Logger.exceptionHandler(code, this.format0(mess, optionalParams))) {
                    return;
                }
                throw new Error(this.format0("[" + code + "] " + mess, optionalParams));
            };
            /**
                @summary Applies default logger formatting to arguments
                @method Exitgames.Common.Logger#format
                @param {string} mess String to start formatting with.
                @param {...any} optionalParams For every additional parameter toString() applies and result added to the end of formatted string after space character.
                @returns {string} Formatted string.
            */
            Logger.prototype.format = function (mess) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                return this.format0(mess, optionalParams);
            };
            /**
                @summary Applies default logger formatting to array of objects.
                @method Exitgames.Common.Logger#format
                @param {string} mess String to start formatting with.
                @param {any[]} optionalParams For every additional parameter toString() applies and result added to the end of formatted string after space character.
                @returns {string} Formatted string.
            */
            Logger.prototype.formatArr = function (mess, optionalParams) { return this.format0(mess, optionalParams); };
            Logger.prototype.log = function (level, msg, optionalParams) {
                if (level >= this.level) {
                    // for global vars console !== undefined throws an error
                    if (typeof console !== "undefined" && msg !== undefined) {
                        try {
                            var logMethod = console[Logger.log_types[level]];
                            if (!logMethod) {
                                logMethod = console["log"];
                            }
                            if (logMethod) {
                                logMethod.apply(console, [this.prefix, msg].concat(optionalParams));
                            }
                        }
                        catch (error) {
                            // silently fail
                        }
                    }
                }
            };
            Logger.prototype.format0 = function (msg, optionalParams) {
                return (this.prefix == "" ? "" : this.prefix + " ") + msg + " " + optionalParams.map(function (x) {
                    if (x !== undefined) {
                        switch (typeof x) {
                            case "object":
                                try {
                                    return JSON.stringify(x);
                                }
                                catch (error) {
                                    return x.toString() + "(" + error + ")";
                                }
                            default:
                                return x.toString();
                        }
                    }
                }).join(" ");
            };
            /**
                @summary Logging levels. Set to restrict log output.
                @member Exitgames.Common.Logger.Level
                @readonly
                @property {number} DEBUG All logging methods enabled.
                @property {number} INFO info(...), warn(...), error(...) methods enabled.
                @property {number} WARN warn(...) and error(...) methods enabled.
                @property {number} ERROR Only error(...) method enabled.
                @property {number} OFF Logging off.
            */
            Logger.Level = {
                //TRACE : 0,
                DEBUG: 1,
                INFO: 2,
                WARN: 3,
                ERROR: 4,
                //FATAL: 5,
                OFF: 6
            };
            Logger.log_types = ["debug", "debug", "info", "warn", "error"];
            return Logger;
        }());
        Common.Logger = Logger;
        var Util = /** @class */ (function () {
            function Util() {
            }
            Util.indexOf = function (arr, item, from) {
                for (var l = arr.length, i = from < 0 ? Math.max(0, l + from) : from || 0; i < l; i++) {
                    if (arr[i] === item) {
                        return i;
                    }
                }
                return -1;
            };
            Util.isArray = function (obj) {
                return Object.prototype.toString.call(obj) === "[object Array]";
            };
            //TODO: naming. could be named mergeHashtable or something more specific
            Util.merge = function (target, additional) {
                for (var i in additional) {
                    if (additional.hasOwnProperty(i)) {
                        target[i] = additional[i];
                    }
                }
            };
            Util.getPropertyOrElse = function (obj, prop, defaultValue) {
                if (obj.hasOwnProperty(prop)) {
                    return obj[prop];
                }
                else {
                    return defaultValue;
                }
            };
            Util.enumValueToName = function (enumObj, value) {
                for (var i in enumObj) {
                    if (value == enumObj[i]) {
                        return i;
                    }
                }
                return "undefined";
            };
            Util.getEnumKeyByValue = function (enumObj, value) {
                for (var i in enumObj) {
                    if (value == enumObj[i]) {
                        return i;
                    }
                }
                return undefined;
            };
            return Util;
        }());
        Common.Util = Util;
    })(Common = Exitgames.Common || (Exitgames.Common = {}));
})(Exitgames || (Exitgames = {}));
/// <reference path="photon-common.ts"/>
/// <reference path="emscripten/EmPhotonPeerLibAPI.d.ts"/>
/**
    Photon
    @namespace Photon
*/
var Photon;
(function (Photon) {
    var TypeCode;
    (function (TypeCode) {
        TypeCode[TypeCode["BYTE"] = 'b'.charCodeAt(0)] = "BYTE";
        TypeCode[TypeCode["SHORT"] = 'k'.charCodeAt(0)] = "SHORT";
        TypeCode[TypeCode["INTEGER"] = 'i'.charCodeAt(0)] = "INTEGER";
        TypeCode[TypeCode["LONG"] = 'l'.charCodeAt(0)] = "LONG";
        TypeCode[TypeCode["FLOAT"] = 'f'.charCodeAt(0)] = "FLOAT";
        TypeCode[TypeCode["DOUBLE"] = 'd'.charCodeAt(0)] = "DOUBLE";
        TypeCode[TypeCode["BOOLEAN"] = 'o'.charCodeAt(0)] = "BOOLEAN";
        TypeCode[TypeCode["STRING"] = 's'.charCodeAt(0)] = "STRING";
        TypeCode[TypeCode["HASHTABLE"] = 'h'.charCodeAt(0)] = "HASHTABLE";
        TypeCode[TypeCode["DICTIONARY"] = 'D'.charCodeAt(0)] = "DICTIONARY";
        TypeCode[TypeCode["OBJECT"] = 'z'.charCodeAt(0)] = "OBJECT";
        TypeCode[TypeCode["ARRAY"] = 'y'.charCodeAt(0)] = "ARRAY";
        TypeCode[TypeCode["BYTEARRAY"] = 'x'.charCodeAt(0)] = "BYTEARRAY";
        TypeCode[TypeCode["PHOTON_COMMAND"] = 'p'.charCodeAt(0)] = "PHOTON_COMMAND";
        TypeCode[TypeCode["EG_NULL"] = '*'.charCodeAt(0)] = "EG_NULL";
        TypeCode[TypeCode["CUSTOM"] = 'c'.charCodeAt(0)] = "CUSTOM";
        TypeCode[TypeCode["UNKNOWN"] = 0] = "UNKNOWN";
    })(TypeCode || (TypeCode = {}));
    var TypeExtType;
    (function (TypeExtType) {
        TypeExtType[TypeExtType["None"] = 0] = "None";
        TypeExtType[TypeExtType["Byte"] = 1] = "Byte";
        TypeExtType[TypeExtType["Short"] = 2] = "Short";
        TypeExtType[TypeExtType["Int"] = 3] = "Int";
        TypeExtType[TypeExtType["Long"] = 4] = "Long";
        TypeExtType[TypeExtType["Float"] = 5] = "Float";
        TypeExtType[TypeExtType["Double"] = 6] = "Double";
        TypeExtType[TypeExtType["String"] = 7] = "String";
        TypeExtType[TypeExtType["Bool"] = 8] = "Bool";
        TypeExtType[TypeExtType["Dict"] = 9] = "Dict";
        TypeExtType[TypeExtType["Object"] = 10] = "Object";
    })(TypeExtType = Photon.TypeExtType || (Photon.TypeExtType = {}));
    // Not exported class does not pollute namespace
    var T = /** @class */ (function () {
        function T(type, value, longHiBytes, keyType, valType) {
            if (longHiBytes === void 0) { longHiBytes = 0; }
            this.type = type;
            this.value = value;
            this.longHiBytes = longHiBytes;
            this.keyType = keyType;
            this.valType = valType;
        }
        T.prototype.ToJS = function () {
            if (this.type == TypeExtType.Long) {
                return this;
            }
            else {
                return this.value;
            }
        };
        return T;
    }());
    // wraps native dict to let user code access it by providing key and value types in access methods
    var NativeDict = /** @class */ (function () {
        function NativeDict(ptr, arrIdx) {
            if (arrIdx === void 0) { arrIdx = -1; }
            this.ptr = ptr;
            this.arrIdx = arrIdx;
        }
        return NativeDict;
    }());
    var NativeDictBase = /** @class */ (function () {
        function NativeDictBase(base) {
            this.base = base;
        }
        NativeDictBase.prototype.Size = function () {
            return Photon.EmTypesLib.DictSize(this.base);
        };
        NativeDictBase.prototype.KeyAt = function (i) {
            return TypeExt.native2js(Photon.EmTypesLib.DictKeyAt(this.base, i));
        };
        NativeDictBase.prototype.ValueAt = function (i) {
            return TypeExt.native2js(Photon.EmTypesLib.DictValueAt(this.base, i));
        };
        return NativeDictBase;
    }());
    Photon.NativeDictBase = NativeDictBase;
    var TypeExt = /** @class */ (function () {
        function TypeExt() {
        }
        TypeExt.Is = function (x) {
            return x instanceof T;
        };
        // for numeric typed objects creators below,
        // pass number to create single value,
        // pass number[] to create array of values
        TypeExt.Byte = function (x) {
            return new T(TypeExtType.Byte, x);
        };
        TypeExt.Short = function (x) {
            return new T(TypeExtType.Short, x);
        };
        TypeExt.Int = function (x) {
            return new T(TypeExtType.Int, x);
        };
        TypeExt.Float = function (x) {
            return new T(TypeExtType.Float, x);
        };
        TypeExt.Double = function (x) {
            return new T(TypeExtType.Double, x);
        };
        // long represented by 2 ints
        // array of longs represented by array of ints of double size: [low0, hi0, low1, hi1...]
        TypeExt.Long = function (x, hi) {
            if (hi === void 0) { hi = 0; }
            return new T(TypeExtType.Long, x, hi);
        };
        // pass js array to create typed array of strings
        // for single values, use string
        TypeExt.String = function (x) {
            return new T(TypeExtType.String, x);
        };
        // pass js array to create array of booleans
        // for single values, use js boolean
        TypeExt.Bool = function (x) {
            return new T(TypeExtType.Bool, x);
        };
        TypeExt.Dict = function (keyType, valType, x) {
            return new T(TypeExtType.Dict, x, 0, keyType, valType);
        };
        TypeExt.GetDict = function (keyType, valType, d) {
            var b = Photon.EmTypesLib.DictBaseFromDict(TypeExt.type2typeCode(keyType), TypeExt.type2typeCode(valType), d.ptr);
            if (b == 0) {
                this.logger.exception(301, "GetDict: Unsupported dictionary key/value types", TypeExtType[keyType] + "(" + keyType + ")/" + TypeExtType[valType] + "(" + valType + ")");
            }
            return new NativeDictBase(b);
        };
        TypeExt.type2typeCode = function (t) {
            switch (t) {
                case TypeExtType.Byte: return TypeCode.BYTE;
                case TypeExtType.Short: return TypeCode.SHORT;
                case TypeExtType.Int: return TypeCode.INTEGER;
                case TypeExtType.Long: return TypeCode.LONG;
                case TypeExtType.Float: return TypeCode.FLOAT;
                case TypeExtType.Double: return TypeCode.DOUBLE;
                case TypeExtType.String: return TypeCode.STRING;
                case TypeExtType.Bool: return TypeCode.BOOLEAN;
                case TypeExtType.Dict: return TypeCode.DICTIONARY;
                case TypeExtType.Object: return TypeCode.OBJECT;
                default: return TypeCode.UNKNOWN;
            }
        };
        TypeExt.js2native = function (nativeRefs, x) {
            var ref = TypeExt.js2native0(nativeRefs, x);
            if (ref != 0) {
                nativeRefs.push(ref);
            }
            return ref;
        };
        TypeExt.js2native0 = function (nativeRefs, x) {
            if (x instanceof T) {
                var y = x;
                if (Array.isArray(y.value)) {
                    var arr = y.value;
                    var typeCode = TypeExt.type2typeCode(y.type);
                    return Photon.EmTypesLib.Array(typeCode, arr);
                }
                else {
                    switch (y.type) {
                        case TypeExtType.Byte: return Photon.EmTypesLib.Byte(x.value);
                        case TypeExtType.Short: return Photon.EmTypesLib.Short(x.value);
                        case TypeExtType.Int: return Photon.EmTypesLib.Int(x.value);
                        case TypeExtType.Long: return Photon.EmTypesLib.Long(x.value, x.longHiBytes);
                        case TypeExtType.Float: return Photon.EmTypesLib.Float(x.value);
                        case TypeExtType.Double: return Photon.EmTypesLib.Double(x.value);
                        case TypeExtType.String: return Photon.EmTypesLib.String(x.value); // normally js string type used for single value
                        case TypeExtType.Bool: return Photon.EmTypesLib.Bool(x.value); // normally js boolean type used for single value
                        case TypeExtType.Dict:
                            var b = Photon.EmTypesLib.DictBase(TypeExt.type2typeCode(x.keyType), TypeExt.type2typeCode(x.valType));
                            if (b == 0) {
                                this.logger.exception(302, "js2native0: Unsupported dictionary key/value types", TypeExtType[x.keyType] + "(" + x.keyType + ")/" + TypeExtType[x.valType] + "(" + x.valType + ")");
                            }
                            for (var k in x.value) {
                                Photon.EmTypesLib.DictBasePut(b, TypeExt.js2native(nativeRefs, k), TypeExt.js2native(nativeRefs, x.value[k]));
                            }
                            return Photon.EmTypesLib.Dict(b);
                        default:
                            TypeExt.logger.exception(303, "js2native: unsupported data type " + y.type + " for value ", x);
                            return 0;
                    }
                }
            }
            else if (Array.isArray(x)) {
                var arr = x;
                var ptrs = [];
                for (var i = 0; i < arr.length; i++) {
                    ptrs.push(TypeExt.js2native(nativeRefs, arr[i]));
                }
                return Photon.EmTypesLib.Array(TypeCode.OBJECT, ptrs);
            }
            else if (typeof x == "number") {
                return Photon.EmTypesLib.Double(x);
            }
            else if (typeof x == "string") {
                return Photon.EmTypesLib.String(x);
            }
            else if (typeof x == "boolean") {
                return Photon.EmTypesLib.Bool(x);
            }
            else if (typeof x == "object") {
                var t = Photon.EmTypesLib.Table();
                for (var k in x) {
                    Photon.EmTypesLib.TablePut(t, TypeExt.js2native(nativeRefs, k), TypeExt.js2native(nativeRefs, x[k]));
                }
                return t;
            }
            else {
                TypeExt.logger.exception(304, "js2native: unsupported data type for value:", x);
            }
            return 0;
        };
        TypeExt.native2js = function (x) {
            var t = Photon.EmTypesLib.GetType(x);
            if (Photon.EmTypesLib.IsArray(x)) {
                var a = [];
                var length = Photon.EmTypesLib.ArraySize(x);
                for (var i = 0; i < length; i++) {
                    switch (t) {
                        case TypeCode.BYTE:
                            a[i] = Photon.EmTypesLib.ByteAt(x, i);
                            break;
                        case TypeCode.SHORT:
                            a[i] = Photon.EmTypesLib.ShortAt(x, i);
                            break;
                        case TypeCode.INTEGER:
                            a[i] = Photon.EmTypesLib.IntAt(x, i);
                            break;
                        case TypeCode.LONG:
                            a[i * 2] = Photon.EmTypesLib.LongLowAt(x, i);
                            a[i * 2 + 1] = Photon.EmTypesLib.LongHighAt(x, i);
                            break;
                        case TypeCode.FLOAT:
                            a[i] = Photon.EmTypesLib.FloatAt(x, i);
                            break;
                        case TypeCode.DOUBLE:
                            a[i] = Photon.EmTypesLib.DoubleAt(x, i);
                            break;
                        case TypeCode.STRING:
                            a[i] = Photon.EmTypesLib.StringAt(x, i);
                            break;
                        case TypeCode.BOOLEAN:
                            a[i] = Photon.EmTypesLib.BoolAt(x, i);
                            break;
                        case TypeCode.HASHTABLE:
                            a[i] = TypeExt.native2js(Photon.EmTypesLib.TableAt(x, i));
                            break;
                        case TypeCode.DICTIONARY:
                            a[i] = new NativeDict(x, i);
                            break;
                        case TypeCode.OBJECT:
                            a[i] = TypeExt.native2js(Photon.EmTypesLib.ObjectAt(x, i));
                            break;
                        case TypeCode.ARRAY:
                        case TypeCode.BYTEARRAY:
                        case TypeCode.PHOTON_COMMAND:
                        case TypeCode.EG_NULL:
                        case TypeCode.CUSTOM:
                        case TypeCode.UNKNOWN:
                        default:
                            this.logger.exception(305, "native2js: Unsupported data type ", t, " for value ", x);
                    }
                }
                if (t == TypeCode.LONG) {
                    return TypeExt.Long(a);
                }
                else {
                    return a;
                }
            }
            switch (t) {
                case TypeCode.BYTE:
                    return Photon.EmTypesLib.GetByte(x);
                case TypeCode.SHORT:
                    return Photon.EmTypesLib.GetShort(x);
                case TypeCode.INTEGER:
                    return Photon.EmTypesLib.GetInt(x);
                case TypeCode.LONG:
                    return TypeExt.Long(Photon.EmTypesLib.GetLongLow(x), Photon.EmTypesLib.GetLongHigh(x));
                case TypeCode.FLOAT:
                    return Photon.EmTypesLib.GetFloat(x);
                case TypeCode.DOUBLE:
                    return Photon.EmTypesLib.GetDouble(x);
                case TypeCode.STRING:
                    return Photon.EmTypesLib.GetString(x);
                case TypeCode.BOOLEAN:
                    return Photon.EmTypesLib.GetBool(x);
                case TypeCode.HASHTABLE:
                    var t = Photon.EmTypesLib.GetTable(x);
                    var table = {};
                    for (var i = 0; i < Photon.EmTypesLib.TableSize(t); i++) {
                        table[this.native2js(Photon.EmTypesLib.TableKeyAt(t, i))] = this.native2js(Photon.EmTypesLib.TableValueAt(t, i));
                    }
                    return table;
                case TypeCode.DICTIONARY:
                    return new NativeDict(x);
                case TypeCode.OBJECT:
                case TypeCode.ARRAY:
                case TypeCode.BYTEARRAY:
                case TypeCode.PHOTON_COMMAND:
                case TypeCode.EG_NULL:
                case TypeCode.CUSTOM:
                case TypeCode.UNKNOWN:
                default:
                    this.logger.exception(306, "native2js: Unsupported data type:", t);
            }
        };
        TypeExt.logger = new Exitgames.Common.Logger("TypeExt");
        return TypeExt;
    }());
    Photon.TypeExt = TypeExt;
})(Photon || (Photon = {}));
/// <reference path="photon-common.ts"/>
/// <reference path="photon-type-ext-em.ts"/>
/**
    Photon
    @namespace Photon
*/
var Photon;
(function (Photon) {
    var EmStatusCode;
    (function (EmStatusCode) {
        EmStatusCode[EmStatusCode["EXCEPTION_ON_CONNECT"] = 1023] = "EXCEPTION_ON_CONNECT";
        EmStatusCode[EmStatusCode["CONNECT"] = 1024] = "CONNECT";
        EmStatusCode[EmStatusCode["DISCONNECT"] = 1025] = "DISCONNECT";
        EmStatusCode[EmStatusCode["EXCEPTION"] = 1026] = "EXCEPTION";
        EmStatusCode[EmStatusCode["QUEUE_OUTGOING_RELIABLE_WARNING"] = 1027] = "QUEUE_OUTGOING_RELIABLE_WARNING";
        EmStatusCode[EmStatusCode["QUEUE_OUTGOING_UNRELIABLE_WARNING"] = 1029] = "QUEUE_OUTGOING_UNRELIABLE_WARNING";
        EmStatusCode[EmStatusCode["SEND_ERROR"] = 1030] = "SEND_ERROR";
        EmStatusCode[EmStatusCode["QUEUE_OUTGOING_ACKS_WARNING"] = 1031] = "QUEUE_OUTGOING_ACKS_WARNING";
        EmStatusCode[EmStatusCode["QUEUE_INCOMING_RELIABLE_WARNING"] = 1033] = "QUEUE_INCOMING_RELIABLE_WARNING";
        EmStatusCode[EmStatusCode["QUEUE_INCOMING_UNRELIABLE_WARNING"] = 1035] = "QUEUE_INCOMING_UNRELIABLE_WARNING";
        EmStatusCode[EmStatusCode["QUEUE_SENT_WARNING"] = 1037] = "QUEUE_SENT_WARNING";
        EmStatusCode[EmStatusCode["INTERNAL_RECEIVE_EXCEPTION"] = 1039] = "INTERNAL_RECEIVE_EXCEPTION";
        EmStatusCode[EmStatusCode["TIMEOUT_DISCONNECT"] = 1040] = "TIMEOUT_DISCONNECT";
        EmStatusCode[EmStatusCode["DISCONNECT_BY_SERVER"] = 1041] = "DISCONNECT_BY_SERVER";
        EmStatusCode[EmStatusCode["DISCONNECT_BY_SERVER_USER_LIMIT"] = 1042] = "DISCONNECT_BY_SERVER_USER_LIMIT";
        EmStatusCode[EmStatusCode["DISCONNECT_BY_SERVER_LOGIC"] = 1043] = "DISCONNECT_BY_SERVER_LOGIC";
        EmStatusCode[EmStatusCode["ENCRYPTION_ESTABLISHED"] = 1048] = "ENCRYPTION_ESTABLISHED";
        EmStatusCode[EmStatusCode["ENCRYPTION_FAILED_TO_ESTABLISH"] = 1049] = "ENCRYPTION_FAILED_TO_ESTABLISH";
    })(EmStatusCode || (EmStatusCode = {}));
    var EmPeerState;
    (function (EmPeerState) {
        EmPeerState[EmPeerState["DISCONNECTED"] = 0] = "DISCONNECTED";
        EmPeerState[EmPeerState["CONNECTING"] = 1] = "CONNECTING";
        EmPeerState[EmPeerState["INITIALIZING_APPLICATION"] = 2] = "INITIALIZING_APPLICATION";
        EmPeerState[EmPeerState["CONNECTED"] = 3] = "CONNECTED";
        EmPeerState[EmPeerState["DISCONNECTING"] = 4] = "DISCONNECTING";
    })(EmPeerState || (EmPeerState = {}));
    /**
        @summary These are the options that can be used as underlying transport protocol.
        @member Photon.ConnectionProtocol
        @readonly
        @property {number} Ws WebSockets connection.
        @property {number} Wss WebSockets Secure connection.
    **/
    var ConnectionProtocol;
    (function (ConnectionProtocol) {
        ConnectionProtocol[ConnectionProtocol["Ws"] = 0] = "Ws";
        ConnectionProtocol[ConnectionProtocol["Wss"] = 1] = "Wss";
    })(ConnectionProtocol = Photon.ConnectionProtocol || (Photon.ConnectionProtocol = {}));
    var PhotonPeer = /** @class */ (function () {
        /**
            @classdesc Instances of the PhotonPeer class are used to connect to a Photon server and communicate with it.
            A PhotonPeer instance allows communication with the Photon Server, which in turn distributes messages to other PhotonPeer clients.
            An application can use more than one PhotonPeer instance, which are treated as separate users on the server.
            Each should have its own listener instance, to separate the operations, callbacks and events.
            @constructor Photon.PhotonPeer
            @param {Photon.ConnectionProtocol} protocol Connection protocol
            @param {string} address Server address:port.
            @param {string} [subprotocol=""] WebSocket protocol.
            @param {string} [debugName=""] Log messages prefixed with this value.
        */
        function PhotonPeer(protocol, address, subprotocol, debugName) {
            if (subprotocol === void 0) { subprotocol = ""; }
            if (debugName === void 0) { debugName = ""; }
            this.protocol = protocol;
            this.address = address;
            this.subprotocol = subprotocol;
            this.closed = false;
            this._peerStatusListeners = {};
            this._eventListeners = {};
            this._responseListeners = {};
            this.lastRtt = 0;
            this.initTimestamp = Date.now();
            this.peer = Photon.EmPhotonPeerLib.Create(protocol == Photon.ConnectionProtocol.Wss, this.onStatusChangedFunc(), this.onOpResponseFunc(), this.onEventFunc(), this.debugReturnFunc());
            this._logger = new Exitgames.Common.Logger(debugName && debugName != "" ? debugName + ":" : "");
        }
        PhotonPeer.prototype.Destroy = function () {
            if (this.timer) {
                clearInterval(this.timer);
            }
            Photon.EmPhotonPeerLib.Destroy(this.peer);
        };
        /**
            @summary Checks if peer is connecting.
            @method Photon.PhotonPeer#isConnecting
            @returns {boolean} True if peer is connecting.
        */
        PhotonPeer.prototype.isConnecting = function () { return Photon.EmPhotonPeerLib.State(this.peer) == EmPeerState.CONNECTING; };
        PhotonPeer.prototype.getLastRtt = function () { return this.lastRtt; };
        /**
            @summary Checks if peer is connected.
            @method Photon.PhotonPeer#isConnected
            @returns {boolean} True if peer is connected.
        */
        PhotonPeer.prototype.isConnected = function () { return Photon.EmPhotonPeerLib.State(this.peer) == EmPeerState.CONNECTED; };
        /**
            @summary Checks if peer is closing.
            @method Photon.PhotonPeer#isClosing
            @returns {boolean} True if peer is closing.
        */
        PhotonPeer.prototype.isClosing = function () { return Photon.EmPhotonPeerLib.State(this.peer) == EmPeerState.DISCONNECTING; };
        /**
            @summary Starts connection to server.
            @method Photon.PhotonPeer#connect
        */
        PhotonPeer.prototype.connect = function (appID) {
            Photon.EmPhotonPeerLib.Connect(this.peer, this.address, appID);
            this.startService();
        };
        PhotonPeer.prototype.startService = function () {
            var peer = this.peer;
            var t = this;
            if (this.timer) {
                clearInterval(this.timer);
            }
            this.timer = setInterval(function () {
                if (!t.closed) {
                    Photon.EmPhotonPeerLib.Service(peer);
                    // t.service();
                }
            }, PhotonPeer.serviceIntervalMs);
        };
        /**
            @summary Disconnects from server.
            @method Photon.PhotonPeer#disconnect
        */
        PhotonPeer.prototype.disconnect = function () {
            Photon.EmPhotonPeerLib.Disconnect(this.peer);
        };
        /**
            @summary Sends operation to the Photon Server.
            @method Photon.PhotonPeer#sendOperation
            @param {number} code Code of operation.
            @param {object} [data] Parameters of operation as a flattened array of key-value pairs: [key1, value1, key2, value2...]
            @param {boolean} [sendReliable=false] Selects if the operation must be acknowledged or not. If false, the operation is not guaranteed to reach the server.
            @param {number} [channelId=0] The channel in which this operation should be sent.
        */
        PhotonPeer.prototype.sendOperation = function (code, data, sendReliable, channelId) {
            if (sendReliable === void 0) { sendReliable = false; }
            if (channelId === void 0) { channelId = 0; }
            if (!Exitgames.Common.Util.isArray(data)) {
                if (data === undefined) {
                    data = [];
                }
                else {
                    this._logger.exception(201, "PhotonPeer[sendOperation] - Trying to send non array data:", data);
                }
            }
            var nativeTempRefs = [];
            var opReq = Photon.EmPhotonPeerLib.OperationRequestCreate(code);
            for (var i = 0; i < data.length; i += 2) {
                var v = Photon.TypeExt.js2native(nativeTempRefs, data[i + 1]);
                if (v != undefined) {
                    Photon.EmPhotonPeerLib.OperationRequestSetParam(opReq, data[i], v);
                }
            }
            for (i = 0; i < nativeTempRefs.length; i++) {
                Photon.EmTypesLib.Destroy(nativeTempRefs[i]);
            }
            Photon.EmPhotonPeerLib.OpCustom(this.peer, opReq, sendReliable, channelId, false);
            Photon.EmPhotonPeerLib.OperationRequestDestroy(opReq);
            this._logger.debug("PhotonPeer[sendOperation] - Sending request:", code);
        };
        /**
            @summary Registers listener for peer status change.
            @method Photon.PhotonPeer#addPeerStatusListener
            @param {PhotonPeer.StatusCodes} statusCode Status change to this value will be listening.
            @param {Function} callback The listener function that processes the status change. This function don't accept any parameters.
        */
        PhotonPeer.prototype.addPeerStatusListener = function (statusCode, callback) {
            this._addListener(this._peerStatusListeners, statusCode, callback);
        };
        /**
            @summary Registers listener for custom event.
            @method Photon.PhotonPeer#addEventListener
            @param {number} eventCode Custom event code.
            @param {Function} callback The listener function that processes the event. This function may accept object with event content.
        */
        PhotonPeer.prototype.addEventListener = function (eventCode, callback) {
            this._addListener(this._eventListeners, eventCode.toString(), callback);
        };
        /**
            @summary Registers listener for operation response.
            @method Photon.PhotonPeer#addResponseListener
            @param {number} operationCode Operation code.
            @param {Function} callback The listener function that processes the event. This function may accept object with operation response content.
        */
        PhotonPeer.prototype.addResponseListener = function (operationCode, callback) {
            this._addListener(this._responseListeners, operationCode.toString(), callback);
        };
        /**
            @summary Removes listener if exists for peer status change.
            @method Photon.PhotonPeer#removePeerStatusListener
            @param {string} statusCode One of PhotonPeer.StatusCodes to remove listener for.
            @param {Function} callback Listener to remove.
        */
        PhotonPeer.prototype.removePeerStatusListener = function (statusCode, callback) {
            this._removeListener(this._peerStatusListeners, statusCode, callback);
        };
        /**
            @summary Removes listener if exists for custom event.
            @method Photon.PhotonPeer#removeEventListener
            @param {number} eventCode Event code to remove to remove listener for.
            @param {Function} callback Listener to remove.
        */
        PhotonPeer.prototype.removeEventListener = function (eventCode, callback) {
            this._removeListener(this._eventListeners, eventCode.toString(), callback);
        };
        /**
            @summary Removes listener if exists for operation response.
            @method Photon.PhotonPeer#removeResponseListener
            @param {number} operationCode Operation code to remove listener for.
            @param {Function} callback Listener to remove.
        */
        PhotonPeer.prototype.removeResponseListener = function (operationCode, callback) {
            this._removeListener(this._responseListeners, operationCode.toString(), callback);
        };
        /**
            @summary Removes all listeners for peer status change specified.
            @method Photon.PhotonPeer#removePeerStatusListenersForCode
            @param {string} statusCode One of PhotonPeer.StatusCodes to remove all listeners for.
        */
        PhotonPeer.prototype.removePeerStatusListenersForCode = function (statusCode) {
            this._removeListenersForCode(this._peerStatusListeners, statusCode);
        };
        /**
            @summary Removes all listeners for custom event specified.
            @method Photon.PhotonPeer#removeEventListenersForCode
            @param {number} eventCode Event code to remove all listeners for.
        */
        PhotonPeer.prototype.removeEventListenersForCode = function (eventCode) {
            this._removeListenersForCode(this._eventListeners, eventCode.toString());
        };
        /**
            @summary Removes all listeners for operation response specified.
            @method Photon.PhotonPeer#removeResponseListenersForCode
            @param {number} operationCode Operation code to remove all listeners for.
        */
        PhotonPeer.prototype.removeResponseListenersForCode = function (operationCode) {
            this._removeListenersForCode(this._responseListeners, operationCode.toString());
        };
        /**
            @summary Sets peer logger level.
            @method Photon.PhotonPeer#setLogLevel
            @param {Exitgames.Common.Logger.Level} level Logging level.
        */
        PhotonPeer.prototype.setLogLevel = function (level) {
            this._logger.setLevel(level);
            var emLevel = (_a = {},
                //[Exitgames.Common.Level.Logger.TRACE]
                _a[Exitgames.Common.Logger.Level.DEBUG] = 4,
                _a[Exitgames.Common.Logger.Level.INFO] = 3,
                _a[Exitgames.Common.Logger.Level.WARN] = 2,
                _a[Exitgames.Common.Logger.Level.ERROR] = 1,
                //[Exitgames.Common.Level.Logger.FATAL]
                _a[Exitgames.Common.Logger.Level.OFF] = 0,
                _a)[level];
            if (emLevel != undefined) {
                Photon.EmPhotonPeerLib.SetDebugOutputLevel(this.peer, emLevel);
            }
            var _a;
        };
        /**
            @summary Called if no listener found for received custom event.
            Override to relay unknown event to user's code or handle known events without listener registration.
            @method Photon.PhotonPeer#onUnhandledEvent
            @param {number} eventCode Code of received event.
            @param {object} [args] Content of received event or empty object.
        */
        PhotonPeer.prototype.onUnhandledEvent = function (eventCode, args) {
            this._logger.warn('PhotonPeer: No handler for event', eventCode, 'registered.');
        };
        /**
            @summary Called if no listener found for received operation response event.
            Override to relay unknown response to user's code or handle known responses without listener registration.
            @method Photon.PhotonPeer#onUnhandledEvent
            @param {number} operationCode Code of received response.
            @param {object} [args] Content of received response or empty object.
        */
        PhotonPeer.prototype.onUnhandledResponse = function (operationCode, args) {
            this._logger.warn('PhotonPeer: No handler for response', operationCode, 'registered.');
        };
        // TODO: lite calls this
        // protected
        PhotonPeer.prototype._dispatchEvent = function (code, args) {
            if (!this._dispatch(this._eventListeners, code.toString(), args, "event")) {
                this.onUnhandledEvent(code, args);
            }
        };
        // TODO: lite calls this
        // protected
        PhotonPeer.prototype._dispatchResponse = function (code, args) {
            if (!this._dispatch(this._responseListeners, code.toString(), args, "response")) {
                this.onUnhandledResponse(code, args);
            }
        };
        PhotonPeer.prototype.nativeByteDict2js = function (x) {
            var d = {};
            for (var i = 0; i < Photon.EmTypesLib.ByteDictSize(x); i++) {
                d[Photon.EmTypesLib.ByteDictKeyAt(x, i)] = Photon.TypeExt.native2js(Photon.EmTypesLib.ByteDictValueAt(x, i));
            }
            return d;
        };
        PhotonPeer.prototype.debugReturnFunc = function () {
            var t = this;
            return function (level, messagePtr) {
                var message = Photon.EmTypesLib.UTF8ToString(messagePtr);
                switch (level) {
                    case 1:
                        t._logger.error(message);
                        break;
                    case 2:
                        t._logger.warn(message);
                        break;
                    case 3:
                        t._logger.info(message);
                        break;
                    case 4:
                        t._logger.debug(message);
                        break;
                    case 4:
                        t._logger.info("[unknown log level " + level + "]" + message);
                }
            };
        };
        PhotonPeer.prototype.onEventFunc = function () {
            var t = this;
            return function (code, content) {
                switch (code) {
                    default:
                        t._dispatchEvent(code, { vals: t.nativeByteDict2js(content) });
                        break;
                }
            };
        };
        PhotonPeer.prototype.onOpResponseFunc = function () {
            var t = this;
            return function (errCode, errMsg, code, content) {
                switch (code) {
                    default:
                        t._dispatchResponse(code, { errCode: errCode, errMsg: Photon.EmTypesLib.Pointer_stringify(errMsg), vals: t.nativeByteDict2js(content) });
                        break;
                }
            };
        };
        PhotonPeer.prototype.onStatusChangedFunc = function () {
            var t = this;
            return function (emStatus) {
                var status = PhotonPeer.StatusCodeEmToJs[emStatus];
                t._logger.debug('PhotonPeer onStatusChangedFunc', emStatus, '->', status);
                if (status) {
                    t._dispatchPeerStatus(status);
                }
            };
        };
        PhotonPeer.prototype._addListener = function (listeners, code, callback) {
            if (!(code in listeners)) {
                listeners[code] = [];
            }
            if (callback && typeof callback === "function") {
                this._logger.debug('PhotonPeer[_addListener] - Adding listener for event', code);
                listeners[code].push(callback);
            }
            else {
                this._logger.error('PhotonPeer[_addListener] - Listener', code, 'is not a function but of type', typeof callback, '. No listener added!');
            }
            return this;
        };
        PhotonPeer.prototype._dispatch = function (listeners, code, args, debugType) {
            if (code in listeners) {
                var events = listeners[code];
                for (var i = 0, l = events.length; i < l; i++) {
                    if (!Exitgames.Common.Util.isArray(args)) {
                        args = [args];
                    }
                    events[i].apply(this, args === undefined ? [] : args);
                }
                return true;
            }
            else {
                return false;
            }
        };
        PhotonPeer.prototype._dispatchPeerStatus = function (code) {
            if (!this._dispatch(this._peerStatusListeners, code, undefined, "peerStatus")) {
                this._logger.warn('PhotonPeer[_dispatchPeerStatus] - No handler for ', code, 'registered.');
            }
        };
        PhotonPeer.prototype._removeListener = function (listeners, code, callback) {
            if ((code in listeners)) {
                var prevLenght = listeners[code].length;
                listeners[code] = listeners[code].filter(function (x) { return x != callback; });
                this._logger.debug('PhotonPeer[_removeListener] - Removing listener for event', code, "removed:", prevLenght - listeners[code].length);
            }
            return this;
        };
        PhotonPeer.prototype._removeListenersForCode = function (listeners, code) {
            this._logger.debug('PhotonPeer[_removeListenersForCode] - Removing all listeners for event', code);
            if (code in listeners) {
                listeners[code] = [];
            }
            return this;
        };
        /**
            @summary Defines how frequently native peer service() gets called. Lower values decrease lattency but take more cpu.
            @member Photon.PhotonPeer#serviceIntervalMs
            @type {number}
            @default 20
        */
        PhotonPeer.serviceIntervalMs = 20;
        /**
            @summary Enum for peer status codes.
            Use to subscribe to status changes.
            @member Photon.PhotonPeer.StatusCodes
            @readonly
            @property {string} connecting Is connecting to server.
            @property {string} connect Connected to server.
            @property {string} connectFailed Connection to server failed.
            @property {string} disconnect Disconnected from server.
            @property {string} connectClosed Connection closed by server.
            @property {string} error General connection error.
            @property {string} timeout Disconnected from server for timeout.
        */
        PhotonPeer.StatusCodes = {
            connecting: "connecting",
            connect: "connect",
            connectFailed: "connectFailed",
            disconnect: "disconnect",
            connectClosed: "connectClosed",
            error: "error",
            timeout: "timeout"
        };
        PhotonPeer.StatusCodeEmToJs = (_a = {},
            _a[EmStatusCode.EXCEPTION_ON_CONNECT] = PhotonPeer.StatusCodes.connectFailed,
            _a[EmStatusCode.CONNECT] = PhotonPeer.StatusCodes.connect,
            _a[EmStatusCode.DISCONNECT] = PhotonPeer.StatusCodes.disconnect,
            _a[EmStatusCode.EXCEPTION] = PhotonPeer.StatusCodes.error,
            _a[EmStatusCode.QUEUE_OUTGOING_RELIABLE_WARNING] = PhotonPeer.StatusCodes.connectFailed,
            _a[EmStatusCode.QUEUE_OUTGOING_UNRELIABLE_WARNING] = PhotonPeer.StatusCodes.connectFailed,
            _a[EmStatusCode.SEND_ERROR] = PhotonPeer.StatusCodes.error,
            //            [EmStatusCode.QUEUE_OUTGOING_ACKS_WARNING]        :  PhotonPeer.StatusCodes.connectFailed,
            //            [EmStatusCode.QUEUE_INCOMING_RELIABLE_WARNING]    :  PhotonPeer.StatusCodes.connectFailed,
            //            [EmStatusCode.QUEUE_INCOMING_UNRELIABLE_WARNING]  :  PhotonPeer.StatusCodes.connectFailed,
            //            [EmStatusCode.QUEUE_SENT_WARNING]                 :  PhotonPeer.StatusCodes.connectFailed,
            _a[EmStatusCode.INTERNAL_RECEIVE_EXCEPTION] = PhotonPeer.StatusCodes.error,
            _a[EmStatusCode.TIMEOUT_DISCONNECT] = PhotonPeer.StatusCodes.timeout,
            _a[EmStatusCode.DISCONNECT_BY_SERVER] = PhotonPeer.StatusCodes.connectClosed,
            _a[EmStatusCode.DISCONNECT_BY_SERVER_USER_LIMIT] = PhotonPeer.StatusCodes.connectClosed,
            _a[EmStatusCode.DISCONNECT_BY_SERVER_LOGIC] = PhotonPeer.StatusCodes.connectClosed,
            //            [EmStatusCode.ENCRYPTION_ESTABLISHED]             :  PhotonPeer.StatusCodes.connectFailed,
            _a[EmStatusCode.ENCRYPTION_FAILED_TO_ESTABLISH] = PhotonPeer.StatusCodes.error,
            _a);
        return PhotonPeer;
    }());
    Photon.PhotonPeer = PhotonPeer;
    var _a;
})(Photon || (Photon = {}));
/**
    Photon Load Balancing API
    @namespace Photon.LoadBalancing
*/
var Photon;
(function (Photon) {
    var LoadBalancing;
    (function (LoadBalancing) {
        var WebFlags = {
            HttpForward: 0x01,
            SendAuthCookie: 0x02,
            SendSync: 0x04,
            SendState: 0x08
        };
        var Actor = /** @class */ (function () {
            /**
                @classdesc Summarizes a "player" within a room, identified (in that room) by ID (or "actorNr"). Extend to implement custom logic.
                @constructor Photon.LoadBalancing.Actor
                @param {string} name Actor name.
                @param {number} actorNr Actor ID.
                @param {boolean} isLocal Actor is local.
            */
            function Actor(name, actorNr, isLocal) {
                this.name = name;
                this.actorNr = actorNr;
                this.isLocal = isLocal;
                this.customProperties = {};
                this.suspended = false;
            }
            // public getLoadBalancingClient() { return this.loadBalancingClient; }
            /**
                @summary Actor's room: the room initialized by client for create room operation or room client connected to.
                @method Photon.LoadBalancing.Actor#getRoom
                @returns {Photon.LoadBalancing.Room} Actor's room.
            */
            Actor.prototype.getRoom = function () { return this.loadBalancingClient.myRoom(); };
            /**
                @summary Raises game custom event.
                @method Photon.LoadBalancing.Actor#raiseEvent
                @param {number} eventCode Identifies this type of event (and the content). Your game's event codes can start with 0.
                @param {object} [data] Custom data you want to send along (use null, if none).
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {number} [options.interestGroup] The ID of the interest group this event goes to (exclusively).
                @property {Photon.LoadBalancing.Constants.EventCaching} [options.cache=EventCaching.DoNotCache] Events can be cached (merged and removed) for players joining later on.
                @property {Photon.LoadBalancing.Constants.ReceiverGroup} [options.receivers=ReceiverGroup.Others] Defines to which group of players the event is passed on.
                @property {number[]} [options.targetActors] Defines the target players who should receive the event (use only for small target groups).
                @property {boolean} [options.webForward=false] Forward to web hook.
            */
            Actor.prototype.raiseEvent = function (eventCode, data, options) {
                if (this.loadBalancingClient) {
                    this.loadBalancingClient.raiseEvent(eventCode, data, options);
                }
            };
            /**
                @summary Sets actor name.
                @method Photon.LoadBalancing.Actor#setName
                @param {string} name Actor name.
            */
            Actor.prototype.setName = function (name) { this.name = name; };
            // properties methods
            /**
                @summary Called on every actor properties update: properties set by client, poperties update from server.
                Override to update custom room state.
                @method Photon.LoadBalancing.Actor#onPropertiesChange
                @param {object} changedCustomProps Key-value map of changed properties.
                @param {boolean} [byClient] true if properties set by client.
            */
            Actor.prototype.onPropertiesChange = function (changedCustomProps, byClient) { };
            /**
                @summary Returns custom property by name.
                @method Photon.LoadBalancing.Actor#getCustomProperty
                @param {string} name Name of the property.
                @returns {object} Property or undefined if property not found.
            */
            Actor.prototype.getCustomProperty = function (name) { return this.customProperties[name]; };
            /**
                @summary Returns custom property by name or default value.
                @method Photon.LoadBalancing.Actor#getCustomPropertyOrElse
                @param {string} name Name of the property.
                @param {object} defaultValue Default property value.
                @returns {object} Property or default value if property not found.
            */
            Actor.prototype.getCustomPropertyOrElse = function (name, defaultValue) { return Exitgames.Common.Util.getPropertyOrElse(this.customProperties, name, defaultValue); };
            /**
                @summary Sets custom property.
                @method Photon.LoadBalancing.Actor#setCustomProperty
                @param {string} name Name of the property.
                @param {object} value Property value.
                @param {boolean} [webForward=false] Forward to web hook.
                @param {object} [expectedValue] Property value expected when update occurs. (CAS : "Check And Swap")
            */
            Actor.prototype.setCustomProperty = function (name, value, webForward, expectedValue) {
                if (webForward === void 0) { webForward = false; }
                this.customProperties[name] = value;
                var props = {};
                props[name] = value;
                var expectedProps;
                if (expectedValue != undefined) {
                    expectedProps = (_a = {}, _a[name] = expectedValue, _a);
                }
                if (this.loadBalancingClient && this.loadBalancingClient.isJoinedToRoom()) {
                    this.loadBalancingClient._setPropertiesOfActor(this.actorNr, props, webForward, expectedProps);
                }
                this.onPropertiesChange(props, true);
                var _a;
            };
            /**
                @summary Sets custom properties.
                @method Photon.LoadBalancing.Actor#setCustomProperties
                @param {object} properties Table of properties to set.
                @param {boolean} [webForward=false] Forward to web hook.
                @param {object} [expectedProperties] Table of properties expected when update occurs. (CAS : "Check And Swap")
            */
            Actor.prototype.setCustomProperties = function (properties, webForward, expectedProperties) {
                if (webForward === void 0) { webForward = false; }
                var props = {};
                for (var name in properties) {
                    this.customProperties[name] = properties[name];
                    props[name] = properties[name];
                }
                if (this.loadBalancingClient && this.loadBalancingClient.isJoinedToRoom()) {
                    this.loadBalancingClient._setPropertiesOfActor(this.actorNr, props, webForward, expectedProperties);
                }
                this.onPropertiesChange(props, true);
            };
            /**
                @summary Returns true if actor is in suspended state.
                @returns {boolean} Actor suspend state.
            **/
            Actor.prototype.isSuspended = function () {
                return this.suspended;
            };
            Actor.prototype._getAllProperties = function () {
                var p = {};
                p[LoadBalancing.Constants.ActorProperties.PlayerName] = this.name;
                for (var k in this.customProperties) {
                    p[k] = this.customProperties[k];
                }
                return p;
            };
            Actor.prototype._setLBC = function (lbc) { this.loadBalancingClient = lbc; };
            Actor.prototype._updateFromResponse = function (vals) {
                this.actorNr = vals[LoadBalancing.Constants.ParameterCode.ActorNr];
                var props = vals[LoadBalancing.Constants.ParameterCode.PlayerProperties];
                if (props !== undefined) {
                    var name = props[LoadBalancing.Constants.ActorProperties.PlayerName];
                    if (name !== undefined) {
                        this.name = name;
                    }
                    this._updateCustomProperties(props);
                }
            };
            Actor.prototype._updateMyActorFromResponse = function (vals) {
                this.actorNr = vals[LoadBalancing.Constants.ParameterCode.ActorNr];
            };
            Actor.prototype._updateCustomProperties = function (vals) {
                for (var p in vals) {
                    this.customProperties[p] = vals[p];
                }
                this.onPropertiesChange(vals, false);
            };
            Actor.prototype._setSuspended = function (s) {
                this.suspended = s;
            };
            Actor._getActorNrFromResponse = function (vals) {
                return vals[LoadBalancing.Constants.ParameterCode.ActorNr];
            };
            return Actor;
        }());
        LoadBalancing.Actor = Actor;
        // readonly room info from server
        var RoomInfo = /** @class */ (function () {
            /**
                @classdesc Used for Room listings of the lobby (not yet joining). Offers the basic info about a room: name, player counts, properties, etc.
                @constructor Photon.LoadBalancing.RoomInfo
                @param {string} name Room name.
            */
            function RoomInfo(name) {
                // standard room properties
                // TODO: access via getters
                /**
                    @summary Room name.
                    @member Photon.LoadBalancing.RoomInfo#name
                    @type {string}
                    @readonly
                */
                this.name = "";
                /**
                    @summary Joined room Game server address.
                    @member Photon.LoadBalancing.RoomInfo#address
                    @type {string}
                    @readonly
                */
                this.address = "";
                /**
                    @summary Max players before room is considered full.
                    @member Photon.LoadBalancing.RoomInfo#maxPlayers
                    @type {number}
                    @readonly
                */
                this.maxPlayers = 0;
                /**
                    @summary Shows the room in the lobby's room list. Makes sense only for local room.
                    @member Photon.LoadBalancing.RoomInfo#isVisible
                    @type {boolean}
                    @readonly
                */
                this.isVisible = true;
                /**
                    @summary Defines if this room can be joined.
                    @member Photon.LoadBalancing.RoomInfo#isOpen
                    @type {boolean}
                    @readonly
                */
                this.isOpen = true;
                /**
                    @summary Count of player currently in room.
                    @member Photon.LoadBalancing.RoomInfo#playerCount
                    @type {number}
                    @readonly
                */
                this.playerCount = 0;
                /**
                    @summary Time in ms indicating how long the room instance will be keeped alive in the server room cache after all clients have left the room.
                    @member Photon.LoadBalancing.RoomInfo#emptyRoomLiveTime
                    @type {number}
                    @readonly
                */
                this.emptyRoomLiveTime = 0;
                /**
                    @summary Time in ms indicating how long suspended player will be kept in the room.
                    @member Photon.LoadBalancing.RoomInfo#suspendedPlayerLiveTime
                    @type {number}
                    @readonly
                **/
                this.suspendedPlayerLiveTime = 0;
                /**
                    @summary Room removed (in room list updates).
                    @member Photon.LoadBalancing.RoomInfo#removed
                    @type {boolean}
                    @readonly
                */
                this.removed = false;
                // TODO: does end user need this?
                this.cleanupCacheOnLeave = false;
                /**
                    @summary Master client set by game server. Note: Not all servers support this currently. If the value of the property is 0, use lowest actorid instead.
                    @member Photon.LoadBalancing.RoomInfo#masterClientId
                    @type { number }
                    @readonly
                */
                this.masterClientId = 0;
                // custom properties
                this._customProperties = {};
                this._propsListedInLobby = [];
                this.name = name;
            }
            /**
                @summary Called on every room properties update: room creation, properties set by client, poperties update from server.
                Override to update custom room state.
                @method Photon.LoadBalancing.RoomInfo#onPropertiesChange
                @param {object} changedCustomProps Key-value map of changed properties.
                @param {boolean} [byClient] true if called on room creation or properties set by client.
            */
            RoomInfo.prototype.onPropertiesChange = function (changedCustomProps, byClient) { };
            /**
                @summary Returns custom property by name.
                @method Photon.LoadBalancing.RoomInfo#getCustomProperty
                @param {string} name Name of the property.
                @returns {object} Property or undefined if property not found.
            */
            RoomInfo.prototype.getCustomProperty = function (prop) { return this._customProperties[prop]; };
            /**
                @summary Returns custom property by name or default value.
                @method Photon.LoadBalancing.RoomInfo#getCustomPropertyOrElse
                @param {string} name Name of the property.
                @param {object} defaultValue Default property value.
                @returns {object} Property or default value if property not found.
            */
            RoomInfo.prototype.getCustomPropertyOrElse = function (prop, defaultValue) { return Exitgames.Common.Util.getPropertyOrElse(this._customProperties, prop, defaultValue); };
            RoomInfo.prototype._updateFromMasterResponse = function (vals) {
                this.address = vals[LoadBalancing.Constants.ParameterCode.Address];
                var name = vals[LoadBalancing.Constants.ParameterCode.RoomName];
                if (name) {
                    this.name = name;
                }
            };
            RoomInfo.prototype._updateFromProps = function (props) {
                if (props) {
                    this.maxPlayers = this.updateIfExists(this.maxPlayers, LoadBalancing.Constants.GameProperties.MaxPlayers, props);
                    this.isVisible = this.updateIfExists(this.isVisible, LoadBalancing.Constants.GameProperties.IsVisible, props);
                    this.isOpen = this.updateIfExists(this.isOpen, LoadBalancing.Constants.GameProperties.IsOpen, props);
                    this.playerCount = this.updateIfExists(this.playerCount, LoadBalancing.Constants.GameProperties.PlayerCount, props);
                    this.removed = this.updateIfExists(this.removed, LoadBalancing.Constants.GameProperties.Removed, props);
                    this._propsListedInLobby = this.updateIfExists(this._propsListedInLobby, LoadBalancing.Constants.GameProperties.PropsListedInLobby, props);
                    this.cleanupCacheOnLeave = this.updateIfExists(this.cleanupCacheOnLeave, LoadBalancing.Constants.GameProperties.CleanupCacheOnLeave, props);
                    this.masterClientId = this.updateIfExists(this.masterClientId, LoadBalancing.Constants.GameProperties.MasterClientId, props);
                    this.emptyRoomLiveTime = this.updateIfExists(this.emptyRoomLiveTime, LoadBalancing.Constants.GameProperties.EmptyRoomTtl, props);
                    this.suspendedPlayerLiveTime = this.updateIfExists(this.suspendedPlayerLiveTime, LoadBalancing.Constants.GameProperties.PlayerTtl, props);
                    var changedProps = {};
                    for (var k in props) {
                        if (parseInt(k).toString() != k) {
                            if (this._customProperties[k] !== props[k]) {
                                this._customProperties[k] = props[k];
                                changedProps[k] = props[k];
                            }
                        }
                    }
                    this.onPropertiesChange(changedProps, false);
                }
            };
            RoomInfo.prototype._updateFromEvent = function (payload) {
                if (payload) {
                    this.masterClientId = this.updateIfExists(this.masterClientId, LoadBalancing.Constants.ParameterCode.MasterClientId, payload);
                }
            };
            RoomInfo.prototype.updateIfExists = function (prevValue, code, props) {
                if (props.hasOwnProperty(code)) {
                    return props[code];
                }
                else {
                    return prevValue;
                }
            };
            return RoomInfo;
        }());
        LoadBalancing.RoomInfo = RoomInfo;
        // joined room with writable properties
        var Room = /** @class */ (function (_super) {
            __extends(Room, _super);
            /**
                @classdesc Represents a room client joins or is joined to. Extend to implement custom logic. Custom properties can be set via setCustomProperty() while being in the room.
                @mixes Photon.LoadBalancing.RoomInfo
                @constructor Photon.LoadBalancing.Room
                @param {string} name Room name.
            */
            function Room(name) {
                return _super.call(this, name) || this;
            }
            // room created from client via factory always has this field set
            //public getLoadBalancingClient() { return this.loadBalancingClient; }
            /**
                @summary Sets custom property
                @method Photon.LoadBalancing.Room#setCustomProperty
                @param {string} name Name of the property.
                @param {object} value Property value.
                @param {boolean} [webForward=false] Forward to web hook.
                @param {object} [expectedValue] Property value expected when update occurs. (CAS : "Check And Swap")
            */
            Room.prototype.setCustomProperty = function (name, value, webForward, expectedValue) {
                if (webForward === void 0) { webForward = false; }
                this._customProperties[name] = value;
                var props = {};
                props[name] = value;
                var expectedProps;
                if (expectedValue != undefined) {
                    expectedProps = (_a = {}, _a[name] = expectedValue, _a);
                }
                if (this.loadBalancingClient && this.loadBalancingClient.isJoinedToRoom()) {
                    this.loadBalancingClient._setPropertiesOfRoom(props, webForward, expectedProps);
                }
                this.onPropertiesChange(props, true);
                var _a;
            };
            /**
                @summary Sets custom property
                @method Photon.LoadBalancing.Room#setCustomProperties
                @param {object} properties Table of properties to set.
                @param {boolean} [webForward=false] Forward to web hook.
                @param {object} [expectedProperties] Table of properties expected when update occurs. (CAS : "Check And Swap")
            */
            Room.prototype.setCustomProperties = function (properties, webForward, expectedProperties) {
                if (webForward === void 0) { webForward = false; }
                var props = {};
                for (var name in properties) {
                    this._customProperties[name] = properties[name];
                    props[name] = properties[name];
                }
                if (this.loadBalancingClient && this.loadBalancingClient.isJoinedToRoom()) {
                    this.loadBalancingClient._setPropertiesOfRoom(props, webForward, expectedProperties);
                }
                this.onPropertiesChange(props, true);
            };
            Room.prototype.setProp = function (name, value) {
                if (this.loadBalancingClient && this.loadBalancingClient.isJoinedToRoom()) {
                    var props = {};
                    props[name] = value;
                    this.loadBalancingClient._setPropertiesOfRoom(props, false, undefined);
                }
            };
            /**
             * @summary Sets rooms visibility in the lobby's room list.
             * @method Photon.LoadBalancing.Room#setIsVisible
             * @param {boolean} isVisible New visibility value.
            */
            Room.prototype.setIsVisible = function (isVisible) {
                if (this.isVisible != isVisible) {
                    this.isVisible = isVisible;
                    this.setProp(LoadBalancing.Constants.GameProperties.IsVisible, isVisible);
                }
            };
            /**
             * @summary Sets if this room can be joined.
             * @method Photon.LoadBalancing.Room#setIsOpen
             * @param {boolean} isOpen New property value.
            */
            Room.prototype.setIsOpen = function (isOpen) {
                if (this.isOpen != isOpen) {
                    this.isOpen = isOpen;
                    this.setProp(LoadBalancing.Constants.GameProperties.IsOpen, isOpen);
                }
            };
            /**
             * @summary Sets max players before room is considered full.
             * @method Photon.LoadBalancing.Room#setMaxPlayers
             * @param {number} maxPlayers New max players value.
            */
            Room.prototype.setMaxPlayers = function (maxPlayers) {
                if (this.maxPlayers != maxPlayers) {
                    this.maxPlayers = maxPlayers;
                    this.setProp(LoadBalancing.Constants.GameProperties.MaxPlayers, maxPlayers);
                }
            };
            /**
             * @summary Sets room live time in the server room cache after all clients have left the room.
             * @method Photon.LoadBalancing.Room#setEmptyRoomLiveTime
             * @param {number} emptyRoomLiveTime New live time value in ms.
            */
            Room.prototype.setEmptyRoomLiveTime = function (emptyRoomLiveTime) {
                if (this.emptyRoomLiveTime != emptyRoomLiveTime) {
                    this.emptyRoomLiveTime = emptyRoomLiveTime;
                    this.setProp(LoadBalancing.Constants.GameProperties.EmptyRoomTtl, emptyRoomLiveTime);
                }
            };
            /**
             * @summary Sets time in ms indicating how long suspended player will be kept in the room.
             * @method Photon.LoadBalancing.Room#setSuspendedPlayerLiveTime
             * @param {number} suspendedPlayerLiveTime New live time value in ms.
            */
            Room.prototype.setSuspendedPlayerLiveTime = function (suspendedPlayerLiveTime) {
                if (this.suspendedPlayerLiveTime != suspendedPlayerLiveTime) {
                    this.suspendedPlayerLiveTime = suspendedPlayerLiveTime;
                    this.setProp(LoadBalancing.Constants.GameProperties.PlayerTtl, suspendedPlayerLiveTime);
                }
            };
            /**
             * @summary Sets expected server plugins.
             * @method Photon.LoadBalancing.Room#setPlugins
             * @param {string[]} plugins New plugins list.
            */
            Room.prototype.setPlugins = function (plugins) {
                this.plugins = plugins;
            };
            /**
                @summary Sets list of the room properties to pass to the RoomInfo list in a lobby.
                @method Photon.LoadBalancing.Room#setPropsListedInLobby
                @param {string[]} props Array of properties names.
            */
            Room.prototype.setPropsListedInLobby = function (props) {
                this._propsListedInLobby = props;
            };
            Room.prototype._setLBC = function (lbc) { this.loadBalancingClient = lbc; };
            return Room;
        }(RoomInfo));
        LoadBalancing.Room = Room;
        var LoadBalancingClient = /** @class */ (function () {
            /**
                @classdesc Implements the Photon LoadBalancing workflow. This class should be extended to handle system or custom events and operation responses.
                @constructor Photon.LoadBalancing.LoadBalancingClient
                @param {Photon.ConnectionProtocol} protocol Connecton protocol.
                @param {string} appId Cloud application ID.
                @param {string} appVersion Cloud application version.
            */
            function LoadBalancingClient(protocol, appId, appVersion) {
                this.appId = appId;
                this.appVersion = appVersion;
                // protected
                this.autoJoinLobby = true; // hardcoded behaviour; inheritor class can override this
                // options mainly keep state between servers
                // set / cleared in connectToNameServer()(connectToRegionMaster()), connect()
                // lobbyName and lobbyType passed to JoinLobby operation (we don't have separate JoinLobby operation and set them in connect())
                this.connectOptions = {};
                // shares lobby info between Master and Game CreateGame calls (createRoomInternal)
                this.createRoomOptions = {};
                // shares options between Master and Game JoinGame operations
                this.joinRoomOptions = {};
                this.roomInfos = new Array();
                this.roomInfosDict = {}; // 'by name' access support
                this.actors = {};
                this.actorsArray = []; // actors 'at index' access support (Scirra/Costruct 2)
                this.lowestActorId = 0; // master client support
                this.userAuthType = LoadBalancing.Constants.CustomAuthenticationType.None;
                this.userAuthParameters = "";
                this.userAuthData = "";
                this.lobbyStatsRequestList = new Array();
                // protected
                this.state = LoadBalancingClient.State.Uninitialized;
                this.logger = new Exitgames.Common.Logger("Client:");
                this.validNextState = {};
                var serverAddress = "";
                if (typeof (protocol) == "number") {
                    this.connectionProtocol = protocol;
                    switch (protocol) {
                        case Photon.ConnectionProtocol.Ws:
                            this.masterServerAddress = "ws://app-eu.exitgamescloud.com:9090";
                            this.nameServerAddress = "ws://ns.exitgames.com:9093";
                            break;
                        case Photon.ConnectionProtocol.Wss:
                            this.masterServerAddress = "wss://app-eu.exitgamescloud.com:19090";
                            this.nameServerAddress = "wss://ns.exitgames.com:19093";
                            break;
                        default:
                            var s0 = "wrong_protocol_error";
                            this.masterServerAddress = s0;
                            this.nameServerAddress = s0;
                            this.logger.error("Wrong protocol: ", protocol);
                            break;
                    }
                }
                else if (typeof (protocol) == "string") {
                    this.connectionProtocol = Photon.ConnectionProtocol.Ws;
                    var s = protocol;
                    this.masterServerAddress = s;
                    this.nameServerAddress = s;
                }
                else {
                    this.connectionProtocol = Photon.ConnectionProtocol.Ws;
                    var s1 = "wrong_protocol_type_error";
                    this.masterServerAddress = s1;
                    this.nameServerAddress = s1;
                    this.logger.error("Wrong protocol type: ", typeof (protocol));
                }
                this.initValidNextState();
                this.currentRoom = this.roomFactoryInternal("");
                this._myActor = this.actorFactoryInternal("", -1, true);
                this.addActor(this._myActor);
            }
            // override to handle system events:
            /**
                @summary Called on client state change. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onStateChange
                @param {Photon.LoadBalancing.LoadBalancingClient.State} state New client state.
            */
            LoadBalancingClient.prototype.onStateChange = function (state) { };
            /**
                @summary Called if client error occures. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onError
                @param {Photon.LoadBalancing.LoadBalancingClient.PeerErrorCode} errorCode Client error code.
                @param {string} errorMsg Error message.
            */
            LoadBalancingClient.prototype.onError = function (errorCode, errorMsg) { };
            /**
                @summary Called on operation response. Override if need custom workflow or response error handling.
                @method Photon.LoadBalancing.LoadBalancingClient#onOperationResponse
                @param {number} errorCode Server error code.
                @param {string} errorMsg Error message.
                @param {number} code Operation code.
                @param {object} content Operation response content.
            */
            LoadBalancingClient.prototype.onOperationResponse = function (errorCode, errorMsg, code, content) { };
            /**
                @summary Called on custom event. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onEvent
                @param {number} code Event code.
                @param {object} content Event content.
                @param {number} actorNr Actor ID event raised by.
            */
            LoadBalancingClient.prototype.onEvent = function (code, content, actorNr) { };
            /**
                @summary Called on room list received from Master server (on connection). Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onRoomList
                @param {{@link Photon.LoadBalancing.RoomInfo}[]} rooms Room list.
            */
            LoadBalancingClient.prototype.onRoomList = function (rooms) { };
            /**
                @summary Called on room list updates received from Master server. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onRoomListUpdate
                @param {{@link Photon.LoadBalancing.RoomInfo}[]} rooms Updated room list.
                @param {{@link Photon.LoadBalancing.RoomInfo}[]} roomsUpdated Rooms whose properties were changed.
                @param {{@link Photon.LoadBalancing.RoomInfo}[]} roomsAdded New rooms in list.
                @param {{@link Photon.LoadBalancing.RoomInfo}[]} roomsRemoved Rooms removed from list.
            */
            LoadBalancingClient.prototype.onRoomListUpdate = function (rooms, roomsUpdated, roomsAdded, roomsRemoved) { };
            // TODO: move to Room? Or remove and use Room.onPropertiesChange only?
            /**
                @summary Called on joined room properties changed event. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onMyRoomPropertiesChange
            */
            LoadBalancingClient.prototype.onMyRoomPropertiesChange = function () { };
            /**
                @summary Called on actor properties changed event. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onActorPropertiesChange
                @param {Photon.LoadBalancing.Actor} actor Actor whose properties were changed.
            */
            LoadBalancingClient.prototype.onActorPropertiesChange = function (actor) { };
            /**
                @summary Called when client joins room. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onJoinRoom
                @param {boolean} createdByMe True if room is created by client.
            */
            LoadBalancingClient.prototype.onJoinRoom = function (createdByMe) { };
            /**
                @summary Called when new actor joins the room client joined to. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onActorJoin
                @param {Photon.LoadBalancing.Actor} actor New actor.
            */
            LoadBalancingClient.prototype.onActorJoin = function (actor) { };
            /**
                @summary Called when actor leaves the room client joined to. Also called for every actor during room cleanup. Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onActorLeave
                @param {Photon.LoadBalancing.Actor} actor Actor left the room.
                @param {boolean} cleanup True if called during room cleanup (e.g. on disconnect).
            */
            LoadBalancingClient.prototype.onActorLeave = function (actor, cleanup) { };
            /**
                @summary Called when actor suspended in the room client joined to.Override to handle it.
                @method Photon.LoadBalancing.LoadBalancingClient#onActorSuspend
                @param {Photon.LoadBalancing.Actor} actor Actor suspended in the room.
            */
            LoadBalancingClient.prototype.onActorSuspend = function (actor) { };
            /**
                @summary Called when {@link Photon.LoadBalancing.LoadBalancingClient#findFriends findFriends} request completed. <br/>
                Override to handle request results.
                @method Photon.LoadBalancing.LoadBalancingClient#onFindFriendsResult
                @param {number} errorCode Result error code. 0 if request is successful.
                @param {string} errorMsg Error message.
                @param {object} friends Table with actors names as keys and friend statuses as values: {name1: friendStatus1, name2: friendStatus2, ... }.
                @property {object} friendStatus Friend status.
                @property {boolean} friendStatus.online Online status.
                @property {string} friendStatus.roomId Joined room.
            */
            LoadBalancingClient.prototype.onFindFriendsResult = function (errorCode, errorMsg, friends) { };
            /**
                @summary Called when lobbies statistics update received. <br/>
                Update can be automated by set up during {@link Photon.LoadBalancing.LoadBalancingClient#connect connect} or requested explicitly by {@link Photon.LoadBalancing.LoadBalancingClient#requestLobbyStats requestLobbyStats}. <br/>
                Override to handle request results.
                @method Photon.LoadBalancing.LoadBalancingClient#onLobbyStats
                @param {number} errorCode Result error code. 0 if request is successful. For automated updates is always 0.
                @param {string} errorMsg Error message. For automated updates is always empty.
                @param {object[]} lobbies Array of lobbies statistics: [lobbyStats1, lobbyStats1, ... ].
                @property {object} lobbyStats Lobby statistics.
                @property {string} lobbyStats.lobbyName Lobby name.
                @property {number} lobbyStats.lobbyType Lobby type.
                @property {number} lobbyStats.peerCount The number of players in the lobby (on Master, not playing).
                @property {number} lobbyStats.gameCount The number of games in the lobby.
            */
            LoadBalancingClient.prototype.onLobbyStats = function (errorCode, errorMsg, lobbies) { };
            /**
                @summary Called when application statistics update received. <br/>
                Override to handle request results.
                @method Photon.LoadBalancing.LoadBalancingClient#onAppStats
                @param {number} errorCode Result error code. Currently is always 0.
                @param {string} errorMsg Error message. Currently is always empty.
                @param {object} stats Application statistics.
                @property {object} stats Application statistics.
                @property {number} stats.peerCount Count of players currently online on Game servers.
                @property {number} stats.masterPeerCount Count of players on Master server (looking for game).
                @property {number} stats.gameCount Count of games currently in use (includes invisible and full rooms, so it doesn't match lobby list).
            */
            LoadBalancingClient.prototype.onAppStats = function (errorCode, errorMsg, stats) { };
            /**
                @summary Called when {@link Photon.LoadBalancing.LoadBalancingClient#getRegions getRegions} request completed.<br/>
                Override to handle request results.
                @param {number} errorCode Result error code. 0 if request is successful.
                @param {string} errorMsg Error message.
                @param {object} regions Object with region codes as keys and Master servers addresses as values
            */
            LoadBalancingClient.prototype.onGetRegionsResult = function (errorCode, errorMsg, regions) { };
            /**
                Called when {@link Photon.LoadBalancing.LoadBalancingClient#webRpc webRpc} request completed.<br/>
                Override to handle request results.
                @param {number} errorCode Result error code. 0 if request is successful.
                @param {string} message Error message if errorCode ~ = 0 or optional message returned by remote procedure.
                @param {string} uriPath Request path.
                @param {number} resultCode Result code returned by remote procedure.
                @param {object} data Data returned by remote procedure.
            **/
            LoadBalancingClient.prototype.onWebRpcResult = function (errorCode, message, uriPath, resultCode, data) { };
            /**
                @summary Override with creation of custom room (extended from Room): { return new CustomRoom(...); }
                @method Photon.LoadBalancing.LoadBalancingClient#roomFactory
                @param {string} name Room name. Pass to super() in custom actor constructor.
            */
            LoadBalancingClient.prototype.roomFactory = function (name) { return new Room(name); };
            /**
                @summary Override with creation of custom actor (extended from Actor): { return new CustomActor(...); }
                @method Photon.LoadBalancing.LoadBalancingClient#actorFactory
                @param {string} name Actor name. Pass to super() in custom room constructor.
                @param {number} actorNr Actor ID. Pass to super() in custom room constructor.
                @param {boolean} isLocal Actor is local. Pass to super() in custom room constructor.
            */
            LoadBalancingClient.prototype.actorFactory = function (name, actorNr, isLocal) { return new Actor(name, actorNr, isLocal); };
            //------------------------
            /**
                @summary Returns local actor.
                Client always has local actor even if not joined.
                @method Photon.LoadBalancing.LoadBalancingClient#myActor
                @returns {Photon.LoadBalancing.Actor} Local actor.
            */
            LoadBalancingClient.prototype.myActor = function () { return this._myActor; };
            /**
                @summary Returns client's room.
                Client always has it's room even if not joined. It's used for room creation operation.
                @method Photon.LoadBalancing.LoadBalancingClient#myRoom
                @returns {Photon.LoadBalancing.Room} Current room.
            */
            LoadBalancingClient.prototype.myRoom = function () { return this.currentRoom; };
            /**
                @summary Returns actors in room client currently joined including local actor.
                @method Photon.LoadBalancing.LoadBalancingClient#myRoomActors
                @returns {object} actorNr -> {@link Photon.LoadBalancing.Actor} map of actors in room.
            */
            LoadBalancingClient.prototype.myRoomActors = function () { return this.actors; };
            /**
                @summary Returns numer of actors in room client currently joined including local actor.
                @method Photon.LoadBalancing.LoadBalancingClient#myRoomActorCount
                @returns {number} Number of actors.
            */
            LoadBalancingClient.prototype.myRoomActorCount = function () { return this.actorsArray.length; };
            LoadBalancingClient.prototype.myRoomActorsArray = function () { return this.actorsArray; }; // actors 'at index' access support (Scirra/Costruct 2)                
            /**
                @summary Actor number of the player who's the master of this Room. Note: This changes when the current master leaves the room.
                @member Photon.LoadBalancing.RoomInfo#masterClientId
                @type {number}
                @readonly
            */
            LoadBalancingClient.prototype.myRoomMasterActorNr = function () {
                if (this.myRoom().masterClientId) {
                    return this.myRoom().masterClientId;
                }
                else {
                    return this.lowestActorId;
                }
            };
            LoadBalancingClient.prototype.lastRtt = function () {
                return this.gamePeer && this.gamePeer.getLastRtt();
            };
            LoadBalancingClient.prototype.roomFactoryInternal = function (name) {
                if (name === void 0) { name = ""; }
                var r = this.roomFactory(name);
                r._setLBC(this);
                return r;
            };
            LoadBalancingClient.prototype.actorFactoryInternal = function (name, actorNr, isLocal) {
                if (name === void 0) { name = ""; }
                if (actorNr === void 0) { actorNr = -1; }
                if (isLocal === void 0) { isLocal = false; }
                var a = this.actorFactory(name, actorNr, isLocal);
                a._setLBC(this);
                return a;
            };
            /**
                @summary Changes default NameServer address and port before connecting to NameServer.
                @method Photon.LoadBalancing.LoadBalancingClient#setNameServerAddress
                @param {string} address New address and port.
            */
            LoadBalancingClient.prototype.setNameServerAddress = function (address) {
                this.nameServerAddress = address;
            };
            /**
                @summary Returns current NameServer address.
                @method Photon.LoadBalancing.LoadBalancingClient#getNameServerAddress
                @returns {string} NameServer address address.
            */
            LoadBalancingClient.prototype.getNameServerAddress = function () {
                return this.nameServerAddress;
            };
            /**
                @summary Changes default Master server address and port before connecting to Master server.
                @method Photon.LoadBalancing.LoadBalancingClient#setMasterServerAddress
                @param {string} address New address and port.
            */
            LoadBalancingClient.prototype.setMasterServerAddress = function (address) {
                this.masterServerAddress = address;
            };
            /**
                @summary Returns current Master server address.
                @method Photon.LoadBalancing.LoadBalancingClient#getMasterServerAddress
                @returns {string} Master server address.
            */
            LoadBalancingClient.prototype.getMasterServerAddress = function () {
                return this.nameServerAddress;
            };
            /**
                @summary Sets optional user id(required by some cloud services)
                @method Photon.LoadBalancing.LoadBalancingClient#setUserId
                @param {string} userId New user id.
            */
            LoadBalancingClient.prototype.setUserId = function (userId) {
                this.userId = userId;
            };
            /**
                @summary Returns previously set user id.
                @method Photon.LoadBalancing.LoadBalancingClient#getUserId
                @returns {string} User id.
            */
            LoadBalancingClient.prototype.getUserId = function () {
                return this.userId;
            };
            /**
                @summary Enables custom authentication and sets it's parameters.
                @method Photon.LoadBalancing.LoadBalancingClient#setCustomAuthentication
                @param {string} authParameters This string must contain any (http get) parameters expected by the used authentication service.
                @param {Photon.LoadBalancing.Constants.CustomAuthenticationType} [authType=Photon.LoadBalancing.Constants.CustomAuthenticationType.Custom] The type of custom authentication provider that should be used.
                @param {any} [authData] The data to be passed-on to the auth service via POST. String passed as is, objects as application/json
            */
            LoadBalancingClient.prototype.setCustomAuthentication = function (authParameters, authType, authData) {
                if (authType === void 0) { authType = Photon.LoadBalancing.Constants.CustomAuthenticationType.Custom; }
                this.userAuthType = authType;
                this.userAuthParameters = authParameters;
                this.userAuthData = authData;
            };
            // TODO: remove backward compatibility (deprecated)
            /**
                @summary Starts connection to Master server.
                @method Photon.LoadBalancing.LoadBalancingClient#connect
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {boolean} [options.keepMasterConnection=false] Don't disconnect from Master server after joining room.
                @property {string} [options.lobbyName] Name of the lobby connect to.
                @property {Photon.LoadBalancing.Constants.LobbyType} [options.lobbyType=LobbyType.Default] Type of the lobby.
                @property {boolean} [options.lobbyStats=false] If true, Master server will be sending lobbies statistics periodically.<br/> Override {@link Photon.LoadBalancing.LoadBalancingClient#onLobbyStats onLobbyStats} to handle request results.<br/>Alternatively, {@link Photon.LoadBalancing.LoadBalancingClient#requestLobbyStats requestLobbyStats} can be used.
                @returns {boolean} True if current client state allows connection.
            */
            LoadBalancingClient.prototype.connect = function (options) {
                // backward compatibility
                if (typeof (options) === "boolean") {
                    if (options) {
                        options = { keepMasterConnection: true };
                    }
                    else {
                        options = { keepMasterConnection: false };
                    }
                }
                //
                if (!options) {
                    options = {};
                }
                if (this.checkNextState(LoadBalancingClient.State.ConnectingToMasterserver, true)) {
                    this.changeState(LoadBalancingClient.State.ConnectingToMasterserver);
                    this.logger.info("Connecting to Master", this.masterServerAddress);
                    // make options copy to protect
                    this.connectOptions = {};
                    for (var k in options)
                        this.connectOptions[k] = options[k];
                    if (this.masterPeer)
                        this.masterPeer.Destroy();
                    this.masterPeer = new MasterPeer(this, this.connectionProtocol, this.masterServerAddress, "");
                    this.initMasterPeer(this.masterPeer);
                    this.masterPeer.connect(this.appId);
                    return true;
                }
                else {
                    return false;
                }
            };
            /**
                @summary Starts connection to NameServer.
                @method Photon.LoadBalancing.LoadBalancingClient#connectToNameServer
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {string} [options.region] Don't disconnect from Master server after joining room.
                @property {string} [options.lobbyName] Name of the lobby connect to.
                @property {Photon.LoadBalancing.Constants.LobbyType} [options.lobbyType=LobbyType.Default] Type of the lobby.
                @property {boolean} [options.lobbyStats=false] If true, Master server will be sending lobbies statistics periodically.<br/> Override {@link Photon.LoadBalancing.LoadBalancingClient#onLobbyStats onLobbyStats} to handle request results.<br/>Alternatively, {@link Photon.LoadBalancing.LoadBalancingClient#requestLobbyStats requestLobbyStats} can be used.
                @property {boolean} [options.keepMasterConnection=false] Don't disconnect from Master server after joining room.
            */
            LoadBalancingClient.prototype.connectToNameServer = function (options) {
                if (!options) {
                    options = {};
                }
                if (this.checkNextState(LoadBalancingClient.State.ConnectingToNameServer, true)) {
                    this.changeState(LoadBalancingClient.State.ConnectingToNameServer);
                    this.logger.info("Connecting to NameServer", this.nameServerAddress);
                    // make options copy to protect
                    this.connectOptions = {};
                    for (var k in options)
                        this.connectOptions[k] = options[k];
                    if (this.nameServerPeer)
                        this.nameServerPeer.Destroy();
                    this.nameServerPeer = new NameServerPeer(this, this.connectionProtocol, this.nameServerAddress, "");
                    this.initNameServerPeer(this.nameServerPeer);
                    this.nameServerPeer.connect(this.appId);
                    return true;
                }
                else {
                    return false;
                }
            };
            LoadBalancingClient.prototype.fillCreateRoomOptions = function (op, options) {
                options = options || {};
                var gp = {};
                if (options.isVisible !== undefined)
                    gp[LoadBalancing.Constants.GameProperties.IsVisible] = options.isVisible;
                if (options.isOpen !== undefined)
                    gp[LoadBalancing.Constants.GameProperties.IsOpen] = options.isOpen;
                if (options.maxPlayers !== undefined)
                    gp[LoadBalancing.Constants.GameProperties.MaxPlayers] = options.maxPlayers;
                if (options.propsListedInLobby !== undefined)
                    gp[LoadBalancing.Constants.GameProperties.PropsListedInLobby] = Photon.TypeExt.String(options.propsListedInLobby);
                if (options.customGameProperties !== undefined) {
                    for (var p in options.customGameProperties) {
                        gp[p] = options.customGameProperties[p];
                    }
                }
                op.push(LoadBalancing.Constants.ParameterCode.GameProperties, gp);
                op.push(LoadBalancing.Constants.ParameterCode.CleanupCacheOnLeave, true); //TODO: make this optional?
                op.push(LoadBalancing.Constants.ParameterCode.Broadcast, true); //TODO: make this optional?
                if (options.emptyRoomLiveTime !== undefined)
                    op.push(LoadBalancing.Constants.ParameterCode.EmptyRoomTTL, Photon.TypeExt.Int(options.emptyRoomLiveTime));
                if (options.suspendedPlayerLiveTime !== undefined)
                    op.push(LoadBalancing.Constants.ParameterCode.PlayerTTL, Photon.TypeExt.Int(options.suspendedPlayerLiveTime));
                if (options.plugins !== undefined)
                    op.push(LoadBalancing.Constants.ParameterCode.Plugins, options.plugins);
                // shold be always set to true by client
                op.push(LoadBalancing.Constants.ParameterCode.CheckUserOnJoin, true);
                if (options.lobbyName) {
                    op.push(LoadBalancing.Constants.ParameterCode.LobbyName);
                    op.push(options.lobbyName);
                    if (options.lobbyType != undefined) {
                        op.push(LoadBalancing.Constants.ParameterCode.LobbyType);
                        op.push(options.lobbyType);
                    }
                }
            };
            /**
                @summary Creates a new room on the server (or fails when the name is already taken). Takes parameters (except name) for new room from myRoom() object. Set them before call.
                @method Photon.LoadBalancing.LoadBalancingClient#createRoomFromMy
                @param {string} [roomName] New room name. Assigned automatically by server if empty or not specified.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {string} [options.lobbyName] Name of the lobby to create room in.
                @property {Photon.LoadBalancing.Constants.LobbyType} [options.lobbyType=LobbyType.Default] Type of the lobby.
            */
            LoadBalancingClient.prototype.createRoomFromMy = function (roomName, options) {
                this.currentRoom.name = roomName ? roomName : "";
                options = this.copyCreateOptionsFromMyRoom(options);
                return this.createRoomInternal(this.masterPeer, options);
            };
            LoadBalancingClient.prototype.copyCreateOptionsFromMyRoom = function (options) {
                options = options || {};
                //retrieve options from my room
                options.isVisible = this.currentRoom.isVisible;
                options.isOpen = this.currentRoom.isOpen;
                options.maxPlayers = this.currentRoom.maxPlayers;
                options.customGameProperties = this.currentRoom._customProperties;
                options.propsListedInLobby = this.currentRoom._propsListedInLobby;
                options.emptyRoomLiveTime = this.currentRoom.emptyRoomLiveTime;
                options.suspendedPlayerLiveTime = this.currentRoom.suspendedPlayerLiveTime;
                options.plugins = this.currentRoom.plugins;
                return options;
            };
            /**
                @summary Creates a new room on the server (or fails when the name is already taken).
                @method Photon.LoadBalancing.LoadBalancingClient#createRoom
                @param {string} [roomName] The name to create a room with. Must be unique and not in use or can't be created. If not specified or null, the server will assign a GUID as name.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {boolean} [options.isVisible=true] Shows the room in the lobby's room list.
                @property {boolean} [options.isOpen=true] Keeps players from joining the room (or opens it to everyone).
                @property {number} [options.maxPlayers=0] Max players before room is considered full (but still listed).
                @property {object} [options.customGameProperties] Custom properties to apply to the room on creation (use string-typed keys but short ones).
                @property {string[]} [options.propsListedInLobby] Defines the custom room properties that get listed in the lobby.
                @property {number} [options.emptyRoomLiveTime=0] Room live time (ms) in the server room cache after all clients have left the room.
                @property {number} [options.suspendedPlayerLiveTime=0] Player live time (ms) in the room after player suspended.
                @property {string[]} [options.plugins] Expected server plugins.
                @property {string} [options.lobbyName] Name of the lobby to create room in.
                @property {Photon.LoadBalancing.Constants.LobbyType} [options.lobbyType=LobbyType.Default] Type of the lobby.
    
            */
            LoadBalancingClient.prototype.createRoom = function (roomName, options) {
                this.currentRoom = this.roomFactoryInternal(roomName ? roomName : "");
                return this.createRoomInternal(this.masterPeer, options);
            };
            /**
                @summary Joins a room by name and sets this player's properties.
                @method Photon.LoadBalancing.LoadBalancingClient#joinRoom
                @param {string} roomName The name of the room to join. Must be existing already, open and non-full or can't be joined.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {string} [options.rejoin=false] Rejoin using current userId.
                @property {boolean} [options.createIfNotExists=false] Create room if not exists.
                @param {object} [createOptions] Room options for creation
                @property {object} createOptions Room options for creation
                @property {boolean} [createOptions.isVisible=true] Shows the room in the lobby's room list.
                @property {boolean} [createOptions.isOpen=true] Keeps players from joining the room (or opens it to everyone).
                @property {number} [createOptions.maxPlayers=0] Max players before room is considered full (but still listed).
                @property {object} [createOptions.customGameProperties] Custom properties to apply to the room on creation (use string-typed keys but short ones).
                @property {string[]} [createOptions.propsListedInLobby] Defines the custom room properties that get listed in the lobby.
                @property {number} [createOptions.emptyRoomLiveTime=0] Room live time (ms) in the server room cache after all clients have left the room.
                @property {number} [createOptions.suspendedPlayerLiveTime=0] Player live time (ms) in the room after player suspended.
                @property {string[]} [createOptions.plugins] Informs the server of the expected plugin setup.
                @property {string} [createOptions.lobbyName=""] Name of the lobby to create room in.
                @property {Photon.LoadBalancing.Constants.LobbyType} [createOptions.lobbyType=LobbyType.Default] Type of the lobby.
    
            */
            LoadBalancingClient.prototype.joinRoom = function (roomName, options, createOptions) {
                var op = [];
                if (options) {
                    if (options.createIfNotExists) {
                        op.push(LoadBalancing.Constants.ParameterCode.JoinMode, LoadBalancingClient.JoinMode.CreateIfNotExists);
                        this.fillCreateRoomOptions(op, createOptions);
                    }
                    if (options.rejoin) {
                        op.push(LoadBalancing.Constants.ParameterCode.JoinMode, LoadBalancingClient.JoinMode.RejoinOnly);
                    }
                }
                this.currentRoom = this.roomFactoryInternal(roomName);
                op.push(LoadBalancing.Constants.ParameterCode.RoomName, roomName);
                this.joinRoomOptions = options || {};
                this.createRoomOptions = createOptions || {};
                this.logger.info("Join Room", roomName, options, createOptions, "...");
                this.masterPeer.sendOperation(LoadBalancing.Constants.OperationCode.JoinGame, op);
                return true;
            };
            /**
                @summary Joins a random, available room.
                This operation fails if all rooms are closed or full.
                @method Photon.LoadBalancing.LoadBalancingClient#joinRandomRoom
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {object} [options.expectedCustomRoomProperties] If specified, a room will only be joined, if it matches these custom properties. Use null to accept rooms with any properties.
                @property {number} [options.expectedMaxPlayers] If specified, filters for a particular maxPlayer setting. Use 0 to accept any maxPlayer value.
                @property {Photon.LoadBalancing.Constants.MatchmakingMode} [options.matchmakingMode=MatchmakingMode.FillRoom] Selects one of the available matchmaking algorithms.
                @property {string} [options.lobbyName] Name of the lobby to search rooms in.
                @property {Photon.LoadBalancing.Constants.LobbyType} [options.lobbyType=LobbyType.Default] Type of the lobby.
                @property {string} [options.sqlLobbyFilter] Basically the "where" clause of a sql statement. Examples: 'C0 = 1 AND C2 > 50'. 'C5 = "Map2" AND C2 > 10 AND C2 < 20'
            */
            LoadBalancingClient.prototype.joinRandomRoom = function (options) {
                var op = [];
                if (options) {
                    if (options.matchingType != undefined && options.matchingType != LoadBalancing.Constants.MatchmakingMode.FillRoom) {
                        op.push(LoadBalancing.Constants.ParameterCode.MatchMakingType);
                        op.push(options.matchingType);
                    }
                    var expectedRoomProperties = {};
                    var propNonEmpty = false;
                    if (options.expectedCustomRoomProperties != undefined) {
                        for (var k in options.expectedCustomRoomProperties) {
                            expectedRoomProperties[k] = options.expectedCustomRoomProperties[k];
                            propNonEmpty = true;
                        }
                    }
                    if (options.expectedMaxPlayers != undefined && options.expectedMaxPlayers > 0) {
                        expectedRoomProperties[LoadBalancing.Constants.GameProperties.MaxPlayers] = options.expectedMaxPlayers;
                        propNonEmpty = true;
                    }
                    if (propNonEmpty) {
                        op.push(LoadBalancing.Constants.ParameterCode.GameProperties);
                        op.push(expectedRoomProperties);
                    }
                    if (options.lobbyName) {
                        op.push(LoadBalancing.Constants.ParameterCode.LobbyName);
                        op.push(options.lobbyName);
                        if (options.lobbyType != undefined) {
                            op.push(LoadBalancing.Constants.ParameterCode.LobbyType);
                            op.push(options.lobbyType);
                        }
                    }
                    if (options.sqlLobbyFilter) {
                        op.push(LoadBalancing.Constants.ParameterCode.Data);
                        op.push(options.sqlLobbyFilter);
                    }
                }
                this.logger.info("Join Random Room", options && options.lobbyName, options && options.lobbyType, "...");
                this.masterPeer.sendOperation(LoadBalancing.Constants.OperationCode.JoinRandomGame, op);
                return true;
            };
            LoadBalancingClient.prototype._setPropertiesOfRoom = function (properties, webForward, expectedProperties) {
                var op = [];
                op.push(LoadBalancing.Constants.ParameterCode.Properties);
                op.push(properties);
                op.push(LoadBalancing.Constants.ParameterCode.Broadcast);
                op.push(true);
                if (webForward) {
                    op.push(LoadBalancing.Constants.ParameterCode.WebFlags);
                    op.push(Photon.TypeExt.Byte(WebFlags.HttpForward));
                }
                if (expectedProperties) {
                    op.push(LoadBalancing.Constants.ParameterCode.ExpectedValues);
                    op.push(expectedProperties);
                }
                this.gamePeer.sendOperation(LoadBalancing.Constants.OperationCode.SetProperties, op);
            };
            LoadBalancingClient.prototype._setPropertiesOfActor = function (actorNr, properties, webForward, expectedProperties) {
                var op = [];
                op.push(LoadBalancing.Constants.ParameterCode.ActorNr);
                op.push(actorNr);
                op.push(LoadBalancing.Constants.ParameterCode.Properties);
                op.push(properties);
                op.push(LoadBalancing.Constants.ParameterCode.Broadcast);
                op.push(true);
                if (webForward) {
                    op.push(LoadBalancing.Constants.ParameterCode.WebFlags);
                    op.push(Photon.TypeExt.Byte(WebFlags.HttpForward));
                }
                if (expectedProperties) {
                    op.push(LoadBalancing.Constants.ParameterCode.ExpectedValues);
                    op.push(expectedProperties);
                }
                this.gamePeer.sendOperation(LoadBalancing.Constants.OperationCode.SetProperties, op);
            };
            /**
                @summary Disconnects from all servers.
                @method Photon.LoadBalancing.LoadBalancingClient#disconnect
            */
            LoadBalancingClient.prototype.disconnect = function () {
                if (this.nameServerPeer) {
                    this.nameServerPeer.disconnect();
                }
                this._cleanupNameServerPeerData();
                if (this.masterPeer) {
                    this.masterPeer.disconnect();
                }
                this._cleanupMasterPeerData();
                if (this.gamePeer) {
                    this.gamePeer.disconnect();
                }
                this._cleanupGamePeerData();
                this.changeState(LoadBalancingClient.State.Disconnected);
            };
            /**
                @summary Disconnects client from Game server keeping player in room (to rejoin later) and connects to Master server if not connected.
                @method Photon.LoadBalancing.LoadBalancingClient#suspendRoom
                @property {object} options Additional options
                @property {boolean} [options.sendAuthCookie] Securely transmit the encrypted object AuthCookie to the web service in PathLeave webhook when available
            */
            LoadBalancingClient.prototype.suspendRoom = function (options) {
                if (this.isJoinedToRoom()) {
                    if (this.gamePeer) {
                        var params = [];
                        if (options) {
                            if (options.sendAuthCookie) {
                                params.push(LoadBalancing.Constants.ParameterCode.WebFlags, Photon.TypeExt.Byte(WebFlags.SendAuthCookie));
                            }
                        }
                        params.push(LoadBalancing.Constants.ParameterCode.IsInactive, true);
                        this.gamePeer.sendOperation(LoadBalancing.Constants.OperationCode.Leave, params);
                        this.gamePeerWaitingForDisconnect = true;
                    }
                    this._cleanupGamePeerData();
                    if (this.isConnectedToMaster()) {
                        this.changeState(LoadBalancingClient.State.JoinedLobby);
                    }
                    else {
                        this.changeState(LoadBalancingClient.State.Disconnected);
                        this.connect(this.connectOptions);
                    }
                }
            };
            /**
                @summary Leaves room and connects to Master server if not connected.
                @method Photon.LoadBalancing.LoadBalancingClient#leaveRoom
                @property {object} options Additional options
                @property {boolean} [options.sendAuthCookie] Securely transmit the encrypted object AuthCookie to the web service in PathLeave webhook when available
            */
            LoadBalancingClient.prototype.leaveRoom = function (options) {
                if (this.isJoinedToRoom()) {
                    if (this.gamePeer) {
                        var params = [];
                        if (options) {
                            if (options.sendAuthCookie) {
                                params.push(LoadBalancing.Constants.ParameterCode.WebFlags, Photon.TypeExt.Byte(WebFlags.SendAuthCookie));
                            }
                        }
                        this.gamePeer.sendOperation(LoadBalancing.Constants.OperationCode.Leave, params);
                        this.gamePeerWaitingForDisconnect = true;
                    }
                    this._cleanupGamePeerData();
                    if (this.isConnectedToMaster()) {
                        this.changeState(LoadBalancingClient.State.JoinedLobby);
                    }
                    else {
                        this.changeState(LoadBalancingClient.State.Disconnected);
                        this.connect(this.connectOptions);
                    }
                }
            };
            /**
                @summary Raises game custom event
                @method Photon.LoadBalancing.LoadBalancingClient#raiseEvent
                @param {number} eventCode Identifies this type of event (and the content). Your game's event codes can start with 0.
                @param {object} [data] Custom data you want to send along (use null, if none).
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {number} [options.interestGroup] The ID of the interest group this event goes to (exclusively).
                @property {Photon.LoadBalancing.Constants.EventCaching} [options.cache=EventCaching.DoNotCache] Events can be cached (merged and removed) for players joining later on.
                @property {Photon.LoadBalancing.Constants.ReceiverGroup} [options.receivers=ReceiverGroup.Others] Defines to which group of players the event is passed on.
                @property {number[]} [options.targetActors] Defines the target players who should receive the event (use only for small target groups).
                @property {boolean} [options.webForward=false] Forward to web hook.
            */
            LoadBalancingClient.prototype.raiseEvent = function (eventCode, data, options) {
                if (this.isJoinedToRoom()) {
                    this.gamePeer.raiseEvent(eventCode, data, options);
                }
            };
            /**
                @summary Changes client's interest groups (for events in room).<br/>
                Note the difference between passing null and []: null won't add/remove any groups, [] will add/remove all (existing) groups.<br/>
                First, removing groups is executed. This way, you could leave all groups and join only the ones provided.
                @method Photon.LoadBalancing.LoadBalancingClient#changeGroups
                @param {number[]} groupsToRemove Groups to remove from interest. Null will not leave any. A [] will remove all.
                @param {number[]} groupsToAdd Groups to add to interest. Null will not add any. A [] will add all current.
            */
            LoadBalancingClient.prototype.changeGroups = function (groupsToRemove, groupsToAdd) {
                if (this.isJoinedToRoom()) {
                    this.logger.debug("Group change:", groupsToRemove, groupsToAdd);
                    this.gamePeer.changeGroups(groupsToRemove, groupsToAdd);
                }
            };
            /**
                @summary Requests Master server for actors online status and joined rooms.<br/>
                Override {@link Photon.LoadBalancing.LoadBalancingClient#onFindFriendsResult onFindFriendsResult} to handle request results.
                @method Photon.LoadBalancing.LoadBalancingClient#findFriends
                @param {string[]} friendsToFind Actors names.
            **/
            LoadBalancingClient.prototype.findFriends = function (friendsToFind) {
                if (this.isConnectedToMaster()) {
                    if (friendsToFind && typeof (friendsToFind) == "object") {
                        this.findFriendsRequestList = new Array();
                        for (var i = 0; i < friendsToFind.length; ++i) {
                            if (typeof (friendsToFind[i]) == "string") {
                                this.findFriendsRequestList[i] = friendsToFind[i];
                            }
                            else {
                                this.logger.error("FindFriends request error:", "Friend name is not a string", i);
                                this.onFindFriendsResult(-1, "Friend name is not a string" + " " + i, {});
                                return;
                            }
                        }
                        this.logger.debug("Find friends:", friendsToFind);
                        this.masterPeer.findFriends(this.findFriendsRequestList);
                    }
                    else {
                        this.logger.error("FindFriends request error:", "Parameter is not an array");
                        this.onFindFriendsResult(-1, "Parameter is not an array", {});
                    }
                }
                else {
                    this.logger.error("FindFriends request error:", "Not connected to Master");
                    this.onFindFriendsResult(LoadBalancingClient.PeerErrorCode.MasterError, "Not connected to Master", {});
                }
            };
            /**
                @summary Requests Master server for lobbies statistics.<br/>
                Override {@link Photon.LoadBalancing.LoadBalancingClient#onLobbyStats onLobbyStats} to handle request results.<br/>
                Alternatively, automated updates can be set up during {@link Photon.LoadBalancing.LoadBalancingClient#connect connect}.
                @method Photon.LoadBalancing.LoadBalancingClient#requestLobbyStats
                @param {any[]} lobbiesToRequest Array of lobbies id pairs [ [lobbyName1, lobbyType1], [lobbyName2, lobbyType2], ... ]. If not specified or null, statistics for all lobbies requested.
    
            **/
            LoadBalancingClient.prototype.requestLobbyStats = function (lobbiesToRequest) {
                if (this.isConnectedToMaster()) {
                    this.lobbyStatsRequestList = new Array();
                    if (lobbiesToRequest) {
                        if (typeof (lobbiesToRequest) == "object") {
                            for (var i = 0; i < lobbiesToRequest.length; ++i) {
                                var l = lobbiesToRequest[i];
                                if (typeof (l) == "object") {
                                    var n = l[0];
                                    if (n) {
                                        var t;
                                        if (l[1] === undefined) {
                                            t = LoadBalancing.Constants.LobbyType.Default;
                                        }
                                        else {
                                            if (typeof (l[1]) == "number") {
                                                t = l[1];
                                            }
                                            else {
                                                this.requestLobbyStatsErr("Lobby type is invalid", i);
                                                return;
                                            }
                                        }
                                        this.lobbyStatsRequestList[i] = [n.toString(), t];
                                    }
                                    else {
                                        this.requestLobbyStatsErr("Lobby name is empty", i);
                                        return;
                                    }
                                }
                                else {
                                    this.requestLobbyStatsErr("Lobby id is not an array", i);
                                    return;
                                }
                            }
                        }
                        else {
                            this.requestLobbyStatsErr("Parameter is not an array");
                            return;
                        }
                    }
                    this.masterPeer.requestLobbyStats(this.lobbyStatsRequestList);
                }
                else {
                    this.logger.error("LobbyState request error:", "Not connected to Master");
                    this.onLobbyStats(LoadBalancingClient.PeerErrorCode.MasterError, "Not connected to Master", []);
                }
            };
            LoadBalancingClient.prototype.requestLobbyStatsErr = function (m, other) {
                if (other === void 0) { other = ""; }
                this.logger.error("LobbyState request error:", m, other);
                this.onLobbyStats(-1, m + " " + other, []);
            };
            /**
                @summary Requests NameServer for regions list.<br/>
                Override {@link Photon.LoadBalancing.LoadBalancingClient#onGetRegionsResult onGetRegionsResult} to handle request results.<br/>
                @method Photon.LoadBalancing.LoadBalancingClient#getRegions
            **/
            LoadBalancingClient.prototype.getRegions = function () {
                if (this.isConnectedToNameServer()) {
                    this.logger.debug("GetRegions...");
                    this.nameServerPeer.getRegions(this.appId);
                }
                else {
                    this.logger.error("GetRegions request error:", "Not connected to NameServer");
                    this.onGetRegionsResult(LoadBalancingClient.PeerErrorCode.NameServerError, "Not connected to NameServer", {});
                }
            };
            /**
                @summary Sends web rpc request to Master server.<br/ >
                Override {@link Photon.LoadBalancing.LoadBalancingClient#onWebRpcResult onWebRpcResult} to handle request results.<br/>
                @method Photon.LoadBalancing.LoadBalancingClient#webRpc
                @param {string} uriPath Request path.
                @param {object} parameters Request parameters.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {boolean} [options.sendAuthCookie] Defines if the authentication cookie gets sent to a WebHook (if setup)
            **/
            LoadBalancingClient.prototype.webRpc = function (uriPath, parameters, options) {
                if (this.isConnectedToMaster()) {
                    this.logger.debug("WebRpc...");
                    this.masterPeer.webRpc(uriPath, parameters, options);
                }
                else if (this.isJoinedToRoom()) {
                    this.logger.debug("WebRpc...");
                    this.gamePeer.webRpc(uriPath, parameters, options);
                }
                else {
                    this.logger.error("WebRpc request error:", "Connected to neither Master nor Game server");
                    this.onWebRpcResult(LoadBalancingClient.PeerErrorCode.MasterError, "Connected to neither Master nor Game server", uriPath, 0, {});
                }
            };
            /**
                @summary Connects to a specific region's Master server, using the NameServer to find the IP.
                Override {@link Photon.LoadBalancing.LoadBalancingClient#onWebRpcResult onWebRpcResult} to handle request results.<br/>
                @method Photon.LoadBalancing.LoadBalancingClient#connectToRegionMaster
                @param {string} region Region connect to Master server of.
                @returns {boolean} True if current client state allows connection.
            **/
            LoadBalancingClient.prototype.connectToRegionMaster = function (region) {
                if (this.isConnectedToNameServer()) {
                    this.logger.debug("Connecting to Region Master", region, "...");
                    this.nameServerPeer.opAuth(this.appId, this.appVersion, this.userAuthType, this.userAuthParameters, this.userAuthData, this.userId, region);
                    return true;
                }
                else if (this.connectToNameServer({ region: region })) {
                    return true;
                }
                else {
                    this.logger.error("Connecting to Region Master error:", "Not connected to NameServer");
                    return false;
                }
            };
            /**
                @summary Checks if client is connected to Master server (usually joined to lobby and receives room list updates).
                @method Photon.LoadBalancing.LoadBalancingClient#isConnectedToMaster
                @returns {boolean} True if client is connected to Master server.
            */
            LoadBalancingClient.prototype.isConnectedToMaster = function () {
                return this.masterPeer && this.masterPeer.isConnected();
            };
            /**
                @summary Checks if client is connected to NameServer server.
                @method Photon.LoadBalancing.LoadBalancingClient#isConnectedToNameServer
                @returns {boolean} True if client is connected to NameServer server.
            */
            LoadBalancingClient.prototype.isConnectedToNameServer = function () {
                return this.nameServerPeer && this.nameServerPeer.isConnected();
            };
            /**
                @summary Checks if client is in lobby and ready to join or create game.
                @method Photon.LoadBalancing.LoadBalancingClient#isInLobby
                @returns {boolean} True if client is in lobby.
            */
            LoadBalancingClient.prototype.isInLobby = function () {
                return this.state == LoadBalancingClient.State.JoinedLobby;
            };
            /**
                @summary Checks if client is joined to game.
                @method Photon.LoadBalancing.LoadBalancingClient#isJoinedToRoom
                @returns {boolean} True if client is joined to game.
            */
            LoadBalancingClient.prototype.isJoinedToRoom = function () {
                return this.state == LoadBalancingClient.State.Joined;
            };
            /**
                @deprecated Use isJoinedToRoom()
            */
            LoadBalancingClient.prototype.isConnectedToGame = function () {
                return this.isJoinedToRoom();
            };
            /**
                @summary Current room list from Master server.
                @method Photon.LoadBalancing.LoadBalancingClient#availableRooms
                @returns {{@link Photon.LoadBalancing.RoomInfo}[]} Current room list
            */
            LoadBalancingClient.prototype.availableRooms = function () { return this.roomInfos; };
            /**
                @summary Sets client logger level
                @method Photon.LoadBalancing.LoadBalancingClient#setLogLevel
                @param {Exitgames.Common.Logger.Level} level Logging level.
            */
            LoadBalancingClient.prototype.setLogLevel = function (level) {
                this.logger.setLevel(level);
                if (this.nameServerPeer) {
                    this.nameServerPeer.setLogLevel(level);
                }
                if (this.masterPeer) {
                    this.masterPeer.setLogLevel(level);
                }
                if (this.gamePeer) {
                    this.gamePeer.setLogLevel(level);
                }
            };
            LoadBalancingClient.prototype.addRoom = function (r) { this.roomInfos.push(r); this.roomInfosDict[r.name] = r; };
            LoadBalancingClient.prototype.clearRooms = function () { this.roomInfos = new Array(); this.roomInfosDict = {}; };
            LoadBalancingClient.prototype.purgeRemovedRooms = function () {
                this.roomInfos = this.roomInfos.filter(function (x) { return !x.removed; });
                for (var n in this.roomInfosDict) {
                    if (this.roomInfosDict[n].removed) {
                        delete this.roomInfosDict[n];
                    }
                }
            };
            LoadBalancingClient.prototype.addActor = function (a) {
                this.actors[a.actorNr] = a;
                this.actorsArray.push(a);
                this.currentRoom.playerCount = this.actorsArray.length;
                if (this.lowestActorId == 0 || this.lowestActorId > a.actorNr)
                    this.lowestActorId = a.actorNr;
            };
            LoadBalancingClient.prototype.removeActor = function (actorNr) {
                delete this.actors[actorNr];
                this.actorsArray = this.actorsArray.filter(function (x) { return x.actorNr != actorNr; });
                this.currentRoom.playerCount = this.actorsArray.length;
                if (this.lowestActorId == actorNr) {
                    if (this.actorsArray.length > 0)
                        this.lowestActorId = this.actorsArray.reduce(function (prev, curr) { return prev.actorNr < curr.actorNr ? prev : curr; }).actorNr;
                    else
                        this.lowestActorId = 0;
                }
            };
            LoadBalancingClient.prototype.clearActors = function () {
                this.actors = {};
                this.actorsArray = [];
                this.currentRoom.playerCount = 0;
                this.lowestActorId = 0;
            };
            LoadBalancingClient.prototype.changeState = function (nextState) {
                this.logger.info("State:", LoadBalancingClient.StateToName(this.state), "->", LoadBalancingClient.StateToName(nextState));
                this.state = nextState;
                this.onStateChange(nextState);
            };
            LoadBalancingClient.prototype.createRoomInternal = function (peer, options) {
                var op = [];
                if (this.currentRoom.name)
                    op.push(LoadBalancing.Constants.ParameterCode.RoomName, this.currentRoom.name);
                this.fillCreateRoomOptions(op, options);
                if (peer === this.masterPeer) {
                    this.createRoomOptions = options;
                }
                if (peer === this.gamePeer) {
                    op.push(LoadBalancing.Constants.ParameterCode.PlayerProperties);
                    op.push(this._myActor._getAllProperties());
                }
                var log = peer == this.gamePeer ? this.gamePeer._logger : this.masterPeer._logger;
                log.info("Create Room", options && options.lobbyName, options && options.lobbyType, "...");
                peer.sendOperation(LoadBalancing.Constants.OperationCode.CreateGame, op);
            };
            LoadBalancingClient.prototype.updateUserIdAndNickname = function (vals, logger) {
                var userId = vals[LoadBalancing.Constants.ParameterCode.UserId];
                if (userId != undefined) {
                    this.setUserId(userId);
                    logger.info("Setting userId sent by server:", userId);
                }
                var nickname = vals[LoadBalancing.Constants.ParameterCode.Nickname];
                if (nickname != undefined) {
                    this.myActor().setName(nickname);
                    logger.info("Setting nickname sent by server:", nickname);
                }
            };
            LoadBalancingClient.prototype.initNameServerPeer = function (np) {
                var _this = this;
                np.setLogLevel(this.logger.getLevel());
                // errors
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.error, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.NameServerError, "NameServer peer error");
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectFailed, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.NameServerConnectFailed, "NameServer peer connect failed. " + _this.nameServerAddress);
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.timeout, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.NameServerTimeout, "NameServer peer timeout");
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connecting, function () {
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connect, function () {
                    np._logger.info("Connected");
                    _this.changeState(LoadBalancingClient.State.ConnectedToNameServer);
                    // connectToRegionMaster inited connection
                    if (_this.connectOptions.region != undefined) {
                        np.opAuth(_this.appId, _this.appVersion, _this.userAuthType, _this.userAuthParameters, _this.userAuthData, _this.userId, _this.connectOptions.region);
                    }
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.disconnect, function () {
                    if (np == _this.nameServerPeer) {
                        _this._cleanupNameServerPeerData();
                        np._logger.info("Disconnected");
                    }
                });
                np.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectClosed, function () {
                    np._logger.info("Server closed connection");
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.NameServerConnectClosed, "NameServer server closed connection");
                });
                // events
                // responses - check operation result. data.errCode
                np.addResponseListener(LoadBalancing.Constants.OperationCode.GetRegions, function (data) {
                    np._logger.debug("resp GetRegions", data);
                    var regions = {};
                    if (data.errCode == 0) {
                        var r = data.vals[LoadBalancing.Constants.ParameterCode.Region];
                        var a = data.vals[LoadBalancing.Constants.ParameterCode.Address];
                        for (var i in r) {
                            regions[r[i]] = a[i];
                        }
                    }
                    else {
                        np._logger.error("GetRegions request error.", data.errCode);
                    }
                    _this.onGetRegionsResult(data.errCode, data.errMsg, regions);
                });
                np.addResponseListener(LoadBalancing.Constants.OperationCode.Authenticate, function (data) {
                    np._logger.debug("resp Authenticate", data);
                    if (data.errCode == 0) {
                        np._logger.info("Authenticated");
                        np.disconnect();
                        _this.updateUserIdAndNickname(data.vals, np._logger);
                        _this.masterServerAddress = data.vals[LoadBalancing.Constants.ParameterCode.Address];
                        np._logger.info("Connecting to Master server", _this.masterServerAddress, "...");
                        _this.connectOptions.userAuthSecret = data.vals[LoadBalancing.Constants.ParameterCode.Secret];
                        _this.connect(_this.connectOptions);
                    }
                    else {
                        _this.changeState(LoadBalancingClient.State.Error);
                        _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.NameServerAuthenticationFailed, "NameServer authentication failed: " + data.errCode + " " + data.errMsg);
                    }
                });
            };
            // protected
            LoadBalancingClient.prototype.initMasterPeer = function (mp) {
                var _this = this;
                mp.setLogLevel(this.logger.getLevel());
                // errors
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.error, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterError, "Master peer error");
                });
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectFailed, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterConnectFailed, "Master peer connect failed: " + _this.masterServerAddress);
                });
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.timeout, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterTimeout, "Master peer error timeout");
                });
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connecting, function () {
                });
                // status
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connect, function () {
                    //TODO: encryption phase
                    mp._logger.info("Connected");
                    var op = [];
                    // if NameSever gave us secret
                    if (_this.connectOptions.userAuthSecret) {
                        op.push(LoadBalancing.Constants.ParameterCode.Secret, _this.connectOptions.userAuthSecret);
                        mp.sendOperation(LoadBalancing.Constants.OperationCode.Authenticate, op);
                        mp._logger.info("Authenticate with secret...");
                    }
                    else {
                        op.push(LoadBalancing.Constants.ParameterCode.ApplicationId);
                        op.push(_this.appId);
                        op.push(LoadBalancing.Constants.ParameterCode.AppVersion);
                        op.push(_this.appVersion);
                        if (_this.userAuthType != LoadBalancing.Constants.CustomAuthenticationType.None) {
                            op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationType, Photon.TypeExt.Byte(_this.userAuthType));
                            op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationParams, _this.userAuthParameters);
                            if (_this.userAuthData) {
                                op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationData, _this.userAuthData);
                            }
                        }
                        if (_this.userId) {
                            op.push(LoadBalancing.Constants.ParameterCode.UserId, _this.userId);
                        }
                        if (_this.connectOptions.lobbyStats) {
                            op.push(LoadBalancing.Constants.ParameterCode.LobbyStats, true);
                        }
                        mp.sendOperation(LoadBalancing.Constants.OperationCode.Authenticate, op);
                        mp._logger.info("Authenticate...");
                    }
                });
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.disconnect, function () {
                    if (mp == _this.masterPeer) {
                        _this._cleanupMasterPeerData();
                        mp._logger.info("Disconnected");
                    }
                });
                mp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectClosed, function () {
                    mp._logger.info("Server closed connection");
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterConnectClosed, "Master server closed connection");
                });
                //events
                mp.addEventListener(LoadBalancing.Constants.EventCode.GameList, function (data) {
                    var gameList = data.vals[LoadBalancing.Constants.ParameterCode.GameList];
                    _this.clearRooms();
                    for (var g in gameList) {
                        var r = new RoomInfo(g);
                        r._updateFromProps(gameList[g]);
                        _this.addRoom(r);
                    }
                    _this.onRoomList(_this.roomInfos);
                    mp._logger.debug("ev GameList", _this.roomInfos, gameList);
                });
                mp.addEventListener(LoadBalancing.Constants.EventCode.GameListUpdate, function (data) {
                    var gameList = data.vals[LoadBalancing.Constants.ParameterCode.GameList];
                    var roomsUpdated = new Array();
                    var roomsAdded = new Array();
                    var roomsRemoved = new Array();
                    for (var g in gameList) {
                        var exist = _this.roomInfos.filter(function (x) { return x.name == g; });
                        if (exist.length > 0) {
                            var r = exist[0];
                            r._updateFromProps(gameList[g]);
                            if (r.removed) {
                                roomsRemoved.push(r);
                            }
                            else {
                                roomsUpdated.push(r);
                            }
                        }
                        else {
                            var ri = new RoomInfo(g);
                            ri._updateFromProps(gameList[g]);
                            _this.addRoom(ri);
                            roomsAdded.push(ri);
                        }
                    }
                    _this.purgeRemovedRooms();
                    _this.onRoomListUpdate(_this.roomInfos, roomsUpdated, roomsAdded, roomsRemoved);
                    mp._logger.debug("ev GameListUpdate:", _this.roomInfos, "u:", roomsUpdated, "a:", roomsAdded, "r:", roomsRemoved, gameList);
                });
                // responses - check operation result: data.errCode
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.Authenticate, function (data) {
                    mp._logger.debug("resp Authenticate", data);
                    if (!data.errCode) {
                        mp._logger.info("Authenticated");
                        _this.updateUserIdAndNickname(data.vals, mp._logger);
                        if (data.vals[LoadBalancing.Constants.ParameterCode.Secret] != undefined) {
                            _this.connectOptions.userAuthSecret = data.vals[LoadBalancing.Constants.ParameterCode.Secret];
                        }
                        _this.changeState(LoadBalancingClient.State.ConnectedToMaster);
                        var op = [];
                        if (_this.connectOptions.lobbyName) {
                            op.push(LoadBalancing.Constants.ParameterCode.LobbyName);
                            op.push(_this.connectOptions.lobbyName);
                            if (_this.connectOptions.lobbyType != undefined) {
                                op.push(LoadBalancing.Constants.ParameterCode.LobbyType);
                                op.push(_this.connectOptions.lobbyType);
                            }
                        }
                        if (_this.autoJoinLobby) {
                            mp.sendOperation(LoadBalancing.Constants.OperationCode.JoinLobby, op);
                            mp._logger.info("Join Lobby", _this.connectOptions.lobbyName, _this.connectOptions.lobbyType, "...");
                        }
                    }
                    else {
                        _this.changeState(LoadBalancingClient.State.Error);
                        _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterAuthenticationFailed, "Master authentication failed: " + data.errCode + " " + data.errMsg);
                    }
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.JoinLobby, function (data) {
                    mp._logger.debug("resp JoinLobby", data);
                    if (!data.errCode) {
                        mp._logger.info("Joined to Lobby");
                        _this.changeState(LoadBalancingClient.State.JoinedLobby);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.JoinLobby, data);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.CreateGame, function (data) {
                    mp._logger.debug("resp CreateGame", data);
                    if (!data.errCode) {
                        _this.currentRoom._updateFromMasterResponse(data.vals);
                        mp._logger.debug("Created/Joined " + _this.currentRoom.name);
                        _this.connectToGameServer(LoadBalancing.Constants.OperationCode.CreateGame);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.CreateGame, data);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.JoinGame, function (data) {
                    mp._logger.debug("resp JoinGame", data);
                    if (!data.errCode) {
                        _this.currentRoom._updateFromMasterResponse(data.vals);
                        mp._logger.debug("Joined " + _this.currentRoom.name);
                        _this.connectToGameServer(LoadBalancing.Constants.OperationCode.JoinGame);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.JoinGame, data);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.JoinRandomGame, function (data) {
                    mp._logger.debug("resp JoinRandomGame", data);
                    if (!data.errCode) {
                        _this.currentRoom._updateFromMasterResponse(data.vals);
                        mp._logger.debug("Joined " + _this.currentRoom.name);
                        _this.connectToGameServer(LoadBalancing.Constants.OperationCode.JoinRandomGame);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.JoinRandomGame, data);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.FindFriends, function (data) {
                    mp._logger.debug("resp FindFriends", data);
                    var res = {};
                    if (!data.errCode) {
                        var onlines = data.vals[LoadBalancing.Constants.ParameterCode.FindFriendsResponseOnlineList] || {};
                        var roomIds = data.vals[LoadBalancing.Constants.ParameterCode.FindFriendsResponseRoomIdList] || {};
                        for (var i = 0; i < _this.findFriendsRequestList.length; ++i) {
                            var name = _this.findFriendsRequestList[i];
                            if (name) {
                                res[name] = { online: onlines[i], roomId: roomIds[i] };
                            }
                        }
                    }
                    else {
                        mp._logger.error("FindFriends request error:", data.errCode);
                    }
                    _this.onFindFriendsResult(data.errCode, data.errMsg, res);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.LobbyStats, function (data) {
                    mp._logger.debug("resp LobbyStats", data);
                    var res = new Array();
                    if (!data.errCode) {
                        var names = data.vals[LoadBalancing.Constants.ParameterCode.LobbyName]; // not inited intentionally
                        var types = data.vals[LoadBalancing.Constants.ParameterCode.LobbyType] || {};
                        var peers = data.vals[LoadBalancing.Constants.ParameterCode.PeerCount] || {};
                        var games = data.vals[LoadBalancing.Constants.ParameterCode.GameCount] || {};
                        if (names) {
                            for (var i = 0; i < names.length; ++i) {
                                res[i] = { lobbyName: names[i], lobbyType: types[i], peerCount: peers[i], gameCount: games[i] };
                            }
                        }
                        else {
                            for (var i = 0; i < _this.lobbyStatsRequestList.length; ++i) {
                                var l = _this.lobbyStatsRequestList[i];
                                res[i] = { lobbyName: l[0], lobbyType: l[1], peerCount: peers[i], gameCount: games[i] };
                            }
                        }
                    }
                    else {
                        mp._logger.error("LobbyStats request error:", data.errCode);
                    }
                    _this.onLobbyStats(data.errCode, data.errMsg, res);
                });
                mp.addEventListener(LoadBalancing.Constants.EventCode.LobbyStats, function (data) {
                    mp._logger.debug("ev LobbyStats", data);
                    var res = new Array();
                    var names = data.vals[LoadBalancing.Constants.ParameterCode.LobbyName]; // not inited intentionally
                    var types = data.vals[LoadBalancing.Constants.ParameterCode.LobbyType] || {};
                    var peers = data.vals[LoadBalancing.Constants.ParameterCode.PeerCount] || {};
                    var games = data.vals[LoadBalancing.Constants.ParameterCode.GameCount] || {};
                    if (names) {
                        for (var i = 0; i < names.length; ++i) {
                            res[i] = { lobbyName: names[i], lobbyType: types[i], peerCount: peers[i], gameCount: games[i] };
                        }
                    }
                    _this.onLobbyStats(0, "", res);
                });
                mp.addEventListener(LoadBalancing.Constants.EventCode.AppStats, function (data) {
                    mp._logger.debug("ev AppStats", data);
                    var res = {
                        peerCount: data.vals[LoadBalancing.Constants.ParameterCode.PeerCount],
                        masterPeerCount: data.vals[LoadBalancing.Constants.ParameterCode.MasterPeerCount],
                        gameCount: data.vals[LoadBalancing.Constants.ParameterCode.GameCount]
                    };
                    _this.onAppStats(0, "", res);
                });
                mp.addResponseListener(LoadBalancing.Constants.OperationCode.Rpc, mp.webRpcHandler(this));
            };
            LoadBalancingClient.prototype.connectToGameServer = function (masterOpCode) {
                if (!this.connectOptions.keepMasterConnection) {
                    this.masterPeer.disconnect();
                }
                if (this.checkNextState(LoadBalancingClient.State.ConnectingToGameserver, true)) {
                    this.logger.info("Connecting to Game", this.currentRoom.address);
                    if (this.gamePeer)
                        this.gamePeer.Destroy();
                    this.gamePeer = new GamePeer(this, this.connectionProtocol, this.currentRoom.address, "");
                    this.gamePeerWaitingForDisconnect = false;
                    this.initGamePeer(this.gamePeer, masterOpCode);
                    this.gamePeer.connect(this.appId);
                    this.changeState(LoadBalancingClient.State.ConnectingToGameserver);
                    return true;
                }
                else {
                    return false;
                }
            };
            LoadBalancingClient.prototype.initGamePeer = function (gp, masterOpCode) {
                var _this = this;
                gp.setLogLevel(this.logger.getLevel());
                // errors
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.error, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.GameError, "Game peer error");
                });
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectFailed, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.GameConnectFailed, "Game peer connect failed: " + _this.currentRoom.address);
                });
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.timeout, function () {
                    _this.changeState(LoadBalancingClient.State.Error);
                    _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.GameTimeout, "Game peer timeout");
                });
                // status
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connect, function () {
                    gp._logger.info("Connected");
                    //TODO: encryption phase
                    var op = [];
                    op.push(LoadBalancing.Constants.ParameterCode.ApplicationId);
                    op.push(_this.appId);
                    op.push(LoadBalancing.Constants.ParameterCode.AppVersion);
                    op.push(_this.appVersion);
                    if (_this.connectOptions.userAuthSecret != undefined) {
                        op.push(LoadBalancing.Constants.ParameterCode.Secret);
                        op.push(_this.connectOptions.userAuthSecret);
                    }
                    if (_this.userAuthType != LoadBalancing.Constants.CustomAuthenticationType.None) {
                        op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationType);
                        op.push(Photon.TypeExt.Byte(_this.userAuthType));
                    }
                    if (_this.userId) {
                        op.push(LoadBalancing.Constants.ParameterCode.UserId, _this.userId);
                    }
                    gp.sendOperation(LoadBalancing.Constants.OperationCode.Authenticate, op);
                    gp._logger.info("Authenticate...");
                });
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.disconnect, function () {
                    if (gp == _this.gamePeer) {
                        _this._cleanupGamePeerData();
                        gp._logger.info("Disconnected");
                    }
                });
                gp.addPeerStatusListener(Photon.PhotonPeer.StatusCodes.connectClosed, function () {
                    gp._logger.info("Server closed connection");
                    if (!_this.gamePeerWaitingForDisconnect) {
                        _this.changeState(LoadBalancingClient.State.Error);
                        _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.MasterConnectClosed, "Game server closed connection");
                    }
                });
                // responses
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.Authenticate, function (data) {
                    gp._logger.debug("resp Authenticate", data);
                    if (!data.errCode) {
                        gp._logger.info("Authenticated");
                        gp._logger.info("Connected");
                        if (masterOpCode == LoadBalancing.Constants.OperationCode.CreateGame) {
                            _this.createRoomInternal(gp, _this.createRoomOptions);
                        }
                        else {
                            var op = [];
                            op.push(LoadBalancing.Constants.ParameterCode.RoomName);
                            op.push(_this.currentRoom.name);
                            op.push(LoadBalancing.Constants.ParameterCode.Broadcast);
                            op.push(true);
                            op.push(LoadBalancing.Constants.ParameterCode.PlayerProperties);
                            op.push(_this._myActor._getAllProperties());
                            if (masterOpCode == LoadBalancing.Constants.OperationCode.JoinGame) {
                                if (_this.joinRoomOptions.createIfNotExists) {
                                    op.push(LoadBalancing.Constants.ParameterCode.JoinMode, LoadBalancingClient.JoinMode.CreateIfNotExists);
                                    _this.fillCreateRoomOptions(op, _this.createRoomOptions);
                                }
                                if (_this.joinRoomOptions.rejoin) {
                                    op.push(LoadBalancing.Constants.ParameterCode.JoinMode, LoadBalancingClient.JoinMode.RejoinOnly);
                                }
                            }
                            gp.sendOperation(LoadBalancing.Constants.OperationCode.JoinGame, op);
                        }
                        _this.changeState(LoadBalancingClient.State.ConnectedToGameserver);
                    }
                    else {
                        _this.changeState(LoadBalancingClient.State.Error);
                        _this._onErrorInternal(LoadBalancingClient.PeerErrorCode.GameAuthenticationFailed, "Game authentication failed: " + data.errCode + " " + data.errMsg);
                    }
                });
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.CreateGame, function (data) {
                    gp._logger.debug("resp CreateGame", data);
                    if (!data.errCode) {
                        _this._myActor._updateMyActorFromResponse(data.vals);
                        gp._logger.info("myActor: ", _this._myActor);
                        _this.currentRoom._updateFromProps(data.vals[LoadBalancing.Constants.ParameterCode.GameProperties]);
                        _this.clearActors();
                        _this.addActor(_this._myActor);
                        _this.changeState(LoadBalancingClient.State.Joined);
                        _this.onJoinRoom(true);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.CreateGame, data);
                });
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.JoinGame, function (data) {
                    gp._logger.debug("resp JoinGame", data);
                    if (!data.errCode) {
                        _this._myActor._updateMyActorFromResponse(data.vals);
                        gp._logger.info("myActor: ", _this._myActor);
                        _this.clearActors();
                        _this.addActor(_this._myActor);
                        var actorList = data.vals[LoadBalancing.Constants.ParameterCode.ActorList];
                        var actorProps = data.vals[LoadBalancing.Constants.ParameterCode.PlayerProperties];
                        if (actorList !== undefined) {
                            for (var i = 0; i < actorList.length; i++) {
                                var actorNr = actorList[i];
                                var props;
                                if (actorProps !== undefined)
                                    props = actorProps[actorNr];
                                var name;
                                if (props !== undefined) {
                                    name = props[LoadBalancing.Constants.ActorProperties.PlayerName];
                                }
                                var a;
                                if (actorNr == _this._myActor.actorNr)
                                    a = _this._myActor;
                                else {
                                    a = _this.actorFactoryInternal(name, actorNr);
                                    _this.addActor(a);
                                }
                                if (props !== undefined) {
                                    a._updateCustomProperties(props);
                                }
                            }
                        }
                        _this.currentRoom._updateFromProps(data.vals[LoadBalancing.Constants.ParameterCode.GameProperties]);
                        _this.changeState(LoadBalancingClient.State.Joined);
                        _this.onJoinRoom(false);
                    }
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.JoinGame, data);
                });
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.SetProperties, function (data) {
                    gp._logger.debug("resp SetProperties", data);
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.SetProperties, data);
                });
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.Leave, function (data) {
                    gp._logger.debug("resp Leave", data);
                    gp.disconnect();
                    _this._onOperationResponseInternal2(LoadBalancing.Constants.OperationCode.Leave, data);
                });
                gp.addResponseListener(LoadBalancing.Constants.OperationCode.Rpc, gp.webRpcHandler(this));
                // events
                gp.addEventListener(LoadBalancing.Constants.EventCode.Join, function (data) {
                    gp._logger.debug("ev Join", data);
                    if (Actor._getActorNrFromResponse(data.vals) === _this._myActor.actorNr) {
                        //this._myActor._updateMyActorFromResponse(data.vals);
                        _this._myActor._updateFromResponse(data.vals);
                        //                    this.addActor(this._myActor);
                        _this.onActorJoin(_this._myActor); // let client read updated properties
                    }
                    else {
                        var actor = _this.actorFactoryInternal();
                        actor._updateFromResponse(data.vals);
                        _this.addActor(actor);
                        _this.onActorJoin(actor);
                    }
                });
                gp.addEventListener(LoadBalancing.Constants.EventCode.Leave, function (data) {
                    gp._logger.debug("ev Leave", data);
                    _this.myRoom()._updateFromEvent(data.vals); // updating masterClientId
                    var actorNr = Actor._getActorNrFromResponse(data.vals);
                    if (actorNr && _this.actors[actorNr]) {
                        var a = _this.actors[actorNr];
                        if (data.vals[LoadBalancing.Constants.ParameterCode.IsInactive]) {
                            a._setSuspended(true);
                            _this.onActorSuspend(a);
                        }
                        else {
                            _this.removeActor(actorNr);
                            _this.onActorLeave(a, false);
                        }
                    }
                });
                gp.addEventListener(LoadBalancing.Constants.EventCode.Disconnect, function (data) {
                    gp._logger.debug("ev Disconnect", data);
                    var actorNr = Actor._getActorNrFromResponse(data.vals);
                    if (actorNr && _this.actors[actorNr]) {
                        var a = _this.actors[actorNr];
                        a._setSuspended(true);
                        _this.onActorSuspend(a);
                    }
                });
                gp.addEventListener(LoadBalancing.Constants.EventCode.PropertiesChanged, function (data) {
                    gp._logger.debug("ev PropertiesChanged", data);
                    var targetActorNr = data.vals[LoadBalancing.Constants.ParameterCode.TargetActorNr];
                    if (targetActorNr !== undefined && targetActorNr > 0) {
                        if (_this.actors[targetActorNr] !== undefined) {
                            var actor = _this.actors[targetActorNr];
                            actor._updateCustomProperties(data.vals[LoadBalancing.Constants.ParameterCode.Properties]);
                            _this.onActorPropertiesChange(actor);
                        }
                    }
                    else {
                        _this.currentRoom._updateFromProps(data.vals[LoadBalancing.Constants.ParameterCode.Properties]);
                        _this.onMyRoomPropertiesChange();
                    }
                });
            };
            LoadBalancingClient.prototype._cleanupNameServerPeerData = function () {
            };
            LoadBalancingClient.prototype._cleanupMasterPeerData = function () {
            };
            LoadBalancingClient.prototype._cleanupGamePeerData = function () {
                for (var i in this.actors) {
                    this.onActorLeave(this.actors[i], true);
                }
                this.clearActors();
                this.addActor(this._myActor);
            };
            LoadBalancingClient.prototype._onOperationResponseInternal2 = function (code, data) {
                if (data.errCode) {
                    this.logger.warn("Operation", code, "error:", data.errMsg, "(" + data.errCode + ")");
                }
                this.onOperationResponse(data.errCode, data.errMsg, code, data.vals);
            };
            LoadBalancingClient.prototype._onErrorInternal = function (errorCode, errorMsg) {
                this.logger.error("Error:", errorCode, errorMsg);
                this.onError(errorCode, errorMsg);
            };
            //TODO: ugly way to init const table
            LoadBalancingClient.prototype.initValidNextState = function () {
                this.validNextState[LoadBalancingClient.State.Error] = [LoadBalancingClient.State.ConnectingToMasterserver, LoadBalancingClient.State.ConnectingToNameServer];
                this.validNextState[LoadBalancingClient.State.Uninitialized] = [LoadBalancingClient.State.ConnectingToMasterserver, LoadBalancingClient.State.ConnectingToNameServer];
                this.validNextState[LoadBalancingClient.State.ConnectedToNameServer] = [LoadBalancingClient.State.ConnectingToMasterserver];
                this.validNextState[LoadBalancingClient.State.Disconnected] = [LoadBalancingClient.State.ConnectingToMasterserver, LoadBalancingClient.State.ConnectingToNameServer];
                this.validNextState[LoadBalancingClient.State.ConnectedToMaster] = [LoadBalancingClient.State.JoinedLobby];
                this.validNextState[LoadBalancingClient.State.JoinedLobby] = [LoadBalancingClient.State.ConnectingToGameserver];
                this.validNextState[LoadBalancingClient.State.ConnectingToGameserver] = [LoadBalancingClient.State.ConnectedToGameserver];
                this.validNextState[LoadBalancingClient.State.ConnectedToGameserver] = [LoadBalancingClient.State.Joined];
            };
            LoadBalancingClient.prototype.checkNextState = function (nextState, dontThrow) {
                if (dontThrow === void 0) { dontThrow = false; }
                var valid = this.validNextState[this.state];
                var res = valid && valid.indexOf(nextState) >= 0;
                if (!res) {
                    if (dontThrow) {
                        this.logger.error("LoadBalancingPeer checkNextState fail: " + LoadBalancingClient.StateToName(this.state) + " -> " + LoadBalancingClient.StateToName(nextState));
                    }
                    else {
                        this.logger.exception(501, "LoadBalancingPeer checkNextState fail: " + LoadBalancingClient.StateToName(this.state) + " -> " + LoadBalancingClient.StateToName(nextState));
                    }
                }
                return res;
            };
            /**
                @summary Converts {@link Photon.LoadBalancing.LoadBalancingClient.State State} element to string name.
                @method Photon.LoadBalancing.LoadBalancingClient.StateToName
                @param {Photon.LoadBalancing.LoadBalancingClient.State} state Client state enum element.
                @returns {string} Specified element name or undefined if not found.
            */
            LoadBalancingClient.StateToName = function (value) { return Exitgames.Common.Util.getEnumKeyByValue(LoadBalancingClient.State, value); };
            LoadBalancingClient.JoinMode = {
                Default: 0,
                CreateIfNotExists: 1,
                //            JoinOrejoin: 2,
                RejoinOnly: 3
            };
            // tsc looses all comments after first static member 
            // jsdoc reads comments from any place within class (and may be from any place in file)
            LoadBalancingClient.PeerErrorCode = {
                /**
                    @summary Enum for client peers error codes.
                    @member Photon.LoadBalancing.LoadBalancingClient.PeerErrorCode
                    @readonly
                    @property {number} Ok No Error.
                    @property {number} MasterError General Master server peer error.
                    @property {number} MasterConnectFailed Master server connection error.
                    @property {number} MasterConnectClosed Disconnected from Master server.
                    @property {number} MasterTimeout Disconnected from Master server for timeout.
                    @property {number} MasterEncryptionEstablishError Master server encryption establishing failed.
                    @property {number} MasterAuthenticationFailed Master server authentication failed.
                    @property {number} GameError General Game server peer error.
                    @property {number} GameConnectFailed Game server connection error.
                    @property {number} GameConnectClosed Disconnected from Game server.
                    @property {number} GameTimeout Disconnected from Game server for timeout.
                    @property {number} GameEncryptionEstablishError Game server encryption establishing failed.
                    @property {number} GameAuthenticationFailed Game server authentication failed.
                    @property {number} NameServerError General NameServer peer error.
                    @property {number} NameServerConnectFailed NameServer connection error.
                    @property {number} NameServerConnectClosed Disconnected from NameServer.
                    @property {number} NameServerTimeout Disconnected from NameServer for timeout.
                    @property {number} NameServerEncryptionEstablishError NameServer encryption establishing failed.
                    @property {number} NameServerAuthenticationFailed NameServer authentication failed.
                 */
                Ok: 0,
                MasterError: 1001,
                MasterConnectFailed: 1002,
                MasterConnectClosed: 1003,
                MasterTimeout: 1004,
                MasterEncryptionEstablishError: 1005,
                MasterAuthenticationFailed: 1101,
                GameError: 2001,
                GameConnectFailed: 2002,
                GameConnectClosed: 2003,
                GameTimeout: 2004,
                GameEncryptionEstablishError: 2005,
                GameAuthenticationFailed: 2101,
                NameServerError: 3001,
                NameServerConnectFailed: 3002,
                NameServerConnectClosed: 3003,
                NameServerTimeout: 3004,
                NameServerEncryptionEstablishError: 3005,
                NameServerAuthenticationFailed: 3101
            };
            LoadBalancingClient.State = {
                /**
                    @summary Enum for client states.
                    @member Photon.LoadBalancing.LoadBalancingClient.State
                    @readonly
                    @property {number} Error Critical error occurred.
                    @property {number} Uninitialized Client is created but not used yet.
                    @property {number} ConnectingToNameServer Connecting to NameServer.
                    @property {number} ConnectedToNameServer Connected to NameServer.
                    @property {number} ConnectingToMasterserver Connecting to Master (includes connect, authenticate and joining the lobby).
                    @property {number} ConnectedToMaster Connected to Master server.
                    @property {number} JoinedLobby Connected to Master and joined lobby. Display room list and join/create rooms at will.
                    @property {number} ConnectingToGameserver Connecting to Game server(client will authenticate and join/create game).
                    @property {number} ConnectedToGameserver Connected to Game server (going to auth and join game).
                    @property {number} Joined The client joined room.
                    @property {number} Disconnected The client is no longer connected (to any server). Connect to Master to go on.
                */
                Error: -1,
                Uninitialized: 0,
                ConnectingToNameServer: 1,
                ConnectedToNameServer: 2,
                ConnectingToMasterserver: 3,
                ConnectedToMaster: 4,
                JoinedLobby: 5,
                ConnectingToGameserver: 6,
                ConnectedToGameserver: 7,
                Joined: 8,
                Disconnected: 10
            };
            return LoadBalancingClient;
        }());
        LoadBalancing.LoadBalancingClient = LoadBalancingClient;
        //TODO: internal
        var LbcPeer = /** @class */ (function (_super) {
            __extends(LbcPeer, _super);
            function LbcPeer() {
                return _super !== null && _super.apply(this, arguments) || this;
            }
            LbcPeer.prototype.webRpc = function (uriPath, parameters, options) {
                var params = [];
                params.push(LoadBalancing.Constants.ParameterCode.UriPath, uriPath);
                params.push(LoadBalancing.Constants.ParameterCode.RpcCallParams, parameters);
                if (options) {
                    if (options.sendAuthCookie) {
                        params.push(LoadBalancing.Constants.ParameterCode.WebFlags, Photon.TypeExt.Byte(WebFlags.SendAuthCookie));
                    }
                }
                this.sendOperation(LoadBalancing.Constants.OperationCode.Rpc, params);
            };
            LbcPeer.prototype.webRpcHandler = function (lbc) {
                var _this = this;
                return function (d) {
                    _this._logger.debug("resp Rpc", d);
                    var uriPath, message, data, resultCode;
                    if (d.errCode == 0) {
                        uriPath = d.vals[LoadBalancing.Constants.ParameterCode.UriPath];
                        data = d.vals[LoadBalancing.Constants.ParameterCode.RpcCallRetData];
                        resultCode = d.vals[LoadBalancing.Constants.ParameterCode.RpcCallRetCode];
                    }
                    else {
                        _this._logger.error("WebRpc request error:", d.errCode);
                    }
                    lbc.onWebRpcResult(d.errCode, d.errMsg, uriPath, resultCode, data);
                };
            };
            return LbcPeer;
        }(Photon.PhotonPeer));
        LoadBalancing.LbcPeer = LbcPeer;
        var NameServerPeer = /** @class */ (function (_super) {
            __extends(NameServerPeer, _super);
            function NameServerPeer(client, protocol, address, subprotocol) {
                var _this = _super.call(this, protocol, address, subprotocol, client.logger.getPrefix() + " NameServer") || this;
                _this.client = client;
                return _this;
            }
            // overrides
            NameServerPeer.prototype.onUnhandledEvent = function (code, args) {
                this.client.onEvent(code, args.vals[LoadBalancing.Constants.ParameterCode.CustomEventContent], args.vals[LoadBalancing.Constants.ParameterCode.ActorNr]);
            };
            NameServerPeer.prototype.onUnhandledResponse = function (code, args) {
                this.client.onOperationResponse(args.errCode, args.errMsg, code, args.vals);
            };
            NameServerPeer.prototype.getRegions = function (appId) {
                var params = [];
                params.push(LoadBalancing.Constants.ParameterCode.ApplicationId, appId);
                this.sendOperation(LoadBalancing.Constants.OperationCode.GetRegions, params, true, 0);
            };
            // this = LBC
            NameServerPeer.prototype.opAuth = function (appId, appVersion, userAuthType, userAuthParameters, userAuthData, userId, region) {
                var op = [];
                op.push(LoadBalancing.Constants.ParameterCode.ApplicationId, appId);
                op.push(LoadBalancing.Constants.ParameterCode.AppVersion, appVersion);
                if (userAuthType != LoadBalancing.Constants.CustomAuthenticationType.None) {
                    op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationType, Photon.TypeExt.Byte(userAuthType));
                    op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationParams, userAuthParameters);
                    if (userAuthData) {
                        op.push(LoadBalancing.Constants.ParameterCode.ClientAuthenticationData, userAuthData);
                    }
                }
                if (userId) {
                    op.push(LoadBalancing.Constants.ParameterCode.UserId, userId);
                }
                //    		if (this.connectOptions.lobbyStats) {
                //    			op.push(Constants.ParameterCode.LobbyStats, true);
                //    		}
                op.push(LoadBalancing.Constants.ParameterCode.Region, region);
                this.sendOperation(LoadBalancing.Constants.OperationCode.Authenticate, op, true, 0);
                this._logger.info("Authenticate...");
            };
            return NameServerPeer;
        }(LbcPeer));
        LoadBalancing.NameServerPeer = NameServerPeer;
        //TODO: internal
        var MasterPeer = /** @class */ (function (_super) {
            __extends(MasterPeer, _super);
            function MasterPeer(client, protocol, address, subprotocol) {
                var _this = _super.call(this, protocol, address, subprotocol, client.logger.getPrefix() + " Master") || this;
                _this.client = client;
                return _this;
            }
            // overrides
            MasterPeer.prototype.onUnhandledEvent = function (code, args) {
                this.client.onEvent(code, args.vals[LoadBalancing.Constants.ParameterCode.CustomEventContent], args.vals[LoadBalancing.Constants.ParameterCode.ActorNr]);
            };
            MasterPeer.prototype.onUnhandledResponse = function (code, args) {
                this.client.onOperationResponse(args.errCode, args.errMsg, code, args.vals);
            };
            MasterPeer.prototype.findFriends = function (friendsToFind) {
                var params = [];
                params.push(LoadBalancing.Constants.ParameterCode.FindFriendsRequestList);
                params.push(friendsToFind);
                this.sendOperation(LoadBalancing.Constants.OperationCode.FindFriends, params);
            };
            MasterPeer.prototype.requestLobbyStats = function (lobbiesToRequest) {
                var params = [];
                if (lobbiesToRequest && lobbiesToRequest.length > 0) {
                    var n = new Array();
                    var t = new Array();
                    for (var i = 0; i < lobbiesToRequest.length; ++i) {
                        n[i] = lobbiesToRequest[i][0];
                        t[i] = lobbiesToRequest[i][1];
                    }
                    params.push(LoadBalancing.Constants.ParameterCode.LobbyName);
                    params.push(n);
                    params.push(LoadBalancing.Constants.ParameterCode.LobbyType);
                    params.push(t);
                }
                this.sendOperation(LoadBalancing.Constants.OperationCode.LobbyStats, params);
            };
            return MasterPeer;
        }(LbcPeer));
        LoadBalancing.MasterPeer = MasterPeer;
        //TODO: internal
        var GamePeer = /** @class */ (function (_super) {
            __extends(GamePeer, _super);
            function GamePeer(client, protocol, address, subprotocol) {
                var _this = _super.call(this, protocol, address, subprotocol, client.logger.getPrefix() + " Game") || this;
                _this.client = client;
                return _this;
            }
            // overrides
            GamePeer.prototype.onUnhandledEvent = function (code, args) {
                this.client.onEvent(code, args.vals[LoadBalancing.Constants.ParameterCode.CustomEventContent], args.vals[LoadBalancing.Constants.ParameterCode.ActorNr]);
            };
            // overrides
            GamePeer.prototype.onUnhandledResponse = function (code, args) {
                this.client.onOperationResponse(args.errCode, args.errMsg, code, args.vals);
            };
            GamePeer.prototype.raiseEvent = function (eventCode, data, options) {
                if (this.client.isJoinedToRoom()) {
                    this._logger.debug("raiseEvent", eventCode, data, options);
                    var params = [LoadBalancing.Constants.ParameterCode.Code, Photon.TypeExt.Byte(eventCode), LoadBalancing.Constants.ParameterCode.Data, data];
                    if (options) {
                        if (options.receivers != undefined && options.receivers !== LoadBalancing.Constants.ReceiverGroup.Others) {
                            params.push(LoadBalancing.Constants.ParameterCode.ReceiverGroup);
                            params.push(Photon.TypeExt.Byte(options.receivers));
                        }
                        if (options.cache != undefined && options.cache !== LoadBalancing.Constants.EventCaching.DoNotCache) {
                            params.push(LoadBalancing.Constants.ParameterCode.Cache);
                            params.push(Photon.TypeExt.Byte(options.cache));
                        }
                        if (options.interestGroup != undefined) {
                            if (this.checkGroupNumber(options.interestGroup)) {
                                params.push(LoadBalancing.Constants.ParameterCode.Group);
                                params.push(Photon.TypeExt.Byte(options.interestGroup));
                            }
                            else {
                                this._logger.exception(502, "raiseEvent - Group not a number: " + options.interestGroup);
                            }
                        }
                        if (options.targetActors != undefined) {
                            params.push(LoadBalancing.Constants.ParameterCode.ActorList);
                            params.push(options.targetActors);
                        }
                        if (options.webForward) {
                            params.push(LoadBalancing.Constants.ParameterCode.WebFlags);
                            params.push(Photon.TypeExt.Byte(WebFlags.HttpForward));
                        }
                    }
                    this.sendOperation(LoadBalancing.Constants.OperationCode.RaiseEvent, params);
                }
                else {
                    throw new Error("raiseEvent - Not joined!");
                }
            };
            GamePeer.prototype.changeGroups = function (groupsToRemove, groupsToAdd) {
                var params = [];
                if (groupsToRemove != null && groupsToRemove != undefined) {
                    this.checkGroupArray(groupsToRemove, "groupsToRemove");
                    params.push(LoadBalancing.Constants.ParameterCode.Remove);
                    params.push(Photon.TypeExt.Byte(groupsToRemove));
                }
                if (groupsToAdd != null && groupsToAdd != undefined) {
                    this.checkGroupArray(groupsToAdd, "groupsToAdd");
                    params.push(LoadBalancing.Constants.ParameterCode.Add);
                    params.push(Photon.TypeExt.Byte(groupsToAdd));
                }
                this.sendOperation(LoadBalancing.Constants.OperationCode.ChangeGroups, params);
            };
            GamePeer.prototype.checkGroupNumber = function (g) {
                return !(typeof (g) != "number" || isNaN(g) || g === Infinity || g === -Infinity);
            };
            GamePeer.prototype.checkGroupArray = function (groups, groupsName) {
                if (Exitgames.Common.Util.isArray(groups)) {
                    for (var i = 0; i < groups.length; ++i) {
                        var g = groups[i];
                        if (this.checkGroupNumber(g)) {
                        }
                        else {
                            this._logger.exception(503, "changeGroups - " + groupsName + " (" + groups + ") not an array of numbers: element " + i + " = " + g);
                        }
                    }
                }
                else {
                    this._logger.exception(504, "changeGroups - groupsToRemove not an array: " + groups);
                }
            };
            return GamePeer;
        }(LbcPeer));
        LoadBalancing.GamePeer = GamePeer;
    })(LoadBalancing = Photon.LoadBalancing || (Photon.LoadBalancing = {}));
})(Photon || (Photon = {}));
/**
    Photon Load Balancing API Constants
    @namespace Photon.LoadBalancing.Constants
*/
var Photon;
(function (Photon) {
    var Lite;
    (function (Lite) {
        var Constants;
        (function (Constants) {
            // Summary:
            //     Lite - keys for parameters of operation requests and responses (short: OpKey).
            //
            // Remarks:
            //     These keys match a definition in the Lite application (part of the server
            //     SDK).  If your game is built as extension of Lite, don't re-use these codes
            //     for your custom events.  These keys are defined per application, so Lite
            //     has different keys than MMO or your custom application. This is why these
            //     are not an enumeration.  Lite and Lite Lobby will use the keys 255 and lower,
            //     to give you room for your own codes.  Keys for operation-parameters could
            //     be assigned on a per operation basis, but it makes sense to have fixed keys
            //     for values which are used throughout the whole application.
            Constants.LiteOpKey = {
                // Summary:
                //     (252) Code for list of players in a room. Currently not used.
                ActorList: 252,
                //
                // Summary:
                //     (254) Code of the Actor of an operation. Used for property get and set.
                ActorNr: 254,
                //
                // Summary:
                //     (249) Code for property set (Hashtable).
                ActorProperties: 249,
                //
                // Summary:
                //     (238) The "Add" operation-parameter can be used to add something to some
                //     list or set. E.g. add groups to player's interest groups.
                Add: 238,
                //
                // Summary:
                //     (250) Code for broadcast parameter of OpSetProperties method.
                Broadcast: 250,
                //
                // Summary:
                //     (247) Code for caching events while raising them.
                Cache: 247,
                //
                // Summary:
                //     (244) Code used when sending some code-related parameter, like OpRaiseEvent's
                //     event-code.
                //
                // Remarks:
                //     This is not the same as the Operation's code, which is no longer sent as
                //     part of the parameter Dictionary in Photon 3.
                Code: 244,
                //
                // Summary:
                //     (245) Code of data of an event. Used in OpRaiseEvent.
                Data: 245,
                //
                // Summary:
                //     (255) Code of the game id (a unique room name). Used in OpJoin.
                GameId: 255,
                //
                // Summary:
                //     (248) Code for property set (Hashtable).
                GameProperties: 248,
                //
                // Summary:
                //     (240) Code for "group" operation-parameter (as used in Op RaiseEvent).
                Group: 240,
                //
                // Summary:
                //     (251) Code for property set (Hashtable). This key is used when sending only
                //     one set of properties.  If either ActorProperties or GameProperties are used
                //     (or both), check those keys.
                Properties: 251,
                //
                // Summary:
                //     (246) Code to select the receivers of events (used in Lite, Operation RaiseEvent).
                ReceiverGroup: 246,
                //
                // Summary:
                //     (239) The "Remove" operation-parameter can be used to remove something from
                //     a list. E.g. remove groups from player's interest groups.
                Remove: 239,
                //
                // Summary:
                //     (253) Code of the target Actor of an operation. Used for property set. Is
                //     0 for game.
                TargetActorNr: 253,
                //
                // Summary:
                //     (236) A parameter indicating how long a room instance should be keeped alive in the 
                //     room cache after all players left the room.
                /// <summary>
                EmptyRoomLiveTime: 236
            };
            // Summary:
            //     Lite - Event codes.  These codes are defined by the Lite application's logic
            //     on the server side.  Other application's won't necessarily use these.
            //
            // Remarks:
            //     If your game is built as extension of Lite, don't re-use these codes for
            //     your custom events.
            Constants.LiteEventCode = {
                // Summary:
                //     (255) Event Join: someone joined the game
                Join: 255,
                //
                // Summary:
                //     (254) Event Leave: someone left the game
                Leave: 254,
                //
                // Summary:
                //     (253) Event PropertiesChanged
                PropertiesChanged: 253
            };
            // Summary:
            //     Lite - Operation Codes.  This enumeration contains the codes that are given
            //     to the Lite Application's operations. Instead of sending "Join", this enables
            //     us to send the byte 255.
            //
            // Remarks:
            //     Other applications (the MMO demo or your own) could define other operations
            //     and other codes.  If your game is built as extension of Lite, don't re-use
            //     these codes for your custom events.
            Constants.LiteOpCode = {
                // Summary:
                //     (248) Operation code to change interest groups in Rooms (Lite application
                //     and extending ones).
                ChangeGroups: 248,
                //
                // Summary:
                //     (251) Operation code for OpGetProperties.
                GetProperties: 251,
                //
                // Summary:
                //     (255) Code for OpJoin, to get into a room.
                Join: 255,
                //
                // Summary:
                //     (254) Code for OpLeave, to get out of a room.
                Leave: 254,
                //
                // Summary:
                //     (253) Code for OpRaiseEvent (not same as eventCode).
                RaiseEvent: 253,
                //
                // Summary:
                //     (252) Code for OpSetProperties.
                SetProperties: 252
            };
        })(Constants = Lite.Constants || (Lite.Constants = {}));
    })(Lite = Photon.Lite || (Photon.Lite = {}));
})(Photon || (Photon = {}));
(function (Photon) {
    var LoadBalancing;
    (function (LoadBalancing) {
        var Constants;
        (function (Constants) {
            var LiteOpKey = Photon.Lite.Constants.LiteOpKey;
            var LiteOpCode = Photon.Lite.Constants.LiteOpCode;
            var LiteEventCode = Photon.Lite.Constants.LiteEventCode;
            /**
                @summary Master and Game servers error codes.
                @member Photon.LoadBalancing.Constants.ErrorCode
                @readonly
                @property {number} Ok No Error.
                @property {number} OperationNotAllowedInCurrentState Operation can't be executed yet.
                @property {number} InvalidOperationCode The operation you called is not implemented on the server (application) you connect to. Make sure you run the fitting applications.
                @property {number} InternalServerError Something went wrong in the server. Try to reproduce and contact Exit Games.
                @property {number} InvalidAuthentication Authentication failed. Possible cause: AppId is unknown to Photon (in cloud service).
                @property {number} GameIdAlreadyExists GameId (name) already in use (can't create another). Change name.
                @property {number} GameFull Game is full. This can when players took over while you joined the game.
                @property {number} GameClosed Game is closed and can't be joined. Join another game.
                @property {number} NoRandomMatchFound Random matchmaking only succeeds if a room exists thats neither closed nor full. Repeat in a few seconds or create a new room.
                @property {number} GameDoesNotExist Join can fail if the room (name) is not existing (anymore). This can happen when players leave while you join.
                @property {number} MaxCcuReached Authorization on the Photon Cloud failed becaus the concurrent users (CCU) limit of the app's subscription is reached.
                @property {number} InvalidRegion Authorization on the Photon Cloud failed because the app's subscription does not allow to use a particular region's server.
            */
            Constants.ErrorCode = {
                Ok: 0,
                // server - Photon low(er) level: <: 0
                /// <summary>
                /// (-3) Operation can't be executed yet (e.g. OpJoin can't be called before being authenticated, RaiseEvent cant be used before getting into a room).
                /// </summary>
                /// <remarks>
                /// Before you call any operations on the Cloud servers, the automated client workflow must complete its authorization.
                /// In PUN, wait until State is: JoinedLobby (with AutoJoinLobby : true) or ConnectedToMaster (AutoJoinLobby : false)
                /// </remarks>
                OperationNotAllowedInCurrentState: -3,
                /// <summary>(-2) The operation you called is not implemented on the server (application) you connect to. Make sure you run the fitting applications.</summary>
                InvalidOperationCode: -2,
                /// <summary>(-1) Something went wrong in the server. Try to reproduce and contact Exit Games.</summary>
                InternalServerError: -1,
                // server - PhotonNetwork: 0x7FFF and down
                // logic-level error codes start with short.max
                /// <summary>(32767) Authentication failed. Possible cause: AppId is unknown to Photon (in cloud service).</summary>
                InvalidAuthentication: 0x7FFF,
                /// <summary>(32766) GameId (name) already in use (can't create another). Change name.</summary>
                GameIdAlreadyExists: 0x7FFF - 1,
                /// <summary>(32765) Game is full. This can when players took over while you joined the game.</summary>
                GameFull: 0x7FFF - 2,
                /// <summary>(32764) Game is closed and can't be joined. Join another game.</summary>
                GameClosed: 0x7FFF - 3,
                // AlreadyMatched: 0x7FFF - 4,
                /// <summary>(32762) Not in use currently.</summary>
                // ServerFull: 0x7FFF - 5,
                /// <summary>(32761) Not in use currently.</summary>
                // UserBlocked: 0x7FFF - 6,
                /// <summary>(32760) Random matchmaking only succeeds if a room exists thats neither closed nor full. Repeat in a few seconds or create a new room.</summary>
                NoRandomMatchFound: 0x7FFF - 7,
                /// <summary>(32758) Join can fail if the room (name) is not existing (anymore). This can happen when players leave while you join.</summary>
                GameDoesNotExist: 0x7FFF - 9,
                /// <summary>(32757) Authorization on the Photon Cloud failed becaus the concurrent users (CCU) limit of the app's subscription is reached.</summary>
                /// <remarks>
                /// Unless you have a plan with "CCU Burst", clients might fail the authentication step during connect. 
                /// Affected client are unable to call operations. Please note that players who end a game and return 
                /// to the Master server will disconnect and re-connect, which means that they just played and are rejected 
                /// in the next minute / re-connect.
                /// This is a temporary measure. Once the CCU is below the limit, players will be able to connect an play again.
                /// 
                /// OpAuthorize is part of connection workflow but only on the Photon Cloud, this error can happen. 
                /// Self-hosted Photon servers with a CCU limited license won't let a client connect at all.
                /// </remarks>
                MaxCcuReached: 0x7FFF - 10,
                /// <summary>(32756) Authorization on the Photon Cloud failed because the app's subscription does not allow to use a particular region's server.</summary>
                /// <remarks>
                /// Some subscription plans for the Photon Cloud are region-bound. Servers of other regions can't be used then.
                /// Check your Master server address and compare it with your Photon Cloud Dashboard's info.
                /// 
                /// OpAuthorize is part of connection workflow but only on the Photon Cloud, this error can happen. 
                /// Self-hosted Photon servers with a CCU limited license won't let a client connect at all.
                /// </remarks>
                InvalidRegion: 0x7FFF - 11,
                /// <summary>
                /// (32755) Custom Authentication of the user failed due to setup reasons (see Cloud Dashboard) or the provided user data (like username or token). Check error message for details.
                /// </summary>
                CustomAuthenticationFailed: 0x7FFF - 12,
                /// <summary>(32753) The Authentication ticket expired. Usually, this is refreshed behind the scenes. Connect (and authorize) again.</summary>
                AuthenticationTicketExpired: 0x7FF1,
                /// <summary>
                /// (32752) A server-side plugin (or webhook) failed to execute and reported an error. Check the OperationResponse.DebugMessage.
                /// </summary>
                PluginReportedError: 0x7FFF - 15,
                /// <summary>
                /// (32751) CreateGame/JoinGame/Join operation fails if expected plugin does not correspond to loaded one.
                /// </summary>
                PluginMismatch: 0x7FFF - 16,
                /// <summary>
                /// (32750) for join requests. Indicates the current peer already called join and is joined to the room.
                /// </summary>
                JoinFailedPeerAlreadyJoined: 32750,
                /// <summary>
                /// (32749)  for join requests. Indicates the list of InactiveActors already contains an actor with the requested ActorNr or UserId.
                /// </summary>
                JoinFailedFoundInactiveJoiner: 32749,
                /// <summary>
                /// (32748) for join requests. Indicates the list of Actors (active and inactive) did not contain an actor with the requested ActorNr or UserId.
                /// </summary>
                JoinFailedWithRejoinerNotFound: 32748,
                /// <summary>
                /// (32747) for join requests. Note: for future use - Indicates the requested UserId was found in the ExcludedList.
                /// </summary>
                JoinFailedFoundExcludedUserId: 32747,
                /// <summary>
                /// (32746) for join requests. Indicates the list of ActiveActors already contains an actor with the requested ActorNr or UserId.
                /// </summary>
                JoinFailedFoundActiveJoiner: 32746,
                /// <summary>
                /// (32745)  for SetProerties and Raisevent (if flag HttpForward is true) requests. Indicates the maximum allowed http requests per minute was reached.
                /// </summary>
                HttpLimitReached: 32745,
                /// <summary>
                /// (32744) for WebRpc requests. Indicates the the call to the external service failed.
                /// </summary>
                ExternalHttpCallFailed: 32744,
                /// <summary>
                /// (32742) Server error during matchmaking with slot reservation. E.g. the reserved slots can not exceed MaxPlayers.
                /// </summary>
                SlotError: 32742,
                /// <summary>
                /// (32741) Server will react with this error if invalid encryption parameters provided by token
                /// </summary>
                InvalidEncryptionParameters: 32741
            };
            /// <summary>
            /// These  values define "well known" properties for an Actor / Player.
            /// </summary>
            /// <remarks>
            /// "Custom properties" have to use a string-type as key. They can be assigned at will.
            /// </remarks>
            Constants.ActorProperties = {
                /// <summary>(255) Name of a player/actor.</summary>
                PlayerName: 255
            };
            /** End user doesn't need this */
            /// <summary>
            /// These  values are for "well known" room/game properties used in Photon Loadbalancing.
            /// </summary>
            /// <remarks>
            /// "Custom properties" have to use a string-type as key. They can be assigned at will.
            /// </remarks>
            Constants.GameProperties = {
                /// <summary>(255) Max number of players that "fit" into this room. 0 is for "unlimited".</summary>
                MaxPlayers: 255,
                /// <summary>(254) Makes this room listed or not in the lobby on Master.</summary>
                IsVisible: 254,
                /// <summary>(253) Allows more players to join a room (or not).</summary>
                IsOpen: 253,
                /// <summary>(252) Current count od players in the room. Used only in the lobby on Master.</summary>
                PlayerCount: 252,
                /// <summary>(251) True if the room is to be removed from room listing (used in update to room list in lobby on Master)</summary>
                Removed: 251,
                /// <summary>(250) A list of the room properties to pass to the RoomInfo list in a lobby. This is used in CreateRoom, which defines this list once per room.</summary>
                PropsListedInLobby: 250,
                /// <summary>Equivalent of Operation Join parameter CleanupCacheOnLeave.</summary>
                CleanupCacheOnLeave: 249,
                /// <summary>(248) Code for MasterClientId, which is synced by server. When sent as op-parameter this is (byte)203. As room property this is (byte)248.</summary>
                /// <remarks>Tightly related to ParameterCode.MasterClientId.</remarks>
                MasterClientId: 248,
                /// <summary>(246) Player Time To Live. How long any player can be inactive (due to disconnect or leave) before the user gets removed from the playerlist (freeing a slot).</summary>
                PlayerTtl: 246,
                /// <summary>(245) Room Time To Live. How long a room stays available (and in server-memory), after the last player becomes inactive. After this time, the room gets persisted or destroyed.</summary>
                EmptyRoomTtl: 245
            };
            /** End user doesn't need this */
            /// <summary>
            /// These values are for events defined by Photon Loadbalancing.
            /// </summary>
            /// <remarks>They start at 255 and go DOWN. Your own in-game events can start at 0.</remarks>
            Constants.EventCode = {
                /// <summary>(230) Initial list of RoomInfos (in lobby on Master)</summary>
                GameList: 230,
                /// <summary>(229) Update of RoomInfos to be merged into "initial" list (in lobby on Master)</summary>
                GameListUpdate: 229,
                /// <summary>(228) Currently not used. State of queueing in case of server-full</summary>
                QueueState: 228,
                /// <summary>(227) Currently not used. Event for matchmaking</summary>
                // Match: 227,
                /// <summary>(226) Event with stats about this application (players, rooms, etc)</summary>
                AppStats: 226,
                /// <summary>(210) Internally used in case of hosting by Azure</summary>
                AzureNodeInfo: 210,
                /// <summary>(255) Event Join: someone joined the game. The new actorNumber is provided as well as the properties of that actor (if set in OpJoin).</summary>
                Join: LiteEventCode.Join,
                /// <summary>(254) Event Leave: The player who left the game can be identified by the actorNumber.</summary>
                Leave: LiteEventCode.Leave,
                /// <summary>(253) When you call OpSetProperties with the broadcast option "on", this event is fired. It contains the properties being set.</summary>
                PropertiesChanged: LiteEventCode.PropertiesChanged,
                /// <summary>(252) When player left game unexpecable and playerTtl > 0 this event is fired</summary>
                Disconnect: 252,
                LobbyStats: 224
            };
            /** End user doesn't need this */
            /// <summary>Codes for parameters of Operations and Events.</summary>
            Constants.ParameterCode = {
                /// <summary>(230) Address of a (Game) server to use.</summary>
                Address: 230,
                /// <summary>(229) Count of players in this application in a rooms (used in stats event)</summary>
                PeerCount: 229,
                /// <summary>(228) Count of games in this application (used in stats event)</summary>
                GameCount: 228,
                /// <summary>(227) Count of players on the Master server (in this app, looking for rooms)</summary>
                MasterPeerCount: 227,
                /// <summary>(225) User's ID</summary>
                UserId: 225,
                /// <summary>(224) Your application's ID: a name on your own Photon or a GUID on the Photon Cloud</summary>
                ApplicationId: 224,
                /// <summary>(223) Not used currently (as "Position"). If you get queued before connect, this is your position</summary>
                Position: 223,
                /// <summary>(223) Modifies the matchmaking algorithm used for OpJoinRandom. Allowed parameter values are defined in enum MatchmakingMode.</summary>
                MatchMakingType: 223,
                /// <summary>(222) List of RoomInfos about open / listed rooms</summary>
                GameList: 222,
                /// <summary>(221) Internally used to establish encryption</summary>
                Secret: 221,
                /// <summary>(220) Version of your application</summary>
                AppVersion: 220,
                /// <summary>(210) Internally used in case of hosting by Azure</summary>
                AzureNodeInfo: 210,
                /// <summary>(209) Internally used in case of hosting by Azure</summary>
                AzureLocalNodeId: 209,
                /// <summary>(208) Internally used in case of hosting by Azure</summary>
                AzureMasterNodeId: 208,
                /// <summary>(255) Code for the gameId/roomName (a unique name per room). Used in OpJoin and similar.</summary>
                RoomName: LiteOpKey.GameId,
                /// <summary>(250) Code for broadcast parameter of OpSetProperties method.</summary>
                Broadcast: LiteOpKey.Broadcast,
                /// <summary>(252) Code for list of players in a room. Currently not used.</summary>
                ActorList: LiteOpKey.ActorList,
                /// <summary>(254) Code of the Actor of an operation. Used for property get and set.</summary>
                ActorNr: LiteOpKey.ActorNr,
                /// <summary>(249) Code for property set (Hashtable).</summary>
                PlayerProperties: LiteOpKey.ActorProperties,
                /// <summary>(245) Code of data/custom content of an event. Used in OpRaiseEvent.</summary>
                CustomEventContent: LiteOpKey.Data,
                /// <summary>(245) Code of data of an event. Used in OpRaiseEvent.</summary>
                Data: LiteOpKey.Data,
                /// <summary>(244) Code used when sending some code-related parameter, like OpRaiseEvent's event-code.</summary>
                /// <remarks>This is not the same as the Operation's code, which is no longer sent as part of the parameter Dictionary in Photon 3.</remarks>
                Code: LiteOpKey.Code,
                /// <summary>(248) Code for property set (Hashtable).</summary>
                GameProperties: LiteOpKey.GameProperties,
                /// <summary>
                /// (251) Code for property-set (Hashtable). This key is used when sending only one set of properties.
                /// If either ActorProperties or GameProperties are used (or both), check those keys.
                /// </summary>
                Properties: LiteOpKey.Properties,
                /// <summary>(253) Code of the target Actor of an operation. Used for property set. Is 0 for game</summary>
                TargetActorNr: LiteOpKey.TargetActorNr,
                /// <summary>(246) Code to select the receivers of events (used in Lite, Operation RaiseEvent).</summary>
                ReceiverGroup: LiteOpKey.ReceiverGroup,
                /// <summary>(247) Code for caching events while raising them.</summary>
                Cache: LiteOpKey.Cache,
                /// <summary>(241) Boolean parameter of CreateGame Operation. If true, server cleans up roomcache of leaving players (their cached events get removed).</summary>
                CleanupCacheOnLeave: 241,
                /// <summary>(240) Code for "group" operation-parameter (as used in Op RaiseEvent).</summary>
                Group: LiteOpKey.Group,
                /// <summary>(239) The "Remove" operation-parameter can be used to remove something from a list. E.g. remove groups from player's interest groups.</summary>
                Remove: LiteOpKey.Remove,
                /// <summary>(238) The "Add" operation-parameter can be used to add something to some list or set. E.g. add groups to player's interest groups.</summary>
                Add: LiteOpKey.Add,
                /// <summary>(236) A parameter indicating how long a room instance should be keeped alive in the room cache after all players left the room.</summary>
                EmptyRoomTTL: LiteOpKey.EmptyRoomLiveTime,
                PlayerTTL: 235,
                Plugins: 204,
                /// <summary>(217) This key's (byte) value defines the target custom authentication type/service the client connects with. Used in OpAuthenticate.</summary>
                ClientAuthenticationType: 217,
                /// <summary>(216) This key's (string) value provides parameters sent to the custom authentication type/service the client connects with. Used in OpAuthenticate.</summary>
                ClientAuthenticationParams: 216,
                ClientAuthenticationData: 214,
                /// <summary>(215) The JoinMode enum defines which variant of joining a room will be executed: Join only if available, create if not exists or re -join.</summary >
                /// <remarks>Replaces CreateIfNotExists which was only a bool -value.</remarks >
                JoinMode: 215,
                /// <summary>(203) Code for MasterClientId, which is synced by server. When sent as op-parameter this is code 203.</summary>
                /// <remarks>Tightly related to GamePropertyKey.MasterClientId.</remarks>
                MasterClientId: 203,
                /// <summary>(1) Used in Op FindFriends request. Value must be string[] of friends to look up.</summary>
                FindFriendsRequestList: 1,
                /// <summary>(1) Used in Op FindFriends response. Contains boolean[] list of online states (false if not online).</summary>
                FindFriendsResponseOnlineList: 1,
                /// <summary>(2) Used in Op FindFriends response. Contains string[] of room names ("" where not known or no room joined).</summary>
                FindFriendsResponseRoomIdList: 2,
                /// <summary>(213) Used in matchmaking-related methods and when creating a room to name a lobby (to join or to attach a room to).</summary>
                LobbyName: 213,
                /// <summary>(212) Used in matchmaking-related methods and when creating a room to define the type of a lobby. Combined with the lobby name this identifies the lobby.</summary>
                LobbyType: 212,
                LobbyStats: 211,
                /// <summary>(210) Used for region values in OpAuth and OpGetRegions.</summary >
                Region: 210,
                IsInactive: 233,
                CheckUserOnJoin: 232,
                /// <summary>(231) Code for "Check And Swap" (CAS) when changing properties.</summary>
                ExpectedValues: 231,
                UriPath: 209,
                RpcCallParams: 208,
                RpcCallRetCode: 207,
                RpcCallRetMessage: 206,
                RpcCallRetData: 208,
                WebFlags: 234,
                // Used by the server in Operation Responses, when it sends the nickname of the client (the user's nickname).
                Nickname: 202
            };
            /**
                @summary Codes for parameters and events used in Photon Load Balancing API.
                @member Photon.LoadBalancing.Constants.OperationCode
                @readonly
                @property {number} Authenticate Authenticates this peer and connects to a virtual application.
                @property {number} JoinLobby Joins lobby (on Master).
                @property {number} LeaveLobby Leaves lobby (on Master).
                @property {number} CreateGame Creates a game (or fails if name exists).
                @property {number} JoinGame Joins room (by name).
                @property {number} JoinRandomGame Joins random room (on Master).
                @property {number} Leave Leaves the room.
                @property {number} RaiseEvent Raises event (in a room, for other actors/players).
                @property {number} SetProperties Sets Properties (of room or actor/player).
                @property {number} GetProperties Gets Properties.
                @property {number} ChangeGroups Changes interest groups in room.
                @property {number} FindFriends Requests Master server for actors online status and joined rooms.
                @property {number} LobbyStats Requests Master server for lobbies statistics.
            */
            Constants.OperationCode = {
                /// <summary>(230) Authenticates this peer and connects to a virtual application</summary>
                Authenticate: 230,
                /// <summary>(229) Joins lobby (on Master)</summary>
                JoinLobby: 229,
                /// <summary>(228) Leaves lobby (on Master)</summary>
                LeaveLobby: 228,
                /// <summary>(227) Creates a game (or fails if name exists)</summary>
                CreateGame: 227,
                /// <summary>(226) Join game (by name)</summary>
                JoinGame: 226,
                /// <summary>(225) Joins random game (on Master)</summary>
                JoinRandomGame: 225,
                // CancelJoinRandom : 224, // obsolete, cause JoinRandom no longer is a "process". now provides result immediately
                /// <summary>(254) Code for OpLeave, to get out of a room.</summary>
                Leave: LiteOpCode.Leave,
                /// <summary>(253) Raise event (in a room, for other actors/players)</summary>
                RaiseEvent: LiteOpCode.RaiseEvent,
                /// <summary>(252) Set Properties (of room or actor/player)</summary>
                SetProperties: LiteOpCode.SetProperties,
                /// <summary>(251) Get Properties</summary>
                GetProperties: LiteOpCode.GetProperties,
                /// <summary>(248) Operation code to change interest groups in Rooms (Lite application and extending ones).</summary>
                ChangeGroups: LiteOpCode.ChangeGroups,
                /// <summary>(222) Request the rooms and online status for a list of friends (by name, which should be unique).</summary>
                FindFriends: 222,
                LobbyStats: 221,
                /// <summary>(220) Gets list of regional servers from a NameServer.</summary>
                GetRegions: 220,
                /// <summary>(219) Rpc Operation.</summary>
                Rpc: 219
            };
            /**
                @summary Options for matchmaking rules for joinRandomGame.
                @member Photon.LoadBalancing.Constants.MatchmakingMode
                @readonly
                @property {number} FillRoom Default. FillRoom Fills up rooms (oldest first) to get players together as fast as possible. Makes most sense with MaxPlayers > 0 and games that can only start with more players.
                @property {number} SerialMatching Distributes players across available rooms sequentially but takes filter into account. Without filter, rooms get players evenly distributed.
                @property {number} RandomMatching Joins a (fully) random room. Expected properties must match but aside from this, any available room might be selected.
            */
            Constants.MatchmakingMode = {
                /// <summary>Fills up rooms (oldest first) to get players together as fast as possible. Default.</summary>
                /// <remarks>Makes most sense with MaxPlayers > 0 and games that can only start with more players.</remarks>
                FillRoom: 0,
                /// <summary>Distributes players across available rooms sequentially but takes filter into account. Without filter, rooms get players evenly distributed.</summary>
                SerialMatching: 1,
                /// <summary>Joins a (fully) random room. Expected properties must match but aside from this, any available room might be selected.</summary>
                RandomMatching: 2
            };
            /**
                @summary Caching options for events.
                @member Photon.LoadBalancing.Constants.EventCaching
                @readonly
                @property {number} DoNotCache Default. Do not cache.
                @property {number} MergeCache Will merge this event's keys with those already cached.
                @property {number} ReplaceCache Replaces the event cache for this eventCode with this event's content.
                @property {number} RemoveCache Removes this event (by eventCode) from the cache.
                @property {number} AddToRoomCache Adds an event to the room's cache.
                @property {number} AddToRoomCacheGlobal Adds this event to the cache for actor 0 (becoming a "globally owned" event in the cache).
                @property {number} RemoveFromRoomCache Remove fitting event from the room's cache.
                @property {number} RemoveFromRoomCacheForActorsLeft Removes events of players who already left the room (cleaning up).
            */
            Constants.EventCaching = {
                // Summary:
                //     Default value (not sent).
                DoNotCache: 0,
                //
                // Summary:
                //     Will merge this event's keys with those already cached.
                MergeCache: 1,
                //
                // Summary:
                //     Replaces the event cache for this eventCode with this event's content.
                ReplaceCache: 2,
                //
                // Summary:
                //     Removes this event (by eventCode) from the cache.
                RemoveCache: 3,
                //
                // Summary:
                //     Adds an event to the room's cache.
                AddToRoomCache: 4,
                //
                // Summary:
                //     Adds this event to the cache for actor 0 (becoming a "globally owned" event
                //     in the cache).
                AddToRoomCacheGlobal: 5,
                //
                // Summary:
                //     Remove fitting event from the room's cache.
                RemoveFromRoomCache: 6,
                //
                // Summary:
                //     Removes events of players who already left the room (cleaning up).
                RemoveFromRoomCacheForActorsLeft: 7
            };
            /**
                @summary Options for choosing room's actors who should receive events.
                @member Photon.LoadBalancing.Constants.ReceiverGroup
                @readonly
                @property {number} Others Default. Anyone else gets my event.
                @property {number} All Everyone in the current room (including this peer) will get this event.
                @property {number} MasterClient The "master client" does not have special rights but is the one who is in this room the longest time.
            */
            Constants.ReceiverGroup = {
                // Summary:
                //     Default value (not sent). Anyone else gets my event.
                Others: 0,
                //
                // Summary:
                //     Everyone in the current room (including this peer) will get this event.
                All: 1,
                //
                // Summary:
                //     The server sends this event only to the actor with the lowest actorNumber.
                //
                // Remarks:
                //     The "master client" does not have special rights but is the one who is in
                //     this room the longest time.
                MasterClient: 2
            };
            /**
                @summary Options for optional "Custom Authentication" services used with Photon.
                @member Photon.LoadBalancing.Constants.CustomAuthenticationType
                @readonly
                @property {number} Custom Default. Use a custom authentification service.
                @property {number} Steam Authenticates users by their Steam Account. Set auth values accordingly.
                @property {number} Facebook Authenticates users by their Facebook Account. Set auth values accordingly.
                @property {number} None Disables custom authentification.
            */
            Constants.CustomAuthenticationType = {
                Custom: 0,
                Steam: 1,
                Facebook: 2,
                None: 255
            };
            /**
                @summary Options of lobby types available. Lobby types might be implemented in certain Photon versions and won't be available on older servers.
                @member Photon.LoadBalancing.Constants.LobbyType
                @readonly
                @property {number} Default This lobby is used unless another is defined by game or JoinRandom. Room-lists will be sent and JoinRandomRoom can filter by matching properties.
                @property {number} SqlLobby This lobby type lists rooms like Default but JoinRandom has a parameter for SQL-like "where" clauses for filtering. This allows bigger, less, or and and combinations.
            **/
            Constants.LobbyType = {
                Default: 0,
                SqlLobby: 2
            };
        })(Constants = LoadBalancing.Constants || (LoadBalancing.Constants = {}));
    })(LoadBalancing = Photon.LoadBalancing || (Photon.LoadBalancing = {}));
})(Photon || (Photon = {}));
/// <reference path="photon-common.ts"/>
/**
    Photon Chat API Constants
    @namespace Photon.Chat.Constants
*/
var Photon;
(function (Photon) {
    var Chat;
    (function (Chat) {
        var Constants;
        (function (Constants) {
            Constants.ParameterCode = {
                Channels: 0,
                Channel: 1,
                Messages: 2,
                Message: 3,
                Senders: 4,
                Sender: 5,
                ChannelUserCount: 6,
                UserId: 225,
                MsgId: 8,
                MsgIds: 9,
                SubscribeResults: 15,
                Status: 10,
                Friends: 11,
                SkipMessage: 12,
                HistoryLength: 14,
                WebFlags: 21
            };
            //- Codes for parameters and events used in Photon Chat API.
            Constants.OperationCode = {
                Subscribe: 0,
                Unsubscribe: 1,
                Publish: 2,
                SendPrivate: 3,
                ChannelHistory: 4,
                UpdateStatus: 5,
                AddFriendds: 6,
                RemoveFriends: 7 // Removes users from the list that should update you of their status.
            };
            Constants.EventCode = {
                ChatMessages: 0,
                Users: 1,
                PrivateMessage: 2,
                FriendsList: 3,
                StatusUpdate: 4,
                Subscribe: 5,
                Unsubscribe: 6
            };
            /**
                @summary Contains commonly used status values for {@link Photon.Chat.ChatClient#setUserStatus}.You can define your own.<br/>
                While "online"(Online and up), the status message will be sent to anyone who has you on his friend list.<br/>
                Define custom online status values as you like with these rules:<br/>
                0: Means "offline".It will be used when you are not connected. In this status, there is no status message.<br/>
                1: Means "invisible" and is sent to friends as "offline". They see status 0, no message but you can chat.<br/>
                2: And any higher value will be treated as "online". Status can be set.<br/>
                @readonly
                @property {number} Offline Offline.
                @property {number} Invisible Offline. Be invisible to everyone. Sends no message.
                @property {number} Online Online and available.
                @property {number} Away Online but not available.
                @property {number} Dnd Do not disturb.
                @property {number} Lfg Looking For Game / Group. Could be used when you want to be invited or do matchmaking.
                @property {number} Playing Could be used when in a room, playing.
                @member Photon.Chat.Constants.UserStatus
            */
            Constants.UserStatus = {
                Offline: 0,
                Invisible: 1,
                Online: 2,
                Away: 3,
                Dnd: 4,
                Lfg: 5,
                Playing: 6
            };
            /**
                @summary Converts {@link Photon.Chat.Constants.UserStatus} element to string name.
                @param {Photon.Chat.Constants.UserStatus} status User status enum element.
                @returns {string} Specified element name or undefined if not found.
                @method Photon.Chat.Constants.UserStatusToName
            */
            function UserStatusToName(status) { return Exitgames.Common.Util.getEnumKeyByValue(Constants.UserStatus, status); }
            Constants.UserStatusToName = UserStatusToName;
        })(Constants = Chat.Constants || (Chat.Constants = {}));
    })(Chat = Photon.Chat || (Photon.Chat = {}));
})(Photon || (Photon = {}));
/// <reference path="photon-loadbalancing.ts"/>
/// <reference path="photon-chat-constants.ts"/>
/**
    Photon Chat API
    @namespace Photon.Chat
*/
var Photon;
(function (Photon) {
    var Chat;
    (function (Chat) {
        var WebFlags = {
            HttpForward: 0x01,
            SendAuthCookie: 0x02,
            SendSync: 0x04,
            SendState: 0x08
        };
        /**
            @class Photon.Chat.Message
            @classdesc Encapsulates chat message data.
        */
        var Message = /** @class */ (function () {
            function Message(sender, content) {
                this.sender = sender;
                this.content = content;
            }
            /**
                @summary Returns message sender.
                @return {string} Message sender.
                @method Photon.Chat.Message#getSender
            */
            Message.prototype.getSender = function () {
                return this.sender;
            };
            /**
                @summary Returns message content.
                @return {any} Message content.
                @method Photon.Chat.Message#getContent
            */
            Message.prototype.getContent = function () {
                return this.content;
            };
            return Message;
        }());
        Chat.Message = Message;
        /**
            @class Photon.Chat.Channel
            @classdesc Represents chat channel.
        */
        var Channel = /** @class */ (function () {
            function Channel(name, isPrivat) {
                this.name = name;
                this.isPrivat = isPrivat;
                this.messages = [];
            }
            /**
                @summary Returns channel name (counterpart user id for private channel).
                @return {string} Channel name.
                @method Photon.Chat.Channel#getName
            */
            Channel.prototype.getName = function () {
                return this.name;
            };
            /**
                @summary Returns true if channel is private.
                @return {boolean} Channel private status.
                @method Photon.Chat.Channel#isPrivate
            */
            Channel.prototype.isPrivate = function () {
                return this.isPrivat;
            };
            /**
                @summary Returns messages cache.
                @return {{@link Photon.Chat.Message}[]} Array of messages.
                @method Photon.Chat.Channel#getMessages
            */
            Channel.prototype.getMessages = function () {
                return this.messages;
            };
            /**
                @summary Returns ID of the last message received.
                @return {number} Last message ID.
                @method Photon.Chat.Channel#getLastId
            */
            Channel.prototype.getLastId = function () {
                return this.lastId;
            };
            /**
                @summary Clears messages cache.
                @method Photon.Chat.Channel#clearMessages
            */
            Channel.prototype.clearMessages = function () {
                this.messages.splice(0);
            };
            // internal
            Channel.prototype.addMessages = function (senders, messages) {
                var newMessages = [];
                for (var i in senders) {
                    if (parseInt(i) < messages.length) {
                        var m = new Message(senders[i], messages[i]);
                        this.messages.push(m);
                        newMessages.push(m);
                    }
                }
                return newMessages;
            };
            return Channel;
        }());
        Chat.Channel = Channel;
        var ChatClient = /** @class */ (function (_super) {
            __extends(ChatClient, _super);
            /**
                @classdesc Implements the Photon Chat API workflow.<br/>
                This class should be extended to handle system or custom events and operation responses.<br/>
                
                @borrows Photon.LoadBalancing.LoadBalancingClient#setCustomAuthentication
                @borrows Photon.LoadBalancing.LoadBalancingClient#connectToNameServer
                @borrows Photon.LoadBalancing.LoadBalancingClient#getRegions
                @borrows Photon.LoadBalancing.LoadBalancingClient#onGetRegionsResult
                @borrows Photon.LoadBalancing.LoadBalancingClient#isConnectedToNameServer
                @borrows Photon.LoadBalancing.LoadBalancingClient#disconnect
                @borrows Photon.LoadBalancing.LoadBalancingClient#setLogLevel
    
                @constructor Photon.Chat.ChatClient
                @param {Photon.ConnectionProtocol} protocol Connecton protocol.
                @param {string} appId Cloud application ID.
                @param {string} appVersion Cloud application version.
            */
            function ChatClient(protocol, appId, appVersion) {
                var _this = _super.call(this, protocol, appId, appVersion) || this;
                _this.publicChannels = {};
                _this.privateChannels = {};
                _this.subscribeRequests = [];
                _this.unsubscribeRequests = [];
                _this.autoJoinLobby = false;
                return _this;
            }
            /**
                @summary Called on client state change. Override to handle it.
                @method Photon.Chat.ChatClient#onStateChange
                @param {Photon.Chat.ChatClient.ChatState} state New client state.
            */
            ChatClient.prototype.onStateChange = function (state) { };
            /**
                @summary Called if client error occures. Override to handle it.
                @method Chat.ChatClient#onError
                @param {Chat.ChatClient.ChatPeerErrorCode} errorCode Client error code.
                @param {string} errorMsg Error message.
            */
            ChatClient.prototype.onError = function (errorCode, errorMsg) { };
            /**
                @summary Called when {@link Photon.Chat.ChatClient#subscribe subscribe} request completed.<br/ >
                Override to handle request results.
                @param {object} results Object with channel names as keys and boolean results as values.
                @method Photon.Chat.ChatClient#onSubscribeResult
            */
            ChatClient.prototype.onSubscribeResult = function (results) { };
            /**
                @summary Called when {@link Photon.Chat.ChatClient#unsubscribe unsubscribe} request completed.<br/ >
                Override to handle request results.
                @param {object} results Object with channel names as keys and boolean results as values.
                @method Photon.Chat.ChatClient#onUnsubscribeResult
            */
            ChatClient.prototype.onUnsubscribeResult = function (results) { };
            /**
                @summary Called when new chat messages received.<br/ >
                Override to handle messages receive event.
                @param {string} channelName Chat channel name.
                @param {{@link Photon.Chat.Message}[]} messages Array of received messages.
                @method Photon.Chat.ChatClient#onChatMessages
            */
            ChatClient.prototype.onChatMessages = function (channelName, messages) { };
            /**
                @summary Called when new private message received.<br/ >
                Override to handle message receive event.
                @param {string} channelName Private channel name(counterpart user id).
                @param {Photon.Chat.Message} message Received message.
                @method Photon.Chat.ChatClient#onPrivateMessage
            */
            ChatClient.prototype.onPrivateMessage = function (channelName, message) { };
            /**
                @summary Called when user from friend list changes state.<br/ >
                Override to handle change state event.
                @param {string} userId User id.
                @param {number} status New User status. Predefined {@link Photon.chat.Constants.UserStatus Constants.UserStatus} or custom.
                @param {boolean} gotMessage True if status message updated.
                @param {string} statusMessage Optional status message (may be null even if gotMessage = true).
                @method Photon.Chat.ChatClient#onUserStatusUpdate
            */
            ChatClient.prototype.onUserStatusUpdate = function (userId, status, gotMessage, statusMessage) { };
            /**
                @summary Connects to a specific region's Master server, using the NameServer to find the IP.
                Override {@link Photon.Chat.ChatClient#onWebRpcResult onWebRpcResult} to handle request results.<br/>
                @method Photon.Chat.ChatClient#connectToRegionFrontEnd
                @param {string} region Region connect to Master server of.
                @returns {boolean} True if current client state allows connection.
            **/
            ChatClient.prototype.connectToRegionFrontEnd = function (region) {
                return this.connectToRegionMaster(region);
            };
            /**
                @summary Returns true if client connected to Front End.When connected, client can send messages, subscribe to channels and so on.
                @return {boolean} True if connected.
                @method Photon.Chat.ChatClient#isConnectedToFrontEnd
            */
            ChatClient.prototype.isConnectedToFrontEnd = function () {
                return this.state == ChatClient.ChatState.ConnectedToFrontEnd;
            };
            /**
                @summary Sends operation to subscribe to a list of channels by name.<br/>
                Override {@link Photon.Chat.ChatClient#onSubscribeResult onSubscribeResult} to handle request results.
                @param {string[]} channelNames Array of channel names to subscribe to.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {number} [options.historyLength] Controls messages history sent on subscription. Not specified or 0: no history. 1 and higher: number of messages in history. -1: all history.
                @property {number[]} [options.lastIds] Array of IDs of last messages received per channel. Useful when resubscribing to receive only messages we missed.
                @return {boolean} True if operation sent.
                @method Photon.Chat.ChatClient#subscribe
            */
            ChatClient.prototype.subscribe = function (channelNames, options) {
                // backward compatibility
                if (typeof (options) == "number") {
                    options = { historyLength: options };
                }
                if (this.isConnectedToFrontEnd()) {
                    this.logger.debug("Subscribe channels:", channelNames);
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Channels, Photon.TypeExt.String(channelNames));
                    if (options) {
                        if (options.historyLength) {
                            params.push(Chat.Constants.ParameterCode.HistoryLength, Photon.TypeExt.Int(options.historyLength));
                        }
                        if (options.lastIds) {
                            params.push(Chat.Constants.ParameterCode.MsgIds, Photon.TypeExt.Int(options.lastIds));
                            if (options.historyLength === undefined) {
                                params.push(Chat.Constants.ParameterCode.HistoryLength, Photon.TypeExt.Int(-1));
                            }
                        }
                    }
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.Subscribe, params);
                    return true;
                }
                else {
                    this.logger.error("subscribe request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Sends operation to unsubscribe from a list of channels by name.<br/ >
                Override {@link Photon.Chat.ChatClient#onUnsubscribeResult onUnsubscribeResult} to handle request results.
                @param {string[]} channelNames Array of channel names to unsubscribe from.
                @return {boolean} True if operation sent.
                @method Photon.Chat.ChatClient#unsubscribe
            */
            ChatClient.prototype.unsubscribe = function (channelNames) {
                if (this.isConnectedToFrontEnd()) {
                    this.logger.debug("Unsubscribe channels:", channelNames);
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Channels, Photon.TypeExt.String(channelNames));
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.Unsubscribe, params);
                    return true;
                }
                else {
                    this.logger.error("unsubscribe request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Sends a message to a public channel.<br/>
                Channel should be subscribed before publishing to it.
                Everyone in that channel will get the message.
                @param {string} channelName Channel name to send message to.
                @param {any} content Text string or arbitrary data to send.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {boolean} [options.webForward] Optionally, private messages can be forwarded as webhooks. Configure webhooks for your Chat app to use this.
                @return {boolean} True if message sent.
                @method Photon.Chat.ChatClient#publishMessage
            */
            ChatClient.prototype.publishMessage = function (channelName, content, options) {
                if (this.isConnectedToFrontEnd()) {
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Channel, channelName);
                    params.push(Chat.Constants.ParameterCode.Message, content);
                    if (options) {
                        if (options.webForward) {
                            params.push(Chat.Constants.ParameterCode.WebFlags);
                            params.push(Photon.TypeExt.Byte(WebFlags.HttpForward));
                        }
                    }
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.Publish, params);
                    return true;
                }
                else {
                    this.logger.error("publishMessage request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Sends a private message to a single target user.<br/>
                @param {string} userId User id to send this message to.
                @param {any} content Text string or arbitrary data to send.
                @param {object} [options] Additional options
                @property {object} options Additional options
                @property {boolean} [options.webForward] Optionally, private messages can be forwarded as webhooks. Configure webhooks for your Chat app to use this.
                @return {boolean} True if message sent.
                @method Photon.Chat.ChatClient#sendPrivateMessage
            */
            ChatClient.prototype.sendPrivateMessage = function (userId, content, options) {
                if (this.isConnectedToFrontEnd()) {
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.UserId, userId);
                    params.push(Chat.Constants.ParameterCode.Message, content);
                    if (options) {
                        if (options.webForward) {
                            params.push(Chat.Constants.ParameterCode.WebFlags);
                            params.push(Photon.TypeExt.Byte(WebFlags.HttpForward));
                        }
                    }
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.SendPrivate, params);
                    return true;
                }
                else {
                    this.logger.error("sendPrivateMessage request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Sets the user's status (pre-defined or custom) and an optional message.<br/>
                The predefined status values can be found in {@link Photon.Chat.Constants.UserStatus Constants.UserStatus}.<br/>
                State UserStatus.Invisible will make you offline for everyone and send no message.
                @param {number} status User status to set.
                @param {string} [message=null] State message.
                @param {boolean} [skipMessage=false] If true { client does not send state message.
                @return {boolean} True if command sent.
                @method Photon.Chat.ChatClient#setUserStatus
            */
            ChatClient.prototype.setUserStatus = function (status, statusMessage, skipMessage) {
                if (statusMessage === void 0) { statusMessage = null; }
                if (skipMessage === void 0) { skipMessage = false; }
                if (this.isConnectedToFrontEnd()) {
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Status, Photon.TypeExt.Int(status));
                    if (skipMessage)
                        params.push(Chat.Constants.ParameterCode.SkipMessage, true);
                    else
                        params.push(Chat.Constants.ParameterCode.Message, statusMessage);
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.UpdateStatus, params);
                    return true;
                }
                else {
                    this.logger.error("setUserStatus request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Adds users to the list on the Chat Server which will send you status updates for those.
                @tparam string[] userIds Array of user ids.
                @return {boolean} True if command sent.
            */
            ChatClient.prototype.addFriends = function (userIds) {
                if (this.isConnectedToFrontEnd()) {
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Friends, Photon.TypeExt.String(userIds));
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.AddFriendds, params);
                    return true;
                }
                else {
                    this.logger.error("addFriends request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Removes users from the list on the Chat Server which will send you status updates for those.
                @tparam string[] friends Array of user ids.
                @return {boolean} True if command sent.
            */
            ChatClient.prototype.removeFriends = function (userIds) {
                if (this.isConnectedToFrontEnd()) {
                    var params = [];
                    params.push(Chat.Constants.ParameterCode.Friends, Photon.TypeExt.String(userIds));
                    this.masterPeer.sendOperation(Chat.Constants.OperationCode.RemoveFriends, params);
                    return true;
                }
                else {
                    this.logger.error("removeFriends request error:", "Not connected to Front End");
                    return false;
                }
            };
            /**
                @summary Returns list of public channels client subscribed to.
                @return Channel[] Array of public channels.
            */
            ChatClient.prototype.getPublicChannels = function () {
                return this.publicChannels;
            };
            /**
                @summary Returns list of channels representing current private conversation.
                @return Channel[] Array of private channels.
            */
            ChatClient.prototype.getPrivateChannels = function () {
                return this.privateChannels;
            };
            // private
            ChatClient.prototype.getOrAddChannel = function (channels, name, isPrivate) {
                if (channels[name] == undefined) {
                    channels[name] = new Channel(name, isPrivate);
                }
                return channels[name];
            };
            // internal
            ChatClient.prototype.initMasterPeer = function (mp) {
                var _this = this;
                _super.prototype.initMasterPeer.call(this, mp);
                // onOperationResponse called if no listener exists
                //mp.addResponseListener(Constants.OperationCode.Publish, (data: any) => {
                //    mp._logger.debug("resp Publish", data.errCode, data.errMsg);
                //});
                //mp.addResponseListener(Constants.OperationCode.SendPrivate, (data: any) => {
                //    mp._logger.debug("resp SendPrivate", data.errCode, data.errMsg);
                //});
                //mp.addResponseListener(Constants.OperationCode.UpdateStatus, (data: any) => {
                //    mp._logger.debug("resp UpdateStatus", data.errCode, data.errMsg);
                //});
                //mp.addResponseListener(Constants.OperationCode.FriendList, (data: any) => {
                //    mp._logger.debug("resp FriendList", data.errCode, data.errMsg);
                //});
                mp.addEventListener(Chat.Constants.EventCode.ChatMessages, function (data) {
                    var senders = data.vals[Chat.Constants.ParameterCode.Senders];
                    var messages = data.vals[Chat.Constants.ParameterCode.Messages];
                    var channelName = data.vals[Chat.Constants.ParameterCode.Channel];
                    var ch = _this.publicChannels[channelName];
                    if (ch) {
                        var newMessages = ch.addMessages(senders, messages);
                        ch.lastId = data.vals[Chat.Constants.ParameterCode.MsgId];
                        _this.onChatMessages(channelName, newMessages);
                    }
                    else {
                        mp._logger.warn("ev ChatMessages: Got message from unsubscribed channel ", channelName);
                    }
                });
                mp.addEventListener(Chat.Constants.EventCode.PrivateMessage, function (data) {
                    var sender = data.vals[Chat.Constants.ParameterCode.Sender];
                    var message = data.vals[Chat.Constants.ParameterCode.Message];
                    var userId = data.vals[Chat.Constants.ParameterCode.UserId];
                    var channelName = "";
                    if (_this.getUserId() == sender)
                        channelName = userId;
                    else
                        channelName = sender;
                    var ch = _this.getOrAddChannel(_this.privateChannels, channelName, true);
                    ch.lastId = data.vals[Chat.Constants.ParameterCode.MsgId];
                    _this.onPrivateMessage(channelName, new Message(sender, message));
                });
                mp.addEventListener(Chat.Constants.EventCode.StatusUpdate, function (data) {
                    var sender = data.vals[Chat.Constants.ParameterCode.Sender];
                    var status = data.vals[Chat.Constants.ParameterCode.Status];
                    var message = data.vals[Chat.Constants.ParameterCode.Message];
                    var gotMessage = message !== undefined;
                    _this.onUserStatusUpdate(sender, status, gotMessage, message);
                });
                mp.addEventListener(Chat.Constants.EventCode.Subscribe, function (data) {
                    mp._logger.debug("ev Subscribe", data);
                    var res = {};
                    var channels = data.vals[Chat.Constants.ParameterCode.Channels] || [];
                    var results = data.vals[Chat.Constants.ParameterCode.SubscribeResults] || [];
                    for (var i in channels) {
                        var ch = channels[i];
                        res[ch] = false;
                        if (i < results.length && results[i]) {
                            _this.getOrAddChannel(_this.publicChannels, ch, false);
                            res[ch] = true;
                        }
                    }
                    _this.onSubscribeResult(res);
                });
                mp.addEventListener(Chat.Constants.EventCode.Unsubscribe, function (data) {
                    mp._logger.debug("ev Unsubscribe", data);
                    var res = {};
                    var channels = data.vals[Chat.Constants.ParameterCode.Channels] || [];
                    for (var i in channels) {
                        var ch = channels[i];
                        delete (_this.publicChannels[ch]);
                        res[ch] = true;
                    }
                    _this.onUnsubscribeResult(res);
                });
            };
            /**
                @summary Converts {@link Photon.Chat.ChatClient.ChatState ChatState} element to string name.
                @method Photon.Chat.ChatClient.StateToName
                @param {Photon.Chat.ChatClient.ChatState} state Client state.
                @returns {string} Specified element name or undefined if not found.
            */
            ChatClient.StateToName = function (value) {
                var x = Exitgames.Common.Util.getEnumKeyByValue(ChatClient.ChatState, value);
                if (x === undefined) {
                    // Super class states support - useless since all states overridden but may help somehow when debugging
                    return Exitgames.Common.Util.getEnumKeyByValue(ChatClient.State, value);
                }
                else {
                    return x;
                }
            };
            ChatClient.ChatPeerErrorCode = {
                /**
                    @summary Enum for client peers error codes.
                    @member Photon.Chat.ChatClient.ChatPeerErrorCode
                    @readonly
                    @property {number} Ok No Error.
                    @property {number} FrontEndError General FrontEnd server peer error.
                    @property {number} FrontEndConnectFailed FrontEnd server connection error.
                    @property {number} FrontEndConnectClosed Disconnected from FrontEnd server.
                    @property {number} FrontEndTimeout Disconnected from FrontEnd server for timeout.
                    @property {number} FrontEndEncryptionEstablishError FrontEnd server encryption establishing failed.
                    @property {number} FrontEndAuthenticationFailed FrontEnd server authentication failed.
                    @property {number} NameServerError General NameServer peer error.
                    @property {number} NameServerConnectFailed NameServer connection error.
                    @property {number} NameServerConnectClosed Disconnected from NameServer.
                    @property {number} NameServerTimeout Disconnected from NameServer for timeout.
                    @property {number} NameServerEncryptionEstablishError NameServer encryption establishing failed.
                    @property {number} NameServerAuthenticationFailed NameServer authentication failed.
                 */
                Ok: 0,
                FrontEndError: 1001,
                FrontEndConnectFailed: 1002,
                FrontEndConnectClosed: 1003,
                FrontEndTimeout: 1004,
                FrontEndEncryptionEstablishError: 1005,
                FrontEndAuthenticationFailed: 1101,
                NameServerError: 3001,
                NameServerConnectFailed: 3002,
                NameServerConnectClosed: 3003,
                NameServerTimeout: 3004,
                NameServerEncryptionEstablishError: 300,
                NameServerAuthenticationFailed: 3101
            };
            ChatClient.ChatState = {
                /**
                    @summary Enum for client states.
                    @member Photon.Chat.ChatClient.ChatState
                    @readonly
                    @property {number} Error Critical error occurred.
                    @property {number} Uninitialized Client is created but not used yet.
                    @property {number} ConnectingToNameServer Connecting to NameServer.
                    @property {number} ConnectedToNameServer Connected to NameServer.
                    @property {number} ConnectingToFrontEnd Connecting to FrontEnd server.
                    @property {number} ConnectedToFrontEnd Connected to FrontEnd server.
                    @property {number} Disconnected The client is no longer connected (to any server).
                */
                Error: -1,
                Uninitialized: 0,
                ConnectingToNameServer: 1,
                ConnectedToNameServer: 2,
                ConnectingToFrontEnd: 3,
                ConnectedToFrontEnd: 4,
                Disconnected: 10
            };
            return ChatClient;
        }(Photon.LoadBalancing.LoadBalancingClient));
        Chat.ChatClient = ChatClient;
    })(Chat = Photon.Chat || (Photon.Chat = {}));
})(Photon || (Photon = {}));
